const __vite__mapDeps=(i,m=__vite__mapDeps,d=(m.f||(m.f=["assets/index-CX2UoIQ2.js","assets/is-plan-event-enabled-DeNtQvA5.js","assets/index-BUG6bmcj.js"])))=>i.map(i=>d[i]);
function Ejn(n,r){for(var o=0;o<r.length;o++){const s=r[o];if(typeof s!="string"&&!Array.isArray(s)){for(const c in s)if(c!=="default"&&!(c in n)){const d=Object.getOwnPropertyDescriptor(s,c);d&&Object.defineProperty(n,c,d.get?d:{enumerable:!0,get:()=>s[c]})}}}return Object.freeze(Object.defineProperty(n,Symbol.toStringTag,{value:"Module"}))}(function(){const r=document.createElement("link").relList;if(r&&r.supports&&r.supports("modulepreload"))return;for(const c of document.querySelectorAll('link[rel="modulepreload"]'))s(c);new MutationObserver(c=>{for(const d of c)if(d.type==="childList")for(const h of d.addedNodes)h.tagName==="LINK"&&h.rel==="modulepreload"&&s(h)}).observe(document,{childList:!0,subtree:!0});function o(c){const d={};return c.integrity&&(d.integrity=c.integrity),c.referrerPolicy&&(d.referrerPolicy=c.referrerPolicy),c.crossOrigin==="use-credentials"?d.credentials="include":c.crossOrigin==="anonymous"?d.credentials="omit":d.credentials="same-origin",d}function s(c){if(c.ep)return;c.ep=!0;const d=o(c);fetch(c.href,d)}})();var pI=typeof globalThis<"u"?globalThis:typeof window<"u"?window:typeof global<"u"?global:typeof self<"u"?self:{};function Y4(n){return n&&n.__esModule&&Object.prototype.hasOwnProperty.call(n,"default")?n.default:n}var mje={exports:{}},Kne={};/**
 * @license React
 * react-jsx-runtime.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var ZMt;function Sjn(){if(ZMt)return Kne;ZMt=1;var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.fragment");function o(s,c,d){var h=null;if(d!==void 0&&(h=""+d),c.key!==void 0&&(h=""+c.key),"key"in c){d={};for(var b in c)b!=="key"&&(d[b]=c[b])}else d=c;return c=d.ref,{$$typeof:n,type:s,key:h,ref:c!==void 0?c:null,props:d}}return Kne.Fragment=r,Kne.jsx=o,Kne.jsxs=o,Kne}var QMt;function _jn(){return QMt||(QMt=1,mje.exports=Sjn()),mje.exports}var H=_jn(),bje={exports:{}},ks={};/**
 * @license React
 * react.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var e8t;function kjn(){if(e8t)return ks;e8t=1;var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.portal"),o=Symbol.for("react.fragment"),s=Symbol.for("react.strict_mode"),c=Symbol.for("react.profiler"),d=Symbol.for("react.consumer"),h=Symbol.for("react.context"),b=Symbol.for("react.forward_ref"),y=Symbol.for("react.suspense"),E=Symbol.for("react.memo"),k=Symbol.for("react.lazy"),x=Symbol.iterator;function O(de){return de===null||typeof de!="object"?null:(de=x&&de[x]||de["@@iterator"],typeof de=="function"?de:null)}var P={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},D=Object.assign,$={};function U(de,We,ht){this.props=de,this.context=We,this.refs=$,this.updater=ht||P}U.prototype.isReactComponent={},U.prototype.setState=function(de,We){if(typeof de!="object"&&typeof de!="function"&&de!=null)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,de,We,"setState")},U.prototype.forceUpdate=function(de){this.updater.enqueueForceUpdate(this,de,"forceUpdate")};function B(){}B.prototype=U.prototype;function G(de,We,ht){this.props=de,this.context=We,this.refs=$,this.updater=ht||P}var K=G.prototype=new B;K.constructor=G,D(K,U.prototype),K.isPureReactComponent=!0;var X=Array.isArray,ae={H:null,A:null,T:null,S:null,V:null},oe=Object.prototype.hasOwnProperty;function he(de,We,ht,Ie,$t,An){return ht=An.ref,{$$typeof:n,type:de,key:We,ref:ht!==void 0?ht:null,props:An}}function ye(de,We){return he(de.type,We,void 0,void 0,void 0,de.props)}function we(de){return typeof de=="object"&&de!==null&&de.$$typeof===n}function ge(de){var We={"=":"=0",":":"=2"};return"$"+de.replace(/[=:]/g,function(ht){return We[ht]})}var Te=/\/+/g;function xe(de,We){return typeof de=="object"&&de!==null&&de.key!=null?ge(""+de.key):We.toString(36)}function Ue(){}function Z(de){switch(de.status){case"fulfilled":return de.value;case"rejected":throw de.reason;default:switch(typeof de.status=="string"?de.then(Ue,Ue):(de.status="pending",de.then(function(We){de.status==="pending"&&(de.status="fulfilled",de.value=We)},function(We){de.status==="pending"&&(de.status="rejected",de.reason=We)})),de.status){case"fulfilled":return de.value;case"rejected":throw de.reason}}throw de}function Re(de,We,ht,Ie,$t){var An=typeof de;(An==="undefined"||An==="boolean")&&(de=null);var Yt=!1;if(de===null)Yt=!0;else switch(An){case"bigint":case"string":case"number":Yt=!0;break;case"object":switch(de.$$typeof){case n:case r:Yt=!0;break;case k:return Yt=de._init,Re(Yt(de._payload),We,ht,Ie,$t)}}if(Yt)return $t=$t(de),Yt=Ie===""?"."+xe(de,0):Ie,X($t)?(ht="",Yt!=null&&(ht=Yt.replace(Te,"$&/")+"/"),Re($t,We,ht,"",function(on){return on})):$t!=null&&(we($t)&&($t=ye($t,ht+($t.key==null||de&&de.key===$t.key?"":(""+$t.key).replace(Te,"$&/")+"/")+Yt)),We.push($t)),1;Yt=0;var Or=Ie===""?".":Ie+":";if(X(de))for(var Nt=0;Nt<de.length;Nt++)Ie=de[Nt],An=Or+xe(Ie,Nt),Yt+=Re(Ie,We,ht,An,$t);else if(Nt=O(de),typeof Nt=="function")for(de=Nt.call(de),Nt=0;!(Ie=de.next()).done;)Ie=Ie.value,An=Or+xe(Ie,Nt++),Yt+=Re(Ie,We,ht,An,$t);else if(An==="object"){if(typeof de.then=="function")return Re(Z(de),We,ht,Ie,$t);throw We=String(de),Error("Objects are not valid as a React child (found: "+(We==="[object Object]"?"object with keys {"+Object.keys(de).join(", ")+"}":We)+"). If you meant to render a collection of children, use an array instead.")}return Yt}function _e(de,We,ht){if(de==null)return de;var Ie=[],$t=0;return Re(de,Ie,"","",function(An){return We.call(ht,An,$t++)}),Ie}function qe(de){if(de._status===-1){var We=de._result;We=We(),We.then(function(ht){(de._status===0||de._status===-1)&&(de._status=1,de._result=ht)},function(ht){(de._status===0||de._status===-1)&&(de._status=2,de._result=ht)}),de._status===-1&&(de._status=0,de._result=We)}if(de._status===1)return de._result.default;throw de._result}var it=typeof reportError=="function"?reportError:function(de){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var We=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof de=="object"&&de!==null&&typeof de.message=="string"?String(de.message):String(de),error:de});if(!window.dispatchEvent(We))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",de);return}console.error(de)};function Xe(){}return ks.Children={map:_e,forEach:function(de,We,ht){_e(de,function(){We.apply(this,arguments)},ht)},count:function(de){var We=0;return _e(de,function(){We++}),We},toArray:function(de){return _e(de,function(We){return We})||[]},only:function(de){if(!we(de))throw Error("React.Children.only expected to receive a single React element child.");return de}},ks.Component=U,ks.Fragment=o,ks.Profiler=c,ks.PureComponent=G,ks.StrictMode=s,ks.Suspense=y,ks.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=ae,ks.__COMPILER_RUNTIME={__proto__:null,c:function(de){return ae.H.useMemoCache(de)}},ks.cache=function(de){return function(){return de.apply(null,arguments)}},ks.cloneElement=function(de,We,ht){if(de==null)throw Error("The argument must be a React element, but you passed "+de+".");var Ie=D({},de.props),$t=de.key,An=void 0;if(We!=null)for(Yt in We.ref!==void 0&&(An=void 0),We.key!==void 0&&($t=""+We.key),We)!oe.call(We,Yt)||Yt==="key"||Yt==="__self"||Yt==="__source"||Yt==="ref"&&We.ref===void 0||(Ie[Yt]=We[Yt]);var Yt=arguments.length-2;if(Yt===1)Ie.children=ht;else if(1<Yt){for(var Or=Array(Yt),Nt=0;Nt<Yt;Nt++)Or[Nt]=arguments[Nt+2];Ie.children=Or}return he(de.type,$t,void 0,void 0,An,Ie)},ks.createContext=function(de){return de={$$typeof:h,_currentValue:de,_currentValue2:de,_threadCount:0,Provider:null,Consumer:null},de.Provider=de,de.Consumer={$$typeof:d,_context:de},de},ks.createElement=function(de,We,ht){var Ie,$t={},An=null;if(We!=null)for(Ie in We.key!==void 0&&(An=""+We.key),We)oe.call(We,Ie)&&Ie!=="key"&&Ie!=="__self"&&Ie!=="__source"&&($t[Ie]=We[Ie]);var Yt=arguments.length-2;if(Yt===1)$t.children=ht;else if(1<Yt){for(var Or=Array(Yt),Nt=0;Nt<Yt;Nt++)Or[Nt]=arguments[Nt+2];$t.children=Or}if(de&&de.defaultProps)for(Ie in Yt=de.defaultProps,Yt)$t[Ie]===void 0&&($t[Ie]=Yt[Ie]);return he(de,An,void 0,void 0,null,$t)},ks.createRef=function(){return{current:null}},ks.forwardRef=function(de){return{$$typeof:b,render:de}},ks.isValidElement=we,ks.lazy=function(de){return{$$typeof:k,_payload:{_status:-1,_result:de},_init:qe}},ks.memo=function(de,We){return{$$typeof:E,type:de,compare:We===void 0?null:We}},ks.startTransition=function(de){var We=ae.T,ht={};ae.T=ht;try{var Ie=de(),$t=ae.S;$t!==null&&$t(ht,Ie),typeof Ie=="object"&&Ie!==null&&typeof Ie.then=="function"&&Ie.then(Xe,it)}catch(An){it(An)}finally{ae.T=We}},ks.unstable_useCacheRefresh=function(){return ae.H.useCacheRefresh()},ks.use=function(de){return ae.H.use(de)},ks.useActionState=function(de,We,ht){return ae.H.useActionState(de,We,ht)},ks.useCallback=function(de,We){return ae.H.useCallback(de,We)},ks.useContext=function(de){return ae.H.useContext(de)},ks.useDebugValue=function(){},ks.useDeferredValue=function(de,We){return ae.H.useDeferredValue(de,We)},ks.useEffect=function(de,We,ht){var Ie=ae.H;if(typeof ht=="function")throw Error("useEffect CRUD overload is not enabled in this build of React.");return Ie.useEffect(de,We)},ks.useId=function(){return ae.H.useId()},ks.useImperativeHandle=function(de,We,ht){return ae.H.useImperativeHandle(de,We,ht)},ks.useInsertionEffect=function(de,We){return ae.H.useInsertionEffect(de,We)},ks.useLayoutEffect=function(de,We){return ae.H.useLayoutEffect(de,We)},ks.useMemo=function(de,We){return ae.H.useMemo(de,We)},ks.useOptimistic=function(de,We){return ae.H.useOptimistic(de,We)},ks.useReducer=function(de,We,ht){return ae.H.useReducer(de,We,ht)},ks.useRef=function(de){return ae.H.useRef(de)},ks.useState=function(de){return ae.H.useState(de)},ks.useSyncExternalStore=function(de,We,ht){return ae.H.useSyncExternalStore(de,We,ht)},ks.useTransition=function(){return ae.H.useTransition()},ks.version="19.1.1",ks}var t8t;function voe(){return t8t||(t8t=1,bje.exports=kjn()),bje.exports}var Q=voe();const Un=Y4(Q),n8t=Ejn({__proto__:null,default:Un},[Q]);var vje={exports:{}},Yne={},wje={exports:{}},yje={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var r8t;function Tjn(){return r8t||(r8t=1,function(n){function r(_e,qe){var it=_e.length;_e.push(qe);e:for(;0<it;){var Xe=it-1>>>1,de=_e[Xe];if(0<c(de,qe))_e[Xe]=qe,_e[it]=de,it=Xe;else break e}}function o(_e){return _e.length===0?null:_e[0]}function s(_e){if(_e.length===0)return null;var qe=_e[0],it=_e.pop();if(it!==qe){_e[0]=it;e:for(var Xe=0,de=_e.length,We=de>>>1;Xe<We;){var ht=2*(Xe+1)-1,Ie=_e[ht],$t=ht+1,An=_e[$t];if(0>c(Ie,it))$t<de&&0>c(An,Ie)?(_e[Xe]=An,_e[$t]=it,Xe=$t):(_e[Xe]=Ie,_e[ht]=it,Xe=ht);else if($t<de&&0>c(An,it))_e[Xe]=An,_e[$t]=it,Xe=$t;else break e}}return qe}function c(_e,qe){var it=_e.sortIndex-qe.sortIndex;return it!==0?it:_e.id-qe.id}if(n.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var d=performance;n.unstable_now=function(){return d.now()}}else{var h=Date,b=h.now();n.unstable_now=function(){return h.now()-b}}var y=[],E=[],k=1,x=null,O=3,P=!1,D=!1,$=!1,U=!1,B=typeof setTimeout=="function"?setTimeout:null,G=typeof clearTimeout=="function"?clearTimeout:null,K=typeof setImmediate<"u"?setImmediate:null;function X(_e){for(var qe=o(E);qe!==null;){if(qe.callback===null)s(E);else if(qe.startTime<=_e)s(E),qe.sortIndex=qe.expirationTime,r(y,qe);else break;qe=o(E)}}function ae(_e){if($=!1,X(_e),!D)if(o(y)!==null)D=!0,oe||(oe=!0,xe());else{var qe=o(E);qe!==null&&Re(ae,qe.startTime-_e)}}var oe=!1,he=-1,ye=5,we=-1;function ge(){return U?!0:!(n.unstable_now()-we<ye)}function Te(){if(U=!1,oe){var _e=n.unstable_now();we=_e;var qe=!0;try{e:{D=!1,$&&($=!1,G(he),he=-1),P=!0;var it=O;try{t:{for(X(_e),x=o(y);x!==null&&!(x.expirationTime>_e&&ge());){var Xe=x.callback;if(typeof Xe=="function"){x.callback=null,O=x.priorityLevel;var de=Xe(x.expirationTime<=_e);if(_e=n.unstable_now(),typeof de=="function"){x.callback=de,X(_e),qe=!0;break t}x===o(y)&&s(y),X(_e)}else s(y);x=o(y)}if(x!==null)qe=!0;else{var We=o(E);We!==null&&Re(ae,We.startTime-_e),qe=!1}}break e}finally{x=null,O=it,P=!1}qe=void 0}}finally{qe?xe():oe=!1}}}var xe;if(typeof K=="function")xe=function(){K(Te)};else if(typeof MessageChannel<"u"){var Ue=new MessageChannel,Z=Ue.port2;Ue.port1.onmessage=Te,xe=function(){Z.postMessage(null)}}else xe=function(){B(Te,0)};function Re(_e,qe){he=B(function(){_e(n.unstable_now())},qe)}n.unstable_IdlePriority=5,n.unstable_ImmediatePriority=1,n.unstable_LowPriority=4,n.unstable_NormalPriority=3,n.unstable_Profiling=null,n.unstable_UserBlockingPriority=2,n.unstable_cancelCallback=function(_e){_e.callback=null},n.unstable_forceFrameRate=function(_e){0>_e||125<_e?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):ye=0<_e?Math.floor(1e3/_e):5},n.unstable_getCurrentPriorityLevel=function(){return O},n.unstable_next=function(_e){switch(O){case 1:case 2:case 3:var qe=3;break;default:qe=O}var it=O;O=qe;try{return _e()}finally{O=it}},n.unstable_requestPaint=function(){U=!0},n.unstable_runWithPriority=function(_e,qe){switch(_e){case 1:case 2:case 3:case 4:case 5:break;default:_e=3}var it=O;O=_e;try{return qe()}finally{O=it}},n.unstable_scheduleCallback=function(_e,qe,it){var Xe=n.unstable_now();switch(typeof it=="object"&&it!==null?(it=it.delay,it=typeof it=="number"&&0<it?Xe+it:Xe):it=Xe,_e){case 1:var de=-1;break;case 2:de=250;break;case 5:de=1073741823;break;case 4:de=1e4;break;default:de=5e3}return de=it+de,_e={id:k++,callback:qe,priorityLevel:_e,startTime:it,expirationTime:de,sortIndex:-1},it>Xe?(_e.sortIndex=it,r(E,_e),o(y)===null&&_e===o(E)&&($?(G(he),he=-1):$=!0,Re(ae,it-Xe))):(_e.sortIndex=de,r(y,_e),D||P||(D=!0,oe||(oe=!0,xe()))),_e},n.unstable_shouldYield=ge,n.unstable_wrapCallback=function(_e){var qe=O;return function(){var it=O;O=qe;try{return _e.apply(this,arguments)}finally{O=it}}}}(yje)),yje}var i8t;function xjn(){return i8t||(i8t=1,wje.exports=Tjn()),wje.exports}var Eje={exports:{}},B1={};/**
 * @license React
 * react-dom.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var o8t;function Cjn(){if(o8t)return B1;o8t=1;var n=voe();function r(y){var E="https://react.dev/errors/"+y;if(1<arguments.length){E+="?args[]="+encodeURIComponent(arguments[1]);for(var k=2;k<arguments.length;k++)E+="&args[]="+encodeURIComponent(arguments[k])}return"Minified React error #"+y+"; visit "+E+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function o(){}var s={d:{f:o,r:function(){throw Error(r(522))},D:o,C:o,L:o,m:o,X:o,S:o,M:o},p:0,findDOMNode:null},c=Symbol.for("react.portal");function d(y,E,k){var x=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:c,key:x==null?null:""+x,children:y,containerInfo:E,implementation:k}}var h=n.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function b(y,E){if(y==="font")return"";if(typeof E=="string")return E==="use-credentials"?E:""}return B1.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=s,B1.createPortal=function(y,E){var k=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!E||E.nodeType!==1&&E.nodeType!==9&&E.nodeType!==11)throw Error(r(299));return d(y,E,null,k)},B1.flushSync=function(y){var E=h.T,k=s.p;try{if(h.T=null,s.p=2,y)return y()}finally{h.T=E,s.p=k,s.d.f()}},B1.preconnect=function(y,E){typeof y=="string"&&(E?(E=E.crossOrigin,E=typeof E=="string"?E==="use-credentials"?E:"":void 0):E=null,s.d.C(y,E))},B1.prefetchDNS=function(y){typeof y=="string"&&s.d.D(y)},B1.preinit=function(y,E){if(typeof y=="string"&&E&&typeof E.as=="string"){var k=E.as,x=b(k,E.crossOrigin),O=typeof E.integrity=="string"?E.integrity:void 0,P=typeof E.fetchPriority=="string"?E.fetchPriority:void 0;k==="style"?s.d.S(y,typeof E.precedence=="string"?E.precedence:void 0,{crossOrigin:x,integrity:O,fetchPriority:P}):k==="script"&&s.d.X(y,{crossOrigin:x,integrity:O,fetchPriority:P,nonce:typeof E.nonce=="string"?E.nonce:void 0})}},B1.preinitModule=function(y,E){if(typeof y=="string")if(typeof E=="object"&&E!==null){if(E.as==null||E.as==="script"){var k=b(E.as,E.crossOrigin);s.d.M(y,{crossOrigin:k,integrity:typeof E.integrity=="string"?E.integrity:void 0,nonce:typeof E.nonce=="string"?E.nonce:void 0})}}else E==null&&s.d.M(y)},B1.preload=function(y,E){if(typeof y=="string"&&typeof E=="object"&&E!==null&&typeof E.as=="string"){var k=E.as,x=b(k,E.crossOrigin);s.d.L(y,k,{crossOrigin:x,integrity:typeof E.integrity=="string"?E.integrity:void 0,nonce:typeof E.nonce=="string"?E.nonce:void 0,type:typeof E.type=="string"?E.type:void 0,fetchPriority:typeof E.fetchPriority=="string"?E.fetchPriority:void 0,referrerPolicy:typeof E.referrerPolicy=="string"?E.referrerPolicy:void 0,imageSrcSet:typeof E.imageSrcSet=="string"?E.imageSrcSet:void 0,imageSizes:typeof E.imageSizes=="string"?E.imageSizes:void 0,media:typeof E.media=="string"?E.media:void 0})}},B1.preloadModule=function(y,E){if(typeof y=="string")if(E){var k=b(E.as,E.crossOrigin);s.d.m(y,{as:typeof E.as=="string"&&E.as!=="script"?E.as:void 0,crossOrigin:k,integrity:typeof E.integrity=="string"?E.integrity:void 0})}else s.d.m(y)},B1.requestFormReset=function(y){s.d.r(y)},B1.unstable_batchedUpdates=function(y,E){return y(E)},B1.useFormState=function(y,E,k){return h.H.useFormState(y,E,k)},B1.useFormStatus=function(){return h.H.useHostTransitionStatus()},B1.version="19.1.1",B1}var a8t;function vKt(){if(a8t)return Eje.exports;a8t=1;function n(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(n)}catch(r){console.error(r)}}return n(),Eje.exports=Cjn(),Eje.exports}/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var s8t;function Ajn(){if(s8t)return Yne;s8t=1;var n=xjn(),r=voe(),o=vKt();function s(w){var _="https://react.dev/errors/"+w;if(1<arguments.length){_+="?args[]="+encodeURIComponent(arguments[1]);for(var C=2;C<arguments.length;C++)_+="&args[]="+encodeURIComponent(arguments[C])}return"Minified React error #"+w+"; visit "+_+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function c(w){return!(!w||w.nodeType!==1&&w.nodeType!==9&&w.nodeType!==11)}function d(w){var _=w,C=w;if(w.alternate)for(;_.return;)_=_.return;else{w=_;do _=w,(_.flags&4098)!==0&&(C=_.return),w=_.return;while(w)}return _.tag===3?C:null}function h(w){if(w.tag===13){var _=w.memoizedState;if(_===null&&(w=w.alternate,w!==null&&(_=w.memoizedState)),_!==null)return _.dehydrated}return null}function b(w){if(d(w)!==w)throw Error(s(188))}function y(w){var _=w.alternate;if(!_){if(_=d(w),_===null)throw Error(s(188));return _!==w?null:w}for(var C=w,j=_;;){var q=C.return;if(q===null)break;var te=q.alternate;if(te===null){if(j=q.return,j!==null){C=j;continue}break}if(q.child===te.child){for(te=q.child;te;){if(te===C)return b(q),w;if(te===j)return b(q),_;te=te.sibling}throw Error(s(188))}if(C.return!==j.return)C=q,j=te;else{for(var Se=!1,Me=q.child;Me;){if(Me===C){Se=!0,C=q,j=te;break}if(Me===j){Se=!0,j=q,C=te;break}Me=Me.sibling}if(!Se){for(Me=te.child;Me;){if(Me===C){Se=!0,C=te,j=q;break}if(Me===j){Se=!0,j=te,C=q;break}Me=Me.sibling}if(!Se)throw Error(s(189))}}if(C.alternate!==j)throw Error(s(190))}if(C.tag!==3)throw Error(s(188));return C.stateNode.current===C?w:_}function E(w){var _=w.tag;if(_===5||_===26||_===27||_===6)return w;for(w=w.child;w!==null;){if(_=E(w),_!==null)return _;w=w.sibling}return null}var k=Object.assign,x=Symbol.for("react.element"),O=Symbol.for("react.transitional.element"),P=Symbol.for("react.portal"),D=Symbol.for("react.fragment"),$=Symbol.for("react.strict_mode"),U=Symbol.for("react.profiler"),B=Symbol.for("react.provider"),G=Symbol.for("react.consumer"),K=Symbol.for("react.context"),X=Symbol.for("react.forward_ref"),ae=Symbol.for("react.suspense"),oe=Symbol.for("react.suspense_list"),he=Symbol.for("react.memo"),ye=Symbol.for("react.lazy"),we=Symbol.for("react.activity"),ge=Symbol.for("react.memo_cache_sentinel"),Te=Symbol.iterator;function xe(w){return w===null||typeof w!="object"?null:(w=Te&&w[Te]||w["@@iterator"],typeof w=="function"?w:null)}var Ue=Symbol.for("react.client.reference");function Z(w){if(w==null)return null;if(typeof w=="function")return w.$$typeof===Ue?null:w.displayName||w.name||null;if(typeof w=="string")return w;switch(w){case D:return"Fragment";case U:return"Profiler";case $:return"StrictMode";case ae:return"Suspense";case oe:return"SuspenseList";case we:return"Activity"}if(typeof w=="object")switch(w.$$typeof){case P:return"Portal";case K:return(w.displayName||"Context")+".Provider";case G:return(w._context.displayName||"Context")+".Consumer";case X:var _=w.render;return w=w.displayName,w||(w=_.displayName||_.name||"",w=w!==""?"ForwardRef("+w+")":"ForwardRef"),w;case he:return _=w.displayName||null,_!==null?_:Z(w.type)||"Memo";case ye:_=w._payload,w=w._init;try{return Z(w(_))}catch{}}return null}var Re=Array.isArray,_e=r.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,qe=o.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,it={pending:!1,data:null,method:null,action:null},Xe=[],de=-1;function We(w){return{current:w}}function ht(w){0>de||(w.current=Xe[de],Xe[de]=null,de--)}function Ie(w,_){de++,Xe[de]=w.current,w.current=_}var $t=We(null),An=We(null),Yt=We(null),Or=We(null);function Nt(w,_){switch(Ie(Yt,_),Ie(An,w),Ie($t,null),_.nodeType){case 9:case 11:w=(w=_.documentElement)&&(w=w.namespaceURI)?ul(w):0;break;default:if(w=_.tagName,_=_.namespaceURI)_=ul(_),w=kl(_,w);else switch(w){case"svg":w=1;break;case"math":w=2;break;default:w=0}}ht($t),Ie($t,w)}function on(){ht($t),ht(An),ht(Yt)}function br(w){w.memoizedState!==null&&Ie(Or,w);var _=$t.current,C=kl(_,w.type);_!==C&&(Ie(An,w),Ie($t,C))}function wi(w){An.current===w&&(ht($t),ht(An)),Or.current===w&&(ht(Or),Mk._currentValue=it)}var no=Object.prototype.hasOwnProperty,Ri=n.unstable_scheduleCallback,ro=n.unstable_cancelCallback,Ko=n.unstable_shouldYield,Ca=n.unstable_requestPaint,qi=n.unstable_now,yi=n.unstable_getCurrentPriorityLevel,lo=n.unstable_ImmediatePriority,Eo=n.unstable_UserBlockingPriority,Fc=n.unstable_NormalPriority,Yo=n.unstable_LowPriority,zl=n.unstable_IdlePriority,ld=n.log,Yc=n.unstable_setDisableYieldValue,Sn=null,Pn=null;function Kr(w){if(typeof ld=="function"&&Yc(w),Pn&&typeof Pn.setStrictMode=="function")try{Pn.setStrictMode(Sn,w)}catch{}}var Ei=Math.clz32?Math.clz32:Cc,Sa=Math.log,bs=Math.LN2;function Cc(w){return w>>>=0,w===0?32:31-(Sa(w)/bs|0)|0}var Ac=256,Xc=4194304;function rc(w){var _=w&42;if(_!==0)return _;switch(w&-w){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:return 128;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return w&4194048;case 4194304:case 8388608:case 16777216:case 33554432:return w&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return w}}function js(w,_,C){var j=w.pendingLanes;if(j===0)return 0;var q=0,te=w.suspendedLanes,Se=w.pingedLanes;w=w.warmLanes;var Me=j&134217727;return Me!==0?(j=Me&~te,j!==0?q=rc(j):(Se&=Me,Se!==0?q=rc(Se):C||(C=Me&~w,C!==0&&(q=rc(C))))):(Me=j&~te,Me!==0?q=rc(Me):Se!==0?q=rc(Se):C||(C=j&~w,C!==0&&(q=rc(C)))),q===0?0:_!==0&&_!==q&&(_&te)===0&&(te=q&-q,C=_&-_,te>=C||te===32&&(C&4194048)!==0)?_:q}function Jc(w,_){return(w.pendingLanes&~(w.suspendedLanes&~w.pingedLanes)&_)===0}function Id(w,_){switch(w){case 1:case 2:case 4:case 8:case 64:return _+250;case 16:case 32:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return _+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function nh(){var w=Ac;return Ac<<=1,(Ac&4194048)===0&&(Ac=256),w}function hu(){var w=Xc;return Xc<<=1,(Xc&62914560)===0&&(Xc=4194304),w}function Qp(w){for(var _=[],C=0;31>C;C++)_.push(w);return _}function Kb(w,_){w.pendingLanes|=_,_!==268435456&&(w.suspendedLanes=0,w.pingedLanes=0,w.warmLanes=0)}function Km(w,_,C,j,q,te){var Se=w.pendingLanes;w.pendingLanes=C,w.suspendedLanes=0,w.pingedLanes=0,w.warmLanes=0,w.expiredLanes&=C,w.entangledLanes&=C,w.errorRecoveryDisabledLanes&=C,w.shellSuspendCounter=0;var Me=w.entanglements,pt=w.expirationTimes,Xt=w.hiddenUpdates;for(C=Se&~C;0<C;){var $n=31-Ei(C),Vn=1<<$n;Me[$n]=0,pt[$n]=-1;var ln=Xt[$n];if(ln!==null)for(Xt[$n]=null,$n=0;$n<ln.length;$n++){var Zt=ln[$n];Zt!==null&&(Zt.lane&=-536870913)}C&=~Vn}j!==0&&S2(w,j,0),te!==0&&q===0&&w.tag!==0&&(w.suspendedLanes|=te&~(Se&~_))}function S2(w,_,C){w.pendingLanes|=_,w.suspendedLanes&=~_;var j=31-Ei(_);w.entangledLanes|=_,w.entanglements[j]=w.entanglements[j]|1073741824|C&4194090}function _2(w,_){var C=w.entangledLanes|=_;for(w=w.entanglements;C;){var j=31-Ei(C),q=1<<j;q&_|w[j]&_&&(w[j]|=_),C&=~q}}function dv(w){switch(w){case 2:w=1;break;case 8:w=4;break;case 32:w=16;break;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:w=128;break;case 268435456:w=134217728;break;default:w=0}return w}function K_(w){return w&=-w,2<w?8<w?(w&134217727)!==0?32:268435456:8:2}function YI(){var w=qe.p;return w!==0?w:(w=window.event,w===void 0?32:hz(w.type))}function n6(w,_){var C=qe.p;try{return qe.p=w,_()}finally{qe.p=C}}var e0=Math.random().toString(36).slice(2),_l="__reactFiber$"+e0,rh="__reactProps$"+e0,t0="__reactContainer$"+e0,k2="__reactEvents$"+e0,r6="__reactListeners$"+e0,i6="__reactHandles$"+e0,XI="__reactResources$"+e0,sy="__reactMarker$"+e0;function rC(w){delete w[_l],delete w[rh],delete w[k2],delete w[r6],delete w[i6]}function fv(w){var _=w[_l];if(_)return _;for(var C=w.parentNode;C;){if(_=C[t0]||C[_l]){if(C=_.alternate,_.child!==null||C!==null&&C.child!==null)for(w=U9(w);w!==null;){if(C=w[_l])return C;w=U9(w)}return _}w=C,C=w.parentNode}return null}function cy(w){if(w=w[_l]||w[t0]){var _=w.tag;if(_===5||_===6||_===13||_===26||_===27||_===3)return w}return null}function uy(w){var _=w.tag;if(_===5||_===26||_===27||_===6)return w.stateNode;throw Error(s(33))}function ly(w){var _=w[XI];return _||(_=w[XI]={hoistableStyles:new Map,hoistableScripts:new Map}),_}function Od(w){w[sy]=!0}var JI=new Set,ZI={};function hv(w,_){dy(w,_),dy(w+"Capture",_)}function dy(w,_){for(ZI[w]=_,w=0;w<_.length;w++)JI.add(_[w])}var T2=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),QI={},x2={};function o6(w){return no.call(x2,w)?!0:no.call(QI,w)?!1:T2.test(w)?x2[w]=!0:(QI[w]=!0,!1)}function fy(w,_,C){if(o6(_))if(C===null)w.removeAttribute(_);else{switch(typeof C){case"undefined":case"function":case"symbol":w.removeAttribute(_);return;case"boolean":var j=_.toLowerCase().slice(0,5);if(j!=="data-"&&j!=="aria-"){w.removeAttribute(_);return}}w.setAttribute(_,""+C)}}function Y_(w,_,C){if(C===null)w.removeAttribute(_);else{switch(typeof C){case"undefined":case"function":case"symbol":case"boolean":w.removeAttribute(_);return}w.setAttribute(_,""+C)}}function Ym(w,_,C,j){if(j===null)w.removeAttribute(C);else{switch(typeof j){case"undefined":case"function":case"symbol":case"boolean":w.removeAttribute(C);return}w.setAttributeNS(_,C,""+j)}}var C2,iC;function hy(w){if(C2===void 0)try{throw Error()}catch(C){var _=C.stack.trim().match(/\n( *(at )?)/);C2=_&&_[1]||"",iC=-1<C.stack.indexOf(`
    at`)?" (<anonymous>)":-1<C.stack.indexOf("@")?"@unknown:0:0":""}return`
`+C2+w+iC}var oC=!1;function py(w,_){if(!w||oC)return"";oC=!0;var C=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var j={DetermineComponentFrameRoot:function(){try{if(_){var Vn=function(){throw Error()};if(Object.defineProperty(Vn.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(Vn,[])}catch(Zt){var ln=Zt}Reflect.construct(w,[],Vn)}else{try{Vn.call()}catch(Zt){ln=Zt}w.call(Vn.prototype)}}else{try{throw Error()}catch(Zt){ln=Zt}(Vn=w())&&typeof Vn.catch=="function"&&Vn.catch(function(){})}}catch(Zt){if(Zt&&ln&&typeof Zt.stack=="string")return[Zt.stack,ln.stack]}return[null,null]}};j.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var q=Object.getOwnPropertyDescriptor(j.DetermineComponentFrameRoot,"name");q&&q.configurable&&Object.defineProperty(j.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var te=j.DetermineComponentFrameRoot(),Se=te[0],Me=te[1];if(Se&&Me){var pt=Se.split(`
`),Xt=Me.split(`
`);for(q=j=0;j<pt.length&&!pt[j].includes("DetermineComponentFrameRoot");)j++;for(;q<Xt.length&&!Xt[q].includes("DetermineComponentFrameRoot");)q++;if(j===pt.length||q===Xt.length)for(j=pt.length-1,q=Xt.length-1;1<=j&&0<=q&&pt[j]!==Xt[q];)q--;for(;1<=j&&0<=q;j--,q--)if(pt[j]!==Xt[q]){if(j!==1||q!==1)do if(j--,q--,0>q||pt[j]!==Xt[q]){var $n=`
`+pt[j].replace(" at new "," at ");return w.displayName&&$n.includes("<anonymous>")&&($n=$n.replace("<anonymous>",w.displayName)),$n}while(1<=j&&0<=q);break}}}finally{oC=!1,Error.prepareStackTrace=C}return(C=w?w.displayName||w.name:"")?hy(C):""}function VM(w){switch(w.tag){case 26:case 27:case 5:return hy(w.type);case 16:return hy("Lazy");case 13:return hy("Suspense");case 19:return hy("SuspenseList");case 0:case 15:return py(w.type,!1);case 11:return py(w.type.render,!1);case 1:return py(w.type,!0);case 31:return hy("Activity");default:return""}}function a6(w){try{var _="";do _+=VM(w),w=w.return;while(w);return _}catch(C){return`
Error generating stack: `+C.message+`
`+C.stack}}function Kg(w){switch(typeof w){case"bigint":case"boolean":case"number":case"string":case"undefined":return w;case"object":return w;default:return""}}function s6(w){var _=w.type;return(w=w.nodeName)&&w.toLowerCase()==="input"&&(_==="checkbox"||_==="radio")}function c6(w){var _=s6(w)?"checked":"value",C=Object.getOwnPropertyDescriptor(w.constructor.prototype,_),j=""+w[_];if(!w.hasOwnProperty(_)&&typeof C<"u"&&typeof C.get=="function"&&typeof C.set=="function"){var q=C.get,te=C.set;return Object.defineProperty(w,_,{configurable:!0,get:function(){return q.call(this)},set:function(Se){j=""+Se,te.call(this,Se)}}),Object.defineProperty(w,_,{enumerable:C.enumerable}),{getValue:function(){return j},setValue:function(Se){j=""+Se},stopTracking:function(){w._valueTracker=null,delete w[_]}}}}function eO(w){w._valueTracker||(w._valueTracker=c6(w))}function u6(w){if(!w)return!1;var _=w._valueTracker;if(!_)return!0;var C=_.getValue(),j="";return w&&(j=s6(w)?w.checked?"true":"false":w.value),w=j,w!==C?(_.setValue(w),!0):!1}function A2(w){if(w=w||(typeof document<"u"?document:void 0),typeof w>"u")return null;try{return w.activeElement||w.body}catch{return w.body}}var WM=/[\n"\\]/g;function Xm(w){return w.replace(WM,function(_){return"\\"+_.charCodeAt(0).toString(16)+" "})}function tO(w,_,C,j,q,te,Se,Me){w.name="",Se!=null&&typeof Se!="function"&&typeof Se!="symbol"&&typeof Se!="boolean"?w.type=Se:w.removeAttribute("type"),_!=null?Se==="number"?(_===0&&w.value===""||w.value!=_)&&(w.value=""+Kg(_)):w.value!==""+Kg(_)&&(w.value=""+Kg(_)):Se!=="submit"&&Se!=="reset"||w.removeAttribute("value"),_!=null?aC(w,Se,Kg(_)):C!=null?aC(w,Se,Kg(C)):j!=null&&w.removeAttribute("value"),q==null&&te!=null&&(w.defaultChecked=!!te),q!=null&&(w.checked=q&&typeof q!="function"&&typeof q!="symbol"),Me!=null&&typeof Me!="function"&&typeof Me!="symbol"&&typeof Me!="boolean"?w.name=""+Kg(Me):w.removeAttribute("name")}function l6(w,_,C,j,q,te,Se,Me){if(te!=null&&typeof te!="function"&&typeof te!="symbol"&&typeof te!="boolean"&&(w.type=te),_!=null||C!=null){if(!(te!=="submit"&&te!=="reset"||_!=null))return;C=C!=null?""+Kg(C):"",_=_!=null?""+Kg(_):C,Me||_===w.value||(w.value=_),w.defaultValue=_}j=j??q,j=typeof j!="function"&&typeof j!="symbol"&&!!j,w.checked=Me?w.checked:!!j,w.defaultChecked=!!j,Se!=null&&typeof Se!="function"&&typeof Se!="symbol"&&typeof Se!="boolean"&&(w.name=Se)}function aC(w,_,C){_==="number"&&A2(w.ownerDocument)===w||w.defaultValue===""+C||(w.defaultValue=""+C)}function X_(w,_,C,j){if(w=w.options,_){_={};for(var q=0;q<C.length;q++)_["$"+C[q]]=!0;for(C=0;C<w.length;C++)q=_.hasOwnProperty("$"+w[C].value),w[C].selected!==q&&(w[C].selected=q),q&&j&&(w[C].defaultSelected=!0)}else{for(C=""+Kg(C),_=null,q=0;q<w.length;q++){if(w[q].value===C){w[q].selected=!0,j&&(w[q].defaultSelected=!0);return}_!==null||w[q].disabled||(_=w[q])}_!==null&&(_.selected=!0)}}function d6(w,_,C){if(_!=null&&(_=""+Kg(_),_!==w.value&&(w.value=_),C==null)){w.defaultValue!==_&&(w.defaultValue=_);return}w.defaultValue=C!=null?""+Kg(C):""}function nO(w,_,C,j){if(_==null){if(j!=null){if(C!=null)throw Error(s(92));if(Re(j)){if(1<j.length)throw Error(s(93));j=j[0]}C=j}C==null&&(C=""),_=C}C=Kg(_),w.defaultValue=C,j=w.textContent,j===C&&j!==""&&j!==null&&(w.value=j)}function gy(w,_){if(_){var C=w.firstChild;if(C&&C===w.lastChild&&C.nodeType===3){C.nodeValue=_;return}}w.textContent=_}var J_=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function f6(w,_,C){var j=_.indexOf("--")===0;C==null||typeof C=="boolean"||C===""?j?w.setProperty(_,""):_==="float"?w.cssFloat="":w[_]="":j?w.setProperty(_,C):typeof C!="number"||C===0||J_.has(_)?_==="float"?w.cssFloat=C:w[_]=(""+C).trim():w[_]=C+"px"}function h6(w,_,C){if(_!=null&&typeof _!="object")throw Error(s(62));if(w=w.style,C!=null){for(var j in C)!C.hasOwnProperty(j)||_!=null&&_.hasOwnProperty(j)||(j.indexOf("--")===0?w.setProperty(j,""):j==="float"?w.cssFloat="":w[j]="");for(var q in _)j=_[q],_.hasOwnProperty(q)&&C[q]!==j&&f6(w,q,j)}else for(var te in _)_.hasOwnProperty(te)&&f6(w,te,_[te])}function p6(w){if(w.indexOf("-")===-1)return!1;switch(w){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var KM=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),YM=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function sC(w){return YM.test(""+w)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":w}var rO=null;function g6(w){return w=w.target||w.srcElement||window,w.correspondingUseElement&&(w=w.correspondingUseElement),w.nodeType===3?w.parentNode:w}var Z_=null,Q_=null;function m6(w){var _=cy(w);if(_&&(w=_.stateNode)){var C=w[rh]||null;e:switch(w=_.stateNode,_.type){case"input":if(tO(w,C.value,C.defaultValue,C.defaultValue,C.checked,C.defaultChecked,C.type,C.name),_=C.name,C.type==="radio"&&_!=null){for(C=w;C.parentNode;)C=C.parentNode;for(C=C.querySelectorAll('input[name="'+Xm(""+_)+'"][type="radio"]'),_=0;_<C.length;_++){var j=C[_];if(j!==w&&j.form===w.form){var q=j[rh]||null;if(!q)throw Error(s(90));tO(j,q.value,q.defaultValue,q.defaultValue,q.checked,q.defaultChecked,q.type,q.name)}}for(_=0;_<C.length;_++)j=C[_],j.form===w.form&&u6(j)}break e;case"textarea":d6(w,C.value,C.defaultValue);break e;case"select":_=C.value,_!=null&&X_(w,!!C.multiple,_,!1)}}}var b6=!1;function XM(w,_,C){if(b6)return w(_,C);b6=!0;try{var j=w(_);return j}finally{if(b6=!1,(Z_!==null||Q_!==null)&&(eN(),Z_&&(_=Z_,w=Q_,Q_=Z_=null,m6(_),w)))for(_=0;_<w.length;_++)m6(w[_])}}function cC(w,_){var C=w.stateNode;if(C===null)return null;var j=C[rh]||null;if(j===null)return null;C=j[_];e:switch(_){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(j=!j.disabled)||(w=w.type,j=!(w==="button"||w==="input"||w==="select"||w==="textarea")),w=!j;break e;default:w=!1}if(w)return null;if(C&&typeof C!="function")throw Error(s(231,_,typeof C));return C}var pv=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),uC=!1;if(pv)try{var lC={};Object.defineProperty(lC,"passive",{get:function(){uC=!0}}),window.addEventListener("test",lC,lC),window.removeEventListener("test",lC,lC)}catch{uC=!1}var my=null,v6=null,iO=null;function JM(){if(iO)return iO;var w,_=v6,C=_.length,j,q="value"in my?my.value:my.textContent,te=q.length;for(w=0;w<C&&_[w]===q[w];w++);var Se=C-w;for(j=1;j<=Se&&_[C-j]===q[te-j];j++);return iO=q.slice(w,1<j?1-j:void 0)}function Zc(w){var _=w.keyCode;return"charCode"in w?(w=w.charCode,w===0&&_===13&&(w=13)):w=_,w===10&&(w=13),32<=w||w===13?w:0}function Bc(){return!0}function ZM(){return!1}function eg(w){function _(C,j,q,te,Se){this._reactName=C,this._targetInst=q,this.type=j,this.nativeEvent=te,this.target=Se,this.currentTarget=null;for(var Me in w)w.hasOwnProperty(Me)&&(C=w[Me],this[Me]=C?C(te):te[Me]);return this.isDefaultPrevented=(te.defaultPrevented!=null?te.defaultPrevented:te.returnValue===!1)?Bc:ZM,this.isPropagationStopped=ZM,this}return k(_.prototype,{preventDefault:function(){this.defaultPrevented=!0;var C=this.nativeEvent;C&&(C.preventDefault?C.preventDefault():typeof C.returnValue!="unknown"&&(C.returnValue=!1),this.isDefaultPrevented=Bc)},stopPropagation:function(){var C=this.nativeEvent;C&&(C.stopPropagation?C.stopPropagation():typeof C.cancelBubble!="unknown"&&(C.cancelBubble=!0),this.isPropagationStopped=Bc)},persist:function(){},isPersistent:Bc}),_}var I2={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(w){return w.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},oO=eg(I2),dC=k({},I2,{view:0,detail:0}),SU=eg(dC),aO,w6,dd,ek=k({},dC,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:n0,button:0,buttons:0,relatedTarget:function(w){return w.relatedTarget===void 0?w.fromElement===w.srcElement?w.toElement:w.fromElement:w.relatedTarget},movementX:function(w){return"movementX"in w?w.movementX:(w!==dd&&(dd&&w.type==="mousemove"?(aO=w.screenX-dd.screenX,w6=w.screenY-dd.screenY):w6=aO=0,dd=w),aO)},movementY:function(w){return"movementY"in w?w.movementY:w6}}),by=eg(ek),QM=k({},ek,{dataTransfer:0}),y6=eg(QM),Jm=k({},dC,{relatedTarget:0}),sO=eg(Jm),e8=k({},I2,{animationName:0,elapsedTime:0,pseudoElement:0}),t8=eg(e8),n8=k({},I2,{clipboardData:function(w){return"clipboardData"in w?w.clipboardData:window.clipboardData}}),r8=eg(n8),i8=k({},I2,{data:0}),tg=eg(i8),_U={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},Zm={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},kU={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function o8(w){var _=this.nativeEvent;return _.getModifierState?_.getModifierState(w):(w=kU[w])?!!_[w]:!1}function n0(){return o8}var cO=k({},dC,{key:function(w){if(w.key){var _=_U[w.key]||w.key;if(_!=="Unidentified")return _}return w.type==="keypress"?(w=Zc(w),w===13?"Enter":String.fromCharCode(w)):w.type==="keydown"||w.type==="keyup"?Zm[w.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:n0,charCode:function(w){return w.type==="keypress"?Zc(w):0},keyCode:function(w){return w.type==="keydown"||w.type==="keyup"?w.keyCode:0},which:function(w){return w.type==="keypress"?Zc(w):w.type==="keydown"||w.type==="keyup"?w.keyCode:0}}),E6=eg(cO),al=k({},ek,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),gv=eg(al),S6=k({},dC,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:n0}),TU=eg(S6),_6=k({},I2,{propertyName:0,elapsedTime:0,pseudoElement:0}),xU=eg(_6),CU=k({},ek,{deltaX:function(w){return"deltaX"in w?w.deltaX:"wheelDeltaX"in w?-w.wheelDeltaX:0},deltaY:function(w){return"deltaY"in w?w.deltaY:"wheelDeltaY"in w?-w.wheelDeltaY:"wheelDelta"in w?-w.wheelDelta:0},deltaZ:0,deltaMode:0}),AU=eg(CU),a8=k({},I2,{newState:0,oldState:0}),uO=eg(a8),tk=[9,13,27,32],k6=pv&&"CompositionEvent"in window,nk=null;pv&&"documentMode"in document&&(nk=document.documentMode);var s8=pv&&"TextEvent"in window&&!nk,lO=pv&&(!k6||nk&&8<nk&&11>=nk),dO=" ",T6=!1;function c8(w,_){switch(w){case"keyup":return tk.indexOf(_.keyCode)!==-1;case"keydown":return _.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function fO(w){return w=w.detail,typeof w=="object"&&"data"in w?w.data:null}var vy=!1;function IU(w,_){switch(w){case"compositionend":return fO(_);case"keypress":return _.which!==32?null:(T6=!0,dO);case"textInput":return w=_.data,w===dO&&T6?null:w;default:return null}}function u8(w,_){if(vy)return w==="compositionend"||!k6&&c8(w,_)?(w=JM(),iO=v6=my=null,vy=!1,w):null;switch(w){case"paste":return null;case"keypress":if(!(_.ctrlKey||_.altKey||_.metaKey)||_.ctrlKey&&_.altKey){if(_.char&&1<_.char.length)return _.char;if(_.which)return String.fromCharCode(_.which)}return null;case"compositionend":return lO&&_.locale!=="ko"?null:_.data;default:return null}}var Yg={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function hO(w){var _=w&&w.nodeName&&w.nodeName.toLowerCase();return _==="input"?!!Yg[w.type]:_==="textarea"}function rk(w,_,C,j){Z_?Q_?Q_.push(j):Q_=[j]:Z_=j,_=cN(_,"onChange"),0<_.length&&(C=new oO("onChange","change",null,C,j),w.push({event:C,listeners:_}))}var ik=null,ok=null;function OU(w){sN(w,0)}function pO(w){var _=uy(w);if(u6(_))return w}function l8(w,_){if(w==="change")return _}var d8=!1;if(pv){var x6;if(pv){var mv="oninput"in document;if(!mv){var f8=document.createElement("div");f8.setAttribute("oninput","return;"),mv=typeof f8.oninput=="function"}x6=mv}else x6=!1;d8=x6&&(!document.documentMode||9<document.documentMode)}function h8(){ik&&(ik.detachEvent("onpropertychange",gO),ok=ik=null)}function gO(w){if(w.propertyName==="value"&&pO(ok)){var _=[];rk(_,ok,w,g6(w)),XM(OU,_)}}function p8(w,_,C){w==="focusin"?(h8(),ik=_,ok=C,ik.attachEvent("onpropertychange",gO)):w==="focusout"&&h8()}function wy(w){if(w==="selectionchange"||w==="keyup"||w==="keydown")return pO(ok)}function mO(w,_){if(w==="click")return pO(_)}function NU(w,_){if(w==="input"||w==="change")return pO(_)}function RU(w,_){return w===_&&(w!==0||1/w===1/_)||w!==w&&_!==_}var Xg=typeof Object.is=="function"?Object.is:RU;function r0(w,_){if(Xg(w,_))return!0;if(typeof w!="object"||w===null||typeof _!="object"||_===null)return!1;var C=Object.keys(w),j=Object.keys(_);if(C.length!==j.length)return!1;for(j=0;j<C.length;j++){var q=C[j];if(!no.call(_,q)||!Xg(w[q],_[q]))return!1}return!0}function Qm(w){for(;w&&w.firstChild;)w=w.firstChild;return w}function C6(w,_){var C=Qm(w);w=0;for(var j;C;){if(C.nodeType===3){if(j=w+C.textContent.length,w<=_&&j>=_)return{node:C,offset:_-w};w=j}e:{for(;C;){if(C.nextSibling){C=C.nextSibling;break e}C=C.parentNode}C=void 0}C=Qm(C)}}function g8(w,_){return w&&_?w===_?!0:w&&w.nodeType===3?!1:_&&_.nodeType===3?g8(w,_.parentNode):"contains"in w?w.contains(_):w.compareDocumentPosition?!!(w.compareDocumentPosition(_)&16):!1:!1}function m8(w){w=w!=null&&w.ownerDocument!=null&&w.ownerDocument.defaultView!=null?w.ownerDocument.defaultView:window;for(var _=A2(w.document);_ instanceof w.HTMLIFrameElement;){try{var C=typeof _.contentWindow.location.href=="string"}catch{C=!1}if(C)w=_.contentWindow;else break;_=A2(w.document)}return _}function A6(w){var _=w&&w.nodeName&&w.nodeName.toLowerCase();return _&&(_==="input"&&(w.type==="text"||w.type==="search"||w.type==="tel"||w.type==="url"||w.type==="password")||_==="textarea"||w.contentEditable==="true")}var PU=pv&&"documentMode"in document&&11>=document.documentMode,ak=null,yy=null,Dt=null,Rn=!1;function En(w,_,C){var j=C.window===C?C.document:C.nodeType===9?C:C.ownerDocument;Rn||ak==null||ak!==A2(j)||(j=ak,"selectionStart"in j&&A6(j)?j={start:j.selectionStart,end:j.selectionEnd}:(j=(j.ownerDocument&&j.ownerDocument.defaultView||window).getSelection(),j={anchorNode:j.anchorNode,anchorOffset:j.anchorOffset,focusNode:j.focusNode,focusOffset:j.focusOffset}),Dt&&r0(Dt,j)||(Dt=j,j=cN(yy,"onSelect"),0<j.length&&(_=new oO("onSelect","select",null,_,C),w.push({event:_,listeners:j}),_.target=ak)))}function ui(w,_){var C={};return C[w.toLowerCase()]=_.toLowerCase(),C["Webkit"+w]="webkit"+_,C["Moz"+w]="moz"+_,C}var Lo={animationend:ui("Animation","AnimationEnd"),animationiteration:ui("Animation","AnimationIteration"),animationstart:ui("Animation","AnimationStart"),transitionrun:ui("Transition","TransitionRun"),transitionstart:ui("Transition","TransitionStart"),transitioncancel:ui("Transition","TransitionCancel"),transitionend:ui("Transition","TransitionEnd")},Ws={},Nd={};pv&&(Nd=document.createElement("div").style,"AnimationEvent"in window||(delete Lo.animationend.animation,delete Lo.animationiteration.animation,delete Lo.animationstart.animation),"TransitionEvent"in window||delete Lo.transitionend.transition);function Ks(w){if(Ws[w])return Ws[w];if(!Lo[w])return w;var _=Lo[w],C;for(C in _)if(_.hasOwnProperty(C)&&C in Nd)return Ws[w]=_[C];return w}var bO=Ks("animationend"),b8=Ks("animationiteration"),Jg=Ks("animationstart"),vO=Ks("transitionrun"),LU=Ks("transitionstart"),sk=Ks("transitioncancel"),fC=Ks("transitionend"),hC=new Map,bc="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll toggle touchMove waiting wheel".split(" ");bc.push("scrollEnd");function Yb(w,_){hC.set(w,_),hv(_,[w])}var I6=new WeakMap;function bp(w,_){if(typeof w=="object"&&w!==null){var C=I6.get(w);return C!==void 0?C:(_={value:w,source:_,stack:a6(_)},I6.set(w,_),_)}return{value:w,source:_,stack:a6(_)}}var eb=[],ck=0,O6=0;function O2(){for(var w=ck,_=O6=ck=0;_<w;){var C=eb[_];eb[_++]=null;var j=eb[_];eb[_++]=null;var q=eb[_];eb[_++]=null;var te=eb[_];if(eb[_++]=null,j!==null&&q!==null){var Se=j.pending;Se===null?q.next=q:(q.next=Se.next,Se.next=q),j.pending=q}te!==0&&pC(C,q,te)}}function Xb(w,_,C,j){eb[ck++]=w,eb[ck++]=_,eb[ck++]=C,eb[ck++]=j,O6|=j,w.lanes|=j,w=w.alternate,w!==null&&(w.lanes|=j)}function uk(w,_,C,j){return Xb(w,_,C,j),gC(w)}function N2(w,_){return Xb(w,null,null,_),gC(w)}function pC(w,_,C){w.lanes|=C;var j=w.alternate;j!==null&&(j.lanes|=C);for(var q=!1,te=w.return;te!==null;)te.childLanes|=C,j=te.alternate,j!==null&&(j.childLanes|=C),te.tag===22&&(w=te.stateNode,w===null||w._visibility&1||(q=!0)),w=te,te=te.return;return w.tag===3?(te=w.stateNode,q&&_!==null&&(q=31-Ei(C),w=te.hiddenUpdates,j=w[q],j===null?w[q]=[_]:j.push(_),_.lane=C|536870912),te):null}function gC(w){if(50<Ok)throw Ok=0,_5=null,Error(s(185));for(var _=w.return;_!==null;)w=_,_=w.return;return w.tag===3?w.stateNode:null}var R2={};function wO(w,_,C,j){this.tag=w,this.key=C,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=_,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=j,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function ng(w,_,C,j){return new wO(w,_,C,j)}function P2(w){return w=w.prototype,!(!w||!w.isReactComponent)}function vp(w,_){var C=w.alternate;return C===null?(C=ng(w.tag,_,w.key,w.mode),C.elementType=w.elementType,C.type=w.type,C.stateNode=w.stateNode,C.alternate=w,w.alternate=C):(C.pendingProps=_,C.type=w.type,C.flags=0,C.subtreeFlags=0,C.deletions=null),C.flags=w.flags&65011712,C.childLanes=w.childLanes,C.lanes=w.lanes,C.child=w.child,C.memoizedProps=w.memoizedProps,C.memoizedState=w.memoizedState,C.updateQueue=w.updateQueue,_=w.dependencies,C.dependencies=_===null?null:{lanes:_.lanes,firstContext:_.firstContext},C.sibling=w.sibling,C.index=w.index,C.ref=w.ref,C.refCleanup=w.refCleanup,C}function mC(w,_){w.flags&=65011714;var C=w.alternate;return C===null?(w.childLanes=0,w.lanes=_,w.child=null,w.subtreeFlags=0,w.memoizedProps=null,w.memoizedState=null,w.updateQueue=null,w.dependencies=null,w.stateNode=null):(w.childLanes=C.childLanes,w.lanes=C.lanes,w.child=C.child,w.subtreeFlags=0,w.deletions=null,w.memoizedProps=C.memoizedProps,w.memoizedState=C.memoizedState,w.updateQueue=C.updateQueue,w.type=C.type,_=C.dependencies,w.dependencies=_===null?null:{lanes:_.lanes,firstContext:_.firstContext}),w}function yO(w,_,C,j,q,te){var Se=0;if(j=w,typeof w=="function")P2(w)&&(Se=1);else if(typeof w=="string")Se=cz(w,C,$t.current)?26:w==="html"||w==="head"||w==="body"?27:5;else e:switch(w){case we:return w=ng(31,C,_,q),w.elementType=we,w.lanes=te,w;case D:return L2(C.children,q,te,_);case $:Se=8,q|=24;break;case U:return w=ng(12,C,_,q|2),w.elementType=U,w.lanes=te,w;case ae:return w=ng(13,C,_,q),w.elementType=ae,w.lanes=te,w;case oe:return w=ng(19,C,_,q),w.elementType=oe,w.lanes=te,w;default:if(typeof w=="object"&&w!==null)switch(w.$$typeof){case B:case K:Se=10;break e;case G:Se=9;break e;case X:Se=11;break e;case he:Se=14;break e;case ye:Se=16,j=null;break e}Se=29,C=Error(s(130,w===null?"null":typeof w,"")),j=null}return _=ng(Se,C,_,q),_.elementType=w,_.type=j,_.lanes=te,_}function L2(w,_,C,j){return w=ng(7,w,j,_),w.lanes=C,w}function N6(w,_,C){return w=ng(6,w,null,_),w.lanes=C,w}function lk(w,_,C){return _=ng(4,w.children!==null?w.children:[],w.key,_),_.lanes=C,_.stateNode={containerInfo:w.containerInfo,pendingChildren:null,implementation:w.implementation},_}var bv=[],Ey=0,EO=null,bC=0,tb=[],nb=0,qu=null,sl=1,vv="";function D2(w,_){bv[Ey++]=bC,bv[Ey++]=EO,EO=w,bC=_}function R6(w,_,C){tb[nb++]=sl,tb[nb++]=vv,tb[nb++]=qu,qu=w;var j=sl;w=vv;var q=32-Ei(j)-1;j&=~(1<<q),C+=1;var te=32-Ei(_)+q;if(30<te){var Se=q-q%5;te=(j&(1<<Se)-1).toString(32),j>>=Se,q-=Se,sl=1<<32-Ei(_)+q|C<<q|j,vv=te+w}else sl=1<<te|C<<q|j,vv=w}function P6(w){w.return!==null&&(D2(w,1),R6(w,1,0))}function vC(w){for(;w===EO;)EO=bv[--Ey],bv[Ey]=null,bC=bv[--Ey],bv[Ey]=null;for(;w===qu;)qu=tb[--nb],tb[nb]=null,vv=tb[--nb],tb[nb]=null,sl=tb[--nb],tb[nb]=null}var Rd=null,cl=null,vs=!1,Jb=null,rg=!1,dk=Error(s(519));function rb(w){var _=Error(s(418,""));throw Sy(bp(_,w)),dk}function v8(w){var _=w.stateNode,C=w.type,j=w.memoizedProps;switch(_[_l]=w,_[rh]=j,C){case"dialog":ss("cancel",_),ss("close",_);break;case"iframe":case"object":case"embed":ss("load",_);break;case"video":case"audio":for(C=0;C<s3.length;C++)ss(s3[C],_);break;case"source":ss("error",_);break;case"img":case"image":case"link":ss("error",_),ss("load",_);break;case"details":ss("toggle",_);break;case"input":ss("invalid",_),l6(_,j.value,j.defaultValue,j.checked,j.defaultChecked,j.type,j.name,!0),eO(_);break;case"select":ss("invalid",_);break;case"textarea":ss("invalid",_),nO(_,j.value,j.defaultValue,j.children),eO(_)}C=j.children,typeof C!="string"&&typeof C!="number"&&typeof C!="bigint"||_.textContent===""+C||j.suppressHydrationWarning===!0||M9(_.textContent,C)?(j.popover!=null&&(ss("beforetoggle",_),ss("toggle",_)),j.onScroll!=null&&ss("scroll",_),j.onScrollEnd!=null&&ss("scrollend",_),j.onClick!=null&&(_.onclick=uN),_=!0):_=!1,_||rb(w)}function w8(w){for(Rd=w.return;Rd;)switch(Rd.tag){case 5:case 13:rg=!1;return;case 27:case 3:rg=!0;return;default:Rd=Rd.return}}function wC(w){if(w!==Rd)return!1;if(!vs)return w8(w),vs=!0,!1;var _=w.tag,C;if((C=_!==3&&_!==27)&&((C=_===5)&&(C=w.type,C=!(C!=="form"&&C!=="button")||dN(w.type,w.memoizedProps)),C=!C),C&&cl&&rb(w),w8(w),_===13){if(w=w.memoizedState,w=w!==null?w.dehydrated:null,!w)throw Error(s(317));e:{for(w=w.nextSibling,_=0;w;){if(w.nodeType===8)if(C=w.data,C==="/$"){if(_===0){cl=o1(w.nextSibling);break e}_--}else C!=="$"&&C!=="$!"&&C!=="$?"||_++;w=w.nextSibling}cl=null}}else _===27?(_=cl,cb(w.type)?(w=j5,j5=null,cl=w):cl=_):cl=Rd?o1(w.stateNode.nextSibling):null;return!0}function yC(){cl=Rd=null,vs=!1}function y8(){var w=Jb;return w!==null&&(fg===null?fg=w:fg.push.apply(fg,w),Jb=null),w}function Sy(w){Jb===null?Jb=[w]:Jb.push(w)}var M2=We(null),_y=null,De=null;function Lh(w,_,C){Ie(M2,_._currentValue),_._currentValue=C}function wp(w){w._currentValue=M2.current,ht(M2)}function ig(w,_,C){for(;w!==null;){var j=w.alternate;if((w.childLanes&_)!==_?(w.childLanes|=_,j!==null&&(j.childLanes|=_)):j!==null&&(j.childLanes&_)!==_&&(j.childLanes|=_),w===C)break;w=w.return}}function ts(w,_,C,j){var q=w.child;for(q!==null&&(q.return=w);q!==null;){var te=q.dependencies;if(te!==null){var Se=q.child;te=te.firstContext;e:for(;te!==null;){var Me=te;te=q;for(var pt=0;pt<_.length;pt++)if(Me.context===_[pt]){te.lanes|=C,Me=te.alternate,Me!==null&&(Me.lanes|=C),ig(te.return,C,w),j||(Se=null);break e}te=Me.next}}else if(q.tag===18){if(Se=q.return,Se===null)throw Error(s(341));Se.lanes|=C,te=Se.alternate,te!==null&&(te.lanes|=C),ig(Se,C,w),Se=null}else Se=q.child;if(Se!==null)Se.return=q;else for(Se=q;Se!==null;){if(Se===w){Se=null;break}if(q=Se.sibling,q!==null){q.return=Se.return,Se=q;break}Se=Se.return}q=Se}}function fk(w,_,C,j){w=null;for(var q=_,te=!1;q!==null;){if(!te){if((q.flags&524288)!==0)te=!0;else if((q.flags&262144)!==0)break}if(q.tag===10){var Se=q.alternate;if(Se===null)throw Error(s(387));if(Se=Se.memoizedProps,Se!==null){var Me=q.type;Xg(q.pendingProps.value,Se.value)||(w!==null?w.push(Me):w=[Me])}}else if(q===Or.current){if(Se=q.alternate,Se===null)throw Error(s(387));Se.memoizedState.memoizedState!==q.memoizedState.memoizedState&&(w!==null?w.push(Mk):w=[Mk])}q=q.return}w!==null&&ts(_,w,C,j),_.flags|=262144}function SO(w){for(w=w.firstContext;w!==null;){if(!Xg(w.context._currentValue,w.memoizedValue))return!0;w=w.next}return!1}function j2(w){_y=w,De=null,w=w.dependencies,w!==null&&(w.firstContext=null)}function Hl(w){return E8(_y,w)}function _O(w,_){return _y===null&&j2(w),E8(w,_)}function E8(w,_){var C=_._currentValue;if(_={context:_,memoizedValue:C,next:null},De===null){if(w===null)throw Error(s(308));De=_,w.dependencies={lanes:0,firstContext:_},w.flags|=524288}else De=De.next=_;return C}var S8=typeof AbortController<"u"?AbortController:function(){var w=[],_=this.signal={aborted:!1,addEventListener:function(C,j){w.push(j)}};this.abort=function(){_.aborted=!0,w.forEach(function(C){return C()})}},_8=n.unstable_scheduleCallback,DU=n.unstable_NormalPriority,Qc={$$typeof:K,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function kO(){return{controller:new S8,data:new Map,refCount:0}}function EC(w){w.refCount--,w.refCount===0&&_8(DU,function(){w.controller.abort()})}var SC=null,L6=0,hk=0,og=null;function MU(w,_){if(SC===null){var C=SC=[];L6=0,hk=aN(),og={status:"pending",value:void 0,then:function(j){C.push(j)}}}return L6++,_.then(k8,k8),_}function k8(){if(--L6===0&&SC!==null){og!==null&&(og.status="fulfilled");var w=SC;SC=null,hk=0,og=null;for(var _=0;_<w.length;_++)(0,w[_])()}}function jU(w,_){var C=[],j={status:"pending",value:null,reason:null,then:function(q){C.push(q)}};return w.then(function(){j.status="fulfilled",j.value=_;for(var q=0;q<C.length;q++)(0,C[q])(_)},function(q){for(j.status="rejected",j.reason=q,q=0;q<C.length;q++)(0,C[q])(void 0)}),j}var T8=_e.S;_e.S=function(w,_){typeof _=="object"&&_!==null&&typeof _.then=="function"&&MU(w,_),T8!==null&&T8(w,_)};var $2=We(null);function wv(){var w=$2.current;return w!==null?w:Nu.pooledCache}function _C(w,_){_===null?Ie($2,$2.current):Ie($2,_.pool)}function x8(){var w=wv();return w===null?null:{parent:Qc._currentValue,pool:w}}var yp=Error(s(460)),C8=Error(s(474)),TO=Error(s(542)),D6={then:function(){}};function A8(w){return w=w.status,w==="fulfilled"||w==="rejected"}function xO(){}function M6(w,_,C){switch(C=w[C],C===void 0?w.push(_):C!==_&&(_.then(xO,xO),_=C),_.status){case"fulfilled":return _.value;case"rejected":throw w=_.reason,O8(w),w;default:if(typeof _.status=="string")_.then(xO,xO);else{if(w=Nu,w!==null&&100<w.shellSuspendCounter)throw Error(s(482));w=_,w.status="pending",w.then(function(j){if(_.status==="pending"){var q=_;q.status="fulfilled",q.value=j}},function(j){if(_.status==="pending"){var q=_;q.status="rejected",q.reason=j}})}switch(_.status){case"fulfilled":return _.value;case"rejected":throw w=_.reason,O8(w),w}throw pk=_,yp}}var pk=null;function I8(){if(pk===null)throw Error(s(459));var w=pk;return pk=null,w}function O8(w){if(w===yp||w===TO)throw Error(s(483))}var i0=!1;function F2(w){w.updateQueue={baseState:w.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function gk(w,_){w=w.updateQueue,_.updateQueue===w&&(_.updateQueue={baseState:w.baseState,firstBaseUpdate:w.firstBaseUpdate,lastBaseUpdate:w.lastBaseUpdate,shared:w.shared,callbacks:null})}function ky(w){return{lane:w,tag:0,payload:null,callback:null,next:null}}function yv(w,_,C){var j=w.updateQueue;if(j===null)return null;if(j=j.shared,(zc&2)!==0){var q=j.pending;return q===null?_.next=_:(_.next=q.next,q.next=_),j.pending=_,_=gC(w),pC(w,null,C),_}return Xb(w,j,_,C),gC(w)}function kC(w,_,C){if(_=_.updateQueue,_!==null&&(_=_.shared,(C&4194048)!==0)){var j=_.lanes;j&=w.pendingLanes,C|=j,_.lanes=C,_2(w,C)}}function Zb(w,_){var C=w.updateQueue,j=w.alternate;if(j!==null&&(j=j.updateQueue,C===j)){var q=null,te=null;if(C=C.firstBaseUpdate,C!==null){do{var Se={lane:C.lane,tag:C.tag,payload:C.payload,callback:null,next:null};te===null?q=te=Se:te=te.next=Se,C=C.next}while(C!==null);te===null?q=te=_:te=te.next=_}else q=te=_;C={baseState:j.baseState,firstBaseUpdate:q,lastBaseUpdate:te,shared:j.shared,callbacks:j.callbacks},w.updateQueue=C;return}w=C.lastBaseUpdate,w===null?C.firstBaseUpdate=_:w.next=_,C.lastBaseUpdate=_}var TC=!1;function Qb(){if(TC){var w=og;if(w!==null)throw w}}function Dh(w,_,C,j){TC=!1;var q=w.updateQueue;i0=!1;var te=q.firstBaseUpdate,Se=q.lastBaseUpdate,Me=q.shared.pending;if(Me!==null){q.shared.pending=null;var pt=Me,Xt=pt.next;pt.next=null,Se===null?te=Xt:Se.next=Xt,Se=pt;var $n=w.alternate;$n!==null&&($n=$n.updateQueue,Me=$n.lastBaseUpdate,Me!==Se&&(Me===null?$n.firstBaseUpdate=Xt:Me.next=Xt,$n.lastBaseUpdate=pt))}if(te!==null){var Vn=q.baseState;Se=0,$n=Xt=pt=null,Me=te;do{var ln=Me.lane&-536870913,Zt=ln!==Me.lane;if(Zt?($s&ln)===ln:(j&ln)===ln){ln!==0&&ln===hk&&(TC=!0),$n!==null&&($n=$n.next={lane:0,tag:Me.tag,payload:Me.payload,callback:null,next:null});e:{var Do=w,Oo=Me;ln=_;var mu=C;switch(Oo.tag){case 1:if(Do=Oo.payload,typeof Do=="function"){Vn=Do.call(mu,Vn,ln);break e}Vn=Do;break e;case 3:Do.flags=Do.flags&-65537|128;case 0:if(Do=Oo.payload,ln=typeof Do=="function"?Do.call(mu,Vn,ln):Do,ln==null)break e;Vn=k({},Vn,ln);break e;case 2:i0=!0}}ln=Me.callback,ln!==null&&(w.flags|=64,Zt&&(w.flags|=8192),Zt=q.callbacks,Zt===null?q.callbacks=[ln]:Zt.push(ln))}else Zt={lane:ln,tag:Me.tag,payload:Me.payload,callback:Me.callback,next:null},$n===null?(Xt=$n=Zt,pt=Vn):$n=$n.next=Zt,Se|=ln;if(Me=Me.next,Me===null){if(Me=q.shared.pending,Me===null)break;Zt=Me,Me=Zt.next,Zt.next=null,q.lastBaseUpdate=Zt,q.shared.pending=null}}while(!0);$n===null&&(pt=Vn),q.baseState=pt,q.firstBaseUpdate=Xt,q.lastBaseUpdate=$n,te===null&&(q.shared.lanes=0),Py|=Se,w.lanes=Se,w.memoizedState=Vn}}function N8(w,_){if(typeof w!="function")throw Error(s(191,w));w.call(_)}function j6(w,_){var C=w.callbacks;if(C!==null)for(w.callbacks=null,w=0;w<C.length;w++)N8(C[w],_)}var B2=We(null),Ty=We(0);function Ev(w,_){w=Rv,Ie(Ty,w),Ie(B2,_),Rv=w|_.baseLanes}function CO(){Ie(Ty,Rv),Ie(B2,B2.current)}function $6(){Rv=Ty.current,ht(B2),ht(Ty)}var e1=0,Wa=null,Uc=null,ws=null,mk=!1,U2=!1,Mh=!1,bk=0,xy=0,ib=null,R8=0;function Cs(){throw Error(s(321))}function xC(w,_){if(_===null)return!1;for(var C=0;C<_.length&&C<w.length;C++)if(!Xg(w[C],_[C]))return!1;return!0}function F6(w,_,C,j,q,te){return e1=te,Wa=_,_.memoizedState=null,_.updateQueue=null,_.lanes=0,_e.H=w===null||w.memoizedState===null?G8:r5,Mh=!1,te=C(j,q),Mh=!1,U2&&(te=L8(_,C,j,q)),P8(w),te}function P8(w){_e.H=jC;var _=Uc!==null&&Uc.next!==null;if(e1=0,ws=Uc=Wa=null,mk=!1,xy=0,ib=null,_)throw Error(s(300));w===null||xf||(w=w.dependencies,w!==null&&SO(w)&&(xf=!0))}function L8(w,_,C,j){Wa=w;var q=0;do{if(U2&&(ib=null),xy=0,U2=!1,25<=q)throw Error(s(301));if(q+=1,ws=Uc=null,w.updateQueue!=null){var te=w.updateQueue;te.lastEffect=null,te.events=null,te.stores=null,te.memoCache!=null&&(te.memoCache.index=0)}_e.H=q8,te=_(C,j)}while(U2);return te}function B6(){var w=_e.H,_=w.useState()[0];return _=typeof _.then=="function"?CC(_):_,w=w.useState()[0],(Uc!==null?Uc.memoizedState:null)!==w&&(Wa.flags|=1024),_}function U6(){var w=bk!==0;return bk=0,w}function z2(w,_,C){_.updateQueue=w.updateQueue,_.flags&=-2053,w.lanes&=~C}function AO(w){if(mk){for(w=w.memoizedState;w!==null;){var _=w.queue;_!==null&&(_.pending=null),w=w.next}mk=!1}e1=0,ws=Uc=Wa=null,U2=!1,xy=bk=0,ib=null}function ag(){var w={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return ws===null?Wa.memoizedState=ws=w:ws=ws.next=w,ws}function Pd(){if(Uc===null){var w=Wa.alternate;w=w!==null?w.memoizedState:null}else w=Uc.next;var _=ws===null?Wa.memoizedState:ws.next;if(_!==null)ws=_,Uc=w;else{if(w===null)throw Wa.alternate===null?Error(s(467)):Error(s(310));Uc=w,w={memoizedState:Uc.memoizedState,baseState:Uc.baseState,baseQueue:Uc.baseQueue,queue:Uc.queue,next:null},ws===null?Wa.memoizedState=ws=w:ws=ws.next=w}return ws}function H2(){return{lastEffect:null,events:null,stores:null,memoCache:null}}function CC(w){var _=xy;return xy+=1,ib===null&&(ib=[]),w=M6(ib,w,_),_=Wa,(ws===null?_.memoizedState:ws.next)===null&&(_=_.alternate,_e.H=_===null||_.memoizedState===null?G8:r5),w}function IO(w){if(w!==null&&typeof w=="object"){if(typeof w.then=="function")return CC(w);if(w.$$typeof===K)return Hl(w)}throw Error(s(438,String(w)))}function AC(w){var _=null,C=Wa.updateQueue;if(C!==null&&(_=C.memoCache),_==null){var j=Wa.alternate;j!==null&&(j=j.updateQueue,j!==null&&(j=j.memoCache,j!=null&&(_={data:j.data.map(function(q){return q.slice()}),index:0})))}if(_==null&&(_={data:[],index:0}),C===null&&(C=H2(),Wa.updateQueue=C),C.memoCache=_,C=_.data[_.index],C===void 0)for(C=_.data[_.index]=Array(w),j=0;j<w;j++)C[j]=ge;return _.index++,C}function o0(w,_){return typeof _=="function"?_(w):_}function OO(w){var _=Pd();return z6(_,Uc,w)}function z6(w,_,C){var j=w.queue;if(j===null)throw Error(s(311));j.lastRenderedReducer=C;var q=w.baseQueue,te=j.pending;if(te!==null){if(q!==null){var Se=q.next;q.next=te.next,te.next=Se}_.baseQueue=q=te,j.pending=null}if(te=w.baseState,q===null)w.memoizedState=te;else{_=q.next;var Me=Se=null,pt=null,Xt=_,$n=!1;do{var Vn=Xt.lane&-536870913;if(Vn!==Xt.lane?($s&Vn)===Vn:(e1&Vn)===Vn){var ln=Xt.revertLane;if(ln===0)pt!==null&&(pt=pt.next={lane:0,revertLane:0,action:Xt.action,hasEagerState:Xt.hasEagerState,eagerState:Xt.eagerState,next:null}),Vn===hk&&($n=!0);else if((e1&ln)===ln){Xt=Xt.next,ln===hk&&($n=!0);continue}else Vn={lane:0,revertLane:Xt.revertLane,action:Xt.action,hasEagerState:Xt.hasEagerState,eagerState:Xt.eagerState,next:null},pt===null?(Me=pt=Vn,Se=te):pt=pt.next=Vn,Wa.lanes|=ln,Py|=ln;Vn=Xt.action,Mh&&C(te,Vn),te=Xt.hasEagerState?Xt.eagerState:C(te,Vn)}else ln={lane:Vn,revertLane:Xt.revertLane,action:Xt.action,hasEagerState:Xt.hasEagerState,eagerState:Xt.eagerState,next:null},pt===null?(Me=pt=ln,Se=te):pt=pt.next=ln,Wa.lanes|=Vn,Py|=Vn;Xt=Xt.next}while(Xt!==null&&Xt!==_);if(pt===null?Se=te:pt.next=Me,!Xg(te,w.memoizedState)&&(xf=!0,$n&&(C=og,C!==null)))throw C;w.memoizedState=te,w.baseState=Se,w.baseQueue=pt,j.lastRenderedState=te}return q===null&&(j.lanes=0),[w.memoizedState,j.dispatch]}function H6(w){var _=Pd(),C=_.queue;if(C===null)throw Error(s(311));C.lastRenderedReducer=w;var j=C.dispatch,q=C.pending,te=_.memoizedState;if(q!==null){C.pending=null;var Se=q=q.next;do te=w(te,Se.action),Se=Se.next;while(Se!==q);Xg(te,_.memoizedState)||(xf=!0),_.memoizedState=te,_.baseQueue===null&&(_.baseState=te),C.lastRenderedState=te}return[te,j]}function G6(w,_,C){var j=Wa,q=Pd(),te=vs;if(te){if(C===void 0)throw Error(s(407));C=C()}else C=_();var Se=!Xg((Uc||q).memoizedState,C);Se&&(q.memoizedState=C,xf=!0),q=q.queue;var Me=RO.bind(null,j,q,w);if(NC(2048,8,Me,[w]),q.getSnapshot!==_||Se||ws!==null&&ws.memoizedState.tag&1){if(j.flags|=2048,vk(9,jO(),D8.bind(null,j,q,C,_),null),Nu===null)throw Error(s(349));te||(e1&124)!==0||NO(j,_,C)}return C}function NO(w,_,C){w.flags|=16384,w={getSnapshot:_,value:C},_=Wa.updateQueue,_===null?(_=H2(),Wa.updateQueue=_,_.stores=[w]):(C=_.stores,C===null?_.stores=[w]:C.push(w))}function D8(w,_,C,j){_.value=C,_.getSnapshot=j,q6(_)&&V6(w)}function RO(w,_,C){return C(function(){q6(_)&&V6(w)})}function q6(w){var _=w.getSnapshot;w=w.value;try{var C=_();return!Xg(w,C)}catch{return!0}}function V6(w){var _=N2(w,2);_!==null&&rm(_,w,2)}function PO(w){var _=ag();if(typeof w=="function"){var C=w;if(w=C(),Mh){Kr(!0);try{C()}finally{Kr(!1)}}}return _.memoizedState=_.baseState=w,_.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:o0,lastRenderedState:w},_}function IC(w,_,C,j){return w.baseState=C,z6(w,Uc,typeof j=="function"?j:o0)}function $U(w,_,C,j,q){if(UO(w))throw Error(s(485));if(w=_.action,w!==null){var te={payload:q,action:w,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(Se){te.listeners.push(Se)}};_e.T!==null?C(!0):te.isTransition=!1,j(te),C=_.pending,C===null?(te.next=_.pending=te,W6(_,te)):(te.next=C.next,_.pending=C.next=te)}}function W6(w,_){var C=_.action,j=_.payload,q=w.state;if(_.isTransition){var te=_e.T,Se={};_e.T=Se;try{var Me=C(q,j),pt=_e.S;pt!==null&&pt(Se,Me),K6(w,_,Me)}catch(Xt){LO(w,_,Xt)}finally{_e.T=te}}else try{te=C(q,j),K6(w,_,te)}catch(Xt){LO(w,_,Xt)}}function K6(w,_,C){C!==null&&typeof C=="object"&&typeof C.then=="function"?C.then(function(j){M8(w,_,j)},function(j){return LO(w,_,j)}):M8(w,_,C)}function M8(w,_,C){_.status="fulfilled",_.value=C,j8(_),w.state=C,_=w.pending,_!==null&&(C=_.next,C===_?w.pending=null:(C=C.next,_.next=C,W6(w,C)))}function LO(w,_,C){var j=w.pending;if(w.pending=null,j!==null){j=j.next;do _.status="rejected",_.reason=C,j8(_),_=_.next;while(_!==j)}w.action=null}function j8(w){w=w.listeners;for(var _=0;_<w.length;_++)(0,w[_])()}function DO(w,_){return _}function Y6(w,_){if(vs){var C=Nu.formState;if(C!==null){e:{var j=Wa;if(vs){if(cl){t:{for(var q=cl,te=rg;q.nodeType!==8;){if(!te){q=null;break t}if(q=o1(q.nextSibling),q===null){q=null;break t}}te=q.data,q=te==="F!"||te==="F"?q:null}if(q){cl=o1(q.nextSibling),j=q.data==="F!";break e}}rb(j)}j=!1}j&&(_=C[0])}}return C=ag(),C.memoizedState=C.baseState=_,j={pending:null,lanes:0,dispatch:null,lastRenderedReducer:DO,lastRenderedState:_},C.queue=j,C=e5.bind(null,Wa,j),j.dispatch=C,j=PO(!1),te=t5.bind(null,Wa,!1,j.queue),j=ag(),q={state:_,dispatch:null,action:w,pending:null},j.queue=q,C=$U.bind(null,Wa,q,te,C),q.dispatch=C,j.memoizedState=w,[_,C,!1]}function MO(w){var _=Pd();return $8(_,Uc,w)}function $8(w,_,C){if(_=z6(w,_,DO)[0],w=OO(o0)[0],typeof _=="object"&&_!==null&&typeof _.then=="function")try{var j=CC(_)}catch(Se){throw Se===yp?TO:Se}else j=_;_=Pd();var q=_.queue,te=q.dispatch;return C!==_.memoizedState&&(Wa.flags|=2048,vk(9,jO(),X6.bind(null,q,C),null)),[j,te,w]}function X6(w,_){w.action=_}function Da(w){var _=Pd(),C=Uc;if(C!==null)return $8(_,C,w);Pd(),_=_.memoizedState,C=Pd();var j=C.queue.dispatch;return C.memoizedState=w,[_,j,!1]}function vk(w,_,C,j){return w={tag:w,create:C,deps:j,inst:_,next:null},_=Wa.updateQueue,_===null&&(_=H2(),Wa.updateQueue=_),C=_.lastEffect,C===null?_.lastEffect=w.next=w:(j=C.next,C.next=w,w.next=j,_.lastEffect=w),w}function jO(){return{destroy:void 0,resource:void 0}}function wk(){return Pd().memoizedState}function OC(w,_,C,j){var q=ag();j=j===void 0?null:j,Wa.flags|=w,q.memoizedState=vk(1|_,jO(),C,j)}function NC(w,_,C,j){var q=Pd();j=j===void 0?null:j;var te=q.memoizedState.inst;Uc!==null&&j!==null&&xC(j,Uc.memoizedState.deps)?q.memoizedState=vk(_,te,C,j):(Wa.flags|=w,q.memoizedState=vk(1|_,te,C,j))}function F8(w,_){OC(8390656,8,w,_)}function Zg(w,_){NC(2048,8,w,_)}function FU(w,_){return NC(4,2,w,_)}function RC(w,_){return NC(4,4,w,_)}function $O(w,_){if(typeof _=="function"){w=w();var C=_(w);return function(){typeof C=="function"?C():_(null)}}if(_!=null)return w=w(),_.current=w,function(){_.current=null}}function J6(w,_,C){C=C!=null?C.concat([w]):null,NC(4,4,$O.bind(null,_,w),C)}function FO(){}function sg(w,_){var C=Pd();_=_===void 0?null:_;var j=C.memoizedState;return _!==null&&xC(_,j[1])?j[0]:(C.memoizedState=[w,_],w)}function Sv(w,_){var C=Pd();_=_===void 0?null:_;var j=C.memoizedState;if(_!==null&&xC(_,j[1]))return j[0];if(j=w(),Mh){Kr(!0);try{w()}finally{Kr(!1)}}return C.memoizedState=[j,_],j}function PC(w,_,C){return C===void 0||(e1&1073741824)!==0?w.memoizedState=_:(w.memoizedState=C,w=f9(),Wa.lanes|=w,Py|=w,C)}function Z6(w,_,C,j){return Xg(C,_)?C:B2.current!==null?(w=PC(w,C,j),Xg(w,_)||(xf=!0),w):(e1&42)===0?(xf=!0,w.memoizedState=C):(w=f9(),Wa.lanes|=w,Py|=w,_)}function LC(w,_,C,j,q){var te=qe.p;qe.p=te!==0&&8>te?te:8;var Se=_e.T,Me={};_e.T=Me,t5(w,!1,_,C);try{var pt=q(),Xt=_e.S;if(Xt!==null&&Xt(Me,pt),pt!==null&&typeof pt=="object"&&typeof pt.then=="function"){var $n=jU(pt,j);G2(w,_,$n,$h(w))}else G2(w,_,j,$h(w))}catch(Vn){G2(w,_,{then:function(){},status:"rejected",reason:Vn},$h())}finally{qe.p=te,_e.T=Se}}function B8(){}function DC(w,_,C,j){if(w.tag!==5)throw Error(s(476));var q=Q6(w).queue;LC(w,q,_,it,C===null?B8:function(){return BO(w),C(j)})}function Q6(w){var _=w.memoizedState;if(_!==null)return _;_={memoizedState:it,baseState:it,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:o0,lastRenderedState:it},next:null};var C={};return _.next={memoizedState:C,baseState:C,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:o0,lastRenderedState:C},next:null},w.memoizedState=_,w=w.alternate,w!==null&&(w.memoizedState=_),_}function BO(w){var _=Q6(w).next.queue;G2(w,_,{},$h())}function MC(){return Hl(Mk)}function _v(){return Pd().memoizedState}function U8(){return Pd().memoizedState}function Cy(w){for(var _=w.return;_!==null;){switch(_.tag){case 24:case 3:var C=$h();w=ky(C);var j=yv(_,w,C);j!==null&&(rm(j,_,C),kC(j,_,C)),_={cache:kO()},w.payload=_;return}_=_.return}}function z8(w,_,C){var j=$h();C={lane:j,revertLane:0,action:C,hasEagerState:!1,eagerState:null,next:null},UO(w)?H8(_,C):(C=uk(w,_,C,j),C!==null&&(rm(C,w,j),n5(C,_,j)))}function e5(w,_,C){var j=$h();G2(w,_,C,j)}function G2(w,_,C,j){var q={lane:j,revertLane:0,action:C,hasEagerState:!1,eagerState:null,next:null};if(UO(w))H8(_,q);else{var te=w.alternate;if(w.lanes===0&&(te===null||te.lanes===0)&&(te=_.lastRenderedReducer,te!==null))try{var Se=_.lastRenderedState,Me=te(Se,C);if(q.hasEagerState=!0,q.eagerState=Me,Xg(Me,Se))return Xb(w,_,q,0),Nu===null&&O2(),!1}catch{}finally{}if(C=uk(w,_,q,j),C!==null)return rm(C,w,j),n5(C,_,j),!0}return!1}function t5(w,_,C,j){if(j={lane:2,revertLane:aN(),action:j,hasEagerState:!1,eagerState:null,next:null},UO(w)){if(_)throw Error(s(479))}else _=uk(w,C,j,2),_!==null&&rm(_,w,2)}function UO(w){var _=w.alternate;return w===Wa||_!==null&&_===Wa}function H8(w,_){U2=mk=!0;var C=w.pending;C===null?_.next=_:(_.next=C.next,C.next=_),w.pending=_}function n5(w,_,C){if((C&4194048)!==0){var j=_.lanes;j&=w.pendingLanes,C|=j,_.lanes=C,_2(w,C)}}var jC={readContext:Hl,use:IO,useCallback:Cs,useContext:Cs,useEffect:Cs,useImperativeHandle:Cs,useLayoutEffect:Cs,useInsertionEffect:Cs,useMemo:Cs,useReducer:Cs,useRef:Cs,useState:Cs,useDebugValue:Cs,useDeferredValue:Cs,useTransition:Cs,useSyncExternalStore:Cs,useId:Cs,useHostTransitionStatus:Cs,useFormState:Cs,useActionState:Cs,useOptimistic:Cs,useMemoCache:Cs,useCacheRefresh:Cs},G8={readContext:Hl,use:IO,useCallback:function(w,_){return ag().memoizedState=[w,_===void 0?null:_],w},useContext:Hl,useEffect:F8,useImperativeHandle:function(w,_,C){C=C!=null?C.concat([w]):null,OC(4194308,4,$O.bind(null,_,w),C)},useLayoutEffect:function(w,_){return OC(4194308,4,w,_)},useInsertionEffect:function(w,_){OC(4,2,w,_)},useMemo:function(w,_){var C=ag();_=_===void 0?null:_;var j=w();if(Mh){Kr(!0);try{w()}finally{Kr(!1)}}return C.memoizedState=[j,_],j},useReducer:function(w,_,C){var j=ag();if(C!==void 0){var q=C(_);if(Mh){Kr(!0);try{C(_)}finally{Kr(!1)}}}else q=_;return j.memoizedState=j.baseState=q,w={pending:null,lanes:0,dispatch:null,lastRenderedReducer:w,lastRenderedState:q},j.queue=w,w=w.dispatch=z8.bind(null,Wa,w),[j.memoizedState,w]},useRef:function(w){var _=ag();return w={current:w},_.memoizedState=w},useState:function(w){w=PO(w);var _=w.queue,C=e5.bind(null,Wa,_);return _.dispatch=C,[w.memoizedState,C]},useDebugValue:FO,useDeferredValue:function(w,_){var C=ag();return PC(C,w,_)},useTransition:function(){var w=PO(!1);return w=LC.bind(null,Wa,w.queue,!0,!1),ag().memoizedState=w,[!1,w]},useSyncExternalStore:function(w,_,C){var j=Wa,q=ag();if(vs){if(C===void 0)throw Error(s(407));C=C()}else{if(C=_(),Nu===null)throw Error(s(349));($s&124)!==0||NO(j,_,C)}q.memoizedState=C;var te={value:C,getSnapshot:_};return q.queue=te,F8(RO.bind(null,j,te,w),[w]),j.flags|=2048,vk(9,jO(),D8.bind(null,j,te,C,_),null),C},useId:function(){var w=ag(),_=Nu.identifierPrefix;if(vs){var C=vv,j=sl;C=(j&~(1<<32-Ei(j)-1)).toString(32)+C,_="«"+_+"R"+C,C=bk++,0<C&&(_+="H"+C.toString(32)),_+="»"}else C=R8++,_="«"+_+"r"+C.toString(32)+"»";return w.memoizedState=_},useHostTransitionStatus:MC,useFormState:Y6,useActionState:Y6,useOptimistic:function(w){var _=ag();_.memoizedState=_.baseState=w;var C={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return _.queue=C,_=t5.bind(null,Wa,!0,C),C.dispatch=_,[w,_]},useMemoCache:AC,useCacheRefresh:function(){return ag().memoizedState=Cy.bind(null,Wa)}},r5={readContext:Hl,use:IO,useCallback:sg,useContext:Hl,useEffect:Zg,useImperativeHandle:J6,useInsertionEffect:FU,useLayoutEffect:RC,useMemo:Sv,useReducer:OO,useRef:wk,useState:function(){return OO(o0)},useDebugValue:FO,useDeferredValue:function(w,_){var C=Pd();return Z6(C,Uc.memoizedState,w,_)},useTransition:function(){var w=OO(o0)[0],_=Pd().memoizedState;return[typeof w=="boolean"?w:CC(w),_]},useSyncExternalStore:G6,useId:_v,useHostTransitionStatus:MC,useFormState:MO,useActionState:MO,useOptimistic:function(w,_){var C=Pd();return IC(C,Uc,w,_)},useMemoCache:AC,useCacheRefresh:U8},q8={readContext:Hl,use:IO,useCallback:sg,useContext:Hl,useEffect:Zg,useImperativeHandle:J6,useInsertionEffect:FU,useLayoutEffect:RC,useMemo:Sv,useReducer:H6,useRef:wk,useState:function(){return H6(o0)},useDebugValue:FO,useDeferredValue:function(w,_){var C=Pd();return Uc===null?PC(C,w,_):Z6(C,Uc.memoizedState,w,_)},useTransition:function(){var w=H6(o0)[0],_=Pd().memoizedState;return[typeof w=="boolean"?w:CC(w),_]},useSyncExternalStore:G6,useId:_v,useHostTransitionStatus:MC,useFormState:Da,useActionState:Da,useOptimistic:function(w,_){var C=Pd();return Uc!==null?IC(C,Uc,w,_):(C.baseState=w,[w,C.queue.dispatch])},useMemoCache:AC,useCacheRefresh:U8},Kd=null,cg=0;function yk(w){var _=cg;return cg+=1,Kd===null&&(Kd=[]),M6(Kd,w,_)}function q2(w,_){_=_.props.ref,w.ref=_!==void 0?_:null}function V2(w,_){throw _.$$typeof===x?Error(s(525)):(w=Object.prototype.toString.call(_),Error(s(31,w==="[object Object]"?"object with keys {"+Object.keys(_).join(", ")+"}":w)))}function kv(w){var _=w._init;return _(w._payload)}function i5(w){function _(zt,Ot){if(w){var Wt=zt.deletions;Wt===null?(zt.deletions=[Ot],zt.flags|=16):Wt.push(Ot)}}function C(zt,Ot){if(!w)return null;for(;Ot!==null;)_(zt,Ot),Ot=Ot.sibling;return null}function j(zt){for(var Ot=new Map;zt!==null;)zt.key!==null?Ot.set(zt.key,zt):Ot.set(zt.index,zt),zt=zt.sibling;return Ot}function q(zt,Ot){return zt=vp(zt,Ot),zt.index=0,zt.sibling=null,zt}function te(zt,Ot,Wt){return zt.index=Wt,w?(Wt=zt.alternate,Wt!==null?(Wt=Wt.index,Wt<Ot?(zt.flags|=67108866,Ot):Wt):(zt.flags|=67108866,Ot)):(zt.flags|=1048576,Ot)}function Se(zt){return w&&zt.alternate===null&&(zt.flags|=67108866),zt}function Me(zt,Ot,Wt,zn){return Ot===null||Ot.tag!==6?(Ot=N6(Wt,zt.mode,zn),Ot.return=zt,Ot):(Ot=q(Ot,Wt),Ot.return=zt,Ot)}function pt(zt,Ot,Wt,zn){var bi=Wt.type;return bi===D?$n(zt,Ot,Wt.props.children,zn,Wt.key):Ot!==null&&(Ot.elementType===bi||typeof bi=="object"&&bi!==null&&bi.$$typeof===ye&&kv(bi)===Ot.type)?(Ot=q(Ot,Wt.props),q2(Ot,Wt),Ot.return=zt,Ot):(Ot=yO(Wt.type,Wt.key,Wt.props,null,zt.mode,zn),q2(Ot,Wt),Ot.return=zt,Ot)}function Xt(zt,Ot,Wt,zn){return Ot===null||Ot.tag!==4||Ot.stateNode.containerInfo!==Wt.containerInfo||Ot.stateNode.implementation!==Wt.implementation?(Ot=lk(Wt,zt.mode,zn),Ot.return=zt,Ot):(Ot=q(Ot,Wt.children||[]),Ot.return=zt,Ot)}function $n(zt,Ot,Wt,zn,bi){return Ot===null||Ot.tag!==7?(Ot=L2(Wt,zt.mode,zn,bi),Ot.return=zt,Ot):(Ot=q(Ot,Wt),Ot.return=zt,Ot)}function Vn(zt,Ot,Wt){if(typeof Ot=="string"&&Ot!==""||typeof Ot=="number"||typeof Ot=="bigint")return Ot=N6(""+Ot,zt.mode,Wt),Ot.return=zt,Ot;if(typeof Ot=="object"&&Ot!==null){switch(Ot.$$typeof){case O:return Wt=yO(Ot.type,Ot.key,Ot.props,null,zt.mode,Wt),q2(Wt,Ot),Wt.return=zt,Wt;case P:return Ot=lk(Ot,zt.mode,Wt),Ot.return=zt,Ot;case ye:var zn=Ot._init;return Ot=zn(Ot._payload),Vn(zt,Ot,Wt)}if(Re(Ot)||xe(Ot))return Ot=L2(Ot,zt.mode,Wt,null),Ot.return=zt,Ot;if(typeof Ot.then=="function")return Vn(zt,yk(Ot),Wt);if(Ot.$$typeof===K)return Vn(zt,_O(zt,Ot),Wt);V2(zt,Ot)}return null}function ln(zt,Ot,Wt,zn){var bi=Ot!==null?Ot.key:null;if(typeof Wt=="string"&&Wt!==""||typeof Wt=="number"||typeof Wt=="bigint")return bi!==null?null:Me(zt,Ot,""+Wt,zn);if(typeof Wt=="object"&&Wt!==null){switch(Wt.$$typeof){case O:return Wt.key===bi?pt(zt,Ot,Wt,zn):null;case P:return Wt.key===bi?Xt(zt,Ot,Wt,zn):null;case ye:return bi=Wt._init,Wt=bi(Wt._payload),ln(zt,Ot,Wt,zn)}if(Re(Wt)||xe(Wt))return bi!==null?null:$n(zt,Ot,Wt,zn,null);if(typeof Wt.then=="function")return ln(zt,Ot,yk(Wt),zn);if(Wt.$$typeof===K)return ln(zt,Ot,_O(zt,Wt),zn);V2(zt,Wt)}return null}function Zt(zt,Ot,Wt,zn,bi){if(typeof zn=="string"&&zn!==""||typeof zn=="number"||typeof zn=="bigint")return zt=zt.get(Wt)||null,Me(Ot,zt,""+zn,bi);if(typeof zn=="object"&&zn!==null){switch(zn.$$typeof){case O:return zt=zt.get(zn.key===null?Wt:zn.key)||null,pt(Ot,zt,zn,bi);case P:return zt=zt.get(zn.key===null?Wt:zn.key)||null,Xt(Ot,zt,zn,bi);case ye:var Ka=zn._init;return zn=Ka(zn._payload),Zt(zt,Ot,Wt,zn,bi)}if(Re(zn)||xe(zn))return zt=zt.get(Wt)||null,$n(Ot,zt,zn,bi,null);if(typeof zn.then=="function")return Zt(zt,Ot,Wt,yk(zn),bi);if(zn.$$typeof===K)return Zt(zt,Ot,Wt,_O(Ot,zn),bi);V2(Ot,zn)}return null}function Do(zt,Ot,Wt,zn){for(var bi=null,Ka=null,Hi=Ot,Vi=Ot=0,Ic=null;Hi!==null&&Vi<Wt.length;Vi++){Hi.index>Vi?(Ic=Hi,Hi=null):Ic=Hi.sibling;var Fs=ln(zt,Hi,Wt[Vi],zn);if(Fs===null){Hi===null&&(Hi=Ic);break}w&&Hi&&Fs.alternate===null&&_(zt,Hi),Ot=te(Fs,Ot,Vi),Ka===null?bi=Fs:Ka.sibling=Fs,Ka=Fs,Hi=Ic}if(Vi===Wt.length)return C(zt,Hi),vs&&D2(zt,Vi),bi;if(Hi===null){for(;Vi<Wt.length;Vi++)Hi=Vn(zt,Wt[Vi],zn),Hi!==null&&(Ot=te(Hi,Ot,Vi),Ka===null?bi=Hi:Ka.sibling=Hi,Ka=Hi);return vs&&D2(zt,Vi),bi}for(Hi=j(Hi);Vi<Wt.length;Vi++)Ic=Zt(Hi,zt,Vi,Wt[Vi],zn),Ic!==null&&(w&&Ic.alternate!==null&&Hi.delete(Ic.key===null?Vi:Ic.key),Ot=te(Ic,Ot,Vi),Ka===null?bi=Ic:Ka.sibling=Ic,Ka=Ic);return w&&Hi.forEach(function(Mv){return _(zt,Mv)}),vs&&D2(zt,Vi),bi}function Oo(zt,Ot,Wt,zn){if(Wt==null)throw Error(s(151));for(var bi=null,Ka=null,Hi=Ot,Vi=Ot=0,Ic=null,Fs=Wt.next();Hi!==null&&!Fs.done;Vi++,Fs=Wt.next()){Hi.index>Vi?(Ic=Hi,Hi=null):Ic=Hi.sibling;var Mv=ln(zt,Hi,Fs.value,zn);if(Mv===null){Hi===null&&(Hi=Ic);break}w&&Hi&&Mv.alternate===null&&_(zt,Hi),Ot=te(Mv,Ot,Vi),Ka===null?bi=Mv:Ka.sibling=Mv,Ka=Mv,Hi=Ic}if(Fs.done)return C(zt,Hi),vs&&D2(zt,Vi),bi;if(Hi===null){for(;!Fs.done;Vi++,Fs=Wt.next())Fs=Vn(zt,Fs.value,zn),Fs!==null&&(Ot=te(Fs,Ot,Vi),Ka===null?bi=Fs:Ka.sibling=Fs,Ka=Fs);return vs&&D2(zt,Vi),bi}for(Hi=j(Hi);!Fs.done;Vi++,Fs=Wt.next())Fs=Zt(Hi,zt,Vi,Fs.value,zn),Fs!==null&&(w&&Fs.alternate!==null&&Hi.delete(Fs.key===null?Vi:Fs.key),Ot=te(Fs,Ot,Vi),Ka===null?bi=Fs:Ka.sibling=Fs,Ka=Fs);return w&&Hi.forEach(function(f3){return _(zt,f3)}),vs&&D2(zt,Vi),bi}function mu(zt,Ot,Wt,zn){if(typeof Wt=="object"&&Wt!==null&&Wt.type===D&&Wt.key===null&&(Wt=Wt.props.children),typeof Wt=="object"&&Wt!==null){switch(Wt.$$typeof){case O:e:{for(var bi=Wt.key;Ot!==null;){if(Ot.key===bi){if(bi=Wt.type,bi===D){if(Ot.tag===7){C(zt,Ot.sibling),zn=q(Ot,Wt.props.children),zn.return=zt,zt=zn;break e}}else if(Ot.elementType===bi||typeof bi=="object"&&bi!==null&&bi.$$typeof===ye&&kv(bi)===Ot.type){C(zt,Ot.sibling),zn=q(Ot,Wt.props),q2(zn,Wt),zn.return=zt,zt=zn;break e}C(zt,Ot);break}else _(zt,Ot);Ot=Ot.sibling}Wt.type===D?(zn=L2(Wt.props.children,zt.mode,zn,Wt.key),zn.return=zt,zt=zn):(zn=yO(Wt.type,Wt.key,Wt.props,null,zt.mode,zn),q2(zn,Wt),zn.return=zt,zt=zn)}return Se(zt);case P:e:{for(bi=Wt.key;Ot!==null;){if(Ot.key===bi)if(Ot.tag===4&&Ot.stateNode.containerInfo===Wt.containerInfo&&Ot.stateNode.implementation===Wt.implementation){C(zt,Ot.sibling),zn=q(Ot,Wt.children||[]),zn.return=zt,zt=zn;break e}else{C(zt,Ot);break}else _(zt,Ot);Ot=Ot.sibling}zn=lk(Wt,zt.mode,zn),zn.return=zt,zt=zn}return Se(zt);case ye:return bi=Wt._init,Wt=bi(Wt._payload),mu(zt,Ot,Wt,zn)}if(Re(Wt))return Do(zt,Ot,Wt,zn);if(xe(Wt)){if(bi=xe(Wt),typeof bi!="function")throw Error(s(150));return Wt=bi.call(Wt),Oo(zt,Ot,Wt,zn)}if(typeof Wt.then=="function")return mu(zt,Ot,yk(Wt),zn);if(Wt.$$typeof===K)return mu(zt,Ot,_O(zt,Wt),zn);V2(zt,Wt)}return typeof Wt=="string"&&Wt!==""||typeof Wt=="number"||typeof Wt=="bigint"?(Wt=""+Wt,Ot!==null&&Ot.tag===6?(C(zt,Ot.sibling),zn=q(Ot,Wt),zn.return=zt,zt=zn):(C(zt,Ot),zn=N6(Wt,zt.mode,zn),zn.return=zt,zt=zn),Se(zt)):C(zt,Ot)}return function(zt,Ot,Wt,zn){try{cg=0;var bi=mu(zt,Ot,Wt,zn);return Kd=null,bi}catch(Hi){if(Hi===yp||Hi===TO)throw Hi;var Ka=ng(29,Hi,null,zt.mode);return Ka.lanes=zn,Ka.return=zt,Ka}finally{}}}var W2=i5(!0),V8=i5(!1),Qg=We(null),Ep=null;function ob(w){var _=w.alternate;Ie(Gl,Gl.current&1),Ie(Qg,w),Ep===null&&(_===null||B2.current!==null||_.memoizedState!==null)&&(Ep=w)}function W8(w){if(w.tag===22){if(Ie(Gl,Gl.current),Ie(Qg,w),Ep===null){var _=w.alternate;_!==null&&_.memoizedState!==null&&(Ep=w)}}else Tv()}function Tv(){Ie(Gl,Gl.current),Ie(Qg,Qg.current)}function a0(w){ht(Qg),Ep===w&&(Ep=null),ht(Gl)}var Gl=We(0);function $C(w){for(var _=w;_!==null;){if(_.tag===13){var C=_.memoizedState;if(C!==null&&(C=C.dehydrated,C===null||C.data==="$?"||_p(C)))return _}else if(_.tag===19&&_.memoizedProps.revealOrder!==void 0){if((_.flags&128)!==0)return _}else if(_.child!==null){_.child.return=_,_=_.child;continue}if(_===w)break;for(;_.sibling===null;){if(_.return===null||_.return===w)return null;_=_.return}_.sibling.return=_.return,_=_.sibling}return null}function K2(w,_,C,j){_=w.memoizedState,C=C(j,_),C=C==null?_:k({},_,C),w.memoizedState=C,w.lanes===0&&(w.updateQueue.baseState=C)}var FC={enqueueSetState:function(w,_,C){w=w._reactInternals;var j=$h(),q=ky(j);q.payload=_,C!=null&&(q.callback=C),_=yv(w,q,j),_!==null&&(rm(_,w,j),kC(_,w,j))},enqueueReplaceState:function(w,_,C){w=w._reactInternals;var j=$h(),q=ky(j);q.tag=1,q.payload=_,C!=null&&(q.callback=C),_=yv(w,q,j),_!==null&&(rm(_,w,j),kC(_,w,j))},enqueueForceUpdate:function(w,_){w=w._reactInternals;var C=$h(),j=ky(C);j.tag=2,_!=null&&(j.callback=_),_=yv(w,j,C),_!==null&&(rm(_,w,C),kC(_,w,C))}};function BC(w,_,C,j,q,te,Se){return w=w.stateNode,typeof w.shouldComponentUpdate=="function"?w.shouldComponentUpdate(j,te,Se):_.prototype&&_.prototype.isPureReactComponent?!r0(C,j)||!r0(q,te):!0}function BU(w,_,C,j){w=_.state,typeof _.componentWillReceiveProps=="function"&&_.componentWillReceiveProps(C,j),typeof _.UNSAFE_componentWillReceiveProps=="function"&&_.UNSAFE_componentWillReceiveProps(C,j),_.state!==w&&FC.enqueueReplaceState(_,_.state,null)}function Ay(w,_){var C=_;if("ref"in _){C={};for(var j in _)j!=="ref"&&(C[j]=_[j])}if(w=w.defaultProps){C===_&&(C=k({},C));for(var q in w)C[q]===void 0&&(C[q]=w[q])}return C}var Y2=typeof reportError=="function"?reportError:function(w){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var _=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof w=="object"&&w!==null&&typeof w.message=="string"?String(w.message):String(w),error:w});if(!window.dispatchEvent(_))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",w);return}console.error(w)};function o5(w){Y2(w)}function zO(w){console.error(w)}function K8(w){Y2(w)}function UC(w,_){try{var C=w.onUncaughtError;C(_.value,{componentStack:_.stack})}catch(j){setTimeout(function(){throw j})}}function s0(w,_,C){try{var j=w.onCaughtError;j(C.value,{componentStack:C.stack,errorBoundary:_.tag===1?_.stateNode:null})}catch(q){setTimeout(function(){throw q})}}function ug(w,_,C){return C=ky(C),C.tag=3,C.payload={element:null},C.callback=function(){UC(w,_)},C}function a5(w){return w=ky(w),w.tag=3,w}function xv(w,_,C,j){var q=C.type.getDerivedStateFromError;if(typeof q=="function"){var te=j.value;w.payload=function(){return q(te)},w.callback=function(){s0(_,C,j)}}var Se=C.stateNode;Se!==null&&typeof Se.componentDidCatch=="function"&&(w.callback=function(){s0(_,C,j),typeof q!="function"&&(Ly===null?Ly=new Set([this]):Ly.add(this));var Me=j.stack;this.componentDidCatch(j.value,{componentStack:Me!==null?Me:""})})}function s5(w,_,C,j,q){if(C.flags|=32768,j!==null&&typeof j=="object"&&typeof j.then=="function"){if(_=C.alternate,_!==null&&fk(_,C,q,!0),C=Qg.current,C!==null){switch(C.tag){case 13:return Ep===null?T5():C.alternate===null&&ql===0&&(ql=3),C.flags&=-257,C.flags|=65536,C.lanes=q,j===D6?C.flags|=16384:(_=C.updateQueue,_===null?C.updateQueue=new Set([j]):_.add(j),A5(w,j,q)),!1;case 22:return C.flags|=65536,j===D6?C.flags|=16384:(_=C.updateQueue,_===null?(_={transitions:null,markerInstances:null,retryQueue:new Set([j])},C.updateQueue=_):(C=_.retryQueue,C===null?_.retryQueue=new Set([j]):C.add(j)),A5(w,j,q)),!1}throw Error(s(435,C.tag))}return A5(w,j,q),T5(),!1}if(vs)return _=Qg.current,_!==null?((_.flags&65536)===0&&(_.flags|=256),_.flags|=65536,_.lanes=q,j!==dk&&(w=Error(s(422),{cause:j}),Sy(bp(w,C)))):(j!==dk&&(_=Error(s(423),{cause:j}),Sy(bp(_,C))),w=w.current.alternate,w.flags|=65536,q&=-q,w.lanes|=q,j=bp(j,C),q=ug(w.stateNode,j,q),Zb(w,q),ql!==4&&(ql=2)),!1;var te=Error(s(520),{cause:j});if(te=bp(te,C),QC===null?QC=[te]:QC.push(te),ql!==4&&(ql=2),_===null)return!0;j=bp(j,C),C=_;do{switch(C.tag){case 3:return C.flags|=65536,w=q&-q,C.lanes|=w,w=ug(C.stateNode,j,w),Zb(C,w),!1;case 1:if(_=C.type,te=C.stateNode,(C.flags&128)===0&&(typeof _.getDerivedStateFromError=="function"||te!==null&&typeof te.componentDidCatch=="function"&&(Ly===null||!Ly.has(te))))return C.flags|=65536,q&=-q,C.lanes|=q,q=a5(q),xv(q,w,C,j),Zb(C,q),!1}C=C.return}while(C!==null);return!1}var Y8=Error(s(461)),xf=!1;function Vu(w,_,C,j){_.child=w===null?V8(_,null,C,j):W2(_,w.child,C,j)}function HO(w,_,C,j,q){C=C.render;var te=_.ref;if("ref"in j){var Se={};for(var Me in j)Me!=="ref"&&(Se[Me]=j[Me])}else Se=j;return j2(_),j=F6(w,_,C,Se,te,q),Me=U6(),w!==null&&!xf?(z2(w,_,q),Cv(w,_,q)):(vs&&Me&&P6(_),_.flags|=1,Vu(w,_,j,q),_.child)}function GO(w,_,C,j,q){if(w===null){var te=C.type;return typeof te=="function"&&!P2(te)&&te.defaultProps===void 0&&C.compare===null?(_.tag=15,_.type=te,qO(w,_,te,j,q)):(w=yO(C.type,null,j,_,_.mode,q),w.ref=_.ref,w.return=_,_.child=w)}if(te=w.child,!t1(w,q)){var Se=te.memoizedProps;if(C=C.compare,C=C!==null?C:r0,C(Se,j)&&w.ref===_.ref)return Cv(w,_,q)}return _.flags|=1,w=vp(te,j),w.ref=_.ref,w.return=_,_.child=w}function qO(w,_,C,j,q){if(w!==null){var te=w.memoizedProps;if(r0(te,j)&&w.ref===_.ref)if(xf=!1,_.pendingProps=j=te,t1(w,q))(w.flags&131072)!==0&&(xf=!0);else return _.lanes=w.lanes,Cv(w,_,q)}return Ek(w,_,C,j,q)}function zC(w,_,C){var j=_.pendingProps,q=j.children,te=w!==null?w.memoizedState:null;if(j.mode==="hidden"){if((_.flags&128)!==0){if(j=te!==null?te.baseLanes|C:C,w!==null){for(q=_.child=w.child,te=0;q!==null;)te=te|q.lanes|q.childLanes,q=q.sibling;_.childLanes=te&~j}else _.childLanes=0,_.child=null;return X2(w,_,j,C)}if((C&536870912)!==0)_.memoizedState={baseLanes:0,cachePool:null},w!==null&&_C(_,te!==null?te.cachePool:null),te!==null?Ev(_,te):CO(),W8(_);else return _.lanes=_.childLanes=536870912,X2(w,_,te!==null?te.baseLanes|C:C,C)}else te!==null?(_C(_,te.cachePool),Ev(_,te),Tv(),_.memoizedState=null):(w!==null&&_C(_,null),CO(),Tv());return Vu(w,_,q,C),_.child}function X2(w,_,C,j){var q=wv();return q=q===null?null:{parent:Qc._currentValue,pool:q},_.memoizedState={baseLanes:C,cachePool:q},w!==null&&_C(_,null),CO(),W8(_),w!==null&&fk(w,_,j,!0),null}function fo(w,_){var C=_.ref;if(C===null)w!==null&&w.ref!==null&&(_.flags|=4194816);else{if(typeof C!="function"&&typeof C!="object")throw Error(s(284));(w===null||w.ref!==C)&&(_.flags|=4194816)}}function Ek(w,_,C,j,q){return j2(_),C=F6(w,_,C,j,void 0,q),j=U6(),w!==null&&!xf?(z2(w,_,q),Cv(w,_,q)):(vs&&j&&P6(_),_.flags|=1,Vu(w,_,C,q),_.child)}function VO(w,_,C,j,q,te){return j2(_),_.updateQueue=null,C=L8(_,j,C,q),P8(w),j=U6(),w!==null&&!xf?(z2(w,_,te),Cv(w,_,te)):(vs&&j&&P6(_),_.flags|=1,Vu(w,_,C,te),_.child)}function Iy(w,_,C,j,q){if(j2(_),_.stateNode===null){var te=R2,Se=C.contextType;typeof Se=="object"&&Se!==null&&(te=Hl(Se)),te=new C(j,te),_.memoizedState=te.state!==null&&te.state!==void 0?te.state:null,te.updater=FC,_.stateNode=te,te._reactInternals=_,te=_.stateNode,te.props=j,te.state=_.memoizedState,te.refs={},F2(_),Se=C.contextType,te.context=typeof Se=="object"&&Se!==null?Hl(Se):R2,te.state=_.memoizedState,Se=C.getDerivedStateFromProps,typeof Se=="function"&&(K2(_,C,Se,j),te.state=_.memoizedState),typeof C.getDerivedStateFromProps=="function"||typeof te.getSnapshotBeforeUpdate=="function"||typeof te.UNSAFE_componentWillMount!="function"&&typeof te.componentWillMount!="function"||(Se=te.state,typeof te.componentWillMount=="function"&&te.componentWillMount(),typeof te.UNSAFE_componentWillMount=="function"&&te.UNSAFE_componentWillMount(),Se!==te.state&&FC.enqueueReplaceState(te,te.state,null),Dh(_,j,te,q),Qb(),te.state=_.memoizedState),typeof te.componentDidMount=="function"&&(_.flags|=4194308),j=!0}else if(w===null){te=_.stateNode;var Me=_.memoizedProps,pt=Ay(C,Me);te.props=pt;var Xt=te.context,$n=C.contextType;Se=R2,typeof $n=="object"&&$n!==null&&(Se=Hl($n));var Vn=C.getDerivedStateFromProps;$n=typeof Vn=="function"||typeof te.getSnapshotBeforeUpdate=="function",Me=_.pendingProps!==Me,$n||typeof te.UNSAFE_componentWillReceiveProps!="function"&&typeof te.componentWillReceiveProps!="function"||(Me||Xt!==Se)&&BU(_,te,j,Se),i0=!1;var ln=_.memoizedState;te.state=ln,Dh(_,j,te,q),Qb(),Xt=_.memoizedState,Me||ln!==Xt||i0?(typeof Vn=="function"&&(K2(_,C,Vn,j),Xt=_.memoizedState),(pt=i0||BC(_,C,pt,j,ln,Xt,Se))?($n||typeof te.UNSAFE_componentWillMount!="function"&&typeof te.componentWillMount!="function"||(typeof te.componentWillMount=="function"&&te.componentWillMount(),typeof te.UNSAFE_componentWillMount=="function"&&te.UNSAFE_componentWillMount()),typeof te.componentDidMount=="function"&&(_.flags|=4194308)):(typeof te.componentDidMount=="function"&&(_.flags|=4194308),_.memoizedProps=j,_.memoizedState=Xt),te.props=j,te.state=Xt,te.context=Se,j=pt):(typeof te.componentDidMount=="function"&&(_.flags|=4194308),j=!1)}else{te=_.stateNode,gk(w,_),Se=_.memoizedProps,$n=Ay(C,Se),te.props=$n,Vn=_.pendingProps,ln=te.context,Xt=C.contextType,pt=R2,typeof Xt=="object"&&Xt!==null&&(pt=Hl(Xt)),Me=C.getDerivedStateFromProps,(Xt=typeof Me=="function"||typeof te.getSnapshotBeforeUpdate=="function")||typeof te.UNSAFE_componentWillReceiveProps!="function"&&typeof te.componentWillReceiveProps!="function"||(Se!==Vn||ln!==pt)&&BU(_,te,j,pt),i0=!1,ln=_.memoizedState,te.state=ln,Dh(_,j,te,q),Qb();var Zt=_.memoizedState;Se!==Vn||ln!==Zt||i0||w!==null&&w.dependencies!==null&&SO(w.dependencies)?(typeof Me=="function"&&(K2(_,C,Me,j),Zt=_.memoizedState),($n=i0||BC(_,C,$n,j,ln,Zt,pt)||w!==null&&w.dependencies!==null&&SO(w.dependencies))?(Xt||typeof te.UNSAFE_componentWillUpdate!="function"&&typeof te.componentWillUpdate!="function"||(typeof te.componentWillUpdate=="function"&&te.componentWillUpdate(j,Zt,pt),typeof te.UNSAFE_componentWillUpdate=="function"&&te.UNSAFE_componentWillUpdate(j,Zt,pt)),typeof te.componentDidUpdate=="function"&&(_.flags|=4),typeof te.getSnapshotBeforeUpdate=="function"&&(_.flags|=1024)):(typeof te.componentDidUpdate!="function"||Se===w.memoizedProps&&ln===w.memoizedState||(_.flags|=4),typeof te.getSnapshotBeforeUpdate!="function"||Se===w.memoizedProps&&ln===w.memoizedState||(_.flags|=1024),_.memoizedProps=j,_.memoizedState=Zt),te.props=j,te.state=Zt,te.context=pt,j=$n):(typeof te.componentDidUpdate!="function"||Se===w.memoizedProps&&ln===w.memoizedState||(_.flags|=4),typeof te.getSnapshotBeforeUpdate!="function"||Se===w.memoizedProps&&ln===w.memoizedState||(_.flags|=1024),j=!1)}return te=j,fo(w,_),j=(_.flags&128)!==0,te||j?(te=_.stateNode,C=j&&typeof C.getDerivedStateFromError!="function"?null:te.render(),_.flags|=1,w!==null&&j?(_.child=W2(_,w.child,null,q),_.child=W2(_,null,C,q)):Vu(w,_,C,q),_.memoizedState=te.state,w=_.child):w=Cv(w,_,q),w}function X8(w,_,C,j){return yC(),_.flags|=256,Vu(w,_,C,j),_.child}var HC={dehydrated:null,treeContext:null,retryLane:0,hydrationErrors:null};function WO(w){return{baseLanes:w,cachePool:x8()}}function Cf(w,_,C){return w=w!==null?w.childLanes&~C:0,_&&(w|=sb),w}function J8(w,_,C){var j=_.pendingProps,q=!1,te=(_.flags&128)!==0,Se;if((Se=te)||(Se=w!==null&&w.memoizedState===null?!1:(Gl.current&2)!==0),Se&&(q=!0,_.flags&=-129),Se=(_.flags&32)!==0,_.flags&=-33,w===null){if(vs){if(q?ob(_):Tv(),vs){var Me=cl,pt;if(pt=Me){e:{for(pt=Me,Me=rg;pt.nodeType!==8;){if(!Me){Me=null;break e}if(pt=o1(pt.nextSibling),pt===null){Me=null;break e}}Me=pt}Me!==null?(_.memoizedState={dehydrated:Me,treeContext:qu!==null?{id:sl,overflow:vv}:null,retryLane:536870912,hydrationErrors:null},pt=ng(18,null,null,0),pt.stateNode=Me,pt.return=_,_.child=pt,Rd=_,cl=null,pt=!0):pt=!1}pt||rb(_)}if(Me=_.memoizedState,Me!==null&&(Me=Me.dehydrated,Me!==null))return _p(Me)?_.lanes=32:_.lanes=536870912,null;a0(_)}return Me=j.children,j=j.fallback,q?(Tv(),q=_.mode,Me=GC({mode:"hidden",children:Me},q),j=L2(j,q,C,null),Me.return=_,j.return=_,Me.sibling=j,_.child=Me,q=_.child,q.memoizedState=WO(C),q.childLanes=Cf(w,Se,C),_.memoizedState=HC,j):(ob(_),c5(_,Me))}if(pt=w.memoizedState,pt!==null&&(Me=pt.dehydrated,Me!==null)){if(te)_.flags&256?(ob(_),_.flags&=-257,_=mt(w,_,C)):_.memoizedState!==null?(Tv(),_.child=w.child,_.flags|=128,_=null):(Tv(),q=j.fallback,Me=_.mode,j=GC({mode:"visible",children:j.children},Me),q=L2(q,Me,C,null),q.flags|=2,j.return=_,q.return=_,j.sibling=q,_.child=j,W2(_,w.child,null,C),j=_.child,j.memoizedState=WO(C),j.childLanes=Cf(w,Se,C),_.memoizedState=HC,_=q);else if(ob(_),_p(Me)){if(Se=Me.nextSibling&&Me.nextSibling.dataset,Se)var Xt=Se.dgst;Se=Xt,j=Error(s(419)),j.stack="",j.digest=Se,Sy({value:j,source:null,stack:null}),_=mt(w,_,C)}else if(xf||fk(w,_,C,!1),Se=(C&w.childLanes)!==0,xf||Se){if(Se=Nu,Se!==null&&(j=C&-C,j=(j&42)!==0?1:dv(j),j=(j&(Se.suspendedLanes|C))!==0?0:j,j!==0&&j!==pt.retryLane))throw pt.retryLane=j,N2(w,j),rm(Se,w,j),Y8;Me.data==="$?"||T5(),_=mt(w,_,C)}else Me.data==="$?"?(_.flags|=192,_.child=w.child,_=null):(w=pt.treeContext,cl=o1(Me.nextSibling),Rd=_,vs=!0,Jb=null,rg=!1,w!==null&&(tb[nb++]=sl,tb[nb++]=vv,tb[nb++]=qu,sl=w.id,vv=w.overflow,qu=_),_=c5(_,j.children),_.flags|=4096);return _}return q?(Tv(),q=j.fallback,Me=_.mode,pt=w.child,Xt=pt.sibling,j=vp(pt,{mode:"hidden",children:j.children}),j.subtreeFlags=pt.subtreeFlags&65011712,Xt!==null?q=vp(Xt,q):(q=L2(q,Me,C,null),q.flags|=2),q.return=_,j.return=_,j.sibling=q,_.child=j,j=q,q=_.child,Me=w.child.memoizedState,Me===null?Me=WO(C):(pt=Me.cachePool,pt!==null?(Xt=Qc._currentValue,pt=pt.parent!==Xt?{parent:Xt,pool:Xt}:pt):pt=x8(),Me={baseLanes:Me.baseLanes|C,cachePool:pt}),q.memoizedState=Me,q.childLanes=Cf(w,Se,C),_.memoizedState=HC,j):(ob(_),C=w.child,w=C.sibling,C=vp(C,{mode:"visible",children:j.children}),C.return=_,C.sibling=null,w!==null&&(Se=_.deletions,Se===null?(_.deletions=[w],_.flags|=16):Se.push(w)),_.child=C,_.memoizedState=null,C)}function c5(w,_){return _=GC({mode:"visible",children:_},w.mode),_.return=w,w.child=_}function GC(w,_){return w=ng(22,w,null,_),w.lanes=0,w.stateNode={_visibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null},w}function mt(w,_,C){return W2(_,w.child,null,C),w=c5(_,_.pendingProps.children),w.flags|=2,_.memoizedState=null,w}function u5(w,_,C){w.lanes|=_;var j=w.alternate;j!==null&&(j.lanes|=_),ig(w.return,_,C)}function l5(w,_,C,j,q){var te=w.memoizedState;te===null?w.memoizedState={isBackwards:_,rendering:null,renderingStartTime:0,last:j,tail:C,tailMode:q}:(te.isBackwards=_,te.rendering=null,te.renderingStartTime=0,te.last=j,te.tail=C,te.tailMode=q)}function Z8(w,_,C){var j=_.pendingProps,q=j.revealOrder,te=j.tail;if(Vu(w,_,j.children,C),j=Gl.current,(j&2)!==0)j=j&1|2,_.flags|=128;else{if(w!==null&&(w.flags&128)!==0)e:for(w=_.child;w!==null;){if(w.tag===13)w.memoizedState!==null&&u5(w,C,_);else if(w.tag===19)u5(w,C,_);else if(w.child!==null){w.child.return=w,w=w.child;continue}if(w===_)break e;for(;w.sibling===null;){if(w.return===null||w.return===_)break e;w=w.return}w.sibling.return=w.return,w=w.sibling}j&=1}switch(Ie(Gl,j),q){case"forwards":for(C=_.child,q=null;C!==null;)w=C.alternate,w!==null&&$C(w)===null&&(q=C),C=C.sibling;C=q,C===null?(q=_.child,_.child=null):(q=C.sibling,C.sibling=null),l5(_,!1,q,C,te);break;case"backwards":for(C=null,q=_.child,_.child=null;q!==null;){if(w=q.alternate,w!==null&&$C(w)===null){_.child=q;break}w=q.sibling,q.sibling=C,C=q,q=w}l5(_,!0,C,null,te);break;case"together":l5(_,!1,null,null,void 0);break;default:_.memoizedState=null}return _.child}function Cv(w,_,C){if(w!==null&&(_.dependencies=w.dependencies),Py|=_.lanes,(C&_.childLanes)===0)if(w!==null){if(fk(w,_,C,!1),(C&_.childLanes)===0)return null}else return null;if(w!==null&&_.child!==w.child)throw Error(s(153));if(_.child!==null){for(w=_.child,C=vp(w,w.pendingProps),_.child=C,C.return=_;w.sibling!==null;)w=w.sibling,C=C.sibling=vp(w,w.pendingProps),C.return=_;C.sibling=null}return _.child}function t1(w,_){return(w.lanes&_)!==0?!0:(w=w.dependencies,!!(w!==null&&SO(w)))}function Sp(w,_,C){switch(_.tag){case 3:Nt(_,_.stateNode.containerInfo),Lh(_,Qc,w.memoizedState.cache),yC();break;case 27:case 5:br(_);break;case 4:Nt(_,_.stateNode.containerInfo);break;case 10:Lh(_,_.type,_.memoizedProps.value);break;case 13:var j=_.memoizedState;if(j!==null)return j.dehydrated!==null?(ob(_),_.flags|=128,null):(C&_.child.childLanes)!==0?J8(w,_,C):(ob(_),w=Cv(w,_,C),w!==null?w.sibling:null);ob(_);break;case 19:var q=(w.flags&128)!==0;if(j=(C&_.childLanes)!==0,j||(fk(w,_,C,!1),j=(C&_.childLanes)!==0),q){if(j)return Z8(w,_,C);_.flags|=128}if(q=_.memoizedState,q!==null&&(q.rendering=null,q.tail=null,q.lastEffect=null),Ie(Gl,Gl.current),j)break;return null;case 22:case 23:return _.lanes=0,zC(w,_,C);case 24:Lh(_,Qc,w.memoizedState.cache)}return Cv(w,_,C)}function KO(w,_,C){if(w!==null)if(w.memoizedProps!==_.pendingProps)xf=!0;else{if(!t1(w,C)&&(_.flags&128)===0)return xf=!1,Sp(w,_,C);xf=(w.flags&131072)!==0}else xf=!1,vs&&(_.flags&1048576)!==0&&R6(_,bC,_.index);switch(_.lanes=0,_.tag){case 16:e:{w=_.pendingProps;var j=_.elementType,q=j._init;if(j=q(j._payload),_.type=j,typeof j=="function")P2(j)?(w=Ay(j,w),_.tag=1,_=Iy(null,_,j,w,C)):(_.tag=0,_=Ek(null,_,j,w,C));else{if(j!=null){if(q=j.$$typeof,q===X){_.tag=11,_=HO(null,_,j,w,C);break e}else if(q===he){_.tag=14,_=GO(null,_,j,w,C);break e}}throw _=Z(j)||j,Error(s(306,_,""))}}return _;case 0:return Ek(w,_,_.type,_.pendingProps,C);case 1:return j=_.type,q=Ay(j,_.pendingProps),Iy(w,_,j,q,C);case 3:e:{if(Nt(_,_.stateNode.containerInfo),w===null)throw Error(s(387));j=_.pendingProps;var te=_.memoizedState;q=te.element,gk(w,_),Dh(_,j,null,C);var Se=_.memoizedState;if(j=Se.cache,Lh(_,Qc,j),j!==te.cache&&ts(_,[Qc],C,!0),Qb(),j=Se.element,te.isDehydrated)if(te={element:j,isDehydrated:!1,cache:Se.cache},_.updateQueue.baseState=te,_.memoizedState=te,_.flags&256){_=X8(w,_,j,C);break e}else if(j!==q){q=bp(Error(s(424)),_),Sy(q),_=X8(w,_,j,C);break e}else{switch(w=_.stateNode.containerInfo,w.nodeType){case 9:w=w.body;break;default:w=w.nodeName==="HTML"?w.ownerDocument.body:w}for(cl=o1(w.firstChild),Rd=_,vs=!0,Jb=null,rg=!0,C=V8(_,null,j,C),_.child=C;C;)C.flags=C.flags&-3|4096,C=C.sibling}else{if(yC(),j===q){_=Cv(w,_,C);break e}Vu(w,_,j,C)}_=_.child}return _;case 26:return fo(w,_),w===null?(C=F5(_.type,null,_.pendingProps,null))?_.memoizedState=C:vs||(C=_.type,w=_.pendingProps,j=tS(Yt.current).createElement(C),j[_l]=_,j[rh]=w,Jd(j,C,w),Od(j),_.stateNode=j):_.memoizedState=F5(_.type,w.memoizedProps,_.pendingProps,w.memoizedState),null;case 27:return br(_),w===null&&vs&&(j=_.stateNode=z9(_.type,_.pendingProps,Yt.current),Rd=_,rg=!0,q=cl,cb(_.type)?(j5=q,cl=o1(j.firstChild)):cl=q),Vu(w,_,_.pendingProps.children,C),fo(w,_),w===null&&(_.flags|=4194304),_.child;case 5:return w===null&&vs&&((q=j=cl)&&(j=B9(j,_.type,_.pendingProps,rg),j!==null?(_.stateNode=j,Rd=_,cl=o1(j.firstChild),rg=!1,q=!0):q=!1),q||rb(_)),br(_),q=_.type,te=_.pendingProps,Se=w!==null?w.memoizedProps:null,j=te.children,dN(q,te)?j=null:Se!==null&&dN(q,Se)&&(_.flags|=32),_.memoizedState!==null&&(q=F6(w,_,B6,null,null,C),Mk._currentValue=q),fo(w,_),Vu(w,_,j,C),_.child;case 6:return w===null&&vs&&((w=C=cl)&&(C=hN(C,_.pendingProps,rg),C!==null?(_.stateNode=C,Rd=_,cl=null,w=!0):w=!1),w||rb(_)),null;case 13:return J8(w,_,C);case 4:return Nt(_,_.stateNode.containerInfo),j=_.pendingProps,w===null?_.child=W2(_,null,j,C):Vu(w,_,j,C),_.child;case 11:return HO(w,_,_.type,_.pendingProps,C);case 7:return Vu(w,_,_.pendingProps,C),_.child;case 8:return Vu(w,_,_.pendingProps.children,C),_.child;case 12:return Vu(w,_,_.pendingProps.children,C),_.child;case 10:return j=_.pendingProps,Lh(_,_.type,j.value),Vu(w,_,j.children,C),_.child;case 9:return q=_.type._context,j=_.pendingProps.children,j2(_),q=Hl(q),j=j(q),_.flags|=1,Vu(w,_,j,C),_.child;case 14:return GO(w,_,_.type,_.pendingProps,C);case 15:return qO(w,_,_.type,_.pendingProps,C);case 19:return Z8(w,_,C);case 31:return j=_.pendingProps,C=_.mode,j={mode:j.mode,children:j.children},w===null?(C=GC(j,C),C.ref=_.ref,_.child=C,C.return=_,_=C):(C=vp(w.child,j),C.ref=_.ref,_.child=C,C.return=_,_=C),_;case 22:return zC(w,_,C);case 24:return j2(_),j=Hl(Qc),w===null?(q=wv(),q===null&&(q=Nu,te=kO(),q.pooledCache=te,te.refCount++,te!==null&&(q.pooledCacheLanes|=C),q=te),_.memoizedState={parent:j,cache:q},F2(_),Lh(_,Qc,q)):((w.lanes&C)!==0&&(gk(w,_),Dh(_,null,null,C),Qb()),q=w.memoizedState,te=_.memoizedState,q.parent!==j?(q={parent:j,cache:j},_.memoizedState=q,_.lanes===0&&(_.memoizedState=_.updateQueue.baseState=q),Lh(_,Qc,j)):(j=te.cache,Lh(_,Qc,j),j!==q.cache&&ts(_,[Qc],C,!0))),Vu(w,_,_.pendingProps.children,C),_.child;case 29:throw _.pendingProps}throw Error(s(156,_.tag))}function Av(w){w.flags|=4}function YO(w,_){if(_.type!=="stylesheet"||(_.state.loading&4)!==0)w.flags&=-16777217;else if(w.flags|=16777216,!Zd(_)){if(_=Qg.current,_!==null&&(($s&4194048)===$s?Ep!==null:($s&62914560)!==$s&&($s&536870912)===0||_!==Ep))throw pk=D6,C8;w.flags|=8192}}function XO(w,_){_!==null&&(w.flags|=4),w.flags&16384&&(_=w.tag!==22?hu():536870912,w.lanes|=_,xk|=_)}function qC(w,_){if(!vs)switch(w.tailMode){case"hidden":_=w.tail;for(var C=null;_!==null;)_.alternate!==null&&(C=_),_=_.sibling;C===null?w.tail=null:C.sibling=null;break;case"collapsed":C=w.tail;for(var j=null;C!==null;)C.alternate!==null&&(j=C),C=C.sibling;j===null?_||w.tail===null?w.tail=null:w.tail.sibling=null:j.sibling=null}}function pu(w){var _=w.alternate!==null&&w.alternate.child===w.child,C=0,j=0;if(_)for(var q=w.child;q!==null;)C|=q.lanes|q.childLanes,j|=q.subtreeFlags&65011712,j|=q.flags&65011712,q.return=w,q=q.sibling;else for(q=w.child;q!==null;)C|=q.lanes|q.childLanes,j|=q.subtreeFlags,j|=q.flags,q.return=w,q=q.sibling;return w.subtreeFlags|=j,w.childLanes=C,_}function Q8(w,_,C){var j=_.pendingProps;switch(vC(_),_.tag){case 31:case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return pu(_),null;case 1:return pu(_),null;case 3:return C=_.stateNode,j=null,w!==null&&(j=w.memoizedState.cache),_.memoizedState.cache!==j&&(_.flags|=2048),wp(Qc),on(),C.pendingContext&&(C.context=C.pendingContext,C.pendingContext=null),(w===null||w.child===null)&&(wC(_)?Av(_):w===null||w.memoizedState.isDehydrated&&(_.flags&256)===0||(_.flags|=1024,y8())),pu(_),null;case 26:return C=_.memoizedState,w===null?(Av(_),C!==null?(pu(_),YO(_,C)):(pu(_),_.flags&=-16777217)):C?C!==w.memoizedState?(Av(_),pu(_),YO(_,C)):(pu(_),_.flags&=-16777217):(w.memoizedProps!==j&&Av(_),pu(_),_.flags&=-16777217),null;case 27:wi(_),C=Yt.current;var q=_.type;if(w!==null&&_.stateNode!=null)w.memoizedProps!==j&&Av(_);else{if(!j){if(_.stateNode===null)throw Error(s(166));return pu(_),null}w=$t.current,wC(_)?v8(_):(w=z9(q,j,C),_.stateNode=w,Av(_))}return pu(_),null;case 5:if(wi(_),C=_.type,w!==null&&_.stateNode!=null)w.memoizedProps!==j&&Av(_);else{if(!j){if(_.stateNode===null)throw Error(s(166));return pu(_),null}if(w=$t.current,wC(_))v8(_);else{switch(q=tS(Yt.current),w){case 1:w=q.createElementNS("http://www.w3.org/2000/svg",C);break;case 2:w=q.createElementNS("http://www.w3.org/1998/Math/MathML",C);break;default:switch(C){case"svg":w=q.createElementNS("http://www.w3.org/2000/svg",C);break;case"math":w=q.createElementNS("http://www.w3.org/1998/Math/MathML",C);break;case"script":w=q.createElement("div"),w.innerHTML="<script><\/script>",w=w.removeChild(w.firstChild);break;case"select":w=typeof j.is=="string"?q.createElement("select",{is:j.is}):q.createElement("select"),j.multiple?w.multiple=!0:j.size&&(w.size=j.size);break;default:w=typeof j.is=="string"?q.createElement(C,{is:j.is}):q.createElement(C)}}w[_l]=_,w[rh]=j;e:for(q=_.child;q!==null;){if(q.tag===5||q.tag===6)w.appendChild(q.stateNode);else if(q.tag!==4&&q.tag!==27&&q.child!==null){q.child.return=q,q=q.child;continue}if(q===_)break e;for(;q.sibling===null;){if(q.return===null||q.return===_)break e;q=q.return}q.sibling.return=q.return,q=q.sibling}_.stateNode=w;e:switch(Jd(w,C,j),C){case"button":case"input":case"select":case"textarea":w=!!j.autoFocus;break e;case"img":w=!0;break e;default:w=!1}w&&Av(_)}}return pu(_),_.flags&=-16777217,null;case 6:if(w&&_.stateNode!=null)w.memoizedProps!==j&&Av(_);else{if(typeof j!="string"&&_.stateNode===null)throw Error(s(166));if(w=Yt.current,wC(_)){if(w=_.stateNode,C=_.memoizedProps,j=null,q=Rd,q!==null)switch(q.tag){case 27:case 5:j=q.memoizedProps}w[_l]=_,w=!!(w.nodeValue===C||j!==null&&j.suppressHydrationWarning===!0||M9(w.nodeValue,C)),w||rb(_)}else w=tS(w).createTextNode(j),w[_l]=_,_.stateNode=w}return pu(_),null;case 13:if(j=_.memoizedState,w===null||w.memoizedState!==null&&w.memoizedState.dehydrated!==null){if(q=wC(_),j!==null&&j.dehydrated!==null){if(w===null){if(!q)throw Error(s(318));if(q=_.memoizedState,q=q!==null?q.dehydrated:null,!q)throw Error(s(317));q[_l]=_}else yC(),(_.flags&128)===0&&(_.memoizedState=null),_.flags|=4;pu(_),q=!1}else q=y8(),w!==null&&w.memoizedState!==null&&(w.memoizedState.hydrationErrors=q),q=!0;if(!q)return _.flags&256?(a0(_),_):(a0(_),null)}if(a0(_),(_.flags&128)!==0)return _.lanes=C,_;if(C=j!==null,w=w!==null&&w.memoizedState!==null,C){j=_.child,q=null,j.alternate!==null&&j.alternate.memoizedState!==null&&j.alternate.memoizedState.cachePool!==null&&(q=j.alternate.memoizedState.cachePool.pool);var te=null;j.memoizedState!==null&&j.memoizedState.cachePool!==null&&(te=j.memoizedState.cachePool.pool),te!==q&&(j.flags|=2048)}return C!==w&&C&&(_.child.flags|=8192),XO(_,_.updateQueue),pu(_),null;case 4:return on(),w===null&&P5(_.stateNode.containerInfo),pu(_),null;case 10:return wp(_.type),pu(_),null;case 19:if(ht(Gl),q=_.memoizedState,q===null)return pu(_),null;if(j=(_.flags&128)!==0,te=q.rendering,te===null)if(j)qC(q,!1);else{if(ql!==0||w!==null&&(w.flags&128)!==0)for(w=_.child;w!==null;){if(te=$C(w),te!==null){for(_.flags|=128,qC(q,!1),w=te.updateQueue,_.updateQueue=w,XO(_,w),_.subtreeFlags=0,w=C,C=_.child;C!==null;)mC(C,w),C=C.sibling;return Ie(Gl,Gl.current&1|2),_.child}w=w.sibling}q.tail!==null&&qi()>Ck&&(_.flags|=128,j=!0,qC(q,!1),_.lanes=4194304)}else{if(!j)if(w=$C(te),w!==null){if(_.flags|=128,j=!0,w=w.updateQueue,_.updateQueue=w,XO(_,w),qC(q,!0),q.tail===null&&q.tailMode==="hidden"&&!te.alternate&&!vs)return pu(_),null}else 2*qi()-q.renderingStartTime>Ck&&C!==536870912&&(_.flags|=128,j=!0,qC(q,!1),_.lanes=4194304);q.isBackwards?(te.sibling=_.child,_.child=te):(w=q.last,w!==null?w.sibling=te:_.child=te,q.last=te)}return q.tail!==null?(_=q.tail,q.rendering=_,q.tail=_.sibling,q.renderingStartTime=qi(),_.sibling=null,w=Gl.current,Ie(Gl,j?w&1|2:w&1),_):(pu(_),null);case 22:case 23:return a0(_),$6(),j=_.memoizedState!==null,w!==null?w.memoizedState!==null!==j&&(_.flags|=8192):j&&(_.flags|=8192),j?(C&536870912)!==0&&(_.flags&128)===0&&(pu(_),_.subtreeFlags&6&&(_.flags|=8192)):pu(_),C=_.updateQueue,C!==null&&XO(_,C.retryQueue),C=null,w!==null&&w.memoizedState!==null&&w.memoizedState.cachePool!==null&&(C=w.memoizedState.cachePool.pool),j=null,_.memoizedState!==null&&_.memoizedState.cachePool!==null&&(j=_.memoizedState.cachePool.pool),j!==C&&(_.flags|=2048),w!==null&&ht($2),null;case 24:return C=null,w!==null&&(C=w.memoizedState.cache),_.memoizedState.cache!==C&&(_.flags|=2048),wp(Qc),pu(_),null;case 25:return null;case 30:return null}throw Error(s(156,_.tag))}function e9(w,_){switch(vC(_),_.tag){case 1:return w=_.flags,w&65536?(_.flags=w&-65537|128,_):null;case 3:return wp(Qc),on(),w=_.flags,(w&65536)!==0&&(w&128)===0?(_.flags=w&-65537|128,_):null;case 26:case 27:case 5:return wi(_),null;case 13:if(a0(_),w=_.memoizedState,w!==null&&w.dehydrated!==null){if(_.alternate===null)throw Error(s(340));yC()}return w=_.flags,w&65536?(_.flags=w&-65537|128,_):null;case 19:return ht(Gl),null;case 4:return on(),null;case 10:return wp(_.type),null;case 22:case 23:return a0(_),$6(),w!==null&&ht($2),w=_.flags,w&65536?(_.flags=w&-65537|128,_):null;case 24:return wp(Qc),null;case 25:return null;default:return null}}function JO(w,_){switch(vC(_),_.tag){case 3:wp(Qc),on();break;case 26:case 27:case 5:wi(_);break;case 4:on();break;case 13:a0(_);break;case 19:ht(Gl);break;case 10:wp(_.type);break;case 22:case 23:a0(_),$6(),w!==null&&ht($2);break;case 24:wp(Qc)}}function VC(w,_){try{var C=_.updateQueue,j=C!==null?C.lastEffect:null;if(j!==null){var q=j.next;C=q;do{if((C.tag&w)===w){j=void 0;var te=C.create,Se=C.inst;j=te(),Se.destroy=j}C=C.next}while(C!==q)}}catch(Me){gu(_,_.return,Me)}}function Oy(w,_,C){try{var j=_.updateQueue,q=j!==null?j.lastEffect:null;if(q!==null){var te=q.next;j=te;do{if((j.tag&w)===w){var Se=j.inst,Me=Se.destroy;if(Me!==void 0){Se.destroy=void 0,q=_;var pt=C,Xt=Me;try{Xt()}catch($n){gu(q,pt,$n)}}}j=j.next}while(j!==te)}}catch($n){gu(_,_.return,$n)}}function t9(w){var _=w.updateQueue;if(_!==null){var C=w.stateNode;try{j6(_,C)}catch(j){gu(w,w.return,j)}}}function WC(w,_,C){C.props=Ay(w.type,w.memoizedProps),C.state=w.memoizedState;try{C.componentWillUnmount()}catch(j){gu(w,_,j)}}function Iv(w,_){try{var C=w.ref;if(C!==null){switch(w.tag){case 26:case 27:case 5:var j=w.stateNode;break;case 30:j=w.stateNode;break;default:j=w.stateNode}typeof C=="function"?w.refCleanup=C(j):C.current=j}}catch(q){gu(w,_,q)}}function em(w,_){var C=w.ref,j=w.refCleanup;if(C!==null)if(typeof j=="function")try{j()}catch(q){gu(w,_,q)}finally{w.refCleanup=null,w=w.alternate,w!=null&&(w.refCleanup=null)}else if(typeof C=="function")try{C(null)}catch(q){gu(w,_,q)}else C.current=null}function ZO(w){var _=w.type,C=w.memoizedProps,j=w.stateNode;try{e:switch(_){case"button":case"input":case"select":case"textarea":C.autoFocus&&j.focus();break e;case"img":C.src?j.src=C.src:C.srcSet&&(j.srcset=C.srcSet)}}catch(q){gu(w,w.return,q)}}function KC(w,_,C){try{var j=w.stateNode;ez(j,w.type,C,_),j[rh]=_}catch(q){gu(w,w.return,q)}}function Sk(w){return w.tag===5||w.tag===3||w.tag===26||w.tag===27&&cb(w.type)||w.tag===4}function YC(w){e:for(;;){for(;w.sibling===null;){if(w.return===null||Sk(w.return))return null;w=w.return}for(w.sibling.return=w.return,w=w.sibling;w.tag!==5&&w.tag!==6&&w.tag!==18;){if(w.tag===27&&cb(w.type)||w.flags&2||w.child===null||w.tag===4)continue e;w.child.return=w,w=w.child}if(!(w.flags&2))return w.stateNode}}function J2(w,_,C){var j=w.tag;if(j===5||j===6)w=w.stateNode,_?(C.nodeType===9?C.body:C.nodeName==="HTML"?C.ownerDocument.body:C).insertBefore(w,_):(_=C.nodeType===9?C.body:C.nodeName==="HTML"?C.ownerDocument.body:C,_.appendChild(w),C=C._reactRootContainer,C!=null||_.onclick!==null||(_.onclick=uN));else if(j!==4&&(j===27&&cb(w.type)&&(C=w.stateNode,_=null),w=w.child,w!==null))for(J2(w,_,C),w=w.sibling;w!==null;)J2(w,_,C),w=w.sibling}function XC(w,_,C){var j=w.tag;if(j===5||j===6)w=w.stateNode,_?C.insertBefore(w,_):C.appendChild(w);else if(j!==4&&(j===27&&cb(w.type)&&(C=w.stateNode),w=w.child,w!==null))for(XC(w,_,C),w=w.sibling;w!==null;)XC(w,_,C),w=w.sibling}function JC(w){var _=w.stateNode,C=w.memoizedProps;try{for(var j=w.type,q=_.attributes;q.length;)_.removeAttributeNode(q[0]);Jd(_,j,C),_[_l]=w,_[rh]=C}catch(te){gu(w,w.return,te)}}var c0=!1,Ys=!1,Ov=!1,n9=typeof WeakSet=="function"?WeakSet:Set,Yd=null;function UU(w,_){if(w=w.containerInfo,i1=oS,w=m8(w),A6(w)){if("selectionStart"in w)var C={start:w.selectionStart,end:w.selectionEnd};else e:{C=(C=w.ownerDocument)&&C.defaultView||window;var j=C.getSelection&&C.getSelection();if(j&&j.rangeCount!==0){C=j.anchorNode;var q=j.anchorOffset,te=j.focusNode;j=j.focusOffset;try{C.nodeType,te.nodeType}catch{C=null;break e}var Se=0,Me=-1,pt=-1,Xt=0,$n=0,Vn=w,ln=null;t:for(;;){for(var Zt;Vn!==C||q!==0&&Vn.nodeType!==3||(Me=Se+q),Vn!==te||j!==0&&Vn.nodeType!==3||(pt=Se+j),Vn.nodeType===3&&(Se+=Vn.nodeValue.length),(Zt=Vn.firstChild)!==null;)ln=Vn,Vn=Zt;for(;;){if(Vn===w)break t;if(ln===C&&++Xt===q&&(Me=Se),ln===te&&++$n===j&&(pt=Se),(Zt=Vn.nextSibling)!==null)break;Vn=ln,ln=Vn.parentNode}Vn=Zt}C=Me===-1||pt===-1?null:{start:Me,end:pt}}else C=null}C=C||{start:0,end:0}}else C=null;for(lN={focusedElem:w,selectionRange:C},oS=!1,Yd=_;Yd!==null;)if(_=Yd,w=_.child,(_.subtreeFlags&1024)!==0&&w!==null)w.return=_,Yd=w;else for(;Yd!==null;){switch(_=Yd,te=_.alternate,w=_.flags,_.tag){case 0:break;case 11:case 15:break;case 1:if((w&1024)!==0&&te!==null){w=void 0,C=_,q=te.memoizedProps,te=te.memoizedState,j=C.stateNode;try{var Do=Ay(C.type,q,C.elementType===C.type);w=j.getSnapshotBeforeUpdate(Do,te),j.__reactInternalSnapshotBeforeUpdate=w}catch(Oo){gu(C,C.return,Oo)}}break;case 3:if((w&1024)!==0){if(w=_.stateNode.containerInfo,C=w.nodeType,C===9)M5(w);else if(C===1)switch(w.nodeName){case"HEAD":case"HTML":case"BODY":M5(w);break;default:w.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if((w&1024)!==0)throw Error(s(163))}if(w=_.sibling,w!==null){w.return=_.return,Yd=w;break}Yd=_.return}}function ZC(w,_,C){var j=C.flags;switch(C.tag){case 0:case 11:case 15:Ny(w,C),j&4&&VC(5,C);break;case 1:if(Ny(w,C),j&4)if(w=C.stateNode,_===null)try{w.componentDidMount()}catch(Se){gu(C,C.return,Se)}else{var q=Ay(C.type,_.memoizedProps);_=_.memoizedState;try{w.componentDidUpdate(q,_,w.__reactInternalSnapshotBeforeUpdate)}catch(Se){gu(C,C.return,Se)}}j&64&&t9(C),j&512&&Iv(C,C.return);break;case 3:if(Ny(w,C),j&64&&(w=C.updateQueue,w!==null)){if(_=null,C.child!==null)switch(C.child.tag){case 27:case 5:_=C.child.stateNode;break;case 1:_=C.child.stateNode}try{j6(w,_)}catch(Se){gu(C,C.return,Se)}}break;case 27:_===null&&j&4&&JC(C);case 26:case 5:Ny(w,C),_===null&&j&4&&ZO(C),j&512&&Iv(C,C.return);break;case 12:Ny(w,C);break;case 13:Ny(w,C),j&4&&i9(w,C),j&64&&(w=C.memoizedState,w!==null&&(w=w.dehydrated,w!==null&&(C=rN.bind(null,C),Pk(w,C))));break;case 22:if(j=C.memoizedState!==null||c0,!j){_=_!==null&&_.memoizedState!==null||Ys,q=c0;var te=Ys;c0=j,(Ys=_)&&!te?Nv(w,C,(C.subtreeFlags&8772)!==0):Ny(w,C),c0=q,Ys=te}break;case 30:break;default:Ny(w,C)}}function r9(w){var _=w.alternate;_!==null&&(w.alternate=null,r9(_)),w.child=null,w.deletions=null,w.sibling=null,w.tag===5&&(_=w.stateNode,_!==null&&rC(_)),w.stateNode=null,w.return=null,w.dependencies=null,w.memoizedProps=null,w.memoizedState=null,w.pendingProps=null,w.stateNode=null,w.updateQueue=null}var Wu=null,lg=!1;function n1(w,_,C){for(C=C.child;C!==null;)d5(w,_,C),C=C.sibling}function d5(w,_,C){if(Pn&&typeof Pn.onCommitFiberUnmount=="function")try{Pn.onCommitFiberUnmount(Sn,C)}catch{}switch(C.tag){case 26:Ys||em(C,_),n1(w,_,C),C.memoizedState?C.memoizedState.count--:C.stateNode&&(C=C.stateNode,C.parentNode.removeChild(C));break;case 27:Ys||em(C,_);var j=Wu,q=lg;cb(C.type)&&(Wu=C.stateNode,lg=!1),n1(w,_,C),c3(C.stateNode),Wu=j,lg=q;break;case 5:Ys||em(C,_);case 6:if(j=Wu,q=lg,Wu=null,n1(w,_,C),Wu=j,lg=q,Wu!==null)if(lg)try{(Wu.nodeType===9?Wu.body:Wu.nodeName==="HTML"?Wu.ownerDocument.body:Wu).removeChild(C.stateNode)}catch(te){gu(C,_,te)}else try{Wu.removeChild(C.stateNode)}catch(te){gu(C,_,te)}break;case 18:Wu!==null&&(lg?(w=Wu,fN(w.nodeType===9?w.body:w.nodeName==="HTML"?w.ownerDocument.body:w,C.stateNode),EN(w)):fN(Wu,C.stateNode));break;case 4:j=Wu,q=lg,Wu=C.stateNode.containerInfo,lg=!0,n1(w,_,C),Wu=j,lg=q;break;case 0:case 11:case 14:case 15:Ys||Oy(2,C,_),Ys||Oy(4,C,_),n1(w,_,C);break;case 1:Ys||(em(C,_),j=C.stateNode,typeof j.componentWillUnmount=="function"&&WC(C,_,j)),n1(w,_,C);break;case 21:n1(w,_,C);break;case 22:Ys=(j=Ys)||C.memoizedState!==null,n1(w,_,C),Ys=j;break;default:n1(w,_,C)}}function i9(w,_){if(_.memoizedState===null&&(w=_.alternate,w!==null&&(w=w.memoizedState,w!==null&&(w=w.dehydrated,w!==null))))try{EN(w)}catch(C){gu(_,_.return,C)}}function zU(w){switch(w.tag){case 13:case 19:var _=w.stateNode;return _===null&&(_=w.stateNode=new n9),_;case 22:return w=w.stateNode,_=w._retryCache,_===null&&(_=w._retryCache=new n9),_;default:throw Error(s(435,w.tag))}}function f5(w,_){var C=zU(w);_.forEach(function(j){var q=KU.bind(null,w,j);C.has(j)||(C.add(j),j.then(q,q))})}function tm(w,_){var C=_.deletions;if(C!==null)for(var j=0;j<C.length;j++){var q=C[j],te=w,Se=_,Me=Se;e:for(;Me!==null;){switch(Me.tag){case 27:if(cb(Me.type)){Wu=Me.stateNode,lg=!1;break e}break;case 5:Wu=Me.stateNode,lg=!1;break e;case 3:case 4:Wu=Me.stateNode.containerInfo,lg=!0;break e}Me=Me.return}if(Wu===null)throw Error(s(160));d5(te,Se,q),Wu=null,lg=!1,te=q.alternate,te!==null&&(te.return=null),q.return=null}if(_.subtreeFlags&13878)for(_=_.child;_!==null;)h5(_,w),_=_.sibling}var r1=null;function h5(w,_){var C=w.alternate,j=w.flags;switch(w.tag){case 0:case 11:case 14:case 15:tm(_,w),dg(w),j&4&&(Oy(3,w,w.return),VC(3,w),Oy(5,w,w.return));break;case 1:tm(_,w),dg(w),j&512&&(Ys||C===null||em(C,C.return)),j&64&&c0&&(w=w.updateQueue,w!==null&&(j=w.callbacks,j!==null&&(C=w.shared.hiddenCallbacks,w.shared.hiddenCallbacks=C===null?j:C.concat(j))));break;case 26:var q=r1;if(tm(_,w),dg(w),j&512&&(Ys||C===null||em(C,C.return)),j&4){var te=C!==null?C.memoizedState:null;if(j=w.memoizedState,C===null)if(j===null)if(w.stateNode===null){e:{j=w.type,C=w.memoizedProps,q=q.ownerDocument||q;t:switch(j){case"title":te=q.getElementsByTagName("title")[0],(!te||te[sy]||te[_l]||te.namespaceURI==="http://www.w3.org/2000/svg"||te.hasAttribute("itemprop"))&&(te=q.createElement(j),q.head.insertBefore(te,q.querySelector("head > title"))),Jd(te,j,C),te[_l]=w,Od(te),j=te;break e;case"link":var Se=K9("link","href",q).get(j+(C.href||""));if(Se){for(var Me=0;Me<Se.length;Me++)if(te=Se[Me],te.getAttribute("href")===(C.href==null||C.href===""?null:C.href)&&te.getAttribute("rel")===(C.rel==null?null:C.rel)&&te.getAttribute("title")===(C.title==null?null:C.title)&&te.getAttribute("crossorigin")===(C.crossOrigin==null?null:C.crossOrigin)){Se.splice(Me,1);break t}}te=q.createElement(j),Jd(te,j,C),q.head.appendChild(te);break;case"meta":if(Se=K9("meta","content",q).get(j+(C.content||""))){for(Me=0;Me<Se.length;Me++)if(te=Se[Me],te.getAttribute("content")===(C.content==null?null:""+C.content)&&te.getAttribute("name")===(C.name==null?null:C.name)&&te.getAttribute("property")===(C.property==null?null:C.property)&&te.getAttribute("http-equiv")===(C.httpEquiv==null?null:C.httpEquiv)&&te.getAttribute("charset")===(C.charSet==null?null:C.charSet)){Se.splice(Me,1);break t}}te=q.createElement(j),Jd(te,j,C),q.head.appendChild(te);break;default:throw Error(s(468,j))}te[_l]=w,Od(te),j=te}w.stateNode=j}else gN(q,w.type,w.stateNode);else w.stateNode=W9(q,j,w.memoizedProps);else te!==j?(te===null?C.stateNode!==null&&(C=C.stateNode,C.parentNode.removeChild(C)):te.count--,j===null?gN(q,w.type,w.stateNode):W9(q,j,w.memoizedProps)):j===null&&w.stateNode!==null&&KC(w,w.memoizedProps,C.memoizedProps)}break;case 27:tm(_,w),dg(w),j&512&&(Ys||C===null||em(C,C.return)),C!==null&&j&4&&KC(w,w.memoizedProps,C.memoizedProps);break;case 5:if(tm(_,w),dg(w),j&512&&(Ys||C===null||em(C,C.return)),w.flags&32){q=w.stateNode;try{gy(q,"")}catch(Zt){gu(w,w.return,Zt)}}j&4&&w.stateNode!=null&&(q=w.memoizedProps,KC(w,q,C!==null?C.memoizedProps:q)),j&1024&&(Ov=!0);break;case 6:if(tm(_,w),dg(w),j&4){if(w.stateNode===null)throw Error(s(162));j=w.memoizedProps,C=w.stateNode;try{C.nodeValue=j}catch(Zt){gu(w,w.return,Zt)}}break;case 3:if(By=null,q=r1,r1=fa(_.containerInfo),tm(_,w),r1=q,dg(w),j&4&&C!==null&&C.memoizedState.isDehydrated)try{EN(_.containerInfo)}catch(Zt){gu(w,w.return,Zt)}Ov&&(Ov=!1,o9(w));break;case 4:j=r1,r1=fa(w.stateNode.containerInfo),tm(_,w),dg(w),r1=j;break;case 12:tm(_,w),dg(w);break;case 13:tm(_,w),dg(w),w.child.flags&8192&&w.memoizedState!==null!=(C!==null&&C.memoizedState!==null)&&(y5=qi()),j&4&&(j=w.updateQueue,j!==null&&(w.updateQueue=null,f5(w,j)));break;case 22:q=w.memoizedState!==null;var pt=C!==null&&C.memoizedState!==null,Xt=c0,$n=Ys;if(c0=Xt||q,Ys=$n||pt,tm(_,w),Ys=$n,c0=Xt,dg(w),j&8192)e:for(_=w.stateNode,_._visibility=q?_._visibility&-2:_._visibility|1,q&&(C===null||pt||c0||Ys||Z2(w)),C=null,_=w;;){if(_.tag===5||_.tag===26){if(C===null){pt=C=_;try{if(te=pt.stateNode,q)Se=te.style,typeof Se.setProperty=="function"?Se.setProperty("display","none","important"):Se.display="none";else{Me=pt.stateNode;var Vn=pt.memoizedProps.style,ln=Vn!=null&&Vn.hasOwnProperty("display")?Vn.display:null;Me.style.display=ln==null||typeof ln=="boolean"?"":(""+ln).trim()}}catch(Zt){gu(pt,pt.return,Zt)}}}else if(_.tag===6){if(C===null){pt=_;try{pt.stateNode.nodeValue=q?"":pt.memoizedProps}catch(Zt){gu(pt,pt.return,Zt)}}}else if((_.tag!==22&&_.tag!==23||_.memoizedState===null||_===w)&&_.child!==null){_.child.return=_,_=_.child;continue}if(_===w)break e;for(;_.sibling===null;){if(_.return===null||_.return===w)break e;C===_&&(C=null),_=_.return}C===_&&(C=null),_.sibling.return=_.return,_=_.sibling}j&4&&(j=w.updateQueue,j!==null&&(C=j.retryQueue,C!==null&&(j.retryQueue=null,f5(w,C))));break;case 19:tm(_,w),dg(w),j&4&&(j=w.updateQueue,j!==null&&(w.updateQueue=null,f5(w,j)));break;case 30:break;case 21:break;default:tm(_,w),dg(w)}}function dg(w){var _=w.flags;if(_&2){try{for(var C,j=w.return;j!==null;){if(Sk(j)){C=j;break}j=j.return}if(C==null)throw Error(s(160));switch(C.tag){case 27:var q=C.stateNode,te=YC(w);XC(w,te,q);break;case 5:var Se=C.stateNode;C.flags&32&&(gy(Se,""),C.flags&=-33);var Me=YC(w);XC(w,Me,Se);break;case 3:case 4:var pt=C.stateNode.containerInfo,Xt=YC(w);J2(w,Xt,pt);break;default:throw Error(s(161))}}catch($n){gu(w,w.return,$n)}w.flags&=-3}_&4096&&(w.flags&=-4097)}function o9(w){if(w.subtreeFlags&1024)for(w=w.child;w!==null;){var _=w;o9(_),_.tag===5&&_.flags&1024&&_.stateNode.reset(),w=w.sibling}}function Ny(w,_){if(_.subtreeFlags&8772)for(_=_.child;_!==null;)ZC(w,_.alternate,_),_=_.sibling}function Z2(w){for(w=w.child;w!==null;){var _=w;switch(_.tag){case 0:case 11:case 14:case 15:Oy(4,_,_.return),Z2(_);break;case 1:em(_,_.return);var C=_.stateNode;typeof C.componentWillUnmount=="function"&&WC(_,_.return,C),Z2(_);break;case 27:c3(_.stateNode);case 26:case 5:em(_,_.return),Z2(_);break;case 22:_.memoizedState===null&&Z2(_);break;case 30:Z2(_);break;default:Z2(_)}w=w.sibling}}function Nv(w,_,C){for(C=C&&(_.subtreeFlags&8772)!==0,_=_.child;_!==null;){var j=_.alternate,q=w,te=_,Se=te.flags;switch(te.tag){case 0:case 11:case 15:Nv(q,te,C),VC(4,te);break;case 1:if(Nv(q,te,C),j=te,q=j.stateNode,typeof q.componentDidMount=="function")try{q.componentDidMount()}catch(Xt){gu(j,j.return,Xt)}if(j=te,q=j.updateQueue,q!==null){var Me=j.stateNode;try{var pt=q.shared.hiddenCallbacks;if(pt!==null)for(q.shared.hiddenCallbacks=null,q=0;q<pt.length;q++)N8(pt[q],Me)}catch(Xt){gu(j,j.return,Xt)}}C&&Se&64&&t9(te),Iv(te,te.return);break;case 27:JC(te);case 26:case 5:Nv(q,te,C),C&&j===null&&Se&4&&ZO(te),Iv(te,te.return);break;case 12:Nv(q,te,C);break;case 13:Nv(q,te,C),C&&Se&4&&i9(q,te);break;case 22:te.memoizedState===null&&Nv(q,te,C),Iv(te,te.return);break;case 30:break;default:Nv(q,te,C)}_=_.sibling}}function p5(w,_){var C=null;w!==null&&w.memoizedState!==null&&w.memoizedState.cachePool!==null&&(C=w.memoizedState.cachePool.pool),w=null,_.memoizedState!==null&&_.memoizedState.cachePool!==null&&(w=_.memoizedState.cachePool.pool),w!==C&&(w!=null&&w.refCount++,C!=null&&EC(C))}function g5(w,_){w=null,_.alternate!==null&&(w=_.alternate.memoizedState.cache),_=_.memoizedState.cache,_!==w&&(_.refCount++,w!=null&&EC(w))}function u0(w,_,C,j){if(_.subtreeFlags&10256)for(_=_.child;_!==null;)a9(w,_,C,j),_=_.sibling}function a9(w,_,C,j){var q=_.flags;switch(_.tag){case 0:case 11:case 15:u0(w,_,C,j),q&2048&&VC(9,_);break;case 1:u0(w,_,C,j);break;case 3:u0(w,_,C,j),q&2048&&(w=null,_.alternate!==null&&(w=_.alternate.memoizedState.cache),_=_.memoizedState.cache,_!==w&&(_.refCount++,w!=null&&EC(w)));break;case 12:if(q&2048){u0(w,_,C,j),w=_.stateNode;try{var te=_.memoizedProps,Se=te.id,Me=te.onPostCommit;typeof Me=="function"&&Me(Se,_.alternate===null?"mount":"update",w.passiveEffectDuration,-0)}catch(pt){gu(_,_.return,pt)}}else u0(w,_,C,j);break;case 13:u0(w,_,C,j);break;case 23:break;case 22:te=_.stateNode,Se=_.alternate,_.memoizedState!==null?te._visibility&2?u0(w,_,C,j):Q2(w,_):te._visibility&2?u0(w,_,C,j):(te._visibility|=2,ab(w,_,C,j,(_.subtreeFlags&10256)!==0)),q&2048&&p5(Se,_);break;case 24:u0(w,_,C,j),q&2048&&g5(_.alternate,_);break;default:u0(w,_,C,j)}}function ab(w,_,C,j,q){for(q=q&&(_.subtreeFlags&10256)!==0,_=_.child;_!==null;){var te=w,Se=_,Me=C,pt=j,Xt=Se.flags;switch(Se.tag){case 0:case 11:case 15:ab(te,Se,Me,pt,q),VC(8,Se);break;case 23:break;case 22:var $n=Se.stateNode;Se.memoizedState!==null?$n._visibility&2?ab(te,Se,Me,pt,q):Q2(te,Se):($n._visibility|=2,ab(te,Se,Me,pt,q)),q&&Xt&2048&&p5(Se.alternate,Se);break;case 24:ab(te,Se,Me,pt,q),q&&Xt&2048&&g5(Se.alternate,Se);break;default:ab(te,Se,Me,pt,q)}_=_.sibling}}function Q2(w,_){if(_.subtreeFlags&10256)for(_=_.child;_!==null;){var C=w,j=_,q=j.flags;switch(j.tag){case 22:Q2(C,j),q&2048&&p5(j.alternate,j);break;case 24:Q2(C,j),q&2048&&g5(j.alternate,j);break;default:Q2(C,j)}_=_.sibling}}var jh=8192;function _k(w){if(w.subtreeFlags&jh)for(w=w.child;w!==null;)s9(w),w=w.sibling}function s9(w){switch(w.tag){case 26:_k(w),w.flags&jh&&w.memoizedState!==null&&uz(r1,w.memoizedState,w.memoizedProps);break;case 5:_k(w);break;case 3:case 4:var _=r1;r1=fa(w.stateNode.containerInfo),_k(w),r1=_;break;case 22:w.memoizedState===null&&(_=w.alternate,_!==null&&_.memoizedState!==null?(_=jh,jh=16777216,_k(w),jh=_):_k(w));break;default:_k(w)}}function c9(w){var _=w.alternate;if(_!==null&&(w=_.child,w!==null)){_.child=null;do _=w.sibling,w.sibling=null,w=_;while(w!==null)}}function kk(w){var _=w.deletions;if((w.flags&16)!==0){if(_!==null)for(var C=0;C<_.length;C++){var j=_[C];Yd=j,l9(j,w)}c9(w)}if(w.subtreeFlags&10256)for(w=w.child;w!==null;)u9(w),w=w.sibling}function u9(w){switch(w.tag){case 0:case 11:case 15:kk(w),w.flags&2048&&Oy(9,w,w.return);break;case 3:kk(w);break;case 12:kk(w);break;case 22:var _=w.stateNode;w.memoizedState!==null&&_._visibility&2&&(w.return===null||w.return.tag!==13)?(_._visibility&=-3,QO(w)):kk(w);break;default:kk(w)}}function QO(w){var _=w.deletions;if((w.flags&16)!==0){if(_!==null)for(var C=0;C<_.length;C++){var j=_[C];Yd=j,l9(j,w)}c9(w)}for(w=w.child;w!==null;){switch(_=w,_.tag){case 0:case 11:case 15:Oy(8,_,_.return),QO(_);break;case 22:C=_.stateNode,C._visibility&2&&(C._visibility&=-3,QO(_));break;default:QO(_)}w=w.sibling}}function l9(w,_){for(;Yd!==null;){var C=Yd;switch(C.tag){case 0:case 11:case 15:Oy(8,C,_);break;case 23:case 22:if(C.memoizedState!==null&&C.memoizedState.cachePool!==null){var j=C.memoizedState.cachePool.pool;j!=null&&j.refCount++}break;case 24:EC(C.memoizedState.cache)}if(j=C.child,j!==null)j.return=C,Yd=j;else e:for(C=w;Yd!==null;){j=Yd;var q=j.sibling,te=j.return;if(r9(j),j===C){Yd=null;break e}if(q!==null){q.return=te,Yd=q;break e}Yd=te}}}var HU={getCacheForType:function(w){var _=Hl(Qc),C=_.data.get(w);return C===void 0&&(C=w(),_.data.set(w,C)),C}},m5=typeof WeakMap=="function"?WeakMap:Map,zc=0,Nu=null,as=null,$s=0,Hc=0,nm=null,Ry=!1,Tk=!1,b5=!1,Rv=0,ql=0,Py=0,eS=0,v5=0,sb=0,xk=0,QC=null,fg=null,w5=!1,y5=0,Ck=1/0,e3=null,Ly=null,ih=0,Dy=null,Ak=null,Ik=0,E5=0,S5=null,d9=null,Ok=0,_5=null;function $h(){if((zc&2)!==0&&$s!==0)return $s&-$s;if(_e.T!==null){var w=hk;return w!==0?w:aN()}return YI()}function f9(){sb===0&&(sb=($s&536870912)===0||vs?nh():536870912);var w=Qg.current;return w!==null&&(w.flags|=32),sb}function rm(w,_,C){(w===Nu&&(Hc===2||Hc===9)||w.cancelPendingCommit!==null)&&(Nk(w,0),My(w,$s,sb,!1)),Kb(w,C),((zc&2)===0||w!==Nu)&&(w===Nu&&((zc&2)===0&&(eS|=C),ql===4&&My(w,$s,sb,!1)),l0(w))}function h9(w,_,C){if((zc&6)!==0)throw Error(s(327));var j=!C&&(_&124)===0&&(_&w.expiredLanes)===0||Jc(w,_),q=j?VU(w,_):x5(w,_,!0),te=j;do{if(q===0){Tk&&!j&&My(w,_,0,!1);break}else{if(C=w.current.alternate,te&&!GU(C)){q=x5(w,_,!1),te=!1;continue}if(q===2){if(te=_,w.errorRecoveryDisabledLanes&te)var Se=0;else Se=w.pendingLanes&-536870913,Se=Se!==0?Se:Se&536870912?536870912:0;if(Se!==0){_=Se;e:{var Me=w;q=QC;var pt=Me.current.memoizedState.isDehydrated;if(pt&&(Nk(Me,Se).flags|=256),Se=x5(Me,Se,!1),Se!==2){if(b5&&!pt){Me.errorRecoveryDisabledLanes|=te,eS|=te,q=4;break e}te=fg,fg=q,te!==null&&(fg===null?fg=te:fg.push.apply(fg,te))}q=Se}if(te=!1,q!==2)continue}}if(q===1){Nk(w,0),My(w,_,0,!0);break}e:{switch(j=w,te=q,te){case 0:case 1:throw Error(s(345));case 4:if((_&4194048)!==_)break;case 6:My(j,_,sb,!Ry);break e;case 2:fg=null;break;case 3:case 5:break;default:throw Error(s(329))}if((_&62914560)===_&&(q=y5+300-qi(),10<q)){if(My(j,_,sb,!Ry),js(j,0,!0)!==0)break e;j.timeoutHandle=j9(p9.bind(null,j,C,fg,e3,w5,_,sb,eS,xk,Ry,te,2,-0,0),q);break e}p9(j,C,fg,e3,w5,_,sb,eS,xk,Ry,te,0,-0,0)}}break}while(!0);l0(w)}function p9(w,_,C,j,q,te,Se,Me,pt,Xt,$n,Vn,ln,Zt){if(w.timeoutHandle=-1,Vn=_.subtreeFlags,(Vn&8192||(Vn&16785408)===16785408)&&(Af={stylesheets:null,count:0,unsuspend:Y9},s9(_),Vn=lz(),Vn!==null)){w.cancelPendingCommit=Vn(t3.bind(null,w,_,te,C,j,q,Se,Me,pt,$n,1,ln,Zt)),My(w,te,Se,!Xt);return}t3(w,_,te,C,j,q,Se,Me,pt)}function GU(w){for(var _=w;;){var C=_.tag;if((C===0||C===11||C===15)&&_.flags&16384&&(C=_.updateQueue,C!==null&&(C=C.stores,C!==null)))for(var j=0;j<C.length;j++){var q=C[j],te=q.getSnapshot;q=q.value;try{if(!Xg(te(),q))return!1}catch{return!1}}if(C=_.child,_.subtreeFlags&16384&&C!==null)C.return=_,_=C;else{if(_===w)break;for(;_.sibling===null;){if(_.return===null||_.return===w)return!0;_=_.return}_.sibling.return=_.return,_=_.sibling}}return!0}function My(w,_,C,j){_&=~v5,_&=~eS,w.suspendedLanes|=_,w.pingedLanes&=~_,j&&(w.warmLanes|=_),j=w.expirationTimes;for(var q=_;0<q;){var te=31-Ei(q),Se=1<<te;j[te]=-1,q&=~Se}C!==0&&S2(w,C,_)}function eN(){return(zc&6)===0?(i3(0),!1):!0}function k5(){if(as!==null){if(Hc===0)var w=as.return;else w=as,De=_y=null,AO(w),Kd=null,cg=0,w=as;for(;w!==null;)JO(w.alternate,w),w=w.return;as=null}}function Nk(w,_){var C=w.timeoutHandle;C!==-1&&(w.timeoutHandle=-1,nz(C)),C=w.cancelPendingCommit,C!==null&&(w.cancelPendingCommit=null,C()),k5(),Nu=w,as=C=vp(w.current,null),$s=_,Hc=0,nm=null,Ry=!1,Tk=Jc(w,_),b5=!1,xk=sb=v5=eS=Py=ql=0,fg=QC=null,w5=!1,(_&8)!==0&&(_|=_&32);var j=w.entangledLanes;if(j!==0)for(w=w.entanglements,j&=_;0<j;){var q=31-Ei(j),te=1<<q;_|=w[q],j&=~te}return Rv=_,O2(),C}function g9(w,_){Wa=null,_e.H=jC,_===yp||_===TO?(_=I8(),Hc=3):_===C8?(_=I8(),Hc=4):Hc=_===Y8?8:_!==null&&typeof _=="object"&&typeof _.then=="function"?6:1,nm=_,as===null&&(ql=1,UC(w,bp(_,w.current)))}function m9(){var w=_e.H;return _e.H=jC,w===null?jC:w}function b9(){var w=_e.A;return _e.A=HU,w}function T5(){ql=4,Ry||($s&4194048)!==$s&&Qg.current!==null||(Tk=!0),(Py&134217727)===0&&(eS&134217727)===0||Nu===null||My(Nu,$s,sb,!1)}function x5(w,_,C){var j=zc;zc|=2;var q=m9(),te=b9();(Nu!==w||$s!==_)&&(e3=null,Nk(w,_)),_=!1;var Se=ql;e:do try{if(Hc!==0&&as!==null){var Me=as,pt=nm;switch(Hc){case 8:k5(),Se=6;break e;case 3:case 2:case 9:case 6:Qg.current===null&&(_=!0);var Xt=Hc;if(Hc=0,nm=null,Rk(w,Me,pt,Xt),C&&Tk){Se=0;break e}break;default:Xt=Hc,Hc=0,nm=null,Rk(w,Me,pt,Xt)}}qU(),Se=ql;break}catch($n){g9(w,$n)}while(!0);return _&&w.shellSuspendCounter++,De=_y=null,zc=j,_e.H=q,_e.A=te,as===null&&(Nu=null,$s=0,O2()),Se}function qU(){for(;as!==null;)C5(as)}function VU(w,_){var C=zc;zc|=2;var j=m9(),q=b9();Nu!==w||$s!==_?(e3=null,Ck=qi()+500,Nk(w,_)):Tk=Jc(w,_);e:do try{if(Hc!==0&&as!==null){_=as;var te=nm;t:switch(Hc){case 1:Hc=0,nm=null,Rk(w,_,te,1);break;case 2:case 9:if(A8(te)){Hc=0,nm=null,w9(_);break}_=function(){Hc!==2&&Hc!==9||Nu!==w||(Hc=7),l0(w)},te.then(_,_);break e;case 3:Hc=7;break e;case 4:Hc=5;break e;case 7:A8(te)?(Hc=0,nm=null,w9(_)):(Hc=0,nm=null,Rk(w,_,te,7));break;case 5:var Se=null;switch(as.tag){case 26:Se=as.memoizedState;case 5:case 27:var Me=as;if(!Se||Zd(Se)){Hc=0,nm=null;var pt=Me.sibling;if(pt!==null)as=pt;else{var Xt=Me.return;Xt!==null?(as=Xt,tN(Xt)):as=null}break t}}Hc=0,nm=null,Rk(w,_,te,5);break;case 6:Hc=0,nm=null,Rk(w,_,te,6);break;case 8:k5(),ql=6;break e;default:throw Error(s(462))}}v9();break}catch($n){g9(w,$n)}while(!0);return De=_y=null,_e.H=j,_e.A=q,zc=C,as!==null?0:(Nu=null,$s=0,O2(),ql)}function v9(){for(;as!==null&&!Ko();)C5(as)}function C5(w){var _=KO(w.alternate,w,Rv);w.memoizedProps=w.pendingProps,_===null?tN(w):as=_}function w9(w){var _=w,C=_.alternate;switch(_.tag){case 15:case 0:_=VO(C,_,_.pendingProps,_.type,void 0,$s);break;case 11:_=VO(C,_,_.pendingProps,_.type.render,_.ref,$s);break;case 5:AO(_);default:JO(C,_),_=as=mC(_,Rv),_=KO(C,_,Rv)}w.memoizedProps=w.pendingProps,_===null?tN(w):as=_}function Rk(w,_,C,j){De=_y=null,AO(_),Kd=null,cg=0;var q=_.return;try{if(s5(w,q,_,C,$s)){ql=1,UC(w,bp(C,w.current)),as=null;return}}catch(te){if(q!==null)throw as=q,te;ql=1,UC(w,bp(C,w.current)),as=null;return}_.flags&32768?(vs||j===1?w=!0:Tk||($s&536870912)!==0?w=!1:(Ry=w=!0,(j===2||j===9||j===3||j===6)&&(j=Qg.current,j!==null&&j.tag===13&&(j.flags|=16384))),y9(_,w)):tN(_)}function tN(w){var _=w;do{if((_.flags&32768)!==0){y9(_,Ry);return}w=_.return;var C=Q8(_.alternate,_,Rv);if(C!==null){as=C;return}if(_=_.sibling,_!==null){as=_;return}as=_=w}while(_!==null);ql===0&&(ql=5)}function y9(w,_){do{var C=e9(w.alternate,w);if(C!==null){C.flags&=32767,as=C;return}if(C=w.return,C!==null&&(C.flags|=32768,C.subtreeFlags=0,C.deletions=null),!_&&(w=w.sibling,w!==null)){as=w;return}as=w=C}while(w!==null);ql=6,as=null}function t3(w,_,C,j,q,te,Se,Me,pt){w.cancelPendingCommit=null;do nN();while(ih!==0);if((zc&6)!==0)throw Error(s(327));if(_!==null){if(_===w.current)throw Error(s(177));if(te=_.lanes|_.childLanes,te|=O6,Km(w,C,te,Se,Me,pt),w===Nu&&(as=Nu=null,$s=0),Ak=_,Dy=w,Ik=C,E5=te,S5=q,d9=j,(_.subtreeFlags&10256)!==0||(_.flags&10256)!==0?(w.callbackNode=null,w.callbackPriority=0,A9(Fc,function(){return T9(),null})):(w.callbackNode=null,w.callbackPriority=0),j=(_.flags&13878)!==0,(_.subtreeFlags&13878)!==0||j){j=_e.T,_e.T=null,q=qe.p,qe.p=2,Se=zc,zc|=4;try{UU(w,_,C)}finally{zc=Se,qe.p=q,_e.T=j}}ih=1,E9(),S9(),_9()}}function E9(){if(ih===1){ih=0;var w=Dy,_=Ak,C=(_.flags&13878)!==0;if((_.subtreeFlags&13878)!==0||C){C=_e.T,_e.T=null;var j=qe.p;qe.p=2;var q=zc;zc|=4;try{h5(_,w);var te=lN,Se=m8(w.containerInfo),Me=te.focusedElem,pt=te.selectionRange;if(Se!==Me&&Me&&Me.ownerDocument&&g8(Me.ownerDocument.documentElement,Me)){if(pt!==null&&A6(Me)){var Xt=pt.start,$n=pt.end;if($n===void 0&&($n=Xt),"selectionStart"in Me)Me.selectionStart=Xt,Me.selectionEnd=Math.min($n,Me.value.length);else{var Vn=Me.ownerDocument||document,ln=Vn&&Vn.defaultView||window;if(ln.getSelection){var Zt=ln.getSelection(),Do=Me.textContent.length,Oo=Math.min(pt.start,Do),mu=pt.end===void 0?Oo:Math.min(pt.end,Do);!Zt.extend&&Oo>mu&&(Se=mu,mu=Oo,Oo=Se);var zt=C6(Me,Oo),Ot=C6(Me,mu);if(zt&&Ot&&(Zt.rangeCount!==1||Zt.anchorNode!==zt.node||Zt.anchorOffset!==zt.offset||Zt.focusNode!==Ot.node||Zt.focusOffset!==Ot.offset)){var Wt=Vn.createRange();Wt.setStart(zt.node,zt.offset),Zt.removeAllRanges(),Oo>mu?(Zt.addRange(Wt),Zt.extend(Ot.node,Ot.offset)):(Wt.setEnd(Ot.node,Ot.offset),Zt.addRange(Wt))}}}}for(Vn=[],Zt=Me;Zt=Zt.parentNode;)Zt.nodeType===1&&Vn.push({element:Zt,left:Zt.scrollLeft,top:Zt.scrollTop});for(typeof Me.focus=="function"&&Me.focus(),Me=0;Me<Vn.length;Me++){var zn=Vn[Me];zn.element.scrollLeft=zn.left,zn.element.scrollTop=zn.top}}oS=!!i1,lN=i1=null}finally{zc=q,qe.p=j,_e.T=C}}w.current=_,ih=2}}function S9(){if(ih===2){ih=0;var w=Dy,_=Ak,C=(_.flags&8772)!==0;if((_.subtreeFlags&8772)!==0||C){C=_e.T,_e.T=null;var j=qe.p;qe.p=2;var q=zc;zc|=4;try{ZC(w,_.alternate,_)}finally{zc=q,qe.p=j,_e.T=C}}ih=3}}function _9(){if(ih===4||ih===3){ih=0,Ca();var w=Dy,_=Ak,C=Ik,j=d9;(_.subtreeFlags&10256)!==0||(_.flags&10256)!==0?ih=5:(ih=0,Ak=Dy=null,k9(w,w.pendingLanes));var q=w.pendingLanes;if(q===0&&(Ly=null),K_(C),_=_.stateNode,Pn&&typeof Pn.onCommitFiberRoot=="function")try{Pn.onCommitFiberRoot(Sn,_,void 0,(_.current.flags&128)===128)}catch{}if(j!==null){_=_e.T,q=qe.p,qe.p=2,_e.T=null;try{for(var te=w.onRecoverableError,Se=0;Se<j.length;Se++){var Me=j[Se];te(Me.value,{componentStack:Me.stack})}}finally{_e.T=_,qe.p=q}}(Ik&3)!==0&&nN(),l0(w),q=w.pendingLanes,(C&4194090)!==0&&(q&42)!==0?w===_5?Ok++:(Ok=0,_5=w):Ok=0,i3(0)}}function k9(w,_){(w.pooledCacheLanes&=_)===0&&(_=w.pooledCache,_!=null&&(w.pooledCache=null,EC(_)))}function nN(w){return E9(),S9(),_9(),T9()}function T9(){if(ih!==5)return!1;var w=Dy,_=E5;E5=0;var C=K_(Ik),j=_e.T,q=qe.p;try{qe.p=32>C?32:C,_e.T=null,C=S5,S5=null;var te=Dy,Se=Ik;if(ih=0,Ak=Dy=null,Ik=0,(zc&6)!==0)throw Error(s(331));var Me=zc;if(zc|=4,u9(te.current),a9(te,te.current,Se,C),zc=Me,i3(0,!1),Pn&&typeof Pn.onPostCommitFiberRoot=="function")try{Pn.onPostCommitFiberRoot(Sn,te)}catch{}return!0}finally{qe.p=q,_e.T=j,k9(w,_)}}function x9(w,_,C){_=bp(C,_),_=ug(w.stateNode,_,2),w=yv(w,_,2),w!==null&&(Kb(w,2),l0(w))}function gu(w,_,C){if(w.tag===3)x9(w,w,C);else for(;_!==null;){if(_.tag===3){x9(_,w,C);break}else if(_.tag===1){var j=_.stateNode;if(typeof _.type.getDerivedStateFromError=="function"||typeof j.componentDidCatch=="function"&&(Ly===null||!Ly.has(j))){w=bp(C,w),C=a5(2),j=yv(_,C,2),j!==null&&(xv(C,j,_,w),Kb(j,2),l0(j));break}}_=_.return}}function A5(w,_,C){var j=w.pingCache;if(j===null){j=w.pingCache=new m5;var q=new Set;j.set(_,q)}else q=j.get(_),q===void 0&&(q=new Set,j.set(_,q));q.has(C)||(b5=!0,q.add(C),w=WU.bind(null,w,_,C),_.then(w,w))}function WU(w,_,C){var j=w.pingCache;j!==null&&j.delete(_),w.pingedLanes|=w.suspendedLanes&C,w.warmLanes&=~C,Nu===w&&($s&C)===C&&(ql===4||ql===3&&($s&62914560)===$s&&300>qi()-y5?(zc&2)===0&&Nk(w,0):v5|=C,xk===$s&&(xk=0)),l0(w)}function C9(w,_){_===0&&(_=hu()),w=N2(w,_),w!==null&&(Kb(w,_),l0(w))}function rN(w){var _=w.memoizedState,C=0;_!==null&&(C=_.retryLane),C9(w,C)}function KU(w,_){var C=0;switch(w.tag){case 13:var j=w.stateNode,q=w.memoizedState;q!==null&&(C=q.retryLane);break;case 19:j=w.stateNode;break;case 22:j=w.stateNode._retryCache;break;default:throw Error(s(314))}j!==null&&j.delete(_),C9(w,C)}function A9(w,_){return Ri(w,_)}var n3=null,jy=null,iN=!1,r3=!1,oN=!1,$y=0;function l0(w){w!==jy&&w.next===null&&(jy===null?n3=jy=w:jy=jy.next=w),r3=!0,iN||(iN=!0,I5())}function i3(w,_){if(!oN&&r3){oN=!0;do for(var C=!1,j=n3;j!==null;){if(w!==0){var q=j.pendingLanes;if(q===0)var te=0;else{var Se=j.suspendedLanes,Me=j.pingedLanes;te=(1<<31-Ei(42|w)+1)-1,te&=q&~(Se&~Me),te=te&201326741?te&201326741|1:te?te|2:0}te!==0&&(C=!0,O9(j,te))}else te=$s,te=js(j,j===Nu?te:0,j.cancelPendingCommit!==null||j.timeoutHandle!==-1),(te&3)===0||Jc(j,te)||(C=!0,O9(j,te));j=j.next}while(C);oN=!1}}function YU(){o3()}function o3(){r3=iN=!1;var w=0;$y!==0&&(tz()&&(w=$y),$y=0);for(var _=qi(),C=null,j=n3;j!==null;){var q=j.next,te=a3(j,_);te===0?(j.next=null,C===null?n3=q:C.next=q,q===null&&(jy=C)):(C=j,(w!==0||(te&3)!==0)&&(r3=!0)),j=q}i3(w)}function a3(w,_){for(var C=w.suspendedLanes,j=w.pingedLanes,q=w.expirationTimes,te=w.pendingLanes&-62914561;0<te;){var Se=31-Ei(te),Me=1<<Se,pt=q[Se];pt===-1?((Me&C)===0||(Me&j)!==0)&&(q[Se]=Id(Me,_)):pt<=_&&(w.expiredLanes|=Me),te&=~Me}if(_=Nu,C=$s,C=js(w,w===_?C:0,w.cancelPendingCommit!==null||w.timeoutHandle!==-1),j=w.callbackNode,C===0||w===_&&(Hc===2||Hc===9)||w.cancelPendingCommit!==null)return j!==null&&j!==null&&ro(j),w.callbackNode=null,w.callbackPriority=0;if((C&3)===0||Jc(w,C)){if(_=C&-C,_===w.callbackPriority)return _;switch(j!==null&&ro(j),K_(C)){case 2:case 8:C=Eo;break;case 32:C=Fc;break;case 268435456:C=zl;break;default:C=Fc}return j=I9.bind(null,w),C=Ri(C,j),w.callbackPriority=_,w.callbackNode=C,_}return j!==null&&j!==null&&ro(j),w.callbackPriority=2,w.callbackNode=null,2}function I9(w,_){if(ih!==0&&ih!==5)return w.callbackNode=null,w.callbackPriority=0,null;var C=w.callbackNode;if(nN()&&w.callbackNode!==C)return null;var j=$s;return j=js(w,w===Nu?j:0,w.cancelPendingCommit!==null||w.timeoutHandle!==-1),j===0?null:(h9(w,j,_),a3(w,qi()),w.callbackNode!=null&&w.callbackNode===C?I9.bind(null,w):null)}function O9(w,_){if(nN())return null;h9(w,_,!0)}function I5(){rz(function(){(zc&6)!==0?Ri(lo,YU):o3()})}function aN(){return $y===0&&($y=nh()),$y}function N9(w){return w==null||typeof w=="symbol"||typeof w=="boolean"?null:typeof w=="function"?w:sC(""+w)}function R9(w,_){var C=_.ownerDocument.createElement("input");return C.name=_.name,C.value=_.value,w.id&&C.setAttribute("form",w.id),_.parentNode.insertBefore(C,_),w=new FormData(w),C.parentNode.removeChild(C),w}function XU(w,_,C,j,q){if(_==="submit"&&C&&C.stateNode===q){var te=N9((q[rh]||null).action),Se=j.submitter;Se&&(_=(_=Se[rh]||null)?N9(_.formAction):Se.getAttribute("formAction"),_!==null&&(te=_,Se=null));var Me=new oO("action","action",null,j,q);w.push({event:Me,listeners:[{instance:null,listener:function(){if(j.defaultPrevented){if($y!==0){var pt=Se?R9(q,Se):new FormData(q);DC(C,{pending:!0,data:pt,method:q.method,action:te},null,pt)}}else typeof te=="function"&&(Me.preventDefault(),pt=Se?R9(q,Se):new FormData(q),DC(C,{pending:!0,data:pt,method:q.method,action:te},te,pt))},currentTarget:q}]})}}for(var O5=0;O5<bc.length;O5++){var N5=bc[O5],JU=N5.toLowerCase(),P9=N5[0].toUpperCase()+N5.slice(1);Yb(JU,"on"+P9)}Yb(bO,"onAnimationEnd"),Yb(b8,"onAnimationIteration"),Yb(Jg,"onAnimationStart"),Yb("dblclick","onDoubleClick"),Yb("focusin","onFocus"),Yb("focusout","onBlur"),Yb(vO,"onTransitionRun"),Yb(LU,"onTransitionStart"),Yb(sk,"onTransitionCancel"),Yb(fC,"onTransitionEnd"),dy("onMouseEnter",["mouseout","mouseover"]),dy("onMouseLeave",["mouseout","mouseover"]),dy("onPointerEnter",["pointerout","pointerover"]),dy("onPointerLeave",["pointerout","pointerover"]),hv("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),hv("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),hv("onBeforeInput",["compositionend","keypress","textInput","paste"]),hv("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),hv("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),hv("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var s3="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),ZU=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(s3));function sN(w,_){_=(_&4)!==0;for(var C=0;C<w.length;C++){var j=w[C],q=j.event;j=j.listeners;e:{var te=void 0;if(_)for(var Se=j.length-1;0<=Se;Se--){var Me=j[Se],pt=Me.instance,Xt=Me.currentTarget;if(Me=Me.listener,pt!==te&&q.isPropagationStopped())break e;te=Me,q.currentTarget=Xt;try{te(q)}catch($n){Y2($n)}q.currentTarget=null,te=pt}else for(Se=0;Se<j.length;Se++){if(Me=j[Se],pt=Me.instance,Xt=Me.currentTarget,Me=Me.listener,pt!==te&&q.isPropagationStopped())break e;te=Me,q.currentTarget=Xt;try{te(q)}catch($n){Y2($n)}q.currentTarget=null,te=pt}}}}function ss(w,_){var C=_[k2];C===void 0&&(C=_[k2]=new Set);var j=w+"__bubble";C.has(j)||(L9(_,w,2,!1),C.add(j))}function R5(w,_,C){var j=0;_&&(j|=4),L9(C,w,j,_)}var hg="_reactListening"+Math.random().toString(36).slice(2);function P5(w){if(!w[hg]){w[hg]=!0,JI.forEach(function(C){C!=="selectionchange"&&(ZU.has(C)||R5(C,!1,w),R5(C,!0,w))});var _=w.nodeType===9?w:w.ownerDocument;_===null||_[hg]||(_[hg]=!0,R5("selectionchange",!1,_))}}function L9(w,_,C,j){switch(hz(_)){case 2:var q=d3;break;case 8:q=oh;break;default:q=q5}C=q.bind(null,_,C,w),q=void 0,!uC||_!=="touchstart"&&_!=="touchmove"&&_!=="wheel"||(q=!0),j?q!==void 0?w.addEventListener(_,C,{capture:!0,passive:q}):w.addEventListener(_,C,!0):q!==void 0?w.addEventListener(_,C,{passive:q}):w.addEventListener(_,C,!1)}function Pv(w,_,C,j,q){var te=j;if((_&1)===0&&(_&2)===0&&j!==null)e:for(;;){if(j===null)return;var Se=j.tag;if(Se===3||Se===4){var Me=j.stateNode.containerInfo;if(Me===q)break;if(Se===4)for(Se=j.return;Se!==null;){var pt=Se.tag;if((pt===3||pt===4)&&Se.stateNode.containerInfo===q)return;Se=Se.return}for(;Me!==null;){if(Se=fv(Me),Se===null)return;if(pt=Se.tag,pt===5||pt===6||pt===26||pt===27){j=te=Se;continue e}Me=Me.parentNode}}j=j.return}XM(function(){var Xt=te,$n=g6(C),Vn=[];e:{var ln=hC.get(w);if(ln!==void 0){var Zt=oO,Do=w;switch(w){case"keypress":if(Zc(C)===0)break e;case"keydown":case"keyup":Zt=E6;break;case"focusin":Do="focus",Zt=sO;break;case"focusout":Do="blur",Zt=sO;break;case"beforeblur":case"afterblur":Zt=sO;break;case"click":if(C.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":Zt=by;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":Zt=y6;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":Zt=TU;break;case bO:case b8:case Jg:Zt=t8;break;case fC:Zt=xU;break;case"scroll":case"scrollend":Zt=SU;break;case"wheel":Zt=AU;break;case"copy":case"cut":case"paste":Zt=r8;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":Zt=gv;break;case"toggle":case"beforetoggle":Zt=uO}var Oo=(_&4)!==0,mu=!Oo&&(w==="scroll"||w==="scrollend"),zt=Oo?ln!==null?ln+"Capture":null:ln;Oo=[];for(var Ot=Xt,Wt;Ot!==null;){var zn=Ot;if(Wt=zn.stateNode,zn=zn.tag,zn!==5&&zn!==26&&zn!==27||Wt===null||zt===null||(zn=cC(Ot,zt),zn!=null&&Oo.push(Xo(Ot,zn,Wt))),mu)break;Ot=Ot.return}0<Oo.length&&(ln=new Zt(ln,Do,null,C,$n),Vn.push({event:ln,listeners:Oo}))}}if((_&7)===0){e:{if(ln=w==="mouseover"||w==="pointerover",Zt=w==="mouseout"||w==="pointerout",ln&&C!==rO&&(Do=C.relatedTarget||C.fromElement)&&(fv(Do)||Do[t0]))break e;if((Zt||ln)&&(ln=$n.window===$n?$n:(ln=$n.ownerDocument)?ln.defaultView||ln.parentWindow:window,Zt?(Do=C.relatedTarget||C.toElement,Zt=Xt,Do=Do?fv(Do):null,Do!==null&&(mu=d(Do),Oo=Do.tag,Do!==mu||Oo!==5&&Oo!==27&&Oo!==6)&&(Do=null)):(Zt=null,Do=Xt),Zt!==Do)){if(Oo=by,zn="onMouseLeave",zt="onMouseEnter",Ot="mouse",(w==="pointerout"||w==="pointerover")&&(Oo=gv,zn="onPointerLeave",zt="onPointerEnter",Ot="pointer"),mu=Zt==null?ln:uy(Zt),Wt=Do==null?ln:uy(Do),ln=new Oo(zn,Ot+"leave",Zt,C,$n),ln.target=mu,ln.relatedTarget=Wt,zn=null,fv($n)===Xt&&(Oo=new Oo(zt,Ot+"enter",Do,C,$n),Oo.target=Wt,Oo.relatedTarget=mu,zn=Oo),mu=zn,Zt&&Do)t:{for(Oo=Zt,zt=Do,Ot=0,Wt=Oo;Wt;Wt=Xd(Wt))Ot++;for(Wt=0,zn=zt;zn;zn=Xd(zn))Wt++;for(;0<Ot-Wt;)Oo=Xd(Oo),Ot--;for(;0<Wt-Ot;)zt=Xd(zt),Wt--;for(;Ot--;){if(Oo===zt||zt!==null&&Oo===zt.alternate)break t;Oo=Xd(Oo),zt=Xd(zt)}Oo=null}else Oo=null;Zt!==null&&Vl(Vn,ln,Zt,Oo,!1),Do!==null&&mu!==null&&Vl(Vn,mu,Do,Oo,!0)}}e:{if(ln=Xt?uy(Xt):window,Zt=ln.nodeName&&ln.nodeName.toLowerCase(),Zt==="select"||Zt==="input"&&ln.type==="file")var bi=l8;else if(hO(ln))if(d8)bi=NU;else{bi=wy;var Ka=p8}else Zt=ln.nodeName,!Zt||Zt.toLowerCase()!=="input"||ln.type!=="checkbox"&&ln.type!=="radio"?Xt&&p6(Xt.elementType)&&(bi=l8):bi=mO;if(bi&&(bi=bi(w,Xt))){rk(Vn,bi,C,$n);break e}Ka&&Ka(w,ln,Xt),w==="focusout"&&Xt&&ln.type==="number"&&Xt.memoizedProps.value!=null&&aC(ln,"number",ln.value)}switch(Ka=Xt?uy(Xt):window,w){case"focusin":(hO(Ka)||Ka.contentEditable==="true")&&(ak=Ka,yy=Xt,Dt=null);break;case"focusout":Dt=yy=ak=null;break;case"mousedown":Rn=!0;break;case"contextmenu":case"mouseup":case"dragend":Rn=!1,En(Vn,C,$n);break;case"selectionchange":if(PU)break;case"keydown":case"keyup":En(Vn,C,$n)}var Hi;if(k6)e:{switch(w){case"compositionstart":var Vi="onCompositionStart";break e;case"compositionend":Vi="onCompositionEnd";break e;case"compositionupdate":Vi="onCompositionUpdate";break e}Vi=void 0}else vy?c8(w,C)&&(Vi="onCompositionEnd"):w==="keydown"&&C.keyCode===229&&(Vi="onCompositionStart");Vi&&(lO&&C.locale!=="ko"&&(vy||Vi!=="onCompositionStart"?Vi==="onCompositionEnd"&&vy&&(Hi=JM()):(my=$n,v6="value"in my?my.value:my.textContent,vy=!0)),Ka=cN(Xt,Vi),0<Ka.length&&(Vi=new tg(Vi,w,null,C,$n),Vn.push({event:Vi,listeners:Ka}),Hi?Vi.data=Hi:(Hi=fO(C),Hi!==null&&(Vi.data=Hi)))),(Hi=s8?IU(w,C):u8(w,C))&&(Vi=cN(Xt,"onBeforeInput"),0<Vi.length&&(Ka=new tg("onBeforeInput","beforeinput",null,C,$n),Vn.push({event:Ka,listeners:Vi}),Ka.data=Hi)),XU(Vn,w,Xt,C,$n)}sN(Vn,_)})}function Xo(w,_,C){return{instance:w,listener:_,currentTarget:C}}function cN(w,_){for(var C=_+"Capture",j=[];w!==null;){var q=w,te=q.stateNode;if(q=q.tag,q!==5&&q!==26&&q!==27||te===null||(q=cC(w,C),q!=null&&j.unshift(Xo(w,q,te)),q=cC(w,_),q!=null&&j.push(Xo(w,q,te))),w.tag===3)return j;w=w.return}return[]}function Xd(w){if(w===null)return null;do w=w.return;while(w&&w.tag!==5&&w.tag!==27);return w||null}function Vl(w,_,C,j,q){for(var te=_._reactName,Se=[];C!==null&&C!==j;){var Me=C,pt=Me.alternate,Xt=Me.stateNode;if(Me=Me.tag,pt!==null&&pt===j)break;Me!==5&&Me!==26&&Me!==27||Xt===null||(pt=Xt,q?(Xt=cC(C,te),Xt!=null&&Se.unshift(Xo(C,Xt,pt))):q||(Xt=cC(C,te),Xt!=null&&Se.push(Xo(C,Xt,pt)))),C=C.return}Se.length!==0&&w.push({event:_,listeners:Se})}var QU=/\r\n?/g,Fy=/\u0000|\uFFFD/g;function D9(w){return(typeof w=="string"?w:""+w).replace(QU,`
`).replace(Fy,"")}function M9(w,_){return _=D9(_),D9(w)===_}function uN(){}function eu(w,_,C,j,q,te){switch(C){case"children":typeof j=="string"?_==="body"||_==="textarea"&&j===""||gy(w,j):(typeof j=="number"||typeof j=="bigint")&&_!=="body"&&gy(w,""+j);break;case"className":Y_(w,"class",j);break;case"tabIndex":Y_(w,"tabindex",j);break;case"dir":case"role":case"viewBox":case"width":case"height":Y_(w,C,j);break;case"style":h6(w,j,te);break;case"data":if(_!=="object"){Y_(w,"data",j);break}case"src":case"href":if(j===""&&(_!=="a"||C!=="href")){w.removeAttribute(C);break}if(j==null||typeof j=="function"||typeof j=="symbol"||typeof j=="boolean"){w.removeAttribute(C);break}j=sC(""+j),w.setAttribute(C,j);break;case"action":case"formAction":if(typeof j=="function"){w.setAttribute(C,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof te=="function"&&(C==="formAction"?(_!=="input"&&eu(w,_,"name",q.name,q,null),eu(w,_,"formEncType",q.formEncType,q,null),eu(w,_,"formMethod",q.formMethod,q,null),eu(w,_,"formTarget",q.formTarget,q,null)):(eu(w,_,"encType",q.encType,q,null),eu(w,_,"method",q.method,q,null),eu(w,_,"target",q.target,q,null)));if(j==null||typeof j=="symbol"||typeof j=="boolean"){w.removeAttribute(C);break}j=sC(""+j),w.setAttribute(C,j);break;case"onClick":j!=null&&(w.onclick=uN);break;case"onScroll":j!=null&&ss("scroll",w);break;case"onScrollEnd":j!=null&&ss("scrollend",w);break;case"dangerouslySetInnerHTML":if(j!=null){if(typeof j!="object"||!("__html"in j))throw Error(s(61));if(C=j.__html,C!=null){if(q.children!=null)throw Error(s(60));w.innerHTML=C}}break;case"multiple":w.multiple=j&&typeof j!="function"&&typeof j!="symbol";break;case"muted":w.muted=j&&typeof j!="function"&&typeof j!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(j==null||typeof j=="function"||typeof j=="boolean"||typeof j=="symbol"){w.removeAttribute("xlink:href");break}C=sC(""+j),w.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",C);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":j!=null&&typeof j!="function"&&typeof j!="symbol"?w.setAttribute(C,""+j):w.removeAttribute(C);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":j&&typeof j!="function"&&typeof j!="symbol"?w.setAttribute(C,""):w.removeAttribute(C);break;case"capture":case"download":j===!0?w.setAttribute(C,""):j!==!1&&j!=null&&typeof j!="function"&&typeof j!="symbol"?w.setAttribute(C,j):w.removeAttribute(C);break;case"cols":case"rows":case"size":case"span":j!=null&&typeof j!="function"&&typeof j!="symbol"&&!isNaN(j)&&1<=j?w.setAttribute(C,j):w.removeAttribute(C);break;case"rowSpan":case"start":j==null||typeof j=="function"||typeof j=="symbol"||isNaN(j)?w.removeAttribute(C):w.setAttribute(C,j);break;case"popover":ss("beforetoggle",w),ss("toggle",w),fy(w,"popover",j);break;case"xlinkActuate":Ym(w,"http://www.w3.org/1999/xlink","xlink:actuate",j);break;case"xlinkArcrole":Ym(w,"http://www.w3.org/1999/xlink","xlink:arcrole",j);break;case"xlinkRole":Ym(w,"http://www.w3.org/1999/xlink","xlink:role",j);break;case"xlinkShow":Ym(w,"http://www.w3.org/1999/xlink","xlink:show",j);break;case"xlinkTitle":Ym(w,"http://www.w3.org/1999/xlink","xlink:title",j);break;case"xlinkType":Ym(w,"http://www.w3.org/1999/xlink","xlink:type",j);break;case"xmlBase":Ym(w,"http://www.w3.org/XML/1998/namespace","xml:base",j);break;case"xmlLang":Ym(w,"http://www.w3.org/XML/1998/namespace","xml:lang",j);break;case"xmlSpace":Ym(w,"http://www.w3.org/XML/1998/namespace","xml:space",j);break;case"is":fy(w,"is",j);break;case"innerText":case"textContent":break;default:(!(2<C.length)||C[0]!=="o"&&C[0]!=="O"||C[1]!=="n"&&C[1]!=="N")&&(C=KM.get(C)||C,fy(w,C,j))}}function L5(w,_,C,j,q,te){switch(C){case"style":h6(w,j,te);break;case"dangerouslySetInnerHTML":if(j!=null){if(typeof j!="object"||!("__html"in j))throw Error(s(61));if(C=j.__html,C!=null){if(q.children!=null)throw Error(s(60));w.innerHTML=C}}break;case"children":typeof j=="string"?gy(w,j):(typeof j=="number"||typeof j=="bigint")&&gy(w,""+j);break;case"onScroll":j!=null&&ss("scroll",w);break;case"onScrollEnd":j!=null&&ss("scrollend",w);break;case"onClick":j!=null&&(w.onclick=uN);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!ZI.hasOwnProperty(C))e:{if(C[0]==="o"&&C[1]==="n"&&(q=C.endsWith("Capture"),_=C.slice(2,q?C.length-7:void 0),te=w[rh]||null,te=te!=null?te[C]:null,typeof te=="function"&&w.removeEventListener(_,te,q),typeof j=="function")){typeof te!="function"&&te!==null&&(C in w?w[C]=null:w.hasAttribute(C)&&w.removeAttribute(C)),w.addEventListener(_,j,q);break e}C in w?w[C]=j:j===!0?w.setAttribute(C,""):fy(w,C,j)}}}function Jd(w,_,C){switch(_){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":ss("error",w),ss("load",w);var j=!1,q=!1,te;for(te in C)if(C.hasOwnProperty(te)){var Se=C[te];if(Se!=null)switch(te){case"src":j=!0;break;case"srcSet":q=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(s(137,_));default:eu(w,_,te,Se,C,null)}}q&&eu(w,_,"srcSet",C.srcSet,C,null),j&&eu(w,_,"src",C.src,C,null);return;case"input":ss("invalid",w);var Me=te=Se=q=null,pt=null,Xt=null;for(j in C)if(C.hasOwnProperty(j)){var $n=C[j];if($n!=null)switch(j){case"name":q=$n;break;case"type":Se=$n;break;case"checked":pt=$n;break;case"defaultChecked":Xt=$n;break;case"value":te=$n;break;case"defaultValue":Me=$n;break;case"children":case"dangerouslySetInnerHTML":if($n!=null)throw Error(s(137,_));break;default:eu(w,_,j,$n,C,null)}}l6(w,te,Me,pt,Xt,Se,q,!1),eO(w);return;case"select":ss("invalid",w),j=Se=te=null;for(q in C)if(C.hasOwnProperty(q)&&(Me=C[q],Me!=null))switch(q){case"value":te=Me;break;case"defaultValue":Se=Me;break;case"multiple":j=Me;default:eu(w,_,q,Me,C,null)}_=te,C=Se,w.multiple=!!j,_!=null?X_(w,!!j,_,!1):C!=null&&X_(w,!!j,C,!0);return;case"textarea":ss("invalid",w),te=q=j=null;for(Se in C)if(C.hasOwnProperty(Se)&&(Me=C[Se],Me!=null))switch(Se){case"value":j=Me;break;case"defaultValue":q=Me;break;case"children":te=Me;break;case"dangerouslySetInnerHTML":if(Me!=null)throw Error(s(91));break;default:eu(w,_,Se,Me,C,null)}nO(w,j,q,te),eO(w);return;case"option":for(pt in C)if(C.hasOwnProperty(pt)&&(j=C[pt],j!=null))switch(pt){case"selected":w.selected=j&&typeof j!="function"&&typeof j!="symbol";break;default:eu(w,_,pt,j,C,null)}return;case"dialog":ss("beforetoggle",w),ss("toggle",w),ss("cancel",w),ss("close",w);break;case"iframe":case"object":ss("load",w);break;case"video":case"audio":for(j=0;j<s3.length;j++)ss(s3[j],w);break;case"image":ss("error",w),ss("load",w);break;case"details":ss("toggle",w);break;case"embed":case"source":case"link":ss("error",w),ss("load",w);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(Xt in C)if(C.hasOwnProperty(Xt)&&(j=C[Xt],j!=null))switch(Xt){case"children":case"dangerouslySetInnerHTML":throw Error(s(137,_));default:eu(w,_,Xt,j,C,null)}return;default:if(p6(_)){for($n in C)C.hasOwnProperty($n)&&(j=C[$n],j!==void 0&&L5(w,_,$n,j,C,void 0));return}}for(Me in C)C.hasOwnProperty(Me)&&(j=C[Me],j!=null&&eu(w,_,Me,j,C,null))}function ez(w,_,C,j){switch(_){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var q=null,te=null,Se=null,Me=null,pt=null,Xt=null,$n=null;for(Zt in C){var Vn=C[Zt];if(C.hasOwnProperty(Zt)&&Vn!=null)switch(Zt){case"checked":break;case"value":break;case"defaultValue":pt=Vn;default:j.hasOwnProperty(Zt)||eu(w,_,Zt,null,j,Vn)}}for(var ln in j){var Zt=j[ln];if(Vn=C[ln],j.hasOwnProperty(ln)&&(Zt!=null||Vn!=null))switch(ln){case"type":te=Zt;break;case"name":q=Zt;break;case"checked":Xt=Zt;break;case"defaultChecked":$n=Zt;break;case"value":Se=Zt;break;case"defaultValue":Me=Zt;break;case"children":case"dangerouslySetInnerHTML":if(Zt!=null)throw Error(s(137,_));break;default:Zt!==Vn&&eu(w,_,ln,Zt,j,Vn)}}tO(w,Se,Me,pt,Xt,$n,te,q);return;case"select":Zt=Se=Me=ln=null;for(te in C)if(pt=C[te],C.hasOwnProperty(te)&&pt!=null)switch(te){case"value":break;case"multiple":Zt=pt;default:j.hasOwnProperty(te)||eu(w,_,te,null,j,pt)}for(q in j)if(te=j[q],pt=C[q],j.hasOwnProperty(q)&&(te!=null||pt!=null))switch(q){case"value":ln=te;break;case"defaultValue":Me=te;break;case"multiple":Se=te;default:te!==pt&&eu(w,_,q,te,j,pt)}_=Me,C=Se,j=Zt,ln!=null?X_(w,!!C,ln,!1):!!j!=!!C&&(_!=null?X_(w,!!C,_,!0):X_(w,!!C,C?[]:"",!1));return;case"textarea":Zt=ln=null;for(Me in C)if(q=C[Me],C.hasOwnProperty(Me)&&q!=null&&!j.hasOwnProperty(Me))switch(Me){case"value":break;case"children":break;default:eu(w,_,Me,null,j,q)}for(Se in j)if(q=j[Se],te=C[Se],j.hasOwnProperty(Se)&&(q!=null||te!=null))switch(Se){case"value":ln=q;break;case"defaultValue":Zt=q;break;case"children":break;case"dangerouslySetInnerHTML":if(q!=null)throw Error(s(91));break;default:q!==te&&eu(w,_,Se,q,j,te)}d6(w,ln,Zt);return;case"option":for(var Do in C)if(ln=C[Do],C.hasOwnProperty(Do)&&ln!=null&&!j.hasOwnProperty(Do))switch(Do){case"selected":w.selected=!1;break;default:eu(w,_,Do,null,j,ln)}for(pt in j)if(ln=j[pt],Zt=C[pt],j.hasOwnProperty(pt)&&ln!==Zt&&(ln!=null||Zt!=null))switch(pt){case"selected":w.selected=ln&&typeof ln!="function"&&typeof ln!="symbol";break;default:eu(w,_,pt,ln,j,Zt)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var Oo in C)ln=C[Oo],C.hasOwnProperty(Oo)&&ln!=null&&!j.hasOwnProperty(Oo)&&eu(w,_,Oo,null,j,ln);for(Xt in j)if(ln=j[Xt],Zt=C[Xt],j.hasOwnProperty(Xt)&&ln!==Zt&&(ln!=null||Zt!=null))switch(Xt){case"children":case"dangerouslySetInnerHTML":if(ln!=null)throw Error(s(137,_));break;default:eu(w,_,Xt,ln,j,Zt)}return;default:if(p6(_)){for(var mu in C)ln=C[mu],C.hasOwnProperty(mu)&&ln!==void 0&&!j.hasOwnProperty(mu)&&L5(w,_,mu,void 0,j,ln);for($n in j)ln=j[$n],Zt=C[$n],!j.hasOwnProperty($n)||ln===Zt||ln===void 0&&Zt===void 0||L5(w,_,$n,ln,j,Zt);return}}for(var zt in C)ln=C[zt],C.hasOwnProperty(zt)&&ln!=null&&!j.hasOwnProperty(zt)&&eu(w,_,zt,null,j,ln);for(Vn in j)ln=j[Vn],Zt=C[Vn],!j.hasOwnProperty(Vn)||ln===Zt||ln==null&&Zt==null||eu(w,_,Vn,ln,j,Zt)}var i1=null,lN=null;function tS(w){return w.nodeType===9?w:w.ownerDocument}function ul(w){switch(w){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function kl(w,_){if(w===0)switch(_){case"svg":return 1;case"math":return 2;default:return 0}return w===1&&_==="foreignObject"?0:w}function dN(w,_){return w==="textarea"||w==="noscript"||typeof _.children=="string"||typeof _.children=="number"||typeof _.children=="bigint"||typeof _.dangerouslySetInnerHTML=="object"&&_.dangerouslySetInnerHTML!==null&&_.dangerouslySetInnerHTML.__html!=null}var D5=null;function tz(){var w=window.event;return w&&w.type==="popstate"?w===D5?!1:(D5=w,!0):(D5=null,!1)}var j9=typeof setTimeout=="function"?setTimeout:void 0,nz=typeof clearTimeout=="function"?clearTimeout:void 0,$9=typeof Promise=="function"?Promise:void 0,rz=typeof queueMicrotask=="function"?queueMicrotask:typeof $9<"u"?function(w){return $9.resolve(null).then(w).catch(F9)}:j9;function F9(w){setTimeout(function(){throw w})}function cb(w){return w==="head"}function fN(w,_){var C=_,j=0,q=0;do{var te=C.nextSibling;if(w.removeChild(C),te&&te.nodeType===8)if(C=te.data,C==="/$"){if(0<j&&8>j){C=j;var Se=w.ownerDocument;if(C&1&&c3(Se.documentElement),C&2&&c3(Se.body),C&4)for(C=Se.head,c3(C),Se=C.firstChild;Se;){var Me=Se.nextSibling,pt=Se.nodeName;Se[sy]||pt==="SCRIPT"||pt==="STYLE"||pt==="LINK"&&Se.rel.toLowerCase()==="stylesheet"||C.removeChild(Se),Se=Me}}if(q===0){w.removeChild(te),EN(_);return}q--}else C==="$"||C==="$?"||C==="$!"?q++:j=C.charCodeAt(0)-48;else j=0;C=te}while(C);EN(_)}function M5(w){var _=w.firstChild;for(_&&_.nodeType===10&&(_=_.nextSibling);_;){var C=_;switch(_=_.nextSibling,C.nodeName){case"HTML":case"HEAD":case"BODY":M5(C),rC(C);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(C.rel.toLowerCase()==="stylesheet")continue}w.removeChild(C)}}function B9(w,_,C,j){for(;w.nodeType===1;){var q=C;if(w.nodeName.toLowerCase()!==_.toLowerCase()){if(!j&&(w.nodeName!=="INPUT"||w.type!=="hidden"))break}else if(j){if(!w[sy])switch(_){case"meta":if(!w.hasAttribute("itemprop"))break;return w;case"link":if(te=w.getAttribute("rel"),te==="stylesheet"&&w.hasAttribute("data-precedence"))break;if(te!==q.rel||w.getAttribute("href")!==(q.href==null||q.href===""?null:q.href)||w.getAttribute("crossorigin")!==(q.crossOrigin==null?null:q.crossOrigin)||w.getAttribute("title")!==(q.title==null?null:q.title))break;return w;case"style":if(w.hasAttribute("data-precedence"))break;return w;case"script":if(te=w.getAttribute("src"),(te!==(q.src==null?null:q.src)||w.getAttribute("type")!==(q.type==null?null:q.type)||w.getAttribute("crossorigin")!==(q.crossOrigin==null?null:q.crossOrigin))&&te&&w.hasAttribute("async")&&!w.hasAttribute("itemprop"))break;return w;default:return w}}else if(_==="input"&&w.type==="hidden"){var te=q.name==null?null:""+q.name;if(q.type==="hidden"&&w.getAttribute("name")===te)return w}else return w;if(w=o1(w.nextSibling),w===null)break}return null}function hN(w,_,C){if(_==="")return null;for(;w.nodeType!==3;)if((w.nodeType!==1||w.nodeName!=="INPUT"||w.type!=="hidden")&&!C||(w=o1(w.nextSibling),w===null))return null;return w}function _p(w){return w.data==="$!"||w.data==="$?"&&w.ownerDocument.readyState==="complete"}function Pk(w,_){var C=w.ownerDocument;if(w.data!=="$?"||C.readyState==="complete")_();else{var j=function(){_(),C.removeEventListener("DOMContentLoaded",j)};C.addEventListener("DOMContentLoaded",j),w._reactRetry=j}}function o1(w){for(;w!=null;w=w.nextSibling){var _=w.nodeType;if(_===1||_===3)break;if(_===8){if(_=w.data,_==="$"||_==="$!"||_==="$?"||_==="F!"||_==="F")break;if(_==="/$")return null}}return w}var j5=null;function U9(w){w=w.previousSibling;for(var _=0;w;){if(w.nodeType===8){var C=w.data;if(C==="$"||C==="$!"||C==="$?"){if(_===0)return w;_--}else C==="/$"&&_++}w=w.previousSibling}return null}function z9(w,_,C){switch(_=tS(C),w){case"html":if(w=_.documentElement,!w)throw Error(s(452));return w;case"head":if(w=_.head,!w)throw Error(s(453));return w;case"body":if(w=_.body,!w)throw Error(s(454));return w;default:throw Error(s(451))}}function c3(w){for(var _=w.attributes;_.length;)w.removeAttributeNode(_[0]);rC(w)}var im=new Map,a1=new Set;function fa(w){return typeof w.getRootNode=="function"?w.getRootNode():w.nodeType===9?w:w.ownerDocument}var d0=qe.d;qe.d={f:om,r:$5,D:H9,C:G9,L:nS,m:iz,X:az,S:oz,M:q9};function om(){var w=d0.f(),_=eN();return w||_}function $5(w){var _=cy(w);_!==null&&_.tag===5&&_.type==="form"?BO(_):d0.r(w)}var ll=typeof document>"u"?null:document;function vc(w,_,C){var j=ll;if(j&&typeof _=="string"&&_){var q=Xm(_);q='link[rel="'+w+'"][href="'+q+'"]',typeof C=="string"&&(q+='[crossorigin="'+C+'"]'),a1.has(q)||(a1.add(q),w={rel:w,crossOrigin:C,href:_},j.querySelector(q)===null&&(_=j.createElement("link"),Jd(_,"link",w),Od(_),j.head.appendChild(_)))}}function H9(w){d0.D(w),vc("dns-prefetch",w,null)}function G9(w,_){d0.C(w,_),vc("preconnect",w,_)}function nS(w,_,C){d0.L(w,_,C);var j=ll;if(j&&w&&_){var q='link[rel="preload"][as="'+Xm(_)+'"]';_==="image"&&C&&C.imageSrcSet?(q+='[imagesrcset="'+Xm(C.imageSrcSet)+'"]',typeof C.imageSizes=="string"&&(q+='[imagesizes="'+Xm(C.imageSizes)+'"]')):q+='[href="'+Xm(w)+'"]';var te=q;switch(_){case"style":te=Lk(w);break;case"script":te=Dk(w)}im.has(te)||(w=k({rel:"preload",href:_==="image"&&C&&C.imageSrcSet?void 0:w,as:_},C),im.set(te,w),j.querySelector(q)!==null||_==="style"&&j.querySelector(rS(te))||_==="script"&&j.querySelector(u3(te))||(_=j.createElement("link"),Jd(_,"link",w),Od(_),j.head.appendChild(_)))}}function iz(w,_){d0.m(w,_);var C=ll;if(C&&w){var j=_&&typeof _.as=="string"?_.as:"script",q='link[rel="modulepreload"][as="'+Xm(j)+'"][href="'+Xm(w)+'"]',te=q;switch(j){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":te=Dk(w)}if(!im.has(te)&&(w=k({rel:"modulepreload",href:w},_),im.set(te,w),C.querySelector(q)===null)){switch(j){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(C.querySelector(u3(te)))return}j=C.createElement("link"),Jd(j,"link",w),Od(j),C.head.appendChild(j)}}}function oz(w,_,C){d0.S(w,_,C);var j=ll;if(j&&w){var q=ly(j).hoistableStyles,te=Lk(w);_=_||"default";var Se=q.get(te);if(!Se){var Me={loading:0,preload:null};if(Se=j.querySelector(rS(te)))Me.loading=5;else{w=k({rel:"stylesheet",href:w,"data-precedence":_},C),(C=im.get(te))&&iS(w,C);var pt=Se=j.createElement("link");Od(pt),Jd(pt,"link",w),pt._p=new Promise(function(Xt,$n){pt.onload=Xt,pt.onerror=$n}),pt.addEventListener("load",function(){Me.loading|=1}),pt.addEventListener("error",function(){Me.loading|=2}),Me.loading|=4,pN(Se,_,j)}Se={type:"stylesheet",instance:Se,count:1,state:Me},q.set(te,Se)}}}function az(w,_){d0.X(w,_);var C=ll;if(C&&w){var j=ly(C).hoistableScripts,q=Dk(w),te=j.get(q);te||(te=C.querySelector(u3(q)),te||(w=k({src:w,async:!0},_),(_=im.get(q))&&B5(w,_),te=C.createElement("script"),Od(te),Jd(te,"link",w),C.head.appendChild(te)),te={type:"script",instance:te,count:1,state:null},j.set(q,te))}}function q9(w,_){d0.M(w,_);var C=ll;if(C&&w){var j=ly(C).hoistableScripts,q=Dk(w),te=j.get(q);te||(te=C.querySelector(u3(q)),te||(w=k({src:w,async:!0,type:"module"},_),(_=im.get(q))&&B5(w,_),te=C.createElement("script"),Od(te),Jd(te,"link",w),C.head.appendChild(te)),te={type:"script",instance:te,count:1,state:null},j.set(q,te))}}function F5(w,_,C,j){var q=(q=Yt.current)?fa(q):null;if(!q)throw Error(s(446));switch(w){case"meta":case"title":return null;case"style":return typeof C.precedence=="string"&&typeof C.href=="string"?(_=Lk(C.href),C=ly(q).hoistableStyles,j=C.get(_),j||(j={type:"style",instance:null,count:0,state:null},C.set(_,j)),j):{type:"void",instance:null,count:0,state:null};case"link":if(C.rel==="stylesheet"&&typeof C.href=="string"&&typeof C.precedence=="string"){w=Lk(C.href);var te=ly(q).hoistableStyles,Se=te.get(w);if(Se||(q=q.ownerDocument||q,Se={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},te.set(w,Se),(te=q.querySelector(rS(w)))&&!te._p&&(Se.instance=te,Se.state.loading=5),im.has(w)||(C={rel:"preload",as:"style",href:C.href,crossOrigin:C.crossOrigin,integrity:C.integrity,media:C.media,hrefLang:C.hrefLang,referrerPolicy:C.referrerPolicy},im.set(w,C),te||sz(q,w,C,Se.state))),_&&j===null)throw Error(s(528,""));return Se}if(_&&j!==null)throw Error(s(529,""));return null;case"script":return _=C.async,C=C.src,typeof C=="string"&&_&&typeof _!="function"&&typeof _!="symbol"?(_=Dk(C),C=ly(q).hoistableScripts,j=C.get(_),j||(j={type:"script",instance:null,count:0,state:null},C.set(_,j)),j):{type:"void",instance:null,count:0,state:null};default:throw Error(s(444,w))}}function Lk(w){return'href="'+Xm(w)+'"'}function rS(w){return'link[rel="stylesheet"]['+w+"]"}function V9(w){return k({},w,{"data-precedence":w.precedence,precedence:null})}function sz(w,_,C,j){w.querySelector('link[rel="preload"][as="style"]['+_+"]")?j.loading=1:(_=w.createElement("link"),j.preload=_,_.addEventListener("load",function(){return j.loading|=1}),_.addEventListener("error",function(){return j.loading|=2}),Jd(_,"link",C),Od(_),w.head.appendChild(_))}function Dk(w){return'[src="'+Xm(w)+'"]'}function u3(w){return"script[async]"+w}function W9(w,_,C){if(_.count++,_.instance===null)switch(_.type){case"style":var j=w.querySelector('style[data-href~="'+Xm(C.href)+'"]');if(j)return _.instance=j,Od(j),j;var q=k({},C,{"data-href":C.href,"data-precedence":C.precedence,href:null,precedence:null});return j=(w.ownerDocument||w).createElement("style"),Od(j),Jd(j,"style",q),pN(j,C.precedence,w),_.instance=j;case"stylesheet":q=Lk(C.href);var te=w.querySelector(rS(q));if(te)return _.state.loading|=4,_.instance=te,Od(te),te;j=V9(C),(q=im.get(q))&&iS(j,q),te=(w.ownerDocument||w).createElement("link"),Od(te);var Se=te;return Se._p=new Promise(function(Me,pt){Se.onload=Me,Se.onerror=pt}),Jd(te,"link",j),_.state.loading|=4,pN(te,C.precedence,w),_.instance=te;case"script":return te=Dk(C.src),(q=w.querySelector(u3(te)))?(_.instance=q,Od(q),q):(j=C,(q=im.get(te))&&(j=k({},C),B5(j,q)),w=w.ownerDocument||w,q=w.createElement("script"),Od(q),Jd(q,"link",j),w.head.appendChild(q),_.instance=q);case"void":return null;default:throw Error(s(443,_.type))}else _.type==="stylesheet"&&(_.state.loading&4)===0&&(j=_.instance,_.state.loading|=4,pN(j,C.precedence,w));return _.instance}function pN(w,_,C){for(var j=C.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),q=j.length?j[j.length-1]:null,te=q,Se=0;Se<j.length;Se++){var Me=j[Se];if(Me.dataset.precedence===_)te=Me;else if(te!==q)break}te?te.parentNode.insertBefore(w,te.nextSibling):(_=C.nodeType===9?C.head:C,_.insertBefore(w,_.firstChild))}function iS(w,_){w.crossOrigin==null&&(w.crossOrigin=_.crossOrigin),w.referrerPolicy==null&&(w.referrerPolicy=_.referrerPolicy),w.title==null&&(w.title=_.title)}function B5(w,_){w.crossOrigin==null&&(w.crossOrigin=_.crossOrigin),w.referrerPolicy==null&&(w.referrerPolicy=_.referrerPolicy),w.integrity==null&&(w.integrity=_.integrity)}var By=null;function K9(w,_,C){if(By===null){var j=new Map,q=By=new Map;q.set(C,j)}else q=By,j=q.get(C),j||(j=new Map,q.set(C,j));if(j.has(w))return j;for(j.set(w,null),C=C.getElementsByTagName(w),q=0;q<C.length;q++){var te=C[q];if(!(te[sy]||te[_l]||w==="link"&&te.getAttribute("rel")==="stylesheet")&&te.namespaceURI!=="http://www.w3.org/2000/svg"){var Se=te.getAttribute(_)||"";Se=w+Se;var Me=j.get(Se);Me?Me.push(te):j.set(Se,[te])}}return j}function gN(w,_,C){w=w.ownerDocument||w,w.head.insertBefore(C,_==="title"?w.querySelector("head > title"):null)}function cz(w,_,C){if(C===1||_.itemProp!=null)return!1;switch(w){case"meta":case"title":return!0;case"style":if(typeof _.precedence!="string"||typeof _.href!="string"||_.href==="")break;return!0;case"link":if(typeof _.rel!="string"||typeof _.href!="string"||_.href===""||_.onLoad||_.onError)break;switch(_.rel){case"stylesheet":return w=_.disabled,typeof _.precedence=="string"&&w==null;default:return!0}case"script":if(_.async&&typeof _.async!="function"&&typeof _.async!="symbol"&&!_.onLoad&&!_.onError&&_.src&&typeof _.src=="string")return!0}return!1}function Zd(w){return!(w.type==="stylesheet"&&(w.state.loading&3)===0)}var Af=null;function Y9(){}function uz(w,_,C){if(Af===null)throw Error(s(475));var j=Af;if(_.type==="stylesheet"&&(typeof C.media!="string"||matchMedia(C.media).matches!==!1)&&(_.state.loading&4)===0){if(_.instance===null){var q=Lk(C.href),te=w.querySelector(rS(q));if(te){w=te._p,w!==null&&typeof w=="object"&&typeof w.then=="function"&&(j.count++,j=l3.bind(j),w.then(j,j)),_.state.loading|=4,_.instance=te,Od(te);return}te=w.ownerDocument||w,C=V9(C),(q=im.get(q))&&iS(C,q),te=te.createElement("link"),Od(te);var Se=te;Se._p=new Promise(function(Me,pt){Se.onload=Me,Se.onerror=pt}),Jd(te,"link",C),_.instance=te}j.stylesheets===null&&(j.stylesheets=new Map),j.stylesheets.set(_,w),(w=_.state.preload)&&(_.state.loading&3)===0&&(j.count++,_=l3.bind(j),w.addEventListener("load",_),w.addEventListener("error",_))}}function lz(){if(Af===null)throw Error(s(475));var w=Af;return w.stylesheets&&w.count===0&&U5(w,w.stylesheets),0<w.count?function(_){var C=setTimeout(function(){if(w.stylesheets&&U5(w,w.stylesheets),w.unsuspend){var j=w.unsuspend;w.unsuspend=null,j()}},6e4);return w.unsuspend=_,function(){w.unsuspend=null,clearTimeout(C)}}:null}function l3(){if(this.count--,this.count===0){if(this.stylesheets)U5(this,this.stylesheets);else if(this.unsuspend){var w=this.unsuspend;this.unsuspend=null,w()}}}var mN=null;function U5(w,_){w.stylesheets=null,w.unsuspend!==null&&(w.count++,mN=new Map,_.forEach(dz,w),mN=null,l3.call(w))}function dz(w,_){if(!(_.state.loading&4)){var C=mN.get(w);if(C)var j=C.get(null);else{C=new Map,mN.set(w,C);for(var q=w.querySelectorAll("link[data-precedence],style[data-precedence]"),te=0;te<q.length;te++){var Se=q[te];(Se.nodeName==="LINK"||Se.getAttribute("media")!=="not all")&&(C.set(Se.dataset.precedence,Se),j=Se)}j&&C.set(null,j)}q=_.instance,Se=q.getAttribute("data-precedence"),te=C.get(Se)||j,te===j&&C.set(null,q),C.set(Se,q),this.count++,j=l3.bind(this),q.addEventListener("load",j),q.addEventListener("error",j),te?te.parentNode.insertBefore(q,te.nextSibling):(w=w.nodeType===9?w.head:w,w.insertBefore(q,w.firstChild)),_.state.loading|=4}}var Mk={$$typeof:K,Provider:null,Consumer:null,_currentValue:it,_currentValue2:it,_threadCount:0};function fz(w,_,C,j,q,te,Se,Me){this.tag=1,this.containerInfo=w,this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Qp(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Qp(0),this.hiddenUpdates=Qp(null),this.identifierPrefix=j,this.onUncaughtError=q,this.onCaughtError=te,this.onRecoverableError=Se,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=Me,this.incompleteTransitions=new Map}function X9(w,_,C,j,q,te,Se,Me,pt,Xt,$n,Vn){return w=new fz(w,_,C,Se,Me,pt,Xt,Vn),_=1,te===!0&&(_|=24),te=ng(3,null,null,_),w.current=te,te.stateNode=w,_=kO(),_.refCount++,w.pooledCache=_,_.refCount++,te.memoizedState={element:j,isDehydrated:C,cache:_},F2(te),w}function J9(w){return w?(w=R2,w):R2}function z5(w,_,C,j,q,te){q=J9(q),j.context===null?j.context=q:j.pendingContext=q,j=ky(_),j.payload={element:C},te=te===void 0?null:te,te!==null&&(j.callback=te),C=yv(w,j,_),C!==null&&(rm(C,w,_),kC(C,w,_))}function H5(w,_){if(w=w.memoizedState,w!==null&&w.dehydrated!==null){var C=w.retryLane;w.retryLane=C!==0&&C<_?C:_}}function bN(w,_){H5(w,_),(w=w.alternate)&&H5(w,_)}function G5(w){if(w.tag===13){var _=N2(w,67108864);_!==null&&rm(_,w,67108864),bN(w,67108864)}}var oS=!0;function d3(w,_,C,j){var q=_e.T;_e.T=null;var te=qe.p;try{qe.p=2,q5(w,_,C,j)}finally{qe.p=te,_e.T=q}}function oh(w,_,C,j){var q=_e.T;_e.T=null;var te=qe.p;try{qe.p=8,q5(w,_,C,j)}finally{qe.p=te,_e.T=q}}function q5(w,_,C,j){if(oS){var q=Z9(j);if(q===null)Pv(w,_,j,V5,C),pz(w,j);else if(qK(q,w,_,C,j))j.stopPropagation();else if(pz(w,j),_&4&&-1<GK.indexOf(w)){for(;q!==null;){var te=cy(q);if(te!==null)switch(te.tag){case 3:if(te=te.stateNode,te.current.memoizedState.isDehydrated){var Se=rc(te.pendingLanes);if(Se!==0){var Me=te;for(Me.pendingLanes|=2,Me.entangledLanes|=2;Se;){var pt=1<<31-Ei(Se);Me.entanglements[1]|=pt,Se&=~pt}l0(te),(zc&6)===0&&(Ck=qi()+500,i3(0))}}break;case 13:Me=N2(te,2),Me!==null&&rm(Me,te,2),eN(),bN(te,2)}if(te=Z9(j),te===null&&Pv(w,_,j,V5,C),te===q)break;q=te}q!==null&&j.stopPropagation()}else Pv(w,_,j,null,C)}}function Z9(w){return w=g6(w),Q9(w)}var V5=null;function Q9(w){if(V5=null,w=fv(w),w!==null){var _=d(w);if(_===null)w=null;else{var C=_.tag;if(C===13){if(w=h(_),w!==null)return w;w=null}else if(C===3){if(_.stateNode.current.memoizedState.isDehydrated)return _.tag===3?_.stateNode.containerInfo:null;w=null}else _!==w&&(w=null)}}return V5=w,null}function hz(w){switch(w){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(yi()){case lo:return 2;case Eo:return 8;case Fc:case Yo:return 32;case zl:return 268435456;default:return 32}default:return 32}}var W5=!1,Uy=null,Lv=null,Wl=null,vN=new Map,jk=new Map,Dv=[],GK="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function pz(w,_){switch(w){case"focusin":case"focusout":Uy=null;break;case"dragenter":case"dragleave":Lv=null;break;case"mouseover":case"mouseout":Wl=null;break;case"pointerover":case"pointerout":vN.delete(_.pointerId);break;case"gotpointercapture":case"lostpointercapture":jk.delete(_.pointerId)}}function wN(w,_,C,j,q,te){return w===null||w.nativeEvent!==te?(w={blockedOn:_,domEventName:C,eventSystemFlags:j,nativeEvent:te,targetContainers:[q]},_!==null&&(_=cy(_),_!==null&&G5(_)),w):(w.eventSystemFlags|=j,_=w.targetContainers,q!==null&&_.indexOf(q)===-1&&_.push(q),w)}function qK(w,_,C,j,q){switch(_){case"focusin":return Uy=wN(Uy,w,_,C,j,q),!0;case"dragenter":return Lv=wN(Lv,w,_,C,j,q),!0;case"mouseover":return Wl=wN(Wl,w,_,C,j,q),!0;case"pointerover":var te=q.pointerId;return vN.set(te,wN(vN.get(te)||null,w,_,C,j,q)),!0;case"gotpointercapture":return te=q.pointerId,jk.set(te,wN(jk.get(te)||null,w,_,C,j,q)),!0}return!1}function yN(w){var _=fv(w.target);if(_!==null){var C=d(_);if(C!==null){if(_=C.tag,_===13){if(_=h(C),_!==null){w.blockedOn=_,n6(w.priority,function(){if(C.tag===13){var j=$h();j=dv(j);var q=N2(C,j);q!==null&&rm(q,C,j),bN(C,j)}});return}}else if(_===3&&C.stateNode.current.memoizedState.isDehydrated){w.blockedOn=C.tag===3?C.stateNode.containerInfo:null;return}}}w.blockedOn=null}function aS(w){if(w.blockedOn!==null)return!1;for(var _=w.targetContainers;0<_.length;){var C=Z9(w.nativeEvent);if(C===null){C=w.nativeEvent;var j=new C.constructor(C.type,C);rO=j,C.target.dispatchEvent(j),rO=null}else return _=cy(C),_!==null&&G5(_),w.blockedOn=C,!1;_.shift()}return!0}function K5(w,_,C){aS(w)&&C.delete(_)}function VK(){W5=!1,Uy!==null&&aS(Uy)&&(Uy=null),Lv!==null&&aS(Lv)&&(Lv=null),Wl!==null&&aS(Wl)&&(Wl=null),vN.forEach(K5),jk.forEach(K5)}function Y5(w,_){w.blockedOn===_&&(w.blockedOn=null,W5||(W5=!0,n.unstable_scheduleCallback(n.unstable_NormalPriority,VK)))}var X5=null;function e7(w){X5!==w&&(X5=w,n.unstable_scheduleCallback(n.unstable_NormalPriority,function(){X5===w&&(X5=null);for(var _=0;_<w.length;_+=3){var C=w[_],j=w[_+1],q=w[_+2];if(typeof j!="function"){if(Q9(j||C)===null)continue;break}var te=cy(C);te!==null&&(w.splice(_,3),_-=3,DC(te,{pending:!0,data:q,method:C.method,action:j},j,q))}}))}function EN(w){function _(pt){return Y5(pt,w)}Uy!==null&&Y5(Uy,w),Lv!==null&&Y5(Lv,w),Wl!==null&&Y5(Wl,w),vN.forEach(_),jk.forEach(_);for(var C=0;C<Dv.length;C++){var j=Dv[C];j.blockedOn===w&&(j.blockedOn=null)}for(;0<Dv.length&&(C=Dv[0],C.blockedOn===null);)yN(C),C.blockedOn===null&&Dv.shift();if(C=(w.ownerDocument||w).$$reactFormReplay,C!=null)for(j=0;j<C.length;j+=3){var q=C[j],te=C[j+1],Se=q[rh]||null;if(typeof te=="function")Se||e7(C);else if(Se){var Me=null;if(te&&te.hasAttribute("formAction")){if(q=te,Se=te[rh]||null)Me=Se.formAction;else if(Q9(q)!==null)continue}else Me=Se.action;typeof Me=="function"?C[j+1]=Me:(C.splice(j,3),j-=3),e7(C)}}}function SN(w){this._internalRoot=w}J5.prototype.render=SN.prototype.render=function(w){var _=this._internalRoot;if(_===null)throw Error(s(409));var C=_.current,j=$h();z5(C,j,w,_,null,null)},J5.prototype.unmount=SN.prototype.unmount=function(){var w=this._internalRoot;if(w!==null){this._internalRoot=null;var _=w.containerInfo;z5(w.current,2,null,w,null,null),eN(),_[t0]=null}};function J5(w){this._internalRoot=w}J5.prototype.unstable_scheduleHydration=function(w){if(w){var _=YI();w={blockedOn:null,target:w,priority:_};for(var C=0;C<Dv.length&&_!==0&&_<Dv[C].priority;C++);Dv.splice(C,0,w),C===0&&yN(w)}};var gz=r.version;if(gz!=="19.1.1")throw Error(s(527,gz,"19.1.1"));qe.findDOMNode=function(w){var _=w._reactInternals;if(_===void 0)throw typeof w.render=="function"?Error(s(188)):(w=Object.keys(w).join(","),Error(s(268,w)));return w=y(_),w=w!==null?E(w):null,w=w===null?null:w.stateNode,w};var WK={bundleType:0,version:"19.1.1",rendererPackageName:"react-dom",currentDispatcherRef:_e,reconcilerVersion:"19.1.1"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var $k=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!$k.isDisabled&&$k.supportsFiber)try{Sn=$k.inject(WK),Pn=$k}catch{}}return Yne.createRoot=function(w,_){if(!c(w))throw Error(s(299));var C=!1,j="",q=o5,te=zO,Se=K8,Me=null;return _!=null&&(_.unstable_strictMode===!0&&(C=!0),_.identifierPrefix!==void 0&&(j=_.identifierPrefix),_.onUncaughtError!==void 0&&(q=_.onUncaughtError),_.onCaughtError!==void 0&&(te=_.onCaughtError),_.onRecoverableError!==void 0&&(Se=_.onRecoverableError),_.unstable_transitionCallbacks!==void 0&&(Me=_.unstable_transitionCallbacks)),_=X9(w,1,!1,null,null,C,j,q,te,Se,Me,null),w[t0]=_.current,P5(w),new SN(_)},Yne.hydrateRoot=function(w,_,C){if(!c(w))throw Error(s(299));var j=!1,q="",te=o5,Se=zO,Me=K8,pt=null,Xt=null;return C!=null&&(C.unstable_strictMode===!0&&(j=!0),C.identifierPrefix!==void 0&&(q=C.identifierPrefix),C.onUncaughtError!==void 0&&(te=C.onUncaughtError),C.onCaughtError!==void 0&&(Se=C.onCaughtError),C.onRecoverableError!==void 0&&(Me=C.onRecoverableError),C.unstable_transitionCallbacks!==void 0&&(pt=C.unstable_transitionCallbacks),C.formState!==void 0&&(Xt=C.formState)),_=X9(w,1,!0,_,C??null,j,q,te,Se,Me,pt,Xt),_.context=J9(null),C=_.current,j=$h(),j=dv(j),q=ky(j),q.callback=null,yv(C,q,j),C=j,_.current.lanes=C,Kb(_,C),l0(_),w[t0]=_.current,P5(w),new J5(_)},Yne.version="19.1.1",Yne}var c8t;function Ijn(){if(c8t)return vje.exports;c8t=1;function n(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(n)}catch(r){console.error(r)}}return n(),vje.exports=Ajn(),vje.exports}var Ojn=Ijn(),Aqe=function(n,r){return Aqe=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(o,s){o.__proto__=s}||function(o,s){for(var c in s)Object.prototype.hasOwnProperty.call(s,c)&&(o[c]=s[c])},Aqe(n,r)};function Njn(n,r){if(typeof r!="function"&&r!==null)throw new TypeError("Class extends value "+String(r)+" is not a constructor or null");Aqe(n,r);function o(){this.constructor=n}n.prototype=r===null?Object.create(r):(o.prototype=r.prototype,new o)}var Bm=function(){return Bm=Object.assign||function(r){for(var o,s=1,c=arguments.length;s<c;s++){o=arguments[s];for(var d in o)Object.prototype.hasOwnProperty.call(o,d)&&(r[d]=o[d])}return r},Bm.apply(this,arguments)};function Rjn(n,r){var o={};for(var s in n)Object.prototype.hasOwnProperty.call(n,s)&&r.indexOf(s)<0&&(o[s]=n[s]);if(n!=null&&typeof Object.getOwnPropertySymbols=="function")for(var c=0,s=Object.getOwnPropertySymbols(n);c<s.length;c++)r.indexOf(s[c])<0&&Object.prototype.propertyIsEnumerable.call(n,s[c])&&(o[s[c]]=n[s[c]]);return o}function OV(n,r,o,s){function c(d){return d instanceof o?d:new o(function(h){h(d)})}return new(o||(o=Promise))(function(d,h){function b(k){try{E(s.next(k))}catch(x){h(x)}}function y(k){try{E(s.throw(k))}catch(x){h(x)}}function E(k){k.done?d(k.value):c(k.value).then(b,y)}E((s=s.apply(n,r||[])).next())})}function NV(n,r){var o={label:0,sent:function(){if(d[0]&1)throw d[1];return d[1]},trys:[],ops:[]},s,c,d,h=Object.create((typeof Iterator=="function"?Iterator:Object).prototype);return h.next=b(0),h.throw=b(1),h.return=b(2),typeof Symbol=="function"&&(h[Symbol.iterator]=function(){return this}),h;function b(E){return function(k){return y([E,k])}}function y(E){if(s)throw new TypeError("Generator is already executing.");for(;h&&(h=0,E[0]&&(o=0)),o;)try{if(s=1,c&&(d=E[0]&2?c.return:E[0]?c.throw||((d=c.return)&&d.call(c),0):c.next)&&!(d=d.call(c,E[1])).done)return d;switch(c=0,d&&(E=[E[0]&2,d.value]),E[0]){case 0:case 1:d=E;break;case 4:return o.label++,{value:E[1],done:!1};case 5:o.label++,c=E[1],E=[0];continue;case 7:E=o.ops.pop(),o.trys.pop();continue;default:if(d=o.trys,!(d=d.length>0&&d[d.length-1])&&(E[0]===6||E[0]===2)){o=0;continue}if(E[0]===3&&(!d||E[1]>d[0]&&E[1]<d[3])){o.label=E[1];break}if(E[0]===6&&o.label<d[1]){o.label=d[1],d=E;break}if(d&&o.label<d[2]){o.label=d[2],o.ops.push(E);break}d[2]&&o.ops.pop(),o.trys.pop();continue}E=r.call(n,o)}catch(k){E=[6,k],c=0}finally{s=d=0}if(E[0]&5)throw E[1];return{value:E[0]?E[1]:void 0,done:!0}}}function Pjn(n,r,o){for(var s=0,c=r.length,d;s<c;s++)(d||!(s in r))&&(d||(d=Array.prototype.slice.call(r,0,s)),d[s]=r[s]);return n.concat(d||Array.prototype.slice.call(r))}function cI(n,r){var o={};for(var s in n)Object.prototype.hasOwnProperty.call(n,s)&&r.indexOf(s)<0&&(o[s]=n[s]);if(n!=null&&typeof Object.getOwnPropertySymbols=="function"){var c=0;for(s=Object.getOwnPropertySymbols(n);c<s.length;c++)r.indexOf(s[c])<0&&Object.prototype.propertyIsEnumerable.call(n,s[c])&&(o[s[c]]=n[s[c]])}return o}var TB=typeof globalThis<"u"?globalThis:typeof window<"u"?window:typeof global<"u"?global:typeof self<"u"?self:{};function TKe(n){return n&&n.__esModule&&Object.prototype.hasOwnProperty.call(n,"default")?n.default:n}function xKe(n,r){return n(r={exports:{}},r.exports),r.exports}var EB=xKe(function(n,r){Object.defineProperty(r,"__esModule",{value:!0});var o=function(){function s(){var c=this;this.locked=new Map,this.addToLocked=function(d,h){var b=c.locked.get(d);b===void 0?h===void 0?c.locked.set(d,[]):c.locked.set(d,[h]):h!==void 0&&(b.unshift(h),c.locked.set(d,b))},this.isLocked=function(d){return c.locked.has(d)},this.lock=function(d){return new Promise(function(h,b){c.isLocked(d)?c.addToLocked(d,h):(c.addToLocked(d),h())})},this.unlock=function(d){var h=c.locked.get(d);if(h!==void 0&&h.length!==0){var b=h.pop();c.locked.set(d,h),b!==void 0&&setTimeout(b,0)}else c.locked.delete(d)}}return s.getInstance=function(){return s.instance===void 0&&(s.instance=new s),s.instance},s}();r.default=function(){return o.getInstance()}});TKe(EB);var Ljn=TKe(xKe(function(n,r){var o=TB&&TB.__awaiter||function(k,x,O,P){return new(O||(O=Promise))(function(D,$){function U(K){try{G(P.next(K))}catch(X){$(X)}}function B(K){try{G(P.throw(K))}catch(X){$(X)}}function G(K){K.done?D(K.value):new O(function(X){X(K.value)}).then(U,B)}G((P=P.apply(k,x||[])).next())})},s=TB&&TB.__generator||function(k,x){var O,P,D,$,U={label:0,sent:function(){if(1&D[0])throw D[1];return D[1]},trys:[],ops:[]};return $={next:B(0),throw:B(1),return:B(2)},typeof Symbol=="function"&&($[Symbol.iterator]=function(){return this}),$;function B(G){return function(K){return function(X){if(O)throw new TypeError("Generator is already executing.");for(;U;)try{if(O=1,P&&(D=2&X[0]?P.return:X[0]?P.throw||((D=P.return)&&D.call(P),0):P.next)&&!(D=D.call(P,X[1])).done)return D;switch(P=0,D&&(X=[2&X[0],D.value]),X[0]){case 0:case 1:D=X;break;case 4:return U.label++,{value:X[1],done:!1};case 5:U.label++,P=X[1],X=[0];continue;case 7:X=U.ops.pop(),U.trys.pop();continue;default:if(D=U.trys,!((D=D.length>0&&D[D.length-1])||X[0]!==6&&X[0]!==2)){U=0;continue}if(X[0]===3&&(!D||X[1]>D[0]&&X[1]<D[3])){U.label=X[1];break}if(X[0]===6&&U.label<D[1]){U.label=D[1],D=X;break}if(D&&U.label<D[2]){U.label=D[2],U.ops.push(X);break}D[2]&&U.ops.pop(),U.trys.pop();continue}X=x.call(k,U)}catch(ae){X=[6,ae],P=0}finally{O=D=0}if(5&X[0])throw X[1];return{value:X[0]?X[1]:void 0,done:!0}}([G,K])}}},c=TB;Object.defineProperty(r,"__esModule",{value:!0});var d="browser-tabs-lock-key",h={key:function(k){return o(c,void 0,void 0,function(){return s(this,function(x){throw new Error("Unsupported")})})},getItem:function(k){return o(c,void 0,void 0,function(){return s(this,function(x){throw new Error("Unsupported")})})},clear:function(){return o(c,void 0,void 0,function(){return s(this,function(k){return[2,window.localStorage.clear()]})})},removeItem:function(k){return o(c,void 0,void 0,function(){return s(this,function(x){throw new Error("Unsupported")})})},setItem:function(k,x){return o(c,void 0,void 0,function(){return s(this,function(O){throw new Error("Unsupported")})})},keySync:function(k){return window.localStorage.key(k)},getItemSync:function(k){return window.localStorage.getItem(k)},clearSync:function(){return window.localStorage.clear()},removeItemSync:function(k){return window.localStorage.removeItem(k)},setItemSync:function(k,x){return window.localStorage.setItem(k,x)}};function b(k){return new Promise(function(x){return setTimeout(x,k)})}function y(k){for(var x="0123456789ABCDEFGHIJKLMNOPQRSTUVWXTZabcdefghiklmnopqrstuvwxyz",O="",P=0;P<k;P++)O+=x[Math.floor(Math.random()*x.length)];return O}var E=function(){function k(x){this.acquiredIatSet=new Set,this.storageHandler=void 0,this.id=Date.now().toString()+y(15),this.acquireLock=this.acquireLock.bind(this),this.releaseLock=this.releaseLock.bind(this),this.releaseLock__private__=this.releaseLock__private__.bind(this),this.waitForSomethingToChange=this.waitForSomethingToChange.bind(this),this.refreshLockWhileAcquired=this.refreshLockWhileAcquired.bind(this),this.storageHandler=x,k.waiters===void 0&&(k.waiters=[])}return k.prototype.acquireLock=function(x,O){return O===void 0&&(O=5e3),o(this,void 0,void 0,function(){var P,D,$,U,B,G,K;return s(this,function(X){switch(X.label){case 0:P=Date.now()+y(4),D=Date.now()+O,$=d+"-"+x,U=this.storageHandler===void 0?h:this.storageHandler,X.label=1;case 1:return Date.now()<D?[4,b(30)]:[3,8];case 2:return X.sent(),U.getItemSync($)!==null?[3,5]:(B=this.id+"-"+x+"-"+P,[4,b(Math.floor(25*Math.random()))]);case 3:return X.sent(),U.setItemSync($,JSON.stringify({id:this.id,iat:P,timeoutKey:B,timeAcquired:Date.now(),timeRefreshed:Date.now()})),[4,b(30)];case 4:return X.sent(),(G=U.getItemSync($))!==null&&(K=JSON.parse(G)).id===this.id&&K.iat===P?(this.acquiredIatSet.add(P),this.refreshLockWhileAcquired($,P),[2,!0]):[3,7];case 5:return k.lockCorrector(this.storageHandler===void 0?h:this.storageHandler),[4,this.waitForSomethingToChange(D)];case 6:X.sent(),X.label=7;case 7:return P=Date.now()+y(4),[3,1];case 8:return[2,!1]}})})},k.prototype.refreshLockWhileAcquired=function(x,O){return o(this,void 0,void 0,function(){var P=this;return s(this,function(D){return setTimeout(function(){return o(P,void 0,void 0,function(){var $,U,B;return s(this,function(G){switch(G.label){case 0:return[4,EB.default().lock(O)];case 1:return G.sent(),this.acquiredIatSet.has(O)?($=this.storageHandler===void 0?h:this.storageHandler,(U=$.getItemSync(x))===null?(EB.default().unlock(O),[2]):((B=JSON.parse(U)).timeRefreshed=Date.now(),$.setItemSync(x,JSON.stringify(B)),EB.default().unlock(O),this.refreshLockWhileAcquired(x,O),[2])):(EB.default().unlock(O),[2])}})})},1e3),[2]})})},k.prototype.waitForSomethingToChange=function(x){return o(this,void 0,void 0,function(){return s(this,function(O){switch(O.label){case 0:return[4,new Promise(function(P){var D=!1,$=Date.now(),U=!1;function B(){if(U||(window.removeEventListener("storage",B),k.removeFromWaiting(B),clearTimeout(G),U=!0),!D){D=!0;var K=50-(Date.now()-$);K>0?setTimeout(P,K):P(null)}}window.addEventListener("storage",B),k.addToWaiting(B);var G=setTimeout(B,Math.max(0,x-Date.now()))})];case 1:return O.sent(),[2]}})})},k.addToWaiting=function(x){this.removeFromWaiting(x),k.waiters!==void 0&&k.waiters.push(x)},k.removeFromWaiting=function(x){k.waiters!==void 0&&(k.waiters=k.waiters.filter(function(O){return O!==x}))},k.notifyWaiters=function(){k.waiters!==void 0&&k.waiters.slice().forEach(function(x){return x()})},k.prototype.releaseLock=function(x){return o(this,void 0,void 0,function(){return s(this,function(O){switch(O.label){case 0:return[4,this.releaseLock__private__(x)];case 1:return[2,O.sent()]}})})},k.prototype.releaseLock__private__=function(x){return o(this,void 0,void 0,function(){var O,P,D,$;return s(this,function(U){switch(U.label){case 0:return O=this.storageHandler===void 0?h:this.storageHandler,P=d+"-"+x,(D=O.getItemSync(P))===null?[2]:($=JSON.parse(D)).id!==this.id?[3,2]:[4,EB.default().lock($.iat)];case 1:U.sent(),this.acquiredIatSet.delete($.iat),O.removeItemSync(P),EB.default().unlock($.iat),k.notifyWaiters(),U.label=2;case 2:return[2]}})})},k.lockCorrector=function(x){for(var O=Date.now()-5e3,P=x,D=[],$=0;;){var U=P.keySync($);if(U===null)break;D.push(U),$++}for(var B=!1,G=0;G<D.length;G++){var K=D[G];if(K.includes(d)){var X=P.getItemSync(K);if(X!==null){var ae=JSON.parse(X);(ae.timeRefreshed===void 0&&ae.timeAcquired<O||ae.timeRefreshed!==void 0&&ae.timeRefreshed<O)&&(P.removeItemSync(K),B=!0)}}}B&&k.notifyWaiters()},k.waiters=void 0,k}();r.default=E}));const Djn={timeoutInSeconds:60},wKt={name:"auth0-spa-js",version:"2.3.0"},yKt=()=>Date.now();let e2=class Iqe extends Error{constructor(r,o){super(o),this.error=r,this.error_description=o,Object.setPrototypeOf(this,Iqe.prototype)}static fromPayload({error:r,error_description:o}){return new Iqe(r,o)}},Mjn=class EKt extends e2{constructor(r,o,s,c=null){super(r,o),this.state=s,this.appState=c,Object.setPrototypeOf(this,EKt.prototype)}},Oqe=class SKt extends e2{constructor(){super("timeout","Timeout"),Object.setPrototypeOf(this,SKt.prototype)}},jjn=class _Kt extends Oqe{constructor(r){super(),this.popup=r,Object.setPrototypeOf(this,_Kt.prototype)}},$jn=class kKt extends e2{constructor(r){super("cancelled","Popup closed"),this.popup=r,Object.setPrototypeOf(this,kKt.prototype)}},Fjn=class TKt extends e2{constructor(r,o,s){super(r,o),this.mfa_token=s,Object.setPrototypeOf(this,TKt.prototype)}},xKt=class CKt extends e2{constructor(r,o){super("missing_refresh_token",`Missing Refresh Token (audience: '${u8t(r,["default"])}', scope: '${u8t(o)}')`),this.audience=r,this.scope=o,Object.setPrototypeOf(this,CKt.prototype)}};function u8t(n,r=[]){return n&&!r.includes(n)?n:""}const X0e=()=>window.crypto,Sje=()=>{const n="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_~.";let r="";return Array.from(X0e().getRandomValues(new Uint8Array(43))).forEach(o=>r+=n[o%n.length]),r},l8t=n=>btoa(n),Nqe=n=>{var{clientId:r}=n,o=cI(n,["clientId"]);return new URLSearchParams((s=>Object.keys(s).filter(c=>s[c]!==void 0).reduce((c,d)=>Object.assign(Object.assign({},c),{[d]:s[d]}),{}))(Object.assign({client_id:r},o))).toString()},d8t=n=>(r=>decodeURIComponent(atob(r).split("").map(o=>"%"+("00"+o.charCodeAt(0).toString(16)).slice(-2)).join("")))(n.replace(/_/g,"/").replace(/-/g,"+")),Bjn=async(n,r)=>{const o=await fetch(n,r);return{ok:o.ok,json:await o.json()}},Ujn=async(n,r,o)=>{const s=new AbortController;let c;return r.signal=s.signal,Promise.race([Bjn(n,r),new Promise((d,h)=>{c=setTimeout(()=>{s.abort(),h(new Error("Timeout when executing 'fetch'"))},o)})]).finally(()=>{clearTimeout(c)})},zjn=async(n,r,o,s,c,d,h)=>{return b={auth:{audience:r,scope:o},timeout:c,fetchUrl:n,fetchOptions:s,useFormData:h},y=d,new Promise(function(E,k){const x=new MessageChannel;x.port1.onmessage=function(O){O.data.error?k(new Error(O.data.error)):E(O.data),x.port1.close()},y.postMessage(b,[x.port2])});var b,y},Hjn=async(n,r,o,s,c,d,h=1e4)=>c?zjn(n,r,o,s,h,c,d):Ujn(n,s,h);async function Gjn(n,r){var{baseUrl:o,timeout:s,audience:c,scope:d,auth0Client:h,useFormData:b}=n,y=cI(n,["baseUrl","timeout","audience","scope","auth0Client","useFormData"]);const E=y.grant_type==="urn:ietf:params:oauth:grant-type:token-exchange",k=Object.assign(Object.assign(Object.assign({},y),E&&c&&{audience:c}),E&&d&&{scope:d}),x=b?Nqe(k):JSON.stringify(k);return await async function(O,P,D,$,U,B,G){let K,X=null;for(let ge=0;ge<3;ge++)try{K=await Hjn(O,D,$,U,B,G,P),X=null;break}catch(Te){X=Te}if(X)throw X;const ae=K.json,{error:oe,error_description:he}=ae,ye=cI(ae,["error","error_description"]),{ok:we}=K;if(!we){const ge=he||`HTTP error. Unable to fetch ${O}`;throw oe==="mfa_required"?new Fjn(oe,ge,ye.mfa_token):oe==="missing_refresh_token"?new xKt(D,$):new e2(oe||"request_error",ge)}return ye}(`${o}/oauth/token`,s,c||"default",d,{method:"POST",body:x,headers:{"Content-Type":b?"application/x-www-form-urlencoded":"application/json","Auth0-Client":btoa(JSON.stringify(h||wKt))}},r,b)}const Xne=(...n)=>{return(r=n.filter(Boolean).join(" ").trim().split(/\s+/),Array.from(new Set(r))).join(" ");var r};let CB=class Rqe{constructor(r,o="@@auth0spajs@@",s){this.prefix=o,this.suffix=s,this.clientId=r.clientId,this.scope=r.scope,this.audience=r.audience}toKey(){return[this.prefix,this.clientId,this.audience,this.scope,this.suffix].filter(Boolean).join("::")}static fromKey(r){const[o,s,c,d]=r.split("::");return new Rqe({clientId:s,scope:d,audience:c},o)}static fromCacheEntry(r){const{scope:o,audience:s,client_id:c}=r;return new Rqe({scope:o,audience:s,clientId:c})}},qjn=class{set(r,o){localStorage.setItem(r,JSON.stringify(o))}get(r){const o=window.localStorage.getItem(r);if(o)try{return JSON.parse(o)}catch{return}}remove(r){localStorage.removeItem(r)}allKeys(){return Object.keys(window.localStorage).filter(r=>r.startsWith("@@auth0spajs@@"))}},AKt=class{constructor(){this.enclosedCache=function(){let r={};return{set(o,s){r[o]=s},get(o){const s=r[o];if(s)return s},remove(o){delete r[o]},allKeys:()=>Object.keys(r)}}()}},Vjn=class{constructor(r,o,s){this.cache=r,this.keyManifest=o,this.nowProvider=s||yKt}async setIdToken(r,o,s){var c;const d=this.getIdTokenCacheKey(r);await this.cache.set(d,{id_token:o,decodedToken:s}),await((c=this.keyManifest)===null||c===void 0?void 0:c.add(d))}async getIdToken(r){const o=await this.cache.get(this.getIdTokenCacheKey(r.clientId));if(!o&&r.scope&&r.audience){const s=await this.get(r);return!s||!s.id_token||!s.decodedToken?void 0:{id_token:s.id_token,decodedToken:s.decodedToken}}if(o)return{id_token:o.id_token,decodedToken:o.decodedToken}}async get(r,o=0){var s;let c=await this.cache.get(r.toKey());if(!c){const b=await this.getCacheKeys();if(!b)return;const y=this.matchExistingCacheKey(r,b);y&&(c=await this.cache.get(y))}if(!c)return;const d=await this.nowProvider(),h=Math.floor(d/1e3);return c.expiresAt-o<h?c.body.refresh_token?(c.body={refresh_token:c.body.refresh_token},await this.cache.set(r.toKey(),c),c.body):(await this.cache.remove(r.toKey()),void await((s=this.keyManifest)===null||s===void 0?void 0:s.remove(r.toKey()))):c.body}async set(r){var o;const s=new CB({clientId:r.client_id,scope:r.scope,audience:r.audience}),c=await this.wrapCacheEntry(r);await this.cache.set(s.toKey(),c),await((o=this.keyManifest)===null||o===void 0?void 0:o.add(s.toKey()))}async clear(r){var o;const s=await this.getCacheKeys();s&&(await s.filter(c=>!r||c.includes(r)).reduce(async(c,d)=>{await c,await this.cache.remove(d)},Promise.resolve()),await((o=this.keyManifest)===null||o===void 0?void 0:o.clear()))}async wrapCacheEntry(r){const o=await this.nowProvider();return{body:r,expiresAt:Math.floor(o/1e3)+r.expires_in}}async getCacheKeys(){var r;return this.keyManifest?(r=await this.keyManifest.get())===null||r===void 0?void 0:r.keys:this.cache.allKeys?this.cache.allKeys():void 0}getIdTokenCacheKey(r){return new CB({clientId:r},"@@auth0spajs@@","@@user@@").toKey()}matchExistingCacheKey(r,o){return o.filter(s=>{var c;const d=CB.fromKey(s),h=new Set(d.scope&&d.scope.split(" ")),b=((c=r.scope)===null||c===void 0?void 0:c.split(" "))||[],y=d.scope&&b.reduce((E,k)=>E&&h.has(k),!0);return d.prefix==="@@auth0spajs@@"&&d.clientId===r.clientId&&d.audience===r.audience&&y})[0]}},Wjn=class{constructor(r,o,s){this.storage=r,this.clientId=o,this.cookieDomain=s,this.storageKey=`a0.spajs.txs.${this.clientId}`}create(r){this.storage.save(this.storageKey,r,{daysUntilExpire:1,cookieDomain:this.cookieDomain})}get(){return this.storage.get(this.storageKey)}remove(){this.storage.remove(this.storageKey,{cookieDomain:this.cookieDomain})}};const Jne=n=>typeof n=="number",Kjn=["iss","aud","exp","nbf","iat","jti","azp","nonce","auth_time","at_hash","c_hash","acr","amr","sub_jwk","cnf","sip_from_tag","sip_date","sip_callid","sip_cseq_num","sip_via_branch","orig","dest","mky","events","toe","txn","rph","sid","vot","vtm"],Yjn=n=>{if(!n.id_token)throw new Error("ID token is required but missing");const r=(d=>{const h=d.split("."),[b,y,E]=h;if(h.length!==3||!b||!y||!E)throw new Error("ID token could not be decoded");const k=JSON.parse(d8t(y)),x={__raw:d},O={};return Object.keys(k).forEach(P=>{x[P]=k[P],Kjn.includes(P)||(O[P]=k[P])}),{encoded:{header:b,payload:y,signature:E},header:JSON.parse(d8t(b)),claims:x,user:O}})(n.id_token);if(!r.claims.iss)throw new Error("Issuer (iss) claim must be a string present in the ID token");if(r.claims.iss!==n.iss)throw new Error(`Issuer (iss) claim mismatch in the ID token; expected "${n.iss}", found "${r.claims.iss}"`);if(!r.user.sub)throw new Error("Subject (sub) claim must be a string present in the ID token");if(r.header.alg!=="RS256")throw new Error(`Signature algorithm of "${r.header.alg}" is not supported. Expected the ID token to be signed with "RS256".`);if(!r.claims.aud||typeof r.claims.aud!="string"&&!Array.isArray(r.claims.aud))throw new Error("Audience (aud) claim must be a string or array of strings present in the ID token");if(Array.isArray(r.claims.aud)){if(!r.claims.aud.includes(n.aud))throw new Error(`Audience (aud) claim mismatch in the ID token; expected "${n.aud}" but was not one of "${r.claims.aud.join(", ")}"`);if(r.claims.aud.length>1){if(!r.claims.azp)throw new Error("Authorized Party (azp) claim must be a string present in the ID token when Audience (aud) claim has multiple values");if(r.claims.azp!==n.aud)throw new Error(`Authorized Party (azp) claim mismatch in the ID token; expected "${n.aud}", found "${r.claims.azp}"`)}}else if(r.claims.aud!==n.aud)throw new Error(`Audience (aud) claim mismatch in the ID token; expected "${n.aud}" but found "${r.claims.aud}"`);if(n.nonce){if(!r.claims.nonce)throw new Error("Nonce (nonce) claim must be a string present in the ID token");if(r.claims.nonce!==n.nonce)throw new Error(`Nonce (nonce) claim mismatch in the ID token; expected "${n.nonce}", found "${r.claims.nonce}"`)}if(n.max_age&&!Jne(r.claims.auth_time))throw new Error("Authentication Time (auth_time) claim must be a number present in the ID token when Max Age (max_age) is specified");if(r.claims.exp==null||!Jne(r.claims.exp))throw new Error("Expiration Time (exp) claim must be a number present in the ID token");if(!Jne(r.claims.iat))throw new Error("Issued At (iat) claim must be a number present in the ID token");const o=n.leeway||60,s=new Date(n.now||Date.now()),c=new Date(0);if(c.setUTCSeconds(r.claims.exp+o),s>c)throw new Error(`Expiration Time (exp) claim error in the ID token; current time (${s}) is after expiration time (${c})`);if(r.claims.nbf!=null&&Jne(r.claims.nbf)){const d=new Date(0);if(d.setUTCSeconds(r.claims.nbf-o),s<d)throw new Error(`Not Before time (nbf) claim in the ID token indicates that this token can't be used just yet. Current time (${s}) is before ${d}`)}if(r.claims.auth_time!=null&&Jne(r.claims.auth_time)){const d=new Date(0);if(d.setUTCSeconds(parseInt(r.claims.auth_time)+n.max_age+o),s>d)throw new Error(`Authentication Time (auth_time) claim in the ID token indicates that too much time has passed since the last end-user authentication. Current time (${s}) is after last auth at ${d}`)}if(n.organization){const d=n.organization.trim();if(d.startsWith("org_")){const h=d;if(!r.claims.org_id)throw new Error("Organization ID (org_id) claim must be a string present in the ID token");if(h!==r.claims.org_id)throw new Error(`Organization ID (org_id) claim mismatch in the ID token; expected "${h}", found "${r.claims.org_id}"`)}else{const h=d.toLowerCase();if(!r.claims.org_name)throw new Error("Organization Name (org_name) claim must be a string present in the ID token");if(h!==r.claims.org_name)throw new Error(`Organization Name (org_name) claim mismatch in the ID token; expected "${h}", found "${r.claims.org_name}"`)}}return r};var AB=xKe(function(n,r){var o=TB&&TB.__assign||function(){return o=Object.assign||function(y){for(var E,k=1,x=arguments.length;k<x;k++)for(var O in E=arguments[k])Object.prototype.hasOwnProperty.call(E,O)&&(y[O]=E[O]);return y},o.apply(this,arguments)};function s(y,E){if(!E)return"";var k="; "+y;return E===!0?k:k+"="+E}function c(y,E,k){return encodeURIComponent(y).replace(/%(23|24|26|2B|5E|60|7C)/g,decodeURIComponent).replace(/\(/g,"%28").replace(/\)/g,"%29")+"="+encodeURIComponent(E).replace(/%(23|24|26|2B|3A|3C|3E|3D|2F|3F|40|5B|5D|5E|60|7B|7D|7C)/g,decodeURIComponent)+function(x){if(typeof x.expires=="number"){var O=new Date;O.setMilliseconds(O.getMilliseconds()+864e5*x.expires),x.expires=O}return s("Expires",x.expires?x.expires.toUTCString():"")+s("Domain",x.domain)+s("Path",x.path)+s("Secure",x.secure)+s("SameSite",x.sameSite)}(k)}function d(y){for(var E={},k=y?y.split("; "):[],x=/(%[\dA-F]{2})+/gi,O=0;O<k.length;O++){var P=k[O].split("="),D=P.slice(1).join("=");D.charAt(0)==='"'&&(D=D.slice(1,-1));try{E[P[0].replace(x,decodeURIComponent)]=D.replace(x,decodeURIComponent)}catch{}}return E}function h(){return d(document.cookie)}function b(y,E,k){document.cookie=c(y,E,o({path:"/"},k))}r.__esModule=!0,r.encode=c,r.parse=d,r.getAll=h,r.get=function(y){return h()[y]},r.set=b,r.remove=function(y,E){b(y,"",o(o({},E),{expires:-1}))}});TKe(AB),AB.encode,AB.parse,AB.getAll;var Xjn=AB.get,IKt=AB.set,OKt=AB.remove;const QV={get(n){const r=Xjn(n);if(r!==void 0)return JSON.parse(r)},save(n,r,o){let s={};window.location.protocol==="https:"&&(s={secure:!0,sameSite:"none"}),o?.daysUntilExpire&&(s.expires=o.daysUntilExpire),o?.cookieDomain&&(s.domain=o.cookieDomain),IKt(n,JSON.stringify(r),s)},remove(n,r){let o={};r?.cookieDomain&&(o.domain=r.cookieDomain),OKt(n,o)}},Jjn={get(n){return QV.get(n)||QV.get(`_legacy_${n}`)},save(n,r,o){let s={};window.location.protocol==="https:"&&(s={secure:!0}),o?.daysUntilExpire&&(s.expires=o.daysUntilExpire),o?.cookieDomain&&(s.domain=o.cookieDomain),IKt(`_legacy_${n}`,JSON.stringify(r),s),QV.save(n,r,o)},remove(n,r){let o={};r?.cookieDomain&&(o.domain=r.cookieDomain),OKt(n,o),QV.remove(n,r),QV.remove(`_legacy_${n}`,r)}},Zjn={get(n){if(typeof sessionStorage>"u")return;const r=sessionStorage.getItem(n);return r!=null?JSON.parse(r):void 0},save(n,r){sessionStorage.setItem(n,JSON.stringify(r))},remove(n){sessionStorage.removeItem(n)}};function Qjn(n,r,o){var s=r===void 0?null:r,c=function(y,E){var k=atob(y);if(E){for(var x=new Uint8Array(k.length),O=0,P=k.length;O<P;++O)x[O]=k.charCodeAt(O);return String.fromCharCode.apply(null,new Uint16Array(x.buffer))}return k}(n,o!==void 0&&o),d=c.indexOf(`
`,10)+1,h=c.substring(d)+(s?"//# sourceMappingURL="+s:""),b=new Blob([h],{type:"application/javascript"});return URL.createObjectURL(b)}var f8t,h8t,p8t,_je,e$n=(f8t="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",h8t=null,p8t=!1,function(n){return _je=_je||Qjn(f8t,h8t,p8t),new Worker(_je,n)});const kje={};let t$n=class{constructor(r,o){this.cache=r,this.clientId=o,this.manifestKey=this.createManifestKeyFrom(this.clientId)}async add(r){var o;const s=new Set(((o=await this.cache.get(this.manifestKey))===null||o===void 0?void 0:o.keys)||[]);s.add(r),await this.cache.set(this.manifestKey,{keys:[...s]})}async remove(r){const o=await this.cache.get(this.manifestKey);if(o){const s=new Set(o.keys);return s.delete(r),s.size>0?await this.cache.set(this.manifestKey,{keys:[...s]}):await this.cache.remove(this.manifestKey)}}get(){return this.cache.get(this.manifestKey)}clear(){return this.cache.remove(this.manifestKey)}createManifestKeyFrom(r){return`@@auth0spajs@@::${r}`}};const n$n={memory:()=>new AKt().enclosedCache,localstorage:()=>new qjn},g8t=n=>n$n[n],m8t=n=>{const{openUrl:r,onRedirect:o}=n,s=cI(n,["openUrl","onRedirect"]);return Object.assign(Object.assign({},s),{openUrl:r===!1||r?r:o})},Tje=new Ljn;let r$n=class{constructor(r){let o,s;if(this.userCache=new AKt().enclosedCache,this.defaultOptions={authorizationParams:{scope:"openid profile email"},useRefreshTokensFallback:!1,useFormData:!0},this._releaseLockOnPageHide=async()=>{await Tje.releaseLock("auth0.lock.getTokenSilently"),window.removeEventListener("pagehide",this._releaseLockOnPageHide)},this.options=Object.assign(Object.assign(Object.assign({},this.defaultOptions),r),{authorizationParams:Object.assign(Object.assign({},this.defaultOptions.authorizationParams),r.authorizationParams)}),typeof window<"u"&&(()=>{if(!X0e())throw new Error("For security reasons, `window.crypto` is required to run `auth0-spa-js`.");if(X0e().subtle===void 0)throw new Error(`
      auth0-spa-js must run on a secure origin. See https://github.com/auth0/auth0-spa-js/blob/main/FAQ.md#why-do-i-get-auth0-spa-js-must-run-on-a-secure-origin for more information.
    `)})(),r.cache&&r.cacheLocation&&console.warn("Both `cache` and `cacheLocation` options have been specified in the Auth0Client configuration; ignoring `cacheLocation` and using `cache`."),r.cache)s=r.cache;else{if(o=r.cacheLocation||"memory",!g8t(o))throw new Error(`Invalid cache location "${o}"`);s=g8t(o)()}this.httpTimeoutMs=r.httpTimeoutInSeconds?1e3*r.httpTimeoutInSeconds:1e4,this.cookieStorage=r.legacySameSiteCookie===!1?QV:Jjn,this.orgHintCookieName=`auth0.${this.options.clientId}.organization_hint`,this.isAuthenticatedCookieName=(h=>`auth0.${h}.is.authenticated`)(this.options.clientId),this.sessionCheckExpiryDays=r.sessionCheckExpiryDays||1;const c=r.useCookiesForTransactions?this.cookieStorage:Zjn;var d;this.scope=Xne("openid",this.options.authorizationParams.scope,this.options.useRefreshTokens?"offline_access":""),this.transactionManager=new Wjn(c,this.options.clientId,this.options.cookieDomain),this.nowProvider=this.options.nowProvider||yKt,this.cacheManager=new Vjn(s,s.allKeys?void 0:new t$n(s,this.options.clientId),this.nowProvider),this.domainUrl=(d=this.options.domain,/^https?:\/\//.test(d)?d:`https://${d}`),this.tokenIssuer=((h,b)=>h?h.startsWith("https://")?h:`https://${h}/`:`${b}/`)(this.options.issuer,this.domainUrl),typeof window<"u"&&window.Worker&&this.options.useRefreshTokens&&o==="memory"&&(this.options.workerUrl?this.worker=new Worker(this.options.workerUrl):this.worker=new e$n)}_url(r){const o=encodeURIComponent(btoa(JSON.stringify(this.options.auth0Client||wKt)));return`${this.domainUrl}${r}&auth0Client=${o}`}_authorizeUrl(r){return this._url(`/authorize?${Nqe(r)}`)}async _verifyIdToken(r,o,s){const c=await this.nowProvider();return Yjn({iss:this.tokenIssuer,aud:this.options.clientId,id_token:r,nonce:o,organization:s,leeway:this.options.leeway,max_age:(d=this.options.authorizationParams.max_age,typeof d!="string"?d:parseInt(d,10)||void 0),now:c});var d}_processOrgHint(r){r?this.cookieStorage.save(this.orgHintCookieName,r,{daysUntilExpire:this.sessionCheckExpiryDays,cookieDomain:this.options.cookieDomain}):this.cookieStorage.remove(this.orgHintCookieName,{cookieDomain:this.options.cookieDomain})}async _prepareAuthorizeUrl(r,o,s){const c=l8t(Sje()),d=l8t(Sje()),h=Sje(),b=(k=>{const x=new Uint8Array(k);return(O=>{const P={"+":"-","/":"_","=":""};return O.replace(/[+/=]/g,D=>P[D])})(window.btoa(String.fromCharCode(...Array.from(x))))})(await(async k=>await X0e().subtle.digest({name:"SHA-256"},new TextEncoder().encode(k)))(h)),y=((k,x,O,P,D,$,U,B)=>Object.assign(Object.assign(Object.assign({client_id:k.clientId},k.authorizationParams),O),{scope:Xne(x,O.scope),response_type:"code",response_mode:B||"query",state:P,nonce:D,redirect_uri:U||k.authorizationParams.redirect_uri,code_challenge:$,code_challenge_method:"S256"}))(this.options,this.scope,r,c,d,b,r.redirect_uri||this.options.authorizationParams.redirect_uri||s,o?.response_mode),E=this._authorizeUrl(y);return{nonce:d,code_verifier:h,scope:y.scope,audience:y.audience||"default",redirect_uri:y.redirect_uri,state:c,url:E}}async loginWithPopup(r,o){var s;if(r=r||{},!(o=o||{}).popup&&(o.popup=(b=>{const y=window.screenX+(window.innerWidth-400)/2,E=window.screenY+(window.innerHeight-600)/2;return window.open(b,"auth0:authorize:popup",`left=${y},top=${E},width=400,height=600,resizable,scrollbars=yes,status=1`)})(""),!o.popup))throw new Error("Unable to open a popup for loginWithPopup - window.open returned `null`");const c=await this._prepareAuthorizeUrl(r.authorizationParams||{},{response_mode:"web_message"},window.location.origin);o.popup.location.href=c.url;const d=await(b=>new Promise((y,E)=>{let k;const x=setInterval(()=>{b.popup&&b.popup.closed&&(clearInterval(x),clearTimeout(O),window.removeEventListener("message",k,!1),E(new $jn(b.popup)))},1e3),O=setTimeout(()=>{clearInterval(x),E(new jjn(b.popup)),window.removeEventListener("message",k,!1)},1e3*(b.timeoutInSeconds||60));k=function(P){if(P.data&&P.data.type==="authorization_response"){if(clearTimeout(O),clearInterval(x),window.removeEventListener("message",k,!1),b.popup.close(),P.data.response.error)return E(e2.fromPayload(P.data.response));y(P.data.response)}},window.addEventListener("message",k)}))(Object.assign(Object.assign({},o),{timeoutInSeconds:o.timeoutInSeconds||this.options.authorizeTimeoutInSeconds||60}));if(c.state!==d.state)throw new e2("state_mismatch","Invalid state");const h=((s=r.authorizationParams)===null||s===void 0?void 0:s.organization)||this.options.authorizationParams.organization;await this._requestToken({audience:c.audience,scope:c.scope,code_verifier:c.code_verifier,grant_type:"authorization_code",code:d.code,redirect_uri:c.redirect_uri},{nonceIn:c.nonce,organization:h})}async getUser(){var r;const o=await this._getIdTokenFromCache();return(r=o?.decodedToken)===null||r===void 0?void 0:r.user}async getIdTokenClaims(){var r;const o=await this._getIdTokenFromCache();return(r=o?.decodedToken)===null||r===void 0?void 0:r.claims}async loginWithRedirect(r={}){var o;const s=m8t(r),{openUrl:c,fragment:d,appState:h}=s,b=cI(s,["openUrl","fragment","appState"]),y=((o=b.authorizationParams)===null||o===void 0?void 0:o.organization)||this.options.authorizationParams.organization,E=await this._prepareAuthorizeUrl(b.authorizationParams||{}),{url:k}=E,x=cI(E,["url"]);this.transactionManager.create(Object.assign(Object.assign(Object.assign({},x),{appState:h}),y&&{organization:y}));const O=d?`${k}#${d}`:k;c?await c(O):window.location.assign(O)}async handleRedirectCallback(r=window.location.href){const o=r.split("?").slice(1);if(o.length===0)throw new Error("There are no query params available for parsing.");const{state:s,code:c,error:d,error_description:h}=(x=>{x.indexOf("#")>-1&&(x=x.substring(0,x.indexOf("#")));const O=new URLSearchParams(x);return{state:O.get("state"),code:O.get("code")||void 0,error:O.get("error")||void 0,error_description:O.get("error_description")||void 0}})(o.join("")),b=this.transactionManager.get();if(!b)throw new e2("missing_transaction","Invalid state");if(this.transactionManager.remove(),d)throw new Mjn(d,h||d,s,b.appState);if(!b.code_verifier||b.state&&b.state!==s)throw new e2("state_mismatch","Invalid state");const y=b.organization,E=b.nonce,k=b.redirect_uri;return await this._requestToken(Object.assign({audience:b.audience,scope:b.scope,code_verifier:b.code_verifier,grant_type:"authorization_code",code:c},k?{redirect_uri:k}:{}),{nonceIn:E,organization:y}),{appState:b.appState}}async checkSession(r){if(!this.cookieStorage.get(this.isAuthenticatedCookieName)){if(!this.cookieStorage.get("auth0.is.authenticated"))return;this.cookieStorage.save(this.isAuthenticatedCookieName,!0,{daysUntilExpire:this.sessionCheckExpiryDays,cookieDomain:this.options.cookieDomain}),this.cookieStorage.remove("auth0.is.authenticated")}try{await this.getTokenSilently(r)}catch{}}async getTokenSilently(r={}){var o;const s=Object.assign(Object.assign({cacheMode:"on"},r),{authorizationParams:Object.assign(Object.assign(Object.assign({},this.options.authorizationParams),r.authorizationParams),{scope:Xne(this.scope,(o=r.authorizationParams)===null||o===void 0?void 0:o.scope)})}),c=await((d,h)=>{let b=kje[h];return b||(b=d().finally(()=>{delete kje[h],b=null}),kje[h]=b),b})(()=>this._getTokenSilently(s),`${this.options.clientId}::${s.authorizationParams.audience}::${s.authorizationParams.scope}`);return r.detailedResponse?c:c?.access_token}async _getTokenSilently(r){const{cacheMode:o}=r,s=cI(r,["cacheMode"]);if(o!=="off"){const c=await this._getEntryFromCache({scope:s.authorizationParams.scope,audience:s.authorizationParams.audience||"default",clientId:this.options.clientId});if(c)return c}if(o!=="cache-only"){if(!await(async(c,d=3)=>{for(let h=0;h<d;h++)if(await c())return!0;return!1})(()=>Tje.acquireLock("auth0.lock.getTokenSilently",5e3),10))throw new Oqe;try{if(window.addEventListener("pagehide",this._releaseLockOnPageHide),o!=="off"){const E=await this._getEntryFromCache({scope:s.authorizationParams.scope,audience:s.authorizationParams.audience||"default",clientId:this.options.clientId});if(E)return E}const c=this.options.useRefreshTokens?await this._getTokenUsingRefreshToken(s):await this._getTokenFromIFrame(s),{id_token:d,access_token:h,oauthTokenScope:b,expires_in:y}=c;return Object.assign(Object.assign({id_token:d,access_token:h},b?{scope:b}:null),{expires_in:y})}finally{await Tje.releaseLock("auth0.lock.getTokenSilently"),window.removeEventListener("pagehide",this._releaseLockOnPageHide)}}}async getTokenWithPopup(r={},o={}){var s;const c=Object.assign(Object.assign({},r),{authorizationParams:Object.assign(Object.assign(Object.assign({},this.options.authorizationParams),r.authorizationParams),{scope:Xne(this.scope,(s=r.authorizationParams)===null||s===void 0?void 0:s.scope)})});return o=Object.assign(Object.assign({},Djn),o),await this.loginWithPopup(c,o),(await this.cacheManager.get(new CB({scope:c.authorizationParams.scope,audience:c.authorizationParams.audience||"default",clientId:this.options.clientId}))).access_token}async isAuthenticated(){return!!await this.getUser()}_buildLogoutUrl(r){r.clientId!==null?r.clientId=r.clientId||this.options.clientId:delete r.clientId;const o=r.logoutParams||{},{federated:s}=o,c=cI(o,["federated"]),d=s?"&federated":"";return this._url(`/v2/logout?${Nqe(Object.assign({clientId:r.clientId},c))}`)+d}async logout(r={}){const o=m8t(r),{openUrl:s}=o,c=cI(o,["openUrl"]);r.clientId===null?await this.cacheManager.clear():await this.cacheManager.clear(r.clientId||this.options.clientId),this.cookieStorage.remove(this.orgHintCookieName,{cookieDomain:this.options.cookieDomain}),this.cookieStorage.remove(this.isAuthenticatedCookieName,{cookieDomain:this.options.cookieDomain}),this.userCache.remove("@@user@@");const d=this._buildLogoutUrl(c);s?await s(d):s!==!1&&window.location.assign(d)}async _getTokenFromIFrame(r){const o=Object.assign(Object.assign({},r.authorizationParams),{prompt:"none"}),s=this.cookieStorage.get(this.orgHintCookieName);s&&!o.organization&&(o.organization=s);const{url:c,state:d,nonce:h,code_verifier:b,redirect_uri:y,scope:E,audience:k}=await this._prepareAuthorizeUrl(o,{response_mode:"web_message"},window.location.origin);try{if(window.crossOriginIsolated)throw new e2("login_required","The application is running in a Cross-Origin Isolated context, silently retrieving a token without refresh token is not possible.");const x=r.timeoutInSeconds||this.options.authorizeTimeoutInSeconds;let O;try{O=new URL(this.domainUrl).origin}catch{O=this.domainUrl}const P=await(($,U,B=60)=>new Promise((G,K)=>{const X=window.document.createElement("iframe");X.setAttribute("width","0"),X.setAttribute("height","0"),X.style.display="none";const ae=()=>{window.document.body.contains(X)&&(window.document.body.removeChild(X),window.removeEventListener("message",oe,!1))};let oe;const he=setTimeout(()=>{K(new Oqe),ae()},1e3*B);oe=function(ye){if(ye.origin!=U||!ye.data||ye.data.type!=="authorization_response")return;const we=ye.source;we&&we.close(),ye.data.response.error?K(e2.fromPayload(ye.data.response)):G(ye.data.response),clearTimeout(he),window.removeEventListener("message",oe,!1),setTimeout(ae,2e3)},window.addEventListener("message",oe,!1),window.document.body.appendChild(X),X.setAttribute("src",$)}))(c,O,x);if(d!==P.state)throw new e2("state_mismatch","Invalid state");const D=await this._requestToken(Object.assign(Object.assign({},r.authorizationParams),{code_verifier:b,code:P.code,grant_type:"authorization_code",redirect_uri:y,timeout:r.authorizationParams.timeout||this.httpTimeoutMs}),{nonceIn:h,organization:o.organization});return Object.assign(Object.assign({},D),{scope:E,oauthTokenScope:D.scope,audience:k})}catch(x){throw x.error==="login_required"&&this.logout({openUrl:!1}),x}}async _getTokenUsingRefreshToken(r){const o=await this.cacheManager.get(new CB({scope:r.authorizationParams.scope,audience:r.authorizationParams.audience||"default",clientId:this.options.clientId}));if(!(o&&o.refresh_token||this.worker)){if(this.options.useRefreshTokensFallback)return await this._getTokenFromIFrame(r);throw new xKt(r.authorizationParams.audience||"default",r.authorizationParams.scope)}const s=r.authorizationParams.redirect_uri||this.options.authorizationParams.redirect_uri||window.location.origin,c=typeof r.timeoutInSeconds=="number"?1e3*r.timeoutInSeconds:null;try{const d=await this._requestToken(Object.assign(Object.assign(Object.assign({},r.authorizationParams),{grant_type:"refresh_token",refresh_token:o&&o.refresh_token,redirect_uri:s}),c&&{timeout:c}));return Object.assign(Object.assign({},d),{scope:r.authorizationParams.scope,oauthTokenScope:d.scope,audience:r.authorizationParams.audience||"default"})}catch(d){if((d.message.indexOf("Missing Refresh Token")>-1||d.message&&d.message.indexOf("invalid refresh token")>-1)&&this.options.useRefreshTokensFallback)return await this._getTokenFromIFrame(r);throw d}}async _saveEntryInCache(r){const{id_token:o,decodedToken:s}=r,c=cI(r,["id_token","decodedToken"]);this.userCache.set("@@user@@",{id_token:o,decodedToken:s}),await this.cacheManager.setIdToken(this.options.clientId,r.id_token,r.decodedToken),await this.cacheManager.set(c)}async _getIdTokenFromCache(){const r=this.options.authorizationParams.audience||"default",o=await this.cacheManager.getIdToken(new CB({clientId:this.options.clientId,audience:r,scope:this.scope})),s=this.userCache.get("@@user@@");return o&&o.id_token===s?.id_token?s:(this.userCache.set("@@user@@",o),o)}async _getEntryFromCache({scope:r,audience:o,clientId:s}){const c=await this.cacheManager.get(new CB({scope:r,audience:o,clientId:s}),60);if(c&&c.access_token){const{access_token:d,oauthTokenScope:h,expires_in:b}=c,y=await this._getIdTokenFromCache();return y&&Object.assign(Object.assign({id_token:y.id_token,access_token:d},h?{scope:h}:null),{expires_in:b})}}async _requestToken(r,o){const{nonceIn:s,organization:c}=o||{},d=await Gjn(Object.assign({baseUrl:this.domainUrl,client_id:this.options.clientId,auth0Client:this.options.auth0Client,useFormData:this.options.useFormData,timeout:this.httpTimeoutMs},r),this.worker),h=await this._verifyIdToken(d.id_token,s,c);return await this._saveEntryInCache(Object.assign(Object.assign(Object.assign(Object.assign({},d),{decodedToken:h,scope:r.scope,audience:r.audience||"default"}),d.scope?{oauthTokenScope:d.scope}:null),{client_id:this.options.clientId})),this.cookieStorage.save(this.isAuthenticatedCookieName,!0,{daysUntilExpire:this.sessionCheckExpiryDays,cookieDomain:this.options.cookieDomain}),this._processOrgHint(c||h.claims.org_id),Object.assign(Object.assign({},d),{decodedToken:h})}async exchangeToken(r){return this._requestToken({grant_type:"urn:ietf:params:oauth:grant-type:token-exchange",subject_token:r.subject_token,subject_token_type:r.subject_token_type,scope:Xne(r.scope,this.scope),audience:r.audience||this.options.authorizationParams.audience})}};var NKt={isAuthenticated:!1,isLoading:!0,error:void 0,user:void 0},b4=function(){throw new Error("You forgot to wrap your component in <Auth0Provider>.")},i$n=Bm(Bm({},NKt),{buildAuthorizeUrl:b4,buildLogoutUrl:b4,getAccessTokenSilently:b4,getAccessTokenWithPopup:b4,getIdTokenClaims:b4,loginWithRedirect:b4,loginWithPopup:b4,logout:b4,handleRedirectCallback:b4}),RKt=Q.createContext(i$n),b8t=function(n){Njn(r,n);function r(o,s){var c=n.call(this,s??o)||this;return c.error=o,c.error_description=s,Object.setPrototypeOf(c,r.prototype),c}return r}(Error),o$n=/[?&]code=[^&]+/,a$n=/[?&]state=[^&]+/,s$n=/[?&]error=[^&]+/,c$n=function(n){return n===void 0&&(n=window.location.search),(o$n.test(n)||s$n.test(n))&&a$n.test(n)},PKt=function(n){return function(r){return r instanceof Error?r:r!==null&&typeof r=="object"&&"error"in r&&typeof r.error=="string"?"error_description"in r&&typeof r.error_description=="string"?new b8t(r.error,r.error_description):new b8t(r.error):new Error(n)}},v8t=PKt("Login failed"),xje=PKt("Get access token failed"),LKt=function(n){var r,o;n?.redirectUri&&(console.warn("Using `redirectUri` has been deprecated, please use `authorizationParams.redirect_uri` instead as `redirectUri` will be no longer supported in a future version"),n.authorizationParams=(r=n.authorizationParams)!==null&&r!==void 0?r:{},n.authorizationParams.redirect_uri=n.redirectUri,delete n.redirectUri),!((o=n?.authorizationParams)===null||o===void 0)&&o.redirectUri&&(console.warn("Using `authorizationParams.redirectUri` has been deprecated, please use `authorizationParams.redirect_uri` instead as `authorizationParams.redirectUri` will be removed in a future version"),n.authorizationParams.redirect_uri=n.authorizationParams.redirectUri,delete n.authorizationParams.redirectUri)},u$n=function(n,r){switch(r.type){case"LOGIN_POPUP_STARTED":return Bm(Bm({},n),{isLoading:!0});case"LOGIN_POPUP_COMPLETE":case"INITIALISED":return Bm(Bm({},n),{isAuthenticated:!!r.user,user:r.user,isLoading:!1,error:void 0});case"HANDLE_REDIRECT_COMPLETE":case"GET_ACCESS_TOKEN_COMPLETE":return n.user===r.user?n:Bm(Bm({},n),{isAuthenticated:!!r.user,user:r.user});case"LOGOUT":return Bm(Bm({},n),{isAuthenticated:!1,user:void 0});case"ERROR":return Bm(Bm({},n),{isLoading:!1,error:r.error})}},l$n=function(n){return LKt(n),Bm(Bm({},n),{auth0Client:{name:"auth0-react",version:"2.4.0"}})},d$n=function(n){var r;window.history.replaceState({},document.title,(r=n?.returnTo)!==null&&r!==void 0?r:window.location.pathname)},f$n=function(n){var r=n.children,o=n.skipRedirectCallback,s=n.onRedirectCallback,c=s===void 0?d$n:s,d=n.context,h=d===void 0?RKt:d,b=Rjn(n,["children","skipRedirectCallback","onRedirectCallback","context"]),y=Q.useState(function(){return new r$n(l$n(b))})[0],E=Q.useReducer(u$n,NKt),k=E[0],x=E[1],O=Q.useRef(!1),P=Q.useCallback(function(oe){return x({type:"ERROR",error:oe}),oe},[]);Q.useEffect(function(){O.current||(O.current=!0,function(){return OV(void 0,void 0,void 0,function(){var oe,he,ye;return NV(this,function(we){switch(we.label){case 0:return we.trys.push([0,7,,8]),oe=void 0,c$n()&&!o?[4,y.handleRedirectCallback()]:[3,3];case 1:return he=we.sent().appState,[4,y.getUser()];case 2:return oe=we.sent(),c(he,oe),[3,6];case 3:return[4,y.checkSession()];case 4:return we.sent(),[4,y.getUser()];case 5:oe=we.sent(),we.label=6;case 6:return x({type:"INITIALISED",user:oe}),[3,8];case 7:return ye=we.sent(),P(v8t(ye)),[3,8];case 8:return[2]}})})}())},[y,c,o,P]);var D=Q.useCallback(function(oe){return LKt(oe),y.loginWithRedirect(oe)},[y]),$=Q.useCallback(function(oe,he){return OV(void 0,void 0,void 0,function(){var ye,we;return NV(this,function(ge){switch(ge.label){case 0:x({type:"LOGIN_POPUP_STARTED"}),ge.label=1;case 1:return ge.trys.push([1,3,,4]),[4,y.loginWithPopup(oe,he)];case 2:return ge.sent(),[3,4];case 3:return ye=ge.sent(),P(v8t(ye)),[2];case 4:return[4,y.getUser()];case 5:return we=ge.sent(),x({type:"LOGIN_POPUP_COMPLETE",user:we}),[2]}})})},[y,P]),U=Q.useCallback(function(){for(var oe=[],he=0;he<arguments.length;he++)oe[he]=arguments[he];return OV(void 0,Pjn([],oe),void 0,function(ye){return ye===void 0&&(ye={}),NV(this,function(we){switch(we.label){case 0:return[4,y.logout(ye)];case 1:return we.sent(),(ye.openUrl||ye.openUrl===!1)&&x({type:"LOGOUT"}),[2]}})})},[y]),B=Q.useCallback(function(oe){return OV(void 0,void 0,void 0,function(){var he,ye,we,ge;return NV(this,function(Te){switch(Te.label){case 0:return Te.trys.push([0,2,3,5]),[4,y.getTokenSilently(oe)];case 1:return he=Te.sent(),[3,5];case 2:throw ye=Te.sent(),xje(ye);case 3:return we=x,ge={type:"GET_ACCESS_TOKEN_COMPLETE"},[4,y.getUser()];case 4:return we.apply(void 0,[(ge.user=Te.sent(),ge)]),[7];case 5:return[2,he]}})})},[y]),G=Q.useCallback(function(oe,he){return OV(void 0,void 0,void 0,function(){var ye,we,ge,Te;return NV(this,function(xe){switch(xe.label){case 0:return xe.trys.push([0,2,3,5]),[4,y.getTokenWithPopup(oe,he)];case 1:return ye=xe.sent(),[3,5];case 2:throw we=xe.sent(),xje(we);case 3:return ge=x,Te={type:"GET_ACCESS_TOKEN_COMPLETE"},[4,y.getUser()];case 4:return ge.apply(void 0,[(Te.user=xe.sent(),Te)]),[7];case 5:return[2,ye]}})})},[y]),K=Q.useCallback(function(){return y.getIdTokenClaims()},[y]),X=Q.useCallback(function(oe){return OV(void 0,void 0,void 0,function(){var he,ye,we;return NV(this,function(ge){switch(ge.label){case 0:return ge.trys.push([0,2,3,5]),[4,y.handleRedirectCallback(oe)];case 1:return[2,ge.sent()];case 2:throw he=ge.sent(),xje(he);case 3:return ye=x,we={type:"HANDLE_REDIRECT_COMPLETE"},[4,y.getUser()];case 4:return ye.apply(void 0,[(we.user=ge.sent(),we)]),[7];case 5:return[2]}})})},[y]),ae=Q.useMemo(function(){return Bm(Bm({},k),{getAccessTokenSilently:B,getAccessTokenWithPopup:G,getIdTokenClaims:K,loginWithRedirect:D,loginWithPopup:$,logout:U,handleRedirectCallback:X})},[k,B,G,K,D,$,U,X]);return Un.createElement(h.Provider,{value:ae},r)},ty=function(n){return n===void 0&&(n=RKt),Q.useContext(n)};function DKt(n){var r=Object.create(null);return function(o){return r[o]===void 0&&(r[o]=n(o)),r[o]}}var h$n=/^((children|dangerouslySetInnerHTML|key|ref|autoFocus|defaultValue|defaultChecked|innerHTML|suppressContentEditableWarning|suppressHydrationWarning|valueLink|abbr|accept|acceptCharset|accessKey|action|allow|allowUserMedia|allowPaymentRequest|allowFullScreen|allowTransparency|alt|async|autoComplete|autoPlay|capture|cellPadding|cellSpacing|challenge|charSet|checked|cite|classID|className|cols|colSpan|content|contentEditable|contextMenu|controls|controlsList|coords|crossOrigin|data|dateTime|decoding|default|defer|dir|disabled|disablePictureInPicture|disableRemotePlayback|download|draggable|encType|enterKeyHint|fetchpriority|fetchPriority|form|formAction|formEncType|formMethod|formNoValidate|formTarget|frameBorder|headers|height|hidden|high|href|hrefLang|htmlFor|httpEquiv|id|inputMode|integrity|is|keyParams|keyType|kind|label|lang|list|loading|loop|low|marginHeight|marginWidth|max|maxLength|media|mediaGroup|method|min|minLength|multiple|muted|name|nonce|noValidate|open|optimum|pattern|placeholder|playsInline|poster|preload|profile|radioGroup|readOnly|referrerPolicy|rel|required|reversed|role|rows|rowSpan|sandbox|scope|scoped|scrolling|seamless|selected|shape|size|sizes|slot|span|spellCheck|src|srcDoc|srcLang|srcSet|start|step|style|summary|tabIndex|target|title|translate|type|useMap|value|width|wmode|wrap|about|datatype|inlist|prefix|property|resource|typeof|vocab|autoCapitalize|autoCorrect|autoSave|color|incremental|fallback|inert|itemProp|itemScope|itemType|itemID|itemRef|on|option|results|security|unselectable|accentHeight|accumulate|additive|alignmentBaseline|allowReorder|alphabetic|amplitude|arabicForm|ascent|attributeName|attributeType|autoReverse|azimuth|baseFrequency|baselineShift|baseProfile|bbox|begin|bias|by|calcMode|capHeight|clip|clipPathUnits|clipPath|clipRule|colorInterpolation|colorInterpolationFilters|colorProfile|colorRendering|contentScriptType|contentStyleType|cursor|cx|cy|d|decelerate|descent|diffuseConstant|direction|display|divisor|dominantBaseline|dur|dx|dy|edgeMode|elevation|enableBackground|end|exponent|externalResourcesRequired|fill|fillOpacity|fillRule|filter|filterRes|filterUnits|floodColor|floodOpacity|focusable|fontFamily|fontSize|fontSizeAdjust|fontStretch|fontStyle|fontVariant|fontWeight|format|from|fr|fx|fy|g1|g2|glyphName|glyphOrientationHorizontal|glyphOrientationVertical|glyphRef|gradientTransform|gradientUnits|hanging|horizAdvX|horizOriginX|ideographic|imageRendering|in|in2|intercept|k|k1|k2|k3|k4|kernelMatrix|kernelUnitLength|kerning|keyPoints|keySplines|keyTimes|lengthAdjust|letterSpacing|lightingColor|limitingConeAngle|local|markerEnd|markerMid|markerStart|markerHeight|markerUnits|markerWidth|mask|maskContentUnits|maskUnits|mathematical|mode|numOctaves|offset|opacity|operator|order|orient|orientation|origin|overflow|overlinePosition|overlineThickness|panose1|paintOrder|pathLength|patternContentUnits|patternTransform|patternUnits|pointerEvents|points|pointsAtX|pointsAtY|pointsAtZ|preserveAlpha|preserveAspectRatio|primitiveUnits|r|radius|refX|refY|renderingIntent|repeatCount|repeatDur|requiredExtensions|requiredFeatures|restart|result|rotate|rx|ry|scale|seed|shapeRendering|slope|spacing|specularConstant|specularExponent|speed|spreadMethod|startOffset|stdDeviation|stemh|stemv|stitchTiles|stopColor|stopOpacity|strikethroughPosition|strikethroughThickness|string|stroke|strokeDasharray|strokeDashoffset|strokeLinecap|strokeLinejoin|strokeMiterlimit|strokeOpacity|strokeWidth|surfaceScale|systemLanguage|tableValues|targetX|targetY|textAnchor|textDecoration|textRendering|textLength|to|transform|u1|u2|underlinePosition|underlineThickness|unicode|unicodeBidi|unicodeRange|unitsPerEm|vAlphabetic|vHanging|vIdeographic|vMathematical|values|vectorEffect|version|vertAdvY|vertOriginX|vertOriginY|viewBox|viewTarget|visibility|widths|wordSpacing|writingMode|x|xHeight|x1|x2|xChannelSelector|xlinkActuate|xlinkArcrole|xlinkHref|xlinkRole|xlinkShow|xlinkTitle|xlinkType|xmlBase|xmlns|xmlnsXlink|xmlLang|xmlSpace|y|y1|y2|yChannelSelector|z|zoomAndPan|for|class|autofocus)|(([Dd][Aa][Tt][Aa]|[Aa][Rr][Ii][Aa]|x)-.*))$/,p$n=DKt(function(n){return h$n.test(n)||n.charCodeAt(0)===111&&n.charCodeAt(1)===110&&n.charCodeAt(2)<91});function g$n(n){if(n.sheet)return n.sheet;for(var r=0;r<document.styleSheets.length;r++)if(document.styleSheets[r].ownerNode===n)return document.styleSheets[r]}function m$n(n){var r=document.createElement("style");return r.setAttribute("data-emotion",n.key),n.nonce!==void 0&&r.setAttribute("nonce",n.nonce),r.appendChild(document.createTextNode("")),r.setAttribute("data-s",""),r}var b$n=function(){function n(o){var s=this;this._insertTag=function(c){var d;s.tags.length===0?s.insertionPoint?d=s.insertionPoint.nextSibling:s.prepend?d=s.container.firstChild:d=s.before:d=s.tags[s.tags.length-1].nextSibling,s.container.insertBefore(c,d),s.tags.push(c)},this.isSpeedy=o.speedy===void 0?!0:o.speedy,this.tags=[],this.ctr=0,this.nonce=o.nonce,this.key=o.key,this.container=o.container,this.prepend=o.prepend,this.insertionPoint=o.insertionPoint,this.before=null}var r=n.prototype;return r.hydrate=function(s){s.forEach(this._insertTag)},r.insert=function(s){this.ctr%(this.isSpeedy?65e3:1)===0&&this._insertTag(m$n(this));var c=this.tags[this.tags.length-1];if(this.isSpeedy){var d=g$n(c);try{d.insertRule(s,d.cssRules.length)}catch{}}else c.appendChild(document.createTextNode(s));this.ctr++},r.flush=function(){this.tags.forEach(function(s){var c;return(c=s.parentNode)==null?void 0:c.removeChild(s)}),this.tags=[],this.ctr=0},n}(),z1="-ms-",Mve="-moz-",zu="-webkit-",MKt="comm",CKe="rule",AKe="decl",v$n="@import",jKt="@keyframes",w$n="@layer",y$n=Math.abs,eye=String.fromCharCode,E$n=Object.assign;function S$n(n,r){return zb(n,0)^45?(((r<<2^zb(n,0))<<2^zb(n,1))<<2^zb(n,2))<<2^zb(n,3):0}function $Kt(n){return n.trim()}function _$n(n,r){return(n=r.exec(n))?n[0]:n}function Hu(n,r,o){return n.replace(r,o)}function Pqe(n,r){return n.indexOf(r)}function zb(n,r){return n.charCodeAt(r)|0}function Aie(n,r,o){return n.slice(r,o)}function uI(n){return n.length}function IKe(n){return n.length}function C1e(n,r){return r.push(n),n}function k$n(n,r){return n.map(r).join("")}var tye=1,zW=1,FKt=0,Qw=0,Wp=0,gK="";function nye(n,r,o,s,c,d,h){return{value:n,root:r,parent:o,type:s,props:c,children:d,line:tye,column:zW,length:h,return:""}}function Zne(n,r){return E$n(nye("",null,null,"",null,null,0),n,{length:-n.length},r)}function T$n(){return Wp}function x$n(){return Wp=Qw>0?zb(gK,--Qw):0,zW--,Wp===10&&(zW=1,tye--),Wp}function l2(){return Wp=Qw<FKt?zb(gK,Qw++):0,zW++,Wp===10&&(zW=1,tye++),Wp}function EI(){return zb(gK,Qw)}function J0e(){return Qw}function woe(n,r){return Aie(gK,n,r)}function Iie(n){switch(n){case 0:case 9:case 10:case 13:case 32:return 5;case 33:case 43:case 44:case 47:case 62:case 64:case 126:case 59:case 123:case 125:return 4;case 58:return 3;case 34:case 39:case 40:case 91:return 2;case 41:case 93:return 1}return 0}function BKt(n){return tye=zW=1,FKt=uI(gK=n),Qw=0,[]}function UKt(n){return gK="",n}function Z0e(n){return $Kt(woe(Qw-1,Lqe(n===91?n+2:n===40?n+1:n)))}function C$n(n){for(;(Wp=EI())&&Wp<33;)l2();return Iie(n)>2||Iie(Wp)>3?"":" "}function A$n(n,r){for(;--r&&l2()&&!(Wp<48||Wp>102||Wp>57&&Wp<65||Wp>70&&Wp<97););return woe(n,J0e()+(r<6&&EI()==32&&l2()==32))}function Lqe(n){for(;l2();)switch(Wp){case n:return Qw;case 34:case 39:n!==34&&n!==39&&Lqe(Wp);break;case 40:n===41&&Lqe(n);break;case 92:l2();break}return Qw}function I$n(n,r){for(;l2()&&n+Wp!==57;)if(n+Wp===84&&EI()===47)break;return"/*"+woe(r,Qw-1)+"*"+eye(n===47?n:l2())}function O$n(n){for(;!Iie(EI());)l2();return woe(n,Qw)}function N$n(n){return UKt(Q0e("",null,null,null,[""],n=BKt(n),0,[0],n))}function Q0e(n,r,o,s,c,d,h,b,y){for(var E=0,k=0,x=h,O=0,P=0,D=0,$=1,U=1,B=1,G=0,K="",X=c,ae=d,oe=s,he=K;U;)switch(D=G,G=l2()){case 40:if(D!=108&&zb(he,x-1)==58){Pqe(he+=Hu(Z0e(G),"&","&\f"),"&\f")!=-1&&(B=-1);break}case 34:case 39:case 91:he+=Z0e(G);break;case 9:case 10:case 13:case 32:he+=C$n(D);break;case 92:he+=A$n(J0e()-1,7);continue;case 47:switch(EI()){case 42:case 47:C1e(R$n(I$n(l2(),J0e()),r,o),y);break;default:he+="/"}break;case 123*$:b[E++]=uI(he)*B;case 125*$:case 59:case 0:switch(G){case 0:case 125:U=0;case 59+k:B==-1&&(he=Hu(he,/\f/g,"")),P>0&&uI(he)-x&&C1e(P>32?y8t(he+";",s,o,x-1):y8t(Hu(he," ","")+";",s,o,x-2),y);break;case 59:he+=";";default:if(C1e(oe=w8t(he,r,o,E,k,c,b,K,X=[],ae=[],x),d),G===123)if(k===0)Q0e(he,r,oe,oe,X,d,x,b,ae);else switch(O===99&&zb(he,3)===110?100:O){case 100:case 108:case 109:case 115:Q0e(n,oe,oe,s&&C1e(w8t(n,oe,oe,0,0,c,b,K,c,X=[],x),ae),c,ae,x,b,s?X:ae);break;default:Q0e(he,oe,oe,oe,[""],ae,0,b,ae)}}E=k=P=0,$=B=1,K=he="",x=h;break;case 58:x=1+uI(he),P=D;default:if($<1){if(G==123)--$;else if(G==125&&$++==0&&x$n()==125)continue}switch(he+=eye(G),G*$){case 38:B=k>0?1:(he+="\f",-1);break;case 44:b[E++]=(uI(he)-1)*B,B=1;break;case 64:EI()===45&&(he+=Z0e(l2())),O=EI(),k=x=uI(K=he+=O$n(J0e())),G++;break;case 45:D===45&&uI(he)==2&&($=0)}}return d}function w8t(n,r,o,s,c,d,h,b,y,E,k){for(var x=c-1,O=c===0?d:[""],P=IKe(O),D=0,$=0,U=0;D<s;++D)for(var B=0,G=Aie(n,x+1,x=y$n($=h[D])),K=n;B<P;++B)(K=$Kt($>0?O[B]+" "+G:Hu(G,/&\f/g,O[B])))&&(y[U++]=K);return nye(n,r,o,c===0?CKe:b,y,E,k)}function R$n(n,r,o){return nye(n,r,o,MKt,eye(T$n()),Aie(n,2,-2),0)}function y8t(n,r,o,s){return nye(n,r,o,AKe,Aie(n,0,s),Aie(n,s+1,-1),s)}function wW(n,r){for(var o="",s=IKe(n),c=0;c<s;c++)o+=r(n[c],c,n,r)||"";return o}function P$n(n,r,o,s){switch(n.type){case w$n:if(n.children.length)break;case v$n:case AKe:return n.return=n.return||n.value;case MKt:return"";case jKt:return n.return=n.value+"{"+wW(n.children,s)+"}";case CKe:n.value=n.props.join(",")}return uI(o=wW(n.children,s))?n.return=n.value+"{"+o+"}":""}function L$n(n){var r=IKe(n);return function(o,s,c,d){for(var h="",b=0;b<r;b++)h+=n[b](o,s,c,d)||"";return h}}function D$n(n){return function(r){r.root||(r=r.return)&&n(r)}}var M$n=function(r,o,s){for(var c=0,d=0;c=d,d=EI(),c===38&&d===12&&(o[s]=1),!Iie(d);)l2();return woe(r,Qw)},j$n=function(r,o){var s=-1,c=44;do switch(Iie(c)){case 0:c===38&&EI()===12&&(o[s]=1),r[s]+=M$n(Qw-1,o,s);break;case 2:r[s]+=Z0e(c);break;case 4:if(c===44){r[++s]=EI()===58?"&\f":"",o[s]=r[s].length;break}default:r[s]+=eye(c)}while(c=l2());return r},$$n=function(r,o){return UKt(j$n(BKt(r),o))},E8t=new WeakMap,F$n=function(r){if(!(r.type!=="rule"||!r.parent||r.length<1)){for(var o=r.value,s=r.parent,c=r.column===s.column&&r.line===s.line;s.type!=="rule";)if(s=s.parent,!s)return;if(!(r.props.length===1&&o.charCodeAt(0)!==58&&!E8t.get(s))&&!c){E8t.set(r,!0);for(var d=[],h=$$n(o,d),b=s.props,y=0,E=0;y<h.length;y++)for(var k=0;k<b.length;k++,E++)r.props[E]=d[y]?h[y].replace(/&\f/g,b[k]):b[k]+" "+h[y]}}},B$n=function(r){if(r.type==="decl"){var o=r.value;o.charCodeAt(0)===108&&o.charCodeAt(2)===98&&(r.return="",r.value="")}};function zKt(n,r){switch(S$n(n,r)){case 5103:return zu+"print-"+n+n;case 5737:case 4201:case 3177:case 3433:case 1641:case 4457:case 2921:case 5572:case 6356:case 5844:case 3191:case 6645:case 3005:case 6391:case 5879:case 5623:case 6135:case 4599:case 4855:case 4215:case 6389:case 5109:case 5365:case 5621:case 3829:return zu+n+n;case 5349:case 4246:case 4810:case 6968:case 2756:return zu+n+Mve+n+z1+n+n;case 6828:case 4268:return zu+n+z1+n+n;case 6165:return zu+n+z1+"flex-"+n+n;case 5187:return zu+n+Hu(n,/(\w+).+(:[^]+)/,zu+"box-$1$2"+z1+"flex-$1$2")+n;case 5443:return zu+n+z1+"flex-item-"+Hu(n,/flex-|-self/,"")+n;case 4675:return zu+n+z1+"flex-line-pack"+Hu(n,/align-content|flex-|-self/,"")+n;case 5548:return zu+n+z1+Hu(n,"shrink","negative")+n;case 5292:return zu+n+z1+Hu(n,"basis","preferred-size")+n;case 6060:return zu+"box-"+Hu(n,"-grow","")+zu+n+z1+Hu(n,"grow","positive")+n;case 4554:return zu+Hu(n,/([^-])(transform)/g,"$1"+zu+"$2")+n;case 6187:return Hu(Hu(Hu(n,/(zoom-|grab)/,zu+"$1"),/(image-set)/,zu+"$1"),n,"")+n;case 5495:case 3959:return Hu(n,/(image-set\([^]*)/,zu+"$1$`$1");case 4968:return Hu(Hu(n,/(.+:)(flex-)?(.*)/,zu+"box-pack:$3"+z1+"flex-pack:$3"),/s.+-b[^;]+/,"justify")+zu+n+n;case 4095:case 3583:case 4068:case 2532:return Hu(n,/(.+)-inline(.+)/,zu+"$1$2")+n;case 8116:case 7059:case 5753:case 5535:case 5445:case 5701:case 4933:case 4677:case 5533:case 5789:case 5021:case 4765:if(uI(n)-1-r>6)switch(zb(n,r+1)){case 109:if(zb(n,r+4)!==45)break;case 102:return Hu(n,/(.+:)(.+)-([^]+)/,"$1"+zu+"$2-$3$1"+Mve+(zb(n,r+3)==108?"$3":"$2-$3"))+n;case 115:return~Pqe(n,"stretch")?zKt(Hu(n,"stretch","fill-available"),r)+n:n}break;case 4949:if(zb(n,r+1)!==115)break;case 6444:switch(zb(n,uI(n)-3-(~Pqe(n,"!important")&&10))){case 107:return Hu(n,":",":"+zu)+n;case 101:return Hu(n,/(.+:)([^;!]+)(;|!.+)?/,"$1"+zu+(zb(n,14)===45?"inline-":"")+"box$3$1"+zu+"$2$3$1"+z1+"$2box$3")+n}break;case 5936:switch(zb(n,r+11)){case 114:return zu+n+z1+Hu(n,/[svh]\w+-[tblr]{2}/,"tb")+n;case 108:return zu+n+z1+Hu(n,/[svh]\w+-[tblr]{2}/,"tb-rl")+n;case 45:return zu+n+z1+Hu(n,/[svh]\w+-[tblr]{2}/,"lr")+n}return zu+n+z1+n+n}return n}var U$n=function(r,o,s,c){if(r.length>-1&&!r.return)switch(r.type){case AKe:r.return=zKt(r.value,r.length);break;case jKt:return wW([Zne(r,{value:Hu(r.value,"@","@"+zu)})],c);case CKe:if(r.length)return k$n(r.props,function(d){switch(_$n(d,/(::plac\w+|:read-\w+)/)){case":read-only":case":read-write":return wW([Zne(r,{props:[Hu(d,/:(read-\w+)/,":"+Mve+"$1")]})],c);case"::placeholder":return wW([Zne(r,{props:[Hu(d,/:(plac\w+)/,":"+zu+"input-$1")]}),Zne(r,{props:[Hu(d,/:(plac\w+)/,":"+Mve+"$1")]}),Zne(r,{props:[Hu(d,/:(plac\w+)/,z1+"input-$1")]})],c)}return""})}},z$n=[U$n],H$n=function(r){var o=r.key;if(o==="css"){var s=document.querySelectorAll("style[data-emotion]:not([data-s])");Array.prototype.forEach.call(s,function($){var U=$.getAttribute("data-emotion");U.indexOf(" ")!==-1&&(document.head.appendChild($),$.setAttribute("data-s",""))})}var c=r.stylisPlugins||z$n,d={},h,b=[];h=r.container||document.head,Array.prototype.forEach.call(document.querySelectorAll('style[data-emotion^="'+o+' "]'),function($){for(var U=$.getAttribute("data-emotion").split(" "),B=1;B<U.length;B++)d[U[B]]=!0;b.push($)});var y,E=[F$n,B$n];{var k,x=[P$n,D$n(function($){k.insert($)})],O=L$n(E.concat(c,x)),P=function(U){return wW(N$n(U),O)};y=function(U,B,G,K){k=G,P(U?U+"{"+B.styles+"}":B.styles),K&&(D.inserted[B.name]=!0)}}var D={key:o,sheet:new b$n({key:o,container:h,nonce:r.nonce,speedy:r.speedy,prepend:r.prepend,insertionPoint:r.insertionPoint}),nonce:r.nonce,inserted:d,registered:{},insert:y};return D.sheet.hydrate(b),D};function Dqe(){return Dqe=Object.assign?Object.assign.bind():function(n){for(var r=1;r<arguments.length;r++){var o=arguments[r];for(var s in o)({}).hasOwnProperty.call(o,s)&&(n[s]=o[s])}return n},Dqe.apply(null,arguments)}var Cje={exports:{}},rl={};/** @license React v16.13.1
 * react-is.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var S8t;function G$n(){if(S8t)return rl;S8t=1;var n=typeof Symbol=="function"&&Symbol.for,r=n?Symbol.for("react.element"):60103,o=n?Symbol.for("react.portal"):60106,s=n?Symbol.for("react.fragment"):60107,c=n?Symbol.for("react.strict_mode"):60108,d=n?Symbol.for("react.profiler"):60114,h=n?Symbol.for("react.provider"):60109,b=n?Symbol.for("react.context"):60110,y=n?Symbol.for("react.async_mode"):60111,E=n?Symbol.for("react.concurrent_mode"):60111,k=n?Symbol.for("react.forward_ref"):60112,x=n?Symbol.for("react.suspense"):60113,O=n?Symbol.for("react.suspense_list"):60120,P=n?Symbol.for("react.memo"):60115,D=n?Symbol.for("react.lazy"):60116,$=n?Symbol.for("react.block"):60121,U=n?Symbol.for("react.fundamental"):60117,B=n?Symbol.for("react.responder"):60118,G=n?Symbol.for("react.scope"):60119;function K(ae){if(typeof ae=="object"&&ae!==null){var oe=ae.$$typeof;switch(oe){case r:switch(ae=ae.type,ae){case y:case E:case s:case d:case c:case x:return ae;default:switch(ae=ae&&ae.$$typeof,ae){case b:case k:case D:case P:case h:return ae;default:return oe}}case o:return oe}}}function X(ae){return K(ae)===E}return rl.AsyncMode=y,rl.ConcurrentMode=E,rl.ContextConsumer=b,rl.ContextProvider=h,rl.Element=r,rl.ForwardRef=k,rl.Fragment=s,rl.Lazy=D,rl.Memo=P,rl.Portal=o,rl.Profiler=d,rl.StrictMode=c,rl.Suspense=x,rl.isAsyncMode=function(ae){return X(ae)||K(ae)===y},rl.isConcurrentMode=X,rl.isContextConsumer=function(ae){return K(ae)===b},rl.isContextProvider=function(ae){return K(ae)===h},rl.isElement=function(ae){return typeof ae=="object"&&ae!==null&&ae.$$typeof===r},rl.isForwardRef=function(ae){return K(ae)===k},rl.isFragment=function(ae){return K(ae)===s},rl.isLazy=function(ae){return K(ae)===D},rl.isMemo=function(ae){return K(ae)===P},rl.isPortal=function(ae){return K(ae)===o},rl.isProfiler=function(ae){return K(ae)===d},rl.isStrictMode=function(ae){return K(ae)===c},rl.isSuspense=function(ae){return K(ae)===x},rl.isValidElementType=function(ae){return typeof ae=="string"||typeof ae=="function"||ae===s||ae===E||ae===d||ae===c||ae===x||ae===O||typeof ae=="object"&&ae!==null&&(ae.$$typeof===D||ae.$$typeof===P||ae.$$typeof===h||ae.$$typeof===b||ae.$$typeof===k||ae.$$typeof===U||ae.$$typeof===B||ae.$$typeof===G||ae.$$typeof===$)},rl.typeOf=K,rl}var _8t;function q$n(){return _8t||(_8t=1,Cje.exports=G$n()),Cje.exports}var Aje,k8t;function V$n(){if(k8t)return Aje;k8t=1;var n=q$n(),r={childContextTypes:!0,contextType:!0,contextTypes:!0,defaultProps:!0,displayName:!0,getDefaultProps:!0,getDerivedStateFromError:!0,getDerivedStateFromProps:!0,mixins:!0,propTypes:!0,type:!0},o={name:!0,length:!0,prototype:!0,caller:!0,callee:!0,arguments:!0,arity:!0},s={$$typeof:!0,render:!0,defaultProps:!0,displayName:!0,propTypes:!0},c={$$typeof:!0,compare:!0,defaultProps:!0,displayName:!0,propTypes:!0,type:!0},d={};d[n.ForwardRef]=s,d[n.Memo]=c;function h(D){return n.isMemo(D)?c:d[D.$$typeof]||r}var b=Object.defineProperty,y=Object.getOwnPropertyNames,E=Object.getOwnPropertySymbols,k=Object.getOwnPropertyDescriptor,x=Object.getPrototypeOf,O=Object.prototype;function P(D,$,U){if(typeof $!="string"){if(O){var B=x($);B&&B!==O&&P(D,B,U)}var G=y($);E&&(G=G.concat(E($)));for(var K=h(D),X=h($),ae=0;ae<G.length;++ae){var oe=G[ae];if(!o[oe]&&!(U&&U[oe])&&!(X&&X[oe])&&!(K&&K[oe])){var he=k($,oe);try{b(D,oe,he)}catch{}}}}return D}return Aje=P,Aje}V$n();var W$n=!0;function HKt(n,r,o){var s="";return o.split(" ").forEach(function(c){n[c]!==void 0?r.push(n[c]+";"):c&&(s+=c+" ")}),s}var OKe=function(r,o,s){var c=r.key+"-"+o.name;(s===!1||W$n===!1)&&r.registered[c]===void 0&&(r.registered[c]=o.styles)},NKe=function(r,o,s){OKe(r,o,s);var c=r.key+"-"+o.name;if(r.inserted[o.name]===void 0){var d=o;do r.insert(o===d?"."+c:"",d,r.sheet,!0),d=d.next;while(d!==void 0)}};function K$n(n){for(var r=0,o,s=0,c=n.length;c>=4;++s,c-=4)o=n.charCodeAt(s)&255|(n.charCodeAt(++s)&255)<<8|(n.charCodeAt(++s)&255)<<16|(n.charCodeAt(++s)&255)<<24,o=(o&65535)*1540483477+((o>>>16)*59797<<16),o^=o>>>24,r=(o&65535)*1540483477+((o>>>16)*59797<<16)^(r&65535)*1540483477+((r>>>16)*59797<<16);switch(c){case 3:r^=(n.charCodeAt(s+2)&255)<<16;case 2:r^=(n.charCodeAt(s+1)&255)<<8;case 1:r^=n.charCodeAt(s)&255,r=(r&65535)*1540483477+((r>>>16)*59797<<16)}return r^=r>>>13,r=(r&65535)*1540483477+((r>>>16)*59797<<16),((r^r>>>15)>>>0).toString(36)}var Y$n={animationIterationCount:1,aspectRatio:1,borderImageOutset:1,borderImageSlice:1,borderImageWidth:1,boxFlex:1,boxFlexGroup:1,boxOrdinalGroup:1,columnCount:1,columns:1,flex:1,flexGrow:1,flexPositive:1,flexShrink:1,flexNegative:1,flexOrder:1,gridRow:1,gridRowEnd:1,gridRowSpan:1,gridRowStart:1,gridColumn:1,gridColumnEnd:1,gridColumnSpan:1,gridColumnStart:1,msGridRow:1,msGridRowSpan:1,msGridColumn:1,msGridColumnSpan:1,fontWeight:1,lineHeight:1,opacity:1,order:1,orphans:1,scale:1,tabSize:1,widows:1,zIndex:1,zoom:1,WebkitLineClamp:1,fillOpacity:1,floodOpacity:1,stopOpacity:1,strokeDasharray:1,strokeDashoffset:1,strokeMiterlimit:1,strokeOpacity:1,strokeWidth:1},X$n=/[A-Z]|^ms/g,J$n=/_EMO_([^_]+?)_([^]*?)_EMO_/g,GKt=function(r){return r.charCodeAt(1)===45},T8t=function(r){return r!=null&&typeof r!="boolean"},Ije=DKt(function(n){return GKt(n)?n:n.replace(X$n,"-$&").toLowerCase()}),x8t=function(r,o){switch(r){case"animation":case"animationName":if(typeof o=="string")return o.replace(J$n,function(s,c,d){return lI={name:c,styles:d,next:lI},c})}return Y$n[r]!==1&&!GKt(r)&&typeof o=="number"&&o!==0?o+"px":o};function Oie(n,r,o){if(o==null)return"";var s=o;if(s.__emotion_styles!==void 0)return s;switch(typeof o){case"boolean":return"";case"object":{var c=o;if(c.anim===1)return lI={name:c.name,styles:c.styles,next:lI},c.name;var d=o;if(d.styles!==void 0){var h=d.next;if(h!==void 0)for(;h!==void 0;)lI={name:h.name,styles:h.styles,next:lI},h=h.next;var b=d.styles+";";return b}return Z$n(n,r,o)}case"function":{if(n!==void 0){var y=lI,E=o(n);return lI=y,Oie(n,r,E)}break}}var k=o;if(r==null)return k;var x=r[k];return x!==void 0?x:k}function Z$n(n,r,o){var s="";if(Array.isArray(o))for(var c=0;c<o.length;c++)s+=Oie(n,r,o[c])+";";else for(var d in o){var h=o[d];if(typeof h!="object"){var b=h;r!=null&&r[b]!==void 0?s+=d+"{"+r[b]+"}":T8t(b)&&(s+=Ije(d)+":"+x8t(d,b)+";")}else if(Array.isArray(h)&&typeof h[0]=="string"&&(r==null||r[h[0]]===void 0))for(var y=0;y<h.length;y++)T8t(h[y])&&(s+=Ije(d)+":"+x8t(d,h[y])+";");else{var E=Oie(n,r,h);switch(d){case"animation":case"animationName":{s+=Ije(d)+":"+E+";";break}default:s+=d+"{"+E+"}"}}}return s}var C8t=/label:\s*([^\s;{]+)\s*(;|$)/g,lI;function RKe(n,r,o){if(n.length===1&&typeof n[0]=="object"&&n[0]!==null&&n[0].styles!==void 0)return n[0];var s=!0,c="";lI=void 0;var d=n[0];if(d==null||d.raw===void 0)s=!1,c+=Oie(o,r,d);else{var h=d;c+=h[0]}for(var b=1;b<n.length;b++)if(c+=Oie(o,r,n[b]),s){var y=d;c+=y[b]}C8t.lastIndex=0;for(var E="",k;(k=C8t.exec(c))!==null;)E+="-"+k[1];var x=K$n(c)+E;return{name:x,styles:c,next:lI}}var Q$n=function(r){return r()},qKt=n8t.useInsertionEffect?n8t.useInsertionEffect:!1,VKt=qKt||Q$n,A8t=qKt||Q.useLayoutEffect,WKt=Q.createContext(typeof HTMLElement<"u"?H$n({key:"css"}):null);WKt.Provider;var PKe=function(r){return Q.forwardRef(function(o,s){var c=Q.useContext(WKt);return r(o,c,s)})},LKe=Q.createContext({}),DKe={}.hasOwnProperty,Mqe="__EMOTION_TYPE_PLEASE_DO_NOT_USE__",eFn=function(r,o){var s={};for(var c in o)DKe.call(o,c)&&(s[c]=o[c]);return s[Mqe]=r,s},tFn=function(r){var o=r.cache,s=r.serialized,c=r.isStringTag;return OKe(o,s,c),VKt(function(){return NKe(o,s,c)}),null},nFn=PKe(function(n,r,o){var s=n.css;typeof s=="string"&&r.registered[s]!==void 0&&(s=r.registered[s]);var c=n[Mqe],d=[s],h="";typeof n.className=="string"?h=HKt(r.registered,d,n.className):n.className!=null&&(h=n.className+" ");var b=RKe(d,void 0,Q.useContext(LKe));h+=r.key+"-"+b.name;var y={};for(var E in n)DKe.call(n,E)&&E!=="css"&&E!==Mqe&&(y[E]=n[E]);return y.className=h,o&&(y.ref=o),Q.createElement(Q.Fragment,null,Q.createElement(tFn,{cache:r,serialized:b,isStringTag:typeof c=="string"}),Q.createElement(c,y))}),rFn=nFn,I8t=function(r,o){var s=arguments;if(o==null||!DKe.call(o,"css"))return Q.createElement.apply(void 0,s);var c=s.length,d=new Array(c);d[0]=rFn,d[1]=eFn(r,o);for(var h=2;h<c;h++)d[h]=s[h];return Q.createElement.apply(null,d)};(function(n){var r;r||(r=n.JSX||(n.JSX={}))})(I8t||(I8t={}));var O8t=PKe(function(n,r){var o=n.styles,s=RKe([o],void 0,Q.useContext(LKe)),c=Q.useRef();return A8t(function(){var d=r.key+"-global",h=new r.sheet.constructor({key:d,nonce:r.sheet.nonce,container:r.sheet.container,speedy:r.sheet.isSpeedy}),b=!1,y=document.querySelector('style[data-emotion="'+d+" "+s.name+'"]');return r.sheet.tags.length&&(h.before=r.sheet.tags[0]),y!==null&&(b=!0,y.setAttribute("data-emotion",d),h.hydrate([y])),c.current=[h,b],function(){h.flush()}},[r]),A8t(function(){var d=c.current,h=d[0],b=d[1];if(b){d[1]=!1;return}if(s.next!==void 0&&NKe(r,s.next,!0),h.tags.length){var y=h.tags[h.tags.length-1].nextElementSibling;h.before=y,h.flush()}r.insert("",s,h,!1)},[r,s.name]),null});function iFn(...n){return function(...o){n.forEach(s=>s?.(...o))}}const oFn=(...n)=>n.map(r=>r?.trim?.()).filter(Boolean).join(" "),aFn=/^on[A-Z]/;function HW(...n){let r={};for(let o of n){for(let s in r){if(aFn.test(s)&&typeof r[s]=="function"&&typeof o[s]=="function"){r[s]=iFn(r[s],o[s]);continue}if(s==="className"||s==="class"){r[s]=oFn(r[s],o[s]);continue}if(s==="style"){r[s]=Object.assign({},r[s]??{},o[s]??{});continue}r[s]=o[s]!==void 0?o[s]:r[s]}for(let s in o)r[s]===void 0&&(r[s]=o[s])}return r}function sFn(n,r){if(n!=null){if(typeof n=="function"){n(r);return}try{n.current=r}catch{throw new Error(`Cannot assign value '${r}' to ref '${n}'`)}}}function cFn(...n){return r=>{n.forEach(o=>{sFn(o,r)})}}function mK(n){const r=Object.assign({},n);for(let o in r)r[o]===void 0&&delete r[o];return r}const Hb=(...n)=>n.filter(Boolean).map(r=>r.trim()).join(" ");function uFn(n){return n.default||n}const Yw=n=>n!=null&&typeof n=="object"&&!Array.isArray(n),B_=n=>typeof n=="string",MKe=n=>typeof n=="function";function lFn(n){const r=Q.version;return!B_(r)||r.startsWith("18.")?n?.ref:n?.props?.ref}const KKt=(...n)=>{const r=n.reduce((o,s)=>(s?.forEach(c=>o.add(c)),o),new Set([]));return Array.from(r)};function dFn(n,r){return`${n} returned \`undefined\`. Seems you forgot to wrap component within ${r}`}function tie(n={}){const{name:r,strict:o=!0,hookName:s="useContext",providerName:c="Provider",errorMessage:d,defaultValue:h}=n,b=Q.createContext(h);b.displayName=r;function y(){const E=Q.useContext(b);if(!E&&o){const k=new Error(d??dFn(s,c));throw k.name="ContextError",Error.captureStackTrace?.(k,y),k}return E}return[b.Provider,y,b]}const[fFn,yoe]=tie({name:"ChakraContext",strict:!0,providerName:"<ChakraProvider />"});function hFn(n){const{value:r,children:o}=n;return H.jsxs(fFn,{value:r,children:[!r._config.disableLayers&&H.jsx(O8t,{styles:r.layers.atRule}),H.jsx(O8t,{styles:r._global}),o]})}const pFn=(n,r)=>{const o={},s={},c=Object.keys(n);for(const d of c)r(d)?s[d]=n[d]:o[d]=n[d];return[s,o]},yW=(n,r)=>{const o=MKe(r)?r:s=>r.includes(s);return pFn(n,o)},gFn=new Set(["htmlWidth","htmlHeight","htmlSize","htmlTranslate"]);function mFn(n){return typeof n=="string"&&gFn.has(n)}function bFn(n,r,o){const{css:s,isValidProperty:c}=yoe(),{children:d,...h}=n,b=Q.useMemo(()=>{const[O,P]=yW(h,G=>o(G,r.variantKeys)),[D,$]=yW(P,r.variantKeys),[U,B]=yW($,c);return{forwardedProps:O,variantProps:D,styleProps:U,elementProps:B}},[r.variantKeys,o,h,c]),{css:y,...E}=b.styleProps,k=Q.useMemo(()=>{const O={...b.variantProps};return r.variantKeys.includes("colorPalette")||(O.colorPalette=h.colorPalette),r.variantKeys.includes("orientation")||(O.orientation=h.orientation),r(O)},[r,b.variantProps,h.colorPalette,h.orientation]);return{styles:Q.useMemo(()=>s(k,...vFn(y),E),[s,k,y,E]),props:{...b.forwardedProps,...b.elementProps,children:d}}}const vFn=n=>(Array.isArray(n)?n:[n]).filter(Boolean).flat(),wFn=uFn(p$n),yFn=wFn,EFn=n=>n!=="theme",SFn=(n,r,o)=>{let s;if(r){const c=r.shouldForwardProp;s=n.__emotion_forwardProp&&c?d=>n.__emotion_forwardProp(d)&&c(d):c}return typeof s!="function"&&o&&(s=n.__emotion_forwardProp),s};let _Fn=typeof document<"u";const N8t=({cache:n,serialized:r,isStringTag:o})=>{OKe(n,r,o);const s=VKt(()=>NKe(n,r,o));if(!_Fn&&s!==void 0){let c=r.name,d=r.next;for(;d!==void 0;)c=Hb(c,d.name),d=d.next;return H.jsx("style",{"data-emotion":Hb(n.key,c),dangerouslySetInnerHTML:{__html:s},nonce:n.sheet.nonce})}return null},R8t={path:["d"],text:["x","y"],circle:["cx","cy","r"],rect:["width","height","x","y","rx","ry"],ellipse:["cx","cy","rx","ry"],g:["transform"],stop:["offset","stopOpacity"]},kFn=(n,r)=>Object.prototype.hasOwnProperty.call(n,r),TFn=(n,r={},o={})=>{if(kFn(R8t,n)){o.forwardProps||(o.forwardProps=[]);const E=R8t[n];o.forwardProps=KKt([...o.forwardProps,...E])}const s=n.__emotion_real===n,c=s&&n.__emotion_base||n;let d,h;o!==void 0&&(d=o.label,h=o.target);let b=[];const y=PKe((E,k,x)=>{const{cva:O,isValidProperty:P}=yoe(),D=r.__cva__?r:O(r),$=CFn(n.__emotion_cva,D),U=Re=>(_e,qe)=>Re.includes(_e)?!0:!qe?.includes(_e)&&!P(_e);!o.shouldForwardProp&&o.forwardProps&&(o.shouldForwardProp=U(o.forwardProps));const B=(Re,_e)=>{const qe=typeof n=="string"&&n.charCodeAt(0)>96?yFn:EFn,it=!_e?.includes(Re)&&!P(Re);return qe(Re)&&it},G=SFn(n,o,s)||B,K=Q.useMemo(()=>Object.assign({},o.defaultProps,mK(E)),[E]),{props:X,styles:ae}=bFn(K,$,G);let oe="",he=[ae],ye=X;if(X.theme==null){ye={};for(let Re in X)ye[Re]=X[Re];ye.theme=Q.useContext(LKe)}typeof X.className=="string"?oe=HKt(k.registered,he,X.className):X.className!=null&&(oe=Hb(oe,X.className));const we=RKe(b.concat(he),k.registered,ye);we.styles&&(oe=Hb(oe,`${k.key}-${we.name}`)),h!==void 0&&(oe=Hb(oe,h));const ge=!G("as");let Te=ge&&X.as||c,xe={};for(let Re in X)if(!(ge&&Re==="as")){if(mFn(Re)){const _e=Re.replace("html","").toLowerCase();xe[_e]=X[Re];continue}G(Re)&&(xe[Re]=X[Re])}let Ue=oe.trim();Ue?xe.className=Ue:Reflect.deleteProperty(xe,"className"),xe.ref=x;const Z=o.forwardAsChild||o.forwardProps?.includes("asChild");if(X.asChild&&!Z){const Re=Q.isValidElement(X.children)?Q.Children.only(X.children):Q.Children.toArray(X.children).find(Q.isValidElement);if(!Re)throw new Error("[chakra-ui > factory] No valid child found");Te=Re.type,xe.children=null,Reflect.deleteProperty(xe,"asChild"),xe=HW(xe,Re.props),xe.ref=cFn(x,lFn(Re))}return xe.as&&Z?(xe.as=void 0,H.jsxs(Q.Fragment,{children:[H.jsx(N8t,{cache:k,serialized:we,isStringTag:typeof Te=="string"}),H.jsx(Te,{asChild:!0,...xe,children:H.jsx(X.as,{children:xe.children})})]})):H.jsxs(Q.Fragment,{children:[H.jsx(N8t,{cache:k,serialized:we,isStringTag:typeof Te=="string"}),H.jsx(Te,{...xe})]})});return y.displayName=d!==void 0?d:`chakra(${typeof c=="string"?c:c.displayName||c.name||"Component"})`,y.__emotion_real=y,y.__emotion_base=c,y.__emotion_forwardProp=o.shouldForwardProp,y.__emotion_cva=r,Object.defineProperty(y,"toString",{value(){return`.${h}`}}),y},Oje=TFn.bind(),Nje=new Map,xFn=new Proxy(Oje,{apply(n,r,o){return Oje(...o)},get(n,r){return Nje.has(r)||Nje.set(r,Oje(r)),Nje.get(r)}}),ol=xFn,CFn=(n,r)=>n&&!r?n:!n&&r?r:n.merge(r),ai=ol("div");ai.displayName="Box";const YKt=Object.freeze({}),XKt=Object.freeze({});function jKe(n){const{key:r,recipe:o}=n,s=yoe();return Q.useMemo(()=>{const c=o||(r!=null?s.getRecipe(r):{});return s.cva(structuredClone(c))},[r,o,s])}const AFn=n=>n.charAt(0).toUpperCase()+n.slice(1);function Wx(n){const{key:r,recipe:o}=n,s=AFn(r||o.className||"Component"),[c,d]=tie({strict:!1,name:`${s}PropsContext`,providerName:`${s}PropsContext`});function h(E){const{unstyled:k,...x}=E,O=jKe({key:r,recipe:x.recipe||o}),[P,D]=Q.useMemo(()=>O.splitVariantProps(x),[O,x]);return{styles:k?YKt:O(P),className:O.className,props:D}}const b=(E,k)=>{const x=ol(E,{},k),O=Q.forwardRef((P,D)=>{const $=d(),U=Q.useMemo(()=>HW($,P),[P,$]),{styles:B,className:G,props:K}=h(U);return H.jsx(x,{...K,ref:D,css:[B,U.css],className:Hb(G,U.className)})});return O.displayName=E.displayName||E.name,O};function y(){return c}return{withContext:b,PropsProvider:c,withPropsProvider:y,usePropsContext:d,useRecipeResult:h}}const IFn=Q.forwardRef(function(r,o){const{templateAreas:s,column:c,row:d,autoFlow:h,autoRows:b,templateRows:y,autoColumns:E,templateColumns:k,inline:x,...O}=r;return H.jsx(ol.div,{...O,ref:o,css:[{display:x?"inline-grid":"grid",gridTemplateAreas:s,gridAutoColumns:E,gridColumn:c,gridRow:d,gridAutoFlow:h,gridAutoRows:b,gridTemplateRows:y,gridTemplateColumns:k},r.css]})});function eve(n){return n==null?[]:Array.isArray(n)?n:[n]}var OFn=n=>Array.from(Array(n).keys()),nie=n=>n[0],$Ke=n=>n[n.length-1],NFn=(n,r)=>n.indexOf(r)!==-1,P4=(n,...r)=>n.concat(r),$B=(n,...r)=>n.filter(o=>!r.includes(o)),sM=n=>Array.from(new Set(n)),Rje=(n,r)=>{const o=new Set(r);return n.filter(s=>!o.has(s))},GW=(n,r)=>NFn(n,r)?$B(n,r):P4(n,r);function JKt(n,r,o={}){const{step:s=1,loop:c=!0}=o,d=r+s,h=n.length,b=h-1;return r===-1?s>0?0:b:d<0?c?b:0:d>=h?c?0:r>h?h:r:d}function RFn(n,r,o={}){return n[JKt(n,r,o)]}function PFn(n,r,o={}){const{step:s=1,loop:c=!0}=o;return JKt(n,r,{step:-s,loop:c})}function LFn(n,r,o={}){return n[PFn(n,r,o)]}function P8t(n,r){return n.reduce(([o,s],c)=>(r(c)?o.push(c):s.push(c),[o,s]),[[],[]])}var L8t=n=>n?.constructor.name==="Array",DFn=(n,r)=>{if(n.length!==r.length)return!1;for(let o=0;o<n.length;o++)if(!d2(n[o],r[o]))return!1;return!0},d2=(n,r)=>{if(Object.is(n,r))return!0;if(n==null&&r!=null||n!=null&&r==null)return!1;if(typeof n?.isEqual=="function"&&typeof r?.isEqual=="function")return n.isEqual(r);if(typeof n=="function"&&typeof r=="function")return n.toString()===r.toString();if(L8t(n)&&L8t(r))return DFn(Array.from(n),Array.from(r));if(typeof n!="object"||typeof r!="object")return!1;const o=Object.keys(r??Object.create(null)),s=o.length;for(let c=0;c<s;c++)if(!Reflect.has(n,o[c]))return!1;for(let c=0;c<s;c++){const d=o[c];if(!d2(n[d],r[d]))return!1}return!0},Ore=n=>Array.isArray(n),MFn=n=>n===!0||n===!1,ZKt=n=>n!=null&&typeof n=="object",IB=n=>ZKt(n)&&!Ore(n),rie=n=>typeof n=="string",cM=n=>typeof n=="function",jFn=n=>n==null,Mx=(n,r)=>Object.prototype.hasOwnProperty.call(n,r),$Fn=n=>Object.prototype.toString.call(n),QKt=Function.prototype.toString,FFn=QKt.call(Object),BFn=n=>{if(!ZKt(n)||$Fn(n)!="[object Object]"||HFn(n))return!1;const r=Object.getPrototypeOf(n);if(r===null)return!0;const o=Mx(r,"constructor")&&r.constructor;return typeof o=="function"&&o instanceof o&&QKt.call(o)==FFn},UFn=n=>typeof n=="object"&&n!==null&&"$$typeof"in n&&"props"in n,zFn=n=>typeof n=="object"&&n!==null&&"__v_isVNode"in n,HFn=n=>UFn(n)||zFn(n),A1e=(n,...r)=>(typeof n=="function"?n(...r):n)??void 0,D8t=n=>n,GFn=n=>n(),qFn=()=>{},Nie=(...n)=>(...r)=>{n.forEach(function(o){o?.(...r)})},eYt=(()=>{let n=0;return()=>(n++,n.toString(36))})();function _4(n,r,...o){if(n in r){const c=r[n];return cM(c)?c(...o):c}const s=new Error(`No matching key: ${JSON.stringify(n)} in ${JSON.stringify(Object.keys(r))}`);throw Error.captureStackTrace?.(s,_4),s}var tYt=(n,r)=>{try{return n()}catch(o){return o instanceof Error&&Error.captureStackTrace?.(o,tYt),r?.()}},{floor:nYt,abs:M8t,round:rye,min:VFn,max:WFn,pow:KFn,sign:YFn}=Math,jqe=n=>Number.isNaN(n),DM=n=>jqe(n)?0:n,rYt=(n,r)=>(n%r+r)%r,XFn=(n,r)=>(n%r+r)%r,JFn=(n,r)=>DM(n)>=r,ZFn=(n,r)=>DM(n)<=r,iYt=(n,r,o)=>{const s=DM(n),c=r==null||s>=r,d=o==null||s<=o;return c&&d},QFn=(n,r,o)=>rye((DM(n)-r)/o)*o+r,W1=(n,r,o)=>VFn(WFn(DM(n),r),o),eBn=(n,r,o)=>(DM(n)-r)/(o-r),tBn=(n,r,o,s)=>W1(QFn(n*(o-r)+r,r,s),r,o),j8t=(n,r)=>{let o=n,s=r.toString(),c=s.indexOf("."),d=c>=0?s.length-c:0;if(d>0){let h=KFn(10,d);o=rye(o*h)/h}return o},Pje=(n,r)=>typeof r=="number"?nYt(n*r+.5)/r:rye(n),$8t=(n,r,o,s)=>{const c=r!=null?Number(r):0,d=Number(o),h=(n-c)%s;let b=M8t(h)*2>=s?n+YFn(h)*(s-M8t(h)):n-h;if(b=j8t(b,s),!jqe(c)&&b<c)b=c;else if(!jqe(d)&&b>d){const y=nYt((d-c)/s),E=c+y*s;b=y<=0||E<c?d:E}return j8t(b,s)},Cd=(n,r=0,o=10)=>{const s=Math.pow(o,r);return rye(n*s)/s},F8t=n=>{if(!Number.isFinite(n))return 0;let r=1,o=0;for(;Math.round(n*r)/r!==n;)r*=10,o+=1;return o},oYt=(n,r,o)=>{let s=r==="+"?n+o:n-o;if(n%1!==0||o%1!==0){const c=10**Math.max(F8t(n),F8t(o));n=Math.round(n*c),o=Math.round(o*c),s=r==="+"?n+o:n-o,s/=c}return s},nBn=(n,r)=>oYt(DM(n),"+",r),rBn=(n,r)=>oYt(DM(n),"-",r),I1e=n=>n!=null?`${n}px`:void 0;function bK(n){if(!BFn(n)||n===void 0)return n;const r=Reflect.ownKeys(n).filter(s=>typeof s=="string"),o={};for(const s of r){const c=n[s];c!==void 0&&(o[s]=bK(c))}return o}function iBn(n,r=Object.is){let o={...n};const s=new Set,c=k=>(s.add(k),()=>s.delete(k)),d=()=>{s.forEach(k=>k())};return{subscribe:c,get:k=>o[k],set:(k,x)=>{r(o[k],x)||(o[k]=x,d())},update:k=>{let x=!1;for(const O in k){const P=k[O];P!==void 0&&!r(o[O],P)&&(o[O]=P,x=!0)}x&&d()},snapshot:()=>({...o})}}function Lje(n,r){const o=performance.now();let s;function c(d){if(d-o>=r){n();return}s=requestAnimationFrame(c)}return s=requestAnimationFrame(c),()=>cancelAnimationFrame(s)}function qW(...n){n.length===1?n[0]:n[1],n.length===2&&n[0]}function aYt(n,r){if(n==null)throw new Error(r())}function oBn(n,r,o){let s=[];for(const c of r)n[c]==null&&s.push(c);if(s.length>0)throw new Error(`[zag-js${` > ${o}`}] missing required props: ${s.join(", ")}`)}function aBn(n,r){return`${n} returned \`undefined\`. Seems you forgot to wrap component within ${r}`}function fu(n={}){const{name:r,strict:o=!0,hookName:s="useContext",providerName:c="Provider",errorMessage:d,defaultValue:h}=n,b=Q.createContext(h);b.displayName=r;function y(){const E=Q.useContext(b);if(!E&&o){const k=new Error(d??aBn(s,c));throw k.name="ContextError",Error.captureStackTrace?.(k,y),k}return E}return[b.Provider,y,b]}const[Dni,qb]=fu({name:"EnvironmentContext",hookName:"useEnvironmentContext",providerName:"<EnvironmentProvider />",strict:!1,defaultValue:{getRootNode:()=>document,getDocument:()=>document,getWindow:()=>window}});function sBn(n){if(!n)return;const r=n.selectionStart??0,o=n.selectionEnd??0;Math.abs(o-r)===0&&r===0&&n.setSelectionRange(n.value.length,n.value.length)}var B8t=n=>Math.max(0,Math.min(1,n)),cBn=(n,r)=>n.map((o,s)=>n[(Math.max(r,0)+s)%n.length]),sYt=()=>{},iye=n=>typeof n=="object"&&n!==null,cYt=2147483647,nr=n=>n?"":void 0,gI=n=>n?"true":void 0,uBn=1,lBn=9,dBn=11,Gb=n=>iye(n)&&n.nodeType===uBn&&typeof n.nodeName=="string",FKe=n=>iye(n)&&n.nodeType===lBn,fBn=n=>iye(n)&&n===n.window,uYt=n=>Gb(n)?n.localName||"":"#document";function hBn(n){return["html","body","#document"].includes(uYt(n))}var pBn=n=>iye(n)&&n.nodeType!==void 0,VW=n=>pBn(n)&&n.nodeType===dBn&&"host"in n,gBn=n=>Gb(n)&&n.localName==="input",eW=n=>!!n?.matches("a[href]"),mBn=n=>Gb(n)?n.offsetWidth>0||n.offsetHeight>0||n.getClientRects().length>0:!1,bBn=/(textarea|select)/;function oye(n){if(n==null||!Gb(n))return!1;try{return gBn(n)&&n.selectionStart!=null||bBn.test(n.localName)||n.isContentEditable||n.getAttribute("contenteditable")==="true"||n.getAttribute("contenteditable")===""}catch{return!1}}function f2(n,r){if(!n||!r||!Gb(n)||!Gb(r))return!1;const o=r.getRootNode?.();if(n===r||n.contains(r))return!0;if(o&&VW(o)){let s=r;for(;s;){if(n===s)return!0;s=s.parentNode||s.host}}return!1}function ny(n){return FKe(n)?n:fBn(n)?n.document:n?.ownerDocument??document}function vBn(n){return ny(n).documentElement}function Jp(n){return VW(n)?Jp(n.host):FKe(n)?n.defaultView??window:Gb(n)?n.ownerDocument?.defaultView??window:window}function lYt(n){let r=n.activeElement;for(;r?.shadowRoot;){const o=r.shadowRoot.activeElement;if(o===r)break;r=o}return r}function wBn(n){if(uYt(n)==="html")return n;const r=n.assignedSlot||n.parentNode||VW(n)&&n.host||vBn(n);return VW(r)?r.host:r}var Dje=new WeakMap;function qB(n){return Dje.has(n)||Dje.set(n,Jp(n).getComputedStyle(n)),Dje.get(n)}var aye=()=>typeof document<"u";function yBn(){return navigator.userAgentData?.platform??navigator.platform}function EBn(){const n=navigator.userAgentData;return n&&Array.isArray(n.brands)?n.brands.map(({brand:r,version:o})=>`${r}/${o}`).join(" "):navigator.userAgent}var BKe=n=>aye()&&n.test(yBn()),dYt=n=>aye()&&n.test(EBn()),SBn=n=>aye()&&n.test(navigator.vendor),U8t=()=>aye()&&!!navigator.maxTouchPoints,_Bn=()=>BKe(/^iPhone/i),kBn=()=>BKe(/^iPad/i)||Eoe()&&navigator.maxTouchPoints>1,sye=()=>_Bn()||kBn(),fYt=()=>Eoe()||sye(),Eoe=()=>BKe(/^Mac/i),cye=()=>fYt()&&SBn(/apple/i),TBn=()=>dYt(/Firefox/i),xBn=()=>dYt(/Android/i);function CBn(n){return n.composedPath?.()??n.nativeEvent?.composedPath?.()}function Yp(n){return CBn(n)?.[0]??n.target}var ABn=n=>f2(n.currentTarget,Yp(n));function z8t(n){const r=n.currentTarget;if(!r||!r.matches("a[href], button[type='submit'], input[type='submit']"))return!1;const s=n.button===1,c=IBn(n);return s||c}function H8t(n){const r=n.currentTarget;if(!r)return!1;const o=r.localName;return n.altKey?o==="a"||o==="button"&&r.type==="submit"||o==="input"&&r.type==="submit":!1}function UKe(n){return LBn(n).isComposing||n.keyCode===229}function IBn(n){return Eoe()?n.metaKey:n.ctrlKey}function OBn(n){return n.key.length===1&&!n.ctrlKey&&!n.metaKey}function NBn(n){return n.pointerType===""&&n.isTrusted?!0:xBn()&&n.pointerType?n.type==="click"&&n.buttons===1:n.detail===0&&!n.pointerType}var EW=n=>n.button===0,hYt=n=>n.button===2||Eoe()&&n.ctrlKey&&n.button===0,k4=n=>n.ctrlKey||n.altKey||n.metaKey,RBn=n=>"touches"in n&&n.touches.length>0,PBn={Up:"ArrowUp",Down:"ArrowDown",Esc:"Escape"," ":"Space",",":"Comma",Left:"ArrowLeft",Right:"ArrowRight"},G8t={ArrowLeft:"ArrowRight",ArrowRight:"ArrowLeft"};function $qe(n,r={}){const{dir:o="ltr",orientation:s="horizontal"}=r;let c=n.key;return c=PBn[c]??c,o==="rtl"&&s==="horizontal"&&c in G8t&&(c=G8t[c]),c}function LBn(n){return n.nativeEvent??n}function OB(n,r="client"){const o=RBn(n)?n.touches[0]||n.changedTouches[0]:n;return{x:o[`${r}X`],y:o[`${r}Y`]}}var Ul=(n,r,o,s)=>{const c=typeof n=="function"?n():n;return c?.addEventListener(r,o,s),()=>{c?.removeEventListener(r,o,s)}};function pYt(n,r){const{type:o="HTMLInputElement",property:s="value"}=r,c=Jp(n)[o].prototype;return Object.getOwnPropertyDescriptor(c,s)??{}}function DBn(n){if(n.localName==="input")return"HTMLInputElement";if(n.localName==="textarea")return"HTMLTextAreaElement";if(n.localName==="select")return"HTMLSelectElement"}function Soe(n,r,o="value"){if(!n)return;const s=DBn(n);s&&pYt(n,{type:s,property:o}).set?.call(n,r),n.setAttribute(o,r)}function MBn(n,r){if(!n)return;pYt(n,{type:"HTMLInputElement",property:"checked"}).set?.call(n,r),r?n.setAttribute("checked",""):n.removeAttribute("checked")}function jBn(n,r){const{value:o,bubbles:s=!0}=r;if(!n)return;const c=Jp(n);n instanceof c.HTMLInputElement&&(Soe(n,`${o}`),n.dispatchEvent(new c.Event("input",{bubbles:s})))}function $Bn(n,r){const{checked:o,bubbles:s=!0}=r;if(!n)return;const c=Jp(n);n instanceof c.HTMLInputElement&&(MBn(n,o),n.dispatchEvent(new c.Event("click",{bubbles:s})))}function FBn(n){return BBn(n)?n.form:n.closest("form")}function BBn(n){return n.matches("textarea, input, select, button")}function UBn(n,r){if(!n)return;const o=FBn(n),s=c=>{c.defaultPrevented||r()};return o?.addEventListener("reset",s,{passive:!0}),()=>o?.removeEventListener("reset",s)}function zBn(n,r){const o=n?.closest("fieldset");if(!o)return;r(o.disabled);const s=Jp(o),c=new s.MutationObserver(()=>r(o.disabled));return c.observe(o,{attributes:!0,attributeFilter:["disabled"]}),()=>c.disconnect()}function uye(n,r){if(!n)return;const{onFieldsetDisabledChange:o,onFormReset:s}=r,c=[UBn(n,s),zBn(n,o)];return()=>c.forEach(d=>d?.())}var gYt=n=>Gb(n)&&n.tagName==="IFRAME",HBn=n=>!Number.isNaN(parseInt(n.getAttribute("tabindex")||"0",10)),GBn=n=>parseInt(n.getAttribute("tabindex")||"0",10)<0,zKe="input:not([type='hidden']):not([disabled]), select:not([disabled]), textarea:not([disabled]), a[href], button:not([disabled]), [tabindex], iframe, object, embed, area[href], audio[controls], video[controls], [contenteditable]:not([contenteditable='false']), details > summary:first-of-type",mYt=(n,r=!1)=>{if(!n)return[];const o=Array.from(n.querySelectorAll(zKe));(r==!0||r=="if-empty"&&o.length===0)&&Gb(n)&&hM(n)&&o.unshift(n);const c=o.filter(hM);return c.forEach((d,h)=>{if(gYt(d)&&d.contentDocument){const b=d.contentDocument.body;c.splice(h,1,...mYt(b))}}),c};function hM(n){return!n||n.closest("[inert]")?!1:n.matches(zKe)&&mBn(n)}function _oe(n,r){if(!n)return[];const o=Array.from(n.querySelectorAll(zKe)),s=o.filter(nM);return r&&nM(n)&&s.unshift(n),s.forEach((c,d)=>{if(gYt(c)&&c.contentDocument){const h=c.contentDocument.body,b=_oe(h);s.splice(d,1,...b)}}),!s.length&&r?o:s}function nM(n){return n!=null&&n.tabIndex>0?!0:hM(n)&&!GBn(n)}function bYt(n,r){const o=_oe(n,r),s=o[0]||null,c=o[o.length-1]||null;return[s,c]}function qBn(n,r){const o=_oe(n),s=n?.ownerDocument||document,c=r??s.activeElement;if(!c)return null;const d=o.indexOf(c);return o[d+1]||null}function Qne(n){return n.tabIndex<0&&(/^(audio|video|details)$/.test(n.localName)||oye(n))&&!HBn(n)?0:n.tabIndex}function Rie(n){const{root:r,getInitialEl:o,filter:s,enabled:c=!0}=n;if(!c)return;let d=null;if(d||(d=typeof o=="function"?o():o),d||(d=r?.querySelector("[data-autofocus],[autofocus]")),!d){const h=_oe(r);d=s?h.filter(s)[0]:h[0]}return d||r||void 0}function VBn(n){const r=n.currentTarget;if(!r)return!1;const[o,s]=bYt(r),c=r.ownerDocument||document;return!(c.activeElement===o&&n.shiftKey||c.activeElement===s&&!n.shiftKey||!o&&!s)}function lye(n){const r=new Set;function o(s){const c=globalThis.requestAnimationFrame(s);r.add(()=>globalThis.cancelAnimationFrame(c))}return o(()=>o(n)),function(){r.forEach(c=>c())}}function ca(n){let r;const o=globalThis.requestAnimationFrame(()=>{r=n()});return()=>{globalThis.cancelAnimationFrame(o),r?.()}}function WBn(n,r,o){const s=ca(()=>{n.removeEventListener(r,c,!0),o()}),c=()=>{s(),o()};return n.addEventListener(r,c,{once:!0,capture:!0}),s}function KBn(n,r){if(!n)return;const{attributes:o,callback:s}=r,c=n.ownerDocument.defaultView||window,d=new c.MutationObserver(h=>{for(const b of h)b.type==="attributes"&&b.attributeName&&o.includes(b.attributeName)&&s(b)});return d.observe(n,{attributes:!0,attributeFilter:o}),()=>d.disconnect()}function koe(n,r){const{defer:o}=r,s=o?ca:d=>d(),c=[];return c.push(s(()=>{const d=typeof n=="function"?n():n;c.push(KBn(d,r))})),()=>{c.forEach(d=>d?.())}}function YBn(n,r){const{callback:o}=r;if(!n)return;const s=n.ownerDocument.defaultView||window,c=new s.MutationObserver(o);return c.observe(n,{childList:!0,subtree:!0}),()=>c.disconnect()}function XBn(n,r){const{defer:o}=r,s=o?ca:d=>d(),c=[];return c.push(s(()=>{const d=typeof n=="function"?n():n;c.push(YBn(d,r))})),()=>{c.forEach(d=>d?.())}}function vYt(n){const r=()=>{const o=Jp(n);n.dispatchEvent(new o.MouseEvent("click"))};TBn()?WBn(n,"keyup",r):queueMicrotask(r)}function jve(n){const r=wBn(n);return hBn(r)?ny(r).body:Gb(r)&&HKe(r)?r:jve(r)}function wYt(n,r=[]){const o=jve(n),s=o===n.ownerDocument.body,c=Jp(o);return s?r.concat(c,c.visualViewport||[],HKe(o)?o:[]):r.concat(o,wYt(o,[]))}var JBn=/auto|scroll|overlay|hidden|clip/,ZBn=new Set(["inline","contents"]);function HKe(n){const r=Jp(n),{overflow:o,overflowX:s,overflowY:c,display:d}=r.getComputedStyle(n);return JBn.test(o+c+s)&&!ZBn.has(d)}function QBn(n){return n.scrollHeight>n.clientHeight||n.scrollWidth>n.clientWidth}function $ve(n,r){const{rootEl:o,...s}=r||{};!n||!o||!HKe(o)||!QBn(o)||n.scrollIntoView(s)}function yYt(n,r){const{left:o,top:s,width:c,height:d}=r.getBoundingClientRect(),h={x:n.x-o,y:n.y-s},b={x:B8t(h.x/c),y:B8t(h.y/d)};function y(E={}){const{dir:k="ltr",orientation:x="horizontal",inverted:O}=E,P=typeof O=="object"?O.x:O,D=typeof O=="object"?O.y:O;return x==="horizontal"?k==="rtl"||P?1-b.x:b.x:D?1-b.y:b.y}return{offset:h,percent:b,getPercentValue:y}}function eUn(n,r){const o=n.body,s="pointerLockElement"in n||"mozPointerLockElement"in n,c=()=>!!n.pointerLockElement;function d(){}function h(y){c(),console.error("PointerLock error occurred:",y),n.exitPointerLock()}if(!s)return;try{o.requestPointerLock()}catch{}const b=[Ul(n,"pointerlockchange",d,!1),Ul(n,"pointerlockerror",h,!1)];return()=>{b.forEach(y=>y()),n.exitPointerLock()}}var cW="default",Fqe="",tve=new WeakMap;function tUn(n={}){const{target:r,doc:o}=n,s=o??document,c=s.documentElement;return sye()?(cW==="default"&&(Fqe=c.style.webkitUserSelect,c.style.webkitUserSelect="none"),cW="disabled"):r&&(tve.set(r,r.style.userSelect),r.style.userSelect="none"),()=>nUn({target:r,doc:s})}function nUn(n={}){const{target:r,doc:o}=n,c=(o??document).documentElement;if(sye()){if(cW!=="disabled")return;cW="restoring",setTimeout(()=>{lye(()=>{cW==="restoring"&&(c.style.webkitUserSelect==="none"&&(c.style.webkitUserSelect=Fqe||""),Fqe="",cW="default")})},300)}else if(r&&tve.has(r)){const d=tve.get(r);r.style.userSelect==="none"&&(r.style.userSelect=d??""),r.getAttribute("style")===""&&r.removeAttribute("style"),tve.delete(r)}}function EYt(n={}){const{defer:r,target:o,...s}=n,c=r?ca:h=>h(),d=[];return d.push(c(()=>{const h=typeof o=="function"?o():o;d.push(tUn({...s,target:h}))})),()=>{d.forEach(h=>h?.())}}function SYt(n,r){const{onPointerMove:o,onPointerUp:s}=r,d=[Ul(n,"pointermove",h=>{const b=OB(h),y=Math.sqrt(b.x**2+b.y**2),E=h.pointerType==="touch"?10:5;if(!(y<E)){if(h.pointerType==="mouse"&&h.button===0){s();return}o({point:b,event:h})}},!1),Ul(n,"pointerup",s,!1),Ul(n,"pointercancel",s,!1),Ul(n,"contextmenu",s,!1),EYt({doc:n})];return()=>{d.forEach(h=>h())}}function rUn(n,r={}){const{triggerElement:o,onFocus:s,onFocusEnter:c}=r,d=n?.ownerDocument||document,h=d.body;function b(y){if(y.key!=="Tab")return;let E=null;const[k,x]=bYt(n,!0),O=qBn(h,o),P=!k&&!x;y.shiftKey&&O===d.activeElement?(c?.(),E=x):y.shiftKey&&(d.activeElement===k||P)?E=o:!y.shiftKey&&d.activeElement===o?(c?.(),E=k):!y.shiftKey&&(d.activeElement===x||P)&&(E=O),E&&(y.preventDefault(),typeof s=="function"?s(E):E.focus())}return Ul(d,"keydown",b,!0)}function iUn(n,r){const{defer:o,triggerElement:s,...c}=r,d=o?ca:b=>b(),h=[];return h.push(d(()=>{const b=typeof n=="function"?n():n,y=typeof s=="function"?s():s;h.push(rUn(b,{triggerElement:y,...c}))})),()=>{h.forEach(b=>b?.())}}function Pie(n,r){return Array.from(n?.querySelectorAll(r)??[])}function vK(n,r){return n?.querySelector(r)??null}var GKe=n=>n.id;function oUn(n,r,o=GKe){return n.find(s=>o(s)===r)}function qKe(n,r,o=GKe){const s=oUn(n,r,o);return s?n.indexOf(s):-1}function aUn(n,r,o=!0){let s=qKe(n,r);return s=o?(s+1)%n.length:Math.min(s+1,n.length-1),n[s]}function sUn(n,r,o=!0){let s=qKe(n,r);return s===-1?o?n[n.length-1]:null:(s=o?(s-1+n.length)%n.length:Math.max(0,s-1),n[s])}function cUn(n,r){const{onEntry:o,measure:s,box:c="border-box"}=r,d=(Array.isArray(n)?n:[n]).filter(Gb),h=Jp(d[0]),b=E=>{const k=d.map(x=>s(x));o({rects:k,entries:E})};b([]);const y=new h.ResizeObserver(b);return d.forEach(E=>y.observe(E,{box:c})),()=>y.disconnect()}var uUn=n=>n.split("").map(r=>{const o=r.charCodeAt(0);return o>0&&o<128?r:o>=128&&o<=255?`/x${o.toString(16)}`.replace("/","\\"):""}).join("").trim(),lUn=n=>uUn(n.dataset?.valuetext??n.textContent??""),dUn=(n,r)=>n.trim().toLowerCase().startsWith(r.toLowerCase());function fUn(n,r,o,s=GKe){const c=o?qKe(n,o,s):-1;let d=o?cBn(n,c):n;return r.length===1&&(d=d.filter(b=>s(b)!==o)),d.find(b=>dUn(lUn(b),r))}function Lie(n,r){if(!n)return sYt;const o=Object.keys(r).reduce((s,c)=>(s[c]=n.style.getPropertyValue(c),s),{});return Object.assign(n.style,r),()=>{Object.assign(n.style,o),n.style.length===0&&n.removeAttribute("style")}}function hUn(n,r,o){if(!n)return sYt;const s=n.style.getPropertyValue(r);return n.style.setProperty(r,o),()=>{n.style.setProperty(r,s),n.style.length===0&&n.removeAttribute("style")}}function pUn(n,r){const{state:o,activeId:s,key:c,timeout:d=350,itemToId:h}=r,b=o.keysSoFar+c,E=b.length>1&&Array.from(b).every(D=>D===b[0])?b[0]:b;let k=n.slice();const x=fUn(k,E,s,h);function O(){clearTimeout(o.timer),o.timer=-1}function P(D){o.keysSoFar=D,O(),D!==""&&(o.timer=+setTimeout(()=>{P(""),O()},d))}return P(b),x}var WW=Object.assign(pUn,{defaultOptions:{keysSoFar:"",timer:-1},isValidEvent:gUn});function gUn(n){return n.key.length===1&&!n.ctrlKey&&!n.metaKey}var mUn={border:"0",clip:"rect(0 0 0 0)",height:"1px",margin:"-1px",overflow:"hidden",padding:"0",position:"absolute",width:"1px",whiteSpace:"nowrap",wordWrap:"normal"};function bUn(n,r,o){const{signal:s}=r;return[new Promise((h,b)=>{const y=setTimeout(()=>{b(new Error(`Timeout of ${o}ms exceeded`))},o);s.addEventListener("abort",()=>{clearTimeout(y),b(new Error("Promise aborted"))}),n.then(E=>{s.aborted||(clearTimeout(y),h(E))}).catch(E=>{s.aborted||(clearTimeout(y),b(E))})}),()=>r.abort()]}function vUn(n,r){const{timeout:o,rootNode:s}=r,c=Jp(s),d=ny(s),h=new c.AbortController;return bUn(new Promise(b=>{const y=n();if(y){b(y);return}const E=new c.MutationObserver(()=>{const k=n();k&&k.isConnected&&(E.disconnect(),b(k))});E.observe(d.body,{childList:!0,subtree:!0})}),h,o)}var wUn=(...n)=>n.map(r=>r?.trim?.()).filter(Boolean).join(" "),yUn=/((?:--)?(?:\w+-?)+)\s*:\s*([^;]*)/g,q8t=n=>{const r={};let o;for(;o=yUn.exec(n);)r[o[1]]=o[2];return r},EUn=(n,r)=>{if(rie(n)){if(rie(r))return`${n};${r}`;n=q8t(n)}else rie(r)&&(r=q8t(r));return Object.assign({},n??{},r??{})};function Fn(...n){let r={};for(let o of n){for(let s in r){if(s.startsWith("on")&&typeof r[s]=="function"&&typeof o[s]=="function"){r[s]=Nie(o[s],r[s]);continue}if(s==="className"||s==="class"){r[s]=wUn(r[s],o[s]);continue}if(s==="style"){r[s]=EUn(r[s],o[s]);continue}r[s]=o[s]!==void 0?o[s]:r[s]}for(let s in o)r[s]===void 0&&(r[s]=o[s])}return r}function V8t(n,r,o){let s=[],c;return d=>{const h=n(d);return(h.length!==s.length||h.some((y,E)=>!d2(s[E],y)))&&(s=h,c=r(...h)),c}}function $I(){return{and:(...n)=>function(o){return n.every(s=>o.guard(s))},or:(...n)=>function(o){return n.some(s=>o.guard(s))},not:n=>function(o){return!o.guard(n)}}}function _Yt(){return{guards:$I(),createMachine:n=>n,choose:n=>function({choose:o}){return o(n)?.actions}}}var tW=(n=>(n.NotStarted="Not Started",n.Started="Started",n.Stopped="Stopped",n))(tW||{}),Mje="__init__";function SUn(n){const r=()=>n.getRootNode?.()??document,o=()=>ny(r()),s=()=>o().defaultView??window,c=()=>lYt(r());return{...n,getRootNode:r,getDoc:o,getWin:s,getActiveElement:c,isActiveElement:b=>b===c(),getById:b=>r().getElementById(b)}}function X4(...n){return r=>{const o=[];for(const s of n)if(typeof s=="function"){const c=s(r);typeof c=="function"&&o.push(c)}else s&&(s.current=r);if(o.length)return()=>{for(const s of o)s()}}}function _Un(n){let r=Object.getOwnPropertyDescriptor(n.props,"ref")?.get,o=r&&"isReactWarning"in r&&r.isReactWarning;return o?n.ref:(r=Object.getOwnPropertyDescriptor(n,"ref")?.get,o=r&&"isReactWarning"in r&&r.isReactWarning,o?n.props.ref:n.props.ref||n.ref)}const jje=n=>{const r=Q.memo(Q.forwardRef((o,s)=>{const{asChild:c,children:d,...h}=o;if(!c)return Q.createElement(n,{...h,ref:s},d);if(!Q.isValidElement(d))return null;const b=Q.Children.only(d),y=_Un(b);return Q.cloneElement(b,{...Fn(h,b.props),ref:s?X4(s,y):y})}));return r.displayName=n.displayName||n.name,r},kUn=()=>{const n=new Map;return new Proxy(jje,{apply(r,o,s){return jje(s[0])},get(r,o){const s=o;return n.has(s)||n.set(s,jje(s)),n.get(s)}})},Xn=kUn(),[Mni,b2]=fu({name:"LocaleContext",hookName:"useLocaleContext",providerName:"<LocaleProvider />",strict:!1,defaultValue:{dir:"ltr",locale:"en-US"}}),{withContext:TUn}=Wx({key:"heading"}),Kx=TUn("h2"),Ou=()=>(n,r)=>r.reduce((o,s)=>{const[c,d]=o,h=s;return d[h]!==void 0&&(c[h]=d[h]),delete d[h],[c,d]},[{},{...n}]),MM=n=>Ou()(n,["immediate","lazyMount","onExitComplete","present","skipAnimationOnMount","unmountOnExit"]);function xUn(n){return new Proxy({},{get(r,o){return o==="style"?s=>n({style:s}).style:n}})}var Wo=()=>n=>Array.from(new Set(n));function CUn(n,r){const{state:o,send:s,context:c}=n,d=o.matches("mounted","unmountSuspended");return{skip:!c.get("initial"),present:d,setNode(h){h&&s({type:"NODE.SET",node:h})},unmount(){s({type:"UNMOUNT"})}}}var AUn={props({props:n}){return{...n,present:!!n.present}},initialState({prop:n}){return n("present")?"mounted":"unmounted"},refs(){return{node:null,styles:null}},context({bindable:n}){return{unmountAnimationName:n(()=>({defaultValue:null})),prevAnimationName:n(()=>({defaultValue:null})),present:n(()=>({defaultValue:!1})),initial:n(()=>({sync:!0,defaultValue:!1}))}},exit:["clearInitial","cleanupNode"],watch({track:n,prop:r,send:o}){n([()=>r("present")],()=>{o({type:"PRESENCE.CHANGED"})})},on:{"NODE.SET":{actions:["setupNode"]},"PRESENCE.CHANGED":{actions:["setInitial","syncPresence"]}},states:{mounted:{on:{UNMOUNT:{target:"unmounted",actions:["clearPrevAnimationName","invokeOnExitComplete"]},"UNMOUNT.SUSPEND":{target:"unmountSuspended"}}},unmountSuspended:{effects:["trackAnimationEvents"],on:{MOUNT:{target:"mounted",actions:["setPrevAnimationName"]},UNMOUNT:{target:"unmounted",actions:["clearPrevAnimationName","invokeOnExitComplete"]}}},unmounted:{on:{MOUNT:{target:"mounted",actions:["setPrevAnimationName"]}}}},implementations:{actions:{setInitial:({context:n})=>{n.get("initial")||queueMicrotask(()=>{n.set("initial",!0)})},clearInitial:({context:n})=>{n.set("initial",!1)},invokeOnExitComplete:({prop:n})=>{n("onExitComplete")?.()},setupNode:({refs:n,event:r})=>{n.get("node")!==r.node&&(n.set("node",r.node),n.set("styles",qB(r.node)))},cleanupNode:({refs:n})=>{n.set("node",null),n.set("styles",null)},syncPresence:({context:n,refs:r,send:o,prop:s})=>{const c=s("present");if(c)return o({type:"MOUNT",src:"presence.changed"});const d=r.get("node");if(!c&&d?.ownerDocument.visibilityState==="hidden")return o({type:"UNMOUNT",src:"visibilitychange"});ca(()=>{const h=O1e(r.get("styles"));n.set("unmountAnimationName",h),h==="none"||h===n.get("prevAnimationName")||r.get("styles")?.display==="none"||r.get("styles")?.animationDuration==="0s"?o({type:"UNMOUNT",src:"presence.changed"}):o({type:"UNMOUNT.SUSPEND"})})},setPrevAnimationName:({context:n,refs:r})=>{ca(()=>{n.set("prevAnimationName",O1e(r.get("styles")))})},clearPrevAnimationName:({context:n})=>{n.set("prevAnimationName",null)}},effects:{trackAnimationEvents:({context:n,refs:r,send:o})=>{const s=r.get("node");if(!s)return;const c=b=>{(b.composedPath?.()?.[0]??b.target)===s&&n.set("prevAnimationName",O1e(r.get("styles")))},d=b=>{const y=O1e(r.get("styles"));Yp(b)===s&&y===n.get("unmountAnimationName")&&o({type:"UNMOUNT",src:"animationend"})};s.addEventListener("animationstart",c),s.addEventListener("animationcancel",d),s.addEventListener("animationend",d);const h=Lie(s,{animationFillMode:"forwards"});return()=>{s.removeEventListener("animationstart",c),s.removeEventListener("animationcancel",d),s.removeEventListener("animationend",d),lye(()=>h())}}}}};function O1e(n){return n?.animationName||"none"}Wo()(["onExitComplete","present","immediate"]);var dye=vKt(),kYt=typeof globalThis.document<"u"?Q.useLayoutEffect:Q.useEffect;function Fve(n){const r=n().value??n().defaultValue,o=n().isEqual??Object.is,[s]=Q.useState(r),[c,d]=Q.useState(s),h=n().value!==void 0,b=Q.useRef(c);b.current=h?n().value:c;const y=Q.useRef(b.current);kYt(()=>{y.current=b.current},[c,n().value]);const E=x=>{const O=y.current,P=cM(x)?x(O):x;n().debug&&console.log(`[bindable > ${n().debug}] setValue`,{next:P,prev:O}),h||d(P),o(P,O)||n().onChange?.(P,O)};function k(){return h?n().value:c}return{initial:s,ref:b,get:k,set(x){(n().sync?dye.flushSync:GFn)(()=>E(x))},invoke(x,O){n().onChange?.(x,O)},hash(x){return n().hash?.(x)??String(x)}}}Fve.cleanup=n=>{Q.useEffect(()=>n,[])};Fve.ref=n=>{const r=Q.useRef(n);return{get:()=>r.current,set:o=>{r.current=o}}};function IUn(n){const r=Q.useRef(n);return{get(o){return r.current[o]},set(o,s){r.current[o]=s}}}var OUn=(n,r)=>{const o=Q.useRef(!1),s=Q.useRef(!1);Q.useEffect(()=>{if(o.current&&s.current)return r();s.current=!0},[...(n??[]).map(c=>typeof c=="function"?c():c)]),Q.useEffect(()=>(o.current=!0,()=>{o.current=!1}),[])};function uv(n,r={}){const o=Q.useMemo(()=>{const{id:Te,ids:xe,getRootNode:Ue}=r;return SUn({id:Te,ids:xe,getRootNode:Ue})},[r]),s=(...Te)=>{n.debug&&console.log(...Te)},c=n.props?.({props:bK(r),scope:o})??r,d=NUn(c),h=n.context?.({prop:d,bindable:Fve,scope:o,flush:W8t,getContext(){return y},getComputed(){return ae},getRefs(){return $},getEvent(){return P()}}),b=TYt(h),y={get(Te){return b.current?.[Te].ref.current},set(Te,xe){b.current?.[Te].set(xe)},initial(Te){return b.current?.[Te].initial},hash(Te){const xe=b.current?.[Te].get();return b.current?.[Te].hash(xe)}},E=Q.useRef(new Map),k=Q.useRef(null),x=Q.useRef(null),O=Q.useRef({type:""}),P=()=>({...O.current,current(){return O.current},previous(){return x.current}}),D=()=>({...oe,matches(...Te){return Te.includes(oe.ref.current)},hasTag(Te){return!!n.states[oe.ref.current]?.tags?.includes(Te)}}),$=IUn(n.refs?.({prop:d,context:y})??{}),U=()=>({state:D(),context:y,event:P(),prop:d,send:ge,action:B,guard:G,track:OUn,refs:$,computed:ae,flush:W8t,scope:o,choose:X}),B=Te=>{const xe=cM(Te)?Te(U()):Te;if(!xe)return;const Ue=xe.map(Z=>{const Re=n.implementations?.actions?.[Z];return Re||qW(`[zag-js] No implementation found for action "${JSON.stringify(Z)}"`),Re});for(const Z of Ue)Z?.(U())},G=Te=>cM(Te)?Te(U()):n.implementations?.guards?.[Te](U()),K=Te=>{const xe=cM(Te)?Te(U()):Te;if(!xe)return;const Ue=xe.map(Re=>{const _e=n.implementations?.effects?.[Re];return _e||qW(`[zag-js] No implementation found for effect "${JSON.stringify(Re)}"`),_e}),Z=[];for(const Re of Ue){const _e=Re?.(U());_e&&Z.push(_e)}return()=>Z.forEach(Re=>Re?.())},X=Te=>eve(Te).find(xe=>{let Ue=!xe.guard;return rie(xe.guard)?Ue=!!G(xe.guard):cM(xe.guard)&&(Ue=xe.guard(U())),Ue}),ae=Te=>{aYt(n.computed,()=>"[zag-js] No computed object found on machine");const xe=n.computed[Te];return xe({context:y,event:P(),prop:d,refs:$,scope:o,computed:ae})},oe=Fve(()=>({defaultValue:n.initialState({prop:d}),onChange(Te,xe){xe&&(E.current.get(xe)?.(),E.current.delete(xe)),xe&&B(n.states[xe]?.exit),B(k.current?.actions);const Ue=K(n.states[Te]?.effects);if(Ue&&E.current.set(Te,Ue),xe===Mje){B(n.entry);const Z=K(n.effects);Z&&E.current.set(Mje,Z)}B(n.states[Te]?.entry)}})),he=Q.useRef(void 0),ye=Q.useRef(tW.NotStarted);kYt(()=>{queueMicrotask(()=>{const Ue=ye.current===tW.Started;ye.current=tW.Started,s(Ue?"rehydrating...":"initializing...");const Z=he.current??oe.initial;oe.invoke(Z,Ue?oe.get():Mje)});const Te=E.current,xe=oe.ref.current;return()=>{s("unmounting..."),he.current=xe,ye.current=tW.Stopped,Te.forEach(Ue=>Ue?.()),E.current=new Map,k.current=null,queueMicrotask(()=>{B(n.exit)})}},[]);const we=()=>"ref"in oe?oe.ref.current:oe.get(),ge=Te=>{queueMicrotask(()=>{if(ye.current!==tW.Started)return;x.current=O.current,O.current=Te;let xe=we();const Ue=n.states[xe].on?.[Te.type]??n.on?.[Te.type],Z=X(Ue);if(!Z)return;k.current=Z;const Re=Z.target??xe;s("transition",Te.type,Z.target||xe,`(${Z.actions})`);const _e=Re!==xe;_e?dye.flushSync(()=>oe.set(Re)):Z.reenter&&!_e?oe.invoke(xe,xe):B(Z.actions??[])})};return n.watch?.(U()),{state:D(),send:ge,context:y,prop:d,scope:o,refs:$,computed:ae,event:P(),getStatus:()=>ye.current}}function TYt(n){const r=Q.useRef(n);return r.current=n,r}function NUn(n){const r=TYt(n);return function(s){return r.current[s]}}function W8t(n){queueMicrotask(()=>{dye.flushSync(()=>n())})}var ry=xUn(n=>n);function RUn(n,r={}){const{sync:o=!1}=r,s=PUn(n);return Q.useCallback((...c)=>o?queueMicrotask(()=>s.current?.(...c)):s.current?.(...c),[o,s])}function PUn(n){const r=Q.useRef(n);return r.current=n,r}const J4=(n={})=>{const{lazyMount:r,unmountOnExit:o,present:s,skipAnimationOnMount:c=!1,...d}=n,h=Q.useRef(!1),b={...d,present:s,onExitComplete:RUn(n.onExitComplete)},y=uv(AUn,b),E=CUn(y);E.present&&(h.current=!0);const k=!E.present&&!h.current&&r||o&&!E.present&&h.current,x=()=>({"data-state":E.skip&&c?void 0:s?"open":"closed",hidden:!E.present});return{ref:E.setNode,getPresenceProps:x,present:E.present,unmounted:k}},[jM,Yx]=fu({name:"PresenceContext",hookName:"usePresenceContext",providerName:"<PresenceProvider />"}),N1e=ol("span"),{withContext:LUn}=Wx({key:"text"}),Fi=LUn("p");function K8t(n,r=[]){const o=Q.useRef(()=>{throw new Error("Cannot call an event handler while rendering.")});return Q.useInsertionEffect(()=>{o.current=n}),Q.useCallback((...s)=>o.current?.(...s),r)}function _M(n={}){const r=K8t(n.onOpen),o=K8t(n.onClose),[s,c]=Q.useState(n.defaultOpen||!1),d=n.open!==void 0?n.open:s,h=n.open!==void 0,b=Q.useCallback(()=>{h||c(!1),o?.()},[h,o]),y=Q.useCallback(()=>{h||c(!0),r?.()},[h,r]),E=Q.useCallback(()=>{d?b():y()},[d,y,b]);return{open:d,onOpen:y,onClose:b,onToggle:E,setOpen:c}}const DUn=n=>Ou()(n,["defaultOpen","disabled","id","ids","lazyMount","onExitComplete","onOpenChange","open","unmountOnExit"]);var Vo=(n,r=[])=>({parts:(...o)=>{if(MUn(r))return Vo(n,o);throw new Error("createAnatomy().parts(...) should only be called once. Did you mean to use .extendWith(...) ?")},extendWith:(...o)=>Vo(n,[...r,...o]),omit:(...o)=>Vo(n,r.filter(s=>!o.includes(s))),rename:o=>Vo(o,r),keys:()=>r,build:()=>[...new Set(r)].reduce((o,s)=>Object.assign(o,{[s]:{selector:[`&[data-scope="${RV(n)}"][data-part="${RV(s)}"]`,`& [data-scope="${RV(n)}"][data-part="${RV(s)}"]`].join(", "),attrs:{"data-scope":RV(n),"data-part":RV(s)}}}),{})}),RV=n=>n.replace(/([A-Z])([A-Z])/g,"$1-$2").replace(/([a-z])([A-Z])/g,"$1-$2").replace(/[\s_]+/g,"-").toLowerCase(),MUn=n=>n.length===0,xYt=Vo("collapsible").parts("root","trigger","content","indicator"),R1e=xYt.build(),jUn=n=>n.ids?.root??`collapsible:${n.id}`,Bqe=n=>n.ids?.content??`collapsible:${n.id}:content`,$Un=n=>n.ids?.trigger??`collapsible:${n.id}:trigger`,P1e=n=>n.getById(Bqe(n));function FUn(n,r){const{state:o,send:s,context:c,scope:d,prop:h}=n,b=o.matches("open")||o.matches("closing"),y=o.matches("open"),{width:E,height:k}=c.get("size"),x=!!h("disabled"),O=!c.get("initial")&&y,P="ltr";return{disabled:x,visible:b,open:y,measureSize(){s({type:"size.measure"})},setOpen(D){o.matches("open")!==D&&s({type:D?"open":"close"})},getRootProps(){return r.element({...R1e.root.attrs,"data-state":y?"open":"closed",dir:P,id:jUn(d)})},getContentProps(){return r.element({...R1e.content.attrs,"data-collapsible":"","data-state":O?void 0:y?"open":"closed",id:Bqe(d),"data-disabled":nr(x),hidden:!b,style:{"--height":k!=null?`${k}px`:void 0,"--width":E!=null?`${E}px`:void 0}})},getTriggerProps(){return r.element({...R1e.trigger.attrs,id:$Un(d),dir:P,type:"button","data-state":y?"open":"closed","data-disabled":nr(x),"aria-controls":Bqe(d),"aria-expanded":b||!1,onClick(D){D.defaultPrevented||x||s({type:y?"close":"open"})}})},getIndicatorProps(){return r.element({...R1e.indicator.attrs,dir:P,"data-state":y?"open":"closed","data-disabled":nr(x)})}}}var BUn={initialState({prop:n}){return n("open")||n("defaultOpen")?"open":"closed"},context({bindable:n}){return{size:n(()=>({defaultValue:{height:0,width:0},sync:!0})),initial:n(()=>({defaultValue:!1}))}},refs(){return{cleanup:void 0,stylesRef:void 0}},watch({track:n,prop:r,action:o}){n([()=>r("open")],()=>{o(["setInitial","computeSize","toggleVisibility"])})},exit:["clearInitial","cleanupNode"],states:{closed:{on:{"controlled.open":{target:"open"},open:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["setInitial","computeSize","invokeOnOpen"]}]}},closing:{effects:["trackExitAnimation"],on:{"controlled.close":{target:"closed"},"controlled.open":{target:"open"},open:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["setInitial","invokeOnOpen"]}],close:[{guard:"isOpenControlled",actions:["invokeOnExitComplete"]},{target:"closed",actions:["setInitial","computeSize","invokeOnExitComplete"]}],"animation.end":{target:"closed",actions:["invokeOnExitComplete","clearInitial"]}}},open:{effects:["trackEnterAnimation"],on:{"controlled.close":{target:"closing"},close:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closing",actions:["setInitial","computeSize","invokeOnClose"]}],"size.measure":{actions:["measureSize"]},"animation.end":{actions:["clearInitial"]}}}},implementations:{guards:{isOpenControlled:({prop:n})=>n("open")!=null},effects:{trackEnterAnimation:({send:n,scope:r})=>{let o;const s=ca(()=>{const c=P1e(r);if(!c)return;const d=qB(c).animationName;if(!d||d==="none"){n({type:"animation.end"});return}const b=y=>{Yp(y)===c&&n({type:"animation.end"})};c.addEventListener("animationend",b),o=()=>{c.removeEventListener("animationend",b)}});return()=>{s(),o?.()}},trackExitAnimation:({send:n,scope:r})=>{let o;const s=ca(()=>{const c=P1e(r);if(!c)return;const d=qB(c).animationName;if(!d||d==="none"){n({type:"animation.end"});return}const b=E=>{Yp(E)===c&&n({type:"animation.end"})};c.addEventListener("animationend",b);const y=Lie(c,{animationFillMode:"forwards"});o=()=>{c.removeEventListener("animationend",b),lye(()=>y())}});return()=>{s(),o?.()}}},actions:{setInitial:({context:n,flush:r})=>{r(()=>{n.set("initial",!0)})},clearInitial:({context:n})=>{n.set("initial",!1)},cleanupNode:({refs:n})=>{n.set("stylesRef",null)},measureSize:({context:n,scope:r})=>{const o=P1e(r);if(!o)return;const{height:s,width:c}=o.getBoundingClientRect();n.set("size",{height:s,width:c})},computeSize:({refs:n,scope:r,context:o})=>{n.get("cleanup")?.();const s=ca(()=>{const c=P1e(r);if(!c)return;const d=c.hidden;c.style.animationName="none",c.style.animationDuration="0s",c.hidden=!1;const h=c.getBoundingClientRect();o.set("size",{height:h.height,width:h.width}),o.get("initial")&&(c.style.animationName="",c.style.animationDuration=""),c.hidden=d});n.set("cleanup",s)},invokeOnOpen:({prop:n})=>{n("onOpenChange")?.({open:!0})},invokeOnClose:({prop:n})=>{n("onOpenChange")?.({open:!1})},invokeOnExitComplete:({prop:n})=>{n("onExitComplete")?.()},toggleVisibility:({prop:n,send:r})=>{r({type:n("open")?"controlled.open":"controlled.close"})}}}};Wo()(["dir","disabled","getRootNode","id","ids","onExitComplete","onOpenChange","defaultOpen","open"]);const UUn=(n={})=>{const{lazyMount:r,unmountOnExit:o,...s}=n,c=Q.useId(),d=Q.useRef(!1),{dir:h}=b2(),{getRootNode:b}=qb(),y={id:c,dir:h,getRootNode:b,...s},E=uv(BUn,y),k=FUn(E,ry);k.visible&&(d.current=!0);const x=!k.visible&&!d.current&&r||o&&!k.visible&&d.current;return{...k,isUnmounted:x}},[zUn,HUn]=fu({name:"CollapsibleContext",hookName:"useCollapsibleContext",providerName:"<CollapsibleProvider />"}),CYt=Q.forwardRef((n,r)=>{const[o,s]=DUn(n),c=UUn(o),d=Fn(c.getRootProps(),s);return H.jsx(zUn,{value:c,children:H.jsx(Xn.div,{...d,ref:r})})});CYt.displayName="CollapsibleRoot";var GUn=Object.defineProperty,qUn=(n,r,o)=>r in n?GUn(n,r,{enumerable:!0,configurable:!0,writable:!0,value:o}):n[r]=o,VKe=(n,r,o)=>qUn(n,r+"",o),VUn=(n,r)=>{if(Object.keys(n).length!==Object.keys(r).length)return!1;for(let o in n)if(n[o]!==r[o])return!1;return!0},WKe=class{toHexInt(){return this.toFormat("rgba").toHexInt()}getChannelValue(r){if(r in this)return this[r];throw new Error("Unsupported color channel: "+r)}getChannelValuePercent(r,o){const s=o??this.getChannelValue(r),{minValue:c,maxValue:d}=this.getChannelRange(r);return eBn(s,c,d)}getChannelPercentValue(r,o){const{minValue:s,maxValue:c,step:d}=this.getChannelRange(r),h=tBn(o,s,c,d);return $8t(h,s,c,d)}withChannelValue(r,o){const{minValue:s,maxValue:c}=this.getChannelRange(r);if(r in this){let d=this.clone();return d[r]=W1(o,s,c),d}throw new Error("Unsupported color channel: "+r)}getColorAxes(r){let{xChannel:o,yChannel:s}=r,c=o||this.getChannels().find(b=>b!==s),d=s||this.getChannels().find(b=>b!==c),h=this.getChannels().find(b=>b!==c&&b!==d);return{xChannel:c,yChannel:d,zChannel:h}}incrementChannel(r,o){const{minValue:s,maxValue:c,step:d}=this.getChannelRange(r),h=$8t(W1(this.getChannelValue(r)+o,s,c),s,c,d);return this.withChannelValue(r,h)}decrementChannel(r,o){return this.incrementChannel(r,-o)}isEqual(r){return VUn(this.toJSON(),r.toJSON())&&this.getChannelValue("alpha")===r.getChannelValue("alpha")}},WUn=/^#[\da-f]+$/i,KUn=/^rgba?\((.*)\)$/,YUn=/[^#]/gi,AYt=class nve extends WKe{constructor(r,o,s,c){super(),this.red=r,this.green=o,this.blue=s,this.alpha=c}static parse(r){let o=[];if(WUn.test(r)&&[4,5,7,9].includes(r.length)){const c=(r.length<6?r.replace(YUn,"$&$&"):r).slice(1).split("");for(;c.length>0;)o.push(parseInt(c.splice(0,2).join(""),16));o[3]=o[3]!==void 0?o[3]/255:void 0}const s=r.match(KUn);return s?.[1]&&(o=s[1].split(",").map(c=>Number(c.trim())).map((c,d)=>W1(c,0,d<3?255:1))),o.length<3?void 0:new nve(o[0],o[1],o[2],o[3]??1)}toString(r){switch(r){case"hex":return"#"+(this.red.toString(16).padStart(2,"0")+this.green.toString(16).padStart(2,"0")+this.blue.toString(16).padStart(2,"0")).toUpperCase();case"hexa":return"#"+(this.red.toString(16).padStart(2,"0")+this.green.toString(16).padStart(2,"0")+this.blue.toString(16).padStart(2,"0")+Math.round(this.alpha*255).toString(16).padStart(2,"0")).toUpperCase();case"rgb":return`rgb(${this.red}, ${this.green}, ${this.blue})`;case"css":case"rgba":return`rgba(${this.red}, ${this.green}, ${this.blue}, ${this.alpha})`;case"hsl":return this.toHSL().toString("hsl");case"hsb":return this.toHSB().toString("hsb");default:return this.toFormat(r).toString(r)}}toFormat(r){switch(r){case"rgba":return this;case"hsba":return this.toHSB();case"hsla":return this.toHSL();default:throw new Error("Unsupported color conversion: rgb -> "+r)}}toHexInt(){return this.red<<16|this.green<<8|this.blue}toHSB(){const r=this.red/255,o=this.green/255,s=this.blue/255,c=Math.min(r,o,s),d=Math.max(r,o,s),h=d-c,b=d===0?0:h/d;let y=0;if(h!==0){switch(d){case r:y=(o-s)/h+(o<s?6:0);break;case o:y=(s-r)/h+2;break;case s:y=(r-o)/h+4;break}y/=6}return new XKe(Cd(y*360,2),Cd(b*100,2),Cd(d*100,2),Cd(this.alpha,2))}toHSL(){const r=this.red/255,o=this.green/255,s=this.blue/255,c=Math.min(r,o,s),d=Math.max(r,o,s),h=(d+c)/2,b=d-c;let y=-1,E=-1;if(b===0)y=E=0;else{switch(E=b/(h<.5?d+c:2-d-c),d){case r:y=(o-s)/b+(o<s?6:0);break;case o:y=(s-r)/b+2;break;case s:y=(r-o)/b+4;break}y/=6}return new YKe(Cd(y*360,2),Cd(E*100,2),Cd(h*100,2),Cd(this.alpha,2))}clone(){return new nve(this.red,this.green,this.blue,this.alpha)}getChannelFormatOptions(r){switch(r){case"red":case"green":case"blue":return{style:"decimal"};case"alpha":return{style:"percent"};default:throw new Error("Unknown color channel: "+r)}}formatChannelValue(r,o){let s=this.getChannelFormatOptions(r),c=this.getChannelValue(r);return new Intl.NumberFormat(o,s).format(c)}getChannelRange(r){switch(r){case"red":case"green":case"blue":return{minValue:0,maxValue:255,step:1,pageSize:17};case"alpha":return{minValue:0,maxValue:1,step:.01,pageSize:.1};default:throw new Error("Unknown color channel: "+r)}}toJSON(){return{r:this.red,g:this.green,b:this.blue,a:this.alpha}}getFormat(){return"rgba"}getChannels(){return nve.colorChannels}};VKe(AYt,"colorChannels",["red","green","blue"]);var KKe=AYt,XUn=/hsl\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%)\)|hsla\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d(.\d+)?)\)/,IYt=class rve extends WKe{constructor(r,o,s,c){super(),this.hue=r,this.saturation=o,this.lightness=s,this.alpha=c}static parse(r){let o;if(o=r.match(XUn)){const[s,c,d,h]=(o[1]??o[2]).split(",").map(b=>Number(b.trim().replace("%","")));return new rve(rYt(s,360),W1(c,0,100),W1(d,0,100),W1(h??1,0,1))}}toString(r){switch(r){case"hex":return this.toRGB().toString("hex");case"hexa":return this.toRGB().toString("hexa");case"hsl":return`hsl(${this.hue}, ${Cd(this.saturation,2)}%, ${Cd(this.lightness,2)}%)`;case"css":case"hsla":return`hsla(${this.hue}, ${Cd(this.saturation,2)}%, ${Cd(this.lightness,2)}%, ${this.alpha})`;case"hsb":return this.toHSB().toString("hsb");case"rgb":return this.toRGB().toString("rgb");default:return this.toFormat(r).toString(r)}}toFormat(r){switch(r){case"hsla":return this;case"hsba":return this.toHSB();case"rgba":return this.toRGB();default:throw new Error("Unsupported color conversion: hsl -> "+r)}}toHSB(){let r=this.saturation/100,o=this.lightness/100,s=o+r*Math.min(o,1-o);return r=s===0?0:2*(1-o/s),new XKe(Cd(this.hue,2),Cd(r*100,2),Cd(s*100,2),Cd(this.alpha,2))}toRGB(){let r=this.hue,o=this.saturation/100,s=this.lightness/100,c=o*Math.min(s,1-s),d=(h,b=(h+r/30)%12)=>s-c*Math.max(Math.min(b-3,9-b,1),-1);return new KKe(Math.round(d(0)*255),Math.round(d(8)*255),Math.round(d(4)*255),Cd(this.alpha,2))}clone(){return new rve(this.hue,this.saturation,this.lightness,this.alpha)}getChannelFormatOptions(r){switch(r){case"hue":return{style:"unit",unit:"degree",unitDisplay:"narrow"};case"saturation":case"lightness":case"alpha":return{style:"percent"};default:throw new Error("Unknown color channel: "+r)}}formatChannelValue(r,o){let s=this.getChannelFormatOptions(r),c=this.getChannelValue(r);return(r==="saturation"||r==="lightness")&&(c/=100),new Intl.NumberFormat(o,s).format(c)}getChannelRange(r){switch(r){case"hue":return{minValue:0,maxValue:360,step:1,pageSize:15};case"saturation":case"lightness":return{minValue:0,maxValue:100,step:1,pageSize:10};case"alpha":return{minValue:0,maxValue:1,step:.01,pageSize:.1};default:throw new Error("Unknown color channel: "+r)}}toJSON(){return{h:this.hue,s:this.saturation,l:this.lightness,a:this.alpha}}getFormat(){return"hsla"}getChannels(){return rve.colorChannels}};VKe(IYt,"colorChannels",["hue","saturation","lightness"]);var YKe=IYt,JUn=/hsb\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%)\)|hsba\(([-+]?\d+(?:.\d+)?\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d+(?:.\d+)?%\s*,\s*[-+]?\d(.\d+)?)\)/,OYt=class ive extends WKe{constructor(r,o,s,c){super(),this.hue=r,this.saturation=o,this.brightness=s,this.alpha=c}static parse(r){let o;if(o=r.match(JUn)){const[s,c,d,h]=(o[1]??o[2]).split(",").map(b=>Number(b.trim().replace("%","")));return new ive(rYt(s,360),W1(c,0,100),W1(d,0,100),W1(h??1,0,1))}}toString(r){switch(r){case"css":return this.toHSL().toString("css");case"hex":return this.toRGB().toString("hex");case"hexa":return this.toRGB().toString("hexa");case"hsb":return`hsb(${this.hue}, ${Cd(this.saturation,2)}%, ${Cd(this.brightness,2)}%)`;case"hsba":return`hsba(${this.hue}, ${Cd(this.saturation,2)}%, ${Cd(this.brightness,2)}%, ${this.alpha})`;case"hsl":return this.toHSL().toString("hsl");case"rgb":return this.toRGB().toString("rgb");default:return this.toFormat(r).toString(r)}}toFormat(r){switch(r){case"hsba":return this;case"hsla":return this.toHSL();case"rgba":return this.toRGB();default:throw new Error("Unsupported color conversion: hsb -> "+r)}}toHSL(){let r=this.saturation/100,o=this.brightness/100,s=o*(1-r/2);return r=s===0||s===1?0:(o-s)/Math.min(s,1-s),new YKe(Cd(this.hue,2),Cd(r*100,2),Cd(s*100,2),Cd(this.alpha,2))}toRGB(){let r=this.hue,o=this.saturation/100,s=this.brightness/100,c=(d,h=(d+r/60)%6)=>s-o*s*Math.max(Math.min(h,4-h,1),0);return new KKe(Math.round(c(5)*255),Math.round(c(3)*255),Math.round(c(1)*255),Cd(this.alpha,2))}clone(){return new ive(this.hue,this.saturation,this.brightness,this.alpha)}getChannelFormatOptions(r){switch(r){case"hue":return{style:"unit",unit:"degree",unitDisplay:"narrow"};case"saturation":case"brightness":case"alpha":return{style:"percent"};default:throw new Error("Unknown color channel: "+r)}}formatChannelValue(r,o){let s=this.getChannelFormatOptions(r),c=this.getChannelValue(r);return(r==="saturation"||r==="brightness")&&(c/=100),new Intl.NumberFormat(o,s).format(c)}getChannelRange(r){switch(r){case"hue":return{minValue:0,maxValue:360,step:1,pageSize:15};case"saturation":case"brightness":return{minValue:0,maxValue:100,step:1,pageSize:10};case"alpha":return{minValue:0,maxValue:1,step:.01,pageSize:.1};default:throw new Error("Unknown color channel: "+r)}}toJSON(){return{h:this.hue,s:this.saturation,b:this.brightness,a:this.alpha}}getFormat(){return"hsba"}getChannels(){return ive.colorChannels}};VKe(OYt,"colorChannels",["hue","saturation","brightness"]);var XKe=OYt,ZUn="aliceblue:f0f8ff,antiquewhite:faebd7,aqua:00ffff,aquamarine:7fffd4,azure:f0ffff,beige:f5f5dc,bisque:ffe4c4,black:000000,blanchedalmond:ffebcd,blue:0000ff,blueviolet:8a2be2,brown:a52a2a,burlywood:deb887,cadetblue:5f9ea0,chartreuse:7fff00,chocolate:d2691e,coral:ff7f50,cornflowerblue:6495ed,cornsilk:fff8dc,crimson:dc143c,cyan:00ffff,darkblue:00008b,darkcyan:008b8b,darkgoldenrod:b8860b,darkgray:a9a9a9,darkgreen:006400,darkkhaki:bdb76b,darkmagenta:8b008b,darkolivegreen:556b2f,darkorange:ff8c00,darkorchid:9932cc,darkred:8b0000,darksalmon:e9967a,darkseagreen:8fbc8f,darkslateblue:483d8b,darkslategray:2f4f4f,darkturquoise:00ced1,darkviolet:9400d3,deeppink:ff1493,deepskyblue:00bfff,dimgray:696969,dodgerblue:1e90ff,firebrick:b22222,floralwhite:fffaf0,forestgreen:228b22,fuchsia:ff00ff,gainsboro:dcdcdc,ghostwhite:f8f8ff,gold:ffd700,goldenrod:daa520,gray:808080,green:008000,greenyellow:adff2f,honeydew:f0fff0,hotpink:ff69b4,indianred:cd5c5c,indigo:4b0082,ivory:fffff0,khaki:f0e68c,lavender:e6e6fa,lavenderblush:fff0f5,lawngreen:7cfc00,lemonchiffon:fffacd,lightblue:add8e6,lightcoral:f08080,lightcyan:e0ffff,lightgoldenrodyellow:fafad2,lightgrey:d3d3d3,lightgreen:90ee90,lightpink:ffb6c1,lightsalmon:ffa07a,lightseagreen:20b2aa,lightskyblue:87cefa,lightslategray:778899,lightsteelblue:b0c4de,lightyellow:ffffe0,lime:00ff00,limegreen:32cd32,linen:faf0e6,magenta:ff00ff,maroon:800000,mediumaquamarine:66cdaa,mediumblue:0000cd,mediumorchid:ba55d3,mediumpurple:9370d8,mediumseagreen:3cb371,mediumslateblue:7b68ee,mediumspringgreen:00fa9a,mediumturquoise:48d1cc,mediumvioletred:c71585,midnightblue:191970,mintcream:f5fffa,mistyrose:ffe4e1,moccasin:ffe4b5,navajowhite:ffdead,navy:000080,oldlace:fdf5e6,olive:808000,olivedrab:6b8e23,orange:ffa500,orangered:ff4500,orchid:da70d6,palegoldenrod:eee8aa,palegreen:98fb98,paleturquoise:afeeee,palevioletred:d87093,papayawhip:ffefd5,peachpuff:ffdab9,peru:cd853f,pink:ffc0cb,plum:dda0dd,powderblue:b0e0e6,purple:800080,rebeccapurple:663399,red:ff0000,rosybrown:bc8f8f,royalblue:4169e1,saddlebrown:8b4513,salmon:fa8072,sandybrown:f4a460,seagreen:2e8b57,seashell:fff5ee,sienna:a0522d,silver:c0c0c0,skyblue:87ceeb,slateblue:6a5acd,slategray:708090,snow:fffafa,springgreen:00ff7f,steelblue:4682b4,tan:d2b48c,teal:008080,thistle:d8bfd8,tomato:ff6347,turquoise:40e0d0,violet:ee82ee,wheat:f5deb3,white:ffffff,whitesmoke:f5f5f5,yellow:ffff00,yellowgreen:9acd32",QUn=n=>{const r=new Map,o=n.split(",");for(let s=0;s<o.length;s++){const[c,d]=o[s].split(":");r.set(c,`#${d}`),c.includes("gray")&&r.set(c.replace("gray","grey"),`#${d}`)}return r},Y8t=QUn(ZUn),Bve=n=>{if(Y8t.has(n))return Bve(Y8t.get(n));const r=KKe.parse(n)||XKe.parse(n)||YKe.parse(n);if(!r){const o=new Error("Invalid color value: "+n);throw Error.captureStackTrace?.(o,Bve),o}return r};const ezn=["top","right","bottom","left"],kM=Math.min,i2=Math.max,Uve=Math.round,L1e=Math.floor,SI=n=>({x:n,y:n}),tzn={left:"right",right:"left",bottom:"top",top:"bottom"},nzn={start:"end",end:"start"};function Uqe(n,r,o){return i2(n,kM(r,o))}function F4(n,r){return typeof n=="function"?n(r):n}function B4(n){return n.split("-")[0]}function wK(n){return n.split("-")[1]}function JKe(n){return n==="x"?"y":"x"}function ZKe(n){return n==="y"?"height":"width"}const rzn=new Set(["top","bottom"]);function mI(n){return rzn.has(B4(n))?"y":"x"}function QKe(n){return JKe(mI(n))}function izn(n,r,o){o===void 0&&(o=!1);const s=wK(n),c=QKe(n),d=ZKe(c);let h=c==="x"?s===(o?"end":"start")?"right":"left":s==="start"?"bottom":"top";return r.reference[d]>r.floating[d]&&(h=zve(h)),[h,zve(h)]}function ozn(n){const r=zve(n);return[zqe(n),r,zqe(r)]}function zqe(n){return n.replace(/start|end/g,r=>nzn[r])}const X8t=["left","right"],J8t=["right","left"],azn=["top","bottom"],szn=["bottom","top"];function czn(n,r,o){switch(n){case"top":case"bottom":return o?r?J8t:X8t:r?X8t:J8t;case"left":case"right":return r?azn:szn;default:return[]}}function uzn(n,r,o,s){const c=wK(n);let d=czn(B4(n),o==="start",s);return c&&(d=d.map(h=>h+"-"+c),r&&(d=d.concat(d.map(zqe)))),d}function zve(n){return n.replace(/left|right|bottom|top/g,r=>tzn[r])}function lzn(n){return{top:0,right:0,bottom:0,left:0,...n}}function NYt(n){return typeof n!="number"?lzn(n):{top:n,right:n,bottom:n,left:n}}function Hve(n){const{x:r,y:o,width:s,height:c}=n;return{width:s,height:c,top:o,left:r,right:r+s,bottom:o+c,x:r,y:o}}function Z8t(n,r,o){let{reference:s,floating:c}=n;const d=mI(r),h=QKe(r),b=ZKe(h),y=B4(r),E=d==="y",k=s.x+s.width/2-c.width/2,x=s.y+s.height/2-c.height/2,O=s[b]/2-c[b]/2;let P;switch(y){case"top":P={x:k,y:s.y-c.height};break;case"bottom":P={x:k,y:s.y+s.height};break;case"right":P={x:s.x+s.width,y:x};break;case"left":P={x:s.x-c.width,y:x};break;default:P={x:s.x,y:s.y}}switch(wK(r)){case"start":P[h]-=O*(o&&E?-1:1);break;case"end":P[h]+=O*(o&&E?-1:1);break}return P}const dzn=async(n,r,o)=>{const{placement:s="bottom",strategy:c="absolute",middleware:d=[],platform:h}=o,b=d.filter(Boolean),y=await(h.isRTL==null?void 0:h.isRTL(r));let E=await h.getElementRects({reference:n,floating:r,strategy:c}),{x:k,y:x}=Z8t(E,s,y),O=s,P={},D=0;for(let $=0;$<b.length;$++){const{name:U,fn:B}=b[$],{x:G,y:K,data:X,reset:ae}=await B({x:k,y:x,initialPlacement:s,placement:O,strategy:c,middlewareData:P,rects:E,platform:h,elements:{reference:n,floating:r}});k=G??k,x=K??x,P={...P,[U]:{...P[U],...X}},ae&&D<=50&&(D++,typeof ae=="object"&&(ae.placement&&(O=ae.placement),ae.rects&&(E=ae.rects===!0?await h.getElementRects({reference:n,floating:r,strategy:c}):ae.rects),{x:k,y:x}=Z8t(E,O,y)),$=-1)}return{x:k,y:x,placement:O,strategy:c,middlewareData:P}};async function Die(n,r){var o;r===void 0&&(r={});const{x:s,y:c,platform:d,rects:h,elements:b,strategy:y}=n,{boundary:E="clippingAncestors",rootBoundary:k="viewport",elementContext:x="floating",altBoundary:O=!1,padding:P=0}=F4(r,n),D=NYt(P),U=b[O?x==="floating"?"reference":"floating":x],B=Hve(await d.getClippingRect({element:(o=await(d.isElement==null?void 0:d.isElement(U)))==null||o?U:U.contextElement||await(d.getDocumentElement==null?void 0:d.getDocumentElement(b.floating)),boundary:E,rootBoundary:k,strategy:y})),G=x==="floating"?{x:s,y:c,width:h.floating.width,height:h.floating.height}:h.reference,K=await(d.getOffsetParent==null?void 0:d.getOffsetParent(b.floating)),X=await(d.isElement==null?void 0:d.isElement(K))?await(d.getScale==null?void 0:d.getScale(K))||{x:1,y:1}:{x:1,y:1},ae=Hve(d.convertOffsetParentRelativeRectToViewportRelativeRect?await d.convertOffsetParentRelativeRectToViewportRelativeRect({elements:b,rect:G,offsetParent:K,strategy:y}):G);return{top:(B.top-ae.top+D.top)/X.y,bottom:(ae.bottom-B.bottom+D.bottom)/X.y,left:(B.left-ae.left+D.left)/X.x,right:(ae.right-B.right+D.right)/X.x}}const fzn=n=>({name:"arrow",options:n,async fn(r){const{x:o,y:s,placement:c,rects:d,platform:h,elements:b,middlewareData:y}=r,{element:E,padding:k=0}=F4(n,r)||{};if(E==null)return{};const x=NYt(k),O={x:o,y:s},P=QKe(c),D=ZKe(P),$=await h.getDimensions(E),U=P==="y",B=U?"top":"left",G=U?"bottom":"right",K=U?"clientHeight":"clientWidth",X=d.reference[D]+d.reference[P]-O[P]-d.floating[D],ae=O[P]-d.reference[P],oe=await(h.getOffsetParent==null?void 0:h.getOffsetParent(E));let he=oe?oe[K]:0;(!he||!await(h.isElement==null?void 0:h.isElement(oe)))&&(he=b.floating[K]||d.floating[D]);const ye=X/2-ae/2,we=he/2-$[D]/2-1,ge=kM(x[B],we),Te=kM(x[G],we),xe=ge,Ue=he-$[D]-Te,Z=he/2-$[D]/2+ye,Re=Uqe(xe,Z,Ue),_e=!y.arrow&&wK(c)!=null&&Z!==Re&&d.reference[D]/2-(Z<xe?ge:Te)-$[D]/2<0,qe=_e?Z<xe?Z-xe:Z-Ue:0;return{[P]:O[P]+qe,data:{[P]:Re,centerOffset:Z-Re-qe,..._e&&{alignmentOffset:qe}},reset:_e}}}),hzn=function(n){return n===void 0&&(n={}),{name:"flip",options:n,async fn(r){var o,s;const{placement:c,middlewareData:d,rects:h,initialPlacement:b,platform:y,elements:E}=r,{mainAxis:k=!0,crossAxis:x=!0,fallbackPlacements:O,fallbackStrategy:P="bestFit",fallbackAxisSideDirection:D="none",flipAlignment:$=!0,...U}=F4(n,r);if((o=d.arrow)!=null&&o.alignmentOffset)return{};const B=B4(c),G=mI(b),K=B4(b)===b,X=await(y.isRTL==null?void 0:y.isRTL(E.floating)),ae=O||(K||!$?[zve(b)]:ozn(b)),oe=D!=="none";!O&&oe&&ae.push(...uzn(b,$,D,X));const he=[b,...ae],ye=await Die(r,U),we=[];let ge=((s=d.flip)==null?void 0:s.overflows)||[];if(k&&we.push(ye[B]),x){const Z=izn(c,h,X);we.push(ye[Z[0]],ye[Z[1]])}if(ge=[...ge,{placement:c,overflows:we}],!we.every(Z=>Z<=0)){var Te,xe;const Z=(((Te=d.flip)==null?void 0:Te.index)||0)+1,Re=he[Z];if(Re&&(!(x==="alignment"?G!==mI(Re):!1)||ge.every(it=>mI(it.placement)===G?it.overflows[0]>0:!0)))return{data:{index:Z,overflows:ge},reset:{placement:Re}};let _e=(xe=ge.filter(qe=>qe.overflows[0]<=0).sort((qe,it)=>qe.overflows[1]-it.overflows[1])[0])==null?void 0:xe.placement;if(!_e)switch(P){case"bestFit":{var Ue;const qe=(Ue=ge.filter(it=>{if(oe){const Xe=mI(it.placement);return Xe===G||Xe==="y"}return!0}).map(it=>[it.placement,it.overflows.filter(Xe=>Xe>0).reduce((Xe,de)=>Xe+de,0)]).sort((it,Xe)=>it[1]-Xe[1])[0])==null?void 0:Ue[0];qe&&(_e=qe);break}case"initialPlacement":_e=b;break}if(c!==_e)return{reset:{placement:_e}}}return{}}}};function Q8t(n,r){return{top:n.top-r.height,right:n.right-r.width,bottom:n.bottom-r.height,left:n.left-r.width}}function e9t(n){return ezn.some(r=>n[r]>=0)}const pzn=function(n){return n===void 0&&(n={}),{name:"hide",options:n,async fn(r){const{rects:o}=r,{strategy:s="referenceHidden",...c}=F4(n,r);switch(s){case"referenceHidden":{const d=await Die(r,{...c,elementContext:"reference"}),h=Q8t(d,o.reference);return{data:{referenceHiddenOffsets:h,referenceHidden:e9t(h)}}}case"escaped":{const d=await Die(r,{...c,altBoundary:!0}),h=Q8t(d,o.floating);return{data:{escapedOffsets:h,escaped:e9t(h)}}}default:return{}}}}},RYt=new Set(["left","top"]);async function gzn(n,r){const{placement:o,platform:s,elements:c}=n,d=await(s.isRTL==null?void 0:s.isRTL(c.floating)),h=B4(o),b=wK(o),y=mI(o)==="y",E=RYt.has(h)?-1:1,k=d&&y?-1:1,x=F4(r,n);let{mainAxis:O,crossAxis:P,alignmentAxis:D}=typeof x=="number"?{mainAxis:x,crossAxis:0,alignmentAxis:null}:{mainAxis:x.mainAxis||0,crossAxis:x.crossAxis||0,alignmentAxis:x.alignmentAxis};return b&&typeof D=="number"&&(P=b==="end"?D*-1:D),y?{x:P*k,y:O*E}:{x:O*E,y:P*k}}const mzn=function(n){return n===void 0&&(n=0),{name:"offset",options:n,async fn(r){var o,s;const{x:c,y:d,placement:h,middlewareData:b}=r,y=await gzn(r,n);return h===((o=b.offset)==null?void 0:o.placement)&&(s=b.arrow)!=null&&s.alignmentOffset?{}:{x:c+y.x,y:d+y.y,data:{...y,placement:h}}}}},bzn=function(n){return n===void 0&&(n={}),{name:"shift",options:n,async fn(r){const{x:o,y:s,placement:c}=r,{mainAxis:d=!0,crossAxis:h=!1,limiter:b={fn:U=>{let{x:B,y:G}=U;return{x:B,y:G}}},...y}=F4(n,r),E={x:o,y:s},k=await Die(r,y),x=mI(B4(c)),O=JKe(x);let P=E[O],D=E[x];if(d){const U=O==="y"?"top":"left",B=O==="y"?"bottom":"right",G=P+k[U],K=P-k[B];P=Uqe(G,P,K)}if(h){const U=x==="y"?"top":"left",B=x==="y"?"bottom":"right",G=D+k[U],K=D-k[B];D=Uqe(G,D,K)}const $=b.fn({...r,[O]:P,[x]:D});return{...$,data:{x:$.x-o,y:$.y-s,enabled:{[O]:d,[x]:h}}}}}},vzn=function(n){return n===void 0&&(n={}),{options:n,fn(r){const{x:o,y:s,placement:c,rects:d,middlewareData:h}=r,{offset:b=0,mainAxis:y=!0,crossAxis:E=!0}=F4(n,r),k={x:o,y:s},x=mI(c),O=JKe(x);let P=k[O],D=k[x];const $=F4(b,r),U=typeof $=="number"?{mainAxis:$,crossAxis:0}:{mainAxis:0,crossAxis:0,...$};if(y){const K=O==="y"?"height":"width",X=d.reference[O]-d.floating[K]+U.mainAxis,ae=d.reference[O]+d.reference[K]-U.mainAxis;P<X?P=X:P>ae&&(P=ae)}if(E){var B,G;const K=O==="y"?"width":"height",X=RYt.has(B4(c)),ae=d.reference[x]-d.floating[K]+(X&&((B=h.offset)==null?void 0:B[x])||0)+(X?0:U.crossAxis),oe=d.reference[x]+d.reference[K]+(X?0:((G=h.offset)==null?void 0:G[x])||0)-(X?U.crossAxis:0);D<ae?D=ae:D>oe&&(D=oe)}return{[O]:P,[x]:D}}}},wzn=function(n){return n===void 0&&(n={}),{name:"size",options:n,async fn(r){var o,s;const{placement:c,rects:d,platform:h,elements:b}=r,{apply:y=()=>{},...E}=F4(n,r),k=await Die(r,E),x=B4(c),O=wK(c),P=mI(c)==="y",{width:D,height:$}=d.floating;let U,B;x==="top"||x==="bottom"?(U=x,B=O===(await(h.isRTL==null?void 0:h.isRTL(b.floating))?"start":"end")?"left":"right"):(B=x,U=O==="end"?"top":"bottom");const G=$-k.top-k.bottom,K=D-k.left-k.right,X=kM($-k[U],G),ae=kM(D-k[B],K),oe=!r.middlewareData.shift;let he=X,ye=ae;if((o=r.middlewareData.shift)!=null&&o.enabled.x&&(ye=K),(s=r.middlewareData.shift)!=null&&s.enabled.y&&(he=G),oe&&!O){const ge=i2(k.left,0),Te=i2(k.right,0),xe=i2(k.top,0),Ue=i2(k.bottom,0);P?ye=D-2*(ge!==0||Te!==0?ge+Te:i2(k.left,k.right)):he=$-2*(xe!==0||Ue!==0?xe+Ue:i2(k.top,k.bottom))}await y({...r,availableWidth:ye,availableHeight:he});const we=await h.getDimensions(b.floating);return D!==we.width||$!==we.height?{reset:{rects:!0}}:{}}}};function fye(){return typeof window<"u"}function yK(n){return PYt(n)?(n.nodeName||"").toLowerCase():"#document"}function h2(n){var r;return(n==null||(r=n.ownerDocument)==null?void 0:r.defaultView)||window}function FI(n){var r;return(r=(PYt(n)?n.ownerDocument:n.document)||window.document)==null?void 0:r.documentElement}function PYt(n){return fye()?n instanceof Node||n instanceof h2(n).Node:!1}function Hx(n){return fye()?n instanceof Element||n instanceof h2(n).Element:!1}function PI(n){return fye()?n instanceof HTMLElement||n instanceof h2(n).HTMLElement:!1}function t9t(n){return!fye()||typeof ShadowRoot>"u"?!1:n instanceof ShadowRoot||n instanceof h2(n).ShadowRoot}const yzn=new Set(["inline","contents"]);function Toe(n){const{overflow:r,overflowX:o,overflowY:s,display:c}=Gx(n);return/auto|scroll|overlay|hidden|clip/.test(r+s+o)&&!yzn.has(c)}const Ezn=new Set(["table","td","th"]);function Szn(n){return Ezn.has(yK(n))}const _zn=[":popover-open",":modal"];function hye(n){return _zn.some(r=>{try{return n.matches(r)}catch{return!1}})}const kzn=["transform","translate","scale","rotate","perspective"],Tzn=["transform","translate","scale","rotate","perspective","filter"],xzn=["paint","layout","strict","content"];function eYe(n){const r=tYe(),o=Hx(n)?Gx(n):n;return kzn.some(s=>o[s]?o[s]!=="none":!1)||(o.containerType?o.containerType!=="normal":!1)||!r&&(o.backdropFilter?o.backdropFilter!=="none":!1)||!r&&(o.filter?o.filter!=="none":!1)||Tzn.some(s=>(o.willChange||"").includes(s))||xzn.some(s=>(o.contain||"").includes(s))}function Czn(n){let r=TM(n);for(;PI(r)&&!KW(r);){if(eYe(r))return r;if(hye(r))return null;r=TM(r)}return null}function tYe(){return typeof CSS>"u"||!CSS.supports?!1:CSS.supports("-webkit-backdrop-filter","none")}const Azn=new Set(["html","body","#document"]);function KW(n){return Azn.has(yK(n))}function Gx(n){return h2(n).getComputedStyle(n)}function pye(n){return Hx(n)?{scrollLeft:n.scrollLeft,scrollTop:n.scrollTop}:{scrollLeft:n.scrollX,scrollTop:n.scrollY}}function TM(n){if(yK(n)==="html")return n;const r=n.assignedSlot||n.parentNode||t9t(n)&&n.host||FI(n);return t9t(r)?r.host:r}function LYt(n){const r=TM(n);return KW(r)?n.ownerDocument?n.ownerDocument.body:n.body:PI(r)&&Toe(r)?r:LYt(r)}function Mie(n,r,o){var s;r===void 0&&(r=[]),o===void 0&&(o=!0);const c=LYt(n),d=c===((s=n.ownerDocument)==null?void 0:s.body),h=h2(c);if(d){const b=Hqe(h);return r.concat(h,h.visualViewport||[],Toe(c)?c:[],b&&o?Mie(b):[])}return r.concat(c,Mie(c,[],o))}function Hqe(n){return n.parent&&Object.getPrototypeOf(n.parent)?n.frameElement:null}function DYt(n){const r=Gx(n);let o=parseFloat(r.width)||0,s=parseFloat(r.height)||0;const c=PI(n),d=c?n.offsetWidth:o,h=c?n.offsetHeight:s,b=Uve(o)!==d||Uve(s)!==h;return b&&(o=d,s=h),{width:o,height:s,$:b}}function nYe(n){return Hx(n)?n:n.contextElement}function SW(n){const r=nYe(n);if(!PI(r))return SI(1);const o=r.getBoundingClientRect(),{width:s,height:c,$:d}=DYt(r);let h=(d?Uve(o.width):o.width)/s,b=(d?Uve(o.height):o.height)/c;return(!h||!Number.isFinite(h))&&(h=1),(!b||!Number.isFinite(b))&&(b=1),{x:h,y:b}}const Izn=SI(0);function MYt(n){const r=h2(n);return!tYe()||!r.visualViewport?Izn:{x:r.visualViewport.offsetLeft,y:r.visualViewport.offsetTop}}function Ozn(n,r,o){return r===void 0&&(r=!1),!o||r&&o!==h2(n)?!1:r}function VB(n,r,o,s){r===void 0&&(r=!1),o===void 0&&(o=!1);const c=n.getBoundingClientRect(),d=nYe(n);let h=SI(1);r&&(s?Hx(s)&&(h=SW(s)):h=SW(n));const b=Ozn(d,o,s)?MYt(d):SI(0);let y=(c.left+b.x)/h.x,E=(c.top+b.y)/h.y,k=c.width/h.x,x=c.height/h.y;if(d){const O=h2(d),P=s&&Hx(s)?h2(s):s;let D=O,$=Hqe(D);for(;$&&s&&P!==D;){const U=SW($),B=$.getBoundingClientRect(),G=Gx($),K=B.left+($.clientLeft+parseFloat(G.paddingLeft))*U.x,X=B.top+($.clientTop+parseFloat(G.paddingTop))*U.y;y*=U.x,E*=U.y,k*=U.x,x*=U.y,y+=K,E+=X,D=h2($),$=Hqe(D)}}return Hve({width:k,height:x,x:y,y:E})}function gye(n,r){const o=pye(n).scrollLeft;return r?r.left+o:VB(FI(n)).left+o}function jYt(n,r){const o=n.getBoundingClientRect(),s=o.left+r.scrollLeft-gye(n,o),c=o.top+r.scrollTop;return{x:s,y:c}}function Nzn(n){let{elements:r,rect:o,offsetParent:s,strategy:c}=n;const d=c==="fixed",h=FI(s),b=r?hye(r.floating):!1;if(s===h||b&&d)return o;let y={scrollLeft:0,scrollTop:0},E=SI(1);const k=SI(0),x=PI(s);if((x||!x&&!d)&&((yK(s)!=="body"||Toe(h))&&(y=pye(s)),PI(s))){const P=VB(s);E=SW(s),k.x=P.x+s.clientLeft,k.y=P.y+s.clientTop}const O=h&&!x&&!d?jYt(h,y):SI(0);return{width:o.width*E.x,height:o.height*E.y,x:o.x*E.x-y.scrollLeft*E.x+k.x+O.x,y:o.y*E.y-y.scrollTop*E.y+k.y+O.y}}function Rzn(n){return Array.from(n.getClientRects())}function Pzn(n){const r=FI(n),o=pye(n),s=n.ownerDocument.body,c=i2(r.scrollWidth,r.clientWidth,s.scrollWidth,s.clientWidth),d=i2(r.scrollHeight,r.clientHeight,s.scrollHeight,s.clientHeight);let h=-o.scrollLeft+gye(n);const b=-o.scrollTop;return Gx(s).direction==="rtl"&&(h+=i2(r.clientWidth,s.clientWidth)-c),{width:c,height:d,x:h,y:b}}const n9t=25;function Lzn(n,r){const o=h2(n),s=FI(n),c=o.visualViewport;let d=s.clientWidth,h=s.clientHeight,b=0,y=0;if(c){d=c.width,h=c.height;const k=tYe();(!k||k&&r==="fixed")&&(b=c.offsetLeft,y=c.offsetTop)}const E=gye(s);if(E<=0){const k=s.ownerDocument,x=k.body,O=getComputedStyle(x),P=k.compatMode==="CSS1Compat"&&parseFloat(O.marginLeft)+parseFloat(O.marginRight)||0,D=Math.abs(s.clientWidth-x.clientWidth-P);D<=n9t&&(d-=D)}else E<=n9t&&(d+=E);return{width:d,height:h,x:b,y}}const Dzn=new Set(["absolute","fixed"]);function Mzn(n,r){const o=VB(n,!0,r==="fixed"),s=o.top+n.clientTop,c=o.left+n.clientLeft,d=PI(n)?SW(n):SI(1),h=n.clientWidth*d.x,b=n.clientHeight*d.y,y=c*d.x,E=s*d.y;return{width:h,height:b,x:y,y:E}}function r9t(n,r,o){let s;if(r==="viewport")s=Lzn(n,o);else if(r==="document")s=Pzn(FI(n));else if(Hx(r))s=Mzn(r,o);else{const c=MYt(n);s={x:r.x-c.x,y:r.y-c.y,width:r.width,height:r.height}}return Hve(s)}function $Yt(n,r){const o=TM(n);return o===r||!Hx(o)||KW(o)?!1:Gx(o).position==="fixed"||$Yt(o,r)}function jzn(n,r){const o=r.get(n);if(o)return o;let s=Mie(n,[],!1).filter(b=>Hx(b)&&yK(b)!=="body"),c=null;const d=Gx(n).position==="fixed";let h=d?TM(n):n;for(;Hx(h)&&!KW(h);){const b=Gx(h),y=eYe(h);!y&&b.position==="fixed"&&(c=null),(d?!y&&!c:!y&&b.position==="static"&&!!c&&Dzn.has(c.position)||Toe(h)&&!y&&$Yt(n,h))?s=s.filter(k=>k!==h):c=b,h=TM(h)}return r.set(n,s),s}function $zn(n){let{element:r,boundary:o,rootBoundary:s,strategy:c}=n;const h=[...o==="clippingAncestors"?hye(r)?[]:jzn(r,this._c):[].concat(o),s],b=h[0],y=h.reduce((E,k)=>{const x=r9t(r,k,c);return E.top=i2(x.top,E.top),E.right=kM(x.right,E.right),E.bottom=kM(x.bottom,E.bottom),E.left=i2(x.left,E.left),E},r9t(r,b,c));return{width:y.right-y.left,height:y.bottom-y.top,x:y.left,y:y.top}}function Fzn(n){const{width:r,height:o}=DYt(n);return{width:r,height:o}}function Bzn(n,r,o){const s=PI(r),c=FI(r),d=o==="fixed",h=VB(n,!0,d,r);let b={scrollLeft:0,scrollTop:0};const y=SI(0);function E(){y.x=gye(c)}if(s||!s&&!d)if((yK(r)!=="body"||Toe(c))&&(b=pye(r)),s){const P=VB(r,!0,d,r);y.x=P.x+r.clientLeft,y.y=P.y+r.clientTop}else c&&E();d&&!s&&c&&E();const k=c&&!s&&!d?jYt(c,b):SI(0),x=h.left+b.scrollLeft-y.x-k.x,O=h.top+b.scrollTop-y.y-k.y;return{x,y:O,width:h.width,height:h.height}}function $je(n){return Gx(n).position==="static"}function i9t(n,r){if(!PI(n)||Gx(n).position==="fixed")return null;if(r)return r(n);let o=n.offsetParent;return FI(n)===o&&(o=o.ownerDocument.body),o}function FYt(n,r){const o=h2(n);if(hye(n))return o;if(!PI(n)){let c=TM(n);for(;c&&!KW(c);){if(Hx(c)&&!$je(c))return c;c=TM(c)}return o}let s=i9t(n,r);for(;s&&Szn(s)&&$je(s);)s=i9t(s,r);return s&&KW(s)&&$je(s)&&!eYe(s)?o:s||Czn(n)||o}const Uzn=async function(n){const r=this.getOffsetParent||FYt,o=this.getDimensions,s=await o(n.floating);return{reference:Bzn(n.reference,await r(n.floating),n.strategy),floating:{x:0,y:0,width:s.width,height:s.height}}};function zzn(n){return Gx(n).direction==="rtl"}const Hzn={convertOffsetParentRelativeRectToViewportRelativeRect:Nzn,getDocumentElement:FI,getClippingRect:$zn,getOffsetParent:FYt,getElementRects:Uzn,getClientRects:Rzn,getDimensions:Fzn,getScale:SW,isElement:Hx,isRTL:zzn};function BYt(n,r){return n.x===r.x&&n.y===r.y&&n.width===r.width&&n.height===r.height}function Gzn(n,r){let o=null,s;const c=FI(n);function d(){var b;clearTimeout(s),(b=o)==null||b.disconnect(),o=null}function h(b,y){b===void 0&&(b=!1),y===void 0&&(y=1),d();const E=n.getBoundingClientRect(),{left:k,top:x,width:O,height:P}=E;if(b||r(),!O||!P)return;const D=L1e(x),$=L1e(c.clientWidth-(k+O)),U=L1e(c.clientHeight-(x+P)),B=L1e(k),K={rootMargin:-D+"px "+-$+"px "+-U+"px "+-B+"px",threshold:i2(0,kM(1,y))||1};let X=!0;function ae(oe){const he=oe[0].intersectionRatio;if(he!==y){if(!X)return h();he?h(!1,he):s=setTimeout(()=>{h(!1,1e-7)},1e3)}he===1&&!BYt(E,n.getBoundingClientRect())&&h(),X=!1}try{o=new IntersectionObserver(ae,{...K,root:c.ownerDocument})}catch{o=new IntersectionObserver(ae,K)}o.observe(n)}return h(!0),d}function qzn(n,r,o,s){s===void 0&&(s={});const{ancestorScroll:c=!0,ancestorResize:d=!0,elementResize:h=typeof ResizeObserver=="function",layoutShift:b=typeof IntersectionObserver=="function",animationFrame:y=!1}=s,E=nYe(n),k=c||d?[...E?Mie(E):[],...Mie(r)]:[];k.forEach(B=>{c&&B.addEventListener("scroll",o,{passive:!0}),d&&B.addEventListener("resize",o)});const x=E&&b?Gzn(E,o):null;let O=-1,P=null;h&&(P=new ResizeObserver(B=>{let[G]=B;G&&G.target===E&&P&&(P.unobserve(r),cancelAnimationFrame(O),O=requestAnimationFrame(()=>{var K;(K=P)==null||K.observe(r)})),o()}),E&&!y&&P.observe(E),P.observe(r));let D,$=y?VB(n):null;y&&U();function U(){const B=VB(n);$&&!BYt($,B)&&o(),$=B,D=requestAnimationFrame(U)}return o(),()=>{var B;k.forEach(G=>{c&&G.removeEventListener("scroll",o),d&&G.removeEventListener("resize",o)}),x?.(),(B=P)==null||B.disconnect(),P=null,y&&cancelAnimationFrame(D)}}const Vzn=mzn,Wzn=bzn,Kzn=hzn,Yzn=wzn,Xzn=pzn,Jzn=fzn,Zzn=vzn,Qzn=(n,r,o)=>{const s=new Map,c={platform:Hzn,...o},d={...c.platform,_c:s};return dzn(n,r,{...c,platform:d})};function o9t(n=0,r=0,o=0,s=0){if(typeof DOMRect=="function")return new DOMRect(n,r,o,s);const c={x:n,y:r,width:o,height:s,top:r,right:n+o,bottom:r+s,left:n};return{...c,toJSON:()=>c}}function eHn(n){if(!n)return o9t();const{x:r,y:o,width:s,height:c}=n;return o9t(r,o,s,c)}function tHn(n,r){return{contextElement:Gb(n)?n:void 0,getBoundingClientRect:()=>{const o=n,s=r?.(o);return s||!o?eHn(s):o.getBoundingClientRect()}}}var ere=n=>({variable:n,reference:`var(${n})`}),T4={arrowSize:ere("--arrow-size"),arrowSizeHalf:ere("--arrow-size-half"),arrowBg:ere("--arrow-background"),transformOrigin:ere("--transform-origin"),arrowOffset:ere("--arrow-offset")},nHn=n=>n==="top"||n==="bottom"?"y":"x";function rHn(n,r){return{name:"transformOrigin",fn(o){const{elements:s,middlewareData:c,placement:d,rects:h,y:b}=o,y=d.split("-")[0],E=nHn(y),k=c.arrow?.x||0,x=c.arrow?.y||0,O=r?.clientWidth||0,P=r?.clientHeight||0,D=k+O/2,$=x+P/2,U=Math.abs(c.shift?.y||0),B=h.reference.height/2,G=P/2,K=n.offset?.mainAxis??n.gutter,X=typeof K=="number"?K+G:K??G,ae=U>X,oe={top:`${D}px calc(100% + ${X}px)`,bottom:`${D}px ${-X}px`,left:`calc(100% + ${X}px) ${$}px`,right:`${-X}px ${$}px`}[y],he=`${D}px ${h.reference.y+B-b}px`,ye=!!n.overlap&&E==="y"&&ae;return s.floating.style.setProperty(T4.transformOrigin.variable,ye?he:oe),{data:{transformOrigin:ye?he:oe}}}}}var iHn={name:"rects",fn({rects:n}){return{data:n}}},oHn=n=>{if(n)return{name:"shiftArrow",fn({placement:r,middlewareData:o}){if(!o.arrow)return{};const{x:s,y:c}=o.arrow,d=r.split("-")[0];return Object.assign(n.style,{left:s!=null?`${s}px`:"",top:c!=null?`${c}px`:"",[d]:`calc(100% + ${T4.arrowOffset.reference})`}),{}}}};function aHn(n){const[r,o]=n.split("-");return{side:r,align:o,hasAlign:o!=null}}function sHn(n){return n.split("-")[0]}var cHn={strategy:"absolute",placement:"bottom",listeners:!0,gutter:8,flip:!0,slide:!0,overlap:!1,sameWidth:!1,fitViewport:!1,overflowPadding:8,arrowPadding:4};function a9t(n,r){const o=n.devicePixelRatio||1;return Math.round(r*o)/o}function rYe(n){return typeof n=="function"?n():n==="clipping-ancestors"?"clippingAncestors":n}function uHn(n,r,o){const s=n||r.createElement("div");return Jzn({element:s,padding:o.arrowPadding})}function lHn(n,r){if(!jFn(r.offset??r.gutter))return Vzn(({placement:o})=>{const s=(n?.clientHeight||0)/2,c=r.offset?.mainAxis??r.gutter,d=typeof c=="number"?c+s:c??s,{hasAlign:h}=aHn(o),b=h?void 0:r.shift,y=r.offset?.crossAxis??b;return bK({crossAxis:y,mainAxis:d,alignmentAxis:r.shift})})}function dHn(n){if(!n.flip)return;const r=rYe(n.boundary);return Kzn({...r?{boundary:r}:void 0,padding:n.overflowPadding,fallbackPlacements:n.flip===!0?void 0:n.flip})}function fHn(n){if(!n.slide&&!n.overlap)return;const r=rYe(n.boundary);return Wzn({...r?{boundary:r}:void 0,mainAxis:n.slide,crossAxis:n.overlap,padding:n.overflowPadding,limiter:Zzn()})}function hHn(n){return Yzn({padding:n.overflowPadding,apply({elements:r,rects:o,availableHeight:s,availableWidth:c}){const d=r.floating,h=Math.round(o.reference.width),b=Math.round(o.reference.height);c=Math.floor(c),s=Math.floor(s),d.style.setProperty("--reference-width",`${h}px`),d.style.setProperty("--reference-height",`${b}px`),d.style.setProperty("--available-width",`${c}px`),d.style.setProperty("--available-height",`${s}px`)}})}function pHn(n){if(n.hideWhenDetached)return Xzn({strategy:"referenceHidden",boundary:rYe(n.boundary)??"clippingAncestors"})}function gHn(n){return n?n===!0?{ancestorResize:!0,ancestorScroll:!0,elementResize:!0,layoutShift:!0}:n:{}}function mHn(n,r,o={}){const s=tHn(n,o.getAnchorRect);if(!r||!s)return;const c=Object.assign({},cHn,o),d=r.querySelector("[data-part=arrow]"),h=[lHn(d,c),dHn(c),fHn(c),uHn(d,r.ownerDocument,c),oHn(d),rHn({gutter:c.gutter,offset:c.offset,overlap:c.overlap},d),hHn(c),pHn(c),iHn],{placement:b,strategy:y,onComplete:E,onPositioned:k}=c,x=async()=>{if(!s||!r)return;const $=await Qzn(s,r,{placement:b,middleware:h,strategy:y});E?.($),k?.({placed:!0});const U=Jp(r),B=a9t(U,$.x),G=a9t(U,$.y);r.style.setProperty("--x",`${B}px`),r.style.setProperty("--y",`${G}px`),c.hideWhenDetached&&($.middlewareData.hide?.referenceHidden?(r.style.setProperty("visibility","hidden"),r.style.setProperty("pointer-events","none")):(r.style.removeProperty("visibility"),r.style.removeProperty("pointer-events")));const K=r.firstElementChild;if(K){const X=qB(K);r.style.setProperty("--z-index",X.zIndex)}},O=async()=>{o.updatePosition?(await o.updatePosition({updatePosition:x,floatingElement:r}),k?.({placed:!0})):await x()},P=gHn(c.listeners),D=c.listeners?qzn(s,r,O,P):qFn;return O(),()=>{D?.(),k?.({placed:!1})}}function g2(n,r,o={}){const{defer:s,...c}=o,d=s?ca:b=>b(),h=[];return h.push(d(()=>{const b=typeof n=="function"?n():n,y=typeof r=="function"?r():r;h.push(mHn(b,y,c))})),()=>{h.forEach(b=>b?.())}}var bHn={bottom:"rotate(45deg)",left:"rotate(135deg)",top:"rotate(225deg)",right:"rotate(315deg)"};function iYe(n={}){const{placement:r,sameWidth:o,fitViewport:s,strategy:c="absolute"}=n;return{arrow:{position:"absolute",width:T4.arrowSize.reference,height:T4.arrowSize.reference,[T4.arrowSizeHalf.variable]:`calc(${T4.arrowSize.reference} / 2)`,[T4.arrowOffset.variable]:`calc(${T4.arrowSizeHalf.reference} * -1)`},arrowTip:{transform:r?bHn[r.split("-")[0]]:void 0,background:T4.arrowBg.reference,top:"0",left:"0",width:"100%",height:"100%",position:"absolute",zIndex:"inherit"},floating:{position:c,isolation:"isolate",minWidth:o?void 0:"max-content",width:o?"var(--reference-width)":void 0,maxWidth:s?"var(--available-width)":void 0,maxHeight:s?"var(--available-height)":void 0,pointerEvents:r?void 0:"none",top:"0px",left:"0px",transform:r?"translate3d(var(--x), var(--y), 0)":"translate3d(0, -100vh, 0)",zIndex:"var(--z-index)"}}}function vHn(n){const r={each(o){for(let s=0;s<n.frames?.length;s+=1){const c=n.frames[s];c&&o(c)}},addEventListener(o,s,c){return r.each(d=>{try{d.document.addEventListener(o,s,c)}catch{}}),()=>{try{r.removeEventListener(o,s,c)}catch{}}},removeEventListener(o,s,c){r.each(d=>{try{d.document.removeEventListener(o,s,c)}catch{}})}};return r}function wHn(n){const r=n.frameElement!=null?n.parent:null;return{addEventListener:(o,s,c)=>{try{r?.addEventListener(o,s,c)}catch{}return()=>{try{r?.removeEventListener(o,s,c)}catch{}}},removeEventListener:(o,s,c)=>{try{r?.removeEventListener(o,s,c)}catch{}}}}var s9t="pointerdown.outside",c9t="focus.outside";function yHn(n){for(const r of n)if(Gb(r)&&hM(r))return!0;return!1}var UYt=n=>"clientY"in n;function EHn(n,r){if(!UYt(r)||!n)return!1;const o=n.getBoundingClientRect();return o.width===0||o.height===0?!1:o.top<=r.clientY&&r.clientY<=o.top+o.height&&o.left<=r.clientX&&r.clientX<=o.left+o.width}function SHn(n,r){return n.y<=r.y&&r.y<=n.y+n.height&&n.x<=r.x&&r.x<=n.x+n.width}function u9t(n,r){if(!r||!UYt(n))return!1;const o=r.scrollHeight>r.clientHeight,s=o&&n.clientX>r.offsetLeft+r.clientWidth,c=r.scrollWidth>r.clientWidth,d=c&&n.clientY>r.offsetTop+r.clientHeight,h={x:r.offsetLeft,y:r.offsetTop,width:r.clientWidth+(o?16:0),height:r.clientHeight+(c?16:0)},b={x:n.clientX,y:n.clientY};return SHn(h,b)?s||d:!1}function _Hn(n,r){const{exclude:o,onFocusOutside:s,onPointerDownOutside:c,onInteractOutside:d,defer:h}=r;if(!n)return;const b=ny(n),y=Jp(n),E=vHn(y),k=wHn(y);function x(G,K){if(!Gb(K)||!K.isConnected||f2(n,K)||EHn(n,G))return!1;const X=b.querySelector(`[aria-controls="${n.id}"]`);if(X){const oe=jve(X);if(u9t(G,oe))return!1}const ae=jve(n);return u9t(G,ae)?!1:!o?.(K)}const O=new Set,P=VW(n?.getRootNode());function D(G){function K(X){const ae=h&&!U8t()?ca:ye=>ye(),oe=X??G,he=oe?.composedPath?.()??[oe?.target];ae(()=>{const ye=P?he[0]:Yp(G);if(!(!n||!x(G,ye))){if(c||d){const we=Nie(c,d);n.addEventListener(s9t,we,{once:!0})}l9t(n,s9t,{bubbles:!1,cancelable:!0,detail:{originalEvent:oe,contextmenu:hYt(oe),focusable:yHn(he),target:ye}})}})}G.pointerType==="touch"?(O.forEach(X=>X()),O.add(Ul(b,"click",K,{once:!0})),O.add(k.addEventListener("click",K,{once:!0})),O.add(E.addEventListener("click",K,{once:!0}))):K()}const $=new Set,U=setTimeout(()=>{$.add(Ul(b,"pointerdown",D,!0)),$.add(k.addEventListener("pointerdown",D,!0)),$.add(E.addEventListener("pointerdown",D,!0))},0);function B(G){(h?ca:X=>X())(()=>{const X=Yp(G);if(!(!n||!x(G,X))){if(s||d){const ae=Nie(s,d);n.addEventListener(c9t,ae,{once:!0})}l9t(n,c9t,{bubbles:!1,cancelable:!0,detail:{originalEvent:G,contextmenu:!1,focusable:hM(X),target:X}})}})}return U8t()||($.add(Ul(b,"focusin",B,!0)),$.add(k.addEventListener("focusin",B,!0)),$.add(E.addEventListener("focusin",B,!0))),()=>{clearTimeout(U),O.forEach(G=>G()),$.forEach(G=>G())}}function zYt(n,r){const{defer:o}=r,s=o?ca:d=>d(),c=[];return c.push(s(()=>{const d=typeof n=="function"?n():n;c.push(_Hn(d,r))})),()=>{c.forEach(d=>d?.())}}function l9t(n,r,o){const s=n.ownerDocument.defaultView||window,c=new s.CustomEvent(r,o);return n.dispatchEvent(c)}function kHn(n,r){const o=s=>{s.key==="Escape"&&(s.isComposing||r?.(s))};return Ul(ny(n),"keydown",o,{capture:!0})}var d9t="layer:request-dismiss",Vw={layers:[],branches:[],count(){return this.layers.length},pointerBlockingLayers(){return this.layers.filter(n=>n.pointerBlocking)},topMostPointerBlockingLayer(){return[...this.pointerBlockingLayers()].slice(-1)[0]},hasPointerBlockingLayer(){return this.pointerBlockingLayers().length>0},isBelowPointerBlockingLayer(n){const r=this.indexOf(n),o=this.topMostPointerBlockingLayer()?this.indexOf(this.topMostPointerBlockingLayer()?.node):-1;return r<o},isTopMost(n){return this.layers[this.count()-1]?.node===n},getNestedLayers(n){return Array.from(this.layers).slice(this.indexOf(n)+1)},isInNestedLayer(n,r){return this.getNestedLayers(n).some(o=>f2(o.node,r))},isInBranch(n){return Array.from(this.branches).some(r=>f2(r,n))},add(n){this.layers.push(n),this.syncLayerIndex()},addBranch(n){this.branches.push(n)},remove(n){const r=this.indexOf(n);r<0||(r<this.count()-1&&this.getNestedLayers(n).forEach(s=>Vw.dismiss(s.node,n)),this.layers.splice(r,1),this.syncLayerIndex())},removeBranch(n){const r=this.branches.indexOf(n);r>=0&&this.branches.splice(r,1)},syncLayerIndex(){this.layers.forEach((n,r)=>{n.node.style.setProperty("--layer-index",`${r}`)})},indexOf(n){return this.layers.findIndex(r=>r.node===n)},dismiss(n,r){const o=this.indexOf(n);if(o===-1)return;const s=this.layers[o];xHn(n,d9t,c=>{s.requestDismiss?.(c),c.defaultPrevented||s?.dismiss()}),THn(n,d9t,{originalLayer:n,targetLayer:r,originalIndex:o,targetIndex:r?this.indexOf(r):-1}),this.syncLayerIndex()},clear(){this.remove(this.layers[0].node)}};function THn(n,r,o){const s=n.ownerDocument.defaultView||window,c=new s.CustomEvent(r,{cancelable:!0,bubbles:!0,detail:o});return n.dispatchEvent(c)}function xHn(n,r,o){n.addEventListener(r,o,{once:!0})}var f9t;function h9t(){Vw.layers.forEach(({node:n})=>{n.style.pointerEvents=Vw.isBelowPointerBlockingLayer(n)?"none":"auto"})}function CHn(n){n.style.pointerEvents=""}function AHn(n,r){const o=ny(n),s=[];return Vw.hasPointerBlockingLayer()&&!o.body.hasAttribute("data-inert")&&(f9t=document.body.style.pointerEvents,queueMicrotask(()=>{o.body.style.pointerEvents="none",o.body.setAttribute("data-inert","")})),r?.forEach(c=>{const[d,h]=vUn(()=>{const b=c();return Gb(b)?b:null},{timeout:1e3});d.then(b=>s.push(Lie(b,{pointerEvents:"auto"}))),s.push(h)}),()=>{Vw.hasPointerBlockingLayer()||(queueMicrotask(()=>{o.body.style.pointerEvents=f9t,o.body.removeAttribute("data-inert"),o.body.style.length===0&&o.body.removeAttribute("style")}),s.forEach(c=>c()))}}function IHn(n,r){const{warnOnMissingNode:o=!0}=r;if(o&&!n){qW("[@zag-js/dismissable] node is `null` or `undefined`");return}if(!n)return;const{onDismiss:s,onRequestDismiss:c,pointerBlocking:d,exclude:h,debug:b}=r,y={dismiss:s,node:n,pointerBlocking:d,requestDismiss:c};Vw.add(y),h9t();function E(D){const $=Yp(D.detail.originalEvent);Vw.isBelowPointerBlockingLayer(n)||Vw.isInBranch($)||(r.onPointerDownOutside?.(D),r.onInteractOutside?.(D),!D.defaultPrevented&&(b&&console.log("onPointerDownOutside:",D.detail.originalEvent),s?.()))}function k(D){const $=Yp(D.detail.originalEvent);Vw.isInBranch($)||(r.onFocusOutside?.(D),r.onInteractOutside?.(D),!D.defaultPrevented&&(b&&console.log("onFocusOutside:",D.detail.originalEvent),s?.()))}function x(D){Vw.isTopMost(n)&&(r.onEscapeKeyDown?.(D),!D.defaultPrevented&&s&&(D.preventDefault(),s()))}function O(D){if(!n)return!1;const $=typeof h=="function"?h():h,U=Array.isArray($)?$:[$],B=r.persistentElements?.map(G=>G()).filter(Gb);return B&&U.push(...B),U.some(G=>f2(G,D))||Vw.isInNestedLayer(n,D)}const P=[d?AHn(n,r.persistentElements):void 0,kHn(n,x),zYt(n,{exclude:O,onFocusOutside:k,onPointerDownOutside:E,defer:r.defer})];return()=>{Vw.remove(n),h9t(),CHn(n),P.forEach(D=>D?.())}}function cU(n,r){const{defer:o}=r,s=o?ca:d=>d(),c=[];return c.push(s(()=>{const d=cM(n)?n():n;c.push(IHn(d,r))})),()=>{c.forEach(d=>d?.())}}function OHn(n,r={}){const{defer:o}=r,s=o?ca:d=>d(),c=[];return c.push(s(()=>{const d=cM(n)?n():n;if(!d){qW("[@zag-js/dismissable] branch node is `null` or `undefined`");return}Vw.addBranch(d),c.push(()=>{Vw.removeBranch(d)})})),()=>{c.forEach(d=>d?.())}}var HYt=Vo("color-picker",["root","label","control","trigger","positioner","content","area","areaThumb","valueText","areaBackground","channelSlider","channelSliderLabel","channelSliderTrack","channelSliderThumb","channelSliderValueText","channelInput","transparencyGrid","swatchGroup","swatchTrigger","swatchIndicator","swatch","eyeDropperTrigger","formatTrigger","formatSelect"]);HYt.build();var NHn=n=>n.ids?.hiddenInput??`color-picker:${n.id}:hidden-input`,RHn=n=>n.ids?.control??`color-picker:${n.id}:control`,PHn=n=>n.ids?.trigger??`color-picker:${n.id}:trigger`,LHn=n=>n.ids?.content??`color-picker:${n.id}:content`,DHn=n=>n.ids?.positioner??`color-picker:${n.id}:positioner`,MHn=n=>n.ids?.formatSelect??`color-picker:${n.id}:format-select`,jHn=n=>n.ids?.area??`color-picker:${n.id}:area`,$Hn=n=>n.ids?.areaThumb??`color-picker:${n.id}:area-thumb`,FHn=(n,r)=>n.ids?.channelSliderTrack?.(r)??`color-picker:${n.id}:slider-track:${r}`,BHn=(n,r)=>n.ids?.channelSliderThumb?.(r)??`color-picker:${n.id}:slider-thumb:${r}`,ove=n=>n.getById(LHn(n)),UHn=n=>n.getById($Hn(n)),zHn=(n,r)=>n.getById(BHn(n,r)),HHn=n=>n.getById(MHn(n)),p9t=n=>n.getById(NHn(n)),GHn=n=>n.getById(jHn(n)),qHn=(n,r,o)=>{const s=GHn(n);if(!s)return;const{getPercentValue:c}=yYt(r,s);return{x:c({dir:o,orientation:"horizontal"}),y:c({orientation:"vertical"})}},VHn=n=>n.getById(RHn(n)),Fje=n=>n.getById(PHn(n)),WHn=n=>n.getById(DHn(n)),KHn=(n,r)=>n.getById(FHn(n,r)),YHn=(n,r,o,s)=>{const c=KHn(n,o);if(!c)return;const{getPercentValue:d}=yYt(r,c);return{x:d({dir:s,orientation:"horizontal"}),y:d({orientation:"vertical"})}},XHn=n=>[...Pie(ove(n),"input[data-channel]"),...Pie(VHn(n),"input[data-channel]")];function JHn(n,r){if(r==null)return"";if(r==="hex")return n.toString("hex");if(r==="css")return n.toString("css");if(r in n)return n.getChannelValue(r).toString();const o=n.getFormat()==="hsla";switch(r){case"hue":return o?n.toFormat("hsla").getChannelValue("hue").toString():n.toFormat("hsba").getChannelValue("hue").toString();case"saturation":return o?n.toFormat("hsla").getChannelValue("saturation").toString():n.toFormat("hsba").getChannelValue("saturation").toString();case"lightness":return n.toFormat("hsla").getChannelValue("lightness").toString();case"brightness":return n.toFormat("hsba").getChannelValue("brightness").toString();case"red":case"green":case"blue":return n.toFormat("rgba").getChannelValue(r).toString();default:return n.getChannelValue(r).toString()}}var g9t=n=>Bve(n),ZHn=/^[0-9a-fA-F]{3,8}$/;function QHn(n){return ZHn.test(n)}function eGn(n){return n.startsWith("#")?n:QHn(n)?`#${n}`:n}var{and:tGn}=$I();tGn("isOpenControlled","closeOnSelect");function m9t(n,r,o){const s=XHn(n);ca(()=>{s.forEach(c=>{const d=c.dataset.channel;Soe(c,JHn(o||r,d))})})}function nGn(n,r){const o=HHn(n);o&&ca(()=>Soe(o,r))}Wo()(["closeOnSelect","dir","disabled","format","defaultFormat","getRootNode","id","ids","initialFocusEl","inline","name","positioning","onFocusOutside","onFormatChange","onInteractOutside","onOpenChange","onPointerDownOutside","onValueChange","onValueChangeEnd","defaultOpen","open","positioning","required","readOnly","value","defaultValue","invalid","openAutoFocus"]);Wo()(["xChannel","yChannel"]);Wo()(["channel","orientation"]);Wo()(["value","disabled"]);Wo()(["value","respectAlpha"]);Wo()(["size"]);const[mye,GYt]=fu({name:"RenderStrategyContext",hookName:"useRenderStrategyContext",providerName:"<RenderStrategyPropsProvider />"}),bye=n=>Ou()(n,["lazyMount","unmountOnExit"]),qYt=Q.forwardRef((n,r)=>{const o=HUn();if(o.isUnmounted)return null;const s=Fn(o.getContentProps(),n);return H.jsx(Xn.div,{...s,ref:r})});qYt.displayName="CollapsibleContent";var VYt=Vo("accordion").parts("root","item","itemTrigger","itemContent","itemIndicator");VYt.build();var WYt=n=>n.ids?.root??`accordion:${n.id}`,KYt=(n,r)=>n.ids?.itemTrigger?.(r)??`accordion:${n.id}:trigger:${r}`,rGn=n=>n.getById(WYt(n)),vye=n=>{const o=`[aria-controls][data-ownedby='${CSS.escape(WYt(n))}']:not([disabled])`;return Pie(rGn(n),o)},iGn=n=>nie(vye(n)),oGn=n=>$Ke(vye(n)),aGn=(n,r)=>aUn(vye(n),KYt(n,r)),sGn=(n,r)=>sUn(vye(n),KYt(n,r)),{and:cGn,not:uGn}=$I();cGn("isExpanded","canToggle"),uGn("isExpanded");Wo()(["collapsible","dir","disabled","getRootNode","id","ids","multiple","onFocusChange","onValueChange","orientation","value","defaultValue"]);Wo()(["value","disabled"]);var Nre=(n,r)=>({x:n,y:r});function lGn(n){const{x:r,y:o,width:s,height:c}=n,d=r+s/2,h=o+c/2;return{x:r,y:o,width:s,height:c,minX:r,minY:o,maxX:r+s,maxY:o+c,midX:d,midY:h,center:Nre(d,h)}}function dGn(n){const r=Nre(n.minX,n.minY),o=Nre(n.maxX,n.minY),s=Nre(n.maxX,n.maxY),c=Nre(n.minX,n.maxY);return{top:r,right:o,bottom:s,left:c}}function fGn(n,r){const o=lGn(n),{top:s,right:c,left:d,bottom:h}=dGn(o),[b]=r.split("-");return{top:[d,s,c,h],right:[s,c,h,d],bottom:[s,d,h,c],left:[c,s,d,h]}[b]}function hGn(n,r){const{x:o,y:s}=r;let c=!1;for(let d=0,h=n.length-1;d<n.length;h=d++){const b=n[d].x,y=n[d].y,E=n[h].x,k=n[h].y;y>s!=k>s&&o<(E-b)*(s-y)/(k-y)+b&&(c=!c)}return c}const[YYt,XYt]=fu({name:"AvatarContext",hookName:"useAvatarContext",providerName:"<AvatarProvider />"}),JYt=Q.forwardRef((n,r)=>{const o=XYt(),s=Fn(o.getFallbackProps(),n);return H.jsx(Xn.span,{...s,ref:r})});JYt.displayName="AvatarFallback";const ZYt=Q.forwardRef((n,r)=>{const o=XYt(),s=Fn(o.getImageProps(),n);return H.jsx(Xn.img,{...s,ref:r})});ZYt.displayName="AvatarImage";var QYt=Vo("avatar").parts("root","image","fallback"),Bje=QYt.build(),eXt=n=>n.ids?.root??`avatar:${n.id}`,tXt=n=>n.ids?.image??`avatar:${n.id}:image`,pGn=n=>n.ids?.fallback??`avatar:${n.id}:fallback`,gGn=n=>n.getById(eXt(n)),Gqe=n=>n.getById(tXt(n));function mGn(n,r){const{state:o,send:s,prop:c,scope:d}=n,h=o.matches("loaded");return{loaded:h,setSrc(b){Gqe(d)?.setAttribute("src",b)},setLoaded(){s({type:"img.loaded",src:"api"})},setError(){s({type:"img.error",src:"api"})},getRootProps(){return r.element({...Bje.root.attrs,dir:c("dir"),id:eXt(d)})},getImageProps(){return r.img({...Bje.image.attrs,hidden:!h,dir:c("dir"),id:tXt(d),"data-state":h?"visible":"hidden",onLoad(){s({type:"img.loaded",src:"element"})},onError(){s({type:"img.error",src:"element"})}})},getFallbackProps(){return r.element({...Bje.fallback.attrs,dir:c("dir"),id:pGn(d),hidden:h,"data-state":h?"hidden":"visible"})}}}var bGn={initialState(){return"loading"},effects:["trackImageRemoval","trackSrcChange"],on:{"src.change":{target:"loading"},"img.unmount":{target:"error"}},states:{loading:{entry:["checkImageStatus"],on:{"img.loaded":{target:"loaded",actions:["invokeOnLoad"]},"img.error":{target:"error",actions:["invokeOnError"]}}},error:{on:{"img.loaded":{target:"loaded",actions:["invokeOnLoad"]}}},loaded:{on:{"img.error":{target:"error",actions:["invokeOnError"]}}}},implementations:{actions:{invokeOnLoad({prop:n}){n("onStatusChange")?.({status:"loaded"})},invokeOnError({prop:n}){n("onStatusChange")?.({status:"error"})},checkImageStatus({send:n,scope:r}){const o=Gqe(r);if(!o?.complete)return;const s=vGn(o)?"img.loaded":"img.error";n({type:s,src:"ssr"})}},effects:{trackImageRemoval({send:n,scope:r}){const o=gGn(r);return XBn(o,{callback(s){Array.from(s[0].removedNodes).find(h=>h.nodeType===Node.ELEMENT_NODE&&h.matches("[data-scope=avatar][data-part=image]"))&&n({type:"img.unmount"})}})},trackSrcChange({send:n,scope:r}){const o=Gqe(r);return koe(o,{attributes:["src","srcset"],callback(){n({type:"src.change"})}})}}}};function vGn(n){return n.complete&&n.naturalWidth!==0&&n.naturalHeight!==0}Wo()(["dir","id","ids","onStatusChange","getRootNode"]);const wGn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv(bGn,c);return mGn(d,ry)},nXt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["id","ids","onStatusChange"]),c=wGn(o),d=Fn(c.getRootProps(),s);return H.jsx(YYt,{value:c,children:H.jsx(Xn.div,{...d,ref:r})})});nXt.displayName="AvatarRoot";const rXt=Q.forwardRef((n,r)=>{const[{value:o},s]=Ou()(n,["value"]),c=Fn(o.getRootProps(),s);return H.jsx(YYt,{value:o,children:H.jsx(Xn.div,{...c,ref:r})})});rXt.displayName="AvatarRootProvider";function yGn(n){return!(n.metaKey||!Eoe()&&n.altKey||n.ctrlKey||n.key==="Control"||n.key==="Shift"||n.key==="Meta")}var EGn=new Set(["checkbox","radio","range","color","file","image","button","submit","reset"]);function SGn(n,r,o){const s=o?Yp(o):null,c=Jp(s);return n=n||s instanceof c.HTMLInputElement&&!EGn.has(s?.type)||s instanceof c.HTMLTextAreaElement||s instanceof c.HTMLElement&&s.isContentEditable,!(n&&r==="keyboard"&&o instanceof c.KeyboardEvent&&!Reflect.has(_Gn,o.key))}var uU=null,qqe=new Set,iie=new Map,WB=!1,Vqe=!1,_Gn={Tab:!0,Escape:!0};function wye(n,r){for(let o of qqe)o(n,r)}function Gve(n){WB=!0,yGn(n)&&(uU="keyboard",wye("keyboard",n))}function R_(n){uU="pointer",(n.type==="mousedown"||n.type==="pointerdown")&&(WB=!0,wye("pointer",n))}function iXt(n){NBn(n)&&(WB=!0,uU="virtual")}function oXt(n){const r=Yp(n);r===Jp(r)||r===ny(r)||(!WB&&!Vqe&&(uU="virtual",wye("virtual",n)),WB=!1,Vqe=!1)}function aXt(){WB=!1,Vqe=!0}function kGn(n){if(typeof window>"u"||iie.get(Jp(n)))return;const r=Jp(n),o=ny(n);let s=r.HTMLElement.prototype.focus;function c(){uU="virtual",wye("virtual",null),WB=!0,s.apply(this,arguments)}Object.defineProperty(r.HTMLElement.prototype,"focus",{configurable:!0,value:c}),o.addEventListener("keydown",Gve,!0),o.addEventListener("keyup",Gve,!0),o.addEventListener("click",iXt,!0),r.addEventListener("focus",oXt,!0),r.addEventListener("blur",aXt,!1),typeof r.PointerEvent<"u"?(o.addEventListener("pointerdown",R_,!0),o.addEventListener("pointermove",R_,!0),o.addEventListener("pointerup",R_,!0)):(o.addEventListener("mousedown",R_,!0),o.addEventListener("mousemove",R_,!0),o.addEventListener("mouseup",R_,!0)),r.addEventListener("beforeunload",()=>{TGn(n)},{once:!0}),iie.set(r,{focus:s})}var TGn=(n,r)=>{const o=Jp(n),s=ny(n);iie.has(o)&&(o.HTMLElement.prototype.focus=iie.get(o).focus,s.removeEventListener("keydown",Gve,!0),s.removeEventListener("keyup",Gve,!0),s.removeEventListener("click",iXt,!0),o.removeEventListener("focus",oXt,!0),o.removeEventListener("blur",aXt,!1),typeof o.PointerEvent<"u"?(s.removeEventListener("pointerdown",R_,!0),s.removeEventListener("pointermove",R_,!0),s.removeEventListener("pointerup",R_,!0)):(s.removeEventListener("mousedown",R_,!0),s.removeEventListener("mousemove",R_,!0),s.removeEventListener("mouseup",R_,!0)),iie.delete(o))};function qve(){return uU==="keyboard"}function sXt(n={}){const{isTextInput:r,autoFocus:o,onChange:s,root:c}=n;kGn(c),s?.({isFocusVisible:o||qve(),modality:uU});const d=(h,b)=>{SGn(!!r,h,b)&&s?.({isFocusVisible:qve(),modality:h})};return qqe.add(d),()=>{qqe.delete(d)}}var cXt=Vo("checkbox").parts("root","label","control","indicator");cXt.build();Wo()(["defaultChecked","checked","dir","disabled","form","getRootNode","id","ids","invalid","name","onCheckedChange","readOnly","required","value"]);const xGn=cXt.extendWith("group"),[CGn,$M]=fu({name:"FieldContext",hookName:"useFieldContext",providerName:"<FieldProvider />",strict:!1});var uXt=Vo("clipboard").parts("root","control","trigger","indicator","input","label");uXt.build();Wo()(["getRootNode","id","ids","value","defaultValue","timeout","onStatusChange","onValueChange"]);Wo()(["copied"]);const AGn=HYt.extendWith("view");var IGn=Object.defineProperty,OGn=(n,r,o)=>r in n?IGn(n,r,{enumerable:!0,configurable:!0,writable:!0,value:o}):n[r]=o,Lr=(n,r,o)=>OGn(n,typeof r!="symbol"?r+"":r,o),ave={itemToValue(n){return typeof n=="string"?n:IB(n)&&Mx(n,"value")?n.value:""},itemToString(n){return typeof n=="string"?n:IB(n)&&Mx(n,"label")?n.label:ave.itemToValue(n)},isItemDisabled(n){return IB(n)&&Mx(n,"disabled")?!!n.disabled:!1}},yye=class lXt{constructor(r){this.options=r,Lr(this,"items"),Lr(this,"indexMap",null),Lr(this,"copy",o=>new lXt({...this.options,items:o??[...this.items]})),Lr(this,"isEqual",o=>d2(this.items,o.items)),Lr(this,"setItems",o=>this.copy(o)),Lr(this,"getValues",(o=this.items)=>{const s=[];for(const c of o){const d=this.getItemValue(c);d!=null&&s.push(d)}return s}),Lr(this,"find",o=>{if(o==null)return null;const s=this.indexOf(o);return s!==-1?this.at(s):null}),Lr(this,"findMany",o=>{const s=[];for(const c of o){const d=this.find(c);d!=null&&s.push(d)}return s}),Lr(this,"at",o=>{if(!this.options.groupBy&&!this.options.groupSort)return this.items[o]??null;let s=0;const c=this.group();for(const[,d]of c)for(const h of d){if(s===o)return h;s++}return null}),Lr(this,"sortFn",(o,s)=>{const c=this.indexOf(o),d=this.indexOf(s);return(c??0)-(d??0)}),Lr(this,"sort",o=>[...o].sort(this.sortFn.bind(this))),Lr(this,"getItemValue",o=>o==null?null:this.options.itemToValue?.(o)??ave.itemToValue(o)),Lr(this,"getItemDisabled",o=>o==null?!1:this.options.isItemDisabled?.(o)??ave.isItemDisabled(o)),Lr(this,"stringifyItem",o=>o==null?null:this.options.itemToString?.(o)??ave.itemToString(o)),Lr(this,"stringify",o=>o==null?null:this.stringifyItem(this.find(o))),Lr(this,"stringifyItems",(o,s=", ")=>{const c=[];for(const d of o){const h=this.stringifyItem(d);h!=null&&c.push(h)}return c.join(s)}),Lr(this,"stringifyMany",(o,s)=>this.stringifyItems(this.findMany(o),s)),Lr(this,"has",o=>this.indexOf(o)!==-1),Lr(this,"hasItem",o=>o==null?!1:this.has(this.getItemValue(o))),Lr(this,"group",()=>{const{groupBy:o,groupSort:s}=this.options;if(!o)return[["",[...this.items]]];const c=new Map;this.items.forEach((h,b)=>{const y=o(h,b);c.has(y)||c.set(y,[]),c.get(y).push(h)});let d=Array.from(c.entries());return s&&d.sort(([h],[b])=>{if(typeof s=="function")return s(h,b);if(Array.isArray(s)){const y=s.indexOf(h),E=s.indexOf(b);return y===-1?1:E===-1?-1:y-E}return s==="asc"?h.localeCompare(b):s==="desc"?b.localeCompare(h):0}),d}),Lr(this,"getNextValue",(o,s=1,c=!1)=>{let d=this.indexOf(o);if(d===-1)return null;for(d=c?Math.min(d+s,this.size-1):d+s;d<=this.size&&this.getItemDisabled(this.at(d));)d++;return this.getItemValue(this.at(d))}),Lr(this,"getPreviousValue",(o,s=1,c=!1)=>{let d=this.indexOf(o);if(d===-1)return null;for(d=c?Math.max(d-s,0):d-s;d>=0&&this.getItemDisabled(this.at(d));)d--;return this.getItemValue(this.at(d))}),Lr(this,"indexOf",o=>{if(o==null)return-1;if(!this.options.groupBy&&!this.options.groupSort)return this.items.findIndex(s=>this.getItemValue(s)===o);if(!this.indexMap){this.indexMap=new Map;let s=0;const c=this.group();for(const[,d]of c)for(const h of d){const b=this.getItemValue(h);b!=null&&this.indexMap.set(b,s),s++}}return this.indexMap.get(o)??-1}),Lr(this,"getByText",(o,s)=>{const c=s!=null?this.indexOf(s):-1,d=o.length===1;for(let h=0;h<this.items.length;h++){const b=this.items[(c+h+1)%this.items.length];if(!(d&&this.getItemValue(b)===s)&&!this.getItemDisabled(b)&&NGn(this.stringifyItem(b),o))return b}}),Lr(this,"search",(o,s)=>{const{state:c,currentValue:d,timeout:h=350}=s,b=c.keysSoFar+o,E=b.length>1&&Array.from(b).every(D=>D===b[0])?b[0]:b,k=this.getByText(E,d),x=this.getItemValue(k);function O(){clearTimeout(c.timer),c.timer=-1}function P(D){c.keysSoFar=D,O(),D!==""&&(c.timer=+setTimeout(()=>{P(""),O()},h))}return P(b),x}),Lr(this,"update",(o,s)=>{let c=this.indexOf(o);return c===-1?this:this.copy([...this.items.slice(0,c),s,...this.items.slice(c+1)])}),Lr(this,"upsert",(o,s,c="append")=>{let d=this.indexOf(o);return d===-1?(c==="append"?this.append:this.prepend)(s):this.copy([...this.items.slice(0,d),s,...this.items.slice(d+1)])}),Lr(this,"insert",(o,...s)=>this.copy(tre(this.items,o,...s))),Lr(this,"insertBefore",(o,...s)=>{let c=this.indexOf(o);if(c===-1)if(this.items.length===0)c=0;else return this;return this.copy(tre(this.items,c,...s))}),Lr(this,"insertAfter",(o,...s)=>{let c=this.indexOf(o);if(c===-1)if(this.items.length===0)c=0;else return this;return this.copy(tre(this.items,c+1,...s))}),Lr(this,"prepend",(...o)=>this.copy(tre(this.items,0,...o))),Lr(this,"append",(...o)=>this.copy(tre(this.items,this.items.length,...o))),Lr(this,"filter",o=>{const s=this.items.filter((c,d)=>o(this.stringifyItem(c),d,c));return this.copy(s)}),Lr(this,"remove",(...o)=>{const s=o.map(c=>typeof c=="string"?c:this.getItemValue(c));return this.copy(this.items.filter(c=>{const d=this.getItemValue(c);return d==null?!1:!s.includes(d)}))}),Lr(this,"move",(o,s)=>{const c=this.indexOf(o);return c===-1?this:this.copy(D1e(this.items,[c],s))}),Lr(this,"moveBefore",(o,...s)=>{let c=this.items.findIndex(h=>this.getItemValue(h)===o);if(c===-1)return this;let d=s.map(h=>this.items.findIndex(b=>this.getItemValue(b)===h)).sort((h,b)=>h-b);return this.copy(D1e(this.items,d,c))}),Lr(this,"moveAfter",(o,...s)=>{let c=this.items.findIndex(h=>this.getItemValue(h)===o);if(c===-1)return this;let d=s.map(h=>this.items.findIndex(b=>this.getItemValue(b)===h)).sort((h,b)=>h-b);return this.copy(D1e(this.items,d,c+1))}),Lr(this,"reorder",(o,s)=>this.copy(D1e(this.items,[o],s))),Lr(this,"compareValue",(o,s)=>{const c=this.indexOf(o),d=this.indexOf(s);return c<d?-1:c>d?1:0}),Lr(this,"range",(o,s)=>{let c=[],d=o;for(;d!=null;){if(this.find(d)&&c.push(d),d===s)return c;d=this.getNextValue(d)}return[]}),Lr(this,"getValueRange",(o,s)=>o&&s?this.compareValue(o,s)<=0?this.range(o,s):this.range(s,o):[]),Lr(this,"toString",()=>{let o="";for(const s of this.items){const c=this.getItemValue(s),d=this.stringifyItem(s),h=this.getItemDisabled(s),b=[c,d,h].filter(Boolean).join(":");o+=b+","}return o}),Lr(this,"toJSON",()=>({size:this.size,first:this.firstValue,last:this.lastValue})),this.items=[...r.items]}get size(){return this.items.length}get firstValue(){let r=0;for(;this.getItemDisabled(this.at(r));)r++;return this.getItemValue(this.at(r))}get lastValue(){let r=this.size-1;for(;this.getItemDisabled(this.at(r));)r--;return this.getItemValue(this.at(r))}*[Symbol.iterator](){yield*this.items}},NGn=(n,r)=>!!n?.toLowerCase().startsWith(r.toLowerCase());function tre(n,r,...o){return[...n.slice(0,r),...o,...n.slice(r)]}function D1e(n,r,o){r=[...r].sort((c,d)=>c-d);const s=r.map(c=>n[c]);for(let c=r.length-1;c>=0;c--)n=[...n.slice(0,r[c]),...n.slice(r[c]+1)];return o=Math.max(0,o-r.filter(c=>c<o).length),[...n.slice(0,o),...s,...n.slice(o)]}function dXt(n,r,o){for(let s=0;s<r.length;s++)n=o.getChildren(n,r.slice(s+1))[r[s]];return n}function RGn(n){const r=PGn(n),o=[],s=new Set;for(const c of r){const d=c.join();s.has(d)||(s.add(d),o.push(c))}return o}function fXt(n,r){for(let o=0;o<Math.min(n.length,r.length);o++){if(n[o]<r[o])return-1;if(n[o]>r[o])return 1}return n.length-r.length}function PGn(n){return n.sort(fXt)}function LGn(n,r){let o;return t2(n,{...r,onEnter:(s,c)=>{if(r.predicate(s,c))return o=s,"stop"}}),o}function DGn(n,r){const o=[];return t2(n,{onEnter:(s,c)=>{r.predicate(s,c)&&o.push(s)},getChildren:r.getChildren}),o}function b9t(n,r){let o;return t2(n,{onEnter:(s,c)=>{if(r.predicate(s,c))return o=[...c],"stop"},getChildren:r.getChildren}),o}function MGn(n,r){let o=r.initialResult;return t2(n,{...r,onEnter:(s,c)=>{o=r.nextResult(o,s,c)}}),o}function jGn(n,r){return MGn(n,{...r,initialResult:[],nextResult:(o,s,c)=>(o.push(...r.transform(s,c)),o)})}function $Gn(n,r){const{predicate:o,create:s,getChildren:c}=r,d=(h,b)=>{const y=c(h,b),E=[];y.forEach((P,D)=>{const $=[...b,D],U=d(P,$);U&&E.push(U)});const k=b.length===0,x=o(h,b),O=E.length>0;return k||x||O?s(h,E,b):null};return d(n,[])||s(n,[],[])}function FGn(n,r){const o=[];let s=0;const c=new Map,d=new Map;return t2(n,{getChildren:r.getChildren,onEnter:(h,b)=>{c.has(h)||c.set(h,s++);const y=r.getChildren(h,b);y.forEach(P=>{d.has(P)||d.set(P,h),c.has(P)||c.set(P,s++)});const E=y.length>0?y.map(P=>c.get(P)):void 0,k=d.get(h),x=k?c.get(k):void 0,O=c.get(h);o.push({...h,_children:E,_parent:x,_index:O})}}),o}function BGn(n,r){return{type:"insert",index:n,nodes:r}}function UGn(n){return{type:"remove",indexes:n}}function oYe(){return{type:"replace"}}function hXt(n){return[n.slice(0,-1),n[n.length-1]]}function pXt(n,r,o=new Map){const[s,c]=hXt(n);for(let h=s.length-1;h>=0;h--){const b=s.slice(0,h).join();switch(o.get(b)?.type){case"remove":continue}o.set(b,oYe())}const d=o.get(s.join());switch(d?.type){case"remove":o.set(s.join(),{type:"removeThenInsert",removeIndexes:d.indexes,insertIndex:c,insertNodes:r});break;default:o.set(s.join(),BGn(c,r))}return o}function gXt(n){const r=new Map,o=new Map;for(const s of n){const c=s.slice(0,-1).join(),d=o.get(c)??[];d.push(s[s.length-1]),o.set(c,d.sort((h,b)=>h-b))}for(const s of n)for(let c=s.length-2;c>=0;c--){const d=s.slice(0,c).join();r.has(d)||r.set(d,oYe())}for(const[s,c]of o)r.set(s,UGn(c));return r}function zGn(n,r){const o=new Map,[s,c]=hXt(n);for(let d=s.length-1;d>=0;d--){const h=s.slice(0,d).join();o.set(h,oYe())}return o.set(s.join(),{type:"removeThenInsert",removeIndexes:[c],insertIndex:c,insertNodes:[r]}),o}function Eye(n,r,o){return HGn(n,{...o,getChildren:(s,c)=>{const d=c.join();switch(r.get(d)?.type){case"replace":case"remove":case"removeThenInsert":case"insert":return o.getChildren(s,c);default:return[]}},transform:(s,c,d)=>{const h=d.join(),b=r.get(h);switch(b?.type){case"remove":return o.create(s,c.filter((k,x)=>!b.indexes.includes(x)),d);case"removeThenInsert":const y=c.filter((k,x)=>!b.removeIndexes.includes(x)),E=b.removeIndexes.reduce((k,x)=>x<k?k-1:k,b.insertIndex);return o.create(s,v9t(y,E,0,...b.insertNodes),d);case"insert":return o.create(s,v9t(c,b.index,0,...b.nodes),d);case"replace":return o.create(s,c,d);default:return s}}})}function v9t(n,r,o,...s){return[...n.slice(0,r),...s,...n.slice(r+o)]}function HGn(n,r){const o={};return t2(n,{...r,onLeave:(s,c)=>{const d=[0,...c],h=d.join(),b=r.transform(s,o[h]??[],c),y=d.slice(0,-1).join(),E=o[y]??[];E.push(b),o[y]=E}}),o[""][0]}function GGn(n,r){const{nodes:o,at:s}=r;if(s.length===0)throw new Error("Can't insert nodes at the root");const c=pXt(s,o);return Eye(n,c,r)}function qGn(n,r){if(r.at.length===0)return r.node;const o=zGn(r.at,r.node);return Eye(n,o,r)}function VGn(n,r){if(r.indexPaths.length===0)return n;for(const s of r.indexPaths)if(s.length===0)throw new Error("Can't remove the root node");const o=gXt(r.indexPaths);return Eye(n,o,r)}function WGn(n,r){if(r.indexPaths.length===0)return n;for(const d of r.indexPaths)if(d.length===0)throw new Error("Can't move the root node");if(r.to.length===0)throw new Error("Can't move nodes to the root");const o=RGn(r.indexPaths),s=o.map(d=>dXt(n,d,r)),c=pXt(r.to,s,gXt(o));return Eye(n,c,r)}function t2(n,r){const{onEnter:o,onLeave:s,getChildren:c}=r;let d=[],h=[{node:n}];const b=r.reuseIndexPath?()=>d:()=>d.slice();for(;h.length>0;){let y=h[h.length-1];if(y.state===void 0){const k=o?.(y.node,b());if(k==="stop")return;y.state=k==="skip"?-1:0}const E=y.children||c(y.node,b());if(y.children||(y.children=E),y.state!==-1){if(y.state<E.length){let x=y.state;d.push(x),h.push({node:E[x]}),y.state=x+1;continue}if(s?.(y.node,b())==="stop")return}d.pop(),h.pop()}}var aYe=class mXt{constructor(r){this.options=r,Lr(this,"rootNode"),Lr(this,"isEqual",o=>d2(this.rootNode,o.rootNode)),Lr(this,"getNodeChildren",o=>this.options.nodeToChildren?.(o)??nW.nodeToChildren(o)??[]),Lr(this,"resolveIndexPath",o=>typeof o=="string"?this.getIndexPath(o):o),Lr(this,"resolveNode",o=>{const s=this.resolveIndexPath(o);return s?this.at(s):void 0}),Lr(this,"getNodeChildrenCount",o=>this.options.nodeToChildrenCount?.(o)??nW.nodeToChildrenCount(o)),Lr(this,"getNodeValue",o=>this.options.nodeToValue?.(o)??nW.nodeToValue(o)),Lr(this,"getNodeDisabled",o=>this.options.isNodeDisabled?.(o)??nW.isNodeDisabled(o)),Lr(this,"stringify",o=>{const s=this.findNode(o);return s?this.stringifyNode(s):null}),Lr(this,"stringifyNode",o=>this.options.nodeToString?.(o)??nW.nodeToString(o)),Lr(this,"getFirstNode",(o=this.rootNode)=>{let s;return t2(o,{getChildren:this.getNodeChildren,onEnter:(c,d)=>{if(!s&&d.length>0&&!this.getNodeDisabled(c))return s=c,"stop"}}),s}),Lr(this,"getLastNode",(o=this.rootNode,s={})=>{let c;return t2(o,{getChildren:this.getNodeChildren,onEnter:(d,h)=>{if(!this.isSameNode(d,o)){if(s.skip?.({value:this.getNodeValue(d),node:d,indexPath:h}))return"skip";h.length>0&&!this.getNodeDisabled(d)&&(c=d)}}}),c}),Lr(this,"at",o=>dXt(this.rootNode,o,{getChildren:this.getNodeChildren})),Lr(this,"findNode",(o,s=this.rootNode)=>LGn(s,{getChildren:this.getNodeChildren,predicate:c=>this.getNodeValue(c)===o})),Lr(this,"findNodes",(o,s=this.rootNode)=>{const c=new Set(o.filter(d=>d!=null));return DGn(s,{getChildren:this.getNodeChildren,predicate:d=>c.has(this.getNodeValue(d))})}),Lr(this,"sort",o=>o.reduce((s,c)=>{const d=this.getIndexPath(c);return d&&s.push({value:c,indexPath:d}),s},[]).sort((s,c)=>fXt(s.indexPath,c.indexPath)).map(({value:s})=>s)),Lr(this,"getIndexPath",o=>b9t(this.rootNode,{getChildren:this.getNodeChildren,predicate:s=>this.getNodeValue(s)===o})),Lr(this,"getValue",o=>{const s=this.at(o);return s?this.getNodeValue(s):void 0}),Lr(this,"getValuePath",o=>{if(!o)return[];const s=[];let c=[...o];for(;c.length>0;){const d=this.at(c);d&&s.unshift(this.getNodeValue(d)),c.pop()}return s}),Lr(this,"getDepth",o=>b9t(this.rootNode,{getChildren:this.getNodeChildren,predicate:c=>this.getNodeValue(c)===o})?.length??0),Lr(this,"isSameNode",(o,s)=>this.getNodeValue(o)===this.getNodeValue(s)),Lr(this,"isRootNode",o=>this.isSameNode(o,this.rootNode)),Lr(this,"contains",(o,s)=>!o||!s?!1:s.slice(0,o.length).every((c,d)=>o[d]===s[d])),Lr(this,"getNextNode",(o,s={})=>{let c=!1,d;return t2(this.rootNode,{getChildren:this.getNodeChildren,onEnter:(h,b)=>{if(this.isRootNode(h))return;const y=this.getNodeValue(h);if(s.skip?.({value:y,node:h,indexPath:b}))return y===o&&(c=!0),"skip";if(c&&!this.getNodeDisabled(h))return d=h,"stop";y===o&&(c=!0)}}),d}),Lr(this,"getPreviousNode",(o,s={})=>{let c,d=!1;return t2(this.rootNode,{getChildren:this.getNodeChildren,onEnter:(h,b)=>{if(this.isRootNode(h))return;const y=this.getNodeValue(h);if(s.skip?.({value:y,node:h,indexPath:b}))return"skip";if(y===o)return d=!0,"stop";this.getNodeDisabled(h)||(c=h)}}),d?c:void 0}),Lr(this,"getParentNodes",o=>{const s=this.resolveIndexPath(o)?.slice();if(!s)return[];const c=[];for(;s.length>0;){s.pop();const d=this.at(s);d&&!this.isRootNode(d)&&c.unshift(d)}return c}),Lr(this,"getDescendantNodes",(o,s)=>{const c=this.resolveNode(o);if(!c)return[];const d=[];return t2(c,{getChildren:this.getNodeChildren,onEnter:(h,b)=>{b.length!==0&&(!s?.withBranch&&this.isBranchNode(h)||d.push(h))}}),d}),Lr(this,"getDescendantValues",(o,s)=>this.getDescendantNodes(o,s).map(d=>this.getNodeValue(d))),Lr(this,"getParentIndexPath",o=>o.slice(0,-1)),Lr(this,"getParentNode",o=>{const s=this.resolveIndexPath(o);return s?this.at(this.getParentIndexPath(s)):void 0}),Lr(this,"visit",o=>{const{skip:s,...c}=o;t2(this.rootNode,{...c,getChildren:this.getNodeChildren,onEnter:(d,h)=>{if(!this.isRootNode(d))return s?.({value:this.getNodeValue(d),node:d,indexPath:h})?"skip":c.onEnter?.(d,h)}})}),Lr(this,"getPreviousSibling",o=>{const s=this.getParentNode(o);if(!s)return;const c=this.getNodeChildren(s);let d=o[o.length-1];for(;--d>=0;){const h=c[d];if(!this.getNodeDisabled(h))return h}}),Lr(this,"getNextSibling",o=>{const s=this.getParentNode(o);if(!s)return;const c=this.getNodeChildren(s);let d=o[o.length-1];for(;++d<c.length;){const h=c[d];if(!this.getNodeDisabled(h))return h}}),Lr(this,"getSiblingNodes",o=>{const s=this.getParentNode(o);return s?this.getNodeChildren(s):[]}),Lr(this,"getValues",(o=this.rootNode)=>jGn(o,{getChildren:this.getNodeChildren,transform:c=>[this.getNodeValue(c)]}).slice(1)),Lr(this,"isValidDepth",(o,s)=>s==null?!0:typeof s=="function"?s(o.length):o.length===s),Lr(this,"isBranchNode",o=>this.getNodeChildren(o).length>0||this.getNodeChildrenCount(o)!=null),Lr(this,"getBranchValues",(o=this.rootNode,s={})=>{let c=[];return t2(o,{getChildren:this.getNodeChildren,onEnter:(d,h)=>{if(h.length===0)return;const b=this.getNodeValue(d);if(s.skip?.({value:b,node:d,indexPath:h}))return"skip";this.isBranchNode(d)&&this.isValidDepth(h,s.depth)&&c.push(this.getNodeValue(d))}}),c}),Lr(this,"flatten",(o=this.rootNode)=>FGn(o,{getChildren:this.getNodeChildren})),Lr(this,"_create",(o,s)=>this.getNodeChildren(o).length>0||s.length>0?{...o,children:s}:{...o}),Lr(this,"_insert",(o,s,c)=>this.copy(GGn(o,{at:s,nodes:c,getChildren:this.getNodeChildren,create:this._create}))),Lr(this,"copy",o=>new mXt({...this.options,rootNode:o})),Lr(this,"_replace",(o,s,c)=>this.copy(qGn(o,{at:s,node:c,getChildren:this.getNodeChildren,create:this._create}))),Lr(this,"_move",(o,s,c)=>this.copy(WGn(o,{indexPaths:s,to:c,getChildren:this.getNodeChildren,create:this._create}))),Lr(this,"_remove",(o,s)=>this.copy(VGn(o,{indexPaths:s,getChildren:this.getNodeChildren,create:this._create}))),Lr(this,"replace",(o,s)=>this._replace(this.rootNode,o,s)),Lr(this,"remove",o=>this._remove(this.rootNode,o)),Lr(this,"insertBefore",(o,s)=>this.getParentNode(o)?this._insert(this.rootNode,o,s):void 0),Lr(this,"insertAfter",(o,s)=>{if(!this.getParentNode(o))return;const d=[...o.slice(0,-1),o[o.length-1]+1];return this._insert(this.rootNode,d,s)}),Lr(this,"move",(o,s)=>this._move(this.rootNode,o,s)),Lr(this,"filter",o=>{const s=$Gn(this.rootNode,{predicate:o,getChildren:this.getNodeChildren,create:this._create});return this.copy(s)}),Lr(this,"toJSON",()=>this.getValues(this.rootNode)),this.rootNode=r.rootNode}},nW={nodeToValue(n){return typeof n=="string"?n:IB(n)&&Mx(n,"value")?n.value:""},nodeToString(n){return typeof n=="string"?n:IB(n)&&Mx(n,"label")?n.label:nW.nodeToValue(n)},isNodeDisabled(n){return IB(n)&&Mx(n,"disabled")?!!n.disabled:!1},nodeToChildren(n){return n.children},nodeToChildrenCount(n){if(IB(n)&&Mx(n,"childrenCount"))return n.childrenCount}};const KGn=n=>new aYe(n);var PV=new WeakMap,M1e=new WeakMap,j1e={},Uje=0,bXt=n=>n&&(n.host||bXt(n.parentNode)),YGn=(n,r)=>r.map(o=>{if(n.contains(o))return o;const s=bXt(o);return s&&n.contains(s)?s:(console.error("[zag-js > ariaHidden] target",o,"in not contained inside",n,". Doing nothing"),null)}).filter(o=>!!o),XGn=new Set(["script","output","status","next-route-announcer"]),JGn=n=>XGn.has(n.localName)||n.role==="status"||n.hasAttribute("aria-live")?!0:n.matches("[data-live-announcer]"),ZGn=(n,r)=>{const{parentNode:o,markerName:s,controlAttribute:c}=r,d=YGn(o,Array.isArray(n)?n:[n]);j1e[s]||(j1e[s]=new WeakMap);const h=j1e[s],b=[],y=new Set,E=new Set(d),k=O=>{!O||y.has(O)||(y.add(O),k(O.parentNode))};d.forEach(k);const x=O=>{!O||E.has(O)||Array.prototype.forEach.call(O.children,P=>{if(y.has(P))x(P);else try{if(JGn(P))return;const $=P.getAttribute(c)==="true",U=(PV.get(P)||0)+1,B=(h.get(P)||0)+1;PV.set(P,U),h.set(P,B),b.push(P),U===1&&$&&M1e.set(P,!0),B===1&&P.setAttribute(s,""),$||P.setAttribute(c,"true")}catch(D){console.error("[zag-js > ariaHidden] cannot operate on ",P,D)}})};return x(o),y.clear(),Uje++,()=>{b.forEach(O=>{const P=PV.get(O)-1,D=h.get(O)-1;PV.set(O,P),h.set(O,D),P||(M1e.has(O)||O.removeAttribute(c),M1e.delete(O)),D||O.removeAttribute(s)}),Uje--,Uje||(PV=new WeakMap,PV=new WeakMap,M1e=new WeakMap,j1e={})}},QGn=n=>(Array.isArray(n)?n[0]:n).ownerDocument.body,eqn=(n,r=QGn(n),o="data-aria-hidden")=>{if(r)return ZGn(n,{parentNode:r,markerName:o,controlAttribute:"aria-hidden"})},tqn=n=>{const r=requestAnimationFrame(()=>n());return()=>cancelAnimationFrame(r)};function sYe(n,r={}){const{defer:o=!0}=r,s=o?tqn:d=>d(),c=[];return c.push(s(()=>{const h=(typeof n=="function"?n():n).filter(Boolean);h.length!==0&&c.push(eqn(h))})),()=>{c.forEach(d=>d?.())}}var vXt=Vo("combobox").parts("root","clearTrigger","content","control","input","item","itemGroup","itemGroupLabel","itemIndicator","itemText","label","list","positioner","trigger");vXt.build();var wXt=n=>new yye(n);wXt.empty=()=>new yye({items:[]});var nqn=n=>n.ids?.control??`combobox:${n.id}:control`,rqn=n=>n.ids?.input??`combobox:${n.id}:input`,iqn=n=>n.ids?.content??`combobox:${n.id}:content`,oqn=n=>n.ids?.positioner??`combobox:${n.id}:popper`,aqn=n=>n.ids?.trigger??`combobox:${n.id}:toggle-btn`,sqn=n=>n.ids?.clearTrigger??`combobox:${n.id}:clear-btn`,YD=n=>n.getById(iqn(n)),rW=n=>n.getById(rqn(n)),w9t=n=>n.getById(oqn(n)),y9t=n=>n.getById(nqn(n)),Rre=n=>n.getById(aqn(n)),E9t=n=>n.getById(sqn(n)),nre=(n,r)=>{if(r==null)return null;const o=`[role=option][data-value="${CSS.escape(r)}"]`;return vK(YD(n),o)},S9t=n=>{const r=rW(n);n.isActiveElement(r)||r?.focus({preventScroll:!0})},cqn=n=>{const r=Rre(n);n.isActiveElement(r)||r?.focus({preventScroll:!0})},{guards:uqn,createMachine:lqn,choose:dqn}=_Yt(),{and:Mm,not:XE}=uqn;lqn({props({props:n}){return{loopFocus:!0,openOnClick:!1,defaultValue:[],closeOnSelect:!n.multiple,allowCustomValue:!1,inputBehavior:"none",selectionBehavior:n.multiple?"clear":"replace",openOnKeyPress:!0,openOnChange:!0,composite:!0,navigate({node:r}){vYt(r)},collection:wXt.empty(),...n,positioning:{placement:"bottom",sameWidth:!0,...n.positioning},translations:{triggerLabel:"Toggle suggestions",clearTriggerLabel:"Clear value",...n.translations}}},initialState({prop:n}){return n("open")||n("defaultOpen")?"suggesting":"idle"},context({prop:n,bindable:r,getContext:o,getEvent:s}){return{currentPlacement:r(()=>({defaultValue:void 0})),value:r(()=>({defaultValue:n("defaultValue"),value:n("value"),isEqual:d2,hash(c){return c.join(",")},onChange(c){const d=o(),h=d.get("selectedItems"),b=n("collection"),y=c.map(E=>h.find(x=>b.getItemValue(x)===E)||b.find(E));d.set("selectedItems",y),n("onValueChange")?.({value:c,items:y})}})),highlightedValue:r(()=>({defaultValue:n("defaultHighlightedValue")||null,value:n("highlightedValue"),onChange(c){const d=n("collection").find(c);n("onHighlightChange")?.({highlightedValue:c,highlightedItem:d})}})),inputValue:r(()=>{let c=n("inputValue")||n("defaultInputValue")||"";const d=n("defaultValue")||n("value")||[];if(!c.trim()&&!n("multiple")){const h=n("collection").stringifyMany(d);c=_4(n("selectionBehavior"),{preserve:c||h,replace:h,clear:""})}return{defaultValue:c,value:n("inputValue"),onChange(h){const b=s(),y=(b.previousEvent||b).src;n("onInputValueChange")?.({inputValue:h,reason:y})}}}),highlightedItem:r(()=>{const c=n("highlightedValue");return{defaultValue:n("collection").find(c)}}),selectedItems:r(()=>{const c=n("value")||n("defaultValue")||[];return{defaultValue:n("collection").findMany(c)}})}},computed:{isInputValueEmpty:({context:n})=>n.get("inputValue").length===0,isInteractive:({prop:n})=>!(n("readOnly")||n("disabled")),autoComplete:({prop:n})=>n("inputBehavior")==="autocomplete",autoHighlight:({prop:n})=>n("inputBehavior")==="autohighlight",hasSelectedItems:({context:n})=>n.get("value").length>0,valueAsString:({context:n,prop:r})=>r("collection").stringifyItems(n.get("selectedItems")),isCustomValue:({context:n,computed:r})=>n.get("inputValue")!==r("valueAsString")},watch({context:n,prop:r,track:o,action:s,send:c}){o([()=>n.hash("value")],()=>{s(["syncSelectedItems"])}),o([()=>n.get("inputValue")],()=>{s(["syncInputValue"])}),o([()=>n.get("highlightedValue")],()=>{s(["syncHighlightedItem","autofillInputValue"])}),o([()=>r("open")],()=>{s(["toggleVisibility"])}),o([()=>r("collection").toString()],()=>{c({type:"CHILDREN_CHANGE"})})},on:{"SELECTED_ITEMS.SYNC":{actions:["syncSelectedItems"]},"HIGHLIGHTED_VALUE.SET":{actions:["setHighlightedValue"]},"HIGHLIGHTED_VALUE.CLEAR":{actions:["clearHighlightedValue"]},"ITEM.SELECT":{actions:["selectItem"]},"ITEM.CLEAR":{actions:["clearItem"]},"VALUE.SET":{actions:["setValue"]},"INPUT_VALUE.SET":{actions:["setInputValue"]},"POSITIONING.SET":{actions:["reposition"]}},entry:dqn([{guard:"autoFocus",actions:["setInitialFocus"]}]),states:{idle:{tags:["idle","closed"],entry:["scrollContentToTop","clearHighlightedValue"],on:{"CONTROLLED.OPEN":{target:"interacting"},"TRIGGER.CLICK":[{guard:"isOpenControlled",actions:["setInitialFocus","highlightFirstSelectedItem","invokeOnOpen"]},{target:"interacting",actions:["setInitialFocus","highlightFirstSelectedItem","invokeOnOpen"]}],"INPUT.CLICK":[{guard:"isOpenControlled",actions:["highlightFirstSelectedItem","invokeOnOpen"]},{target:"interacting",actions:["highlightFirstSelectedItem","invokeOnOpen"]}],"INPUT.FOCUS":{target:"focused"},OPEN:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"interacting",actions:["invokeOnOpen"]}],"VALUE.CLEAR":{target:"focused",actions:["clearInputValue","clearSelectedItems","setInitialFocus"]}}},focused:{tags:["focused","closed"],entry:["scrollContentToTop","clearHighlightedValue"],on:{"CONTROLLED.OPEN":[{guard:"isChangeEvent",target:"suggesting"},{target:"interacting"}],"INPUT.CHANGE":[{guard:Mm("isOpenControlled","openOnChange"),actions:["setInputValue","invokeOnOpen","highlightFirstItemIfNeeded"]},{guard:"openOnChange",target:"suggesting",actions:["setInputValue","invokeOnOpen","highlightFirstItemIfNeeded"]},{actions:["setInputValue"]}],"LAYER.INTERACT_OUTSIDE":{target:"idle"},"INPUT.ESCAPE":{guard:Mm("isCustomValue",XE("allowCustomValue")),actions:["revertInputValue"]},"INPUT.BLUR":{target:"idle"},"INPUT.CLICK":[{guard:"isOpenControlled",actions:["highlightFirstSelectedItem","invokeOnOpen"]},{target:"interacting",actions:["highlightFirstSelectedItem","invokeOnOpen"]}],"TRIGGER.CLICK":[{guard:"isOpenControlled",actions:["setInitialFocus","highlightFirstSelectedItem","invokeOnOpen"]},{target:"interacting",actions:["setInitialFocus","highlightFirstSelectedItem","invokeOnOpen"]}],"INPUT.ARROW_DOWN":[{guard:Mm("isOpenControlled","autoComplete"),actions:["invokeOnOpen"]},{guard:"autoComplete",target:"interacting",actions:["invokeOnOpen"]},{guard:"isOpenControlled",actions:["highlightFirstOrSelectedItem","invokeOnOpen"]},{target:"interacting",actions:["highlightFirstOrSelectedItem","invokeOnOpen"]}],"INPUT.ARROW_UP":[{guard:"autoComplete",target:"interacting",actions:["invokeOnOpen"]},{guard:"autoComplete",target:"interacting",actions:["invokeOnOpen"]},{target:"interacting",actions:["highlightLastOrSelectedItem","invokeOnOpen"]},{target:"interacting",actions:["highlightLastOrSelectedItem","invokeOnOpen"]}],OPEN:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"interacting",actions:["invokeOnOpen"]}],"VALUE.CLEAR":{actions:["clearInputValue","clearSelectedItems"]}}},interacting:{tags:["open","focused"],entry:["setInitialFocus"],effects:["scrollToHighlightedItem","trackDismissableLayer","trackPlacement","hideOtherElements"],on:{"CONTROLLED.CLOSE":[{guard:"restoreFocus",target:"focused",actions:["setFinalFocus"]},{target:"idle"}],CHILDREN_CHANGE:[{guard:"isHighlightedItemRemoved",actions:["clearHighlightedValue"]},{actions:["scrollToHighlightedItem"]}],"INPUT.HOME":{actions:["highlightFirstItem"]},"INPUT.END":{actions:["highlightLastItem"]},"INPUT.ARROW_DOWN":[{guard:Mm("autoComplete","isLastItemHighlighted"),actions:["clearHighlightedValue","scrollContentToTop"]},{actions:["highlightNextItem"]}],"INPUT.ARROW_UP":[{guard:Mm("autoComplete","isFirstItemHighlighted"),actions:["clearHighlightedValue"]},{actions:["highlightPrevItem"]}],"INPUT.ENTER":[{guard:Mm("isOpenControlled","isCustomValue",XE("hasHighlightedItem"),XE("allowCustomValue")),actions:["revertInputValue","invokeOnClose"]},{guard:Mm("isCustomValue",XE("hasHighlightedItem"),XE("allowCustomValue")),target:"focused",actions:["revertInputValue","invokeOnClose"]},{guard:Mm("isOpenControlled","closeOnSelect"),actions:["selectHighlightedItem","invokeOnClose"]},{guard:"closeOnSelect",target:"focused",actions:["selectHighlightedItem","invokeOnClose","setFinalFocus"]},{actions:["selectHighlightedItem"]}],"INPUT.CHANGE":[{guard:"autoComplete",target:"suggesting",actions:["setInputValue"]},{target:"suggesting",actions:["clearHighlightedValue","setInputValue"]}],"ITEM.POINTER_MOVE":{actions:["setHighlightedValue"]},"ITEM.POINTER_LEAVE":{actions:["clearHighlightedValue"]},"ITEM.CLICK":[{guard:Mm("isOpenControlled","closeOnSelect"),actions:["selectItem","invokeOnClose"]},{guard:"closeOnSelect",target:"focused",actions:["selectItem","invokeOnClose","setFinalFocus"]},{actions:["selectItem"]}],"LAYER.ESCAPE":[{guard:Mm("isOpenControlled","autoComplete"),actions:["syncInputValue","invokeOnClose"]},{guard:"autoComplete",target:"focused",actions:["syncInputValue","invokeOnClose"]},{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"focused",actions:["invokeOnClose","setFinalFocus"]}],"TRIGGER.CLICK":[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"focused",actions:["invokeOnClose"]}],"LAYER.INTERACT_OUTSIDE":[{guard:Mm("isOpenControlled","isCustomValue",XE("allowCustomValue")),actions:["revertInputValue","invokeOnClose"]},{guard:Mm("isCustomValue",XE("allowCustomValue")),target:"idle",actions:["revertInputValue","invokeOnClose"]},{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"idle",actions:["invokeOnClose"]}],CLOSE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"focused",actions:["invokeOnClose","setFinalFocus"]}],"VALUE.CLEAR":[{guard:"isOpenControlled",actions:["clearInputValue","clearSelectedItems","invokeOnClose"]},{target:"focused",actions:["clearInputValue","clearSelectedItems","invokeOnClose","setFinalFocus"]}]}},suggesting:{tags:["open","focused"],effects:["trackDismissableLayer","scrollToHighlightedItem","trackPlacement","hideOtherElements"],entry:["setInitialFocus"],on:{"CONTROLLED.CLOSE":[{guard:"restoreFocus",target:"focused",actions:["setFinalFocus"]},{target:"idle"}],CHILDREN_CHANGE:[{guard:"autoHighlight",actions:["highlightFirstItem"]},{guard:"isHighlightedItemRemoved",actions:["clearHighlightedValue"]}],"INPUT.ARROW_DOWN":{target:"interacting",actions:["highlightNextItem"]},"INPUT.ARROW_UP":{target:"interacting",actions:["highlightPrevItem"]},"INPUT.HOME":{target:"interacting",actions:["highlightFirstItem"]},"INPUT.END":{target:"interacting",actions:["highlightLastItem"]},"INPUT.ENTER":[{guard:Mm("isOpenControlled","isCustomValue",XE("hasHighlightedItem"),XE("allowCustomValue")),actions:["revertInputValue","invokeOnClose"]},{guard:Mm("isCustomValue",XE("hasHighlightedItem"),XE("allowCustomValue")),target:"focused",actions:["revertInputValue","invokeOnClose"]},{guard:Mm("isOpenControlled","closeOnSelect"),actions:["selectHighlightedItem","invokeOnClose"]},{guard:"closeOnSelect",target:"focused",actions:["selectHighlightedItem","invokeOnClose","setFinalFocus"]},{actions:["selectHighlightedItem"]}],"INPUT.CHANGE":{actions:["setInputValue"]},"LAYER.ESCAPE":[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"focused",actions:["invokeOnClose"]}],"ITEM.POINTER_MOVE":{target:"interacting",actions:["setHighlightedValue"]},"ITEM.POINTER_LEAVE":{actions:["clearHighlightedValue"]},"LAYER.INTERACT_OUTSIDE":[{guard:Mm("isOpenControlled","isCustomValue",XE("allowCustomValue")),actions:["revertInputValue","invokeOnClose"]},{guard:Mm("isCustomValue",XE("allowCustomValue")),target:"idle",actions:["revertInputValue","invokeOnClose"]},{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"idle",actions:["invokeOnClose"]}],"TRIGGER.CLICK":[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"focused",actions:["invokeOnClose"]}],"ITEM.CLICK":[{guard:Mm("isOpenControlled","closeOnSelect"),actions:["selectItem","invokeOnClose"]},{guard:"closeOnSelect",target:"focused",actions:["selectItem","invokeOnClose","setFinalFocus"]},{actions:["selectItem"]}],CLOSE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"focused",actions:["invokeOnClose","setFinalFocus"]}],"VALUE.CLEAR":[{guard:"isOpenControlled",actions:["clearInputValue","clearSelectedItems","invokeOnClose"]},{target:"focused",actions:["clearInputValue","clearSelectedItems","invokeOnClose","setFinalFocus"]}]}}},implementations:{guards:{isInputValueEmpty:({computed:n})=>n("isInputValueEmpty"),autoComplete:({computed:n,prop:r})=>n("autoComplete")&&!r("multiple"),autoHighlight:({computed:n})=>n("autoHighlight"),isFirstItemHighlighted:({prop:n,context:r})=>n("collection").firstValue===r.get("highlightedValue"),isLastItemHighlighted:({prop:n,context:r})=>n("collection").lastValue===r.get("highlightedValue"),isCustomValue:({computed:n})=>n("isCustomValue"),allowCustomValue:({prop:n})=>!!n("allowCustomValue"),hasHighlightedItem:({context:n})=>n.get("highlightedValue")!=null,closeOnSelect:({prop:n})=>!!n("closeOnSelect"),isOpenControlled:({prop:n})=>n("open")!=null,openOnChange:({prop:n,context:r})=>{const o=n("openOnChange");return MFn(o)?o:!!o?.({inputValue:r.get("inputValue")})},restoreFocus:({event:n})=>n.restoreFocus==null?!0:!!n.restoreFocus,isChangeEvent:({event:n})=>n.previousEvent?.type==="INPUT.CHANGE",autoFocus:({prop:n})=>!!n("autoFocus"),isHighlightedItemRemoved:({prop:n,context:r})=>!n("collection").has(r.get("highlightedValue"))},effects:{trackDismissableLayer({send:n,prop:r,scope:o}){return r("disableLayer")?void 0:cU(()=>YD(o),{defer:!0,exclude:()=>[rW(o),Rre(o),E9t(o)],onFocusOutside:r("onFocusOutside"),onPointerDownOutside:r("onPointerDownOutside"),onInteractOutside:r("onInteractOutside"),onEscapeKeyDown(c){c.preventDefault(),c.stopPropagation(),n({type:"LAYER.ESCAPE",src:"escape-key"})},onDismiss(){n({type:"LAYER.INTERACT_OUTSIDE",src:"interact-outside",restoreFocus:!1})}})},hideOtherElements({scope:n}){return sYe([rW(n),YD(n),Rre(n),E9t(n)])},trackPlacement({context:n,prop:r,scope:o}){const s=()=>y9t(o)||Rre(o),c=()=>w9t(o);return n.set("currentPlacement",r("positioning").placement),g2(s,c,{...r("positioning"),defer:!0,onComplete(d){n.set("currentPlacement",d.placement)}})},scrollToHighlightedItem({context:n,prop:r,scope:o,event:s}){const c=rW(o);let d=[];const h=E=>{const k=s.current().type.includes("POINTER"),x=n.get("highlightedValue");if(k||!x)return;const O=YD(o),P=r("scrollToIndexFn");if(P){const U=r("collection").indexOf(x);P({index:U,immediate:E,getElement:()=>nre(o,x)});return}const D=nre(o,x),$=ca(()=>{$ve(D,{rootEl:O,block:"nearest"})});d.push($)},b=ca(()=>h(!0));d.push(b);const y=koe(c,{attributes:["aria-activedescendant"],callback:()=>h(!1)});return d.push(y),()=>{d.forEach(E=>E())}}},actions:{reposition({context:n,prop:r,scope:o,event:s}){g2(()=>y9t(o),()=>w9t(o),{...r("positioning"),...s.options,defer:!0,listeners:!1,onComplete(h){n.set("currentPlacement",h.placement)}})},setHighlightedValue({context:n,event:r}){r.value!=null&&n.set("highlightedValue",r.value)},clearHighlightedValue({context:n}){n.set("highlightedValue",null)},selectHighlightedItem(n){const{context:r,prop:o}=n,s=o("collection"),c=r.get("highlightedValue");if(!c||!s.has(c))return;const d=o("multiple")?GW(r.get("value"),c):[c];o("onSelect")?.({value:d,itemValue:c}),r.set("value",d);const h=_4(o("selectionBehavior"),{preserve:r.get("inputValue"),replace:s.stringifyMany(d),clear:""});r.set("inputValue",h)},scrollToHighlightedItem({context:n,prop:r,scope:o}){lye(()=>{const s=n.get("highlightedValue");if(s==null)return;const c=nre(o,s),d=YD(o),h=r("scrollToIndexFn");if(h){const b=r("collection").indexOf(s);h({index:b,immediate:!0,getElement:()=>nre(o,s)});return}$ve(c,{rootEl:d,block:"nearest"})})},selectItem(n){const{context:r,event:o,flush:s,prop:c}=n;o.value!=null&&s(()=>{const d=c("multiple")?GW(r.get("value"),o.value):[o.value];c("onSelect")?.({value:d,itemValue:o.value}),r.set("value",d);const h=_4(c("selectionBehavior"),{preserve:r.get("inputValue"),replace:c("collection").stringifyMany(d),clear:""});r.set("inputValue",h)})},clearItem(n){const{context:r,event:o,flush:s,prop:c}=n;o.value!=null&&s(()=>{const d=$B(r.get("value"),o.value);r.set("value",d);const h=_4(c("selectionBehavior"),{preserve:r.get("inputValue"),replace:c("collection").stringifyMany(d),clear:""});r.set("inputValue",h)})},setInitialFocus({scope:n}){ca(()=>{S9t(n)})},setFinalFocus({scope:n}){ca(()=>{Rre(n)?.dataset.focusable==null?S9t(n):cqn(n)})},syncInputValue({context:n,scope:r,event:o}){const s=rW(r);s&&(s.value=n.get("inputValue"),queueMicrotask(()=>{o.current().type!=="INPUT.CHANGE"&&sBn(s)}))},setInputValue({context:n,event:r}){n.set("inputValue",r.value)},clearInputValue({context:n}){n.set("inputValue","")},revertInputValue({context:n,prop:r,computed:o}){const s=r("selectionBehavior"),c=_4(s,{replace:o("hasSelectedItems")?o("valueAsString"):"",preserve:n.get("inputValue"),clear:""});n.set("inputValue",c)},setValue(n){const{context:r,flush:o,event:s,prop:c}=n;o(()=>{r.set("value",s.value);const d=_4(c("selectionBehavior"),{preserve:r.get("inputValue"),replace:c("collection").stringifyMany(s.value),clear:""});r.set("inputValue",d)})},clearSelectedItems(n){const{context:r,flush:o,prop:s}=n;o(()=>{r.set("value",[]);const c=_4(s("selectionBehavior"),{preserve:r.get("inputValue"),replace:s("collection").stringifyMany([]),clear:""});r.set("inputValue",c)})},scrollContentToTop({prop:n,scope:r}){const o=n("scrollToIndexFn");if(o){const s=n("collection").firstValue;o({index:0,immediate:!0,getElement:()=>nre(r,s)})}else{const s=YD(r);if(!s)return;s.scrollTop=0}},invokeOnOpen({prop:n,event:r}){const o=_9t(r);n("onOpenChange")?.({open:!0,reason:o})},invokeOnClose({prop:n,event:r}){const o=_9t(r);n("onOpenChange")?.({open:!1,reason:o})},highlightFirstItem({context:n,prop:r,scope:o}){(YD(o)?queueMicrotask:ca)(()=>{const c=r("collection").firstValue;c&&n.set("highlightedValue",c)})},highlightFirstItemIfNeeded({computed:n,action:r}){n("autoHighlight")&&r(["highlightFirstItem"])},highlightLastItem({context:n,prop:r,scope:o}){(YD(o)?queueMicrotask:ca)(()=>{const c=r("collection").lastValue;c&&n.set("highlightedValue",c)})},highlightNextItem({context:n,prop:r}){let o=null;const s=n.get("highlightedValue"),c=r("collection");s?(o=c.getNextValue(s),!o&&r("loopFocus")&&(o=c.firstValue)):o=c.firstValue,o&&n.set("highlightedValue",o)},highlightPrevItem({context:n,prop:r}){let o=null;const s=n.get("highlightedValue"),c=r("collection");s?(o=c.getPreviousValue(s),!o&&r("loopFocus")&&(o=c.lastValue)):o=c.lastValue,o&&n.set("highlightedValue",o)},highlightFirstSelectedItem({context:n,prop:r}){ca(()=>{const[o]=r("collection").sort(n.get("value"));o&&n.set("highlightedValue",o)})},highlightFirstOrSelectedItem({context:n,prop:r,computed:o}){ca(()=>{let s=null;o("hasSelectedItems")?s=r("collection").sort(n.get("value"))[0]:s=r("collection").firstValue,s&&n.set("highlightedValue",s)})},highlightLastOrSelectedItem({context:n,prop:r,computed:o}){ca(()=>{const s=r("collection");let c=null;o("hasSelectedItems")?c=s.sort(n.get("value"))[0]:c=s.lastValue,c&&n.set("highlightedValue",c)})},autofillInputValue({context:n,computed:r,prop:o,event:s,scope:c}){const d=rW(c),h=o("collection");if(!r("autoComplete")||!d||!s.keypress)return;const b=h.stringify(n.get("highlightedValue"));ca(()=>{d.value=b||n.get("inputValue")})},syncSelectedItems(n){queueMicrotask(()=>{const{context:r,prop:o}=n,s=o("collection"),c=r.get("value"),d=c.map(b=>r.get("selectedItems").find(E=>s.getItemValue(E)===b)||s.find(b));r.set("selectedItems",d);const h=_4(o("selectionBehavior"),{preserve:r.get("inputValue"),replace:s.stringifyMany(c),clear:""});r.set("inputValue",h)})},syncHighlightedItem({context:n,prop:r}){const o=r("collection").find(n.get("highlightedValue"));n.set("highlightedItem",o)},toggleVisibility({event:n,send:r,prop:o}){r({type:o("open")?"CONTROLLED.OPEN":"CONTROLLED.CLOSE",previousEvent:n})}}}});function _9t(n){return(n.previousEvent||n).src}Wo()(["allowCustomValue","autoFocus","closeOnSelect","collection","composite","defaultHighlightedValue","defaultInputValue","defaultOpen","defaultValue","dir","disabled","disableLayer","form","getRootNode","highlightedValue","id","ids","inputBehavior","inputValue","invalid","loopFocus","multiple","name","navigate","onFocusOutside","onHighlightChange","onInputValueChange","onInteractOutside","onOpenChange","onOpenChange","onPointerDownOutside","onSelect","onValueChange","open","openOnChange","openOnClick","openOnKeyPress","placeholder","positioning","readOnly","required","scrollToIndexFn","selectionBehavior","translations","value"]);Wo()(["htmlFor"]);Wo()(["id"]);Wo()(["item","persistFocus"]);const fqn=vXt.extendWith("empty"),[yXt,FM]=fu({name:"DialogContext",hookName:"useDialogContext",providerName:"<DialogProvider />"}),EXt=Q.forwardRef((n,r)=>{const o=FM(),s=GYt(),c=J4({...s,present:o.open}),d=Fn(o.getBackdropProps(),c.getPresenceProps(),n);return c.unmounted?null:H.jsx(Xn.div,{...d,ref:X4(c.ref,r)})});EXt.displayName="DialogBackdrop";const SXt=Q.forwardRef((n,r)=>{const o=FM(),s=Fn(o.getCloseTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});SXt.displayName="DialogCloseTrigger";const _Xt=Q.forwardRef((n,r)=>{const o=FM(),s=Yx(),c=Fn(o.getContentProps(),s.getPresenceProps(),n);return s.unmounted?null:H.jsx(Xn.div,{...c,ref:X4(s.ref,r)})});_Xt.displayName="DialogContent";const kXt=Q.forwardRef((n,r)=>{const o=FM(),s=Fn(o.getDescriptionProps(),n);return H.jsx(Xn.div,{...s,ref:r})});kXt.displayName="DialogDescription";const TXt=Q.forwardRef((n,r)=>{const o=FM(),s=Fn(o.getPositionerProps(),n);return Yx().unmounted?null:H.jsx(Xn.div,{...s,ref:r})});TXt.displayName="DialogPositioner";var hqn=Object.defineProperty,pqn=(n,r,o)=>r in n?hqn(n,r,{enumerable:!0,configurable:!0,writable:!0,value:o}):n[r]=o,Zf=(n,r,o)=>pqn(n,typeof r!="symbol"?r+"":r,o),k9t={activateTrap(n,r){if(n.length>0){const s=n[n.length-1];s!==r&&s.pause()}const o=n.indexOf(r);o===-1||n.splice(o,1),n.push(r)},deactivateTrap(n,r){const o=n.indexOf(r);o!==-1&&n.splice(o,1),n.length>0&&n[n.length-1].unpause()}},gqn=[],mqn=class{constructor(n,r){Zf(this,"trapStack"),Zf(this,"config"),Zf(this,"doc"),Zf(this,"state",{containers:[],containerGroups:[],tabbableGroups:[],nodeFocusedBeforeActivation:null,mostRecentlyFocusedNode:null,active:!1,paused:!1,delayInitialFocusTimer:void 0,recentNavEvent:void 0}),Zf(this,"listenerCleanups",[]),Zf(this,"handleFocus",s=>{const c=Yp(s),d=this.findContainerIndex(c,s)>=0;if(d||FKe(c))d&&(this.state.mostRecentlyFocusedNode=c);else{s.stopImmediatePropagation();let h,b=!0;if(this.state.mostRecentlyFocusedNode)if(Qne(this.state.mostRecentlyFocusedNode)>0){const y=this.findContainerIndex(this.state.mostRecentlyFocusedNode),{tabbableNodes:E}=this.state.containerGroups[y];if(E.length>0){const k=E.findIndex(x=>x===this.state.mostRecentlyFocusedNode);k>=0&&(this.config.isKeyForward(this.state.recentNavEvent)?k+1<E.length&&(h=E[k+1],b=!1):k-1>=0&&(h=E[k-1],b=!1))}}else this.state.containerGroups.some(y=>y.tabbableNodes.some(E=>Qne(E)>0))||(b=!1);else b=!1;b&&(h=this.findNextNavNode({target:this.state.mostRecentlyFocusedNode,isBackward:this.config.isKeyBackward(this.state.recentNavEvent)})),h?this.tryFocus(h):this.tryFocus(this.state.mostRecentlyFocusedNode||this.getInitialFocusNode())}this.state.recentNavEvent=void 0}),Zf(this,"handlePointerDown",s=>{const c=Yp(s);if(!(this.findContainerIndex(c,s)>=0)){if(ire(this.config.clickOutsideDeactivates,s)){this.deactivate({returnFocus:this.config.returnFocusOnDeactivate});return}ire(this.config.allowOutsideClick,s)||s.preventDefault()}}),Zf(this,"handleClick",s=>{const c=Yp(s);this.findContainerIndex(c,s)>=0||ire(this.config.clickOutsideDeactivates,s)||ire(this.config.allowOutsideClick,s)||(s.preventDefault(),s.stopImmediatePropagation())}),Zf(this,"handleTabKey",s=>{if(this.config.isKeyForward(s)||this.config.isKeyBackward(s)){this.state.recentNavEvent=s;const c=this.config.isKeyBackward(s),d=this.findNextNavNode({event:s,isBackward:c});if(!d)return;rre(s)&&s.preventDefault(),this.tryFocus(d)}}),Zf(this,"handleEscapeKey",s=>{bqn(s)&&ire(this.config.escapeDeactivates,s)!==!1&&(s.preventDefault(),this.deactivate())}),Zf(this,"_mutationObserver"),Zf(this,"setupMutationObserver",()=>{const s=this.doc.defaultView||window;this._mutationObserver=new s.MutationObserver(c=>{c.some(h=>Array.from(h.removedNodes).some(y=>y===this.state.mostRecentlyFocusedNode))&&this.tryFocus(this.getInitialFocusNode())})}),Zf(this,"updateObservedNodes",()=>{this._mutationObserver?.disconnect(),this.state.active&&!this.state.paused&&this.state.containers.map(s=>{this._mutationObserver?.observe(s,{subtree:!0,childList:!0})})}),Zf(this,"getInitialFocusNode",()=>{let s=this.getNodeForOption("initialFocus",{hasFallback:!0});if(s===!1)return!1;if(s===void 0||s&&!hM(s))if(this.findContainerIndex(this.doc.activeElement)>=0)s=this.doc.activeElement;else{const c=this.state.tabbableGroups[0];s=c&&c.firstTabbableNode||this.getNodeForOption("fallbackFocus")}else s===null&&(s=this.getNodeForOption("fallbackFocus"));if(!s)throw new Error("Your focus-trap needs to have at least one focusable element");return s.isConnected||(s=this.getNodeForOption("fallbackFocus")),s}),Zf(this,"tryFocus",s=>{if(s!==!1&&s!==lYt(this.doc)){if(!s||!s.focus){this.tryFocus(this.getInitialFocusNode());return}s.focus({preventScroll:!!this.config.preventScroll}),this.state.mostRecentlyFocusedNode=s,vqn(s)&&s.select()}}),Zf(this,"deactivate",s=>{if(!this.state.active)return this;const c={onDeactivate:this.config.onDeactivate,onPostDeactivate:this.config.onPostDeactivate,checkCanReturnFocus:this.config.checkCanReturnFocus,...s};clearTimeout(this.state.delayInitialFocusTimer),this.state.delayInitialFocusTimer=void 0,this.removeListeners(),this.state.active=!1,this.state.paused=!1,this.updateObservedNodes(),k9t.deactivateTrap(this.trapStack,this);const d=this.getOption(c,"onDeactivate"),h=this.getOption(c,"onPostDeactivate"),b=this.getOption(c,"checkCanReturnFocus"),y=this.getOption(c,"returnFocus","returnFocusOnDeactivate");d?.();const E=()=>{T9t(()=>{if(y){const k=this.getReturnFocusNode(this.state.nodeFocusedBeforeActivation);this.tryFocus(k)}h?.()})};if(y&&b){const k=this.getReturnFocusNode(this.state.nodeFocusedBeforeActivation);return b(k).then(E,E),this}return E(),this}),Zf(this,"pause",s=>{if(this.state.paused||!this.state.active)return this;const c=this.getOption(s,"onPause"),d=this.getOption(s,"onPostPause");return this.state.paused=!0,c?.(),this.removeListeners(),this.updateObservedNodes(),d?.(),this}),Zf(this,"unpause",s=>{if(!this.state.paused||!this.state.active)return this;const c=this.getOption(s,"onUnpause"),d=this.getOption(s,"onPostUnpause");return this.state.paused=!1,c?.(),this.updateTabbableNodes(),this.addListeners(),this.updateObservedNodes(),d?.(),this}),Zf(this,"updateContainerElements",s=>(this.state.containers=Array.isArray(s)?s.filter(Boolean):[s].filter(Boolean),this.state.active&&this.updateTabbableNodes(),this.updateObservedNodes(),this)),Zf(this,"getReturnFocusNode",s=>{const c=this.getNodeForOption("setReturnFocus",{params:[s]});return c||(c===!1?!1:s)}),Zf(this,"getOption",(s,c,d)=>s&&s[c]!==void 0?s[c]:this.config[d||c]),Zf(this,"getNodeForOption",(s,{hasFallback:c=!1,params:d=[]}={})=>{let h=this.config[s];if(typeof h=="function"&&(h=h(...d)),h===!0&&(h=void 0),!h){if(h===void 0||h===!1)return h;throw new Error(`\`${s}\` was specified but was not a node, or did not return a node`)}let b=h;if(typeof h=="string"){try{b=this.doc.querySelector(h)}catch(y){throw new Error(`\`${s}\` appears to be an invalid selector; error="${y.message}"`)}if(!b&&!c)throw new Error(`\`${s}\` as selector refers to no known node`)}return b}),Zf(this,"findNextNavNode",s=>{const{event:c,isBackward:d=!1}=s,h=s.target||Yp(c);this.updateTabbableNodes();let b=null;if(this.state.tabbableGroups.length>0){const y=this.findContainerIndex(h,c),E=y>=0?this.state.containerGroups[y]:void 0;if(y<0)d?b=this.state.tabbableGroups[this.state.tabbableGroups.length-1].lastTabbableNode:b=this.state.tabbableGroups[0].firstTabbableNode;else if(d){let k=this.state.tabbableGroups.findIndex(({firstTabbableNode:x})=>h===x);if(k<0&&(E?.container===h||hM(h)&&!nM(h)&&!E?.nextTabbableNode(h,!1))&&(k=y),k>=0){const x=k===0?this.state.tabbableGroups.length-1:k-1,O=this.state.tabbableGroups[x];b=Qne(h)>=0?O.lastTabbableNode:O.lastDomTabbableNode}else rre(c)||(b=E?.nextTabbableNode(h,!1))}else{let k=this.state.tabbableGroups.findIndex(({lastTabbableNode:x})=>h===x);if(k<0&&(E?.container===h||hM(h)&&!nM(h)&&!E?.nextTabbableNode(h))&&(k=y),k>=0){const x=k===this.state.tabbableGroups.length-1?0:k+1,O=this.state.tabbableGroups[x];b=Qne(h)>=0?O.firstTabbableNode:O.firstDomTabbableNode}else rre(c)||(b=E?.nextTabbableNode(h))}}else b=this.getNodeForOption("fallbackFocus");return b}),this.trapStack=r.trapStack||gqn;const o={returnFocusOnDeactivate:!0,escapeDeactivates:!0,delayInitialFocus:!0,isKeyForward(s){return rre(s)&&!s.shiftKey},isKeyBackward(s){return rre(s)&&s.shiftKey},...r};this.doc=o.document||ny(Array.isArray(n)?n[0]:n),this.config=o,this.updateContainerElements(n),this.setupMutationObserver()}get active(){return this.state.active}get paused(){return this.state.paused}findContainerIndex(n,r){const o=typeof r?.composedPath=="function"?r.composedPath():void 0;return this.state.containerGroups.findIndex(({container:s,tabbableNodes:c})=>s.contains(n)||o?.includes(s)||c.find(d=>d===n))}updateTabbableNodes(){if(this.state.containerGroups=this.state.containers.map(n=>{const r=_oe(n),o=mYt(n),s=r.length>0?r[0]:void 0,c=r.length>0?r[r.length-1]:void 0,d=o.find(E=>nM(E)),h=o.slice().reverse().find(E=>nM(E)),b=!!r.find(E=>Qne(E)>0);function y(E,k=!0){const x=r.indexOf(E);return x<0?k?o.slice(o.indexOf(E)+1).find(O=>nM(O)):o.slice(0,o.indexOf(E)).reverse().find(O=>nM(O)):r[x+(k?1:-1)]}return{container:n,tabbableNodes:r,focusableNodes:o,posTabIndexesFound:b,firstTabbableNode:s,lastTabbableNode:c,firstDomTabbableNode:d,lastDomTabbableNode:h,nextTabbableNode:y}}),this.state.tabbableGroups=this.state.containerGroups.filter(n=>n.tabbableNodes.length>0),this.state.tabbableGroups.length<=0&&!this.getNodeForOption("fallbackFocus"))throw new Error("Your focus-trap must have at least one container with at least one tabbable node in it at all times");if(this.state.containerGroups.find(n=>n.posTabIndexesFound)&&this.state.containerGroups.length>1)throw new Error("At least one node with a positive tabindex was found in one of your focus-trap's multiple containers. Positive tabindexes are only supported in single-container focus-traps.")}addListeners(){if(this.state.active)return k9t.activateTrap(this.trapStack,this),this.state.delayInitialFocusTimer=this.config.delayInitialFocus?T9t(()=>{this.tryFocus(this.getInitialFocusNode())}):this.tryFocus(this.getInitialFocusNode()),this.listenerCleanups.push(Ul(this.doc,"focusin",this.handleFocus,!0),Ul(this.doc,"mousedown",this.handlePointerDown,{capture:!0,passive:!1}),Ul(this.doc,"touchstart",this.handlePointerDown,{capture:!0,passive:!1}),Ul(this.doc,"click",this.handleClick,{capture:!0,passive:!1}),Ul(this.doc,"keydown",this.handleTabKey,{capture:!0,passive:!1}),Ul(this.doc,"keydown",this.handleEscapeKey)),this}removeListeners(){if(this.state.active)return this.listenerCleanups.forEach(n=>n()),this.listenerCleanups=[],this}activate(n){if(this.state.active)return this;const r=this.getOption(n,"onActivate"),o=this.getOption(n,"onPostActivate"),s=this.getOption(n,"checkCanFocusTrap");s||this.updateTabbableNodes(),this.state.active=!0,this.state.paused=!1,this.state.nodeFocusedBeforeActivation=this.doc.activeElement||null,r?.();const c=()=>{s&&this.updateTabbableNodes(),this.addListeners(),this.updateObservedNodes(),o?.()};return s?(s(this.state.containers.concat()).then(c,c),this):(c(),this)}},rre=n=>n.key==="Tab",ire=(n,...r)=>typeof n=="function"?n(...r):n,bqn=n=>!n.isComposing&&n.key==="Escape",T9t=n=>setTimeout(n,0),vqn=n=>n.localName==="input"&&"select"in n&&typeof n.select=="function";function xXt(n,r={}){let o;const s=ca(()=>{const c=typeof n=="function"?n():n;if(c){o=new mqn(c,{escapeDeactivates:!1,allowOutsideClick:!0,preventScroll:!0,returnFocusOnDeactivate:!0,delayInitialFocus:!1,fallbackFocus:c,...r,document:ny(c)});try{o.activate()}catch{}}});return function(){o?.deactivate(),s()}}var zje="data-scroll-lock";function wqn(n){const r=n.getBoundingClientRect().left;return Math.round(r)+n.scrollLeft?"paddingLeft":"paddingRight"}function CXt(n){const r=n??document,o=r.defaultView??window,{documentElement:s,body:c}=r;if(c.hasAttribute(zje))return;const h=o.innerWidth-s.clientWidth;c.setAttribute(zje,"");const b=()=>hUn(s,"--scrollbar-width",`${h}px`),y=wqn(s),E=()=>Lie(c,{overflow:"hidden",[y]:`${h}px`}),k=()=>{const{scrollX:O,scrollY:P,visualViewport:D}=o,$=D?.offsetLeft??0,U=D?.offsetTop??0,B=Lie(c,{position:"fixed",overflow:"hidden",top:`${-(P-Math.floor(U))}px`,left:`${-(O-Math.floor($))}px`,right:"0",[y]:`${h}px`});return()=>{B?.(),o.scrollTo({left:O,top:P,behavior:"instant"})}},x=[b(),sye()?k():E()];return()=>{x.forEach(O=>O?.()),c.removeAttribute(zje)}}var cYe=Vo("dialog").parts("trigger","backdrop","positioner","content","title","description","closeTrigger"),tB=cYe.build(),AXt=n=>n.ids?.positioner??`dialog:${n.id}:positioner`,IXt=n=>n.ids?.backdrop??`dialog:${n.id}:backdrop`,Wqe=n=>n.ids?.content??`dialog:${n.id}:content`,OXt=n=>n.ids?.trigger??`dialog:${n.id}:trigger`,Kqe=n=>n.ids?.title??`dialog:${n.id}:title`,Yqe=n=>n.ids?.description??`dialog:${n.id}:description`,NXt=n=>n.ids?.closeTrigger??`dialog:${n.id}:close`,$1e=n=>n.getById(Wqe(n)),yqn=n=>n.getById(AXt(n)),Eqn=n=>n.getById(IXt(n)),Sqn=n=>n.getById(OXt(n)),_qn=n=>n.getById(Kqe(n)),kqn=n=>n.getById(Yqe(n)),Tqn=n=>n.getById(NXt(n));function xqn(n,r){const{state:o,send:s,context:c,prop:d,scope:h}=n,b=d("aria-label"),y=o.matches("open");return{open:y,setOpen(E){o.matches("open")!==E&&s({type:E?"OPEN":"CLOSE"})},getTriggerProps(){return r.button({...tB.trigger.attrs,dir:d("dir"),id:OXt(h),"aria-haspopup":"dialog",type:"button","aria-expanded":y,"data-state":y?"open":"closed","aria-controls":Wqe(h),onClick(E){E.defaultPrevented||s({type:"TOGGLE"})}})},getBackdropProps(){return r.element({...tB.backdrop.attrs,dir:d("dir"),hidden:!y,id:IXt(h),"data-state":y?"open":"closed"})},getPositionerProps(){return r.element({...tB.positioner.attrs,dir:d("dir"),id:AXt(h),style:{pointerEvents:y?void 0:"none"}})},getContentProps(){const E=c.get("rendered");return r.element({...tB.content.attrs,dir:d("dir"),role:d("role"),hidden:!y,id:Wqe(h),tabIndex:-1,"data-state":y?"open":"closed","aria-modal":!0,"aria-label":b||void 0,"aria-labelledby":b||!E.title?void 0:Kqe(h),"aria-describedby":E.description?Yqe(h):void 0})},getTitleProps(){return r.element({...tB.title.attrs,dir:d("dir"),id:Kqe(h)})},getDescriptionProps(){return r.element({...tB.description.attrs,dir:d("dir"),id:Yqe(h)})},getCloseTriggerProps(){return r.button({...tB.closeTrigger.attrs,dir:d("dir"),id:NXt(h),type:"button",onClick(E){E.defaultPrevented||(E.stopPropagation(),s({type:"CLOSE"}))}})}}}var Cqn={props({props:n,scope:r}){const o=n.role==="alertdialog",s=o?()=>Tqn(r):void 0,c=typeof n.modal=="boolean"?n.modal:!0;return{role:"dialog",modal:c,trapFocus:c,preventScroll:c,closeOnInteractOutside:!o,closeOnEscape:!0,restoreFocus:!0,initialFocusEl:s,...n}},initialState({prop:n}){return n("open")||n("defaultOpen")?"open":"closed"},context({bindable:n}){return{rendered:n(()=>({defaultValue:{title:!0,description:!0}}))}},watch({track:n,action:r,prop:o}){n([()=>o("open")],()=>{r(["toggleVisibility"])})},states:{open:{entry:["checkRenderedElements","syncZIndex"],effects:["trackDismissableElement","trapFocus","preventScroll","hideContentBelow"],on:{"CONTROLLED.CLOSE":{target:"closed"},CLOSE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],TOGGLE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}]}},closed:{on:{"CONTROLLED.OPEN":{target:"open"},OPEN:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}],TOGGLE:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}]}}},implementations:{guards:{isOpenControlled:({prop:n})=>n("open")!=null},effects:{trackDismissableElement({scope:n,send:r,prop:o}){return cU(()=>$1e(n),{defer:!0,pointerBlocking:o("modal"),exclude:[Sqn(n)],onInteractOutside(c){o("onInteractOutside")?.(c),o("closeOnInteractOutside")||c.preventDefault()},persistentElements:o("persistentElements"),onFocusOutside:o("onFocusOutside"),onPointerDownOutside:o("onPointerDownOutside"),onRequestDismiss:o("onRequestDismiss"),onEscapeKeyDown(c){o("onEscapeKeyDown")?.(c),o("closeOnEscape")||c.preventDefault()},onDismiss(){r({type:"CLOSE",src:"interact-outside"})}})},preventScroll({scope:n,prop:r}){if(r("preventScroll"))return CXt(n.getDoc())},trapFocus({scope:n,prop:r}){return r("trapFocus")?xXt(()=>$1e(n),{preventScroll:!0,returnFocusOnDeactivate:!!r("restoreFocus"),initialFocus:r("initialFocusEl"),setReturnFocus:s=>r("finalFocusEl")?.()??s}):void 0},hideContentBelow({scope:n,prop:r}){return r("modal")?sYe(()=>[$1e(n)],{defer:!0}):void 0}},actions:{checkRenderedElements({context:n,scope:r}){ca(()=>{n.set("rendered",{title:!!_qn(r),description:!!kqn(r)})})},syncZIndex({scope:n}){ca(()=>{const r=$1e(n);if(!r)return;const o=qB(r);[yqn(n),Eqn(n)].forEach(c=>{c?.style.setProperty("--z-index",o.zIndex),c?.style.setProperty("--layer-index",o.getPropertyValue("--layer-index"))})})},invokeOnClose({prop:n}){n("onOpenChange")?.({open:!1})},invokeOnOpen({prop:n}){n("onOpenChange")?.({open:!0})},toggleVisibility({prop:n,send:r,event:o}){r({type:n("open")?"CONTROLLED.OPEN":"CONTROLLED.CLOSE",previousEvent:o})}}}};Wo()(["aria-label","closeOnEscape","closeOnInteractOutside","dir","finalFocusEl","getRootNode","getRootNode","id","id","ids","initialFocusEl","modal","onEscapeKeyDown","onFocusOutside","onInteractOutside","onOpenChange","onPointerDownOutside","onRequestDismiss","defaultOpen","open","persistentElements","preventScroll","restoreFocus","role","trapFocus"]);const Aqn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,getRootNode:o,dir:s,...n},d=uv(Cqn,c);return xqn(d,ry)},Iqn=n=>{const[r,{children:o,...s}]=MM(n),[c]=bye(r),d=Aqn(s),h=J4(Fn({present:d.open},r));return H.jsx(yXt,{value:d,children:H.jsx(mye,{value:c,children:H.jsx(jM,{value:h,children:o})})})},Oqn=n=>{const[r,{value:o,children:s}]=MM(n),[c]=bye(r),d=J4(Fn({present:o.open},r));return H.jsx(yXt,{value:o,children:H.jsx(mye,{value:c,children:H.jsx(jM,{value:d,children:s})})})},RXt=Q.forwardRef((n,r)=>{const o=FM(),s=Fn(o.getTitleProps(),n);return H.jsx(Xn.h2,{...s,ref:r})});RXt.displayName="DialogTitle";const PXt=Q.forwardRef((n,r)=>{const o=FM(),s=Yx(),c=Fn({...o.getTriggerProps(),"aria-controls":s.unmounted?void 0:o.getTriggerProps()["aria-controls"]},n);return H.jsx(Xn.button,{...c,ref:r})});PXt.displayName="DialogTrigger";const[LXt,BM]=fu({name:"EditableContext",hookName:"useEditableContext",providerName:"<EditableProvider />"}),DXt=Q.forwardRef((n,r)=>{const o=BM(),s=Fn(o.getAreaProps(),n);return H.jsx(Xn.div,{...s,ref:r})});DXt.displayName="EditableArea";const MXt=Q.forwardRef((n,r)=>{const o=BM(),s=Fn(o.getCancelTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});MXt.displayName="EditableCancelTrigger";const jXt=Q.forwardRef((n,r)=>{const o=BM(),s=Fn(o.getControlProps(),n);return H.jsx(Xn.div,{...s,ref:r})});jXt.displayName="EditableControl";const $Xt=Q.forwardRef((n,r)=>{const o=BM(),s=Fn(o.getEditTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});$Xt.displayName="EditableEditTrigger";const FXt=Q.forwardRef((n,r)=>{const o=BM(),s=Fn(o.getInputProps(),n),c=$M();return H.jsx(Xn.input,{"aria-describedby":c?.ariaDescribedby,...s,ref:r})});FXt.displayName="EditableInput";const BXt=Q.forwardRef((n,r)=>{const o=BM(),s=Fn(o.getPreviewProps(),n);return H.jsx(Xn.span,{...s,ref:r})});BXt.displayName="EditablePreview";var UXt=Vo("editable").parts("root","area","label","preview","input","editTrigger","submitTrigger","cancelTrigger","control"),v4=UXt.build(),Nqn=n=>n.ids?.root??`editable:${n.id}`,Rqn=n=>n.ids?.area??`editable:${n.id}:area`,Pqn=n=>n.ids?.label??`editable:${n.id}:label`,zXt=n=>n.ids?.preview??`editable:${n.id}:preview`,Xqe=n=>n.ids?.input??`editable:${n.id}:input`,Lqn=n=>n.ids?.control??`editable:${n.id}:control`,HXt=n=>n.ids?.submitTrigger??`editable:${n.id}:submit`,GXt=n=>n.ids?.cancelTrigger??`editable:${n.id}:cancel`,qXt=n=>n.ids?.editTrigger??`editable:${n.id}:edit`,F1e=n=>n.getById(Xqe(n)),Dqn=n=>n.getById(zXt(n)),Mqn=n=>n.getById(HXt(n)),jqn=n=>n.getById(GXt(n)),$qn=n=>n.getById(qXt(n));function Fqn(n,r){const{state:o,context:s,send:c,prop:d,scope:h,computed:b}=n,y=d("disabled"),E=b("isInteractive"),k=d("readOnly"),x=d("invalid"),O=d("autoResize"),P=d("translations"),D=o.matches("edit"),$=d("placeholder"),U=typeof $=="string"?{edit:$,preview:$}:$,B=s.get("value"),G=B.trim()==="",K=G?U?.preview??"":B;return{editing:D,empty:G,value:B,valueText:K,setValue(X){c({type:"VALUE.SET",value:X,src:"setValue"})},clearValue(){c({type:"VALUE.SET",value:"",src:"clearValue"})},edit(){E&&c({type:"EDIT"})},cancel(){E&&c({type:"CANCEL"})},submit(){E&&c({type:"SUBMIT"})},getRootProps(){return r.element({...v4.root.attrs,id:Nqn(h),dir:d("dir")})},getAreaProps(){return r.element({...v4.area.attrs,id:Rqn(h),dir:d("dir"),style:O?{display:"inline-grid"}:void 0,"data-focus":nr(D),"data-disabled":nr(y),"data-placeholder-shown":nr(G)})},getLabelProps(){return r.label({...v4.label.attrs,id:Pqn(h),dir:d("dir"),htmlFor:Xqe(h),"data-focus":nr(D),"data-invalid":nr(x),onClick(){if(D)return;Dqn(h)?.focus({preventScroll:!0})}})},getInputProps(){return r.input({...v4.input.attrs,dir:d("dir"),"aria-label":P?.input,name:d("name"),form:d("form"),id:Xqe(h),hidden:O?void 0:!D,placeholder:U?.edit,maxLength:d("maxLength"),required:d("required"),disabled:y,"data-disabled":nr(y),readOnly:k,"data-readonly":nr(k),"aria-invalid":gI(x),"data-invalid":nr(x),"data-autoresize":nr(O),defaultValue:B,size:O?1:void 0,onChange(X){c({type:"VALUE.SET",src:"input.change",value:X.currentTarget.value})},onKeyDown(X){if(X.defaultPrevented||UKe(X))return;const oe={Escape(){c({type:"CANCEL"}),X.preventDefault()},Enter(he){if(!b("submitOnEnter"))return;const{localName:ye}=he.currentTarget;if(ye==="textarea"){if(!(fYt()?he.metaKey:he.ctrlKey))return;c({type:"SUBMIT",src:"keydown.enter"});return}ye==="input"&&!he.shiftKey&&!he.metaKey&&(c({type:"SUBMIT",src:"keydown.enter"}),he.preventDefault())}}[X.key];oe&&oe(X)},style:O?{gridArea:"1 / 1 / auto / auto",visibility:D?void 0:"hidden"}:void 0})},getPreviewProps(){return r.element({id:zXt(h),...v4.preview.attrs,dir:d("dir"),"data-placeholder-shown":nr(G),"aria-readonly":gI(k),"data-readonly":nr(y),"data-disabled":nr(y),"aria-disabled":gI(y),"aria-invalid":gI(x),"data-invalid":nr(x),"aria-label":P?.edit,"data-autoresize":nr(O),children:K,hidden:O?void 0:D,tabIndex:E?0:void 0,onClick(){E&&d("activationMode")==="click"&&c({type:"EDIT",src:"click"})},onFocus(){E&&d("activationMode")==="focus"&&c({type:"EDIT",src:"focus"})},onDoubleClick(X){X.defaultPrevented||E&&d("activationMode")==="dblclick"&&c({type:"EDIT",src:"dblclick"})},style:O?{whiteSpace:"pre",userSelect:"none",gridArea:"1 / 1 / auto / auto",visibility:D?"hidden":void 0,overflow:"hidden",textOverflow:"ellipsis"}:void 0})},getEditTriggerProps(){return r.button({...v4.editTrigger.attrs,id:qXt(h),dir:d("dir"),"aria-label":P?.edit,hidden:D,type:"button",disabled:y,onClick(X){X.defaultPrevented||E&&c({type:"EDIT",src:"edit.click"})}})},getControlProps(){return r.element({id:Lqn(h),...v4.control.attrs,dir:d("dir")})},getSubmitTriggerProps(){return r.button({...v4.submitTrigger.attrs,dir:d("dir"),id:HXt(h),"aria-label":P?.submit,hidden:!D,disabled:y,type:"button",onClick(X){X.defaultPrevented||E&&c({type:"SUBMIT",src:"submit.click"})}})},getCancelTriggerProps(){return r.button({...v4.cancelTrigger.attrs,dir:d("dir"),"aria-label":P?.cancel,id:GXt(h),hidden:!D,type:"button",disabled:y,onClick(X){X.defaultPrevented||E&&c({type:"CANCEL",src:"cancel.click"})}})}}}var Bqn={props({props:n}){return{activationMode:"focus",submitMode:"both",defaultValue:"",selectOnFocus:!0,...n,translations:{input:"editable input",edit:"edit",submit:"submit",cancel:"cancel",...n.translations}}},initialState({prop:n}){return n("edit")||n("defaultEdit")?"edit":"preview"},entry:["focusInputIfNeeded"],context:({bindable:n,prop:r})=>({value:n(()=>({defaultValue:r("defaultValue"),value:r("value"),onChange(o){return r("onValueChange")?.({value:o})}})),previousValue:n(()=>({defaultValue:""}))}),watch({track:n,action:r,context:o,prop:s}){n([()=>o.get("value")],()=>{r(["syncInputValue"])}),n([()=>s("edit")],()=>{r(["toggleEditing"])})},computed:{submitOnEnter({prop:n}){const r=n("submitMode");return r==="both"||r==="enter"},submitOnBlur({prop:n}){const r=n("submitMode");return r==="both"||r==="blur"},isInteractive({prop:n}){return!(n("disabled")||n("readOnly"))}},on:{"VALUE.SET":{actions:["setValue"]}},states:{preview:{entry:["blurInputIfNeeded"],on:{"CONTROLLED.EDIT":{target:"edit",actions:["setPreviousValue","focusInput"]},EDIT:[{guard:"isEditControlled",actions:["invokeOnEdit"]},{target:"edit",actions:["setPreviousValue","focusInput","invokeOnEdit"]}]}},edit:{effects:["trackInteractOutside"],on:{"CONTROLLED.PREVIEW":[{guard:"isSubmitEvent",target:"preview",actions:["setPreviousValue","restoreFocus","invokeOnSubmit"]},{target:"preview",actions:["revertValue","restoreFocus","invokeOnCancel"]}],CANCEL:[{guard:"isEditControlled",actions:["invokeOnPreview"]},{target:"preview",actions:["revertValue","restoreFocus","invokeOnCancel","invokeOnPreview"]}],SUBMIT:[{guard:"isEditControlled",actions:["invokeOnPreview"]},{target:"preview",actions:["setPreviousValue","restoreFocus","invokeOnSubmit","invokeOnPreview"]}]}}},implementations:{guards:{isEditControlled:({prop:n})=>n("edit")!=null,isSubmitEvent:({event:n})=>n.previousEvent?.type==="SUBMIT"},effects:{trackInteractOutside({send:n,scope:r,prop:o,computed:s}){return zYt(F1e(r),{exclude(c){return[jqn(r),Mqn(r)].some(h=>f2(h,c))},onFocusOutside:o("onFocusOutside"),onPointerDownOutside:o("onPointerDownOutside"),onInteractOutside(c){if(o("onInteractOutside")?.(c),c.defaultPrevented)return;const{focusable:d}=c.detail;n({type:s("submitOnBlur")?"SUBMIT":"CANCEL",src:"interact-outside",focusable:d})}})}},actions:{restoreFocus({event:n,scope:r,prop:o}){n.focusable||ca(()=>{(o("finalFocusEl")?.()??$qn(r))?.focus({preventScroll:!0})})},clearValue({context:n}){n.set("value","")},focusInputIfNeeded({action:n,prop:r}){(r("edit")||r("defaultEdit"))&&n(["focusInput"])},focusInput({scope:n,prop:r}){ca(()=>{const o=F1e(n);o&&(r("selectOnFocus")?o.select():o.focus({preventScroll:!0}))})},invokeOnCancel({prop:n,context:r}){const o=r.get("previousValue");n("onValueRevert")?.({value:o})},invokeOnSubmit({prop:n,context:r}){const o=r.get("value");n("onValueCommit")?.({value:o})},invokeOnEdit({prop:n}){n("onEditChange")?.({edit:!0})},invokeOnPreview({prop:n}){n("onEditChange")?.({edit:!1})},toggleEditing({prop:n,send:r,event:o}){r({type:n("edit")?"CONTROLLED.EDIT":"CONTROLLED.PREVIEW",previousEvent:o})},syncInputValue({context:n,scope:r}){const o=F1e(r);o&&Soe(o,n.get("value"))},setValue({context:n,prop:r,event:o}){const s=r("maxLength"),c=s!=null?o.value.slice(0,s):o.value;n.set("value",c)},setPreviousValue({context:n}){n.set("previousValue",n.get("value"))},revertValue({context:n}){const r=n.get("previousValue");r&&n.set("value",r)},blurInputIfNeeded({scope:n}){F1e(n)?.blur()}}}};Wo()(["activationMode","autoResize","dir","disabled","finalFocusEl","form","getRootNode","id","ids","invalid","maxLength","name","onEditChange","onFocusOutside","onInteractOutside","onPointerDownOutside","onValueChange","onValueCommit","onValueRevert","placeholder","readOnly","required","selectOnFocus","edit","defaultEdit","submitMode","translations","defaultValue","value"]);const Uqn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c=$M(),d={id:r,ids:{label:c?.ids.label,input:c?.ids.control},dir:s,disabled:c?.disabled,invalid:c?.invalid,readOnly:c?.readOnly,required:c?.required,getRootNode:o,...n},h=uv(Bqn,d);return Fqn(h,ry)},VXt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["activationMode","autoResize","defaultEdit","defaultValue","disabled","edit","finalFocusEl","form","id","ids","invalid","maxLength","name","onEditChange","onFocusOutside","onInteractOutside","onPointerDownOutside","onValueChange","onValueCommit","onValueRevert","placeholder","readOnly","required","selectOnFocus","submitMode","translations","value"]),c=Uqn(o),d=Fn(c.getRootProps(),s);return H.jsx(LXt,{value:c,children:H.jsx(Xn.div,{...d,ref:r})})});VXt.displayName="EditableRoot";const WXt=Q.forwardRef((n,r)=>{const[{value:o},s]=Ou()(n,["value"]),c=Fn(o.getRootProps(),s);return H.jsx(LXt,{value:o,children:H.jsx(Xn.div,{...c,ref:r})})});WXt.displayName="EditableRootProvider";const KXt=Q.forwardRef((n,r)=>{const o=BM(),s=Fn(o.getSubmitTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});KXt.displayName="EditableSubmitTrigger";const YXt=Q.forwardRef((n,r)=>{const o=$M(),s=Fn(o.getErrorTextProps(),n);return o?.invalid?H.jsx(Xn.span,{...s,ref:r}):null});YXt.displayName="FieldErrorText";const XXt=Q.forwardRef((n,r)=>{const o=$M(),s=Fn(o?.getHelperTextProps(),n);return H.jsx(Xn.span,{...s,ref:r})});XXt.displayName="FieldHelperText";const JXt=Q.forwardRef((n,r)=>{const o=$M(),s=Fn(o?.getInputProps(),n);return H.jsx(Xn.input,{...s,ref:r})});JXt.displayName="FieldInput";const ZXt=Q.forwardRef((n,r)=>{const o=$M(),s=Fn(o?.getLabelProps(),n);return H.jsx(Xn.label,{...s,ref:r})});ZXt.displayName="FieldLabel";const QXt=typeof window<"u"?Q.useLayoutEffect:Q.useEffect,[zqn,Sye]=fu({name:"FieldsetContext",hookName:"useFieldsetContext",providerName:"<FieldsetProvider />",strict:!1}),eJt=Vo("field").parts("root","errorText","helperText","input","label","select","textarea","requiredIndicator"),zD=eJt.build(),Hqn=(n={})=>{const r=Sye(),o=qb(),{ids:s,disabled:c=!!r?.disabled,invalid:d=!1,readOnly:h=!1,required:b=!1}=n,[y,E]=Q.useState(!1),[k,x]=Q.useState(!1),O=Q.useId(),P=n.id??O,D=Q.useRef(null),$=s?.control??`field::${P}`,U=s?.errorText??`field::${P}::error-text`,B=s?.helperText??`field::${P}::helper-text`,G=s?.label??`field::${P}::label`;QXt(()=>{const Ue=D.current;if(!Ue)return;const Z=()=>{const qe=o.getRootNode();E(!!qe.getElementById(U)),x(!!qe.getElementById(B))};Z();const Re=o.getWindow(),_e=new Re.MutationObserver(Z);return _e.observe(Ue,{childList:!0,subtree:!0}),()=>_e.disconnect()},[o,U,B]);const K=Q.useMemo(()=>{const Ue=[];return y&&d&&Ue.push(U),k&&Ue.push(B),Ue.join(" ")||void 0},[d,U,B,y,k]),X=Q.useMemo(()=>()=>({...zD.root.attrs,id:$,ref:D,role:"group","data-disabled":nr(c),"data-invalid":nr(d),"data-readonly":nr(h)}),[c,d,h,$]),ae=Q.useMemo(()=>()=>({...zD.label.attrs,id:G,"data-disabled":nr(c),"data-invalid":nr(d),"data-readonly":nr(h),htmlFor:P}),[c,d,h,P,G]),oe=Q.useMemo(()=>()=>({"aria-describedby":K,"aria-invalid":gI(d),"data-invalid":nr(d),"data-required":nr(b),"data-readonly":nr(h),id:P,required:b,disabled:c,readOnly:h}),[K,d,b,h,P,c]),he=Q.useMemo(()=>()=>({...oe(),...zD.input.attrs}),[oe]),ye=Q.useMemo(()=>()=>({...oe(),...zD.textarea.attrs}),[oe]),we=Q.useMemo(()=>()=>({...oe(),...zD.select.attrs}),[oe]),ge=Q.useMemo(()=>()=>({id:B,...zD.helperText.attrs,"data-disabled":nr(c)}),[c,B]),Te=Q.useMemo(()=>()=>({id:U,...zD.errorText.attrs,"aria-live":"polite"}),[U]),xe=Q.useMemo(()=>()=>({"aria-hidden":!0,...zD.requiredIndicator.attrs}),[]);return{ariaDescribedby:K,ids:{root:$,control:P,label:G,errorText:U,helperText:B},refs:{rootRef:D},disabled:c,invalid:d,readOnly:h,required:b,getLabelProps:ae,getRootProps:X,getInputProps:he,getTextareaProps:ye,getSelectProps:we,getHelperTextProps:ge,getErrorTextProps:Te,getRequiredIndicatorProps:xe}},tJt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["id","ids","disabled","invalid","readOnly","required"]),c=Hqn(o),d=Fn(c.getRootProps(),s);return H.jsx(CGn,{value:c,children:H.jsx(Xn.div,{...d,ref:X4(r,c.refs.rootRef)})})});tJt.displayName="FieldRoot";var Gqn=n=>{if(!n)return;const r=qB(n),o=Jp(n),s=ny(n),c=()=>{requestAnimationFrame(()=>{n.style.height="auto";let E;r.boxSizing==="content-box"?E=n.scrollHeight-(parseFloat(r.paddingTop)+parseFloat(r.paddingBottom)):E=n.scrollHeight+parseFloat(r.borderTopWidth)+parseFloat(r.borderBottomWidth),r.maxHeight!=="none"&&E>parseFloat(r.maxHeight)?(r.overflowY==="hidden"&&(n.style.overflowY="scroll"),E=parseFloat(r.maxHeight)):r.overflowY!=="hidden"&&(n.style.overflowY="hidden"),n.style.height=`${E}px`})};n.addEventListener("input",c),n.form?.addEventListener("reset",c);const d=Object.getPrototypeOf(n),h=Object.getOwnPropertyDescriptor(d,"value");Object.defineProperty(n,"value",{...h,set(){h?.set?.apply(this,arguments),c()}});const b=new o.ResizeObserver(()=>{requestAnimationFrame(()=>c())});b.observe(n);const y=new o.MutationObserver(()=>c());return y.observe(n,{attributes:!0,attributeFilter:["rows","placeholder"]}),s.fonts?.addEventListener("loadingdone",c),()=>{n.removeEventListener("input",c),n.form?.removeEventListener("reset",c),s.fonts?.removeEventListener("loadingdone",c),b.disconnect(),y.disconnect()}};const nJt=Q.forwardRef((n,r)=>{const{autoresize:o,...s}=n,c=Q.useRef(null),d=$M(),h=Fn(d?.getTextareaProps(),{style:{resize:o?"none":void 0}},s);return Q.useEffect(()=>{if(o)return Gqn(c.current)},[o]),H.jsx(Xn.textarea,{...h,ref:X4(r,c)})});nJt.displayName="FieldTextarea";const rJt=Q.forwardRef((n,r)=>{const o=Sye(),s=Fn(o.getErrorTextProps(),n);return o.invalid?H.jsx(Xn.span,{...s,ref:r}):null});rJt.displayName="FieldsetErrorText";const iJt=Q.forwardRef((n,r)=>{const o=Sye(),s=Fn(o.getHelperTextProps(),n);return H.jsx(Xn.span,{...s,ref:r})});iJt.displayName="FieldsetHelperText";const oJt=Q.forwardRef((n,r)=>{const o=Sye(),s=Fn(o.getLegendProps(),n);return H.jsx(Xn.div,{...s,ref:r})});oJt.displayName="FieldsetLegend";const aJt=Vo("fieldset").parts("root","errorText","helperText","legend"),B1e=aJt.build(),qqn=(n={})=>{const{disabled:r=!1,invalid:o=!1}=n,s=qb(),c=Q.useRef(!1),d=Q.useRef(!1),h=Q.useId(),b=n.id??h,y=Q.useRef(null),E=`fieldset::${b}::legend`,k=`fieldset::${b}::error-text`,x=`fieldset::${b}::helper-text`;QXt(()=>{const B=y.current;if(!B)return;const G=()=>{const ae=s.getRootNode();c.current=!!ae.getElementById(k),d.current=!!ae.getElementById(x)};G();const K=s.getWindow(),X=new K.MutationObserver(G);return X.observe(B,{childList:!0,subtree:!0}),()=>X.disconnect()},[s,k,x]);const O=Q.useMemo(()=>{const B=[];return c.current&&o&&B.push(k),d.current&&B.push(x),B.join(" ")||void 0},[o,k,x]),P=Q.useMemo(()=>()=>({...B1e.root.attrs,ref:y,disabled:r,"data-disabled":nr(r),"data-invalid":nr(o),"aria-labelledby":E,"aria-describedby":O}),[r,o,E,O]),D=Q.useMemo(()=>()=>({id:E,...B1e.legend.attrs,"data-disabled":nr(r),"data-invalid":nr(o)}),[E,r,o]),$=Q.useMemo(()=>()=>({id:x,...B1e.helperText.attrs}),[x]),U=Q.useMemo(()=>()=>({id:k,...B1e.errorText.attrs,"aria-live":"polite"}),[k]);return{refs:{rootRef:y},disabled:r,invalid:o,getRootProps:P,getLegendProps:D,getHelperTextProps:$,getErrorTextProps:U}},sJt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["id","disabled","invalid"]),c=qqn(o),d=Fn(c.getRootProps(),s);return H.jsx(zqn,{value:c,children:H.jsx(Xn.fieldset,{...d,ref:X4(r,c.refs.rootRef)})})});sJt.displayName="FieldsetRoot";var cJt=Vo("file-upload").parts("root","dropzone","item","itemDeleteTrigger","itemGroup","itemName","itemPreview","itemPreviewImage","itemSizeText","label","trigger","clearTrigger");cJt.build();Wo()(["accept","acceptedFiles","allowDrop","capture","defaultAcceptedFiles","dir","directory","disabled","getRootNode","id","ids","invalid","locale","maxFiles","maxFileSize","minFileSize","name","onFileAccept","onFileChange","onFileReject","preventDocumentDrop","required","transformFiles","translations","validate"]);Wo()(["file"]);var uJt=Vo("hoverCard").parts("arrow","arrowTip","trigger","positioner","content");uJt.build();var Vqn=n=>n.ids?.trigger??`hover-card:${n.id}:trigger`,Wqn=n=>n.ids?.content??`hover-card:${n.id}:content`,Kqn=n=>n.ids?.positioner??`hover-card:${n.id}:popper`,Hje=n=>n.getById(Vqn(n)),Yqn=n=>n.getById(Wqn(n)),x9t=n=>n.getById(Kqn(n)),{not:U1e,and:C9t}=$I();C9t("isOpenControlled",U1e("isPointer")),U1e("isPointer"),C9t("isOpenControlled",U1e("isPointer")),U1e("isPointer");Wo()(["closeDelay","dir","getRootNode","id","ids","disabled","onOpenChange","defaultOpen","open","openDelay","positioning","onInteractOutside","onPointerDownOutside","onFocusOutside"]);var lJt=Vo("tree-view").parts("branch","branchContent","branchControl","branchIndentGuide","branchIndicator","branchText","branchTrigger","item","itemIndicator","itemText","label","nodeCheckbox","root","tree"),Bw=lJt.build(),dJt=n=>new aYe(n);dJt.empty=()=>new aYe({rootNode:{children:[]}});var Xqn=n=>n.ids?.root??`tree:${n.id}:root`,A9t=n=>n.ids?.label??`tree:${n.id}:label`,fJt=(n,r)=>n.ids?.node?.(r)??`tree:${n.id}:node:${r}`,Gje=n=>n.ids?.tree??`tree:${n.id}:tree`,XD=(n,r)=>{r!=null&&n.getById(fJt(n,r))?.focus()};function hJt(n,r,o){const s=n.getNodeValue(r);if(!n.isBranchNode(r))return o.includes(s);const c=n.getDescendantValues(s),d=c.every(b=>o.includes(b)),h=c.some(b=>o.includes(b));return d?!0:h?"indeterminate":!1}function Jqn(n,r,o){const s=n.getDescendantValues(r),c=s.every(d=>o.includes(d));return sM(c?$B(o,...s):P4(o,...s))}function Zqn(n,r){const o=new Map;return n.visit({onEnter:s=>{const c=n.getNodeValue(s),d=n.isBranchNode(s),h=hJt(n,s,r);o.set(c,{type:d?"branch":"leaf",checked:h})}}),o}function Qqn(n,r){const{context:o,scope:s,computed:c,prop:d,send:h}=n,b=d("collection"),y=Array.from(o.get("expandedValue")),E=Array.from(o.get("selectedValue")),k=Array.from(o.get("checkedValue")),x=c("isTypingAhead"),O=o.get("focusedValue"),P=o.get("loadingStatus");function D($){const{node:U,indexPath:B}=$,G=b.getNodeValue(U),K=b.getFirstNode(),X=K?b.getNodeValue(K):null;return{id:fJt(s,G),value:G,indexPath:B,valuePath:b.getValuePath(B),disabled:!!U.disabled,focused:O==null?X==G:O===G,selected:E.includes(G),expanded:y.includes(G),loading:P[G]==="loading",depth:B.length,isBranch:b.isBranchNode(U),get checked(){return hJt(b,U,k)}}}return{collection:b,expandedValue:y,selectedValue:E,checkedValue:k,toggleChecked($,U){h({type:"CHECKED.TOGGLE",value:$,isBranch:U})},setChecked($){h({type:"CHECKED.SET",value:$})},clearChecked(){h({type:"CHECKED.CLEAR"})},getCheckedMap(){return Zqn(b,k)},expand($){h({type:$?"BRANCH.EXPAND":"EXPANDED.ALL",value:$})},collapse($){h({type:$?"BRANCH.COLLAPSE":"EXPANDED.CLEAR",value:$})},deselect($){h({type:$?"NODE.DESELECT":"SELECTED.CLEAR",value:$})},select($){h({type:$?"NODE.SELECT":"SELECTED.ALL",value:$,isTrusted:!1})},getVisibleNodes(){return c("visibleNodes").map(({node:$})=>$)},focus($){XD(s,$)},selectParent($){const U=b.getParentNode($);if(!U)return;const B=P4(E,b.getNodeValue(U));h({type:"SELECTED.SET",value:B,src:"select.parent"})},expandParent($){const U=b.getParentNode($);if(!U)return;const B=P4(y,b.getNodeValue(U));h({type:"EXPANDED.SET",value:B,src:"expand.parent"})},setExpandedValue($){const U=sM($);h({type:"EXPANDED.SET",value:U})},setSelectedValue($){const U=sM($);h({type:"SELECTED.SET",value:U})},getRootProps(){return r.element({...Bw.root.attrs,id:Xqn(s),dir:d("dir")})},getLabelProps(){return r.element({...Bw.label.attrs,id:A9t(s),dir:d("dir")})},getTreeProps(){return r.element({...Bw.tree.attrs,id:Gje(s),dir:d("dir"),role:"tree","aria-label":"Tree View","aria-labelledby":A9t(s),"aria-multiselectable":d("selectionMode")==="multiple"||void 0,tabIndex:-1,onKeyDown($){if($.defaultPrevented||UKe($))return;const U=Yp($);if(oye(U))return;const B=U?.closest("[data-part=branch-control], [data-part=item]");if(!B)return;const G=B.dataset.value;if(G==null){console.warn("[zag-js/tree-view] Node id not found for node",B);return}const K=B.matches("[data-part=branch-control]"),X={ArrowDown(he){k4(he)||(he.preventDefault(),h({type:"NODE.ARROW_DOWN",id:G,shiftKey:he.shiftKey}))},ArrowUp(he){k4(he)||(he.preventDefault(),h({type:"NODE.ARROW_UP",id:G,shiftKey:he.shiftKey}))},ArrowLeft(he){k4(he)||B.dataset.disabled||(he.preventDefault(),h({type:K?"BRANCH_NODE.ARROW_LEFT":"NODE.ARROW_LEFT",id:G}))},ArrowRight(he){!K||B.dataset.disabled||(he.preventDefault(),h({type:"BRANCH_NODE.ARROW_RIGHT",id:G}))},Home(he){k4(he)||(he.preventDefault(),h({type:"NODE.HOME",id:G,shiftKey:he.shiftKey}))},End(he){k4(he)||(he.preventDefault(),h({type:"NODE.END",id:G,shiftKey:he.shiftKey}))},Space(he){B.dataset.disabled||(x?h({type:"TREE.TYPEAHEAD",key:he.key}):X.Enter?.(he))},Enter(he){B.dataset.disabled||eW(U)&&k4(he)||(h({type:K?"BRANCH_NODE.CLICK":"NODE.CLICK",id:G,src:"keyboard"}),eW(U)||he.preventDefault())},"*"(he){B.dataset.disabled||(he.preventDefault(),h({type:"SIBLINGS.EXPAND",id:G}))},a(he){!he.metaKey||B.dataset.disabled||(he.preventDefault(),h({type:"SELECTED.ALL",moveFocus:!0}))}},ae=$qe($,{dir:d("dir")}),oe=X[ae];if(oe){oe($);return}WW.isValidEvent($)&&(h({type:"TREE.TYPEAHEAD",key:$.key,id:G}),$.preventDefault())}})},getNodeState:D,getItemProps($){const U=D($);return r.element({...Bw.item.attrs,id:U.id,dir:d("dir"),"data-ownedby":Gje(s),"data-path":$.indexPath.join("/"),"data-value":U.value,tabIndex:U.focused?0:-1,"data-focus":nr(U.focused),role:"treeitem","aria-current":U.selected?"true":void 0,"aria-selected":U.disabled?void 0:U.selected,"data-selected":nr(U.selected),"aria-disabled":gI(U.disabled),"data-disabled":nr(U.disabled),"aria-level":U.depth,"data-depth":U.depth,style:{"--depth":U.depth},onFocus(B){B.stopPropagation(),h({type:"NODE.FOCUS",id:U.value})},onClick(B){if(U.disabled||!EW(B)||eW(B.currentTarget)&&k4(B))return;const G=B.metaKey||B.ctrlKey;h({type:"NODE.CLICK",id:U.value,shiftKey:B.shiftKey,ctrlKey:G}),B.stopPropagation(),eW(B.currentTarget)||B.preventDefault()}})},getItemTextProps($){const U=D($);return r.element({...Bw.itemText.attrs,"data-disabled":nr(U.disabled),"data-selected":nr(U.selected),"data-focus":nr(U.focused)})},getItemIndicatorProps($){const U=D($);return r.element({...Bw.itemIndicator.attrs,"aria-hidden":!0,"data-disabled":nr(U.disabled),"data-selected":nr(U.selected),"data-focus":nr(U.focused),hidden:!U.selected})},getBranchProps($){const U=D($);return r.element({...Bw.branch.attrs,"data-depth":U.depth,dir:d("dir"),"data-branch":U.value,role:"treeitem","data-ownedby":Gje(s),"data-value":U.value,"aria-level":U.depth,"aria-selected":U.disabled?void 0:U.selected,"data-path":$.indexPath.join("/"),"data-selected":nr(U.selected),"aria-expanded":U.expanded,"data-state":U.expanded?"open":"closed","aria-disabled":gI(U.disabled),"data-disabled":nr(U.disabled),"data-loading":nr(U.loading),"aria-busy":gI(U.loading),style:{"--depth":U.depth}})},getBranchIndicatorProps($){const U=D($);return r.element({...Bw.branchIndicator.attrs,"aria-hidden":!0,"data-state":U.expanded?"open":"closed","data-disabled":nr(U.disabled),"data-selected":nr(U.selected),"data-focus":nr(U.focused),"data-loading":nr(U.loading)})},getBranchTriggerProps($){const U=D($);return r.element({...Bw.branchTrigger.attrs,role:"button",dir:d("dir"),"data-disabled":nr(U.disabled),"data-state":U.expanded?"open":"closed","data-value":U.value,"data-loading":nr(U.loading),disabled:U.loading,onClick(B){U.disabled||U.loading||(h({type:"BRANCH_TOGGLE.CLICK",id:U.value}),B.stopPropagation())}})},getBranchControlProps($){const U=D($);return r.element({...Bw.branchControl.attrs,role:"button",id:U.id,dir:d("dir"),tabIndex:U.focused?0:-1,"data-path":$.indexPath.join("/"),"data-state":U.expanded?"open":"closed","data-disabled":nr(U.disabled),"data-selected":nr(U.selected),"data-focus":nr(U.focused),"data-value":U.value,"data-depth":U.depth,"data-loading":nr(U.loading),"aria-busy":gI(U.loading),onFocus(B){h({type:"NODE.FOCUS",id:U.value}),B.stopPropagation()},onClick(B){if(U.disabled||U.loading||!EW(B)||eW(B.currentTarget)&&k4(B))return;const G=B.metaKey||B.ctrlKey;h({type:"BRANCH_NODE.CLICK",id:U.value,shiftKey:B.shiftKey,ctrlKey:G}),B.stopPropagation()}})},getBranchTextProps($){const U=D($);return r.element({...Bw.branchText.attrs,dir:d("dir"),"data-disabled":nr(U.disabled),"data-state":U.expanded?"open":"closed","data-loading":nr(U.loading)})},getBranchContentProps($){const U=D($);return r.element({...Bw.branchContent.attrs,role:"group",dir:d("dir"),"data-state":U.expanded?"open":"closed","data-depth":U.depth,"data-path":$.indexPath.join("/"),"data-value":U.value,hidden:!U.expanded})},getBranchIndentGuideProps($){const U=D($);return r.element({...Bw.branchIndentGuide.attrs,"data-depth":U.depth})},getNodeCheckboxProps($){const U=D($),B=U.checked;return r.element({...Bw.nodeCheckbox.attrs,tabIndex:-1,role:"checkbox","data-state":B===!0?"checked":B===!1?"unchecked":"indeterminate","aria-checked":B===!0?"true":B===!1?"false":"mixed","data-disabled":nr(U.disabled),onClick(G){if(G.defaultPrevented||U.disabled||!EW(G))return;h({type:"CHECKED.TOGGLE",value:U.value,isBranch:U.isBranch}),G.stopPropagation(),G.currentTarget.closest("[role=treeitem]")?.focus({preventScroll:!0})}})}}}function z1e(n,r){const{context:o,prop:s,refs:c}=n;if(!s("loadChildren")){o.set("expandedValue",$=>sM(P4($,...r)));return}const d=o.get("loadingStatus"),[h,b]=P8t(r,$=>d[$]==="loaded");if(h.length>0&&o.set("expandedValue",$=>sM(P4($,...h))),b.length===0)return;const y=s("collection"),[E,k]=P8t(b,$=>{const U=y.findNode($);return y.getNodeChildren(U).length>0});if(E.length>0&&o.set("expandedValue",$=>sM(P4($,...E))),k.length===0)return;o.set("loadingStatus",$=>({...$,...k.reduce((U,B)=>({...U,[B]:"loading"}),{})}));const x=k.map($=>{const U=y.getIndexPath($),B=y.getValuePath(U),G=y.findNode($);return{id:$,indexPath:U,valuePath:B,node:G}}),O=c.get("pendingAborts"),P=s("loadChildren");aYt(P,()=>"[zag-js/tree-view] `loadChildren` is required for async expansion");const D=x.map(({id:$,indexPath:U,valuePath:B,node:G})=>{const K=O.get($);K&&(K.abort(),O.delete($));const X=new AbortController;return O.set($,X),P({valuePath:B,indexPath:U,node:G,signal:X.signal})});Promise.allSettled(D).then($=>{const U=[],B=[],G=o.get("loadingStatus");let K=s("collection");$.forEach((X,ae)=>{const{id:oe,indexPath:he,node:ye,valuePath:we}=x[ae];X.status==="fulfilled"?(G[oe]="loaded",U.push(oe),K=K.replace(he,{...ye,children:X.value})):(O.delete(oe),Reflect.deleteProperty(G,oe),B.push({node:ye,error:X.reason,indexPath:he,valuePath:we}))}),o.set("loadingStatus",G),U.length&&(o.set("expandedValue",X=>sM(P4(X,...U))),s("onLoadChildrenComplete")?.({collection:K})),B.length&&s("onLoadChildrenError")?.({nodes:B})})}function HD(n){const{prop:r,context:o}=n;return function({indexPath:c}){return r("collection").getValuePath(c).slice(0,-1).some(h=>!o.get("expandedValue").includes(h))}}var{and:eI}=$I(),eVn={props({props:n}){return{selectionMode:"single",collection:dJt.empty(),typeahead:!0,expandOnClick:!0,defaultExpandedValue:[],defaultSelectedValue:[],...n}},initialState(){return"idle"},context({prop:n,bindable:r,getContext:o}){return{expandedValue:r(()=>({defaultValue:n("defaultExpandedValue"),value:n("expandedValue"),isEqual:d2,onChange(s){const d=o().get("focusedValue");n("onExpandedChange")?.({expandedValue:s,focusedValue:d,get expandedNodes(){return n("collection").findNodes(s)}})}})),selectedValue:r(()=>({defaultValue:n("defaultSelectedValue"),value:n("selectedValue"),isEqual:d2,onChange(s){const d=o().get("focusedValue");n("onSelectionChange")?.({selectedValue:s,focusedValue:d,get selectedNodes(){return n("collection").findNodes(s)}})}})),focusedValue:r(()=>({defaultValue:n("defaultFocusedValue")||null,value:n("focusedValue"),onChange(s){n("onFocusChange")?.({focusedValue:s,get focusedNode(){return s?n("collection").findNode(s):null}})}})),loadingStatus:r(()=>({defaultValue:{}})),checkedValue:r(()=>({defaultValue:n("defaultCheckedValue")||[],value:n("checkedValue"),isEqual:d2,onChange(s){n("onCheckedChange")?.({checkedValue:s})}}))}},refs(){return{typeaheadState:{...WW.defaultOptions},pendingAborts:new Map}},computed:{isMultipleSelection:({prop:n})=>n("selectionMode")==="multiple",isTypingAhead:({refs:n})=>n.get("typeaheadState").keysSoFar.length>0,visibleNodes:({prop:n,context:r})=>{const o=[];return n("collection").visit({skip:HD({prop:n,context:r}),onEnter:(s,c)=>{o.push({node:s,indexPath:c})}}),o}},on:{"EXPANDED.SET":{actions:["setExpanded"]},"EXPANDED.CLEAR":{actions:["clearExpanded"]},"EXPANDED.ALL":{actions:["expandAllBranches"]},"BRANCH.EXPAND":{actions:["expandBranches"]},"BRANCH.COLLAPSE":{actions:["collapseBranches"]},"SELECTED.SET":{actions:["setSelected"]},"SELECTED.ALL":[{guard:eI("isMultipleSelection","moveFocus"),actions:["selectAllNodes","focusTreeLastNode"]},{guard:"isMultipleSelection",actions:["selectAllNodes"]}],"SELECTED.CLEAR":{actions:["clearSelected"]},"NODE.SELECT":{actions:["selectNode"]},"NODE.DESELECT":{actions:["deselectNode"]},"CHECKED.TOGGLE":{actions:["toggleChecked"]},"CHECKED.SET":{actions:["setChecked"]},"CHECKED.CLEAR":{actions:["clearChecked"]}},exit:["clearPendingAborts"],states:{idle:{on:{"NODE.FOCUS":{actions:["setFocusedNode"]},"NODE.ARROW_DOWN":[{guard:eI("isShiftKey","isMultipleSelection"),actions:["focusTreeNextNode","extendSelectionToNextNode"]},{actions:["focusTreeNextNode"]}],"NODE.ARROW_UP":[{guard:eI("isShiftKey","isMultipleSelection"),actions:["focusTreePrevNode","extendSelectionToPrevNode"]},{actions:["focusTreePrevNode"]}],"NODE.ARROW_LEFT":{actions:["focusBranchNode"]},"BRANCH_NODE.ARROW_LEFT":[{guard:"isBranchExpanded",actions:["collapseBranch"]},{actions:["focusBranchNode"]}],"BRANCH_NODE.ARROW_RIGHT":[{guard:eI("isBranchFocused","isBranchExpanded"),actions:["focusBranchFirstNode"]},{actions:["expandBranch"]}],"SIBLINGS.EXPAND":{actions:["expandSiblingBranches"]},"NODE.HOME":[{guard:eI("isShiftKey","isMultipleSelection"),actions:["extendSelectionToFirstNode","focusTreeFirstNode"]},{actions:["focusTreeFirstNode"]}],"NODE.END":[{guard:eI("isShiftKey","isMultipleSelection"),actions:["extendSelectionToLastNode","focusTreeLastNode"]},{actions:["focusTreeLastNode"]}],"NODE.CLICK":[{guard:eI("isCtrlKey","isMultipleSelection"),actions:["toggleNodeSelection"]},{guard:eI("isShiftKey","isMultipleSelection"),actions:["extendSelectionToNode"]},{actions:["selectNode"]}],"BRANCH_NODE.CLICK":[{guard:eI("isCtrlKey","isMultipleSelection"),actions:["toggleNodeSelection"]},{guard:eI("isShiftKey","isMultipleSelection"),actions:["extendSelectionToNode"]},{guard:"expandOnClick",actions:["selectNode","toggleBranchNode"]},{actions:["selectNode"]}],"BRANCH_TOGGLE.CLICK":{actions:["toggleBranchNode"]},"TREE.TYPEAHEAD":{actions:["focusMatchedNode"]}}}},implementations:{guards:{isBranchFocused:({context:n,event:r})=>n.get("focusedValue")===r.id,isBranchExpanded:({context:n,event:r})=>n.get("expandedValue").includes(r.id),isShiftKey:({event:n})=>n.shiftKey,isCtrlKey:({event:n})=>n.ctrlKey,hasSelectedItems:({context:n})=>n.get("selectedValue").length>0,isMultipleSelection:({prop:n})=>n("selectionMode")==="multiple",moveFocus:({event:n})=>!!n.moveFocus,expandOnClick:({prop:n})=>!!n("expandOnClick")},actions:{selectNode({context:n,event:r}){const o=r.id||r.value;n.set("selectedValue",s=>o==null?s:!r.isTrusted&&Ore(o)?s.concat(...o):[Ore(o)?$Ke(o):o].filter(Boolean))},deselectNode({context:n,event:r}){const o=eve(r.id||r.value);n.set("selectedValue",s=>$B(s,...o))},setFocusedNode({context:n,event:r}){n.set("focusedValue",r.id)},clearFocusedNode({context:n}){n.set("focusedValue",null)},clearSelectedItem({context:n}){n.set("selectedValue",[])},toggleBranchNode({context:n,event:r,action:o}){const s=n.get("expandedValue").includes(r.id);o(s?["collapseBranch"]:["expandBranch"])},expandBranch(n){const{event:r}=n;z1e(n,[r.id])},expandBranches(n){const{context:r,event:o}=n,s=eve(o.value);z1e(n,Rje(s,r.get("expandedValue")))},collapseBranch({context:n,event:r}){n.set("expandedValue",o=>$B(o,r.id))},collapseBranches(n){const{context:r,event:o}=n,s=eve(o.value);r.set("expandedValue",c=>$B(c,...s))},setExpanded({context:n,event:r}){Ore(r.value)&&n.set("expandedValue",r.value)},clearExpanded({context:n}){n.set("expandedValue",[])},setSelected({context:n,event:r}){Ore(r.value)&&n.set("selectedValue",r.value)},clearSelected({context:n}){n.set("selectedValue",[])},focusTreeFirstNode({prop:n,scope:r}){const o=n("collection"),s=o.getFirstNode(),c=o.getNodeValue(s);XD(r,c)},focusTreeLastNode(n){const{prop:r,scope:o}=n,s=r("collection"),c=s.getLastNode(void 0,{skip:HD(n)}),d=s.getNodeValue(c);XD(o,d)},focusBranchFirstNode({event:n,prop:r,scope:o}){const s=r("collection"),c=s.findNode(n.id),d=s.getFirstNode(c),h=s.getNodeValue(d);XD(o,h)},focusTreeNextNode(n){const{event:r,prop:o,scope:s}=n,c=o("collection"),d=c.getNextNode(r.id,{skip:HD(n)});if(!d)return;const h=c.getNodeValue(d);XD(s,h)},focusTreePrevNode(n){const{event:r,prop:o,scope:s}=n,c=o("collection"),d=c.getPreviousNode(r.id,{skip:HD(n)});if(!d)return;const h=c.getNodeValue(d);XD(s,h)},focusBranchNode({event:n,prop:r,scope:o}){const s=r("collection"),c=s.getParentNode(n.id),d=c?s.getNodeValue(c):void 0;XD(o,d)},selectAllNodes({context:n,prop:r}){n.set("selectedValue",r("collection").getValues())},focusMatchedNode(n){const{context:r,prop:o,refs:s,event:c,scope:d,computed:h}=n,y=h("visibleNodes").map(({node:k})=>({textContent:o("collection").stringifyNode(k),id:o("collection").getNodeValue(k)})),E=WW(y,{state:s.get("typeaheadState"),activeId:r.get("focusedValue"),key:c.key});XD(d,E?.id)},toggleNodeSelection({context:n,event:r}){const o=GW(n.get("selectedValue"),r.id);n.set("selectedValue",o)},expandAllBranches(n){const{context:r,prop:o}=n,s=o("collection").getBranchValues(),c=Rje(s,r.get("expandedValue"));z1e(n,c)},expandSiblingBranches(n){const{context:r,event:o,prop:s}=n,c=s("collection"),d=c.getIndexPath(o.id);if(!d)return;const b=c.getSiblingNodes(d).map(E=>c.getNodeValue(E)),y=Rje(b,r.get("expandedValue"));z1e(n,y)},extendSelectionToNode(n){const{context:r,event:o,prop:s,computed:c}=n,d=s("collection"),h=nie(r.get("selectedValue"))||d.getNodeValue(d.getFirstNode()),b=o.id;let y=[h,b],E=0;c("visibleNodes").forEach(({node:x})=>{const O=d.getNodeValue(x);E===1&&y.push(O),(O===h||O===b)&&E++}),r.set("selectedValue",sM(y))},extendSelectionToNextNode(n){const{context:r,event:o,prop:s}=n,c=s("collection"),d=c.getNextNode(o.id,{skip:HD(n)});if(!d)return;const h=new Set(r.get("selectedValue")),b=c.getNodeValue(d);b!=null&&(h.has(o.id)&&h.has(b)?h.delete(o.id):h.has(b)||h.add(b),r.set("selectedValue",Array.from(h)))},extendSelectionToPrevNode(n){const{context:r,event:o,prop:s}=n,c=s("collection"),d=c.getPreviousNode(o.id,{skip:HD(n)});if(!d)return;const h=new Set(r.get("selectedValue")),b=c.getNodeValue(d);b!=null&&(h.has(o.id)&&h.has(b)?h.delete(o.id):h.has(b)||h.add(b),r.set("selectedValue",Array.from(h)))},extendSelectionToFirstNode(n){const{context:r,prop:o}=n,s=o("collection"),c=nie(r.get("selectedValue")),d=[];s.visit({skip:HD(n),onEnter:h=>{const b=s.getNodeValue(h);if(d.push(b),b===c)return"stop"}}),r.set("selectedValue",d)},extendSelectionToLastNode(n){const{context:r,prop:o}=n,s=o("collection"),c=nie(r.get("selectedValue")),d=[];let h=!1;s.visit({skip:HD(n),onEnter:b=>{const y=s.getNodeValue(b);y===c&&(h=!0),h&&d.push(y)}}),r.set("selectedValue",d)},clearPendingAborts({refs:n}){const r=n.get("pendingAborts");r.forEach(o=>o.abort()),r.clear()},toggleChecked({context:n,event:r,prop:o}){const s=o("collection");n.set("checkedValue",c=>r.isBranch?Jqn(s,r.value,c):GW(c,r.value))},setChecked({context:n,event:r}){n.set("checkedValue",r.value)},clearChecked({context:n}){n.set("checkedValue",[])}}}};Wo()(["ids","collection","dir","expandedValue","expandOnClick","defaultFocusedValue","focusedValue","getRootNode","id","onExpandedChange","onFocusChange","onSelectionChange","checkedValue","selectedValue","selectionMode","typeahead","defaultExpandedValue","defaultSelectedValue","defaultCheckedValue","onCheckedChange","onLoadChildrenComplete","onLoadChildrenError","loadChildren"]);Wo()(["node","indexPath"]);const tVn=n=>{const r=Q.useId(),{dir:o}=b2(),{getRootNode:s}=qb(),c={id:r,dir:o,getRootNode:s,...n},d=uv(eVn,c);return Qqn(d,ry)},[pJt,lv]=fu({name:"TreeViewContext",hookName:"useTreeViewContext",providerName:"<TreeViewProvider />"}),nVn=(n,r)=>{const[o,s]=bye(n),[c,d]=Ou()(s,["checkedValue","collection","defaultCheckedValue","defaultExpandedValue","defaultFocusedValue","defaultSelectedValue","expandedValue","expandOnClick","focusedValue","id","ids","loadChildren","onCheckedChange","onExpandedChange","onFocusChange","onLoadChildrenComplete","onLoadChildrenError","onSelectionChange","selectedValue","selectionMode","typeahead"]),h=tVn(c),b=Fn(h.getRootProps(),d);return H.jsx(pJt,{value:h,children:H.jsx(mye,{value:o,children:H.jsx(Xn.div,{...b,ref:r})})})},rVn=Q.forwardRef(nVn),iVn=(n,r)=>{const[o,s]=bye(n),[{value:c},d]=Ou()(s,["value"]),h=Fn(c.getRootProps(),d);return H.jsx(pJt,{value:c,children:H.jsx(mye,{value:o,children:H.jsx(Xn.div,{...h,ref:r})})})},oVn=Q.forwardRef(iVn),gJt=Q.forwardRef((n,r)=>{const o=lv(),s=Fn(o.getTreeProps(),n);return H.jsx(Xn.div,{...s,ref:r})});gJt.displayName="TreeViewTree";const[aVn,mJt]=fu({name:"TreeViewNodeContext",hookName:"useTreeViewNodeContext",providerName:"<TreeViewNodeProvider />"}),[sVn,Xx]=fu({name:"TreeViewNodePropsContext",hookName:"useTreeViewNodePropsContext",providerName:"<TreeViewItemProvider />"}),bJt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=mJt(),d=GYt(),h=Fn(o.getBranchProps(s),n),b=o.getBranchContentProps(s);return H.jsx(CYt,{ref:r,open:c.expanded,ids:{content:b.id},...d,...h})});bJt.displayName="TreeViewBranch";const cVn=Ou(),vJt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=o.getBranchContentProps(s),[,d]=cVn(c,["hidden","data-state"]),h=Fn(d,n);return H.jsx(qYt,{ref:r,...h})});vJt.displayName="TreeViewBranchContent";const wJt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getBranchControlProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});wJt.displayName="TreeViewBranchControl";const yJt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getBranchIndentGuideProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});yJt.displayName="TreeViewBranchIndentGuide";const EJt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getBranchIndicatorProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});EJt.displayName="TreeViewBranchIndicator";const SJt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getBranchTextProps(s),n);return H.jsx(Xn.span,{...c,ref:r})});SJt.displayName="TreeViewBranchText";const _Jt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getItemProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});_Jt.displayName="TreeViewItem";const kJt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getItemTextProps(s),n);return H.jsx(Xn.span,{...c,ref:r})});kJt.displayName="TreeViewItemText";function uVn(n){const[r,o]=Ou()(n,["indexPath","node"]),c=lv().getNodeState(r);return H.jsx(aVn,{value:c,children:H.jsx(sVn,{value:r,children:o.children})})}var TJt=Vo("listbox").parts("label","input","item","itemText","itemIndicator","itemGroup","itemGroupLabel","content","root","valueText");TJt.build();Wo()(["collection","defaultHighlightedValue","defaultValue","dir","disabled","deselectable","disallowSelectAll","getRootNode","highlightedValue","id","ids","loopFocus","onHighlightChange","onSelect","onValueChange","orientation","scrollToIndexFn","selectionMode","selectOnHighlight","typeahead","value"]);Wo()(["item","highlightOnHover"]);Wo()(["id"]);Wo()(["htmlFor"]);const lVn=TJt.extendWith("empty"),[xJt,qm]=fu({name:"MenuContext",hookName:"useMenuContext",providerName:"<MenuProvider />",strict:!1}),CJt=Q.forwardRef((n,r)=>{const o=qm(),s=Fn(o.getArrowProps(),n);return H.jsx(Xn.div,{...s,ref:r})});CJt.displayName="MenuArrow";const AJt=Q.forwardRef((n,r)=>{const o=qm(),s=Fn(o.getArrowTipProps(),n);return H.jsx(Xn.div,{...s,ref:r})});AJt.displayName="MenuArrowTip";const[uYe,$ni]=fu({name:"MenuItemContext",hookName:"useMenuItemContext",providerName:"<MenuItemProvider />"}),[lYe,IJt]=fu({name:"MenuItemPropsContext",hookName:"useMenuItemPropsContext",providerName:"<MenuItemPropsProvider />"}),OJt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["checked","closeOnSelect","disabled","onCheckedChange","value","valueText"]),c={...o,type:"checkbox"},d=qm(),h=Fn(d.getOptionItemProps(c),s),b=d.getOptionItemState(c);return H.jsx(lYe,{value:c,children:H.jsx(uYe,{value:b,children:H.jsx(Xn.div,{...h,ref:r})})})});OJt.displayName="MenuCheckboxItem";const NJt=Q.forwardRef((n,r)=>{const o=qm(),s=Yx(),c=Fn(o.getContentProps(),s.getPresenceProps(),n);return s.unmounted?null:H.jsx(Xn.div,{...c,ref:X4(s.ref,r)})});NJt.displayName="MenuContent";const RJt=Q.forwardRef((n,r)=>{const o=qm(),s=Fn(o.getContextTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});RJt.displayName="MenuContextTrigger";const PJt=Q.forwardRef((n,r)=>{const o=qm(),s=Fn(o.getIndicatorProps(),n);return H.jsx(Xn.div,{...s,ref:r})});PJt.displayName="MenuIndicator";const LJt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["closeOnSelect","disabled","value","valueText","onSelect"]),c=qm(),d=Fn(c.getItemProps(o),s),h=c.getItemState(o);return Q.useEffect(()=>c.addItemListener({id:h.id,onSelect:o.onSelect}),[h.id,o.onSelect]),H.jsx(lYe,{value:o,children:H.jsx(uYe,{value:h,children:H.jsx(Xn.div,{...d,ref:r})})})});LJt.displayName="MenuItem";const[DJt,MJt]=fu({name:"MenuItemGroupContext",hookName:"useMenuItemGroupContext",providerName:"<MenuItemGroupProvider />"}),jJt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["id"]),c=qm(),h={id:Q.useId(),...o},b=Fn(c.getItemGroupProps(h),s);return H.jsx(DJt,{value:h,children:H.jsx(Xn.div,{...b,ref:r})})});jJt.displayName="MenuItemGroup";const $Jt=Q.forwardRef((n,r)=>{const o=qm(),s=MJt(),c=Fn(o.getItemGroupLabelProps({htmlFor:s.id}),n);return H.jsx(Xn.div,{...c,ref:r})});$Jt.displayName="MenuItemGroupLabel";const FJt=Q.forwardRef((n,r)=>{const o=qm(),s=IJt(),c=Fn(o.getItemIndicatorProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});FJt.displayName="MenuItemIndicator";const BJt=Q.forwardRef((n,r)=>{const o=qm(),s=IJt(),c=Fn(o.getItemTextProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});BJt.displayName="MenuItemText";const UJt=Q.forwardRef((n,r)=>{const o=qm(),s=Fn(o.getPositionerProps(),n);return Yx().unmounted?null:H.jsx(Xn.div,{...s,ref:r})});UJt.displayName="MenuPositioner";const zJt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["closeOnSelect","disabled","value","valueText"]),c=qm(),d=MJt(),h={...o,checked:d.value===o.value,type:"radio",onCheckedChange:()=>d.onValueChange?.({value:o.value})},b=Fn(c.getOptionItemProps(h),s),y=c.getOptionItemState(h);return H.jsx(lYe,{value:h,children:H.jsx(uYe,{value:y,children:H.jsx(Xn.div,{...b,ref:r})})})});zJt.displayName="MenuRadioItem";const HJt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["id","onValueChange","value"]),c=qm(),h={id:Q.useId(),...o},b=Fn(c.getItemGroupProps({id:h.id}),s);return H.jsx(DJt,{value:h,children:H.jsx(Xn.div,{...b,ref:r})})});HJt.displayName="MenuRadioItemGroup";const GJt=n=>{const r=Q.useRef(n),o=Q.useRef(!1);Q.useEffect(()=>{r.current=n}),Q.useEffect(()=>{o.current!==!0&&(o.current=!0,r.current())},[])};var qJt=Vo("menu").parts("arrow","arrowTip","content","contextTrigger","indicator","item","itemGroup","itemGroupLabel","itemIndicator","itemText","positioner","separator","trigger","triggerItem"),nv=qJt.build(),Vve=n=>n.ids?.trigger??`menu:${n.id}:trigger`,VJt=n=>n.ids?.contextTrigger??`menu:${n.id}:ctx-trigger`,oie=n=>n.ids?.content??`menu:${n.id}:content`,dVn=n=>n.ids?.arrow??`menu:${n.id}:arrow`,WJt=n=>n.ids?.positioner??`menu:${n.id}:popper`,fVn=(n,r)=>n.ids?.group?.(r)??`menu:${n.id}:group:${r}`,jie=(n,r)=>`${n.id}/${r}`,nB=n=>n?.dataset.value??null,I9t=(n,r)=>n.ids?.groupLabel?.(r)??`menu:${n.id}:group-label:${r}`,JD=n=>n.getById(oie(n)),O9t=n=>n.getById(WJt(n)),H1e=n=>n.getById(Vve(n)),KJt=(n,r)=>r?n.getById(jie(n,r)):null,qje=n=>n.getById(VJt(n)),xoe=n=>{const o=`[role^="menuitem"][data-ownedby=${CSS.escape(oie(n))}]:not([data-disabled])`;return Pie(JD(n),o)},hVn=n=>nie(xoe(n)),pVn=n=>$Ke(xoe(n)),dYe=(n,r)=>r?n.id===r||n.dataset.value===r:!1,gVn=(n,r)=>{const o=xoe(n),s=o.findIndex(c=>dYe(c,r.value));return RFn(o,s,{loop:r.loop??r.loopFocus})},mVn=(n,r)=>{const o=xoe(n),s=o.findIndex(c=>dYe(c,r.value));return LFn(o,s,{loop:r.loop??r.loopFocus})},bVn=(n,r)=>{const o=xoe(n),s=o.find(c=>dYe(c,r.value));return WW(o,{state:r.typeaheadState,key:r.key,activeId:s?.id??null})},G1e=n=>Gb(n)&&(n.dataset.disabled===""||n.hasAttribute("disabled")),vVn=n=>!!n?.getAttribute("role")?.startsWith("menuitem")&&!!n?.hasAttribute("aria-controls"),Jqe="menu:select";function wVn(n,r){if(!n)return;const o=Jp(n),s=new o.CustomEvent(Jqe,{detail:{value:r}});n.dispatchEvent(s)}function yVn(n,r){const{context:o,send:s,state:c,computed:d,prop:h,scope:b}=n,y=c.hasTag("open"),E=o.get("isSubmenu"),k=d("isTypingAhead"),x=h("composite"),O=o.get("currentPlacement"),P=o.get("anchorPoint"),D=o.get("highlightedValue"),$=iYe({...h("positioning"),placement:P?"bottom":O});function U(X){return{id:jie(b,X.value),disabled:!!X.disabled,highlighted:D===X.value}}function B(X){const ae=X.valueText??X.value;return{...X,id:X.value,valueText:ae}}function G(X){return{...U(B(X)),checked:!!X.checked}}function K(X){const{closeOnSelect:ae,valueText:oe,value:he}=X,ye=U(X),we=jie(b,he);return r.element({...nv.item.attrs,id:we,role:"menuitem","aria-disabled":gI(ye.disabled),"data-disabled":nr(ye.disabled),"data-ownedby":oie(b),"data-highlighted":nr(ye.highlighted),"data-value":he,"data-valuetext":oe,onDragStart(ge){ge.currentTarget.matches("a[href]")&&ge.preventDefault()},onPointerMove(ge){if(ye.disabled||ge.pointerType!=="mouse")return;const Te=ge.currentTarget;ye.highlighted||s({type:"ITEM_POINTERMOVE",id:we,target:Te,closeOnSelect:ae})},onPointerLeave(ge){if(ye.disabled||ge.pointerType!=="mouse"||!n.event.previous()?.type.includes("POINTER"))return;const xe=ge.currentTarget;s({type:"ITEM_POINTERLEAVE",id:we,target:xe,closeOnSelect:ae})},onPointerDown(ge){if(ye.disabled)return;const Te=ge.currentTarget;s({type:"ITEM_POINTERDOWN",target:Te,id:we,closeOnSelect:ae})},onClick(ge){if(H8t(ge)||z8t(ge)||ye.disabled)return;const Te=ge.currentTarget;s({type:"ITEM_CLICK",target:Te,id:we,closeOnSelect:ae})}})}return{highlightedValue:D,open:y,setOpen(X){c.hasTag("open")!==X&&s({type:X?"OPEN":"CLOSE"})},setHighlightedValue(X){s({type:"HIGHLIGHTED.SET",value:X})},setParent(X){s({type:"PARENT.SET",value:X,id:X.prop("id")})},setChild(X){s({type:"CHILD.SET",value:X,id:X.prop("id")})},reposition(X={}){s({type:"POSITIONING.SET",options:X})},addItemListener(X){const ae=b.getById(X.id);if(!ae)return;const oe=()=>X.onSelect?.();return ae.addEventListener(Jqe,oe),()=>ae.removeEventListener(Jqe,oe)},getContextTriggerProps(){return r.element({...nv.contextTrigger.attrs,dir:h("dir"),id:VJt(b),"data-state":y?"open":"closed",onPointerDown(X){if(X.pointerType==="mouse")return;const ae=OB(X);s({type:"CONTEXT_MENU_START",point:ae})},onPointerCancel(X){X.pointerType!=="mouse"&&s({type:"CONTEXT_MENU_CANCEL"})},onPointerMove(X){X.pointerType!=="mouse"&&s({type:"CONTEXT_MENU_CANCEL"})},onPointerUp(X){X.pointerType!=="mouse"&&s({type:"CONTEXT_MENU_CANCEL"})},onContextMenu(X){const ae=OB(X);s({type:"CONTEXT_MENU",point:ae}),X.preventDefault()},style:{WebkitTouchCallout:"none",WebkitUserSelect:"none",userSelect:"none"}})},getTriggerItemProps(X){const ae=X.getTriggerProps();return Fn(K({value:ae.id}),ae)},getTriggerProps(){return r.button({...E?nv.triggerItem.attrs:nv.trigger.attrs,"data-placement":o.get("currentPlacement"),type:"button",dir:h("dir"),id:Vve(b),"data-uid":h("id"),"aria-haspopup":x?"menu":"dialog","aria-controls":oie(b),"aria-expanded":y||void 0,"data-state":y?"open":"closed",onPointerMove(X){if(X.pointerType!=="mouse"||G1e(X.currentTarget)||!E)return;const oe=OB(X);s({type:"TRIGGER_POINTERMOVE",target:X.currentTarget,point:oe})},onPointerLeave(X){if(G1e(X.currentTarget)||X.pointerType!=="mouse"||!E)return;const ae=OB(X);s({type:"TRIGGER_POINTERLEAVE",target:X.currentTarget,point:ae})},onPointerDown(X){G1e(X.currentTarget)||hYt(X)||X.preventDefault()},onClick(X){X.defaultPrevented||G1e(X.currentTarget)||s({type:"TRIGGER_CLICK",target:X.currentTarget})},onBlur(){s({type:"TRIGGER_BLUR"})},onFocus(){s({type:"TRIGGER_FOCUS"})},onKeyDown(X){if(X.defaultPrevented)return;const ae={ArrowDown(){s({type:"ARROW_DOWN"})},ArrowUp(){s({type:"ARROW_UP"})},Enter(){s({type:"ARROW_DOWN",src:"enter"})},Space(){s({type:"ARROW_DOWN",src:"space"})}},oe=$qe(X,{orientation:"vertical",dir:h("dir")}),he=ae[oe];he&&(X.preventDefault(),he(X))}})},getIndicatorProps(){return r.element({...nv.indicator.attrs,dir:h("dir"),"data-state":y?"open":"closed"})},getPositionerProps(){return r.element({...nv.positioner.attrs,dir:h("dir"),id:WJt(b),style:$.floating})},getArrowProps(){return r.element({id:dVn(b),...nv.arrow.attrs,dir:h("dir"),style:$.arrow})},getArrowTipProps(){return r.element({...nv.arrowTip.attrs,dir:h("dir"),style:$.arrowTip})},getContentProps(){return r.element({...nv.content.attrs,id:oie(b),"aria-label":h("aria-label"),hidden:!y,"data-state":y?"open":"closed",role:x?"menu":"dialog",tabIndex:0,dir:h("dir"),"aria-activedescendant":d("highlightedId")||void 0,"aria-labelledby":Vve(b),"data-placement":O,onPointerEnter(X){X.pointerType==="mouse"&&s({type:"MENU_POINTERENTER"})},onKeyDown(X){if(X.defaultPrevented||!ABn(X))return;const ae=Yp(X);if(!(ae?.closest("[role=menu]")===X.currentTarget||ae===X.currentTarget))return;if(X.key==="Tab"&&!VBn(X)){X.preventDefault();return}const he=KJt(b,D),ye={ArrowDown(){s({type:"ARROW_DOWN"})},ArrowUp(){s({type:"ARROW_UP"})},ArrowLeft(){s({type:"ARROW_LEFT"})},ArrowRight(){s({type:"ARROW_RIGHT"})},Enter(){s({type:"ENTER"}),D!=null&&eW(he)&&h("navigate")?.({value:D,node:he,href:he.href})},Space(Te){k?s({type:"TYPEAHEAD",key:Te.key}):ye.Enter?.(Te)},Home(){s({type:"HOME"})},End(){s({type:"END"})}},we=$qe(X,{dir:h("dir")}),ge=ye[we];if(ge){ge(X),X.stopPropagation(),X.preventDefault();return}h("typeahead")&&OBn(X)&&(k4(X)||oye(ae)||(s({type:"TYPEAHEAD",key:X.key}),X.preventDefault()))}})},getSeparatorProps(){return r.element({...nv.separator.attrs,role:"separator",dir:h("dir"),"aria-orientation":"horizontal"})},getItemState:U,getItemProps:K,getOptionItemState:G,getOptionItemProps(X){const{type:ae,disabled:oe,closeOnSelect:he}=X,ye=B(X),we=G(X);return{...K(ye),...r.element({"data-type":ae,...nv.item.attrs,dir:h("dir"),"data-value":ye.value,role:`menuitem${ae}`,"aria-checked":!!we.checked,"data-state":we.checked?"checked":"unchecked",onClick(ge){if(oe||H8t(ge)||z8t(ge))return;const Te=ge.currentTarget;s({type:"ITEM_CLICK",target:Te,option:ye,closeOnSelect:he})}})}},getItemIndicatorProps(X){const ae=G(D8t(X)),oe=ae.checked?"checked":"unchecked";return r.element({...nv.itemIndicator.attrs,dir:h("dir"),"data-disabled":nr(ae.disabled),"data-highlighted":nr(ae.highlighted),"data-state":Mx(X,"checked")?oe:void 0,hidden:Mx(X,"checked")?!ae.checked:void 0})},getItemTextProps(X){const ae=G(D8t(X)),oe=ae.checked?"checked":"unchecked";return r.element({...nv.itemText.attrs,dir:h("dir"),"data-disabled":nr(ae.disabled),"data-highlighted":nr(ae.highlighted),"data-state":Mx(X,"checked")?oe:void 0})},getItemGroupLabelProps(X){return r.element({...nv.itemGroupLabel.attrs,id:I9t(b,X.htmlFor),dir:h("dir")})},getItemGroupProps(X){return r.element({id:fVn(b,X.id),...nv.itemGroup.attrs,dir:h("dir"),"aria-labelledby":I9t(b,X.id),role:"group"})}}}var{not:I_,and:LV,or:EVn}=$I(),SVn={props({props:n}){return{closeOnSelect:!0,typeahead:!0,composite:!0,loopFocus:!1,navigate(r){vYt(r.node)},...n,positioning:{placement:"bottom-start",gutter:8,...n.positioning}}},initialState({prop:n}){return n("open")||n("defaultOpen")?"open":"idle"},context({bindable:n,prop:r}){return{suspendPointer:n(()=>({defaultValue:!1})),highlightedValue:n(()=>({defaultValue:r("defaultHighlightedValue")||null,value:r("highlightedValue"),onChange(o){r("onHighlightChange")?.({highlightedValue:o})}})),lastHighlightedValue:n(()=>({defaultValue:null})),currentPlacement:n(()=>({defaultValue:void 0})),intentPolygon:n(()=>({defaultValue:null})),anchorPoint:n(()=>({defaultValue:null,hash(o){return`x: ${o?.x}, y: ${o?.y}`}})),isSubmenu:n(()=>({defaultValue:!1}))}},refs(){return{parent:null,children:{},typeaheadState:{...WW.defaultOptions},positioningOverride:{}}},computed:{isRtl:({prop:n})=>n("dir")==="rtl",isTypingAhead:({refs:n})=>n.get("typeaheadState").keysSoFar!=="",highlightedId:({context:n,scope:r,refs:o})=>kVn(o.get("children"),n.get("highlightedValue"),r)},watch({track:n,action:r,context:o,prop:s}){n([()=>o.get("isSubmenu")],()=>{r(["setSubmenuPlacement"])}),n([()=>o.hash("anchorPoint")],()=>{o.get("anchorPoint")&&r(["reposition"])}),n([()=>s("open")],()=>{r(["toggleVisibility"])})},on:{"PARENT.SET":{actions:["setParentMenu"]},"CHILD.SET":{actions:["setChildMenu"]},OPEN:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}],OPEN_AUTOFOCUS:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["highlightFirstItem","invokeOnOpen"]}],CLOSE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],"HIGHLIGHTED.RESTORE":{actions:["restoreHighlightedItem"]},"HIGHLIGHTED.SET":{actions:["setHighlightedItem"]}},states:{idle:{tags:["closed"],on:{"CONTROLLED.OPEN":{target:"open"},"CONTROLLED.CLOSE":{target:"closed"},CONTEXT_MENU_START:{target:"opening:contextmenu",actions:["setAnchorPoint"]},CONTEXT_MENU:[{guard:"isOpenControlled",actions:["setAnchorPoint","invokeOnOpen"]},{target:"open",actions:["setAnchorPoint","invokeOnOpen"]}],TRIGGER_CLICK:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}],TRIGGER_FOCUS:{guard:I_("isSubmenu"),target:"closed"},TRIGGER_POINTERMOVE:{guard:"isSubmenu",target:"opening"}}},"opening:contextmenu":{tags:["closed"],effects:["waitForLongPress"],on:{"CONTROLLED.OPEN":{target:"open"},"CONTROLLED.CLOSE":{target:"closed"},CONTEXT_MENU_CANCEL:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],"LONG_PRESS.OPEN":[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}]}},opening:{tags:["closed"],effects:["waitForOpenDelay"],on:{"CONTROLLED.OPEN":{target:"open"},"CONTROLLED.CLOSE":{target:"closed"},BLUR:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],TRIGGER_POINTERLEAVE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],"DELAY.OPEN":[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}]}},closing:{tags:["open"],effects:["trackPointerMove","trackInteractOutside","waitForCloseDelay"],on:{"CONTROLLED.OPEN":{target:"open"},"CONTROLLED.CLOSE":{target:"closed",actions:["focusParentMenu","restoreParentHighlightedItem"]},MENU_POINTERENTER:{target:"open",actions:["clearIntentPolygon"]},POINTER_MOVED_AWAY_FROM_SUBMENU:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["focusParentMenu","restoreParentHighlightedItem"]}],"DELAY.CLOSE":[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["focusParentMenu","restoreParentHighlightedItem","invokeOnClose"]}]}},closed:{tags:["closed"],entry:["clearHighlightedItem","focusTrigger","resumePointer","clearAnchorPoint"],on:{"CONTROLLED.OPEN":[{guard:EVn("isOpenAutoFocusEvent","isArrowDownEvent"),target:"open",actions:["highlightFirstItem"]},{guard:"isArrowUpEvent",target:"open",actions:["highlightLastItem"]},{target:"open"}],CONTEXT_MENU_START:{target:"opening:contextmenu",actions:["setAnchorPoint"]},CONTEXT_MENU:[{guard:"isOpenControlled",actions:["setAnchorPoint","invokeOnOpen"]},{target:"open",actions:["setAnchorPoint","invokeOnOpen"]}],TRIGGER_CLICK:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}],TRIGGER_POINTERMOVE:{guard:"isTriggerItem",target:"opening"},TRIGGER_BLUR:{target:"idle"},ARROW_DOWN:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["highlightFirstItem","invokeOnOpen"]}],ARROW_UP:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["highlightLastItem","invokeOnOpen"]}]}},open:{tags:["open"],effects:["trackInteractOutside","trackPositioning","scrollToHighlightedItem"],entry:["focusMenu","resumePointer"],on:{"CONTROLLED.CLOSE":[{target:"closed",guard:"isArrowLeftEvent",actions:["focusParentMenu"]},{target:"closed"}],TRIGGER_CLICK:[{guard:LV(I_("isTriggerItem"),"isOpenControlled"),actions:["invokeOnClose"]},{guard:I_("isTriggerItem"),target:"closed",actions:["invokeOnClose"]}],CONTEXT_MENU:{actions:["setAnchorPoint","focusMenu"]},ARROW_UP:{actions:["highlightPrevItem","focusMenu"]},ARROW_DOWN:{actions:["highlightNextItem","focusMenu"]},ARROW_LEFT:[{guard:LV("isSubmenu","isOpenControlled"),actions:["invokeOnClose"]},{guard:"isSubmenu",target:"closed",actions:["focusParentMenu","invokeOnClose"]}],HOME:{actions:["highlightFirstItem","focusMenu"]},END:{actions:["highlightLastItem","focusMenu"]},ARROW_RIGHT:{guard:"isTriggerItemHighlighted",actions:["openSubmenu"]},ENTER:[{guard:"isTriggerItemHighlighted",actions:["openSubmenu"]},{actions:["clickHighlightedItem"]}],ITEM_POINTERMOVE:[{guard:I_("isPointerSuspended"),actions:["setHighlightedItem","focusMenu"]},{actions:["setLastHighlightedItem"]}],ITEM_POINTERLEAVE:{guard:LV(I_("isPointerSuspended"),I_("isTriggerItem")),actions:["clearHighlightedItem"]},ITEM_CLICK:[{guard:LV(I_("isTriggerItemHighlighted"),I_("isHighlightedItemEditable"),"closeOnSelect","isOpenControlled"),actions:["invokeOnSelect","setOptionState","closeRootMenu","invokeOnClose"]},{guard:LV(I_("isTriggerItemHighlighted"),I_("isHighlightedItemEditable"),"closeOnSelect"),target:"closed",actions:["invokeOnSelect","setOptionState","closeRootMenu","invokeOnClose"]},{guard:LV(I_("isTriggerItemHighlighted"),I_("isHighlightedItemEditable")),actions:["invokeOnSelect","setOptionState"]},{actions:["setHighlightedItem"]}],TRIGGER_POINTERMOVE:{guard:"isTriggerItem",actions:["setIntentPolygon"]},TRIGGER_POINTERLEAVE:{target:"closing"},ITEM_POINTERDOWN:{actions:["setHighlightedItem"]},TYPEAHEAD:{actions:["highlightMatchedItem"]},FOCUS_MENU:{actions:["focusMenu"]},"POSITIONING.SET":{actions:["reposition"]}}}},implementations:{guards:{closeOnSelect:({prop:n,event:r})=>!!(r?.closeOnSelect??n("closeOnSelect")),isTriggerItem:({event:n})=>vVn(n.target),isTriggerItemHighlighted:({event:n,scope:r,computed:o})=>!!(n.target??r.getById(o("highlightedId")))?.hasAttribute("aria-controls"),isSubmenu:({context:n})=>n.get("isSubmenu"),isPointerSuspended:({context:n})=>n.get("suspendPointer"),isHighlightedItemEditable:({scope:n,computed:r})=>oye(n.getById(r("highlightedId"))),isOpenControlled:({prop:n})=>n("open")!==void 0,isArrowLeftEvent:({event:n})=>n.previousEvent?.type==="ARROW_LEFT",isArrowUpEvent:({event:n})=>n.previousEvent?.type==="ARROW_UP",isArrowDownEvent:({event:n})=>n.previousEvent?.type==="ARROW_DOWN",isOpenAutoFocusEvent:({event:n})=>n.previousEvent?.type==="OPEN_AUTOFOCUS"},effects:{waitForOpenDelay({send:n}){const r=setTimeout(()=>{n({type:"DELAY.OPEN"})},100);return()=>clearTimeout(r)},waitForCloseDelay({send:n}){const r=setTimeout(()=>{n({type:"DELAY.CLOSE"})},300);return()=>clearTimeout(r)},waitForLongPress({send:n}){const r=setTimeout(()=>{n({type:"LONG_PRESS.OPEN"})},700);return()=>clearTimeout(r)},trackPositioning({context:n,prop:r,scope:o,refs:s}){if(qje(o))return;const c={...r("positioning"),...s.get("positioningOverride")};n.set("currentPlacement",c.placement);const d=()=>O9t(o);return g2(H1e(o),d,{...c,defer:!0,onComplete(h){n.set("currentPlacement",h.placement)}})},trackInteractOutside({refs:n,scope:r,prop:o,context:s,send:c}){const d=()=>JD(r);let h=!0;return cU(d,{defer:!0,exclude:[H1e(r)],onInteractOutside:o("onInteractOutside"),onRequestDismiss:o("onRequestDismiss"),onFocusOutside(b){o("onFocusOutside")?.(b);const y=Yp(b.detail.originalEvent);if(f2(qje(r),y)){b.preventDefault();return}},onEscapeKeyDown(b){o("onEscapeKeyDown")?.(b),s.get("isSubmenu")&&b.preventDefault(),N9t({parent:n.get("parent")})},onPointerDownOutside(b){o("onPointerDownOutside")?.(b);const y=Yp(b.detail.originalEvent);if(f2(qje(r),y)&&b.detail.contextmenu){b.preventDefault();return}h=!b.detail.focusable},onDismiss(){c({type:"CLOSE",src:"interact-outside",restoreFocus:h})}})},trackPointerMove({context:n,scope:r,send:o,refs:s,flush:c}){const d=s.get("parent");c(()=>{d.context.set("suspendPointer",!0)});const h=r.getDoc();return Ul(h,"pointermove",b=>{_Vn(n.get("intentPolygon"),{x:b.clientX,y:b.clientY})||(o({type:"POINTER_MOVED_AWAY_FROM_SUBMENU"}),d.context.set("suspendPointer",!1))})},scrollToHighlightedItem({event:n,scope:r,computed:o}){const s=()=>{if(n.type.startsWith("ITEM_POINTER"))return;const d=r.getById(o("highlightedId")),h=JD(r);$ve(d,{rootEl:h,block:"nearest"})};return ca(()=>s()),koe(()=>JD(r),{defer:!0,attributes:["aria-activedescendant"],callback:s})}},actions:{setAnchorPoint({context:n,event:r}){n.set("anchorPoint",o=>d2(o,r.point)?o:r.point)},setSubmenuPlacement({context:n,computed:r,refs:o}){if(!n.get("isSubmenu"))return;const s=r("isRtl")?"left-start":"right-start";o.set("positioningOverride",{placement:s,gutter:0})},reposition({context:n,scope:r,prop:o,event:s,refs:c}){const d=()=>O9t(r),h=n.get("anchorPoint"),b=h?()=>({width:0,height:0,...h}):void 0,y={...o("positioning"),...c.get("positioningOverride")};g2(H1e(r),d,{...y,defer:!0,getAnchorRect:b,...s.options??{},listeners:!1,onComplete(E){n.set("currentPlacement",E.placement)}})},setOptionState({event:n}){if(!n.option)return;const{checked:r,onCheckedChange:o,type:s}=n.option;s==="radio"?o?.(!0):s==="checkbox"&&o?.(!r)},clickHighlightedItem({scope:n,computed:r}){const o=n.getById(r("highlightedId"));!o||o.dataset.disabled||queueMicrotask(()=>o.click())},setIntentPolygon({context:n,scope:r,event:o}){const s=JD(r),c=n.get("currentPlacement");if(!s||!c)return;const d=s.getBoundingClientRect(),h=fGn(d,c);if(!h)return;const y=sHn(c)==="right"?-5:5;n.set("intentPolygon",[{...o.point,x:o.point.x+y},...h])},clearIntentPolygon({context:n}){n.set("intentPolygon",null)},clearAnchorPoint({context:n}){n.set("anchorPoint",null)},resumePointer({refs:n,flush:r}){const o=n.get("parent");o&&r(()=>{o.context.set("suspendPointer",!1)})},setHighlightedItem({context:n,event:r}){const o=r.value||nB(r.target);n.set("highlightedValue",o)},clearHighlightedItem({context:n}){n.set("highlightedValue",null)},focusMenu({scope:n}){ca(()=>{const r=JD(n);Rie({root:r,enabled:!f2(r,n.getActiveElement()),filter(s){return!s.role?.startsWith("menuitem")}})?.focus({preventScroll:!0})})},highlightFirstItem({context:n,scope:r}){(JD(r)?queueMicrotask:ca)(()=>{const s=hVn(r);s&&n.set("highlightedValue",nB(s))})},highlightLastItem({context:n,scope:r}){(JD(r)?queueMicrotask:ca)(()=>{const s=pVn(r);s&&n.set("highlightedValue",nB(s))})},highlightNextItem({context:n,scope:r,event:o,prop:s}){const c=gVn(r,{loop:o.loop,value:n.get("highlightedValue"),loopFocus:s("loopFocus")});n.set("highlightedValue",nB(c))},highlightPrevItem({context:n,scope:r,event:o,prop:s}){const c=mVn(r,{loop:o.loop,value:n.get("highlightedValue"),loopFocus:s("loopFocus")});n.set("highlightedValue",nB(c))},invokeOnSelect({context:n,prop:r,scope:o}){const s=n.get("highlightedValue");if(s==null)return;const c=KJt(o,s);wVn(c,s),r("onSelect")?.({value:s})},focusTrigger({scope:n,context:r,event:o}){r.get("isSubmenu")||r.get("anchorPoint")||o.restoreFocus===!1||queueMicrotask(()=>H1e(n)?.focus({preventScroll:!0}))},highlightMatchedItem({scope:n,context:r,event:o,refs:s}){const c=bVn(n,{key:o.key,value:r.get("highlightedValue"),typeaheadState:s.get("typeaheadState")});c&&r.set("highlightedValue",nB(c))},setParentMenu({refs:n,event:r,context:o}){n.set("parent",r.value),o.set("isSubmenu",!0)},setChildMenu({refs:n,event:r}){const o=n.get("children");o[r.id]=r.value,n.set("children",o)},closeRootMenu({refs:n}){N9t({parent:n.get("parent")})},openSubmenu({refs:n,scope:r,computed:o}){const c=r.getById(o("highlightedId"))?.getAttribute("data-uid"),d=n.get("children");(c?d[c]:null)?.send({type:"OPEN_AUTOFOCUS"})},focusParentMenu({refs:n}){n.get("parent")?.send({type:"FOCUS_MENU"})},setLastHighlightedItem({context:n,event:r}){n.set("lastHighlightedValue",nB(r.target))},restoreHighlightedItem({context:n}){n.get("lastHighlightedValue")&&(n.set("highlightedValue",n.get("lastHighlightedValue")),n.set("lastHighlightedValue",null))},restoreParentHighlightedItem({refs:n}){n.get("parent")?.send({type:"HIGHLIGHTED.RESTORE"})},invokeOnOpen({prop:n}){n("onOpenChange")?.({open:!0})},invokeOnClose({prop:n}){n("onOpenChange")?.({open:!1})},toggleVisibility({prop:n,event:r,send:o}){o({type:n("open")?"CONTROLLED.OPEN":"CONTROLLED.CLOSE",previousEvent:r})}}}};function N9t(n){let r=n.parent;for(;r&&r.context.get("isSubmenu");)r=r.refs.get("parent");r?.send({type:"CLOSE"})}function _Vn(n,r){return n?hGn(n,r):!1}function kVn(n,r,o){const s=Object.keys(n).length>0;if(!r)return null;if(!s)return jie(o,r);for(const c in n){const d=n[c],h=Vve(d.scope);if(h===r)return h}return jie(o,r)}Wo()(["anchorPoint","aria-label","closeOnSelect","composite","defaultHighlightedValue","defaultOpen","dir","getRootNode","highlightedValue","id","ids","loopFocus","navigate","onEscapeKeyDown","onFocusOutside","onHighlightChange","onInteractOutside","onOpenChange","onPointerDownOutside","onRequestDismiss","onSelect","open","positioning","typeahead"]);Wo()(["closeOnSelect","disabled","value","valueText"]);Wo()(["htmlFor"]);Wo()(["id"]);Wo()(["checked","closeOnSelect","disabled","onCheckedChange","type","value","valueText"]);const TVn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv(SVn,c);return{api:yVn(d,ry),service:d}},[YJt,XJt]=fu({name:"MenuMachineContext",hookName:"useMenuMachineContext",providerName:"<MenuMachineProvider />",strict:!1}),[JJt,xVn]=fu({name:"MenuMachineContext",hookName:"useMenuMachineContext",providerName:"<MenuMachineProvider />",strict:!1}),CVn=n=>{const[r,o]=MM(n),[s,c]=Ou()(o,["anchorPoint","aria-label","closeOnSelect","composite","defaultHighlightedValue","defaultOpen","highlightedValue","id","ids","loopFocus","navigate","onEscapeKeyDown","onFocusOutside","onHighlightChange","onInteractOutside","onOpenChange","onPointerDownOutside","onRequestDismiss","onSelect","open","positioning","typeahead"]),d=qm(),h=XJt(),{api:b,service:y}=TVn(s),E=J4(Fn({present:b.open},r));GJt(()=>{h&&d&&(d.setChild(y),b.setParent(h))});const k=Q.useCallback(()=>d?.getTriggerItemProps(b),[b,d]);return H.jsx(JJt,{value:k,children:H.jsx(YJt,{value:y,children:H.jsx(xJt,{value:b,children:H.jsx(jM,{value:E,...c})})})})},AVn=n=>{const r=qm(),o=XJt(),[s,{value:c,children:d}]=MM(n),{api:h,service:b}=c,y=J4(Fn({present:h.open},s));GJt(()=>{o&&r&&(r.setChild(b),h.setParent(o))});const E=Q.useCallback(()=>r?.getTriggerItemProps(h),[h,r]);return H.jsx(JJt,{value:E,children:H.jsx(YJt,{value:b,children:H.jsx(xJt,{value:h,children:H.jsx(jM,{value:y,children:d})})})})},ZJt=Q.forwardRef((n,r)=>{const o=qm(),s=Fn(o.getSeparatorProps(),n);return H.jsx(Xn.hr,{...s,ref:r})});ZJt.displayName="MenuSeparator";const QJt=Q.forwardRef((n,r)=>{const o=qm(),s=Yx(),c=Fn({...o.getTriggerProps(),"aria-controls":s.unmounted?void 0:o.getTriggerProps()["aria-controls"]},n);return H.jsx(Xn.button,{...c,ref:r})});QJt.displayName="MenuTrigger";const eZt=Q.forwardRef((n,r)=>{const o=xVn(),s=Fn(o?.()??{},n);return H.jsx(Xn.div,{...s,ref:r})});eZt.displayName="MenuTriggerItem";let Vje=new Map,Zqe=!1;try{Zqe=new Intl.NumberFormat("de-DE",{signDisplay:"exceptZero"}).resolvedOptions().signDisplay==="exceptZero"}catch{}let Wve=!1;try{Wve=new Intl.NumberFormat("de-DE",{style:"unit",unit:"degree"}).resolvedOptions().style==="unit"}catch{}const tZt={degree:{narrow:{default:"°","ja-JP":" 度","zh-TW":"度","sl-SI":" °"}}};class IVn{format(r){let o="";if(!Zqe&&this.options.signDisplay!=null?o=NVn(this.numberFormatter,this.options.signDisplay,r):o=this.numberFormatter.format(r),this.options.style==="unit"&&!Wve){var s;let{unit:c,unitDisplay:d="short",locale:h}=this.resolvedOptions();if(!c)return o;let b=(s=tZt[c])===null||s===void 0?void 0:s[d];o+=b[h]||b.default}return o}formatToParts(r){return this.numberFormatter.formatToParts(r)}formatRange(r,o){if(typeof this.numberFormatter.formatRange=="function")return this.numberFormatter.formatRange(r,o);if(o<r)throw new RangeError("End date must be >= start date");return`${this.format(r)} – ${this.format(o)}`}formatRangeToParts(r,o){if(typeof this.numberFormatter.formatRangeToParts=="function")return this.numberFormatter.formatRangeToParts(r,o);if(o<r)throw new RangeError("End date must be >= start date");let s=this.numberFormatter.formatToParts(r),c=this.numberFormatter.formatToParts(o);return[...s.map(d=>({...d,source:"startRange"})),{type:"literal",value:" – ",source:"shared"},...c.map(d=>({...d,source:"endRange"}))]}resolvedOptions(){let r=this.numberFormatter.resolvedOptions();return!Zqe&&this.options.signDisplay!=null&&(r={...r,signDisplay:this.options.signDisplay}),!Wve&&this.options.style==="unit"&&(r={...r,style:"unit",unit:this.options.unit,unitDisplay:this.options.unitDisplay}),r}constructor(r,o={}){this.numberFormatter=OVn(r,o),this.options=o}}function OVn(n,r={}){let{numberingSystem:o}=r;if(o&&n.includes("-nu-")&&(n.includes("-u-")||(n+="-u-"),n+=`-nu-${o}`),r.style==="unit"&&!Wve){var s;let{unit:h,unitDisplay:b="short"}=r;if(!h)throw new Error('unit option must be provided with style: "unit"');if(!(!((s=tZt[h])===null||s===void 0)&&s[b]))throw new Error(`Unsupported unit ${h} with unitDisplay = ${b}`);r={...r,style:"decimal"}}let c=n+(r?Object.entries(r).sort((h,b)=>h[0]<b[0]?-1:1).join():"");if(Vje.has(c))return Vje.get(c);let d=new Intl.NumberFormat(n,r);return Vje.set(c,d),d}function NVn(n,r,o){if(r==="auto")return n.format(o);if(r==="never")return n.format(Math.abs(o));{let s=!1;if(r==="always"?s=o>0||Object.is(o,0):r==="exceptZero"&&(Object.is(o,-0)||Object.is(o,0)?o=Math.abs(o):s=o>0),s){let c=n.format(-o),d=n.format(o),h=c.replace(d,"").replace(/\u200e|\u061C/,"");return[...h].length!==1&&console.warn("@react-aria/i18n polyfill for NumberFormat signDisplay: Unsupported case"),c.replace(d,"!!!").replace(h,"+").replace("!!!",d)}else return n.format(o)}}const RVn=new RegExp("^.*\\(.*\\).*$"),PVn=["latn","arab","hanidec","deva","beng","fullwide"];class nZt{parse(r){return Wje(this.locale,this.options,r).parse(r)}isValidPartialNumber(r,o,s){return Wje(this.locale,this.options,r).isValidPartialNumber(r,o,s)}getNumberingSystem(r){return Wje(this.locale,this.options,r).options.numberingSystem}constructor(r,o={}){this.locale=r,this.options=o}}const R9t=new Map;function Wje(n,r,o){let s=P9t(n,r);if(!n.includes("-nu-")&&!s.isValidPartialNumber(o)){for(let c of PVn)if(c!==s.options.numberingSystem){let d=P9t(n+(n.includes("-u-")?"-nu-":"-u-nu-")+c,r);if(d.isValidPartialNumber(o))return d}}return s}function P9t(n,r){let o=n+(r?Object.entries(r).sort((c,d)=>c[0]<d[0]?-1:1).join():""),s=R9t.get(o);return s||(s=new LVn(n,r),R9t.set(o,s)),s}class LVn{parse(r){let o=this.sanitize(r);if(this.symbols.group&&(o=ore(o,this.symbols.group,"")),this.symbols.decimal&&(o=o.replace(this.symbols.decimal,".")),this.symbols.minusSign&&(o=o.replace(this.symbols.minusSign,"-")),o=o.replace(this.symbols.numeral,this.symbols.index),this.options.style==="percent"){let h=o.indexOf("-");o=o.replace("-",""),o=o.replace("+","");let b=o.indexOf(".");b===-1&&(b=o.length),o=o.replace(".",""),b-2===0?o=`0.${o}`:b-2===-1?o=`0.0${o}`:b-2===-2?o="0.00":o=`${o.slice(0,b-2)}.${o.slice(b-2)}`,h>-1&&(o=`-${o}`)}let s=o?+o:NaN;if(isNaN(s))return NaN;if(this.options.style==="percent"){var c,d;let h={...this.options,style:"decimal",minimumFractionDigits:Math.min(((c=this.options.minimumFractionDigits)!==null&&c!==void 0?c:0)+2,20),maximumFractionDigits:Math.min(((d=this.options.maximumFractionDigits)!==null&&d!==void 0?d:0)+2,20)};return new nZt(this.locale,h).parse(new IVn(this.locale,h).format(s))}return this.options.currencySign==="accounting"&&RVn.test(r)&&(s=-1*s),s}sanitize(r){return r=r.replace(this.symbols.literals,""),this.symbols.minusSign&&(r=r.replace("-",this.symbols.minusSign)),this.options.numberingSystem==="arab"&&(this.symbols.decimal&&(r=r.replace(",",this.symbols.decimal),r=r.replace("،",this.symbols.decimal)),this.symbols.group&&(r=ore(r,".",this.symbols.group))),this.options.locale==="fr-FR"&&this.symbols.group&&(r=ore(r," ",this.symbols.group),r=ore(r,/\u00A0/g,this.symbols.group)),r}isValidPartialNumber(r,o=-1/0,s=1/0){return r=this.sanitize(r),this.symbols.minusSign&&r.startsWith(this.symbols.minusSign)&&o<0?r=r.slice(this.symbols.minusSign.length):this.symbols.plusSign&&r.startsWith(this.symbols.plusSign)&&s>0&&(r=r.slice(this.symbols.plusSign.length)),this.symbols.group&&r.startsWith(this.symbols.group)||this.symbols.decimal&&r.indexOf(this.symbols.decimal)>-1&&this.options.maximumFractionDigits===0?!1:(this.symbols.group&&(r=ore(r,this.symbols.group,"")),r=r.replace(this.symbols.numeral,""),this.symbols.decimal&&(r=r.replace(this.symbols.decimal,"")),r.length===0)}constructor(r,o={}){this.locale=r,o.roundingIncrement!==1&&o.roundingIncrement!=null&&(o.maximumFractionDigits==null&&o.minimumFractionDigits==null?(o.maximumFractionDigits=0,o.minimumFractionDigits=0):o.maximumFractionDigits==null?o.maximumFractionDigits=o.minimumFractionDigits:o.minimumFractionDigits==null&&(o.minimumFractionDigits=o.maximumFractionDigits)),this.formatter=new Intl.NumberFormat(r,o),this.options=this.formatter.resolvedOptions(),this.symbols=MVn(r,this.formatter,this.options,o);var s,c;this.options.style==="percent"&&(((s=this.options.minimumFractionDigits)!==null&&s!==void 0?s:0)>18||((c=this.options.maximumFractionDigits)!==null&&c!==void 0?c:0)>18)&&console.warn("NumberParser cannot handle percentages with greater than 18 decimal places, please reduce the number in your options.")}}const L9t=new Set(["decimal","fraction","integer","minusSign","plusSign","group"]),DVn=[0,4,2,1,11,20,3,7,100,21,.1,1.1];function MVn(n,r,o,s){var c,d,h,b;let y=new Intl.NumberFormat(n,{...o,minimumSignificantDigits:1,maximumSignificantDigits:21,roundingIncrement:1,roundingPriority:"auto",roundingMode:"halfExpand"}),E=y.formatToParts(-10000.111),k=y.formatToParts(10000.111),x=DVn.map(ge=>y.formatToParts(ge));var O;let P=(O=(c=E.find(ge=>ge.type==="minusSign"))===null||c===void 0?void 0:c.value)!==null&&O!==void 0?O:"-",D=(d=k.find(ge=>ge.type==="plusSign"))===null||d===void 0?void 0:d.value;!D&&(s?.signDisplay==="exceptZero"||s?.signDisplay==="always")&&(D="+");let U=(h=new Intl.NumberFormat(n,{...o,minimumFractionDigits:2,maximumFractionDigits:2}).formatToParts(.001).find(ge=>ge.type==="decimal"))===null||h===void 0?void 0:h.value,B=(b=E.find(ge=>ge.type==="group"))===null||b===void 0?void 0:b.value,G=E.filter(ge=>!L9t.has(ge.type)).map(ge=>D9t(ge.value)),K=x.flatMap(ge=>ge.filter(Te=>!L9t.has(Te.type)).map(Te=>D9t(Te.value))),X=[...new Set([...G,...K])].sort((ge,Te)=>Te.length-ge.length),ae=X.length===0?new RegExp("[\\p{White_Space}]","gu"):new RegExp(`${X.join("|")}|[\\p{White_Space}]`,"gu"),oe=[...new Intl.NumberFormat(o.locale,{useGrouping:!1}).format(9876543210)].reverse(),he=new Map(oe.map((ge,Te)=>[ge,Te])),ye=new RegExp(`[${oe.join("")}]`,"g");return{minusSign:P,plusSign:D,decimal:U,group:B,literals:ae,numeral:ye,index:ge=>String(he.get(ge))}}function ore(n,r,o){return n.replaceAll?n.replaceAll(r,o):n.split(r).join(o)}function D9t(n){return n.replace(/[.*+?^${}()|[\]\\]/g,"\\$&")}var rZt=Vo("numberInput").parts("root","label","input","control","valueText","incrementTrigger","decrementTrigger","scrubber");rZt.build();var jVn=n=>n.ids?.input??`number-input:${n.id}:input`,$Vn=n=>n.ids?.incrementTrigger??`number-input:${n.id}:inc`,FVn=n=>n.ids?.decrementTrigger??`number-input:${n.id}:dec`,iZt=n=>`number-input:${n.id}:cursor`,q1e=n=>n.getById(jVn(n)),BVn=n=>n.getById($Vn(n)),UVn=n=>n.getById(FVn(n)),oZt=n=>n.getDoc().getElementById(iZt(n)),zVn=(n,r)=>{let o=null;return r==="increment"&&(o=BVn(n)),r==="decrement"&&(o=UVn(n)),o},HVn=(n,r)=>{if(!cye())return VVn(n,r),()=>{oZt(n)?.remove()}},GVn=n=>{const r=n.getDoc(),o=r.documentElement,s=r.body;return s.style.pointerEvents="none",o.style.userSelect="none",o.style.cursor="ew-resize",()=>{s.style.pointerEvents="",o.style.userSelect="",o.style.cursor="",o.style.length||o.removeAttribute("style"),s.style.length||s.removeAttribute("style")}},qVn=(n,r)=>{const{point:o,isRtl:s,event:c}=r,d=n.getWin(),h=Pje(c.movementX,d.devicePixelRatio),b=Pje(c.movementY,d.devicePixelRatio);let y=h>0?"increment":h<0?"decrement":null;s&&y==="increment"&&(y="decrement"),s&&y==="decrement"&&(y="increment");const E={x:o.x+h,y:o.y+b},k=d.innerWidth,x=Pje(7.5,d.devicePixelRatio);return E.x=XFn(E.x+x,k)-x,{hint:y,point:E}},VVn=(n,r)=>{const o=n.getDoc(),s=o.createElement("div");s.className="scrubber--cursor",s.id=iZt(n),Object.assign(s.style,{width:"15px",height:"15px",position:"fixed",pointerEvents:"none",left:"0px",top:"0px",zIndex:cYt,transform:r?`translate3d(${r.x}px, ${r.y}px, 0px)`:void 0,willChange:"transform"}),s.innerHTML=`
      <svg width="46" height="15" style="left: -15.5px; position: absolute; top: 0; filter: drop-shadow(rgba(0, 0, 0, 0.4) 0px 1px 1.1px);">
        <g transform="translate(2 3)">
          <path fill-rule="evenodd" d="M 15 4.5L 15 2L 11.5 5.5L 15 9L 15 6.5L 31 6.5L 31 9L 34.5 5.5L 31 2L 31 4.5Z" style="stroke-width: 2px; stroke: white;"></path>
          <path fill-rule="evenodd" d="M 15 4.5L 15 2L 11.5 5.5L 15 9L 15 6.5L 31 6.5L 31 9L 34.5 5.5L 31 2L 31 4.5Z"></path>
        </g>
      </svg>`,o.body.appendChild(s)};function WVn(n){if(!(!n||n.ownerDocument.activeElement!==n))try{const{selectionStart:r,selectionEnd:o,value:s}=n,c=s.substring(0,r),d=s.substring(o);return{start:r,end:o,value:s,beforeTxt:c,afterTxt:d}}catch{}}function KVn(n,r){if(!(!n||n.ownerDocument.activeElement!==n)){if(!r){n.setSelectionRange(n.value.length,n.value.length);return}try{const{value:o}=n,{beforeTxt:s="",afterTxt:c="",start:d}=r;let h=o.length;if(o.endsWith(c))h=o.length-c.length;else if(o.startsWith(s))h=s.length;else if(d!=null){const b=s[d-1],y=o.indexOf(b,d-1);y!==-1&&(h=y+1)}n.setSelectionRange(h,h)}catch{}}}var YVn=(n,r={})=>new Intl.NumberFormat(n,r),XVn=(n,r={})=>new nZt(n,r),Kje=(n,r)=>{const{prop:o,computed:s}=r;return o("formatOptions")?n===""?Number.NaN:s("parser").parse(n):parseFloat(n)},rB=(n,r)=>{const{prop:o,computed:s}=r;return Number.isNaN(n)?"":o("formatOptions")?s("formatter").format(n):n.toString()},JVn=(n,r)=>{let o=n!==void 0&&!Number.isNaN(n)?n:1;return r?.style==="percent"&&(n===void 0||Number.isNaN(n))&&(o=.01),o},{choose:ZVn,guards:QVn,createMachine:eWn}=_Yt(),{not:M9t,and:j9t}=QVn;eWn({props({props:n}){const r=JVn(n.step,n.formatOptions);return{dir:"ltr",locale:"en-US",focusInputOnChange:!0,clampValueOnBlur:!n.allowOverflow,allowOverflow:!1,inputMode:"decimal",pattern:"-?[0-9]*(.[0-9]+)?",defaultValue:"",step:r,min:Number.MIN_SAFE_INTEGER,max:Number.MAX_SAFE_INTEGER,spinOnPress:!0,...n,translations:{incrementLabel:"increment value",decrementLabel:"decrease value",...n.translations}}},initialState(){return"idle"},context({prop:n,bindable:r,getComputed:o}){return{value:r(()=>({defaultValue:n("defaultValue"),value:n("value"),onChange(s){const c=o(),d=Kje(s,{computed:c,prop:n});n("onValueChange")?.({value:s,valueAsNumber:d})}})),hint:r(()=>({defaultValue:null})),scrubberCursorPoint:r(()=>({defaultValue:null,hash(s){return s?`x:${s.x}, y:${s.y}`:""}})),fieldsetDisabled:r(()=>({defaultValue:!1}))}},computed:{isRtl:({prop:n})=>n("dir")==="rtl",valueAsNumber:({context:n,computed:r,prop:o})=>Kje(n.get("value"),{computed:r,prop:o}),formattedValue:({computed:n,prop:r})=>rB(n("valueAsNumber"),{computed:n,prop:r}),isAtMin:({computed:n,prop:r})=>ZFn(n("valueAsNumber"),r("min")),isAtMax:({computed:n,prop:r})=>JFn(n("valueAsNumber"),r("max")),isOutOfRange:({computed:n,prop:r})=>!iYt(n("valueAsNumber"),r("min"),r("max")),isValueEmpty:({context:n})=>n.get("value")==="",isDisabled:({prop:n,context:r})=>!!n("disabled")||r.get("fieldsetDisabled"),canIncrement:({prop:n,computed:r})=>n("allowOverflow")||!r("isAtMax"),canDecrement:({prop:n,computed:r})=>n("allowOverflow")||!r("isAtMin"),valueText:({prop:n,context:r})=>n("translations").valueText?.(r.get("value")),formatter:V8t(({prop:n})=>[n("locale"),n("formatOptions")],(n,r)=>YVn(n,r)),parser:V8t(({prop:n})=>[n("locale"),n("formatOptions")],(n,r)=>XVn(n,r))},watch({track:n,action:r,context:o,computed:s,prop:c}){n([()=>o.get("value"),()=>c("locale")],()=>{r(["syncInputElement"])}),n([()=>s("isOutOfRange")],()=>{r(["invokeOnInvalid"])}),n([()=>o.hash("scrubberCursorPoint")],()=>{r(["setVirtualCursorPosition"])})},effects:["trackFormControl"],on:{"VALUE.SET":{actions:["setRawValue"]},"VALUE.CLEAR":{actions:["clearValue"]},"VALUE.INCREMENT":{actions:["increment"]},"VALUE.DECREMENT":{actions:["decrement"]}},states:{idle:{on:{"TRIGGER.PRESS_DOWN":[{guard:"isTouchPointer",target:"before:spin",actions:["setHint"]},{target:"before:spin",actions:["focusInput","invokeOnFocus","setHint"]}],"SCRUBBER.PRESS_DOWN":{target:"scrubbing",actions:["focusInput","invokeOnFocus","setHint","setCursorPoint"]},"INPUT.FOCUS":{target:"focused",actions:["focusInput","invokeOnFocus"]}}},focused:{tags:["focus"],effects:["attachWheelListener"],on:{"TRIGGER.PRESS_DOWN":[{guard:"isTouchPointer",target:"before:spin",actions:["setHint"]},{target:"before:spin",actions:["focusInput","setHint"]}],"SCRUBBER.PRESS_DOWN":{target:"scrubbing",actions:["focusInput","setHint","setCursorPoint"]},"INPUT.ARROW_UP":{actions:["increment"]},"INPUT.ARROW_DOWN":{actions:["decrement"]},"INPUT.HOME":{actions:["decrementToMin"]},"INPUT.END":{actions:["incrementToMax"]},"INPUT.CHANGE":{actions:["setValue","setHint"]},"INPUT.BLUR":[{guard:j9t("clampValueOnBlur",M9t("isInRange")),target:"idle",actions:["setClampedValue","clearHint","invokeOnBlur"]},{guard:M9t("isInRange"),target:"idle",actions:["setFormattedValue","clearHint","invokeOnBlur","invokeOnInvalid"]},{target:"idle",actions:["setFormattedValue","clearHint","invokeOnBlur"]}],"INPUT.ENTER":{actions:["setFormattedValue","clearHint","invokeOnBlur"]}}},"before:spin":{tags:["focus"],effects:["trackButtonDisabled","waitForChangeDelay"],entry:ZVn([{guard:"isIncrementHint",actions:["increment"]},{guard:"isDecrementHint",actions:["decrement"]}]),on:{CHANGE_DELAY:{target:"spinning",guard:j9t("isInRange","spinOnPress")},"TRIGGER.PRESS_UP":[{guard:"isTouchPointer",target:"focused",actions:["clearHint"]},{target:"focused",actions:["focusInput","clearHint"]}]}},spinning:{tags:["focus"],effects:["trackButtonDisabled","spinValue"],on:{SPIN:[{guard:"isIncrementHint",actions:["increment"]},{guard:"isDecrementHint",actions:["decrement"]}],"TRIGGER.PRESS_UP":{target:"focused",actions:["focusInput","clearHint"]}}},scrubbing:{tags:["focus"],effects:["activatePointerLock","trackMousemove","setupVirtualCursor","preventTextSelection"],on:{"SCRUBBER.POINTER_UP":{target:"focused",actions:["focusInput","clearCursorPoint"]},"SCRUBBER.POINTER_MOVE":[{guard:"isIncrementHint",actions:["increment","setCursorPoint"]},{guard:"isDecrementHint",actions:["decrement","setCursorPoint"]}]}}},implementations:{guards:{clampValueOnBlur:({prop:n})=>n("clampValueOnBlur"),spinOnPress:({prop:n})=>!!n("spinOnPress"),isInRange:({computed:n})=>!n("isOutOfRange"),isDecrementHint:({context:n,event:r})=>(r.hint??n.get("hint"))==="decrement",isIncrementHint:({context:n,event:r})=>(r.hint??n.get("hint"))==="increment",isTouchPointer:({event:n})=>n.pointerType==="touch"},effects:{waitForChangeDelay({send:n}){const r=setTimeout(()=>{n({type:"CHANGE_DELAY"})},300);return()=>clearTimeout(r)},spinValue({send:n}){const r=setInterval(()=>{n({type:"SPIN"})},50);return()=>clearInterval(r)},trackFormControl({context:n,scope:r}){const o=q1e(r);return uye(o,{onFieldsetDisabledChange(s){n.set("fieldsetDisabled",s)},onFormReset(){n.set("value",n.initial("value"))}})},setupVirtualCursor({context:n,scope:r}){const o=n.get("scrubberCursorPoint");return HVn(r,o)},preventTextSelection({scope:n}){return GVn(n)},trackButtonDisabled({context:n,scope:r,send:o}){const s=n.get("hint"),c=zVn(r,s);return koe(c,{attributes:["disabled"],callback(){o({type:"TRIGGER.PRESS_UP",src:"attr"})}})},attachWheelListener({scope:n,send:r,prop:o}){const s=q1e(n);if(!s||!n.isActiveElement(s)||!o("allowMouseWheel"))return;function c(d){d.preventDefault();const h=Math.sign(d.deltaY)*-1;h===1?r({type:"VALUE.INCREMENT"}):h===-1&&r({type:"VALUE.DECREMENT"})}return Ul(s,"wheel",c,{passive:!1})},activatePointerLock({scope:n}){if(!cye())return eUn(n.getDoc())},trackMousemove({scope:n,send:r,context:o,computed:s}){const c=n.getDoc();function d(b){const y=o.get("scrubberCursorPoint"),E=s("isRtl"),k=qVn(n,{point:y,isRtl:E,event:b});k.hint&&r({type:"SCRUBBER.POINTER_MOVE",hint:k.hint,point:k.point})}function h(){r({type:"SCRUBBER.POINTER_UP"})}return Nie(Ul(c,"mousemove",d,!1),Ul(c,"mouseup",h,!1))}},actions:{focusInput({scope:n,prop:r}){if(!r("focusInputOnChange"))return;const o=q1e(n);n.isActiveElement(o)||ca(()=>o?.focus({preventScroll:!0}))},increment({context:n,event:r,prop:o,computed:s}){let c=nBn(s("valueAsNumber"),r.step??o("step"));o("allowOverflow")||(c=W1(c,o("min"),o("max"))),n.set("value",rB(c,{computed:s,prop:o}))},decrement({context:n,event:r,prop:o,computed:s}){let c=rBn(s("valueAsNumber"),r.step??o("step"));o("allowOverflow")||(c=W1(c,o("min"),o("max"))),n.set("value",rB(c,{computed:s,prop:o}))},setClampedValue({context:n,prop:r,computed:o}){const s=W1(o("valueAsNumber"),r("min"),r("max"));n.set("value",rB(s,{computed:o,prop:r}))},setRawValue({context:n,event:r,prop:o,computed:s}){let c=Kje(r.value,{computed:s,prop:o});o("allowOverflow")||(c=W1(c,o("min"),o("max"))),n.set("value",rB(c,{computed:s,prop:o}))},setValue({context:n,event:r}){const o=r.target?.value??r.value;n.set("value",o)},clearValue({context:n}){n.set("value","")},incrementToMax({context:n,prop:r,computed:o}){const s=rB(r("max"),{computed:o,prop:r});n.set("value",s)},decrementToMin({context:n,prop:r,computed:o}){const s=rB(r("min"),{computed:o,prop:r});n.set("value",s)},setHint({context:n,event:r}){n.set("hint",r.hint)},clearHint({context:n}){n.set("hint",null)},invokeOnFocus({computed:n,prop:r}){r("onFocusChange")?.({focused:!0,value:n("formattedValue"),valueAsNumber:n("valueAsNumber")})},invokeOnBlur({computed:n,prop:r}){r("onFocusChange")?.({focused:!1,value:n("formattedValue"),valueAsNumber:n("valueAsNumber")})},invokeOnInvalid({computed:n,prop:r,event:o}){if(o.type==="INPUT.CHANGE")return;const s=n("valueAsNumber")>r("max")?"rangeOverflow":"rangeUnderflow";r("onValueInvalid")?.({reason:s,value:n("formattedValue"),valueAsNumber:n("valueAsNumber")})},syncInputElement({context:n,event:r,computed:o,scope:s}){const c=r.type.endsWith("CHANGE")?n.get("value"):o("formattedValue"),d=q1e(s),h=WVn(d);ca(()=>{Soe(d,c),KVn(d,h)})},setFormattedValue({context:n,computed:r}){n.set("value",r("formattedValue"))},setCursorPoint({context:n,event:r}){n.set("scrubberCursorPoint",r.point)},clearCursorPoint({context:n}){n.set("scrubberCursorPoint",null)},setVirtualCursorPosition({context:n,scope:r}){const o=oZt(r),s=n.get("scrubberCursorPoint");!o||!s||(o.style.transform=`translate3d(${s.x}px, ${s.y}px, 0px)`)}}}});Wo()(["allowMouseWheel","allowOverflow","clampValueOnBlur","dir","disabled","focusInputOnChange","form","formatOptions","getRootNode","id","ids","inputMode","invalid","locale","max","min","name","onFocusChange","onValueChange","onValueInvalid","pattern","required","readOnly","spinOnPress","step","translations","value","defaultValue"]);var aZt=Vo("pinInput").parts("root","label","input","control");aZt.build();Wo()(["autoFocus","blurOnComplete","count","defaultValue","dir","disabled","form","getRootNode","id","ids","invalid","mask","name","onValueChange","onValueComplete","onValueInvalid","otp","pattern","placeholder","readOnly","required","selectOnFocus","translations","type","value"]);const[sZt,BI]=fu({name:"PopoverContext",hookName:"usePopoverContext",providerName:"<PopoverProvider />"}),cZt=Q.forwardRef((n,r)=>{const o=BI(),s=Fn(o.getAnchorProps(),n);return H.jsx(Xn.div,{...s,ref:r})});cZt.displayName="PopoverAnchor";const uZt=Q.forwardRef((n,r)=>{const o=BI(),s=Fn(o.getArrowProps(),n);return H.jsx(Xn.div,{...s,ref:r})});uZt.displayName="PopoverArrow";const lZt=Q.forwardRef((n,r)=>{const o=BI(),s=Fn(o.getArrowTipProps(),n);return H.jsx(Xn.div,{...s,ref:r})});lZt.displayName="PopoverArrowTip";const dZt=Q.forwardRef((n,r)=>{const o=BI(),s=Fn(o.getCloseTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});dZt.displayName="PopoverCloseTrigger";const fZt=Q.forwardRef((n,r)=>{const o=BI(),s=Yx(),c=Fn(o.getContentProps(),s.getPresenceProps(),n);return s.unmounted?null:H.jsx(Xn.div,{...c,ref:X4(s.ref,r)})});fZt.displayName="PopoverContent";const hZt=Q.forwardRef((n,r)=>{const o=BI(),s=Fn(o.getDescriptionProps(),n);return H.jsx(Xn.div,{...s,ref:r})});hZt.displayName="PopoverDescription";const pZt=Q.forwardRef((n,r)=>{const o=BI(),s=Fn(o.getIndicatorProps(),n);return H.jsx(Xn.div,{...s,ref:r})});pZt.displayName="PopoverIndicator";const gZt=Q.forwardRef((n,r)=>{const o=BI(),s=Yx(),c=Fn(o.getPositionerProps(),n);return s.unmounted?null:H.jsx(Xn.div,{...c,ref:r})});gZt.displayName="PopoverPositioner";var mZt=Vo("popover").parts("arrow","arrowTip","anchor","trigger","indicator","positioner","content","title","description","closeTrigger"),tI=mZt.build(),bZt=n=>n.ids?.anchor??`popover:${n.id}:anchor`,vZt=n=>n.ids?.trigger??`popover:${n.id}:trigger`,Qqe=n=>n.ids?.content??`popover:${n.id}:content`,wZt=n=>n.ids?.positioner??`popover:${n.id}:popper`,tWn=n=>n.ids?.arrow??`popover:${n.id}:arrow`,eVe=n=>n.ids?.title??`popover:${n.id}:title`,tVe=n=>n.ids?.description??`popover:${n.id}:desc`,nWn=n=>n.ids?.closeTrigger??`popover:${n.id}:close`,$9t=n=>n.getById(bZt(n)),DV=n=>n.getById(vZt(n)),MV=n=>n.getById(Qqe(n)),F9t=n=>n.getById(wZt(n)),rWn=n=>n.getById(eVe(n)),iWn=n=>n.getById(tVe(n));function oWn(n,r){const{state:o,context:s,send:c,computed:d,prop:h,scope:b}=n,y=o.matches("open"),E=s.get("currentPlacement"),k=d("currentPortalled"),x=s.get("renderedElements"),O=iYe({...h("positioning"),placement:E});return{portalled:k,open:y,setOpen(P){o.matches("open")!==P&&c({type:P?"OPEN":"CLOSE"})},reposition(P={}){c({type:"POSITIONING.SET",options:P})},getArrowProps(){return r.element({id:tWn(b),...tI.arrow.attrs,dir:h("dir"),style:O.arrow})},getArrowTipProps(){return r.element({...tI.arrowTip.attrs,dir:h("dir"),style:O.arrowTip})},getAnchorProps(){return r.element({...tI.anchor.attrs,dir:h("dir"),id:bZt(b)})},getTriggerProps(){return r.button({...tI.trigger.attrs,dir:h("dir"),type:"button","data-placement":E,id:vZt(b),"aria-haspopup":"dialog","aria-expanded":y,"data-state":y?"open":"closed","aria-controls":Qqe(b),onPointerDown(P){EW(P)&&cye()&&P.currentTarget.focus()},onClick(P){P.defaultPrevented||c({type:"TOGGLE"})},onBlur(P){c({type:"TRIGGER_BLUR",target:P.relatedTarget})}})},getIndicatorProps(){return r.element({...tI.indicator.attrs,dir:h("dir"),"data-state":y?"open":"closed"})},getPositionerProps(){return r.element({id:wZt(b),...tI.positioner.attrs,dir:h("dir"),style:O.floating})},getContentProps(){return r.element({...tI.content.attrs,dir:h("dir"),id:Qqe(b),tabIndex:-1,role:"dialog",hidden:!y,"data-state":y?"open":"closed","data-expanded":nr(y),"aria-labelledby":x.title?eVe(b):void 0,"aria-describedby":x.description?tVe(b):void 0,"data-placement":E})},getTitleProps(){return r.element({...tI.title.attrs,id:eVe(b),dir:h("dir")})},getDescriptionProps(){return r.element({...tI.description.attrs,id:tVe(b),dir:h("dir")})},getCloseTriggerProps(){return r.button({...tI.closeTrigger.attrs,dir:h("dir"),id:nWn(b),type:"button","aria-label":"close",onClick(P){P.defaultPrevented||(P.stopPropagation(),c({type:"CLOSE"}))}})}}}var aWn={props({props:n}){return{closeOnInteractOutside:!0,closeOnEscape:!0,autoFocus:!0,modal:!1,portalled:!0,...n,positioning:{placement:"bottom",...n.positioning}}},initialState({prop:n}){return n("open")||n("defaultOpen")?"open":"closed"},context({bindable:n}){return{currentPlacement:n(()=>({defaultValue:void 0})),renderedElements:n(()=>({defaultValue:{title:!0,description:!0}}))}},computed:{currentPortalled:({prop:n})=>!!n("modal")||!!n("portalled")},watch({track:n,prop:r,action:o}){n([()=>r("open")],()=>{o(["toggleVisibility"])})},entry:["checkRenderedElements"],states:{closed:{on:{"CONTROLLED.OPEN":{target:"open",actions:["setInitialFocus"]},TOGGLE:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen","setInitialFocus"]}],OPEN:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen","setInitialFocus"]}]}},open:{effects:["trapFocus","preventScroll","hideContentBelow","trackPositioning","trackDismissableElement","proxyTabFocus"],on:{"CONTROLLED.CLOSE":{target:"closed",actions:["setFinalFocus"]},CLOSE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose","setFinalFocus"]}],TOGGLE:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],"POSITIONING.SET":{actions:["reposition"]}}}},implementations:{guards:{isOpenControlled:({prop:n})=>n("open")!=null},effects:{trackPositioning({context:n,prop:r,scope:o}){n.set("currentPlacement",r("positioning").placement);const s=$9t(o)??DV(o);return g2(s,()=>F9t(o),{...r("positioning"),defer:!0,onComplete(d){n.set("currentPlacement",d.placement)}})},trackDismissableElement({send:n,prop:r,scope:o}){const s=()=>MV(o);let c=!0;return cU(s,{pointerBlocking:r("modal"),exclude:DV(o),defer:!0,onEscapeKeyDown(d){r("onEscapeKeyDown")?.(d),!r("closeOnEscape")&&d.preventDefault()},onInteractOutside(d){r("onInteractOutside")?.(d),!d.defaultPrevented&&(c=!(d.detail.focusable||d.detail.contextmenu),r("closeOnInteractOutside")||d.preventDefault())},onPointerDownOutside:r("onPointerDownOutside"),onFocusOutside:r("onFocusOutside"),persistentElements:r("persistentElements"),onRequestDismiss:r("onRequestDismiss"),onDismiss(){n({type:"CLOSE",src:"interact-outside",restoreFocus:c})}})},proxyTabFocus({prop:n,scope:r}){return n("modal")||!n("portalled")?void 0:iUn(()=>MV(r),{triggerElement:DV(r),defer:!0,onFocus(s){s.focus({preventScroll:!0})}})},hideContentBelow({prop:n,scope:r}){return n("modal")?sYe(()=>[MV(r),DV(r)],{defer:!0}):void 0},preventScroll({prop:n,scope:r}){if(n("modal"))return CXt(r.getDoc())},trapFocus({prop:n,scope:r}){return n("modal")?xXt(()=>MV(r),{initialFocus:()=>Rie({root:MV(r),getInitialEl:n("initialFocusEl"),enabled:n("autoFocus")})}):void 0}},actions:{reposition({event:n,prop:r,scope:o,context:s}){const c=$9t(o)??DV(o);g2(c,()=>F9t(o),{...r("positioning"),...n.options,defer:!0,listeners:!1,onComplete(h){s.set("currentPlacement",h.placement)}})},checkRenderedElements({context:n,scope:r}){ca(()=>{Object.assign(n.get("renderedElements"),{title:!!rWn(r),description:!!iWn(r)})})},setInitialFocus({prop:n,scope:r}){n("modal")||ca(()=>{Rie({root:MV(r),getInitialEl:n("initialFocusEl"),enabled:n("autoFocus")})?.focus({preventScroll:!0})})},setFinalFocus({event:n,scope:r}){const o=n.restoreFocus??n.previousEvent?.restoreFocus;o!=null&&!o||ca(()=>{DV(r)?.focus({preventScroll:!0})})},invokeOnOpen({prop:n,flush:r}){r(()=>{n("onOpenChange")?.({open:!0})})},invokeOnClose({prop:n,flush:r}){r(()=>{n("onOpenChange")?.({open:!1})})},toggleVisibility({event:n,send:r,prop:o}){r({type:o("open")?"CONTROLLED.OPEN":"CONTROLLED.CLOSE",previousEvent:n})}}}};Wo()(["autoFocus","closeOnEscape","closeOnInteractOutside","dir","getRootNode","id","ids","initialFocusEl","modal","onEscapeKeyDown","onFocusOutside","onInteractOutside","onOpenChange","onPointerDownOutside","onRequestDismiss","defaultOpen","open","persistentElements","portalled","positioning"]);const sWn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv(aWn,c);return oWn(d,ry)},cWn=n=>{const[r,{children:o,...s}]=MM(n),c=sWn(s),d=J4(Fn({present:c.open},r));return H.jsx(sZt,{value:c,children:H.jsx(jM,{value:d,children:o})})},uWn=n=>{const[r,{value:o,children:s}]=MM(n),c=J4(Fn({present:o.open},r));return H.jsx(sZt,{value:o,children:H.jsx(jM,{value:c,children:s})})},yZt=Q.forwardRef((n,r)=>{const o=BI(),s=Fn(o.getTitleProps(),n);return H.jsx(Xn.div,{...s,ref:r})});yZt.displayName="PopoverTitle";const EZt=Q.forwardRef((n,r)=>{const o=BI(),s=Yx(),c=Fn({...o.getTriggerProps(),"aria-controls":s.unmounted?void 0:o.getTriggerProps()["aria-controls"]},n);return H.jsx(Xn.button,{...c,ref:r})});EZt.displayName="PopoverTrigger";const lU=n=>{const{children:r,disabled:o}=n,[s,c]=Q.useState(n.container?.current),d=Q.useSyncExternalStore(dWn,()=>!1,()=>!0),{getRootNode:h}=qb();if(Q.useEffect(()=>{c(()=>n.container?.current)},[n.container]),d||o)return H.jsx(H.Fragment,{children:r});const b=s??lWn(h);return H.jsx(H.Fragment,{children:Q.Children.map(r,y=>dye.createPortal(y,b))})},lWn=n=>{const r=n?.(),o=r.getRootNode();return VW(o)?o:ny(r).body},dWn=()=>()=>{};var fYe=Vo("progress").parts("root","label","track","range","valueText","view","circle","circleTrack","circleRange");fYe.build();Wo()(["dir","getRootNode","id","ids","max","min","orientation","translations","value","onValueChange","defaultValue","formatOptions","locale"]);var SZt=Vo("qr-code").parts("root","frame","pattern","overlay","downloadTrigger");SZt.build();Wo()(["ids","defaultValue","value","id","encoding","dir","getRootNode","onValueChange","pixelSize"]);const[_Zt,_ye]=fu({name:"RadioGroupContext",hookName:"useRadioGroupContext",providerName:"<RadioGroupProvider />"}),[fWn,hWn]=fu({name:"RadioGroupItemContext",hookName:"useRadioGroupItemContext",providerName:"<RadioGroupItemProvider />"}),[pWn,kZt]=fu({name:"RadioGroupItemPropsContext",hookName:"useRadioGroupItemPropsContext",providerName:"<RadioGroupItemPropsProvider />"}),TZt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["value","disabled","invalid"]),c=_ye(),d=Fn(c.getItemProps(o),s),h=c.getItemState(o);return H.jsx(fWn,{value:h,children:H.jsx(pWn,{value:o,children:H.jsx(Xn.label,{...d,ref:r})})})});TZt.displayName="RadioGroupItem";const gWn=n=>n.children(hWn()),hYe=Q.forwardRef((n,r)=>{const o=_ye(),s=kZt(),c=Fn(o.getItemControlProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});hYe.displayName="RadioGroupItemControl";const xZt=Q.forwardRef((n,r)=>{const o=_ye(),s=kZt(),c=Fn(o.getItemTextProps(s),n);return H.jsx(Xn.span,{...c,ref:r})});xZt.displayName="RadioGroupItemText";const CZt=Q.forwardRef((n,r)=>{const o=_ye(),s=Fn(o.getLabelProps(),n);return H.jsx(Xn.label,{...s,ref:r})});CZt.displayName="RadioGroupLabel";var pYe=Vo("radio-group").parts("root","label","item","itemText","itemControl","indicator"),jV=pYe.build(),Kve=n=>n.ids?.root??`radio-group:${n.id}`,B9t=n=>n.ids?.label??`radio-group:${n.id}:label`,AZt=(n,r)=>n.ids?.item?.(r)??`radio-group:${n.id}:radio:${r}`,nVe=(n,r)=>n.ids?.itemHiddenInput?.(r)??`radio-group:${n.id}:radio:input:${r}`,mWn=(n,r)=>n.ids?.itemControl?.(r)??`radio-group:${n.id}:radio:control:${r}`,bWn=(n,r)=>n.ids?.itemLabel?.(r)??`radio-group:${n.id}:radio:label:${r}`,IZt=n=>n.ids?.indicator??`radio-group:${n.id}:indicator`,kye=n=>n.getById(Kve(n)),vWn=(n,r)=>n.getById(nVe(n,r)),wWn=n=>n.getById(IZt(n)),yWn=n=>kye(n)?.querySelector("input:not(:disabled)"),EWn=n=>kye(n)?.querySelector("input:not(:disabled):checked"),U9t=n=>{const o=`input[type=radio][data-ownedby='${CSS.escape(Kve(n))}']:not([disabled])`;return Pie(kye(n),o)},SWn=(n,r)=>{if(r)return n.getById(AZt(n,r))},_Wn=n=>({left:n?.offsetLeft??0,top:n?.offsetTop??0,width:n?.offsetWidth??0,height:n?.offsetHeight??0}),kWn=n=>({width:`${n.width}px`,height:`${n.height}px`,left:`${n.left}px`,top:`${n.top}px`});function OZt(n,r){const{context:o,send:s,computed:c,prop:d,scope:h,refs:b}=n,y=c("isDisabled"),E=d("readOnly");function k(P){return{value:P.value,invalid:!!P.invalid,disabled:!!P.disabled||y,checked:o.get("value")===P.value,focused:o.get("focusedValue")===P.value,focusVisible:b.get("focusVisibleValue")===P.value,hovered:o.get("hoveredValue")===P.value,active:o.get("activeValue")===P.value}}function x(P){const D=k(P);return{"data-focus":nr(D.focused),"data-focus-visible":nr(D.focusVisible),"data-disabled":nr(D.disabled),"data-readonly":nr(E),"data-state":D.checked?"checked":"unchecked","data-hover":nr(D.hovered),"data-invalid":nr(D.invalid),"data-orientation":d("orientation"),"data-ssr":nr(o.get("ssr"))}}const O=()=>{(EWn(h)??yWn(h))?.focus()};return{focus:O,value:o.get("value"),setValue(P){s({type:"SET_VALUE",value:P,isTrusted:!1})},clearValue(){s({type:"SET_VALUE",value:null,isTrusted:!1})},getRootProps(){return r.element({...jV.root.attrs,role:"radiogroup",id:Kve(h),"aria-labelledby":B9t(h),"data-orientation":d("orientation"),"data-disabled":nr(y),"aria-orientation":d("orientation"),dir:d("dir"),style:{position:"relative"}})},getLabelProps(){return r.element({...jV.label.attrs,dir:d("dir"),"data-orientation":d("orientation"),"data-disabled":nr(y),id:B9t(h),onClick:O})},getItemState:k,getItemProps(P){const D=k(P);return r.label({...jV.item.attrs,dir:d("dir"),id:AZt(h,P.value),htmlFor:nVe(h,P.value),...x(P),onPointerMove(){D.disabled||D.hovered||s({type:"SET_HOVERED",value:P.value,hovered:!0})},onPointerLeave(){D.disabled||s({type:"SET_HOVERED",value:null})},onPointerDown($){D.disabled||EW($)&&(D.focused&&$.pointerType==="mouse"&&$.preventDefault(),s({type:"SET_ACTIVE",value:P.value,active:!0}))},onPointerUp(){D.disabled||s({type:"SET_ACTIVE",value:null})},onClick(){!D.disabled&&cye()&&vWn(h,P.value)?.focus()}})},getItemTextProps(P){return r.element({...jV.itemText.attrs,dir:d("dir"),id:bWn(h,P.value),...x(P)})},getItemControlProps(P){const D=k(P);return r.element({...jV.itemControl.attrs,dir:d("dir"),id:mWn(h,P.value),"data-active":nr(D.active),"aria-hidden":!0,...x(P)})},getItemHiddenInputProps(P){const D=k(P);return r.input({"data-ownedby":Kve(h),id:nVe(h,P.value),type:"radio",name:d("name")||d("id"),form:d("form"),value:P.value,onClick($){if(E){$.preventDefault();return}$.currentTarget.checked&&s({type:"SET_VALUE",value:P.value,isTrusted:!0})},onBlur(){s({type:"SET_FOCUSED",value:null,focused:!1})},onFocus(){const $=qve();s({type:"SET_FOCUSED",value:P.value,focused:!0,focusVisible:$})},onKeyDown($){$.defaultPrevented||$.key===" "&&s({type:"SET_ACTIVE",value:P.value,active:!0})},onKeyUp($){$.defaultPrevented||$.key===" "&&s({type:"SET_ACTIVE",value:null})},disabled:D.disabled,defaultChecked:D.checked,style:mUn})},getIndicatorProps(){const P=o.get("indicatorRect");return r.element({id:IZt(h),...jV.indicator.attrs,dir:d("dir"),hidden:o.get("value")==null,"data-disabled":nr(y),"data-orientation":d("orientation"),style:{"--transition-property":"left, top, width, height","--left":P?.left,"--top":P?.top,"--width":P?.width,"--height":P?.height,position:"absolute",willChange:"var(--transition-property)",transitionProperty:"var(--transition-property)",transitionDuration:o.get("canIndicatorTransition")?"var(--transition-duration, 150ms)":"0ms",transitionTimingFunction:"var(--transition-timing-function)",[d("orientation")==="horizontal"?"left":"top"]:d("orientation")==="horizontal"?"var(--left)":"var(--top)"}})}}}var{not:TWn}=$I(),NZt={props({props:n}){return{orientation:"vertical",...n}},initialState(){return"idle"},context({prop:n,bindable:r}){return{value:r(()=>({defaultValue:n("defaultValue"),value:n("value"),onChange(o){n("onValueChange")?.({value:o})}})),activeValue:r(()=>({defaultValue:null})),focusedValue:r(()=>({defaultValue:null})),hoveredValue:r(()=>({defaultValue:null})),indicatorRect:r(()=>({defaultValue:{}})),canIndicatorTransition:r(()=>({defaultValue:!1})),fieldsetDisabled:r(()=>({defaultValue:!1})),ssr:r(()=>({defaultValue:!0}))}},refs(){return{indicatorCleanup:null,focusVisibleValue:null}},computed:{isDisabled:({prop:n,context:r})=>!!n("disabled")||r.get("fieldsetDisabled")},entry:["syncIndicatorRect","syncSsr"],exit:["cleanupObserver"],effects:["trackFormControlState","trackFocusVisible"],watch({track:n,action:r,context:o}){n([()=>o.get("value")],()=>{r(["setIndicatorTransition","syncIndicatorRect","syncInputElements"])})},on:{SET_VALUE:[{guard:TWn("isTrusted"),actions:["setValue","dispatchChangeEvent"]},{actions:["setValue"]}],SET_HOVERED:{actions:["setHovered"]},SET_ACTIVE:{actions:["setActive"]},SET_FOCUSED:{actions:["setFocused"]}},states:{idle:{}},implementations:{guards:{isTrusted:({event:n})=>!!n.isTrusted},effects:{trackFormControlState({context:n,scope:r}){return uye(kye(r),{onFieldsetDisabledChange(o){n.set("fieldsetDisabled",o)},onFormReset(){n.set("value",n.initial("value"))}})},trackFocusVisible({scope:n}){return sXt({root:n.getRootNode?.()})}},actions:{setValue({context:n,event:r}){n.set("value",r.value)},setHovered({context:n,event:r}){n.set("hoveredValue",r.value)},setActive({context:n,event:r}){n.set("activeValue",r.value)},setFocused({context:n,event:r,refs:o}){n.set("focusedValue",r.value),o.set("focusVisibleValue",r.focusVisible?r.value:null)},syncInputElements({context:n,scope:r}){U9t(r).forEach(s=>{s.checked=s.value===n.get("value")})},setIndicatorTransition({context:n}){n.set("canIndicatorTransition",rie(n.get("value")))},cleanupObserver({refs:n}){n.get("indicatorCleanup")?.()},syncSsr({context:n}){n.set("ssr",!1)},syncIndicatorRect({context:n,scope:r,refs:o}){if(o.get("indicatorCleanup")?.(),!wWn(r))return;const s=n.get("value"),c=SWn(r,s);if(s==null||!c){n.set("canIndicatorTransition",!1),n.set("indicatorRect",{});return}const d=cUn([c],{measure(h){return _Wn(h)},onEntry({rects:h}){n.set("indicatorRect",kWn(h[0]))}});o.set("indicatorCleanup",d)},dispatchChangeEvent({context:n,scope:r}){U9t(r).forEach(s=>{const c=s.value===n.get("value");c!==s.checked&&$Bn(s,{checked:c})})}}}};Wo()(["dir","disabled","form","getRootNode","id","ids","name","onValueChange","orientation","readOnly","value","defaultValue"]);Wo()(["value","disabled","invalid"]);const xWn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv(NZt,c);return OZt(d,ry)},RZt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["defaultValue","disabled","form","id","ids","name","onValueChange","orientation","readOnly","value"]),c=xWn(o),d=Fn(c.getRootProps(),s);return H.jsx(_Zt,{value:c,children:H.jsx(Xn.div,{...d,ref:r})})});RZt.displayName="RadioGroupRoot";const PZt=Q.forwardRef((n,r)=>{const[{value:o},s]=Ou()(n,["value"]),c=Fn(o.getRootProps(),s);return H.jsx(_Zt,{value:o,children:H.jsx(Xn.div,{...c,ref:r})})});PZt.displayName="RadioGroupRootProvider";var LZt=Vo("rating-group").parts("root","label","item","control");LZt.build();Wo()(["allowHalf","autoFocus","count","dir","disabled","form","getRootNode","id","ids","name","onHoverChange","onValueChange","required","readOnly","translations","value","defaultValue"]);Wo()(["index"]);const[DZt,Coe]=fu({name:"ScrollAreaContext",hookName:"useScrollAreaContext",providerName:"<ScrollAreaProvider />"}),MZt=Q.forwardRef((n,r)=>{const o=Coe(),s=Fn(o.getContentProps(),n);return H.jsx(Xn.div,{...s,ref:r})});MZt.displayName="ScrollAreaContent";const jZt=Q.forwardRef((n,r)=>{const o=Coe(),s=Fn(o.getCornerProps(),n);return H.jsx(Xn.div,{...s,ref:r})});jZt.displayName="ScrollAreaCorner";var CWn=Object.defineProperty,AWn=(n,r,o)=>r in n?CWn(n,r,{enumerable:!0,configurable:!0,writable:!0,value:o}):n[r]=o,Yje=(n,r,o)=>AWn(n,typeof r!="symbol"?r+"":r,o),$Zt=Vo("scroll-area").parts("root","viewport","content","scrollbar","thumb","corner"),$V=$Zt.build(),Nx=n=>n.ids?.root??`scroll-area-${n.id}`,FZt=n=>n.ids?.viewport??`scroll-area-${n.id}:viewport`,BZt=n=>n.ids?.content??`scroll-area-${n.id}:content`,gYe=n=>n.getById(Nx(n)),aI=n=>n.getById(FZt(n)),IWn=n=>n.getById(BZt(n)),Pre=n=>vK(gYe(n),`[data-part=scrollbar][data-orientation=horizontal][data-ownedby="${Nx(n)}"]`),Lre=n=>vK(gYe(n),`[data-part=scrollbar][data-orientation=vertical][data-ownedby="${Nx(n)}"]`),Xje=n=>vK(Pre(n),`[data-part=thumb][data-orientation=horizontal][data-ownedby="${Nx(n)}"]`),Jje=n=>vK(Lre(n),`[data-part=thumb][data-orientation=vertical][data-ownedby="${Nx(n)}"]`),OWn=n=>vK(gYe(n),`[data-part=corner][data-ownedby="${Nx(n)}"]`);function NWn(n,r){if(!n)return RWn;let o=0,s=0;const c=n.scrollWidth-n.clientWidth;c>r&&(o=Math.min(1,Math.max(0,n.scrollLeft/c)));const d=n.scrollHeight-n.clientHeight;return d>r&&(s=Math.min(1,Math.max(0,n.scrollTop/d))),{x:o,y:s}}var RWn={x:0,y:0},PWn=300,LWn=n=>n*(2-n);function UZt(n,r={}){const{top:o,left:s,duration:c=PWn,easing:d=LWn,onComplete:h}=r;if(!n)return;const b={startTime:0,startScrollTop:n.scrollTop,startScrollLeft:n.scrollLeft,targetScrollTop:o??n.scrollTop,targetScrollLeft:s??n.scrollLeft};let y=!1;const E=()=>{b.rafId&&(cancelAnimationFrame(b.rafId),b.rafId=void 0),y=!0},k=x=>{if(y)return;b.startTime===0&&(b.startTime=x);const O=x-b.startTime,P=Math.min(O/c,1),D=d(P),$=b.targetScrollTop-b.startScrollTop,U=b.targetScrollLeft-b.startScrollLeft;n.scrollTop=b.startScrollTop+$*D,n.scrollLeft=b.startScrollLeft+U*D,P<1?b.rafId=requestAnimationFrame(k):h?.()};return b.rafId=requestAnimationFrame(k),E}function DWn(n,r={}){if(!n)return;const{top:o,left:s,behavior:c="smooth",easing:d,duration:h}=r;if(c==="smooth")UZt(n,{top:o,left:s,easing:d,duration:h});else{const b=bK({behavior:c,top:o,left:s});n.scrollTo(b)}}function MWn(n,r,o,s="smooth",c,d){if(!n)return;const h=n.scrollWidth-n.clientWidth,b=n.scrollHeight-n.clientHeight,y=o==="rtl";let E,k;switch(r){case"top":E=0;break;case"bottom":E=b;break;case"left":y?k=n.scrollLeft<=0?-h:0:k=0;break;case"right":y?k=n.scrollLeft<=0?0:h:k=h;break}if(s==="smooth")UZt(n,{top:E,left:k,easing:c,duration:d});else{const x=bK({left:k,top:E,behavior:s});n.scrollTo(x)}}function jWn(n,r){const{state:o,send:s,context:c,prop:d,scope:h}=n,b=o.matches("dragging"),y=c.get("hovering"),E=c.get("cornerSize"),k=c.get("thumbSize"),x=c.get("hiddenState"),O=c.get("atSides");return{isAtTop:O.top,isAtBottom:O.bottom,isAtLeft:O.left,isAtRight:O.right,hasOverflowX:!x.scrollbarXHidden,hasOverflowY:!x.scrollbarYHidden,getScrollProgress(){return NWn(aI(h),0)},scrollToEdge(P){const{edge:D,behavior:$}=P;return MWn(aI(h),D,d("dir"),$)},scrollTo(P){return DWn(aI(h),P)},getScrollbarState(P){const D=P.orientation==="horizontal";return{hovering:y,dragging:b,scrolling:c.get(D?"scrollingX":"scrollingY"),hidden:D?x.scrollbarXHidden:x.scrollbarYHidden}},getRootProps(){return r.element({...$V.root.attrs,id:Nx(h),dir:d("dir"),role:"presentation","data-overflow-x":nr(!x.scrollbarXHidden),"data-overflow-y":nr(!x.scrollbarYHidden),onPointerEnter(P){const D=Yp(P);f2(P.currentTarget,D)&&s({type:"root.pointerenter",pointerType:P.pointerType})},onPointerMove(P){const D=Yp(P);f2(P.currentTarget,D)&&s({type:"root.pointerenter",pointerType:P.pointerType})},onPointerDown({pointerType:P}){s({type:"root.pointerdown",pointerType:P})},onPointerLeave(P){f2(P.currentTarget,P.relatedTarget)||s({type:"root.pointerleave"})},style:{position:"relative","--corner-width":I1e(E?.width),"--corner-height":I1e(E?.height),"--thumb-width":I1e(k?.width),"--thumb-height":I1e(k?.height)}})},getViewportProps(){const P=()=>{s({type:"user.scroll"})};return r.element({...$V.viewport.attrs,role:"presentation","data-ownedby":Nx(h),id:FZt(h),"data-at-top":nr(O.top),"data-at-bottom":nr(O.bottom),"data-at-left":nr(O.left),"data-at-right":nr(O.right),"data-overflow-x":nr(!x.scrollbarXHidden),"data-overflow-y":nr(!x.scrollbarYHidden),tabIndex:x.scrollbarXHidden||x.scrollbarYHidden?void 0:0,style:{overflow:"auto"},onScroll(D){s({type:"viewport.scroll",target:D.currentTarget})},onWheel:P,onTouchMove:P,onPointerMove:P,onPointerEnter:P,onKeyDown:P})},getContentProps(){return r.element({...$V.content.attrs,id:BZt(h),role:"presentation","data-overflow-x":nr(!x.scrollbarXHidden),"data-overflow-y":nr(!x.scrollbarYHidden),style:{minWidth:"fit-content"}})},getScrollbarProps(P={}){const{orientation:D="vertical"}=P;return r.element({...$V.scrollbar.attrs,"data-ownedby":Nx(h),"data-orientation":D,"data-scrolling":nr(c.get(D==="horizontal"?"scrollingX":"scrollingY")),"data-hover":nr(y),"data-dragging":nr(b),"data-overflow-x":nr(!x.scrollbarXHidden),"data-overflow-y":nr(!x.scrollbarYHidden),onPointerUp(){s({type:"scrollbar.pointerup",orientation:D})},onPointerDown($){if($.button!==0||$.currentTarget!==$.target)return;const U=OB($);s({type:"scrollbar.pointerdown",orientation:D,point:U}),$.stopPropagation()},style:{position:"absolute",touchAction:"none",WebkitUserSelect:"none",userSelect:"none",...D==="vertical"&&{top:0,bottom:"var(--corner-height)",insetInlineEnd:0},...D==="horizontal"&&{insetInlineStart:0,insetInlineEnd:"var(--corner-width)",bottom:0}}})},getThumbProps(P={}){const{orientation:D="vertical"}=P;return r.element({...$V.thumb.attrs,"data-ownedby":Nx(h),"data-orientation":D,"data-hover":nr(y),"data-dragging":nr(b),onPointerDown($){if($.button!==0)return;const U=OB($);s({type:"thumb.pointerdown",orientation:D,point:U})},style:{...D==="vertical"&&{height:"var(--thumb-height)"},...D==="horizontal"&&{width:"var(--thumb-width)"}}})},getCornerProps(){return r.element({...$V.corner.attrs,"data-ownedby":Nx(h),"data-hover":nr(y),"data-state":x.cornerHidden?"hidden":"visible","data-overflow-x":nr(!x.scrollbarXHidden),"data-overflow-y":nr(!x.scrollbarYHidden),style:{position:"absolute",bottom:0,insetInlineEnd:0,width:"var(--corner-width)",height:"var(--corner-height)"}})}}}function O_(n,r,o){if(!n)return 0;const s=qB(n),c=o==="x"?"Left":"Top",d=o==="x"?"Right":"Bottom";return parseFloat(s[`${r}${c}`])+parseFloat(s[`${r}${d}`])}function z9t(n,r){const o=n.scrollTop,s=n.scrollLeft,c=r==="rtl",d=1,h=n.scrollHeight-n.clientHeight>d,b=n.scrollWidth-n.clientWidth>d,y=n.scrollWidth-n.clientWidth,E=n.scrollHeight-n.clientHeight;let k=!1,x=!1,O=!1,P=!1;return b&&(c&&s<=0?(k=Math.abs(s)>=y-d,x=Math.abs(s)<=d):(k=s<=d,x=s>=y-d)),h&&(O=o<=d,P=o>=E-d),{top:O,right:x,bottom:P,left:k}}var are=0,Zje=class{constructor(){Yje(this,"currentId",are),Yje(this,"clear",()=>{this.currentId!==are&&(clearTimeout(this.currentId),this.currentId=are)}),Yje(this,"disposeEffect",()=>this.clear)}start(n,r){this.clear(),this.currentId=setTimeout(()=>{this.currentId=are,r()},n)}isStarted(){return this.currentId!==are}},H9t=20,V1e=1e3,$Wn={props({props:n}){return{id:"sv",...n}},context({bindable:n}){return{scrollingX:n(()=>({defaultValue:!1})),scrollingY:n(()=>({defaultValue:!1})),hovering:n(()=>({defaultValue:!1})),dragging:n(()=>({defaultValue:!1})),touchModality:n(()=>({defaultValue:!1})),atSides:n(()=>({defaultValue:{top:!0,right:!1,bottom:!1,left:!0}})),cornerSize:n(()=>({defaultValue:{width:0,height:0}})),thumbSize:n(()=>({defaultValue:{width:0,height:0}})),hiddenState:n(()=>({defaultValue:{scrollbarYHidden:!1,scrollbarXHidden:!1,cornerHidden:!1},hash(r){return`Y:${r.scrollbarYHidden} X:${r.scrollbarXHidden} C:${r.cornerHidden}`}}))}},refs(){return{orientation:"vertical",scrollPosition:{x:0,y:0},scrollYTimeout:new Zje,scrollXTimeout:new Zje,scrollEndTimeout:new Zje,startX:0,startY:0,startScrollTop:0,startScrollLeft:0,programmaticScroll:!0}},initialState(){return"idle"},watch({track:n,prop:r,context:o,send:s}){n([()=>r("dir"),()=>o.hash("hiddenState")],()=>{s({type:"thumb.measure"})})},effects:["trackContentResize","trackViewportVisibility","trackWheelEvent"],entry:["checkHovering"],exit:["clearTimeouts"],on:{"thumb.measure":{actions:["setThumbSize"]},"viewport.scroll":{actions:["setThumbSize","setScrolling","setProgrammaticScroll"]},"root.pointerenter":{actions:["setTouchModality","setHovering"]},"root.pointerdown":{actions:["setTouchModality"]},"root.pointerleave":{actions:["clearHovering"]}},states:{idle:{on:{"scrollbar.pointerdown":{target:"dragging",actions:["scrollToPointer","startDragging"]},"thumb.pointerdown":{target:"dragging",actions:["startDragging"]}}},dragging:{effects:["trackPointerMove"],on:{"thumb.pointermove":{actions:["setDraggingScroll"]},"scrollbar.pointerup":{target:"idle",actions:["stopDragging"]},"thumb.pointerup":{target:"idle",actions:["clearScrolling","stopDragging"]}}}},implementations:{actions:{setTouchModality({context:n,event:r}){n.set("touchModality",r.pointerType==="touch")},setHovering({context:n}){n.set("hovering",!0)},clearHovering({context:n}){n.set("hovering",!1)},setProgrammaticScroll({refs:n}){n.get("scrollEndTimeout").start(100,()=>{n.set("programmaticScroll",!0)})},clearScrolling({context:n,event:r}){n.set(r.orientation==="vertical"?"scrollingY":"scrollingX",!1)},setThumbSize({context:n,scope:r,prop:o}){const s=aI(r);if(!s)return;const c=s.scrollHeight,d=s.scrollWidth;if(c===0||d===0)return;const h=Lre(r),b=Pre(r),y=Jje(r),E=Xje(r),k=s.clientHeight,x=s.clientWidth,O=s.scrollTop,P=s.scrollLeft,D=k>=c,$=x>=d,U=x/d,B=k/c,G=$?0:x,K=D?0:k,X=O_(b,"padding","x"),ae=O_(h,"padding","y"),oe=O_(E,"margin","x"),he=O_(y,"margin","y"),ye=G-X-oe,we=K-ae-he,ge=b?Math.min(b.offsetWidth,ye):ye,Te=h?Math.min(h.offsetHeight,we):we,xe=Math.max(H9t,ge*U),Ue=Math.max(H9t,Te*B);if(n.set("thumbSize",Re=>Re.height===Ue&&Re.width===xe?Re:{width:xe,height:Ue}),h&&y){const Re=h.offsetHeight-Ue-ae-he,_e=O/(c-k),qe=Math.min(Re,Math.max(0,_e*Re));y.style.transform=`translate3d(0,${qe}px,0)`}if(b&&E){const Re=b.offsetWidth-xe-X-oe,_e=P/(d-x),qe=o("dir")==="rtl"?W1(_e*Re,-Re,0):W1(_e*Re,0,Re);E.style.transform=`translate3d(${qe}px,0,0)`}if(OWn(r)){if($||D)n.set("cornerSize",{width:0,height:0});else if(!$&&!D){const Re=h?.offsetWidth||0,_e=b?.offsetHeight||0;n.set("cornerSize",{width:Re,height:_e})}}n.set("hiddenState",Re=>{const _e=D||$;return Re.scrollbarYHidden===D&&Re.scrollbarXHidden===$&&Re.cornerHidden===_e?Re:{scrollbarYHidden:D,scrollbarXHidden:$,cornerHidden:_e}}),n.set("atSides",Re=>{const _e=z9t(s,o("dir"));return d2(Re,_e)?Re:_e})},checkHovering({scope:n,context:r}){aI(n)?.matches(":hover")&&r.set("hovering",!0)},setScrolling({event:n,refs:r,context:o,prop:s}){const c={x:n.target.scrollLeft,y:n.target.scrollTop},d=r.get("scrollPosition"),h=c.x-d.x,b=c.y-d.y;r.set("scrollPosition",c),o.set("atSides",y=>{const E=z9t(n.target,s("dir"));return d2(y,E)?y:E}),b!==0&&(o.set("scrollingY",!0),r.get("scrollYTimeout").start(V1e,()=>{o.set("scrollingY",!1)})),h!==0&&(o.set("scrollingX",!0),r.get("scrollXTimeout").start(V1e,()=>{o.set("scrollingX",!1)}))},scrollToPointer({event:n,scope:r,prop:o}){const s=aI(r);if(!s)return;const c=Jje(r),d=Lre(r),h=Xje(r),b=Pre(r),y=n.point;if(c&&d&&n.orientation==="vertical"){const E=O_(c,"margin","y"),k=O_(d,"padding","y"),x=c.offsetHeight,O=d.getBoundingClientRect(),P=y.y-O.top-x/2-k+E/2,D=s.scrollHeight,$=s.clientHeight,U=d.offsetHeight-x-k-E,G=P/U*(D-$);s.scrollTop=G}if(h&&b&&n.orientation==="horizontal"){const E=O_(h,"margin","x"),k=O_(b,"padding","x"),x=h.offsetWidth,O=b.getBoundingClientRect(),P=y.x-O.left-x/2-k+E/2,D=s.scrollWidth,$=s.clientWidth,U=b.offsetWidth-x-k-E,B=P/U;let G;o("dir")==="rtl"?(G=(1-B)*(D-$),s.scrollLeft<=0&&(G=-G)):G=B*(D-$),s.scrollLeft=G}},startDragging({event:n,refs:r,scope:o}){r.set("startX",n.point.x),r.set("startY",n.point.y),r.set("orientation",n.orientation);const s=aI(o);s&&(r.set("startScrollTop",s.scrollTop),r.set("startScrollLeft",s.scrollLeft))},setDraggingScroll({event:n,refs:r,scope:o,context:s}){const c=r.get("startY"),d=r.get("startX"),h=r.get("startScrollTop"),b=r.get("startScrollLeft"),y=n.point,E=y.y-c,k=y.x-d,x=aI(o);if(!x)return;const O=x.scrollHeight,P=x.clientHeight,D=x.scrollWidth,$=x.clientWidth,U=r.get("orientation"),B=Jje(o),G=Lre(o);if(B&&G&&U==="vertical"){const ae=O_(G,"padding","y"),oe=O_(B,"margin","y"),he=B.offsetHeight,ye=G.offsetHeight-he-ae-oe,we=E/ye;x.scrollTop=h+we*(O-P),s.set("scrollingY",!0),r.get("scrollYTimeout").start(V1e,()=>{s.set("scrollingY",!1)})}const K=Xje(o),X=Pre(o);if(K&&X&&U==="horizontal"){const ae=O_(X,"padding","x"),oe=O_(K,"margin","x"),he=K.offsetWidth,ye=X.offsetWidth-he-ae-oe,we=k/ye;x.scrollLeft=b+we*(D-$),s.set("scrollingX",!0),r.get("scrollXTimeout").start(V1e,()=>{s.set("scrollingX",!1)})}},stopDragging({refs:n}){n.set("orientation",null)},clearTimeouts({refs:n}){n.get("scrollYTimeout").clear(),n.get("scrollXTimeout").clear(),n.get("scrollEndTimeout").clear()}},effects:{trackContentResize({scope:n,send:r}){const o=IWn(n);if(!o)return;const s=n.getWin(),c=new s.ResizeObserver(()=>{setTimeout(()=>{r({type:"thumb.measure"})},1)});return c.observe(o),()=>{c.disconnect()}},trackViewportVisibility({scope:n,send:r}){const o=n.getWin(),s=aI(n);if(!s)return;const c=new o.IntersectionObserver(d=>{d.forEach(h=>{h.intersectionRatio>0&&(r({type:"thumb.measure"}),c.disconnect())})});return c.observe(s),()=>{c.disconnect()}},trackWheelEvent({scope:n}){const r=Lre(n),o=Pre(n);if(!r&&!o)return;const s=c=>{const d=aI(n);if(!d||c.ctrlKey)return;const h=c.currentTarget.dataset.orientation;if(h==="vertical"){const b=d.scrollHeight>d.clientHeight,y=d.scrollTop===0&&c.deltaY<0,E=d.scrollTop===d.scrollHeight-d.clientHeight&&c.deltaY>0;if(!(b&&c.deltaY!==0&&!(y||E)))return;c.preventDefault(),d.scrollTop+=c.deltaY}else if(h==="horizontal"){const b=d.scrollWidth>d.clientWidth,y=d.scrollLeft===0&&c.deltaX<0,E=d.scrollLeft===d.scrollWidth-d.clientWidth&&c.deltaX>0;if(!(b&&c.deltaX!==0&&!(y||E)))return;c.preventDefault(),d.scrollLeft+=c.deltaX}};return Nie(r&&Ul(r,"wheel",s,{passive:!1}),o&&Ul(o,"wheel",s,{passive:!1}))},trackPointerMove({scope:n,send:r,refs:o}){const s=n.getDoc(),c=o.get("orientation");return SYt(s,{onPointerMove({point:d}){r({type:"thumb.pointermove",orientation:c,point:d})},onPointerUp(){r({type:"thumb.pointerup",orientation:c})}})}}}};Wo()(["dir","getRootNode","ids","id"]);const FWn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv($Wn,c);return jWn(d,ry)},zZt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["id","ids"]),c=FWn(o),d=Fn(c.getRootProps(),s);return H.jsx(DZt,{value:c,children:H.jsx(Xn.div,{...d,ref:r})})});zZt.displayName="ScrollAreaRoot";const HZt=Q.forwardRef((n,r)=>{const[{value:o},s]=Ou()(n,["value"]),c=Fn(o.getRootProps(),s);return H.jsx(DZt,{value:o,children:H.jsx(Xn.div,{...c,ref:r})})});HZt.displayName="ScrollAreaRootProvider";const[BWn,UWn]=fu({name:"ScrollAreaScrollbarContext",hookName:"useScrollAreaScrollbarContext",providerName:"<ScrollAreaScrollbarProvider />"}),GZt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["orientation"]),c=Coe(),d=Fn(c.getScrollbarProps(o),s);return H.jsx(BWn,{value:o,children:H.jsx(Xn.div,{...d,ref:r})})});GZt.displayName="ScrollAreaScrollbar";const qZt=Q.forwardRef((n,r)=>{const o=Coe(),s=UWn(),c=Fn(o.getThumbProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});qZt.displayName="ScrollAreaThumb";const VZt=Q.forwardRef((n,r)=>{const o=Coe(),s=Fn(o.getViewportProps(),n);return H.jsx(Xn.div,{...s,ref:r})});VZt.displayName="ScrollAreaViewport";const[WZt,Tye]=fu({name:"SegmentGroupContext",hookName:"useSegmentGroupContext",providerName:"<SegmentGroupProvider />"}),KZt=pYe.rename("segment-group"),Aoe=KZt.build(),YZt=Q.forwardRef((n,r)=>{const o=Tye(),s=Fn(o.getIndicatorProps(),Aoe.indicator.attrs,n);return H.jsx(Xn.div,{...s,ref:r})});YZt.displayName="SegmentGroupIndicator";const[zWn,Fni]=fu({name:"SegmentGroupItemContext",hookName:"useSegmentGroupItemContext",providerName:"<SegmentGroupItemProvider />"}),[HWn,XZt]=fu({name:"SegmentGroupItemPropsContext",hookName:"useSegmentGroupItemPropsContext",providerName:"<SegmentGroupItemPropsProvider />"}),JZt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["value","disabled","invalid"]),c=Tye(),d=Fn(c.getItemProps(o),Aoe.item.attrs,s),h=c.getItemState(o);return H.jsx(HWn,{value:o,children:H.jsx(zWn,{value:h,children:H.jsx(Xn.label,{...d,ref:r})})})});JZt.displayName="SegmentGroupItem";const ZZt=Q.forwardRef((n,r)=>{const o=Tye(),s=XZt(),c=Fn(o.getItemHiddenInputProps(s),n);return H.jsx(Xn.input,{...c,ref:r})});ZZt.displayName="SegmentGroupItemHiddenInput";const QZt=Q.forwardRef((n,r)=>{const o=Tye(),s=XZt(),c=Fn(o.getItemTextProps(s),Aoe.itemText.attrs,n);return H.jsx(Xn.span,{...c,ref:r})});QZt.displayName="SegmentGroupItemText";const GWn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv(NZt,c);return OZt(d,ry)},eQt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["defaultValue","disabled","form","id","ids","name","onValueChange","orientation","readOnly","value"]),c=GWn(o),d=Fn(c.getRootProps(),Aoe.root.attrs,s);return H.jsx(WZt,{value:c,children:H.jsx(Xn.div,{...d,ref:r})})});eQt.displayName="SegmentGroupRoot";const tQt=Q.forwardRef((n,r)=>{const[{value:o},s]=Ou()(n,["value"]),c=Fn(o.getRootProps(),Aoe.root.attrs,s);return H.jsx(WZt,{value:o,children:H.jsx(Xn.div,{...c,ref:r})})});tQt.displayName="SegmentGroupRootProvider";var nQt=Vo("select").parts("label","positioner","trigger","indicator","clearTrigger","item","itemText","itemIndicator","itemGroup","itemGroupLabel","list","content","root","control","valueText");nQt.build();var rQt=n=>new yye(n);rQt.empty=()=>new yye({items:[]});var qWn=n=>n.ids?.content??`select:${n.id}:content`,VWn=n=>n.ids?.trigger??`select:${n.id}:trigger`,WWn=n=>n.ids?.clearTrigger??`select:${n.id}:clear-trigger`,KWn=(n,r)=>n.ids?.item?.(r)??`select:${n.id}:option:${r}`,YWn=n=>n.ids?.hiddenSelect??`select:${n.id}:select`,XWn=n=>n.ids?.positioner??`select:${n.id}:positioner`,Qje=n=>n.getById(YWn(n)),sre=n=>n.getById(qWn(n)),W1e=n=>n.getById(VWn(n)),JWn=n=>n.getById(WWn(n)),G9t=n=>n.getById(XWn(n)),e$e=(n,r)=>r==null?null:n.getById(KWn(n,r)),{and:cre,not:iB,or:ZWn}=$I();ZWn("isTriggerArrowDownEvent","isTriggerEnterEvent"),cre(iB("multiple"),"hasSelectedItems"),iB("multiple"),cre(iB("multiple"),"hasSelectedItems"),iB("multiple"),iB("multiple"),iB("multiple"),iB("multiple"),cre("closeOnSelect","isOpenControlled"),cre("hasHighlightedItem","loop","isLastItemHighlighted"),cre("hasHighlightedItem","loop","isFirstItemHighlighted");function q9t(n){const r=n.restoreFocus??n.previousEvent?.restoreFocus;return r==null||!!r}Wo()(["closeOnSelect","collection","composite","defaultHighlightedValue","defaultOpen","defaultValue","deselectable","dir","disabled","form","getRootNode","highlightedValue","id","ids","invalid","loopFocus","multiple","name","onFocusOutside","onHighlightChange","onInteractOutside","onOpenChange","onPointerDownOutside","onSelect","onValueChange","open","positioning","readOnly","required","scrollToIndexFn","value"]);Wo()(["item","persistFocus"]);Wo()(["id"]);Wo()(["htmlFor"]);var iQt=Vo("slider").parts("root","label","thumb","valueText","track","range","control","markerGroup","marker","draggingIndicator");iQt.build();Wo()(["aria-label","aria-labelledby","dir","disabled","form","getAriaValueText","getRootNode","id","ids","invalid","max","min","minStepsBetweenThumbs","name","onFocusChange","onValueChange","onValueChangeEnd","orientation","origin","readOnly","step","thumbAlignment","thumbAlignment","thumbSize","value","defaultValue"]);Wo()(["index","name"]);const[oQt,Z4]=fu({name:"StepsContext",hookName:"useStepsContext",providerName:"<StepsProvider />"}),aQt=Q.forwardRef((n,r)=>{const o=Z4(),s=Fn(o.getContentProps({index:o.count}),n);return H.jsx(Xn.div,{...s,ref:r})});aQt.displayName="StepsCompletedContent";const sQt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["index"]),c=Z4(),d=Fn(c.getContentProps(o),s);return H.jsx(Xn.div,{...d,ref:r})});sQt.displayName="StepsContent";const[QWn,mYe]=fu({name:"StepsItemPropsContext",hookName:"useStepsItemPropsContext",providerName:"<StepsItemPropsProvider />"}),cQt=Q.forwardRef((n,r)=>{const o=Z4(),s=mYe(),c=Fn(o.getIndicatorProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});cQt.displayName="StepsIndicator";const[eKn,tKn]=fu({name:"StepsItemContext",hookName:"useStepsItemContext",providerName:"<StepsItem />"}),uQt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["index"]),c=Z4(),d=Fn(c.getItemProps(o),s),h=c.getItemState(o);return H.jsx(QWn,{value:o,children:H.jsx(eKn,{value:h,children:H.jsx(Xn.div,{...d,ref:r})})})});uQt.displayName="StepsItem";const bYe=n=>n.children(tKn());bYe.displayName="StepsItemContext";const lQt=Q.forwardRef((n,r)=>{const o=Z4(),s=Fn(o.getListProps(),n);return H.jsx(Xn.div,{...s,ref:r})});lQt.displayName="StepsList";const dQt=Q.forwardRef((n,r)=>{const o=Z4(),s=Fn(o.getNextTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});dQt.displayName="StepsNextTrigger";const fQt=Q.forwardRef((n,r)=>{const o=Z4(),s=Fn(o.getPrevTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});fQt.displayName="StepsPrevTrigger";var nKn=Vo("steps").parts("root","list","item","trigger","indicator","separator","content","nextTrigger","prevTrigger","progress"),nI=nKn.build(),rKn=n=>n.ids?.root??`steps:${n.id}`,iKn=n=>n.ids?.list??`steps:${n.id}:list`,V9t=(n,r)=>n.ids?.triggerId?.(r)??`steps:${n.id}:trigger:${r}`,oKn=(n,r)=>n.ids?.contentId?.(r)??`steps:${n.id}:content:${r}`;function aKn(n,r){const{context:o,send:s,computed:c,prop:d,scope:h}=n,b=o.get("step"),y=d("count"),E=c("percent"),k=c("hasNextStep"),x=c("hasPrevStep"),O=B=>({triggerId:V9t(h,B.index),contentId:oKn(h,B.index),current:B.index===b,completed:B.index<b,incomplete:B.index>b,index:B.index,first:B.index===0,last:B.index===y-1}),P=()=>{s({type:"STEP.NEXT",src:"next.trigger.click"})},D=()=>{s({type:"STEP.PREV",src:"prev.trigger.click"})},$=()=>{s({type:"STEP.RESET",src:"reset.trigger.click"})},U=B=>{s({type:"STEP.SET",value:B,src:"api.setValue"})};return{value:b,count:y,percent:E,hasNextStep:k,hasPrevStep:x,isCompleted:c("completed"),goToNextStep:P,goToPrevStep:D,resetStep:$,getItemState:O,setStep:U,getRootProps(){return r.element({...nI.root.attrs,id:rKn(h),dir:d("dir"),"data-orientation":d("orientation"),style:{"--percent":`${E}%`}})},getListProps(){const G=OFn(y).map((K,X)=>V9t(h,X));return r.element({...nI.list.attrs,dir:d("dir"),id:iKn(h),role:"tablist","aria-owns":G.join(" "),"aria-orientation":d("orientation"),"data-orientation":d("orientation")})},getItemProps(B){const G=O(B);return r.element({...nI.item.attrs,dir:d("dir"),"aria-current":G.current?"step":void 0,"data-orientation":d("orientation")})},getTriggerProps(B){const G=O(B);return r.button({...nI.trigger.attrs,id:G.triggerId,role:"tab",dir:d("dir"),tabIndex:!d("linear")||G.current?0:-1,"aria-selected":G.current,"aria-controls":G.contentId,"data-state":G.current?"open":"closed","data-orientation":d("orientation"),"data-complete":nr(G.completed),"data-current":nr(G.current),"data-incomplete":nr(G.incomplete),onClick(K){K.defaultPrevented||d("linear")||s({type:"STEP.SET",value:B.index,src:"trigger.click"})}})},getContentProps(B){const G=O(B);return r.element({...nI.content.attrs,dir:d("dir"),id:G.contentId,role:"tabpanel",tabIndex:0,hidden:!G.current,"data-state":G.current?"open":"closed","data-orientation":d("orientation"),"aria-labelledby":G.triggerId})},getIndicatorProps(B){const G=O(B);return r.element({...nI.indicator.attrs,dir:d("dir"),"aria-hidden":!0,"data-complete":nr(G.completed),"data-current":nr(G.current),"data-incomplete":nr(G.incomplete)})},getSeparatorProps(B){const G=O(B);return r.element({...nI.separator.attrs,dir:d("dir"),"data-orientation":d("orientation"),"data-complete":nr(G.completed),"data-current":nr(G.current),"data-incomplete":nr(G.incomplete)})},getNextTriggerProps(){return r.button({...nI.nextTrigger.attrs,dir:d("dir"),type:"button",disabled:!k,onClick(B){B.defaultPrevented||P()}})},getPrevTriggerProps(){return r.button({dir:d("dir"),...nI.prevTrigger.attrs,type:"button",disabled:!x,onClick(B){B.defaultPrevented||D()}})},getProgressProps(){return r.element({dir:d("dir"),...nI.progress.attrs,role:"progressbar","aria-valuenow":E,"aria-valuemin":0,"aria-valuemax":100,"aria-valuetext":`${E}% complete`,"data-complete":nr(E===100)})}}}var sKn={props({props:n}){return{defaultStep:0,count:1,linear:!1,orientation:"horizontal",...n}},context({prop:n,bindable:r}){return{step:r(()=>({defaultValue:n("defaultStep"),value:n("step"),onChange(o){n("onStepChange")?.({step:o}),o==n("count")&&n("onStepComplete")?.()}}))}},computed:{percent:({context:n,prop:r})=>n.get("step")/r("count")*100,hasNextStep:({context:n,prop:r})=>n.get("step")<r("count"),hasPrevStep:({context:n})=>n.get("step")>0,completed:({context:n,prop:r})=>n.get("step")===r("count")},initialState(){return"idle"},entry:["validateStep"],states:{idle:{on:{"STEP.SET":{actions:["setStep"]},"STEP.NEXT":{actions:["goToNextStep"]},"STEP.PREV":{actions:["goToPrevStep"]},"STEP.RESET":{actions:["resetStep"]}}}},implementations:{actions:{goToNextStep({context:n,prop:r}){const o=Math.min(n.get("step")+1,r("count"));n.set("step",o)},goToPrevStep({context:n}){const r=Math.max(n.get("step")-1,0);n.set("step",r)},resetStep({context:n}){n.set("step",0)},setStep({context:n,event:r}){n.set("step",r.value)},validateStep({context:n,prop:r}){cKn(r("count"),n.get("step"))}}}},cKn=(n,r)=>{if(!iYt(r,0,n))throw new RangeError(`[zag-js/steps] step index ${r} is out of bounds`)};Wo()(["count","dir","getRootNode","id","ids","linear","onStepChange","onStepComplete","orientation","step","defaultStep"]);function uKn(n){const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv(sKn,c);return aKn(d,ry)}const hQt=Q.forwardRef((n,r)=>{const[o,s]=Ou()(n,["defaultStep","id","ids","count","linear","onStepChange","onStepComplete","orientation","step"]),c=uKn(o),d=Fn(c.getRootProps(),s);return H.jsx(oQt,{value:c,children:H.jsx(Xn.div,{...d,ref:r})})});hQt.displayName="StepsRoot";const pQt=Q.forwardRef((n,r)=>{const[{value:o},s]=Ou()(n,["value"]),c=Fn(o.getRootProps(),s);return H.jsx(oQt,{value:o,children:H.jsx(Xn.div,{...c,ref:r,children:n.children})})});pQt.displayName="StepsRootProvider";const gQt=Q.forwardRef((n,r)=>{const o=Z4(),s=mYe(),c=Fn(o.getSeparatorProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});gQt.displayName="StepsSeparator";const mQt=Q.forwardRef((n,r)=>{const o=Z4(),s=mYe(),c=Fn(o.getTriggerProps(s),n);return H.jsx(Xn.button,{...c,ref:r})});mQt.displayName="StepsTrigger";var bQt=Vo("switch").parts("root","label","control","thumb");bQt.build();Wo()(["checked","defaultChecked","dir","disabled","form","getRootNode","id","ids","invalid","label","name","onCheckedChange","readOnly","required","value"]);var lKn=Vo("toast").parts("group","root","title","description","actionTrigger","closeTrigger"),iW=lKn.build(),dKn=n=>`toast-group:${n}`,W9t=(n,r)=>n.getById(`toast-group:${r}`),vQt=n=>`toast:${n.id}`,K9t=n=>n.getById(vQt(n)),Y9t=n=>`toast:${n.id}:title`,X9t=n=>`toast:${n.id}:description`,fKn=n=>`toast${n.id}:close`,J9t={info:5e3,error:5e3,success:2e3,loading:1/0,DEFAULT:5e3};function t$e(n,r){return n??J9t[r]??J9t.DEFAULT}var hKn=n=>typeof n=="string"?{left:n,right:n,bottom:n,top:n}:n;function pKn(n,r){const{prop:o,computed:s,context:c}=n,{offsets:d,gap:h}=o("store").attrs,b=c.get("heights"),y=hKn(d),E=o("dir")==="rtl",k=r.replace("-start",E?"-right":"-left").replace("-end",E?"-left":"-right"),x=k.includes("right"),O=k.includes("left"),P={position:"fixed",pointerEvents:s("count")>0?void 0:"none",display:"flex",flexDirection:"column","--gap":`${h}px`,"--first-height":`${b[0]?.height||0}px`,zIndex:cYt};let D="center";if(x&&(D="flex-end"),O&&(D="flex-start"),P.alignItems=D,k.includes("top")){const $=y.top;P.top=`max(env(safe-area-inset-top, 0px), ${$})`}if(k.includes("bottom")){const $=y.bottom;P.bottom=`max(env(safe-area-inset-bottom, 0px), ${$})`}if(!k.includes("left")){const $=y.right;P.insetInlineEnd=`calc(env(safe-area-inset-right, 0px) + ${$})`}if(!k.includes("right")){const $=y.left;P.insetInlineStart=`calc(env(safe-area-inset-left, 0px) + ${$})`}return P}function gKn(n,r){const{prop:o,context:s,computed:c}=n,d=o("parent"),h=d.computed("placement"),{gap:b}=d.prop("store").attrs,[y]=h.split("-"),E=s.get("mounted"),k=s.get("remainingTime"),x=c("height"),O=c("frontmost"),P=!O,D=!o("stacked"),$=o("stacked"),B=o("type")==="loading"?Number.MAX_SAFE_INTEGER:k,G=c("heightIndex")*b+c("heightBefore"),K={position:"absolute",pointerEvents:"auto","--opacity":"0","--remove-delay":`${o("removeDelay")}ms`,"--duration":`${B}ms`,"--initial-height":`${x}px`,"--offset":`${G}px`,"--index":o("index"),"--z-index":c("zIndex"),"--lift-amount":"calc(var(--lift) * var(--gap))","--y":"100%","--x":"0"},X=ae=>Object.assign(K,ae);return y==="top"?X({top:"0","--sign":"-1","--y":"-100%","--lift":"1"}):y==="bottom"&&X({bottom:"0","--sign":"1","--y":"100%","--lift":"-1"}),E&&(X({"--y":"0","--opacity":"1"}),$&&X({"--y":"calc(var(--lift) * var(--offset))","--height":"var(--initial-height)"})),r||X({"--opacity":"0",pointerEvents:"none"}),P&&D&&(X({"--base-scale":"var(--index) * 0.05 + 1","--y":"calc(var(--lift-amount) * var(--index))","--scale":"calc(-1 * var(--base-scale))","--height":"var(--first-height)"}),r||X({"--y":"calc(var(--sign) * 40%)"})),P&&$&&!r&&X({"--y":"calc(var(--lift) * var(--offset) + var(--lift) * -100%)"}),O&&!r&&X({"--y":"calc(var(--lift) * -100%)"}),K}function mKn(n,r){const{computed:o}=n,s={position:"absolute",inset:"0",scale:"1 2",pointerEvents:r?"none":"auto"},c=d=>Object.assign(s,d);return o("frontmost")&&!r&&c({height:"calc(var(--initial-height) + 80%)"}),s}function bKn(){return{position:"absolute",left:"0",height:"calc(var(--gap) + 2px)",bottom:"100%",width:"100%"}}function vKn(n,r){const{context:o,prop:s,send:c,refs:d,computed:h}=n;return{getCount(){return o.get("toasts").length},getToasts(){return o.get("toasts")},getGroupProps(b={}){const{label:y="Notifications"}=b,{hotkey:E}=s("store").attrs,k=E.join("+").replace(/Key/g,"").replace(/Digit/g,""),x=h("placement"),[O,P="center"]=x.split("-");return r.element({...iW.group.attrs,dir:s("dir"),tabIndex:-1,"aria-label":`${x} ${y} ${k}`,id:dKn(x),"data-placement":x,"data-side":O,"data-align":P,"aria-live":"polite",role:"region",style:pKn(n,x),onMouseMove(){c({type:"REGION.POINTER_ENTER",placement:x})},onMouseLeave(){c({type:"REGION.POINTER_LEAVE",placement:x})},onFocus(D){c({type:"REGION.FOCUS",target:D.relatedTarget})},onBlur(D){d.get("isFocusWithin")&&!f2(D.currentTarget,D.relatedTarget)&&queueMicrotask(()=>c({type:"REGION.BLUR"}))}})},subscribe(b){return s("store").subscribe(()=>b(o.get("toasts")))}}}var wKn={props({props:n}){return{dir:"ltr",id:eYt(),...n,store:n.store}},initialState({prop:n}){return n("store").attrs.overlap?"overlap":"stack"},refs(){return{lastFocusedEl:null,isFocusWithin:!1,dismissableCleanup:void 0}},context({bindable:n}){return{toasts:n(()=>({defaultValue:[],sync:!0,hash:r=>r.map(o=>o.id).join(",")})),heights:n(()=>({defaultValue:[],sync:!0}))}},computed:{count:({context:n})=>n.get("toasts").length,overlap:({prop:n})=>n("store").attrs.overlap,placement:({prop:n})=>n("store").attrs.placement},effects:["subscribeToStore","trackDocumentVisibility","trackHotKeyPress"],watch({track:n,context:r,action:o}){n([()=>r.hash("toasts")],()=>{queueMicrotask(()=>{o(["collapsedIfEmpty","setDismissableBranch"])})})},exit:["clearDismissableBranch","clearLastFocusedEl"],on:{"DOC.HOTKEY":{actions:["focusRegionEl"]},"REGION.BLUR":[{guard:"isOverlapping",target:"overlap",actions:["collapseToasts","resumeToasts","restoreLastFocusedEl"]},{target:"stack",actions:["resumeToasts","restoreLastFocusedEl"]}],"TOAST.REMOVE":{actions:["removeToast","removeHeight"]},"TOAST.PAUSE":{actions:["pauseToasts"]}},states:{stack:{on:{"REGION.POINTER_LEAVE":[{guard:"isOverlapping",target:"overlap",actions:["resumeToasts","collapseToasts"]},{actions:["resumeToasts"]}],"REGION.OVERLAP":{target:"overlap",actions:["collapseToasts"]},"REGION.FOCUS":{actions:["setLastFocusedEl","pauseToasts"]},"REGION.POINTER_ENTER":{actions:["pauseToasts"]}}},overlap:{on:{"REGION.STACK":{target:"stack",actions:["expandToasts"]},"REGION.POINTER_ENTER":{target:"stack",actions:["pauseToasts","expandToasts"]},"REGION.FOCUS":{target:"stack",actions:["setLastFocusedEl","pauseToasts","expandToasts"]}}}},implementations:{guards:{isOverlapping:({computed:n})=>n("overlap")},effects:{subscribeToStore({context:n,prop:r}){return r("store").subscribe(o=>{if(o.dismiss){n.set("toasts",s=>s.filter(c=>c.id!==o.id));return}n.set("toasts",s=>{const c=s.findIndex(d=>d.id===o.id);return c!==-1?[...s.slice(0,c),{...s[c],...o},...s.slice(c+1)]:[o,...s]})})},trackHotKeyPress({prop:n,send:r}){return Ul(document,"keydown",s=>{const{hotkey:c}=n("store").attrs;c.every(h=>s[h]||s.code===h)&&r({type:"DOC.HOTKEY"})},{capture:!0})},trackDocumentVisibility({prop:n,send:r,scope:o}){const{pauseOnPageIdle:s}=n("store").attrs;if(!s)return;const c=o.getDoc();return Ul(c,"visibilitychange",()=>{const d=c.visibilityState==="hidden";r({type:d?"PAUSE_ALL":"RESUME_ALL"})})}},actions:{setDismissableBranch({refs:n,context:r,computed:o,scope:s}){const c=r.get("toasts"),d=o("placement"),h=c.length>0;if(!h){n.get("dismissableCleanup")?.();return}if(h&&n.get("dismissableCleanup"))return;const y=OHn(()=>W9t(s,d),{defer:!0});n.set("dismissableCleanup",y)},clearDismissableBranch({refs:n}){n.get("dismissableCleanup")?.()},focusRegionEl({scope:n,computed:r}){queueMicrotask(()=>{W9t(n,r("placement"))?.focus()})},pauseToasts({prop:n}){n("store").pause()},resumeToasts({prop:n}){n("store").resume()},expandToasts({prop:n}){n("store").expand()},collapseToasts({prop:n}){n("store").collapse()},removeToast({prop:n,event:r}){n("store").remove(r.id)},removeHeight({event:n,context:r}){n?.id!=null&&queueMicrotask(()=>{r.set("heights",o=>o.filter(s=>s.id!==n.id))})},collapsedIfEmpty({send:n,computed:r}){!r("overlap")||r("count")>1||n({type:"REGION.OVERLAP"})},setLastFocusedEl({refs:n,event:r}){n.get("isFocusWithin")||!r.target||(n.set("isFocusWithin",!0),n.set("lastFocusedEl",r.target))},restoreLastFocusedEl({refs:n}){n.get("lastFocusedEl")&&(n.get("lastFocusedEl")?.focus({preventScroll:!0}),n.set("lastFocusedEl",null),n.set("isFocusWithin",!1))},clearLastFocusedEl({refs:n}){n.get("lastFocusedEl")&&(n.get("lastFocusedEl")?.focus({preventScroll:!0}),n.set("lastFocusedEl",null),n.set("isFocusWithin",!1))}}}};function yKn(n,r){const{state:o,send:s,prop:c,scope:d,context:h,computed:b}=n,y=o.hasTag("visible"),E=o.hasTag("paused"),k=h.get("mounted"),x=b("frontmost"),O=c("parent").computed("placement"),P=c("type"),D=c("stacked"),$=c("title"),U=c("description"),B=c("action"),[G,K="center"]=O.split("-");return{type:P,title:$,description:U,placement:O,visible:y,paused:E,closable:!!c("closable"),pause(){s({type:"PAUSE"})},resume(){s({type:"RESUME"})},dismiss(){s({type:"DISMISS",src:"programmatic"})},getRootProps(){return r.element({...iW.root.attrs,dir:c("dir"),id:vQt(d),"data-state":y?"open":"closed","data-type":P,"data-placement":O,"data-align":K,"data-side":G,"data-mounted":nr(k),"data-paused":nr(E),"data-first":nr(x),"data-sibling":nr(!x),"data-stack":nr(D),"data-overlap":nr(!D),role:"status","aria-atomic":"true","aria-describedby":U?X9t(d):void 0,"aria-labelledby":$?Y9t(d):void 0,tabIndex:0,style:gKn(n,y),onKeyDown(X){X.defaultPrevented||X.key=="Escape"&&(s({type:"DISMISS",src:"keyboard"}),X.preventDefault())}})},getGhostBeforeProps(){return r.element({"data-ghost":"before",style:mKn(n,y)})},getGhostAfterProps(){return r.element({"data-ghost":"after",style:bKn()})},getTitleProps(){return r.element({...iW.title.attrs,id:Y9t(d)})},getDescriptionProps(){return r.element({...iW.description.attrs,id:X9t(d)})},getActionTriggerProps(){return r.button({...iW.actionTrigger.attrs,type:"button",onClick(X){X.defaultPrevented||(B?.onClick?.(),s({type:"DISMISS",src:"user"}))}})},getCloseTriggerProps(){return r.button({id:fKn(d),...iW.closeTrigger.attrs,type:"button","aria-label":"Dismiss notification",onClick(X){X.defaultPrevented||s({type:"DISMISS",src:"user"})}})}}}var{not:EKn}=$I(),SKn={props({props:n}){return oBn(n,["id","type","parent","removeDelay"],"toast"),{closable:!0,...n,duration:t$e(n.duration,n.type)}},initialState({prop:n}){return n("type")==="loading"||n("duration")===1/0?"visible:persist":"visible"},context({prop:n,bindable:r}){return{remainingTime:r(()=>({defaultValue:t$e(n("duration"),n("type"))})),createdAt:r(()=>({defaultValue:Date.now()})),mounted:r(()=>({defaultValue:!1})),initialHeight:r(()=>({defaultValue:0}))}},refs(){return{closeTimerStartTime:Date.now(),lastCloseStartTimerStartTime:0}},computed:{zIndex:({prop:n})=>{const r=n("parent").context.get("toasts"),o=r.findIndex(s=>s.id===n("id"));return r.length-o},height:({prop:n})=>n("parent").context.get("heights").find(s=>s.id===n("id"))?.height??0,heightIndex:({prop:n})=>n("parent").context.get("heights").findIndex(o=>o.id===n("id")),frontmost:({prop:n})=>n("index")===0,heightBefore:({prop:n})=>{const r=n("parent").context.get("heights"),o=r.findIndex(s=>s.id===n("id"));return r.reduce((s,c,d)=>d>=o?s:s+c.height,0)},shouldPersist:({prop:n})=>n("type")==="loading"||n("duration")===1/0},watch({track:n,prop:r,send:o}){n([()=>r("message")],()=>{const s=r("message");s&&o({type:s,src:"programmatic"})}),n([()=>r("type"),()=>r("duration")],()=>{o({type:"UPDATE"})})},on:{UPDATE:[{guard:"shouldPersist",target:"visible:persist",actions:["resetCloseTimer"]},{target:"visible:updating",actions:["resetCloseTimer"]}],MEASURE:{actions:["measureHeight"]}},entry:["setMounted","measureHeight","invokeOnVisible"],effects:["trackHeight"],states:{"visible:updating":{tags:["visible","updating"],effects:["waitForNextTick"],on:{SHOW:{target:"visible"}}},"visible:persist":{tags:["visible","paused"],on:{RESUME:{guard:EKn("isLoadingType"),target:"visible",actions:["setCloseTimer"]},DISMISS:{target:"dismissing"}}},visible:{tags:["visible"],effects:["waitForDuration"],on:{DISMISS:{target:"dismissing"},PAUSE:{target:"visible:persist",actions:["syncRemainingTime"]}}},dismissing:{entry:["invokeOnDismiss"],effects:["waitForRemoveDelay"],on:{REMOVE:{target:"unmounted",actions:["notifyParentToRemove"]}}},unmounted:{entry:["invokeOnUnmount"]}},implementations:{effects:{waitForRemoveDelay({prop:n,send:r}){return Lje(()=>{r({type:"REMOVE",src:"timer"})},n("removeDelay"))},waitForDuration({send:n,context:r,computed:o}){if(!o("shouldPersist"))return Lje(()=>{n({type:"DISMISS",src:"timer"})},r.get("remainingTime"))},waitForNextTick({send:n}){return Lje(()=>{n({type:"SHOW",src:"timer"})},0)},trackHeight({scope:n,prop:r}){let o;return ca(()=>{const s=K9t(n);if(!s)return;const c=()=>{const b=s.style.height;s.style.height="auto";const y=s.getBoundingClientRect().height;s.style.height=b;const E={id:r("id"),height:y};Z9t(r("parent"),E)},d=n.getWin(),h=new d.MutationObserver(c);h.observe(s,{childList:!0,subtree:!0,characterData:!0}),o=()=>h.disconnect()}),()=>o?.()}},guards:{isLoadingType:({prop:n})=>n("type")==="loading",shouldPersist:({computed:n})=>n("shouldPersist")},actions:{setMounted({context:n}){ca(()=>{n.set("mounted",!0)})},measureHeight({scope:n,prop:r,context:o}){queueMicrotask(()=>{const s=K9t(n);if(!s)return;const c=s.style.height;s.style.height="auto";const d=s.getBoundingClientRect().height;s.style.height=c,o.set("initialHeight",d);const h={id:r("id"),height:d};Z9t(r("parent"),h)})},setCloseTimer({refs:n}){n.set("closeTimerStartTime",Date.now())},resetCloseTimer({context:n,refs:r,prop:o}){r.set("closeTimerStartTime",Date.now()),n.set("remainingTime",t$e(o("duration"),o("type")))},syncRemainingTime({context:n,refs:r}){n.set("remainingTime",o=>{const s=r.get("closeTimerStartTime"),c=Date.now()-s;return r.set("lastCloseStartTimerStartTime",Date.now()),o-c})},notifyParentToRemove({prop:n}){n("parent").send({type:"TOAST.REMOVE",id:n("id")})},invokeOnDismiss({prop:n,event:r}){n("onStatusChange")?.({status:"dismissing",src:r.src})},invokeOnUnmount({prop:n}){n("onStatusChange")?.({status:"unmounted"})},invokeOnVisible({prop:n}){n("onStatusChange")?.({status:"visible"})}}}};function Z9t(n,r){const{id:o,height:s}=r;n.context.set("heights",c=>c.find(h=>h.id===o)?c.map(h=>h.id===o?{...h,height:s}:h):[{id:o,height:s},...c])}var _Kn=(n,r)=>({...r,...bK(n)});function kKn(n){const r=_Kn(n,{placement:"bottom",overlap:!1,max:24,gap:16,offsets:"1rem",hotkey:["altKey","KeyT"],removeDelay:200,pauseOnPageIdle:!0});let o=[],s=[],c=new Set,d=[];const h=xe=>(o.push(xe),()=>{const Ue=o.indexOf(xe);o.splice(Ue,1)}),b=xe=>(o.forEach(Ue=>Ue(xe)),xe),y=xe=>{if(s.length>=r.max){d.push(xe);return}b(xe),s.unshift(xe)},E=()=>{for(;d.length>0&&s.length<r.max;){const xe=d.shift();xe&&(b(xe),s.unshift(xe))}},k=xe=>{const Ue=xe.id??`toast:${eYt()}`,Z=s.find(Re=>Re.id===Ue);return c.has(Ue)&&c.delete(Ue),Z?s=s.map(Re=>Re.id===Ue?b({...Re,...xe,id:Ue}):Re):y({id:Ue,duration:r.duration,removeDelay:r.removeDelay,type:"info",...xe,stacked:!r.overlap,gap:r.gap}),Ue},x=xe=>(c.add(xe),xe?(o.forEach(Ue=>Ue({id:xe,dismiss:!0})),s=s.filter(Ue=>Ue.id!==xe),E()):(s.forEach(Ue=>{o.forEach(Z=>Z({id:Ue.id,dismiss:!0}))}),s=[],d=[]),xe);return{attrs:r,subscribe:h,create:k,update:(xe,Ue)=>k({id:xe,...Ue}),remove:x,dismiss:xe=>{xe!=null?s=s.map(Ue=>Ue.id===xe?b({...Ue,message:"DISMISS"}):Ue):s=s.map(Ue=>b({...Ue,message:"DISMISS"}))},error:xe=>k({...xe,type:"error"}),success:xe=>k({...xe,type:"success"}),info:xe=>k({...xe,type:"info"}),warning:xe=>k({...xe,type:"warning"}),loading:xe=>k({...xe,type:"loading"}),getVisibleToasts:()=>s.filter(xe=>!c.has(xe.id)),getCount:()=>s.length,promise:(xe,Ue,Z={})=>{if(!Ue||!Ue.loading){qW("[zag-js > toast] toaster.promise() requires at least a 'loading' option to be specified");return}const Re=k({...Z,...Ue.loading,promise:xe,type:"loading"});let _e=!0,qe;const it=A1e(xe).then(async de=>{if(qe=["resolve",de],TKn(de)&&!de.ok){_e=!1;const We=A1e(Ue.error,`HTTP Error! status: ${de.status}`);k({...Z,...We,id:Re,type:"error"})}else if(Ue.success!==void 0){_e=!1;const We=A1e(Ue.success,de);k({...Z,...We,id:Re,type:"success"})}}).catch(async de=>{if(qe=["reject",de],Ue.error!==void 0){_e=!1;const We=A1e(Ue.error,de);k({...Z,...We,id:Re,type:"error"})}}).finally(()=>{_e&&x(Re),Ue.finally?.()});return{id:Re,unwrap:()=>new Promise((de,We)=>it.then(()=>qe[0]==="reject"?We(qe[1]):de(qe[1])).catch(We))}},pause:xe=>{xe!=null?s=s.map(Ue=>Ue.id===xe?b({...Ue,message:"PAUSE"}):Ue):s=s.map(Ue=>b({...Ue,message:"PAUSE"}))},resume:xe=>{xe!=null?s=s.map(Ue=>Ue.id===xe?b({...Ue,message:"RESUME"}):Ue):s=s.map(Ue=>b({...Ue,message:"RESUME"}))},isVisible:xe=>!c.has(xe)&&!!s.find(Ue=>Ue.id===xe),isDismissed:xe=>c.has(xe),expand:()=>{s=s.map(xe=>b({...xe,stacked:!0}))},collapse:()=>{s=s.map(xe=>b({...xe,stacked:!1}))}}}var TKn=n=>n&&typeof n=="object"&&"ok"in n&&typeof n.ok=="boolean"&&"status"in n&&typeof n.status=="number",Q9t={connect:vKn,machine:wKn};const xKn=n=>kKn(n),[CKn,EK]=fu({name:"ToastContext",hookName:"useToastContext",providerName:"<ToastProvider />"}),wQt=Q.forwardRef((n,r)=>{const o=EK(),s=Fn(o.getActionTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});wQt.displayName="ToastActionTrigger";const yQt=Q.forwardRef((n,r)=>{const o=EK(),s=Fn(o.getCloseTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});yQt.displayName="ToastCloseTrigger";const EQt=Q.forwardRef((n,r)=>{const o=EK(),s=Fn(o.getDescriptionProps(),n);return H.jsx(Xn.div,{...s,ref:r})});EQt.displayName="ToastDescription";const SQt=Q.forwardRef((n,r)=>{const o=EK(),s=Fn(o.getRootProps(),n);return H.jsxs("div",{...s,ref:r,children:[H.jsx("div",{...o.getGhostBeforeProps()}),n.children,H.jsx("div",{...o.getGhostAfterProps()})]})});SQt.displayName="ToastRoot";const _Qt=Q.forwardRef((n,r)=>{const o=EK(),s=Fn(o.getTitleProps(),n);return H.jsx(Xn.div,{...s,ref:r})});_Qt.displayName="ToastTitle";const kQt=Q.forwardRef((n,r)=>{const{toaster:o,children:s,...c}=n,d=b2(),h=qb(),b=uv(Q9t.machine,{store:o,id:Q.useId(),dir:d?.dir,getRootNode:h?.getRootNode}),y=Q9t.connect(b,ry),E=Fn(y.getGroupProps(),c);return H.jsx(Xn.div,{...E,ref:r,children:y.getToasts().map((k,x)=>H.jsx(TQt,{value:k,parent:b,index:x,children:O=>s(O)},k.id))})});kQt.displayName="Toaster";const TQt=n=>{const r=qb(),o={...n.value,parent:n.parent,index:n.index,getRootNode:r.getRootNode},s=uv(SKn,{...o}),c=yKn(s,ry);return H.jsx(CKn,{value:c,children:n.children(n.value)})};TQt.displayName="ToastActor";const[xQt,Ioe]=fu({name:"TooltipContext",hookName:"useTooltipContext",providerName:"<TooltipProvider />"}),CQt=Q.forwardRef((n,r)=>{const o=Ioe(),s=Fn(o.getArrowProps(),n);return H.jsx(Xn.div,{...s,ref:r})});CQt.displayName="TooltipArrow";const AQt=Q.forwardRef((n,r)=>{const o=Ioe(),s=Fn(o.getArrowTipProps(),n);return H.jsx(Xn.div,{...s,ref:r})});AQt.displayName="TooltipArrowTip";const IQt=Q.forwardRef((n,r)=>{const o=Ioe(),s=Yx(),c=Fn(o.getContentProps(),s.getPresenceProps(),n);return s.unmounted?null:H.jsx(Xn.div,{...c,ref:X4(s.ref,r)})});IQt.displayName="TooltipContent";const OQt=Q.forwardRef((n,r)=>{const o=Ioe(),s=Fn(o.getPositionerProps(),n);return Yx().unmounted?null:H.jsx(Xn.div,{...s,ref:r})});OQt.displayName="TooltipPositioner";var NQt=Vo("tooltip").parts("trigger","arrow","arrowTip","positioner","content"),ure=NQt.build(),RQt=n=>n.ids?.trigger??`tooltip:${n.id}:trigger`,AKn=n=>n.ids?.content??`tooltip:${n.id}:content`,IKn=n=>n.ids?.arrow??`tooltip:${n.id}:arrow`,PQt=n=>n.ids?.positioner??`tooltip:${n.id}:popper`,n$e=n=>n.getById(RQt(n)),e7t=n=>n.getById(PQt(n)),x4=iBn({id:null});function OKn(n,r){const{state:o,context:s,send:c,scope:d,prop:h,event:b}=n,y=h("id"),E=!!h("aria-label"),k=o.matches("open","closing"),x=RQt(d),O=AKn(d),P=h("disabled"),D=iYe({...h("positioning"),placement:s.get("currentPlacement")});return{open:k,setOpen($){o.matches("open","closing")!==$&&c({type:$?"open":"close"})},reposition($={}){c({type:"positioning.set",options:$})},getTriggerProps(){return r.button({...ure.trigger.attrs,id:x,dir:h("dir"),"data-expanded":nr(k),"data-state":k?"open":"closed","aria-describedby":k?O:void 0,onClick($){$.defaultPrevented||P||h("closeOnClick")&&c({type:"close",src:"trigger.click"})},onFocus($){queueMicrotask(()=>{$.defaultPrevented||P||b.src!=="trigger.pointerdown"&&qve()&&c({type:"open",src:"trigger.focus"})})},onBlur($){$.defaultPrevented||P||y===x4.get("id")&&c({type:"close",src:"trigger.blur"})},onPointerDown($){$.defaultPrevented||P||EW($)&&h("closeOnPointerDown")&&y===x4.get("id")&&c({type:"close",src:"trigger.pointerdown"})},onPointerMove($){$.defaultPrevented||P||$.pointerType!=="touch"&&c({type:"pointer.move"})},onPointerLeave(){P||c({type:"pointer.leave"})},onPointerCancel(){P||c({type:"pointer.leave"})}})},getArrowProps(){return r.element({id:IKn(d),...ure.arrow.attrs,dir:h("dir"),style:D.arrow})},getArrowTipProps(){return r.element({...ure.arrowTip.attrs,dir:h("dir"),style:D.arrowTip})},getPositionerProps(){return r.element({id:PQt(d),...ure.positioner.attrs,dir:h("dir"),style:D.floating})},getContentProps(){return r.element({...ure.content.attrs,dir:h("dir"),hidden:!k,"data-state":k?"open":"closed",role:E?void 0:"tooltip",id:E?void 0:O,"data-placement":s.get("currentPlacement"),onPointerEnter(){c({type:"content.pointer.move"})},onPointerLeave(){c({type:"content.pointer.leave"})},style:{pointerEvents:h("interactive")?"auto":"none"}})}}}var{and:NKn,not:t7t}=$I(),RKn={initialState:({prop:n})=>n("open")||n("defaultOpen")?"open":"closed",props({props:n}){const r=n.closeOnClick??!0,o=n.closeOnPointerDown??r;return{id:"x",openDelay:1e3,closeDelay:500,closeOnEscape:!0,interactive:!1,closeOnScroll:!0,disabled:!1,...n,closeOnPointerDown:o,closeOnClick:r,positioning:{placement:"bottom",...n.positioning}}},effects:["trackFocusVisible","trackStore"],context:({bindable:n})=>({currentPlacement:n(()=>({defaultValue:void 0})),hasPointerMoveOpened:n(()=>({defaultValue:!1}))}),watch({track:n,action:r,prop:o}){n([()=>o("disabled")],()=>{r(["closeIfDisabled"])}),n([()=>o("open")],()=>{r(["toggleVisibility"])})},states:{closed:{entry:["clearGlobalId"],on:{"controlled.open":{target:"open"},open:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}],"pointer.leave":{actions:["clearPointerMoveOpened"]},"pointer.move":[{guard:NKn("noVisibleTooltip",t7t("hasPointerMoveOpened")),target:"opening"},{guard:t7t("hasPointerMoveOpened"),target:"open",actions:["setPointerMoveOpened","invokeOnOpen"]}]}},opening:{effects:["trackScroll","trackPointerlockChange","waitForOpenDelay"],on:{"after.openDelay":[{guard:"isOpenControlled",actions:["setPointerMoveOpened","invokeOnOpen"]},{target:"open",actions:["setPointerMoveOpened","invokeOnOpen"]}],"controlled.open":{target:"open"},"controlled.close":{target:"closed"},open:[{guard:"isOpenControlled",actions:["invokeOnOpen"]},{target:"open",actions:["invokeOnOpen"]}],"pointer.leave":[{guard:"isOpenControlled",actions:["clearPointerMoveOpened","invokeOnClose","toggleVisibility"]},{target:"closed",actions:["clearPointerMoveOpened","invokeOnClose"]}],close:[{guard:"isOpenControlled",actions:["invokeOnClose","toggleVisibility"]},{target:"closed",actions:["invokeOnClose"]}]}},open:{effects:["trackEscapeKey","trackScroll","trackPointerlockChange","trackPositioning"],entry:["setGlobalId"],on:{"controlled.close":{target:"closed"},close:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],"pointer.leave":[{guard:"isVisible",target:"closing",actions:["clearPointerMoveOpened"]},{guard:"isOpenControlled",actions:["clearPointerMoveOpened","invokeOnClose"]},{target:"closed",actions:["clearPointerMoveOpened","invokeOnClose"]}],"content.pointer.leave":{guard:"isInteractive",target:"closing"},"positioning.set":{actions:["reposition"]}}},closing:{effects:["trackPositioning","waitForCloseDelay"],on:{"after.closeDelay":[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],"controlled.close":{target:"closed"},"controlled.open":{target:"open"},close:[{guard:"isOpenControlled",actions:["invokeOnClose"]},{target:"closed",actions:["invokeOnClose"]}],"pointer.move":[{guard:"isOpenControlled",actions:["setPointerMoveOpened","invokeOnOpen","toggleVisibility"]},{target:"open",actions:["setPointerMoveOpened","invokeOnOpen"]}],"content.pointer.move":{guard:"isInteractive",target:"open"},"positioning.set":{actions:["reposition"]}}}},implementations:{guards:{noVisibleTooltip:()=>x4.get("id")===null,isVisible:({prop:n})=>n("id")===x4.get("id"),isInteractive:({prop:n})=>!!n("interactive"),hasPointerMoveOpened:({context:n})=>n.get("hasPointerMoveOpened"),isOpenControlled:({prop:n})=>n("open")!==void 0},actions:{setGlobalId:({prop:n})=>{x4.set("id",n("id"))},clearGlobalId:({prop:n})=>{n("id")===x4.get("id")&&x4.set("id",null)},invokeOnOpen:({prop:n})=>{n("onOpenChange")?.({open:!0})},invokeOnClose:({prop:n})=>{n("onOpenChange")?.({open:!1})},closeIfDisabled:({prop:n,send:r})=>{n("disabled")&&r({type:"close",src:"disabled.change"})},reposition:({context:n,event:r,prop:o,scope:s})=>{if(r.type!=="positioning.set")return;const c=()=>e7t(s);return g2(n$e(s),c,{...o("positioning"),...r.options,defer:!0,listeners:!1,onComplete(d){n.set("currentPlacement",d.placement)}})},toggleVisibility:({prop:n,event:r,send:o})=>{queueMicrotask(()=>{o({type:n("open")?"controlled.open":"controlled.close",previousEvent:r})})},setPointerMoveOpened:({context:n})=>{n.set("hasPointerMoveOpened",!0)},clearPointerMoveOpened:({context:n})=>{n.set("hasPointerMoveOpened",!1)}},effects:{trackFocusVisible:({scope:n})=>sXt({root:n.getRootNode?.()}),trackPositioning:({context:n,prop:r,scope:o})=>{n.get("currentPlacement")||n.set("currentPlacement",r("positioning").placement);const s=()=>e7t(o);return g2(n$e(o),s,{...r("positioning"),defer:!0,onComplete(c){n.set("currentPlacement",c.placement)}})},trackPointerlockChange:({send:n,scope:r})=>{const o=r.getDoc();return Ul(o,"pointerlockchange",()=>n({type:"close",src:"pointerlock:change"}),!1)},trackScroll:({send:n,prop:r,scope:o})=>{if(!r("closeOnScroll"))return;const s=n$e(o);if(!s)return;const d=wYt(s).map(h=>Ul(h,"scroll",()=>{n({type:"close",src:"scroll"})},{passive:!0,capture:!0}));return()=>{d.forEach(h=>h?.())}},trackStore:({prop:n,send:r})=>{let o;return queueMicrotask(()=>{o=x4.subscribe(()=>{x4.get("id")!==n("id")&&r({type:"close",src:"id.change"})})}),()=>o?.()},trackEscapeKey:({send:n,prop:r})=>r("closeOnEscape")?Ul(document,"keydown",s=>{UKe(s)||s.key==="Escape"&&(s.stopPropagation(),n({type:"close",src:"keydown.escape"}))},!0):void 0,waitForOpenDelay:({send:n,prop:r})=>{const o=setTimeout(()=>{n({type:"after.openDelay"})},r("openDelay"));return()=>clearTimeout(o)},waitForCloseDelay:({send:n,prop:r})=>{const o=setTimeout(()=>{n({type:"after.closeDelay"})},r("closeDelay"));return()=>clearTimeout(o)}}}};Wo()(["aria-label","closeDelay","closeOnEscape","closeOnPointerDown","closeOnScroll","closeOnClick","dir","disabled","getRootNode","id","ids","interactive","onOpenChange","defaultOpen","open","openDelay","positioning"]);const PKn=n=>{const r=Q.useId(),{getRootNode:o}=qb(),{dir:s}=b2(),c={id:r,dir:s,getRootNode:o,...n},d=uv(RKn,c);return OKn(d,ry)},LKn=n=>{const[r,{children:o,...s}]=MM(n),c=PKn(s),d=J4(Fn({present:c.open},r));return H.jsx(xQt,{value:c,children:H.jsx(jM,{value:d,children:o})})},DKn=n=>{const[r,{value:o,children:s}]=MM(n),c=J4(Fn({present:o.open},r));return H.jsx(xQt,{value:o,children:H.jsx(jM,{value:c,children:s})})},LQt=Q.forwardRef((n,r)=>{const o=Ioe(),s=Fn(o.getTriggerProps(),n);return H.jsx(Xn.button,{...s,ref:r})});LQt.displayName="TooltipTrigger";const DQt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getBranchTriggerProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});DQt.displayName="TreeViewBranchTrigger";const MQt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getItemIndicatorProps(s),n);return H.jsx(Xn.div,{...c,ref:r})});MQt.displayName="TreeViewItemIndicator";const jQt=Q.forwardRef((n,r)=>{const o=lv(),s=Fn(o.getLabelProps(),n);return H.jsx(Xn.h3,{...s,ref:r})});jQt.displayName="TreeViewLabel";const MKn=n=>n.children(mJt()),$Qt=Q.forwardRef((n,r)=>{const o=lv(),s=Xx(),c=Fn(o.getNodeCheckboxProps(s),n);return H.jsx(Xn.span,{...c,ref:r})});$Qt.displayName="TreeViewNodeCheckbox";function vYe(n,r=[]){const o=Object.assign({},n);for(const s of r)s in o&&delete o[s];return o}const jKn=(n,r)=>{if(!n||typeof n!="string")return{invalid:!0,value:n};const[o,s]=n.split("/");if(!o||!s||o==="currentBg")return{invalid:!0,value:o};const c=r(`colors.${o}`),d=r.raw(`opacity.${s}`)?.value;if(!d&&isNaN(Number(s)))return{invalid:!0,value:o};const h=d?Number(d)*100+"%":`${s}%`,b=c??o;return{invalid:!1,color:b,value:`color-mix(in srgb, ${b} ${h}, transparent)`}},ad=n=>(r,o)=>{const s=o.utils.colorMix(r);if(s.invalid)return{[n]:r};const c="--mix-"+n;return{[c]:s.value,[n]:`var(${c}, ${s.color})`}};function rVe(n){if(n===null||typeof n!="object")return n;if(Array.isArray(n))return n.map(o=>rVe(o));const r=Object.create(Object.getPrototypeOf(n));for(const o of Object.keys(n))r[o]=rVe(n[o]);return r}function iVe(n,r){if(r==null)return n;for(const o of Object.keys(r))if(!(r[o]===void 0||o==="__proto__"))if(!Yw(n[o])&&Yw(r[o]))Object.assign(n,{[o]:r[o]});else if(n[o]&&Yw(r[o]))iVe(n[o],r[o]);else if(Array.isArray(r[o])&&Array.isArray(n[o])){let s=0;for(;s<r[o].length;s++)Yw(n[o][s])&&Yw(r[o][s])?iVe(n[o][s],r[o][s]):n[o][s]=r[o][s]}else Object.assign(n,{[o]:r[o]});return n}function SK(n,...r){for(const o of r)iVe(n,o);return n}const oVe=n=>n!=null;function LI(n,r,o={}){const{stop:s,getKey:c}=o;function d(h,b=[]){if(Yw(h)||Array.isArray(h)){const y={};for(const[E,k]of Object.entries(h)){const x=c?.(E,k)??E,O=[...b,x];if(s?.(h,O))return r(h,b);const P=d(k,O);oVe(P)&&(y[x]=P)}return y}return r(h,b)}return d(n)}function wYe(n,r){return Array.isArray(n)?n.map(o=>oVe(o)?r(o):o):Yw(n)?LI(n,o=>r(o)):oVe(n)?r(n):n}const K1e=["value","type","description"],$Kn=n=>n&&typeof n=="object"&&!Array.isArray(n),FQt=(...n)=>{const r=SK({},...n.map(rVe));return r.theme?.tokens&&LI(r.theme.tokens,o=>{const d=Object.keys(o).filter(b=>!K1e.includes(b)).length>0,h=K1e.some(b=>o[b]!=null);return d&&h&&(o.DEFAULT||(o.DEFAULT={}),K1e.forEach(b=>{var y;o[b]!=null&&((y=o.DEFAULT)[b]||(y[b]=o[b]),delete o[b])})),o},{stop(o){return $Kn(o)&&Object.keys(o).some(s=>K1e.includes(s)||s!==s.toLowerCase()&&s!==s.toUpperCase())}}),r},FKn=n=>n,Rh=n=>n,Bo=n=>n,BKn=n=>n,UKn=n=>n,_K=n=>n,zKn=n=>n,HKn=n=>n,GKn=n=>n;function BQt(){const n=r=>r;return new Proxy(n,{get(){return n}})}const Zp=BQt(),yYe=BQt(),EYe=n=>n,qKn=/[^a-zA-Z0-9_\u0081-\uffff-]/g;function VKn(n){return`${n}`.replace(qKn,r=>`\\${r}`)}const WKn=/[A-Z]/g;function KKn(n){return n.replace(WKn,r=>`-${r.toLowerCase()}`)}function UQt(n,r={}){const{fallback:o="",prefix:s=""}=r,c=KKn(["-",s,VKn(n)].filter(Boolean).join("-"));return{var:c,ref:`var(${c}${o?`, ${o}`:""})`}}const YKn=n=>/^var\(--.+\)$/.test(n),jm=(n,r)=>r!=null?`${n}(${r})`:r,oB=n=>{if(YKn(n)||n==null)return n;const r=typeof n=="string"&&!n.endsWith("deg");return typeof n=="number"||r?`${n}deg`:n},n7t=n=>({values:["outside","inside","mixed","none"],transform(r,{token:o}){const s=o("colors.colorPalette.focusRing");return{inside:{"--focus-ring-color":s,[n]:{outlineOffset:"0px",outlineWidth:"var(--focus-ring-width, 1px)",outlineColor:"var(--focus-ring-color)",outlineStyle:"var(--focus-ring-style, solid)",borderColor:"var(--focus-ring-color)"}},outside:{"--focus-ring-color":s,[n]:{outlineWidth:"var(--focus-ring-width, 2px)",outlineOffset:"var(--focus-ring-offset, 2px)",outlineStyle:"var(--focus-ring-style, solid)",outlineColor:"var(--focus-ring-color)"}},mixed:{"--focus-ring-color":s,[n]:{outlineWidth:"var(--focus-ring-width, 3px)",outlineStyle:"var(--focus-ring-style, solid)",outlineColor:"color-mix(in srgb, var(--focus-ring-color), transparent 60%)",borderColor:"var(--focus-ring-color)"}},none:{"--focus-ring-color":s,[n]:{outline:"none"}}}[r]??{}}}),XKn=ad("borderColor"),w4=n=>({transition:n,transitionTimingFunction:"cubic-bezier(0.4, 0, 0.2, 1)",transitionDuration:"150ms"}),JKn=FKn({hover:["@media (hover: hover)","&:is(:hover, [data-hover]):not(:disabled, [data-disabled])"],active:"&:is(:active, [data-active]):not(:disabled, [data-disabled], [data-state=open])",focus:"&:is(:focus, [data-focus])",focusWithin:"&:is(:focus-within, [data-focus-within])",focusVisible:"&:is(:focus-visible, [data-focus-visible])",disabled:"&:is(:disabled, [disabled], [data-disabled], [aria-disabled=true])",visited:"&:visited",target:"&:target",readOnly:"&:is([data-readonly], [aria-readonly=true], [readonly])",readWrite:"&:read-write",empty:"&:is(:empty, [data-empty])",checked:"&:is(:checked, [data-checked], [aria-checked=true], [data-state=checked])",enabled:"&:enabled",expanded:"&:is([aria-expanded=true], [data-expanded], [data-state=expanded])",highlighted:"&[data-highlighted]",complete:"&[data-complete]",incomplete:"&[data-incomplete]",dragging:"&[data-dragging]",before:"&::before",after:"&::after",firstLetter:"&::first-letter",firstLine:"&::first-line",marker:"&::marker",selection:"&::selection",file:"&::file-selector-button",backdrop:"&::backdrop",first:"&:first-of-type",last:"&:last-of-type",notFirst:"&:not(:first-of-type)",notLast:"&:not(:last-of-type)",only:"&:only-child",even:"&:nth-of-type(even)",odd:"&:nth-of-type(odd)",peerFocus:".peer:is(:focus, [data-focus]) ~ &",peerHover:".peer:is(:hover, [data-hover]):not(:disabled, [data-disabled]) ~ &",peerActive:".peer:is(:active, [data-active]):not(:disabled, [data-disabled]) ~ &",peerFocusWithin:".peer:focus-within ~ &",peerFocusVisible:".peer:is(:focus-visible, [data-focus-visible]) ~ &",peerDisabled:".peer:is(:disabled, [disabled], [data-disabled]) ~ &",peerChecked:".peer:is(:checked, [data-checked], [aria-checked=true], [data-state=checked]) ~ &",peerInvalid:".peer:is(:invalid, [data-invalid], [aria-invalid=true]) ~ &",peerExpanded:".peer:is([aria-expanded=true], [data-expanded], [data-state=expanded]) ~ &",peerPlaceholderShown:".peer:placeholder-shown ~ &",groupFocus:".group:is(:focus, [data-focus]) &",groupHover:".group:is(:hover, [data-hover]):not(:disabled, [data-disabled]) &",groupActive:".group:is(:active, [data-active]):not(:disabled, [data-disabled]) &",groupFocusWithin:".group:focus-within &",groupFocusVisible:".group:is(:focus-visible, [data-focus-visible]) &",groupDisabled:".group:is(:disabled, [disabled], [data-disabled]) &",groupChecked:".group:is(:checked, [data-checked], [aria-checked=true], [data-state=checked]) &",groupExpanded:".group:is([aria-expanded=true], [data-expanded], [data-state=expanded]) &",groupInvalid:".group:invalid &",indeterminate:"&:is(:indeterminate, [data-indeterminate], [aria-checked=mixed], [data-state=indeterminate])",required:"&:is([data-required], [aria-required=true])",valid:"&:is([data-valid], [data-state=valid])",invalid:"&:is([data-invalid], [aria-invalid=true], [data-state=invalid])",autofill:"&:autofill",inRange:"&:is(:in-range, [data-in-range])",outOfRange:"&:is(:out-of-range, [data-outside-range])",placeholder:"&::placeholder, &[data-placeholder]",placeholderShown:"&:is(:placeholder-shown, [data-placeholder-shown])",pressed:"&:is([aria-pressed=true], [data-pressed])",selected:"&:is([aria-selected=true], [data-selected])",grabbed:"&:is([aria-grabbed=true], [data-grabbed])",underValue:"&[data-state=under-value]",overValue:"&[data-state=over-value]",atValue:"&[data-state=at-value]",default:"&:default",optional:"&:optional",open:"&:is([open], [data-open], [data-state=open])",closed:"&:is([closed], [data-closed], [data-state=closed])",fullscreen:"&:is(:fullscreen, [data-fullscreen])",loading:"&:is([data-loading], [aria-busy=true])",hidden:"&:is([hidden], [data-hidden])",current:"&[data-current]",currentPage:"&[aria-current=page]",currentStep:"&[aria-current=step]",today:"&[data-today]",unavailable:"&[data-unavailable]",rangeStart:"&[data-range-start]",rangeEnd:"&[data-range-end]",now:"&[data-now]",topmost:"&[data-topmost]",motionReduce:"@media (prefers-reduced-motion: reduce)",motionSafe:"@media (prefers-reduced-motion: no-preference)",print:"@media print",landscape:"@media (orientation: landscape)",portrait:"@media (orientation: portrait)",dark:".dark &, .dark .chakra-theme:not(.light) &",light:":root &, .light &",osDark:"@media (prefers-color-scheme: dark)",osLight:"@media (prefers-color-scheme: light)",highContrast:"@media (forced-colors: active)",lessContrast:"@media (prefers-contrast: less)",moreContrast:"@media (prefers-contrast: more)",ltr:"[dir=ltr] &",rtl:"[dir=rtl] &",scrollbar:"&::-webkit-scrollbar",scrollbarThumb:"&::-webkit-scrollbar-thumb",scrollbarTrack:"&::-webkit-scrollbar-track",horizontal:"&[data-orientation=horizontal]",vertical:"&[data-orientation=vertical]",icon:"& :where(svg)",starting:"@starting-style"}),uW=UQt("bg-currentcolor"),r7t=n=>n===uW.ref||n==="currentBg",od=n=>({...n("colors"),currentBg:uW}),ZKn=EYe({conditions:JKn,utilities:{background:{values:od,shorthand:["bg"],transform(n,r){if(r7t(r.raw))return{background:uW.ref};const o=ad("background")(n,r);return{...o,[uW.var]:o?.background}}},backgroundColor:{values:od,shorthand:["bgColor"],transform(n,r){if(r7t(r.raw))return{backgroundColor:uW.ref};const o=ad("backgroundColor")(n,r);return{...o,[uW.var]:o?.backgroundColor}}},backgroundSize:{shorthand:["bgSize"]},backgroundPosition:{shorthand:["bgPos"]},backgroundRepeat:{shorthand:["bgRepeat"]},backgroundAttachment:{shorthand:["bgAttachment"]},backgroundClip:{shorthand:["bgClip"],values:["text"],transform(n){return n==="text"?{color:"transparent",backgroundClip:"text"}:{backgroundClip:n}}},backgroundGradient:{shorthand:["bgGradient"],values(n){return{...n("gradients"),"to-t":"linear-gradient(to top, var(--gradient))","to-tr":"linear-gradient(to top right, var(--gradient))","to-r":"linear-gradient(to right, var(--gradient))","to-br":"linear-gradient(to bottom right, var(--gradient))","to-b":"linear-gradient(to bottom, var(--gradient))","to-bl":"linear-gradient(to bottom left, var(--gradient))","to-l":"linear-gradient(to left, var(--gradient))","to-tl":"linear-gradient(to top left, var(--gradient))"}},transform(n){return{"--gradient-stops":"var(--gradient-from), var(--gradient-to)","--gradient":"var(--gradient-via-stops, var(--gradient-stops))",backgroundImage:n}}},gradientFrom:{values:od,transform:ad("--gradient-from")},gradientTo:{values:od,transform:ad("--gradient-to")},gradientVia:{values:od,transform(n,r){return{...ad("--gradient-via")(n,r),"--gradient-via-stops":"var(--gradient-from), var(--gradient-via), var(--gradient-to)"}}},backgroundImage:{values(n){return{...n("gradients"),...n("assets")}},shorthand:["bgImg","bgImage"]},border:{values:"borders"},borderTop:{values:"borders"},borderLeft:{values:"borders"},borderBlockStart:{values:"borders"},borderRight:{values:"borders"},borderBottom:{values:"borders"},borderBlockEnd:{values:"borders"},borderInlineStart:{values:"borders",shorthand:["borderStart"]},borderInlineEnd:{values:"borders",shorthand:["borderEnd"]},borderInline:{values:"borders",shorthand:["borderX"]},borderBlock:{values:"borders",shorthand:["borderY"]},borderColor:{values:od,transform:ad("borderColor")},borderTopColor:{values:od,transform:ad("borderTopColor")},borderBlockStartColor:{values:od,transform:ad("borderBlockStartColor")},borderBottomColor:{values:od,transform:ad("borderBottomColor")},borderBlockEndColor:{values:od,transform:ad("borderBlockEndColor")},borderLeftColor:{values:od,transform:ad("borderLeftColor")},borderInlineStartColor:{values:od,shorthand:["borderStartColor"],transform:ad("borderInlineStartColor")},borderRightColor:{values:od,transform:ad("borderRightColor")},borderInlineEndColor:{values:od,shorthand:["borderEndColor"],transform:ad("borderInlineEndColor")},borderStyle:{values:"borderStyles"},borderTopStyle:{values:"borderStyles"},borderBlockStartStyle:{values:"borderStyles"},borderBottomStyle:{values:"borderStyles"},borderBlockEndStyle:{values:"borderStyles"},borderInlineStartStyle:{values:"borderStyles",shorthand:["borderStartStyle"]},borderInlineEndStyle:{values:"borderStyles",shorthand:["borderEndStyle"]},borderLeftStyle:{values:"borderStyles"},borderRightStyle:{values:"borderStyles"},borderRadius:{values:"radii",shorthand:["rounded"]},borderTopLeftRadius:{values:"radii",shorthand:["roundedTopLeft"]},borderStartStartRadius:{values:"radii",shorthand:["roundedStartStart","borderTopStartRadius"]},borderEndStartRadius:{values:"radii",shorthand:["roundedEndStart","borderBottomStartRadius"]},borderTopRightRadius:{values:"radii",shorthand:["roundedTopRight"]},borderStartEndRadius:{values:"radii",shorthand:["roundedStartEnd","borderTopEndRadius"]},borderEndEndRadius:{values:"radii",shorthand:["roundedEndEnd","borderBottomEndRadius"]},borderBottomLeftRadius:{values:"radii",shorthand:["roundedBottomLeft"]},borderBottomRightRadius:{values:"radii",shorthand:["roundedBottomRight"]},borderInlineStartRadius:{values:"radii",property:"borderRadius",shorthand:["roundedStart","borderStartRadius"],transform:n=>({borderStartStartRadius:n,borderEndStartRadius:n})},borderInlineEndRadius:{values:"radii",property:"borderRadius",shorthand:["roundedEnd","borderEndRadius"],transform:n=>({borderStartEndRadius:n,borderEndEndRadius:n})},borderTopRadius:{values:"radii",property:"borderRadius",shorthand:["roundedTop"],transform:n=>({borderTopLeftRadius:n,borderTopRightRadius:n})},borderBottomRadius:{values:"radii",property:"borderRadius",shorthand:["roundedBottom"],transform:n=>({borderBottomLeftRadius:n,borderBottomRightRadius:n})},borderLeftRadius:{values:"radii",property:"borderRadius",shorthand:["roundedLeft"],transform:n=>({borderTopLeftRadius:n,borderBottomLeftRadius:n})},borderRightRadius:{values:"radii",property:"borderRadius",shorthand:["roundedRight"],transform:n=>({borderTopRightRadius:n,borderBottomRightRadius:n})},borderWidth:{values:"borderWidths"},borderBlockStartWidth:{values:"borderWidths"},borderTopWidth:{values:"borderWidths"},borderBottomWidth:{values:"borderWidths"},borderBlockEndWidth:{values:"borderWidths"},borderRightWidth:{values:"borderWidths"},borderInlineWidth:{values:"borderWidths",shorthand:["borderXWidth"]},borderInlineStartWidth:{values:"borderWidths",shorthand:["borderStartWidth"]},borderInlineEndWidth:{values:"borderWidths",shorthand:["borderEndWidth"]},borderLeftWidth:{values:"borderWidths"},borderBlockWidth:{values:"borderWidths",shorthand:["borderYWidth"]},color:{values:od,transform:ad("color")},fill:{values:od,transform:ad("fill")},stroke:{values:od,transform:ad("stroke")},accentColor:{values:od,transform:ad("accentColor")},divideX:{values:{type:"string"},transform(n){return{"& > :not(style, [hidden]) ~ :not(style, [hidden])":{borderInlineStartWidth:n,borderInlineEndWidth:"0px"}}}},divideY:{values:{type:"string"},transform(n){return{"& > :not(style, [hidden]) ~ :not(style, [hidden])":{borderTopWidth:n,borderBottomWidth:"0px"}}}},divideColor:{values:od,transform(n,r){return{"& > :not(style, [hidden]) ~ :not(style, [hidden])":XKn(n,r)}}},divideStyle:{property:"borderStyle",transform(n){return{"& > :not(style, [hidden]) ~ :not(style, [hidden])":{borderStyle:n}}}},boxShadow:{values:"shadows",shorthand:["shadow"]},boxShadowColor:{values:od,transform:ad("--shadow-color"),shorthand:["shadowColor"]},mixBlendMode:{shorthand:["blendMode"]},backgroundBlendMode:{shorthand:["bgBlendMode"]},opacity:{values:"opacity"},filter:{transform(n){return n!=="auto"?{filter:n}:{filter:"var(--blur) var(--brightness) var(--contrast) var(--grayscale) var(--hue-rotate) var(--invert) var(--saturate) var(--sepia) var(--drop-shadow)"}}},blur:{values:"blurs",transform:n=>({"--blur":jm("blur",n)})},brightness:{transform:n=>({"--brightness":jm("brightness",n)})},contrast:{transform:n=>({"--contrast":jm("contrast",n)})},grayscale:{transform:n=>({"--grayscale":jm("grayscale",n)})},hueRotate:{transform:n=>({"--hue-rotate":jm("hue-rotate",oB(n))})},invert:{transform:n=>({"--invert":jm("invert",n)})},saturate:{transform:n=>({"--saturate":jm("saturate",n)})},sepia:{transform:n=>({"--sepia":jm("sepia",n)})},dropShadow:{transform:n=>({"--drop-shadow":jm("drop-shadow",n)})},backdropFilter:{transform(n){return n!=="auto"?{backdropFilter:n}:{backdropFilter:"var(--backdrop-blur) var(--backdrop-brightness) var(--backdrop-contrast) var(--backdrop-grayscale) var(--backdrop-hue-rotate) var(--backdrop-invert) var(--backdrop-opacity) var(--backdrop-saturate) var(--backdrop-sepia)"}}},backdropBlur:{values:"blurs",transform:n=>({"--backdrop-blur":jm("blur",n)})},backdropBrightness:{transform:n=>({"--backdrop-brightness":jm("brightness",n)})},backdropContrast:{transform:n=>({"--backdrop-contrast":jm("contrast",n)})},backdropGrayscale:{transform:n=>({"--backdrop-grayscale":jm("grayscale",n)})},backdropHueRotate:{transform:n=>({"--backdrop-hue-rotate":jm("hue-rotate",oB(n))})},backdropInvert:{transform:n=>({"--backdrop-invert":jm("invert",n)})},backdropOpacity:{transform:n=>({"--backdrop-opacity":jm("opacity",n)})},backdropSaturate:{transform:n=>({"--backdrop-saturate":jm("saturate",n)})},backdropSepia:{transform:n=>({"--backdrop-sepia":jm("sepia",n)})},flexBasis:{values:"sizes"},gap:{values:"spacing"},rowGap:{values:"spacing",shorthand:["gapY"]},columnGap:{values:"spacing",shorthand:["gapX"]},flexDirection:{shorthand:["flexDir"]},gridGap:{values:"spacing"},gridColumnGap:{values:"spacing"},gridRowGap:{values:"spacing"},outlineColor:{values:od,transform:ad("outlineColor")},focusRing:n7t("&:is(:focus, [data-focus])"),focusVisibleRing:n7t("&:is(:focus-visible, [data-focus-visible])"),focusRingColor:{values:od,transform:ad("--focus-ring-color")},focusRingOffset:{values:"spacing",transform:n=>({"--focus-ring-offset":n})},focusRingWidth:{values:"borderWidths",property:"outlineWidth",transform:n=>({"--focus-ring-width":n})},focusRingStyle:{values:"borderStyles",property:"outlineStyle",transform:n=>({"--focus-ring-style":n})},aspectRatio:{values:"aspectRatios"},width:{values:"sizes",shorthand:["w"]},inlineSize:{values:"sizes"},height:{values:"sizes",shorthand:["h"]},blockSize:{values:"sizes"},boxSize:{values:"sizes",property:"width",transform:n=>({width:n,height:n})},minWidth:{values:"sizes",shorthand:["minW"]},minInlineSize:{values:"sizes"},minHeight:{values:"sizes",shorthand:["minH"]},minBlockSize:{values:"sizes"},maxWidth:{values:"sizes",shorthand:["maxW"]},maxInlineSize:{values:"sizes"},maxHeight:{values:"sizes",shorthand:["maxH"]},maxBlockSize:{values:"sizes"},hideFrom:{values:"breakpoints",transform:(n,{raw:r,token:o})=>({[o.raw(`breakpoints.${r}`)?`@breakpoint ${r}`:`@media screen and (min-width: ${n})`]:{display:"none"}})},hideBelow:{values:"breakpoints",transform(n,{raw:r,token:o}){return{[o.raw(`breakpoints.${r}`)?`@breakpoint ${r}Down`:`@media screen and (max-width: ${n})`]:{display:"none"}}}},overscrollBehavior:{shorthand:["overscroll"]},overscrollBehaviorX:{shorthand:["overscrollX"]},overscrollBehaviorY:{shorthand:["overscrollY"]},scrollbar:{values:["visible","hidden"],transform(n){switch(n){case"visible":return{msOverflowStyle:"auto",scrollbarWidth:"auto","&::-webkit-scrollbar":{display:"block"}};case"hidden":return{msOverflowStyle:"none",scrollbarWidth:"none","&::-webkit-scrollbar":{display:"none"}};default:return{}}}},scrollbarColor:{values:od,transform:ad("scrollbarColor")},scrollbarGutter:{values:"spacing"},scrollbarWidth:{values:"sizes"},scrollMargin:{values:"spacing"},scrollMarginTop:{values:"spacing"},scrollMarginBottom:{values:"spacing"},scrollMarginLeft:{values:"spacing"},scrollMarginRight:{values:"spacing"},scrollMarginX:{values:"spacing",transform:n=>({scrollMarginLeft:n,scrollMarginRight:n})},scrollMarginY:{values:"spacing",transform:n=>({scrollMarginTop:n,scrollMarginBottom:n})},scrollPadding:{values:"spacing"},scrollPaddingTop:{values:"spacing"},scrollPaddingBottom:{values:"spacing"},scrollPaddingLeft:{values:"spacing"},scrollPaddingRight:{values:"spacing"},scrollPaddingInline:{values:"spacing",shorthand:["scrollPaddingX"]},scrollPaddingBlock:{values:"spacing",shorthand:["scrollPaddingY"]},scrollSnapType:{values:{none:"none",x:"x var(--scroll-snap-strictness)",y:"y var(--scroll-snap-strictness)",both:"both var(--scroll-snap-strictness)"}},scrollSnapStrictness:{values:["mandatory","proximity"],transform:n=>({"--scroll-snap-strictness":n})},scrollSnapMargin:{values:"spacing"},scrollSnapMarginTop:{values:"spacing"},scrollSnapMarginBottom:{values:"spacing"},scrollSnapMarginLeft:{values:"spacing"},scrollSnapMarginRight:{values:"spacing"},listStylePosition:{shorthand:["listStylePos"]},listStyleImage:{values:"assets",shorthand:["listStyleImg"]},position:{shorthand:["pos"]},zIndex:{values:"zIndex"},inset:{values:"spacing"},insetInline:{values:"spacing",shorthand:["insetX"]},insetBlock:{values:"spacing",shorthand:["insetY"]},top:{values:"spacing"},insetBlockStart:{values:"spacing"},bottom:{values:"spacing"},insetBlockEnd:{values:"spacing"},left:{values:"spacing"},right:{values:"spacing"},insetInlineStart:{values:"spacing",shorthand:["insetStart"]},insetInlineEnd:{values:"spacing",shorthand:["insetEnd"]},ring:{transform(n){return{"--ring-offset-shadow":"var(--ring-inset) 0 0 0 var(--ring-offset-width) var(--ring-offset-color)","--ring-shadow":"var(--ring-inset) 0 0 0 calc(var(--ring-width) + var(--ring-offset-width)) var(--ring-color)","--ring-width":n,boxShadow:"var(--ring-offset-shadow), var(--ring-shadow), var(--shadow, 0 0 #0000)"}}},ringColor:{values:od,transform:ad("--ring-color")},ringOffset:{transform:n=>({"--ring-offset-width":n})},ringOffsetColor:{values:od,transform:ad("--ring-offset-color")},ringInset:{transform:n=>({"--ring-inset":n})},margin:{values:"spacing",shorthand:["m"]},marginTop:{values:"spacing",shorthand:["mt"]},marginBlockStart:{values:"spacing"},marginRight:{values:"spacing",shorthand:["mr"]},marginBottom:{values:"spacing",shorthand:["mb"]},marginBlockEnd:{values:"spacing"},marginLeft:{values:"spacing",shorthand:["ml"]},marginInlineStart:{values:"spacing",shorthand:["ms","marginStart"]},marginInlineEnd:{values:"spacing",shorthand:["me","marginEnd"]},marginInline:{values:"spacing",shorthand:["mx","marginX"]},marginBlock:{values:"spacing",shorthand:["my","marginY"]},padding:{values:"spacing",shorthand:["p"]},paddingTop:{values:"spacing",shorthand:["pt"]},paddingRight:{values:"spacing",shorthand:["pr"]},paddingBottom:{values:"spacing",shorthand:["pb"]},paddingBlockStart:{values:"spacing"},paddingBlockEnd:{values:"spacing"},paddingLeft:{values:"spacing",shorthand:["pl"]},paddingInlineStart:{values:"spacing",shorthand:["ps","paddingStart"]},paddingInlineEnd:{values:"spacing",shorthand:["pe","paddingEnd"]},paddingInline:{values:"spacing",shorthand:["px","paddingX"]},paddingBlock:{values:"spacing",shorthand:["py","paddingY"]},textDecoration:{shorthand:["textDecor"]},textDecorationColor:{values:od,transform:ad("textDecorationColor")},textShadow:{values:"shadows"},transform:{transform:n=>{let r=n;return n==="auto"&&(r="translateX(var(--translate-x, 0)) translateY(var(--translate-y, 0)) rotate(var(--rotate, 0)) scaleX(var(--scale-x, 1)) scaleY(var(--scale-y, 1)) skewX(var(--skew-x, 0)) skewY(var(--skew-y, 0))"),n==="auto-gpu"&&(r="translate3d(var(--translate-x, 0), var(--translate-y, 0), 0) rotate(var(--rotate, 0)) scaleX(var(--scale-x, 1)) scaleY(var(--scale-y, 1)) skewX(var(--skew-x, 0)) skewY(var(--skew-y, 0))"),{transform:r}}},skewX:{transform:n=>({"--skew-x":oB(n)})},skewY:{transform:n=>({"--skew-y":oB(n)})},scaleX:{transform:n=>({"--scale-x":n})},scaleY:{transform:n=>({"--scale-y":n})},scale:{transform(n){return n!=="auto"?{scale:n}:{scale:"var(--scale-x, 1) var(--scale-y, 1)"}}},spaceXReverse:{values:{type:"boolean"},transform(n){return{"& > :not(style, [hidden]) ~ :not(style, [hidden])":{"--space-x-reverse":n?"1":void 0}}}},spaceX:{property:"marginInlineStart",values:"spacing",transform:n=>({"& > :not(style, [hidden]) ~ :not(style, [hidden])":{"--space-x-reverse":"0",marginInlineStart:`calc(${n} * calc(1 - var(--space-x-reverse)))`,marginInlineEnd:`calc(${n} * var(--space-x-reverse))`}})},spaceYReverse:{values:{type:"boolean"},transform(n){return{"& > :not(style, [hidden]) ~ :not(style, [hidden])":{"--space-y-reverse":n?"1":void 0}}}},spaceY:{property:"marginTop",values:"spacing",transform:n=>({"& > :not(style, [hidden]) ~ :not(style, [hidden])":{"--space-y-reverse":"0",marginTop:`calc(${n} * calc(1 - var(--space-y-reverse)))`,marginBottom:`calc(${n} * var(--space-y-reverse))`}})},rotate:{transform(n){return n!=="auto"?{rotate:oB(n)}:{rotate:"var(--rotate-x, 0) var(--rotate-y, 0) var(--rotate-z, 0)"}}},rotateX:{transform(n){return{"--rotate-x":oB(n)}}},rotateY:{transform(n){return{"--rotate-y":oB(n)}}},translate:{transform(n){return n!=="auto"?{translate:n}:{translate:"var(--translate-x) var(--translate-y)"}}},translateX:{values:"spacing",transform:n=>({"--translate-x":n})},translateY:{values:"spacing",transform:n=>({"--translate-y":n})},transition:{values:["all","common","colors","opacity","position","backgrounds","size","shadow","transform"],transform(n){switch(n){case"all":return w4("all");case"position":return w4("left, right, top, bottom, inset-inline, inset-block");case"colors":return w4("color, background-color, border-color, text-decoration-color, fill, stroke");case"opacity":return w4("opacity");case"shadow":return w4("box-shadow");case"transform":return w4("transform");case"size":return w4("width, height");case"backgrounds":return w4("background, background-color, background-image, background-position");case"common":return w4("color, background-color, border-color, text-decoration-color, fill, stroke, opacity, box-shadow, transform, filter, backdrop-filter");default:return{transition:n}}}},transitionDuration:{values:"durations"},transitionProperty:{values:{common:"background-color, border-color, color, fill, stroke, opacity, box-shadow, translate, transform",colors:"background-color, border-color, color, fill, stroke",size:"width, height",position:"left, right, top, bottom, inset-inline, inset-block",background:"background, background-color, background-image, background-position"}},transitionTimingFunction:{values:"easings"},animation:{values:"animations"},animationDuration:{values:"durations"},animationDelay:{values:"durations"},animationTimingFunction:{values:"easings"},fontFamily:{values:"fonts"},fontSize:{values:"fontSizes"},fontWeight:{values:"fontWeights"},lineHeight:{values:"lineHeights"},letterSpacing:{values:"letterSpacings"},textIndent:{values:"spacing"},truncate:{values:{type:"boolean"},transform(n){return n===!0?{overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap"}:{}}},lineClamp:{transform(n){return n==="none"?{WebkitLineClamp:"unset"}:{overflow:"hidden",display:"-webkit-box",WebkitLineClamp:n,WebkitBoxOrient:"vertical",textWrap:"wrap"}}},borderSpacing:{values:n=>({...n("spacing"),auto:"var(--border-spacing-x, 0) var(--border-spacing-y, 0)"})},borderSpacingX:{values:"spacing",transform(n){return{"--border-spacing-x":n}}},borderSpacingY:{values:"spacing",transform(n){return{"--border-spacing-y":n}}},srOnly:{values:{type:"boolean"},transform(n){return QKn[n]||{}}},debug:{values:{type:"boolean"},transform(n){return n?{outline:"1px solid blue !important","& > *":{outline:"1px solid red !important"}}:{}}},caretColor:{values:od,transform:ad("caretColor")},cursor:{values:"cursor"}}}),QKn={true:{position:"absolute",width:"1px",height:"1px",padding:"0",margin:"-1px",overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",borderWidth:"0"},false:{position:"static",width:"auto",height:"auto",padding:"0",margin:"0",overflow:"visible",clip:"auto",whiteSpace:"normal"}};var eYn="",tYn=eYn.split(","),nYn="WebkitAppearance,WebkitBorderBefore,WebkitBorderBeforeColor,WebkitBorderBeforeStyle,WebkitBorderBeforeWidth,WebkitBoxReflect,WebkitLineClamp,WebkitMask,WebkitMaskAttachment,WebkitMaskClip,WebkitMaskComposite,WebkitMaskImage,WebkitMaskOrigin,WebkitMaskPosition,WebkitMaskPositionX,WebkitMaskPositionY,WebkitMaskRepeat,WebkitMaskRepeatX,WebkitMaskRepeatY,WebkitMaskSize,WebkitOverflowScrolling,WebkitTapHighlightColor,WebkitTextFillColor,WebkitTextStroke,WebkitTextStrokeColor,WebkitTextStrokeWidth,WebkitTouchCallout,WebkitUserModify,WebkitUserSelect,accentColor,alignContent,alignItems,alignSelf,alignTracks,all,anchorName,anchorScope,animation,animationComposition,animationDelay,animationDirection,animationDuration,animationFillMode,animationIterationCount,animationName,animationPlayState,animationRange,animationRangeEnd,animationRangeStart,animationTimeline,animationTimingFunction,appearance,aspectRatio,backdropFilter,backfaceVisibility,background,backgroundAttachment,backgroundBlendMode,backgroundClip,backgroundColor,backgroundImage,backgroundOrigin,backgroundPosition,backgroundPositionX,backgroundPositionY,backgroundRepeat,backgroundSize,blockSize,border,borderBlock,borderBlockColor,borderBlockEnd,borderBlockEndColor,borderBlockEndStyle,borderBlockEndWidth,borderBlockStart,borderBlockStartColor,borderBlockStartStyle,borderBlockStartWidth,borderBlockStyle,borderBlockWidth,borderBottom,borderBottomColor,borderBottomLeftRadius,borderBottomRightRadius,borderBottomStyle,borderBottomWidth,borderCollapse,borderColor,borderEndEndRadius,borderEndStartRadius,borderImage,borderImageOutset,borderImageRepeat,borderImageSlice,borderImageSource,borderImageWidth,borderInline,borderInlineColor,borderInlineEnd,borderInlineEndColor,borderInlineEndStyle,borderInlineEndWidth,borderInlineStart,borderInlineStartColor,borderInlineStartStyle,borderInlineStartWidth,borderInlineStyle,borderInlineWidth,borderLeft,borderLeftColor,borderLeftStyle,borderLeftWidth,borderRadius,borderRight,borderRightColor,borderRightStyle,borderRightWidth,borderSpacing,borderStartEndRadius,borderStartStartRadius,borderStyle,borderTop,borderTopColor,borderTopLeftRadius,borderTopRightRadius,borderTopStyle,borderTopWidth,borderWidth,bottom,boxAlign,boxDecorationBreak,boxDirection,boxFlex,boxFlexGroup,boxLines,boxOrdinalGroup,boxOrient,boxPack,boxShadow,boxSizing,breakAfter,breakBefore,breakInside,captionSide,caret,caretColor,caretShape,clear,clip,clipPath,clipRule,color,colorInterpolationFilters,colorScheme,columnCount,columnFill,columnGap,columnRule,columnRuleColor,columnRuleStyle,columnRuleWidth,columnSpan,columnWidth,columns,contain,containIntrinsicBlockSize,containIntrinsicHeight,containIntrinsicInlineSize,containIntrinsicSize,containIntrinsicWidth,container,containerName,containerType,content,contentVisibility,counterIncrement,counterReset,counterSet,cursor,cx,cy,d,direction,display,dominantBaseline,emptyCells,fieldSizing,fill,fillOpacity,fillRule,filter,flex,flexBasis,flexDirection,flexFlow,flexGrow,flexShrink,flexWrap,float,floodColor,floodOpacity,font,fontFamily,fontFeatureSettings,fontKerning,fontLanguageOverride,fontOpticalSizing,fontPalette,fontSize,fontSizeAdjust,fontSmooth,fontStretch,fontStyle,fontSynthesis,fontSynthesisPosition,fontSynthesisSmallCaps,fontSynthesisStyle,fontSynthesisWeight,fontVariant,fontVariantAlternates,fontVariantCaps,fontVariantEastAsian,fontVariantEmoji,fontVariantLigatures,fontVariantNumeric,fontVariantPosition,fontVariationSettings,fontWeight,forcedColorAdjust,gap,grid,gridArea,gridAutoColumns,gridAutoFlow,gridAutoRows,gridColumn,gridColumnEnd,gridColumnGap,gridColumnStart,gridGap,gridRow,gridRowEnd,gridRowGap,gridRowStart,gridTemplate,gridTemplateAreas,gridTemplateColumns,gridTemplateRows,hangingPunctuation,height,hyphenateCharacter,hyphenateLimitChars,hyphens,imageOrientation,imageRendering,imageResolution,imeMode,initialLetter,initialLetterAlign,inlineSize,inset,insetBlock,insetBlockEnd,insetBlockStart,insetInline,insetInlineEnd,insetInlineStart,interpolateSize,isolation,justifyContent,justifyItems,justifySelf,justifyTracks,left,letterSpacing,lightingColor,lineBreak,lineClamp,lineHeight,lineHeightStep,listStyle,listStyleImage,listStylePosition,listStyleType,margin,marginBlock,marginBlockEnd,marginBlockStart,marginBottom,marginInline,marginInlineEnd,marginInlineStart,marginLeft,marginRight,marginTop,marginTrim,marker,markerEnd,markerMid,markerStart,mask,maskBorder,maskBorderMode,maskBorderOutset,maskBorderRepeat,maskBorderSlice,maskBorderSource,maskBorderWidth,maskClip,maskComposite,maskImage,maskMode,maskOrigin,maskPosition,maskRepeat,maskSize,maskType,masonryAutoFlow,mathDepth,mathShift,mathStyle,maxBlockSize,maxHeight,maxInlineSize,maxLines,maxWidth,minBlockSize,minHeight,minInlineSize,minWidth,mixBlendMode,objectFit,objectPosition,offset,offsetAnchor,offsetDistance,offsetPath,offsetPosition,offsetRotate,opacity,order,orphans,outline,outlineColor,outlineOffset,outlineStyle,outlineWidth,overflow,overflowAnchor,overflowBlock,overflowClipBox,overflowClipMargin,overflowInline,overflowWrap,overflowX,overflowY,overlay,overscrollBehavior,overscrollBehaviorBlock,overscrollBehaviorInline,overscrollBehaviorX,overscrollBehaviorY,padding,paddingBlock,paddingBlockEnd,paddingBlockStart,paddingBottom,paddingInline,paddingInlineEnd,paddingInlineStart,paddingLeft,paddingRight,paddingTop,page,pageBreakAfter,pageBreakBefore,pageBreakInside,paintOrder,perspective,perspectiveOrigin,placeContent,placeItems,placeSelf,pointerEvents,position,positionAnchor,positionArea,positionTry,positionTryFallbacks,positionTryOrder,positionVisibility,printColorAdjust,quotes,r,resize,right,rotate,rowGap,rubyAlign,rubyMerge,rubyPosition,rx,ry,scale,scrollBehavior,scrollMargin,scrollMarginBlock,scrollMarginBlockEnd,scrollMarginBlockStart,scrollMarginBottom,scrollMarginInline,scrollMarginInlineEnd,scrollMarginInlineStart,scrollMarginLeft,scrollMarginRight,scrollMarginTop,scrollPadding,scrollPaddingBlock,scrollPaddingBlockEnd,scrollPaddingBlockStart,scrollPaddingBottom,scrollPaddingInline,scrollPaddingInlineEnd,scrollPaddingInlineStart,scrollPaddingLeft,scrollPaddingRight,scrollPaddingTop,scrollSnapAlign,scrollSnapCoordinate,scrollSnapDestination,scrollSnapPointsX,scrollSnapPointsY,scrollSnapStop,scrollSnapType,scrollSnapTypeX,scrollSnapTypeY,scrollTimeline,scrollTimelineAxis,scrollTimelineName,scrollbarColor,scrollbarGutter,scrollbarWidth,shapeImageThreshold,shapeMargin,shapeOutside,shapeRendering,stopColor,stopOpacity,stroke,strokeDasharray,strokeDashoffset,strokeLinecap,strokeLinejoin,strokeMiterlimit,strokeOpacity,strokeWidth,tabSize,tableLayout,textAlign,textAlignLast,textAnchor,textBox,textBoxEdge,textBoxTrim,textCombineUpright,textDecoration,textDecorationColor,textDecorationLine,textDecorationSkip,textDecorationSkipInk,textDecorationStyle,textDecorationThickness,textEmphasis,textEmphasisColor,textEmphasisPosition,textEmphasisStyle,textIndent,textJustify,textOrientation,textOverflow,textRendering,textShadow,textSizeAdjust,textSpacingTrim,textTransform,textUnderlineOffset,textUnderlinePosition,textWrap,textWrapMode,textWrapStyle,timelineScope,top,touchAction,transform,transformBox,transformOrigin,transformStyle,transition,transitionBehavior,transitionDelay,transitionDuration,transitionProperty,transitionTimingFunction,translate,unicodeBidi,userSelect,vectorEffect,verticalAlign,viewTimeline,viewTimelineAxis,viewTimelineInset,viewTimelineName,viewTransitionName,visibility,whiteSpace,whiteSpaceCollapse,widows,width,willChange,wordBreak,wordSpacing,wordWrap,writingMode,x,y,zIndex,zoom,alignmentBaseline,baselineShift,colorInterpolation,colorRendering,glyphOrientationVertical",rYn=nYn.split(",").concat(tYn),iYn=new Map(rYn.map(n=>[n,!0]));function oYn(n){const r=Object.create(null);return o=>(r[o]===void 0&&(r[o]=n(o)),r[o])}var aYn=/&|@/,sYn=oYn(n=>iYn.has(n)||n.startsWith("--")||aYn.test(n));function zQt(n,r){const o={};return LI(n,(s,c)=>{s&&(o[c.join(".")]=s.value)},{stop:r}),o}var r$e,i7t;function cYn(){if(i7t)return r$e;i7t=1,r$e=d,d.default=d,d.stable=E,d.stableStringify=E;var n="[...]",r="[Circular]",o=[],s=[];function c(){return{depthLimit:Number.MAX_SAFE_INTEGER,edgesLimit:Number.MAX_SAFE_INTEGER}}function d(O,P,D,$){typeof $>"u"&&($=c()),b(O,"",0,[],void 0,0,$);var U;try{s.length===0?U=JSON.stringify(O,P,D):U=JSON.stringify(O,x(P),D)}catch{return JSON.stringify("[unable to serialize, circular reference is too complex to analyze]")}finally{for(;o.length!==0;){var B=o.pop();B.length===4?Object.defineProperty(B[0],B[1],B[3]):B[0][B[1]]=B[2]}}return U}function h(O,P,D,$){var U=Object.getOwnPropertyDescriptor($,D);U.get!==void 0?U.configurable?(Object.defineProperty($,D,{value:O}),o.push([$,D,P,U])):s.push([P,D,O]):($[D]=O,o.push([$,D,P]))}function b(O,P,D,$,U,B,G){B+=1;var K;if(typeof O=="object"&&O!==null){for(K=0;K<$.length;K++)if($[K]===O){h(r,O,P,U);return}if(typeof G.depthLimit<"u"&&B>G.depthLimit){h(n,O,P,U);return}if(typeof G.edgesLimit<"u"&&D+1>G.edgesLimit){h(n,O,P,U);return}if($.push(O),Array.isArray(O))for(K=0;K<O.length;K++)b(O[K],K,K,$,O,B,G);else{var X=Object.keys(O);for(K=0;K<X.length;K++){var ae=X[K];b(O[ae],ae,K,$,O,B,G)}}$.pop()}}function y(O,P){return O<P?-1:O>P?1:0}function E(O,P,D,$){typeof $>"u"&&($=c());var U=k(O,"",0,[],void 0,0,$)||O,B;try{s.length===0?B=JSON.stringify(U,P,D):B=JSON.stringify(U,x(P),D)}catch{return JSON.stringify("[unable to serialize, circular reference is too complex to analyze]")}finally{for(;o.length!==0;){var G=o.pop();G.length===4?Object.defineProperty(G[0],G[1],G[3]):G[0][G[1]]=G[2]}}return B}function k(O,P,D,$,U,B,G){B+=1;var K;if(typeof O=="object"&&O!==null){for(K=0;K<$.length;K++)if($[K]===O){h(r,O,P,U);return}try{if(typeof O.toJSON=="function")return}catch{return}if(typeof G.depthLimit<"u"&&B>G.depthLimit){h(n,O,P,U);return}if(typeof G.edgesLimit<"u"&&D+1>G.edgesLimit){h(n,O,P,U);return}if($.push(O),Array.isArray(O))for(K=0;K<O.length;K++)k(O[K],K,K,$,O,B,G);else{var X={},ae=Object.keys(O).sort(y);for(K=0;K<ae.length;K++){var oe=ae[K];k(O[oe],oe,K,$,O,B,G),X[oe]=O[oe]}if(typeof U<"u")o.push([U,P,O]),U[P]=X;else return X}$.pop()}}function x(O){return O=typeof O<"u"?O:function(P,D){return D},function(P,D){if(s.length>0)for(var $=0;$<s.length;$++){var U=s[$];if(U[1]===P&&U[0]===D){D=U[2],s.splice($,1);break}}return O.call(this,P,D)}}return r$e}var uYn=cYn();const lYn=Y4(uYn),jx=n=>{const r=Object.create(null);function o(...s){const c=s.map(d=>lYn(d)).join("|");return r[c]===void 0&&(r[c]=n(...s)),r[c]}return o},HQt=16,Yve="px",SYe="em",aie="rem";function GQt(n=""){const r=new RegExp(String.raw`-?\d+(?:\.\d+|\d*)`),o=new RegExp(`${Yve}|${SYe}|${aie}`);return n.match(new RegExp(`${r.source}(${o.source})`))?.[1]}function qQt(n=""){if(typeof n=="number")return`${n}px`;const r=GQt(n);if(!r||r===Yve)return n;if(r===SYe||r===aie)return`${parseFloat(n)*HQt}${Yve}`}function VQt(n=""){const r=GQt(n);if(!r||r===aie)return n;if(r===SYe)return`${parseFloat(n)}${aie}`;if(r===Yve)return`${parseFloat(n)/HQt}${aie}`}const dYn=n=>n.charAt(0).toUpperCase()+n.slice(1);function fYn(n){const r=hYn(n),o=Object.fromEntries(r);function s(O){return o[O]}function c(O){return FV(s(O))}function d(){const O=Object.keys(o),P=pYn(O),D=O.flatMap($=>{const U=s($),B=[`${$}Down`,FV({max:sve(U.min)})],G=[$,FV({min:U.min})],K=[`${$}Only`,c($)];return[G,K,B]}).filter(([,$])=>$!=="").concat(P.map(([$,U])=>{const B=s($),G=s(U);return[`${$}To${dYn(U)}`,FV({min:B.min,max:sve(G.min)})]}));return Object.fromEntries(D)}function h(){const O=d();return Object.fromEntries(Object.entries(O))}const b=h(),y=O=>b[O];function E(){return["base",...Object.keys(o)]}function k(O){return FV({min:s(O).min})}function x(O){return FV({max:sve(s(O).min)})}return{values:Object.values(o),only:c,keys:E,conditions:b,getCondition:y,up:k,down:x}}function sve(n){const r=parseFloat(qQt(n)??"")-.04;return VQt(`${r}px`)}function hYn(n){return Object.entries(n).sort(([,o],[,s])=>parseInt(o,10)<parseInt(s,10)?-1:1).map(([o,s],c,d)=>{let h=null;return c<=d.length-1&&(h=d[c+1]?.[1]),h!=null&&(h=sve(h)),[o,{name:o,min:VQt(s),max:h}]})}function pYn(n){const r=[];return n.forEach((o,s)=>{let c=s;c++;let d=n[c];for(;d;)r.push([o,d]),c++,d=n[c]}),r}function FV({min:n,max:r}){return n==null&&r==null?"":["@media screen",n&&`(min-width: ${n})`,r&&`(max-width: ${r})`].filter(Boolean).join(" and ")}const gYn=(n,r)=>Object.fromEntries(Object.entries(n).map(([o,s])=>r(o,s))),mYn=n=>{const{breakpoints:r,conditions:o={}}=n,s=gYn(o,(k,x)=>[`_${k}`,x]),c=Object.assign({},s,r.conditions);function d(){return Object.keys(c)}function h(k){return d().includes(k)||/^@|&|&$/.test(k)||k.startsWith("_")}function b(k){return k.filter(x=>x!=="base").sort((x,O)=>{const P=h(x),D=h(O);return P&&!D?1:!P&&D?-1:0})}function y(k){return k.startsWith("@breakpoint")?r.getCondition(k.replace("@breakpoint ","")):k}function E(k){return Reflect.get(c,k)||k}return{keys:d,sort:b,has:h,resolve:E,breakpoints:r.keys(),expandAtRule:y}},WQt=n=>({minMax:new RegExp(`(!?\\(\\s*min(-device-)?-${n})(.|
)+\\(\\s*max(-device)?-${n}`,"i"),min:new RegExp(`\\(\\s*min(-device)?-${n}`,"i"),maxMin:new RegExp(`(!?\\(\\s*max(-device)?-${n})(.|
)+\\(\\s*min(-device)?-${n}`,"i"),max:new RegExp(`\\(\\s*max(-device)?-${n}`,"i")}),bYn=WQt("width"),vYn=WQt("height"),KQt=n=>({isMin:l7t(n.minMax,n.maxMin,n.min),isMax:l7t(n.maxMin,n.minMax,n.max)}),{isMin:aVe,isMax:o7t}=KQt(bYn),{isMin:sVe,isMax:a7t}=KQt(vYn),s7t=/print/i,c7t=/^print$/i,wYn=/(-?\d*\.?\d+)(ch|em|ex|px|rem)/,yYn=/(\d)/,Dre=Number.MAX_VALUE,EYn={ch:8.8984375,em:16,rem:16,ex:8.296875,px:1};function u7t(n){const r=wYn.exec(n)||(aVe(n)||sVe(n)?yYn.exec(n):null);if(!r)return Dre;if(r[0]==="0")return 0;const o=parseFloat(r[1]),s=r[2];return o*(EYn[s]||1)}function l7t(n,r,o){return s=>n.test(s)||!r.test(s)&&o.test(s)}function SYn(n,r){const o=s7t.test(n),s=c7t.test(n),c=s7t.test(r),d=c7t.test(r);return o&&c?!s&&d?1:s&&!d?-1:n.localeCompare(r):o?1:c?-1:null}const _Yn=jx((n,r)=>{const o=SYn(n,r);if(o!==null)return o;const s=aVe(n)||sVe(n),c=o7t(n)||a7t(n),d=aVe(r)||sVe(r),h=o7t(r)||a7t(r);if(s&&h)return-1;if(c&&d)return 1;const b=u7t(n),y=u7t(r);return b===Dre&&y===Dre?n.localeCompare(r):b===Dre?1:y===Dre?-1:b!==y?b>y?c?-1:1:c?1:-1:n.localeCompare(r)});function d7t(n){return n.sort(([r],[o])=>_Yn(r,o))}function YQt(n){const r=[],o=[],s={};for(const[h,b]of Object.entries(n))h.startsWith("@media")?r.push([h,b]):h.startsWith("@container")?o.push([h,b]):Yw(b)?s[h]=YQt(b):s[h]=b;const c=d7t(r),d=d7t(o);return{...s,...Object.fromEntries(c),...Object.fromEntries(d)}}const XQt=/\s*!(important)?/i,kYn=n=>B_(n)?XQt.test(n):!1,TYn=n=>B_(n)?n.replace(XQt,"").trim():n;function JQt(n){const{transform:r,conditions:o,normalize:s}=n,c=AYn(n);return jx(function(...h){const b=c(...h),y=s(b),E=Object.create(null);return LI(y,(k,x)=>{const O=kYn(k);if(k==null)return;const[P,...D]=o.sort(x).map(o.resolve);O&&(k=TYn(k));let $=r(P,k)??Object.create(null);$=LI($,U=>B_(U)&&O?`${U} !important`:U,{getKey:U=>o.expandAtRule(U)}),xYn(E,D.flat(),$)}),YQt(E)})}function xYn(n,r,o){let s=n;for(const c of r)c&&(s[c]||(s[c]=Object.create(null)),s=s[c]);SK(s,o)}function CYn(...n){return n.filter(r=>Yw(r)&&Object.keys(mK(r)).length>0)}function AYn(n){function r(o){const s=CYn(...o);return s.length===1?s:s.map(c=>n.normalize(c))}return jx(function(...s){return SK({},...r(s))})}const ZQt=n=>({base:{},variants:{},defaultVariants:{},compoundVariants:[],...n});function IYn(n){const{css:r,conditions:o,normalize:s,layers:c}=n;function d(b={}){const{base:y,variants:E,defaultVariants:k,compoundVariants:x}=ZQt(b),O=JQt({conditions:o,normalize:s,transform(G,K){return E[G]?.[K]}}),P=(G={})=>{const K=s({...k,...mK(G)});let X={...y};SK(X,O(K));const ae=h(x,K);return c.wrap("recipes",r(X,ae))},D=Object.keys(E),$=G=>{const K=vYe(G,["recipe"]),[X,ae]=yW(K,D);return D.includes("colorPalette")||(X.colorPalette=G.colorPalette||k.colorPalette),D.includes("orientation")&&(ae.orientation=G.orientation),[X,ae]},U=Object.fromEntries(Object.entries(E).map(([G,K])=>[G,Object.keys(K)]));return Object.assign(G=>r(P(G)),{className:b.className,__cva__:!0,variantMap:U,variantKeys:D,raw:P,config:b,splitVariantProps:$,merge(G){return d(OYn(n)(this,G))}})}function h(b,y){let E={};return b.forEach(k=>{Object.entries(k).every(([O,P])=>O==="css"?!0:(Array.isArray(P)?P:[P]).some($=>y[O]===$))&&(E=r(E,k.css))}),E}return d}function OYn(n){const{css:r}=n;return function(s,c){const d=ZQt(c.config),h=KKt(s.variantKeys,Object.keys(c.variants)),b=r(s.base,d.base),y=Object.fromEntries(h.map(O=>[O,r(s.config.variants[O],d.variants[O])])),E=SK(s.config.defaultVariants,d.defaultVariants),k=[...s.compoundVariants,...d.compoundVariants];return{className:Hb(s.className,c.className),base:b,variants:y,defaultVariants:E,compoundVariants:k}}}const NYn={reset:"reset",base:"base",tokens:"tokens",recipes:"recipes"},f7t={reset:0,base:1,tokens:2,recipes:3};function RYn(n){const r=n.layers??NYn,s=Object.values(r).sort((c,d)=>f7t[c]-f7t[d]);return{names:s,atRule:`@layer ${s.join(", ")};`,wrap(c,d){return n.disableLayers?d:{[`@layer ${r[c]}`]:d}}}}function PYn(n){const{utility:r,normalize:o}=n,{hasShorthand:s,resolveShorthand:c}=r;return function(d){return LI(d,o,{stop:h=>Array.isArray(h),getKey:s?c:void 0})}}function LYn(n){const{preflight:r}=n;if(!r)return{};const{scope:o="",level:s="parent"}=Yw(r)?r:{};let c="";o&&s==="parent"?c=`${o} `:o&&s==="element"&&(c=`&${o}`);const d={"*":{margin:"0px",padding:"0px",font:"inherit",wordWrap:"break-word",WebkitTapHighlightColor:"transparent"},"*, *::before, *::after, *::backdrop":{boxSizing:"border-box",borderWidth:"0px",borderStyle:"solid",borderColor:"var(--global-color-border, currentColor)"},hr:{height:"0px",color:"inherit",borderTopWidth:"1px"},body:{minHeight:"100dvh",position:"relative"},img:{borderStyle:"none"},"img, svg, video, canvas, audio, iframe, embed, object":{display:"block",verticalAlign:"middle"},iframe:{border:"none"},"img, video":{maxWidth:"100%",height:"auto"},"p, h1, h2, h3, h4, h5, h6":{overflowWrap:"break-word"},"ol, ul":{listStyle:"none"},"code, kbd, pre, samp":{fontSize:"1em"},"button, [type='button'], [type='reset'], [type='submit']":{WebkitAppearance:"button",backgroundColor:"transparent",backgroundImage:"none"},"button, input, optgroup, select, textarea":{color:"inherit"},"button, select":{textTransform:"none"},table:{textIndent:"0px",borderColor:"inherit",borderCollapse:"collapse"},"*::placeholder":{opacity:"unset",color:"#9ca3af",userSelect:"none"},textarea:{resize:"vertical"},summary:{display:"list-item"},small:{fontSize:"80%"},"sub, sup":{fontSize:"75%",lineHeight:0,position:"relative",verticalAlign:"baseline"},sub:{bottom:"-0.25em"},sup:{top:"-0.5em"},dialog:{padding:"0px"},a:{color:"inherit",textDecoration:"inherit"},"abbr:where([title])":{textDecoration:"underline dotted"},"b, strong":{fontWeight:"bolder"},"code, kbd, samp, pre":{fontSize:"1em","--font-mono-fallback":"ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, 'Liberation Mono', 'Courier New'",fontFamily:"var(--global-font-mono, var(--font-mono-fallback))"},'input[type="text"], input[type="email"], input[type="search"], input[type="password"]':{WebkitAppearance:"none",MozAppearance:"none"},"input[type='search']":{WebkitAppearance:"textfield",outlineOffset:"-2px"},"::-webkit-search-decoration, ::-webkit-search-cancel-button":{WebkitAppearance:"none"},"::-webkit-file-upload-button":{WebkitAppearance:"button",font:"inherit"},'input[type="number"]::-webkit-inner-spin-button, input[type="number"]::-webkit-outer-spin-button':{height:"auto"},"input[type='number']":{MozAppearance:"textfield"},":-moz-ui-invalid":{boxShadow:"none"},":-moz-focusring":{outline:"auto"},"[hidden]:where(:not([hidden='until-found']))":{display:"none !important"}},h={[o||"html"]:{lineHeight:1.5,"--font-fallback":"ui-sans-serif, system-ui, -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, 'Helvetica Neue', Arial, 'Noto Sans', sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji', 'Segoe UI Symbol', 'Noto Color Emoji'",WebkitTextSizeAdjust:"100%",WebkitFontSmoothing:"antialiased",MozOsxFontSmoothing:"grayscale",textRendering:"optimizeLegibility",touchAction:"manipulation",MozTabSize:"4",tabSize:"4",fontFamily:"var(--global-font-body, var(--font-fallback))"}};if(s==="element"){const b=Object.entries(d).reduce((y,[E,k])=>(y[E]={[c]:k},y),{});Object.assign(h,b)}else c?h[c]=d:Object.assign(h,d);return h}function DYn(n){const{conditions:r,isValidProperty:o}=n;return function(c){return LI(c,d=>d,{getKey:(d,h)=>Yw(h)&&!r.has(d)&&!o(d)?MYn(d).map(b=>"&"+b).join(", "):d})}}function MYn(n){const r=[];let o=0,s="",c=!1;for(let d=0;d<n.length;d++){const h=n[d];if(h==="\\"&&!c){c=!0,s+=h;continue}if(c){c=!1,s+=h;continue}h==="("?o++:h===")"&&o--,h===","&&o===0?(r.push(s.trim()),s=""):s+=h}return s&&r.push(s.trim()),r}const jYn=(n={})=>{const r=c=>({base:n.base?.[c]??{},variants:{},defaultVariants:n.defaultVariants??{},compoundVariants:n.compoundVariants?$Yn(n.compoundVariants,c):[]}),s=(n.slots??[]).map(c=>[c,r(c)]);for(const[c,d]of Object.entries(n.variants??{}))for(const[h,b]of Object.entries(d))s.forEach(([y,E])=>{var k;(k=E.variants)[c]??(k[c]={}),E.variants[c][h]=b[y]??{}});return Object.fromEntries(s)},$Yn=(n,r)=>n.filter(o=>o.css[r]).map(o=>({...o,css:o.css[r]}));function FYn(n){const{cva:r}=n;return function(s={}){const c=Object.entries(jYn(s)).map(([x,O])=>[x,r(O)]);function d(x){const O=c.map(([P,D])=>[P,D(x)]);return Object.fromEntries(O)}const h=s.variants??{},b=Object.keys(h);function y(x){const O=vYe(x,["recipe"]),[P,D]=yW(O,b);return b.includes("colorPalette")||(P.colorPalette=x.colorPalette||s.defaultVariants?.colorPalette),b.includes("orientation")&&(D.orientation=x.orientation),[P,D]}const E=Object.fromEntries(Object.entries(h).map(([x,O])=>[x,Object.keys(O)]));let k={};return s.className&&(k=Object.fromEntries(s.slots.map(x=>[x,`${s.className}__${x}`]))),Object.assign(d,{variantMap:E,variantKeys:b,splitVariantProps:y,classNameMap:k})}}const BYn=()=>n=>Array.from(new Set(n)),UYn=/([\0-\x1f\x7f]|^-?\d)|^-$|^-|[^\x80-\uFFFF\w-]/g,zYn=function(n,r){return r?n==="\0"?"�":n==="-"&&n.length===1?"\\-":n.slice(0,-1)+"\\"+n.charCodeAt(n.length-1).toString(16):"\\"+n},QQt=n=>(n+"").replace(UYn,zYn),een=(n,r)=>{let o="",s=0,c="char",d="",h="";const b=[];for(;s<n.length;){const y=n[s];if(y==="{"){const k=n.indexOf("}",s);if(k===-1)break;const x=n.slice(s+1,k),O=r(x);o+=O??x,s=k+1;continue}if(c==="token"&&y===","){n[s]===""&&s++,c="fallback",b.push(c);const k=r(d);k?.endsWith(")")&&(o+=k.slice(0,-1)),d="",h="";continue}if(c==="fallback"&&h+y===", var("){const O=HYn(n.slice(s+1))+s+1,P=n.slice(s+1,O);if(O===-1)break;o+=", var("+P+")",s=O+1,c=b.pop()??c,h="";continue}if(c==="token"||c==="fallback"){if(s++,y===")"){c=b.pop()??c??"char",h+=y;const k=d&&(r(d)??QQt(d));if(h){if(h=h.slice(1).trim(),!h.startsWith("token(")&&h.endsWith(")")&&(h=h.slice(0,-1)),h.includes("token(")){const O=een(h,r);O&&(h=O.slice(0,-1))}else if(h){const O=r(h);O&&(h=O)}}const x=o.at(-1);h?x?.trim()?o+=k.slice(0,-1)+(", "+h+")"):o+=h:o+=k||")",d="",h="",c="char";continue}c==="token"&&(d+=y),c==="fallback"&&(h+=y);continue}const E=n.indexOf("token(",s);if(E!==-1){const k=E+6;o+=n.slice(s,E),s=k,c="token",b.push(c);continue}o+=y,s++}return o},HYn=n=>{let r=0;const o=["("];for(;r<n.length;){const s=n[r];if(s==="(")o.push(s);else if(s===")"&&(o.pop(),o.length===0))return r;r++}return r};function ten(n){const r={};return n.forEach((o,s)=>{o instanceof Map?r[s]=Object.fromEntries(o):r[s]=o}),r}const nen=/({([^}]*)})/g,GYn=/[{}]/g,qYn=/\w+\.\w+/,ren=n=>{if(!B_(n))return[];const r=n.match(nen);return r?r.map(o=>o.replace(GYn,"")).map(o=>o.trim()):[]},VYn=n=>nen.test(n);function ien(n){if(!n.extensions?.references)return n.extensions?.cssVar?.ref??n.value;const r=n.extensions.references??{};return n.value=Object.keys(r).reduce((o,s)=>{const c=r[s];if(c.extensions.conditions)return o;const d=ien(c);return o.replace(`{${s}}`,d)},n.value),delete n.extensions.references,n.value}function oen(n){return Yw(n)&&n.reference?n.reference:String(n)}const xye=(n,...r)=>r.map(oen).join(` ${n} `).replace(/calc/g,""),h7t=(...n)=>`calc(${xye("+",...n)})`,p7t=(...n)=>`calc(${xye("-",...n)})`,cVe=(...n)=>`calc(${xye("*",...n)})`,g7t=(...n)=>`calc(${xye("/",...n)})`,m7t=n=>{const r=oen(n);return r!=null&&!Number.isNaN(parseFloat(r))?String(r).startsWith("-")?String(r).slice(1):`-${r}`:cVe(r,-1)},oW=Object.assign(n=>({add:(...r)=>oW(h7t(n,...r)),subtract:(...r)=>oW(p7t(n,...r)),multiply:(...r)=>oW(cVe(n,...r)),divide:(...r)=>oW(g7t(n,...r)),negate:()=>oW(m7t(n)),toString:()=>n.toString()}),{add:h7t,subtract:p7t,multiply:cVe,divide:g7t,negate:m7t}),WYn={enforce:"pre",transform(n){const{prefix:r,allTokens:o,formatCssVar:s,formatTokenName:c,registerToken:d}=n;o.filter(({extensions:b})=>b.category==="spacing").forEach(b=>{const y=b.path.slice(),E=s(y,r);if(B_(b.value)&&b.value==="0rem")return;const k=structuredClone(b);Object.assign(k.extensions,{negative:!0,prop:`-${b.extensions.prop}`,originalPath:y}),k.value=oW.negate(E.ref);const x=k.path[k.path.length-1];x!=null&&(k.path[k.path.length-1]=`-${x}`),k.path&&(k.name=c(k.path)),d(k)})}},KYn=new Set(["spacing","sizes","borderWidths","fontSizes","radii"]),YYn={enforce:"post",transform(n){n.allTokens.filter(o=>KYn.has(o.extensions.category)&&!o.extensions.negative).forEach(o=>{Object.assign(o.extensions,{pixelValue:qQt(o.value)})})}},XYn={enforce:"post",transform(n){const{allTokens:r,registerToken:o,formatTokenName:s}=n,c=r.filter(({extensions:b})=>b.category==="colors"),d=new Map,h=new Map;c.forEach(b=>{const{colorPalette:y}=b.extensions;y&&(y.keys.forEach(E=>{d.set(s(E),E)}),y.roots.forEach(E=>{const k=s(E),x=h.get(k)||[];if(x.push(b),h.set(k,x),b.extensions.default&&E.length===1){const O=y.keys[0]?.filter(Boolean);if(!O.length)return;const P=E.concat(O);d.set(s(P),[])}}))}),d.forEach(b=>{const y=["colors","colorPalette",...b].filter(Boolean),E=s(y),k=s(y.slice(1));o({name:E,value:E,originalValue:E,path:y,extensions:{condition:"base",originalPath:y,category:"colors",prop:k,virtual:!0}},"pre")})}},JYn={enforce:"post",transform(n){n.allTokens=n.allTokens.filter(r=>r.value!=="")}},ZYn=[WYn,XYn,YYn,JYn],QYn={type:"extensions",enforce:"pre",name:"tokens/css-var",transform(n,r){const{prefix:o,formatCssVar:s}=r,{negative:c,originalPath:d}=n.extensions,h=c?d:n.path;return{cssVar:s(h.filter(Boolean),o)}}},eXn={enforce:"post",type:"value",name:"tokens/conditionals",transform(n,r){const{prefix:o,formatCssVar:s}=r,c=ren(n.value);return c.length&&c.forEach(d=>{const h=s(d.split("."),o);n.value=n.value.replace(`{${h.ref}}`,h)}),n.value}},tXn={type:"extensions",enforce:"pre",name:"tokens/colors/colorPalette",match(n){return n.extensions.category==="colors"&&!n.extensions.virtual},transform(n,r){let o=n.path.slice();if(o.pop(),o.shift(),o.length===0){const b=[...n.path];b.shift(),o=b}if(o.length===0)return{};const s=o.reduce((b,y,E,k)=>{const x=k.slice(0,E+1);return b.push(x),b},[]),c=o[0],d=r.formatTokenName(o),h=n.path.slice(n.path.indexOf(c)+1).reduce((b,y,E,k)=>(b.push(k.slice(E)),b),[]);return h.length===0&&h.push([""]),{colorPalette:{value:d,roots:s,keys:h}}}},nXn=[QYn,eXn,tXn],b7t=n=>Yw(n)&&Object.prototype.hasOwnProperty.call(n,"value");function rXn(n){return n?{breakpoints:wYe(n,r=>({value:r})),sizes:Object.fromEntries(Object.entries(n).map(([r,o])=>[`breakpoint-${r}`,{value:o}]))}:{breakpoints:{},sizes:{}}}function iXn(n){const{prefix:r="",tokens:o={},semanticTokens:s={},breakpoints:c={}}=n,d=Nt=>Nt.join("."),h=(Nt,on)=>UQt(Nt.join("-"),{prefix:on}),b=[],y=new Map,E=new Map,k=new Map,x=new Map,O=new Map,P=new Map,D=new Map,$=new Map,U=[];function B(Nt,on){b.push(Nt),y.set(Nt.name,Nt),on&&$.forEach(br=>{br.enforce===on&&de(br,Nt)})}const G=rXn(c),K=mK({...o,breakpoints:G.breakpoints,sizes:{...o.sizes,...G.sizes}});function X(){LI(K,(Nt,on)=>{const br=on.includes("DEFAULT");on=v7t(on);const wi=on[0],no=d(on),Ri=B_(Nt)?{value:Nt}:Nt,ro={value:Ri.value,originalValue:Ri.value,name:no,path:on,extensions:{condition:"base",originalPath:on,category:wi,prop:d(on.slice(1))}};br&&(ro.extensions.default=!0),B(ro)},{stop:b7t}),LI(s,(Nt,on)=>{const br=on.includes("DEFAULT");on=aen(v7t(on));const wi=on[0],no=d(on),Ri=B_(Nt.value)?{value:{base:Nt.value}}:Nt,ro={value:Ri.value.base||"",originalValue:Ri.value.base||"",name:no,path:on,extensions:{originalPath:on,category:wi,conditions:Ri.value,condition:"base",prop:d(on.slice(1))}};br&&(ro.extensions.default=!0),B(ro)},{stop:b7t})}function ae(Nt){return y.get(Nt)}function oe(Nt){const{condition:on}=Nt.extensions;on&&(E.has(on)||E.set(on,new Set),E.get(on).add(Nt))}function he(Nt){const{category:on,prop:br}=Nt.extensions;on&&(D.has(on)||D.set(on,new Map),D.get(on).set(br,Nt))}function ye(Nt){const{condition:on,negative:br,virtual:wi,cssVar:no}=Nt.extensions;br||wi||!on||!no||(k.has(on)||k.set(on,new Map),k.get(on).set(no.var,Nt.value))}function we(Nt){const{category:on,prop:br,cssVar:wi,negative:no}=Nt.extensions;if(!on)return;P.has(on)||P.set(on,new Map);const Ri=no?Nt.extensions.conditions?Nt.originalValue:Nt.value:wi.ref;P.get(on).set(br,Ri),O.set([on,br].join("."),Ri)}function ge(Nt){const{colorPalette:on,virtual:br,default:wi}=Nt.extensions;!on||br||on.roots.forEach(no=>{const Ri=d(no);x.has(Ri)||x.set(Ri,new Map);const ro=aXn([...Nt.path],[...no]),Ko=d(ro),Ca=ae(Ko);if(!Ca||!Ca.extensions.cssVar)return;const{var:qi}=Ca.extensions.cssVar;if(x.get(Ri).set(qi,Nt.extensions.cssVar.ref),wi&&no.length===1){const yi=d(["colors","colorPalette"]),lo=ae(yi);if(!lo)return;const Eo=d(Nt.path),Fc=ae(Eo);if(!Fc)return;const Yo=on.keys[0]?.filter(Boolean);if(!Yo.length)return;const zl=d(no.concat(Yo));x.has(zl)||x.set(zl,new Map),x.get(zl).set(lo.extensions.cssVar.var,Fc.extensions.cssVar.ref)}})}let Te={};function xe(){b.forEach(Nt=>{oe(Nt),he(Nt),ye(Nt),we(Nt),ge(Nt)}),Te=ten(P)}const Ue=(Nt,on)=>{if(!Nt||typeof Nt!="string")return{invalid:!0,value:Nt};const[br,wi]=Nt.split("/");if(!br||!wi)return{invalid:!0,value:br};const no=on(br),Ri=ae(`opacity.${wi}`)?.value;if(!Ri&&isNaN(Number(wi)))return{invalid:!0,value:br};const ro=Ri?Number(Ri)*100+"%":`${wi}%`,Ko=no??br;return{invalid:!1,color:Ko,value:`color-mix(in srgb, ${Ko} ${ro}, transparent)`}},Z=jx((Nt,on)=>O.get(Nt)??on),Re=jx(Nt=>Te[Nt]||null),_e=jx(Nt=>een(Nt,on=>{if(!on)return;if(on.includes("/")){const wi=Ue(on,no=>Z(no));if(wi.invalid)throw new Error("Invalid color mix at "+on+": "+wi.value);return wi.value}const br=Z(on);return br||(qYn.test(on)?QQt(on):on)})),qe={prefix:r,allTokens:b,tokenMap:y,registerToken:B,getByName:ae,formatTokenName:d,formatCssVar:h,flatMap:O,cssVarMap:k,categoryMap:D,colorPaletteMap:x,getVar:Z,getCategoryValues:Re,expandReferenceInValue:_e};function it(...Nt){Nt.forEach(on=>{$.set(on.name,on)})}function Xe(...Nt){U.push(...Nt)}function de(Nt,on){if(on.extensions.references||MKe(Nt.match)&&!Nt.match(on))return;const wi=(no=>Nt.transform(no,qe))(on);switch(!0){case Nt.type==="extensions":Object.assign(on.extensions,wi);break;case Nt.type==="value":on.value=wi;break;default:on[Nt.type]=wi;break}}function We(Nt){U.forEach(on=>{on.enforce===Nt&&on.transform(qe)})}function ht(Nt){$.forEach(on=>{on.enforce===Nt&&b.forEach(br=>{de(on,br)})})}function Ie(){b.forEach(Nt=>{const on=oXn(Nt);!on||on.length===0||on.forEach(br=>{B(br)})})}function $t(Nt){return ren(Nt).map(br=>ae(br)).filter(Boolean)}function An(){b.forEach(Nt=>{if(!VYn(Nt.value))return;const on=$t(Nt.value);Nt.extensions.references=on.reduce((br,wi)=>(br[wi.name]=wi,br),{})})}function Yt(){b.forEach(Nt=>{ien(Nt)})}function Or(){We("pre"),ht("pre"),Ie(),An(),Yt(),We("post"),ht("post"),xe()}return X(),it(...nXn),Xe(...ZYn),Or(),qe}function v7t(n){return n[0]==="DEFAULT"?n:n.filter(r=>r!=="DEFAULT")}function aen(n){return n.filter(r=>r!=="base")}function oXn(n){if(!n.extensions.conditions)return;const{conditions:r}=n.extensions,o=[];return LI(r,(s,c)=>{const d=aen(c);if(!d.length)return;const h=structuredClone(n);h.value=s,h.extensions.condition=d.join(":"),o.push(h)}),o}function aXn(n,r){const o=n.findIndex((s,c)=>r.every((d,h)=>n[c+h]===d));return o===-1||(n.splice(o,r.length),n.splice(o,0,"colorPalette")),n}BYn()(["aspectRatios","zIndex","opacity","colors","fonts","fontSizes","fontWeights","lineHeights","letterSpacings","sizes","shadows","spacing","radii","cursor","borders","borderWidths","borderStyles","durations","easings","animations","blurs","gradients","breakpoints","assets"]);function sXn(n){return Object.fromEntries(Object.entries(n).map(([r,o])=>[r,o]))}function cXn(n){const r=sXn(n.config),o=n.tokens,s=new Map,c=new Map;function d(ye,we){r[ye]=we,h(ye,we)}const h=(ye,we)=>{const ge=$(we);ge&&(c.set(ye,ge),x(ye,we))},b=()=>{for(const[ye,we]of Object.entries(r))we&&h(ye,we)},y=()=>{for(const[ye,we]of Object.entries(r)){const{shorthand:ge}=we??{};if(!ge)continue;(Array.isArray(ge)?ge:[ge]).forEach(xe=>s.set(xe,ye))}},E=()=>{const ye=ten(o.colorPaletteMap);d("colorPalette",{values:Object.keys(ye),transform:jx(we=>ye[we])})},k=new Map,x=(ye,we)=>{if(!we)return;const ge=$(we,xe=>`type:Tokens["${xe}"]`);if(typeof ge=="object"&&ge.type){k.set(ye,new Set([`type:${ge.type}`]));return}if(ge){const xe=new Set(Object.keys(ge));k.set(ye,xe)}const Te=k.get(ye)??new Set;we.property&&k.set(ye,Te.add(`CssProperties["${we.property}"]`))},O=()=>{for(const[ye,we]of Object.entries(r))we&&x(ye,we)},P=(ye,we)=>{const ge=k.get(ye)??new Set;k.set(ye,new Set([...ge,...we]))},D=()=>{const ye=new Map;for(const[we,ge]of k.entries()){if(ge.size===0){ye.set(we,["string"]);continue}const Te=Array.from(ge).map(xe=>xe.startsWith("CssProperties")?xe:xe.startsWith("type:")?xe.replace("type:",""):JSON.stringify(xe));ye.set(we,Te)}return ye},$=(ye,we)=>{const{values:ge}=ye,Te=xe=>{const Ue=we?.(xe);return Ue?{[Ue]:Ue}:void 0};return B_(ge)?Te?.(ge)??o.getCategoryValues(ge)??{}:Array.isArray(ge)?ge.reduce((xe,Ue)=>(xe[Ue]=Ue,xe),{}):MKe(ge)?ge(we?Te:o.getCategoryValues):ge},U=jx((ye,we)=>({[ye]:ye.startsWith("--")?o.getVar(we,we):we})),B=Object.assign(o.getVar,{raw:ye=>o.getByName(ye)}),G=jx((ye,we)=>{const ge=ae(ye);B_(we)&&!we.includes("_EMO_")&&(we=o.expandReferenceInValue(we));const Te=r[ge];if(!Te)return U(ge,we);const xe=c.get(ge)?.[we];if(!Te.transform)return U(ye,xe??we);const Ue=Z=>jKn(Z,B);return Te.transform(xe??we,{raw:we,token:B,utils:{colorMix:Ue}})});function K(){y(),E(),b(),O()}K();const X=s.size>0,ae=jx(ye=>s.get(ye)??ye);return{keys:()=>[...Array.from(s.keys()),...Object.keys(r)],hasShorthand:X,transform:G,shorthands:s,resolveShorthand:ae,register:d,getTypes:D,addPropertyType:P}}const Uw={};function sen(...n){const r=FQt(...n),{theme:o={},utilities:s={},globalCss:c={},cssVarsRoot:d=":where(:root, :host)",cssVarsPrefix:h="chakra",preflight:b}=r,y=RYn(r),E=iXn({breakpoints:o.breakpoints,tokens:o.tokens,semanticTokens:o.semanticTokens,prefix:h}),k=fYn(o.breakpoints??Uw),x=mYn({conditions:r.conditions??Uw,breakpoints:k}),O=cXn({config:s,tokens:E});function P(){const{textStyles:Xe,layerStyles:de,animationStyles:We}=o,ht=mK({textStyle:Xe,layerStyle:de,animationStyle:We});for(const[Ie,$t]of Object.entries(ht)){const An=zQt($t??Uw,cen);O.register(Ie,{values:Object.keys(An),transform(Yt){return K(An[Yt])}})}}P(),O.addPropertyType("animationName",Object.keys(o.keyframes??Uw));const D=new Set(["css",...O.keys(),...x.keys()]),$=jx(Xe=>D.has(Xe)||sYn(Xe)),U=Xe=>Array.isArray(Xe)?Xe.reduce((de,We,ht)=>{const Ie=x.breakpoints[ht];return We!=null&&(de[Ie]=We),de},{}):Xe,B=PYn({utility:O,normalize:U}),G=DYn({conditions:x,isValidProperty:$}),K=JQt({transform:O.transform,conditions:x,normalize:B}),X=IYn({css:K,conditions:x,normalize:B,layers:y}),ae=FYn({cva:X});function oe(){const Xe={};for(const[de,We]of E.cssVarMap.entries()){const ht=Object.fromEntries(We);if(Object.keys(ht).length===0)continue;const Ie=de==="base"?d:x.resolve(de),$t=Ie.startsWith("@"),An=K(G({[Ie]:$t?{[d]:ht}:ht}));SK(Xe,An)}return y.wrap("tokens",Xe)}function he(){const Xe=Object.fromEntries(Object.entries(o.keyframes??Uw).map(([We,ht])=>[`@keyframes ${We}`,ht])),de=Object.assign({},Xe,K(G(c)));return y.wrap("base",de)}function ye(Xe){return yW(Xe,$)}function we(){const Xe=LYn({preflight:b});return y.wrap("reset",Xe)}const ge=uXn(E),Te=(Xe,de)=>ge.get(Xe)?.value||de;Te.var=(Xe,de)=>ge.get(Xe)?.variable||de;function xe(Xe,de){return o.recipes?.[Xe]??de}function Ue(Xe,de){return o.slotRecipes?.[Xe]??de}function Z(Xe){return Object.hasOwnProperty.call(o.recipes??Uw,Xe)}function Re(Xe){return Object.hasOwnProperty.call(o.slotRecipes??Uw,Xe)}function _e(Xe){return Z(Xe)||Re(Xe)}const qe=[we(),he(),oe()],it={layerStyles:i$e(o.layerStyles??Uw),textStyles:i$e(o.textStyles??Uw),animationStyles:i$e(o.animationStyles??Uw),tokens:w7t(E,Object.keys(o.tokens??Uw),(Xe,de)=>!Xe.extensions.conditions&&!de.includes("colorPalette")),semanticTokens:w7t(E,Object.keys(o.semanticTokens??Uw),Xe=>!!Xe.extensions.conditions),keyframes:y7t(o.keyframes??Uw),breakpoints:y7t(o.breakpoints??Uw)};return{$$chakra:!0,_config:r,_global:qe,breakpoints:k,tokens:E,conditions:x,utility:O,token:Te,properties:D,layers:y,isValidProperty:$,splitCssProps:ye,normalizeValue:U,getTokenCss:oe,getGlobalCss:he,getPreflightCss:we,css:K,cva:X,sva:ae,getRecipe:xe,getSlotRecipe:Ue,hasRecipe:_e,isRecipe:Z,isSlotRecipe:Re,query:it}}function uXn(n){const r=new Map;return n.allTokens.forEach(o=>{const{cssVar:s,virtual:c,conditions:d}=o.extensions,h=d||c?s.ref:o.value;r.set(o.name,{value:h,variable:s.ref})}),r}const cen=n=>Yw(n)&&"value"in n,i$e=n=>({list(){return Object.keys(zQt(n,cen))},search(r){return this.list().filter(o=>o.includes(r))}}),w7t=(n,r,o)=>({categoryKeys:r,list(s){return Array.from(n.categoryMap.get(s)?.entries()??[]).reduce((c,[d,h])=>(o(h,d)&&c.push(d),c),[])},search(s,c){return this.list(s).filter(d=>d.includes(c))}}),y7t=n=>({list(){return Object.keys(n)},search(r){return this.list().filter(o=>o.includes(r))}}),lXn={sm:"480px",md:"768px",lg:"1024px",xl:"1280px","2xl":"1536px"},o$e="var(--chakra-empty,/*!*/ /*!*/)",dXn=UKn({"*":{fontFeatureSettings:'"cv11"',"--ring-inset":o$e,"--ring-offset-width":"0px","--ring-offset-color":"#fff","--ring-color":"rgba(66, 153, 225, 0.6)","--ring-offset-shadow":"0 0 #0000","--ring-shadow":"0 0 #0000",...Object.fromEntries(["brightness","contrast","grayscale","hue-rotate","invert","saturate","sepia","drop-shadow"].map(n=>[`--${n}`,o$e])),...Object.fromEntries(["blur","brightness","contrast","grayscale","hue-rotate","invert","opacity","saturate","sepia"].map(n=>[`--backdrop-${n}`,o$e])),"--global-font-mono":"fonts.mono","--global-font-body":"fonts.body","--global-color-border":"colors.border"},html:{color:"fg",bg:"bg",lineHeight:"1.5",colorPalette:"gray"},"*::placeholder, *[data-placeholder]":{color:"fg.muted/80"},"*::selection":{bg:"colorPalette.emphasized/80"}}),fXn=GKn({"fill.muted":{value:{background:"colorPalette.muted",color:"colorPalette.fg"}},"fill.subtle":{value:{background:"colorPalette.subtle",color:"colorPalette.fg"}},"fill.surface":{value:{background:"colorPalette.subtle",color:"colorPalette.fg",boxShadow:"0 0 0px 1px var(--shadow-color)",boxShadowColor:"colorPalette.muted"}},"fill.solid":{value:{background:"colorPalette.solid",color:"colorPalette.contrast"}},"outline.subtle":{value:{color:"colorPalette.fg",boxShadow:"inset 0 0 0px 1px var(--shadow-color)",boxShadowColor:"colorPalette.subtle"}},"outline.solid":{value:{borderWidth:"1px",borderColor:"colorPalette.solid",color:"colorPalette.fg"}},"indicator.bottom":{value:{position:"relative","--indicator-color-fallback":"colors.colorPalette.solid",_before:{content:'""',position:"absolute",bottom:"var(--indicator-offset-y, 0)",insetInline:"var(--indicator-offset-x, 0)",height:"var(--indicator-thickness, 2px)",background:"var(--indicator-color, var(--indicator-color-fallback))"}}},"indicator.top":{value:{position:"relative","--indicator-color-fallback":"colors.colorPalette.solid",_before:{content:'""',position:"absolute",top:"var(--indicator-offset-y, 0)",insetInline:"var(--indicator-offset-x, 0)",height:"var(--indicator-thickness, 2px)",background:"var(--indicator-color, var(--indicator-color-fallback))"}}},"indicator.start":{value:{position:"relative","--indicator-color-fallback":"colors.colorPalette.solid",_before:{content:'""',position:"absolute",insetInlineStart:"var(--indicator-offset-x, 0)",insetBlock:"var(--indicator-offset-y, 0)",width:"var(--indicator-thickness, 2px)",background:"var(--indicator-color, var(--indicator-color-fallback))"}}},"indicator.end":{value:{position:"relative","--indicator-color-fallback":"colors.colorPalette.solid",_before:{content:'""',position:"absolute",insetInlineEnd:"var(--indicator-offset-x, 0)",insetBlock:"var(--indicator-offset-y, 0)",width:"var(--indicator-thickness, 2px)",background:"var(--indicator-color, var(--indicator-color-fallback))"}}},disabled:{value:{opacity:"0.5",cursor:"not-allowed"}},none:{value:{}}}),hXn=HKn({"slide-fade-in":{value:{transformOrigin:"var(--transform-origin)","&[data-placement^=top]":{animationName:"slide-from-bottom, fade-in"},"&[data-placement^=bottom]":{animationName:"slide-from-top, fade-in"},"&[data-placement^=left]":{animationName:"slide-from-right, fade-in"},"&[data-placement^=right]":{animationName:"slide-from-left, fade-in"}}},"slide-fade-out":{value:{transformOrigin:"var(--transform-origin)","&[data-placement^=top]":{animationName:"slide-to-bottom, fade-out"},"&[data-placement^=bottom]":{animationName:"slide-to-top, fade-out"},"&[data-placement^=left]":{animationName:"slide-to-right, fade-out"},"&[data-placement^=right]":{animationName:"slide-to-left, fade-out"}}},"scale-fade-in":{value:{transformOrigin:"var(--transform-origin)",animationName:"scale-in, fade-in"}},"scale-fade-out":{value:{transformOrigin:"var(--transform-origin)",animationName:"scale-out, fade-out"}}}),_Ye=Rh({className:"chakra-badge",base:{display:"inline-flex",alignItems:"center",borderRadius:"l2",gap:"1",fontWeight:"medium",fontVariantNumeric:"tabular-nums",whiteSpace:"nowrap",userSelect:"none"},variants:{variant:{solid:{bg:"colorPalette.solid",color:"colorPalette.contrast"},subtle:{bg:"colorPalette.subtle",color:"colorPalette.fg"},outline:{color:"colorPalette.fg",shadow:"inset 0 0 0px 1px var(--shadow-color)",shadowColor:"colorPalette.muted"},surface:{bg:"colorPalette.subtle",color:"colorPalette.fg",shadow:"inset 0 0 0px 1px var(--shadow-color)",shadowColor:"colorPalette.muted"},plain:{color:"colorPalette.fg"}},size:{xs:{textStyle:"2xs",px:"1",minH:"4"},sm:{textStyle:"xs",px:"1.5",minH:"5"},md:{textStyle:"sm",px:"2",minH:"6"},lg:{textStyle:"sm",px:"2.5",minH:"7"}}},defaultVariants:{variant:"subtle",size:"sm"}}),pXn=Rh({className:"chakra-button",base:{display:"inline-flex",appearance:"none",alignItems:"center",justifyContent:"center",userSelect:"none",position:"relative",borderRadius:"l2",whiteSpace:"nowrap",verticalAlign:"middle",borderWidth:"1px",borderColor:"transparent",cursor:"button",flexShrink:"0",outline:"0",lineHeight:"1.2",isolation:"isolate",fontWeight:"medium",transitionProperty:"common",transitionDuration:"moderate",focusVisibleRing:"outside",_disabled:{layerStyle:"disabled"},_icon:{flexShrink:"0"}},variants:{size:{"2xs":{h:"6",minW:"6",textStyle:"xs",px:"2",gap:"1",_icon:{width:"3.5",height:"3.5"}},xs:{h:"8",minW:"8",textStyle:"xs",px:"2.5",gap:"1",_icon:{width:"4",height:"4"}},sm:{h:"9",minW:"9",px:"3.5",textStyle:"sm",gap:"2",_icon:{width:"4",height:"4"}},md:{h:"10",minW:"10",textStyle:"sm",px:"4",gap:"2",_icon:{width:"5",height:"5"}},lg:{h:"11",minW:"11",textStyle:"md",px:"5",gap:"3",_icon:{width:"5",height:"5"}},xl:{h:"12",minW:"12",textStyle:"md",px:"5",gap:"2.5",_icon:{width:"5",height:"5"}},"2xl":{h:"16",minW:"16",textStyle:"lg",px:"7",gap:"3",_icon:{width:"6",height:"6"}}},variant:{solid:{bg:"colorPalette.solid",color:"colorPalette.contrast",borderColor:"transparent",_hover:{bg:"colorPalette.solid/90"},_expanded:{bg:"colorPalette.solid/90"}},subtle:{bg:"colorPalette.subtle",color:"colorPalette.fg",borderColor:"transparent",_hover:{bg:"colorPalette.muted"},_expanded:{bg:"colorPalette.muted"}},surface:{bg:"colorPalette.subtle",color:"colorPalette.fg",shadow:"0 0 0px 1px var(--shadow-color)",shadowColor:"colorPalette.muted",_hover:{bg:"colorPalette.muted"},_expanded:{bg:"colorPalette.muted"}},outline:{borderWidth:"1px",borderColor:"colorPalette.muted",color:"colorPalette.fg",_hover:{bg:"colorPalette.subtle"},_expanded:{bg:"colorPalette.subtle"}},ghost:{bg:"transparent",color:"colorPalette.fg",_hover:{bg:"colorPalette.subtle"},_expanded:{bg:"colorPalette.subtle"}},plain:{color:"colorPalette.fg"}}},defaultVariants:{size:"md",variant:"solid"}}),Fb=Rh({className:"chakra-checkmark",base:{display:"inline-flex",alignItems:"center",justifyContent:"center",flexShrink:"0",color:"white",borderWidth:"1px",borderColor:"transparent",borderRadius:"l1",cursor:"checkbox",focusVisibleRing:"outside",_icon:{boxSize:"full"},_invalid:{colorPalette:"red",borderColor:"border.error"},_disabled:{opacity:"0.5",cursor:"disabled"}},variants:{size:{xs:{boxSize:"3"},sm:{boxSize:"4"},md:{boxSize:"5",p:"0.5"},lg:{boxSize:"6",p:"0.5"}},variant:{solid:{borderColor:"border.emphasized","&:is([data-state=checked], [data-state=indeterminate])":{bg:"colorPalette.solid",color:"colorPalette.contrast",borderColor:"colorPalette.solid"}},outline:{borderColor:"border","&:is([data-state=checked], [data-state=indeterminate])":{color:"colorPalette.fg",borderColor:"colorPalette.solid"}},subtle:{bg:"colorPalette.muted",borderColor:"colorPalette.muted","&:is([data-state=checked], [data-state=indeterminate])":{color:"colorPalette.fg"}},plain:{"&:is([data-state=checked], [data-state=indeterminate])":{color:"colorPalette.fg"}},inverted:{borderColor:"border",color:"colorPalette.fg","&:is([data-state=checked], [data-state=indeterminate])":{borderColor:"colorPalette.solid"}}},filled:{true:{bg:"bg"}}},defaultVariants:{variant:"solid",size:"md"}}),{variants:gXn,defaultVariants:mXn}=_Ye,bXn=Rh({className:"chakra-code",base:{fontFamily:"mono",alignItems:"center",display:"inline-flex",borderRadius:"l2"},variants:gXn,defaultVariants:mXn}),uen=Rh({className:"color-swatch",base:{boxSize:"var(--swatch-size)",shadow:"inset 0 0 0 1px rgba(0, 0, 0, 0.1)","--checker-size":"8px","--checker-bg":"colors.bg","--checker-fg":"colors.bg.emphasized",background:"linear-gradient(var(--color), var(--color)), repeating-conic-gradient(var(--checker-fg) 0%, var(--checker-fg) 25%, var(--checker-bg) 0%, var(--checker-bg) 50%) 0% 50% / var(--checker-size) var(--checker-size) !important",display:"inline-flex",alignItems:"center",justifyContent:"center",flexShrink:"0"},variants:{size:{"2xs":{"--swatch-size":"sizes.3.5"},xs:{"--swatch-size":"sizes.4"},sm:{"--swatch-size":"sizes.4.5"},md:{"--swatch-size":"sizes.5"},lg:{"--swatch-size":"sizes.6"},xl:{"--swatch-size":"sizes.7"},"2xl":{"--swatch-size":"sizes.8"},inherit:{"--swatch-size":"inherit"},full:{"--swatch-size":"100%"}},shape:{square:{borderRadius:"none"},circle:{borderRadius:"full"},rounded:{borderRadius:"l1"}}},defaultVariants:{size:"md",shape:"rounded"}}),vXn=Rh({className:"chakra-container",base:{position:"relative",maxWidth:"8xl",w:"100%",mx:"auto",px:{base:"4",md:"6",lg:"8"}},variants:{centerContent:{true:{display:"flex",flexDirection:"column",alignItems:"center"}},fluid:{true:{maxWidth:"full"}}}}),wXn=Rh({className:"chakra-heading",base:{fontFamily:"heading",fontWeight:"semibold"},variants:{size:{xs:{textStyle:"xs"},sm:{textStyle:"sm"},md:{textStyle:"md"},lg:{textStyle:"lg"},xl:{textStyle:"xl"},"2xl":{textStyle:"2xl"},"3xl":{textStyle:"3xl"},"4xl":{textStyle:"4xl"},"5xl":{textStyle:"5xl"},"6xl":{textStyle:"6xl"},"7xl":{textStyle:"7xl"}}},defaultVariants:{size:"xl"}}),yXn=Rh({className:"chakra-icon",base:{display:"inline-block",lineHeight:"1em",flexShrink:"0",color:"currentcolor",verticalAlign:"middle"},variants:{size:{inherit:{},xs:{boxSize:"3"},sm:{boxSize:"4"},md:{boxSize:"5"},lg:{boxSize:"6"},xl:{boxSize:"7"},"2xl":{boxSize:"8"}}},defaultVariants:{size:"inherit"}}),qp=Rh({className:"chakra-input",base:{width:"100%",minWidth:"0",outline:"0",position:"relative",appearance:"none",textAlign:"start",borderRadius:"l2",_disabled:{layerStyle:"disabled"},height:"var(--input-height)",minW:"var(--input-height)","--focus-color":"colors.colorPalette.focusRing","--error-color":"colors.border.error",_invalid:{focusRingColor:"var(--error-color)",borderColor:"var(--error-color)"}},variants:{size:{"2xs":{textStyle:"xs",px:"2","--input-height":"sizes.7"},xs:{textStyle:"xs",px:"2","--input-height":"sizes.8"},sm:{textStyle:"sm",px:"2.5","--input-height":"sizes.9"},md:{textStyle:"sm",px:"3","--input-height":"sizes.10"},lg:{textStyle:"md",px:"4","--input-height":"sizes.11"},xl:{textStyle:"md",px:"4.5","--input-height":"sizes.12"},"2xl":{textStyle:"lg",px:"5","--input-height":"sizes.16"}},variant:{outline:{bg:"transparent",borderWidth:"1px",borderColor:"border",focusVisibleRing:"inside",focusRingColor:"var(--focus-color)"},subtle:{borderWidth:"1px",borderColor:"transparent",bg:"bg.muted",focusVisibleRing:"inside",focusRingColor:"var(--focus-color)"},flushed:{bg:"transparent",borderBottomWidth:"1px",borderBottomColor:"border",borderRadius:"0",px:"0",_focusVisible:{borderColor:"var(--focus-color)",boxShadow:"0px 1px 0px 0px var(--focus-color)",_invalid:{borderColor:"var(--error-color)",boxShadow:"0px 1px 0px 0px var(--error-color)"}}}}},defaultVariants:{size:"md",variant:"outline"}}),EXn=Rh({className:"chakra-input-addon",base:{flex:"0 0 auto",width:"auto",display:"flex",alignItems:"center",whiteSpace:"nowrap",alignSelf:"stretch",borderRadius:"l2"},variants:{size:qp.variants.size,variant:{outline:{borderWidth:"1px",borderColor:"border",bg:"bg.muted"},subtle:{borderWidth:"1px",borderColor:"transparent",bg:"bg.emphasized"},flushed:{borderBottom:"1px solid",borderColor:"inherit",borderRadius:"0",px:"0",bg:"transparent"}}},defaultVariants:{size:"md",variant:"outline"}}),SXn=Rh({className:"chakra-kbd",base:{display:"inline-flex",alignItems:"center",fontWeight:"medium",fontFamily:"mono",flexShrink:"0",whiteSpace:"nowrap",wordSpacing:"-0.5em",userSelect:"none",px:"1",borderRadius:"l2"},variants:{variant:{raised:{bg:"colorPalette.subtle",color:"colorPalette.fg",borderWidth:"1px",borderBottomWidth:"2px",borderColor:"colorPalette.muted"},outline:{borderWidth:"1px",color:"colorPalette.fg"},subtle:{bg:"colorPalette.muted",color:"colorPalette.fg"},plain:{color:"colorPalette.fg"}},size:{sm:{textStyle:"xs",height:"4.5"},md:{textStyle:"sm",height:"5"},lg:{textStyle:"md",height:"6"}}},defaultVariants:{size:"md",variant:"raised"}}),_Xn=Rh({className:"chakra-link",base:{display:"inline-flex",alignItems:"center",outline:"none",gap:"1.5",cursor:"pointer",borderRadius:"l1",focusRing:"outside"},variants:{variant:{underline:{color:"colorPalette.fg",textDecoration:"underline",textUnderlineOffset:"3px",textDecorationColor:"currentColor/20"},plain:{color:"colorPalette.fg",_hover:{textDecoration:"underline",textUnderlineOffset:"3px",textDecorationColor:"currentColor/20"}}}},defaultVariants:{variant:"plain"}}),kXn=Rh({className:"chakra-mark",base:{bg:"transparent",color:"inherit",whiteSpace:"nowrap"},variants:{variant:{subtle:{bg:"colorPalette.subtle",color:"inherit"},solid:{bg:"colorPalette.solid",color:"colorPalette.contrast"},text:{fontWeight:"medium"},plain:{}}}}),Bb=Rh({className:"chakra-radiomark",base:{display:"inline-flex",alignItems:"center",justifyContent:"center",flexShrink:0,verticalAlign:"top",color:"white",borderWidth:"1px",borderColor:"transparent",borderRadius:"full",cursor:"radio",_focusVisible:{outline:"2px solid",outlineColor:"colorPalette.focusRing",outlineOffset:"2px"},_invalid:{colorPalette:"red",borderColor:"red.500"},_disabled:{opacity:"0.5",cursor:"disabled"},"& .dot":{height:"100%",width:"100%",borderRadius:"full",bg:"currentColor",scale:"0.4"}},variants:{variant:{solid:{borderWidth:"1px",borderColor:"border.emphasized",_checked:{bg:"colorPalette.solid",color:"colorPalette.contrast",borderColor:"colorPalette.solid"}},subtle:{borderWidth:"1px",bg:"colorPalette.muted",borderColor:"colorPalette.muted",color:"transparent",_checked:{color:"colorPalette.fg"}},outline:{borderWidth:"1px",borderColor:"inherit",_checked:{color:"colorPalette.fg",borderColor:"colorPalette.solid"},"& .dot":{scale:"0.6"}},inverted:{bg:"bg",borderWidth:"1px",borderColor:"inherit",_checked:{color:"colorPalette.solid",borderColor:"currentcolor"}}},size:{xs:{boxSize:"3"},sm:{boxSize:"4"},md:{boxSize:"5"},lg:{boxSize:"6"}},filled:{true:{bg:"bg"}}},defaultVariants:{variant:"solid",size:"md"}}),TXn=Rh({className:"chakra-separator",base:{display:"block",borderColor:"border"},variants:{variant:{solid:{borderStyle:"solid"},dashed:{borderStyle:"dashed"},dotted:{borderStyle:"dotted"}},orientation:{vertical:{borderInlineStartWidth:"var(--separator-thickness)"},horizontal:{borderTopWidth:"var(--separator-thickness)"}},size:{xs:{"--separator-thickness":"0.5px"},sm:{"--separator-thickness":"1px"},md:{"--separator-thickness":"2px"},lg:{"--separator-thickness":"3px"}}},defaultVariants:{size:"sm",variant:"solid",orientation:"horizontal"}}),xXn=Rh({className:"chakra-skeleton",base:{},variants:{loading:{true:{borderRadius:"l2",boxShadow:"none",backgroundClip:"padding-box",cursor:"default",color:"transparent",pointerEvents:"none",userSelect:"none",flexShrink:"0","&::before, &::after, *":{visibility:"hidden"}},false:{background:"unset",animation:"fade-in var(--fade-duration, 0.1s) ease-out !important"}},variant:{pulse:{background:"bg.emphasized",animation:"pulse",animationDuration:"var(--duration, 1.2s)"},shine:{"--animate-from":"200%","--animate-to":"-200%","--start-color":"colors.bg.muted","--end-color":"colors.bg.emphasized",backgroundImage:"linear-gradient(270deg,var(--start-color),var(--end-color),var(--end-color),var(--start-color))",backgroundSize:"400% 100%",animation:"bg-position var(--duration, 5s) ease-in-out infinite"},none:{animation:"none"}}},defaultVariants:{variant:"pulse",loading:!0}}),CXn=Rh({className:"chakra-skip-nav",base:{display:"inline-flex",bg:"bg.panel",padding:"2.5",borderRadius:"l2",fontWeight:"semibold",focusVisibleRing:"outside",textStyle:"sm",userSelect:"none",border:"0",height:"1px",width:"1px",margin:"-1px",outline:"0",overflow:"hidden",position:"absolute",clip:"rect(0 0 0 0)",_focusVisible:{clip:"auto",width:"auto",height:"auto",position:"fixed",top:"6",insetStart:"6"}}}),AXn=Rh({className:"chakra-spinner",base:{display:"inline-block",borderColor:"currentColor",borderStyle:"solid",borderWidth:"2px",borderRadius:"full",width:"var(--spinner-size)",height:"var(--spinner-size)",animation:"spin",animationDuration:"slowest","--spinner-track-color":"transparent",borderBottomColor:"var(--spinner-track-color)",borderInlineStartColor:"var(--spinner-track-color)"},variants:{size:{inherit:{"--spinner-size":"1em"},xs:{"--spinner-size":"sizes.3"},sm:{"--spinner-size":"sizes.4"},md:{"--spinner-size":"sizes.5"},lg:{"--spinner-size":"sizes.8"},xl:{"--spinner-size":"sizes.10"}}},defaultVariants:{size:"md"}}),IXn=Rh({className:"chakra-textarea",base:{width:"100%",minWidth:"0",outline:"0",position:"relative",appearance:"none",textAlign:"start",borderRadius:"l2",_disabled:{layerStyle:"disabled"},"--focus-color":"colors.colorPalette.focusRing","--error-color":"colors.border.error",_invalid:{focusRingColor:"var(--error-color)",borderColor:"var(--error-color)"}},variants:{size:{xs:{textStyle:"xs",px:"2",py:"1.5",scrollPaddingBottom:"1.5"},sm:{textStyle:"sm",px:"2.5",py:"2",scrollPaddingBottom:"2"},md:{textStyle:"sm",px:"3",py:"2",scrollPaddingBottom:"2"},lg:{textStyle:"md",px:"4",py:"3",scrollPaddingBottom:"3"},xl:{textStyle:"md",px:"4.5",py:"3.5",scrollPaddingBottom:"3.5"}},variant:{outline:{bg:"transparent",borderWidth:"1px",borderColor:"border",focusVisibleRing:"inside"},subtle:{borderWidth:"1px",borderColor:"transparent",bg:"bg.muted",focusVisibleRing:"inside"},flushed:{bg:"transparent",borderBottomWidth:"1px",borderBottomColor:"border",borderRadius:"0",px:"0",_focusVisible:{borderColor:"var(--focus-color)",boxShadow:"0px 1px 0px 0px var(--focus-color)"}}}},defaultVariants:{size:"md",variant:"outline"}}),OXn={badge:_Ye,button:pXn,code:bXn,container:vXn,heading:wXn,input:qp,inputAddon:EXn,kbd:SXn,link:_Xn,mark:kXn,separator:TXn,skeleton:xXn,skipNavLink:CXn,spinner:AXn,textarea:IXn,icon:yXn,checkmark:Fb,radiomark:Bb,colorSwatch:uen},NXn=yYe.colors({bg:{DEFAULT:{value:{_light:"{colors.white}",_dark:"{colors.black}"}},subtle:{value:{_light:"{colors.gray.50}",_dark:"{colors.gray.950}"}},muted:{value:{_light:"{colors.gray.100}",_dark:"{colors.gray.900}"}},emphasized:{value:{_light:"{colors.gray.200}",_dark:"{colors.gray.800}"}},inverted:{value:{_light:"{colors.black}",_dark:"{colors.white}"}},panel:{value:{_light:"{colors.white}",_dark:"{colors.gray.950}"}},error:{value:{_light:"{colors.red.50}",_dark:"{colors.red.950}"}},warning:{value:{_light:"{colors.orange.50}",_dark:"{colors.orange.950}"}},success:{value:{_light:"{colors.green.50}",_dark:"{colors.green.950}"}},info:{value:{_light:"{colors.blue.50}",_dark:"{colors.blue.950}"}}},fg:{DEFAULT:{value:{_light:"{colors.black}",_dark:"{colors.gray.50}"}},muted:{value:{_light:"{colors.gray.600}",_dark:"{colors.gray.400}"}},subtle:{value:{_light:"{colors.gray.400}",_dark:"{colors.gray.500}"}},inverted:{value:{_light:"{colors.gray.50}",_dark:"{colors.black}"}},error:{value:{_light:"{colors.red.500}",_dark:"{colors.red.400}"}},warning:{value:{_light:"{colors.orange.600}",_dark:"{colors.orange.300}"}},success:{value:{_light:"{colors.green.600}",_dark:"{colors.green.300}"}},info:{value:{_light:"{colors.blue.600}",_dark:"{colors.blue.300}"}}},border:{DEFAULT:{value:{_light:"{colors.gray.200}",_dark:"{colors.gray.800}"}},muted:{value:{_light:"{colors.gray.100}",_dark:"{colors.gray.900}"}},subtle:{value:{_light:"{colors.gray.50}",_dark:"{colors.gray.950}"}},emphasized:{value:{_light:"{colors.gray.300}",_dark:"{colors.gray.700}"}},inverted:{value:{_light:"{colors.gray.800}",_dark:"{colors.gray.200}"}},error:{value:{_light:"{colors.red.500}",_dark:"{colors.red.400}"}},warning:{value:{_light:"{colors.orange.500}",_dark:"{colors.orange.400}"}},success:{value:{_light:"{colors.green.500}",_dark:"{colors.green.400}"}},info:{value:{_light:"{colors.blue.500}",_dark:"{colors.blue.400}"}}},gray:{contrast:{value:{_light:"{colors.white}",_dark:"{colors.black}"}},fg:{value:{_light:"{colors.gray.800}",_dark:"{colors.gray.200}"}},subtle:{value:{_light:"{colors.gray.100}",_dark:"{colors.gray.900}"}},muted:{value:{_light:"{colors.gray.200}",_dark:"{colors.gray.800}"}},emphasized:{value:{_light:"{colors.gray.300}",_dark:"{colors.gray.700}"}},solid:{value:{_light:"{colors.gray.900}",_dark:"{colors.white}"}},focusRing:{value:{_light:"{colors.gray.400}",_dark:"{colors.gray.400}"}}},red:{contrast:{value:{_light:"white",_dark:"white"}},fg:{value:{_light:"{colors.red.700}",_dark:"{colors.red.300}"}},subtle:{value:{_light:"{colors.red.100}",_dark:"{colors.red.900}"}},muted:{value:{_light:"{colors.red.200}",_dark:"{colors.red.800}"}},emphasized:{value:{_light:"{colors.red.300}",_dark:"{colors.red.700}"}},solid:{value:{_light:"{colors.red.600}",_dark:"{colors.red.600}"}},focusRing:{value:{_light:"{colors.red.500}",_dark:"{colors.red.500}"}}},orange:{contrast:{value:{_light:"white",_dark:"black"}},fg:{value:{_light:"{colors.orange.700}",_dark:"{colors.orange.300}"}},subtle:{value:{_light:"{colors.orange.100}",_dark:"{colors.orange.900}"}},muted:{value:{_light:"{colors.orange.200}",_dark:"{colors.orange.800}"}},emphasized:{value:{_light:"{colors.orange.300}",_dark:"{colors.orange.700}"}},solid:{value:{_light:"{colors.orange.600}",_dark:"{colors.orange.500}"}},focusRing:{value:{_light:"{colors.orange.500}",_dark:"{colors.orange.500}"}}},green:{contrast:{value:{_light:"white",_dark:"white"}},fg:{value:{_light:"{colors.green.700}",_dark:"{colors.green.300}"}},subtle:{value:{_light:"{colors.green.100}",_dark:"{colors.green.900}"}},muted:{value:{_light:"{colors.green.200}",_dark:"{colors.green.800}"}},emphasized:{value:{_light:"{colors.green.300}",_dark:"{colors.green.700}"}},solid:{value:{_light:"{colors.green.600}",_dark:"{colors.green.600}"}},focusRing:{value:{_light:"{colors.green.500}",_dark:"{colors.green.500}"}}},blue:{contrast:{value:{_light:"white",_dark:"white"}},fg:{value:{_light:"{colors.blue.700}",_dark:"{colors.blue.300}"}},subtle:{value:{_light:"{colors.blue.100}",_dark:"{colors.blue.900}"}},muted:{value:{_light:"{colors.blue.200}",_dark:"{colors.blue.800}"}},emphasized:{value:{_light:"{colors.blue.300}",_dark:"{colors.blue.700}"}},solid:{value:{_light:"{colors.blue.600}",_dark:"{colors.blue.600}"}},focusRing:{value:{_light:"{colors.blue.500}",_dark:"{colors.blue.500}"}}},yellow:{contrast:{value:{_light:"black",_dark:"black"}},fg:{value:{_light:"{colors.yellow.800}",_dark:"{colors.yellow.300}"}},subtle:{value:{_light:"{colors.yellow.100}",_dark:"{colors.yellow.900}"}},muted:{value:{_light:"{colors.yellow.200}",_dark:"{colors.yellow.800}"}},emphasized:{value:{_light:"{colors.yellow.300}",_dark:"{colors.yellow.700}"}},solid:{value:{_light:"{colors.yellow.300}",_dark:"{colors.yellow.300}"}},focusRing:{value:{_light:"{colors.yellow.500}",_dark:"{colors.yellow.500}"}}},teal:{contrast:{value:{_light:"white",_dark:"white"}},fg:{value:{_light:"{colors.teal.700}",_dark:"{colors.teal.300}"}},subtle:{value:{_light:"{colors.teal.100}",_dark:"{colors.teal.900}"}},muted:{value:{_light:"{colors.teal.200}",_dark:"{colors.teal.800}"}},emphasized:{value:{_light:"{colors.teal.300}",_dark:"{colors.teal.700}"}},solid:{value:{_light:"{colors.teal.600}",_dark:"{colors.teal.600}"}},focusRing:{value:{_light:"{colors.teal.500}",_dark:"{colors.teal.500}"}}},purple:{contrast:{value:{_light:"white",_dark:"white"}},fg:{value:{_light:"{colors.purple.700}",_dark:"{colors.purple.300}"}},subtle:{value:{_light:"{colors.purple.100}",_dark:"{colors.purple.900}"}},muted:{value:{_light:"{colors.purple.200}",_dark:"{colors.purple.800}"}},emphasized:{value:{_light:"{colors.purple.300}",_dark:"{colors.purple.700}"}},solid:{value:{_light:"{colors.purple.600}",_dark:"{colors.purple.600}"}},focusRing:{value:{_light:"{colors.purple.500}",_dark:"{colors.purple.500}"}}},pink:{contrast:{value:{_light:"white",_dark:"white"}},fg:{value:{_light:"{colors.pink.700}",_dark:"{colors.pink.300}"}},subtle:{value:{_light:"{colors.pink.100}",_dark:"{colors.pink.900}"}},muted:{value:{_light:"{colors.pink.200}",_dark:"{colors.pink.800}"}},emphasized:{value:{_light:"{colors.pink.300}",_dark:"{colors.pink.700}"}},solid:{value:{_light:"{colors.pink.600}",_dark:"{colors.pink.600}"}},focusRing:{value:{_light:"{colors.pink.500}",_dark:"{colors.pink.500}"}}},cyan:{contrast:{value:{_light:"white",_dark:"white"}},fg:{value:{_light:"{colors.cyan.700}",_dark:"{colors.cyan.300}"}},subtle:{value:{_light:"{colors.cyan.100}",_dark:"{colors.cyan.900}"}},muted:{value:{_light:"{colors.cyan.200}",_dark:"{colors.cyan.800}"}},emphasized:{value:{_light:"{colors.cyan.300}",_dark:"{colors.cyan.700}"}},solid:{value:{_light:"{colors.cyan.600}",_dark:"{colors.cyan.600}"}},focusRing:{value:{_light:"{colors.cyan.500}",_dark:"{colors.cyan.500}"}}}}),RXn=yYe.radii({l1:{value:"{radii.xs}"},l2:{value:"{radii.sm}"},l3:{value:"{radii.md}"}}),PXn=yYe.shadows({xs:{value:{_light:"0px 1px 2px {colors.gray.900/10}, 0px 0px 1px {colors.gray.900/20}",_dark:"0px 1px 1px {black/64}, 0px 0px 1px inset {colors.gray.300/20}"}},sm:{value:{_light:"0px 2px 4px {colors.gray.900/10}, 0px 0px 1px {colors.gray.900/30}",_dark:"0px 2px 4px {black/64}, 0px 0px 1px inset {colors.gray.300/30}"}},md:{value:{_light:"0px 4px 8px {colors.gray.900/10}, 0px 0px 1px {colors.gray.900/30}",_dark:"0px 4px 8px {black/64}, 0px 0px 1px inset {colors.gray.300/30}"}},lg:{value:{_light:"0px 8px 16px {colors.gray.900/10}, 0px 0px 1px {colors.gray.900/30}",_dark:"0px 8px 16px {black/64}, 0px 0px 1px inset {colors.gray.300/30}"}},xl:{value:{_light:"0px 16px 24px {colors.gray.900/10}, 0px 0px 1px {colors.gray.900/30}",_dark:"0px 16px 24px {black/64}, 0px 0px 1px inset {colors.gray.300/30}"}},"2xl":{value:{_light:"0px 24px 40px {colors.gray.900/16}, 0px 0px 1px {colors.gray.900/30}",_dark:"0px 24px 40px {black/64}, 0px 0px 1px inset {colors.gray.300/30}"}},inner:{value:{_light:"inset 0 2px 4px 0 {black/5}",_dark:"inset 0 2px 4px 0 black"}},inset:{value:{_light:"inset 0 0 0 1px {black/5}",_dark:"inset 0 0 0 1px {colors.gray.300/5}"}}}),LXn=VYt.extendWith("itemBody"),DXn=Vo("action-bar").parts("positioner","content","separator","selectionTrigger","closeTrigger"),MXn=Vo("alert").parts("title","description","root","indicator","content"),jXn=Vo("breadcrumb").parts("link","currentLink","item","list","root","ellipsis","separator"),$Xn=Vo("blockquote").parts("root","icon","content","caption"),FXn=Vo("card").parts("root","header","body","footer","title","description"),BXn=Vo("checkbox-card",["root","control","label","description","addon","indicator","content"]),UXn=Vo("data-list").parts("root","item","itemLabel","itemValue"),zXn=cYe.extendWith("header","body","footer","backdrop"),HXn=cYe.extendWith("header","body","footer","backdrop"),GXn=UXt.extendWith("textarea"),qXn=Vo("empty-state",["root","content","indicator","title","description"]),VXn=eJt.extendWith("requiredIndicator"),WXn=aJt.extendWith("content"),KXn=cJt.extendWith("itemContent","dropzoneContent","fileText"),YXn=Vo("list").parts("root","item","indicator"),XXn=qJt.extendWith("itemCommand"),JXn=Vo("select").parts("root","field","indicator"),ZXn=mZt.extendWith("header","body","footer"),len=pYe.extendWith("itemAddon","itemIndicator"),QXn=len.extendWith("itemContent","itemDescription"),eJn=LZt.extendWith("itemIndicator"),tJn=nQt.extendWith("indicatorGroup"),nJn=fqn.extendWith("indicatorGroup","empty"),rJn=iQt.extendWith("markerIndicator"),iJn=Vo("stat").parts("root","label","helpText","valueText","valueUnit","indicator"),oJn=Vo("status").parts("root","indicator"),aJn=Vo("steps",["root","list","item","trigger","indicator","separator","content","title","description","nextTrigger","prevTrigger","progress"]),sJn=bQt.extendWith("indicator"),cJn=Vo("table").parts("root","header","body","row","columnHeader","cell","footer","caption"),uJn=Vo("toast").parts("root","title","description","indicator","closeTrigger","actionTrigger"),lJn=Vo("tabs").parts("root","trigger","list","content","contentGroup","indicator"),dJn=Vo("tag").parts("root","label","closeTrigger","startElement","endElement"),fJn=Vo("timeline").parts("root","item","content","separator","indicator","connector","title","description"),hJn=AGn.extendWith("channelText"),pJn=Vo("code-block",["root","content","title","header","footer","control","overlay","code","codeText","copyTrigger","copyIndicator","collapseTrigger","collapseIndicator","collapseText"]);uXt.extendWith("valueText");const gJn=lVn,mJn=Bo({className:"chakra-accordion",slots:LXn.keys(),base:{root:{width:"full","--accordion-radius":"radii.l2"},item:{overflowAnchor:"none"},itemTrigger:{display:"flex",alignItems:"center",textAlign:"start",width:"full",outline:"0",gap:"3",fontWeight:"medium",borderRadius:"var(--accordion-radius)",_focusVisible:{outline:"2px solid",outlineColor:"colorPalette.focusRing"},_disabled:{layerStyle:"disabled"}},itemBody:{pt:"var(--accordion-padding-y)",pb:"calc(var(--accordion-padding-y) * 2)"},itemContent:{overflow:"hidden",borderRadius:"var(--accordion-radius)",_open:{animationName:"expand-height, fade-in",animationDuration:"moderate"},_closed:{animationName:"collapse-height, fade-out",animationDuration:"moderate"}},itemIndicator:{transition:"rotate 0.2s",transformOrigin:"center",color:"fg.subtle",_open:{rotate:"180deg"},_icon:{width:"1.2em",height:"1.2em"}}},variants:{variant:{outline:{item:{borderBottomWidth:"1px"}},subtle:{itemTrigger:{px:"var(--accordion-padding-x)"},itemContent:{px:"var(--accordion-padding-x)"},item:{borderRadius:"var(--accordion-radius)",_open:{bg:"colorPalette.subtle"}}},enclosed:{root:{borderWidth:"1px",borderRadius:"var(--accordion-radius)",divideY:"1px",overflow:"hidden"},itemTrigger:{px:"var(--accordion-padding-x)"},itemContent:{px:"var(--accordion-padding-x)"},item:{_open:{bg:"bg.subtle"}}},plain:{}},size:{sm:{root:{"--accordion-padding-x":"spacing.3","--accordion-padding-y":"spacing.2"},itemTrigger:{textStyle:"sm",py:"var(--accordion-padding-y)"}},md:{root:{"--accordion-padding-x":"spacing.4","--accordion-padding-y":"spacing.2"},itemTrigger:{textStyle:"md",py:"var(--accordion-padding-y)"}},lg:{root:{"--accordion-padding-x":"spacing.4.5","--accordion-padding-y":"spacing.2.5"},itemTrigger:{textStyle:"lg",py:"var(--accordion-padding-y)"}}}},defaultVariants:{size:"md",variant:"outline"}}),bJn=Bo({className:"chakra-action-bar",slots:DXn.keys(),base:{positioner:{position:"fixed",display:"flex",justifyContent:"center",pointerEvents:"none",insetInline:"0",top:"unset",bottom:"calc(env(safe-area-inset-bottom) + 20px)"},content:{bg:"bg.panel",shadow:"md",display:"flex",alignItems:"center",gap:"3",borderRadius:"l3",py:"2.5",px:"3",pointerEvents:"auto",translate:"calc(-1 * var(--scrollbar-width) / 2) 0px",_open:{animationName:"slide-from-bottom, fade-in",animationDuration:"moderate"},_closed:{animationName:"slide-to-bottom, fade-out",animationDuration:"faster"}},separator:{width:"1px",height:"5",bg:"border"},selectionTrigger:{display:"inline-flex",alignItems:"center",gap:"2",alignSelf:"stretch",textStyle:"sm",px:"4",py:"1",borderRadius:"l2",borderWidth:"1px",borderStyle:"dashed"}}}),vJn=Bo({slots:MXn.keys(),className:"chakra-alert",base:{root:{width:"full",display:"flex",alignItems:"flex-start",position:"relative",borderRadius:"l3"},title:{fontWeight:"medium"},description:{display:"inline"},indicator:{display:"inline-flex",alignItems:"center",justifyContent:"center",flexShrink:"0",width:"1em",height:"1em",_icon:{boxSize:"full"}},content:{display:"flex",flex:"1",gap:"1"}},variants:{status:{info:{root:{colorPalette:"blue"}},warning:{root:{colorPalette:"orange"}},success:{root:{colorPalette:"green"}},error:{root:{colorPalette:"red"}},neutral:{root:{colorPalette:"gray"}}},inline:{true:{content:{display:"inline-flex",flexDirection:"row",alignItems:"center"}},false:{content:{display:"flex",flexDirection:"column"}}},variant:{subtle:{root:{bg:"colorPalette.subtle",color:"colorPalette.fg"}},surface:{root:{bg:"colorPalette.subtle",color:"colorPalette.fg",shadow:"inset 0 0 0px 1px var(--shadow-color)",shadowColor:"colorPalette.muted"},indicator:{color:"colorPalette.fg"}},outline:{root:{color:"colorPalette.fg",shadow:"inset 0 0 0px 1px var(--shadow-color)",shadowColor:"colorPalette.muted"},indicator:{color:"colorPalette.fg"}},solid:{root:{bg:"colorPalette.solid",color:"colorPalette.contrast"},indicator:{color:"colorPalette.contrast"}}},size:{sm:{root:{gap:"2",px:"3",py:"3",textStyle:"xs"},indicator:{textStyle:"lg"}},md:{root:{gap:"3",px:"4",py:"4",textStyle:"sm"},indicator:{textStyle:"xl"}},lg:{root:{gap:"3",px:"4",py:"4",textStyle:"md"},indicator:{textStyle:"2xl"}}}},defaultVariants:{status:"info",variant:"subtle",size:"md",inline:!1}}),wJn=Bo({slots:QYt.keys(),className:"chakra-avatar",base:{root:{display:"inline-flex",alignItems:"center",justifyContent:"center",fontWeight:"medium",position:"relative",verticalAlign:"top",flexShrink:"0",userSelect:"none",width:"var(--avatar-size)",height:"var(--avatar-size)",fontSize:"var(--avatar-font-size)",borderRadius:"var(--avatar-radius)","&[data-group-item]":{borderWidth:"2px",borderColor:"bg"}},image:{width:"100%",height:"100%",objectFit:"cover",borderRadius:"var(--avatar-radius)"},fallback:{lineHeight:"1",textTransform:"uppercase",fontWeight:"medium",fontSize:"var(--avatar-font-size)",borderRadius:"var(--avatar-radius)"}},variants:{size:{full:{root:{"--avatar-size":"100%","--avatar-font-size":"100%"}},"2xs":{root:{"--avatar-font-size":"fontSizes.2xs","--avatar-size":"sizes.6"}},xs:{root:{"--avatar-font-size":"fontSizes.xs","--avatar-size":"sizes.8"}},sm:{root:{"--avatar-font-size":"fontSizes.sm","--avatar-size":"sizes.9"}},md:{root:{"--avatar-font-size":"fontSizes.md","--avatar-size":"sizes.10"}},lg:{root:{"--avatar-font-size":"fontSizes.md","--avatar-size":"sizes.11"}},xl:{root:{"--avatar-font-size":"fontSizes.lg","--avatar-size":"sizes.12"}},"2xl":{root:{"--avatar-font-size":"fontSizes.xl","--avatar-size":"sizes.16"}}},variant:{solid:{root:{bg:"colorPalette.solid",color:"colorPalette.contrast"}},subtle:{root:{bg:"colorPalette.muted",color:"colorPalette.fg"}},outline:{root:{color:"colorPalette.fg",borderWidth:"1px",borderColor:"colorPalette.muted"}}},shape:{square:{},rounded:{root:{"--avatar-radius":"radii.l3"}},full:{root:{"--avatar-radius":"radii.full"}}},borderless:{true:{root:{"&[data-group-item]":{borderWidth:"0px"}}}}},defaultVariants:{size:"md",shape:"full",variant:"subtle"}}),yJn=Bo({className:"chakra-blockquote",slots:$Xn.keys(),base:{root:{position:"relative",display:"flex",flexDirection:"column",gap:"2"},caption:{textStyle:"sm",color:"fg.muted"},icon:{boxSize:"5"}},variants:{justify:{start:{root:{alignItems:"flex-start",textAlign:"start"}},center:{root:{alignItems:"center",textAlign:"center"}},end:{root:{alignItems:"flex-end",textAlign:"end"}}},variant:{subtle:{root:{paddingX:"5",borderStartWidth:"4px",borderStartColor:"colorPalette.muted"},icon:{color:"colorPalette.fg"}},solid:{root:{paddingX:"5",borderStartWidth:"4px",borderStartColor:"colorPalette.solid"},icon:{color:"colorPalette.solid"}},plain:{root:{paddingX:"5"},icon:{color:"colorPalette.solid"}}}},defaultVariants:{variant:"subtle",justify:"start"}}),EJn=Bo({className:"chakra-breadcrumb",slots:jXn.keys(),base:{list:{display:"flex",alignItems:"center",wordBreak:"break-word",color:"fg.muted",listStyle:"none"},link:{outline:"0",textDecoration:"none",borderRadius:"l1",focusRing:"outside",display:"inline-flex",alignItems:"center",gap:"2"},item:{display:"inline-flex",alignItems:"center"},separator:{color:"fg.muted",opacity:"0.8",_icon:{boxSize:"1em"},_rtl:{rotate:"180deg"}},ellipsis:{display:"inline-flex",alignItems:"center",justifyContent:"center",_icon:{boxSize:"1em"}}},variants:{variant:{underline:{link:{color:"colorPalette.fg",textDecoration:"underline",textUnderlineOffset:"0.2em",textDecorationColor:"colorPalette.muted"},currentLink:{color:"colorPalette.fg"}},plain:{link:{color:"fg.muted",_hover:{color:"fg"}},currentLink:{color:"fg"}}},size:{sm:{list:{gap:"1",textStyle:"xs"}},md:{list:{gap:"1.5",textStyle:"sm"}},lg:{list:{gap:"2",textStyle:"md"}}}},defaultVariants:{variant:"plain",size:"md"}}),SJn=Bo({className:"chakra-card",slots:FXn.keys(),base:{root:{display:"flex",flexDirection:"column",position:"relative",minWidth:"0",wordWrap:"break-word",borderRadius:"l3",color:"fg",textAlign:"start"},title:{fontWeight:"semibold"},description:{color:"fg.muted",fontSize:"sm"},header:{paddingInline:"var(--card-padding)",paddingTop:"var(--card-padding)",display:"flex",flexDirection:"column",gap:"1.5"},body:{padding:"var(--card-padding)",flex:"1",display:"flex",flexDirection:"column"},footer:{display:"flex",alignItems:"center",gap:"2",paddingInline:"var(--card-padding)",paddingBottom:"var(--card-padding)"}},variants:{size:{sm:{root:{"--card-padding":"spacing.4"},title:{textStyle:"md"}},md:{root:{"--card-padding":"spacing.6"},title:{textStyle:"lg"}},lg:{root:{"--card-padding":"spacing.7"},title:{textStyle:"xl"}}},variant:{elevated:{root:{bg:"bg.panel",boxShadow:"md"}},outline:{root:{bg:"bg.panel",borderWidth:"1px",borderColor:"border"}},subtle:{root:{bg:"bg.muted"}}}},defaultVariants:{variant:"outline",size:"md"}}),_Jn=Bo({slots:xGn.keys(),className:"chakra-checkbox",base:{root:{display:"inline-flex",gap:"2",alignItems:"center",verticalAlign:"top",position:"relative"},control:Fb.base,label:{fontWeight:"medium",userSelect:"none",_disabled:{opacity:"0.5"}}},variants:{size:{xs:{root:{gap:"1.5"},label:{textStyle:"xs"},control:Fb.variants?.size?.xs},sm:{root:{gap:"2"},label:{textStyle:"sm"},control:Fb.variants?.size?.sm},md:{root:{gap:"2.5"},label:{textStyle:"sm"},control:Fb.variants?.size?.md},lg:{root:{gap:"3"},label:{textStyle:"md"},control:Fb.variants?.size?.lg}},variant:{outline:{control:Fb.variants?.variant?.outline},solid:{control:Fb.variants?.variant?.solid},subtle:{control:Fb.variants?.variant?.subtle}}},defaultVariants:{variant:"solid",size:"md"}}),kJn=Bo({slots:BXn.keys(),className:"chakra-checkbox-card",base:{root:{display:"flex",flexDirection:"column",userSelect:"none",position:"relative",borderRadius:"l2",flex:"1",focusVisibleRing:"outside",_disabled:{opacity:"0.8"},_invalid:{outline:"2px solid",outlineColor:"border.error"}},control:{display:"inline-flex",flex:"1",position:"relative",borderRadius:"inherit",justifyContent:"var(--checkbox-card-justify)",alignItems:"var(--checkbox-card-align)"},label:{fontWeight:"medium",display:"flex",alignItems:"center",gap:"2",flex:"1",_disabled:{opacity:"0.5"}},description:{opacity:"0.64",textStyle:"sm",_disabled:{opacity:"0.5"}},addon:{_disabled:{opacity:"0.5"}},indicator:Fb.base,content:{display:"flex",flexDirection:"column",flex:"1",gap:"1",justifyContent:"var(--checkbox-card-justify)",alignItems:"var(--checkbox-card-align)"}},variants:{size:{sm:{root:{textStyle:"sm"},control:{padding:"3",gap:"1.5"},addon:{px:"3",py:"1.5",borderTopWidth:"1px"},indicator:Fb.variants?.size.sm},md:{root:{textStyle:"sm"},control:{padding:"4",gap:"2.5"},addon:{px:"4",py:"2",borderTopWidth:"1px"},indicator:Fb.variants?.size.md},lg:{root:{textStyle:"md"},control:{padding:"4",gap:"3.5"},addon:{px:"4",py:"2",borderTopWidth:"1px"},indicator:Fb.variants?.size.lg}},variant:{surface:{root:{borderWidth:"1px",borderColor:"border",_checked:{bg:"colorPalette.subtle",color:"colorPalette.fg",borderColor:"colorPalette.muted"},_disabled:{bg:"bg.muted"}},indicator:Fb.variants?.variant.solid},subtle:{root:{bg:"bg.muted"},control:{_checked:{bg:"colorPalette.muted",color:"colorPalette.fg"}},indicator:Fb.variants?.variant.plain},outline:{root:{borderWidth:"1px",borderColor:"border",_checked:{boxShadow:"0 0 0 1px var(--shadow-color)",boxShadowColor:"colorPalette.solid",borderColor:"colorPalette.solid"}},indicator:Fb.variants?.variant.solid},solid:{root:{borderWidth:"1px",_checked:{bg:"colorPalette.solid",color:"colorPalette.contrast",borderColor:"colorPalette.solid"}},indicator:Fb.variants?.variant.inverted}},justify:{start:{root:{"--checkbox-card-justify":"flex-start"}},end:{root:{"--checkbox-card-justify":"flex-end"}},center:{root:{"--checkbox-card-justify":"center"}}},align:{start:{root:{"--checkbox-card-align":"flex-start"},content:{textAlign:"start"}},end:{root:{"--checkbox-card-align":"flex-end"},content:{textAlign:"end"}},center:{root:{"--checkbox-card-align":"center"},content:{textAlign:"center"}}},orientation:{vertical:{control:{flexDirection:"column"}},horizontal:{control:{flexDirection:"row"}}}},defaultVariants:{size:"md",variant:"outline",align:"start",orientation:"horizontal"}}),TJn=Bo({slots:pJn.keys(),className:"code-block",base:{root:{colorPalette:"gray",rounded:"var(--code-block-radius)",overflow:"hidden",bg:"bg",color:"fg",borderWidth:"1px","--code-block-max-height":"320px","--code-block-bg":"colors.bg","--code-block-fg":"colors.fg","--code-block-obscured-opacity":"0.5","--code-block-obscured-blur":"1px","--code-block-line-number-width":"sizes.3","--code-block-line-number-margin":"spacing.4","--code-block-highlight-bg":"{colors.teal.focusRing/20}","--code-block-highlight-border":"colors.teal.focusRing","--code-block-highlight-added-bg":"{colors.green.focusRing/20}","--code-block-highlight-added-border":"colors.green.focusRing","--code-block-highlight-removed-bg":"{colors.red.focusRing/20}","--code-block-highlight-removed-border":"colors.red.focusRing"},header:{display:"flex",alignItems:"center",gap:"2",position:"relative",px:"var(--code-block-padding)",minH:"var(--code-block-header-height)",mb:"calc(var(--code-block-padding) / 2 * -1)"},title:{display:"inline-flex",alignItems:"center",gap:"1.5",flex:"1",color:"fg.muted"},control:{gap:"1.5",display:"inline-flex",alignItems:"center"},footer:{display:"flex",alignItems:"center",justifyContent:"center",gap:"2",px:"var(--code-block-padding)",minH:"var(--code-block-header-height)"},content:{position:"relative",colorScheme:"dark",overflow:"hidden",borderBottomRadius:"var(--code-block-radius)",maxHeight:"var(--code-block-max-height)","& ::selection":{bg:"blue.500/40"},_expanded:{maxHeight:"unset"}},overlay:{"--bg":"{colors.black/50}",display:"flex",alignItems:"flex-end",justifyContent:"center",padding:"4",bgImage:"linear-gradient(0deg,var(--bg) 25%,transparent 100%)",color:"white",minH:"5rem",pos:"absolute",bottom:"0",insetInline:"0",zIndex:"1",fontWeight:"medium",_expanded:{display:"none"}},code:{fontFamily:"mono",lineHeight:"tall",whiteSpace:"pre",counterReset:"line 0"},codeText:{px:"var(--code-block-padding)",py:"var(--code-block-padding)",position:"relative",display:"block",width:"100%","&[data-has-focused]":{"& [data-line]:not([data-focused])":{transitionProperty:"opacity, filter",transitionDuration:"moderate",transitionTimingFunction:"ease-in-out",opacity:"var(--code-block-obscured-opacity)",filter:"blur(var(--code-block-obscured-blur))"},"&:hover":{"--code-block-obscured-opacity":"1","--code-block-obscured-blur":"0px"}},"&[data-has-line-numbers][data-plaintext]":{paddingInlineStart:"calc(var(--code-block-line-number-width) + var(--code-block-line-number-margin) + var(--code-block-padding))"},"& [data-line]":{position:"relative","--highlight-bg":"var(--code-block-highlight-bg)","--highlight-border":"var(--code-block-highlight-border)","&[data-highlight], &[data-diff]":{display:"inline-block",width:"full","&:after":{content:"''",display:"block",position:"absolute",insetStart:"calc(var(--code-block-padding) * -1)",insetEnd:"0px",width:"calc(100% + var(--code-block-padding) * 2)",height:"100%",bg:"var(--highlight-bg)",borderStartWidth:"2px",borderStartColor:"var(--highlight-border)"}},"&[data-diff='added']":{"--highlight-bg":"var(--code-block-highlight-added-bg)","--highlight-border":"var(--code-block-highlight-added-border)"},"&[data-diff='removed']":{"--highlight-bg":"var(--code-block-highlight-removed-bg)","--highlight-border":"var(--code-block-highlight-removed-border)"}},"&[data-word-wrap]":{"&[data-plaintext], & [data-line]":{whiteSpace:"pre-wrap",wordBreak:"break-all"}},"&[data-has-line-numbers]":{"--content":"counter(line)","& [data-line]:before":{content:"var(--content)",counterIncrement:"line",width:"var(--code-block-line-number-width)",marginRight:"var(--code-block-line-number-margin)",display:"inline-block",textAlign:"end",userSelect:"none",opacity:.4},"& [data-diff='added']:before":{content:"'+'"},"& [data-diff='removed']:before":{content:"'-'"}}}},variants:{size:{sm:{root:{"--code-block-padding":"spacing.4","--code-block-radius":"radii.md","--code-block-header-height":"sizes.8"},title:{textStyle:"xs"},code:{fontSize:"xs"}},md:{root:{"--code-block-padding":"spacing.4","--code-block-radius":"radii.lg","--code-block-header-height":"sizes.10"},title:{textStyle:"xs"},code:{fontSize:"sm"}},lg:{root:{"--code-block-padding":"spacing.5","--code-block-radius":"radii.xl","--code-block-header-height":"sizes.12"},title:{textStyle:"sm"},code:{fontSize:"sm"}}}},defaultVariants:{size:"md"}}),xJn=Bo({slots:xYt.keys(),className:"chakra-collapsible",base:{content:{overflow:"hidden",_open:{animationName:"expand-height, fade-in",animationDuration:"moderate"},_closed:{animationName:"collapse-height, fade-out",animationDuration:"moderate"}}}}),CJn=Bo({className:"colorPicker",slots:hJn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"1.5"},label:{color:"fg",fontWeight:"medium",textStyle:"sm",_disabled:{opacity:"0.5"}},valueText:{textAlign:"start"},control:{display:"flex",alignItems:"center",flexDirection:"row",gap:"2",position:"relative"},swatchTrigger:{display:"flex",alignItems:"center",justifyContent:"center"},trigger:{display:"flex",alignItems:"center",justifyContent:"center",flexDirection:"row",flexShrink:"0",gap:"2",textStyle:"sm",minH:"var(--input-height)",minW:"var(--input-height)",px:"1",rounded:"l2",_disabled:{opacity:"0.5"},"--focus-color":"colors.colorPalette.focusRing","&:focus-visible":{borderColor:"var(--focus-color)",outline:"1px solid var(--focus-color)"},"&[data-fit-content]":{"--input-height":"unset",px:"0",border:"0"}},content:{display:"flex",flexDirection:"column",bg:"bg.panel",borderRadius:"l3",boxShadow:"lg",width:"64",p:"4",gap:"3",zIndex:"dropdown",_open:{animationStyle:"slide-fade-in",animationDuration:"fast"},_closed:{animationStyle:"slide-fade-out",animationDuration:"faster"}},area:{height:"180px",borderRadius:"l2",overflow:"hidden"},areaThumb:{borderRadius:"full",height:"var(--thumb-size)",width:"var(--thumb-size)",borderWidth:"2px",borderColor:"white",shadow:"sm",focusVisibleRing:"mixed",focusRingColor:"white"},areaBackground:{height:"full"},channelSlider:{borderRadius:"l2",flex:"1"},channelSliderTrack:{height:"var(--slider-height)",borderRadius:"inherit",boxShadow:"inset 0 0 0 1px rgba(0,0,0,0.1)"},channelText:{textStyle:"xs",color:"fg.muted",fontWeight:"medium",textTransform:"capitalize"},swatchGroup:{display:"flex",flexDirection:"row",flexWrap:"wrap",gap:"2"},swatch:{...uen.base,borderRadius:"l1"},swatchIndicator:{color:"white",rounded:"full"},channelSliderThumb:{borderRadius:"full",height:"var(--thumb-size)",width:"var(--thumb-size)",borderWidth:"2px",borderColor:"white",shadow:"sm",transform:"translate(-50%, -50%)",focusVisibleRing:"outside",focusRingOffset:"1px"},channelInput:{...qp.base,"&::-webkit-inner-spin-button, &::-webkit-outer-spin-button":{WebkitAppearance:"none",margin:0}},formatSelect:{textStyle:"xs",textTransform:"uppercase",borderWidth:"1px",minH:"6",focusRing:"inside",rounded:"l2"},transparencyGrid:{borderRadius:"l2"},view:{display:"flex",flexDirection:"column",gap:"2"}},variants:{size:{"2xs":{channelInput:qp.variants?.size?.["2xs"],swatch:{"--swatch-size":"sizes.4.5"},trigger:{"--input-height":"sizes.7"},area:{"--thumb-size":"sizes.3"},channelSlider:{"--slider-height":"sizes.3","--thumb-size":"sizes.3"}},xs:{channelInput:qp.variants?.size?.xs,swatch:{"--swatch-size":"sizes.5"},trigger:{"--input-height":"sizes.8"},area:{"--thumb-size":"sizes.3.5"},channelSlider:{"--slider-height":"sizes.3.5","--thumb-size":"sizes.3.5"}},sm:{channelInput:qp.variants?.size?.sm,swatch:{"--swatch-size":"sizes.6"},trigger:{"--input-height":"sizes.9"},area:{"--thumb-size":"sizes.3.5"},channelSlider:{"--slider-height":"sizes.3.5","--thumb-size":"sizes.3.5"}},md:{channelInput:qp.variants?.size?.md,swatch:{"--swatch-size":"sizes.7"},trigger:{"--input-height":"sizes.10"},area:{"--thumb-size":"sizes.3.5"},channelSlider:{"--slider-height":"sizes.3.5","--thumb-size":"sizes.3.5"}},lg:{channelInput:qp.variants?.size?.lg,swatch:{"--swatch-size":"sizes.7"},trigger:{"--input-height":"sizes.11"},area:{"--thumb-size":"sizes.3.5"},channelSlider:{"--slider-height":"sizes.3.5","--thumb-size":"sizes.3.5"}},xl:{channelInput:qp.variants?.size?.xl,swatch:{"--swatch-size":"sizes.8"},trigger:{"--input-height":"sizes.12"},area:{"--thumb-size":"sizes.3.5"},channelSlider:{"--slider-height":"sizes.3.5","--thumb-size":"sizes.3.5"}},"2xl":{channelInput:qp.variants?.size?.["2xl"],swatch:{"--swatch-size":"sizes.10"},trigger:{"--input-height":"sizes.16"},area:{"--thumb-size":"sizes.3.5"},channelSlider:{"--slider-height":"sizes.3.5","--thumb-size":"sizes.3.5"}}},variant:{outline:{channelInput:qp.variants?.variant?.outline,trigger:{borderWidth:"1px"}},subtle:{channelInput:qp.variants?.variant?.subtle,trigger:{borderWidth:"1px",borderColor:"transparent",bg:"bg.muted"}}}},defaultVariants:{size:"md",variant:"outline"}}),AJn=Bo({className:"chakra-combobox",slots:nJn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"1.5",width:"full"},label:{fontWeight:"medium",userSelect:"none",textStyle:"sm",_disabled:{layerStyle:"disabled"}},input:{display:"flex",alignItems:"center",justifyContent:"space-between",background:"bg.panel",width:"full",minH:"var(--combobox-input-height)",px:"var(--combobox-input-padding-x)","--input-height":"var(--combobox-input-height)",borderRadius:"l2",outline:0,userSelect:"none",textAlign:"start",_placeholderShown:{color:"fg.muted"},_disabled:{layerStyle:"disabled"},"--focus-color":"colors.colorPalette.focusRing","--error-color":"colors.border.error",_invalid:{focusRingColor:"var(--error-color)",borderColor:"var(--error-color)"}},trigger:{display:"inline-flex",alignItems:"center",justifyContent:"center","--input-height":"var(--combobox-input-height)"},clearTrigger:{color:"fg.muted",pointerEvents:"auto",focusVisibleRing:"inside",focusRingWidth:"2px",rounded:"l1"},control:{pos:"relative"},indicatorGroup:{display:"flex",alignItems:"center",justifyContent:"center",gap:"1",pos:"absolute",insetEnd:"0",top:"0",bottom:"0",px:"var(--combobox-input-padding-x)",_icon:{boxSize:"var(--combobox-indicator-size)"},"[data-disabled] &":{opacity:.5}},content:{background:"bg.panel",display:"flex",flexDirection:"column",zIndex:"dropdown",borderRadius:"l2",outline:0,maxH:"96",overflowY:"auto",boxShadow:"md",_open:{animationStyle:"slide-fade-in",animationDuration:"fast"},_closed:{animationStyle:"slide-fade-out",animationDuration:"0s"},"&[data-empty]:not(:has([data-scope=combobox][data-part=empty]))":{opacity:0}},item:{position:"relative",userSelect:"none",display:"flex",alignItems:"center",gap:"2",py:"var(--combobox-item-padding-y)",px:"var(--combobox-item-padding-x)",cursor:"option",justifyContent:"space-between",flex:"1",textAlign:"start",borderRadius:"l1",_highlighted:{bg:"bg.emphasized/60"},_disabled:{pointerEvents:"none",opacity:"0.5"},_icon:{boxSize:"var(--combobox-indicator-size)"}},empty:{py:"var(--combobox-item-padding-y)",px:"var(--combobox-item-padding-x)"},itemText:{flex:"1"},itemGroup:{pb:"var(--combobox-item-padding-y)",_last:{pb:"0"}},itemGroupLabel:{fontWeight:"medium",py:"var(--combobox-item-padding-y)",px:"var(--combobox-item-padding-x)"}},variants:{variant:{outline:{input:{bg:"transparent",borderWidth:"1px",borderColor:"border",focusVisibleRing:"inside"}},subtle:{input:{borderWidth:"1px",borderColor:"transparent",bg:"bg.muted",focusVisibleRing:"inside"}},flushed:{input:{bg:"transparent",borderBottomWidth:"1px",borderBottomColor:"border",borderRadius:"0",px:"0",_focusVisible:{borderColor:"var(--focus-color)",boxShadow:"0px 1px 0px 0px var(--focus-color)"}},indicatorGroup:{px:"0"}}},size:{xs:{root:{"--combobox-input-height":"sizes.8","--combobox-input-padding-x":"spacing.2","--combobox-indicator-size":"sizes.3.5"},input:{textStyle:"xs"},content:{"--combobox-item-padding-x":"spacing.1.5","--combobox-item-padding-y":"spacing.1","--combobox-indicator-size":"sizes.3.5",p:"1",textStyle:"xs"},trigger:{textStyle:"xs",gap:"1"}},sm:{root:{"--combobox-input-height":"sizes.9","--combobox-input-padding-x":"spacing.2.5","--combobox-indicator-size":"sizes.4"},input:{textStyle:"sm"},content:{"--combobox-item-padding-x":"spacing.2","--combobox-item-padding-y":"spacing.1.5","--combobox-indicator-size":"sizes.4",p:"1",textStyle:"sm"},trigger:{textStyle:"sm",gap:"1"}},md:{root:{"--combobox-input-height":"sizes.10","--combobox-input-padding-x":"spacing.3","--combobox-indicator-size":"sizes.4"},input:{textStyle:"sm"},content:{"--combobox-item-padding-x":"spacing.2","--combobox-item-padding-y":"spacing.1.5","--combobox-indicator-size":"sizes.4",p:"1",textStyle:"sm"},itemIndicator:{display:"flex",alignItems:"center",justifyContent:"center"},trigger:{textStyle:"sm",gap:"2"}},lg:{root:{"--combobox-input-height":"sizes.12","--combobox-input-padding-x":"spacing.4","--combobox-indicator-size":"sizes.5"},input:{textStyle:"md"},content:{"--combobox-item-padding-y":"spacing.2","--combobox-item-padding-x":"spacing.3","--combobox-indicator-size":"sizes.5",p:"1.5",textStyle:"md"},trigger:{textStyle:"md",py:"3",gap:"2"}}}},defaultVariants:{size:"md",variant:"outline"}}),IJn=Bo({slots:UXn.keys(),className:"chakra-data-list",base:{itemLabel:{display:"flex",alignItems:"center",gap:"1"},itemValue:{display:"flex",minWidth:"0",flex:"1"}},variants:{orientation:{horizontal:{root:{display:"flex",flexDirection:"column"},item:{display:"inline-flex",alignItems:"center",gap:"4"},itemLabel:{minWidth:"120px"}},vertical:{root:{display:"flex",flexDirection:"column"},item:{display:"flex",flexDirection:"column",gap:"1"}}},size:{sm:{root:{gap:"3"},item:{textStyle:"xs"}},md:{root:{gap:"4"},item:{textStyle:"sm"}},lg:{root:{gap:"5"},item:{textStyle:"md"}}},variant:{subtle:{itemLabel:{color:"fg.muted"}},bold:{itemLabel:{fontWeight:"medium"},itemValue:{color:"fg.muted"}}}},defaultVariants:{size:"md",orientation:"vertical",variant:"subtle"}}),OJn=Bo({slots:zXn.keys(),className:"chakra-dialog",base:{backdrop:{bg:"blackAlpha.500",pos:"fixed",left:0,top:0,w:"100dvw",h:"100dvh",zIndex:"var(--z-index)",_open:{animationName:"fade-in",animationDuration:"slow"},_closed:{animationName:"fade-out",animationDuration:"moderate"}},positioner:{display:"flex",width:"100dvw",height:"100dvh",position:"fixed",left:0,top:0,"--dialog-z-index":"zIndex.modal",zIndex:"calc(var(--dialog-z-index) + var(--layer-index, 0))",justifyContent:"center",overscrollBehaviorY:"none"},content:{display:"flex",flexDirection:"column",position:"relative",width:"100%",outline:0,borderRadius:"l3",textStyle:"sm",my:"var(--dialog-margin, var(--dialog-base-margin))","--dialog-z-index":"zIndex.modal",zIndex:"calc(var(--dialog-z-index) + var(--layer-index, 0))",bg:"bg.panel",boxShadow:"lg",_open:{animationDuration:"moderate"},_closed:{animationDuration:"faster"}},header:{display:"flex",gap:"2",flex:0,px:"6",pt:"6",pb:"4"},body:{flex:"1",px:"6",pt:"2",pb:"6"},footer:{display:"flex",alignItems:"center",justifyContent:"flex-end",gap:"3",px:"6",pt:"2",pb:"4"},title:{textStyle:"lg",fontWeight:"semibold"},description:{color:"fg.muted"},closeTrigger:{pos:"absolute",top:"2",insetEnd:"2"}},variants:{placement:{center:{positioner:{alignItems:"center"},content:{"--dialog-base-margin":"auto",mx:"auto"}},top:{positioner:{alignItems:"flex-start"},content:{"--dialog-base-margin":"spacing.16",mx:"auto"}},bottom:{positioner:{alignItems:"flex-end"},content:{"--dialog-base-margin":"spacing.16",mx:"auto"}}},scrollBehavior:{inside:{positioner:{overflow:"hidden"},content:{maxH:"calc(100% - 7.5rem)"},body:{overflow:"auto"}},outside:{positioner:{overflow:"auto",pointerEvents:"auto"}}},size:{xs:{content:{maxW:"sm"}},sm:{content:{maxW:"md"}},md:{content:{maxW:"lg"}},lg:{content:{maxW:"2xl"}},xl:{content:{maxW:"4xl"}},cover:{positioner:{padding:"10"},content:{width:"100%",height:"100%","--dialog-margin":"0"}},full:{content:{maxW:"100dvw",minH:"100dvh","--dialog-margin":"0",borderRadius:"0"}}},motionPreset:{scale:{content:{_open:{animationName:"scale-in, fade-in"},_closed:{animationName:"scale-out, fade-out"}}},"slide-in-bottom":{content:{_open:{animationName:"slide-from-bottom, fade-in"},_closed:{animationName:"slide-to-bottom, fade-out"}}},"slide-in-top":{content:{_open:{animationName:"slide-from-top, fade-in"},_closed:{animationName:"slide-to-top, fade-out"}}},"slide-in-left":{content:{_open:{animationName:"slide-from-left, fade-in"},_closed:{animationName:"slide-to-left, fade-out"}}},"slide-in-right":{content:{_open:{animationName:"slide-from-right, fade-in"},_closed:{animationName:"slide-to-right, fade-out"}}},none:{}}},defaultVariants:{size:"md",scrollBehavior:"outside",placement:"top",motionPreset:"scale"}}),NJn=Bo({slots:HXn.keys(),className:"chakra-drawer",base:{backdrop:{bg:"blackAlpha.500",pos:"fixed",insetInlineStart:0,top:0,w:"100vw",h:"100dvh",zIndex:"overlay",_open:{animationName:"fade-in",animationDuration:"slow"},_closed:{animationName:"fade-out",animationDuration:"moderate"}},positioner:{display:"flex",width:"100vw",height:"100dvh",position:"fixed",insetInlineStart:0,top:0,zIndex:"modal",overscrollBehaviorY:"none"},content:{display:"flex",flexDirection:"column",position:"relative",width:"100%",outline:0,zIndex:"modal",textStyle:"sm",maxH:"100dvh",color:"inherit",bg:"bg.panel",boxShadow:"lg",_open:{animationDuration:"slowest",animationTimingFunction:"ease-in-smooth"},_closed:{animationDuration:"slower",animationTimingFunction:"ease-in-smooth"}},header:{display:"flex",alignItems:"center",gap:"2",flex:0,px:"6",pt:"6",pb:"4"},body:{px:"6",py:"2",flex:"1",overflow:"auto"},footer:{display:"flex",alignItems:"center",justifyContent:"flex-end",gap:"3",px:"6",pt:"2",pb:"4"},title:{flex:"1",textStyle:"lg",fontWeight:"semibold"},description:{color:"fg.muted"},closeTrigger:{pos:"absolute",top:"3",insetEnd:"2"}},variants:{size:{xs:{content:{maxW:"xs"}},sm:{content:{maxW:"md"}},md:{content:{maxW:"lg"}},lg:{content:{maxW:"2xl"}},xl:{content:{maxW:"4xl"}},full:{content:{maxW:"100vw",h:"100dvh"}}},placement:{start:{positioner:{justifyContent:"flex-start",alignItems:"stretch"},content:{_open:{animationName:{base:"slide-from-left-full, fade-in",_rtl:"slide-from-right-full, fade-in"}},_closed:{animationName:{base:"slide-to-left-full, fade-out",_rtl:"slide-to-right-full, fade-out"}}}},end:{positioner:{justifyContent:"flex-end",alignItems:"stretch"},content:{_open:{animationName:{base:"slide-from-right-full, fade-in",_rtl:"slide-from-left-full, fade-in"}},_closed:{animationName:{base:"slide-to-right-full, fade-out",_rtl:"slide-to-left-full, fade-out"}}}},top:{positioner:{justifyContent:"stretch",alignItems:"flex-start"},content:{maxW:"100%",_open:{animationName:"slide-from-top-full, fade-in"},_closed:{animationName:"slide-to-top-full, fade-out"}}},bottom:{positioner:{justifyContent:"stretch",alignItems:"flex-end"},content:{maxW:"100%",_open:{animationName:"slide-from-bottom-full, fade-in"},_closed:{animationName:"slide-to-bottom-full, fade-out"}}}},contained:{true:{positioner:{padding:"4"},content:{borderRadius:"l3"}}}},defaultVariants:{size:"xs",placement:"end"}}),E7t=_K({fontSize:"inherit",fontWeight:"inherit",textAlign:"inherit",bg:"transparent",borderRadius:"l2"}),RJn=Bo({slots:GXn.keys(),className:"chakra-editable",base:{root:{display:"inline-flex",alignItems:"center",position:"relative",gap:"1.5",width:"full"},preview:{...E7t,py:"1",px:"1",display:"inline-flex",alignItems:"center",transitionProperty:"common",transitionDuration:"moderate",cursor:"text",_hover:{bg:"bg.muted"},_disabled:{userSelect:"none"}},input:{...E7t,outline:"0",py:"1",px:"1",transitionProperty:"common",transitionDuration:"normal",width:"full",focusVisibleRing:"inside",focusRingWidth:"2px",_placeholder:{opacity:.6}},control:{display:"inline-flex",alignItems:"center",gap:"1.5"}},variants:{size:{sm:{root:{textStyle:"sm"},preview:{minH:"8"},input:{minH:"8"}},md:{root:{textStyle:"sm"},preview:{minH:"9"},input:{minH:"9"}},lg:{root:{textStyle:"md"},preview:{minH:"10"},input:{minH:"10"}}}},defaultVariants:{size:"md"}}),PJn=Bo({slots:qXn.keys(),className:"chakra-empty-state",base:{root:{width:"full"},content:{display:"flex",flexDirection:"column",alignItems:"center",justifyContent:"center"},indicator:{display:"flex",alignItems:"center",justifyContent:"center",color:"fg.subtle",_icon:{boxSize:"1em"}},title:{fontWeight:"semibold"},description:{textStyle:"sm",color:"fg.muted"}},variants:{size:{sm:{root:{px:"4",py:"6"},title:{textStyle:"md"},content:{gap:"4"},indicator:{textStyle:"2xl"}},md:{root:{px:"8",py:"12"},title:{textStyle:"lg"},content:{gap:"6"},indicator:{textStyle:"4xl"}},lg:{root:{px:"12",py:"16"},title:{textStyle:"xl"},content:{gap:"8"},indicator:{textStyle:"6xl"}}}},defaultVariants:{size:"md"}}),LJn=Bo({className:"chakra-field",slots:VXn.keys(),base:{requiredIndicator:{color:"fg.error",lineHeight:"1"},root:{display:"flex",width:"100%",position:"relative",gap:"1.5"},label:{display:"flex",alignItems:"center",textAlign:"start",textStyle:"sm",fontWeight:"medium",gap:"1",userSelect:"none",_disabled:{opacity:"0.5"}},errorText:{display:"inline-flex",alignItems:"center",fontWeight:"medium",gap:"1",color:"fg.error",textStyle:"xs"},helperText:{color:"fg.muted",textStyle:"xs"}},variants:{orientation:{vertical:{root:{flexDirection:"column",alignItems:"flex-start"}},horizontal:{root:{flexDirection:"row",alignItems:"center",justifyContent:"space-between"},label:{flex:"0 0 var(--field-label-width, 80px)"}}}},defaultVariants:{orientation:"vertical"}}),DJn=Bo({className:"fieldset",slots:WXn.keys(),base:{root:{display:"flex",flexDirection:"column",width:"full"},content:{display:"flex",flexDirection:"column",width:"full"},legend:{color:"fg",fontWeight:"medium",_disabled:{opacity:"0.5"}},helperText:{color:"fg.muted",textStyle:"sm"},errorText:{display:"inline-flex",alignItems:"center",color:"fg.error",gap:"2",fontWeight:"medium",textStyle:"sm"}},variants:{size:{sm:{root:{spaceY:"2"},content:{gap:"1.5"},legend:{textStyle:"sm"}},md:{root:{spaceY:"4"},content:{gap:"4"},legend:{textStyle:"sm"}},lg:{root:{spaceY:"6"},content:{gap:"4"},legend:{textStyle:"md"}}}},defaultVariants:{size:"md"}}),MJn=Bo({className:"chakra-file-upload",slots:KXn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"4",width:"100%",alignItems:"flex-start"},label:{fontWeight:"medium",textStyle:"sm"},dropzone:{background:"bg",borderRadius:"l3",borderWidth:"2px",borderStyle:"dashed",display:"flex",alignItems:"center",flexDirection:"column",gap:"4",justifyContent:"center",minHeight:"2xs",px:"3",py:"2",transition:"backgrounds",focusVisibleRing:"outside",_hover:{bg:"bg.subtle"},_dragging:{bg:"colorPalette.subtle",borderStyle:"solid",borderColor:"colorPalette.solid"}},dropzoneContent:{display:"flex",flexDirection:"column",alignItems:"center",textAlign:"center",gap:"1",textStyle:"sm"},item:{pos:"relative",textStyle:"sm",animationName:"fade-in",animationDuration:"moderate",background:"bg",borderRadius:"l2",borderWidth:"1px",width:"100%",display:"flex",alignItems:"center",gap:"3",p:"4"},itemGroup:{width:"100%",display:"flex",flexDirection:"column",gap:"3",_empty:{display:"none"}},itemName:{color:"fg",fontWeight:"medium",lineClamp:"1"},itemContent:{display:"flex",flexDirection:"column",gap:"0.5",flex:"1"},itemSizeText:{color:"fg.muted",textStyle:"xs"},itemDeleteTrigger:{display:"flex",alignItems:"center",justifyContent:"center",alignSelf:"flex-start",boxSize:"5",p:"2px",color:"fg.muted",cursor:"button"},itemPreview:{color:"fg.muted",_icon:{boxSize:"4.5"}}},defaultVariants:{}}),jJn=Bo({className:"chakra-hover-card",slots:uJt.keys(),base:{content:{position:"relative",display:"flex",flexDirection:"column",textStyle:"sm","--hovercard-bg":"colors.bg.panel",bg:"var(--hovercard-bg)",boxShadow:"lg",maxWidth:"80",borderRadius:"l3",zIndex:"popover",transformOrigin:"var(--transform-origin)",outline:"0",_open:{animationStyle:"slide-fade-in",animationDuration:"fast"},_closed:{animationStyle:"slide-fade-out",animationDuration:"faster"}},arrow:{"--arrow-size":"sizes.3","--arrow-background":"var(--hovercard-bg)"},arrowTip:{borderTopWidth:"0.5px",borderInlineStartWidth:"0.5px"}},variants:{size:{xs:{content:{padding:"3"}},sm:{content:{padding:"4"}},md:{content:{padding:"5"}},lg:{content:{padding:"6"}}}},defaultVariants:{size:"md"}}),$Jn=Bo({className:"chakra-list",slots:YXn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"var(--list-gap)","& :where(ul, ol)":{marginTop:"var(--list-gap)"}},item:{whiteSpace:"normal",display:"list-item"},indicator:{marginEnd:"2",minHeight:"1lh",flexShrink:0,display:"inline-block",verticalAlign:"middle"}},variants:{variant:{marker:{root:{listStyle:"revert"},item:{_marker:{color:"fg.subtle"}}},plain:{item:{alignItems:"flex-start",display:"inline-flex"}}},align:{center:{item:{alignItems:"center"}},start:{item:{alignItems:"flex-start"}},end:{item:{alignItems:"flex-end"}}}},defaultVariants:{variant:"marker"}}),FJn=Bo({className:"chakra-listbox",slots:gJn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"1.5",width:"full"},content:{display:"flex",maxH:"96",p:"1",gap:"1",textStyle:"sm",outline:"none",scrollPadding:"1",_horizontal:{flexDirection:"row",overflowX:"auto"},_vertical:{flexDirection:"column",overflowY:"auto"},"--listbox-item-padding-x":"spacing.2","--listbox-item-padding-y":"spacing.1.5"},item:{position:"relative",userSelect:"none",display:"flex",alignItems:"center",gap:"2",cursor:"pointer",justifyContent:"space-between",flex:"1",textAlign:"start",borderRadius:"l1",py:"var(--listbox-item-padding-y)",px:"var(--listbox-item-padding-x)",_highlighted:{outline:"2px solid",outlineColor:"border.emphasized"},_disabled:{pointerEvents:"none",opacity:"0.5"}},empty:{py:"var(--listbox-item-padding-y)",px:"var(--listbox-item-padding-x)"},itemText:{flex:"1"},itemGroup:{mt:"1.5",_first:{mt:"0"}},itemGroupLabel:{py:"1.5",px:"2",fontWeight:"medium"},label:{fontWeight:"medium",userSelect:"none",textStyle:"sm",_disabled:{layerStyle:"disabled"}},valueText:{lineClamp:"1",maxW:"80%"},itemIndicator:{display:"flex",alignItems:"center",justifyContent:"center",_icon:{boxSize:"4"}}},variants:{variant:{subtle:{content:{bg:"bg.panel",borderWidth:"1px",borderRadius:"l2"},item:{_hover:{bg:"bg.emphasized/60"},_selected:{bg:"bg.muted"}}},solid:{content:{bg:"bg.panel",borderWidth:"1px",borderRadius:"l2"},item:{_selected:{bg:"colorPalette.solid",color:"colorPalette.contrast"}}},plain:{}}},defaultVariants:{variant:"subtle"}}),BJn=Bo({className:"chakra-menu",slots:XXn.keys(),base:{content:{outline:0,bg:"bg.panel",boxShadow:"lg",color:"fg",maxHeight:"var(--available-height)","--menu-z-index":"zIndex.dropdown",zIndex:"calc(var(--menu-z-index) + var(--layer-index, 0))",borderRadius:"l2",overflow:"hidden",overflowY:"auto",_open:{animationStyle:"slide-fade-in",animationDuration:"fast"},_closed:{animationStyle:"slide-fade-out",animationDuration:"faster"}},item:{textDecoration:"none",color:"fg",userSelect:"none",borderRadius:"l1",width:"100%",display:"flex",cursor:"menuitem",alignItems:"center",textAlign:"start",position:"relative",flex:"0 0 auto",outline:0,_disabled:{layerStyle:"disabled"},"&[data-type]":{ps:"8"}},itemText:{flex:"1"},itemIndicator:{position:"absolute",insetStart:"2",transform:"translateY(-50%)",top:"50%"},itemGroupLabel:{px:"2",py:"1.5",fontWeight:"semibold",textStyle:"sm"},indicator:{display:"inline-flex",alignItems:"center",justifyContent:"center",flexShrink:"0"},itemCommand:{opacity:"0.6",textStyle:"xs",ms:"auto",ps:"4",letterSpacing:"widest",fontFamily:"inherit"},separator:{height:"1px",bg:"bg.muted",my:"1",mx:"-1"}},variants:{variant:{subtle:{item:{_highlighted:{bg:"bg.emphasized/60"}}},solid:{item:{_highlighted:{bg:"colorPalette.solid",color:"colorPalette.contrast"}}}},size:{sm:{content:{minW:"8rem",padding:"1",scrollPadding:"1"},item:{gap:"1",textStyle:"xs",py:"1",px:"1.5"}},md:{content:{minW:"8rem",padding:"1.5",scrollPadding:"1.5"},item:{gap:"2",textStyle:"sm",py:"1.5",px:"2"}}}},defaultVariants:{size:"md",variant:"subtle"}}),cve=Bo({className:"chakra-select",slots:tJn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"1.5",width:"full"},trigger:{display:"flex",alignItems:"center",justifyContent:"space-between",width:"full",minH:"var(--select-trigger-height)","--input-height":"var(--select-trigger-height)",px:"var(--select-trigger-padding-x)",borderRadius:"l2",userSelect:"none",textAlign:"start",focusVisibleRing:"inside",_placeholderShown:{color:"fg.muted/80"},_disabled:{layerStyle:"disabled"},_invalid:{borderColor:"border.error"}},indicatorGroup:{display:"flex",alignItems:"center",gap:"1",pos:"absolute",insetEnd:"0",top:"0",bottom:"0",px:"var(--select-trigger-padding-x)",pointerEvents:"none"},indicator:{display:"flex",alignItems:"center",justifyContent:"center",color:{base:"fg.muted",_disabled:"fg.subtle",_invalid:"fg.error"}},content:{background:"bg.panel",display:"flex",flexDirection:"column",zIndex:"dropdown",borderRadius:"l2",outline:0,maxH:"96",overflowY:"auto",boxShadow:"md",_open:{animationStyle:"slide-fade-in",animationDuration:"fast"},_closed:{animationStyle:"slide-fade-out",animationDuration:"fastest"}},item:{position:"relative",userSelect:"none",display:"flex",alignItems:"center",gap:"2",cursor:"option",justifyContent:"space-between",flex:"1",textAlign:"start",borderRadius:"l1",_highlighted:{bg:"bg.emphasized/60"},_disabled:{pointerEvents:"none",opacity:"0.5"},_icon:{width:"4",height:"4"}},control:{pos:"relative"},itemText:{flex:"1"},itemGroup:{_first:{mt:"0"}},itemGroupLabel:{py:"1",fontWeight:"medium"},label:{fontWeight:"medium",userSelect:"none",textStyle:"sm",_disabled:{layerStyle:"disabled"}},valueText:{lineClamp:"1",maxW:"80%"},clearTrigger:{color:"fg.muted",pointerEvents:"auto",focusVisibleRing:"inside",focusRingWidth:"2px",rounded:"l1"}},variants:{variant:{outline:{trigger:{bg:"transparent",borderWidth:"1px",borderColor:"border",_expanded:{borderColor:"border.emphasized"}}},subtle:{trigger:{borderWidth:"1px",borderColor:"transparent",bg:"bg.muted"}}},size:{xs:{root:{"--select-trigger-height":"sizes.8","--select-trigger-padding-x":"spacing.2"},content:{p:"1",gap:"1",textStyle:"xs"},trigger:{textStyle:"xs",gap:"1"},item:{py:"1",px:"2"},itemGroupLabel:{py:"1",px:"2"},indicator:{_icon:{width:"3.5",height:"3.5"}}},sm:{root:{"--select-trigger-height":"sizes.9","--select-trigger-padding-x":"spacing.2.5"},content:{p:"1",textStyle:"sm"},trigger:{textStyle:"sm",gap:"1"},indicator:{_icon:{width:"4",height:"4"}},item:{py:"1",px:"1.5"},itemGroup:{mt:"1"},itemGroupLabel:{py:"1",px:"1.5"}},md:{root:{"--select-trigger-height":"sizes.10","--select-trigger-padding-x":"spacing.3"},content:{p:"1",textStyle:"sm"},itemGroup:{mt:"1.5"},item:{py:"1.5",px:"2"},itemIndicator:{display:"flex",alignItems:"center",justifyContent:"center"},itemGroupLabel:{py:"1.5",px:"2"},trigger:{textStyle:"sm",gap:"2"},indicator:{_icon:{width:"4",height:"4"}}},lg:{root:{"--select-trigger-height":"sizes.12","--select-trigger-padding-x":"spacing.4"},content:{p:"1.5",textStyle:"md"},itemGroup:{mt:"2"},item:{py:"2",px:"3"},itemGroupLabel:{py:"2",px:"3"},trigger:{textStyle:"md",py:"3",gap:"2"},indicator:{_icon:{width:"5",height:"5"}}}}},defaultVariants:{size:"md",variant:"outline"}}),UJn=Bo({className:"chakra-native-select",slots:JXn.keys(),base:{root:{height:"fit-content",display:"flex",width:"100%",position:"relative"},field:{width:"100%",minWidth:"0",outline:"0",appearance:"none",borderRadius:"l2","--error-color":"colors.border.error","--input-height":"var(--select-field-height)",height:"var(--select-field-height)",_disabled:{layerStyle:"disabled"},_invalid:{focusRingColor:"var(--error-color)",borderColor:"var(--error-color)"},focusVisibleRing:"inside",lineHeight:"normal","& > option, & > optgroup":{bg:"bg"}},indicator:{position:"absolute",display:"inline-flex",alignItems:"center",justifyContent:"center",pointerEvents:"none",top:"50%",transform:"translateY(-50%)",height:"100%",color:"fg.muted",_disabled:{opacity:"0.5"},_invalid:{color:"fg.error"},_icon:{width:"1em",height:"1em"}}},variants:{variant:{outline:{field:cve.variants?.variant.outline.trigger},subtle:{field:cve.variants?.variant.subtle.trigger},plain:{field:{bg:"transparent",color:"fg",focusRingWidth:"2px"}}},size:{xs:{root:{"--select-field-height":"sizes.8"},field:{textStyle:"xs",ps:"2",pe:"6"},indicator:{textStyle:"sm",insetEnd:"1.5"}},sm:{root:{"--select-field-height":"sizes.9"},field:{textStyle:"sm",ps:"2.5",pe:"8"},indicator:{textStyle:"md",insetEnd:"2"}},md:{root:{"--select-field-height":"sizes.10"},field:{textStyle:"sm",ps:"3",pe:"8"},indicator:{textStyle:"lg",insetEnd:"2"}},lg:{root:{"--select-field-height":"sizes.11"},field:{textStyle:"md",ps:"4",pe:"8"},indicator:{textStyle:"xl",insetEnd:"3"}},xl:{root:{"--select-field-height":"sizes.12"},field:{textStyle:"md",ps:"4.5",pe:"10"},indicator:{textStyle:"xl",insetEnd:"3"}}}},defaultVariants:cve.defaultVariants});function uVe(n,r){const o={};for(const s in n){const c=r(s,n[s]);o[c[0]]=c[1]}return o}const S7t=_K({display:"flex",justifyContent:"center",alignItems:"center",flex:"1",userSelect:"none",cursor:"button",lineHeight:"1",color:"fg.muted","--stepper-base-radius":"radii.l1","--stepper-radius":"calc(var(--stepper-base-radius) + 1px)",_icon:{boxSize:"1em"},_disabled:{opacity:"0.5"},_hover:{bg:"bg.muted"},_active:{bg:"bg.emphasized"}}),zJn=Bo({className:"chakra-number-input",slots:rZt.keys(),base:{root:{position:"relative",zIndex:"0",isolation:"isolate"},input:{...qp.base,verticalAlign:"top",pe:"calc(var(--stepper-width) + 0.5rem)"},control:{display:"flex",flexDirection:"column",position:"absolute",top:"0",insetEnd:"0px",margin:"1px",width:"var(--stepper-width)",height:"calc(100% - 2px)",zIndex:"1",borderStartWidth:"1px",divideY:"1px"},incrementTrigger:{...S7t,borderTopEndRadius:"var(--stepper-radius)"},decrementTrigger:{...S7t,borderBottomEndRadius:"var(--stepper-radius)"},valueText:{fontWeight:"medium",fontFeatureSettings:"pnum",fontVariantNumeric:"proportional-nums"}},variants:{size:{xs:{input:qp.variants.size.xs,control:{fontSize:"2xs","--stepper-width":"sizes.4"}},sm:{input:qp.variants.size.sm,control:{fontSize:"xs","--stepper-width":"sizes.5"}},md:{input:qp.variants.size.md,control:{fontSize:"sm","--stepper-width":"sizes.6"}},lg:{input:qp.variants.size.lg,control:{fontSize:"sm","--stepper-width":"sizes.6"}}},variant:uVe(qp.variants.variant,(n,r)=>[n,{input:r}])},defaultVariants:{size:"md",variant:"outline"}}),{variants:_7t,defaultVariants:HJn}=qp,GJn=Bo({className:"chakra-pin-input",slots:aZt.keys(),base:{input:{...qp.base,textAlign:"center",width:"var(--input-height)"},control:{display:"inline-flex",gap:"2",isolation:"isolate"}},variants:{size:uVe(_7t.size,(n,r)=>[n,{input:{...r,px:"1"}}]),variant:uVe(_7t.variant,(n,r)=>[n,{input:r}]),attached:{true:{control:{gap:"0",spaceX:"-1px"},input:{_notFirst:{borderStartRadius:"0"},_notLast:{borderEndRadius:"0"},_focusVisible:{zIndex:"1"}}}}},defaultVariants:HJn}),qJn=Bo({className:"chakra-popover",slots:ZXn.keys(),base:{content:{position:"relative",display:"flex",flexDirection:"column",textStyle:"sm","--popover-bg":"colors.bg.panel",bg:"var(--popover-bg)",boxShadow:"lg","--popover-size":"sizes.xs","--popover-mobile-size":"calc(100dvw - 1rem)",width:{base:"min(var(--popover-mobile-size), var(--popover-size))",sm:"var(--popover-size)"},borderRadius:"l3","--popover-z-index":"zIndex.popover",zIndex:"calc(var(--popover-z-index) + var(--layer-index, 0))",outline:"0",transformOrigin:"var(--transform-origin)",maxHeight:"var(--available-height)",_open:{animationStyle:"scale-fade-in",animationDuration:"fast"},_closed:{animationStyle:"scale-fade-out",animationDuration:"faster"}},header:{paddingInline:"var(--popover-padding)",paddingTop:"var(--popover-padding)"},body:{padding:"var(--popover-padding)",flex:"1"},footer:{display:"flex",alignItems:"center",paddingInline:"var(--popover-padding)",paddingBottom:"var(--popover-padding)"},arrow:{"--arrow-size":"sizes.3","--arrow-background":"var(--popover-bg)"},arrowTip:{borderTopWidth:"1px",borderInlineStartWidth:"1px"}},variants:{size:{xs:{content:{"--popover-padding":"spacing.3"}},sm:{content:{"--popover-padding":"spacing.4"}},md:{content:{"--popover-padding":"spacing.5"}},lg:{content:{"--popover-padding":"spacing.6"}}}},defaultVariants:{size:"md"}}),VJn=Bo({slots:fYe.keys(),className:"chakra-progress",base:{root:{textStyle:"sm",position:"relative"},track:{overflow:"hidden",position:"relative"},range:{display:"flex",alignItems:"center",justifyContent:"center",transitionProperty:"width, height",transitionDuration:"slow",height:"100%",bgColor:"var(--track-color)",_indeterminate:{"--animate-from-x":"-40%","--animate-to-x":"100%",position:"absolute",willChange:"left",minWidth:"50%",animation:"position 1s ease infinite normal none running",backgroundImage:"linear-gradient(to right, transparent 0%, var(--track-color) 50%, transparent 100%)"}},label:{display:"inline-flex",fontWeight:"medium",alignItems:"center",gap:"1"},valueText:{textStyle:"xs",lineHeight:"1",fontWeight:"medium"}},variants:{variant:{outline:{track:{shadow:"inset",bgColor:"bg.muted"},range:{bgColor:"colorPalette.solid"}},subtle:{track:{bgColor:"colorPalette.muted"},range:{bgColor:"colorPalette.solid/72"}}},shape:{square:{},rounded:{track:{borderRadius:"l1"}},full:{track:{borderRadius:"full"}}},striped:{true:{range:{backgroundImage:"linear-gradient(45deg, var(--stripe-color) 25%, transparent 25%, transparent 50%, var(--stripe-color) 50%, var(--stripe-color) 75%, transparent 75%, transparent)",backgroundSize:"var(--stripe-size) var(--stripe-size)","--stripe-size":"1rem","--stripe-color":{_light:"rgba(255, 255, 255, 0.3)",_dark:"rgba(0, 0, 0, 0.3)"}}}},animated:{true:{range:{"--animate-from":"var(--stripe-size)",animation:"bg-position 1s linear infinite"}}},size:{xs:{track:{h:"1.5"}},sm:{track:{h:"2"}},md:{track:{h:"2.5"}},lg:{track:{h:"3"}},xl:{track:{h:"4"}}}},defaultVariants:{variant:"outline",size:"md",shape:"rounded"}}),WJn=Bo({className:"chakra-progress-circle",slots:fYe.keys(),base:{root:{display:"inline-flex",textStyle:"sm",position:"relative"},circle:{_indeterminate:{animation:"spin 2s linear infinite"}},circleTrack:{"--track-color":"colors.colorPalette.muted",stroke:"var(--track-color)"},circleRange:{stroke:"colorPalette.solid",transitionProperty:"stroke-dashoffset, stroke-dasharray",transitionDuration:"0.6s",_indeterminate:{animation:"circular-progress 1.5s linear infinite"}},label:{display:"inline-flex"},valueText:{lineHeight:"1",fontWeight:"medium",letterSpacing:"tight",fontVariantNumeric:"tabular-nums"}},variants:{size:{xs:{circle:{"--size":"24px","--thickness":"4px"},valueText:{textStyle:"2xs"}},sm:{circle:{"--size":"32px","--thickness":"5px"},valueText:{textStyle:"2xs"}},md:{circle:{"--size":"40px","--thickness":"6px"},valueText:{textStyle:"xs"}},lg:{circle:{"--size":"48px","--thickness":"7px"},valueText:{textStyle:"sm"}},xl:{circle:{"--size":"64px","--thickness":"8px"},valueText:{textStyle:"sm"}}}},defaultVariants:{size:"md"}}),KJn=Bo({slots:SZt.keys(),className:"chakra-qr-code",base:{root:{position:"relative",width:"fit-content","--qr-code-overlay-size":"calc(var(--qr-code-size) / 3)"},frame:{width:"var(--qr-code-size)",height:"var(--qr-code-size)",fill:"currentColor"},overlay:{display:"flex",alignItems:"center",justifyContent:"center",width:"var(--qr-code-overlay-size)",height:"var(--qr-code-overlay-size)",padding:"1",bg:"bg",rounded:"l1"}},variants:{size:{"2xs":{root:{"--qr-code-size":"40px"}},xs:{root:{"--qr-code-size":"64px"}},sm:{root:{"--qr-code-size":"80px"}},md:{root:{"--qr-code-size":"120px"}},lg:{root:{"--qr-code-size":"160px"}},xl:{root:{"--qr-code-size":"200px"}},"2xl":{root:{"--qr-code-size":"240px"}},full:{root:{"--qr-code-size":"100%"}}}},defaultVariants:{size:"md"}}),YJn=Bo({className:"chakra-radio-card",slots:QXn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"1.5",isolation:"isolate"},item:{flex:"1",display:"flex",flexDirection:"column",userSelect:"none",position:"relative",borderRadius:"l2",_focus:{bg:"colorPalette.muted/20"},_disabled:{opacity:"0.8",borderColor:"border.disabled"},_checked:{zIndex:"1"}},label:{display:"inline-flex",fontWeight:"medium",textStyle:"sm",_disabled:{opacity:"0.5"}},itemText:{fontWeight:"medium",flex:"1"},itemDescription:{opacity:"0.64",textStyle:"sm"},itemControl:{display:"inline-flex",flex:"1",pos:"relative",rounded:"inherit",justifyContent:"var(--radio-card-justify)",alignItems:"var(--radio-card-align)",_disabled:{bg:"bg.muted"}},itemIndicator:Bb.base,itemAddon:{roundedBottom:"inherit",_disabled:{color:"fg.muted"}},itemContent:{display:"flex",flexDirection:"column",flex:"1",gap:"1",justifyContent:"var(--radio-card-justify)",alignItems:"var(--radio-card-align)"}},variants:{size:{sm:{item:{textStyle:"sm"},itemControl:{padding:"3",gap:"1.5"},itemAddon:{px:"3",py:"1.5",borderTopWidth:"1px"},itemIndicator:Bb.variants?.size.sm},md:{item:{textStyle:"sm"},itemControl:{padding:"4",gap:"2.5"},itemAddon:{px:"4",py:"2",borderTopWidth:"1px"},itemIndicator:Bb.variants?.size.md},lg:{item:{textStyle:"md"},itemControl:{padding:"4",gap:"3.5"},itemAddon:{px:"4",py:"2",borderTopWidth:"1px"},itemIndicator:Bb.variants?.size.lg}},variant:{surface:{item:{borderWidth:"1px",_checked:{bg:"colorPalette.subtle",color:"colorPalette.fg",borderColor:"colorPalette.muted"}},itemIndicator:Bb.variants?.variant.solid},subtle:{item:{bg:"bg.muted"},itemControl:{_checked:{bg:"colorPalette.muted",color:"colorPalette.fg"}},itemIndicator:Bb.variants?.variant.outline},outline:{item:{borderWidth:"1px",_checked:{boxShadow:"0 0 0 1px var(--shadow-color)",boxShadowColor:"colorPalette.solid",borderColor:"colorPalette.solid"}},itemIndicator:Bb.variants?.variant.solid},solid:{item:{borderWidth:"1px",_checked:{bg:"colorPalette.solid",color:"colorPalette.contrast",borderColor:"colorPalette.solid"}},itemIndicator:Bb.variants?.variant.inverted}},justify:{start:{item:{"--radio-card-justify":"flex-start"}},end:{item:{"--radio-card-justify":"flex-end"}},center:{item:{"--radio-card-justify":"center"}}},align:{start:{item:{"--radio-card-align":"flex-start"},itemControl:{textAlign:"start"}},end:{item:{"--radio-card-align":"flex-end"},itemControl:{textAlign:"end"}},center:{item:{"--radio-card-align":"center"},itemControl:{textAlign:"center"}}},orientation:{vertical:{itemControl:{flexDirection:"column"}},horizontal:{itemControl:{flexDirection:"row"}}}},defaultVariants:{size:"md",variant:"outline",align:"start",orientation:"horizontal"}}),XJn=Bo({className:"chakra-radio-group",slots:len.keys(),base:{item:{display:"inline-flex",alignItems:"center",position:"relative",fontWeight:"medium",_disabled:{cursor:"disabled"}},itemControl:Bb.base,label:{userSelect:"none",textStyle:"sm",_disabled:{opacity:"0.5"}}},variants:{variant:{outline:{itemControl:Bb.variants?.variant?.outline},subtle:{itemControl:Bb.variants?.variant?.subtle},solid:{itemControl:Bb.variants?.variant?.solid}},size:{xs:{item:{textStyle:"xs",gap:"1.5"},itemControl:Bb.variants?.size?.xs},sm:{item:{textStyle:"sm",gap:"2"},itemControl:Bb.variants?.size?.sm},md:{item:{textStyle:"sm",gap:"2.5"},itemControl:Bb.variants?.size?.md},lg:{item:{textStyle:"md",gap:"3"},itemControl:Bb.variants?.size?.lg}}},defaultVariants:{size:"md",variant:"solid"}}),JJn=Bo({className:"chakra-rating-group",slots:eJn.keys(),base:{root:{display:"inline-flex"},control:{display:"inline-flex",alignItems:"center"},item:{display:"inline-flex",alignItems:"center",justifyContent:"center",userSelect:"none"},itemIndicator:{display:"inline-flex",alignItems:"center",justifyContent:"center",width:"1em",height:"1em",position:"relative","--clip-path":{base:"inset(0 50% 0 0)",_rtl:"inset(0 0 0 50%)"},_icon:{stroke:"currentColor",width:"100%",height:"100%",display:"inline-block",flexShrink:0,position:"absolute",left:0,top:0},"& [data-bg]":{color:"bg.emphasized"},"& [data-fg]":{color:"transparent"},"&[data-highlighted]:not([data-half])":{"& [data-fg]":{color:"colorPalette.solid"}},"&[data-half]":{"& [data-fg]":{color:"colorPalette.solid",clipPath:"var(--clip-path)"}}}},variants:{size:{xs:{item:{textStyle:"sm"}},sm:{item:{textStyle:"md"}},md:{item:{textStyle:"xl"}},lg:{item:{textStyle:"2xl"}}}},defaultVariants:{size:"md"}}),ZJn=Bo({className:"chakra-scroll-area",slots:$Zt.keys(),base:{root:{display:"flex",flexDirection:"column",width:"100%",height:"100%",position:"relative",overflow:"hidden","--scrollbar-margin":"2px","--scrollbar-click-area":"calc(var(--scrollbar-size) + calc(var(--scrollbar-margin) * 2))"},viewport:{display:"flex",flexDirection:"column",height:"100%",width:"100%",borderRadius:"inherit",WebkitOverflowScrolling:"touch",scrollbarWidth:"none","&::-webkit-scrollbar":{display:"none"}},content:{minWidth:"100%"},scrollbar:{display:"flex",userSelect:"none",touchAction:"none",borderRadius:"full",colorPalette:"gray",transition:"opacity 150ms 300ms",position:"relative",margin:"var(--scrollbar-margin)","&:not([data-overflow-x], [data-overflow-y])":{display:"none"},bg:"{colors.colorPalette.solid/10}","--thumb-bg":"{colors.colorPalette.solid/25}","&:is(:hover, :active)":{"--thumb-bg":"{colors.colorPalette.solid/50}"},_before:{content:'""',position:"absolute"},_vertical:{width:"var(--scrollbar-size)",flexDirection:"column","&::before":{width:"var(--scrollbar-click-area)",height:"100%",insetInlineStart:"calc(var(--scrollbar-margin) * -1)"}},_horizontal:{height:"var(--scrollbar-size)",flexDirection:"row","&::before":{height:"var(--scrollbar-click-area)",width:"100%",top:"calc(var(--scrollbar-margin) * -1)"}}},thumb:{borderRadius:"inherit",bg:"var(--thumb-bg)",transition:"backgrounds",_vertical:{width:"full"},_horizontal:{height:"full"}},corner:{bg:"bg.muted",margin:"var(--scrollbar-margin)",opacity:0,transition:"opacity 150ms 300ms","&[data-hover]":{transitionDelay:"0ms",opacity:1}}},variants:{variant:{hover:{scrollbar:{opacity:"0","&[data-hover], &[data-scrolling]":{opacity:"1",transitionDuration:"faster",transitionDelay:"0ms"}}},always:{scrollbar:{opacity:"1"}}},size:{xs:{root:{"--scrollbar-size":"sizes.1"}},sm:{root:{"--scrollbar-size":"sizes.1.5"}},md:{root:{"--scrollbar-size":"sizes.2"}},lg:{root:{"--scrollbar-size":"sizes.3"}}}},defaultVariants:{size:"md",variant:"hover"}}),QJn=Bo({className:"chakra-segment-group",slots:KZt.keys(),base:{root:{"--segment-radius":"radii.l2",borderRadius:"l2",display:"inline-flex",boxShadow:"inset",minW:"max-content",textAlign:"center",position:"relative",isolation:"isolate",bg:"bg.muted",_vertical:{flexDirection:"column"}},item:{display:"flex",alignItems:"center",justifyContent:"center",userSelect:"none",fontSize:"sm",position:"relative",color:"fg",borderRadius:"var(--segment-radius)",_disabled:{opacity:"0.5"},"&:has(input:focus-visible)":{focusRing:"outside"},_before:{content:'""',position:"absolute",bg:"border",transition:"opacity 0.2s"},_horizontal:{_before:{insetInlineStart:0,insetBlock:"1.5",width:"1px"}},_vertical:{_before:{insetBlockStart:0,insetInline:"1.5",height:"1px"}},"& + &[data-state=checked], &[data-state=checked] + &, &:first-of-type":{_before:{opacity:"0"}},"&[data-state=checked][data-ssr]":{shadow:"sm",bg:"bg",borderRadius:"var(--segment-radius)"}},indicator:{shadow:"sm",pos:"absolute",bg:{_light:"bg",_dark:"bg.emphasized"},width:"var(--width)",height:"var(--height)",top:"var(--top)",left:"var(--left)",zIndex:-1,borderRadius:"var(--segment-radius)"}},variants:{size:{xs:{item:{textStyle:"xs",px:"3",gap:"1",height:"6"}},sm:{item:{textStyle:"sm",px:"4",gap:"2",height:"8"}},md:{item:{textStyle:"sm",px:"4",gap:"2",height:"10"}},lg:{item:{textStyle:"md",px:"4.5",gap:"3",height:"11"}}}},defaultVariants:{size:"md"}}),eZn=Bo({className:"chakra-slider",slots:rJn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"1",textStyle:"sm",position:"relative",isolation:"isolate",touchAction:"none"},label:{fontWeight:"medium",textStyle:"sm"},control:{display:"inline-flex",alignItems:"center",position:"relative"},track:{overflow:"hidden",borderRadius:"full",flex:"1"},range:{width:"inherit",height:"inherit",_disabled:{bg:"border.emphasized!"}},markerGroup:{position:"absolute!",zIndex:"1"},marker:{"--marker-bg":{base:"white",_underValue:"colors.bg"},display:"flex",alignItems:"center",gap:"calc(var(--slider-thumb-size) / 2)",color:"fg.muted",textStyle:"xs"},markerIndicator:{width:"var(--slider-marker-size)",height:"var(--slider-marker-size)",borderRadius:"full",bg:"var(--marker-bg)"},thumb:{width:"var(--slider-thumb-size)",height:"var(--slider-thumb-size)",display:"flex",alignItems:"center",justifyContent:"center",outline:0,zIndex:"2",borderRadius:"full",_focusVisible:{ring:"2px",ringColor:"colorPalette.focusRing",ringOffset:"2px",ringOffsetColor:"bg"}}},variants:{size:{sm:{root:{"--slider-thumb-size":"sizes.4","--slider-track-size":"sizes.1.5","--slider-marker-center":"6px","--slider-marker-size":"sizes.1","--slider-marker-inset":"3px"}},md:{root:{"--slider-thumb-size":"sizes.5","--slider-track-size":"sizes.2","--slider-marker-center":"8px","--slider-marker-size":"sizes.1","--slider-marker-inset":"4px"}},lg:{root:{"--slider-thumb-size":"sizes.6","--slider-track-size":"sizes.2.5","--slider-marker-center":"9px","--slider-marker-size":"sizes.1.5","--slider-marker-inset":"5px"}}},variant:{outline:{track:{shadow:"inset",bg:"bg.emphasized/72"},range:{bg:"colorPalette.solid"},thumb:{borderWidth:"2px",borderColor:"colorPalette.solid",bg:"bg",_disabled:{bg:"border.emphasized",borderColor:"border.emphasized"}}},solid:{track:{bg:"colorPalette.subtle",_disabled:{bg:"bg.muted"}},range:{bg:"colorPalette.solid"},thumb:{bg:"colorPalette.solid",_disabled:{bg:"border.emphasized"}}}},orientation:{vertical:{root:{display:"inline-flex"},control:{flexDirection:"column",height:"100%",minWidth:"var(--slider-thumb-size)","&[data-has-mark-label], &:has(.chakra-slider__marker-label)":{marginEnd:"4"}},track:{width:"var(--slider-track-size)"},thumb:{left:"50%",translate:"-50% 0"},markerGroup:{insetStart:"var(--slider-marker-center)",insetBlock:"var(--slider-marker-inset)"},marker:{flexDirection:"row"}},horizontal:{control:{flexDirection:"row",width:"100%",minHeight:"var(--slider-thumb-size)","&[data-has-mark-label], &:has(.chakra-slider__marker-label)":{marginBottom:"4"}},track:{height:"var(--slider-track-size)"},thumb:{top:"50%",translate:"0 -50%"},markerGroup:{top:"var(--slider-marker-center)",insetInline:"var(--slider-marker-inset)"},marker:{flexDirection:"column"}}}},defaultVariants:{size:"md",variant:"outline",orientation:"horizontal"}}),tZn=Bo({className:"chakra-stat",slots:iJn.keys(),base:{root:{display:"flex",flexDirection:"column",gap:"1",position:"relative",flex:"1"},label:{display:"inline-flex",gap:"1.5",alignItems:"center",color:"fg.muted",textStyle:"sm"},helpText:{color:"fg.muted",textStyle:"xs"},valueUnit:{color:"fg.muted",textStyle:"xs",fontWeight:"initial",letterSpacing:"initial"},valueText:{verticalAlign:"baseline",fontWeight:"semibold",letterSpacing:"tight",fontFeatureSettings:"pnum",fontVariantNumeric:"proportional-nums",display:"inline-flex",gap:"1"},indicator:{display:"inline-flex",alignItems:"center",justifyContent:"center",marginEnd:1,"& :where(svg)":{w:"1em",h:"1em"},"&[data-type=up]":{color:"fg.success"},"&[data-type=down]":{color:"fg.error"}}},variants:{size:{sm:{valueText:{textStyle:"xl"}},md:{valueText:{textStyle:"2xl"}},lg:{valueText:{textStyle:"3xl"}}}},defaultVariants:{size:"md"}}),nZn=Bo({className:"chakra-status",slots:oJn.keys(),base:{root:{display:"inline-flex",alignItems:"center",gap:"2"},indicator:{width:"0.64em",height:"0.64em",flexShrink:0,borderRadius:"full",forcedColorAdjust:"none",bg:"colorPalette.solid"}},variants:{size:{sm:{root:{textStyle:"xs"}},md:{root:{textStyle:"sm"}},lg:{root:{textStyle:"md"}}}},defaultVariants:{size:"md"}}),rZn=Bo({className:"chakra-steps",slots:aJn.keys(),base:{root:{display:"flex",width:"full"},list:{display:"flex",justifyContent:"space-between","--steps-gutter":"spacing.3","--steps-thickness":"2px"},title:{fontWeight:"medium",color:"fg"},description:{color:"fg.muted"},separator:{bg:"border",flex:"1"},indicator:{display:"flex",justifyContent:"center",alignItems:"center",flexShrink:"0",borderRadius:"full",fontWeight:"medium",width:"var(--steps-size)",height:"var(--steps-size)",_icon:{flexShrink:"0",width:"var(--steps-icon-size)",height:"var(--steps-icon-size)"}},item:{position:"relative",display:"flex",gap:"3",flex:"1 0 0","&:last-of-type":{flex:"initial","& [data-part=separator]":{display:"none"}}},trigger:{display:"flex",alignItems:"center",gap:"3",textAlign:"start",focusVisibleRing:"outside",borderRadius:"l2"},content:{focusVisibleRing:"outside"}},variants:{orientation:{vertical:{root:{flexDirection:"row",height:"100%"},list:{flexDirection:"column",alignItems:"flex-start"},separator:{position:"absolute",width:"var(--steps-thickness)",height:"100%",maxHeight:"calc(100% - var(--steps-size) - var(--steps-gutter) * 2)",top:"calc(var(--steps-size) + var(--steps-gutter))",insetStart:"calc(var(--steps-size) / 2 - 1px)"},item:{alignItems:"flex-start"}},horizontal:{root:{flexDirection:"column",width:"100%"},list:{flexDirection:"row",alignItems:"center"},separator:{width:"100%",height:"var(--steps-thickness)",marginX:"var(--steps-gutter)"},item:{alignItems:"center"}}},variant:{solid:{indicator:{_incomplete:{borderWidth:"var(--steps-thickness)"},_current:{bg:"colorPalette.muted",borderWidth:"var(--steps-thickness)",borderColor:"colorPalette.solid",color:"colorPalette.fg"},_complete:{bg:"colorPalette.solid",borderColor:"colorPalette.solid",color:"colorPalette.contrast"}},separator:{_complete:{bg:"colorPalette.solid"}}},subtle:{indicator:{_incomplete:{bg:"bg.muted"},_current:{bg:"colorPalette.muted",color:"colorPalette.fg"},_complete:{bg:"colorPalette.emphasized",color:"colorPalette.fg"}},separator:{_complete:{bg:"colorPalette.emphasized"}}}},size:{xs:{root:{gap:"2.5"},list:{"--steps-size":"sizes.6","--steps-icon-size":"sizes.3.5",textStyle:"xs"},title:{textStyle:"sm"}},sm:{root:{gap:"3"},list:{"--steps-size":"sizes.8","--steps-icon-size":"sizes.4",textStyle:"xs"},title:{textStyle:"sm"}},md:{root:{gap:"4"},list:{"--steps-size":"sizes.10","--steps-icon-size":"sizes.4",textStyle:"sm"},title:{textStyle:"sm"}},lg:{root:{gap:"6"},list:{"--steps-size":"sizes.11","--steps-icon-size":"sizes.5",textStyle:"md"},title:{textStyle:"md"}}}},defaultVariants:{size:"md",variant:"solid",orientation:"horizontal"}}),iZn=Bo({slots:sJn.keys(),className:"chakra-switch",base:{root:{display:"inline-flex",gap:"2.5",alignItems:"center",position:"relative",verticalAlign:"middle","--switch-diff":"calc(var(--switch-width) - var(--switch-height))","--switch-x":{base:"var(--switch-diff)",_rtl:"calc(var(--switch-diff) * -1)"}},label:{lineHeight:"1",userSelect:"none",fontSize:"sm",fontWeight:"medium",_disabled:{opacity:"0.5"}},indicator:{position:"absolute",height:"var(--switch-height)",width:"var(--switch-height)",fontSize:"var(--switch-indicator-font-size)",fontWeight:"medium",flexShrink:0,userSelect:"none",display:"grid",placeContent:"center",transition:"inset-inline-start 0.12s ease",insetInlineStart:"calc(var(--switch-x) - 2px)",_checked:{insetInlineStart:"2px"}},control:{display:"inline-flex",gap:"0.5rem",flexShrink:0,justifyContent:"flex-start",cursor:"switch",borderRadius:"full",position:"relative",width:"var(--switch-width)",height:"var(--switch-height)",transition:"backgrounds",_disabled:{opacity:"0.5",cursor:"not-allowed"},_invalid:{outline:"2px solid",outlineColor:"border.error",outlineOffset:"2px"}},thumb:{display:"flex",alignItems:"center",justifyContent:"center",flexShrink:0,transitionProperty:"translate",transitionDuration:"fast",borderRadius:"inherit",_checked:{translate:"var(--switch-x) 0"}}},variants:{variant:{solid:{control:{borderRadius:"full",bg:"bg.emphasized",focusVisibleRing:"outside",_checked:{bg:"colorPalette.solid"}},thumb:{bg:"white",width:"var(--switch-height)",height:"var(--switch-height)",scale:"0.8",boxShadow:"sm",_checked:{bg:"colorPalette.contrast"}}},raised:{control:{borderRadius:"full",height:"calc(var(--switch-height) / 2)",bg:"bg.muted",boxShadow:"inset",_checked:{bg:"colorPalette.solid/60"}},thumb:{width:"var(--switch-height)",height:"var(--switch-height)",position:"relative",top:"calc(var(--switch-height) * -0.25)",bg:"white",boxShadow:"xs",focusVisibleRing:"outside",_checked:{bg:"colorPalette.solid"}}}},size:{xs:{root:{"--switch-width":"sizes.6","--switch-height":"sizes.3","--switch-indicator-font-size":"fontSizes.xs"}},sm:{root:{"--switch-width":"sizes.8","--switch-height":"sizes.4","--switch-indicator-font-size":"fontSizes.xs"}},md:{root:{"--switch-width":"sizes.10","--switch-height":"sizes.5","--switch-indicator-font-size":"fontSizes.sm"}},lg:{root:{"--switch-width":"sizes.12","--switch-height":"sizes.6","--switch-indicator-font-size":"fontSizes.md"}}}},defaultVariants:{variant:"solid",size:"md"}}),oZn=Bo({className:"chakra-table",slots:cJn.keys(),base:{root:{fontVariantNumeric:"lining-nums tabular-nums",borderCollapse:"collapse",width:"full",textAlign:"start",verticalAlign:"top"},row:{_selected:{bg:"colorPalette.subtle"}},cell:{textAlign:"start",alignItems:"center"},columnHeader:{fontWeight:"medium",textAlign:"start",color:"fg"},caption:{fontWeight:"medium",textStyle:"xs"},footer:{fontWeight:"medium"}},variants:{interactive:{true:{body:{"& tr":{_hover:{bg:"colorPalette.subtle"}}}}},stickyHeader:{true:{header:{"& :where(tr)":{top:"var(--table-sticky-offset, 0)",position:"sticky",zIndex:1}}}},striped:{true:{row:{"&:nth-of-type(odd) td":{bg:"bg.muted"}}}},showColumnBorder:{true:{columnHeader:{"&:not(:last-of-type)":{borderInlineEndWidth:"1px"}},cell:{"&:not(:last-of-type)":{borderInlineEndWidth:"1px"}}}},variant:{line:{columnHeader:{borderBottomWidth:"1px"},cell:{borderBottomWidth:"1px"},row:{bg:"bg"}},outline:{root:{boxShadow:"0 0 0 1px {colors.border}",overflow:"hidden"},columnHeader:{borderBottomWidth:"1px"},header:{bg:"bg.muted"},row:{"&:not(:last-of-type)":{borderBottomWidth:"1px"}},footer:{borderTopWidth:"1px"}}},size:{sm:{root:{textStyle:"sm"},columnHeader:{px:"2",py:"2"},cell:{px:"2",py:"2"}},md:{root:{textStyle:"sm"},columnHeader:{px:"3",py:"3"},cell:{px:"3",py:"3"}},lg:{root:{textStyle:"md"},columnHeader:{px:"4",py:"3"},cell:{px:"4",py:"3"}}}},defaultVariants:{variant:"line",size:"md"}}),aZn=Bo({slots:lJn.keys(),className:"chakra-tabs",base:{root:{"--tabs-trigger-radius":"radii.l2",position:"relative",_horizontal:{display:"block"},_vertical:{display:"flex"}},list:{display:"inline-flex",position:"relative",isolation:"isolate","--tabs-indicator-shadow":"shadows.xs","--tabs-indicator-bg":"colors.bg",minH:"var(--tabs-height)",_horizontal:{flexDirection:"row"},_vertical:{flexDirection:"column"}},trigger:{outline:"0",minW:"var(--tabs-height)",height:"var(--tabs-height)",display:"flex",alignItems:"center",fontWeight:"medium",position:"relative",cursor:"button",gap:"2",_focusVisible:{zIndex:1,outline:"2px solid",outlineColor:"colorPalette.focusRing"},_disabled:{cursor:"not-allowed",opacity:.5}},content:{focusVisibleRing:"inside",_horizontal:{width:"100%",pt:"var(--tabs-content-padding)"},_vertical:{height:"100%",ps:"var(--tabs-content-padding)"}},indicator:{width:"var(--width)",height:"var(--height)",borderRadius:"var(--tabs-indicator-radius)",bg:"var(--tabs-indicator-bg)",shadow:"var(--tabs-indicator-shadow)",zIndex:-1}},variants:{fitted:{true:{list:{display:"flex"},trigger:{flex:1,textAlign:"center",justifyContent:"center"}}},justify:{start:{list:{justifyContent:"flex-start"}},center:{list:{justifyContent:"center"}},end:{list:{justifyContent:"flex-end"}}},size:{sm:{root:{"--tabs-height":"sizes.9","--tabs-content-padding":"spacing.3"},trigger:{py:"1",px:"3",textStyle:"sm"}},md:{root:{"--tabs-height":"sizes.10","--tabs-content-padding":"spacing.4"},trigger:{py:"2",px:"4",textStyle:"sm"}},lg:{root:{"--tabs-height":"sizes.11","--tabs-content-padding":"spacing.4.5"},trigger:{py:"2",px:"4.5",textStyle:"md"}}},variant:{line:{list:{display:"flex",borderColor:"border",_horizontal:{borderBottomWidth:"1px"},_vertical:{borderEndWidth:"1px"}},trigger:{color:"fg.muted",_disabled:{_active:{bg:"initial"}},_selected:{color:"fg",_horizontal:{layerStyle:"indicator.bottom","--indicator-offset-y":"-1px","--indicator-color":"colors.colorPalette.solid"},_vertical:{layerStyle:"indicator.end","--indicator-offset-x":"-1px"}}}},subtle:{trigger:{borderRadius:"var(--tabs-trigger-radius)",color:"fg.muted",_selected:{bg:"colorPalette.subtle",color:"colorPalette.fg"}}},enclosed:{list:{bg:"bg.muted",padding:"1",borderRadius:"l3",minH:"calc(var(--tabs-height) - 4px)"},trigger:{justifyContent:"center",color:"fg.muted",borderRadius:"var(--tabs-trigger-radius)",_selected:{bg:"bg",color:"colorPalette.fg",shadow:"xs"}}},outline:{list:{"--line-thickness":"1px","--line-offset":"calc(var(--line-thickness) * -1)",borderColor:"border",display:"flex",_horizontal:{_before:{content:'""',position:"absolute",bottom:"0px",width:"100%",borderBottomWidth:"var(--line-thickness)",borderBottomColor:"border"}},_vertical:{_before:{content:'""',position:"absolute",insetInline:"var(--line-offset)",height:"calc(100% - calc(var(--line-thickness) * 2))",borderEndWidth:"var(--line-thickness)",borderEndColor:"border"}}},trigger:{color:"fg.muted",borderWidth:"1px",borderColor:"transparent",_selected:{bg:"currentBg",color:"colorPalette.fg"},_horizontal:{borderTopRadius:"var(--tabs-trigger-radius)",marginBottom:"var(--line-offset)",marginEnd:{_notLast:"var(--line-offset)"},_selected:{borderColor:"border",borderBottomColor:"transparent"}},_vertical:{borderStartRadius:"var(--tabs-trigger-radius)",marginEnd:"var(--line-offset)",marginBottom:{_notLast:"var(--line-offset)"},_selected:{borderColor:"border",borderEndColor:"transparent"}}}},plain:{trigger:{color:"fg.muted",_selected:{color:"colorPalette.fg"},borderRadius:"var(--tabs-trigger-radius)","&[data-selected][data-ssr]":{bg:"var(--tabs-indicator-bg)",shadow:"var(--tabs-indicator-shadow)",borderRadius:"var(--tabs-indicator-radius)"}}}}},defaultVariants:{size:"md",variant:"line"}}),Y1e=_Ye.variants?.variant,sZn=Bo({slots:dJn.keys(),className:"chakra-tag",base:{root:{display:"inline-flex",alignItems:"center",verticalAlign:"top",maxWidth:"100%",userSelect:"none",borderRadius:"l2",focusVisibleRing:"outside"},label:{lineClamp:"1"},closeTrigger:{display:"flex",alignItems:"center",justifyContent:"center",outline:"0",borderRadius:"l1",color:"currentColor",focusVisibleRing:"inside",focusRingWidth:"2px"},startElement:{flexShrink:0,boxSize:"var(--tag-element-size)",ms:"var(--tag-element-offset)","&:has([data-scope=avatar])":{boxSize:"var(--tag-avatar-size)",ms:"calc(var(--tag-element-offset) * 1.5)"},_icon:{boxSize:"100%"}},endElement:{flexShrink:0,boxSize:"var(--tag-element-size)",me:"var(--tag-element-offset)",_icon:{boxSize:"100%"},"&:has(button)":{ms:"calc(var(--tag-element-offset) * -1)"}}},variants:{size:{sm:{root:{px:"1.5",minH:"4.5",gap:"1","--tag-avatar-size":"spacing.3","--tag-element-size":"spacing.3","--tag-element-offset":"-2px"},label:{textStyle:"xs"}},md:{root:{px:"1.5",minH:"5",gap:"1","--tag-avatar-size":"spacing.3.5","--tag-element-size":"spacing.3.5","--tag-element-offset":"-2px"},label:{textStyle:"xs"}},lg:{root:{px:"2",minH:"6",gap:"1.5","--tag-avatar-size":"spacing.4.5","--tag-element-size":"spacing.4","--tag-element-offset":"-3px"},label:{textStyle:"sm"}},xl:{root:{px:"2.5",minH:"8",gap:"1.5","--tag-avatar-size":"spacing.6","--tag-element-size":"spacing.4.5","--tag-element-offset":"-4px"},label:{textStyle:"sm"}}},variant:{subtle:{root:Y1e?.subtle},solid:{root:Y1e?.solid},outline:{root:Y1e?.outline},surface:{root:Y1e?.surface}}},defaultVariants:{size:"md",variant:"surface"}}),cZn=Bo({slots:fJn.keys(),className:"chakra-timeline",base:{root:{display:"flex",flexDirection:"column",width:"full","--timeline-thickness":"1px","--timeline-gutter":"4px"},item:{display:"flex",position:"relative",alignItems:"flex-start",flexShrink:0,gap:"4",_last:{"& :where(.chakra-timeline__separator)":{display:"none"}}},separator:{position:"absolute",borderStartWidth:"var(--timeline-thickness)",ms:"calc(-1 * var(--timeline-thickness) / 2)",insetInlineStart:"calc(var(--timeline-indicator-size) / 2)",insetBlock:"0",borderColor:"border"},indicator:{outline:"2px solid {colors.bg}",position:"relative",flexShrink:"0",boxSize:"var(--timeline-indicator-size)",fontSize:"var(--timeline-font-size)",display:"flex",alignItems:"center",justifyContent:"center",borderRadius:"full",fontWeight:"medium"},connector:{alignSelf:"stretch",position:"relative"},content:{pb:"6",display:"flex",flexDirection:"column",width:"full",gap:"2"},title:{display:"flex",fontWeight:"medium",flexWrap:"wrap",gap:"1.5",alignItems:"center",mt:"var(--timeline-margin)"},description:{color:"fg.muted",textStyle:"xs"}},variants:{variant:{subtle:{indicator:{bg:"colorPalette.muted"}},solid:{indicator:{bg:"colorPalette.solid",color:"colorPalette.contrast"}},outline:{indicator:{bg:"currentBg",borderWidth:"1px",borderColor:"colorPalette.muted"}},plain:{}},size:{sm:{root:{"--timeline-indicator-size":"sizes.4","--timeline-font-size":"fontSizes.2xs"},title:{textStyle:"xs"}},md:{root:{"--timeline-indicator-size":"sizes.5","--timeline-font-size":"fontSizes.xs"},title:{textStyle:"sm"}},lg:{root:{"--timeline-indicator-size":"sizes.6","--timeline-font-size":"fontSizes.xs"},title:{mt:"0.5",textStyle:"sm"}},xl:{root:{"--timeline-indicator-size":"sizes.8","--timeline-font-size":"fontSizes.sm"},title:{mt:"1.5",textStyle:"sm"}}}},defaultVariants:{size:"md",variant:"solid"}}),uZn=Bo({slots:uJn.keys(),className:"chakra-toast",base:{root:{width:"full",display:"flex",alignItems:"flex-start",position:"relative",gap:"3",py:"4",ps:"4",pe:"6",borderRadius:"l2",translate:"var(--x) var(--y)",scale:"var(--scale)",zIndex:"var(--z-index)",height:"var(--height)",opacity:"var(--opacity)",willChange:"translate, opacity, scale",transition:"translate 400ms, scale 400ms, opacity 400ms, height 400ms, box-shadow 200ms",transitionTimingFunction:"cubic-bezier(0.21, 1.02, 0.73, 1)",_closed:{transition:"translate 400ms, scale 400ms, opacity 200ms",transitionTimingFunction:"cubic-bezier(0.06, 0.71, 0.55, 1)"},bg:"bg.panel",color:"fg",boxShadow:"xl","--toast-trigger-bg":"colors.bg.muted","&[data-type=warning]":{bg:"orange.solid",color:"orange.contrast","--toast-trigger-bg":"{white/10}","--toast-border-color":"{white/40}"},"&[data-type=success]":{bg:"green.solid",color:"green.contrast","--toast-trigger-bg":"{white/10}","--toast-border-color":"{white/40}"},"&[data-type=error]":{bg:"red.solid",color:"red.contrast","--toast-trigger-bg":"{white/10}","--toast-border-color":"{white/40}"}},title:{fontWeight:"medium",textStyle:"sm",marginEnd:"2"},description:{display:"inline",textStyle:"sm",opacity:"0.8"},indicator:{flexShrink:"0",boxSize:"5"},actionTrigger:{textStyle:"sm",fontWeight:"medium",height:"8",px:"3",borderRadius:"l2",alignSelf:"center",borderWidth:"1px",borderColor:"var(--toast-border-color, inherit)",transition:"background 200ms",_hover:{bg:"var(--toast-trigger-bg)"}},closeTrigger:{position:"absolute",top:"1",insetEnd:"1",padding:"1",display:"inline-flex",alignItems:"center",justifyContent:"center",color:"{currentColor/60}",borderRadius:"l2",textStyle:"md",transition:"background 200ms",_icon:{boxSize:"1em"}}}}),lZn=Bo({slots:NQt.keys(),className:"chakra-tooltip",base:{content:{"--tooltip-bg":"colors.bg.inverted",bg:"var(--tooltip-bg)",color:"fg.inverted",px:"2.5",py:"1",borderRadius:"l2",fontWeight:"medium",textStyle:"xs",boxShadow:"md",maxW:"xs",zIndex:"tooltip",transformOrigin:"var(--transform-origin)",_open:{animationStyle:"scale-fade-in",animationDuration:"fast"},_closed:{animationStyle:"scale-fade-out",animationDuration:"fast"}},arrow:{"--arrow-size":"sizes.2","--arrow-background":"var(--tooltip-bg)"},arrowTip:{borderTopWidth:"1px",borderInlineStartWidth:"1px",borderColor:"var(--tooltip-bg)"}}}),k7t=_K({display:"flex",alignItems:"center",gap:"var(--tree-item-gap)",rounded:"l2",userSelect:"none",position:"relative","--tree-depth":"calc(var(--depth) - 1)","--tree-indentation-offset":"calc(var(--tree-indentation) * var(--tree-depth))","--tree-icon-offset":"calc(var(--tree-icon-size) * var(--tree-depth) * 0.5)","--tree-offset":"calc(var(--tree-padding-inline) + var(--tree-indentation-offset) + var(--tree-icon-offset))",ps:"var(--tree-offset)",pe:"var(--tree-padding-inline)",py:"var(--tree-padding-block)",focusVisibleRing:"inside",focusRingColor:"border.emphasized",focusRingWidth:"2px","&:hover, &:focus-visible":{bg:"bg.muted"},_disabled:{layerStyle:"disabled"}}),T7t=_K({flex:"1"}),x7t=_K({_selected:{bg:"colorPalette.subtle",color:"colorPalette.fg"}}),C7t=_K({_selected:{layerStyle:"fill.solid"}}),dZn=Bo({slots:lJt.keys(),className:"chakra-tree-view",base:{root:{width:"full",display:"flex",flexDirection:"column",gap:"2"},tree:{display:"flex",flexDirection:"column","--tree-item-gap":"spacing.2",_icon:{boxSize:"var(--tree-icon-size)"}},label:{fontWeight:"medium",textStyle:"sm"},branch:{position:"relative"},branchContent:{position:"relative"},branchIndentGuide:{height:"100%",width:"1px",bg:"border",position:"absolute","--tree-depth":"calc(var(--depth) - 1)","--tree-indentation-offset":"calc(var(--tree-indentation) * var(--tree-depth))","--tree-offset":"calc(var(--tree-padding-inline) + var(--tree-indentation-offset))","--tree-icon-offset":"calc(var(--tree-icon-size) * 0.5 * var(--depth))",insetInlineStart:"calc(var(--tree-offset) + var(--tree-icon-offset))",zIndex:"1"},branchIndicator:{color:"fg.muted",transformOrigin:"center",transitionDuration:"normal",transitionProperty:"transform",transitionTimingFunction:"default",_open:{transform:"rotate(90deg)"}},branchTrigger:{display:"inline-flex",alignItems:"center",justifyContent:"center"},branchControl:k7t,item:k7t,itemText:T7t,branchText:T7t,nodeCheckbox:{display:"inline-flex"}},variants:{size:{md:{tree:{textStyle:"sm","--tree-indentation":"spacing.4","--tree-padding-inline":"spacing.3","--tree-padding-block":"spacing.1.5","--tree-icon-size":"spacing.4"}},sm:{tree:{textStyle:"sm","--tree-indentation":"spacing.4","--tree-padding-inline":"spacing.3","--tree-padding-block":"spacing.1","--tree-icon-size":"spacing.3"}},xs:{tree:{textStyle:"xs","--tree-indentation":"spacing.4","--tree-padding-inline":"spacing.2","--tree-padding-block":"spacing.1","--tree-icon-size":"spacing.3"}}},variant:{subtle:{branchControl:x7t,item:x7t},solid:{branchControl:C7t,item:C7t}},animateContent:{true:{branchContent:{_open:{animationName:"expand-height, fade-in",animationDuration:"moderate"},_closed:{animationName:"collapse-height, fade-out",animationDuration:"moderate"}}}}},defaultVariants:{size:"md",variant:"subtle"}}),fZn={accordion:mJn,actionBar:bJn,alert:vJn,avatar:wJn,blockquote:yJn,breadcrumb:EJn,card:SJn,checkbox:_Jn,checkboxCard:kJn,codeBlock:TJn,collapsible:xJn,dataList:IJn,dialog:OJn,drawer:NJn,editable:RJn,emptyState:PJn,field:LJn,fieldset:DJn,fileUpload:MJn,hoverCard:jJn,list:$Jn,listbox:FJn,menu:BJn,nativeSelect:UJn,numberInput:zJn,pinInput:GJn,popover:qJn,progress:VJn,progressCircle:WJn,radioCard:YJn,radioGroup:XJn,ratingGroup:JJn,scrollArea:ZJn,segmentGroup:QJn,select:cve,combobox:AJn,slider:eZn,stat:tZn,steps:rZn,switch:iZn,table:oZn,tabs:aZn,tag:sZn,toast:uZn,tooltip:lZn,status:nZn,timeline:cZn,colorPicker:CJn,qrCode:KJn,treeView:dZn},hZn=zKn({"2xs":{value:{fontSize:"2xs",lineHeight:"0.75rem"}},xs:{value:{fontSize:"xs",lineHeight:"1rem"}},sm:{value:{fontSize:"sm",lineHeight:"1.25rem"}},md:{value:{fontSize:"md",lineHeight:"1.5rem"}},lg:{value:{fontSize:"lg",lineHeight:"1.75rem"}},xl:{value:{fontSize:"xl",lineHeight:"1.875rem"}},"2xl":{value:{fontSize:"2xl",lineHeight:"2rem"}},"3xl":{value:{fontSize:"3xl",lineHeight:"2.375rem"}},"4xl":{value:{fontSize:"4xl",lineHeight:"2.75rem",letterSpacing:"-0.025em"}},"5xl":{value:{fontSize:"5xl",lineHeight:"3.75rem",letterSpacing:"-0.025em"}},"6xl":{value:{fontSize:"6xl",lineHeight:"4.5rem",letterSpacing:"-0.025em"}},"7xl":{value:{fontSize:"7xl",lineHeight:"5.75rem",letterSpacing:"-0.025em"}},none:{value:{}},label:{value:{fontSize:"sm",lineHeight:"1.25rem",fontWeight:"medium"}}}),pZn=Zp.animations({spin:{value:"spin 1s linear infinite"},ping:{value:"ping 1s cubic-bezier(0, 0, 0.2, 1) infinite"},pulse:{value:"pulse 2s cubic-bezier(0.4, 0, 0.6, 1) infinite"},bounce:{value:"bounce 1s infinite"}}),gZn=Zp.aspectRatios({square:{value:"1 / 1"},landscape:{value:"4 / 3"},portrait:{value:"3 / 4"},wide:{value:"16 / 9"},ultrawide:{value:"18 / 5"},golden:{value:"1.618 / 1"}}),mZn=Zp.blurs({none:{value:" "},sm:{value:"4px"},md:{value:"8px"},lg:{value:"12px"},xl:{value:"16px"},"2xl":{value:"24px"},"3xl":{value:"40px"},"4xl":{value:"64px"}}),bZn=Zp.borders({xs:{value:"0.5px solid"},sm:{value:"1px solid"},md:{value:"2px solid"},lg:{value:"4px solid"},xl:{value:"8px solid"}}),vZn=Zp.colors({transparent:{value:"transparent"},current:{value:"currentColor"},black:{value:"#09090B"},white:{value:"#FFFFFF"},whiteAlpha:{50:{value:"rgba(255, 255, 255, 0.04)"},100:{value:"rgba(255, 255, 255, 0.06)"},200:{value:"rgba(255, 255, 255, 0.08)"},300:{value:"rgba(255, 255, 255, 0.16)"},400:{value:"rgba(255, 255, 255, 0.24)"},500:{value:"rgba(255, 255, 255, 0.36)"},600:{value:"rgba(255, 255, 255, 0.48)"},700:{value:"rgba(255, 255, 255, 0.64)"},800:{value:"rgba(255, 255, 255, 0.80)"},900:{value:"rgba(255, 255, 255, 0.92)"},950:{value:"rgba(255, 255, 255, 0.95)"}},blackAlpha:{50:{value:"rgba(0, 0, 0, 0.04)"},100:{value:"rgba(0, 0, 0, 0.06)"},200:{value:"rgba(0, 0, 0, 0.08)"},300:{value:"rgba(0, 0, 0, 0.16)"},400:{value:"rgba(0, 0, 0, 0.24)"},500:{value:"rgba(0, 0, 0, 0.36)"},600:{value:"rgba(0, 0, 0, 0.48)"},700:{value:"rgba(0, 0, 0, 0.64)"},800:{value:"rgba(0, 0, 0, 0.80)"},900:{value:"rgba(0, 0, 0, 0.92)"},950:{value:"rgba(0, 0, 0, 0.95)"}},gray:{50:{value:"#fafafa"},100:{value:"#f4f4f5"},200:{value:"#e4e4e7"},300:{value:"#d4d4d8"},400:{value:"#a1a1aa"},500:{value:"#71717a"},600:{value:"#52525b"},700:{value:"#3f3f46"},800:{value:"#27272a"},900:{value:"#18181b"},950:{value:"#111111"}},red:{50:{value:"#fef2f2"},100:{value:"#fee2e2"},200:{value:"#fecaca"},300:{value:"#fca5a5"},400:{value:"#f87171"},500:{value:"#ef4444"},600:{value:"#dc2626"},700:{value:"#991919"},800:{value:"#511111"},900:{value:"#300c0c"},950:{value:"#1f0808"}},orange:{50:{value:"#fff7ed"},100:{value:"#ffedd5"},200:{value:"#fed7aa"},300:{value:"#fdba74"},400:{value:"#fb923c"},500:{value:"#f97316"},600:{value:"#ea580c"},700:{value:"#92310a"},800:{value:"#6c2710"},900:{value:"#3b1106"},950:{value:"#220a04"}},yellow:{50:{value:"#fefce8"},100:{value:"#fef9c3"},200:{value:"#fef08a"},300:{value:"#fde047"},400:{value:"#facc15"},500:{value:"#eab308"},600:{value:"#ca8a04"},700:{value:"#845209"},800:{value:"#713f12"},900:{value:"#422006"},950:{value:"#281304"}},green:{50:{value:"#f0fdf4"},100:{value:"#dcfce7"},200:{value:"#bbf7d0"},300:{value:"#86efac"},400:{value:"#4ade80"},500:{value:"#22c55e"},600:{value:"#16a34a"},700:{value:"#116932"},800:{value:"#124a28"},900:{value:"#042713"},950:{value:"#03190c"}},teal:{50:{value:"#f0fdfa"},100:{value:"#ccfbf1"},200:{value:"#99f6e4"},300:{value:"#5eead4"},400:{value:"#2dd4bf"},500:{value:"#14b8a6"},600:{value:"#0d9488"},700:{value:"#0c5d56"},800:{value:"#114240"},900:{value:"#032726"},950:{value:"#021716"}},blue:{50:{value:"#eff6ff"},100:{value:"#dbeafe"},200:{value:"#bfdbfe"},300:{value:"#a3cfff"},400:{value:"#60a5fa"},500:{value:"#3b82f6"},600:{value:"#2563eb"},700:{value:"#173da6"},800:{value:"#1a3478"},900:{value:"#14204a"},950:{value:"#0c142e"}},cyan:{50:{value:"#ecfeff"},100:{value:"#cffafe"},200:{value:"#a5f3fc"},300:{value:"#67e8f9"},400:{value:"#22d3ee"},500:{value:"#06b6d4"},600:{value:"#0891b2"},700:{value:"#0c5c72"},800:{value:"#134152"},900:{value:"#072a38"},950:{value:"#051b24"}},purple:{50:{value:"#faf5ff"},100:{value:"#f3e8ff"},200:{value:"#e9d5ff"},300:{value:"#d8b4fe"},400:{value:"#c084fc"},500:{value:"#a855f7"},600:{value:"#9333ea"},700:{value:"#641ba3"},800:{value:"#4a1772"},900:{value:"#2f0553"},950:{value:"#1a032e"}},pink:{50:{value:"#fdf2f8"},100:{value:"#fce7f3"},200:{value:"#fbcfe8"},300:{value:"#f9a8d4"},400:{value:"#f472b6"},500:{value:"#ec4899"},600:{value:"#db2777"},700:{value:"#a41752"},800:{value:"#6d0e34"},900:{value:"#45061f"},950:{value:"#2c0514"}}}),wZn=Zp.cursor({button:{value:"pointer"},checkbox:{value:"default"},disabled:{value:"not-allowed"},menuitem:{value:"default"},option:{value:"default"},radio:{value:"default"},slider:{value:"default"},switch:{value:"pointer"}}),yZn=Zp.durations({fastest:{value:"50ms"},faster:{value:"100ms"},fast:{value:"150ms"},moderate:{value:"200ms"},slow:{value:"300ms"},slower:{value:"400ms"},slowest:{value:"500ms"}}),EZn=Zp.easings({"ease-in":{value:"cubic-bezier(0.42, 0, 1, 1)"},"ease-out":{value:"cubic-bezier(0, 0, 0.58, 1)"},"ease-in-out":{value:"cubic-bezier(0.42, 0, 0.58, 1)"},"ease-in-smooth":{value:"cubic-bezier(0.32, 0.72, 0, 1)"}}),SZn=Zp.fontSizes({"2xs":{value:"0.625rem"},xs:{value:"0.75rem"},sm:{value:"0.875rem"},md:{value:"1rem"},lg:{value:"1.125rem"},xl:{value:"1.25rem"},"2xl":{value:"1.5rem"},"3xl":{value:"1.875rem"},"4xl":{value:"2.25rem"},"5xl":{value:"3rem"},"6xl":{value:"3.75rem"},"7xl":{value:"4.5rem"},"8xl":{value:"6rem"},"9xl":{value:"8rem"}}),_Zn=Zp.fontWeights({thin:{value:"100"},extralight:{value:"200"},light:{value:"300"},normal:{value:"400"},medium:{value:"500"},semibold:{value:"600"},bold:{value:"700"},extrabold:{value:"800"},black:{value:"900"}}),A7t='-apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol"',kZn=Zp.fonts({heading:{value:`Inter, ${A7t}`},body:{value:`Inter, ${A7t}`},mono:{value:'SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",monospace'}}),TZn=BKn({spin:{"0%":{transform:"rotate(0deg)"},"100%":{transform:"rotate(360deg)"}},pulse:{"50%":{opacity:"0.5"}},ping:{"75%, 100%":{transform:"scale(2)",opacity:"0"}},bounce:{"0%, 100%":{transform:"translateY(-25%)",animationTimingFunction:"cubic-bezier(0.8,0,1,1)"},"50%":{transform:"none",animationTimingFunction:"cubic-bezier(0,0,0.2,1)"}},"bg-position":{from:{backgroundPosition:"var(--animate-from, 1rem) 0"},to:{backgroundPosition:"var(--animate-to, 0) 0"}},position:{from:{insetInlineStart:"var(--animate-from-x)",insetBlockStart:"var(--animate-from-y)"},to:{insetInlineStart:"var(--animate-to-x)",insetBlockStart:"var(--animate-to-y)"}},"circular-progress":{"0%":{strokeDasharray:"1, 400",strokeDashoffset:"0"},"50%":{strokeDasharray:"400, 400",strokeDashoffset:"-100%"},"100%":{strokeDasharray:"400, 400",strokeDashoffset:"-260%"}},"expand-height":{from:{height:"0"},to:{height:"var(--height)"}},"collapse-height":{from:{height:"var(--height)"},to:{height:"0"}},"expand-width":{from:{width:"0"},to:{width:"var(--width)"}},"collapse-width":{from:{height:"var(--width)"},to:{height:"0"}},"fade-in":{from:{opacity:0},to:{opacity:1}},"fade-out":{from:{opacity:1},to:{opacity:0}},"slide-from-left-full":{from:{translate:"-100% 0"},to:{translate:"0 0"}},"slide-from-right-full":{from:{translate:"100% 0"},to:{translate:"0 0"}},"slide-from-top-full":{from:{translate:"0 -100%"},to:{translate:"0 0"}},"slide-from-bottom-full":{from:{translate:"0 100%"},to:{translate:"0 0"}},"slide-to-left-full":{from:{translate:"0 0"},to:{translate:"-100% 0"}},"slide-to-right-full":{from:{translate:"0 0"},to:{translate:"100% 0"}},"slide-to-top-full":{from:{translate:"0 0"},to:{translate:"0 -100%"}},"slide-to-bottom-full":{from:{translate:"0 0"},to:{translate:"0 100%"}},"slide-from-top":{"0%":{translate:"0 -0.5rem"},to:{translate:"0"}},"slide-from-bottom":{"0%":{translate:"0 0.5rem"},to:{translate:"0"}},"slide-from-left":{"0%":{translate:"-0.5rem 0"},to:{translate:"0"}},"slide-from-right":{"0%":{translate:"0.5rem 0"},to:{translate:"0"}},"slide-to-top":{"0%":{translate:"0"},to:{translate:"0 -0.5rem"}},"slide-to-bottom":{"0%":{translate:"0"},to:{translate:"0 0.5rem"}},"slide-to-left":{"0%":{translate:"0"},to:{translate:"-0.5rem 0"}},"slide-to-right":{"0%":{translate:"0"},to:{translate:"0.5rem 0"}},"scale-in":{from:{scale:"0.95"},to:{scale:"1"}},"scale-out":{from:{scale:"1"},to:{scale:"0.95"}}}),xZn=Zp.letterSpacings({tighter:{value:"-0.05em"},tight:{value:"-0.025em"},wide:{value:"0.025em"},wider:{value:"0.05em"},widest:{value:"0.1em"}}),CZn=Zp.lineHeights({shorter:{value:1.25},short:{value:1.375},moderate:{value:1.5},tall:{value:1.625},taller:{value:2}}),AZn=Zp.radii({none:{value:"0"},"2xs":{value:"0.0625rem"},xs:{value:"0.125rem"},sm:{value:"0.25rem"},md:{value:"0.375rem"},lg:{value:"0.5rem"},xl:{value:"0.75rem"},"2xl":{value:"1rem"},"3xl":{value:"1.5rem"},"4xl":{value:"2rem"},full:{value:"9999px"}}),den=Zp.spacing({.5:{value:"0.125rem"},1:{value:"0.25rem"},1.5:{value:"0.375rem"},2:{value:"0.5rem"},2.5:{value:"0.625rem"},3:{value:"0.75rem"},3.5:{value:"0.875rem"},4:{value:"1rem"},4.5:{value:"1.125rem"},5:{value:"1.25rem"},6:{value:"1.5rem"},7:{value:"1.75rem"},8:{value:"2rem"},9:{value:"2.25rem"},10:{value:"2.5rem"},11:{value:"2.75rem"},12:{value:"3rem"},14:{value:"3.5rem"},16:{value:"4rem"},20:{value:"5rem"},24:{value:"6rem"},28:{value:"7rem"},32:{value:"8rem"},36:{value:"9rem"},40:{value:"10rem"},44:{value:"11rem"},48:{value:"12rem"},52:{value:"13rem"},56:{value:"14rem"},60:{value:"15rem"},64:{value:"16rem"},72:{value:"18rem"},80:{value:"20rem"},96:{value:"24rem"}}),IZn=Zp.sizes({"3xs":{value:"14rem"},"2xs":{value:"16rem"},xs:{value:"20rem"},sm:{value:"24rem"},md:{value:"28rem"},lg:{value:"32rem"},xl:{value:"36rem"},"2xl":{value:"42rem"},"3xl":{value:"48rem"},"4xl":{value:"56rem"},"5xl":{value:"64rem"},"6xl":{value:"72rem"},"7xl":{value:"80rem"},"8xl":{value:"90rem"}}),OZn=Zp.sizes({max:{value:"max-content"},min:{value:"min-content"},fit:{value:"fit-content"},prose:{value:"60ch"},full:{value:"100%"},dvh:{value:"100dvh"},svh:{value:"100svh"},lvh:{value:"100lvh"},dvw:{value:"100dvw"},svw:{value:"100svw"},lvw:{value:"100lvw"},vw:{value:"100vw"},vh:{value:"100vh"}}),NZn=Zp.sizes({"1/2":{value:"50%"},"1/3":{value:"33.333333%"},"2/3":{value:"66.666667%"},"1/4":{value:"25%"},"3/4":{value:"75%"},"1/5":{value:"20%"},"2/5":{value:"40%"},"3/5":{value:"60%"},"4/5":{value:"80%"},"1/6":{value:"16.666667%"},"2/6":{value:"33.333333%"},"3/6":{value:"50%"},"4/6":{value:"66.666667%"},"5/6":{value:"83.333333%"},"1/12":{value:"8.333333%"},"2/12":{value:"16.666667%"},"3/12":{value:"25%"},"4/12":{value:"33.333333%"},"5/12":{value:"41.666667%"},"6/12":{value:"50%"},"7/12":{value:"58.333333%"},"8/12":{value:"66.666667%"},"9/12":{value:"75%"},"10/12":{value:"83.333333%"},"11/12":{value:"91.666667%"}}),RZn=Zp.sizes({...IZn,...den,...NZn,...OZn}),PZn=Zp.zIndex({hide:{value:-1},base:{value:0},docked:{value:10},dropdown:{value:1e3},sticky:{value:1100},banner:{value:1200},overlay:{value:1300},modal:{value:1400},popover:{value:1500},skipNav:{value:1600},toast:{value:1700},tooltip:{value:1800},max:{value:2147483647}}),LZn={aspectRatios:gZn,animations:pZn,blurs:mZn,borders:bZn,colors:vZn,durations:yZn,easings:EZn,fonts:kZn,fontSizes:SZn,fontWeights:_Zn,letterSpacings:xZn,lineHeights:CZn,radii:AZn,spacing:den,sizes:RZn,zIndex:PZn,cursor:wZn},DZn={colors:NXn,shadows:PXn,radii:RXn},MZn="chakra",jZn=":where(html, .chakra-theme)",$Zn=EYe({preflight:!0,cssVarsPrefix:MZn,cssVarsRoot:jZn,globalCss:dXn,theme:{breakpoints:lXn,keyframes:TZn,tokens:LZn,semanticTokens:DZn,recipes:OXn,slotRecipes:fZn,textStyles:hZn,layerStyles:fXn,animationStyles:hXn}}),fen=FQt(ZKn,$Zn);sen(fen);function hen(n){const{key:r,recipe:o}=n,s=yoe();return Q.useMemo(()=>{const c=o||(r!=null?s.getSlotRecipe(r):{});return s.sva(structuredClone(c))},[r,o,s])}const FZn=n=>n.charAt(0).toUpperCase()+n.slice(1),J1=n=>{const{key:r,recipe:o}=n,s=FZn(r||o.className||"Component"),[c,d]=tie({name:`${s}StylesContext`,errorMessage:`use${s}Styles returned is 'undefined'. Seems you forgot to wrap the components in "<${s}.Root />" `}),[h,b]=tie({name:`${s}ClassNameContext`,errorMessage:`use${s}ClassNames returned is 'undefined'. Seems you forgot to wrap the components in "<${s}.Root />" `,strict:!1}),[y,E]=tie({strict:!1,name:`${s}PropsContext`,providerName:`${s}PropsContext`,defaultValue:{}});function k(D){const{unstyled:$,...U}=D,B=hen({key:r,recipe:U.recipe||o}),[G,K]=Q.useMemo(()=>B.splitVariantProps(U),[U,B]);return{styles:Q.useMemo(()=>$?XKt:B(G),[$,G,B]),classNames:B.classNameMap,props:K}}function x(D,$={}){const{defaultProps:U}=$,B=G=>{const K=E(),X=Q.useMemo(()=>HW(U,K,G),[K,G]),{styles:ae,classNames:oe,props:he}=k(X);return H.jsx(c,{value:ae,children:H.jsx(h,{value:oe,children:H.jsx(D,{...he})})})};return B.displayName=D.displayName||D.name,B}return{StylesProvider:c,ClassNamesProvider:h,PropsProvider:y,usePropsContext:E,useRecipeResult:k,withProvider:(D,$,U)=>{const{defaultProps:B,...G}=U??{},K=ol(D,{},G),X=Q.forwardRef((ae,oe)=>{const he=E(),ye=Q.useMemo(()=>HW(B??{},he,ae),[he,ae]),{styles:we,props:ge,classNames:Te}=k(ye),xe=Te[$],Ue=H.jsx(c,{value:we,children:H.jsx(h,{value:Te,children:H.jsx(K,{ref:oe,...ge,css:[we[$],ye.css],className:Hb(ye.className,xe)})})});return U?.wrapElement?.(Ue,ye)??Ue});return X.displayName=D.displayName||D.name,X},withContext:(D,$,U)=>{const B=ol(D,{},U),G=Q.forwardRef((K,X)=>{const{unstyled:ae,...oe}=K,he=d(),we=b()?.[$];return H.jsx(B,{...oe,css:[!ae&&$?he[$]:void 0,K.css],ref:X,className:Hb(K.className,we)})});return G.displayName=D.displayName||D.name,G},withRootProvider:x,useStyles:d,useClassNames:b}};function pen(n,r){const o=yoe();return Q.useMemo(()=>(Array.isArray(r)?r:[r]).map(c=>o.token(`${n}.${c}`,c)),[o,n,r])}const gen=ol("div",{base:{position:"absolute",display:"flex",alignItems:"center",justifyContent:"center"},variants:{axis:{horizontal:{insetStart:"50%",translate:"-50%",_rtl:{translate:"50%"}},vertical:{top:"50%",translate:"0 -50%"},both:{insetStart:"50%",top:"50%",translate:"-50% -50%",_rtl:{translate:"50% -50%"}}}},defaultVariants:{axis:"both"}});gen.displayName="AbsoluteCenter";const men=n=>H.jsx(ol.svg,{viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:"2",strokeLinecap:"round",strokeLinejoin:"round",...n,children:H.jsx("path",{d:"M20 6 9 17l-5-5"})}),BZn=n=>H.jsx(ol.svg,{stroke:"currentColor",fill:"currentColor",strokeWidth:"0",viewBox:"0 0 24 24",...n,children:H.jsx("path",{d:"M12 22C6.47715 22 2 17.5228 2 12C2 6.47715 6.47715 2 12 2C17.5228 2 22 6.47715 22 12C22 17.5228 17.5228 22 12 22ZM12 20C16.4183 20 20 16.4183 20 12C20 7.58172 16.4183 4 12 4C7.58172 4 4 7.58172 4 12C4 16.4183 7.58172 20 12 20ZM11.0026 16L6.75999 11.7574L8.17421 10.3431L11.0026 13.1716L16.6595 7.51472L18.0737 8.92893L11.0026 16Z"})}),I7t=n=>H.jsx(ol.svg,{stroke:"currentColor",fill:"currentColor",strokeWidth:"0",viewBox:"0 0 24 24",...n,children:H.jsx("path",{d:"M12 22C6.47715 22 2 17.5228 2 12C2 6.47715 6.47715 2 12 2C17.5228 2 22 6.47715 22 12C22 17.5228 17.5228 22 12 22ZM12 20C16.4183 20 20 16.4183 20 12C20 7.58172 16.4183 4 12 4C7.58172 4 4 7.58172 4 12C4 16.4183 7.58172 20 12 20ZM11 15H13V17H11V15ZM11 7H13V13H11V7Z"})}),ben=n=>H.jsx(ol.svg,{viewBox:"0 0 24 24",fill:"currentColor",...n,children:H.jsx("path",{fillRule:"evenodd",clipRule:"evenodd",d:"M18.7071 6.70711C19.0976 6.31658 19.0976 5.68342 18.7071 5.29289C18.3166 4.90237 17.6834 4.90237 17.2929 5.29289L12 10.5858L6.70711 5.29289C6.31658 4.90237 5.68342 4.90237 5.29289 5.29289C4.90237 5.68342 4.90237 6.31658 5.29289 6.70711L10.5858 12L5.29289 17.2929C4.90237 17.6834 4.90237 18.3166 5.29289 18.7071C5.68342 19.0976 6.31658 19.0976 6.70711 18.7071L12 13.4142L17.2929 18.7071C17.6834 19.0976 18.3166 19.0976 18.7071 18.7071C19.0976 18.3166 19.0976 17.6834 18.7071 17.2929L13.4142 12L18.7071 6.70711Z"})}),_W=n=>n?"":void 0,UZn=ol("div",{base:{display:"inline-flex",gap:"var(--group-gap, 0.5rem)",isolation:"isolate",position:"relative","& [data-group-item]":{_focusVisible:{zIndex:1}}},variants:{orientation:{horizontal:{flexDirection:"row"},vertical:{flexDirection:"column"}},attached:{true:{gap:"0!"}},grow:{true:{display:"flex","& > *":{flex:1}}},stacking:{"first-on-top":{"& > [data-group-item]":{zIndex:"calc(var(--group-count) - var(--group-index))"}},"last-on-top":{"& > [data-group-item]":{zIndex:"var(--group-index)"}}}},compoundVariants:[{orientation:"horizontal",attached:!0,css:{"& > *[data-first]":{borderEndRadius:"0!",marginEnd:"-1px"},"& > *[data-between]":{borderRadius:"0!",marginEnd:"-1px"},"& > *[data-last]":{borderStartRadius:"0!"}}},{orientation:"vertical",attached:!0,css:{"& > *[data-first]":{borderBottomRadius:"0!",marginBottom:"-1px"},"& > *[data-between]":{borderRadius:"0!",marginBottom:"-1px"},"& > *[data-last]":{borderTopRadius:"0!"}}}],defaultVariants:{orientation:"horizontal"}}),ven=Q.memo(Q.forwardRef(function(r,o){const{align:s="center",justify:c="flex-start",children:d,wrap:h,skip:b,...y}=r,E=Q.useMemo(()=>{let k=Q.Children.toArray(d).filter(Q.isValidElement);if(k.length===1)return k;const x=k.filter(P=>!b?.(P)),O=x.length;return x.length===1?k:k.map(P=>{const D=P.props;if(b?.(P))return P;const $=x.indexOf(P);return Q.cloneElement(P,{...D,"data-group-item":"","data-first":_W($===0),"data-last":_W($===O-1),"data-between":_W($>0&&$<O-1),style:{"--group-count":O,"--group-index":$,...D?.style??{}}})})},[d,b]);return H.jsx(UZn,{ref:o,alignItems:s,justifyContent:c,flexWrap:h,...y,className:Hb("chakra-group",r.className),children:E})})),{withProvider:wen,withContext:zZn,useStyles:HZn,useClassNames:GZn,PropsProvider:qZn}=J1({key:"avatar"});wen(rXt,"root",{forwardAsChild:!0});const VZn=wen(nXt,"root",{forwardAsChild:!0}),WZn=ol(JYt,{},{forwardAsChild:!0});function KZn(n){return n.children||n.asChild?n.children:n.name?YZn(n.name):H.jsx(ZZn,{})}function YZn(n){const r=n.trim().split(" "),o=r[0]!=null?r[0]:"",s=r.length>1?r[r.length-1]:"";return o&&s?`${o.charAt(0)}${s.charAt(0)}`:o.charAt(0)}const XZn=Q.forwardRef(function(r,o){const s=HZn(),c=GZn(),{name:d,...h}=r;return H.jsx(WZn,{ref:o,...h,className:Hb(r.className,c.fallback),css:[s.fallback,r.css],children:KZn(r)})}),JZn=zZn(ZYt,"image",{forwardAsChild:!0,defaultProps:{draggable:"false",referrerPolicy:"no-referrer"}}),ZZn=Q.forwardRef(function(r,o){return H.jsx(ol.svg,{stroke:"currentColor",fill:"currentColor",strokeWidth:"0",viewBox:"0 0 24 24",height:"1.2em",width:"1.2em",ref:o,...r,children:H.jsx("path",{d:"M20 22H18V20C18 18.3431 16.6569 17 15 17H9C7.34315 17 6 18.3431 6 20V22H4V20C4 17.2386 6.23858 15 9 15H15C17.7614 15 20 17.2386 20 20V22ZM12 13C8.68629 13 6 10.3137 6 7C6 3.68629 8.68629 1 12 1C15.3137 1 18 3.68629 18 7C18 10.3137 15.3137 13 12 13ZM12 11C14.2091 11 16 9.20914 16 7C16 4.79086 14.2091 3 12 3C9.79086 3 8 4.79086 8 7C8 9.20914 9.79086 11 12 11Z"})})});Q.forwardRef(function(r,o){const s=hen({key:"avatar"}),[c,d]=Q.useMemo(()=>s.splitVariantProps(r),[r,s]);return H.jsx(qZn,{value:c,children:H.jsx(ven,{gap:"0",spaceX:"-3",ref:o,...d})})});const{withContext:QZn}=Wx({key:"spinner"}),kK=QZn("span"),eQn=Q.forwardRef(function(r,o){const{spinner:s=H.jsx(kK,{size:"inherit",borderWidth:"0.125em",color:"inherit"}),spinnerPlacement:c="start",children:d,text:h,visible:b=!0,...y}=r;return b?h?H.jsxs(N1e,{ref:o,display:"contents",...y,children:[c==="start"&&s,h,c==="end"&&s]}):s?H.jsxs(N1e,{ref:o,display:"contents",...y,children:[H.jsx(gen,{display:"inline-flex",children:s}),H.jsx(N1e,{visibility:"hidden",display:"contents",children:d})]}):H.jsx(N1e,{ref:o,display:"contents",...y,children:d}):d}),{useRecipeResult:tQn,usePropsContext:nQn}=Wx({key:"button"}),ud=Q.forwardRef(function(r,o){const s=nQn(),c=Q.useMemo(()=>HW(s,r),[s,r]),d=tQn(c),{loading:h,loadingText:b,children:y,spinner:E,spinnerPlacement:k,...x}=d.props;return H.jsx(ol.button,{type:"button",ref:o,...x,"data-loading":_W(h),disabled:h||x.disabled,className:Hb(d.className,c.className),css:[d.styles,c.css],children:!c.asChild&&h?H.jsx(eQn,{spinner:E,text:b,spinnerPlacement:k,children:y}):y})}),Gm=Q.forwardRef(function(r,o){return H.jsx(ud,{px:"0",py:"0",_icon:{fontSize:"1.2em"},ref:o,...r})}),kYe=ol("div",{base:{display:"flex",alignItems:"center",justifyContent:"center"},variants:{inline:{true:{display:"inline-flex"}}}});kYe.displayName="Center";const{withContext:rQn}=Wx({key:"code"}),lVe=rQn("code");function iQn(n){const{gap:r,direction:o}=n,s={column:{marginY:r,marginX:0,borderInlineStartWidth:0,borderTopWidth:"1px"},"column-reverse":{marginY:r,marginX:0,borderInlineStartWidth:0,borderTopWidth:"1px"},row:{marginX:r,marginY:0,borderInlineStartWidth:"1px",borderTopWidth:0},"row-reverse":{marginX:r,marginY:0,borderInlineStartWidth:"1px",borderTopWidth:0}};return{"&":wYe(o,c=>s[c])}}function oQn(n){return Q.Children.toArray(n).filter(r=>Q.isValidElement(r))}const Cye=Q.forwardRef(function(r,o){const{direction:s="column",align:c,justify:d,gap:h="0.5rem",wrap:b,children:y,separator:E,className:k,...x}=r,O=Q.useMemo(()=>iQn({gap:h,direction:s}),[h,s]),P=Q.useMemo(()=>Q.isValidElement(E)?oQn(y).map((D,$,U)=>{const B=typeof D.key<"u"?D.key:$,G=E,K=Q.cloneElement(G,{css:[O,G.props.css]});return H.jsxs(Q.Fragment,{children:[D,$===U.length-1?null:K]},B)}):y,[y,E,O]);return H.jsx(ol.div,{ref:o,display:"flex",alignItems:c,justifyContent:d,flexDirection:s,flexWrap:b,gap:E?void 0:h,className:Hb("chakra-stack",k),...x,children:P})}),{withContext:aQn}=Wx({key:"container"}),sQn=aQn("div"),{withRootProvider:yen,withContext:UI}=J1({key:"dialog"});yen(Oqn,{defaultProps:{unmountOnExit:!0,lazyMount:!0}});const cQn=yen(Iqn,{defaultProps:{unmountOnExit:!0,lazyMount:!0}}),O7t=UI(PXt,"trigger",{forwardAsChild:!0}),uQn=UI(TXt,"positioner",{forwardAsChild:!0}),lQn=UI(_Xt,"content",{forwardAsChild:!0});UI(kXt,"description",{forwardAsChild:!0});UI(RXt,"title",{forwardAsChild:!0});const dQn=UI(SXt,"closeTrigger",{forwardAsChild:!0});Q.forwardRef(function(r,o){const s=FM();return H.jsx(ol.button,{...r,ref:o,onClick:()=>s.setOpen(!1)})});const fQn=UI(EXt,"backdrop",{forwardAsChild:!0}),hQn=UI("div","body"),pQn=UI("div","footer"),gQn=UI("div","header"),{withProvider:Een,withContext:UM}=J1({key:"editable"});Een(WXt,"root",{forwardAsChild:!0});const mQn=Een(VXt,"root",{forwardAsChild:!0}),bQn=UM(BXt,"preview",{forwardAsChild:!0}),vQn=UM(FXt,"input",{forwardAsChild:!0}),wQn=Q.forwardRef(function(r,o){const s=BM(),c=HW(s.getInputProps(),r);return H.jsx(Xn.textarea,{ref:o,...c})});UM(wQn,"input",{forwardAsChild:!0});UM(jXt,"control",{forwardAsChild:!0});UM(DXt,"area",{forwardAsChild:!0});UM($Xt,"editTrigger",{forwardAsChild:!0});UM(KXt,"submitTrigger",{forwardAsChild:!0});UM(MXt,"cancelTrigger",{forwardAsChild:!0});const{useRecipeResult:yQn}=Wx({key:"icon"}),EQn=Q.forwardRef(function(r,o){const{styles:s,className:c,props:d}=yQn({asChild:!r.as,...r});return H.jsx(ol.svg,{ref:o,focusable:!1,"aria-hidden":"true",...d,css:[s,r.css],className:Hb(c,r.className)})});function SQn(n){const{viewBox:r="0 0 24 24",d:o,displayName:s,defaultProps:c={}}=n,d=Q.Children.toArray(n.path),h=Q.forwardRef((b,y)=>H.jsx(EQn,{ref:y,asChild:!1,viewBox:r,...c,...b,children:d.length?d:H.jsx("path",{fill:"currentColor",d:o})}));return h.displayName=s,h}const{withProvider:_Qn,withContext:TYe,useStyles:kQn,useClassNames:TQn}=J1({key:"field"}),sie=_Qn(tJt,"root",{forwardAsChild:!0}),a$e=TYe(ZXt,"label",{forwardAsChild:!0});TYe(XXt,"helperText",{forwardAsChild:!0});TYe(YXt,"errorText",{forwardAsChild:!0});SQn({d:"M11.983,0a12.206,12.206,0,0,0-8.51,3.653A11.8,11.8,0,0,0,0,12.207,11.779,11.779,0,0,0,11.8,24h.214A12.111,12.111,0,0,0,24,11.791h0A11.766,11.766,0,0,0,11.983,0ZM10.5,16.542a1.476,1.476,0,0,1,1.449-1.53h.027a1.527,1.527,0,0,1,1.523,1.47,1.475,1.475,0,0,1-1.449,1.53h-.027A1.529,1.529,0,0,1,10.5,16.542ZM11,12.5v-6a1,1,0,0,1,2,0v6a1,1,0,1,1-2,0Z"});const xQn=Q.forwardRef(function(r,o){const{fallback:s,children:c="*",...d}=r,h=$M(),b=TQn(),y=kQn();return h?.required?H.jsx(ol.span,{ref:o,"aria-hidden":"true",...d,className:Hb(b.requiredIndicator,r.className),css:[y.requiredIndicator,r.css],children:c}):s}),{withProvider:CQn,withContext:Aye}=J1({key:"fieldset"}),Sen=CQn(sJt,"root");Aye(rJt,"errorText");Aye(iJt,"helperText");Aye(oJt,"legend");Aye(Xn.div,"content");function AQn(n){const{each:r,fallback:o,children:s}=n;return r?.length===0?o||null:r?.map(s)}const Iu=Q.forwardRef(function(r,o){const{direction:s,align:c,justify:d,wrap:h,basis:b,grow:y,shrink:E,inline:k,...x}=r;return H.jsx(ol.div,{ref:o,...x,css:{display:k?"inline-flex":"flex",flexDirection:s,alignItems:c,justifyContent:d,flexWrap:h,flexBasis:b,flexGrow:y,flexShrink:E,...r.css}})});function N7t(n){return wYe(n,r=>r==="auto"?"auto":`span ${r}/span ${r}`)}const xYe=Q.forwardRef(function(r,o){const{area:s,colSpan:c,colStart:d,colEnd:h,rowEnd:b,rowSpan:y,rowStart:E,...k}=r,x=Q.useMemo(()=>mK({gridArea:s,gridColumn:N7t(c),gridRow:N7t(y),gridColumnStart:d,gridColumnEnd:h,gridRowStart:E,gridRowEnd:b}),[s,c,d,h,b,y,E]);return H.jsx(ol.div,{ref:o,css:[x,r.css],...k})}),Iye=Q.forwardRef(function(r,o){const{align:s,fit:c="cover",...d}=r;return H.jsx(ol.img,{ref:o,objectFit:c,objectPosition:s,className:Hb("chakra-image",r.className),...d})}),{withContext:IQn}=Wx({key:"input"}),OQn=IQn(JXt),R7t=Q.forwardRef(function({unstyled:r,...o},s){const c=jKe({key:"inputAddon",recipe:o.recipe}),[d,h]=c.splitVariantProps(o),b=r?XKt:c(d);return H.jsx(ol.div,{ref:s,...h,css:[b,o.css]})}),s$e=ol("div",{base:{display:"flex",alignItems:"center",justifyContent:"center",position:"absolute",zIndex:2,color:"fg.muted",height:"full",fontSize:"sm",px:"3"},variants:{placement:{start:{insetInlineStart:"0"},end:{insetInlineEnd:"0"}}}}),NQn=Q.forwardRef(function(r,o){const{startElement:s,startElementProps:c,endElement:d,endElementProps:h,startAddon:b,startAddonProps:y,endAddon:E,endAddonProps:k,children:x,startOffset:O="0px",endOffset:P="0px",...D}=r,$=Q.Children.only(x),U=!!(b||E);return H.jsxs(ven,{width:"full",ref:o,attached:U,skip:B=>B.type===s$e,...D,children:[b&&H.jsx(R7t,{...y,children:b}),s&&H.jsx(s$e,{pointerEvents:"none",...c,children:s}),Q.cloneElement($,{...s&&{ps:`calc(var(--input-height) - ${O})`},...d&&{pe:`calc(var(--input-height) - ${P})`},...x.props}),d&&H.jsx(s$e,{placement:"end",...h,children:d}),E&&H.jsx(R7t,{...k,children:E})]})}),{withContext:RQn}=Wx({key:"link"}),_I=RQn("a"),{withProvider:PQn,withContext:_en}=J1({key:"list"}),Xve=PQn("ul","root",{defaultProps:{role:"list"}}),CYe=_en("li","item");_en("span","indicator");const{withRootProvider:ken,withContext:Vm}=J1({key:"menu"});ken(AVn);const AYe=ken(CVn,{defaultProps:{lazyMount:!0,unmountOnExit:!0}}),LQn=Vm(QJt,"trigger",{forwardAsChild:!0}),Ten=Vm(RJt,"contextTrigger",{forwardAsChild:!0}),IYe=Vm(UJt,"positioner",{forwardAsChild:!0});Vm(ZJt,"separator",{forwardAsChild:!0});const OYe=Vm(NJt,"content",{forwardAsChild:!0}),DQn=Vm(AJt,"arrowTip",{forwardAsChild:!0});Vm(CJt,"arrow",{forwardAsChild:!0,defaultProps:{children:H.jsx(DQn,{})}});Vm(PJt,"indicator",{forwardAsChild:!0});Vm(jJt,"itemGroup",{forwardAsChild:!0});Vm($Jt,"itemGroupLabel",{forwardAsChild:!0});const NB=Vm(LJt,"item",{forwardAsChild:!0});Vm(eZt,"item",{forwardAsChild:!0});Vm(BJt,"itemText",{forwardAsChild:!0});Vm("kbd","itemCommand");Vm(FJt,"itemIndicator",{forwardAsChild:!0,defaultProps:{children:H.jsx(men,{boxSize:"4"})}});Vm(OJt,"item",{forwardAsChild:!0});Vm(HJt,"itemGroup",{forwardAsChild:!0});Vm(zJt,"item",{forwardAsChild:!0});const{withRootProvider:xen,withContext:v2}=J1({key:"popover"});xen(uWn);const MQn=xen(cWn),P7t=v2(EZt,"trigger",{forwardAsChild:!0}),jQn=v2(gZt,"positioner",{forwardAsChild:!0}),$Qn=v2(fZt,"content",{forwardAsChild:!0}),FQn=v2(lZt,"arrowTip",{forwardAsChild:!0});v2(uZt,"arrow",{forwardAsChild:!0,defaultProps:{children:H.jsx(FQn,{})}});const BQn=v2(dZt,"closeTrigger",{forwardAsChild:!0});v2(pZt,"indicator",{forwardAsChild:!0});v2(yZt,"title",{forwardAsChild:!0});v2(hZt,"description",{forwardAsChild:!0});v2("footer","footer");v2("header","header");v2("div","body");v2(cZt,void 0,{forwardAsChild:!0});const UQn=Q.forwardRef(function(r,o){const s=jKe({key:"radiomark",recipe:r.recipe}),[c,d]=s.splitVariantProps(r),{checked:h,disabled:b,unstyled:y,children:E,...k}=d,x=y?YKt:s(c),O={ref:o,"data-checked":_W(h),"data-disabled":_W(b),...k,css:[x,r.css]};return H.jsx(ol.span,{...O,children:h&&H.jsx("span",{className:"dot"})})}),{withProvider:Cen,withContext:Oye,useStyles:zQn}=J1({key:"radioGroup"});Cen(PZt,"root",{forwardAsChild:!0});const NYe=Cen(RZt,"root",{forwardAsChild:!0});Oye(CZt,"label",{forwardAsChild:!0});const RYe=Oye(TZt,"item",{forwardAsChild:!0});Oye(xZt,"itemText",{forwardAsChild:!0});Oye(hYe,"itemControl",{forwardAsChild:!0});Q.forwardRef(function(r,o){const s=zQn();return H.jsx(gWn,{children:c=>H.jsx(hYe,{asChild:!0,children:H.jsx(UQn,{ref:o,unstyled:!0,...r,checked:c.checked,disabled:c.disabled,css:[s.itemControl,r.css]})})})});const{withProvider:Aen,withContext:Ooe}=J1({key:"scrollArea"});Aen(HZt,"root",{forwardAsChild:!0});const Ien=Aen(zZt,"root",{forwardAsChild:!0}),Oen=Ooe(VZt,"viewport",{forwardAsChild:!0});Ooe(MZt,"content",{forwardAsChild:!0});const PYe=Ooe(qZt,"thumb",{forwardAsChild:!0}),Nen=Ooe(GZt,"scrollbar",{forwardAsChild:!0,defaultProps:{children:H.jsx(PYe,{})}});Ooe(jZt,"corner",{forwardAsChild:!0});const{withProvider:Ren,withContext:LYe}=J1({key:"segmentGroup"});Ren(tQt,"root",{forwardAsChild:!0});const HQn=Ren(eQt,"root",{forwardAsChild:!0,forwardProps:["orientation"],defaultProps:{orientation:"horizontal"}}),GQn=LYe(JZt,"item",{forwardAsChild:!0}),qQn=LYe(QZt,"itemText",{forwardAsChild:!0}),VQn=LYe(YZt,"indicator",{forwardAsChild:!0});function WQn(n){return n.map(r=>typeof r=="string"?{value:r,label:r}:r)}const KQn=n=>{const{items:r,...o}=n,s=Q.useMemo(()=>WQn(r),[r]);return H.jsx(AQn,{each:s,children:c=>H.jsxs(GQn,{value:c.value,disabled:c.disabled,...o,children:[H.jsx(qQn,{children:c.label}),H.jsx(YQn,{})]},c.value)})},YQn=ZZt,{useRecipeResult:XQn}=Wx({key:"separator"}),dVe=Q.forwardRef(function(r,o){const{styles:s,className:c,props:d}=XQn(r),h=r.orientation||"horizontal";return H.jsx(ol.span,{ref:o,role:B_(h)?"separator":"presentation","aria-orientation":B_(h)?h:void 0,...vYe(d,["orientation"]),className:Hb(c,r.className),css:[s,r.css]})}),Pen=ol("div",{base:{flex:1,justifySelf:"stretch",alignSelf:"stretch"}});Pen.displayName="Spacer";const $c=Q.forwardRef(function(r,o){return H.jsx(Cye,{align:"center",...r,direction:"row",ref:o})}),Vd=Q.forwardRef(function(r,o){return H.jsx(Cye,{align:"center",...r,direction:"column",ref:o})}),{withProvider:Len,withContext:Jx}=J1({key:"steps"});Len(pQt,"root",{forwardAsChild:!0});const JQn=Len(hQt,"root",{forwardAsChild:!0}),ZQn=Jx(lQt,"list",{forwardAsChild:!0}),QQn=Jx(uQt,"item",{forwardAsChild:!0});Jx(mQt,"trigger",{forwardAsChild:!0});Jx(sQt,"content",{forwardAsChild:!0});Jx(aQt,"content");const eer=Q.forwardRef(function(r,o){return H.jsx(bYe,{children:s=>H.jsx(ol.div,{ref:o,...r,children:s.index+1})})});Jx("div","title");Jx("div","description");const ter=Jx(gQt,"separator",{forwardAsChild:!0}),Den=n=>H.jsx(bYe,{children:r=>r.current?H.jsx(H.Fragment,{children:n.current??n.incomplete}):r.completed?H.jsx(H.Fragment,{children:n.complete}):H.jsx(H.Fragment,{children:n.incomplete??n.current})}),ner=Jx(cQt,"indicator",{forwardAsChild:!0,defaultProps:{children:H.jsx(Den,{complete:H.jsx(men,{}),incomplete:H.jsx(eer,{})})}});Jx(dQt,"nextTrigger",{forwardAsChild:!0});Jx(fQt,"prevTrigger",{forwardAsChild:!0});const{withProvider:rer,withContext:Nye}=J1({key:"tag"}),Men=rer("div","root"),jen=Nye("span","label"),ier=Nye("button","closeTrigger",{defaultProps:{children:H.jsx(ben,{})}}),oer=Nye("span","startElement"),aer=Nye("span","endElement"),{withContext:ser}=Wx({key:"textarea"}),DYe=ser(nJt),{withProvider:cer,withContext:Rye,useStyles:uer}=J1({key:"toast"}),ler=ol(kQt,{},{forwardAsChild:!0}),der=cer(SQt,"root",{forwardAsChild:!0}),fer=Rye(yQt,"closeTrigger",{forwardAsChild:!0,defaultProps:{children:H.jsx(ben,{})}}),her=Rye(_Qt,"title",{forwardAsChild:!0}),per=Rye(EQt,"description",{forwardAsChild:!0}),ger=Rye(wQt,"actionTrigger",{forwardAsChild:!0}),mer={warning:I7t,success:BZn,error:I7t},ber=Q.forwardRef(function(r,o){const{unstyled:s,...c}=r,d=EK(),h=uer(),b=mer[d.type];return b?H.jsx(b,{ref:o,...c,css:[s?void 0:h.indicator,r.css]}):null}),{withRootProvider:$en,withContext:Noe}=J1({key:"tooltip"});$en(DKn);const ver=$en(LKn,{defaultProps:{lazyMount:!0,unmountOnExit:!0}}),wer=Noe(LQt,"trigger",{forwardAsChild:!0}),yer=Noe(OQt,"positioner",{forwardAsChild:!0}),Eer=Noe(IQt,"content",{forwardAsChild:!0}),Fen=Noe(AQt,"arrowTip",{forwardAsChild:!0}),Ser=Noe(CQt,"arrow",{forwardAsChild:!0,defaultProps:{children:H.jsx(Fen,{})}}),{withProvider:Ben,withContext:w2}=J1({key:"treeView"});Ben(oVn,"root",{forwardAsChild:!0});const _er=Ben(rVn,"root",{forwardAsChild:!0}),ker=w2(bJt,"branch",{forwardAsChild:!0}),Ter=w2(vJt,"branchContent",{forwardAsChild:!0}),xer=w2(wJt,"branchControl",{forwardAsChild:!0}),Cer=w2(DQt,"branchTrigger",{forwardAsChild:!0}),Aer=w2(EJt,"branchIndicator",{forwardAsChild:!0}),Ier=w2(SJt,"branchText",{forwardAsChild:!0}),Oer=w2(yJt,"branchIndentGuide",{forwardAsChild:!0}),Ner=w2(_Jt,"item",{forwardAsChild:!0});w2(MQt,"itemIndicator",{forwardAsChild:!0});w2(kJt,"itemText",{forwardAsChild:!0});w2(jQt,"label",{forwardAsChild:!0});const Rer=w2(gJt,"tree",{forwardAsChild:!0});w2($Qt,"nodeCheckbox",{forwardAsChild:!0});function Per(n){const{render:r,indentGuide:o,branchProps:s,branchContentProps:c}=n,d=lv(),h=(b,y)=>H.jsx(uVn,{node:b,indexPath:y,children:H.jsx(MKn,{children:E=>E.isBranch?H.jsxs(ker,{...s,children:[r({node:b,indexPath:y,nodeState:E}),H.jsxs(Ter,{...c,children:[o,d.collection.getNodeChildren(b).map((k,x)=>h(k,[...y,x]))]})]}):r({node:b,indexPath:y,nodeState:E})})},y.join("."));return H.jsx(H.Fragment,{children:d.collection.getNodeChildren(d.collection.rootNode).map((b,y)=>h(b,[y]))})}var TK=class{constructor(){this.listeners=new Set,this.subscribe=this.subscribe.bind(this)}subscribe(n){return this.listeners.add(n),this.onSubscribe(),()=>{this.listeners.delete(n),this.onUnsubscribe()}}hasListeners(){return this.listeners.size>0}onSubscribe(){}onUnsubscribe(){}},KB=typeof window>"u"||"Deno"in globalThis;function ov(){}function Ler(n,r){return typeof n=="function"?n(r):n}function fVe(n){return typeof n=="number"&&n>=0&&n!==1/0}function Uen(n,r){return Math.max(n+(r||0)-Date.now(),0)}function pM(n,r){return typeof n=="function"?n(r):n}function N_(n,r){return typeof n=="function"?n(r):n}function L7t(n,r){const{type:o="all",exact:s,fetchStatus:c,predicate:d,queryKey:h,stale:b}=n;if(h){if(s){if(r.queryHash!==MYe(h,r.options))return!1}else if(!$ie(r.queryKey,h))return!1}if(o!=="all"){const y=r.isActive();if(o==="active"&&!y||o==="inactive"&&y)return!1}return!(typeof b=="boolean"&&r.isStale()!==b||c&&c!==r.state.fetchStatus||d&&!d(r))}function D7t(n,r){const{exact:o,status:s,predicate:c,mutationKey:d}=n;if(d){if(!r.options.mutationKey)return!1;if(o){if(YB(r.options.mutationKey)!==YB(d))return!1}else if(!$ie(r.options.mutationKey,d))return!1}return!(s&&r.state.status!==s||c&&!c(r))}function MYe(n,r){return(r?.queryKeyHashFn||YB)(n)}function YB(n){return JSON.stringify(n,(r,o)=>hVe(o)?Object.keys(o).sort().reduce((s,c)=>(s[c]=o[c],s),{}):o)}function $ie(n,r){return n===r?!0:typeof n!=typeof r?!1:n&&r&&typeof n=="object"&&typeof r=="object"?Object.keys(r).every(o=>$ie(n[o],r[o])):!1}function zen(n,r){if(n===r)return n;const o=M7t(n)&&M7t(r);if(o||hVe(n)&&hVe(r)){const s=o?n:Object.keys(n),c=s.length,d=o?r:Object.keys(r),h=d.length,b=o?[]:{},y=new Set(s);let E=0;for(let k=0;k<h;k++){const x=o?k:d[k];(!o&&y.has(x)||o)&&n[x]===void 0&&r[x]===void 0?(b[x]=void 0,E++):(b[x]=zen(n[x],r[x]),b[x]===n[x]&&n[x]!==void 0&&E++)}return c===h&&E===c?n:b}return r}function Jve(n,r){if(!r||Object.keys(n).length!==Object.keys(r).length)return!1;for(const o in n)if(n[o]!==r[o])return!1;return!0}function M7t(n){return Array.isArray(n)&&n.length===Object.keys(n).length}function hVe(n){if(!j7t(n))return!1;const r=n.constructor;if(r===void 0)return!0;const o=r.prototype;return!(!j7t(o)||!o.hasOwnProperty("isPrototypeOf")||Object.getPrototypeOf(n)!==Object.prototype)}function j7t(n){return Object.prototype.toString.call(n)==="[object Object]"}function Der(n){return new Promise(r=>{setTimeout(r,n)})}function pVe(n,r,o){return typeof o.structuralSharing=="function"?o.structuralSharing(n,r):o.structuralSharing!==!1?zen(n,r):r}function Mer(n,r,o=0){const s=[...n,r];return o&&s.length>o?s.slice(1):s}function jer(n,r,o=0){const s=[r,...n];return o&&s.length>o?s.slice(0,-1):s}var jYe=Symbol();function Hen(n,r){return!n.queryFn&&r?.initialPromise?()=>r.initialPromise:!n.queryFn||n.queryFn===jYe?()=>Promise.reject(new Error(`Missing queryFn: '${n.queryHash}'`)):n.queryFn}function Gen(n,r){return typeof n=="function"?n(...r):!!n}var $er=class extends TK{#t;#e;#n;constructor(){super(),this.#n=n=>{if(!KB&&window.addEventListener){const r=()=>n();return window.addEventListener("visibilitychange",r,!1),()=>{window.removeEventListener("visibilitychange",r)}}}}onSubscribe(){this.#e||this.setEventListener(this.#n)}onUnsubscribe(){this.hasListeners()||(this.#e?.(),this.#e=void 0)}setEventListener(n){this.#n=n,this.#e?.(),this.#e=n(r=>{typeof r=="boolean"?this.setFocused(r):this.onFocus()})}setFocused(n){this.#t!==n&&(this.#t=n,this.onFocus())}onFocus(){const n=this.isFocused();this.listeners.forEach(r=>{r(n)})}isFocused(){return typeof this.#t=="boolean"?this.#t:globalThis.document?.visibilityState!=="hidden"}},$Ye=new $er,Fer=class extends TK{#t=!0;#e;#n;constructor(){super(),this.#n=n=>{if(!KB&&window.addEventListener){const r=()=>n(!0),o=()=>n(!1);return window.addEventListener("online",r,!1),window.addEventListener("offline",o,!1),()=>{window.removeEventListener("online",r),window.removeEventListener("offline",o)}}}}onSubscribe(){this.#e||this.setEventListener(this.#n)}onUnsubscribe(){this.hasListeners()||(this.#e?.(),this.#e=void 0)}setEventListener(n){this.#n=n,this.#e?.(),this.#e=n(this.setOnline.bind(this))}setOnline(n){this.#t!==n&&(this.#t=n,this.listeners.forEach(o=>{o(n)}))}isOnline(){return this.#t}},Zve=new Fer;function gVe(){let n,r;const o=new Promise((c,d)=>{n=c,r=d});o.status="pending",o.catch(()=>{});function s(c){Object.assign(o,c),delete o.resolve,delete o.reject}return o.resolve=c=>{s({status:"fulfilled",value:c}),n(c)},o.reject=c=>{s({status:"rejected",reason:c}),r(c)},o}function Ber(n){return Math.min(1e3*2**n,3e4)}function qen(n){return(n??"online")==="online"?Zve.isOnline():!0}var Ven=class extends Error{constructor(n){super("CancelledError"),this.revert=n?.revert,this.silent=n?.silent}};function c$e(n){return n instanceof Ven}function Wen(n){let r=!1,o=0,s=!1,c;const d=gVe(),h=$=>{s||(O(new Ven($)),n.abort?.())},b=()=>{r=!0},y=()=>{r=!1},E=()=>$Ye.isFocused()&&(n.networkMode==="always"||Zve.isOnline())&&n.canRun(),k=()=>qen(n.networkMode)&&n.canRun(),x=$=>{s||(s=!0,n.onSuccess?.($),c?.(),d.resolve($))},O=$=>{s||(s=!0,n.onError?.($),c?.(),d.reject($))},P=()=>new Promise($=>{c=U=>{(s||E())&&$(U)},n.onPause?.()}).then(()=>{c=void 0,s||n.onContinue?.()}),D=()=>{if(s)return;let $;const U=o===0?n.initialPromise:void 0;try{$=U??n.fn()}catch(B){$=Promise.reject(B)}Promise.resolve($).then(x).catch(B=>{if(s)return;const G=n.retry??(KB?0:3),K=n.retryDelay??Ber,X=typeof K=="function"?K(o,B):K,ae=G===!0||typeof G=="number"&&o<G||typeof G=="function"&&G(o,B);if(r||!ae){O(B);return}o++,n.onFail?.(o,B),Der(X).then(()=>E()?void 0:P()).then(()=>{r?O(B):D()})})};return{promise:d,cancel:h,continue:()=>(c?.(),d),cancelRetry:b,continueRetry:y,canStart:k,start:()=>(k()?D():P().then(D),d)}}var Uer=n=>setTimeout(n,0);function zer(){let n=[],r=0,o=b=>{b()},s=b=>{b()},c=Uer;const d=b=>{r?n.push(b):c(()=>{o(b)})},h=()=>{const b=n;n=[],b.length&&c(()=>{s(()=>{b.forEach(y=>{o(y)})})})};return{batch:b=>{let y;r++;try{y=b()}finally{r--,r||h()}return y},batchCalls:b=>(...y)=>{d(()=>{b(...y)})},schedule:d,setNotifyFunction:b=>{o=b},setBatchNotifyFunction:b=>{s=b},setScheduler:b=>{c=b}}}var qg=zer(),Ken=class{#t;destroy(){this.clearGcTimeout()}scheduleGc(){this.clearGcTimeout(),fVe(this.gcTime)&&(this.#t=setTimeout(()=>{this.optionalRemove()},this.gcTime))}updateGcTime(n){this.gcTime=Math.max(this.gcTime||0,n??(KB?1/0:300*1e3))}clearGcTimeout(){this.#t&&(clearTimeout(this.#t),this.#t=void 0)}},Her=class extends Ken{#t;#e;#n;#r;#i;#a;#s;constructor(n){super(),this.#s=!1,this.#a=n.defaultOptions,this.setOptions(n.options),this.observers=[],this.#r=n.client,this.#n=this.#r.getQueryCache(),this.queryKey=n.queryKey,this.queryHash=n.queryHash,this.#t=Ger(this.options),this.state=n.state??this.#t,this.scheduleGc()}get meta(){return this.options.meta}get promise(){return this.#i?.promise}setOptions(n){this.options={...this.#a,...n},this.updateGcTime(this.options.gcTime)}optionalRemove(){!this.observers.length&&this.state.fetchStatus==="idle"&&this.#n.remove(this)}setData(n,r){const o=pVe(this.state.data,n,this.options);return this.#o({data:o,type:"success",dataUpdatedAt:r?.updatedAt,manual:r?.manual}),o}setState(n,r){this.#o({type:"setState",state:n,setStateOptions:r})}cancel(n){const r=this.#i?.promise;return this.#i?.cancel(n),r?r.then(ov).catch(ov):Promise.resolve()}destroy(){super.destroy(),this.cancel({silent:!0})}reset(){this.destroy(),this.setState(this.#t)}isActive(){return this.observers.some(n=>N_(n.options.enabled,this)!==!1)}isDisabled(){return this.getObserversCount()>0?!this.isActive():this.options.queryFn===jYe||this.state.dataUpdateCount+this.state.errorUpdateCount===0}isStatic(){return this.getObserversCount()>0?this.observers.some(n=>pM(n.options.staleTime,this)==="static"):!1}isStale(){return this.getObserversCount()>0?this.observers.some(n=>n.getCurrentResult().isStale):this.state.data===void 0||this.state.isInvalidated}isStaleByTime(n=0){return this.state.data===void 0?!0:n==="static"?!1:this.state.isInvalidated?!0:!Uen(this.state.dataUpdatedAt,n)}onFocus(){this.observers.find(r=>r.shouldFetchOnWindowFocus())?.refetch({cancelRefetch:!1}),this.#i?.continue()}onOnline(){this.observers.find(r=>r.shouldFetchOnReconnect())?.refetch({cancelRefetch:!1}),this.#i?.continue()}addObserver(n){this.observers.includes(n)||(this.observers.push(n),this.clearGcTimeout(),this.#n.notify({type:"observerAdded",query:this,observer:n}))}removeObserver(n){this.observers.includes(n)&&(this.observers=this.observers.filter(r=>r!==n),this.observers.length||(this.#i&&(this.#s?this.#i.cancel({revert:!0}):this.#i.cancelRetry()),this.scheduleGc()),this.#n.notify({type:"observerRemoved",query:this,observer:n}))}getObserversCount(){return this.observers.length}invalidate(){this.state.isInvalidated||this.#o({type:"invalidate"})}fetch(n,r){if(this.state.fetchStatus!=="idle"){if(this.state.data!==void 0&&r?.cancelRefetch)this.cancel({silent:!0});else if(this.#i)return this.#i.continueRetry(),this.#i.promise}if(n&&this.setOptions(n),!this.options.queryFn){const y=this.observers.find(E=>E.options.queryFn);y&&this.setOptions(y.options)}const o=new AbortController,s=y=>{Object.defineProperty(y,"signal",{enumerable:!0,get:()=>(this.#s=!0,o.signal)})},c=()=>{const y=Hen(this.options,r),k=(()=>{const x={client:this.#r,queryKey:this.queryKey,meta:this.meta};return s(x),x})();return this.#s=!1,this.options.persister?this.options.persister(y,k,this):y(k)},h=(()=>{const y={fetchOptions:r,options:this.options,queryKey:this.queryKey,client:this.#r,state:this.state,fetchFn:c};return s(y),y})();this.options.behavior?.onFetch(h,this),this.#e=this.state,(this.state.fetchStatus==="idle"||this.state.fetchMeta!==h.fetchOptions?.meta)&&this.#o({type:"fetch",meta:h.fetchOptions?.meta});const b=y=>{c$e(y)&&y.silent||this.#o({type:"error",error:y}),c$e(y)||(this.#n.config.onError?.(y,this),this.#n.config.onSettled?.(this.state.data,y,this)),this.scheduleGc()};return this.#i=Wen({initialPromise:r?.initialPromise,fn:h.fetchFn,abort:o.abort.bind(o),onSuccess:y=>{if(y===void 0){b(new Error(`${this.queryHash} data is undefined`));return}try{this.setData(y)}catch(E){b(E);return}this.#n.config.onSuccess?.(y,this),this.#n.config.onSettled?.(y,this.state.error,this),this.scheduleGc()},onError:b,onFail:(y,E)=>{this.#o({type:"failed",failureCount:y,error:E})},onPause:()=>{this.#o({type:"pause"})},onContinue:()=>{this.#o({type:"continue"})},retry:h.options.retry,retryDelay:h.options.retryDelay,networkMode:h.options.networkMode,canRun:()=>!0}),this.#i.start()}#o(n){const r=o=>{switch(n.type){case"failed":return{...o,fetchFailureCount:n.failureCount,fetchFailureReason:n.error};case"pause":return{...o,fetchStatus:"paused"};case"continue":return{...o,fetchStatus:"fetching"};case"fetch":return{...o,...Yen(o.data,this.options),fetchMeta:n.meta??null};case"success":return this.#e=void 0,{...o,data:n.data,dataUpdateCount:o.dataUpdateCount+1,dataUpdatedAt:n.dataUpdatedAt??Date.now(),error:null,isInvalidated:!1,status:"success",...!n.manual&&{fetchStatus:"idle",fetchFailureCount:0,fetchFailureReason:null}};case"error":const s=n.error;return c$e(s)&&s.revert&&this.#e?{...this.#e,fetchStatus:"idle"}:{...o,error:s,errorUpdateCount:o.errorUpdateCount+1,errorUpdatedAt:Date.now(),fetchFailureCount:o.fetchFailureCount+1,fetchFailureReason:s,fetchStatus:"idle",status:"error"};case"invalidate":return{...o,isInvalidated:!0};case"setState":return{...o,...n.state}}};this.state=r(this.state),qg.batch(()=>{this.observers.forEach(o=>{o.onQueryUpdate()}),this.#n.notify({query:this,type:"updated",action:n})})}};function Yen(n,r){return{fetchFailureCount:0,fetchFailureReason:null,fetchStatus:qen(r.networkMode)?"fetching":"paused",...n===void 0&&{error:null,status:"pending"}}}function Ger(n){const r=typeof n.initialData=="function"?n.initialData():n.initialData,o=r!==void 0,s=o?typeof n.initialDataUpdatedAt=="function"?n.initialDataUpdatedAt():n.initialDataUpdatedAt:0;return{data:r,dataUpdateCount:0,dataUpdatedAt:o?s??Date.now():0,error:null,errorUpdateCount:0,errorUpdatedAt:0,fetchFailureCount:0,fetchFailureReason:null,fetchMeta:null,isInvalidated:!1,status:o?"success":"pending",fetchStatus:"idle"}}var qer=class extends TK{constructor(n={}){super(),this.config=n,this.#t=new Map}#t;build(n,r,o){const s=r.queryKey,c=r.queryHash??MYe(s,r);let d=this.get(c);return d||(d=new Her({client:n,queryKey:s,queryHash:c,options:n.defaultQueryOptions(r),state:o,defaultOptions:n.getQueryDefaults(s)}),this.add(d)),d}add(n){this.#t.has(n.queryHash)||(this.#t.set(n.queryHash,n),this.notify({type:"added",query:n}))}remove(n){const r=this.#t.get(n.queryHash);r&&(n.destroy(),r===n&&this.#t.delete(n.queryHash),this.notify({type:"removed",query:n}))}clear(){qg.batch(()=>{this.getAll().forEach(n=>{this.remove(n)})})}get(n){return this.#t.get(n)}getAll(){return[...this.#t.values()]}find(n){const r={exact:!0,...n};return this.getAll().find(o=>L7t(r,o))}findAll(n={}){const r=this.getAll();return Object.keys(n).length>0?r.filter(o=>L7t(n,o)):r}notify(n){qg.batch(()=>{this.listeners.forEach(r=>{r(n)})})}onFocus(){qg.batch(()=>{this.getAll().forEach(n=>{n.onFocus()})})}onOnline(){qg.batch(()=>{this.getAll().forEach(n=>{n.onOnline()})})}},Ver=class extends Ken{#t;#e;#n;constructor(n){super(),this.mutationId=n.mutationId,this.#e=n.mutationCache,this.#t=[],this.state=n.state||Xen(),this.setOptions(n.options),this.scheduleGc()}setOptions(n){this.options=n,this.updateGcTime(this.options.gcTime)}get meta(){return this.options.meta}addObserver(n){this.#t.includes(n)||(this.#t.push(n),this.clearGcTimeout(),this.#e.notify({type:"observerAdded",mutation:this,observer:n}))}removeObserver(n){this.#t=this.#t.filter(r=>r!==n),this.scheduleGc(),this.#e.notify({type:"observerRemoved",mutation:this,observer:n})}optionalRemove(){this.#t.length||(this.state.status==="pending"?this.scheduleGc():this.#e.remove(this))}continue(){return this.#n?.continue()??this.execute(this.state.variables)}async execute(n){const r=()=>{this.#r({type:"continue"})};this.#n=Wen({fn:()=>this.options.mutationFn?this.options.mutationFn(n):Promise.reject(new Error("No mutationFn found")),onFail:(c,d)=>{this.#r({type:"failed",failureCount:c,error:d})},onPause:()=>{this.#r({type:"pause"})},onContinue:r,retry:this.options.retry??0,retryDelay:this.options.retryDelay,networkMode:this.options.networkMode,canRun:()=>this.#e.canRun(this)});const o=this.state.status==="pending",s=!this.#n.canStart();try{if(o)r();else{this.#r({type:"pending",variables:n,isPaused:s}),await this.#e.config.onMutate?.(n,this);const d=await this.options.onMutate?.(n);d!==this.state.context&&this.#r({type:"pending",context:d,variables:n,isPaused:s})}const c=await this.#n.start();return await this.#e.config.onSuccess?.(c,n,this.state.context,this),await this.options.onSuccess?.(c,n,this.state.context),await this.#e.config.onSettled?.(c,null,this.state.variables,this.state.context,this),await this.options.onSettled?.(c,null,n,this.state.context),this.#r({type:"success",data:c}),c}catch(c){try{throw await this.#e.config.onError?.(c,n,this.state.context,this),await this.options.onError?.(c,n,this.state.context),await this.#e.config.onSettled?.(void 0,c,this.state.variables,this.state.context,this),await this.options.onSettled?.(void 0,c,n,this.state.context),c}finally{this.#r({type:"error",error:c})}}finally{this.#e.runNext(this)}}#r(n){const r=o=>{switch(n.type){case"failed":return{...o,failureCount:n.failureCount,failureReason:n.error};case"pause":return{...o,isPaused:!0};case"continue":return{...o,isPaused:!1};case"pending":return{...o,context:n.context,data:void 0,failureCount:0,failureReason:null,error:null,isPaused:n.isPaused,status:"pending",variables:n.variables,submittedAt:Date.now()};case"success":return{...o,data:n.data,failureCount:0,failureReason:null,error:null,status:"success",isPaused:!1};case"error":return{...o,data:void 0,error:n.error,failureCount:o.failureCount+1,failureReason:n.error,isPaused:!1,status:"error"}}};this.state=r(this.state),qg.batch(()=>{this.#t.forEach(o=>{o.onMutationUpdate(n)}),this.#e.notify({mutation:this,type:"updated",action:n})})}};function Xen(){return{context:void 0,data:void 0,error:null,failureCount:0,failureReason:null,isPaused:!1,status:"idle",variables:void 0,submittedAt:0}}var Wer=class extends TK{constructor(n={}){super(),this.config=n,this.#t=new Set,this.#e=new Map,this.#n=0}#t;#e;#n;build(n,r,o){const s=new Ver({mutationCache:this,mutationId:++this.#n,options:n.defaultMutationOptions(r),state:o});return this.add(s),s}add(n){this.#t.add(n);const r=X1e(n);if(typeof r=="string"){const o=this.#e.get(r);o?o.push(n):this.#e.set(r,[n])}this.notify({type:"added",mutation:n})}remove(n){if(this.#t.delete(n)){const r=X1e(n);if(typeof r=="string"){const o=this.#e.get(r);if(o)if(o.length>1){const s=o.indexOf(n);s!==-1&&o.splice(s,1)}else o[0]===n&&this.#e.delete(r)}}this.notify({type:"removed",mutation:n})}canRun(n){const r=X1e(n);if(typeof r=="string"){const s=this.#e.get(r)?.find(c=>c.state.status==="pending");return!s||s===n}else return!0}runNext(n){const r=X1e(n);return typeof r=="string"?this.#e.get(r)?.find(s=>s!==n&&s.state.isPaused)?.continue()??Promise.resolve():Promise.resolve()}clear(){qg.batch(()=>{this.#t.forEach(n=>{this.notify({type:"removed",mutation:n})}),this.#t.clear(),this.#e.clear()})}getAll(){return Array.from(this.#t)}find(n){const r={exact:!0,...n};return this.getAll().find(o=>D7t(r,o))}findAll(n={}){return this.getAll().filter(r=>D7t(n,r))}notify(n){qg.batch(()=>{this.listeners.forEach(r=>{r(n)})})}resumePausedMutations(){const n=this.getAll().filter(r=>r.state.isPaused);return qg.batch(()=>Promise.all(n.map(r=>r.continue().catch(ov))))}};function X1e(n){return n.options.scope?.id}function $7t(n){return{onFetch:(r,o)=>{const s=r.options,c=r.fetchOptions?.meta?.fetchMore?.direction,d=r.state.data?.pages||[],h=r.state.data?.pageParams||[];let b={pages:[],pageParams:[]},y=0;const E=async()=>{let k=!1;const x=D=>{Object.defineProperty(D,"signal",{enumerable:!0,get:()=>(r.signal.aborted?k=!0:r.signal.addEventListener("abort",()=>{k=!0}),r.signal)})},O=Hen(r.options,r.fetchOptions),P=async(D,$,U)=>{if(k)return Promise.reject();if($==null&&D.pages.length)return Promise.resolve(D);const G=(()=>{const oe={client:r.client,queryKey:r.queryKey,pageParam:$,direction:U?"backward":"forward",meta:r.options.meta};return x(oe),oe})(),K=await O(G),{maxPages:X}=r.options,ae=U?jer:Mer;return{pages:ae(D.pages,K,X),pageParams:ae(D.pageParams,$,X)}};if(c&&d.length){const D=c==="backward",$=D?Ker:F7t,U={pages:d,pageParams:h},B=$(s,U);b=await P(U,B,D)}else{const D=n??d.length;do{const $=y===0?h[0]??s.initialPageParam:F7t(s,b);if(y>0&&$==null)break;b=await P(b,$),y++}while(y<D)}return b};r.options.persister?r.fetchFn=()=>r.options.persister?.(E,{client:r.client,queryKey:r.queryKey,meta:r.options.meta,signal:r.signal},o):r.fetchFn=E}}}function F7t(n,{pages:r,pageParams:o}){const s=r.length-1;return r.length>0?n.getNextPageParam(r[s],r,o[s],o):void 0}function Ker(n,{pages:r,pageParams:o}){return r.length>0?n.getPreviousPageParam?.(r[0],r,o[0],o):void 0}var Yer=class{#t;#e;#n;#r;#i;#a;#s;#o;constructor(n={}){this.#t=n.queryCache||new qer,this.#e=n.mutationCache||new Wer,this.#n=n.defaultOptions||{},this.#r=new Map,this.#i=new Map,this.#a=0}mount(){this.#a++,this.#a===1&&(this.#s=$Ye.subscribe(async n=>{n&&(await this.resumePausedMutations(),this.#t.onFocus())}),this.#o=Zve.subscribe(async n=>{n&&(await this.resumePausedMutations(),this.#t.onOnline())}))}unmount(){this.#a--,this.#a===0&&(this.#s?.(),this.#s=void 0,this.#o?.(),this.#o=void 0)}isFetching(n){return this.#t.findAll({...n,fetchStatus:"fetching"}).length}isMutating(n){return this.#e.findAll({...n,status:"pending"}).length}getQueryData(n){const r=this.defaultQueryOptions({queryKey:n});return this.#t.get(r.queryHash)?.state.data}ensureQueryData(n){const r=this.defaultQueryOptions(n),o=this.#t.build(this,r),s=o.state.data;return s===void 0?this.fetchQuery(n):(n.revalidateIfStale&&o.isStaleByTime(pM(r.staleTime,o))&&this.prefetchQuery(r),Promise.resolve(s))}getQueriesData(n){return this.#t.findAll(n).map(({queryKey:r,state:o})=>{const s=o.data;return[r,s]})}setQueryData(n,r,o){const s=this.defaultQueryOptions({queryKey:n}),d=this.#t.get(s.queryHash)?.state.data,h=Ler(r,d);if(h!==void 0)return this.#t.build(this,s).setData(h,{...o,manual:!0})}setQueriesData(n,r,o){return qg.batch(()=>this.#t.findAll(n).map(({queryKey:s})=>[s,this.setQueryData(s,r,o)]))}getQueryState(n){const r=this.defaultQueryOptions({queryKey:n});return this.#t.get(r.queryHash)?.state}removeQueries(n){const r=this.#t;qg.batch(()=>{r.findAll(n).forEach(o=>{r.remove(o)})})}resetQueries(n,r){const o=this.#t;return qg.batch(()=>(o.findAll(n).forEach(s=>{s.reset()}),this.refetchQueries({type:"active",...n},r)))}cancelQueries(n,r={}){const o={revert:!0,...r},s=qg.batch(()=>this.#t.findAll(n).map(c=>c.cancel(o)));return Promise.all(s).then(ov).catch(ov)}invalidateQueries(n,r={}){return qg.batch(()=>(this.#t.findAll(n).forEach(o=>{o.invalidate()}),n?.refetchType==="none"?Promise.resolve():this.refetchQueries({...n,type:n?.refetchType??n?.type??"active"},r)))}refetchQueries(n,r={}){const o={...r,cancelRefetch:r.cancelRefetch??!0},s=qg.batch(()=>this.#t.findAll(n).filter(c=>!c.isDisabled()&&!c.isStatic()).map(c=>{let d=c.fetch(void 0,o);return o.throwOnError||(d=d.catch(ov)),c.state.fetchStatus==="paused"?Promise.resolve():d}));return Promise.all(s).then(ov)}fetchQuery(n){const r=this.defaultQueryOptions(n);r.retry===void 0&&(r.retry=!1);const o=this.#t.build(this,r);return o.isStaleByTime(pM(r.staleTime,o))?o.fetch(r):Promise.resolve(o.state.data)}prefetchQuery(n){return this.fetchQuery(n).then(ov).catch(ov)}fetchInfiniteQuery(n){return n.behavior=$7t(n.pages),this.fetchQuery(n)}prefetchInfiniteQuery(n){return this.fetchInfiniteQuery(n).then(ov).catch(ov)}ensureInfiniteQueryData(n){return n.behavior=$7t(n.pages),this.ensureQueryData(n)}resumePausedMutations(){return Zve.isOnline()?this.#e.resumePausedMutations():Promise.resolve()}getQueryCache(){return this.#t}getMutationCache(){return this.#e}getDefaultOptions(){return this.#n}setDefaultOptions(n){this.#n=n}setQueryDefaults(n,r){this.#r.set(YB(n),{queryKey:n,defaultOptions:r})}getQueryDefaults(n){const r=[...this.#r.values()],o={};return r.forEach(s=>{$ie(n,s.queryKey)&&Object.assign(o,s.defaultOptions)}),o}setMutationDefaults(n,r){this.#i.set(YB(n),{mutationKey:n,defaultOptions:r})}getMutationDefaults(n){const r=[...this.#i.values()],o={};return r.forEach(s=>{$ie(n,s.mutationKey)&&Object.assign(o,s.defaultOptions)}),o}defaultQueryOptions(n){if(n._defaulted)return n;const r={...this.#n.queries,...this.getQueryDefaults(n.queryKey),...n,_defaulted:!0};return r.queryHash||(r.queryHash=MYe(r.queryKey,r)),r.refetchOnReconnect===void 0&&(r.refetchOnReconnect=r.networkMode!=="always"),r.throwOnError===void 0&&(r.throwOnError=!!r.suspense),!r.networkMode&&r.persister&&(r.networkMode="offlineFirst"),r.queryFn===jYe&&(r.enabled=!1),r}defaultMutationOptions(n){return n?._defaulted?n:{...this.#n.mutations,...n?.mutationKey&&this.getMutationDefaults(n.mutationKey),...n,_defaulted:!0}}clear(){this.#t.clear(),this.#e.clear()}},Xer=class extends TK{constructor(n,r){super(),this.options=r,this.#t=n,this.#o=null,this.#s=gVe(),this.options.experimental_prefetchInRender||this.#s.reject(new Error("experimental_prefetchInRender feature flag is not enabled")),this.bindMethods(),this.setOptions(r)}#t;#e=void 0;#n=void 0;#r=void 0;#i;#a;#s;#o;#g;#f;#h;#u;#l;#c;#p=new Set;bindMethods(){this.refetch=this.refetch.bind(this)}onSubscribe(){this.listeners.size===1&&(this.#e.addObserver(this),B7t(this.#e,this.options)?this.#d():this.updateResult(),this.#w())}onUnsubscribe(){this.hasListeners()||this.destroy()}shouldFetchOnReconnect(){return mVe(this.#e,this.options,this.options.refetchOnReconnect)}shouldFetchOnWindowFocus(){return mVe(this.#e,this.options,this.options.refetchOnWindowFocus)}destroy(){this.listeners=new Set,this.#y(),this.#E(),this.#e.removeObserver(this)}setOptions(n){const r=this.options,o=this.#e;if(this.options=this.#t.defaultQueryOptions(n),this.options.enabled!==void 0&&typeof this.options.enabled!="boolean"&&typeof this.options.enabled!="function"&&typeof N_(this.options.enabled,this.#e)!="boolean")throw new Error("Expected enabled to be a boolean or a callback that returns a boolean");this.#S(),this.#e.setOptions(this.options),r._defaulted&&!Jve(this.options,r)&&this.#t.getQueryCache().notify({type:"observerOptionsUpdated",query:this.#e,observer:this});const s=this.hasListeners();s&&U7t(this.#e,o,this.options,r)&&this.#d(),this.updateResult(),s&&(this.#e!==o||N_(this.options.enabled,this.#e)!==N_(r.enabled,this.#e)||pM(this.options.staleTime,this.#e)!==pM(r.staleTime,this.#e))&&this.#m();const c=this.#b();s&&(this.#e!==o||N_(this.options.enabled,this.#e)!==N_(r.enabled,this.#e)||c!==this.#c)&&this.#v(c)}getOptimisticResult(n){const r=this.#t.getQueryCache().build(this.#t,n),o=this.createResult(r,n);return Zer(this,o)&&(this.#r=o,this.#a=this.options,this.#i=this.#e.state),o}getCurrentResult(){return this.#r}trackResult(n,r){return new Proxy(n,{get:(o,s)=>(this.trackProp(s),r?.(s),Reflect.get(o,s))})}trackProp(n){this.#p.add(n)}getCurrentQuery(){return this.#e}refetch({...n}={}){return this.fetch({...n})}fetchOptimistic(n){const r=this.#t.defaultQueryOptions(n),o=this.#t.getQueryCache().build(this.#t,r);return o.fetch().then(()=>this.createResult(o,r))}fetch(n){return this.#d({...n,cancelRefetch:n.cancelRefetch??!0}).then(()=>(this.updateResult(),this.#r))}#d(n){this.#S();let r=this.#e.fetch(this.options,n);return n?.throwOnError||(r=r.catch(ov)),r}#m(){this.#y();const n=pM(this.options.staleTime,this.#e);if(KB||this.#r.isStale||!fVe(n))return;const o=Uen(this.#r.dataUpdatedAt,n)+1;this.#u=setTimeout(()=>{this.#r.isStale||this.updateResult()},o)}#b(){return(typeof this.options.refetchInterval=="function"?this.options.refetchInterval(this.#e):this.options.refetchInterval)??!1}#v(n){this.#E(),this.#c=n,!(KB||N_(this.options.enabled,this.#e)===!1||!fVe(this.#c)||this.#c===0)&&(this.#l=setInterval(()=>{(this.options.refetchIntervalInBackground||$Ye.isFocused())&&this.#d()},this.#c))}#w(){this.#m(),this.#v(this.#b())}#y(){this.#u&&(clearTimeout(this.#u),this.#u=void 0)}#E(){this.#l&&(clearInterval(this.#l),this.#l=void 0)}createResult(n,r){const o=this.#e,s=this.options,c=this.#r,d=this.#i,h=this.#a,y=n!==o?n.state:this.#n,{state:E}=n;let k={...E},x=!1,O;if(r._optimisticResults){const ye=this.hasListeners(),we=!ye&&B7t(n,r),ge=ye&&U7t(n,o,r,s);(we||ge)&&(k={...k,...Yen(E.data,n.options)}),r._optimisticResults==="isRestoring"&&(k.fetchStatus="idle")}let{error:P,errorUpdatedAt:D,status:$}=k;O=k.data;let U=!1;if(r.placeholderData!==void 0&&O===void 0&&$==="pending"){let ye;c?.isPlaceholderData&&r.placeholderData===h?.placeholderData?(ye=c.data,U=!0):ye=typeof r.placeholderData=="function"?r.placeholderData(this.#h?.state.data,this.#h):r.placeholderData,ye!==void 0&&($="success",O=pVe(c?.data,ye,r),x=!0)}if(r.select&&O!==void 0&&!U)if(c&&O===d?.data&&r.select===this.#g)O=this.#f;else try{this.#g=r.select,O=r.select(O),O=pVe(c?.data,O,r),this.#f=O,this.#o=null}catch(ye){this.#o=ye}this.#o&&(P=this.#o,O=this.#f,D=Date.now(),$="error");const B=k.fetchStatus==="fetching",G=$==="pending",K=$==="error",X=G&&B,ae=O!==void 0,he={status:$,fetchStatus:k.fetchStatus,isPending:G,isSuccess:$==="success",isError:K,isInitialLoading:X,isLoading:X,data:O,dataUpdatedAt:k.dataUpdatedAt,error:P,errorUpdatedAt:D,failureCount:k.fetchFailureCount,failureReason:k.fetchFailureReason,errorUpdateCount:k.errorUpdateCount,isFetched:k.dataUpdateCount>0||k.errorUpdateCount>0,isFetchedAfterMount:k.dataUpdateCount>y.dataUpdateCount||k.errorUpdateCount>y.errorUpdateCount,isFetching:B,isRefetching:B&&!G,isLoadingError:K&&!ae,isPaused:k.fetchStatus==="paused",isPlaceholderData:x,isRefetchError:K&&ae,isStale:FYe(n,r),refetch:this.refetch,promise:this.#s,isEnabled:N_(r.enabled,n)!==!1};if(this.options.experimental_prefetchInRender){const ye=Te=>{he.status==="error"?Te.reject(he.error):he.data!==void 0&&Te.resolve(he.data)},we=()=>{const Te=this.#s=he.promise=gVe();ye(Te)},ge=this.#s;switch(ge.status){case"pending":n.queryHash===o.queryHash&&ye(ge);break;case"fulfilled":(he.status==="error"||he.data!==ge.value)&&we();break;case"rejected":(he.status!=="error"||he.error!==ge.reason)&&we();break}}return he}updateResult(){const n=this.#r,r=this.createResult(this.#e,this.options);if(this.#i=this.#e.state,this.#a=this.options,this.#i.data!==void 0&&(this.#h=this.#e),Jve(r,n))return;this.#r=r;const o=()=>{if(!n)return!0;const{notifyOnChangeProps:s}=this.options,c=typeof s=="function"?s():s;if(c==="all"||!c&&!this.#p.size)return!0;const d=new Set(c??this.#p);return this.options.throwOnError&&d.add("error"),Object.keys(this.#r).some(h=>{const b=h;return this.#r[b]!==n[b]&&d.has(b)})};this.#_({listeners:o()})}#S(){const n=this.#t.getQueryCache().build(this.#t,this.options);if(n===this.#e)return;const r=this.#e;this.#e=n,this.#n=n.state,this.hasListeners()&&(r?.removeObserver(this),n.addObserver(this))}onQueryUpdate(){this.updateResult(),this.hasListeners()&&this.#w()}#_(n){qg.batch(()=>{n.listeners&&this.listeners.forEach(r=>{r(this.#r)}),this.#t.getQueryCache().notify({query:this.#e,type:"observerResultsUpdated"})})}};function Jer(n,r){return N_(r.enabled,n)!==!1&&n.state.data===void 0&&!(n.state.status==="error"&&r.retryOnMount===!1)}function B7t(n,r){return Jer(n,r)||n.state.data!==void 0&&mVe(n,r,r.refetchOnMount)}function mVe(n,r,o){if(N_(r.enabled,n)!==!1&&pM(r.staleTime,n)!=="static"){const s=typeof o=="function"?o(n):o;return s==="always"||s!==!1&&FYe(n,r)}return!1}function U7t(n,r,o,s){return(n!==r||N_(s.enabled,n)===!1)&&(!o.suspense||n.state.status!=="error")&&FYe(n,o)}function FYe(n,r){return N_(r.enabled,n)!==!1&&n.isStaleByTime(pM(r.staleTime,n))}function Zer(n,r){return!Jve(n.getCurrentResult(),r)}var Qer=class extends TK{#t;#e=void 0;#n;#r;constructor(r,o){super(),this.#t=r,this.setOptions(o),this.bindMethods(),this.#i()}bindMethods(){this.mutate=this.mutate.bind(this),this.reset=this.reset.bind(this)}setOptions(r){const o=this.options;this.options=this.#t.defaultMutationOptions(r),Jve(this.options,o)||this.#t.getMutationCache().notify({type:"observerOptionsUpdated",mutation:this.#n,observer:this}),o?.mutationKey&&this.options.mutationKey&&YB(o.mutationKey)!==YB(this.options.mutationKey)?this.reset():this.#n?.state.status==="pending"&&this.#n.setOptions(this.options)}onUnsubscribe(){this.hasListeners()||this.#n?.removeObserver(this)}onMutationUpdate(r){this.#i(),this.#a(r)}getCurrentResult(){return this.#e}reset(){this.#n?.removeObserver(this),this.#n=void 0,this.#i(),this.#a()}mutate(r,o){return this.#r=o,this.#n?.removeObserver(this),this.#n=this.#t.getMutationCache().build(this.#t,this.options),this.#n.addObserver(this),this.#n.execute(r)}#i(){const r=this.#n?.state??Xen();this.#e={...r,isPending:r.status==="pending",isSuccess:r.status==="success",isError:r.status==="error",isIdle:r.status==="idle",mutate:this.mutate,reset:this.reset}}#a(r){qg.batch(()=>{if(this.#r&&this.hasListeners()){const o=this.#e.variables,s=this.#e.context;r?.type==="success"?(this.#r.onSuccess?.(r.data,o,s),this.#r.onSettled?.(r.data,null,o,s)):r?.type==="error"&&(this.#r.onError?.(r.error,o,s),this.#r.onSettled?.(void 0,r.error,o,s))}this.listeners.forEach(o=>{o(this.#e)})})}},Jen=Q.createContext(void 0),Roe=n=>{const r=Q.useContext(Jen);if(!r)throw new Error("No QueryClient set, use QueryClientProvider to set one");return r},etr=({client:n,children:r})=>(Q.useEffect(()=>(n.mount(),()=>{n.unmount()}),[n]),H.jsx(Jen.Provider,{value:n,children:r})),Zen=Q.createContext(!1),ttr=()=>Q.useContext(Zen);Zen.Provider;function ntr(){let n=!1;return{clearReset:()=>{n=!1},reset:()=>{n=!0},isReset:()=>n}}var rtr=Q.createContext(ntr()),itr=()=>Q.useContext(rtr),otr=(n,r)=>{(n.suspense||n.throwOnError||n.experimental_prefetchInRender)&&(r.isReset()||(n.retryOnMount=!1))},atr=n=>{Q.useEffect(()=>{n.clearReset()},[n])},str=({result:n,errorResetBoundary:r,throwOnError:o,query:s,suspense:c})=>n.isError&&!r.isReset()&&!n.isFetching&&s&&(c&&n.data===void 0||Gen(o,[n.error,s])),ctr=n=>{if(n.suspense){const r=s=>s==="static"?s:Math.max(s??1e3,1e3),o=n.staleTime;n.staleTime=typeof o=="function"?(...s)=>r(o(...s)):r(o),typeof n.gcTime=="number"&&(n.gcTime=Math.max(n.gcTime,1e3))}},utr=(n,r)=>n.isLoading&&n.isFetching&&!r,ltr=(n,r)=>n?.suspense&&r.isPending,z7t=(n,r,o)=>r.fetchOptimistic(n).catch(()=>{o.clearReset()});function dtr(n,r,o){const s=ttr(),c=itr(),d=Roe(),h=d.defaultQueryOptions(n);d.getDefaultOptions().queries?._experimental_beforeQuery?.(h),h._optimisticResults=s?"isRestoring":"optimistic",ctr(h),otr(h,c),atr(c);const b=!d.getQueryCache().get(h.queryHash),[y]=Q.useState(()=>new r(d,h)),E=y.getOptimisticResult(h),k=!s&&n.subscribed!==!1;if(Q.useSyncExternalStore(Q.useCallback(x=>{const O=k?y.subscribe(qg.batchCalls(x)):ov;return y.updateResult(),O},[y,k]),()=>y.getCurrentResult(),()=>y.getCurrentResult()),Q.useEffect(()=>{y.setOptions(h)},[h,y]),ltr(h,E))throw z7t(h,y,c);if(str({result:E,errorResetBoundary:c,throwOnError:h.throwOnError,query:d.getQueryCache().get(h.queryHash),suspense:h.suspense}))throw E.error;return d.getDefaultOptions().queries?._experimental_afterQuery?.(h,E),h.experimental_prefetchInRender&&!KB&&utr(E,s)&&(b?z7t(h,y,c):d.getQueryCache().get(h.queryHash)?.promise)?.catch(ov).finally(()=>{y.updateResult()}),h.notifyOnChangeProps?E:y.trackResult(E)}function xK(n,r){return dtr(n,Xer)}function BYe(n,r){const o=Roe(),[s]=Q.useState(()=>new Qer(o,n));Q.useEffect(()=>{s.setOptions(n)},[s,n]);const c=Q.useSyncExternalStore(Q.useCallback(h=>s.subscribe(qg.batchCalls(h)),[s]),()=>s.getCurrentResult(),()=>s.getCurrentResult()),d=Q.useCallback((h,b)=>{s.mutate(h,b).catch(ov)},[s]);if(c.error&&Gen(s.options.throwOnError,[c.error]))throw c.error;return{...c,mutate:d,mutateAsync:c.mutate}}const ftr="modulepreload",htr=function(n){return"/"+n},H7t={},Ix=function(r,o,s){let c=Promise.resolve();if(o&&o.length>0){let y=function(E){return Promise.all(E.map(k=>Promise.resolve(k).then(x=>({status:"fulfilled",value:x}),x=>({status:"rejected",reason:x}))))};document.getElementsByTagName("link");const h=document.querySelector("meta[property=csp-nonce]"),b=h?.nonce||h?.getAttribute("nonce");c=y(o.map(E=>{if(E=htr(E),E in H7t)return;H7t[E]=!0;const k=E.endsWith(".css"),x=k?'[rel="stylesheet"]':"";if(document.querySelector(`link[href="${E}"]${x}`))return;const O=document.createElement("link");if(O.rel=k?"stylesheet":ftr,k||(O.as="script"),O.crossOrigin="",O.href=E,b&&O.setAttribute("nonce",b),document.head.appendChild(O),k)return new Promise((P,D)=>{O.addEventListener("load",P),O.addEventListener("error",()=>D(new Error(`Unable to preload CSS for ${E}`)))})}))}function d(h){const b=new Event("vite:preloadError",{cancelable:!0});if(b.payload=h,window.dispatchEvent(b),!b.defaultPrevented)throw h}return c.then(h=>{for(const b of h||[])b.status==="rejected"&&d(b.reason);return r().catch(d)})};var ptr=function(){return null};/**
 * react-router v7.7.1
 *
 * Copyright (c) Remix Software Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE.md file in the root directory of this source tree.
 *
 * @license MIT
 */var G7t="popstate";function gtr(n={}){function r(s,c){let{pathname:d,search:h,hash:b}=s.location;return bVe("",{pathname:d,search:h,hash:b},c.state&&c.state.usr||null,c.state&&c.state.key||"default")}function o(s,c){return typeof c=="string"?c:Fie(c)}return btr(r,o,null,n)}function Tf(n,r){if(n===!1||n===null||typeof n>"u")throw new Error(r)}function U_(n,r){if(!n){typeof console<"u"&&console.warn(r);try{throw new Error(r)}catch{}}}function mtr(){return Math.random().toString(36).substring(2,10)}function q7t(n,r){return{usr:n.state,key:n.key,idx:r}}function bVe(n,r,o=null,s){return{pathname:typeof n=="string"?n:n.pathname,search:"",hash:"",...typeof r=="string"?CK(r):r,state:o,key:r&&r.key||s||mtr()}}function Fie({pathname:n="/",search:r="",hash:o=""}){return r&&r!=="?"&&(n+=r.charAt(0)==="?"?r:"?"+r),o&&o!=="#"&&(n+=o.charAt(0)==="#"?o:"#"+o),n}function CK(n){let r={};if(n){let o=n.indexOf("#");o>=0&&(r.hash=n.substring(o),n=n.substring(0,o));let s=n.indexOf("?");s>=0&&(r.search=n.substring(s),n=n.substring(0,s)),n&&(r.pathname=n)}return r}function btr(n,r,o,s={}){let{window:c=document.defaultView,v5Compat:d=!1}=s,h=c.history,b="POP",y=null,E=k();E==null&&(E=0,h.replaceState({...h.state,idx:E},""));function k(){return(h.state||{idx:null}).idx}function x(){b="POP";let U=k(),B=U==null?null:U-E;E=U,y&&y({action:b,location:$.location,delta:B})}function O(U,B){b="PUSH";let G=bVe($.location,U,B);E=k()+1;let K=q7t(G,E),X=$.createHref(G);try{h.pushState(K,"",X)}catch(ae){if(ae instanceof DOMException&&ae.name==="DataCloneError")throw ae;c.location.assign(X)}d&&y&&y({action:b,location:$.location,delta:1})}function P(U,B){b="REPLACE";let G=bVe($.location,U,B);E=k();let K=q7t(G,E),X=$.createHref(G);h.replaceState(K,"",X),d&&y&&y({action:b,location:$.location,delta:0})}function D(U){return vtr(U)}let $={get action(){return b},get location(){return n(c,h)},listen(U){if(y)throw new Error("A history only accepts one active listener");return c.addEventListener(G7t,x),y=U,()=>{c.removeEventListener(G7t,x),y=null}},createHref(U){return r(c,U)},createURL:D,encodeLocation(U){let B=D(U);return{pathname:B.pathname,search:B.search,hash:B.hash}},push:O,replace:P,go(U){return h.go(U)}};return $}function vtr(n,r=!1){let o="http://localhost";typeof window<"u"&&(o=window.location.origin!=="null"?window.location.origin:window.location.href),Tf(o,"No window.location.(origin|href) available to create URL");let s=typeof n=="string"?n:Fie(n);return s=s.replace(/ $/,"%20"),!r&&s.startsWith("//")&&(s=o+s),new URL(s,o)}function Qen(n,r,o="/"){return wtr(n,r,o,!1)}function wtr(n,r,o,s){let c=typeof r=="string"?CK(r):r,d=U4(c.pathname||"/",o);if(d==null)return null;let h=etn(n);ytr(h);let b=null;for(let y=0;b==null&&y<h.length;++y){let E=Ntr(d);b=Itr(h[y],E,s)}return b}function etn(n,r=[],o=[],s=""){let c=(d,h,b)=>{let y={relativePath:b===void 0?d.path||"":b,caseSensitive:d.caseSensitive===!0,childrenIndex:h,route:d};y.relativePath.startsWith("/")&&(Tf(y.relativePath.startsWith(s),`Absolute route path "${y.relativePath}" nested under path "${s}" is not valid. An absolute child route path must start with the combined path of all its parent routes.`),y.relativePath=y.relativePath.slice(s.length));let E=j4([s,y.relativePath]),k=o.concat(y);d.children&&d.children.length>0&&(Tf(d.index!==!0,`Index routes must not have child routes. Please remove all child routes from route path "${E}".`),etn(d.children,r,k,E)),!(d.path==null&&!d.index)&&r.push({path:E,score:Ctr(E,d.index),routesMeta:k})};return n.forEach((d,h)=>{if(d.path===""||!d.path?.includes("?"))c(d,h);else for(let b of ttn(d.path))c(d,h,b)}),r}function ttn(n){let r=n.split("/");if(r.length===0)return[];let[o,...s]=r,c=o.endsWith("?"),d=o.replace(/\?$/,"");if(s.length===0)return c?[d,""]:[d];let h=ttn(s.join("/")),b=[];return b.push(...h.map(y=>y===""?d:[d,y].join("/"))),c&&b.push(...h),b.map(y=>n.startsWith("/")&&y===""?"/":y)}function ytr(n){n.sort((r,o)=>r.score!==o.score?o.score-r.score:Atr(r.routesMeta.map(s=>s.childrenIndex),o.routesMeta.map(s=>s.childrenIndex)))}var Etr=/^:[\w-]+$/,Str=3,_tr=2,ktr=1,Ttr=10,xtr=-2,V7t=n=>n==="*";function Ctr(n,r){let o=n.split("/"),s=o.length;return o.some(V7t)&&(s+=xtr),r&&(s+=_tr),o.filter(c=>!V7t(c)).reduce((c,d)=>c+(Etr.test(d)?Str:d===""?ktr:Ttr),s)}function Atr(n,r){return n.length===r.length&&n.slice(0,-1).every((s,c)=>s===r[c])?n[n.length-1]-r[r.length-1]:0}function Itr(n,r,o=!1){let{routesMeta:s}=n,c={},d="/",h=[];for(let b=0;b<s.length;++b){let y=s[b],E=b===s.length-1,k=d==="/"?r:r.slice(d.length)||"/",x=Qve({path:y.relativePath,caseSensitive:y.caseSensitive,end:E},k),O=y.route;if(!x&&E&&o&&!s[s.length-1].route.index&&(x=Qve({path:y.relativePath,caseSensitive:y.caseSensitive,end:!1},k)),!x)return null;Object.assign(c,x.params),h.push({params:c,pathname:j4([d,x.pathname]),pathnameBase:Dtr(j4([d,x.pathnameBase])),route:O}),x.pathnameBase!=="/"&&(d=j4([d,x.pathnameBase]))}return h}function Qve(n,r){typeof n=="string"&&(n={path:n,caseSensitive:!1,end:!0});let[o,s]=Otr(n.path,n.caseSensitive,n.end),c=r.match(o);if(!c)return null;let d=c[0],h=d.replace(/(.)\/+$/,"$1"),b=c.slice(1);return{params:s.reduce((E,{paramName:k,isOptional:x},O)=>{if(k==="*"){let D=b[O]||"";h=d.slice(0,d.length-D.length).replace(/(.)\/+$/,"$1")}const P=b[O];return x&&!P?E[k]=void 0:E[k]=(P||"").replace(/%2F/g,"/"),E},{}),pathname:d,pathnameBase:h,pattern:n}}function Otr(n,r=!1,o=!0){U_(n==="*"||!n.endsWith("*")||n.endsWith("/*"),`Route path "${n}" will be treated as if it were "${n.replace(/\*$/,"/*")}" because the \`*\` character must always follow a \`/\` in the pattern. To get rid of this warning, please change the route path to "${n.replace(/\*$/,"/*")}".`);let s=[],c="^"+n.replace(/\/*\*?$/,"").replace(/^\/*/,"/").replace(/[\\.*+^${}|()[\]]/g,"\\$&").replace(/\/:([\w-]+)(\?)?/g,(h,b,y)=>(s.push({paramName:b,isOptional:y!=null}),y?"/?([^\\/]+)?":"/([^\\/]+)"));return n.endsWith("*")?(s.push({paramName:"*"}),c+=n==="*"||n==="/*"?"(.*)$":"(?:\\/(.+)|\\/*)$"):o?c+="\\/*$":n!==""&&n!=="/"&&(c+="(?:(?=\\/|$))"),[new RegExp(c,r?void 0:"i"),s]}function Ntr(n){try{return n.split("/").map(r=>decodeURIComponent(r).replace(/\//g,"%2F")).join("/")}catch(r){return U_(!1,`The URL path "${n}" could not be decoded because it is a malformed URL segment. This is probably due to a bad percent encoding (${r}).`),n}}function U4(n,r){if(r==="/")return n;if(!n.toLowerCase().startsWith(r.toLowerCase()))return null;let o=r.endsWith("/")?r.length-1:r.length,s=n.charAt(o);return s&&s!=="/"?null:n.slice(o)||"/"}function Rtr(n,r="/"){let{pathname:o,search:s="",hash:c=""}=typeof n=="string"?CK(n):n;return{pathname:o?o.startsWith("/")?o:Ptr(o,r):r,search:Mtr(s),hash:jtr(c)}}function Ptr(n,r){let o=r.replace(/\/+$/,"").split("/");return n.split("/").forEach(c=>{c===".."?o.length>1&&o.pop():c!=="."&&o.push(c)}),o.length>1?o.join("/"):"/"}function u$e(n,r,o,s){return`Cannot include a '${n}' character in a manually specified \`to.${r}\` field [${JSON.stringify(s)}].  Please separate it out to the \`to.${o}\` field. Alternatively you may provide the full path as a string in <Link to="..."> and the router will parse it for you.`}function Ltr(n){return n.filter((r,o)=>o===0||r.route.path&&r.route.path.length>0)}function UYe(n){let r=Ltr(n);return r.map((o,s)=>s===r.length-1?o.pathname:o.pathnameBase)}function zYe(n,r,o,s=!1){let c;typeof n=="string"?c=CK(n):(c={...n},Tf(!c.pathname||!c.pathname.includes("?"),u$e("?","pathname","search",c)),Tf(!c.pathname||!c.pathname.includes("#"),u$e("#","pathname","hash",c)),Tf(!c.search||!c.search.includes("#"),u$e("#","search","hash",c)));let d=n===""||c.pathname==="",h=d?"/":c.pathname,b;if(h==null)b=o;else{let x=r.length-1;if(!s&&h.startsWith("..")){let O=h.split("/");for(;O[0]==="..";)O.shift(),x-=1;c.pathname=O.join("/")}b=x>=0?r[x]:"/"}let y=Rtr(c,b),E=h&&h!=="/"&&h.endsWith("/"),k=(d||h===".")&&o.endsWith("/");return!y.pathname.endsWith("/")&&(E||k)&&(y.pathname+="/"),y}var j4=n=>n.join("/").replace(/\/\/+/g,"/"),Dtr=n=>n.replace(/\/+$/,"").replace(/^\/*/,"/"),Mtr=n=>!n||n==="?"?"":n.startsWith("?")?n:"?"+n,jtr=n=>!n||n==="#"?"":n.startsWith("#")?n:"#"+n;function $tr(n){return n!=null&&typeof n.status=="number"&&typeof n.statusText=="string"&&typeof n.internal=="boolean"&&"data"in n}var ntn=["POST","PUT","PATCH","DELETE"];new Set(ntn);var Ftr=["GET",...ntn];new Set(Ftr);var AK=Q.createContext(null);AK.displayName="DataRouter";var Pye=Q.createContext(null);Pye.displayName="DataRouterState";Q.createContext(!1);var rtn=Q.createContext({isTransitioning:!1});rtn.displayName="ViewTransition";var Btr=Q.createContext(new Map);Btr.displayName="Fetchers";var Utr=Q.createContext(null);Utr.displayName="Await";var Zx=Q.createContext(null);Zx.displayName="Navigation";var Poe=Q.createContext(null);Poe.displayName="Location";var Qx=Q.createContext({outlet:null,matches:[],isDataRoute:!1});Qx.displayName="Route";var HYe=Q.createContext(null);HYe.displayName="RouteError";function ztr(n,{relative:r}={}){Tf(IK(),"useHref() may be used only in the context of a <Router> component.");let{basename:o,navigator:s}=Q.useContext(Zx),{hash:c,pathname:d,search:h}=Loe(n,{relative:r}),b=d;return o!=="/"&&(b=d==="/"?o:j4([o,d])),s.createHref({pathname:b,search:h,hash:c})}function IK(){return Q.useContext(Poe)!=null}function zI(){return Tf(IK(),"useLocation() may be used only in the context of a <Router> component."),Q.useContext(Poe).location}var itn="You should call navigate() in a React.useEffect(), not when your component is first rendered.";function otn(n){Q.useContext(Zx).static||Q.useLayoutEffect(n)}function y2(){let{isDataRoute:n}=Q.useContext(Qx);return n?tnr():Htr()}function Htr(){Tf(IK(),"useNavigate() may be used only in the context of a <Router> component.");let n=Q.useContext(AK),{basename:r,navigator:o}=Q.useContext(Zx),{matches:s}=Q.useContext(Qx),{pathname:c}=zI(),d=JSON.stringify(UYe(s)),h=Q.useRef(!1);return otn(()=>{h.current=!0}),Q.useCallback((y,E={})=>{if(U_(h.current,itn),!h.current)return;if(typeof y=="number"){o.go(y);return}let k=zYe(y,JSON.parse(d),c,E.relative==="path");n==null&&r!=="/"&&(k.pathname=k.pathname==="/"?r:j4([r,k.pathname])),(E.replace?o.replace:o.push)(k,E.state,E)},[r,o,d,c,n])}Q.createContext(null);function G_(){let{matches:n}=Q.useContext(Qx),r=n[n.length-1];return r?r.params:{}}function Loe(n,{relative:r}={}){let{matches:o}=Q.useContext(Qx),{pathname:s}=zI(),c=JSON.stringify(UYe(o));return Q.useMemo(()=>zYe(n,JSON.parse(c),s,r==="path"),[n,c,s,r])}function Gtr(n,r){return atn(n,r)}function atn(n,r,o,s){Tf(IK(),"useRoutes() may be used only in the context of a <Router> component.");let{navigator:c}=Q.useContext(Zx),{matches:d}=Q.useContext(Qx),h=d[d.length-1],b=h?h.params:{},y=h?h.pathname:"/",E=h?h.pathnameBase:"/",k=h&&h.route;{let B=k&&k.path||"";stn(y,!k||B.endsWith("*")||B.endsWith("*?"),`You rendered descendant <Routes> (or called \`useRoutes()\`) at "${y}" (under <Route path="${B}">) but the parent route path has no trailing "*". This means if you navigate deeper, the parent won't match anymore and therefore the child routes will never render.

Please change the parent <Route path="${B}"> to <Route path="${B==="/"?"*":`${B}/*`}">.`)}let x=zI(),O;if(r){let B=typeof r=="string"?CK(r):r;Tf(E==="/"||B.pathname?.startsWith(E),`When overriding the location using \`<Routes location>\` or \`useRoutes(routes, location)\`, the location pathname must begin with the portion of the URL pathname that was matched by all parent routes. The current pathname base is "${E}" but pathname "${B.pathname}" was given in the \`location\` prop.`),O=B}else O=x;let P=O.pathname||"/",D=P;if(E!=="/"){let B=E.replace(/^\//,"").split("/");D="/"+P.replace(/^\//,"").split("/").slice(B.length).join("/")}let $=Qen(n,{pathname:D});U_(k||$!=null,`No routes matched location "${O.pathname}${O.search}${O.hash}" `),U_($==null||$[$.length-1].route.element!==void 0||$[$.length-1].route.Component!==void 0||$[$.length-1].route.lazy!==void 0,`Matched leaf route at location "${O.pathname}${O.search}${O.hash}" does not have an element or Component. This means it will render an <Outlet /> with a null value by default resulting in an "empty" page.`);let U=Ytr($&&$.map(B=>Object.assign({},B,{params:Object.assign({},b,B.params),pathname:j4([E,c.encodeLocation?c.encodeLocation(B.pathname).pathname:B.pathname]),pathnameBase:B.pathnameBase==="/"?E:j4([E,c.encodeLocation?c.encodeLocation(B.pathnameBase).pathname:B.pathnameBase])})),d,o,s);return r&&U?Q.createElement(Poe.Provider,{value:{location:{pathname:"/",search:"",hash:"",state:null,key:"default",...O},navigationType:"POP"}},U):U}function qtr(){let n=enr(),r=$tr(n)?`${n.status} ${n.statusText}`:n instanceof Error?n.message:JSON.stringify(n),o=n instanceof Error?n.stack:null,s="rgba(200,200,200, 0.5)",c={padding:"0.5rem",backgroundColor:s},d={padding:"2px 4px",backgroundColor:s},h=null;return console.error("Error handled by React Router default ErrorBoundary:",n),h=Q.createElement(Q.Fragment,null,Q.createElement("p",null,"💿 Hey developer 👋"),Q.createElement("p",null,"You can provide a way better UX than this when your app throws errors by providing your own ",Q.createElement("code",{style:d},"ErrorBoundary")," or"," ",Q.createElement("code",{style:d},"errorElement")," prop on your route.")),Q.createElement(Q.Fragment,null,Q.createElement("h2",null,"Unexpected Application Error!"),Q.createElement("h3",{style:{fontStyle:"italic"}},r),o?Q.createElement("pre",{style:c},o):null,h)}var Vtr=Q.createElement(qtr,null),Wtr=class extends Q.Component{constructor(n){super(n),this.state={location:n.location,revalidation:n.revalidation,error:n.error}}static getDerivedStateFromError(n){return{error:n}}static getDerivedStateFromProps(n,r){return r.location!==n.location||r.revalidation!=="idle"&&n.revalidation==="idle"?{error:n.error,location:n.location,revalidation:n.revalidation}:{error:n.error!==void 0?n.error:r.error,location:r.location,revalidation:n.revalidation||r.revalidation}}componentDidCatch(n,r){console.error("React Router caught the following error during render",n,r)}render(){return this.state.error!==void 0?Q.createElement(Qx.Provider,{value:this.props.routeContext},Q.createElement(HYe.Provider,{value:this.state.error,children:this.props.component})):this.props.children}};function Ktr({routeContext:n,match:r,children:o}){let s=Q.useContext(AK);return s&&s.static&&s.staticContext&&(r.route.errorElement||r.route.ErrorBoundary)&&(s.staticContext._deepestRenderedBoundaryId=r.route.id),Q.createElement(Qx.Provider,{value:n},o)}function Ytr(n,r=[],o=null,s=null){if(n==null){if(!o)return null;if(o.errors)n=o.matches;else if(r.length===0&&!o.initialized&&o.matches.length>0)n=o.matches;else return null}let c=n,d=o?.errors;if(d!=null){let y=c.findIndex(E=>E.route.id&&d?.[E.route.id]!==void 0);Tf(y>=0,`Could not find a matching route for errors on route IDs: ${Object.keys(d).join(",")}`),c=c.slice(0,Math.min(c.length,y+1))}let h=!1,b=-1;if(o)for(let y=0;y<c.length;y++){let E=c[y];if((E.route.HydrateFallback||E.route.hydrateFallbackElement)&&(b=y),E.route.id){let{loaderData:k,errors:x}=o,O=E.route.loader&&!k.hasOwnProperty(E.route.id)&&(!x||x[E.route.id]===void 0);if(E.route.lazy||O){h=!0,b>=0?c=c.slice(0,b+1):c=[c[0]];break}}}return c.reduceRight((y,E,k)=>{let x,O=!1,P=null,D=null;o&&(x=d&&E.route.id?d[E.route.id]:void 0,P=E.route.errorElement||Vtr,h&&(b<0&&k===0?(stn("route-fallback",!1,"No `HydrateFallback` element provided to render during initial hydration"),O=!0,D=null):b===k&&(O=!0,D=E.route.hydrateFallbackElement||null)));let $=r.concat(c.slice(0,k+1)),U=()=>{let B;return x?B=P:O?B=D:E.route.Component?B=Q.createElement(E.route.Component,null):E.route.element?B=E.route.element:B=y,Q.createElement(Ktr,{match:E,routeContext:{outlet:y,matches:$,isDataRoute:o!=null},children:B})};return o&&(E.route.ErrorBoundary||E.route.errorElement||k===0)?Q.createElement(Wtr,{location:o.location,revalidation:o.revalidation,component:P,error:x,children:U(),routeContext:{outlet:null,matches:$,isDataRoute:!0}}):U()},null)}function GYe(n){return`${n} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function Xtr(n){let r=Q.useContext(AK);return Tf(r,GYe(n)),r}function Jtr(n){let r=Q.useContext(Pye);return Tf(r,GYe(n)),r}function Ztr(n){let r=Q.useContext(Qx);return Tf(r,GYe(n)),r}function qYe(n){let r=Ztr(n),o=r.matches[r.matches.length-1];return Tf(o.route.id,`${n} can only be used on routes that contain a unique "id"`),o.route.id}function Qtr(){return qYe("useRouteId")}function enr(){let n=Q.useContext(HYe),r=Jtr("useRouteError"),o=qYe("useRouteError");return n!==void 0?n:r.errors?.[o]}function tnr(){let{router:n}=Xtr("useNavigate"),r=qYe("useNavigate"),o=Q.useRef(!1);return otn(()=>{o.current=!0}),Q.useCallback(async(c,d={})=>{U_(o.current,itn),o.current&&(typeof c=="number"?n.navigate(c):await n.navigate(c,{fromRouteId:r,...d}))},[n,r])}var W7t={};function stn(n,r,o){!r&&!W7t[n]&&(W7t[n]=!0,U_(!1,o))}Q.memo(nnr);function nnr({routes:n,future:r,state:o}){return atn(n,void 0,o,r)}function K7t({to:n,replace:r,state:o,relative:s}){Tf(IK(),"<Navigate> may be used only in the context of a <Router> component.");let{static:c}=Q.useContext(Zx);U_(!c,"<Navigate> must not be used on the initial render in a <StaticRouter>. This is a no-op, but you should modify your code so the <Navigate> is only ever rendered in response to some user interaction or state change.");let{matches:d}=Q.useContext(Qx),{pathname:h}=zI(),b=y2(),y=zYe(n,UYe(d),h,s==="path"),E=JSON.stringify(y);return Q.useEffect(()=>{b(JSON.parse(E),{replace:r,state:o,relative:s})},[b,E,s,r,o]),null}function SB(n){Tf(!1,"A <Route> is only ever to be used as the child of <Routes> element, never rendered directly. Please wrap your <Route> in a <Routes>.")}function rnr({basename:n="/",children:r=null,location:o,navigationType:s="POP",navigator:c,static:d=!1}){Tf(!IK(),"You cannot render a <Router> inside another <Router>. You should never have more than one in your app.");let h=n.replace(/^\/*/,"/"),b=Q.useMemo(()=>({basename:h,navigator:c,static:d,future:{}}),[h,c,d]);typeof o=="string"&&(o=CK(o));let{pathname:y="/",search:E="",hash:k="",state:x=null,key:O="default"}=o,P=Q.useMemo(()=>{let D=U4(y,h);return D==null?null:{location:{pathname:D,search:E,hash:k,state:x,key:O},navigationType:s}},[h,y,E,k,x,O,s]);return U_(P!=null,`<Router basename="${h}"> is not able to match the URL "${y}${E}${k}" because it does not start with the basename, so the <Router> won't render anything.`),P==null?null:Q.createElement(Zx.Provider,{value:b},Q.createElement(Poe.Provider,{children:r,value:P}))}function inr({children:n,location:r}){return Gtr(vVe(n),r)}function vVe(n,r=[]){let o=[];return Q.Children.forEach(n,(s,c)=>{if(!Q.isValidElement(s))return;let d=[...r,c];if(s.type===Q.Fragment){o.push.apply(o,vVe(s.props.children,d));return}Tf(s.type===SB,`[${typeof s.type=="string"?s.type:s.type.name}] is not a <Route> component. All component children of <Routes> must be a <Route> or <React.Fragment>`),Tf(!s.props.index||!s.props.children,"An index route cannot have child routes.");let h={id:s.props.id||d.join("-"),caseSensitive:s.props.caseSensitive,element:s.props.element,Component:s.props.Component,index:s.props.index,path:s.props.path,loader:s.props.loader,action:s.props.action,hydrateFallbackElement:s.props.hydrateFallbackElement,HydrateFallback:s.props.HydrateFallback,errorElement:s.props.errorElement,ErrorBoundary:s.props.ErrorBoundary,hasErrorBoundary:s.props.hasErrorBoundary===!0||s.props.ErrorBoundary!=null||s.props.errorElement!=null,shouldRevalidate:s.props.shouldRevalidate,handle:s.props.handle,lazy:s.props.lazy};s.props.children&&(h.children=vVe(s.props.children,d)),o.push(h)}),o}var uve="get",lve="application/x-www-form-urlencoded";function Lye(n){return n!=null&&typeof n.tagName=="string"}function onr(n){return Lye(n)&&n.tagName.toLowerCase()==="button"}function anr(n){return Lye(n)&&n.tagName.toLowerCase()==="form"}function snr(n){return Lye(n)&&n.tagName.toLowerCase()==="input"}function cnr(n){return!!(n.metaKey||n.altKey||n.ctrlKey||n.shiftKey)}function unr(n,r){return n.button===0&&(!r||r==="_self")&&!cnr(n)}function wVe(n=""){return new URLSearchParams(typeof n=="string"||Array.isArray(n)||n instanceof URLSearchParams?n:Object.keys(n).reduce((r,o)=>{let s=n[o];return r.concat(Array.isArray(s)?s.map(c=>[o,c]):[[o,s]])},[]))}function lnr(n,r){let o=wVe(n);return r&&r.forEach((s,c)=>{o.has(c)||r.getAll(c).forEach(d=>{o.append(c,d)})}),o}var J1e=null;function dnr(){if(J1e===null)try{new FormData(document.createElement("form"),0),J1e=!1}catch{J1e=!0}return J1e}var fnr=new Set(["application/x-www-form-urlencoded","multipart/form-data","text/plain"]);function l$e(n){return n!=null&&!fnr.has(n)?(U_(!1,`"${n}" is not a valid \`encType\` for \`<Form>\`/\`<fetcher.Form>\` and will default to "${lve}"`),null):n}function hnr(n,r){let o,s,c,d,h;if(anr(n)){let b=n.getAttribute("action");s=b?U4(b,r):null,o=n.getAttribute("method")||uve,c=l$e(n.getAttribute("enctype"))||lve,d=new FormData(n)}else if(onr(n)||snr(n)&&(n.type==="submit"||n.type==="image")){let b=n.form;if(b==null)throw new Error('Cannot submit a <button> or <input type="submit"> without a <form>');let y=n.getAttribute("formaction")||b.getAttribute("action");if(s=y?U4(y,r):null,o=n.getAttribute("formmethod")||b.getAttribute("method")||uve,c=l$e(n.getAttribute("formenctype"))||l$e(b.getAttribute("enctype"))||lve,d=new FormData(b,n),!dnr()){let{name:E,type:k,value:x}=n;if(k==="image"){let O=E?`${E}.`:"";d.append(`${O}x`,"0"),d.append(`${O}y`,"0")}else E&&d.append(E,x)}}else{if(Lye(n))throw new Error('Cannot submit element that is not <form>, <button>, or <input type="submit|image">');o=uve,s=null,c=lve,h=n}return d&&c==="text/plain"&&(h=d,d=void 0),{action:s,method:o.toLowerCase(),encType:c,formData:d,body:h}}Object.getOwnPropertyNames(Object.prototype).sort().join("\0");function VYe(n,r){if(n===!1||n===null||typeof n>"u")throw new Error(r)}function pnr(n,r,o){let s=typeof n=="string"?new URL(n,typeof window>"u"?"server://singlefetch/":window.location.origin):n;return s.pathname==="/"?s.pathname=`_root.${o}`:r&&U4(s.pathname,r)==="/"?s.pathname=`${r.replace(/\/$/,"")}/_root.${o}`:s.pathname=`${s.pathname.replace(/\/$/,"")}.${o}`,s}async function gnr(n,r){if(n.id in r)return r[n.id];try{let o=await import(n.module);return r[n.id]=o,o}catch(o){return console.error(`Error loading route module \`${n.module}\`, reloading page...`),console.error(o),window.__reactRouterContext&&window.__reactRouterContext.isSpaMode,window.location.reload(),new Promise(()=>{})}}function mnr(n){return n==null?!1:n.href==null?n.rel==="preload"&&typeof n.imageSrcSet=="string"&&typeof n.imageSizes=="string":typeof n.rel=="string"&&typeof n.href=="string"}async function bnr(n,r,o){let s=await Promise.all(n.map(async c=>{let d=r.routes[c.route.id];if(d){let h=await gnr(d,o);return h.links?h.links():[]}return[]}));return Enr(s.flat(1).filter(mnr).filter(c=>c.rel==="stylesheet"||c.rel==="preload").map(c=>c.rel==="stylesheet"?{...c,rel:"prefetch",as:"style"}:{...c,rel:"prefetch"}))}function Y7t(n,r,o,s,c,d){let h=(y,E)=>o[E]?y.route.id!==o[E].route.id:!0,b=(y,E)=>o[E].pathname!==y.pathname||o[E].route.path?.endsWith("*")&&o[E].params["*"]!==y.params["*"];return d==="assets"?r.filter((y,E)=>h(y,E)||b(y,E)):d==="data"?r.filter((y,E)=>{let k=s.routes[y.route.id];if(!k||!k.hasLoader)return!1;if(h(y,E)||b(y,E))return!0;if(y.route.shouldRevalidate){let x=y.route.shouldRevalidate({currentUrl:new URL(c.pathname+c.search+c.hash,window.origin),currentParams:o[0]?.params||{},nextUrl:new URL(n,window.origin),nextParams:y.params,defaultShouldRevalidate:!0});if(typeof x=="boolean")return x}return!0}):[]}function vnr(n,r,{includeHydrateFallback:o}={}){return wnr(n.map(s=>{let c=r.routes[s.route.id];if(!c)return[];let d=[c.module];return c.clientActionModule&&(d=d.concat(c.clientActionModule)),c.clientLoaderModule&&(d=d.concat(c.clientLoaderModule)),o&&c.hydrateFallbackModule&&(d=d.concat(c.hydrateFallbackModule)),c.imports&&(d=d.concat(c.imports)),d}).flat(1))}function wnr(n){return[...new Set(n)]}function ynr(n){let r={},o=Object.keys(n).sort();for(let s of o)r[s]=n[s];return r}function Enr(n,r){let o=new Set;return new Set(r),n.reduce((s,c)=>{let d=JSON.stringify(ynr(c));return o.has(d)||(o.add(d),s.push({key:d,link:c})),s},[])}function ctn(){let n=Q.useContext(AK);return VYe(n,"You must render this element inside a <DataRouterContext.Provider> element"),n}function Snr(){let n=Q.useContext(Pye);return VYe(n,"You must render this element inside a <DataRouterStateContext.Provider> element"),n}var WYe=Q.createContext(void 0);WYe.displayName="FrameworkContext";function utn(){let n=Q.useContext(WYe);return VYe(n,"You must render this element inside a <HydratedRouter> element"),n}function _nr(n,r){let o=Q.useContext(WYe),[s,c]=Q.useState(!1),[d,h]=Q.useState(!1),{onFocus:b,onBlur:y,onMouseEnter:E,onMouseLeave:k,onTouchStart:x}=r,O=Q.useRef(null);Q.useEffect(()=>{if(n==="render"&&h(!0),n==="viewport"){let $=B=>{B.forEach(G=>{h(G.isIntersecting)})},U=new IntersectionObserver($,{threshold:.5});return O.current&&U.observe(O.current),()=>{U.disconnect()}}},[n]),Q.useEffect(()=>{if(s){let $=setTimeout(()=>{h(!0)},100);return()=>{clearTimeout($)}}},[s]);let P=()=>{c(!0)},D=()=>{c(!1),h(!1)};return o?n!=="intent"?[d,O,{}]:[d,O,{onFocus:lre(b,P),onBlur:lre(y,D),onMouseEnter:lre(E,P),onMouseLeave:lre(k,D),onTouchStart:lre(x,P)}]:[!1,O,{}]}function lre(n,r){return o=>{n&&n(o),o.defaultPrevented||r(o)}}function knr({page:n,...r}){let{router:o}=ctn(),s=Q.useMemo(()=>Qen(o.routes,n,o.basename),[o.routes,n,o.basename]);return s?Q.createElement(xnr,{page:n,matches:s,...r}):null}function Tnr(n){let{manifest:r,routeModules:o}=utn(),[s,c]=Q.useState([]);return Q.useEffect(()=>{let d=!1;return bnr(n,r,o).then(h=>{d||c(h)}),()=>{d=!0}},[n,r,o]),s}function xnr({page:n,matches:r,...o}){let s=zI(),{manifest:c,routeModules:d}=utn(),{basename:h}=ctn(),{loaderData:b,matches:y}=Snr(),E=Q.useMemo(()=>Y7t(n,r,y,c,s,"data"),[n,r,y,c,s]),k=Q.useMemo(()=>Y7t(n,r,y,c,s,"assets"),[n,r,y,c,s]),x=Q.useMemo(()=>{if(n===s.pathname+s.search+s.hash)return[];let D=new Set,$=!1;if(r.forEach(B=>{let G=c.routes[B.route.id];!G||!G.hasLoader||(!E.some(K=>K.route.id===B.route.id)&&B.route.id in b&&d[B.route.id]?.shouldRevalidate||G.hasClientLoader?$=!0:D.add(B.route.id))}),D.size===0)return[];let U=pnr(n,h,"data");return $&&D.size>0&&U.searchParams.set("_routes",r.filter(B=>D.has(B.route.id)).map(B=>B.route.id).join(",")),[U.pathname+U.search]},[h,b,s,c,E,r,n,d]),O=Q.useMemo(()=>vnr(k,c),[k,c]),P=Tnr(k);return Q.createElement(Q.Fragment,null,x.map(D=>Q.createElement("link",{key:D,rel:"prefetch",as:"fetch",href:D,...o})),O.map(D=>Q.createElement("link",{key:D,rel:"modulepreload",href:D,...o})),P.map(({key:D,link:$})=>Q.createElement("link",{key:D,...$})))}function Cnr(...n){return r=>{n.forEach(o=>{typeof o=="function"?o(r):o!=null&&(o.current=r)})}}var ltn=typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u";try{ltn&&(window.__reactRouterVersion="7.7.1")}catch{}function Anr({basename:n,children:r,window:o}){let s=Q.useRef();s.current==null&&(s.current=gtr({window:o,v5Compat:!0}));let c=s.current,[d,h]=Q.useState({action:c.action,location:c.location}),b=Q.useCallback(y=>{Q.startTransition(()=>h(y))},[h]);return Q.useLayoutEffect(()=>c.listen(b),[c,b]),Q.createElement(rnr,{basename:n,children:r,location:d.location,navigationType:d.action,navigator:c})}var dtn=/^(?:[a-z][a-z0-9+.-]*:|\/\/)/i,KYe=Q.forwardRef(function({onClick:r,discover:o="render",prefetch:s="none",relative:c,reloadDocument:d,replace:h,state:b,target:y,to:E,preventScrollReset:k,viewTransition:x,...O},P){let{basename:D}=Q.useContext(Zx),$=typeof E=="string"&&dtn.test(E),U,B=!1;if(typeof E=="string"&&$&&(U=E,ltn))try{let we=new URL(window.location.href),ge=E.startsWith("//")?new URL(we.protocol+E):new URL(E),Te=U4(ge.pathname,D);ge.origin===we.origin&&Te!=null?E=Te+ge.search+ge.hash:B=!0}catch{U_(!1,`<Link to="${E}"> contains an invalid URL which will probably break when clicked - please update to a valid URL path.`)}let G=ztr(E,{relative:c}),[K,X,ae]=_nr(s,O),oe=Rnr(E,{replace:h,state:b,target:y,preventScrollReset:k,relative:c,viewTransition:x});function he(we){r&&r(we),we.defaultPrevented||oe(we)}let ye=Q.createElement("a",{...O,...ae,href:U||G,onClick:B||d?r:he,ref:Cnr(P,X),target:y,"data-discover":!$&&o==="render"?"true":void 0});return K&&!$?Q.createElement(Q.Fragment,null,ye,Q.createElement(knr,{page:G})):ye});KYe.displayName="Link";var Inr=Q.forwardRef(function({"aria-current":r="page",caseSensitive:o=!1,className:s="",end:c=!1,style:d,to:h,viewTransition:b,children:y,...E},k){let x=Loe(h,{relative:E.relative}),O=zI(),P=Q.useContext(Pye),{navigator:D,basename:$}=Q.useContext(Zx),U=P!=null&&jnr(x)&&b===!0,B=D.encodeLocation?D.encodeLocation(x).pathname:x.pathname,G=O.pathname,K=P&&P.navigation&&P.navigation.location?P.navigation.location.pathname:null;o||(G=G.toLowerCase(),K=K?K.toLowerCase():null,B=B.toLowerCase()),K&&$&&(K=U4(K,$)||K);const X=B!=="/"&&B.endsWith("/")?B.length-1:B.length;let ae=G===B||!c&&G.startsWith(B)&&G.charAt(X)==="/",oe=K!=null&&(K===B||!c&&K.startsWith(B)&&K.charAt(B.length)==="/"),he={isActive:ae,isPending:oe,isTransitioning:U},ye=ae?r:void 0,we;typeof s=="function"?we=s(he):we=[s,ae?"active":null,oe?"pending":null,U?"transitioning":null].filter(Boolean).join(" ");let ge=typeof d=="function"?d(he):d;return Q.createElement(KYe,{...E,"aria-current":ye,className:we,ref:k,style:ge,to:h,viewTransition:b},typeof y=="function"?y(he):y)});Inr.displayName="NavLink";var Onr=Q.forwardRef(({discover:n="render",fetcherKey:r,navigate:o,reloadDocument:s,replace:c,state:d,method:h=uve,action:b,onSubmit:y,relative:E,preventScrollReset:k,viewTransition:x,...O},P)=>{let D=Dnr(),$=Mnr(b,{relative:E}),U=h.toLowerCase()==="get"?"get":"post",B=typeof b=="string"&&dtn.test(b),G=K=>{if(y&&y(K),K.defaultPrevented)return;K.preventDefault();let X=K.nativeEvent.submitter,ae=X?.getAttribute("formmethod")||h;D(X||K.currentTarget,{fetcherKey:r,method:ae,navigate:o,replace:c,state:d,relative:E,preventScrollReset:k,viewTransition:x})};return Q.createElement("form",{ref:P,method:U,action:$,onSubmit:s?y:G,...O,"data-discover":!B&&n==="render"?"true":void 0})});Onr.displayName="Form";function Nnr(n){return`${n} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function ftn(n){let r=Q.useContext(AK);return Tf(r,Nnr(n)),r}function Rnr(n,{target:r,replace:o,state:s,preventScrollReset:c,relative:d,viewTransition:h}={}){let b=y2(),y=zI(),E=Loe(n,{relative:d});return Q.useCallback(k=>{if(unr(k,r)){k.preventDefault();let x=o!==void 0?o:Fie(y)===Fie(E);b(n,{replace:x,state:s,preventScrollReset:c,relative:d,viewTransition:h})}},[y,b,E,o,s,r,n,c,d,h])}function Dye(n){U_(typeof URLSearchParams<"u","You cannot use the `useSearchParams` hook in a browser that does not support the URLSearchParams API. If you need to support Internet Explorer 11, we recommend you load a polyfill such as https://github.com/ungap/url-search-params.");let r=Q.useRef(wVe(n)),o=Q.useRef(!1),s=zI(),c=Q.useMemo(()=>lnr(s.search,o.current?null:r.current),[s.search]),d=y2(),h=Q.useCallback((b,y)=>{const E=wVe(typeof b=="function"?b(new URLSearchParams(c)):b);o.current=!0,d("?"+E,y)},[d,c]);return[c,h]}var Pnr=0,Lnr=()=>`__${String(++Pnr)}__`;function Dnr(){let{router:n}=ftn("useSubmit"),{basename:r}=Q.useContext(Zx),o=Qtr();return Q.useCallback(async(s,c={})=>{let{action:d,method:h,encType:b,formData:y,body:E}=hnr(s,r);if(c.navigate===!1){let k=c.fetcherKey||Lnr();await n.fetch(k,o,c.action||d,{preventScrollReset:c.preventScrollReset,formData:y,body:E,formMethod:c.method||h,formEncType:c.encType||b,flushSync:c.flushSync})}else await n.navigate(c.action||d,{preventScrollReset:c.preventScrollReset,formData:y,body:E,formMethod:c.method||h,formEncType:c.encType||b,replace:c.replace,state:c.state,fromRouteId:o,flushSync:c.flushSync,viewTransition:c.viewTransition})},[n,r,o])}function Mnr(n,{relative:r}={}){let{basename:o}=Q.useContext(Zx),s=Q.useContext(Qx);Tf(s,"useFormAction must be used inside a RouteContext");let[c]=s.matches.slice(-1),d={...Loe(n||".",{relative:r})},h=zI();if(n==null){d.search=h.search;let b=new URLSearchParams(d.search),y=b.getAll("index");if(y.some(k=>k==="")){b.delete("index"),y.filter(x=>x).forEach(x=>b.append("index",x));let k=b.toString();d.search=k?`?${k}`:""}}return(!n||n===".")&&c.route.index&&(d.search=d.search?d.search.replace(/^\?/,"?index&"):"?index"),o!=="/"&&(d.pathname=d.pathname==="/"?o:j4([o,d.pathname])),Fie(d)}function jnr(n,{relative:r}={}){let o=Q.useContext(rtn);Tf(o!=null,"`useViewTransitionState` must be used within `react-router-dom`'s `RouterProvider`.  Did you accidentally import `RouterProvider` from `react-router`?");let{basename:s}=ftn("useViewTransitionState"),c=Loe(n,{relative:r});if(!o.isTransitioning)return!1;let d=U4(o.currentLocation.pathname,s)||o.currentLocation.pathname,h=U4(o.nextLocation.pathname,s)||o.nextLocation.pathname;return Qve(c.pathname,h)!=null||Qve(c.pathname,d)!=null}/*!
 * Font Awesome Free 6.7.2 by @fontawesome - https://fontawesome.com
 * License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License)
 * Copyright 2024 Fonticons, Inc.
 */function $nr(n,r,o){return(r=Bnr(r))in n?Object.defineProperty(n,r,{value:o,enumerable:!0,configurable:!0,writable:!0}):n[r]=o,n}function X7t(n,r){var o=Object.keys(n);if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(n);r&&(s=s.filter(function(c){return Object.getOwnPropertyDescriptor(n,c).enumerable})),o.push.apply(o,s)}return o}function Hr(n){for(var r=1;r<arguments.length;r++){var o=arguments[r]!=null?arguments[r]:{};r%2?X7t(Object(o),!0).forEach(function(s){$nr(n,s,o[s])}):Object.getOwnPropertyDescriptors?Object.defineProperties(n,Object.getOwnPropertyDescriptors(o)):X7t(Object(o)).forEach(function(s){Object.defineProperty(n,s,Object.getOwnPropertyDescriptor(o,s))})}return n}function Fnr(n,r){if(typeof n!="object"||!n)return n;var o=n[Symbol.toPrimitive];if(o!==void 0){var s=o.call(n,r);if(typeof s!="object")return s;throw new TypeError("@@toPrimitive must return a primitive value.")}return(r==="string"?String:Number)(n)}function Bnr(n){var r=Fnr(n,"string");return typeof r=="symbol"?r:r+""}const J7t=()=>{};let YYe={},htn={},ptn=null,gtn={mark:J7t,measure:J7t};try{typeof window<"u"&&(YYe=window),typeof document<"u"&&(htn=document),typeof MutationObserver<"u"&&(ptn=MutationObserver),typeof performance<"u"&&(gtn=performance)}catch{}const{userAgent:Z7t=""}=YYe.navigator||{},xM=YYe,qd=htn,Q7t=ptn,Z1e=gtn;xM.document;const Q4=!!qd.documentElement&&!!qd.head&&typeof qd.addEventListener=="function"&&typeof qd.createElement=="function",mtn=~Z7t.indexOf("MSIE")||~Z7t.indexOf("Trident/");var Unr=/fa(s|r|l|t|d|dr|dl|dt|b|k|kd|ss|sr|sl|st|sds|sdr|sdl|sdt)?[\-\ ]/,znr=/Font ?Awesome ?([56 ]*)(Solid|Regular|Light|Thin|Duotone|Brands|Free|Pro|Sharp Duotone|Sharp|Kit)?.*/i,btn={classic:{fa:"solid",fas:"solid","fa-solid":"solid",far:"regular","fa-regular":"regular",fal:"light","fa-light":"light",fat:"thin","fa-thin":"thin",fab:"brands","fa-brands":"brands"},duotone:{fa:"solid",fad:"solid","fa-solid":"solid","fa-duotone":"solid",fadr:"regular","fa-regular":"regular",fadl:"light","fa-light":"light",fadt:"thin","fa-thin":"thin"},sharp:{fa:"solid",fass:"solid","fa-solid":"solid",fasr:"regular","fa-regular":"regular",fasl:"light","fa-light":"light",fast:"thin","fa-thin":"thin"},"sharp-duotone":{fa:"solid",fasds:"solid","fa-solid":"solid",fasdr:"regular","fa-regular":"regular",fasdl:"light","fa-light":"light",fasdt:"thin","fa-thin":"thin"}},Hnr={GROUP:"duotone-group",PRIMARY:"primary",SECONDARY:"secondary"},vtn=["fa-classic","fa-duotone","fa-sharp","fa-sharp-duotone"],X1="classic",Mye="duotone",Gnr="sharp",qnr="sharp-duotone",wtn=[X1,Mye,Gnr,qnr],Vnr={classic:{900:"fas",400:"far",normal:"far",300:"fal",100:"fat"},duotone:{900:"fad",400:"fadr",300:"fadl",100:"fadt"},sharp:{900:"fass",400:"fasr",300:"fasl",100:"fast"},"sharp-duotone":{900:"fasds",400:"fasdr",300:"fasdl",100:"fasdt"}},Wnr={"Font Awesome 6 Free":{900:"fas",400:"far"},"Font Awesome 6 Pro":{900:"fas",400:"far",normal:"far",300:"fal",100:"fat"},"Font Awesome 6 Brands":{400:"fab",normal:"fab"},"Font Awesome 6 Duotone":{900:"fad",400:"fadr",normal:"fadr",300:"fadl",100:"fadt"},"Font Awesome 6 Sharp":{900:"fass",400:"fasr",normal:"fasr",300:"fasl",100:"fast"},"Font Awesome 6 Sharp Duotone":{900:"fasds",400:"fasdr",normal:"fasdr",300:"fasdl",100:"fasdt"}},Knr=new Map([["classic",{defaultShortPrefixId:"fas",defaultStyleId:"solid",styleIds:["solid","regular","light","thin","brands"],futureStyleIds:[],defaultFontWeight:900}],["sharp",{defaultShortPrefixId:"fass",defaultStyleId:"solid",styleIds:["solid","regular","light","thin"],futureStyleIds:[],defaultFontWeight:900}],["duotone",{defaultShortPrefixId:"fad",defaultStyleId:"solid",styleIds:["solid","regular","light","thin"],futureStyleIds:[],defaultFontWeight:900}],["sharp-duotone",{defaultShortPrefixId:"fasds",defaultStyleId:"solid",styleIds:["solid","regular","light","thin"],futureStyleIds:[],defaultFontWeight:900}]]),Ynr={classic:{solid:"fas",regular:"far",light:"fal",thin:"fat",brands:"fab"},duotone:{solid:"fad",regular:"fadr",light:"fadl",thin:"fadt"},sharp:{solid:"fass",regular:"fasr",light:"fasl",thin:"fast"},"sharp-duotone":{solid:"fasds",regular:"fasdr",light:"fasdl",thin:"fasdt"}},Xnr=["fak","fa-kit","fakd","fa-kit-duotone"],ejt={kit:{fak:"kit","fa-kit":"kit"},"kit-duotone":{fakd:"kit-duotone","fa-kit-duotone":"kit-duotone"}},Jnr=["kit"],Znr={kit:{"fa-kit":"fak"}},Qnr=["fak","fakd"],err={kit:{fak:"fa-kit"}},tjt={kit:{kit:"fak"},"kit-duotone":{"kit-duotone":"fakd"}},Q1e={GROUP:"duotone-group",SWAP_OPACITY:"swap-opacity",PRIMARY:"primary",SECONDARY:"secondary"},trr=["fa-classic","fa-duotone","fa-sharp","fa-sharp-duotone"],nrr=["fak","fa-kit","fakd","fa-kit-duotone"],rrr={"Font Awesome Kit":{400:"fak",normal:"fak"},"Font Awesome Kit Duotone":{400:"fakd",normal:"fakd"}},irr={classic:{"fa-brands":"fab","fa-duotone":"fad","fa-light":"fal","fa-regular":"far","fa-solid":"fas","fa-thin":"fat"},duotone:{"fa-regular":"fadr","fa-light":"fadl","fa-thin":"fadt"},sharp:{"fa-solid":"fass","fa-regular":"fasr","fa-light":"fasl","fa-thin":"fast"},"sharp-duotone":{"fa-solid":"fasds","fa-regular":"fasdr","fa-light":"fasdl","fa-thin":"fasdt"}},orr={classic:["fas","far","fal","fat","fad"],duotone:["fadr","fadl","fadt"],sharp:["fass","fasr","fasl","fast"],"sharp-duotone":["fasds","fasdr","fasdl","fasdt"]},yVe={classic:{fab:"fa-brands",fad:"fa-duotone",fal:"fa-light",far:"fa-regular",fas:"fa-solid",fat:"fa-thin"},duotone:{fadr:"fa-regular",fadl:"fa-light",fadt:"fa-thin"},sharp:{fass:"fa-solid",fasr:"fa-regular",fasl:"fa-light",fast:"fa-thin"},"sharp-duotone":{fasds:"fa-solid",fasdr:"fa-regular",fasdl:"fa-light",fasdt:"fa-thin"}},arr=["fa-solid","fa-regular","fa-light","fa-thin","fa-duotone","fa-brands"],EVe=["fa","fas","far","fal","fat","fad","fadr","fadl","fadt","fab","fass","fasr","fasl","fast","fasds","fasdr","fasdl","fasdt",...trr,...arr],srr=["solid","regular","light","thin","duotone","brands"],ytn=[1,2,3,4,5,6,7,8,9,10],crr=ytn.concat([11,12,13,14,15,16,17,18,19,20]),urr=[...Object.keys(orr),...srr,"2xs","xs","sm","lg","xl","2xl","beat","border","fade","beat-fade","bounce","flip-both","flip-horizontal","flip-vertical","flip","fw","inverse","layers-counter","layers-text","layers","li","pull-left","pull-right","pulse","rotate-180","rotate-270","rotate-90","rotate-by","shake","spin-pulse","spin-reverse","spin","stack-1x","stack-2x","stack","ul",Q1e.GROUP,Q1e.SWAP_OPACITY,Q1e.PRIMARY,Q1e.SECONDARY].concat(ytn.map(n=>"".concat(n,"x"))).concat(crr.map(n=>"w-".concat(n))),lrr={"Font Awesome 5 Free":{900:"fas",400:"far"},"Font Awesome 5 Pro":{900:"fas",400:"far",normal:"far",300:"fal"},"Font Awesome 5 Brands":{400:"fab",normal:"fab"},"Font Awesome 5 Duotone":{900:"fad"}};const z4="___FONT_AWESOME___",SVe=16,Etn="fa",Stn="svg-inline--fa",XB="data-fa-i2svg",_Ve="data-fa-pseudo-element",drr="data-fa-pseudo-element-pending",XYe="data-prefix",JYe="data-icon",njt="fontawesome-i2svg",frr="async",hrr=["HTML","HEAD","STYLE","SCRIPT"],_tn=(()=>{try{return!0}catch{return!1}})();function Doe(n){return new Proxy(n,{get(r,o){return o in r?r[o]:r[X1]}})}const ktn=Hr({},btn);ktn[X1]=Hr(Hr(Hr(Hr({},{"fa-duotone":"duotone"}),btn[X1]),ejt.kit),ejt["kit-duotone"]);const prr=Doe(ktn),kVe=Hr({},Ynr);kVe[X1]=Hr(Hr(Hr(Hr({},{duotone:"fad"}),kVe[X1]),tjt.kit),tjt["kit-duotone"]);const rjt=Doe(kVe),TVe=Hr({},yVe);TVe[X1]=Hr(Hr({},TVe[X1]),err.kit);const ZYe=Doe(TVe),xVe=Hr({},irr);xVe[X1]=Hr(Hr({},xVe[X1]),Znr.kit);Doe(xVe);const grr=Unr,Ttn="fa-layers-text",mrr=znr,brr=Hr({},Vnr);Doe(brr);const vrr=["class","data-prefix","data-icon","data-fa-transform","data-fa-mask"],d$e=Hnr,wrr=[...Jnr,...urr],cie=xM.FontAwesomeConfig||{};function yrr(n){var r=qd.querySelector("script["+n+"]");if(r)return r.getAttribute(n)}function Err(n){return n===""?!0:n==="false"?!1:n==="true"?!0:n}qd&&typeof qd.querySelector=="function"&&[["data-family-prefix","familyPrefix"],["data-css-prefix","cssPrefix"],["data-family-default","familyDefault"],["data-style-default","styleDefault"],["data-replacement-class","replacementClass"],["data-auto-replace-svg","autoReplaceSvg"],["data-auto-add-css","autoAddCss"],["data-auto-a11y","autoA11y"],["data-search-pseudo-elements","searchPseudoElements"],["data-observe-mutations","observeMutations"],["data-mutate-approach","mutateApproach"],["data-keep-original-source","keepOriginalSource"],["data-measure-performance","measurePerformance"],["data-show-missing-icons","showMissingIcons"]].forEach(r=>{let[o,s]=r;const c=Err(yrr(o));c!=null&&(cie[s]=c)});const xtn={styleDefault:"solid",familyDefault:X1,cssPrefix:Etn,replacementClass:Stn,autoReplaceSvg:!0,autoAddCss:!0,autoA11y:!0,searchPseudoElements:!1,observeMutations:!0,mutateApproach:"async",keepOriginalSource:!0,measurePerformance:!1,showMissingIcons:!0};cie.familyPrefix&&(cie.cssPrefix=cie.familyPrefix);const YW=Hr(Hr({},xtn),cie);YW.autoReplaceSvg||(YW.observeMutations=!1);const Co={};Object.keys(xtn).forEach(n=>{Object.defineProperty(Co,n,{enumerable:!0,set:function(r){YW[n]=r,uie.forEach(o=>o(Co))},get:function(){return YW[n]}})});Object.defineProperty(Co,"familyPrefix",{enumerable:!0,set:function(n){YW.cssPrefix=n,uie.forEach(r=>r(Co))},get:function(){return YW.cssPrefix}});xM.FontAwesomeConfig=Co;const uie=[];function Srr(n){return uie.push(n),()=>{uie.splice(uie.indexOf(n),1)}}const GD=SVe,bI={size:16,x:0,y:0,rotate:0,flipX:!1,flipY:!1};function _rr(n){if(!n||!Q4)return;const r=qd.createElement("style");r.setAttribute("type","text/css"),r.innerHTML=n;const o=qd.head.childNodes;let s=null;for(let c=o.length-1;c>-1;c--){const d=o[c],h=(d.tagName||"").toUpperCase();["STYLE","LINK"].indexOf(h)>-1&&(s=d)}return qd.head.insertBefore(r,s),n}const krr="0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";function Bie(){let n=12,r="";for(;n-- >0;)r+=krr[Math.random()*62|0];return r}function OK(n){const r=[];for(let o=(n||[]).length>>>0;o--;)r[o]=n[o];return r}function QYe(n){return n.classList?OK(n.classList):(n.getAttribute("class")||"").split(" ").filter(r=>r)}function Ctn(n){return"".concat(n).replace(/&/g,"&amp;").replace(/"/g,"&quot;").replace(/'/g,"&#39;").replace(/</g,"&lt;").replace(/>/g,"&gt;")}function Trr(n){return Object.keys(n||{}).reduce((r,o)=>r+"".concat(o,'="').concat(Ctn(n[o]),'" '),"").trim()}function jye(n){return Object.keys(n||{}).reduce((r,o)=>r+"".concat(o,": ").concat(n[o].trim(),";"),"")}function eXe(n){return n.size!==bI.size||n.x!==bI.x||n.y!==bI.y||n.rotate!==bI.rotate||n.flipX||n.flipY}function xrr(n){let{transform:r,containerWidth:o,iconWidth:s}=n;const c={transform:"translate(".concat(o/2," 256)")},d="translate(".concat(r.x*32,", ").concat(r.y*32,") "),h="scale(".concat(r.size/16*(r.flipX?-1:1),", ").concat(r.size/16*(r.flipY?-1:1),") "),b="rotate(".concat(r.rotate," 0 0)"),y={transform:"".concat(d," ").concat(h," ").concat(b)},E={transform:"translate(".concat(s/2*-1," -256)")};return{outer:c,inner:y,path:E}}function Crr(n){let{transform:r,width:o=SVe,height:s=SVe,startCentered:c=!1}=n,d="";return c&&mtn?d+="translate(".concat(r.x/GD-o/2,"em, ").concat(r.y/GD-s/2,"em) "):c?d+="translate(calc(-50% + ".concat(r.x/GD,"em), calc(-50% + ").concat(r.y/GD,"em)) "):d+="translate(".concat(r.x/GD,"em, ").concat(r.y/GD,"em) "),d+="scale(".concat(r.size/GD*(r.flipX?-1:1),", ").concat(r.size/GD*(r.flipY?-1:1),") "),d+="rotate(".concat(r.rotate,"deg) "),d}var Arr=`:root, :host {
  --fa-font-solid: normal 900 1em/1 "Font Awesome 6 Free";
  --fa-font-regular: normal 400 1em/1 "Font Awesome 6 Free";
  --fa-font-light: normal 300 1em/1 "Font Awesome 6 Pro";
  --fa-font-thin: normal 100 1em/1 "Font Awesome 6 Pro";
  --fa-font-duotone: normal 900 1em/1 "Font Awesome 6 Duotone";
  --fa-font-duotone-regular: normal 400 1em/1 "Font Awesome 6 Duotone";
  --fa-font-duotone-light: normal 300 1em/1 "Font Awesome 6 Duotone";
  --fa-font-duotone-thin: normal 100 1em/1 "Font Awesome 6 Duotone";
  --fa-font-brands: normal 400 1em/1 "Font Awesome 6 Brands";
  --fa-font-sharp-solid: normal 900 1em/1 "Font Awesome 6 Sharp";
  --fa-font-sharp-regular: normal 400 1em/1 "Font Awesome 6 Sharp";
  --fa-font-sharp-light: normal 300 1em/1 "Font Awesome 6 Sharp";
  --fa-font-sharp-thin: normal 100 1em/1 "Font Awesome 6 Sharp";
  --fa-font-sharp-duotone-solid: normal 900 1em/1 "Font Awesome 6 Sharp Duotone";
  --fa-font-sharp-duotone-regular: normal 400 1em/1 "Font Awesome 6 Sharp Duotone";
  --fa-font-sharp-duotone-light: normal 300 1em/1 "Font Awesome 6 Sharp Duotone";
  --fa-font-sharp-duotone-thin: normal 100 1em/1 "Font Awesome 6 Sharp Duotone";
}

svg:not(:root).svg-inline--fa, svg:not(:host).svg-inline--fa {
  overflow: visible;
  box-sizing: content-box;
}

.svg-inline--fa {
  display: var(--fa-display, inline-block);
  height: 1em;
  overflow: visible;
  vertical-align: -0.125em;
}
.svg-inline--fa.fa-2xs {
  vertical-align: 0.1em;
}
.svg-inline--fa.fa-xs {
  vertical-align: 0em;
}
.svg-inline--fa.fa-sm {
  vertical-align: -0.0714285705em;
}
.svg-inline--fa.fa-lg {
  vertical-align: -0.2em;
}
.svg-inline--fa.fa-xl {
  vertical-align: -0.25em;
}
.svg-inline--fa.fa-2xl {
  vertical-align: -0.3125em;
}
.svg-inline--fa.fa-pull-left {
  margin-right: var(--fa-pull-margin, 0.3em);
  width: auto;
}
.svg-inline--fa.fa-pull-right {
  margin-left: var(--fa-pull-margin, 0.3em);
  width: auto;
}
.svg-inline--fa.fa-li {
  width: var(--fa-li-width, 2em);
  top: 0.25em;
}
.svg-inline--fa.fa-fw {
  width: var(--fa-fw-width, 1.25em);
}

.fa-layers svg.svg-inline--fa {
  bottom: 0;
  left: 0;
  margin: auto;
  position: absolute;
  right: 0;
  top: 0;
}

.fa-layers-counter, .fa-layers-text {
  display: inline-block;
  position: absolute;
  text-align: center;
}

.fa-layers {
  display: inline-block;
  height: 1em;
  position: relative;
  text-align: center;
  vertical-align: -0.125em;
  width: 1em;
}
.fa-layers svg.svg-inline--fa {
  transform-origin: center center;
}

.fa-layers-text {
  left: 50%;
  top: 50%;
  transform: translate(-50%, -50%);
  transform-origin: center center;
}

.fa-layers-counter {
  background-color: var(--fa-counter-background-color, #ff253a);
  border-radius: var(--fa-counter-border-radius, 1em);
  box-sizing: border-box;
  color: var(--fa-inverse, #fff);
  line-height: var(--fa-counter-line-height, 1);
  max-width: var(--fa-counter-max-width, 5em);
  min-width: var(--fa-counter-min-width, 1.5em);
  overflow: hidden;
  padding: var(--fa-counter-padding, 0.25em 0.5em);
  right: var(--fa-right, 0);
  text-overflow: ellipsis;
  top: var(--fa-top, 0);
  transform: scale(var(--fa-counter-scale, 0.25));
  transform-origin: top right;
}

.fa-layers-bottom-right {
  bottom: var(--fa-bottom, 0);
  right: var(--fa-right, 0);
  top: auto;
  transform: scale(var(--fa-layers-scale, 0.25));
  transform-origin: bottom right;
}

.fa-layers-bottom-left {
  bottom: var(--fa-bottom, 0);
  left: var(--fa-left, 0);
  right: auto;
  top: auto;
  transform: scale(var(--fa-layers-scale, 0.25));
  transform-origin: bottom left;
}

.fa-layers-top-right {
  top: var(--fa-top, 0);
  right: var(--fa-right, 0);
  transform: scale(var(--fa-layers-scale, 0.25));
  transform-origin: top right;
}

.fa-layers-top-left {
  left: var(--fa-left, 0);
  right: auto;
  top: var(--fa-top, 0);
  transform: scale(var(--fa-layers-scale, 0.25));
  transform-origin: top left;
}

.fa-1x {
  font-size: 1em;
}

.fa-2x {
  font-size: 2em;
}

.fa-3x {
  font-size: 3em;
}

.fa-4x {
  font-size: 4em;
}

.fa-5x {
  font-size: 5em;
}

.fa-6x {
  font-size: 6em;
}

.fa-7x {
  font-size: 7em;
}

.fa-8x {
  font-size: 8em;
}

.fa-9x {
  font-size: 9em;
}

.fa-10x {
  font-size: 10em;
}

.fa-2xs {
  font-size: 0.625em;
  line-height: 0.1em;
  vertical-align: 0.225em;
}

.fa-xs {
  font-size: 0.75em;
  line-height: 0.0833333337em;
  vertical-align: 0.125em;
}

.fa-sm {
  font-size: 0.875em;
  line-height: 0.0714285718em;
  vertical-align: 0.0535714295em;
}

.fa-lg {
  font-size: 1.25em;
  line-height: 0.05em;
  vertical-align: -0.075em;
}

.fa-xl {
  font-size: 1.5em;
  line-height: 0.0416666682em;
  vertical-align: -0.125em;
}

.fa-2xl {
  font-size: 2em;
  line-height: 0.03125em;
  vertical-align: -0.1875em;
}

.fa-fw {
  text-align: center;
  width: 1.25em;
}

.fa-ul {
  list-style-type: none;
  margin-left: var(--fa-li-margin, 2.5em);
  padding-left: 0;
}
.fa-ul > li {
  position: relative;
}

.fa-li {
  left: calc(-1 * var(--fa-li-width, 2em));
  position: absolute;
  text-align: center;
  width: var(--fa-li-width, 2em);
  line-height: inherit;
}

.fa-border {
  border-color: var(--fa-border-color, #eee);
  border-radius: var(--fa-border-radius, 0.1em);
  border-style: var(--fa-border-style, solid);
  border-width: var(--fa-border-width, 0.08em);
  padding: var(--fa-border-padding, 0.2em 0.25em 0.15em);
}

.fa-pull-left {
  float: left;
  margin-right: var(--fa-pull-margin, 0.3em);
}

.fa-pull-right {
  float: right;
  margin-left: var(--fa-pull-margin, 0.3em);
}

.fa-beat {
  animation-name: fa-beat;
  animation-delay: var(--fa-animation-delay, 0s);
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 1s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, ease-in-out);
}

.fa-bounce {
  animation-name: fa-bounce;
  animation-delay: var(--fa-animation-delay, 0s);
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 1s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, cubic-bezier(0.28, 0.84, 0.42, 1));
}

.fa-fade {
  animation-name: fa-fade;
  animation-delay: var(--fa-animation-delay, 0s);
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 1s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, cubic-bezier(0.4, 0, 0.6, 1));
}

.fa-beat-fade {
  animation-name: fa-beat-fade;
  animation-delay: var(--fa-animation-delay, 0s);
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 1s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, cubic-bezier(0.4, 0, 0.6, 1));
}

.fa-flip {
  animation-name: fa-flip;
  animation-delay: var(--fa-animation-delay, 0s);
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 1s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, ease-in-out);
}

.fa-shake {
  animation-name: fa-shake;
  animation-delay: var(--fa-animation-delay, 0s);
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 1s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, linear);
}

.fa-spin {
  animation-name: fa-spin;
  animation-delay: var(--fa-animation-delay, 0s);
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 2s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, linear);
}

.fa-spin-reverse {
  --fa-animation-direction: reverse;
}

.fa-pulse,
.fa-spin-pulse {
  animation-name: fa-spin;
  animation-direction: var(--fa-animation-direction, normal);
  animation-duration: var(--fa-animation-duration, 1s);
  animation-iteration-count: var(--fa-animation-iteration-count, infinite);
  animation-timing-function: var(--fa-animation-timing, steps(8));
}

@media (prefers-reduced-motion: reduce) {
  .fa-beat,
.fa-bounce,
.fa-fade,
.fa-beat-fade,
.fa-flip,
.fa-pulse,
.fa-shake,
.fa-spin,
.fa-spin-pulse {
    animation-delay: -1ms;
    animation-duration: 1ms;
    animation-iteration-count: 1;
    transition-delay: 0s;
    transition-duration: 0s;
  }
}
@keyframes fa-beat {
  0%, 90% {
    transform: scale(1);
  }
  45% {
    transform: scale(var(--fa-beat-scale, 1.25));
  }
}
@keyframes fa-bounce {
  0% {
    transform: scale(1, 1) translateY(0);
  }
  10% {
    transform: scale(var(--fa-bounce-start-scale-x, 1.1), var(--fa-bounce-start-scale-y, 0.9)) translateY(0);
  }
  30% {
    transform: scale(var(--fa-bounce-jump-scale-x, 0.9), var(--fa-bounce-jump-scale-y, 1.1)) translateY(var(--fa-bounce-height, -0.5em));
  }
  50% {
    transform: scale(var(--fa-bounce-land-scale-x, 1.05), var(--fa-bounce-land-scale-y, 0.95)) translateY(0);
  }
  57% {
    transform: scale(1, 1) translateY(var(--fa-bounce-rebound, -0.125em));
  }
  64% {
    transform: scale(1, 1) translateY(0);
  }
  100% {
    transform: scale(1, 1) translateY(0);
  }
}
@keyframes fa-fade {
  50% {
    opacity: var(--fa-fade-opacity, 0.4);
  }
}
@keyframes fa-beat-fade {
  0%, 100% {
    opacity: var(--fa-beat-fade-opacity, 0.4);
    transform: scale(1);
  }
  50% {
    opacity: 1;
    transform: scale(var(--fa-beat-fade-scale, 1.125));
  }
}
@keyframes fa-flip {
  50% {
    transform: rotate3d(var(--fa-flip-x, 0), var(--fa-flip-y, 1), var(--fa-flip-z, 0), var(--fa-flip-angle, -180deg));
  }
}
@keyframes fa-shake {
  0% {
    transform: rotate(-15deg);
  }
  4% {
    transform: rotate(15deg);
  }
  8%, 24% {
    transform: rotate(-18deg);
  }
  12%, 28% {
    transform: rotate(18deg);
  }
  16% {
    transform: rotate(-22deg);
  }
  20% {
    transform: rotate(22deg);
  }
  32% {
    transform: rotate(-12deg);
  }
  36% {
    transform: rotate(12deg);
  }
  40%, 100% {
    transform: rotate(0deg);
  }
}
@keyframes fa-spin {
  0% {
    transform: rotate(0deg);
  }
  100% {
    transform: rotate(360deg);
  }
}
.fa-rotate-90 {
  transform: rotate(90deg);
}

.fa-rotate-180 {
  transform: rotate(180deg);
}

.fa-rotate-270 {
  transform: rotate(270deg);
}

.fa-flip-horizontal {
  transform: scale(-1, 1);
}

.fa-flip-vertical {
  transform: scale(1, -1);
}

.fa-flip-both,
.fa-flip-horizontal.fa-flip-vertical {
  transform: scale(-1, -1);
}

.fa-rotate-by {
  transform: rotate(var(--fa-rotate-angle, 0));
}

.fa-stack {
  display: inline-block;
  vertical-align: middle;
  height: 2em;
  position: relative;
  width: 2.5em;
}

.fa-stack-1x,
.fa-stack-2x {
  bottom: 0;
  left: 0;
  margin: auto;
  position: absolute;
  right: 0;
  top: 0;
  z-index: var(--fa-stack-z-index, auto);
}

.svg-inline--fa.fa-stack-1x {
  height: 1em;
  width: 1.25em;
}
.svg-inline--fa.fa-stack-2x {
  height: 2em;
  width: 2.5em;
}

.fa-inverse {
  color: var(--fa-inverse, #fff);
}

.sr-only,
.fa-sr-only {
  position: absolute;
  width: 1px;
  height: 1px;
  padding: 0;
  margin: -1px;
  overflow: hidden;
  clip: rect(0, 0, 0, 0);
  white-space: nowrap;
  border-width: 0;
}

.sr-only-focusable:not(:focus),
.fa-sr-only-focusable:not(:focus) {
  position: absolute;
  width: 1px;
  height: 1px;
  padding: 0;
  margin: -1px;
  overflow: hidden;
  clip: rect(0, 0, 0, 0);
  white-space: nowrap;
  border-width: 0;
}

.svg-inline--fa .fa-primary {
  fill: var(--fa-primary-color, currentColor);
  opacity: var(--fa-primary-opacity, 1);
}

.svg-inline--fa .fa-secondary {
  fill: var(--fa-secondary-color, currentColor);
  opacity: var(--fa-secondary-opacity, 0.4);
}

.svg-inline--fa.fa-swap-opacity .fa-primary {
  opacity: var(--fa-secondary-opacity, 0.4);
}

.svg-inline--fa.fa-swap-opacity .fa-secondary {
  opacity: var(--fa-primary-opacity, 1);
}

.svg-inline--fa mask .fa-primary,
.svg-inline--fa mask .fa-secondary {
  fill: black;
}`;function Atn(){const n=Etn,r=Stn,o=Co.cssPrefix,s=Co.replacementClass;let c=Arr;if(o!==n||s!==r){const d=new RegExp("\\.".concat(n,"\\-"),"g"),h=new RegExp("\\--".concat(n,"\\-"),"g"),b=new RegExp("\\.".concat(r),"g");c=c.replace(d,".".concat(o,"-")).replace(h,"--".concat(o,"-")).replace(b,".".concat(s))}return c}let ijt=!1;function f$e(){Co.autoAddCss&&!ijt&&(_rr(Atn()),ijt=!0)}var Irr={mixout(){return{dom:{css:Atn,insertCss:f$e}}},hooks(){return{beforeDOMElementCreation(){f$e()},beforeI2svg(){f$e()}}}};const H4=xM||{};H4[z4]||(H4[z4]={});H4[z4].styles||(H4[z4].styles={});H4[z4].hooks||(H4[z4].hooks={});H4[z4].shims||(H4[z4].shims=[]);var vI=H4[z4];const Itn=[],Otn=function(){qd.removeEventListener("DOMContentLoaded",Otn),ewe=1,Itn.map(n=>n())};let ewe=!1;Q4&&(ewe=(qd.documentElement.doScroll?/^loaded|^c/:/^loaded|^i|^c/).test(qd.readyState),ewe||qd.addEventListener("DOMContentLoaded",Otn));function Orr(n){Q4&&(ewe?setTimeout(n,0):Itn.push(n))}function Moe(n){const{tag:r,attributes:o={},children:s=[]}=n;return typeof n=="string"?Ctn(n):"<".concat(r," ").concat(Trr(o),">").concat(s.map(Moe).join(""),"</").concat(r,">")}function ojt(n,r,o){if(n&&n[r]&&n[r][o])return{prefix:r,iconName:o,icon:n[r][o]}}var h$e=function(r,o,s,c){var d=Object.keys(r),h=d.length,b=o,y,E,k;for(s===void 0?(y=1,k=r[d[0]]):(y=0,k=s);y<h;y++)E=d[y],k=b(k,r[E],E,r);return k};function Nrr(n){const r=[];let o=0;const s=n.length;for(;o<s;){const c=n.charCodeAt(o++);if(c>=55296&&c<=56319&&o<s){const d=n.charCodeAt(o++);(d&64512)==56320?r.push(((c&1023)<<10)+(d&1023)+65536):(r.push(c),o--)}else r.push(c)}return r}function CVe(n){const r=Nrr(n);return r.length===1?r[0].toString(16):null}function Rrr(n,r){const o=n.length;let s=n.charCodeAt(r),c;return s>=55296&&s<=56319&&o>r+1&&(c=n.charCodeAt(r+1),c>=56320&&c<=57343)?(s-55296)*1024+c-56320+65536:s}function ajt(n){return Object.keys(n).reduce((r,o)=>{const s=n[o];return!!s.icon?r[s.iconName]=s.icon:r[o]=s,r},{})}function AVe(n,r){let o=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};const{skipHooks:s=!1}=o,c=ajt(r);typeof vI.hooks.addPack=="function"&&!s?vI.hooks.addPack(n,ajt(r)):vI.styles[n]=Hr(Hr({},vI.styles[n]||{}),c),n==="fas"&&AVe("fa",r)}const{styles:Uie,shims:Prr}=vI,Ntn=Object.keys(ZYe),Lrr=Ntn.reduce((n,r)=>(n[r]=Object.keys(ZYe[r]),n),{});let tXe=null,Rtn={},Ptn={},Ltn={},Dtn={},Mtn={};function Drr(n){return~wrr.indexOf(n)}function Mrr(n,r){const o=r.split("-"),s=o[0],c=o.slice(1).join("-");return s===n&&c!==""&&!Drr(c)?c:null}const jtn=()=>{const n=s=>h$e(Uie,(c,d,h)=>(c[h]=h$e(d,s,{}),c),{});Rtn=n((s,c,d)=>(c[3]&&(s[c[3]]=d),c[2]&&c[2].filter(b=>typeof b=="number").forEach(b=>{s[b.toString(16)]=d}),s)),Ptn=n((s,c,d)=>(s[d]=d,c[2]&&c[2].filter(b=>typeof b=="string").forEach(b=>{s[b]=d}),s)),Mtn=n((s,c,d)=>{const h=c[2];return s[d]=d,h.forEach(b=>{s[b]=d}),s});const r="far"in Uie||Co.autoFetchSvg,o=h$e(Prr,(s,c)=>{const d=c[0];let h=c[1];const b=c[2];return h==="far"&&!r&&(h="fas"),typeof d=="string"&&(s.names[d]={prefix:h,iconName:b}),typeof d=="number"&&(s.unicodes[d.toString(16)]={prefix:h,iconName:b}),s},{names:{},unicodes:{}});Ltn=o.names,Dtn=o.unicodes,tXe=$ye(Co.styleDefault,{family:Co.familyDefault})};Srr(n=>{tXe=$ye(n.styleDefault,{family:Co.familyDefault})});jtn();function nXe(n,r){return(Rtn[n]||{})[r]}function jrr(n,r){return(Ptn[n]||{})[r]}function RB(n,r){return(Mtn[n]||{})[r]}function $tn(n){return Ltn[n]||{prefix:null,iconName:null}}function $rr(n){const r=Dtn[n],o=nXe("fas",n);return r||(o?{prefix:"fas",iconName:o}:null)||{prefix:null,iconName:null}}function CM(){return tXe}const Ftn=()=>({prefix:null,iconName:null,rest:[]});function Frr(n){let r=X1;const o=Ntn.reduce((s,c)=>(s[c]="".concat(Co.cssPrefix,"-").concat(c),s),{});return wtn.forEach(s=>{(n.includes(o[s])||n.some(c=>Lrr[s].includes(c)))&&(r=s)}),r}function $ye(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const{family:o=X1}=r,s=prr[o][n];if(o===Mye&&!n)return"fad";const c=rjt[o][n]||rjt[o][s],d=n in vI.styles?n:null;return c||d||null}function Brr(n){let r=[],o=null;return n.forEach(s=>{const c=Mrr(Co.cssPrefix,s);c?o=c:s&&r.push(s)}),{iconName:o,rest:r}}function sjt(n){return n.sort().filter((r,o,s)=>s.indexOf(r)===o)}function Fye(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const{skipLookups:o=!1}=r;let s=null;const c=EVe.concat(nrr),d=sjt(n.filter(x=>c.includes(x))),h=sjt(n.filter(x=>!EVe.includes(x))),b=d.filter(x=>(s=x,!vtn.includes(x))),[y=null]=b,E=Frr(d),k=Hr(Hr({},Brr(h)),{},{prefix:$ye(y,{family:E})});return Hr(Hr(Hr({},k),Grr({values:n,family:E,styles:Uie,config:Co,canonical:k,givenPrefix:s})),Urr(o,s,k))}function Urr(n,r,o){let{prefix:s,iconName:c}=o;if(n||!s||!c)return{prefix:s,iconName:c};const d=r==="fa"?$tn(c):{},h=RB(s,c);return c=d.iconName||h||c,s=d.prefix||s,s==="far"&&!Uie.far&&Uie.fas&&!Co.autoFetchSvg&&(s="fas"),{prefix:s,iconName:c}}const zrr=wtn.filter(n=>n!==X1||n!==Mye),Hrr=Object.keys(yVe).filter(n=>n!==X1).map(n=>Object.keys(yVe[n])).flat();function Grr(n){const{values:r,family:o,canonical:s,givenPrefix:c="",styles:d={},config:h={}}=n,b=o===Mye,y=r.includes("fa-duotone")||r.includes("fad"),E=h.familyDefault==="duotone",k=s.prefix==="fad"||s.prefix==="fa-duotone";if(!b&&(y||E||k)&&(s.prefix="fad"),(r.includes("fa-brands")||r.includes("fab"))&&(s.prefix="fab"),!s.prefix&&zrr.includes(o)&&(Object.keys(d).find(O=>Hrr.includes(O))||h.autoFetchSvg)){const O=Knr.get(o).defaultShortPrefixId;s.prefix=O,s.iconName=RB(s.prefix,s.iconName)||s.iconName}return(s.prefix==="fa"||c==="fa")&&(s.prefix=CM()||"fas"),s}class qrr{constructor(){this.definitions={}}add(){for(var r=arguments.length,o=new Array(r),s=0;s<r;s++)o[s]=arguments[s];const c=o.reduce(this._pullDefinitions,{});Object.keys(c).forEach(d=>{this.definitions[d]=Hr(Hr({},this.definitions[d]||{}),c[d]),AVe(d,c[d]);const h=ZYe[X1][d];h&&AVe(h,c[d]),jtn()})}reset(){this.definitions={}}_pullDefinitions(r,o){const s=o.prefix&&o.iconName&&o.icon?{0:o}:o;return Object.keys(s).map(c=>{const{prefix:d,iconName:h,icon:b}=s[c],y=b[2];r[d]||(r[d]={}),y.length>0&&y.forEach(E=>{typeof E=="string"&&(r[d][E]=b)}),r[d][h]=b}),r}}let cjt=[],lW={};const kW={},Vrr=Object.keys(kW);function Wrr(n,r){let{mixoutsTo:o}=r;return cjt=n,lW={},Object.keys(kW).forEach(s=>{Vrr.indexOf(s)===-1&&delete kW[s]}),cjt.forEach(s=>{const c=s.mixout?s.mixout():{};if(Object.keys(c).forEach(d=>{typeof c[d]=="function"&&(o[d]=c[d]),typeof c[d]=="object"&&Object.keys(c[d]).forEach(h=>{o[d]||(o[d]={}),o[d][h]=c[d][h]})}),s.hooks){const d=s.hooks();Object.keys(d).forEach(h=>{lW[h]||(lW[h]=[]),lW[h].push(d[h])})}s.provides&&s.provides(kW)}),o}function IVe(n,r){for(var o=arguments.length,s=new Array(o>2?o-2:0),c=2;c<o;c++)s[c-2]=arguments[c];return(lW[n]||[]).forEach(h=>{r=h.apply(null,[r,...s])}),r}function JB(n){for(var r=arguments.length,o=new Array(r>1?r-1:0),s=1;s<r;s++)o[s-1]=arguments[s];(lW[n]||[]).forEach(d=>{d.apply(null,o)})}function AM(){const n=arguments[0],r=Array.prototype.slice.call(arguments,1);return kW[n]?kW[n].apply(null,r):void 0}function OVe(n){n.prefix==="fa"&&(n.prefix="fas");let{iconName:r}=n;const o=n.prefix||CM();if(r)return r=RB(o,r)||r,ojt(Btn.definitions,o,r)||ojt(vI.styles,o,r)}const Btn=new qrr,Krr=()=>{Co.autoReplaceSvg=!1,Co.observeMutations=!1,JB("noAuto")},Yrr={i2svg:function(){let n=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};return Q4?(JB("beforeI2svg",n),AM("pseudoElements2svg",n),AM("i2svg",n)):Promise.reject(new Error("Operation requires a DOM of some kind."))},watch:function(){let n=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};const{autoReplaceSvgRoot:r}=n;Co.autoReplaceSvg===!1&&(Co.autoReplaceSvg=!0),Co.observeMutations=!0,Orr(()=>{Jrr({autoReplaceSvgRoot:r}),JB("watch",n)})}},Xrr={icon:n=>{if(n===null)return null;if(typeof n=="object"&&n.prefix&&n.iconName)return{prefix:n.prefix,iconName:RB(n.prefix,n.iconName)||n.iconName};if(Array.isArray(n)&&n.length===2){const r=n[1].indexOf("fa-")===0?n[1].slice(3):n[1],o=$ye(n[0]);return{prefix:o,iconName:RB(o,r)||r}}if(typeof n=="string"&&(n.indexOf("".concat(Co.cssPrefix,"-"))>-1||n.match(grr))){const r=Fye(n.split(" "),{skipLookups:!0});return{prefix:r.prefix||CM(),iconName:RB(r.prefix,r.iconName)||r.iconName}}if(typeof n=="string"){const r=CM();return{prefix:r,iconName:RB(r,n)||n}}}},E2={noAuto:Krr,config:Co,dom:Yrr,parse:Xrr,library:Btn,findIconDefinition:OVe,toHtml:Moe},Jrr=function(){let n=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};const{autoReplaceSvgRoot:r=qd}=n;(Object.keys(vI.styles).length>0||Co.autoFetchSvg)&&Q4&&Co.autoReplaceSvg&&E2.dom.i2svg({node:r})};function Bye(n,r){return Object.defineProperty(n,"abstract",{get:r}),Object.defineProperty(n,"html",{get:function(){return n.abstract.map(o=>Moe(o))}}),Object.defineProperty(n,"node",{get:function(){if(!Q4)return;const o=qd.createElement("div");return o.innerHTML=n.html,o.children}}),n}function Zrr(n){let{children:r,main:o,mask:s,attributes:c,styles:d,transform:h}=n;if(eXe(h)&&o.found&&!s.found){const{width:b,height:y}=o,E={x:b/y/2,y:.5};c.style=jye(Hr(Hr({},d),{},{"transform-origin":"".concat(E.x+h.x/16,"em ").concat(E.y+h.y/16,"em")}))}return[{tag:"svg",attributes:c,children:r}]}function Qrr(n){let{prefix:r,iconName:o,children:s,attributes:c,symbol:d}=n;const h=d===!0?"".concat(r,"-").concat(Co.cssPrefix,"-").concat(o):d;return[{tag:"svg",attributes:{style:"display: none;"},children:[{tag:"symbol",attributes:Hr(Hr({},c),{},{id:h}),children:s}]}]}function rXe(n){const{icons:{main:r,mask:o},prefix:s,iconName:c,transform:d,symbol:h,title:b,maskId:y,titleId:E,extra:k,watchable:x=!1}=n,{width:O,height:P}=o.found?o:r,D=Qnr.includes(s),$=[Co.replacementClass,c?"".concat(Co.cssPrefix,"-").concat(c):""].filter(ae=>k.classes.indexOf(ae)===-1).filter(ae=>ae!==""||!!ae).concat(k.classes).join(" ");let U={children:[],attributes:Hr(Hr({},k.attributes),{},{"data-prefix":s,"data-icon":c,class:$,role:k.attributes.role||"img",xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 ".concat(O," ").concat(P)})};const B=D&&!~k.classes.indexOf("fa-fw")?{width:"".concat(O/P*16*.0625,"em")}:{};x&&(U.attributes[XB]=""),b&&(U.children.push({tag:"title",attributes:{id:U.attributes["aria-labelledby"]||"title-".concat(E||Bie())},children:[b]}),delete U.attributes.title);const G=Hr(Hr({},U),{},{prefix:s,iconName:c,main:r,mask:o,maskId:y,transform:d,symbol:h,styles:Hr(Hr({},B),k.styles)}),{children:K,attributes:X}=o.found&&r.found?AM("generateAbstractMask",G)||{children:[],attributes:{}}:AM("generateAbstractIcon",G)||{children:[],attributes:{}};return G.children=K,G.attributes=X,h?Qrr(G):Zrr(G)}function ujt(n){const{content:r,width:o,height:s,transform:c,title:d,extra:h,watchable:b=!1}=n,y=Hr(Hr(Hr({},h.attributes),d?{title:d}:{}),{},{class:h.classes.join(" ")});b&&(y[XB]="");const E=Hr({},h.styles);eXe(c)&&(E.transform=Crr({transform:c,startCentered:!0,width:o,height:s}),E["-webkit-transform"]=E.transform);const k=jye(E);k.length>0&&(y.style=k);const x=[];return x.push({tag:"span",attributes:y,children:[r]}),d&&x.push({tag:"span",attributes:{class:"sr-only"},children:[d]}),x}function eir(n){const{content:r,title:o,extra:s}=n,c=Hr(Hr(Hr({},s.attributes),o?{title:o}:{}),{},{class:s.classes.join(" ")}),d=jye(s.styles);d.length>0&&(c.style=d);const h=[];return h.push({tag:"span",attributes:c,children:[r]}),o&&h.push({tag:"span",attributes:{class:"sr-only"},children:[o]}),h}const{styles:p$e}=vI;function NVe(n){const r=n[0],o=n[1],[s]=n.slice(4);let c=null;return Array.isArray(s)?c={tag:"g",attributes:{class:"".concat(Co.cssPrefix,"-").concat(d$e.GROUP)},children:[{tag:"path",attributes:{class:"".concat(Co.cssPrefix,"-").concat(d$e.SECONDARY),fill:"currentColor",d:s[0]}},{tag:"path",attributes:{class:"".concat(Co.cssPrefix,"-").concat(d$e.PRIMARY),fill:"currentColor",d:s[1]}}]}:c={tag:"path",attributes:{fill:"currentColor",d:s}},{found:!0,width:r,height:o,icon:c}}const tir={found:!1,width:512,height:512};function nir(n,r){!_tn&&!Co.showMissingIcons&&n&&console.error('Icon with name "'.concat(n,'" and prefix "').concat(r,'" is missing.'))}function RVe(n,r){let o=r;return r==="fa"&&Co.styleDefault!==null&&(r=CM()),new Promise((s,c)=>{if(o==="fa"){const d=$tn(n)||{};n=d.iconName||n,r=d.prefix||r}if(n&&r&&p$e[r]&&p$e[r][n]){const d=p$e[r][n];return s(NVe(d))}nir(n,r),s(Hr(Hr({},tir),{},{icon:Co.showMissingIcons&&n?AM("missingIconAbstract")||{}:{}}))})}const ljt=()=>{},PVe=Co.measurePerformance&&Z1e&&Z1e.mark&&Z1e.measure?Z1e:{mark:ljt,measure:ljt},Mre='FA "6.7.2"',rir=n=>(PVe.mark("".concat(Mre," ").concat(n," begins")),()=>Utn(n)),Utn=n=>{PVe.mark("".concat(Mre," ").concat(n," ends")),PVe.measure("".concat(Mre," ").concat(n),"".concat(Mre," ").concat(n," begins"),"".concat(Mre," ").concat(n," ends"))};var iXe={begin:rir,end:Utn};const dve=()=>{};function djt(n){return typeof(n.getAttribute?n.getAttribute(XB):null)=="string"}function iir(n){const r=n.getAttribute?n.getAttribute(XYe):null,o=n.getAttribute?n.getAttribute(JYe):null;return r&&o}function oir(n){return n&&n.classList&&n.classList.contains&&n.classList.contains(Co.replacementClass)}function air(){return Co.autoReplaceSvg===!0?fve.replace:fve[Co.autoReplaceSvg]||fve.replace}function sir(n){return qd.createElementNS("http://www.w3.org/2000/svg",n)}function cir(n){return qd.createElement(n)}function ztn(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const{ceFn:o=n.tag==="svg"?sir:cir}=r;if(typeof n=="string")return qd.createTextNode(n);const s=o(n.tag);return Object.keys(n.attributes||[]).forEach(function(d){s.setAttribute(d,n.attributes[d])}),(n.children||[]).forEach(function(d){s.appendChild(ztn(d,{ceFn:o}))}),s}function uir(n){let r=" ".concat(n.outerHTML," ");return r="".concat(r,"Font Awesome fontawesome.com "),r}const fve={replace:function(n){const r=n[0];if(r.parentNode)if(n[1].forEach(o=>{r.parentNode.insertBefore(ztn(o),r)}),r.getAttribute(XB)===null&&Co.keepOriginalSource){let o=qd.createComment(uir(r));r.parentNode.replaceChild(o,r)}else r.remove()},nest:function(n){const r=n[0],o=n[1];if(~QYe(r).indexOf(Co.replacementClass))return fve.replace(n);const s=new RegExp("".concat(Co.cssPrefix,"-.*"));if(delete o[0].attributes.id,o[0].attributes.class){const d=o[0].attributes.class.split(" ").reduce((h,b)=>(b===Co.replacementClass||b.match(s)?h.toSvg.push(b):h.toNode.push(b),h),{toNode:[],toSvg:[]});o[0].attributes.class=d.toSvg.join(" "),d.toNode.length===0?r.removeAttribute("class"):r.setAttribute("class",d.toNode.join(" "))}const c=o.map(d=>Moe(d)).join(`
`);r.setAttribute(XB,""),r.innerHTML=c}};function fjt(n){n()}function Htn(n,r){const o=typeof r=="function"?r:dve;if(n.length===0)o();else{let s=fjt;Co.mutateApproach===frr&&(s=xM.requestAnimationFrame||fjt),s(()=>{const c=air(),d=iXe.begin("mutate");n.map(c),d(),o()})}}let oXe=!1;function Gtn(){oXe=!0}function LVe(){oXe=!1}let twe=null;function hjt(n){if(!Q7t||!Co.observeMutations)return;const{treeCallback:r=dve,nodeCallback:o=dve,pseudoElementsCallback:s=dve,observeMutationsRoot:c=qd}=n;twe=new Q7t(d=>{if(oXe)return;const h=CM();OK(d).forEach(b=>{if(b.type==="childList"&&b.addedNodes.length>0&&!djt(b.addedNodes[0])&&(Co.searchPseudoElements&&s(b.target),r(b.target)),b.type==="attributes"&&b.target.parentNode&&Co.searchPseudoElements&&s(b.target.parentNode),b.type==="attributes"&&djt(b.target)&&~vrr.indexOf(b.attributeName))if(b.attributeName==="class"&&iir(b.target)){const{prefix:y,iconName:E}=Fye(QYe(b.target));b.target.setAttribute(XYe,y||h),E&&b.target.setAttribute(JYe,E)}else oir(b.target)&&o(b.target)})}),Q4&&twe.observe(c,{childList:!0,attributes:!0,characterData:!0,subtree:!0})}function lir(){twe&&twe.disconnect()}function dir(n){const r=n.getAttribute("style");let o=[];return r&&(o=r.split(";").reduce((s,c)=>{const d=c.split(":"),h=d[0],b=d.slice(1);return h&&b.length>0&&(s[h]=b.join(":").trim()),s},{})),o}function fir(n){const r=n.getAttribute("data-prefix"),o=n.getAttribute("data-icon"),s=n.innerText!==void 0?n.innerText.trim():"";let c=Fye(QYe(n));return c.prefix||(c.prefix=CM()),r&&o&&(c.prefix=r,c.iconName=o),c.iconName&&c.prefix||(c.prefix&&s.length>0&&(c.iconName=jrr(c.prefix,n.innerText)||nXe(c.prefix,CVe(n.innerText))),!c.iconName&&Co.autoFetchSvg&&n.firstChild&&n.firstChild.nodeType===Node.TEXT_NODE&&(c.iconName=n.firstChild.data)),c}function hir(n){const r=OK(n.attributes).reduce((c,d)=>(c.name!=="class"&&c.name!=="style"&&(c[d.name]=d.value),c),{}),o=n.getAttribute("title"),s=n.getAttribute("data-fa-title-id");return Co.autoA11y&&(o?r["aria-labelledby"]="".concat(Co.replacementClass,"-title-").concat(s||Bie()):(r["aria-hidden"]="true",r.focusable="false")),r}function pir(){return{iconName:null,title:null,titleId:null,prefix:null,transform:bI,symbol:!1,mask:{iconName:null,prefix:null,rest:[]},maskId:null,extra:{classes:[],styles:{},attributes:{}}}}function pjt(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{styleParser:!0};const{iconName:o,prefix:s,rest:c}=fir(n),d=hir(n),h=IVe("parseNodeAttributes",{},n);let b=r.styleParser?dir(n):[];return Hr({iconName:o,title:n.getAttribute("title"),titleId:n.getAttribute("data-fa-title-id"),prefix:s,transform:bI,mask:{iconName:null,prefix:null,rest:[]},maskId:null,symbol:!1,extra:{classes:c,styles:b,attributes:d}},h)}const{styles:gir}=vI;function qtn(n){const r=Co.autoReplaceSvg==="nest"?pjt(n,{styleParser:!1}):pjt(n);return~r.extra.classes.indexOf(Ttn)?AM("generateLayersText",n,r):AM("generateSvgReplacementMutation",n,r)}function mir(){return[...Xnr,...EVe]}function gjt(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:null;if(!Q4)return Promise.resolve();const o=qd.documentElement.classList,s=k=>o.add("".concat(njt,"-").concat(k)),c=k=>o.remove("".concat(njt,"-").concat(k)),d=Co.autoFetchSvg?mir():vtn.concat(Object.keys(gir));d.includes("fa")||d.push("fa");const h=[".".concat(Ttn,":not([").concat(XB,"])")].concat(d.map(k=>".".concat(k,":not([").concat(XB,"])"))).join(", ");if(h.length===0)return Promise.resolve();let b=[];try{b=OK(n.querySelectorAll(h))}catch{}if(b.length>0)s("pending"),c("complete");else return Promise.resolve();const y=iXe.begin("onTree"),E=b.reduce((k,x)=>{try{const O=qtn(x);O&&k.push(O)}catch(O){_tn||O.name==="MissingIcon"&&console.error(O)}return k},[]);return new Promise((k,x)=>{Promise.all(E).then(O=>{Htn(O,()=>{s("active"),s("complete"),c("pending"),typeof r=="function"&&r(),y(),k()})}).catch(O=>{y(),x(O)})})}function bir(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:null;qtn(n).then(o=>{o&&Htn([o],r)})}function vir(n){return function(r){let o=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const s=(r||{}).icon?r:OVe(r||{});let{mask:c}=o;return c&&(c=(c||{}).icon?c:OVe(c||{})),n(s,Hr(Hr({},o),{},{mask:c}))}}const wir=function(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const{transform:o=bI,symbol:s=!1,mask:c=null,maskId:d=null,title:h=null,titleId:b=null,classes:y=[],attributes:E={},styles:k={}}=r;if(!n)return;const{prefix:x,iconName:O,icon:P}=n;return Bye(Hr({type:"icon"},n),()=>(JB("beforeDOMElementCreation",{iconDefinition:n,params:r}),Co.autoA11y&&(h?E["aria-labelledby"]="".concat(Co.replacementClass,"-title-").concat(b||Bie()):(E["aria-hidden"]="true",E.focusable="false")),rXe({icons:{main:NVe(P),mask:c?NVe(c.icon):{found:!1,width:null,height:null,icon:{}}},prefix:x,iconName:O,transform:Hr(Hr({},bI),o),symbol:s,title:h,maskId:d,titleId:b,extra:{attributes:E,styles:k,classes:y}})))};var yir={mixout(){return{icon:vir(wir)}},hooks(){return{mutationObserverCallbacks(n){return n.treeCallback=gjt,n.nodeCallback=bir,n}}},provides(n){n.i2svg=function(r){const{node:o=qd,callback:s=()=>{}}=r;return gjt(o,s)},n.generateSvgReplacementMutation=function(r,o){const{iconName:s,title:c,titleId:d,prefix:h,transform:b,symbol:y,mask:E,maskId:k,extra:x}=o;return new Promise((O,P)=>{Promise.all([RVe(s,h),E.iconName?RVe(E.iconName,E.prefix):Promise.resolve({found:!1,width:512,height:512,icon:{}})]).then(D=>{let[$,U]=D;O([r,rXe({icons:{main:$,mask:U},prefix:h,iconName:s,transform:b,symbol:y,maskId:k,title:c,titleId:d,extra:x,watchable:!0})])}).catch(P)})},n.generateAbstractIcon=function(r){let{children:o,attributes:s,main:c,transform:d,styles:h}=r;const b=jye(h);b.length>0&&(s.style=b);let y;return eXe(d)&&(y=AM("generateAbstractTransformGrouping",{main:c,transform:d,containerWidth:c.width,iconWidth:c.width})),o.push(y||c.icon),{children:o,attributes:s}}}},Eir={mixout(){return{layer(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const{classes:o=[]}=r;return Bye({type:"layer"},()=>{JB("beforeDOMElementCreation",{assembler:n,params:r});let s=[];return n(c=>{Array.isArray(c)?c.map(d=>{s=s.concat(d.abstract)}):s=s.concat(c.abstract)}),[{tag:"span",attributes:{class:["".concat(Co.cssPrefix,"-layers"),...o].join(" ")},children:s}]})}}}},Sir={mixout(){return{counter(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const{title:o=null,classes:s=[],attributes:c={},styles:d={}}=r;return Bye({type:"counter",content:n},()=>(JB("beforeDOMElementCreation",{content:n,params:r}),eir({content:n.toString(),title:o,extra:{attributes:c,styles:d,classes:["".concat(Co.cssPrefix,"-layers-counter"),...s]}})))}}}},_ir={mixout(){return{text(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const{transform:o=bI,title:s=null,classes:c=[],attributes:d={},styles:h={}}=r;return Bye({type:"text",content:n},()=>(JB("beforeDOMElementCreation",{content:n,params:r}),ujt({content:n,transform:Hr(Hr({},bI),o),title:s,extra:{attributes:d,styles:h,classes:["".concat(Co.cssPrefix,"-layers-text"),...c]}})))}}},provides(n){n.generateLayersText=function(r,o){const{title:s,transform:c,extra:d}=o;let h=null,b=null;if(mtn){const y=parseInt(getComputedStyle(r).fontSize,10),E=r.getBoundingClientRect();h=E.width/y,b=E.height/y}return Co.autoA11y&&!s&&(d.attributes["aria-hidden"]="true"),Promise.resolve([r,ujt({content:r.innerHTML,width:h,height:b,transform:c,title:s,extra:d,watchable:!0})])}}};const kir=new RegExp('"',"ug"),mjt=[1105920,1112319],bjt=Hr(Hr(Hr(Hr({},{FontAwesome:{normal:"fas",400:"fas"}}),Wnr),lrr),rrr),DVe=Object.keys(bjt).reduce((n,r)=>(n[r.toLowerCase()]=bjt[r],n),{}),Tir=Object.keys(DVe).reduce((n,r)=>{const o=DVe[r];return n[r]=o[900]||[...Object.entries(o)][0][1],n},{});function xir(n){const r=n.replace(kir,""),o=Rrr(r,0),s=o>=mjt[0]&&o<=mjt[1],c=r.length===2?r[0]===r[1]:!1;return{value:CVe(c?r[0]:r),isSecondary:s||c}}function Cir(n,r){const o=n.replace(/^['"]|['"]$/g,"").toLowerCase(),s=parseInt(r),c=isNaN(s)?"normal":s;return(DVe[o]||{})[c]||Tir[o]}function vjt(n,r){const o="".concat(drr).concat(r.replace(":","-"));return new Promise((s,c)=>{if(n.getAttribute(o)!==null)return s();const h=OK(n.children).filter(O=>O.getAttribute(_Ve)===r)[0],b=xM.getComputedStyle(n,r),y=b.getPropertyValue("font-family"),E=y.match(mrr),k=b.getPropertyValue("font-weight"),x=b.getPropertyValue("content");if(h&&!E)return n.removeChild(h),s();if(E&&x!=="none"&&x!==""){const O=b.getPropertyValue("content");let P=Cir(y,k);const{value:D,isSecondary:$}=xir(O),U=E[0].startsWith("FontAwesome");let B=nXe(P,D),G=B;if(U){const K=$rr(D);K.iconName&&K.prefix&&(B=K.iconName,P=K.prefix)}if(B&&!$&&(!h||h.getAttribute(XYe)!==P||h.getAttribute(JYe)!==G)){n.setAttribute(o,G),h&&n.removeChild(h);const K=pir(),{extra:X}=K;X.attributes[_Ve]=r,RVe(B,P).then(ae=>{const oe=rXe(Hr(Hr({},K),{},{icons:{main:ae,mask:Ftn()},prefix:P,iconName:G,extra:X,watchable:!0})),he=qd.createElementNS("http://www.w3.org/2000/svg","svg");r==="::before"?n.insertBefore(he,n.firstChild):n.appendChild(he),he.outerHTML=oe.map(ye=>Moe(ye)).join(`
`),n.removeAttribute(o),s()}).catch(c)}else s()}else s()})}function Air(n){return Promise.all([vjt(n,"::before"),vjt(n,"::after")])}function Iir(n){return n.parentNode!==document.head&&!~hrr.indexOf(n.tagName.toUpperCase())&&!n.getAttribute(_Ve)&&(!n.parentNode||n.parentNode.tagName!=="svg")}function wjt(n){if(Q4)return new Promise((r,o)=>{const s=OK(n.querySelectorAll("*")).filter(Iir).map(Air),c=iXe.begin("searchPseudoElements");Gtn(),Promise.all(s).then(()=>{c(),LVe(),r()}).catch(()=>{c(),LVe(),o()})})}var Oir={hooks(){return{mutationObserverCallbacks(n){return n.pseudoElementsCallback=wjt,n}}},provides(n){n.pseudoElements2svg=function(r){const{node:o=qd}=r;Co.searchPseudoElements&&wjt(o)}}};let yjt=!1;var Nir={mixout(){return{dom:{unwatch(){Gtn(),yjt=!0}}}},hooks(){return{bootstrap(){hjt(IVe("mutationObserverCallbacks",{}))},noAuto(){lir()},watch(n){const{observeMutationsRoot:r}=n;yjt?LVe():hjt(IVe("mutationObserverCallbacks",{observeMutationsRoot:r}))}}}};const Ejt=n=>{let r={size:16,x:0,y:0,flipX:!1,flipY:!1,rotate:0};return n.toLowerCase().split(" ").reduce((o,s)=>{const c=s.toLowerCase().split("-"),d=c[0];let h=c.slice(1).join("-");if(d&&h==="h")return o.flipX=!0,o;if(d&&h==="v")return o.flipY=!0,o;if(h=parseFloat(h),isNaN(h))return o;switch(d){case"grow":o.size=o.size+h;break;case"shrink":o.size=o.size-h;break;case"left":o.x=o.x-h;break;case"right":o.x=o.x+h;break;case"up":o.y=o.y-h;break;case"down":o.y=o.y+h;break;case"rotate":o.rotate=o.rotate+h;break}return o},r)};var Rir={mixout(){return{parse:{transform:n=>Ejt(n)}}},hooks(){return{parseNodeAttributes(n,r){const o=r.getAttribute("data-fa-transform");return o&&(n.transform=Ejt(o)),n}}},provides(n){n.generateAbstractTransformGrouping=function(r){let{main:o,transform:s,containerWidth:c,iconWidth:d}=r;const h={transform:"translate(".concat(c/2," 256)")},b="translate(".concat(s.x*32,", ").concat(s.y*32,") "),y="scale(".concat(s.size/16*(s.flipX?-1:1),", ").concat(s.size/16*(s.flipY?-1:1),") "),E="rotate(".concat(s.rotate," 0 0)"),k={transform:"".concat(b," ").concat(y," ").concat(E)},x={transform:"translate(".concat(d/2*-1," -256)")},O={outer:h,inner:k,path:x};return{tag:"g",attributes:Hr({},O.outer),children:[{tag:"g",attributes:Hr({},O.inner),children:[{tag:o.icon.tag,children:o.icon.children,attributes:Hr(Hr({},o.icon.attributes),O.path)}]}]}}}};const g$e={x:0,y:0,width:"100%",height:"100%"};function Sjt(n){let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:!0;return n.attributes&&(n.attributes.fill||r)&&(n.attributes.fill="black"),n}function Pir(n){return n.tag==="g"?n.children:[n]}var Lir={hooks(){return{parseNodeAttributes(n,r){const o=r.getAttribute("data-fa-mask"),s=o?Fye(o.split(" ").map(c=>c.trim())):Ftn();return s.prefix||(s.prefix=CM()),n.mask=s,n.maskId=r.getAttribute("data-fa-mask-id"),n}}},provides(n){n.generateAbstractMask=function(r){let{children:o,attributes:s,main:c,mask:d,maskId:h,transform:b}=r;const{width:y,icon:E}=c,{width:k,icon:x}=d,O=xrr({transform:b,containerWidth:k,iconWidth:y}),P={tag:"rect",attributes:Hr(Hr({},g$e),{},{fill:"white"})},D=E.children?{children:E.children.map(Sjt)}:{},$={tag:"g",attributes:Hr({},O.inner),children:[Sjt(Hr({tag:E.tag,attributes:Hr(Hr({},E.attributes),O.path)},D))]},U={tag:"g",attributes:Hr({},O.outer),children:[$]},B="mask-".concat(h||Bie()),G="clip-".concat(h||Bie()),K={tag:"mask",attributes:Hr(Hr({},g$e),{},{id:B,maskUnits:"userSpaceOnUse",maskContentUnits:"userSpaceOnUse"}),children:[P,U]},X={tag:"defs",children:[{tag:"clipPath",attributes:{id:G},children:Pir(x)},K]};return o.push(X,{tag:"rect",attributes:Hr({fill:"currentColor","clip-path":"url(#".concat(G,")"),mask:"url(#".concat(B,")")},g$e)}),{children:o,attributes:s}}}},Dir={provides(n){let r=!1;xM.matchMedia&&(r=xM.matchMedia("(prefers-reduced-motion: reduce)").matches),n.missingIconAbstract=function(){const o=[],s={fill:"currentColor"},c={attributeType:"XML",repeatCount:"indefinite",dur:"2s"};o.push({tag:"path",attributes:Hr(Hr({},s),{},{d:"M156.5,447.7l-12.6,29.5c-18.7-9.5-35.9-21.2-51.5-34.9l22.7-22.7C127.6,430.5,141.5,440,156.5,447.7z M40.6,272H8.5 c1.4,21.2,5.4,41.7,11.7,61.1L50,321.2C45.1,305.5,41.8,289,40.6,272z M40.6,240c1.4-18.8,5.2-37,11.1-54.1l-29.5-12.6 C14.7,194.3,10,216.7,8.5,240H40.6z M64.3,156.5c7.8-14.9,17.2-28.8,28.1-41.5L69.7,92.3c-13.7,15.6-25.5,32.8-34.9,51.5 L64.3,156.5z M397,419.6c-13.9,12-29.4,22.3-46.1,30.4l11.9,29.8c20.7-9.9,39.8-22.6,56.9-37.6L397,419.6z M115,92.4 c13.9-12,29.4-22.3,46.1-30.4l-11.9-29.8c-20.7,9.9-39.8,22.6-56.8,37.6L115,92.4z M447.7,355.5c-7.8,14.9-17.2,28.8-28.1,41.5 l22.7,22.7c13.7-15.6,25.5-32.9,34.9-51.5L447.7,355.5z M471.4,272c-1.4,18.8-5.2,37-11.1,54.1l29.5,12.6 c7.5-21.1,12.2-43.5,13.6-66.8H471.4z M321.2,462c-15.7,5-32.2,8.2-49.2,9.4v32.1c21.2-1.4,41.7-5.4,61.1-11.7L321.2,462z M240,471.4c-18.8-1.4-37-5.2-54.1-11.1l-12.6,29.5c21.1,7.5,43.5,12.2,66.8,13.6V471.4z M462,190.8c5,15.7,8.2,32.2,9.4,49.2h32.1 c-1.4-21.2-5.4-41.7-11.7-61.1L462,190.8z M92.4,397c-12-13.9-22.3-29.4-30.4-46.1l-29.8,11.9c9.9,20.7,22.6,39.8,37.6,56.9 L92.4,397z M272,40.6c18.8,1.4,36.9,5.2,54.1,11.1l12.6-29.5C317.7,14.7,295.3,10,272,8.5V40.6z M190.8,50 c15.7-5,32.2-8.2,49.2-9.4V8.5c-21.2,1.4-41.7,5.4-61.1,11.7L190.8,50z M442.3,92.3L419.6,115c12,13.9,22.3,29.4,30.5,46.1 l29.8-11.9C470,128.5,457.3,109.4,442.3,92.3z M397,92.4l22.7-22.7c-15.6-13.7-32.8-25.5-51.5-34.9l-12.6,29.5 C370.4,72.1,384.4,81.5,397,92.4z"})});const d=Hr(Hr({},c),{},{attributeName:"opacity"}),h={tag:"circle",attributes:Hr(Hr({},s),{},{cx:"256",cy:"364",r:"28"}),children:[]};return r||h.children.push({tag:"animate",attributes:Hr(Hr({},c),{},{attributeName:"r",values:"28;14;28;28;14;28;"})},{tag:"animate",attributes:Hr(Hr({},d),{},{values:"1;0;1;1;0;1;"})}),o.push(h),o.push({tag:"path",attributes:Hr(Hr({},s),{},{opacity:"1",d:"M263.7,312h-16c-6.6,0-12-5.4-12-12c0-71,77.4-63.9,77.4-107.8c0-20-17.8-40.2-57.4-40.2c-29.1,0-44.3,9.6-59.2,28.7 c-3.9,5-11.1,6-16.2,2.4l-13.1-9.2c-5.6-3.9-6.9-11.8-2.6-17.2c21.2-27.2,46.4-44.7,91.2-44.7c52.3,0,97.4,29.8,97.4,80.2 c0,67.6-77.4,63.5-77.4,107.8C275.7,306.6,270.3,312,263.7,312z"}),children:r?[]:[{tag:"animate",attributes:Hr(Hr({},d),{},{values:"1;0;0;0;0;1;"})}]}),r||o.push({tag:"path",attributes:Hr(Hr({},s),{},{opacity:"0",d:"M232.5,134.5l7,168c0.3,6.4,5.6,11.5,12,11.5h9c6.4,0,11.7-5.1,12-11.5l7-168c0.3-6.8-5.2-12.5-12-12.5h-23 C237.7,122,232.2,127.7,232.5,134.5z"}),children:[{tag:"animate",attributes:Hr(Hr({},d),{},{values:"0;0;1;1;0;0;"})}]}),{tag:"g",attributes:{class:"missing"},children:o}}}},Mir={hooks(){return{parseNodeAttributes(n,r){const o=r.getAttribute("data-fa-symbol"),s=o===null?!1:o===""?!0:o;return n.symbol=s,n}}}},jir=[Irr,yir,Eir,Sir,_ir,Oir,Nir,Rir,Lir,Dir,Mir];Wrr(jir,{mixoutsTo:E2});E2.noAuto;E2.config;E2.library;E2.dom;const MVe=E2.parse;E2.findIconDefinition;E2.toHtml;const $ir=E2.icon;E2.layer;E2.text;E2.counter;var m$e={exports:{}},b$e,_jt;function Fir(){if(_jt)return b$e;_jt=1;var n="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED";return b$e=n,b$e}var v$e,kjt;function Bir(){if(kjt)return v$e;kjt=1;var n=Fir();function r(){}function o(){}return o.resetWarningCache=r,v$e=function(){function s(h,b,y,E,k,x){if(x!==n){var O=new Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw O.name="Invariant Violation",O}}s.isRequired=s;function c(){return s}var d={array:s,bigint:s,bool:s,func:s,number:s,object:s,string:s,symbol:s,any:s,arrayOf:c,element:s,elementType:s,instanceOf:c,node:s,objectOf:c,oneOf:c,oneOfType:c,shape:c,exact:c,checkPropTypes:o,resetWarningCache:r};return d.PropTypes=d,d},v$e}var Tjt;function Uir(){return Tjt||(Tjt=1,m$e.exports=Bir()()),m$e.exports}var zir=Uir();const qs=Y4(zir);var Hir={};function jVe(n,r){(r==null||r>n.length)&&(r=n.length);for(var o=0,s=Array(r);o<r;o++)s[o]=n[o];return s}function Gir(n){if(Array.isArray(n))return n}function qir(n){if(Array.isArray(n))return jVe(n)}function rM(n,r,o){return(r=Qir(r))in n?Object.defineProperty(n,r,{value:o,enumerable:!0,configurable:!0,writable:!0}):n[r]=o,n}function Vir(n){if(typeof Symbol<"u"&&n[Symbol.iterator]!=null||n["@@iterator"]!=null)return Array.from(n)}function Wir(n,r){var o=n==null?null:typeof Symbol<"u"&&n[Symbol.iterator]||n["@@iterator"];if(o!=null){var s,c,d,h,b=[],y=!0,E=!1;try{if(d=(o=o.call(n)).next,r!==0)for(;!(y=(s=d.call(o)).done)&&(b.push(s.value),b.length!==r);y=!0);}catch(k){E=!0,c=k}finally{try{if(!y&&o.return!=null&&(h=o.return(),Object(h)!==h))return}finally{if(E)throw c}}return b}}function Kir(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Yir(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function xjt(n,r){var o=Object.keys(n);if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(n);r&&(s=s.filter(function(c){return Object.getOwnPropertyDescriptor(n,c).enumerable})),o.push.apply(o,s)}return o}function dI(n){for(var r=1;r<arguments.length;r++){var o=arguments[r]!=null?arguments[r]:{};r%2?xjt(Object(o),!0).forEach(function(s){rM(n,s,o[s])}):Object.getOwnPropertyDescriptors?Object.defineProperties(n,Object.getOwnPropertyDescriptors(o)):xjt(Object(o)).forEach(function(s){Object.defineProperty(n,s,Object.getOwnPropertyDescriptor(o,s))})}return n}function Xir(n,r){if(n==null)return{};var o,s,c=Jir(n,r);if(Object.getOwnPropertySymbols){var d=Object.getOwnPropertySymbols(n);for(s=0;s<d.length;s++)o=d[s],r.indexOf(o)===-1&&{}.propertyIsEnumerable.call(n,o)&&(c[o]=n[o])}return c}function Jir(n,r){if(n==null)return{};var o={};for(var s in n)if({}.hasOwnProperty.call(n,s)){if(r.indexOf(s)!==-1)continue;o[s]=n[s]}return o}function Cjt(n,r){return Gir(n)||Wir(n,r)||Vtn(n,r)||Kir()}function $Ve(n){return qir(n)||Vir(n)||Vtn(n)||Yir()}function Zir(n,r){if(typeof n!="object"||!n)return n;var o=n[Symbol.toPrimitive];if(o!==void 0){var s=o.call(n,r);if(typeof s!="object")return s;throw new TypeError("@@toPrimitive must return a primitive value.")}return(r==="string"?String:Number)(n)}function Qir(n){var r=Zir(n,"string");return typeof r=="symbol"?r:r+""}function nwe(n){"@babel/helpers - typeof";return nwe=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(r){return typeof r}:function(r){return r&&typeof Symbol=="function"&&r.constructor===Symbol&&r!==Symbol.prototype?"symbol":typeof r},nwe(n)}function Vtn(n,r){if(n){if(typeof n=="string")return jVe(n,r);var o={}.toString.call(n).slice(8,-1);return o==="Object"&&n.constructor&&(o=n.constructor.name),o==="Map"||o==="Set"?Array.from(n):o==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(o)?jVe(n,r):void 0}}var eor="7.0.0-alpha1",FVe;try{var tor=require("@fortawesome/fontawesome-svg-core/package.json");FVe=tor.version}catch{FVe=Hir.FA_VERSION||"7.0.0-alpha8"}function nor(n){var r=n.beat,o=n.fade,s=n.beatFade,c=n.bounce,d=n.shake,h=n.flash,b=n.spin,y=n.spinPulse,E=n.spinReverse,k=n.pulse,x=n.fixedWidth,O=n.inverse,P=n.border,D=n.listItem,$=n.flip,U=n.size,B=n.rotation,G=n.pull,K=n.swapOpacity,X=n.rotateBy,ae=n.widthAuto,oe=ror(FVe,eor),he=rM(rM(rM(rM(rM(rM({"fa-beat":r,"fa-fade":o,"fa-beat-fade":s,"fa-bounce":c,"fa-shake":d,"fa-flash":h,"fa-spin":b,"fa-spin-reverse":E,"fa-spin-pulse":y,"fa-pulse":k,"fa-fw":x,"fa-inverse":O,"fa-border":P,"fa-li":D,"fa-flip":$===!0,"fa-flip-horizontal":$==="horizontal"||$==="both","fa-flip-vertical":$==="vertical"||$==="both"},"fa-".concat(U),typeof U<"u"&&U!==null),"fa-rotate-".concat(B),typeof B<"u"&&B!==null&&B!==0),"fa-pull-".concat(G),typeof G<"u"&&G!==null),"fa-swap-opacity",K),"fa-rotate-by",oe&&X),"fa-width-auto",oe&&ae);return Object.keys(he).map(function(ye){return he[ye]?ye:null}).filter(function(ye){return ye})}function ror(n,r){for(var o=n.split("-"),s=Cjt(o,2),c=s[0],d=s[1],h=r.split("-"),b=Cjt(h,2),y=b[0],E=b[1],k=c.split("."),x=y.split("."),O=0;O<Math.max(k.length,x.length);O++){var P=k[O]||"0",D=x[O]||"0",$=parseInt(P,10),U=parseInt(D,10);if($!==U)return $>U}for(var B=0;B<Math.max(k.length,x.length);B++){var G=k[B]||"0",K=x[B]||"0";if(G!==K&&G.length!==K.length)return G.length<K.length}return!(d&&!E)}function ior(n){return n=n-0,n===n}function Wtn(n){return ior(n)?n:(n=n.replace(/[\-_\s]+(.)?/g,function(r,o){return o?o.toUpperCase():""}),n.substr(0,1).toLowerCase()+n.substr(1))}var oor=["style"];function aor(n){return n.charAt(0).toUpperCase()+n.slice(1)}function sor(n){return n.split(";").map(function(r){return r.trim()}).filter(function(r){return r}).reduce(function(r,o){var s=o.indexOf(":"),c=Wtn(o.slice(0,s)),d=o.slice(s+1).trim();return c.startsWith("webkit")?r[aor(c)]=d:r[c]=d,r},{})}function Ktn(n,r){var o=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};if(typeof r=="string")return r;var s=(r.children||[]).map(function(y){return Ktn(n,y)}),c=Object.keys(r.attributes||{}).reduce(function(y,E){var k=r.attributes[E];switch(E){case"class":y.attrs.className=k,delete r.attributes.class;break;case"style":y.attrs.style=sor(k);break;default:E.indexOf("aria-")===0||E.indexOf("data-")===0?y.attrs[E.toLowerCase()]=k:y.attrs[Wtn(E)]=k}return y},{attrs:{}}),d=o.style,h=d===void 0?{}:d,b=Xir(o,oor);return c.attrs.style=dI(dI({},c.attrs.style),h),n.apply(void 0,[r.tag,dI(dI({},c.attrs),b)].concat($Ve(s)))}var Ytn=!1;try{Ytn=!0}catch{}function cor(){if(!Ytn&&console&&typeof console.error=="function"){var n;(n=console).error.apply(n,arguments)}}function Ajt(n){if(n&&nwe(n)==="object"&&n.prefix&&n.iconName&&n.icon)return n;if(MVe.icon)return MVe.icon(n);if(n===null)return null;if(n&&nwe(n)==="object"&&n.prefix&&n.iconName)return n;if(Array.isArray(n)&&n.length===2)return{prefix:n[0],iconName:n[1]};if(typeof n=="string")return{prefix:"fas",iconName:n}}function w$e(n,r){return Array.isArray(r)&&r.length>0||!Array.isArray(r)&&r?rM({},n,r):{}}var Ijt={border:!1,className:"",mask:null,maskId:null,fixedWidth:!1,inverse:!1,flip:!1,icon:null,listItem:!1,pull:null,pulse:!1,rotation:null,rotateBy:!1,size:null,spin:!1,spinPulse:!1,spinReverse:!1,beat:!1,fade:!1,beatFade:!1,bounce:!1,shake:!1,symbol:!1,title:"",titleId:null,transform:null,swapOpacity:!1,widthAuto:!1},aXe=Un.forwardRef(function(n,r){var o=dI(dI({},Ijt),n),s=o.icon,c=o.mask,d=o.symbol,h=o.className,b=o.title,y=o.titleId,E=o.maskId,k=Ajt(s),x=w$e("classes",[].concat($Ve(nor(o)),$Ve((h||"").split(" ")))),O=w$e("transform",typeof o.transform=="string"?MVe.transform(o.transform):o.transform),P=w$e("mask",Ajt(c)),D=$ir(k,dI(dI(dI(dI({},x),O),P),{},{symbol:d,title:b,titleId:y,maskId:E}));if(!D)return cor("Could not find icon",k),null;var $=D.abstract,U={ref:r};return Object.keys(o).forEach(function(B){Ijt.hasOwnProperty(B)||(U[B]=o[B])}),uor($[0],U)});aXe.displayName="FontAwesomeIcon";aXe.propTypes={beat:qs.bool,border:qs.bool,beatFade:qs.bool,bounce:qs.bool,className:qs.string,fade:qs.bool,flash:qs.bool,mask:qs.oneOfType([qs.object,qs.array,qs.string]),maskId:qs.string,fixedWidth:qs.bool,inverse:qs.bool,flip:qs.oneOf([!0,!1,"horizontal","vertical","both"]),icon:qs.oneOfType([qs.object,qs.array,qs.string]),listItem:qs.bool,pull:qs.oneOf(["right","left"]),pulse:qs.bool,rotation:qs.oneOf([0,90,180,270]),rotateBy:qs.bool,shake:qs.bool,size:qs.oneOf(["2xs","xs","sm","lg","xl","2xl","1x","2x","3x","4x","5x","6x","7x","8x","9x","10x"]),spin:qs.bool,spinPulse:qs.bool,spinReverse:qs.bool,symbol:qs.oneOfType([qs.bool,qs.string]),title:qs.string,titleId:qs.string,transform:qs.oneOfType([qs.string,qs.object]),swapOpacity:qs.bool,widthAuto:qs.bool};var uor=Ktn.bind(null,Un.createElement);function xa({color:n,...r}){const[o]=pen("colors",n?[n]:[]);return H.jsx(aXe,{...r,color:o||n})}/*!
 * Font Awesome Pro 6.7.2 by @fontawesome - https://fontawesome.com
 * License - https://fontawesome.com/license (Commercial License)
 * Copyright 2024 Fonticons, Inc.
 */const lor={prefix:"fal",iconName:"comments",icon:[640,512,[128490,61670],"f086","M32 176c0-74.8 73.7-144 176-144s176 69.2 176 144s-73.7 144-176 144c-15.3 0-30.6-1.9-46.3-5c-3.5-.7-7.1-.2-10.2 1.4c-6.1 3.1-12 6-18 8.7c-28.4 12.9-60.2 23.1-91.5 26c14.9-19 26.8-39.7 37.6-59.9c3.3-6.1 2.3-13.6-2.5-18.6C50 244.2 32 213.1 32 176zM208 0C93.1 0 0 78.9 0 176c0 44.2 19.8 80.1 46 110c-11.7 21-24 40.6-39.5 57.5c0 0 0 0 0 0l-.1 .1c-6.5 7-8.2 17.1-4.4 25.8C5.8 378.3 14.4 384 24 384c43 0 86.5-13.3 122.7-29.7c4.9-2.2 9.6-4.5 14.3-6.8c15.3 2.8 30.9 4.6 47 4.6c114.9 0 208-78.9 208-176S322.9 0 208 0zM447.4 160.5C541.6 167 608 233 608 304c0 37.1-18 68.2-45.1 96.6c-4.8 5-5.8 12.5-2.5 18.6c10.9 20.2 22.7 40.8 37.6 59.9c-31.3-3-63.2-13.2-91.5-26c-6-2.7-11.9-5.6-18-8.7c-3.2-1.6-6.8-2.1-10.2-1.4c-15.6 3.1-30.9 5-46.3 5c-68.2 0-123.6-30.7-153.1-73.3c-11 3-22.3 5.2-33.8 6.8C279 439.8 349.9 480 432 480c16.1 0 31.7-1.8 47-4.6c4.6 2.3 9.4 4.6 14.3 6.8C529.5 498.7 573 512 616 512c9.6 0 18.2-5.7 22-14.5c3.8-8.7 2-18.9-4.4-25.8l-.1-.1s0 0 0 0c-15.5-17-27.8-36.5-39.5-57.5c26.2-29.9 46-65.8 46-110c0-94.4-87.8-171.5-198.2-175.8c2.8 10.4 4.7 21.2 5.6 32.3z"]},Xtn={prefix:"fal",iconName:"circle-exclamation",icon:[512,512,["exclamation-circle"],"f06a","M256 32a224 224 0 1 1 0 448 224 224 0 1 1 0-448zm0 480A256 256 0 1 0 256 0a256 256 0 1 0 0 512zm0-384c-8.8 0-16 7.2-16 16l0 128c0 8.8 7.2 16 16 16s16-7.2 16-16l0-128c0-8.8-7.2-16-16-16zm24 224a24 24 0 1 0 -48 0 24 24 0 1 0 48 0z"]},dor=Xtn,hor={prefix:"fal",iconName:"ghost",icon:[384,512,[128123],"f6e2","M242.7 506c3.3 3.8 8.2 6 13.3 6s9.9-2.2 13.3-6l29.3-33.4c11.6-13.3 32.1-14 44.5-1.5l2.3 2.3c4.2 4.2 10 6.6 16 6.6c12.5 0 22.6-10.1 22.6-22.6L384 192C384 86 298 0 192 0S0 86 0 192L0 457.4C0 469.9 10.1 480 22.6 480c6 0 11.8-2.4 16-6.6l2.3-2.3c12.5-12.5 32.9-11.8 44.5 1.5L114.7 506c3.3 3.8 8.2 6 13.3 6s9.9-2.2 13.3-6l26.6-30.5c12.7-14.6 35.4-14.6 48.2 0L242.7 506zM32 438.1L32 192c0-88.4 71.6-160 160-160s160 71.6 160 160l0 246.1c-25.1-14-57.6-9.3-77.6 13.4L256 472.5l-15.8-18.1c-25.5-29.1-70.8-29.1-96.3 0L128 472.5l-18.4-21.1C89.6 428.7 57.1 424 32 438.1zM152 192a24 24 0 1 0 -48 0 24 24 0 1 0 48 0zm104 24a24 24 0 1 0 0-48 24 24 0 1 0 0 48z"]},por={prefix:"fal",iconName:"share-nodes",icon:[448,512,["share-alt"],"f1e0","M448 112c0 44.2-35.8 80-80 80c-22.9 0-43.6-9.6-58.1-25l-151 75.5c.8 4.4 1.1 8.9 1.1 13.5s-.4 9.1-1.1 13.5l151 75.5c14.6-15.4 35.2-25 58.1-25c44.2 0 80 35.8 80 80s-35.8 80-80 80s-80-35.8-80-80c0-9.7 1.7-19 4.9-27.7L147.2 299.5c-14.3 22-39 36.5-67.2 36.5c-44.2 0-80-35.8-80-80s35.8-80 80-80c28.2 0 52.9 14.5 67.2 36.5l145.7-72.9c-3.2-8.6-4.9-17.9-4.9-27.7c0-44.2 35.8-80 80-80s80 35.8 80 80zM80 304a48 48 0 1 0 0-96 48 48 0 1 0 0 96zM416 112a48 48 0 1 0 -96 0 48 48 0 1 0 96 0zM368 448a48 48 0 1 0 0-96 48 48 0 1 0 0 96z"]},gor=por,mor={prefix:"fal",iconName:"ear",icon:[384,512,[128066],"f5f0","M352 192c0-88.4-71.6-160-160-160S32 103.6 32 192l0 176c0 61.9 50.1 112 112 112s112-50.1 112-112c0-16.4 8.2-32.2 22.5-41.4C322.8 298.1 352 248.5 352 192zm32 0c0 67.8-35.1 127.4-88.2 161.5c-4.9 3.2-7.8 8.6-7.8 14.5c0 79.5-64.5 144-144 144S0 447.5 0 368L0 192C0 86 86 0 192 0S384 86 384 192zm-272 0l0 6.3c0 10.7 5.3 20.7 14.2 26.6l8.9 5.9c15.5 10.4 24.9 27.8 24.9 46.5c0 16.6-7.4 32.3-20.1 42.9l-33.7 28c-6.8 5.7-16.9 4.7-22.5-2s-4.7-16.9 2-22.5l33.7-28c5.4-4.5 8.6-11.2 8.6-18.3c0-8-4-15.4-10.6-19.9l-8.9-5.9C90.7 239.7 80 219.7 80 198.3l0-6.3c0-61.9 50.1-112 112-112s112 50.1 112 112l0 16c0 8.8-7.2 16-16 16s-16-7.2-16-16l0-16c0-44.2-35.8-80-80-80s-80 35.8-80 80z"]},bor={prefix:"fal",iconName:"circle-play",icon:[512,512,[61469,"play-circle"],"f144","M256 32a224 224 0 1 1 0 448 224 224 0 1 1 0-448zm0 480A256 256 0 1 0 256 0a256 256 0 1 0 0 512zm72.7-256L208 329.7l0-147.5L328.7 256zM212.5 147.5c-7.4-4.5-16.7-4.7-24.3-.5s-12.3 12.3-12.3 20.9l0 176c0 8.7 4.7 16.7 12.3 20.9s16.8 4.1 24.3-.5l144-88c7.1-4.4 11.5-12.1 11.5-20.5s-4.4-16.1-11.5-20.5l-144-88z"]},BVe=bor,vor={prefix:"fal",iconName:"circle-check",icon:[512,512,[61533,"check-circle"],"f058","M256 32a224 224 0 1 1 0 448 224 224 0 1 1 0-448zm0 480A256 256 0 1 0 256 0a256 256 0 1 0 0 512zM363.3 203.3c6.2-6.2 6.2-16.4 0-22.6s-16.4-6.2-22.6 0L224 297.4l-52.7-52.7c-6.2-6.2-16.4-6.2-22.6 0s-6.2 16.4 0 22.6l64 64c6.2 6.2 16.4 6.2 22.6 0l128-128z"]},sXe=vor,wor={prefix:"fal",iconName:"arrows-rotate",icon:[512,512,[128472,"refresh","sync"],"f021","M69.4 210.6C89.8 126.5 165.6 64 256 64c71.1 0 133.1 38.6 166.3 96L368 160c-8.8 0-16 7.2-16 16s7.2 16 16 16l80.7 0 15.3 0c8.8 0 16-7.2 16-16l0-96c0-8.8-7.2-16-16-16s-16 7.2-16 16l0 60.6C408.8 75.5 337.5 32 256 32C149.6 32 60.5 106.2 37.7 205.6C35.5 215.2 43.1 224 53 224c7.9 0 14.6-5.7 16.5-13.4zm373.2 90.9C422.2 385.5 346.4 448 256 448c-71.1 0-133.1-38.6-166.3-96l54.5 0c8.8 0 16-7.2 16-16s-7.2-16-16-16l-80.9 0-15.1 0c-8.8 0-16 7.2-16 16l0 96c0 8.8 7.2 16 16 16s16-7.2 16-16l0-60.2C103.4 436.6 174.7 480 256 480c106.4 0 195.5-74.2 218.3-173.6c2.2-9.6-5.4-18.4-15.3-18.4c-7.9 0-14.6 5.7-16.5 13.4z"]},Jtn=wor,yor={prefix:"fal",iconName:"question",icon:[320,512,[10067,10068,61736],"3f","M144 64c-53 0-96 43-96 96c0 8.8-7.2 16-16 16s-16-7.2-16-16C16 89.3 73.3 32 144 32l32 0c70.7 0 128 57.3 128 128c0 35.3-16.6 68.4-44.8 89.6l-60.8 45.6C184.3 305.8 176 322.4 176 340l0 12c0 8.8-7.2 16-16 16s-16-7.2-16-16l0-12c0-27.7 13-53.8 35.2-70.4L240 224c20.1-15.1 32-38.8 32-64c0-53-43-96-96-96l-32 0zm-8 384a24 24 0 1 1 48 0 24 24 0 1 1 -48 0z"]},NK={prefix:"fal",iconName:"code",icon:[640,512,[],"f121","M405.1 .8c-8.4-2.8-17.4 1.7-20.2 10.1l-160 480c-2.8 8.4 1.7 17.4 10.1 20.2s17.4-1.7 20.2-10.1l160-480c2.8-8.4-1.7-17.4-10.1-20.2zM172 117.4c-5.9-6.6-16-7.2-22.6-1.3l-144 128C2 247.1 0 251.4 0 256s2 8.9 5.4 12l144 128c6.6 5.9 16.7 5.3 22.6-1.3s5.3-16.7-1.3-22.6L40.1 256 170.6 140c6.6-5.9 7.2-16 1.3-22.6zm296.1 0c-5.9 6.6-5.3 16.7 1.3 22.6L599.9 256 469.4 372c-6.6 5.9-7.2 16-1.3 22.6s16 7.2 22.6 1.3l144-128c3.4-3 5.4-7.4 5.4-12s-2-8.9-5.4-12l-144-128c-6.6-5.9-16.7-5.3-22.6 1.3z"]},Ztn={prefix:"fal",iconName:"arrow-up-right-from-square",icon:[512,512,["external-link"],"f08e","M336 0c-8.8 0-16 7.2-16 16s7.2 16 16 16l121.4 0L212.7 276.7c-6.2 6.2-6.2 16.4 0 22.6s16.4 6.2 22.6 0L480 54.6 480 176c0 8.8 7.2 16 16 16s16-7.2 16-16l0-160c0-8.8-7.2-16-16-16L336 0zM64 32C28.7 32 0 60.7 0 96L0 448c0 35.3 28.7 64 64 64l352 0c35.3 0 64-28.7 64-64l0-144c0-8.8-7.2-16-16-16s-16 7.2-16 16l0 144c0 17.7-14.3 32-32 32L64 480c-17.7 0-32-14.3-32-32L32 96c0-17.7 14.3-32 32-32l144 0c8.8 0 16-7.2 16-16s-7.2-16-16-16L64 32z"]},Uye=Ztn,Eor={prefix:"fal",iconName:"heart",icon:[512,512,[128153,128154,128155,128156,128420,129293,129294,129505,9829,10084,61578],"f004","M244 130.6l-12-13.5-4.2-4.7c-26-29.2-65.3-42.8-103.8-35.8c-53.3 9.7-92 56.1-92 110.3l0 3.5c0 32.3 13.4 63.1 37.1 85.1L253 446.8c.8 .7 1.9 1.2 3 1.2s2.2-.4 3-1.2L443 275.5c23.6-22 37-52.8 37-85.1l0-3.5c0-54.2-38.7-100.6-92-110.3c-38.5-7-77.8 6.6-103.8 35.8l-4.2 4.7-12 13.5c-3 3.4-7.4 5.4-12 5.4s-8.9-2-12-5.4zm34.9-57.1C311 48.4 352.7 37.7 393.7 45.1C462.2 57.6 512 117.3 512 186.9l0 3.5c0 36-13.1 70.6-36.6 97.5c-3.4 3.8-6.9 7.5-10.7 11l-184 171.3c-.8 .8-1.7 1.5-2.6 2.2c-6.3 4.9-14.1 7.5-22.1 7.5c-9.2 0-18-3.5-24.8-9.7L47.2 299c-3.8-3.5-7.3-7.2-10.7-11C13.1 261 0 226.4 0 190.4l0-3.5C0 117.3 49.8 57.6 118.3 45.1c40.9-7.4 82.6 3.2 114.7 28.4c6.7 5.3 13 11.1 18.7 17.6l4.2 4.7 4.2-4.7c4.2-4.7 8.6-9.1 13.3-13.1c1.8-1.5 3.6-3 5.4-4.5z"]},Qtn={prefix:"fal",iconName:"circle-question",icon:[512,512,[62108,"question-circle"],"f059","M480 256A224 224 0 1 0 32 256a224 224 0 1 0 448 0zM0 256a256 256 0 1 1 512 0A256 256 0 1 1 0 256zm168.7-86.2c6.5-24.6 28.7-41.8 54.2-41.8l57.1 0c35.5 0 64 29 64 64.3c0 24-13.4 46.2-34.9 57.2L272 268.3l0 19.7c0 8.8-7.2 16-16 16s-16-7.2-16-16l0-29.5c0-6 3.4-11.5 8.7-14.3l45.8-23.4c10.7-5.4 17.5-16.6 17.5-28.7c0-17.8-14.4-32.3-32-32.3l-57.1 0c-10.9 0-20.5 7.4-23.2 17.9l-.2 .7c-2.2 8.5-11 13.7-19.5 11.4s-13.7-11-11.4-19.5l.2-.7zM232 352a24 24 0 1 1 48 0 24 24 0 1 1 -48 0z"]},Sor={prefix:"fal",iconName:"location-crosshairs",icon:[512,512,["location"],"f601","M256 0c8.8 0 16 7.2 16 16l0 48.7c93.3 7.7 167.6 82.1 175.3 175.3l48.7 0c8.8 0 16 7.2 16 16s-7.2 16-16 16l-48.7 0c-7.7 93.3-82.1 167.6-175.3 175.3l0 48.7c0 8.8-7.2 16-16 16s-16-7.2-16-16l0-48.7C146.7 439.6 72.4 365.3 64.7 272L16 272c-8.8 0-16-7.2-16-16s7.2-16 16-16l48.7 0C72.4 146.7 146.7 72.4 240 64.7L240 16c0-8.8 7.2-16 16-16zM96 256a160 160 0 1 0 320 0A160 160 0 1 0 96 256zm224 0a64 64 0 1 0 -128 0 64 64 0 1 0 128 0zm-160 0a96 96 0 1 1 192 0 96 96 0 1 1 -192 0z"]},UVe={prefix:"fal",iconName:"comment",icon:[512,512,[128489,61669],"f075","M256 64C125.8 64 32 148.6 32 240c0 37.1 15.5 70.6 40 100c5.2 6.3 8.4 14.8 7.4 23.9c-3.1 27-11.4 52.5-25.7 76.3c-.5 .9-1.1 1.8-1.6 2.6c11.1-2.9 22.2-7 32.7-11.5L91.2 446l-6.4-14.7c17-7.4 33-16.7 48.4-27.4c8.5-5.9 19.4-7.5 29.2-4.2C193 410.1 224.1 416 256 416c130.2 0 224-84.6 224-176s-93.8-176-224-176zM0 240C0 125.2 114.5 32 256 32s256 93.2 256 208s-114.5 208-256 208c-36 0-70.5-6.7-103.8-17.9c-.2-.1-.5 0-.7 .1c-16.9 11.7-34.7 22.1-53.9 30.5C73.6 471.1 44.7 480 16 480c-6.5 0-12.3-3.9-14.8-9.8s-1.1-12.8 3.4-17.4c8.1-8.2 15.2-18.2 21.7-29c11.7-19.6 18.7-40.6 21.3-63.1c0 0-.1-.1-.1-.2C19.6 327.1 0 286.6 0 240z"]},_or={prefix:"fal",iconName:"paperclip",icon:[448,512,[128206],"f0c6","M380.7 67.3c-29.1-29.1-76.3-29.1-105.4 0l-192 192c-46.8 46.8-46.8 122.6 0 169.4s122.6 46.8 169.4 0l152-152c6.2-6.2 16.4-6.2 22.6 0s6.2 16.4 0 22.6l-152 152c-59.3 59.3-155.4 59.3-214.6 0S1.4 296 60.7 236.7l192-192c41.6-41.6 109-41.6 150.6 0s41.6 109 0 150.6L219.5 379.1c-28.2 28.2-74.6 25.6-99.6-5.5c-21.3-26.6-19.2-65 4.9-89.1L276.7 132.7c6.2-6.2 16.4-6.2 22.6 0s6.2 16.4 0 22.6L147.5 307.1c-12.6 12.6-13.7 32.6-2.6 46.5c13 16.2 37.2 17.6 51.9 2.9L380.7 172.7c29.1-29.1 29.1-76.3 0-105.4z"]},kor={prefix:"fal",iconName:"arrow-rotate-left",icon:[512,512,[8634,"arrow-left-rotate","arrow-rotate-back","arrow-rotate-backward","undo"],"f0e2","M48 192c-8.8 0-16-7.2-16-16L32 48c0-8.8 7.2-16 16-16s16 7.2 16 16l0 92.6C103.2 75.5 174.5 32 256 32c123.7 0 224 100.3 224 224s-100.3 224-224 224c-79.9 0-150-41.8-189.7-104.8C59.8 364.8 67.7 352 80 352c6 0 11.4 3.2 14.7 8.2C128.9 413 188.4 448 256 448c106 0 192-86 192-192s-86-192-192-192c-71.1 0-133.1 38.6-166.3 96l86.3 0c8.8 0 16 7.2 16 16s-7.2 16-16 16L48 192z"]},Tor=kor,enn={prefix:"fal",iconName:"minus",icon:[448,512,[8211,8722,10134,"subtract"],"f068","M432 256c0 8.8-7.2 16-16 16L32 272c-8.8 0-16-7.2-16-16s7.2-16 16-16l384 0c8.8 0 16 7.2 16 16z"]},xor={prefix:"fal",iconName:"arrow-right-arrow-left",icon:[448,512,[8644,"exchange"],"f0ec","M443.3 139.3c6.2-6.2 6.2-16.4 0-22.6l-96-96c-6.2-6.2-16.4-6.2-22.6 0s-6.2 16.4 0 22.6L393.4 112 16 112c-8.8 0-16 7.2-16 16s7.2 16 16 16l377.4 0-68.7 68.7c-6.2 6.2-6.2 16.4 0 22.6s16.4 6.2 22.6 0l96-96zm-342.6 352c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6L54.6 400 432 400c8.8 0 16-7.2 16-16s-7.2-16-16-16L54.6 368l68.7-68.7c6.2-6.2 6.2-16.4 0-22.6s-16.4-6.2-22.6 0l-96 96c-6.2 6.2-6.2 16.4 0 22.6l96 96z"]},Cor={prefix:"fal",iconName:"code-merge",icon:[448,512,[],"f387","M80 32a48 48 0 1 1 0 96 48 48 0 1 1 0-96zM96.9 158.2C133 150.5 160 118.4 160 80c0-44.2-35.8-80-80-80S0 35.8 0 80c0 38.7 27.5 71 64 78.4l0 195.2C27.5 361 0 393.3 0 432c0 44.2 35.8 80 80 80s80-35.8 80-80c0-38.7-27.5-71-64-78.4l0-119.1c26.4 32.6 66.8 53.5 112 53.5l81.6 0c7.4 36.5 39.7 64 78.4 64c44.2 0 80-35.8 80-80s-35.8-80-80-80c-38.7 0-71 27.5-78.4 64L208 256c-57 0-104.1-42.6-111.1-97.8zM416 272a48 48 0 1 1 -96 0 48 48 0 1 1 96 0zM128 432a48 48 0 1 1 -96 0 48 48 0 1 1 96 0z"]},Aor={prefix:"fal",iconName:"hand-wave",icon:[512,512,[],"e1a7","M354.3 95.8c8.7-1.2 14.8-9.4 13.6-18.1l-1.2-8.5c-5-35.2-32.7-62.9-67.9-67.9L290.3 .2c-8.7-1.2-16.9 4.8-18.1 13.6s4.8 16.9 13.6 18.1l8.5 1.2c21.1 3 37.7 19.6 40.7 40.7l1.2 8.5c1.3 8.7 9.4 14.8 18.1 13.6zM13.7 368.2C5 369.4-1.1 377.5 .2 386.3l1.2 8.5c5 35.2 32.7 62.9 67.9 67.9l8.5 1.2c8.7 1.3 16.9-4.8 18.1-13.6s-4.8-16.9-13.6-18.1L73.8 431c-21.1-3-37.7-19.6-40.7-40.7l-1.2-8.5c-1.2-8.7-9.4-14.8-18.1-13.6zM150.4 62.4c-21-10.8-47.5-7.3-65.1 10.3s-21 44.1-10.3 65.1c-8.7 2.6-16.8 7.3-23.7 14.1c-21.9 21.9-21.9 57.3 0 79.2l9.5 9.5c-7.6 2.7-14.8 7.1-20.9 13.2c-21.8 21.8-21.9 57.2-.2 79.1L162.9 457.1c34.8 35.1 82.2 54.9 131.7 54.9C397 512 480 429 480 326.6l0-13.8c0 0 0 0 0 0l0-.7c0 0 0 0 0 0L480 152c0-30.9-25.1-56-56-56s-56 25.1-56 56l0 11L243.7 38.7c-21.9-21.9-57.3-21.9-79.2 0c-6.8 6.8-11.5 15-14.1 23.7zM108 95.3c9.4-9.4 24.6-9.4 33.9 0l8.8 8.8 .1 .1L288 241.4c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6L187.1 95.3l-.1-.1c-9.3-9.4-9.3-24.5 .1-33.9c9.4-9.4 24.6-9.4 33.9 0L372.7 212.9c4.6 4.6 11.5 5.9 17.4 3.5s9.9-8.3 9.9-14.8l0-49.6c0-13.3 10.7-24 24-24s24 10.7 24 24l0 160c0 0 0 0 0 0l0 .7c0 0 0 0 0 0l0 13.9C448 411.3 379.3 480 294.6 480c-40.9 0-80.1-16.4-109-45.4L62.6 310.3c-9.3-9.4-9.3-24.6 .1-34c9.4-9.4 24.7-9.4 34.1 0l36 36 .1 .1 42.1 42.1c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6l-42.2-42.2L105.6 240c0 0 0 0-.1-.1L74 208.4c-9.4-9.4-9.4-24.6 0-33.9c9.3-9.3 24.4-9.4 33.8-.1L231.4 297.9c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6l-46.5-46.5-.1-.1L108 129.2c-9.4-9.4-9.4-24.6 0-33.9z"]},Ior={prefix:"fal",iconName:"thumbs-down",icon:[512,512,[128078,61576],"f165","M288.8 430.3c3.5 12.8 16.7 20.3 29.5 16.8s20.3-16.7 16.8-29.5l-4.5-16.4c-5.5-20.2-13.9-39.3-24.7-56.9c-3.1-4.9-3.2-11.1-.4-16.2s8.2-8.2 14-8.2L448 320c17.7 0 32-14.3 32-32c0-11.3-5.9-21.3-14.8-27c-7.2-4.6-9.5-13.9-5.3-21.3c2.6-4.6 4.1-10 4.1-15.7c0-12.4-7-23.1-17.3-28.5c-4.2-2.2-7.3-6.1-8.3-10.8s.1-9.5 3-13.2c4.2-5.4 6.7-12.2 6.7-19.5c0-14.2-9.2-26.3-22.1-30.4c-7.8-2.5-12.4-10.6-10.7-18.6c.5-2.2 .7-4.5 .7-6.9c0-17.7-14.3-32-32-32l-89.5 0c-15.8 0-31.2 4.7-44.4 13.4l-38.5 25.7c-9 6-16.6 13.7-22.4 22.6c-4.9 7.4-14.8 9.4-22.2 4.6s-9.4-14.8-4.6-22.2c8.1-12.3 18.7-23.1 31.4-31.6l38.5-25.7c18.4-12.3 40-18.8 62.1-18.8L384 32c35.3 0 64 28.7 64 64l0 .6c19.1 11.1 32 31.7 32 55.4c0 8.7-1.8 17.1-4.9 24.7C487.9 188.4 496 205.2 496 224c0 6.5-1 12.8-2.8 18.7C504.8 254.3 512 270.3 512 288c0 35.3-28.7 64-64 64l-101.6 0c6.2 13.1 11.3 26.7 15.1 40.9l4.5 16.4c8.1 29.8-9.5 60.6-39.3 68.8s-60.6-9.5-68.8-39.3l-4.5-16.4c-8.9-32.6-29.6-60.8-58.2-79l-3.1-2 8.2-12.9-8.2 12.9c-11.8-7.5-21.7-17.1-29.5-28.2c-5.1-7.2-3.3-17.2 4-22.3s17.2-3.3 22.3 4c5.4 7.7 12.2 14.4 20.4 19.5l3.1 2c35.3 22.4 60.9 57.2 71.9 97.5l4.5 16.4zM32 352l64 0 0-224-64 0 0 224zM0 352L0 128c0-17.7 14.3-32 32-32l64 0c17.7 0 32 14.3 32 32l0 224c0 17.7-14.3 32-32 32l-64 0c-17.7 0-32-14.3-32-32z"]},Oor={prefix:"fal",iconName:"file",icon:[384,512,[128196,128459,61462],"f15b","M352 448l0-256-112 0c-26.5 0-48-21.5-48-48l0-112L64 32C46.3 32 32 46.3 32 64l0 384c0 17.7 14.3 32 32 32l256 0c17.7 0 32-14.3 32-32zm-.5-288c-.7-2.8-2.1-5.4-4.2-7.4L231.4 36.7c-2.1-2.1-4.6-3.5-7.4-4.2L224 144c0 8.8 7.2 16 16 16l111.5 0zM0 64C0 28.7 28.7 0 64 0L220.1 0c12.7 0 24.9 5.1 33.9 14.1L369.9 129.9c9 9 14.1 21.2 14.1 33.9L384 448c0 35.3-28.7 64-64 64L64 512c-35.3 0-64-28.7-64-64L0 64z"]},Nor={prefix:"fal",iconName:"arrow-down",icon:[384,512,[8595],"f063","M180.7 475.3c6.2 6.2 16.4 6.2 22.6 0l176-176c6.2-6.2 6.2-16.4 0-22.6s-16.4-6.2-22.6 0L208 425.4 208 48c0-8.8-7.2-16-16-16s-16 7.2-16 16l0 377.4L27.3 276.7c-6.2-6.2-16.4-6.2-22.6 0s-6.2 16.4 0 22.6l176 176z"]},cXe={prefix:"fal",iconName:"check-to-slot",icon:[576,512,["vote-yea"],"f772","M144 64c-8.8 0-16 7.2-16 16l0 304 320 0 0-304c0-8.8-7.2-16-16-16L144 64zM480 80l0 304 16 0c8.8 0 16 7.2 16 16s-7.2 16-16 16l-16 0-32 0-320 0-32 0-16 0c-8.8 0-16-7.2-16-16s7.2-16 16-16l16 0L96 80c0-26.5 21.5-48 48-48l288 0c26.5 0 48 21.5 48 48zM48 320l16 0 0 32-16 0c-8.8 0-16 7.2-16 16l0 64c0 8.8 7.2 16 16 16l480 0c8.8 0 16-7.2 16-16l0-64c0-8.8-7.2-16-16-16l-16 0 0-32 16 0c26.5 0 48 21.5 48 48l0 64c0 26.5-21.5 48-48 48L48 480c-26.5 0-48-21.5-48-48l0-64c0-26.5 21.5-48 48-48zM379.3 174L272.6 280.6c-6.2 6.2-16.4 6.2-22.6 0l-53.3-53.3c-6.2-6.2-6.2-16.4 0-22.6s16.4-6.2 22.6 0l42 42 95.4-95.4c6.2-6.2 16.4-6.2 22.6 0s6.2 16.4 0 22.6z"]},Ror={prefix:"fal",iconName:"link",icon:[640,512,[128279,"chain"],"f0c1","M591.5 256c50-50 50-131 0-181s-131-50-181 0L387.9 97.6c-6.2 6.2-6.2 16.4 0 22.6s16.4 6.2 22.6 0l22.6-22.6c37.5-37.5 98.3-37.5 135.8 0s37.5 98.3 0 135.8L444.3 357.9c-37.4 37.4-98.1 37.4-135.6 0c-35.6-35.6-37.6-92.6-4.7-130.6l5.3-6.1c5.8-6.7 5.1-16.8-1.6-22.6s-16.8-5.1-22.6 1.6l-5.3 6.1c-43.9 50.7-41.2 126.7 6.2 174.1c49.9 49.9 130.9 49.9 180.8 0L591.5 256zM48.5 256c-50 50-50 131 0 181s131 50 181 0l22.6-22.6c6.2-6.2 6.2-16.4 0-22.6s-16.4-6.2-22.6 0l-22.6 22.6c-37.5 37.5-98.3 37.5-135.8 0s-37.5-98.3 0-135.8L195.7 154.1c37.4-37.4 98.1-37.4 135.6 0c35.6 35.6 37.6 92.6 4.7 130.6l-5.3 6.1c-5.8 6.7-5.1 16.8 1.6 22.6s16.8 5.1 22.6-1.6l5.3-6.1c43.9-50.7 41.2-126.7-6.2-174.1C303.9 81.5 223 81.5 173 131.4L48.5 256z"]},Por={prefix:"fal",iconName:"plus",icon:[448,512,[10133,61543,"add"],"2b","M240 64c0-8.8-7.2-16-16-16s-16 7.2-16 16l0 176L32 240c-8.8 0-16 7.2-16 16s7.2 16 16 16l176 0 0 176c0 8.8 7.2 16 16 16s16-7.2 16-16l0-176 176 0c8.8 0 16-7.2 16-16s-7.2-16-16-16l-176 0 0-176z"]},Lor={prefix:"fal",iconName:"expand",icon:[448,512,[],"f065","M144 32c8.8 0 16 7.2 16 16s-7.2 16-16 16L32 64l0 112c0 8.8-7.2 16-16 16s-16-7.2-16-16L0 48c0-8.8 7.2-16 16-16l128 0zM0 336c0-8.8 7.2-16 16-16s16 7.2 16 16l0 112 112 0c8.8 0 16 7.2 16 16s-7.2 16-16 16L16 480c-8.8 0-16-7.2-16-16L0 336zM432 32c8.8 0 16 7.2 16 16l0 128c0 8.8-7.2 16-16 16s-16-7.2-16-16l0-112L304 64c-8.8 0-16-7.2-16-16s7.2-16 16-16l128 0zM416 336c0-8.8 7.2-16 16-16s16 7.2 16 16l0 128c0 8.8-7.2 16-16 16l-128 0c-8.8 0-16-7.2-16-16s7.2-16 16-16l112 0 0-112z"]},tnn={prefix:"fal",iconName:"xmark",icon:[384,512,[128473,10005,10006,10060,215,"close","multiply","remove","times"],"f00d","M324.5 411.1c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6L214.6 256 347.1 123.5c6.2-6.2 6.2-16.4 0-22.6s-16.4-6.2-22.6 0L192 233.4 59.6 100.9c-6.2-6.2-16.4-6.2-22.6 0s-6.2 16.4 0 22.6L169.4 256 36.9 388.5c-6.2 6.2-6.2 16.4 0 22.6s16.4 6.2 22.6 0L192 278.6 324.5 411.1z"]},Dor={prefix:"fal",iconName:"chevron-left",icon:[320,512,[9001],"f053","M20.7 267.3c-6.2-6.2-6.2-16.4 0-22.6l192-192c6.2-6.2 16.4-6.2 22.6 0s6.2 16.4 0 22.6L54.6 256 235.3 436.7c6.2 6.2 6.2 16.4 0 22.6s-16.4 6.2-22.6 0l-192-192z"]},nnn={prefix:"fal",iconName:"chevron-right",icon:[320,512,[9002],"f054","M299.3 244.7c6.2 6.2 6.2 16.4 0 22.6l-192 192c-6.2 6.2-16.4 6.2-22.6 0s-6.2-16.4 0-22.6L265.4 256 84.7 75.3c-6.2-6.2-6.2-16.4 0-22.6s16.4-6.2 22.6 0l192 192z"]},Mor={prefix:"fal",iconName:"clock-rotate-left",icon:[512,512,["history"],"f1da","M32 132l0-84c0-8.8-7.2-16-16-16S0 39.2 0 48L0 176c0 8.8 7.2 16 16 16l128 0c8.8 0 16-7.2 16-16s-7.2-16-16-16l-90.4 0C89.5 84.3 166.7 32 256 32c123.7 0 224 100.3 224 224s-100.3 224-224 224c-73.3 0-138.3-35.2-179.2-89.6c-5.3-7.1-15.3-8.5-22.4-3.2s-8.5 15.3-3.2 22.4C97.9 471.8 172.2 512 256 512c141.4 0 256-114.6 256-256S397.4 0 256 0C159.6 0 75.7 53.3 32 132zm224-4c-8.8 0-16 7.2-16 16l0 112c0 4.2 1.7 8.3 4.7 11.3l80 80c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6L272 249.4 272 144c0-8.8-7.2-16-16-16z"]},jor={prefix:"fal",iconName:"clone",icon:[512,512,[],"f24d","M64 480l224 0c17.7 0 32-14.3 32-32l0-64 32 0 0 64c0 35.3-28.7 64-64 64L64 512c-35.3 0-64-28.7-64-64L0 224c0-35.3 28.7-64 64-64l64 0 0 32-64 0c-17.7 0-32 14.3-32 32l0 224c0 17.7 14.3 32 32 32zM224 320l224 0c17.7 0 32-14.3 32-32l0-224c0-17.7-14.3-32-32-32L224 32c-17.7 0-32 14.3-32 32l0 224c0 17.7 14.3 32 32 32zm-64-32l0-224c0-35.3 28.7-64 64-64L448 0c35.3 0 64 28.7 64 64l0 224c0 35.3-28.7 64-64 64l-224 0c-35.3 0-64-28.7-64-64z"]},uXe={prefix:"fal",iconName:"check",icon:[448,512,[10003,10004],"f00c","M443.3 100.7c6.2 6.2 6.2 16.4 0 22.6l-272 272c-6.2 6.2-16.4 6.2-22.6 0l-144-144c-6.2-6.2-6.2-16.4 0-22.6s16.4-6.2 22.6 0L160 361.4 420.7 100.7c6.2-6.2 16.4-6.2 22.6 0z"]},$or={prefix:"fal",iconName:"triangle-exclamation",icon:[512,512,[9888,"exclamation-triangle","warning"],"f071","M34.5 420.4c-1.6 2.8-2.5 6-2.5 9.3c0 10.2 8.2 18.4 18.4 18.4l411.2 0c10.2 0 18.4-8.2 18.4-18.4c0-3.3-.9-6.4-2.5-9.3L276.5 75.8C272.2 68.5 264.4 64 256 64s-16.2 4.5-20.5 11.8L34.5 420.4zM6.9 404.2l201-344.6C217.9 42.5 236.2 32 256 32s38.1 10.5 48.1 27.6l201 344.6c4.5 7.7 6.9 16.5 6.9 25.4c0 27.8-22.6 50.4-50.4 50.4L50.4 480C22.6 480 0 457.4 0 429.6c0-8.9 2.4-17.7 6.9-25.4zM256 160c8.8 0 16 7.2 16 16l0 128c0 8.8-7.2 16-16 16s-16-7.2-16-16l0-128c0-8.8 7.2-16 16-16zM232 384a24 24 0 1 1 48 0 24 24 0 1 1 -48 0z"]},rnn={prefix:"fal",iconName:"circle-xmark",icon:[512,512,[61532,"times-circle","xmark-circle"],"f057","M256 32a224 224 0 1 1 0 448 224 224 0 1 1 0-448zm0 480A256 256 0 1 0 256 0a256 256 0 1 0 0 512zM180.7 180.7c-6.2 6.2-6.2 16.4 0 22.6L233.4 256l-52.7 52.7c-6.2 6.2-6.2 16.4 0 22.6s16.4 6.2 22.6 0L256 278.6l52.7 52.7c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6L278.6 256l52.7-52.7c6.2-6.2 6.2-16.4 0-22.6s-16.4-6.2-22.6 0L256 233.4l-52.7-52.7c-6.2-6.2-16.4-6.2-22.6 0z"]},inn={prefix:"fal",iconName:"thumbs-up",icon:[512,512,[128077,61575],"f164","M288.8 81.7c3.5-12.8 16.7-20.3 29.5-16.8s20.3 16.7 16.8 29.5l-4.5 16.4c-5.5 20.2-13.9 39.3-24.7 56.9c-3.1 4.9-3.2 11.1-.4 16.2s8.2 8.2 14 8.2L448 192c17.7 0 32 14.3 32 32c0 11.3-5.9 21.3-14.8 27c-7.2 4.6-9.5 13.9-5.3 21.3c2.6 4.6 4.1 10 4.1 15.7c0 12.4-7 23.1-17.3 28.5c-4.2 2.2-7.3 6.1-8.3 10.8s.1 9.5 3 13.2c4.2 5.4 6.7 12.2 6.7 19.5c0 14.2-9.2 26.3-22.1 30.4c-7.8 2.5-12.4 10.6-10.7 18.6c.5 2.2 .7 4.5 .7 6.9c0 17.7-14.3 32-32 32l-89.5 0c-15.8 0-31.2-4.7-44.4-13.4l-38.5-25.7c-9-6-16.6-13.7-22.4-22.6c-4.9-7.4-14.8-9.4-22.2-4.6s-9.4 14.8-4.6 22.2c8.1 12.3 18.7 23.1 31.4 31.6l38.5 25.7c18.4 12.3 40 18.8 62.1 18.8l89.5 0c35.3 0 64-28.7 64-64l0-.6c19.1-11.1 32-31.7 32-55.4c0-8.7-1.8-17.1-4.9-24.7C487.9 323.6 496 306.8 496 288c0-6.5-1-12.8-2.8-18.7C504.8 257.7 512 241.7 512 224c0-35.3-28.7-64-64-64l-101.6 0c6.2-13.1 11.3-26.7 15.1-40.9l4.5-16.4c8.1-29.8-9.5-60.6-39.3-68.8s-60.6 9.5-68.8 39.3l-4.5 16.4c-8.9 32.6-29.6 60.8-58.2 79l-3.1 2c-11.8 7.5-21.7 17.1-29.5 28.2c-5.1 7.2-3.3 17.2 4 22.3s17.2 3.3 22.3-4c5.4-7.7 12.2-14.4 20.4-19.5l3.1-2c35.3-22.4 60.9-57.2 71.9-97.5l4.5-16.4zM32 224l64 0 0 224-64 0 0-224zM0 224L0 448c0 17.7 14.3 32 32 32l64 0c17.7 0 32-14.3 32-32l0-224c0-17.7-14.3-32-32-32l-64 0c-17.7 0-32 14.3-32 32z"]},For={prefix:"fal",iconName:"comment-xmark",icon:[512,512,["comment-times"],"f4b5","M32 240c0-91.4 93.8-176 224-176s224 84.6 224 176s-93.8 176-224 176c-32 0-63-5.9-93.5-16.2c-9.9-3.3-20.7-1.8-29.2 4.2c-15.4 10.7-31.4 20-48.4 27.4L91.2 446l-6.4-14.7c-10.5 4.6-21.5 8.6-32.7 11.5c.5-.9 1.1-1.8 1.6-2.6C68 416.5 76.3 391 79.4 363.9c1-9.2-2.1-17.7-7.4-23.9c-24.5-29.4-40-62.9-40-100zM256 32C114.5 32 0 125.2 0 240c0 46.6 19.6 87.1 47.4 120.5c0 .1 .1 .1 .1 .2c-2.6 22.6-9.5 43.5-21.3 63.1c-6.5 10.8-13.5 20.8-21.7 29c-4.5 4.6-5.9 11.5-3.4 17.4s8.3 9.8 14.8 9.8c28.7 0 57.6-8.9 81.6-19.3c19.2-8.4 37.1-18.7 53.9-30.5c.2-.1 .5-.2 .7-.1C185.5 441.3 220 448 256 448c141.5 0 256-93.2 256-208S397.5 32 256 32zM180.7 164.7c-6.2 6.2-6.2 16.4 0 22.6L233.4 240l-52.7 52.7c-6.2 6.2-6.2 16.4 0 22.6s16.4 6.2 22.6 0L256 262.6l52.7 52.7c6.2 6.2 16.4 6.2 22.6 0s6.2-16.4 0-22.6L278.6 240l52.7-52.7c6.2-6.2 6.2-16.4 0-22.6s-16.4-6.2-22.6 0L256 217.4l-52.7-52.7c-6.2-6.2-16.4-6.2-22.6 0z"]},Bor={prefix:"fal",iconName:"eyes",icon:[640,512,[],"e367","M304 256c0 55.6-16.9 104.7-42.8 139.2C235.3 429.7 202.1 448 168 448s-67.3-18.3-93.2-52.8c-7-9.3-13.3-19.6-18.8-30.8c7.6 2.4 15.6 3.7 24 3.7c44.2 0 80-35.8 80-80s-35.8-80-80-80c-17.1 0-32.9 5.4-45.9 14.5c5.3-41.7 20.2-78.3 40.7-105.7C100.7 82.3 133.9 64 168 64s67.3 18.3 93.2 52.8C287.1 151.3 304 200.4 304 256zm16-95.5C293.1 84.5 235.2 32 168 32C75.2 32 0 132.3 0 256S75.2 480 168 480c67.2 0 125.1-52.5 152-128.5C346.9 427.5 404.8 480 472 480c92.8 0 168-100.3 168-224s-75.2-224-168-224c-67.2 0-125.1 52.5-152 128.5zm58.8 234.7c-7-9.3-13.3-19.6-18.8-30.8c7.6 2.4 15.6 3.7 24 3.7c44.2 0 80-35.8 80-80s-35.8-80-80-80c-17.1 0-32.9 5.4-45.9 14.5c5.3-41.7 20.2-78.3 40.7-105.7C404.7 82.3 437.9 64 472 64s67.3 18.3 93.2 52.8C591.1 151.3 608 200.4 608 256s-16.9 104.7-42.8 139.2S506.1 448 472 448s-67.3-18.3-93.2-52.8zM32 288a48 48 0 1 1 96 0 48 48 0 1 1 -96 0zm352-48a48 48 0 1 1 0 96 48 48 0 1 1 0-96z"]};/*!
 * Font Awesome Pro 6.7.2 by @fontawesome - https://fontawesome.com
 * License - https://fontawesome.com/license (Commercial License)
 * Copyright 2024 Fonticons, Inc.
 */const Uor={prefix:"fas",iconName:"paper-plane-top",icon:[512,512,["paper-plane-alt","send"],"e20a","M57.6 35.8C23.1 20.6-11.3 57.4 6.1 90.9l63 121.2c4.4 8.4 12.6 14.1 22 15.3c0 0 0 0 0 0L266 249.3c3.4 .4 6 3.3 6 6.7s-2.6 6.3-6 6.7L91.1 284.6s0 0 0 0c-9.4 1.2-17.6 6.9-22 15.3L6.1 421.1c-17.4 33.5 17 70.2 51.6 55.1L492.9 285.3c25.5-11.2 25.5-47.4 0-58.6L57.6 35.8z"]},zor={prefix:"fas",iconName:"square",icon:[448,512,[9632,9723,9724,61590],"f0c8","M0 96C0 60.7 28.7 32 64 32H384c35.3 0 64 28.7 64 64V416c0 35.3-28.7 64-64 64H64c-35.3 0-64-28.7-64-64V96z"]},Hor={prefix:"fas",iconName:"arrow-up",icon:[384,512,[8593],"f062","M214.6 41.4c-12.5-12.5-32.8-12.5-45.3 0l-160 160c-12.5 12.5-12.5 32.8 0 45.3s32.8 12.5 45.3 0L160 141.2 160 448c0 17.7 14.3 32 32 32s32-14.3 32-32l0-306.7L329.4 246.6c12.5 12.5 32.8 12.5 45.3 0s12.5-32.8 0-45.3l-160-160z"]},Kp=xKn({placement:"bottom-end",pauseOnPageIdle:!0}),Gor=()=>{const n=Q.useCallback((r,o)=>o?H.jsx(xa,{icon:o}):r==="loading"?H.jsx(kK,{size:"sm",color:"blue.solid"}):H.jsx(ber,{}),[]);return H.jsx(lU,{children:H.jsx(ler,{toaster:Kp,insetInline:{mdDown:"4"},"data-testid":"toaster",children:r=>H.jsxs(der,{width:{md:"sm"},children:[n(r.type||"",r.meta?.icon),H.jsxs(Cye,{gap:"1",flex:"1",maxWidth:"100%",children:[r.title&&H.jsx(her,{asChild:!0,children:H.jsx(Kx,{color:"inherit",children:r.title})}),r.description&&H.jsx(per,{children:r.description})]}),r.action&&H.jsx(ger,{children:r.action.label}),r.closable&&H.jsx(fer,{})]})})})},onn={PROJECT_SERVICE_URL:""},ann={CREATE_PROJECT:"/api/projects",GET_PROJECT:n=>`/api/projects/${n}`,JOB_EVENTS:n=>`/api/job-events/${n}`},Vb={COPILOT:"/api/copilot",COPILOT_HISTORY:"/api/copilot/history",BOT_BUILDER:"/api",BOT_DATA:"/api/data",ASSISTANT_INFO:"/api/assistant",JOB_EVENTS:"/api/job-events",BOT_FILES:"/api/files",DOWNLOAD_PROJECT:"/api/download",COMMITS:"/api/commits",TRAIN:"/api/train",COPILOT_MODE:"/api/copilot/mode"};var y$e={exports:{}},E$e={},S$e={exports:{}},_$e={};/**
 * @license React
 * use-sync-external-store-shim.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Ojt;function qor(){if(Ojt)return _$e;Ojt=1;var n=voe();function r(x,O){return x===O&&(x!==0||1/x===1/O)||x!==x&&O!==O}var o=typeof Object.is=="function"?Object.is:r,s=n.useState,c=n.useEffect,d=n.useLayoutEffect,h=n.useDebugValue;function b(x,O){var P=O(),D=s({inst:{value:P,getSnapshot:O}}),$=D[0].inst,U=D[1];return d(function(){$.value=P,$.getSnapshot=O,y($)&&U({inst:$})},[x,P,O]),c(function(){return y($)&&U({inst:$}),x(function(){y($)&&U({inst:$})})},[x]),h(P),P}function y(x){var O=x.getSnapshot;x=x.value;try{var P=O();return!o(x,P)}catch{return!0}}function E(x,O){return O()}var k=typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"?E:b;return _$e.useSyncExternalStore=n.useSyncExternalStore!==void 0?n.useSyncExternalStore:k,_$e}var Njt;function Vor(){return Njt||(Njt=1,S$e.exports=qor()),S$e.exports}/**
 * @license React
 * use-sync-external-store-shim/with-selector.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Rjt;function Wor(){if(Rjt)return E$e;Rjt=1;var n=voe(),r=Vor();function o(E,k){return E===k&&(E!==0||1/E===1/k)||E!==E&&k!==k}var s=typeof Object.is=="function"?Object.is:o,c=r.useSyncExternalStore,d=n.useRef,h=n.useEffect,b=n.useMemo,y=n.useDebugValue;return E$e.useSyncExternalStoreWithSelector=function(E,k,x,O,P){var D=d(null);if(D.current===null){var $={hasValue:!1,value:null};D.current=$}else $=D.current;D=b(function(){function B(oe){if(!G){if(G=!0,K=oe,oe=O(oe),P!==void 0&&$.hasValue){var he=$.value;if(P(he,oe))return X=he}return X=oe}if(he=X,s(K,oe))return he;var ye=O(oe);return P!==void 0&&P(he,ye)?(K=oe,he):(K=oe,X=ye)}var G=!1,K,X,ae=x===void 0?null:x;return[function(){return B(k())},ae===null?void 0:function(){return B(ae())}]},[k,x,O,P]);var U=c(E,D[0],D[1]);return h(function(){$.hasValue=!0,$.value=U},[U]),y(U),U},E$e}var Pjt;function Kor(){return Pjt||(Pjt=1,y$e.exports=Wor()),y$e.exports}var snn=Kor();const Yor=Y4(snn),TW=new WeakMap,hve=new WeakMap,rwe={current:[]};let k$e=!1;const e0e=new Set,Ljt=new Map;function cnn(n){const r=Array.from(n).sort((o,s)=>o instanceof xW&&o.options.deps.includes(s)?1:s instanceof xW&&s.options.deps.includes(o)?-1:0);for(const o of r){if(rwe.current.includes(o))continue;rwe.current.push(o),o.recompute();const s=hve.get(o);if(s)for(const c of s){const d=TW.get(c);d&&cnn(d)}}}function Xor(n){n.listeners.forEach(r=>r({prevVal:n.prevState,currentVal:n.state}))}function Jor(n){n.listeners.forEach(r=>r({prevVal:n.prevState,currentVal:n.state}))}function Zor(n){if(e0e.add(n),!k$e)try{for(k$e=!0;e0e.size>0;){const r=Array.from(e0e);e0e.clear();for(const o of r){const s=Ljt.get(o)??o.prevState;o.prevState=s,Xor(o)}for(const o of r){const s=TW.get(o);s&&(rwe.current.push(o),cnn(s))}for(const o of r){const s=TW.get(o);if(s)for(const c of s)Jor(c)}}}finally{k$e=!1,rwe.current=[],Ljt.clear()}}function Qor(n){return typeof n=="function"}class IM{constructor(r,o){this.listeners=new Set,this.subscribe=s=>{var c,d;this.listeners.add(s);const h=(d=(c=this.options)==null?void 0:c.onSubscribe)==null?void 0:d.call(c,s,this);return()=>{this.listeners.delete(s),h?.()}},this.prevState=r,this.state=r,this.options=o}setState(r){var o,s,c;this.prevState=this.state,(o=this.options)!=null&&o.updateFn?this.state=this.options.updateFn(this.prevState)(r):Qor(r)?this.state=r(this.prevState):this.state=r,(c=(s=this.options)==null?void 0:s.onUpdate)==null||c.call(s),Zor(this)}}class xW{constructor(r){this.listeners=new Set,this._subscriptions=[],this.lastSeenDepValues=[],this.getDepVals=()=>{const o=[],s=[];for(const c of this.options.deps)o.push(c.prevState),s.push(c.state);return this.lastSeenDepValues=s,{prevDepVals:o,currDepVals:s,prevVal:this.prevState??void 0}},this.recompute=()=>{var o,s;this.prevState=this.state;const{prevDepVals:c,currDepVals:d,prevVal:h}=this.getDepVals();this.state=this.options.fn({prevDepVals:c,currDepVals:d,prevVal:h}),(s=(o=this.options).onUpdate)==null||s.call(o)},this.checkIfRecalculationNeededDeeply=()=>{for(const d of this.options.deps)d instanceof xW&&d.checkIfRecalculationNeededDeeply();let o=!1;const s=this.lastSeenDepValues,{currDepVals:c}=this.getDepVals();for(let d=0;d<c.length;d++)if(c[d]!==s[d]){o=!0;break}o&&this.recompute()},this.mount=()=>(this.registerOnGraph(),this.checkIfRecalculationNeededDeeply(),()=>{this.unregisterFromGraph();for(const o of this._subscriptions)o()}),this.subscribe=o=>{var s,c;this.listeners.add(o);const d=(c=(s=this.options).onSubscribe)==null?void 0:c.call(s,o,this);return()=>{this.listeners.delete(o),d?.()}},this.options=r,this.state=r.fn({prevDepVals:void 0,prevVal:void 0,currDepVals:this.getDepVals().currDepVals})}registerOnGraph(r=this.options.deps){for(const o of r)if(o instanceof xW)o.registerOnGraph(),this.registerOnGraph(o.options.deps);else if(o instanceof IM){let s=TW.get(o);s||(s=new Set,TW.set(o,s)),s.add(this);let c=hve.get(this);c||(c=new Set,hve.set(this,c)),c.add(o)}}unregisterFromGraph(r=this.options.deps){for(const o of r)if(o instanceof xW)this.unregisterFromGraph(o.options.deps);else if(o instanceof IM){const s=TW.get(o);s&&s.delete(this);const c=hve.get(this);c&&c.delete(o)}}}function Io(n,r=o=>o){return snn.useSyncExternalStoreWithSelector(n.subscribe,()=>n.state,()=>n.state,r,ear)}function ear(n,r){if(Object.is(n,r))return!0;if(typeof n!="object"||n===null||typeof r!="object"||r===null)return!1;if(n instanceof Map&&r instanceof Map){if(n.size!==r.size)return!1;for(const[s,c]of n)if(!r.has(s)||!Object.is(c,r.get(s)))return!1;return!0}if(n instanceof Set&&r instanceof Set){if(n.size!==r.size)return!1;for(const s of n)if(!r.has(s))return!1;return!0}if(n instanceof Date&&r instanceof Date)return n.getTime()===r.getTime();const o=Object.keys(n);if(o.length!==Object.keys(r).length)return!1;for(let s=0;s<o.length;s++)if(!Object.prototype.hasOwnProperty.call(r,o[s])||!Object.is(n[o[s]],r[o[s]]))return!1;return!0}const unn=Object.freeze({status:"aborted"});function rn(n,r,o){function s(b,y){var E;Object.defineProperty(b,"_zod",{value:b._zod??{},enumerable:!1}),(E=b._zod).traits??(E.traits=new Set),b._zod.traits.add(n),r(b,y);for(const k in h.prototype)k in b||Object.defineProperty(b,k,{value:h.prototype[k].bind(b)});b._zod.constr=h,b._zod.def=y}const c=o?.Parent??Object;class d extends c{}Object.defineProperty(d,"name",{value:n});function h(b){var y;const E=o?.Parent?new d:this;s(E,b),(y=E._zod).deferred??(y.deferred=[]);for(const k of E._zod.deferred)k();return E}return Object.defineProperty(h,"init",{value:s}),Object.defineProperty(h,Symbol.hasInstance,{value:b=>o?.Parent&&b instanceof o.Parent?!0:b?._zod?.traits?.has(n)}),Object.defineProperty(h,"name",{value:n}),h}const lnn=Symbol("zod_brand");class XW extends Error{constructor(){super("Encountered Promise during synchronous parse. Use .parseAsync() instead.")}}const iwe={};function cv(n){return n&&Object.assign(iwe,n),iwe}function tar(n){return n}function nar(n){return n}function rar(n){}function iar(n){throw new Error}function oar(n){}function lXe(n){const r=Object.values(n).filter(s=>typeof s=="number");return Object.entries(n).filter(([s,c])=>r.indexOf(+s)===-1).map(([s,c])=>c)}function ti(n,r="|"){return n.map(o=>es(o)).join(r)}function owe(n,r){return typeof r=="bigint"?r.toString():r}function zye(n){return{get value(){{const r=n();return Object.defineProperty(this,"value",{value:r}),r}}}}function dU(n){return n==null}function Hye(n){const r=n.startsWith("^")?1:0,o=n.endsWith("$")?n.length-1:n.length;return n.slice(r,o)}function dnn(n,r){const o=(n.toString().split(".")[1]||"").length,s=r.toString();let c=(s.split(".")[1]||"").length;if(c===0&&/\d?e-\d?/.test(s)){const y=s.match(/\d?e-(\d?)/);y?.[1]&&(c=Number.parseInt(y[1]))}const d=o>c?o:c,h=Number.parseInt(n.toFixed(d).replace(".","")),b=Number.parseInt(r.toFixed(d).replace(".",""));return h%b/10**d}const Djt=Symbol("evaluating");function du(n,r,o){let s;Object.defineProperty(n,r,{get(){if(s!==Djt)return s===void 0&&(s=Djt,s=o()),s},set(c){Object.defineProperty(n,r,{value:c})},configurable:!0})}function HI(n,r,o){Object.defineProperty(n,r,{value:o,writable:!0,enumerable:!0,configurable:!0})}function zM(...n){const r={};for(const o of n){const s=Object.getOwnPropertyDescriptors(o);Object.assign(r,s)}return Object.defineProperties({},r)}function aar(n){return zM(n._zod.def)}function sar(n,r){return r?r.reduce((o,s)=>o?.[s],n):n}function car(n){const r=Object.keys(n),o=r.map(s=>n[s]);return Promise.all(o).then(s=>{const c={};for(let d=0;d<r.length;d++)c[r[d]]=s[d];return c})}function uar(n=10){const r="abcdefghijklmnopqrstuvwxyz";let o="";for(let s=0;s<n;s++)o+=r[Math.floor(Math.random()*r.length)];return o}function zVe(n){return JSON.stringify(n)}const dXe="captureStackTrace"in Error?Error.captureStackTrace:(...n)=>{};function zie(n){return typeof n=="object"&&n!==null&&!Array.isArray(n)}const fnn=zye(()=>{if(typeof navigator<"u"&&navigator?.userAgent?.includes("Cloudflare"))return!1;try{const n=Function;return new n(""),!0}catch{return!1}});function Hie(n){if(zie(n)===!1)return!1;const r=n.constructor;if(r===void 0)return!0;const o=r.prototype;return!(zie(o)===!1||Object.prototype.hasOwnProperty.call(o,"isPrototypeOf")===!1)}function lar(n){let r=0;for(const o in n)Object.prototype.hasOwnProperty.call(n,o)&&r++;return r}const dar=n=>{const r=typeof n;switch(r){case"undefined":return"undefined";case"string":return"string";case"number":return Number.isNaN(n)?"nan":"number";case"boolean":return"boolean";case"function":return"function";case"bigint":return"bigint";case"symbol":return"symbol";case"object":return Array.isArray(n)?"array":n===null?"null":n.then&&typeof n.then=="function"&&n.catch&&typeof n.catch=="function"?"promise":typeof Map<"u"&&n instanceof Map?"map":typeof Set<"u"&&n instanceof Set?"set":typeof Date<"u"&&n instanceof Date?"date":typeof File<"u"&&n instanceof File?"file":"object";default:throw new Error(`Unknown data type: ${r}`)}},awe=new Set(["string","number","symbol"]),hnn=new Set(["string","number","bigint","boolean","symbol","undefined"]);function OM(n){return n.replace(/[.*+?^${}()|[\]\\]/g,"\\$&")}function eC(n,r,o){const s=new n._zod.constr(r??n._zod.def);return(!r||o?.parent)&&(s._zod.parent=n),s}function gr(n){const r=n;if(!r)return{};if(typeof r=="string")return{error:()=>r};if(r?.message!==void 0){if(r?.error!==void 0)throw new Error("Cannot specify both `message` and `error` params");r.error=r.message}return delete r.message,typeof r.error=="string"?{...r,error:()=>r.error}:r}function far(n){let r;return new Proxy({},{get(o,s,c){return r??(r=n()),Reflect.get(r,s,c)},set(o,s,c,d){return r??(r=n()),Reflect.set(r,s,c,d)},has(o,s){return r??(r=n()),Reflect.has(r,s)},deleteProperty(o,s){return r??(r=n()),Reflect.deleteProperty(r,s)},ownKeys(o){return r??(r=n()),Reflect.ownKeys(r)},getOwnPropertyDescriptor(o,s){return r??(r=n()),Reflect.getOwnPropertyDescriptor(r,s)},defineProperty(o,s,c){return r??(r=n()),Reflect.defineProperty(r,s,c)}})}function es(n){return typeof n=="bigint"?n.toString()+"n":typeof n=="string"?`"${n}"`:`${n}`}function pnn(n){return Object.keys(n).filter(r=>n[r]._zod.optin==="optional"&&n[r]._zod.optout==="optional")}const gnn={safeint:[Number.MIN_SAFE_INTEGER,Number.MAX_SAFE_INTEGER],int32:[-2147483648,2147483647],uint32:[0,4294967295],float32:[-34028234663852886e22,34028234663852886e22],float64:[-Number.MAX_VALUE,Number.MAX_VALUE]},mnn={int64:[BigInt("-9223372036854775808"),BigInt("9223372036854775807")],uint64:[BigInt(0),BigInt("18446744073709551615")]};function bnn(n,r){const o=n._zod.def,s=zM(n._zod.def,{get shape(){const c={};for(const d in r){if(!(d in o.shape))throw new Error(`Unrecognized key: "${d}"`);r[d]&&(c[d]=o.shape[d])}return HI(this,"shape",c),c},checks:[]});return eC(n,s)}function vnn(n,r){const o=n._zod.def,s=zM(n._zod.def,{get shape(){const c={...n._zod.def.shape};for(const d in r){if(!(d in o.shape))throw new Error(`Unrecognized key: "${d}"`);r[d]&&delete c[d]}return HI(this,"shape",c),c},checks:[]});return eC(n,s)}function wnn(n,r){if(!Hie(r))throw new Error("Invalid input to extend: expected a plain object");const o=zM(n._zod.def,{get shape(){const s={...n._zod.def.shape,...r};return HI(this,"shape",s),s},checks:[]});return eC(n,o)}function ynn(n,r){const o=zM(n._zod.def,{get shape(){const s={...n._zod.def.shape,...r._zod.def.shape};return HI(this,"shape",s),s},get catchall(){return r._zod.def.catchall},checks:[]});return eC(n,o)}function Enn(n,r,o){const s=zM(r._zod.def,{get shape(){const c=r._zod.def.shape,d={...c};if(o)for(const h in o){if(!(h in c))throw new Error(`Unrecognized key: "${h}"`);o[h]&&(d[h]=n?new n({type:"optional",innerType:c[h]}):c[h])}else for(const h in c)d[h]=n?new n({type:"optional",innerType:c[h]}):c[h];return HI(this,"shape",d),d},checks:[]});return eC(r,s)}function Snn(n,r,o){const s=zM(r._zod.def,{get shape(){const c=r._zod.def.shape,d={...c};if(o)for(const h in o){if(!(h in d))throw new Error(`Unrecognized key: "${h}"`);o[h]&&(d[h]=new n({type:"nonoptional",innerType:c[h]}))}else for(const h in c)d[h]=new n({type:"nonoptional",innerType:c[h]});return HI(this,"shape",d),d},checks:[]});return eC(r,s)}function CW(n,r=0){for(let o=r;o<n.issues.length;o++)if(n.issues[o]?.continue!==!0)return!0;return!1}function $x(n,r){return r.map(o=>{var s;return(s=o).path??(s.path=[]),o.path.unshift(n),o})}function jre(n){return typeof n=="string"?n:n?.message}function qx(n,r,o){const s={...n,path:n.path??[]};if(!n.message){const c=jre(n.inst?._zod.def?.error?.(n))??jre(r?.error?.(n))??jre(o.customError?.(n))??jre(o.localeError?.(n))??"Invalid input";s.message=c}return delete s.inst,delete s.continue,r?.reportInput||delete s.input,s}function Gye(n){return n instanceof Set?"set":n instanceof Map?"map":n instanceof File?"file":"unknown"}function qye(n){return Array.isArray(n)?"array":typeof n=="string"?"string":"unknown"}function JW(...n){const[r,o,s]=n;return typeof r=="string"?{message:r,code:"custom",input:o,inst:s}:{...r}}function har(n){return Object.entries(n).filter(([r,o])=>Number.isNaN(Number.parseInt(r,10))).map(r=>r[1])}class par{constructor(...r){}}const gar=Object.freeze(Object.defineProperty({__proto__:null,BIGINT_FORMAT_RANGES:mnn,Class:par,NUMBER_FORMAT_RANGES:gnn,aborted:CW,allowsEval:fnn,assert:oar,assertEqual:tar,assertIs:rar,assertNever:iar,assertNotEqual:nar,assignProp:HI,cached:zye,captureStackTrace:dXe,cleanEnum:har,cleanRegex:Hye,clone:eC,cloneDef:aar,createTransparentProxy:far,defineLazy:du,esc:zVe,escapeRegex:OM,extend:wnn,finalizeIssue:qx,floatSafeRemainder:dnn,getElementAtPath:sar,getEnumValues:lXe,getLengthableOrigin:qye,getParsedType:dar,getSizableOrigin:Gye,isObject:zie,isPlainObject:Hie,issue:JW,joinValues:ti,jsonStringifyReplacer:owe,merge:ynn,mergeDefs:zM,normalizeParams:gr,nullish:dU,numKeys:lar,omit:vnn,optionalKeys:pnn,partial:Enn,pick:bnn,prefixIssues:$x,primitiveTypes:hnn,promiseAllObject:car,propertyKeyTypes:awe,randomString:uar,required:Snn,stringifyPrimitive:es,unwrapMessage:jre},Symbol.toStringTag,{value:"Module"})),_nn=(n,r)=>{n.name="$ZodError",Object.defineProperty(n,"_zod",{value:n._zod,enumerable:!1}),Object.defineProperty(n,"issues",{value:r,enumerable:!1}),n.message=JSON.stringify(r,owe,2),Object.defineProperty(n,"toString",{value:()=>n.message,enumerable:!1})},fXe=rn("$ZodError",_nn),joe=rn("$ZodError",_nn,{Parent:Error});function hXe(n,r=o=>o.message){const o={},s=[];for(const c of n.issues)c.path.length>0?(o[c.path[0]]=o[c.path[0]]||[],o[c.path[0]].push(r(c))):s.push(r(c));return{formErrors:s,fieldErrors:o}}function pXe(n,r){const o=r||function(d){return d.message},s={_errors:[]},c=d=>{for(const h of d.issues)if(h.code==="invalid_union"&&h.errors.length)h.errors.map(b=>c({issues:b}));else if(h.code==="invalid_key")c({issues:h.issues});else if(h.code==="invalid_element")c({issues:h.issues});else if(h.path.length===0)s._errors.push(o(h));else{let b=s,y=0;for(;y<h.path.length;){const E=h.path[y];y===h.path.length-1?(b[E]=b[E]||{_errors:[]},b[E]._errors.push(o(h))):b[E]=b[E]||{_errors:[]},b=b[E],y++}}};return c(n),s}function knn(n,r){const o=r||function(d){return d.message},s={errors:[]},c=(d,h=[])=>{var b,y;for(const E of d.issues)if(E.code==="invalid_union"&&E.errors.length)E.errors.map(k=>c({issues:k},E.path));else if(E.code==="invalid_key")c({issues:E.issues},E.path);else if(E.code==="invalid_element")c({issues:E.issues},E.path);else{const k=[...h,...E.path];if(k.length===0){s.errors.push(o(E));continue}let x=s,O=0;for(;O<k.length;){const P=k[O],D=O===k.length-1;typeof P=="string"?(x.properties??(x.properties={}),(b=x.properties)[P]??(b[P]={errors:[]}),x=x.properties[P]):(x.items??(x.items=[]),(y=x.items)[P]??(y[P]={errors:[]}),x=x.items[P]),D&&x.errors.push(o(E)),O++}}};return c(n),s}function Tnn(n){const r=[],o=n.map(s=>typeof s=="object"?s.key:s);for(const s of o)typeof s=="number"?r.push(`[${s}]`):typeof s=="symbol"?r.push(`[${JSON.stringify(String(s))}]`):/[^\w$]/.test(s)?r.push(`[${JSON.stringify(s)}]`):(r.length&&r.push("."),r.push(s));return r.join("")}function xnn(n){const r=[],o=[...n.issues].sort((s,c)=>(s.path??[]).length-(c.path??[]).length);for(const s of o)r.push(`✖ ${s.message}`),s.path?.length&&r.push(`  → at ${Tnn(s.path)}`);return r.join(`
`)}const gXe=n=>(r,o,s,c)=>{const d=s?Object.assign(s,{async:!1}):{async:!1},h=r._zod.run({value:o,issues:[]},d);if(h instanceof Promise)throw new XW;if(h.issues.length){const b=new(c?.Err??n)(h.issues.map(y=>qx(y,d,cv())));throw dXe(b,c?.callee),b}return h.value},HVe=gXe(joe),mXe=n=>async(r,o,s,c)=>{const d=s?Object.assign(s,{async:!0}):{async:!0};let h=r._zod.run({value:o,issues:[]},d);if(h instanceof Promise&&(h=await h),h.issues.length){const b=new(c?.Err??n)(h.issues.map(y=>qx(y,d,cv())));throw dXe(b,c?.callee),b}return h.value},GVe=mXe(joe),bXe=n=>(r,o,s)=>{const c=s?{...s,async:!1}:{async:!1},d=r._zod.run({value:o,issues:[]},c);if(d instanceof Promise)throw new XW;return d.issues.length?{success:!1,error:new(n??fXe)(d.issues.map(h=>qx(h,c,cv())))}:{success:!0,data:d.value}},Cnn=bXe(joe),vXe=n=>async(r,o,s)=>{const c=s?Object.assign(s,{async:!0}):{async:!0};let d=r._zod.run({value:o,issues:[]},c);return d instanceof Promise&&(d=await d),d.issues.length?{success:!1,error:new n(d.issues.map(h=>qx(h,c,cv())))}:{success:!0,data:d.value}},Ann=vXe(joe),Inn=/^[cC][^\s-]{8,}$/,Onn=/^[0-9a-z]+$/,Nnn=/^[0-9A-HJKMNP-TV-Za-hjkmnp-tv-z]{26}$/,Rnn=/^[0-9a-vA-V]{20}$/,Pnn=/^[A-Za-z0-9]{27}$/,Lnn=/^[a-zA-Z0-9_-]{21}$/,Dnn=/^P(?:(\d+W)|(?!.*W)(?=\d|T\d)(\d+Y)?(\d+M)?(\d+D)?(T(?=\d)(\d+H)?(\d+M)?(\d+([.,]\d+)?S)?)?)$/,mar=/^[-+]?P(?!$)(?:(?:[-+]?\d+Y)|(?:[-+]?\d+[.,]\d+Y$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:(?:[-+]?\d+W)|(?:[-+]?\d+[.,]\d+W$))?(?:(?:[-+]?\d+D)|(?:[-+]?\d+[.,]\d+D$))?(?:T(?=[\d+-])(?:(?:[-+]?\d+H)|(?:[-+]?\d+[.,]\d+H$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:[-+]?\d+(?:[.,]\d+)?S)?)??$/,Mnn=/^([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})$/,ZW=n=>n?new RegExp(`^([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-${n}[0-9a-fA-F]{3}-[89abAB][0-9a-fA-F]{3}-[0-9a-fA-F]{12})$`):/^([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[1-8][0-9a-fA-F]{3}-[89abAB][0-9a-fA-F]{3}-[0-9a-fA-F]{12}|00000000-0000-0000-0000-000000000000)$/,bar=ZW(4),war=ZW(6),yar=ZW(7),jnn=/^(?!\.)(?!.*\.\.)([A-Za-z0-9_'+\-\.]*)[A-Za-z0-9_+-]@([A-Za-z0-9][A-Za-z0-9\-]*\.)+[A-Za-z]{2,}$/,Ear=/^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$/,Sar=/^(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/,_ar=/^[^\s@"]{1,64}@[^\s@]{1,255}$/u,kar=/^[^\s@"]{1,64}@[^\s@]{1,255}$/u,Tar=/^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$/,xar="^(\\p{Extended_Pictographic}|\\p{Emoji_Component})+$";function $nn(){return new RegExp(xar,"u")}const Fnn=/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])$/,Bnn=/^(([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}|::|([0-9a-fA-F]{1,4})?::([0-9a-fA-F]{1,4}:?){0,6})$/,Unn=/^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\/([0-9]|[1-2][0-9]|3[0-2])$/,znn=/^(([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}|::|([0-9a-fA-F]{1,4})?::([0-9a-fA-F]{1,4}:?){0,6})\/(12[0-8]|1[01][0-9]|[1-9]?[0-9])$/,Hnn=/^$|^(?:[0-9a-zA-Z+/]{4})*(?:(?:[0-9a-zA-Z+/]{2}==)|(?:[0-9a-zA-Z+/]{3}=))?$/,wXe=/^[A-Za-z0-9_-]*$/,yXe=/^(?=.{1,253}\.?$)[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[-0-9a-zA-Z]{0,61}[0-9a-zA-Z])?)*\.?$/,Car=/^([a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]{2,}$/,Gnn=/^\+(?:[0-9]){6,14}[0-9]$/,qnn="(?:(?:\\d\\d[2468][048]|\\d\\d[13579][26]|\\d\\d0[48]|[02468][048]00|[13579][26]00)-02-29|\\d{4}-(?:(?:0[13578]|1[02])-(?:0[1-9]|[12]\\d|3[01])|(?:0[469]|11)-(?:0[1-9]|[12]\\d|30)|(?:02)-(?:0[1-9]|1\\d|2[0-8])))",Vnn=new RegExp(`^${qnn}$`);function Wnn(n){const r="(?:[01]\\d|2[0-3]):[0-5]\\d";return typeof n.precision=="number"?n.precision===-1?`${r}`:n.precision===0?`${r}:[0-5]\\d`:`${r}:[0-5]\\d\\.\\d{${n.precision}}`:`${r}(?::[0-5]\\d(?:\\.\\d+)?)?`}function Knn(n){return new RegExp(`^${Wnn(n)}$`)}function Ynn(n){const r=Wnn({precision:n.precision}),o=["Z"];n.local&&o.push(""),n.offset&&o.push("([+-](?:[01]\\d|2[0-3]):[0-5]\\d)");const s=`${r}(?:${o.join("|")})`;return new RegExp(`^${qnn}T(?:${s})$`)}const Xnn=n=>{const r=n?`[\\s\\S]{${n?.minimum??0},${n?.maximum??""}}`:"[\\s\\S]*";return new RegExp(`^${r}$`)},Jnn=/^\d+n?$/,Znn=/^\d+$/,Qnn=/^-?\d+(?:\.\d+)?/i,ern=/true|false/i,trn=/null/i,nrn=/undefined/i,rrn=/^[^A-Z]*$/,irn=/^[^a-z]*$/,orn=Object.freeze(Object.defineProperty({__proto__:null,base64:Hnn,base64url:wXe,bigint:Jnn,boolean:ern,browserEmail:Tar,cidrv4:Unn,cidrv6:znn,cuid:Inn,cuid2:Onn,date:Vnn,datetime:Ynn,domain:Car,duration:Dnn,e164:Gnn,email:jnn,emoji:$nn,extendedDuration:mar,guid:Mnn,hostname:yXe,html5Email:Ear,idnEmail:kar,integer:Znn,ipv4:Fnn,ipv6:Bnn,ksuid:Pnn,lowercase:rrn,nanoid:Lnn,null:trn,number:Qnn,rfc5322Email:Sar,string:Xnn,time:Knn,ulid:Nnn,undefined:nrn,unicodeEmail:_ar,uppercase:irn,uuid:ZW,uuid4:bar,uuid6:war,uuid7:yar,xid:Rnn},Symbol.toStringTag,{value:"Module"})),Ph=rn("$ZodCheck",(n,r)=>{var o;n._zod??(n._zod={}),n._zod.def=r,(o=n._zod).onattach??(o.onattach=[])}),arn={number:"number",bigint:"bigint",object:"date"},EXe=rn("$ZodCheckLessThan",(n,r)=>{Ph.init(n,r);const o=arn[typeof r.value];n._zod.onattach.push(s=>{const c=s._zod.bag,d=(r.inclusive?c.maximum:c.exclusiveMaximum)??Number.POSITIVE_INFINITY;r.value<d&&(r.inclusive?c.maximum=r.value:c.exclusiveMaximum=r.value)}),n._zod.check=s=>{(r.inclusive?s.value<=r.value:s.value<r.value)||s.issues.push({origin:o,code:"too_big",maximum:r.value,input:s.value,inclusive:r.inclusive,inst:n,continue:!r.abort})}}),SXe=rn("$ZodCheckGreaterThan",(n,r)=>{Ph.init(n,r);const o=arn[typeof r.value];n._zod.onattach.push(s=>{const c=s._zod.bag,d=(r.inclusive?c.minimum:c.exclusiveMinimum)??Number.NEGATIVE_INFINITY;r.value>d&&(r.inclusive?c.minimum=r.value:c.exclusiveMinimum=r.value)}),n._zod.check=s=>{(r.inclusive?s.value>=r.value:s.value>r.value)||s.issues.push({origin:o,code:"too_small",minimum:r.value,input:s.value,inclusive:r.inclusive,inst:n,continue:!r.abort})}}),srn=rn("$ZodCheckMultipleOf",(n,r)=>{Ph.init(n,r),n._zod.onattach.push(o=>{var s;(s=o._zod.bag).multipleOf??(s.multipleOf=r.value)}),n._zod.check=o=>{if(typeof o.value!=typeof r.value)throw new Error("Cannot mix number and bigint in multiple_of check.");(typeof o.value=="bigint"?o.value%r.value===BigInt(0):dnn(o.value,r.value)===0)||o.issues.push({origin:typeof o.value,code:"not_multiple_of",divisor:r.value,input:o.value,inst:n,continue:!r.abort})}}),crn=rn("$ZodCheckNumberFormat",(n,r)=>{Ph.init(n,r),r.format=r.format||"float64";const o=r.format?.includes("int"),s=o?"int":"number",[c,d]=gnn[r.format];n._zod.onattach.push(h=>{const b=h._zod.bag;b.format=r.format,b.minimum=c,b.maximum=d,o&&(b.pattern=Znn)}),n._zod.check=h=>{const b=h.value;if(o){if(!Number.isInteger(b)){h.issues.push({expected:s,format:r.format,code:"invalid_type",continue:!1,input:b,inst:n});return}if(!Number.isSafeInteger(b)){b>0?h.issues.push({input:b,code:"too_big",maximum:Number.MAX_SAFE_INTEGER,note:"Integers must be within the safe integer range.",inst:n,origin:s,continue:!r.abort}):h.issues.push({input:b,code:"too_small",minimum:Number.MIN_SAFE_INTEGER,note:"Integers must be within the safe integer range.",inst:n,origin:s,continue:!r.abort});return}}b<c&&h.issues.push({origin:"number",input:b,code:"too_small",minimum:c,inclusive:!0,inst:n,continue:!r.abort}),b>d&&h.issues.push({origin:"number",input:b,code:"too_big",maximum:d,inst:n})}}),urn=rn("$ZodCheckBigIntFormat",(n,r)=>{Ph.init(n,r);const[o,s]=mnn[r.format];n._zod.onattach.push(c=>{const d=c._zod.bag;d.format=r.format,d.minimum=o,d.maximum=s}),n._zod.check=c=>{const d=c.value;d<o&&c.issues.push({origin:"bigint",input:d,code:"too_small",minimum:o,inclusive:!0,inst:n,continue:!r.abort}),d>s&&c.issues.push({origin:"bigint",input:d,code:"too_big",maximum:s,inst:n})}}),lrn=rn("$ZodCheckMaxSize",(n,r)=>{var o;Ph.init(n,r),(o=n._zod.def).when??(o.when=s=>{const c=s.value;return!dU(c)&&c.size!==void 0}),n._zod.onattach.push(s=>{const c=s._zod.bag.maximum??Number.POSITIVE_INFINITY;r.maximum<c&&(s._zod.bag.maximum=r.maximum)}),n._zod.check=s=>{const c=s.value;c.size<=r.maximum||s.issues.push({origin:Gye(c),code:"too_big",maximum:r.maximum,input:c,inst:n,continue:!r.abort})}}),drn=rn("$ZodCheckMinSize",(n,r)=>{var o;Ph.init(n,r),(o=n._zod.def).when??(o.when=s=>{const c=s.value;return!dU(c)&&c.size!==void 0}),n._zod.onattach.push(s=>{const c=s._zod.bag.minimum??Number.NEGATIVE_INFINITY;r.minimum>c&&(s._zod.bag.minimum=r.minimum)}),n._zod.check=s=>{const c=s.value;c.size>=r.minimum||s.issues.push({origin:Gye(c),code:"too_small",minimum:r.minimum,input:c,inst:n,continue:!r.abort})}}),frn=rn("$ZodCheckSizeEquals",(n,r)=>{var o;Ph.init(n,r),(o=n._zod.def).when??(o.when=s=>{const c=s.value;return!dU(c)&&c.size!==void 0}),n._zod.onattach.push(s=>{const c=s._zod.bag;c.minimum=r.size,c.maximum=r.size,c.size=r.size}),n._zod.check=s=>{const c=s.value,d=c.size;if(d===r.size)return;const h=d>r.size;s.issues.push({origin:Gye(c),...h?{code:"too_big",maximum:r.size}:{code:"too_small",minimum:r.size},inclusive:!0,exact:!0,input:s.value,inst:n,continue:!r.abort})}}),hrn=rn("$ZodCheckMaxLength",(n,r)=>{var o;Ph.init(n,r),(o=n._zod.def).when??(o.when=s=>{const c=s.value;return!dU(c)&&c.length!==void 0}),n._zod.onattach.push(s=>{const c=s._zod.bag.maximum??Number.POSITIVE_INFINITY;r.maximum<c&&(s._zod.bag.maximum=r.maximum)}),n._zod.check=s=>{const c=s.value;if(c.length<=r.maximum)return;const h=qye(c);s.issues.push({origin:h,code:"too_big",maximum:r.maximum,inclusive:!0,input:c,inst:n,continue:!r.abort})}}),prn=rn("$ZodCheckMinLength",(n,r)=>{var o;Ph.init(n,r),(o=n._zod.def).when??(o.when=s=>{const c=s.value;return!dU(c)&&c.length!==void 0}),n._zod.onattach.push(s=>{const c=s._zod.bag.minimum??Number.NEGATIVE_INFINITY;r.minimum>c&&(s._zod.bag.minimum=r.minimum)}),n._zod.check=s=>{const c=s.value;if(c.length>=r.minimum)return;const h=qye(c);s.issues.push({origin:h,code:"too_small",minimum:r.minimum,inclusive:!0,input:c,inst:n,continue:!r.abort})}}),grn=rn("$ZodCheckLengthEquals",(n,r)=>{var o;Ph.init(n,r),(o=n._zod.def).when??(o.when=s=>{const c=s.value;return!dU(c)&&c.length!==void 0}),n._zod.onattach.push(s=>{const c=s._zod.bag;c.minimum=r.length,c.maximum=r.length,c.length=r.length}),n._zod.check=s=>{const c=s.value,d=c.length;if(d===r.length)return;const h=qye(c),b=d>r.length;s.issues.push({origin:h,...b?{code:"too_big",maximum:r.length}:{code:"too_small",minimum:r.length},inclusive:!0,exact:!0,input:s.value,inst:n,continue:!r.abort})}}),$oe=rn("$ZodCheckStringFormat",(n,r)=>{var o,s;Ph.init(n,r),n._zod.onattach.push(c=>{const d=c._zod.bag;d.format=r.format,r.pattern&&(d.patterns??(d.patterns=new Set),d.patterns.add(r.pattern))}),r.pattern?(o=n._zod).check??(o.check=c=>{r.pattern.lastIndex=0,!r.pattern.test(c.value)&&c.issues.push({origin:"string",code:"invalid_format",format:r.format,input:c.value,...r.pattern?{pattern:r.pattern.toString()}:{},inst:n,continue:!r.abort})}):(s=n._zod).check??(s.check=()=>{})}),mrn=rn("$ZodCheckRegex",(n,r)=>{$oe.init(n,r),n._zod.check=o=>{r.pattern.lastIndex=0,!r.pattern.test(o.value)&&o.issues.push({origin:"string",code:"invalid_format",format:"regex",input:o.value,pattern:r.pattern.toString(),inst:n,continue:!r.abort})}}),brn=rn("$ZodCheckLowerCase",(n,r)=>{r.pattern??(r.pattern=rrn),$oe.init(n,r)}),vrn=rn("$ZodCheckUpperCase",(n,r)=>{r.pattern??(r.pattern=irn),$oe.init(n,r)}),wrn=rn("$ZodCheckIncludes",(n,r)=>{Ph.init(n,r);const o=OM(r.includes),s=new RegExp(typeof r.position=="number"?`^.{${r.position}}${o}`:o);r.pattern=s,n._zod.onattach.push(c=>{const d=c._zod.bag;d.patterns??(d.patterns=new Set),d.patterns.add(s)}),n._zod.check=c=>{c.value.includes(r.includes,r.position)||c.issues.push({origin:"string",code:"invalid_format",format:"includes",includes:r.includes,input:c.value,inst:n,continue:!r.abort})}}),yrn=rn("$ZodCheckStartsWith",(n,r)=>{Ph.init(n,r);const o=new RegExp(`^${OM(r.prefix)}.*`);r.pattern??(r.pattern=o),n._zod.onattach.push(s=>{const c=s._zod.bag;c.patterns??(c.patterns=new Set),c.patterns.add(o)}),n._zod.check=s=>{s.value.startsWith(r.prefix)||s.issues.push({origin:"string",code:"invalid_format",format:"starts_with",prefix:r.prefix,input:s.value,inst:n,continue:!r.abort})}}),Ern=rn("$ZodCheckEndsWith",(n,r)=>{Ph.init(n,r);const o=new RegExp(`.*${OM(r.suffix)}$`);r.pattern??(r.pattern=o),n._zod.onattach.push(s=>{const c=s._zod.bag;c.patterns??(c.patterns=new Set),c.patterns.add(o)}),n._zod.check=s=>{s.value.endsWith(r.suffix)||s.issues.push({origin:"string",code:"invalid_format",format:"ends_with",suffix:r.suffix,input:s.value,inst:n,continue:!r.abort})}});function Mjt(n,r,o){n.issues.length&&r.issues.push(...$x(o,n.issues))}const Srn=rn("$ZodCheckProperty",(n,r)=>{Ph.init(n,r),n._zod.check=o=>{const s=r.schema._zod.run({value:o.value[r.property],issues:[]},{});if(s instanceof Promise)return s.then(c=>Mjt(c,o,r.property));Mjt(s,o,r.property)}}),_rn=rn("$ZodCheckMimeType",(n,r)=>{Ph.init(n,r);const o=new Set(r.mime);n._zod.onattach.push(s=>{s._zod.bag.mime=r.mime}),n._zod.check=s=>{o.has(s.value.type)||s.issues.push({code:"invalid_value",values:r.mime,input:s.value.type,inst:n,continue:!r.abort})}}),krn=rn("$ZodCheckOverwrite",(n,r)=>{Ph.init(n,r),n._zod.check=o=>{o.value=r.tx(o.value)}});class Trn{constructor(r=[]){this.content=[],this.indent=0,this&&(this.args=r)}indented(r){this.indent+=1,r(this),this.indent-=1}write(r){if(typeof r=="function"){r(this,{execution:"sync"}),r(this,{execution:"async"});return}const s=r.split(`
`).filter(h=>h),c=Math.min(...s.map(h=>h.length-h.trimStart().length)),d=s.map(h=>h.slice(c)).map(h=>" ".repeat(this.indent*2)+h);for(const h of d)this.content.push(h)}compile(){const r=Function,o=this?.args,c=[...(this?.content??[""]).map(d=>`  ${d}`)];return new r(...o,c.join(`
`))}}const xrn={major:4,minor:0,patch:14},Qa=rn("$ZodType",(n,r)=>{var o;n??(n={}),n._zod.def=r,n._zod.bag=n._zod.bag||{},n._zod.version=xrn;const s=[...n._zod.def.checks??[]];n._zod.traits.has("$ZodCheck")&&s.unshift(n);for(const c of s)for(const d of c._zod.onattach)d(n);if(s.length===0)(o=n._zod).deferred??(o.deferred=[]),n._zod.deferred?.push(()=>{n._zod.run=n._zod.parse});else{const c=(d,h,b)=>{let y=CW(d),E;for(const k of h){if(k._zod.def.when){if(!k._zod.def.when(d))continue}else if(y)continue;const x=d.issues.length,O=k._zod.check(d);if(O instanceof Promise&&b?.async===!1)throw new XW;if(E||O instanceof Promise)E=(E??Promise.resolve()).then(async()=>{await O,d.issues.length!==x&&(y||(y=CW(d,x)))});else{if(d.issues.length===x)continue;y||(y=CW(d,x))}}return E?E.then(()=>d):d};n._zod.run=(d,h)=>{const b=n._zod.parse(d,h);if(b instanceof Promise){if(h.async===!1)throw new XW;return b.then(y=>c(y,s,h))}return c(b,s,h)}}n["~standard"]={validate:c=>{try{const d=Cnn(n,c);return d.success?{value:d.data}:{issues:d.error?.issues}}catch{return Ann(n,c).then(h=>h.success?{value:h.data}:{issues:h.error?.issues})}},vendor:"zod",version:1}}),Foe=rn("$ZodString",(n,r)=>{Qa.init(n,r),n._zod.pattern=[...n?._zod.bag?.patterns??[]].pop()??Xnn(n._zod.bag),n._zod.parse=(o,s)=>{if(r.coerce)try{o.value=String(o.value)}catch{}return typeof o.value=="string"||o.issues.push({expected:"string",code:"invalid_type",input:o.value,inst:n}),o}}),Ad=rn("$ZodStringFormat",(n,r)=>{$oe.init(n,r),Foe.init(n,r)}),Crn=rn("$ZodGUID",(n,r)=>{r.pattern??(r.pattern=Mnn),Ad.init(n,r)}),Arn=rn("$ZodUUID",(n,r)=>{if(r.version){const s={v1:1,v2:2,v3:3,v4:4,v5:5,v6:6,v7:7,v8:8}[r.version];if(s===void 0)throw new Error(`Invalid UUID version: "${r.version}"`);r.pattern??(r.pattern=ZW(s))}else r.pattern??(r.pattern=ZW());Ad.init(n,r)}),Irn=rn("$ZodEmail",(n,r)=>{r.pattern??(r.pattern=jnn),Ad.init(n,r)}),Orn=rn("$ZodURL",(n,r)=>{Ad.init(n,r),n._zod.check=o=>{try{const s=o.value.trim(),c=new URL(s);r.hostname&&(r.hostname.lastIndex=0,r.hostname.test(c.hostname)||o.issues.push({code:"invalid_format",format:"url",note:"Invalid hostname",pattern:yXe.source,input:o.value,inst:n,continue:!r.abort})),r.protocol&&(r.protocol.lastIndex=0,r.protocol.test(c.protocol.endsWith(":")?c.protocol.slice(0,-1):c.protocol)||o.issues.push({code:"invalid_format",format:"url",note:"Invalid protocol",pattern:r.protocol.source,input:o.value,inst:n,continue:!r.abort})),r.normalize?o.value=c.href:o.value=s;return}catch{o.issues.push({code:"invalid_format",format:"url",input:o.value,inst:n,continue:!r.abort})}}}),Nrn=rn("$ZodEmoji",(n,r)=>{r.pattern??(r.pattern=$nn()),Ad.init(n,r)}),Rrn=rn("$ZodNanoID",(n,r)=>{r.pattern??(r.pattern=Lnn),Ad.init(n,r)}),Prn=rn("$ZodCUID",(n,r)=>{r.pattern??(r.pattern=Inn),Ad.init(n,r)}),Lrn=rn("$ZodCUID2",(n,r)=>{r.pattern??(r.pattern=Onn),Ad.init(n,r)}),Drn=rn("$ZodULID",(n,r)=>{r.pattern??(r.pattern=Nnn),Ad.init(n,r)}),Mrn=rn("$ZodXID",(n,r)=>{r.pattern??(r.pattern=Rnn),Ad.init(n,r)}),jrn=rn("$ZodKSUID",(n,r)=>{r.pattern??(r.pattern=Pnn),Ad.init(n,r)}),$rn=rn("$ZodISODateTime",(n,r)=>{r.pattern??(r.pattern=Ynn(r)),Ad.init(n,r)}),Frn=rn("$ZodISODate",(n,r)=>{r.pattern??(r.pattern=Vnn),Ad.init(n,r)}),Brn=rn("$ZodISOTime",(n,r)=>{r.pattern??(r.pattern=Knn(r)),Ad.init(n,r)}),Urn=rn("$ZodISODuration",(n,r)=>{r.pattern??(r.pattern=Dnn),Ad.init(n,r)}),zrn=rn("$ZodIPv4",(n,r)=>{r.pattern??(r.pattern=Fnn),Ad.init(n,r),n._zod.onattach.push(o=>{const s=o._zod.bag;s.format="ipv4"})}),Hrn=rn("$ZodIPv6",(n,r)=>{r.pattern??(r.pattern=Bnn),Ad.init(n,r),n._zod.onattach.push(o=>{const s=o._zod.bag;s.format="ipv6"}),n._zod.check=o=>{try{new URL(`http://[${o.value}]`)}catch{o.issues.push({code:"invalid_format",format:"ipv6",input:o.value,inst:n,continue:!r.abort})}}}),Grn=rn("$ZodCIDRv4",(n,r)=>{r.pattern??(r.pattern=Unn),Ad.init(n,r)}),qrn=rn("$ZodCIDRv6",(n,r)=>{r.pattern??(r.pattern=znn),Ad.init(n,r),n._zod.check=o=>{const[s,c]=o.value.split("/");try{if(!c)throw new Error;const d=Number(c);if(`${d}`!==c)throw new Error;if(d<0||d>128)throw new Error;new URL(`http://[${s}]`)}catch{o.issues.push({code:"invalid_format",format:"cidrv6",input:o.value,inst:n,continue:!r.abort})}}});function _Xe(n){if(n==="")return!0;if(n.length%4!==0)return!1;try{return atob(n),!0}catch{return!1}}const Vrn=rn("$ZodBase64",(n,r)=>{r.pattern??(r.pattern=Hnn),Ad.init(n,r),n._zod.onattach.push(o=>{o._zod.bag.contentEncoding="base64"}),n._zod.check=o=>{_Xe(o.value)||o.issues.push({code:"invalid_format",format:"base64",input:o.value,inst:n,continue:!r.abort})}});function Wrn(n){if(!wXe.test(n))return!1;const r=n.replace(/[-_]/g,s=>s==="-"?"+":"/"),o=r.padEnd(Math.ceil(r.length/4)*4,"=");return _Xe(o)}const Krn=rn("$ZodBase64URL",(n,r)=>{r.pattern??(r.pattern=wXe),Ad.init(n,r),n._zod.onattach.push(o=>{o._zod.bag.contentEncoding="base64url"}),n._zod.check=o=>{Wrn(o.value)||o.issues.push({code:"invalid_format",format:"base64url",input:o.value,inst:n,continue:!r.abort})}}),Yrn=rn("$ZodE164",(n,r)=>{r.pattern??(r.pattern=Gnn),Ad.init(n,r)});function Xrn(n,r=null){try{const o=n.split(".");if(o.length!==3)return!1;const[s]=o;if(!s)return!1;const c=JSON.parse(atob(s));return!("typ"in c&&c?.typ!=="JWT"||!c.alg||r&&(!("alg"in c)||c.alg!==r))}catch{return!1}}const Jrn=rn("$ZodJWT",(n,r)=>{Ad.init(n,r),n._zod.check=o=>{Xrn(o.value,r.alg)||o.issues.push({code:"invalid_format",format:"jwt",input:o.value,inst:n,continue:!r.abort})}}),Zrn=rn("$ZodCustomStringFormat",(n,r)=>{Ad.init(n,r),n._zod.check=o=>{r.fn(o.value)||o.issues.push({code:"invalid_format",format:r.format,input:o.value,inst:n,continue:!r.abort})}}),kXe=rn("$ZodNumber",(n,r)=>{Qa.init(n,r),n._zod.pattern=n._zod.bag.pattern??Qnn,n._zod.parse=(o,s)=>{if(r.coerce)try{o.value=Number(o.value)}catch{}const c=o.value;if(typeof c=="number"&&!Number.isNaN(c)&&Number.isFinite(c))return o;const d=typeof c=="number"?Number.isNaN(c)?"NaN":Number.isFinite(c)?void 0:"Infinity":void 0;return o.issues.push({expected:"number",code:"invalid_type",input:c,inst:n,...d?{received:d}:{}}),o}}),Qrn=rn("$ZodNumber",(n,r)=>{crn.init(n,r),kXe.init(n,r)}),TXe=rn("$ZodBoolean",(n,r)=>{Qa.init(n,r),n._zod.pattern=ern,n._zod.parse=(o,s)=>{if(r.coerce)try{o.value=!!o.value}catch{}const c=o.value;return typeof c=="boolean"||o.issues.push({expected:"boolean",code:"invalid_type",input:c,inst:n}),o}}),xXe=rn("$ZodBigInt",(n,r)=>{Qa.init(n,r),n._zod.pattern=Jnn,n._zod.parse=(o,s)=>{if(r.coerce)try{o.value=BigInt(o.value)}catch{}return typeof o.value=="bigint"||o.issues.push({expected:"bigint",code:"invalid_type",input:o.value,inst:n}),o}}),ein=rn("$ZodBigInt",(n,r)=>{urn.init(n,r),xXe.init(n,r)}),tin=rn("$ZodSymbol",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value;return typeof c=="symbol"||o.issues.push({expected:"symbol",code:"invalid_type",input:c,inst:n}),o}}),nin=rn("$ZodUndefined",(n,r)=>{Qa.init(n,r),n._zod.pattern=nrn,n._zod.values=new Set([void 0]),n._zod.optin="optional",n._zod.optout="optional",n._zod.parse=(o,s)=>{const c=o.value;return typeof c>"u"||o.issues.push({expected:"undefined",code:"invalid_type",input:c,inst:n}),o}}),rin=rn("$ZodNull",(n,r)=>{Qa.init(n,r),n._zod.pattern=trn,n._zod.values=new Set([null]),n._zod.parse=(o,s)=>{const c=o.value;return c===null||o.issues.push({expected:"null",code:"invalid_type",input:c,inst:n}),o}}),iin=rn("$ZodAny",(n,r)=>{Qa.init(n,r),n._zod.parse=o=>o}),swe=rn("$ZodUnknown",(n,r)=>{Qa.init(n,r),n._zod.parse=o=>o}),oin=rn("$ZodNever",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>(o.issues.push({expected:"never",code:"invalid_type",input:o.value,inst:n}),o)}),ain=rn("$ZodVoid",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value;return typeof c>"u"||o.issues.push({expected:"void",code:"invalid_type",input:c,inst:n}),o}}),sin=rn("$ZodDate",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{if(r.coerce)try{o.value=new Date(o.value)}catch{}const c=o.value,d=c instanceof Date;return d&&!Number.isNaN(c.getTime())||o.issues.push({expected:"date",code:"invalid_type",input:c,...d?{received:"Invalid Date"}:{},inst:n}),o}});function jjt(n,r,o){n.issues.length&&r.issues.push(...$x(o,n.issues)),r.value[o]=n.value}const CXe=rn("$ZodArray",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value;if(!Array.isArray(c))return o.issues.push({expected:"array",code:"invalid_type",input:c,inst:n}),o;o.value=Array(c.length);const d=[];for(let h=0;h<c.length;h++){const b=c[h],y=r.element._zod.run({value:b,issues:[]},s);y instanceof Promise?d.push(y.then(E=>jjt(E,o,h))):jjt(y,o,h)}return d.length?Promise.all(d).then(()=>o):o}});function t0e(n,r,o,s){n.issues.length&&r.issues.push(...$x(o,n.issues)),n.value===void 0?o in s&&(r.value[o]=void 0):r.value[o]=n.value}const cin=rn("$ZodObject",(n,r)=>{Qa.init(n,r);const o=zye(()=>{const x=Object.keys(r.shape);for(const P of x)if(!(r.shape[P]instanceof Qa))throw new Error(`Invalid element at key "${P}": expected a Zod schema`);const O=pnn(r.shape);return{shape:r.shape,keys:x,keySet:new Set(x),numKeys:x.length,optionalKeys:new Set(O)}});du(n._zod,"propValues",()=>{const x=r.shape,O={};for(const P in x){const D=x[P]._zod;if(D.values){O[P]??(O[P]=new Set);for(const $ of D.values)O[P].add($)}}return O});const s=x=>{const O=new Trn(["shape","payload","ctx"]),P=o.value,D=G=>{const K=zVe(G);return`shape[${K}]._zod.run({ value: input[${K}], issues: [] }, ctx)`};O.write("const input = payload.value;");const $=Object.create(null);let U=0;for(const G of P.keys)$[G]=`key_${U++}`;O.write("const newResult = {}");for(const G of P.keys){const K=$[G],X=zVe(G);O.write(`const ${K} = ${D(G)};`),O.write(`
        if (${K}.issues.length) {
          payload.issues = payload.issues.concat(${K}.issues.map(iss => ({
            ...iss,
            path: iss.path ? [${X}, ...iss.path] : [${X}]
          })));
        }

        if (${K}.value === undefined) {
          if (${X} in input) {
            newResult[${X}] = undefined;
          }
        } else {
          newResult[${X}] = ${K}.value;
        }
      `)}O.write("payload.value = newResult;"),O.write("return payload;");const B=O.compile();return(G,K)=>B(x,G,K)};let c;const d=zie,h=!iwe.jitless,y=h&&fnn.value,E=r.catchall;let k;n._zod.parse=(x,O)=>{k??(k=o.value);const P=x.value;if(!d(P))return x.issues.push({expected:"object",code:"invalid_type",input:P,inst:n}),x;const D=[];if(h&&y&&O?.async===!1&&O.jitless!==!0)c||(c=s(r.shape)),x=c(x,O);else{x.value={};const K=k.shape;for(const X of k.keys){const oe=K[X]._zod.run({value:P[X],issues:[]},O);oe instanceof Promise?D.push(oe.then(he=>t0e(he,x,X,P))):t0e(oe,x,X,P)}}if(!E)return D.length?Promise.all(D).then(()=>x):x;const $=[],U=k.keySet,B=E._zod,G=B.def.type;for(const K of Object.keys(P)){if(U.has(K))continue;if(G==="never"){$.push(K);continue}const X=B.run({value:P[K],issues:[]},O);X instanceof Promise?D.push(X.then(ae=>t0e(ae,x,K,P))):t0e(X,x,K,P)}return $.length&&x.issues.push({code:"unrecognized_keys",keys:$,input:P,inst:n}),D.length?Promise.all(D).then(()=>x):x}});function $jt(n,r,o,s){for(const d of n)if(d.issues.length===0)return r.value=d.value,r;const c=n.filter(d=>!CW(d));return c.length===1?(r.value=c[0].value,c[0]):(r.issues.push({code:"invalid_union",input:r.value,inst:o,errors:n.map(d=>d.issues.map(h=>qx(h,s,cv())))}),r)}const AXe=rn("$ZodUnion",(n,r)=>{Qa.init(n,r),du(n._zod,"optin",()=>r.options.some(c=>c._zod.optin==="optional")?"optional":void 0),du(n._zod,"optout",()=>r.options.some(c=>c._zod.optout==="optional")?"optional":void 0),du(n._zod,"values",()=>{if(r.options.every(c=>c._zod.values))return new Set(r.options.flatMap(c=>Array.from(c._zod.values)))}),du(n._zod,"pattern",()=>{if(r.options.every(c=>c._zod.pattern)){const c=r.options.map(d=>d._zod.pattern);return new RegExp(`^(${c.map(d=>Hye(d.source)).join("|")})$`)}});const o=r.options.length===1,s=r.options[0]._zod.run;n._zod.parse=(c,d)=>{if(o)return s(c,d);let h=!1;const b=[];for(const y of r.options){const E=y._zod.run({value:c.value,issues:[]},d);if(E instanceof Promise)b.push(E),h=!0;else{if(E.issues.length===0)return E;b.push(E)}}return h?Promise.all(b).then(y=>$jt(y,c,n,d)):$jt(b,c,n,d)}}),uin=rn("$ZodDiscriminatedUnion",(n,r)=>{AXe.init(n,r);const o=n._zod.parse;du(n._zod,"propValues",()=>{const c={};for(const d of r.options){const h=d._zod.propValues;if(!h||Object.keys(h).length===0)throw new Error(`Invalid discriminated union option at index "${r.options.indexOf(d)}"`);for(const[b,y]of Object.entries(h)){c[b]||(c[b]=new Set);for(const E of y)c[b].add(E)}}return c});const s=zye(()=>{const c=r.options,d=new Map;for(const h of c){const b=h._zod.propValues?.[r.discriminator];if(!b||b.size===0)throw new Error(`Invalid discriminated union option at index "${r.options.indexOf(h)}"`);for(const y of b){if(d.has(y))throw new Error(`Duplicate discriminator value "${String(y)}"`);d.set(y,h)}}return d});n._zod.parse=(c,d)=>{const h=c.value;if(!zie(h))return c.issues.push({code:"invalid_type",expected:"object",input:h,inst:n}),c;const b=s.value.get(h?.[r.discriminator]);return b?b._zod.run(c,d):r.unionFallback?o(c,d):(c.issues.push({code:"invalid_union",errors:[],note:"No matching discriminator",discriminator:r.discriminator,input:h,path:[r.discriminator],inst:n}),c)}}),lin=rn("$ZodIntersection",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value,d=r.left._zod.run({value:c,issues:[]},s),h=r.right._zod.run({value:c,issues:[]},s);return d instanceof Promise||h instanceof Promise?Promise.all([d,h]).then(([y,E])=>Fjt(o,y,E)):Fjt(o,d,h)}});function qVe(n,r){if(n===r)return{valid:!0,data:n};if(n instanceof Date&&r instanceof Date&&+n==+r)return{valid:!0,data:n};if(Hie(n)&&Hie(r)){const o=Object.keys(r),s=Object.keys(n).filter(d=>o.indexOf(d)!==-1),c={...n,...r};for(const d of s){const h=qVe(n[d],r[d]);if(!h.valid)return{valid:!1,mergeErrorPath:[d,...h.mergeErrorPath]};c[d]=h.data}return{valid:!0,data:c}}if(Array.isArray(n)&&Array.isArray(r)){if(n.length!==r.length)return{valid:!1,mergeErrorPath:[]};const o=[];for(let s=0;s<n.length;s++){const c=n[s],d=r[s],h=qVe(c,d);if(!h.valid)return{valid:!1,mergeErrorPath:[s,...h.mergeErrorPath]};o.push(h.data)}return{valid:!0,data:o}}return{valid:!1,mergeErrorPath:[]}}function Fjt(n,r,o){if(r.issues.length&&n.issues.push(...r.issues),o.issues.length&&n.issues.push(...o.issues),CW(n))return n;const s=qVe(r.value,o.value);if(!s.valid)throw new Error(`Unmergable intersection. Error path: ${JSON.stringify(s.mergeErrorPath)}`);return n.value=s.data,n}const Vye=rn("$ZodTuple",(n,r)=>{Qa.init(n,r);const o=r.items,s=o.length-[...o].reverse().findIndex(c=>c._zod.optin!=="optional");n._zod.parse=(c,d)=>{const h=c.value;if(!Array.isArray(h))return c.issues.push({input:h,inst:n,expected:"tuple",code:"invalid_type"}),c;c.value=[];const b=[];if(!r.rest){const E=h.length>o.length,k=h.length<s-1;if(E||k)return c.issues.push({...E?{code:"too_big",maximum:o.length}:{code:"too_small",minimum:o.length},input:h,inst:n,origin:"array"}),c}let y=-1;for(const E of o){if(y++,y>=h.length&&y>=s)continue;const k=E._zod.run({value:h[y],issues:[]},d);k instanceof Promise?b.push(k.then(x=>n0e(x,c,y))):n0e(k,c,y)}if(r.rest){const E=h.slice(o.length);for(const k of E){y++;const x=r.rest._zod.run({value:k,issues:[]},d);x instanceof Promise?b.push(x.then(O=>n0e(O,c,y))):n0e(x,c,y)}}return b.length?Promise.all(b).then(()=>c):c}});function n0e(n,r,o){n.issues.length&&r.issues.push(...$x(o,n.issues)),r.value[o]=n.value}const din=rn("$ZodRecord",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value;if(!Hie(c))return o.issues.push({expected:"record",code:"invalid_type",input:c,inst:n}),o;const d=[];if(r.keyType._zod.values){const h=r.keyType._zod.values;o.value={};for(const y of h)if(typeof y=="string"||typeof y=="number"||typeof y=="symbol"){const E=r.valueType._zod.run({value:c[y],issues:[]},s);E instanceof Promise?d.push(E.then(k=>{k.issues.length&&o.issues.push(...$x(y,k.issues)),o.value[y]=k.value})):(E.issues.length&&o.issues.push(...$x(y,E.issues)),o.value[y]=E.value)}let b;for(const y in c)h.has(y)||(b=b??[],b.push(y));b&&b.length>0&&o.issues.push({code:"unrecognized_keys",input:c,inst:n,keys:b})}else{o.value={};for(const h of Reflect.ownKeys(c)){if(h==="__proto__")continue;const b=r.keyType._zod.run({value:h,issues:[]},s);if(b instanceof Promise)throw new Error("Async schemas not supported in object keys currently");if(b.issues.length){o.issues.push({code:"invalid_key",origin:"record",issues:b.issues.map(E=>qx(E,s,cv())),input:h,path:[h],inst:n}),o.value[b.value]=b.value;continue}const y=r.valueType._zod.run({value:c[h],issues:[]},s);y instanceof Promise?d.push(y.then(E=>{E.issues.length&&o.issues.push(...$x(h,E.issues)),o.value[b.value]=E.value})):(y.issues.length&&o.issues.push(...$x(h,y.issues)),o.value[b.value]=y.value)}}return d.length?Promise.all(d).then(()=>o):o}}),fin=rn("$ZodMap",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value;if(!(c instanceof Map))return o.issues.push({expected:"map",code:"invalid_type",input:c,inst:n}),o;const d=[];o.value=new Map;for(const[h,b]of c){const y=r.keyType._zod.run({value:h,issues:[]},s),E=r.valueType._zod.run({value:b,issues:[]},s);y instanceof Promise||E instanceof Promise?d.push(Promise.all([y,E]).then(([k,x])=>{Bjt(k,x,o,h,c,n,s)})):Bjt(y,E,o,h,c,n,s)}return d.length?Promise.all(d).then(()=>o):o}});function Bjt(n,r,o,s,c,d,h){n.issues.length&&(awe.has(typeof s)?o.issues.push(...$x(s,n.issues)):o.issues.push({code:"invalid_key",origin:"map",input:c,inst:d,issues:n.issues.map(b=>qx(b,h,cv()))})),r.issues.length&&(awe.has(typeof s)?o.issues.push(...$x(s,r.issues)):o.issues.push({origin:"map",code:"invalid_element",input:c,inst:d,key:s,issues:r.issues.map(b=>qx(b,h,cv()))})),o.value.set(n.value,r.value)}const hin=rn("$ZodSet",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value;if(!(c instanceof Set))return o.issues.push({input:c,inst:n,expected:"set",code:"invalid_type"}),o;const d=[];o.value=new Set;for(const h of c){const b=r.valueType._zod.run({value:h,issues:[]},s);b instanceof Promise?d.push(b.then(y=>Ujt(y,o))):Ujt(b,o)}return d.length?Promise.all(d).then(()=>o):o}});function Ujt(n,r){n.issues.length&&r.issues.push(...n.issues),r.value.add(n.value)}const pin=rn("$ZodEnum",(n,r)=>{Qa.init(n,r);const o=lXe(r.entries),s=new Set(o);n._zod.values=s,n._zod.pattern=new RegExp(`^(${o.filter(c=>awe.has(typeof c)).map(c=>typeof c=="string"?OM(c):c.toString()).join("|")})$`),n._zod.parse=(c,d)=>{const h=c.value;return s.has(h)||c.issues.push({code:"invalid_value",values:o,input:h,inst:n}),c}}),gin=rn("$ZodLiteral",(n,r)=>{if(Qa.init(n,r),r.values.length===0)throw new Error("Cannot create literal schema with no valid values");n._zod.values=new Set(r.values),n._zod.pattern=new RegExp(`^(${r.values.map(o=>typeof o=="string"?OM(o):o?OM(o.toString()):String(o)).join("|")})$`),n._zod.parse=(o,s)=>{const c=o.value;return n._zod.values.has(c)||o.issues.push({code:"invalid_value",values:r.values,input:c,inst:n}),o}}),min=rn("$ZodFile",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=o.value;return c instanceof File||o.issues.push({expected:"file",code:"invalid_type",input:c,inst:n}),o}}),IXe=rn("$ZodTransform",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=r.transform(o.value,o);if(s.async)return(c instanceof Promise?c:Promise.resolve(c)).then(h=>(o.value=h,o));if(c instanceof Promise)throw new XW;return o.value=c,o}});function zjt(n,r){return n.issues.length&&r===void 0?{issues:[],value:void 0}:n}const bin=rn("$ZodOptional",(n,r)=>{Qa.init(n,r),n._zod.optin="optional",n._zod.optout="optional",du(n._zod,"values",()=>r.innerType._zod.values?new Set([...r.innerType._zod.values,void 0]):void 0),du(n._zod,"pattern",()=>{const o=r.innerType._zod.pattern;return o?new RegExp(`^(${Hye(o.source)})?$`):void 0}),n._zod.parse=(o,s)=>{if(r.innerType._zod.optin==="optional"){const c=r.innerType._zod.run(o,s);return c instanceof Promise?c.then(d=>zjt(d,o.value)):zjt(c,o.value)}return o.value===void 0?o:r.innerType._zod.run(o,s)}}),vin=rn("$ZodNullable",(n,r)=>{Qa.init(n,r),du(n._zod,"optin",()=>r.innerType._zod.optin),du(n._zod,"optout",()=>r.innerType._zod.optout),du(n._zod,"pattern",()=>{const o=r.innerType._zod.pattern;return o?new RegExp(`^(${Hye(o.source)}|null)$`):void 0}),du(n._zod,"values",()=>r.innerType._zod.values?new Set([...r.innerType._zod.values,null]):void 0),n._zod.parse=(o,s)=>o.value===null?o:r.innerType._zod.run(o,s)}),win=rn("$ZodDefault",(n,r)=>{Qa.init(n,r),n._zod.optin="optional",du(n._zod,"values",()=>r.innerType._zod.values),n._zod.parse=(o,s)=>{if(o.value===void 0)return o.value=r.defaultValue,o;const c=r.innerType._zod.run(o,s);return c instanceof Promise?c.then(d=>Hjt(d,r)):Hjt(c,r)}});function Hjt(n,r){return n.value===void 0&&(n.value=r.defaultValue),n}const yin=rn("$ZodPrefault",(n,r)=>{Qa.init(n,r),n._zod.optin="optional",du(n._zod,"values",()=>r.innerType._zod.values),n._zod.parse=(o,s)=>(o.value===void 0&&(o.value=r.defaultValue),r.innerType._zod.run(o,s))}),Ein=rn("$ZodNonOptional",(n,r)=>{Qa.init(n,r),du(n._zod,"values",()=>{const o=r.innerType._zod.values;return o?new Set([...o].filter(s=>s!==void 0)):void 0}),n._zod.parse=(o,s)=>{const c=r.innerType._zod.run(o,s);return c instanceof Promise?c.then(d=>Gjt(d,n)):Gjt(c,n)}});function Gjt(n,r){return!n.issues.length&&n.value===void 0&&n.issues.push({code:"invalid_type",expected:"nonoptional",input:n.value,inst:r}),n}const Sin=rn("$ZodSuccess",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>{const c=r.innerType._zod.run(o,s);return c instanceof Promise?c.then(d=>(o.value=d.issues.length===0,o)):(o.value=c.issues.length===0,o)}}),_in=rn("$ZodCatch",(n,r)=>{Qa.init(n,r),du(n._zod,"optin",()=>r.innerType._zod.optin),du(n._zod,"optout",()=>r.innerType._zod.optout),du(n._zod,"values",()=>r.innerType._zod.values),n._zod.parse=(o,s)=>{const c=r.innerType._zod.run(o,s);return c instanceof Promise?c.then(d=>(o.value=d.value,d.issues.length&&(o.value=r.catchValue({...o,error:{issues:d.issues.map(h=>qx(h,s,cv()))},input:o.value}),o.issues=[]),o)):(o.value=c.value,c.issues.length&&(o.value=r.catchValue({...o,error:{issues:c.issues.map(d=>qx(d,s,cv()))},input:o.value}),o.issues=[]),o)}}),kin=rn("$ZodNaN",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>((typeof o.value!="number"||!Number.isNaN(o.value))&&o.issues.push({input:o.value,inst:n,expected:"nan",code:"invalid_type"}),o)}),OXe=rn("$ZodPipe",(n,r)=>{Qa.init(n,r),du(n._zod,"values",()=>r.in._zod.values),du(n._zod,"optin",()=>r.in._zod.optin),du(n._zod,"optout",()=>r.out._zod.optout),du(n._zod,"propValues",()=>r.in._zod.propValues),n._zod.parse=(o,s)=>{const c=r.in._zod.run(o,s);return c instanceof Promise?c.then(d=>qjt(d,r,s)):qjt(c,r,s)}});function qjt(n,r,o){return n.issues.length?n:r.out._zod.run({value:n.value,issues:n.issues},o)}const Tin=rn("$ZodReadonly",(n,r)=>{Qa.init(n,r),du(n._zod,"propValues",()=>r.innerType._zod.propValues),du(n._zod,"values",()=>r.innerType._zod.values),du(n._zod,"optin",()=>r.innerType._zod.optin),du(n._zod,"optout",()=>r.innerType._zod.optout),n._zod.parse=(o,s)=>{const c=r.innerType._zod.run(o,s);return c instanceof Promise?c.then(Vjt):Vjt(c)}});function Vjt(n){return n.value=Object.freeze(n.value),n}const xin=rn("$ZodTemplateLiteral",(n,r)=>{Qa.init(n,r);const o=[];for(const s of r.parts)if(s instanceof Qa){if(!s._zod.pattern)throw new Error(`Invalid template literal part, no pattern found: ${[...s._zod.traits].shift()}`);const c=s._zod.pattern instanceof RegExp?s._zod.pattern.source:s._zod.pattern;if(!c)throw new Error(`Invalid template literal part: ${s._zod.traits}`);const d=c.startsWith("^")?1:0,h=c.endsWith("$")?c.length-1:c.length;o.push(c.slice(d,h))}else if(s===null||hnn.has(typeof s))o.push(OM(`${s}`));else throw new Error(`Invalid template literal part: ${s}`);n._zod.pattern=new RegExp(`^${o.join("")}$`),n._zod.parse=(s,c)=>typeof s.value!="string"?(s.issues.push({input:s.value,inst:n,expected:"template_literal",code:"invalid_type"}),s):(n._zod.pattern.lastIndex=0,n._zod.pattern.test(s.value)||s.issues.push({input:s.value,inst:n,code:"invalid_format",format:r.format??"template_literal",pattern:n._zod.pattern.source}),s)}),Cin=rn("$ZodPromise",(n,r)=>{Qa.init(n,r),n._zod.parse=(o,s)=>Promise.resolve(o.value).then(c=>r.innerType._zod.run({value:c,issues:[]},s))}),Ain=rn("$ZodLazy",(n,r)=>{Qa.init(n,r),du(n._zod,"innerType",()=>r.getter()),du(n._zod,"pattern",()=>n._zod.innerType._zod.pattern),du(n._zod,"propValues",()=>n._zod.innerType._zod.propValues),du(n._zod,"optin",()=>n._zod.innerType._zod.optin??void 0),du(n._zod,"optout",()=>n._zod.innerType._zod.optout??void 0),n._zod.parse=(o,s)=>n._zod.innerType._zod.run(o,s)}),Iin=rn("$ZodCustom",(n,r)=>{Ph.init(n,r),Qa.init(n,r),n._zod.parse=(o,s)=>o,n._zod.check=o=>{const s=o.value,c=r.fn(s);if(c instanceof Promise)return c.then(d=>Wjt(d,o,s,n));Wjt(c,o,s,n)}});function Wjt(n,r,o,s){if(!n){const c={code:"custom",input:o,inst:s,path:[...s._zod.def.path??[]],continue:!s._zod.def.abort};s._zod.def.params&&(c.params=s._zod.def.params),r.issues.push(JW(c))}}const Aar=()=>{const n={string:{unit:"حرف",verb:"أن يحوي"},file:{unit:"بايت",verb:"أن يحوي"},array:{unit:"عنصر",verb:"أن يحوي"},set:{unit:"عنصر",verb:"أن يحوي"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"مدخل",email:"بريد إلكتروني",url:"رابط",emoji:"إيموجي",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"تاريخ ووقت بمعيار ISO",date:"تاريخ بمعيار ISO",time:"وقت بمعيار ISO",duration:"مدة بمعيار ISO",ipv4:"عنوان IPv4",ipv6:"عنوان IPv6",cidrv4:"مدى عناوين بصيغة IPv4",cidrv6:"مدى عناوين بصيغة IPv6",base64:"نَص بترميز base64-encoded",base64url:"نَص بترميز base64url-encoded",json_string:"نَص على هيئة JSON",e164:"رقم هاتف بمعيار E.164",jwt:"JWT",template_literal:"مدخل"};return c=>{switch(c.code){case"invalid_type":return`مدخلات غير مقبولة: يفترض إدخال ${c.expected}، ولكن تم إدخال ${o(c.input)}`;case"invalid_value":return c.values.length===1?`مدخلات غير مقبولة: يفترض إدخال ${es(c.values[0])}`:`اختيار غير مقبول: يتوقع انتقاء أحد هذه الخيارات: ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?` أكبر من اللازم: يفترض أن تكون ${c.origin??"القيمة"} ${d} ${c.maximum.toString()} ${h.unit??"عنصر"}`:`أكبر من اللازم: يفترض أن تكون ${c.origin??"القيمة"} ${d} ${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`أصغر من اللازم: يفترض لـ ${c.origin} أن يكون ${d} ${c.minimum.toString()} ${h.unit}`:`أصغر من اللازم: يفترض لـ ${c.origin} أن يكون ${d} ${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`نَص غير مقبول: يجب أن يبدأ بـ "${c.prefix}"`:d.format==="ends_with"?`نَص غير مقبول: يجب أن ينتهي بـ "${d.suffix}"`:d.format==="includes"?`نَص غير مقبول: يجب أن يتضمَّن "${d.includes}"`:d.format==="regex"?`نَص غير مقبول: يجب أن يطابق النمط ${d.pattern}`:`${s[d.format]??c.format} غير مقبول`}case"not_multiple_of":return`رقم غير مقبول: يجب أن يكون من مضاعفات ${c.divisor}`;case"unrecognized_keys":return`معرف${c.keys.length>1?"ات":""} غريب${c.keys.length>1?"ة":""}: ${ti(c.keys,"، ")}`;case"invalid_key":return`معرف غير مقبول في ${c.origin}`;case"invalid_union":return"مدخل غير مقبول";case"invalid_element":return`مدخل غير مقبول في ${c.origin}`;default:return"مدخل غير مقبول"}}};function Iar(){return{localeError:Aar()}}const Oar=()=>{const n={string:{unit:"simvol",verb:"olmalıdır"},file:{unit:"bayt",verb:"olmalıdır"},array:{unit:"element",verb:"olmalıdır"},set:{unit:"element",verb:"olmalıdır"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"input",email:"email address",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO datetime",date:"ISO date",time:"ISO time",duration:"ISO duration",ipv4:"IPv4 address",ipv6:"IPv6 address",cidrv4:"IPv4 range",cidrv6:"IPv6 range",base64:"base64-encoded string",base64url:"base64url-encoded string",json_string:"JSON string",e164:"E.164 number",jwt:"JWT",template_literal:"input"};return c=>{switch(c.code){case"invalid_type":return`Yanlış dəyər: gözlənilən ${c.expected}, daxil olan ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Yanlış dəyər: gözlənilən ${es(c.values[0])}`:`Yanlış seçim: aşağıdakılardan biri olmalıdır: ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Çox böyük: gözlənilən ${c.origin??"dəyər"} ${d}${c.maximum.toString()} ${h.unit??"element"}`:`Çox böyük: gözlənilən ${c.origin??"dəyər"} ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Çox kiçik: gözlənilən ${c.origin} ${d}${c.minimum.toString()} ${h.unit}`:`Çox kiçik: gözlənilən ${c.origin} ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Yanlış mətn: "${d.prefix}" ilə başlamalıdır`:d.format==="ends_with"?`Yanlış mətn: "${d.suffix}" ilə bitməlidir`:d.format==="includes"?`Yanlış mətn: "${d.includes}" daxil olmalıdır`:d.format==="regex"?`Yanlış mətn: ${d.pattern} şablonuna uyğun olmalıdır`:`Yanlış ${s[d.format]??c.format}`}case"not_multiple_of":return`Yanlış ədəd: ${c.divisor} ilə bölünə bilən olmalıdır`;case"unrecognized_keys":return`Tanınmayan açar${c.keys.length>1?"lar":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`${c.origin} daxilində yanlış açar`;case"invalid_union":return"Yanlış dəyər";case"invalid_element":return`${c.origin} daxilində yanlış dəyər`;default:return"Yanlış dəyər"}}};function Nar(){return{localeError:Oar()}}function Kjt(n,r,o,s){const c=Math.abs(n),d=c%10,h=c%100;return h>=11&&h<=19?s:d===1?r:d>=2&&d<=4?o:s}const Rar=()=>{const n={string:{unit:{one:"сімвал",few:"сімвалы",many:"сімвалаў"},verb:"мець"},array:{unit:{one:"элемент",few:"элементы",many:"элементаў"},verb:"мець"},set:{unit:{one:"элемент",few:"элементы",many:"элементаў"},verb:"мець"},file:{unit:{one:"байт",few:"байты",many:"байтаў"},verb:"мець"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"лік";case"object":{if(Array.isArray(c))return"масіў";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"увод",email:"email адрас",url:"URL",emoji:"эмодзі",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO дата і час",date:"ISO дата",time:"ISO час",duration:"ISO працягласць",ipv4:"IPv4 адрас",ipv6:"IPv6 адрас",cidrv4:"IPv4 дыяпазон",cidrv6:"IPv6 дыяпазон",base64:"радок у фармаце base64",base64url:"радок у фармаце base64url",json_string:"JSON радок",e164:"нумар E.164",jwt:"JWT",template_literal:"увод"};return c=>{switch(c.code){case"invalid_type":return`Няправільны ўвод: чакаўся ${c.expected}, атрымана ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Няправільны ўвод: чакалася ${es(c.values[0])}`:`Няправільны варыянт: чакаўся адзін з ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);if(h){const b=Number(c.maximum),y=Kjt(b,h.unit.one,h.unit.few,h.unit.many);return`Занадта вялікі: чакалася, што ${c.origin??"значэнне"} павінна ${h.verb} ${d}${c.maximum.toString()} ${y}`}return`Занадта вялікі: чакалася, што ${c.origin??"значэнне"} павінна быць ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);if(h){const b=Number(c.minimum),y=Kjt(b,h.unit.one,h.unit.few,h.unit.many);return`Занадта малы: чакалася, што ${c.origin} павінна ${h.verb} ${d}${c.minimum.toString()} ${y}`}return`Занадта малы: чакалася, што ${c.origin} павінна быць ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Няправільны радок: павінен пачынацца з "${d.prefix}"`:d.format==="ends_with"?`Няправільны радок: павінен заканчвацца на "${d.suffix}"`:d.format==="includes"?`Няправільны радок: павінен змяшчаць "${d.includes}"`:d.format==="regex"?`Няправільны радок: павінен адпавядаць шаблону ${d.pattern}`:`Няправільны ${s[d.format]??c.format}`}case"not_multiple_of":return`Няправільны лік: павінен быць кратным ${c.divisor}`;case"unrecognized_keys":return`Нераспазнаны ${c.keys.length>1?"ключы":"ключ"}: ${ti(c.keys,", ")}`;case"invalid_key":return`Няправільны ключ у ${c.origin}`;case"invalid_union":return"Няправільны ўвод";case"invalid_element":return`Няправільнае значэнне ў ${c.origin}`;default:return"Няправільны ўвод"}}};function Par(){return{localeError:Rar()}}const Lar=()=>{const n={string:{unit:"caràcters",verb:"contenir"},file:{unit:"bytes",verb:"contenir"},array:{unit:"elements",verb:"contenir"},set:{unit:"elements",verb:"contenir"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"entrada",email:"adreça electrònica",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"data i hora ISO",date:"data ISO",time:"hora ISO",duration:"durada ISO",ipv4:"adreça IPv4",ipv6:"adreça IPv6",cidrv4:"rang IPv4",cidrv6:"rang IPv6",base64:"cadena codificada en base64",base64url:"cadena codificada en base64url",json_string:"cadena JSON",e164:"número E.164",jwt:"JWT",template_literal:"entrada"};return c=>{switch(c.code){case"invalid_type":return`Tipus invàlid: s'esperava ${c.expected}, s'ha rebut ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Valor invàlid: s'esperava ${es(c.values[0])}`:`Opció invàlida: s'esperava una de ${ti(c.values," o ")}`;case"too_big":{const d=c.inclusive?"com a màxim":"menys de",h=r(c.origin);return h?`Massa gran: s'esperava que ${c.origin??"el valor"} contingués ${d} ${c.maximum.toString()} ${h.unit??"elements"}`:`Massa gran: s'esperava que ${c.origin??"el valor"} fos ${d} ${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?"com a mínim":"més de",h=r(c.origin);return h?`Massa petit: s'esperava que ${c.origin} contingués ${d} ${c.minimum.toString()} ${h.unit}`:`Massa petit: s'esperava que ${c.origin} fos ${d} ${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Format invàlid: ha de començar amb "${d.prefix}"`:d.format==="ends_with"?`Format invàlid: ha d'acabar amb "${d.suffix}"`:d.format==="includes"?`Format invàlid: ha d'incloure "${d.includes}"`:d.format==="regex"?`Format invàlid: ha de coincidir amb el patró ${d.pattern}`:`Format invàlid per a ${s[d.format]??c.format}`}case"not_multiple_of":return`Número invàlid: ha de ser múltiple de ${c.divisor}`;case"unrecognized_keys":return`Clau${c.keys.length>1?"s":""} no reconeguda${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Clau invàlida a ${c.origin}`;case"invalid_union":return"Entrada invàlida";case"invalid_element":return`Element invàlid a ${c.origin}`;default:return"Entrada invàlida"}}};function Dar(){return{localeError:Lar()}}const Mar=()=>{const n={string:{unit:"znaků",verb:"mít"},file:{unit:"bajtů",verb:"mít"},array:{unit:"prvků",verb:"mít"},set:{unit:"prvků",verb:"mít"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"číslo";case"string":return"řetězec";case"boolean":return"boolean";case"bigint":return"bigint";case"function":return"funkce";case"symbol":return"symbol";case"undefined":return"undefined";case"object":{if(Array.isArray(c))return"pole";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"regulární výraz",email:"e-mailová adresa",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"datum a čas ve formátu ISO",date:"datum ve formátu ISO",time:"čas ve formátu ISO",duration:"doba trvání ISO",ipv4:"IPv4 adresa",ipv6:"IPv6 adresa",cidrv4:"rozsah IPv4",cidrv6:"rozsah IPv6",base64:"řetězec zakódovaný ve formátu base64",base64url:"řetězec zakódovaný ve formátu base64url",json_string:"řetězec ve formátu JSON",e164:"číslo E.164",jwt:"JWT",template_literal:"vstup"};return c=>{switch(c.code){case"invalid_type":return`Neplatný vstup: očekáváno ${c.expected}, obdrženo ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Neplatný vstup: očekáváno ${es(c.values[0])}`:`Neplatná možnost: očekávána jedna z hodnot ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Hodnota je příliš velká: ${c.origin??"hodnota"} musí mít ${d}${c.maximum.toString()} ${h.unit??"prvků"}`:`Hodnota je příliš velká: ${c.origin??"hodnota"} musí být ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Hodnota je příliš malá: ${c.origin??"hodnota"} musí mít ${d}${c.minimum.toString()} ${h.unit??"prvků"}`:`Hodnota je příliš malá: ${c.origin??"hodnota"} musí být ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Neplatný řetězec: musí začínat na "${d.prefix}"`:d.format==="ends_with"?`Neplatný řetězec: musí končit na "${d.suffix}"`:d.format==="includes"?`Neplatný řetězec: musí obsahovat "${d.includes}"`:d.format==="regex"?`Neplatný řetězec: musí odpovídat vzoru ${d.pattern}`:`Neplatný formát ${s[d.format]??c.format}`}case"not_multiple_of":return`Neplatné číslo: musí být násobkem ${c.divisor}`;case"unrecognized_keys":return`Neznámé klíče: ${ti(c.keys,", ")}`;case"invalid_key":return`Neplatný klíč v ${c.origin}`;case"invalid_union":return"Neplatný vstup";case"invalid_element":return`Neplatná hodnota v ${c.origin}`;default:return"Neplatný vstup"}}};function jar(){return{localeError:Mar()}}const $ar=()=>{const n={string:{unit:"tegn",verb:"havde"},file:{unit:"bytes",verb:"havde"},array:{unit:"elementer",verb:"indeholdt"},set:{unit:"elementer",verb:"indeholdt"}},r={string:"streng",number:"tal",boolean:"boolean",array:"liste",object:"objekt",set:"sæt",file:"fil"};function o(h){return n[h]??null}function s(h){return r[h]??h}const c=h=>{const b=typeof h;switch(b){case"number":return Number.isNaN(h)?"NaN":"tal";case"object":return Array.isArray(h)?"liste":h===null?"null":Object.getPrototypeOf(h)!==Object.prototype&&h.constructor?h.constructor.name:"objekt"}return b},d={regex:"input",email:"e-mailadresse",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO dato- og klokkeslæt",date:"ISO-dato",time:"ISO-klokkeslæt",duration:"ISO-varighed",ipv4:"IPv4-område",ipv6:"IPv6-område",cidrv4:"IPv4-spektrum",cidrv6:"IPv6-spektrum",base64:"base64-kodet streng",base64url:"base64url-kodet streng",json_string:"JSON-streng",e164:"E.164-nummer",jwt:"JWT",template_literal:"input"};return h=>{switch(h.code){case"invalid_type":return`Ugyldigt input: forventede ${s(h.expected)}, fik ${s(c(h.input))}`;case"invalid_value":return h.values.length===1?`Ugyldig værdi: forventede ${es(h.values[0])}`:`Ugyldigt valg: forventede en af følgende ${ti(h.values,"|")}`;case"too_big":{const b=h.inclusive?"<=":"<",y=o(h.origin),E=s(h.origin);return y?`For stor: forventede ${E??"value"} ${y.verb} ${b} ${h.maximum.toString()} ${y.unit??"elementer"}`:`For stor: forventede ${E??"value"} havde ${b} ${h.maximum.toString()}`}case"too_small":{const b=h.inclusive?">=":">",y=o(h.origin),E=s(h.origin);return y?`For lille: forventede ${E} ${y.verb} ${b} ${h.minimum.toString()} ${y.unit}`:`For lille: forventede ${E} havde ${b} ${h.minimum.toString()}`}case"invalid_format":{const b=h;return b.format==="starts_with"?`Ugyldig streng: skal starte med "${b.prefix}"`:b.format==="ends_with"?`Ugyldig streng: skal ende med "${b.suffix}"`:b.format==="includes"?`Ugyldig streng: skal indeholde "${b.includes}"`:b.format==="regex"?`Ugyldig streng: skal matche mønsteret ${b.pattern}`:`Ugyldig ${d[b.format]??h.format}`}case"not_multiple_of":return`Ugyldigt tal: skal være deleligt med ${h.divisor}`;case"unrecognized_keys":return`${h.keys.length>1?"Ukendte nøgler":"Ukendt nøgle"}: ${ti(h.keys,", ")}`;case"invalid_key":return`Ugyldig nøgle i ${h.origin}`;case"invalid_union":return"Ugyldigt input: matcher ingen af de tilladte typer";case"invalid_element":return`Ugyldig værdi i ${h.origin}`;default:return"Ugyldigt input"}}};function Far(){return{localeError:$ar()}}const Bar=()=>{const n={string:{unit:"Zeichen",verb:"zu haben"},file:{unit:"Bytes",verb:"zu haben"},array:{unit:"Elemente",verb:"zu haben"},set:{unit:"Elemente",verb:"zu haben"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"Zahl";case"object":{if(Array.isArray(c))return"Array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"Eingabe",email:"E-Mail-Adresse",url:"URL",emoji:"Emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO-Datum und -Uhrzeit",date:"ISO-Datum",time:"ISO-Uhrzeit",duration:"ISO-Dauer",ipv4:"IPv4-Adresse",ipv6:"IPv6-Adresse",cidrv4:"IPv4-Bereich",cidrv6:"IPv6-Bereich",base64:"Base64-codierter String",base64url:"Base64-URL-codierter String",json_string:"JSON-String",e164:"E.164-Nummer",jwt:"JWT",template_literal:"Eingabe"};return c=>{switch(c.code){case"invalid_type":return`Ungültige Eingabe: erwartet ${c.expected}, erhalten ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Ungültige Eingabe: erwartet ${es(c.values[0])}`:`Ungültige Option: erwartet eine von ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Zu groß: erwartet, dass ${c.origin??"Wert"} ${d}${c.maximum.toString()} ${h.unit??"Elemente"} hat`:`Zu groß: erwartet, dass ${c.origin??"Wert"} ${d}${c.maximum.toString()} ist`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Zu klein: erwartet, dass ${c.origin} ${d}${c.minimum.toString()} ${h.unit} hat`:`Zu klein: erwartet, dass ${c.origin} ${d}${c.minimum.toString()} ist`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Ungültiger String: muss mit "${d.prefix}" beginnen`:d.format==="ends_with"?`Ungültiger String: muss mit "${d.suffix}" enden`:d.format==="includes"?`Ungültiger String: muss "${d.includes}" enthalten`:d.format==="regex"?`Ungültiger String: muss dem Muster ${d.pattern} entsprechen`:`Ungültig: ${s[d.format]??c.format}`}case"not_multiple_of":return`Ungültige Zahl: muss ein Vielfaches von ${c.divisor} sein`;case"unrecognized_keys":return`${c.keys.length>1?"Unbekannte Schlüssel":"Unbekannter Schlüssel"}: ${ti(c.keys,", ")}`;case"invalid_key":return`Ungültiger Schlüssel in ${c.origin}`;case"invalid_union":return"Ungültige Eingabe";case"invalid_element":return`Ungültiger Wert in ${c.origin}`;default:return"Ungültige Eingabe"}}};function Uar(){return{localeError:Bar()}}const zar=n=>{const r=typeof n;switch(r){case"number":return Number.isNaN(n)?"NaN":"number";case"object":{if(Array.isArray(n))return"array";if(n===null)return"null";if(Object.getPrototypeOf(n)!==Object.prototype&&n.constructor)return n.constructor.name}}return r},Har=()=>{const n={string:{unit:"characters",verb:"to have"},file:{unit:"bytes",verb:"to have"},array:{unit:"items",verb:"to have"},set:{unit:"items",verb:"to have"}};function r(s){return n[s]??null}const o={regex:"input",email:"email address",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO datetime",date:"ISO date",time:"ISO time",duration:"ISO duration",ipv4:"IPv4 address",ipv6:"IPv6 address",cidrv4:"IPv4 range",cidrv6:"IPv6 range",base64:"base64-encoded string",base64url:"base64url-encoded string",json_string:"JSON string",e164:"E.164 number",jwt:"JWT",template_literal:"input"};return s=>{switch(s.code){case"invalid_type":return`Invalid input: expected ${s.expected}, received ${zar(s.input)}`;case"invalid_value":return s.values.length===1?`Invalid input: expected ${es(s.values[0])}`:`Invalid option: expected one of ${ti(s.values,"|")}`;case"too_big":{const c=s.inclusive?"<=":"<",d=r(s.origin);return d?`Too big: expected ${s.origin??"value"} to have ${c}${s.maximum.toString()} ${d.unit??"elements"}`:`Too big: expected ${s.origin??"value"} to be ${c}${s.maximum.toString()}`}case"too_small":{const c=s.inclusive?">=":">",d=r(s.origin);return d?`Too small: expected ${s.origin} to have ${c}${s.minimum.toString()} ${d.unit}`:`Too small: expected ${s.origin} to be ${c}${s.minimum.toString()}`}case"invalid_format":{const c=s;return c.format==="starts_with"?`Invalid string: must start with "${c.prefix}"`:c.format==="ends_with"?`Invalid string: must end with "${c.suffix}"`:c.format==="includes"?`Invalid string: must include "${c.includes}"`:c.format==="regex"?`Invalid string: must match pattern ${c.pattern}`:`Invalid ${o[c.format]??s.format}`}case"not_multiple_of":return`Invalid number: must be a multiple of ${s.divisor}`;case"unrecognized_keys":return`Unrecognized key${s.keys.length>1?"s":""}: ${ti(s.keys,", ")}`;case"invalid_key":return`Invalid key in ${s.origin}`;case"invalid_union":return"Invalid input";case"invalid_element":return`Invalid value in ${s.origin}`;default:return"Invalid input"}}};function Oin(){return{localeError:Har()}}const Gar=n=>{const r=typeof n;switch(r){case"number":return Number.isNaN(n)?"NaN":"nombro";case"object":{if(Array.isArray(n))return"tabelo";if(n===null)return"senvalora";if(Object.getPrototypeOf(n)!==Object.prototype&&n.constructor)return n.constructor.name}}return r},qar=()=>{const n={string:{unit:"karaktrojn",verb:"havi"},file:{unit:"bajtojn",verb:"havi"},array:{unit:"elementojn",verb:"havi"},set:{unit:"elementojn",verb:"havi"}};function r(s){return n[s]??null}const o={regex:"enigo",email:"retadreso",url:"URL",emoji:"emoĝio",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO-datotempo",date:"ISO-dato",time:"ISO-tempo",duration:"ISO-daŭro",ipv4:"IPv4-adreso",ipv6:"IPv6-adreso",cidrv4:"IPv4-rango",cidrv6:"IPv6-rango",base64:"64-ume kodita karaktraro",base64url:"URL-64-ume kodita karaktraro",json_string:"JSON-karaktraro",e164:"E.164-nombro",jwt:"JWT",template_literal:"enigo"};return s=>{switch(s.code){case"invalid_type":return`Nevalida enigo: atendiĝis ${s.expected}, riceviĝis ${Gar(s.input)}`;case"invalid_value":return s.values.length===1?`Nevalida enigo: atendiĝis ${es(s.values[0])}`:`Nevalida opcio: atendiĝis unu el ${ti(s.values,"|")}`;case"too_big":{const c=s.inclusive?"<=":"<",d=r(s.origin);return d?`Tro granda: atendiĝis ke ${s.origin??"valoro"} havu ${c}${s.maximum.toString()} ${d.unit??"elementojn"}`:`Tro granda: atendiĝis ke ${s.origin??"valoro"} havu ${c}${s.maximum.toString()}`}case"too_small":{const c=s.inclusive?">=":">",d=r(s.origin);return d?`Tro malgranda: atendiĝis ke ${s.origin} havu ${c}${s.minimum.toString()} ${d.unit}`:`Tro malgranda: atendiĝis ke ${s.origin} estu ${c}${s.minimum.toString()}`}case"invalid_format":{const c=s;return c.format==="starts_with"?`Nevalida karaktraro: devas komenciĝi per "${c.prefix}"`:c.format==="ends_with"?`Nevalida karaktraro: devas finiĝi per "${c.suffix}"`:c.format==="includes"?`Nevalida karaktraro: devas inkluzivi "${c.includes}"`:c.format==="regex"?`Nevalida karaktraro: devas kongrui kun la modelo ${c.pattern}`:`Nevalida ${o[c.format]??s.format}`}case"not_multiple_of":return`Nevalida nombro: devas esti oblo de ${s.divisor}`;case"unrecognized_keys":return`Nekonata${s.keys.length>1?"j":""} ŝlosilo${s.keys.length>1?"j":""}: ${ti(s.keys,", ")}`;case"invalid_key":return`Nevalida ŝlosilo en ${s.origin}`;case"invalid_union":return"Nevalida enigo";case"invalid_element":return`Nevalida valoro en ${s.origin}`;default:return"Nevalida enigo"}}};function Var(){return{localeError:qar()}}const War=()=>{const n={string:{unit:"caracteres",verb:"tener"},file:{unit:"bytes",verb:"tener"},array:{unit:"elementos",verb:"tener"},set:{unit:"elementos",verb:"tener"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"número";case"object":{if(Array.isArray(c))return"arreglo";if(c===null)return"nulo";if(Object.getPrototypeOf(c)!==Object.prototype)return c.constructor.name}}return d},s={regex:"entrada",email:"dirección de correo electrónico",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"fecha y hora ISO",date:"fecha ISO",time:"hora ISO",duration:"duración ISO",ipv4:"dirección IPv4",ipv6:"dirección IPv6",cidrv4:"rango IPv4",cidrv6:"rango IPv6",base64:"cadena codificada en base64",base64url:"URL codificada en base64",json_string:"cadena JSON",e164:"número E.164",jwt:"JWT",template_literal:"entrada"};return c=>{switch(c.code){case"invalid_type":return`Entrada inválida: se esperaba ${c.expected}, recibido ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Entrada inválida: se esperaba ${es(c.values[0])}`:`Opción inválida: se esperaba una de ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Demasiado grande: se esperaba que ${c.origin??"valor"} tuviera ${d}${c.maximum.toString()} ${h.unit??"elementos"}`:`Demasiado grande: se esperaba que ${c.origin??"valor"} fuera ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Demasiado pequeño: se esperaba que ${c.origin} tuviera ${d}${c.minimum.toString()} ${h.unit}`:`Demasiado pequeño: se esperaba que ${c.origin} fuera ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Cadena inválida: debe comenzar con "${d.prefix}"`:d.format==="ends_with"?`Cadena inválida: debe terminar en "${d.suffix}"`:d.format==="includes"?`Cadena inválida: debe incluir "${d.includes}"`:d.format==="regex"?`Cadena inválida: debe coincidir con el patrón ${d.pattern}`:`Inválido ${s[d.format]??c.format}`}case"not_multiple_of":return`Número inválido: debe ser múltiplo de ${c.divisor}`;case"unrecognized_keys":return`Llave${c.keys.length>1?"s":""} desconocida${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Llave inválida en ${c.origin}`;case"invalid_union":return"Entrada inválida";case"invalid_element":return`Valor inválido en ${c.origin}`;default:return"Entrada inválida"}}};function Kar(){return{localeError:War()}}const Yar=()=>{const n={string:{unit:"کاراکتر",verb:"داشته باشد"},file:{unit:"بایت",verb:"داشته باشد"},array:{unit:"آیتم",verb:"داشته باشد"},set:{unit:"آیتم",verb:"داشته باشد"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"عدد";case"object":{if(Array.isArray(c))return"آرایه";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"ورودی",email:"آدرس ایمیل",url:"URL",emoji:"ایموجی",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"تاریخ و زمان ایزو",date:"تاریخ ایزو",time:"زمان ایزو",duration:"مدت زمان ایزو",ipv4:"IPv4 آدرس",ipv6:"IPv6 آدرس",cidrv4:"IPv4 دامنه",cidrv6:"IPv6 دامنه",base64:"base64-encoded رشته",base64url:"base64url-encoded رشته",json_string:"JSON رشته",e164:"E.164 عدد",jwt:"JWT",template_literal:"ورودی"};return c=>{switch(c.code){case"invalid_type":return`ورودی نامعتبر: می‌بایست ${c.expected} می‌بود، ${o(c.input)} دریافت شد`;case"invalid_value":return c.values.length===1?`ورودی نامعتبر: می‌بایست ${es(c.values[0])} می‌بود`:`گزینه نامعتبر: می‌بایست یکی از ${ti(c.values,"|")} می‌بود`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`خیلی بزرگ: ${c.origin??"مقدار"} باید ${d}${c.maximum.toString()} ${h.unit??"عنصر"} باشد`:`خیلی بزرگ: ${c.origin??"مقدار"} باید ${d}${c.maximum.toString()} باشد`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`خیلی کوچک: ${c.origin} باید ${d}${c.minimum.toString()} ${h.unit} باشد`:`خیلی کوچک: ${c.origin} باید ${d}${c.minimum.toString()} باشد`}case"invalid_format":{const d=c;return d.format==="starts_with"?`رشته نامعتبر: باید با "${d.prefix}" شروع شود`:d.format==="ends_with"?`رشته نامعتبر: باید با "${d.suffix}" تمام شود`:d.format==="includes"?`رشته نامعتبر: باید شامل "${d.includes}" باشد`:d.format==="regex"?`رشته نامعتبر: باید با الگوی ${d.pattern} مطابقت داشته باشد`:`${s[d.format]??c.format} نامعتبر`}case"not_multiple_of":return`عدد نامعتبر: باید مضرب ${c.divisor} باشد`;case"unrecognized_keys":return`کلید${c.keys.length>1?"های":""} ناشناس: ${ti(c.keys,", ")}`;case"invalid_key":return`کلید ناشناس در ${c.origin}`;case"invalid_union":return"ورودی نامعتبر";case"invalid_element":return`مقدار نامعتبر در ${c.origin}`;default:return"ورودی نامعتبر"}}};function Xar(){return{localeError:Yar()}}const Jar=()=>{const n={string:{unit:"merkkiä",subject:"merkkijonon"},file:{unit:"tavua",subject:"tiedoston"},array:{unit:"alkiota",subject:"listan"},set:{unit:"alkiota",subject:"joukon"},number:{unit:"",subject:"luvun"},bigint:{unit:"",subject:"suuren kokonaisluvun"},int:{unit:"",subject:"kokonaisluvun"},date:{unit:"",subject:"päivämäärän"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"säännöllinen lauseke",email:"sähköpostiosoite",url:"URL-osoite",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO-aikaleima",date:"ISO-päivämäärä",time:"ISO-aika",duration:"ISO-kesto",ipv4:"IPv4-osoite",ipv6:"IPv6-osoite",cidrv4:"IPv4-alue",cidrv6:"IPv6-alue",base64:"base64-koodattu merkkijono",base64url:"base64url-koodattu merkkijono",json_string:"JSON-merkkijono",e164:"E.164-luku",jwt:"JWT",template_literal:"templaattimerkkijono"};return c=>{switch(c.code){case"invalid_type":return`Virheellinen tyyppi: odotettiin ${c.expected}, oli ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Virheellinen syöte: täytyy olla ${es(c.values[0])}`:`Virheellinen valinta: täytyy olla yksi seuraavista: ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Liian suuri: ${h.subject} täytyy olla ${d}${c.maximum.toString()} ${h.unit}`.trim():`Liian suuri: arvon täytyy olla ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Liian pieni: ${h.subject} täytyy olla ${d}${c.minimum.toString()} ${h.unit}`.trim():`Liian pieni: arvon täytyy olla ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Virheellinen syöte: täytyy alkaa "${d.prefix}"`:d.format==="ends_with"?`Virheellinen syöte: täytyy loppua "${d.suffix}"`:d.format==="includes"?`Virheellinen syöte: täytyy sisältää "${d.includes}"`:d.format==="regex"?`Virheellinen syöte: täytyy vastata säännöllistä lauseketta ${d.pattern}`:`Virheellinen ${s[d.format]??c.format}`}case"not_multiple_of":return`Virheellinen luku: täytyy olla luvun ${c.divisor} monikerta`;case"unrecognized_keys":return`${c.keys.length>1?"Tuntemattomat avaimet":"Tuntematon avain"}: ${ti(c.keys,", ")}`;case"invalid_key":return"Virheellinen avain tietueessa";case"invalid_union":return"Virheellinen unioni";case"invalid_element":return"Virheellinen arvo joukossa";default:return"Virheellinen syöte"}}};function Zar(){return{localeError:Jar()}}const Qar=()=>{const n={string:{unit:"caractères",verb:"avoir"},file:{unit:"octets",verb:"avoir"},array:{unit:"éléments",verb:"avoir"},set:{unit:"éléments",verb:"avoir"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"nombre";case"object":{if(Array.isArray(c))return"tableau";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"entrée",email:"adresse e-mail",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"date et heure ISO",date:"date ISO",time:"heure ISO",duration:"durée ISO",ipv4:"adresse IPv4",ipv6:"adresse IPv6",cidrv4:"plage IPv4",cidrv6:"plage IPv6",base64:"chaîne encodée en base64",base64url:"chaîne encodée en base64url",json_string:"chaîne JSON",e164:"numéro E.164",jwt:"JWT",template_literal:"entrée"};return c=>{switch(c.code){case"invalid_type":return`Entrée invalide : ${c.expected} attendu, ${o(c.input)} reçu`;case"invalid_value":return c.values.length===1?`Entrée invalide : ${es(c.values[0])} attendu`:`Option invalide : une valeur parmi ${ti(c.values,"|")} attendue`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Trop grand : ${c.origin??"valeur"} doit ${h.verb} ${d}${c.maximum.toString()} ${h.unit??"élément(s)"}`:`Trop grand : ${c.origin??"valeur"} doit être ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Trop petit : ${c.origin} doit ${h.verb} ${d}${c.minimum.toString()} ${h.unit}`:`Trop petit : ${c.origin} doit être ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Chaîne invalide : doit commencer par "${d.prefix}"`:d.format==="ends_with"?`Chaîne invalide : doit se terminer par "${d.suffix}"`:d.format==="includes"?`Chaîne invalide : doit inclure "${d.includes}"`:d.format==="regex"?`Chaîne invalide : doit correspondre au modèle ${d.pattern}`:`${s[d.format]??c.format} invalide`}case"not_multiple_of":return`Nombre invalide : doit être un multiple de ${c.divisor}`;case"unrecognized_keys":return`Clé${c.keys.length>1?"s":""} non reconnue${c.keys.length>1?"s":""} : ${ti(c.keys,", ")}`;case"invalid_key":return`Clé invalide dans ${c.origin}`;case"invalid_union":return"Entrée invalide";case"invalid_element":return`Valeur invalide dans ${c.origin}`;default:return"Entrée invalide"}}};function esr(){return{localeError:Qar()}}const tsr=()=>{const n={string:{unit:"caractères",verb:"avoir"},file:{unit:"octets",verb:"avoir"},array:{unit:"éléments",verb:"avoir"},set:{unit:"éléments",verb:"avoir"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"entrée",email:"adresse courriel",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"date-heure ISO",date:"date ISO",time:"heure ISO",duration:"durée ISO",ipv4:"adresse IPv4",ipv6:"adresse IPv6",cidrv4:"plage IPv4",cidrv6:"plage IPv6",base64:"chaîne encodée en base64",base64url:"chaîne encodée en base64url",json_string:"chaîne JSON",e164:"numéro E.164",jwt:"JWT",template_literal:"entrée"};return c=>{switch(c.code){case"invalid_type":return`Entrée invalide : attendu ${c.expected}, reçu ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Entrée invalide : attendu ${es(c.values[0])}`:`Option invalide : attendu l'une des valeurs suivantes ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"≤":"<",h=r(c.origin);return h?`Trop grand : attendu que ${c.origin??"la valeur"} ait ${d}${c.maximum.toString()} ${h.unit}`:`Trop grand : attendu que ${c.origin??"la valeur"} soit ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?"≥":">",h=r(c.origin);return h?`Trop petit : attendu que ${c.origin} ait ${d}${c.minimum.toString()} ${h.unit}`:`Trop petit : attendu que ${c.origin} soit ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Chaîne invalide : doit commencer par "${d.prefix}"`:d.format==="ends_with"?`Chaîne invalide : doit se terminer par "${d.suffix}"`:d.format==="includes"?`Chaîne invalide : doit inclure "${d.includes}"`:d.format==="regex"?`Chaîne invalide : doit correspondre au motif ${d.pattern}`:`${s[d.format]??c.format} invalide`}case"not_multiple_of":return`Nombre invalide : doit être un multiple de ${c.divisor}`;case"unrecognized_keys":return`Clé${c.keys.length>1?"s":""} non reconnue${c.keys.length>1?"s":""} : ${ti(c.keys,", ")}`;case"invalid_key":return`Clé invalide dans ${c.origin}`;case"invalid_union":return"Entrée invalide";case"invalid_element":return`Valeur invalide dans ${c.origin}`;default:return"Entrée invalide"}}};function nsr(){return{localeError:tsr()}}const rsr=()=>{const n={string:{unit:"אותיות",verb:"לכלול"},file:{unit:"בייטים",verb:"לכלול"},array:{unit:"פריטים",verb:"לכלול"},set:{unit:"פריטים",verb:"לכלול"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"קלט",email:"כתובת אימייל",url:"כתובת רשת",emoji:"אימוג'י",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"תאריך וזמן ISO",date:"תאריך ISO",time:"זמן ISO",duration:"משך זמן ISO",ipv4:"כתובת IPv4",ipv6:"כתובת IPv6",cidrv4:"טווח IPv4",cidrv6:"טווח IPv6",base64:"מחרוזת בבסיס 64",base64url:"מחרוזת בבסיס 64 לכתובות רשת",json_string:"מחרוזת JSON",e164:"מספר E.164",jwt:"JWT",template_literal:"קלט"};return c=>{switch(c.code){case"invalid_type":return`קלט לא תקין: צריך ${c.expected}, התקבל ${o(c.input)}`;case"invalid_value":return c.values.length===1?`קלט לא תקין: צריך ${es(c.values[0])}`:`קלט לא תקין: צריך אחת מהאפשרויות  ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`גדול מדי: ${c.origin??"value"} צריך להיות ${d}${c.maximum.toString()} ${h.unit??"elements"}`:`גדול מדי: ${c.origin??"value"} צריך להיות ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`קטן מדי: ${c.origin} צריך להיות ${d}${c.minimum.toString()} ${h.unit}`:`קטן מדי: ${c.origin} צריך להיות ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`מחרוזת לא תקינה: חייבת להתחיל ב"${d.prefix}"`:d.format==="ends_with"?`מחרוזת לא תקינה: חייבת להסתיים ב "${d.suffix}"`:d.format==="includes"?`מחרוזת לא תקינה: חייבת לכלול "${d.includes}"`:d.format==="regex"?`מחרוזת לא תקינה: חייבת להתאים לתבנית ${d.pattern}`:`${s[d.format]??c.format} לא תקין`}case"not_multiple_of":return`מספר לא תקין: חייב להיות מכפלה של ${c.divisor}`;case"unrecognized_keys":return`מפתח${c.keys.length>1?"ות":""} לא מזוה${c.keys.length>1?"ים":"ה"}: ${ti(c.keys,", ")}`;case"invalid_key":return`מפתח לא תקין ב${c.origin}`;case"invalid_union":return"קלט לא תקין";case"invalid_element":return`ערך לא תקין ב${c.origin}`;default:return"קלט לא תקין"}}};function isr(){return{localeError:rsr()}}const osr=()=>{const n={string:{unit:"karakter",verb:"legyen"},file:{unit:"byte",verb:"legyen"},array:{unit:"elem",verb:"legyen"},set:{unit:"elem",verb:"legyen"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"szám";case"object":{if(Array.isArray(c))return"tömb";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"bemenet",email:"email cím",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO időbélyeg",date:"ISO dátum",time:"ISO idő",duration:"ISO időintervallum",ipv4:"IPv4 cím",ipv6:"IPv6 cím",cidrv4:"IPv4 tartomány",cidrv6:"IPv6 tartomány",base64:"base64-kódolt string",base64url:"base64url-kódolt string",json_string:"JSON string",e164:"E.164 szám",jwt:"JWT",template_literal:"bemenet"};return c=>{switch(c.code){case"invalid_type":return`Érvénytelen bemenet: a várt érték ${c.expected}, a kapott érték ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Érvénytelen bemenet: a várt érték ${es(c.values[0])}`:`Érvénytelen opció: valamelyik érték várt ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Túl nagy: ${c.origin??"érték"} mérete túl nagy ${d}${c.maximum.toString()} ${h.unit??"elem"}`:`Túl nagy: a bemeneti érték ${c.origin??"érték"} túl nagy: ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Túl kicsi: a bemeneti érték ${c.origin} mérete túl kicsi ${d}${c.minimum.toString()} ${h.unit}`:`Túl kicsi: a bemeneti érték ${c.origin} túl kicsi ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Érvénytelen string: "${d.prefix}" értékkel kell kezdődnie`:d.format==="ends_with"?`Érvénytelen string: "${d.suffix}" értékkel kell végződnie`:d.format==="includes"?`Érvénytelen string: "${d.includes}" értéket kell tartalmaznia`:d.format==="regex"?`Érvénytelen string: ${d.pattern} mintának kell megfelelnie`:`Érvénytelen ${s[d.format]??c.format}`}case"not_multiple_of":return`Érvénytelen szám: ${c.divisor} többszörösének kell lennie`;case"unrecognized_keys":return`Ismeretlen kulcs${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Érvénytelen kulcs ${c.origin}`;case"invalid_union":return"Érvénytelen bemenet";case"invalid_element":return`Érvénytelen érték: ${c.origin}`;default:return"Érvénytelen bemenet"}}};function asr(){return{localeError:osr()}}const ssr=()=>{const n={string:{unit:"karakter",verb:"memiliki"},file:{unit:"byte",verb:"memiliki"},array:{unit:"item",verb:"memiliki"},set:{unit:"item",verb:"memiliki"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"input",email:"alamat email",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"tanggal dan waktu format ISO",date:"tanggal format ISO",time:"jam format ISO",duration:"durasi format ISO",ipv4:"alamat IPv4",ipv6:"alamat IPv6",cidrv4:"rentang alamat IPv4",cidrv6:"rentang alamat IPv6",base64:"string dengan enkode base64",base64url:"string dengan enkode base64url",json_string:"string JSON",e164:"angka E.164",jwt:"JWT",template_literal:"input"};return c=>{switch(c.code){case"invalid_type":return`Input tidak valid: diharapkan ${c.expected}, diterima ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Input tidak valid: diharapkan ${es(c.values[0])}`:`Pilihan tidak valid: diharapkan salah satu dari ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Terlalu besar: diharapkan ${c.origin??"value"} memiliki ${d}${c.maximum.toString()} ${h.unit??"elemen"}`:`Terlalu besar: diharapkan ${c.origin??"value"} menjadi ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Terlalu kecil: diharapkan ${c.origin} memiliki ${d}${c.minimum.toString()} ${h.unit}`:`Terlalu kecil: diharapkan ${c.origin} menjadi ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`String tidak valid: harus dimulai dengan "${d.prefix}"`:d.format==="ends_with"?`String tidak valid: harus berakhir dengan "${d.suffix}"`:d.format==="includes"?`String tidak valid: harus menyertakan "${d.includes}"`:d.format==="regex"?`String tidak valid: harus sesuai pola ${d.pattern}`:`${s[d.format]??c.format} tidak valid`}case"not_multiple_of":return`Angka tidak valid: harus kelipatan dari ${c.divisor}`;case"unrecognized_keys":return`Kunci tidak dikenali ${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Kunci tidak valid di ${c.origin}`;case"invalid_union":return"Input tidak valid";case"invalid_element":return`Nilai tidak valid di ${c.origin}`;default:return"Input tidak valid"}}};function csr(){return{localeError:ssr()}}const usr=n=>{const r=typeof n;switch(r){case"number":return Number.isNaN(n)?"NaN":"númer";case"object":{if(Array.isArray(n))return"fylki";if(n===null)return"null";if(Object.getPrototypeOf(n)!==Object.prototype&&n.constructor)return n.constructor.name}}return r},lsr=()=>{const n={string:{unit:"stafi",verb:"að hafa"},file:{unit:"bæti",verb:"að hafa"},array:{unit:"hluti",verb:"að hafa"},set:{unit:"hluti",verb:"að hafa"}};function r(s){return n[s]??null}const o={regex:"gildi",email:"netfang",url:"vefslóð",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO dagsetning og tími",date:"ISO dagsetning",time:"ISO tími",duration:"ISO tímalengd",ipv4:"IPv4 address",ipv6:"IPv6 address",cidrv4:"IPv4 range",cidrv6:"IPv6 range",base64:"base64-encoded strengur",base64url:"base64url-encoded strengur",json_string:"JSON strengur",e164:"E.164 tölugildi",jwt:"JWT",template_literal:"gildi"};return s=>{switch(s.code){case"invalid_type":return`Rangt gildi: Þú slóst inn ${usr(s.input)} þar sem á að vera ${s.expected}`;case"invalid_value":return s.values.length===1?`Rangt gildi: gert ráð fyrir ${es(s.values[0])}`:`Ógilt val: má vera eitt af eftirfarandi ${ti(s.values,"|")}`;case"too_big":{const c=s.inclusive?"<=":"<",d=r(s.origin);return d?`Of stórt: gert er ráð fyrir að ${s.origin??"gildi"} hafi ${c}${s.maximum.toString()} ${d.unit??"hluti"}`:`Of stórt: gert er ráð fyrir að ${s.origin??"gildi"} sé ${c}${s.maximum.toString()}`}case"too_small":{const c=s.inclusive?">=":">",d=r(s.origin);return d?`Of lítið: gert er ráð fyrir að ${s.origin} hafi ${c}${s.minimum.toString()} ${d.unit}`:`Of lítið: gert er ráð fyrir að ${s.origin} sé ${c}${s.minimum.toString()}`}case"invalid_format":{const c=s;return c.format==="starts_with"?`Ógildur strengur: verður að byrja á "${c.prefix}"`:c.format==="ends_with"?`Ógildur strengur: verður að enda á "${c.suffix}"`:c.format==="includes"?`Ógildur strengur: verður að innihalda "${c.includes}"`:c.format==="regex"?`Ógildur strengur: verður að fylgja mynstri ${c.pattern}`:`Rangt ${o[c.format]??s.format}`}case"not_multiple_of":return`Röng tala: verður að vera margfeldi af ${s.divisor}`;case"unrecognized_keys":return`Óþekkt ${s.keys.length>1?"ir lyklar":"ur lykill"}: ${ti(s.keys,", ")}`;case"invalid_key":return`Rangur lykill í ${s.origin}`;case"invalid_union":return"Rangt gildi";case"invalid_element":return`Rangt gildi í ${s.origin}`;default:return"Rangt gildi"}}};function dsr(){return{localeError:lsr()}}const fsr=()=>{const n={string:{unit:"caratteri",verb:"avere"},file:{unit:"byte",verb:"avere"},array:{unit:"elementi",verb:"avere"},set:{unit:"elementi",verb:"avere"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"numero";case"object":{if(Array.isArray(c))return"vettore";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"input",email:"indirizzo email",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"data e ora ISO",date:"data ISO",time:"ora ISO",duration:"durata ISO",ipv4:"indirizzo IPv4",ipv6:"indirizzo IPv6",cidrv4:"intervallo IPv4",cidrv6:"intervallo IPv6",base64:"stringa codificata in base64",base64url:"URL codificata in base64",json_string:"stringa JSON",e164:"numero E.164",jwt:"JWT",template_literal:"input"};return c=>{switch(c.code){case"invalid_type":return`Input non valido: atteso ${c.expected}, ricevuto ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Input non valido: atteso ${es(c.values[0])}`:`Opzione non valida: atteso uno tra ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Troppo grande: ${c.origin??"valore"} deve avere ${d}${c.maximum.toString()} ${h.unit??"elementi"}`:`Troppo grande: ${c.origin??"valore"} deve essere ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Troppo piccolo: ${c.origin} deve avere ${d}${c.minimum.toString()} ${h.unit}`:`Troppo piccolo: ${c.origin} deve essere ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Stringa non valida: deve iniziare con "${d.prefix}"`:d.format==="ends_with"?`Stringa non valida: deve terminare con "${d.suffix}"`:d.format==="includes"?`Stringa non valida: deve includere "${d.includes}"`:d.format==="regex"?`Stringa non valida: deve corrispondere al pattern ${d.pattern}`:`Invalid ${s[d.format]??c.format}`}case"not_multiple_of":return`Numero non valido: deve essere un multiplo di ${c.divisor}`;case"unrecognized_keys":return`Chiav${c.keys.length>1?"i":"e"} non riconosciut${c.keys.length>1?"e":"a"}: ${ti(c.keys,", ")}`;case"invalid_key":return`Chiave non valida in ${c.origin}`;case"invalid_union":return"Input non valido";case"invalid_element":return`Valore non valido in ${c.origin}`;default:return"Input non valido"}}};function hsr(){return{localeError:fsr()}}const psr=()=>{const n={string:{unit:"文字",verb:"である"},file:{unit:"バイト",verb:"である"},array:{unit:"要素",verb:"である"},set:{unit:"要素",verb:"である"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"数値";case"object":{if(Array.isArray(c))return"配列";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"入力値",email:"メールアドレス",url:"URL",emoji:"絵文字",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO日時",date:"ISO日付",time:"ISO時刻",duration:"ISO期間",ipv4:"IPv4アドレス",ipv6:"IPv6アドレス",cidrv4:"IPv4範囲",cidrv6:"IPv6範囲",base64:"base64エンコード文字列",base64url:"base64urlエンコード文字列",json_string:"JSON文字列",e164:"E.164番号",jwt:"JWT",template_literal:"入力値"};return c=>{switch(c.code){case"invalid_type":return`無効な入力: ${c.expected}が期待されましたが、${o(c.input)}が入力されました`;case"invalid_value":return c.values.length===1?`無効な入力: ${es(c.values[0])}が期待されました`:`無効な選択: ${ti(c.values,"、")}のいずれかである必要があります`;case"too_big":{const d=c.inclusive?"以下である":"より小さい",h=r(c.origin);return h?`大きすぎる値: ${c.origin??"値"}は${c.maximum.toString()}${h.unit??"要素"}${d}必要があります`:`大きすぎる値: ${c.origin??"値"}は${c.maximum.toString()}${d}必要があります`}case"too_small":{const d=c.inclusive?"以上である":"より大きい",h=r(c.origin);return h?`小さすぎる値: ${c.origin}は${c.minimum.toString()}${h.unit}${d}必要があります`:`小さすぎる値: ${c.origin}は${c.minimum.toString()}${d}必要があります`}case"invalid_format":{const d=c;return d.format==="starts_with"?`無効な文字列: "${d.prefix}"で始まる必要があります`:d.format==="ends_with"?`無効な文字列: "${d.suffix}"で終わる必要があります`:d.format==="includes"?`無効な文字列: "${d.includes}"を含む必要があります`:d.format==="regex"?`無効な文字列: パターン${d.pattern}に一致する必要があります`:`無効な${s[d.format]??c.format}`}case"not_multiple_of":return`無効な数値: ${c.divisor}の倍数である必要があります`;case"unrecognized_keys":return`認識されていないキー${c.keys.length>1?"群":""}: ${ti(c.keys,"、")}`;case"invalid_key":return`${c.origin}内の無効なキー`;case"invalid_union":return"無効な入力";case"invalid_element":return`${c.origin}内の無効な値`;default:return"無効な入力"}}};function gsr(){return{localeError:psr()}}const msr=()=>{const n={string:{unit:"តួអក្សរ",verb:"គួរមាន"},file:{unit:"បៃ",verb:"គួរមាន"},array:{unit:"ធាតុ",verb:"គួរមាន"},set:{unit:"ធាតុ",verb:"គួរមាន"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"មិនមែនជាលេខ (NaN)":"លេខ";case"object":{if(Array.isArray(c))return"អារេ (Array)";if(c===null)return"គ្មានតម្លៃ (null)";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"ទិន្នន័យបញ្ចូល",email:"អាសយដ្ឋានអ៊ីមែល",url:"URL",emoji:"សញ្ញាអារម្មណ៍",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"កាលបរិច្ឆេទ និងម៉ោង ISO",date:"កាលបរិច្ឆេទ ISO",time:"ម៉ោង ISO",duration:"រយៈពេល ISO",ipv4:"អាសយដ្ឋាន IPv4",ipv6:"អាសយដ្ឋាន IPv6",cidrv4:"ដែនអាសយដ្ឋាន IPv4",cidrv6:"ដែនអាសយដ្ឋាន IPv6",base64:"ខ្សែអក្សរអ៊ិកូដ base64",base64url:"ខ្សែអក្សរអ៊ិកូដ base64url",json_string:"ខ្សែអក្សរ JSON",e164:"លេខ E.164",jwt:"JWT",template_literal:"ទិន្នន័យបញ្ចូល"};return c=>{switch(c.code){case"invalid_type":return`ទិន្នន័យបញ្ចូលមិនត្រឹមត្រូវ៖ ត្រូវការ ${c.expected} ប៉ុន្តែទទួលបាន ${o(c.input)}`;case"invalid_value":return c.values.length===1?`ទិន្នន័យបញ្ចូលមិនត្រឹមត្រូវ៖ ត្រូវការ ${es(c.values[0])}`:`ជម្រើសមិនត្រឹមត្រូវ៖ ត្រូវជាមួយក្នុងចំណោម ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`ធំពេក៖ ត្រូវការ ${c.origin??"តម្លៃ"} ${d} ${c.maximum.toString()} ${h.unit??"ធាតុ"}`:`ធំពេក៖ ត្រូវការ ${c.origin??"តម្លៃ"} ${d} ${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`តូចពេក៖ ត្រូវការ ${c.origin} ${d} ${c.minimum.toString()} ${h.unit}`:`តូចពេក៖ ត្រូវការ ${c.origin} ${d} ${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`ខ្សែអក្សរមិនត្រឹមត្រូវ៖ ត្រូវចាប់ផ្តើមដោយ "${d.prefix}"`:d.format==="ends_with"?`ខ្សែអក្សរមិនត្រឹមត្រូវ៖ ត្រូវបញ្ចប់ដោយ "${d.suffix}"`:d.format==="includes"?`ខ្សែអក្សរមិនត្រឹមត្រូវ៖ ត្រូវមាន "${d.includes}"`:d.format==="regex"?`ខ្សែអក្សរមិនត្រឹមត្រូវ៖ ត្រូវតែផ្គូផ្គងនឹងទម្រង់ដែលបានកំណត់ ${d.pattern}`:`មិនត្រឹមត្រូវ៖ ${s[d.format]??c.format}`}case"not_multiple_of":return`លេខមិនត្រឹមត្រូវ៖ ត្រូវតែជាពហុគុណនៃ ${c.divisor}`;case"unrecognized_keys":return`រកឃើញសោមិនស្គាល់៖ ${ti(c.keys,", ")}`;case"invalid_key":return`សោមិនត្រឹមត្រូវនៅក្នុង ${c.origin}`;case"invalid_union":return"ទិន្នន័យមិនត្រឹមត្រូវ";case"invalid_element":return`ទិន្នន័យមិនត្រឹមត្រូវនៅក្នុង ${c.origin}`;default:return"ទិន្នន័យមិនត្រឹមត្រូវ"}}};function bsr(){return{localeError:msr()}}const vsr=()=>{const n={string:{unit:"문자",verb:"to have"},file:{unit:"바이트",verb:"to have"},array:{unit:"개",verb:"to have"},set:{unit:"개",verb:"to have"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"입력",email:"이메일 주소",url:"URL",emoji:"이모지",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO 날짜시간",date:"ISO 날짜",time:"ISO 시간",duration:"ISO 기간",ipv4:"IPv4 주소",ipv6:"IPv6 주소",cidrv4:"IPv4 범위",cidrv6:"IPv6 범위",base64:"base64 인코딩 문자열",base64url:"base64url 인코딩 문자열",json_string:"JSON 문자열",e164:"E.164 번호",jwt:"JWT",template_literal:"입력"};return c=>{switch(c.code){case"invalid_type":return`잘못된 입력: 예상 타입은 ${c.expected}, 받은 타입은 ${o(c.input)}입니다`;case"invalid_value":return c.values.length===1?`잘못된 입력: 값은 ${es(c.values[0])} 이어야 합니다`:`잘못된 옵션: ${ti(c.values,"또는 ")} 중 하나여야 합니다`;case"too_big":{const d=c.inclusive?"이하":"미만",h=d==="미만"?"이어야 합니다":"여야 합니다",b=r(c.origin),y=b?.unit??"요소";return b?`${c.origin??"값"}이 너무 큽니다: ${c.maximum.toString()}${y} ${d}${h}`:`${c.origin??"값"}이 너무 큽니다: ${c.maximum.toString()} ${d}${h}`}case"too_small":{const d=c.inclusive?"이상":"초과",h=d==="이상"?"이어야 합니다":"여야 합니다",b=r(c.origin),y=b?.unit??"요소";return b?`${c.origin??"값"}이 너무 작습니다: ${c.minimum.toString()}${y} ${d}${h}`:`${c.origin??"값"}이 너무 작습니다: ${c.minimum.toString()} ${d}${h}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`잘못된 문자열: "${d.prefix}"(으)로 시작해야 합니다`:d.format==="ends_with"?`잘못된 문자열: "${d.suffix}"(으)로 끝나야 합니다`:d.format==="includes"?`잘못된 문자열: "${d.includes}"을(를) 포함해야 합니다`:d.format==="regex"?`잘못된 문자열: 정규식 ${d.pattern} 패턴과 일치해야 합니다`:`잘못된 ${s[d.format]??c.format}`}case"not_multiple_of":return`잘못된 숫자: ${c.divisor}의 배수여야 합니다`;case"unrecognized_keys":return`인식할 수 없는 키: ${ti(c.keys,", ")}`;case"invalid_key":return`잘못된 키: ${c.origin}`;case"invalid_union":return"잘못된 입력";case"invalid_element":return`잘못된 값: ${c.origin}`;default:return"잘못된 입력"}}};function wsr(){return{localeError:vsr()}}const ysr=()=>{const n={string:{unit:"знаци",verb:"да имаат"},file:{unit:"бајти",verb:"да имаат"},array:{unit:"ставки",verb:"да имаат"},set:{unit:"ставки",verb:"да имаат"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"број";case"object":{if(Array.isArray(c))return"низа";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"внес",email:"адреса на е-пошта",url:"URL",emoji:"емоџи",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO датум и време",date:"ISO датум",time:"ISO време",duration:"ISO времетраење",ipv4:"IPv4 адреса",ipv6:"IPv6 адреса",cidrv4:"IPv4 опсег",cidrv6:"IPv6 опсег",base64:"base64-енкодирана низа",base64url:"base64url-енкодирана низа",json_string:"JSON низа",e164:"E.164 број",jwt:"JWT",template_literal:"внес"};return c=>{switch(c.code){case"invalid_type":return`Грешен внес: се очекува ${c.expected}, примено ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Invalid input: expected ${es(c.values[0])}`:`Грешана опција: се очекува една ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Премногу голем: се очекува ${c.origin??"вредноста"} да има ${d}${c.maximum.toString()} ${h.unit??"елементи"}`:`Премногу голем: се очекува ${c.origin??"вредноста"} да биде ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Премногу мал: се очекува ${c.origin} да има ${d}${c.minimum.toString()} ${h.unit}`:`Премногу мал: се очекува ${c.origin} да биде ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Неважечка низа: мора да започнува со "${d.prefix}"`:d.format==="ends_with"?`Неважечка низа: мора да завршува со "${d.suffix}"`:d.format==="includes"?`Неважечка низа: мора да вклучува "${d.includes}"`:d.format==="regex"?`Неважечка низа: мора да одгоара на патернот ${d.pattern}`:`Invalid ${s[d.format]??c.format}`}case"not_multiple_of":return`Грешен број: мора да биде делив со ${c.divisor}`;case"unrecognized_keys":return`${c.keys.length>1?"Непрепознаени клучеви":"Непрепознаен клуч"}: ${ti(c.keys,", ")}`;case"invalid_key":return`Грешен клуч во ${c.origin}`;case"invalid_union":return"Грешен внес";case"invalid_element":return`Грешна вредност во ${c.origin}`;default:return"Грешен внес"}}};function Esr(){return{localeError:ysr()}}const Ssr=()=>{const n={string:{unit:"aksara",verb:"mempunyai"},file:{unit:"bait",verb:"mempunyai"},array:{unit:"elemen",verb:"mempunyai"},set:{unit:"elemen",verb:"mempunyai"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"nombor";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"input",email:"alamat e-mel",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"tarikh masa ISO",date:"tarikh ISO",time:"masa ISO",duration:"tempoh ISO",ipv4:"alamat IPv4",ipv6:"alamat IPv6",cidrv4:"julat IPv4",cidrv6:"julat IPv6",base64:"string dikodkan base64",base64url:"string dikodkan base64url",json_string:"string JSON",e164:"nombor E.164",jwt:"JWT",template_literal:"input"};return c=>{switch(c.code){case"invalid_type":return`Input tidak sah: dijangka ${c.expected}, diterima ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Input tidak sah: dijangka ${es(c.values[0])}`:`Pilihan tidak sah: dijangka salah satu daripada ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Terlalu besar: dijangka ${c.origin??"nilai"} ${h.verb} ${d}${c.maximum.toString()} ${h.unit??"elemen"}`:`Terlalu besar: dijangka ${c.origin??"nilai"} adalah ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Terlalu kecil: dijangka ${c.origin} ${h.verb} ${d}${c.minimum.toString()} ${h.unit}`:`Terlalu kecil: dijangka ${c.origin} adalah ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`String tidak sah: mesti bermula dengan "${d.prefix}"`:d.format==="ends_with"?`String tidak sah: mesti berakhir dengan "${d.suffix}"`:d.format==="includes"?`String tidak sah: mesti mengandungi "${d.includes}"`:d.format==="regex"?`String tidak sah: mesti sepadan dengan corak ${d.pattern}`:`${s[d.format]??c.format} tidak sah`}case"not_multiple_of":return`Nombor tidak sah: perlu gandaan ${c.divisor}`;case"unrecognized_keys":return`Kunci tidak dikenali: ${ti(c.keys,", ")}`;case"invalid_key":return`Kunci tidak sah dalam ${c.origin}`;case"invalid_union":return"Input tidak sah";case"invalid_element":return`Nilai tidak sah dalam ${c.origin}`;default:return"Input tidak sah"}}};function _sr(){return{localeError:Ssr()}}const ksr=()=>{const n={string:{unit:"tekens"},file:{unit:"bytes"},array:{unit:"elementen"},set:{unit:"elementen"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"getal";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"invoer",email:"emailadres",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO datum en tijd",date:"ISO datum",time:"ISO tijd",duration:"ISO duur",ipv4:"IPv4-adres",ipv6:"IPv6-adres",cidrv4:"IPv4-bereik",cidrv6:"IPv6-bereik",base64:"base64-gecodeerde tekst",base64url:"base64 URL-gecodeerde tekst",json_string:"JSON string",e164:"E.164-nummer",jwt:"JWT",template_literal:"invoer"};return c=>{switch(c.code){case"invalid_type":return`Ongeldige invoer: verwacht ${c.expected}, ontving ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Ongeldige invoer: verwacht ${es(c.values[0])}`:`Ongeldige optie: verwacht één van ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Te lang: verwacht dat ${c.origin??"waarde"} ${d}${c.maximum.toString()} ${h.unit??"elementen"} bevat`:`Te lang: verwacht dat ${c.origin??"waarde"} ${d}${c.maximum.toString()} is`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Te kort: verwacht dat ${c.origin} ${d}${c.minimum.toString()} ${h.unit} bevat`:`Te kort: verwacht dat ${c.origin} ${d}${c.minimum.toString()} is`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Ongeldige tekst: moet met "${d.prefix}" beginnen`:d.format==="ends_with"?`Ongeldige tekst: moet op "${d.suffix}" eindigen`:d.format==="includes"?`Ongeldige tekst: moet "${d.includes}" bevatten`:d.format==="regex"?`Ongeldige tekst: moet overeenkomen met patroon ${d.pattern}`:`Ongeldig: ${s[d.format]??c.format}`}case"not_multiple_of":return`Ongeldig getal: moet een veelvoud van ${c.divisor} zijn`;case"unrecognized_keys":return`Onbekende key${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Ongeldige key in ${c.origin}`;case"invalid_union":return"Ongeldige invoer";case"invalid_element":return`Ongeldige waarde in ${c.origin}`;default:return"Ongeldige invoer"}}};function Tsr(){return{localeError:ksr()}}const xsr=()=>{const n={string:{unit:"tegn",verb:"å ha"},file:{unit:"bytes",verb:"å ha"},array:{unit:"elementer",verb:"å inneholde"},set:{unit:"elementer",verb:"å inneholde"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"tall";case"object":{if(Array.isArray(c))return"liste";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"input",email:"e-postadresse",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO dato- og klokkeslett",date:"ISO-dato",time:"ISO-klokkeslett",duration:"ISO-varighet",ipv4:"IPv4-område",ipv6:"IPv6-område",cidrv4:"IPv4-spekter",cidrv6:"IPv6-spekter",base64:"base64-enkodet streng",base64url:"base64url-enkodet streng",json_string:"JSON-streng",e164:"E.164-nummer",jwt:"JWT",template_literal:"input"};return c=>{switch(c.code){case"invalid_type":return`Ugyldig input: forventet ${c.expected}, fikk ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Ugyldig verdi: forventet ${es(c.values[0])}`:`Ugyldig valg: forventet en av ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`For stor(t): forventet ${c.origin??"value"} til å ha ${d}${c.maximum.toString()} ${h.unit??"elementer"}`:`For stor(t): forventet ${c.origin??"value"} til å ha ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`For lite(n): forventet ${c.origin} til å ha ${d}${c.minimum.toString()} ${h.unit}`:`For lite(n): forventet ${c.origin} til å ha ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Ugyldig streng: må starte med "${d.prefix}"`:d.format==="ends_with"?`Ugyldig streng: må ende med "${d.suffix}"`:d.format==="includes"?`Ugyldig streng: må inneholde "${d.includes}"`:d.format==="regex"?`Ugyldig streng: må matche mønsteret ${d.pattern}`:`Ugyldig ${s[d.format]??c.format}`}case"not_multiple_of":return`Ugyldig tall: må være et multiplum av ${c.divisor}`;case"unrecognized_keys":return`${c.keys.length>1?"Ukjente nøkler":"Ukjent nøkkel"}: ${ti(c.keys,", ")}`;case"invalid_key":return`Ugyldig nøkkel i ${c.origin}`;case"invalid_union":return"Ugyldig input";case"invalid_element":return`Ugyldig verdi i ${c.origin}`;default:return"Ugyldig input"}}};function Csr(){return{localeError:xsr()}}const Asr=()=>{const n={string:{unit:"harf",verb:"olmalıdır"},file:{unit:"bayt",verb:"olmalıdır"},array:{unit:"unsur",verb:"olmalıdır"},set:{unit:"unsur",verb:"olmalıdır"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"numara";case"object":{if(Array.isArray(c))return"saf";if(c===null)return"gayb";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"giren",email:"epostagâh",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO hengâmı",date:"ISO tarihi",time:"ISO zamanı",duration:"ISO müddeti",ipv4:"IPv4 nişânı",ipv6:"IPv6 nişânı",cidrv4:"IPv4 menzili",cidrv6:"IPv6 menzili",base64:"base64-şifreli metin",base64url:"base64url-şifreli metin",json_string:"JSON metin",e164:"E.164 sayısı",jwt:"JWT",template_literal:"giren"};return c=>{switch(c.code){case"invalid_type":return`Fâsit giren: umulan ${c.expected}, alınan ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Fâsit giren: umulan ${es(c.values[0])}`:`Fâsit tercih: mûteberler ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Fazla büyük: ${c.origin??"value"}, ${d}${c.maximum.toString()} ${h.unit??"elements"} sahip olmalıydı.`:`Fazla büyük: ${c.origin??"value"}, ${d}${c.maximum.toString()} olmalıydı.`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Fazla küçük: ${c.origin}, ${d}${c.minimum.toString()} ${h.unit} sahip olmalıydı.`:`Fazla küçük: ${c.origin}, ${d}${c.minimum.toString()} olmalıydı.`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Fâsit metin: "${d.prefix}" ile başlamalı.`:d.format==="ends_with"?`Fâsit metin: "${d.suffix}" ile bitmeli.`:d.format==="includes"?`Fâsit metin: "${d.includes}" ihtivâ etmeli.`:d.format==="regex"?`Fâsit metin: ${d.pattern} nakşına uymalı.`:`Fâsit ${s[d.format]??c.format}`}case"not_multiple_of":return`Fâsit sayı: ${c.divisor} katı olmalıydı.`;case"unrecognized_keys":return`Tanınmayan anahtar ${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`${c.origin} için tanınmayan anahtar var.`;case"invalid_union":return"Giren tanınamadı.";case"invalid_element":return`${c.origin} için tanınmayan kıymet var.`;default:return"Kıymet tanınamadı."}}};function Isr(){return{localeError:Asr()}}const Osr=()=>{const n={string:{unit:"توکي",verb:"ولري"},file:{unit:"بایټس",verb:"ولري"},array:{unit:"توکي",verb:"ولري"},set:{unit:"توکي",verb:"ولري"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"عدد";case"object":{if(Array.isArray(c))return"ارې";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"ورودي",email:"بریښنالیک",url:"یو آر ال",emoji:"ایموجي",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"نیټه او وخت",date:"نېټه",time:"وخت",duration:"موده",ipv4:"د IPv4 پته",ipv6:"د IPv6 پته",cidrv4:"د IPv4 ساحه",cidrv6:"د IPv6 ساحه",base64:"base64-encoded متن",base64url:"base64url-encoded متن",json_string:"JSON متن",e164:"د E.164 شمېره",jwt:"JWT",template_literal:"ورودي"};return c=>{switch(c.code){case"invalid_type":return`ناسم ورودي: باید ${c.expected} وای, مګر ${o(c.input)} ترلاسه شو`;case"invalid_value":return c.values.length===1?`ناسم ورودي: باید ${es(c.values[0])} وای`:`ناسم انتخاب: باید یو له ${ti(c.values,"|")} څخه وای`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`ډیر لوی: ${c.origin??"ارزښت"} باید ${d}${c.maximum.toString()} ${h.unit??"عنصرونه"} ولري`:`ډیر لوی: ${c.origin??"ارزښت"} باید ${d}${c.maximum.toString()} وي`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`ډیر کوچنی: ${c.origin} باید ${d}${c.minimum.toString()} ${h.unit} ولري`:`ډیر کوچنی: ${c.origin} باید ${d}${c.minimum.toString()} وي`}case"invalid_format":{const d=c;return d.format==="starts_with"?`ناسم متن: باید د "${d.prefix}" سره پیل شي`:d.format==="ends_with"?`ناسم متن: باید د "${d.suffix}" سره پای ته ورسيږي`:d.format==="includes"?`ناسم متن: باید "${d.includes}" ولري`:d.format==="regex"?`ناسم متن: باید د ${d.pattern} سره مطابقت ولري`:`${s[d.format]??c.format} ناسم دی`}case"not_multiple_of":return`ناسم عدد: باید د ${c.divisor} مضرب وي`;case"unrecognized_keys":return`ناسم ${c.keys.length>1?"کلیډونه":"کلیډ"}: ${ti(c.keys,", ")}`;case"invalid_key":return`ناسم کلیډ په ${c.origin} کې`;case"invalid_union":return"ناسمه ورودي";case"invalid_element":return`ناسم عنصر په ${c.origin} کې`;default:return"ناسمه ورودي"}}};function Nsr(){return{localeError:Osr()}}const Rsr=()=>{const n={string:{unit:"znaków",verb:"mieć"},file:{unit:"bajtów",verb:"mieć"},array:{unit:"elementów",verb:"mieć"},set:{unit:"elementów",verb:"mieć"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"liczba";case"object":{if(Array.isArray(c))return"tablica";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"wyrażenie",email:"adres email",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"data i godzina w formacie ISO",date:"data w formacie ISO",time:"godzina w formacie ISO",duration:"czas trwania ISO",ipv4:"adres IPv4",ipv6:"adres IPv6",cidrv4:"zakres IPv4",cidrv6:"zakres IPv6",base64:"ciąg znaków zakodowany w formacie base64",base64url:"ciąg znaków zakodowany w formacie base64url",json_string:"ciąg znaków w formacie JSON",e164:"liczba E.164",jwt:"JWT",template_literal:"wejście"};return c=>{switch(c.code){case"invalid_type":return`Nieprawidłowe dane wejściowe: oczekiwano ${c.expected}, otrzymano ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Nieprawidłowe dane wejściowe: oczekiwano ${es(c.values[0])}`:`Nieprawidłowa opcja: oczekiwano jednej z wartości ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Za duża wartość: oczekiwano, że ${c.origin??"wartość"} będzie mieć ${d}${c.maximum.toString()} ${h.unit??"elementów"}`:`Zbyt duż(y/a/e): oczekiwano, że ${c.origin??"wartość"} będzie wynosić ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Za mała wartość: oczekiwano, że ${c.origin??"wartość"} będzie mieć ${d}${c.minimum.toString()} ${h.unit??"elementów"}`:`Zbyt mał(y/a/e): oczekiwano, że ${c.origin??"wartość"} będzie wynosić ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Nieprawidłowy ciąg znaków: musi zaczynać się od "${d.prefix}"`:d.format==="ends_with"?`Nieprawidłowy ciąg znaków: musi kończyć się na "${d.suffix}"`:d.format==="includes"?`Nieprawidłowy ciąg znaków: musi zawierać "${d.includes}"`:d.format==="regex"?`Nieprawidłowy ciąg znaków: musi odpowiadać wzorcowi ${d.pattern}`:`Nieprawidłow(y/a/e) ${s[d.format]??c.format}`}case"not_multiple_of":return`Nieprawidłowa liczba: musi być wielokrotnością ${c.divisor}`;case"unrecognized_keys":return`Nierozpoznane klucze${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Nieprawidłowy klucz w ${c.origin}`;case"invalid_union":return"Nieprawidłowe dane wejściowe";case"invalid_element":return`Nieprawidłowa wartość w ${c.origin}`;default:return"Nieprawidłowe dane wejściowe"}}};function Psr(){return{localeError:Rsr()}}const Lsr=()=>{const n={string:{unit:"caracteres",verb:"ter"},file:{unit:"bytes",verb:"ter"},array:{unit:"itens",verb:"ter"},set:{unit:"itens",verb:"ter"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"número";case"object":{if(Array.isArray(c))return"array";if(c===null)return"nulo";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"padrão",email:"endereço de e-mail",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"data e hora ISO",date:"data ISO",time:"hora ISO",duration:"duração ISO",ipv4:"endereço IPv4",ipv6:"endereço IPv6",cidrv4:"faixa de IPv4",cidrv6:"faixa de IPv6",base64:"texto codificado em base64",base64url:"URL codificada em base64",json_string:"texto JSON",e164:"número E.164",jwt:"JWT",template_literal:"entrada"};return c=>{switch(c.code){case"invalid_type":return`Tipo inválido: esperado ${c.expected}, recebido ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Entrada inválida: esperado ${es(c.values[0])}`:`Opção inválida: esperada uma das ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Muito grande: esperado que ${c.origin??"valor"} tivesse ${d}${c.maximum.toString()} ${h.unit??"elementos"}`:`Muito grande: esperado que ${c.origin??"valor"} fosse ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Muito pequeno: esperado que ${c.origin} tivesse ${d}${c.minimum.toString()} ${h.unit}`:`Muito pequeno: esperado que ${c.origin} fosse ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Texto inválido: deve começar com "${d.prefix}"`:d.format==="ends_with"?`Texto inválido: deve terminar com "${d.suffix}"`:d.format==="includes"?`Texto inválido: deve incluir "${d.includes}"`:d.format==="regex"?`Texto inválido: deve corresponder ao padrão ${d.pattern}`:`${s[d.format]??c.format} inválido`}case"not_multiple_of":return`Número inválido: deve ser múltiplo de ${c.divisor}`;case"unrecognized_keys":return`Chave${c.keys.length>1?"s":""} desconhecida${c.keys.length>1?"s":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Chave inválida em ${c.origin}`;case"invalid_union":return"Entrada inválida";case"invalid_element":return`Valor inválido em ${c.origin}`;default:return"Campo inválido"}}};function Dsr(){return{localeError:Lsr()}}function Yjt(n,r,o,s){const c=Math.abs(n),d=c%10,h=c%100;return h>=11&&h<=19?s:d===1?r:d>=2&&d<=4?o:s}const Msr=()=>{const n={string:{unit:{one:"символ",few:"символа",many:"символов"},verb:"иметь"},file:{unit:{one:"байт",few:"байта",many:"байт"},verb:"иметь"},array:{unit:{one:"элемент",few:"элемента",many:"элементов"},verb:"иметь"},set:{unit:{one:"элемент",few:"элемента",many:"элементов"},verb:"иметь"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"число";case"object":{if(Array.isArray(c))return"массив";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"ввод",email:"email адрес",url:"URL",emoji:"эмодзи",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO дата и время",date:"ISO дата",time:"ISO время",duration:"ISO длительность",ipv4:"IPv4 адрес",ipv6:"IPv6 адрес",cidrv4:"IPv4 диапазон",cidrv6:"IPv6 диапазон",base64:"строка в формате base64",base64url:"строка в формате base64url",json_string:"JSON строка",e164:"номер E.164",jwt:"JWT",template_literal:"ввод"};return c=>{switch(c.code){case"invalid_type":return`Неверный ввод: ожидалось ${c.expected}, получено ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Неверный ввод: ожидалось ${es(c.values[0])}`:`Неверный вариант: ожидалось одно из ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);if(h){const b=Number(c.maximum),y=Yjt(b,h.unit.one,h.unit.few,h.unit.many);return`Слишком большое значение: ожидалось, что ${c.origin??"значение"} будет иметь ${d}${c.maximum.toString()} ${y}`}return`Слишком большое значение: ожидалось, что ${c.origin??"значение"} будет ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);if(h){const b=Number(c.minimum),y=Yjt(b,h.unit.one,h.unit.few,h.unit.many);return`Слишком маленькое значение: ожидалось, что ${c.origin} будет иметь ${d}${c.minimum.toString()} ${y}`}return`Слишком маленькое значение: ожидалось, что ${c.origin} будет ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Неверная строка: должна начинаться с "${d.prefix}"`:d.format==="ends_with"?`Неверная строка: должна заканчиваться на "${d.suffix}"`:d.format==="includes"?`Неверная строка: должна содержать "${d.includes}"`:d.format==="regex"?`Неверная строка: должна соответствовать шаблону ${d.pattern}`:`Неверный ${s[d.format]??c.format}`}case"not_multiple_of":return`Неверное число: должно быть кратным ${c.divisor}`;case"unrecognized_keys":return`Нераспознанн${c.keys.length>1?"ые":"ый"} ключ${c.keys.length>1?"и":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Неверный ключ в ${c.origin}`;case"invalid_union":return"Неверные входные данные";case"invalid_element":return`Неверное значение в ${c.origin}`;default:return"Неверные входные данные"}}};function jsr(){return{localeError:Msr()}}const $sr=()=>{const n={string:{unit:"znakov",verb:"imeti"},file:{unit:"bajtov",verb:"imeti"},array:{unit:"elementov",verb:"imeti"},set:{unit:"elementov",verb:"imeti"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"število";case"object":{if(Array.isArray(c))return"tabela";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"vnos",email:"e-poštni naslov",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO datum in čas",date:"ISO datum",time:"ISO čas",duration:"ISO trajanje",ipv4:"IPv4 naslov",ipv6:"IPv6 naslov",cidrv4:"obseg IPv4",cidrv6:"obseg IPv6",base64:"base64 kodiran niz",base64url:"base64url kodiran niz",json_string:"JSON niz",e164:"E.164 številka",jwt:"JWT",template_literal:"vnos"};return c=>{switch(c.code){case"invalid_type":return`Neveljaven vnos: pričakovano ${c.expected}, prejeto ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Neveljaven vnos: pričakovano ${es(c.values[0])}`:`Neveljavna možnost: pričakovano eno izmed ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Preveliko: pričakovano, da bo ${c.origin??"vrednost"} imelo ${d}${c.maximum.toString()} ${h.unit??"elementov"}`:`Preveliko: pričakovano, da bo ${c.origin??"vrednost"} ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Premajhno: pričakovano, da bo ${c.origin} imelo ${d}${c.minimum.toString()} ${h.unit}`:`Premajhno: pričakovano, da bo ${c.origin} ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Neveljaven niz: mora se začeti z "${d.prefix}"`:d.format==="ends_with"?`Neveljaven niz: mora se končati z "${d.suffix}"`:d.format==="includes"?`Neveljaven niz: mora vsebovati "${d.includes}"`:d.format==="regex"?`Neveljaven niz: mora ustrezati vzorcu ${d.pattern}`:`Neveljaven ${s[d.format]??c.format}`}case"not_multiple_of":return`Neveljavno število: mora biti večkratnik ${c.divisor}`;case"unrecognized_keys":return`Neprepoznan${c.keys.length>1?"i ključi":" ključ"}: ${ti(c.keys,", ")}`;case"invalid_key":return`Neveljaven ključ v ${c.origin}`;case"invalid_union":return"Neveljaven vnos";case"invalid_element":return`Neveljavna vrednost v ${c.origin}`;default:return"Neveljaven vnos"}}};function Fsr(){return{localeError:$sr()}}const Bsr=()=>{const n={string:{unit:"tecken",verb:"att ha"},file:{unit:"bytes",verb:"att ha"},array:{unit:"objekt",verb:"att innehålla"},set:{unit:"objekt",verb:"att innehålla"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"antal";case"object":{if(Array.isArray(c))return"lista";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"reguljärt uttryck",email:"e-postadress",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO-datum och tid",date:"ISO-datum",time:"ISO-tid",duration:"ISO-varaktighet",ipv4:"IPv4-intervall",ipv6:"IPv6-intervall",cidrv4:"IPv4-spektrum",cidrv6:"IPv6-spektrum",base64:"base64-kodad sträng",base64url:"base64url-kodad sträng",json_string:"JSON-sträng",e164:"E.164-nummer",jwt:"JWT",template_literal:"mall-literal"};return c=>{switch(c.code){case"invalid_type":return`Ogiltig inmatning: förväntat ${c.expected}, fick ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Ogiltig inmatning: förväntat ${es(c.values[0])}`:`Ogiltigt val: förväntade en av ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`För stor(t): förväntade ${c.origin??"värdet"} att ha ${d}${c.maximum.toString()} ${h.unit??"element"}`:`För stor(t): förväntat ${c.origin??"värdet"} att ha ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`För lite(t): förväntade ${c.origin??"värdet"} att ha ${d}${c.minimum.toString()} ${h.unit}`:`För lite(t): förväntade ${c.origin??"värdet"} att ha ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Ogiltig sträng: måste börja med "${d.prefix}"`:d.format==="ends_with"?`Ogiltig sträng: måste sluta med "${d.suffix}"`:d.format==="includes"?`Ogiltig sträng: måste innehålla "${d.includes}"`:d.format==="regex"?`Ogiltig sträng: måste matcha mönstret "${d.pattern}"`:`Ogiltig(t) ${s[d.format]??c.format}`}case"not_multiple_of":return`Ogiltigt tal: måste vara en multipel av ${c.divisor}`;case"unrecognized_keys":return`${c.keys.length>1?"Okända nycklar":"Okänd nyckel"}: ${ti(c.keys,", ")}`;case"invalid_key":return`Ogiltig nyckel i ${c.origin??"värdet"}`;case"invalid_union":return"Ogiltig input";case"invalid_element":return`Ogiltigt värde i ${c.origin??"värdet"}`;default:return"Ogiltig input"}}};function Usr(){return{localeError:Bsr()}}const zsr=()=>{const n={string:{unit:"எழுத்துக்கள்",verb:"கொண்டிருக்க வேண்டும்"},file:{unit:"பைட்டுகள்",verb:"கொண்டிருக்க வேண்டும்"},array:{unit:"உறுப்புகள்",verb:"கொண்டிருக்க வேண்டும்"},set:{unit:"உறுப்புகள்",verb:"கொண்டிருக்க வேண்டும்"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"எண் அல்லாதது":"எண்";case"object":{if(Array.isArray(c))return"அணி";if(c===null)return"வெறுமை";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"உள்ளீடு",email:"மின்னஞ்சல் முகவரி",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO தேதி நேரம்",date:"ISO தேதி",time:"ISO நேரம்",duration:"ISO கால அளவு",ipv4:"IPv4 முகவரி",ipv6:"IPv6 முகவரி",cidrv4:"IPv4 வரம்பு",cidrv6:"IPv6 வரம்பு",base64:"base64-encoded சரம்",base64url:"base64url-encoded சரம்",json_string:"JSON சரம்",e164:"E.164 எண்",jwt:"JWT",template_literal:"input"};return c=>{switch(c.code){case"invalid_type":return`தவறான உள்ளீடு: எதிர்பார்க்கப்பட்டது ${c.expected}, பெறப்பட்டது ${o(c.input)}`;case"invalid_value":return c.values.length===1?`தவறான உள்ளீடு: எதிர்பார்க்கப்பட்டது ${es(c.values[0])}`:`தவறான விருப்பம்: எதிர்பார்க்கப்பட்டது ${ti(c.values,"|")} இல் ஒன்று`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`மிக பெரியது: எதிர்பார்க்கப்பட்டது ${c.origin??"மதிப்பு"} ${d}${c.maximum.toString()} ${h.unit??"உறுப்புகள்"} ஆக இருக்க வேண்டும்`:`மிக பெரியது: எதிர்பார்க்கப்பட்டது ${c.origin??"மதிப்பு"} ${d}${c.maximum.toString()} ஆக இருக்க வேண்டும்`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`மிகச் சிறியது: எதிர்பார்க்கப்பட்டது ${c.origin} ${d}${c.minimum.toString()} ${h.unit} ஆக இருக்க வேண்டும்`:`மிகச் சிறியது: எதிர்பார்க்கப்பட்டது ${c.origin} ${d}${c.minimum.toString()} ஆக இருக்க வேண்டும்`}case"invalid_format":{const d=c;return d.format==="starts_with"?`தவறான சரம்: "${d.prefix}" இல் தொடங்க வேண்டும்`:d.format==="ends_with"?`தவறான சரம்: "${d.suffix}" இல் முடிவடைய வேண்டும்`:d.format==="includes"?`தவறான சரம்: "${d.includes}" ஐ உள்ளடக்க வேண்டும்`:d.format==="regex"?`தவறான சரம்: ${d.pattern} முறைபாட்டுடன் பொருந்த வேண்டும்`:`தவறான ${s[d.format]??c.format}`}case"not_multiple_of":return`தவறான எண்: ${c.divisor} இன் பலமாக இருக்க வேண்டும்`;case"unrecognized_keys":return`அடையாளம் தெரியாத விசை${c.keys.length>1?"கள்":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`${c.origin} இல் தவறான விசை`;case"invalid_union":return"தவறான உள்ளீடு";case"invalid_element":return`${c.origin} இல் தவறான மதிப்பு`;default:return"தவறான உள்ளீடு"}}};function Hsr(){return{localeError:zsr()}}const Gsr=()=>{const n={string:{unit:"ตัวอักษร",verb:"ควรมี"},file:{unit:"ไบต์",verb:"ควรมี"},array:{unit:"รายการ",verb:"ควรมี"},set:{unit:"รายการ",verb:"ควรมี"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"ไม่ใช่ตัวเลข (NaN)":"ตัวเลข";case"object":{if(Array.isArray(c))return"อาร์เรย์ (Array)";if(c===null)return"ไม่มีค่า (null)";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"ข้อมูลที่ป้อน",email:"ที่อยู่อีเมล",url:"URL",emoji:"อิโมจิ",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"วันที่เวลาแบบ ISO",date:"วันที่แบบ ISO",time:"เวลาแบบ ISO",duration:"ช่วงเวลาแบบ ISO",ipv4:"ที่อยู่ IPv4",ipv6:"ที่อยู่ IPv6",cidrv4:"ช่วง IP แบบ IPv4",cidrv6:"ช่วง IP แบบ IPv6",base64:"ข้อความแบบ Base64",base64url:"ข้อความแบบ Base64 สำหรับ URL",json_string:"ข้อความแบบ JSON",e164:"เบอร์โทรศัพท์ระหว่างประเทศ (E.164)",jwt:"โทเคน JWT",template_literal:"ข้อมูลที่ป้อน"};return c=>{switch(c.code){case"invalid_type":return`ประเภทข้อมูลไม่ถูกต้อง: ควรเป็น ${c.expected} แต่ได้รับ ${o(c.input)}`;case"invalid_value":return c.values.length===1?`ค่าไม่ถูกต้อง: ควรเป็น ${es(c.values[0])}`:`ตัวเลือกไม่ถูกต้อง: ควรเป็นหนึ่งใน ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"ไม่เกิน":"น้อยกว่า",h=r(c.origin);return h?`เกินกำหนด: ${c.origin??"ค่า"} ควรมี${d} ${c.maximum.toString()} ${h.unit??"รายการ"}`:`เกินกำหนด: ${c.origin??"ค่า"} ควรมี${d} ${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?"อย่างน้อย":"มากกว่า",h=r(c.origin);return h?`น้อยกว่ากำหนด: ${c.origin} ควรมี${d} ${c.minimum.toString()} ${h.unit}`:`น้อยกว่ากำหนด: ${c.origin} ควรมี${d} ${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`รูปแบบไม่ถูกต้อง: ข้อความต้องขึ้นต้นด้วย "${d.prefix}"`:d.format==="ends_with"?`รูปแบบไม่ถูกต้อง: ข้อความต้องลงท้ายด้วย "${d.suffix}"`:d.format==="includes"?`รูปแบบไม่ถูกต้อง: ข้อความต้องมี "${d.includes}" อยู่ในข้อความ`:d.format==="regex"?`รูปแบบไม่ถูกต้อง: ต้องตรงกับรูปแบบที่กำหนด ${d.pattern}`:`รูปแบบไม่ถูกต้อง: ${s[d.format]??c.format}`}case"not_multiple_of":return`ตัวเลขไม่ถูกต้อง: ต้องเป็นจำนวนที่หารด้วย ${c.divisor} ได้ลงตัว`;case"unrecognized_keys":return`พบคีย์ที่ไม่รู้จัก: ${ti(c.keys,", ")}`;case"invalid_key":return`คีย์ไม่ถูกต้องใน ${c.origin}`;case"invalid_union":return"ข้อมูลไม่ถูกต้อง: ไม่ตรงกับรูปแบบยูเนียนที่กำหนดไว้";case"invalid_element":return`ข้อมูลไม่ถูกต้องใน ${c.origin}`;default:return"ข้อมูลไม่ถูกต้อง"}}};function qsr(){return{localeError:Gsr()}}const Vsr=n=>{const r=typeof n;switch(r){case"number":return Number.isNaN(n)?"NaN":"number";case"object":{if(Array.isArray(n))return"array";if(n===null)return"null";if(Object.getPrototypeOf(n)!==Object.prototype&&n.constructor)return n.constructor.name}}return r},Wsr=()=>{const n={string:{unit:"karakter",verb:"olmalı"},file:{unit:"bayt",verb:"olmalı"},array:{unit:"öğe",verb:"olmalı"},set:{unit:"öğe",verb:"olmalı"}};function r(s){return n[s]??null}const o={regex:"girdi",email:"e-posta adresi",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO tarih ve saat",date:"ISO tarih",time:"ISO saat",duration:"ISO süre",ipv4:"IPv4 adresi",ipv6:"IPv6 adresi",cidrv4:"IPv4 aralığı",cidrv6:"IPv6 aralığı",base64:"base64 ile şifrelenmiş metin",base64url:"base64url ile şifrelenmiş metin",json_string:"JSON dizesi",e164:"E.164 sayısı",jwt:"JWT",template_literal:"Şablon dizesi"};return s=>{switch(s.code){case"invalid_type":return`Geçersiz değer: beklenen ${s.expected}, alınan ${Vsr(s.input)}`;case"invalid_value":return s.values.length===1?`Geçersiz değer: beklenen ${es(s.values[0])}`:`Geçersiz seçenek: aşağıdakilerden biri olmalı: ${ti(s.values,"|")}`;case"too_big":{const c=s.inclusive?"<=":"<",d=r(s.origin);return d?`Çok büyük: beklenen ${s.origin??"değer"} ${c}${s.maximum.toString()} ${d.unit??"öğe"}`:`Çok büyük: beklenen ${s.origin??"değer"} ${c}${s.maximum.toString()}`}case"too_small":{const c=s.inclusive?">=":">",d=r(s.origin);return d?`Çok küçük: beklenen ${s.origin} ${c}${s.minimum.toString()} ${d.unit}`:`Çok küçük: beklenen ${s.origin} ${c}${s.minimum.toString()}`}case"invalid_format":{const c=s;return c.format==="starts_with"?`Geçersiz metin: "${c.prefix}" ile başlamalı`:c.format==="ends_with"?`Geçersiz metin: "${c.suffix}" ile bitmeli`:c.format==="includes"?`Geçersiz metin: "${c.includes}" içermeli`:c.format==="regex"?`Geçersiz metin: ${c.pattern} desenine uymalı`:`Geçersiz ${o[c.format]??s.format}`}case"not_multiple_of":return`Geçersiz sayı: ${s.divisor} ile tam bölünebilmeli`;case"unrecognized_keys":return`Tanınmayan anahtar${s.keys.length>1?"lar":""}: ${ti(s.keys,", ")}`;case"invalid_key":return`${s.origin} içinde geçersiz anahtar`;case"invalid_union":return"Geçersiz değer";case"invalid_element":return`${s.origin} içinde geçersiz değer`;default:return"Geçersiz değer"}}};function Ksr(){return{localeError:Wsr()}}const Ysr=()=>{const n={string:{unit:"символів",verb:"матиме"},file:{unit:"байтів",verb:"матиме"},array:{unit:"елементів",verb:"матиме"},set:{unit:"елементів",verb:"матиме"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"число";case"object":{if(Array.isArray(c))return"масив";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"вхідні дані",email:"адреса електронної пошти",url:"URL",emoji:"емодзі",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"дата та час ISO",date:"дата ISO",time:"час ISO",duration:"тривалість ISO",ipv4:"адреса IPv4",ipv6:"адреса IPv6",cidrv4:"діапазон IPv4",cidrv6:"діапазон IPv6",base64:"рядок у кодуванні base64",base64url:"рядок у кодуванні base64url",json_string:"рядок JSON",e164:"номер E.164",jwt:"JWT",template_literal:"вхідні дані"};return c=>{switch(c.code){case"invalid_type":return`Неправильні вхідні дані: очікується ${c.expected}, отримано ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Неправильні вхідні дані: очікується ${es(c.values[0])}`:`Неправильна опція: очікується одне з ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Занадто велике: очікується, що ${c.origin??"значення"} ${h.verb} ${d}${c.maximum.toString()} ${h.unit??"елементів"}`:`Занадто велике: очікується, що ${c.origin??"значення"} буде ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Занадто мале: очікується, що ${c.origin} ${h.verb} ${d}${c.minimum.toString()} ${h.unit}`:`Занадто мале: очікується, що ${c.origin} буде ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Неправильний рядок: повинен починатися з "${d.prefix}"`:d.format==="ends_with"?`Неправильний рядок: повинен закінчуватися на "${d.suffix}"`:d.format==="includes"?`Неправильний рядок: повинен містити "${d.includes}"`:d.format==="regex"?`Неправильний рядок: повинен відповідати шаблону ${d.pattern}`:`Неправильний ${s[d.format]??c.format}`}case"not_multiple_of":return`Неправильне число: повинно бути кратним ${c.divisor}`;case"unrecognized_keys":return`Нерозпізнаний ключ${c.keys.length>1?"і":""}: ${ti(c.keys,", ")}`;case"invalid_key":return`Неправильний ключ у ${c.origin}`;case"invalid_union":return"Неправильні вхідні дані";case"invalid_element":return`Неправильне значення у ${c.origin}`;default:return"Неправильні вхідні дані"}}};function Xsr(){return{localeError:Ysr()}}const Jsr=()=>{const n={string:{unit:"حروف",verb:"ہونا"},file:{unit:"بائٹس",verb:"ہونا"},array:{unit:"آئٹمز",verb:"ہونا"},set:{unit:"آئٹمز",verb:"ہونا"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"نمبر";case"object":{if(Array.isArray(c))return"آرے";if(c===null)return"نل";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"ان پٹ",email:"ای میل ایڈریس",url:"یو آر ایل",emoji:"ایموجی",uuid:"یو یو آئی ڈی",uuidv4:"یو یو آئی ڈی وی 4",uuidv6:"یو یو آئی ڈی وی 6",nanoid:"نینو آئی ڈی",guid:"جی یو آئی ڈی",cuid:"سی یو آئی ڈی",cuid2:"سی یو آئی ڈی 2",ulid:"یو ایل آئی ڈی",xid:"ایکس آئی ڈی",ksuid:"کے ایس یو آئی ڈی",datetime:"آئی ایس او ڈیٹ ٹائم",date:"آئی ایس او تاریخ",time:"آئی ایس او وقت",duration:"آئی ایس او مدت",ipv4:"آئی پی وی 4 ایڈریس",ipv6:"آئی پی وی 6 ایڈریس",cidrv4:"آئی پی وی 4 رینج",cidrv6:"آئی پی وی 6 رینج",base64:"بیس 64 ان کوڈڈ سٹرنگ",base64url:"بیس 64 یو آر ایل ان کوڈڈ سٹرنگ",json_string:"جے ایس او این سٹرنگ",e164:"ای 164 نمبر",jwt:"جے ڈبلیو ٹی",template_literal:"ان پٹ"};return c=>{switch(c.code){case"invalid_type":return`غلط ان پٹ: ${c.expected} متوقع تھا، ${o(c.input)} موصول ہوا`;case"invalid_value":return c.values.length===1?`غلط ان پٹ: ${es(c.values[0])} متوقع تھا`:`غلط آپشن: ${ti(c.values,"|")} میں سے ایک متوقع تھا`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`بہت بڑا: ${c.origin??"ویلیو"} کے ${d}${c.maximum.toString()} ${h.unit??"عناصر"} ہونے متوقع تھے`:`بہت بڑا: ${c.origin??"ویلیو"} کا ${d}${c.maximum.toString()} ہونا متوقع تھا`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`بہت چھوٹا: ${c.origin} کے ${d}${c.minimum.toString()} ${h.unit} ہونے متوقع تھے`:`بہت چھوٹا: ${c.origin} کا ${d}${c.minimum.toString()} ہونا متوقع تھا`}case"invalid_format":{const d=c;return d.format==="starts_with"?`غلط سٹرنگ: "${d.prefix}" سے شروع ہونا چاہیے`:d.format==="ends_with"?`غلط سٹرنگ: "${d.suffix}" پر ختم ہونا چاہیے`:d.format==="includes"?`غلط سٹرنگ: "${d.includes}" شامل ہونا چاہیے`:d.format==="regex"?`غلط سٹرنگ: پیٹرن ${d.pattern} سے میچ ہونا چاہیے`:`غلط ${s[d.format]??c.format}`}case"not_multiple_of":return`غلط نمبر: ${c.divisor} کا مضاعف ہونا چاہیے`;case"unrecognized_keys":return`غیر تسلیم شدہ کی${c.keys.length>1?"ز":""}: ${ti(c.keys,"، ")}`;case"invalid_key":return`${c.origin} میں غلط کی`;case"invalid_union":return"غلط ان پٹ";case"invalid_element":return`${c.origin} میں غلط ویلیو`;default:return"غلط ان پٹ"}}};function Zsr(){return{localeError:Jsr()}}const Qsr=()=>{const n={string:{unit:"ký tự",verb:"có"},file:{unit:"byte",verb:"có"},array:{unit:"phần tử",verb:"có"},set:{unit:"phần tử",verb:"có"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"số";case"object":{if(Array.isArray(c))return"mảng";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"đầu vào",email:"địa chỉ email",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ngày giờ ISO",date:"ngày ISO",time:"giờ ISO",duration:"khoảng thời gian ISO",ipv4:"địa chỉ IPv4",ipv6:"địa chỉ IPv6",cidrv4:"dải IPv4",cidrv6:"dải IPv6",base64:"chuỗi mã hóa base64",base64url:"chuỗi mã hóa base64url",json_string:"chuỗi JSON",e164:"số E.164",jwt:"JWT",template_literal:"đầu vào"};return c=>{switch(c.code){case"invalid_type":return`Đầu vào không hợp lệ: mong đợi ${c.expected}, nhận được ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Đầu vào không hợp lệ: mong đợi ${es(c.values[0])}`:`Tùy chọn không hợp lệ: mong đợi một trong các giá trị ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Quá lớn: mong đợi ${c.origin??"giá trị"} ${h.verb} ${d}${c.maximum.toString()} ${h.unit??"phần tử"}`:`Quá lớn: mong đợi ${c.origin??"giá trị"} ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Quá nhỏ: mong đợi ${c.origin} ${h.verb} ${d}${c.minimum.toString()} ${h.unit}`:`Quá nhỏ: mong đợi ${c.origin} ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Chuỗi không hợp lệ: phải bắt đầu bằng "${d.prefix}"`:d.format==="ends_with"?`Chuỗi không hợp lệ: phải kết thúc bằng "${d.suffix}"`:d.format==="includes"?`Chuỗi không hợp lệ: phải bao gồm "${d.includes}"`:d.format==="regex"?`Chuỗi không hợp lệ: phải khớp với mẫu ${d.pattern}`:`${s[d.format]??c.format} không hợp lệ`}case"not_multiple_of":return`Số không hợp lệ: phải là bội số của ${c.divisor}`;case"unrecognized_keys":return`Khóa không được nhận dạng: ${ti(c.keys,", ")}`;case"invalid_key":return`Khóa không hợp lệ trong ${c.origin}`;case"invalid_union":return"Đầu vào không hợp lệ";case"invalid_element":return`Giá trị không hợp lệ trong ${c.origin}`;default:return"Đầu vào không hợp lệ"}}};function ecr(){return{localeError:Qsr()}}const tcr=()=>{const n={string:{unit:"字符",verb:"包含"},file:{unit:"字节",verb:"包含"},array:{unit:"项",verb:"包含"},set:{unit:"项",verb:"包含"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"非数字(NaN)":"数字";case"object":{if(Array.isArray(c))return"数组";if(c===null)return"空值(null)";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"输入",email:"电子邮件",url:"URL",emoji:"表情符号",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO日期时间",date:"ISO日期",time:"ISO时间",duration:"ISO时长",ipv4:"IPv4地址",ipv6:"IPv6地址",cidrv4:"IPv4网段",cidrv6:"IPv6网段",base64:"base64编码字符串",base64url:"base64url编码字符串",json_string:"JSON字符串",e164:"E.164号码",jwt:"JWT",template_literal:"输入"};return c=>{switch(c.code){case"invalid_type":return`无效输入：期望 ${c.expected}，实际接收 ${o(c.input)}`;case"invalid_value":return c.values.length===1?`无效输入：期望 ${es(c.values[0])}`:`无效选项：期望以下之一 ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`数值过大：期望 ${c.origin??"值"} ${d}${c.maximum.toString()} ${h.unit??"个元素"}`:`数值过大：期望 ${c.origin??"值"} ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`数值过小：期望 ${c.origin} ${d}${c.minimum.toString()} ${h.unit}`:`数值过小：期望 ${c.origin} ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`无效字符串：必须以 "${d.prefix}" 开头`:d.format==="ends_with"?`无效字符串：必须以 "${d.suffix}" 结尾`:d.format==="includes"?`无效字符串：必须包含 "${d.includes}"`:d.format==="regex"?`无效字符串：必须满足正则表达式 ${d.pattern}`:`无效${s[d.format]??c.format}`}case"not_multiple_of":return`无效数字：必须是 ${c.divisor} 的倍数`;case"unrecognized_keys":return`出现未知的键(key): ${ti(c.keys,", ")}`;case"invalid_key":return`${c.origin} 中的键(key)无效`;case"invalid_union":return"无效输入";case"invalid_element":return`${c.origin} 中包含无效值(value)`;default:return"无效输入"}}};function ncr(){return{localeError:tcr()}}const rcr=()=>{const n={string:{unit:"字元",verb:"擁有"},file:{unit:"位元組",verb:"擁有"},array:{unit:"項目",verb:"擁有"},set:{unit:"項目",verb:"擁有"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"number";case"object":{if(Array.isArray(c))return"array";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"輸入",email:"郵件地址",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"ISO 日期時間",date:"ISO 日期",time:"ISO 時間",duration:"ISO 期間",ipv4:"IPv4 位址",ipv6:"IPv6 位址",cidrv4:"IPv4 範圍",cidrv6:"IPv6 範圍",base64:"base64 編碼字串",base64url:"base64url 編碼字串",json_string:"JSON 字串",e164:"E.164 數值",jwt:"JWT",template_literal:"輸入"};return c=>{switch(c.code){case"invalid_type":return`無效的輸入值：預期為 ${c.expected}，但收到 ${o(c.input)}`;case"invalid_value":return c.values.length===1?`無效的輸入值：預期為 ${es(c.values[0])}`:`無效的選項：預期為以下其中之一 ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`數值過大：預期 ${c.origin??"值"} 應為 ${d}${c.maximum.toString()} ${h.unit??"個元素"}`:`數值過大：預期 ${c.origin??"值"} 應為 ${d}${c.maximum.toString()}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`數值過小：預期 ${c.origin} 應為 ${d}${c.minimum.toString()} ${h.unit}`:`數值過小：預期 ${c.origin} 應為 ${d}${c.minimum.toString()}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`無效的字串：必須以 "${d.prefix}" 開頭`:d.format==="ends_with"?`無效的字串：必須以 "${d.suffix}" 結尾`:d.format==="includes"?`無效的字串：必須包含 "${d.includes}"`:d.format==="regex"?`無效的字串：必須符合格式 ${d.pattern}`:`無效的 ${s[d.format]??c.format}`}case"not_multiple_of":return`無效的數字：必須為 ${c.divisor} 的倍數`;case"unrecognized_keys":return`無法識別的鍵值${c.keys.length>1?"們":""}：${ti(c.keys,"、")}`;case"invalid_key":return`${c.origin} 中有無效的鍵值`;case"invalid_union":return"無效的輸入值";case"invalid_element":return`${c.origin} 中有無效的值`;default:return"無效的輸入值"}}};function icr(){return{localeError:rcr()}}const ocr=()=>{const n={string:{unit:"àmi",verb:"ní"},file:{unit:"bytes",verb:"ní"},array:{unit:"nkan",verb:"ní"},set:{unit:"nkan",verb:"ní"}};function r(c){return n[c]??null}const o=c=>{const d=typeof c;switch(d){case"number":return Number.isNaN(c)?"NaN":"nọ́mbà";case"object":{if(Array.isArray(c))return"akopọ";if(c===null)return"null";if(Object.getPrototypeOf(c)!==Object.prototype&&c.constructor)return c.constructor.name}}return d},s={regex:"ẹ̀rọ ìbáwọlé",email:"àdírẹ́sì ìmẹ́lì",url:"URL",emoji:"emoji",uuid:"UUID",uuidv4:"UUIDv4",uuidv6:"UUIDv6",nanoid:"nanoid",guid:"GUID",cuid:"cuid",cuid2:"cuid2",ulid:"ULID",xid:"XID",ksuid:"KSUID",datetime:"àkókò ISO",date:"ọjọ́ ISO",time:"àkókò ISO",duration:"àkókò tó pé ISO",ipv4:"àdírẹ́sì IPv4",ipv6:"àdírẹ́sì IPv6",cidrv4:"àgbègbè IPv4",cidrv6:"àgbègbè IPv6",base64:"ọ̀rọ̀ tí a kọ́ ní base64",base64url:"ọ̀rọ̀ base64url",json_string:"ọ̀rọ̀ JSON",e164:"nọ́mbà E.164",jwt:"JWT",template_literal:"ẹ̀rọ ìbáwọlé"};return c=>{switch(c.code){case"invalid_type":return`Ìbáwọlé aṣìṣe: a ní láti fi ${c.expected}, àmọ̀ a rí ${o(c.input)}`;case"invalid_value":return c.values.length===1?`Ìbáwọlé aṣìṣe: a ní láti fi ${es(c.values[0])}`:`Àṣàyàn aṣìṣe: yan ọ̀kan lára ${ti(c.values,"|")}`;case"too_big":{const d=c.inclusive?"<=":"<",h=r(c.origin);return h?`Tó pọ̀ jù: a ní láti jẹ́ pé ${c.origin??"iye"} ${h.verb} ${d}${c.maximum} ${h.unit}`:`Tó pọ̀ jù: a ní láti jẹ́ ${d}${c.maximum}`}case"too_small":{const d=c.inclusive?">=":">",h=r(c.origin);return h?`Kéré ju: a ní láti jẹ́ pé ${c.origin} ${h.verb} ${d}${c.minimum} ${h.unit}`:`Kéré ju: a ní láti jẹ́ ${d}${c.minimum}`}case"invalid_format":{const d=c;return d.format==="starts_with"?`Ọ̀rọ̀ aṣìṣe: gbọ́dọ̀ bẹ̀rẹ̀ pẹ̀lú "${d.prefix}"`:d.format==="ends_with"?`Ọ̀rọ̀ aṣìṣe: gbọ́dọ̀ parí pẹ̀lú "${d.suffix}"`:d.format==="includes"?`Ọ̀rọ̀ aṣìṣe: gbọ́dọ̀ ní "${d.includes}"`:d.format==="regex"?`Ọ̀rọ̀ aṣìṣe: gbọ́dọ̀ bá àpẹẹrẹ mu ${d.pattern}`:`Aṣìṣe: ${s[d.format]??c.format}`}case"not_multiple_of":return`Nọ́mbà aṣìṣe: gbọ́dọ̀ jẹ́ èyà pípín ti ${c.divisor}`;case"unrecognized_keys":return`Bọtìnì àìmọ̀: ${ti(c.keys,", ")}`;case"invalid_key":return`Bọtìnì aṣìṣe nínú ${c.origin}`;case"invalid_union":return"Ìbáwọlé aṣìṣe";case"invalid_element":return`Iye aṣìṣe nínú ${c.origin}`;default:return"Ìbáwọlé aṣìṣe"}}};function acr(){return{localeError:ocr()}}const Nin=Object.freeze(Object.defineProperty({__proto__:null,ar:Iar,az:Nar,be:Par,ca:Dar,cs:jar,da:Far,de:Uar,en:Oin,eo:Var,es:Kar,fa:Xar,fi:Zar,fr:esr,frCA:nsr,he:isr,hu:asr,id:csr,is:dsr,it:hsr,ja:gsr,kh:bsr,ko:wsr,mk:Esr,ms:_sr,nl:Tsr,no:Csr,ota:Isr,pl:Psr,ps:Nsr,pt:Dsr,ru:jsr,sl:Fsr,sv:Usr,ta:Hsr,th:qsr,tr:Ksr,ua:Xsr,ur:Zsr,vi:ecr,yo:acr,zhCN:ncr,zhTW:icr},Symbol.toStringTag,{value:"Module"})),Rin=Symbol("ZodOutput"),Pin=Symbol("ZodInput");class NXe{constructor(){this._map=new Map,this._idmap=new Map}add(r,...o){const s=o[0];if(this._map.set(r,s),s&&typeof s=="object"&&"id"in s){if(this._idmap.has(s.id))throw new Error(`ID ${s.id} already exists in the registry`);this._idmap.set(s.id,r)}return this}clear(){return this._map=new Map,this._idmap=new Map,this}remove(r){const o=this._map.get(r);return o&&typeof o=="object"&&"id"in o&&this._idmap.delete(o.id),this._map.delete(r),this}get(r){const o=r._zod.parent;if(o){const s={...this.get(o)??{}};delete s.id;const c={...s,...this._map.get(r)};return Object.keys(c).length?c:void 0}return this._map.get(r)}has(r){return this._map.has(r)}}function RXe(){return new NXe}const PB=RXe();function Lin(n,r){return new n({type:"string",...gr(r)})}function Din(n,r){return new n({type:"string",coerce:!0,...gr(r)})}function PXe(n,r){return new n({type:"string",format:"email",check:"string_format",abort:!1,...gr(r)})}function cwe(n,r){return new n({type:"string",format:"guid",check:"string_format",abort:!1,...gr(r)})}function LXe(n,r){return new n({type:"string",format:"uuid",check:"string_format",abort:!1,...gr(r)})}function DXe(n,r){return new n({type:"string",format:"uuid",check:"string_format",abort:!1,version:"v4",...gr(r)})}function MXe(n,r){return new n({type:"string",format:"uuid",check:"string_format",abort:!1,version:"v6",...gr(r)})}function jXe(n,r){return new n({type:"string",format:"uuid",check:"string_format",abort:!1,version:"v7",...gr(r)})}function $Xe(n,r){return new n({type:"string",format:"url",check:"string_format",abort:!1,...gr(r)})}function FXe(n,r){return new n({type:"string",format:"emoji",check:"string_format",abort:!1,...gr(r)})}function BXe(n,r){return new n({type:"string",format:"nanoid",check:"string_format",abort:!1,...gr(r)})}function UXe(n,r){return new n({type:"string",format:"cuid",check:"string_format",abort:!1,...gr(r)})}function zXe(n,r){return new n({type:"string",format:"cuid2",check:"string_format",abort:!1,...gr(r)})}function HXe(n,r){return new n({type:"string",format:"ulid",check:"string_format",abort:!1,...gr(r)})}function GXe(n,r){return new n({type:"string",format:"xid",check:"string_format",abort:!1,...gr(r)})}function qXe(n,r){return new n({type:"string",format:"ksuid",check:"string_format",abort:!1,...gr(r)})}function VXe(n,r){return new n({type:"string",format:"ipv4",check:"string_format",abort:!1,...gr(r)})}function WXe(n,r){return new n({type:"string",format:"ipv6",check:"string_format",abort:!1,...gr(r)})}function KXe(n,r){return new n({type:"string",format:"cidrv4",check:"string_format",abort:!1,...gr(r)})}function YXe(n,r){return new n({type:"string",format:"cidrv6",check:"string_format",abort:!1,...gr(r)})}function XXe(n,r){return new n({type:"string",format:"base64",check:"string_format",abort:!1,...gr(r)})}function JXe(n,r){return new n({type:"string",format:"base64url",check:"string_format",abort:!1,...gr(r)})}function ZXe(n,r){return new n({type:"string",format:"e164",check:"string_format",abort:!1,...gr(r)})}function QXe(n,r){return new n({type:"string",format:"jwt",check:"string_format",abort:!1,...gr(r)})}const Min={Any:null,Minute:-1,Second:0,Millisecond:3,Microsecond:6};function jin(n,r){return new n({type:"string",format:"datetime",check:"string_format",offset:!1,local:!1,precision:null,...gr(r)})}function $in(n,r){return new n({type:"string",format:"date",check:"string_format",...gr(r)})}function Fin(n,r){return new n({type:"string",format:"time",check:"string_format",precision:null,...gr(r)})}function Bin(n,r){return new n({type:"string",format:"duration",check:"string_format",...gr(r)})}function Uin(n,r){return new n({type:"number",checks:[],...gr(r)})}function zin(n,r){return new n({type:"number",coerce:!0,checks:[],...gr(r)})}function Hin(n,r){return new n({type:"number",check:"number_format",abort:!1,format:"safeint",...gr(r)})}function Gin(n,r){return new n({type:"number",check:"number_format",abort:!1,format:"float32",...gr(r)})}function qin(n,r){return new n({type:"number",check:"number_format",abort:!1,format:"float64",...gr(r)})}function Vin(n,r){return new n({type:"number",check:"number_format",abort:!1,format:"int32",...gr(r)})}function Win(n,r){return new n({type:"number",check:"number_format",abort:!1,format:"uint32",...gr(r)})}function Kin(n,r){return new n({type:"boolean",...gr(r)})}function Yin(n,r){return new n({type:"boolean",coerce:!0,...gr(r)})}function Xin(n,r){return new n({type:"bigint",...gr(r)})}function Jin(n,r){return new n({type:"bigint",coerce:!0,...gr(r)})}function Zin(n,r){return new n({type:"bigint",check:"bigint_format",abort:!1,format:"int64",...gr(r)})}function Qin(n,r){return new n({type:"bigint",check:"bigint_format",abort:!1,format:"uint64",...gr(r)})}function eon(n,r){return new n({type:"symbol",...gr(r)})}function ton(n,r){return new n({type:"undefined",...gr(r)})}function non(n,r){return new n({type:"null",...gr(r)})}function ron(n){return new n({type:"any"})}function uwe(n){return new n({type:"unknown"})}function ion(n,r){return new n({type:"never",...gr(r)})}function oon(n,r){return new n({type:"void",...gr(r)})}function aon(n,r){return new n({type:"date",...gr(r)})}function son(n,r){return new n({type:"date",coerce:!0,...gr(r)})}function con(n,r){return new n({type:"nan",...gr(r)})}function ZB(n,r){return new EXe({check:"less_than",...gr(r),value:n,inclusive:!1})}function Ux(n,r){return new EXe({check:"less_than",...gr(r),value:n,inclusive:!0})}function QB(n,r){return new SXe({check:"greater_than",...gr(r),value:n,inclusive:!1})}function s2(n,r){return new SXe({check:"greater_than",...gr(r),value:n,inclusive:!0})}function uon(n){return QB(0,n)}function lon(n){return ZB(0,n)}function don(n){return Ux(0,n)}function fon(n){return s2(0,n)}function Gie(n,r){return new srn({check:"multiple_of",...gr(r),value:n})}function Wye(n,r){return new lrn({check:"max_size",...gr(r),maximum:n})}function qie(n,r){return new drn({check:"min_size",...gr(r),minimum:n})}function eJe(n,r){return new frn({check:"size_equals",...gr(r),size:n})}function Kye(n,r){return new hrn({check:"max_length",...gr(r),maximum:n})}function QW(n,r){return new prn({check:"min_length",...gr(r),minimum:n})}function Yye(n,r){return new grn({check:"length_equals",...gr(r),length:n})}function tJe(n,r){return new mrn({check:"string_format",format:"regex",...gr(r),pattern:n})}function nJe(n){return new brn({check:"string_format",format:"lowercase",...gr(n)})}function rJe(n){return new vrn({check:"string_format",format:"uppercase",...gr(n)})}function iJe(n,r){return new wrn({check:"string_format",format:"includes",...gr(r),includes:n})}function oJe(n,r){return new yrn({check:"string_format",format:"starts_with",...gr(r),prefix:n})}function aJe(n,r){return new Ern({check:"string_format",format:"ends_with",...gr(r),suffix:n})}function hon(n,r,o){return new Srn({check:"property",property:n,schema:r,...gr(o)})}function sJe(n,r){return new _rn({check:"mime_type",mime:n,...gr(r)})}function fU(n){return new krn({check:"overwrite",tx:n})}function cJe(n){return fU(r=>r.normalize(n))}function uJe(){return fU(n=>n.trim())}function lJe(){return fU(n=>n.toLowerCase())}function dJe(){return fU(n=>n.toUpperCase())}function fJe(n,r,o){return new n({type:"array",element:r,...gr(o)})}function scr(n,r,o){return new n({type:"union",options:r,...gr(o)})}function ccr(n,r,o,s){return new n({type:"union",options:o,discriminator:r,...gr(s)})}function ucr(n,r,o){return new n({type:"intersection",left:r,right:o})}function pon(n,r,o,s){const c=o instanceof Qa,d=c?s:o,h=c?o:null;return new n({type:"tuple",items:r,rest:h,...gr(d)})}function lcr(n,r,o,s){return new n({type:"record",keyType:r,valueType:o,...gr(s)})}function dcr(n,r,o,s){return new n({type:"map",keyType:r,valueType:o,...gr(s)})}function fcr(n,r,o){return new n({type:"set",valueType:r,...gr(o)})}function hcr(n,r,o){const s=Array.isArray(r)?Object.fromEntries(r.map(c=>[c,c])):r;return new n({type:"enum",entries:s,...gr(o)})}function pcr(n,r,o){return new n({type:"enum",entries:r,...gr(o)})}function gcr(n,r,o){return new n({type:"literal",values:Array.isArray(r)?r:[r],...gr(o)})}function gon(n,r){return new n({type:"file",...gr(r)})}function mcr(n,r){return new n({type:"transform",transform:r})}function bcr(n,r){return new n({type:"optional",innerType:r})}function vcr(n,r){return new n({type:"nullable",innerType:r})}function wcr(n,r,o){return new n({type:"default",innerType:r,get defaultValue(){return typeof o=="function"?o():o}})}function ycr(n,r,o){return new n({type:"nonoptional",innerType:r,...gr(o)})}function Ecr(n,r){return new n({type:"success",innerType:r})}function Scr(n,r,o){return new n({type:"catch",innerType:r,catchValue:typeof o=="function"?o:()=>o})}function _cr(n,r,o){return new n({type:"pipe",in:r,out:o})}function kcr(n,r){return new n({type:"readonly",innerType:r})}function Tcr(n,r,o){return new n({type:"template_literal",parts:r,...gr(o)})}function xcr(n,r){return new n({type:"lazy",getter:r})}function Ccr(n,r){return new n({type:"promise",innerType:r})}function mon(n,r,o){const s=gr(o);return s.abort??(s.abort=!0),new n({type:"custom",check:"custom",fn:r,...s})}function bon(n,r,o){return new n({type:"custom",check:"custom",fn:r,...gr(o)})}function von(n){const r=won(o=>(o.addIssue=s=>{if(typeof s=="string")o.issues.push(JW(s,o.value,r._zod.def));else{const c=s;c.fatal&&(c.continue=!1),c.code??(c.code="custom"),c.input??(c.input=o.value),c.inst??(c.inst=r),c.continue??(c.continue=!r._zod.def.abort),o.issues.push(JW(c))}},n(o.value,o)));return r}function won(n,r){const o=new Ph({check:"custom",...gr(r)});return o._zod.check=n,o}function yon(n,r){const o=gr(r);let s=o.truthy??["true","1","yes","on","y","enabled"],c=o.falsy??["false","0","no","off","n","disabled"];o.case!=="sensitive"&&(s=s.map(D=>typeof D=="string"?D.toLowerCase():D),c=c.map(D=>typeof D=="string"?D.toLowerCase():D));const d=new Set(s),h=new Set(c),b=n.Pipe??OXe,y=n.Boolean??TXe,E=n.String??Foe,k=n.Transform??IXe,x=new k({type:"transform",transform:(D,$)=>{let U=D;return o.case!=="sensitive"&&(U=U.toLowerCase()),d.has(U)?!0:h.has(U)?!1:($.issues.push({code:"invalid_value",expected:"stringbool",values:[...d,...h],input:$.value,inst:x,continue:!1}),{})},error:o.error}),O=new b({type:"pipe",in:new E({type:"string",error:o.error}),out:x,error:o.error});return new b({type:"pipe",in:O,out:new y({type:"boolean",error:o.error}),error:o.error})}function hJe(n,r,o,s={}){const c=gr(s),d={...gr(s),check:"string_format",type:"string",format:r,fn:typeof o=="function"?o:b=>o.test(b),...c};return o instanceof RegExp&&(d.pattern=o),new n(d)}class Eon{constructor(r){this._def=r,this.def=r}implement(r){if(typeof r!="function")throw new Error("implement() must be called with a function");const o=(...s)=>{const c=this._def.input?HVe(this._def.input,s,void 0,{callee:o}):s;if(!Array.isArray(c))throw new Error("Invalid arguments schema: not an array or tuple schema.");const d=r(...c);return this._def.output?HVe(this._def.output,d,void 0,{callee:o}):d};return o}implementAsync(r){if(typeof r!="function")throw new Error("implement() must be called with a function");const o=async(...s)=>{const c=this._def.input?await GVe(this._def.input,s,void 0,{callee:o}):s;if(!Array.isArray(c))throw new Error("Invalid arguments schema: not an array or tuple schema.");const d=await r(...c);return this._def.output?GVe(this._def.output,d,void 0,{callee:o}):d};return o}input(...r){const o=this.constructor;return Array.isArray(r[0])?new o({type:"function",input:new Vye({type:"tuple",items:r[0],rest:r[1]}),output:this._def.output}):new o({type:"function",input:r[0],output:this._def.output})}output(r){const o=this.constructor;return new o({type:"function",input:this._def.input,output:r})}}function Son(n){return new Eon({type:"function",input:Array.isArray(n?.input)?pon(Vye,n?.input):n?.input??fJe(CXe,uwe(swe)),output:n?.output??uwe(swe)})}class VVe{constructor(r){this.counter=0,this.metadataRegistry=r?.metadata??PB,this.target=r?.target??"draft-2020-12",this.unrepresentable=r?.unrepresentable??"throw",this.override=r?.override??(()=>{}),this.io=r?.io??"output",this.seen=new Map}process(r,o={path:[],schemaPath:[]}){var s;const c=r._zod.def,d={guid:"uuid",url:"uri",datetime:"date-time",json_string:"json-string",regex:""},h=this.seen.get(r);if(h)return h.count++,o.schemaPath.includes(r)&&(h.cycle=o.path),h.schema;const b={schema:{},count:1,cycle:void 0,path:o.path};this.seen.set(r,b);const y=r._zod.toJSONSchema?.();if(y)b.schema=y;else{const x={...o,schemaPath:[...o.schemaPath,r],path:o.path},O=r._zod.parent;if(O)b.ref=O,this.process(O,x),this.seen.get(O).isParent=!0;else{const P=b.schema;switch(c.type){case"string":{const D=P;D.type="string";const{minimum:$,maximum:U,format:B,patterns:G,contentEncoding:K}=r._zod.bag;if(typeof $=="number"&&(D.minLength=$),typeof U=="number"&&(D.maxLength=U),B&&(D.format=d[B]??B,D.format===""&&delete D.format),K&&(D.contentEncoding=K),G&&G.size>0){const X=[...G];X.length===1?D.pattern=X[0].source:X.length>1&&(b.schema.allOf=[...X.map(ae=>({...this.target==="draft-7"||this.target==="draft-4"?{type:"string"}:{},pattern:ae.source}))])}break}case"number":{const D=P,{minimum:$,maximum:U,format:B,multipleOf:G,exclusiveMaximum:K,exclusiveMinimum:X}=r._zod.bag;typeof B=="string"&&B.includes("int")?D.type="integer":D.type="number",typeof X=="number"&&(this.target==="draft-4"?(D.minimum=X,D.exclusiveMinimum=!0):D.exclusiveMinimum=X),typeof $=="number"&&(D.minimum=$,typeof X=="number"&&this.target!=="draft-4"&&(X>=$?delete D.minimum:delete D.exclusiveMinimum)),typeof K=="number"&&(this.target==="draft-4"?(D.maximum=K,D.exclusiveMaximum=!0):D.exclusiveMaximum=K),typeof U=="number"&&(D.maximum=U,typeof K=="number"&&this.target!=="draft-4"&&(K<=U?delete D.maximum:delete D.exclusiveMaximum)),typeof G=="number"&&(D.multipleOf=G);break}case"boolean":{const D=P;D.type="boolean";break}case"bigint":{if(this.unrepresentable==="throw")throw new Error("BigInt cannot be represented in JSON Schema");break}case"symbol":{if(this.unrepresentable==="throw")throw new Error("Symbols cannot be represented in JSON Schema");break}case"null":{P.type="null";break}case"any":break;case"unknown":break;case"undefined":{if(this.unrepresentable==="throw")throw new Error("Undefined cannot be represented in JSON Schema");break}case"void":{if(this.unrepresentable==="throw")throw new Error("Void cannot be represented in JSON Schema");break}case"never":{P.not={};break}case"date":{if(this.unrepresentable==="throw")throw new Error("Date cannot be represented in JSON Schema");break}case"array":{const D=P,{minimum:$,maximum:U}=r._zod.bag;typeof $=="number"&&(D.minItems=$),typeof U=="number"&&(D.maxItems=U),D.type="array",D.items=this.process(c.element,{...x,path:[...x.path,"items"]});break}case"object":{const D=P;D.type="object",D.properties={};const $=c.shape;for(const G in $)D.properties[G]=this.process($[G],{...x,path:[...x.path,"properties",G]});const U=new Set(Object.keys($)),B=new Set([...U].filter(G=>{const K=c.shape[G]._zod;return this.io==="input"?K.optin===void 0:K.optout===void 0}));B.size>0&&(D.required=Array.from(B)),c.catchall?._zod.def.type==="never"?D.additionalProperties=!1:c.catchall?c.catchall&&(D.additionalProperties=this.process(c.catchall,{...x,path:[...x.path,"additionalProperties"]})):this.io==="output"&&(D.additionalProperties=!1);break}case"union":{const D=P;D.anyOf=c.options.map(($,U)=>this.process($,{...x,path:[...x.path,"anyOf",U]}));break}case"intersection":{const D=P,$=this.process(c.left,{...x,path:[...x.path,"allOf",0]}),U=this.process(c.right,{...x,path:[...x.path,"allOf",1]}),B=K=>"allOf"in K&&Object.keys(K).length===1,G=[...B($)?$.allOf:[$],...B(U)?U.allOf:[U]];D.allOf=G;break}case"tuple":{const D=P;D.type="array";const $=c.items.map((G,K)=>this.process(G,{...x,path:[...x.path,"prefixItems",K]}));if(this.target==="draft-2020-12"?D.prefixItems=$:D.items=$,c.rest){const G=this.process(c.rest,{...x,path:[...x.path,"items"]});this.target==="draft-2020-12"?D.items=G:D.additionalItems=G}c.rest&&(D.items=this.process(c.rest,{...x,path:[...x.path,"items"]}));const{minimum:U,maximum:B}=r._zod.bag;typeof U=="number"&&(D.minItems=U),typeof B=="number"&&(D.maxItems=B);break}case"record":{const D=P;D.type="object",this.target!=="draft-4"&&(D.propertyNames=this.process(c.keyType,{...x,path:[...x.path,"propertyNames"]})),D.additionalProperties=this.process(c.valueType,{...x,path:[...x.path,"additionalProperties"]});break}case"map":{if(this.unrepresentable==="throw")throw new Error("Map cannot be represented in JSON Schema");break}case"set":{if(this.unrepresentable==="throw")throw new Error("Set cannot be represented in JSON Schema");break}case"enum":{const D=P,$=lXe(c.entries);$.every(U=>typeof U=="number")&&(D.type="number"),$.every(U=>typeof U=="string")&&(D.type="string"),D.enum=$;break}case"literal":{const D=P,$=[];for(const U of c.values)if(U===void 0){if(this.unrepresentable==="throw")throw new Error("Literal `undefined` cannot be represented in JSON Schema")}else if(typeof U=="bigint"){if(this.unrepresentable==="throw")throw new Error("BigInt literals cannot be represented in JSON Schema");$.push(Number(U))}else $.push(U);if($.length!==0)if($.length===1){const U=$[0];D.type=U===null?"null":typeof U,this.target==="draft-4"?D.enum=[U]:D.const=U}else $.every(U=>typeof U=="number")&&(D.type="number"),$.every(U=>typeof U=="string")&&(D.type="string"),$.every(U=>typeof U=="boolean")&&(D.type="string"),$.every(U=>U===null)&&(D.type="null"),D.enum=$;break}case"file":{const D=P,$={type:"string",format:"binary",contentEncoding:"binary"},{minimum:U,maximum:B,mime:G}=r._zod.bag;U!==void 0&&($.minLength=U),B!==void 0&&($.maxLength=B),G?G.length===1?($.contentMediaType=G[0],Object.assign(D,$)):D.anyOf=G.map(K=>({...$,contentMediaType:K})):Object.assign(D,$);break}case"transform":{if(this.unrepresentable==="throw")throw new Error("Transforms cannot be represented in JSON Schema");break}case"nullable":{const D=this.process(c.innerType,x);P.anyOf=[D,{type:"null"}];break}case"nonoptional":{this.process(c.innerType,x),b.ref=c.innerType;break}case"success":{const D=P;D.type="boolean";break}case"default":{this.process(c.innerType,x),b.ref=c.innerType,P.default=JSON.parse(JSON.stringify(c.defaultValue));break}case"prefault":{this.process(c.innerType,x),b.ref=c.innerType,this.io==="input"&&(P._prefault=JSON.parse(JSON.stringify(c.defaultValue)));break}case"catch":{this.process(c.innerType,x),b.ref=c.innerType;let D;try{D=c.catchValue(void 0)}catch{throw new Error("Dynamic catch values are not supported in JSON Schema")}P.default=D;break}case"nan":{if(this.unrepresentable==="throw")throw new Error("NaN cannot be represented in JSON Schema");break}case"template_literal":{const D=P,$=r._zod.pattern;if(!$)throw new Error("Pattern not found in template literal");D.type="string",D.pattern=$.source;break}case"pipe":{const D=this.io==="input"?c.in._zod.def.type==="transform"?c.out:c.in:c.out;this.process(D,x),b.ref=D;break}case"readonly":{this.process(c.innerType,x),b.ref=c.innerType,P.readOnly=!0;break}case"promise":{this.process(c.innerType,x),b.ref=c.innerType;break}case"optional":{this.process(c.innerType,x),b.ref=c.innerType;break}case"lazy":{const D=r._zod.innerType;this.process(D,x),b.ref=D;break}case"custom":{if(this.unrepresentable==="throw")throw new Error("Custom types cannot be represented in JSON Schema");break}}}}const E=this.metadataRegistry.get(r);return E&&Object.assign(b.schema,E),this.io==="input"&&Hg(r)&&(delete b.schema.examples,delete b.schema.default),this.io==="input"&&b.schema._prefault&&((s=b.schema).default??(s.default=b.schema._prefault)),delete b.schema._prefault,this.seen.get(r).schema}emit(r,o){const s={cycles:o?.cycles??"ref",reused:o?.reused??"inline",external:o?.external??void 0},c=this.seen.get(r);if(!c)throw new Error("Unprocessed schema. This is a bug in Zod.");const d=k=>{const x=this.target==="draft-2020-12"?"$defs":"definitions";if(s.external){const $=s.external.registry.get(k[0])?.id,U=s.external.uri??(G=>G);if($)return{ref:U($)};const B=k[1].defId??k[1].schema.id??`schema${this.counter++}`;return k[1].defId=B,{defId:B,ref:`${U("__shared")}#/${x}/${B}`}}if(k[1]===c)return{ref:"#"};const P=`#/${x}/`,D=k[1].schema.id??`__schema${this.counter++}`;return{defId:D,ref:P+D}},h=k=>{if(k[1].schema.$ref)return;const x=k[1],{ref:O,defId:P}=d(k);x.def={...x.schema},P&&(x.defId=P);const D=x.schema;for(const $ in D)delete D[$];D.$ref=O};if(s.cycles==="throw")for(const k of this.seen.entries()){const x=k[1];if(x.cycle)throw new Error(`Cycle detected: #/${x.cycle?.join("/")}/<root>

Set the \`cycles\` parameter to \`"ref"\` to resolve cyclical schemas with defs.`)}for(const k of this.seen.entries()){const x=k[1];if(r===k[0]){h(k);continue}if(s.external){const P=s.external.registry.get(k[0])?.id;if(r!==k[0]&&P){h(k);continue}}if(this.metadataRegistry.get(k[0])?.id){h(k);continue}if(x.cycle){h(k);continue}if(x.count>1&&s.reused==="ref"){h(k);continue}}const b=(k,x)=>{const O=this.seen.get(k),P=O.def??O.schema,D={...P};if(O.ref===null)return;const $=O.ref;if(O.ref=null,$){b($,x);const U=this.seen.get($).schema;U.$ref&&(x.target==="draft-7"||x.target==="draft-4")?(P.allOf=P.allOf??[],P.allOf.push(U)):(Object.assign(P,U),Object.assign(P,D))}O.isParent||this.override({zodSchema:k,jsonSchema:P,path:O.path??[]})};for(const k of[...this.seen.entries()].reverse())b(k[0],{target:this.target});const y={};if(this.target==="draft-2020-12"?y.$schema="https://json-schema.org/draft/2020-12/schema":this.target==="draft-7"?y.$schema="http://json-schema.org/draft-07/schema#":this.target==="draft-4"?y.$schema="http://json-schema.org/draft-04/schema#":console.warn(`Invalid target: ${this.target}`),s.external?.uri){const k=s.external.registry.get(r)?.id;if(!k)throw new Error("Schema is missing an `id` property");y.$id=s.external.uri(k)}Object.assign(y,c.def);const E=s.external?.defs??{};for(const k of this.seen.entries()){const x=k[1];x.def&&x.defId&&(E[x.defId]=x.def)}s.external||Object.keys(E).length>0&&(this.target==="draft-2020-12"?y.$defs=E:y.definitions=E);try{return JSON.parse(JSON.stringify(y))}catch{throw new Error("Error converting schema to JSON.")}}}function _on(n,r){if(n instanceof NXe){const s=new VVe(r),c={};for(const b of n._idmap.entries()){const[y,E]=b;s.process(E)}const d={},h={registry:n,uri:r?.uri,defs:c};for(const b of n._idmap.entries()){const[y,E]=b;d[y]=s.emit(E,{...r,external:h})}if(Object.keys(c).length>0){const b=s.target==="draft-2020-12"?"$defs":"definitions";d.__shared={[b]:c}}return{schemas:d}}const o=new VVe(r);return o.process(n),o.emit(n,r)}function Hg(n,r){const o=r??{seen:new Set};if(o.seen.has(n))return!1;o.seen.add(n);const c=n._zod.def;switch(c.type){case"string":case"number":case"bigint":case"boolean":case"date":case"symbol":case"undefined":case"null":case"any":case"unknown":case"never":case"void":case"literal":case"enum":case"nan":case"file":case"template_literal":return!1;case"array":return Hg(c.element,o);case"object":{for(const d in c.shape)if(Hg(c.shape[d],o))return!0;return!1}case"union":{for(const d of c.options)if(Hg(d,o))return!0;return!1}case"intersection":return Hg(c.left,o)||Hg(c.right,o);case"tuple":{for(const d of c.items)if(Hg(d,o))return!0;return!!(c.rest&&Hg(c.rest,o))}case"record":return Hg(c.keyType,o)||Hg(c.valueType,o);case"map":return Hg(c.keyType,o)||Hg(c.valueType,o);case"set":return Hg(c.valueType,o);case"promise":case"optional":case"nonoptional":case"nullable":case"readonly":return Hg(c.innerType,o);case"lazy":return Hg(c.getter(),o);case"default":return Hg(c.innerType,o);case"prefault":return Hg(c.innerType,o);case"custom":return!1;case"transform":return!0;case"pipe":return Hg(c.in,o)||Hg(c.out,o);case"success":return!1;case"catch":return!1}throw new Error(`Unknown schema type: ${c.type}`)}const Acr=Object.freeze(Object.defineProperty({__proto__:null},Symbol.toStringTag,{value:"Module"})),Icr=Object.freeze(Object.defineProperty({__proto__:null,$ZodAny:iin,$ZodArray:CXe,$ZodAsyncError:XW,$ZodBase64:Vrn,$ZodBase64URL:Krn,$ZodBigInt:xXe,$ZodBigIntFormat:ein,$ZodBoolean:TXe,$ZodCIDRv4:Grn,$ZodCIDRv6:qrn,$ZodCUID:Prn,$ZodCUID2:Lrn,$ZodCatch:_in,$ZodCheck:Ph,$ZodCheckBigIntFormat:urn,$ZodCheckEndsWith:Ern,$ZodCheckGreaterThan:SXe,$ZodCheckIncludes:wrn,$ZodCheckLengthEquals:grn,$ZodCheckLessThan:EXe,$ZodCheckLowerCase:brn,$ZodCheckMaxLength:hrn,$ZodCheckMaxSize:lrn,$ZodCheckMimeType:_rn,$ZodCheckMinLength:prn,$ZodCheckMinSize:drn,$ZodCheckMultipleOf:srn,$ZodCheckNumberFormat:crn,$ZodCheckOverwrite:krn,$ZodCheckProperty:Srn,$ZodCheckRegex:mrn,$ZodCheckSizeEquals:frn,$ZodCheckStartsWith:yrn,$ZodCheckStringFormat:$oe,$ZodCheckUpperCase:vrn,$ZodCustom:Iin,$ZodCustomStringFormat:Zrn,$ZodDate:sin,$ZodDefault:win,$ZodDiscriminatedUnion:uin,$ZodE164:Yrn,$ZodEmail:Irn,$ZodEmoji:Nrn,$ZodEnum:pin,$ZodError:fXe,$ZodFile:min,$ZodFunction:Eon,$ZodGUID:Crn,$ZodIPv4:zrn,$ZodIPv6:Hrn,$ZodISODate:Frn,$ZodISODateTime:$rn,$ZodISODuration:Urn,$ZodISOTime:Brn,$ZodIntersection:lin,$ZodJWT:Jrn,$ZodKSUID:jrn,$ZodLazy:Ain,$ZodLiteral:gin,$ZodMap:fin,$ZodNaN:kin,$ZodNanoID:Rrn,$ZodNever:oin,$ZodNonOptional:Ein,$ZodNull:rin,$ZodNullable:vin,$ZodNumber:kXe,$ZodNumberFormat:Qrn,$ZodObject:cin,$ZodOptional:bin,$ZodPipe:OXe,$ZodPrefault:yin,$ZodPromise:Cin,$ZodReadonly:Tin,$ZodRealError:joe,$ZodRecord:din,$ZodRegistry:NXe,$ZodSet:hin,$ZodString:Foe,$ZodStringFormat:Ad,$ZodSuccess:Sin,$ZodSymbol:tin,$ZodTemplateLiteral:xin,$ZodTransform:IXe,$ZodTuple:Vye,$ZodType:Qa,$ZodULID:Drn,$ZodURL:Orn,$ZodUUID:Arn,$ZodUndefined:nin,$ZodUnion:AXe,$ZodUnknown:swe,$ZodVoid:ain,$ZodXID:Mrn,$brand:lnn,$constructor:rn,$input:Pin,$output:Rin,Doc:Trn,JSONSchema:Acr,JSONSchemaGenerator:VVe,NEVER:unn,TimePrecision:Min,_any:ron,_array:fJe,_base64:XXe,_base64url:JXe,_bigint:Xin,_boolean:Kin,_catch:Scr,_check:won,_cidrv4:KXe,_cidrv6:YXe,_coercedBigint:Jin,_coercedBoolean:Yin,_coercedDate:son,_coercedNumber:zin,_coercedString:Din,_cuid:UXe,_cuid2:zXe,_custom:mon,_date:aon,_default:wcr,_discriminatedUnion:ccr,_e164:ZXe,_email:PXe,_emoji:FXe,_endsWith:aJe,_enum:hcr,_file:gon,_float32:Gin,_float64:qin,_gt:QB,_gte:s2,_guid:cwe,_includes:iJe,_int:Hin,_int32:Vin,_int64:Zin,_intersection:ucr,_ipv4:VXe,_ipv6:WXe,_isoDate:$in,_isoDateTime:jin,_isoDuration:Bin,_isoTime:Fin,_jwt:QXe,_ksuid:qXe,_lazy:xcr,_length:Yye,_literal:gcr,_lowercase:nJe,_lt:ZB,_lte:Ux,_map:dcr,_max:Ux,_maxLength:Kye,_maxSize:Wye,_mime:sJe,_min:s2,_minLength:QW,_minSize:qie,_multipleOf:Gie,_nan:con,_nanoid:BXe,_nativeEnum:pcr,_negative:lon,_never:ion,_nonnegative:fon,_nonoptional:ycr,_nonpositive:don,_normalize:cJe,_null:non,_nullable:vcr,_number:Uin,_optional:bcr,_overwrite:fU,_parse:gXe,_parseAsync:mXe,_pipe:_cr,_positive:uon,_promise:Ccr,_property:hon,_readonly:kcr,_record:lcr,_refine:bon,_regex:tJe,_safeParse:bXe,_safeParseAsync:vXe,_set:fcr,_size:eJe,_startsWith:oJe,_string:Lin,_stringFormat:hJe,_stringbool:yon,_success:Ecr,_superRefine:von,_symbol:eon,_templateLiteral:Tcr,_toLowerCase:lJe,_toUpperCase:dJe,_transform:mcr,_trim:uJe,_tuple:pon,_uint32:Win,_uint64:Qin,_ulid:HXe,_undefined:ton,_union:scr,_unknown:uwe,_uppercase:rJe,_url:$Xe,_uuid:LXe,_uuidv4:DXe,_uuidv6:MXe,_uuidv7:jXe,_void:oon,_xid:GXe,clone:eC,config:cv,flattenError:hXe,formatError:pXe,function:Son,globalConfig:iwe,globalRegistry:PB,isValidBase64:_Xe,isValidBase64URL:Wrn,isValidJWT:Xrn,locales:Nin,parse:HVe,parseAsync:GVe,prettifyError:xnn,regexes:orn,registry:RXe,safeParse:Cnn,safeParseAsync:Ann,toDotPath:Tnn,toJSONSchema:_on,treeifyError:knn,util:gar,version:xrn},Symbol.toStringTag,{value:"Module"})),pJe=rn("ZodISODateTime",(n,r)=>{$rn.init(n,r),Wd.init(n,r)});function kon(n){return jin(pJe,n)}const gJe=rn("ZodISODate",(n,r)=>{Frn.init(n,r),Wd.init(n,r)});function Ton(n){return $in(gJe,n)}const mJe=rn("ZodISOTime",(n,r)=>{Brn.init(n,r),Wd.init(n,r)});function xon(n){return Fin(mJe,n)}const bJe=rn("ZodISODuration",(n,r)=>{Urn.init(n,r),Wd.init(n,r)});function Con(n){return Bin(bJe,n)}const Ocr=Object.freeze(Object.defineProperty({__proto__:null,ZodISODate:gJe,ZodISODateTime:pJe,ZodISODuration:bJe,ZodISOTime:mJe,date:Ton,datetime:kon,duration:Con,time:xon},Symbol.toStringTag,{value:"Module"})),Aon=(n,r)=>{fXe.init(n,r),n.name="ZodError",Object.defineProperties(n,{format:{value:o=>pXe(n,o)},flatten:{value:o=>hXe(n,o)},addIssue:{value:o=>{n.issues.push(o),n.message=JSON.stringify(n.issues,owe,2)}},addIssues:{value:o=>{n.issues.push(...o),n.message=JSON.stringify(n.issues,owe,2)}},isEmpty:{get(){return n.issues.length===0}}})},Ncr=rn("ZodError",Aon),Boe=rn("ZodError",Aon,{Parent:Error}),Ion=gXe(Boe),Oon=mXe(Boe),Non=bXe(Boe),Ron=vXe(Boe),Vs=rn("ZodType",(n,r)=>(Qa.init(n,r),n.def=r,Object.defineProperty(n,"_def",{value:r}),n.check=(...o)=>n.clone({...r,checks:[...r.checks??[],...o.map(s=>typeof s=="function"?{_zod:{check:s,def:{check:"custom"},onattach:[]}}:s)]}),n.clone=(o,s)=>eC(n,o,s),n.brand=()=>n,n.register=(o,s)=>(o.add(n,s),n),n.parse=(o,s)=>Ion(n,o,s,{callee:n.parse}),n.safeParse=(o,s)=>Non(n,o,s),n.parseAsync=async(o,s)=>Oon(n,o,s,{callee:n.parseAsync}),n.safeParseAsync=async(o,s)=>Ron(n,o,s),n.spa=n.safeParseAsync,n.refine=(o,s)=>n.check(dan(o,s)),n.superRefine=o=>n.check(fan(o)),n.overwrite=o=>n.check(fU(o)),n.optional=()=>dwe(n),n.nullable=()=>fwe(n),n.nullish=()=>dwe(fwe(n)),n.nonoptional=o=>tan(n,o),n.array=()=>mp(n),n.or=o=>D4([n,o]),n.and=o=>Gon(n,o),n.transform=o=>hwe(n,GJe(o)),n.default=o=>Zon(n,o),n.prefault=o=>ean(n,o),n.catch=o=>ian(n,o),n.pipe=o=>hwe(n,o),n.readonly=()=>san(n),n.describe=o=>{const s=n.clone();return PB.add(s,{description:o}),s},Object.defineProperty(n,"description",{get(){return PB.get(n)?.description},configurable:!0}),n.meta=(...o)=>{if(o.length===0)return PB.get(n);const s=n.clone();return PB.add(s,o[0]),s},n.isOptional=()=>n.safeParse(void 0).success,n.isNullable=()=>n.safeParse(null).success,n)),vJe=rn("_ZodString",(n,r)=>{Foe.init(n,r),Vs.init(n,r);const o=n._zod.bag;n.format=o.format??null,n.minLength=o.minimum??null,n.maxLength=o.maximum??null,n.regex=(...s)=>n.check(tJe(...s)),n.includes=(...s)=>n.check(iJe(...s)),n.startsWith=(...s)=>n.check(oJe(...s)),n.endsWith=(...s)=>n.check(aJe(...s)),n.min=(...s)=>n.check(QW(...s)),n.max=(...s)=>n.check(Kye(...s)),n.length=(...s)=>n.check(Yye(...s)),n.nonempty=(...s)=>n.check(QW(1,...s)),n.lowercase=s=>n.check(nJe(s)),n.uppercase=s=>n.check(rJe(s)),n.trim=()=>n.check(uJe()),n.normalize=(...s)=>n.check(cJe(...s)),n.toLowerCase=()=>n.check(lJe()),n.toUpperCase=()=>n.check(dJe())}),Xye=rn("ZodString",(n,r)=>{Foe.init(n,r),vJe.init(n,r),n.email=o=>n.check(PXe(wJe,o)),n.url=o=>n.check($Xe(yJe,o)),n.jwt=o=>n.check(QXe(DJe,o)),n.emoji=o=>n.check(FXe(EJe,o)),n.guid=o=>n.check(cwe(lwe,o)),n.uuid=o=>n.check(LXe(L4,o)),n.uuidv4=o=>n.check(DXe(L4,o)),n.uuidv6=o=>n.check(MXe(L4,o)),n.uuidv7=o=>n.check(jXe(L4,o)),n.nanoid=o=>n.check(BXe(SJe,o)),n.guid=o=>n.check(cwe(lwe,o)),n.cuid=o=>n.check(UXe(_Je,o)),n.cuid2=o=>n.check(zXe(kJe,o)),n.ulid=o=>n.check(HXe(TJe,o)),n.base64=o=>n.check(XXe(RJe,o)),n.base64url=o=>n.check(JXe(PJe,o)),n.xid=o=>n.check(GXe(xJe,o)),n.ksuid=o=>n.check(qXe(CJe,o)),n.ipv4=o=>n.check(VXe(AJe,o)),n.ipv6=o=>n.check(WXe(IJe,o)),n.cidrv4=o=>n.check(KXe(OJe,o)),n.cidrv6=o=>n.check(YXe(NJe,o)),n.e164=o=>n.check(ZXe(LJe,o)),n.datetime=o=>n.check(kon(o)),n.date=o=>n.check(Ton(o)),n.time=o=>n.check(xon(o)),n.duration=o=>n.check(Con(o))});function Tr(n){return Lin(Xye,n)}const Wd=rn("ZodStringFormat",(n,r)=>{Ad.init(n,r),vJe.init(n,r)}),wJe=rn("ZodEmail",(n,r)=>{Irn.init(n,r),Wd.init(n,r)});function Rcr(n){return PXe(wJe,n)}const lwe=rn("ZodGUID",(n,r)=>{Crn.init(n,r),Wd.init(n,r)});function Pcr(n){return cwe(lwe,n)}const L4=rn("ZodUUID",(n,r)=>{Arn.init(n,r),Wd.init(n,r)});function Lcr(n){return LXe(L4,n)}function Dcr(n){return DXe(L4,n)}function Mcr(n){return MXe(L4,n)}function jcr(n){return jXe(L4,n)}const yJe=rn("ZodURL",(n,r)=>{Orn.init(n,r),Wd.init(n,r)});function $cr(n){return $Xe(yJe,n)}const EJe=rn("ZodEmoji",(n,r)=>{Nrn.init(n,r),Wd.init(n,r)});function Fcr(n){return FXe(EJe,n)}const SJe=rn("ZodNanoID",(n,r)=>{Rrn.init(n,r),Wd.init(n,r)});function Bcr(n){return BXe(SJe,n)}const _Je=rn("ZodCUID",(n,r)=>{Prn.init(n,r),Wd.init(n,r)});function Ucr(n){return UXe(_Je,n)}const kJe=rn("ZodCUID2",(n,r)=>{Lrn.init(n,r),Wd.init(n,r)});function zcr(n){return zXe(kJe,n)}const TJe=rn("ZodULID",(n,r)=>{Drn.init(n,r),Wd.init(n,r)});function Hcr(n){return HXe(TJe,n)}const xJe=rn("ZodXID",(n,r)=>{Mrn.init(n,r),Wd.init(n,r)});function Gcr(n){return GXe(xJe,n)}const CJe=rn("ZodKSUID",(n,r)=>{jrn.init(n,r),Wd.init(n,r)});function qcr(n){return qXe(CJe,n)}const AJe=rn("ZodIPv4",(n,r)=>{zrn.init(n,r),Wd.init(n,r)});function Vcr(n){return VXe(AJe,n)}const IJe=rn("ZodIPv6",(n,r)=>{Hrn.init(n,r),Wd.init(n,r)});function Wcr(n){return WXe(IJe,n)}const OJe=rn("ZodCIDRv4",(n,r)=>{Grn.init(n,r),Wd.init(n,r)});function Kcr(n){return KXe(OJe,n)}const NJe=rn("ZodCIDRv6",(n,r)=>{qrn.init(n,r),Wd.init(n,r)});function Ycr(n){return YXe(NJe,n)}const RJe=rn("ZodBase64",(n,r)=>{Vrn.init(n,r),Wd.init(n,r)});function Xcr(n){return XXe(RJe,n)}const PJe=rn("ZodBase64URL",(n,r)=>{Krn.init(n,r),Wd.init(n,r)});function Jcr(n){return JXe(PJe,n)}const LJe=rn("ZodE164",(n,r)=>{Yrn.init(n,r),Wd.init(n,r)});function Zcr(n){return ZXe(LJe,n)}const DJe=rn("ZodJWT",(n,r)=>{Jrn.init(n,r),Wd.init(n,r)});function Qcr(n){return QXe(DJe,n)}const MJe=rn("ZodCustomStringFormat",(n,r)=>{Zrn.init(n,r),Wd.init(n,r)});function eur(n,r,o={}){return hJe(MJe,n,r,o)}function tur(n){return hJe(MJe,"hostname",yXe,n)}const Jye=rn("ZodNumber",(n,r)=>{kXe.init(n,r),Vs.init(n,r),n.gt=(s,c)=>n.check(QB(s,c)),n.gte=(s,c)=>n.check(s2(s,c)),n.min=(s,c)=>n.check(s2(s,c)),n.lt=(s,c)=>n.check(ZB(s,c)),n.lte=(s,c)=>n.check(Ux(s,c)),n.max=(s,c)=>n.check(Ux(s,c)),n.int=s=>n.check(WVe(s)),n.safe=s=>n.check(WVe(s)),n.positive=s=>n.check(QB(0,s)),n.nonnegative=s=>n.check(s2(0,s)),n.negative=s=>n.check(ZB(0,s)),n.nonpositive=s=>n.check(Ux(0,s)),n.multipleOf=(s,c)=>n.check(Gie(s,c)),n.step=(s,c)=>n.check(Gie(s,c)),n.finite=()=>n;const o=n._zod.bag;n.minValue=Math.max(o.minimum??Number.NEGATIVE_INFINITY,o.exclusiveMinimum??Number.NEGATIVE_INFINITY)??null,n.maxValue=Math.min(o.maximum??Number.POSITIVE_INFINITY,o.exclusiveMaximum??Number.POSITIVE_INFINITY)??null,n.isInt=(o.format??"").includes("int")||Number.isSafeInteger(o.multipleOf??.5),n.isFinite=!0,n.format=o.format??null});function G4(n){return Uin(Jye,n)}const RK=rn("ZodNumberFormat",(n,r)=>{Qrn.init(n,r),Jye.init(n,r)});function WVe(n){return Hin(RK,n)}function nur(n){return Gin(RK,n)}function rur(n){return qin(RK,n)}function iur(n){return Vin(RK,n)}function our(n){return Win(RK,n)}const Zye=rn("ZodBoolean",(n,r)=>{TXe.init(n,r),Vs.init(n,r)});function V1(n){return Kin(Zye,n)}const Qye=rn("ZodBigInt",(n,r)=>{xXe.init(n,r),Vs.init(n,r),n.gte=(s,c)=>n.check(s2(s,c)),n.min=(s,c)=>n.check(s2(s,c)),n.gt=(s,c)=>n.check(QB(s,c)),n.gte=(s,c)=>n.check(s2(s,c)),n.min=(s,c)=>n.check(s2(s,c)),n.lt=(s,c)=>n.check(ZB(s,c)),n.lte=(s,c)=>n.check(Ux(s,c)),n.max=(s,c)=>n.check(Ux(s,c)),n.positive=s=>n.check(QB(BigInt(0),s)),n.negative=s=>n.check(ZB(BigInt(0),s)),n.nonpositive=s=>n.check(Ux(BigInt(0),s)),n.nonnegative=s=>n.check(s2(BigInt(0),s)),n.multipleOf=(s,c)=>n.check(Gie(s,c));const o=n._zod.bag;n.minValue=o.minimum??null,n.maxValue=o.maximum??null,n.format=o.format??null});function aur(n){return Xin(Qye,n)}const jJe=rn("ZodBigIntFormat",(n,r)=>{ein.init(n,r),Qye.init(n,r)});function sur(n){return Zin(jJe,n)}function cur(n){return Qin(jJe,n)}const Pon=rn("ZodSymbol",(n,r)=>{tin.init(n,r),Vs.init(n,r)});function uur(n){return eon(Pon,n)}const Lon=rn("ZodUndefined",(n,r)=>{nin.init(n,r),Vs.init(n,r)});function lur(n){return ton(Lon,n)}const Don=rn("ZodNull",(n,r)=>{rin.init(n,r),Vs.init(n,r)});function $Je(n){return non(Don,n)}const Mon=rn("ZodAny",(n,r)=>{iin.init(n,r),Vs.init(n,r)});function jon(){return ron(Mon)}const $on=rn("ZodUnknown",(n,r)=>{swe.init(n,r),Vs.init(n,r)});function Vie(){return uwe($on)}const Fon=rn("ZodNever",(n,r)=>{oin.init(n,r),Vs.init(n,r)});function FJe(n){return ion(Fon,n)}const Bon=rn("ZodVoid",(n,r)=>{ain.init(n,r),Vs.init(n,r)});function dur(n){return oon(Bon,n)}const BJe=rn("ZodDate",(n,r)=>{sin.init(n,r),Vs.init(n,r),n.min=(s,c)=>n.check(s2(s,c)),n.max=(s,c)=>n.check(Ux(s,c));const o=n._zod.bag;n.minDate=o.minimum?new Date(o.minimum):null,n.maxDate=o.maximum?new Date(o.maximum):null});function fur(n){return aon(BJe,n)}const Uon=rn("ZodArray",(n,r)=>{CXe.init(n,r),Vs.init(n,r),n.element=r.element,n.min=(o,s)=>n.check(QW(o,s)),n.nonempty=o=>n.check(QW(1,o)),n.max=(o,s)=>n.check(Kye(o,s)),n.length=(o,s)=>n.check(Yye(o,s)),n.unwrap=()=>n.element});function mp(n,r){return fJe(Uon,n,r)}function hur(n){const r=n._zod.def.shape;return qI(Object.keys(r))}const eEe=rn("ZodObject",(n,r)=>{cin.init(n,r),Vs.init(n,r),du(n,"shape",()=>r.shape),n.keyof=()=>GI(Object.keys(n._zod.def.shape)),n.catchall=o=>n.clone({...n._zod.def,catchall:o}),n.passthrough=()=>n.clone({...n._zod.def,catchall:Vie()}),n.loose=()=>n.clone({...n._zod.def,catchall:Vie()}),n.strict=()=>n.clone({...n._zod.def,catchall:FJe()}),n.strip=()=>n.clone({...n._zod.def,catchall:void 0}),n.extend=o=>wnn(n,o),n.merge=o=>ynn(n,o),n.pick=o=>bnn(n,o),n.omit=o=>vnn(n,o),n.partial=(...o)=>Enn(qJe,n,o[0]),n.required=(...o)=>Snn(VJe,n,o[0])});function xs(n,r){const o={type:"object",get shape(){return HI(this,"shape",{...n}),this.shape},...gr(r)};return new eEe(o)}function pur(n,r){return new eEe({type:"object",get shape(){return HI(this,"shape",{...n}),this.shape},catchall:FJe(),...gr(r)})}function gur(n,r){return new eEe({type:"object",get shape(){return HI(this,"shape",{...n}),this.shape},catchall:Vie(),...gr(r)})}const UJe=rn("ZodUnion",(n,r)=>{AXe.init(n,r),Vs.init(n,r),n.options=r.options});function D4(n,r){return new UJe({type:"union",options:n,...gr(r)})}const zon=rn("ZodDiscriminatedUnion",(n,r)=>{UJe.init(n,r),uin.init(n,r)});function mur(n,r,o){return new zon({type:"union",options:r,discriminator:n,...gr(o)})}const Hon=rn("ZodIntersection",(n,r)=>{lin.init(n,r),Vs.init(n,r)});function Gon(n,r){return new Hon({type:"intersection",left:n,right:r})}const qon=rn("ZodTuple",(n,r)=>{Vye.init(n,r),Vs.init(n,r),n.rest=o=>n.clone({...n._zod.def,rest:o})});function bur(n,r,o){const s=r instanceof Qa,c=s?o:r,d=s?r:null;return new qon({type:"tuple",items:n,rest:d,...gr(c)})}const zJe=rn("ZodRecord",(n,r)=>{din.init(n,r),Vs.init(n,r),n.keyType=r.keyType,n.valueType=r.valueType});function hU(n,r,o){return new zJe({type:"record",keyType:n,valueType:r,...gr(o)})}function vur(n,r,o){const s=eC(n);return s._zod.values=void 0,new zJe({type:"record",keyType:s,valueType:r,...gr(o)})}const Von=rn("ZodMap",(n,r)=>{fin.init(n,r),Vs.init(n,r),n.keyType=r.keyType,n.valueType=r.valueType});function wur(n,r,o){return new Von({type:"map",keyType:n,valueType:r,...gr(o)})}const Won=rn("ZodSet",(n,r)=>{hin.init(n,r),Vs.init(n,r),n.min=(...o)=>n.check(qie(...o)),n.nonempty=o=>n.check(qie(1,o)),n.max=(...o)=>n.check(Wye(...o)),n.size=(...o)=>n.check(eJe(...o))});function yur(n,r){return new Won({type:"set",valueType:n,...gr(r)})}const Wie=rn("ZodEnum",(n,r)=>{pin.init(n,r),Vs.init(n,r),n.enum=r.entries,n.options=Object.values(r.entries);const o=new Set(Object.keys(r.entries));n.extract=(s,c)=>{const d={};for(const h of s)if(o.has(h))d[h]=r.entries[h];else throw new Error(`Key ${h} not found in enum`);return new Wie({...r,checks:[],...gr(c),entries:d})},n.exclude=(s,c)=>{const d={...r.entries};for(const h of s)if(o.has(h))delete d[h];else throw new Error(`Key ${h} not found in enum`);return new Wie({...r,checks:[],...gr(c),entries:d})}});function GI(n,r){const o=Array.isArray(n)?Object.fromEntries(n.map(s=>[s,s])):n;return new Wie({type:"enum",entries:o,...gr(r)})}function Eur(n,r){return new Wie({type:"enum",entries:n,...gr(r)})}const Kon=rn("ZodLiteral",(n,r)=>{gin.init(n,r),Vs.init(n,r),n.values=new Set(r.values),Object.defineProperty(n,"value",{get(){if(r.values.length>1)throw new Error("This schema contains multiple valid literal values. Use `.values` instead.");return r.values[0]}})});function qI(n,r){return new Kon({type:"literal",values:Array.isArray(n)?n:[n],...gr(r)})}const Yon=rn("ZodFile",(n,r)=>{min.init(n,r),Vs.init(n,r),n.min=(o,s)=>n.check(qie(o,s)),n.max=(o,s)=>n.check(Wye(o,s)),n.mime=(o,s)=>n.check(sJe(Array.isArray(o)?o:[o],s))});function Sur(n){return gon(Yon,n)}const HJe=rn("ZodTransform",(n,r)=>{IXe.init(n,r),Vs.init(n,r),n._zod.parse=(o,s)=>{o.addIssue=d=>{if(typeof d=="string")o.issues.push(JW(d,o.value,r));else{const h=d;h.fatal&&(h.continue=!1),h.code??(h.code="custom"),h.input??(h.input=o.value),h.inst??(h.inst=n),o.issues.push(JW(h))}};const c=r.transform(o.value,o);return c instanceof Promise?c.then(d=>(o.value=d,o)):(o.value=c,o)}});function GJe(n){return new HJe({type:"transform",transform:n})}const qJe=rn("ZodOptional",(n,r)=>{bin.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType});function dwe(n){return new qJe({type:"optional",innerType:n})}const Xon=rn("ZodNullable",(n,r)=>{vin.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType});function fwe(n){return new Xon({type:"nullable",innerType:n})}function _ur(n){return dwe(fwe(n))}const Jon=rn("ZodDefault",(n,r)=>{win.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType,n.removeDefault=n.unwrap});function Zon(n,r){return new Jon({type:"default",innerType:n,get defaultValue(){return typeof r=="function"?r():r}})}const Qon=rn("ZodPrefault",(n,r)=>{yin.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType});function ean(n,r){return new Qon({type:"prefault",innerType:n,get defaultValue(){return typeof r=="function"?r():r}})}const VJe=rn("ZodNonOptional",(n,r)=>{Ein.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType});function tan(n,r){return new VJe({type:"nonoptional",innerType:n,...gr(r)})}const nan=rn("ZodSuccess",(n,r)=>{Sin.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType});function kur(n){return new nan({type:"success",innerType:n})}const ran=rn("ZodCatch",(n,r)=>{_in.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType,n.removeCatch=n.unwrap});function ian(n,r){return new ran({type:"catch",innerType:n,catchValue:typeof r=="function"?r:()=>r})}const oan=rn("ZodNaN",(n,r)=>{kin.init(n,r),Vs.init(n,r)});function Tur(n){return con(oan,n)}const WJe=rn("ZodPipe",(n,r)=>{OXe.init(n,r),Vs.init(n,r),n.in=r.in,n.out=r.out});function hwe(n,r){return new WJe({type:"pipe",in:n,out:r})}const aan=rn("ZodReadonly",(n,r)=>{Tin.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType});function san(n){return new aan({type:"readonly",innerType:n})}const can=rn("ZodTemplateLiteral",(n,r)=>{xin.init(n,r),Vs.init(n,r)});function xur(n,r){return new can({type:"template_literal",parts:n,...gr(r)})}const uan=rn("ZodLazy",(n,r)=>{Ain.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.getter()});function KJe(n){return new uan({type:"lazy",getter:n})}const lan=rn("ZodPromise",(n,r)=>{Cin.init(n,r),Vs.init(n,r),n.unwrap=()=>n._zod.def.innerType});function Cur(n){return new lan({type:"promise",innerType:n})}const tEe=rn("ZodCustom",(n,r)=>{Iin.init(n,r),Vs.init(n,r)});function Aur(n){const r=new Ph({check:"custom"});return r._zod.check=n,r}function Iur(n,r){return mon(tEe,n??(()=>!0),r)}function dan(n,r={}){return bon(tEe,n,r)}function fan(n){return von(n)}function Our(n,r={error:`Input not instance of ${n.name}`}){const o=new tEe({type:"custom",check:"custom",fn:s=>s instanceof n,abort:!0,...gr(r)});return o._zod.bag.Class=n,o}const Nur=(...n)=>yon({Pipe:WJe,Boolean:Zye,String:Xye,Transform:HJe},...n);function Rur(n){const r=KJe(()=>D4([Tr(n),G4(),V1(),$Je(),mp(r),hU(Tr(),r)]));return r}function Pur(n,r){return hwe(GJe(n),r)}const Lur={invalid_type:"invalid_type",too_big:"too_big",too_small:"too_small",invalid_format:"invalid_format",not_multiple_of:"not_multiple_of",unrecognized_keys:"unrecognized_keys",invalid_union:"invalid_union",invalid_key:"invalid_key",invalid_element:"invalid_element",invalid_value:"invalid_value",custom:"custom"};function Dur(n){cv({customError:n})}function Mur(){return cv().customError}var KVe;KVe||(KVe={});function jur(n){return Din(Xye,n)}function $ur(n){return zin(Jye,n)}function Fur(n){return Yin(Zye,n)}function Bur(n){return Jin(Qye,n)}function Uur(n){return son(BJe,n)}const zur=Object.freeze(Object.defineProperty({__proto__:null,bigint:Bur,boolean:Fur,date:Uur,number:$ur,string:jur},Symbol.toStringTag,{value:"Module"}));cv(Oin());const _f=Object.freeze(Object.defineProperty({__proto__:null,$brand:lnn,$input:Pin,$output:Rin,NEVER:unn,TimePrecision:Min,ZodAny:Mon,ZodArray:Uon,ZodBase64:RJe,ZodBase64URL:PJe,ZodBigInt:Qye,ZodBigIntFormat:jJe,ZodBoolean:Zye,ZodCIDRv4:OJe,ZodCIDRv6:NJe,ZodCUID:_Je,ZodCUID2:kJe,ZodCatch:ran,ZodCustom:tEe,ZodCustomStringFormat:MJe,ZodDate:BJe,ZodDefault:Jon,ZodDiscriminatedUnion:zon,ZodE164:LJe,ZodEmail:wJe,ZodEmoji:EJe,ZodEnum:Wie,ZodError:Ncr,ZodFile:Yon,get ZodFirstPartyTypeKind(){return KVe},ZodGUID:lwe,ZodIPv4:AJe,ZodIPv6:IJe,ZodISODate:gJe,ZodISODateTime:pJe,ZodISODuration:bJe,ZodISOTime:mJe,ZodIntersection:Hon,ZodIssueCode:Lur,ZodJWT:DJe,ZodKSUID:CJe,ZodLazy:uan,ZodLiteral:Kon,ZodMap:Von,ZodNaN:oan,ZodNanoID:SJe,ZodNever:Fon,ZodNonOptional:VJe,ZodNull:Don,ZodNullable:Xon,ZodNumber:Jye,ZodNumberFormat:RK,ZodObject:eEe,ZodOptional:qJe,ZodPipe:WJe,ZodPrefault:Qon,ZodPromise:lan,ZodReadonly:aan,ZodRealError:Boe,ZodRecord:zJe,ZodSet:Won,ZodString:Xye,ZodStringFormat:Wd,ZodSuccess:nan,ZodSymbol:Pon,ZodTemplateLiteral:can,ZodTransform:HJe,ZodTuple:qon,ZodType:Vs,ZodULID:TJe,ZodURL:yJe,ZodUUID:L4,ZodUndefined:Lon,ZodUnion:UJe,ZodUnknown:$on,ZodVoid:Bon,ZodXID:xJe,_ZodString:vJe,_default:Zon,any:jon,array:mp,base64:Xcr,base64url:Jcr,bigint:aur,boolean:V1,catch:ian,check:Aur,cidrv4:Kcr,cidrv6:Ycr,clone:eC,coerce:zur,config:cv,core:Icr,cuid:Ucr,cuid2:zcr,custom:Iur,date:fur,discriminatedUnion:mur,e164:Zcr,email:Rcr,emoji:Fcr,endsWith:aJe,enum:GI,file:Sur,flattenError:hXe,float32:nur,float64:rur,formatError:pXe,function:Son,getErrorMap:Mur,globalRegistry:PB,gt:QB,gte:s2,guid:Pcr,hostname:tur,includes:iJe,instanceof:Our,int:WVe,int32:iur,int64:sur,intersection:Gon,ipv4:Vcr,ipv6:Wcr,iso:Ocr,json:Rur,jwt:Qcr,keyof:hur,ksuid:qcr,lazy:KJe,length:Yye,literal:qI,locales:Nin,looseObject:gur,lowercase:nJe,lt:ZB,lte:Ux,map:wur,maxLength:Kye,maxSize:Wye,mime:sJe,minLength:QW,minSize:qie,multipleOf:Gie,nan:Tur,nanoid:Bcr,nativeEnum:Eur,negative:lon,never:FJe,nonnegative:fon,nonoptional:tan,nonpositive:don,normalize:cJe,null:$Je,nullable:fwe,nullish:_ur,number:G4,object:xs,optional:dwe,overwrite:fU,parse:Ion,parseAsync:Oon,partialRecord:vur,pipe:hwe,positive:uon,prefault:ean,preprocess:Pur,prettifyError:xnn,promise:Cur,property:hon,readonly:san,record:hU,refine:dan,regex:tJe,regexes:orn,registry:RXe,safeParse:Non,safeParseAsync:Ron,set:yur,setErrorMap:Dur,size:eJe,startsWith:oJe,strictObject:pur,string:Tr,stringFormat:eur,stringbool:Nur,success:kur,superRefine:fan,symbol:uur,templateLiteral:xur,toJSONSchema:_on,toLowerCase:lJe,toUpperCase:dJe,transform:GJe,treeifyError:knn,trim:uJe,tuple:bur,uint32:our,uint64:cur,ulid:Hcr,undefined:lur,union:D4,unknown:Vie,uppercase:rJe,url:$cr,uuid:Lcr,uuidv4:Dcr,uuidv6:Mcr,uuidv7:jcr,void:dur,xid:Gcr},Symbol.toStringTag,{value:"Module"}));var Pa=(n=>(n.Call="CALL",n.CollectInformation="COLLECT_INFORMATION",n.Condition="CONDITION",n.CustomAction="CUSTOM_ACTION",n.Link="LINK",n.Logic="LOGIC",n.Message="MESSAGE",n.SetSlots="SET_SLOTS",n.Start="START",n))(Pa||{}),gM=(n=>(n.IF="IF",n.ELSE="ELSE",n))(gM||{}),AW=(n=>(n.Auto="AUTO",n.Custom="CUSTOM",n))(AW||{}),av=(n=>(n.COLLECT_INFORMATION="COLLECT_INFORMATION",n.MESSAGE="MESSAGE",n.CONDITION="CONDITION",n.CUSTOM_ACTION="CUSTOM_ACTION",n.LINK="LINK",n.START="START",n.LOGIC="LOGIC",n.SET_SLOTS="SET_SLOTS",n.CALL="CALL",n))(av||{}),Sl=(n=>(n.Action="ACTION",n.ActiveLoop="ACTIVE_LOOP",n.FlowCancelled="FLOW_CANCELLED",n.FlowCompleted="FLOW_COMPLETED",n.FlowInterrupted="FLOW_INTERRUPTED",n.FlowResumed="FLOW_RESUMED",n.FlowStarted="FLOW_STARTED",n.Form="FORM",n.ResetSlots="RESET_SLOTS",n.Restart="RESTART",n.SessionStarted="SESSION_STARTED",n.Slot="SLOT",n))(Sl||{}),Vx=(n=>(n.Bot="BOT",n.User="USER",n))(Vx||{}),kI=(n=>(n.User="user",n.Copilot="copilot",n.Internal="copilot_internal",n))(kI||{}),Kie=(n=>(n.Positive="positive",n.Negative="negative",n))(Kie||{}),$b=(n=>(n.Copilot="copilot",n.OutOfScopeDetection="out_of_scope_detection",n.RoleplayDetection="roleplay_detection",n.Reference="reference",n.GuardrailsPolicyViolation="guardrails_policy_violation",n.GuardrailsBlocked="guardrails_blocked",n.ErrorFallback="error_fallback",n.KnowledgeBaseAccessRequested="knowledge_base_access_requested",n.TrainingErrorLog="training_error_log",n.CopilotTrainingSuccess="copilot_training_success",n.Commit="commit",n.CopilotTextContentPartStart="copilot_text_content_part_start",n.CopilotTextContentPartDelta="copilot_text_content_part_delta",n.CopilotTextContentPartEnd="copilot_text_content_part_end",n.Reasoning="reasoning",n.Exception="exception",n.McpToolCall="mcp_tool_call",n.TaskPlanning="task_planning",n))($b||{}),Yie=(n=>(n.Token="token",n.Complete="complete",n))(Yie||{}),han=(n=>(n.Log="log",n))(han||{});const Hur=GI(["called","running","completed","failed"]),Gur=GI(["copilot","out_of_scope_detection","roleplay_detection","reference","guardrails_policy_violation","guardrails_blocked","error_fallback","knowledge_base_access_requested","training_error_log","copilot_training_success","commit","copilot_text_content_part_start","copilot_text_content_part_delta","copilot_text_content_part_end","reasoning","exception","mcp_tool_call","task_planning"],{error:n=>{if(n.code==="invalid_value"){const{input:r,values:o}=n;return`Invalid enum value. Expected one of: ${o?.join(" | ")||""}, received: ${r}`}}}),qur=GI(["token","complete"]),pan=xs({index:G4(),title:Tr(),url:Tr()}),Vur=xs({type:qI("log"),content:Tr(),context:Tr().optional()}),Wur=xs({sha:Tr(),short_sha:Tr(),author:Tr(),email:Tr(),timestamp:G4(),message:Tr(),training_success:V1()}),gan=xs({id:Tr(),content:Tr(),status:GI(["pending","in_progress","completed","cancelled"])}),Kur=xs({type:qI("plan"),tasks:mp(gan)}),man=xs({response_category:Gur.optional(),completeness:qur,content:Tr().optional(),references:mp(pan).optional(),status:Tr().optional(),logs:mp(Vur).optional(),commit:Wur.optional(),tool_name:Tr().optional(),tool_status:Hur.optional(),tasks:mp(gan).optional()}),Yur=xs({type:qI("text"),text:Tr()}),Xur=xs({type:qI("file"),file_path:Tr(),file_content:Tr()}),Jur=xs({type:qI("link"),url:Tr(),label:Tr()}),Zur=xs({type:qI("button"),payload:Tr(),label:Tr()}),Qur=xs({type:qI("event"),name:Tr(),metadata:xs({active_flow:Tr().optional(),utter_action:Tr().optional(),flow_id:Tr().optional(),execution_success:V1().optional(),step_id:Tr().optional(),rawEvent:jon().optional(),reset:V1().optional(),was_modified_by_studio:V1().optional(),metadata:xs({rephrase:V1().optional(),rephrasePrompt:Tr().optional()}).optional(),parseData:Vie()}),timestamp:G4(),event:GI(["user","stack","action","bot","slot"]),conversation_id:Tr(),text:Tr(),data:xs({attachment:xs({type:Tr(),payload:xs({src:Tr(),url:Tr(),elements:Tr()})}),quick_replies:mp(xs({image_url:Tr().optional(),payload:Tr(),title:Tr()})),buttons:mp(xs({image_url:Tr().optional(),payload:Tr(),title:Tr()})),elements:Tr(),image:Tr(),text:Tr()}),update:Tr(),flow_id:Tr(),step_id:Tr(),value:Tr(),parse_data:xs({intent_ranking:mp(xs({name:Tr(),confidence:G4()})),commands:Tr().optional()})}),elr=xs({type:qI("references"),references:mp(pan)}),tlr=D4([Yur,Xur,Qur,elr,Jur,Kur,Zur]),nlr=xs({role:GI(["user","copilot","copilot_internal"]),content:mp(tlr)});xs({session_id:Tr(),message:nlr,chat_id:Tr().optional()});const rlr=_f.object({project_id:_f.string(),url:_f.string(),available_until:_f.string().nullable(),status:_f.string().optional(),machine_id:_f.string().nullable().optional(),project_state:_f.string().nullable().optional(),job_id:_f.string().nullable().optional(),backup_url:_f.string().nullable().optional()});_f.object({job_id:_f.string(),status:_f.string()});_f.object({status:_f.string(),message:_f.string().optional(),result:_f.object({project_id:_f.string(),url:_f.string(),available_until:_f.string().nullable(),status:_f.string().optional(),machine_id:_f.string().nullable().optional(),project_state:_f.string().nullable().optional(),job_id:_f.string().nullable().optional(),backup_url:_f.string().nullable().optional()}).optional()});const ilr=xs({status:Tr(),auth_required:V1().optional(),files_changed_since_training:V1().optional(),model_trained_at:Tr().optional()}),olr=xs({assistant_id:Tr()}),ban=xs({mode:GI(["legacy","agent_sdk"])}),alr=hU(Tr(),Tr());var YJe=(n=>(n.Added="A",n.Modified="M",n.Deleted="D",n.Renamed="R",n))(YJe||{});const slr=xs({files:hU(Tr(),xs({status:GI(["A","M","D","R"]),content_original:Tr(),content_modified:Tr(),path_original:Tr().nullable(),path_modified:Tr().nullable()}))});var FB=(n=>(n.Finance="finance",n.Telco="telco",n.Basic="basic",n))(FB||{}),Ox=(n=>(n.None="none",n.WrongAnswer="wrong-answer",n.TooLong="too-long",n.OffTopic="off-topic",n.BrokenLinks="broken-links",n.WrongFlag="wrong-flag",n.Other="other",n))(Ox||{}),IW=(n=>(n.Cloud="cloud",n.Standalone="standalone",n.StudioEmbed="studio",n))(IW||{}),q4=(n=>(n.Chat="CHAT",n.Code="CODE",n))(q4||{});const XJe=xs({job_id:Tr(),status:Tr()});var zm=(n=>(n[n.NoFlow=0]="NoFlow",n[n.Mobile=1]="Mobile",n[n.Cubes=2]="Cubes",n[n.CubesError=3]="CubesError",n[n.CubesPause=4]="CubesPause",n))(zm||{});const clr=()=>({messages:[],attachments:[],isLoading:!1,sessionId:null,streamingMessage:null,shouldScrollOnStreamStart:!1}),il=new IM(clr());var YVe=function(n,r){return YVe=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(o,s){o.__proto__=s}||function(o,s){for(var c in s)Object.prototype.hasOwnProperty.call(s,c)&&(o[c]=s[c])},YVe(n,r)};function iy(n,r){if(typeof r!="function"&&r!==null)throw new TypeError("Class extends value "+String(r)+" is not a constructor or null");YVe(n,r);function o(){this.constructor=n}n.prototype=r===null?Object.create(r):(o.prototype=r.prototype,new o)}var zr=function(){return zr=Object.assign||function(r){for(var o,s=1,c=arguments.length;s<c;s++){o=arguments[s];for(var d in o)Object.prototype.hasOwnProperty.call(o,d)&&(r[d]=o[d])}return r},zr.apply(this,arguments)};function van(n,r){var o={};for(var s in n)Object.prototype.hasOwnProperty.call(n,s)&&r.indexOf(s)<0&&(o[s]=n[s]);if(n!=null&&typeof Object.getOwnPropertySymbols=="function")for(var c=0,s=Object.getOwnPropertySymbols(n);c<s.length;c++)r.indexOf(s[c])<0&&Object.prototype.propertyIsEnumerable.call(n,s[c])&&(o[s[c]]=n[s[c]]);return o}function wo(n,r,o,s){function c(d){return d instanceof o?d:new o(function(h){h(d)})}return new(o||(o=Promise))(function(d,h){function b(k){try{E(s.next(k))}catch(x){h(x)}}function y(k){try{E(s.throw(k))}catch(x){h(x)}}function E(k){k.done?d(k.value):c(k.value).then(b,y)}E((s=s.apply(n,r||[])).next())})}function yo(n,r){var o={label:0,sent:function(){if(d[0]&1)throw d[1];return d[1]},trys:[],ops:[]},s,c,d,h=Object.create((typeof Iterator=="function"?Iterator:Object).prototype);return h.next=b(0),h.throw=b(1),h.return=b(2),typeof Symbol=="function"&&(h[Symbol.iterator]=function(){return this}),h;function b(E){return function(k){return y([E,k])}}function y(E){if(s)throw new TypeError("Generator is already executing.");for(;h&&(h=0,E[0]&&(o=0)),o;)try{if(s=1,c&&(d=E[0]&2?c.return:E[0]?c.throw||((d=c.return)&&d.call(c),0):c.next)&&!(d=d.call(c,E[1])).done)return d;switch(c=0,d&&(E=[E[0]&2,d.value]),E[0]){case 0:case 1:d=E;break;case 4:return o.label++,{value:E[1],done:!1};case 5:o.label++,c=E[1],E=[0];continue;case 7:E=o.ops.pop(),o.trys.pop();continue;default:if(d=o.trys,!(d=d.length>0&&d[d.length-1])&&(E[0]===6||E[0]===2)){o=0;continue}if(E[0]===3&&(!d||E[1]>d[0]&&E[1]<d[3])){o.label=E[1];break}if(E[0]===6&&o.label<d[1]){o.label=d[1],d=E;break}if(d&&o.label<d[2]){o.label=d[2],o.ops.push(E);break}d[2]&&o.ops.pop(),o.trys.pop();continue}E=r.call(n,o)}catch(k){E=[6,k],c=0}finally{s=d=0}if(E[0]&5)throw E[1];return{value:E[0]?E[1]:void 0,done:!0}}}function Vg(n,r,o){if(o||arguments.length===2)for(var s=0,c=r.length,d;s<c;s++)(d||!(s in r))&&(d||(d=Array.prototype.slice.call(r,0,s)),d[s]=r[s]);return n.concat(d||Array.prototype.slice.call(r))}function XVe(n,r,o){r.split&&(r=r.split("."));for(var s=0,c=r.length,d=n,h,b;s<c&&(b=""+r[s++],!(b==="__proto__"||b==="constructor"||b==="prototype"));)d=d[b]=s===c?o:typeof(h=d[b])==typeof r?h:r[s]*0!==0||~(""+r[s]).indexOf(".")?{}:[]}var ulr=function(n,r){return Object.keys(n).filter(function(o){return r(o,n[o])}).reduce(function(o,s){return o[s]=n[s],o},{})},eU=function(n){iy(r,n);function r(o,s){var c=n.call(this,"".concat(o," ").concat(s))||this;return c.field=o,c}return r}(Error);function NM(n){return typeof n=="string"}function JVe(n){return typeof n=="number"}function OW(n){return typeof n=="function"}function llr(n){return n!=null}function Fx(n){return Object.prototype.toString.call(n).slice(8,-1).toLowerCase()==="object"}var JJe="is not a string",ZJe="is not an object",dlr="is nil";function flr(n){if(!llr(n))throw new eU("Event",dlr);if(typeof n!="object")throw new eU("Event",ZJe)}function hlr(n){if(!NM(n.type))throw new eU(".type",JJe)}function plr(n){if(!NM(n.event))throw new eU(".event",JJe)}function glr(n){if(!Fx(n.properties))throw new eU(".properties",ZJe)}function mlr(n){if(!Fx(n.traits))throw new eU(".traits",ZJe)}function blr(n){if(!NM(n.messageId))throw new eU(".messageId",JJe)}function vlr(n){flr(n),hlr(n),blr(n),n.type==="track"&&(plr(n),glr(n)),["group","identify"].includes(n.type)&&mlr(n)}var wlr=function(){function n(r){var o,s;this.settings=r,this.createMessageId=r.createMessageId,this.onEventMethodCall=(o=r.onEventMethodCall)!==null&&o!==void 0?o:function(){},this.onFinishedEvent=(s=r.onFinishedEvent)!==null&&s!==void 0?s:function(){}}return n}(),ylr=function(){function n(r){this.settings=new wlr(r)}return n.prototype.track=function(r,o,s,c){return this.settings.onEventMethodCall({type:"track",options:s}),this.normalize(zr(zr({},this.baseEvent()),{event:r,type:"track",properties:o??{},options:zr({},s),integrations:zr({},c)}))},n.prototype.page=function(r,o,s,c,d){var h;this.settings.onEventMethodCall({type:"page",options:c});var b={type:"page",properties:zr({},s),options:zr({},c),integrations:zr({},d)};return r!==null&&(b.category=r,b.properties=(h=b.properties)!==null&&h!==void 0?h:{},b.properties.category=r),o!==null&&(b.name=o),this.normalize(zr(zr({},this.baseEvent()),b))},n.prototype.screen=function(r,o,s,c,d){this.settings.onEventMethodCall({type:"screen",options:c});var h={type:"screen",properties:zr({},s),options:zr({},c),integrations:zr({},d)};return r!==null&&(h.category=r),o!==null&&(h.name=o),this.normalize(zr(zr({},this.baseEvent()),h))},n.prototype.identify=function(r,o,s,c){return this.settings.onEventMethodCall({type:"identify",options:s}),this.normalize(zr(zr({},this.baseEvent()),{type:"identify",userId:r,traits:o??{},options:zr({},s),integrations:c}))},n.prototype.group=function(r,o,s,c){return this.settings.onEventMethodCall({type:"group",options:s}),this.normalize(zr(zr({},this.baseEvent()),{type:"group",traits:o??{},options:zr({},s),integrations:zr({},c),groupId:r}))},n.prototype.alias=function(r,o,s,c){this.settings.onEventMethodCall({type:"alias",options:s});var d={userId:r,type:"alias",options:zr({},s),integrations:zr({},c)};return o!==null&&(d.previousId=o),r===void 0?this.normalize(zr(zr({},d),this.baseEvent())):this.normalize(zr(zr({},this.baseEvent()),d))},n.prototype.baseEvent=function(){return{integrations:{},options:{}}},n.prototype.context=function(r){var o,s=["userId","anonymousId","timestamp","messageId"];delete r.integrations;var c=Object.keys(r),d=(o=r.context)!==null&&o!==void 0?o:{},h={};return c.forEach(function(b){b!=="context"&&(s.includes(b)?XVe(h,b,r[b]):XVe(d,b,r[b]))}),[d,h]},n.prototype.normalize=function(r){var o,s,c=Object.keys((o=r.integrations)!==null&&o!==void 0?o:{}).reduce(function(O,P){var D,$;return zr(zr({},O),(D={},D[P]=!!(!(($=r.integrations)===null||$===void 0)&&$[P]),D))},{});r.options=ulr(r.options||{},function(O,P){return P!==void 0});var d=zr(zr({},c),(s=r.options)===null||s===void 0?void 0:s.integrations),h=r.options?this.context(r.options):[],b=h[0],y=h[1],E=r.options,k=van(r,["options"]),x=zr(zr(zr(zr({timestamp:new Date},k),{context:b,integrations:d}),y),{messageId:E.messageId||this.settings.createMessageId()});return this.settings.onFinishedEvent(x),vlr(x),x},n}();function Elr(n,r){return new Promise(function(o,s){var c=setTimeout(function(){s(Error("Promise timed out"))},r);n.then(function(d){return clearTimeout(c),o(d)}).catch(s)})}function Slr(n){return new Promise(function(r){return setTimeout(r,n)})}function _lr(n,r,o){var s=function(){try{return Promise.resolve(r(n))}catch(c){return Promise.reject(c)}};return Slr(o).then(function(){return Elr(s(),1e3)}).catch(function(c){n?.log("warn","Callback Error",{error:c}),n?.stats.increment("callback_error")}).then(function(){return n})}var wan=function(){var n,r,o=!1,s=new Promise(function(c,d){n=function(){for(var h=[],b=0;b<arguments.length;b++)h[b]=arguments[b];o=!0,c.apply(void 0,h)},r=function(){for(var h=[],b=0;b<arguments.length;b++)h[b]=arguments[b];o=!0,d.apply(void 0,h)}});return{resolve:n,reject:r,promise:s,isSettled:function(){return o}}},QJe=function(){function n(r){var o;this.callbacks={},this.warned=!1,this.maxListeners=(o=r?.maxListeners)!==null&&o!==void 0?o:10}return n.prototype.warnIfPossibleMemoryLeak=function(r){this.warned||this.maxListeners&&this.callbacks[r].length>this.maxListeners&&(console.warn("Event Emitter: Possible memory leak detected; ".concat(String(r)," has exceeded ").concat(this.maxListeners," listeners.")),this.warned=!0)},n.prototype.on=function(r,o){return this.callbacks[r]?(this.callbacks[r].push(o),this.warnIfPossibleMemoryLeak(r)):this.callbacks[r]=[o],this},n.prototype.once=function(r,o){var s=this,c=function(){for(var d=[],h=0;h<arguments.length;h++)d[h]=arguments[h];s.off(r,c),o.apply(s,d)};return this.on(r,c),this},n.prototype.off=function(r,o){var s,c=(s=this.callbacks[r])!==null&&s!==void 0?s:[],d=c.filter(function(h){return h!==o});return this.callbacks[r]=d,this},n.prototype.emit=function(r){for(var o=this,s,c=[],d=1;d<arguments.length;d++)c[d-1]=arguments[d];var h=(s=this.callbacks[r])!==null&&s!==void 0?s:[];return h.forEach(function(b){b.apply(o,c)}),this},n}();function klr(n){var r=Math.random()+1,o=n.minTimeout,s=o===void 0?500:o,c=n.factor,d=c===void 0?2:c,h=n.attempt,b=n.maxTimeout,y=b===void 0?1/0:b;return Math.min(r*s*Math.pow(d,h),y)}var yan="onRemoveFromFuture",eZe=function(n){iy(r,n);function r(o,s,c){var d=n.call(this)||this;return d.future=[],d.maxAttempts=o,d.queue=s,d.seen=c??{},d}return r.prototype.push=function(){for(var o=this,s=[],c=0;c<arguments.length;c++)s[c]=arguments[c];var d=s.map(function(h){var b=o.updateAttempts(h);return b>o.maxAttempts||o.includes(h)?!1:(o.queue.push(h),!0)});return this.queue=this.queue.sort(function(h,b){return o.getAttempts(h)-o.getAttempts(b)}),d},r.prototype.pushWithBackoff=function(o,s){var c=this;if(s===void 0&&(s=0),s==0&&this.getAttempts(o)===0)return this.push(o)[0];var d=this.updateAttempts(o);if(d>this.maxAttempts||this.includes(o))return!1;var h=klr({attempt:d-1});return s>0&&h<s&&(h=s),setTimeout(function(){c.queue.push(o),c.future=c.future.filter(function(b){return b.id!==o.id}),c.emit(yan)},h),this.future.push(o),!0},r.prototype.getAttempts=function(o){var s;return(s=this.seen[o.id])!==null&&s!==void 0?s:0},r.prototype.updateAttempts=function(o){return this.seen[o.id]=this.getAttempts(o)+1,this.getAttempts(o)},r.prototype.includes=function(o){return this.queue.includes(o)||this.future.includes(o)||!!this.queue.find(function(s){return s.id===o.id})||!!this.future.find(function(s){return s.id===o.id})},r.prototype.pop=function(){return this.queue.shift()},Object.defineProperty(r.prototype,"length",{get:function(){return this.queue.length},enumerable:!1,configurable:!0}),Object.defineProperty(r.prototype,"todo",{get:function(){return this.queue.length+this.future.length},enumerable:!1,configurable:!0}),r}(QJe),LB=256,pve=[],r0e;for(;LB--;)pve[LB]=(LB+256).toString(16).substring(1);function tZe(){var n=0,r,o="";if(!r0e||LB+16>256){for(r0e=Array(n=256);n--;)r0e[n]=256*Math.random()|0;n=LB=0}for(;n<16;n++)r=r0e[LB+n],n==6?o+=pve[r&15|64]:n==8?o+=pve[r&63|128]:o+=pve[r],n&1&&n>1&&n<11&&(o+="-");return LB++,o}var Tlr=function(){function n(){this._logs=[]}return n.prototype.log=function(r,o,s){var c=new Date;this._logs.push({level:r,message:o,time:c,extras:s})},Object.defineProperty(n.prototype,"logs",{get:function(){return this._logs},enumerable:!1,configurable:!0}),n.prototype.flush=function(){if(this.logs.length>1){var r=this._logs.reduce(function(o,s){var c,d,h,b=zr(zr({},s),{json:JSON.stringify(s.extras,null," "),extras:s.extras});delete b.time;var y=(h=(d=s.time)===null||d===void 0?void 0:d.toISOString())!==null&&h!==void 0?h:"";return o[y]&&(y="".concat(y,"-").concat(Math.random())),zr(zr({},o),(c={},c[y]=b,c))},{});console.table?console.table(r):console.log(r)}else this.logs.forEach(function(o){var s=o.level,c=o.message,d=o.extras;s==="info"||s==="debug"?console.log(c,d??""):console[s](c,d??"")});this._logs=[]},n}(),xlr=function(n){var r={gauge:"g",counter:"c"};return r[n]},Ean=function(){function n(){this.metrics=[]}return n.prototype.increment=function(r,o,s){o===void 0&&(o=1),this.metrics.push({metric:r,value:o,tags:s??[],type:"counter",timestamp:Date.now()})},n.prototype.gauge=function(r,o,s){this.metrics.push({metric:r,value:o,tags:s??[],type:"gauge",timestamp:Date.now()})},n.prototype.flush=function(){var r=this.metrics.map(function(o){return zr(zr({},o),{tags:o.tags.join(",")})});console.table?console.table(r):console.log(r),this.metrics=[]},n.prototype.serialize=function(){return this.metrics.map(function(r){return{m:r.metric,v:r.value,t:r.tags,k:xlr(r.type),e:r.timestamp}})},n}(),Clr=function(n){iy(r,n);function r(){return n!==null&&n.apply(this,arguments)||this}return r.prototype.gauge=function(){},r.prototype.increment=function(){},r.prototype.flush=function(){},r.prototype.serialize=function(){return[]},r}(Ean),eK=function(){function n(r){var o,s,c;this.retry=(o=r.retry)!==null&&o!==void 0?o:!0,this.type=(s=r.type)!==null&&s!==void 0?s:"plugin Error",this.reason=(c=r.reason)!==null&&c!==void 0?c:""}return n}(),pwe=function(){function n(r,o,s,c){o===void 0&&(o=tZe()),s===void 0&&(s=new Clr),c===void 0&&(c=new Tlr),this.attempts=0,this.event=r,this._id=o,this.logger=c,this.stats=s}return n.system=function(){},n.prototype.isSame=function(r){return r.id===this.id},n.prototype.cancel=function(r){throw r||new eK({reason:"Context Cancel"})},n.prototype.log=function(r,o,s){this.logger.log(r,o,s)},Object.defineProperty(n.prototype,"id",{get:function(){return this._id},enumerable:!1,configurable:!0}),n.prototype.updateEvent=function(r,o){var s;if(r.split(".")[0]==="integrations"){var c=r.split(".")[1];if(((s=this.event.integrations)===null||s===void 0?void 0:s[c])===!1)return this.event}return XVe(this.event,r,o),this.event},n.prototype.failedDelivery=function(){return this._failedDelivery},n.prototype.setFailedDelivery=function(r){this._failedDelivery=r},n.prototype.logs=function(){return this.logger.logs},n.prototype.flush=function(){this.logger.flush(),this.stats.flush()},n.prototype.toJSON=function(){return{id:this._id,event:this.event,logs:this.logger.logs,metrics:this.stats.metrics}},n}();function Alr(n,r){var o={};return n.forEach(function(s){var c,d=void 0;{var h=s[r];d=typeof h!="string"?JSON.stringify(h):h}d!==void 0&&(o[d]=Vg(Vg([],(c=o[d])!==null&&c!==void 0?c:[],!0),[s],!1))}),o}var Ilr=function(n){return typeof n=="object"&&n!==null&&"then"in n&&typeof n.then=="function"},Olr=function(){var n,r,o=0;return{done:function(){return n},run:function(s){var c=s();return Ilr(c)&&(++o===1&&(n=new Promise(function(d){return r=d})),c.finally(function(){return--o===0&&r()})),c}}};function Nlr(n){return wo(this,void 0,void 0,function(){var r;return yo(this,function(o){switch(o.label){case 0:return o.trys.push([0,2,,3]),[4,n()];case 1:return[2,o.sent()];case 2:return r=o.sent(),[2,Promise.reject(r)];case 3:return[2]}})})}function lie(n,r){n.log("debug","plugin",{plugin:r.name});var o=new Date().getTime(),s=r[n.event.type];if(s===void 0)return Promise.resolve(n);var c=Nlr(function(){return s.apply(r,[n])}).then(function(d){var h=new Date().getTime()-o;return d.stats.gauge("plugin_time",h,["plugin:".concat(r.name)]),d}).catch(function(d){if(d instanceof eK&&d.type==="middleware_cancellation")throw d;return d instanceof eK?(n.log("warn",d.type,{plugin:r.name,error:d}),d):(n.log("error","plugin Error",{plugin:r.name,error:d}),n.stats.increment("plugin_error",1,["plugin:".concat(r.name)]),d)});return c}function Rlr(n,r){return lie(n,r).then(function(o){if(o instanceof pwe)return o;n.log("debug","Context canceled"),n.stats.increment("context_canceled"),n.cancel(o)})}var Plr=function(n){iy(r,n);function r(o){var s=n.call(this)||this;return s.criticalTasks=Olr(),s.plugins=[],s.failedInitializations=[],s.flushing=!1,s.queue=o,s.queue.on(yan,function(){s.scheduleFlush(0)}),s}return r.prototype.register=function(o,s,c){return wo(this,void 0,void 0,function(){var d,h,b=this;return yo(this,function(y){switch(y.label){case 0:return this.plugins.push(s),d=function(E){b.failedInitializations.push(s.name),b.emit("initialization_failure",s),console.warn(s.name,E),o.log("warn","Failed to load destination",{plugin:s.name,error:E}),b.plugins=b.plugins.filter(function(k){return k!==s})},s.type==="destination"&&s.name!=="Segment.io"?(s.load(o,c).catch(d),[3,4]):[3,1];case 1:return y.trys.push([1,3,,4]),[4,s.load(o,c)];case 2:return y.sent(),[3,4];case 3:return h=y.sent(),d(h),[3,4];case 4:return[2]}})})},r.prototype.deregister=function(o,s,c){return wo(this,void 0,void 0,function(){var d;return yo(this,function(h){switch(h.label){case 0:return h.trys.push([0,3,,4]),s.unload?[4,Promise.resolve(s.unload(o,c))]:[3,2];case 1:h.sent(),h.label=2;case 2:return this.plugins=this.plugins.filter(function(b){return b.name!==s.name}),[3,4];case 3:return d=h.sent(),o.log("warn","Failed to unload destination",{plugin:s.name,error:d}),[3,4];case 4:return[2]}})})},r.prototype.dispatch=function(o){return wo(this,void 0,void 0,function(){var s;return yo(this,function(c){return o.log("debug","Dispatching"),o.stats.increment("message_dispatched"),this.queue.push(o),s=this.subscribeToDelivery(o),this.scheduleFlush(0),[2,s]})})},r.prototype.subscribeToDelivery=function(o){return wo(this,void 0,void 0,function(){var s=this;return yo(this,function(c){return[2,new Promise(function(d){var h=function(b,y){b.isSame(o)&&(s.off("flush",h),d(b))};s.on("flush",h)})]})})},r.prototype.dispatchSingle=function(o){return wo(this,void 0,void 0,function(){var s=this;return yo(this,function(c){return o.log("debug","Dispatching"),o.stats.increment("message_dispatched"),this.queue.updateAttempts(o),o.attempts=1,[2,this.deliver(o).catch(function(d){var h=s.enqueuRetry(d,o);return h?s.subscribeToDelivery(o):(o.setFailedDelivery({reason:d}),o)})]})})},r.prototype.isEmpty=function(){return this.queue.length===0},r.prototype.scheduleFlush=function(o){var s=this;o===void 0&&(o=500),!this.flushing&&(this.flushing=!0,setTimeout(function(){s.flush().then(function(){setTimeout(function(){s.flushing=!1,s.queue.length&&s.scheduleFlush(0)},0)})},o))},r.prototype.deliver=function(o){return wo(this,void 0,void 0,function(){var s,c,d,h;return yo(this,function(b){switch(b.label){case 0:return[4,this.criticalTasks.done()];case 1:b.sent(),s=Date.now(),b.label=2;case 2:return b.trys.push([2,4,,5]),[4,this.flushOne(o)];case 3:return o=b.sent(),c=Date.now()-s,this.emit("delivery_success",o),o.stats.gauge("delivered",c),o.log("debug","Delivered",o.event),[2,o];case 4:throw d=b.sent(),h=d,o.log("error","Failed to deliver",h),this.emit("delivery_failure",o,h),o.stats.increment("delivery_failed"),d;case 5:return[2]}})})},r.prototype.enqueuRetry=function(o,s){var c=!(o instanceof eK)||o.retry;return c?this.queue.pushWithBackoff(s):!1},r.prototype.flush=function(){return wo(this,void 0,void 0,function(){var o,s,c;return yo(this,function(d){switch(d.label){case 0:if(this.queue.length===0)return[2,[]];if(o=this.queue.pop(),!o)return[2,[]];o.attempts=this.queue.getAttempts(o),d.label=1;case 1:return d.trys.push([1,3,,4]),[4,this.deliver(o)];case 2:return o=d.sent(),this.emit("flush",o,!0),[3,4];case 3:return s=d.sent(),c=this.enqueuRetry(s,o),c||(o.setFailedDelivery({reason:s}),this.emit("flush",o,!1)),[2,[]];case 4:return[2,[o]]}})})},r.prototype.isReady=function(){return!0},r.prototype.availableExtensions=function(o){var s=this.plugins.filter(function(P){var D,$,U;if(P.type!=="destination"&&P.name!=="Segment.io")return!0;var B=void 0;return(D=P.alternativeNames)===null||D===void 0||D.forEach(function(G){o[G]!==void 0&&(B=o[G])}),(U=($=o[P.name])!==null&&$!==void 0?$:B)!==null&&U!==void 0?U:(P.name==="Segment.io"?!0:o.All)!==!1}),c=Alr(s,"type"),d=c.before,h=d===void 0?[]:d,b=c.enrichment,y=b===void 0?[]:b,E=c.destination,k=E===void 0?[]:E,x=c.after,O=x===void 0?[]:x;return{before:h,enrichment:y,destinations:k,after:O}},r.prototype.flushOne=function(o){var s,c;return wo(this,void 0,void 0,function(){var d,h,b,y,E,k,D,x,O,P,D,$,U,B,G;return yo(this,function(K){switch(K.label){case 0:if(!this.isReady())throw new Error("Not ready");o.attempts>1&&this.emit("delivery_retry",o),d=this.availableExtensions((s=o.event.integrations)!==null&&s!==void 0?s:{}),h=d.before,b=d.enrichment,y=0,E=h,K.label=1;case 1:return y<E.length?(k=E[y],[4,Rlr(o,k)]):[3,4];case 2:D=K.sent(),D instanceof pwe&&(o=D),this.emit("message_enriched",o,k),K.label=3;case 3:return y++,[3,1];case 4:x=0,O=b,K.label=5;case 5:return x<O.length?(P=O[x],[4,lie(o,P)]):[3,8];case 6:D=K.sent(),D instanceof pwe&&(o=D),this.emit("message_enriched",o,P),K.label=7;case 7:return x++,[3,5];case 8:return $=this.availableExtensions((c=o.event.integrations)!==null&&c!==void 0?c:{}),U=$.destinations,B=$.after,[4,new Promise(function(X,ae){setTimeout(function(){var oe=U.map(function(he){return lie(o,he)});Promise.all(oe).then(X).catch(ae)},0)})];case 9:return K.sent(),o.stats.increment("message_delivered"),this.emit("message_delivered",o),G=B.map(function(X){return lie(o,X)}),[4,Promise.all(G)];case 10:return K.sent(),[2,o]}})})},r}(QJe),Llr=function(n,r){var o=Date.now()-n;return Math.max((r??300)-o,0)};function Dlr(n,r,o,s){return wo(this,void 0,void 0,function(){var c,d;return yo(this,function(h){switch(h.label){case 0:return o.emit("dispatch_start",n),c=Date.now(),r.isEmpty()?[4,r.dispatchSingle(n)]:[3,2];case 1:return d=h.sent(),[3,4];case 2:return[4,r.dispatch(n)];case 3:d=h.sent(),h.label=4;case 4:return s?.callback?[4,_lr(d,s.callback,Llr(c,s.timeout))]:[3,6];case 5:d=h.sent(),h.label=6;case 6:return s?.debug&&d.flush(),[2,d]}})})}function Mlr(n,r,o,s){var c,d=[n,r,o,s],h=Fx(n)?n.event:n;if(!h||!NM(h))throw new Error("Event missing");var b=Fx(n)?(c=n.properties)!==null&&c!==void 0?c:{}:Fx(r)?r:{},y={};OW(o)||(y=o??{}),Fx(n)&&!OW(r)&&(y=r??{});var E=d.find(OW);return[h,b,y,E]}function Xjt(n,r,o,s,c){var d,h,b=null,y=null,E=[n,r,o,s,c],k=E.filter(NM);k.length===1?NM(E[1])?(y=E[1],b=null):(y=k[0],b=null):k.length===2&&(typeof E[0]=="string"&&(b=E[0]),typeof E[1]=="string"&&(y=E[1]));var x=E.find(OW),O=E.filter(Fx);return O.length===1?Fx(E[2])?(h={},d=E[2]):Fx(E[3])?(d={},h=E[3]):(d=O[0],h={}):O.length===2&&(d=O[0],h=O[1]),[b,y,d??(d={}),h??(h={}),x]}var Jjt=function(n){return function(){for(var r,o,s,c=[],d=0;d<arguments.length;d++)c[d]=arguments[d];for(var h={},b=["callback","options","traits","id"],y=0,E=c;y<E.length;y++){var k=E[y],x=b.pop();if(x==="id"){if(NM(k)||JVe(k)){h.id=k.toString();continue}if(k==null)continue;x=b.pop()}if((x==="traits"||x==="options")&&(k==null||Fx(k))&&(h[x]=k),OW(k)){h.callback=k;break}}return[(r=h.id)!==null&&r!==void 0?r:n.id(),(o=h.traits)!==null&&o!==void 0?o:{},(s=h.options)!==null&&s!==void 0?s:{},h.callback]}};function jlr(n,r,o,s){JVe(n)&&(n=n.toString()),JVe(r)&&(r=r.toString());var c=[n,r,o,s],d=c.filter(NM),h=d[0],b=h===void 0?n:h,y=d[1],E=y===void 0?null:y,k=c.filter(Fx)[0],x=k===void 0?{}:k,O=c.find(OW);return[b,E,x,O]}function nZe(){return typeof window<"u"}function Uni(){return!nZe()}function $lr(){return nZe()?window.navigator.onLine:!0}function Xie(){return!$lr()}function Flr(n,r){return r=r||{},new Promise(function(o,s){var c=new XMLHttpRequest,d=[],h=[],b={},y=function(){return{ok:(c.status/100|0)==2,statusText:c.statusText,status:c.status,url:c.responseURL,text:function(){return Promise.resolve(c.responseText)},json:function(){return Promise.resolve(c.responseText).then(JSON.parse)},blob:function(){return Promise.resolve(new Blob([c.response]))},clone:y,headers:{keys:function(){return d},entries:function(){return h},get:function(k){return b[k.toLowerCase()]},has:function(k){return k.toLowerCase()in b}}}};for(var E in c.open(r.method||"get",n,!0),c.onload=function(){c.getAllResponseHeaders().replace(/^(.*?):[^\S\n]*([\s\S]*?)$/gm,function(k,x,O){d.push(x=x.toLowerCase()),h.push([x,O]),b[x]=b[x]?b[x]+","+O:O}),o(y())},c.onerror=s,c.withCredentials=r.credentials=="include",r.headers)c.setRequestHeader(E,r.headers[E]);c.send(r.body||null)})}var rZe=function(){return typeof globalThis<"u"?globalThis:typeof self<"u"?self:typeof window<"u"?window:typeof global<"u"?global:null},nEe=function(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var o=rZe();return(o&&o.fetch||Flr).apply(void 0,n)},rEe="1.81.1",San="api.segment.io/v1",Blr=function(n,r,o){var s=r.reduce(function(c,d){var h=d.split(":"),b=h[0],y=h[1];return c[b]=y,c},{});return{type:"Counter",metric:n,value:1,tags:zr(zr({},s),{library:"analytics.js",library_version:"npm:next-".concat(rEe)})}};function T$e(n){console.error("Error sending segment performance metrics",n)}var Ulr=function(){function n(r){var o=this,s,c,d,h,b;if(this.host=(s=r?.host)!==null&&s!==void 0?s:San,this.sampleRate=(c=r?.sampleRate)!==null&&c!==void 0?c:1,this.flushTimer=(d=r?.flushTimer)!==null&&d!==void 0?d:30*1e3,this.maxQueueSize=(h=r?.maxQueueSize)!==null&&h!==void 0?h:20,this.protocol=(b=r?.protocol)!==null&&b!==void 0?b:"https",this.queue=[],this.sampleRate>0){var y=!1,E=function(){y||(y=!0,o.flush().catch(T$e),y=!1,setTimeout(E,o.flushTimer))};E()}}return n.prototype.increment=function(r,o){if(r.includes("analytics_js.")&&o.length!==0&&!(Math.random()>this.sampleRate)&&!(this.queue.length>=this.maxQueueSize)){var s=Blr(r,o);this.queue.push(s),r.includes("error")&&this.flush().catch(T$e)}},n.prototype.flush=function(){return wo(this,void 0,void 0,function(){var r=this;return yo(this,function(o){switch(o.label){case 0:return this.queue.length<=0?[2]:[4,this.send().catch(function(s){T$e(s),r.sampleRate=0})];case 1:return o.sent(),[2]}})})},n.prototype.send=function(){return wo(this,void 0,void 0,function(){var r,o,s;return yo(this,function(c){return r={series:this.queue},this.queue=[],o={"Content-Type":"text/plain"},s="".concat(this.protocol,"://").concat(this.host,"/m"),[2,nEe(s,{headers:o,body:JSON.stringify(r),method:"POST"})]})})},n}(),i0e,_an=function(n){iy(r,n);function r(){return n!==null&&n.apply(this,arguments)||this}return r.initRemoteMetrics=function(o){i0e=new Ulr(o)},r.prototype.increment=function(o,s,c){n.prototype.increment.call(this,o,s,c),i0e?.increment(o,c??[])},r}(Ean),$4=function(n){iy(r,n);function r(o,s){return n.call(this,o,s,new _an)||this}return r.system=function(){return new this({type:"track",event:"system"})},r}(pwe);function zlr(n){return Object.prototype.toString.call(n).slice(8,-1).toLowerCase()==="object"}var kan="bpc",Tan=function(n,r,o,s,c,d){return{__t:kan,c:r,p:s,u:n,s:o,t:c,r:d}},Hlr=Object.keys(Tan("","","","","",""));function Glr(n){if(!zlr(n)||n.__t!==kan)return!1;for(var r in n)if(!Hlr.includes(r))return!1;return!0}var qlr=function(n,r){return n.indexOf("?")>-1?n:n+r},Vlr=function(n){var r=n.indexOf("#");return r===-1?n:n.slice(0,r)},Wlr=function(n){try{return new URL(n).pathname}catch{return n[0]==="/"?n:"/"+n}},xan=function(n){var r=n.c,o=n.p,s=n.s,c=n.u,d=n.r,h=n.t,b=r?Wlr(r):o,y=r?qlr(r,s):Vlr(c);return{path:b,referrer:d,search:s,title:h,url:y}},Can=function(){var n=document.querySelector("link[rel='canonical']");return Tan(location.href,n&&n.getAttribute("href")||void 0,location.search,location.pathname,document.title,document.referrer)},Klr=function(){return xan(Can())};function Ylr(n,r){return Object.assign.apply(Object,Vg([{}],r.map(function(o){var s;if(n&&Object.prototype.hasOwnProperty.call(n,o))return s={},s[o]=n[o],s}),!1))}var BV=function(n,r){r===void 0&&(r=Klr());var o=n.context,s;n.type==="page"&&(s=n.properties&&Ylr(n.properties,Object.keys(r)),n.properties=zr(zr(zr({},r),n.properties),n.name?{name:n.name}:{})),o.page=zr(zr(zr({},r),s),o.page)},Xlr=function(n){iy(r,n);function r(o){var s=n.call(this,{createMessageId:function(){return"ajs-next-".concat(Date.now(),"-").concat(tZe())},onEventMethodCall:function(c){var d=c.options;s.maybeUpdateAnonId(d)},onFinishedEvent:function(c){return s.addIdentity(c),c}})||this;return s.user=o,s}return r.prototype.maybeUpdateAnonId=function(o){o?.anonymousId&&this.user.anonymousId(o.anonymousId)},r.prototype.addIdentity=function(o){this.user.id()&&(o.userId=this.user.id()),this.user.anonymousId()&&(o.anonymousId=this.user.anonymousId())},r.prototype.track=function(o,s,c,d,h){var b=n.prototype.track.call(this,o,s,c,d);return BV(b,h),b},r.prototype.page=function(o,s,c,d,h,b){var y=n.prototype.page.call(this,o,s,c,d,h);return BV(y,b),y},r.prototype.screen=function(o,s,c,d,h,b){var y=n.prototype.screen.call(this,o,s,c,d,h);return BV(y,b),y},r.prototype.identify=function(o,s,c,d,h){var b=n.prototype.identify.call(this,o,s,c,d);return BV(b,h),b},r.prototype.group=function(o,s,c,d,h){var b=n.prototype.group.call(this,o,s,c,d);return BV(b,h),b},r.prototype.alias=function(o,s,c,d,h){var b=n.prototype.alias.call(this,o,s,c,d);return BV(b,h),b},r}(ylr),Jlr=function(n){return"addMiddleware"in n&&n.type==="destination"},TI={getItem:function(){},setItem:function(){},removeItem:function(){}};try{TI=nZe()&&window.localStorage?window.localStorage:TI}catch(n){console.warn("Unable to access localStorage",n)}function Aan(n){var r=TI.getItem(n);return(r?JSON.parse(r):[]).map(function(o){return new $4(o.event,o.id)})}function Zlr(n,r){var o=Aan(n),s=Vg(Vg([],r,!0),o,!0),c=s.reduce(function(d,h){var b;return zr(zr({},d),(b={},b[h.id]=h,b))},{});TI.setItem(n,JSON.stringify(Object.values(c)))}function Ian(n){var r=TI.getItem(n);return r?JSON.parse(r):{}}function Qlr(n,r){var o=Ian(n);TI.setItem(n,JSON.stringify(zr(zr({},o),r)))}function Zjt(n){TI.removeItem(n)}var edr=function(){return new Date().getTime()};function ZVe(n,r,o){o===void 0&&(o=0);var s=50,c="persisted-queue:v1:".concat(n,":lock"),d=function(E){return new Date().getTime()>E},h=TI.getItem(c),b=h?JSON.parse(h):null,y=b===null||d(b);if(y){TI.setItem(c,JSON.stringify(edr()+s)),r(),TI.removeItem(c);return}!y&&o<3?setTimeout(function(){ZVe(n,r,o+1)},s):console.error("Unable to retrieve lock")}var iZe=function(n){iy(r,n);function r(o,s){var c=n.call(this,o,[])||this,d="persisted-queue:v1:".concat(s,":items"),h="persisted-queue:v1:".concat(s,":seen"),b=[],y={};return ZVe(s,function(){try{b=Aan(d),y=Ian(h),Zjt(d),Zjt(h),c.queue=Vg(Vg([],b,!0),c.queue,!0),c.seen=zr(zr({},y),c.seen)}catch(E){console.error(E)}}),window.addEventListener("pagehide",function(){if(c.todo>0){var E=Vg(Vg([],c.queue,!0),c.future,!0);try{ZVe(s,function(){Zlr(d,E),Qlr(h,c.seen)})}catch(k){console.error(k)}}}),c}return r}(eZe),tdr=function(n){iy(r,n);function r(o){return n.call(this,typeof o=="string"?new iZe(4,o):o)||this}return r.prototype.flush=function(){return wo(this,void 0,void 0,function(){return yo(this,function(o){return Xie()?[2,[]]:[2,n.prototype.flush.call(this)]})})},r}(Plr);function oZe(n){for(var r=n.constructor.prototype,o=0,s=Object.getOwnPropertyNames(r);o<s.length;o++){var c=s[o];if(c!=="constructor"){var d=Object.getOwnPropertyDescriptor(n.constructor.prototype,c);d&&typeof d.value=="function"&&(n[c]=n[c].bind(n))}}return n}/*! js-cookie v3.0.1 | MIT */function o0e(n){for(var r=1;r<arguments.length;r++){var o=arguments[r];for(var s in o)n[s]=o[s]}return n}var ndr={read:function(n){return n[0]==='"'&&(n=n.slice(1,-1)),n.replace(/(%[\dA-F]{2})+/gi,decodeURIComponent)},write:function(n){return encodeURIComponent(n).replace(/%(2[346BF]|3[AC-F]|40|5[BDE]|60|7[BCD])/g,decodeURIComponent)}};function QVe(n,r){function o(c,d,h){if(!(typeof document>"u")){h=o0e({},r,h),typeof h.expires=="number"&&(h.expires=new Date(Date.now()+h.expires*864e5)),h.expires&&(h.expires=h.expires.toUTCString()),c=encodeURIComponent(c).replace(/%(2[346B]|5E|60|7C)/g,decodeURIComponent).replace(/[()]/g,escape);var b="";for(var y in h)h[y]&&(b+="; "+y,h[y]!==!0&&(b+="="+h[y].split(";")[0]));return document.cookie=c+"="+n.write(d,c)+b}}function s(c){if(!(typeof document>"u"||arguments.length&&!c)){for(var d=document.cookie?document.cookie.split("; "):[],h={},b=0;b<d.length;b++){var y=d[b].split("="),E=y.slice(1).join("=");try{var k=decodeURIComponent(y[0]);if(h[k]=n.read(E,k),c===k)break}catch{}}return c?h[c]:h}}return Object.create({set:o,get:s,remove:function(c,d){o(c,"",o0e({},d,{expires:-1}))},withAttributes:function(c){return QVe(this.converter,o0e({},this.attributes,c))},withConverter:function(c){return QVe(o0e({},this.converter,c),this.attributes)}},{attributes:{value:Object.freeze(r)},converter:{value:Object.freeze(n)}})}var O4=QVe(ndr,{path:"/"});function rdr(n){var r=n.hostname,o=r.split("."),s=o[o.length-1],c=[];if(o.length===4&&parseInt(s,10)>0||o.length<=1)return c;for(var d=o.length-2;d>=0;--d)c.push(o.slice(d).join("."));return c}function idr(n){try{return new URL(n)}catch{return}}function Oan(n){var r=idr(n);if(r)for(var o=rdr(r),s=0;s<o.length;++s){var c="__tld__",d=o[s],h={domain:"."+d};try{if(O4.set(c,"1",h),O4.get(c))return O4.remove(c,h),d}catch{return}}}var odr=365,Nan=function(){function n(r){r===void 0&&(r=n.defaults),this.options=zr(zr({},n.defaults),r)}return Object.defineProperty(n,"defaults",{get:function(){return{maxage:odr,domain:Oan(window.location.href),path:"/",sameSite:"Lax"}},enumerable:!1,configurable:!0}),n.prototype.opts=function(){return{sameSite:this.options.sameSite,expires:this.options.maxage,domain:this.options.domain,path:this.options.path,secure:this.options.secure}},n.prototype.get=function(r){var o;try{var s=O4.get(r);if(s==null)return null;try{return(o=JSON.parse(s))!==null&&o!==void 0?o:null}catch{return s??null}}catch{return null}},n.prototype.set=function(r,o){typeof o=="string"?O4.set(r,o,this.opts()):o===null?O4.remove(r,this.opts()):O4.set(r,JSON.stringify(o),this.opts())},n.prototype.remove=function(r){return O4.remove(r,this.opts())},n}(),adr=function(){function n(){}return n.prototype.localStorageWarning=function(r,o){console.warn("Unable to access ".concat(r,", localStorage may be ").concat(o))},n.prototype.get=function(r){var o;try{var s=localStorage.getItem(r);if(s===null)return null;try{return(o=JSON.parse(s))!==null&&o!==void 0?o:null}catch{return s??null}}catch{return this.localStorageWarning(r,"unavailable"),null}},n.prototype.set=function(r,o){try{localStorage.setItem(r,JSON.stringify(o))}catch{this.localStorageWarning(r,"full")}},n.prototype.remove=function(r){try{return localStorage.removeItem(r)}catch{this.localStorageWarning(r,"unavailable")}},n}(),aZe=function(){function n(){this.cache={}}return n.prototype.get=function(r){var o;return(o=this.cache[r])!==null&&o!==void 0?o:null},n.prototype.set=function(r,o){this.cache[r]=o},n.prototype.remove=function(r){delete this.cache[r]},n}(),Ww={Cookie:"cookie",LocalStorage:"localStorage",Memory:"memory"};function Ran(n){return n&&n.stores&&Array.isArray(n.stores)&&n.stores.every(function(r){return Object.values(Ww).includes(r)})}function sdr(n){return typeof n=="object"&&n.name!==void 0}var x$e=function(n,r,o,s){console.warn("".concat(n.constructor.name,": Can't ").concat(r,' key "').concat(o,'" | Err: ').concat(s))},BB=function(){function n(r){this.stores=r}return n.prototype.get=function(r){for(var o=null,s=0,c=this.stores;s<c.length;s++){var d=c[s];try{if(o=d.get(r),o!=null)return o}catch(h){x$e(d,"get",r,h)}}return null},n.prototype.set=function(r,o){this.stores.forEach(function(s){try{s.set(r,o)}catch(c){x$e(s,"set",r,c)}})},n.prototype.clear=function(r){this.stores.forEach(function(o){try{o.remove(r)}catch(s){x$e(o,"remove",r,s)}})},n.prototype.getAndSync=function(r){var o=this.get(r),s=typeof o=="number"?o.toString():o;return this.set(r,s),s},n}();function eWe(n){var r=n.map(function(o){var s,c;switch(sdr(o)?(s=o.name,c=o.settings):s=o,s){case Ww.Cookie:return new Nan(c);case Ww.LocalStorage:return new adr;case Ww.Memory:return new aZe;default:throw new Error("Unknown Store Type: ".concat(o))}});return r}function Pan(n,r){return n.map(function(o){return r&&o===Ww.Cookie?{name:o,settings:r}:o})}var aB={persist:!0,cookie:{key:"ajs_user_id",oldKey:"ajs_user"},localStorage:{key:"ajs_user_traits"}},Lan=function(){function n(r,o){r===void 0&&(r=aB);var s=this,c,d,h,b;this.options={},this.id=function(E){if(s.options.disable)return null;var k=s.identityStore.getAndSync(s.idKey);if(E!==void 0){s.identityStore.set(s.idKey,E);var x=E!==k&&k!==null&&E!==null;x&&s.anonymousId(null)}var O=s.identityStore.getAndSync(s.idKey);if(O)return O;var P=s.legacyUserStore.get(aB.cookie.oldKey);return P?typeof P=="object"?P.id:P:null},this.anonymousId=function(E){var k,x;if(s.options.disable)return null;if(E===void 0){var O=(k=s.identityStore.getAndSync(s.anonKey))!==null&&k!==void 0?k:(x=s.legacySIO())===null||x===void 0?void 0:x[0];if(O)return O}return E===null?(s.identityStore.set(s.anonKey,null),s.identityStore.getAndSync(s.anonKey)):(s.identityStore.set(s.anonKey,E??tZe()),s.identityStore.getAndSync(s.anonKey))},this.traits=function(E){var k;if(!s.options.disable)return E===null&&(E={}),E&&s.traitsStore.set(s.traitsKey,E??{}),(k=s.traitsStore.get(s.traitsKey))!==null&&k!==void 0?k:{}},this.options=zr(zr({},aB),r),this.cookieOptions=o,this.idKey=(d=(c=r.cookie)===null||c===void 0?void 0:c.key)!==null&&d!==void 0?d:aB.cookie.key,this.traitsKey=(b=(h=r.localStorage)===null||h===void 0?void 0:h.key)!==null&&b!==void 0?b:aB.localStorage.key,this.anonKey="ajs_anonymous_id",this.identityStore=this.createStorage(this.options,o),this.legacyUserStore=this.createStorage(this.options,o,function(E){return E===Ww.Cookie}),this.traitsStore=this.createStorage(this.options,o,function(E){return E!==Ww.Cookie});var y=this.legacyUserStore.get(aB.cookie.oldKey);y&&typeof y=="object"&&(y.id&&this.id(y.id),y.traits&&this.traits(y.traits)),oZe(this)}return n.prototype.legacySIO=function(){var r=this.legacyUserStore.get("_sio");if(!r)return null;var o=r.split("----"),s=o[0],c=o[1];return[s,c]},n.prototype.identify=function(r,o){if(!this.options.disable){o=o??{};var s=this.id();(s===null||s===r)&&(o=zr(zr({},this.traits()),o)),r&&this.id(r),this.traits(o)}},n.prototype.logout=function(){this.anonymousId(null),this.id(null),this.traits({})},n.prototype.reset=function(){this.logout(),this.identityStore.clear(this.idKey),this.identityStore.clear(this.anonKey),this.traitsStore.clear(this.traitsKey)},n.prototype.load=function(){return new n(this.options,this.cookieOptions)},n.prototype.save=function(){return!0},n.prototype.createStorage=function(r,o,s){var c=[Ww.LocalStorage,Ww.Cookie,Ww.Memory];return r.disable?new BB([]):r.persist?(r.storage!==void 0&&r.storage!==null&&Ran(r.storage)&&(c=r.storage.stores),r.localStorageFallbackDisabled&&(c=c.filter(function(d){return d!==Ww.LocalStorage})),s&&(c=c.filter(s)),new BB(eWe(Pan(c,o)))):new BB([new aZe])},n.defaults=aB,n}(),Qjt={persist:!0,cookie:{key:"ajs_group_id"},localStorage:{key:"ajs_group_properties"}},cdr=function(n){iy(r,n);function r(o,s){o===void 0&&(o=Qjt);var c=n.call(this,zr(zr({},Qjt),o),s)||this;return c.anonymousId=function(d){},oZe(c),c}return r}(Lan),sZe="analytics";function Dan(){return window[sZe]}function udr(n){sZe=n}function ldr(n){window[sZe]=n}var ddr=function(n){return typeof n=="object"&&n!==null&&"then"in n&&typeof n.then=="function"},Man=function(n,r,o){o.getAndRemove(n).forEach(function(s){iEe(r,s).catch(console.error)})},fdr=function(n,r){return wo(void 0,void 0,void 0,function(){var o,s,c;return yo(this,function(d){switch(d.label){case 0:o=0,s=r.getAndRemove("addSourceMiddleware"),d.label=1;case 1:return o<s.length?(c=s[o],[4,iEe(n,c).catch(console.error)]):[3,4];case 2:d.sent(),d.label=3;case 3:return o++,[3,1];case 4:return[2]}})})},hdr=function(n,r){return wo(void 0,void 0,void 0,function(){var o,s,c;return yo(this,function(d){switch(d.label){case 0:o=0,s=r.getAndRemove("register"),d.label=1;case 1:return o<s.length?(c=s[o],[4,iEe(n,c).catch(console.error)]):[3,4];case 2:d.sent(),d.label=3;case 3:return o++,[3,1];case 4:return[2]}})})},pdr=Man.bind(void 0,"on"),gdr=Man.bind(void 0,"setAnonymousId"),mdr=function(n,r){Object.keys(r.calls).forEach(function(o){r.getAndRemove(o).forEach(function(s){setTimeout(function(){iEe(n,s).catch(console.error)},0)})})},UV=function(n){if(jan(n)){var r=n.pop();return xan(r)}},jan=function(n){var r=n[n.length-1];return Glr(r)},tWe=function(){function n(r,o,s,c){s===void 0&&(s=function(){}),c===void 0&&(c=console.error),this.method=r,this.resolve=s,this.reject=c,this.called=!1,this.args=o}return n}(),bdr=function(){function n(){for(var r=[],o=0;o<arguments.length;o++)r[o]=arguments[o];this._callMap={},this.add.apply(this,r)}return Object.defineProperty(n.prototype,"calls",{get:function(){return this._pushSnippetWindowBuffer(),this._callMap},set:function(r){this._callMap=r},enumerable:!1,configurable:!0}),n.prototype.get=function(r){var o;return(o=this.calls[r])!==null&&o!==void 0?o:[]},n.prototype.getAndRemove=function(r){var o=this.get(r);return this.calls[r]=[],o},n.prototype.add=function(){for(var r=this,o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];o.forEach(function(c){var d=["track","screen","alias","group","page","identify"];d.includes(c.method)&&!jan(c.args)&&(c.args=Vg(Vg([],c.args,!0),[Can()],!1)),r.calls[c.method]?r.calls[c.method].push(c):r.calls[c.method]=[c]})},n.prototype.clear=function(){this._pushSnippetWindowBuffer(),this.calls={}},n.prototype.toArray=function(){var r;return(r=[]).concat.apply(r,Object.values(this.calls))},n.prototype._pushSnippetWindowBuffer=function(){},n}();function iEe(n,r){return wo(this,void 0,void 0,function(){var o,s;return yo(this,function(c){switch(c.label){case 0:return c.trys.push([0,3,,4]),r.called?[2,void 0]:(r.called=!0,o=n[r.method].apply(n,r.args),ddr(o)?[4,o]:[3,2]);case 1:c.sent(),c.label=2;case 2:return r.resolve(o),[3,4];case 3:return s=c.sent(),r.reject(s),[3,4];case 4:return[2]}})})}var vdr=function(){function n(r){var o=this;this.trackSubmit=this._createMethod("trackSubmit"),this.trackClick=this._createMethod("trackClick"),this.trackLink=this._createMethod("trackLink"),this.pageView=this._createMethod("pageview"),this.identify=this._createMethod("identify"),this.reset=this._createMethod("reset"),this.group=this._createMethod("group"),this.track=this._createMethod("track"),this.ready=this._createMethod("ready"),this.alias=this._createMethod("alias"),this.debug=this._createChainableMethod("debug"),this.page=this._createMethod("page"),this.once=this._createChainableMethod("once"),this.off=this._createChainableMethod("off"),this.on=this._createChainableMethod("on"),this.addSourceMiddleware=this._createMethod("addSourceMiddleware"),this.setAnonymousId=this._createMethod("setAnonymousId"),this.addDestinationMiddleware=this._createMethod("addDestinationMiddleware"),this.screen=this._createMethod("screen"),this.register=this._createMethod("register"),this.deregister=this._createMethod("deregister"),this.user=this._createMethod("user"),this.VERSION=rEe,this._preInitBuffer=new bdr,this._promise=r(this._preInitBuffer),this._promise.then(function(s){var c=s[0],d=s[1];o.instance=c,o.ctx=d}).catch(function(){})}return n.prototype.then=function(){for(var r,o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return(r=this._promise).then.apply(r,o)},n.prototype.catch=function(){for(var r,o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return(r=this._promise).catch.apply(r,o)},n.prototype.finally=function(){for(var r,o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return(r=this._promise).finally.apply(r,o)},n.prototype._createMethod=function(r){var o=this;return function(){for(var s,c=[],d=0;d<arguments.length;d++)c[d]=arguments[d];if(o.instance){var h=(s=o.instance)[r].apply(s,c);return Promise.resolve(h)}return new Promise(function(b,y){o._preInitBuffer.add(new tWe(r,c,b,y))})}},n.prototype._createChainableMethod=function(r){var o=this;return function(){for(var s,c=[],d=0;d<arguments.length;d++)c[d]=arguments[d];return o.instance?((s=o.instance)[r].apply(s,c),o):(o._preInitBuffer.add(new tWe(r,c)),o)}},n}(),dre={},sB={},zV={},C$e={exports:{}},e$t;function oEe(){return e$t||(e$t=1,function(n){n.exports=r(o),n.exports.find=n.exports,n.exports.replace=function(b,y,E,k){return r(c).call(this,b,y,E,k),b},n.exports.del=function(b,y,E){return r(s).call(this,b,y,null,E),b};function r(b){return function(y,E,k,x){var O=x&&h(x.normalizer)?x.normalizer:d;E=O(E);for(var P,D=!1;!D;)$();function $(){for(P in y){var U=O(P);if(E.indexOf(U)===0){var B=E.substr(U.length);if(B.charAt(0)==="."||B.length===0){E=B.substr(1);var G=y[P];if(G==null){D=!0;return}if(!E.length){D=!0;return}y=G;return}}}P=void 0,D=!0}if(P)return y==null?y:b(y,P,k)}}function o(b,y){if(b.hasOwnProperty(y))return b[y]}function s(b,y){return b.hasOwnProperty(y)&&delete b[y],b}function c(b,y,E){return b.hasOwnProperty(y)&&(b[y]=E),b}function d(b){return b.replace(/[^a-zA-Z0-9\.]+/g,"").toLowerCase()}function h(b){return typeof b=="function"}}(C$e)),C$e.exports}var t$t;function wdr(){if(t$t)return zV;t$t=1;var n=zV&&zV.__importDefault||function(c){return c&&c.__esModule?c:{default:c}};Object.defineProperty(zV,"__esModule",{value:!0});var r=n(oEe());function o(c,d){return function(){var h=this.traits(),b=this.properties?this.properties():{};return r.default(h,"address."+c)||r.default(h,c)||(d?r.default(h,"address."+d):null)||(d?r.default(h,d):null)||r.default(b,"address."+c)||r.default(b,c)||(d?r.default(b,"address."+d):null)||(d?r.default(b,d):null)}}function s(c){c.zip=o("postalCode","zip"),c.country=o("country"),c.street=o("street"),c.state=o("state"),c.city=o("city"),c.region=o("region")}return zV.default=s,zV}var fre={},n$t;function ydr(){if(n$t)return fre;n$t=1,Object.defineProperty(fre,"__esModule",{value:!0}),fre.clone=void 0;function n(r){if(typeof r!="object")return r;if(Object.prototype.toString.call(r)==="[object Object]"){var o={};for(var s in r)Object.prototype.hasOwnProperty.call(r,s)&&(o[s]=n(r[s]));return o}else return Array.isArray(r)?r.map(n):r}return fre.clone=n,fre}var a0e={},r$t;function Edr(){if(r$t)return a0e;r$t=1,Object.defineProperty(a0e,"__esModule",{value:!0});var n={Salesforce:!0};function r(o){return!n[o]}return a0e.default=r,a0e}var s0e={},i$t;function $an(){if(i$t)return s0e;i$t=1;var n=/^(\d{4})(?:-?(\d{2})(?:-?(\d{2}))?)?(?:([ T])(\d{2}):?(\d{2})(?::?(\d{2})(?:[,\.](\d{1,}))?)?(?:(Z)|([+\-])(\d{2})(?::?(\d{2}))?)?)?$/;return s0e.parse=function(r){var o=[1,5,6,7,11,12],s=n.exec(r),c=0;if(!s)return new Date(r);for(var d=0,h;h=o[d];d++)s[h]=parseInt(s[h],10)||0;s[2]=parseInt(s[2],10)||1,s[3]=parseInt(s[3],10)||1,s[2]--,s[8]=s[8]?(s[8]+"00").substring(0,3):0,s[4]===" "?c=new Date().getTimezoneOffset():s[9]!=="Z"&&s[10]&&(c=s[11]*60+s[12],s[10]==="+"&&(c=0-c));var b=Date.UTC(s[1],s[2],s[3],s[5],s[6]+c,s[7],s[8]);return new Date(b)},s0e.is=function(r,o){return typeof r!="string"||o&&/^\d{4}-\d{2}-\d{2}/.test(r)===!1?!1:n.test(r)},s0e}var c0e={},o$t;function Sdr(){if(o$t)return c0e;o$t=1;var n=/\d{13}/;return c0e.is=function(r){return n.test(r)},c0e.parse=function(r){return r=parseInt(r,10),new Date(r)},c0e}var u0e={},a$t;function _dr(){if(a$t)return u0e;a$t=1;var n=/\d{10}/;return u0e.is=function(r){return n.test(r)},u0e.parse=function(r){var o=parseInt(r,10)*1e3;return new Date(o)},u0e}var A$e,s$t;function cZe(){if(s$t)return A$e;s$t=1;var n=$an(),r=Sdr(),o=_dr(),s=Object.prototype,c=s.toString;function d(y){return c.call(y)==="[object Date]"}function h(y){return c.call(y)==="[object Number]"}A$e=function(E){return d(E)?E:h(E)?new Date(b(E)):n.is(E)?n.parse(E):r.is(E)?r.parse(E):o.is(E)?o.parse(E):new Date(E)};function b(y){return y<315576e5?y*1e3:y}return A$e}var I$e,c$t;function kdr(){if(c$t)return I$e;c$t=1;var n=$an();I$e=r;function r(c,d){return d===void 0&&(d=!0),c&&typeof c=="object"?o(c,d):Array.isArray(c)?s(c,d):n.is(c,d)?n.parse(c):c}function o(c,d){return Object.keys(c).forEach(function(h){c[h]=r(c[h],d)}),c}function s(c,d){return c.forEach(function(h,b){c[b]=r(h,d)}),c}return I$e}var u$t;function pU(){if(u$t)return sB;u$t=1;var n=sB&&sB.__importDefault||function(k){return k&&k.__esModule?k:{default:k}};Object.defineProperty(sB,"__esModule",{value:!0}),sB.Facade=void 0;var r=n(wdr()),o=ydr(),s=n(Edr()),c=n(cZe()),d=n(oEe()),h=n(kdr());function b(k,x){x=x||{},this.raw=o.clone(k),"clone"in x||(x.clone=!0),x.clone&&(k=o.clone(k)),"traverse"in x||(x.traverse=!0),"timestamp"in k?k.timestamp=c.default(k.timestamp):k.timestamp=new Date,x.traverse&&h.default(k),this.opts=x,this.obj=k}sB.Facade=b;var y=b.prototype;y.proxy=function(k){var x=k.split(".");k=x.shift();var O=this[k]||this.obj[k];return O&&(typeof O=="function"&&(O=O.call(this)||{}),x.length===0?this.opts.clone?E(O):O:(O=d.default(O,x.join(".")),this.opts.clone?E(O):O))},y.field=function(k){var x=this.obj[k];return this.opts.clone?E(x):x},b.proxy=function(k){return function(){return this.proxy(k)}},b.field=function(k){return function(){return this.field(k)}},b.multi=function(k){return function(){var x=this.proxy(k+"s");if(Array.isArray(x))return x;var O=this.proxy(k);return O&&(O=[this.opts.clone?o.clone(O):O]),O||[]}},b.one=function(k){return function(){var x=this.proxy(k);if(x)return x;var O=this.proxy(k+"s");if(Array.isArray(O))return O[0]}},y.json=function(){var k=this.opts.clone?o.clone(this.obj):this.obj;return this.type&&(k.type=this.type()),k},y.rawEvent=function(){return this.raw},y.options=function(k){var x=this.obj.options||this.obj.context||{},O=this.opts.clone?o.clone(x):x;if(!k)return O;if(this.enabled(k)){var P=this.integrations(),D=P[k]||d.default(P,k);return typeof D!="object"&&(D=d.default(this.options(),k)),typeof D=="object"?D:{}}},y.context=y.options,y.enabled=function(k){var x=this.proxy("options.providers.all");typeof x!="boolean"&&(x=this.proxy("options.all")),typeof x!="boolean"&&(x=this.proxy("integrations.all")),typeof x!="boolean"&&(x=!0);var O=x&&s.default(k),P=this.integrations();if(P.providers&&P.providers.hasOwnProperty(k)&&(O=P.providers[k]),P.hasOwnProperty(k)){var D=P[k];typeof D=="boolean"?O=D:O=!0}return!!O},y.integrations=function(){return this.obj.integrations||this.proxy("options.providers")||this.options()},y.active=function(){var k=this.proxy("options.active");return k==null&&(k=!0),k},y.anonymousId=function(){return this.field("anonymousId")||this.field("sessionId")},y.sessionId=y.anonymousId,y.groupId=b.proxy("options.groupId"),y.traits=function(k){var x=this.proxy("options.traits")||{},O=this.userId();k=k||{},O&&(x.id=O);for(var P in k)if(Object.prototype.hasOwnProperty.call(k,P)){var D=this[P]==null?this.proxy("options.traits."+P):this[P]();if(D==null)continue;x[k[P]]=D,delete x[P]}return x},y.library=function(){var k=this.proxy("options.library");return k?typeof k=="string"?{name:k,version:null}:k:{name:"unknown",version:null}},y.device=function(){var k=this.proxy("context.device");(typeof k!="object"||k===null)&&(k={});var x=this.library().name;return k.type||(x.indexOf("ios")>-1&&(k.type="ios"),x.indexOf("android")>-1&&(k.type="android")),k},y.userAgent=b.proxy("context.userAgent"),y.timezone=b.proxy("context.timezone"),y.timestamp=b.field("timestamp"),y.channel=b.field("channel"),y.ip=b.proxy("context.ip"),y.userId=b.field("userId"),r.default(y);function E(k){return o.clone(k)}return sB}var cB={},l0e={exports:{}},l$t;function gU(){return l$t||(l$t=1,typeof Object.create=="function"?l0e.exports=function(r,o){o&&(r.super_=o,r.prototype=Object.create(o.prototype,{constructor:{value:r,enumerable:!1,writable:!0,configurable:!0}}))}:l0e.exports=function(r,o){if(o){r.super_=o;var s=function(){};s.prototype=o.prototype,r.prototype=new s,r.prototype.constructor=r}}),l0e.exports}var d$t;function Tdr(){if(d$t)return cB;d$t=1;var n=cB&&cB.__importDefault||function(c){return c&&c.__esModule?c:{default:c}};Object.defineProperty(cB,"__esModule",{value:!0}),cB.Alias=void 0;var r=n(gU()),o=pU();function s(c,d){o.Facade.call(this,c,d)}return cB.Alias=s,r.default(s,o.Facade),s.prototype.action=function(){return"alias"},s.prototype.type=s.prototype.action,s.prototype.previousId=function(){return this.field("previousId")||this.field("from")},s.prototype.from=s.prototype.previousId,s.prototype.userId=function(){return this.field("userId")||this.field("to")},s.prototype.to=s.prototype.userId,cB}var uB={},d0e={},f$t;function aEe(){if(f$t)return d0e;f$t=1,Object.defineProperty(d0e,"__esModule",{value:!0});var n=/.+\@.+\..+/;function r(o){return n.test(o)}return d0e.default=r,d0e}var h$t;function xdr(){if(h$t)return uB;h$t=1;var n=uB&&uB.__importDefault||function(b){return b&&b.__esModule?b:{default:b}};Object.defineProperty(uB,"__esModule",{value:!0}),uB.Group=void 0;var r=n(gU()),o=n(aEe()),s=n(cZe()),c=pU();function d(b,y){c.Facade.call(this,b,y)}uB.Group=d,r.default(d,c.Facade);var h=d.prototype;return h.action=function(){return"group"},h.type=h.action,h.groupId=c.Facade.field("groupId"),h.created=function(){var b=this.proxy("traits.createdAt")||this.proxy("traits.created")||this.proxy("properties.createdAt")||this.proxy("properties.created");if(b)return s.default(b)},h.email=function(){var b=this.proxy("traits.email");if(b)return b;var y=this.groupId();if(o.default(y))return y},h.traits=function(b){var y=this.properties(),E=this.groupId();b=b||{},E&&(y.id=E);for(var k in b)if(Object.prototype.hasOwnProperty.call(b,k)){var x=this[k]==null?this.proxy("traits."+k):this[k]();if(x==null)continue;y[b[k]]=x,delete y[k]}return y},h.name=c.Facade.proxy("traits.name"),h.industry=c.Facade.proxy("traits.industry"),h.employees=c.Facade.proxy("traits.employees"),h.properties=function(){return this.field("traits")||this.field("properties")||{}},uB}var lB={},p$t;function Fan(){if(p$t)return lB;p$t=1;var n=lB&&lB.__importDefault||function(E){return E&&E.__esModule?E:{default:E}};Object.defineProperty(lB,"__esModule",{value:!0}),lB.Identify=void 0;var r=pU(),o=n(oEe()),s=n(gU()),c=n(aEe()),d=n(cZe()),h=function(E){return E.trim()};function b(E,k){r.Facade.call(this,E,k)}lB.Identify=b,s.default(b,r.Facade);var y=b.prototype;return y.action=function(){return"identify"},y.type=y.action,y.traits=function(E){var k=this.field("traits")||{},x=this.userId();E=E||{},x&&(k.id=x);for(var O in E){var P=this[O]==null?this.proxy("traits."+O):this[O]();P!=null&&(k[E[O]]=P,O!==E[O]&&delete k[O])}return k},y.email=function(){var E=this.proxy("traits.email");if(E)return E;var k=this.userId();if(c.default(k))return k},y.created=function(){var E=this.proxy("traits.created")||this.proxy("traits.createdAt");if(E)return d.default(E)},y.companyCreated=function(){var E=this.proxy("traits.company.created")||this.proxy("traits.company.createdAt");if(E)return d.default(E)},y.companyName=function(){return this.proxy("traits.company.name")},y.name=function(){var E=this.proxy("traits.name");if(typeof E=="string")return h(E);var k=this.firstName(),x=this.lastName();if(k&&x)return h(k+" "+x)},y.firstName=function(){var E=this.proxy("traits.firstName");if(typeof E=="string")return h(E);var k=this.proxy("traits.name");if(typeof k=="string")return h(k).split(" ")[0]},y.lastName=function(){var E=this.proxy("traits.lastName");if(typeof E=="string")return h(E);var k=this.proxy("traits.name");if(typeof k=="string"){var x=h(k).indexOf(" ");if(x!==-1)return h(k.substr(x+1))}},y.uid=function(){return this.userId()||this.username()||this.email()},y.description=function(){return this.proxy("traits.description")||this.proxy("traits.background")},y.age=function(){var E=this.birthday(),k=o.default(this.traits(),"age");if(k!=null)return k;if(E instanceof Date){var x=new Date;return x.getFullYear()-E.getFullYear()}},y.avatar=function(){var E=this.traits();return o.default(E,"avatar")||o.default(E,"photoUrl")||o.default(E,"avatarUrl")},y.position=function(){var E=this.traits();return o.default(E,"position")||o.default(E,"jobTitle")},y.username=r.Facade.proxy("traits.username"),y.website=r.Facade.one("traits.website"),y.websites=r.Facade.multi("traits.website"),y.phone=r.Facade.one("traits.phone"),y.phones=r.Facade.multi("traits.phone"),y.address=r.Facade.proxy("traits.address"),y.gender=r.Facade.proxy("traits.gender"),y.birthday=r.Facade.proxy("traits.birthday"),lB}var dB={},g$t;function uZe(){if(g$t)return dB;g$t=1;var n=dB&&dB.__importDefault||function(E){return E&&E.__esModule?E:{default:E}};Object.defineProperty(dB,"__esModule",{value:!0}),dB.Track=void 0;var r=n(gU()),o=pU(),s=Fan(),c=n(aEe()),d=n(oEe());function h(E,k){o.Facade.call(this,E,k)}dB.Track=h,r.default(h,o.Facade);var b=h.prototype;b.action=function(){return"track"},b.type=b.action,b.event=o.Facade.field("event"),b.value=o.Facade.proxy("properties.value"),b.category=o.Facade.proxy("properties.category"),b.id=o.Facade.proxy("properties.id"),b.productId=function(){return this.proxy("properties.product_id")||this.proxy("properties.productId")},b.promotionId=function(){return this.proxy("properties.promotion_id")||this.proxy("properties.promotionId")},b.cartId=function(){return this.proxy("properties.cart_id")||this.proxy("properties.cartId")},b.checkoutId=function(){return this.proxy("properties.checkout_id")||this.proxy("properties.checkoutId")},b.paymentId=function(){return this.proxy("properties.payment_id")||this.proxy("properties.paymentId")},b.couponId=function(){return this.proxy("properties.coupon_id")||this.proxy("properties.couponId")},b.wishlistId=function(){return this.proxy("properties.wishlist_id")||this.proxy("properties.wishlistId")},b.reviewId=function(){return this.proxy("properties.review_id")||this.proxy("properties.reviewId")},b.orderId=function(){return this.proxy("properties.id")||this.proxy("properties.order_id")||this.proxy("properties.orderId")},b.sku=o.Facade.proxy("properties.sku"),b.tax=o.Facade.proxy("properties.tax"),b.name=o.Facade.proxy("properties.name"),b.price=o.Facade.proxy("properties.price"),b.total=o.Facade.proxy("properties.total"),b.repeat=o.Facade.proxy("properties.repeat"),b.coupon=o.Facade.proxy("properties.coupon"),b.shipping=o.Facade.proxy("properties.shipping"),b.discount=o.Facade.proxy("properties.discount"),b.shippingMethod=function(){return this.proxy("properties.shipping_method")||this.proxy("properties.shippingMethod")},b.paymentMethod=function(){return this.proxy("properties.payment_method")||this.proxy("properties.paymentMethod")},b.description=o.Facade.proxy("properties.description"),b.plan=o.Facade.proxy("properties.plan"),b.subtotal=function(){var E=d.default(this.properties(),"subtotal"),k=this.total()||this.revenue();if(E)return E;if(!k)return 0;if(this.total()){var x=this.tax();x&&(k-=x),x=this.shipping(),x&&(k-=x),x=this.discount(),x&&(k+=x)}return k},b.products=function(){var E=this.properties(),k=d.default(E,"products");return Array.isArray(k)?k.filter(function(x){return x!==null}):[]},b.quantity=function(){var E=this.obj.properties||{};return E.quantity||1},b.currency=function(){var E=this.obj.properties||{};return E.currency||"USD"},b.referrer=function(){return this.proxy("context.referrer.url")||this.proxy("context.page.referrer")||this.proxy("properties.referrer")},b.query=o.Facade.proxy("options.query"),b.properties=function(E){var k=this.field("properties")||{};E=E||{};for(var x in E)if(Object.prototype.hasOwnProperty.call(E,x)){var O=this[x]==null?this.proxy("properties."+x):this[x]();if(O==null)continue;k[E[x]]=O,delete k[x]}return k},b.username=function(){return this.proxy("traits.username")||this.proxy("properties.username")||this.userId()||this.sessionId()},b.email=function(){var E=this.proxy("traits.email")||this.proxy("properties.email")||this.proxy("options.traits.email");if(E)return E;var k=this.userId();if(c.default(k))return k},b.revenue=function(){var E=this.proxy("properties.revenue"),k=this.event(),x=/^[ _]?completed[ _]?order[ _]?|^[ _]?order[ _]?completed[ _]?$/i;return!E&&k&&k.match(x)&&(E=this.proxy("properties.total")),y(E)},b.cents=function(){var E=this.revenue();return typeof E!="number"?this.value()||0:E*100},b.identify=function(){var E=this.json();return E.traits=this.traits(),new s.Identify(E,this.opts)};function y(E){if(E){if(typeof E=="number")return E;if(typeof E=="string"&&(E=E.replace(/\$/g,""),E=parseFloat(E),!isNaN(E)))return E}}return dB}var fB={},m$t;function Ban(){if(m$t)return fB;m$t=1;var n=fB&&fB.__importDefault||function(b){return b&&b.__esModule?b:{default:b}};Object.defineProperty(fB,"__esModule",{value:!0}),fB.Page=void 0;var r=n(gU()),o=pU(),s=uZe(),c=n(aEe());function d(b,y){o.Facade.call(this,b,y)}fB.Page=d,r.default(d,o.Facade);var h=d.prototype;return h.action=function(){return"page"},h.type=h.action,h.category=o.Facade.field("category"),h.name=o.Facade.field("name"),h.title=o.Facade.proxy("properties.title"),h.path=o.Facade.proxy("properties.path"),h.url=o.Facade.proxy("properties.url"),h.referrer=function(){return this.proxy("context.referrer.url")||this.proxy("context.page.referrer")||this.proxy("properties.referrer")},h.properties=function(b){var y=this.field("properties")||{},E=this.category(),k=this.name();b=b||{},E&&(y.category=E),k&&(y.name=k);for(var x in b)if(Object.prototype.hasOwnProperty.call(b,x)){var O=this[x]==null?this.proxy("properties."+x):this[x]();if(O==null)continue;y[b[x]]=O,x!==b[x]&&delete y[x]}return y},h.email=function(){var b=this.proxy("context.traits.email")||this.proxy("properties.email");if(b)return b;var y=this.userId();if(c.default(y))return y},h.fullName=function(){var b=this.category(),y=this.name();return y&&b?b+" "+y:y},h.event=function(b){return b?"Viewed "+b+" Page":"Loaded a Page"},h.track=function(b){var y=this.json();return y.event=this.event(b),y.timestamp=this.timestamp(),y.properties=this.properties(),new s.Track(y,this.opts)},fB}var hB={},b$t;function Cdr(){if(b$t)return hB;b$t=1;var n=hB&&hB.__importDefault||function(d){return d&&d.__esModule?d:{default:d}};Object.defineProperty(hB,"__esModule",{value:!0}),hB.Screen=void 0;var r=n(gU()),o=Ban(),s=uZe();function c(d,h){o.Page.call(this,d,h)}return hB.Screen=c,r.default(c,o.Page),c.prototype.action=function(){return"screen"},c.prototype.type=c.prototype.action,c.prototype.event=function(d){return d?"Viewed "+d+" Screen":"Loaded a Screen"},c.prototype.track=function(d){var h=this.json();return h.event=this.event(d),h.timestamp=this.timestamp(),h.properties=this.properties(),new s.Track(h,this.opts)},hB}var pB={},v$t;function Adr(){if(v$t)return pB;v$t=1;var n=pB&&pB.__importDefault||function(c){return c&&c.__esModule?c:{default:c}};Object.defineProperty(pB,"__esModule",{value:!0}),pB.Delete=void 0;var r=n(gU()),o=pU();function s(c,d){o.Facade.call(this,c,d)}return pB.Delete=s,r.default(s,o.Facade),s.prototype.type=function(){return"delete"},pB}var w$t;function Idr(){return w$t||(w$t=1,function(n){var r=dre&&dre.__assign||function(){return r=Object.assign||function(k){for(var x,O=1,P=arguments.length;O<P;O++){x=arguments[O];for(var D in x)Object.prototype.hasOwnProperty.call(x,D)&&(k[D]=x[D])}return k},r.apply(this,arguments)};Object.defineProperty(n,"__esModule",{value:!0}),n.Delete=n.Screen=n.Page=n.Track=n.Identify=n.Group=n.Alias=n.Facade=void 0;var o=pU();Object.defineProperty(n,"Facade",{enumerable:!0,get:function(){return o.Facade}});var s=Tdr();Object.defineProperty(n,"Alias",{enumerable:!0,get:function(){return s.Alias}});var c=xdr();Object.defineProperty(n,"Group",{enumerable:!0,get:function(){return c.Group}});var d=Fan();Object.defineProperty(n,"Identify",{enumerable:!0,get:function(){return d.Identify}});var h=uZe();Object.defineProperty(n,"Track",{enumerable:!0,get:function(){return h.Track}});var b=Ban();Object.defineProperty(n,"Page",{enumerable:!0,get:function(){return b.Page}});var y=Cdr();Object.defineProperty(n,"Screen",{enumerable:!0,get:function(){return y.Screen}});var E=Adr();Object.defineProperty(n,"Delete",{enumerable:!0,get:function(){return E.Delete}}),n.default=r(r({},o.Facade),{Alias:s.Alias,Group:c.Group,Identify:d.Identify,Track:h.Track,Page:b.Page,Screen:y.Screen,Delete:E.Delete})}(dre)),dre}var gB=Idr();function gwe(n,r){var o=new gB.Facade(n,r);return n.type==="track"&&(o=new gB.Track(n,r)),n.type==="identify"&&(o=new gB.Identify(n,r)),n.type==="page"&&(o=new gB.Page(n,r)),n.type==="alias"&&(o=new gB.Alias(n,r)),n.type==="group"&&(o=new gB.Group(n,r)),n.type==="screen"&&(o=new gB.Screen(n,r)),Object.defineProperty(o,"obj",{value:n,writable:!0}),o}var Odr=function(n){var r=!1;window.addEventListener("pagehide",function(){r||(r=!0,n(r))}),document.addEventListener("visibilitychange",function(){if(document.visibilityState=="hidden"){if(r)return;r=!0}else r=!1;n(r)})},Uan=function(n){iy(r,n);function r(o,s){var c=n.call(this,o)||this;return c.retryTimeout=s,c.name="RateLimitError",c}return r}(Error),zan=function(n){return zr({"Content-Type":"text/plain"},typeof n=="function"?n():n)},Ndr=500,Rdr=64;function lZe(n){var r=encodeURI(JSON.stringify(n)).split(/%..|./).length-1;return r/1024}function Pdr(n){return lZe(n)>=Ndr-50}function Ldr(n){return lZe(n)>=Rdr-10}function Ddr(n){var r=[],o=0;return n.forEach(function(s){var c=lZe(r[o]);c>=64&&o++,r[o]?r[o].push(s):r[o]=[s]}),r}function Mdr(n,r){var o,s,c=[],d=!1,h=(o=r?.size)!==null&&o!==void 0?o:10,b=(s=r?.timeout)!==null&&s!==void 0?s:5e3,y=0;function E(D){var $;if(D.length!==0){var U=($=D[0])===null||$===void 0?void 0:$.writeKey,B=D.map(function(G){var K=G;K.sentAt;var X=van(K,["sentAt"]);return X});return nEe("https://".concat(n,"/b"),{credentials:r?.credentials,keepalive:r?.keepalive||d,headers:zan(r?.headers),method:"post",body:JSON.stringify({writeKey:U,batch:B,sentAt:new Date().toISOString()}),priority:r?.priority}).then(function(G){var K;if(G.status>=500)throw new Error("Bad response from server: ".concat(G.status));if(G.status===429){var X=(K=G.headers)===null||K===void 0?void 0:K.get("x-ratelimit-reset"),ae=typeof X=="string"?parseInt(X)*1e3:b;throw new Uan("Rate limit exceeded: ".concat(G.status),ae)}})}}function k(D){var $;return D===void 0&&(D=1),wo(this,void 0,void 0,function(){var U;return yo(this,function(B){return c.length?(U=c,c=[],[2,($=E(U))===null||$===void 0?void 0:$.catch(function(G){var K,X=$4.system();X.log("error","Error sending batch",G),D<=((K=r?.maxRetries)!==null&&K!==void 0?K:10)&&(G.name==="RateLimitError"&&(y=G.retryTimeout),c.push.apply(c,U),c.map(function(ae){if("_metadata"in ae){var oe=ae;oe._metadata=zr(zr({},oe._metadata),{retryCount:D})}}),O(D+1))})]):[2]})})}var x;function O(D){D===void 0&&(D=1),!x&&(x=setTimeout(function(){x=void 0,k(D).catch(console.error)},y||b),y=0)}Odr(function(D){if(d=D,d&&c.length){var $=Ddr(c).map(E);Promise.all($).catch(console.error)}});function P(D,$){return wo(this,void 0,void 0,function(){var U;return yo(this,function(B){return c.push($),U=c.length>=h||Pdr(c)||r?.keepalive&&Ldr(c),[2,U||d?k():O()]})})}return{dispatch:P}}function jdr(n){function r(o,s){return nEe(o,{credentials:n?.credentials,keepalive:n?.keepalive,headers:zan(n?.headers),method:"post",body:JSON.stringify(s),priority:n?.priority}).then(function(c){var d;if(c.status>=500)throw new Error("Bad response from server: ".concat(c.status));if(c.status===429){var h=(d=c.headers)===null||d===void 0?void 0:d.get("x-ratelimit-reset"),b=h?parseInt(h)*1e3:5e3;throw new Uan("Rate limit exceeded: ".concat(c.status),b)}})}return{dispatch:r}}function $dr(n,r,o,s,c){var d,h=n.user();delete r.options,r.writeKey=o?.apiKey,r.userId=r.userId||h.id(),r.anonymousId=r.anonymousId||h.anonymousId(),r.sentAt=new Date;var b=n.queue.failedInitializations||[];b.length>0&&(r._metadata={failedInitializations:b}),c!=null&&(c.attempts>1&&(r._metadata=zr(zr({},r._metadata),{retryCount:c.attempts})),c.attempts++);var y=[],E=[];for(var k in s){var x=s[k];k==="Segment.io"&&y.push(k),x.bundlingStatus==="bundled"&&y.push(k),x.bundlingStatus==="unbundled"&&E.push(k)}for(var O=0,P=o?.unbundledIntegrations||[];O<P.length;O++){var D=P[O];E.includes(D)||E.push(D)}var $=(d=o?.maybeBundledConfigIds)!==null&&d!==void 0?d:{},U=[];return y.sort().forEach(function(B){var G;((G=$[B])!==null&&G!==void 0?G:[]).forEach(function(K){U.push(K)})}),o?.addBundledMetadata!==!1&&(r._metadata=zr(zr({},r._metadata),{bundled:y.sort(),unbundled:E.sort(),bundledIds:U})),r}var Fdr=function(n,r){return wo(void 0,void 0,void 0,function(){var o;return yo(this,function(s){return o=function(c){return wo(void 0,void 0,void 0,function(){var d;return yo(this,function(h){switch(h.label){case 0:return n(c)?(d=o,[4,r()]):[3,2];case 1:return[2,d.apply(void 0,[h.sent()])];case 2:return[2]}})})},[2,o(void 0)]})})};function Bdr(n,r){return wo(this,void 0,void 0,function(){var o,s=this;return yo(this,function(c){switch(c.label){case 0:return o=[],Xie()?[2,r]:[4,Fdr(function(){return r.length>0&&!Xie()},function(){return wo(s,void 0,void 0,function(){var d,h,b;return yo(this,function(y){switch(y.label){case 0:return d=r.pop(),d?[4,lie(d,n)]:[2];case 1:return h=y.sent(),b=h instanceof $4,b||o.push(d),[2]}})})})];case 1:return c.sent(),o.map(function(d){return r.pushWithBackoff(d)}),[2,r]}})})}function HV(n,r,o,s){var c=this;n||setTimeout(function(){return wo(c,void 0,void 0,function(){var d,h;return yo(this,function(b){switch(b.label){case 0:return d=!0,[4,Bdr(o,r)];case 1:return h=b.sent(),d=!1,r.todo>0&&s(d,h,o,s),[2]}})})},Math.random()*5e3)}function Udr(n,r){var o,s,c,d,h=n.user();return r.previousId=(c=(s=(o=r.previousId)!==null&&o!==void 0?o:r.from)!==null&&s!==void 0?s:h.id())!==null&&c!==void 0?c:h.anonymousId(),r.userId=(d=r.userId)!==null&&d!==void 0?d:r.to,delete r.from,delete r.to,r}var zdr=function(n){return n.name==="Segment.io"};function Hdr(n,r,o){var s,c,d;window.addEventListener("pagehide",function(){b.push.apply(b,Array.from(y)),y.clear()});var h=(s=r?.apiKey)!==null&&s!==void 0?s:"",b=n.options.disableClientPersistence?new eZe(n.queue.queue.maxAttempts,[]):new iZe(n.queue.queue.maxAttempts,"".concat(h,":dest-Segment.io")),y=new Set,E=!1,k=(c=r?.apiHost)!==null&&c!==void 0?c:San,x=(d=r?.protocol)!==null&&d!==void 0?d:"https",O="".concat(x,"://").concat(k),P=r?.deliveryStrategy,D=P&&"strategy"in P&&P.strategy==="batching"?Mdr(k,P.config):jdr(P?.config);function $(B){return wo(this,void 0,void 0,function(){var G,K;return yo(this,function(X){return Xie()?(b.push(B),HV(E,b,U,HV),[2,B]):(y.add(B),G=B.event.type.charAt(0),K=gwe(B.event).json(),B.event.type==="track"&&delete K.traits,B.event.type==="alias"&&(K=Udr(n,K)),b.getAttempts(B)>=b.maxAttempts?(y.delete(B),[2,B]):[2,D.dispatch("".concat(O,"/").concat(G),$dr(n,K,r,o,B)).then(function(){return B}).catch(function(ae){if(B.log("error","Error sending event",ae),ae.name==="RateLimitError"){var oe=ae.retryTimeout;b.pushWithBackoff(B,oe)}else b.pushWithBackoff(B);return HV(E,b,U,HV),B}).finally(function(){y.delete(B)})])})})}var U={metadata:{writeKey:h,apiHost:k,protocol:x},name:"Segment.io",type:"destination",version:"0.1.0",isLoaded:function(){return!0},load:function(){return Promise.resolve()},track:$,identify:$,page:$,alias:$,group:$,screen:$};return b.todo&&HV(E,b,U,HV),U}var ZD="This is being deprecated and will be not be available in future releases of Analytics JS",O$e=rZe(),N$e=O$e?.analytics;function Gdr(n,r,o){r===void 0&&(r=!1),o===void 0&&(o=!1);var s=r?10:1,c=o?new eZe(s,[]):new iZe(s,n);return new tdr(c)}var qdr=function(){function n(r,o){var s;this.timeout=300,this._getSegmentPluginMetadata=function(){var d;return(d=o.plugins.find(zdr))===null||d===void 0?void 0:d.metadata},this.writeKey=r.writeKey;var c={integrations:{"Segment.io":{apiKey:""}}};this.cdnSettings=(s=r.cdnSettings)!==null&&s!==void 0?s:c,this.cdnURL=r.cdnURL}return Object.defineProperty(n.prototype,"apiHost",{get:function(){var r,o;return(o=(r=this._getSegmentPluginMetadata)===null||r===void 0?void 0:r.call(this))===null||o===void 0?void 0:o.apiHost},enumerable:!1,configurable:!0}),n}();function rI(){console.warn(ZD)}var Han=function(n){iy(r,n);function r(o,s,c,d,h){var b=this,y,E;b=n.call(this)||this,b._debug=!1,b.initialized=!1,b.user=function(){return b._user},b.init=b.initialize.bind(b),b.log=rI,b.addIntegrationMiddleware=rI,b.listeners=rI,b.addEventListener=rI,b.removeAllListeners=rI,b.removeListener=rI,b.removeEventListener=rI,b.hasListeners=rI,b.add=rI,b.addIntegration=rI;var k=s?.cookie,x=(y=s?.disableClientPersistence)!==null&&y!==void 0?y:!1;b.queue=c??Gdr("".concat(o.writeKey,":event-queue"),s?.retryQueue,x),b.settings=new qdr(o,b.queue);var O=s?.storage;return b._universalStorage=b.createStore(x,O,k),b._user=d??new Lan(zr({persist:!x,storage:s?.storage},s?.user),k).load(),b._group=h??new cdr(zr({persist:!x,storage:s?.storage},s?.group),k).load(),b.eventFactory=new Xlr(b._user),b.integrations=(E=s?.integrations)!==null&&E!==void 0?E:{},b.options=s??{},oZe(b),b}return r.prototype.createStore=function(o,s,c){return o?new BB([new aZe]):s&&Ran(s)?new BB(eWe(Pan(s.stores,c))):new BB(eWe([Ww.LocalStorage,{name:Ww.Cookie,settings:c},Ww.Memory]))},Object.defineProperty(r.prototype,"storage",{get:function(){return this._universalStorage},enumerable:!1,configurable:!0}),r.prototype.track=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d,h,b,y,E,k,x=this;return yo(this,function(O){return c=UV(o),d=Mlr.apply(void 0,o),h=d[0],b=d[1],y=d[2],E=d[3],k=this.eventFactory.track(h,b,y,this.integrations,c),[2,this._dispatch(k,E).then(function(P){return x.emit("track",h,P.event.properties,P.event.options),P})]})})},r.prototype.page=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d,h,b,y,E,k,x,O=this;return yo(this,function(P){return c=UV(o),d=Xjt.apply(void 0,o),h=d[0],b=d[1],y=d[2],E=d[3],k=d[4],x=this.eventFactory.page(h,b,y,E,this.integrations,c),[2,this._dispatch(x,k).then(function(D){return O.emit("page",h,b,D.event.properties,D.event.options),D})]})})},r.prototype.identify=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d,h,b,y,E,k,x=this;return yo(this,function(O){return c=UV(o),d=Jjt(this._user).apply(void 0,o),h=d[0],b=d[1],y=d[2],E=d[3],this._user.identify(h,b),k=this.eventFactory.identify(this._user.id(),this._user.traits(),y,this.integrations,c),[2,this._dispatch(k,E).then(function(P){return x.emit("identify",P.event.userId,P.event.traits,P.event.options),P})]})})},r.prototype.group=function(){for(var o=this,s=[],c=0;c<arguments.length;c++)s[c]=arguments[c];var d=UV(s);if(s.length===0)return this._group;var h=Jjt(this._group).apply(void 0,s),b=h[0],y=h[1],E=h[2],k=h[3];this._group.identify(b,y);var x=this._group.id(),O=this._group.traits(),P=this.eventFactory.group(x,O,E,this.integrations,d);return this._dispatch(P,k).then(function(D){return o.emit("group",D.event.groupId,D.event.traits,D.event.options),D})},r.prototype.alias=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d,h,b,y,E,k,x=this;return yo(this,function(O){return c=UV(o),d=jlr.apply(void 0,o),h=d[0],b=d[1],y=d[2],E=d[3],k=this.eventFactory.alias(h,b,y,this.integrations,c),[2,this._dispatch(k,E).then(function(P){return x.emit("alias",h,b,P.event.options),P})]})})},r.prototype.screen=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d,h,b,y,E,k,x,O=this;return yo(this,function(P){return c=UV(o),d=Xjt.apply(void 0,o),h=d[0],b=d[1],y=d[2],E=d[3],k=d[4],x=this.eventFactory.screen(h,b,y,E,this.integrations,c),[2,this._dispatch(x,k).then(function(D){return O.emit("screen",h,b,D.event.properties,D.event.options),D})]})})},r.prototype.trackClick=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d;return yo(this,function(h){switch(h.label){case 0:return[4,Ix(()=>import("./auto-track-Bxl6oEUq.js"),[])];case 1:return c=h.sent(),[2,(d=c.link).call.apply(d,Vg([this],o,!1))]}})})},r.prototype.trackLink=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d;return yo(this,function(h){switch(h.label){case 0:return[4,Ix(()=>import("./auto-track-Bxl6oEUq.js"),[])];case 1:return c=h.sent(),[2,(d=c.link).call.apply(d,Vg([this],o,!1))]}})})},r.prototype.trackSubmit=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d;return yo(this,function(h){switch(h.label){case 0:return[4,Ix(()=>import("./auto-track-Bxl6oEUq.js"),[])];case 1:return c=h.sent(),[2,(d=c.form).call.apply(d,Vg([this],o,!1))]}})})},r.prototype.trackForm=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d;return yo(this,function(h){switch(h.label){case 0:return[4,Ix(()=>import("./auto-track-Bxl6oEUq.js"),[])];case 1:return c=h.sent(),[2,(d=c.form).call.apply(d,Vg([this],o,!1))]}})})},r.prototype.register=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d,h=this;return yo(this,function(b){switch(b.label){case 0:return c=$4.system(),d=o.map(function(y){return h.queue.register(c,y,h)}),[4,Promise.all(d)];case 1:return b.sent(),[2,c]}})})},r.prototype.deregister=function(){for(var o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];return wo(this,void 0,void 0,function(){var c,d,h=this;return yo(this,function(b){switch(b.label){case 0:return c=$4.system(),d=o.map(function(y){var E=h.queue.plugins.find(function(k){return k.name===y});if(E)return h.queue.deregister(c,E,h);c.log("warn","plugin ".concat(y," not found"))}),[4,Promise.all(d)];case 1:return b.sent(),[2,c]}})})},r.prototype.debug=function(o){return o===!1&&localStorage.getItem("debug")&&localStorage.removeItem("debug"),this._debug=o,this},r.prototype.reset=function(){this._user.reset(),this._group.reset(),this.emit("reset")},r.prototype.timeout=function(o){this.settings.timeout=o},r.prototype._dispatch=function(o,s){return wo(this,void 0,void 0,function(){var c;return yo(this,function(d){return c=new $4(o),c.stats.increment("analytics_js.invoke",1,[o.type]),Xie()&&!this.options.retryQueue?[2,c]:[2,Dlr(c,this.queue,this,{callback:s,debug:this._debug,timeout:this.settings.timeout})]})})},r.prototype.addSourceMiddleware=function(o){return wo(this,void 0,void 0,function(){var s=this;return yo(this,function(c){switch(c.label){case 0:return[4,this.queue.criticalTasks.run(function(){return wo(s,void 0,void 0,function(){var d,h,b;return yo(this,function(y){switch(y.label){case 0:return[4,Ix(()=>Promise.resolve().then(()=>cfr),void 0)];case 1:return d=y.sent().sourceMiddlewarePlugin,h={},this.queue.plugins.forEach(function(E){if(E.type==="destination")return h[E.name]=!0}),b=d(o,h),[4,this.register(b)];case 2:return y.sent(),[2]}})})})];case 1:return c.sent(),[2,this]}})})},r.prototype.addDestinationMiddleware=function(o){for(var s=[],c=1;c<arguments.length;c++)s[c-1]=arguments[c];return this.queue.plugins.filter(Jlr).forEach(function(d){(o==="*"||d.name.toLowerCase()===o.toLowerCase())&&d.addMiddleware.apply(d,s)}),Promise.resolve(this)},r.prototype.setAnonymousId=function(o){return this._user.anonymousId(o)},r.prototype.queryString=function(o){return wo(this,void 0,void 0,function(){var s;return yo(this,function(c){switch(c.label){case 0:return this.options.useQueryString===!1?[2,[]]:[4,Ix(()=>import("./index-GiVmMbST.js"),[])];case 1:return s=c.sent().queryString,[2,s(this,o)]}})})},r.prototype.use=function(o){return o(this),this},r.prototype.ready=function(o){return o===void 0&&(o=function(s){return s}),wo(this,void 0,void 0,function(){return yo(this,function(s){return[2,Promise.all(this.queue.plugins.map(function(c){return c.ready?c.ready():Promise.resolve()})).then(function(c){return o(c),c})]})})},r.prototype.noConflict=function(){return console.warn(ZD),ldr(N$e??this),this},r.prototype.normalize=function(o){return console.warn(ZD),this.eventFactory.normalize(o)},Object.defineProperty(r.prototype,"failedInitializations",{get:function(){return console.warn(ZD),this.queue.failedInitializations},enumerable:!1,configurable:!0}),Object.defineProperty(r.prototype,"VERSION",{get:function(){return rEe},enumerable:!1,configurable:!0}),r.prototype.initialize=function(o,s){return wo(this,void 0,void 0,function(){return yo(this,function(c){return console.warn(ZD),[2,Promise.resolve(this)]})})},r.prototype.pageview=function(o){return wo(this,void 0,void 0,function(){return yo(this,function(s){switch(s.label){case 0:return console.warn(ZD),[4,this.page({path:o})];case 1:return s.sent(),[2,this]}})})},Object.defineProperty(r.prototype,"plugins",{get:function(){var o;return console.warn(ZD),(o=this._plugins)!==null&&o!==void 0?o:{}},enumerable:!1,configurable:!0}),Object.defineProperty(r.prototype,"Integrations",{get:function(){console.warn(ZD);var o=this.queue.plugins.filter(function(s){return s.type==="destination"}).reduce(function(s,c){var d="".concat(c.name.toLowerCase().replace(".","").split(" ").join("-"),"Integration"),h=window[d];if(!h)return s;var b=h.Integration;return b?(s[c.name]=b,s):(s[c.name]=h,s)},{});return o},enumerable:!1,configurable:!0}),r.prototype.push=function(o){var s=this,c=o.shift();c&&!s[c]||s[c].apply(this,o)},r}(QJe),y$t=function(n){iy(r,n);function r(){var o=n.call(this,{writeKey:""},{disableClientPersistence:!0})||this;return o.initialized=!0,o}return r}(Han),E$t={};function Gan(){return typeof process>"u"||!E$t?{}:E$t}var Vdr=/(https:\/\/.*)\/analytics\.js\/v1\/(?:.*?)\/(?:platform|analytics.*)?/,Wdr=function(){var n,r=Array.prototype.slice.call(document.querySelectorAll("script"));return r.forEach(function(o){var s,c=(s=o.getAttribute("src"))!==null&&s!==void 0?s:"",d=Vdr.exec(c);d&&d[1]&&(n=d[1])}),n},gve,Kdr=function(){var n,r=gve??((n=Dan())===null||n===void 0?void 0:n._cdn);return r},Ydr=function(n){var r=Dan();r&&(r._cdn=n),gve=n},dZe=function(){var n=Kdr();if(n)return n;var r=Wdr();return r||"https://cdn.segment.com"},zni=function(){var n=dZe();return"".concat(n,"/next-integrations")};function Xdr(n,r){var o,s=Object.entries((o=r.integrations)!==null&&o!==void 0?o:{}).reduce(function(c,d){var h,b,y=d[0],E=d[1];return typeof E=="object"?zr(zr({},c),(h={},h[y]=E,h)):zr(zr({},c),(b={},b[y]={},b))},{});return Object.entries(n.integrations).reduce(function(c,d){var h,b=d[0],y=d[1];return zr(zr({},c),(h={},h[b]=zr(zr({},y),s[b]),h))},{})}function Jdr(n){try{return decodeURIComponent(n.replace(/\+/g," "))}catch{return n}}function Zdr(n){return wo(this,void 0,void 0,function(){var r;return yo(this,function(o){return r=navigator.userAgentData,r?n?[2,r.getHighEntropyValues(n).catch(function(){return r.toJSON()})]:[2,r.toJSON()]:[2,void 0]})})}var hre;function Qdr(){if(hre)return hre;var n=Oan(window.location.href);return hre={expires:31536e6,secure:!1,path:"/"},n&&(hre.domain=n),hre}function efr(n){var r={btid:"dataxu",urid:"millennial-media"};n.startsWith("?")&&(n=n.substring(1)),n=n.replace(/\?/g,"&");for(var o=n.split("&"),s=0,c=o;s<c.length;s++){var d=c[s],h=d.split("="),b=h[0],y=h[1];if(r[b])return{id:y,type:r[b]}}}function tfr(n){return n.startsWith("?")&&(n=n.substring(1)),n=n.replace(/\?/g,"&"),n.split("&").reduce(function(r,o){var s=o.split("="),c=s[0],d=s[1],h=d===void 0?"":d;if(c.includes("utm_")&&c.length>4){var b=c.slice(4);b==="campaign"&&(b="name"),r[b]=Jdr(h)}return r},{})}function nfr(){var n=O4.get("_ga");if(n&&n.startsWith("amp"))return n}function rfr(n,r,o){var s,c=new BB(o?[]:[new Nan(Qdr())]),d=c.get("s:context.referrer"),h=(s=efr(n))!==null&&s!==void 0?s:d;h&&(r&&(r.referrer=zr(zr({},r.referrer),h)),c.set("s:context.referrer",h))}var ifr=function(n){try{var r=new URLSearchParams;return Object.entries(n).forEach(function(o){var s=o[0],c=o[1];Array.isArray(c)?c.forEach(function(d){return r.append(s,d)}):r.append(s,c)}),r.toString()}catch{return""}},ofr=function(){function n(){var r=this;this.name="Page Enrichment",this.type="before",this.version="0.1.0",this.isLoaded=function(){return!0},this.load=function(o,s){return wo(r,void 0,void 0,function(){var c;return yo(this,function(d){switch(d.label){case 0:this.instance=s,d.label=1;case 1:return d.trys.push([1,3,,4]),c=this,[4,Zdr(this.instance.options.highEntropyValuesClientHints)];case 2:return c.userAgentData=d.sent(),[3,4];case 3:return d.sent(),[3,4];case 4:return[2,Promise.resolve()]}})})},this.enrich=function(o){var s,c,d=o.event.context,h=d.page.search||"",b=typeof h=="object"?ifr(h):h;d.userAgent=navigator.userAgent,d.userAgentData=r.userAgentData;var y=navigator.userLanguage||navigator.language;typeof d.locale>"u"&&typeof y<"u"&&(d.locale=y),(s=d.library)!==null&&s!==void 0||(d.library={name:"analytics.js",version:"".concat("npm:next","-").concat(rEe)}),b&&!d.campaign&&(d.campaign=tfr(b));var E=nfr();E&&(d.amp={id:E}),rfr(b,d,(c=r.instance.options.disableClientPersistence)!==null&&c!==void 0?c:!1);try{d.timezone=Intl.DateTimeFormat().resolvedOptions().timeZone}catch{}return o},this.track=this.enrich,this.identify=this.enrich,this.page=this.enrich,this.group=this.enrich,this.alias=this.enrich,this.screen=this.enrich}return n}(),afr=new ofr;function qan(n){var r=Array.prototype.slice.call(window.document.querySelectorAll("script"));return r.find(function(o){return o.src===n})}function R$e(n,r){var o=qan(n);if(o!==void 0){var s=o?.getAttribute("status");if(s==="loaded")return Promise.resolve(o);if(s==="loading")return new Promise(function(c,d){o.addEventListener("load",function(){return c(o)}),o.addEventListener("error",function(h){return d(h)})})}return new Promise(function(c,d){var h,b=window.document.createElement("script");b.type="text/javascript",b.src=n,b.async=!0,b.setAttribute("status","loading");for(var y=0,E=Object.entries({});y<E.length;y++){var k=E[y],x=k[0],O=k[1];b.setAttribute(x,O)}b.onload=function(){b.onerror=b.onload=null,b.setAttribute("status","loaded"),c(b)},b.onerror=function(){b.onerror=b.onload=null,b.setAttribute("status","error"),d(new Error("Failed to load ".concat(n)))};var P=window.document.querySelector("script");P?(h=P.parentElement)===null||h===void 0||h.insertBefore(b,P):window.document.head.appendChild(b)})}function Hni(n){var r=qan(n);return r!==void 0&&r.remove(),Promise.resolve()}function Van(n,r,o){return wo(this,void 0,void 0,function(){function s(E,k){return wo(this,void 0,void 0,function(){var x,O,P;return yo(this,function(D){switch(D.label){case 0:return x=!1,O=null,[4,k({payload:gwe(E,{clone:!0,traverse:!1}),integration:n,next:function($){x=!0,$===null&&(O=null),$&&(O=$.obj)}})];case 1:return D.sent(),!x&&O!==null&&(O=O,O.integrations=zr(zr({},E.integrations),(P={},P[n]=!1,P))),[2,O]}})})}var c,d,h,b,y;return yo(this,function(E){switch(E.label){case 0:c=gwe(r,{clone:!0,traverse:!1}).rawEvent(),d=0,h=o,E.label=1;case 1:return d<h.length?(b=h[d],[4,s(c,b)]):[3,4];case 2:if(y=E.sent(),y===null)return[2,null];c=y,E.label=3;case 3:return d++,[3,1];case 4:return[2,c]}})})}function sfr(n,r){function o(s){return wo(this,void 0,void 0,function(){var c;return yo(this,function(d){switch(d.label){case 0:return c=!1,[4,n({payload:gwe(s.event,{clone:!0,traverse:!1}),integrations:r??{},next:function(h){c=!0,h&&(s.event=h.obj)}})];case 1:if(d.sent(),!c)throw new eK({retry:!1,type:"middleware_cancellation",reason:"Middleware `next` function skipped"});return[2,s]}})})}return{name:"Source Middleware ".concat(n.name),type:"before",version:"0.1.0",isLoaded:function(){return!0},load:function(s){return Promise.resolve(s)},track:o,page:o,screen:o,identify:o,alias:o,group:o}}const cfr=Object.freeze(Object.defineProperty({__proto__:null,applyDestinationMiddleware:Van,sourceMiddlewarePlugin:sfr},Symbol.toStringTag,{value:"Module"}));function f0e(n,r){var o=r.methodName,s=r.integrationName,c=r.type,d=r.didError,h=d===void 0?!1:d;n.stats.increment("analytics_js.integration.invoke".concat(h?".error":""),1,["method:".concat(o),"integration_name:".concat(s),"type:".concat(c)])}var ufr=function(){function n(r,o){this.version="1.0.0",this.alternativeNames=[],this.loadPromise=wan(),this.middleware=[],this.alias=this._createMethod("alias"),this.group=this._createMethod("group"),this.identify=this._createMethod("identify"),this.page=this._createMethod("page"),this.screen=this._createMethod("screen"),this.track=this._createMethod("track"),this.action=o,this.name=r,this.type=o.type,this.alternativeNames.push(o.name)}return n.prototype.addMiddleware=function(){for(var r,o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];this.type==="destination"&&(r=this.middleware).push.apply(r,o)},n.prototype.transform=function(r){return wo(this,void 0,void 0,function(){var o;return yo(this,function(s){switch(s.label){case 0:return[4,Van(this.name,r.event,this.middleware)];case 1:return o=s.sent(),o===null&&r.cancel(new eK({retry:!1,reason:"dropped by destination middleware"})),[2,new $4(o)]}})})},n.prototype._createMethod=function(r){var o=this;return function(s){return wo(o,void 0,void 0,function(){var c,d;return yo(this,function(h){switch(h.label){case 0:return this.action[r]?(c=s,this.type!=="destination"?[3,2]:[4,this.transform(s)]):[2,s];case 1:c=h.sent(),h.label=2;case 2:return h.trys.push([2,5,,6]),[4,this.ready()];case 3:if(!h.sent())throw new Error("Something prevented the destination from getting ready");return f0e(s,{integrationName:this.action.name,methodName:r,type:"action"}),[4,this.action[r](c)];case 4:return h.sent(),[3,6];case 5:throw d=h.sent(),f0e(s,{integrationName:this.action.name,methodName:r,type:"action",didError:!0}),d;case 6:return[2,s]}})})}},n.prototype.isLoaded=function(){return this.action.isLoaded()},n.prototype.ready=function(){return wo(this,void 0,void 0,function(){return yo(this,function(r){switch(r.label){case 0:return r.trys.push([0,2,,3]),[4,this.loadPromise.promise];case 1:return r.sent(),[2,!0];case 2:return r.sent(),[2,!1];case 3:return[2]}})})},n.prototype.load=function(r,o){return wo(this,void 0,void 0,function(){var s,c,d,h;return yo(this,function(b){switch(b.label){case 0:if(this.loadPromise.isSettled())return[2,this.loadPromise.promise];b.label=1;case 1:return b.trys.push([1,3,,4]),f0e(r,{integrationName:this.action.name,methodName:"load",type:"action"}),s=this.action.load(r,o),d=(c=this.loadPromise).resolve,[4,s];case 2:return d.apply(c,[b.sent()]),[2,s];case 3:throw h=b.sent(),f0e(r,{integrationName:this.action.name,methodName:"load",type:"action",didError:!0}),this.loadPromise.reject(h),h;case 4:return[2]}})})},n.prototype.unload=function(r,o){var s,c;return(c=(s=this.action).unload)===null||c===void 0?void 0:c.call(s,r,o)},n}();function lfr(n){if(!Array.isArray(n))throw new Error("Not a valid list of plugins");var r=["load","isLoaded","name","version","type"];return n.forEach(function(o){r.forEach(function(s){var c;if(o[s]===void 0)throw new Error("Plugin: ".concat((c=o.name)!==null&&c!==void 0?c:"unknown"," missing required function ").concat(s))})}),!0}function dfr(n,r){var o=n[r.creationName],s=n[r.name];return n.All===!1&&!o&&!s||o===!1||s===!1}function ffr(n,r){return wo(this,void 0,void 0,function(){var o,s,c,d,h,b;return yo(this,function(y){switch(y.label){case 0:if(y.trys.push([0,9,,10]),o=new RegExp("https://cdn.segment.(com|build)"),s=dZe(),!r)return[3,6];c=n.url.split("/"),d=c[c.length-2],h=n.url.replace(d,btoa(d).replace(/=/g,"")),y.label=1;case 1:return y.trys.push([1,3,,5]),[4,R$e(h.replace(o,s))];case 2:return y.sent(),[3,5];case 3:return y.sent(),[4,R$e(n.url.replace(o,s))];case 4:return y.sent(),[3,5];case 5:return[3,8];case 6:return[4,R$e(n.url.replace(o,s))];case 7:y.sent(),y.label=8;case 8:return typeof window[n.libraryName]=="function"?[2,window[n.libraryName]]:[3,10];case 9:throw b=y.sent(),console.error("Failed to create PluginFactory",n),b;case 10:return[2]}})})}function hfr(n,r,o,s,c,d){var h,b,y;return wo(this,void 0,void 0,function(){var E,k,x,O=this;return yo(this,function(P){switch(P.label){case 0:return E=[],k=(b=(h=n.middlewareSettings)===null||h===void 0?void 0:h.routingRules)!==null&&b!==void 0?b:[],x=((y=n.remotePlugins)!==null&&y!==void 0?y:[]).map(function(D){return wo(O,void 0,void 0,function(){var $,U,B,G,K,X,ae;return yo(this,function(oe){switch(oe.label){case 0:if(dfr(r,D))return[2];oe.label=1;case 1:return oe.trys.push([1,6,,7]),U=d?.find(function(he){var ye=he.pluginName;return ye===D.name}),U?[3,3]:[4,ffr(D,s?.obfuscate)];case 2:U=oe.sent(),oe.label=3;case 3:return $=U,$?(B=o[D.name],[4,$(zr(zr({},D.settings),B))]):[3,5];case 4:G=oe.sent(),K=Array.isArray(G)?G:[G],lfr(K),X=k.filter(function(he){return he.destinationName===D.creationName}),K.forEach(function(he){var ye=new ufr(D.creationName,he);X.length&&c&&ye.addMiddleware(c),E.push(ye)}),oe.label=5;case 5:return[3,7];case 6:return ae=oe.sent(),console.warn("Failed to load Remote Plugin",ae),[3,7];case 7:return[2]}})})}),[4,Promise.all(x)];case 1:return P.sent(),[2,E.filter(Boolean)]}})})}var P$e,S$t,pfr=rZe(),_$t=(P$e=(S$t=pfr).__SEGMENT_INSPECTOR__)!==null&&P$e!==void 0?P$e:S$t.__SEGMENT_INSPECTOR__={},gfr=function(n){var r;return(r=_$t.attach)===null||r===void 0?void 0:r.call(_$t,n)};function mfr(n,r){return nEe("".concat(r,"/v1/projects/").concat(n,"/settings")).then(function(o){return o.ok?o.json():o.text().then(function(s){throw new Error(s)})}).catch(function(o){throw console.error(o.message),o})}function bfr(n){return Gan().NODE_ENV!=="test"&&Object.keys(n.integrations).length>1}function vfr(n){var r,o,s;return Gan().NODE_ENV!=="test"&&((s=(o=(r=n.middlewareSettings)===null||r===void 0?void 0:r.routingRules)===null||o===void 0?void 0:o.length)!==null&&s!==void 0?s:0)>0}function wfr(n,r){gdr(n,r),pdr(n,r)}function yfr(n,r,o){return wo(this,void 0,void 0,function(){return yo(this,function(s){switch(s.label){case 0:return[4,Sfr(n,r)];case 1:return s.sent(),mdr(n,o),[2]}})})}var Efr=function(){var n,r,o=(n=window.location.hash)!==null&&n!==void 0?n:"",s=(r=window.location.search)!==null&&r!==void 0?r:"",c=s.length?s:o.replace(/(?=#).*(?=\?)/,"");return c},Sfr=function(n,r){return wo(void 0,void 0,void 0,function(){return yo(this,function(o){switch(o.label){case 0:return r.includes("ajs_")?[4,n.queryString(r).catch(console.error)]:[3,2];case 1:o.sent(),o.label=2;case 2:return[2]}})})};function _fr(n,r,o,s,c,d,h){var b,y,E;return c===void 0&&(c=[]),wo(this,void 0,void 0,function(){var k,x,O,P,D,$,U,B,G,K,X,ae,oe,he,ye,we=this;return yo(this,function(ge){switch(ge.label){case 0:return wfr(o,h),k=c?.filter(function(Te){return typeof Te=="object"}),x=c?.filter(function(Te){return typeof Te=="function"&&typeof Te.pluginName=="string"}),vfr(r)?[4,Ix(()=>import("./index-7O0TZzl3.js"),[]).then(function(Te){return Te.tsubMiddleware(r.middlewareSettings.routingRules)})]:[3,2];case 1:return P=ge.sent(),[3,3];case 2:P=void 0,ge.label=3;case 3:return O=P,bfr(r)||d.length>0?[4,Ix(()=>import("./index-CX2UoIQ2.js"),__vite__mapDeps([0,1])).then(function(Te){return Te.ajsDestinations(n,r,o.integrations,s,O,d)})]:[3,5];case 4:return $=ge.sent(),[3,6];case 5:$=[],ge.label=6;case 6:return D=$,r.legacyVideoPluginsEnabled?[4,Ix(()=>import("./index-J_vsoDtB.js"),[]).then(function(Te){return Te.loadLegacyVideoPlugins(o)})]:[3,8];case 7:ge.sent(),ge.label=8;case 8:return!((b=s.plan)===null||b===void 0)&&b.track?[4,Ix(()=>import("./index-BUG6bmcj.js"),__vite__mapDeps([2,1])).then(function(Te){var xe;return Te.schemaFilter((xe=s.plan)===null||xe===void 0?void 0:xe.track,r)})]:[3,10];case 9:return B=ge.sent(),[3,11];case 10:B=void 0,ge.label=11;case 11:return U=B,G=Xdr(r,s),[4,hfr(r,o.integrations,G,s,O,x).catch(function(){return[]})];case 12:return K=ge.sent(),X=Vg(Vg([afr],D,!0),K,!0),U&&X.push(U),ae=((y=s.integrations)===null||y===void 0?void 0:y.All)===!1&&!s.integrations["Segment.io"]||s.integrations&&s.integrations["Segment.io"]===!1,ae?[3,14]:(he=(oe=X).push,[4,Hdr(o,G["Segment.io"],r.integrations)]);case 13:he.apply(oe,[ge.sent()]),ge.label=14;case 14:return[4,o.register.apply(o,Vg(Vg([],X,!1),k,!1))];case 15:return ye=ge.sent(),[4,hdr(o,h)];case 16:return ge.sent(),Object.entries((E=r.enabledMiddleware)!==null&&E!==void 0?E:{}).some(function(Te){var xe=Te[1];return xe})?[4,Ix(()=>import("./index-YCKUjqLe.js"),[]).then(function(Te){var xe=Te.remoteMiddlewares;return wo(we,void 0,void 0,function(){var Ue,Z;return yo(this,function(Re){switch(Re.label){case 0:return[4,xe(ye,r,s.obfuscate)];case 1:return Ue=Re.sent(),Z=Ue.map(function(_e){return o.addSourceMiddleware(_e)}),[2,Promise.all(Z)]}})})})]:[3,18];case 17:ge.sent(),ge.label=18;case 18:return[4,fdr(o,h)];case 19:return ge.sent(),[2,ye]}})})}function kfr(n,r,o){var s,c,d,h,b,y,E,k,x;return r===void 0&&(r={}),wo(this,void 0,void 0,function(){var O,P,D,$,U,B,G,K,X,ae,oe;return yo(this,function(he){switch(he.label){case 0:return r.disable===!0?[2,[new y$t,$4.system()]]:(r.globalAnalyticsKey&&udr(r.globalAnalyticsKey),n.cdnURL&&Ydr(n.cdnURL),r.initialPageview&&o.add(new tWe("page",[])),O=Efr(),P=(s=n.cdnURL)!==null&&s!==void 0?s:dZe(),(c=n.cdnSettings)!==null&&c!==void 0?($=c,[3,3]):[3,1]);case 1:return[4,mfr(n.writeKey,P)];case 2:$=he.sent(),he.label=3;case 3:return D=$,r.updateCDNSettings&&(D=r.updateCDNSettings(D)),typeof r.disable!="function"?[3,5]:[4,r.disable(D)];case 4:if(U=he.sent(),U)return[2,[new y$t,$4.system()]];he.label=5;case 5:return B=(h=(d=D.integrations["Segment.io"])===null||d===void 0?void 0:d.retryQueue)!==null&&h!==void 0?h:!0,r=zr({retryQueue:B},r),G=new Han(zr(zr({},n),{cdnSettings:D,cdnURL:P}),r),gfr(G),K=(b=n.plugins)!==null&&b!==void 0?b:[],X=(y=n.classicIntegrations)!==null&&y!==void 0?y:[],ae=(E=r.integrations)===null||E===void 0?void 0:E["Segment.io"],_an.initRemoteMetrics(zr(zr({},D.metrics),{host:(k=ae?.apiHost)!==null&&k!==void 0?k:(x=D.metrics)===null||x===void 0?void 0:x.host,protocol:ae?.protocol})),[4,_fr(n.writeKey,D,G,r,K,X,o)];case 6:return oe=he.sent(),G.initialized=!0,G.emit("initialize",n,r),[4,yfr(G,O,o)];case 7:return he.sent(),[2,[G,oe]]}})})}var Tfr=function(n){iy(r,n);function r(){var o=this,s=wan(),c=s.promise,d=s.resolve;return o=n.call(this,function(h){return c.then(function(b){var y=b[0],E=b[1];return kfr(y,E,h)})})||this,o._resolveLoadStart=function(h,b){return d([h,b])},o}return r.prototype.load=function(o,s){return s===void 0&&(s={}),this._resolveLoadStart(o,s),this},r.load=function(o,s){return s===void 0&&(s={}),new r().load(o,s)},r.standalone=function(o,s){return r.load({writeKey:o},s).then(function(c){return c[0]})},r}(vdr);const jb=[];for(let n=0;n<256;++n)jb.push((n+256).toString(16).slice(1));function xfr(n,r=0){return(jb[n[r+0]]+jb[n[r+1]]+jb[n[r+2]]+jb[n[r+3]]+"-"+jb[n[r+4]]+jb[n[r+5]]+"-"+jb[n[r+6]]+jb[n[r+7]]+"-"+jb[n[r+8]]+jb[n[r+9]]+"-"+jb[n[r+10]]+jb[n[r+11]]+jb[n[r+12]]+jb[n[r+13]]+jb[n[r+14]]+jb[n[r+15]]).toLowerCase()}let L$e;const Cfr=new Uint8Array(16);function Afr(){if(!L$e){if(typeof crypto>"u"||!crypto.getRandomValues)throw new Error("crypto.getRandomValues() not supported. See https://github.com/uuidjs/uuid#getrandomvalues-not-supported");L$e=crypto.getRandomValues.bind(crypto)}return L$e(Cfr)}const Ifr=typeof crypto<"u"&&crypto.randomUUID&&crypto.randomUUID.bind(crypto),k$t={randomUUID:Ifr};function n2(n,r,o){if(k$t.randomUUID&&!n)return k$t.randomUUID();n=n||{};const s=n.random??n.rng?.()??Afr();if(s.length<16)throw new Error("Random bytes length must be >= 16");return s[6]=s[6]&15|64,s[8]=s[8]&63|128,xfr(s)}const NW=typeof __SENTRY_DEBUG__>"u"||__SENTRY_DEBUG__,RM=globalThis,die="9.44.1";function Wan(){return fZe(RM),RM}function fZe(n){const r=n.__SENTRY__=n.__SENTRY__||{};return r.version=r.version||die,r[die]=r[die]||{}}function hZe(n,r,o=RM){const s=o.__SENTRY__=o.__SENTRY__||{},c=s[die]=s[die]||{};return c[n]||(c[n]=r())}const Ofr="Sentry Logger ",T$t={};function Nfr(n){if(!("console"in RM))return n();const r=RM.console,o={},s=Object.keys(T$t);s.forEach(c=>{const d=T$t[c];o[c]=r[c],r[c]=d});try{return n()}finally{s.forEach(c=>{r[c]=o[c]})}}function Rfr(){gZe().enabled=!0}function Pfr(){gZe().enabled=!1}function Kan(){return gZe().enabled}function Lfr(...n){pZe("log",...n)}function Dfr(...n){pZe("warn",...n)}function Mfr(...n){pZe("error",...n)}function pZe(n,...r){NW&&Kan()&&Nfr(()=>{RM.console[n](`${Ofr}[${n}]:`,...r)})}function gZe(){return NW?hZe("loggerSettings",()=>({enabled:!1})):{enabled:!1}}const mve={enable:Rfr,disable:Pfr,isEnabled:Kan,log:Lfr,warn:Dfr,error:Mfr},jfr=Object.prototype.toString;function $fr(n,r){return jfr.call(n)===`[object ${r}]`}function Ffr(n){return $fr(n,"Object")}function Bfr(n){return!!(n?.then&&typeof n.then=="function")}function Ufr(n,r=0){return typeof n!="string"||r===0||n.length<=r?n:`${n.slice(0,r)}...`}function zfr(n,r,o){try{Object.defineProperty(n,r,{value:o,writable:!0,configurable:!0})}catch{NW&&mve.log(`Failed to add non-enumerable property "${r}" to object`,n)}}function Hfr(){const n=RM;return n.crypto||n.msCrypto}function fie(n=Hfr()){let r=()=>Math.random()*16;try{if(n?.randomUUID)return n.randomUUID().replace(/-/g,"");n?.getRandomValues&&(r=()=>{const o=new Uint8Array(1);return n.getRandomValues(o),o[0]})}catch{}return("10000000100040008000"+1e11).replace(/[018]/g,o=>(o^(r()&15)>>o/4).toString(16))}const Yan=1e3;function Xan(){return Date.now()/Yan}function Gfr(){const{performance:n}=RM;if(!n?.now||!n.timeOrigin)return Xan;const r=n.timeOrigin;return()=>(r+n.now())/Yan}let x$t;function qfr(){return(x$t??(x$t=Gfr()))()}function Vfr(n,r={}){if(r.user&&(!n.ipAddress&&r.user.ip_address&&(n.ipAddress=r.user.ip_address),!n.did&&!r.did&&(n.did=r.user.id||r.user.email||r.user.username)),n.timestamp=r.timestamp||qfr(),r.abnormal_mechanism&&(n.abnormal_mechanism=r.abnormal_mechanism),r.ignoreDuration&&(n.ignoreDuration=r.ignoreDuration),r.sid&&(n.sid=r.sid.length===32?r.sid:fie()),r.init!==void 0&&(n.init=r.init),!n.did&&r.did&&(n.did=`${r.did}`),typeof r.started=="number"&&(n.started=r.started),n.ignoreDuration)n.duration=void 0;else if(typeof r.duration=="number")n.duration=r.duration;else{const o=n.timestamp-n.started;n.duration=o>=0?o:0}r.release&&(n.release=r.release),r.environment&&(n.environment=r.environment),!n.ipAddress&&r.ipAddress&&(n.ipAddress=r.ipAddress),!n.userAgent&&r.userAgent&&(n.userAgent=r.userAgent),typeof r.errors=="number"&&(n.errors=r.errors),r.status&&(n.status=r.status)}function Jan(n,r,o=2){if(!r||typeof r!="object"||o<=0)return r;if(n&&Object.keys(r).length===0)return n;const s={...n};for(const c in r)Object.prototype.hasOwnProperty.call(r,c)&&(s[c]=Jan(s[c],r[c],o-1));return s}function C$t(){return fie()}const nWe="_sentrySpan";function A$t(n,r){r?zfr(n,nWe,r):delete n[nWe]}function I$t(n){return n[nWe]}const Wfr=100;class PM{constructor(){this._notifyingListeners=!1,this._scopeListeners=[],this._eventProcessors=[],this._breadcrumbs=[],this._attachments=[],this._user={},this._tags={},this._extra={},this._contexts={},this._sdkProcessingMetadata={},this._propagationContext={traceId:C$t(),sampleRand:Math.random()}}clone(){const r=new PM;return r._breadcrumbs=[...this._breadcrumbs],r._tags={...this._tags},r._extra={...this._extra},r._contexts={...this._contexts},this._contexts.flags&&(r._contexts.flags={values:[...this._contexts.flags.values]}),r._user=this._user,r._level=this._level,r._session=this._session,r._transactionName=this._transactionName,r._fingerprint=this._fingerprint,r._eventProcessors=[...this._eventProcessors],r._attachments=[...this._attachments],r._sdkProcessingMetadata={...this._sdkProcessingMetadata},r._propagationContext={...this._propagationContext},r._client=this._client,r._lastEventId=this._lastEventId,A$t(r,I$t(this)),r}setClient(r){this._client=r}setLastEventId(r){this._lastEventId=r}getClient(){return this._client}lastEventId(){return this._lastEventId}addScopeListener(r){this._scopeListeners.push(r)}addEventProcessor(r){return this._eventProcessors.push(r),this}setUser(r){return this._user=r||{email:void 0,id:void 0,ip_address:void 0,username:void 0},this._session&&Vfr(this._session,{user:r}),this._notifyScopeListeners(),this}getUser(){return this._user}setTags(r){return this._tags={...this._tags,...r},this._notifyScopeListeners(),this}setTag(r,o){return this._tags={...this._tags,[r]:o},this._notifyScopeListeners(),this}setExtras(r){return this._extra={...this._extra,...r},this._notifyScopeListeners(),this}setExtra(r,o){return this._extra={...this._extra,[r]:o},this._notifyScopeListeners(),this}setFingerprint(r){return this._fingerprint=r,this._notifyScopeListeners(),this}setLevel(r){return this._level=r,this._notifyScopeListeners(),this}setTransactionName(r){return this._transactionName=r,this._notifyScopeListeners(),this}setContext(r,o){return o===null?delete this._contexts[r]:this._contexts[r]=o,this._notifyScopeListeners(),this}setSession(r){return r?this._session=r:delete this._session,this._notifyScopeListeners(),this}getSession(){return this._session}update(r){if(!r)return this;const o=typeof r=="function"?r(this):r,s=o instanceof PM?o.getScopeData():Ffr(o)?r:void 0,{tags:c,extra:d,user:h,contexts:b,level:y,fingerprint:E=[],propagationContext:k}=s||{};return this._tags={...this._tags,...c},this._extra={...this._extra,...d},this._contexts={...this._contexts,...b},h&&Object.keys(h).length&&(this._user=h),y&&(this._level=y),E.length&&(this._fingerprint=E),k&&(this._propagationContext=k),this}clear(){return this._breadcrumbs=[],this._tags={},this._extra={},this._user={},this._contexts={},this._level=void 0,this._transactionName=void 0,this._fingerprint=void 0,this._session=void 0,A$t(this,void 0),this._attachments=[],this.setPropagationContext({traceId:C$t(),sampleRand:Math.random()}),this._notifyScopeListeners(),this}addBreadcrumb(r,o){const s=typeof o=="number"?o:Wfr;if(s<=0)return this;const c={timestamp:Xan(),...r,message:r.message?Ufr(r.message,2048):r.message};return this._breadcrumbs.push(c),this._breadcrumbs.length>s&&(this._breadcrumbs=this._breadcrumbs.slice(-s),this._client?.recordDroppedEvent("buffer_overflow","log_item")),this._notifyScopeListeners(),this}getLastBreadcrumb(){return this._breadcrumbs[this._breadcrumbs.length-1]}clearBreadcrumbs(){return this._breadcrumbs=[],this._notifyScopeListeners(),this}addAttachment(r){return this._attachments.push(r),this}clearAttachments(){return this._attachments=[],this}getScopeData(){return{breadcrumbs:this._breadcrumbs,attachments:this._attachments,contexts:this._contexts,tags:this._tags,extra:this._extra,user:this._user,level:this._level,fingerprint:this._fingerprint||[],eventProcessors:this._eventProcessors,propagationContext:this._propagationContext,sdkProcessingMetadata:this._sdkProcessingMetadata,transactionName:this._transactionName,span:I$t(this)}}setSDKProcessingMetadata(r){return this._sdkProcessingMetadata=Jan(this._sdkProcessingMetadata,r,2),this}setPropagationContext(r){return this._propagationContext=r,this}getPropagationContext(){return this._propagationContext}captureException(r,o){const s=o?.event_id||fie();if(!this._client)return NW&&mve.warn("No client configured on scope - will not capture exception!"),s;const c=new Error("Sentry syntheticException");return this._client.captureException(r,{originalException:r,syntheticException:c,...o,event_id:s},this),s}captureMessage(r,o,s){const c=s?.event_id||fie();if(!this._client)return NW&&mve.warn("No client configured on scope - will not capture message!"),c;const d=new Error(r);return this._client.captureMessage(r,o,{originalException:r,syntheticException:d,...s,event_id:c},this),c}captureEvent(r,o){const s=o?.event_id||fie();return this._client?(this._client.captureEvent(r,{...o,event_id:s},this),s):(NW&&mve.warn("No client configured on scope - will not capture event!"),s)}_notifyScopeListeners(){this._notifyingListeners||(this._notifyingListeners=!0,this._scopeListeners.forEach(r=>{r(this)}),this._notifyingListeners=!1)}}function Kfr(){return hZe("defaultCurrentScope",()=>new PM)}function Yfr(){return hZe("defaultIsolationScope",()=>new PM)}class Xfr{constructor(r,o){let s;r?s=r:s=new PM;let c;o?c=o:c=new PM,this._stack=[{scope:s}],this._isolationScope=c}withScope(r){const o=this._pushScope();let s;try{s=r(o)}catch(c){throw this._popScope(),c}return Bfr(s)?s.then(c=>(this._popScope(),c),c=>{throw this._popScope(),c}):(this._popScope(),s)}getClient(){return this.getStackTop().client}getScope(){return this.getStackTop().scope}getIsolationScope(){return this._isolationScope}getStackTop(){return this._stack[this._stack.length-1]}_pushScope(){const r=this.getScope().clone();return this._stack.push({client:this.getClient(),scope:r}),r}_popScope(){return this._stack.length<=1?!1:!!this._stack.pop()}}function tK(){const n=Wan(),r=fZe(n);return r.stack=r.stack||new Xfr(Kfr(),Yfr())}function Jfr(n){return tK().withScope(n)}function Zfr(n,r){const o=tK();return o.withScope(()=>(o.getStackTop().scope=n,r(n)))}function O$t(n){return tK().withScope(()=>n(tK().getIsolationScope()))}function Qfr(){return{withIsolationScope:O$t,withScope:Jfr,withSetScope:Zfr,withSetIsolationScope:(n,r)=>O$t(r),getCurrentScope:()=>tK().getScope(),getIsolationScope:()=>tK().getIsolationScope()}}function ehr(n){const r=fZe(n);return r.acs?r.acs:Qfr()}function thr(){const n=Wan();return ehr(n).getCurrentScope()}function nhr(n){if(n)return rhr(n)?{captureContext:n}:ohr(n)?{captureContext:n}:n}function rhr(n){return n instanceof PM||typeof n=="function"}const ihr=["user","level","extra","contexts","tags","fingerprint","propagationContext"];function ohr(n){return Object.keys(n).some(r=>ihr.includes(r))}function ahr(n,r){return thr().captureException(n,nhr(r))}var o2=(n=>(n[n.Collect=0]="Collect",n[n.Action=1]="Action",n[n.Link=2]="Link",n[n.SetSlots=3]="SetSlots",n[n.Call=4]="Call",n[n.Noop=5]="Noop",n))(o2||{});function D$e(n){return n.if!==void 0&&n.then!==void 0}const shr=xs({if:Tr(),utter:Tr()}),chr=xs({noop:V1(),collect:Tr(),action:Tr(),link:Tr(),set_slots:mp(hU(Tr(),D4([Tr(),V1(),G4(),$Je()]))),call:Tr(),id:Tr(),utter:Tr(),rejections:mp(shr),ask_before_filling:V1(),force_slot_filling:V1(),silence_timeout:G4(),reset_after_flow_ends:V1(),description:Tr()}).partial(),bve=chr.extend({next:KJe(()=>D4([mp(bve),Tr(),mp(D4([xs({if:Tr(),then:D4([mp(bve),Tr()])}),xs({else:D4([mp(bve),Tr()])})]))])).optional()}).refine(n=>n.noop&&!n.collect&&!n.action&&!n.link&&!n.set_slots&&!n.call||n.collect&&!n.action&&!n.link&&!n.set_slots&&!n.call&&!n.noop||n.action&&!n.collect&&!n.link&&!n.set_slots&&!n.call&&!n.noop||n.set_slots&&!n.collect&&!n.action&&!n.link&&!n.call&&!n.noop||n.link&&!n.collect&&!n.action&&!n.set_slots&&!n.call&&!n.noop||n.call&&!n.collect&&!n.action&&!n.link&&!n.set_slots&&!n.noop).transform(n=>{const r={...n};return n.collect?{...r,askBeforeFilling:n.ask_before_filling,forceSlotFilling:n.force_slot_filling,resetAfterFlowEnds:n.reset_after_flow_ends,silenceTimeout:n.silence_timeout}:r}),uhr=xs({intent:Tr().transform(n=>({name:n,confidence_threshold:0})).or(xs({name:Tr(),confidence_threshold:G4().optional().default(0)}))}),lhr=hU(Tr(),xs({name:Tr()})),dhr=xs({name:Tr().optional(),description:Tr(),translation:lhr.optional(),persisted_slots:mp(Tr()).nullable().optional().transform(n=>n===null?void 0:n),steps:mp(bve),nlu_trigger:mp(uhr).nullable().optional().transform(n=>n===null?void 0:n),if:Tr().or(V1()).optional(),always_include_in_prompt:V1().optional(),file_path:Tr().optional()}),fhr=xs({flows:hU(Tr(),dhr)}),hhr=async({projectUrl:n,requestBody:r,token:o})=>{const s=await e6(),c={"Content-Type":"application/json","X-User-Id":String(s)};o&&(c.Authorization=`Bearer ${o}`);const d={method:"POST",headers:c,body:JSON.stringify({...r})},h=await fetch(`${n}${Vb.COPILOT}`,d);if(!h.ok){const b=await h.json().catch(()=>({})),y={};for(const[E,k]of Object.entries(d.headers))y[E]=E.toLowerCase()==="authorization"?"*****":k;throw new Error(`HTTP error! status: ${h.status},
      request: ${JSON.stringify(d.body,null,2)},
      request heareds: ${JSON.stringify(y,null,2)},
      response: ${JSON.stringify(b,null,2)}`)}if(!h.body)throw new Error("Response body is null");return h.body},phr=async({projectUrl:n})=>V_(`${n}${Vb.COPILOT_MODE}`,{method:"GET"},void 0,ban),ghr=async({projectUrl:n,mode:r})=>V_(`${n}${Vb.COPILOT_MODE}`,{method:"POST"},{mode:r},ban),mhr=async({projectUrl:n})=>V_(`${n}${Vb.BOT_BUILDER}`,{method:"GET"},void 0,ilr),bhr=async({projectUrl:n})=>V_(`${n}${Vb.BOT_DATA}`,{method:"GET"},void 0,fhr),vhr=async({projectUrl:n})=>V_(`${n}${Vb.BOT_FILES}`,{method:"GET"},void 0,alr),whr=async({projectUrl:n})=>{const r=await e6(),o={"Content-Type":"application/json","X-User-Id":String(r)};return V_(`${n}${Vb.TRAIN}`,{method:"POST",headers:o},void 0,XJe)},yhr=async({projectUrl:n,commitSha:r})=>V_(`${n}${Vb.COMMITS}/${r}/diff-with-contents`,{method:"GET"},void 0,slr),Ehr=async({projectUrl:n,files:r,token:o})=>{const s=await e6(),c={"Content-Type":"application/json","X-User-Id":String(s)};return o&&(c.Authorization=`Bearer ${o}`),V_(`${n}${Vb.BOT_FILES}`,{method:"POST",headers:c},r,XJe)},Shr=async({projectUrl:n})=>V_(`${n}${Vb.ASSISTANT_INFO}`,{method:"GET"},void 0,olr),_hr=async({projectUrl:n,projectId:r,token:o})=>{const s=new URL(`${n}${Vb.DOWNLOAD_PROJECT}`);s.searchParams.set("project_name",r);const c=await e6();return j5r(s.toString(),{method:"GET",headers:{Authorization:`Bearer ${o}`,"X-User-Id":String(c)}})},khr=async({projectUrl:n,commit_sha:r})=>{const o=await e6(),s={"Content-Type":"application/json","X-User-Id":String(o)};return V_(`${n}${Vb.COMMITS}/${r}/rollback`,{method:"POST",headers:s},void 0,XJe)},Thr=async({projectUrl:n,token:r})=>{const o=await e6(),s={"X-User-Id":String(o)};r&&(s.Authorization=`Bearer ${r}`);const c=new URL(`${n}${Vb.COPILOT_HISTORY}`);return(await V_(c.toString(),{method:"GET",headers:s})).messages},xhr=["abap"],Chr=["ampl"],Ahr=["g4"],Ihr=["apl","dyalog"],Ohr=["asp","asax","ascx","ashx","asmx","aspx","axd"],Nhr=["dats","hats","sats"],Rhr=["as"],Phr=["adb","ada","ads"],Lhr=["agda"],Dhr=["als"],Mhr=["apacheconf"],jhr=["cls"],$hr=["applescript","scpt"],Fhr=["arc"],Bhr=["ino"],Uhr=["asciidoc","adoc","asc"],zhr=["aj"],Hhr=["asm","a51","nasm"],Ghr=["aug"],qhr=["ahk","ahkl"],Vhr=["au3"],Whr=["awk","auk","gawk","mawk","nawk"],Khr=["bat","cmd"],Yhr=["befunge"],Xhr=["y"],Jhr=["bb"],Zhr=["bb","decls"],Qhr=["bmx"],epr=["bsv"],tpr=["boo"],npr=["b","bf"],rpr=["brs"],ipr=["bro"],opr=["c","cats","h","idc","w"],apr=["clp"],spr=["cmake","cmake.in"],cpr=["cob","cbl","ccp","cobol","cpy"],upr=["css"],lpr=["mss"],dpr=["ceylon"],fpr=["chpl"],hpr=["ck"],ppr=["cirru"],gpr=["clw"],mpr=["icl","dcl"],bpr=["clj","boot","cl2","cljc","cljs","cljs.hl","cljscm","cljx","hic"],vpr=["coffee","_coffee","cjsx","cson","iced"],wpr=["cfm","cfml"],ypr=["cl"],Epr=["coq","v"],Spr=["creole"],_pr=["cr"],kpr=["feature"],Tpr=["cu","cuh"],xpr=["cy"],Cpr=["pyx","pxd","pxi"],Apr=["d","di"],Ipr=["dm"],Opr=["d"],Npr=["dart"],Rpr=["diff","patch"],Ppr=["dockerfile"],Lpr=["djs"],Dpr=["dylan","dyl","intr","lid"],Mpr=["E"],jpr=["ecl","eclxml"],$pr=["sch","brd"],Fpr=["e"],Bpr=["ex","exs"],Upr=["elm"],zpr=["em","emberscript"],Hpr=["erl","es","escript","hrl"],Gpr=["fx","flux"],qpr=["f90","f","f03","f08","f77","f95","for","fpp"],Vpr=["factor"],Wpr=["fy","fancypack"],Kpr=["fan"],Ypr=["fs"],Xpr=["for"],Jpr=["fth","4th","f","for","forth","fr","frt","fs"],Zpr=["fr"],Qpr=["gms"],egr=["g","gap","gd","gi","tst"],tgr=["s"],ngr=["gd"],rgr=["glsl","fp","frag","frg","fs","fshader","geo","geom","glslv","gshader","shader","vert","vrx","vshader"],igr=["kid"],ogr=["glf"],agr=["gp","gnu","gnuplot","plot","plt"],sgr=["go"],cgr=["golo"],ugr=["gs","gst","gsx","vark"],lgr=["grace"],dgr=["gradle"],fgr=["man","1","2","3","4","5","6","7"],hgr=["groovy","grt","gtpl","gvy"],pgr=["html","htm","html.hl","st","xht","xhtml"],ggr=["http"],mgr=["hh","php"],bgr=["haml","haml.deface"],vgr=["handlebars","hbs"],wgr=["hb"],ygr=["hs","hsc"],Egr=["hx","hxsl"],Sgr=["hy"],_gr=["pro","dlm"],kgr=["ini","cfg","prefs","pro","properties"],Tgr=["idr","lidr"],xgr=["io"],Cgr=["ik"],Agr=["thy"],Igr=["ijs"],Ogr=["flex","jflex"],Ngr=["json","lock"],Rgr=["json5"],Pgr=["jsonld"],Lgr=["jq"],Dgr=["jade"],Mgr=["j"],jgr=["java"],$gr=["js","_js","bones","es6","frag","gs","jake","jsb","jsfl","jsm","jss","jsx","njs","pac","sjs","ssjs","sublime-build","sublime-commands","sublime-completions","sublime-keymap","sublime-macro","sublime-menu","sublime-mousemap","sublime-project","sublime-settings","sublime-theme","sublime-workspace","sublime_metrics","sublime_session","xsjs","xsjslib"],Fgr=["jl"],Bgr=["krl"],Ugr=["sch"],zgr=["kit"],Hgr=["kt","ktm","kts"],Ggr=["lfe"],qgr=["ll"],Vgr=["lol"],Wgr=["lsl"],Kgr=["lvproj"],Ygr=["lasso","las","lasso8","lasso9","ldml"],Xgr=["latte"],Jgr=["lean","hlean"],Zgr=["less"],Qgr=["ly","ily"],emr=["b","m"],tmr=["liquid"],nmr=["ls","_ls"],rmr=["xm","x","xi"],imr=["lgt","logtalk"],omr=["lookml"],amr=["ls"],smr=["lua","fcgi","nse","pd_lua","rbxs","wlua"],cmr=["mumps","m"],umr=["mtml"],lmr=["muf","m"],dmr=["mak","d","mk"],fmr=["mako","mao"],hmr=["md","markdown","mkd","mkdn","mkdown","ron"],pmr=["mask"],gmr=["mathematica","cdf","m","ma","nb","nbp","wl","wlt"],mmr=["matlab","m"],bmr=["maxpat","maxhelp","maxproj","mxt","pat"],vmr=["mediawiki"],wmr=["m","moo"],ymr=["minid"],Emr=["druby","duby","mir","mirah"],Smr=["mo"],_mr=["monkey"],kmr=["moo"],Tmr=["moon"],xmr=["myt"],Cmr=["nl"],Amr=["nsi","nsh"],Imr=["n"],Omr=["axs","axi"],Nmr=["nlogo"],Rmr=["nl","lisp","lsp"],Pmr=["nginxconf"],Lmr=["nim","nimrod"],Dmr=["ninja"],Mmr=["nit"],jmr=["nix"],$mr=["nu"],Fmr=["numpy","numpyw","numsc"],Bmr=["ml","eliom","eliomi","ml4","mli","mll","mly"],Umr=["objdump"],zmr=["omgrofl"],Hmr=["opa"],Gmr=["opal"],qmr=["cl","opencl"],Vmr=["scad"],Wmr=["org"],Kmr=["ox","oxh","oxo"],Ymr=["oxygene"],Xmr=["oz"],Jmr=["pwn"],Zmr=["php","aw","ctp","fcgi","php3","php4","php5","phpt"],Qmr=["pls","pkb","pks","plb","plsql","sql"],ebr=["sql"],tbr=["pan"],nbr=["psc"],rbr=["parrot"],ibr=["pas","dfm","dpr","lpr","pp"],obr=["pl","cgi","fcgi","perl","ph","plx","pm","pod","psgi","t"],abr=["6pl","6pm","nqp","p6","p6l","p6m","pl","pl6","pm","pm6","t"],sbr=["pig"],cbr=["pike","pmod"],ubr=["pod"],lbr=["pogo"],dbr=["ps","eps"],fbr=["ps1","psd1","psm1"],hbr=["pde"],pbr=["pl","ecl","pro","prolog"],gbr=["pp"],mbr=["pb","pbi"],bbr=["purs"],vbr=["py","cgi","fcgi","gyp","lmi","pyde","pyp","pyt","pyw","tac","wsgi","xpy"],wbr=["qml"],ybr=["pro","pri"],Ebr=["r","rd","rsx"],Sbr=["raml"],_br=["rdoc"],kbr=["rbbas","rbfrm","rbmnu","rbres","rbtbar","rbuistate"],Tbr=["rhtml"],xbr=["rmd"],Cbr=["rkt","rktd","rktl","scrbl"],Abr=["reb","r","r2","r3","rebol"],Ibr=["red","reds"],Obr=["cw"],Nbr=["rs","rsh"],Rbr=["robot"],Pbr=["rg"],Lbr=["rb","builder","fcgi","gemspec","god","irbrc","jbuilder","mspec","pluginspec","podspec","rabl","rake","rbuild","rbw","rbx","ru","ruby","thor","watchr"],Dbr=["rs"],Mbr=["sas"],jbr=["scss"],$br=["sparql","rq"],Fbr=["sqf","hqf"],Bbr=["sql","cql","ddl","prc","tab","udf","viw"],Ubr=["sql","db2"],zbr=["ston"],Hbr=["svg"],Gbr=["sage","sagews"],qbr=["sls"],Vbr=["sass"],Wbr=["scala","sbt","sc"],Kbr=["scaml"],Ybr=["scm","sld","sls","sps","ss"],Xbr=["sci","sce","tst"],Jbr=["self"],Zbr=["sh","bash","bats","cgi","command","fcgi","ksh","tmux","tool","zsh"],Qbr=["sh-session"],e1r=["shen"],t1r=["sl"],n1r=["slim"],r1r=["st","cs"],i1r=["tpl"],o1r=["sp","sma"],a1r=["nut"],s1r=["do","ado","doh","ihlp","mata","matah","sthlp"],c1r=["styl"],u1r=["scd","sc"],l1r=["swift"],d1r=["sv","svh","vh"],f1r=["toml"],h1r=["txl"],p1r=["tcl","adp","tm"],g1r=["tcsh","csh"],m1r=["tex","aux","bbx","bib","cbx","cls","dtx","ins","lbx","ltx","mkii","mkiv","mkvi","sty","toc"],b1r=["tea"],v1r=["txt","fr"],w1r=["textile"],y1r=["thrift"],E1r=["t","tu"],S1r=["ttl"],_1r=["twig"],k1r=["ts"],T1r=["uc"],x1r=["vcl"],C1r=["vhdl","vhd","vhf","vhi","vho","vhs","vht","vhw"],A1r=["vala","vapi"],I1r=["v","veo"],O1r=["vim"],N1r=["volt"],R1r=["webidl"],P1r=["xc"],L1r=["xml","ant","axml","ccxml","clixml","cproject","csproj","ct","dita","ditamap","ditaval","dll.config","filters","fsproj","fxml","glade","grxml","ivy","jelly","kml","launch","mm","mxml","nproj","nuspec","odd","osm","plist","pluginspec","ps1xml","psc1","pt","rdf","rss","scxml","srdf","storyboard","stTheme","sublime-snippet","targets","tmCommand","tml","tmLanguage","tmPreferences","tmSnippet","tmTheme","ts","ui","urdf","vbproj","vcxproj","vxml","wsdl","wsf","wxi","wxl","wxs","x3d","xacro","xaml","xib","xlf","xliff","xmi","xml.dist","xsd","xul","zcml"],D1r=["xpl","xproc"],M1r=["xquery","xq","xql","xqm","xqy"],j1r=["xs"],$1r=["xslt","xsl"],F1r=["xojo_code","xojo_menu","xojo_report","xojo_script","xojo_toolbar","xojo_window"],B1r=["xtend"],U1r=["yml","reek","rviz","yaml"],z1r=["zep"],H1r=["zimpl","zmpl","zpl"],G1r=["desktop","desktop.in"],q1r=["ec","eh"],V1r=["edn"],W1r=["fish"],K1r=["mu"],Y1r=["nc"],X1r=["ooc"],J1r=["rst","rest"],Z1r=["wisp"],Q1r=["prg"],e0r={abap:xhr,"ags script":["asc","ash"],ampl:Chr,antlr:Ahr,apl:Ihr,asp:Ohr,ats:Nhr,actionscript:Rhr,ada:Phr,agda:Lhr,alloy:Dhr,apacheconf:Mhr,apex:jhr,applescript:$hr,arc:Fhr,arduino:Bhr,asciidoc:Uhr,aspectj:zhr,assembly:Hhr,augeas:Ghr,autohotkey:qhr,autoit:Vhr,awk:Whr,batchfile:Khr,befunge:Yhr,bison:Xhr,bitbake:Jhr,blitzbasic:Zhr,blitzmax:Qhr,bluespec:epr,boo:tpr,brainfuck:npr,brightscript:rpr,bro:ipr,c:opr,"c#":["cs","cshtml","csx"],"c++":["cpp","c++","cc","cp","cxx","h","h++","hh","hpp","hxx","inl","ipp","tcc","tpp"],"c-objdump":["c-objdump"],"c2hs haskell":["chs"],clips:apr,cmake:spr,cobol:cpr,css:upr,"cap'n proto":["capnp"],cartocss:lpr,ceylon:dpr,chapel:fpr,chuck:hpr,cirru:ppr,clarion:gpr,clean:mpr,clojure:bpr,coffeescript:vpr,coldfusion:wpr,"coldfusion cfc":["cfc"],"common lisp":["lisp","asd","cl","lsp","ny","podsl"],"component pascal":["cp","cps"],cool:ypr,coq:Epr,"cpp-objdump":["cppobjdump","c++-objdump","c++objdump","cpp-objdump","cxx-objdump"],creole:Spr,crystal:_pr,cucumber:kpr,cuda:Tpr,cycript:xpr,cython:Cpr,d:Apr,"d-objdump":["d-objdump"],"digital command language":["com"],dm:Ipr,dtrace:Opr,"darcs patch":["darcspatch","dpatch"],dart:Npr,diff:Rpr,dockerfile:Ppr,dogescript:Lpr,dylan:Dpr,e:Mpr,ecl:jpr,eagle:$pr,"ecere projects":["epj"],eiffel:Fpr,elixir:Bpr,elm:Upr,"emacs lisp":["el","emacs","emacs.desktop"],emberscript:zpr,erlang:Hpr,"f#":["fs","fsi","fsx"],flux:Gpr,fortran:qpr,factor:Vpr,fancy:Wpr,fantom:Kpr,filterscript:Ypr,formatted:Xpr,forth:Jpr,frege:Zpr,"g-code":["g","gco","gcode"],gams:Qpr,gap:egr,gas:tgr,gdscript:ngr,glsl:rgr,"game maker language":["gml"],genshi:igr,"gentoo ebuild":["ebuild"],"gentoo eclass":["eclass"],"gettext catalog":["po","pot"],glyph:ogr,gnuplot:agr,go:sgr,golo:cgr,gosu:ugr,grace:lgr,gradle:dgr,"grammatical framework":["gf"],"graph modeling language":["gml"],"graphviz (dot)":["dot","gv"],groff:fgr,groovy:hgr,"groovy server pages":["gsp"],html:pgr,"html+django":["mustache","jinja"],"html+erb":["erb","erb.deface"],"html+php":["phtml"],http:ggr,hack:mgr,haml:bgr,handlebars:vgr,harbour:wgr,haskell:ygr,haxe:Egr,hy:Sgr,idl:_gr,"igor pro":["ipf"],ini:kgr,"irc log":["irclog","weechatlog"],idris:Tgr,"inform 7":["ni","i7x"],"inno setup":["iss"],io:xgr,ioke:Cgr,isabelle:Agr,j:Igr,jflex:Ogr,json:Ngr,json5:Rgr,jsonld:Pgr,jsoniq:Lgr,jade:Dgr,jasmin:Mgr,java:jgr,"java server pages":["jsp"],javascript:$gr,julia:Fgr,krl:Bgr,kicad:Ugr,kit:zgr,kotlin:Hgr,lfe:Ggr,llvm:qgr,lolcode:Vgr,lsl:Wgr,labview:Kgr,lasso:Ygr,latte:Xgr,lean:Jgr,less:Zgr,lilypond:Qgr,limbo:emr,"linker script":["ld","lds"],liquid:tmr,"literate agda":["lagda"],"literate coffeescript":["litcoffee"],"literate haskell":["lhs"],livescript:nmr,logos:rmr,logtalk:imr,lookml:omr,loomscript:amr,lua:smr,m:cmr,mtml:umr,muf:lmr,makefile:dmr,mako:fmr,markdown:hmr,mask:pmr,mathematica:gmr,matlab:mmr,max:bmr,mediawiki:vmr,mercury:wmr,minid:ymr,mirah:Emr,modelica:Smr,"module management system":["mms","mmk"],monkey:_mr,moocode:kmr,moonscript:Tmr,myghty:xmr,nl:Cmr,nsis:Amr,nemerle:Imr,netlinx:Omr,"netlinx+erb":["axs.erb","axi.erb"],netlogo:Nmr,newlisp:Rmr,nginx:Pmr,nimrod:Lmr,ninja:Dmr,nit:Mmr,nix:jmr,nu:$mr,numpy:Fmr,ocaml:Bmr,objdump:Umr,"objective-c":["m","h"],"objective-c++":["mm"],"objective-j":["j","sj"],omgrofl:zmr,opa:Hmr,opal:Gmr,opencl:qmr,"openedge abl":["p","cls"],openscad:Vmr,org:Wmr,ox:Kmr,oxygene:Ymr,oz:Xmr,pawn:Jmr,php:Zmr,plsql:Qmr,plpgsql:ebr,pan:tbr,papyrus:nbr,parrot:rbr,"parrot assembly":["pasm"],"parrot internal representation":["pir"],pascal:ibr,perl:obr,perl6:abr,piglatin:sbr,pike:cbr,pod:ubr,pogoscript:lbr,postscript:dbr,powershell:fbr,processing:hbr,prolog:pbr,"propeller spin":["spin"],"protocol buffer":["proto"],"public key":["asc","pub"],puppet:gbr,"pure data":["pd"],purebasic:mbr,purescript:bbr,python:vbr,"python traceback":["pytb"],qml:wbr,qmake:ybr,r:Ebr,raml:Sbr,rdoc:_br,realbasic:kbr,rhtml:Tbr,rmarkdown:xbr,racket:Cbr,"ragel in ruby host":["rl"],"raw token data":["raw"],rebol:Abr,red:Ibr,redcode:Obr,renderscript:Nbr,robotframework:Rbr,rouge:Pbr,ruby:Lbr,rust:Dbr,sas:Mbr,scss:jbr,sparql:$br,sqf:Fbr,sql:Bbr,sqlpl:Ubr,ston:zbr,svg:Hbr,sage:Gbr,saltstack:qbr,sass:Vbr,scala:Wbr,scaml:Kbr,scheme:Ybr,scilab:Xbr,self:Jbr,shell:Zbr,shellsession:Qbr,shen:e1r,slash:t1r,slim:n1r,smalltalk:r1r,smarty:i1r,sourcepawn:o1r,squirrel:a1r,"standard ml":["ML","fun","sig","sml"],stata:s1r,stylus:c1r,supercollider:u1r,swift:l1r,systemverilog:d1r,toml:f1r,txl:h1r,tcl:p1r,tcsh:g1r,tex:m1r,tea:b1r,text:v1r,textile:w1r,thrift:y1r,turing:E1r,turtle:S1r,twig:_1r,typescript:k1r,"unified parallel c":["upc"],unrealscript:T1r,vcl:x1r,vhdl:C1r,vala:A1r,verilog:I1r,viml:O1r,"visual basic":["vb","bas","cls","frm","frx","vba","vbhtml","vbs"],volt:N1r,"web ontology language":["owl"],webidl:R1r,xc:P1r,xml:L1r,xproc:D1r,xquery:M1r,xs:j1r,xslt:$1r,xojo:F1r,xtend:B1r,yaml:U1r,zephir:z1r,zimpl:H1r,desktop:G1r,ec:q1r,edn:V1r,fish:W1r,mupad:K1r,nesc:Y1r,ooc:X1r,restructuredtext:J1r,wisp:Z1r,xbase:Q1r},t0r=["abap"],n0r=["ags script","asciidoc","public key"],r0r=["ags script"],i0r=["ampl"],o0r=["antlr"],a0r=["apl"],s0r=["apl"],c0r=["asp"],u0r=["asp"],l0r=["asp"],d0r=["asp"],f0r=["asp"],h0r=["asp"],p0r=["asp"],g0r=["ats"],m0r=["ats"],b0r=["ats"],v0r=["actionscript"],w0r=["ada"],y0r=["ada"],E0r=["ada"],S0r=["agda"],_0r=["alloy"],k0r=["apacheconf"],T0r=["apex","openedge abl","tex","visual basic"],x0r=["applescript"],C0r=["applescript"],A0r=["arc"],I0r=["arduino"],O0r=["asciidoc"],N0r=["asciidoc"],R0r=["aspectj"],P0r=["assembly"],L0r=["assembly"],D0r=["assembly"],M0r=["augeas"],j0r=["autohotkey"],$0r=["autohotkey"],F0r=["autoit"],B0r=["awk"],U0r=["awk"],z0r=["awk"],H0r=["awk"],G0r=["awk"],q0r=["batchfile"],V0r=["batchfile"],W0r=["befunge"],K0r=["bison"],Y0r=["bitbake","blitzbasic"],X0r=["blitzbasic"],J0r=["blitzmax"],Z0r=["bluespec"],Q0r=["boo"],evr=["brainfuck","limbo"],tvr=["brainfuck"],nvr=["brightscript"],rvr=["bro"],ivr=["c"],ovr=["c"],avr=["c","c++","objective-c"],svr=["c"],cvr=["c"],uvr=["c#","smalltalk"],lvr=["c#"],dvr=["c#"],fvr=["c++"],hvr=["c++"],pvr=["c++","component pascal"],gvr=["c++"],mvr=["c++","hack"],bvr=["c++"],vvr=["c++"],wvr=["c++"],yvr=["c++"],Evr=["c++"],Svr=["c++"],_vr=["c2hs haskell"],kvr=["clips"],Tvr=["cmake"],xvr=["cobol"],Cvr=["cobol"],Avr=["cobol"],Ivr=["cobol"],Ovr=["cobol"],Nvr=["css"],Rvr=["cap'n proto"],Pvr=["cartocss"],Lvr=["ceylon"],Dvr=["chapel"],Mvr=["chuck"],jvr=["cirru"],$vr=["clarion"],Fvr=["clean"],Bvr=["clean"],Uvr=["clojure"],zvr=["clojure"],Hvr=["clojure"],Gvr=["clojure"],qvr=["clojure"],Vvr=["clojure"],Wvr=["clojure"],Kvr=["clojure"],Yvr=["coffeescript"],Xvr=["coffeescript"],Jvr=["coffeescript"],Zvr=["coffeescript"],Qvr=["coffeescript"],ewr=["coldfusion"],twr=["coldfusion"],nwr=["coldfusion cfc"],rwr=["common lisp","newlisp"],iwr=["common lisp"],owr=["common lisp","cool","opencl"],awr=["common lisp","newlisp"],swr=["common lisp"],cwr=["common lisp"],uwr=["component pascal"],lwr=["coq"],dwr=["coq","verilog"],fwr=["cpp-objdump"],hwr=["creole"],pwr=["crystal"],gwr=["cucumber"],mwr=["cuda"],bwr=["cuda"],vwr=["cycript"],wwr=["cython"],ywr=["cython"],Ewr=["cython"],Swr=["d","dtrace","makefile"],_wr=["d"],kwr=["digital command language"],Twr=["dm"],xwr=["darcs patch"],Cwr=["darcs patch"],Awr=["dart"],Iwr=["diff"],Owr=["diff"],Nwr=["dockerfile"],Rwr=["dogescript"],Pwr=["dylan"],Lwr=["dylan"],Dwr=["dylan"],Mwr=["dylan"],jwr=["e"],$wr=["ecl","prolog"],Fwr=["ecl"],Bwr=["eagle","kicad"],Uwr=["eagle"],zwr=["ecere projects"],Hwr=["eiffel"],Gwr=["elixir"],qwr=["elixir"],Vwr=["elm"],Wwr=["emacs lisp"],Kwr=["emacs lisp"],Ywr=["emberscript"],Xwr=["emberscript"],Jwr=["erlang"],Zwr=["erlang"],Qwr=["erlang"],eyr=["erlang"],tyr=["f#","filterscript","forth","glsl"],nyr=["f#"],ryr=["f#"],iyr=["flux"],oyr=["flux"],ayr=["fortran"],syr=["fortran","forth"],cyr=["fortran"],uyr=["fortran"],lyr=["fortran"],dyr=["fortran"],fyr=["fortran"],hyr=["factor"],pyr=["fancy"],gyr=["fancy"],myr=["fantom"],byr=["forth"],vyr=["forth"],wyr=["forth","frege","text"],yyr=["forth"],Eyr=["g-code","gap"],Syr=["g-code"],_yr=["g-code"],kyr=["gams"],Tyr=["gap"],xyr=["gap","gdscript"],Cyr=["gap"],Ayr=["gap","scilab"],Iyr=["gas"],Oyr=["glsl"],Nyr=["glsl"],Ryr=["glsl","javascript"],Pyr=["glsl"],Lyr=["glsl"],Dyr=["glsl"],Myr=["glsl"],jyr=["glsl"],$yr=["glsl"],Fyr=["glsl"],Byr=["glsl"],Uyr=["glsl"],zyr=["glsl"],Hyr=["game maker language","graph modeling language"],Gyr=["genshi"],qyr=["gentoo ebuild"],Vyr=["gentoo eclass"],Wyr=["gettext catalog"],Kyr=["gettext catalog"],Yyr=["glyph"],Xyr=["gnuplot"],Jyr=["gnuplot"],Zyr=["gnuplot"],Qyr=["gnuplot"],eEr=["gnuplot"],tEr=["go"],nEr=["golo"],rEr=["gosu","javascript"],iEr=["gosu"],oEr=["gosu"],aEr=["gosu"],sEr=["grace"],cEr=["gradle"],uEr=["grammatical framework"],lEr=["graphviz (dot)"],dEr=["graphviz (dot)"],fEr=["groff"],hEr=["groovy"],pEr=["groovy"],gEr=["groovy"],mEr=["groovy"],bEr=["groovy server pages"],vEr=["html"],wEr=["html"],yEr=["html","smalltalk"],EEr=["html"],SEr=["html"],_Er=["html+django"],kEr=["html+django"],TEr=["html+erb"],xEr=["html+php"],CEr=["http"],AEr=["hack","php"],IEr=["haml"],OEr=["handlebars"],NEr=["handlebars"],REr=["harbour"],PEr=["haskell"],LEr=["haskell"],DEr=["haxe"],MEr=["haxe"],jEr=["hy"],$Er=["idl","ini","prolog","qmake"],FEr=["idl"],BEr=["igor pro"],UEr=["ini"],zEr=["ini"],HEr=["ini"],GEr=["ini"],qEr=["irc log"],VEr=["irc log"],WEr=["idris"],KEr=["idris"],YEr=["inform 7"],XEr=["inform 7"],JEr=["inno setup"],ZEr=["io"],QEr=["ioke"],e2r=["isabelle"],t2r=["j"],n2r=["jflex"],r2r=["jflex"],i2r=["json"],o2r=["json"],a2r=["json5"],s2r=["jsonld"],c2r=["jsoniq"],u2r=["jade"],l2r=["jasmin","objective-j"],d2r=["java"],f2r=["java server pages"],h2r=["javascript"],p2r=["javascript"],g2r=["javascript"],m2r=["javascript"],b2r=["javascript"],v2r=["javascript"],w2r=["javascript"],y2r=["javascript"],E2r=["javascript"],S2r=["javascript"],_2r=["javascript"],k2r=["javascript"],T2r=["javascript"],x2r=["javascript"],C2r=["javascript"],A2r=["javascript"],I2r=["javascript"],O2r=["javascript"],N2r=["julia"],R2r=["krl"],P2r=["kit"],L2r=["kotlin"],D2r=["kotlin"],M2r=["kotlin"],j2r=["lfe"],$2r=["llvm"],F2r=["lolcode"],B2r=["lsl"],U2r=["labview"],z2r=["lasso"],H2r=["lasso"],G2r=["lasso"],q2r=["lasso"],V2r=["lasso"],W2r=["latte"],K2r=["lean"],Y2r=["lean"],X2r=["less"],J2r=["lilypond"],Z2r=["lilypond"],Q2r=["limbo","m","muf","mathematica","matlab","mercury","objective-c"],eSr=["linker script"],tSr=["linker script"],nSr=["liquid"],rSr=["literate agda"],iSr=["literate coffeescript"],oSr=["literate haskell"],aSr=["livescript","loomscript"],sSr=["livescript"],cSr=["logos"],uSr=["logos"],lSr=["logos"],dSr=["logtalk"],fSr=["logtalk"],hSr=["lookml"],pSr=["lua"],gSr=["lua","php","perl","python","ruby","shell"],mSr=["lua"],bSr=["lua"],vSr=["lua"],wSr=["lua"],ySr=["m"],ESr=["mtml"],SSr=["muf"],_Sr=["makefile"],kSr=["makefile"],TSr=["mako"],xSr=["mako"],CSr=["markdown"],ASr=["markdown"],ISr=["markdown"],OSr=["markdown"],NSr=["markdown"],RSr=["markdown"],PSr=["mask"],LSr=["mathematica"],DSr=["mathematica"],MSr=["mathematica"],jSr=["mathematica"],$Sr=["mathematica"],FSr=["mathematica"],BSr=["mathematica"],USr=["matlab"],zSr=["max"],HSr=["max"],GSr=["max"],qSr=["max"],VSr=["max"],WSr=["mediawiki"],KSr=["mercury","moocode"],YSr=["minid"],XSr=["mirah"],JSr=["mirah"],ZSr=["mirah"],QSr=["mirah"],e_r=["modelica"],t_r=["module management system"],n_r=["module management system"],r_r=["monkey"],i_r=["moonscript"],o_r=["myghty"],a_r=["nl","newlisp"],s_r=["nsis"],c_r=["nsis"],u_r=["nemerle"],l_r=["netlinx"],d_r=["netlinx"],f_r=["netlogo"],h_r=["nginx"],p_r=["nimrod"],g_r=["nimrod"],m_r=["ninja"],b_r=["nit"],v_r=["nix"],w_r=["nu"],y_r=["numpy"],E_r=["numpy"],S_r=["numpy"],__r=["ocaml"],k_r=["ocaml"],T_r=["ocaml"],x_r=["ocaml"],C_r=["ocaml"],A_r=["ocaml"],I_r=["ocaml"],O_r=["objdump"],N_r=["objective-c++","xml"],R_r=["objective-j"],P_r=["omgrofl"],L_r=["opa"],D_r=["opal"],M_r=["opencl"],j_r=["openedge abl"],$_r=["openscad"],F_r=["org"],B_r=["ox"],U_r=["ox"],z_r=["ox"],H_r=["oxygene"],G_r=["oz"],q_r=["pawn"],V_r=["php"],W_r=["php"],K_r=["php"],Y_r=["php"],X_r=["php"],J_r=["php"],Z_r=["plsql"],Q_r=["plsql"],ekr=["plsql"],tkr=["plsql"],nkr=["plsql"],rkr=["plsql","plpgsql","sql","sqlpl"],ikr=["pan"],okr=["papyrus"],akr=["parrot"],skr=["parrot assembly"],ckr=["parrot internal representation"],ukr=["pascal"],lkr=["pascal"],dkr=["pascal"],fkr=["pascal"],hkr=["pascal","puppet"],pkr=["perl","perl6","prolog"],gkr=["perl","python","shell"],mkr=["perl"],bkr=["perl"],vkr=["perl"],wkr=["perl","perl6"],ykr=["perl","pod"],Ekr=["perl"],Skr=["perl","perl6","turing"],_kr=["perl6"],kkr=["perl6"],Tkr=["perl6"],xkr=["perl6"],Ckr=["perl6"],Akr=["perl6"],Ikr=["piglatin"],Okr=["pike"],Nkr=["pike"],Rkr=["pogoscript"],Pkr=["postscript"],Lkr=["postscript"],Dkr=["powershell"],Mkr=["powershell"],jkr=["powershell"],$kr=["processing"],Fkr=["prolog"],Bkr=["propeller spin"],Ukr=["protocol buffer"],zkr=["public key"],Hkr=["pure data"],Gkr=["purebasic"],qkr=["purebasic"],Vkr=["purescript"],Wkr=["python"],Kkr=["python"],Ykr=["python"],Xkr=["python"],Jkr=["python"],Zkr=["python"],Qkr=["python"],eTr=["python"],tTr=["python"],nTr=["python"],rTr=["python traceback"],iTr=["qml"],oTr=["qmake"],aTr=["r","rebol"],sTr=["r"],cTr=["r"],uTr=["raml"],lTr=["rdoc"],dTr=["realbasic"],fTr=["realbasic"],hTr=["realbasic"],pTr=["realbasic"],gTr=["realbasic"],mTr=["realbasic"],bTr=["rhtml"],vTr=["rmarkdown"],wTr=["racket"],yTr=["racket"],ETr=["racket"],STr=["racket"],_Tr=["ragel in ruby host"],kTr=["raw token data"],TTr=["rebol"],xTr=["rebol"],CTr=["rebol"],ATr=["rebol"],ITr=["red"],OTr=["red"],NTr=["redcode"],RTr=["renderscript","rust"],PTr=["renderscript"],LTr=["robotframework"],DTr=["rouge"],MTr=["ruby"],jTr=["ruby"],$Tr=["ruby"],FTr=["ruby"],BTr=["ruby"],UTr=["ruby"],zTr=["ruby"],HTr=["ruby","xml"],GTr=["ruby"],qTr=["ruby"],VTr=["ruby"],WTr=["ruby"],KTr=["ruby"],YTr=["ruby"],XTr=["ruby"],JTr=["ruby"],ZTr=["ruby"],QTr=["ruby"],exr=["sas"],txr=["scss"],nxr=["sparql"],rxr=["sparql"],ixr=["sqf"],oxr=["sqf"],axr=["sql"],sxr=["sql"],cxr=["sql"],uxr=["sql"],lxr=["sql"],dxr=["sql"],fxr=["sqlpl"],hxr=["ston"],pxr=["svg"],gxr=["sage"],mxr=["sage"],bxr=["saltstack","scheme"],vxr=["sass"],wxr=["scala"],yxr=["scala"],Exr=["scala","supercollider"],Sxr=["scaml"],_xr=["scheme"],kxr=["scheme"],Txr=["scheme"],xxr=["scheme"],Cxr=["scilab"],Axr=["scilab"],Ixr=["self"],Oxr=["shell"],Nxr=["shell"],Rxr=["shell"],Pxr=["shell"],Lxr=["shell"],Dxr=["shell"],Mxr=["shell"],jxr=["shell"],$xr=["shen"],Fxr=["slash"],Bxr=["slim"],Uxr=["smarty"],zxr=["sourcepawn"],Hxr=["sourcepawn"],Gxr=["squirrel"],qxr=["standard ml"],Vxr=["standard ml"],Wxr=["standard ml"],Kxr=["standard ml"],Yxr=["stata"],Xxr=["stata"],Jxr=["stata"],Zxr=["stata"],Qxr=["stata"],eCr=["stata"],tCr=["stylus"],nCr=["supercollider"],rCr=["swift"],iCr=["systemverilog"],oCr=["systemverilog"],aCr=["systemverilog"],sCr=["toml"],cCr=["txl"],uCr=["tcl"],lCr=["tcl"],dCr=["tcl"],fCr=["tcsh"],hCr=["tcsh"],pCr=["tex"],gCr=["tex"],mCr=["tex"],bCr=["tex"],vCr=["tex"],wCr=["tex"],yCr=["tex"],ECr=["tex"],SCr=["tex"],_Cr=["tex"],kCr=["tex"],TCr=["tex"],xCr=["tex"],CCr=["tex"],ACr=["tea"],ICr=["text"],OCr=["textile"],NCr=["thrift"],RCr=["turing"],PCr=["turtle"],LCr=["twig"],DCr=["typescript","xml"],MCr=["unified parallel c"],jCr=["unrealscript"],$Cr=["vcl"],FCr=["vhdl"],BCr=["vhdl"],UCr=["vhdl"],zCr=["vhdl"],HCr=["vhdl"],GCr=["vhdl"],qCr=["vhdl"],VCr=["vhdl"],WCr=["vala"],KCr=["vala"],YCr=["verilog"],XCr=["viml"],JCr=["visual basic"],ZCr=["visual basic"],QCr=["visual basic"],e3r=["visual basic"],t3r=["visual basic"],n3r=["visual basic"],r3r=["visual basic"],i3r=["volt"],o3r=["web ontology language"],a3r=["webidl"],s3r=["xc"],c3r=["xml"],u3r=["xml"],l3r=["xml"],d3r=["xml"],f3r=["xml"],h3r=["xml"],p3r=["xml"],g3r=["xml"],m3r=["xml"],b3r=["xml"],v3r=["xml"],w3r=["xml"],y3r=["xml"],E3r=["xml"],S3r=["xml"],_3r=["xml"],k3r=["xml"],T3r=["xml"],x3r=["xml"],C3r=["xml"],A3r=["xml"],I3r=["xml"],O3r=["xml"],N3r=["xml"],R3r=["xml"],P3r=["xml"],L3r=["xml"],D3r=["xml"],M3r=["xml"],j3r=["xml"],$3r=["xml"],F3r=["xml"],B3r=["xml"],U3r=["xml"],z3r=["xml"],H3r=["xml"],G3r=["xml"],q3r=["xml"],V3r=["xml"],W3r=["xml"],K3r=["xml"],Y3r=["xml"],X3r=["xml"],J3r=["xml"],Z3r=["xml"],Q3r=["xml"],eAr=["xml"],tAr=["xml"],nAr=["xml"],rAr=["xml"],iAr=["xml"],oAr=["xml"],aAr=["xml"],sAr=["xml"],cAr=["xml"],uAr=["xml"],lAr=["xml"],dAr=["xml"],fAr=["xml"],hAr=["xml"],pAr=["xml"],gAr=["xml"],mAr=["xproc"],bAr=["xproc"],vAr=["xquery"],wAr=["xquery"],yAr=["xquery"],EAr=["xquery"],SAr=["xquery"],_Ar=["xs"],kAr=["xslt"],TAr=["xslt"],xAr=["xojo"],CAr=["xojo"],AAr=["xojo"],IAr=["xojo"],OAr=["xojo"],NAr=["xojo"],RAr=["xtend"],PAr=["yaml"],LAr=["yaml"],DAr=["yaml"],MAr=["yaml"],jAr=["zephir"],$Ar=["zimpl"],FAr=["zimpl"],BAr=["zimpl"],UAr=["desktop"],zAr=["ec"],HAr=["ec"],GAr=["edn"],qAr=["fish"],VAr=["mupad"],WAr=["nesc"],KAr=["ooc"],YAr=["restructuredtext"],XAr=["restructuredtext"],JAr=["wisp"],ZAr=["xbase"],QAr={1:["groff"],2:["groff"],3:["groff"],4:["groff"],5:["groff"],6:["groff"],7:["groff"],abap:t0r,asc:n0r,ash:r0r,ampl:i0r,g4:o0r,apl:a0r,dyalog:s0r,asp:c0r,asax:u0r,ascx:l0r,ashx:d0r,asmx:f0r,aspx:h0r,axd:p0r,dats:g0r,hats:m0r,sats:b0r,as:v0r,adb:w0r,ada:y0r,ads:E0r,agda:S0r,als:_0r,apacheconf:k0r,cls:T0r,applescript:x0r,scpt:C0r,arc:A0r,ino:I0r,asciidoc:O0r,adoc:N0r,aj:R0r,asm:P0r,a51:L0r,nasm:D0r,aug:M0r,ahk:j0r,ahkl:$0r,au3:F0r,awk:B0r,auk:U0r,gawk:z0r,mawk:H0r,nawk:G0r,bat:q0r,cmd:V0r,befunge:W0r,y:K0r,bb:Y0r,decls:X0r,bmx:J0r,bsv:Z0r,boo:Q0r,b:evr,bf:tvr,brs:nvr,bro:rvr,c:ivr,cats:ovr,h:avr,idc:svr,w:cvr,cs:uvr,cshtml:lvr,csx:dvr,cpp:fvr,"c++":["c++"],cc:hvr,cp:pvr,cxx:gvr,"h++":["c++"],hh:mvr,hpp:bvr,hxx:vvr,inl:wvr,ipp:yvr,tcc:Evr,tpp:Svr,"c-objdump":["c-objdump"],chs:_vr,clp:kvr,cmake:Tvr,"cmake.in":["cmake"],cob:xvr,cbl:Cvr,ccp:Avr,cobol:Ivr,cpy:Ovr,css:Nvr,capnp:Rvr,mss:Pvr,ceylon:Lvr,chpl:Dvr,ck:Mvr,cirru:jvr,clw:$vr,icl:Fvr,dcl:Bvr,clj:Uvr,boot:zvr,cl2:Hvr,cljc:Gvr,cljs:qvr,"cljs.hl":["clojure"],cljscm:Vvr,cljx:Wvr,hic:Kvr,coffee:Yvr,_coffee:Xvr,cjsx:Jvr,cson:Zvr,iced:Qvr,cfm:ewr,cfml:twr,cfc:nwr,lisp:rwr,asd:iwr,cl:owr,lsp:awr,ny:swr,podsl:cwr,cps:uwr,coq:lwr,v:dwr,cppobjdump:fwr,"c++-objdump":["cpp-objdump"],"c++objdump":["cpp-objdump"],"cpp-objdump":["cpp-objdump"],"cxx-objdump":["cpp-objdump"],creole:hwr,cr:pwr,feature:gwr,cu:mwr,cuh:bwr,cy:vwr,pyx:wwr,pxd:ywr,pxi:Ewr,d:Swr,di:_wr,"d-objdump":["d-objdump"],com:kwr,dm:Twr,darcspatch:xwr,dpatch:Cwr,dart:Awr,diff:Iwr,patch:Owr,dockerfile:Nwr,djs:Rwr,dylan:Pwr,dyl:Lwr,intr:Dwr,lid:Mwr,E:jwr,ecl:$wr,eclxml:Fwr,sch:Bwr,brd:Uwr,epj:zwr,e:Hwr,ex:Gwr,exs:qwr,elm:Vwr,el:Wwr,emacs:Kwr,"emacs.desktop":["emacs lisp"],em:Ywr,emberscript:Xwr,erl:Jwr,es:Zwr,escript:Qwr,hrl:eyr,fs:tyr,fsi:nyr,fsx:ryr,fx:iyr,flux:oyr,f90:ayr,f:syr,f03:cyr,f08:uyr,f77:lyr,f95:dyr,for:["fortran","formatted","forth"],fpp:fyr,factor:hyr,fy:pyr,fancypack:gyr,fan:myr,fth:byr,"4th":["forth"],forth:vyr,fr:wyr,frt:yyr,g:Eyr,gco:Syr,gcode:_yr,gms:kyr,gap:Tyr,gd:xyr,gi:Cyr,tst:Ayr,s:Iyr,glsl:Oyr,fp:Nyr,frag:Ryr,frg:Pyr,fshader:Lyr,geo:Dyr,geom:Myr,glslv:jyr,gshader:$yr,shader:Fyr,vert:Byr,vrx:Uyr,vshader:zyr,gml:Hyr,kid:Gyr,ebuild:qyr,eclass:Vyr,po:Wyr,pot:Kyr,glf:Yyr,gp:Xyr,gnu:Jyr,gnuplot:Zyr,plot:Qyr,plt:eEr,go:tEr,golo:nEr,gs:rEr,gst:iEr,gsx:oEr,vark:aEr,grace:sEr,gradle:cEr,gf:uEr,dot:lEr,gv:dEr,man:fEr,groovy:hEr,grt:pEr,gtpl:gEr,gvy:mEr,gsp:bEr,html:vEr,htm:wEr,"html.hl":["html"],st:yEr,xht:EEr,xhtml:SEr,mustache:_Er,jinja:kEr,erb:TEr,"erb.deface":["html+erb"],phtml:xEr,http:CEr,php:AEr,haml:IEr,"haml.deface":["haml"],handlebars:OEr,hbs:NEr,hb:REr,hs:PEr,hsc:LEr,hx:DEr,hxsl:MEr,hy:jEr,pro:$Er,dlm:FEr,ipf:BEr,ini:UEr,cfg:zEr,prefs:HEr,properties:GEr,irclog:qEr,weechatlog:VEr,idr:WEr,lidr:KEr,ni:YEr,i7x:XEr,iss:JEr,io:ZEr,ik:QEr,thy:e2r,ijs:t2r,flex:n2r,jflex:r2r,json:i2r,lock:o2r,json5:a2r,jsonld:s2r,jq:c2r,jade:u2r,j:l2r,java:d2r,jsp:f2r,js:h2r,_js:p2r,bones:g2r,es6:m2r,jake:b2r,jsb:v2r,jsfl:w2r,jsm:y2r,jss:E2r,jsx:S2r,njs:_2r,pac:k2r,sjs:T2r,ssjs:x2r,"sublime-build":["javascript"],"sublime-commands":["javascript"],"sublime-completions":["javascript"],"sublime-keymap":["javascript"],"sublime-macro":["javascript"],"sublime-menu":["javascript"],"sublime-mousemap":["javascript"],"sublime-project":["javascript"],"sublime-settings":["javascript"],"sublime-theme":["javascript"],"sublime-workspace":["javascript"],sublime_metrics:C2r,sublime_session:A2r,xsjs:I2r,xsjslib:O2r,jl:N2r,krl:R2r,kit:P2r,kt:L2r,ktm:D2r,kts:M2r,lfe:j2r,ll:$2r,lol:F2r,lsl:B2r,lvproj:U2r,lasso:z2r,las:H2r,lasso8:G2r,lasso9:q2r,ldml:V2r,latte:W2r,lean:K2r,hlean:Y2r,less:X2r,ly:J2r,ily:Z2r,m:Q2r,ld:eSr,lds:tSr,liquid:nSr,lagda:rSr,litcoffee:iSr,lhs:oSr,ls:aSr,_ls:sSr,xm:cSr,x:uSr,xi:lSr,lgt:dSr,logtalk:fSr,lookml:hSr,lua:pSr,fcgi:gSr,nse:mSr,pd_lua:bSr,rbxs:vSr,wlua:wSr,mumps:ySr,mtml:ESr,muf:SSr,mak:_Sr,mk:kSr,mako:TSr,mao:xSr,md:CSr,markdown:ASr,mkd:ISr,mkdn:OSr,mkdown:NSr,ron:RSr,mask:PSr,mathematica:LSr,cdf:DSr,ma:MSr,nb:jSr,nbp:$Sr,wl:FSr,wlt:BSr,matlab:USr,maxpat:zSr,maxhelp:HSr,maxproj:GSr,mxt:qSr,pat:VSr,mediawiki:WSr,moo:KSr,minid:YSr,druby:XSr,duby:JSr,mir:ZSr,mirah:QSr,mo:e_r,mms:t_r,mmk:n_r,monkey:r_r,moon:i_r,myt:o_r,nl:a_r,nsi:s_r,nsh:c_r,n:u_r,axs:l_r,axi:d_r,"axs.erb":["netlinx+erb"],"axi.erb":["netlinx+erb"],nlogo:f_r,nginxconf:h_r,nim:p_r,nimrod:g_r,ninja:m_r,nit:b_r,nix:v_r,nu:w_r,numpy:y_r,numpyw:E_r,numsc:S_r,ml:__r,eliom:k_r,eliomi:T_r,ml4:x_r,mli:C_r,mll:A_r,mly:I_r,objdump:O_r,mm:N_r,sj:R_r,omgrofl:P_r,opa:L_r,opal:D_r,opencl:M_r,p:j_r,scad:$_r,org:F_r,ox:B_r,oxh:U_r,oxo:z_r,oxygene:H_r,oz:G_r,pwn:q_r,aw:V_r,ctp:W_r,php3:K_r,php4:Y_r,php5:X_r,phpt:J_r,pls:Z_r,pkb:Q_r,pks:ekr,plb:tkr,plsql:nkr,sql:rkr,pan:ikr,psc:okr,parrot:akr,pasm:skr,pir:ckr,pas:ukr,dfm:lkr,dpr:dkr,lpr:fkr,pp:hkr,pl:pkr,cgi:gkr,perl:mkr,ph:bkr,plx:vkr,pm:wkr,pod:ykr,psgi:Ekr,t:Skr,"6pl":["perl6"],"6pm":["perl6"],nqp:_kr,p6:kkr,p6l:Tkr,p6m:xkr,pl6:Ckr,pm6:Akr,pig:Ikr,pike:Okr,pmod:Nkr,pogo:Rkr,ps:Pkr,eps:Lkr,ps1:Dkr,psd1:Mkr,psm1:jkr,pde:$kr,prolog:Fkr,spin:Bkr,proto:Ukr,pub:zkr,pd:Hkr,pb:Gkr,pbi:qkr,purs:Vkr,py:Wkr,gyp:Kkr,lmi:Ykr,pyde:Xkr,pyp:Jkr,pyt:Zkr,pyw:Qkr,tac:eTr,wsgi:tTr,xpy:nTr,pytb:rTr,qml:iTr,pri:oTr,r:aTr,rd:sTr,rsx:cTr,raml:uTr,rdoc:lTr,rbbas:dTr,rbfrm:fTr,rbmnu:hTr,rbres:pTr,rbtbar:gTr,rbuistate:mTr,rhtml:bTr,rmd:vTr,rkt:wTr,rktd:yTr,rktl:ETr,scrbl:STr,rl:_Tr,raw:kTr,reb:TTr,r2:xTr,r3:CTr,rebol:ATr,red:ITr,reds:OTr,cw:NTr,rs:RTr,rsh:PTr,robot:LTr,rg:DTr,rb:MTr,builder:jTr,gemspec:$Tr,god:FTr,irbrc:BTr,jbuilder:UTr,mspec:zTr,pluginspec:HTr,podspec:GTr,rabl:qTr,rake:VTr,rbuild:WTr,rbw:KTr,rbx:YTr,ru:XTr,ruby:JTr,thor:ZTr,watchr:QTr,sas:exr,scss:txr,sparql:nxr,rq:rxr,sqf:ixr,hqf:oxr,cql:axr,ddl:sxr,prc:cxr,tab:uxr,udf:lxr,viw:dxr,db2:fxr,ston:hxr,svg:pxr,sage:gxr,sagews:mxr,sls:bxr,sass:vxr,scala:wxr,sbt:yxr,sc:Exr,scaml:Sxr,scm:_xr,sld:kxr,sps:Txr,ss:xxr,sci:Cxr,sce:Axr,self:Ixr,sh:Oxr,bash:Nxr,bats:Rxr,command:Pxr,ksh:Lxr,tmux:Dxr,tool:Mxr,zsh:jxr,"sh-session":["shellsession"],shen:$xr,sl:Fxr,slim:Bxr,tpl:Uxr,sp:zxr,sma:Hxr,nut:Gxr,ML:qxr,fun:Vxr,sig:Wxr,sml:Kxr,do:["stata"],ado:Yxr,doh:Xxr,ihlp:Jxr,mata:Zxr,matah:Qxr,sthlp:eCr,styl:tCr,scd:nCr,swift:rCr,sv:iCr,svh:oCr,vh:aCr,toml:sCr,txl:cCr,tcl:uCr,adp:lCr,tm:dCr,tcsh:fCr,csh:hCr,tex:pCr,aux:gCr,bbx:mCr,bib:bCr,cbx:vCr,dtx:wCr,ins:yCr,lbx:ECr,ltx:SCr,mkii:_Cr,mkiv:kCr,mkvi:TCr,sty:xCr,toc:CCr,tea:ACr,txt:ICr,textile:OCr,thrift:NCr,tu:RCr,ttl:PCr,twig:LCr,ts:DCr,upc:MCr,uc:jCr,vcl:$Cr,vhdl:FCr,vhd:BCr,vhf:UCr,vhi:zCr,vho:HCr,vhs:GCr,vht:qCr,vhw:VCr,vala:WCr,vapi:KCr,veo:YCr,vim:XCr,vb:JCr,bas:ZCr,frm:QCr,frx:e3r,vba:t3r,vbhtml:n3r,vbs:r3r,volt:i3r,owl:o3r,webidl:a3r,xc:s3r,xml:c3r,ant:u3r,axml:l3r,ccxml:d3r,clixml:f3r,cproject:h3r,csproj:p3r,ct:g3r,dita:m3r,ditamap:b3r,ditaval:v3r,"dll.config":["xml"],filters:w3r,fsproj:y3r,fxml:E3r,glade:S3r,grxml:_3r,ivy:k3r,jelly:T3r,kml:x3r,launch:C3r,mxml:A3r,nproj:I3r,nuspec:O3r,odd:N3r,osm:R3r,plist:P3r,ps1xml:L3r,psc1:D3r,pt:M3r,rdf:j3r,rss:$3r,scxml:F3r,srdf:B3r,storyboard:U3r,stTheme:z3r,"sublime-snippet":["xml"],targets:H3r,tmCommand:G3r,tml:q3r,tmLanguage:V3r,tmPreferences:W3r,tmSnippet:K3r,tmTheme:Y3r,ui:X3r,urdf:J3r,vbproj:Z3r,vcxproj:Q3r,vxml:eAr,wsdl:tAr,wsf:nAr,wxi:rAr,wxl:iAr,wxs:oAr,x3d:aAr,xacro:sAr,xaml:cAr,xib:uAr,xlf:lAr,xliff:dAr,xmi:fAr,"xml.dist":["xml"],xsd:hAr,xul:pAr,zcml:gAr,xpl:mAr,xproc:bAr,xquery:vAr,xq:wAr,xql:yAr,xqm:EAr,xqy:SAr,xs:_Ar,xslt:kAr,xsl:TAr,xojo_code:xAr,xojo_menu:CAr,xojo_report:AAr,xojo_script:IAr,xojo_toolbar:OAr,xojo_window:NAr,xtend:RAr,yml:PAr,reek:LAr,rviz:DAr,yaml:MAr,zep:jAr,zimpl:$Ar,zmpl:FAr,zpl:BAr,desktop:UAr,"desktop.in":["desktop"],ec:zAr,eh:HAr,edn:GAr,fish:qAr,mu:VAr,nc:WAr,ooc:KAr,rst:YAr,rest:XAr,wisp:JAr,prg:ZAr};/*!
 * lang-map <https://github.com/jonschlinkert/lang-map>
 *
 * Copyright (c) 2014-2015, Jon Schlinkert.
 * Licensed under the MIT license.
 */var M$e,N$t;function eIr(){if(N$t)return M$e;N$t=1;function n(){var o={};return o.extensions||(o.extensions=e0r),o.languages||(o.languages=QAr),o}n.extensions=function(s){s=r(s);var c=n().languages,d=n().extensions;return d[s]||d[c[s]]||[s]},n.languages=function(s){s=r(s);var c=n().languages,d=n().extensions;return c[s]||c[d[s]]||[s]};function r(o){return o.charAt(0)==="."&&(o=o.slice(1)),o.toLowerCase()}return M$e=n,M$e}var tIr=eIr();const nIr=Y4(tIr);function sEe(n){return Array.isArray(n)}function cEe(n){return n!==null&&typeof n=="object"&&(n.constructor===void 0||n.constructor.name==="Object")}function rIr(n,r){return JSON.stringify(n)===JSON.stringify(r)}function Zan(n){return n.slice(0,n.length-1)}function iIr(n){return n[n.length-1]}function Qan(n){return typeof n=="object"&&n!==null}function mZe(n){if(sEe(n)){const r=n.slice();return Object.getOwnPropertySymbols(n).forEach(o=>{r[o]=n[o]}),r}if(cEe(n)){const r={...n};return Object.getOwnPropertySymbols(n).forEach(o=>{r[o]=n[o]}),r}return n}function bZe(n,r,o){if(n[r]===o)return n;const s=mZe(n);return s[r]=o,s}function Uoe(n,r){let o=n,s=0;for(;s<r.length;)cEe(o)?o=o[r[s]]:sEe(o)?o=o[Number.parseInt(r[s])]:o=void 0,s++;return o}function zoe(n,r,o){let s=arguments.length>3&&arguments[3]!==void 0?arguments[3]:!1;if(r.length===0)return o;const c=r[0],d=zoe(n?n[c]:void 0,r.slice(1),o,s);if(cEe(n)||sEe(n))return bZe(n,c,d);if(s){const h=oIr.test(c)?[]:{};return h[c]=d,h}throw new Error("Path does not exist")}const oIr=/^\d+$/;function esn(n,r,o){if(r.length===0)return o(n);if(!Qan(n))throw new Error("Path doesn't exist");const s=r[0],c=esn(n[s],r.slice(1),o);return bZe(n,s,c)}function vZe(n,r){if(r.length===0)return n;if(!Qan(n))throw new Error("Path does not exist");if(r.length===1){const c=r[0];if(!(c in n))return n;const d=mZe(n);return sEe(d)&&d.splice(Number.parseInt(c),1),cEe(d)&&delete d[c],d}const o=r[0],s=vZe(n[o],r.slice(1));return bZe(n,o,s)}function wZe(n,r,o){const s=r.slice(0,r.length-1),c=r[r.length-1];return esn(n,s,d=>{if(!Array.isArray(d))throw new TypeError(`Array expected at path ${JSON.stringify(s)}`);const h=mZe(d);return h.splice(Number.parseInt(c),0,o),h})}function yZe(n,r){return n===void 0?!1:r.length===0?!0:n===null?!1:yZe(n[r[0]],r.slice(1))}function tsn(n){const r=n.split("/");return r.shift(),r.map(o=>o.replace(/~1/g,"/").replace(/~0/g,"~"))}function j$e(n){return n.map(aIr).join("")}function aIr(n){return`/${String(n).replace(/~/g,"~0").replace(/\//g,"~1")}`}function sIr(n,r,o){let s=n;for(let c=0;c<r.length;c++){gIr(r[c]);let d=r[c];const h=mIr(s,d.path);if(d.op==="add")s=lIr(s,h,d.value);else if(d.op==="remove")s=uIr(s,h);else if(d.op==="replace")s=cIr(s,h,d.value);else if(d.op==="copy")s=dIr(s,h,R$t(d.from));else if(d.op==="move")s=fIr(s,h,R$t(d.from));else if(d.op==="test")hIr(s,h,d.value);else throw new Error(`Unknown JSONPatch operation ${JSON.stringify(d)}`)}return s}function cIr(n,r,o){return yZe(n,r)?zoe(n,r,o):n}function uIr(n,r){return vZe(n,r)}function lIr(n,r,o){return EZe(n,r)?wZe(n,r,o):zoe(n,r,o)}function dIr(n,r,o){const s=Uoe(n,o);return EZe(n,r)?wZe(n,r,s):zoe(n,r,s)}function fIr(n,r,o){const s=Uoe(n,o),c=vZe(n,o);return EZe(c,r)?wZe(c,r,s):zoe(c,r,s)}function hIr(n,r,o){if(o===void 0)throw new Error(`Test failed: no value provided (path: "${j$e(r)}")`);if(!yZe(n,r))throw new Error(`Test failed: path not found (path: "${j$e(r)}")`);const s=Uoe(n,r);if(!rIr(s,o))throw new Error(`Test failed, value differs (path: "${j$e(r)}")`)}function EZe(n,r){if(r.length===0)return!1;const o=Uoe(n,Zan(r));return Array.isArray(o)}function pIr(n,r){if(iIr(r)!=="-")return r;const o=Zan(r),s=Uoe(n,o);return o.concat(s.length)}function gIr(n){if(!["add","remove","replace","copy","move","test"].includes(n.op))throw new Error(`Unknown JSONPatch op ${JSON.stringify(n.op)}`);if(typeof n.path!="string")throw new Error(`Required property "path" missing or not a string in operation ${JSON.stringify(n)}`);if((n.op==="copy"||n.op==="move")&&typeof n.from!="string")throw new Error(`Required property "from" missing or not a string in operation ${JSON.stringify(n)}`)}function mIr(n,r){return pIr(n,tsn(r))}function R$t(n){return tsn(n)}const bIr={[Sl.FlowCancelled]:" cancelled",[Sl.FlowCompleted]:" completed",[Sl.FlowStarted]:" started",[Sl.FlowInterrupted]:" interrupted",[Sl.FlowResumed]:" resumed"},vIr=n=>["flow_hashes","dialogue_stack"].includes(n),wIr=n=>["action_run_slot_rejections"].includes(n),yIr=n=>["session_started","user_featurization"].includes(n),EIr=n=>["/session_start","/silence_timeout"].includes(n),SIr=n=>["pattern_collect_information","pattern_user_silence"].includes(n);function nK(n){return n.__typename==="ConversationEvent"}function _Ir(n,r){if(r.update){const o=JSON.parse(r.update);return sIr(n,o)}return n}function kIr(n){return n.length>0&&n[n.length-1].flow_id!==void 0}const TIr=n=>{let r=[],o={};for(const s of n)if(nK(s)&&s.conversationEventType===Sl.Restart)r=[],o={};else if(CIr(s)){if(r=_Ir(r,s),!kIr(r))continue;const c=r[r.length-1];(!o[c.flow_id]||c.step_id==="START")&&(o[c.flow_id]=[]),o[c.flow_id].push(c.step_id)}return o},xIr=n=>n.filter(r=>nK(r)?r.conversationEventType===Sl.Slot?!0:r.conversationEventType===Sl.FlowStarted||r.conversationEventType===Sl.FlowCompleted||r.conversationEventType===Sl.FlowCancelled?!r.flowId?.startsWith("pattern_"):r.conversationEventType===Sl.SessionStarted:!1);function UB(n){return n.type===Vx.Bot||n.type===Vx.User}function CIr(n){return n.__typename==="StackEvent"}function AIr(n){const r={buttons:[],quickReplies:[]};if(n.text){const s=new RegExp(/\[(.*?)\]\s*\((.*?)\)/).exec(n.text);s&&(r.attachment=s[2])}return n.attachment&&(n.attachment.type==="image"&&(r.image=n.attachment.payload.src),n.attachment.type==="link"&&(r.attachment=n.attachment.payload.url),n.attachment.type==="template"&&(r.elements=n.attachment.payload.elements)),n.quick_replies&&(r.quickReplies=n.quick_replies.map(o=>({title:o.title,payload:o.payload,imageUrl:o.image_url}))),n.buttons&&(r.quickReplies=n.buttons.map(o=>({title:o.title,payload:o.payload,imageUrl:o.image_url}))),n.elements&&(r.elements=n.elements),r.custom=IIr(n),r}function IIr(n){const r=["text","attachment","quick_replies","buttons","elements","image"],o={};for(const s in n)n.hasOwnProperty(s)&&!r.includes(s)&&n[s]!==null&&(o[s]=n[s]);if(Object.keys(o).length>0)return o}function nsn(n){const r=n.map(s=>{s.metadata||(s.metadata={}),s.metadata.flow_id=s.metadata.active_flow||s.metadata.flow_id;const c=new Date(s.timestamp*1e3).toISOString();switch(s.event){case"user":{const d=s.parse_data?.intent_ranking?.map(h=>({id:n2(),name:h.name,confidence:h.confidence}))||[];return{id:n2(),conversationId:s.conversation_id,machineLearningModelId:"",entities:[],intents:d,text:s.text,tokens:[],timestamp:c,type:Vx.User,__typename:"Utterance",rephrase:!1,rephrasePrompt:null,commands:s.parse_data?.commands,metadata:{...s.metadata,parseData:s.parse_data,rawEvent:s}}}case"bot":return{id:n2(),conversationId:s.conversation_id,machineLearningModelId:"",entities:[],intents:[],text:s.text||"",responseData:AIr({...s.data,text:s.text}),tokens:[],timestamp:c,type:Vx.Bot,commands:null,rephrase:s.metadata?.metadata?.rephrase||!1,rephrasePrompt:s.metadata?.metadata?.rephrasePrompt||null,metadata:{...s.metadata,rawEvent:s},__typename:"Utterance"};case"stack":return{id:n2(),metadata:{...s.metadata,rawEvent:s},update:s.update||"",timestamp:c,__typename:"StackEvent"};case"action":return{id:n2(),actionText:s.name,conversationEventType:s.event.toUpperCase(),flowId:s.flow_id||s.metadata?.active_flow,metadata:{...s.metadata,rawEvent:s},name:s.name,slotValue:void 0,stepId:s.step_id||s.metadata?.step_id,timestamp:c,__typename:"ConversationEvent"};case"slot":return{id:n2(),actionText:void 0,conversationEventType:s.event.toUpperCase(),flowId:s.flow_id||s.metadata?.active_flow,metadata:{...s.metadata,rawEvent:s},name:s.name,slotValue:s.value,stepId:s.step_id||s.metadata?.step_id,timestamp:c,__typename:"ConversationEvent"};default:return{id:n2(),actionText:void 0,conversationEventType:s.event.toUpperCase(),flowId:s.flow_id||s.metadata?.active_flow,metadata:{...s.metadata,rawEvent:s},name:s.name,slotValue:void 0,stepId:s.step_id||s.metadata?.step_id,timestamp:c,__typename:"ConversationEvent"}}}),o={flow_id:void 0,step_id:void 0};for(const s of r){if(!s.metadata?.flow_id){s.metadata.flow_id=o.flow_id,s.metadata.step_id=o.step_id;continue}o.flow_id=s.metadata.flow_id,o.step_id=s.metadata.step_id}return r}const Va=(n,r)=>{ahr(n,r)},OIr=n=>{const r=n.trim();if(!r.startsWith("data: "))return null;try{const o=r.slice(6),s=JSON.parse(o);return man.parse(s)}catch(o){return Va(o,{tags:{component:"parseSSEDataLine",action:"parseSSEDataLine"},extra:{line:r}}),null}},NIr=n=>{const r=n.split(`
`),o=r.pop()||"",s=[];for(const c of r){const d=OIr(c);d&&s.push(d)}return{events:s,remainingBuffer:o}},RIr=n=>{if(n.event_data?.event_data){const r=n.event_data.event_data;return r.event=n.event,r}else{if(n.event_data)return{...n.event_data,event:n.event};{const{type:r,event_data:o,...s}=n;return s}}},PIr=n=>{const r=[];for(const o of n)(o.type==="event"||o.type==="text"||o.type==="file"||o.type==="references"||o.type==="link"||o.type==="logs"||o.type==="commit"||o.type==="plan")&&r.push(o);return r},rsn=(n,r="en-US")=>{try{let o;if(n instanceof Date)o=n;else if(typeof n=="string"||typeof n=="number"){const s=(typeof n=="string"?Number.parseFloat(n):n)*1e3;o=new Date(s)}else return console.error("Invalid timestamp type:",typeof n,n),"Invalid date";return Number.isNaN(o.getTime())?(console.error("Invalid date:",n),"Invalid date"):o.toLocaleString(r,{month:"short",day:"2-digit",year:"numeric",hour:"2-digit",minute:"2-digit",second:"2-digit",hour12:!1,timeZoneName:"short"})}catch(o){return console.error("Error formatting timestamp:",n,o),"Invalid date"}},LIr=n=>(n.getTime()/1e3).toString();function h0e(n){throw new Error('Could not dynamically require "'+n+'". Please configure the dynamicRequireTargets or/and ignoreDynamicRequires option of @rollup/plugin-commonjs appropriately for this require call to work.')}var $$e={exports:{}},P$t;function DIr(){return P$t||(P$t=1,function(n,r){(function(o){n.exports=o()})(function(){return function(){function o(s,c,d){function h(E,k){if(!c[E]){if(!s[E]){var x=typeof h0e=="function"&&h0e;if(!k&&x)return x(E,!0);if(b)return b(E,!0);var O=new Error("Cannot find module '"+E+"'");throw O.code="MODULE_NOT_FOUND",O}var P=c[E]={exports:{}};s[E][0].call(P.exports,function(D){var $=s[E][1][D];return h($||D)},P,P.exports,o,s,c,d)}return c[E].exports}for(var b=typeof h0e=="function"&&h0e,y=0;y<d.length;y++)h(d[y]);return h}return o}()({1:[function(o,s,c){Object.defineProperty(c,"__esModule",{value:!0}),c.default=void 0;function d(O){"@babel/helpers - typeof";return d=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(P){return typeof P}:function(P){return P&&typeof Symbol=="function"&&P.constructor===Symbol&&P!==Symbol.prototype?"symbol":typeof P},d(O)}function h(O,P){if(!(O instanceof P))throw new TypeError("Cannot call a class as a function")}function b(O,P){for(var D=0;D<P.length;D++){var $=P[D];$.enumerable=$.enumerable||!1,$.configurable=!0,"value"in $&&($.writable=!0),Object.defineProperty(O,E($.key),$)}}function y(O,P,D){return P&&b(O.prototype,P),Object.defineProperty(O,"prototype",{writable:!1}),O}function E(O){var P=k(O,"string");return d(P)=="symbol"?P:P+""}function k(O,P){if(d(O)!="object"||!O)return O;var D=O[Symbol.toPrimitive];if(D!==void 0){var $=D.call(O,P);if(d($)!="object")return $;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(O)}c.default=function(){function O(){var P=this,D=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},$=D.defaultLayoutOptions,U=$===void 0?{}:$,B=D.algorithms,G=B===void 0?["layered","stress","mrtree","radial","force","disco","sporeOverlap","sporeCompaction","rectpacking"]:B,K=D.workerFactory,X=D.workerUrl;if(h(this,O),this.defaultLayoutOptions=U,this.initialized=!1,typeof X>"u"&&typeof K>"u")throw new Error("Cannot construct an ELK without both 'workerUrl' and 'workerFactory'.");var ae=K;typeof X<"u"&&typeof K>"u"&&(ae=function(ye){return new Worker(ye)});var oe=ae(X);if(typeof oe.postMessage!="function")throw new TypeError("Created worker does not provide the required 'postMessage' function.");this.worker=new x(oe),this.worker.postMessage({cmd:"register",algorithms:G}).then(function(he){return P.initialized=!0}).catch(console.err)}return y(O,[{key:"layout",value:function(D){var $=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},U=$.layoutOptions,B=U===void 0?this.defaultLayoutOptions:U,G=$.logging,K=G===void 0?!1:G,X=$.measureExecutionTime,ae=X===void 0?!1:X;return D?this.worker.postMessage({cmd:"layout",graph:D,layoutOptions:B,options:{logging:K,measureExecutionTime:ae}}):Promise.reject(new Error("Missing mandatory parameter 'graph'."))}},{key:"knownLayoutAlgorithms",value:function(){return this.worker.postMessage({cmd:"algorithms"})}},{key:"knownLayoutOptions",value:function(){return this.worker.postMessage({cmd:"options"})}},{key:"knownLayoutCategories",value:function(){return this.worker.postMessage({cmd:"categories"})}},{key:"terminateWorker",value:function(){this.worker&&this.worker.terminate()}}])}();var x=function(){function O(P){var D=this;if(h(this,O),P===void 0)throw new Error("Missing mandatory parameter 'worker'.");this.resolvers={},this.worker=P,this.worker.onmessage=function($){setTimeout(function(){D.receive(D,$)},0)}}return y(O,[{key:"postMessage",value:function(D){var $=this.id||0;this.id=$+1,D.id=$;var U=this;return new Promise(function(B,G){U.resolvers[$]=function(K,X){K?(U.convertGwtStyleError(K),G(K)):B(X)},U.worker.postMessage(D)})}},{key:"receive",value:function(D,$){var U=$.data,B=D.resolvers[U.id];B&&(delete D.resolvers[U.id],U.error?B(U.error):B(null,U.data))}},{key:"terminate",value:function(){this.worker&&this.worker.terminate()}},{key:"convertGwtStyleError",value:function(D){if(D){var $=D.__java$exception;$&&($.cause&&$.cause.backingJsObject&&(D.cause=$.cause.backingJsObject,this.convertGwtStyleError(D.cause)),delete D.__java$exception)}}}])}()},{}],2:[function(o,s,c){(function(d){(function(){var h;typeof window<"u"?h=window:typeof d<"u"?h=d:typeof self<"u"&&(h=self);var b;function y(){}function E(){}function k(){}function x(){}function O(){}function P(){}function D(){}function $(){}function U(){}function B(){}function G(){}function K(){}function X(){}function ae(){}function oe(){}function he(){}function ye(){}function we(){}function ge(){}function Te(){}function xe(){}function Ue(){}function Z(){}function Re(){}function _e(){}function qe(){}function it(){}function Xe(){}function de(){}function We(){}function ht(){}function Ie(){}function $t(){}function An(){}function Yt(){}function Or(){}function Nt(){}function on(){}function br(){}function wi(){}function no(){}function Ri(){}function ro(){}function Ko(){}function Ca(){}function qi(){}function yi(){}function lo(){}function Eo(){}function Fc(){}function Yo(){}function zl(){}function ld(){}function Yc(){}function Sn(){}function Pn(){}function Kr(){}function Ei(){}function Sa(){}function bs(){}function Cc(){}function Ac(){}function Xc(){}function rc(){}function js(){}function Jc(){}function Id(){}function nh(){}function hu(){}function Qp(){}function Kb(){}function Km(){}function S2(){}function _2(){}function dv(){}function K_(){}function YI(){}function n6(){}function e0(){}function _l(){}function rh(){}function t0(){}function k2(){}function r6(){}function i6(){}function XI(){}function sy(){}function rC(){}function fv(){}function cy(){}function uy(){}function ly(){}function Od(){}function JI(){}function ZI(){}function hv(){}function dy(){}function T2(){}function QI(){}function x2(){}function o6(){}function fy(){}function Y_(){}function Ym(){}function C2(){}function iC(){}function hy(){}function oC(){}function py(){}function VM(){}function a6(){}function Kg(){}function s6(){}function c6(){}function eO(){}function u6(){}function A2(){}function WM(){}function Xm(){}function tO(){}function l6(){}function aC(){}function X_(){}function d6(){}function nO(){}function gy(){}function J_(){}function f6(){}function h6(){}function p6(){}function KM(){}function YM(){}function sC(){}function rO(){}function g6(){}function Z_(){}function Q_(){}function m6(){}function b6(){}function XM(){}function cC(){}function pv(){}function uC(){}function lC(){}function my(){}function v6(){}function iO(){}function JM(){}function Zc(){}function Bc(){}function ZM(){}function eg(){}function I2(){}function oO(){}function dC(){}function SU(){}function aO(){}function w6(){}function dd(){}function ek(){}function by(){}function QM(){}function y6(){}function Jm(){}function sO(){}function e8(){}function t8(){}function n8(){}function r8(){}function i8(){}function tg(){}function _U(){}function Zm(){}function kU(){}function o8(){}function n0(){}function cO(){}function E6(){}function al(){}function gv(){}function S6(){}function TU(){}function _6(){}function xU(){}function CU(){}function AU(){}function a8(){}function uO(){}function tk(){}function k6(){}function nk(){}function s8(){}function lO(){}function dO(){}function T6(){}function c8(){}function fO(){}function vy(){}function IU(){}function u8(){}function Yg(){}function hO(){}function rk(){}function ik(){}function ok(){}function OU(){}function pO(){}function l8(){}function d8(){}function x6(){}function mv(){}function f8(){}function h8(){}function gO(){}function p8(){}function wy(){}function mO(){}function NU(){}function RU(){}function Xg(){}function r0(){}function Qm(){}function C6(){}function g8(){}function m8(){}function A6(){}function PU(){}function ak(){}function yy(){}function Dt(){}function Rn(){}function En(){}function ui(){}function Lo(){}function Ws(){}function Nd(){}function Ks(){}function bO(){}function b8(){}function Jg(){}function vO(){}function LU(){}function sk(){}function fC(){}function hC(){}function bc(){}function Yb(){}function I6(){}function bp(){}function eb(){}function ck(){}function O6(){}function O2(){}function Xb(){}function uk(){}function N2(){}function pC(){}function gC(){}function R2(){}function wO(){}function ng(){}function P2(){}function vp(){}function mC(){}function yO(){}function L2(){}function N6(){}function lk(){}function bv(){}function Ey(){}function EO(){}function bC(){}function tb(){}function nb(){}function qu(){}function sl(){}function vv(){}function D2(){}function R6(){}function P6(){}function vC(){}function Rd(){}function cl(){}function vs(){}function Jb(){}function rg(){}function dk(){}function rb(){}function v8(){}function w8(){}function wC(){}function yC(){}function y8(){}function Sy(){}function M2(){}function _y(){}function De(){}function Lh(){}function wp(){}function ig(){}function ts(){}function fk(){}function SO(){}function j2(){}function Hl(){}function _O(){}function E8(){}function S8(){}function _8(){}function DU(){}function Qc(){}function kO(){}function EC(){}function SC(){}function L6(){}function hk(){}function og(){}function MU(){}function k8(){}function jU(){}function T8(){}function $2(){}function wv(){}function _C(){}function x8(){}function yp(){}function C8(){}function TO(){}function D6(){}function A8(){}function xO(){}function M6(){}function pk(){}function I8(){}function O8(){}function i0(){}function F2(){}function gk(){}function ky(){}function yv(){}function kC(){}function Zb(){}function TC(){}function Qb(){}function Dh(){}function N8(){}function j6(){}function B2(){}function Ty(){}function Ev(){}function CO(){}function $6(){}function e1(){}function Wa(){}function Uc(){}function ws(){}function mk(){}function U2(){}function Mh(){}function bk(){}function xy(){}function ib(){}function R8(){}function Cs(){}function xC(){}function F6(){}function P8(){}function L8(){}function B6(){}function U6(){}function z2(){}function AO(){}function ag(){}function Pd(){}function H2(){}function CC(){}function IO(){}function AC(){}function o0(){}function OO(){}function z6(){}function H6(){}function G6(){}function NO(){}function D8(){}function RO(){}function q6(){}function V6(){}function PO(){}function IC(){}function $U(){}function W6(){}function K6(){}function M8(){}function LO(){}function j8(){}function DO(){}function Y6(){}function MO(){}function $8(){}function X6(){}function Da(){}function vk(){}function jO(){}function wk(){}function OC(){}function NC(){}function F8(){}function Zg(){}function FU(){}function RC(){}function $O(){}function J6(){}function FO(){}function sg(){}function Sv(){}function PC(){}function Z6(){}function LC(){}function B8(){}function DC(){}function Q6(){}function BO(){}function MC(){}function _v(){}function U8(){}function Cy(){}function z8(){}function e5(){}function G2(){}function t5(){}function UO(){}function H8(){}function n5(){}function jC(){}function G8(){}function r5(){}function q8(){}function Kd(){}function cg(){}function yk(){}function q2(){}function V2(){}function kv(){}function i5(){}function W2(){}function V8(){}function Qg(){}function Ep(){}function ob(){}function W8(){}function Tv(){}function a0(){}function Gl(){}function $C(){}function K2(){}function FC(){}function BC(){}function BU(){}function Ay(){}function Y2(){}function o5(){}function zO(){}function K8(){}function UC(){}function s0(){}function ug(){}function a5(){}function xv(){}function s5(){}function Y8(){}function xf(){}function Vu(){}function HO(){}function GO(){}function qO(){}function zC(){}function X2(){}function fo(){}function Ek(){}function VO(){}function Iy(){}function X8(){}function HC(){}function WO(){}function Cf(){}function J8(){}function c5(){}function GC(){}function mt(){}function u5(){}function l5(){}function Z8(){}function Cv(){}function t1(){}function Sp(){}function KO(){}function Av(){}function YO(){}function XO(){}function qC(){}function pu(){}function Q8(){}function e9(){}function JO(){}function VC(){}function Oy(){}function t9(){}function WC(){}function Iv(){}function em(){}function ZO(){}function KC(){}function Sk(){}function YC(){}function J2(){}function XC(){}function JC(){}function c0(){}function Ys(){}function Ov(){}function n9(){}function Yd(){}function UU(){}function ZC(){}function r9(){}function Wu(){}function lg(){}function n1(){}function d5(){}function i9(){}function zU(){}function f5(){}function tm(){}function r1(){}function h5(){}function dg(){}function o9(){}function Ny(){}function Z2(){}function Nv(){}function p5(){}function g5(){}function u0(){}function a9(){}function ab(){}function Q2(){}function jh(){}function _k(){}function s9(){}function c9(){}function kk(){}function u9(){}function QO(){}function l9(){}function HU(){}function m5(){}function zc(){}function Nu(){}function as(){}function $s(){}function Hc(){}function nm(){}function Ry(){}function Tk(){}function b5(){}function Rv(){}function ql(){}function Py(){}function eS(){}function v5(){}function sb(){}function xk(){}function QC(){}function fg(){}function w5(){}function y5(){}function Ck(){}function e3(){}function Ly(){}function ih(){}function Dy(){}function Ak(){}function Ik(){}function E5(){}function S5(){}function d9(){}function Ok(){}function _5(){}function $h(){}function f9(){}function rm(){}function h9(){}function p9(){}function GU(){}function My(){}function eN(){}function k5(){}function Nk(){}function g9(){}function m9(){}function b9(){}function T5(){}function x5(){}function qU(){}function VU(){}function v9(){}function C5(){}function w9(){}function Rk(){}function tN(){}function y9(){}function t3(){}function E9(){}function S9(){}function _9(){}function k9(){}function nN(){}function T9(){}function x9(){}function gu(){}function A5(){}function WU(){}function C9(){}function rN(){}function KU(){}function A9(){}function n3(){}function jy(){}function iN(){}function r3(){}function oN(){}function $y(){}function l0(){}function i3(){}function YU(){}function o3(){}function a3(){}function I9(){}function O9(){}function I5(){}function aN(){}function N9(){}function R9(){}function XU(){}function O5(){}function N5(){}function JU(){}function P9(){}function s3(){}function ZU(){}function sN(){a7()}function ss(){SAe()}function R5(){Rz()}function hg(){Qh()}function P5(){R3e()}function L9(){dG()}function Pv(){Nz()}function Xo(){WP()}function cN(){aH()}function Xd(){xot()}function Vl(){dR()}function QU(){Ypt()}function Fy(){cw()}function D9(){Uft()}function M9(){Cgt()}function uN(){Pxe()}function eu(){Lxe()}function L5(){kY()}function Jd(){xOe()}function ez(){ot()}function i1(){zft()}function lN(){pbt()}function tS(){gbt()}function ul(){Hht()}function kl(){mbt()}function dN(){lR()}function D5(){Ec()}function tz(){qft()}function j9(){bbt()}function nz(){uw()}function $9(){Fbt()}function rz(){FIe()}function F9(){Hft()}function cb(){bwt()}function fN(){Gft()}function M5(){$St()}function B9(){QAe()}function hN(){y1()}function _p(){N0t()}function Pk(){KIe()}function o1(){Ag()}function j5(){mG()}function U9(){Nde()}function z9(){Ble()}function c3(){_J()}function im(){GZ()}function a1(){Rr()}function fa(){rZ()}function d0(){u3e()}function om(){dQ()}function $5(){Uue()}function ll(){Cut()}function vc(){JIe()}function H9(e){an(e)}function G9(e){this.a=e}function nS(e){this.a=e}function iz(e){this.a=e}function oz(e){this.a=e}function az(e){this.a=e}function q9(e){this.a=e}function F5(e){this.a=e}function Lk(e){this.a=e}function rS(e){this.a=e}function V9(e){this.a=e}function sz(e){this.a=e}function Dk(e){this.a=e}function u3(e){this.a=e}function W9(e){this.c=e}function pN(e){this.a=e}function iS(e){this.a=e}function B5(e){this.a=e}function By(e){this.a=e}function K9(e){this.a=e}function gN(e){this.a=e}function cz(e){this.a=e}function Zd(e){this.a=e}function Af(e){this.a=e}function Y9(e){this.b=e}function uz(e){this.a=e}function lz(e){this.a=e}function l3(e){this.a=e}function mN(e){this.a=e}function U5(e){this.a=e}function dz(e){this.a=e}function Mk(e){this.a=e}function fz(e){this.a=e}function X9(e){this.a=e}function J9(e){this.a=e}function z5(e){this.a=e}function H5(e){this.a=e}function bN(e){this.a=e}function G5(e){this.a=e}function oS(e){this.a=e}function d3(e){this.a=e}function oh(){this.a=[]}function q5(e,t){e.a=t}function Z9(e,t){e.a=t}function V5(e,t){e.b=t}function Q9(e,t){e.c=t}function hz(e,t){e.c=t}function W5(e,t){e.d=t}function Uy(e,t){e.d=t}function Lv(e,t){e.j=t}function Wl(e,t){e.k=t}function vN(e,t){e.c=t}function jk(e,t){e.c=t}function Dv(e,t){e.a=t}function GK(e,t){e.a=t}function pz(e,t){e.f=t}function wN(e,t){e.a=t}function qK(e,t){e.b=t}function yN(e,t){e.d=t}function aS(e,t){e.i=t}function K5(e,t){e.o=t}function VK(e,t){e.r=t}function Y5(e,t){e.a=t}function X5(e,t){e.b=t}function e7(e,t){e.e=t}function EN(e,t){e.f=t}function SN(e,t){e.g=t}function J5(e,t){e.e=t}function gz(e,t){e.f=t}function WK(e,t){e.f=t}function $k(e,t){e.a=t}function w(e,t){e.b=t}function _(e,t){e.n=t}function C(e,t){e.a=t}function j(e,t){e.a=t}function q(e,t){e.a=t}function te(e,t){e.c=t}function Se(e,t){e.c=t}function Me(e,t){e.c=t}function pt(e,t){e.d=t}function Xt(e,t){e.d=t}function $n(e,t){e.e=t}function Vn(e,t){e.e=t}function ln(e,t){e.g=t}function Zt(e,t){e.f=t}function Do(e,t){e.j=t}function Oo(e,t){e.a=t}function mu(e,t){e.a=t}function zt(e,t){e.b=t}function Ot(e){e.b=e.a}function Wt(e){e.c=e.d.d}function zn(e){this.a=e}function bi(e){this.a=e}function Ka(e){this.a=e}function Hi(e){this.a=e}function Vi(e){this.a=e}function Ic(e){this.a=e}function Fs(e){this.a=e}function Mv(e){this.a=e}function f3(e){this.a=e}function KK(e){this.a=e}function kp(e){this.a=e}function Fk(e){this.d=e}function zy(e){this.a=e}function nae(e){this.a=e}function rae(e){this.a=e}function YK(e){this.a=e}function iae(e){this.a=e}function mz(e){this.a=e}function h3(e){this.b=e}function p3(e){this.b=e}function bz(e){this.b=e}function ne(e){this.c=e}function vz(e){this.c=e}function XK(e){this.a=e}function wr(e){this.a=e}function oae(e){this.a=e}function JK(e){this.a=e}function ZK(e){this.a=e}function QK(e){this.a=e}function eY(e){this.a=e}function tY(e){this.a=e}function aae(e){this.c=e}function g3(e){this.a=e}function I(e){this.a=e}function M(e){this.a=e}function Y(e){this.a=e}function fe(e){this.a=e}function Ce(e){this.a=e}function ze(e){this.a=e}function dt(e){this.a=e}function It(e){this.a=e}function qt(e){this.a=e}function Hn(e){this.a=e}function qn(e){this.a=e}function sr(e){this.a=e}function ri(e){this.a=e}function Ji(e){this.a=e}function Uo(e){this.a=e}function ns(e){this.a=e}function zo(e){this.a=e}function rs(e){this.a=e}function Xs(e){this.a=e}function s1(e){this.a=e}function pg(e){this.a=e}function c1(e){this.a=e}function Qd(e){this.a=e}function jv(e){this.a=e}function Hy(e){this.a=e}function wz(e){this.a=e}function ZEe(e){this.a=e}function t7(e){this.a=e}function nY(e){this.e=e}function Z5(e){this.a=e}function eet(e){this.a=e}function _N(e){this.a=e}function QEe(e){this.a=e}function tet(e){this.a=e}function net(e){this.a=e}function ret(e){this.a=e}function iet(e){this.a=e}function oet(e){this.a=e}function aet(e){this.a=e}function set(e){this.a=e}function cet(e){this.a=e}function uet(e){this.a=e}function det(e){this.a=e}function fet(e){this.a=e}function e2e(e){this.a=e}function het(e){this.a=e}function pet(e){this.a=e}function get(e){this.a=e}function met(e){this.a=e}function bet(e){this.a=e}function vet(e){this.a=e}function wet(e){this.a=e}function yet(e){this.a=e}function Eet(e){this.a=e}function _et(e){this.a=e}function ket(e){this.a=e}function Tet(e){this.a=e}function xet(e){this.a=e}function Cet(e){this.a=e}function Aet(e){this.a=e}function Iet(e){this.a=e}function Oet(e){this.a=e}function Net(e){this.a=e}function Ret(e){this.a=e}function Pet(e){this.a=e}function Let(e){this.a=e}function Det(e){this.a=e}function Met(e){this.a=e}function jet(e){this.a=e}function $et(e){this.a=e}function Fet(e){this.a=e}function Bet(e){this.a=e}function Uet(e){this.a=e}function zet(e){this.a=e}function Het(e){this.a=e}function Get(e){this.a=e}function qet(e){this.a=e}function Vet(e){this.a=e}function Wet(e){this.a=e}function Ket(e){this.a=e}function Yet(e){this.a=e}function Xet(e){this.a=e}function Jet(e){this.a=e}function Zet(e){this.a=e}function Qet(e){this.a=e}function ett(e){this.c=e}function ttt(e){this.b=e}function ntt(e){this.a=e}function rtt(e){this.a=e}function itt(e){this.a=e}function ott(e){this.a=e}function att(e){this.a=e}function stt(e){this.a=e}function ctt(e){this.a=e}function utt(e){this.a=e}function ltt(e){this.a=e}function dtt(e){this.a=e}function ftt(e){this.a=e}function htt(e){this.a=e}function ptt(e){this.a=e}function gtt(e){this.a=e}function mtt(e){this.a=e}function btt(e){this.a=e}function vtt(e){this.a=e}function wtt(e){this.a=e}function ytt(e){this.a=e}function Ett(e){this.a=e}function Stt(e){this.a=e}function _tt(e){this.a=e}function ktt(e){this.a=e}function Ttt(e){this.a=e}function xtt(e){this.a=e}function Ctt(e){this.a=e}function Att(e){this.a=e}function u1(e){this.a=e}function m3(e){this.a=e}function Itt(e){this.a=e}function Ott(e){this.a=e}function Ntt(e){this.a=e}function Rtt(e){this.a=e}function Ptt(e){this.a=e}function Ltt(e){this.a=e}function Dtt(e){this.a=e}function Mtt(e){this.a=e}function jtt(e){this.a=e}function $tt(e){this.a=e}function Ftt(e){this.a=e}function Btt(e){this.a=e}function Utt(e){this.a=e}function ztt(e){this.a=e}function Htt(e){this.a=e}function Gtt(e){this.a=e}function qtt(e){this.a=e}function Vtt(e){this.a=e}function Wtt(e){this.a=e}function Ktt(e){this.a=e}function Ytt(e){this.a=e}function Xtt(e){this.a=e}function Jtt(e){this.a=e}function Ztt(e){this.a=e}function Qtt(e){this.a=e}function ent(e){this.a=e}function rY(e){this.a=e}function tnt(e){this.f=e}function nnt(e){this.a=e}function rnt(e){this.a=e}function int(e){this.a=e}function ont(e){this.a=e}function ant(e){this.a=e}function snt(e){this.a=e}function cnt(e){this.a=e}function unt(e){this.a=e}function lnt(e){this.a=e}function dnt(e){this.a=e}function fnt(e){this.a=e}function hnt(e){this.a=e}function pnt(e){this.a=e}function gnt(e){this.a=e}function mnt(e){this.a=e}function bnt(e){this.a=e}function vnt(e){this.a=e}function wnt(e){this.a=e}function ynt(e){this.a=e}function Ent(e){this.a=e}function Snt(e){this.a=e}function _nt(e){this.a=e}function knt(e){this.a=e}function Tnt(e){this.a=e}function xnt(e){this.a=e}function Cnt(e){this.a=e}function Ant(e){this.a=e}function sae(e){this.a=e}function Si(e){this.b=e}function t2e(e){this.a=e}function Int(e){this.a=e}function Ont(e){this.a=e}function Nnt(e){this.a=e}function Rnt(e){this.a=e}function Pnt(e){this.a=e}function Lnt(e){this.a=e}function Dnt(e){this.a=e}function Mnt(e){this.a=e}function yz(e){this.a=e}function jnt(e){this.a=e}function $nt(e){this.a=e}function iY(e){this.a=e}function oY(e){this.a=e}function Fnt(e){this.b=e}function n2e(e){this.c=e}function aY(e){this.e=e}function Bnt(e){this.d=e}function cae(e){this.a=e}function Unt(e){this.a=e}function r2e(e){this.a=e}function i2e(e){this.a=e}function sS(e){this.e=e}function $hn(){this.a=0}function et(){kse(this)}function Wn(){nu(this)}function uae(){kdt(this)}function znt(){}function cS(){this.c=F7e}function Hnt(e,t){e.b+=t}function Fhn(e,t){t.Wb(e)}function Bhn(e){return e.a}function Uhn(e){return e.a}function zhn(e){return e.a}function Hhn(e){return e.a}function Ghn(e){return e.a}function ie(e){return e.e}function qhn(){return null}function Vhn(){return null}function Whn(){H2e(),r7n()}function Khn(e){e.b.Mf(e.e)}function Gnt(e){e.b=new Cae}function n7(e,t){e.b=t-e.b}function r7(e,t){e.a=t-e.a}function Yhn(e,t){No(t,e)}function qnt(e,t){t.gd(e.a)}function vn(e,t){e.push(t)}function Vnt(e,t){e.sort(t)}function Xhn(e,t,i){e.Wd(i,t)}function Ez(e,t){e.e=t,t.b=e}function o2e(e){lm(),this.a=e}function Wnt(e){lm(),this.a=e}function Knt(e){lm(),this.a=e}function lae(e){_S(),this.a=e}function Ynt(e){AP(),Hpe.je(e)}function Gy(){hct.call(this)}function dae(){Gy.call(this)}function a2e(){Gy.call(this)}function s2e(){hct.call(this)}function Xnt(){Gy.call(this)}function Sz(){Gy.call(this)}function Tl(){Gy.call(this)}function kN(){Gy.call(this)}function cr(){Gy.call(this)}function If(){Gy.call(this)}function Jnt(){Gy.call(this)}function ic(){Gy.call(this)}function Znt(){Gy.call(this)}function Qnt(){this.a=this}function sY(){this.Bb|=256}function ert(){this.b=new ast}function c2e(){c2e=Z,new Wn}function trt(){a2e.call(this)}function cY(e,t){Qe(e.a,t)}function Jhn(e,t){kAe(e.c,t)}function Zhn(e,t){Go(e.b,t)}function Qhn(e,t){OZ(e.a,t)}function epn(e,t){hle(e.a,t)}function Q5(e,t){Zr(e.e,t)}function TN(e){KZ(e.c,e.b)}function tpn(e,t){e.kc().Nb(t)}function Bk(e,t){e.length=t}function Mo(){this.a=new Wn}function nrt(){this.a=new Wn}function u2e(){this.a=new yot}function uY(){this.a=new et}function fae(){this.a=new et}function l2e(){this.a=new et}function ah(){this.a=new Cc}function qy(){this.a=new zpt}function d2e(){this.a=new Dft}function f2e(){this.a=new nut}function h2e(){this.a=new rh}function rrt(){this.a=new aht}function irt(){this.a=new et}function ort(){this.a=new et}function art(){this.a=new et}function p2e(){this.a=new et}function srt(){this.d=new et}function crt(){this.a=new Mo}function urt(){this.a=new Wn}function lrt(){this.b=new Wn}function drt(){this.b=new et}function g2e(){this.e=new et}function frt(){this.a=new Fy}function hrt(){this.d=new et}function m2e(e){this.a=gCn(e)}function xN(){znt.call(this)}function i7(){znt.call(this)}function hae(){i7.call(this)}function b2e(){xN.call(this)}function v2e(){uY.call(this)}function prt(){hX.call(this)}function grt(){p2e.call(this)}function mrt(){et.call(this)}function brt(){lft.call(this)}function vrt(){lft.call(this)}function wrt(){S2e.call(this)}function yrt(){S2e.call(this)}function Ert(){S2e.call(this)}function Srt(){_2e.call(this)}function fd(){ki.call(this)}function o7(){t9.call(this)}function w2e(){t9.call(this)}function _rt(){$rt.call(this)}function krt(){$rt.call(this)}function Trt(){Wn.call(this)}function xrt(){Wn.call(this)}function Crt(){Wn.call(this)}function pae(){dbt.call(this)}function Art(){Mo.call(this)}function Irt(){sY.call(this)}function gae(){JSe.call(this)}function y2e(){Wn.call(this)}function mae(){JSe.call(this)}function bae(){Wn.call(this)}function Ort(){Wn.call(this)}function E2e(){ZC.call(this)}function Nrt(){E2e.call(this)}function Rrt(){ZC.call(this)}function Prt(){P9.call(this)}function S2e(){this.a=new Mo}function Lrt(){this.a=new Wn}function _2e(){this.a=new Wn}function Drt(){this.a=new et}function Mrt(){this.j=new et}function k2e(){this.a=new H8}function jrt(){this.a=new Dit}function CN(){this.a=new ki}function $rt(){this.a=new Ov}function a7(){a7=Z,Dpe=new E}function vae(){vae=Z,Mpe=new Brt}function wae(){wae=Z,jpe=new Frt}function Frt(){iS.call(this,"")}function Brt(){iS.call(this,"")}function Urt(e){Mmt.call(this,e)}function zrt(e){Mmt.call(this,e)}function T2e(e){q9.call(this,e)}function x2e(e){Wit.call(this,e)}function npn(e){Wit.call(this,e)}function rpn(e){x2e.call(this,e)}function ipn(e){x2e.call(this,e)}function opn(e){x2e.call(this,e)}function Hrt(e){rue.call(this,e)}function Grt(e){rue.call(this,e)}function qrt(e){Tut.call(this,e)}function Vrt(e){Y2e.call(this,e)}function s7(e){yY.call(this,e)}function C2e(e){yY.call(this,e)}function Wrt(e){yY.call(this,e)}function A2e(e){RNn.call(this,e)}function I2e(e){A2e.call(this,e)}function oc(e){zlt.call(this,e)}function Krt(e){oc.call(this,e)}function AN(){d3.call(this,{})}function yae(e){dP(),this.a=e}function Yrt(e){e.b=null,e.c=0}function apn(e,t){e.e=t,F2t(e,t)}function spn(e,t){e.a=t,wRn(e)}function Eae(e,t,i){e.a[t.g]=i}function cpn(e,t,i){$In(i,e,t)}function upn(e,t){Hbn(t.i,e.n)}function Xrt(e,t){QTn(e).Ad(t)}function lpn(e,t){return e*e/t}function Jrt(e,t){return e.g-t.g}function dpn(e,t){e.a.ec().Kc(t)}function fpn(e){return new oS(e)}function hpn(e){return new cT(e)}function Zrt(){Zrt=Z,p6e=new y}function O2e(){O2e=Z,g6e=new ge}function lY(){lY=Z,L$=new X}function dY(){dY=Z,Bpe=new Zst}function Qrt(){Qrt=Z,jAt=new oe}function Sae(e){Xxe(),this.a=e}function eit(e){xut(),this.a=e}function uS(e){rce(),this.f=e}function _ae(e){rce(),this.f=e}function fY(e){oc.call(this,e)}function Ru(e){oc.call(this,e)}function tit(e){oc.call(this,e)}function kae(e){zlt.call(this,e)}function eP(e){oc.call(this,e)}function dn(e){oc.call(this,e)}function As(e){oc.call(this,e)}function nit(e){oc.call(this,e)}function IN(e){oc.call(this,e)}function f0(e){oc.call(this,e)}function bu(e){an(e),this.a=e}function c7(e){Lke(e,e.length)}function N2e(e){return dE(e),e}function lt(e){return an(e),e}function ke(e){return an(e),e}function _z(e){return an(e),e}function R2e(e){return an(e),e}function ppn(e){return an(e),e}function u7(e){return e.b==e.c}function Uk(e){return!!e&&e.b}function gpn(e){return!!e&&e.k}function mpn(e){return!!e&&e.j}function h0(e){oc.call(this,e)}function ON(e){oc.call(this,e)}function am(e){oc.call(this,e)}function vr(e){oc.call(this,e)}function Tae(e){oc.call(this,e)}function xae(e){d_e.call(this,e,0)}function Cae(){wTe.call(this,12,3)}function Aae(){this.a=dr(or(Fu))}function NN(e){lm(),this.a=or(e)}function rit(e,t){e._d(t),t.$d(e)}function bpn(e,t,i){e.c.Cf(t,i)}function se(e,t){return Lft(e,t)}function vpn(e,t){return e.a-t.a}function wpn(e,t){return e.a-t.a}function ypn(e,t){return t in e.a}function Epn(e,t){return e.a-t.a}function sm(e){dn.call(this,e)}function P2e(e){Ru.call(this,e)}function p0(){bi.call(this,"")}function l7(){bi.call(this,"")}function $v(){bi.call(this,"")}function RN(){bi.call(this,"")}function L2e(e){Ru.call(this,e)}function tP(e){h3.call(this,e)}function Iae(e){cX.call(this,e)}function iit(e){tP.call(this,e)}function oit(){oit=Z,XAt=lOn()}function hY(){hY=Z,AP()}function ait(){throw ie(new cr)}function D2e(){throw ie(new cr)}function sit(){throw ie(new cr)}function Spn(){throw ie(new cr)}function _pn(){throw ie(new cr)}function kpn(){throw ie(new cr)}function pY(e,t){return bue(e,t)}function cit(e){return e.a?e.b:0}function Tpn(e){return e.a?e.b:0}function xpn(e,t){return e.g=t,e}function Cpn(e,t){return e.b=t,e}function uit(e,t){return e.c=t,e}function lit(e,t){return e.f=t,e}function M2e(e,t){return e.f=t,e}function j2e(e,t){return e.a=t,e}function Apn(e,t){return e.k=t,e}function $2e(e,t){return e.e=t,e}function Ipn(e,t){return e.e=t,e}function F2e(e,t){return e.a=t,e}function Opn(e,t){return e.f=t,e}function Npn(e,t){e.b=new ja(t)}function Rpn(e,t,i){t.Ne(e.a[i])}function Ppn(e,t,i){t.Ad(e.a[i])}function Lpn(e,t){tf(),t.n.a+=e}function Dpn(e,t){cw(),cc(t,e)}function Mpn(e,t){return e.b-t.b}function jpn(e,t){return e.g-t.g}function $pn(e,t){return e.s-t.s}function Fpn(e,t){return e?0:t-1}function dit(e,t){return e?0:t-1}function Bpn(e,t){return e?t-1:0}function Upn(e,t){return t.pg(e)}function zk(e,t){return e.b=t,e}function gY(e,t){return e.a=t,e}function d7(e,t){return e.a=t,e}function Hk(e,t){return e.c=t,e}function nP(e,t){return e.c=t,e}function Gk(e,t){return e.d=t,e}function qk(e,t){return e.e=t,e}function B2e(e,t){return e.f=t,e}function rP(e,t){return e.b=t,e}function Lt(e,t){return e.b=t,e}function vt(e,t){return e.c=t,e}function wt(e,t){return e.d=t,e}function yt(e,t){return e.e=t,e}function zpn(e,t){return e.f=t,e}function Et(e,t){return e.g=t,e}function St(e,t){return e.a=t,e}function _t(e,t){return e.i=t,e}function kt(e,t){return e.j=t,e}function Hpn(e,t){return e.j=t,e}function fit(e,t){return e.k=t,e}function hit(e){Gdt.call(this,e)}function U2e(e){Gdt.call(this,e)}function mY(e){Hse.call(this,e)}function Fv(e){OS.call(this,e)}function pit(e){Pce.call(this,e)}function git(e){Pce.call(this,e)}function mit(){HSe.call(this,"")}function ha(){this.a=0,this.b=0}function bit(){this.b=0,this.a=0}function vit(e,t){e.b=0,vT(e,t)}function Gpn(e,t){e.c=t,e.b=!0}function qpn(e,t,i){Kvn(e.a,t,i)}function wit(e,t){return e.c._b(t)}function Tp(e){return e.e&&e.e()}function Oae(e){return e?e.d:null}function yit(e,t){return E0t(e.b,t)}function Vpn(e){return e?e.g:null}function Wpn(e){return e?e.i:null}function Vy(e){return d1(e),e.o}function Eit(){throw ie(new cr)}function Sit(){Sit=Z,SMt=kIn()}function z2e(){z2e=Z,Mc=mRn()}function H2e(){H2e=Z,qE=fj()}function Bv(){Bv=Z,$Dt=SIn()}function _it(){_it=Z,Qo=ROn()}function iP(){iP=Z,$7e=_In()}function kit(){throw ie(new cr)}function Tit(){throw ie(new cr)}function xit(){throw ie(new cr)}function Cit(){throw ie(new cr)}function Ait(){throw ie(new cr)}function bY(e){this.a=new PN(e)}function G2e(e){hkt(),g7n(this,e)}function g0(e){this.a=new dce(e)}function b3(e,t){for(;e.Pe(t););}function q2e(e,t){for(;e.zd(t););}function v3(e,t){return e.a+=t,e}function Nae(e,t){return e.a+=t,e}function m0(e,t){return e.a+=t,e}function lS(e,t){return e.a+=t,e}function oP(e){return Kv(e),e.a}function kz(e){return e.b!=e.d.c}function Iit(e){return e.l|e.m<<22}function V2e(e,t){return e.d[t.p]}function Oit(e,t){return G5n(e,t)}function W2e(e,t,i){e.splice(t,i)}function Nit(e){e.c?tSt(e):nSt(e)}function vY(e){this.a=0,this.b=e}function Rit(){this.a=new Gj(xMe)}function Pit(){this.a=new Gj(XMe)}function Lit(){this.b=new Gj(Ibe)}function Dit(){this.b=new Gj(Ibe)}function Mit(){this.b=new Gj(qDe)}function jit(){throw ie(new cr)}function $it(){throw ie(new cr)}function Fit(){throw ie(new cr)}function Bit(){throw ie(new cr)}function Uit(){throw ie(new cr)}function zit(){throw ie(new cr)}function Hit(){throw ie(new cr)}function Git(){throw ie(new cr)}function qit(){throw ie(new cr)}function Vit(){throw ie(new cr)}function Kpn(){throw ie(new ic)}function Ypn(){throw ie(new ic)}function Tz(e){this.a=new vot(e)}function wY(e){this.b=f(or(e),91)}function K2e(e,t){this.b=e,this.c=t}function aP(e,t){this.e=e,this.d=t}function Wit(e){QSe(e.dc()),this.c=e}function Kit(e,t){this.a=e,this.b=t}function Yit(e,t){this.a=e,this.b=t}function Xit(e,t){this.a=e,this.b=t}function Jit(e,t){this.a=e,this.b=t}function Zit(e,t){this.a=e,this.b=t}function Qit(e,t){this.a=e,this.b=t}function eot(e,t){this.a=e,this.b=t}function tot(e,t){this.b=e,this.a=t}function dS(e,t){this.g=e,this.i=t}function not(e,t){this.b=e,this.a=t}function xz(e,t){N3.call(this,e,t)}function sP(e,t){xz.call(this,e,t)}function rot(e,t){this.b=e,this.a=t}function iot(e,t){this.b=e,this.a=t}function oot(e,t){this.a=e,this.b=t}function ur(e,t){this.f=e,this.g=t}function yY(e){this.b=f(or(e),49)}function Rae(e,t){this.a=e,this.b=t}function aot(e,t){this.a=e,this.f=t}function sot(e){this.a=f(or(e),15)}function Y2e(e){this.a=f(or(e),15)}function cot(e,t){this.b=e,this.c=t}function uot(e){this.a=f(or(e),91)}function Xpn(e,t){this.a=e,this.b=t}function lot(e,t){this.a=e,this.b=t}function dot(e,t){return Yu(e.b,t)}function fot(e,t){return e>t&&t<n$}function Cz(e){return zg[e]!=-1}function Az(e){return!e||_dt(e)}function f7(e){return e.f.c+e.i.c}function Jpn(e){return sdt(),e?MAt:DAt}function Iz(e,t){return e.b.Gc(t)}function hot(e,t){return e.b.Hc(t)}function pot(e,t){return e.b.Oc(t)}function got(e,t){return e.c.uc(t)}function mot(e,t){return ei(e.c,t)}function bot(e,t){return e.b.Gc(t)}function Zpn(e,t){Ede(je(e.a),t)}function Qpn(e,t){Ede(je(e.a),t)}function xp(e,t){return e.a._b(t)}function vot(e){Ckn(this,e,MNn())}function PN(e){w1t.call(this,e,0)}function LN(e){this.c=e,mmt(this)}function ki(){$st(this),nf(this)}function wot(){Lpe==null&&(Lpe=[])}function yot(){dce.call(this,null)}function Pae(e,t){this.a=e,this.b=t}function EY(){this.q=new h.Date}function egn(){Cee!=0&&(Cee=0),Aee=-1}function cP(){cP=Z,z6e=new wi}function b0(){b0=Z,UAt=new Ue}function uP(){uP=Z,tD=new it}function Vk(){Vk=Z,Ype=new ist}function SY(){SY=Z,ZAt=new ost}function Wy(){Wy=Z,zR=new no}function X2e(){hue.call(this,null)}function Zi(e,t){Kv(e),e.a.Nb(t)}function tgn(e,t){return e.a.Vc(t)}function ngn(e,t){return e.a.Wc(t)}function Lae(e,t){return e.a.Yc(t)}function Dae(e,t){return e.a.Zc(t)}function rgn(e,t){return e.Fc(t),e}function ign(e,t){return e.a.f=t,e}function Eot(e,t){return e.a.d=t,e}function Sot(e,t){return e.a.g=t,e}function _ot(e,t){return e.a.j=t,e}function Fh(e,t){return e.a.a=t,e}function Bh(e,t){return e.a.d=t,e}function Uh(e,t){return e.a.e=t,e}function zh(e,t){return e.a.g=t,e}function Oz(e,t){return e.a.f=t,e}function ogn(e){return e.b=!1,e}function J2e(){J2e=Z,h5e=new hv}function kot(){kot=Z,cIt=new Sn}function Z2e(){Z2e=Z,IOt=new rC}function Tot(){Tot=Z,OOt=new Od}function Q2e(){Q2e=Z,NOt=new Ilt}function xot(){xot=Z,XOt=new r8}function Cot(){Cot=Z,BOt=new hy}function h7(){h7=Z,zOt=new oC}function Nz(){Nz=Z,VOt=new Q_}function Rz(){Rz=Z,qOt=new ha}function _Y(){_Y=Z,c6t=new t1}function p7(){p7=Z,iNt=new N6}function kY(){kY=Z,wA=new yC}function TY(){TY=Z,Obe=new Iot}function xY(){xY=Z,Nbe=new Ict}function g7(){g7=Z,Zq=new Rdt}function Aot(){hbt(),this.c=new Cae}function Iot(){ur.call(this,fxt,0)}function agn(e,t,i,a){b9n(a,e,t,i)}function sgn(e,t,i,a){e5n(a,e,t,i)}function cgn(e,t,i,a){Z3n(e,a,t,i)}function ugn(e,t,i){Os(e.d,t.f,i)}function Wk(e,t){pE(e.c.b,t.c,t)}function m7(e,t){pE(e.c.c,t.b,t)}function b7(e,t){return e.a=t.g,e}function lgn(e){return e.e.a+e.f.a}function dgn(e){return e.e.b+e.f.b}function fgn(e){return(e.c+e.a)/2}function hgn(e){return e.b?e.b:e.a}function Oot(e,t){return rIn(e.a,t)}function fS(){fS=Z,MDt=new Oy}function Not(){Not=Z,GDt=new UU}function Rot(){Rot=Z,rMt=new xrt}function Pot(){Pot=Z,iMt=new Crt}function Uv(){Uv=Z,xh=new y2e}function eSe(){eSe=Z,j7e=new Ort}function CY(){CY=Z,Fne=new Wn}function tSe(){tSe=Z,K7e=new et}function v7(){v7=Z,v1e=new zst}function v0(){v0=Z,TV=new Hst}function Mae(){Mae=Z,vMt=new h9}function is(){is=Z,yMt=new p9}function zv(){zv=Z,Fw=new ZU}function pgn(e){h.clearTimeout(e)}function nSe(e){xce(),Tut.call(this,e)}function ggn(e,t){Mft.call(this,e,t)}function AY(e,t){this.d=e,this.e=t}function rSe(e,t){Plt(e,e.length,t)}function Pz(e,t){adt(e,e.length,t)}function Lot(e,t){return oce(e.a,t)}function jae(e,t){return Eu(e,t)>0}function $ae(e,t){return Eu(e,t)<0}function Dot(e,t){return Yu(e.e,t)}function w7(e,t){return e.a.get(t)}function mgn(e,t){return t.split(e)}function Fae(e){return e.b<e.d.gc()}function iSe(e){return an(e),!1}function IY(e){return f(e,43).jd()}function oSe(e){this.a=f(or(e),228)}function OY(e){Bt.call(this,e,21)}function NY(e,t){ur.call(this,e,t)}function Bae(e,t){ur.call(this,e,t)}function Mot(e,t){this.b=e,this.a=t}function jot(e,t){this.a=e,this.b=t}function $ot(e,t){this.a=e,this.b=t}function Fot(e,t){this.a=e,this.b=t}function Bot(e,t){this.a=e,this.b=t}function DN(e,t){this.a=e,this.b=t}function Uot(e,t){this.b=e,this.a=t}function aSe(e,t){this.b=e,this.a=t}function Uae(e,t){ur.call(this,e,t)}function zae(e,t){ur.call(this,e,t)}function Hae(e,t){ur.call(this,e,t)}function RY(e,t){ur.call(this,e,t)}function sSe(e,t){this.b=e,this.a=t}function zot(e,t){this.b=e,this.a=t}function cSe(e,t){ur.call(this,e,t)}function Gae(e,t){ur.call(this,e,t)}function Lz(e,t){ur.call(this,e,t)}function y7(e,t,i){e.splice(t,0,i)}function bgn(e,t,i){e.Mb(i)&&t.Ad(i)}function vgn(e,t,i){t.Ne(e.a.We(i))}function wgn(e,t,i){t.Bd(e.a.Xe(i))}function ygn(e,t,i){t.Ad(e.a.Kb(i))}function Egn(e,t){return Cl(e.c,t)}function Sgn(e,t){return Cl(e.e,t)}function PY(e,t){ur.call(this,e,t)}function LY(e,t){ur.call(this,e,t)}function MN(e,t){ur.call(this,e,t)}function uSe(e,t){ur.call(this,e,t)}function ho(e,t){ur.call(this,e,t)}function DY(e,t){ur.call(this,e,t)}function lSe(e,t){this.b=e,this.a=t}function lP(e,t){this.c=e,this.d=t}function Hot(e,t){this.e=e,this.d=t}function Got(e,t){this.a=e,this.b=t}function qot(e,t){this.a=e,this.b=t}function Vot(e,t){this.a=e,this.b=t}function Wot(e,t){this.a=e,this.b=t}function Kot(e,t){this.a=e,this.b=t}function Yot(e,t){this.a=e,this.b=t}function Xot(e,t){this.a=e,this.b=t}function Jot(e,t){this.a=e,this.b=t}function Zot(e,t){this.a=e,this.b=t}function Qot(e,t){this.b=e,this.a=t}function eat(e,t){this.b=e,this.a=t}function tat(e,t){this.b=e,this.a=t}function nat(e,t){this.b=t,this.c=e}function MY(e,t){ur.call(this,e,t)}function Dz(e,t){ur.call(this,e,t)}function qae(e,t){ur.call(this,e,t)}function E7(e,t){ur.call(this,e,t)}function dSe(e,t){ur.call(this,e,t)}function Vae(e,t){ur.call(this,e,t)}function Wae(e,t){ur.call(this,e,t)}function Mz(e,t){ur.call(this,e,t)}function fSe(e,t){ur.call(this,e,t)}function Kae(e,t){ur.call(this,e,t)}function S7(e,t){ur.call(this,e,t)}function hSe(e,t){ur.call(this,e,t)}function _7(e,t){ur.call(this,e,t)}function k7(e,t){ur.call(this,e,t)}function Kk(e,t){ur.call(this,e,t)}function Yae(e,t){ur.call(this,e,t)}function Xae(e,t){ur.call(this,e,t)}function pSe(e,t){ur.call(this,e,t)}function jz(e,t){ur.call(this,e,t)}function gSe(e,t){ur.call(this,e,t)}function w3(e,t){ur.call(this,e,t)}function Jae(e,t){ur.call(this,e,t)}function jY(e,t){ur.call(this,e,t)}function $z(e,t){ur.call(this,e,t)}function Zae(e,t){ur.call(this,e,t)}function Yk(e,t){ur.call(this,e,t)}function $Y(e,t){ur.call(this,e,t)}function mSe(e,t){ur.call(this,e,t)}function Qae(e,t){ur.call(this,e,t)}function ese(e,t){ur.call(this,e,t)}function tse(e,t){ur.call(this,e,t)}function nse(e,t){ur.call(this,e,t)}function rse(e,t){ur.call(this,e,t)}function ise(e,t){ur.call(this,e,t)}function bSe(e,t){ur.call(this,e,t)}function rat(e,t){this.b=e,this.a=t}function iat(e,t){this.a=e,this.b=t}function oat(e,t){this.a=e,this.b=t}function aat(e,t){this.a=e,this.b=t}function vSe(e,t){ur.call(this,e,t)}function wSe(e,t){ur.call(this,e,t)}function sat(e,t){this.a=e,this.b=t}function cat(e,t){this.b=e,this.d=t}function uat(){Vce(),this.b=new Mo}function lat(){iQ(),this.a=new Mo}function dat(){mX(),this.a=new P_e}function fat(){return oit(),new XAt}function Fz(e){return Ln(e.a),e.b}function ose(e){return T4n(e,e.c),e}function _gn(e,t){return vP(),t!=e}function kgn(e,t){return Cl(t.b,e)}function hat(e,t){I_n(e.a,f(t,11))}function ySe(e,t){ur.call(this,e,t)}function ESe(e,t){ur.call(this,e,t)}function FY(e,t){ur.call(this,e,t)}function hS(e,t){ur.call(this,e,t)}function ase(e,t){ur.call(this,e,t)}function BY(e,t){ur.call(this,e,t)}function SSe(e,t){ur.call(this,e,t)}function _Se(e,t){ur.call(this,e,t)}function Bz(e,t){ur.call(this,e,t)}function kSe(e,t){ur.call(this,e,t)}function sse(e,t){ur.call(this,e,t)}function UY(e,t){ur.call(this,e,t)}function cse(e,t){ur.call(this,e,t)}function use(e,t){ur.call(this,e,t)}function zY(e,t){ur.call(this,e,t)}function lse(e,t){ur.call(this,e,t)}function TSe(e,t){ur.call(this,e,t)}function dse(e,t){ur.call(this,e,t)}function pat(e,t){this.a=e,this.b=t}function gat(e,t){this.a=e,this.b=t}function mat(e,t){this.a=e,this.b=t}function xSe(e,t){this.a=e,this.b=t}function bat(e,t){this.b=e,this.a=t}function vat(e,t){this.b=e,this.a=t}function wat(e,t){this.b=e,this.a=t}function yat(e,t){this.b=e,this.a=t}function Eat(e,t){this.b=e,this.a=t}function Sat(e,t){this.a=e,this.b=t}function _at(e,t){this.a=e,this.b=t}function fse(e,t){ur.call(this,e,t)}function Uz(e,t){ur.call(this,e,t)}function hse(e,t){ur.call(this,e,t)}function CSe(e,t){ur.call(this,e,t)}function zz(e,t){ur.call(this,e,t)}function ASe(e,t){ur.call(this,e,t)}function Hz(e,t){ur.call(this,e,t)}function jN(e,t){ur.call(this,e,t)}function T7(e,t){ur.call(this,e,t)}function x7(e,t){ur.call(this,e,t)}function Gz(e,t){ur.call(this,e,t)}function HY(e,t){ur.call(this,e,t)}function pse(e,t){ur.call(this,e,t)}function GY(e,t){ur.call(this,e,t)}function C7(e,t){ur.call(this,e,t)}function gse(e,t){ur.call(this,e,t)}function qY(e,t){ur.call(this,e,t)}function y3(e,t){ur.call(this,e,t)}function qz(e,t){ur.call(this,e,t)}function A7(e,t){ur.call(this,e,t)}function I7(e,t){ur.call(this,e,t)}function Vz(e,t){ur.call(this,e,t)}function mse(e,t){ur.call(this,e,t)}function VY(e,t){ur.call(this,e,t)}function E3(e,t){ur.call(this,e,t)}function bse(e,t){ur.call(this,e,t)}function kat(e,t){ur.call(this,e,t)}function WY(e,t){ur.call(this,e,t)}function Tat(e,t){this.a=e,this.b=t}function xat(e,t){this.a=e,this.b=t}function Cat(e,t){this.a=e,this.b=t}function Ke(e,t){this.a=e,this.b=t}function Aat(e,t){this.a=e,this.b=t}function Iat(e,t){this.a=e,this.b=t}function Oat(e,t){this.a=e,this.b=t}function Ya(e,t){this.a=e,this.b=t}function Nat(e,t){this.a=e,this.b=t}function Rat(e,t){this.a=e,this.b=t}function Pat(e,t){this.a=e,this.b=t}function Lat(e,t){this.a=e,this.b=t}function Dat(e,t){this.a=e,this.b=t}function Mat(e,t){this.a=e,this.b=t}function jat(e,t){this.b=e,this.a=t}function $at(e,t){this.b=e,this.a=t}function Fat(e,t){this.b=e,this.a=t}function Bat(e,t){this.b=e,this.a=t}function Uat(e,t){this.a=e,this.b=t}function zat(e,t){this.a=e,this.b=t}function Hat(e,t){this.a=e,this.b=t}function Gat(e,t){this.a=e,this.b=t}function qat(e,t){this.f=e,this.c=t}function ISe(e,t){this.i=e,this.g=t}function KY(e,t){this.a=e,this.b=t}function Vat(e,t){this.a=e,this.b=t}function YY(e,t){ur.call(this,e,t)}function $N(e,t){ur.call(this,e,t)}function Wat(e,t){this.a=e,this.b=t}function Kat(e,t){this.a=e,this.b=t}function OSe(e,t){this.d=e,this.e=t}function Yat(e,t){this.d=e,this.b=t}function Xat(e,t){this.e=e,this.a=t}function Wz(e,t){this.b=e,this.c=t}function Jat(e,t){this.a=e,this.b=t}function NSe(e,t){e.i=null,qJ(e,t)}function Tgn(e,t){e&&qr(SV,e,t)}function Zat(e,t){return Ele(e.a,t)}function RSe(e,t){return Cl(e.g,t)}function xgn(e,t){return-e.b.$e(t)}function XY(e){return oG(e.c,e.b)}function ac(e){return e?e.kd():null}function Ae(e){return e??null}function Xk(e){return typeof e===xR}function Jk(e){return typeof e===AOe}function ta(e){return typeof e===sfe}function Hv(e,t){return e.Pd().Xb(t)}function JY(e,t){return eTn(e.Jc(),t)}function O7(e,t){return Eu(e,t)==0}function ZY(e,t){return Eu(e,t)>=0}function N7(e,t){return Eu(e,t)!=0}function Cgn(e){return""+(an(e),e)}function Agn(e,t){M_n(new Nn(e),t)}function Ign(e,t,i){jwt(t,mde(e,i))}function Ogn(e,t,i){jwt(t,mde(e,i))}function Qat(){bTe(),_ke.call(this)}function est(e){return vd(e),e.d.gc()}function QY(e){return z7(e==null),e}function PSe(e){return Gt(e,0),null}function R7(e,t){return e.a+=""+t,e}function ys(e,t){return e.a+=""+t,e}function P7(e,t){return e.a+=""+t,e}function vu(e,t){return e.a+=""+t,e}function xr(e,t){return e.a+=""+t,e}function LSe(e,t){return e.a+=""+t,e}function S3(e,t){Wi(e,t,e.a,e.a.a)}function Ma(e,t){Wi(e,t,e.c.b,e.c)}function tst(e,t){Oke.call(this,e,t)}function nst(e,t){Oke.call(this,e,t)}function eX(e,t){Oke.call(this,e,t)}function rst(e){nu(this),mj(this,e)}function ist(){this.b=0,this.a=!1}function ost(){this.b=0,this.a=!1}function ast(){this.b=new PN(mT(12))}function sst(e,t){e.q.setTime(oE(t))}function Pt(e,t){Mi.call(this,e,t)}function vse(e,t){rT.call(this,e,t)}function _3(e,t){ISe.call(this,e,t)}function cst(e,t){Zz.call(this,e,t)}function wse(e,t){GP.call(this,e,t)}function Dr(e,t){CY(),qr(Fne,e,t)}function ust(e,t){return Ae(e)===Ae(t)}function Ngn(e){return or(e),new L7(e)}function lst(e,t){return e.a.a.a.cc(t)}function Rgn(e,t){return e.j[t.p]==2}function pS(e,t){return e==t?0:e?1:-1}function tX(e){return e<10?"0"+e:""+e}function dst(e){return Ku(e.l,e.m,e.h)}function Js(e){return typeof e===AOe}function yse(e,t){return uh(e.a,0,t)}function FN(e){return Aa((an(e),e))}function Pgn(e){return Aa((an(e),e))}function Lgn(e,t){return di(e.a,t.a)}function DSe(e,t){return tu(e.a,t.a)}function Dgn(e,t){return tdt(e.a,t.a)}function cm(e,t){return e.indexOf(t)}function MSe(e,t){LP(e,0,e.length,t)}function fst(e,t){return e.a=t.g+1,e}function Cp(e){return e.a=0,e.b=0,e}function hst(e){this.a=fat(),this.b=e}function pst(e){this.a=fat(),this.b=e}function L7(e){this.a=e,sN.call(this)}function gst(e){this.a=e,sN.call(this)}function mst(){mst=Z,_5t=hr(Jbt())}function bst(){bst=Z,UIt=hr(Ole())}function vst(){vst=Z,YOt=hr(O2t())}function gg(){gg=Z,h.Math.log(2)}function jSe(){jSe=Z,c2e(),m6e=new Wn}function ub(){JK.call(this,new Zv)}function wst(){IX.call(this,0,0,0,0)}function $Se(e){return Oc(e.b.b,e,0)}function yst(e,t){return tu(e.g,t.g)}function Mgn(e,t){return tu(e.g,t.g)}function jgn(e,t){return di(t.f,e.f)}function nX(e){return lf(new _o,e)}function Est(e){return e==PA||e==RA}function $gn(e){return e==a_||e==YT}function Fgn(e){return e==a_||e==KT}function l1(e){return e==Ds||e==Gs}function k3(e){return e==Wf||e==jg}function T3(e){return e!=Fg&&e!=Pw}function Ld(e){return e.rh()&&e.sh()}function Sst(e){return Cce(f(e,124))}function x3(e,t,i){RS(e,t),NS(e,i)}function Of(e,t,i){md(e,t),bd(e,i)}function Ese(e,t,i){MJ(e,t),DJ(e,i)}function Kz(e,t,i){U3(e,t),z3(e,i)}function Yz(e,t,i){B3(e,t),H3(e,i)}function Sse(e,t){YP(e,t),jP(e,e.D)}function Bgn(e,t){return Qe(e.a,t),t}function Ugn(e,t){return Qe(e.c,t),e}function FSe(e,t){return lf(e.a,t),e}function zgn(e,t){return tf(),t.a+=e}function Hgn(e,t){return tf(),t.a+=e}function Ggn(e,t){return tf(),t.c+=e}function _st(e,t){return new GP(t,e)}function qgn(e,t){return new GP(t,e)}function BSe(e){return ivn(e.b.Jc(),e.a)}function C3(e){return e!=null?Ti(e):0}function Vgn(e,t){return e[e.length]=t}function Wgn(e,t){return e[e.length]=t}function Kgn(e,t){Lg||(e.b=t)}function ja(e){this.a=e.a,this.b=e.b}function _se(e){qat.call(this,e,!0)}function gS(e){IX.call(this,e,e,e,e)}function kst(){NY.call(this,"Head",1)}function Tst(){NY.call(this,"Tail",3)}function BN(){qh.call(this,0,0,0,0)}function xst(e,t,i){S_e.call(this,e,t,i)}function Gv(e){ym(),rTn.call(this,e)}function Cst(e){Mu(e.Qf(),new zo(e))}function Ast(){Ast=Z,zme=new Aj(i1e)}function kse(e){e.c=Pe(Ro,Jt,1,0,5,1)}function Ist(e,t,i){io(e.c[t.g],t.g,i)}function Ygn(e,t,i){f(e.c,71).Di(t,i)}function Xgn(e,t,i){Of(i,i.i+e,i.j+t)}function pa(e,t){Mi.call(this,e.b,t)}function Jgn(e,t){Qn(iu(e.a),Yft(t))}function Zgn(e,t){Qn(gd(e.a),Xft(t))}function Qgn(e,t){return bT(t,bg(e))}function emn(e,t){return bT(t,bg(e))}function tmn(e,t){return BJ(Sce(e.f),t)}function nmn(e,t){return BJ(Sce(e.n),t)}function rmn(e,t){return BJ(Sce(e.p),t)}function imn(e){return e==null?0:Ti(e)}function D7(e){Xr(),sS.call(this,e)}function Ost(e){this.a=e,tke.call(this,e)}function Tse(e){this.a=e,wY.call(this,e)}function xse(e){this.a=e,wY.call(this,e)}function Nst(e,t){_S(),this.a=e,this.b=t}function omn(e,t){or(t),P3(e).Ic(new O)}function Cse(e,t,i){return io(e,t,i),i}function Ku(e,t,i){return{l:e,m:t,h:i}}function tu(e,t){return e<t?-1:e>t?1:0}function USe(e,t){return Eu(e,t)>0?e:t}function sc(e){return e.a<e.c.c.length}function zSe(e){return e.a<e.c.a.length}function na(e,t){ace(e.c,e.c.length,t)}function amn(e,t){e.a!=null&&hat(t,e.a)}function Rst(e,t){return e.a?e.b:t.Ue()}function smn(){throw ie(new f0(EAt))}function cmn(){throw ie(new f0(SAt))}function umn(){throw ie(new f0(SAt))}function lmn(){throw ie(new f0(EAt))}function Pst(){new Wn,new Wn,new Wn}function lr(){lr=Z,new Pst,new et}function Nf(){Nf=Z,Mb=(Not(),GDt)}function rX(e){e.j=Pe(I6e,Ye,320,0,0,1)}function Lst(){NY.call(this,"Range",2)}function Dst(e){uTe.call(this),this.b=e}function HSe(e){R_e.call(this),this.a=e}function Mst(e){this.b=new e0,this.a=e}function iX(e){this.b=e,this.a=new et}function jst(){H3e(),this.a=new Gj(f5e)}function $st(e){e.a=new An,e.c=new An}function Fst(e){e.a=new Wn,e.d=new Wn}function dmn(e,t){return qr(e.a,t.a,t)}function fmn(e,t,i){return qr(e.g,i,t)}function hmn(e,t,i){return qr(e.k,i,t)}function A3(e,t,i){return l3e(t,i,e.c)}function GSe(e,t){return Oe(bn(e.i,t))}function qSe(e,t){return Oe(bn(e.j,t))}function VSe(e){return new Ke(e.c,e.d)}function pmn(e){return new Ke(e.c,e.d)}function $a(e){return new Ke(e.a,e.b)}function Bst(e,t){return D8n(e.a,t,null)}function gmn(e){Ia(e,null),ua(e,null)}function Ust(e){Gce(e,null),qce(e,null)}function zst(){Zz.call(this,null,null)}function Hst(){lX.call(this,null,null)}function WSe(e){this.a=e,Wn.call(this)}function mmn(e){this.b=(Ht(),new vz(e))}function oX(e){return e.Ob()?e.Pb():null}function bmn(e,t){return new uct(e.Jc(),t)}function Ase(e,t){return DTn(e.Jc(),t)!=-1}function M7(e,t){return KMn(e.c,e.b,t)}function me(e,t){return e!=null&&Lle(e,t)}function vmn(e,t){e.q.setHours(t),Yj(e,t)}function wmn(e,t){return rf(),t.n.b+=e}function ymn(e,t,i){return P8n(e,t,i),i}function Qe(e,t){return vn(e.c,t),!0}function Gst(e,t){e.c&&(hke(t),yft(t))}function Emn(e,t,i){e.c._c(t,f(i,135))}function Smn(e,t,i){e.c.Ri(t,f(i,135))}function qst(e,t){Jn(e),e.Fc(f(t,15))}function KSe(e,t){return e.a.Ac(t)!=null}function Ise(e,t){return e.a.Ac(t)!=null}function Ose(e,t,i){return e.a[t.g][i.g]}function _mn(e,t){return e.e[t.c.p][t.p]}function kmn(e,t){return e.c[t.c.p][t.p]}function mg(e,t){return e.a[t.c.p][t.p]}function Tmn(e,t){return e.j[t.p]=N6n(t)}function xmn(e,t){return ke(Oe(t.a))<=e}function Cmn(e,t){return ke(Oe(t.a))>=e}function Amn(e,t){return YTe(e.f,t.Ng())}function Vst(e){return bm(e,0,e.length)}function UN(e,t){return e.a*t.a+e.b*t.b}function Imn(e,t){return e.a<zse(t)?-1:1}function Omn(e,t){return YTe(e.b,t.Ng())}function Nmn(e,t,i){return i?t!=0:t!=e-1}function Wst(e,t,i){e.a=t^1502,e.b=i^$fe}function Rmn(e,t,i){return e.a=t,e.b=i,e}function qv(e,t){return e.a*=t,e.b*=t,e}function j7(e,t,i){return io(e.g,t,i),i}function Pmn(e,t,i,a){io(e.a[t.g],i.g,a)}function So(e,t,i){lH.call(this,e,t,i)}function aX(e,t,i){So.call(this,e,t,i)}function YSe(e,t,i){yX.call(this,e,t,i)}function Kst(e,t,i){yX.call(this,e,t,i)}function Yst(e,t,i){YSe.call(this,e,t,i)}function xl(e,t,i){So.call(this,e,t,i)}function Xst(e,t,i){aX.call(this,e,t,i)}function XSe(e,t,i){lH.call(this,e,t,i)}function I3(e,t,i){lH.call(this,e,t,i)}function Jst(e,t,i){XSe.call(this,e,t,i)}function Zst(){ur.call(this,"INSTANCE",0)}function O3(){this.a=Pe(Ro,Jt,1,8,5,1)}function JSe(){this.Bb|=256,this.Bb|=512}function Nn(e){this.i=e,this.f=this.i.j}function Vv(e){this.c=e,this.a=this.c.a}function N3(e,t){this.a=e,wY.call(this,t)}function ZSe(e,t){return fCn(e,new $v,t).a}function QSe(e){if(!e)throw ie(new Sz)}function e_e(e){if(!e)throw ie(new Tl)}function t_e(){t_e=Z,t_e(),QAt=new Nt}function Qst(){Qst=Z,Mae(),wMt=new vc}function dP(){dP=Z,F6e=new yae(null)}function Lmn(e){bH(e,KCt),QZ(e,O9n(e))}function ect(e){e.a=f(fn(e.b.a,4),128)}function tct(e){e.a=f(fn(e.b.a,4),128)}function nct(e){e.b.Qb(),--e.d.f.d,PX(e.d)}function n_e(e){this.a=e,W9.call(this,e.d)}function rct(e,t){this.a=e,xae.call(this,t)}function ict(e,t){this.a=e,xae.call(this,t)}function oct(e,t){this.a=e,xae.call(this,t)}function r_e(e,t){this.a=t,xae.call(this,e)}function act(e,t){this.a=t,rue.call(this,e)}function sct(e,t){this.a=e,rue.call(this,t)}function cct(e,t){this.a=t,yY.call(this,e)}function uct(e,t){this.a=t,yY.call(this,e)}function Cn(e,t){return or(t),new cct(e,t)}function lct(e,t){return new Byt(e.a,e.b,t)}function sX(e,t){return e.lastIndexOf(t)}function i_e(e,t,i){return e.indexOf(t,i)}function $7(e){return e==null?bl:Qs(e)}function Dmn(e){return e==null?null:e.name}function Mmn(e){return e.l+e.m*OR+e.h*yE}function jmn(e){return kz(e.a)?Jft(e):null}function dct(e){iS.call(this,f(or(e),34))}function fct(e){iS.call(this,f(or(e),34))}function sh(e){bi.call(this,(an(e),e))}function ef(e){bi.call(this,(an(e),e))}function Nse(e){JK.call(this,new oxe(e))}function cX(e){h3.call(this,e),this.a=e}function o_e(e){p3.call(this,e),this.a=e}function a_e(e){tP.call(this,e),this.a=e}function hct(){rX(this),YX(this),this.he()}function pct(e){this.a=e,bz.call(this,e)}function Dd(e){return Ln(e.a!=null),e.a}function gct(e,t){return Qe(t.a,e.a),e.a}function mct(e,t){return Qe(t.b,e.a),e.a}function mS(e,t){return Qe(t.a,e.a),e.a}function Xz(e,t,i){return Hue(e,t,t,i),e}function uX(e,t){return++e.b,Qe(e.a,t)}function s_e(e,t){return++e.b,hl(e.a,t)}function $mn(e,t){return di(e.c.d,t.c.d)}function Fmn(e,t){return di(e.c.c,t.c.c)}function Bmn(e,t){return di(e.n.a,t.n.a)}function dl(e,t){return f(si(e.b,t),15)}function Umn(e,t){return e.n.b=(an(t),t)}function zmn(e,t){return e.n.b=(an(t),t)}function Cl(e,t){return!!t&&e.b[t.g]==t}function F7(e){return sc(e.a)||sc(e.b)}function bS(e){return e.$H||(e.$H=++Q7n)}function zN(e){return e.a!=null?e.a:null}function Hmn(e,t){return di(e.e.b,t.e.b)}function Gmn(e,t){return di(e.e.a,t.e.a)}function qmn(e,t,i){return Kht(e,t,i,e.b)}function c_e(e,t,i){return Kht(e,t,i,e.c)}function u_e(e,t,i){f(DH(e,t),22).Ec(i)}function l_e(e,t){v3e.call(this,e,t,null)}function Rse(e,t){rce(),this.f=t,this.d=e}function bct(){p7(),this.b=new Oet(this)}function Vmn(e){return tf(),!!e&&!e.dc()}function Jz(e){me(e,157)&&f(e,157).li()}function Wmn(e,t,i){hle(e.a,i),OZ(e.a,t)}function Zz(e,t){v7(),this.a=e,this.b=t}function lX(e,t){v0(),this.b=e,this.c=t}function d_e(e,t){XTe(t,e),this.d=e,this.c=t}function w0(e){var t;t=e.a,e.a=e.b,e.b=t}function mn(){mn=Z,yw=!1,QL=!0}function dX(){dX=Z,Mee=new Mi($Tt,0)}function Kmn(e){return new hT(3,e)}function Ymn(e){return dY(),fr((Vft(),NAt),e)}function Pse(e,t){return new Sut(e,e.gc(),t)}function Xmn(e,t){return uce(e.c).Kd().Xb(t)}function fP(e,t,i){return i_e(e,gl(t),i)}function vct(e,t,i,a){Ake.call(this,e,t,i,a)}function HN(e){this.d=e,Nn.call(this,e)}function GN(e){this.c=e,Nn.call(this,e)}function Qz(e){this.c=e,HN.call(this,e)}function wct(e){nke.call(this,e,null,null)}function yct(e){if(!e)throw ie(new If)}function lb(e){return uf(e,$T),new Du(e)}function Hh(e){return Ln(e.b!=0),e.c.b.c}function hP(e){return Ln(e.b!=0),e.a.a.c}function Jmn(e,t){return Hue(e,t,t+1,""),e}function pP(e,t,i){var a;a=e.dd(t),a.Rb(i)}function Ect(e){return e.b=f(iTe(e.a),43)}function Zmn(e){return e.f!=null?e.f:""+e.g}function Lse(e){return e.f!=null?e.f:""+e.g}function Qmn(e,t){return e>0?t/(e*e):t*100}function ebn(e,t){return e>0?t*t/e:t*t*100}function li(e,t){return!!e.q&&Yu(e.q,t)}function Zk(e,t){return f(Eg(e.a,t),34)}function Sct(e){return AP(),parseInt(e)||-1}function tbn(e,t){return cw(),jt(e,t.e,t)}function nbn(e,t,i){return _Y(),i.Kg(e,t)}function rbn(e){return y1(),e.e.a+e.f.a/2}function ibn(e,t,i){return y1(),i.e.a-e*t}function obn(e){return y1(),e.e.b+e.f.b/2}function abn(e,t,i){return y1(),i.e.b-e*t}function f_e(e){if(!e)throw ie(new Sz)}function Qk(e){if(!e)throw ie(new Tl)}function sbn(e){if(!e)throw ie(new dae)}function Ln(e){if(!e)throw ie(new ic)}function _ct(e){e.d=new wct(e),e.e=new Wn}function kct(){this.a=new CS,this.b=new CS}function Tct(e){this.c=e,this.a=1,this.b=1}function Dse(e){Vk(),this.b=e,this.a=!0}function xct(e){SY(),this.b=e,this.a=!0}function Cct(e){ofe(),Gnt(this),this.Df(e)}function Act(e){ki.call(this),dj(this,e)}function Ict(){ur.call(this,"GROW_TREE",0)}function Oct(e){this.c=e,md(e,0),bd(e,0)}function fX(e){return e.a=-e.a,e.b=-e.b,e}function h_e(e,t){return e.a=t.a,e.b=t.b,e}function eT(e,t,i){return e.a+=t,e.b+=i,e}function Nct(e,t,i){return e.a-=t,e.b-=i,e}function cbn(e,t,i){_J(),e.nf(t)&&i.Ad(e)}function ubn(e,t,i){return Qe(t,j0t(e,i))}function lbn(e,t){return e.c.Ec(f(t,135))}function dbn(e,t){return f(bn(e.e,t),27)}function fbn(e,t){return f(bn(e.e,t),27)}function Md(e,t,i){Ol.call(this,e,t,i,2)}function Rct(e,t){v0(),p_e.call(this,e,t)}function p_e(e,t){v0(),lX.call(this,e,t)}function Pct(e,t){v0(),lX.call(this,e,t)}function Lct(e,t){v7(),Zz.call(this,e,t)}function Mse(e,t){Nf(),NX.call(this,e,t)}function g_e(e,t){Nf(),NX.call(this,e,t)}function Dct(e,t){Nf(),NX.call(this,e,t)}function Mct(e,t){Nf(),Mse.call(this,e,t)}function m_e(e,t){Nf(),Mse.call(this,e,t)}function jct(e,t){Nf(),m_e.call(this,e,t)}function $ct(e,t){Nf(),g_e.call(this,e,t)}function hbn(e,t,i){xj(iu(e.a),t,Yft(i))}function pbn(e,t,i){xj(gd(e.a),t,Xft(i))}function b_e(e,t,i){return pQ(LH(e,t),i)}function gbn(e,t,i){return t.wl(e.e,e.c,i)}function mbn(e,t,i){return t.xl(e.e,e.c,i)}function jse(e,t){return iw(e.e,f(t,52))}function Fct(e,t){return an(e),e+zse(t)}function bbn(e){return e==null?null:Qs(e)}function vbn(e){return e==null?null:Qs(e)}function wbn(e){return e==null?null:y9n(e)}function ybn(e){return e==null?null:zNn(e)}function d1(e){e.o==null&&a6n(e)}function ut(e){return z7(e==null||Xk(e)),e}function Oe(e){return z7(e==null||Jk(e)),e}function dr(e){return z7(e==null||ta(e)),e}function v_e(e){this.q=new h.Date(oE(e))}function eH(e,t){this.c=e,aP.call(this,e,t)}function B7(e,t){this.a=e,eH.call(this,e,t)}function Ebn(e,t){this.d=e,Wt(this),this.b=t}function w_e(e,t){hue.call(this,e),this.a=t}function y_e(e,t){hue.call(this,e),this.a=t}function E_e(){dbt.call(this),this.Bb|=Xa}function S_e(e,t,i){EJ.call(this,e,t,i,null)}function Bct(e,t,i){EJ.call(this,e,t,i,null)}function Sbn(e,t,i){return e.Le(t,i)<=0?i:t}function _bn(e,t,i){return e.Le(t,i)<=0?t:i}function Uct(e){return an(e),e?1231:1237}function $se(e){return f(ct(e.a,e.b),292)}function kbn(e,t){return f(Eg(e.b,t),149)}function Tbn(e,t){return f(Eg(e.c,t),232)}function zct(e){return new Ke(e.c,e.d+e.a)}function Hct(e){return rf(),Est(f(e,203))}function tT(){tT=Z,V6e=At((Ud(),GE))}function R3(){R3=Z,MA=new Si("root")}function gP(){gP=Z,kV=new _rt,new krt}function __e(e,t,i){return i=hf(e,t,3,i),i}function k_e(e,t,i){return i=hf(e,t,6,i),i}function T_e(e,t,i){return i=hf(e,t,9,i),i}function um(e,t){return bH(t,WOe),e.f=t,e}function xbn(e,t){return Rz(),UP(t.d.i,e)}function Cbn(e,t){return dR(),new kSt(t,e)}function Gct(e,t){Lg||Qe(e.a,t)}function Abn(e,t){t.a?A4n(e,t):Ise(e.a,t.b)}function tH(e,t,i){++e.j,e.qj(),fue(e,t,i)}function qct(e,t,i){++e.j,e.nj(t,e.Wi(t,i))}function Vct(e,t,i){var a;a=e.dd(t),a.Rb(i)}function Wct(e,t,i){return sOe(e.c,e.b,t,i)}function x_e(e,t){return f(bn(e.b,t),275)}function C_e(e,t){return(t&Qr)%e.d.length}function A_e(e,t){n2e.call(this,e),this.a=t}function Fse(e,t){n2e.call(this,e),this.a=t}function Mi(e,t){Si.call(this,e),this.a=t}function Kct(e,t){this.c=e,OS.call(this,t)}function Yct(e,t){this.a=e,Fnt.call(this,t)}function nH(e,t){this.a=e,Fnt.call(this,t)}function Xct(e){this.a=(uf(e,$T),new Du(e))}function Jct(e){this.a=(uf(e,$T),new Du(e))}function Ibn(e){return mr(Jy(e,32))^mr(e)}function Bse(e){return String.fromCharCode(e)}function Obn(e){return e==null?null:e.message}function Nbn(e,t,i){return e.apply(t,i)}function I_e(e,t,i){return ZN(e,f(t,23),i)}function Zct(e,t,i){return e.a+=bm(t,0,i),e}function O_e(e,t){var i;return i=e.e,e.e=t,i}function N_e(e,t){var i;return i=t,!!e.De(i)}function Rbn(e,t){var i;i=e[jfe],i.call(e,t)}function Pbn(e,t){var i;i=e[jfe],i.call(e,t)}function nT(e,t){e.a._c(e.b,t),++e.b,e.c=-1}function Qct(e){nu(e.e),e.d.b=e.d,e.d.a=e.d}function rH(e){e.b?rH(e.b):e.f.c.yc(e.e,e.d)}function iH(e){return!e.a&&(e.a=new he),e.a}function eut(){this.a=new et,this.b=new et}function tut(){this.a=new rh,this.b=new ert}function nut(){this.b=new ha,this.c=new et}function R_e(){this.n=new ha,this.o=new ha}function hX(){this.n=new xN,this.i=new BN}function P_e(){this.b=new Zv,this.a=new Zv}function rut(){this.a=new L5,this.b=new M2}function iut(){this.b=new Mo,this.a=new Mo}function out(){this.a=new et,this.d=new et}function aut(){this.b=new Mit,this.a=new P8}function sut(){this.b=new Wn,this.a=new Wn}function cut(){hX.call(this),this.a=new ha}function L_e(e,t,i,a){IX.call(this,e,t,i,a)}function Lbn(e,t,i){Wy(),q5(e,t.Te(e.a,i))}function Dbn(e,t){return Rz(),!UP(t.d.i,e)}function Mbn(e,t){return mn(),e==t?0:e?1:-1}function jbn(e,t){return e.n.a=(an(t),t+10)}function $bn(e,t){return e.n.a=(an(t),t+10)}function Gh(e,t){return pY(new Array(t),e)}function Fbn(e,t){return t==e||lL(ZZ(t),e)}function Bbn(e){return e<100?null:new Fv(e)}function uut(e,t){return qr(e.a,t,"")==null}function Ubn(e,t){var i;return i=t.mi(e.a),i}function ii(e,t){return e.a+=t.a,e.b+=t.b,e}function Ho(e,t){return e.a-=t.a,e.b-=t.b,e}function D_e(e,t,i){return i=hf(e,t,11,i),i}function zbn(e){return Bk(e.j.c,0),e.a=-1,e}function Hbn(e,t){l1(e.f)?Q4n(e,t):HOn(e,t)}function lut(e,t,i){i!=null&&zJ(t,qle(e,i))}function dut(e,t,i){i!=null&&HJ(t,qle(e,i))}function qN(e,t,i,a){Be.call(this,e,t,i,a)}function M_e(e,t,i,a){Be.call(this,e,t,i,a)}function fut(e,t,i,a){M_e.call(this,e,t,i,a)}function hut(e,t,i,a){FX.call(this,e,t,i,a)}function Use(e,t,i,a){FX.call(this,e,t,i,a)}function j_e(e,t,i,a){FX.call(this,e,t,i,a)}function put(e,t,i,a){Use.call(this,e,t,i,a)}function $_e(e,t,i,a){Use.call(this,e,t,i,a)}function Qt(e,t,i,a){j_e.call(this,e,t,i,a)}function gut(e,t,i,a){$_e.call(this,e,t,i,a)}function mut(e,t,i,a){Rke.call(this,e,t,i,a)}function rT(e,t){Ru.call(this,T$+e+TE+t)}function but(e,t){cvn.call(this,e,e.length,t)}function F_e(e,t){return e.gk().si().ni(e,t)}function B_e(e,t){return e.gk().si().pi(e,t)}function Gbn(e,t){return e.b.zd(new $ot(e,t))}function qbn(e,t){return e.b.zd(new Fot(e,t))}function vut(e,t){return e.b.zd(new Bot(e,t))}function Vbn(e,t){return e.e=f(e.d.Kb(t),158)}function U_e(e,t,i){return e.lastIndexOf(t,i)}function Wbn(e,t,i){return di(e[t.a],e[i.a])}function Kbn(e,t){return tu(e.a.d.p,t.a.d.p)}function Ybn(e,t){return tu(t.a.d.p,e.a.d.p)}function Xbn(e,t){return $e(t,(ot(),Dq),e)}function wut(e,t){return an(e),Ae(e)===Ae(t)}function Mt(e,t){return an(e),Ae(e)===Ae(t)}function Jbn(e,t){return di(e.b.e.a,t.b.e.a)}function Zbn(e,t){return di(e.c.e.a,t.c.e.a)}function Qbn(e,t){return di(e.c-e.s,t.c-t.s)}function yut(e){return e.c?Oc(e.c.a,e,0):-1}function Eut(e,t){return me(t,15)&&oSt(e.c,t)}function e1n(e,t){Lg||t&&(e.d=t)}function z_e(e,t){this.c=e,bce.call(this,e,t)}function Sut(e,t,i){this.a=e,d_e.call(this,t,i)}function _ut(e){this.c=e,eX.call(this,FG,0)}function kut(e,t,i){this.c=t,this.b=i,this.a=e}function oH(e){vP(),this.d=e,this.a=new O3}function Tut(e){lm(),this.a=(Ht(),new tP(e))}function xut(){xut=Z,YDt=Pe(Ro,Jt,1,0,5,1)}function Cut(){Cut=Z,hMt=Pe(Ro,Jt,1,0,5,1)}function H_e(){H_e=Z,pMt=Pe(Ro,Jt,1,0,5,1)}function lm(){lm=Z,new o2e((Ht(),Ht(),Ga))}function mP(e){if(!e)throw ie(new As(IOe))}function G_e(e,t){if(!e)throw ie(new dn(t))}function t1n(e,t,i){return f(e.c,71).Tk(t,i)}function pX(e,t,i){return f(e.c,71).Uk(t,i)}function n1n(e,t,i){return gbn(e,f(t,339),i)}function q_e(e,t,i){return mbn(e,f(t,339),i)}function r1n(e,t,i){return Wwt(e,f(t,339),i)}function Aut(e,t,i){return nNn(e,f(t,339),i)}function U7(e,t){return t==null?null:ST(e.b,t)}function VN(e){return e==HE||e==Lb||e==uu}function zse(e){return Jk(e)?(an(e),e):e.se()}function gX(e){return!isNaN(e)&&!isFinite(e)}function Hse(e){$st(this),nf(this),Ba(this,e)}function Kl(e){kse(this),rke(this.c,0,e.Nc())}function Iut(e){hd(e.a),nxe(e.c,e.b),e.b=null}function Out(e){this.a=e,b0(),Nc(Date.now())}function Rf(e,t,i){this.a=e,this.b=t,this.c=i}function Nut(e,t,i){this.a=e,this.b=t,this.c=i}function Rut(e,t,i){this.d=e,this.b=i,this.a=t}function i1n(e,t,i){return Wy(),e.a.Wd(t,i),t}function Gse(e){var t;return t=new nh,t.e=e,t}function V_e(e){var t;return t=new srt,t.b=e,t}function qse(){qse=Z,$6e=new Yt,JAt=new Or}function Vse(e,t){if(e!=t)throw ie(new If)}function o1n(e){return Np(),fr((wgt(),pIt),e)}function a1n(e){return Yl(),fr((ygt(),mIt),e)}function s1n(e){return DP(),fr((Rgt(),eIt),e)}function c1n(e){return fh(),fr((vgt(),aIt),e)}function u1n(e){return pl(),fr((Egt(),vIt),e)}function l1n(e){return wQ(),fr((bst(),UIt),e)}function d1n(e){return DS(),fr((Mgt(),HIt),e)}function f1n(e){return qP(),fr((jgt(),POt),e)}function h1n(e){return OJ(),fr((spt(),ZIt),e)}function p1n(e){return cj(),fr((Sgt(),kOt),e)}function g1n(e){return wa(),fr((vmt(),AOt),e)}function m1n(e){return hE(),fr((Lgt(),FOt),e)}function b1n(e){return cn(),fr((Wbt(),HOt),e)}function v1n(e){return BP(),fr((cpt(),KOt),e)}function w1n(e){return ka(),fr((vst(),YOt),e)}function Wse(e){IX.call(this,e.d,e.c,e.a,e.b)}function W_e(e){IX.call(this,e.d,e.c,e.a,e.b)}function y1n(e){return nZ(),fr((Dgt(),ZOt),e)}function aH(){aH=Z,zee=new w6,Hee=new dd}function mX(){mX=Z,eNt=new AU,QOt=new a8}function tf(){tf=Z,oNt=new eb,aNt=new ck}function bX(e,t,i){this.e=t,this.b=e,this.d=i}function Put(e,t,i){this.a=e,this.b=t,this.c=i}function Lut(e,t,i){this.a=e,this.b=t,this.c=i}function K_e(e,t,i){this.a=e,this.b=t,this.c=i}function vX(e,t,i){this.b=e,this.a=t,this.c=i}function Dut(e,t,i){this.b=e,this.a=t,this.c=i}function Y_e(e,t,i){this.b=e,this.c=t,this.a=i}function E1n(e){return CT(),fr((ymt(),dNt),e)}function S1n(e){return LS(),fr((Xpt(),mNt),e)}function _1n(e){return _g(),fr(($gt(),uNt),e)}function k1n(e){return aQ(),fr((Kbt(),bNt),e)}function T1n(e){return ER(),fr((Ymt(),vNt),e)}function x1n(e){return SJ(),fr((upt(),wNt),e)}function C1n(e){return hj(),fr((Jpt(),yNt),e)}function A1n(e){return FJ(),fr((Zpt(),ENt),e)}function I1n(e){return R0(),fr((wmt(),SNt),e)}function O1n(e){return FH(),fr((lpt(),_Nt),e)}function N1n(e){return lE(),fr((Qpt(),TNt),e)}function R1n(e){return qZ(),fr((Vmt(),xNt),e)}function P1n(e){return RH(),fr((fpt(),CNt),e)}function L1n(e){return gG(),fr((Kmt(),ANt),e)}function D1n(e){return fL(),fr((Wmt(),INt),e)}function M1n(e){return us(),fr((E1t(),ONt),e)}function j1n(e){return zP(),fr((lgt(),NNt),e)}function $1n(e){return w1(),fr((ugt(),PNt),e)}function F1n(e){return lJ(),fr((dpt(),LNt),e)}function B1n(e){return Ff(),fr((Smt(),DNt),e)}function U1n(e){return pJ(),fr((hpt(),MNt),e)}function z1n(e){return D0(),fr((o1t(),l6t),e)}function H1n(e){return Sj(),fr((rgt(),d6t),e)}function G1n(e){return xT(),fr((Fgt(),f6t),e)}function q1n(e){return Nj(),fr((Emt(),h6t),e)}function V1n(e){return gh(),fr((S1t(),p6t),e)}function W1n(e){return Pp(),fr((Bgt(),g6t),e)}function K1n(e){return MH(),fr((ppt(),m6t),e)}function Y1n(e){return cs(),fr((egt(),v6t),e)}function X1n(e){return QJ(),fr((tgt(),w6t),e)}function J1n(e){return Ej(),fr((ngt(),y6t),e)}function Z1n(e){return JP(),fr((igt(),E6t),e)}function Q1n(e){return $J(),fr((ogt(),S6t),e)}function e0n(e){return eZ(),fr((agt(),_6t),e)}function t0n(e){return PS(),fr((mgt(),F6t),e)}function n0n(e){return oj(),fr((gpt(),G6t),e)}function r0n(e){return fm(),fr((bpt(),J6t),e)}function i0n(e){return vg(),fr((mpt(),Q6t),e)}function o0n(e){return Ip(),fr((vpt(),p5t),e)}function a0n(e){return AS(),fr((wpt(),E5t),e)}function s0n(e){return yR(),fr((Vgt(),S5t),e)}function c0n(e){return Wj(),fr((mst(),_5t),e)}function u0n(e){return yj(),fr((sgt(),k5t),e)}function l0n(e){return wj(),fr((qgt(),W5t),e)}function d0n(e){return uJ(),fr((ypt(),K5t),e)}function f0n(e){return GJ(),fr((Ept(),Q5t),e)}function h0n(e){return BZ(),fr((_mt(),tPt),e)}function p0n(e){return ZH(),fr((cgt(),rPt),e)}function g0n(e){return kJ(),fr((Spt(),nPt),e)}function m0n(e){return RZ(),fr((Ggt(),_Pt),e)}function b0n(e){return JJ(),fr((dgt(),kPt),e)}function v0n(e){return wZ(),fr((fgt(),TPt),e)}function w0n(e){return zZ(),fr((Wgt(),CPt),e)}function y0n(e){return gZ(),fr((bgt(),OPt),e)}function E0n(e){return gJ(),fr((Tpt(),ZPt),e)}function S0n(e){return TJ(),fr((pgt(),QPt),e)}function _0n(e){return lZ(),fr((ggt(),eLt),e)}function k0n(e){return xY(),fr((Uht(),aLt),e)}function T0n(e){return rG(),fr((hgt(),rLt),e)}function x0n(e){return TY(),fr((zht(),iLt),e)}function C0n(e){return Hj(),fr((kmt(),nLt),e)}function A0n(e){return CH(),fr((_pt(),sLt),e)}function I0n(e){return bG(),fr((Tmt(),cLt),e)}function O0n(e){return g7(),fr((Bht(),kLt),e)}function N0n(e){return XH(),fr((kpt(),TLt),e)}function R0n(e){return wm(),fr((Imt(),NLt),e)}function P0n(e){return bE(),fr((Ybt(),PLt),e)}function L0n(e){return vb(),fr((Xmt(),LLt),e)}function D0n(e){return mE(),fr((Jmt(),BLt),e)}function M0n(e){return ko(),fr((xmt(),uDt),e)}function j0n(e){return XP(),fr((Ygt(),lDt),e)}function $0n(e){return Sg(),fr((kgt(),dDt),e)}function F0n(e){return k1(),fr((Xgt(),fDt),e)}function B0n(e){return HZ(),fr((Zmt(),hDt),e)}function U0n(e){return _1(),fr((_gt(),gDt),e)}function z0n(e){return df(),fr((Jgt(),bDt),e)}function H0n(e){return NT(),fr((Qbt(),vDt),e)}function G0n(e){return K3(),fr((Amt(),wDt),e)}function q0n(e){return oa(),fr((Qmt(),yDt),e)}function V0n(e){return Xl(),fr((ebt(),EDt),e)}function W0n(e){return tt(),fr((Cmt(),SDt),e)}function K0n(e){return lj(),fr((xgt(),CDt),e)}function Y0n(e){return Ud(),fr((Kgt(),ADt),e)}function X0n(e){return yd(),fr((a1t(),IDt),e)}function J0n(e){return gR(),fr((Tgt(),ODt),e)}function Z0n(e){return Jse(),fr((xpt(),NDt),e)}function Q0n(e){return pZ(),fr((Zgt(),RDt),e)}function evn(e){return iZ(),fr((Qgt(),DDt),e)}function tvn(e,t){return an(e),e+(an(t),t)}function Kse(e,t){this.c=e,this.a=t,this.b=t-e}function Mut(e,t,i){this.a=e,this.b=t,this.c=i}function X_e(e,t,i){this.a=e,this.b=t,this.c=i}function J_e(e,t,i){this.a=e,this.b=t,this.c=i}function Yse(e,t,i){this.a=e,this.b=t,this.c=i}function jut(e,t,i){this.a=e,this.b=t,this.c=i}function $ut(e,t,i){this.a=e,this.b=t,this.c=i}function Fut(e,t,i){this.a=e,this.c=t,this.b=i}function y0(e,t,i){this.e=e,this.a=t,this.c=i}function Xse(e,t,i){Nf(),Kke.call(this,e,t,i)}function Z_e(e,t,i){Nf(),Kke.call(this,e,t,i)}function But(e,t,i){Nf(),Xse.call(this,e,t,i)}function Q_e(e,t,i){Nf(),Xse.call(this,e,t,i)}function Uut(e,t,i){Nf(),Q_e.call(this,e,t,i)}function zut(e,t,i){Nf(),Z_e.call(this,e,t,i)}function eke(e,t,i){Nf(),Kke.call(this,e,t,i)}function Hut(e,t,i){Nf(),eke.call(this,e,t,i)}function Gut(e,t,i){Nf(),dTe.call(this,e,t,i)}function nvn(e){return pL(),fr((Xbt(),KDt),e)}function sH(e,t){return or(e),or(t),new Yit(e,t)}function WN(e,t){return or(e),or(t),new elt(e,t)}function rvn(e,t){return or(e),or(t),new tlt(e,t)}function ivn(e,t){return or(e),or(t),new iot(e,t)}function f(e,t){return z7(e==null||Lle(e,t)),e}function bP(e){var t;return t=new et,kue(t,e),t}function ovn(e){var t;return t=new Mo,kue(t,e),t}function qut(e){var t;return t=new u2e,jue(t,e),t}function cH(e){var t;return t=new ki,jue(t,e),t}function avn(e){return!e.e&&(e.e=new et),e.e}function svn(e){return!e.c&&(e.c=new ab),e.c}function Jse(){Jse=Z,h7e=new flt,p7e=new Rlt}function vP(){vP=Z,HDe=(tt(),pn),Kte=_n}function cvn(e,t,i){Wlt.call(this,t,i),this.a=e}function Vut(e,t){this.c=e,this.b=t,this.a=!1}function tke(e){this.d=e,Wt(this),this.b=Jvn(e.d)}function Wut(){this.a=";,;",this.b="",this.c=""}function Kut(e,t,i){this.b=e,tst.call(this,t,i)}function nke(e,t,i){this.c=e,AY.call(this,t,i)}function rke(e,t,i){KAe(i,0,e,t,i.length,!1)}function ike(e,t,i){lP.call(this,e,t),this.b=i}function oke(e,t,i,a,u){e.d=t,e.c=i,e.a=a,e.b=u}function db(e,t,i,a,u){e.b=t,e.c=i,e.d=a,e.a=u}function uvn(e,t){t&&(e.b=t,e.a=(Kv(t),t.a))}function uH(e,t){if(!e)throw ie(new dn(t))}function KN(e,t){if(!e)throw ie(new As(t))}function ake(e,t){if(!e)throw ie(new tit(t))}function lvn(e,t){return kY(),tu(e.d.p,t.d.p)}function dvn(e,t){return y1(),di(e.e.b,t.e.b)}function fvn(e,t){return y1(),di(e.e.a,t.e.a)}function Zse(e){return Ln(e.b!=0),lh(e,e.a.a)}function hvn(e){return Ln(e.b!=0),lh(e,e.c.b)}function ske(e){var t,i;t=e.b,i=e.c,e.b=i,e.c=t}function cke(e){var t,i;i=e.d,t=e.a,e.d=t,e.a=i}function Yut(e,t,i,a,u){e.c=t,e.d=i,e.b=a,e.a=u}function wX(e,t){return ukn(e),e.a*=t,e.b*=t,e}function uke(e,t){return t<0?e.g=-1:e.g=t,e}function lH(e,t,i){OSe.call(this,e,t),this.c=i}function lke(e){H_e(),ZC.call(this),this.$h(e)}function Xut(e,t,i){this.a=e,_3.call(this,t,i)}function Jut(e,t,i){this.a=e,_3.call(this,t,i)}function wP(e,t,i){var a;a=new cT(i),Wh(e,t,a)}function pvn(e,t){var i;return i=e.c,Fxe(e,t),i}function yX(e,t,i){OSe.call(this,e,t),this.c=i}function Zut(){RP(),Rwn.call(this,(Uv(),xh))}function Qut(){this.b=ke(Oe(st((Qh(),cge))))}function Qse(e){this.b=e,this.a=Yy(this.b.a).Md()}function elt(e,t){this.b=e,this.a=t,sN.call(this)}function tlt(e,t){this.a=e,this.b=t,sN.call(this)}function gvn(e,t){return tu(slt(e.d),slt(t.d))}function EX(e,t){return t&&GX(e,t.d)?t:null}function mvn(e,t){return t==(tt(),pn)?e.c:e.d}function bvn(e){return new Ke(e.c+e.b,e.d+e.a)}function nlt(e){return e!=null&&!wle(e,zF,HF)}function dH(e){return e==0||isNaN(e)?e:e<0?-1:1}function vvn(e,t){return VX(),zle(e,t),new Idt(e,t)}function wvn(e,t){return(I0t(e)<<4|I0t(t))&To}function Mr(e,t){return Wi(e,t,e.c.b,e.c),!0}function SX(e){var t;return t=e.slice(),bue(t,e)}function _X(e){var t;return t=e.n,e.a.b+t.d+t.a}function rlt(e){var t;return t=e.n,e.e.b+t.d+t.a}function dke(e){var t;return t=e.n,e.e.a+t.b+t.c}function ilt(e){return Xr(),new fb(0,e)}function olt(){olt=Z,y1e=(Ht(),new mz(Tpe))}function alt(){alt=Z,k6e=Pe(ba,Ye,17,256,0,1)}function kX(){kX=Z,new E3e((wae(),jpe),(vae(),Mpe))}function fke(){db(this,!1,!1,!1,!1)}function hke(e){e.a.b=e.b,e.b.a=e.a,e.a=e.b=null}function ece(e){return e.e.c.length+e.g.c.length}function slt(e){return e.e.c.length-e.g.c.length}function clt(e){return e.b.c.length-e.e.c.length}function yvn(e,t){return e.c<t.c?-1:e.c==t.c?0:1}function vS(e,t){return Dj(e,new lP(t.a,t.b))}function Evn(e,t,i){return qr(e.b,f(i.b,18),t)}function Svn(e,t,i){return qr(e.b,f(i.b,18),t)}function _vn(e,t,i){return y1(),i.e.b+i.f.b+e*t}function kvn(e,t,i){return y1(),i.e.a+i.f.a+e*t}function Tvn(e){return rf(),(tt(),tl).Gc(e.j)}function xvn(e){return e.a?e.a:Ice(e)}function z7(e){if(!e)throw ie(new eP(null))}function Cvn(e,t){return Qe(e,new Ke(t.a,t.b))}function Avn(e){return!Fa(e)&&e.c.i.c==e.d.i.c}function ult(e){return lR(),Pe(Zte,eee,39,e,0,1)}function pke(e,t,i,a){p3e.call(this,e,t,i,a,0,0)}function TX(e,t){Wz.call(this,e,t),this.a=this}function llt(e,t){Wlt.call(this,t,1040),this.a=e}function dlt(e){H_e(),lke.call(this,e),this.a=-1}function flt(){kat.call(this,"COUNT_CHILDREN",0)}function gke(e){this.c=e,this.b=this.c.d.vc().Jc()}function hlt(e){return e.e.Pd().gc()*e.c.Pd().gc()}function tce(e,t,i){return new kut(Lwn(e).Ze(),i,t)}function mke(e,t,i){var a;return a=e[t],e[t]=i,a}function Bi(e,t){var i;return i=_ce(e,t),i.i=2,i}function xX(e,t){var i;return++e.j,i=e.zj(t),i}function Sr(e,t,i){return e.a=-1,u_e(e,t.g,i),e}function iT(e,t,i,a,u,l){return $wt(e,t,i,a,u,0,l)}function Ivn(e,t){$xe(e,t==null?null:(an(t),t))}function Ovn(e,t){Bxe(e,t==null?null:(an(t),t))}function Nvn(e,t){Bxe(e,t==null?null:(an(t),t))}function bke(e){if(e.c.e!=e.a)throw ie(new If)}function vke(e){if(e.e.c!=e.b)throw ie(new If)}function CX(e){for(or(e);e.Ob();)e.Pb(),e.Qb()}function nce(e){_S(),this.a=(Ht(),new mz(or(e)))}function Rvn(e){e.a.jd(),f(e.a.kd(),16).gc(),D2e()}function Pvn(e,t){return an(e),zH(e,(an(t),t))}function Lvn(e,t){return an(t),zH(t,(an(e),e))}function plt(e,t){return e.a+=bm(t,0,t.length),e}function ct(e,t){return Gt(t,e.c.length),e.c[t]}function glt(e,t){return Gt(t,e.a.length),e.a[t]}function Ft(e,t){Wy(),hue.call(this,e),this.a=t}function wke(e){JK.call(this,new Zv),Ba(this,e)}function mlt(e){this.c=e,this.a=new LN(this.c.a)}function blt(e){this.a=e,this.c=new Wn,BTn(this)}function H7(e){this.a=new PN(e.gc()),Ba(this,e)}function vlt(e){return e.q?e.q:(Ht(),Ht(),Cb)}function wlt(e){return e.c-f(ct(e.a,e.b),292).b}function ylt(e,t){return e?0:h.Math.max(0,t-1)}function Dvn(e,t,i){return tu(t.d[e.g],i.d[e.g])}function Mvn(e,t,i){return tu(e.d[t.p],e.d[i.p])}function jvn(e,t,i){return tu(e.d[t.p],e.d[i.p])}function $vn(e,t,i){return tu(e.d[t.p],e.d[i.p])}function Fvn(e,t,i){return tu(e.d[t.p],e.d[i.p])}function AX(e,t,i){return h.Math.min(i/e,1/t)}function jd(e){return e.c?e.c.f:e.e.b}function Al(e){return e.c?e.c.g:e.e.a}function Bvn(e,t){return e.a==null&&fSt(e),e.a[t]}function Elt(e){var t;return t=Fwt(e),t?Elt(t):e}function Uvn(e,t){var i;for(i=0;i<t;++i)e[i]=-1}function IX(e,t,i,a){oke(this,e,t,i,a)}function OX(e){OS.call(this,e.gc()),Qi(this,e)}function NX(e,t){Nf(),aY.call(this,t),this.a=e}function G7(e,t,i){this.a=e,So.call(this,t,i,2)}function fb(e,t){Xr(),sS.call(this,e),this.a=t}function Slt(e){this.b=new ki,this.a=e,this.c=-1}function _lt(){_lt=Z,IAt=new Nlt(Pe(Ro,Jt,1,0,5,1))}function klt(){klt=Z,T6e=Pe(i_,Ye,190,256,0,1)}function Tlt(){Tlt=Z,A6e=Pe(o_,Ye,191,256,0,1)}function xlt(){xlt=Z,E6e=Pe(BR,Ye,219,256,0,1)}function Clt(){Clt=Z,_6e=Pe(D$,Ye,179,128,0,1)}function rce(){rce=Z,jDt=new pa((Rr(),Sd),0)}function RX(e,t){return Xr(),new $ke(e,t)}function ice(e,t){return e==null?t==null:Mt(e,t)}function zvn(e,t){return e==null?t==null:SZ(e,t)}function oce(e,t){return me(t,23)&&Cl(e,f(t,23))}function Alt(e,t){return me(t,23)&&nkn(e,f(t,23))}function Is(e,t){return J1t(e,t,kSn(e,e.b.Ae(t)))}function Hvn(e,t){return e.a.get(t)!==void 0}function yke(e){return wd(e,26)*r$+wd(e,27)*i$}function Eke(e){return Array.isArray(e)&&e.Qm===xe}function PX(e){e.b?PX(e.b):e.d.dc()&&e.f.c.Ac(e.e)}function mr(e){return Js(e)?e|0:Iit(e)}function xi(e,t){var i;return i=At(e),fCe(i,t),i}function Gvn(e,t){return!e&&(e=[]),e[e.length]=t,e}function Ky(e,t,i){fT(t,e.c.length),y7(e.c,t,i)}function ace(e,t,i){q1t(0,t,e.length),LP(e,0,t,i)}function qvn(e,t){sce(e,Ho(new Ke(t.a,t.b),e.c))}function sce(e,t){ii(e.c,t),e.b.c+=t.a,e.b.d+=t.b}function cce(e,t){this.b=new ki,this.a=e,this.c=t}function Ilt(){this.b=new ZI,this.c=new uft(this)}function Ske(){this.d=new Sa,this.e=new cft(this)}function _ke(){bTe(),this.f=new ki,this.e=new ki}function Olt(){rf(),this.k=new Wn,this.d=new Mo}function Nlt(e){d_e.call(this,0,0),this.a=e,this.b=0}function Rlt(){kat.call(this,"LOOKAHEAD_LAYOUT",1)}function Vvn(e,t,i){Lj(i,e,1),Qe(t,new eat(i,e))}function Wvn(e,t,i){n2t(i,e,1),Qe(t,new Kot(i,e))}function Kvn(e,t,i){return Go(e,new DN(t.a,i.a))}function Yvn(e,t,i){return-tu(e.f[t.p],e.f[i.p])}function Plt(e,t,i){var a;for(a=0;a<t;++a)e[a]=i}function LX(e,t,i){var a;e&&(a=e.i,a.d=t,a.a=i)}function DX(e,t,i){var a;e&&(a=e.i,a.c=t,a.b=i)}function Il(e,t,i){return e.a=-1,u_e(e,t.g+1,i),e}function kke(e,t,i){return i=hf(e,f(t,52),3,i),i}function Tke(e,t,i){return i=hf(e,f(t,52),7,i),i}function Llt(e,t,i){this.a=e,aX.call(this,t,i,22)}function Dlt(e,t,i){this.a=e,aX.call(this,t,i,14)}function Mlt(e,t,i,a){Nf(),vht.call(this,e,t,i,a)}function jlt(e,t,i,a){Nf(),vht.call(this,e,t,i,a)}function Xvn(e,t){(t.Bb&Dc)!=0&&!e.a.o&&(e.a.o=t)}function $lt(e){return e!=null&&gce(e)&&e.Qm!==xe}function xke(e){return!Array.isArray(e)&&e.Qm===xe}function Jvn(e){return me(e,15)?f(e,15).cd():e.Jc()}function Cke(e){return e.Oc(Pe(Ro,Jt,1,e.gc(),5,1))}function q7(e,t){return f3n(LH(e,t))?t.vi():null}function Flt(e,t){return or(t),e.a.Hd(t)&&!e.b.Hd(t)}function Yy(e){return e.c?e.c:e.c=e.Qd()}function uce(e){return e.d?e.d:e.d=e.Rd()}function Blt(e){var t;return t=e.f,t||(e.f=e.Cc())}function YN(e){var t;return t=e.i,t||(e.i=e.bc())}function Ult(e){return Xr(),new Bce(10,e,0)}function Xy(e){zv(),this.a=0,this.b=e-1,this.c=1}function zlt(e){rX(this),this.g=e,YX(this),this.he()}function Ake(e,t,i,a){this.a=e,EJ.call(this,e,t,i,a)}function Tn(e){this.a=(_lt(),IAt),this.d=f(or(e),49)}function MX(e){if(e.e.j!=e.d)throw ie(new If)}function V7(e){return Js(e)?""+e:iSt(e)}function hb(e,t){return tw(yyt(Js(e)?dh(e):e,t))}function wS(e,t){return tw(rIe(Js(e)?dh(e):e,t))}function Jy(e,t){return tw(ZNn(Js(e)?dh(e):e,t))}function Zvn(e,t){return Mbn((an(e),e),(an(t),t))}function lce(e,t){return di((an(e),e),(an(t),t))}function Qvn(e,t){return Ku(e.l&t.l,e.m&t.m,e.h&t.h)}function ewn(e,t){return Ku(e.l|t.l,e.m|t.m,e.h|t.h)}function twn(e,t){return Ku(e.l^t.l,e.m^t.m,e.h^t.h)}function oT(e,t){return tG(e,(an(t),new I(t)))}function XN(e,t){return tG(e,(an(t),new M(t)))}function Hlt(e){return Nl(),f(e,11).g.c.length!=0}function Glt(e){return Nl(),f(e,11).e.c.length!=0}function nwn(e,t){return dR(),di(t.a.o.a,e.a.o.a)}function qlt(e,t,i){return Gkt(e,f(t,11),f(i,11))}function rwn(e){return e.e?QTe(e.e):null}function Ike(e){e.d||(e.d=e.b.Jc(),e.c=e.b.gc())}function iwn(e,t,i){e.a.Mb(i)&&(e.b=!0,t.Ad(i))}function yP(e,t){if(e<0||e>=t)throw ie(new trt)}function Vlt(e){this.a=jvt(e.a),this.b=new Kl(e.b)}function jX(e){this.b=new Du(11),this.a=(TS(),e)}function dce(e){this.b=null,this.a=(TS(),e||M6e)}function Oke(e,t){this.e=e,this.d=(t&64)!=0?t|Sm:t}function Wlt(e,t){this.c=0,this.d=e,this.b=t|64|Sm}function Klt(e){this.b=e,HN.call(this,e),ect(this)}function Ylt(e){this.b=e,Qz.call(this,e),tct(this)}function Jo(e,t,i,a,u){gue.call(this,e,t,i,a,u,-1)}function W7(e,t,i,a,u){NH.call(this,e,t,i,a,u,-1)}function Be(e,t,i,a){So.call(this,e,t,i),this.b=a}function E0(e,t,i,a){var u;u=e.i,u.i=t,u.a=i,u.b=a}function Nke(e){var t;for(t=e;t.f;)t=t.f;return t}function $X(e){return new Ke(e.c+e.b/2,e.d+e.a/2)}function fce(e,t){return t.Rh()?iw(e.b,f(t,52)):t}function K7(e){return Xl(),!e.Gc(M1)&&!e.Gc(Lw)}function Xlt(e,t,i){return bL(),zue(e,t)&&zue(e,i)}function own(e,t,i){t.of(i,ke(Oe(bn(e.b,i)))*e.a)}function awn(e,t){t.Sg("General 'Rotator",1),s9n(e)}function Rke(e,t,i,a){this.b=e,So.call(this,t,i,a)}function aT(e,t,i){this.a=e,qN.call(this,t,i,5,6)}function Jlt(e,t){this.b=e,W9.call(this,e.b),this.a=t}function FX(e,t,i,a){lH.call(this,e,t,i),this.b=a}function Zlt(e){qat.call(this,e,!1),this.a=!1}function Pke(e,t){_S(),Xpn.call(this,e,cZ(new bu(t)))}function BX(e,t){return Xr(),new Wke(e,t,0)}function hce(e,t){return Xr(),new Wke(6,e,t)}function swn(e,t){return Mt(e.substr(0,t.length),t)}function Yu(e,t){return ta(t)?Mce(e,t):!!Is(e.f,t)}function va(e,t){for(an(t);e.Ob();)t.Ad(e.Pb())}function Zy(e,t,i){ym(),this.e=e,this.d=t,this.a=i}function Ne(e,t,i,a){q1t(t,i,e.length),cwn(e,t,i,a)}function cwn(e,t,i,a){var u;for(u=t;u<i;++u)e[u]=a}function Lke(e,t){var i;for(i=0;i<t;++i)e[i]=!1}function EP(e,t){var i;return i=e.a.gc(),XTe(t,i),i-t}function JN(e){var t;return t=vj(e),Ln(t!=null),t}function uwn(e){var t;return t=Xxn(e),Ln(t!=null),t}function Go(e,t){var i;return i=e.a.yc(t,e),i==null}function fH(e,t){return e.a.yc(t,(mn(),yw))==null}function pce(e){return new Ft(null,vwn(e,e.length))}function pb(e){return new Tn(new r_e(e.a.length,e.a))}function lwn(e){return Ku(~e.l&Ed,~e.m&Ed,~e.h&A1)}function gce(e){return typeof e===DG||typeof e===cfe}function dwn(e,t){return e>0?h.Math.log(e/t):-100}function Dke(e,t){return t?Ba(e,t):!1}function Qlt(e){if(!e)throw ie(new ic);return e.d}function Mke(e,t){if(e==null)throw ie(new IN(t))}function ZN(e,t,i){return Yh(e.a,t),mke(e.b,t.g,i)}function fwn(e,t,i){yP(i,e.a.c.length),af(e.a,i,t)}function hwn(e,t,i){f9n(e.a,e.b,e.d,e.c,f(t,168),i)}function hH(e,t){qst(e,me(t,159)?t:f(t,1982).Ol())}function Ap(e,t,i){return hxn(e,t.g,i),Yh(e.c,t),e}function pwn(e){return oA(e,(ko(),Ds)),e.d=!0,e}function mce(e){return!e.j&&Do(e,pLn(e.g,e.b)),e.j}function gwn(e,t,i){var a;return a=gT(e,t),Qce(e,t,i),a}function jke(e,t,i){this.c=e,this.a=t,Ht(),this.b=i}function bce(e,t){this.d=e,Nn.call(this,e),this.e=t}function $ke(e,t){sS.call(this,1),this.a=e,this.b=t}function Y7(e){Qk(e.b!=-1),_0(e.c,e.a=e.b),e.b=-1}function nu(e){e.f=new hst(e),e.i=new pst(e),++e.g}function edt(e){e.a=null,e.e=null,nu(e.b),e.d=0,++e.c}function X7(e){return e.e==0?e:new Zy(-e.e,e.d,e.a)}function mwn(e){return e==Gi?mq:e==qo?"-INF":""+e}function bwn(e){return e==Gi?mq:e==qo?"-INF":""+e}function vwn(e,t){return lkn(t,e.length),new llt(e,t)}function wwn(e,t){return di(e.c.c+e.c.b,t.c.c+t.c.b)}function tdt(e,t){return Eu(e,t)<0?-1:Eu(e,t)>0?1:0}function ndt(e,t){return e.c?ndt(e.c,t):Qe(e.b,t),e}function UX(e,t){Wi(e.d,t,e.b.b,e.b),++e.a,e.c=null}function sT(e){this.d=(an(e),e),this.a=0,this.c=FG}function rdt(e){uTe.call(this),this.a=e,Qe(e.a,this)}function vce(e,t){this.d=Zxn(e),this.c=t,this.a=.5*t}function idt(e,t,i){zbt.call(this,t,i,!1),this.e=e}function SP(e,t,i,a,u){ode(e,f(si(t.k,i),15),i,a,u)}function odt(e,t,i,a,u){for(;t<i;)a[u++]=_a(e,t++)}function adt(e,t,i){var a;for(a=0;a<t;++a)io(e,a,i)}function Qi(e,t){return e.Pi()&&(t=Nft(e,t)),e.Ci(t)}function yS(e,t){Zi(Lu(e.Mc(),new rg),new Uet(t))}function ywn(e,t){return di(e.e.a+e.f.a,t.e.a+t.f.a)}function Ewn(e,t){return di(e.e.b+e.f.b,t.e.b+t.f.b)}function wce(e){return h.Math.abs(e.d.e-e.e.e)-e.a}function jn(e){return(e.i==null&&Em(e),e.i).length}function Swn(e,t,i){return f(e.c.fd(t,f(i,135)),43)}function _wn(){return dY(),ce(se(OAt,1),Ve,534,0,[Bpe])}function sdt(){sdt=Z,DAt=new G5(!1),MAt=new G5(!0)}function Wv(){Wv=Z,sn=c4n(),Ut(),Ex&&wIn()}function zX(e){e.s=NaN,e.c=NaN,HEt(e,e.e),HEt(e,e.j)}function Fke(e){e?uAe(e,(b0(),UAt)):xyn((b0(),e))}function kwn(e,t){++e.j,ide(e,e.i,t),u4n(e,f(t,339))}function yce(e,t){return t=e.Vk(null,t),nyt(e,null,t)}function cdt(e,t,i){return Xr(),new tht(e,t,i)}function Ece(e){var t;return t=e.j,t||(e.j=new Mk(e))}function udt(e){var t;return t=e.f,t||(e.f=new n_e(e))}function Bke(e){var t;return t=e.k,t||(e.k=new F5(e))}function HX(e){var t;return t=e.k,t||(e.k=new F5(e))}function J7(e){var t;return t=e.g,t||(e.g=new q9(e))}function Twn(e){var t;return t=e.i,t||(e.i=new V9(e))}function Sce(e){var t;return t=e.d,t||(e.d=new By(e))}function _ce(e,t){var i;return i=new uCe,i.j=e,i.d=t,i}function R(e,t,i){var a;return a=_ce(e,t),rbt(i,a),a}function Uke(e,t){var i;return i=e.a.gc(),_H(t,i),i-1-t}function ldt(e,t){return _H(t,e.c.b.c.gc()),new Kit(e,t)}function zke(e){return me(e,601)?e:new Tft(e)}function or(e){if(e==null)throw ie(new kN);return e}function cT(e){if(e==null)throw ie(new kN);this.a=e}function ddt(e){c2e(),this.b=new et,this.a=e,f8n(this,e)}function Hke(e){this.b=e,this.a=f(ch(this.b.a.e),226)}function fdt(){Zv.call(this),this.a=!0,this.b=!0}function Gke(e,t){return gn(t,e.length+1),e.substr(t)}function io(e,t,i){return sbn(i==null||ZLn(e,i)),e[t]=i}function _P(e,t){return e.a+=String.fromCharCode(t),e}function Qy(e,t){return e.a+=String.fromCharCode(t),e}function kce(e,t){for(an(t);e.c<e.d;)e.Qe(t,e.c++)}function hdt(e,t){return f(Yy(e.a).Kd().Xb(t),43).jd()}function bn(e,t){return ta(t)?wu(e,t):ac(Is(e.f,t))}function xwn(e,t){return dR(),f(Es(e,t.d),15).Ec(t)}function Cwn(e,t,i,a){return i==0||(i-a)/i<e.e||t>=e.g}function Tce(e,t,i){var a;return a=Lue(e,t,i),LIe(e,a)}function pdt(e,t){var i;i=console[e],i.call(console,t)}function Awn(e,t,i){f(t.b,68),Mu(t.a,new X_e(e,i,t))}function Iwn(e,t){QN(e,new cT(t.f!=null?t.f:""+t.g))}function Own(e,t){QN(e,new cT(t.f!=null?t.f:""+t.g))}function gdt(e,t){var i;++e.j,i=e.Bj(),e.oj(e.Wi(i,t))}function QN(e,t){var i;i=e.a.length,gT(e,i),Qce(e,i,t)}function Nwn(e){return!e.a&&(e.a=new So(Dw,e,4)),e.a}function ES(e){return!e.d&&(e.d=new So(gs,e,1)),e.d}function mdt(e){return me(e,101)&&(f(e,19).Bb&Dc)!=0}function Z7(e){return h.Math.sqrt(e.a*e.a+e.b*e.b)}function qke(e){this.d=e,this.c=e.a.d.a,this.b=e.a.e.g}function bdt(e){this.c=e,this.a=new ki,this.b=new ki}function ru(e){this.c=new ha,this.a=new et,this.b=e}function vdt(e){this.b=new et,this.a=new et,this.c=e}function Vke(e,t,i){this.a=e,n2e.call(this,t),this.b=i}function wdt(e,t,i){this.a=e,OTe.call(this,8,t,null,i)}function Wke(e,t,i){sS.call(this,e),this.a=t,this.b=i}function Kke(e,t,i){aY.call(this,t),this.a=e,this.b=i}function ydt(e){this.c=e,this.b=this.c.a,this.a=this.c.e}function Rwn(e){this.a=(an(kr),kr),this.b=e,new y2e}function hd(e){Qk(e.c!=-1),e.d.ed(e.c),e.b=e.c,e.c=-1}function Edt(e){tTe(e.a),e.b=Pe(Ro,Jt,1,e.b.length,5,1)}function GX(e,t){return Jle(e.c,e.f,t,e.b,e.a,e.e,e.d)}function SS(e,t){return yP(t,e.a.c.length),ct(e.a,t)}function dm(e,t){return Ae(e)===Ae(t)||e!=null&&ei(e,t)}function Sdt(e){return ch(e),me(e,470)?f(e,470):Qs(e)}function _dt(e){return e?e.dc():!e.Jc().Ob()}function Pwn(e){return _x?Mce(_x,e):!1}function Lwn(e){return 0>=e?new X2e:Okn(e-1)}function ra(e){return!e.a&&e.c?e.c.b:e.a}function ch(e){if(e==null)throw ie(new kN);return e}function an(e){if(e==null)throw ie(new kN);return e}function Q7(e){e.c?e.c.Ye():(e.d=!0,u5n(e))}function Kv(e){e.c?Kv(e.c):(ow(e),e.d=!0)}function kdt(e){e.b=!1,e.c=!1,e.d=!1,e.a=!1}function Tdt(e){var t,i;return t=e.c.i.c,i=e.d.i.c,t==i}function Dwn(e,t){var i;i=e.Eh(t),i>=0?e.gi(i):FAe(e,t)}function xdt(e,t){e.c<0||e.b.b<e.c?Ma(e.b,t):e.a.vf(t)}function Mwn(e,t){Qn((!e.a&&(e.a=new nH(e,e)),e.a),t)}function _S(){_S=Z,lm(),$pe=new Fce((Ht(),Ht(),Ga))}function xce(){xce=Z,lm(),h6e=new nSe((Ht(),Ht(),Ree))}function jwn(e,t,i){return hR(),i.Jg(e,f(t.jd(),144))}function $wn(e,t){return tu(t.j.c.length,e.j.c.length)}function Cdt(e,t){this.b=e,bce.call(this,e,t),ect(this)}function Adt(e,t){this.b=e,z_e.call(this,e,t),tct(this)}function Fwn(e,t){sce(f(t.b,68),e),Mu(t.a,new pg(e))}function Bwn(e){if(e.p!=3)throw ie(new Tl);return e.e}function Uwn(e){if(e.p!=4)throw ie(new Tl);return e.e}function zwn(e){if(e.p!=6)throw ie(new Tl);return e.f}function Hwn(e){if(e.p!=3)throw ie(new Tl);return e.j}function Gwn(e){if(e.p!=4)throw ie(new Tl);return e.j}function qwn(e){if(e.p!=6)throw ie(new Tl);return e.k}function Yke(e){return!e.b&&(e.b=new iY(new bae)),e.b}function kS(e){return e.c==-2&&Me(e,oNn(e.g,e.b)),e.c}function kP(e,t){var i;return i=_ce("",e),i.n=t,i.i=1,i}function qX(e,t,i,a){dS.call(this,e,i),this.a=t,this.f=a}function Xke(e,t,i,a){dS.call(this,e,t),this.d=i,this.a=a}function Idt(e,t){mmn.call(this,Ikn(or(e),or(t))),this.a=t}function _o(){Mrt.call(this),Bk(this.j.c,0),this.a=-1}function Odt(){mAe.call(this,CE,(Sit(),SMt)),QMn(this)}function Ndt(){mAe.call(this,bh,(iP(),$7e)),aMn(this)}function Rdt(){ur.call(this,"DELAUNAY_TRIANGULATION",0)}function Vwn(e){return String.fromCharCode.apply(null,e)}function qr(e,t,i){return ta(t)?Os(e,t,i):ml(e.f,t,i)}function Jke(e){return Ht(),e?e.Me():(TS(),TS(),j6e)}function Wwn(e){return uf(e,ffe),NJ(Ua(Ua(5,e),e/10|0))}function Pdt(e,t){return kX(),new E3e(new fct(e),new dct(t))}function Pu(e,t){return Ae(e)===Ae(t)||e!=null&&ei(e,t)}function Es(e,t){return oce(e.a,t)?e.b[f(t,23).g]:null}function Ldt(e,t){e.u.Gc((Xl(),M1))&&n4n(e,t),JSn(e,t)}function Ddt(e,t){var i;return i=new ru(e),vn(t.c,i),i}function TP(e){return!e.a&&(e.a=new iit(e.c.vc())),e.a}function Mdt(e){return!e.b&&(e.b=new tP(e.c.ec())),e.b}function jdt(e){return!e.d&&(e.d=new h3(e.c.Bc())),e.d}function Cce(e){return!e.n&&(e.n=new Be(hc,e,1,7)),e.n}function Kwn(e){return!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b}function Zke(e){return!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c}function pH(e){return e.j.c.length=0,tTe(e.c),zbn(e.a),e}function Fr(e,t,i,a){return Hbt(e,t,i,!1),uZ(e,a),e}function gb(e,t){for(;t-- >0;)e=e<<1|(e<0?1:0);return e}function Qke(e){return!e.c&&(e.c=new Be(_d,e,9,9)),e.c}function VX(){VX=Z,AAt=new I2e(ce(se(AE,1),yQ,43,0,[]))}function Ywn(){return xY(),ce(se(oLt,1),Ve,543,0,[Nbe])}function Xwn(){return TY(),ce(se(k8e,1),Ve,475,0,[Obe])}function Jwn(){return g7(),ce(se(q8e,1),Ve,524,0,[Zq])}function $dt(e,t){GCe(e,ke(E1(t,"x")),ke(E1(t,"y")))}function Fdt(e,t){GCe(e,ke(E1(t,"x")),ke(E1(t,"y")))}function _a(e,t){return gn(t,e.length),e.charCodeAt(t)}function eTe(e,t){this.b=e,this.c=t,this.a=new LN(this.b)}function ej(e){this.d=e,this.a=this.d.b,this.b=this.d.c}function Ace(e,t,i){this.a=DOe,this.d=e,this.b=t,this.c=i}function Bdt(e,t){this.d=(an(e),e),this.a=16449,this.c=t}function gH(e){return e.f==XL&&Zt(e,sIn(e.g,e.b)),e.f}function xP(e){return e.e==XL&&Vn(e,w3n(e.g,e.b)),e.e}function P3(e){var t;return t=e.b,!t&&(e.b=t=new Lk(e)),t}function tTe(e){var t;for(t=e.Jc();t.Ob();)t.Pb(),t.Qb()}function Zwn(e,t,i){var a;a=f(e.d.Kb(i),158),a&&a.Nb(t)}function Qwn(e,t){return new fft(f(or(e),51),f(or(t),51))}function eyn(e,t){return J2e(),di((an(e),e),(an(t),t))}function Jr(e,t){return ow(e),new Ft(e,new dxe(t,e.a))}function Lu(e,t){return ow(e),new Ft(e,new ZTe(t,e.a))}function WX(e,t){return ow(e),new y_e(e,new Gpt(t,e.a))}function uT(e,t){return ow(e),new w_e(e,new qpt(t,e.a))}function tyn(e,t){return di(e.d.c+e.d.b/2,t.d.c+t.d.b/2)}function nyn(e,t){return di(e.g.c+e.g.b/2,t.g.c+t.g.b/2)}function ryn(e){tf();var t;t=f(e.g,9),t.n.a=e.d.c+t.d.b}function Ice(e){var t;return t=Rkn(e),t||null}function Udt(){oc.call(this,"There is no more element.")}function eR(e,t,i,a){ur.call(this,e,t),this.a=i,this.b=a}function nTe(e,t,i,a){this.c=e,this.b=t,this.a=i,this.d=a}function zdt(e,t,i,a){this.c=e,this.b=t,this.d=i,this.a=a}function qh(e,t,i,a){this.c=e,this.d=t,this.b=i,this.a=a}function KX(e,t,i,a){this.a=e,this.c=t,this.b=i,this.d=a}function Hdt(e,t,i,a){this.a=e,this.d=t,this.c=i,this.b=a}function Oce(e,t,i,a){this.a=e,this.e=t,this.d=i,this.c=a}function Gdt(e){this.a=new et,this.e=Pe(pr,Ye,54,e,0,2)}function iyn(e){return e!=null&&Iz(Bne,e.toLowerCase())}function oyn(e,t){return di(Al(e)*jd(e),Al(t)*jd(t))}function ayn(e,t){return di(Al(e)*jd(e),Al(t)*jd(t))}function syn(e,t,i){nMn(e.a,i),STn(i),F4n(e.b,i),SMn(t,i)}function Nce(e,t,i){var a,u;return a=EOe(e),u=t.pi(i,a),u}function f1(e){var t,i;return i=(t=new cS,t),$P(i,e),i}function Rce(e){var t,i;return i=(t=new cS,t),wAe(i,e),i}function qdt(e){return!e.b&&(e.b=new Be(vo,e,12,3)),e.b}function CP(e){if(vd(e.d),e.d.d!=e.c)throw ie(new If)}function Vdt(e,t,i,a){this.a=e,this.c=t,this.d=i,this.b=a}function Wdt(e,t,i,a){this.a=e,this.b=t,this.d=i,this.c=a}function Kdt(e,t,i,a){this.a=e,this.b=t,this.c=i,this.d=a}function Ydt(e,t,i,a){this.a=e,this.b=t,this.c=i,this.d=a}function eE(e,t,i,a){this.e=e,this.a=t,this.c=i,this.d=a}function Xdt(e,t,i,a){Nf(),Bpt.call(this,t,i,a),this.a=e}function Jdt(e,t,i,a){Nf(),Bpt.call(this,t,i,a),this.a=e}function Zdt(e,t){this.a=e,Ebn.call(this,e,f(e.d,15).dd(t))}function Pce(e){this.f=e,this.c=this.f.e,e.f>0&&Nwt(this)}function Qdt(e,t,i,a){this.b=e,this.c=a,eX.call(this,t,i)}function nf(e){e.a.a=e.c,e.c.b=e.a,e.a.b=e.c.a=null,e.b=0}function rTe(e,t){return e.b=t.b,e.c=t.c,e.d=t.d,e.a=t.a,e}function YX(e){return e.n&&(e.e!==iTt&&e.he(),e.j=null),e}function eft(e){return z7(e==null||gce(e)&&e.Qm!==xe),e}function iTe(e){return Ln(e.b<e.d.gc()),e.d.Xb(e.c=e.b++)}function cyn(e,t){return f(V(e,(Ge(),XR)),15).Ec(t),t}function uyn(e,t){return jt(e,f(V(t,(ot(),bD)),17),t)}function lyn(e){return HS(e)&&lt(ut(Fe(e,(ot(),PE))))}function tR(e){var t;return t=e.f,t||(e.f=new aP(e,e.c))}function dyn(e,t,i){return p7(),_Cn(f(bn(e.e,t),514),i)}function fyn(e,t,i){e.i=0,e.e=0,t!=i&&R1t(e,t,i)}function hyn(e,t,i){e.i=0,e.e=0,t!=i&&P1t(e,t,i)}function tft(e,t){e.e&&!e.e.a&&(Hnt(e.e,t),tft(e.e,t))}function nft(e,t){e.d&&!e.d.a&&(Hnt(e.d,t),nft(e.d,t))}function rft(e,t){this.g=e,this.d=ce(se(Ib,1),j0,9,0,[t])}function oTe(e,t){new ki,this.a=new fd,this.b=e,this.c=t}function ift(e){this.b=new et,$o(this.b,this.b),this.a=e}function oft(){dde(this,new om),this.wb=(Wv(),sn),iP()}function pyn(e){return f(e.jd(),144).Ng()+":"+Qs(e.kd())}function aft(e,t){return W3(e.j,t.s,t.c)+W3(t.e,e.s,e.c)}function gyn(e,t){return-di(Al(e)*jd(e),Al(t)*jd(t))}function tE(e,t,i){var a,u;a=zse(i),u=new oS(a),Wh(e,t,u)}function myn(e,t){return Fd(e),Fd(t),Jrt(f(e,23),f(t,23))}function byn(e){e.Sg("No crossing minimization",1),e.Tg()}function vyn(e){hY(),h.setTimeout(function(){throw e},0)}function je(e){return!e.s&&(e.s=new Be(El,e,21,17)),e.s}function ci(e){return!e.q&&(e.q=new Be(Th,e,11,10)),e.q}function XX(e){return!e.a&&(e.a=new Be(Ur,e,10,11)),e.a}function JX(e,t){if(e==null)throw ie(new IN(t));return e}function Ht(){Ht=Z,Ga=new qe,Cb=new Xe,Ree=new _e}function TS(){TS=Z,M6e=new We,Kpe=new We,j6e=new ht}function aTe(){aTe=Z,uIt=new Pn,dIt=new Ske,lIt=new Kr}function AP(){AP=Z;var e,t;t=!QCn(),e=new K,Hpe=t?new G:e}function wyn(e){return or(e),Wvt(new Tn(Cn(e.a.Jc(),new B)))}function yyn(e){return new rct(e,e.e.Pd().gc()*e.c.Pd().gc())}function Eyn(e){return new ict(e,e.e.Pd().gc()*e.c.Pd().gc())}function Lce(e){return me(e,16)?new H7(f(e,16)):ovn(e.Jc())}function Dce(e){return Ht(),me(e,59)?new Iae(e):new cX(e)}function Mce(e,t){return t==null?!!Is(e.f,null):Hvn(e.i,t)}function sTe(e){return e&&e.hashCode?e.hashCode():bS(e)}function cTe(e){this.b=e,this.c=e,e.e=null,e.c=null,this.a=1}function sft(e){this.b=e,this.a=new g0(f(or(new n6),51))}function cft(e){this.c=e,this.b=new g0(f(or(new Ei),51))}function uft(e){this.c=e,this.b=new g0(f(or(new JI),51))}function xS(){this.a=new fd,this.b=(uf(3,$T),new Du(3))}function uTe(){this.c=new ha,this.d=new ha,this.e=new ha}function lft(){this.b=new Mo,this.d=new ki,this.e=new v2e}function ZX(e){this.c=e.c,this.d=e.d,this.b=e.b,this.a=e.a}function dft(e,t,i,a){OSe.call(this,t,i),this.b=e,this.a=a}function fft(e,t){rpn.call(this,new dce(e)),this.a=e,this.b=t}function hft(e,t,i,a,u,l){this.a=e,Oue.call(this,t,i,a,u,l)}function pft(e,t,i,a,u,l){this.a=e,Oue.call(this,t,i,a,u,l)}function jce(e,t,i,a,u,l){NH.call(this,e,t,i,a,u,l?-2:-1)}function gft(e,t,i){return e.f?e.f.cf(t,i):!1}function Bs(e){return e.u||(pd(e),e.u=new Yct(e,e)),e.u}function Syn(e,t){var i;return i=KSe(e.a,t),i&&(t.d=null),i}function fl(e){var t;return t=f(fn(e,16),29),t||e.ei()}function QX(e,t){var i;return i=Vy(e.Om),t==null?i:i+": "+t}function S0(e,t){this.e=e,this.a=Ro,this.b=ySt(t),this.c=t}function mft(e,t){return me(t,144)&&Mt(e.b,f(t,144).Ng())}function _yn(e,t){return e.a?t.Ch().Jc():f(t.Ch(),71).Fi()}function kyn(e,t,i){return ke(Oe(i.a))<=e&&ke(Oe(i.b))>=t}function Tyn(e,t,i){return i>=0&&Mt(e.substr(i,t.length),t)}function Yv(e,t,i,a,u,l,g){return new nue(e.e,t,i,a,u,l,g)}function $ce(e,t,i,a){io(e.c[t.g],t.g,i),io(e.b[t.g],t.g,a)}function mH(e,t,i,a){io(e.c[t.g],i.g,a),io(e.c[i.g],t.g,a)}function uh(e,t,i){return ga(t,i,e.length),e.substr(t,i-t)}function bft(e,t){var i;return i=e.b.Oc(t),npt(i,e.b.gc()),i}function bH(e,t){if(e==null)throw ie(new IN(t));return e}function Fce(e){this.a=(Ht(),me(e,59)?new Iae(e):new cX(e))}function xyn(e){e&&QX(e,e.ge())}function vft(e,t){hX.call(this),Exe(this),this.a=e,this.c=t}function Cyn(){return OJ(),ce(se(X6e,1),Ve,421,0,[age,jee])}function Ayn(){return BP(),ce(se(WOt,1),Ve,501,0,[kq,yge])}function Iyn(){return fm(),ce(se(X6t,1),Ve,507,0,[y_,G0])}function Oyn(){return vg(),ce(se(Z6t,1),Ve,506,0,[$E,Mg])}function Nyn(){return oj(),ce(se(H6t,1),Ve,516,0,[dF,lF])}function Ryn(){return MH(),ce(se(ODe,1),Ve,418,0,[Lme,Dme])}function Pyn(){return SJ(),ce(se(lPe,1),Ve,419,0,[uPe,Dge])}function Lyn(){return FH(),ce(se(vPe,1),Ve,416,0,[hte,bPe])}function Dyn(){return lJ(),ce(se(LPe,1),Ve,417,0,[Xge,PPe])}function Myn(){return RH(),ce(se(EPe,1),Ve,474,0,[yPe,gte])}function jyn(){return gJ(),ce(se(b8e,1),Ve,476,0,[xbe,m8e])}function $yn(){return pJ(),ce(se(KPe,1),Ve,420,0,[ame,ome])}function Fyn(){return uJ(),ce(se(TMe,1),Ve,422,0,[rbe,kMe])}function Byn(){return GJ(),ce(se(xMe,1),Ve,485,0,[rne,jA])}function Uyn(){return XH(),ce(se(W8e,1),Ve,424,0,[mne,V8e])}function zyn(){return CH(),ce(se(T8e,1),Ve,425,0,[Rbe,pne])}function Hyn(){return kJ(),ce(se(AMe,1),Ve,423,0,[CMe,ube])}function Gyn(){return Ip(),ce(se(h5t,1),Ve,513,0,[dx,Cw])}function qyn(){return AS(),ce(se(y5t,1),Ve,451,0,[Aw,LA])}function Vyn(){return Jse(),ce(se(Nne,1),Ve,343,0,[h7e,p7e])}function eJ(e,t,i,a){return i>=0?e.Qh(t,i,a):e.yh(null,i,a)}function tJ(e){return e.b.b==0?e.a.uf():Zse(e.b)}function Wyn(e){if(e.p!=5)throw ie(new Tl);return mr(e.f)}function Kyn(e){if(e.p!=5)throw ie(new Tl);return mr(e.k)}function lTe(e){return Ae(e.a)===Ae((Uue(),b1e))&&WMn(e),e.a}function nJ(e,t){return JX(e,"set1"),JX(t,"set2"),new lot(e,t)}function wft(e,t){Y5(this,new Ke(e.a,e.b)),X5(this,cH(t))}function CS(){ipn.call(this,new PN(mT(12))),QSe(!0),this.a=2}function Bce(e,t,i){Xr(),sS.call(this,e),this.b=t,this.a=i}function dTe(e,t,i){Nf(),aY.call(this,t),this.a=e,this.b=i}function yft(e){var t;t=e.c.d.b,e.b=t,e.a=e.c.d,t.a=e.c.d.b=e}function Yyn(e,t){e.b=t,e.c>0&&e.b>0&&(e.g=AX(e.c,e.b,e.a))}function Xyn(e,t){e.c=t,e.c>0&&e.b>0&&(e.g=AX(e.c,e.b,e.a))}function Jyn(e){return e.b==0?null:(Ln(e.b!=0),lh(e,e.a.a))}function wu(e,t){return t==null?ac(Is(e.f,null)):w7(e.i,t)}function Eft(e,t,i,a,u){return new hde(e,(DP(),Zpe),t,i,a,u)}function rJ(e,t){return ipt(t),vkn(e,Pe(pr,Pr,30,t,15,1),t)}function Zyn(e,t){lSt(e,!0),Mu(e.e.Pf(),new Y_e(e,!0,t))}function vH(e){var t;Tkn(e.a),Cst(e.a),t=new rs(e.a),t3e(t)}function Qyn(e,t){var i;return i=fkn(e.f,t),ii(fX(i),e.f.d)}function Sft(e,t){var i,a;return i=t,a=new on,tkt(e,i,a),a.d}function Uce(e,t,i,a){var u;u=new cut,t.a[i.g]=u,ZN(e.b,a,u)}function iJ(e,t,i){this.c=new et,this.e=e,this.f=t,this.b=i}function fTe(e,t,i){this.a=new et,this.e=e,this.f=t,this.c=i}function _ft(e,t,i){this.i=new et,this.b=e,this.g=t,this.a=i}function kft(e){hX.call(this),Exe(this),this.a=e,this.c=!0}function Tft(e){this.a=f(or(e),274),this.b=(Ht(),new a_e(e))}function hTe(e,t,i){var a;a=e.Eh(t),a>=0?e.Zh(a,i):pIe(e,t,i)}function L3(e,t,i){oJ(),e&&qr(p1e,e,t),e&&qr(SV,e,i)}function eEn(e,t){return y1(),f(V(t,(Ec(),Lm)),17).a==e}function Aa(e){return Math.max(Math.min(e,Qr),-2147483648)|0}function tEn(e,t){var i=zpe[e.charCodeAt(0)];return i??e}function lT(e){return e.Db>>16!=3?null:f(e.Cb,27)}function bg(e){return e.Db>>16!=9?null:f(e.Cb,27)}function xft(e){return e.Db>>16!=6?null:f(e.Cb,85)}function nEn(e){if(e.ye())return null;var t=e.n;return xee[t]}function h1(e,t){if(t==null)throw ie(new kN);return o3n(e,t)}function rEn(e,t){return u1t(e,t)?(abt(e),!0):!1}function iEn(e){function t(){}return t.prototype=e||{},new t}function Cft(e){var t;return t=new bY(mT(e.length)),fCe(t,e),t}function Aft(e,t){return e.b+=t.b,e.c+=t.c,e.d+=t.d,e.a+=t.a,e}function Ift(e,t){var i;for(or(t),i=e.a;i;i=i.c)t.Wd(i.g,i.i)}function wH(e,t){var i;i=e.q.getHours(),e.q.setDate(t),Yj(e,i)}function zce(e,t){return h.Math.abs(e)<h.Math.abs(t)?e:t}function oEn(e,t){return mX(),mn(),f(t.a,17).a<e}function aEn(e,t){return mX(),mn(),f(t.b,17).a<e}function Oft(e,t){var i;return i=e.Eh(t),i>=0?e.Sh(i):wde(e,t)}function Hce(e,t,i){var a;a=O1t(e,t,i),e.b=new YJ(a.c.length)}function fm(){fm=Z,y_=new vSe(RR,0),G0=new vSe(NR,1)}function vg(){vg=Z,$E=new wSe(Gfe,0),Mg=new wSe("UP",1)}function AS(){AS=Z,Aw=new ESe(NR,0),LA=new ESe(RR,1)}function Nft(e,t){var i;return i=new wke(t),oyt(i,e),new Kl(i)}function pTe(e){var t;return t=e.d,t=e.$i(e.f),Qn(e,t),t.Ob()}function sEn(e){if(e.p!=0)throw ie(new Tl);return N7(e.f,0)}function cEn(e){if(e.p!=0)throw ie(new Tl);return N7(e.k,0)}function Rft(e){return e.Db>>16!=7?null:f(e.Cb,240)}function gTe(e){return e.Db>>16!=7?null:f(e.Cb,173)}function so(e){return e.Db>>16!=11?null:f(e.Cb,27)}function dT(e){return e.Db>>16!=17?null:f(e.Cb,29)}function IP(e){return e.Db>>16!=6?null:f(e.Cb,240)}function Pft(e){return e.Db>>16!=3?null:f(e.Cb,154)}function Lft(e,t){var i=e.a=e.a||[];return i[t]||(i[t]=e.te(t))}function uEn(e,t){var i;i=e.q.getHours(),e.q.setMonth(t),Yj(e,i)}function mTe(e){var t;return ow(e),t=new Mo,Jr(e,new qt(t))}function oJ(){oJ=Z,p1e=new Wn,SV=new Wn,Tgn(YAt,new JC)}function Dft(){this.a=new ub,this.e=new Mo,this.g=0,this.i=0}function Mft(e,t){rX(this),this.f=t,this.g=e,YX(this),this.he()}function jft(e,t){this.a=e,this.c=$a(this.a),this.b=new ZX(t)}function $ft(e,t,i){this.a=t,this.c=e,this.b=(or(i),new Kl(i))}function Fft(e,t,i){this.a=t,this.c=e,this.b=(or(i),new Kl(i))}function Bft(e){this.a=e,this.b=Pe(B6t,Ye,1989,e.e.length,0,2)}function lEn(e){return!e.a&&(e.a=new Be(Ur,e,10,11)),e.a.i>0}function tj(e,t,i,a,u,l){return new v1(e.e,t,e.Ij(),i,a,u,l)}function Os(e,t,i){return t==null?ml(e.f,null,i):MS(e.i,t,i)}function Ia(e,t){e.c&&hl(e.c.g,e),e.c=t,e.c&&Qe(e.c.g,e)}function jo(e,t){e.c&&hl(e.c.a,e),e.c=t,e.c&&Qe(e.c.a,e)}function cc(e,t){e.i&&hl(e.i.j,e),e.i=t,e.i&&Qe(e.i.j,e)}function ua(e,t){e.d&&hl(e.d.e,e),e.d=t,e.d&&Qe(e.d.e,e)}function fT(e,t){if(e<0||e>t)throw ie(new Ru(GOe+e+qOe+t))}function dEn(e,t,i){return Qe(e.a,(VX(),zle(t,i),new dS(t,i))),e}function Gce(e,t){e.a&&hl(e.a.k,e),e.a=t,e.a&&Qe(e.a.k,e)}function qce(e,t){e.b&&hl(e.b.f,e),e.b=t,e.b&&Qe(e.b.f,e)}function nE(e,t,i,a){this.c=e,this.d=a,Gce(this,t),qce(this,i)}function Uft(){Uft=Z,C6t=Il(new _o,(wa(),zs),(ka(),Tq))}function zft(){zft=Z,k6t=Il(new _o,(wa(),zs),(ka(),Tq))}function bTe(){bTe=Z,T6t=Il(new _o,(wa(),zs),(ka(),Tq))}function Hft(){Hft=Z,q6t=Sr(new _o,(wa(),zs),(ka(),z$))}function rf(){rf=Z,K6t=Sr(new _o,(wa(),zs),(ka(),z$))}function Gft(){Gft=Z,Y6t=Sr(new _o,(wa(),zs),(ka(),z$))}function Vce(){Vce=Z,e5t=Sr(new _o,(wa(),zs),(ka(),z$))}function qft(){qft=Z,Y5t=Il(new _o,(yR(),hF),(Wj(),WDe))}function Vft(){Vft=Z,NAt=hr((dY(),ce(se(OAt,1),Ve,534,0,[Bpe])))}function fEn(e){return ole(),mn(),f(e.a,82).d.e!=0}function hEn(e,t){return di(f(e.c,65).c.e.b,f(t.c,65).c.e.b)}function pEn(e,t){return di(f(e.c,65).c.e.a,f(t.c,65).c.e.a)}function Wft(e,t){Awn(e,e.b,e.c),f(e.b.b,68),t&&f(t.b,68).b}function gEn(e,t){VCe(e,t),me(e.Cb,88)&&OT(pd(f(e.Cb,88)),2)}function Wce(e,t){me(e.Cb,88)&&OT(pd(f(e.Cb,88)),4),ju(e,t)}function aJ(e,t){me(e.Cb,184)&&(f(e.Cb,184).tb=null),ju(e,t)}function Kft(e,t){var i;return i=f(ST(tR(e.a),t),16),i?i.gc():0}function mEn(e,t){var i,a;i=t.c,a=i!=null,a&&QN(e,new cT(t.c))}function Yft(e){var t,i;return i=(iP(),t=new cS,t),$P(i,e),i}function Xft(e){var t,i;return i=(iP(),t=new cS,t),$P(i,e),i}function Jft(e){for(var t;;)if(t=e.Pb(),!e.Ob())return t}function yH(e){return ym(),Eu(e,0)>=0?aw(e):X7(aw(T0(e)))}function bEn(e,t){Wst(e,mr(ia(wS(t,24),kQ)),mr(ia(t,kQ)))}function Zft(e,t,i){var a;S0t(t,i,e.c.length),a=i-t,W2e(e.c,t,a)}function Xv(e,t,i){S0t(t,i,e.gc()),this.c=e,this.a=t,this.b=i-t}function nR(e){this.c=new ki,this.b=e.b,this.d=e.c,this.a=e.a}function Kce(e){this.a=h.Math.cos(e),this.b=h.Math.sin(e)}function Bt(e,t){this.b=(an(e),e),this.a=(t&FT)==0?t|64|Sm:t}function Gt(e,t){if(e<0||e>=t)throw ie(new Ru(GOe+e+qOe+t))}function gn(e,t){if(e<0||e>=t)throw ie(new L2e(GOe+e+qOe+t))}function yu(e,t){return is(),_ue(t)?new TX(t,e):new Wz(t,e)}function Qft(e,t,i){return new hde(e,(DP(),Jpe),null,!1,t,i)}function eht(e,t,i){return new hde(e,(DP(),Qpe),t,i,null,!1)}function vEn(e,t,i){var a;a=new ja(i.d),ii(a,e),GCe(t,a.a,a.b)}function vTe(e){var t;return ow(e),t=(TS(),TS(),Kpe),RJ(e,t)}function D3(e){return p7(),me(e.g,9)?f(e.g,9):null}function wEn(){return hj(),ce(se(fPe,1),Ve,323,0,[G$,Aq,dPe])}function yEn(){return FJ(),ce(se(pPe,1),Ve,347,0,[hPe,fte,Mge])}function EEn(){return lE(),ce(se(kNt,1),Ve,446,0,[Fge,lD,SA])}function SEn(){return zP(),ce(se(Yge,1),Ve,298,0,[Wge,Kge,Iq])}function _En(){return w1(),ce(se(RNt,1),Ve,312,0,[Oq,kA,WR])}function kEn(){return Sj(),ce(se(yDe,1),Ve,349,0,[Cme,wDe,Hte])}function TEn(){return cs(),ce(se(b6t,1),Ve,449,0,[sF,td,Uu])}function xEn(){return QJ(),ce(se(PDe,1),Ve,379,0,[NDe,Mme,RDe])}function CEn(){return Ej(),ce(se(LDe,1),Ve,345,0,[$me,jme,Uq])}function AEn(){return JP(),ce(se(MDe,1),Ve,346,0,[Fme,DDe,cF])}function IEn(){return $J(),ce(se(FDe,1),Ve,348,0,[$De,Bme,jDe])}function OEn(){return eZ(),ce(se(BDe,1),Ve,380,0,[Ume,kD,lx])}function NEn(){return LS(),ce(se(xge,1),Ve,365,0,[c_,_w,s_])}function REn(){return yj(),ce(se(rMe,1),Ve,383,0,[nMe,Gme,Gq])}function PEn(){return PS(),ce(se($6t,1),Ve,450,0,[zq,uF,Wte])}function LEn(){return ZH(),ce(se(NMe,1),Ve,384,0,[ine,IMe,OMe])}function DEn(){return JJ(),ce(se(YMe,1),Ve,299,0,[pbe,KMe,WMe])}function MEn(){return wZ(),ce(se(XMe,1),Ve,433,0,[vF,sne,gbe])}function jEn(){return Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])}function $En(){return Yl(),ce(se(gIt,1),Ve,457,0,[Om,Ew,tp])}function FEn(){return pl(),ce(se(bIt,1),Ve,458,0,[jp,Sw,np])}function BEn(){return fh(),ce(se(Ql,1),Ve,131,0,[H6e,Zl,G6e])}function UEn(){return cj(),ce(se(d5e,1),Ve,382,0,[lge,uge,dge])}function zEn(){return gZ(),ce(se(QMe,1),Ve,385,0,[ZMe,Sbe,JMe])}function HEn(){return TJ(),ce(se(y8e,1),Ve,426,0,[Cbe,w8e,v8e])}function GEn(){return lZ(),ce(se(_8e,1),Ve,427,0,[E8e,S8e,Abe])}function qEn(){return rG(),ce(se(Ibe,1),Ve,432,0,[dne,fne,hne])}function VEn(){return _1(),ce(se(r7e,1),Ve,342,0,[kne,W0,PF])}function WEn(){return Sg(),ce(se(q9e,1),Ve,276,0,[PD,gx,LD])}function KEn(){return gR(),ce(se(f7e,1),Ve,278,0,[d7e,bx,One])}function YEn(){return lj(),ce(se(u7e,1),Ve,297,0,[pV,a1e,c7e])}function wg(e){return uc(ce(se(Zo,1),Ye,8,0,[e.i.n,e.n,e.a]))}function XEn(e){return e.a==(RP(),Gne)&&q(e,CPn(e.g,e.b)),e.a}function rR(e){return e.d==(RP(),Gne)&&Xt(e,yDn(e.g,e.b)),e.d}function JEn(e,t){var i;return i=(an(e),e).g,f_e(!!i),an(t),i(t)}function wTe(e,t){npn.call(this,new PN(mT(e))),uf(t,Zkt),this.a=t}function tht(e,t,i){sS.call(this,25),this.b=e,this.a=t,this.c=i}function of(e){Xr(),sS.call(this,e),this.c=!1,this.a=!1}function ZEn(e){if(e.p!=2)throw ie(new Tl);return mr(e.f)&To}function QEn(e){if(e.p!=2)throw ie(new Tl);return mr(e.k)&To}function e2n(e){return e.Db>>16!=6?null:f(Sde(e),240)}function nht(e,t){var i,a;return a=EP(e,t),i=e.a.dd(a),new cot(e,i)}function t2n(e,t,i){var a;return a=Zj(e,t,!1),a.b<=t&&a.a<=i}function rht(e,t,i){var a;a=new B2,a.b=t,a.a=i,++t.b,Qe(e.d,a)}function n2n(e){var t;return t=f(SX(e.b),10),new Rf(e.a,t,e.c)}function r2n(e){return Ua(hb(Nc(wd(e,32)),32),Nc(wd(e,32)))}function ia(e,t){return tw(Qvn(Js(e)?dh(e):e,Js(t)?dh(t):t))}function hm(e,t){return tw(ewn(Js(e)?dh(e):e,Js(t)?dh(t):t))}function Yce(e,t){return tw(twn(Js(e)?dh(e):e,Js(t)?dh(t):t))}function yTe(e,t){return Alt(e.a,t)?mke(e.b,f(t,23).g,null):null}function rE(e){return or(e),me(e,16)?new Kl(f(e,16)):bP(e.Jc())}function i2n(e,t){return uj(),e.c==t.c?di(t.d,e.d):di(e.c,t.c)}function o2n(e,t){return uj(),e.c==t.c?di(e.d,t.d):di(e.c,t.c)}function a2n(e,t){return uj(),e.c==t.c?di(t.d,e.d):di(t.c,e.c)}function s2n(e,t){return uj(),e.c==t.c?di(e.d,t.d):di(t.c,e.c)}function c2n(e,t){e.b=e.b|t.b,e.c=e.c|t.c,e.d=e.d|t.d,e.a=e.a|t.a}function ee(e){return Ln(e.a<e.c.c.length),e.b=e.a++,e.c.c[e.b]}function ETe(e){return e.b==null||e.b.length==0?"n_"+e.a:"n_"+e.b}function EH(e){return f(kg(e,Pe(Ib,j0,9,e.c.length,0,1)),199)}function STe(e,t){return e&&e.equals?e.equals(t):Ae(e)===Ae(t)}function iE(e){return e.c==null||e.c.length==0?"n_"+e.g:"n_"+e.c}function iht(e,t){var i;for(i=e+"";i.length<t;)i="0"+i;return i}function u2n(e,t){var i;i=f(bn(e.g,t),60),Mu(t.d,new Jot(e,i))}function l2n(e,t){var i;i=ke(Oe(e.a.mf((Rr(),Ene)))),_kt(e,t,i)}function d2n(e,t){var i,a;return i=rwt(e),a=rwt(t),i<a?-1:i>a?1:0}function oht(e,t){var i,a;return i=vue(t),a=i,f(bn(e.c,a),17).a}function Xce(e,t,i){var a;a=e.d[t.p],e.d[t.p]=e.d[i.p],e.d[i.p]=a}function f2n(e,t,i){var a;e.n&&t&&i&&(a=new VC,Qe(e.e,a))}function Jce(e,t){if(Go(e.a,t),t.d)throw ie(new oc(STt));t.d=e}function _Te(e,t){this.a=new et,this.d=new et,this.f=e,this.c=t}function aht(){this.c=new jst,this.a=new Hpt,this.b=new lrt,Cot()}function sht(){hR(),this.b=new Wn,this.a=new Wn,this.c=new et}function cht(e,t,i){this.d=e,this.j=t,this.e=i,this.o=-1,this.p=3}function uht(e,t,i){this.d=e,this.k=t,this.f=i,this.o=-1,this.p=5}function lht(e,t,i,a,u,l){VTe.call(this,e,t,i,a,u),l&&(this.o=-2)}function dht(e,t,i,a,u,l){kxe.call(this,e,t,i,a,u),l&&(this.o=-2)}function fht(e,t,i,a,u,l){Txe.call(this,e,t,i,a,u),l&&(this.o=-2)}function hht(e,t,i,a,u,l){qTe.call(this,e,t,i,a,u),l&&(this.o=-2)}function pht(e,t,i,a,u,l){Axe.call(this,e,t,i,a,u),l&&(this.o=-2)}function ght(e,t,i,a,u,l){xxe.call(this,e,t,i,a,u),l&&(this.o=-2)}function mht(e,t,i,a,u,l){Cxe.call(this,e,t,i,a,u),l&&(this.o=-2)}function bht(e,t,i,a,u,l){GTe.call(this,e,t,i,a,u),l&&(this.o=-2)}function vht(e,t,i,a){aY.call(this,i),this.b=e,this.c=t,this.d=a}function wht(e,t){this.f=e,this.a=(RP(),Hne),this.c=Hne,this.b=t}function yht(e,t){this.g=e,this.d=(RP(),Gne),this.a=Gne,this.b=t}function Eht(e,t){Zy.call(this,1,2,ce(se(pr,1),Pr,30,15,[e,t]))}function kTe(e,t){!e.c&&(e.c=new eo(e,0)),hQ(e.c,(fi(),qF),t)}function h2n(e,t){return c6n(e,t,me(t,101)&&(f(t,19).Bb&Xa)!=0)}function p2n(e,t){return tdt(Nc(e.q.getTime()),Nc(t.q.getTime()))}function Sht(e){return tce(e.e.Pd().gc()*e.c.Pd().gc(),16,new Dk(e))}function g2n(e){return!!e.u&&iu(e.u.a).i!=0&&!(e.n&&$le(e.n))}function m2n(e){return!!e.a&&gd(e.a.a).i!=0&&!(e.b&&Fle(e.b))}function TTe(e,t){return t==0?!!e.o&&e.o.f!=0:Rle(e,t)}function _ht(e){return Ln(e.b.b!=e.d.a),e.c=e.b=e.b.b,--e.a,e.c.c}function kht(e){return e.a?e.e.length==0?e.a.a:e.a.a+(""+e.e):e.c}function nj(e){for(;e.d>0&&e.a[--e.d]==0;);e.a[e.d++]==0&&(e.e=0)}function la(e,t){this.a=e,Fk.call(this,e),fT(t,e.gc()),this.b=t}function Tht(e){this.a=Pe(Ro,Jt,1,dCe(h.Math.max(8,e))<<1,5,1)}function xht(e){$ue.call(this,e,(DP(),Xpe),null,!1,null,!1)}function Cht(e,t){var i;return i=1-t,e.a[i]=VJ(e.a[i],i),VJ(e,t)}function Aht(e,t){var i,a;return a=ia(e,ls),i=hb(t,32),hm(i,a)}function b2n(e,t,i){var a;return a=f(e.Zb().xc(t),16),!!a&&a.Gc(i)}function Iht(e,t,i){var a;return a=f(e.Zb().xc(t),16),!!a&&a.Kc(i)}function Oht(e,t,i){var a;a=(or(e),new Kl(e)),XAn(new $ft(a,t,i))}function SH(e,t,i){var a;a=(or(e),new Kl(e)),JAn(new Fft(a,t,i))}function v2n(e,t,i){e.a=t,e.c=i,e.b.a.$b(),nf(e.d),Bk(e.e.a.c,0)}function Nht(e,t){var i;e.e=new k2e,i=RT(t),na(i,e.c),Q2t(e,i,0)}function w2n(e,t){return new Yse(t,Nct($a(t.e),e,e),(mn(),!0))}function y2n(e,t){return lR(),f(V(t,(Ec(),DA)),17).a>=e.gc()}function E2n(e){return rf(),!Fa(e)&&!(!Fa(e)&&e.c.i.c==e.d.i.c)}function pm(e){return f(kg(e,Pe(vge,Jfe,18,e.c.length,0,1)),469)}function S2n(e){M0t((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a),new q8)}function xTe(){var e,t,i;return t=(i=(e=new cS,e),i),Qe(K7e,t),t}function wc(e,t,i,a,u,l){return Hbt(e,t,i,l),jCe(e,a),$Ce(e,u),e}function Rht(e,t,i,a){return e.a+=""+uh(t==null?bl:Qs(t),i,a),e}function _H(e,t){if(e<0||e>=t)throw ie(new Ru(DRn(e,t)));return e}function Pht(e,t,i){if(e<0||t<e||t>i)throw ie(new Ru(rRn(e,t,i)))}function Je(e,t,i,a){var u;u=new s5,u.a=t,u.b=i,u.c=a,Mr(e.b,u)}function po(e,t,i,a){var u;u=new s5,u.a=t,u.b=i,u.c=a,Mr(e.a,u)}function _2n(e,t,i){var a;a=S3n();try{return Nbn(e,t,i)}finally{TSn(a)}}function oE(e){var t;return Js(e)?(t=e,t==-0?0:t):$_n(e)}function Lht(e,t){return me(t,43)?Gle(e.a,f(t,43)):!1}function Dht(e,t){return me(t,43)?Gle(e.a,f(t,43)):!1}function Mht(e,t){return me(t,43)?Gle(e.a,f(t,43)):!1}function k2n(e){return P3(e).dc()?!1:(omn(e,new P),!0)}function sJ(e){var t;return Kv(e),t=new $t,b3(e.a,new dt(t)),t}function CTe(e){var t;return Kv(e),t=new Ie,b3(e.a,new ze(t)),t}function T2n(e){if(!("stack"in e))try{throw e}catch{}return e}function x2n(e,t){return e.a<=e.b?(t.Bd(e.a++),!0):!1}function C2n(e,t,i){e.d&&hl(e.d.e,e),e.d=t,e.d&&Ky(e.d.e,i,e)}function ATe(e,t,i){this.d=new Zet(this),this.e=e,this.i=t,this.f=i}function ITe(e,t,i,a){Fst(this),this.c=e,this.e=t,this.f=i,this.b=a}function rj(e,t){o7n(t,e),ske(e.d),ske(f(V(e,(ot(),Mte)),213))}function Zce(e,t){a7n(t,e),cke(e.d),cke(f(V(e,(ot(),Mte)),213))}function IS(e,t){var i,a;return i=h1(e,t),a=null,i&&(a=i.ne()),a}function OP(e,t){var i,a;return i=gT(e,t),a=null,i&&(a=i.qe()),a}function ij(e,t){var i,a;return i=h1(e,t),a=null,i&&(a=i.qe()),a}function p1(e,t){var i,a;return i=h1(e,t),a=null,i&&(a=xAe(i)),a}function A2n(e,t,i){var a;return a=oL(i),cQ(e.n,a,t),cQ(e.o,t,i),t}function OTe(e,t,i,a){this.d=e,this.n=t,this.g=i,this.o=a,this.p=-1}function jht(e){var t;t=e.Ch(),this.a=me(t,71)?f(t,71).Fi():t.Jc()}function I2n(e){return tce(e.e.Pd().gc()*e.c.Pd().gc(),273,new sz(e))}function cJ(e){return new Du((uf(e,ffe),NJ(Ua(Ua(5,e),e/10|0))))}function $ht(e){return f(kg(e,Pe(GOt,XTt,11,e.c.length,0,1)),1988)}function Fht(e,t,i,a){return me(i,59)?new vct(e,t,i,a):new Ake(e,t,i,a)}function uJ(){uJ=Z,rbe=new SSe("DFS",0),kMe=new SSe("BFS",1)}function lJ(){lJ=Z,Xge=new pSe(PL,0),PPe=new pSe("TOP_LEFT",1)}function Bht(){Bht=Z,kLt=hr((g7(),ce(se(q8e,1),Ve,524,0,[Zq])))}function Uht(){Uht=Z,aLt=hr((xY(),ce(se(oLt,1),Ve,543,0,[Nbe])))}function zht(){zht=Z,iLt=hr((TY(),ce(se(k8e,1),Ve,475,0,[Obe])))}function Hht(){Hht=Z,zDe=Pdt(rt(1),rt(4)),UDe=Pdt(rt(1),rt(2))}function O2n(){return DS(),ce(se(zIt,1),Ve,406,0,[yq,wq,ige,oge])}function N2n(){return qP(),ce(se(ROt,1),Ve,405,0,[a_,YT,KT,bA])}function R2n(){return DP(),ce(se(Pee,1),Ve,305,0,[Xpe,Jpe,Zpe,Qpe])}function P2n(){return hE(),ce(se(g5e,1),Ve,381,0,[XT,p5e,mge,bge])}function L2n(){return nZ(),ce(se(JOt,1),Ve,364,0,[kge,ate,ste,xq])}function NP(e){return me(e,16)?f(e,16).dc():!e.Jc().Ob()}function iR(e){return p7(),me(e.g,152)?f(e.g,152):null}function Ght(e){if(e.e.g!=e.b)throw ie(new If);return!!e.c&&e.d>0}function Zn(e){return Ln(e.b!=e.d.c),e.c=e.b,e.b=e.b.a,++e.a,e.c.c}function NTe(e,t){an(t),io(e.a,e.c,t),e.c=e.c+1&e.a.length-1,lwt(e)}function Jv(e,t){an(t),e.b=e.b-1&e.a.length-1,io(e.a,e.b,t),lwt(e)}function D2n(e,t){var i;i=t.a,Ia(i,t.c.d),ua(i,t.d.d),wT(i.a,e.n)}function RTe(e,t){var i;return i=f(Eg(e.b,t),67),!i&&(i=new ki),i}function M2n(e){var t;return e.j==(tt(),Bn)&&(t=zEt(e),Cl(t,_n))}function j2n(){return _g(),ce(se(cNt,1),Ve,410,0,[VR,rD,iD,Tge])}function $2n(){return xT(),ce(se(Ime,1),Ve,203,0,[Gte,Ame,PA,RA])}function F2n(){return Pp(),ce(se(IDe,1),Ve,266,0,[Pm,ADe,Rme,Pme])}function B2n(){return yR(),ce(se(qDe,1),Ve,396,0,[Jte,fF,hF,pF])}function qht(e,t){return f(Dd(oT(f(si(e.k,t),15).Mc(),wA)),112)}function Vht(e,t){return f(Dd(XN(f(si(e.k,t),15).Mc(),wA)),112)}function U2n(e){return new Bt(pkn(f(e.a.kd(),16).gc(),e.a.jd()),16)}function z2n(e,t){return UN(new Ke(t.e.a+t.f.a/2,t.e.b+t.f.b/2),e)}function Wht(e,t){var i;for(i=e.j.c.length;i<t;i++)Qe(e.j,e.Lg())}function Kht(e,t,i,a){var u;return u=a[t.g][i.g],ke(Oe(V(e.a,u)))}function Yht(e,t,i,a){this.e=null,this.c=e,this.d=t,this.a=i,this.b=a}function PTe(e,t,i,a,u){this.i=e,this.a=t,this.e=i,this.j=a,this.f=u}function H2n(e){e.a=null,e.e=null,Bk(e.b.c,0),Bk(e.f.c,0),e.c=null}function oj(){oj=Z,dF=new bSe("UPPER",0),lF=new bSe("LOWER",1)}function G2n(){return wj(),ce(se(_Me,1),Ve,350,0,[nbe,nne,tbe,ebe])}function q2n(){return RZ(),ce(se(VMe,1),Ve,351,0,[hbe,GMe,qMe,HMe])}function V2n(){return zZ(),ce(se(xPt,1),Ve,398,0,[wbe,mbe,vbe,bbe])}function W2n(){return Ud(),ce(se($F,1),Ve,378,0,[mV,GE,gV,mx])}function K2n(){return XP(),ce(se(G9e,1),Ve,275,0,[cV,_ne,z9e,H9e])}function Y2n(){return k1(),ce(se(i1e,1),Ve,220,0,[r1e,uV,DD,c4])}function X2n(){return df(),ce(se(mDt,1),Ve,289,0,[dV,Rb,Nw,lV])}function J2n(){return pZ(),ce(se(v7e,1),Ve,321,0,[c1e,g7e,b7e,m7e])}function Z2n(){return iZ(),ce(se(LDt,1),Ve,404,0,[u1e,y7e,w7e,E7e])}function Q2n(e){return oJ(),Yu(p1e,e)?f(bn(p1e,e),337).Og():null}function Pf(e,t,i){return t<0?wde(e,i):f(i,69).tk().yk(e,e.di(),t)}function eSn(e,t,i){var a;return a=oL(i),cQ(e.f,a,t),qr(e.g,t,i),t}function tSn(e,t,i){var a;return a=oL(i),cQ(e.p,a,t),qr(e.q,t,i),t}function Xht(e){var t,i;return t=(Bv(),i=new WC,i),e&&QZ(t,e),t}function LTe(e){var t;return t=e.Zi(e.i),e.i>0&&au(e.g,0,t,0,e.i),t}function g1(e){var t;for(t=0;e.Ob();)e.Pb(),t=Ua(t,1);return NJ(t)}function Jht(e,t){var i;return i=new RN,e.Ed(i),i.a+="..",t.Fd(i),i.a}function Zht(e,t){CY();var i;return i=f(bn(Fne,e),58),!i||i.ck(t)}function nSn(e){if(e.p!=1)throw ie(new Tl);return mr(e.f)<<24>>24}function rSn(e){if(e.p!=1)throw ie(new Tl);return mr(e.k)<<24>>24}function iSn(e){if(e.p!=7)throw ie(new Tl);return mr(e.f)<<16>>16}function oSn(e){if(e.p!=7)throw ie(new Tl);return mr(e.k)<<16>>16}function M3(e,t){return t.e==0||e.e==0?M$:(vL(),xde(e,t))}function Qht(e,t,i){e.i=0,e.e=0,t!=i&&(P1t(e,t,i),R1t(e,t,i))}function aSn(e,t,i,a,u){var l;l=S6n(u,i,a),Qe(t,CRn(u,l)),ENn(e,u,t)}function sSn(e,t,i){var a;a=f(bn(e.g,i),60),Qe(e.a.c,new Ya(t,a))}function cSn(e,t,i){return lce(Oe(ac(Is(e.f,t))),Oe(ac(Is(e.f,i))))}function uSn(e,t){return f(t==null?ac(Is(e.f,null)):w7(e.i,t),286)}function ept(e,t){return Ae(t)===Ae(e)?"(this Map)":t==null?bl:Qs(t)}function DTe(e,t){return e==(cn(),Yi)&&t==Yi?4:e==Yi||t==Yi?8:32}function dJ(e,t,i){return lQ(e,t,i,me(t,101)&&(f(t,19).Bb&Xa)!=0)}function lSn(e,t,i){return EL(e,t,i,me(t,101)&&(f(t,19).Bb&Xa)!=0)}function dSn(e,t,i){return g6n(e,t,i,me(t,101)&&(f(t,19).Bb&Xa)!=0)}function mb(e,t){is();var i;return i=f(e,69).sk(),LNn(i,t),i.ul(t)}function tpt(e,t){var i;for(i=t;i;)eT(e,i.i,i.j),i=so(i);return e}function npt(e,t){var i;for(i=0;i<t;++i)io(e,i,new YK(f(e[i],43)))}function m1(e,t){ym(),Zy.call(this,e,1,ce(se(pr,1),Pr,30,15,[t]))}function hT(e,t){Xr(),sS.call(this,e),this.a=t,this.c=-1,this.b=-1}function rpt(e,t){uTe.call(this),this.a=e,this.b=t,Qe(this.a.b,this)}function aj(e){return Ln(e.a<e.c.a.length),e.b=e.a,mmt(e),e.c.b[e.b]}function iu(e){return e.n||(pd(e),e.n=new Llt(e,gs,e),Bs(e)),e.n}function ipt(e){if(e<0)throw ie(new nit("Negative array size: "+e))}function Qce(e,t,i){if(i){var a=i.me();i=a(i)}else i=void 0;e.a[t]=i}function fSn(e,t,i){if(i){var a=i.me();e.a[t]=a(i)}else delete e.a[t]}function opt(e,t){WP();var i;return i=e.j.g-t.j.g,i!=0?i:0}function hSn(e,t){return b0(),Qn(je(e.a),t)}function pSn(e,t){return b0(),Qn(je(e.a),t)}function MTe(e,t){var i;i=e.q.getHours(),e.q.setFullYear(t+hw),Yj(e,i)}function gSn(e,t){var i;return i=t!=null?wu(e,t):ac(Is(e.f,t)),QY(i)}function mSn(e,t){var i;return i=t!=null?wu(e,t):ac(Is(e.f,t)),QY(i)}function eue(e){return e.c&&e.d?ETe(e.c)+"->"+ETe(e.d):"e_"+bS(e)}function jTe(e){e.b!=e.c&&(e.a=Pe(Ro,Jt,1,8,5,1),e.b=0,e.c=0)}function fJ(e){var t;return e?new wke(e):(t=new ub,jue(t,e),t)}function bSn(e,t){var i;return ipt(t),i=e.slice(0,t),i.length=t,bue(i,e)}function vSn(e,t){var i;for(i=e.d-1;i>=0&&e.a[i]===t[i];i--);return i<0}function pT(e,t,i,a){cht.call(this,1,i,a),this.c=e,this.b=t}function tue(e,t,i,a){uht.call(this,1,i,a),this.c=e,this.b=t}function nue(e,t,i,a,u,l,g){Oue.call(this,t,a,u,l,g),this.c=e,this.a=i}function aE(e,t,i){this.e=e,this.a=Ro,this.b=ySt(t),this.c=t,this.d=i}function rue(e){this.e=e,this.c=this.e.a,this.b=this.e.g,this.d=this.e.i}function $Te(e){this.d=e,this.b=this.d.a.entries(),this.a=this.b.next()}function apt(e){this.c=e,this.a=f(ph(e),155),this.b=this.a.gk().si()}function Zv(){Wn.call(this),_ct(this),this.d.b=this.d,this.d.a=this.d}function Wi(e,t,i,a){var u;u=new An,u.c=t,u.b=i,u.a=a,a.b=i.a=u,++e.b}function $d(e,t){var i;return t.b.Kb(tmt(e,t.c.Ve(),(i=new qn(t),i)))}function hJ(e){lAe(),Wst(this,mr(ia(wS(e,24),kQ)),mr(ia(e,kQ)))}function spt(){spt=Z,ZIt=hr((OJ(),ce(se(X6e,1),Ve,421,0,[age,jee])))}function cpt(){cpt=Z,KOt=hr((BP(),ce(se(WOt,1),Ve,501,0,[kq,yge])))}function upt(){upt=Z,wNt=hr((SJ(),ce(se(lPe,1),Ve,419,0,[uPe,Dge])))}function lpt(){lpt=Z,_Nt=hr((FH(),ce(se(vPe,1),Ve,416,0,[hte,bPe])))}function dpt(){dpt=Z,LNt=hr((lJ(),ce(se(LPe,1),Ve,417,0,[Xge,PPe])))}function fpt(){fpt=Z,CNt=hr((RH(),ce(se(EPe,1),Ve,474,0,[yPe,gte])))}function hpt(){hpt=Z,MNt=hr((pJ(),ce(se(KPe,1),Ve,420,0,[ame,ome])))}function ppt(){ppt=Z,m6t=hr((MH(),ce(se(ODe,1),Ve,418,0,[Lme,Dme])))}function gpt(){gpt=Z,G6t=hr((oj(),ce(se(H6t,1),Ve,516,0,[dF,lF])))}function mpt(){mpt=Z,Q6t=hr((vg(),ce(se(Z6t,1),Ve,506,0,[$E,Mg])))}function bpt(){bpt=Z,J6t=hr((fm(),ce(se(X6t,1),Ve,507,0,[y_,G0])))}function vpt(){vpt=Z,p5t=hr((Ip(),ce(se(h5t,1),Ve,513,0,[dx,Cw])))}function wpt(){wpt=Z,E5t=hr((AS(),ce(se(y5t,1),Ve,451,0,[Aw,LA])))}function ypt(){ypt=Z,K5t=hr((uJ(),ce(se(TMe,1),Ve,422,0,[rbe,kMe])))}function Ept(){Ept=Z,Q5t=hr((GJ(),ce(se(xMe,1),Ve,485,0,[rne,jA])))}function Spt(){Spt=Z,nPt=hr((kJ(),ce(se(AMe,1),Ve,423,0,[CMe,ube])))}function _pt(){_pt=Z,sLt=hr((CH(),ce(se(T8e,1),Ve,425,0,[Rbe,pne])))}function kpt(){kpt=Z,TLt=hr((XH(),ce(se(W8e,1),Ve,424,0,[mne,V8e])))}function Tpt(){Tpt=Z,ZPt=hr((gJ(),ce(se(b8e,1),Ve,476,0,[xbe,m8e])))}function pJ(){pJ=Z,ame=new gSe(Lp,0),ome=new gSe("ALTERNATING",1)}function gJ(){gJ=Z,xbe=new TSe(cNe,0),m8e=new TSe("TARGET_WIDTH",1)}function mJ(){mJ=Z,Lg=!0,nIt=!1,rIt=!1,oIt=!1,iIt=!1}function xpt(){xpt=Z,NDt=hr((Jse(),ce(se(Nne,1),Ve,343,0,[h7e,p7e])))}function wSn(){return ko(),ce(se(NF,1),Ve,86,0,[$g,Gs,Ds,jg,Wf])}function ySn(){return tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])}function ESn(e,t,i){return f(t==null?ml(e.f,null,i):MS(e.i,t,i),286)}function SSn(e){return(e.k==(cn(),Yi)||e.k==bo)&&li(e,(Ge(),K$))}function Cpt(e,t){var i;return i=t.jd(),new dS(i,e.e.pc(i,f(t.kd(),16)))}function Apt(e,t){var i,a;return i=t.jd(),a=e.De(i),!!a&&Pu(a.e,t.kd())}function Oa(e,t){var i,a;for(an(t),a=e.Jc();a.Ob();)i=a.Pb(),t.Ad(i)}function _Sn(e,t){var i,a;a=!1;do i=_1t(e,t),a=a|i;while(i);return a}function FTe(e,t){var i,a;for(i=t,a=0;i>0;)a+=e.a[i],i-=i&-i;return a}function Ipt(e,t){var i;for(i=t;i;)eT(e,-i.i,-i.j),i=so(i);return e}function af(e,t,i){var a;return a=(Gt(t,e.c.length),e.c[t]),e.c[t]=i,a}function BTe(e,t,i){e.a.c.length=0,ZMn(e,t,i),e.a.c.length==0||ILn(e,t)}function kH(e){e.i=0,Pz(e.b,null),Pz(e.c,null),e.a=null,e.e=null,++e.g}function kSn(e,t){var i;return i=e.a.get(t),i??Pe(Ro,Jt,1,0,5,1)}function Fa(e){return!e.c||!e.d?!1:!!e.c.i&&e.c.i==e.d.i}function iue(e){mJ(),!Lg&&(this.c=e,this.e=!0,this.a=new et)}function Opt(e){Ukt(),Gnt(this),this.a=new ki,yCe(this,e),Mr(this.a,e)}function Npt(){kse(this),this.b=new Ke(Gi,Gi),this.a=new Ke(qo,qo)}function Rpt(e,t){this.c=0,this.b=t,nst.call(this,e,17493),this.a=this.c}function UTe(e){ggn.call(this,e==null?bl:Qs(e),me(e,80)?f(e,80):null)}function TSn(e){e&&B_n((O2e(),g6e)),--Cee,e&&Aee!=-1&&(pgn(Aee),Aee=-1)}function j3(e,t){return(ow(e),oP(new Ft(e,new dxe(t,e.a)))).zd(zR)}function xSn(){return wa(),ce(se(f5e,1),Ve,360,0,[rp,Ab,su,cu,zs])}function CSn(){return CT(),ce(se(lNt,1),Ve,368,0,[Cq,lte,dte,ute,cte])}function ASn(){return BZ(),ce(se(ePt,1),Ve,361,0,[sbe,ibe,cbe,obe,abe])}function ISn(){return Hj(),ce(se(tLt,1),Ve,366,0,[$A,n4,kF,_F,Jq])}function OSn(){return bG(),ce(se(I8e,1),Ve,325,0,[x8e,Pbe,A8e,Lbe,C8e])}function NSn(){return Nj(),ce(se(SDe,1),Ve,324,0,[EDe,Ome,Nme,oF,aF])}function RSn(){return R0(),ce(se(mPe,1),Ve,344,0,[jge,gPe,$ge,uD,cD])}function PSn(){return Ff(),ce(se(WPe,1),Ve,171,0,[Lq,X$,Tw,J$,tx])}function LSn(){return wm(),ce(se(Ob,1),Ve,170,0,[en,Ao,Dm,FE,L1])}function DSn(){return K3(),ce(se(DF,1),Ve,255,0,[Rw,fV,i7e,LF,o7e])}function MSn(e){return hY(),function(){return _2n(e,this,arguments)}}function zTe(e,t){return me(t,149)?Mt(e.c,f(t,149).c):!1}function Ppt(e){var t;return e.c||(t=e.r,me(t,88)&&(e.c=f(t,29))),e.c}function pd(e){return e.t||(e.t=new Lnt(e),xj(new eit(e),0,e.t)),e.t}function jSn(e){return e.e=3,e.d=e.Yb(),e.e!=2?(e.e=0,!0):!1}function oue(e,t){var i;return i=f(Eg(e.d,t),21),i||f(Eg(e.e,t),21)}function $Sn(e,t){var i,a;i=f(Oxn(e.c,t),16),i&&(a=i.gc(),i.$b(),e.d-=a)}function FSn(e){var t,i,a,u;for(i=e,a=0,u=i.length;a<u;++a)t=i[a],Q7(t)}function aue(e){var t,i,a;return t=e&Ed,i=e>>22&Ed,a=e<0?A1:0,Ku(t,i,a)}function Lpt(e){var t;return t=e.length,Mt(wn.substr(wn.length-t,t),e)}function Dn(e){if(Kn(e))return e.c=e.a,e.a.Pb();throw ie(new ic)}function oR(e,t){return t==0||e.e==0?e:t>0?tvt(e,t):B2t(e,-t)}function HTe(e,t){return t==0||e.e==0?e:t>0?B2t(e,t):tvt(e,-t)}function Dpt(e){this.b=e,GN.call(this,e),this.a=f(fn(this.b.a,4),128)}function Mpt(e){this.b=e,Nn.call(this,e),this.a=f(fn(this.b.a,4),128)}function Vh(e,t,i,a,u){Fpt.call(this,t,a,u),this.c=e,this.b=i}function GTe(e,t,i,a,u){Fpt.call(this,t,a,u),this.c=e,this.a=i}function qTe(e,t,i,a,u){cht.call(this,t,a,u),this.c=e,this.a=i}function VTe(e,t,i,a,u){uht.call(this,t,a,u),this.c=e,this.a=i}function WTe(e,t,i,a){var u;for(u=0;u<Dee;u++)DX(e.a[u][t.g],i,a[t.g])}function KTe(e,t,i,a){var u;for(u=0;u<nge;u++)LX(e.a[t.g][u],i,a[t.g])}function sue(e,t,i){var a;return a=e.Eh(t),a>=0?e.Hh(a,i,!0):GS(e,t,i)}function BSn(e,t,i){return di(UN(sL(e),$a(t.b)),UN(sL(e),$a(i.b)))}function USn(e,t,i){return di(UN(sL(e),$a(t.e)),UN(sL(e),$a(i.e)))}function zSn(e,t){return h.Math.min(Qv(t.a,e.d.d.c),Qv(t.b,e.d.d.c))}function jpt(e,t,i){var a;a=new WSe(e.a),mj(a,e.a.a),ml(a.f,t,i),e.a.a=a}function cue(e){var t;return t=new xS,Rc(t,e),$e(t,(ot(),Ls),null),t}function $pt(e){var t,i;return t=e.c.i,i=e.d.i,t.k==(cn(),bo)&&i.k==bo}function sj(e){var t,i;++e.j,t=e.g,i=e.i,e.g=null,e.i=0,e.Li(i,t),e.Ki()}function TH(e,t){e.Yi(e.i+1),j7(e,e.i,e.Wi(e.i,t)),e.Ji(e.i++,t),e.Ki()}function si(e,t){var i;return i=f(e.c.xc(t),16),!i&&(i=e.ic(t)),e.pc(t,i)}function YTe(e,t){var i,a;return i=(an(e),e),a=(an(t),t),i==a?0:i<a?-1:1}function HSn(e,t){var i;i=e.q.getHours()+(t/60|0),e.q.setMinutes(t),Yj(e,i)}function GSn(e){var t;return t=(or(e),e?new Kl(e):bP(e.Jc())),vG(t),cZ(t)}function gm(e){var t,i;return or(e),t=Wwn(e.length),i=new Du(t),fCe(i,e),i}function _0(e,t){var i;return i=(Gt(t,e.c.length),e.c[t]),W2e(e.c,t,1),i}function b1(e,t){return e.a?xr(e.a,e.b):e.a=new ef(e.d),P7(e.a,t),e}function XTe(e,t){if(e<0||e>t)throw ie(new Ru(DAe(e,t,"index")));return e}function uue(e,t,i,a){var u;return u=Pe(pr,Pr,30,t,15,1),$On(u,e,t,i,a),u}function JTe(e){var t;return t=e.e+e.f,isNaN(t)&&gX(e.d)?e.d:t}function RP(){RP=Z;var e,t;Hne=(iP(),t=new sY,t),Gne=(e=new pae,e)}function bJ(){bJ=Z,TF=new Si("org.eclipse.elk.labels.labelManager")}function Ip(){Ip=Z,dx=new ySe("REGULAR",0),Cw=new ySe("CRITICAL",1)}function yg(e){this.c=e,this.a=new ne(this.c.a),this.b=new ne(this.c.b)}function Fpt(e,t,i){this.d=e,this.k=t?1:0,this.f=i?1:0,this.o=-1,this.p=0}function Bpt(e,t,i){aY.call(this,i),this.b=e,this.c=t,this.d=(_le(),w1e)}function Upt(e,t,i){this.a=e,this.c=t,this.d=i,Qe(t.e,this),Qe(i.b,this)}function vJ(){this.e=new et,this.c=new et,this.d=new et,this.b=new et}function zpt(){this.g=new l2e,this.b=new l2e,this.a=new et,this.k=new et}function Hpt(){this.a=new p2e,this.b=new grt,this.d=new C2,this.e=new Ym}function Gpt(e,t){nst.call(this,t.xd(),t.wd()&-6),an(e),this.a=e,this.b=t}function qpt(e,t){tst.call(this,t.xd(),t.wd()&-6),an(e),this.a=e,this.b=t}function ZTe(e,t){eX.call(this,t.xd(),t.wd()&-6),an(e),this.a=e,this.b=t}function wJ(e,t,i){this.a=e,this.b=t,this.c=i,Qe(e.t,this),Qe(t.i,this)}function xH(){this.b=new ki,this.a=new ki,this.b=new ki,this.a=new ki}function qSn(e,t){var i;return i=d8n(e,t),e.b=new YJ(i.c.length),OMn(e,i)}function VSn(e,t,i){var a;return++e.e,--e.f,a=f(e.d[t].ed(i),135),a.kd()}function Vpt(e){var t;return e.a||(t=e.r,me(t,155)&&(e.a=f(t,155))),e.a}function QTe(e){if(e.a){if(e.e)return QTe(e.e)}else return e;return null}function WSn(e,t){return e.p<t.p?1:e.p>t.p?-1:0}function Wpt(e,t){return Yu(e.a,t)?(aR(e.a,t),!0):!1}function lue(e){var t;return t=e.b,t.b==0?null:f(Xh(t,0),65).b}function KSn(e){var t,i;return t=e.jd(),i=f(e.kd(),16),sH(i.Lc(),new az(t))}function yJ(e,t){return an(t),e.c<e.d?(e.Qe(t,e.c++),!0):!1}function YSn(e){var t;return ow(e),t=new Kut(e,e.a.e,e.a.d|4),new w_e(e,t)}function Kpt(e){var t;for(Kv(e),t=0;e.a.zd(new Ko);)t=Ua(t,1);return t}function aR(e,t){return ta(t)?t==null?ZAe(e.f,null):t1t(e.i,t):ZAe(e.f,t)}function XSn(e,t){t.Sg("Label management",1),QY(V(e,(bJ(),TF))),t.Tg()}function sR(e,t,i){return _H(t,e.e.Pd().gc()),_H(i,e.c.Pd().gc()),e.a[t][i]}function exe(e,t,i){var a,u;for(a=0,u=0;u<t.length;u++)a+=e.rg(t[u],a,i)}function JSn(e,t){var i;e.C&&(i=f(Es(e.b,t),126).n,i.d=e.C.d,i.a=e.C.a)}function txe(e){var t,i,a,u;u=e.d,t=e.a,i=e.b,a=e.c,e.d=i,e.a=a,e.b=u,e.c=t}function EJ(e,t,i,a){this.f=e,this.e=t,this.d=i,this.b=a,this.c=a?a.d:null}function Du(e){kse(this),uH(e>=0,"Initial capacity must not be negative")}function Ypt(){Ypt=Z,iPe=new Mi("separateLayerConnections",(nZ(),kge))}function CH(){CH=Z,Rbe=new CSe("FIXED",0),pne=new CSe("CENTER_NODE",1)}function SJ(){SJ=Z,uPe=new dSe("QUADRATIC",0),Dge=new dSe("SCANLINE",1)}function Xpt(){Xpt=Z,mNt=hr((LS(),ce(se(xge,1),Ve,365,0,[c_,_w,s_])))}function Jpt(){Jpt=Z,yNt=hr((hj(),ce(se(fPe,1),Ve,323,0,[G$,Aq,dPe])))}function Zpt(){Zpt=Z,ENt=hr((FJ(),ce(se(pPe,1),Ve,347,0,[hPe,fte,Mge])))}function Qpt(){Qpt=Z,TNt=hr((lE(),ce(se(kNt,1),Ve,446,0,[Fge,lD,SA])))}function egt(){egt=Z,v6t=hr((cs(),ce(se(b6t,1),Ve,449,0,[sF,td,Uu])))}function tgt(){tgt=Z,w6t=hr((QJ(),ce(se(PDe,1),Ve,379,0,[NDe,Mme,RDe])))}function ngt(){ngt=Z,y6t=hr((Ej(),ce(se(LDe,1),Ve,345,0,[$me,jme,Uq])))}function rgt(){rgt=Z,d6t=hr((Sj(),ce(se(yDe,1),Ve,349,0,[Cme,wDe,Hte])))}function igt(){igt=Z,E6t=hr((JP(),ce(se(MDe,1),Ve,346,0,[Fme,DDe,cF])))}function ogt(){ogt=Z,S6t=hr(($J(),ce(se(FDe,1),Ve,348,0,[$De,Bme,jDe])))}function agt(){agt=Z,_6t=hr((eZ(),ce(se(BDe,1),Ve,380,0,[Ume,kD,lx])))}function sgt(){sgt=Z,k5t=hr((yj(),ce(se(rMe,1),Ve,383,0,[nMe,Gme,Gq])))}function cgt(){cgt=Z,rPt=hr((ZH(),ce(se(NMe,1),Ve,384,0,[ine,IMe,OMe])))}function ugt(){ugt=Z,PNt=hr((w1(),ce(se(RNt,1),Ve,312,0,[Oq,kA,WR])))}function lgt(){lgt=Z,NNt=hr((zP(),ce(se(Yge,1),Ve,298,0,[Wge,Kge,Iq])))}function dgt(){dgt=Z,kPt=hr((JJ(),ce(se(YMe,1),Ve,299,0,[pbe,KMe,WMe])))}function fgt(){fgt=Z,TPt=hr((wZ(),ce(se(XMe,1),Ve,433,0,[vF,sne,gbe])))}function hgt(){hgt=Z,rLt=hr((rG(),ce(se(Ibe,1),Ve,432,0,[dne,fne,hne])))}function pgt(){pgt=Z,QPt=hr((TJ(),ce(se(y8e,1),Ve,426,0,[Cbe,w8e,v8e])))}function ggt(){ggt=Z,eLt=hr((lZ(),ce(se(_8e,1),Ve,427,0,[E8e,S8e,Abe])))}function mgt(){mgt=Z,F6t=hr((PS(),ce(se($6t,1),Ve,450,0,[zq,uF,Wte])))}function bgt(){bgt=Z,OPt=hr((gZ(),ce(se(QMe,1),Ve,385,0,[ZMe,Sbe,JMe])))}function vgt(){vgt=Z,aIt=hr((fh(),ce(se(Ql,1),Ve,131,0,[H6e,Zl,G6e])))}function wgt(){wgt=Z,pIt=hr((Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])))}function ygt(){ygt=Z,mIt=hr((Yl(),ce(se(gIt,1),Ve,457,0,[Om,Ew,tp])))}function Egt(){Egt=Z,vIt=hr((pl(),ce(se(bIt,1),Ve,458,0,[jp,Sw,np])))}function Sgt(){Sgt=Z,kOt=hr((cj(),ce(se(d5e,1),Ve,382,0,[lge,uge,dge])))}function _gt(){_gt=Z,gDt=hr((_1(),ce(se(r7e,1),Ve,342,0,[kne,W0,PF])))}function kgt(){kgt=Z,dDt=hr((Sg(),ce(se(q9e,1),Ve,276,0,[PD,gx,LD])))}function Tgt(){Tgt=Z,ODt=hr((gR(),ce(se(f7e,1),Ve,278,0,[d7e,bx,One])))}function xgt(){xgt=Z,CDt=hr((lj(),ce(se(u7e,1),Ve,297,0,[pV,a1e,c7e])))}function mm(e,t){return!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),Ele(e.o,t)}function ZSn(e){return!e.g&&(e.g=new Ys),!e.g.d&&(e.g.d=new Nnt(e)),e.g.d}function QSn(e){return!e.g&&(e.g=new Ys),!e.g.a&&(e.g.a=new Rnt(e)),e.g.a}function e_n(e){return!e.g&&(e.g=new Ys),!e.g.b&&(e.g.b=new Ont(e)),e.g.b}function AH(e){return!e.g&&(e.g=new Ys),!e.g.c&&(e.g.c=new Pnt(e)),e.g.c}function t_n(e,t,i,a){return i&&(a=i.Nh(t,Pi(i.zh(),e.c.rk()),null,a)),a}function n_n(e,t,i,a){return i&&(a=i.Ph(t,Pi(i.zh(),e.c.rk()),null,a)),a}function due(e,t,i,a){var u;return u=Pe(pr,Pr,30,t+1,15,1),lDn(u,e,t,i,a),u}function Pe(e,t,i,a,u,l){var g;return g=twt(u,a),u!=10&&ce(se(e,l),t,i,u,g),g}function r_n(e,t,i){var a,u;for(u=new GP(t,e),a=0;a<i;++a)AZ(u);return u}function fue(e,t,i){var a,u;if(i!=null)for(a=0;a<t;++a)u=i[a],e.Ni(a,u)}function nxe(e,t){var i;return i=new on,i.c=!0,i.d=t.kd(),tkt(e,t.jd(),i)}function i_n(e,t){var i;i=e.q.getHours()+(t/3600|0),e.q.setSeconds(t),Yj(e,i)}function rxe(e,t){var i,a;return i=t,a=Dce(bP(new wue(e,i))),CX(new wue(e,i)),a}function o_n(e,t,i,a){r_t(e,t,i,EL(e,t,a,me(t,101)&&(f(t,19).Bb&Xa)!=0))}function ixe(e,t,i){f(e.b,68),f(e.b,68),f(e.b,68),Mu(e.a,new Put(i,t,e))}function ga(e,t,i){if(e<0||t>i||t<e)throw ie(new L2e(TQ+e+HOe+t+BOe+i))}function hue(e){e?(this.c=e,this.b=null):(this.c=null,this.b=new et)}function oxe(e){w1t.call(this,e,0),_ct(this),this.d.b=this.d,this.d.a=this.d}function pue(e,t){AY.call(this,e,t),this.a=Pe(X7n,yQ,435,2,0,1),this.b=!0}function axe(e){this.e=e,this.b=this.e.a.entries(),this.a=Pe(Ro,Jt,1,0,5,1)}function Cgt(){Cgt=Z,x6t=Il(Sr(new _o,(wa(),rp),(ka(),Ege)),zs,Tq)}function a_n(){return gG(),ce(se(xPe,1),Ve,279,0,[mte,_Pe,TPe,SPe,kPe,qge])}function s_n(){return fL(),ce(se(RPe,1),Ve,280,0,[q$,APe,NPe,OPe,IPe,CPe])}function c_n(){return qZ(),ce(se(wPe,1),Ve,281,0,[Uge,Bge,Hge,zge,Gge,pte])}function u_n(){return ER(),ce(se(sD,1),Ve,231,0,[aD,H$,oD,QT,EA,yA])}function l_n(){return vb(),ce(se(Z8e,1),Ve,254,0,[Bbe,Qq,eV,wne,bne,vne])}function d_n(){return mE(),ce(se(Gbe,1),Ve,296,0,[Hbe,IF,AF,zbe,xF,CF])}function f_n(){return HZ(),ce(se(J9e,1),Ve,322,0,[o1e,K9e,X9e,V9e,Y9e,W9e])}function h_n(){return oa(),ce(se(a7e,1),Ve,100,0,[Pw,Fg,MD,HE,Lb,uu])}function p_n(){return Xl(),ce(se(Tne,1),Ve,277,0,[Lw,M1,hV,jF,MF,u4])}function Agt(e,t,i){Xr();var a;return a=dw(e,t),i&&a&&Pwn(e)&&(a=null),a}function Op(e,t){return bH(t,"Horizontal alignment cannot be null"),e.b=t,e}function Qn(e,t){return e.Pi()&&e.Gc(t)?!1:(e.Ei(t),!0)}function sxe(e){var t;return!e.a&&e.b!=-1&&(t=e.c.zh(),e.a=Vt(t,e.b)),e.a}function Qv(e,t){var i,a;return i=e.a-t.a,a=e.b-t.b,h.Math.sqrt(i*i+a*a)}function cxe(e,t,i){var a;e.d[t.g]=i,a=e.g.c,a[t.g]=h.Math.max(a[t.g],i+1)}function uxe(e,t,i){var a;a=e.b[i.c.p][i.p],a.b+=t.b,a.c+=t.c,a.a+=t.a,++a.a}function k0(e,t,i){var a,u;for(u=e.Jc();u.Ob();)a=f(u.Pb(),36),wL(a,t,i)}function lxe(e,t){var i,a;for(a=t.Jc();a.Ob();)i=f(a.Pb(),36),OSt(e,i,0,0)}function g_n(e){var t,i;for(i=tr(e.a,0);i.b!=i.d.c;)t=f(Zn(i),65),zAe(t)}function m_n(e,t){var i;return i=bn(e.k,t),vIe(e,t),Lde(e,t),cIe(e,t,i),null}function PP(e,t){var i;return i=e.Eh(t),i>=0?e.Hh(i,!0,!0):GS(e,t,!0)}function IH(e,t){var i,a,u;return u=e.r,a=e.d,i=Zj(e,t,!0),i.b!=u||i.a!=a}function Igt(e,t){return Dot(e.e,t)||pE(e.e,t,new xvt(t)),f(Eg(e.e,t),112)}function sf(e,t,i,a){return an(e),an(t),an(i),an(a),new jke(e,t,new br)}function OH(e,t,i){var a,u;return u=(a=gL(e.b,t),a),u?pQ(LH(e,u),i):null}function b_n(e,t,i){var a,u,l;a=h1(e,i),u=null,a&&(u=xAe(a)),l=u,kvt(t,i,l)}function v_n(e,t,i){var a,u,l;a=h1(e,i),u=null,a&&(u=xAe(a)),l=u,kvt(t,i,l)}function Ol(e,t,i,a){this.Zj(),this.a=t,this.b=e,this.c=new Rke(this,t,i,a)}function gue(e,t,i,a,u,l){OTe.call(this,t,a,u,l),this.c=e,this.b=i}function NH(e,t,i,a,u,l){OTe.call(this,t,a,u,l),this.c=e,this.a=i}function dxe(e,t){eX.call(this,t.xd(),t.wd()&-16449),an(e),this.a=e,this.c=t}function w_n(e,t){e.a.Le(t.d,e.b)>0&&(Qe(e.c,new ike(t.c,t.d,e.d)),e.b=t.d)}function mue(e){e.a=Pe(pr,Pr,30,e.b+1,15,1),e.c=Pe(pr,Pr,30,e.b,15,1),e.d=0}function y_n(e){_J(),f(e.mf((Rr(),px)),181).Ec((Xl(),hV)),e.of(e1e,null)}function _J(){_J=Z,ALt=new K8,OLt=new UC,ILt=sxn((Rr(),e1e),ALt,Ow,OLt)}function kJ(){kJ=Z,CMe=new kSe("LEAF_NUMBER",0),ube=new kSe("NODE_SIZE",1)}function cR(e){var t;return or(e),me(e,204)?(t=f(e,204),t):new Af(e)}function E_n(e){for(;!e.a;)if(!vut(e.c,new It(e)))return!1;return!0}function S_n(e){var t;if(!e.a)throw ie(new Udt);return t=e.a,e.a=so(e.a),t}function __n(e){if(e.b<=0)throw ie(new ic);return--e.b,e.a-=e.c.c,rt(e.a)}function fxe(e,t){if(e.g==null||t>=e.i)throw ie(new vse(t,e.i));return e.g[t]}function Ogt(e,t,i){if(KP(e,i),i!=null&&!e.ck(i))throw ie(new dae);return i}function k_n(e,t,i){var a;return a=O1t(e,t,i),e.b=new YJ(a.c.length),xIe(e,a)}function Ngt(e,t){var i,a;return a=t/e.c.Pd().gc()|0,i=t%e.c.Pd().gc(),sR(e,a,i)}function LP(e,t,i,a){var u;a=(TS(),a||M6e),u=e.slice(t,i),MAe(u,e,t,i,-t,a)}function Lf(e,t,i,a,u){return t<0?GS(e,i,a):f(i,69).tk().vk(e,e.di(),t,a,u)}function T_n(e,t){return di(ke(Oe(V(e,(Ge(),d_)))),ke(Oe(V(t,d_))))}function Rgt(){Rgt=Z,eIt=hr((DP(),ce(se(Pee,1),Ve,305,0,[Xpe,Jpe,Zpe,Qpe])))}function DP(){DP=Z,Xpe=new NY("All",0),Jpe=new kst,Zpe=new Lst,Qpe=new Tst}function Yl(){Yl=Z,Om=new zae(NR,0),Ew=new zae(PL,1),tp=new zae(RR,2)}function Pgt(){Pgt=Z,dQ(),nje=Gi,jMt=qo,rje=new f3(Gi),$Mt=new f3(qo)}function Lgt(){Lgt=Z,FOt=hr((hE(),ce(se(g5e,1),Ve,381,0,[XT,p5e,mge,bge])))}function Dgt(){Dgt=Z,ZOt=hr((nZ(),ce(se(JOt,1),Ve,364,0,[kge,ate,ste,xq])))}function Mgt(){Mgt=Z,HIt=hr((DS(),ce(se(zIt,1),Ve,406,0,[yq,wq,ige,oge])))}function jgt(){jgt=Z,POt=hr((qP(),ce(se(ROt,1),Ve,405,0,[a_,YT,KT,bA])))}function $gt(){$gt=Z,uNt=hr((_g(),ce(se(cNt,1),Ve,410,0,[VR,rD,iD,Tge])))}function Fgt(){Fgt=Z,f6t=hr((xT(),ce(se(Ime,1),Ve,203,0,[Gte,Ame,PA,RA])))}function Bgt(){Bgt=Z,g6t=hr((Pp(),ce(se(IDe,1),Ve,266,0,[Pm,ADe,Rme,Pme])))}function RH(){RH=Z,yPe=new hSe(Lp,0),gte=new hSe("IMPROVE_STRAIGHTNESS",1)}function PH(e){var t;for(t=e.p+1;t<e.c.a.c.length;++t)--f(ct(e.c.a,t),9).p}function Ugt(e){var t;if(e.kl())for(t=e.i-1;t>=0;--t)be(e,t);return LTe(e)}function hxe(e){var t,i;if(!e.b)return null;for(i=e.b;t=i.a[0];)i=t;return i}function zgt(e){var t,i;if(!e.b)return null;for(i=e.b;t=i.a[1];)i=t;return i}function x_n(e){return me(e,179)?""+f(e,179).a:e==null?null:Qs(e)}function C_n(e){return me(e,179)?""+f(e,179).a:e==null?null:Qs(e)}function Hgt(e,t){if(t.a)throw ie(new oc(STt));Go(e.a,t),t.a=e,!e.j&&(e.j=t)}function A_n(e,t){return vP(),Qe(e,new Ya(t,rt(t.e.c.length+t.g.c.length)))}function I_n(e,t){return vP(),Qe(e,new Ya(t,rt(t.e.c.length+t.g.c.length)))}function O_n(e,t){return new Yse(t,eT($a(t.e),t.f.a+e,t.f.b+e),(mn(),!1))}function bue(e,t){return $H(t)!=10&&ce(Fd(t),t.Pm,t.__elementTypeId$,$H(t),e),e}function Ggt(){Ggt=Z,_Pt=hr((RZ(),ce(se(VMe,1),Ve,351,0,[hbe,GMe,qMe,HMe])))}function qgt(){qgt=Z,W5t=hr((wj(),ce(se(_Me,1),Ve,350,0,[nbe,nne,tbe,ebe])))}function Vgt(){Vgt=Z,S5t=hr((yR(),ce(se(qDe,1),Ve,396,0,[Jte,fF,hF,pF])))}function Wgt(){Wgt=Z,CPt=hr((zZ(),ce(se(xPt,1),Ve,398,0,[wbe,mbe,vbe,bbe])))}function Kgt(){Kgt=Z,ADt=hr((Ud(),ce(se($F,1),Ve,378,0,[mV,GE,gV,mx])))}function Ygt(){Ygt=Z,lDt=hr((XP(),ce(se(G9e,1),Ve,275,0,[cV,_ne,z9e,H9e])))}function Xgt(){Xgt=Z,fDt=hr((k1(),ce(se(i1e,1),Ve,220,0,[r1e,uV,DD,c4])))}function Jgt(){Jgt=Z,bDt=hr((df(),ce(se(mDt,1),Ve,289,0,[dV,Rb,Nw,lV])))}function Zgt(){Zgt=Z,RDt=hr((pZ(),ce(se(v7e,1),Ve,321,0,[c1e,g7e,b7e,m7e])))}function Qgt(){Qgt=Z,DDt=hr((iZ(),ce(se(LDt,1),Ve,404,0,[u1e,y7e,w7e,E7e])))}function TJ(){TJ=Z,Cbe=new dse(jxt,0),w8e=new dse(NRe,1),v8e=new dse(Lp,2)}function N_n(e,t,i,a,u){return an(e),an(t),an(i),an(a),an(u),new jke(e,t,a)}function emt(e,t){var i;return i=f(aR(e.e,t),391),i?(hke(i),i.e):null}function hl(e,t){var i;return i=Oc(e,t,0),i==-1?!1:(_0(e,i),!0)}function tmt(e,t,i){var a;return Kv(e),a=new ro,a.a=t,e.a.Nb(new Uot(a,i)),a.a}function R_n(e){var t;return Kv(e),t=Pe(sa,_s,30,0,15,1),b3(e.a,new Ce(t)),t}function pxe(e){var t;if(!Mue(e))throw ie(new ic);return e.e=1,t=e.d,e.d=null,t}function T0(e){var t;return Js(e)&&(t=0-e,!isNaN(t))?t:tw(VP(e))}function Oc(e,t,i){for(;i<e.c.length;++i)if(Pu(t,e.c[i]))return i;return-1}function vue(e){var t,i;return i=f(ct(e.j,0),11),t=f(V(i,(Ge(),oi)),11),t}function wue(e,t){var i;this.f=e,this.b=t,i=f(bn(e.b,t),260),this.c=i?i.b:null}function nmt(){tf(),this.b=new Wn,this.f=new Wn,this.g=new Wn,this.e=new Wn}function xJ(e){rX(this),this.g=e?QX(e,e.ge()):null,this.f=e,YX(this),this.he()}function yue(e){var t;t=e.gj(),t!=null&&e.d!=-1&&f(t,93).th(e),e.i&&e.i.lj()}function rmt(e){Qk(!!e.c),Vse(e.f.g,e.d),e.c.Qb(),e.c=null,e.b=nCe(e),e.d=e.f.g}function gd(e){return e.b||(e.b=new Dlt(e,gs,e),!e.a&&(e.a=new nH(e,e))),e.b}function LH(e,t){var i,a;return i=f(t,675),a=i.ti(),!a&&i.wi(a=new Xat(e,t)),a}function Ns(e,t){var i,a;return i=f(t,677),a=i.Xk(),!a&&i._k(a=new yht(e,t)),a}function gxe(e,t){p7();var i,a;return i=iR(e),a=iR(t),!!i&&!!a&&!Nvt(i.k,a.k)}function DH(e,t){if(t<0)throw ie(new Ru(Zxt+t));return Wht(e,t+1),ct(e.j,t)}function imt(e,t,i,a){if(!e)throw ie(new dn(Vj(t,ce(se(Ro,1),Jt,1,5,[i,a]))))}function CJ(e,t){return Pu(t,ct(e.f,0))||Pu(t,ct(e.f,1))||Pu(t,ct(e.f,2))}function cj(){cj=Z,lge=new Gae("XY",0),uge=new Gae("X",1),dge=new Gae("Y",2)}function MH(){MH=Z,Lme=new mSe("INPUT_ORDER",0),Dme=new mSe("PORT_DEGREE",1)}function pl(){pl=Z,jp=new Hae("TOP",0),Sw=new Hae(PL,1),np=new Hae(KOe,2)}function omt(e,t){W6e=new _l,GIt=t,$$=e,f($$.b,68),ixe($$,W6e,null),L_t($$)}function AJ(e){var t,i;return i=e.length,t=Pe(Xf,_m,30,i,15,1),odt(e,0,i,t,0),t}function jH(e){xlt();var t,i;return t=e+128,i=E6e[t],!i&&(i=E6e[t]=new zn(e)),i}function amt(e){return Vse(e.d.a.e.g,e.b),Ln(e.c!=e.d.a.d),e.a=e.c,e.c=e.c.a,e.a}function smt(e){(this.q?this.q:(Ht(),Ht(),Cb)).zc(e.q?e.q:(Ht(),Ht(),Cb))}function mxe(e){var t;return e.Kh()||(t=jn(e.zh())-e.fi(),e.Wh().Jk(t)),e.vh()}function bxe(e){var t;return t=uR(fn(e,32)),t==null&&(Xu(e),t=uR(fn(e,32))),t}function Eue(e,t){var i;return i=Pi(e.d,t),i>=0?CZ(e,i,!0,!0):GS(e,t,!0)}function P_n(e,t){VN(f(f(e.f,27).mf((Rr(),OF)),100))&&M0t(Qke(f(e.f,27)),t)}function cmt(e,t){md(e,t==null||gX((an(t),t))||isNaN((an(t),t))?0:(an(t),t))}function umt(e,t){bd(e,t==null||gX((an(t),t))||isNaN((an(t),t))?0:(an(t),t))}function lmt(e,t){RS(e,t==null||gX((an(t),t))||isNaN((an(t),t))?0:(an(t),t))}function dmt(e,t){NS(e,t==null||gX((an(t),t))||isNaN((an(t),t))?0:(an(t),t))}function fmt(e,t){for(;t[0]<e.length&&cm(` 	\r
`,gl(_a(e,t[0])))>=0;)++t[0]}function hmt(e,t,i,a){Xr(),sS.call(this,26),this.c=e,this.a=t,this.d=i,this.b=a}function v1(e,t,i,a,u,l,g){Oue.call(this,t,a,u,l,g),this.c=e,this.b=i}function Sue(e,t,i){var a;return a=e.g[t],j7(e,t,e.Wi(t,i)),e.Oi(t,i,a),e.Ki(),a}function IJ(e,t){var i;return i=e.bd(t),i>=0?(e.ed(i),!0):!1}function Eg(e,t){var i;return i=f(bn(e.e,t),391),i?(Gst(e,i),i.e):null}function _ue(e){var t;return e.d!=e.r&&(t=ph(e),e.e=!!t&&t.ik()==$3t,e.d=t),e.e}function kue(e,t){var i;for(or(e),or(t),i=!1;t.Ob();)i=i|e.Ec(t.Pb());return i}function Tue(e){var t;return t=e.b.c.length==0?null:ct(e.b,0),t!=null&&Pue(e,0),t}function Zs(e,t){var i,a;return ow(e),a=new ZTe(t,e.a),i=new _ut(a),new Ft(e,i)}function w1(){w1=Z,Oq=new Xae(Lp,0),kA=new Xae("TOP",1),WR=new Xae(KOe,2)}function uj(){uj=Z,MOt=new dy,jOt=new T2,LOt=new QI,DOt=new x2,$Ot=new o6}function $H(e){return e.__elementTypeCategory$==null?10:e.__elementTypeCategory$}function L_n(e,t,i){return UN(new Ke(i.e.a+i.f.a/2,i.e.b+i.f.b/2),e)==(an(t),t)}function D_n(e,t){var i,a,u;u=t.c.i,i=f(bn(e.f,u),60),a=i.d.c-i.e.c,Yxe(t.a,a,0)}function vxe(e,t){var i;for(++e.d,++e.c[t],i=t+1;i<e.a.length;)++e.a[i],i+=i&-i}function bb(e,t,i){var a,u;for(a=10,u=0;u<i-1;u++)t<a&&(e.a+="0"),a*=10;e.a+=t}function M_n(e,t){var i;for(i=0;e.e!=e.i.gc();)hwn(t,Mn(e),rt(i)),i!=Qr&&++i}function pmt(e){var t,i;return t=e/60|0,i=e%60,i==0?""+t:""+t+":"+(""+i)}function gT(e,t){var i=e.a[t],a=(Xue(),Gpe)[typeof i];return a?a(i):ICe(typeof i)}function j_n(e){switch(e.g){case 0:return Qr;case 1:return Ra;default:return 0}}function $_n(e){return z3e(e,(MP(),w6e))<0?-Mmn(VP(e)):e.l+e.m*OR+e.h*yE}function wxe(e){if(!e.c)throw ie(new ic);return e.e=e.a=e.c,e.c=e.c.c,++e.d,e.a.f}function gmt(e){if(!e.e)throw ie(new ic);return e.c=e.a=e.e,e.e=e.e.e,--e.d,e.a.f}function mmt(e){var t;for(++e.a,t=e.c.a.length;e.a<t;++e.a)if(e.c.b[e.a])return}function F_n(e){var t,i;if(e.a){i=null;do t=e.a,e.a=null,i=Pyt(t,i);while(e.a);e.a=i}}function B_n(e){var t,i;if(e.b){i=null;do t=e.b,e.b=null,i=Pyt(t,i);while(e.b);e.b=i}}function U_n(e,t){var i,a;for(a=t.c,i=a+1;i<=t.f;i++)e.a[i]>e.a[a]&&(a=i);return a}function z_n(e,t){var i;return i=jS(e.e.c,t.e.c),i==0?di(e.e.d,t.e.d):i}function bmt(e,t){return t==(qse(),qse(),JAt)?e.toLocaleLowerCase():e.toLowerCase()}function H_n(e,t){return me(t,101)&&(f(t,19).Bb&Xa)!=0?new wse(t,e):new GP(t,e)}function G_n(e,t){return me(t,101)&&(f(t,19).Bb&Xa)!=0?new wse(t,e):new GP(t,e)}function vmt(){vmt=Z,AOt=hr((wa(),ce(se(f5e,1),Ve,360,0,[rp,Ab,su,cu,zs])))}function wmt(){wmt=Z,SNt=hr((R0(),ce(se(mPe,1),Ve,344,0,[jge,gPe,$ge,uD,cD])))}function ymt(){ymt=Z,dNt=hr((CT(),ce(se(lNt,1),Ve,368,0,[Cq,lte,dte,ute,cte])))}function Emt(){Emt=Z,h6t=hr((Nj(),ce(se(SDe,1),Ve,324,0,[EDe,Ome,Nme,oF,aF])))}function Smt(){Smt=Z,DNt=hr((Ff(),ce(se(WPe,1),Ve,171,0,[Lq,X$,Tw,J$,tx])))}function _mt(){_mt=Z,tPt=hr((BZ(),ce(se(ePt,1),Ve,361,0,[sbe,ibe,cbe,obe,abe])))}function kmt(){kmt=Z,nLt=hr((Hj(),ce(se(tLt,1),Ve,366,0,[$A,n4,kF,_F,Jq])))}function Tmt(){Tmt=Z,cLt=hr((bG(),ce(se(I8e,1),Ve,325,0,[x8e,Pbe,A8e,Lbe,C8e])))}function xmt(){xmt=Z,uDt=hr((ko(),ce(se(NF,1),Ve,86,0,[$g,Gs,Ds,jg,Wf])))}function Cmt(){Cmt=Z,SDt=hr((tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])))}function Amt(){Amt=Z,wDt=hr((K3(),ce(se(DF,1),Ve,255,0,[Rw,fV,i7e,LF,o7e])))}function Imt(){Imt=Z,NLt=hr((wm(),ce(se(Ob,1),Ve,170,0,[en,Ao,Dm,FE,L1])))}function FH(){FH=Z,hte=new fSe("READING_DIRECTION",0),bPe=new fSe("ROTATION",1)}function OJ(){OJ=Z,age=new cSe("EADES",0),jee=new cSe("FRUCHTERMAN_REINGOLD",1)}function MP(){MP=Z,b6e=Ku(Ed,Ed,524287),$At=Ku(0,0,zG),v6e=aue(1),aue(2),w6e=aue(0)}function xue(e,t){var i;return i=f(bn(e.a,t),148),i||(i=new k2,qr(e.a,t,i)),i}function Omt(e){var t;return t=f(V(e,(Ge(),u_)),314),t?t.a==e:!1}function Nmt(e){var t;return t=f(V(e,(Ge(),u_)),314),t?t.i==e:!1}function Rmt(e,t){return an(t),Ike(e),e.d.Ob()?(t.Ad(e.d.Pb()),!0):!1}function NJ(e){return Eu(e,Qr)>0?Qr:Eu(e,Ra)<0?Ra:mr(e)}function mT(e){return e<3?(uf(e,nTt),e+1):e<n$?Aa(e/.75+1):Qr}function Wh(e,t,i){var a;if(t==null)throw ie(new kN);return a=h1(e,t),fSn(e,t,i),a}function q_n(e,t,i){var a,u;for(u=new ne(i);u.a<u.c.c.length;)a=ee(u),oAe(e,t,a)}function V_n(e,t,i){var a;return a=e.a.e[f(t.a,9).p]-e.a.e[f(i.a,9).p],Aa(dH(a))}function Vt(e,t){var i;return i=(e.i==null&&Em(e),e.i),t>=0&&t<i.length?i[t]:null}function W_n(e,t){var i;return i=t>0?t-1:t,fit(Hpn(tbt(uke(new CN,i),e.n),e.j),e.k)}function Rs(e){var t,i;i=(t=new gae,t),Qn((!e.q&&(e.q=new Be(Th,e,11,10)),e.q),i)}function Pmt(e){this.g=e,this.f=new et,this.a=h.Math.min(this.g.c.c,this.g.d.c)}function Lmt(e){this.b=new et,this.a=new et,this.c=new et,this.d=new et,this.e=e}function Dmt(e,t,i){hX.call(this),Exe(this),this.a=e,this.c=i,this.b=t.d,this.f=t.e}function Mmt(e){this.d=e,this.c=e.c.vc().Jc(),this.b=null,this.a=null,this.e=(dY(),Bpe)}function OS(e){if(e<0)throw ie(new dn("Illegal Capacity: "+e));this.g=this.Zi(e)}function K_n(){return aQ(),ce(se(cPe,1),Ve,272,0,[Age,Nge,Cge,Lge,Oge,Ige,Pge,Rge])}function Y_n(){return cn(),ce(se(wge,1),Ve,247,0,[Yi,uo,bo,xu,Vc,Dg,_q,U$])}function X_n(){return pL(),ce(se($ne,1),Ve,243,0,[h1e,Mne,jne,Dne,f1e,Lne,Pne,d1e])}function J_n(){return bE(),ce(se(RLt,1),Ve,282,0,[Y8e,Po,Hs,i4,ya,Li,r4,Nb])}function Z_n(e,t){return lR(),-tu(f(V(e,(Ec(),DA)),17).a,f(V(t,DA),17).a)}function jmt(e,t){return!!gj(e,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))))}function yxe(e){return((e.i&2)!=0?"interface ":(e.i&1)!=0?"":"class ")+(d1(e),e.o)}function RJ(e,t){var i;return ow(e),i=new Qdt(e,e.a.xd(),e.a.wd()|4,t),new Ft(e,i)}function Q_n(e,t){var i,a;return i=f(ST(e.d,t),16),i?(a=t,e.e.pc(a,i)):null}function PJ(e,t){var i,a;for(a=e.Jc();a.Ob();)i=f(a.Pb(),72),$e(i,(Ge(),JR),t)}function ekn(e){var t;return t=ke(Oe(V(e,(ot(),z0)))),t<0&&(t=0,$e(e,z0,t)),t}function tkn(e,t){VN(f(V(f(e.e,9),(ot(),Xi)),100))&&(Ht(),na(f(e.e,9).j,t))}function $mt(e,t){pi(e,(wb(),Ebe),t.f),pi(e,IPt,t.e),pi(e,ybe,t.d),pi(e,APt,t.c)}function Cue(e){var t;Qk(!!e.c),t=e.c.a,lh(e.d,e.c),e.b==e.c?e.b=t:--e.a,e.c=null}function Fmt(e){return e.a>=-.01&&e.a<=Og&&(e.a=0),e.b>=-.01&&e.b<=Og&&(e.b=0),e}function $3(e){bL();var t,i;for(i=LRe,t=0;t<e.length;t++)e[t]>i&&(i=e[t]);return i}function Bmt(e,t){var i;if(i=kG(e.zh(),t),!i)throw ie(new dn(vw+t+rpe));return i}function bT(e,t){var i;for(i=e;so(i);)if(i=so(i),i==t)return!0;return!1}function nkn(e,t){return t&&e.b[t.g]==t?(io(e.b,t.g,null),--e.c,!0):!1}function lh(e,t){var i;return i=t.c,t.a.b=t.b,t.b.a=t.a,t.a=t.b=null,t.c=null,--e.b,i}function Mu(e,t){var i,a,u,l;for(an(t),a=e.c,u=0,l=a.length;u<l;++u)i=a[u],t.Ad(i)}function rkn(e,t){var i,a,u;for(a=t.a.jd(),i=f(t.a.kd(),16).gc(),u=0;u<i;u++)e.Ad(a)}function ikn(e,t,i){var a;a=h.Math.max(0,e.b/2-.5),Lj(i,a,1),Qe(t,new Vot(i,a))}function okn(e,t,i,a){var u;e.j=-1,YAe(e,CAe(e,t,i),(is(),u=f(t,69).sk(),u.ul(a)))}function Umt(e,t,i,a,u,l){var g;g=cue(a),Ia(g,u),ua(g,l),jt(e.a,a,new vX(g,t,i.f))}function Aue(e,t){var i,a,u,l;for(a=e.d,u=0,l=a.length;u<l;++u)i=a[u],mg(e.g,i).a=t}function akn(e,t,i){var a,u,l;for(u=t[i],a=0;a<u.length;a++)l=u[a],e.e[l.c.p][l.p]=a}function skn(e){var t;for(t=0;t<e.c.length;t++)(Gt(t,e.c.length),f(e.c[t],11)).p=t}function ckn(e){var t,i;for(t=e.a.d.j,i=e.c.d.j;t!=i;)Yh(e.b,t),t=fZ(t);Yh(e.b,t)}function ukn(e){var t;return t=h.Math.sqrt(e.a*e.a+e.b*e.b),t>0&&(e.a/=t,e.b/=t),e}function zmt(e,t,i){var a,u;a=t;do u=ke(e.p[a.p])+i,e.p[a.p]=u,a=e.a[a.p];while(a!=t)}function lkn(e,t){if(0>e||e>t)throw ie(new P2e("fromIndex: 0, toIndex: "+e+BOe+t))}function be(e,t){if(e.g==null||t>=e.i)throw ie(new vse(t,e.i));return e.Ti(t,e.g[t])}function Sg(){Sg=Z,PD=new pse(PL,0),gx=new pse("HEAD",1),LD=new pse("TAIL",2)}function Np(){Np=Z,_c=new Uae("BEGIN",0),Bu=new Uae(PL,1),kc=new Uae("END",2)}function Hmt(){Hmt=Z,Dee=(Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])).length,nge=Dee}function Exe(e){e.b=(Yl(),Ew),e.f=(pl(),Sw),e.d=(uf(2,$T),new Du(2)),e.e=new ha}function LJ(e){this.b=(or(e),new Kl(e)),this.a=new et,this.d=new et,this.e=new ha}function cf(e){var t;return e.w?e.w:(t=e2n(e),t&&!t.Rh()&&(e.w=t),t)}function dkn(e){var t;return e==null?null:(t=f(e,195),QOn(t,t.length))}function uR(e){var t;return z7(e==null||Array.isArray(e)&&(t=$H(e),!(t>=14&&t<=16))),e}function fkn(e,t){var i;return i=Ho($a(f(bn(e.g,t),8)),VSe(f(bn(e.f,t),456).b)),i}function Gmt(e,t,i){var a=function(){return e.apply(a,arguments)};return t.apply(a,i),a}function hkn(e,t){return cw(),tu(e.b.c.length-e.e.c.length,t.b.c.length-t.e.c.length)}function F3(e,t){return Wpn(pj(e,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15)))))}function Sxe(e,t){return gg(),Kh(fw),h.Math.abs(e-t)<=fw||e==t||isNaN(e)&&isNaN(t)}function _xe(e,t){return gg(),Kh(fw),h.Math.abs(e-t)<=fw||e==t||isNaN(e)&&isNaN(t)}function pkn(e,t){Ht();var i,a;for(a=new et,i=0;i<e;++i)vn(a.c,t);return new Iae(a)}function qmt(e){return ow(e),KN(!0,"n may not be negative"),new Ft(e,new fbt(e.a))}function y1(){y1=Z,Z5t=vm(vm(vm(b7(new _o,(yR(),pF)),(Wj(),JDe)),VDe),XDe)}function lR(){lR=Z,X5t=vm(vm(vm(b7(new _o,(yR(),fF)),(Wj(),Hme)),YDe),QDe)}function Vmt(){Vmt=Z,xNt=hr((qZ(),ce(se(wPe,1),Ve,281,0,[Uge,Bge,Hge,zge,Gge,pte])))}function Wmt(){Wmt=Z,INt=hr((fL(),ce(se(RPe,1),Ve,280,0,[q$,APe,NPe,OPe,IPe,CPe])))}function Kmt(){Kmt=Z,ANt=hr((gG(),ce(se(xPe,1),Ve,279,0,[mte,_Pe,TPe,SPe,kPe,qge])))}function Ymt(){Ymt=Z,vNt=hr((ER(),ce(se(sD,1),Ve,231,0,[aD,H$,oD,QT,EA,yA])))}function Xmt(){Xmt=Z,LLt=hr((vb(),ce(se(Z8e,1),Ve,254,0,[Bbe,Qq,eV,wne,bne,vne])))}function Jmt(){Jmt=Z,BLt=hr((mE(),ce(se(Gbe,1),Ve,296,0,[Hbe,IF,AF,zbe,xF,CF])))}function Zmt(){Zmt=Z,hDt=hr((HZ(),ce(se(J9e,1),Ve,322,0,[o1e,K9e,X9e,V9e,Y9e,W9e])))}function Qmt(){Qmt=Z,yDt=hr((oa(),ce(se(a7e,1),Ve,100,0,[Pw,Fg,MD,HE,Lb,uu])))}function ebt(){ebt=Z,EDt=hr((Xl(),ce(se(Tne,1),Ve,277,0,[Lw,M1,hV,jF,MF,u4])))}function gkn(){return NT(),ce(se(ps,1),Ve,94,0,[sp,D1,cp,lp,Pb,Sh,bf,up,Eh])}function lj(){lj=Z,pV=new mse(I4e,0),a1e=new mse("PARENT",1),c7e=new mse("ROOT",2)}function tbt(e,t){return e.n=t,e.n?(e.f=new et,e.e=new et):(e.f=null,e.e=null),e}function NS(e,t){var i;i=e.f,e.f=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,3,i,e.f))}function RS(e,t){var i;i=e.g,e.g=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,4,i,e.g))}function md(e,t){var i;i=e.i,e.i=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,5,i,e.i))}function bd(e,t){var i;i=e.j,e.j=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,6,i,e.j))}function B3(e,t){var i;i=e.j,e.j=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,1,i,e.j))}function DJ(e,t){var i;i=e.b,e.b=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,1,i,e.b))}function U3(e,t){var i;i=e.b,e.b=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,3,i,e.b))}function z3(e,t){var i;i=e.c,e.c=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,4,i,e.c))}function H3(e,t){var i;i=e.k,e.k=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,2,i,e.k))}function MJ(e,t){var i;i=e.a,e.a=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new pT(e,0,i,e.a))}function x0(e,t){var i;i=e.s,e.s=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new tue(e,4,i,e.s))}function vT(e,t){var i;i=e.t,e.t=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new tue(e,5,i,e.t))}function Iue(e,t){var i;i=e.d,e.d=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new tue(e,2,i,e.d))}function jP(e,t){var i;i=e.F,e.F=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,5,i,t))}function $P(e,t){var i,a;a=e.a,i=Vxn(e,t,null),a!=t&&!e.e&&(i=kL(e,t,i)),i&&i.lj()}function mkn(e,t){var i,a,u;return i=(a=(Bv(),u=new KC,u),t&&yIe(a,t),a),jxe(i,e),i}function nbt(e,t,i){var a;return a=oL(i),qr(e.c,a,t),qr(e.d,t,i),qr(e.e,t,lT(t)),t}function E1(e,t){var i,a;return i=t in e.a,i&&(a=h1(e,t).pe(),a)?a.a:null}function rbt(e,t){if(e){t.n=e;var i=nEn(t);if(!i){xee[e]=[t];return}i.Om=t}}function ibt(e,t,i){if(KP(e,i),!e.hl()&&i!=null&&!e.ck(i))throw ie(new dae);return i}function fr(e,t){var i;return an(t),i=e[":"+t],uH(!!i,"Enum constant undefined: "+t),i}function er(e,t,i,a,u,l){var g;return g=_ce(e,t),rbt(i,g),g.i=u?8:0,g.f=a,g.e=u,g.g=l,g}function kxe(e,t,i,a,u){this.d=t,this.k=a,this.f=u,this.o=-1,this.p=1,this.c=e,this.a=i}function Txe(e,t,i,a,u){this.d=t,this.k=a,this.f=u,this.o=-1,this.p=2,this.c=e,this.a=i}function xxe(e,t,i,a,u){this.d=t,this.k=a,this.f=u,this.o=-1,this.p=6,this.c=e,this.a=i}function Cxe(e,t,i,a,u){this.d=t,this.k=a,this.f=u,this.o=-1,this.p=7,this.c=e,this.a=i}function Axe(e,t,i,a,u){this.d=t,this.j=a,this.e=u,this.o=-1,this.p=4,this.c=e,this.a=i}function obt(e,t){var i,a,u,l;for(a=t,u=0,l=a.length;u<l;++u)i=a[u],Hgt(e.a,i);return e}function Df(e){var t,i,a,u;for(i=e,a=0,u=i.length;a<u;++a)t=i[a],or(t);return new gst(e)}function Ixe(e){var t;return t=Ho($a(e.d.d),e.c.d),uL(t,e.c.e.a,e.c.e.b),ii(t,e.c.d)}function Oxe(e){var t;return t=Ho($a(e.c.d),e.d.d),uL(t,e.d.e.a,e.d.e.b),ii(t,e.d.d)}function bkn(e){var t=/function(?:\s+([\w$]+))?\s*\(/,i=t.exec(e);return i&&i[1]||hfe}function BH(e,t){var i;return i=f(bn((CY(),Fne),e),58),i?i.dk(t):Pe(Ro,Jt,1,t,5,1)}function vkn(e,t,i){var a,u;return u=e.length,a=h.Math.min(i,u),KAe(e,0,t,0,a,!0),t}function wkn(e,t){var i,a;return i=e.c,a=t.e[e.p],a>0?f(ct(i.a,a-1),9):null}function jJ(e,t){e.b=h.Math.max(e.b,t.d),e.e+=t.r+(e.a.c.length==0?0:e.c),Qe(e.a,t)}function abt(e){Qk(e.c>=0),c3n(e.d,e.c)<0&&(e.a=e.a-1&e.d.a.length-1,e.b=e.d.c),e.c=-1}function dj(e,t){var i,a,u,l;for(a=t,u=0,l=a.length;u<l;++u)i=a[u],Wi(e,i,e.c.b,e.c)}function ykn(){wot();for(var e=Lpe,t=0;t<arguments.length;t++)e.push(arguments[t])}function Ekn(e,t){t.Sg("Hierarchical port constraint processing",1),U3n(e),S7n(e),t.Tg()}function UH(e,t){var i;if(i=e.gc(),t<0||t>i)throw ie(new rT(t,i));return new z_e(e,t)}function Ss(e,t,i){return fE(e,t),ju(e,i),x0(e,0),vT(e,1),N0(e,!0),O0(e,!0),e}function sbt(e,t){var i;return me(t,43)?e.c.Kc(t):(i=Ele(e,t),yZ(e,t),i)}function Kh(e){if(!(e>=0))throw ie(new dn("tolerance ("+e+") must be >= 0"));return e}function fj(){return $be||($be=new wSt,pR($be,ce(se(HR,1),Jt,146,0,[new a1]))),$be}function $J(){$J=Z,$De=new nse("NO",0),Bme=new nse(cNe,1),jDe=new nse("LOOK_BACK",2)}function FJ(){FJ=Z,hPe=new Wae("ARD",0),fte=new Wae("MSD",1),Mge=new Wae("MANUAL",2)}function cs(){cs=Z,sF=new Qae(s$,0),td=new Qae("INPUT",1),Uu=new Qae("OUTPUT",2)}function PS(){PS=Z,zq=new ise("BARYCENTER",0),uF=new ise(sxt,1),Wte=new ise(cxt,2)}function Skn(){return D0(),ce(se(vDe,1),Ve,265,0,[xme,gDe,mDe,Tme,pDe,bDe,zte,SD,_D])}function _kn(){return yd(),ce(se(l7e,1),Ve,264,0,[jD,vV,xne,FF,Cne,Ine,Ane,s1e,bV])}function BJ(e,t){return Vpn(gj(e.a,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15)))))}function kkn(e){var t,i,a,u;for(i=e.a,a=0,u=i.length;a<u;++a)t=i[a],adt(t,t.length,null)}function UJ(e){var t,i;for(i=e.c.Bc().Jc();i.Ob();)t=f(i.Pb(),16),t.$b();e.c.$b(),e.d=0}function Tkn(e){var t,i;for(i=new ne(swt(e));i.a<i.c.c.length;)t=f(ee(i),680),t.Zf()}function cbt(e,t){var i,a;for(i=0,a=e.gc();i<a;++i)if(Pu(t,e.Xb(i)))return i;return-1}function Nxe(e){var t,i;if(e==0)return 32;for(i=0,t=1;(t&e)==0;t<<=1)++i;return i}function uf(e,t){if(e<0)throw ie(new dn(t+" cannot be negative but was: "+e));return e}function ubt(e){Z2e(),this.g=new Wn,this.f=new Wn,this.b=new Wn,this.c=new CS,this.i=e}function Rxe(){this.f=new ha,this.d=new b2e,this.c=new ha,this.a=new et,this.b=new et}function lbt(e,t,i,a){this.Zj(),this.a=t,this.b=e,this.c=null,this.c=new mut(this,t,i,a)}function Oue(e,t,i,a,u){this.d=e,this.n=t,this.g=i,this.o=a,this.p=-1,u||(this.o=-2-a-1)}function dbt(){JSe.call(this),this.n=-1,this.g=null,this.i=null,this.j=null,this.Bb|=ep}function fbt(e){eX.call(this,e.yd(64)?USe(0,hh(e.xd(),1)):FG,e.wd()),this.b=1,this.a=e}function xkn(e,t){var i,a,u;u=t.c.i,i=f(bn(e.f,u),60),a=i.d.c-i.e.c,Mu(t.b,new Fet(a))}function hj(){hj=Z,G$=new Vae("LAYER_SWEEP",0),Aq=new Vae(the,1),dPe=new Vae(Lp,2)}function Pxe(){Pxe=Z,A6t=Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)}function hbt(){hbt=Z,I6t=Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)}function Lxe(){Lxe=Z,O6t=Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)}function pbt(){pbt=Z,N6t=Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)}function gbt(){gbt=Z,R6t=Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)}function mbt(){mbt=Z,P6t=Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)}function bbt(){bbt=Z,M6t=Il(Sr(Sr(new _o,(wa(),su),(ka(),Zee)),cu,Wee),zs,Jee)}function vbt(){vbt=Z,FAt=ce(se(pr,1),Pr,30,15,[0,8,4,12,2,10,6,14,1,9,5,13,3,11,7,15])}function Dxe(e,t){var i;i=e.b,e.b=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,0,i,e.b))}function Mxe(e,t){var i;i=e.c,e.c=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,1,i,e.c))}function Nue(e,t){var i;i=e.c,e.c=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,4,i,e.c))}function FP(e,t){var i;i=e.k,e.k=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,2,i,e.k))}function Rue(e,t){var i;i=e.D,e.D=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,2,i,e.D))}function zJ(e,t){var i;i=e.f,e.f=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,8,i,e.f))}function HJ(e,t){var i;i=e.i,e.i=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,7,i,e.i))}function jxe(e,t){var i;i=e.a,e.a=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,8,i,e.a))}function $xe(e,t){var i;i=e.d,e.d=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,1,i,e.d))}function Fxe(e,t){var i;i=e.c,e.c=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,1,i,e.c))}function Bxe(e,t){var i;i=e.b,e.b=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,0,i,e.b))}function Ckn(e,t,i){var a;e.b=t,e.a=i,a=(e.a&512)==512?new Prt:new P9,e.c=tPn(a,e.b,e.a)}function wbt(e,t){return x1(e.e,t)?(is(),_ue(t)?new TX(t,e):new Wz(t,e)):new Jat(t,e)}function Akn(e,t){return lR(),f(V(t,(Ec(),DA)),17).a<e.gc()&&f(V(t,DA),17).a>=0}function Ikn(e,t){Ht();var i;return i=new PN(1),ta(e)?Os(i,e,t):ml(i.f,e,t),new vz(i)}function Okn(e){var t,i;return 0>e?new X2e:(t=e+1,i=new Rpt(t,e),new y_e(null,i))}function ew(){ew=Z,qR=new Si(eNe),Bee=new Si(HTt),B$=new Si(GTt),Eq=new Si(qTt)}function GJ(){GJ=Z,rne=new _Se("P1_NODE_PLACEMENT",0),jA=new _Se("P2_EDGE_ROUTING",1)}function LS(){LS=Z,c_=new qae("START",0),_w=new qae("MIDDLE",1),s_=new qae("END",2)}function Nkn(e,t){var i;i=new _l,f(t.b,68),f(t.b,68),f(t.b,68),Mu(t.a,new J_e(e,i,t))}function ybt(e,t){var i,a;for(a=t.vc().Jc();a.Ob();)i=f(a.Pb(),43),hG(e,i.jd(),i.kd())}function Uxe(e,t){var i;i=e.d,e.d=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,11,i,e.d))}function qJ(e,t){var i;i=e.j,e.j=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,13,i,e.j))}function zxe(e,t){var i;i=e.b,e.b=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,21,i,e.b))}function Hxe(e,t){e.r>0&&e.c<e.r&&(e.c+=t,e.i&&e.i.d>0&&e.g!=0&&Hxe(e.i,t/e.r*e.i.d))}function Ebt(e,t){var i;return me(t,8)?(i=f(t,8),e.a==i.a&&e.b==i.b):!1}function Rkn(e){var t;return t=V(e,(Ge(),oi)),me(t,173)?B0t(f(t,173)):null}function Sbt(e){var t;return e=h.Math.max(e,2),t=dCe(e),e>t?(t<<=1,t>0?t:n$):t}function G3(e){var t;return Vse(e.f.g,e.d),Ln(e.b),e.c=e.a,t=f(e.a.Pb(),43),e.b=nCe(e),t}function Gxe(e){var t;return e.b==null?(v0(),v0(),TV):(t=e.rl()?e.ql():e.pl(),t)}function _bt(e,t){var i;return i=t==null?-1:Oc(e.b,t,0),i<0?!1:(Pue(e,i),!0)}function Yh(e,t){var i;return an(t),i=t.g,e.b[i]?!1:(io(e.b,i,t),++e.c,!0)}function VJ(e,t){var i,a;return i=1-t,a=e.a[i],e.a[i]=a.a[t],a.a[t]=e,e.b=!0,a.b=!1,a}function Pue(e,t){var i;i=_0(e.b,e.b.c.length-1),t<e.b.c.length&&(af(e.b,t,i),eEt(e,t))}function Pkn(e,t){(mJ(),Lg?null:t.c).length==0&&Gct(t,new Ca),Os(e.a,Lg?null:t.c,t)}function Lkn(e,t){Zi(Jr(new Ft(null,new Bt(new Ic(e.b),1)),new Tat(e,t)),new Cat(e,t))}function kbt(e,t,i){ATe.call(this,e,t,i),this.a=new Wn,this.b=new Wn,this.d=new att(this)}function Tbt(e){Ske.call(this),this.b=ke(Oe(V(e,(ot(),op)))),this.a=f(V(e,P1),220)}function Dkn(e,t){var i,a;return i=f(V(e,(ot(),sx)),8),a=f(V(t,sx),8),di(i.b,a.b)}function Mkn(e,t){var i,a;a=new et,i=t;do vn(a.c,i),i=f(bn(e.k,i),18);while(i);return a}function Lue(e,t,i){var a;return a=new et,_Ie(e,t,a,i,!0,!0),e.b=new YJ(a.c.length),a}function xbt(e,t){var i;return i=f(bn(e.c,t),453),i||(i=new hrt,i.c=t,qr(e.c,i.c,i)),i}function Due(e,t){var i=e.a,a=0;for(var u in i)i.hasOwnProperty(u)&&(t[a++]=u);return t}function Mue(e){switch(e_e(e.e!=3),e.e){case 2:return!1;case 0:return!0}return jSn(e)}function jue(e,t){var i;return me(t,16)?(i=f(t,16),e.Fc(i)):kue(e,f(or(t),20).Jc())}function qxe(e){var t,i;for(i=new Nn(e);i.e!=i.i.gc();)t=f(Mn(i),27),md(t,0),bd(t,0)}function sE(e,t){var i,a;for(i=e.Nc(),LP(i,0,i.length,t),a=0;a<i.length;a++)e.fd(a,i[a])}function Cbt(e,t,i){var a,u,l;for(l=e.a.length-1,u=e.b,a=0;a<i;u=u+1&l,++a)io(t,a,e.a[u])}function Abt(e){var t,i;for(i=e.c.a.ec().Jc();i.Ob();)t=f(i.Pb(),217),GK(t,new Vlt(t.f))}function Vxe(e){var t,i;for(i=e.c.a.ec().Jc();i.Ob();)t=f(i.Pb(),217),pz(t,new hyt(t.e))}function Ti(e){return ta(e)?C0(e):Jk(e)?FN(e):Xk(e)?Uct(e):xke(e)?e.Hb():Eke(e)?bS(e):sTe(e)}function jkn(e,t){t.Sg(axt,1),Zi(Zs(new Ft(null,new Bt(e.b,16)),new X_),new d6),t.Tg()}function Ki(e,t){var i,a;i=(a=new pae,a),i.n=t,Qn((!e.s&&(e.s=new Be(El,e,21,17)),e.s),i)}function Ci(e,t){var i,a;a=(i=new E_e,i),a.n=t,Qn((!e.s&&(e.s=new Be(El,e,21,17)),e.s),a)}function ju(e,t){var i;i=e.zb,e.zb=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,1,i,e.zb))}function WJ(e,t){var i;i=e.xb,e.xb=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,3,i,e.xb))}function KJ(e,t){var i;i=e.yb,e.yb=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,2,i,e.yb))}function YJ(e){this.b=e,this.a=Pe(pr,Pr,30,e+1,15,1),this.c=Pe(pr,Pr,30,e,15,1),this.d=0}function Ibt(e){this.e=e,this.d=new bY(mT(YN(this.e).gc())),this.c=this.e.a,this.b=this.e.c}function Obt(){this.c=new vY(0),this.b=new vY(PRe),this.d=new vY(Ixt),this.a=new vY(Oxt)}function BP(){BP=Z,kq=new uSe("TO_INTERNAL_LTR",0),yge=new uSe("TO_INPUT_DIRECTION",1)}function dR(){dR=Z,_ge=new Mi("edgelabelcenterednessanalysis.includelabel",(mn(),yw))}function Wxe(e){return e.a<54?e.f<0?-1:e.f>0?1:0:(!e.c&&(e.c=yH(Nc(e.f))),e.c).e}function Nbt(e,t){t?e.B==null&&(e.B=e.D,e.D=null):e.B!=null&&(e.D=e.B,e.B=null)}function $kn(e,t){typeof window===DG&&typeof window.$gwt===DG&&(window.$gwt[e]=t)}function Fkn(e,t){return qP(),e==a_&&t==YT||e==YT&&t==a_||e==bA&&t==KT||e==KT&&t==bA}function Bkn(e,t){return qP(),e==a_&&t==KT||e==a_&&t==bA||e==YT&&t==bA||e==YT&&t==KT}function Rbt(e,t){return gg(),Kh(Og),h.Math.abs(0-t)<=Og||t==0||isNaN(0)&&isNaN(t)?0:e/t}function Pbt(e,t){return ke(Oe(Dd(tG(Lu(new Ft(null,new Bt(e.c.b,16)),new Wet(e)),t))))}function Kxe(e,t){return ke(Oe(Dd(tG(Lu(new Ft(null,new Bt(e.c.b,16)),new Vet(e)),t))))}function Ukn(){return us(),ce(se(Vge,1),Ve,259,0,[bte,Vf,V$,vte,hD,_A,W$,dD,fD,wte])}function zkn(){return gh(),ce(se(CDe,1),Ve,242,0,[qte,Fq,Bq,kDe,TDe,_De,xDe,Vte,w_,ux])}function oo(e,t,i,a,u,l,g,m,v,S,T,A,L){return Kyt(e,t,i,a,u,l,g,m,v,S,T,A,L),gle(e,!1),e}function $ue(e,t,i,a,u,l){var g;this.c=e,g=new et,A3e(e,g,t,e.b,i,a,u,l),this.a=new la(g,0)}function Hkn(e,t,i,a){var u;u=new AN,tE(u,"x",$Z(e,t,a.a)),tE(u,"y",FZ(e,t,a.b)),QN(i,u)}function Gkn(e,t,i,a){var u;u=new AN,tE(u,"x",$Z(e,t,a.a)),tE(u,"y",FZ(e,t,a.b)),QN(i,u)}function Yxe(e,t,i){var a,u;for(u=tr(e,0);u.b!=u.d.c;)a=f(Zn(u),8),a.a+=t,a.b+=i;return e}function pj(e,t,i){var a;for(a=e.b[i&e.f];a;a=a.b)if(i==a.a&&dm(t,a.g))return a;return null}function gj(e,t,i){var a;for(a=e.c[i&e.f];a;a=a.d)if(i==a.f&&dm(t,a.i))return a;return null}function Lbt(e){var t,i,a;for(t=0,a=e.Jc();a.Ob();)i=a.Pb(),t+=i!=null?Ti(i):0,t=~~t;return t}function Ba(e,t){var i,a,u;for(an(t),i=!1,u=t.Jc();u.Ob();)a=u.Pb(),i=i|e.Ec(a);return i}function qkn(e,t){var i,a;return i=f(Fe(e,(uw(),one)),17),a=f(Fe(t,one),17),tu(i.a,a.a)}function Vkn(e,t){var i;t.Sg("General Compactor",1),i=zCn(f(Fe(e,(uw(),dbe)),384)),i.Ag(e)}function Wkn(e,t,i){i.Sg("DFS Treeifying phase",1),e3n(e,t),x5n(e,t),e.a=null,e.b=null,i.Tg()}function Mf(e,t,i,a,u,l,g){ur.call(this,e,t),this.d=i,this.e=a,this.c=u,this.b=l,this.a=gm(g)}function Kkn(e,t,i){var a,u,l;for(a=0,u=0;u<i;u++)l=t[u],e[u]=l<<1|a,a=l>>>31;a!=0&&(e[i]=a)}function Dbt(e){var t;return e==0?"UTC":(e<0?(e=-e,t="UTC+"):t="UTC-",t+pmt(e))}function Fue(e){var t;return t=CTe(e),O7(t.a,0)?(Vk(),Vk(),Ype):(Vk(),new Dse(t.b))}function Bue(e){var t;return t=CTe(e),O7(t.a,0)?(Vk(),Vk(),Ype):(Vk(),new Dse(t.c))}function Ykn(e){var t;return t=sJ(e),O7(t.a,0)?(SY(),SY(),ZAt):(SY(),new xct(t.b))}function Xkn(e){return e.b.c.i.k==(cn(),bo)?f(V(e.b.c.i,(Ge(),oi)),11):e.b.c}function Mbt(e){return e.b.d.i.k==(cn(),bo)?f(V(e.b.d.i,(Ge(),oi)),11):e.b.d}function jbt(e){switch(e.g){case 2:return tt(),pn;case 4:return tt(),_n;default:return e}}function $bt(e){switch(e.g){case 1:return tt(),Bn;case 3:return tt(),hn;default:return e}}function Jkn(e,t){var i;return i=fAe(e),HAe(new Ke(i.c,i.d),new Ke(i.b,i.a),e.Kf(),t,e.$f())}function Zkn(e,t){t.Sg(axt,1),t3e(ogn(new rs((h7(),new Oce(e,!1,!1,new py))))),t.Tg()}function Nl(){Nl=Z,E5e=new a6,w5e=new Kg,y5e=new s6,v5e=new c6,S5e=new eO,_5e=new u6}function Uue(){Uue=Z,GF=new Art,b1e=ce(se(El,1),pA,178,0,[]),gMt=ce(se(Th,1),Z4e,62,0,[])}function Fbt(){Fbt=Z,U6t=vm(fst(Sr(Sr(new _o,(wa(),su),(ka(),Zee)),cu,Wee),zs),Jee)}function Xxe(){Xxe=Z,j6t=vm(fst(Sr(Sr(new _o,(wa(),su),(ka(),Zee)),cu,Wee),zs),Jee)}function zH(e,t){return mn(),ta(e)?YTe(e,dr(t)):Jk(e)?lce(e,Oe(t)):Xk(e)?Zvn(e,ut(t)):e.Dd(t)}function mj(e,t){var i,a;for(an(t),a=t.vc().Jc();a.Ob();)i=f(a.Pb(),43),e.yc(i.jd(),i.kd())}function Qkn(e,t,i){var a;for(a=i.Jc();a.Ob();)if(!dJ(e,t,a.Pb()))return!1;return!0}function eTn(e,t){var i;for(or(t);e.Ob();)if(i=e.Pb(),!tCe(f(i,9)))return!1;return!0}function tTn(e,t,i,a,u){var l;return i&&(l=Pi(t.zh(),e.c),u=i.Nh(t,-1-(l==-1?a:l),null,u)),u}function nTn(e,t,i,a,u){var l;return i&&(l=Pi(t.zh(),e.c),u=i.Ph(t,-1-(l==-1?a:l),null,u)),u}function Bbt(e){var t;if(e.b==-2){if(e.e==0)t=-1;else for(t=0;e.a[t]==0;t++);e.b=t}return e.b}function rTn(e){if(an(e),e.length==0)throw ie(new sm("Zero length BigInteger"));bLn(this,e)}function Jxe(e){this.i=e.gc(),this.i>0&&(this.g=this.Zi(this.i+(this.i/8|0)+1),e.Oc(this.g))}function Ubt(e,t,i){this.g=e,this.d=t,this.e=i,this.a=new et,GRn(this),Ht(),na(this.a,null)}function zbt(e,t,i){this.b=new Wn,this.f=new Wn,this.c=(Sj(),Hte),this.d=e,this.c=t,this.a=i}function Zxe(e,t,i,a){zbt.call(this,t,i,a),this.e=Pe(Ib,j0,9,e.a.c.length,0,1),kg(e.a,this.e)}function Qxe(e,t){t.q=e,e.d=h.Math.max(e.d,t.r),e.b+=t.d+(e.a.c.length==0?0:e.c),Qe(e.a,t)}function zue(e,t){var i,a,u,l;return u=e.c,i=e.c+e.b,l=e.d,a=e.d+e.a,t.a>u&&t.a<i&&t.b>l&&t.b<a}function wT(e,t){var i,a;for(a=tr(e,0);a.b!=a.d.c;)i=f(Zn(a),8),i.a+=t.a,i.b+=t.b;return e}function iTn(e){var t,i,a;for(a=0,i=new ne(e.b);i.a<i.c.c.length;)t=f(ee(i),25),t.p=a,++a}function oTn(e){var t,i,a;return e.j==(tt(),hn)&&(t=zEt(e),i=Cl(t,_n),a=Cl(t,pn),a||a&&i)}function HH(e,t){var i,a;for(an(t),a=e.vc().Jc();a.Ob();)i=f(a.Pb(),43),t.Wd(i.jd(),i.kd())}function XJ(e,t){var i;me(t,91)?(f(e.c,76).Dk(),i=f(t,91),ybt(e,i)):f(e.c,76).Wb(t)}function Hbt(e,t,i,a){me(e.Cb,184)&&(f(e.Cb,184).tb=null),ju(e,i),t&&a4n(e,t),a&&e.dl(!0)}function aTn(e,t){return t<e.b.gc()?f(e.b.Xb(t),9):t==e.b.gc()?e.a:f(ct(e.e,t-e.b.gc()-1),9)}function Hue(e,t,i,a){var u;u=e.a.length,i>u?i=u:gn(t,i+1),e.a=uh(e.a,0,t)+(""+a)+Gke(e.a,i)}function sTn(e,t){var i;t.Sg("Edge and layer constraint edge reversal",1),i=DPn(e),r9n(i),t.Tg()}function cTn(e,t){var i,a;for(a=new ne(t.b);a.a<a.c.c.length;)i=f(ee(a),25),e.a[i.p]=mNn(i)}function yT(e){this.f=e,this.e=new $Te(this.f.i),this.a=this.e,this.b=nCe(this),this.d=this.f.g}function eo(e,t){yX.call(this,mMt,e,t),this.b=this,this.a=Ju(e.zh(),Vt(this.e.zh(),this.c))}function Gbt(e,t){var i;return i=jIe(fj(),e),i?(pi(t,(Rr(),VA),i),!0):!1}function cE(e,t,i){var a,u;return u=f(U7(e.d,t),17),a=f(U7(e.b,i),17),!u||!a?null:sR(e,u.a,a.a)}function lf(e,t){var i;for(i=0;i<t.j.c.length;i++)f(DH(e,i),22).Fc(f(DH(t,i),16));return e}function uTn(){var e;return ege||(ege=new nrt,e=new iue(""),Kgn(e,(cP(),z6e)),Pkn(ege,e)),ege}function qbt(e){return tT(),e.A.Gc((Ud(),mx))&&!e.B.Gc((yd(),vV))?V0t(e):null}function Vbt(){this.a=f(st((Qh(),$ee)),17).a,this.c=ke(Oe(st(Fee))),this.b=ke(Oe(st(sge)))}function Wbt(){Wbt=Z,HOt=hr((cn(),ce(se(wge,1),Ve,247,0,[Yi,uo,bo,xu,Vc,Dg,_q,U$])))}function Kbt(){Kbt=Z,bNt=hr((aQ(),ce(se(cPe,1),Ve,272,0,[Age,Nge,Cge,Lge,Oge,Ige,Pge,Rge])))}function Ybt(){Ybt=Z,PLt=hr((bE(),ce(se(RLt,1),Ve,282,0,[Y8e,Po,Hs,i4,ya,Li,r4,Nb])))}function Xbt(){Xbt=Z,KDt=hr((pL(),ce(se($ne,1),Ve,243,0,[h1e,Mne,jne,Dne,f1e,Lne,Pne,d1e])))}function Jbt(){return Wj(),ce(se(djn,1),Ve,262,0,[Hme,YDe,QDe,eMe,ZDe,KDe,tMe,VDe,JDe,XDe,WDe])}function lTn(e,t,i){return di(UN(sL(e),new Ke(t.e.a,t.e.b)),UN(sL(e),new Ke(i.e.a,i.e.b)))}function dTn(e,t,i){return e==(PS(),Wte)?new k8:wd(t,1)!=0?new U2e(i.length):new hit(i.length)}function Zr(e,t){var i,a,u;if(i=e.ph(),i!=null&&e.sh())for(a=0,u=i.length;a<u;++a)i[a].aj(t)}function UP(e,t){var i,a;for(i=e,a=ra(i).e;a;){if(i=a,i==t)return!0;a=ra(i).e}return!1}function tw(e){var t;return t=e.h,t==0?e.l+e.m*OR:t==A1?e.l+e.m*OR-yE:e}function fTn(e,t,i){var a,u;return a=e.a.f[t.p],u=e.a.f[i.p],a<u?-1:a==u?0:1}function hTn(e,t){var i,a;for(a=new ne(t);a.a<a.c.c.length;)i=f(ee(a),72),Qe(e.d,i),bNn(e,i)}function Zbt(e){var t;e.d==null?(++e.e,e.f=0,h0t(null)):(++e.e,t=e.d,e.d=null,e.f=0,h0t(t))}function pTn(e){var t;if(e.a==e.b.a)throw ie(new ic);return t=e.a,e.c=t,e.a=f(ch(e.a.e),226),t}function gTn(e){Ast();var t;return Lot(zme,e)||(t=new Ev,t.a=e,I_e(zme,e,t)),f(Es(zme,e),630)}function JJ(){JJ=Z,pbe=new cse(Lp,0),KMe=new cse("POLAR_COORDINATE",1),WMe=new cse("ID",2)}function zP(){zP=Z,Wge=new Yae("ONE_SIDED",0),Kge=new Yae("TWO_SIDED",1),Iq=new Yae("OFF",2)}function HP(){HP=Z,F$=new Mi("debugSVG",(mn(),!1)),K6e=new Mi("overlapsExisted",!0)}function Qbt(){Qbt=Z,vDt=hr((NT(),ce(se(ps,1),Ve,94,0,[sp,D1,cp,lp,Pb,Sh,bf,up,Eh])))}function mTn(e,t){var i,a;for(a=new Nn(e);a.e!=a.i.gc();)i=f(Mn(a),27),Of(i,i.i+t.b,i.j+t.d)}function yc(e,t){var i,a;return i=(a=new $5,a),i.G=t,!e.rb&&(e.rb=new aT(e,Up,e)),Qn(e.rb,i),i}function Br(e,t){var i,a;return i=(a=new sY,a),i.G=t,!e.rb&&(e.rb=new aT(e,Up,e)),Qn(e.rb,i),i}function GH(e){var t,i,a;for(a=0,i=new LN(e.a);i.a<i.c.a.length;)t=aj(i),e.b.Gc(t)&&++a;return a}function bTn(e){var t,i,a;for(t=1,a=e.Jc();a.Ob();)i=a.Pb(),t=31*t+(i==null?0:Ti(i)),t=~~t;return t}function qH(e){var t;return e<128?(Clt(),t=_6e[e],!t&&(t=_6e[e]=new Ka(e)),t):new Ka(e)}function Nc(e){return HG<e&&e<yE?e<0?h.Math.ceil(e):h.Math.floor(e):tw(c5n(e))}function e1t(e,t){e.a=Ua(e.a,1),e.c=h.Math.min(e.c,t),e.b=h.Math.max(e.b,t),e.d=Ua(e.d,t)}function t1t(e,t){var i;return i=e.a.get(t),i===void 0?++e.d:(Pbn(e.a,t),--e.c,++e.b.g),i}function fn(e,t){var i;return(e.Db&t)!=0?(i=Wle(e,t),i==-1?e.Eb:uR(e.Eb)[i]):null}function ZJ(e){var t;if(e.g>1||e.Ob())return++e.a,e.g=0,t=e.i,e.Ob(),t;throw ie(new ic)}function vTn(e,t){var i;try{t.be()}catch(a){if(a=go(a),me(a,80))i=a,vn(e.c,i);else throw ie(a)}}function Rc(e,t){var i;return t&&(i=t.lf(),i.dc()||(e.q?mj(e.q,i):e.q=new rst(i))),e}function wTn(e,t){var i,a,u;return i=t.p-e.p,i==0?(a=e.f.a*e.f.b,u=t.f.a*t.f.b,di(a,u)):i}function dh(e){var t,i,a,u;return u=e,a=0,u<0&&(u+=yE,a=A1),i=Aa(u/OR),t=Aa(u-i*OR),Ku(t,i,a)}function yTn(e){return e.b.c.length!=0&&f(ct(e.b,0),72).a?f(ct(e.b,0),72).a:Ice(e)}function eCe(e,t){switch(t){case 1:return!!e.n&&e.n.i!=0;case 2:return e.k!=null}return TTe(e,t)}function n1t(e){switch(e.a.g){case 1:return new lat;case 3:return new uwt;default:return new M5}}function r1t(e,t){e.c&&(ASt(e,t,!0),Zi(new Ft(null,new Bt(t,16)),new Qet(e))),ASt(e,t,!1)}function uE(e,t,i){this.b=(an(e),e),this.d=(an(t),t),this.e=(an(i),i),this.c=this.d+(""+this.e)}function GP(e,t){this.b=e,this.e=t,this.d=t.j,this.f=(is(),f(e,69).uk()),this.k=Ju(t.e.zh(),e)}function VH(e){this.n=new et,this.e=new ki,this.j=new ki,this.k=new et,this.f=new et,this.p=e}function i1t(e){e.r=new Mo,e.w=new Mo,e.t=new et,e.i=new et,e.d=new Mo,e.a=new BN,e.c=new Wn}function DS(){DS=Z,yq=new RY("UP",0),wq=new RY(Gfe,1),ige=new RY(NR,2),oge=new RY(RR,3)}function QJ(){QJ=Z,NDe=new Zae("EQUALLY",0),Mme=new Zae("NORTH",1),RDe=new Zae("NORTH_SOUTH",2)}function o1t(){o1t=Z,l6t=hr((D0(),ce(se(vDe,1),Ve,265,0,[xme,gDe,mDe,Tme,pDe,bDe,zte,SD,_D])))}function a1t(){a1t=Z,IDt=hr((yd(),ce(se(l7e,1),Ve,264,0,[jD,vV,xne,FF,Cne,Ine,Ane,s1e,bV])))}function _g(){_g=Z,VR=new MY(PL,0),rD=new MY(NR,1),iD=new MY(RR,2),Tge=new MY("TOP",3)}function eZ(){eZ=Z,Ume=new rse("OFF",0),kD=new rse("SINGLE_EDGE",1),lx=new rse("MULTI_EDGE",2)}function bj(e,t,i,a,u){io(e.c[t.g],i.g,a),io(e.c[i.g],t.g,a),io(e.b[t.g],i.g,u),io(e.b[i.g],t.g,u)}function tZ(e,t,i){var a,u,l,g;g=ra(e),a=g.d,u=g.c,l=e.n,t&&(l.a=l.a-a.b-u.a),i&&(l.b=l.b-a.d-u.b)}function ETn(e,t){var i,a;return i=e.j,a=t.j,i!=a?i.g-a.g:e.p==t.p?0:i==(tt(),hn)?e.p-t.p:t.p-e.p}function STn(e){var t,i;for(k9n(e),i=new ne(e.d);i.a<i.c.c.length;)t=f(ee(i),105),t.i&&KNn(t)}function tCe(e){var t;return t=f(V(e,(Ge(),Tc)),64),e.k==(cn(),bo)&&(t==(tt(),pn)||t==_n)}function _Tn(e,t,i){var a,u;u=f(V(e,(ot(),Ls)),77),u&&(a=new fd,ale(a,0,u),wT(a,i),Ba(t,a))}function kTn(e,t,i){var a,u,l,g;return l=null,g=t,u=IS(g,dpe),a=new Nat(e,i),l=(Pwt(a.a,a.b,u),u),l}function WH(e,t){var i,a;return a=(i=new Q2,i),ju(a,t),Qn((!e.A&&(e.A=new xl(nl,e,7)),e.A),a),a}function Gue(e,t,i,a){return i==1?(!e.n&&(e.n=new Be(hc,e,1,7)),za(e.n,t,a)):yAe(e,t,i,a)}function TTn(e,t,i,a){f(i.b,68),f(i.b,68),f(a.b,68),f(a.b,68),f(a.b,68),Mu(a.a,new X_e(e,t,a))}function xTn(e,t){e.d==(ko(),Ds)||e.d==Wf?f(t.a,60).c.Ec(f(t.b,60)):f(t.b,60).c.Ec(f(t.a,60))}function KH(e,t){var i,a;for(an(t),a=t.Jc();a.Ob();)if(i=a.Pb(),!e.Gc(i))return!1;return!0}function CTn(e,t){var i,a,u;return i=e.l+t.l,a=e.m+t.m+(i>>22),u=e.h+t.h+(a>>22),Ku(i&Ed,a&Ed,u&A1)}function s1t(e,t){var i,a,u;return i=e.l-t.l,a=e.m-t.m+(i>>22),u=e.h-t.h+(a>>22),Ku(i&Ed,a&Ed,u&A1)}function ATn(e){var t;if(e){if(t=e,t.dc())throw ie(new ic);return t.Xb(t.gc()-1)}return Jft(e.Jc())}function que(e){var t;return(!e.a||(e.Bb&1)==0&&e.a.Rh())&&(t=ph(e),me(t,155)&&(e.a=f(t,155))),e.a}function go(e){var t;return me(e,80)?e:(t=e&&e.__java$exception,t||(t=new Y1t(e),Ynt(t)),t)}function Vue(e){if(me(e,193))return f(e,124);if(e)return null;throw ie(new IN(WCt))}function c1t(e){switch(e.g){case 0:return new Ep;case 1:return new ob;case 2:default:return null}}function ITn(e){switch(f(V(e,(ot(),el)),171).g){case 2:case 4:return!0;default:return!1}}function u1t(e,t){if(t==null)return!1;for(;e.a!=e.b;)if(ei(t,sZ(e)))return!0;return!1}function nCe(e){return e.a.Ob()?!0:e.a!=e.e?!1:(e.a=new axe(e.f.f),e.a.Ob())}function $o(e,t){var i,a;return i=t.Nc(),a=i.length,a==0?!1:(rke(e.c,e.c.length,i),!0)}function OTn(e,t){var i,a;return i=e.c,a=t.e[e.p],a<i.a.c.length-1?f(ct(i.a,a+1),9):null}function l1t(e,t){var i,a;for(a=new ne(e.b);a.a<a.c.c.length;)i=f(ee(a),72),$e(i,(Ge(),JR),t)}function NTn(e,t,i){var a,u;for(u=new ne(e.b);u.a<u.c.c.length;)a=f(ee(u),27),Of(a,a.i+t,a.j+i)}function Wue(e){var t,i,a;for(t=new ki,a=tr(e.d,0);a.b!=a.d.c;)i=f(Zn(a),65),Mr(t,i.c);return t}function rCe(e){var t,i,a,u;for(t=1,i=0,u=e.gc();i<u;++i)a=e.Si(i),t=31*t+(a==null?0:Ti(a));return t}function d1t(e){var t,i;return i=(t=new Q2,t),ju(i,"T"),Qn((!e.d&&(e.d=new xl(nl,e,11)),e.d),i),i}function f1t(e,t){var i;return i=lb(t.a.gc()),Zi(RJ(new Ft(null,new Bt(t,1)),e.i),new _at(e,i)),i}function h1t(e,t,i,a){var u;return _H(t,e.e.Pd().gc()),_H(i,e.c.Pd().gc()),u=e.a[t][i],io(e.a[t],i,a),u}function ce(e,t,i,a,u){return u.Om=e,u.Pm=t,u.Qm=xe,u.__elementTypeId$=i,u.__elementTypeCategory$=a,u}function RTn(e,t,i,a,u){return bL(),h.Math.min(Z_t(e,t,i,a,u),Z_t(i,a,e,t,fX(new Ke(u.a,u.b))))}function p1t(e,t){return!e||!t||e==t?!1:rvt(e.d.c,t.d.c+t.d.b)&&rvt(t.d.c,e.d.c+e.d.b)}function PTn(){return mJ(),Lg?new iue(null):FEt(uTn(),"com.google.common.base.Strings")}function g1t(e,t){if(!e)throw ie(new dn(Vj("value already present: %s",ce(se(Ro,1),Jt,1,5,[t]))))}function m1t(e,t,i){var a;nu(e.a),Mu(i.i,new Vtt(e)),a=new iX(f(bn(e.a,t.b),68)),gvt(e,a,t),i.f=a}function LTn(e){var t;VS(e,!0),t=M0,li(e,(ot(),wD))&&(t+=f(V(e,wD),17).a),$e(e,wD,rt(t))}function Kue(e){var t,i,a,u;for(u=new et,a=e.Jc();a.Ob();)i=f(a.Pb(),27),t=RT(i),$o(u,t);return u}function fR(e,t){var i,a;if(a=0,e<64&&e<=t)for(t=t<64?t:63,i=e;i<=t;i++)a=hm(a,hb(1,i));return a}function DTn(e,t){var i,a;for(JX(t,"predicate"),a=0;e.Ob();a++)if(i=e.Pb(),t.Lb(i))return a;return-1}function iCe(e,t){switch(t){case 0:!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),e.o.c.$b();return}pde(e,t)}function b1t(e){switch(e.g){case 1:return Nw;case 2:return Rb;case 3:return lV;default:return dV}}function oCe(e){var t,i;return i=(Bv(),t=new em,t),e&&Qn((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a),i),i}function MTn(e){var t;return t=new ae,t.a=e,t.b=UTn(e),t.c=Pe(ft,Ye,2,2,6,1),t.c[0]=Dbt(e),t.c[1]=Dbt(e),t}function aCe(e){Ht();var t,i,a;for(a=0,i=e.Jc();i.Ob();)t=i.Pb(),a=a+(t!=null?Ti(t):0),a=a|0;return a}function v1t(e){var t,i,a;return i=e.n,a=e.o,t=e.d,new qh(i.a-t.b,i.b-t.d,a.a+(t.b+t.c),a.b+(t.d+t.a))}function jTn(e,t){return!e||!t||e==t?!1:jS(e.b.c,t.b.c+t.b.b)<0&&jS(t.b.c,e.b.c+e.b.b)<0}function YH(e,t,i){switch(i.g){case 2:e.b=t;break;case 1:e.c=t;break;case 4:e.d=t;break;case 3:e.a=t}}function sCe(e,t){this.c=e,this.d=t,this.b=this.d/this.c.c.Pd().gc()|0,this.a=this.d%this.c.c.Pd().gc()}function w1t(e,t){uH(e>=0,"Negative initial capacity"),uH(t>=0,"Non-positive load factor"),nu(this)}function $Tn(e,t){var i,a;return i=f(f(bn(e.g,t.a),47).a,68),a=f(f(bn(e.g,t.b),47).a,68),o_t(i,a)}function FTn(e){bL();var t,i,a;for(i=Pe(Zo,Ye,8,2,0,1),a=0,t=0;t<2;t++)a+=.5,i[t]=yAn(a,e);return i}function cCe(e,t,i){var a;if(a=e.gc(),t>a)throw ie(new rT(t,a));return e.Pi()&&(i=Nft(e,i)),e.Bi(t,i)}function Yue(e,t,i){return e>=128?!1:e<64?N7(ia(hb(1,e),i),0):N7(ia(hb(1,e-64),t),0)}function y1t(e,t){for(;e.g==null&&!e.c?pTe(e):e.g==null||e.i!=0&&f(e.g[e.i-1],49).Ob();)t.Ai(XZ(e))}function uCe(){this.o=null,this.k=null,this.j=null,this.d=null,this.b=null,this.n=null,this.a=null}function qP(){qP=Z,a_=new PY("Q1",0),YT=new PY("Q4",1),KT=new PY("Q2",2),bA=new PY("Q3",3)}function nZ(){nZ=Z,kge=new DY(Lp,0),ate=new DY(uxt,1),ste=new DY(lxt,2),xq=new DY("BOTH",3)}function lE(){lE=Z,Fge=new Kae(Lp,0),lD=new Kae("INCOMING_ONLY",1),SA=new Kae("OUTGOING_ONLY",2)}function XH(){XH=Z,mne=new ASe("MINIMUM_SPANNING_TREE",0),V8e=new ASe("MAXIMUM_SPANNING_TREE",1)}function hR(){hR=Z,CLt=new $C,xLt=new BC}function E1t(){E1t=Z,ONt=hr((us(),ce(se(Vge,1),Ve,259,0,[bte,Vf,V$,vte,hD,_A,W$,dD,fD,wte])))}function S1t(){S1t=Z,p6t=hr((gh(),ce(se(CDe,1),Ve,242,0,[qte,Fq,Bq,kDe,TDe,_De,xDe,Vte,w_,ux])))}function Xue(){Xue=Z,Gpe={boolean:Jpn,number:fpn,string:hpn,object:Jyt,function:Jyt,undefined:Vhn}}function BTn(e){var t,i,a,u;for(i=e.a,a=0,u=i.length;a<u;++a)t=i[a],I1t(e,t,(tt(),Bn)),I1t(e,t,hn)}function _1t(e,t){var i,a,u,l;for(i=!1,a=e.a[t].length,l=0;l<a-1;l++)u=l+1,i=i|t3n(e,t,l,u);return i}function lCe(e){var t;return(!e.c||(e.Bb&1)==0&&(e.c.Db&64)!=0)&&(t=ph(e),me(t,88)&&(e.c=f(t,29))),e.c}function dCe(e){var t;if(e<0)return Ra;if(e==0)return 0;for(t=n$;(t&e)==0;t>>=1);return t}function UTn(e){var t;return e==0?"Etc/GMT":(e<0?(e=-e,t="Etc/GMT-"):t="Etc/GMT+",t+pmt(e))}function VP(e){var t,i,a;return t=~e.l+1&Ed,i=~e.m+(t==0?1:0)&Ed,a=~e.h+(t==0&&i==0?1:0)&A1,Ku(t,i,a)}function Jue(e){var t,i,a;t=~e.l+1&Ed,i=~e.m+(t==0?1:0)&Ed,a=~e.h+(t==0&&i==0?1:0)&A1,e.l=t,e.m=i,e.h=a}function k1t(e){var t,i;return t=e.t-e.k[e.o.p]*e.d+e.j[e.o.p]>e.f,i=e.u+e.e[e.o.p]*e.d>e.f*e.s*e.d,t||i}function vj(e){var t;return t=e.a[e.b],t==null?null:(io(e.a,e.b,null),e.b=e.b+1&e.a.length-1,t)}function T1t(e){var t,i;return i=new vJ,Rc(i,e),$e(i,(ew(),qR),e),t=new Wn,TDn(e,i,t),X8n(e,i,t),i}function $e(e,t,i){return i==null?(!e.q&&(e.q=new Wn),aR(e.q,t)):(!e.q&&(e.q=new Wn),qr(e.q,t,i)),e}function JH(e,t,i){return i==null?(!e.q&&(e.q=new Wn),aR(e.q,t)):(!e.q&&(e.q=new Wn),qr(e.q,t,i)),e}function x1t(e,t){return UN(sL(f(V(t,(Ec(),E_)),86)),new Ke(e.c.e.a-e.b.e.a,e.c.e.b-e.b.e.b))<=0}function C1t(e,t,i){var a,u;return a=new pue(t,i),u=new on,e.b=G2t(e,e.b,a,u),u.b||++e.c,e.b.b=!1,u.d}function A1t(e){var t,i;return i=EG(e.h),i==32?(t=EG(e.m),t==32?EG(e.l)+32:t+20-10):i-12}function nw(e){var t,i;for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),82),t.g.c=-t.g.c-t.g.b;tQ(e)}function rw(e){var t,i;for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),60),t.d.c=-t.d.c-t.d.b;wIe(e)}function zTn(e){var t,i,a;return t=f(e.e&&e.e(),10),a=(i=t.slice(),f(bue(i,t),10)),new Rf(t,a,t.length)}function I1t(e,t,i){var a,u,l,g;for(g=kj(t,i),l=0,u=g.Jc();u.Ob();)a=f(u.Pb(),11),qr(e.c,a,rt(l++))}function fCe(e,t){Ht();var i,a,u,l,g;for(g=!1,a=t,u=0,l=a.length;u<l;++u)i=a[u],g=g|e.Ec(i);return g}function uc(e){var t,i,a,u,l;for(t=new ha,a=e,u=0,l=a.length;u<l;++u)i=a[u],t.a+=i.a,t.b+=i.b;return t}function Zue(e,t,i){var a,u,l,g;return l=null,g=t,u=IS(g,"labels"),a=new Uat(e,i),l=(t2t(a.a,a.b,u),u),l}function O1t(e,t,i){var a;return a=new et,_Ie(e,t,a,(tt(),_n),!0,!1),_Ie(e,i,a,pn,!1,!1),a}function HTn(e,t,i,a){var u;return u=lIe(e,t,i,a),!u&&(u=fle(e,i,a),u&&!iA(e,t,u))?null:u}function GTn(e,t,i,a){var u;return u=uIe(e,t,i,a),!u&&(u=Wxn(e,i,a),u&&!iA(e,t,u))?null:u}function N1t(e,t){var i;for(i=0;i<e.a.a.length;i++)if(!f(glt(e.a,i),177).Lb(t))return!1;return!0}function qTn(e,t,i){if(or(t),i.Ob())for(LSe(t,Sdt(i.Pb()));i.Ob();)LSe(t,e.a),LSe(t,Sdt(i.Pb()));return t}function VTn(e,t,i,a,u){var l;return l=rIe(e,t),i&&Jue(l),u&&(e=xAn(e,t),a?ww=VP(e):ww=Ku(e.l,e.m,e.h)),l}function hCe(e){Ht();var t,i,a;for(a=1,i=e.Jc();i.Ob();)t=i.Pb(),a=31*a+(t!=null?Ti(t):0),a=a|0;return a}function S1(e){var t;return Kv(e),t=new ro,e.a.zd(t)?(dP(),new yae(an(t.a))):(dP(),dP(),F6e)}function dE(e){var t,i;for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),82),t.f.$b();spn(e.b,e),R2t(e)}function WTn(e,t){switch(t.g){case 0:me(e.b,626)||(e.b=new Vbt);break;case 1:me(e.b,627)||(e.b=new Qut)}}function pCe(e,t){switch(t){case 7:return!!e.e&&e.e.i!=0;case 8:return!!e.d&&e.d.i!=0}return o3e(e,t)}function R1t(e,t,i){e.g=gde(e,t,(tt(),_n),e.b),e.d=gde(e,i,_n,e.b),!(e.g.c==0||e.d.c==0)&&pyt(e)}function P1t(e,t,i){e.g=gde(e,t,(tt(),pn),e.j),e.d=gde(e,i,pn,e.j),!(e.g.c==0||e.d.c==0)&&pyt(e)}function KTn(e,t,i){return!oP(Jr(new Ft(null,new Bt(e.c,16)),new Y(new xat(t,i)))).zd((Wy(),zR))}function WP(){WP=Z,aPe=new u8,sPe=new Yg,rNt=new hO,nNt=new rk,tNt=new ik,oPe=(an(tNt),new de)}function Que(e,t,i){this.g=e,this.e=new ha,this.f=new ha,this.d=new ki,this.b=new ki,this.a=t,this.c=i}function ele(e,t,i,a){this.b=new et,this.n=new et,this.i=a,this.j=i,this.s=e,this.t=t,this.r=0,this.d=0}function gCe(e,t,i){this.q=new h.Date,this.q.setFullYear(e+hw,t,i),this.q.setHours(0,0,0,0),Yj(this,0)}function L1t(e,t){this.b=e,_3.call(this,(f(be(je((Wv(),sn).o),10),19),t.i),t.g),this.a=(Uue(),b1e)}function mCe(e,t,i){switch(t){case 0:!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),XJ(e.o,i);return}bde(e,t,i)}function D1t(e){switch(e.g){case 0:return new FC;default:throw ie(new dn(iee+(e.f!=null?e.f:""+e.g)))}}function M1t(e){switch(e.g){case 0:return new Gl;default:throw ie(new dn(iee+(e.f!=null?e.f:""+e.g)))}}function YTn(){Xr();var e;return _1e||(e=Kmn(dw("M",!0)),e=RX(dw("M",!1),e),_1e=e,_1e)}function bCe(e){var t;return e.b<=0?!1:(t=cm("MLydhHmsSDkK",gl(_a(e.c,0))),t>1||t>=0&&e.b<3)}function j1t(e){return Lg?Pe(tIt,vTt,563,0,0,1):f(kg(e.a,Pe(tIt,vTt,563,e.a.c.length,0,1)),835)}function XTn(e){var t,i;t=f(V(e,(Ge(),ed)),9),t&&(i=t.c,hl(i.a,t),i.a.c.length==0&&hl(ra(t).b,i))}function C0(e){var t,i;for(t=0,i=0;i<e.length;i++)t=(t<<5)-t+(gn(i,e.length),e.charCodeAt(i))|0;return t}function vCe(e){var t,i,a;for(t=new fd,a=tr(e,0);a.b!=a.d.c;)i=f(Zn(a),8),pP(t,0,new ja(i));return t}function $1t(e,t){var i,a,u;for(u=e.b;u;){if(i=e.a.Le(t,u.d),i==0)return u;a=i<0?0:1,u=u.a[a]}return null}function pR(e,t){var i,a,u,l,g;for(a=t,u=0,l=a.length;u<l;++u)i=a[u],g=new bdt(e),i.tf(g),$Dn(g);nu(e.f)}function JTn(e,t,i,a,u){var l,g;if(g=e.length,l=i.length,t<0||a<0||u<0||t+u>g||a+u>l)throw ie(new a2e)}function ZTn(e,t){var i,a;return i=f(Fe(e,(Ag(),Yq)),17).a,a=f(Fe(t,Yq),17).a,i==a||i<a?-1:i>a?1:0}function QTn(e){return Qe(e.c,(hR(),CLt)),_xe(e.a,ke(Oe(st((Sle(),Ute)))))?new Sp:new Ztt(e)}function exn(e){for(;!e.d||!e.d.Ob();)if(e.b&&!u7(e.b))e.d=f(JN(e.b),49);else return null;return e.d}function wCe(e){switch(e.g){case 1:return Ixt;default:case 2:return 0;case 3:return Oxt;case 4:return PRe}}function yCe(e,t){return A2t(e,t)?(jt(e.b,f(V(t,(Ge(),R1)),22),t),Mr(e.a,t),!0):!1}function tle(e,t){var i;return t===e?!0:me(t,228)?(i=f(t,228),ei(e.Zb(),i.Zb())):!1}function txn(e,t){return li(e,(Ge(),hi))&&li(t,hi)?f(V(t,hi),17).a-f(V(e,hi),17).a:0}function nxn(e,t){return li(e,(Ge(),hi))&&li(t,hi)?f(V(e,hi),17).a-f(V(t,hi),17).a:0}function KP(e,t){if(!e.Ii()&&t==null)throw ie(new dn("The 'no null' constraint is violated"));return t}function q3(e,t,i){var a,u;return u=(a=new gae,a),Ss(u,t,i),Qn((!e.q&&(e.q=new Be(Th,e,11,10)),e.q),u),u}function nle(e){var t,i,a,u;for(u=mgn(qDt,e),i=u.length,a=Pe(ft,Ye,2,i,6,1),t=0;t<i;++t)a[t]=u[t];return a}function F1t(e,t,i){var a,u;this.g=e,this.c=t,this.a=this,this.d=this,u=Sbt(i),a=Pe(RAt,BG,226,u,0,1),this.b=a}function ECe(e,t,i){var a;for(a=i-1;a>=0&&e[a]===t[a];a--);return a<0?0:$ae(ia(e[a],ls),ia(t[a],ls))?-1:1}function SCe(e,t){var i;t*2+1>=e.b.c.length||(SCe(e,2*t+1),i=2*t+2,i<e.b.c.length&&SCe(e,i),eEt(e,t))}function rxn(e,t){var i,a;for(a=tr(e,0);a.b!=a.d.c;)i=f(Zn(a),217),i.e.length>0&&(t.Ad(i),i.i&&Yxn(i))}function ixn(e,t){var i;return!e||e==t||!li(t,(Ge(),l_))?!1:(i=f(V(t,(Ge(),l_)),9),i!=e)}function rle(e){switch(e.i){case 2:return!0;case 1:return!1;case-1:++e.c;default:return e.Xl()}}function B1t(e,t,i){return e.d[t.p][i.p]||(eAn(e,t,i),e.d[t.p][i.p]=!0,e.d[i.p][t.p]=!0),e.a[t.p][i.p]}function YP(e,t){e.D==null&&e.B!=null&&(e.D=e.B,e.B=null),Rue(e,t==null?null:(an(t),t)),e.C&&e.el(null)}function U1t(e,t){var i;return i=new Ode((e.f&256)!=0,e.i,e.a,e.d,(e.f&16)!=0,e.j,e.g,t),e.e!=null||(i.c=e),i}function ile(e,t){var i,a;return a=f(fn(e.a,4),128),i=Pe(g1e,ype,412,t,0,1),a!=null&&au(a,0,i,0,a.length),i}function oxn(e,t){var i,a;for(a=e.Zb().Bc().Jc();a.Ob();)if(i=f(a.Pb(),16),i.Gc(t))return!0;return!1}function z1t(e,t,i){var a,u,l,g;for(an(i),g=!1,l=e.dd(t),u=i.Jc();u.Ob();)a=u.Pb(),l.Rb(a),g=!0;return g}function axn(e,t){var i;return e===t?!0:me(t,91)?(i=f(t,91),_Ae(Yy(e),i.vc())):!1}function H1t(e,t,i){var a,u;for(u=i.Jc();u.Ob();)if(a=f(u.Pb(),43),e.ze(t,a.kd()))return!0;return!1}function sxn(e,t,i,a){return VX(),new I2e(ce(se(AE,1),yQ,43,0,[(zle(e,t),new dS(e,t)),(zle(i,a),new dS(i,a))]))}function ole(){ole=Z,hge=Cft(ce(se(NF,1),Ve,86,0,[(ko(),Ds),Gs])),pge=Cft(ce(se(NF,1),Ve,86,0,[Wf,jg]))}function rZ(){rZ=Z,n7e=new gS(15),pDt=new pa((Rr(),V0),n7e),RF=o4,Z9e=zLt,Q9e=UE,t7e=GA,e7e=hx}function iZ(){iZ=Z,u1e=new YY("ELK",0),y7e=new YY("JSON",1),w7e=new YY("DOT",2),E7e=new YY("SVG",3)}function wj(){wj=Z,nbe=new BY(aNe,0),nne=new BY(Pxt,1),tbe=new BY("FAN",2),ebe=new BY("CONSTRAINT",3)}function yj(){yj=Z,nMe=new ase(Lp,0),Gme=new ase("MIDDLE_TO_MIDDLE",1),Gq=new ase("AVOID_OVERLAP",2)}function ZH(){ZH=Z,ine=new sse(Lp,0),IMe=new sse("RADIAL_COMPACTION",1),OMe=new sse("WEDGE_COMPACTION",2)}function Ej(){Ej=Z,$me=new ese("STACKED",0),jme=new ese("REVERSE_STACKED",1),Uq=new ese("SEQUENCED",2)}function fh(){fh=Z,H6e=new Bae("CONCURRENT",0),Zl=new Bae("IDENTITY_FINISH",1),G6e=new Bae("UNORDERED",2)}function _1(){_1=Z,kne=new gse(I4e,0),W0=new gse("INCLUDE_CHILDREN",1),PF=new gse("SEPARATE_CHILDREN",2)}function ale(e,t,i){var a,u,l;for(a=new ki,l=tr(i,0);l.b!=l.d.c;)u=f(Zn(l),8),Mr(a,new ja(u));z1t(e,t,a)}function cxn(e){var t,i,a;for(t=0,a=Pe(Zo,Ye,8,e.b,0,1),i=tr(e,0);i.b!=i.d.c;)a[t++]=f(Zn(i),8);return a}function uxn(e,t){var i,a;return i=f(e.d.Ac(t),16),i?(a=e.e.hc(),a.Fc(i),e.e.d-=i.gc(),i.$b(),a):null}function G1t(e){var t;if(t=e.a.c.length,t>0)return yP(t-1,e.a.c.length),_0(e.a,t-1);throw ie(new Jnt)}function _Ce(e,t){var i,a;if(a=e.c[t],a!=0)for(e.c[t]=0,e.d-=a,i=t+1;i<e.a.length;)e.a[i]-=a,i+=i&-i}function lxn(e,t){var i;t.Sg("Min Size Preprocessing",1),i=RAe(e),pi(e,(wb(),yF),i.a),pi(e,cne,i.b),t.Tg()}function dxn(e,t){var i;i=st((Sle(),Ute))!=null&&t.Qg()!=null?ke(Oe(t.Qg()))/ke(Oe(st(Ute))):1,qr(e.b,t,i)}function fxn(e,t){var i;return i=Ua(e,t),$ae(Yce(e,t),0)|ZY(Yce(e,i),0)?i:Ua(FG,Yce(Jy(i,63),1))}function kCe(e){var t;return t=(!e.a&&(e.a=new Be(j1,e,9,5)),e.a),t.i!=0?hgn(f(be(t,0),679)):null}function ET(e){var t;++e.j,e.i==0?e.g=null:e.i<e.g.length&&(t=e.g,e.g=e.Zi(e.i),au(t,0,e.g,0,e.i))}function hxn(e,t,i){if(t<0)throw ie(new Ru(Zxt+t));t<e.j.c.length?af(e.j,t,i):(Wht(e,t),Qe(e.j,i))}function q1t(e,t,i){if(e>t)throw ie(new dn(TQ+e+wTt+t));if(e<0||t>i)throw ie(new P2e(TQ+e+HOe+t+BOe+i))}function V1t(e){if(!e.a||(e.a.i&8)==0)throw ie(new As("Enumeration class expected for layout option "+e.f))}function W1t(e){Mft.call(this,"The given string does not match the expected format for individual spacings.",e)}function K1t(){opn.call(this,new oxe(mT(16))),uf(2,Zkt),this.b=2,this.a=new Xke(null,null,0,null),Ez(this.a,this.a)}function Y1t(e){Zrt(),rX(this),YX(this),this.e=e,F2t(this,e),this.g=e==null?bl:Qs(e),this.a="",this.b=e,this.a=""}function X1t(e,t,i,a){Fst(this),this.c=Pe(Ib,j0,9,e.a.c.length,0,1),this.e=t,kg(e.a,this.c),this.f=i,this.b=a}function TCe(e,t,i){var a,u;return fT(t,e.c.length),a=i.Nc(),u=a.length,u==0?!1:(rke(e.c,t,a),!0)}function pxn(e,t){var i,a;for(i=e.a.length-1;t!=e.b;)a=t-1&i,io(e.a,t,e.a[a]),t=a;io(e.a,e.b,null),e.b=e.b+1&i}function gxn(e,t){var i,a;for(i=e.a.length-1,e.c=e.c-1&i;t!=e.c;)a=t+1&i,io(e.a,t,e.a[a]),t=a;io(e.a,e.c,null)}function mxn(e){var t,i;if(e==null)return null;for(t=0,i=e.length;t<i;t++)if(!nlt(e[t]))return e[t];return null}function oZ(e){var t,i,a,u,l;for(l=1,i=e,a=0,u=i.length;a<u;++a)t=i[a],l=31*l+(t!=null?Ti(t):0),l=l|0;return l}function hr(e){var t,i,a,u,l;for(t={},a=e,u=0,l=a.length;u<l;++u)i=a[u],t[":"+(i.f!=null?i.f:""+i.g)]=i;return t}function J1t(e,t,i){var a,u,l,g;for(u=i,l=0,g=u.length;l<g;++l)if(a=u[l],e.b.ze(t,a.jd()))return a;return null}function sle(e,t){return!e||t&&!e.j||me(e,126)&&f(e,126).a.b==0?0:e.ff()}function aZ(e,t){return!e||t&&!e.k||me(e,126)&&f(e,126).a.a==0?0:e.gf()}function V3(e){return(e.c!=e.b.b||e.i!=e.g.b)&&(Bk(e.a.c,0),$o(e.a,e.b),$o(e.a,e.g),e.c=e.b.b,e.i=e.g.b),e.a}function Z1t(e){var t,i,a;for(a=0,i=new Tn(Cn(e.a.Jc(),new B));Kn(i);)t=f(Dn(i),18),t.c.i==t.d.i||++a;return a}function bxn(e){var t;for(or(e),G_e(!0,"numberToAdvance must be nonnegative"),t=0;t<0&&Kn(e);t++)Dn(e);return t}function Q1t(e,t){return li(e,(Ge(),hi))&&li(t,hi)?tu(f(V(e,hi),17).a,f(V(t,hi),17).a):0}function e0t(e,t){var i,a,u;i=e,u=0;do{if(i==t)return u;if(a=i.e,!a)throw ie(new Sz);i=ra(a),++u}while(!0)}function t0t(e,t){var i,a,u;for(u=t-e.f,a=new ne(e.d);a.a<a.c.c.length;)i=f(ee(a),316),evt(i,i.e,i.f+u);e.f=t}function fE(e,t){var i,a,u;a=e.Vk(t,null),u=null,t&&(u=(iP(),i=new cS,i),$P(u,e.r)),a=xg(e,u,a),a&&a.lj()}function vxn(e){var t;return t=new Pmt(e),SH(e.a,$Ot,new bu(ce(se(Sq,1),Jt,374,0,[t]))),t.d&&Qe(t.f,t.d),t.f}function n0t(e,t){var i;for(i=0;i<t.length;i++)if(e==(gn(i,t.length),t.charCodeAt(i)))return!0;return!1}function wxn(e,t){return t<e.length&&(gn(t,e.length),e.charCodeAt(t)!=63)&&(gn(t,e.length),e.charCodeAt(t)!=35)}function Fd(e){return ta(e)?ft:Jk(e)?mo:Xk(e)?to:xke(e)||Eke(e)?e.Om:e.Om||Array.isArray(e)&&se(PAt,1)||PAt}function yxn(e){return e?(e.i&1)!=0?e==kd?to:e==pr?ba:e==kx?eD:e==sa?mo:e==x_?i_:e==JA?o_:e==$l?BR:D$:e:null}function Exn(e,t){var i,a;for(a=wd(e.d,1)!=0,i=!0;i;)i=!1,i=t.c.kg(t.e,a),i=i|TG(e,t,a,!1),a=!a;Vxe(e)}function r0t(e,t,i,a){var u,l;e.a=t,l=a?0:1,e.f=(u=new Ayt(e.c,e.a,i,l),new uSt(i,e.a,u,e.e,e.b,e.c==(PS(),uF)))}function xCe(e,t){var i,a,u;return a=!1,i=t.q.d,t.d<e.b&&(u=bIe(t.q,e.b),t.q.d>u&&(Bwt(t.q,u),a=i!=t.q.d)),a}function i0t(e,t){var i,a,u,l,g,m,v,S;return v=t.i,S=t.j,a=e.f,u=a.i,l=a.j,g=v-u,m=S-l,i=h.Math.sqrt(g*g+m*m),i}function o0t(e,t){var i;for(i=0;i<t.length;i++)if(e==(gn(i,t.length),t.charCodeAt(i)))return!0;return!1}function Sxn(e){var t,i;for(i=new ne(e.r);i.a<i.c.c.length;)if(t=f(ee(i),9),e.n[t.p]<=0)return t;return null}function _xn(e){var t,i;if(e==null)return!1;for(t=0,i=e.length;t<i;t++)if(!nlt(e[t]))return!1;return!0}function kxn(e){var t;if(!(e.c.c<0?e.a>=e.c.b:e.a<=e.c.b))throw ie(new ic);return t=e.a,e.a+=e.c.c,++e.b,rt(t)}function QH(e,t){var i,a;return i=f(e.c.Ac(t),16),i?(a=e.hc(),a.Fc(i),e.d-=i.gc(),i.$b(),e.mc(a)):e.jc()}function CCe(e,t){var i,a;return a=EZ(e),a||(i=(Jde(),o2t(t)),a=new Bnt(i),Qn(a.Bl(),e)),a}function XP(){XP=Z,cV=new HY(I4e,0),_ne=new HY("CONTAINER",1),z9e=new HY("PARENT",2),H9e=new HY("ROOT",3)}function JP(){JP=Z,Fme=new tse("CONSERVATIVE",0),DDe=new tse("CONSERVATIVE_SOFT",1),cF=new tse("SLOPPY",2)}function Sj(){Sj=Z,Cme=new Jae("DUMMY_NODE_OVER",0),wDe=new Jae("DUMMY_NODE_UNDER",1),Hte=new Jae("EQUAL",2)}function ACe(){this.a=new W8,this.f=new Ftt(this),this.b=new Btt(this),this.i=new Utt(this),this.e=new ztt(this)}function ICe(e){throw Xue(),ie(new Krt("Unexpected typeof result '"+e+"'; please report this bug to the GWT team"))}function a0t(e){switch(e.i){case-2:return!0;case-1:return!1;case 1:--e.c;default:return e.Yl()}}function s0t(e){var t;if(e.c!=0)return e.c;for(t=0;t<e.a.length;t++)e.c=e.c*33+(e.a[t]&-1);return e.c=e.c*e.e,e.c}function sZ(e){var t;return Ln(e.a!=e.b),t=e.d.a[e.a],yct(e.b==e.d.c&&t!=null),e.c=e.a,e.a=e.a+1&e.d.a.length-1,t}function cle(e){var t;return t=new HSe(e.a),Rc(t,e),$e(t,(Ge(),oi),e),t.o.a=e.g,t.o.b=e.f,t.n.a=e.i,t.n.b=e.j,t}function Txn(e){var t;return t=nX(M6t),f(V(e,(Ge(),Cu)),22).Gc((us(),hD))&&Sr(t,(wa(),su),(ka(),tte)),t}function xxn(e){var t,i,a,u;for(u=new Mo,a=new ne(e);a.a<a.c.c.length;)i=f(ee(a),27),t=dPn(i),Ba(u,t);return u}function Cxn(e,t,i){var a,u;for(u=t.a.a.ec().Jc();u.Ob();)if(a=f(u.Pb(),60),gft(e,a,i))return!0;return!1}function Axn(e,t,i,a){var u,l;for(l=e.Jc();l.Ob();)u=f(l.Pb(),72),u.n.a=t.a+(a.a-u.o.a)/2,u.n.b=t.b,t.b+=u.o.b+i}function Ixn(e,t,i){var a;a=new IEt(e,t),jt(e.r,t.$f(),a),i&&!K7(e.u)&&(a.c=new kft(e.d),Mu(t.Pf(),new Xs(a)))}function Eu(e,t){var i;return Js(e)&&Js(t)&&(i=e-t,!isNaN(i))?i:z3e(Js(e)?dh(e):e,Js(t)?dh(t):t)}function ule(e,t,i){return h.Math.abs(t-e)<ZQ||h.Math.abs(i-e)<ZQ?!0:t-e>ZQ?e-i>ZQ:i-e>ZQ}function ST(e,t){or(e);try{return e.xc(t)}catch(i){if(i=go(i),me(i,211)||me(i,169))return null;throw ie(i)}}function Oxn(e,t){or(e);try{return e.Ac(t)}catch(i){if(i=go(i),me(i,211)||me(i,169))return null;throw ie(i)}}function c0t(e){switch(e.g){case 0:return new W2;case 1:return new Qg;case 2:return new V8;default:return null}}function OCe(e,t){var i,a,u;for(u=1,i=e,a=t>=0?t:-t;a>0;)a%2==0?(i*=i,a=a/2|0):(u*=i,a-=1);return t<0?1/u:u}function Nxn(e,t){var i,a,u;for(u=1,i=e,a=t>=0?t:-t;a>0;)a%2==0?(i*=i,a=a/2|0):(u*=i,a-=1);return t<0?1/u:u}function iw(e,t){var i,a,u,l;return l=(u=e?EZ(e):null,Xyt((a=t,u&&u.Dl(),a))),l==t&&(i=EZ(e),i&&i.Dl()),l}function u0t(e,t,i){var a,u;return u=e.b,e.b=t,(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,3,u,t),i?i.kj(a):i=a),i}function l0t(e,t,i){var a,u;return u=e.f,e.f=t,(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,0,u,t),i?i.kj(a):i=a),i}function NCe(e,t,i){var a,u;return u=e.a,e.a=t,(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,1,u,t),i?i.kj(a):i=a),i}function d0t(e,t,i){var a,u;for(u=e.Jc();u.Ob();){if(a=f(u.Pb(),9),a==t)return-1;if(a==i)return 1}return 0}function f0t(e,t){var i;return i=new Tg(e),Wl(i,(cn(),uo)),$e(i,(Ge(),oi),t),$e(i,(ot(),Xi),(oa(),uu)),i}function lle(e){return(tt(),wl).Gc(e.j)?ke(Oe(V(e,(Ge(),pD)))):uc(ce(se(Zo,1),Ye,8,0,[e.i.n,e.n,e.a])).b}function Rxn(e,t,i,a,u){t==0||a==0||(t==1?u[a]=S3e(u,i,a,e[0]):a==1?u[t]=S3e(u,e,t,i[0]):C4n(e,i,u,t,a))}function Pxn(e,t,i,a,u,l,g,m){var v;for(v=i;l<g;)v>=a||t<i&&m.Le(e[t],e[v])<=0?io(u,l++,e[t++]):io(u,l++,e[v++])}function Lxn(e,t,i){var a,u,l,g;for(a=i/e.gc(),u=0,g=e.Jc();g.Ob();)l=f(g.Pb(),186),t0t(l,l.f+a*u),DIn(l,t,a),++u}function Dxn(e){var t,i,a;for(a=0,i=new ne(e.a);i.a<i.c.c.length;)t=f(ee(i),172),a=h.Math.max(a,t.g);return a}function Mxn(e){var t,i,a;for(a=new ne(e.b);a.a<a.c.c.length;)i=f(ee(a),217),t=i.c.ig()?i.f:i.a,t&&gMn(t,i.j)}function h0t(e){var t,i;if(e!=null)for(i=0;i<e.length;++i)t=e[i],t&&(f(t.g,371),t.i)}function cZ(e){var t;switch(e.gc()){case 0:return $pe;case 1:return new nce(or(e.Xb(0)));default:return t=e,new Fce(t)}}function p0t(e){switch(lm(),e.gc()){case 0:return xce(),h6e;case 1:return new NN(e.Jc().Pb());default:return new nSe(e)}}function A0(e){switch(lm(),e.c){case 0:return xce(),h6e;case 1:return new NN(oEt(new LN(e)));default:return new qrt(e)}}function rt(e){var t,i;return e>-129&&e<128?(alt(),t=e+128,i=k6e[t],!i&&(i=k6e[t]=new KK(e)),i):new KK(e)}function ZP(e){var t,i;return e>-129&&e<128?(Tlt(),t=e+128,i=A6e[t],!i&&(i=A6e[t]=new Fs(e)),i):new Fs(e)}function g0t(e,t){var i;e.c.length!=0&&(i=f(kg(e,Pe(Ib,j0,9,e.c.length,0,1)),199),MSe(i,new oO),wEt(i,t))}function m0t(e,t){var i;e.c.length!=0&&(i=f(kg(e,Pe(Ib,j0,9,e.c.length,0,1)),199),MSe(i,new dC),wEt(i,t))}function b0t(e,t){var i;e.a.c.length>0&&(i=f(ct(e.a,e.a.c.length-1),561),yCe(i,t))||Qe(e.a,new Opt(t))}function jxn(e){tf();var t,i;t=e.d.c-e.e.c,i=f(e.g,152),Mu(i.b,new Ret(t)),Mu(i.c,new Pet(t)),Oa(i.i,new Let(t))}function v0t(e){var t;return t=new $v,t.a+="VerticalSegment ",vu(t,e.e),t.a+=" ",xr(t,ZSe(new Aae,new ne(e.k))),t.a}function $xn(e,t){var i;e.c=t,e.a=jCn(t),e.a<54&&(e.f=(i=t.d>1?Aht(t.a[0],t.a[1]):Aht(t.a[0],0),oE(t.e>0?i:T0(i))))}function dle(e,t){var i,a,u;for(i=0,u=lc(e,t).Jc();u.Ob();)a=f(u.Pb(),11),i+=V(a,(Ge(),ed))!=null?1:0;return i}function W3(e,t,i){var a,u,l;for(a=0,l=tr(e,0);l.b!=l.d.c&&(u=ke(Oe(Zn(l))),!(u>i));)u>=t&&++a;return a}function Fxn(e){var t;return t=f(Eg(e.c.c,""),232),t||(t=new nR(nP(rP(new xv,""),"Other")),pE(e.c.c,"",t)),t}function _j(e){var t;return(e.Db&64)!=0?Jh(e):(t=new sh(Jh(e)),t.a+=" (name: ",ys(t,e.zb),t.a+=")",t.a)}function RCe(e,t,i){var a,u;return u=e.sb,e.sb=t,(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,4,u,t),i?i.kj(a):i=a),i}function eG(e,t,i){var a;e.Yi(e.i+1),a=e.Wi(t,i),t!=e.i&&au(e.g,t,e.g,t+1,e.i-t),io(e.g,t,a),++e.i,e.Ji(t,i),e.Ki()}function PCe(e,t,i){var a,u;return u=e.r,e.r=t,(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,8,u,e.r),i?i.kj(a):i=a),i}function Bxn(e,t,i){var a,u;return a=new v1(e.e,3,13,null,(u=t.c,u||(Ut(),Ug)),P0(e,t),!1),i?i.kj(a):i=a,i}function Uxn(e,t,i){var a,u;return a=new v1(e.e,4,13,(u=t.c,u||(Ut(),Ug)),null,P0(e,t),!1),i?i.kj(a):i=a,i}function zxn(e,t){var i,a,u,l;if(t.bj(e.a),l=f(fn(e.a,8),1981),l!=null)for(i=l,a=0,u=i.length;a<u;++a)null.Rm()}function I0(e,t){var i,a;return i=f(t,676),a=i.bl(),!a&&i.cl(a=me(t,88)?new Yat(e,f(t,29)):new wht(e,f(t,155))),a}function w0t(e){return e-=e>>1&1431655765,e=(e>>2&858993459)+(e&858993459),e=(e>>4)+e&252645135,e+=e>>8,e+=e>>16,e&63}function LCe(e,t){switch(t){case 1:!e.n&&(e.n=new Be(hc,e,1,7)),Jn(e.n);return;case 2:FP(e,null);return}iCe(e,t)}function kj(e,t){switch(t.g){case 2:case 1:return lc(e,t);case 3:case 4:return Bd(lc(e,t))}return Ht(),Ht(),Ga}function tG(e,t){var i;return i=new ro,e.a.zd(i)?(dP(),new yae(an(tmt(e,i.a,t)))):(Kv(e),dP(),dP(),F6e)}function y0t(e){var t;return Eu(e,0)<0&&(e=tw(lwn(Js(e)?dh(e):e))),t=mr(Jy(e,32)),64-(t!=0?EG(t):EG(mr(e))+32)}function ei(e,t){return ta(e)?Mt(e,t):Jk(e)?wut(e,t):Xk(e)?(an(e),Ae(e)===Ae(t)):xke(e)?e.Fb(t):Eke(e)?ust(e,t):STe(e,t)}function Hxn(e,t){var i;return t.a&&(i=t.a.a.length,e.a?xr(e.a,e.b):e.a=new ef(e.d),Rht(e.a,t.a,t.d.length,i)),e}function Gxn(e){wQ();var t,i,a,u;for(i=Ole(),a=0,u=i.length;a<u;++a)if(t=i[a],Oc(t.a,e,0)!=-1)return t;return rge}function E0t(e,t){or(e);try{return e._b(t)}catch(i){if(i=go(i),me(i,211)||me(i,169))return!1;throw ie(i)}}function DCe(e,t){or(e);try{return e.Gc(t)}catch(i){if(i=go(i),me(i,211)||me(i,169))return!1;throw ie(i)}}function qxn(e,t){or(e);try{return e.Kc(t)}catch(i){if(i=go(i),me(i,211)||me(i,169))return!1;throw ie(i)}}function MCe(e,t,i,a){switch(t){case 1:return!e.n&&(e.n=new Be(hc,e,1,7)),e.n;case 2:return e.k}return N3e(e,t,i,a)}function jf(e,t,i){if(t<0)FAe(e,i);else{if(!i.ok())throw ie(new dn(vw+i.ve()+S$));f(i,69).tk().Bk(e,e.di(),t)}}function S0t(e,t,i){if(e<0||t>i)throw ie(new Ru(TQ+e+HOe+t+", size: "+i));if(e>t)throw ie(new dn(TQ+e+wTt+t))}function O0(e,t){var i;i=(e.Bb&256)!=0,t?e.Bb|=256:e.Bb&=-257,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,2,i,t))}function jCe(e,t){var i;i=(e.Bb&256)!=0,t?e.Bb|=256:e.Bb&=-257,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,8,i,t))}function $Ce(e,t){var i;i=(e.Bb&512)!=0,t?e.Bb|=512:e.Bb&=-513,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,9,i,t))}function N0(e,t){var i;i=(e.Bb&512)!=0,t?e.Bb|=512:e.Bb&=-513,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,3,i,t))}function uZ(e,t){var i;i=(e.Bb&256)!=0,t?e.Bb|=256:e.Bb&=-257,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,8,i,t))}function Vxn(e,t,i){var a,u;return u=e.a,e.a=t,(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,5,u,e.a),i?nAe(i,a):i=a),i}function _0t(e){var t;return(e.Db&64)!=0?Jh(e):(t=new sh(Jh(e)),t.a+=" (source: ",ys(t,e.d),t.a+=")",t.a)}function Tj(e,t){var i;return e.b==-1&&e.a&&(i=e.a.mk(),e.b=i?e.c.Dh(e.a.Ij(),i):Pi(e.c.zh(),e.a)),e.c.uh(e.b,t)}function k0t(e,t){var i,a;for(a=new Nn(e);a.e!=a.i.gc();)if(i=f(Mn(a),29),Ae(t)===Ae(i))return!0;return!1}function FCe(e){var t,i;return t=e.k,t==(cn(),bo)?(i=f(V(e,(Ge(),Tc)),64),i==(tt(),hn)||i==Bn):!1}function T0t(e){var t;return t=CTe(e),O7(t.a,0)?(Vk(),Vk(),Ype):(Vk(),new Dse(jae(t.a,0)?JTe(t)/oE(t.a):0))}function nG(e,t){this.e=t,this.a=y0t(e),this.a<54?this.f=oE(e):this.c=(ym(),Eu(e,0)>=0?aw(e):X7(aw(T0(e))))}function x0t(e,t,i,a,u,l){this.e=new et,this.f=(cs(),sF),Qe(this.e,e),this.d=t,this.a=i,this.b=a,this.f=u,this.c=l}function xj(e,t,i){var a;if(a=e.gc(),t>a)throw ie(new rT(t,a));if(e.Pi()&&e.Gc(i))throw ie(new dn(lq));e.Di(t,i)}function C0t(e,t){var i,a;for(a=new Nn(e);a.e!=a.i.gc();)if(i=f(Mn(a),141),Ae(t)===Ae(i))return!0;return!1}function Wxn(e,t,i){var a,u,l;return l=(u=gL(e.b,t),u),l&&(a=f(pQ(LH(e,l),""),29),a)?uIe(e,a,t,i):null}function fle(e,t,i){var a,u,l;return l=(u=gL(e.b,t),u),l&&(a=f(pQ(LH(e,l),""),29),a)?lIe(e,a,t,i):null}function Kxn(e){var t,i,a;for(a=0,i=e.length,t=0;t<i;t++)e[t]==32||e[t]==13||e[t]==10||e[t]==9||(e[a++]=e[t]);return a}function Yxn(e){var t;e.g&&(t=e.c.ig()?e.f:e.a,kIe(t.a,e.o,!0),kIe(t.a,e.o,!1),$e(e.o,(ot(),Xi),(oa(),HE)))}function Xxn(e){var t;return t=e.a[e.c-1&e.a.length-1],t==null?null:(e.c=e.c-1&e.a.length-1,io(e.a,e.c,null),t)}function A0t(e){var t,i;for(i=e.p.a.ec().Jc();i.Ob();)if(t=f(i.Pb(),216),t.f&&e.b[t.c]<-1e-10)return t;return null}function I0t(e){return e>=65&&e<=70?e-65+10:e>=97&&e<=102?e-97+10:e>=48&&e<=57?e-48:0}function di(e,t){return e<t?-1:e>t?1:e==t?e==0?di(1/e,1/t):0:isNaN(e)?isNaN(t)?0:1:-1}function Jxn(e,t){var i;if(i=kG(e,t),me(i,330))return f(i,35);throw ie(new dn(vw+t+"' is not a valid attribute"))}function BCe(e,t){switch(e.b.g){case 0:case 1:return t;case 2:case 3:return new qh(t.d,0,t.a,t.b);default:return null}}function Zxn(e){switch(e){case 0:return new Ert;case 1:return new wrt;case 2:return new yrt;default:throw ie(new Sz)}}function O0t(e){switch(e.g){case 2:return Gs;case 1:return Ds;case 4:return jg;case 3:return Wf;default:return $g}}function Qxn(e){switch(f(V(e,(Ge(),OE)),312).g){case 1:$e(e,OE,(w1(),WR));break;case 2:$e(e,OE,(w1(),kA))}}function N0t(){N0t=Z,J5t=vm(vm(b7(vm(vm(b7(Sr(new _o,(yR(),fF),(Wj(),Hme)),hF),ZDe),eMe),pF),KDe),tMe)}function rG(){rG=Z,dne=new hse("P1_STRUCTURE",0),fne=new hse("P2_PROCESSING_ORDER",1),hne=new hse("P3_EXECUTION",2)}function lZ(){lZ=Z,E8e=new fse("EQUAL_BETWEEN_STRUCTURES",0),S8e=new fse("TO_ASPECT_RATIO",1),Abe=new fse(Lp,2)}function gR(){gR=Z,d7e=new bse("PARALLEL_NODE",0),bx=new bse("HIERARCHICAL_NODE",1),One=new bse("ROOT_NODE",2)}function k1(){k1=Z,r1e=new GY(s$,0),uV=new GY("POLYLINE",1),DD=new GY("ORTHOGONAL",2),c4=new GY("SPLINES",3)}function ko(){ko=Z,$g=new Gz(s$,0),Gs=new Gz(RR,1),Ds=new Gz(NR,2),jg=new Gz(Gfe,3),Wf=new Gz("UP",4)}function eCn(e,t){t.Sg("Sort end labels",1),Zi(Jr(Zs(new Ft(null,new Bt(e.b,16)),new m6),new b6),new XM),t.Tg()}function QP(e,t){var i;i=(e.Bb&ep)!=0,t?e.Bb|=ep:e.Bb&=-1025,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,10,i,t))}function eL(e,t){var i;i=(e.Bb&FT)!=0,t?e.Bb|=FT:e.Bb&=-4097,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,12,i,t))}function tL(e,t){var i;i=(e.Bb&pw)!=0,t?e.Bb|=pw:e.Bb&=-2049,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,11,i,t))}function nL(e,t){var i;i=(e.Bb&Ml)!=0,t?e.Bb|=Ml:e.Bb&=-8193,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,15,i,t))}function hle(e,t){var i,a,u;e.d==null?(++e.e,--e.f):(u=t.jd(),i=t.xi(),a=(i&Qr)%e.d.length,VSn(e,a,d2t(e,a,i,u)))}function Cj(e,t,i){var a,u;return e.Mj()?(u=e.Nj(),a=yde(e,t,i),e.Gj(e.Fj(7,rt(i),a,t,u)),a):yde(e,t,i)}function tCn(e){var t,i;for(i=a2t(cf(e)).Jc();i.Ob();)if(t=dr(i.Pb()),Kj(e,t))return gSn((Rot(),rMt),t);return null}function nCn(e){var t,i,a;for(t=new et,a=new ne(e.b);a.a<a.c.c.length;)i=f(ee(a),586),$o(t,f(i.Af(),16));return t}function ple(e,t){var i,a;for(a=new ne(t);a.a<a.c.c.length;)i=f(ee(a),47),hl(e.b.b,i.b),Syn(f(i.a,194),f(i.b,82))}function rCn(e){var t,i;for(i=Qy(new $v,91),t=!0;e.Ob();)t||(i.a+=Fu),t=!1,vu(i,e.Pb());return(i.a+="]",i).a}function iCn(e,t){var i;return i=di(e.b.c,t.b.c),i!=0||(i=di(e.a.a,t.a.a),i!=0)?i:di(e.a.b,t.a.b)}function dZ(e){var t,i;return i=f(V(e,(ot(),mf)),86),i==(ko(),$g)?(t=ke(Oe(V(e,Tte))),t>=1?Gs:jg):i}function oCn(e){switch(f(V(e,(ot(),P1)),220).g){case 1:return new wC;case 3:return new _y;default:return new w8}}function UCe(e){switch(e.g){case 1:return pn;case 2:return hn;case 3:return _n;case 4:return Bn;default:return fc}}function iG(e){switch(e.g){case 1:return Bn;case 2:return pn;case 3:return hn;case 4:return _n;default:return fc}}function fZ(e){switch(e.g){case 1:return _n;case 2:return Bn;case 3:return pn;case 4:return hn;default:return fc}}function ow(e){if(e.c)ow(e.c);else if(e.d)throw ie(new As("Stream already terminated, can't be modified or used"))}function aCn(e){var t;if(!e.a)throw ie(new As("Cannot offset an unassigned cut."));t=e.c-e.b,e.b+=t,nft(e,t),tft(e,t)}function R0t(e,t,i){var a,u;if(e.c)gAe(e.c,t,i);else for(u=new ne(e.b);u.a<u.c.c.length;)a=f(ee(u),162),R0t(a,t,i)}function zCe(e,t){var i,a,u;for(an(t),i=!1,a=new ne(e);a.a<a.c.c.length;)u=ee(a),t.Gc(u)&&(Y7(a),i=!0);return i}function sCn(e,t){var i,a,u,l,g;for(g=Ju(e.e.zh(),t),l=0,i=f(e.g,122),u=0;u<e.i;++u)a=i[u],g.Zl(a.Ik())&&++l;return l}function MS(e,t,i){var a;return a=e.a.get(t),e.a.set(t,i===void 0?null:i),a===void 0?(++e.c,++e.b.g):++e.d,a}function cCn(e,t,i){var a,u;for(u=e.a.ec().Jc();u.Ob();)if(a=f(u.Pb(),9),KH(i,f(ct(t,a.p),16)))return a;return null}function uCn(e,t,i){var a,u;for(u=me(t,101)&&(f(t,19).Bb&Xa)!=0?new wse(t,e):new GP(t,e),a=0;a<i;++a)AZ(u);return u}function HCe(e,t,i){var a;return a=0,t&&(k3(e.a)?a+=t.f.a/2:a+=t.f.b/2),i&&(k3(e.a)?a+=i.f.a/2:a+=i.f.b/2),a}function lCn(e,t,i){var a;a=i,!a&&(a=uke(new CN,0)),a.Sg(WTt,2),Fvt(e.b,t,a.bh(1)),HMn(e,t,a.bh(1)),F9n(t,a.bh(1)),a.Tg()}function GCe(e,t,i){var a,u;return a=(Bv(),u=new Iv,u),MJ(a,t),DJ(a,i),e&&Qn((!e.a&&(e.a=new So(yf,e,5)),e.a),a),a}function dCn(e,t,i){var a,u;return a=f(t.mf(e.a),34),u=f(i.mf(e.a),34),a!=null&&u!=null?zH(a,u):a!=null?-1:u!=null?1:0}function lc(e,t){var i;return e.i||$Ae(e),i=f(Es(e.g,t),47),i?new Xv(e.j,f(i.a,17).a,f(i.b,17).a):(Ht(),Ht(),Ga)}function Ua(e,t){var i;return Js(e)&&Js(t)&&(i=e+t,HG<i&&i<yE)?i:tw(CTn(Js(e)?dh(e):e,Js(t)?dh(t):t))}function Na(e,t){var i;return Js(e)&&Js(t)&&(i=e*t,HG<i&&i<yE)?i:tw(AMn(Js(e)?dh(e):e,Js(t)?dh(t):t))}function hh(e,t){var i;return Js(e)&&Js(t)&&(i=e-t,HG<i&&i<yE)?i:tw(s1t(Js(e)?dh(e):e,Js(t)?dh(t):t))}function fCn(e,t,i){var a;try{qTn(e,t,i)}catch(u){throw u=go(u),me(u,590)?(a=u,ie(new UTe(a))):ie(u)}return t}function qCe(e){var t;return t=Pe(Xf,_m,30,2,15,1),e-=Xa,t[0]=(e>>10)+GG&To,t[1]=(e&1023)+56320&To,bm(t,0,t.length)}function rL(e,t){var i;i=(e.Bb&Sm)!=0,t?e.Bb|=Sm:e.Bb&=-16385,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,16,i,t))}function gle(e,t){var i;i=(e.Bb&Dc)!=0,t?e.Bb|=Dc:e.Bb&=-32769,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,18,i,t))}function VCe(e,t){var i;i=(e.Bb&Dc)!=0,t?e.Bb|=Dc:e.Bb&=-32769,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,18,i,t))}function WCe(e,t){var i;i=(e.Bb&Xa)!=0,t?e.Bb|=Xa:e.Bb&=-65537,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Vh(e,1,20,i,t))}function mle(e){var t;return(e.Db&64)!=0?Jh(e):(t=new sh(Jh(e)),t.a+=" (identifier: ",ys(t,e.k),t.a+=")",t.a)}function T1(e){var t,i,a;for(t=new et,a=new ne(e.j);a.a<a.c.c.length;)i=f(ee(a),11),Qe(t,i.b);return or(t),new L7(t)}function Fo(e){var t,i,a;for(t=new et,a=new ne(e.j);a.a<a.c.c.length;)i=f(ee(a),11),Qe(t,i.e);return or(t),new L7(t)}function ao(e){var t,i,a;for(t=new et,a=new ne(e.j);a.a<a.c.c.length;)i=f(ee(a),11),Qe(t,i.g);return or(t),new L7(t)}function hCn(e){var t,i;for(i=j4n(cf(dT(e))).Jc();i.Ob();)if(t=dr(i.Pb()),Kj(e,t))return mSn((Pot(),iMt),t);return null}function pCn(e){var t,i,a;for(i=0,a=e.length;i<a;i++)if(e[i]==null)throw ie(new IN("at index "+i));return t=e,new bu(t)}function P0t(e){var t,i,a,u;for(t=new Jct(e.Pd().gc()),u=0,a=cR(e.Pd().Jc());a.Ob();)i=a.Pb(),dEn(t,i,rt(u++));return HNn(t.a)}function gCn(e){var t;return t=AT(e),t>34028234663852886e22?Gi:t<-34028234663852886e22?qo:t}function At(e){var t,i,a,u;return i=(t=f(Tp((a=e.Om,u=a.f,u==ar?a:u)),10),new Rf(t,f(Gh(t,t.length),10),0)),Yh(i,e),i}function KCe(e,t){e.n.c.length==0&&Qe(e.n,new iJ(e.s,e.t,e.i)),Qe(e.b,t),C3e(f(ct(e.n,e.n.c.length-1),208),t),u_t(e,t)}function mCn(e,t,i){var a;i.Sg("Straight Line Edge Routing",1),i.ah(t,HRe),a=f(Fe(t,(R3(),MA)),27),S_t(e,a),i.ah(t,tee)}function ble(e,t){var i,a,u;for(u=0,a=f(t.Kb(e),20).Jc();a.Ob();)i=f(a.Pb(),18),lt(ut(V(i,(Ge(),U0))))||++u;return u}function bCn(e,t){var i,a,u;a=D3(t),u=ke(Oe(kT(a,(ot(),op)))),i=h.Math.max(0,u/2-.5),Lj(t,i,1),Qe(e,new Qot(t,i))}function vCn(e,t){return Mt(t.b&&t.c?iE(t.b)+"->"+iE(t.c):"e_"+Ti(t),e.b&&e.c?iE(e.b)+"->"+iE(e.c):"e_"+Ti(e))}function wCn(e,t){return Mt(t.b&&t.c?iE(t.b)+"->"+iE(t.c):"e_"+Ti(t),e.b&&e.c?iE(e.b)+"->"+iE(e.c):"e_"+Ti(e))}function jS(e,t){return gg(),Kh(fw),h.Math.abs(e-t)<=fw||e==t||isNaN(e)&&isNaN(t)?0:e<t?-1:e>t?1:pS(isNaN(e),isNaN(t))}function yCn(e){return ole(),mn(),!!(D0t(f(e.a,82).j,f(e.b,86))||f(e.a,82).d.e!=0&&D0t(f(e.a,82).j,f(e.b,86)))}function vle(){lAe();var e,t,i;i=Y7n+++Date.now(),e=Aa(h.Math.floor(i*VG))&kQ,t=Aa(i-e*FOe),this.a=e^1502,this.b=t^$fe}function ECn(e,t,i,a){var u,l,g;for(u=t+1;u<i;++u)for(l=u;l>t&&a.Le(e[l-1],e[l])>0;--l)g=e[l],io(e,l,e[l-1]),io(e,l-1,g)}function Tt(e,t){var i,a,u,l,g;if(i=t.f,pE(e.c.d,i,t),t.g!=null)for(u=t.g,l=0,g=u.length;l<g;++l)a=u[l],pE(e.c.e,a,t)}function L0t(e,t){var i,a;for(i=tr(e,0);i.b!=i.d.c;){if(a=_z(Oe(Zn(i))),a==t)return;if(a>t){_ht(i);break}}UX(i,t)}function SCn(e,t){var i;if(i=kG(e.zh(),t),me(i,101))return f(i,19);throw ie(new dn(vw+t+"' is not a valid reference"))}function hZ(e,t){if(t==e.d)return e.e;if(t==e.e)return e.d;throw ie(new dn("Node "+t+" not part of edge "+e))}function _Cn(e,t){switch(t.g){case 2:return e.b;case 1:return e.c;case 4:return e.d;case 3:return e.a;default:return!1}}function D0t(e,t){switch(t.g){case 2:return e.b;case 1:return e.c;case 4:return e.d;case 3:return e.a;default:return!1}}function YCe(e,t,i,a){switch(t){case 3:return e.f;case 4:return e.g;case 5:return e.i;case 6:return e.j}return MCe(e,t,i,a)}function $f(e,t,i,a){if(t<0)pIe(e,i,a);else{if(!i.ok())throw ie(new dn(vw+i.ve()+S$));f(i,69).tk().zk(e,e.di(),t,a)}}function vd(e){var t;if(e.b){if(vd(e.b),e.b.d!=e.c)throw ie(new If)}else e.d.dc()&&(t=f(e.f.c.xc(e.e),16),t&&(e.d=t))}function kCn(e){return e.k!=(cn(),Yi)?!1:j3(new Ft(null,new sT(new Tn(Cn(ao(e).a.Jc(),new B)))),new D2)}function TCn(e){return e.e==null?e:(!e.c&&(e.c=new Ode((e.f&256)!=0,e.i,e.a,e.d,(e.f&16)!=0,e.j,e.g,null)),e.c)}function xCn(e,t){return e.h==zG&&e.m==0&&e.l==0?(t&&(ww=Ku(0,0,0)),dst((MP(),v6e))):(t&&(ww=Ku(e.l,e.m,e.h)),Ku(0,0,0))}function Qs(e){var t;return Array.isArray(e)&&e.Qm===xe?Vy(Fd(e))+"@"+(t=Ti(e)>>>0,t.toString(16)):e.toString()}function Aj(e){var t;this.a=(t=f(e.e&&e.e(),10),new Rf(t,f(Gh(t,t.length),10),0)),this.b=Pe(Ro,Jt,1,this.a.a.length,5,1)}function XCe(e){Sle(),this.c=gm(ce(se(pjn,1),Jt,824,0,[c6t])),this.b=new Wn,this.a=e,qr(this.b,Ute,1),Mu(u6t,new Jtt(this))}function Ff(){Ff=Z,Lq=new jz(Lp,0),X$=new jz("FIRST",1),Tw=new jz(uxt,2),J$=new jz("LAST",3),tx=new jz(lxt,4)}function pZ(){pZ=Z,c1e=new WY(NRe,0),g7e=new WY("GROUP_DEC",1),b7e=new WY("GROUP_MIXED",2),m7e=new WY("GROUP_INC",3)}function gZ(){gZ=Z,ZMe=new lse("ASPECT_RATIO_DRIVEN",0),Sbe=new lse("MAX_SCALE_DRIVEN",1),JMe=new lse("AREA_DRIVEN",2)}function JCe(e,t,i,a,u){var l,g,m;for(g=u;t.b!=t.c;)l=f(JN(t),9),m=f(lc(l,a).Xb(0),11),e.d[m.p]=g++,vn(i.c,m);return g}function CCn(e){var t,i,a,u,l;return l=dAe(e),i=Az(e.c),a=!i,a&&(u=new oh,Wh(l,"knownLayouters",u),t=new Cnt(u),Oa(e.c,t)),l}function ZCe(e,t){var i,a,u,l,g,m;for(a=0,i=0,l=t,g=0,m=l.length;g<m;++g)u=l[g],u>0&&(a+=u,++i);return i>1&&(a+=e.d*(i-1)),a}function QCe(e){var t,i,a;for(a=new p0,a.a+="[",t=0,i=e.gc();t<i;)ys(a,$7(e.Si(t))),++t<i&&(a.a+=Fu);return a.a+="]",a.a}function wle(e,t,i){var a,u;for(a=0,u=e.length;a<u;a++)if(Yue((gn(a,e.length),e.charCodeAt(a)),t,i))return!0;return!1}function ACn(e){var t;return e==null?!0:(t=e.length,t>0&&(gn(t-1,e.length),e.charCodeAt(t-1)==58)&&!wle(e,zF,HF))}function e3e(e,t){var i;return Ae(e)===Ae(t)?!0:me(t,90)?(i=f(t,90),e.e==i.e&&e.d==i.d&&vSn(e,i.a)):!1}function mR(e){switch(tt(),e.g){case 4:return hn;case 1:return _n;case 3:return Bn;case 2:return pn;default:return fc}}function ICn(e){var t,i;if(e.b)return e.b;for(i=Lg?null:e.d;i;){if(t=Lg?null:i.b,t)return t;i=Lg?null:i.d}return cP(),z6e}function OCn(e,t){var i,a,u;for(a=YPn(e,t),u=a[a.length-1]/2,i=0;i<a.length;i++)if(a[i]>=u)return t.c+i;return t.c+t.b.gc()}function M0t(e,t){gP();var i,a,u,l;for(a=Ugt(e),u=t,LP(a,0,a.length,u),i=0;i<a.length;i++)l=aIn(e,a[i],i),i!=l&&Cj(e,i,l)}function t3e(e){var t,i,a;for(a=ke(Oe(e.a.mf((Rr(),Ene)))),i=new ne(e.a.Qf());i.a<i.c.c.length;)t=f(ee(i),680),_kt(e,t,a)}function NCn(e){tT();var t,i,a,u;for(t=e.o.b,a=f(f(si(e.r,(tt(),Bn)),22),83).Jc();a.Ob();)i=f(a.Pb(),115),u=i.e,u.b+=t}function RCn(e,t,i){var a,u;for(u=e.a.b,a=u.c.length;a<i;a++)Ky(u,0,new ru(e.a));jo(t,f(ct(u,u.c.length-i),25)),e.b[t.p]=i}function yle(e,t){var i,a;for(a=new ne(t);a.a<a.c.c.length;)i=f(ee(a),47),Qe(e.b.b,f(i.b,82)),Jce(f(i.a,194),f(i.b,82))}function j0t(e,t){var i;return e.d?Yu(e.b,t)?f(bn(e.b,t),44):(i=t.bg(),qr(e.b,t,i),i):t.bg()}function PCn(e){var t=e.e;function i(a){return!a||a.length==0?"":"	"+a.join(`
	`)}return t&&(t.stack||i(e[pfe]))}function Bd(e){var t,i;return me(e,308)?(i=GSn(f(e,308)),t=i,t):me(e,429)?f(e,429).a:me(e,59)?new Vrt(e):new Y2e(e)}function LCn(e,t,i,a,u){rf(),Zh(Uh(Bh(Fh(zh(new ah,0),u.d.e-e),t),u.d)),Zh(Uh(Bh(Fh(zh(new ah,0),i-u.a.e),u.a),a))}function $0t(e,t){var i,a,u,l;t&&(u=E1(t,"x"),i=new wnt(e),B3(i.a,(an(u),u)),l=E1(t,"y"),a=new ynt(e),H3(a.a,(an(l),l)))}function F0t(e,t){var i,a,u,l;t&&(u=E1(t,"x"),i=new Snt(e),U3(i.a,(an(u),u)),l=E1(t,"y"),a=new _nt(e),z3(a.a,(an(l),l)))}function DCn(e,t){var i,a,u,l;for(u=new Du(t.gc()),a=t.Jc();a.Ob();)i=a.Pb(),l=$de(e,f(i,57)),l&&vn(u.c,l);return u}function _T(e,t,i){var a,u;for(u=e.Jc();u.Ob();)if(a=u.Pb(),Ae(t)===Ae(a)||t!=null&&ei(t,a))return i&&u.Qb(),!0;return!1}function n3e(e,t,i){var a,u;if(++e.j,i.dc())return!1;for(u=i.Jc();u.Ob();)a=u.Pb(),e.nj(t,e.Wi(t,a)),++t;return!0}function MCn(e,t,i,a){var u,l;if(l=i-t,l<3)for(;l<3;)e*=10,++l;else{for(u=1;l>3;)u*=10,--l;e=(e+(u>>1))/u|0}return a.i=e,!0}function jCn(e){var t,i,a;return e.e==0?0:(t=e.d<<5,i=e.a[e.d-1],e.e<0&&(a=Bbt(e),a==e.d-1&&(--i,i=i|0)),t-=EG(i),t)}function $Cn(e){var t,i,a;return e<Nee.length?Nee[e]:(i=e>>5,t=e&31,a=Pe(pr,Pr,30,i+1,15,1),a[i]=1<<t,new Zy(1,i+1,a))}function B0t(e){var t,i,a;return i=e.hh(),i?(t=e.Ah(),me(t,173)&&(a=B0t(f(t,173)),a!=null)?a+"."+i:i):null}function r3e(e){var t;return!e.c&&(e.c=new _2),na(e.d,new dv),mPn(e),t=aPn(e),Zi(new Ft(null,new Bt(e.d,16)),new s1(e)),t}function FCn(e,t){t.Sg("End label post-processing",1),Zi(Jr(Zs(new Ft(null,new Bt(e.b,16)),new p6),new KM),new YM),t.Tg()}function BCn(e){var t,i,a,u;for(t=(e.j==null&&(e.j=(AP(),u=Hpe.ke(e),GIn(u))),e.j),i=0,a=t.length;i<a;++i);}function i3e(e){var t,i,a,u;for(t=new oh,u=new p3(e.b.Jc());u.b.Ob();)a=f(u.b.Pb(),685),i=XNn(a),gwn(t,t.a.length,i);return t.a}function U0t(e,t){var i,a;if(t){for(i=0;i<e.i;++i)if(a=f(e.g[i],370),a.jj(t))return!1;return Qn(e,t)}else return!1}function o3e(e,t){switch(t){case 3:return e.f!=0;case 4:return e.g!=0;case 5:return e.i!=0;case 6:return e.j!=0}return eCe(e,t)}function a3e(e,t){switch(t){case 3:NS(e,0);return;case 4:RS(e,0);return;case 5:md(e,0);return;case 6:bd(e,0);return}LCe(e,t)}function z0t(e){switch(e.g){case 0:return new e5;case 1:return new UO;default:throw ie(new dn(Zfe+(e.f!=null?e.f:""+e.g)))}}function UCn(e){switch(e.g){case 0:return new G2;case 1:return new t5;default:throw ie(new dn(Bhe+(e.f!=null?e.f:""+e.g)))}}function zCn(e){switch(e.g){case 1:return new Q6;case 2:return new kct;default:throw ie(new dn(Bhe+(e.f!=null?e.f:""+e.g)))}}function H0t(e){switch(e.g){case 0:return new _2e;case 1:return new Srt;default:throw ie(new dn(iee+(e.f!=null?e.f:""+e.g)))}}function s3e(e){_S();var t;switch(t=e.Nc(),t.length){case 0:return $pe;case 1:return new nce(or(t[0]));default:return new Fce(pCn(t))}}function $S(e,t){switch(t.g){case 1:return WN(e.j,(Nl(),w5e));case 2:return WN(e.j,(Nl(),E5e));default:return Ht(),Ht(),Ga}}function HCn(e,t){var i;return i=f($d(e,sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),i.Oc(ult(i.gc()))}function G0t(e,t){var i,a;if(a=new g3(e.a.$c(t,!0)),a.a.gc()<=1)throw ie(new kN);return i=a.a.ec().Jc(),i.Pb(),f(i.Pb(),39)}function Pi(e,t){var i,a,u;if(i=(e.i==null&&Em(e),e.i),a=t.Ij(),a!=-1){for(u=i.length;a<u;++a)if(i[a]==t)return a}return-1}function Ele(e,t){var i,a,u;return e.f>0?(e.Yj(),a=t==null?0:Ti(t),u=(a&Qr)%e.d.length,i=d2t(e,u,a,t),i!=-1):!1}function c3e(e,t){var i;return e.i>0&&(t.length<e.i&&(i=BH(Fd(t).c,e.i),t=i),au(e.g,0,t,0,e.i)),t.length>e.i&&io(t,e.i,null),t}function mZ(e){var t;return(e.Db&64)!=0?_j(e):(t=new sh(_j(e)),t.a+=" (instanceClassName: ",ys(t,e.D),t.a+=")",t.a)}function q0t(e){var t,i;t=e.d==(ER(),oD),i=tAe(e),t&&!i||!t&&i?$e(e.a,(ot(),Nm),(vb(),eV)):$e(e.a,(ot(),Nm),(vb(),Qq))}function GCn(e){_J(),f(e.mf((Rr(),Ow)),181).Gc((yd(),Ane))&&(f(e.mf(px),181).Ec((Xl(),u4)),f(e.mf(Ow),181).Kc(Ane))}function Sle(){Sle=Z,_Y(),Ute=(ot(),v_),u6t=gm(ce(se(Fbe,1),ORe,144,0,[jq,op,OA,b_,cx,yme,yD,ED,Eme,rF,IA,jE,NA]))}function qCn(e){var t,i,a,u,l;for(i=f(e.g,674),a=e.i-1;a>=0;--a)for(t=i[a],u=0;u<a;++u)if(l=i[u],n_t(e,t,l)){wR(e,a);break}}function oG(e,t){var i,a,u,l;for(l=Ju(e.e.zh(),t),i=f(e.g,122),u=0;u<e.i;++u)if(a=i[u],l.Zl(a.Ik()))return!1;return!0}function VCn(e,t,i){var a,u;return a=ke(e.p[t.i.p])+ke(e.d[t.i.p])+t.n.b+t.a.b,u=ke(e.p[i.i.p])+ke(e.d[i.i.p])+i.n.b+i.a.b,u-a}function WCn(e,t,i){var a,u;for(a=ia(i,ls),u=0;Eu(a,0)!=0&&u<t;u++)a=Ua(a,ia(e[u],ls)),e[u]=mr(a),a=wS(a,32);return mr(a)}function bZ(e){var t,i,a,u;for(u=0,i=0,a=e.length;i<a;i++)t=(gn(i,e.length),e.charCodeAt(i)),t<64&&(u=hm(u,hb(1,t)));return u}function $u(e,t,i){var a,u,l;return e.Mj()?(a=e.i,l=e.Nj(),eG(e,a,t),u=e.Fj(3,null,t,a,l),i?i.kj(u):i=u):eG(e,e.i,t),i}function KCn(e,t,i){var a,u;return a=new v1(e.e,3,10,null,(u=t.c,me(u,88)?f(u,29):(Ut(),Ch)),P0(e,t),!1),i?i.kj(a):i=a,i}function YCn(e,t,i){var a,u;return a=new v1(e.e,4,10,(u=t.c,me(u,88)?f(u,29):(Ut(),Ch)),null,P0(e,t),!1),i?i.kj(a):i=a,i}function V0t(e){tT();var t;return t=new ja(f(e.e.mf((Rr(),GA)),8)),e.B.Gc((yd(),jD))&&(t.a<=0&&(t.a=20),t.b<=0&&(t.b=20)),t}function vZ(e,t){jSe();var i,a;return i=iH((lY(),lY(),L$)),a=null,t==i&&(a=f(wu(m6e,e),609)),a||(a=new ddt(e),t==i&&Os(m6e,e,a)),a}function W0t(e){xT();var t;return(e.q?e.q:(Ht(),Ht(),Cb))._b((ot(),g_))?t=f(V(e,g_),203):t=f(V(ra(e),nF),203),t}function aw(e){ym();var t,i;return i=mr(e),t=mr(Jy(e,32)),t!=0?new Eht(i,t):i>10||i<0?new m1(1,i):GAt[i]}function Rp(e,t){var i,a,u;return e.f>0&&(e.Yj(),a=t==null?0:Ti(t),u=(a&Qr)%e.d.length,i=VAe(e,u,a,t),i)?i.kd():null}function K0t(e,t){var i,a,u;return me(t,43)?(i=f(t,43),a=i.jd(),u=ST(e.Pc(),a),dm(u,i.kd())&&(u!=null||e.Pc()._b(a))):!1}function Y0t(e){if(e.b==null){for(;e.a.Ob();)if(e.b=e.a.Pb(),!f(e.b,52).Fh())return!0;return e.b=null,!1}else return!0}function X0t(e,t,i){Hmt(),prt.call(this),this.a=iT(hIt,[Ye,VOe],[587,215],0,[Dee,nge],2),this.c=new BN,this.g=e,this.f=t,this.d=i}function J0t(e){this.e=Pe(pr,Pr,30,e.length,15,1),this.c=Pe(kd,Tb,30,e.length,16,1),this.b=Pe(kd,Tb,30,e.length,16,1),this.f=0}function XCn(e){var t,i;for(e.j=Pe(sa,_s,30,e.p.c.length,15,1),i=new ne(e.p);i.a<i.c.c.length;)t=f(ee(i),9),e.j[t.p]=t.o.b/e.i}function JCn(e){var t,i,a,u;for(a=yRn(e),na(a,VOt),u=e.d,u.c.length=0,i=new ne(a);i.a<i.c.c.length;)t=f(ee(i),452),$o(u,t.b)}function aG(e,t){var i,a;for(a=t.length,i=0;i<a;i+=2)Su(e,(gn(i,t.length),t.charCodeAt(i)),(gn(i+1,t.length),t.charCodeAt(i+1)))}function Z0t(e,t){var i,a;e.a=Ua(e.a,1),e.c=h.Math.min(e.c,t),e.b=h.Math.max(e.b,t),e.d+=t,i=t-e.f,a=e.e+i,e.f=a-e.e-i,e.e=a}function ZCn(e,t){t.Sg("Min Size Postprocessing",1),pi(e,(wb(),fx),h.Math.max(ke(Oe(Fe(e,fx))),ke(Oe(Fe(e,yF))))),t.Tg()}function Q0t(e,t){e.Uj();try{e.d._c(e.e++,t),e.f=e.d.j,e.g=-1}catch(i){throw i=go(i),me(i,97)?ie(new If):ie(i)}}function _le(){_le=Z,w1e=new dg,B7e=new o9,U7e=new Ny,z7e=new Z2,H7e=new Nv,G7e=new p5,q7e=new g5,V7e=new u0,W7e=new a9}function u3e(){u3e=Z,s7e=new gS(15),kDt=new pa((Rr(),V0),s7e),xDt=new pa(zE,15),TDt=new pa(t1e,rt(0)),_Dt=new pa(ID,DL)}function QCn(){return Error.stackTraceLimit>0?(h.Error.stackTraceLimit=Error.stackTraceLimit=64,!0):"stack"in new Error}function kle(e,t){var i;return Js(e)&&Js(t)&&(i=e%t,HG<i&&i<yE)?i:tw((dOe(Js(e)?dh(e):e,Js(t)?dh(t):t,!0),ww))}function Ij(e,t){var i;t9n(t),i=f(V(e,(ot(),Rte)),281),i&&$e(e,Rte,YAn(i)),w0(e.c),w0(e.f),txe(e.d),txe(f(V(e,Mte),213))}function kT(e,t){var i,a;return a=null,li(e,(ot(),Fte))&&(i=f(V(e,Fte),102),i.nf(t)&&(a=i.mf(t))),a==null&&(a=V(ra(e),t)),a}function e3n(e,t){var i,a,u,l;for(l=t.b.b,e.a=new ki,e.b=Pe(pr,Pr,30,l,15,1),i=0,u=tr(t.b,0);u.b!=u.d.c;)a=f(Zn(u),39),a.g=i++}function evt(e,t,i){var a,u,l,g;for(l=t-e.e,g=i-e.f,u=new ne(e.a);u.a<u.c.c.length;)a=f(ee(u),172),cG(a,a.s+l,a.t+g);e.e=t,e.f=i}function l3e(e,t,i){var a,u,l,g,m;return g=e.k,m=t.k,a=i[g.g][m.g],u=Oe(kT(e,a)),l=Oe(kT(t,a)),h.Math.max((an(u),u),(an(l),l))}function d3e(e,t,i){var a,u;a=f(wu(BD,t),121),u=f(wu(YF,t),121),i?(Os(BD,e,a),Os(YF,e,u)):(Os(YF,e,a),Os(BD,e,u))}function tvt(e,t){var i,a,u,l;return i=t>>5,t&=31,u=e.d+i+(t==0?0:1),a=Pe(pr,Pr,30,u,15,1),VOn(a,e.a,i,t),l=new Zy(e.e,u,a),nj(l),l}function iL(e,t,i){var a,u,l;for(u=null,l=e.b;l;){if(a=e.a.Le(t,l.d),i&&a==0)return l;a>=0?l=l.a[1]:(u=l,l=l.a[0])}return u}function sG(e,t,i){var a,u,l;for(u=null,l=e.b;l;){if(a=e.a.Le(t,l.d),i&&a==0)return l;a<=0?l=l.a[0]:(u=l,l=l.a[1])}return u}function t3n(e,t,i,a){var u,l,g;return u=!1,_8n(e.f,i,a)&&(k3n(e.f,e.a[t][i],e.a[t][a]),l=e.a[t],g=l[a],l[a]=l[i],l[i]=g,u=!0),u}function nvt(e,t,i){var a,u,l,g;for(u=f(bn(e.b,i),183),a=0,g=new ne(t.j);g.a<g.c.c.length;)l=f(ee(g),112),u[l.d.p]&&++a;return a}function Tle(e,t){for(var i=0;!t[i]||t[i]=="";)i++;for(var a=t[i++];i<t.length;i++)!t[i]||t[i]==""||(a+=e+t[i]);return a}function n3n(e,t){return gg(),gg(),Kh(fw),(h.Math.abs(e-t)<=fw||e==t||isNaN(e)&&isNaN(t)?0:e<t?-1:e>t?1:pS(isNaN(e),isNaN(t)))>0}function f3e(e,t){return gg(),gg(),Kh(fw),(h.Math.abs(e-t)<=fw||e==t||isNaN(e)&&isNaN(t)?0:e<t?-1:e>t?1:pS(isNaN(e),isNaN(t)))<0}function rvt(e,t){return gg(),gg(),Kh(fw),(h.Math.abs(e-t)<=fw||e==t||isNaN(e)&&isNaN(t)?0:e<t?-1:e>t?1:pS(isNaN(e),isNaN(t)))<=0}function r3n(e){var t;t=e.a;do t=f(Dn(new Tn(Cn(ao(t).a.Jc(),new B))),18).d.i,t.k==(cn(),uo)&&Qe(e.e,t);while(t.k==(cn(),uo))}function i3n(e,t){var i,a,u;for(a=new Tn(Cn(ao(e).a.Jc(),new B));Kn(a);)if(i=f(Dn(a),18),u=i.d.i,u.c==t)return!1;return!0}function ivt(e){var t,i;return t=f(fn(e.a,4),128),t!=null?(i=Pe(g1e,ype,412,t.length,0,1),au(t,0,i,0,t.length),i):YDt}function ovt(e){var t,i,a,u,l;if(e==null)return null;for(l=new et,i=nle(e),a=0,u=i.length;a<u;++a)t=i[a],Qe(l,_u(t,!0));return l}function avt(e){var t,i,a,u,l;if(e==null)return null;for(l=new et,i=nle(e),a=0,u=i.length;a<u;++a)t=i[a],Qe(l,_u(t,!0));return l}function svt(e){var t,i,a,u,l;if(e==null)return null;for(l=new et,i=nle(e),a=0,u=i.length;a<u;++a)t=i[a],Qe(l,_u(t,!0));return l}function bm(e,t,i){var a,u,l,g;for(l=t+i,ga(t,l,e.length),g="",u=t;u<l;)a=h.Math.min(u+1e4,l),g+=Vwn(e.slice(u,a)),u=a;return g}function xle(e){var t;e.c!=0&&(t=f(ct(e.a,e.b),292),t.b==1?(++e.b,e.b<e.a.c.length&&Ot(f(ct(e.a,e.b),292))):--t.b,--e.c)}function wZ(){wZ=Z,vF=new use("P1_WIDTH_APPROXIMATION",0),sne=new use("P2_PACKING",1),gbe=new use("P3_WHITESPACE_ELIMINATION",2)}function Pp(){Pp=Z,Pm=new $Y(Lp,0),ADe=new $Y("NODES_AND_EDGES",1),Rme=new $Y("PREFER_EDGES",2),Pme=new $Y("PREFER_NODES",3)}function Ud(){Ud=Z,mV=new VY("PORTS",0),GE=new VY("PORT_LABELS",1),gV=new VY("NODE_LABELS",2),mx=new VY("MINIMUM_SIZE",3)}function R0(){R0=Z,jge=new Mz(cNe,0),gPe=new Mz(fxt,1),$ge=new Mz(the,2),uD=new Mz(aNe,3),cD=new Mz("GREEDY_MODEL_ORDER",4)}function h3e(e){switch(e.g){case 12:case 13:case 14:case 15:case 16:case 17:case 18:case 19:case 20:return!0;default:return!1}}function p3e(e,t,i,a,u,l){this.a=e,this.c=t,this.b=i,this.f=a,this.d=u,this.e=l,this.c>0&&this.b>0&&(this.g=AX(this.c,this.b,this.a))}function o3n(e,t){var i=e.a,a;t=String(t),i.hasOwnProperty(t)&&(a=i[t]);var u=(Xue(),Gpe)[typeof a],l=u?u(a):ICe(typeof a);return l}function oL(e){var t,i,a;if(a=null,t=Am in e.a,i=!t,i)throw ie(new am("Every element must have an id."));return a=SR(h1(e,Am)),a}function FS(e){var t,i;for(i=Ryt(e),t=null;e.c==2;)Yr(e),t||(t=(Xr(),Xr(),new D7(2)),vE(t,i),i=t),i.Gm(Ryt(e));return i}function yZ(e,t){var i,a,u;return e.Yj(),a=t==null?0:Ti(t),u=(a&Qr)%e.d.length,i=VAe(e,u,a,t),i?(sbt(e,i),i.kd()):null}function cvt(e,t){var i,a,u;if(e.c)RS(e.c,t);else for(i=t-Al(e),u=new ne(e.d);u.a<u.c.c.length;)a=f(ee(u),162),cvt(a,Al(a)+i)}function uvt(e,t){var i,a,u;if(e.c)NS(e.c,t);else for(i=t-jd(e),u=new ne(e.a);u.a<u.c.c.length;)a=f(ee(u),162),uvt(a,jd(a)+i)}function a3n(e,t,i,a){var u,l;t.c.length!=0&&(u=$6n(i,a),l=WRn(t),Zi(RJ(new Ft(null,new Bt(l,1)),new U2),new Hdt(e,i,u,a)))}function s3n(e,t){if(t.c==e)return t.d;if(t.d==e)return t.c;throw ie(new dn("Input edge is not connected to the input port."))}function lvt(e,t){return e.e>t.e?1:e.e<t.e?-1:e.d>t.d?e.e:e.d<t.d?-t.e:e.e*ECe(e.a,t.a,e.d)}function dvt(e){return e>=48&&e<48+h.Math.min(10,10)?e-48:e>=97&&e<97?e-97+10:e>=65&&e<65?e-65+10:-1}function vm(e,t){if(e.a<0)throw ie(new As("Did not call before(...) or after(...) before calling add(...)."));return u_e(e,e.a,t),e}function g3e(e){return oJ(),me(e,161)?f(bn(SV,YAt),293).Pg(e):Yu(SV,Fd(e))?f(bn(SV,Fd(e)),293).Pg(e):null}function Xu(e){var t,i;return(e.Db&32)==0&&(i=(t=f(fn(e,16),29),jn(t||e.ei())-jn(e.ei())),i!=0&&bR(e,32,Pe(Ro,Jt,1,i,5,1))),e}function bR(e,t,i){var a;(e.Db&t)!=0?i==null?x4n(e,t):(a=Wle(e,t),a==-1?e.Eb=i:io(uR(e.Eb),a,i)):i!=null&&U5n(e,t,i)}function TT(e){var t,i;return Eu(e,-129)>0&&Eu(e,128)<0?(klt(),t=mr(e)+128,i=T6e[t],!i&&(i=T6e[t]=new Mv(e)),i):new Mv(e)}function fvt(e,t){var i;return Ae(t)===Ae(e)?!0:!me(t,22)||(i=f(t,22),i.gc()!=e.gc())?!1:e.Hc(i)}function EZ(e){var t,i,a;if(a=e.Fh(),!a)for(t=0,i=e.Lh();i;i=i.Lh()){if(++t>Lfe)return i.Mh();if(a=i.Fh(),a||i==e)break}return a}function c3n(e,t){var i,a,u,l;return a=e.a.length-1,i=t-e.b&a,l=e.c-t&a,u=e.c-e.b&a,yct(i<u),i>=l?(gxn(e,t),-1):(pxn(e,t),1)}function u3n(e,t,i){var a,u,l;return a=f(be(gd(e.a),t),87),l=(u=a.c,u||(Ut(),Ug)),(l.Rh()?iw(e.b,f(l,52)):l)==i?uQ(a):$P(a,i),l}function m3e(e){switch(typeof e){case sfe:return C0(e);case AOe:return FN(e);case xR:return Uct(e);default:return e==null?0:bS(e)}}function l3n(e,t){return e.e<t.e?-1:e.e>t.e?1:e.f<t.f?-1:e.f>t.f?1:Ti(e)-Ti(t)}function SZ(e,t){return an(e),t==null?!1:Mt(e,t)?!0:e.length==t.length&&Mt(e.toLowerCase(),t.toLowerCase())}function d3n(e){if(SZ(qL,e))return mn(),QL;if(SZ(epe,e))return mn(),yw;throw ie(new dn("Expecting true or false"))}function hE(){hE=Z,XT=new LY(Lp,0),p5e=new LY("INSIDE_PORT_SIDE_GROUPS",1),mge=new LY("GROUP_MODEL_ORDER",2),bge=new LY(aNe,3)}function f3n(e){var t;return e.b||Gpn(e,(t=Ubn(e.e,e.a),!t||!Mt(epe,Rp((!t.b&&(t.b=new Md((Ut(),Za),xc,t)),t.b),"qualified")))),e.c}function h3n(e,t){var i,a;for(i=(gn(t,e.length),e.charCodeAt(t)),a=t+1;a<e.length&&(gn(a,e.length),e.charCodeAt(a)==i);)++a;return a-t}function p3n(e,t){(!t&&console.groupCollapsed!=null?console.groupCollapsed:console.group??console.log).call(console,e)}function g3n(e,t,i,a){a==e,f(i.b,68),f(i.b,68),f(a.b,68),f(a.b,68).c.b,ixe(a,t,e)}function b3e(e,t,i){t.b=h.Math.max(t.b,-i.a),t.c=h.Math.max(t.c,i.a-e.a),t.d=h.Math.max(t.d,-i.b),t.a=h.Math.max(t.a,i.b-e.b)}function v3e(e,t,i){this.c=e,this.f=new et,this.e=new ha,this.j=new fke,this.n=new fke,this.b=t,this.g=new qh(t.c,t.d,t.b,t.a),this.a=i}function Cle(e){var t,i,a,u;for(this.a=new ub,this.d=new Mo,this.e=0,i=e,a=0,u=i.length;a<u;++a)t=i[a],!this.f&&(this.f=t),Jce(this,t)}function hvt(e){ym(),e.length==0?(this.e=0,this.d=1,this.a=ce(se(pr,1),Pr,30,15,[0])):(this.e=1,this.d=e.length,this.a=e,nj(this))}function m3n(e){var t,i,a,u,l,g;for(g=f(V(e,(Ge(),oi)),11),$e(g,pD,e.i.n.b),t=pm(e.e),a=t,u=0,l=a.length;u<l;++u)i=a[u],ua(i,g)}function b3n(e){var t,i,a,u,l,g;for(i=f(V(e,(Ge(),oi)),11),$e(i,pD,e.i.n.b),t=pm(e.g),u=t,l=0,g=u.length;l<g;++l)a=u[l],Ia(a,i)}function pvt(e,t,i){var a,u,l;a=t.c.p,l=t.p,e.b[a][l]=new rft(e,t),i&&(e.a[a][l]=new ett(t),u=f(V(t,(Ge(),l_)),9),u&&jt(e.d,u,t))}function gvt(e,t,i){var a,u,l;for(l=new ne(i.a);l.a<l.c.c.length;)u=f(ee(l),224),a=new iX(f(bn(e.a,u.b),68)),Qe(t.a,a),gvt(e,a,u)}function v3n(e,t){Vce();var i,a;for(a=new Tn(Cn(T1(e).a.Jc(),new B));Kn(a);)if(i=f(Dn(a),18),i.d.i==t||i.c.i==t)return i;return null}function mvt(e,t){var i,a;return i=t.mi(e.a),i&&(a=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),Gr)),a!=null)?a:t.ve()}function w3n(e,t){var i,a;return i=t.mi(e.a),i&&(a=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),Gr)),a!=null)?a:t.ve()}function y3n(e,t){var i,a;return i=tu(e.a.c.p,t.a.c.p),i!=0?i:(a=tu(e.a.d.i.p,t.a.d.i.p),a!=0?a:tu(t.a.d.p,e.a.d.p))}function E3n(e,t){var i,a,u,l;for(a=0,u=t.gc();a<u;++a)i=t.Ql(a),me(i,101)&&(f(i,19).Bb&Dc)!=0&&(l=t.Rl(a),l!=null&&$de(e,f(l,57)))}function S3n(){var e;return Cee!=0&&(e=Date.now(),e-LAt>2e3&&(LAt=e,Aee=h.setTimeout(egn,10))),Cee++==0?(F_n((O2e(),g6e)),!0):!1}function _3n(e,t,i){var a;(nIt?(ICn(e),!0):rIt||oIt?(cP(),!0):iIt&&(cP(),!1))&&(a=new Out(t),a.b=i,xNn(e,a))}function Oj(e,t,i){prt.call(this),this.a=Pe(hIt,VOe,215,(Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])).length,0,1),this.b=e,this.d=t,this.c=i}function bvt(e){this.d=new et,this.e=new Zv,this.c=Pe(pr,Pr,30,(tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])).length,15,1),this.b=e}function vvt(e){var t;this.d=new et,this.j=new ha,this.g=new ha,t=e.g.b,this.f=f(V(ra(t),(ot(),mf)),86),this.e=ke(Oe(TZ(t,cx)))}function Ale(e,t){var i;i=!e.A.Gc((Ud(),GE))||e.q==(oa(),uu),e.u.Gc((Xl(),M1))?i?R9n(e,t):dkt(e,t):e.u.Gc(Lw)&&(i?Q8n(e,t):xkt(e,t))}function wvt(e){var t;Ae(Fe(e,(Rr(),UA)))===Ae((_1(),kne))&&(so(e)?(t=f(Fe(so(e),UA),342),pi(e,UA,t)):pi(e,UA,PF))}function k3n(e,t,i){var a,u;ude(e.e,t,i,(tt(),pn)),ude(e.i,t,i,_n),e.a&&(u=f(V(t,(Ge(),oi)),11),a=f(V(i,oi),11),Xce(e.g,u,a))}function T3n(e){var t,i;return li(e.d.i,(ot(),bD))?(t=f(V(e.c.i,bD),17),i=f(V(e.d.i,bD),17),tu(t.a,i.a)>0):!1}function yvt(e,t,i){return new qh(h.Math.min(e.a,t.a)-i/2,h.Math.min(e.b,t.b)-i/2,h.Math.abs(e.a-t.a)+i,h.Math.abs(e.b-t.b)+i)}function w3e(e,t,i){var a;switch(a=i[e.g][t],e.g){case 1:case 3:return new Ke(0,a);case 2:case 4:return new Ke(a,0);default:return null}}function x3n(e,t){var i;if(i=F3(e.o,t),i==null)throw ie(new am("Node did not exist in input."));return vIe(e,t),Lde(e,t),cIe(e,t,i),null}function aL(e,t){var i,a;if(++e.j,t!=null&&(i=(a=e.a.Cb,me(a,98)?f(a,98).ph():null),g4n(t,i))){bR(e.a,4,i);return}bR(e.a,4,f(t,128))}function C3n(e,t,i){var a,u,l,g;return l=t.j,g=i.j,l!=g?l.g-g.g:(a=e.f[t.p],u=e.f[i.p],a==0&&u==0?0:a==0?-1:u==0?1:di(a,u))}function Evt(e,t,i){var a,u,l;if(!i[t.d])for(i[t.d]=!0,u=new ne(V3(t));u.a<u.c.c.length;)a=f(ee(u),216),l=hZ(a,t),Evt(e,l,i)}function Svt(e,t){var i,a,u,l;for(i=0,u=new ne(t.a);u.a<u.c.c.length;)a=f(ee(u),9),l=a.o.a+a.d.c+a.d.b+e.j,i=h.Math.max(i,l);return i}function _vt(e,t){var i,a;for(a=e.a.length,t.length<a&&(t=pY(new Array(a),t)),i=0;i<a;++i)io(t,i,e.a[i]);return t.length>a&&io(t,a,null),t}function kg(e,t){var i,a;for(a=e.c.length,t.length<a&&(t=pY(new Array(a),t)),i=0;i<a;++i)io(t,i,e.c[i]);return t.length>a&&io(t,a,null),t}function Ile(e,t,i,a){var u;if(u=e.length,t>=u)return u;for(t=t>0?t:0;t<u&&!Yue((gn(t,e.length),e.charCodeAt(t)),i,a);t++);return t}function A3n(e,t,i,a){var u,l,g;return u=new v1(e.e,1,13,(g=t.c,g||(Ut(),Ug)),(l=i.c,l||(Ut(),Ug)),P0(e,t),!1),a?a.kj(u):a=u,a}function kvt(e,t,i){var a,u,l,g,m,v;return a=null,m=pOe(fj(),t),l=null,m&&(u=null,v=fOe(m,i),g=null,v!=null&&(g=e.of(m,v)),u=g,l=u),a=l,a}function pE(e,t,i){var a,u,l;return u=f(bn(e.e,t),391),u?(l=O_e(u,i),Gst(e,u),l):(a=new nke(e,t,i),qr(e.e,t,a),yft(a),null)}function I3n(e,t){var i;if(i=bn(e.q,t),i==null)throw ie(new am("Port did not exist in input."));return vIe(e,t),Lde(e,t),cIe(e,t,i),null}function Tvt(e,t,i){var a,u;u=f(tJ(t.f),222);try{u.kf(e,i),xdt(t.f,u)}catch(l){throw l=go(l),me(l,103)?(a=l,ie(a)):ie(l)}}function O3n(e){var t;if(e==null)return null;if(t=h6n(_u(e,!0)),t==null)throw ie(new Tae("Invalid hexBinary value: '"+e+"'"));return t}function _Z(e,t,i){var a;t.a.length>0&&(Qe(e.b,new Vut(t.a,i)),a=t.a.length,0<a?t.a=uh(t.a,0,0):0>a&&(t.a+=Vst(Pe(Xf,_m,30,-a,15,1))))}function Ole(){return wQ(),ce(se(BIt,1),Ve,166,0,[$It,jIt,FIt,IIt,AIt,OIt,PIt,RIt,NIt,MIt,DIt,LIt,xIt,TIt,CIt,_It,SIt,kIt,yIt,wIt,EIt,rge])}function sL(e){switch(e.g){case 4:return new Ke(0,-1);case 1:return new Ke(1,0);case 2:return new Ke(-1,0);default:return new Ke(0,1)}}function Nle(e){switch(e.g){case 1:return ko(),Wf;case 4:return ko(),Ds;case 2:return ko(),Gs;case 3:return ko(),jg}return ko(),$g}function N3n(e){var t;switch(t=e.ej(null),t){case 10:return 0;case 15:return 1;case 14:return 2;case 11:return 3;case 21:return 4}return-1}function wm(){wm=Z,en=new Hz("PARENTS",0),Ao=new Hz("NODES",1),Dm=new Hz("EDGES",2),FE=new Hz("PORTS",3),L1=new Hz("LABELS",4)}function K3(){K3=Z,Rw=new qz("DISTRIBUTED",0),fV=new qz("JUSTIFIED",1),i7e=new qz("BEGIN",2),LF=new qz(PL,3),o7e=new qz("END",4)}function R3n(e,t,i){var a;switch(a=i.q.getFullYear()-hw+hw,a<0&&(a=-a),t){case 1:e.a+=a;break;case 2:bb(e,a%100,2);break;default:bb(e,a,t)}}function tr(e,t){var i,a;if(fT(t,e.b),t>=e.b>>1)for(a=e.c,i=e.b;i>t;--i)a=a.b;else for(a=e.a.a,i=0;i<t;++i)a=a.a;return new Rut(e,t,a)}function xvt(e){this.b=new et,this.e=new et,this.d=e,this.a=!oP(Jr(new Ft(null,new sT(new yg(e.b))),new Y(new R6))).zd((Wy(),zR))}function Cvt(e,t){var i,a,u;u=t.d.i,a=u.k,!(a==(cn(),Yi)||a==Dg)&&(i=new Tn(Cn(ao(u).a.Jc(),new B)),Kn(i)&&qr(e.k,t,f(Dn(i),18)))}function P3n(e,t){return y1(),di((e.a.b==0?new Ke(e.c.e.a,e.c.e.b):f(hP(e.a),8)).b,(t.a.b==0?new Ke(t.c.e.a,t.c.e.b):f(hP(t.a),8)).b)}function L3n(e,t){return y1(),di((e.a.b==0?new Ke(e.c.e.a,e.c.e.b):f(hP(e.a),8)).a,(t.a.b==0?new Ke(t.c.e.a,t.c.e.b):f(hP(t.a),8)).a)}function D3n(e,t){return y1(),di((e.a.b==0?new Ke(e.b.e.a,e.b.e.b):f(Hh(e.a),8)).a,(t.a.b==0?new Ke(t.b.e.a,t.b.e.b):f(Hh(t.a),8)).a)}function M3n(e,t){return y1(),di((e.a.b==0?new Ke(e.b.e.a,e.b.e.b):f(Hh(e.a),8)).b,(t.a.b==0?new Ke(t.b.e.a,t.b.e.b):f(Hh(t.a),8)).b)}function Rle(e,t){var i,a,u;return a=Vt(e.zh(),t),i=t-e.fi(),i<0?(u=e.Eh(a),u>=0?e.Sh(u):wde(e,a)):i<0?wde(e,a):f(a,69).tk().yk(e,e.di(),i)}function Avt(e){var t,i,a;for(a=(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),e.o),i=a.c.Jc();i.e!=i.i.gc();)t=f(i.Vj(),43),t.kd();return AH(a)}function st(e){var t;if(me(e.a,4)){if(t=g3e(e.a),t==null)throw ie(new As(eCt+e.b+"'. "+Qxt+(d1(_V),_V.k)+S4e));return t}else return e.a}function j3n(e){var t;if(e==null)return null;if(t=U9n(_u(e,!0)),t==null)throw ie(new Tae("Invalid base64Binary value: '"+e+"'"));return t}function Mn(e){var t;try{return t=e.i.Xb(e.e),e.Uj(),e.g=e.e++,t}catch(i){throw i=go(i),me(i,97)?(e.Uj(),ie(new ic)):ie(i)}}function Ple(e){var t;try{return t=e.c.Si(e.e),e.Uj(),e.g=e.e++,t}catch(i){throw i=go(i),me(i,97)?(e.Uj(),ie(new ic)):ie(i)}}function kZ(e){var t,i,a,u;for(u=0,i=0,a=e.length;i<a;i++)t=(gn(i,e.length),e.charCodeAt(i)),t>=64&&t<128&&(u=hm(u,hb(1,t-64)));return u}function TZ(e,t){var i,a;return a=null,li(e,(Rr(),a4))&&(i=f(V(e,a4),102),i.nf(t)&&(a=i.mf(t))),a==null&&ra(e)&&(a=V(ra(e),t)),a}function $3n(e,t){var i,a,u;for(u=new Du(t.gc()),a=t.Jc();a.Ob();)i=f(a.Pb(),291),i.c==i.f?hL(e,i,i.c):tRn(e,i)||vn(u.c,i);return u}function Ivt(e,t){var i,a,u;for(i=e.o,u=f(f(si(e.r,t),22),83).Jc();u.Ob();)a=f(u.Pb(),115),a.e.a=UAn(a,i.a),a.e.b=i.b*ke(Oe(a.b.mf(Mee)))}function F3n(e,t){var i,a,u,l;return u=e.k,i=ke(Oe(V(e,(Ge(),d_)))),l=t.k,a=ke(Oe(V(t,d_))),l!=(cn(),bo)?-1:u!=bo?1:i==a?0:i<a?-1:1}function B3n(e,t){var i,a;return i=f(f(bn(e.g,t.a),47).a,68),a=f(f(bn(e.g,t.b),47).a,68),Qv(t.a,t.b)-Qv(t.a,VSe(i.b))-Qv(t.b,VSe(a.b))}function xZ(e){switch(J2e(),this.c=new et,this.d=e,e.g){case 0:case 2:this.a=Jke(h5e),this.b=Gi;break;case 3:case 1:this.a=h5e,this.b=qo}}function U3n(e){var t;VN(f(V(e,(ot(),Xi)),100))&&(t=e.b,EEt((Gt(0,t.c.length),f(t.c[0],25))),EEt(f(ct(t,t.c.length-1),25)))}function z3n(e,t){t.Sg("Self-Loop post-processing",1),Zi(Jr(Jr(Zs(new Ft(null,new Bt(e.b,16)),new h8),new gO),new p8),new wy),t.Tg()}function Ovt(e,t,i){var a,u;if(e.c)md(e.c,e.c.i+t),bd(e.c,e.c.j+i);else for(u=new ne(e.b);u.a<u.c.c.length;)a=f(ee(u),162),Ovt(a,t,i)}function H3n(e){var t,i,a;for(a=e.c.a,e.p=(or(a),new Kl(a)),i=new ne(a);i.a<i.c.c.length;)t=f(ee(i),9),t.p=_Rn(t).a;Ht(),na(e.p,new MU)}function G3n(e,t){var i,a;if(e.j.length!=t.j.length)return!1;for(i=0,a=e.j.length;i<a;i++)if(!Mt(e.j[i],t.j[i]))return!1;return!0}function Nvt(e,t){Ht();var i,a,u,l;for(i=e,l=t,me(e,22)&&!me(t,22)&&(i=t,l=e),u=i.Jc();u.Ob();)if(a=u.Pb(),l.Gc(a))return!1;return!0}function q3n(e,t,i,a){if(t.a<a.a)return!0;if(t.a==a.a){if(t.b<a.b)return!0;if(t.b==a.b&&e.b>i.b)return!0}return!1}function V3n(e,t){var i;return i=f(V(e,(ot(),Ls)),77),Ase(t,BOt)?i?nf(i):(i=new fd,$e(e,Ls,i)):i&&$e(e,Ls,null),i}function Rvt(e){var t;return t=new $v,t.a+="n",e.k!=(cn(),Yi)&&xr(xr((t.a+="(",t),Lse(e.k).toLowerCase()),")"),xr((t.a+="_",t),lG(e)),t.a}function Nj(){Nj=Z,EDe=new $z(NRe,0),Ome=new $z(the,1),Nme=new $z("LINEAR_SEGMENTS",2),oF=new $z("BRANDES_KOEPF",3),aF=new $z(xxt,4)}function vR(e,t,i,a){var u;return i>=0?e.Oh(t,i,a):(e.Lh()&&(a=(u=e.Bh(),u>=0?e.wh(a):e.Lh().Ph(e,-1-u,null,a))),e.yh(t,i,a))}function y3e(e,t){switch(t){case 7:!e.e&&(e.e=new Qt(vo,e,7,4)),Jn(e.e);return;case 8:!e.d&&(e.d=new Qt(vo,e,8,5)),Jn(e.d);return}a3e(e,t)}function pi(e,t,i){return i==null?(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),yZ(e.o,t)):(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),hG(e.o,t,i)),e}function Xh(e,t){var i;i=e.dd(t);try{return i.Pb()}catch(a){throw a=go(a),me(a,111)?ie(new Ru("Can't get element "+t)):ie(a)}}function Pvt(e,t){var i;switch(i=f(Es(e.b,t),126).n,t.g){case 1:e.t>=0&&(i.d=e.t);break;case 3:e.t>=0&&(i.a=e.t)}e.C&&(i.b=e.C.b,i.c=e.C.c)}function W3n(e){var t;t=e.a;do t=f(Dn(new Tn(Cn(Fo(t).a.Jc(),new B))),18).c.i,t.k==(cn(),uo)&&e.b.Ec(t);while(t.k==(cn(),uo));e.b=Bd(e.b)}function Lvt(e,t){var i,a,u;for(u=e,a=new Tn(Cn(Fo(t).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),i.c.i.c&&(u=h.Math.max(u,i.c.i.c.p));return u}function K3n(e,t){var i,a,u;for(u=0,a=f(f(si(e.r,t),22),83).Jc();a.Ob();)i=f(a.Pb(),115),u+=i.d.d+i.b.Kf().b+i.d.a,a.Ob()&&(u+=e.w);return u}function Y3n(e,t){var i,a,u;for(u=0,a=f(f(si(e.r,t),22),83).Jc();a.Ob();)i=f(a.Pb(),115),u+=i.d.b+i.b.Kf().a+i.d.c,a.Ob()&&(u+=e.w);return u}function Dvt(e){var t,i,a,u;if(a=0,u=RT(e),u.c.length==0)return 1;for(i=new ne(u);i.a<i.c.c.length;)t=f(ee(i),27),a+=Dvt(t);return a}function Mvt(e){var t,i;for(this.b=new et,this.c=e,this.a=!1,i=new ne(e.a);i.a<i.c.c.length;)t=f(ee(i),9),this.a=this.a|t.k==(cn(),Yi)}function X3n(e,t,i){var a,u,l,g;return a=e.bd(t),a!=-1&&(e.Mj()?(l=e.Nj(),g=xX(e,a),u=e.Fj(4,g,null,a,l),i?i.kj(u):i=u):xX(e,a)),i}function za(e,t,i){var a,u,l,g;return a=e.bd(t),a!=-1&&(e.Mj()?(l=e.Nj(),g=wR(e,a),u=e.Fj(4,g,null,a,l),i?i.kj(u):i=u):wR(e,a)),i}function J3n(e,t,i,a){var u,l,g;i.Th(t)&&(is(),_ue(t)?(u=f(i.Ih(t),159),E3n(e,u)):(l=(g=t,g?f(a,52).ci(g):null),l&&Fhn(i.Ih(t),l)))}function CZ(e,t,i,a){var u,l,g;return l=Vt(e.zh(),t),u=t-e.fi(),u<0?(g=e.Eh(l),g>=0?e.Hh(g,i,!0):GS(e,l,i)):f(l,69).tk().vk(e,e.di(),u,i,a)}function Z3n(e,t,i,a){var u,l;l=t.nf((Rr(),HA))?f(t.mf(HA),22):e.j,u=Gxn(l),u!=(wQ(),rge)&&(i&&!h3e(u)||kAe(p6n(e,u,a),t))}function Lle(e,t){return ta(e)?!!CAt[t]:e.Pm?!!e.Pm[t]:Jk(e)?!!xAt[t]:Xk(e)?!!TAt[t]:!1}function Q3n(e){switch(e.g){case 1:return DS(),yq;case 3:return DS(),wq;case 2:return DS(),oge;case 4:return DS(),ige;default:return null}}function eAn(e,t,i){if(e.e)switch(e.b){case 1:fyn(e.c,t,i);break;case 0:hyn(e.c,t,i)}else Qht(e.c,t,i);e.a[t.p][i.p]=e.c.i,e.a[i.p][t.p]=e.c.e}function jvt(e){var t,i;if(e==null)return null;for(i=Pe(Ib,Ye,199,e.length,0,2),t=0;t<i.length;t++)i[t]=f(bSn(e[t],e[t].length),199);return i}function $vt(e){var t;t=f(V(e,(Ec(),bMe)),107),$e(e,(_i(),TD),new Ke(0,0)),LSt(new xH,e,t.b-ke(Oe(V(e,Kme))),t.d-ke(Oe(V(e,Yme))))}function AZ(e){var t;if(rle(e))return MX(e),e.rl()&&(t=Bj(e.e,e.b,e.c,e.a,e.j),e.j=t),e.g=e.a,++e.a,++e.c,e.i=0,e.j;throw ie(new ic)}function Fvt(e,t,i){i.Sg("Compound graph preprocessor",1),e.a=new CS,skt(e,t,null),pMn(e,t),Y6n(e),$e(t,(Ge(),$Pe),e.a),e.a=null,nu(e.b),i.Tg()}function Bvt(e,t,i){var a,u,l;for(u=new Tn(Cn((t?Fo(e):ao(e)).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),l=t?a.c.i:a.d.i,l.k==(cn(),Vc)&&jo(l,i)}function xT(){xT=Z,Gte=new jY(Lp,0),Ame=new jY("PORT_POSITION",1),PA=new jY("NODE_SIZE_WHERE_SPACE_PERMITS",2),RA=new jY("NODE_SIZE",3)}function tAn(e,t){var i,a,u;for(t.Sg("Untreeify",1),i=f(V(e,(_i(),cMe)),15),u=i.Jc();u.Ob();)a=f(u.Pb(),65),Mr(a.b.d,a),Mr(a.c.b,a);t.Tg()}function vb(){vb=Z,Bbe=new T7("AUTOMATIC",0),Qq=new T7(NR,1),eV=new T7(RR,2),wne=new T7("TOP",3),bne=new T7(KOe,4),vne=new T7(PL,5)}function Y3(e,t,i){var a,u;if(u=e.gc(),t>=u)throw ie(new rT(t,u));if(e.Pi()&&(a=e.bd(i),a>=0&&a!=t))throw ie(new dn(lq));return e.Ui(t,i)}function P0(e,t){var i,a,u;if(u=vwt(e,t),u>=0)return u;if(e.ll()){for(a=0;a<e.i;++a)if(i=e.ml(f(e.g[a],57)),Ae(i)===Ae(t))return a}return-1}function E3e(e,t){if(this.a=f(or(e),252),this.b=f(or(t),252),e.Cd(t)>0||e==(vae(),Mpe)||t==(wae(),jpe))throw ie(new dn("Invalid range: "+Jht(e,t)))}function S3e(e,t,i,a){vL();var u,l;for(u=0,l=0;l<i;l++)u=Ua(Na(ia(t[l],ls),ia(a,ls)),ia(mr(u),ls)),e[l]=mr(u),u=Jy(u,32);return mr(u)}function _3e(e,t,i){var a,u;for(u=0,a=0;a<nge;a++)u=h.Math.max(u,sle(e.a[t.g][a],i));return t==(Np(),Bu)&&e.b&&(u=h.Math.max(u,e.b.b)),u}function IZ(e,t){var i,a;if(f_e(t>0),(t&-t)==t)return Aa(t*wd(e,31)*4656612873077393e-25);do i=wd(e,31),a=i%t;while(i-a+(t-1)<0);return Aa(a)}function nAn(e,t){var i,a,u;for(i=mS(new qy,e),u=new ne(t);u.a<u.c.c.length;)a=f(ee(u),123),Zh(Uh(Bh(zh(Fh(new ah,0),0),i),a));return i}function cG(e,t,i){var a,u;for(NTn(e,t-e.s,i-e.t),u=new ne(e.n);u.a<u.c.c.length;)a=f(ee(u),208),J5(a,a.e+t-e.s),gz(a,a.f+i-e.t);e.s=t,e.t=i}function rAn(e,t,i){switch(i.g){case 1:e.a=t.a/2,e.b=0;break;case 2:e.a=t.a,e.b=t.b/2;break;case 3:e.a=t.a/2,e.b=t.b;break;case 4:e.a=0,e.b=t.b/2}}function uG(e,t,i,a){var u,l;for(u=t;u<e.c.length;u++)if(l=(Gt(u,e.c.length),f(e.c[u],11)),i.Mb(l))vn(a.c,l);else return u;return e.c.length}function iAn(e){var t,i,a;for(a=f(si(e.a,(CT(),lte)),15).Jc();a.Ob();)i=f(a.Pb(),105),t=P3e(i),SP(e,i,t[0],(LS(),s_),0),SP(e,i,t[1],c_,1)}function oAn(e){var t,i,a;for(a=f(si(e.a,(CT(),dte)),15).Jc();a.Ob();)i=f(a.Pb(),105),t=P3e(i),SP(e,i,t[0],(LS(),s_),0),SP(e,i,t[1],c_,1)}function Dle(e){switch(e.g){case 0:return null;case 1:return new Obt;case 2:return new k2e;default:throw ie(new dn(Bhe+(e.f!=null?e.f:""+e.g)))}}function Uvt(e){var t;return t=ke(Oe(Fe(e,(Rr(),ND))))*h.Math.sqrt((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a).i),new Ke(t,t/ke(Oe(Fe(e,sV))))}function Mle(e){var t;return e.f&&e.f.Rh()&&(t=f(e.f,52),e.f=f(iw(e,t),84),e.f!=t&&(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,9,8,t,e.f))),e.f}function jle(e){var t;return e.i&&e.i.Rh()&&(t=f(e.i,52),e.i=f(iw(e,t),84),e.i!=t&&(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,9,7,t,e.i))),e.i}function os(e){var t;return e.b&&(e.b.Db&64)!=0&&(t=e.b,e.b=f(iw(e,t),19),e.b!=t&&(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,9,21,t,e.b))),e.b}function OZ(e,t){var i,a,u;e.d==null?(++e.e,++e.f):(a=t.xi(),o5n(e,e.f+1),u=(a&Qr)%e.d.length,i=e.d[u],!i&&(i=e.d[u]=e.ak()),i.Ec(t),++e.f)}function k3e(e,t,i){var a;return t.qk()?!1:t.Fk()!=-2?(a=t.fk(),a==null?i==null:ei(a,i)):t.nk()==e.e.zh()&&i==null}function NZ(){var e;uf(16,nTt),e=Sbt(16),this.b=Pe(Fpe,BG,304,e,0,1),this.c=Pe(Fpe,BG,304,e,0,1),this.a=null,this.e=null,this.i=0,this.f=e-1,this.g=0}function Tg(e){R_e.call(this),this.k=(cn(),Yi),this.j=(uf(6,$T),new Du(6)),this.b=(uf(2,$T),new Du(2)),this.d=new hae,this.f=new b2e,this.a=e}function aAn(e){var t,i,a,u,l;for(i=0,u=new ne(e.a);u.a<u.c.c.length;)a=f(ee(u),123),a.d=i++;return t=cNn(e),l=null,t.c.length>1&&(l=nAn(e,t)),l}function sAn(e){var t,i,a;for(t=0,a=new ne(e.c.a);a.a<a.c.c.length;)i=f(ee(a),9),t+=g1(new Tn(Cn(ao(i).a.Jc(),new B)));return t/e.c.a.c.length}function zvt(e,t){var i,a,u;for(e.b[t.g]=1,a=tr(t.d,0);a.b!=a.d.c;)i=f(Zn(a),65),u=i.c,e.b[u.g]==1?Mr(e.a,i):e.b[u.g]==2?e.b[u.g]=1:zvt(e,u)}function cAn(e){var t,i;e.c.length<=1||(t=J2t(e,(tt(),Bn)),Dyt(e,f(t.a,17).a,f(t.b,17).a),i=J2t(e,pn),Dyt(e,f(i.a,17).a,f(i.b,17).a))}function uAn(e,t,i){var a,u;for(u=e.a.b,a=u.c.length;a<i;a++)Ky(u,u.c.length,new ru(e.a));jo(t,(Gt(i-1,u.c.length),f(u.c[i-1],25))),e.b[t.p]=i}function Hvt(e,t,i,a){var u,l,g;for(u=f(si(a?e.a:e.b,t),22),g=u.Jc();g.Ob();)if(l=f(g.Pb(),27),oQ(e,i,l))return!0;return!1}function $le(e){var t,i;for(i=new Nn(e);i.e!=i.i.gc();)if(t=f(Mn(i),87),t.e||(!t.d&&(t.d=new So(gs,t,1)),t.d).i!=0)return!0;return!1}function Fle(e){var t,i;for(i=new Nn(e);i.e!=i.i.gc();)if(t=f(Mn(i),87),t.e||(!t.d&&(t.d=new So(gs,t,1)),t.d).i!=0)return!0;return!1}function Ble(){Ble=Z,G8e=(Nde(),B8e),H8e=new gS(8),new pa((Rr(),V0),H8e),new pa(zE,8),_Lt=$8e,U8e=gLt,z8e=mLt,SLt=new pa(nV,(mn(),!1))}function lAn(e,t,i){var a;i.Sg("Shrinking tree compaction",1),lt(ut(V(t,(HP(),F$))))?(Nkn(e,t.f),omt(t.f,(a=t.c,a))):omt(t.f,t.c),i.Tg()}function T3e(e,t,i,a){switch(t){case 7:return!e.e&&(e.e=new Qt(vo,e,7,4)),e.e;case 8:return!e.d&&(e.d=new Qt(vo,e,8,5)),e.d}return YCe(e,t,i,a)}function Ule(e){var t;return e.a&&e.a.Rh()&&(t=f(e.a,52),e.a=f(iw(e,t),141),e.a!=t&&(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,9,5,t,e.a))),e.a}function gE(e){return e<48||e>102?-1:e<=57?e-48:e<65?-1:e<=70?e-65+10:e<97?-1:e-97+10}function zle(e,t){if(e==null)throw ie(new IN("null key in entry: null="+t));if(t==null)throw ie(new IN("null value in entry: "+e+"=null"))}function dAn(e,t){for(var i,a;e.Ob();)if(!t.Ob()||(i=e.Pb(),a=t.Pb(),!(Ae(i)===Ae(a)||i!=null&&ei(i,a))))return!1;return!t.Ob()}function Gvt(e,t){var i;return i=ce(se(sa,1),_s,30,15,[sle(e.a[0],t),sle(e.a[1],t),sle(e.a[2],t)]),e.d&&(i[0]=h.Math.max(i[0],i[2]),i[2]=i[0]),i}function qvt(e,t){var i;return i=ce(se(sa,1),_s,30,15,[aZ(e.a[0],t),aZ(e.a[1],t),aZ(e.a[2],t)]),e.d&&(i[0]=h.Math.max(i[0],i[2]),i[2]=i[0]),i}function x3e(e,t,i){VN(f(V(t,(ot(),Xi)),100))||(BTe(e,t,L0(t,i)),BTe(e,t,L0(t,(tt(),Bn))),BTe(e,t,L0(t,hn)),Ht(),na(t.j,new Jet(e)))}function Vvt(e){var t,i;for(e.c||JMn(e),i=new fd,t=new ne(e.a),ee(t);t.a<t.c.c.length;)Mr(i,f(ee(t),407).a);return Ln(i.b!=0),lh(i,i.c.b),i}function fAn(e,t,i){var a,u,l,g,m;for(m=e.r+t,e.r+=t,e.d+=i,a=i/e.n.c.length,u=0,g=new ne(e.n);g.a<g.c.c.length;)l=f(ee(g),208),i6n(l,m,a,u),++u}function hAn(e){var t,i,a;for(e.b.a.$b(),e.a=Pe(Lee,Jt,60,e.c.c.a.b.c.length,0,1),t=0,a=new ne(e.c.c.a.b);a.a<a.c.c.length;)i=f(ee(a),60),i.f=t++}function pAn(e){var t,i,a;for(e.b.a.$b(),e.a=Pe(fge,Jt,82,e.c.a.a.b.c.length,0,1),t=0,a=new ne(e.c.a.a.b);a.a<a.c.c.length;)i=f(ee(a),82),i.i=t++}function gAn(e){var t,i,a,u,l;for(a=new ne(e.b);a.a<a.c.c.length;)for(i=f(ee(a),25),t=0,l=new ne(i.a);l.a<l.c.c.length;)u=f(ee(l),9),u.p=t++}function mAn(e,t,i){var a,u,l,g;for(g=0,a=i/e.a.c.length,l=new ne(e.a);l.a<l.c.c.length;)u=f(ee(l),172),cG(u,u.s,u.t+g*a),fAn(u,e.d-u.r+t,a),++g}function Wvt(e){var t;if(t=bxn(e),!Kn(e))throw ie(new Ru("position (0) must be less than the number of elements that remained ("+t+")"));return Dn(e)}function bAn(e,t){var i;return e.a||(i=Pe(sa,_s,30,0,15,1),b3(e.b.a,new fe(i)),Vnt(i,Gmt(Re.prototype.Ke,Re,[])),e.a=new but(i,e.d)),yJ(e.a,t)}function Kvt(e){switch(e.g){case 1:return tt(),pn;case 4:return tt(),hn;case 3:return tt(),_n;case 2:return tt(),Bn;default:return tt(),fc}}function vAn(e,t,i){t.k==(cn(),Yi)&&i.k==uo&&(e.d=dle(t,(tt(),Bn)),e.b=dle(t,hn)),i.k==Yi&&t.k==uo&&(e.d=dle(i,(tt(),hn)),e.b=dle(i,Bn))}function Hle(e,t){var i,a,u;for(u=lc(e,t),a=u.Jc();a.Ob();)if(i=f(a.Pb(),11),V(i,(Ge(),ed))!=null||F7(new yg(i.b)))return!0;return!1}function wAn(e,t,i){i.Sg("Linear segments node placement",1),e.b=f(V(t,(Ge(),CA)),313),_7n(e,t),QDn(e,t),vMn(e,t),Q9n(e),e.a=null,e.b=null,i.Tg()}function C3e(e,t){return md(t,e.e+e.d+(e.c.c.length==0?0:e.b)),bd(t,e.f),e.a=h.Math.max(e.a,t.f),e.d+=t.g+(e.c.c.length==0?0:e.b),Qe(e.c,t),!0}function yAn(e,t){var i,a,u,l,g,m;for(u=t.length-1,g=0,m=0,a=0;a<=u;a++)l=t[a],i=NNn(u,a)*OCe(1-e,u-a)*OCe(e,a),g+=l.a*i,m+=l.b*i;return new Ke(g,m)}function Yvt(e,t){var i,a,u,l,g;for(i=t.gc(),e.Yi(e.i+i),l=t.Jc(),g=e.i,e.i+=i,a=g;a<e.i;++a)u=l.Pb(),j7(e,a,e.Wi(a,u)),e.Ji(a,u),e.Ki();return i!=0}function EAn(e,t,i){var a,u,l;return e.Mj()?(a=e.Bj(),l=e.Nj(),++e.j,e.nj(a,e.Wi(a,t)),u=e.Fj(3,null,t,a,l),i?i.kj(u):i=u):qct(e,e.Bj(),t),i}function SAn(e,t,i){var a,u,l;return a=f(be(iu(e.a),t),87),l=(u=a.c,me(u,88)?f(u,29):(Ut(),Ch)),((l.Db&64)!=0?iw(e.b,l):l)==i?uQ(a):$P(a,i),l}function _An(e){var t;return e==null?null:new Gv((t=_u(e,!0),t.length>0&&(gn(0,t.length),t.charCodeAt(0)==43)?(gn(1,t.length+1),t.substr(1)):t))}function kAn(e){var t;return e==null?null:new Gv((t=_u(e,!0),t.length>0&&(gn(0,t.length),t.charCodeAt(0)==43)?(gn(1,t.length+1),t.substr(1)):t))}function A3e(e,t,i,a,u,l,g,m){var v,S;a&&(v=a.a[0],v&&A3e(e,t,i,v,u,l,g,m),Jle(e,i,a.d,u,l,g,m)&&t.Ec(a),S=a.a[1],S&&A3e(e,t,i,S,u,l,g,m))}function Rj(e,t){var i,a,u,l;for(l=e.gc(),t.length<l&&(t=pY(new Array(l),t)),u=t,a=e.Jc(),i=0;i<l;++i)io(u,i,a.Pb());return t.length>l&&io(t,l,null),t}function TAn(e,t){var i,a;if(a=e.gc(),t==null){for(i=0;i<a;i++)if(e.Xb(i)==null)return i}else for(i=0;i<a;i++)if(ei(t,e.Xb(i)))return i;return-1}function Gle(e,t){var i,a,u;return i=t.jd(),u=t.kd(),a=e.xc(i),!(!(Ae(u)===Ae(a)||u!=null&&ei(u,a))||a==null&&!e._b(i))}function xAn(e,t){var i,a,u;return t<=22?(i=e.l&(1<<t)-1,a=u=0):t<=44?(i=e.l,a=e.m&(1<<t-22)-1,u=0):(i=e.l,a=e.m,u=e.h&(1<<t-44)-1),Ku(i,a,u)}function CAn(e,t){switch(t.g){case 1:return e.f.n.d+e.t;case 3:return e.f.n.a+e.t;case 2:return e.f.n.c+e.s;case 4:return e.f.n.b+e.s;default:return 0}}function AAn(e,t){var i,a;switch(a=t.c,i=t.a,e.b.g){case 0:i.d=e.e-a.a-a.d;break;case 1:i.d+=e.e;break;case 2:i.c=e.e-a.a-a.d;break;case 3:i.c=e.e+a.d}}function RZ(){RZ=Z,hbe=new UY(Lp,0),GMe=new UY(hxt,1),qMe=new UY("EDGE_LENGTH_BY_POSITION",2),HMe=new UY("CROSSING_MINIMIZATION_BY_POSITION",3)}function qle(e,t){var i,a;if(i=f(F3(e.n,t),27),i)return i;if(a=f(F3(e.p,t),124),a)return a;throw ie(new am("Referenced shape does not exist: "+t))}function Xvt(e,t){if(e.g==-1)throw ie(new Tl);e.Uj();try{e.d.fd(e.g,t),e.f=e.d.j}catch(i){throw i=go(i),me(i,97)?ie(new If):ie(i)}}function I3e(e,t){var i,a;if(me(t,252)){a=f(t,252);try{return i=e.Cd(a),i==0}catch(u){if(u=go(u),me(u,211))return!1;throw ie(u)}}return!1}function IAn(e,t){if(e.c==t)return e.d;if(e.d==t)return e.c;throw ie(new dn("Node 'one' must be either source or target of edge 'edge'."))}function OAn(e,t){if(e.c.i==t)return e.d.i;if(e.d.i==t)return e.c.i;throw ie(new dn("Node "+t+" is neither source nor target of edge "+e))}function NAn(e,t,i){i.Sg("Self-Loop ordering",1),Zi(Lu(Jr(Jr(Zs(new Ft(null,new Bt(t.b,16)),new pO),new l8),new d8),new x6),new xet(e)),i.Tg()}function RAn(e,t){var i;switch(t.g){case 2:case 4:i=e.a,e.c.d.n.b<i.d.n.b&&(i=e.c),E0(e,t,(_g(),Tge),i);break;case 1:case 3:E0(e,t,(_g(),VR),null)}}function Vle(e,t,i,a,u,l){var g,m,v,S,T;for(g=OIn(t,i,l),m=i==(tt(),hn)||i==pn?-1:1,S=e[i.g],T=0;T<S.length;T++)v=S[T],v>0&&(v+=u),S[T]=g,g+=m*(v+a)}function Jvt(e){var t,i,a;for(a=e.f,e.n=Pe(sa,_s,30,a,15,1),e.d=Pe(sa,_s,30,a,15,1),t=0;t<a;t++)i=f(ct(e.c.b,t),25),e.n[t]=Svt(e,i),e.d[t]=k2t(e,i)}function Zvt(e,t,i,a){var u;this.c=e,this.d=t,u=new ki,Wi(u,i,u.c.b,u.c),this.a=u,this.b=f(V(a,(Ec(),E_)),86),this.e=ke(Oe(V(a,yMe))),Hkt(this)}function Wle(e,t){var i,a,u;for(u=0,a=2;a<t;a<<=1)(e.Db&a)!=0&&++u;if(u==0){for(i=t<<=1;i<=128;i<<=1)if((e.Db&i)!=0)return 0;return-1}else return u}function Qvt(e,t){var i,a,u,l,g;for(g=Ju(e.e.zh(),t),l=null,i=f(e.g,122),u=0;u<e.i;++u)a=i[u],g.Zl(a.Ik())&&(!l&&(l=new J2),Qn(l,a));l&&Ckt(e,l)}function ewt(e){var t,i,a;if(!e)return null;if(e.dc())return"";for(a=new p0,i=e.Jc();i.Ob();)t=i.Pb(),ys(a,dr(t)),a.a+=" ";return yse(a,a.a.length-1)}function twt(e,t){var i=new Array(t),a;switch(e){case 14:case 15:a=0;break;case 16:a=!1;break;default:return i}for(var u=0;u<t;++u)i[u]=a;return i}function BS(e){var t,i,a;for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),60),t.c.$b();l1(e.d)?a=e.a.c:a=e.a.d,Mu(a,new ns(e)),e.c.bf(e),H2t(e)}function nwt(e){var t,i,a,u;for(i=new ne(e.e.c);i.a<i.c.c.length;){for(t=f(ee(i),288),u=new ne(t.b);u.a<u.c.c.length;)a=f(ee(u),442),oOe(a);NEt(t)}}function rwt(e){var t,i,a;for(a=Qr,i=new ne(e.a);i.a<i.c.c.length;)t=f(ee(i),9),li(t,(Ge(),hi))&&(a=h.Math.min(a,f(V(t,hi),17).a));return a}function PZ(e){var t,i,a,u,l;for(a=0,l=0,u=0,i=new ne(e.a);i.a<i.c.c.length;)t=f(ee(i),172),l=h.Math.max(l,t.r),a+=t.d+(u>0?e.c:0),++u;e.b=a,e.d=l}function iwt(e,t){var i;return i=ce(se(sa,1),_s,30,15,[_3e(e,(Np(),_c),t),_3e(e,Bu,t),_3e(e,kc,t)]),e.f&&(i[0]=h.Math.max(i[0],i[2]),i[2]=i[0]),i}function owt(e){var t;li(e,(ot(),p_))&&(t=f(V(e,p_),22),t.Gc((NT(),sp))?(t.Kc(sp),t.Ec(cp)):t.Gc(cp)&&(t.Kc(cp),t.Ec(sp)))}function awt(e){var t;li(e,(ot(),p_))&&(t=f(V(e,p_),22),t.Gc((NT(),lp))?(t.Kc(lp),t.Ec(Sh)):t.Gc(Sh)&&(t.Kc(Sh),t.Ec(lp)))}function Kle(e,t,i,a){var u,l,g,m;return e.a==null&&ONn(e,t),g=t.b.j.c.length,l=i.d.p,m=a.d.p,u=m-1,u<0&&(u=g-1),l<=u?e.a[u]-e.a[l]:e.a[g-1]-e.a[l]+e.a[u]}function PAn(e){var t;for(t=0;t<e.a.c.length;t++)if(li(f(ct(e.a,t),9),(ot(),eF))&&!lt(ut(V(f(ct(e.a,t),9),eF))))return!1;return!0}function LAn(e,t,i){var a,u,l,g,m;for(l=e.c,m=i?t:e,a=i?e:t,u=m.p+1;u<a.p;++u)if(g=f(ct(l.a,u),9),!(g.k==(cn(),Dg)||bIn(g)))return!1;return!0}function DAn(e){var t,i;if(!e.b)for(e.b=cJ(f(e.f,27).ih().i),i=new Nn(f(e.f,27).ih());i.e!=i.i.gc();)t=f(Mn(i),153),Qe(e.b,new _ae(t));return e.b}function MAn(e){var t,i;if(!e.e)for(e.e=cJ(Qke(f(e.f,27)).i),i=new Nn(Qke(f(e.f,27)));i.e!=i.i.gc();)t=f(Mn(i),124),Qe(e.e,new tnt(t));return e.e}function swt(e){var t,i;if(!e.a)for(e.a=cJ(XX(f(e.f,27)).i),i=new Nn(XX(f(e.f,27)));i.e!=i.i.gc();)t=f(Mn(i),27),Qe(e.a,new Rse(e,t));return e.a}function US(e){var t;if(!e.C&&(e.D!=null||e.B!=null))if(t=q8n(e),t)e.el(t);else try{e.el(null)}catch(i){if(i=go(i),!me(i,63))throw ie(i)}return e.C}function jAn(e){switch(e.q.g){case 5:Lwt(e,(tt(),hn)),Lwt(e,Bn);break;case 4:B_t(e,(tt(),hn)),B_t(e,Bn);break;default:Uyt(e,(tt(),hn)),Uyt(e,Bn)}}function $An(e){switch(e.q.g){case 5:Dwt(e,(tt(),_n)),Dwt(e,pn);break;case 4:U_t(e,(tt(),_n)),U_t(e,pn);break;default:zyt(e,(tt(),_n)),zyt(e,pn)}}function X3(e,t){var i,a,u;for(u=new ha,a=e.Jc();a.Ob();)i=f(a.Pb(),36),wL(i,u.a,0),u.a+=i.f.a+t,u.b=h.Math.max(u.b,i.f.b);return u.b>0&&(u.b+=t),u}function LZ(e,t){var i,a,u;for(u=new ha,a=e.Jc();a.Ob();)i=f(a.Pb(),36),wL(i,0,u.b),u.b+=i.f.b+t,u.a=h.Math.max(u.a,i.f.a);return u.a>0&&(u.a+=t),u}function cwt(e,t){var i,a;if(t.length==0)return 0;for(i=Tce(e.a,t[0],(tt(),pn)),i+=Tce(e.a,t[t.length-1],_n),a=0;a<t.length;a++)i+=wNn(e,a,t);return i}function uwt(){qj(),this.c=new et,this.i=new et,this.e=new ub,this.f=new ub,this.g=new ub,this.j=new et,this.a=new et,this.b=new Wn,this.k=new Wn}function Yle(e,t){var i,a;return e.Db>>16==6?e.Cb.Ph(e,5,Bp,t):(a=os(f(Vt((i=f(fn(e,16),29),i||e.ei()),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function FAn(e){AP();var t=e.e;if(t&&t.stack){var i=t.stack,a=t+`
`;return i.substring(0,a.length)==a&&(i=i.substring(a.length)),i.split(`
`)}return[]}function BAn(e){var t;return t=(vbt(),FAt),t[e>>>28]|t[e>>24&15]<<4|t[e>>20&15]<<8|t[e>>16&15]<<12|t[e>>12&15]<<16|t[e>>8&15]<<20|t[e>>4&15]<<24|t[e&15]<<28}function lwt(e){var t,i,a;e.b==e.c&&(a=e.a.length,i=dCe(h.Math.max(8,a))<<1,e.b!=0?(t=Gh(e.a,i),Cbt(e,t,a),e.a=t,e.b=0):Bk(e.a,i),e.c=a)}function UAn(e,t){var i;return i=e.b,i.nf((Rr(),Sd))?i.$f()==(tt(),pn)?-i.Kf().a-ke(Oe(i.mf(Sd))):t+ke(Oe(i.mf(Sd))):i.$f()==(tt(),pn)?-i.Kf().a:t}function lG(e){var t;return e.b.c.length!=0&&f(ct(e.b,0),72).a?f(ct(e.b,0),72).a:(t=Ice(e),t??""+(e.c?Oc(e.c.a,e,0):-1))}function DZ(e){var t;return e.f.c.length!=0&&f(ct(e.f,0),72).a?f(ct(e.f,0),72).a:(t=Ice(e),t??""+(e.i?Oc(e.i.j,e,0):-1))}function zAn(e,t){var i,a;if(t<0||t>=e.gc())return null;for(i=t;i<e.gc();++i)if(a=f(e.Xb(i),132),i==e.gc()-1||!a.o)return new Ya(rt(i),a);return null}function O3e(e){var t,i,a,u,l;for(l=0,u=qo,a=0,i=new ne(e.a);i.a<i.c.c.length;)t=f(ee(i),172),l+=t.r+(a>0?e.c:0),u=h.Math.max(u,t.d),++a;e.e=l,e.b=u}function HAn(e){var t,i;if(!e.b)for(e.b=cJ(f(e.f,124).ih().i),i=new Nn(f(e.f,124).ih());i.e!=i.i.gc();)t=f(Mn(i),153),Qe(e.b,new _ae(t));return e.b}function GAn(e,t){var i,a,u;if(t.dc())return gP(),gP(),kV;for(i=new Kct(e,t.gc()),u=new Nn(e);u.e!=u.i.gc();)a=Mn(u),t.Gc(a)&&Qn(i,a);return i}function N3e(e,t,i,a){return t==0?a?(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),e.o):(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),AH(e.o)):CZ(e,t,i,a)}function Xle(e){var t,i;if(e.rb)for(t=0,i=e.rb.i;t<i;++t)Jz(be(e.rb,t));if(e.vb)for(t=0,i=e.vb.i;t<i;++t)Jz(be(e.vb,t));q7((Rl(),Ea),e),e.Bb|=1}function Ai(e,t,i,a,u,l,g,m,v,S,T,A,L,F){return Kyt(e,t,a,null,u,l,g,m,v,S,L,!0,F),VCe(e,T),me(e.Cb,88)&&OT(pd(f(e.Cb,88)),2),i&&zxe(e,i),WCe(e,A),e}function qAn(e){var t,i;if(e==null)return null;i=0;try{i=ff(e,Ra,Qr)&To}catch(a){if(a=go(a),me(a,130))t=AJ(e),i=t[0];else throw ie(a)}return qH(i)}function VAn(e){var t,i;if(e==null)return null;i=0;try{i=ff(e,Ra,Qr)&To}catch(a){if(a=go(a),me(a,130))t=AJ(e),i=t[0];else throw ie(a)}return qH(i)}function WAn(e,t){var i,a,u;return u=e.h-t.h,u<0||(i=e.l-t.l,a=e.m-t.m+(i>>22),u+=a>>22,u<0)?!1:(e.l=i&Ed,e.m=a&Ed,e.h=u&A1,!0)}function Jle(e,t,i,a,u,l,g){var m,v;return!(t.Re()&&(v=e.a.Le(i,a),v<0||!u&&v==0)||t.Se()&&(m=e.a.Le(i,l),m>0||!g&&m==0))}function KAn(e,t){WP();var i;if(i=e.j.g-t.j.g,i!=0)return 0;switch(e.j.g){case 2:return ble(t,sPe)-ble(e,sPe);case 4:return ble(e,aPe)-ble(t,aPe)}return 0}function YAn(e){switch(e.g){case 0:return Bge;case 1:return Uge;case 2:return zge;case 3:return Hge;case 4:return pte;case 5:return Gge;default:return null}}function Ps(e,t,i){var a,u;return a=(u=new mae,fE(u,t),ju(u,i),Qn((!e.c&&(e.c=new Be(S_,e,12,10)),e.c),u),u),x0(a,0),vT(a,1),N0(a,!0),O0(a,!0),a}function wR(e,t){var i,a;if(t>=e.i)throw ie(new vse(t,e.i));return++e.j,i=e.g[t],a=e.i-t-1,a>0&&au(e.g,t+1,e.g,t,a),io(e.g,--e.i,null),e.Ni(t,i),e.Ki(),i}function dwt(e,t){var i,a;return e.Db>>16==17?e.Cb.Ph(e,21,kh,t):(a=os(f(Vt((i=f(fn(e,16),29),i||e.ei()),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function XAn(e){var t,i,a,u;for(Ht(),na(e.c,e.a),u=new ne(e.c);u.a<u.c.c.length;)for(a=ee(u),i=new ne(e.b);i.a<i.c.c.length;)t=f(ee(i),678),t._e(a)}function JAn(e){var t,i,a,u;for(Ht(),na(e.c,e.a),u=new ne(e.c);u.a<u.c.c.length;)for(a=ee(u),i=new ne(e.b);i.a<i.c.c.length;)t=f(ee(i),374),t._e(a)}function ZAn(e){var t,i,a,u,l;for(u=Qr,l=null,a=new ne(e.d);a.a<a.c.c.length;)i=f(ee(a),216),i.d.j^i.e.j&&(t=i.e.e-i.d.e-i.a,t<u&&(u=t,l=i));return l}function R3e(){R3e=Z,bOt=new Pt(tNe,(mn(),!1)),pOt=new Pt(nNe,100),i5e=(cj(),lge),gOt=new Pt(rNe,i5e),mOt=new Pt(iNe,Tm),vOt=new Pt(oNe,rt(Qr))}function fwt(e,t){var i,a,u;for(a=new Tn(Cn(T1(e).a.Jc(),new B));Kn(a);)return i=f(Dn(a),18),u=f(t.Kb(i),9),new nS(or(u.n.b+u.o.b/2));return a7(),a7(),Dpe}function hwt(e,t,i){var a,u,l,g,m,v,S,T;for(S=0,u=e.a[t],l=0,g=u.length;l<g;++l)for(a=u[l],T=kj(a,i),v=T.Jc();v.Ob();)m=f(v.Pb(),11),qr(e.f,m,rt(S++))}function pwt(e,t,i){var a,u,l,g;if(i)for(u=i.a.length,a=new Xy(u),g=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);g.Ob();)l=f(g.Pb(),17),jt(e,t,SR(gT(i,l.a)))}function gwt(e,t,i){var a,u,l,g;if(i)for(u=i.a.length,a=new Xy(u),g=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);g.Ob();)l=f(g.Pb(),17),jt(e,t,SR(gT(i,l.a)))}function P3e(e){cw();var t;return t=f(Rj(YN(e.k),Pe(Ja,qc,64,2,0,1)),125),LP(t,0,t.length,null),t[0]==(tt(),hn)&&t[1]==pn&&(io(t,0,pn),io(t,1,hn)),t}function mwt(e,t,i){var a,u,l;return u=S5n(e,t,i),l=xIe(e,u),mue(e.b),Xce(e,t,i),Ht(),na(u,new stt(e)),a=xIe(e,u),mue(e.b),Xce(e,i,t),new Ya(rt(l),rt(a))}function bwt(){bwt=Z,V6t=Sr(new _o,(wa(),zs),(ka(),z$)),Yte=new Mi("linearSegments.inputPrio",rt(0)),Xte=new Mi("linearSegments.outputPrio",rt(0))}function yR(){yR=Z,Jte=new FY("P1_TREEIFICATION",0),fF=new FY("P2_NODE_ORDERING",1),hF=new FY("P3_NODE_PLACEMENT",2),pF=new FY("P4_EDGE_ROUTING",3)}function QAn(e){var t,i,a,u;for(i=0,t=0,u=new Nn(e);u.e!=u.i.gc();)a=f(Mn(u),27),i=h.Math.max(a.g+a.i,i),t=h.Math.max(a.f+a.j,t);return new Ke(i,t)}function eIn(e,t){var i,a,u,l;for(l=0,a=new ne(e);a.a<a.c.c.length;)i=f(ee(a),27),l+=h.Math.pow(i.g*i.f-t,2);return u=h.Math.sqrt(l/(e.c.length-1)),u}function df(){df=Z,dV=new qY("UNKNOWN",0),Rb=new qY("ABOVE",1),Nw=new qY("BELOW",2),lV=new qY("INLINE",3),new Mi("org.eclipse.elk.labelSide",dV)}function tIn(e,t){if(me(t,268))return fbn(e,f(t,85));if(me(t,273))return f(t,273);throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[t])))))}function vwt(e,t){var i;if(e.Vi()&&t!=null){for(i=0;i<e.i;++i)if(ei(t,e.g[i]))return i}else for(i=0;i<e.i;++i)if(Ae(e.g[i])===Ae(t))return i;return-1}function nIn(e,t,i){var a,u;return t.c==(cs(),Uu)&&i.c==td?-1:t.c==td&&i.c==Uu?1:(a=e0t(t.a,e.a),u=e0t(i.a,e.a),t.c==Uu?u-a:a-u)}function sw(e,t,i){if(i&&(t<0||t>i.a.c.length))throw ie(new dn("index must be >= 0 and <= layer node count"));e.c&&hl(e.c.a,e),e.c=i,i&&Ky(i.a,t,e)}function wwt(e,t){this.c=new Wn,this.a=e,this.b=t,this.d=f(V(e,(Ge(),CA)),313),Ae(V(e,(ot(),YLe)))===Ae((RH(),gte))?this.e=new vrt:this.e=new brt}function Zle(e,t){var i,a;i=e.dd(t);try{return a=i.Pb(),i.Qb(),a}catch(u){throw u=go(u),me(u,111)?ie(new Ru("Can't remove element "+t)):ie(u)}}function rIn(e,t){var i,a,u;if(a=new EY,u=new gCe(a.q.getFullYear()-hw,a.q.getMonth(),a.q.getDate()),i=FDn(e,t,u),i==0||i<t.length)throw ie(new dn(t));return u}function L3e(e,t){var i,a,u;for(an(t),f_e(t!=e),u=e.b.c.length,a=t.Jc();a.Ob();)i=a.Pb(),Qe(e.b,an(i));return u!=e.b.c.length?(SCe(e,0),!0):!1}function dG(){dG=Z,c5e=(Rr(),iV),new pa(qbe,(mn(),!0)),yOt=UE,EOt=GA,SOt=Ow,wOt=HA,l5e=oV,_Ot=px,s5e=(R3e(),bOt),o5e=gOt,a5e=mOt,u5e=vOt,Uee=pOt}function iIn(e,t){if(t==e.c)return e.d;if(t==e.d)return e.c;throw ie(new dn("'port' must be either the source port or target port of the edge."))}function oIn(e,t,i){var a,u;switch(u=e.o,a=e.d,t.g){case 1:return-a.d-i;case 3:return u.b+a.a+i;case 2:return u.a+a.c+i;case 4:return-a.b-i;default:return 0}}function D3e(e,t,i,a){var u,l,g,m;for(jo(t,f(a.Xb(0),25)),m=a.hd(1,a.gc()),l=f(i.Kb(t),20).Jc();l.Ob();)u=f(l.Pb(),18),g=u.c.i==t?u.d.i:u.c.i,D3e(e,g,i,m)}function ywt(e){var t;return t=new Wn,li(e,(Ge(),rme))?f(V(e,rme),91):(Zi(Jr(new Ft(null,new Bt(e.j,16)),new ui),new Iet(t)),$e(e,rme,t),t)}function Ewt(e,t){var i,a,u,l,g;for(a=0,u=0,i=0,g=new ne(e);g.a<g.c.c.length;)l=f(ee(g),186),a=h.Math.max(a,l.e),u+=l.b+(i>0?t:0),++i;return new Ke(a,u)}function Pj(e,t){var i,a;return a=null,e.nf((Rr(),a4))&&(i=f(e.mf(a4),102),i.nf(t)&&(a=i.mf(t))),a==null&&e.Rf()&&(a=e.Rf().mf(t)),a==null&&(a=st(t)),a}function M3e(e,t){var i,a;return e.Db>>16==6?e.Cb.Ph(e,6,vo,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Gc(),Rne)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function j3e(e,t){var i,a;return e.Db>>16==7?e.Cb.Ph(e,1,wV,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Gc(),_7e)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function $3e(e,t){var i,a;return e.Db>>16==9?e.Cb.Ph(e,9,Ur,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Gc(),T7e)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function Swt(e,t){var i,a;return e.Db>>16==5?e.Cb.Ph(e,9,Une,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Ut(),X0)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function _wt(e,t){var i,a;return e.Db>>16==7?e.Cb.Ph(e,6,Bp,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Ut(),Z0)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function F3e(e,t){var i,a;return e.Db>>16==3?e.Cb.Ph(e,0,EV,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Ut(),Y0)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function B3e(e,t){var i,a;return e.Db>>16==3?e.Cb.Ph(e,12,Ur,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Gc(),S7e)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function aIn(e,t,i){var a,u,l;for(i<0&&(i=0),l=e.i,u=i;u<l;u++)if(a=be(e,u),t==null){if(a==null)return u}else if(Ae(t)===Ae(a)||ei(t,a))return u;return-1}function sIn(e,t){var i,a;return i=t.mi(e.a),i?(a=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),See)),Mt(pq,a)?q7(e,cf(t.nk())):a):null}function cL(e,t){var i,a;if(t){if(t==e)return!0;for(i=0,a=f(t,52).Lh();a&&a!=t;a=a.Lh()){if(++i>Lfe)return cL(e,a);if(a==e)return!0}}return!1}function cIn(e){switch(dX(),e.q.g){case 5:dEt(e,(tt(),hn)),dEt(e,Bn);break;case 4:m2t(e,(tt(),hn)),m2t(e,Bn);break;default:gkt(e,(tt(),hn)),gkt(e,Bn)}}function uIn(e){switch(dX(),e.q.g){case 5:OEt(e,(tt(),_n)),OEt(e,pn);break;case 4:Ivt(e,(tt(),_n)),Ivt(e,pn);break;default:mkt(e,(tt(),_n)),mkt(e,pn)}}function lIn(e){var t,i;t=f(V(e,(Qh(),sOt)),17),t?(i=t.a,i==0?$e(e,(ew(),Bee),new vle):$e(e,(ew(),Bee),new hJ(i))):$e(e,(ew(),Bee),new hJ(1))}function dIn(e,t){var i;switch(i=e.i,t.g){case 1:return-(e.n.b+e.o.b);case 2:return e.n.a-i.o.a;case 3:return e.n.b-i.o.b;case 4:return-(e.n.a+e.o.a)}return 0}function fIn(e,t){switch(e.g){case 0:return t==(Ff(),Tw)?ate:ste;case 1:return t==(Ff(),Tw)?ate:xq;case 2:return t==(Ff(),Tw)?xq:ste;default:return xq}}function fG(e,t){var i,a,u;for(hl(e.a,t),e.e-=t.r+(e.a.c.length==0?0:e.c),u=GRe,a=new ne(e.a);a.a<a.c.c.length;)i=f(ee(a),172),u=h.Math.max(u,i.d);e.b=u}function hIn(e,t,i){var a,u,l;if(l=f(bn(e.r,t),297),u=null,l)switch(l.g){case 2:a=Oe(bn(e.i,t)),u=(an(i),i+(an(a),a));break;default:u=i}else u=i;return u}function pIn(e,t,i){var a,u,l;if(l=f(bn(e.r,t),297),u=null,l)switch(l.g){case 2:a=Oe(bn(e.j,t)),u=(an(i),i+(an(a),a));break;default:u=i}else u=i;return u}function ph(e){var t;return(e.Bb&1)==0&&e.r&&e.r.Rh()&&(t=f(e.r,52),e.r=f(iw(e,t),141),e.r!=t&&(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,9,8,t,e.r))),e.r}function Qle(e,t,i){var a;return a=ce(se(sa,1),_s,30,15,[hAe(e,(Np(),_c),t,i),hAe(e,Bu,t,i),hAe(e,kc,t,i)]),e.f&&(a[0]=h.Math.max(a[0],a[2]),a[2]=a[0]),a}function gIn(e,t){var i,a,u;if(u=$3n(e,t),u.c.length!=0)for(na(u,new al),i=u.c.length,a=0;a<i;a++)hL(e,(Gt(a,u.c.length),f(u.c[a],291)),pPn(e,u,a))}function Lj(e,t,i){var a,u;a=t*i,me(e.g,152)?(u=iR(e),u.f.d?u.f.a||(e.d.a+=a+Og):(e.d.d-=a+Og,e.d.a+=a+Og)):me(e.g,9)&&(e.d.d-=a,e.d.a+=2*a)}function mIn(e){var t,i,a,u;for(u=f(si(e.a,(CT(),cte)),15).Jc();u.Ob();)for(a=f(u.Pb(),105),i=YN(a.k).Jc();i.Ob();)t=f(i.Pb(),64),SP(e,a,t,(LS(),_w),1)}function bIn(e){var t,i;if(e.k==(cn(),uo)){for(i=new Tn(Cn(T1(e).a.Jc(),new B));Kn(i);)if(t=f(Dn(i),18),!Fa(t)&&e.c==K3e(t,e).c)return!0}return!1}function vIn(e){var t,i;if(e.k==(cn(),uo)){for(i=new Tn(Cn(T1(e).a.Jc(),new B));Kn(i);)if(t=f(Dn(i),18),!Fa(t)&&t.c.i.c==t.d.i.c)return!0}return!1}function U3e(e,t){var i,a;return e.Db>>16==11?e.Cb.Ph(e,10,Ur,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Gc(),k7e)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function kwt(e,t){var i,a;return e.Db>>16==10?e.Cb.Ph(e,11,kh,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Ut(),J0)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function Twt(e,t){var i,a;return e.Db>>16==10?e.Cb.Ph(e,12,Th,t):(a=os(f(Vt((i=f(fn(e,16),29),i||(Ut(),Sx)),e.Db>>16),19)),e.Cb.Ph(e,a.n,a.f,t))}function xwt(e,t){var i,a,u,l,g;if(t)for(u=t.a.length,i=new Xy(u),g=(i.b-i.a)*i.c<0?(zv(),Fw):new Vv(i);g.Ob();)l=f(g.Pb(),17),a=OP(t,l.a),a&&l2t(e,a)}function wIn(){tSe();var e,t;for(C7n((Wv(),sn)),m7n(sn),Xle(sn),F7e=(Ut(),Ug),t=new ne(K7e);t.a<t.c.c.length;)e=f(ee(t),246),kL(e,Ug,null);return!0}function z3e(e,t){var i,a,u,l,g,m,v,S;return v=e.h>>19,S=t.h>>19,v!=S?S-v:(u=e.h,m=t.h,u!=m?u-m:(a=e.m,g=t.m,a!=g?a-g:(i=e.l,l=t.l,i-l)))}function Cwt(e,t,i){var a,u,l,g,m;for(u=e[i.g],m=new ne(t.d);m.a<m.c.c.length;)g=f(ee(m),105),l=g.i,l&&l.i==i&&(a=g.d[i.g],u[a]=h.Math.max(u[a],l.j.b))}function yIn(e,t){var i,a,u,l,g;for(a=0,u=0,i=0,g=new ne(t.d);g.a<g.c.c.length;)l=f(ee(g),316),PZ(l),a=h.Math.max(a,l.b),u+=l.d+(i>0?e.b:0),++i;t.b=a,t.e=u}function Awt(e){var t,i,a;if(a=e.b,fot(e.i,a.length)){for(i=a.length*2,e.b=Pe(Fpe,BG,304,i,0,1),e.c=Pe(Fpe,BG,304,i,0,1),e.f=i-1,e.i=0,t=e.a;t;t=t.c)yG(e,t,t);++e.g}}function Dj(e,t){return e.b.a=h.Math.min(e.b.a,t.c),e.b.b=h.Math.min(e.b.b,t.d),e.a.a=h.Math.max(e.a.a,t.c),e.a.b=h.Math.max(e.a.b,t.d),vn(e.c,t),!0}function EIn(e,t,i){var a;a=t.c.i,a.k==(cn(),uo)?($e(e,(Ge(),$p),f(V(a,$p),11)),$e(e,yh,f(V(a,yh),11))):($e(e,(Ge(),$p),t.c),$e(e,yh,i.d))}function uL(e,t,i){bL();var a,u,l,g,m,v;return g=t/2,l=i/2,a=h.Math.abs(e.a),u=h.Math.abs(e.b),m=1,v=1,a>g&&(m=g/a),u>l&&(v=l/u),qv(e,h.Math.min(m,v)),e}function SIn(){dQ();var e,t;try{if(t=f(Z3e((Uv(),xh),WL),2059),t)return t}catch(i){if(i=go(i),me(i,103))e=i,Fke((lr(),e));else throw ie(i)}return new ZO}function _In(){dQ();var e,t;try{if(t=f(Z3e((Uv(),xh),bh),1986),t)return t}catch(i){if(i=go(i),me(i,103))e=i,Fke((lr(),e));else throw ie(i)}return new jh}function kIn(){Pgt();var e,t;try{if(t=f(Z3e((Uv(),xh),CE),2068),t)return t}catch(i){if(i=go(i),me(i,103))e=i,Fke((lr(),e));else throw ie(i)}return new GU}function TIn(e,t,i){var a,u;return u=e.e,e.e=t,(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,4,u,t),i?i.kj(a):i=a),u!=t&&(t?i=kL(e,JZ(e,t),i):i=kL(e,e.a,i)),i}function Iwt(){EY.call(this),this.e=-1,this.a=!1,this.p=Ra,this.k=-1,this.c=-1,this.b=-1,this.g=!1,this.f=-1,this.j=-1,this.n=-1,this.i=-1,this.d=-1,this.o=Ra}function xIn(e,t){var i,a,u;if(a=e.b.d.d,e.a||(a+=e.b.d.a),u=t.b.d.d,t.a||(u+=t.b.d.a),i=di(a,u),i==0){if(!e.a&&t.a)return-1;if(!t.a&&e.a)return 1}return i}function CIn(e,t){var i,a,u;if(a=e.b.b.d,e.a||(a+=e.b.b.a),u=t.b.b.d,t.a||(u+=t.b.b.a),i=di(a,u),i==0){if(!e.a&&t.a)return-1;if(!t.a&&e.a)return 1}return i}function AIn(e,t){var i,a,u;if(a=e.b.g.d,e.a||(a+=e.b.g.a),u=t.b.g.d,t.a||(u+=t.b.g.a),i=di(a,u),i==0){if(!e.a&&t.a)return-1;if(!t.a&&e.a)return 1}return i}function H3e(){H3e=Z,TOt=Il(Sr(Sr(Sr(new _o,(wa(),cu),(ka(),V5e)),cu,W5e),zs,K5e),zs,D5e),COt=Sr(Sr(new _o,cu,A5e),cu,M5e),xOt=Il(new _o,zs,$5e)}function IIn(e){var t,i,a,u,l;for(t=f(V(e,(Ge(),K$)),91),l=e.n,a=t.Bc().Jc();a.Ob();)i=f(a.Pb(),315),u=i.i,u.c+=l.a,u.d+=l.b,i.c?tSt(i):nSt(i);$e(e,K$,null)}function OIn(e,t,i){var a,u;switch(u=e.b,a=u.d,t.g){case 1:return-a.d-i;case 2:return u.o.a+a.c+i;case 3:return u.o.b+a.a+i;case 4:return-a.b-i;default:return-1}}function Owt(e,t){var i,a;for(a=new ne(t);a.a<a.c.c.length;)i=f(ee(a),9),e.c[i.c.p][i.p].a=yke(e.i),e.c[i.c.p][i.p].d=ke(e.c[i.c.p][i.p].a),e.c[i.c.p][i.p].b=1}function NIn(e,t){var i,a,u,l;for(l=0,a=new ne(e);a.a<a.c.c.length;)i=f(ee(a),162),l+=h.Math.pow(Al(i)*jd(i)-t,2);return u=h.Math.sqrt(l/(e.c.length-1)),u}function RIn(e,t,i){var a,u;for(i.Sg("Interactive node placement",1),e.a=f(V(t,(Ge(),CA)),313),u=new ne(t.b);u.a<u.c.c.length;)a=f(ee(u),25),ePn(e,a);i.Tg()}function PIn(e){var t,i,a,u,l;if(a=0,u=$R,e.b)for(t=0;t<360;t++)i=t*.017453292519943295,HIe(e,e.d,0,0,_E,i),l=e.b.Bg(e.d),l<u&&(a=i,u=l);HIe(e,e.d,0,0,_E,a)}function LIn(e,t){var i,a,u,l;for(l=new Wn,t.e=null,t.f=null,a=new ne(t.i);a.a<a.c.c.length;)i=f(ee(a),68),u=f(bn(e.g,i.a),47),i.a=$X(i.b),qr(l,i.a,u);e.g=l}function DIn(e,t,i){var a,u,l,g,m,v;for(u=t-e.e,l=u/e.d.c.length,g=0,v=new ne(e.d);v.a<v.c.c.length;)m=f(ee(v),316),a=e.b-m.b+i,evt(m,m.e+g*l,m.f),mAn(m,l,a),++g}function Nwt(e){var t;if(e.f.Yj(),e.b!=-1){if(++e.b,t=e.f.d[e.a],e.b<t.i)return;++e.a}for(;e.a<e.f.d.length;++e.a)if(t=e.f.d[e.a],t&&t.i!=0){e.b=0;return}e.b=-1}function MIn(e,t){var i,a,u;for(u=t.c.length,i=LRn(e,u==0?"":(Gt(0,t.c.length),dr(t.c[0]))),a=1;a<u&&i;++a)i=f(i,52).Vh((Gt(a,t.c.length),dr(t.c[a])));return i}function Rwt(e,t,i,a){var u,l,g;return l=XPn(e,t,i,a),g=LIe(e,l),ude(e,t,i,a),mue(e.b),Ht(),na(l,new ctt(e)),u=LIe(e,l),ude(e,i,t,a),mue(e.b),new Ya(rt(g),rt(u))}function jIn(e,t){var i;t.Sg("Delaunay triangulation",1),i=new et,Mu(e.i,new Gtt(i)),lt(ut(V(e,(HP(),F$)))),e.e?Ba(e.e,Pkt(i)):e.e=Pkt(i),t.Tg()}function $In(e,t,i){var a,u;for(Yz(e,e.j+t,e.k+i),u=new Nn((!e.a&&(e.a=new So(yf,e,5)),e.a));u.e!=u.i.gc();)a=f(Mn(u),369),Ese(a,a.a+t,a.b+i);Kz(e,e.b+t,e.c+i)}function G3e(e,t,i,a){switch(i){case 7:return!e.e&&(e.e=new Qt(vo,e,7,4)),$u(e.e,t,a);case 8:return!e.d&&(e.d=new Qt(vo,e,8,5)),$u(e.d,t,a)}return ade(e,t,i,a)}function q3e(e,t,i,a){switch(i){case 7:return!e.e&&(e.e=new Qt(vo,e,7,4)),za(e.e,t,a);case 8:return!e.d&&(e.d=new Qt(vo,e,8,5)),za(e.d,t,a)}return Gue(e,t,i,a)}function Pwt(e,t,i){var a,u,l,g,m;if(i)for(l=i.a.length,a=new Xy(l),m=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);m.Ob();)g=f(m.Pb(),17),u=OP(i,g.a),u&&lEt(e,u,t)}function hG(e,t,i){var a,u,l,g,m;return e.Yj(),l=t==null?0:Ti(t),e.f>0&&(g=(l&Qr)%e.d.length,u=VAe(e,g,l,t),u)?(m=u.ld(i),m):(a=e._j(l,t,i),e.c.Ec(a),null)}function V3e(e,t){var i,a,u,l;switch(I0(e,t).Hl()){case 3:case 2:{for(i=rA(t),u=0,l=i.i;u<l;++u)if(a=f(be(i,u),35),kS(Ns(e,a))==5)return a;break}}return null}function FIn(e){var t,i,a,u,l;if(fot(e.f,e.b.length))for(a=Pe(RAt,BG,226,e.b.length*2,0,1),e.b=a,u=a.length-1,i=e.a;i!=e;i=i.Zd())l=f(i,226),t=l.d&u,l.a=a[t],a[t]=l}function Lwt(e,t){var i,a,u,l;for(l=0,u=f(f(si(e.r,t),22),83).Jc();u.Ob();)a=f(u.Pb(),115),l=h.Math.max(l,a.e.a+a.b.Kf().a);i=f(Es(e.b,t),126),i.n.b=0,i.a.a=l}function Dwt(e,t){var i,a,u,l;for(i=0,l=f(f(si(e.r,t),22),83).Jc();l.Ob();)u=f(l.Pb(),115),i=h.Math.max(i,u.e.b+u.b.Kf().b);a=f(Es(e.b,t),126),a.n.d=0,a.a.b=i}function BIn(e){var t,i,a;for(i=new ne(e.p);i.a<i.c.c.length;)t=f(ee(i),9),t.k==(cn(),Yi)&&(a=t.o.b,e.i=h.Math.min(e.i,a),e.g=h.Math.max(e.g,a))}function Mwt(e,t,i){var a,u,l;for(l=new ne(t);l.a<l.c.c.length;)a=f(ee(l),9),e.c[a.c.p][a.p].e=!1;for(u=new ne(t);u.a<u.c.c.length;)a=f(ee(u),9),bOe(e,a,i)}function UIn(e){var t,i;return i=f(V(e,(Ge(),Cu)),22),t=nX(c5t),i.Gc((us(),_A))&&lf(t,d5t),i.Gc(W$)&&lf(t,f5t),i.Gc(dD)&&lf(t,u5t),i.Gc(fD)&&lf(t,l5t),t}function W3e(e){if(e<0)throw ie(new dn("The input must be positive"));return e<X8e.length?oE(X8e[e]):h.Math.sqrt(_E*e)*(Nxn(e,e)/OCe(2.718281828459045,e))}function lL(e,t){var i;if(e.Vi()&&t!=null){for(i=0;i<e.i;++i)if(ei(t,e.g[i]))return!0}else for(i=0;i<e.i;++i)if(Ae(e.g[i])===Ae(t))return!0;return!1}function zIn(e,t){if(t==null){for(;e.a.Ob();)if(f(e.a.Pb(),43).kd()==null)return!0}else for(;e.a.Ob();)if(ei(t,f(e.a.Pb(),43).kd()))return!0;return!1}function HIn(e,t){var i,a,u;return t===e?!0:me(t,663)?(u=f(t,1992),fvt((a=e.g,a||(e.g=new rS(e))),(i=u.g,i||(u.g=new rS(u))))):!1}function GIn(e){var t,i,a,u;for(t="$z",i="nz",u=h.Math.min(e.length,5),a=u-1;a>=0;a--)if(Mt(e[a].d,t)||Mt(e[a].d,i)){e.length>=a+1&&e.splice(0,a+1);break}return e}function pG(e,t){var i;return Js(e)&&Js(t)&&(i=e/t,HG<i&&i<yE)?i<0?h.Math.ceil(i):h.Math.floor(i):tw(dOe(Js(e)?dh(e):e,Js(t)?dh(t):t,!1))}function K3e(e,t){if(t==e.c.i)return e.d.i;if(t==e.d.i)return e.c.i;throw ie(new dn("'node' must either be the source node or target node of the edge."))}function qIn(e){var t,i,a,u;if(u=f(V(e,(Ge(),jPe)),36),u){for(a=new ha,t=ra(e.c.i);t!=u;)i=t.e,t=ra(i),eT(ii(ii(a,i.n),t.c),t.d.b,t.d.d);return a}return qOt}function VIn(e){var t;t=f(V(e,(Ge(),f_)),333),Zi(Zs(new Ft(null,new Bt(t.d,16)),new mO),new Cet(e)),Zi(Jr(new Ft(null,new Bt(t.d,16)),new NU),new Aet(e))}function ede(e,t){var i,a,u,l;for(u=t?ao(e):Fo(e),a=new Tn(Cn(u.a.Jc(),new B));Kn(a);)if(i=f(Dn(a),18),l=K3e(i,e),l.k==(cn(),uo)&&l.c!=e.c)return l;return null}function tde(e,t,i){var a,u;a=W3(t.j,i.s,i.c)+W3(i.e,t.s,t.c),u=W3(i.j,t.s,t.c)+W3(t.e,i.s,i.c),a==u?a>0&&(e.b+=2,e.a+=a):(e.b+=1,e.a+=h.Math.min(a,u))}function jwt(e,t){var i,a;if(a=!1,ta(t)&&(a=!0,QN(e,new cT(dr(t)))),a||me(t,241)&&(a=!0,QN(e,(i=zse(f(t,241)),new oS(i)))),!a)throw ie(new kae(U4e))}function WIn(e,t,i,a){var u,l,g;return u=new v1(e.e,1,10,(g=t.c,me(g,88)?f(g,29):(Ut(),Ch)),(l=i.c,me(l,88)?f(l,29):(Ut(),Ch)),P0(e,t),!1),a?a.kj(u):a=u,a}function Y3e(e){var t,i;switch(f(V(ra(e),(ot(),ULe)),417).g){case 0:return t=e.n,i=e.o,new Ke(t.a+i.a/2,t.b+i.b/2);case 1:return new ja(e.n);default:return null}}function gG(){gG=Z,mte=new _7(Lp,0),_Pe=new _7("LEFTUP",1),TPe=new _7("RIGHTUP",2),SPe=new _7("LEFTDOWN",3),kPe=new _7("RIGHTDOWN",4),qge=new _7("BALANCED",5)}function KIn(e,t,i){var a,u,l;if(a=di(e.a[t.p],e.a[i.p]),a==0){if(u=f(V(t,(Ge(),XR)),15),l=f(V(i,XR),15),u.Gc(i))return-1;if(l.Gc(t))return 1}return a}function YIn(e){switch(e.g){case 1:return new MC;case 2:return new _v;case 3:return new BO;case 0:return null;default:throw ie(new dn(Bhe+(e.f!=null?e.f:""+e.g)))}}function X3e(e,t,i){switch(t){case 1:!e.n&&(e.n=new Be(hc,e,1,7)),Jn(e.n),!e.n&&(e.n=new Be(hc,e,1,7)),Qi(e.n,f(i,16));return;case 2:FP(e,dr(i));return}mCe(e,t,i)}function J3e(e,t,i){switch(t){case 3:NS(e,ke(Oe(i)));return;case 4:RS(e,ke(Oe(i)));return;case 5:md(e,ke(Oe(i)));return;case 6:bd(e,ke(Oe(i)));return}X3e(e,t,i)}function MZ(e,t,i){var a,u,l;l=(a=new mae,a),u=xg(l,t,null),u&&u.lj(),ju(l,i),Qn((!e.c&&(e.c=new Be(S_,e,12,10)),e.c),l),x0(l,0),vT(l,1),N0(l,!0),O0(l,!0)}function Z3e(e,t){var i,a,u;return i=w7(e.i,t),me(i,240)?(u=f(i,240),u.vi()==null,u.si()):me(i,491)?(a=f(i,1983),u=a.b,u):null}function XIn(e,t,i,a){var u,l;return or(t),or(i),l=f(U7(e.d,t),17),imt(!!l,"Row %s not in %s",t,e.e),u=f(U7(e.b,i),17),imt(!!u,"Column %s not in %s",i,e.c),h1t(e,l.a,u.a,a)}function $wt(e,t,i,a,u,l,g){var m,v,S,T,A;if(T=u[l],S=l==g-1,m=S?a:0,A=twt(m,T),a!=10&&ce(se(e,g-l),t[l],i[l],m,A),!S)for(++l,v=0;v<T;++v)A[v]=$wt(e,t,i,a,u,l,g);return A}function JIn(e){var t,i,a,u,l;for(a=new yT(new Vi(e.b).a);a.b;)i=G3(a),t=f(i.jd(),9),l=f(f(i.kd(),47).a,9),u=f(f(i.kd(),47).b,8),ii(Cp(t.n),ii($a(l.n),u))}function ZIn(e,t){var i,a,u,l;for(l=new ne(t.a);l.a<l.c.c.length;)for(u=f(ee(l),9),c7(e.d),a=new Tn(Cn(ao(u).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),UEt(e,u,i.d.i)}function Mj(e){if(e.g==-1)throw ie(new Tl);e.Uj();try{e.i.ed(e.g),e.f=e.i.j,e.g<e.e&&--e.e,e.g=-1}catch(t){throw t=go(t),me(t,97)?ie(new If):ie(t)}}function QIn(e){var t,i,a,u;for(u=-1,a=0,i=new ne(e);i.a<i.c.c.length;){if(t=f(ee(i),248),t.c==(cs(),td)){u=a==0?0:a-1;break}else a==e.c.length-1&&(u=a);a+=1}return u}function eOn(e){var t,i,a,u;for(u=0,t=0,a=new ne(e.c);a.a<a.c.c.length;)i=f(ee(a),27),md(i,e.e+u),bd(i,e.f),u+=i.g+e.b,t=h.Math.max(t,i.f+e.b);e.d=u-e.b,e.a=t-e.b}function J3(e){var t,i,a;for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),60),a=t.d.c,t.d.c=t.d.d,t.d.d=a,a=t.d.b,t.d.b=t.d.a,t.d.a=a,a=t.b.a,t.b.a=t.b.b,t.b.b=a;wIe(e)}function Z3(e){var t,i,a;for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),82),a=t.g.c,t.g.c=t.g.d,t.g.d=a,a=t.g.b,t.g.b=t.g.a,t.g.a=a,a=t.e.a,t.e.a=t.e.b,t.e.b=a;tQ(e)}function tOn(e){var t,i,a,u,l;for(l=YN(e.k),i=(tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])),a=0,u=i.length;a<u;++a)if(t=i[a],t!=fc&&!l.Gc(t))return t;return null}function nde(e,t){var i,a;return a=f(zN(S1(Jr(new Ft(null,new Bt(t.j,16)),new De))),11),a&&(i=f(ct(a.e,0),18),i)?f(V(i,(Ge(),hi)),17).a:j_n(e.c)}function Fwt(e){var t,i;return i=kde(e),NP(i)?null:(t=(or(i),f(Wvt(new Tn(Cn(i.a.Jc(),new B))),85)),Us(f(be((!t.b&&(t.b=new Qt(Yn,t,4,7)),t.b),0),84)))}function jZ(e){var t;return e.o||(t=e.rk(),t?e.o=new Vke(e,e,null):e.Zk()?e.o=new A_e(e,null):kS(Ns((Rl(),Ea),e))==1?e.o=new apt(e):e.o=new Fse(e,null)),e.o}function nOn(e,t,i,a){var u,l,g,m,v;i.Th(t)&&(u=(g=t,g?f(a,52).ci(g):null),u&&(v=i.Ih(t),m=t.t,m>1||m==-1?(l=f(v,15),u.Wb(DCn(e,l))):u.Wb($de(e,f(v,57)))))}function rOn(e,t,i,a){wot();var u=Lpe;function l(){for(var g=0;g<u.length;g++)u[g]()}if(e)try{YMt(l)()}catch(g){e(t,g)}else YMt(l)()}function iOn(e,t){var i,a,u,l;for(u=(l=new Ic(e.b).a.vc().Jc(),new Hi(l));u.a.Ob();)if(a=(i=f(u.a.Pb(),43),f(i.jd(),34)),DSe(t,f(a,17))<0)return!1;return!0}function oOn(e,t){var i,a,u,l;for(u=(l=new Ic(e.b).a.vc().Jc(),new Hi(l));u.a.Ob();)if(a=(i=f(u.a.Pb(),43),f(i.jd(),34)),DSe(t,f(a,17))>0)return!1;return!0}function aOn(e){switch(f(V(e.b,(ot(),LLe)),379).g){case 1:Zi(Lu(Zs(new Ft(null,new Bt(e.d,16)),new nb),new qu),new sl);break;case 2:MPn(e);break;case 0:PRn(e)}}function sOn(e,t,i){var a,u,l;for(a=i,!a&&(a=new CN),a.Sg("Layout",e.a.c.length),l=new ne(e.a);l.a<l.c.c.length;){if(u=f(ee(l),44),a.Yg())return;u.If(t,a.bh(1))}a.Tg()}function cOn(e,t){var i,a;for(hl(e.b,t),a=new ne(e.n);a.a<a.c.c.length;)if(i=f(ee(a),208),Oc(i.c,t,0)!=-1){hl(i.c,t),eOn(i),i.c.c.length==0&&hl(e.n,i);break}__t(e)}function Bwt(e,t){var i,a,u,l,g;for(g=e.f,u=0,l=0,a=new ne(e.a);a.a<a.c.c.length;)i=f(ee(a),172),cG(i,e.e,g),IH(i,t),l=h.Math.max(l,i.r),g+=i.d+e.c,u=g;e.d=l,e.b=u}function mE(){mE=Z,Hbe=new x7("V_TOP",0),IF=new x7("V_CENTER",1),AF=new x7("V_BOTTOM",2),zbe=new x7("H_LEFT",3),xF=new x7("H_CENTER",4),CF=new x7("H_RIGHT",5)}function Q3e(e){var t;return(e.Db&64)!=0?mZ(e):(t=new sh(mZ(e)),t.a+=" (abstract: ",m0(t,(e.Bb&256)!=0),t.a+=", interface: ",m0(t,(e.Bb&512)!=0),t.a+=")",t.a)}function uOn(e){var t;e.c==null&&(t=Ae(e.b)===Ae(p6e)?null:e.b,e.d=t==null?bl:$lt(t)?Dmn(eft(t)):ta(t)?POe:Vy(Fd(t)),e.a=e.a+": "+($lt(t)?Obn(eft(t)):t+""),e.c="("+e.d+") "+e.a)}function lOn(){function e(){try{return new Map().entries().next().done}catch{return!1}}return typeof Map===cfe&&Map.prototype.entries&&e()?Map:w9n()}function dOn(e,t){var i,a,u,l;for(l=new la(e.e,0),i=0;l.b<l.d.gc();){if(a=ke((Ln(l.b<l.d.gc()),Oe(l.d.Xb(l.c=l.b++)))),u=a-t,u>Mhe)return i;u>-1e-6&&++i}return i}function $Z(e,t,i){if(me(t,268))return M6n(e,f(t,85),i);if(me(t,273))return hIn(e,f(t,273),i);throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[t,i])))))}function FZ(e,t,i){if(me(t,268))return j6n(e,f(t,85),i);if(me(t,273))return pIn(e,f(t,273),i);throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[t,i])))))}function eAe(e,t){var i;t!=e.b?(i=null,e.b&&(i=eJ(e.b,e,-4,i)),t&&(i=vR(t,e,-4,i)),i=u0t(e,t,i),i&&i.lj()):(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,3,t,t))}function Uwt(e,t){var i;t!=e.f?(i=null,e.f&&(i=eJ(e.f,e,-1,i)),t&&(i=vR(t,e,-1,i)),i=l0t(e,t,i),i&&i.lj()):(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,0,t,t))}function fOn(e,t,i,a){var u,l,g,m;return Ld(e.e)&&(u=t.Ik(),m=t.kd(),l=i.kd(),g=Yv(e,1,u,m,l,u.Gk()?EL(e,u,l,me(u,101)&&(f(u,19).Bb&Xa)!=0):-1,!0),a?a.kj(g):a=g),a}function zwt(e){var t,i,a;if(e==null)return null;if(i=f(e,15),i.dc())return"";for(a=new p0,t=i.Jc();t.Ob();)ys(a,(fi(),dr(t.Pb()))),a.a+=" ";return yse(a,a.a.length-1)}function Hwt(e){var t,i,a;if(e==null)return null;if(i=f(e,15),i.dc())return"";for(a=new p0,t=i.Jc();t.Ob();)ys(a,(fi(),dr(t.Pb()))),a.a+=" ";return yse(a,a.a.length-1)}function hOn(e,t){var i,a,u,l,g;for(l=new ne(t.a);l.a<l.c.c.length;)for(u=f(ee(l),9),a=new Tn(Cn(Fo(u).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),g=i.c.i.p,e.n[g]=e.n[g]-1}function pOn(e,t,i){var a,u;return a=e.c[t.c.p][t.p],u=e.c[i.c.p][i.p],a.a!=null&&u.a!=null?lce(a.a,u.a):a.a!=null?-1:u.a!=null?1:0}function gOn(e,t,i){return i.Sg("Tree layout",1),pH(e.b),Ap(e.b,(yR(),Jte),Jte),Ap(e.b,fF,fF),Ap(e.b,hF,hF),Ap(e.b,pF,pF),e.a=NG(e.b,t),sOn(e,t,i.bh(1)),i.Tg(),t}function Gwt(e,t){var i,a,u,l,g,m;if(t)for(l=t.a.length,i=new Xy(l),m=(i.b-i.a)*i.c<0?(zv(),Fw):new Vv(i);m.Ob();)g=f(m.Pb(),17),u=OP(t,g.a),a=new dnt(e),Fdt(a.a,u)}function qwt(e,t){var i,a,u,l,g,m;if(t)for(l=t.a.length,i=new Xy(l),m=(i.b-i.a)*i.c<0?(zv(),Fw):new Vv(i);m.Ob();)g=f(m.Pb(),17),u=OP(t,g.a),a=new nnt(e),$dt(a.a,u)}function mOn(e){var t;if(e!=null&&e.length>0&&_a(e,e.length-1)==33)try{return t=o2t(uh(e,0,e.length-1)),t.e==null}catch(i){if(i=go(i),!me(i,32))throw ie(i)}return!1}function bOn(e,t,i){var a,u,l;switch(a=ra(t),u=dZ(a),l=new ou,cc(l,t),i.g){case 1:No(l,iG(mR(u)));break;case 2:No(l,mR(u))}return $e(l,(ot(),ox),Oe(V(e,ox))),l}function tAe(e){var t,i;return t=f(Dn(new Tn(Cn(Fo(e.a).a.Jc(),new B))),18),i=f(Dn(new Tn(Cn(ao(e.a).a.Jc(),new B))),18),lt(ut(V(t,(Ge(),U0))))||lt(ut(V(i,U0)))}function CT(){CT=Z,Cq=new Dz("ONE_SIDE",0),lte=new Dz("TWO_SIDES_CORNER",1),dte=new Dz("TWO_SIDES_OPPOSING",2),ute=new Dz("THREE_SIDES",3),cte=new Dz("FOUR_SIDES",4)}function Vwt(e,t){var i,a,u,l;for(l=new et,u=0,a=t.Jc();a.Ob();){for(i=rt(f(a.Pb(),17).a+u);i.a<e.f&&!Bvn(e,i.a);)i=rt(i.a+1),++u;if(i.a>=e.f)break;vn(l.c,i)}return l}function vOn(e){var t,i;for(i=new ne(e.e.b);i.a<i.c.c.length;)t=f(ee(i),25),B8n(e,t);Zi(Jr(Zs(Zs(new Ft(null,new Bt(e.e.b,16)),new i0),new Zb),new TC),new Stt(e))}function nAe(e,t){return t?e.jj(t)?!1:e.i?e.i.kj(t):me(t,150)?(e.i=f(t,150),!0):(e.i=new c0,e.i.kj(t)):!1}function Wwt(e,t,i){var a,u,l;return a=t.Ik(),l=t.kd(),u=a.Gk()?Yv(e,3,a,null,l,EL(e,a,l,me(a,101)&&(f(a,19).Bb&Xa)!=0),!0):Yv(e,1,a,a.fk(),l,-1,!0),i?i.kj(u):i=u,i}function wOn(e){if(e=_u(e,!0),Mt(qL,e)||Mt("1",e))return mn(),QL;if(Mt(epe,e)||Mt("0",e))return mn(),yw;throw ie(new Tae("Invalid boolean value: '"+e+"'"))}function rAe(e,t,i){var a,u,l;for(u=e.vc().Jc();u.Ob();)if(a=f(u.Pb(),43),l=a.jd(),Ae(t)===Ae(l)||t!=null&&ei(t,l))return i&&(a=new AY(a.jd(),a.kd()),u.Qb()),a;return null}function yOn(e){tT();var t,i,a;e.B.Gc((yd(),xne))&&(a=e.f.i,t=new ZX(e.a.c),i=new xN,i.b=t.c-a.c,i.d=t.d-a.d,i.c=a.c+a.b-(t.c+t.b),i.a=a.d+a.a-(t.d+t.a),e.e.Yf(i))}function Kwt(e,t,i,a){var u,l,g;for(g=h.Math.min(i,d_t(f(e.b,68),t,i,a)),l=new ne(e.a);l.a<l.c.c.length;)u=f(ee(l),224),u!=t&&(g=h.Math.min(g,Kwt(u,t,g,a)));return g}function iAe(e){var t,i,a,u;for(u=Pe(Ib,Ye,199,e.b.c.length,0,2),a=new la(e.b,0);a.b<a.d.gc();)t=(Ln(a.b<a.d.gc()),f(a.d.Xb(a.c=a.b++),25)),i=a.b-1,u[i]=EH(t.a);return u}function oAe(e,t,i){var a,u,l;a=f(Eg(e.a,i),34),a!=null&&(l=f(Eg(e.b,a),67),_T(l,i,!0)),u=f(Eg(e.b,t),67),u||(u=new ki,pE(e.b,t,u)),Wi(u,i,u.c.b,u.c),pE(e.a,i,t)}function rde(e,t,i,a,u){var l,g,m,v;for(g=Ipn(F2e(V_e(Q3n(i)),a),oIn(e,i,u)),v=L0(e,i).Jc();v.Ob();)m=f(v.Pb(),11),t[m.p]&&(l=t[m.p].i,Qe(g.d,new cce(l,BCe(g,l))));r3e(g)}function aAe(e,t){this.f=new Wn,this.b=new Wn,this.j=new Wn,this.a=e,this.c=t,this.c>0&&hwt(this,this.c-1,(tt(),_n)),this.c<this.a.length-1&&hwt(this,this.c+1,(tt(),pn))}function EOn(e,t){var i,a,u,l,g;for(l=new ne(t.d);l.a<l.c.c.length;)for(u=f(ee(l),105),g=f(bn(e.c,u),116).o,a=new LN(u.b);a.a<a.c.a.length;)i=f(aj(a),64),cxe(u,i,g)}function sAe(e){e.length>0&&e[0].length>0&&(this.c=lt(ut(V(ra(e[0][0]),(Ge(),BPe))))),this.a=Pe(L6t,Ye,2063,e.length,0,2),this.b=Pe(D6t,Ye,2064,e.length,0,2),this.d=new K1t}function SOn(e){return e.c.length==0?!1:(Gt(0,e.c.length),f(e.c[0],18)).c.i.k==(cn(),uo)?!0:j3(Lu(new Ft(null,new Bt(e,16)),new C8),new TO)}function Ywt(e,t){var i,a,u,l,g,m,v;for(m=RT(t),l=t.f,v=t.g,g=h.Math.sqrt(l*l+v*v),u=0,a=new ne(m);a.a<a.c.c.length;)i=f(ee(a),27),u+=Ywt(e,i);return h.Math.max(u,g)}function oa(){oa=Z,Pw=new A7(s$,0),Fg=new A7("FREE",1),MD=new A7("FIXED_SIDE",2),HE=new A7("FIXED_ORDER",3),Lb=new A7("FIXED_RATIO",4),uu=new A7("FIXED_POS",5)}function _On(e,t){var i,a,u;if(i=t.mi(e.a),i){for(u=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),vh)),a=1;a<(Rl(),Z7e).length;++a)if(Mt(Z7e[a],u))return a}return 0}function kOn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],b1(l,""+t);return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function TOn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],b1(l,""+t);return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function Xwt(e){var t,i,a;for(a=new uE(Fu,"{","}"),i=e.vc().Jc();i.Ob();)t=f(i.Pb(),43),b1(a,ept(e,t.jd())+"="+ept(e,t.kd()));return a.a?a.e.length==0?a.a.a:a.a.a+(""+a.e):a.c}function xOn(e){for(var t,i,a,u;!u7(e.o);)i=f(JN(e.o),47),a=f(i.a,123),t=f(i.b,216),u=hZ(t,a),t.e==a?(uX(u.g,t),a.e=u.e+t.a):(uX(u.b,t),a.e=u.e-t.a),Qe(e.e.a,a)}function cAe(e,t){var i,a,u;for(i=null,u=f(t.Kb(e),20).Jc();u.Ob();)if(a=f(u.Pb(),18),!i)i=a.c.i==e?a.d.i:a.c.i;else if((a.c.i==e?a.d.i:a.c.i)!=i)return!1;return!0}function Jwt(e,t){var i,a,u,l,g;for(i=y2t(e,!1,t),u=new ne(i);u.a<u.c.c.length;)a=f(ee(u),133),a.d==0?(Gce(a,null),qce(a,null)):(l=a.a,g=a.b,Gce(a,g),qce(a,l))}function COn(e){var t,i;return t=new _o,lf(t,g5t),i=f(V(e,(Ge(),Cu)),22),i.Gc((us(),W$))&&lf(t,w5t),i.Gc(dD)&&lf(t,m5t),i.Gc(_A)&&lf(t,v5t),i.Gc(fD)&&lf(t,b5t),t}function uAe(e,t,i){var a,u,l,g,m;for(BCn(e),u=(e.k==null&&(e.k=Pe(Upe,Ye,80,0,0,1)),e.k),l=0,g=u.length;l<g;++l)a=u[l],uAe(a);m=e.f,m&&uAe(m)}function AOn(e){var t,i,a,u;for(GMn(e),i=new Tn(Cn(T1(e).a.Jc(),new B));Kn(i);)t=f(Dn(i),18),a=t.c.i==e,u=a?t.d:t.c,a?ua(t,null):Ia(t,null),$e(t,(Ge(),GPe),u),O4n(e,u.i)}function IOn(e,t,i,a){var u,l;switch(l=t.i,u=i[l.g][e.d[l.g]],l.g){case 1:u-=a+t.j.b,t.g.b=u;break;case 3:u+=a,t.g.b=u;break;case 4:u-=a+t.j.a,t.g.a=u;break;case 2:u+=a,t.g.a=u}}function OOn(e,t,i){var a,u,l,g;for(g=Oc(e.e,t,0),l=new g2e,l.b=i,a=new la(e.e,g);a.b<a.d.gc();)u=(Ln(a.b<a.d.gc()),f(a.d.Xb(a.c=a.b++),9)),u.p=i,Qe(l.e,u),hd(a);return l}function NOn(e){var t,i,a;for(i=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));i.e!=i.i.gc();)if(t=f(Mn(i),27),a=kde(t),!Kn(new Tn(Cn(a.a.Jc(),new B))))return t;return null}function BZ(){BZ=Z,sbe=new Bz("OVERLAP_REMOVAL",0),ibe=new Bz(jxt,1),cbe=new Bz("ROTATION",2),obe=new Bz("GRAPH_SIZE_CALCULATION",3),abe=new Bz("OUTGOING_EDGE_ANGLES",4)}function ROn(){var e;return VDt?f(gL((Uv(),xh),WL),2061):(e=f(me(wu((Uv(),xh),WL),552)?wu(xh,WL):new qEt,552),VDt=!0,x8n(e),B7n(e),Xle(e),Os(xh,WL,e),e)}function ide(e,t,i){var a,u;if(e.j==0)return i;if(u=f(ibt(e,t,i),74),a=i.Ik(),!a.ok()||!e.a.Zl(a))throw ie(new oc("Invalid entry feature '"+a.nk().zb+"."+a.ve()+"'"));return u}function POn(e,t){var i,a,u,l,g,m,v,S;for(m=e.a,v=0,S=m.length;v<S;++v)for(g=m[v],a=g,u=0,l=a.length;u<l;++u)if(i=a[u],Ae(t)===Ae(i)||t!=null&&ei(t,i))return!0;return!1}function LOn(e){var t,i,a;return Eu(e,0)>=0?(i=pG(e,SQ),a=kle(e,SQ)):(t=Jy(e,1),i=pG(t,5e8),a=kle(t,5e8),a=Ua(hb(a,1),ia(e,1))),hm(hb(a,32),ia(i,ls))}function Zwt(e,t,i,a){var u,l,g,m,v;switch(v=e.b,l=t.d,g=l.j,m=w3e(g,v.d[g.g],i),u=ii($a(l.n),l.a),l.j.g){case 1:case 3:m.a+=u.a;break;case 2:case 4:m.b+=u.b}Wi(a,m,a.c.b,a.c)}function DOn(e,t,i,a){var u,l,g,m,v;for(u=null,l=0,m=new ne(t);m.a<m.c.c.length;)g=f(ee(m),27),v=g.i+g.g,e<g.j+g.f+a&&(u?i.i-v<i.i-l&&(u=g):u=g,l=u.i+u.g);return u?l+a:0}function MOn(e,t,i,a){var u,l,g,m,v;for(l=null,u=0,m=new ne(t);m.a<m.c.c.length;)g=f(ee(m),27),v=g.j+g.f,e<g.i+g.g+a&&(l?i.j-v<i.j-u&&(l=g):l=g,u=l.j+l.f);return l?u+a:0}function jOn(e){var t,i,a;for(t=!1,a=e.b.c.length,i=0;i<a;i++)bCe(f(ct(e.b,i),431))?!t&&i+1<a&&bCe(f(ct(e.b,i+1),431))&&(t=!0,f(ct(e.b,i),431).a=!0):t=!1}function $On(e,t,i,a,u){var l,g;for(l=0,g=0;g<u;g++)l=Ua(l,hh(ia(t[g],ls),ia(a[g],ls))),e[g]=mr(l),l=wS(l,32);for(;g<i;g++)l=Ua(l,ia(t[g],ls)),e[g]=mr(l),l=wS(l,32)}function FOn(e,t){vL();var i,a;for(a=(ym(),Oee),i=e;t>1;t>>=1)(t&1)!=0&&(a=M3(a,i)),i.d==1?i=M3(i,i):i=new hvt(BSt(i.a,i.d,Pe(pr,Pr,30,i.d<<1,15,1)));return a=M3(a,i),a}function lAe(){lAe=Z;var e,t,i,a;for(B6e=Pe(sa,_s,30,25,15,1),U6e=Pe(sa,_s,30,33,15,1),a=152587890625e-16,t=32;t>=0;t--)U6e[t]=a,a*=.5;for(i=1,e=24;e>=0;e--)B6e[e]=i,i*=.5}function BOn(e){var t,i;if(lt(ut(Fe(e,(ot(),ix))))){for(i=new Tn(Cn(lw(e).a.Jc(),new B));Kn(i);)if(t=f(Dn(i),85),HS(t)&&lt(ut(Fe(t,PE))))return!0}return!1}function Qwt(e){var t,i,a,u;for(t=new ki,i=new ki,u=tr(e,0);u.b!=u.d.c;)a=f(Zn(u),11),a.e.c.length==0?Wi(i,a,i.c.b,i.c):Wi(t,a,t.c.b,t.c);return Bd(t).Fc(i),t}function eyt(e,t){var i,a,u;Go(e.f,t)&&(t.b=e,a=t.c,Oc(e.j,a,0)!=-1||Qe(e.j,a),u=t.d,Oc(e.j,u,0)!=-1||Qe(e.j,u),i=t.a.b,i.c.length!=0&&(!e.i&&(e.i=new vvt(e)),hTn(e.i,i)))}function UOn(e){var t,i,a,u,l;return i=e.c.d,a=i.j,u=e.d.d,l=u.j,a==l?i.p<u.p?0:1:fZ(a)==l?0:UCe(a)==l?1:(t=e.b,Cl(t.b,fZ(a))?0:1)}function UZ(e,t){var i,a,u,l,g,m;u=e,g=ij(u,"layoutOptions"),!g&&(g=ij(u,$Ct)),g&&(m=g,a=null,m&&(a=(l=Due(m,Pe(ft,Ye,2,0,6,1)),new Pae(m,l))),a&&(i=new Fat(m,t),Oa(a,i)))}function Us(e){if(me(e,206))return f(e,27);if(me(e,193))return bg(f(e,124));throw ie(e?new f0("Only support nodes and ports."):new IN(WCt))}function zOn(e,t,i,a){return(t>=0&&Mt(e.substr(t,3),"GMT")||t>=0&&Mt(e.substr(t,3),"UTC"))&&(i[0]=t+3),VIe(e,i,a)}function HOn(e,t){var i,a,u,l,g;for(l=e.g.a,g=e.g.b,a=new ne(e.d);a.a<a.c.c.length;)i=f(ee(a),72),u=i.n,u.a=l,e.i==(tt(),hn)?u.b=g+e.j.b-i.o.b:u.b=g,ii(u,t),l+=i.o.a+e.e}function tyt(e,t,i){if(e.b)throw ie(new As("The task is already done."));return e.p!=null?!1:(e.p=t,e.r=i,e.k&&(e.o=(b0(),Na(Nc(Date.now()),M0))),!0)}function dAe(e){var t,i,a,u,l,g,m;return m=new AN,i=e.Ng(),u=i!=null,u&&wP(m,Am,e.Ng()),a=e.ve(),l=a!=null,l&&wP(m,Gr,e.ve()),t=e.Mg(),g=t!=null,g&&wP(m,"description",e.Mg()),m}function nyt(e,t,i){var a,u,l;return l=e.q,e.q=t,(e.Db&4)!=0&&(e.Db&1)==0&&(u=new Jo(e,1,9,l,t),i?i.kj(u):i=u),t?(a=t.c,a!=e.r&&(i=e.Vk(a,i))):e.r&&(i=e.Vk(null,i)),i}function GOn(e,t,i){var a,u,l,g,m;for(i=(m=t,vR(m,e.e,-1-e.c,i)),g=Yke(e.a),l=(a=new yT(new Vi(g.a).a),new oY(a));l.a.b;)u=f(G3(l.a).jd(),87),i=kL(u,JZ(u,e.a),i);return i}function qOn(e,t,i){var a,u,l,g,m;for(i=(m=t,eJ(m,e.e,-1-e.c,i)),g=Yke(e.a),l=(a=new yT(new Vi(g.a).a),new oY(a));l.a.b;)u=f(G3(l.a).jd(),87),i=kL(u,JZ(u,e.a),i);return i}function VOn(e,t,i,a){var u,l,g;if(a==0)au(t,0,e,i,e.length-i);else for(g=32-a,e[e.length-1]=0,l=e.length-1;l>i;l--)e[l]|=t[l-i-1]>>>g,e[l-1]=t[l-i-1]<<a;for(u=0;u<i;u++)e[u]=0}function WOn(e){var t,i,a,u,l;for(t=0,i=0,l=e.Jc();l.Ob();)a=f(l.Pb(),115),t=h.Math.max(t,a.d.b),i=h.Math.max(i,a.d.c);for(u=e.Jc();u.Ob();)a=f(u.Pb(),115),a.d.b=t,a.d.c=i}function KOn(e){var t,i,a,u,l;for(i=0,t=0,l=e.Jc();l.Ob();)a=f(l.Pb(),115),i=h.Math.max(i,a.d.d),t=h.Math.max(t,a.d.a);for(u=e.Jc();u.Ob();)a=f(u.Pb(),115),a.d.d=i,a.d.a=t}function ode(e,t,i,a,u){var l,g;l=f($d(Jr(t.Mc(),new bv),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),vG(l),g=f(cE(e.b,i,a),15),u==0?g.ad(0,l):g.Fc(l)}function YOn(e,t,i){i.Sg("Grow Tree",1),e.b=t.f,lt(ut(V(t,(HP(),F$))))?(e.c=new _l,Wft(e,null)):e.c=new _l,e.a=!1,N2t(e,t.f),$e(t,K6e,(mn(),!!e.a)),i.Tg()}function zS(e){var t;this.d=new Wn,this.c=e.c,this.e=e.d,this.b=e.b,this.f=new Slt(e.e),this.a=e.a,e.f?this.g=e.f:this.g=(t=f(Tp($ne),10),new Rf(t,f(Gh(t,t.length),10),0))}function fAe(e){var t,i,a,u;for(t=null,u=new ne(e.Pf());u.a<u.c.c.length;)a=f(ee(u),187),i=new qh(a.Jf().a,a.Jf().b,a.Kf().a,a.Kf().b),t?jj(t,i):t=i;return!t&&(t=new BN),t}function ade(e,t,i,a){var u,l;return i==1?(!e.n&&(e.n=new Be(hc,e,1,7)),$u(e.n,t,a)):(l=f(Vt((u=f(fn(e,16),29),u||e.ei()),i),69),l.tk().wk(e,Xu(e),i-jn(e.ei()),t,a))}function sde(e,t,i){var a,u,l,g,m;for(a=i.gc(),e.Yi(e.i+a),m=e.i-t,m>0&&au(e.g,t,e.g,t+a,m),g=i.Jc(),e.i+=a,u=0;u<a;++u)l=g.Pb(),j7(e,t,e.Wi(t,l)),e.Ji(t,l),e.Ki(),++t;return a!=0}function xg(e,t,i){var a;return t!=e.q?(e.q&&(i=eJ(e.q,e,-10,i)),t&&(i=vR(t,e,-10,i)),i=nyt(e,t,i)):(e.Db&4)!=0&&(e.Db&1)==0&&(a=new Jo(e,1,9,t,t),i?i.kj(a):i=a),i}function cde(e,t,i,a){return G_e((i&Sm)==0,"flatMap does not support SUBSIZED characteristic"),G_e((i&4)==0,"flatMap does not support SORTED characteristic"),or(e),or(t),new Yht(e,t,i,a)}function XOn(e,t){Mke(t,"Cannot suppress a null exception."),uH(t!=e,"Exception can not suppress itself."),!e.i&&(e.k==null?e.k=ce(se(Upe,1),Ye,80,0,[t]):e.k[e.k.length]=t)}function JOn(e,t){var i;if(i=yst(e.b.$f(),t.b.$f()),i!=0)return i;switch(e.b.$f().g){case 1:case 2:return tu(e.b.Lf(),t.b.Lf());case 3:case 4:return tu(t.b.Lf(),e.b.Lf())}return 0}function ZOn(e){var t,i,a;for(a=e.e.c.length,e.a=iT(pr,[Ye,Pr],[54,30],15,[a,a],2),i=new ne(e.c);i.a<i.c.c.length;)t=f(ee(i),288),e.a[t.c.a][t.d.a]+=f(V(t,(Qh(),GR)),17).a}function QOn(e,t){var i,a,u,l,g;if(e==null)return null;for(g=Pe(Xf,_m,30,2*t,15,1),a=0,u=0;a<t;++a)i=e[a]>>4&15,l=e[a]&15,g[u++]=x7e[i],g[u++]=x7e[l];return bm(g,0,g.length)}function gl(e){var t,i;return e>=Xa?(t=GG+(e-Xa>>10&1023)&To,i=56320+(e-Xa&1023)&To,String.fromCharCode(t)+(""+String.fromCharCode(i))):String.fromCharCode(e&To)}function eNn(e,t){tT();var i,a,u,l;return u=f(f(si(e.r,t),22),83),u.gc()>=2?(a=f(u.Jc().Pb(),115),i=e.u.Gc((Xl(),jF)),l=e.u.Gc(u4),!a.a&&!i&&(u.gc()==2||l)):!1}function ryt(e,t,i,a,u){var l,g,m;for(l=W2t(e,t,i,a,u),m=!1;!l;)YZ(e,u,!0),m=!0,l=W2t(e,t,i,a,u);m&&YZ(e,u,!1),g=Kue(u),g.c.length!=0&&(e.d&&e.d.Eg(g),ryt(e,u,i,a,g))}function zZ(){zZ=Z,wbe=new zY("NODE_SIZE_REORDERER",0),mbe=new zY("INTERACTIVE_NODE_REORDERER",1),vbe=new zY("MIN_SIZE_PRE_PROCESSOR",2),bbe=new zY("MIN_SIZE_POST_PROCESSOR",3)}function HZ(){HZ=Z,o1e=new C7(Lp,0),K9e=new C7("DIRECTED",1),X9e=new C7("UNDIRECTED",2),V9e=new C7("ASSOCIATION",3),Y9e=new C7("GENERALIZATION",4),W9e=new C7("DEPENDENCY",5)}function tNn(e,t){var i;if(!bg(e))throw ie(new As(kCt));switch(i=bg(e),t.g){case 1:return-(e.j+e.f);case 2:return e.i-i.g;case 3:return e.j-i.f;case 4:return-(e.i+e.g)}return 0}function nNn(e,t,i){var a,u,l;return a=t.Ik(),l=t.kd(),u=a.Gk()?Yv(e,4,a,l,null,EL(e,a,l,me(a,101)&&(f(a,19).Bb&Xa)!=0),!0):Yv(e,a.qk()?2:1,a,l,a.fk(),-1,!0),i?i.kj(u):i=u,i}function dL(e,t){var i,a;for(an(t),a=e.b.c.length,Qe(e.b,t);a>0;){if(i=a,a=(a-1)/2|0,e.a.Le(ct(e.b,a),t)<=0)return af(e.b,i,t),!0;af(e.b,i,ct(e.b,a))}return af(e.b,a,t),!0}function hAe(e,t,i,a){var u,l;if(u=0,i)u=aZ(e.a[i.g][t.g],a);else for(l=0;l<Dee;l++)u=h.Math.max(u,aZ(e.a[l][t.g],a));return t==(Np(),Bu)&&e.b&&(u=h.Math.max(u,e.b.a)),u}function rNn(e,t){var i,a,u,l,g,m;return u=e.i,l=t.i,!u||!l||u.i!=l.i||u.i==(tt(),_n)||u.i==(tt(),pn)?!1:(g=u.g.a,i=g+u.j.a,m=l.g.a,a=m+l.j.a,g<=a&&i>=m)}function iyt(e){switch(e.g){case 0:return new V2;case 1:return new i5;default:throw ie(new dn("No implementation is available for the width approximator "+(e.f!=null?e.f:""+e.g)))}}function pAe(e,t,i,a){var u;if(u=!1,ta(a)&&(u=!0,wP(t,i,dr(a))),u||Xk(a)&&(u=!0,pAe(e,t,i,a)),u||me(a,241)&&(u=!0,tE(t,i,f(a,241))),!u)throw ie(new kae(U4e))}function iNn(e,t){var i,a,u;if(i=t.mi(e.a),i&&(u=Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),mh),u!=null)){for(a=1;a<(Rl(),X7e).length;++a)if(Mt(X7e[a],u))return a}return 0}function oNn(e,t){var i,a,u;if(i=t.mi(e.a),i&&(u=Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),mh),u!=null)){for(a=1;a<(Rl(),J7e).length;++a)if(Mt(J7e[a],u))return a}return 0}function oyt(e,t){var i,a,u,l;if(an(t),l=e.a.gc(),l<t.gc())for(i=e.a.ec().Jc();i.Ob();)a=i.Pb(),t.Gc(a)&&i.Qb();else for(u=t.Jc();u.Ob();)a=u.Pb(),e.a.Ac(a)!=null;return l!=e.a.gc()}function ayt(e){var t,i;switch(i=$a(uc(ce(se(Zo,1),Ye,8,0,[e.i.n,e.n,e.a]))),t=e.i.d,e.j.g){case 1:i.b-=t.d;break;case 2:i.a+=t.c;break;case 3:i.b+=t.a;break;case 4:i.a-=t.b}return i}function aNn(e){var t;for(t=(dR(),f(Dn(new Tn(Cn(Fo(e).a.Jc(),new B))),18).c.i);t.k==(cn(),uo);)$e(t,(Ge(),Pq),(mn(),!0)),t=f(Dn(new Tn(Cn(Fo(t).a.Jc(),new B))),18).c.i}function ude(e,t,i,a){var u,l,g,m;for(m=kj(t,a),g=m.Jc();g.Ob();)u=f(g.Pb(),11),e.d[u.p]=e.d[u.p]+e.c[i.p];for(m=kj(i,a),l=m.Jc();l.Ob();)u=f(l.Pb(),11),e.d[u.p]=e.d[u.p]-e.c[t.p]}function gAe(e,t,i){var a,u;for(u=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));u.e!=u.i.gc();)a=f(Mn(u),27),Of(a,a.i+t,a.j+i);Oa((!e.b&&(e.b=new Be(vo,e,12,3)),e.b),new Aat(t,i))}function sNn(e,t,i,a){var u,l;for(l=t,u=l.d==null||e.a.Le(i.d,l.d)>0?1:0;l.a[u]!=i;)l=l.a[u],u=e.a.Le(i.d,l.d)>0?1:0;l.a[u]=a,a.b=i.b,a.a[0]=i.a[0],a.a[1]=i.a[1],i.a[0]=null,i.a[1]=null}function cNn(e){var t,i,a,u;for(t=new et,i=Pe(kd,Tb,30,e.a.c.length,16,1),Lke(i,i.length),u=new ne(e.a);u.a<u.c.c.length;)a=f(ee(u),123),i[a.d]||(vn(t.c,a),Evt(e,a,i));return t}function syt(e,t){var i,a,u,l,g;for(u=t==1?pge:hge,a=u.a.ec().Jc();a.Ob();)for(i=f(a.Pb(),86),g=f(si(e.f.c,i),22).Jc();g.Ob();)l=f(g.Pb(),47),hl(e.b.b,l.b),hl(e.b.a,f(l.b,82).d)}function uNn(e,t){var i;t.Sg("Hierarchical port position processing",1),i=e.b,i.c.length>0&&DSt((Gt(0,i.c.length),f(i.c[0],25)),e),i.c.length>1&&DSt(f(ct(i,i.c.length-1),25),e),t.Tg()}function lNn(e){Xl();var t,i;return t=xi(M1,ce(se(Tne,1),Ve,277,0,[Lw])),!(GH(nJ(t,e))>1||(i=xi(jF,ce(se(Tne,1),Ve,277,0,[MF,u4])),GH(nJ(i,e))>1))}function mAe(e,t){var i;i=wu((Uv(),xh),e),me(i,491)?Os(xh,e,new Wat(this,t)):Os(xh,e,this),dde(this,t),t==(iP(),$7e)?(this.wb=f(this,1984),f(t,1986)):this.wb=(Wv(),sn)}function dNn(e){var t,i,a;if(e==null)return null;for(t=null,i=0;i<UF.length;++i)try{return Oot(UF[i],e)}catch(u){if(u=go(u),me(u,32))a=u,t=a;else throw ie(u)}throw ie(new xJ(t))}function cyt(){cyt=Z,WAt=ce(se(ft,1),Ye,2,6,["Sun","Mon","Tue","Wed","Thu","Fri","Sat"]),KAt=ce(se(ft,1),Ye,2,6,["Jan","Feb","Mar","Apr",AR,"Jun","Jul","Aug","Sep","Oct","Nov","Dec"])}function V(e,t){var i,a;return a=(!e.q&&(e.q=new Wn),bn(e.q,t)),a??(i=t.Qg(),me(i,4)&&(i==null?(!e.q&&(e.q=new Wn),aR(e.q,t)):(!e.q&&(e.q=new Wn),qr(e.q,t,i))),i)}function wa(){wa=Z,rp=new Lz("P1_CYCLE_BREAKING",0),Ab=new Lz("P2_LAYERING",1),su=new Lz("P3_NODE_ORDERING",2),cu=new Lz("P4_NODE_PLACEMENT",3),zs=new Lz("P5_EDGE_ROUTING",4)}function fNn(e,t){uj();var i;if(e.c==t.c){if(e.b==t.b||Fkn(e.b,t.b)){if(i=$gn(e.b)?1:-1,e.a&&!t.a)return i;if(!e.a&&t.a)return-i}return tu(e.b.g,t.b.g)}else return di(e.c,t.c)}function hNn(e,t,i,a){var u,l,g,m,v;return g=mwt(e.a,t,i),m=f(g.a,17).a,l=f(g.b,17).a,a&&(v=f(V(t,(Ge(),ed)),9),u=f(V(i,ed),9),v&&u&&(Qht(e.b,v,u),m+=e.b.i,l+=e.b.e)),m>l}function uyt(e,t){var i,a,u;if(vAe(e,t))return!0;for(a=new ne(t);a.a<a.c.c.length;)if(i=f(ee(a),27),u=Fwt(i),oQ(e,i,u)||i0t(e,i)-e.g<=e.a)return!0;return!1}function mG(){mG=Z,gne=(Nde(),B8e),jbe=ELt,Mbe=yLt,P8e=bLt,Dbe=wLt,R8e=new gS(8),lLt=new pa((Rr(),V0),R8e),dLt=new pa(zE,8),fLt=$8e,O8e=hLt,N8e=pLt,uLt=new pa(nV,(mn(),!1))}function GZ(){GZ=Z,r9e=new gS(15),$Lt=new pa((Rr(),V0),r9e),FLt=new pa(zE,15),i9e=new pa(aV,rt(0)),e9e=VLt,MLt=UE,jLt=Ow,Q8e=new pa(ID,tCt),t9e=iV,n9e=GA,Ube=HLt,DLt=tV}function pNn(e,t){if(me(t,206))return x3n(e,f(t,27));if(me(t,193))return I3n(e,f(t,124));if(t)return null;throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[t])))))}function gNn(e,t){if(me(t,359))return m_n(e,f(t,153));if(me(t,268))return iMn(e,f(t,85));if(t)return null;throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[t])))))}function lde(e){if((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b).i!=1||(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c).i!=1)throw ie(new dn(YCt));return Us(f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84))}function lyt(e){if((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b).i!=1||(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c).i!=1)throw ie(new dn(YCt));return Us(f(be((!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c),0),84))}function bAe(e,t,i){var a,u,l;if(++e.j,u=e.Bj(),t>=u||t<0)throw ie(new Ru(ppe+t+TE+u));if(i>=u||i<0)throw ie(new Ru(gpe+i+TE+u));return t!=i?a=(l=e.zj(i),e.nj(t,l),l):a=e.uj(i),a}function dyt(e){var t,i,a;if(a=e,e)for(t=0,i=e.Ah();i;i=i.Ah()){if(++t>Lfe)return dyt(i);if(a=i,i==e)throw ie(new As("There is a cycle in the containment hierarchy of "+e))}return a}function Cg(e){var t,i,a;for(a=new uE(Fu,"[","]"),i=e.Jc();i.Ob();)t=i.Pb(),b1(a,Ae(t)===Ae(e)?"(this Collection)":t==null?bl:Qs(t));return a.a?a.e.length==0?a.a.a:a.a.a+(""+a.e):a.c}function vAe(e,t){var i,a;if(a=!1,t.gc()<2)return!1;for(i=0;i<t.gc();i++)i<t.gc()-1?a=a|oQ(e,f(t.Xb(i),27),f(t.Xb(i+1),27)):a=a|oQ(e,f(t.Xb(i),27),f(t.Xb(0),27));return a}function fyt(e,t){var i;t!=e.a?(i=null,e.a&&(i=f(e.a,52).Ph(e,4,Bp,i)),t&&(i=f(t,52).Nh(e,4,Bp,i)),i=NCe(e,t,i),i&&i.lj()):(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,1,t,t))}function wAe(e,t){var i;t!=e.e?(e.e&&Wpt(Yke(e.e),e),t&&(!t.b&&(t.b=new iY(new bae)),uut(t.b,e)),i=TIn(e,t,null),i&&i.lj()):(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,4,t,t))}function mNn(e){var t,i,a,u;if(k3(f(V(e.b,(ot(),mf)),86)))return 0;for(t=0,a=new ne(e.a);a.a<a.c.c.length;)i=f(ee(a),9),i.k==(cn(),Yi)&&(u=i.o.a,t=h.Math.max(t,u));return t}function bNn(e,t){var i;i=t.o,l1(e.f)?(e.j.a=h.Math.max(e.j.a,i.a),e.j.b+=i.b,e.d.c.length>1&&(e.j.b+=e.e)):(e.j.a+=i.a,e.j.b=h.Math.max(e.j.b,i.b),e.d.c.length>1&&(e.j.a+=e.e))}function cw(){cw=Z,hNt=ce(se(Ja,1),qc,64,0,[(tt(),hn),_n,Bn]),fNt=ce(se(Ja,1),qc,64,0,[_n,Bn,pn]),pNt=ce(se(Ja,1),qc,64,0,[Bn,pn,hn]),gNt=ce(se(Ja,1),qc,64,0,[pn,hn,_n])}function vNn(e,t,i,a){var u,l,g,m,v,S,T;if(g=e.c.d,m=e.d.d,g.j!=m.j)for(T=e.b,u=g.j,v=null;u!=m.j;)v=t==0?fZ(u):UCe(u),l=w3e(u,T.d[u.g],i),S=w3e(v,T.d[v.g],i),Mr(a,ii(l,S)),u=v}function hyt(e){var t,i,a,u,l,g,m,v,S;for(this.a=jvt(e),this.b=new et,i=e,a=0,u=i.length;a<u;++a)for(t=i[a],l=new et,Qe(this.b,l),m=t,v=0,S=m.length;v<S;++v)g=m[v],Qe(l,new Kl(g.j))}function wNn(e,t,i){var a,u,l;return l=0,a=i[t],t<i.length-1&&(u=i[t+1],e.b[t]?(l=j7n(e.d,a,u),l+=Tce(e.a,a,(tt(),_n)),l+=Tce(e.a,u,pn)):l=k_n(e.a,a,u)),e.c[t]&&(l+=qSn(e.a,a)),l}function yNn(e,t,i,a,u){var l,g,m,v;for(v=null,m=new ne(a);m.a<m.c.c.length;)if(g=f(ee(m),441),g!=i&&Oc(g.e,u,0)!=-1){v=g;break}l=cue(u),Ia(l,i.b),ua(l,v.b),jt(e.a,u,new vX(l,t,i.f))}function pyt(e){for(;e.g.c!=0&&e.d.c!=0;)$se(e.g).c>$se(e.d).c?(e.i+=e.g.c,xle(e.d)):$se(e.d).c>$se(e.g).c?(e.e+=e.d.c,xle(e.g)):(e.i+=wlt(e.g),e.e+=wlt(e.d),xle(e.g),xle(e.d))}function ENn(e,t,i){var a,u,l,g;for(l=t.q,g=t.r,new nE((Ip(),Cw),t,l,1),new nE(Cw,l,g,1),u=new ne(i);u.a<u.c.c.length;)a=f(ee(u),116),a!=l&&a!=t&&a!=g&&(cOe(e.a,a,t),cOe(e.a,a,g))}function gyt(e,t,i,a){e.a.d=h.Math.min(t,i),e.a.a=h.Math.max(t,a)-e.a.d,t<i?(e.b=.5*(t+i),e.g=jhe*e.b+.9*t,e.f=jhe*e.b+.9*i):(e.b=.5*(t+a),e.g=jhe*e.b+.9*a,e.f=jhe*e.b+.9*t)}function SNn(e){var t,i,a,u;if(e.b!=0){for(t=new ki,u=tr(e,0);u.b!=u.d.c;)a=f(Zn(u),39),Ba(t,Wue(a)),i=a.e,i.a=f(V(a,(_i(),Vq)),17).a,i.b=f(V(a,Wq),17).a;return t}return new ki}function _Nn(e){switch(f(V(e,(ot(),el)),171).g){case 1:$e(e,el,(Ff(),J$));break;case 2:$e(e,el,(Ff(),tx));break;case 3:$e(e,el,(Ff(),X$));break;case 4:$e(e,el,(Ff(),Tw))}}function kNn(e,t,i){var a;i.Sg("Self-Loop routing",1),a=oCn(t),QY(V(t,(bJ(),TF))),Zi(Lu(Jr(Jr(Zs(new Ft(null,new Bt(t.b,16)),new C6),new g8),new m8),new A6),new qot(e,a)),i.Tg()}function fL(){fL=Z,q$=new k7(Lp,0),APe=new k7(NR,1),NPe=new k7(RR,2),OPe=new k7("LEFT_RIGHT_CONSTRAINT_LOCKING",3),IPe=new k7("LEFT_RIGHT_CONNECTION_LOCKING",4),CPe=new k7(hxt,5)}function myt(e,t,i){var a,u,l,g,m,v,S;m=i.a/2,l=i.b/2,a=h.Math.abs(t.a-e.a),u=h.Math.abs(t.b-e.b),v=1,S=1,a>m&&(v=m/a),u>l&&(S=l/u),g=h.Math.min(v,S),e.a+=g*(t.a-e.a),e.b+=g*(t.b-e.b)}function TNn(e,t,i,a,u){var l,g;for(g=!1,l=f(ct(i.b,0),27);CDn(e,t,l,a,u)&&(g=!0,cOn(i,l),i.b.c.length!=0);)l=f(ct(i.b,0),27);return i.b.c.length==0&&fG(i.j,i),g&&PZ(t.q),g}function yAe(e,t,i,a){var u,l;return i==0?(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),pX(e.o,t,a)):(l=f(Vt((u=f(fn(e,16),29),u||e.ei()),i),69),l.tk().xk(e,Xu(e),i-jn(e.ei()),t,a))}function dde(e,t){var i;t!=e.sb?(i=null,e.sb&&(i=f(e.sb,52).Ph(e,1,BF,i)),t&&(i=f(t,52).Nh(e,1,BF,i)),i=RCe(e,t,i),i&&i.lj()):(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,4,t,t))}function byt(e,t){var i,a,u,l;if(t)u=E1(t,"x"),i=new cnt(e),U3(i.a,(an(u),u)),l=E1(t,"y"),a=new unt(e),z3(a.a,(an(l),l));else throw ie(new am("All edge sections need an end point."))}function vyt(e,t){var i,a,u,l;if(t)u=E1(t,"x"),i=new ont(e),B3(i.a,(an(u),u)),l=E1(t,"y"),a=new ant(e),H3(a.a,(an(l),l));else throw ie(new am("All edge sections need a start point."))}function xNn(e,t){var i,a,u,l,g,m,v;for(a=j1t(e),l=0,m=a.length;l<m;++l)Eyt(t);for(v=!Lg&&e.e?Lg?null:e.d:null;v;){for(i=j1t(v),u=0,g=i.length;u<g;++u)Eyt(t);v=!Lg&&v.e?Lg?null:v.d:null}}function wyt(e,t){var i,a;a=f(V(t,(ot(),Xi)),100),$e(t,(Ge(),qPe),a),i=t.e,i&&(Zi(new Ft(null,new Bt(i.a,16)),new Hy(e)),Zi(Zs(new Ft(null,new Bt(i.b,16)),new sy),new wz(e)))}function CNn(e){var t,i,a,u;if(t=!1,li(e,(Ge(),K$)))for(i=f(V(e,K$),91),u=new ne(e.j);u.a<u.c.c.length;)a=f(ee(u),11),_6n(a)&&(t||(zRn(ra(e)),t=!0),JCn(f(i.xc(a),315)))}function ANn(e){var t,i,a,u,l,g,m,v,S;return S=dAe(e),i=e.e,l=i!=null,l&&wP(S,cee,e.e),m=e.k,g=!!m,g&&wP(S,"type",Lse(e.k)),a=Az(e.j),u=!a,u&&(v=new oh,Wh(S,fpe,v),t=new Ant(v),Oa(e.j,t)),S}function INn(e){var t,i,a,u;for(u=Qy((uf(e.gc(),"size"),new RN),123),a=!0,i=Yy(e).Jc();i.Ob();)t=f(i.Pb(),43),a||(u.a+=Fu),a=!1,vu(Qy(vu(u,t.jd()),61),t.kd());return(u.a+="}",u).a}function yyt(e,t){var i,a,u;return t&=63,t<22?(i=e.l<<t,a=e.m<<t|e.l>>22-t,u=e.h<<t|e.m>>22-t):t<44?(i=0,a=e.l<<t-22,u=e.m<<t-22|e.l>>44-t):(i=0,a=0,u=e.l<<t-44),Ku(i&Ed,a&Ed,u&A1)}function AT(e){if(y6e==null&&(y6e=new RegExp("^\\s*[+-]?(NaN|Infinity|((\\d+\\.?\\d*)|(\\.\\d+))([eE][+-]?\\d+)?[dDfF]?)\\s*$")),!y6e.test(e))throw ie(new sm(XS+e+'"'));return parseFloat(e)}function Eyt(e){var t,i,a;t=Mt(typeof console,Ufe)?null:new zl,t&&(cP(),i=(a=900,a>=M0?"error":a>=900?"warn":a>=800?"info":"log"),pdt(i,e.a),e.b&&EIe(t,i,e.b,"Exception: ",!0))}function Syt(e,t){var i,a,u,l,g;for(u=t==1?pge:hge,a=u.a.ec().Jc();a.Ob();)for(i=f(a.Pb(),86),g=f(si(e.f.c,i),22).Jc();g.Ob();)l=f(g.Pb(),47),Qe(e.b.b,f(l.b,82)),Qe(e.b.a,f(l.b,82).d)}function ONn(e,t){var i,a,u,l;for(l=t.b.j,e.a=Pe(pr,Pr,30,l.c.length,15,1),u=0,a=0;a<l.c.length;a++)i=(Gt(a,l.c.length),f(l.c[a],11)),i.e.c.length==0&&i.g.c.length==0?u+=1:u+=3,e.a[a]=u}function qZ(){qZ=Z,Uge=new S7("ALWAYS_UP",0),Bge=new S7("ALWAYS_DOWN",1),Hge=new S7("DIRECTION_UP",2),zge=new S7("DIRECTION_DOWN",3),Gge=new S7("SMART_UP",4),pte=new S7("SMART_DOWN",5)}function NNn(e,t){if(e<0||t<0)throw ie(new dn("k and n must be positive"));if(t>e)throw ie(new dn("k must be smaller than n"));return t==0||t==e?1:e==0?0:W3e(e)/(W3e(t)*W3e(e-t))}function EAe(e,t){var i,a,u,l;for(i=new _se(e);i.g==null&&!i.c?pTe(i):i.g==null||i.i!=0&&f(i.g[i.i-1],49).Ob();)if(l=f(XZ(i),57),me(l,173))for(a=f(l,173),u=0;u<t.length;u++)t[u].Ig(a)}function fde(e){var t;return(e.Db&64)!=0?mle(e):(t=new sh(mle(e)),t.a+=" (height: ",v3(t,e.f),t.a+=", width: ",v3(t,e.g),t.a+=", x: ",v3(t,e.i),t.a+=", y: ",v3(t,e.j),t.a+=")",t.a)}function RNn(e){var t,i,a,u,l,g,m;for(t=new Zv,a=e,u=0,l=a.length;u<l;++u)if(i=a[u],g=or(i.jd()),m=pE(t,g,or(i.kd())),m!=null)throw ie(new dn("duplicate key: "+g));this.b=(Ht(),new vz(t))}function PNn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],b1(l,String.fromCharCode(t));return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function SAe(){SAe=Z,Y6e=(OJ(),jee),VIt=new Pt(CQ,Y6e),rt(1),qIt=new Pt(ZOe,rt(300)),rt(0),YIt=new Pt(QOe,rt(0)),XIt=new Pt(Vfe,Tm),WIt=new Pt(Wfe,5),JIt=jee,KIt=age}function LNn(e,t){var i;if(t!=null&&!e.c.Ek().ck(t))throw i=me(t,57)?f(t,57).zh().zb:Vy(Fd(t)),ie(new eP(vw+e.c.ve()+"'s type '"+e.c.Ek().ve()+"' does not permit a value of type '"+i+"'"))}function DNn(e,t,i){var a,u;for(u=new la(e.b,0);u.b<u.d.gc();)a=(Ln(u.b<u.d.gc()),f(u.d.Xb(u.c=u.b++),72)),Ae(V(a,(Ge(),HPe)))===Ae(t)&&(IAe(a.n,ra(e.c.i),i),hd(u),Qe(t.b,a))}function _yt(e){var t,i;return i=h.Math.sqrt((e.k==null&&(e.k=Kxe(e,new DU)),ke(e.k)/(e.b*(e.g==null&&(e.g=Pbt(e,new _8)),ke(e.g))))),t=mr(Nc(h.Math.round(i))),t=h.Math.min(t,e.f),t}function MNn(){var e,t,i;for(t=0,e=0;e<1;e++){if(i=JAe((gn(e,1),"X".charCodeAt(e))),i==0)throw ie(new vr((gn(e,2),"Unknown Option: "+"X".substr(e))));t|=i}return t}function jNn(e){var t,i,a,u,l,g;for(t=new O3,i=new O3,Jv(t,e),Jv(i,e);i.b!=i.c;)for(u=f(JN(i),36),g=new ne(u.a);g.a<g.c.c.length;)l=f(ee(g),9),l.e&&(a=l.e,Jv(t,a),Jv(i,a));return t}function ou(){Nl(),R_e.call(this),this.j=(tt(),fc),this.a=new ha,new hae,this.f=(uf(2,$T),new Du(2)),this.e=(uf(4,$T),new Du(4)),this.g=(uf(4,$T),new Du(4)),this.b=new Got(this.e,this.g)}function $Nn(e,t){var i,a;return!(lt(ut(V(t,(Ge(),U0))))||(a=t.c.i,e==(Ff(),X$)&&a.k==(cn(),Vc))||(i=f(V(a,(ot(),el)),171),i==Tw))}function FNn(e,t){var i,a;return!(lt(ut(V(t,(Ge(),U0))))||(a=t.d.i,e==(Ff(),J$)&&a.k==(cn(),Vc))||(i=f(V(a,(ot(),el)),171),i==tx))}function BNn(e,t){var i,a,u,l,g,m,v;for(g=e.d,v=e.o,m=new qh(-g.b,-g.d,g.b+v.a+g.c,g.d+v.b+g.a),a=t,u=0,l=a.length;u<l;++u)i=a[u],i&&jj(m,i.i);g.b=-m.c,g.d=-m.d,g.c=m.b-g.b-v.a,g.a=m.a-g.d-v.b}function UNn(e,t){if(t.a)switch(f(V(t.b,(Ge(),qPe)),100).g){case 0:case 1:aOn(t);case 2:Zi(new Ft(null,new Bt(t.d,16)),new mv),N4n(e.a,t)}else Zi(new Ft(null,new Bt(t.d,16)),new mv)}function bG(){bG=Z,x8e=new zz("CENTER_DISTANCE",0),Pbe=new zz("CIRCLE_UNDERLAP",1),A8e=new zz("RECTANGLE_UNDERLAP",2),Lbe=new zz("INVERTED_OVERLAP",3),C8e=new zz("MINIMUM_ROOT_DISTANCE",4)}function zNn(e){AIe();var t,i,a,u,l;if(e==null)return null;for(a=e.length,u=a*2,t=Pe(Xf,_m,30,u,15,1),i=0;i<a;i++)l=e[i],l<0&&(l+=256),t[i*2]=Vne[l>>4],t[i*2+1]=Vne[l&15];return bm(t,0,t.length)}function HNn(e){VX();var t,i,a;switch(a=e.c.length,a){case 0:return AAt;case 1:return t=f(oEt(new ne(e)),43),vvn(t.jd(),t.kd());default:return i=f(kg(e,Pe(AE,yQ,43,e.c.length,0,1)),174),new A2e(i)}}function L0(e,t){switch(t.g){case 1:return WN(e.j,(Nl(),y5e));case 2:return WN(e.j,(Nl(),v5e));case 3:return WN(e.j,(Nl(),S5e));case 4:return WN(e.j,(Nl(),_5e));default:return Ht(),Ht(),Ga}}function GNn(e,t){var i,a,u;i=mvn(t,e.e),a=f(bn(e.g.f,i),17).a,u=e.a.c.length-1,e.a.c.length!=0&&f(ct(e.a,u),292).c==a?(++f(ct(e.a,u),292).a,++f(ct(e.a,u),292).b):Qe(e.a,new Tct(a))}function qNn(e,t,i){var a,u;return a=HPn(e,t,i),a!=0?a:li(t,(Ge(),hi))&&li(i,hi)?(u=tu(f(V(t,hi),17).a,f(V(i,hi),17).a),u<0?xG(e,t,i):u>0&&xG(e,i,t),u):S4n(e,t,i)}function uw(){uw=Z,EPt=(Rr(),o4),SPt=zE,bPt=UE,vPt=GA,wPt=Ow,mPt=HA,jMe=oV,yPt=px,lbe=(FIe(),oPt),dbe=aPt,FMe=lPt,fbe=hPt,BMe=dPt,UMe=fPt,$Me=sPt,one=cPt,ane=uPt,Kq=pPt,zMe=gPt,MMe=iPt}function kyt(e,t){var i,a,u,l,g;if(e.e<=t||t2n(e,e.g,t))return e.g;for(l=e.r,a=e.g,g=e.r,u=(l-a)/2+a;a+1<l;)i=Zj(e,u,!1),i.b<=u&&i.a<=t?(g=u,l=u):a=u,u=(l-a)/2+a;return g}function VNn(e,t,i){var a;a=Y2t(e,t,!0),tyt(i,"Recursive Graph Layout",a),EAe(t,ce(se(K8e,1),Jt,521,0,[new c3])),mm(t,(Rr(),VA))||EAe(t,ce(se(K8e,1),Jt,521,0,[new a5])),COe(e,t,null,i),Tyt(i)}function Tyt(e){var t;if(e.p==null)throw ie(new As("The task has not begun yet."));e.b||(e.k&&(t=(b0(),Na(Nc(Date.now()),M0)),e.q=oE(hh(t,e.o))*1e-9),e.c<e.r&&Hxe(e,e.r-e.c),e.b=!0)}function WNn(e){var t,i,a;for(a=new fd,Mr(a,new Ke(e.j,e.k)),i=new Nn((!e.a&&(e.a=new So(yf,e,5)),e.a));i.e!=i.i.gc();)t=f(Mn(i),369),Mr(a,new Ke(t.a,t.b));return Mr(a,new Ke(e.b,e.c)),a}function xyt(e,t,i,a,u){var l,g,m,v,S,T;if(u)for(v=u.a.length,l=new Xy(v),T=(l.b-l.a)*l.c<0?(zv(),Fw):new Vv(l);T.Ob();)S=f(T.Pb(),17),m=OP(u,S.a),g=new Ydt(e,t,i,a),zSt(g.a,g.b,g.c,g.d,m)}function _Ae(e,t){var i;if(Ae(e)===Ae(t))return!0;if(me(t,22)){i=f(t,22);try{return e.gc()==i.gc()&&e.Hc(i)}catch(a){if(a=go(a),me(a,169)||me(a,211))return!1;throw ie(a)}}return!1}function hde(e,t,i,a,u,l){switch(this.c=e,t.g){case 2:if(e.a.Le(u,i)<0)throw ie(new dn(UOe+u+bTt+i));break;case 1:e.a.Le(u,u);break;case 3:e.a.Le(i,i)}this.f=t,this.b=i,this.a=a,this.e=u,this.d=l}function kAe(e,t){var i;Qe(e.d,t),i=t.Kf(),e.c?(e.e.a=h.Math.max(e.e.a,i.a),e.e.b+=i.b,e.d.c.length>1&&(e.e.b+=e.a)):(e.e.a+=i.a,e.e.b=h.Math.max(e.e.b,i.b),e.d.c.length>1&&(e.e.a+=e.a))}function KNn(e){var t,i,a,u;switch(u=e.i,t=u.b,a=u.j,i=u.g,u.a.g){case 0:i.a=(e.g.b.o.a-a.a)/2;break;case 1:i.a=t.d.n.a+t.d.a.a;break;case 2:i.a=t.d.n.a+t.d.a.a-a.a;break;case 3:i.b=t.d.n.b+t.d.a.b}}function YNn(e,t,i){var a,u,l;for(u=new Tn(Cn(T1(i).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),!Fa(a)&&!(!Fa(a)&&a.c.i.c==a.d.i.c)&&(l=w2t(e,a,i,new mrt),l.c.length>1&&vn(t.c,l))}function Cyt(e,t,i,a,u){if(a<t||u<i)throw ie(new dn("The highx must be bigger then lowx and the highy must be bigger then lowy"));return e.a<t?e.a=t:e.a>a&&(e.a=a),e.b<i?e.b=i:e.b>u&&(e.b=u),e}function XNn(e){if(me(e,149))return f5n(f(e,149));if(me(e,232))return CCn(f(e,232));if(me(e,21))return ANn(f(e,21));throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[e])))))}function JNn(e,t,i,a,u){var l,g,m;for(l=!0,g=0;g<a;g++)l=l&i[g]==0;if(u==0)au(i,a,e,0,t),g=t;else{for(m=32-u,l=l&i[g]<<m==0,g=0;g<t-1;g++)e[g]=i[g+a]>>>u|i[g+a+1]<<m;e[g]=i[g+a]>>>u,++g}return l}function TAe(e,t,i,a){var u,l,g;if(t.k==(cn(),uo)){for(l=new Tn(Cn(Fo(t).a.Jc(),new B));Kn(l);)if(u=f(Dn(l),18),g=u.c.i.k,g==uo&&e.c.a[u.c.i.c.p]==a&&e.c.a[t.c.p]==i)return!0}return!1}function ZNn(e,t){var i,a,u,l;return t&=63,i=e.h&A1,t<22?(l=i>>>t,u=e.m>>t|i<<22-t,a=e.l>>t|e.m<<22-t):t<44?(l=0,u=i>>>t-22,a=e.m>>t-22|e.h<<44-t):(l=0,u=0,a=i>>>t-44),Ku(a&Ed,u&Ed,l&A1)}function Ayt(e,t,i,a){var u;this.b=a,this.e=e==(PS(),uF),u=t[i],this.d=iT(kd,[Ye,Tb],[183,30],16,[u.length,u.length],2),this.a=iT(pr,[Ye,Pr],[54,30],15,[u.length,u.length],2),this.c=new aAe(t,i)}function QNn(e){var t,i,a;for(e.k=new wTe((tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])).length,e.j.c.length),a=new ne(e.j);a.a<a.c.c.length;)i=f(ee(a),112),t=i.d.j,jt(e.k,t,i);e.e=Q6n(YN(e.k))}function Iyt(e,t){var i,a,u;Go(e.d,t),i=new F6,qr(e.c,t,i),i.f=lle(t.c),i.a=lle(t.d),i.d=(qj(),u=t.c.i.k,u==(cn(),Yi)||u==Dg),i.e=(a=t.d.i.k,a==Yi||a==Dg),i.b=t.c.j==(tt(),pn),i.c=t.d.j==_n}function eRn(e){var t,i,a,u,l;for(l=Qr,u=Qr,a=new ne(V3(e));a.a<a.c.c.length;)i=f(ee(a),216),t=i.e.e-i.d.e,i.e==e&&t<u?u=t:t<l&&(l=t);return u==Qr&&(u=-1),l==Qr&&(l=-1),new Ya(rt(u),rt(l))}function tRn(e,t){var i,a,u,l;for(i=t.a.o.a,l=new Xv(ra(t.a).b,t.c,t.f+1),u=new Fk(l);u.b<u.d.gc();)if(a=(Ln(u.b<u.d.gc()),f(u.d.Xb(u.c=u.b++),25)),a.c.a>=i)return hL(e,t,a.p),!0;return!1}function Q3(e,t,i,a){var u,l,g,m,v,S;for(g=i.length,l=0,u=-1,S=bmt((gn(t,e.length+1),e.substr(t)),(qse(),$6e)),m=0;m<g;++m)v=i[m].length,v>l&&swn(S,bmt(i[m],$6e))&&(u=m,l=v);return u>=0&&(a[0]=t+l),u}function nRn(e,t,i){var a,u,l,g,m,v,S,T;l=e.d.p,m=l.e,v=l.r,e.g=new oH(v),g=e.d.o.c.p,a=g>0?m[g-1]:Pe(Ib,j0,9,0,0,1),u=m[g],S=g<m.length-1?m[g+1]:Pe(Ib,j0,9,0,0,1),T=t==i-1,T?Hce(e.g,u,S):Hce(e.g,a,u)}function Oyt(e){var t;return(e.Db&64)!=0?fde(e):(t=new ef(R4e),!e.a||xr(xr((t.a+=' "',t),e.a),'"'),xr(lS(xr(lS(xr(lS(xr(lS((t.a+=" (",t),e.i),","),e.j)," | "),e.g),","),e.f),")"),t.a)}function Nyt(e,t,i){var a,u,l,g,m;for(m=Ju(e.e.zh(),t),u=f(e.g,122),a=0,g=0;g<e.i;++g)if(l=u[g],m.Zl(l.Ik())){if(a==i)return PT(e,g),is(),f(t,69).uk()?l:l.kd();++a}throw ie(new Ru(T$+i+TE+a))}function Ryt(e){var t,i,a;if(t=e.c,t==2||t==7||t==1)return Xr(),Xr(),XF;for(a=_Oe(e),i=null;(t=e.c)!=2&&t!=7&&t!=1;)i||(i=(Xr(),Xr(),new D7(1)),vE(i,a),a=i),vE(i,_Oe(e));return a}function rRn(e,t,i){return e<0||e>i?DAe(e,i,"start index"):t<0||t>i?DAe(t,i,"end index"):Vj("end index (%s) must not be less than start index (%s)",ce(se(Ro,1),Jt,1,5,[rt(t),rt(e)]))}function Pyt(e,t){var i,a,u,l;for(a=0,u=e.length;a<u;a++){l=e[a];try{l[1]?l[0].Rm()&&(t=Gvn(t,l)):l[0].Rm()}catch(g){if(g=go(g),me(g,80))i=g,hY(),vyn(me(i,472)?f(i,472).ie():i);else throw ie(g)}}return t}function hL(e,t,i){var a,u,l;for(i!=t.c+t.b.gc()&&XDn(t.a,aTn(t,i-t.c)),l=t.a.c.p,e.a[l]=h.Math.max(e.a[l],t.a.o.a),u=f(V(t.a,(Ge(),Y$)),15).Jc();u.Ob();)a=f(u.Pb(),72),$e(a,_ge,(mn(),!0))}function iRn(e,t){var i,a,u;u=y5n(t),$e(t,(Ge(),eme),u),u&&(a=Qr,Is(e.f,u)&&(a=f(ac(Is(e.f,u)),17).a),i=f(ct(t.g,0),18),lt(ut(V(i,U0)))||qr(e,u,rt(h.Math.min(f(V(i,hi),17).a,a))))}function oRn(e){var t,i,a,u;for(i=Qr,u=!0,t=0;t<e.a.c.length;t++)li(f(ct(e.a,t),9),(ot(),gD))&&(u=!1,a=f(V(f(ct(e.a,t),9),gD),17).a,i=i<a?i:a);return u&&(i=f(st((ot(),gD)),17).a),i}function Lyt(e,t,i){var a,u,l,g,m;for(t.p=-1,m=$S(t,(cs(),Uu)).Jc();m.Ob();)for(g=f(m.Pb(),11),u=new ne(g.g);u.a<u.c.c.length;)a=f(ee(u),18),l=a.d.i,t!=l&&(l.p<0?i.Ec(a):l.p>0&&Lyt(e,l,i));t.p=0}function aRn(e){var t,i,a,u;for(t=Qy(xr(new ef("Predicates."),"and"),40),i=!0,u=new Fk(e);u.b<u.d.gc();)a=(Ln(u.b<u.d.gc()),u.d.Xb(u.c=u.b++)),i||(t.a+=","),t.a+=""+a,i=!1;return(t.a+=")",t).a}function Dyt(e,t,i){var a,u,l;if(!(i<=t+2))for(u=(i-t)/2|0,a=0;a<u;++a)l=(Gt(t+a,e.c.length),f(e.c[t+a],11)),af(e,t+a,(Gt(i-a-1,e.c.length),f(e.c[i-a-1],11))),Gt(i-a-1,e.c.length),e.c[i-a-1]=l}function Myt(e,t){var i,a,u;if(t.c.length!=0){for(i=uyt(e,t),u=!1;!i;)YZ(e,t,!0),u=!0,i=uyt(e,t);u&&YZ(e,t,!1),a=Kue(t),e.b&&e.b.Eg(a),e.a=i0t(e,(Gt(0,t.c.length),f(t.c[0],27))),Myt(e,a)}}function pde(e,t){var i,a,u;if(a=Vt(e.zh(),t),i=t-e.fi(),i<0)if(a)if(a.ok())u=e.Eh(a),u>=0?e.gi(u):FAe(e,a);else throw ie(new dn(vw+a.ve()+S$));else throw ie(new dn(DCt+t+MCt));else jf(e,i,a)}function xAe(e){var t,i;if(i=null,t=!1,me(e,210)&&(t=!0,i=f(e,210).a),t||me(e,263)&&(t=!0,i=""+f(e,263).a),t||me(e,477)&&(t=!0,i=""+f(e,477).a),!t)throw ie(new kae(U4e));return i}function CAe(e,t,i){var a,u,l,g,m,v;for(v=Ju(e.e.zh(),t),a=0,m=e.i,u=f(e.g,122),g=0;g<e.i;++g)if(l=u[g],v.Zl(l.Ik())){if(i==a)return g;++a,m=g+1}if(i==a)return m;throw ie(new Ru(T$+i+TE+a))}function sRn(e,t){var i,a,u,l;if(e.f.c.length==0)return null;for(l=new BN,a=new ne(e.f);a.a<a.c.c.length;)i=f(ee(a),72),u=i.o,l.b=h.Math.max(l.b,u.a),l.a+=u.b;return l.a+=(e.f.c.length-1)*t,l}function AAe(e){var t,i,a,u;for(e.e=0,u=tr(e.f,0);u.b!=u.d.c;)a=f(Zn(u),9),a.p>=e.d.b.c.length&&(t=new ru(e.d),t.p=a.p-1,Qe(e.d.b,t),i=new ru(e.d),i.p=a.p,Qe(e.d.b,i)),jo(a,f(ct(e.d.b,a.p),25))}function cRn(e){var t,i,a,u;for(i=new ki,Ba(i,e.o),a=new v2e;i.b!=0;)t=f(i.b==0?null:(Ln(i.b!=0),lh(i,i.a.a)),499),u=Skt(e,t,!0),u&&Qe(a.a,t);for(;a.a.c.length!=0;)t=f(G1t(a),499),Skt(e,t,!1)}function bt(e){var t;this.c=new ki,this.f=e.e,this.e=e.d,this.i=e.g,this.d=e.c,this.b=e.b,this.k=e.j,this.a=e.a,e.i?this.j=e.i:this.j=(t=f(Tp(Ob),10),new Rf(t,f(Gh(t,t.length),10),0)),this.g=e.f}function bE(){bE=Z,Y8e=new jN(s$,0),Po=new jN("BOOLEAN",1),Hs=new jN("INT",2),i4=new jN("STRING",3),ya=new jN("DOUBLE",4),Li=new jN("ENUM",5),r4=new jN("ENUMSET",6),Nb=new jN("OBJECT",7)}function jj(e,t){var i,a,u,l,g;a=h.Math.min(e.c,t.c),l=h.Math.min(e.d,t.d),u=h.Math.max(e.c+e.b,t.c+t.b),g=h.Math.max(e.d+e.a,t.d+t.a),u<a&&(i=a,a=u,u=i),g<l&&(i=l,l=g,g=i),Yut(e,a,l,u-a,g-l)}function jyt(e,t){var i,a;if(e.f){for(;t.Ob();)if(i=f(t.Pb(),74),a=i.Ik(),me(a,101)&&(f(a,19).Bb&Dc)!=0&&(!e.e||a.mk()!=$D||a.Ij()!=0)&&i.kd()!=null)return t.Ub(),!0;return!1}else return t.Ob()}function $yt(e,t){var i,a;if(e.f){for(;t.Sb();)if(i=f(t.Ub(),74),a=i.Ik(),me(a,101)&&(f(a,19).Bb&Dc)!=0&&(!e.e||a.mk()!=$D||a.Ij()!=0)&&i.kd()!=null)return t.Pb(),!0;return!1}else return t.Sb()}function Rl(){Rl=Z,J7e=ce(se(ft,1),Ye,2,6,[t6e,hq,yee,Q3t,Eee,_pe,cee]),X7e=ce(se(ft,1),Ye,2,6,[t6e,"empty",hq,fq,"elementOnly"]),Z7e=ce(se(ft,1),Ye,2,6,[t6e,"preserve","replace",Mp]),Ea=new Zut}function IAe(e,t,i){var a,u,l;if(t!=i){a=t;do ii(e,a.c),u=a.e,u&&(l=a.d,eT(e,l.b,l.d),ii(e,u.n),a=ra(u));while(u);a=i;do Ho(e,a.c),u=a.e,u&&(l=a.d,Nct(e,l.b,l.d),Ho(e,u.n),a=ra(u));while(u)}}function gde(e,t,i,a){var u,l,g,m,v;if(a.f.c+a.i.c==0)for(g=e.a[e.c],m=0,v=g.length;m<v;++m)l=g[m],qr(a,l,new Ubt(e,l,i));return u=f(ac(Is(a.f,t)),662),u.b=0,u.c=u.f,u.c==0||Ot(f(ct(u.a,u.b),292)),u}function Fyt(e){var t;this.j=new et,this.f=new Mo,this.b=(t=f(Tp(Ja),10),new Rf(t,f(Gh(t,t.length),10),0)),this.d=Pe(pr,Pr,30,(tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])).length,15,1),this.g=e}function ER(){ER=Z,aD=new E7("MEDIAN_LAYER",0),H$=new E7("TAIL_LAYER",1),oD=new E7("HEAD_LAYER",2),QT=new E7("SPACE_EFFICIENT_LAYER",3),EA=new E7("WIDEST_LAYER",4),yA=new E7("CENTER_LAYER",5)}function OAe(e,t,i){var a,u,l;if(!e.b[t.g]){for(e.b[t.g]=!0,a=i,!a&&(a=new xH),Mr(a.b,t),l=e.a[t.g].Jc();l.Ob();)u=f(l.Pb(),65),u.b!=t&&OAe(e,u.b,a),u.c!=t&&OAe(e,u.c,a),Mr(a.a,u);return a}return null}function uRn(e){switch(e.g){case 0:case 1:case 2:return tt(),hn;case 3:case 4:case 5:return tt(),Bn;case 6:case 7:case 8:return tt(),pn;case 9:case 10:case 11:return tt(),_n;default:return tt(),fc}}function lRn(e,t){var i;return e.c.length==0?!1:(i=W0t((Gt(0,e.c.length),f(e.c[0],18)).c.i),rf(),i==(xT(),PA)||i==RA?!0:j3(Lu(new Ft(null,new Bt(e,16)),new D6),new gtt(t)))}function mde(e,t){if(me(t,206))return nmn(e,f(t,27));if(me(t,193))return rmn(e,f(t,124));if(me(t,440))return tmn(e,f(t,168));throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[t])))))}function Byt(e,t,i){var a,u;if(this.f=e,a=f(bn(e.b,t),260),u=a?a.a:0,XTe(i,u),i>=(u/2|0))for(this.e=a?a.c:null,this.d=u;i++<u;)gmt(this);else for(this.c=a?a.b:null;i-- >0;)wxe(this);this.b=t,this.a=null}function dRn(e,t){var i,a;t.a?O5n(e,t):(i=f(Dae(e.b,t.b),60),i&&i==e.a[t.b.f]&&i.a&&i.a!=t.b.a&&i.c.Ec(t.b),a=f(Lae(e.b,t.b),60),a&&e.a[a.f]==t.b&&a.a&&a.a!=t.b.a&&t.b.c.Ec(a),Ise(e.b,t.b))}function Uyt(e,t){var i,a;if(i=f(Es(e.b,t),126),f(f(si(e.r,t),22),83).dc()){i.n.b=0,i.n.c=0;return}i.n.b=e.C.b,i.n.c=e.C.c,e.A.Gc((Ud(),GE))&&mSt(e,t),a=Y3n(e,t),Ade(e,t)==(K3(),Rw)&&(a+=2*e.w),i.a.a=a}function zyt(e,t){var i,a;if(i=f(Es(e.b,t),126),f(f(si(e.r,t),22),83).dc()){i.n.d=0,i.n.a=0;return}i.n.d=e.C.d,i.n.a=e.C.a,e.A.Gc((Ud(),GE))&&bSt(e,t),a=K3n(e,t),Ade(e,t)==(K3(),Rw)&&(a+=2*e.w),i.a.b=a}function fRn(e,t){var i,a,u,l;for(l=new et,a=new ne(t);a.a<a.c.c.length;)i=f(ee(a),68),Qe(l,new sSe(i,!0)),Qe(l,new sSe(i,!1));u=new sft(e),u.a.a.$b(),Oht(l,e.b,new bu(ce(se(sIt,1),Jt,678,0,[u])))}function hRn(e,t){var i,a,u;t.Sg("End label pre-processing",1),i=ke(Oe(V(e,(ot(),OA)))),a=ke(Oe(V(e,cx))),u=k3(f(V(e,mf),86)),Zi(Zs(new Ft(null,new Bt(e.b,16)),new rO),new Lut(i,a,u)),t.Tg()}function NAe(e,t){var i,a,u;if(!e.d[t.p]){for(e.d[t.p]=!0,e.a[t.p]=!0,a=new Tn(Cn(ao(t).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),!Fa(i)&&(u=i.d.i,e.a[u.p]?Qe(e.b,i):NAe(e,u));e.a[t.p]=!1}}function Hyt(e,t,i){var a;switch(a=0,f(V(t,(ot(),el)),171).g){case 2:a=2*-i+e.a,++e.a;break;case 1:a=-i;break;case 3:a=i;break;case 4:a=2*i+e.b,++e.b}return li(t,(Ge(),hi))&&(a+=f(V(t,hi),17).a),a}function Gyt(e,t,i){var a,u,l;for(i.yc(t,e),Qe(e.n,t),l=e.p.xg(t),t.j==e.p.yg()?L0t(e.e,l):L0t(e.j,l),zX(e),u=pb(Df(ce(se(Gf,1),Jt,20,0,[new Z5(t),new _N(t)])));Kn(u);)a=f(Dn(u),11),i._b(a)||Gyt(e,a,i)}function pRn(e,t,i){var a,u,l;for(i.Sg("Processor set neighbors",1),e.a=t.b.b==0?1:t.b.b,u=null,a=tr(t.b,0);!u&&a.b!=a.d.c;)l=f(Zn(a),39),lt(ut(V(l,(_i(),Iw))))&&(u=l);u&&Z2t(e,new u1(u),i),i.Tg()}function RAe(e){var t,i,a;return i=f(Fe(e,(Rr(),UE)),22),i.Gc((Ud(),mx))?(a=f(Fe(e,Ow),22),t=new ja(f(Fe(e,GA),8)),a.Gc((yd(),jD))&&(t.a<=0&&(t.a=20),t.b<=0&&(t.b=20)),t):new ha}function PAe(e){var t,i,a;if(!e.b){for(a=new f5,i=new GN(Jj(e));i.e!=i.i.gc();)t=f(Ple(i),19),(t.Bb&Dc)!=0&&Qn(a,t);ET(a),e.b=new _3((f(be(je((Wv(),sn).o),8),19),a.i),a.g),pd(e).b&=-9}return e.b}function IT(e){var t,i,a;for(i=e.length,a=0;a<i&&(gn(a,e.length),e.charCodeAt(a)<=32);)++a;for(t=i;t>a&&(gn(t-1,e.length),e.charCodeAt(t-1)<=32);)--t;return a>0||t<i?(ga(a,t,e.length),e.substr(a,t-a)):e}function gRn(e,t){var i,a,u,l,g,m,v,S;v=f(Rj(YN(t.k),Pe(Ja,qc,64,2,0,1)),125),S=t.g,i=Vht(t,v[0]),u=qht(t,v[1]),a=Kle(e,S,i,u),l=Vht(t,v[1]),m=qht(t,v[0]),g=Kle(e,S,l,m),a<=g?(t.a=i,t.c=u):(t.a=l,t.c=m)}function qyt(e,t,i,a,u){var l,g,m,v,S;if(t)for(m=t.Jc();m.Ob();)for(g=f(m.Pb(),9),S=RIe(g,(cs(),Uu),i).Jc();S.Ob();)v=f(S.Pb(),11),l=f(ac(Is(u.f,v)),116),l||(l=new VH(e.d),vn(a.c,l),Gyt(l,v,u))}function vG(e){var t;Ht();var i,a,u,l,g,m;if(me(e,59))for(l=0,u=e.gc()-1;l<u;++l,--u)t=e.Xb(l),e.fd(l,e.Xb(u)),e.fd(u,t);else for(i=e.cd(),g=e.dd(e.gc());i.Tb()<g.Vb();)a=i.Pb(),m=g.Ub(),i.Wb(m),g.Wb(a)}function VZ(e,t){var i,a,u,l,g,m;for(m=0,l=new O3,Jv(l,t);l.b!=l.c;)for(g=f(JN(l),217),m+=cwt(g.d,g.e),u=new ne(g.b);u.a<u.c.c.length;)a=f(ee(u),36),i=f(ct(e.b,a.p),217),i.s||(m+=VZ(e,i));return m}function Vyt(e,t,i){var a,u;i1t(this),t==(AS(),Aw)?Go(this.r,e.c):Go(this.w,e.c),i==Aw?Go(this.r,e.d):Go(this.w,e.d),Iyt(this,e),a=lle(e.c),u=lle(e.d),gyt(this,a,u,u),this.o=(qj(),h.Math.abs(a-u)<.2)}function Wyt(e,t,i){var a,u,l,g,m,v;if(m=f(fn(e.a,8),1981),m!=null)for(u=m,l=0,g=u.length;l<g;++l)null.Rm();a=i,(e.a.Db&1)==0&&(v=new wdt(e,i,t),a.aj(v)),me(a,671)?f(a,671).cj(e.a):a._i()==e.a&&a.bj(null)}function mRn(){var e;return FMt?f(gL((Uv(),xh),CE),1990):(v9n(),e=f(me(wu((Uv(),xh),CE),577)?wu(xh,CE):new Odt,577),FMt=!0,i7n(e),$7n(e),qr((eSe(),j7e),e,new My),Xle(e),Os(xh,CE,e),e)}function bRn(e,t,i,a){var u;return u=Q3(e,i,ce(se(ft,1),Ye,2,6,[xfe,Cfe,Afe,Ife,Ofe,Nfe,Rfe]),t),u<0&&(u=Q3(e,i,ce(se(ft,1),Ye,2,6,["Sun","Mon","Tue","Wed","Thu","Fri","Sat"]),t)),u<0?!1:(a.d=u,!0)}function vRn(e,t,i,a){var u;return u=Q3(e,i,ce(se(ft,1),Ye,2,6,[xfe,Cfe,Afe,Ife,Ofe,Nfe,Rfe]),t),u<0&&(u=Q3(e,i,ce(se(ft,1),Ye,2,6,["Sun","Mon","Tue","Wed","Thu","Fri","Sat"]),t)),u<0?!1:(a.d=u,!0)}function ml(e,t,i){var a,u,l,g;if(g=e.b.Ae(t),u=(a=e.a.get(g),a??Pe(Ro,Jt,1,0,5,1)),u.length==0)e.a.set(g,u);else if(l=J1t(e,t,u),l)return l.ld(i);return io(u,u.length,new AY(t,i)),++e.c,++e.b.g,null}function wRn(e){var t,i,a;for(b6n(e),a=new et,i=new ne(e.a.a.b);i.a<i.c.c.length;)t=f(ee(i),82),Qe(a,new lSe(t,!0)),Qe(a,new lSe(t,!1));pAn(e.c),SH(a,e.b,new bu(ce(se(Sq,1),Jt,374,0,[e.c]))),P4n(e)}function wG(e,t){var i,a,u;for(u=new et,a=new ne(e.c.a.b);a.a<a.c.c.length;)i=f(ee(a),60),t.Lb(i)&&(Qe(u,new aSe(i,!0)),Qe(u,new aSe(i,!1)));hAn(e.e),Oht(u,e.d,new bu(ce(se(sIt,1),Jt,678,0,[e.e])))}function yRn(e){var t,i,a,u;for(i=new Wn,u=new ne(e.d);u.a<u.c.c.length;)a=f(ee(u),187),t=f(a.mf((Ge(),YR)),18),Is(i.f,t)||qr(i,t,new ift(t)),Qe(f(ac(Is(i.f,t)),452).b,a);return new Kl(new kp(i))}function ERn(e,t){var i,a,u,l,g;for(a=new Tht(e.j.c.length),i=null,l=new ne(e.j);l.a<l.c.c.length;)u=f(ee(l),11),u.j!=i&&(a.b==a.c||p2t(a,i,t),jTe(a),i=u.j),g=KEt(u),g&&NTe(a,g);a.b==a.c||p2t(a,i,t)}function SRn(e,t){var i,a,u;for(a=new la(e.b,0);a.b<a.d.gc();)i=(Ln(a.b<a.d.gc()),f(a.d.Xb(a.c=a.b++),72)),u=f(V(i,(ot(),Rm)),276),u==(Sg(),gx)&&(hd(a),Qe(t.b,i),li(i,(Ge(),YR))||$e(i,YR,e))}function _Rn(e){var t,i,a,u,l;for(t=g1(new Tn(Cn(ao(e).a.Jc(),new B))),u=new Tn(Cn(Fo(e).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),i=a.c.i,l=g1(new Tn(Cn(ao(i).a.Jc(),new B))),t=h.Math.max(t,l);return rt(t)}function LAe(e,t,i){var a,u,l;a=f(Fe(e,(Rr(),tV)),22),u=0,l=0,t.a>i.a&&(a.Gc((mE(),xF))?u=(t.a-i.a)/2:a.Gc(CF)&&(u=t.a-i.a)),t.b>i.b&&(a.Gc((mE(),IF))?l=(t.b-i.b)/2:a.Gc(AF)&&(l=t.b-i.b)),gAe(e,u,l)}function Kyt(e,t,i,a,u,l,g,m,v,S,T,A,L){me(e.Cb,88)&&OT(pd(f(e.Cb,88)),4),ju(e,i),e.f=g,eL(e,m),tL(e,v),QP(e,S),nL(e,T),N0(e,A),rL(e,L),O0(e,!0),x0(e,u),e.Wk(l),fE(e,t),a!=null&&(e.i=null,qJ(e,a))}function DAe(e,t,i){if(e<0)return Vj(Wkt,ce(se(Ro,1),Jt,1,5,[i,rt(e)]));if(t<0)throw ie(new dn(Kkt+t));return Vj("%s (%s) must not be greater than size (%s)",ce(se(Ro,1),Jt,1,5,[i,rt(e),rt(t)]))}function MAe(e,t,i,a,u,l){var g,m,v,S;if(g=a-i,g<7){ECn(t,i,a,l);return}if(v=i+u,m=a+u,S=v+(m-v>>1),MAe(t,e,v,S,-u,l),MAe(t,e,S,m,-u,l),l.Le(e[S-1],e[S])<=0){for(;i<a;)io(t,i++,e[v++]);return}Pxn(e,v,S,m,t,i,a,l)}function kRn(e,t){var i,a,u,l,g,m,v;for(v=t.d,u=t.b.j,m=new ne(v);m.a<m.c.c.length;)for(g=f(ee(m),105),l=Pe(kd,Tb,30,u.c.length,16,1),qr(e.b,g,l),i=g.a.d.p-1,a=g.c.d.p;i!=a;)i=(i+1)%u.c.length,l[i]=!0}function TRn(e,t){if(Pxe(),li(e,(Ge(),hi))&&li(t,hi))return tu(f(V(e,hi),17).a,f(V(t,hi),17).a);throw ie(new ON("The BF model order layer assigner requires all real nodes to have a model order."))}function xRn(e,t){if(Lxe(),li(e,(Ge(),hi))&&li(t,hi))return tu(f(V(e,hi),17).a,f(V(t,hi),17).a);throw ie(new ON("The DF model order layer assigner requires all real nodes to have a model order."))}function CRn(e,t){for(e.r=new VH(e.p),VK(e.r,e),Ba(e.r.j,e.j),nf(e.j),Mr(e.j,t),Mr(e.r.e,t),zX(e),zX(e.r);e.f.c.length!=0;)Ust(f(ct(e.f,0),133));for(;e.k.c.length!=0;)Ust(f(ct(e.k,0),133));return e.r}function bde(e,t,i){var a,u,l;if(u=Vt(e.zh(),t),a=t-e.fi(),a<0)if(u)if(u.ok())l=e.Eh(u),l>=0?e.Zh(l,i):pIe(e,u,i);else throw ie(new dn(vw+u.ve()+S$));else throw ie(new dn(DCt+t+MCt));else $f(e,a,u,i)}function Yyt(e){var t,i;if(e.f){for(;e.n>0;){if(t=f(e.k.Xb(e.n-1),74),i=t.Ik(),me(i,101)&&(f(i,19).Bb&Dc)!=0&&(!e.e||i.mk()!=$D||i.Ij()!=0)&&t.kd()!=null)return!0;--e.n}return!1}else return e.n>0}function Xyt(e){var t,i,a,u;if(i=f(e,52).Xh(),i)try{if(a=null,t=gL((Uv(),xh),FSt(TCn(i))),t&&(u=t.Yh(),u&&(a=u.Cl(ppn(i.e)))),a&&a!=e)return Xyt(a)}catch(l){if(l=go(l),!me(l,63))throw ie(l)}return e}function ARn(e,t,i){var a,u,l;i.Sg("Remove overlaps",1),i.ah(t,HRe),a=f(Fe(t,(R3(),MA)),27),e.f=a,e.a=Dle(f(Fe(t,(uw(),Kq)),299)),u=Oe(Fe(t,(Rr(),zE))),SN(e,(an(u),u)),l=RT(a),okt(e,t,l,i),i.ah(t,tee)}function IRn(e){var t,i,a;if(lt(ut(Fe(e,(Rr(),rV))))){for(a=new et,i=new Tn(Cn(lw(e).a.Jc(),new B));Kn(i);)t=f(Dn(i),85),HS(t)&&lt(ut(Fe(t,Wbe)))&&vn(a.c,t);return a}else return Ht(),Ht(),Ga}function Jyt(e){if(!e)return Qrt(),jAt;var t=e.valueOf?e.valueOf():e;if(t!==e){var i=Gpe[typeof t];return i?i(t):ICe(typeof t)}else return e instanceof Array||e instanceof h.Array?new bN(e):new d3(e)}function Zyt(e,t,i){var a,u,l;switch(l=e.o,a=f(Es(e.p,i),251),u=a.i,u.b=Fj(a),u.a=$j(a),u.b=h.Math.max(u.b,l.a),u.b>l.a&&!t&&(u.b=l.a),u.c=-(u.b-l.a)/2,i.g){case 1:u.d=-u.a;break;case 3:u.d=l.b}Ude(a),zde(a)}function Qyt(e,t,i){var a,u,l;switch(l=e.o,a=f(Es(e.p,i),251),u=a.i,u.b=Fj(a),u.a=$j(a),u.a=h.Math.max(u.a,l.b),u.a>l.b&&!t&&(u.a=l.b),u.d=-(u.a-l.b)/2,i.g){case 4:u.c=-u.b;break;case 2:u.c=l.a}Ude(a),zde(a)}function ORn(e,t){var i,a,u;return me(t.g,9)&&f(t.g,9).k==(cn(),bo)?Gi:(u=iR(t),u?h.Math.max(0,e.b/2-.5):(i=D3(t),i?(a=ke(Oe(kT(i,(ot(),v_)))),h.Math.max(0,a/2-.5)):Gi))}function NRn(e,t){var i,a,u;return me(t.g,9)&&f(t.g,9).k==(cn(),bo)?Gi:(u=iR(t),u?h.Math.max(0,e.b/2-.5):(i=D3(t),i?(a=ke(Oe(kT(i,(ot(),v_)))),h.Math.max(0,a/2-.5)):Gi))}function RRn(e,t){var i,a,u,l,g;if(!t.dc()){if(u=f(t.Xb(0),132),t.gc()==1){j2t(e,u,u,1,0,t);return}for(i=1;i<t.gc();)(u.j||!u.o)&&(l=zAn(t,i),l&&(a=f(l.a,17).a,g=f(l.b,132),j2t(e,u,g,i,a,t),i=a+1,u=g))}}function PRn(e){var t,i,a,u,l,g;for(g=new Kl(e.d),na(g,new P6),t=(aQ(),ce(se(cPe,1),Ve,272,0,[Age,Nge,Cge,Lge,Oge,Ige,Pge,Rge])),i=0,l=new ne(g);l.a<l.c.c.length;)u=f(ee(l),105),a=t[i%t.length],E4n(u,a),++i}function jAe(e,t){var i,a,u,l,g,m,v,S,T,A;return T=null,A=e,g=E1(A,"x"),i=new hnt(t),cmt(i.a,g),m=E1(A,"y"),a=new pnt(t),umt(a.a,m),v=E1(A,cpe),u=new gnt(t),lmt(u.a,v),S=E1(A,spe),l=new mnt(t),T=(dmt(l.a,S),S),T}function OT(e,t){hSt(e,t),(e.b&1)!=0&&(e.a.a=null),(e.b&2)!=0&&(e.a.f=null),(e.b&4)!=0&&(e.a.g=null,e.a.i=null),(e.b&16)!=0&&(e.a.d=null,e.a.e=null),(e.b&8)!=0&&(e.a.b=null),(e.b&32)!=0&&(e.a.j=null,e.a.c=null)}function LRn(e,t){var i,a,u;if(u=0,t.length>0)try{u=ff(t,Ra,Qr)}catch(l){throw l=go(l),me(l,130)?(a=l,ie(new xJ(a))):ie(l)}return i=(!e.a&&(e.a=new cae(e)),e.a),u<i.i&&u>=0?f(be(i,u),57):null}function DRn(e,t){if(e<0)return Vj(Wkt,ce(se(Ro,1),Jt,1,5,["index",rt(e)]));if(t<0)throw ie(new dn(Kkt+t));return Vj("%s (%s) must be less than size (%s)",ce(se(Ro,1),Jt,1,5,["index",rt(e),rt(t)]))}function MRn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],l.a?xr(l.a,l.b):l.a=new ef(l.d),P7(l.a,""+t);return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function jRn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],l.a?xr(l.a,l.b):l.a=new ef(l.d),P7(l.a,""+t);return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function $Rn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],l.a?xr(l.a,l.b):l.a=new ef(l.d),P7(l.a,""+t);return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function FRn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],l.a?xr(l.a,l.b):l.a=new ef(l.d),P7(l.a,""+t);return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function eEt(e,t){var i,a,u,l,g,m;for(i=e.b.c.length,u=ct(e.b,t);t*2+1<i&&(a=(l=2*t+1,g=l+1,m=l,g<i&&e.a.Le(ct(e.b,g),ct(e.b,l))<0&&(m=g),m),!(e.a.Le(u,ct(e.b,a))<0));)af(e.b,t,ct(e.b,a)),t=a;af(e.b,t,u)}function vde(e,t,i){var a,u;return a=i.d,u=i.e,e.g[a.d]<=e.i[t.d]&&e.i[t.d]<=e.i[a.d]&&e.g[u.d]<=e.i[t.d]&&e.i[t.d]<=e.i[u.d]?!(e.i[a.d]<e.i[u.d]):e.i[a.d]<e.i[u.d]}function BRn(e,t){var i;if(i=f(V(t,(ot(),Q$)),323),i!=e)throw ie(new ON("The hierarchy aware processor "+i+" in child node "+t+" is only allowed if the root node specifies the same hierarchical processor."))}function URn(e,t){var i,a,u,l,g;for(a=(!t.s&&(t.s=new Be(El,t,21,17)),t.s),l=null,u=0,g=a.i;u<g;++u)switch(i=f(be(a,u),178),kS(Ns(e,i))){case 2:case 3:!l&&(l=new et),vn(l.c,i)}return l||(Ht(),Ht(),Ga)}function tEt(e,t,i){var a,u,l,g,m,v;for(v=Gi,l=new ne(x2t(e.b));l.a<l.c.c.length;)for(u=f(ee(l),176),m=new ne(x2t(t.b));m.a<m.c.c.length;)g=f(ee(m),176),a=RTn(u.a,u.b,g.a,g.b,i),v=h.Math.min(v,a);return v}function No(e,t){if(!t)throw ie(new kN);if(e.j=t,!e.d)switch(e.j.g){case 1:e.a.a=e.o.a/2,e.a.b=0;break;case 2:e.a.a=e.o.a,e.a.b=e.o.b/2;break;case 3:e.a.a=e.o.a/2,e.a.b=e.o.b;break;case 4:e.a.a=0,e.a.b=e.o.b/2}}function zRn(e){var t,i,a,u,l,g,m;for(a=0,i=new ne(e.b);i.a<i.c.c.length;)for(t=f(ee(i),25),l=new ne(t.a);l.a<l.c.c.length;)for(u=f(ee(l),9),u.p=a++,m=new ne(u.j);m.a<m.c.c.length;)g=f(ee(m),11),g.p=a++}function HRn(e,t){tf();var i,a,u,l,g,m;for(i=null,g=t.Jc();g.Ob();)l=f(g.Pb(),132),!l.o&&(a=pmn(l.a),u=bvn(l.a),m=new Xj(a,u,null,f(l.d.a.ec().Jc().Pb(),18)),Qe(m.c,l.a),vn(e.c,m),i&&Qe(i.d,m),i=m)}function GRn(e){var t,i,a,u,l,g;for(g=kj(e.d,e.e),l=g.Jc();l.Ob();)for(u=f(l.Pb(),11),a=e.e==(tt(),pn)?u.e:u.g,i=new ne(a);i.a<i.c.c.length;)t=f(ee(i),18),!Fa(t)&&t.c.i.c!=t.d.i.c&&(GNn(e,t),++e.f,++e.c)}function nEt(e,t){var i,a;if(t.dc())return Ht(),Ht(),Ga;for(a=new et,Qe(a,rt(Ra)),i=1;i<e.f;++i)e.a==null&&fSt(e),e.a[i]&&Qe(a,rt(i));return a.c.length==1?(Ht(),Ht(),Ga):(Qe(a,rt(Qr)),ODn(t,a))}function qRn(e,t){var i,a,u,l,g,m,v;g=t.c.i.k!=(cn(),Yi),v=g?t.d:t.c,i=iIn(t,v).i,u=f(bn(e.k,v),123),a=e.i[i.p].a,yut(v.i)<(i.c?Oc(i.c.a,i,0):-1)?(l=u,m=a):(l=a,m=u),Zh(Uh(Bh(zh(Fh(new ah,0),4),l),m))}function rEt(e,t,i){var a,u,l,g,m,v;if(i)for(u=i.a.length,a=new Xy(u),m=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);m.Ob();)g=f(m.Pb(),17),v=qle(e,SR(gT(i,g.a))),v&&(l=(!t.b&&(t.b=new Qt(Yn,t,4,7)),t.b),Qn(l,v))}function iEt(e,t,i){var a,u,l,g,m,v;if(i)for(u=i.a.length,a=new Xy(u),m=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);m.Ob();)g=f(m.Pb(),17),v=qle(e,SR(gT(i,g.a))),v&&(l=(!t.c&&(t.c=new Qt(Yn,t,5,8)),t.c),Qn(l,v))}function yG(e,t,i){var a,u;a=t.a&e.f,t.b=e.b[a],e.b[a]=t,u=t.f&e.f,t.d=e.c[u],e.c[u]=t,i?(t.e=i.e,t.e?t.e.c=t:e.a=t,t.c=i.c,t.c?t.c.e=t:e.e=t):(t.e=e.e,t.c=null,e.e?e.e.c=t:e.a=t,e.e=t),++e.i,++e.g}function oEt(e){var t,i,a;if(t=e.Pb(),!e.Ob())return t;for(a=vu(xr(new $v,"expected one element but was: <"),t),i=0;i<4&&e.Ob();i++)vu((a.a+=Fu,a),e.Pb());throw e.Ob()&&(a.a+=", ..."),a.a+=">",ie(new dn(a.a))}function VRn(e){var t,i;return i=-e.a,t=ce(se(Xf,1),_m,30,15,[43,48,48,48,48]),i<0&&(t[0]=45,i=-i),t[1]=t[1]+((i/60|0)/10|0)&To,t[2]=t[2]+(i/60|0)%10&To,t[3]=t[3]+(i%60/10|0)&To,t[4]=t[4]+i%10&To,bm(t,0,t.length)}function $Ae(e){var t,i,a,u;for(e.g=new Aj(f(or(Ja),295)),a=0,i=(tt(),hn),t=0;t<e.j.c.length;t++)u=f(ct(e.j,t),11),u.j!=i&&(a!=t&&ZN(e.g,i,new Ya(rt(a),rt(t))),i=u.j,a=t);ZN(e.g,i,new Ya(rt(a),rt(t)))}function FAe(e,t){var i,a,u;if(u=iA((Rl(),Ea),e.zh(),t),u)is(),f(u,69).uk()||(u=rR(Ns(Ea,u))),a=(i=e.Eh(u),f(i>=0?e.Hh(i,!0,!0):GS(e,u,!0),159)),f(a,218).Wl(t);else throw ie(new dn(vw+t.ve()+S$))}function BAe(e){var t,i;return e>-0x800000000000&&e<0x800000000000?e==0?0:(t=e<0,t&&(e=-e),i=Aa(h.Math.floor(h.Math.log(e)/.6931471805599453)),(!t||e!=h.Math.pow(2,i))&&++i,i):y0t(Nc(e))}function WRn(e){var t,i,a,u,l,g,m;for(l=new ub,i=new ne(e);i.a<i.c.c.length;)t=f(ee(i),133),g=t.a,m=t.b,!(l.a._b(g)||l.a._b(m))&&(u=g,a=m,g.e.b+g.j.b>2&&m.e.b+m.j.b<=2&&(u=m,a=g),l.a.yc(u,l),u.q=a);return l}function KRn(e,t,i){i.Sg("Eades radial",1),i.ah(t,tee),e.d=f(Fe(t,(R3(),MA)),27),e.c=ke(Oe(Fe(t,(uw(),ane)))),e.e=Dle(f(Fe(t,Kq),299)),e.a=UCn(f(Fe(t,zMe),423)),e.b=YIn(f(Fe(t,$Me),351)),PIn(e),i.ah(t,tee)}function YRn(e,t){if(t.Sg("Target Width Setter",1),mm(e,(Ag(),Tbe)))pi(e,(wb(),fx),Oe(Fe(e,Tbe)));else throw ie(new h0("A target width has to be set if the TargetWidthWidthApproximator should be used."));t.Tg()}function aEt(e,t){var i,a,u;return a=new Tg(e),Rc(a,t),$e(a,(Ge(),_te),t),$e(a,(ot(),Xi),(oa(),uu)),$e(a,Nm,(vb(),vne)),Wl(a,(cn(),bo)),i=new ou,cc(i,a),No(i,(tt(),pn)),u=new ou,cc(u,a),No(u,_n),a}function sEt(e){switch(e.g){case 0:return new Sae((PS(),zq));case 1:return new j9;case 2:return new $9;default:throw ie(new dn("No implementation is available for the crossing minimizer "+(e.f!=null?e.f:""+e.g)))}}function cEt(e,t){var i,a,u,l,g;for(e.c[t.p]=!0,Qe(e.a,t),g=new ne(t.j);g.a<g.c.c.length;)for(l=f(ee(g),11),a=new yg(l.b);sc(a.a)||sc(a.b);)i=f(sc(a.a)?ee(a.a):ee(a.b),18),u=s3n(l,i).i,e.c[u.p]||cEt(e,u)}function uEt(e){var t,i,a,u,l,g,m;for(g=0,i=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));i.e!=i.i.gc();)t=f(Mn(i),27),m=t.g,u=t.f,a=h.Math.sqrt(m*m+u*u),g=h.Math.max(a,g),l=uEt(t),g=h.Math.max(l,g);return g}function Xl(){Xl=Z,Lw=new I7("OUTSIDE",0),M1=new I7("INSIDE",1),hV=new I7("NEXT_TO_PORT_IF_POSSIBLE",2),jF=new I7("ALWAYS_SAME_SIDE",3),MF=new I7("ALWAYS_OTHER_SAME_SIDE",4),u4=new I7("SPACE_EFFICIENT",5)}function lEt(e,t,i){var a,u,l,g,m,v;return a=A2n(e,(u=(Bv(),l=new o7,l),i&&eQ(u,i),u),t),FP(a,p1(t,Am)),UZ(t,a),e6n(t,a),jAe(t,a),g=t,m=IS(g,"ports"),v=new zat(e,a),E2t(v.a,v.b,m),Zue(e,t,a),kTn(e,t,a),a}function XRn(e){var t,i;return i=-e.a,t=ce(se(Xf,1),_m,30,15,[43,48,48,58,48,48]),i<0&&(t[0]=45,i=-i),t[1]=t[1]+((i/60|0)/10|0)&To,t[2]=t[2]+(i/60|0)%10&To,t[4]=t[4]+(i%60/10|0)&To,t[5]=t[5]+i%10&To,bm(t,0,t.length)}function JRn(e){var t;return t=ce(se(Xf,1),_m,30,15,[71,77,84,45,48,48,58,48,48]),e<=0&&(t[3]=43,e=-e),t[4]=t[4]+((e/60|0)/10|0)&To,t[5]=t[5]+(e/60|0)%10&To,t[7]=t[7]+(e%60/10|0)&To,t[8]=t[8]+e%10&To,bm(t,0,t.length)}function ZRn(e){var t,i,a,u,l;if(e==null)return bl;for(l=new uE(Fu,"[","]"),i=e,a=0,u=i.length;a<u;++a)t=i[a],l.a?xr(l.a,l.b):l.a=new ef(l.d),P7(l.a,""+V7(t));return l.a?l.e.length==0?l.a.a:l.a.a+(""+l.e):l.c}function UAe(e,t){var i,a,u;for(u=Qr,a=new ne(V3(t));a.a<a.c.c.length;)i=f(ee(a),216),i.f&&!e.c[i.c]&&(e.c[i.c]=!0,u=h.Math.min(u,UAe(e,hZ(i,t))));return e.i[t.d]=e.j,e.g[t.d]=h.Math.min(u,e.j++),e.g[t.d]}function dEt(e,t){var i,a,u;for(u=f(f(si(e.r,t),22),83).Jc();u.Ob();)a=f(u.Pb(),115),a.e.b=(i=a.b,i.nf((Rr(),Sd))?i.$f()==(tt(),hn)?-i.Kf().b-ke(Oe(i.mf(Sd))):ke(Oe(i.mf(Sd))):i.$f()==(tt(),hn)?-i.Kf().b:0)}function zAe(e){var t,i,a,u,l;t=e.a,i=e.b,u=e.c,a=new Ke(i.e.a+i.f.a/2,i.e.b+i.f.b/2),l=new Ke(u.e.a+u.f.a/2,u.e.b+u.f.b/2),Wi(t,a,t.a,t.a.a),Wi(t,l,t.c.b,t.c),myt(a,f(Xh(t,1),8),e.b.f),myt(l,f(Xh(t,t.b-2),8),e.c.f)}function SR(e){var t,i;if(i=!1,me(e,210))return i=!0,f(e,210).a;if(!i&&me(e,263)&&(t=f(e,263).a%1==0,t))return i=!0,rt(Pgn(f(e,263).a));throw ie(new am("Id must be a string or an integer: '"+e+"'."))}function QRn(e,t){var i,a,u,l,g,m;for(l=null,u=new Zlt((!e.a&&(e.a=new cae(e)),e.a));WAe(u);)if(i=f(XZ(u),57),a=(g=i.zh(),m=(rA(g),g.o),!m||!i.Th(m)?null:F_e(que(m),i.Ih(m))),a!=null&&Mt(a,t)){l=i;break}return l}function fEt(e,t,i){var a,u,l,g,m;if(uf(i,"occurrences"),i==0)return m=f(ST(tR(e.a),t),16),m?m.gc():0;if(g=f(ST(tR(e.a),t),16),!g)return 0;if(l=g.gc(),i>=l)g.$b();else for(u=g.Jc(),a=0;a<i;a++)u.Pb(),u.Qb();return l}function e4n(e,t,i){var a,u,l,g;return uf(i,"oldCount"),uf(0,"newCount"),a=f(ST(tR(e.a),t),16),(a?a.gc():0)==i?(uf(0,"count"),u=(l=f(ST(tR(e.a),t),16),l?l.gc():0),g=-u,g>0?D2e():g<0&&fEt(e,t,-g),!0):!1}function $j(e){var t,i,a,u,l,g,m;if(m=0,e.b==0){for(g=Gvt(e,!0),t=0,a=g,u=0,l=a.length;u<l;++u)i=a[u],i>0&&(m+=i,++t);t>1&&(m+=e.c*(t-1))}else m=cit(Fue(uT(Jr(pce(e.a),new js),new Ac)));return m>0?m+e.n.d+e.n.a:0}function Fj(e){var t,i,a,u,l,g,m;if(m=0,e.b==0)m=cit(Fue(uT(Jr(pce(e.a),new Xc),new rc)));else{for(g=qvt(e,!0),t=0,a=g,u=0,l=a.length;u<l;++u)i=a[u],i>0&&(m+=i,++t);t>1&&(m+=e.c*(t-1))}return m>0?m+e.n.b+e.n.c:0}function t4n(e){var t,i;if(e.c.length!=2)throw ie(new As("Order only allowed for two paths."));t=(Gt(0,e.c.length),f(e.c[0],18)),i=(Gt(1,e.c.length),f(e.c[1],18)),t.d.i!=i.c.i&&(e.c.length=0,vn(e.c,i),vn(e.c,t))}function hEt(e,t,i){var a;for(x3(i,t.g,t.f),Of(i,t.i,t.j),a=0;a<(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i;a++)hEt(e,f(be((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a),a),27),f(be((!i.a&&(i.a=new Be(Ur,i,10,11)),i.a),a),27))}function n4n(e,t){var i,a,u,l;for(l=f(Es(e.b,t),126),i=l.a,u=f(f(si(e.r,t),22),83).Jc();u.Ob();)a=f(u.Pb(),115),a.c&&(i.a=h.Math.max(i.a,dke(a.c)));if(i.a>0)switch(t.g){case 2:l.n.c=e.s;break;case 4:l.n.b=e.s}}function r4n(e,t){var i,a,u;return i=f(V(t,(Qh(),GR)),17).a-f(V(e,GR),17).a,i==0?(a=Ho($a(f(V(e,(ew(),Eq)),8)),f(V(e,B$),8)),u=Ho($a(f(V(t,Eq),8)),f(V(t,B$),8)),di(a.a*a.b,u.a*u.b)):i}function i4n(e,t){var i,a,u;return i=f(V(t,(Ec(),tne)),17).a-f(V(e,tne),17).a,i==0?(a=Ho($a(f(V(e,(_i(),qq)),8)),f(V(e,TD),8)),u=Ho($a(f(V(t,qq),8)),f(V(t,TD),8)),di(a.a*a.b,u.a*u.b)):i}function pEt(e){var t,i;return i=new $v,i.a+="e_",t=yTn(e),t!=null&&(i.a+=""+t),e.c&&e.d&&(xr((i.a+=" ",i),DZ(e.c)),xr(vu((i.a+="[",i),e.c.i),"]"),xr((i.a+=Xfe,i),DZ(e.d)),xr(vu((i.a+="[",i),e.d.i),"]")),i.a}function gEt(e){switch(e.g){case 0:return new tz;case 1:return new dN;case 2:return new _p;case 3:return new hN;default:throw ie(new dn("No implementation is available for the layout phase "+(e.f!=null?e.f:""+e.g)))}}function HAe(e,t,i,a,u){var l;switch(l=0,u.g){case 1:l=h.Math.max(0,t.b+e.b-(i.b+a));break;case 3:l=h.Math.max(0,-e.b-a);break;case 2:l=h.Math.max(0,-e.a-a);break;case 4:l=h.Math.max(0,t.a+e.a-(i.a+a))}return l}function mEt(e,t,i){var a,u,l,g,m;if(i)for(u=i.a.length,a=new Xy(u),m=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);m.Ob();)g=f(m.Pb(),17),l=OP(i,g.a),M4e in l.a||fpe in l.a?rLn(e,l,t):p7n(e,l,t),Lmn(f(bn(e.c,oL(l)),85))}function GAe(e){var t,i;switch(e.b){case-1:return!0;case 0:return i=e.t,i>1||i==-1?(e.b=-1,!0):(t=ph(e),t&&(is(),t.ik()==$3t)?(e.b=-1,!0):(e.b=1,!1));default:case 1:return!1}}function qAe(e,t){var i,a,u,l;if(Yr(e),e.c!=0||e.a!=123)throw ie(new vr(yr((lr(),a3t))));if(l=t==112,a=e.d,i=fP(e.i,125,a),i<0)throw ie(new vr(yr((lr(),s3t))));return u=uh(e.i,a,i),e.d=i+1,Agt(u,l,(e.e&512)==512)}function o4n(e){var t,i,a,u,l,g,m;for(m=lb(e.c.length),u=new ne(e);u.a<u.c.c.length;){for(a=f(ee(u),9),g=new Mo,l=ao(a),i=new Tn(Cn(l.a.Jc(),new B));Kn(i);)t=f(Dn(i),18),t.c.i==t.d.i||Go(g,t.d.i);vn(m.c,g)}return m}function a4n(e,t){var i,a,u;if(!t)Rue(e,null),jP(e,null);else if((t.i&4)!=0)for(a="[]",i=t.c;;i=i.c){if((i.i&4)==0){u=R2e((d1(i),i.o+a)),Rue(e,u),jP(e,u);break}a+="[]"}else u=R2e((d1(t),t.o)),Rue(e,u),jP(e,u);e.el(t)}function Bj(e,t,i,a,u){var l,g,m,v;return v=jse(e,f(u,57)),Ae(v)!==Ae(u)?(m=f(e.g[i],74),l=mb(t,v),j7(e,i,ide(e,i,l)),Ld(e.e)&&(g=Yv(e,9,l.Ik(),u,v,a,!1),nAe(g,new v1(e.e,9,e.c,m,l,a,!1)),yue(g)),v):u}function s4n(e,t,i){var a,u,l,g,m,v;for(a=f(si(e.c,t),15),u=f(si(e.c,i),15),l=a.dd(a.gc()),g=u.dd(u.gc());l.Sb()&&g.Sb();)if(m=f(l.Ub(),17),v=f(g.Ub(),17),m!=v)return tu(m.a,v.a);return!l.Ob()&&!g.Ob()?0:l.Ob()?1:-1}function bEt(e,t){var i,a,u;try{return u=JEn(e.a,t),u}catch(l){if(l=go(l),me(l,32)){try{if(a=ff(t,Ra,Qr),i=Tp(e.a),a>=0&&a<i.length)return i[a]}catch(g){if(g=go(g),!me(g,130))throw ie(g)}return null}else throw ie(l)}}function wde(e,t){var i,a,u;if(u=iA((Rl(),Ea),e.zh(),t),u)return is(),f(u,69).uk()||(u=rR(Ns(Ea,u))),a=(i=e.Eh(u),f(i>=0?e.Hh(i,!0,!0):GS(e,u,!0),159)),f(a,218).Tl(t);throw ie(new dn(vw+t.ve()+rpe))}function c4n(){tSe();var e;return bMt?f(gL((Uv(),xh),bh),1984):(Dr(AE,new d9),M8n(),e=f(me(wu((Uv(),xh),bh),545)?wu(xh,bh):new Ndt,545),bMt=!0,L7n(e),U7n(e),qr((eSe(),j7e),e,new _k),Os(xh,bh,e),e)}function u4n(e,t){var i,a,u,l;e.j=-1,Ld(e.e)?(i=e.i,l=e.i!=0,TH(e,t),a=new v1(e.e,3,e.c,null,t,i,l),u=t.wl(e.e,e.c,null),u=Wwt(e,t,u),u?(u.kj(a),u.lj()):Zr(e.e,a)):(TH(e,t),u=t.wl(e.e,e.c,null),u&&u.lj())}function WZ(e,t){var i,a,u;if(u=0,a=t[0],a>=e.length)return-1;for(i=(gn(a,e.length),e.charCodeAt(a));i>=48&&i<=57&&(u=u*10+(i-48),++a,!(a>=e.length));)i=(gn(a,e.length),e.charCodeAt(a));return a>t[0]?t[0]=a:u=-1,u}function l4n(e,t,i){var a,u,l,g,m;g=e.c,m=e.d,l=uc(ce(se(Zo,1),Ye,8,0,[g.i.n,g.n,g.a])).b,u=(l+uc(ce(se(Zo,1),Ye,8,0,[m.i.n,m.n,m.a])).b)/2,a=null,g.j==(tt(),_n)?a=new Ke(t+g.i.c.c.a+i,u):a=new Ke(t-i,u),pP(e.a,0,a)}function HS(e){var t,i,a,u;for(t=null,a=pb(Df(ce(se(Gf,1),Jt,20,0,[(!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c)])));Kn(a);)if(i=f(Dn(a),84),u=Us(i),!t)t=u;else if(t!=u)return!1;return!0}function yde(e,t,i){var a;if(++e.j,t>=e.i)throw ie(new Ru(ppe+t+TE+e.i));if(i>=e.i)throw ie(new Ru(gpe+i+TE+e.i));return a=e.g[i],t!=i&&(t<i?au(e.g,t,e.g,t+1,i-t):au(e.g,i+1,e.g,i,t-i),io(e.g,t,a),e.Mi(t,a,i),e.Ki()),a}function jt(e,t,i){var a;if(a=f(e.c.xc(t),16),a)return a.Ec(i)?(++e.d,!0):!1;if(a=e.ic(t),a.Ec(i))return++e.d,e.c.yc(t,a),!0;throw ie(new UTe("New Collection violated the Collection spec"))}function EG(e){var t,i,a;return e<0?0:e==0?32:(a=-(e>>16),t=a>>16&16,i=16-t,e=e>>t,a=e-256,t=a>>16&8,i+=t,e<<=t,a=e-FT,t=a>>16&4,i+=t,e<<=t,a=e-Sm,t=a>>16&2,i+=t,e<<=t,a=e>>14,t=a&~(a>>1),i+2-t)}function d4n(e,t){var i,a,u;for(u=new et,a=tr(t.a,0);a.b!=a.d.c;)i=f(Zn(a),65),i.c.g==e.g&&Ae(V(i.b,(Ec(),Lm)))!==Ae(V(i.c,Lm))&&!j3(new Ft(null,new Bt(u,16)),new Ctt(i))&&vn(u.c,i);return na(u,new B6),u}function vEt(e,t,i){var a,u,l,g;return me(t,151)&&me(i,151)?(l=f(t,151),g=f(i,151),e.a[l.a][g.a]+e.a[g.a][l.a]):me(t,249)&&me(i,249)&&(a=f(t,249),u=f(i,249),a.a==u.a)?f(V(u.a,(Qh(),GR)),17).a:0}function wEt(e,t){var i,a,u,l,g,m,v,S;for(S=ke(Oe(V(t,(ot(),rF)))),v=e[0].n.a+e[0].o.a+e[0].d.c+S,m=1;m<e.length;m++)a=e[m].n,u=e[m].o,i=e[m].d,l=a.a-i.b-v,l<0&&(a.a-=l),g=t.f,g.a=h.Math.max(g.a,a.a+u.a),v=a.a+u.a+i.c+S}function f4n(e,t){var i,a,u,l,g,m;return a=f(f(bn(e.g,t.a),47).a,68),u=f(f(bn(e.g,t.b),47).a,68),l=a.b,g=u.b,i=wMn(l,g),i>=0?i:(m=Z7(Ho(new Ke(g.c+g.b/2,g.d+g.a/2),new Ke(l.c+l.b/2,l.d+l.a/2))),-(KSt(l,g)-1)*m)}function h4n(e,t,i){var a;Zi(new Ft(null,(!i.a&&(i.a=new Be(Ii,i,6,6)),new Bt(i.a,16))),new Iat(e,t)),Zi(new Ft(null,(!i.n&&(i.n=new Be(hc,i,1,7)),new Bt(i.n,16))),new Oat(e,t)),a=f(Fe(i,(Rr(),zA)),77),a&&Yxe(a,e,t)}function GS(e,t,i){var a,u,l;if(l=iA((Rl(),Ea),e.zh(),t),l)return is(),f(l,69).uk()||(l=rR(Ns(Ea,l))),u=(a=e.Eh(l),f(a>=0?e.Hh(a,!0,!0):GS(e,l,!0),159)),f(u,218).Pl(t,i);throw ie(new dn(vw+t.ve()+rpe))}function VAe(e,t,i,a){var u,l,g,m,v;if(u=e.d[t],u){if(l=u.g,v=u.i,a!=null){for(m=0;m<v;++m)if(g=f(l[m],135),g.xi()==i&&ei(a,g.jd()))return g}else for(m=0;m<v;++m)if(g=f(l[m],135),Ae(g.jd())===Ae(a))return g}return null}function p4n(e,t){var i,a,u,l,g;for(a=(!t.s&&(t.s=new Be(El,t,21,17)),t.s),l=null,u=0,g=a.i;u<g;++u)switch(i=f(be(a,u),178),kS(Ns(e,i))){case 4:case 5:case 6:{!l&&(l=new et),vn(l.c,i);break}}return l||(Ht(),Ht(),Ga)}function SG(e,t){var i;if(t<0)throw ie(new fY("Negative exponent"));if(t==0)return Oee;if(t==1||e3e(e,Oee)||e3e(e,M$))return e;if(!WEt(e,0)){for(i=1;!WEt(e,i);)++i;return M3($Cn(i*t),SG(HTe(e,i),t))}return FOn(e,t)}function g4n(e,t){var i,a,u;if(Ae(e)===Ae(t))return!0;if(e==null||t==null||e.length!=t.length)return!1;for(i=0;i<e.length;++i)if(a=e[i],u=t[i],!(Ae(a)===Ae(u)||a!=null&&ei(a,u)))return!1;return!0}function yEt(e){Q2e();var t,i,a;for(this.b=NOt,this.c=(ko(),$g),this.f=(Tot(),OOt),this.a=e,$2e(this,new fv),tQ(this),a=new ne(e.b);a.a<a.c.c.length;)i=f(ee(a),82),i.d||(t=new Cle(ce(se(fge,1),Jt,82,0,[i])),Qe(e.a,t))}function m4n(e,t,i){var a,u,l,g,m,v;if(!e||e.c.length==0)return null;for(l=new vft(t,!i),u=new ne(e);u.a<u.c.c.length;)a=f(ee(u),72),kAe(l,(h7(),new nY(a)));return g=l.i,g.a=(v=l.n,l.e.b+v.d+v.a),g.b=(m=l.n,l.e.a+m.b+m.c),l}function EEt(e){var t,i,a,u,l,g,m;for(m=EH(e.a),MSe(m,new ZM),i=null,u=m,l=0,g=u.length;l<g&&(a=u[l],a.k==(cn(),bo));++l)t=f(V(a,(Ge(),Tc)),64),!(t!=(tt(),pn)&&t!=_n)&&(i&&f(V(i,XR),15).Ec(a),i=a)}function b4n(e,t,i){var a,u,l,g,m,v,S;v=(Gt(t,e.c.length),f(e.c[t],335)),_0(e,t),v.b/2>=i&&(a=t,S=(v.c+v.a)/2,g=S-i,v.c<=S-i&&(u=new Kse(v.c,g),Ky(e,a++,u)),m=S+i,m<=v.a&&(l=new Kse(m,v.a),fT(a,e.c.length),y7(e.c,a,l)))}function SEt(e,t,i){var a,u,l,g,m,v;if(!t.dc()){for(u=new ki,v=t.Jc();v.Ob();)for(m=f(v.Pb(),39),qr(e.a,rt(m.g),rt(i)),g=(a=tr(new u1(m).a.d,0),new m3(a));kz(g.a);)l=f(Zn(g.a),65).c,Wi(u,l,u.c.b,u.c);SEt(e,u,i+1)}}function WAe(e){var t;if(!e.c&&e.g==null)e.d=e.$i(e.f),Qn(e,e.d),t=e.d;else{if(e.g==null)return!0;if(e.i==0)return!1;t=f(e.g[e.i-1],49)}return t==e.b&&null.Sm>=null.Rm()?(XZ(e),WAe(e)):t.Ob()}function _Et(e){if(this.a=e,e.c.i.k==(cn(),bo))this.c=e.c,this.d=f(V(e.c.i,(Ge(),Tc)),64);else if(e.d.i.k==bo)this.c=e.d,this.d=f(V(e.d.i,(Ge(),Tc)),64);else throw ie(new dn("Edge "+e+" is not an external edge."))}function kEt(e,t){var i,a,u;u=e.b,e.b=t,(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,3,u,e.b)),t?t!=e&&(ju(e,t.zb),Iue(e,t.d),i=(a=t.c,a??t.zb),Nue(e,i==null||Mt(i,t.zb)?null:i)):(ju(e,null),Iue(e,0),Nue(e,null))}function TEt(e,t){var i;this.e=(_S(),or(e),_S(),s3e(e)),this.c=(or(t),s3e(t)),QSe(this.e.Pd().dc()==this.c.Pd().dc()),this.d=P0t(this.e),this.b=P0t(this.c),i=iT(Ro,[Ye,Jt],[5,1],5,[this.e.Pd().gc(),this.c.Pd().gc()],2),this.a=i,kkn(this)}function xEt(e){!zpe&&(zpe=G9n());var t=e.replace(/[\x00-\x1f\xad\u0600-\u0603\u06dd\u070f\u17b4\u17b5\u200b-\u200f\u2028-\u202e\u2060-\u2064\u206a-\u206f\ufeff\ufff9-\ufffb"\\]/g,function(i){return tEn(i)});return'"'+t+'"'}function KAe(e,t,i,a,u,l){var g,m,v,S,T;if(u!=0)for(Ae(e)===Ae(i)&&(e=e.slice(t,t+u),t=0),v=i,m=t,S=t+u;m<S;)g=h.Math.min(m+1e4,S),u=g-m,T=e.slice(m,g),T.splice(0,0,a,l?u:0),Array.prototype.splice.apply(v,T),m=g,a+=u}function CEt(e){aTe();var t,i;for(this.b=uIt,this.c=dIt,this.g=(kot(),cIt),this.d=(ko(),$g),this.a=e,wIe(this),i=new ne(e.b);i.a<i.c.c.length;)t=f(ee(i),60),!t.a&&gct(obt(new d2e,ce(se(Lee,1),Jt,60,0,[t])),e),t.e=new ZX(t.d)}function v4n(e){var t,i,a,u,l,g;for(u=e.e.c.length,a=Pe(qf,BT,15,u,0,1),g=new ne(e.e);g.a<g.c.c.length;)l=f(ee(g),151),a[l.a]=new ki;for(i=new ne(e.c);i.a<i.c.c.length;)t=f(ee(i),288),a[t.c.a].Ec(t),a[t.d.a].Ec(t);return a}function w4n(e,t){var i,a,u,l,g;if(i=f(fn(e.a,4),128),g=i==null?0:i.length,t>=g)throw ie(new rT(t,g));return u=i[t],g==1?a=null:(a=Pe(g1e,ype,412,g-1,0,1),au(i,0,a,0,t),l=g-t-1,l>0&&au(i,t+1,a,t,l)),aL(e,a),Wyt(e,t,u),u}function AEt(e){var t,i;if(e.f){for(;e.n<e.o;){if(t=f(e.j?e.j.Xi(e.n):e.k.Xb(e.n),74),i=t.Ik(),me(i,101)&&(f(i,19).Bb&Dc)!=0&&(!e.e||i.mk()!=$D||i.Ij()!=0)&&t.kd()!=null)return!0;++e.n}return!1}else return e.n<e.o}function _R(){_R=Z,f4=f(be(je((z2e(),Mc).qb),6),35),d4=f(be(je(Mc.qb),3),35),E1e=f(be(je(Mc.qb),4),35),S1e=f(be(je(Mc.qb),5),19),jZ(f4),jZ(d4),jZ(E1e),jZ(S1e),EMt=new bu(ce(se(El,1),pA,178,0,[f4,d4]))}function IEt(e,t){var i;this.d=new i7,this.b=t,this.e=new ja(t.Jf()),i=e.u.Gc((Xl(),hV)),e.u.Gc(M1)?e.F?this.a=i&&!t._f():this.a=!0:e.u.Gc(Lw)?i?this.a=!(t.Sf().Jc().Ob()||t.Uf().Jc().Ob()):this.a=!1:this.a=!1}function OEt(e,t){var i,a,u,l;for(i=e.o.a,l=f(f(si(e.r,t),22),83).Jc();l.Ob();)u=f(l.Pb(),115),u.e.a=(a=u.b,a.nf((Rr(),Sd))?a.$f()==(tt(),pn)?-a.Kf().a-ke(Oe(a.mf(Sd))):i+ke(Oe(a.mf(Sd))):a.$f()==(tt(),pn)?-a.Kf().a:i)}function NEt(e){var t,i,a,u,l,g,m;if(a=e.a.c.length,a>0)for(g=e.c.d,m=e.d.d,u=qv(Ho(new Ke(m.a,m.b),g),1/(a+1)),l=new Ke(g.a,g.b),i=new ne(e.a);i.a<i.c.c.length;)t=f(ee(i),249),t.d.a=l.a+u.a,t.d.b=l.b+u.b,l.a+=u.a,l.b+=u.b}function REt(e,t){var i,a,u,l;i=f(V(e,(ot(),mf)),86),l=f(Fe(t,vD),64),u=f(V(e,Xi),100),u!=(oa(),Fg)&&u!=Pw?l==(tt(),fc)&&(l=QIe(t,i),l==fc&&(l=mR(i))):(a=R_t(t),a>0?l=mR(i):l=iG(mR(i))),pi(t,vD,l)}function PEt(e,t){var i,a;if(e.c.length!=0){if(e.c.length==2)TR((Gt(0,e.c.length),f(e.c[0],9)),(df(),Rb)),TR((Gt(1,e.c.length),f(e.c[1],9)),Nw);else for(a=new ne(e);a.a<a.c.c.length;)i=f(ee(a),9),TR(i,t);e.c.length=0}}function y4n(e,t){var i;t.Sg("Partition preprocessing",1),i=f($d(Jr(Zs(Jr(new Ft(null,new Bt(e.a,16)),new c8),new fO),new vy),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),Zi(i.Mc(),new IU),t.Tg()}function E4n(e,t){var i,a,u,l,g;for(g=e.j,t.a!=t.b&&na(g,new vC),u=g.c.length/2|0,a=0;a<u;a++)l=(Gt(a,g.c.length),f(g.c[a],112)),l.c&&No(l.d,t.a);for(i=u;i<g.c.length;i++)l=(Gt(i,g.c.length),f(g.c[i],112)),l.c&&No(l.d,t.b)}function S4n(e,t,i){var a,u,l;return a=e.c[t.c.p][t.p],u=e.c[i.c.p][i.p],a.a!=null&&u.a!=null?(l=lce(a.a,u.a),l<0?xG(e,t,i):l>0&&xG(e,i,t),l):a.a!=null?(xG(e,t,i),-1):u.a!=null?(xG(e,i,t),1):0}function LEt(e){Vce();var t,i,a,u,l,g,m;for(i=new Zv,u=new ne(e.e.b);u.a<u.c.c.length;)for(a=f(ee(u),25),g=new ne(a.a);g.a<g.c.c.length;)l=f(ee(g),9),m=e.g[l.p],t=f(Eg(i,m),15),t||(t=new et,pE(i,m,t)),t.Ec(l);return i}function _4n(e,t){var i,a,u,l,g;for(u=t.b.b,e.a=Pe(qf,BT,15,u,0,1),e.b=Pe(kd,Tb,30,u,16,1),g=tr(t.b,0);g.b!=g.d.c;)l=f(Zn(g),39),e.a[l.g]=new ki;for(a=tr(t.a,0);a.b!=a.d.c;)i=f(Zn(a),65),e.a[i.b.g].Ec(i),e.a[i.c.g].Ec(i)}function DEt(e,t){var i,a,u,l;e.Mj()?(i=e.Bj(),l=e.Nj(),++e.j,e.nj(i,e.Wi(i,t)),a=e.Fj(3,null,t,i,l),e.Jj()?(u=e.Kj(t,null),u?(u.kj(a),u.lj()):e.Gj(a)):e.Gj(a)):(gdt(e,t),e.Jj()&&(u=e.Kj(t,null),u&&u.lj()))}function YAe(e,t,i){var a,u,l;e.Mj()?(l=e.Nj(),eG(e,t,i),a=e.Fj(3,null,i,t,l),e.Jj()?(u=e.Kj(i,null),e.Qj()&&(u=e.Rj(i,u)),u?(u.kj(a),u.lj()):e.Gj(a)):e.Gj(a)):(eG(e,t,i),e.Jj()&&(u=e.Kj(i,null),u&&u.lj()))}function KZ(e,t){var i,a,u,l,g;for(g=Ju(e.e.zh(),t),u=new J2,i=f(e.g,122),l=e.i;--l>=0;)a=i[l],g.Zl(a.Ik())&&Qn(u,a);!Ckt(e,u)&&Ld(e.e)&&Q5(e,t.Gk()?Yv(e,6,t,(Ht(),Ga),null,-1,!1):Yv(e,t.qk()?2:1,t,null,null,-1,!1))}function k4n(e,t){var i,a,u,l,g;return e.a==(fL(),q$)?!0:(l=t.a.c,i=t.a.c+t.a.b,!(t.j&&(a=t.A,g=a.c.c.a-a.o.a/2,u=l-(a.n.a+a.o.a),u>g)||t.q&&(a=t.C,g=a.c.c.a-a.o.a/2,u=a.n.a-i,u>g)))}function MEt(e,t,i){var a,u,l,g,m,v;for(a=0,v=i,t||(a=i*(e.c.length-1),v*=-1),l=new ne(e);l.a<l.c.c.length;){for(u=f(ee(l),9),$e(u,(ot(),Nm),(vb(),vne)),u.o.a=a,m=L0(u,(tt(),_n)).Jc();m.Ob();)g=f(m.Pb(),11),g.n.a=a;a+=v}}function jEt(e){var t;return(e.Db&64)!=0?Jh(e):(t=new sh(Jh(e)),t.a+=" (startX: ",v3(t,e.j),t.a+=", startY: ",v3(t,e.k),t.a+=", endX: ",v3(t,e.b),t.a+=", endY: ",v3(t,e.c),t.a+=", identifier: ",ys(t,e.d),t.a+=")",t.a)}function XAe(e){var t;return(e.Db&64)!=0?_j(e):(t=new sh(_j(e)),t.a+=" (ordered: ",m0(t,(e.Bb&256)!=0),t.a+=", unique: ",m0(t,(e.Bb&512)!=0),t.a+=", lowerBound: ",Nae(t,e.s),t.a+=", upperBound: ",Nae(t,e.t),t.a+=")",t.a)}function $Et(e,t,i,a,u,l,g,m){var v;return me(e.Cb,88)&&OT(pd(f(e.Cb,88)),4),ju(e,i),e.f=a,eL(e,u),tL(e,l),QP(e,g),nL(e,!1),N0(e,!0),rL(e,m),O0(e,!0),x0(e,0),e.b=0,vT(e,1),v=xg(e,t,null),v&&v.lj(),gle(e,!1),e}function FEt(e,t){var i,a,u,l;return i=f(wu(e.a,t),509),i||(a=new iue(t),u=(mJ(),Lg?null:a.c),l=uh(u,0,h.Math.max(0,sX(u,gl(46)))),e1n(a,FEt(e,l)),(Lg?null:a.c).length==0&&Gct(a,new Ca),Os(e.a,Lg?null:a.c,a),a)}function BEt(){this.a=new YC,this.n=new NZ,this.p=new NZ,this.c=new Wn,this.f=new NZ,this.o=new NZ,this.q=new Wn,this.d=new Wn,this.g=new Wn,this.k=new Wn,this.e=new Wn,this.i=new Wn,this.j=new Wn,this.r=new Wn,this.b=new Wn}function kR(e,t,i){var a,u,l,g,m,v;return m=e.ml(i),m!=i?(g=e.g[t],v=m,j7(e,t,e.Wi(t,v)),l=g,e.Oi(t,v,l),e.Zk()&&(a=i,u=e.Lj(a,null),!f(m,52).Lh()&&(u=e.Kj(v,u)),u&&u.lj()),Ld(e.e)&&Q5(e,e.Fj(9,i,m,t,!1)),m):i}function T4n(e,t){var i,a,u,l;for(a=new ne(e.a.a);a.a<a.c.c.length;)i=f(ee(a),194),i.g=!0;for(l=new ne(e.a.b);l.a<l.c.c.length;)u=f(ee(l),82),u.k=lt(ut(e.e.Kb(new Ya(u,t)))),u.d.g=u.d.g&lt(ut(e.e.Kb(new Ya(u,t))));return e}function UEt(e,t,i){var a,u,l,g,m;if(!e.d[i.p]){for(u=new Tn(Cn(ao(i).a.Jc(),new B));Kn(u);){for(a=f(Dn(u),18),m=a.d.i,g=new Tn(Cn(Fo(m).a.Jc(),new B));Kn(g);)l=f(Dn(g),18),l.c.i==t&&(e.a[l.p]=!0);UEt(e,t,m)}e.d[i.p]=!0}}function x4n(e,t){var i,a,u,l,g,m,v;if(a=w0t(e.Db&254),a==1)e.Eb=null;else if(l=uR(e.Eb),a==2)u=Wle(e,t),e.Eb=l[u==0?1:0];else{for(g=Pe(Ro,Jt,1,a-1,5,1),i=2,m=0,v=0;i<=128;i<<=1)i==t?++m:(e.Db&i)!=0&&(g[v++]=l[m++]);e.Eb=g}e.Db&=~t}function JAe(e){var t;switch(t=0,e){case 105:t=2;break;case 109:t=8;break;case 115:t=4;break;case 120:t=16;break;case 117:t=32;break;case 119:t=64;break;case 70:t=256;break;case 72:t=128;break;case 88:t=512;break;case 44:t=ep}return t}function C4n(e,t,i,a,u){var l,g,m,v;if(Ae(e)===Ae(t)&&a==u){BSt(e,a,i);return}for(m=0;m<a;m++){for(g=0,l=e[m],v=0;v<u;v++)g=Ua(Ua(Na(ia(l,ls),ia(t[v],ls)),ia(i[m+v],ls)),ia(mr(g),ls)),i[m+v]=mr(g),g=Jy(g,32);i[m+u]=mr(g)}}function zEt(e){var t,i,a,u,l;if(i=(t=f(Tp(Ja),10),new Rf(t,f(Gh(t,t.length),10),0)),l=f(V(e,(Ge(),ed)),9),l)for(u=new ne(l.j);u.a<u.c.c.length;)a=f(ee(u),11),Ae(V(a,oi))===Ae(e)&&F7(new yg(a.b))&&Yh(i,a.j);return i}function HEt(e,t){t.b!=0&&(isNaN(e.s)?e.s=ke((Ln(t.b!=0),Oe(t.a.a.c))):e.s=h.Math.min(e.s,ke((Ln(t.b!=0),Oe(t.a.a.c)))),isNaN(e.c)?e.c=ke((Ln(t.b!=0),Oe(t.c.b.c))):e.c=h.Math.max(e.c,ke((Ln(t.b!=0),Oe(t.c.b.c)))))}function Uj(e){var t,i,a,u;for(t=null,a=pb(Df(ce(se(Gf,1),Jt,20,0,[(!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c)])));Kn(a);)if(i=f(Dn(a),84),u=Us(i),!t)t=so(u);else if(t!=so(u))return!0;return!1}function Ede(e,t){var i,a,u,l;e.Mj()?(i=e.i,l=e.Nj(),TH(e,t),a=e.Fj(3,null,t,i,l),e.Jj()?(u=e.Kj(t,null),e.Qj()&&(u=e.Rj(t,u)),u?(u.kj(a),u.lj()):e.Gj(a)):e.Gj(a)):(TH(e,t),e.Jj()&&(u=e.Kj(t,null),u&&u.lj()))}function A4n(e,t){var i,a,u,l;if(l=fH(e.a,t.b),!l)throw ie(new As("Invalid hitboxes for scanline overlap calculation."));for(u=!1,a=e.a.a.ec().Jc();a.Ob();)if(i=f(a.Pb(),68),jTn(t.b,i))qpn(e.b.a,t.b,i),u=!0;else if(u)break}function I4n(e){var t;if(!e.a)throw ie(new As("IDataType class expected for layout option "+e.f));if(t=Q2n(e.a),t==null)throw ie(new As("Couldn't create new instance of property '"+e.f+"'. "+Qxt+(d1(_V),_V.k)+S4e));return f(t,411)}function Sde(e){var t,i,a,u,l;return l=e.Lh(),l&&l.Rh()&&(u=iw(e,l),u!=l)?(i=e.Bh(),a=(t=e.Bh(),t>=0?e.wh(null):e.Lh().Ph(e,-1-t,null,null)),e.xh(f(u,52),i),a&&a.lj(),e.rh()&&e.sh()&&i>-1&&Zr(e,new Jo(e,9,i,l,u)),u):l}function ZAe(e,t){var i,a,u,l,g;for(l=e.b.Ae(t),a=(i=e.a.get(l),i??Pe(Ro,Jt,1,0,5,1)),g=0;g<a.length;g++)if(u=a[g],e.b.ze(t,u.jd()))return a.length==1?(a.length=0,Rbn(e.a,l)):a.splice(g,1),--e.c,++e.b.g,u.kd();return null}function GEt(e){var t,i,a,u,l,g,m,v;for(g=0,l=e.f.e,a=0;a<l.c.length;++a)for(m=(Gt(a,l.c.length),f(l.c[a],151)),u=a+1;u<l.c.length;++u)v=(Gt(u,l.c.length),f(l.c[u],151)),i=Qv(m.d,v.d),t=i-e.a[m.a][v.a],g+=e.i[m.a][v.a]*t*t;return g}function O4n(e,t){var i;if(!li(t,(ot(),el))&&(i=fIn(f(V(t,iPe),364),f(V(e,el),171)),$e(t,iPe,i),!Kn(new Tn(Cn(T1(t).a.Jc(),new B)))))switch(i.g){case 1:$e(t,el,(Ff(),X$));break;case 2:$e(t,el,(Ff(),J$))}}function N4n(e,t){var i;$Pn(e),e.a=(i=new Cae,Zi(new Ft(null,new Bt(t.d,16)),new zet(i)),i),ZPn(e,f(V(t.b,(ot(),pme)),345)),iAn(e),$4n(e),mIn(e),oAn(e),BMn(e,t),Zi(Zs(new Ft(null,Sht(Twn(e.b).a)),new cl),new vs),t.a=!1,e.a=null}function QAe(){QAe=Z,x5t=new Pt(DRe,(mn(),!1)),C5t=new Pt(MRe,7),rt(0),N5t=new Pt(jRe,rt(0)),I5t=new Pt($Re,rt(-1)),dMe=(wj(),nbe),R5t=new Pt(FRe,dMe),uMe=(yj(),Gq),A5t=new Pt(BRe,uMe),lMe=(uJ(),rbe),O5t=new Pt(URe,lMe)}function qEt(){mAe.call(this,WL,(Bv(),$Dt)),this.p=null,this.a=null,this.f=null,this.n=null,this.g=null,this.c=null,this.i=null,this.j=null,this.d=null,this.b=null,this.e=null,this.k=null,this.o=null,this.s=null,this.q=!1,this.r=!1}function pL(){pL=Z,h1e=new $N(pxt,0),Mne=new $N("INSIDE_SELF_LOOPS",1),jne=new $N("MULTI_EDGES",2),Dne=new $N("EDGE_LABELS",3),f1e=new $N("PORTS",4),Lne=new $N("COMPOUND",5),Pne=new $N("CLUSTERS",6),d1e=new $N("DISCONNECTED",7)}function VEt(e,t,i){var a,u,l;e.Mj()?(l=e.Nj(),++e.j,e.nj(t,e.Wi(t,i)),a=e.Fj(3,null,i,t,l),e.Jj()?(u=e.Kj(i,null),u?(u.kj(a),u.lj()):e.Gj(a)):e.Gj(a)):(++e.j,e.nj(t,e.Wi(t,i)),e.Jj()&&(u=e.Kj(i,null),u&&u.lj()))}function WEt(e,t){var i,a,u;if(t==0)return(e.a[0]&1)!=0;if(t<0)throw ie(new fY("Negative bit address"));if(u=t>>5,u>=e.d)return e.e<0;if(i=e.a[u],t=1<<(t&31),e.e<0){if(a=Bbt(e),u<a)return!1;a==u?i=-i:i=~i}return(i&t)!=0}function KEt(e){var t,i,a,u,l;if(u=new et,l=GSt(e,u),t=f(V(e,(Ge(),ed)),9),t)for(a=new ne(t.j);a.a<a.c.c.length;)i=f(ee(a),11),Ae(V(i,oi))===Ae(e)&&(l=h.Math.max(l,GSt(i,u)));return u.c.length==0||$e(e,TA,l),l!=-1?u:null}function R4n(e,t,i,a){var u;f(i.b,68),f(i.b,68),f(a.b,68),f(a.b,68),u=Ho($a(f(i.b,68).c),f(a.b,68).c),wX(u,tEt(f(i.b,68),f(a.b,68),u)),f(a.b,68),f(a.b,68),f(a.b,68).c.a+u.a,f(a.b,68).c.b+u.b,f(a.b,68),Mu(a.a,new J_e(e,t,a))}function eIe(e,t){var i,a,u,l,g,m,v;if(l=t.e,l){for(i=Sde(l),a=f(e.g,674),g=0;g<e.i;++g)if(v=a[g],Ule(v)==i&&(u=(!v.d&&(v.d=new So(gs,v,1)),v.d),m=f(i.Ih(Dde(l,l.Cb,l.Db>>16)),15).bd(l),m<u.i))return eIe(e,f(be(u,m),87))}return t}function N(e,t,i){var a=xee,u,l=a[e],g=l instanceof Array?l[0]:null;l&&!g?p=l:(p=(u=t&&t.prototype,!u&&(u=xee[t]),iEn(u)),p.Pm=i,!t&&(p.Qm=xe),a[e]=p);for(var m=3;m<arguments.length;++m)arguments[m].prototype=p;g&&(p.Om=g)}function Kn(e){for(var t;!f(or(e.a),49).Ob();){if(e.d=exn(e),!e.d)return!1;if(e.a=f(e.d.Pb(),49),me(e.a,37)){if(t=f(e.a,37),e.a=t.a,!e.b&&(e.b=new O3),Jv(e.b,e.d),t.b)for(;!u7(t.b);)Jv(e.b,f(uwn(t.b),49));e.d=t.d}}return!0}function tIe(e,t){var i,a,u,l;for(u=1,t.j=!0,l=null,a=new ne(V3(t));a.a<a.c.c.length;)i=f(ee(a),216),e.c[i.c]||(e.c[i.c]=!0,l=hZ(i,t),i.f?u+=tIe(e,l):!l.j&&i.a==i.e.e-i.d.e&&(i.f=!0,Go(e.p,i),u+=tIe(e,l)));return u}function P4n(e){var t,i,a;for(i=new ne(e.a.a.b);i.a<i.c.c.length;)t=f(ee(i),82),a=(an(0),0),a>0&&(!(l1(e.a.c)&&t.n.d)&&!(k3(e.a.c)&&t.n.b)&&(t.g.d+=h.Math.max(0,a/2-.5)),!(l1(e.a.c)&&t.n.a)&&!(k3(e.a.c)&&t.n.c)&&(t.g.a-=a-1))}function YEt(e,t,i){var a,u,l,g,m,v;l=f(ct(t.e,0),18).c,a=l.i,u=a.k,v=f(ct(i.g,0),18).d,g=v.i,m=g.k,u==(cn(),uo)?$e(e,(Ge(),$p),f(V(a,$p),11)):$e(e,(Ge(),$p),l),m==uo?$e(e,(Ge(),yh),f(V(g,yh),11)):$e(e,(Ge(),yh),v)}function XEt(e,t){var i,a,u,l,g,m;for(l=new ne(e.b);l.a<l.c.c.length;)for(u=f(ee(l),25),m=new ne(u.a);m.a<m.c.c.length;)for(g=f(ee(m),9),g.k==(cn(),Vc)&&TR(g,t),a=new Tn(Cn(ao(g).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),l1t(i,t)}function L4n(e,t){var i,a,u;for(t.Sg("Layer constraint preprocessing",1),i=new et,u=new la(e.a,0);u.b<u.d.gc();)a=(Ln(u.b<u.d.gc()),f(u.d.Xb(u.c=u.b++),9)),ITn(a)&&(AOn(a),vn(i.c,a),hd(u));i.c.length==0||$e(e,(Ge(),Zge),i),t.Tg()}function nIe(e){var t,i,a;this.c=e,a=f(V(e,(ot(),mf)),86),t=ke(Oe(V(e,Tte))),i=ke(Oe(V(e,sDe))),a==(ko(),Ds)||a==Gs||a==$g?this.b=t*i:this.b=1/(t*i),this.j=ke(Oe(V(e,NA))),this.e=ke(Oe(V(e,v_))),this.f=e.b.c.length}function D4n(e){var t,i;for(e.e=Pe(pr,Pr,30,e.p.c.length,15,1),e.k=Pe(pr,Pr,30,e.p.c.length,15,1),i=new ne(e.p);i.a<i.c.c.length;)t=f(ee(i),9),e.e[t.p]=g1(new Tn(Cn(Fo(t).a.Jc(),new B))),e.k[t.p]=g1(new Tn(Cn(ao(t).a.Jc(),new B)))}function M4n(e){var t,i,a,u,l,g;for(u=0,e.q=new et,t=new Mo,g=new ne(e.p);g.a<g.c.c.length;){for(l=f(ee(g),9),l.p=u,a=new Tn(Cn(ao(l).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),Go(t,i.d.i);t.a.Ac(l)!=null,Qe(e.q,new H7(t)),t.a.$b(),++u}}function rIe(e,t){var i,a,u,l,g;return t&=63,i=e.h,a=(i&zG)!=0,a&&(i|=-1048576),t<22?(g=i>>t,l=e.m>>t|i<<22-t,u=e.l>>t|e.m<<22-t):t<44?(g=a?A1:0,l=i>>t-22,u=e.m>>t-22|i<<44-t):(g=a?A1:0,l=a?Ed:0,u=i>>t-44),Ku(u&Ed,l&Ed,g&A1)}function JEt(e,t){var i,a,u,l,g,m,v,S,T;if(e.a.f>0&&me(t,43)&&(e.a.Yj(),S=f(t,43),v=S.jd(),l=v==null?0:Ti(v),g=C_e(e.a,l),i=e.a.d[g],i)){for(a=f(i.g,371),T=i.i,m=0;m<T;++m)if(u=a[m],u.xi()==l&&u.Fb(S))return JEt(e,S),!0}return!1}function j4n(e){var t,i,a,u,l,g,m;if(t=e.mi(bh),t&&(m=dr(Rp((!t.b&&(t.b=new Md((Ut(),Za),xc,t)),t.b),"settingDelegates")),m!=null)){for(i=new et,u=DT(m,"\\w+"),l=0,g=u.length;l<g;++l)a=u[l],vn(i.c,a);return i}return Ht(),Ht(),Ga}function $4n(e){var t,i,a,u;for(u=f(si(e.a,(CT(),ute)),15).Jc();u.Ob();)a=f(u.Pb(),105),i=(t=YN(a.k),t.Gc((tt(),hn))?t.Gc(_n)?t.Gc(Bn)?t.Gc(pn)?null:hNt:gNt:pNt:fNt),SP(e,a,i[0],(LS(),s_),0),SP(e,a,i[1],_w,1),SP(e,a,i[2],c_,1)}function F4n(e,t){var i,a;i=lLn(t),Q5n(e,t,i),Jwt(e.a,f(V(ra(t.b),(Ge(),ZR)),233)),OLn(e),EOn(e,t),a=Pe(pr,Pr,30,t.b.j.c.length,15,1),tfe(e,t,(tt(),hn),a,i),tfe(e,t,_n,a,i),tfe(e,t,Bn,a,i),tfe(e,t,pn,a,i),e.a=null,e.c=null,e.b=null}function iIe(e,t,i){switch(t){case 7:!e.e&&(e.e=new Qt(vo,e,7,4)),Jn(e.e),!e.e&&(e.e=new Qt(vo,e,7,4)),Qi(e.e,f(i,16));return;case 8:!e.d&&(e.d=new Qt(vo,e,8,5)),Jn(e.d),!e.d&&(e.d=new Qt(vo,e,8,5)),Qi(e.d,f(i,16));return}J3e(e,t,i)}function oIe(e,t){var i,a,u,l,g;if(Ae(t)===Ae(e))return!0;if(!me(t,15)||(g=f(t,15),e.gc()!=g.gc()))return!1;for(l=g.Jc(),a=e.Jc();a.Ob();)if(i=a.Pb(),u=l.Pb(),!(Ae(i)===Ae(u)||i!=null&&ei(i,u)))return!1;return!0}function B4n(e,t){var i,a,u,l;for(l=f($d(Zs(Zs(new Ft(null,new Bt(t.b,16)),new by),new QM),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),l.Ic(new y6),i=0,u=l.Jc();u.Ob();)a=f(u.Pb(),11),a.p==-1&&sIe(e,a,i++)}function U4n(e,t){var i,a,u,l,g;for(t.Sg("Port side processing",1),g=new ne(e.a);g.a<g.c.c.length;)u=f(ee(g),9),g_t(u);for(a=new ne(e.b);a.a<a.c.c.length;)for(i=f(ee(a),25),l=new ne(i.a);l.a<l.c.c.length;)u=f(ee(l),9),g_t(u);t.Tg()}function ZEt(e){switch(e.g){case 0:return new fN;case 1:return new F9;case 2:return new cb;case 3:return new uat;case 4:return new Olt;default:throw ie(new dn("No implementation is available for the node placer "+(e.f!=null?e.f:""+e.g)))}}function QEt(e){switch(e.g){case 0:return new _ke;case 1:return new i1;case 2:return new M9;case 3:return new D9;case 4:return new Qat;default:throw ie(new dn("No implementation is available for the cycle breaker "+(e.f!=null?e.f:""+e.g)))}}function z4n(e,t){var i,a,u,l,g,m;for(a=new Zv,g=fJ(new bu(e.g)),l=g.a.ec().Jc();l.Ob();){if(u=f(l.Pb(),9),!u){t._g("There are no classes in a balanced layout.");break}m=e.j[u.p],i=f(Eg(a,m),15),i||(i=new et,pE(a,m,i)),i.Ec(u)}return a}function H4n(e,t){var i,a,u,l,g;a=new ki,Wi(a,t,a.c.b,a.c);do for(i=(Ln(a.b!=0),f(lh(a,a.a.a),39)),e.b[i.g]=1,l=tr(i.d,0);l.b!=l.d.c;)u=f(Zn(l),65),g=u.c,e.b[g.g]==1?Mr(e.a,u):e.b[g.g]==2?e.b[g.g]=1:Wi(a,g,a.c.b,a.c);while(a.b!=0)}function aIe(e,t){if(me(t,206))return so(f(t,27));if(me(t,193))return bg(f(t,124));if(me(t,359))return gTe(f(t,153));if(me(t,268))return dbn(e,f(t,85));throw ie(new dn(KL+Cg(new bu(ce(se(Ro,1),Jt,1,5,[t])))))}function G4n(e,t,i){var a;a=null,t&&(a=t.d),Dj(e,new lP(t.n.a-a.b+i.a,t.n.b-a.d+i.b)),Dj(e,new lP(t.n.a-a.b+i.a,t.n.b+t.o.b+a.a+i.b)),Dj(e,new lP(t.n.a+t.o.a+a.c+i.a,t.n.b-a.d+i.b)),Dj(e,new lP(t.n.a+t.o.a+a.c+i.a,t.n.b+t.o.b+a.a+i.b))}function sIe(e,t,i){var a,u,l;for(t.p=i,l=pb(Df(ce(se(Gf,1),Jt,20,0,[new Z5(t),new _N(t)])));Kn(l);)a=f(Dn(l),11),a.p==-1&&sIe(e,a,i);if(t.i.k==(cn(),uo))for(u=new ne(t.i.j);u.a<u.c.c.length;)a=f(ee(u),11),a!=t&&a.p==-1&&sIe(e,a,i)}function e2t(e){var t,i,a,u,l;if(u=f($d(mTe(vTe(e)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),a=$R,u.gc()>=2)for(i=u.Jc(),t=Oe(i.Pb());i.Ob();)l=t,t=Oe(i.Pb()),a=h.Math.min(a,(an(t),t-(an(l),l)));return a}function q4n(e,t){var i,a,u;for(u=new et,a=tr(t.a,0);a.b!=a.d.c;)i=f(Zn(a),65),i.b.g==e.g&&!Mt(i.b.c,QQ)&&Ae(V(i.b,(Ec(),Lm)))!==Ae(V(i.c,Lm))&&!j3(new Ft(null,new Bt(u,16)),new Att(i))&&vn(u.c,i);return na(u,new ag),u}function V4n(e,t){var i,a,u;if(Ae(t)===Ae(or(e)))return!0;if(!me(t,15)||(a=f(t,15),u=e.gc(),u!=a.gc()))return!1;if(me(a,59)){for(i=0;i<u;i++)if(!dm(e.Xb(i),a.Xb(i)))return!1;return!0}else return dAn(e.Jc(),a.Jc())}function W4n(e,t,i,a,u,l){var g,m,v,S;for(m=!oP(Jr(e.Mc(),new Y(new n0))).zd((Wy(),zR)),g=e,l==(ko(),Wf)&&(g=Bd(g)),S=g.Jc();S.Ob();)v=f(S.Pb(),72),v.n.a=t.a,m?v.n.b=t.b+(a.b-v.o.b)/2:u?v.n.b=t.b:v.n.b=t.b+a.b-v.o.b,t.a+=v.o.a+i}function t2t(e,t,i){var a,u,l,g,m,v,S;if(i)for(l=i.a.length,a=new Xy(l),m=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);m.Ob();)g=f(m.Pb(),17),v=OP(i,g.a),v&&(S=mkn(p1(v,lpe),t),qr(e.k,S,v),u=Am in v.a,u&&FP(S,p1(v,Am)),UZ(v,S),jAe(v,S))}function K4n(e,t,i){var a,u,l,g,m;if(m=i,!m&&(m=uke(new CN,0)),m.Sg(WTt,1),X_t(e.c,t),g=j8n(e.a,t),g.gc()==1)O_t(f(g.Xb(0),36),m);else for(l=1/g.gc(),u=g.Jc();u.Ob();){if(a=f(u.Pb(),36),i.Yg())return;O_t(a,m.bh(l))}bpn(e.a,g,t),SPn(t),m.Tg()}function n2t(e,t,i){var a,u,l,g,m;if(u=e.f,!u&&(u=f(e.a.a.ec().Jc().Pb(),60)),Lj(u,t,i),e.a.a.gc()!=1)for(a=t*i,g=e.a.a.ec().Jc();g.Ob();)l=f(g.Pb(),60),l!=u&&(m=iR(l),m.f.d?(l.d.d+=a+Og,l.d.a-=a+Og):m.f.a&&(l.d.a-=a+Og))}function Y4n(e,t,i,a){var u,l,g,m,v,S,T,A;u=i,T=t,l=T;do l=e.a[l.p],m=(A=e.g[l.p],ke(e.p[A.p])+ke(e.d[l.p])-l.d.d),v=wkn(l,a),v&&(g=(S=e.g[v.p],ke(e.p[S.p])+ke(e.d[v.p])+v.o.b+v.d.a),u=h.Math.min(u,m-(g+A3(e.k,l,v))));while(T!=l);return u}function X4n(e,t,i,a){var u,l,g,m,v,S,T,A;u=i,T=t,l=T;do l=e.a[l.p],g=(A=e.g[l.p],ke(e.p[A.p])+ke(e.d[l.p])+l.o.b+l.d.a),v=OTn(l,a),v&&(m=(S=e.g[v.p],ke(e.p[S.p])+ke(e.d[v.p])-v.d.d),u=h.Math.min(u,m-(g+A3(e.k,l,v))));while(T!=l);return u}function r2t(e,t){var i;if(t.Sg("Equal Whitespace Eliminator",1),mm(e,(wb(),une)))Lxn(f(Fe(e,une),15),ke(Oe(Fe(e,t4))),(i=ke(Oe(Fe(e,wF))),ke(Oe(Fe(e,(Ag(),SF)))),i));else throw ie(new h0("The graph does not contain rows."));t.Tg()}function Fe(e,t){var i,a;return a=(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),Rp(e.o,t)),a??(i=t.Qg(),me(i,4)&&(i==null?(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),yZ(e.o,t)):(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),hG(e.o,t,i))),i)}function NT(){NT=Z,sp=new y3("H_LEFT",0),D1=new y3("H_CENTER",1),cp=new y3("H_RIGHT",2),lp=new y3("V_TOP",3),Pb=new y3("V_CENTER",4),Sh=new y3("V_BOTTOM",5),bf=new y3("INSIDE",6),up=new y3("OUTSIDE",7),Eh=new y3("H_PRIORITY",8)}function J4n(e,t){var i,a,u,l,g,m,v;if(!t.f)throw ie(new dn("The input edge is not a tree edge."));for(l=null,u=Qr,a=new ne(e.d);a.a<a.c.c.length;)i=f(ee(a),216),m=i.d,v=i.e,vde(e,m,t)&&!vde(e,v,t)&&(g=v.e-m.e-i.a,g<u&&(u=g,l=i));return l}function Z4n(e){var t,i,a,u,l,g;if(!(e.f.e.c.length<=1)){t=0,u=GEt(e),i=Gi;do{for(t>0&&(u=i),g=new ne(e.f.e);g.a<g.c.c.length;)l=f(ee(g),151),!lt(ut(V(l,(dG(),s5e))))&&(a=eDn(e,l),ii(Cp(l.d),a));i=GEt(e)}while(!Cwn(e,t++,u,i))}}function Q4n(e,t){var i,a,u,l,g;for(l=e.g.a,g=e.g.b,a=new ne(e.d);a.a<a.c.c.length;)i=f(ee(a),72),u=i.n,e.a==(_g(),rD)||e.i==(tt(),_n)?u.a=l:e.a==iD||e.i==(tt(),pn)?u.a=l+e.j.a-i.o.a:u.a=l+(e.j.a-i.o.a)/2,u.b=g,ii(u,t),g+=i.o.b+e.e}function e6n(e,t){var i,a,u,l,g,m,v,S,T,A;S=e,v=ij(S,"individualSpacings"),v&&(a=mm(t,(Rr(),a4)),g=!a,g&&(u=new JO,pi(t,a4,u)),m=f(Fe(t,a4),377),A=v,l=null,A&&(l=(T=Due(A,Pe(ft,Ye,2,0,6,1)),new Pae(A,T))),l&&(i=new Bat(A,m),Oa(l,i)))}function t6n(e,t){var i,a,u,l,g,m,v,S,T,A,L;return v=null,A=e,T=null,(qCt in A.a||VCt in A.a||see in A.a)&&(S=null,L=oCe(t),g=ij(A,qCt),i=new fnt(L),$0t(i.a,g),m=ij(A,VCt),a=new Ent(L),F0t(a.a,m),l=IS(A,see),u=new knt(L),S=(qwt(u.a,l),l),T=S),v=T,v}function n6n(e,t){var i,a,u;if(t===e)return!0;if(me(t,537)){if(u=f(t,828),e.a.d!=u.a.d||P3(e).gc()!=P3(u).gc())return!1;for(a=P3(u).Jc();a.Ob();)if(i=f(a.Pb(),413),Kft(e,i.a.jd())!=f(i.a.kd(),16).gc())return!1;return!0}return!1}function r6n(e,t){return e.c<t.c?-1:e.c>t.c?1:e.b<t.b?-1:e.b>t.b?1:e.a!=t.a?Ti(e.a)-Ti(t.a):e.d==(oj(),dF)&&t.d==lF?-1:e.d==lF&&t.d==dF?1:0}function _de(e){var t,i,a,u,l,g,m,v;for(u=Gi,a=qo,i=new ne(e.e.b);i.a<i.c.c.length;)for(t=f(ee(i),25),g=new ne(t.a);g.a<g.c.c.length;)l=f(ee(g),9),v=ke(e.p[l.p]),m=v+ke(e.b[e.g[l.p].p]),u=h.Math.min(u,v),a=h.Math.max(a,m);return a-u}function i2t(e,t){var i,a,u,l,g;return l=t.a,l.c.i==t.b?g=l.d:g=l.c,l.c.i==t.b?a=l.c:a=l.d,u=VCn(e.a,g,a),u>0&&u<$R?(i=Y4n(e.a,a.i,u,e.c),zmt(e.a,a.i,-i),i>0):u<0&&-u<$R?(i=X4n(e.a,a.i,-u,e.c),zmt(e.a,a.i,i),i>0):!1}function i6n(e,t,i,a){var u,l,g,m,v,S,T,A;for(u=(t-e.d)/e.c.c.length,l=0,e.a+=i,e.d=t,A=new ne(e.c);A.a<A.c.c.length;)T=f(ee(A),27),S=T.g,v=T.f,md(T,T.i+l*u),bd(T,T.j+a*i),RS(T,T.g+u),NS(T,e.a),++l,m=T.g,g=T.f,LAe(T,new Ke(m,g),new Ke(S,v))}function o6n(e){var t,i,a,u,l,g,m;if(e==null)return null;for(m=e.length,u=(m+1)/2|0,g=Pe($l,hA,30,u,15,1),m%2!=0&&(g[--u]=NIe((gn(m-1,e.length),e.charCodeAt(m-1)))),i=0,a=0;i<u;++i)t=NIe(_a(e,a++)),l=NIe(_a(e,a++)),g[i]=(t<<4|l)<<24>>24;return g}function a6n(e){if(e.xe()){var t=e.c;t.ye()?e.o="["+t.n:t.xe()?e.o="["+t.ve():e.o="[L"+t.ve()+";",e.b=t.ue()+"[]",e.k=t.we()+"[]";return}var i=e.j,a=e.d;a=a.split("/"),e.o=Tle(".",[i,Tle("$",a)]),e.b=Tle(".",[i,Tle(".",a)]),e.k=a[a.length-1]}function s6n(e,t){var i,a,u,l,g;for(g=null,l=new ne(e.e.a);l.a<l.c.c.length;)if(u=f(ee(l),123),u.b.a.c.length==u.g.a.c.length){for(a=u.e,g=eRn(u),i=u.e-f(g.a,17).a+1;i<u.e+f(g.b,17).a;i++)t[i]<t[a]&&(a=i);t[a]<t[u.e]&&(--t[u.e],++t[a],u.e=a)}}function cIe(e,t,i){var a,u,l,g,m,v,S,T;S=f(i,147),T=aIe(e,t),l=null,a=null,T&&(a=$Z(e,T,t.i)),a!=null?l=a:(m=t.i,l=m),tE(S,"x",l),g=null,u=null,T&&(u=FZ(e,T,t.j)),u!=null?g=u:(v=t.j,g=v),tE(S,"y",g),tE(S,cpe,t.g),tE(S,spe,t.f)}function o2t(e){Jde();var t,i,a,u;return a=cm(e,gl(35)),t=a==-1?e:(ga(0,a,e.length),e.substr(0,a)),i=a==-1?null:(gn(a+1,e.length+1),e.substr(a+1)),u=uSn(M7e,t),u?i!=null&&(u=U1t(u,(an(i),i))):(u=I7n(t),ESn(M7e,t,u),i!=null&&(u=U1t(u,i))),u}function uIe(e,t,i,a){var u,l,g,m,v;for(u=BIe(e,t),m=0,v=u.gc();m<v;++m)if(l=f(u.Xb(m),178),Mt(a,xP(Ns(e,l)))){if(g=gH(Ns(e,l)),i==null){if(g==null)return l}else if(Mt(i,g))return l}return null}function lIe(e,t,i,a){var u,l,g,m,v;for(u=Mde(e,t),m=0,v=u.gc();m<v;++m)if(l=f(u.Xb(m),178),Mt(a,xP(Ns(e,l)))){if(g=gH(Ns(e,l)),i==null){if(g==null)return l}else if(Mt(i,g))return l}return null}function c6n(e,t,i){var a,u,l,g,m,v;if(g=new J2,m=Ju(e.e.zh(),t),a=f(e.g,122),is(),f(t,69).uk())for(l=0;l<e.i;++l)u=a[l],m.Zl(u.Ik())&&Qn(g,u);else for(l=0;l<e.i;++l)u=a[l],m.Zl(u.Ik())&&(v=u.kd(),Qn(g,i?Bj(e,t,l,g.i,v):v));return LTe(g)}function a2t(e){var t,i,a,u,l,g,m;if(e&&(t=e.mi(bh),t&&(g=dr(Rp((!t.b&&(t.b=new Md((Ut(),Za),xc,t)),t.b),"conversionDelegates")),g!=null))){for(m=new et,a=DT(g,"\\w+"),u=0,l=a.length;u<l;++u)i=a[u],vn(m.c,i);return m}return Ht(),Ht(),Ga}function s2t(e,t){var i,a,u,l,g,m,v,S;for(g=t==1?pge:hge,l=g.a.ec().Jc();l.Ob();)for(u=f(l.Pb(),86),v=f(si(e.f.c,u),22).Jc();v.Ob();)switch(m=f(v.Pb(),47),a=f(m.b,82),S=f(m.a,194),i=S.c,u.g){case 2:case 1:a.g.d+=i;break;case 4:case 3:a.g.c+=i}}function u6n(e,t){var i,a,u,l,g;for(i=new Aj(sD),u=(ER(),ce(se(sD,1),Ve,231,0,[aD,H$,oD,QT,EA,yA])),l=0,g=u.length;l<g;++l)a=u[l],I_e(i,a,new et);return Zi(Lu(Jr(Zs(new Ft(null,new Bt(e.b,16)),new cO),new E6),new bet(t)),new vet(i)),i}function c2t(e,t){var i,a,u;for(i=1;i<e.c.length;i++){for(u=(Gt(i,e.c.length),f(e.c[i],9)),a=i;a>0&&LG(t,(Gt(a-1,e.c.length),f(e.c[a-1],9)),u)>0;)af(e,a,(Gt(a-1,e.c.length),f(e.c[a-1],9))),--a;Gt(a,e.c.length),e.c[a]=u}t.b=new Wn,t.f=new Wn}function u2t(e,t,i){var a,u,l;for(a=1;a<e.c.length;a++){for(l=(Gt(a,e.c.length),f(e.c[a],9)),u=a;u>0&&t.Le((Gt(u-1,e.c.length),f(e.c[u-1],9)),l)>0;)af(e,u,(Gt(u-1,e.c.length),f(e.c[u-1],9))),--u;Gt(u,e.c.length),e.c[u]=l}i.a=new Wn,i.b=new Wn}function YZ(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(l=t.Jc();l.Ob();)u=f(l.Pb(),27),T=u.i+u.g/2,L=u.j+u.f/2,v=e.f,g=v.i+v.g/2,m=v.j+v.f/2,S=T-g,A=L-m,a=h.Math.sqrt(S*S+A*A),S*=e.e/a,A*=e.e/a,i?(T-=S,L-=A):(T+=S,L+=A),md(u,T-u.g/2),bd(u,L-u.f/2)}function eA(e){var t,i,a;if(!e.c&&e.b!=null){for(t=e.b.length-4;t>=0;t-=2)for(i=0;i<=t;i+=2)(e.b[i]>e.b[i+2]||e.b[i]===e.b[i+2]&&e.b[i+1]>e.b[i+3])&&(a=e.b[i+2],e.b[i+2]=e.b[i],e.b[i]=a,a=e.b[i+3],e.b[i+3]=e.b[i+1],e.b[i+1]=a);e.c=!0}}function l6n(e,t){var i,a,u,l,g,m,v,S,T;for(S=-1,T=0,g=e,m=0,v=g.length;m<v;++m){for(l=g[m],i=new idt(S==-1?e[0]:e[S],t,(Sj(),Hte)),a=0;a<l.length;a++)for(u=a+1;u<l.length;u++)li(l[a],(Ge(),hi))&&li(l[u],hi)&&LG(i,l[a],l[u])>0&&++T;++S}return T}function Jh(e){var t,i;return i=new ef(Vy(e.Om)),i.a+="@",xr(i,(t=Ti(e)>>>0,t.toString(16))),e.Rh()?(i.a+=" (eProxyURI: ",vu(i,e.Xh()),e.Gh()&&(i.a+=" eClass: ",vu(i,e.Gh())),i.a+=")"):e.Gh()&&(i.a+=" (eClass: ",vu(i,e.Gh()),i.a+=")"),i.a}function zj(e){var t,i,a,u;if(e.e)throw ie(new As((d1(tge),zfe+tge.k+Hfe)));for(e.d==(ko(),$g)&&gQ(e,Ds),i=new ne(e.a.a);i.a<i.c.c.length;)t=f(ee(i),317),t.g=t.i;for(u=new ne(e.a.b);u.a<u.c.c.length;)a=f(ee(u),60),a.i=qo;return e.b.af(e),e}function d6n(e,t){var i,a,u,l,g,m,v,S,T;for(v=lb(e.c-e.b&e.a.length-1),S=null,T=null,l=new ej(e);l.a!=l.b;)u=f(sZ(l),9),i=(m=f(V(u,(Ge(),$p)),11),m?m.i:null),a=(g=f(V(u,yh),11),g?g.i:null),(S!=i||T!=a)&&(PEt(v,t),S=i,T=a),vn(v.c,u);PEt(v,t)}function f6n(e,t){var i,a,u,l,g;if(t<2*e.b)throw ie(new dn("The knot vector must have at least two time the dimension elements."));for(e.f=1,u=0;u<e.b;u++)Qe(e.e,0);for(g=t+1-2*e.b,i=g,l=1;l<g;l++)Qe(e.e,l/i);if(e.d)for(a=0;a<e.b;a++)Qe(e.e,1)}function l2t(e,t){var i,a,u,l,g,m,v,S,T;if(S=t,T=f(BJ(Sce(e.o),S),27),!T)throw u=p1(S,Am),m="Unable to find elk node for json object '"+u,v=m+"' Panic!",ie(new am(v));l=IS(S,"edges"),i=new Rat(e,T),mEt(i.a,i.b,l),g=IS(S,dpe),a=new int(e),xwt(a.a,g)}function d2t(e,t,i,a){var u,l,g,m,v;if(a!=null){if(u=e.d[t],u){for(l=u.g,v=u.i,m=0;m<v;++m)if(g=f(l[m],135),g.xi()==i&&ei(a,g.jd()))return m}}else if(u=e.d[t],u){for(l=u.g,v=u.i,m=0;m<v;++m)if(g=f(l[m],135),Ae(g.jd())===Ae(a))return m}return-1}function gL(e,t){var i,a,u;return i=t==null?ac(Is(e.f,null)):w7(e.i,t),me(i,240)?(u=f(i,240),u.vi()==null,u):me(i,491)?(a=f(i,1983),u=a.a,u&&(u.yb==null||(t==null?ml(e.f,null,u):MS(e.i,t,u))),u):null}function h6n(e){AIe();var t,i,a,u,l,g,m;if(e==null||(u=e.length,u%2!=0))return null;for(t=AJ(e),l=u/2|0,i=Pe($l,hA,30,l,15,1),a=0;a<l;a++){if(g=KF[t[a*2]],g==-1||(m=KF[t[a*2+1]],m==-1))return null;i[a]=(g<<4|m)<<24>>24}return i}function p6n(e,t,i){var a,u,l;if(u=f(Es(e.i,t),315),!u)if(u=new Dmt(e.d,t,i),ZN(e.i,t,u),h3e(t))Pmn(e.a,t.c,t.b,u);else switch(l=uRn(t),a=f(Es(e.p,l),251),l.g){case 1:case 3:u.j=!0,Eae(a,t.b,u);break;case 4:case 2:u.k=!0,Eae(a,t.c,u)}return u}function g6n(e,t,i,a){var u,l,g,m,v,S;if(m=new J2,v=Ju(e.e.zh(),t),u=f(e.g,122),is(),f(t,69).uk())for(g=0;g<e.i;++g)l=u[g],v.Zl(l.Ik())&&Qn(m,l);else for(g=0;g<e.i;++g)l=u[g],v.Zl(l.Ik())&&(S=l.kd(),Qn(m,a?Bj(e,t,g,m.i,S):S));return c3e(m,i)}function f2t(e,t){var i,a,u,l,g,m,v,S;if(u=e.b[t.p],u>=0)return u;for(l=1,m=new ne(t.j);m.a<m.c.c.length;)for(g=f(ee(m),11),a=new ne(g.g);a.a<a.c.c.length;)i=f(ee(a),18),S=i.d.i,t!=S&&(v=f2t(e,S),l=h.Math.max(l,v+1));return RCn(e,t,l),l}function h2t(e,t){var i,a,u,l,g,m,v,S;if(u=e.b[t.p],u>=0)return u;for(l=1,m=new ne(t.j);m.a<m.c.c.length;)for(g=f(ee(m),11),a=new ne(g.e);a.a<a.c.c.length;)i=f(ee(a),18),S=i.c.i,t!=S&&(v=h2t(e,S),l=h.Math.max(l,v+1));return uAn(e,t,l),l}function m6n(e){var t,i,a,u;for(u=e.length,t=null,a=0;a<u;a++)i=(gn(a,e.length),e.charCodeAt(a)),cm(".*+?{[()|\\^$",gl(i))>=0?(t||(t=new l7,a>0&&ys(t,(ga(0,a,e.length),e.substr(0,a)))),t.a+="\\",_P(t,i&To)):t&&_P(t,i&To);return t?t.a:e}function b6n(e){var t,i,a;for(i=new ne(e.a.a.b);i.a<i.c.c.length;)t=f(ee(i),82),a=(an(0),0),a>0&&(!(l1(e.a.c)&&t.n.d)&&!(k3(e.a.c)&&t.n.b)&&(t.g.d-=h.Math.max(0,a/2-.5)),!(l1(e.a.c)&&t.n.a)&&!(k3(e.a.c)&&t.n.c)&&(t.g.a+=h.Math.max(0,a-1)))}function p2t(e,t,i){var a,u;if((e.c-e.b&e.a.length-1)==2)t==(tt(),hn)||t==_n?(PJ(f(vj(e),15),(df(),Rb)),PJ(f(vj(e),15),Nw)):(PJ(f(vj(e),15),(df(),Nw)),PJ(f(vj(e),15),Rb));else for(u=new ej(e);u.a!=u.b;)a=f(sZ(u),15),PJ(a,i)}function v6n(e,t){var i,a,u,l,g,m,v;for(u=bP(new t2e(e)),m=new la(u,u.c.length),l=bP(new t2e(t)),v=new la(l,l.c.length),g=null;m.b>0&&v.b>0&&(i=(Ln(m.b>0),f(m.a.Xb(m.c=--m.b),27)),a=(Ln(v.b>0),f(v.a.Xb(v.c=--v.b),27)),i==a);)g=i;return g}function w6n(e,t){var i,a,u,l;for(t.Sg("Self-Loop pre-processing",1),a=new ne(e.a);a.a<a.c.c.length;)i=f(ee(a),9),kCn(i)&&(u=(l=new bvt(i),$e(i,(Ge(),f_),l),LLn(l),l),Zi(Lu(Zs(new Ft(null,new Bt(u.d,16)),new Xg),new r0),new Qm),P5n(u));t.Tg()}function g2t(e,t,i){var a,u,l,g;oht(e,t)>oht(e,i)?(a=lc(i,(tt(),_n)),e.d=a.dc()?0:ece(f(a.Xb(0),11)),g=lc(t,pn),e.b=g.dc()?0:ece(f(g.Xb(0),11))):(u=lc(i,(tt(),pn)),e.d=u.dc()?0:ece(f(u.Xb(0),11)),l=lc(t,_n),e.b=l.dc()?0:ece(f(l.Xb(0),11)))}function y6n(e){var t,i,a,u,l,g,m,v;t=!0,u=null,l=null;e:for(v=new ne(e.a);v.a<v.c.c.length;)for(m=f(ee(v),9),a=new Tn(Cn(Fo(m).a.Jc(),new B));Kn(a);){if(i=f(Dn(a),18),u&&u!=m){t=!1;break e}if(u=m,g=i.c.i,l&&l!=g){t=!1;break e}l=g}return t}function E6n(e,t,i){var a,u,l,g,m,v,S,T;for(T=(a=f(t.e&&t.e(),10),new Rf(a,f(Gh(a,a.length),10),0)),v=DT(i,"[\\[\\]\\s,]+"),l=v,g=0,m=l.length;g<m;++g)if(u=l[g],IT(u).length!=0){if(S=bEt(e,u),S==null)return null;Yh(T,f(S,23))}return T}function m2t(e,t){var i,a,u,l;for(i=e.o.a,l=f(f(si(e.r,t),22),83).Jc();l.Ob();)u=f(l.Pb(),115),u.e.a=i*ke(Oe(u.b.mf(Mee))),u.e.b=(a=u.b,a.nf((Rr(),Sd))?a.$f()==(tt(),hn)?-a.Kf().b-ke(Oe(a.mf(Sd))):ke(Oe(a.mf(Sd))):a.$f()==(tt(),hn)?-a.Kf().b:0)}function S6n(e,t,i){var a,u,l,g,m,v;for(l=-1,m=-1,g=0;g<t.c.length&&(u=(Gt(g,t.c.length),f(t.c[g],335)),!(u.c>e.c));g++)u.a>=e.s&&(l<0&&(l=g),m=g);return v=(e.s+e.c)/2,l>=0&&(a=tLn(e,t,l,m),v=fgn((Gt(a,t.c.length),f(t.c[a],335))),b4n(t,a,i)),v}function ir(e,t,i){var a,u,l,g,m,v,S;for(g=(l=new Wu,l),$xe(g,(an(t),t)),S=(!g.b&&(g.b=new Md((Ut(),Za),xc,g)),g.b),v=1;v<i.length;v+=2)hG(S,i[v-1],i[v]);for(a=(!e.Ab&&(e.Ab=new Be(Ir,e,0,3)),e.Ab),m=0;m<0;++m)u=Nwn(f(be(a,a.i-1),582)),a=u;Qn(a,g)}function dIe(e,t,i,a,u,l){var g,m,v;if(!u[t.a]){for(u[t.a]=!0,g=a,!g&&(g=new vJ),Qe(g.e,t),v=l[t.a].Jc();v.Ob();)m=f(v.Pb(),288),!(m.d==i||m.c==i)&&(m.c!=t&&dIe(e,m.c,t,g,u,l),m.d!=t&&dIe(e,m.d,t,g,u,l),Qe(g.c,m),$o(g.d,m.b));return g}return null}function _6n(e){var t,i,a,u,l,g,m;for(t=0,u=new ne(e.e);u.a<u.c.c.length;)a=f(ee(u),18),i=j3(new Ft(null,new Bt(a.b,16)),new cC),i&&++t;for(g=new ne(e.g);g.a<g.c.c.length;)l=f(ee(g),18),m=j3(new Ft(null,new Bt(l.b,16)),new pv),m&&++t;return t>=2}function k6n(e,t,i,a,u){var l,g,m,v,S,T;for(l=e.c.d.j,g=f(Xh(i,0),8),T=1;T<i.b;T++)S=f(Xh(i,T),8),Wi(a,g,a.c.b,a.c),m=qv(ii(new ja(g),S),.5),v=qv(new Kce(wCe(l)),u),ii(m,v),Wi(a,m,a.c.b,a.c),g=S,l=t==0?fZ(l):UCe(l);Mr(a,(Ln(i.b!=0),f(i.c.b.c,8)))}function T6n(e){NT();var t,i,a;return i=xi(bf,ce(se(ps,1),Ve,94,0,[up])),!(GH(nJ(i,e))>1||(t=xi(sp,ce(se(ps,1),Ve,94,0,[D1,cp])),GH(nJ(t,e))>1)||(a=xi(lp,ce(se(ps,1),Ve,94,0,[Pb,Sh])),GH(nJ(a,e))>1))}function b2t(e){var t,i,a,u,l,g,m;for(t=0,a=new ne(e.a);a.a<a.c.c.length;)for(i=f(ee(a),9),l=new Tn(Cn(ao(i).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),e==u.d.i.c&&u.c.j==(tt(),pn)&&(g=wg(u.c).b,m=wg(u.d).b,t=h.Math.max(t,h.Math.abs(m-g)));return t}function fIe(e,t,i){var a,u,l;for(l=new ne(e.t);l.a<l.c.c.length;)a=f(ee(l),270),a.b.s<0&&a.c>0&&(a.b.n-=a.c,a.b.n<=0&&a.b.u>0&&Mr(t,a.b));for(u=new ne(e.i);u.a<u.c.c.length;)a=f(ee(u),270),a.a.s<0&&a.c>0&&(a.a.u-=a.c,a.a.u<=0&&a.a.n>0&&Mr(i,a.a))}function XZ(e){var t,i,a,u,l;if(e.g==null&&(e.d=e.$i(e.f),Qn(e,e.d),e.c))return l=e.f,l;if(t=f(e.g[e.i-1],49),u=t.Pb(),e.e=t,i=e.$i(u),i.Ob())e.d=i,Qn(e,i);else for(e.d=null;!t.Ob()&&(io(e.g,--e.i,null),e.i!=0);)a=f(e.g[e.i-1],49),t=a;return u}function x6n(e,t){var i,a,u,l,g,m;if(a=t,u=a.Ik(),x1(e.e,u)){if(u.Pi()&&dJ(e,u,a.kd()))return!1}else for(m=Ju(e.e.zh(),u),i=f(e.g,122),l=0;l<e.i;++l)if(g=i[l],m.Zl(g.Ik()))return ei(g,a)?!1:(f(Y3(e,l,t),74),!0);return Qn(e,t)}function cn(){cn=Z,Yi=new MN("NORMAL",0),uo=new MN("LONG_EDGE",1),bo=new MN("EXTERNAL_PORT",2),xu=new MN("NORTH_SOUTH_PORT",3),Vc=new MN("LABEL",4),Dg=new MN("BREAKING_POINT",5),_q=new MN("PLACEHOLDER",6),U$=new MN("NONSHIFTING_PLACEHOLDER",7)}function C6n(e,t,i,a){var u,l,g,m;for(u=new Tg(e),Wl(u,(cn(),Vc)),$e(u,(Ge(),oi),t),$e(u,Y$,a),$e(u,(ot(),Xi),(oa(),uu)),$e(u,$p,t.c),$e(u,yh,t.d),Cde(t,u),m=h.Math.floor(i/2),g=new ne(u.j);g.a<g.c.c.length;)l=f(ee(g),11),l.n.b=m;return u}function v2t(e,t,i){var a,u,l,g,m;for(i.Sg("ELK Force",1),lt(ut(Fe(t,(Qh(),Q6e))))||vH((a=new ri((fS(),new uS(t))),a)),m=T1t(t),lIn(m),WTn(e,f(V(m,Z6e),421)),g=l_t(e.a,m),l=g.Jc();l.Ob();)u=f(l.Pb(),234),cDn(e.b,u,i.bh(1/g.gc()));m=Ekt(g),Nkt(m),i.Tg()}function hIe(e,t,i){switch(i.g){case 1:return new Ke(t.a,h.Math.min(e.d.b,t.b));case 2:return new Ke(h.Math.max(e.c.a,t.a),t.b);case 3:return new Ke(t.a,h.Math.max(e.c.b,t.b));case 4:return new Ke(h.Math.min(t.a,e.d.a),t.b)}return new Ke(t.a,t.b)}function A6n(e,t){var i,a,u,l,g;if(t.Sg("Breaking Point Processor",1),Y8n(e),lt(ut(V(e,(ot(),fDe))))){for(u=new ne(e.b);u.a<u.c.c.length;)for(a=f(ee(u),25),i=0,g=new ne(a.a);g.a<g.c.c.length;)l=f(ee(g),9),l.p=i++;WDn(e),M2t(e,!0),M2t(e,!1)}t.Tg()}function kde(e){var t,i,a;for(t=lb(1+(!e.c&&(e.c=new Be(_d,e,9,9)),e.c).i),Qe(t,(!e.d&&(e.d=new Qt(vo,e,8,5)),e.d)),a=new Nn((!e.c&&(e.c=new Be(_d,e,9,9)),e.c));a.e!=a.i.gc();)i=f(Mn(a),124),Qe(t,(!i.d&&(i.d=new Qt(vo,i,8,5)),i.d));return or(t),new L7(t)}function lw(e){var t,i,a;for(t=lb(1+(!e.c&&(e.c=new Be(_d,e,9,9)),e.c).i),Qe(t,(!e.e&&(e.e=new Qt(vo,e,7,4)),e.e)),a=new Nn((!e.c&&(e.c=new Be(_d,e,9,9)),e.c));a.e!=a.i.gc();)i=f(Mn(a),124),Qe(t,(!i.e&&(i.e=new Qt(vo,i,7,4)),i.e));return or(t),new L7(t)}function I6n(e){var t,i,a,u;if(e==null)return null;if(a=_u(e,!0),u=mq.length,Mt(a.substr(a.length-u,u),mq)){if(i=a.length,i==4){if(t=(gn(0,a.length),a.charCodeAt(0)),t==43)return nje;if(t==45)return jMt}else if(i==3)return nje}return AT(a)}function O6n(e,t,i,a){var u,l,g,m,v,S,T,A,L;for(A=a?(tt(),pn):(tt(),_n),u=!1,v=t[i],S=0,T=v.length;S<T;++S)m=v[S],!VN(f(V(m,(ot(),Xi)),100))&&(g=m.e,L=!lc(m,A).dc()&&!!g,L&&(l=iAe(g),e.b=new aAe(l,a?0:l.length-1)),u=u|fPn(e,m,A,L));return u}function w2t(e,t,i,a){var u,l,g;if(g=K3e(t,i),vn(a.c,t),e.j[g.p]==-1||e.j[g.p]==2||e.a[t.p])return a;for(e.j[g.p]=-1,l=new Tn(Cn(T1(g).a.Jc(),new B));Kn(l);)if(u=f(Dn(l),18),!(!(!Fa(u)&&!(!Fa(u)&&u.c.i.c==u.d.i.c))||u==t))return w2t(e,u,g,a);return a}function N6n(e){var t,i,a,u;for(t=0,i=0,u=new ne(e.j);u.a<u.c.c.length;)if(a=f(ee(u),11),t=mr(Ua(t,Kpt(Jr(new Ft(null,new Bt(a.e,16)),new yv)))),i=mr(Ua(i,Kpt(Jr(new Ft(null,new Bt(a.g,16)),new kC)))),t>1||i>1)return 2;return t+i==1?2:0}function wd(e,t){var i,a,u,l,g,m;return l=e.a*$fe+e.b*1502,m=e.b*$fe+11,i=h.Math.floor(m*VG),l+=i,m-=i*FOe,l%=FOe,e.a=l,e.b=m,t<=24?h.Math.floor(e.a*B6e[t]):(u=e.a*(1<<t-24),g=h.Math.floor(e.b*U6e[t]),a=u+g,a>=2147483648&&(a-=4294967296),a)}function y2t(e,t,i){var a,u,l,g,m,v,S;for(l=new et,S=new ki,g=new ki,jDn(e,S,g,t),h8n(e,S,g,t,i),v=new ne(e);v.a<v.c.c.length;)for(m=f(ee(v),116),u=new ne(m.k);u.a<u.c.c.length;)a=f(ee(u),133),(!t||a.c==(Ip(),Cw))&&m.g>a.b.g&&vn(l.c,a);return l}function R6n(e,t,i){var a,u,l,g,m,v;for(m=e.c,g=(i.q?i.q:(Ht(),Ht(),Cb)).vc().Jc();g.Ob();)l=f(g.Pb(),43),a=!oP(Jr(new Ft(null,new Bt(m,16)),new Y(new Sat(t,l)))).zd((Wy(),zR)),a&&(v=l.kd(),me(v,4)&&(u=g3e(v),u!=null&&(v=u)),t.of(f(l.jd(),144),v))}function P6n(e,t){var i,a,u,l;for(t.Sg("Resize child graph to fit parent.",1),a=new ne(e.b);a.a<a.c.c.length;)i=f(ee(a),25),$o(e.a,i.a),i.a.c.length=0;for(l=new ne(e.a);l.a<l.c.c.length;)u=f(ee(l),9),jo(u,null);e.b.c.length=0,Y5n(e),e.e&&BLn(e.e,e),t.Tg()}function L6n(e,t){var i,a,u,l,g;for(t.Sg("Edge joining",1),i=lt(ut(V(e,(ot(),Sme)))),u=new ne(e.b);u.a<u.c.c.length;)for(a=f(ee(u),25),g=new la(a.a,0);g.b<g.d.gc();)l=(Ln(g.b<g.d.gc()),f(g.d.Xb(g.c=g.b++),9)),l.k==(cn(),uo)&&(efe(l,i),hd(g));t.Tg()}function D6n(e,t,i){var a,u;if(pH(e.b),Ap(e.b,(rG(),dne),(g7(),Zq)),Ap(e.b,fne,t.g),Ap(e.b,hne,t.a),e.a=NG(e.b,t),i.Sg("Compaction by shrinking a tree",e.a.c.length),t.i.c.length>1)for(u=new ne(e.a);u.a<u.c.c.length;)a=f(ee(u),44),a.If(t,i.bh(1));i.Tg()}function pIe(e,t,i){var a,u,l;if(l=iA((Rl(),Ea),e.zh(),t),l){if(is(),!f(l,69).uk()&&(l=rR(Ns(Ea,l)),!l))throw ie(new dn(vw+t.ve()+S$));u=(a=e.Eh(l),f(a>=0?e.Hh(a,!0,!0):GS(e,l,!0),159)),f(u,218).Ul(t,i)}else throw ie(new dn(vw+t.ve()+S$))}function M6n(e,t,i){var a,u,l,g,m,v;if(v=x_e(e,f(bn(e.e,t),27)),m=null,v)switch(v.g){case 3:a=GSe(e,lT(t)),m=(an(i),i+(an(a),a));break;case 2:u=GSe(e,lT(t)),g=(an(i),i+(an(u),u)),l=GSe(e,f(bn(e.e,t),27)),m=g-(an(l),l);break;default:m=i}else m=i;return m}function j6n(e,t,i){var a,u,l,g,m,v;if(v=x_e(e,f(bn(e.e,t),27)),m=null,v)switch(v.g){case 3:a=qSe(e,lT(t)),m=(an(i),i+(an(a),a));break;case 2:u=qSe(e,lT(t)),g=(an(i),i+(an(u),u)),l=qSe(e,f(bn(e.e,t),27)),m=g-(an(l),l);break;default:m=i}else m=i;return m}function JZ(e,t){var i,a,u,l,g;if(t){for(l=me(e.Cb,88)||me(e.Cb,101),g=!l&&me(e.Cb,330),a=new Nn((!t.a&&(t.a=new G7(t,gs,t)),t.a));a.e!=a.i.gc();)if(i=f(Mn(a),87),u=uQ(i),l?me(u,88):g?me(u,155):u)return u;return l?(Ut(),Ch):(Ut(),Ug)}else return null}function $6n(e,t){var i,a,u,l,g;for(i=new et,u=Zs(new Ft(null,new Bt(e,16)),new Mh),l=Zs(new Ft(null,new Bt(e,16)),new bk),g=R_n(YSn(uT(h5n(ce(se(Z7n,1),Jt,827,0,[u,l])),new xy))),a=1;a<g.length;a++)g[a]-g[a-1]>=2*t&&Qe(i,new Kse(g[a-1]+t,g[a]-t));return i}function E2t(e,t,i){var a,u,l,g,m,v,S,T;if(i)for(l=i.a.length,a=new Xy(l),m=(a.b-a.a)*a.c<0?(zv(),Fw):new Vv(a);m.Ob();)g=f(m.Pb(),17),u=OP(i,g.a),u&&(v=tSn(e,(S=(Bv(),T=new w2e,T),t&&gIe(S,t),S),u),FP(v,p1(u,Am)),UZ(u,v),jAe(u,v),Zue(e,u,v))}function ZZ(e){var t,i,a,u,l,g;if(!e.j){if(g=new tm,t=GF,l=t.a.yc(e,t),l==null){for(a=new Nn(Bs(e));a.e!=a.i.gc();)i=f(Mn(a),29),u=ZZ(i),Qi(g,u),Qn(g,i);t.a.Ac(e)!=null}ET(g),e.j=new _3((f(be(je((Wv(),sn).o),11),19),g.i),g.g),pd(e).b&=-33}return e.j}function F6n(e){var t,i,a,u;if(e==null)return null;if(a=_u(e,!0),u=mq.length,Mt(a.substr(a.length-u,u),mq)){if(i=a.length,i==4){if(t=(gn(0,a.length),a.charCodeAt(0)),t==43)return rje;if(t==45)return $Mt}else if(i==3)return rje}return new m2e(a)}function B6n(e){var t,i,a;return i=e.l,(i&i-1)!=0||(a=e.m,(a&a-1)!=0)||(t=e.h,(t&t-1)!=0)||t==0&&a==0&&i==0?-1:t==0&&a==0&&i!=0?Nxe(i):t==0&&a!=0&&i==0?Nxe(a)+22:t!=0&&a==0&&i==0?Nxe(t)+44:-1}function tA(e,t){var i,a,u,l,g;for(u=t.a&e.f,l=null,a=e.b[u];;a=a.b){if(a==t){l?l.b=t.b:e.b[u]=t.b;break}l=a}for(g=t.f&e.f,l=null,i=e.c[g];;i=i.d){if(i==t){l?l.d=t.d:e.c[g]=t.d;break}l=i}t.e?t.e.c=t.c:e.a=t.c,t.c?t.c.e=t.e:e.e=t.e,--e.i,++e.g}function U6n(e,t){var i;t.d?t.d.b=t.b:e.a=t.b,t.b?t.b.d=t.d:e.e=t.d,!t.e&&!t.c?(i=f(ch(f(aR(e.b,t.a),260)),260),i.a=0,++e.c):(i=f(ch(f(bn(e.b,t.a),260)),260),--i.a,t.e?t.e.c=t.c:i.b=f(ch(t.c),495),t.c?t.c.e=t.e:i.c=f(ch(t.e),495)),--e.d}function Tde(e,t){var i,a,u,l;for(l=new la(e,0),i=(Ln(l.b<l.d.gc()),f(l.d.Xb(l.c=l.b++),143));l.b<l.d.gc();)a=(Ln(l.b<l.d.gc()),f(l.d.Xb(l.c=l.b++),143)),u=new ike(a.c,i.d,t),Ln(l.b>0),l.a.Xb(l.c=--l.b),nT(l,u),Ln(l.b<l.d.gc()),l.d.Xb(l.c=l.b++),u.a=!1,i=a}function S2t(e){var t,i,a,u,l,g;for(u=f(V(e,(Ge(),yte)),11),g=new ne(e.j);g.a<g.c.c.length;){for(l=f(ee(g),11),a=new ne(l.g);a.a<a.c.c.length;)return t=f(ee(a),18),ua(t,u),l;for(i=new ne(l.e);i.a<i.c.c.length;)return t=f(ee(i),18),Ia(t,u),l}return null}function _2t(e,t,i){var a,u,l,g,m,v;for(v=f(Zk(e.a,t),17).a,i?oAe(e.a,rt(v+1),t):oAe(e.a,rt(v-1),t),g=new ub,u=new Tn(Cn((i?ao(t):Fo(t)).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),i?l=a.d.i:l=a.c.i,Ae(Zk(e.a,l))===Ae(Zk(e.a,t))&&(m=g.a.yc(l,g),m==null);return g}function z6n(e,t,i){var a,u;a=Nc(i.q.getTime()),Eu(a,0)<0?(u=M0-mr(kle(T0(a),M0)),u==M0&&(u=0)):u=mr(kle(a,M0)),t==1?(u=h.Math.min((u+50)/100|0,9),Qy(e,48+u&To)):t==2?(u=h.Math.min((u+5)/10|0,99),bb(e,u,2)):(bb(e,u,3),t>3&&bb(e,0,t-3))}function H6n(e){var t,i,a,u;return Ae(V(e,(ot(),rx)))===Ae((_1(),W0))?!e.e&&Ae(V(e,Dq))!==Ae((zP(),Iq)):(a=f(V(e,dme),298),u=lt(ut(V(e,fme)))||Ae(V(e,Q$))===Ae((hj(),Aq)),t=f(V(e,ALe),17).a,i=e.a.c.length,!u&&a!=(zP(),Iq)&&(t==0||t>i))}function G6n(e,t){var i,a,u,l,g,m,v;for(u=e.Jc();u.Ob();)for(a=f(u.Pb(),9),m=new ou,cc(m,a),No(m,(tt(),_n)),$e(m,(Ge(),kte),(mn(),!0)),g=t.Jc();g.Ob();)l=f(g.Pb(),9),v=new ou,cc(v,l),No(v,pn),$e(v,kte,!0),i=new xS,$e(i,kte,!0),Ia(i,m),ua(i,v)}function q6n(e){var t,i;for(i=0;i<e.c.length&&!(clt((Gt(i,e.c.length),f(e.c[i],112)))>0);i++);if(i>0&&i<e.c.length-1)return i;for(t=0;t<e.c.length&&!(clt((Gt(t,e.c.length),f(e.c[t],112)))>0);t++);return t>0&&i<e.c.length-1?t:e.c.length/2|0}function k2t(e,t){var i,a,u,l,g,m,v;for(u=0,g=new ne(t.a);g.a<g.c.c.length;)for(l=f(ee(g),9),u+=l.o.b+l.d.a+l.d.d+e.e,a=new Tn(Cn(Fo(l).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),i.c.i.k==(cn(),xu)&&(v=i.c.i,m=f(V(v,(Ge(),oi)),9),u+=m.o.b+m.d.a+m.d.d);return u}function T2t(e,t){var i,a;if(t!=e.Cb||e.Db>>16!=6&&t){if(cL(e,t))throw ie(new dn(_$+jEt(e)));a=null,e.Cb&&(a=(i=e.Db>>16,i>=0?M3e(e,a):e.Cb.Ph(e,-1-i,null,a))),t&&(a=vR(t,e,6,a)),a=k_e(e,t,a),a&&a.lj()}else(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,6,t,t))}function gIe(e,t){var i,a;if(t!=e.Cb||e.Db>>16!=9&&t){if(cL(e,t))throw ie(new dn(_$+ESt(e)));a=null,e.Cb&&(a=(i=e.Db>>16,i>=0?$3e(e,a):e.Cb.Ph(e,-1-i,null,a))),t&&(a=vR(t,e,9,a)),a=T_e(e,t,a),a&&a.lj()}else(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,9,t,t))}function QZ(e,t){var i,a;if(t!=e.Cb||e.Db>>16!=3&&t){if(cL(e,t))throw ie(new dn(_$+k_t(e)));a=null,e.Cb&&(a=(i=e.Db>>16,i>=0?B3e(e,a):e.Cb.Ph(e,-1-i,null,a))),t&&(a=vR(t,e,12,a)),a=__e(e,t,a),a&&a.lj()}else(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,3,t,t))}function eQ(e,t){var i,a;if(t!=e.Cb||e.Db>>16!=11&&t){if(cL(e,t))throw ie(new dn(_$+MIe(e)));a=null,e.Cb&&(a=(i=e.Db>>16,i>=0?U3e(e,a):e.Cb.Ph(e,-1-i,null,a))),t&&(a=vR(t,e,10,a)),a=D_e(e,t,a),a&&a.lj()}else(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,11,t,t))}function mL(e){var t,i,a,u,l;if(a=ph(e),l=e.j,l==null&&a)return e.Gk()?null:a.fk();if(me(a,155)){if(i=a.gk(),i&&(u=i.si(),u!=e.i)){if(t=f(a,155),t.kk())try{e.g=u.pi(t,l)}catch(g){if(g=go(g),me(g,80))e.g=null;else throw ie(g)}e.i=u}return e.g}return null}function x2t(e){var t;return t=new et,Qe(t,new DN(new Ke(e.c,e.d),new Ke(e.c+e.b,e.d))),Qe(t,new DN(new Ke(e.c,e.d),new Ke(e.c,e.d+e.a))),Qe(t,new DN(new Ke(e.c+e.b,e.d+e.a),new Ke(e.c+e.b,e.d))),Qe(t,new DN(new Ke(e.c+e.b,e.d+e.a),new Ke(e.c,e.d+e.a))),t}function V6n(e){var t,i,a;if(e==null)return bl;try{return Qs(e)}catch(u){if(u=go(u),me(u,103))return t=u,a=Vy(Fd(e))+"@"+(i=(b0(),m3e(e)>>>0),i.toString(16)),_3n(PTn(),(cP(),"Exception during lenientFormat for "+a),t),"<"+a+" threw "+Vy(t.Om)+">";throw ie(u)}}function W6n(e,t,i,a){var u,l,g,m;u=nvt(e,t,i),l=nvt(e,i,t),g=f(bn(e.c,t),116),m=f(bn(e.c,i),116),u<l?new nE((Ip(),dx),g,m,l-u):l<u?new nE((Ip(),dx),m,g,u-l):(u!=0||!(!t.i||!i.i)&&a[t.i.c][i.i.c])&&(new nE((Ip(),dx),g,m,0),new nE(dx,m,g,0))}function K6n(e,t,i){var a,u,l,g,m;for(i.Sg("Breaking Point Removing",1),e.a=f(V(t,(ot(),P1)),220),l=new ne(t.b);l.a<l.c.c.length;)for(u=f(ee(l),25),m=new ne(rE(u.a));m.a<m.c.c.length;)g=f(ee(m),9),Omt(g)&&(a=f(V(g,(Ge(),u_)),314),!a.d&&fkt(e,a));i.Tg()}function Hj(){Hj=Z,$A=new Uz("CANDIDATE_POSITION_LAST_PLACED_RIGHT",0),n4=new Uz("CANDIDATE_POSITION_LAST_PLACED_BELOW",1),kF=new Uz("CANDIDATE_POSITION_WHOLE_DRAWING_RIGHT",2),_F=new Uz("CANDIDATE_POSITION_WHOLE_DRAWING_BELOW",3),Jq=new Uz("WHOLE_DRAWING",4)}function Y6n(e){var t,i,a,u;for(a=new yT(new Vi(e.b).a);a.b;)i=G3(a),u=f(i.jd(),11),t=f(i.kd(),9),$e(t,(Ge(),oi),u),$e(u,ed,t),$e(u,Rq,(mn(),!0)),No(u,f(V(t,Tc),64)),V(t,Tc),$e(u.i,(ot(),Xi),(oa(),MD)),f(V(ra(u.i),Cu),22).Ec((us(),hD))}function X6n(e){var t,i,a,u,l,g,m;for(l=new ki,u=new ne(e.d.a);u.a<u.c.c.length;)a=f(ee(u),123),a.b.a.c.length==0&&Wi(l,a,l.c.b,l.c);if(l.b>1)for(t=mS((i=new qy,++e.b,i),e.d),m=tr(l,0);m.b!=m.d.c;)g=f(Zn(m),123),Zh(Uh(Bh(zh(Fh(new ah,1),0),t),g))}function J6n(e,t,i){var a,u,l,g,m,v;if(l=0,g=0,e.c)for(v=new ne(e.d.i.j);v.a<v.c.c.length;)m=f(ee(v),11),l+=m.e.c.length;else l=1;if(e.d)for(v=new ne(e.c.i.j);v.a<v.c.c.length;)m=f(ee(v),11),g+=m.g.c.length;else g=1;return u=Aa(dH(g-l)),a=(i+t)/2+(i-t)*(.4*u),a}function Z6n(e,t){var i,a,u,l,g,m;for(t.Sg("Partition postprocessing",1),a=new ne(e.b);a.a<a.c.c.length;)for(i=f(ee(a),25),l=new ne(i.a);l.a<l.c.c.length;)for(u=f(ee(l),9),m=new ne(u.j);m.a<m.c.c.length;)g=f(ee(m),11),lt(ut(V(g,(Ge(),kte))))&&Y7(m);t.Tg()}function Q6n(e){CT();var t,i;if(e.Gc((tt(),fc)))throw ie(new dn("Port sides must not contain UNDEFINED"));switch(e.gc()){case 1:return Cq;case 2:return t=e.Gc(_n)&&e.Gc(pn),i=e.Gc(hn)&&e.Gc(Bn),t||i?dte:lte;case 3:return ute;case 4:return cte;default:return null}}function C2t(e,t,i){return bL(),zue(e,t)&&zue(e,i)?!1:nfe(new Ke(e.c,e.d),new Ke(e.c+e.b,e.d),t,i)||nfe(new Ke(e.c+e.b,e.d),new Ke(e.c+e.b,e.d+e.a),t,i)||nfe(new Ke(e.c+e.b,e.d+e.a),new Ke(e.c,e.d+e.a),t,i)||nfe(new Ke(e.c,e.d+e.a),new Ke(e.c,e.d),t,i)}function mIe(e,t){var i,a,u,l;if(!e.dc()){for(i=0,a=e.gc();i<a;++i)if(l=dr(e.Xb(i)),l==null?t==null:Mt(l.substr(0,3),"!##")?t!=null&&(u=t.length,!Mt(l.substr(l.length-u,u),t)||l.length!=t.length+3)&&!Mt(CE,t):Mt(l,Tpe)&&!Mt(CE,t)||Mt(l,t))return!0}return!1}function e5n(e,t,i,a){var u,l,g,m,v,S;for(g=e.j.c.length,v=Pe(tjn,VOe,315,g,0,1),m=0;m<g;m++)l=f(ct(e.j,m),11),l.p=m,v[m]=m4n(KEt(l),i,a);for(C5n(e,v,i,t,a),S=new Wn,u=0;u<v.length;u++)v[u]&&qr(S,f(ct(e.j,u),11),v[u]);S.f.c+S.i.c!=0&&($e(e,(Ge(),K$),S),BNn(e,v))}function t5n(e,t,i){var a,u,l;for(u=new ne(e.a.b);u.a<u.c.c.length;)if(a=f(ee(u),60),l=D3(a),l&&l.k==(cn(),bo))switch(f(V(l,(Ge(),Tc)),64).g){case 4:l.n.a=t.a;break;case 2:l.n.a=i.a-(l.o.a+l.d.c);break;case 1:l.n.b=t.b;break;case 3:l.n.b=i.b-(l.o.b+l.d.a)}}function n5n(e,t,i){var a,u,l;for(i.Sg("Processor determine the height for each level",1),e.a=t.b.b==0?1:t.b.b,u=null,a=tr(t.b,0);!u&&a.b!=a.d.c;)l=f(Zn(a),39),lt(ut(V(l,(_i(),Iw))))&&(u=l);u&&p_t(e,gm(ce(se(Zte,1),eee,39,0,[u])),i,f(V(t,(Ec(),E_)),86)),i.Tg()}function r5n(e){var t,i,a,u,l,g;for(a=(Bv(),l=new o7,l),_G(a,e),i=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));i.e!=i.i.gc();)t=f(Mn(i),27),g=(u=new o7,u),eQ(g,a),x3(g,t.g,t.f),FP(g,t.k),Of(g,t.i,t.j),Qn((!a.a&&(a.a=new Be(Ur,a,10,11)),a.a),g),_G(g,t);return a}function i5n(e,t,i){var a,u,l,g,m;return u=f(Fe(t,(GZ(),i9e)),17),!u&&(u=rt(0)),l=f(Fe(i,i9e),17),!l&&(l=rt(0)),u.a>l.a?-1:u.a<l.a?1:e.a&&(a=di(t.j,i.j),a!=0||(a=di(t.i,i.i),a!=0))?a:(g=t.g*t.f,m=i.g*i.f,di(g,m))}function o5n(e,t){var i,a,u,l,g,m,v,S,T,A;if(++e.e,v=e.d==null?0:e.d.length,t>v){for(T=e.d,e.d=Pe(A7e,X4e,66,2*v+4,0,1),l=0;l<v;++l)if(S=T[l],S)for(a=S.g,A=S.i,m=0;m<A;++m)u=f(a[m],135),g=C_e(e,u.xi()),i=e.d[g],!i&&(i=e.d[g]=e.ak()),i.Ec(u);return!0}else return!1}function a5n(e,t,i){var a,u,l,g,m,v;if(u=i,l=u.Ik(),x1(e.e,l)){if(l.Pi()){for(a=f(e.g,122),g=0;g<e.i;++g)if(m=a[g],ei(m,u)&&g!=t)throw ie(new dn(lq))}}else for(v=Ju(e.e.zh(),l),a=f(e.g,122),g=0;g<e.i;++g)if(m=a[g],v.Zl(m.Ik()))throw ie(new dn(gq));xj(e,t,i)}function A2t(e,t){var i,a,u,l,g,m;for(i=f(V(t,(Ge(),R1)),22),g=f(si((ofe(),gi),i),22),m=f(si(Ar,i),22),l=g.Jc();l.Ob();)if(a=f(l.Pb(),22),!f(si(e.b,a),15).dc())return!1;for(u=m.Jc();u.Ob();)if(a=f(u.Pb(),22),!f(si(e.b,a),15).dc())return!1;return!0}function s5n(e,t,i){e.d=0,e.b=0,t.k==(cn(),xu)&&i.k==xu&&f(V(t,(Ge(),oi)),9)==f(V(i,oi),9)&&(vue(t).j==(tt(),hn)?g2t(e,t,i):g2t(e,i,t)),t.k==xu&&i.k==uo?vue(t).j==(tt(),hn)?e.d=1:e.b=1:i.k==xu&&t.k==uo&&(vue(i).j==(tt(),hn)?e.b=1:e.d=1),vAn(e,t,i)}function bIe(e,t){var i,a,u,l,g,m,v,S,T;if(e.a.c.length==1)return kyt(f(ct(e.a,0),172),t);for(g=Dxn(e),v=0,S=e.d,l=g,T=e.d,m=(S-l)/2+l;l+1<S;){for(v=0,a=new ne(e.a);a.a<a.c.c.length;)i=f(ee(a),172),v+=(u=Zj(i,m,!1),u.a);v<t?(T=m,S=m):l=m,m=(S-l)/2+l}return T}function _G(e,t){var i,a,u,l,g;if(!t)return e;if(me(t,338))for(u=f(t,338),l=(!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),e.o),a=u.eh().c.Jc();a.e!=a.i.gc();)i=f(a.Vj(),43),g=i.kd(),hG(l,f(i.jd(),144),g);else!e.o&&(e.o=new Ol((Gc(),Db),K0,e,0)),ybt(e.o,t.lf());return e}function vIe(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;return T=null,z=aIe(e,t),v=null,z&&(v=tIn(e,z)),W=v,i=null,u=Oe(bn(e.i,W)),u!=null?i=u:i=rt(0),L=i,a=null,l=Oe(bn(e.j,W)),l!=null?a=l:a=rt(0),F=a,S=t.i,g=Fct(S,L),qr(e.i,t,g),A=t.j,m=Fct(A,F),T=Oe(qr(e.j,t,m)),T}function c5n(e){var t,i,a,u,l;return isNaN(e)?(MP(),w6e):e<-9223372036854776e3?(MP(),$At):e>=9223372036854776e3?(MP(),b6e):(u=!1,e<0&&(u=!0,e=-e),a=0,e>=yE&&(a=Aa(e/yE),e-=a*yE),i=0,e>=OR&&(i=Aa(e/OR),e-=i*OR),t=Aa(e),l=Ku(t,i,a),u&&Jue(l),l)}function u5n(e){var t,i,a,u,l;if(l=new et,Mu(e.b,new sr(l)),e.b.c.length=0,l.c.length!=0){for(t=(Gt(0,l.c.length),f(l.c[0],80)),i=1,a=l.c.length;i<a;++i)u=(Gt(i,l.c.length),f(l.c[i],80)),u!=t&&XOn(t,u);if(me(t,63))throw ie(f(t,63));if(me(t,294))throw ie(f(t,294))}}function l5n(e,t){var i,a,u,l;for(i=!t||!e.u.Gc((Xl(),M1)),l=0,u=new ne(e.e.Vf());u.a<u.c.c.length;){if(a=f(ee(u),831),a.$f()==(tt(),fc))throw ie(new dn("Label and node size calculator can only be used with ports that have port sides assigned."));a.Of(l++),Ixn(e,a,i)}}function wIe(e){var t,i,a,u,l;for(i=new ne(e.a.a);i.a<i.c.c.length;){for(t=f(ee(i),317),t.j=null,l=t.a.a.ec().Jc();l.Ob();)a=f(l.Pb(),60),Cp(a.b),(!t.j||a.d.c<t.j.d.c)&&(t.j=a);for(u=t.a.a.ec().Jc();u.Ob();)a=f(u.Pb(),60),a.b.a=a.d.c-t.j.d.c,a.b.b=a.d.d-t.j.d.d}return e}function tQ(e){var t,i,a,u,l;for(i=new ne(e.a.a);i.a<i.c.c.length;){for(t=f(ee(i),194),t.f=null,l=t.a.a.ec().Jc();l.Ob();)a=f(l.Pb(),82),Cp(a.e),(!t.f||a.g.c<t.f.g.c)&&(t.f=a);for(u=t.a.a.ec().Jc();u.Ob();)a=f(u.Pb(),82),a.e.a=a.g.c-t.f.g.c,a.e.b=a.g.d-t.f.g.d}return e}function d5n(e,t,i){var a,u,l,g,m;for(g=kj(e,i),m=Pe(Ib,j0,9,t.length,0,1),a=0,l=g.Jc();l.Ob();)u=f(l.Pb(),11),lt(ut(V(u,(Ge(),Rq))))&&(m[a++]=f(V(u,ed),9));if(a<t.length)throw ie(new As("Expected "+t.length+" hierarchical ports, but found only "+a+"."));return m}function f5n(e){var t,i,a,u,l,g,m,v,S,T,A;return A=dAe(e),t=e.a,v=t!=null,v&&wP(A,"category",e.a),u=Az(new Ic(e.d)),g=!u,g&&(S=new oh,Wh(A,"knownOptions",S),i=new Tnt(S),Oa(new Ic(e.d),i)),l=Az(e.g),m=!l,m&&(T=new oh,Wh(A,"supportedFeatures",T),a=new xnt(T),Oa(e.g,a)),A}function h5n(e){var t,i,a,u,l,g,m,v,S;for(a=!1,t=336,i=0,l=new Xct(e.length),m=e,v=0,S=m.length;v<S;++v)g=m[v],a=a|(ow(g),!1),u=(Kv(g),g.a),Qe(l.a,or(u)),t&=u.wd(),i=fxn(i,u.xd());return f(f(ndt(new Ft(null,cde(new Bt((_S(),s3e(l.a)),16),new Te,t,i)),new H5(e)),672),827)}function p5n(e,t){var i;e.d&&(t.c!=e.e.c||Bkn(e.e.b,t.b))&&(Qe(e.f,e.d),e.a=e.d.c+e.d.b,e.d=null,e.e=null),Fgn(t.b)?e.c=t:e.b=t,(t.b==(qP(),a_)&&!t.a||t.b==KT&&t.a||t.b==bA&&t.a||t.b==YT&&!t.a)&&e.c&&e.b&&(i=new qh(e.a,e.c.d,t.c-e.a,e.b.d-e.c.d),e.d=i,e.e=t)}function yIe(e,t){var i,a;if(t!=e.Cb||e.Db>>16!=7&&t){if(cL(e,t))throw ie(new dn(_$+Oyt(e)));a=null,e.Cb&&(a=(i=e.Db>>16,i>=0?j3e(e,a):e.Cb.Ph(e,-1-i,null,a))),t&&(a=f(t,52).Nh(e,1,wV,a)),a=Tke(e,t,a),a&&a.lj()}else(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,7,t,t))}function I2t(e,t){var i,a;if(t!=e.Cb||e.Db>>16!=3&&t){if(cL(e,t))throw ie(new dn(_$+_0t(e)));a=null,e.Cb&&(a=(i=e.Db>>16,i>=0?F3e(e,a):e.Cb.Ph(e,-1-i,null,a))),t&&(a=f(t,52).Nh(e,0,EV,a)),a=kke(e,t,a),a&&a.lj()}else(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,3,t,t))}function xde(e,t){vL();var i,a,u,l,g,m,v,S,T;return t.d>e.d&&(m=e,e=t,t=m),t.d<63?Z5n(e,t):(g=(e.d&-2)<<4,S=HTe(e,g),T=HTe(t,g),a=Vde(e,oR(S,g)),u=Vde(t,oR(T,g)),v=xde(S,T),i=xde(a,u),l=xde(Vde(S,a),Vde(u,T)),l=Qde(Qde(l,v),i),l=oR(l,g),v=oR(v,g<<1),Qde(Qde(v,l),i))}function D0(){D0=Z,xme=new w3(xxt,0),gDe=new w3("LONGEST_PATH",1),mDe=new w3("LONGEST_PATH_SOURCE",2),Tme=new w3("COFFMAN_GRAHAM",3),pDe=new w3(the,4),bDe=new w3("STRETCH_WIDTH",5),zte=new w3("MIN_WIDTH",6),SD=new w3("BF_MODEL_ORDER",7),_D=new w3("DF_MODEL_ORDER",8)}function g5n(e,t){var i,a,u,l,g,m;if(!e.tb){for(l=(!e.rb&&(e.rb=new aT(e,Up,e)),e.rb),m=new PN(l.i),u=new Nn(l);u.e!=u.i.gc();)a=f(Mn(u),141),g=a.ve(),i=f(g==null?ml(m.f,null,a):MS(m.i,g,a),141),i&&(g==null?ml(m.f,null,i):MS(m.i,g,i));e.tb=m}return f(wu(e.tb,t),141)}function kG(e,t){var i,a,u,l,g;if((e.i==null&&Em(e),e.i).length,!e.p){for(g=new PN((3*e.g.i/2|0)+1),u=new GN(e.g);u.e!=u.i.gc();)a=f(Ple(u),178),l=a.ve(),i=f(l==null?ml(g.f,null,a):MS(g.i,l,a),178),i&&(l==null?ml(g.f,null,i):MS(g.i,l,i));e.p=g}return f(wu(e.p,t),178)}function EIe(e,t,i,a,u){var l,g,m,v,S;for(p3n(a+QX(i,i.ge()),u),pdt(t,PCn(i)),l=i.f,l&&EIe(e,t,l,"Caused by: ",!1),m=(i.k==null&&(i.k=Pe(Upe,Ye,80,0,0,1)),i.k),v=0,S=m.length;v<S;++v)g=m[v],EIe(e,t,g,"Suppressed: ",!1);console.groupEnd!=null&&console.groupEnd.call(console)}function TG(e,t,i,a){var u,l,g,m,v;for(v=t.e,m=v.length,g=t.q.sg(v,i?0:m-1,i),u=v[i?0:m-1],g=g|pSt(e,u,i,a),l=i?1:m-2;i?l<m:l>=0;l+=i?1:-1)g=g|t.c.jg(v,l,i,a&&!lt(ut(V(t.j,(Ge(),kw))))&&!lt(ut(V(t.j,(Ge(),xA))))),g=g|t.q.sg(v,l,i),g=g|pSt(e,v[l],i,a);return Go(e.c,t),g}function nQ(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(T=$ht(e.j),A=0,L=T.length;A<L;++A){if(S=T[A],i==(cs(),td)||i==sF)for(v=pm(S.g),u=v,l=0,g=u.length;l<g;++l)a=u[l],FNn(t,a)&&VS(a,!0);if(i==Uu||i==sF)for(m=pm(S.e),u=m,l=0,g=u.length;l<g;++l)a=u[l],$Nn(t,a)&&VS(a,!0)}}function O2t(){return ka(),ce(se(ajn,1),Ve,79,0,[L5e,N5e,vA,Ege,Q5e,Kee,ite,ZT,J5e,z5e,Y5e,JT,Z5e,F5e,ePe,T5e,Zee,Sge,Vee,tte,nPe,ete,x5e,X5e,rPe,nte,k5e,tPe,Wee,M5e,W5e,V5e,ote,I5e,qee,Xee,A5e,nD,G5e,B5e,K5e,z$,R5e,O5e,q5e,U5e,Jee,rte,C5e,Qee,H5e,Yee,j5e,D5e,Tq,Gee,$5e,P5e])}function m5n(e){var t,i;switch(t=null,i=null,tOn(e).g){case 1:t=(tt(),_n),i=pn;break;case 2:t=(tt(),Bn),i=hn;break;case 3:t=(tt(),pn),i=_n;break;case 4:t=(tt(),hn),i=Bn}Dv(e,f(Dd(XN(f(si(e.k,t),15).Mc(),wA)),112)),jk(e,f(Dd(oT(f(si(e.k,i),15).Mc(),wA)),112))}function Gj(e){var t;if(Mrt.call(this),this.i=new zO,this.g=e,this.f=f(e.e&&e.e(),10).length,this.f==0)throw ie(new dn("There must be at least one phase in the phase enumeration."));this.c=(t=f(Tp(this.g),10),new Rf(t,f(Gh(t,t.length),10),0)),this.a=new _o,this.b=new Wn}function b5n(e){var t,i,a,u,l,g;if(u=f(ct(e.j,0),11),u.e.c.length+u.g.c.length==0)e.n.a=0;else{for(g=0,a=pb(Df(ce(se(Gf,1),Jt,20,0,[new Z5(u),new _N(u)])));Kn(a);)i=f(Dn(a),11),g+=i.i.n.a+i.n.a+i.a.a;t=f(V(e,(ot(),m_)),8),l=t?t.a:0,e.n.a=g/(u.e.c.length+u.g.c.length)-l}}function N2t(e,t){var i,a,u;for(a=new ne(t.a);a.a<a.c.c.length;)i=f(ee(a),224),sce(f(i.b,68),Ho($a(f(t.b,68).c),f(t.b,68).a)),u=KSt(f(t.b,68).b,f(i.b,68).b),u>1&&(e.a=!0),qvn(f(i.b,68),ii($a(f(t.b,68).c),qv(Ho($a(f(i.b,68).a),f(t.b,68).a),u))),Wft(e,t),N2t(e,i)}function R2t(e){var t,i,a,u,l,g,m;for(l=new ne(e.a.a);l.a<l.c.c.length;)a=f(ee(l),194),a.e=0,a.d.a.$b();for(u=new ne(e.a.a);u.a<u.c.c.length;)for(a=f(ee(u),194),i=a.a.a.ec().Jc();i.Ob();)for(t=f(i.Pb(),82),m=t.f.Jc();m.Ob();)g=f(m.Pb(),82),g.d!=a&&(Go(a.d,g),++g.d.e)}function v5n(e,t){var i,a,u,l,g,m,v;for(t.Sg("Constraints Postprocessor",1),g=0,l=new ne(e.b);l.a<l.c.c.length;){for(u=f(ee(l),25),v=0,m=!1,a=new ne(u.a);a.a<a.c.c.length;)i=f(ee(a),9),i.k==(cn(),Yi)&&(m=!0,$e(i,(ot(),Lte),rt(g)),$e(i,Ote,rt(v)),++v);m&&++g}t.Tg()}function w5n(e){var t,i,a,u,l,g,m,v;for(v=e.j.c.length,i=0,t=v,u=2*v,m=new ne(e.j);m.a<m.c.c.length;)switch(g=f(ee(m),11),g.j.g){case 2:case 4:g.p=-1;break;case 1:case 3:a=g.e.c.length,l=g.g.c.length,a>0&&l>0?g.p=t++:a>0?g.p=i++:l>0?g.p=u++:g.p=i++}Ht(),na(e.j,new nk)}function y5n(e){var t,i;i=null,t=f(ct(e.g,0),18);do{if(i=t.d.i,li(i,(Ge(),yh)))return f(V(i,yh),11).i;if(i.k!=(cn(),Yi)&&Kn(new Tn(Cn(ao(i).a.Jc(),new B))))t=f(Dn(new Tn(Cn(ao(i).a.Jc(),new B))),18);else if(i.k!=Yi)return null}while(i&&i.k!=(cn(),Yi));return i}function E5n(e,t){var i,a,u,l,g,m,v,S,T;for(m=t.j,g=t.g,v=f(ct(m,m.c.length-1),112),T=(Gt(0,m.c.length),f(m.c[0],112)),S=Kle(e,g,v,T),l=1;l<m.c.length;l++)i=(Gt(l-1,m.c.length),f(m.c[l-1],112)),u=(Gt(l,m.c.length),f(m.c[l],112)),a=Kle(e,g,i,u),a>S&&(v=i,T=u,S=a);t.a=T,t.c=v}function S5n(e,t,i){var a,u,l,g,m,v,S;for(S=new g0(new ltt(e)),g=ce(se(GOt,1),XTt,11,0,[t,i]),m=0,v=g.length;m<v;++m)for(l=g[m],S.a.yc(l,(mn(),yw))==null,u=new yg(l.b);sc(u.a)||sc(u.b);)a=f(sc(u.a)?ee(u.a):ee(u.b),18),a.c==a.d||fH(S,l==a.c?a.d:a.c);return or(S),new Kl(S)}function Zh(e){if(!e.a.d||!e.a.e)throw ie(new As((d1(fIt),fIt.k+" must have a source and target "+(d1(q6e),q6e.k)+" specified.")));if(e.a.d==e.a.e)throw ie(new As("Network simplex does not support self-loops: "+e.a+" "+e.a.d+" "+e.a.e));return uX(e.a.d.g,e.a),uX(e.a.e.b,e.a),e.a}function _5n(e,t,i){var a,u,l,g,m;for(i.Sg("Longest path layering",1),e.a=t,m=e.a.a,e.b=Pe(pr,Pr,30,m.c.length,15,1),a=0,g=new ne(m);g.a<g.c.c.length;)u=f(ee(g),9),u.p=a,e.b[a]=-1,++a;for(l=new ne(m);l.a<l.c.c.length;)u=f(ee(l),9),f2t(e,u);m.c.length=0,e.a=null,e.b=null,i.Tg()}function P2t(e,t,i){var a,u,l,g,m,v;if(a=0,t.b!=0&&i.b!=0){l=tr(t,0),g=tr(i,0),m=ke(Oe(Zn(l))),v=ke(Oe(Zn(g))),u=!0;do{if(m>v-e.b&&m<v+e.b)return-1;m>v-e.a&&m<v+e.a&&++a,m<=v&&l.b!=l.d.c?m=ke(Oe(Zn(l))):v<=m&&g.b!=g.d.c?v=ke(Oe(Zn(g))):u=!1}while(u)}return a}function L2t(e,t){var i,a;return pH(e.a),Ap(e.a,(GJ(),rne),rne),Ap(e.a,jA,jA),a=new _o,Sr(a,jA,(BZ(),sbe)),Ae(Fe(t,(uw(),dbe)))!==Ae((ZH(),ine))&&Sr(a,jA,ibe),lt(ut(Fe(t,FMe)))&&Sr(a,jA,cbe),Sr(a,jA,obe),lt(ut(Fe(t,UMe)))&&Il(a,jA,abe),FSe(e.a,a),i=NG(e.a,t),i}function k5n(e,t,i,a){var u,l,g,m,v,S,T,A,L;for(v=0,T=new ne(e.a);T.a<T.c.c.length;){for(S=f(ee(T),9),m=0,l=new Tn(Cn(Fo(S).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),A=wg(u.c).b,L=wg(u.d).b,m=h.Math.max(m,h.Math.abs(L-A));v=h.Math.max(v,m)}return g=a*h.Math.min(1,t/i)*v,g}function T5n(e){var t,i;for(i=new Tn(Cn(ao(e).a.Jc(),new B));Kn(i);)if(t=f(Dn(i),18),t.d.i.k!=(cn(),Vc))throw ie(new h0(ehe+lG(e)+"' has its layer constraint set to LAST, but has at least one outgoing edge that  does not go to a LAST_SEPARATE node. That must not happen."))}function x5n(e,t){var i,a,u,l,g;for(g=f(V(t,(Ec(),wMe)),422),l=tr(t.b,0);l.b!=l.d.c;)if(u=f(Zn(l),39),e.b[u.g]==0){switch(g.g){case 0:zvt(e,u);break;case 1:H4n(e,u)}e.b[u.g]=2}for(a=tr(e.a,0);a.b!=a.d.c;)i=f(Zn(a),65),_T(i.b.d,i,!0),_T(i.c.b,i,!0);$e(t,(_i(),cMe),e.a)}function SIe(e){var t;return t=new l7,(e&256)!=0&&(t.a+="F"),(e&128)!=0&&(t.a+="H"),(e&512)!=0&&(t.a+="X"),(e&2)!=0&&(t.a+="i"),(e&8)!=0&&(t.a+="m"),(e&4)!=0&&(t.a+="s"),(e&32)!=0&&(t.a+="u"),(e&64)!=0&&(t.a+="w"),(e&16)!=0&&(t.a+="x"),(e&ep)!=0&&(t.a+=","),R2e(t.a)}function C5n(e,t,i,a,u){var l,g,m,v;for(v=(l=f(Tp(Ja),10),new Rf(l,f(Gh(l,l.length),10),0)),m=new ne(e.j);m.a<m.c.c.length;)g=f(ee(m),11),t[g.p]&&(D9n(g,t[g.p],a),Yh(v,g.j));u?(rde(e,t,(tt(),_n),2*i,a),rde(e,t,pn,2*i,a)):(rde(e,t,(tt(),hn),2*i,a),rde(e,t,Bn,2*i,a))}function A5n(e,t){var i,a,u,l,g,m,v;for(u=new et,i=0;i<=e.j;i++)a=new ru(t),a.p=e.j-i,vn(u.c,a);for(m=new ne(e.p);m.a<m.c.c.length;)g=f(ee(m),9),jo(g,f(ct(u,e.j-e.g[g.p]),25));for(l=new ne(u);l.a<l.c.c.length;)v=f(ee(l),25),v.a.c.length==0&&Y7(l);t.b.c.length=0,$o(t.b,u)}function I5n(e,t){var i,a,u,l,g,m;t.Sg(Kxt,1),u=f(Fe(e,(Ag(),EF)),107),l=(!e.a&&(e.a=new Be(Ur,e,10,11)),e.a),g=QAn(l),m=h.Math.max(g.a,ke(Oe(Fe(e,(wb(),yF))))-(u.b+u.c)),a=h.Math.max(g.b,ke(Oe(Fe(e,cne)))-(u.d+u.a)),i=a-g.b,pi(e,wF,i),pi(e,t4,m),pi(e,CD,a+i),t.Tg()}function rQ(e){var t,i;if((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i==0)return oCe(e);for(t=f(be((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a),0),168),Jn((!t.a&&(t.a=new So(yf,t,5)),t.a)),B3(t,0),H3(t,0),U3(t,0),z3(t,0),i=(!e.a&&(e.a=new Be(Ii,e,6,6)),e.a);i.i>1;)PT(i,i.i-1);return t}function Ju(e,t){is();var i,a,u,l;return t?t==(fi(),MMt)||(t==kMt||t==VE||t==_Mt)&&e!=tje?new vOe(e,t):(a=f(t,677),i=a.Xk(),i||(xP(Ns((Rl(),Ea),t)),i=a.Xk()),l=(!i.i&&(i.i=new Wn),i.i),u=f(ac(Is(l.f,e)),1987),!u&&qr(l,e,u=new vOe(e,t)),u):yMt}function O5n(e,t){var i,a;if(a=fH(e.b,t.b),!a)throw ie(new As("Invalid hitboxes for scanline constraint calculation."));(p1t(t.b,f(ngn(e.b,t.b),60))||p1t(t.b,f(tgn(e.b,t.b),60)))&&b0(),e.a[t.b.f]=f(Dae(e.b,t.b),60),i=f(Lae(e.b,t.b),60),i&&(e.a[i.f]=t.b)}function N5n(e,t){var i,a,u,l,g,m,v,S,T;for(v=f(V(e,(Ge(),oi)),11),S=uc(ce(se(Zo,1),Ye,8,0,[v.i.n,v.n,v.a])).a,T=e.i.n.b,i=pm(e.e),u=i,l=0,g=u.length;l<g;++l)a=u[l],ua(a,v),Ma(a.a,new Ke(S,T)),t&&(m=f(V(a,(ot(),Ls)),77),m||(m=new fd,$e(a,Ls,m)),Mr(m,new Ke(S,T)))}function R5n(e,t){var i,a,u,l,g,m,v,S,T;for(u=f(V(e,(Ge(),oi)),11),S=uc(ce(se(Zo,1),Ye,8,0,[u.i.n,u.n,u.a])).a,T=e.i.n.b,i=pm(e.g),g=i,m=0,v=g.length;m<v;++m)l=g[m],Ia(l,u),S3(l.a,new Ke(S,T)),t&&(a=f(V(l,(ot(),Ls)),77),a||(a=new fd,$e(l,Ls,a)),Mr(a,new Ke(S,T)))}function P5n(e){var t,i,a,u,l,g,m,v,S;if(a=e.b,l=a.e,g=VN(f(V(a,(ot(),Xi)),100)),i=!!l&&f(V(l,(Ge(),Cu)),22).Gc((us(),Vf)),!(g||i))for(S=(m=new kp(e.e).a.vc().Jc(),new zy(m));S.a.Ob();)v=(t=f(S.a.Pb(),43),f(t.kd(),112)),v.a&&(u=v.d,cc(u,null),v.c=!0,e.a=!0)}function L5n(e,t){var i,a,u,l;for(t.Sg("Semi-Interactive Crossing Minimization Processor",1),i=!1,u=new ne(e.b);u.a<u.c.c.length;)a=f(ee(u),25),l=tG(RJ(Jr(Jr(new Ft(null,new Bt(a.a,16)),new ak),new yy),new Dt),new Rn),i=i|l.a!=null;i&&$e(e,(Ge(),BPe),(mn(),!0)),t.Tg()}function D5n(e,t){var i,a,u,l,g,m;for(e.b=new et,e.d=f(V(t,(Ge(),ZR)),233),e.e=r2n(e.d),l=new ki,u=gm(ce(se(UOt,1),KTt,36,0,[t])),g=0;g<u.c.length;)a=(Gt(g,u.c.length),f(u.c[g],36)),a.p=g++,i=new q_t(a,e.a,e.b),$o(u,i.b),Qe(e.b,i),i.s&&(m=tr(l,0),UX(m,i));return e.c=new Mo,l}function M5n(e,t){var i,a,u,l,g,m;for(g=f(f(si(e.r,t),22),83).Jc();g.Ob();)l=f(g.Pb(),115),i=l.c?dke(l.c):0,i>0?l.a?(m=l.b.Kf().a,i>m&&(u=(i-m)/2,l.d.b=u,l.d.c=u)):l.d.c=e.s+i:K7(e.u)&&(a=fAe(l.b),a.c<0&&(l.d.b=-a.c),a.c+a.b>l.b.Kf().a&&(l.d.c=a.c+a.b-l.b.Kf().a))}function j5n(e,t){var i,a,u,l,g;g=new et,i=t;do l=f(bn(e.b,i),132),l.B=i.c,l.D=i.d,vn(g.c,l),i=f(bn(e.k,i),18);while(i);return a=(Gt(0,g.c.length),f(g.c[0],132)),a.j=!0,a.A=f(a.d.a.ec().Jc().Pb(),18).c.i,u=f(ct(g,g.c.length-1),132),u.q=!0,u.C=f(u.d.a.ec().Jc().Pb(),18).d.i,g}function $5n(e){var t,i;i=f(V(e,(ot(),el)),171),t=f(V(e,(Ge(),OE)),312),i==(Ff(),Tw)?($e(e,el,Lq),$e(e,OE,(w1(),kA))):i==tx?($e(e,el,Lq),$e(e,OE,(w1(),WR))):t==(w1(),kA)?($e(e,el,Tw),$e(e,OE,Oq)):t==WR&&($e(e,el,tx),$e(e,OE,Oq))}function iQ(){iQ=Z,Hq=new $6,c5t=Sr(new _o,(wa(),su),(ka(),Vee)),d5t=Il(Sr(new _o,su,ete),zs,Qee),f5t=vm(vm(b7(Il(Sr(new _o,rp,ite),zs,rte),cu),nte),ote),u5t=Il(Sr(Sr(Sr(new _o,Ab,Kee),cu,Xee),cu,nD),zs,Yee),l5t=Il(Sr(Sr(new _o,cu,nD),cu,qee),zs,Gee)}function qj(){qj=Z,g5t=Sr(Il(new _o,(wa(),zs),(ka(),j5e)),su,Vee),w5t=vm(vm(b7(Il(Sr(new _o,rp,ite),zs,rte),cu),nte),ote),m5t=Il(Sr(Sr(Sr(new _o,Ab,Kee),cu,Xee),cu,nD),zs,Yee),v5t=Sr(Sr(new _o,su,ete),zs,Qee),b5t=Il(Sr(Sr(new _o,cu,nD),cu,qee),zs,Gee)}function F5n(e,t,i,a,u){var l,g;(!Fa(t)&&t.c.i.c==t.d.i.c||!Ebt(uc(ce(se(Zo,1),Ye,8,0,[u.i.n,u.n,u.a])),i))&&!Fa(t)&&(t.c==u?pP(t.a,0,new ja(i)):Mr(t.a,new ja(i)),a&&!xp(e.a,i)&&(g=f(V(t,(ot(),Ls)),77),g||(g=new fd,$e(t,Ls,g)),l=new ja(i),Wi(g,l,g.c.b,g.c),Go(e.a,l)))}function D2t(e,t){var i,a,u,l;for(l=mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))),i=l&e.b.length-1,u=null,a=e.b[i];a;u=a,a=a.a)if(a.d==l&&dm(a.i,t))return u?u.a=a.a:e.b[i]=a.a,rit(f(ch(a.c),588),f(ch(a.f),588)),Ez(f(ch(a.b),226),f(ch(a.e),226)),--e.f,++e.e,!0;return!1}function B5n(e){var t,i;for(i=new Tn(Cn(Fo(e).a.Jc(),new B));Kn(i);)if(t=f(Dn(i),18),t.c.i.k!=(cn(),Vc))throw ie(new h0(ehe+lG(e)+"' has its layer constraint set to FIRST, but has at least one incoming edge that  does not come from a FIRST_SEPARATE node. That must not happen."))}function M2t(e,t){var i,a,u,l,g,m,v,S,T,A,L;u=t?new ig:new Hl,l=!1;do for(l=!1,S=t?Bd(e.b):e.b,v=S.Jc();v.Ob();)for(m=f(v.Pb(),25),L=rE(m.a),t||Bd(L),A=new ne(L);A.a<A.c.c.length;)T=f(ee(A),9),u.Mb(T)&&(a=T,i=f(V(T,(Ge(),u_)),314),g=t?i.b:i.k,l=sSt(a,g,t,!1));while(l)}function U5n(e,t,i){var a,u,l,g,m,v,S;if(u=w0t(e.Db&254),u==0)e.Eb=i;else{if(u==1)m=Pe(Ro,Jt,1,2,5,1),l=Wle(e,t),l==0?(m[0]=i,m[1]=e.Eb):(m[0]=e.Eb,m[1]=i);else for(m=Pe(Ro,Jt,1,u+1,5,1),g=uR(e.Eb),a=2,v=0,S=0;a<=128;a<<=1)a==t?m[S++]=i:(e.Db&a)!=0&&(m[S++]=g[v++]);e.Eb=m}e.Db|=t}function j2t(e,t,i,a,u,l){var g,m,v,S,T,A,L,F,z,W,J,ue;for(T=a,t.j&&t.o?(F=f(bn(e.f,t.A),60),W=F.d.c+F.d.b,--T):W=t.a.c+t.a.b,A=u,i.q&&i.o?(F=f(bn(e.f,i.C),60),S=F.d.c,++A):S=i.a.c,J=S-W,v=h.Math.max(2,A-T),m=J/v,z=W+m,L=T;L<A;++L)g=f(l.Xb(L),132),ue=g.a.b,g.a.c=z-ue/2,z+=m}function _Ie(e,t,i,a,u,l){var g,m,v,S,T,A;for(S=i.c.length,l&&(e.c=Pe(pr,Pr,30,t.length,15,1)),g=u?0:t.length-1;u?g<t.length:g>=0;g+=u?1:-1){for(m=t[g],v=a==(tt(),_n)?u?lc(m,a):Bd(lc(m,a)):u?Bd(lc(m,a)):lc(m,a),l&&(e.c[m.p]=v.gc()),A=v.Jc();A.Ob();)T=f(A.Pb(),11),e.d[T.p]=S++;$o(i,v)}}function $2t(e,t,i){var a,u,l,g,m,v,S,T;for(l=ke(Oe(e.b.Jc().Pb())),S=ke(Oe(ATn(t.b))),a=qv($a(e.a),S-i),u=qv($a(t.a),i-l),T=ii(a,u),qv(T,1/(S-l)),this.a=T,this.b=new et,m=!0,g=e.b.Jc(),g.Pb();g.Ob();)v=ke(Oe(g.Pb())),m&&v-i>Mhe&&(this.b.Ec(i),m=!1),this.b.Ec(v);m&&this.b.Ec(i)}function z5n(e){var t,i,a,u;if(iLn(e,e.n),e.d.c.length>0){for(c7(e.c);tIe(e,f(ee(new ne(e.e.a)),123))<e.e.a.c.length;){for(t=ZAn(e),u=t.e.e-t.d.e-t.a,t.e.j&&(u=-u),a=new ne(e.e.a);a.a<a.c.c.length;)i=f(ee(a),123),i.j&&(i.e+=u);c7(e.c)}c7(e.c),UAe(e,f(ee(new ne(e.e.a)),123)),F_t(e)}}function H5n(e,t,i){var a,u,l,g,m;for(i.Sg("Longest path to source layering",1),e.a=t,m=e.a.a,e.b=Pe(pr,Pr,30,m.c.length,15,1),a=0,g=new ne(m);g.a<g.c.c.length;)u=f(ee(g),9),u.p=a,e.b[a]=-1,++a;for(l=new ne(m);l.a<l.c.c.length;)u=f(ee(l),9),h2t(e,u);m.c.length=0,e.a=null,e.b=null,i.Tg()}function G5n(e,t){hR();var i,a;if(i=oue(fj(),t.Ng()),i){if(a=i.j,me(e,206))return lEn(f(e,27))?Cl(a,(wm(),Ao))||Cl(a,en):Cl(a,(wm(),Ao));if(me(e,268))return Cl(a,(wm(),Dm));if(me(e,193))return Cl(a,(wm(),FE));if(me(e,359))return Cl(a,(wm(),L1))}return!0}function q5n(e,t,i){var a,u,l,g,m,v;if(u=i,l=u.Ik(),x1(e.e,l)){if(l.Pi()){for(a=f(e.g,122),g=0;g<e.i;++g)if(m=a[g],ei(m,u)&&g!=t)throw ie(new dn(lq))}}else for(v=Ju(e.e.zh(),l),a=f(e.g,122),g=0;g<e.i;++g)if(m=a[g],v.Zl(m.Ik())&&g!=t)throw ie(new dn(gq));return f(Y3(e,t,i),74)}function F2t(e,t){if(t instanceof Object)try{if(t.__java$exception=e,navigator.userAgent.toLowerCase().indexOf("msie")!=-1&&$doc.documentMode<9)return;var i=e;Object.defineProperties(t,{cause:{get:function(){var a=i.fe();return a&&a.de()}},suppressed:{get:function(){return i.ee()}}})}catch{}}function B2t(e,t){var i,a,u,l,g;if(a=t>>5,t&=31,a>=e.d)return e.e<0?(ym(),HAt):(ym(),M$);if(l=e.d-a,u=Pe(pr,Pr,30,l+1,15,1),JNn(u,l,e.a,a,t),e.e<0){for(i=0;i<a&&e.a[i]==0;i++);if(i<a||t>0&&e.a[i]<<32-t!=0){for(i=0;i<l&&u[i]==-1;i++)u[i]=0;i==l&&++l,++u[i]}}return g=new Zy(e.e,l,u),nj(g),g}function V5n(e,t,i,a){var u,l,g,m,v;return m=Us(f(be((!t.b&&(t.b=new Qt(Yn,t,4,7)),t.b),0),84)),v=Us(f(be((!t.c&&(t.c=new Qt(Yn,t,5,8)),t.c),0),84)),so(m)==so(v)||bT(v,m)?null:(g=lT(t),g==i?a:(l=f(bn(e.a,g),9),l&&(u=l.e,u)?u:null))}function W5n(e,t){var i;switch(i=f(V(e,(ot(),Rte)),281),t.Sg("Label side selection ("+i+")",1),i.g){case 0:XEt(e,(df(),Rb));break;case 1:XEt(e,(df(),Nw));break;case 2:m_t(e,(df(),Rb));break;case 3:m_t(e,(df(),Nw));break;case 4:K2t(e,(df(),Rb));break;case 5:K2t(e,(df(),Nw))}t.Tg()}function x1(e,t){is();var i,a,u;return t.Gk()?!0:t.Fk()==-2?t==(_R(),f4)||t==d4||t==E1e||t==S1e?!0:(u=e.zh(),Pi(u,t)>=0?!1:(i=iA((Rl(),Ea),u,t),i?(a=i.Fk(),(a>1||a==-1)&&kS(Ns(Ea,i))!=3):!0)):!1}function kIe(e,t,i){var a,u,l,g,m,v;if(a=Bpn(i,e.length),g=e[a],l=dit(i,g.length),g[l].k==(cn(),bo))for(v=t.j,u=0;u<v.c.length;u++)m=(Gt(u,v.c.length),f(v.c[u],11)),(i?m.j==(tt(),_n):m.j==(tt(),pn))&&lt(ut(V(m,(Ge(),Rq))))&&(af(v,u,f(V(g[l],(Ge(),oi)),11)),l+=i?1:-1)}function K5n(e,t){var i,a,u,l,g,m,v,S;t.Sg("Greedy Width Approximator",1),i=ke(Oe(Fe(e,(Ag(),lne)))),v=f(Fe(e,EF),107),l=f(Fe(e,h8e),385),g=lt(ut(Fe(e,f8e))),m=ke(Oe(Fe(e,SF))),S=(!e.a&&(e.a=new Be(Ur,e,10,11)),e.a),qxe(S),u=new Mut(i,l,g),a=nLn(u,S,m,v),pi(e,(wb(),fx),a.c),t.Tg()}function U2t(e){if(e.g==null)switch(e.p){case 0:e.g=sEn(e)?(mn(),QL):(mn(),yw);break;case 1:e.g=jH(nSn(e));break;case 2:e.g=qH(ZEn(e));break;case 3:e.g=Bwn(e);break;case 4:e.g=new f3(Uwn(e));break;case 6:e.g=TT(zwn(e));break;case 5:e.g=rt(Wyn(e));break;case 7:e.g=ZP(iSn(e))}return e.g}function TIe(e){if(e.n==null)switch(e.p){case 0:e.n=cEn(e)?(mn(),QL):(mn(),yw);break;case 1:e.n=jH(rSn(e));break;case 2:e.n=qH(QEn(e));break;case 3:e.n=Hwn(e);break;case 4:e.n=new f3(Gwn(e));break;case 6:e.n=TT(qwn(e));break;case 5:e.n=rt(Kyn(e));break;case 7:e.n=ZP(oSn(e))}return e.n}function z2t(e,t,i,a){var u,l,g,m,v;if(m=(is(),f(t,69).uk()),x1(e.e,t)){if(t.Pi()&&lQ(e,t,a,me(t,101)&&(f(t,19).Bb&Xa)!=0))throw ie(new dn(lq))}else for(v=Ju(e.e.zh(),t),u=f(e.g,122),g=0;g<e.i;++g)if(l=u[g],v.Zl(l.Ik()))throw ie(new dn(gq));xj(e,CAe(e,t,i),m?f(a,74):mb(t,a))}function H2t(e){var t,i,a,u,l,g,m;for(l=new ne(e.a.a);l.a<l.c.c.length;)a=f(ee(l),317),a.g=0,a.i=0,a.e.a.$b();for(u=new ne(e.a.a);u.a<u.c.c.length;)for(a=f(ee(u),317),i=a.a.a.ec().Jc();i.Ob();)for(t=f(i.Pb(),60),m=t.c.Jc();m.Ob();)g=f(m.Pb(),60),g.a!=a&&(Go(a.e,g),++g.a.g,++g.a.i)}function Y5n(e){var t,i,a,u,l;u=f(V(e,(ot(),DE)),22),l=f(V(e,$te),22),i=new Ke(e.f.a+e.d.b+e.d.c,e.f.b+e.d.d+e.d.a),t=new ja(i),u.Gc((Ud(),mx))&&(a=f(V(e,mD),8),l.Gc((yd(),jD))&&(a.a<=0&&(a.a=20),a.b<=0&&(a.b=20)),t.a=h.Math.max(i.a,a.a),t.b=h.Math.max(i.b,a.b)),HDn(e,i,t)}function X5n(e,t){var i,a,u;t.a?(fH(e.b,t.b),e.a[t.b.i]=f(Dae(e.b,t.b),82),i=f(Lae(e.b,t.b),82),i&&(e.a[i.i]=t.b)):(a=f(Dae(e.b,t.b),82),a&&a==e.a[t.b.i]&&a.d&&a.d!=t.b.d&&a.f.Ec(t.b),u=f(Lae(e.b,t.b),82),u&&e.a[u.i]==t.b&&u.d&&u.d!=t.b.d&&t.b.f.Ec(u),Ise(e.b,t.b))}function Cde(e,t){var i,a,u,l,g,m;return l=e.d,m=ke(Oe(V(e,(ot(),z0)))),m<0&&(m=0,$e(e,z0,m)),t.o.b=m,g=h.Math.floor(m/2),a=new ou,No(a,(tt(),pn)),cc(a,t),a.n.b=g,u=new ou,No(u,_n),cc(u,t),u.n.b=g,ua(e,a),i=new xS,Rc(i,e),$e(i,Ls,null),Ia(i,u),ua(i,l),oDn(t,e,i),SRn(e,i),i}function J5n(e){var t,i;return i=f(V(e,(Ge(),Cu)),22),t=new _o,i.Gc((us(),V$))&&(lf(t,i5t),lf(t,GDe)),(i.Gc(hD)||lt(ut(V(e,(ot(),mme)))))&&(lf(t,GDe),i.Gc(_A)&&lf(t,a5t)),i.Gc(Vf)&&lf(t,r5t),i.Gc(W$)&&lf(t,s5t),i.Gc(vte)&&lf(t,o5t),i.Gc(dD)&&lf(t,t5t),i.Gc(fD)&&lf(t,n5t),t}function Z5n(e,t){var i,a,u,l,g,m,v,S,T,A,L;return a=e.d,l=t.d,m=a+l,v=e.e!=t.e?-1:1,m==2?(T=Na(ia(e.a[0],ls),ia(t.a[0],ls)),L=mr(T),A=mr(Jy(T,32)),A==0?new m1(v,L):new Zy(v,2,ce(se(pr,1),Pr,30,15,[L,A]))):(i=e.a,u=t.a,g=Pe(pr,Pr,30,m,15,1),Rxn(i,a,u,l,g),S=new Zy(v,m,g),nj(S),S)}function G2t(e,t,i,a){var u,l;if(t){if(u=e.a.Le(i.d,t.d),u==0)return a.d=O_e(t,i.e),a.b=!0,t;l=u<0?0:1,t.a[l]=G2t(e,t.a[l],i,a),Uk(t.a[l])&&(Uk(t.a[1-l])?(t.b=!0,t.a[0].b=!1,t.a[1].b=!1):Uk(t.a[l].a[l])?t=VJ(t,1-l):Uk(t.a[l].a[1-l])&&(t=Cht(t,1-l)))}else return i;return t}function q2t(e,t,i){var a,u,l,g;u=e.i,a=e.n,WTe(e,(Np(),_c),u.c+a.b,i),WTe(e,kc,u.c+u.b-a.c-i[2],i),g=u.b-a.b-a.c,i[0]>0&&(i[0]+=e.d,g-=i[0]),i[2]>0&&(i[2]+=e.d,g-=i[2]),l=h.Math.max(0,g),i[1]=h.Math.max(i[1],g),WTe(e,Bu,u.c+a.b+i[0]-(i[1]-g)/2,i),t==Bu&&(e.c.b=l,e.c.c=u.c+a.b+(l-g)/2)}function V2t(){this.c=Pe(sa,_s,30,(tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])).length,15,1),this.b=Pe(sa,_s,30,ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn]).length,15,1),this.a=Pe(sa,_s,30,ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn]).length,15,1),rSe(this.c,Gi),rSe(this.b,qo),rSe(this.a,qo)}function Su(e,t,i){var a,u,l,g;if(t<=i?(u=t,l=i):(u=i,l=t),a=0,e.b==null)e.b=Pe(pr,Pr,30,2,15,1),e.b[0]=u,e.b[1]=l,e.c=!0;else{if(a=e.b.length,e.b[a-1]+1==u){e.b[a-1]=l;return}g=Pe(pr,Pr,30,a+2,15,1),au(e.b,0,g,0,a),e.b=g,e.b[a-1]>=u&&(e.c=!1,e.a=!1),e.b[a++]=u,e.b[a]=l,e.c||eA(e)}}function Q5n(e,t,i){var a,u,l,g,m,v,S;for(S=t.d,e.a=new Du(S.c.length),e.c=new Wn,m=new ne(S);m.a<m.c.c.length;)g=f(ee(m),105),l=new VH(null),Qe(e.a,l),qr(e.c,g,l);for(e.b=new Wn,kRn(e,t),a=0;a<S.c.length-1;a++)for(v=f(ct(t.d,a),105),u=a+1;u<S.c.length;u++)W6n(e,v,f(ct(t.d,u),105),i)}function ePn(e,t){var i,a,u,l,g,m,v;for(i=qo,m=(cn(),Yi),u=new ne(t.a);u.a<u.c.c.length;)a=f(ee(u),9),l=a.k,l!=Yi&&(g=Oe(V(a,(Ge(),zPe))),g==null?(i=h.Math.max(i,0),a.n.b=i+c_e(e.a,l,m)):a.n.b=(an(g),g)),v=c_e(e.a,l,m),a.n.b<i+v+a.d.d&&(a.n.b=i+v+a.d.d),i=a.n.b+a.o.b+a.d.a,m=l}function RT(e){var t,i,a,u,l;for(u=new et,t=new H7((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a)),a=new Tn(Cn(lw(e).a.Jc(),new B));Kn(a);)i=f(Dn(a),85),me(be((!i.b&&(i.b=new Qt(Yn,i,4,7)),i.b),0),193)||(l=Us(f(be((!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c),0),84)),t.a._b(l)||vn(u.c,l));return u}function tPn(e,t,i){var a,u,l;if(e.e=i,e.d=0,e.b=0,e.f=1,e.i=t,(e.e&16)==16&&(e.i=cLn(e.i)),e.j=e.i.length,Yr(e),l=FS(e),e.d!=e.j)throw ie(new vr(yr((lr(),JCt))));if(e.g){for(a=0;a<e.g.a.c.length;a++)if(u=f(SS(e.g,a),575),e.f<=u.a)throw ie(new vr(yr((lr(),ZCt))));e.g.a.c.length=0}return l}function nPn(e,t){var i,a,u,l,g,m,v;for(t.Sg("Comment post-processing",1),l=new ne(e.b);l.a<l.c.c.length;){for(u=f(ee(l),25),a=new et,m=new ne(u.a);m.a<m.c.c.length;)g=f(ee(m),9),v=f(V(g,(Ge(),QR)),15),i=f(V(g,KR),15),(v||i)&&(c9n(g,v,i),v&&$o(a,v),i&&$o(a,i));$o(u.a,a)}t.Tg()}function W2t(e,t,i,a,u){var l,g,m,v,S,T;if(e.d&&e.d.Eg(u),l=f(u.Xb(0),27),Hvt(e,i,l,!1)||(g=f(u.Xb(u.gc()-1),27),Hvt(e,a,g,!0))||vAe(e,u))return!0;for(T=u.Jc();T.Ob();)for(S=f(T.Pb(),27),v=t.Jc();v.Ob();)if(m=f(v.Pb(),27),oQ(e,S,m))return!0;return!1}function rPn(e,t,i){var a,u,l,g,m,v,S,T,A,L;L=t.c.length,A=(S=e.Eh(i),f(S>=0?e.Hh(S,!1,!0):GS(e,i,!1),61));e:for(l=A.Jc();l.Ob();){for(u=f(l.Pb(),57),T=0;T<L;++T)if(g=(Gt(T,t.c.length),f(t.c[T],74)),v=g.kd(),m=g.Ik(),a=u.Jh(m,!1),v==null?a!=null:!ei(v,a))continue e;return u}return null}function iPn(e,t,i,a){var u,l,g,m;for(u=f(L0(t,(tt(),pn)).Jc().Pb(),11),l=f(L0(t,_n).Jc().Pb(),11),m=new ne(e.j);m.a<m.c.c.length;){for(g=f(ee(m),11);g.e.c.length!=0;)ua(f(ct(g.e,0),18),u);for(;g.g.c.length!=0;)Ia(f(ct(g.g,0),18),l)}i||$e(t,(Ge(),$p),null),a||$e(t,(Ge(),yh),null)}function K2t(e,t){var i,a,u,l,g,m,v;for(i=new O3,l=new ne(e.b);l.a<l.c.c.length;){for(u=f(ee(l),25),v=!0,a=0,m=new ne(u.a);m.a<m.c.c.length;)switch(g=f(ee(m),9),g.k.g){case 4:++a;case 1:NTe(i,g);break;case 0:ERn(g,t);default:i.b==i.c||MSt(i,a,v,!1,t),v=!1,a=0}i.b==i.c||MSt(i,a,v,!0,t)}}function oPn(e){var t,i,a,u,l,g,m,v,S;for(e.a=new P_e,S=0,u=0,a=new ne(e.i.b);a.a<a.c.c.length;){for(t=f(ee(a),25),t.p=u,v=new ne(t.a);v.a<v.c.c.length;)m=f(ee(v),9),m.p=S,++S;++u}for(l=e.r==(gh(),w_),g=l?eNt:QOt,i=new ne(e.i.b);i.a<i.c.c.length;)t=f(ee(i),25),na(t.a,g),q_n(e.a,rt(t.p),t.a)}function xIe(e,t){var i,a,u,l,g,m;for(i=0,m=new ne(t);m.a<m.c.c.length;){for(g=f(ee(m),11),_Ce(e.b,e.d[g.p]),u=new yg(g.b);sc(u.a)||sc(u.b);)a=f(sc(u.a)?ee(u.a):ee(u.b),18),l=V2e(e,g==a.c?a.d:a.c),l>e.d[g.p]&&(i+=FTe(e.b,l),Jv(e.a,rt(l)));for(;!u7(e.a);)vxe(e.b,f(JN(e.a),17).a)}return i}function Y2t(e,t,i){var a,u,l,g;for(l=(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i,u=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));u.e!=u.i.gc();)a=f(Mn(u),27),(!a.a&&(a.a=new Be(Ur,a,10,11)),a.a).i==0||(l+=Y2t(e,a,!1));if(i)for(g=so(t);g;)l+=(!g.a&&(g.a=new Be(Ur,g,10,11)),g.a).i,g=so(g);return l}function PT(e,t){var i,a,u,l;return e.Mj()?(a=null,u=e.Nj(),e.Qj()&&(a=e.Sj(e.Xi(t),null)),i=e.Fj(4,l=wR(e,t),null,t,u),e.Jj()&&l!=null&&(a=e.Lj(l,a)),a?(a.kj(i),a.lj()):e.Gj(i),l):(l=wR(e,t),e.Jj()&&l!=null&&(a=e.Lj(l,null),a&&a.lj()),l)}function aPn(e){var t,i,a,u,l,g,m,v,S,T;for(S=e.a,t=new Mo,v=0,a=new ne(e.d);a.a<a.c.c.length;){for(i=f(ee(a),225),T=0,sE(i.b,new Km),g=tr(i.b,0);g.b!=g.d.c;)l=f(Zn(g),225),t.a._b(l)&&(u=i.c,m=l.c,T<m.d+m.a+S&&T+u.a+S>m.d&&(T=m.d+m.a+S));i.c.d=T,t.a.yc(i,t),v=h.Math.max(v,i.c.d+i.c.a)}return v}function sPn(e,t,i){var a,u,l,g,m,v;for(g=f(V(e,(Ge(),Zge)),15).Jc();g.Ob();){switch(l=f(g.Pb(),9),f(V(l,(ot(),el)),171).g){case 2:jo(l,t);break;case 4:jo(l,i)}for(u=new Tn(Cn(T1(l).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),!(a.c&&a.d)&&(m=!a.d,v=f(V(a,GPe),11),m?ua(a,v):Ia(a,v))}}function us(){us=Z,bte=new Kk("COMMENTS",0),Vf=new Kk("EXTERNAL_PORTS",1),V$=new Kk("HYPEREDGES",2),vte=new Kk("HYPERNODES",3),hD=new Kk("NON_FREE_PORTS",4),_A=new Kk("NORTH_SOUTH_PORTS",5),W$=new Kk(pxt,6),dD=new Kk("CENTER_LABELS",7),fD=new Kk("END_LABELS",8),wte=new Kk("PARTITIONS",9)}function cPn(e,t,i,a,u){return a<0?(a=Q3(e,u,ce(se(ft,1),Ye,2,6,[gfe,mfe,bfe,vfe,AR,wfe,yfe,Efe,Sfe,_fe,kfe,Tfe]),t),a<0&&(a=Q3(e,u,ce(se(ft,1),Ye,2,6,["Jan","Feb","Mar","Apr",AR,"Jun","Jul","Aug","Sep","Oct","Nov","Dec"]),t)),a<0?!1:(i.k=a,!0)):a>0?(i.k=a-1,!0):!1}function uPn(e,t,i,a,u){return a<0?(a=Q3(e,u,ce(se(ft,1),Ye,2,6,[gfe,mfe,bfe,vfe,AR,wfe,yfe,Efe,Sfe,_fe,kfe,Tfe]),t),a<0&&(a=Q3(e,u,ce(se(ft,1),Ye,2,6,["Jan","Feb","Mar","Apr",AR,"Jun","Jul","Aug","Sep","Oct","Nov","Dec"]),t)),a<0?!1:(i.k=a,!0)):a>0?(i.k=a-1,!0):!1}function lPn(e,t,i,a,u,l){var g,m,v,S;if(m=32,a<0){if(t[0]>=e.length||(m=_a(e,t[0]),m!=43&&m!=45)||(++t[0],a=WZ(e,t),a<0))return!1;m==45&&(a=-a)}return m==32&&t[0]-i==2&&u.b==2&&(v=new EY,S=v.q.getFullYear()-hw+hw-80,g=S%100,l.a=a==g,a+=(S/100|0)*100+(a<g?100:0)),l.p=a,!0}function X2t(e,t){var i,a,u,l,g;so(e)&&(g=f(V(t,(ot(),DE)),181),Ae(Fe(e,Xi))===Ae((oa(),Pw))&&pi(e,Xi,Fg),a=(fS(),new uS(so(e))),l=new Rse(so(e)?new uS(so(e)):null,e),u=ukt(a,l,!1,!0),Yh(g,(Ud(),mx)),i=f(V(t,mD),8),i.a=h.Math.max(u.a,i.a),i.b=h.Math.max(u.b,i.b))}function dPn(e){var t,i,a,u,l,g;for(l=new Mo,t=new H7((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a)),u=new Tn(Cn(lw(e).a.Jc(),new B));Kn(u);)a=f(Dn(u),85),me(be((!a.b&&(a.b=new Qt(Yn,a,4,7)),a.b),0),193)||(g=Us(f(be((!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c),0),84)),t.a._b(g)||(i=l.a.yc(g,l),i==null));return l}function bL(){bL=Z,X8e=ce(se(x_,1),dTt,30,14,[1,1,2,6,24,120,720,5040,40320,362880,3628800,39916800,479001600,6227020800,87178291200,1307674368e3,{l:3506176,m:794077,h:1},{l:884736,m:916411,h:20},{l:3342336,m:3912489,h:363},{l:589824,m:3034138,h:6914},{l:3407872,m:1962506,h:138294}]),h.Math.pow(2,-65)}function vL(){vL=Z;var e,t;for(UR=Pe(mA,Ye,90,32,0,1),j$=Pe(mA,Ye,90,32,0,1),e=1,t=0;t<=18;t++)UR[t]=(ym(),Eu(e,0)>=0?aw(e):X7(aw(T0(e)))),j$[t]=ZY(hb(e,t),0)?aw(hb(e,t)):X7(aw(T0(hb(e,t)))),e=Na(e,5);for(;t<j$.length;t++)UR[t]=M3(UR[t-1],UR[1]),j$[t]=M3(j$[t-1],(ym(),Wpe))}function J2t(e,t){var i,a,u,l,g;if(e.c.length==0)return new Ya(rt(0),rt(0));for(i=(Gt(0,e.c.length),f(e.c[0],11)).j,g=0,l=t.g,a=t.g+1;g<e.c.length-1&&i.g<l;)++g,i=(Gt(g,e.c.length),f(e.c[g],11)).j;for(u=g;u<e.c.length-1&&i.g<a;)++u,i=(Gt(g,e.c.length),f(e.c[g],11)).j;return new Ya(rt(g),rt(u))}function fPn(e,t,i,a){var u,l,g,m,v,S,T;v=lc(t,i),(i==(tt(),Bn)||i==pn)&&(v=Bd(v)),g=!1;do for(u=!1,l=0;l<v.gc()-1;l++)S=f(v.Xb(l),11),m=f(v.Xb(l+1),11),hNn(e,S,m,a)&&(g=!0,Xce(e.a,f(v.Xb(l),11),f(v.Xb(l+1),11)),T=f(v.Xb(l+1),11),v.fd(l+1,f(v.Xb(l),11)),v.fd(l,T),u=!0);while(u);return g}function hPn(e,t,i){var a,u,l,g;for(i.Sg(Nxt,1),u=f($d(Jr(new Ft(null,new Bt(t.b,16)),new IC),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),SEt(e,u,0),g=tr(t.b,0);g.b!=g.d.c;)l=f(Zn(g),39),a=bn(e.a,rt(l.g))!=null?f(bn(e.a,rt(l.g)),17).a:0,$e(l,(Ec(),Lm),rt(a));i.Tg()}function oQ(e,t,i){var a,u,l,g,m,v,S,T;return m=t.i-e.g/2,v=i.i-e.g/2,S=t.j-e.g/2,T=i.j-e.g/2,l=t.g+e.g,g=i.g+e.g,a=t.f+e.g,u=i.f+e.g,m<v+g&&v<m&&S<T+u&&T<S||v<m+l&&m<v&&T<S+a&&S<T||m<v+g&&v<m&&S<T&&T<S+a?!0:v<m+l&&m<v&&S<T+u&&T<S}function pPn(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(l=t.c.length,g=(Gt(i,t.c.length),f(t.c[i],291)),m=g.a.o.a,A=g.c,L=0,S=g.c;S<=g.f;S++){if(m<=e.a[S])return S;for(T=e.a[S],v=null,u=i+1;u<l;u++)a=(Gt(u,t.c.length),f(t.c[u],291)),a.c<=S&&a.f>=S&&(v=a);v&&(T=h.Math.max(T,v.a.o.a)),T>L&&(A=S,L=T)}return A}function gPn(e){var t,i,a,u,l,g;for(i=new ne(e.b);i.a<i.c.c.length;)for(t=f(ee(i),25),l=!1,g=!1,u=new ne(t.a);u.a<u.c.c.length;)if(a=f(ee(u),9),!l&&li(a,(ot(),gD))&&($e(t,(ot(),gD),f(V(a,gD),17)),l=!0),!g&&li(a,(ot(),eF))&&($e(t,(ot(),eF),ut(V(a,eF))),g=!0),l&&g)return}function mPn(e){var t,i,a,u,l,g,m;for(l=new g0(f(or(new K_),51)),m=qo,i=new ne(e.d);i.a<i.c.c.length;){for(t=f(ee(i),225),m=t.c.c;l.a.gc()!=0&&(g=f(l.a.Rc(),225),g.c.c+g.c.b<m);)l.a.Ac(g)!=null;for(u=l.a.ec().Jc();u.Ob();)a=f(u.Pb(),225),Mr(a.b,t),Mr(t.b,a);l.a.yc(t,(mn(),yw))==null}}function Z2t(e,t,i){var a,u,l,g,m;if(!NP(t)){for(m=i.bh((me(t,16)?f(t,16).gc():g1(t.Jc()))/e.a|0),m.Sg(Rxt,1),g=new W6,l=null,u=t.Jc();u.Ob();)a=f(u.Pb(),39),g=Df(ce(se(Gf,1),Jt,20,0,[g,new u1(a)])),l&&($e(l,(_i(),T5t),a),$e(a,Xme,l),lue(a)==lue(l)&&($e(l,Jme,a),$e(a,Qte,l))),l=a;m.Tg(),Z2t(e,g,i)}}function bPn(e,t){var i,a,u;if(t==null){for(a=(!e.a&&(e.a=new Be(j1,e,9,5)),new Nn(e.a));a.e!=a.i.gc();)if(i=f(Mn(a),679),u=i.c,(u??i.zb)==null)return i}else for(a=(!e.a&&(e.a=new Be(j1,e,9,5)),new Nn(e.a));a.e!=a.i.gc();)if(i=f(Mn(a),679),Mt(t,(u=i.c,u??i.zb)))return i;return null}function Ade(e,t){var i;switch(i=null,t.g){case 1:e.e.nf((Rr(),Xbe))&&(i=f(e.e.mf(Xbe),255));break;case 3:e.e.nf((Rr(),Jbe))&&(i=f(e.e.mf(Jbe),255));break;case 2:e.e.nf((Rr(),Ybe))&&(i=f(e.e.mf(Ybe),255));break;case 4:e.e.nf((Rr(),Zbe))&&(i=f(e.e.mf(Zbe),255))}return!i&&(i=f(e.e.mf((Rr(),x9e)),255)),i}function Q2t(e,t,i){var a,u,l,g,m,v;for(u=i,l=0,m=new ne(t);m.a<m.c.c.length;)g=f(ee(m),27),pi(g,(uw(),one),rt(u++)),v=RT(g),a=h.Math.atan2(g.j+g.f/2,g.i+g.g/2),a+=a<0?_E:0,a<.7853981633974483||a>zxt?na(v,e.b):a<=zxt&&a>Hxt?na(v,e.d):a<=Hxt&&a>Gxt?na(v,e.c):a<=Gxt&&na(v,e.a),l=Q2t(e,v,l);return u}function eSt(e,t,i,a){var u,l,g,m,v,S;for(u=(a.c+a.a)/2,nf(t.j),Mr(t.j,u),nf(i.e),Mr(i.e,u),S=new bit,m=new ne(e.f);m.a<m.c.c.length;)l=f(ee(m),133),v=l.a,tde(S,t,v),tde(S,i,v);for(g=new ne(e.k);g.a<g.c.c.length;)l=f(ee(g),133),v=l.b,tde(S,t,v),tde(S,i,v);return S.b+=2,S.a+=aft(t,e.q),S.a+=aft(e.q,i),S}function vPn(e,t,i){var a;i.Sg("Processor arrange node",1),lt(ut(V(t,(Ec(),hMe)))),a=f(Dd(S1(Jr(new Ft(null,new Bt(t.b,16)),new RC))),39),e.a=f(V(t,SMe),350),e.a==(wj(),tbe)||e.a==nne?lkt(e,new bu(ce(se(Zte,1),eee,39,0,[a])),i.bh(1)):e.a==ebe&&jkt(e,new bu(ce(se(Zte,1),eee,39,0,[a])),i.bh(1)),i.Tg()}function tSt(e){var t,i,a,u,l,g,m;for(i=e.i,t=e.n,m=i.d,e.f==(pl(),Sw)?m+=(i.a-e.e.b)/2:e.f==np&&(m+=i.a-e.e.b),u=new ne(e.d);u.a<u.c.c.length;){switch(a=f(ee(u),187),g=a.Kf(),l=new ha,l.b=m,m+=g.b+e.a,e.b.g){case 0:l.a=i.c+t.b;break;case 1:l.a=i.c+t.b+(i.b-g.a)/2;break;case 2:l.a=i.c+i.b-t.c-g.a}a.Mf(l)}}function nSt(e){var t,i,a,u,l,g,m;for(i=e.i,t=e.n,m=i.c,e.b==(Yl(),Ew)?m+=(i.b-e.e.a)/2:e.b==tp&&(m+=i.b-e.e.a),u=new ne(e.d);u.a<u.c.c.length;){switch(a=f(ee(u),187),g=a.Kf(),l=new ha,l.a=m,m+=g.a+e.a,e.f.g){case 0:l.b=i.d+t.d;break;case 1:l.b=i.d+t.d+(i.a-g.b)/2;break;case 2:l.b=i.d+i.a-t.a-g.b}a.Mf(l)}}function wPn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;T=i.a.c,g=i.a.c+i.a.b,l=f(bn(i.c,t),454),F=l.f,z=l.a,v=new Ke(T,F),A=new Ke(g,z),u=T,i.p||(u+=e.c),u+=i.F+i.v*e.b,S=new Ke(u,F),L=new Ke(u,z),dj(t.a,ce(se(Zo,1),Ye,8,0,[v,S])),m=i.d.a.gc()>1,m&&(a=new Ke(u,i.b),Mr(t.a,a)),dj(t.a,ce(se(Zo,1),Ye,8,0,[L,A]))}function CIe(e,t,i){var a,u;for(t<e.d.b.c.length?(e.b=f(ct(e.d.b,t),25),e.a=f(ct(e.d.b,t-1),25),e.c=t):(e.a=new ru(e.d),e.a.p=t-1,Qe(e.d.b,e.a),e.b=new ru(e.d),e.b.p=t,Qe(e.d.b,e.b),e.c=t),jo(i,e.b),u=new Tn(Cn(Fo(i).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),!a.c.i.c&&a.c.i.k==(cn(),Vc)&&jo(a.c.i,e.a)}function yPn(e,t){var i,a,u,l;for(l=lc(t,(tt(),Bn)).Jc();l.Ob();)a=f(l.Pb(),11),i=f(V(a,(Ge(),ed)),9),i&&Zh(Uh(Bh(zh(Fh(new ah,0),.1),e.i[t.p].d),e.i[i.p].a));for(u=lc(t,hn).Jc();u.Ob();)a=f(u.Pb(),11),i=f(V(a,(Ge(),ed)),9),i&&Zh(Uh(Bh(zh(Fh(new ah,0),.1),e.i[i.p].d),e.i[t.p].a))}function rSt(e){Wk(e,new zS(qk(zk(Gk(Hk(new ug,aee),"ELK Randomizer"),'Distributes the nodes randomly on the plane, leading to very obfuscating layouts. Can be useful to demonstrate the power of "real" layout algorithms.'),new XO))),Je(e,aee,zT,s7e),Je(e,aee,UT,15),Je(e,aee,AQ,rt(0)),Je(e,aee,ML,DL)}function AIe(){AIe=Z;var e,t,i,a,u,l;for(KF=Pe($l,hA,30,255,15,1),Vne=Pe(Xf,_m,30,16,15,1),t=0;t<255;t++)KF[t]=-1;for(i=57;i>=48;i--)KF[i]=i-48<<24>>24;for(a=70;a>=65;a--)KF[a]=a-65+10<<24>>24;for(u=102;u>=97;u--)KF[u]=u-97+10<<24>>24;for(l=0;l<10;l++)Vne[l]=48+l&To;for(e=10;e<=15;e++)Vne[e]=65+e-10&To}function EPn(e,t){t.Sg("Process graph bounds",1),$e(e,(_i(),Kme),Fz(Bue(uT(new Ft(null,new Bt(e.b,16)),new G6)))),$e(e,Yme,Fz(Bue(uT(new Ft(null,new Bt(e.b,16)),new NO)))),$e(e,aMe,Fz(Fue(uT(new Ft(null,new Bt(e.b,16)),new D8)))),$e(e,sMe,Fz(Fue(uT(new Ft(null,new Bt(e.b,16)),new RO)))),t.Tg()}function SPn(e){var t,i,a,u,l;u=f(V(e,(ot(),DE)),22),l=f(V(e,$te),22),i=new Ke(e.f.a+e.d.b+e.d.c,e.f.b+e.d.d+e.d.a),t=new ja(i),u.Gc((Ud(),mx))&&(a=f(V(e,mD),8),l.Gc((yd(),jD))&&(a.a<=0&&(a.a=20),a.b<=0&&(a.b=20)),t.a=h.Math.max(i.a,a.a),t.b=h.Math.max(i.b,a.b)),lt(ut(V(e,vme)))||zDn(e,i,t)}function _Pn(e){var t,i,a,u,l,g,m;for(t=!1,i=0,u=new ne(e.d.b);u.a<u.c.c.length;)for(a=f(ee(u),25),a.p=i++,g=new ne(a.a);g.a<g.c.c.length;)l=f(ee(g),9),!t&&!NP(T1(l))&&(t=!0);m=xi((ko(),$g),ce(se(NF,1),Ve,86,0,[Ds,Gs])),t||(Yh(m,Wf),Yh(m,jg)),e.a=new Lmt(m),nu(e.f),nu(e.b),nu(e.e),nu(e.g)}function Ide(e){var t,i,a,u,l,g;if(!e.c){if(g=new lg,t=GF,l=t.a.yc(e,t),l==null){for(a=new Nn(iu(e));a.e!=a.i.gc();)i=f(Mn(a),87),u=uQ(i),me(u,88)&&Qi(g,Ide(f(u,29))),Qn(g,i);t.a.Ac(e)!=null,t.a.gc()==0}qCn(g),ET(g),e.c=new _3((f(be(je((Wv(),sn).o),15),19),g.i),g.g),pd(e).b&=-33}return e.c}function IIe(e){var t;if(e.c!=10)throw ie(new vr(yr((lr(),dee))));switch(t=e.a,t){case 110:t=10;break;case 114:t=13;break;case 116:t=9;break;case 92:case 124:case 46:case 94:case 45:case 63:case 42:case 43:case 123:case 125:case 40:case 41:case 91:case 93:break;default:throw ie(new vr(yr((lr(),Hf))))}return t}function iSt(e){var t,i,a,u,l;if(e.l==0&&e.m==0&&e.h==0)return"0";if(e.h==zG&&e.m==0&&e.l==0)return"-9223372036854775808";if(e.h>>19!=0)return"-"+iSt(VP(e));for(i=e,a="";!(i.l==0&&i.m==0&&i.h==0);){if(u=aue(SQ),i=dOe(i,u,!0),t=""+Iit(ww),!(i.l==0&&i.m==0&&i.h==0))for(l=9-t.length;l>0;l--)t="0"+t;a=t+a}return a}function kPn(){if(!Object.create||!Object.getOwnPropertyNames)return!1;var e="__proto__",t=Object.create(null);if(t[e]!==void 0)return!1;var i=Object.getOwnPropertyNames(t);return!(i.length!=0||(t[e]=42,t[e]!==42)||Object.getOwnPropertyNames(t).length==0)}function TPn(e,t,i){var a,u,l,g,m,v,S,T,A;for(a=i.c,u=i.d,m=wg(t.c),v=wg(t.d),a==t.c?(m=hIe(e,m,u),v=ayt(t.d)):(m=ayt(t.c),v=hIe(e,v,u)),S=new mY(t.a),Wi(S,m,S.a,S.a.a),Wi(S,v,S.c.b,S.c),g=t.c==a,A=new ort,l=0;l<S.b-1;++l)T=new Ya(f(Xh(S,l),8),f(Xh(S,l+1),8)),g&&l==0||!g&&l==S.b-2?A.b=T:Qe(A.a,T);return A}function xPn(e,t){var i,a,u,l;if(l=e.j.g-t.j.g,l!=0)return l;if(i=f(V(e,(ot(),H0)),17),a=f(V(t,H0),17),i&&a&&(u=i.a-a.a,u!=0))return u;switch(e.j.g){case 1:return di(e.n.a,t.n.a);case 2:return di(e.n.b,t.n.b);case 3:return di(t.n.a,e.n.a);case 4:return di(t.n.b,e.n.b);default:throw ie(new As(sNe))}}function OIe(e,t,i,a){var u,l,g,m,v;if(g1((aH(),new Tn(Cn(T1(t).a.Jc(),new B))))>=e.a||!cAe(t,i))return-1;if(NP(f(a.Kb(t),20)))return 1;for(u=0,g=f(a.Kb(t),20).Jc();g.Ob();)if(l=f(g.Pb(),18),v=l.c.i==t?l.d.i:l.c.i,m=OIe(e,v,i,a),m==-1||(u=h.Math.max(u,m),u>e.c-1))return-1;return u+1}function Ag(){Ag=Z,lne=new pa((Rr(),ID),1.3),KPt=new pa(hx,(mn(),!1)),u8e=new gS(15),EF=new pa(V0,u8e),SF=new pa(zE,15),GPt=tV,WPt=UE,YPt=GA,XPt=Ow,VPt=HA,_be=oV,JPt=px,h8e=(KIe(),UPt),f8e=BPt,Tbe=HPt,p8e=zPt,c8e=jPt,kbe=MPt,s8e=DPt,d8e=FPt,o8e=iV,qPt=Kbe,Yq=RPt,i8e=NPt,Xq=PPt,l8e=$Pt,a8e=LPt}function oSt(e,t){var i,a,u,l,g,m;if(Ae(t)===Ae(e))return!0;if(!me(t,15)||(a=f(t,15),m=e.gc(),a.gc()!=m))return!1;if(g=a.Jc(),e.Vi()){for(i=0;i<m;++i)if(u=e.Si(i),l=g.Pb(),u==null?l!=null:!ei(u,l))return!1}else for(i=0;i<m;++i)if(u=e.Si(i),l=g.Pb(),Ae(u)!==Ae(l))return!1;return!0}function aSt(e,t){var i,a,u,l,g,m;if(e.f>0){if(e.Yj(),t!=null){for(l=0;l<e.d.length;++l)if(i=e.d[l],i){for(a=f(i.g,371),m=i.i,g=0;g<m;++g)if(u=a[g],ei(t,u.kd()))return!0}}else for(l=0;l<e.d.length;++l)if(i=e.d[l],i){for(a=f(i.g,371),m=i.i,g=0;g<m;++g)if(u=a[g],Ae(t)===Ae(u.kd()))return!0}}return!1}function CPn(e,t){var i,a,u;return i=t.mi(e.a),i&&(u=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),"affiliation")),u!=null)?(a=sX(u,gl(35)),a==-1?fle(e,q7(e,cf(t.nk())),u):a==0?fle(e,null,(gn(1,u.length+1),u.substr(1))):fle(e,(ga(0,a,u.length),u.substr(0,a)),(gn(a+1,u.length+1),u.substr(a+1)))):null}function aQ(){aQ=Z,Age=new eR("NORTH",0,(tt(),hn),hn),Nge=new eR("SOUTH",1,Bn,Bn),Cge=new eR("EAST",2,_n,_n),Lge=new eR("WEST",3,pn,pn),Oge=new eR("NORTH_WEST_CORNER",4,pn,hn),Ige=new eR("NORTH_EAST_CORNER",5,hn,_n),Pge=new eR("SOUTH_WEST_CORNER",6,Bn,pn),Rge=new eR("SOUTH_EAST_CORNER",7,_n,Bn)}function APn(e,t,i){var a,u,l,g;i.Sg("Orthogonally routing hierarchical port edges",1),e.a=0,a=dMn(t),v8n(t,a),i8n(e,t,a),m9n(t),u=f(V(t,(ot(),Xi)),100),l=t.b,akt((Gt(0,l.c.length),f(l.c[0],25)),u,t),akt(f(ct(l,l.c.length-1),25),u,t),g=t.b,w_t((Gt(0,g.c.length),f(g.c[0],25))),w_t(f(ct(g,g.c.length-1),25)),i.Tg()}function NIe(e){switch(e){case 48:case 49:case 50:case 51:case 52:case 53:case 54:case 55:case 56:case 57:return e-48<<24>>24;case 97:case 98:case 99:case 100:case 101:case 102:return e-97+10<<24>>24;case 65:case 66:case 67:case 68:case 69:case 70:return e-65+10<<24>>24;default:throw ie(new sm("Invalid hexadecimal"))}}function sSt(e,t,i,a){var u,l,g,m,v,S;for(v=ede(e,i),S=ede(t,i),u=!1;v&&S&&(a||LAn(v,S,i));)g=ede(v,i),m=ede(S,i),PH(t),PH(e),l=v.c,efe(v,!1),efe(S,!1),i?(sw(t,S.p,l),t.p=S.p,sw(e,v.p+1,l),e.p=v.p):(sw(e,v.p,l),e.p=v.p,sw(t,S.p+1,l),t.p=S.p),jo(v,null),jo(S,null),v=g,S=m,u=!0;return u}function cSt(e){switch(e.g){case 0:return new kl;case 1:return new lN;case 3:return new Aot;case 4:return new og;case 5:return new iut;case 6:return new ul;case 2:return new tS;case 7:return new uN;case 8:return new eu;default:throw ie(new dn("No implementation is available for the layerer "+(e.f!=null?e.f:""+e.g)))}}function IPn(e,t,i,a){var u,l,g,m,v;for(u=!1,l=!1,m=new ne(a.j);m.a<m.c.c.length;)g=f(ee(m),11),Ae(V(g,(Ge(),oi)))===Ae(i)&&(g.g.c.length==0?g.e.c.length==0||(u=!0):l=!0);return v=0,u&&u^l?v=i.j==(tt(),hn)?-e.e[a.c.p][a.p]:t-e.e[a.c.p][a.p]:l&&u^l?v=e.e[a.c.p][a.p]+1:u&&l&&(v=i.j==(tt(),hn)?0:t/2),v}function Ode(e,t,i,a,u,l,g,m){var v,S,T;for(v=0,t!=null&&(v^=C0(t.toLowerCase())),i!=null&&(v^=C0(i)),a!=null&&(v^=C0(a)),g!=null&&(v^=C0(g)),m!=null&&(v^=C0(m)),S=0,T=l.length;S<T;S++)v^=C0(l[S]);e?v|=256:v&=-257,u?v|=16:v&=-17,this.f=v,this.i=t==null?null:(an(t),t),this.a=i,this.d=a,this.j=l,this.g=g,this.e=m}function RIe(e,t,i){var a,u;switch(u=null,t.g){case 1:u=(Nl(),w5e);break;case 2:u=(Nl(),E5e)}switch(a=null,i.g){case 1:a=(Nl(),y5e);break;case 2:a=(Nl(),v5e);break;case 3:a=(Nl(),S5e);break;case 4:a=(Nl(),_5e)}return u&&a?WN(e.j,new G9(new bu(ce(se(G7n,1),Jt,177,0,[f(or(u),177),f(or(a),177)])))):(Ht(),Ht(),Ga)}function OPn(e){var t,i,a;switch(t=f(V(e,(ot(),mD)),8),$e(e,mD,new Ke(t.b,t.a)),f(V(e,Nm),254).g){case 1:$e(e,Nm,(vb(),wne));break;case 2:$e(e,Nm,(vb(),bne));break;case 3:$e(e,Nm,(vb(),Qq));break;case 4:$e(e,Nm,(vb(),eV))}(e.q?e.q:(Ht(),Ht(),Cb))._b(sx)&&(i=f(V(e,sx),8),a=i.a,i.a=i.b,i.b=a)}function uSt(e,t,i,a,u,l){if(this.b=i,this.d=u,e>=t.length)throw ie(new Ru("Greedy SwitchDecider: Free layer not in graph."));this.c=t[e],this.e=new oH(a),Lue(this.e,this.c,(tt(),pn)),this.i=new oH(a),Lue(this.i,this.c,_n),this.f=new blt(this.c),this.a=!l&&u.i&&!u.s&&this.c[0].k==(cn(),bo),this.a&&nRn(this,e,t.length)}function lSt(e,t){var i,a,u,l,g,m;l=!e.B.Gc((yd(),bV)),g=e.B.Gc(s1e),e.a=new X0t(g,l,e.c),e.n&&rTe(e.a.n,e.n),Eae(e.g,(Np(),Bu),e.a),t||(a=new Oj(1,l,e.c),a.n.a=e.k,ZN(e.p,(tt(),hn),a),u=new Oj(1,l,e.c),u.n.d=e.k,ZN(e.p,Bn,u),m=new Oj(0,l,e.c),m.n.c=e.k,ZN(e.p,pn,m),i=new Oj(0,l,e.c),i.n.b=e.k,ZN(e.p,_n,i))}function NPn(e){var t,i,a;switch(t=f(V(e.d,(ot(),P1)),220),t.g){case 2:i=c7n(e);break;case 3:i=(a=new et,Zi(Jr(Lu(Zs(Zs(new Ft(null,new Bt(e.d.b,16)),new mC),new yO),new L2),new bp),new Bet(a)),a);break;default:throw ie(new As("Compaction not supported for "+t+" edges."))}PMn(e,i),Oa(new Ic(e.g),new Met(e))}function RPn(e,t){var i,a,u,l,g,m,v;if(t.Sg("Process directions",1),i=f(V(e,(Ec(),E_)),86),i!=(ko(),jg))for(u=tr(e.b,0);u.b!=u.d.c;){switch(a=f(Zn(u),39),m=f(V(a,(_i(),Vq)),17).a,v=f(V(a,Wq),17).a,i.g){case 4:v*=-1;break;case 1:l=m,m=v,v=l;break;case 2:g=m,m=-v,v=g}$e(a,Vq,rt(m)),$e(a,Wq,rt(v))}t.Tg()}function PPn(e){var t,i,a,u,l,g,m,v;for(v=new Npt,m=new ne(e.a);m.a<m.c.c.length;)if(g=f(ee(m),9),g.k!=(cn(),bo)){for(G4n(v,g,new ha),l=new Tn(Cn(ao(g).a.Jc(),new B));Kn(l);)if(u=f(Dn(l),18),!(u.c.i.k==bo||u.d.i.k==bo))for(a=tr(u.a,0);a.b!=a.d.c;)i=f(Zn(a),8),t=i,Dj(v,new lP(t.a,t.b))}return v}function Nde(){Nde=Z,B8e=new Si(Whe),F8e=(g7(),Zq),$8e=new Pt(Xhe,F8e),j8e=(XH(),mne),ELt=new Pt(m4e,j8e),M8e=(bG(),Pbe),yLt=new Pt(b4e,M8e),bLt=new Pt(Khe,null),D8e=(CH(),pne),wLt=new Pt(Yhe,D8e),L8e=(TY(),Obe),hLt=new Pt(v4e,L8e),pLt=new Pt(w4e,(mn(),!1)),gLt=new Pt(y4e,rt(64)),mLt=new Pt(E4e,!0),vLt=Rbe}function dSt(e,t){var i,a,u,l,g,m,v,S,T,A;for(e.p=1,u=e.c,A=new ub,T=$S(e,(cs(),Uu)).Jc();T.Ob();)for(S=f(T.Pb(),11),a=new ne(S.g);a.a<a.c.c.length;)i=f(ee(a),18),v=i.d.i,e!=v&&(l=v.c,l.p<=u.p&&(g=u.p+1,g==t.b.c.length?(m=new ru(t),m.p=g,Qe(t.b,m),jo(v,m)):(m=f(ct(t.b,g),25),jo(v,m)),A.a.yc(v,A)));return A}function LPn(e,t){var i,a;if(i=f(V(e,(_i(),qme)),15),!i||i.gc()<1)return null;if(i.gc()==1)return f(i.Xb(0),39);switch(a=null,t.g){case 2:a=f(Dd(XN(i.Mc(),new OO)),39);break;case 1:a=f(Dd(oT(i.Mc(),new CC)),39);break;case 4:a=f(Dd(XN(i.Mc(),new IO)),39);break;case 3:a=f(Dd(oT(i.Mc(),new AC)),39)}return a}function fSt(e){var t,i,a,u,l,g;if(e.a==null)if(e.a=Pe(kd,Tb,30,e.c.b.c.length,16,1),e.a[0]=!1,li(e.c,(ot(),kme)))for(a=f(V(e.c,kme),15),i=a.Jc();i.Ob();)t=f(i.Pb(),17).a,t>0&&t<e.a.length&&(e.a[t]=!1);else for(g=new ne(e.c.b),g.a<g.c.c.length&&ee(g),u=1;g.a<g.c.c.length;)l=f(ee(g),25),e.a[u++]=y6n(l)}function wb(){wb=Z,wF=new Si("additionalHeight"),CD=new Si("drawingHeight"),t4=new Si("drawingWidth"),cne=new Si("minHeight"),yF=new Si("minWidth"),une=new Si("rows"),fx=new Si("targetWidth"),Ebe=new Mi("minRowIncrease",0),IPt=new Mi("maxRowIncrease",0),ybe=new Mi("minRowDecrease",0),APt=new Mi("maxRowDecrease",0)}function hSt(e,t){var i,a,u,l;switch(u=e.b,t){case 1:{e.b|=1,e.b|=4,e.b|=8;break}case 2:{e.b|=2,e.b|=4,e.b|=8;break}case 4:{e.b|=1,e.b|=2,e.b|=4,e.b|=8;break}case 3:{e.b|=16,e.b|=8;break}case 0:{e.b|=32,e.b|=16,e.b|=8,e.b|=1,e.b|=2,e.b|=4;break}}if(e.b!=u&&e.c)for(a=new Nn(e.c);a.e!=a.i.gc();)l=f(Mn(a),468),i=pd(l),OT(i,t)}function pSt(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z;for(u=!1,g=t,m=0,v=g.length;m<v;++m)l=g[m],lt((mn(),!!l.e))&&!f(ct(e.b,l.e.p),217).s&&(u=u|(S=l.e,T=f(ct(e.b,S.p),217),A=T.e,L=dit(i,A.length),F=A[L][0],F.k==(cn(),bo)?A[L]=d5n(l,A[L],i?(tt(),pn):(tt(),_n)):T.c.kg(A,i),z=TG(e,T,i,a),kIe(T.e,T.o,i),z));return u}function gSt(e,t){var i,a,u,l,g;for(l=(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i,u=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));u.e!=u.i.gc();)a=f(Mn(u),27),Ae(Fe(a,(Rr(),UA)))!==Ae((_1(),PF))&&(g=f(Fe(t,VA),149),i=f(Fe(a,VA),149),(g==i||g&&zTe(g,i))&&(!a.a&&(a.a=new Be(Ur,a,10,11)),a.a).i!=0&&(l+=gSt(e,a)));return l}function DPn(e){var t,i,a,u,l;for(l=new Du(e.a.c.length),u=new ne(e.a);u.a<u.c.c.length;){switch(a=f(ee(u),9),i=f(V(a,(ot(),el)),171),t=null,i.g){case 1:case 2:t=(lE(),SA);break;case 3:case 4:t=(lE(),lD)}t?($e(a,(Ge(),Ste),(lE(),SA)),t==lD?nQ(a,i,(cs(),td)):t==SA&&nQ(a,i,(cs(),Uu))):vn(l.c,a)}return l}function MPn(e){var t,i,a,u,l,g,m;for(a=0,m=0,g=new ne(e.d);g.a<g.c.c.length;)l=f(ee(g),105),u=f($d(Jr(new Ft(null,new Bt(l.j,16)),new Rd),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),i=null,a<=m?(i=(tt(),hn),a+=u.gc()):m<a&&(i=(tt(),Bn),m+=u.gc()),t=i,Zi(Lu(u.Mc(),new vv),new Het(t))}function jPn(e,t){var i;return i=new k2,t&&Rc(i,f(bn(e.a,wV),102)),me(t,273)&&Rc(i,f(bn(e.a,yV),102)),me(t,359)?(Rc(i,f(bn(e.a,hc),102)),i):(me(t,84)&&Rc(i,f(bn(e.a,Yn),102)),me(t,206)?(Rc(i,f(bn(e.a,Ur),102)),i):me(t,193)?(Rc(i,f(bn(e.a,_d),102)),i):(me(t,268)&&Rc(i,f(bn(e.a,vo),102)),i))}function $Pn(e){var t,i,a,u,l,g,m,v;for(e.b=new TEt(new bu((tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn]))),new bu((LS(),ce(se(xge,1),Ve,365,0,[c_,_w,s_])))),g=ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn]),m=0,v=g.length;m<v;++m)for(l=g[m],i=ce(se(xge,1),Ve,365,0,[c_,_w,s_]),a=0,u=i.length;a<u;++a)t=i[a],XIn(e.b,l,t,new et)}function mSt(e,t){var i,a,u,l,g,m,v,S,T,A;if(g=f(f(si(e.r,t),22),83),m=e.u.Gc((Xl(),Lw)),i=e.u.Gc(jF),a=e.u.Gc(MF),S=e.u.Gc(u4),A=e.B.Gc((yd(),Ine)),T=!i&&!a&&(S||g.gc()==2),M5n(e,t),u=null,v=null,m){for(l=g.Jc(),u=f(l.Pb(),115),v=u;l.Ob();)v=f(l.Pb(),115);u.d.b=0,v.d.c=0,T&&!u.a&&(u.d.c=0)}A&&(WOn(g),m&&(u.d.b=0,v.d.c=0))}function bSt(e,t){var i,a,u,l,g,m,v,S,T,A;if(g=f(f(si(e.r,t),22),83),m=e.u.Gc((Xl(),Lw)),i=e.u.Gc(jF),a=e.u.Gc(MF),v=e.u.Gc(u4),A=e.B.Gc((yd(),Ine)),S=!i&&!a&&(v||g.gc()==2),JLn(e,t),T=null,u=null,m){for(l=g.Jc(),T=f(l.Pb(),115),u=T;l.Ob();)u=f(l.Pb(),115);T.d.d=0,u.d.a=0,S&&!T.a&&(T.d.a=0)}A&&(KOn(g),m&&(T.d.d=0,u.d.a=0))}function vSt(e,t,i){var a,u,l,g,m,v,S,T;if(u=t.k,t.p>=0)return!1;if(t.p=i.b,Qe(i.e,t),u==(cn(),uo)||u==xu){for(g=new ne(t.j);g.a<g.c.c.length;)for(l=f(ee(g),11),T=(a=new ne(new _N(l).a.g),new QEe(a));sc(T.a);)if(S=f(ee(T.a),18).d,m=S.i,v=m.k,t.c!=m.c&&(v==uo||v==xu)&&vSt(e,m,i))return!0}return!0}function sQ(e){var t;return(e.Db&64)!=0?XAe(e):(t=new sh(XAe(e)),t.a+=" (changeable: ",m0(t,(e.Bb&ep)!=0),t.a+=", volatile: ",m0(t,(e.Bb&pw)!=0),t.a+=", transient: ",m0(t,(e.Bb&FT)!=0),t.a+=", defaultValueLiteral: ",ys(t,e.j),t.a+=", unsettable: ",m0(t,(e.Bb&Ml)!=0),t.a+=", derived: ",m0(t,(e.Bb&Sm)!=0),t.a+=")",t.a)}function FPn(e,t){var i,a,u,l,g;return u=t.mi(e.a),u&&(a=(!u.b&&(u.b=new Md((Ut(),Za),xc,u)),u.b),i=dr(Rp(a,Ta)),i!=null&&(l=i.lastIndexOf("#"),g=l==-1?b_e(e,t.gk(),i):l==0?OH(e,null,(gn(1,i.length+1),i.substr(1))):OH(e,(ga(0,l,i.length),i.substr(0,l)),(gn(l+1,i.length+1),i.substr(l+1))),me(g,155)))?f(g,155):null}function BPn(e,t){var i,a,u,l,g;return a=t.mi(e.a),a&&(i=(!a.b&&(a.b=new Md((Ut(),Za),xc,a)),a.b),l=dr(Rp(i,kpe)),l!=null&&(u=l.lastIndexOf("#"),g=u==-1?b_e(e,t.gk(),l):u==0?OH(e,null,(gn(1,l.length+1),l.substr(1))):OH(e,(ga(0,u,l.length),l.substr(0,u)),(gn(u+1,l.length+1),l.substr(u+1))),me(g,155)))?f(g,155):null}function UPn(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(t.Sg("Restoring reversed edges",1),v=new ne(e.b);v.a<v.c.c.length;)for(m=f(ee(v),25),T=new ne(m.a);T.a<T.c.c.length;)for(S=f(ee(T),9),L=new ne(S.j);L.a<L.c.c.length;)for(A=f(ee(L),11),g=pm(A.g),a=g,u=0,l=a.length;u<l;++u)i=a[u],lt(ut(V(i,(Ge(),U0))))&&VS(i,!1);t.Tg()}function zPn(e,t,i,a){var u,l,g,m,v;for(v=Pe(sa,Ye,108,(tt(),ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn])).length,0,2),l=ce(se(Ja,1),qc,64,0,[fc,hn,_n,Bn,pn]),g=0,m=l.length;g<m;++g)u=l[g],v[u.g]=Pe(sa,_s,30,e.c[u.g],15,1);return Cwt(v,e,hn),Cwt(v,e,Bn),Vle(v,e,hn,t,i,a),Vle(v,e,_n,t,i,a),Vle(v,e,Bn,t,i,a),Vle(v,e,pn,t,i,a),v}function HPn(e,t,i){if(Yu(e.a,t)){if(xp(f(bn(e.a,t),50),i))return 1}else qr(e.a,t,new Mo);if(Yu(e.a,i)){if(xp(f(bn(e.a,i),50),t))return-1}else qr(e.a,i,new Mo);if(Yu(e.b,t)){if(xp(f(bn(e.b,t),50),i))return-1}else qr(e.b,t,new Mo);if(Yu(e.b,i)){if(xp(f(bn(e.b,i),50),t))return 1}else qr(e.b,i,new Mo);return 0}function GPn(e){var t,i,a,u,l,g;e.q==(oa(),Lb)||e.q==uu||(u=e.f.n.d+_X(f(Es(e.b,(tt(),hn)),126))+e.c,t=e.f.n.a+_X(f(Es(e.b,Bn),126))+e.c,a=f(Es(e.b,_n),126),g=f(Es(e.b,pn),126),l=h.Math.max(0,a.n.d-u),l=h.Math.max(l,g.n.d-u),i=h.Math.max(0,a.n.a-t),i=h.Math.max(i,g.n.a-t),a.n.d=l,g.n.d=l,a.n.a=i,g.n.a=i)}function PIe(e,t,i,a){var u,l,g,m,v,S;if(i==null){for(u=f(e.g,122),m=0;m<e.i;++m)if(g=u[m],g.Ik()==t)return za(e,g,a)}return l=(is(),f(t,69).uk()?f(i,74):mb(t,i)),Ld(e.e)?(S=!oG(e,t),a=$u(e,l,a),v=t.Gk()?Yv(e,3,t,null,i,EL(e,t,i,me(t,101)&&(f(t,19).Bb&Xa)!=0),S):Yv(e,1,t,t.fk(),i,-1,S),a?a.kj(v):a=v):a=$u(e,l,a),a}function wSt(){this.b=new Zv,this.d=new Zv,this.e=new Zv,this.c=new Zv,this.a=new Wn,this.f=new Wn,L3(Zo,new s0,new Y8),L3(J8e,new Iy,new X8),L3(m5e,new HC,new WO),L3(b5e,new Cf,new J8),L3(PDt,new c5,new GC),L3(q7n,new xf,new Vu),L3(K7n,new HO,new GO),L3(V7n,new qO,new zC),L3(W7n,new X2,new fo),L3(J7n,new Ek,new VO)}function Vj(e,t){var i,a,u,l,g;for(e=e==null?bl:(an(e),e),u=0;u<t.length;u++)t[u]=V6n(t[u]);for(i=new RN,g=0,a=0;a<t.length&&(l=e.indexOf("%s",g),l!=-1);)i.a+=""+uh(e==null?bl:(an(e),e),g,l),vu(i,t[a++]),g=l+2;if(Rht(i,e,g,e.length),a<t.length){for(i.a+=" [",vu(i,t[a++]);a<t.length;)i.a+=Fu,vu(i,t[a++]);i.a+="]"}return i.a}function LIe(e,t){var i,a,u,l,g,m,v;for(i=0,v=new ne(t);v.a<v.c.c.length;){for(m=f(ee(v),11),_Ce(e.b,e.d[m.p]),g=0,u=new yg(m.b);sc(u.a)||sc(u.b);)a=f(sc(u.a)?ee(u.a):ee(u.b),18),Tdt(a)?(l=V2e(e,m==a.c?a.d:a.c),l>e.d[m.p]&&(i+=FTe(e.b,l),Jv(e.a,rt(l)))):++g;for(i+=e.b.d*g;!u7(e.a);)vxe(e.b,f(JN(e.a),17).a)}return i}function ySt(e){var t,i,a,u,l,g;return l=0,t=ph(e),t.hk()&&(l|=4),(e.Bb&Ml)!=0&&(l|=2),me(e,101)?(i=f(e,19),u=os(i),(i.Bb&Dc)!=0&&(l|=32),u&&(jn(dT(u)),l|=8,g=u.t,(g>1||g==-1)&&(l|=16),(u.Bb&Dc)!=0&&(l|=64)),(i.Bb&Xa)!=0&&(l|=pw),l|=ep):me(t,455)?l|=512:(a=t.hk(),a&&(a.i&1)!=0&&(l|=256)),(e.Bb&512)!=0&&(l|=128),l}function qPn(e,t){var i;return e.f==y1e?(i=kS(Ns((Rl(),Ea),t)),e.e?i==4&&t!=(_R(),f4)&&t!=(_R(),d4)&&t!=(_R(),E1e)&&t!=(_R(),S1e):i==2):e.d&&(e.d.Gc(t)||e.d.Gc(rR(Ns((Rl(),Ea),t)))||e.d.Gc(iA((Rl(),Ea),e.b,t)))?!0:e.f&&mIe((Rl(),e.f),gH(Ns(Ea,t)))?(i=kS(Ns(Ea,t)),e.e?i==4:i==2):!1}function VPn(e,t){var i,a,u,l,g,m,v,S;for(l=new et,t.b.c.length=0,i=f($d(vTe(new Ft(null,new Bt(new Ic(e.a.b),1))),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),u=i.Jc();u.Ob();)if(a=f(u.Pb(),17),g=RTe(e.a,a),g.b!=0)for(m=new ru(t),vn(l.c,m),m.p=a.a,S=tr(g,0);S.b!=S.d.c;)v=f(Zn(S),9),jo(v,m);$o(t.b,l)}function WPn(e){var t,i,a,u,l,g,m,v,S,T,A,L,F;for(L=-1,F=0,S=e,T=0,A=S.length;T<A;++T){for(v=S[T],l=v,g=0,m=l.length;g<m;++g)for(u=l[g],t=new ITe(L==-1?e[0]:e[L],f(V(ra(u),(ot(),ip)),266),ywt(u),lt(ut(V(ra(u),ume)))),i=0;i<u.j.c.length;i++)for(a=i+1;a<u.j.c.length;a++)qlt(t,f(ct(u.j,i),11),f(ct(u.j,a),11))>0&&++F;++L}return F}function KPn(e,t,i,a){var u,l,g,m,v,S,T,A;return g=f(Fe(i,(Rr(),o4)),8),v=g.a,T=g.b+e,u=h.Math.atan2(T,v),u<0&&(u+=_E),u+=t,u>_E&&(u-=_E),m=f(Fe(a,o4),8),S=m.a,A=m.b+e,l=h.Math.atan2(A,S),l<0&&(l+=_E),l+=t,l>_E&&(l-=_E),gg(),Kh(1e-10),h.Math.abs(u-l)<=1e-10||u==l||isNaN(u)&&isNaN(l)?0:u<l?-1:u>l?1:pS(isNaN(u),isNaN(l))}function Rde(e){var t,i,a,u,l,g,m;for(m=new Wn,a=new ne(e.a.b);a.a<a.c.c.length;)t=f(ee(a),60),qr(m,t,new et);for(u=new ne(e.a.b);u.a<u.c.c.length;)for(t=f(ee(u),60),t.i=qo,g=t.c.Jc();g.Ob();)l=f(g.Pb(),60),f(ac(Is(m.f,l)),15).Ec(t);for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),60),t.c.$b(),t.c=f(ac(Is(m.f,t)),15);H2t(e)}function Pde(e){var t,i,a,u,l,g,m;for(m=new Wn,a=new ne(e.a.b);a.a<a.c.c.length;)t=f(ee(a),82),qr(m,t,new et);for(u=new ne(e.a.b);u.a<u.c.c.length;)for(t=f(ee(u),82),t.o=qo,g=t.f.Jc();g.Ob();)l=f(g.Pb(),82),f(ac(Is(m.f,l)),15).Ec(t);for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),82),t.f.$b(),t.f=f(ac(Is(m.f,t)),15);R2t(e)}function YPn(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(v=ra(t.a),u=ke(Oe(V(v,(ot(),jE))))*2,T=ke(Oe(V(v,NA))),S=h.Math.max(u,T),l=Pe(sa,_s,30,t.f-t.c+1,15,1),a=-S,i=0,m=t.b.Jc();m.Ob();)g=f(m.Pb(),9),a+=e.a[g.c.p]+S,l[i++]=a;for(a+=e.a[t.a.c.p]+S,l[i++]=a,L=new ne(t.e);L.a<L.c.c.length;)A=f(ee(L),9),a+=e.a[A.c.p]+S,l[i++]=a;return l}function XPn(e,t,i,a){var u,l,g,m,v,S,T,A,L;for(L=new g0(new utt(e)),m=ce(se(Ib,1),j0,9,0,[t,i]),v=0,S=m.length;v<S;++v)for(g=m[v],A=kj(g,a).Jc();A.Ob();)for(T=f(A.Pb(),11),l=new yg(T.b);sc(l.a)||sc(l.b);)u=f(sc(l.a)?ee(l.a):ee(l.b),18),Fa(u)||(L.a.yc(T,(mn(),yw))==null,Tdt(u)&&fH(L,T==u.c?u.d:u.c));return or(L),new Kl(L)}function DIe(e,t,i,a){var u,l,g;t&&(l=ke(Oe(V(t,(_i(),q0))))+a,g=i+ke(Oe(V(t,ene)))/2,$e(t,Vq,rt(mr(Nc(h.Math.round(l))))),$e(t,Wq,rt(mr(Nc(h.Math.round(g))))),t.d.b==0||DIe(e,f(oX((u=tr(new u1(t).a.d,0),new m3(u))),39),i+ke(Oe(V(t,ene)))+e.b,a+ke(Oe(V(t,xD)))),V(t,Jme)!=null&&DIe(e,f(V(t,Jme),39),i,a))}function JPn(e,t){var i,a,u,l;if(l=f(Fe(e,(Rr(),qA)),64).g-f(Fe(t,qA),64).g,l!=0)return l;if(i=f(Fe(e,Qbe),17),a=f(Fe(t,Qbe),17),i&&a&&(u=i.a-a.a,u!=0))return u;switch(f(Fe(e,qA),64).g){case 1:return di(e.i,t.i);case 2:return di(e.j,t.j);case 3:return di(t.i,e.i);case 4:return di(t.j,e.j);default:throw ie(new As(sNe))}}function MIe(e){var t,i,a;return(e.Db&64)!=0?fde(e):(t=new ef(P4e),i=e.k,i?xr(xr((t.a+=' "',t),i),'"'):(!e.n&&(e.n=new Be(hc,e,1,7)),e.n.i>0&&(a=(!e.n&&(e.n=new Be(hc,e,1,7)),f(be(e.n,0),153)).a,!a||xr(xr((t.a+=' "',t),a),'"'))),xr(lS(xr(lS(xr(lS(xr(lS((t.a+=" (",t),e.i),","),e.j)," | "),e.g),","),e.f),")"),t.a)}function ESt(e){var t,i,a;return(e.Db&64)!=0?fde(e):(t=new ef(L4e),i=e.k,i?xr(xr((t.a+=' "',t),i),'"'):(!e.n&&(e.n=new Be(hc,e,1,7)),e.n.i>0&&(a=(!e.n&&(e.n=new Be(hc,e,1,7)),f(be(e.n,0),153)).a,!a||xr(xr((t.a+=' "',t),a),'"'))),xr(lS(xr(lS(xr(lS(xr(lS((t.a+=" (",t),e.i),","),e.j)," | "),e.g),","),e.f),")"),t.a)}function ZPn(e,t){var i,a,u,l,g;for(t==(Ej(),jme)&&vG(f(si(e.a,(CT(),Cq)),15)),u=f(si(e.a,(CT(),Cq)),15).Jc();u.Ob();)switch(a=f(u.Pb(),105),i=f(ct(a.j,0),112).d.j,l=new Kl(a.j),na(l,new dk),t.g){case 2:ode(e,l,i,(LS(),_w),1);break;case 1:case 0:g=q6n(l),ode(e,new Xv(l,0,g),i,(LS(),_w),0),ode(e,new Xv(l,g,l.c.length),i,_w,1)}}function QPn(e){var t,i,a,u,l,g,m;for(u=f(V(e,(Ge(),l_)),9),a=e.j,i=(Gt(0,a.c.length),f(a.c[0],11)),g=new ne(u.j);g.a<g.c.c.length;)if(l=f(ee(g),11),Ae(l)===Ae(V(i,oi))){l.j==(tt(),hn)&&e.p>u.p?(No(l,Bn),l.d&&(m=l.o.b,t=l.a.b,l.a.b=m-t)):l.j==Bn&&u.p>e.p&&(No(l,hn),l.d&&(m=l.o.b,t=l.a.b,l.a.b=-(m-t)));break}return u}function jIe(e,t){var i,a,u,l,g,m,v;if(t==null||t.length==0)return null;if(u=f(wu(e.a,t),149),!u){for(a=(m=new kp(e.b).a.vc().Jc(),new zy(m));a.a.Ob();)if(i=(l=f(a.a.Pb(),43),f(l.kd(),149)),g=i.c,v=t.length,Mt(g.substr(g.length-v,v),t)&&(t.length==g.length||_a(g,g.length-t.length-1)==46)){if(u)return null;u=i}u&&Os(e.a,t,u)}return u}function wL(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(l=new Ke(t,i),T=new ne(e.a);T.a<T.c.c.length;)for(S=f(ee(T),9),ii(S.n,l),L=new ne(S.j);L.a<L.c.c.length;)for(A=f(ee(L),11),u=new ne(A.g);u.a<u.c.c.length;)for(a=f(ee(u),18),wT(a.a,l),g=f(V(a,(ot(),Ls)),77),g&&wT(g,l),v=new ne(a.b);v.a<v.c.c.length;)m=f(ee(v),72),ii(m.n,l)}function eLn(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(l=new Ke(t,i),T=new ne(e.a);T.a<T.c.c.length;)for(S=f(ee(T),9),ii(S.n,l),L=new ne(S.j);L.a<L.c.c.length;)for(A=f(ee(L),11),u=new ne(A.g);u.a<u.c.c.length;)for(a=f(ee(u),18),wT(a.a,l),g=f(V(a,(ot(),Ls)),77),g&&wT(g,l),v=new ne(a.b);v.a<v.c.c.length;)m=f(ee(v),72),ii(m.n,l)}function SSt(e){if((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b).i==0)throw ie(new ON("Edges must have a source."));if((!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c).i==0)throw ie(new ON("Edges must have a target."));if(!e.b&&(e.b=new Qt(Yn,e,4,7)),!(e.b.i<=1&&(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c.i<=1)))throw ie(new ON("Hyperedges are not supported."))}function $Ie(e,t){var i,a,u,l,g,m,v,S,T,A;for(A=0,l=new O3,Jv(l,t);l.b!=l.c;)for(v=f(JN(l),217),S=0,T=f(V(t.j,(ot(),ip)),266),g=ke(Oe(V(t.j,Ate))),m=ke(Oe(V(t.j,CLe))),T!=(Pp(),Pm)&&(S+=g*l6n(v.e,T),S+=m*WPn(v.e)),A+=cwt(v.d,v.e)+S,u=new ne(v.b);u.a<u.c.c.length;)a=f(ee(u),36),i=f(ct(e.b,a.p),217),i.s||(A+=VZ(e,i));return A}function _St(e){var t,i,a;li(e,(ot(),p_))&&(a=f(V(e,p_),22),!a.dc()&&(i=(t=f(Tp(ps),10),new Rf(t,f(Gh(t,t.length),10),0)),a.Gc((NT(),bf))?Yh(i,bf):Yh(i,up),a.Gc(Eh)||Yh(i,Eh),a.Gc(sp)?Yh(i,lp):a.Gc(D1)?Yh(i,Pb):a.Gc(cp)&&Yh(i,Sh),a.Gc(lp)?Yh(i,sp):a.Gc(Pb)?Yh(i,D1):a.Gc(Sh)&&Yh(i,cp),$e(e,p_,i)))}function kSt(e,t){var i,a;for(this.b=new et,this.e=new et,this.a=e,this.d=t,W3n(this),r3n(this),this.b.dc()?this.c=e.c.p:this.c=f(this.b.Xb(0),9).c.p,this.e.c.length==0?this.f=e.c.p:this.f=f(ct(this.e,this.e.c.length-1),9).c.p,a=f(V(e,(Ge(),Y$)),15).Jc();a.Ob();)if(i=f(a.Pb(),72),li(i,(ot(),Nte))){this.d=f(V(i,Nte),231);break}}function zd(e,t,i){var a,u,l,g,m,v,S,T;for(a=f(bn(e.b,t),50),l=f(bn(e.b,i),50),u=f(bn(e.f,t),50),g=f(bn(e.f,i),50),a.a.yc(i,a),g.a.yc(t,g),T=l.a.ec().Jc();T.Ob();)S=f(T.Pb(),9),a.a.yc(S,a),Go(f(bn(e.f,S),50),t),Ba(f(bn(e.f,S),50),u);for(v=u.a.ec().Jc();v.Ob();)m=f(v.Pb(),9),g.a.yc(m,g),Go(f(bn(e.b,m),50),i),Ba(f(bn(e.b,m),50),l)}function xG(e,t,i){var a,u,l,g,m,v,S,T;for(a=f(bn(e.a,t),50),l=f(bn(e.a,i),50),u=f(bn(e.b,t),50),g=f(bn(e.b,i),50),a.a.yc(i,a),g.a.yc(t,g),T=l.a.ec().Jc();T.Ob();)S=f(T.Pb(),9),a.a.yc(S,a),Go(f(bn(e.b,S),50),t),Ba(f(bn(e.b,S),50),u);for(v=u.a.ec().Jc();v.Ob();)m=f(v.Pb(),9),g.a.yc(m,g),Go(f(bn(e.a,m),50),i),Ba(f(bn(e.a,m),50),l)}function ym(){ym=Z;var e;for(Oee=new m1(1,1),Wpe=new m1(1,10),M$=new m1(0,0),HAt=new m1(-1,1),GAt=ce(se(mA,1),Ye,90,0,[M$,Oee,new m1(1,2),new m1(1,3),new m1(1,4),new m1(1,5),new m1(1,6),new m1(1,7),new m1(1,8),new m1(1,9),Wpe]),Nee=Pe(mA,Ye,90,32,0,1),e=0;e<Nee.length;e++)Nee[e]=ZY(hb(1,e),0)?aw(hb(1,e)):X7(aw(T0(hb(1,e))))}function TSt(e,t,i,a,u,l,g){if(e.c=a.Jf().a,e.d=a.Jf().b,u&&(e.c+=u.Jf().a,e.d+=u.Jf().b),e.b=t.Kf().a,e.a=t.Kf().b,!u)i?e.c-=g+t.Kf().a:e.c+=a.Kf().a+g;else switch(u.$f().g){case 0:case 2:e.c+=u.Kf().a+g+l.a+g;break;case 4:e.c-=g+l.a+g+t.Kf().a;break;case 1:e.c+=u.Kf().a+g,e.d-=g+l.b+g+t.Kf().b;break;case 3:e.c+=u.Kf().a+g,e.d+=u.Kf().b+g+l.b+g}}function tLn(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z;if(l=i,i<a)for(L=(F=new VH(e.p),z=new VH(e.p),Ba(F.e,e.e),F.q=e.q,F.r=z,zX(F),Ba(z.j,e.j),z.r=F,zX(z),new Ya(F,z)),A=f(L.a,116),T=f(L.b,116),u=(Gt(l,t.c.length),f(t.c[l],335)),g=eSt(e,A,T,u),S=i+1;S<=a;S++)m=(Gt(S,t.c.length),f(t.c[S],335)),v=eSt(e,A,T,m),q3n(m,v,u,g)&&(u=m,g=v,l=S);return l}function nLn(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z;for(g=f(be(t,0),27),md(g,0),bd(g,0),L=new et,vn(L.c,g),m=g,l=new pke(e.a,g.g,g.f,(Hj(),Jq)),F=1;F<t.i;F++)z=f(be(t,F),27),v=Gde(e,$A,z,m,l,L,i),S=Gde(e,n4,z,m,l,L,i),T=Gde(e,kF,z,m,l,L,i),A=Gde(e,_F,z,m,l,L,i),u=kMn(e,v,S,T,A,z,m,a),md(z,u.d),bd(z,u.e),WK(u,Jq),l=u,m=z,vn(L.c,z);return l}function rLn(e,t,i){var a,u,l,g,m,v,S,T,A,L;if(T=null,L=t,A=nbt(e,Xht(i),L),FP(A,p1(L,Am)),g=IS(L,M4e),a=new Pat(e,A),rEt(a.a,a.b,g),m=IS(L,fpe),u=new Lat(e,A),iEt(u.a,u.b,m),(!A.b&&(A.b=new Qt(Yn,A,4,7)),A.b).i==0||(!A.c&&(A.c=new Qt(Yn,A,5,8)),A.c).i==0)throw l=p1(L,Am),v=zCt+l,S=v+YL,ie(new am(S));return UZ(L,A),e7n(e,L,A),T=Zue(e,L,A),T}function iLn(e,t){var i,a,u,l,g,m,v;for(u=Pe(pr,Pr,30,e.e.a.c.length,15,1),g=new ne(e.e.a);g.a<g.c.c.length;)l=f(ee(g),123),u[l.d]+=l.b.a.c.length;for(m=cH(t);m.b!=0;)for(l=f(m.b==0?null:(Ln(m.b!=0),lh(m,m.a.a)),123),a=cR(new ne(l.g.a));a.Ob();)i=f(a.Pb(),216),v=i.e,v.e=h.Math.max(v.e,l.e+i.a),--u[v.d],u[v.d]==0&&Wi(m,v,m.c.b,m.c)}function xSt(e){var t,i,a,u,l,g,m,v,S,T,A;for(i=Ra,u=Qr,m=new ne(e.e.a);m.a<m.c.c.length;)l=f(ee(m),123),u=h.Math.min(u,l.e),i=h.Math.max(i,l.e);for(t=Pe(pr,Pr,30,i-u+1,15,1),g=new ne(e.e.a);g.a<g.c.c.length;)l=f(ee(g),123),l.e-=u,++t[l.e];if(a=0,e.k!=null)for(S=e.k,T=0,A=S.length;T<A&&(v=S[T],t[a++]+=v,t.length!=a);++T);return t}function oLn(e,t){var i,a,u,l,g,m;if(t.Sg("Edge routing",1),u=f(V(e,(Ec(),Zme)),383),u==(yj(),Gme))g_n(e);else if(u==Gq)for(f(Dd(S1(Jr(new Ft(null,new Bt(e.b,16)),new AO))),39),l=ke(Oe(V(e,yMe))),g=ke(Oe(V(e,gMe))),m=f(V(e,E_),86),n7n(e,m,l),P7n(e,m,l,g),M7n(e,m,l,g),a=tr(e.a,0);a.b!=a.d.c;)i=f(Zn(a),65),i.a.b<2&&zAe(i);t.Tg()}function CSt(e){switch(e.d){case 9:case 8:return!0;case 3:case 5:case 4:case 6:return!1;case 7:return f(TIe(e),17).a==e.o;case 1:case 2:{if(e.o==-2)return!1;switch(e.p){case 0:case 1:case 2:case 6:case 5:case 7:return O7(e.k,e.f);case 3:case 4:return e.j==e.e;default:return e.n==null?e.g==null:ei(e.n,e.g)}}default:return!1}}function ASt(e,t,i){var a,u,l,g,m,v,S,T,A;for(l=new Du(t.c.length),S=new ne(t);S.a<S.c.c.length;)g=f(ee(S),9),Qe(l,e.b[g.c.p][g.p]);for(DMn(e,l,i),A=null;A=O8n(l);)EDn(e,f(A.a,238),f(A.b,238),l);for(t.c.length=0,u=new ne(l);u.a<u.c.c.length;)for(a=f(ee(u),238),m=a.d,v=0,T=m.length;v<T;++v)g=m[v],vn(t.c,g),e.a[g.c.p][g.p].a=mg(a.g,a.d[0]).a}function aLn(e,t){var i,a,u;switch(t.Sg("Breaking Point Insertion",1),a=new nIe(e),f(V(e,(ot(),_me)),347).g){case 2:u=new Qc;break;case 0:u=new fk;break;default:u=new kO}if(i=u.mg(e,a),lt(ut(V(e,dDe)))&&(i=qMn(e,i)),!u.ng()&&li(e,Bte))switch(f(V(e,Bte),348).g){case 2:i=nEt(a,i);break;case 1:i=Vwt(a,i)}if(i.dc()){t.Tg();return}s7n(e,i),t.Tg()}function ISt(e){Wk(e,new zS(qk(zk(Gk(Hk(new ug,y$),"ELK Fixed"),"Keeps the current layout as it is, without any automatic modification. Optional coordinates can be given for nodes and edge bend points."),new YO))),Je(e,y$,zT,n7e),Je(e,y$,YQ,st(RF)),Je(e,y$,T4e,st(Z9e)),Je(e,y$,uA,st(Q9e)),Je(e,y$,DR,st(t7e)),Je(e,y$,jL,st(e7e))}function cQ(e,t,i){var a,u,l,g,m;if(a=mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))),m=mr(Na(Sb,gb(mr(Na(i==null?0:Ti(i),_b)),15))),l=pj(e,t,a),l&&m==l.f&&dm(i,l.i))return i;if(g=gj(e,i,m),g)throw ie(new dn("value already present: "+i));return u=new qX(t,a,i,m),l?(tA(e,l),yG(e,u,l),l.e=null,l.c=null,l.i):(yG(e,u,null),Awt(e),null)}function sLn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;T=i.a.c,g=i.a.c+i.a.b,l=f(bn(i.c,t),454),F=l.f,z=l.a,l.b?v=new Ke(g,F):v=new Ke(T,F),l.c?A=new Ke(T,z):A=new Ke(g,z),u=T,i.p||(u+=e.c),u+=i.F+i.v*e.b,S=new Ke(u,F),L=new Ke(u,z),dj(t.a,ce(se(Zo,1),Ye,8,0,[v,S])),m=i.d.a.gc()>1,m&&(a=new Ke(u,i.b),Mr(t.a,a)),dj(t.a,ce(se(Zo,1),Ye,8,0,[L,A]))}function gh(){gh=Z,qte=new Yk(Lp,0),Fq=new Yk("NIKOLOV",1),Bq=new Yk("NIKOLOV_PIXEL",2),kDe=new Yk("NIKOLOV_IMPROVED",3),TDe=new Yk("NIKOLOV_IMPROVED_PIXEL",4),_De=new Yk("DUMMYNODE_PERCENTAGE",5),xDe=new Yk("NODECOUNT_PERCENTAGE",6),Vte=new Yk("NO_BOUNDARY",7),w_=new Yk("MODEL_ORDER_LEFT_TO_RIGHT",8),ux=new Yk("MODEL_ORDER_RIGHT_TO_LEFT",9)}function Lde(e,t){var i,a,u,l,g,m,v,S,T,A,L,F;return T=null,L=aIe(e,t),a=null,m=f(Fe(t,(Rr(),YLt)),297),m?a=m:a=(lj(),pV),F=a,F==(lj(),pV)&&(u=null,S=f(bn(e.r,L),297),S?u=S:u=a1e,F=u),qr(e.r,t,F),l=null,v=f(Fe(t,KLt),275),v?l=v:l=(XP(),cV),A=l,A==(XP(),cV)&&(g=null,i=f(bn(e.b,L),275),i?g=i:g=_ne,A=g),T=f(qr(e.b,t,A),275),T}function cLn(e){var t,i,a,u,l;for(a=e.length,t=new l7,l=0;l<a;)if(i=_a(e,l++),!(i==9||i==10||i==12||i==13||i==32)){if(i==35){for(;l<a&&(i=_a(e,l++),!(i==13||i==10)););continue}i==92&&l<a?(u=(gn(l,e.length),e.charCodeAt(l)))==35||u==9||u==10||u==12||u==13||u==32?(_P(t,u&To),++l):(t.a+="\\",_P(t,u&To),++l):_P(t,i&To)}return t.a}function FIe(){FIe=Z,iPt=new Pt(WRe,(mn(),!1)),cPt=new Pt(KRe,rt(0)),uPt=new Pt(YRe,0),lPt=new Pt(nee,!1),RMe=(ZH(),ine),aPt=new Pt(Uhe,RMe),rt(0),oPt=new Pt(zhe,rt(1)),LMe=(JJ(),pbe),pPt=new Pt(XRe,LMe),DMe=(kJ(),ube),gPt=new Pt(JRe,DMe),PMe=(RZ(),hbe),sPt=new Pt(ZRe,PMe),hPt=new Pt(Hhe,0),dPt=new Pt(Ghe,!1),fPt=new Pt(QRe,!1)}function uLn(e,t){var i,a,u;for(a=new ne(t);a.a<a.c.c.length;)if(i=f(ee(a),27),jt(e.a,i,i),jt(e.b,i,i),u=RT(i),u.c.length!=0)for(e.d&&e.d.Eg(u),jt(e.a,i,(Gt(0,u.c.length),f(u.c[0],27))),jt(e.b,i,f(ct(u,u.c.length-1),27));Kue(u).c.length!=0;)u=Kue(u),e.d&&e.d.Eg(u),jt(e.a,i,(Gt(0,u.c.length),f(u.c[0],27))),jt(e.b,i,f(ct(u,u.c.length-1),27))}function Dde(e,t,i){var a,u,l,g,m,v;if(t)if(i<=-1){if(a=Vt(t.zh(),-1-i),me(a,101))return f(a,19);for(g=f(t.Ih(a),159),m=0,v=g.gc();m<v;++m)if(Ae(g.Rl(m))===Ae(e)&&(u=g.Ql(m),me(u,101)&&(l=f(u,19),(l.Bb&Dc)!=0)))return l;throw ie(new As("The containment feature could not be located"))}else return os(f(Vt(e.zh(),i),19));else return null}function lLn(e){var t,i,a,u,l,g,m,v,S,T;for(i=0,m=new ne(e.d);m.a<m.c.c.length;)g=f(ee(m),105),g.i&&(g.i.c=i++);for(t=iT(kd,[Ye,Tb],[183,30],16,[i,i],2),T=e.d,u=0;u<T.c.length;u++)if(v=(Gt(u,T.c.length),f(T.c[u],105)),v.i)for(l=u+1;l<T.c.length;l++)S=(Gt(l,T.c.length),f(T.c[l],105)),S.i&&(a=rNn(v,S),t[v.i.c][S.i.c]=a,t[S.i.c][v.i.c]=a);return t}function Wj(){Wj=Z,Hme=new hS("ROOT_PROC",0),YDe=new hS("FAN_PROC",1),QDe=new hS("LEVEL_PROC",2),eMe=new hS("NEIGHBORS_PROC",3),ZDe=new hS("LEVEL_HEIGHT",4),KDe=new hS("DIRECTION_PROC",5),tMe=new hS("NODE_POSITION_PROC",6),VDe=new hS("COMPACTION_PROC",7),JDe=new hS("LEVEL_COORDS",8),XDe=new hS("GRAPH_BOUNDS_PROC",9),WDe=new hS("DETREEIFYING_PROC",10)}function BIe(e,t){var i,a,u,l,g,m,v,S,T,A;for(A=Bs(t),S=null,u=!1,m=0,T=iu(A.a).i;m<T;++m)g=f(RG(A,m,(l=f(be(iu(A.a),m),87),v=l.c,me(v,88)?f(v,29):(Ut(),Ch))),29),i=BIe(e,g),i.dc()||(S?(u||(u=!0,S=new OX(S)),S.Fc(i)):S=i);return a=URn(e,t),a.dc()?S||(Ht(),Ht(),Ga):S?(u||(S=new OX(S)),S.Fc(a),S):a}function Mde(e,t){var i,a,u,l,g,m,v,S,T,A;for(A=Bs(t),S=null,a=!1,m=0,T=iu(A.a).i;m<T;++m)l=f(RG(A,m,(u=f(be(iu(A.a),m),87),v=u.c,me(v,88)?f(v,29):(Ut(),Ch))),29),i=Mde(e,l),i.dc()||(S?(a||(a=!0,S=new OX(S)),S.Fc(i)):S=i);return g=p4n(e,t),g.dc()?S||(Ht(),Ht(),Ga):S?(a||(S=new OX(S)),S.Fc(g),S):g}function dLn(e,t){var i,a,u,l,g;for(e.c==null||e.c.length<t.c.length?e.c=Pe(kd,Tb,30,t.c.length,16,1):c7(e.c),e.a=new et,a=0,g=new ne(t);g.a<g.c.c.length;)u=f(ee(g),9),u.p=a++;for(i=new ki,l=new ne(t);l.a<l.c.c.length;)u=f(ee(l),9),e.c[u.p]||(cEt(e,u),i.b==0||(Ln(i.b!=0),f(i.a.a.c,15)).gc()<e.a.c.length?S3(i,e.a):Ma(i,e.a),e.a=new et);return i}function CG(e,t,i){var a,u,l,g,m,v;if(me(t,74))return za(e,t,i);for(m=null,l=null,a=f(e.g,122),g=0;g<e.i;++g)if(u=a[g],ei(t,u.kd())&&(l=u.Ik(),me(l,101)&&(f(l,19).Bb&Dc)!=0)){m=u;break}return m&&(Ld(e.e)&&(v=l.Gk()?Yv(e,4,l,t,null,EL(e,l,t,me(l,101)&&(f(l,19).Bb&Xa)!=0),!0):Yv(e,l.qk()?2:1,l,t,l.fk(),-1,!0),i?i.kj(v):i=v),i=CG(e,m,i)),i}function fLn(e,t,i){var a,u,l,g;if(g=Ju(e.e.zh(),t),a=f(e.g,122),is(),f(t,69).uk()){for(l=0;l<e.i;++l)if(u=a[l],g.Zl(u.Ik())&&ei(u,i))return PT(e,l),!0}else if(i!=null){for(l=0;l<e.i;++l)if(u=a[l],g.Zl(u.Ik())&&ei(i,u.kd()))return PT(e,l),!0}else for(l=0;l<e.i;++l)if(u=a[l],g.Zl(u.Ik())&&u.kd()==null)return PT(e,l),!0;return!1}function OSt(e,t,i,a){var u,l,g,m,v,S,T,A,L,F;for(g=eT(t.c,i,a),A=new ne(t.a);A.a<A.c.c.length;){for(T=f(ee(A),9),ii(T.n,g),F=new ne(T.j);F.a<F.c.c.length;)for(L=f(ee(F),11),l=new ne(L.g);l.a<l.c.c.length;)for(u=f(ee(l),18),wT(u.a,g),m=f(V(u,(ot(),Ls)),77),m&&wT(m,g),S=new ne(u.b);S.a<S.c.c.length;)v=f(ee(S),72),ii(v.n,g);Qe(e.a,T),T.a=e}}function hLn(e,t){var i,a,u,l,g;if(t.Sg("Node and Port Label Placement and Node Sizing",1),Cst((h7(),new Oce(e,!0,!0,new _U))),f(V(e,(Ge(),Cu)),22).Gc((us(),Vf)))for(l=f(V(e,(ot(),ax)),22),u=l.Gc((Xl(),hV)),g=lt(ut(V(e,tDe))),a=new ne(e.b);a.a<a.c.c.length;)i=f(ee(a),25),Zi(Jr(new Ft(null,new Bt(i.a,16)),new Zm),new Dut(l,u,g));t.Tg()}function NSt(e){Wk(e,new zS(qk(zk(Gk(Hk(new ug,HL),"ELK SPOrE Overlap Removal"),'A node overlap removal algorithm proposed by Nachmanson et al. in "Node overlap removal by growing a tree".'),new a0))),Je(e,HL,Whe,st(G8e)),Je(e,HL,zT,H8e),Je(e,HL,UT,8),Je(e,HL,Xhe,st(_Lt)),Je(e,HL,y4e,st(U8e)),Je(e,HL,E4e,st(z8e)),Je(e,HL,aq,(mn(),!1))}function pLn(e,t){var i,a,u,l,g,m,v;if(i=t.mi(e.a),i&&(v=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),n6e)),v!=null)){for(a=new et,l=DT(v,"\\w"),g=0,m=l.length;g<m;++g)u=l[g],Mt(u,"##other")?Qe(a,"!##"+q7(e,cf(t.nk()))):Mt(u,"##local")?a.c.push(null):Mt(u,pq)?Qe(a,q7(e,cf(t.nk()))):vn(a.c,u);return a}return Ht(),Ht(),Ga}function nA(e,t,i,a){this.e=e,this.k=f(V(e,(Ge(),CA)),313),this.g=Pe(Ib,j0,9,t,0,1),this.b=Pe(mo,Ye,341,t,7,1),this.a=Pe(Ib,j0,9,t,0,1),this.d=Pe(mo,Ye,341,t,7,1),this.j=Pe(Ib,j0,9,t,0,1),this.i=Pe(mo,Ye,341,t,7,1),this.p=Pe(mo,Ye,341,t,7,1),this.n=Pe(to,Ye,471,t,8,1),Pz(this.n,(mn(),!1)),this.f=Pe(to,Ye,471,t,8,1),Pz(this.f,!0),this.o=i,this.c=a}function AG(e){var t,i,a,u,l,g,m,v;if(e.d)throw ie(new As((d1(gge),zfe+gge.k+Hfe)));for(e.c==(ko(),$g)&&oA(e,Ds),i=new ne(e.a.a);i.a<i.c.c.length;)t=f(ee(i),194),t.e=0;for(g=new ne(e.a.b);g.a<g.c.c.length;)for(l=f(ee(g),82),l.o=qo,u=l.f.Jc();u.Ob();)a=f(u.Pb(),82),++a.d.e;for(o9n(e),v=new ne(e.a.b);v.a<v.c.c.length;)m=f(ee(v),82),m.k=!0;return e}function gLn(e,t){var i,a,u,l,g,m,v,S;for(m=new Fyt(e),i=new ki,Wi(i,t,i.c.b,i.c);i.b!=0;){for(a=f(i.b==0?null:(Ln(i.b!=0),lh(i,i.a.a)),112),a.d.p=1,g=new ne(a.e);g.a<g.c.c.length;)u=f(ee(g),336),eyt(m,u),S=u.d,S.d.p==0&&Wi(i,S,i.c.b,i.c);for(l=new ne(a.b);l.a<l.c.c.length;)u=f(ee(l),336),eyt(m,u),v=u.c,v.d.p==0&&Wi(i,v,i.c.b,i.c)}return m}function RSt(e){var t,i,a,u,l;if(a=ke(Oe(Fe(e,(Rr(),QLt)))),a!=1)for(x3(e,a*e.g,a*e.f),i=Ngn(rvn((!e.c&&(e.c=new Be(_d,e,9,9)),e.c),new Q8)),l=pb(Df(ce(se(Gf,1),Jt,20,0,[(!e.n&&(e.n=new Be(hc,e,1,7)),e.n),(!e.c&&(e.c=new Be(_d,e,9,9)),e.c),i])));Kn(l);)u=f(Dn(l),273),u.oh(a*u.lh(),a*u.mh()),u.nh(a*u.kh(),a*u.jh()),t=f(u.mf(A9e),8),t&&(t.a*=a,t.b*=a)}function UIe(e,t,i){var a,u,l,g,m;if(g=(is(),f(t,69).uk()),x1(e.e,t)){if(t.Pi()&&lQ(e,t,i,me(t,101)&&(f(t,19).Bb&Xa)!=0))return!1}else for(m=Ju(e.e.zh(),t),a=f(e.g,122),l=0;l<e.i;++l)if(u=a[l],m.Zl(u.Ik()))return(g?ei(u,i):i==null?u.kd()==null:ei(i,u.kd()))?!1:(f(Y3(e,l,g?f(i,74):mb(t,i)),74),!0);return Qn(e,g?f(i,74):mb(t,i))}function mLn(e,t,i,a,u){var l,g,m,v,S,T,A,L;for(g=new ne(e.b);g.a<g.c.c.length;)for(l=f(ee(g),25),L=EH(l.a),S=L,T=0,A=S.length;T<A;++T)switch(v=S[T],f(V(v,(ot(),el)),171).g){case 1:B5n(v),jo(v,t),Bvt(v,!0,a);break;case 3:T5n(v),jo(v,i),Bvt(v,!1,u)}for(m=new la(e.b,0);m.b<m.d.gc();)(Ln(m.b<m.d.gc()),f(m.d.Xb(m.c=m.b++),25)).a.c.length==0&&hd(m)}function bLn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(F=t.length,v=F,gn(0,t.length),t.charCodeAt(0)==45?(A=-1,L=1,--F):(A=1,L=0),l=(qde(),VAt)[10],u=F/l|0,J=F%l,J!=0&&++u,m=Pe(pr,Pr,30,u,15,1),i=qAt[8],g=0,z=L+(J==0?l:J),W=L;W<v;W=z,z=W+l)a=ff((ga(W,z,t.length),t.substr(W,z-W)),Ra,Qr),S=(vL(),S3e(m,m,g,i)),S+=WCn(m,g,a),m[g++]=S;T=g,e.e=A,e.d=T,e.a=m,nj(e)}function vLn(e){var t,i,a,u,l,g,m,v,S,T,A,L;for(m=e.i,u=lt(ut(V(m,(ot(),ix)))),T=0,a=0,S=new ne(e.g);S.a<S.c.c.length;)v=f(ee(S),18),g=Fa(v),l=g&&u&&lt(ut(V(v,PE))),L=v.d.i,g&&l?++a:g&&!l?++T:ra(L).e==m?++a:++T;for(i=new ne(e.e);i.a<i.c.c.length;)t=f(ee(i),18),g=Fa(t),l=g&&u&&lt(ut(V(t,PE))),A=t.c.i,g&&l?++T:g&&!l?++a:ra(A).e==m?++T:++a;return T-a}function PSt(e,t){var i,a,u,l,g,m;if(!t.dc())if(f(t.Xb(0),291).d==(ER(),QT))gIn(e,t);else for(a=t.Jc();a.Ob();){switch(i=f(a.Pb(),291),i.d.g){case 5:hL(e,i,OCn(e,i));break;case 0:hL(e,i,(g=i.f-i.c+1,m=(g-1)/2|0,i.c+m));break;case 4:hL(e,i,U_n(e,i));break;case 2:q0t(i),hL(e,i,(l=tAe(i),l?i.c:i.f));break;case 1:q0t(i),hL(e,i,(u=tAe(i),u?i.f:i.c))}aNn(i.a)}}function zIe(e,t,i,a){var u,l,g;return g=new aot(t,i),e.a?a?(u=f(ch(f(bn(e.b,t),260)),260),++u.a,g.d=a.d,g.e=a.e,g.b=a,g.c=a,a.e?a.e.c=g:u.b=g,a.d?a.d.b=g:e.a=g,a.d=g,a.e=g):(f(ch(e.e),495).b=g,g.d=e.e,e.e=g,u=f(bn(e.b,t),260),u?(++u.a,l=u.c,l.c=g,g.e=l,u.c=g):(qr(e.b,t,u=new cTe(g)),++e.c)):(e.a=e.e=g,qr(e.b,t,new cTe(g)),++e.c),++e.d,g}function jde(e,t){var i,a,u,l,g;if(t.Sg("Network simplex",1),e.e.a.c.length<1){t.Tg();return}for(l=new ne(e.e.a);l.a<l.c.c.length;)u=f(ee(l),123),u.e=0;for(g=e.e.a.c.length>=40,g&&aDn(e),fMn(e),z5n(e),i=A0t(e),a=0;i&&a<e.f;)_Ln(e,i,J4n(e,i)),i=A0t(e),++a;g&&xOn(e),e.a?s6n(e,xSt(e)):xSt(e),e.b=null,e.d=null,e.p=null,e.c=null,e.g=null,e.i=null,e.n=null,e.o=null,t.Tg()}function wLn(e,t){var i,a,u,l,g,m,v;if(!t.e){for(t.e=!0,a=t.d.a.ec().Jc();a.Ob();){if(i=f(a.Pb(),18),t.o&&t.d.a.gc()<=1){g=t.a.c,m=t.a.c+t.a.b,v=new Ke(g+(m-g)/2,t.b),Mr(f(t.d.a.ec().Jc().Pb(),18).a,v);continue}if(u=f(bn(t.c,i),454),u.b||u.c){sLn(e,i,t);continue}l=e.d==(JP(),cF)&&(u.d||u.e)&&k4n(e,t)&&t.d.a.gc()<=1,l?G8n(i,t):wPn(e,i,t)}t.k&&Oa(t.d,new JM)}}function HIe(e,t,i,a,u,l){var g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(L=l,m=(a+u)/2+L,J=i*h.Math.cos(m),ue=i*h.Math.sin(m),pe=J-t.g/2,ve=ue-t.f/2,md(t,pe),bd(t,ve),A=e.a.Cg(t),W=2*h.Math.acos(i/i+e.c),W<u-a?(F=W/A,g=(a+u-W)/2):(F=(u-a)/A,g=a),z=RT(t),e.e&&(e.e.Dg(e.d),e.e.Eg(z)),S=new ne(z);S.a<S.c.c.length;)v=f(ee(S),27),T=e.a.Cg(v),HIe(e,v,i+e.c,g,g+F*T,l),g+=F*T}function yLn(e,t,i){var a;switch(a=i.q.getMonth(),t){case 5:xr(e,ce(se(ft,1),Ye,2,6,["J","F","M","A","M","J","J","A","S","O","N","D"])[a]);break;case 4:xr(e,ce(se(ft,1),Ye,2,6,[gfe,mfe,bfe,vfe,AR,wfe,yfe,Efe,Sfe,_fe,kfe,Tfe])[a]);break;case 3:xr(e,ce(se(ft,1),Ye,2,6,["Jan","Feb","Mar","Apr",AR,"Jun","Jul","Aug","Sep","Oct","Nov","Dec"])[a]);break;default:bb(e,a+1,t)}}function ELn(e,t,i,a){var u,l,g,m,v,S,T,A,L;for(v=new Ke(i,a),Ho(v,f(V(t,(ew(),B$)),8)),L=new ne(t.e);L.a<L.c.c.length;)A=f(ee(L),151),ii(A.d,v),Qe(e.e,A);for(m=new ne(t.c);m.a<m.c.c.length;){for(g=f(ee(m),288),l=new ne(g.a);l.a<l.c.c.length;)u=f(ee(l),249),ii(u.d,v);Qe(e.c,g)}for(T=new ne(t.d);T.a<T.c.c.length;)S=f(ee(T),442),ii(S.d,v),Qe(e.d,S)}function GIe(e,t){var i,a,u,l,g,m,v,S;for(v=new ne(t.j);v.a<v.c.c.length;)for(m=f(ee(v),11),u=new yg(m.b);sc(u.a)||sc(u.b);)a=f(sc(u.a)?ee(u.a):ee(u.b),18),i=a.c==m?a.d:a.c,l=i.i,t!=l&&(S=f(V(a,(ot(),wD)),17).a,S<0&&(S=0),g=l.p,e.b[g]==0&&(a.d==i?(e.a[g]-=S+1,e.a[g]<=0&&e.c[g]>0&&Mr(e.f,l)):(e.c[g]-=S+1,e.c[g]<=0&&e.a[g]>0&&Mr(e.e,l))))}function LSt(e,t,i,a){var u,l,g,m,v,S,T;for(v=new Ke(i,a),Ho(v,f(V(t,(_i(),TD)),8)),T=tr(t.b,0);T.b!=T.d.c;)S=f(Zn(T),39),ii(S.e,v),Mr(e.b,S);for(m=f($d(mTe(new Ft(null,new Bt(t.a,16))),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15).Jc();m.Ob();){for(g=f(m.Pb(),65),l=tr(g.a,0);l.b!=l.d.c;)u=f(Zn(l),8),u.a+=v.a,u.b+=v.b;Mr(e.a,g)}}function qIe(e,t){var i,a,u,l;if(0<(me(e,16)?f(e,16).gc():g1(e.Jc()))){if(u=t,1<u){for(--u,l=new U6,a=e.Jc();a.Ob();)i=f(a.Pb(),39),l=Df(ce(se(Gf,1),Jt,20,0,[l,new u1(i)]));return qIe(l,u)}if(u<0){for(l=new z2,a=e.Jc();a.Ob();)i=f(a.Pb(),39),l=Df(ce(se(Gf,1),Jt,20,0,[l,new u1(i)]));if(0<(me(l,16)?f(l,16).gc():g1(l.Jc())))return qIe(l,u)}}return f(oX(e.Jc()),39)}function SLn(e,t,i){var a,u,l,g;for(i.Sg("Processor order nodes",2),e.b=ke(Oe(V(t,(Ec(),Qme)))),e.a=f(V(t,E_),86),e.a==(ko(),$g)&&(e.a=jg,$e(t,E_,e.a)),u=new ki,g=tr(t.b,0);g.b!=g.d.c;)l=f(Zn(g),39),lt(ut(V(l,(_i(),Iw))))&&Wi(u,l,u.c.b,u.c);a=(Ln(u.b!=0),f(u.a.a.c,39)),I_t(e,a),i.dh(1),DIe(e,a,0-ke(Oe(V(a,(_i(),ene))))/2,0),i.dh(1),i.Tg()}function yd(){yd=Z,jD=new E3("DEFAULT_MINIMUM_SIZE",0),vV=new E3("MINIMUM_SIZE_ACCOUNTS_FOR_PADDING",1),xne=new E3("COMPUTE_PADDING",2),FF=new E3("OUTSIDE_NODE_LABELS_OVERHANG",3),Cne=new E3("PORTS_OVERHANG",4),Ine=new E3("UNIFORM_PORT_SPACING",5),Ane=new E3("SPACE_EFFICIENT_PORT_LABELS",6),s1e=new E3("FORCE_TABULAR_NODE_LABELS",7),bV=new E3("ASYMMETRICAL",8)}function $de(e,t){var i,a,u,l,g,m,v,S;if(t){if(i=(l=t.zh(),l?cf(l).si().oi(l):null),i){for(pE(e,t,i),u=t.zh(),v=0,S=(u.i==null&&Em(u),u.i).length;v<S;++v)m=(a=(u.i==null&&Em(u),u.i),v>=0&&v<a.length?a[v]:null),m.ok()&&!m.pk()&&(me(m,330)?J3n(e,f(m,35),t,i):(g=f(m,19),(g.Bb&Dc)!=0&&nOn(e,g,t,i)));t.Rh()&&f(i,52).ai(f(t,52).Xh())}return i}else return null}function _Ln(e,t,i){var a,u,l;if(!t.f)throw ie(new dn("Given leave edge is no tree edge."));if(i.f)throw ie(new dn("Given enter edge is a tree edge already."));for(t.f=!1,KSe(e.p,t),i.f=!0,Go(e.p,i),a=i.e.e-i.d.e-i.a,vde(e,i.e,t)||(a=-a),l=new ne(e.e.a);l.a<l.c.c.length;)u=f(ee(l),123),vde(e,u,t)||(u.e+=a);e.j=1,c7(e.c),UAe(e,f(ee(new ne(e.e.a)),123)),F_t(e)}function DSt(e,t){var i,a,u,l,g,m;if(m=f(V(t,(ot(),Xi)),100),m==(oa(),Lb)||m==uu)for(u=new Ke(t.f.a+t.d.b+t.d.c,t.f.b+t.d.d+t.d.a).b,g=new ne(e.a);g.a<g.c.c.length;)l=f(ee(g),9),l.k==(cn(),bo)&&(i=f(V(l,(Ge(),Tc)),64),!(i!=(tt(),_n)&&i!=pn)&&(a=ke(Oe(V(l,d_))),m==Lb&&(a*=u),l.n.b=a-f(V(l,m_),8).b,tZ(l,!1,!0)))}function MSt(e,t,i,a,u){i&&(!a||(e.c-e.b&e.a.length-1)>1)&&t==1&&f(e.a[e.b],9).k==(cn(),Vc)?TR(f(e.a[e.b],9),(df(),Rb)):a&&(!i||(e.c-e.b&e.a.length-1)>1)&&t==1&&f(e.a[e.c-1&e.a.length-1],9).k==(cn(),Vc)?TR(f(e.a[e.c-1&e.a.length-1],9),(df(),Nw)):(e.c-e.b&e.a.length-1)==2?(TR(f(vj(e),9),(df(),Rb)),TR(f(vj(e),9),Nw)):d6n(e,u),jTe(e)}function kLn(e){var t,i,a,u,l,g,m,v;for(v=new Wn,t=new fae,g=e.Jc();g.Ob();)u=f(g.Pb(),9),m=mS(Oz(new qy,u),t),ml(v.f,u,m);for(l=e.Jc();l.Ob();)for(u=f(l.Pb(),9),a=new Tn(Cn(ao(u).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),!Fa(i)&&Zh(Uh(Bh(Fh(zh(new ah,h.Math.max(1,f(V(i,(ot(),rDe)),17).a)),1),f(bn(v,i.c.i),123)),f(bn(v,i.d.i),123)));return t}function jSt(e,t,i,a){var u,l,g,m,v,S,T,A,L,F;if(akn(e,t,i),l=t[i],F=a?(tt(),pn):(tt(),_n),Nmn(t.length,i,a)){for(u=t[a?i-1:i+1],exe(e,u,a?(cs(),Uu):(cs(),td)),v=l,T=0,L=v.length;T<L;++T)g=v[T],x3e(e,g,F);for(exe(e,l,a?(cs(),td):(cs(),Uu)),m=u,S=0,A=m.length;S<A;++S)g=m[S],g.e||x3e(e,g,iG(F))}else for(m=l,S=0,A=m.length;S<A;++S)g=m[S],x3e(e,g,F);return!1}function TLn(e,t,i,a,u){var l,g,m,v,S,T,A;for(Ht(),na(e,new Cv),m=new la(e,0),A=new et,l=0;m.b<m.d.gc();)g=(Ln(m.b<m.d.gc()),f(m.d.Xb(m.c=m.b++),162)),A.c.length!=0&&Al(g)*jd(g)>l*2?(T=new LJ(A),S=Al(g)/jd(g),v=ife(T,t,new xN,i,a,u,S),ii(Cp(T.e),v),A.c.length=0,l=0,vn(A.c,T),vn(A.c,g),l=Al(T)*jd(T)+Al(g)*jd(g)):(vn(A.c,g),l+=Al(g)*jd(g));return A}function xLn(e,t){var i,a,u,l,g,m,v;for(t.Sg("Port order processing",1),v=f(V(e,(ot(),nDe)),418),a=new ne(e.b);a.a<a.c.c.length;)for(i=f(ee(a),25),l=new ne(i.a);l.a<l.c.c.length;)u=f(ee(l),9),g=f(V(u,Xi),100),m=u.j,g==(oa(),HE)||g==Lb||g==uu?(Ht(),na(m,oPe)):g!=Fg&&g!=Pw&&(Ht(),na(m,rNt),cAn(m),v==(MH(),Dme)&&na(m,nNt)),u.i=!0,$Ae(u);t.Tg()}function CLn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;if(Ld(e.e)){if(t!=i&&(u=f(e.g,122),F=u[i],g=F.Ik(),x1(e.e,g))){for(z=Ju(e.e.zh(),g),v=-1,m=-1,a=0,S=0,A=t>i?t:i;S<=A;++S)S==i?m=a++:(l=u[S],T=z.Zl(l.Ik()),S==t&&(v=S==A&&!T?a-1:a),T&&++a);return L=f(Cj(e,t,i),74),m!=v&&Q5(e,new NH(e.e,7,g,rt(m),F.kd(),v)),L}}else return f(yde(e,t,i),74);return f(Cj(e,t,i),74)}function $St(){$St=Z,i5t=Sr(new _o,(wa(),cu),(ka(),G5e)),GDe=Sr(new _o,su,Vee),a5t=Il(Sr(new _o,su,ete),zs,Qee),r5t=Il(Sr(Sr(new _o,su,F5e),cu,B5e),zs,U5e),s5t=vm(vm(b7(Il(Sr(new _o,rp,ite),zs,rte),cu),nte),ote),o5t=Il(new _o,zs,q5e),t5t=Il(Sr(Sr(Sr(new _o,Ab,Kee),cu,Xee),cu,nD),zs,Yee),n5t=Il(Sr(Sr(new _o,cu,nD),cu,qee),zs,Gee)}function ALn(e,t,i,a,u,l){var g,m,v,S,T,A,L;for(S=A1t(t)-A1t(e),g=yyt(t,S),v=Ku(0,0,0);S>=0&&(m=WAn(e,g),!(m&&(S<22?v.l|=1<<S:S<44?v.m|=1<<S-22:v.h|=1<<S-44,e.l==0&&e.m==0&&e.h==0)));)T=g.m,A=g.h,L=g.l,g.h=A>>>1,g.m=T>>>1|(A&1)<<21,g.l=L>>>1|(T&1)<<21,--S;return i&&Jue(v),l&&(a?(ww=VP(e),u&&(ww=s1t(ww,(MP(),v6e)))):ww=Ku(e.l,e.m,e.h)),v}function ILn(e,t){var i,a,u,l,g,m,v,S,T,A;for(S=e.e[t.c.p][t.p]+1,v=t.c.a.c.length+1,m=new ne(e.a);m.a<m.c.c.length;){for(g=f(ee(m),11),A=0,l=0,u=pb(Df(ce(se(Gf,1),Jt,20,0,[new Z5(g),new _N(g)])));Kn(u);)a=f(Dn(u),11),a.i.c==t.c&&(A+=_mn(e,a.i)+1,++l);i=A/l,T=g.j,T==(tt(),_n)?i<S?e.f[g.p]=e.c-i:e.f[g.p]=e.b+(v-i):T==pn&&(i<S?e.f[g.p]=e.b+i:e.f[g.p]=e.c-(v-i))}}function ff(e,t,i){var a,u,l,g,m;if(e==null)throw ie(new sm(bl));for(l=e.length,g=l>0&&(gn(0,e.length),e.charCodeAt(0)==45||(gn(0,e.length),e.charCodeAt(0)==43))?1:0,a=g;a<l;a++)if(dvt((gn(a,e.length),e.charCodeAt(a)))==-1)throw ie(new sm(XS+e+'"'));if(m=parseInt(e,10),u=m<t,isNaN(m))throw ie(new sm(XS+e+'"'));if(u||m>i)throw ie(new sm(XS+e+'"'));return m}function OLn(e){var t,i,a,u,l,g,m;for(g=new ki,l=new ne(e.a);l.a<l.c.c.length;)u=f(ee(l),116),yN(u,u.f.c.length),aS(u,u.k.c.length),u.i==0&&(u.o=0,Wi(g,u,g.c.b,g.c));for(;g.b!=0;)for(u=f(g.b==0?null:(Ln(g.b!=0),lh(g,g.a.a)),116),a=u.o+1,i=new ne(u.f);i.a<i.c.c.length;)t=f(ee(i),133),m=t.a,K5(m,h.Math.max(m.o,a)),aS(m,m.i-1),m.i==0&&Wi(g,m,g.c.b,g.c)}function NLn(e){var t,i,a,u,l,g,m,v;for(g=new ne(e);g.a<g.c.c.length;){for(l=f(ee(g),85),a=Us(f(be((!l.b&&(l.b=new Qt(Yn,l,4,7)),l.b),0),84)),m=a.i,v=a.j,u=f(be((!l.a&&(l.a=new Be(Ii,l,6,6)),l.a),0),168),Yz(u,u.j+m,u.k+v),Kz(u,u.b+m,u.c+v),i=new Nn((!u.a&&(u.a=new So(yf,u,5)),u.a));i.e!=i.i.gc();)t=f(Mn(i),369),Ese(t,t.a+m,t.b+v);Yxe(f(Fe(l,(Rr(),zA)),77),m,v)}}function yL(e){var t;switch(e){case 100:return aA(P$,!0);case 68:return aA(P$,!1);case 119:return aA(Ape,!0);case 87:return aA(Ape,!1);case 115:return aA(Ipe,!0);case 83:return aA(Ipe,!1);case 99:return aA(Ope,!0);case 67:return aA(Ope,!1);case 105:return aA(Npe,!0);case 73:return aA(Npe,!1);default:throw ie(new oc((t=e,mAt+t.toString(16))))}}function RLn(e){var t,i,a,u,l;switch(u=f(ct(e.a,0),9),t=new Tg(e),Qe(e.a,t),t.o.a=h.Math.max(1,u.o.a),t.o.b=h.Math.max(1,u.o.b),t.n.a=u.n.a,t.n.b=u.n.b,f(V(u,(Ge(),Tc)),64).g){case 4:t.n.a+=2;break;case 1:t.n.b+=2;break;case 2:t.n.a-=2;break;case 3:t.n.b-=2}return a=new ou,cc(a,t),i=new xS,l=f(ct(u.j,0),11),Ia(i,l),ua(i,a),ii(Cp(a.n),l.n),ii(Cp(a.a),l.a),t}function Bf(e,t,i,a){var u,l,g,m,v,S,T,A,L,F;for(u=t,m=i,a<0&&(u=i,m=t),l=f(bn(e.a,u),50),v=f(bn(e.a,m),50),g=f(bn(e.d,u),50),S=f(bn(e.d,m),50),l.a.yc(m,l),S.a.yc(u,S),F=v.a.ec().Jc();F.Ob();)L=f(F.Pb(),11),l.a.yc(L,l),Go(f(bn(e.d,L),50),u),Ba(f(bn(e.d,L),50),g);for(A=g.a.ec().Jc();A.Ob();)T=f(A.Pb(),11),S.a.yc(T,S),Go(f(bn(e.a,T),50),m),Ba(f(bn(e.a,T),50),v)}function PLn(e,t,i){var a,u,l,g,m;for(l=0,u=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));u.e!=u.i.gc();)a=f(Mn(u),27),g="",(!a.n&&(a.n=new Be(hc,a,1,7)),a.n).i==0||(g=f(be((!a.n&&(a.n=new Be(hc,a,1,7)),a.n),0),153).a),m=new Que(l++,t,g),Rc(m,a),$e(m,(_i(),mF),a),m.e.b=a.j+a.f/2,m.f.a=h.Math.max(a.g,1),m.e.a=a.i+a.g/2,m.f.b=h.Math.max(a.f,1),Mr(t.b,m),ml(i.f,a,m)}function VIe(e,t,i){var a,u,l,g;if(t[0]>=e.length)return i.o=0,!0;switch(_a(e,t[0])){case 43:u=1;break;case 45:u=-1;break;default:return i.o=0,!0}if(++t[0],l=t[0],g=WZ(e,t),g==0&&t[0]==l)return!1;if(t[0]<e.length&&_a(e,t[0])==58){if(a=g*60,++t[0],l=t[0],g=WZ(e,t),g==0&&t[0]==l)return!1;a+=g}else a=g,a<24&&t[0]-l<=2?a*=60:a=a%100+(a/100|0)*60;return a*=u,i.o=-a,!0}function LLn(e){var t,i,a,u,l,g,m,v,S;for(g=new et,a=new Tn(Cn(ao(e.b).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),Fa(i)&&Qe(g,new Upt(i,Igt(e,i.c),Igt(e,i.d)));for(S=(l=new kp(e.e).a.vc().Jc(),new zy(l));S.a.Ob();)m=(t=f(S.a.Pb(),43),f(t.kd(),112)),m.d.p=0;for(v=(u=new kp(e.e).a.vc().Jc(),new zy(u));v.a.Ob();)m=(t=f(v.a.Pb(),43),f(t.kd(),112)),m.d.p==0&&Qe(e.d,gLn(e,m))}function DLn(e,t){var i,a,u,l,g,m,v,S,T;for(g=new ne(t.b);g.a<g.c.c.length;)for(l=f(ee(g),25),S=new ne(l.a);S.a<S.c.c.length;){for(v=f(ee(S),9),T=new et,m=0,a=new Tn(Cn(Fo(v).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),!(Fa(i)||!Fa(i)&&i.c.i.c==i.d.i.c)&&(u=f(V(i,(ot(),e4)),17).a,u>m&&(m=u,T.c.length=0),u==m&&Qe(T,new Ya(i.c.i,i)));Ht(),na(T,e.c),Ky(e.b,v.p,T)}}function MLn(e,t){var i,a,u,l,g,m,v,S,T;for(g=new ne(t.b);g.a<g.c.c.length;)for(l=f(ee(g),25),S=new ne(l.a);S.a<S.c.c.length;){for(v=f(ee(S),9),T=new et,m=0,a=new Tn(Cn(ao(v).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),!(Fa(i)||!Fa(i)&&i.c.i.c==i.d.i.c)&&(u=f(V(i,(ot(),e4)),17).a,u>m&&(m=u,T.c.length=0),u==m&&Qe(T,new Ya(i.d.i,i)));Ht(),na(T,e.c),Ky(e.f,v.p,T)}}function jLn(e){var t,i,a,u,l,g,m;for(l=bg(e),u=new Nn((!e.e&&(e.e=new Qt(vo,e,7,4)),e.e));u.e!=u.i.gc();)if(a=f(Mn(u),85),m=Us(f(be((!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c),0),84)),!bT(m,l))return!0;for(i=new Nn((!e.d&&(e.d=new Qt(vo,e,8,5)),e.d));i.e!=i.i.gc();)if(t=f(Mn(i),85),g=Us(f(be((!t.b&&(t.b=new Qt(Yn,t,4,7)),t.b),0),84)),!bT(g,l))return!0;return!1}function $Ln(e){var t,i,a,u,l;a=f(V(e,(Ge(),oi)),27),l=f(Fe(a,(ot(),DE)),181).Gc((Ud(),GE)),e.e||(u=f(V(e,Cu),22),t=new Ke(e.f.a+e.d.b+e.d.c,e.f.b+e.d.d+e.d.a),u.Gc((us(),Vf))?(pi(a,Xi,(oa(),uu)),WS(a,t.a,t.b,!1,!0)):lt(ut(Fe(a,vme)))||WS(a,t.a,t.b,!0,!0)),l?pi(a,DE,At(GE)):pi(a,DE,(i=f(Tp($F),10),new Rf(i,f(Gh(i,i.length),10),0)))}function FLn(e,t){var i,a,u,l,g,m,v,S;if(S=ut(V(t,(Ec(),z5t))),S==null||(an(S),S)){for(_4n(e,t),u=new et,v=tr(t.b,0);v.b!=v.d.c;)g=f(Zn(v),39),i=OAe(e,g,null),i&&(Rc(i,t),vn(u.c,i));if(e.a=null,e.b=null,u.c.length>1)for(a=new ne(u);a.a<a.c.c.length;)for(i=f(ee(a),120),l=0,m=tr(i.b,0);m.b!=m.d.c;)g=f(Zn(m),39),g.g=l++;return u}return gm(ce(se(ljn,1),FTt,120,0,[t]))}function BLn(e,t){var i,a,u,l,g,m;for(u=new ne(t.a);u.a<u.c.c.length;)a=f(ee(u),9),l=V(a,(Ge(),oi)),me(l,11)&&(g=f(l,11),m=i_t(t,a,g.o.a,g.o.b),g.n.a=m.a,g.n.b=m.b,No(g,f(V(a,Tc),64)));i=new Ke(t.f.a+t.d.b+t.d.c,t.f.b+t.d.d+t.d.a),f(V(t,(Ge(),Cu)),22).Gc((us(),Vf))?($e(e,(ot(),Xi),(oa(),uu)),f(V(ra(e),Cu),22).Ec(hD),G_t(e,i,!1)):G_t(e,i,!0)}function ULn(e){var t,i,a,u,l,g,m,v;for(v=new fd,t=tr(e,0),m=null,i=f(Zn(t),8),u=f(Zn(t),8);t.b!=t.d.c;)m=i,i=u,u=f(Zn(t),8),l=Fmt(Ho(new Ke(m.a,m.b),i)),g=Fmt(Ho(new Ke(u.a,u.b),i)),a=10,a=h.Math.min(a,h.Math.abs(l.a+l.b)/2),a=h.Math.min(a,h.Math.abs(g.a+g.b)/2),l.a=dH(l.a)*a,l.b=dH(l.b)*a,g.a=dH(g.a)*a,g.b=dH(g.b)*a,Mr(v,ii(l,i)),Mr(v,ii(g,i));return v}function zLn(e,t,i){var a,u,l,g,m,v;if(i.Sg("Minimize Crossings "+e.a,1),a=t.b.c.length==0||!oP(Jr(new Ft(null,new Bt(t.b,16)),new Y(new jU))).zd((Wy(),zR)),v=t.b.c.length==1&&f(ct(t.b,0),25).a.c.length==1,l=Ae(V(t,(ot(),rx)))===Ae((_1(),W0)),a||v&&!l){i.Tg();return}u=D5n(e,t),g=(m=f(Xh(u,0),217),m.c.ig()?m.c.cg()?new itt(e):new ott(e):new rtt(e)),rxn(u,g),Mxn(e),i.Tg()}function hf(e,t,i,a){var u,l,g,m,v;return g=e.Lh(),v=e.Fh(),u=null,v?t&&(Dde(e,t,i).Bb&Xa)==0?(a=za(v.Bl(),e,a),e._h(null),u=t.Mh()):v=null:(g&&(v=g.Mh()),t&&(u=t.Mh())),v!=u&&v&&v.Fl(e),m=e.Bh(),e.xh(t,i),v!=u&&u&&u.El(e),e.rh()&&e.sh()&&(g&&m>=0&&m!=i&&(l=new Jo(e,1,m,g,null),a?a.kj(l):a=l),i>=0&&(l=new Jo(e,1,i,m==i?g:null,t),a?a.kj(l):a=l)),a}function FSt(e){var t,i,a;if(e.b==null){if(a=new p0,e.i!=null&&(ys(a,e.i),a.a+=":"),(e.f&256)!=0){for((e.f&256)!=0&&e.a!=null&&(iyn(e.i)||(a.a+="//"),ys(a,e.a)),e.d!=null&&(a.a+="/",ys(a,e.d)),(e.f&16)!=0&&(a.a+="/"),t=0,i=e.j.length;t<i;t++)t!=0&&(a.a+="/"),ys(a,e.j[t]);e.g!=null&&(a.a+="?",ys(a,e.g))}else ys(a,e.a);e.e!=null&&(a.a+="#",ys(a,e.e)),e.b=a.a}return e.b}function HLn(e,t,i,a,u){var l,g,m,v;l=new Tg(e),Wl(l,(cn(),xu)),$e(l,(ot(),Xi),(oa(),uu)),$e(l,(Ge(),oi),t.c.i),g=new ou,$e(g,oi,t.c),No(g,u),cc(g,l),$e(t.c,ed,l),m=new Tg(e),Wl(m,xu),$e(m,Xi,uu),$e(m,oi,t.d.i),v=new ou,$e(v,oi,t.d),No(v,u),cc(v,m),$e(t.d,ed,m),Ia(t,g),ua(t,v),fT(0,i.c.length),y7(i.c,0,l),vn(a.c,m),$e(l,Ete,rt(1)),$e(m,Ete,rt(1))}function GLn(e,t,i,a){var u,l,g,m,v;if(v=mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))),u=mr(Na(Sb,gb(mr(Na(i==null?0:Ti(i),_b)),15))),m=gj(e,t,v),g=pj(e,i,u),m&&u==m.a&&dm(i,m.g))return i;if(g)throw ie(new dn("key already present: "+i));return m&&tA(e,m),g&&tA(e,g),l=new qX(i,u,t,v),yG(e,l,g),g&&(g.e=null,g.c=null),m&&(m.e=null,m.c=null),Awt(e),m?m.g:null}function BSt(e,t,i){var a,u,l,g,m;for(l=0;l<t;l++){for(a=0,m=l+1;m<t;m++)a=Ua(Ua(Na(ia(e[l],ls),ia(e[m],ls)),ia(i[l+m],ls)),ia(mr(a),ls)),i[l+m]=mr(a),a=Jy(a,32);i[l+t]=mr(a)}for(Kkn(i,i,t<<1),a=0,u=0,g=0;u<t;++u,g++)a=Ua(Ua(Na(ia(e[u],ls),ia(e[u],ls)),ia(i[g],ls)),ia(mr(a),ls)),i[g]=mr(a),a=Jy(a,32),++g,a=Ua(a,ia(i[g],ls)),i[g]=mr(a),a=Jy(a,32);return i}function USt(e,t,i){var a,u,l,g,m,v,S,T;if(!NP(t)){for(v=ke(Oe(kT(i.c,(ot(),rF)))),S=f(kT(i.c,$q),139),!S&&(S=new i7),a=i.a,u=null,m=t.Jc();m.Ob();)g=f(m.Pb(),11),T=0,u?(T=v,T+=u.o.b):T=S.d,l=mS(Oz(new qy,g),e.f),qr(e.k,g,l),Zh(Uh(Bh(Fh(zh(new ah,0),Aa(h.Math.ceil(T))),a),l)),u=g,a=l;Zh(Uh(Bh(Fh(zh(new ah,0),Aa(h.Math.ceil(S.a+u.o.b))),a),i.d))}}function qLn(e,t,i,a,u,l,g,m){var v,S,T,A,L,F;return F=!1,L=l-i.s,T=i.t-t.f+(S=Zj(i,L,!1),S.a),a.g+m>L?!1:(A=(v=Zj(a,L,!1),v.a),T+m+A<=t.b&&(IH(i,l-i.s),i.c=!0,IH(a,l-i.s),cG(a,i.s,i.t+i.d+m),a.k=!0,Qxe(i.q,a),F=!0,u&&(jJ(t,a),a.j=t,e.c.length>g&&(fG((Gt(g,e.c.length),f(e.c[g],186)),a),(Gt(g,e.c.length),f(e.c[g],186)).a.c.length==0&&_0(e,g)))),F)}function VLn(e,t){var i,a,u,l,g,m;if(t.Sg("Partition midprocessing",1),u=new CS,Zi(Jr(new Ft(null,new Bt(e.a,16)),new lO),new Eet(u)),u.d!=0){for(m=f($d(vTe((l=u.i,new Ft(null,(l||(u.i=new N3(u,u.c))).Lc()))),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),a=m.Jc(),i=f(a.Pb(),17);a.Ob();)g=f(a.Pb(),17),G6n(f(si(u,i),22),f(si(u,g),22)),i=g;t.Tg()}}function Kj(e,t){var i,a,u,l,g;if(e.Ab){if(e.Ab){if(g=e.Ab.i,g>0){if(u=f(e.Ab.g,1979),t==null){for(l=0;l<g;++l)if(i=u[l],i.d==null)return i}else for(l=0;l<g;++l)if(i=u[l],Mt(t,i.d))return i}}else if(t==null){for(a=new Nn(e.Ab);a.e!=a.i.gc();)if(i=f(Mn(a),582),i.d==null)return i}else for(a=new Nn(e.Ab);a.e!=a.i.gc();)if(i=f(Mn(a),582),Mt(t,i.d))return i}return null}function zSt(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;F=eSn(e,oCe(t),u),Uxe(F,p1(u,Am)),b=null,z=u,W=ij(z,UCt),J=new rnt(F),vyt(J.a,W),ue=ij(z,"endPoint"),pe=new snt(F),byt(pe.a,ue),ve=IS(z,see),Ee=new lnt(F),Gwt(Ee.a,ve),A=p1(u,$4e),l=new Dat(e,F),lut(l.a,l.b,A),L=p1(u,j4e),g=new Mat(e,F),dut(g.a,g.b,L),S=IS(u,B4e),m=new jat(i,F),pwt(m.b,m.a,S),T=IS(u,F4e),v=new $at(a,F),gwt(v.b,v.a,T)}function HSt(e,t,i){var a,u,l,g,m,v,S,T;if(t.p==0){for(t.p=1,g=i,g||(u=new et,l=(a=f(Tp(Ja),10),new Rf(a,f(Gh(a,a.length),10),0)),g=new Ya(u,l)),f(g.a,15).Ec(t),t.k==(cn(),bo)&&f(g.b,22).Ec(f(V(t,(Ge(),Tc)),64)),v=new ne(t.j);v.a<v.c.c.length;)for(m=f(ee(v),11),T=pb(Df(ce(se(Gf,1),Jt,20,0,[new Z5(m),new _N(m)])));Kn(T);)S=f(Dn(T),11),HSt(e,S.i,g);return g}return null}function WIe(e,t,i){var a,u,l,g,m;switch(m=null,t.g){case 1:for(u=new ne(e.j);u.a<u.c.c.length;)if(a=f(ee(u),11),lt(ut(V(a,(Ge(),Qge)))))return a;m=new ou,$e(m,(Ge(),Qge),(mn(),!0));break;case 2:for(g=new ne(e.j);g.a<g.c.c.length;)if(l=f(ee(g),11),lt(ut(V(l,(Ge(),tme)))))return l;m=new ou,$e(m,(Ge(),tme),(mn(),!0))}return m&&(cc(m,e),No(m,i),rAn(m.n,e.o,i)),m}function GSt(e,t){var i,a,u,l,g,m;for(m=-1,g=new ki,a=new yg(e.b);sc(a.a)||sc(a.b);){for(i=f(sc(a.a)?ee(a.a):ee(a.b),18),m=h.Math.max(m,ke(Oe(V(i,(ot(),z0))))),i.c==e?Zi(Jr(new Ft(null,new Bt(i.b,16)),new g6),new uet(g)):Zi(Jr(new Ft(null,new Bt(i.b,16)),new Z_),new det(g)),l=tr(g,0);l.b!=l.d.c;)u=f(Zn(l),72),li(u,(Ge(),YR))||$e(u,YR,i);$o(t,g),nf(g)}return m}function qS(e,t,i,a,u){var l,g,m,v,S;m=u?a.b:a.a,!xp(e.a,a)&&(S=m>i.s&&m<i.c,v=!1,i.e.b!=0&&i.j.b!=0&&(v=v|(h.Math.abs(m-ke(Oe(hP(i.e))))<Tm&&h.Math.abs(m-ke(Oe(hP(i.j))))<Tm),v=v|(h.Math.abs(m-ke(Oe(Hh(i.e))))<Tm&&h.Math.abs(m-ke(Oe(Hh(i.j))))<Tm)),(S||v)&&(g=f(V(t,(ot(),Ls)),77),g||(g=new fd,$e(t,Ls,g)),l=new ja(a),Wi(g,l,g.c.b,g.c),Go(e.a,l)))}function WLn(e,t,i){var a,u,l,g,m,v,S,T,A;for(i.Sg("Processor set coordinates",1),e.a=t.b.b==0?1:t.b.b,S=null,a=tr(t.b,0);!S&&a.b!=a.d.c;)A=f(Zn(a),39),lt(ut(V(A,(_i(),Iw))))&&(S=A,v=A.e,v.a=f(V(A,Vq),17).a,v.b=f(V(A,Wq),17).a);m=Wue(S),T=1;do m=SNn((u=m,i.bh(T),u)),T=m.b/e.a|0;while(m.b!=0);for(g=tr(t.b,0);g.b!=g.d.c;)l=f(Zn(g),39),Ho(l.e,new Ke(l.f.a/2,l.f.b/2));i.Tg()}function KLn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(i=!1,T=ke(Oe(V(t,(ot(),v_)))),z=fw*T,u=new ne(t.b);u.a<u.c.c.length;)for(a=f(ee(u),25),S=new ne(a.a),l=f(ee(S),9),A=Nke(e.a[l.p]);S.a<S.c.c.length;)m=f(ee(S),9),L=Nke(e.a[m.p]),A!=L&&(F=A3(e.b,l,m),g=l.n.b+l.o.b+l.d.a+A.a+F,v=m.n.b-m.d.d+L.a,g>v+z&&(W=A.g+L.g,L.a=(L.g*L.a+A.g*A.a)/W,L.g=W,A.f=L,i=!0)),l=m,A=L;return i}function YLn(e,t,i){var a,u,l,g,m,v,S,T;for(i.Sg(Nxt,1),nu(e.b),nu(e.a),m=null,l=tr(t.b,0);!m&&l.b!=l.d.c;)S=f(Zn(l),39),lt(ut(V(S,(_i(),Iw))))&&(m=S);for(v=new ki,Wi(v,m,v.c.b,v.c),bkt(e,v),T=tr(t.b,0);T.b!=T.d.c;)S=f(Zn(T),39),g=dr(V(S,(_i(),gF))),u=wu(e.b,g)!=null?f(wu(e.b,g),17).a:0,$e(S,Wme,rt(u)),a=1+(wu(e.a,g)!=null?f(wu(e.a,g),17).a:0),$e(S,oMe,rt(a));i.Tg()}function qSt(e){Wk(e,new zS(qk(zk(Gk(Hk(new ug,t_),"ELK Box"),"Algorithm for packing of unconnected boxes, i.e. graphs without edges."),new u5))),Je(e,t_,zT,r9e),Je(e,t_,UT,15),Je(e,t_,XG,rt(0)),Je(e,t_,_4e,st(e9e)),Je(e,t_,uA,st(MLt)),Je(e,t_,LR,st(jLt)),Je(e,t_,ML,tCt),Je(e,t_,l$,st(t9e)),Je(e,t_,DR,st(n9e)),Je(e,t_,k4e,st(Ube)),Je(e,t_,KQ,st(DLt))}function VSt(e,t){var i,a,u,l,g,m,v,S,T;if(u=e.i,g=u.o.a,l=u.o.b,g<=0&&l<=0)return tt(),fc;switch(S=e.n.a,T=e.n.b,m=e.o.a,i=e.o.b,t.g){case 2:case 1:if(S<0)return tt(),pn;if(S+m>g)return tt(),_n;break;case 4:case 3:if(T<0)return tt(),hn;if(T+i>l)return tt(),Bn}return v=(S+m/2)/g,a=(T+i/2)/l,v+a<=1&&v-a<=0?(tt(),pn):v+a>=1&&v-a>=0?(tt(),_n):a<.5?(tt(),hn):(tt(),Bn)}function WSt(e,t,i,a,u,l,g){var m,v,S,T,A,L;for(L=new BN,S=t.Jc();S.Ob();)for(m=f(S.Pb(),832),A=new ne(m.Pf());A.a<A.c.c.length;)T=f(ee(A),187),Ae(T.mf((Rr(),Vbe)))===Ae((Sg(),LD))&&(TSt(L,T,!1,a,u,l,g),jj(e,L));for(v=i.Jc();v.Ob();)for(m=f(v.Pb(),832),A=new ne(m.Pf());A.a<A.c.c.length;)T=f(ee(A),187),Ae(T.mf((Rr(),Vbe)))===Ae((Sg(),gx))&&(TSt(L,T,!0,a,u,l,g),jj(e,L))}function XLn(e,t,i){var a,u,l,g,m,v,S;for(g=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));g.e!=g.i.gc();)for(l=f(Mn(g),27),u=new Tn(Cn(lw(l).a.Jc(),new B));Kn(u);)a=f(Dn(u),85),!Uj(a)&&!Uj(a)&&!HS(a)&&(v=f(ac(Is(i.f,l)),39),S=f(bn(i,Us(f(be((!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c),0),84))),39),v&&S&&(m=new oTe(v,S),$e(m,(_i(),mF),a),Rc(m,a),Mr(v.d,m),Mr(S.b,m),Mr(t.a,m)))}function JLn(e,t){var i,a,u,l,g,m,v,S;for(v=f(f(si(e.r,t),22),83).Jc();v.Ob();)m=f(v.Pb(),115),u=m.c?rlt(m.c):0,u>0?m.a?(S=m.b.Kf().b,u>S&&(e.v||m.c.d.c.length==1?(g=(u-S)/2,m.d.d=g,m.d.a=g):(i=f(ct(m.c.d,0),187).Kf().b,a=(i-S)/2,m.d.d=h.Math.max(0,a),m.d.a=u-a-S))):m.d.a=e.t+u:K7(e.u)&&(l=fAe(m.b),l.d<0&&(m.d.d=-l.d),l.d+l.a>m.b.Kf().b&&(m.d.a=l.d+l.a-m.b.Kf().b))}function Qh(){Qh=Z,GR=new pa((Rr(),aV),rt(1)),Fee=new pa(zE,80),uOt=new pa(P9e,5),QIt=new pa(ID,DL),sOt=new pa(t1e,rt(1)),cOt=new pa(n1e,(mn(),!0)),e5e=new gS(50),oOt=new pa(V0,e5e),J6e=iV,t5e=OF,eOt=new pa(qbe,!1),Q6e=oV,rOt=hx,iOt=Ow,nOt=UE,tOt=HA,aOt=px,Z6e=(SAe(),VIt),cge=XIt,$ee=qIt,sge=WIt,n5e=YIt,fOt=RD,hOt=Sne,dOt=ND,lOt=sV,r5e=(gR(),bx),new pa(s4,r5e)}function ZLn(e,t){var i;switch($H(e)){case 6:return ta(t);case 7:return Jk(t);case 8:return Xk(t);case 3:return Array.isArray(t)&&(i=$H(t),!(i>=14&&i<=16));case 11:return t!=null&&typeof t===cfe;case 12:return t!=null&&(typeof t===DG||typeof t==cfe);case 0:return Lle(t,e.__elementTypeId$);case 2:return gce(t)&&t.Qm!==xe;case 1:return gce(t)&&t.Qm!==xe||Lle(t,e.__elementTypeId$);default:return!0}}function QLn(e){var t,i,a,u;a=e.o,tT(),e.A.dc()||ei(e.A,V6e)?u=a.a:(e.D?u=h.Math.max(a.a,Fj(e.f)):u=Fj(e.f),e.A.Gc((Ud(),gV))&&!e.B.Gc((yd(),FF))&&(u=h.Math.max(u,Fj(f(Es(e.p,(tt(),hn)),251))),u=h.Math.max(u,Fj(f(Es(e.p,Bn),251)))),t=qbt(e),t&&(u=h.Math.max(u,t.a))),lt(ut(e.e.Rf().mf((Rr(),hx))))?a.a=h.Math.max(a.a,u):a.a=u,i=e.f.i,i.c=0,i.b=u,Ude(e.f)}function KSt(e,t){var i,a,u,l;return a=h.Math.min(h.Math.abs(e.c-(t.c+t.b)),h.Math.abs(e.c+e.b-t.c)),l=h.Math.min(h.Math.abs(e.d-(t.d+t.a)),h.Math.abs(e.d+e.a-t.d)),i=h.Math.abs(e.c+e.b/2-(t.c+t.b/2)),i>e.b/2+t.b/2||(u=h.Math.abs(e.d+e.a/2-(t.d+t.a/2)),u>e.a/2+t.a/2)?1:i==0&&u==0?0:i==0?l/u+1:u==0?a/i+1:h.Math.min(a/i,l/u)+1}function eDn(e,t){var i,a,u,l,g,m,v;for(l=0,m=0,v=0,u=new ne(e.f.e);u.a<u.c.c.length;)a=f(ee(u),151),t!=a&&(g=e.i[t.a][a.a],l+=g,i=Qv(t.d,a.d),i>0&&e.d!=(cj(),dge)&&(m+=g*(a.d.a+e.a[t.a][a.a]*(t.d.a-a.d.a)/i)),i>0&&e.d!=(cj(),uge)&&(v+=g*(a.d.b+e.a[t.a][a.a]*(t.d.b-a.d.b)/i)));switch(e.d.g){case 1:return new Ke(m/l,t.d.b);case 2:return new Ke(t.d.a,v/l);default:return new Ke(m/l,v/l)}}function YSt(e){var t,i,a,u,l,g;for(i=(!e.a&&(e.a=new So(yf,e,5)),e.a).i+2,g=new Du(i),Qe(g,new Ke(e.j,e.k)),Zi(new Ft(null,(!e.a&&(e.a=new So(yf,e,5)),new Bt(e.a,16))),new Qtt(g)),Qe(g,new Ke(e.b,e.c)),t=1;t<g.c.length-1;)a=(Gt(t-1,g.c.length),f(g.c[t-1],8)),u=(Gt(t,g.c.length),f(g.c[t],8)),l=(Gt(t+1,g.c.length),f(g.c[t+1],8)),a.a==u.a&&u.a==l.a||a.b==u.b&&u.b==l.b?_0(g,t):++t;return g}function XSt(e,t){WP();var i,a,u,l,g;if(g=f(V(e.i,(ot(),Xi)),100),l=e.j.g-t.j.g,l!=0||!(g==(oa(),HE)||g==Lb||g==uu))return 0;if(g==(oa(),HE)&&(i=f(V(e,H0),17),a=f(V(t,H0),17),i&&a&&(u=i.a-a.a,u!=0)))return u;switch(e.j.g){case 1:return di(e.n.a,t.n.a);case 2:return di(e.n.b,t.n.b);case 3:return di(t.n.a,e.n.a);case 4:return di(t.n.b,e.n.b);default:throw ie(new As(sNe))}}function JSt(e,t){var i,a,u,l,g,m,v;for(i=mct(_ot(Eot(Sot(new f2e,t),new ZX(t.e)),aNt),e.a),t.j.c.length==0||Hgt(f(ct(t.j,0),60).a,i),v=new uae,qr(e.e,i,v),g=new Mo,m=new Mo,l=new ne(t.k);l.a<l.c.c.length;)u=f(ee(l),18),Go(g,u.c),Go(m,u.d);a=g.a.gc()-m.a.gc(),a<0?(YH(v,!0,(ko(),Ds)),YH(v,!1,Gs)):a>0&&(YH(v,!1,(ko(),Ds)),YH(v,!0,Gs)),Mu(t.g,new Zot(e,i)),qr(e.g,t,i)}function KIe(){KIe=Z,$Pt=new Pt(n4e,(mn(),!1)),rt(-1),NPt=new Pt(r4e,rt(-1)),rt(-1),RPt=new Pt(i4e,rt(-1)),PPt=new Pt(o4e,!1),LPt=new Pt(a4e,!1),r8e=(gJ(),xbe),zPt=new Pt(s4e,r8e),HPt=new Pt(c4e,-1),n8e=(gZ(),Sbe),UPt=new Pt(u4e,n8e),BPt=new Pt(l4e,!0),e8e=(TJ(),Cbe),jPt=new Pt(d4e,e8e),MPt=new Pt(f4e,!1),rt(1),DPt=new Pt(h4e,rt(1)),t8e=(lZ(),Abe),FPt=new Pt(p4e,t8e)}function ZSt(){ZSt=Z;var e;for(x6e=ce(se(pr,1),Pr,30,15,[-1,-1,30,19,15,13,11,11,10,9,9,8,8,8,8,7,7,7,7,7,7,7,6,6,6,6,6,6,6,6,6,6,6,6,6,6,5]),qpe=Pe(pr,Pr,30,37,15,1),BAt=ce(se(pr,1),Pr,30,15,[-1,-1,63,40,32,28,25,23,21,20,19,19,18,18,17,17,16,16,16,15,15,15,15,14,14,14,14,14,14,13,13,13,13,13,13,13,13]),C6e=Pe(x_,dTt,30,37,14,1),e=2;e<=36;e++)qpe[e]=Aa(h.Math.pow(e,x6e[e])),C6e[e]=pG(FG,qpe[e])}function tDn(e){var t;if((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i!=1)throw ie(new dn(TCt+(!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i));return t=new fd,Vue(f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84))&&Ba(t,Mkt(e,Vue(f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84)),!1)),Vue(f(be((!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c),0),84))&&Ba(t,Mkt(e,Vue(f(be((!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c),0),84)),!0)),t}function QSt(e,t){var i,a,u,l,g;for(t.d?u=e.a.c==(fm(),y_)?Fo(t.b):ao(t.b):u=e.a.c==(fm(),G0)?Fo(t.b):ao(t.b),l=!1,a=new Tn(Cn(u.a.Jc(),new B));Kn(a);)if(i=f(Dn(a),18),g=lt(e.a.f[e.a.g[t.b.p].p]),!(!g&&!Fa(i)&&i.c.i.c==i.d.i.c)&&!(lt(e.a.n[e.a.g[t.b.p].p])||lt(e.a.n[e.a.g[t.b.p].p]))&&(l=!0,xp(e.b,e.a.g[OAn(i,t.b).p])))return t.c=!0,t.a=i,t;return t.c=l,t.a=null,t}function YIe(e,t,i){var a,u,l,g,m,v,S;if(a=i.gc(),a==0)return!1;if(e.Mj())if(v=e.Nj(),n3e(e,t,i),g=a==1?e.Fj(3,null,i.Jc().Pb(),t,v):e.Fj(5,null,i,t,v),e.Jj()){for(m=a<100?null:new Fv(a),l=t+a,u=t;u<l;++u)S=e.uj(u),m=e.Kj(S,m),m=m;m?(m.kj(g),m.lj()):e.Gj(g)}else e.Gj(g);else if(n3e(e,t,i),e.Jj()){for(m=a<100?null:new Fv(a),l=t+a,u=t;u<l;++u)m=e.Kj(e.uj(u),m);m&&m.lj()}return!0}function e_t(e,t,i){var a,u,l,g,m;return e.Mj()?(u=null,l=e.Nj(),a=e.Fj(1,m=(g=e.Aj(t,e.Wi(t,i)),g),i,t,l),e.Jj()&&!(e.Vi()&&m?ei(m,i):Ae(m)===Ae(i))&&(m&&(u=e.Lj(m,u)),u=e.Kj(i,u)),u?(u.kj(a),u.lj()):e.Gj(a),m):(m=(g=e.Aj(t,e.Wi(t,i)),g),e.Jj()&&!(e.Vi()&&m?ei(m,i):Ae(m)===Ae(i))&&(u=null,m&&(u=e.Lj(m,null)),u=e.Kj(i,u),u&&u.lj()),m)}function XIe(e,t){var i,a,u,l,g,m,v,S,T;if(e.e=t,e.f=f(V(t,(ew(),Bee)),233),ZOn(t),e.d=h.Math.max(t.e.c.length*16+t.c.c.length,256),!lt(ut(V(t,(Qh(),J6e)))))for(T=e.e.e.c.length,v=new ne(t.e);v.a<v.c.c.length;)m=f(ee(v),151),S=m.d,S.a=yke(e.f)*T,S.b=yke(e.f)*T;for(i=t.b,l=new ne(t.c);l.a<l.c.c.length;)if(u=f(ee(l),288),a=f(V(u,n5e),17).a,a>0){for(g=0;g<a;g++)Qe(i,new rdt(u));NEt(u)}}function nDn(e,t){var i,a,u,l,g,m,v,S,T,A;for(t.Sg("Hypernodes processing",1),u=new ne(e.b);u.a<u.c.c.length;)for(a=f(ee(u),25),m=new ne(a.a);m.a<m.c.c.length;)if(g=f(ee(m),9),lt(ut(V(g,(ot(),Pte))))&&g.j.c.length<=2){for(A=0,T=0,i=0,l=0,S=new ne(g.j);S.a<S.c.c.length;)switch(v=f(ee(S),11),v.j.g){case 1:++A;break;case 2:++T;break;case 3:++i;break;case 4:++l}A==0&&i==0&&t7n(e,g,l<=T)}t.Tg()}function t_t(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z;if(L=new cT(e.Xg()),Wh(t,Gr,L),i&&!e.Vg().a.dc())for(T=new oh,Wh(t,"logs",T),m=0,z=new p3(e.Vg().b.Jc());z.b.Ob();)F=dr(z.b.Pb()),A=new cT(F),gT(T,m),Qce(T,m,A),++m;if(a&&(S=new oS(e.Ug()),Wh(t,"executionTime",S)),!e.Wg().a.dc())for(g=new oh,Wh(t,dpe,g),m=0,l=new p3(e.Wg().b.Jc());l.b.Ob();)u=f(l.b.Pb(),847),v=new AN,gT(g,m),Qce(g,m,v),t_t(u,v,i,a),++m}function JIe(){JIe=Z,Mae(),KMt=new s3,ce(se(l4,2),Ye,373,0,[ce(se(l4,1),Tee,584,0,[new Tz(aAt)])]),ce(se(l4,2),Ye,373,0,[ce(se(l4,1),Tee,584,0,[new Tz(a6e)])]),ce(se(l4,2),Ye,373,0,[ce(se(l4,1),Tee,584,0,[new Tz(sAt)]),ce(se(l4,1),Tee,584,0,[new Tz(a6e)])]),new Gv("-1"),ce(se(l4,2),Ye,373,0,[ce(se(l4,1),Tee,584,0,[new Tz("\\c+")])]),new Gv("0"),new Gv("0"),new Gv("1"),new Gv("0"),new Gv(hAt)}function rDn(e,t,i){var a,u,l,g,m,v,S,T,A;for(i.Sg("Hyperedge merging",1),B4n(e,t),v=new la(t.b,0);v.b<v.d.gc();)if(m=(Ln(v.b<v.d.gc()),f(v.d.Xb(v.c=v.b++),25)),T=m.a,T.c.length!=0)for(a=null,u=null,l=null,g=null,S=0;S<T.c.length;S++)a=(Gt(S,T.c.length),f(T.c[S],9)),u=a.k,u==(cn(),uo)&&g==uo&&(A=YDn(a,l),A.a&&(iPn(a,l,A.b,A.c),Gt(S,T.c.length),W2e(T.c,S,1),--S,a=l,u=g)),l=a,g=u;i.Tg()}function iDn(e,t,i,a,u){var l,g,m,v,S,T,A;for(g=new ne(t);g.a<g.c.c.length;){if(l=f(ee(g),18),v=l.c,i.a._b(v))S=(AS(),Aw);else if(a.a._b(v))S=(AS(),LA);else throw ie(new dn("Source port must be in one of the port sets."));if(T=l.d,i.a._b(T))A=(AS(),Aw);else if(a.a._b(T))A=(AS(),LA);else throw ie(new dn("Target port must be in one of the port sets."));m=new Vyt(l,S,A),qr(e.b,l,m),vn(u.c,m)}}function uQ(e){var t,i;return e.c&&e.c.Rh()&&(i=f(e.c,52),e.c=f(iw(e,i),141),e.c!=i&&((e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,9,2,i,e.c)),me(e.Cb,400)?e.Db>>16==-15&&e.Cb.Uh()&&yue(new gue(e.Cb,9,13,i,e.c,P0(gd(f(e.Cb,62)),e))):me(e.Cb,88)&&e.Db>>16==-23&&e.Cb.Uh()&&(t=e.c,me(t,88)||(t=(Ut(),Ch)),me(i,88)||(i=(Ut(),Ch)),yue(new gue(e.Cb,9,10,i,t,P0(iu(f(e.Cb,29)),e)))))),e.c}function n_t(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;if(t==i)return!0;if(t=eIe(e,t),i=eIe(e,i),a=Ule(t),a){if(T=Ule(i),T!=a)return T?(v=a.jk(),z=T.jk(),v==z&&v!=null):!1;if(g=(!t.d&&(t.d=new So(gs,t,1)),t.d),l=g.i,L=(!i.d&&(i.d=new So(gs,i,1)),i.d),l==L.i){for(S=0;S<l;++S)if(u=f(be(g,S),87),A=f(be(L,S),87),!n_t(e,u,A))return!1}return!0}else return m=t.e,F=i.e,m==F}function r_t(e,t,i,a){var u,l,g,m,v,S,T,A;if(x1(e.e,t)){for(A=Ju(e.e.zh(),t),l=f(e.g,122),T=null,v=-1,m=-1,u=0,S=0;S<e.i;++S)g=l[S],A.Zl(g.Ik())&&(u==i&&(v=S),u==a&&(m=S,T=g.kd()),++u);if(v==-1)throw ie(new Ru(ppe+i+TE+u));if(m==-1)throw ie(new Ru(gpe+a+TE+u));return Cj(e,v,m),Ld(e.e)&&Q5(e,Yv(e,7,t,rt(a),T,i,!0)),T}else throw ie(new dn("The feature must be many-valued to support move"))}function i_t(e,t,i,a){var u,l,g,m,v;switch(v=new ja(t.n),v.a+=t.o.a/2,v.b+=t.o.b/2,m=ke(Oe(V(t,(ot(),ox)))),l=e.f,g=e.d,u=e.c,f(V(t,(Ge(),Tc)),64).g){case 1:v.a+=g.b+u.a-i/2,v.b=-a-m,t.n.b=-(g.d+m+u.b);break;case 2:v.a=l.a+g.b+g.c+m,v.b+=g.d+u.b-a/2,t.n.a=l.a+g.c+m-u.a;break;case 3:v.a+=g.b+u.a-i/2,v.b=l.b+g.d+g.a+m,t.n.b=l.b+g.a+m-u.b;break;case 4:v.a=-i-m,v.b+=g.d+u.b-a/2,t.n.a=-(g.b+m+u.a)}return v}function oDn(e,t,i){var a,u;a=t.c.i,u=i.d.i,a.k==(cn(),uo)?($e(e,(Ge(),$p),f(V(a,$p),11)),$e(e,yh,f(V(a,yh),11)),$e(e,ex,ut(V(a,ex)))):a.k==Vc?($e(e,(Ge(),$p),f(V(a,$p),11)),$e(e,yh,f(V(a,yh),11)),$e(e,ex,(mn(),!0))):u.k==Vc?($e(e,(Ge(),$p),f(V(u,$p),11)),$e(e,yh,f(V(u,yh),11)),$e(e,ex,(mn(),!0))):($e(e,(Ge(),$p),t.c),$e(e,yh,i.d))}function aDn(e){var t,i,a,u,l,g,m;for(e.o=new O3,a=new ki,g=new ne(e.e.a);g.a<g.c.c.length;)l=f(ee(g),123),V3(l).c.length==1&&Wi(a,l,a.c.b,a.c);for(;a.b!=0;)l=f(a.b==0?null:(Ln(a.b!=0),lh(a,a.a.a)),123),V3(l).c.length!=0&&(t=f(ct(V3(l),0),216),i=l.g.a.c.length>0,m=hZ(t,l),s_e(i?m.b:m.g,t),V3(m).c.length==1&&Wi(a,m,a.c.b,a.c),u=new Ya(l,t),Jv(e.o,u),hl(e.e.a,l))}function o_t(e,t){var i,a,u,l,g,m,v;return a=h.Math.abs($X(e.b).a-$X(t.b).a),m=h.Math.abs($X(e.b).b-$X(t.b).b),u=0,v=0,i=1,g=1,a>e.b.b/2+t.b.b/2&&(u=h.Math.min(h.Math.abs(e.b.c-(t.b.c+t.b.b)),h.Math.abs(e.b.c+e.b.b-t.b.c)),i=1-u/a),m>e.b.a/2+t.b.a/2&&(v=h.Math.min(h.Math.abs(e.b.d-(t.b.d+t.b.a)),h.Math.abs(e.b.d+e.b.a-t.b.d)),g=1-v/m),l=h.Math.min(i,g),(1-l)*h.Math.sqrt(a*a+m*m)}function sDn(e){var t,i,a,u;for(rfe(e,e.e,e.f,(AS(),Aw),!0,e.c,e.i),rfe(e,e.e,e.f,Aw,!1,e.c,e.i),rfe(e,e.e,e.f,LA,!0,e.c,e.i),rfe(e,e.e,e.f,LA,!1,e.c,e.i),iDn(e,e.c,e.e,e.f,e.i),a=new la(e.i,0);a.b<a.d.gc();)for(t=(Ln(a.b<a.d.gc()),f(a.d.Xb(a.c=a.b++),132)),u=new la(e.i,a.b);u.b<u.d.gc();)i=(Ln(u.b<u.d.gc()),f(u.d.Xb(u.c=u.b++),132)),VMn(t,i);d7n(e.i,f(V(e.d,(Ge(),ZR)),233)),A8n(e.i)}function Fde(e,t){var i,a;if(t!=null){if(a=US(e),a)if((a.i&1)!=0){if(a==kd)return Xk(t);if(a==pr)return me(t,17);if(a==kx)return me(t,160);if(a==$l)return me(t,219);if(a==Xf)return me(t,179);if(a==sa)return Jk(t);if(a==JA)return me(t,191);if(a==x_)return me(t,190)}else return CY(),i=f(bn(Fne,a),58),!i||i.ck(t);else if(me(t,57))return e.al(f(t,57))}return!1}function ZIe(){ZIe=Z;var e,t,i,a,u,l,g,m,v;for(zg=Pe($l,hA,30,255,15,1),Q0=Pe(Xf,_m,30,64,15,1),t=0;t<255;t++)zg[t]=-1;for(i=90;i>=65;i--)zg[i]=i-65<<24>>24;for(a=122;a>=97;a--)zg[a]=a-97+26<<24>>24;for(u=57;u>=48;u--)zg[u]=u-48+52<<24>>24;for(zg[43]=62,zg[47]=63,l=0;l<=25;l++)Q0[l]=65+l&To;for(g=26,v=0;g<=51;++g,v++)Q0[g]=97+v&To;for(e=52,m=0;e<=61;++e,m++)Q0[e]=48+m&To;Q0[62]=43,Q0[63]=47}function a_t(e,t){var i,a,u,l,g,m;return u=Wxe(e),m=Wxe(t),u==m?e.e==t.e&&e.a<54&&t.a<54?e.f<t.f?-1:e.f>t.f?1:0:(a=e.e-t.e,i=(e.d>0?e.d:h.Math.floor((e.a-1)*fTt)+1)-(t.d>0?t.d:h.Math.floor((t.a-1)*fTt)+1),i>a+1?u:i<a-1?-u:(l=(!e.c&&(e.c=yH(Nc(e.f))),e.c),g=(!t.c&&(t.c=yH(Nc(t.f))),t.c),a<0?l=M3(l,v_t(-a)):a>0&&(g=M3(g,v_t(a))),lvt(l,g))):u<m?-1:1}function s_t(e){var t,i,a,u,l,g;return a=new Rxe,Rc(a,e),Ae(V(a,(ot(),mf)))===Ae((ko(),$g))&&$e(a,mf,dZ(a)),V(a,(bJ(),TF))==null&&(g=f(dyt(e),173),$e(a,TF,QY(g.mf(TF)))),$e(a,(Ge(),oi),e),$e(a,Cu,(t=f(Tp(Vge),10),new Rf(t,f(Gh(t,t.length),10),0))),u=Z8n((so(e)&&(fS(),new uS(so(e))),fS(),new Rse(so(e)?new uS(so(e)):null,e)),Gs),l=f(V(a,JLe),107),i=a.d,Aft(i,l),Aft(i,u),a}function cDn(e,t,i){var a,u,l,g,m,v,S,T;for(i.Sg(zTt,1),e.qf(t),l=0;e.sf(l)&&!i.Yg();){for(e.rf(),T=pb(Df(ce(se(Gf,1),Jt,20,0,[t.e,t.d,t.b])));Kn(T);)for(v=f(Dn(T),310),m=pb(Df(ce(se(Gf,1),Jt,20,0,[t.e,t.d,t.b])));Kn(m);)g=f(Dn(m),310),g!=v&&(u=e.pf(g,v),u&&ii(v.c,u));for(S=pb(Df(ce(se(Gf,1),Jt,20,0,[t.e,t.d,t.b])));Kn(S);)v=f(Dn(S),310),a=v.c,Cyt(a,-e.d,-e.d,e.d,e.d),ii(v.d,a),a.a=0,a.b=0;++l}i.Tg()}function uDn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F;if(e.dc())return new ha;for(S=0,A=0,u=e.Jc();u.Ob();)a=f(u.Pb(),36),l=a.f,S=h.Math.max(S,l.a),A+=l.a*l.b;for(S=h.Math.max(S,h.Math.sqrt(A)*ke(Oe(V(f(e.Jc().Pb(),36),(ot(),Tte))))),L=0,F=0,v=0,i=t,m=e.Jc();m.Ob();)g=f(m.Pb(),36),T=g.f,L+T.a>S&&(L=0,F+=v+t,v=0),wL(g,L,F),i=h.Math.max(i,L+T.a),v=h.Math.max(v,T.b),L+=T.a+t;return new Ke(i+t,F+v+t)}function QIe(e,t){var i,a,u,l,g,m,v;if(!bg(e))throw ie(new As(kCt));if(a=bg(e),l=a.g,u=a.f,l<=0&&u<=0)return tt(),fc;switch(m=e.i,v=e.j,t.g){case 2:case 1:if(m<0)return tt(),pn;if(m+e.g>l)return tt(),_n;break;case 4:case 3:if(v<0)return tt(),hn;if(v+e.f>u)return tt(),Bn}return g=(m+e.g/2)/l,i=(v+e.f/2)/u,g+i<=1&&g-i<=0?(tt(),pn):g+i>=1&&g-i>=0?(tt(),_n):i<.5?(tt(),hn):(tt(),Bn)}function lDn(e,t,i,a,u){var l,g;if(l=Ua(ia(t[0],ls),ia(a[0],ls)),e[0]=mr(l),l=wS(l,32),i>=u){for(g=1;g<u;g++)l=Ua(l,Ua(ia(t[g],ls),ia(a[g],ls))),e[g]=mr(l),l=wS(l,32);for(;g<i;g++)l=Ua(l,ia(t[g],ls)),e[g]=mr(l),l=wS(l,32)}else{for(g=1;g<i;g++)l=Ua(l,Ua(ia(t[g],ls),ia(a[g],ls))),e[g]=mr(l),l=wS(l,32);for(;g<u;g++)l=Ua(l,ia(a[g],ls)),e[g]=mr(l),l=wS(l,32)}Eu(l,0)!=0&&(e[g]=mr(l))}function dDn(e,t){var i,a,u,l,g,m,v,S,T;for(t.Sg("Layer constraint edge reversal",1),g=new ne(e.b);g.a<g.c.c.length;){for(l=f(ee(g),25),T=-1,i=new et,S=EH(l.a),u=0;u<S.length;u++)a=f(V(S[u],(Ge(),OE)),312),T==-1?a!=(w1(),kA)&&(T=u):a==(w1(),kA)&&(jo(S[u],null),sw(S[u],T++,l)),a==(w1(),WR)&&vn(i.c,S[u]);for(v=new ne(i);v.a<v.c.c.length;)m=f(ee(v),9),jo(m,null),jo(m,l)}t.Tg()}function LT(e){Xr();var t,i,a,u,l,g;if(e.e!=4&&e.e!=5)throw ie(new dn("Token#complementRanges(): must be RANGE: "+e.e));for(l=e,eA(l),Qj(l),a=l.b.length+2,l.b[0]==0&&(a-=2),i=l.b[l.b.length-1],i==JL&&(a-=2),u=new of(4),u.b=Pe(pr,Pr,30,a,15,1),g=0,l.b[0]>0&&(u.b[g++]=0,u.b[g++]=l.b[0]-1),t=1;t<l.b.length-2;t+=2)u.b[g++]=l.b[t]+1,u.b[g++]=l.b[t+1]-1;return i!=JL&&(u.b[g++]=i+1,u.b[g]=JL),u.a=!0,u}function fDn(e,t){var i,a,u,l,g,m,v,S,T;for(t.Sg("Hierarchical port dummy size processing",1),v=new et,T=new et,a=ke(Oe(V(e,(ot(),IA)))),i=a*2,l=new ne(e.b);l.a<l.c.c.length;){for(u=f(ee(l),25),v.c.length=0,T.c.length=0,m=new ne(u.a);m.a<m.c.c.length;)g=f(ee(m),9),g.k==(cn(),bo)&&(S=f(V(g,(Ge(),Tc)),64),S==(tt(),hn)?vn(v.c,g):S==Bn&&vn(T.c,g));MEt(v,!0,i),MEt(T,!1,i)}t.Tg()}function Bde(e,t,i){var a,u,l,g,m,v,S,T;if(a=i.gc(),a==0)return!1;if(e.Mj())if(S=e.Nj(),sde(e,t,i),g=a==1?e.Fj(3,null,i.Jc().Pb(),t,S):e.Fj(5,null,i,t,S),e.Jj()){for(m=a<100?null:new Fv(a),l=t+a,u=t;u<l;++u)T=e.g[u],m=e.Kj(T,m),m=e.Rj(T,m);m?(m.kj(g),m.lj()):e.Gj(g)}else e.Gj(g);else if(sde(e,t,i),e.Jj()){for(m=a<100?null:new Fv(a),l=t+a,u=t;u<l;++u)v=e.g[u],m=e.Kj(v,m);m&&m.lj()}return!0}function eOe(e,t,i,a){var u,l,g,m,v;for(g=new ne(e.k);g.a<g.c.c.length;)u=f(ee(g),133),(!a||u.c==(Ip(),Cw))&&(v=u.b,v.g<0&&u.d>0&&(yN(v,v.d-u.d),u.c==(Ip(),Cw)&&wN(v,v.a-u.d),v.d<=0&&v.i>0&&Wi(t,v,t.c.b,t.c)));for(l=new ne(e.f);l.a<l.c.c.length;)u=f(ee(l),133),(!a||u.c==(Ip(),Cw))&&(m=u.a,m.g<0&&u.d>0&&(aS(m,m.i-u.d),u.c==(Ip(),Cw)&&qK(m,m.b-u.d),m.i<=0&&m.d>0&&Wi(i,m,i.c.b,i.c)))}function hDn(e,t,i,a,u){var l,g,m,v,S,T,A,L,F;for(Ht(),na(e,new Z8),g=cH(e),F=new et,L=new et,m=null,v=0;g.b!=0;)l=f(g.b==0?null:(Ln(g.b!=0),lh(g,g.a.a)),162),!m||Al(m)*jd(m)/2<Al(l)*jd(l)?(m=l,vn(F.c,l)):(v+=Al(l)*jd(l),vn(L.c,l),L.c.length>1&&(v>Al(m)*jd(m)/2||g.b==0)&&(A=new LJ(L),T=Al(m)/jd(m),S=ife(A,t,new xN,i,a,u,T),ii(Cp(A.e),S),m=A,vn(F.c,A),v=0,L.c.length=0));return $o(F,L),F}function au(e,t,i,a,u){b0();var l,g,m,v,S,T,A;if(Mke(e,"src"),Mke(i,"dest"),A=Fd(e),v=Fd(i),ake((A.i&4)!=0,"srcType is not an array"),ake((v.i&4)!=0,"destType is not an array"),T=A.c,g=v.c,ake((T.i&1)!=0?T==g:(g.i&1)==0,"Array types don't match"),JTn(e,t,i,a,u),(T.i&1)==0&&A!=v)if(S=uR(e),l=uR(i),Ae(e)===Ae(i)&&t<a)for(t+=u,m=a+u;m-- >a;)io(l,m,S[--t]);else for(m=a+u;a<m;)io(l,a++,S[t++]);else KAe(e,t,i,a,u,!0)}function c_t(e,t){var i,a,u,l,g,m,v,S,T;switch(t.Sg("Box layout",2),u=_z(Oe(Fe(e,(GZ(),FLt)))),l=f(Fe(e,$Lt),107),i=lt(ut(Fe(e,e9e))),a=lt(ut(Fe(e,t9e))),f(Fe(e,Ube),321).g){case 0:g=(T=new Kl((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a)),Ht(),na(T,new Xtt(a)),T),m=RAe(e),v=Oe(Fe(e,Q8e)),(v==null||(an(v),v<=0))&&(v=1.3),S=$9n(g,u,l,m.a,m.b,i,(an(v),v)),WS(e,S.a,S.b,!1,!0);break;default:MMn(e,u,l,i)}t.Tg()}function pDn(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z;for(L=dOn(e,i),v=0;v<t;v++){for(nT(u,i),F=new et,z=(Ln(a.b<a.d.gc()),f(a.d.Xb(a.c=a.b++),407)),T=L+v;T<e.b;T++)m=z,z=(Ln(a.b<a.d.gc()),f(a.d.Xb(a.c=a.b++),407)),Qe(F,new $2t(m,z,i));for(A=L+v;A<e.b;A++)Ln(a.b>0),a.a.Xb(a.c=--a.b),A>L+v&&hd(a);for(g=new ne(F);g.a<g.c.c.length;)l=f(ee(g),407),nT(a,l);if(v<t-1)for(S=L+v;S<e.b;S++)Ln(a.b>0),a.a.Xb(a.c=--a.b)}}function gDn(){Xr();var e,t,i,a,u,l;if(k1e)return k1e;for(e=new of(4),MT(e,dw(Rpe,!0)),t$(e,dw("M",!0)),t$(e,dw("C",!0)),l=new of(4),a=0;a<11;a++)Su(l,a,a);return t=new of(4),MT(t,dw("M",!0)),Su(t,4448,4607),Su(t,65438,65439),u=new D7(2),vE(u,e),vE(u,XF),i=new D7(2),i.Gm(RX(l,dw("L",!0))),i.Gm(t),i=new hT(3,i),i=new $ke(u,i),k1e=i,k1e}function DT(e,t){var i,a,u,l,g,m,v,S;for(i=new RegExp(t,"g"),v=Pe(ft,Ye,2,0,6,1),a=0,S=e,l=null;;)if(m=i.exec(S),m==null||S==""){v[a]=S;break}else g=m.index,v[a]=(ga(0,g,S.length),S.substr(0,g)),S=uh(S,g+m[0].length,S.length),i.lastIndex=0,l==S&&(v[a]=(ga(0,1,S.length),S.substr(0,1)),S=(gn(1,S.length+1),S.substr(1))),l=S,++a;if(e.length>0){for(u=v.length;u>0&&v[u-1]=="";)--u;u<v.length&&(v.length=u)}return v}function Ec(){Ec=Z,vMe=new gS(20),bMe=new pa((Rr(),V0),vMe),Qme=new pa(zE,20),yMe=new pa(L9e,3),P5t=new pa(ID,DL),tne=new pa(aV,rt(1)),z5t=new pa(n1e,(mn(),!0)),hMe=nV,pMe=(ko(),$g),E_=new pa(BE,pMe),L5t=iV,D5t=Kbe,j5t=UE,$5t=hx,F5t=GA,B5t=Ow,M5t=HA,mMe=oV,U5t=px,SMe=(QAe(),R5t),wMe=O5t,q5t=RD,V5t=Sne,G5t=ND,H5t=sV,EMe=(gR(),bx),new pa(s4,EMe),DA=I5t,Zme=A5t,Lm=N5t,fMe=x5t,gMe=C5t}function mDn(e){var t,i,a,u,l,g,m,v,S,T,A;for(A=new Ty,A.d=0,g=new ne(e.b);g.a<g.c.c.length;)l=f(ee(g),25),A.d+=l.a.c.length;for(a=0,u=0,A.a=Pe(pr,Pr,30,e.b.c.length,15,1),S=0,T=0,A.e=Pe(pr,Pr,30,A.d,15,1),i=new ne(e.b);i.a<i.c.c.length;)for(t=f(ee(i),25),t.p=a++,A.a[t.p]=u++,T=0,v=new ne(t.a);v.a<v.c.c.length;)m=f(ee(v),9),m.p=S++,A.e[m.p]=T++;return A.c=new _tt(A),A.b=lb(A.d),DLn(A,e),A.f=lb(A.d),MLn(A,e),A}function bDn(e){var t,i;if(t=dr(Fe(e,(Rr(),AD))),!Gbt(t,e)&&!mm(e,VA)&&((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a).i!=0||lt(ut(Fe(e,rV)))))if(t==null||IT(t).length==0){if(!Gbt(wn,e))throw i=xr(xr(new ef("Unable to load default layout algorithm "),wn)," for unconfigured node "),mQ(e,i),ie(new h0(i.a))}else throw i=xr(xr(new ef("Layout algorithm '"),t),"' not found for "),mQ(e,i),ie(new h0(i.a))}function Ude(e){var t,i,a,u,l,g,m,v,S,T,A,L,F;if(i=e.i,t=e.n,e.b==0)for(F=i.c+t.b,L=i.b-t.b-t.c,g=e.a,v=0,T=g.length;v<T;++v)u=g[v],DX(u,F,L);else a=qvt(e,!1),DX(e.a[0],i.c+t.b,a[0]),DX(e.a[2],i.c+i.b-t.c-a[2],a[2]),A=i.b-t.b-t.c,a[0]>0&&(A-=a[0]+e.c,a[0]+=e.c),a[2]>0&&(A-=a[2]+e.c),a[1]=h.Math.max(a[1],A),DX(e.a[1],i.c+t.b+a[0]-(a[1]-A)/2,a[1]);for(l=e.a,m=0,S=l.length;m<S;++m)u=l[m],me(u,332)&&f(u,332).hf()}function vDn(e){var t,i,a,u,l,g,m,v,S,T;for(T=Pe(pr,Pr,30,e.b.c.length+1,15,1),S=new Mo,a=0,l=new ne(e.b);l.a<l.c.c.length;){for(u=f(ee(l),25),T[a++]=S.a.gc(),v=new ne(u.a);v.a<v.c.c.length;)for(g=f(ee(v),9),i=new Tn(Cn(ao(g).a.Jc(),new B));Kn(i);)t=f(Dn(i),18),S.a.yc(t,S);for(m=new ne(u.a);m.a<m.c.c.length;)for(g=f(ee(m),9),i=new Tn(Cn(Fo(g).a.Jc(),new B));Kn(i);)t=f(Dn(i),18),S.a.Ac(t)!=null}return T}function u_t(e,t){var i,a,u,l;for(l=f(ct(e.n,e.n.c.length-1),208).d,e.p=h.Math.min(e.p,t.g),e.r=h.Math.max(e.r,l),e.g=h.Math.max(e.g,t.g+(e.b.c.length==1?0:e.i)),e.o=h.Math.min(e.o,t.f),e.e+=t.f+(e.b.c.length==1?0:e.i),e.f=h.Math.max(e.f,t.f),u=e.n.c.length>0?(e.n.c.length-1)*e.i:0,a=new ne(e.n);a.a<a.c.c.length;)i=f(ee(a),208),u+=i.a;e.d=u,e.a=e.e/e.b.c.length-e.i*((e.b.c.length-1)/e.b.c.length),O3e(e.j)}function l_t(e,t){var i,a,u,l,g,m,v,S,T,A;if(T=ut(V(t,(Qh(),cOt))),T==null||(an(T),T)){for(A=Pe(kd,Tb,30,t.e.c.length,16,1),g=v4n(t),u=new ki,S=new ne(t.e);S.a<S.c.c.length;)m=f(ee(S),151),i=dIe(e,m,null,null,A,g),i&&(Rc(i,t),Wi(u,i,u.c.b,u.c));if(u.b>1)for(a=tr(u,0);a.b!=a.d.c;)for(i=f(Zn(a),234),l=0,v=new ne(i.e);v.a<v.c.c.length;)m=f(ee(v),151),m.a=l++;return u}return gm(ce(se(rjn,1),FTt,234,0,[t]))}function Em(e){var t,i,a,u,l,g,m;if(!e.g){if(m=new n1,t=GF,g=t.a.yc(e,t),g==null){for(a=new Nn(Bs(e));a.e!=a.i.gc();)i=f(Mn(a),29),Qi(m,Em(i));t.a.Ac(e)!=null,t.a.gc()==0}for(u=m.i,l=(!e.s&&(e.s=new Be(El,e,21,17)),new Nn(e.s));l.e!=l.i.gc();++u)_(f(Mn(l),448),u);Qi(m,(!e.s&&(e.s=new Be(El,e,21,17)),e.s)),ET(m),e.g=new L1t(e,m),e.i=f(m.g,253),e.i==null&&(e.i=b1e),e.p=null,pd(e).b&=-5}return e.g}function wDn(e,t){var i,a,u,l,g,m,v,S,T;if(i=t.mi(e.a),i&&(v=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),"memberTypes")),v!=null)){for(S=new et,l=DT(v,"\\w"),g=0,m=l.length;g<m;++g)u=l[g],a=u.lastIndexOf("#"),T=a==-1?b_e(e,t.gk(),u):a==0?OH(e,null,(gn(1,u.length+1),u.substr(1))):OH(e,(ga(0,a,u.length),u.substr(0,a)),(gn(a+1,u.length+1),u.substr(a+1))),me(T,155)&&Qe(S,f(T,155));return S}return Ht(),Ht(),Ga}function zde(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z;if(a=e.i,i=e.n,e.b==0)t=Gvt(e,!1),LX(e.a[0],a.d+i.d,t[0]),LX(e.a[2],a.d+a.a-i.a-t[2],t[2]),L=a.a-i.d-i.a,A=L,t[0]>0&&(t[0]+=e.c,A-=t[0]),t[2]>0&&(A-=t[2]+e.c),t[1]=h.Math.max(t[1],A),LX(e.a[1],a.d+i.d+t[0]-(t[1]-A)/2,t[1]);else for(z=a.d+i.d,F=a.a-i.d-i.a,g=e.a,v=0,T=g.length;v<T;++v)u=g[v],LX(u,z,F);for(l=e.a,m=0,S=l.length;m<S;++m)u=l[m],me(u,332)&&f(u,332).jf()}function lQ(e,t,i,a){var u,l,g,m,v;if(v=Ju(e.e.zh(),t),u=f(e.g,122),is(),f(t,69).uk()){for(g=0;g<e.i;++g)if(l=u[g],v.Zl(l.Ik())&&ei(l,i))return!0}else if(i!=null){for(m=0;m<e.i;++m)if(l=u[m],v.Zl(l.Ik())&&ei(i,l.kd()))return!0;if(a){for(g=0;g<e.i;++g)if(l=u[g],v.Zl(l.Ik())&&Ae(i)===Ae(jse(e,f(l.kd(),57))))return!0}}else for(g=0;g<e.i;++g)if(l=u[g],v.Zl(l.Ik())&&l.kd()==null)return!1;return!1}function yDn(e,t){var i,a,u,l,g,m;if(i=t.mi(e.a),i&&(m=dr(Rp((!i.b&&(i.b=new Md((Ut(),Za),xc,i)),i.b),cee)),m!=null))switch(u=sX(m,gl(35)),a=t.nk(),u==-1?(g=q7(e,cf(a)),l=m):u==0?(g=null,l=(gn(1,m.length+1),m.substr(1))):(g=(ga(0,u,m.length),m.substr(0,u)),l=(gn(u+1,m.length+1),m.substr(u+1))),kS(Ns(e,t))){case 2:case 3:return GTn(e,a,g,l);case 0:case 4:case 5:case 6:return HTn(e,a,g,l)}return null}function d_t(e,t,i,a){var u,l,g,m;for(m=i,g=new ne(t.a);g.a<g.c.c.length;){if(l=f(ee(g),224),u=f(l.b,68),jS(e.b.c,u.b.c+u.b.b)<=0&&jS(u.b.c,e.b.c+e.b.b)<=0&&jS(e.b.d,u.b.d+u.b.a)<=0&&jS(u.b.d,e.b.d+e.b.a)<=0){if(jS(u.b.c,e.b.c+e.b.b)==0&&a.a<0||jS(u.b.c+u.b.b,e.b.c)==0&&a.a>0||jS(u.b.d,e.b.d+e.b.a)==0&&a.b<0||jS(u.b.d+u.b.a,e.b.d)==0&&a.b>0){m=0;break}}else m=h.Math.min(m,tEt(e,u,a));m=h.Math.min(m,d_t(e,l,m,a))}return m}function tOe(e,t){var i,a,u,l,g,m,v;if(e.b<2)throw ie(new dn("The vector chain must contain at least a source and a target point."));for(u=(Ln(e.b!=0),f(e.a.a.c,8)),Yz(t,u.a,u.b),v=new HN((!t.a&&(t.a=new So(yf,t,5)),t.a)),g=tr(e,1);g.a<e.b-1;)m=f(Zn(g),8),v.e!=v.i.gc()?i=f(Mn(v),369):(i=(Bv(),a=new Iv,a),Q0t(v,i)),Ese(i,m.a,m.b);for(;v.e!=v.i.gc();)Mn(v),Mj(v);l=(Ln(e.b!=0),f(e.c.b.c,8)),Kz(t,l.a,l.b)}function f_t(e,t,i,a){var u,l,g,m,v,S;if(S=Ju(e.e.zh(),t),g=f(e.g,122),x1(e.e,t)){if(t.Pi()&&(l=EL(e,t,a,me(t,101)&&(f(t,19).Bb&Xa)!=0),l>=0&&l!=i))throw ie(new dn(lq));for(u=0,v=0;v<e.i;++v)if(m=g[v],S.Zl(m.Ik())){if(u==i)return f(Y3(e,v,(is(),f(t,69).uk()?f(a,74):mb(t,a))),74);++u}throw ie(new Ru(T$+i+TE+u))}else{for(v=0;v<e.i;++v)if(m=g[v],S.Zl(m.Ik()))return is(),f(t,69).uk()?m:m.kd();return null}}function h_t(e,t){var i,a,u,l,g,m,v,S,T;for(i=0,u=new ne((Gt(0,e.c.length),f(e.c[0],105)).g.b.j);u.a<u.c.c.length;)a=f(ee(u),11),a.p=i++;for(t==(tt(),hn)?na(e,new rb):na(e,new v8),m=0,T=e.c.length-1;m<T;)g=(Gt(m,e.c.length),f(e.c[m],105)),S=(Gt(T,e.c.length),f(e.c[T],105)),l=t==hn?g.c:g.a,v=t==hn?S.a:S.c,E0(g,t,(_g(),iD),l),E0(S,t,rD,v),++m,--T;m==T&&E0((Gt(m,e.c.length),f(e.c[m],105)),t,(_g(),VR),null)}function EDn(e,t,i,a){var u,l,g,m,v,S;for(g=new z_t(e,t,i),v=new la(a,0),u=!1;v.b<v.d.gc();)m=(Ln(v.b<v.d.gc()),f(v.d.Xb(v.c=v.b++),238)),m==t||m==i?hd(v):!u&&ke(mg(m.g,m.d[0]).a)>ke(mg(g.g,g.d[0]).a)?(Ln(v.b>0),v.a.Xb(v.c=--v.b),nT(v,g),u=!0):m.e&&m.e.gc()>0&&(l=(!m.e&&(m.e=new et),m.e).Kc(t),S=(!m.e&&(m.e=new et),m.e).Kc(i),(l||S)&&((!m.e&&(m.e=new et),m.e).Ec(g),++g.c));u||vn(a.c,g)}function SDn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;return A=e.a.i+e.a.g/2,L=e.a.i+e.a.g/2,z=t.i+t.g/2,J=t.j+t.f/2,m=new Ke(z,J),S=f(Fe(t,(Rr(),o4)),8),S.a=S.a+A,S.b=S.b+L,l=(m.b-S.b)/(m.a-S.a),a=m.b-l*m.a,W=i.i+i.g/2,ue=i.j+i.f/2,v=new Ke(W,ue),T=f(Fe(i,o4),8),T.a=T.a+A,T.b=T.b+L,g=(v.b-T.b)/(v.a-T.a),u=v.b-g*v.a,F=(a-u)/(g-l),S.a<F&&m.a<F||F<S.a&&F<m.a?!1:!(T.a<F&&v.a<F||F<T.a&&F<v.a)}function p_t(e,t,i,a){var u,l,g,m,v,S;if(!NP(t)){if(S=i.bh((me(t,16)?f(t,16).gc():g1(t.Jc()))/e.a|0),S.Sg(Rxt,1),v=new PO,m=0,a==(ko(),Ds)||a==Gs)for(g=t.Jc();g.Ob();)u=f(g.Pb(),39),v=Df(ce(se(Gf,1),Jt,20,0,[v,new u1(u)])),m<u.f.a&&(m=u.f.a);else for(g=t.Jc();g.Ob();)u=f(g.Pb(),39),v=Df(ce(se(Gf,1),Jt,20,0,[v,new u1(u)])),m<u.f.b&&(m=u.f.b);for(l=t.Jc();l.Ob();)u=f(l.Pb(),39),$e(u,(_i(),ene),m);S.Tg(),p_t(e,v,i,a)}}function nOe(e,t,i){var a,u,l,g,m,v,S,T;this.a=e,this.b=t,this.c=i,this.e=gm(ce(se(ejn,1),Jt,176,0,[new DN(e,t),new DN(t,i),new DN(i,e)])),this.f=gm(ce(se(Zo,1),Ye,8,0,[e,t,i])),this.d=(a=Ho($a(this.b),this.a),u=Ho($a(this.c),this.a),l=Ho($a(this.c),this.b),g=a.a*(this.a.a+this.b.a)+a.b*(this.a.b+this.b.b),m=u.a*(this.a.a+this.c.a)+u.b*(this.a.b+this.c.b),v=2*(a.a*l.b-a.b*l.a),S=(u.b*g-a.b*m)/v,T=(a.a*m-u.a*g)/v,new Ke(S,T))}function VS(e,t){var i,a,u,l,g,m;for(l=e.c,g=e.d,Ia(e,null),ua(e,null),t&&lt(ut(V(g,(Ge(),Qge))))?Ia(e,WIe(g.i,(cs(),Uu),(tt(),_n))):Ia(e,g),t&&lt(ut(V(l,(Ge(),tme))))?ua(e,WIe(l.i,(cs(),td),(tt(),pn))):ua(e,l),a=new ne(e.b);a.a<a.c.c.length;)i=f(ee(a),72),u=f(V(i,(ot(),Rm)),276),u==(Sg(),LD)?$e(i,Rm,gx):u==gx&&$e(i,Rm,LD);m=lt(ut(V(e,(Ge(),U0)))),$e(e,U0,(mn(),!m)),e.a=vCe(e.a)}function _Dn(e,t){var i,a,u,l,g;return i=sL(f(V(t,(Ec(),E_)),86)),e.b.b==0?null:(g=f($d(Lu(new Ft(null,new Bt(e.b,16)),new Pd),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),l=f($d(Jr(new Ft(null,new Bt(t.b,16)),new Ttt(g)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[Zl]))),15),u=Oe(Dd(oT(Lu(l.Mc(),new xtt(i)),(TS(),TS(),Kpe)))),a=f(Dd(S1(Jr(l.Mc(),new mat(i,u)))),39),a)}function kDn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;i=mS(new qy,e.f),S=e.i[t.c.i.p],F=e.i[t.d.i.p],v=t.c,L=t.d,m=v.a.b,A=L.a.b,S.b||(m+=v.n.b),F.b||(A+=L.n.b),T=Aa(h.Math.max(0,m-A)),g=Aa(h.Math.max(0,A-m)),z=(W=h.Math.max(1,f(V(t,(ot(),e4)),17).a),J=DTe(t.c.i.k,t.d.i.k),W*J),u=Zh(Uh(Bh(Fh(zh(new ah,z),g),i),f(bn(e.k,t.c),123))),l=Zh(Uh(Bh(Fh(zh(new ah,z),T),i),f(bn(e.k,t.d),123))),a=new iat(u,l),e.c[t.p]=a}function TDn(e,t,i){var a,u,l,g,m,v;for(a=0,l=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));l.e!=l.i.gc();)u=f(Mn(l),27),g="",(!u.n&&(u.n=new Be(hc,u,1,7)),u.n).i==0||(g=f(be((!u.n&&(u.n=new Be(hc,u,1,7)),u.n),0),153).a),m=new Dst(g),Rc(m,u),$e(m,(ew(),qR),u),m.a=a++,m.d.a=u.i+u.g/2,m.d.b=u.j+u.f/2,m.e.a=h.Math.max(u.g,1),m.e.b=h.Math.max(u.f,1),Qe(t.e,m),ml(i.f,u,m),v=f(Fe(u,(Qh(),t5e)),100),v==(oa(),Pw)&&(v=Fg)}function g_t(e){var t,i,a;if(T3(f(V(e,(ot(),Xi)),100)))for(i=new ne(e.j);i.a<i.c.c.length;)t=f(ee(i),11),t.j==(tt(),fc)&&(a=f(V(t,(Ge(),ed)),9),a?No(t,f(V(a,Tc),64)):t.e.c.length-t.g.c.length<0?No(t,_n):No(t,pn));else{for(i=new ne(e.j);i.a<i.c.c.length;)t=f(ee(i),11),a=f(V(t,(Ge(),ed)),9),a?No(t,f(V(a,Tc),64)):t.e.c.length-t.g.c.length<0?No(t,(tt(),_n)):No(t,(tt(),pn));$e(e,Xi,(oa(),MD))}}function xDn(e,t){var i,a,u,l,g,m,v;t.Sg("Layer constraint postprocessing",1),v=e.b,v.c.length!=0&&(a=(Gt(0,v.c.length),f(v.c[0],25)),g=f(ct(v,v.c.length-1),25),i=new ru(e),l=new ru(e),mLn(e,a,g,i,l),i.a.c.length==0||(fT(0,v.c.length),y7(v.c,0,i)),l.a.c.length==0||vn(v.c,l)),li(e,(Ge(),Zge))&&(u=new ru(e),m=new ru(e),sPn(e,u,m),u.a.c.length==0||(fT(0,v.c.length),y7(v.c,0,u)),m.a.c.length==0||vn(v.c,m)),t.Tg()}function IG(e){var t,i,a;switch(e){case 91:case 93:case 45:case 94:case 44:case 92:a="\\"+String.fromCharCode(e&To);break;case 12:a="\\f";break;case 10:a="\\n";break;case 13:a="\\r";break;case 9:a="\\t";break;case 27:a="\\e";break;default:e<32?(i=(t=e>>>0,"0"+t.toString(16)),a="\\x"+uh(i,i.length-2,i.length)):e>=Xa?(i=(t=e>>>0,"0"+t.toString(16)),a="\\v"+uh(i,i.length-6,i.length)):a=""+String.fromCharCode(e&To)}return a}function m_t(e,t){var i,a,u,l,g,m,v,S,T;for(l=new ne(e.b);l.a<l.c.c.length;)for(u=f(ee(l),25),m=new ne(u.a);m.a<m.c.c.length;)for(g=f(ee(m),9),g.k==(cn(),Vc)&&(v=(S=f(Dn(new Tn(Cn(Fo(g).a.Jc(),new B))),18),T=f(Dn(new Tn(Cn(ao(g).a.Jc(),new B))),18),!lt(ut(V(S,(Ge(),U0))))||!lt(ut(V(T,U0)))?t:b1t(t)),TR(g,v)),a=new Tn(Cn(ao(g).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),v=lt(ut(V(i,(Ge(),U0))))?b1t(t):t,l1t(i,v)}function rOe(e){var t,i,a,u,l,g;for(this.e=new et,this.a=new et,i=e.b-1;i<3;i++)pP(e,0,f(Xh(e,0),8));if(e.b<4)throw ie(new dn("At (least dimension + 1) control points are necessary!"));for(this.b=3,this.d=!0,this.c=!1,f6n(this,e.b+this.b-1),g=new et,l=new ne(this.e),t=0;t<this.b-1;t++)Qe(g,Oe(ee(l)));for(u=tr(e,0);u.b!=u.d.c;)a=f(Zn(u),8),Qe(g,Oe(ee(l))),Qe(this.a,new wft(a,g)),Gt(0,g.c.length),g.c.splice(0,1)}function CDn(e,t,i,a,u){var l,g,m;if(i.f>=t.o&&i.f<=t.f||t.a*.5<=i.f&&t.a*1.5>=i.f){if(g=f(ct(t.n,t.n.c.length-1),208),g.e+g.d+i.g+u<=a&&(l=f(ct(t.n,t.n.c.length-1),208),l.f-e.f+i.f<=e.b||e.a.c.length==1))return KCe(t,i),!0;if(t.s+i.g<=a&&(t.t+t.d+i.f+u<=e.b||e.a.c.length==1))return Qe(t.b,i),m=f(ct(t.n,t.n.c.length-1),208),Qe(t.n,new iJ(t.s,m.f+m.a+t.i,t.i)),C3e(f(ct(t.n,t.n.c.length-1),208),i),u_t(t,i),!0}return!1}function b_t(e,t,i){var a,u,l,g;return e.Mj()?(u=null,l=e.Nj(),a=e.Fj(1,g=Sue(e,t,i),i,t,l),e.Jj()&&!(e.Vi()&&g!=null?ei(g,i):Ae(g)===Ae(i))?(g!=null&&(u=e.Lj(g,u)),u=e.Kj(i,u),e.Qj()&&(u=e.Tj(g,i,u)),u?(u.kj(a),u.lj()):e.Gj(a)):(e.Qj()&&(u=e.Tj(g,i,u)),u?(u.kj(a),u.lj()):e.Gj(a)),g):(g=Sue(e,t,i),e.Jj()&&!(e.Vi()&&g!=null?ei(g,i):Ae(g)===Ae(i))&&(u=null,g!=null&&(u=e.Lj(g,null)),u=e.Kj(i,u),u&&u.lj()),g)}function ADn(e,t){var i,a,u,l,g;if(t.Sg("Path-Like Graph Wrapping",1),e.b.c.length==0){t.Tg();return}if(u=new nIe(e),g=(u.i==null&&(u.i=Kxe(u,new S8)),ke(u.i)*u.f),i=g/(u.i==null&&(u.i=Kxe(u,new S8)),ke(u.i)),u.b>i){t.Tg();return}switch(f(V(e,(ot(),_me)),347).g){case 2:l=new Qc;break;case 0:l=new fk;break;default:l=new kO}if(a=l.mg(e,u),!l.ng())switch(f(V(e,Bte),348).g){case 2:a=nEt(u,a);break;case 1:a=Vwt(u,a)}yMn(e,u,a),t.Tg()}function Yj(e,t){var i,a,u,l,g,m,v,S;t%=24,e.q.getHours()!=t&&(a=new h.Date(e.q.getTime()),a.setDate(a.getDate()+1),m=e.q.getTimezoneOffset()-a.getTimezoneOffset(),m>0&&(v=m/60|0,S=m%60,u=e.q.getDate(),i=e.q.getHours(),i+v>=24&&++u,l=new h.Date(e.q.getFullYear(),e.q.getMonth(),u,t+v,e.q.getMinutes()+S,e.q.getSeconds(),e.q.getMilliseconds()),e.q.setTime(l.getTime()))),g=e.q.getTime(),e.q.setTime(g+36e5),e.q.getHours()!=t&&e.q.setTime(g)}function IDn(e,t){var i,a,u,l;if(bEn(e.d,e.e),e.c.a.$b(),ke(Oe(V(t.j,(ot(),Ate))))!=0||ke(Oe(V(t.j,Ate)))!=0)for(i=$R,Ae(V(t.j,ip))!==Ae((Pp(),Pm))&&$e(t.j,(Ge(),kw),(mn(),!0)),l=f(V(t.j,iF),17).a,u=0;u<l&&(a=UDn(e,t),!(a<i&&(i=a,Abt(e),i==0)));u++);else for(i=Qr,Ae(V(t.j,ip))!==Ae((Pp(),Pm))&&$e(t.j,(Ge(),kw),(mn(),!0)),l=f(V(t.j,iF),17).a,u=0;u<l&&(a=E_t(e,t),!(a<i&&(i=a,Abt(e),i==0)));u++);}function ODn(e,t){var i,a,u,l,g,m,v,S;for(g=new et,m=0,i=0,v=0;m<t.c.length-1&&i<e.gc();){for(a=f(e.Xb(i),17).a+v;(Gt(m+1,t.c.length),f(t.c[m+1],17)).a<a;)++m;for(S=0,l=a-(Gt(m,t.c.length),f(t.c[m],17)).a,u=(Gt(m+1,t.c.length),f(t.c[m+1],17)).a-a,l>u&&++S,Qe(g,(Gt(m+S,t.c.length),f(t.c[m+S],17))),v+=(Gt(m+S,t.c.length),f(t.c[m+S],17)).a-a,++i;i<e.gc()&&f(e.Xb(i),17).a+v<=(Gt(m+S,t.c.length),f(t.c[m+S],17)).a;)++i;m+=1+S}return g}function NDn(e,t){var i,a,u,l,g;for(g=new Tn(Cn(Fo(t).a.Jc(),new B));Kn(g);)if(l=f(Dn(g),18),e.f.b==0?(u=l.c.i.k==(cn(),Yi)&&!!l.c.i.c&&l.c.i.c.p==e.c,Kn(new Tn(Cn(Fo(l.c.i).a.Jc(),new B)))?(i=f(Dn(new Tn(Cn(Fo(l.c.i).a.Jc(),new B))),18).c.i.c,a=l.c.i.k==Vc&&!!i&&i.p==e.c):a=!1):(u=l.c.i.k==(cn(),Yi)&&l.c.i.p==e.c,a=l.c.i.k==Vc&&f(Dn(new Tn(Cn(Fo(l.c.i).a.Jc(),new B))),18).c.i.p==e.c),u||a)return!0;return!1}function RDn(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(L=new et,ue=fJ(a),J=t*e.a,A=0,z=0,l=new Mo,g=new Mo,m=new et,pe=0,ve=0,F=0,W=0,S=0,T=0;ue.a.gc()!=0;)v=cCn(ue,u,g),v&&(ue.a.Ac(v)!=null,vn(m.c,v),l.a.yc(v,l),z=e.f[v.p],pe+=e.e[v.p]-z*e.b,A=e.c[v.p],ve+=A*e.b,T+=z*e.b,W+=e.e[v.p]),(!v||ue.a.gc()==0||pe>=J&&e.e[v.p]>z*e.b||ve>=i*J)&&(vn(L.c,m),m=new et,Ba(g,l),l.a.$b(),S-=T,F=h.Math.max(F,S*e.b+W),S+=ve,pe=ve,ve=0,T=0,W=0);return new Ya(F,L)}function Hde(e){var t,i,a,u,l,g,m;if(!e.d){if(m=new zU,t=GF,l=t.a.yc(e,t),l==null){for(a=new Nn(Bs(e));a.e!=a.i.gc();)i=f(Mn(a),29),Qi(m,Hde(i));t.a.Ac(e)!=null,t.a.gc()==0}for(g=m.i,u=(!e.q&&(e.q=new Be(Th,e,11,10)),new Nn(e.q));u.e!=u.i.gc();++g)f(Mn(u),400);Qi(m,(!e.q&&(e.q=new Be(Th,e,11,10)),e.q)),ET(m),e.d=new _3((f(be(je((Wv(),sn).o),9),19),m.i),m.g),e.e=f(m.g,673),e.e==null&&(e.e=gMt),pd(e).b&=-17}return e.d}function EL(e,t,i,a){var u,l,g,m,v,S;if(S=Ju(e.e.zh(),t),v=0,u=f(e.g,122),is(),f(t,69).uk()){for(g=0;g<e.i;++g)if(l=u[g],S.Zl(l.Ik())){if(ei(l,i))return v;++v}}else if(i!=null){for(m=0;m<e.i;++m)if(l=u[m],S.Zl(l.Ik())){if(ei(i,l.kd()))return v;++v}if(a){for(v=0,g=0;g<e.i;++g)if(l=u[g],S.Zl(l.Ik())){if(Ae(i)===Ae(jse(e,f(l.kd(),57))))return v;++v}}}else for(g=0;g<e.i;++g)if(l=u[g],S.Zl(l.Ik())){if(l.kd()==null)return v;++v}return-1}function PDn(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W;if(i.Th(t)&&(T=(F=t,F?f(a,52).ci(F):null),T))if(W=i.Jh(t,e.a),z=t.t,z>1||z==-1)if(A=f(W,71),L=f(T,71),A.dc())L.$b();else for(g=!!os(t),l=0,m=e.a?A.Jc():A.Fi();m.Ob();)S=f(m.Pb(),57),u=f(Eg(e,S),57),u?(g?(v=L.bd(u),v==-1?L.Di(l,u):l!=v&&L.Ri(l,u)):L.Di(l,u),++l):e.b&&!g&&(L.Di(l,S),++l);else W==null?T.Wb(null):(u=Eg(e,W),u==null?e.b&&!os(t)&&T.Wb(W):T.Wb(u))}function LDn(e,t){var i,a,u,l,g,m,v,S;for(i=new aO,u=new Tn(Cn(Fo(t).a.Jc(),new B));Kn(u);)if(a=f(Dn(u),18),!Fa(a)&&(m=a.c.i,cAe(m,Hee))){if(S=OIe(e,m,Hee,zee),S==-1)continue;i.b=h.Math.max(i.b,S),!i.a&&(i.a=new et),Qe(i.a,m)}for(g=new Tn(Cn(ao(t).a.Jc(),new B));Kn(g);)if(l=f(Dn(g),18),!Fa(l)&&(v=l.d.i,cAe(v,zee))){if(S=OIe(e,v,zee,Hee),S==-1)continue;i.d=h.Math.max(i.d,S),!i.c&&(i.c=new et),Qe(i.c,v)}return i}function DDn(e,t,i,a){var u,l,g,m,v,S,T;if(i.d.i!=t.i){for(u=new Tg(e),Wl(u,(cn(),uo)),$e(u,(Ge(),oi),i),$e(u,(ot(),Xi),(oa(),uu)),vn(a.c,u),g=new ou,cc(g,u),No(g,(tt(),pn)),m=new ou,cc(m,u),No(m,_n),T=i.d,ua(i,g),l=new xS,Rc(l,i),$e(l,Ls,null),Ia(l,m),ua(l,T),S=new la(i.b,0);S.b<S.d.gc();)v=(Ln(S.b<S.d.gc()),f(S.d.Xb(S.c=S.b++),72)),Ae(V(v,Rm))===Ae((Sg(),gx))&&($e(v,YR,i),hd(S),Qe(l.b,v));YEt(u,g,m)}}function MDn(e,t,i,a){var u,l,g,m,v,S,T;if(i.c.i!=t.i)for(u=new Tg(e),Wl(u,(cn(),uo)),$e(u,(Ge(),oi),i),$e(u,(ot(),Xi),(oa(),uu)),vn(a.c,u),g=new ou,cc(g,u),No(g,(tt(),pn)),m=new ou,cc(m,u),No(m,_n),ua(i,g),l=new xS,Rc(l,i),$e(l,Ls,null),Ia(l,m),ua(l,t),YEt(u,g,m),S=new la(i.b,0);S.b<S.d.gc();)v=(Ln(S.b<S.d.gc()),f(S.d.Xb(S.c=S.b++),72)),T=f(V(v,Rm),276),T==(Sg(),gx)&&(li(v,YR)||$e(v,YR,i),hd(S),Qe(l.b,v))}function v_t(e){vL();var t,i,a,u;if(t=Aa(e),e<j$.length)return j$[t];if(e<=50)return SG((ym(),Wpe),t);if(e<=M0)return oR(SG(UR[1],t),t);if(e>1e6)throw ie(new fY("power of ten too big"));if(e<=Qr)return oR(SG(UR[1],t),t);for(a=SG(UR[1],Qr),u=a,i=Nc(e-Qr),t=Aa(e%Qr);Eu(i,Qr)>0;)u=M3(u,a),i=hh(i,Qr);for(u=M3(u,SG(UR[1],t)),u=oR(u,Qr),i=Nc(e-Qr);Eu(i,Qr)>0;)u=oR(u,Qr),i=hh(i,Qr);return u=oR(u,t),u}function w_t(e){var t,i,a,u,l,g,m,v,S,T;for(v=new ne(e.a);v.a<v.c.c.length;)if(m=f(ee(v),9),m.k==(cn(),bo)&&(u=f(V(m,(Ge(),Tc)),64),u==(tt(),_n)||u==pn))for(a=new Tn(Cn(T1(m).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),t=i.a,t.b!=0&&(S=i.c,S.i==m&&(l=(Ln(t.b!=0),f(t.a.a.c,8)),l.b=uc(ce(se(Zo,1),Ye,8,0,[S.i.n,S.n,S.a])).b),T=i.d,T.i==m&&(g=(Ln(t.b!=0),f(t.c.b.c,8)),g.b=uc(ce(se(Zo,1),Ye,8,0,[T.i.n,T.n,T.a])).b))}function Xj(e,t,i,a){var u,l,g;if(this.j=new et,this.k=new et,this.b=new et,this.c=new et,this.e=new BN,this.i=new fd,this.f=new uae,this.d=new et,this.g=new et,Qe(this.b,e),Qe(this.b,t),this.e.c=h.Math.min(e.a,t.a),this.e.d=h.Math.min(e.b,t.b),this.e.b=h.Math.abs(e.a-t.a),this.e.a=h.Math.abs(e.b-t.b),u=f(V(a,(ot(),Ls)),77),u)for(g=tr(u,0);g.b!=g.d.c;)l=f(Zn(g),8),Sxe(l.a,e.a)&&Mr(this.i,l);i&&Qe(this.j,i),Qe(this.k,a)}function jDn(e,t,i,a){var u,l,g,m,v,S,T;for(m=-1,T=new ne(e);T.a<T.c.c.length;)S=f(ee(T),116),S.g=m--,u=mr(sJ(WX(Jr(new Ft(null,new Bt(S.f,16)),new e1),new Wa)).d),l=mr(sJ(WX(Jr(new Ft(null,new Bt(S.k,16)),new Uc),new ws)).d),g=u,v=l,a||(g=mr(sJ(WX(new Ft(null,new Bt(S.f,16)),new mk)).d),v=mr(sJ(WX(new Ft(null,new Bt(S.k,16)),new CO)).d)),S.d=g,S.a=u,S.i=v,S.b=l,v==0?Wi(i,S,i.c.b,i.c):g==0&&Wi(t,S,t.c.b,t.c)}function TR(e,t){var i,a,u,l,g,m;if(e.k==(cn(),Vc)&&(i=e.k==Vc&&!oP(Jr(f(V(e,(Ge(),Y$)),15).Mc(),new Y(new VM))).zd((Wy(),zR))?(df(),lV):t,$e(e,(Ge(),JR),i),i!=(df(),Nw)))for(a=f(V(e,oi),18),m=ke(Oe(V(a,(ot(),z0)))),g=0,i==Rb?g=e.o.b-h.Math.ceil(m/2):i==lV&&(g=h.Math.ceil(e.o.b-ke(Oe(V(ra(e),OA)))-m)/2,e.o.b-=ke(Oe(V(ra(e),OA))),e.o.b-=m),l=new ne(e.j);l.a<l.c.c.length;)u=f(ee(l),11),u.n.b=g}function y_t(e,t,i){var a,u,l,g,m,v,S,T,A;for(u=!0,g=new ne(e.b);g.a<g.c.c.length;){for(l=f(ee(g),25),S=qo,T=null,v=new ne(l.a);v.a<v.c.c.length;)if(m=f(ee(v),9),A=ke(t.p[m.p])+ke(t.d[m.p])-m.d.d,a=ke(t.p[m.p])+ke(t.d[m.p])+m.o.b+m.d.a,A>S&&a>S)T=m,S=ke(t.p[m.p])+ke(t.d[m.p])+m.o.b+m.d.a;else{u=!1,i.Zg()&&i._g("bk node placement breaks on "+m+" which should have been after "+T);break}if(!u)break}return i.Zg()&&i._g(t+" is feasible: "+u),u}function iOe(e,t,i,a){var u,l,g,m,v,S,T,A,L;if(l=new Tg(e),Wl(l,(cn(),xu)),$e(l,(ot(),Xi),(oa(),uu)),u=0,t){for(g=new ou,$e(g,(Ge(),oi),t),$e(l,oi,t.i),No(g,(tt(),pn)),cc(g,l),L=pm(t.e),S=L,T=0,A=S.length;T<A;++T)v=S[T],ua(v,g);$e(t,ed,l),++u}if(i){for(m=new ou,$e(l,(Ge(),oi),i.i),$e(m,oi,i),No(m,(tt(),_n)),cc(m,l),L=pm(i.g),S=L,T=0,A=S.length;T<A;++T)v=S[T],Ia(v,m);$e(i,ed,l),++u}return $e(l,(Ge(),Ete),rt(u)),vn(a.c,l),l}function $Dn(e){var t,i,a,u,l,g,m,v,S,T,A,L,F;for(i=(S=new kp(e.c.b).a.vc().Jc(),new zy(S));i.a.Ob();)t=(m=f(i.a.Pb(),43),f(m.kd(),149)),u=t.a,u==null&&(u=""),a=Tbn(e.c,u),!a&&u.length==0&&(a=Fxn(e)),a&&!_T(a.c,t,!1)&&Mr(a.c,t);for(g=tr(e.a,0);g.b!=g.d.c;)l=f(Zn(g),473),T=oue(e.c,l.a),F=oue(e.c,l.b),T&&F&&Mr(T.c,new Ya(F,l.c));for(nf(e.a),L=tr(e.b,0);L.b!=L.d.c;)A=f(Zn(L),473),t=kbn(e.c,A.a),v=oue(e.c,A.b),t&&v&&ugn(t,v,A.c);nf(e.b)}function FDn(e,t,i){var a,u,l,g,m,v,S,T,A;for(g=new Iwt,S=ce(se(pr,1),Pr,30,15,[0]),u=-1,l=0,a=0,v=0;v<e.b.c.length;++v)if(T=f(ct(e.b,v),431),T.b>0){if(u<0&&T.a&&(u=v,l=S[0],a=0),u>=0){if(m=T.b,v==u&&(m-=a++,m==0))return 0;if(!ykt(t,S,T,m,g)){v=u-1,S[0]=l;continue}}else if(u=-1,!ykt(t,S,T,0,g))return 0}else{if(u=-1,_a(T.c,0)==32){if(A=S[0],fmt(t,S),S[0]>A)continue}else if(Tyn(t,T.c,S[0])){S[0]+=T.c.length;continue}return 0}return E9n(g,i)?S[0]:0}function BDn(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(T=new jX(new c1(i)),m=Pe(kd,Tb,30,e.f.e.c.length,16,1),Lke(m,m.length),i[t.a]=0,S=new ne(e.f.e);S.a<S.c.c.length;)v=f(ee(S),151),v.a!=t.a&&(i[v.a]=Qr),KN(dL(T,v),NL);for(;T.b.c.length!=0;)for(A=f(Tue(T),151),m[A.a]=!0,l=lct(new Rae(e.b,A),0);l.c;)u=f(wxe(l),288),L=IAn(u,A),!m[L.a]&&(li(u,(dG(),Uee))?g=ke(Oe(V(u,Uee))):g=e.c,a=i[A.a]+g,a<i[L.a]&&(i[L.a]=a,_bt(T,L),KN(dL(T,L),NL)))}function oOe(e){var t,i,a,u,l,g,m,v;i=lt(ut(V(e,(Qh(),eOt)))),l=e.a.c.d,m=e.a.d.d,i?(g=qv(Ho(new Ke(m.a,m.b),l),.5),v=qv($a(e.e),.5),t=Ho(ii(new Ke(l.a,l.b),g),v),h_e(e.d,t)):(u=ke(Oe(V(e.a,uOt))),a=e.d,l.a>=m.a?l.b>=m.b?(a.a=m.a+(l.a-m.a)/2+u,a.b=m.b+(l.b-m.b)/2-u-e.e.b):(a.a=m.a+(l.a-m.a)/2+u,a.b=l.b+(m.b-l.b)/2+u):l.b>=m.b?(a.a=l.a+(m.a-l.a)/2+u,a.b=m.b+(l.b-m.b)/2+u):(a.a=l.a+(m.a-l.a)/2+u,a.b=l.b+(m.b-l.b)/2-u-e.e.b))}function Jj(e){var t,i,a,u,l,g,m,v;if(!e.f){if(v=new d5,m=new d5,t=GF,g=t.a.yc(e,t),g==null){for(l=new Nn(Bs(e));l.e!=l.i.gc();)u=f(Mn(l),29),Qi(v,Jj(u));t.a.Ac(e)!=null,t.a.gc()==0}for(a=(!e.s&&(e.s=new Be(El,e,21,17)),new Nn(e.s));a.e!=a.i.gc();)i=f(Mn(a),178),me(i,101)&&Qn(m,f(i,19));ET(m),e.r=new Jut(e,(f(be(je((Wv(),sn).o),6),19),m.i),m.g),Qi(v,e.r),ET(v),e.f=new _3((f(be(je(sn.o),5),19),v.i),v.g),pd(e).b&=-3}return e.f}function rA(e){var t,i,a,u,l,g,m,v;if(!e.a){if(e.o=null,v=new Dnt(e),t=new i9,i=GF,m=i.a.yc(e,i),m==null){for(g=new Nn(Bs(e));g.e!=g.i.gc();)l=f(Mn(g),29),Qi(v,rA(l));i.a.Ac(e)!=null,i.a.gc()==0}for(u=(!e.s&&(e.s=new Be(El,e,21,17)),new Nn(e.s));u.e!=u.i.gc();)a=f(Mn(u),178),me(a,330)&&Qn(t,f(a,35));ET(t),e.k=new Xut(e,(f(be(je((Wv(),sn).o),7),19),t.i),t.g),Qi(v,e.k),ET(v),e.a=new _3((f(be(je(sn.o),4),19),v.i),v.g),pd(e).b&=-2}return e.a}function dQ(){dQ=Z,x7e=ce(se(Xf,1),_m,30,15,[48,49,50,51,52,53,54,55,56,57,65,66,67,68,69,70]),qDt=new RegExp(`[
\r\f]+`);try{UF=ce(se(vjn,1),Jt,2060,0,[new yz((jSe(),vZ("yyyy-MM-dd'T'HH:mm:ss'.'SSSZ",iH((lY(),lY(),L$))))),new yz(vZ("yyyy-MM-dd'T'HH:mm:ss'.'SSS",iH(L$))),new yz(vZ("yyyy-MM-dd'T'HH:mm:ss",iH(L$))),new yz(vZ("yyyy-MM-dd'T'HH:mm",iH(L$))),new yz(vZ("yyyy-MM-dd",iH(L$)))])}catch(e){if(e=go(e),!me(e,80))throw ie(e)}}function UDn(e,t){var i,a,u,l;if(u=wd(e.d,1)!=0,a=$Ie(e,t),a==0&&lt(ut(V(t.j,(Ge(),kw)))))return 0;!lt(ut(V(t.j,(Ge(),kw))))&&!lt(ut(V(t.j,xA)))||Ae(V(t.j,(ot(),ip)))===Ae((Pp(),Pm))?t.c.kg(t.e,u):u=lt(ut(V(t.j,kw))),TG(e,t,u,!0),lt(ut(V(t.j,xA)))&&$e(t.j,xA,(mn(),!1)),lt(ut(V(t.j,kw)))&&($e(t.j,kw,(mn(),!1)),$e(t.j,xA,!0)),i=$Ie(e,t);do{if(Vxe(e),i==0)return 0;u=!u,l=i,TG(e,t,u,!1),i=$Ie(e,t)}while(l>i);return l}function E_t(e,t){var i,a,u,l;if(u=wd(e.d,1)!=0,a=VZ(e,t),a==0&&lt(ut(V(t.j,(Ge(),kw)))))return 0;!lt(ut(V(t.j,(Ge(),kw))))&&!lt(ut(V(t.j,xA)))||Ae(V(t.j,(ot(),ip)))===Ae((Pp(),Pm))?t.c.kg(t.e,u):u=lt(ut(V(t.j,kw))),TG(e,t,u,!0),lt(ut(V(t.j,xA)))&&$e(t.j,xA,(mn(),!1)),lt(ut(V(t.j,kw)))&&($e(t.j,kw,(mn(),!1)),$e(t.j,xA,!0)),i=VZ(e,t);do{if(Vxe(e),i==0)return 0;u=!u,l=i,TG(e,t,u,!1),i=VZ(e,t)}while(l>i);return l}function zDn(e,t,i){var a,u,l,g,m;if(a=f(V(e,(ot(),lme)),22),i.a>t.a&&(a.Gc((mE(),xF))?e.c.a+=(i.a-t.a)/2:a.Gc(CF)&&(e.c.a+=i.a-t.a)),i.b>t.b&&(a.Gc((mE(),IF))?e.c.b+=(i.b-t.b)/2:a.Gc(AF)&&(e.c.b+=i.b-t.b)),f(V(e,(Ge(),Cu)),22).Gc((us(),Vf))&&(i.a>t.a||i.b>t.b))for(m=new ne(e.a);m.a<m.c.c.length;)g=f(ee(m),9),g.k==(cn(),bo)&&(u=f(V(g,Tc),64),u==(tt(),_n)?g.n.a+=i.a-t.a:u==Bn&&(g.n.b+=i.b-t.b));l=e.d,e.f.a=i.a-l.b-l.c,e.f.b=i.b-l.d-l.a}function HDn(e,t,i){var a,u,l,g,m;if(a=f(V(e,(ot(),lme)),22),i.a>t.a&&(a.Gc((mE(),xF))?e.c.a+=(i.a-t.a)/2:a.Gc(CF)&&(e.c.a+=i.a-t.a)),i.b>t.b&&(a.Gc((mE(),IF))?e.c.b+=(i.b-t.b)/2:a.Gc(AF)&&(e.c.b+=i.b-t.b)),f(V(e,(Ge(),Cu)),22).Gc((us(),Vf))&&(i.a>t.a||i.b>t.b))for(g=new ne(e.a);g.a<g.c.c.length;)l=f(ee(g),9),l.k==(cn(),bo)&&(u=f(V(l,Tc),64),u==(tt(),_n)?l.n.a+=i.a-t.a:u==Bn&&(l.n.b+=i.b-t.b));m=e.d,e.f.a=i.a-m.b-m.c,e.f.b=i.b-m.d-m.a}function GDn(e){var t,i,a,u,l,g,m,v,S,T,A,L;for(t=LEt(e),T=(m=new Ic(t).a.vc().Jc(),new Hi(m));T.a.Ob();){for(S=(u=f(T.a.Pb(),43),f(u.jd(),9)),A=0,L=0,A=S.d.d,L=S.o.b+S.d.a,e.d[S.p]=0,i=S;(l=e.a[i.p])!=S;)a=v3n(i,l),v=0,e.c==(fm(),G0)?v=a.d.n.b+a.d.a.b-a.c.n.b-a.c.a.b:v=a.c.n.b+a.c.a.b-a.d.n.b-a.d.a.b,g=ke(e.d[i.p])+v,e.d[l.p]=g,A=h.Math.max(A,l.d.d-g),L=h.Math.max(L,g+l.o.b+l.d.a),i=l;i=S;do e.d[i.p]=ke(e.d[i.p])+A,i=e.a[i.p];while(i!=S);e.b[S.p]=A+L}}function aOe(e,t,i,a){var u,l,g,m,v,S,T,A,L;return v=Ho(new Ke(i.a,i.b),e),S=v.a*t.b-v.b*t.a,T=t.a*a.b-t.b*a.a,A=(v.a*a.b-v.b*a.a)/T,L=S/T,T==0?S==0?(u=ii(new Ke(i.a,i.b),qv(new Ke(a.a,a.b),.5)),l=Qv(e,u),g=Qv(ii(new Ke(e.a,e.b),t),u),m=h.Math.sqrt(a.a*a.a+a.b*a.b)*.5,l<g&&l<=m?new Ke(e.a,e.b):g<=m?ii(new Ke(e.a,e.b),t):null):null:A>=0&&A<=1&&L>=0&&L<=1?ii(new Ke(e.a,e.b),qv(new Ke(t.a,t.b),A)):null}function Zj(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(l=0,g=e.t,u=0,a=0,v=0,L=0,A=0,i&&(e.n.c.length=0,Qe(e.n,new iJ(e.s,e.t,e.i))),m=0,T=new ne(e.b);T.a<T.c.c.length;)S=f(ee(T),27),l+S.g+(m>0?e.i:0)>t&&v>0&&(l=0,g+=v+e.i,u=h.Math.max(u,L),a+=v+e.i,v=0,L=0,i&&(++A,Qe(e.n,new iJ(e.s,g,e.i))),m=0),L+=S.g+(m>0?e.i:0),v=h.Math.max(v,S.f),i&&C3e(f(ct(e.n,A),208),S),l+=S.g+(m>0?e.i:0),++m;return u=h.Math.max(u,L),a+=v,i&&(e.r=u,e.d=a,O3e(e.j)),new qh(e.s,e.t,u,a)}function qDn(e){var t,i,a,u,l,g,m,v,S,T,A,L;if(m=e.d,A=f(V(e,(Ge(),QR)),15),t=f(V(e,KR),15),!(!A&&!t)){if(l=ke(Oe(kT(e,(ot(),wme)))),g=ke(Oe(kT(e,iDe))),L=0,A){for(S=0,u=A.Jc();u.Ob();)a=f(u.Pb(),9),S=h.Math.max(S,a.o.b),L+=a.o.a;L+=l*(A.gc()-1),m.d+=S+g}if(i=0,t){for(S=0,u=t.Jc();u.Ob();)a=f(u.Pb(),9),S=h.Math.max(S,a.o.b),i+=a.o.a;i+=l*(t.gc()-1),m.a+=S+g}v=h.Math.max(L,i),v>e.o.a&&(T=(v-e.o.a)/2,m.b=h.Math.max(m.b,T),m.c=h.Math.max(m.c,T))}}function sOe(e,t,i,a){var u,l,g,m,v,S,T;if(T=Ju(e.e.zh(),t),u=0,l=f(e.g,122),v=null,is(),f(t,69).uk()){for(m=0;m<e.i;++m)if(g=l[m],T.Zl(g.Ik())){if(ei(g,i)){v=g;break}++u}}else if(i!=null){for(m=0;m<e.i;++m)if(g=l[m],T.Zl(g.Ik())){if(ei(i,g.kd())){v=g;break}++u}}else for(m=0;m<e.i;++m)if(g=l[m],T.Zl(g.Ik())){if(g.kd()==null){v=g;break}++u}return v&&(Ld(e.e)&&(S=t.Gk()?new nue(e.e,4,t,i,null,u,!0):Yv(e,t.qk()?2:1,t,i,t.fk(),-1,!0),a?a.kj(S):a=S),a=CG(e,v,a)),a}function Gde(e,t,i,a,u,l,g){var m,v,S,T,A,L,F,z,W;switch(z=0,W=0,v=u.c,m=u.b,T=i.f,F=i.g,t.g){case 0:z=a.i+a.g+g,e.c?W=MOn(z,l,a,g):W=a.j,L=h.Math.max(v,z+F),S=h.Math.max(m,W+T);break;case 1:W=a.j+a.f+g,e.c?z=DOn(W,l,a,g):z=a.i,L=h.Math.max(v,z+F),S=h.Math.max(m,W+T);break;case 2:z=v+g,W=0,L=v+g+F,S=h.Math.max(m,T);break;case 3:z=0,W=m+g,L=h.Math.max(v,F),S=m+g+T;break;default:throw ie(new dn("IllegalPlacementOption."))}return A=new p3e(e.a,L,S,t,z,W),A}function VDn(e){var t,i,a,u,l,g,m,v;for(l=new sht,Ugn(l,(hR(),xLt)),a=(u=Due(e,Pe(ft,Ye,2,0,6,1)),new Fk(new bu(new Pae(e,u).b)));a.b<a.d.gc();)i=(Ln(a.b<a.d.gc()),dr(a.d.Xb(a.c=a.b++))),g=pOe(qE,i),g&&(t=h1(e,i),t.re()?m=t.re().a:t.oe()?m=""+t.oe().a:t.pe()?m=""+t.pe().a:m=t.Ib(),v=fOe(g,m),v!=null&&((Cl(g.j,(wm(),Ao))||Cl(g.j,en))&&JH(xue(l,Ur),g,v),Cl(g.j,Dm)&&JH(xue(l,vo),g,v),Cl(g.j,FE)&&JH(xue(l,_d),g,v),Cl(g.j,L1)&&JH(xue(l,hc),g,v)));return l}function OG(e,t,i){var a,u,l,g,m,v,S,T;if(u=f(e.g,122),x1(e.e,t))return is(),f(t,69).uk()?new TX(t,e):new Wz(t,e);for(S=Ju(e.e.zh(),t),a=0,m=0;m<e.i;++m){if(l=u[m],g=l.Ik(),S.Zl(g)){if(is(),f(t,69).uk())return l;if(g==(_R(),f4)||g==d4){for(v=new ef(Qs(l.kd()));++m<e.i;)l=u[m],g=l.Ik(),(g==f4||g==d4)&&xr(v,Qs(l.kd()));return B_e(f(t.Ek(),155),v.a)}else return T=l.kd(),T!=null&&i&&me(t,101)&&(f(t,19).Bb&Xa)!=0&&(T=Bj(e,t,m,a,T)),T}++a}return t.fk()}function WDn(e){var t,i,a,u,l,g,m,v;for(u=new ne(e.b);u.a<u.c.c.length;)for(a=f(ee(u),25),g=new ne(rE(a.a));g.a<g.c.c.length;)if(l=f(ee(g),9),Nmt(l)&&(i=f(V(l,(Ge(),u_)),314),!i.g&&i.d))for(t=i,v=i.d;v;)sSt(v.i,v.k,!1,!0),PH(t.a),PH(v.i),PH(v.k),PH(v.b),ua(v.c,t.c.d),ua(t.c,null),jo(t.a,null),jo(v.i,null),jo(v.k,null),jo(v.b,null),m=new PTe(t.i,v.a,t.e,v.j,v.f),m.k=t.k,m.n=t.n,m.b=t.b,m.c=v.c,m.g=t.g,m.d=v.d,$e(t.i,u_,m),$e(v.a,u_,m),v=v.d,t=m}function fQ(e,t,i,a){var u,l,g,m,v,S;if(v=Ju(e.e.zh(),t),l=f(e.g,122),x1(e.e,t)){for(u=0,m=0;m<e.i;++m)if(g=l[m],v.Zl(g.Ik())){if(u==i)return is(),f(t,69).uk()?g:(S=g.kd(),S!=null&&a&&me(t,101)&&(f(t,19).Bb&Xa)!=0&&(S=Bj(e,t,m,u,S)),S);++u}throw ie(new Ru(T$+i+TE+u))}else{for(u=0,m=0;m<e.i;++m){if(g=l[m],v.Zl(g.Ik()))return is(),f(t,69).uk()?g:(S=g.kd(),S!=null&&a&&me(t,101)&&(f(t,19).Bb&Xa)!=0&&(S=Bj(e,t,m,u,S)),S);++u}return t.fk()}}function qde(){qde=Z,qAt=ce(se(pr,1),Pr,30,15,[Ra,1162261467,n$,1220703125,362797056,1977326743,n$,387420489,SQ,214358881,429981696,815730721,1475789056,170859375,268435456,410338673,612220032,893871739,128e7,1801088541,113379904,148035889,191102976,244140625,308915776,387420489,481890304,594823321,729e6,887503681,n$,1291467969,1544804416,1838265625,60466176]),VAt=ce(se(pr,1),Pr,30,15,[-1,-1,31,19,15,13,11,11,10,9,9,8,8,8,8,7,7,7,7,7,7,7,6,6,6,6,6,6,6,6,6,6,6,6,6,6,5])}function Vde(e,t){var i,a,u,l,g,m,v,S,T,A;if(g=e.e,v=t.e,v==0)return e;if(g==0)return t.e==0?t:new Zy(-t.e,t.d,t.a);if(l=e.d,m=t.d,l+m==2)return i=ia(e.a[0],ls),a=ia(t.a[0],ls),g<0&&(i=T0(i)),v<0&&(a=T0(a)),ym(),ZY(hh(i,a),0)?aw(hh(i,a)):X7(aw(T0(hh(i,a))));if(u=l!=m?l>m?1:-1:ECe(e.a,t.a,l),u==-1)A=-v,T=g==v?uue(t.a,m,e.a,l):due(t.a,m,e.a,l);else if(A=g,g==v){if(u==0)return ym(),M$;T=uue(e.a,l,t.a,m)}else T=due(e.a,l,t.a,m);return S=new Zy(A,T.length,T),nj(S),S}function KDn(e,t){var i,a,u,l;if(l=s_t(t),!t.c&&(t.c=new Be(_d,t,9,9)),Zi(new Ft(null,(!t.c&&(t.c=new Be(_d,t,9,9)),new Bt(t.c,16))),new tet(l)),u=f(V(l,(Ge(),Cu)),22),F8n(t,u),u.Gc((us(),Vf)))for(a=new Nn((!t.c&&(t.c=new Be(_d,t,9,9)),t.c));a.e!=a.i.gc();)i=f(Mn(a),124),l9n(e,t,l,i);return f(Fe(t,(ot(),DE)),181).gc()!=0&&X2t(t,l),lt(ut(V(l,QLe)))&&u.Ec(wte),li(l,Mq)&&Xrt(new XCe(ke(Oe(V(l,Mq)))),l),Ae(Fe(t,rx))===Ae((_1(),W0))?D7n(e,t,l):E7n(e,t,l),l}function _u(e,t){var i,a,u,l,g,m,v;if(e==null)return null;if(l=e.length,l==0)return"";for(v=Pe(Xf,_m,30,l,15,1),ga(0,l,e.length),ga(0,l,v.length),odt(e,0,l,v,0),i=null,m=t,u=0,g=0;u<l;u++)a=v[u],Vkt(),a<=32&&(le[a]&2)!=0?m?(!i&&(i=new sh(e)),Jmn(i,u-g++)):(m=t,a!=32&&(!i&&(i=new sh(e)),Hue(i,u-g,u-g+1," "))):m=!1;return m?i?(l=i.a.length,l>0?uh(i.a,0,l-1):""):(ga(0,l-1,e.length),e.substr(0,l-1)):i?i.a:e}function S_t(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(a=new Tn(Cn(lw(t).a.Jc(),new B));Kn(a);)i=f(Dn(a),85),me(be((!i.b&&(i.b=new Qt(Yn,i,4,7)),i.b),0),193)||(v=Us(f(be((!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c),0),84)),Uj(i)||(g=t.i+t.g/2,m=t.j+t.f/2,T=v.i+v.g/2,A=v.j+v.f/2,L=new ha,L.a=T-g,L.b=A-m,l=new Ke(L.a,L.b),uL(l,t.g,t.f),L.a-=l.a,L.b-=l.b,g=T-L.a,m=A-L.b,S=new Ke(L.a,L.b),uL(S,v.g,v.f),L.a-=S.a,L.b-=S.b,T=g+L.a,A=m+L.b,u=rQ(i),B3(u,g),H3(u,m),U3(u,T),z3(u,A),S_t(e,v)))}function MT(e,t){var i,a,u,l,g;if(g=f(t,136),eA(e),eA(g),g.b!=null){if(e.c=!0,e.b==null){e.b=Pe(pr,Pr,30,g.b.length,15,1),au(g.b,0,e.b,0,g.b.length);return}for(l=Pe(pr,Pr,30,e.b.length+g.b.length,15,1),i=0,a=0,u=0;i<e.b.length||a<g.b.length;)i>=e.b.length?(l[u++]=g.b[a++],l[u++]=g.b[a++]):a>=g.b.length?(l[u++]=e.b[i++],l[u++]=e.b[i++]):g.b[a]<e.b[i]||g.b[a]===e.b[i]&&g.b[a+1]<e.b[i+1]?(l[u++]=g.b[a++],l[u++]=g.b[a++]):(l[u++]=e.b[i++],l[u++]=e.b[i++]);e.b=l}}function YDn(e,t){var i,a,u,l,g,m,v,S,T,A;return i=lt(ut(V(e,(Ge(),ex)))),m=lt(ut(V(t,ex))),a=f(V(e,$p),11),v=f(V(t,$p),11),u=f(V(e,yh),11),S=f(V(t,yh),11),T=!!a&&a==v,A=!!u&&u==S,!i&&!m?new K_e(f(ee(new ne(e.j)),11).p==f(ee(new ne(t.j)),11).p,T,A):(l=(!lt(ut(V(e,ex)))||lt(ut(V(e,Pq))))&&(!lt(ut(V(t,ex)))||lt(ut(V(t,Pq)))),g=(!lt(ut(V(e,ex)))||!lt(ut(V(e,Pq))))&&(!lt(ut(V(t,ex)))||!lt(ut(V(t,Pq)))),new K_e(T&&l||A&&g,T,A))}function __t(e){var t,i,a,u,l,g,m,v;for(a=0,i=0,v=new ki,t=0,m=new ne(e.n);m.a<m.c.c.length;)g=f(ee(m),208),g.c.c.length==0?Wi(v,g,v.c.b,v.c):(a=h.Math.max(a,g.d),i+=g.a+(t>0?e.i:0)),++t;for(zCe(e.n,v),e.d=i,e.r=a,e.g=0,e.f=0,e.e=0,e.o=Gi,e.p=Gi,l=new ne(e.b);l.a<l.c.c.length;)u=f(ee(l),27),e.p=h.Math.min(e.p,u.g),e.g=h.Math.max(e.g,u.g),e.f=h.Math.max(e.f,u.f),e.o=h.Math.min(e.o,u.f),e.e+=u.f+e.i;e.a=e.e/e.b.c.length-e.i*((e.b.c.length-1)/e.b.c.length),O3e(e.j)}function k_t(e){var t,i,a,u;return(e.Db&64)!=0?mle(e):(t=new ef(N4e),a=e.k,a?xr(xr((t.a+=' "',t),a),'"'):(!e.n&&(e.n=new Be(hc,e,1,7)),e.n.i>0&&(u=(!e.n&&(e.n=new Be(hc,e,1,7)),f(be(e.n,0),153)).a,!u||xr(xr((t.a+=' "',t),u),'"'))),i=(!e.b&&(e.b=new Qt(Yn,e,4,7)),!(e.b.i<=1&&(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c.i<=1))),i?t.a+=" [":t.a+=" ",xr(t,ZSe(new Aae,new Nn(e.b))),i&&(t.a+="]"),t.a+=Xfe,i&&(t.a+="["),xr(t,ZSe(new Aae,new Nn(e.c))),i&&(t.a+="]"),t.a)}function XDn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;for(Le=e.c,He=t.c,i=Oc(Le.a,e,0),a=Oc(He.a,t,0),ve=f($S(e,(cs(),td)).Jc().Pb(),11),Ct=f($S(e,Uu).Jc().Pb(),11),Ee=f($S(t,td).Jc().Pb(),11),Kt=f($S(t,Uu).Jc().Pb(),11),ue=pm(ve.e),nt=pm(Ct.g),pe=pm(Ee.e),xt=pm(Kt.g),sw(e,a,He),g=pe,T=0,z=g.length;T<z;++T)u=g[T],ua(u,ve);for(m=xt,A=0,W=m.length;A<W;++A)u=m[A],Ia(u,Ct);for(sw(t,i,Le),v=ue,L=0,J=v.length;L<J;++L)u=v[L],ua(u,Ee);for(l=nt,S=0,F=l.length;S<F;++S)u=l[S],Ia(u,Kt)}function JDn(e){var t,i,a,u,l,g,m;for(g=f(Fe(e,(R3(),MA)),27),a=new Nn((!g.e&&(g.e=new Qt(vo,g,7,4)),g.e));a.e!=a.i.gc();)i=f(Mn(a),85),m=new Ke(f(be((!i.a&&(i.a=new Be(Ii,i,6,6)),i.a),0),168).j,f(be((!i.a&&(i.a=new Be(Ii,i,6,6)),i.a),0),168).k),l=new Ke(f(be((!i.a&&(i.a=new Be(Ii,i,6,6)),i.a),0),168).b,f(be((!i.a&&(i.a=new Be(Ii,i,6,6)),i.a),0),168).c),u=new Ke(l.a-m.a,l.b-m.b),t=h.Math.atan2(u.b,u.a),f(be((!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c),0),84).of((uw(),fbe),t)}function ZDn(e,t){var i,a,u,l,g,m,v,S,T;for(t.Sg("Interactive Node Reorderer",1),T=(!e.a&&(e.a=new Be(Ur,e,10,11)),e.a),m=new et,u=new Nn(T);u.e!=u.i.gc();)i=f(Mn(u),27),mm(i,(Ag(),Yq))&&vn(m.c,i);for(l=new ne(m);l.a<l.c.c.length;)i=f(ee(l),27),IJ(T,i);for(Ht(),na(m,new jC),g=new ne(m);g.a<g.c.c.length;)i=f(ee(g),27),S=f(Fe(i,(Ag(),Yq)),17).a,S=h.Math.min(S,T.i),xj(T,S,i);for(v=0,a=new Nn(T);a.e!=a.i.gc();)i=f(Mn(a),27),pi(i,(Ag(),i8e),rt(v)),++v;t.Tg()}function QDn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(m=Pe(pr,Pr,30,t.b.c.length,15,1),S=Pe(wge,Ve,247,t.b.c.length,0,1),v=Pe(Ib,j0,9,t.b.c.length,0,1),A=e.a,L=0,F=A.length;L<F;++L){for(T=A[L],W=0,g=new ne(T.e);g.a<g.c.c.length;)u=f(ee(g),9),a=$Se(u.c),++m[a],z=ke(Oe(V(t,(ot(),op)))),m[a]>0&&v[a]&&(z=A3(e.b,v[a],u)),W=h.Math.max(W,u.c.c.b+z);for(l=new ne(T.e);l.a<l.c.c.length;)u=f(ee(l),9),u.n.b=W+u.d.d,i=u.c,i.c.b=W+u.d.d+u.o.b+u.d.a,S[Oc(i.b.b,i,0)]=u.k,v[Oc(i.b.b,i,0)]=u}}function cOe(e,t,i){var a,u,l,g,m,v,S,T;return h.Math.abs(t.s-t.c)<Tm||h.Math.abs(i.s-i.c)<Tm?0:(a=P2t(e,t.j,i.e),u=P2t(e,i.j,t.e),l=a==-1||u==-1,g=0,l?(a==-1&&(new nE((Ip(),Cw),i,t,1),++g),u==-1&&(new nE((Ip(),Cw),t,i,1),++g)):(m=W3(t.j,i.s,i.c),m+=W3(i.e,t.s,t.c),v=W3(i.j,t.s,t.c),v+=W3(t.e,i.s,i.c),S=a+16*m,T=u+16*v,S<T?new nE((Ip(),dx),t,i,T-S):S>T?new nE((Ip(),dx),i,t,S-T):S>0&&T>0&&(new nE((Ip(),dx),t,i,0),new nE(dx,i,t,0))),g)}function eMn(e,t,i){var a,u,l;for(e.a=new et,l=tr(t.b,0);l.b!=l.d.c;){for(u=f(Zn(l),39);f(V(u,(Ec(),Lm)),17).a>e.a.c.length-1;)Qe(e.a,new Ya($R,LRe));a=f(V(u,Lm),17).a,i==(ko(),Ds)||i==Gs?(u.e.a<ke(Oe(f(ct(e.a,a),47).a))&&$k(f(ct(e.a,a),47),u.e.a),u.e.a+u.f.a>ke(Oe(f(ct(e.a,a),47).b))&&w(f(ct(e.a,a),47),u.e.a+u.f.a)):(u.e.b<ke(Oe(f(ct(e.a,a),47).a))&&$k(f(ct(e.a,a),47),u.e.b),u.e.b+u.f.b>ke(Oe(f(ct(e.a,a),47).b))&&w(f(ct(e.a,a),47),u.e.b+u.f.b))}}function T_t(e,t,i,a){var u,l,g,m,v,S,T;if(l=dZ(a),m=lt(ut(V(a,(ot(),WLe)))),(m||lt(ut(V(e,Pte))))&&!T3(f(V(e,Xi),100)))u=mR(l),v=WIe(e,i,i==(cs(),Uu)?u:iG(u));else switch(v=new ou,cc(v,e),t?(T=v.n,T.a=t.a-e.n.a,T.b=t.b-e.n.b,Cyt(T,0,0,e.o.a,e.o.b),No(v,VSt(v,l))):(u=mR(l),No(v,i==(cs(),Uu)?u:iG(u))),g=f(V(a,(Ge(),Cu)),22),S=v.j,l.g){case 2:case 1:(S==(tt(),hn)||S==Bn)&&g.Ec((us(),_A));break;case 4:case 3:(S==(tt(),_n)||S==pn)&&g.Ec((us(),_A))}return v}function x_t(e,t){var i,a,u,l,g,m;for(g=new yT(new Vi(e.f.b).a);g.b;){if(l=G3(g),u=f(l.jd(),586),t==1){if(u.yf()!=(ko(),Wf)&&u.yf()!=jg)continue}else if(u.yf()!=(ko(),Ds)&&u.yf()!=Gs)continue;switch(a=f(f(l.kd(),47).b,82),m=f(f(l.kd(),47).a,194),i=m.c,u.yf().g){case 2:a.g.c=e.e.a,a.g.b=h.Math.max(1,a.g.b+i);break;case 1:a.g.c=a.g.c+i,a.g.b=h.Math.max(1,a.g.b-i);break;case 4:a.g.d=e.e.b,a.g.a=h.Math.max(1,a.g.a+i);break;case 3:a.g.d=a.g.d+i,a.g.a=h.Math.max(1,a.g.a-i)}}}function tMn(e,t){var i,a,u,l,g,m,v,S,T,A;for(t.Sg("Simple node placement",1),A=f(V(e,(Ge(),CA)),313),m=0,l=new ne(e.b);l.a<l.c.c.length;){for(a=f(ee(l),25),g=a.c,g.b=0,i=null,S=new ne(a.a);S.a<S.c.c.length;)v=f(ee(S),9),i&&(g.b+=l3e(v,i,A.c)),g.b+=v.d.d+v.o.b+v.d.a,i=v;m=h.Math.max(m,g.b)}for(u=new ne(e.b);u.a<u.c.c.length;)for(a=f(ee(u),25),g=a.c,T=(m-g.b)/2,i=null,S=new ne(a.a);S.a<S.c.c.length;)v=f(ee(S),9),i&&(T+=l3e(v,i,A.c)),T+=v.d.d,v.n.b=T,T+=v.o.b+v.d.a,i=v;t.Tg()}function C_t(e){Wk(e,new zS(qk(zk(Gk(Hk(new ug,QS),"ELK SPOrE Compaction"),"ShrinkTree is a compaction algorithm that maintains the topology of a layout. The relocation of diagram elements is based on contracting a spanning tree."),new Tv))),Je(e,QS,Whe,st(gne)),Je(e,QS,m4e,st(jbe)),Je(e,QS,b4e,st(Mbe)),Je(e,QS,Khe,st(P8e)),Je(e,QS,Yhe,st(Dbe)),Je(e,QS,zT,R8e),Je(e,QS,UT,8),Je(e,QS,Xhe,st(fLt)),Je(e,QS,v4e,st(O8e)),Je(e,QS,w4e,st(N8e)),Je(e,QS,aq,(mn(),!1))}function nMn(e,t){var i,a,u,l;for(skn(t.b.j),Zi(Lu(new Ft(null,new Bt(t.d,16)),new y8),new Sy),l=new ne(t.d);l.a<l.c.c.length;){switch(u=f(ee(l),105),u.e.g){case 0:i=f(ct(u.j,0),112).d.j,Dv(u,f(Dd(XN(f(si(u.k,i),15).Mc(),wA)),112)),jk(u,f(Dd(oT(f(si(u.k,i),15).Mc(),wA)),112));break;case 1:a=P3e(u),Dv(u,f(Dd(XN(f(si(u.k,a[0]),15).Mc(),wA)),112)),jk(u,f(Dd(oT(f(si(u.k,a[1]),15).Mc(),wA)),112));break;case 2:gRn(e,u);break;case 3:m5n(u);break;case 4:E5n(e,u)}ckn(u)}e.a=null}function Wde(e,t,i){var a,u,l,g,m,v,S,T;return a=e.a.o==(vg(),Mg)?Gi:qo,m=QSt(e,new cat(t,i)),!m.a&&m.c?(Mr(e.d,m),a):m.a?(u=m.a.c,v=m.a.d,i?(S=e.a.c==(fm(),y_)?v:u,l=e.a.c==y_?u:v,g=e.a.g[l.i.p],T=ke(e.a.p[g.p])+ke(e.a.d[l.i.p])+l.n.b+l.a.b-ke(e.a.d[S.i.p])-S.n.b-S.a.b):(S=e.a.c==(fm(),G0)?v:u,l=e.a.c==G0?u:v,T=ke(e.a.p[e.a.g[l.i.p].p])+ke(e.a.d[l.i.p])+l.n.b+l.a.b-ke(e.a.d[S.i.p])-S.n.b-S.a.b),e.a.n[e.a.g[u.i.p].p]=(mn(),!0),e.a.n[e.a.g[v.i.p].p]=!0,T):a}function rMn(e,t,i,a){var u,l,g,m,v,S,T,A;if(a.gc()==0)return!1;if(v=(is(),f(t,69).uk()),g=v?a:new OS(a.gc()),x1(e.e,t)){if(t.Pi())for(T=a.Jc();T.Ob();)S=T.Pb(),lQ(e,t,S,me(t,101)&&(f(t,19).Bb&Xa)!=0)||(l=mb(t,S),g.Ec(l));else if(!v)for(T=a.Jc();T.Ob();)S=T.Pb(),l=mb(t,S),g.Ec(l)}else{for(A=Ju(e.e.zh(),t),u=f(e.g,122),m=0;m<e.i;++m)if(l=u[m],A.Zl(l.Ik()))throw ie(new dn(gq));if(a.gc()>1)throw ie(new dn(gq));v||(l=mb(t,a.Jc().Pb()),g.Ec(l))}return cCe(e,CAe(e,t,i),g)}function hQ(e,t,i){var a,u,l,g,m,v,S,T;if(x1(e.e,t))v=(is(),f(t,69).uk()?new TX(t,e):new Wz(t,e)),KZ(v.c,v.b),M7(v,f(i,16));else{for(T=Ju(e.e.zh(),t),a=f(e.g,122),g=0;g<e.i;++g)if(u=a[g],l=u.Ik(),T.Zl(l)){if(l==(_R(),f4)||l==d4){for(S=k3e(e,t,i),m=g,S?PT(e,g):++g;g<e.i;)u=a[g],l=u.Ik(),l==f4||l==d4?PT(e,g):++g;S||f(Y3(e,m,mb(t,i)),74)}else k3e(e,t,i)?PT(e,g):f(Y3(e,g,(is(),f(t,69).uk()?f(i,74):mb(t,i))),74);return}k3e(e,t,i)||Qn(e,(is(),f(t,69).uk()?f(i,74):mb(t,i)))}}function A_t(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(A=0;A<t.length;A++){for(m=e.Jc();m.Ob();)l=f(m.Pb(),229),l.fg(A,t);for(L=0;L<t[A].length;L++){for(v=e.Jc();v.Ob();)l=f(v.Pb(),229),l.gg(A,L,t);for(W=t[A][L].j,F=0;F<W.c.length;F++){for(S=e.Jc();S.Ob();)l=f(S.Pb(),229),l.hg(A,L,F,t);for(z=(Gt(F,W.c.length),f(W.c[F],11)),i=0,u=new yg(z.b);sc(u.a)||sc(u.b);)for(a=f(sc(u.a)?ee(u.a):ee(u.b),18),T=e.Jc();T.Ob();)l=f(T.Pb(),229),l.eg(A,L,F,i++,a,t)}}}for(g=e.Jc();g.Ob();)l=f(g.Pb(),229),l.dg()}function iMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;if(z=f(bn(e.d,t),147),!z)throw ie(new am("Edge did not exist in input."));return Lde(e,t),A=oL(z),g=Az((!t.a&&(t.a=new Be(Ii,t,6,6)),t.a)),v=!g,v&&(W=new oh,a=new Wdt(e,A,W,t),Agn((!t.a&&(t.a=new Be(Ii,t,6,6)),t.a),a),Wh(z,hpe,null),Wh(z,hpe,W)),l=mm(t,(Rr(),zA)),l&&(L=f(Fe(t,zA),77),m=!L||_dt(L),S=!m,S&&(F=new oh,u=new jut(e,t,F),Oa(L,u),Wh(z,"junctionPoints",F))),i=x_e(e,f(bn(e.e,t),27)),T=i==(XP(),_ne),T&&wP(z,"container",lT(t).k),null}function oMn(e,t){var i,a,u,l,g,m,v;for(e.b=ke(Oe(V(t,(ot(),IA)))),e.c=ke(Oe(V(t,jE))),e.d=f(V(t,gme),346),e.a=f(V(t,Cte),280),gAn(t),m=f($d(Jr(Jr(Zs(Zs(new Ft(null,new Bt(t.b,16)),new lC),new my),new v6),new iO),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),u=m.Jc();u.Ob();)i=f(u.Pb(),18),g=f(V(i,(Ge(),NE)),15),g.Ic(new fet(e)),$e(i,NE,null);for(a=m.Jc();a.Ob();)i=f(a.Pb(),18),v=f(V(i,(Ge(),VPe)),18),l=f(V(i,AA),15),v7n(e,l,v),$e(i,AA,null)}function Kde(e,t){var i,a,u,l,g,m,v;if(e.a){if(m=e.a.ve(),v=null,m!=null?t.a+=""+m:(g=e.a.jk(),g!=null&&(l=cm(g,gl(91)),l!=-1?(v=(gn(l,g.length+1),g.substr(l)),t.a+=""+uh(g==null?bl:(an(g),g),0,l)):t.a+=""+g)),e.d&&e.d.i!=0){for(u=!0,t.a+="<",a=new Nn(e.d);a.e!=a.i.gc();)i=f(Mn(a),87),u?u=!1:t.a+=Fu,Kde(i,t);t.a+=">"}v!=null&&(t.a+=""+v)}else e.e?(m=e.e.zb,m!=null&&(t.a+=""+m)):(t.a+="?",e.b?(t.a+=" super ",Kde(e.b,t)):e.f&&(t.a+=" extends ",Kde(e.f,t)))}function aMn(e){e.b=null,e.a=null,e.o=null,e.q=null,e.v=null,e.w=null,e.B=null,e.p=null,e.Q=null,e.R=null,e.S=null,e.T=null,e.U=null,e.V=null,e.W=null,e.bb=null,e.eb=null,e.ab=null,e.H=null,e.db=null,e.c=null,e.d=null,e.f=null,e.n=null,e.r=null,e.s=null,e.u=null,e.G=null,e.J=null,e.e=null,e.j=null,e.i=null,e.g=null,e.k=null,e.t=null,e.F=null,e.I=null,e.L=null,e.M=null,e.O=null,e.P=null,e.$=null,e.N=null,e.Z=null,e.cb=null,e.K=null,e.D=null,e.A=null,e.C=null,e._=null,e.fb=null,e.X=null,e.Y=null,e.gb=!1,e.hb=!1}function sMn(e){var t,i,a,u;if(a=afe((!e.c&&(e.c=yH(Nc(e.f))),e.c),0),e.e==0||e.a==0&&e.f!=-1&&e.e<0)return a;if(t=Wxe(e)<0?1:0,i=e.e,u=(a.length+1+h.Math.abs(Aa(e.e)),new RN),t==1&&(u.a+="-"),e.e>0)if(i-=a.length-t,i>=0){for(u.a+="0.";i>IE.length;i-=IE.length)plt(u,IE);Zct(u,IE,Aa(i)),xr(u,(gn(t,a.length+1),a.substr(t)))}else i=t-i,xr(u,uh(a,t,Aa(i))),u.a+=".",xr(u,Gke(a,Aa(i)));else{for(xr(u,(gn(t,a.length+1),a.substr(t)));i<-IE.length;i+=IE.length)plt(u,IE);Zct(u,IE,Aa(-i))}return u.a}function Yde(e){var t,i,a,u,l,g,m,v,S;return!(e.k!=(cn(),Yi)||e.j.c.length<=1||(l=f(V(e,(ot(),Xi)),100),l==(oa(),uu))||(u=(xT(),(e.q?e.q:(Ht(),Ht(),Cb))._b(g_)?a=f(V(e,g_),203):a=f(V(ra(e),nF),203),a),u==Gte)||!(u==PA||u==RA)&&(g=ke(Oe(kT(e,rF))),t=f(V(e,$q),139),!t&&(t=new L_e(g,g,g,g)),S=lc(e,(tt(),pn)),v=t.d+t.a+(S.gc()-1)*g,v>e.o.b||(i=lc(e,_n),m=t.d+t.a+(i.gc()-1)*g,m>e.o.b)))}function cMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;t.Sg("Orthogonal edge routing",1),S=ke(Oe(V(e,(ot(),NA)))),i=ke(Oe(V(e,IA))),a=ke(Oe(V(e,jE))),L=new vce(0,i),J=0,g=new la(e.b,0),m=null,T=null,v=null,A=null;do T=g.b<g.d.gc()?(Ln(g.b<g.d.gc()),f(g.d.Xb(g.c=g.b++),25)):null,A=T?T.a:null,m&&(lOe(m,J),J+=m.c.a),W=m?J+a:J,z=gOe(L,e,v,A,W),u=!m||JY(v,(iQ(),Hq)),l=!T||JY(A,(iQ(),Hq)),z>0?(F=(z-1)*i,m&&(F+=a),T&&(F+=a),F<S&&!u&&!l&&(F=S),J+=F):!u&&!l&&(J+=S),m=T,v=A;while(T);e.f.a=J,t.Tg()}function pQ(e,t){var i,a,u,l,g,m,v,S,T,A;if(T=null,e.d&&(T=f(wu(e.d,t),141)),!T){if(l=e.a.ri(),A=l.i,!e.d||f7(e.d)!=A){for(v=new Wn,e.d&&mj(v,e.d),S=v.f.c+v.i.c,m=S;m<A;++m)a=f(be(l,m),141),u=I0(e.e,a).ve(),i=f(u==null?ml(v.f,null,a):MS(v.i,u,a),141),i&&i!=a&&(u==null?ml(v.f,null,i):MS(v.i,u,i));if(v.f.c+v.i.c!=A)for(g=0;g<S;++g)a=f(be(l,g),141),u=I0(e.e,a).ve(),i=f(u==null?ml(v.f,null,a):MS(v.i,u,a),141),i&&i!=a&&(u==null?ml(v.f,null,i):MS(v.i,u,i));e.d=v}T=f(wu(e.d,t),141)}return T}function Xde(e,t,i,a,u,l,g){var m,v,S,T,A,L,F;return A=lt(ut(V(t,(ot(),KLe)))),L=null,l==(cs(),td)&&a.c.i==i?L=a.c:l==Uu&&a.d.i==i&&(L=a.d),S=g,!S||!A||L?(T=(tt(),fc),L?T=L.j:T3(f(V(i,Xi),100))&&(T=l==td?pn:_n),v=uMn(e,t,i,l,T,a),m=cue((ra(i),a)),l==td?(Ia(m,f(ct(v.j,0),11)),ua(m,u)):(Ia(m,u),ua(m,f(ct(v.j,0),11))),S=new x0t(a,m,v,f(V(v,(Ge(),oi)),11),l,!L)):(Qe(S.e,a),F=h.Math.max(ke(Oe(V(S.d,z0))),ke(Oe(V(a,z0)))),$e(S.d,z0,F)),jt(e.a,a,new vX(S.d,t,l)),S}function Jde(){Jde=Z;var e;M7e=new Trt,eMt=Pe(ft,Ye,2,0,6,1),ZDt=hm(fR(33,58),fR(1,26)),QDt=hm(fR(97,122),fR(65,90)),N7e=fR(48,57),XDt=hm(ZDt,0),JDt=hm(QDt,N7e),R7e=hm(hm(0,fR(1,6)),fR(33,38)),P7e=hm(hm(N7e,fR(65,70)),fR(97,102)),tMt=hm(XDt,kZ("-_.!~*'()")),nMt=hm(JDt,bZ("-_.!~*'()")),kZ(T3t),bZ(T3t),hm(tMt,kZ(";:@&=+$,")),hm(nMt,bZ(";:@&=+$,")),L7e=kZ(":/?#"),D7e=bZ(":/?#"),zF=kZ("/?#"),HF=bZ("/?#"),e=new Mo,e.a.yc("jar",e),e.a.yc("zip",e),e.a.yc("archive",e),Bne=(Ht(),new tP(e))}function uMn(e,t,i,a,u,l){var g,m,v,S,T,A;return g=null,S=a==(cs(),td)?l.c:l.d,v=dZ(t),S.i==i?(g=f(bn(e.b,S),9),g||(g=PG(S,f(V(i,(ot(),Xi)),100),u,vLn(S),null,S.n,S.o,v,t),$e(g,(Ge(),oi),S),qr(e.b,S,g))):(g=PG((T=new k2,A=ke(Oe(V(t,(ot(),op))))/2,JH(T,ox,A),T),f(V(i,Xi),100),u,a==td?-1:1,null,new ha,new Ke(0,0),v,t),m=bOn(g,i,a),$e(g,(Ge(),oi),m),qr(e.b,m,g)),f(V(t,(Ge(),Cu)),22).Ec((us(),Vf)),T3(f(V(t,(ot(),Xi)),100))?$e(t,Xi,(oa(),MD)):$e(t,Xi,(oa(),Fg)),g}function SL(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;m=0,z=0,v=rJ(e.g,e.g.length),l=e.e,g=e.j,a=e.b,u=e.c;do{for(F=0,T=new ne(e.q);T.a<T.c.c.length;)S=f(ee(T),9),L=Y_t(e,S),i=!0,(e.r==(gh(),Fq)||e.r==Bq)&&(i=lt(ut(L.b))),f(L.a,17).a<0&&i?(++F,v=rJ(e.g,e.g.length),e.e=e.e+f(L.a,17).a,z+=l-e.e,l=e.e+f(L.a,17).a,g=e.j,a=rE(e.b),u=rE(e.c)):(e.g=rJ(v,v.length),e.e=l,e.b=(or(a),a?new Kl(a):bP(new ne(a))),e.c=(or(u),u?new Kl(u):bP(new ne(u))),e.j=g);++m,A=F!=0&&lt(ut(t.Kb(new Ya(rt(z),rt(m)))))}while(A)}function lMn(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct;return g=e.f,L=t.f,m=g==(Hj(),n4)||g==_F,F=L==n4||L==_F,v=g==$A||g==kF,z=L==$A||L==kF,S=g==$A||g==n4,W=L==$A||L==n4,m&&F?e.f==_F?e:t:v&&z?e.f==kF?e:t:S&&W?(g==$A?(A=e,T=t):(A=t,T=e),l=(J=i.j+i.f,ue=A.e+a.f,pe=h.Math.max(J,ue),ve=pe-h.Math.min(i.j,A.e),Ee=A.d+a.g-i.i,Ee*ve),u=(Le=i.i+i.g,He=T.d+a.g,nt=h.Math.max(Le,He),xt=nt-h.Math.min(i.i,T.d),Ct=T.e+a.f-i.j,xt*Ct),l<=u?e.f==$A?e:t:e.f==n4?e:t):e}function I_t(e,t){var i,a,u,l,g,m,v,S,T,A;if($e(t,(_i(),xD),0),v=f(V(t,Qte),39),t.d.b==0)v?(T=ke(Oe(V(v,q0)))+e.b+HCe(e,v,t),$e(t,q0,T)):$e(t,q0,0);else{for(a=(l=tr(new u1(t).a.d,0),new m3(l));kz(a.a);)i=f(Zn(a.a),65).c,I_t(e,i);m=f(oX((g=tr(new u1(t).a.d,0),new m3(g))),39),A=f(jmn((u=tr(new u1(t).a.d,0),new m3(u))),39),S=(ke(Oe(V(A,q0)))+ke(Oe(V(m,q0))))/2,v?(T=ke(Oe(V(v,q0)))+e.b+HCe(e,v,t),$e(t,q0,T),$e(t,xD,ke(Oe(V(t,q0)))-S),k8n(e,t)):$e(t,q0,S)}}function dMn(e){var t,i,a,u,l,g,m,v,S,T,A,L,F;if(S=new et,!li(e,(Ge(),Jge)))return S;for(a=f(V(e,Jge),15).Jc();a.Ob();)t=f(a.Pb(),9),RMn(t,e),vn(S.c,t);for(l=new ne(e.b);l.a<l.c.c.length;)for(u=f(ee(l),25),m=new ne(u.a);m.a<m.c.c.length;)g=f(ee(m),9),g.k==(cn(),bo)&&(v=f(V(g,_te),9),v&&(T=new ou,cc(T,g),A=f(V(g,Tc),64),No(T,A),L=f(ct(v.j,0),11),F=new xS,Ia(F,T),ua(F,L)));for(i=new ne(S);i.a<i.c.c.length;)t=f(ee(i),9),jo(t,f(ct(e.b,e.b.c.length-1),25));return S}function fMn(e){var t,i,a,u,l,g,m,v,S,T,A;for(T=e.e.a.c.length,g=new ne(e.e.a);g.a<g.c.c.length;)l=f(ee(g),123),l.j=!1;for(e.i=Pe(pr,Pr,30,T,15,1),e.g=Pe(pr,Pr,30,T,15,1),e.n=new et,u=0,A=new et,v=new ne(e.e.a);v.a<v.c.c.length;)m=f(ee(v),123),m.d=u++,m.b.a.c.length==0&&Qe(e.n,m),$o(A,m.g);for(t=0,a=new ne(A);a.a<a.c.c.length;)i=f(ee(a),216),i.c=t++,i.f=!1;S=A.c.length,e.b==null||e.b.length<S?(e.b=Pe(sa,_s,30,S,15,1),e.c=Pe(kd,Tb,30,S,16,1)):c7(e.c),e.d=A,e.p=new Nse(mT(e.d.c.length)),e.j=1}function hMn(e,t){var i,a,u,l,g,m,v,S,T;if(!(t.e.c.length<=1)){for(e.f=t,e.d=f(V(e.f,(dG(),o5e)),382),e.g=f(V(e.f,u5e),17).a,e.e=ke(Oe(V(e.f,a5e))),e.c=ke(Oe(V(e.f,Uee))),edt(e.b),u=new ne(e.f.c);u.a<u.c.c.length;)a=f(ee(u),288),zIe(e.b,a.c,a,null),zIe(e.b,a.d,a,null);for(m=e.f.e.c.length,e.a=iT(sa,[Ye,_s],[108,30],15,[m,m],2),S=new ne(e.f.e);S.a<S.c.c.length;)v=f(ee(S),151),BDn(e,v,e.a[v.a]);for(e.i=iT(sa,[Ye,_s],[108,30],15,[m,m],2),l=0;l<m;++l)for(g=0;g<m;++g)i=e.a[l][g],T=1/(i*i),e.i[l][g]=T}}function O_t(e,t){var i,a,u,l,g,m,v,S,T,A,L,F;if(m=t.$g(),m||t.Sg(zTt,1),i=f(V(e,(Ge(),nme)),15),g=1/i.gc(),t.Zg())for(t._g("ELK Layered uses the following "+i.gc()+" modules:"),F=0,L=i.Jc();L.Ob();)T=f(L.Pb(),44),a=(F<10?"0":"")+F++,t._g("   Slot "+a+": "+Vy(Fd(T)));for(A=i.Jc();A.Ob();){if(T=f(A.Pb(),44),t.Yg())return;T.If(e,t.bh(g))}for(l=new ne(e.b);l.a<l.c.c.length;)u=f(ee(l),25),$o(e.a,u.a),u.a.c.length=0;for(S=new ne(e.a);S.a<S.c.c.length;)v=f(ee(S),9),jo(v,null);e.b.c.length=0,m||t.Tg()}function N_t(e,t,i){var a,u,l,g,m,v,S,T,A;for(A=new et,T=new _Te(0,i),l=0,jJ(T,new ele(0,0,T,i)),u=0,S=new Nn(e);S.e!=S.i.gc();)v=f(Mn(S),27),a=f(ct(T.a,T.a.c.length-1),172),m=u+v.g+(f(ct(T.a,0),172).b.c.length==0?0:i),(m>t||lt(ut(Fe(v,(Ag(),Xq)))))&&(u=0,l+=T.b+i,vn(A.c,T),T=new _Te(l,i),a=new ele(0,T.f,T,i),jJ(T,a),u=0),a.b.c.length==0||!lt(ut(Fe(so(v),(Ag(),kbe))))&&(v.f>=a.o&&v.f<=a.f||a.a*.5<=v.f&&a.a*1.5>=v.f)?KCe(a,v):(g=new ele(a.s+a.r+i,T.f,T,i),jJ(T,g),KCe(g,v)),u=v.i+v.g;return vn(A.c,T),A}function Qj(e){var t,i,a,u;if(!(e.b==null||e.b.length<=2)&&!e.a){for(t=0,u=0;u<e.b.length;){for(t!=u?(e.b[t]=e.b[u++],e.b[t+1]=e.b[u++]):u+=2,i=e.b[t+1];u<e.b.length&&!(i+1<e.b[u]);)if(i+1==e.b[u])e.b[t+1]=e.b[u+1],i=e.b[t+1],u+=2;else if(i>=e.b[u+1])u+=2;else if(i<e.b[u+1])e.b[t+1]=e.b[u+1],i=e.b[t+1],u+=2;else throw ie(new oc("Token#compactRanges(): Internel Error: ["+e.b[t]+","+e.b[t+1]+"] ["+e.b[u]+","+e.b[u+1]+"]"));t+=2}t!=e.b.length&&(a=Pe(pr,Pr,30,t,15,1),au(e.b,0,a,0,t),e.b=a),e.a=!0}}function pMn(e,t){var i,a,u,l,g,m,v;for(g=YN(e.a).Jc();g.Ob();){if(l=f(g.Pb(),18),l.b.c.length>0)for(a=new Kl(f(si(e.a,l),22)),Ht(),na(a,new ZEe(t)),u=new la(l.b,0);u.b<u.d.gc();){switch(i=(Ln(u.b<u.d.gc()),f(u.d.Xb(u.c=u.b++),72)),m=-1,f(V(i,(ot(),Rm)),276).g){case 1:m=a.c.length-1;break;case 0:m=QIn(a);break;case 2:m=0}m!=-1&&(v=(Gt(m,a.c.length),f(a.c[m],248)),Qe(v.b.b,i),f(V(ra(v.b.c.i),(Ge(),Cu)),22).Ec((us(),fD)),f(V(ra(v.b.c.i),Cu),22).Ec(dD),hd(u),$e(i,HPe,l))}Ia(l,null),ua(l,null)}}function gMn(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(T=new et,L=new Mo,g=t.b,u=0;u<g.c.length;u++){for(S=(Gt(u,g.c.length),f(g.c[u],25)).a,T.c.length=0,l=0;l<S.c.length;l++)m=e.a[u][l],m.p=l,m.k==(cn(),xu)&&vn(T.c,m),af(f(ct(t.b,u),25).a,l,m),m.j.c.length=0,$o(m.j,f(f(ct(e.b,u),15).Xb(l),16)),VN(f(V(m,(ot(),Xi)),100))||$e(m,Xi,(oa(),HE));for(a=new ne(T);a.a<a.c.c.length;)i=f(ee(a),9),A=QPn(i),L.a.yc(A,L),L.a.yc(i,L)}for(v=L.a.ec().Jc();v.Ob();)m=f(v.Pb(),9),Ht(),na(m.j,(WP(),oPe)),m.i=!0,$Ae(m)}function mMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F;if(t.Sg("Edge splitting",1),e.b.c.length<=2){t.Tg();return}for(l=new la(e.b,0),g=(Ln(l.b<l.d.gc()),f(l.d.Xb(l.c=l.b++),25));l.b<l.d.gc();)for(u=g,g=(Ln(l.b<l.d.gc()),f(l.d.Xb(l.c=l.b++),25)),v=new ne(u.a);v.a<v.c.c.length;)for(m=f(ee(v),9),T=new ne(m.j);T.a<T.c.c.length;)for(S=f(ee(T),11),a=new ne(S.g);a.a<a.c.c.length;)i=f(ee(a),18),L=i.d,A=L.i.c,A!=u&&A!=g&&Cde(i,(F=new Tg(e),Wl(F,(cn(),uo)),$e(F,(Ge(),oi),i),$e(F,(ot(),Xi),(oa(),uu)),jo(F,g),F));t.Tg()}function R_t(e){var t,i,a,u,l,g,m,v,S,T,A,L;for(t=bg(e),l=lt(ut(Fe(t,(ot(),ix)))),T=0,u=0,S=new Nn((!e.e&&(e.e=new Qt(vo,e,7,4)),e.e));S.e!=S.i.gc();)v=f(Mn(S),85),m=HS(v),g=m&&l&&lt(ut(Fe(v,PE))),L=Us(f(be((!v.c&&(v.c=new Qt(Yn,v,5,8)),v.c),0),84)),m&&g?++u:m&&!g?++T:so(L)==t||L==t?++u:++T;for(a=new Nn((!e.d&&(e.d=new Qt(vo,e,8,5)),e.d));a.e!=a.i.gc();)i=f(Mn(a),85),m=HS(i),g=m&&l&&lt(ut(Fe(i,PE))),A=Us(f(be((!i.b&&(i.b=new Qt(Yn,i,4,7)),i.b),0),84)),m&&g?++T:m&&!g?++u:so(A)==t||A==t?++T:++u;return T-u}function bMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F;for(i.Sg("Depth-first cycle removal",1),A=t.a,T=A.c.length,e.c=new et,e.d=Pe(kd,Tb,30,T,16,1),e.a=Pe(kd,Tb,30,T,16,1),e.b=new et,g=0,S=new ne(A);S.a<S.c.c.length;)v=f(ee(S),9),v.p=g,NP(Fo(v))&&Qe(e.c,v),++g;for(F=new ne(e.c);F.a<F.c.c.length;)L=f(ee(F),9),NAe(e,L);for(l=0;l<T;l++)e.d[l]||(m=(Gt(l,A.c.length),f(A.c[l],9)),NAe(e,m));for(u=new ne(e.b);u.a<u.c.c.length;)a=f(ee(u),18),VS(a,!0),$e(t,(Ge(),Nq),(mn(),!0));e.c=null,e.d=null,e.a=null,e.b=null,i.Tg()}function P_t(e){var t,i,a,u,l;return e.g!=null?e.g:e.a<32?(e.g=P9n(Nc(e.f),Aa(e.e)),e.g):(u=afe((!e.c&&(e.c=yH(Nc(e.f))),e.c),0),e.e==0?u:(t=(!e.c&&(e.c=yH(Nc(e.f))),e.c).e<0?2:1,i=u.length,a=-e.e+i-t,l=new $v,l.a+=""+u,e.e>0&&a>=-6?a>=0?Xz(l,i-Aa(e.e),"."):(Hue(l,t-1,t-1,"0."),Xz(l,t+1,bm(IE,0,-Aa(a)-1))):(i-t>=1&&(Xz(l,t,"."),++i),Xz(l,i,"E"),a>0&&Xz(l,++i,"+"),Xz(l,++i,""+V7(Nc(a)))),e.g=l.a,e.g))}function vMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt;a=ke(Oe(V(t,(ot(),XLe)))),Le=f(V(t,iF),17).a,L=4,u=3,He=20/Le,F=!1,v=0,g=Qr;do{for(l=v!=1,A=v!=0,nt=0,J=e.a,pe=0,Ee=J.length;pe<Ee;++pe)z=J[pe],z.f=null,i9n(e,z,l,A,a),nt+=h.Math.abs(z.a);do m=KLn(e,t);while(m);for(W=e.a,ue=0,ve=W.length;ue<ve;++ue)if(z=W[ue],i=Nke(z).a,i!=0)for(T=new ne(z.e);T.a<T.c.c.length;)S=f(ee(T),9),S.n.b+=i;v==0||v==1?(--L,L<=0&&(nt<g||-L>Le)?(v=2,g=Qr):v==0?(v=1,g=nt):(v=0,g=nt)):(F=nt>=g||g-nt<He,g=nt,F&&--u)}while(!(F&&u<=0))}function Zde(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;for(z=new Wn,l=e.a.ec().Jc();l.Ob();)a=f(l.Pb(),176),qr(z,a,i.$e(a));for(g=(or(e),e?new Kl(e):bP(e.a.ec().Jc())),na(g,new Uo(z)),m=fJ(g),v=new iX(t),F=new Wn,ml(F.f,t,v);m.a.gc()!=0;){for(S=null,T=null,A=null,u=m.a.ec().Jc();u.Ob();)if(a=f(u.Pb(),176),ke(Oe(ac(Is(z.f,a))))<=Gi){if(Yu(F,a.a)&&!Yu(F,a.b)){T=a.b,A=a.a,S=a;break}if(Yu(F,a.b)&&!Yu(F,a.a)){T=a.a,A=a.b,S=a;break}}if(!S)break;L=new iX(T),Qe(f(ac(Is(F.f,A)),224).a,L),ml(F.f,T,L),m.a.Ac(S)!=null}return v}function wMn(e,t){bL();var i,a,u,l,g,m;return l=t.c-(e.c+e.b),u=e.c-(t.c+t.b),g=e.d-(t.d+t.a),i=t.d-(e.d+e.a),a=h.Math.max(u,l),m=h.Math.max(g,i),gg(),Kh(Cm),(h.Math.abs(a)<=Cm||a==0||isNaN(a)&&isNaN(0)?0:a<0?-1:a>0?1:pS(isNaN(a),isNaN(0)))>=0^(Kh(Cm),(h.Math.abs(m)<=Cm||m==0||isNaN(m)&&isNaN(0)?0:m<0?-1:m>0?1:pS(isNaN(m),isNaN(0)))>=0)?h.Math.max(m,a):(Kh(Cm),(h.Math.abs(a)<=Cm||a==0||isNaN(a)&&isNaN(0)?0:a<0?-1:a>0?1:pS(isNaN(a),isNaN(0)))>0?h.Math.sqrt(m*m+a*a):-h.Math.sqrt(m*m+a*a))}function vE(e,t){var i,a,u,l,g,m;if(t){if(!e.a&&(e.a=new uY),e.e==2){cY(e.a,t);return}if(t.e==1){for(u=0;u<t.Mm();u++)vE(e,t.Im(u));return}if(m=e.a.a.c.length,m==0){cY(e.a,t);return}if(g=f(SS(e.a,m-1),121),!((g.e==0||g.e==10)&&(t.e==0||t.e==10))){cY(e.a,t);return}l=t.e==0?2:t.Jm().length,g.e==0?(i=new l7,a=g.Hm(),a>=Xa?ys(i,qCe(a)):_P(i,a&To),g=new Bce(10,null,0),fwn(e.a,g,m-1)):(i=(g.Jm().length+l,new l7),ys(i,g.Jm())),t.e==0?(a=t.Hm(),a>=Xa?ys(i,qCe(a)):_P(i,a&To)):ys(i,t.Jm()),f(g,515).b=i.a}}function yMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;if(!i.dc()){for(m=0,L=0,a=i.Jc(),z=f(a.Pb(),17).a;m<t.f;){if(m==z&&(L=0,a.Ob()?z=f(a.Pb(),17).a:z=t.f+1),m!=L){for(J=f(ct(e.b,m),25),F=f(ct(e.b,L),25),W=rE(J.a),A=new ne(W);A.a<A.c.c.length;)if(T=f(ee(A),9),sw(T,F.a.c.length,F),L==0)for(g=rE(Fo(T)),l=new ne(g);l.a<l.c.c.length;)u=f(ee(l),18),VS(u,!0),$e(e,(Ge(),Nq),(mn(),!0)),$_t(e,u,1)}++L,++m}for(v=new la(e.b,0);v.b<v.d.gc();)S=(Ln(v.b<v.d.gc()),f(v.d.Xb(v.c=v.b++),25)),S.a.c.length==0&&hd(v)}}function EMn(e,t,i){var a,u,l;if(u=f(V(t,(ot(),Cte)),280),u!=(fL(),q$)){switch(i.Sg("Horizontal Compaction",1),e.a=t,l=new nmt,a=new CEt((l.d=t,l.c=f(V(l.d,P1),220),_Pn(l),S8n(l),NPn(l),l.a)),xpn(a,e.b),f(V(t,xLe),419).g){case 1:uit(a,new Tbt(e.a));break;default:uit(a,(aTe(),lIt))}switch(u.g){case 1:zj(a);break;case 2:zj(gQ(a,(ko(),Gs)));break;case 3:zj(lit(gQ(zj(a),(ko(),Gs)),new I6));break;case 4:zj(lit(gQ(zj(a),(ko(),Gs)),new Net(l)));break;case 5:zj(Cpn(a,iNt))}gQ(a,(ko(),Ds)),a.e=!0,g9n(l),i.Tg()}}function SMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(g=t.b,T=g.o,v=g.d,a=ke(Oe(TZ(g,(ot(),op)))),u=ke(Oe(TZ(g,OA))),S=ke(Oe(TZ(g,Eme))),m=new hae,oke(m,v.d,v.c,v.a,v.b),L=zPn(t,a,u,S),ue=new ne(t.d);ue.a<ue.c.c.length;){for(J=f(ee(ue),105),z=J.f.a.ec().Jc();z.Ob();)F=f(z.Pb(),336),l=F.a,A=UOn(F),i=(pe=new fd,Zwt(F,F.c,L,pe),vNn(F,A,L,pe),Zwt(F,F.d,L,pe),pe),i=e.lg(F,A,i),nf(l.a),Ba(l.a,i),Zi(new Ft(null,new Bt(i,16)),new tat(T,m));W=J.i,W&&(IOn(J,W,L,u),ve=new ja(W.g),b3e(T,m,ve),ii(ve,W.j),b3e(T,m,ve))}oke(v,m.d,m.c,m.a,m.b)}function _Mn(e){var t,i,a,u;u=e.o,tT(),e.A.dc()||ei(e.A,V6e)?t=u.b:(e.D?t=h.Math.max(u.b,$j(e.f)):t=$j(e.f),e.A.Gc((Ud(),gV))&&!e.B.Gc((yd(),FF))&&(t=h.Math.max(t,$j(f(Es(e.p,(tt(),_n)),251))),t=h.Math.max(t,$j(f(Es(e.p,pn),251)))),i=qbt(e),i&&(t=h.Math.max(t,i.b)),e.A.Gc(mV)&&(e.q==(oa(),Lb)||e.q==uu)&&(t=h.Math.max(t,_X(f(Es(e.b,(tt(),_n)),126))),t=h.Math.max(t,_X(f(Es(e.b,pn),126))))),lt(ut(e.e.Rf().mf((Rr(),hx))))?u.b=h.Math.max(u.b,t):u.b=t,a=e.f.i,a.d=0,a.a=t,zde(e.f)}function kMn(e,t,i,a,u,l,g,m){var v,S,T,A;switch(v=gm(ce(se(hjn,1),Jt,237,0,[t,i,a,u])),A=null,e.b.g){case 1:A=gm(ce(se(g8e,1),Jt,520,0,[new kv,new yk,new q2]));break;case 0:A=gm(ce(se(g8e,1),Jt,520,0,[new q2,new yk,new kv]));break;case 2:A=gm(ce(se(g8e,1),Jt,520,0,[new yk,new kv,new q2]))}for(T=new ne(A);T.a<T.c.c.length;)S=f(ee(T),520),v.c.length>1&&(v=S.Fg(v,e.a,m));return v.c.length==1?f(ct(v,v.c.length-1),237):v.c.length==2?lMn((Gt(0,v.c.length),f(v.c[0],237)),(Gt(1,v.c.length),f(v.c[1],237)),g,l):null}function TMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;u=new d3(e),l=new BEt,a=(kH(l.n),kH(l.p),nu(l.c),kH(l.f),kH(l.o),nu(l.q),nu(l.d),nu(l.g),nu(l.k),nu(l.e),nu(l.i),nu(l.j),nu(l.r),nu(l.b),L=lEt(l,u,null),l2t(l,u),L),t&&(v=new d3(t),g=VDn(v),EAe(a,ce(se(K8e,1),Jt,521,0,[g]))),A=!1,T=!1,i&&(v=new d3(i),lee in v.a&&(A=h1(v,lee).oe().a),XCt in v.a&&(T=h1(v,XCt).oe().a)),S=fit(tbt(new CN,A),T),VNn(new Ay,a,S),lee in u.a&&Wh(u,lee,null),(A||T)&&(m=new AN,t_t(S,m,A,T),Wh(u,lee,m)),F=new bnt(l),y1t(new _se(a),F),z=new vnt(l),y1t(new _se(a),z)}function xMn(e,t,i){var a,u,l,g,m,v,S;for(i.Sg("Find roots",1),e.a.c.length=0,u=tr(t.b,0);u.b!=u.d.c;)a=f(Zn(u),39),a.b.b==0&&($e(a,(_i(),Iw),(mn(),!0)),Qe(e.a,a));switch(e.a.c.length){case 0:l=new Que(0,t,"DUMMY_ROOT"),$e(l,(_i(),Iw),(mn(),!0)),$e(l,Vme,!0),Mr(t.b,l);break;case 1:break;default:for(g=new Que(0,t,QQ),v=new ne(e.a);v.a<v.c.c.length;)m=f(ee(v),39),S=new oTe(g,m),$e(S,(_i(),Vme),(mn(),!0)),Mr(g.a.a,S),Mr(g.d,S),Mr(m.b,S),$e(m,Iw,!1);$e(g,(_i(),Iw),(mn(),!0)),$e(g,Vme,!0),Mr(t.b,g)}i.Tg()}function L_t(e){var t,i,a,u,l,g;for(Mu(e.a,new YI),i=new ne(e.a);i.a<i.c.c.length;)t=f(ee(i),224),a=Ho($a(f(e.b,68).c),f(t.b,68).c),GIt?(g=f(e.b,68).b,l=f(t.b,68).b,h.Math.abs(a.a)>=h.Math.abs(a.b)?(a.b=0,l.d+l.a>g.d&&l.d<g.d+g.a&&wX(a,h.Math.max(g.c-(l.c+l.b),l.c-(g.c+g.b)))):(a.a=0,l.c+l.b>g.c&&l.c<g.c+g.b&&wX(a,h.Math.max(g.d-(l.d+l.a),l.d-(g.d+g.a))))):wX(a,o_t(f(e.b,68),f(t.b,68))),u=h.Math.sqrt(a.a*a.a+a.b*a.b),u=Kwt($$,t,u,a),wX(a,u),sce(f(t.b,68),a),Mu(t.a,new pg(a)),f($$.b,68),ixe($$,W6e,t)}function CMn(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z;for(e.f=new fae,S=0,u=0,g=new ne(e.e.b);g.a<g.c.c.length;)for(l=f(ee(g),25),v=new ne(l.a);v.a<v.c.c.length;){for(m=f(ee(v),9),m.p=S++,a=new Tn(Cn(ao(m).a.Jc(),new B));Kn(a);)i=f(Dn(a),18),i.p=u++;for(t=Yde(m),L=new ne(m.j);L.a<L.c.c.length;)A=f(ee(L),11),t&&(z=A.a.b,z!=h.Math.floor(z)&&(T=z-oE(Nc(h.Math.round(z))),A.a.b-=T)),F=A.n.b+A.a.b,F!=h.Math.floor(F)&&(T=F-oE(Nc(h.Math.round(F))),A.n.b-=T)}e.g=S,e.b=u,e.i=Pe(ujn,Jt,401,S,0,1),e.c=Pe(cjn,Jt,639,u,0,1),e.d.a.$b()}function Jn(e){var t,i,a,u,l,g,m,v,S;if(e.Mj())if(v=e.Nj(),e.i>0){if(t=new ISe(e.i,e.g),i=e.i,l=i<100?null:new Fv(i),e.Qj())for(a=0;a<e.i;++a)g=e.g[a],l=e.Sj(g,l);if(sj(e),u=i==1?e.Fj(4,be(t,0),null,0,v):e.Fj(6,t,null,-1,v),e.Jj()){for(a=new GN(t);a.e!=a.i.gc();)l=e.Lj(Ple(a),l);l?(l.kj(u),l.lj()):e.Gj(u)}else l?(l.kj(u),l.lj()):e.Gj(u)}else sj(e),e.Gj(e.Fj(6,(Ht(),Ga),null,-1,v));else if(e.Jj())if(e.i>0){for(m=e.g,S=e.i,sj(e),l=S<100?null:new Fv(S),a=0;a<S;++a)g=m[a],l=e.Lj(g,l);l&&l.lj()}else sj(e);else sj(e)}function uOe(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(i1t(this),i==(AS(),Aw)?Go(this.r,e):Go(this.w,e),T=Gi,S=qo,g=t.a.ec().Jc();g.Ob();)u=f(g.Pb(),47),m=f(u.a,451),a=f(u.b,18),v=a.c,v==e&&(v=a.d),m==Aw?Go(this.r,v):Go(this.w,v),L=(tt(),wl).Gc(v.j)?ke(Oe(V(v,(Ge(),pD)))):uc(ce(se(Zo,1),Ye,8,0,[v.i.n,v.n,v.a])).b,T=h.Math.min(T,L),S=h.Math.max(S,L);for(A=(tt(),wl).Gc(e.j)?ke(Oe(V(e,(Ge(),pD)))):uc(ce(se(Zo,1),Ye,8,0,[e.i.n,e.n,e.a])).b,gyt(this,A,T,S),l=t.a.ec().Jc();l.Ob();)u=f(l.Pb(),47),Iyt(this,f(u.b,18));this.o=!1}function AMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In;return i=e.l&8191,a=e.l>>13|(e.m&15)<<9,u=e.m>>4&8191,l=e.m>>17|(e.h&255)<<5,g=(e.h&1048320)>>8,m=t.l&8191,v=t.l>>13|(t.m&15)<<9,S=t.m>>4&8191,T=t.m>>17|(t.h&255)<<5,A=(t.h&1048320)>>8,xt=i*m,Ct=a*m,Kt=u*m,tn=l*m,In=g*m,v!=0&&(Ct+=i*v,Kt+=a*v,tn+=u*v,In+=l*v),S!=0&&(Kt+=i*S,tn+=a*S,In+=u*S),T!=0&&(tn+=i*T,In+=a*T),A!=0&&(In+=i*A),F=xt&Ed,z=(Ct&511)<<13,L=F+z,J=xt>>22,ue=Ct>>9,pe=(Kt&262143)<<4,ve=(tn&31)<<17,W=J+ue+pe+ve,Le=Kt>>18,He=tn>>5,nt=(In&4095)<<8,Ee=Le+He+nt,W+=L>>22,L&=Ed,Ee+=W>>22,W&=Ed,Ee&=A1,Ku(L,W,Ee)}function D_t(e){var t,i,a,u,l,g,m;if(m=f(ct(e.j,0),11),m.g.c.length!=0&&m.e.c.length!=0)throw ie(new As("Interactive layout does not support NORTH/SOUTH ports with incoming _and_ outgoing edges."));if(m.g.c.length!=0){for(l=Gi,i=new ne(m.g);i.a<i.c.c.length;)t=f(ee(i),18),g=t.d.i,a=f(V(g,(ot(),Dte)),139),l=h.Math.min(l,g.n.a-a.b);return new nS(or(l))}if(m.e.c.length!=0){for(u=qo,i=new ne(m.e);i.a<i.c.c.length;)t=f(ee(i),18),g=t.c.i,a=f(V(g,(ot(),Dte)),139),u=h.Math.max(u,g.n.a+g.o.a+a.c);return new nS(or(u))}return a7(),a7(),Dpe}function IMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(i.Sg("Interactive cycle breaking",1),A=new et,F=new ne(t.a);F.a<F.c.c.length;)for(L=f(ee(F),9),L.p=1,z=Y3e(L).a,T=$S(L,(cs(),Uu)).Jc();T.Ob();)for(S=f(T.Pb(),11),l=new ne(S.g);l.a<l.c.c.length;)a=f(ee(l),18),W=a.d.i,W!=L&&(J=Y3e(W).a,J<z&&vn(A.c,a));for(g=new ne(A);g.a<g.c.c.length;)a=f(ee(g),18),VS(a,!0);for(A.c.length=0,v=new ne(t.a);v.a<v.c.c.length;)m=f(ee(v),9),m.p>0&&Lyt(e,m,A);for(u=new ne(A);u.a<u.c.c.length;)a=f(ee(u),18),VS(a,!0);A.c.length=0,i.Tg()}function M_t(e,t){var i,a,u,l,g,m,v;if(e.ll()){if(e.i>4)if(e.ck(t)){if(e.Zk()){if(u=f(t,52),a=u.Ah(),v=a==e.e&&(e.jl()?u.uh(u.Bh(),e.fl())==e.gl():-1-u.Bh()==e.Ij()),e.kl()&&!v&&!a&&u.Fh()){for(l=0;l<e.i;++l)if(i=e.ml(f(e.g[l],57)),Ae(i)===Ae(t))return!0}return v}else if(e.jl()&&!e.il()){if(g=f(t,57).Ih(os(f(e.Ik(),19))),Ae(g)===Ae(e.e))return!0;if(g==null||!f(g,57).Rh())return!1}}else return!1;if(m=lL(e,t),e.kl()&&!m){for(l=0;l<e.i;++l)if(u=e.ml(f(e.g[l],57)),Ae(u)===Ae(t))return!0}return m}else return lL(e,t)}function OMn(e,t){var i,a,u,l,g,m,v,S,T;for(i=0,T=new et,m=new ne(t);m.a<m.c.c.length;){switch(g=f(ee(m),11),_Ce(e.b,e.d[g.p]),T.c.length=0,g.i.k.g){case 0:a=f(V(g,(Ge(),ed)),9),Mu(a.j,new dtt(T));break;case 1:amn(S1(Jr(new Ft(null,new Bt(g.i.j,16)),new ftt(g))),new htt(T));break;case 3:u=f(V(g,(Ge(),oi)),11),Qe(T,new Ya(u,rt(g.e.c.length+g.g.c.length)))}for(S=new ne(T);S.a<S.c.c.length;)v=f(ee(S),47),l=V2e(e,f(v.a,11)),l>e.d[g.p]&&(i+=FTe(e.b,l)*f(v.b,17).a,Jv(e.a,rt(l)));for(;!u7(e.a);)vxe(e.b,f(JN(e.a),17).a)}return i}function NMn(e){var t,i,a;for(pR(qE,ce(se(HR,1),Jt,146,0,[new a1])),i=new bN(e),a=0;a<i.a.length;++a)t=gT(i,a).re().a,Mt(t,"layered")?pR(qE,ce(se(HR,1),Jt,146,0,[new Jd])):Mt(t,"force")?pR(qE,ce(se(HR,1),Jt,146,0,[new ss])):Mt(t,"stress")?pR(qE,ce(se(HR,1),Jt,146,0,[new P5])):Mt(t,"mrtree")?pR(qE,ce(se(HR,1),Jt,146,0,[new B9])):Mt(t,"radial")?pR(qE,ce(se(HR,1),Jt,146,0,[new rz])):Mt(t,"sporeOverlap")||Mt(t,"sporeCompaction")?pR(qE,ce(se(HR,1),Jt,146,0,[new U9])):Mt(t,"rectpacking")&&pR(qE,ce(se(HR,1),Jt,146,0,[new Pk]))}function RMn(e,t){var i,a,u,l,g,m,v,S,T,A;if(T=f(V(e,(Ge(),Tc)),64),a=f(ct(e.j,0),11),T==(tt(),hn)?No(a,Bn):T==Bn&&No(a,hn),f(V(t,(ot(),DE)),181).Gc((Ud(),GE))){if(v=ke(Oe(V(e,yD))),S=ke(Oe(V(e,ED))),g=ke(Oe(V(e,cx))),m=f(V(t,ax),22),m.Gc((Xl(),M1)))for(i=S,A=e.o.a/2-a.n.a,l=new ne(a.f);l.a<l.c.c.length;)u=f(ee(l),72),u.n.b=i,u.n.a=A-u.o.a/2,i+=u.o.b+g;else if(m.Gc(Lw))for(l=new ne(a.f);l.a<l.c.c.length;)u=f(ee(l),72),u.n.a=v+e.o.a-a.n.a;l2n(new rs((h7(),new Oce(t,!1,!1,new py))),new bX(null,e,!1))}}function PMn(e,t){var i,a,u,l,g,m,v,S,T;if(t.c.length!=0){for(Ht(),ace(t.c,t.c.length,null),u=new ne(t),a=f(ee(u),152);u.a<u.c.c.length;)i=f(ee(u),152),Sxe(a.e.c,i.e.c)&&!(f3e(zct(a.e).b,i.e.d)||f3e(zct(i.e).b,a.e.d))?a=($o(a.k,i.k),$o(a.b,i.b),$o(a.c,i.c),Ba(a.i,i.i),$o(a.d,i.d),$o(a.j,i.j),l=h.Math.min(a.e.c,i.e.c),g=h.Math.min(a.e.d,i.e.d),m=h.Math.max(a.e.c+a.e.b,i.e.c+i.e.b),v=m-l,S=h.Math.max(a.e.d+a.e.a,i.e.d+i.e.a),T=S-g,Yut(a.e,l,g,v,T),c2n(a.f,i.f),!a.a&&(a.a=i.a),$o(a.g,i.g),Qe(a.g,i),a):(JSt(e,a),a=i);JSt(e,a)}}function LMn(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le;for(v=new et,l=new ne(t.a);l.a<l.c.c.length;)for(u=f(ee(l),9),m=new ne(u.j);m.a<m.c.c.length;){for(g=f(ee(m),11),T=null,ve=pm(g.g),Ee=0,Le=ve.length;Ee<Le;++Ee)pe=ve[Ee],UP(pe.d.i,i)||(ue=Xde(e,t,i,pe,pe.c,(cs(),Uu),T),ue!=T&&vn(v.c,ue),ue.c&&(T=ue));for(S=null,z=pm(g.e),W=0,J=z.length;W<J;++W)F=z[W],UP(F.c.i,i)||(ue=Xde(e,t,i,F,F.d,(cs(),td),S),ue!=S&&vn(v.c,ue),ue.c&&(S=ue))}for(L=new ne(v);L.a<L.c.c.length;)A=f(ee(L),441),Oc(t.a,A.a,0)!=-1||Qe(t.a,A.a),A.c&&vn(a.c,A)}function DMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F;for(m=new ne(t);m.a<m.c.c.length;)l=f(ee(m),238),l.e=null,l.c=0;for(v=null,g=new ne(t);g.a<g.c.c.length;)if(l=f(ee(g),238),A=l.d[0],!(i&&A.k!=(cn(),Yi))){for(F=f(V(A,(Ge(),XR)),15).Jc();F.Ob();)L=f(F.Pb(),9),(!i||L.k==(cn(),Yi))&&((!l.e&&(l.e=new et),l.e).Ec(e.b[L.c.p][L.p]),++e.b[L.c.p][L.p].c);if(!i&&A.k==(cn(),Yi)){if(v)for(T=f(si(e.d,v),22).Jc();T.Ob();)for(S=f(T.Pb(),9),u=f(si(e.d,A),22).Jc();u.Ob();)a=f(u.Pb(),9),avn(e.b[S.c.p][S.p]).Ec(e.b[a.c.p][a.p]),++e.b[a.c.p][a.p].c;v=A}}}function MMn(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J;for(A=new ja(f(Fe(e,(GZ(),n9e)),8)),A.a=h.Math.max(A.a-i.b-i.c,0),A.b=h.Math.max(A.b-i.d-i.a,0),u=Oe(Fe(e,Q8e)),(u==null||(an(u),u<=0))&&(u=1.3),m=new et,z=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));z.e!=z.i.gc();)F=f(Mn(z),27),g=new Oct(F),vn(m.c,g);switch(L=f(Fe(e,Ube),321),L.g){case 3:J=TLn(m,t,A.a,A.b,(S=a,an(u),S));break;case 1:J=hDn(m,t,A.a,A.b,(T=a,an(u),T));break;default:J=FMn(m,t,A.a,A.b,(v=a,an(u),v))}l=new LJ(J),W=ife(l,t,i,A.a,A.b,a,(an(u),u)),WS(e,W.a,W.b,!1,!0)}function jMn(e,t,i,a){var u,l,g,m,v,S;if(m=e.j,m==(tt(),fc)&&t!=(oa(),Fg)&&t!=(oa(),Pw)&&(m=VSt(e,i),No(e,m),!(e.q?e.q:(Ht(),Ht(),Cb))._b((ot(),ox))&&m!=fc&&(e.n.a!=0||e.n.b!=0)&&$e(e,ox,dIn(e,m))),t==(oa(),Lb)){switch(S=0,m.g){case 1:case 3:l=e.i.o.a,l>0&&(S=e.n.a/l);break;case 2:case 4:u=e.i.o.b,u>0&&(S=e.n.b/u)}$e(e,(Ge(),d_),S)}if(v=e.o,g=e.a,a)g.a=a.a,g.b=a.b,e.d=!0;else if(t!=Fg&&t!=Pw&&m!=fc)switch(m.g){case 1:g.a=v.a/2;break;case 2:g.a=v.a,g.b=v.b/2;break;case 3:g.a=v.a/2,g.b=v.b;break;case 4:g.b=v.b/2}else g.a=v.a/2,g.b=v.b/2}function e$(e){var t,i,a,u,l,g,m,v,S,T;if(e.Mj())if(T=e.Bj(),v=e.Nj(),T>0)if(t=new Jxe(e.mj()),i=T,l=i<100?null:new Fv(i),tH(e,i,t.g),u=i==1?e.Fj(4,be(t,0),null,0,v):e.Fj(6,t,null,-1,v),e.Jj()){for(a=new Nn(t);a.e!=a.i.gc();)l=e.Lj(Mn(a),l);l?(l.kj(u),l.lj()):e.Gj(u)}else l?(l.kj(u),l.lj()):e.Gj(u);else tH(e,e.Bj(),e.Cj()),e.Gj(e.Fj(6,(Ht(),Ga),null,-1,v));else if(e.Jj())if(T=e.Bj(),T>0){for(m=e.Cj(),S=T,tH(e,T,m),l=S<100?null:new Fv(S),a=0;a<S;++a)g=m[a],l=e.Lj(g,l);l&&l.lj()}else tH(e,e.Bj(),e.Cj());else tH(e,e.Bj(),e.Cj())}function $Mn(e,t){var i,a,u,l,g,m,v,S;for(t.Sg("Sort By Input Model "+V(e,(ot(),ip)),1),l=0,u=new ne(e.b);u.a<u.c.c.length;){for(a=f(ee(u),25),a.p=l,S=l==0?0:l-1,v=f(ct(e.b,S),25),i=new Zxe(v,f(V(e,ip),266),f(V(e,cme),349),!0),c2t(a.a,i),m=new ne(a.a);m.a<m.c.c.length;)g=f(ee(m),9),Ae(V(g,Xi))!==Ae((oa(),HE))&&Ae(V(g,Xi))!==Ae(uu)&&(Ht(),na(g.j,new X1t(v,f(V(e,ip),266),ywt(g),lt(ut(V(e,ume))))),t._g("Node "+g+" ports: "+g.j));i=new Zxe(v,f(V(e,ip),266),f(V(e,cme),349),!1),c2t(a.a,i),t._g("Layer "+l+": "+a),++l}t.Tg()}function FMn(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J;for(m=Pe(sa,_s,30,e.c.length,15,1),L=new jX(new l5),L3e(L,e),S=0,W=new et;L.b.c.length!=0;)if(g=f(L.b.c.length==0?null:ct(L.b,0),162),S>1&&Al(g)*jd(g)/2>m[0]){for(l=0;l<W.c.length-1&&Al(g)*jd(g)/2>m[l];)++l;z=new Xv(W,0,l+1),A=new LJ(z),T=Al(g)/jd(g),v=ife(A,t,new xN,i,a,u,T),ii(Cp(A.e),v),KN(dL(L,A),NL),F=new Xv(W,l+1,W.c.length),L3e(L,F),W.c.length=0,S=0,Plt(m,m.length,0)}else J=L.b.c.length==0?null:ct(L.b,0),J!=null&&Pue(L,0),S>0&&(m[S]=m[S-1]),m[S]+=Al(g)*jd(g),++S,vn(W.c,g);return W}function BMn(e,t){var i,a,u,l;i=t.b,l=new Kl(i.j),u=0,a=i.j,a.c.length=0,yS(f(cE(e.b,(tt(),hn),(LS(),c_)),15),i),u=uG(l,u,new Ey,a),yS(f(cE(e.b,hn,_w),15),i),u=uG(l,u,new lk,a),yS(f(cE(e.b,hn,s_),15),i),yS(f(cE(e.b,_n,c_),15),i),yS(f(cE(e.b,_n,_w),15),i),u=uG(l,u,new EO,a),yS(f(cE(e.b,_n,s_),15),i),yS(f(cE(e.b,Bn,c_),15),i),u=uG(l,u,new bC,a),yS(f(cE(e.b,Bn,_w),15),i),u=uG(l,u,new tb,a),yS(f(cE(e.b,Bn,s_),15),i),yS(f(cE(e.b,pn,c_),15),i),u=uG(l,u,new Jb,a),yS(f(cE(e.b,pn,_w),15),i),yS(f(cE(e.b,pn,s_),15),i)}function UMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(i.Sg("Model order cycle breaking",1),e.a=0,e.b=0,F=new et,T=t.a.c.length,S=new ne(t.a);S.a<S.c.c.length;)v=f(ee(S),9),li(v,(Ge(),hi))&&(T=h.Math.max(T,f(V(v,hi),17).a+1));for(W=new ne(t.a);W.a<W.c.c.length;)for(z=f(ee(W),9),g=Hyt(e,z,T),L=$S(z,(cs(),Uu)).Jc();L.Ob();)for(A=f(L.Pb(),11),l=new ne(A.g);l.a<l.c.c.length;)a=f(ee(l),18),J=a.d.i,m=Hyt(e,J,T),m<g&&vn(F.c,a);for(u=new ne(F);u.a<u.c.c.length;)a=f(ee(u),18),VS(a,!0),$e(t,(Ge(),Nq),(mn(),!0));F.c.length=0,i.Tg()}function zMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(t.Sg("Layer size calculation",1),T=Gi,S=qo,u=!1,m=new ne(e.b);m.a<m.c.c.length;)if(g=f(ee(m),25),v=g.c,v.a=0,v.b=0,g.a.c.length!=0){for(u=!0,L=new ne(g.a);L.a<L.c.c.length;)A=f(ee(L),9),z=A.o,F=A.d,v.a=h.Math.max(v.a,z.a+F.b+F.c);a=f(ct(g.a,0),9),W=a.n.b-a.d.d,a.k==(cn(),bo)&&(W-=f(V(e,(ot(),$q)),139).d),l=f(ct(g.a,g.a.c.length-1),9),i=l.n.b+l.o.b+l.d.a,l.k==bo&&(i+=f(V(e,(ot(),$q)),139).a),v.b=i-W,T=h.Math.min(T,W),S=h.Math.max(S,i)}u||(T=0,S=0),e.f.b=S-T,e.c.b-=T,t.Tg()}function j_t(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(g=t.d,m=i.d;g.a-m.a==0&&g.b-m.b==0;)v=!1,me(t,249)&&me(i,249)&&!v?(S=f(t,249).a,T=Ho(new ja(Oxe(S)),Ixe(S)),a=2,u=new Ke(T.a/h.Math.sqrt(T.a*T.a+T.b*T.b)*a,-T.b/h.Math.sqrt(T.a*T.a+T.b*T.b)*a),ii(g,u),A=f(i,249).a,L=Ho(new ja(Oxe(A)),Ixe(A)),a=T==L?-2:2,l=new Ke(L.a/h.Math.sqrt(L.a*L.a+L.b*L.b)*a,-(L.b/h.Math.sqrt(L.a*L.a+L.b*L.b))*a),ii(g,l),v=!0):(g.a+=wd(e,26)*r$+wd(e,27)*i$-.5,g.b+=wd(e,26)*r$+wd(e,27)*i$-.5,m.a+=wd(e,26)*r$+wd(e,27)*i$-.5,m.b+=wd(e,26)*r$+wd(e,27)*i$-.5)}function HMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(S=jNn(t),W=f(V(t,(ot(),Q$)),323),Oa(S,new Qd(W)),J=f(V(t,Dq),298),Oa(S,new jv(J)),z=0,T=new et,l=new ej(S);l.a!=l.b;)u=f(sZ(l),36),X_t(e.c,u),L=f(V(u,(Ge(),nme)),15),z+=L.gc(),a=L.Jc(),Qe(T,new Ya(u,a));for(i.Sg("Recursive hierarchical layout",z),F=f(f(ct(T,T.c.length-1),47).b,49);F.Ob();)for(v=new ne(T);v.a<v.c.c.length;)for(m=f(ee(v),47),L=f(m.b,49),g=f(m.a,36);L.Ob();)if(A=f(L.Pb(),44),me(A,498)){if(g.e)break;A.If(g,i.bh(1));break}else A.If(g,i.bh(1));i.Tg()}function lOe(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;for(l=0,g=0,S=new ne(e.a);S.a<S.c.c.length;)m=f(ee(S),9),l=h.Math.max(l,m.d.b),g=h.Math.max(g,m.d.c);for(v=new ne(e.a);v.a<v.c.c.length;){switch(m=f(ee(v),9),i=f(V(m,(ot(),Nm)),254),i.g){case 1:z=0;break;case 2:z=1;break;case 5:z=.5;break;default:for(a=0,A=0,F=new ne(m.j);F.a<F.c.c.length;)L=f(ee(F),11),L.e.c.length==0||++a,L.g.c.length==0||++A;a+A==0?z=.5:z=A/(a+A)}J=e.c,T=m.o.a,ue=(J.a-T)*z,z>.5?ue-=g*2*(z-.5):z<.5&&(ue+=l*2*(.5-z)),u=m.d.b,ue<u&&(ue=u),W=m.d.c,ue>J.a-W-T&&(ue=J.a-W-T),m.n.a=t+ue}}function GMn(e){var t,i,a,u,l;if(a=f(V(e,(ot(),el)),171),a==(Ff(),Tw)){for(i=new Tn(Cn(Fo(e).a.Jc(),new B));Kn(i);)if(t=f(Dn(i),18),!$pt(t))throw ie(new h0(ehe+lG(e)+"' has its layer constraint set to FIRST_SEPARATE, but has at least one incoming edge. FIRST_SEPARATE nodes must not have incoming edges."))}else if(a==tx){for(l=new Tn(Cn(ao(e).a.Jc(),new B));Kn(l);)if(u=f(Dn(l),18),!$pt(u))throw ie(new h0(ehe+lG(e)+"' has its layer constraint set to LAST_SEPARATE, but has at least one outgoing edge. LAST_SEPARATE nodes must not have outgoing edges."))}}function qMn(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;for(L=new et,u=new et,W=null,m=t.Jc();m.Ob();)g=f(m.Pb(),17),l=new ttt(g.a),vn(u.c,l),W&&(l.d=W,W.e=l),W=l;for(ve=vDn(e),T=0;T<u.c.length;++T){for(F=null,J=QTe((Gt(0,u.c.length),f(u.c[0],645))),i=null,a=Gi,A=1;A<e.b.c.length;++A)ue=J?h.Math.abs(J.b-A):h.Math.abs(A-F.b)+1,z=F?h.Math.abs(A-F.b):ue+1,z<ue?(S=F,v=z):(S=J,v=ue),pe=(Ee=ke(Oe(V(e,(ot(),lDe)))),ve[A]+h.Math.pow(v,Ee)),pe<a&&(a=pe,i=S,i.c=A),J&&A==J.b&&(F=J,J=rwn(J));i&&(Qe(L,rt(i.c)),i.a=!0,aCn(i))}return Ht(),ace(L.c,L.c.length,null),L}function NG(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;if(e.e&&e.c.c<e.f)throw ie(new As("Expected "+e.f+" phases to be configured; only found "+e.c.c));for(T=f(Tp(e.g),10),F=lb(e.f),l=T,m=0,S=l.length;m<S;++m)a=l[m],A=f(DH(e,a.g),188),A?Qe(F,f(j0t(e,A),106)):F.c.push(null);for(z=new _o,Zi(Jr(Lu(Jr(new Ft(null,new Bt(F,16)),new Y2),new Wtt(t)),new o5),new Ktt(z)),lf(z,e.a),i=new et,u=T,g=0,v=u.length;g<v;++g)a=u[g],$o(i,f1t(e,Lce(f(DH(z,a.g),20)))),L=f(ct(F,a.g),106),L&&vn(i.c,L);return $o(i,f1t(e,Lce(f(DH(z,T[T.length-1].g+1),20)))),i}function dOe(e,t,i){var a,u,l,g,m,v;if(t.l==0&&t.m==0&&t.h==0)throw ie(new fY("divide by zero"));if(e.l==0&&e.m==0&&e.h==0)return i&&(ww=Ku(0,0,0)),Ku(0,0,0);if(t.h==zG&&t.m==0&&t.l==0)return xCn(e,i);if(v=!1,t.h>>19!=0&&(t=VP(t),v=!v),g=B6n(t),l=!1,u=!1,a=!1,e.h==zG&&e.m==0&&e.l==0)if(u=!0,l=!0,g==-1)e=dst((MP(),b6e)),a=!0,v=!v;else return m=rIe(e,g),v&&Jue(m),i&&(ww=Ku(0,0,0)),m;else e.h>>19!=0&&(l=!0,e=VP(e),a=!0,v=!v);return g!=-1?VTn(e,g,v,l,i):z3e(e,t)<0?(i&&(l?ww=VP(e):ww=Ku(e.l,e.m,e.h)),Ku(0,0,0)):ALn(a?e:Ku(e.l,e.m,e.h),t,v,l,u,i)}function Qde(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;if(g=e.e,v=t.e,g==0)return t;if(v==0)return e;if(l=e.d,m=t.d,l+m==2)return i=ia(e.a[0],ls),a=ia(t.a[0],ls),g==v?(T=Ua(i,a),z=mr(T),F=mr(Jy(T,32)),F==0?new m1(g,z):new Zy(g,2,ce(se(pr,1),Pr,30,15,[z,F]))):(ym(),ZY(g<0?hh(a,i):hh(i,a),0)?aw(g<0?hh(a,i):hh(i,a)):X7(aw(T0(g<0?hh(a,i):hh(i,a)))));if(g==v)L=g,A=l>=m?due(e.a,l,t.a,m):due(t.a,m,e.a,l);else{if(u=l!=m?l>m?1:-1:ECe(e.a,t.a,l),u==0)return ym(),M$;u==1?(L=g,A=uue(e.a,l,t.a,m)):(L=v,A=uue(t.a,m,e.a,l))}return S=new Zy(L,A.length,A),nj(S),S}function VMn(e,t){var i,a,u,l,g,m,v;if(!(e.g>t.f||t.g>e.f)){for(i=0,a=0,g=e.w.a.ec().Jc();g.Ob();)u=f(g.Pb(),11),ule(uc(ce(se(Zo,1),Ye,8,0,[u.i.n,u.n,u.a])).b,t.g,t.f)&&++i;for(m=e.r.a.ec().Jc();m.Ob();)u=f(m.Pb(),11),ule(uc(ce(se(Zo,1),Ye,8,0,[u.i.n,u.n,u.a])).b,t.g,t.f)&&--i;for(v=t.w.a.ec().Jc();v.Ob();)u=f(v.Pb(),11),ule(uc(ce(se(Zo,1),Ye,8,0,[u.i.n,u.n,u.a])).b,e.g,e.f)&&++a;for(l=t.r.a.ec().Jc();l.Ob();)u=f(l.Pb(),11),ule(uc(ce(se(Zo,1),Ye,8,0,[u.i.n,u.n,u.a])).b,e.g,e.f)&&--a;i<a?new wJ(e,t,a-i):a<i?new wJ(t,e,i-a):(new wJ(t,e,0),new wJ(e,t,0))}}function WMn(e){var t,i,a,u,l,g,m,v,S,T;for(t=new n1,i=new n1,S=Mt(fq,(u=Kj(e.b,kr),u?dr(Rp((!u.b&&(u.b=new Md((Ut(),Za),xc,u)),u.b),mh)):null)),v=0;v<e.i;++v)m=f(e.g[v],178),me(m,101)?(g=f(m,19),(g.Bb&Dc)!=0?((g.Bb&Sm)==0||!S&&(l=Kj(g,kr),(l?dr(Rp((!l.b&&(l.b=new Md((Ut(),Za),xc,l)),l.b),cee)):null)==null))&&Qn(t,g):(T=os(g),T&&(T.Bb&Dc)!=0||((g.Bb&Sm)==0||!S&&(a=Kj(g,kr),(a?dr(Rp((!a.b&&(a.b=new Md((Ut(),Za),xc,a)),a.b),cee)):null)==null))&&Qn(i,g))):(is(),f(m,69).uk()&&(m.pk()||(Qn(t,m),Qn(i,m))));ET(t),ET(i),e.a=f(t.g,253),f(i.g,253)}function iA(e,t,i){var a,u,l,g,m,v,S,T,A;if(Pi(t,i)>=0)return i;switch(kS(Ns(e,i))){case 2:{if(Mt("",I0(e,i.nk()).ve())){if(v=gH(Ns(e,i)),m=xP(Ns(e,i)),T=uIe(e,t,v,m),T)return T;for(u=BIe(e,t),g=0,A=u.gc();g<A;++g)if(T=f(u.Xb(g),178),mIe(mce(Ns(e,T)),v))return T}return null}case 4:{if(Mt("",I0(e,i.nk()).ve())){for(a=i;a;a=XEn(Ns(e,a)))if(S=gH(Ns(e,a)),m=xP(Ns(e,a)),T=lIe(e,t,S,m),T)return T;if(v=gH(Ns(e,i)),Mt(CE,v))return V3e(e,t);for(l=Mde(e,t),g=0,A=l.gc();g<A;++g)if(T=f(l.Xb(g),178),mIe(mce(Ns(e,T)),v))return T}return null}default:return null}}function KMn(e,t,i){var a,u,l,g,m,v,S,T;if(i.gc()==0)return!1;if(m=(is(),f(t,69).uk()),l=m?i:new OS(i.gc()),x1(e.e,t)){if(t.Pi())for(S=i.Jc();S.Ob();)v=S.Pb(),lQ(e,t,v,me(t,101)&&(f(t,19).Bb&Xa)!=0)||(u=mb(t,v),l.Gc(u)||l.Ec(u));else if(!m)for(S=i.Jc();S.Ob();)v=S.Pb(),u=mb(t,v),l.Ec(u)}else{if(i.gc()>1)throw ie(new dn(gq));for(T=Ju(e.e.zh(),t),a=f(e.g,122),g=0;g<e.i;++g)if(u=a[g],T.Zl(u.Ik())){if(i.Gc(m?u:u.kd()))return!1;for(S=i.Jc();S.Ob();)v=S.Pb(),f(Y3(e,g,m?f(v,74):mb(t,v)),74);return!0}m||(u=mb(t,i.Jc().Pb()),l.Ec(u))}return Qi(e,l)}function YMn(e,t){var i,a,u,l,g,m,v,S,T;for(T=new ki,m=(S=new kp(e.c).a.vc().Jc(),new zy(S));m.a.Ob();)l=(u=f(m.a.Pb(),43),f(u.kd(),453)),l.b==0&&Wi(T,l,T.c.b,T.c);for(;T.b!=0;)for(l=f(T.b==0?null:(Ln(T.b!=0),lh(T,T.a.a)),453),l.a==null&&(l.a=0),a=new ne(l.d);a.a<a.c.c.length;)i=f(ee(a),647),i.b.a==null?i.b.a=ke(l.a)+i.a:t.o==(vg(),$E)?i.b.a=h.Math.min(ke(i.b.a),ke(l.a)+i.a):i.b.a=h.Math.max(ke(i.b.a),ke(l.a)+i.a),--i.b.b,i.b.b==0&&Mr(T,i.b);for(g=(v=new kp(e.c).a.vc().Jc(),new zy(v));g.a.Ob();)l=(u=f(g.a.Pb(),43),f(u.kd(),453)),t.i[l.c.p]=l.a}function XMn(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z;for(T=i+t.c.c.a,F=new ne(t.j);F.a<F.c.c.length;){if(L=f(ee(F),11),u=uc(ce(se(Zo,1),Ye,8,0,[L.i.n,L.n,L.a])),t.k==(cn(),xu)&&(m=f(V(L,(Ge(),oi)),11),u.a=uc(ce(se(Zo,1),Ye,8,0,[m.i.n,m.n,m.a])).a,t.n.a=u.a),g=new Ke(0,u.b),L.j==(tt(),_n))g.a=T;else if(L.j==pn)g.a=i;else continue;if(z=h.Math.abs(u.a-g.a),!(z<=a&&!vIn(t)))for(l=L.g.c.length+L.e.c.length>1,S=new yg(L.b);sc(S.a)||sc(S.b);)v=f(sc(S.a)?ee(S.a):ee(S.b),18),A=v.c==L?v.d:v.c,h.Math.abs(uc(ce(se(Zo,1),Ye,8,0,[A.i.n,A.n,A.a])).b-g.b)>1&&F5n(e,v,g,l,L)}}function JMn(e){var t,i,a,u,l,g;if(u=new la(e.e,0),a=new la(e.a,0),e.d)for(i=0;i<e.b;i++)Ln(u.b<u.d.gc()),u.d.Xb(u.c=u.b++);else for(i=0;i<e.b-1;i++)Ln(u.b<u.d.gc()),u.d.Xb(u.c=u.b++),hd(u);for(t=ke((Ln(u.b<u.d.gc()),Oe(u.d.Xb(u.c=u.b++))));e.f-t>Mhe;){for(l=t,g=0;h.Math.abs(t-l)<Mhe;)++g,t=ke((Ln(u.b<u.d.gc()),Oe(u.d.Xb(u.c=u.b++)))),Ln(a.b<a.d.gc()),a.d.Xb(a.c=a.b++);g<e.b&&(Ln(u.b>0),u.a.Xb(u.c=--u.b),pDn(e,e.b-g,l,a,u),Ln(u.b<u.d.gc()),u.d.Xb(u.c=u.b++)),Ln(a.b>0),a.a.Xb(a.c=--a.b)}if(!e.d)for(i=0;i<e.b-1;i++)Ln(u.b<u.d.gc()),u.d.Xb(u.c=u.b++),hd(u);e.d=!0,e.c=!0}function fi(){fi=Z,tje=(z2e(),Mc).b,kMt=f(be(je(Mc.b),0),35),VE=f(be(je(Mc.b),1),35),_Mt=f(be(je(Mc.b),2),35),KA=Mc.bb,f(be(je(Mc.bb),0),35),f(be(je(Mc.bb),1),35),YA=Mc.fb,qF=f(be(je(Mc.fb),0),35),f(be(je(Mc.fb),1),35),f(be(je(Mc.fb),2),19),T_=Mc.qb,MMt=f(be(je(Mc.qb),0),35),f(be(je(Mc.qb),1),19),f(be(je(Mc.qb),2),19),AV=f(be(je(Mc.qb),3),35),IV=f(be(je(Mc.qb),4),35),WF=f(be(je(Mc.qb),6),35),VF=f(be(je(Mc.qb),5),19),TMt=Mc.j,xMt=Mc.k,CMt=Mc.q,AMt=Mc.w,IMt=Mc.B,OMt=Mc.A,NMt=Mc.C,RMt=Mc.D,PMt=Mc._,LMt=Mc.cb,DMt=Mc.hb}function ZMn(e,t,i){var a,u,l,g,m,v,S,T,A,L,F;e.c=0,e.b=0,a=2*t.c.a.c.length+1;e:for(A=i.Jc();A.Ob();){if(T=f(A.Pb(),11),m=T.j==(tt(),hn)||T.j==Bn,F=0,m){if(L=f(V(T,(Ge(),ed)),9),!L)continue;F+=IPn(e,a,T,L)}else{for(S=new ne(T.g);S.a<S.c.c.length;)if(v=f(ee(S),18),u=v.d,u.i.c==t.c){Qe(e.a,T);continue e}else F+=e.g[u.p];for(g=new ne(T.e);g.a<g.c.c.length;)if(l=f(ee(g),18),u=l.c,u.i.c==t.c){Qe(e.a,T);continue e}else F-=e.g[u.p]}T.e.c.length+T.g.c.length>0?(e.f[T.p]=F/(T.e.c.length+T.g.c.length),e.c=h.Math.min(e.c,e.f[T.p]),e.b=h.Math.max(e.b,e.f[T.p])):m&&(e.f[T.p]=F)}}function QMn(e){e.b=null,e.bb=null,e.fb=null,e.qb=null,e.a=null,e.c=null,e.d=null,e.e=null,e.f=null,e.n=null,e.M=null,e.L=null,e.Q=null,e.R=null,e.K=null,e.db=null,e.eb=null,e.g=null,e.i=null,e.j=null,e.k=null,e.gb=null,e.o=null,e.p=null,e.q=null,e.r=null,e.$=null,e.ib=null,e.S=null,e.T=null,e.t=null,e.s=null,e.u=null,e.v=null,e.w=null,e.B=null,e.A=null,e.C=null,e.D=null,e.F=null,e.G=null,e.H=null,e.I=null,e.J=null,e.P=null,e.Z=null,e.U=null,e.V=null,e.W=null,e.X=null,e.Y=null,e._=null,e.ab=null,e.cb=null,e.hb=null,e.nb=null,e.lb=null,e.mb=null,e.ob=null,e.pb=null,e.jb=null,e.kb=null,e.N=!1,e.O=!1}function e8n(e,t,i){var a,u,l,g;for(i.Sg("Graph transformation ("+e.a+")",1),g=rE(t.a),l=new ne(t.b);l.a<l.c.c.length;)u=f(ee(l),25),$o(g,u.a);if(a=f(V(t,(ot(),OLe)),416),a==(FH(),hte))switch(f(V(t,mf),86).g){case 2:rj(t,g);break;case 3:Ij(t,g);break;case 4:e.a==(BP(),kq)?(Ij(t,g),Zce(t,g)):(Zce(t,g),Ij(t,g))}else if(e.a==(BP(),kq))switch(f(V(t,mf),86).g){case 2:rj(t,g),Zce(t,g);break;case 3:Ij(t,g),rj(t,g);break;case 4:rj(t,g),Ij(t,g)}else switch(f(V(t,mf),86).g){case 2:rj(t,g),Zce(t,g);break;case 3:rj(t,g),Ij(t,g);break;case 4:Ij(t,g),rj(t,g)}i.Tg()}function t8n(e){var t,i,a,u,l,g,m,v;for(l=new ne(e.a.b);l.a<l.c.c.length;)u=f(ee(l),82),u.b.c=u.g.c,u.b.d=u.g.d;for(v=new Ke(Gi,Gi),t=new Ke(qo,qo),a=new ne(e.a.b);a.a<a.c.c.length;)i=f(ee(a),82),v.a=h.Math.min(v.a,i.g.c),v.b=h.Math.min(v.b,i.g.d),t.a=h.Math.max(t.a,i.g.c+i.g.b),t.b=h.Math.max(t.b,i.g.d+i.g.a);for(m=HX(e.c).a.nc();m.Ob();)g=f(m.Pb(),47),i=f(g.b,82),v.a=h.Math.min(v.a,i.g.c),v.b=h.Math.min(v.b,i.g.d),t.a=h.Math.max(t.a,i.g.c+i.g.b),t.b=h.Math.max(t.b,i.g.d+i.g.a);e.d=fX(new Ke(v.a,v.b)),e.e=Ho(new Ke(t.a,t.b),v),e.a.a.c.length=0,e.a.b.c.length=0}function n8n(e){uj();var t,i,a,u,l,g,m;for(m=new irt,i=new ne(e);i.a<i.c.c.length;)t=f(ee(i),143),(!m.b||t.c>=m.b.c)&&(m.b=t),(!m.c||t.c<=m.c.c)&&(m.d=m.c,m.c=t),(!m.e||t.d>=m.e.d)&&(m.e=t),(!m.f||t.d<=m.f.d)&&(m.f=t);return a=new xZ((qP(),a_)),SH(e,jOt,new bu(ce(se(Sq,1),Jt,374,0,[a]))),g=new xZ(YT),SH(e,MOt,new bu(ce(se(Sq,1),Jt,374,0,[g]))),u=new xZ(KT),SH(e,DOt,new bu(ce(se(Sq,1),Jt,374,0,[u]))),l=new xZ(bA),SH(e,LOt,new bu(ce(se(Sq,1),Jt,374,0,[l]))),Tde(a.c,a_),Tde(u.c,KT),Tde(l.c,bA),Tde(g.c,YT),m.a.c.length=0,$o(m.a,a.c),$o(m.a,Bd(u.c)),$o(m.a,l.c),$o(m.a,Bd(g.c)),m}function r8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;for(t.Sg(Kxt,1),F=ke(Oe(Fe(e,(wb(),fx)))),g=ke(Oe(Fe(e,(Ag(),SF)))),m=f(Fe(e,EF),107),qxe((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a)),T=N_t((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a),F,g),!e.a&&(e.a=new Be(Ur,e,10,11)),S=new ne(T);S.a<S.c.c.length;)for(v=f(ee(S),186),u=new ne(v.a);u.a<u.c.c.length;)a=f(ee(u),172),L=new fTe(a.s,a.t,ke(Oe(Fe(e,SF)))),Qxe(L,a),Qe(v.d,L);A=Ewt(T,g),z=h.Math.max(A.a,ke(Oe(Fe(e,yF)))-(m.b+m.c)),l=h.Math.max(A.b,ke(Oe(Fe(e,cne)))-(m.d+m.a)),i=l-A.b,pi(e,wF,i),pi(e,t4,z),pi(e,CD,l+i),pi(e,une,T),t.Tg()}function i8n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W;for(S=new ub,T=new ub,z=new ub,W=new ub,v=ke(Oe(V(t,(ot(),v_)))),l=ke(Oe(V(t,op))),m=new ne(i);m.a<m.c.c.length;)if(g=f(ee(m),9),A=f(V(g,(Ge(),Tc)),64),A==(tt(),hn))for(T.a.yc(g,T),u=new Tn(Cn(Fo(g).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),Go(S,a.c.i);else if(A==Bn)for(W.a.yc(g,W),u=new Tn(Cn(Fo(g).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),Go(z,a.c.i);S.a.gc()!=0&&(L=new vce(2,l),F=gOe(L,t,S,T,-v-t.c.b),F>0&&(e.a=v+(F-1)*l,t.c.b+=e.a,t.f.b+=e.a)),z.a.gc()!=0&&(L=new vce(1,l),F=gOe(L,t,z,W,t.f.b+v-t.c.b),F>0&&(t.f.b+=v+(F-1)*l))}function $_t(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;for(T=ke(Oe(V(e,(ot(),b_)))),a=ke(Oe(V(e,aDe))),L=new JO,$e(L,b_,T+a),S=t,ue=S.d,W=S.c.i,pe=S.d.i,J=$Se(W.c),ve=$Se(pe.c),u=new et,A=J;A<=ve;A++)m=new Tg(e),Wl(m,(cn(),uo)),$e(m,(Ge(),oi),S),$e(m,Xi,(oa(),uu)),$e(m,Fte,L),F=f(ct(e.b,A),25),A==J?sw(m,F.a.c.length-i,F):jo(m,F),Ee=ke(Oe(V(S,z0))),Ee<0&&(Ee=0,$e(S,z0,Ee)),m.o.b=Ee,z=h.Math.floor(Ee/2),g=new ou,No(g,(tt(),pn)),cc(g,m),g.n.b=z,v=new ou,No(v,_n),cc(v,m),v.n.b=z,ua(S,g),l=new xS,Rc(l,S),$e(l,Ls,null),Ia(l,v),ua(l,ue),EIn(m,S,l),vn(u.c,l),S=l;return u}function o8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;if(W=t.b.c.length,!(W<3)){for(F=Pe(pr,Pr,30,W,15,1),A=0,T=new ne(t.b);T.a<T.c.c.length;)S=f(ee(T),25),F[A++]=S.a.c.length;for(L=new la(t.b,2),a=1;a<W-1;a++)for(i=(Ln(L.b<L.d.gc()),f(L.d.Xb(L.c=L.b++),25)),z=new ne(i.a),l=0,m=0,v=0;v<F[a+1];v++)if(ve=f(ee(z),9),v==F[a+1]-1||TAe(e,ve,a+1,a)){for(g=F[a]-1,TAe(e,ve,a+1,a)&&(g=e.c.e[f(f(f(ct(e.c.b,ve.p),15).Xb(0),47).a,9).p]);m<=v;){if(pe=f(ct(i.a,m),9),!TAe(e,pe,a+1,a))for(ue=f(ct(e.c.b,pe.p),15).Jc();ue.Ob();)J=f(ue.Pb(),47),u=e.c.e[f(J.a,9).p],(u<l||u>g)&&Go(e.b,f(J.b,18));++m}l=g}}}function efe(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(v=f(L0(e,(tt(),pn)).Jc().Pb(),11).e,F=f(L0(e,_n).Jc().Pb(),11).g,m=v.c.length,ve=wg(f(ct(e.j,0),11));m-- >0;){for(W=(Gt(0,v.c.length),f(v.c[0],18)),u=(Gt(0,F.c.length),f(F.c[0],18)),pe=u.d.e,l=Oc(pe,u,0),C2n(W,u.d,l),Ia(u,null),ua(u,null),z=W.a,t&&Mr(z,new ja(ve)),a=tr(u.a,0);a.b!=a.d.c;)i=f(Zn(a),8),Mr(z,new ja(i));for(ue=W.b,L=new ne(u.b);L.a<L.c.c.length;)A=f(ee(L),72),vn(ue.c,A);if(J=f(V(W,(ot(),Ls)),77),g=f(V(u,Ls),77),g)for(J||(J=new fd,$e(W,Ls,J)),T=tr(g,0);T.b!=T.d.c;)S=f(Zn(T),8),Mr(J,new ja(S))}}function fOe(e,t){var i;if(t==null||Mt(t,bl)||t.length==0&&e.k!=(bE(),r4))return null;switch(e.k.g){case 1:return SZ(t,qL)?(mn(),QL):SZ(t,epe)?(mn(),yw):null;case 2:try{return rt(ff(t,Ra,Qr))}catch(a){if(a=go(a),me(a,130))return null;throw ie(a)}case 4:try{return AT(t)}catch(a){if(a=go(a),me(a,130))return null;throw ie(a)}case 3:return t;case 5:return V1t(e),bEt(e,t);case 6:return V1t(e),E6n(e,e.a,t);case 7:try{return i=I4n(e),i.ag(t),i}catch(a){if(a=go(a),me(a,32))return null;throw ie(a)}default:throw ie(new As("Invalid type set for this layout option."))}}function hOe(e){var t;switch(e.d){case 1:{if(e.Pj())return e.o!=-2;break}case 2:{if(e.Pj())return e.o==-2;break}case 3:case 5:case 4:case 6:case 7:return e.o>-2;default:return!1}switch(t=e.Oj(),e.p){case 0:return t!=null&&lt(ut(t))!=N7(e.k,0);case 1:return t!=null&&f(t,219).a!=mr(e.k)<<24>>24;case 2:return t!=null&&f(t,179).a!=(mr(e.k)&To);case 6:return t!=null&&N7(f(t,190).a,e.k);case 5:return t!=null&&f(t,17).a!=mr(e.k);case 7:return t!=null&&f(t,191).a!=mr(e.k)<<16>>16;case 3:return t!=null&&ke(Oe(t))!=e.j;case 4:return t!=null&&f(t,160).a!=e.j;default:return t==null?e.n!=null:!ei(t,e.n)}}function RG(e,t,i){var a,u,l,g;return e.ll()&&e.kl()&&(g=fce(e,f(i,57)),Ae(g)!==Ae(i))?(e.uj(t),e.Aj(t,Ogt(e,t,g)),e.Zk()&&(l=(u=f(i,52),e.jl()?e.hl()?u.Ph(e.b,os(f(Vt(fl(e.b),e.Ij()),19)).n,f(Vt(fl(e.b),e.Ij()).Ek(),29).hk(),null):u.Ph(e.b,Pi(u.zh(),os(f(Vt(fl(e.b),e.Ij()),19))),null,null):u.Ph(e.b,-1-e.Ij(),null,null)),!f(g,52).Lh()&&(l=(a=f(g,52),e.jl()?e.hl()?a.Nh(e.b,os(f(Vt(fl(e.b),e.Ij()),19)).n,f(Vt(fl(e.b),e.Ij()).Ek(),29).hk(),l):a.Nh(e.b,Pi(a.zh(),os(f(Vt(fl(e.b),e.Ij()),19))),null,l):a.Nh(e.b,-1-e.Ij(),null,l))),l&&l.lj()),Ld(e.b)&&e.Gj(e.Fj(9,i,g,t,!1)),g):i}function F_t(e){var t,i,a,u,l,g,m,v,S,T;for(a=new et,g=new ne(e.e.a);g.a<g.c.c.length;){for(u=f(ee(g),123),T=0,u.k.c.length=0,i=new ne(V3(u));i.a<i.c.c.length;)t=f(ee(i),216),t.f&&(Qe(u.k,t),++T);T==1&&vn(a.c,u)}for(l=new ne(a);l.a<l.c.c.length;)for(u=f(ee(l),123);u.k.c.length==1;){for(S=f(ee(new ne(u.k)),216),e.b[S.c]=S.g,m=S.d,v=S.e,i=new ne(V3(u));i.a<i.c.c.length;)t=f(ee(i),216),ei(t,S)||(t.f?m==t.d||v==t.e?e.b[S.c]-=e.b[t.c]-t.g:e.b[S.c]+=e.b[t.c]-t.g:u==m?t.d==u?e.b[S.c]+=t.g:e.b[S.c]-=t.g:t.d==u?e.b[S.c]-=t.g:e.b[S.c]+=t.g);hl(m.k,S),hl(v.k,S),m==u?u=S.e:u=S.d}}function B_t(e,t){var i,a,u,l,g,m,v,S,T,A,L,F;if(i=f(Es(e.b,t),126),v=f(f(si(e.r,t),22),83),v.dc()){i.n.b=0,i.n.c=0;return}for(S=e.u.Gc((Xl(),M1)),g=0,m=v.Jc(),T=null,A=0,L=0;m.Ob();)a=f(m.Pb(),115),u=ke(Oe(a.b.mf((dX(),Mee)))),l=a.b.Kf().a,e.A.Gc((Ud(),GE))&&mSt(e,t),T?(F=L+T.d.c+e.w+a.d.b,g=h.Math.max(g,(gg(),Kh(Og),h.Math.abs(A-u)<=Og||A==u||isNaN(A)&&isNaN(u)?0:F/(u-A)))):e.C&&e.C.b>0&&(g=h.Math.max(g,Rbt(e.C.b+a.d.b,u))),T=a,A=u,L=l;e.C&&e.C.c>0&&(F=L+e.C.c,S&&(F+=T.d.c),g=h.Math.max(g,(gg(),Kh(Og),h.Math.abs(A-1)<=Og||A==1||isNaN(A)&&isNaN(1)?0:F/(1-A)))),i.n.b=0,i.a.a=g}function U_t(e,t){var i,a,u,l,g,m,v,S,T,A,L,F;if(i=f(Es(e.b,t),126),v=f(f(si(e.r,t),22),83),v.dc()){i.n.d=0,i.n.a=0;return}for(S=e.u.Gc((Xl(),M1)),g=0,e.A.Gc((Ud(),GE))&&bSt(e,t),m=v.Jc(),T=null,L=0,A=0;m.Ob();)a=f(m.Pb(),115),l=ke(Oe(a.b.mf((dX(),Mee)))),u=a.b.Kf().b,T?(F=A+T.d.a+e.w+a.d.d,g=h.Math.max(g,(gg(),Kh(Og),h.Math.abs(L-l)<=Og||L==l||isNaN(L)&&isNaN(l)?0:F/(l-L)))):e.C&&e.C.d>0&&(g=h.Math.max(g,Rbt(e.C.d+a.d.d,l))),T=a,L=l,A=u;e.C&&e.C.a>0&&(F=A+e.C.a,S&&(F+=T.d.a),g=h.Math.max(g,(gg(),Kh(Og),h.Math.abs(L-1)<=Og||L==1||isNaN(L)&&isNaN(1)?0:F/(1-L)))),i.n.d=0,i.a.b=g}function z_t(e,t,i){var a,u,l,g,m,v;for(this.g=e,m=t.d.length,v=i.d.length,this.d=Pe(Ib,j0,9,m+v,0,1),g=0;g<m;g++)this.d[g]=t.d[g];for(l=0;l<v;l++)this.d[m+l]=i.d[l];if(t.e){if(this.e=cH(t.e),this.e.Kc(i),i.e)for(u=i.e.Jc();u.Ob();)a=f(u.Pb(),238),a!=t&&(this.e.Gc(a)?--a.c:this.e.Ec(a))}else i.e&&(this.e=cH(i.e),this.e.Kc(t));this.f=t.f+i.f,this.a=t.a+i.a,this.a>0?Aue(this,this.f/this.a):mg(t.g,t.d[0]).a!=null&&mg(i.g,i.d[0]).a!=null?Aue(this,(ke(mg(t.g,t.d[0]).a)+ke(mg(i.g,i.d[0]).a))/2):mg(t.g,t.d[0]).a!=null?Aue(this,mg(t.g,t.d[0]).a):mg(i.g,i.d[0]).a!=null&&Aue(this,mg(i.g,i.d[0]).a)}function a8n(e,t,i,a,u,l,g,m){var v,S,T,A,L,F,z,W,J,ue;if(z=!1,S=bIe(i.q,t.f+t.b-i.q.f),F=a.f>t.b&&m,ue=u-(i.q.e+S-g),A=(v=Zj(a,ue,!1),v.a),F&&A>a.f)return!1;if(F){for(L=0,J=new ne(t.d);J.a<J.c.c.length;)W=f(ee(J),316),L+=bIe(W,a.f)+g;ue=u-L}return ue<a.g||(T=l==e.c.length-1&&ue>=(Gt(l,e.c.length),f(e.c[l],186)).e,!F&&A>t.b&&!T)?!1:((T||F||A<=t.b)&&(T&&A>t.b?(i.d=A,IH(i,kyt(i,A))):(Bwt(i.q,S),i.c=!0),IH(a,u-(i.s+i.r)),cG(a,i.q.e+i.q.d,t.f),jJ(t,a),e.c.length>l&&(fG((Gt(l,e.c.length),f(e.c[l],186)),a),(Gt(l,e.c.length),f(e.c[l],186)).a.c.length==0&&_0(e,l)),z=!0),z)}function s8n(e,t){var i,a,u,l,g,m,v,S,T,A;for(e.a=new vdt(zTn(NF)),a=new ne(t.a);a.a<a.c.c.length;){for(i=f(ee(a),834),m=new Cle(ce(se(fge,1),Jt,82,0,[])),Qe(e.a.a,m),S=new ne(i.d);S.a<S.c.c.length;)v=f(ee(S),119),T=new l_e(e,v),mOe(T,f(V(i.c,(Ge(),R1)),22)),Yu(e.g,i)||(qr(e.g,i,new Ke(v.c,v.d)),qr(e.f,i,T)),Qe(e.a.b,T),Jce(m,T);for(g=new ne(i.b);g.a<g.c.c.length;)l=f(ee(g),586),T=new l_e(e,l.Bf()),qr(e.b,l,new Ya(m,T)),mOe(T,f(V(i.c,(Ge(),R1)),22)),l.zf()&&(A=new v3e(e,l.zf(),1),mOe(A,f(V(i.c,R1),22)),u=new Cle(ce(se(fge,1),Jt,82,0,[])),Jce(u,A),jt(e.c,l.yf(),new Ya(m,A)))}return e.a}function c8n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;if(g=i.Ik(),me(g,101)&&(f(g,19).Bb&Xa)!=0&&(L=f(i.kd(),52),W=iw(e.e,L),W!=L)){if(T=mb(g,W),j7(e,t,ide(e,t,T)),A=null,Ld(e.e)&&(a=iA((Rl(),Ea),e.e.zh(),g),a!=Vt(e.e.zh(),e.c))){for(J=Ju(e.e.zh(),g),m=0,l=f(e.g,122),v=0;v<t;++v)u=l[v],J.Zl(u.Ik())&&++m;A=new nue(e.e,9,a,L,W,m,!1),A.kj(new v1(e.e,9,e.c,i,T,t,!1))}return z=f(g,19),F=os(z),F?(A=L.Ph(e.e,Pi(L.zh(),F),null,A),A=f(W,52).Nh(e.e,Pi(W.zh(),F),null,A)):(z.Bb&Dc)!=0&&(S=-1-Pi(e.e.zh(),z),A=L.Ph(e.e,S,null,null),!f(W,52).Lh()&&(A=f(W,52).Nh(e.e,S,null,A))),A&&A.lj(),T}return i}function H_t(e){var t;this.a=e,t=(cn(),ce(se(wge,1),Ve,247,0,[Yi,uo,bo,xu,Vc,Dg,_q,U$])).length,this.b=iT(Fbe,[Ye,ORe],[585,144],0,[t,t],2),this.c=iT(Fbe,[Ye,ORe],[585,144],0,[t,t],2),$ce(this,Yi,(ot(),v_),NA),bj(this,Yi,uo,b_,jE),mH(this,Yi,xu,b_),mH(this,Yi,bo,b_),bj(this,Yi,Vc,v_,NA),$ce(this,uo,op,IA),mH(this,uo,xu,op),mH(this,uo,bo,op),bj(this,uo,Vc,b_,jE),Ist(this,xu,op),mH(this,xu,bo,op),mH(this,xu,Vc,yme),Ist(this,bo,rF),bj(this,bo,Vc,ED,yD),$ce(this,Vc,op,op),$ce(this,Dg,op,IA),bj(this,Dg,Yi,b_,jE),bj(this,Dg,Vc,b_,jE),bj(this,Dg,uo,b_,jE)}function u8n(e,t,i,a,u,l){var g;if(!(t==null||!wle(t,L7e,D7e)))throw ie(new dn("invalid scheme: "+t));if(!e&&!(i!=null&&cm(i,gl(35))==-1&&i.length>0&&(gn(0,i.length),i.charCodeAt(0)!=47)))throw ie(new dn("invalid opaquePart: "+i));if(e&&!(t!=null&&Iz(Bne,t.toLowerCase()))&&!(i==null||!wle(i,zF,HF)))throw ie(new dn(x3t+i));if(e&&t!=null&&Iz(Bne,t.toLowerCase())&&!mOn(i))throw ie(new dn(x3t+i));if(!ACn(a))throw ie(new dn("invalid device: "+a));if(!_xn(u))throw g=u==null?"invalid segments: null":"invalid segment: "+mxn(u),ie(new dn(g));if(!(l==null||cm(l,gl(35))==-1))throw ie(new dn("invalid query: "+l))}function G_t(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;if(L=new ja(e.o),ue=t.a/L.a,m=t.b/L.b,W=t.a-L.a,l=t.b-L.b,i)for(u=Ae(V(e,(ot(),Xi)))===Ae((oa(),uu)),z=new ne(e.j);z.a<z.c.c.length;)switch(F=f(ee(z),11),F.j.g){case 1:u||(F.n.a*=ue);break;case 2:F.n.a+=W,u||(F.n.b*=m);break;case 3:u||(F.n.a*=ue),F.n.b+=l;break;case 4:u||(F.n.b*=m)}for(S=new ne(e.b);S.a<S.c.c.length;)v=f(ee(S),72),T=v.n.a+v.o.a/2,A=v.n.b+v.o.b/2,J=T/L.a,g=A/L.b,J+g>=1&&(J-g>0&&A>=0?(v.n.a+=W,v.n.b+=l*g):J-g<0&&T>=0&&(v.n.a+=W*J,v.n.b+=l));e.o.a=t.a,e.o.b=t.b,$e(e,(ot(),DE),(Ud(),a=f(Tp($F),10),new Rf(a,f(Gh(a,a.length),10),0)))}function l8n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;if(i.Sg("Network simplex layering",1),e.b=t,ue=f(V(t,(ot(),iF)),17).a*4,J=e.b.a,J.c.length<1){i.Tg();return}for(l=dLn(e,J),W=null,u=tr(l,0);u.b!=u.d.c;){for(a=f(Zn(u),15),m=ue*Aa(h.Math.sqrt(a.gc())),g=kLn(a),jde(j2e(Apn(M2e(Gse(g),m),W),!0),i.bh(1)),L=e.b.b,z=new ne(g.a);z.a<z.c.c.length;){for(F=f(ee(z),123);L.c.length<=F.e;)Ky(L,L.c.length,new ru(e.b));T=f(F.f,9),jo(T,f(ct(L,F.e),25))}if(l.b>1)for(W=Pe(pr,Pr,30,e.b.b.c.length,15,1),A=0,S=new ne(e.b.b);S.a<S.c.c.length;)v=f(ee(S),25),W[A++]=v.a.c.length}J.c.length=0,e.a=null,e.b=null,e.c=null,i.Tg()}function d8n(e,t){var i,a,u,l,g,m,v,S,T,A;for(T=new et,A=new O3,l=null,u=0,a=0;a<t.length;++a)switch(i=t[a],ixn(l,i)&&(u=JCe(e,A,T,Kte,u)),li(i,(Ge(),l_))&&(l=f(V(i,l_),9)),i.k.g){case 0:for(v=BSe(WN(lc(i,(tt(),hn)),new _C));Mue(v);)g=f(pxe(v),11),e.d[g.p]=u++,vn(T.c,g);for(u=JCe(e,A,T,Kte,u),S=BSe(WN(lc(i,Bn),new _C));Mue(S);)g=f(pxe(S),11),e.d[g.p]=u++,vn(T.c,g);break;case 3:lc(i,HDe).dc()||(g=f(lc(i,HDe).Xb(0),11),e.d[g.p]=u++,vn(T.c,g)),lc(i,Kte).dc()||Jv(A,i);break;case 1:for(m=lc(i,(tt(),pn)).Jc();m.Ob();)g=f(m.Pb(),11),e.d[g.p]=u++,vn(T.c,g);lc(i,_n).Ic(new rat(A,i))}return JCe(e,A,T,Kte,u),T}function pOe(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;if(t==null||t.length==0)return null;if(l=f(wu(e.f,t),21),!l){for(u=(F=new kp(e.d).a.vc().Jc(),new zy(F));u.a.Ob();)if(i=(g=f(u.a.Pb(),43),f(g.kd(),21)),m=i.f,z=t.length,Mt(m.substr(m.length-z,z),t)&&(t.length==m.length||_a(m,m.length-t.length-1)==46)){if(l)return null;l=i}if(!l){for(a=(L=new kp(e.d).a.vc().Jc(),new zy(L));a.a.Ob();)if(i=(g=f(a.a.Pb(),43),f(g.kd(),21)),A=i.g,A!=null){for(v=A,S=0,T=v.length;S<T;++S)if(m=v[S],z=t.length,Mt(m.substr(m.length-z,z),t)&&(t.length==m.length||_a(m,m.length-t.length-1)==46)){if(l)return null;l=i}}}l&&Os(e.f,t,l)}return l}function f8n(e,t){var i,a,u,l,g;for(i=new RN,g=!1,l=0;l<t.length;l++){if(a=(gn(l,t.length),t.charCodeAt(l)),a==32){for(_Z(e,i,0),i.a+=" ",_Z(e,i,0);l+1<t.length&&(gn(l+1,t.length),t.charCodeAt(l+1)==32);)++l;continue}if(g){a==39?l+1<t.length&&(gn(l+1,t.length),t.charCodeAt(l+1)==39)?(i.a+=String.fromCharCode(a),++l):g=!1:i.a+=String.fromCharCode(a);continue}if(cm("GyMLdkHmsSEcDahKzZv",gl(a))>0){_Z(e,i,0),i.a+=String.fromCharCode(a),u=h3n(t,l),_Z(e,i,u),l+=u-1;continue}a==39?l+1<t.length&&(gn(l+1,t.length),t.charCodeAt(l+1)==39)?(i.a+="'",++l):g=!0:i.a+=String.fromCharCode(a)}_Z(e,i,0),jOn(e)}function h8n(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(ve=qut(e),v=new et,l=e.c.length,S=l-1,T=l+1;ve.a.gc()!=0;){for(;i.b!=0;)ue=(Ln(i.b!=0),f(lh(i,i.a.a),116)),ve.a.Ac(ue)!=null,ue.g=S--,eOe(ue,t,i,a);for(;t.b!=0;)pe=(Ln(t.b!=0),f(lh(t,t.a.a),116)),ve.a.Ac(pe)!=null,pe.g=T++,eOe(pe,t,i,a);for(m=Ra,W=ve.a.ec().Jc();W.Ob();){if(z=f(W.Pb(),116),!a&&z.b>0&&z.a<=0){v.c.length=0,vn(v.c,z);break}F=z.i-z.d,F>=m&&(F>m&&(v.c.length=0,m=F),vn(v.c,z))}v.c.length!=0&&(g=f(ct(v,IZ(u,v.c.length)),116),ve.a.Ac(g)!=null,g.g=T++,eOe(g,t,i,a),v.c.length=0)}for(J=e.c.length+1,L=new ne(e);L.a<L.c.c.length;)A=f(ee(L),116),A.g<l&&(A.g=A.g+J)}function q_t(e,t,i){var a,u,l,g;this.j=e,this.e=iAe(e),this.o=this.j.e,this.i=!!this.o,this.p=this.i?f(ct(i,ra(this.o).p),217):null,u=f(V(e,(Ge(),Cu)),22),this.g=u.Gc((us(),Vf)),this.b=new et,this.d=new J0t(this.e),g=f(V(this.j,ZR),233),this.q=dTn(t,g,this.e),this.k=new Bft(this),l=gm(ce(se(sNt,1),Jt,229,0,[this,this.d,this.k,this.q])),t==(PS(),zq)&&!lt(ut(V(e,(ot(),h_))))?(a=new sAe(this.e),vn(l.c,a),this.c=new ATe(a,g,f(this.q,402))):t==zq&&lt(ut(V(e,(ot(),h_))))?(a=new sAe(this.e),vn(l.c,a),this.c=new kbt(a,g,f(this.q,402))):this.c=new nat(t,this),Qe(l,this.c),A_t(l,this.e),this.s=Z9n(this.k)}function p8n(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(u=new ne(e.a.b);u.a<u.c.c.length;)for(i=f(ee(u),25),v=new ne(i.a);v.a<v.c.c.length;)m=f(ee(v),9),t.j[m.p]=m,t.i[m.p]=t.o==(vg(),Mg)?qo:Gi;for(nu(e.c),g=e.a.b,t.c==(fm(),G0)&&(g=Bd(g)),v2n(e.e,t,e.b),Pz(t.p,null),l=g.Jc();l.Ob();)for(i=f(l.Pb(),25),S=i.a,t.o==(vg(),Mg)&&(S=Bd(S)),L=S.Jc();L.Ob();)A=f(L.Pb(),9),t.g[A.p]==A&&Akt(e,A,t);for(YMn(e,t),a=g.Jc();a.Ob();)for(i=f(a.Pb(),25),L=new ne(i.a);L.a<L.c.c.length;)A=f(ee(L),9),t.p[A.p]=t.p[t.g[A.p].p],A==t.g[A.p]&&(T=ke(t.i[t.j[A.p].p]),(t.o==(vg(),Mg)&&T>qo||t.o==$E&&T<Gi)&&(t.p[A.p]=ke(t.p[A.p])+T));e.e.vg()}function g8n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;if(lt(ut(V(i,(ot(),ix)))))for(m=new ne(i.j);m.a<m.c.c.length;)for(g=f(ee(m),11),L=pm(g.g),S=L,T=0,A=S.length;T<A;++T)v=S[T],l=v.d.i==i,u=l&&lt(ut(V(v,PE))),u&&(z=v.c,F=f(bn(e.b,z),9),F||(F=PG(z,(oa(),Fg),z.j,-1,null,null,z.o,f(V(t,mf),86),t),$e(F,(Ge(),oi),z),qr(e.b,z,F),Qe(t.a,F)),J=v.d,W=f(bn(e.b,J),9),W||(W=PG(J,(oa(),Fg),J.j,1,null,null,J.o,f(V(t,mf),86),t),$e(W,(Ge(),oi),J),qr(e.b,J,W),Qe(t.a,W)),a=cue(v),Ia(a,f(ct(F.j,0),11)),ua(a,f(ct(W.j,0),11)),jt(e.a,v,new vX(a,t,(cs(),Uu))),f(V(t,(Ge(),Cu)),22).Ec((us(),Vf)))}function m8n(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J;for(l=new et,S=new ne(a);S.a<S.c.c.length;)if(m=f(ee(S),441),g=null,m.f==(cs(),Uu))for(z=new ne(m.e);z.a<z.c.c.length;)F=f(ee(z),18),J=F.d.i,ra(J)==t?Umt(e,t,m,F,m.b,F.d):!i||UP(J,i)?yNn(e,t,m,a,F):(L=Xde(e,t,i,F,m.b,Uu,g),L!=g&&vn(l.c,L),L.c&&(g=L));else for(A=new ne(m.e);A.a<A.c.c.length;)if(T=f(ee(A),18),W=T.c.i,ra(W)==t)Umt(e,t,m,T,T.c,m.b);else{if(!i||UP(W,i))continue;L=Xde(e,t,i,T,m.b,td,g),L!=g&&vn(l.c,L),L.c&&(g=L)}for(v=new ne(l);v.a<v.c.c.length;)m=f(ee(v),441),Oc(t.a,m.a,0)!=-1||Qe(t.a,m.a),m.c&&vn(u.c,m)}function b8n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;for(u=f(V(e,(_i(),mF)),27),S=Qr,T=Qr,m=Ra,v=Ra,ve=tr(e.b,0);ve.b!=ve.d.c;)ue=f(Zn(ve),39),F=ue.e,z=ue.f,S=h.Math.min(S,F.a-z.a/2),T=h.Math.min(T,F.b-z.b/2),m=h.Math.max(m,F.a+z.a/2),v=h.Math.max(v,F.b+z.b/2);for(L=f(Fe(u,(Ec(),bMe)),107),pe=tr(e.b,0);pe.b!=pe.d.c;)ue=f(Zn(pe),39),A=V(ue,mF),me(A,206)&&(l=f(A,27),Of(l,ue.e.a,ue.e.b),_G(l,ue));for(J=tr(e.a,0);J.b!=J.d.c;)W=f(Zn(J),65),a=f(V(W,mF),85),a&&(t=W.a,i=rQ(a),tOe(t,i));Ee=m-S+(L.b+L.c),g=v-T+(L.d+L.a),lt(ut(Fe(u,(Rr(),hx))))||WS(u,Ee,g,!1,!1),pi(u,BA,Ee-(L.b+L.c)),pi(u,FA,g-(L.d+L.a))}function V_t(e){var t,i,a,u,l,g,m;for(t=0,l=new ne(e.b.a);l.a<l.c.c.length;)a=f(ee(l),194),a.b=0,a.c=0;for(Syt(e,0),yle(e,e.g),tQ(e.c),N2e(e.c),i=(ko(),Ds),AG(ose(oA(AG(ose(oA(AG(oA(e.c,i)),O0t(i)))),i))),oA(e.c,Ds),ple(e,e.g),syt(e,0),x_t(e,0),s2t(e,1),Syt(e,1),yle(e,e.d),tQ(e.c),g=new ne(e.b.a);g.a<g.c.c.length;)a=f(ee(g),194),t+=h.Math.abs(a.c);for(m=new ne(e.b.a);m.a<m.c.c.length;)a=f(ee(m),194),a.b=0,a.c=0;for(i=Wf,AG(ose(oA(AG(ose(oA(AG(N2e(oA(e.c,i))),O0t(i)))),i))),oA(e.c,Ds),ple(e,e.d),syt(e,1),x_t(e,1),s2t(e,0),N2e(e.c),u=new ne(e.b.a);u.a<u.c.c.length;)a=f(ee(u),194),t+=h.Math.abs(a.c);return t}function v8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(i=f(V(e,(ot(),Xi)),100),g=e.f,l=e.d,m=g.a+l.b+l.c,v=0-l.d-e.c.b,T=g.b+l.d+l.a-e.c.b,S=new et,A=new et,u=new ne(t);u.a<u.c.c.length;){switch(a=f(ee(u),9),i.g){case 1:case 2:case 3:b5n(a);break;case 4:L=f(V(a,m_),8),F=L?L.a:0,a.n.a=m*ke(Oe(V(a,(Ge(),d_))))-F,tZ(a,!0,!1);break;case 5:z=f(V(a,m_),8),W=z?z.a:0,a.n.a=ke(Oe(V(a,(Ge(),d_))))-W,tZ(a,!0,!1),g.a=h.Math.max(g.a,a.n.a+a.o.a/2)}switch(f(V(a,(Ge(),Tc)),64).g){case 1:a.n.b=v,vn(S.c,a);break;case 3:a.n.b=T,vn(A.c,a)}}switch(i.g){case 1:case 2:g0t(S,e),g0t(A,e);break;case 3:m0t(S,e),m0t(A,e)}}function w8n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;for(i.Sg("Label dummy switching",1),a=f(V(t,(ot(),Nte)),231),iTn(t),u=u6n(t,a),e.a=Pe(sa,_s,30,t.b.c.length,15,1),m=(ER(),ce(se(sD,1),Ve,231,0,[aD,H$,oD,QT,EA,yA])),T=0,F=m.length;T<F;++T)if(l=m[T],(l==EA||l==yA||l==QT)&&!f(Cl(u.a,l)?u.b[l.g]:null,15).dc()){cTn(e,t);break}for(v=ce(se(sD,1),Ve,231,0,[aD,H$,oD,QT,EA,yA]),A=0,z=v.length;A<z;++A)l=v[A],l==EA||l==yA||l==QT||PSt(e,f(Cl(u.a,l)?u.b[l.g]:null,15));for(g=ce(se(sD,1),Ve,231,0,[aD,H$,oD,QT,EA,yA]),S=0,L=g.length;S<L;++S)l=g[S],(l==EA||l==yA||l==QT)&&PSt(e,f(Cl(u.a,l)?u.b[l.g]:null,15));e.a=null,i.Tg()}function W_t(e,t){var i,a,u,l,g,m,v,S,T;if(S=t,!(S.b==null||e.b==null)){for(eA(e),Qj(e),eA(S),Qj(S),i=Pe(pr,Pr,30,e.b.length+S.b.length,15,1),T=0,a=0,g=0;a<e.b.length&&g<S.b.length;)if(u=e.b[a],l=e.b[a+1],m=S.b[g],v=S.b[g+1],l<m)a+=2;else if(l>=m&&u<=v)m<=u&&l<=v?(i[T++]=u,i[T++]=l,a+=2):m<=u?(i[T++]=u,i[T++]=v,e.b[a]=v+1,g+=2):l<=v?(i[T++]=m,i[T++]=l,a+=2):(i[T++]=m,i[T++]=v,e.b[a]=v+1);else if(v<u)g+=2;else throw ie(new oc("Token#intersectRanges(): Internal Error: ["+e.b[a]+","+e.b[a+1]+"] & ["+S.b[g]+","+S.b[g+1]+"]"));for(;a<e.b.length;)i[T++]=e.b[a++],i[T++]=e.b[a++];e.b=Pe(pr,Pr,30,T,15,1),au(i,0,e.b,0,T)}}function y8n(e){var t,i,a,u,l,g,m;for(t=new et,e.g=new et,e.d=new et,g=new yT(new Vi(e.f.b).a);g.b;)l=G3(g),Qe(t,f(f(l.kd(),47).b,82)),l1(f(l.jd(),586).yf())?Qe(e.d,f(l.kd(),47)):Qe(e.g,f(l.kd(),47));for(yle(e,e.d),yle(e,e.g),e.c=new yEt(e.b),Opn(e.c,(Z2e(),IOt)),ple(e,e.d),ple(e,e.g),$o(t,e.c.a.b),e.e=new Ke(Gi,Gi),e.a=new Ke(qo,qo),a=new ne(t);a.a<a.c.c.length;)i=f(ee(a),82),e.e.a=h.Math.min(e.e.a,i.g.c),e.e.b=h.Math.min(e.e.b,i.g.d),e.a.a=h.Math.max(e.a.a,i.g.c+i.g.b),e.a.b=h.Math.max(e.a.b,i.g.d+i.g.a);$2e(e.c,new uy),m=0;do u=V_t(e),++m;while((m<2||u>fw)&&m<10);$2e(e.c,new ly),V_t(e),pwn(e.c),t8n(e.f)}function E8n(e,t){var i,a,u,l,g,m,v,S,T,A,L;switch(e.k.g){case 1:if(a=f(V(e,(Ge(),oi)),18),i=f(V(a,UPe),77),i?lt(ut(V(a,U0)))&&(i=vCe(i)):i=new fd,S=f(V(e,$p),11),S){if(T=uc(ce(se(Zo,1),Ye,8,0,[S.i.n,S.n,S.a])),t<=T.a)return T.b;Wi(i,T,i.a,i.a.a)}if(A=f(V(e,yh),11),A){if(L=uc(ce(se(Zo,1),Ye,8,0,[A.i.n,A.n,A.a])),L.a<=t)return L.b;Wi(i,L,i.c.b,i.c)}if(i.b>=2){for(v=tr(i,0),g=f(Zn(v),8),m=f(Zn(v),8);m.a<t&&v.b!=v.d.c;)g=m,m=f(Zn(v),8);return g.b+(t-g.a)/(m.a-g.a)*(m.b-g.b)}break;case 3:switch(l=f(V(f(ct(e.j,0),11),(Ge(),oi)),11),u=l.i,l.j.g){case 1:return u.n.b;case 3:return u.n.b+u.o.b}}return Y3e(e).b}function S8n(e){var t,i,a,u,l,g,m,v,S,T,A;for(g=new ne(e.d.b);g.a<g.c.c.length;)for(l=f(ee(g),25),v=new ne(l.a);v.a<v.c.c.length;){if(m=f(ee(v),9),lt(ut(V(m,(ot(),Z$))))&&!NP(T1(m))){a=f(wyn(T1(m)),18),T=a.c.i,T==m&&(T=a.d.i),A=new Ya(T,Ho($a(m.n),T.n)),qr(e.b,m,A);continue}u=new qh(m.n.a-m.d.b,m.n.b-m.d.d,m.o.a+m.d.b+m.d.c,m.o.b+m.d.d+m.d.a),t=mct(_ot(Eot(Sot(new f2e,m),u),oNt),e.a),gct(ign(obt(new d2e,ce(se(Lee,1),Jt,60,0,[t])),t),e.a),S=new uae,qr(e.e,t,S),i=g1(new Tn(Cn(Fo(m).a.Jc(),new B)))-g1(new Tn(Cn(ao(m).a.Jc(),new B))),i<0?YH(S,!0,(ko(),Ds)):i>0&&YH(S,!0,(ko(),Gs)),m.k==(cn(),bo)&&kdt(S),qr(e.f,m,t)}}function K_t(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe;for(S=Gi,T=Gi,m=qo,v=qo,L=new ne(t.i);L.a<L.c.c.length;)A=f(ee(L),68),u=f(f(bn(e.g,A.a),47).b,27),Of(u,A.b.c,A.b.d),S=h.Math.min(S,u.i),T=h.Math.min(T,u.j),m=h.Math.max(m,u.i+u.g),v=h.Math.max(v,u.j+u.f);for(F=f(Fe(e.c,(mG(),lLt)),107),WS(e.c,m-S+(F.b+F.c),v-T+(F.d+F.a),!0,!0),gAe(e.c,-S+F.b,-T+F.d),a=new Nn(qdt(e.c));a.e!=a.i.gc();)i=f(Mn(a),85),g=rQ(i),z=lde(i),J=lyt(i),W=new Ke(z.i+z.g/2,z.j+z.f/2),l=new Ke(J.i+J.g/2,J.j+J.f/2),ue=Ho(new Ke(l.a,l.b),W),uL(ue,z.g,z.f),ii(W,ue),pe=Ho(new Ke(W.a,W.b),l),uL(pe,J.g,J.f),ii(l,pe),Yz(g,W.a,W.b),Kz(g,l.a,l.b)}function Y_t(e,t){var i,a,u,l,g,m,v,S,T,A;for(v=!0,u=0,S=e.g[t.p],T=t.o.b+e.o,i=e.d[t.p][2],af(e.b,S,rt(f(ct(e.b,S),17).a-1+i)),af(e.c,S,ke(Oe(ct(e.c,S)))-T+i*e.f),++S,S>=e.j?(++e.j,Qe(e.b,rt(1)),Qe(e.c,T)):(a=e.d[t.p][1],af(e.b,S,rt(f(ct(e.b,S),17).a+1-a)),af(e.c,S,ke(Oe(ct(e.c,S)))+T-a*e.f)),(e.r==(gh(),Fq)&&(f(ct(e.b,S),17).a>e.k||f(ct(e.b,S-1),17).a>e.k)||e.r==Bq&&(ke(Oe(ct(e.c,S)))>e.n||ke(Oe(ct(e.c,S-1)))>e.n))&&(v=!1),g=new Tn(Cn(Fo(t).a.Jc(),new B));Kn(g);)l=f(Dn(g),18),m=l.c.i,e.g[m.p]==S&&(A=Y_t(e,m),u=u+f(A.a,17).a,v=v&&lt(ut(A.b)));return e.g[t.p]=S,u=u+e.d[t.p][0],new Ya(rt(u),(mn(),!!v))}function _8n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;return L=e.c[t],F=e.c[i],z=f(V(L,(Ge(),XR)),15),!!z&&z.gc()!=0&&z.Gc(F)||(W=L.k!=(cn(),uo)&&F.k!=uo,J=f(V(L,l_),9),ue=f(V(F,l_),9),pe=J!=ue,ve=!!J&&J!=L||!!ue&&ue!=F,Ee=Hle(L,(tt(),hn)),Le=Hle(F,Bn),ve=ve|(Hle(L,Bn)||Hle(F,hn)),He=ve&&pe||Ee||Le,W&&He)||L.k==(cn(),xu)&&F.k==Yi||F.k==(cn(),xu)&&L.k==Yi?!1:(T=e.c[t],l=e.c[i],u=Rwt(e.e,T,l,(tt(),pn)),v=Rwt(e.i,T,l,_n),s5n(e.f,T,l),S=B1t(e.b,T,l)+f(u.a,17).a+f(v.a,17).a+e.f.d,m=B1t(e.b,l,T)+f(u.b,17).a+f(v.b,17).a+e.f.b,e.a&&(A=f(V(T,oi),11),g=f(V(l,oi),11),a=mwt(e.g,A,g),S+=f(a.a,17).a,m+=f(a.b,17).a),S>m)}function X_t(e,t){var i,a,u,l,g;i=ke(Oe(V(t,(ot(),op)))),i<2&&$e(t,op,2),a=f(V(t,mf),86),a==(ko(),$g)&&$e(t,mf,dZ(t)),u=f(V(t,n6t),17),u.a==0?$e(t,(Ge(),ZR),new vle):$e(t,(Ge(),ZR),new hJ(u.a)),l=ut(V(t,tF)),l==null&&$e(t,tF,(mn(),Ae(V(t,P1))===Ae((k1(),DD)))),Zi(new Ft(null,new Bt(t.a,16)),new Hy(e)),Zi(Zs(new Ft(null,new Bt(t.b,16)),new sy),new wz(e)),g=new H_t(t),$e(t,(Ge(),CA),g),pH(e.a),Ap(e.a,(wa(),rp),f(V(t,RE),188)),Ap(e.a,Ab,f(V(t,LE),188)),Ap(e.a,su,f(V(t,Q$),188)),Ap(e.a,cu,f(V(t,jte),188)),Ap(e.a,zs,gTn(f(V(t,P1),220))),FSe(e.a,b7n(t)),$e(t,nme,NG(e.a,t))}function gOe(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue;for(A=new Wn,g=new et,qyt(e,i,e.d.yg(),g,A),qyt(e,a,e.d.zg(),g,A),e.b=.2*(W=e2t(Zs(new Ft(null,new Bt(g,16)),new ib)),J=e2t(Zs(new Ft(null,new Bt(g,16)),new R8)),h.Math.min(W,J)),l=0,m=0;m<g.c.length-1;m++)for(v=(Gt(m,g.c.length),f(g.c[m],116)),z=m+1;z<g.c.length;z++)l+=cOe(e,v,(Gt(z,g.c.length),f(g.c[z],116)));for(L=f(V(t,(Ge(),ZR)),233),l>=2&&(ue=y2t(g,!0,L),!e.e&&(e.e=new ktt(e)),a3n(e.e,ue,g,e.b)),Jwt(g,L),I8n(g),F=-1,T=new ne(g);T.a<T.c.c.length;)S=f(ee(T),116),!(h.Math.abs(S.s-S.c)<Tm)&&(F=h.Math.max(F,S.o),e.d.wg(S,u,e.c));return e.d.a.a.$b(),F+1}function k8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;for(A=f(oX((g=tr(new u1(t).a.d,0),new m3(g))),39),z=A?f(V(A,(_i(),Xme)),39):null,u=1;A&&z;){for(v=0,Ee=0,i=A,a=z,m=0;m<u;m++)i=lue(i),a=lue(a),Ee+=ke(Oe(V(i,(_i(),xD)))),v+=ke(Oe(V(a,xD)));if(ve=ke(Oe(V(z,(_i(),q0)))),pe=ke(Oe(V(A,q0))),L=HCe(e,A,z),F=ve+v+e.b+L-pe-Ee,0<F){for(S=t,T=0;S&&S!=a;)++T,S=f(V(S,Qte),39);if(S)for(ue=F/T,S=t;S!=a;)J=ke(Oe(V(S,q0)))+F,$e(S,q0,J),W=ke(Oe(V(S,xD)))+F,$e(S,xD,W),F-=ue,S=f(V(S,Qte),39);else return}++u,A.d.b==0?A=qIe(new u1(t),u):A=f(oX((l=tr(new u1(A).a.d,0),new m3(l))),39),z=A?f(V(A,Xme),39):null}}function T8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;for(t.Sg("Label dummy removal",1),a=ke(Oe(V(e,(ot(),OA)))),u=ke(Oe(V(e,cx))),S=f(V(e,mf),86),v=new ne(e.b);v.a<v.c.c.length;)for(m=f(ee(v),25),A=new la(m.a,0);A.b<A.d.gc();)T=(Ln(A.b<A.d.gc()),f(A.d.Xb(A.c=A.b++),9)),T.k==(cn(),Vc)&&(L=f(V(T,(Ge(),oi)),18),z=ke(Oe(V(L,z0))),g=Ae(V(T,JR))===Ae((df(),Nw)),i=new ja(T.n),g&&(i.b+=z+a),l=new Ke(T.o.a,T.o.b+(T.k==Vc&&!oP(Jr(f(V(T,Y$),15).Mc(),new Y(new VM))).zd((Wy(),zR))?0:-z-a)),F=f(V(T,Y$),15),S==(ko(),Wf)||S==jg?W4n(F,i,u,l,g,S):Axn(F,i,u,l),$o(L.b,F),efe(T,Ae(V(e,P1))===Ae((k1(),uV))),hd(A));t.Tg()}function x8n(e){e.q||(e.q=!0,e.p=yc(e,0),e.a=yc(e,1),Ci(e.a,0),e.f=yc(e,2),Ci(e.f,1),Ki(e.f,2),e.n=yc(e,3),Ki(e.n,3),Ki(e.n,4),Ki(e.n,5),Ki(e.n,6),e.g=yc(e,4),Ci(e.g,7),Ki(e.g,8),e.c=yc(e,5),Ci(e.c,7),Ci(e.c,8),e.i=yc(e,6),Ci(e.i,9),Ci(e.i,10),Ci(e.i,11),Ci(e.i,12),Ki(e.i,13),e.j=yc(e,7),Ci(e.j,9),e.d=yc(e,8),Ci(e.d,3),Ci(e.d,4),Ci(e.d,5),Ci(e.d,6),Ki(e.d,7),Ki(e.d,8),Ki(e.d,9),Ki(e.d,10),e.b=yc(e,9),Ki(e.b,0),Ki(e.b,1),e.e=yc(e,10),Ki(e.e,1),Ki(e.e,2),Ki(e.e,3),Ki(e.e,4),Ci(e.e,5),Ci(e.e,6),Ci(e.e,7),Ci(e.e,8),Ci(e.e,9),Ci(e.e,10),Ki(e.e,11),e.k=yc(e,11),Ki(e.k,0),Ki(e.k,1),e.o=Br(e,12),e.s=Br(e,13))}function mOe(e,t){t.dc()&&db(e.j,!0,!0,!0,!0),ei(t,(tt(),dp))&&db(e.j,!0,!0,!0,!1),ei(t,Kf)&&db(e.j,!1,!0,!0,!0),ei(t,vf)&&db(e.j,!0,!0,!1,!0),ei(t,_h)&&db(e.j,!0,!1,!0,!0),ei(t,nd)&&db(e.j,!1,!0,!0,!1),ei(t,Yf)&&db(e.j,!1,!0,!1,!0),ei(t,wf)&&db(e.j,!0,!1,!1,!0),ei(t,fp)&&db(e.j,!0,!1,!0,!1),ei(t,wl)&&db(e.j,!0,!0,!0,!0),ei(t,tl)&&db(e.j,!0,!0,!0,!0),ei(t,wl)&&db(e.j,!0,!0,!0,!0),ei(t,vl)&&db(e.j,!0,!0,!0,!0),ei(t,yl)&&db(e.j,!0,!0,!0,!0),ei(t,rd)&&db(e.j,!0,!0,!0,!0),ei(t,Au)&&db(e.j,!0,!0,!0,!0)}function J_t(e,t,i){var a,u,l,g,m,v,S,T,A;if(e.a!=t.gk())throw ie(new dn(VL+t.ve()+n_));if(a=I0((Rl(),Ea),t).Gl(),a)return a.gk().si().ni(a,i);if(g=I0(Ea,t).Il(),g){if(i==null)return null;if(m=f(i,15),m.dc())return"";for(A=new p0,l=m.Jc();l.Ob();)u=l.Pb(),ys(A,g.gk().si().ni(g,u)),A.a+=" ";return yse(A,A.a.length-1)}if(T=I0(Ea,t).Jl(),!T.dc()){for(S=T.Jc();S.Ob();)if(v=f(S.Pb(),155),v.ck(i))try{if(A=v.gk().si().ni(v,i),A!=null)return A}catch(L){if(L=go(L),!me(L,103))throw ie(L)}throw ie(new dn("Invalid value: '"+i+"' for datatype :"+t.ve()))}return f(t,826).lk(),i==null?null:me(i,179)?""+f(i,179).a:Fd(i)==Iee?Bst(UF[0],f(i,205)):Qs(i)}function C8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;for(t.Sg("Interactive crossing minimization",1),g=0,l=new ne(e.b);l.a<l.c.c.length;)a=f(ee(l),25),a.p=g++;for(L=iAe(e),J=new U2e(L.length),A_t(new bu(ce(se(sNt,1),Jt,229,0,[J])),L),W=0,g=0,u=new ne(e.b);u.a<u.c.c.length;){for(a=f(ee(u),25),i=0,A=0,T=new ne(a.a);T.a<T.c.c.length;)for(v=f(ee(T),9),v.n.a>0&&(i+=v.n.a+v.o.a/2,++A),z=new ne(v.j);z.a<z.c.c.length;)F=f(ee(z),11),F.p=W++;for(A>0&&(i/=A),ue=Pe(sa,_s,30,a.a.c.length,15,1),m=0,S=new ne(a.a);S.a<S.c.c.length;)v=f(ee(S),9),v.p=m++,ue[v.p]=E8n(v,i),v.k==(cn(),uo)&&$e(v,(Ge(),zPe),ue[v.p]);Ht(),na(a.a,new ntt(ue)),jSt(J,L,g,!0),++g}t.Tg()}function A8n(e){var t,i,a,u,l,g,m,v,S,T;for(S=new ki,m=new ki,l=new ne(e);l.a<l.c.c.length;)a=f(ee(l),132),a.v=0,a.n=a.i.c.length,a.u=a.t.c.length,a.n==0&&Wi(S,a,S.c.b,S.c),a.u==0&&a.r.a.gc()==0&&Wi(m,a,m.c.b,m.c);for(g=-1;S.b!=0;)for(a=f(Zle(S,0),132),i=new ne(a.t);i.a<i.c.c.length;)t=f(ee(i),270),T=t.b,T.v=h.Math.max(T.v,a.v+1),g=h.Math.max(g,T.v),--T.n,T.n==0&&Wi(S,T,S.c.b,S.c);if(g>-1){for(u=tr(m,0);u.b!=u.d.c;)a=f(Zn(u),132),a.v=g;for(;m.b!=0;)for(a=f(Zle(m,0),132),i=new ne(a.i);i.a<i.c.c.length;)t=f(ee(i),270),v=t.a,v.r.a.gc()==0&&(v.v=h.Math.min(v.v,a.v-1),--v.u,v.u==0&&Wi(m,v,m.c.b,m.c))}}function I8n(e){var t,i,a,u,l,g,m,v,S,T;for(S=new et,m=new et,g=new ne(e);g.a<g.c.c.length;)u=f(ee(g),116),yN(u,u.f.c.length),aS(u,u.k.c.length),u.d==0&&vn(S.c,u),u.i==0&&u.e.b==0&&vn(m.c,u);for(a=-1;S.c.length!=0;)for(u=f(_0(S,0),116),i=new ne(u.k);i.a<i.c.c.length;)t=f(ee(i),133),T=t.b,K5(T,h.Math.max(T.o,u.o+1)),a=h.Math.max(a,T.o),yN(T,T.d-1),T.d==0&&vn(S.c,T);if(a>-1){for(l=new ne(m);l.a<l.c.c.length;)u=f(ee(l),116),u.o=a;for(;m.c.length!=0;)for(u=f(_0(m,0),116),i=new ne(u.f);i.a<i.c.c.length;)t=f(ee(i),133),v=t.a,!(v.e.b>0)&&(K5(v,h.Math.min(v.o,u.o-1)),aS(v,v.i-1),v.i==0&&vn(m.c,v))}}function Z_t(e,t,i,a,u){var l,g,m,v;return v=Gi,g=!1,m=aOe(e,Ho(new Ke(t.a,t.b),e),ii(new Ke(i.a,i.b),u),Ho(new Ke(a.a,a.b),i)),l=!!m&&!(h.Math.abs(m.a-e.a)<=e_&&h.Math.abs(m.b-e.b)<=e_||h.Math.abs(m.a-t.a)<=e_&&h.Math.abs(m.b-t.b)<=e_),m=aOe(e,Ho(new Ke(t.a,t.b),e),i,u),m&&((h.Math.abs(m.a-e.a)<=e_&&h.Math.abs(m.b-e.b)<=e_)==(h.Math.abs(m.a-t.a)<=e_&&h.Math.abs(m.b-t.b)<=e_)||l?v=h.Math.min(v,Z7(Ho(m,i))):g=!0),m=aOe(e,Ho(new Ke(t.a,t.b),e),a,u),m&&(g||(h.Math.abs(m.a-e.a)<=e_&&h.Math.abs(m.b-e.b)<=e_)==(h.Math.abs(m.a-t.a)<=e_&&h.Math.abs(m.b-t.b)<=e_)||l)&&(v=h.Math.min(v,Z7(Ho(m,a)))),v}function Q_t(e){Wk(e,new zS(gY(qk(zk(Gk(Hk(new ug,gw),VTt),"Minimizes the stress within a layout using stress majorization. Stress exists if the euclidean distance between a pair of nodes doesn't match their graph theoretic distance, that is, the shortest path between the two nodes. The method allows to specify individual edge lengths."),new i6),Zu))),Je(e,gw,l$,st(c5e)),Je(e,gw,IQ,(mn(),!0)),Je(e,gw,uA,st(yOt)),Je(e,gw,DR,st(EOt)),Je(e,gw,LR,st(SOt)),Je(e,gw,$L,st(wOt)),Je(e,gw,d$,st(l5e)),Je(e,gw,FL,st(_Ot)),Je(e,gw,tNe,st(s5e)),Je(e,gw,rNe,st(o5e)),Je(e,gw,iNe,st(a5e)),Je(e,gw,oNe,st(u5e)),Je(e,gw,nNe,st(Uee))}function O8n(e){var t,i,a,u,l,g,m,v;for(t=null,a=new ne(e);a.a<a.c.c.length;)i=f(ee(a),238),ke(mg(i.g,i.d[0]).a),i.b=null,i.e&&i.e.gc()>0&&i.c==0&&(!t&&(t=new et),vn(t.c,i));if(t)for(;t.c.length!=0;){if(i=f(_0(t,0),238),i.b&&i.b.c.length>0){for(l=(!i.b&&(i.b=new et),new ne(i.b));l.a<l.c.c.length;)if(u=f(ee(l),238),_z(mg(u.g,u.d[0]).a)==_z(mg(i.g,i.d[0]).a)){if(Oc(e,u,0)>Oc(e,i,0))return new Ya(u,i)}else if(ke(mg(u.g,u.d[0]).a)>ke(mg(i.g,i.d[0]).a))return new Ya(u,i)}for(m=(!i.e&&(i.e=new et),i.e).Jc();m.Ob();)g=f(m.Pb(),238),v=(!g.b&&(g.b=new et),g.b),fT(0,v.c.length),y7(v.c,0,i),g.c==v.c.length&&vn(t.c,g)}return null}function t$(e,t){var i,a,u,l,g,m,v,S,T;if(t.e==5){W_t(e,t);return}if(S=t,!(S.b==null||e.b==null)){for(eA(e),Qj(e),eA(S),Qj(S),i=Pe(pr,Pr,30,e.b.length+S.b.length,15,1),T=0,a=0,g=0;a<e.b.length&&g<S.b.length;)if(u=e.b[a],l=e.b[a+1],m=S.b[g],v=S.b[g+1],l<m)i[T++]=e.b[a++],i[T++]=e.b[a++];else if(l>=m&&u<=v)m<=u&&l<=v?a+=2:m<=u?(e.b[a]=v+1,g+=2):l<=v?(i[T++]=u,i[T++]=m-1,a+=2):(i[T++]=u,i[T++]=m-1,e.b[a]=v+1,g+=2);else if(v<u)g+=2;else throw ie(new oc("Token#subtractRanges(): Internal Error: ["+e.b[a]+","+e.b[a+1]+"] - ["+S.b[g]+","+S.b[g+1]+"]"));for(;a<e.b.length;)i[T++]=e.b[a++],i[T++]=e.b[a++];e.b=Pe(pr,Pr,30,T,15,1),au(i,0,e.b,0,T)}}function N8n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;for(a=new ne(e.e.b);a.a<a.c.c.length;)for(i=f(ee(a),25),l=new ne(i.a);l.a<l.c.c.length;)if(u=f(ee(l),9),F=e.i[u.p],S=F.a.e,v=F.d.e,u.n.b=S,ue=v-S-u.o.b,t=Yde(u),L=(xT(),(u.q?u.q:(Ht(),Ht(),Cb))._b((ot(),g_))?A=f(V(u,g_),203):A=f(V(ra(u),nF),203),A),t&&(L==PA||L==RA)&&(u.o.b+=ue),t&&(L==Ame||L==PA||L==RA)){for(W=new ne(u.j);W.a<W.c.c.length;)z=f(ee(W),11),(tt(),tl).Gc(z.j)&&(T=f(bn(e.k,z),123),z.n.b=T.e-S);for(m=new ne(u.b);m.a<m.c.c.length;)g=f(ee(m),72),J=f(V(u,p_),22),J.Gc((NT(),Sh))?g.n.b+=ue:J.Gc(Pb)&&(g.n.b+=ue/2);(L==PA||L==RA)&&lc(u,(tt(),Bn)).Ic(new vtt(ue))}}function R8n(e){var t,i,a,u,l,g,m;if(!e.A.dc()){if(e.A.Gc((Ud(),mV))&&(f(Es(e.b,(tt(),hn)),126).k=!0,f(Es(e.b,Bn),126).k=!0,t=e.q!=(oa(),Lb)&&e.q!=uu,Lv(f(Es(e.b,_n),126),t),Lv(f(Es(e.b,pn),126),t),Lv(e.g,t),e.A.Gc(GE)&&(f(Es(e.b,hn),126).j=!0,f(Es(e.b,Bn),126).j=!0,f(Es(e.b,_n),126).k=!0,f(Es(e.b,pn),126).k=!0,e.g.k=!0)),e.A.Gc(gV))for(e.a.j=!0,e.a.k=!0,e.g.j=!0,e.g.k=!0,m=e.B.Gc((yd(),FF)),u=Ole(),l=0,g=u.length;l<g;++l)a=u[l],i=f(Es(e.i,a),315),i&&(h3e(a)?(i.j=!0,i.k=!0):(i.j=!m,i.k=!m));e.A.Gc(mx)&&e.B.Gc((yd(),vV))&&(e.g.j=!0,e.g.j=!0,e.a.j||(e.a.j=!0,e.a.k=!0,e.a.e=!0))}}function P8n(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(S=new et,v=new ne(t.a);v.a<v.c.c.length;)for(g=f(ee(v),9),L=lc(g,(tt(),_n)).Jc();L.Ob();)for(A=f(L.Pb(),11),u=new ne(A.g);u.a<u.c.c.length;)a=f(ee(u),18),!(!Fa(a)&&a.c.i.c==a.d.i.c||Fa(a)||a.d.i.c!=i)&&vn(S.c,a);for(m=Bd(i.a).Jc();m.Ob();)for(g=f(m.Pb(),9),L=lc(g,(tt(),pn)).Jc();L.Ob();)for(A=f(L.Pb(),11),u=new ne(A.e);u.a<u.c.c.length;)if(a=f(ee(u),18),!(!Fa(a)&&a.c.i.c==a.d.i.c||Fa(a)||a.c.i.c!=t)&&S.c.length!=0){for(T=new la(S,S.c.length),l=(Ln(T.b>0),f(T.a.Xb(T.c=--T.b),18));l!=a&&T.b>0;)e.a[l.p]=!0,e.a[a.p]=!0,l=(Ln(T.b>0),f(T.a.Xb(T.c=--T.b),18));T.b>0&&hd(T)}}function ekt(e,t,i){var a,u,l,g,m,v,S,T,A,L;if(i)for(a=-1,T=new la(t,0);T.b<T.d.gc();){if(m=(Ln(T.b<T.d.gc()),f(T.d.Xb(T.c=T.b++),9)),A=e.c[m.c.p][m.p].a,A==null){for(g=a+1,l=new la(t,T.b);l.b<l.d.gc();)if(L=kmn(e,(Ln(l.b<l.d.gc()),f(l.d.Xb(l.c=l.b++),9))).a,L!=null){g=(an(L),L);break}A=(a+g)/2,e.c[m.c.p][m.p].a=A,e.c[m.c.p][m.p].d=(an(A),A),e.c[m.c.p][m.p].b=1}a=(an(A),A)}else{for(u=0,S=new ne(t);S.a<S.c.c.length;)m=f(ee(S),9),e.c[m.c.p][m.p].a!=null&&(u=h.Math.max(u,ke(e.c[m.c.p][m.p].a)));for(u+=2,v=new ne(t);v.a<v.c.c.length;)m=f(ee(v),9),e.c[m.c.p][m.p].a==null&&(A=wd(e.i,24)*VG*u-1,e.c[m.c.p][m.p].a=A,e.c[m.c.p][m.p].d=A,e.c[m.c.p][m.p].b=1)}}function L8n(e){var t,i,a,u,l,g,m,v,S,T,A,L;for(u=new ne(e.a.a.b);u.a<u.c.c.length;)for(a=f(ee(u),60),v=a.c.Jc();v.Ob();)m=f(v.Pb(),60),a.a!=m.a&&(l1(e.a.d)?A=e.a.g.df(a,m):A=e.a.g.ef(a,m),l=a.b.a+a.d.b+A-m.b.a,l=h.Math.ceil(l),l=h.Math.max(0,l),gxe(a,m)?(g=mS(new qy,e.d),S=Aa(h.Math.ceil(m.b.a-a.b.a)),t=S-(m.b.a-a.b.a),T=iR(a).a,i=a,T||(T=iR(m).a,t=-t,i=m),T&&(i.b.a-=t,T.n.a-=t),Zh(Uh(Bh(zh(Fh(new ah,h.Math.max(0,S)),1),g),e.c[a.a.d])),Zh(Uh(Bh(zh(Fh(new ah,h.Math.max(0,-S)),1),g),e.c[m.a.d]))):(L=1,(me(a.g,152)&&me(m.g,9)||me(m.g,152)&&me(a.g,9))&&(L=2),Zh(Uh(Bh(zh(Fh(new ah,Aa(l)),L),e.c[a.a.d]),e.c[m.a.d]))))}function tkt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F;if(!e.b)return!1;for(g=null,L=null,v=new pue(null,null),u=1,v.a[1]=e.b,A=v;A.a[u];)S=u,m=L,L=A,A=A.a[u],a=e.a.Le(t,A.d),u=a<0?0:1,a==0&&(!i.c||Pu(A.e,i.d))&&(g=A),!(A&&A.b)&&!Uk(A.a[u])&&(Uk(A.a[1-u])?L=L.a[S]=VJ(A,u):Uk(A.a[1-u])||(F=L.a[1-S],F&&(!Uk(F.a[1-S])&&!Uk(F.a[S])?(L.b=!1,F.b=!0,A.b=!0):(l=m.a[1]==L?1:0,Uk(F.a[S])?m.a[l]=Cht(L,S):Uk(F.a[1-S])&&(m.a[l]=VJ(L,S)),A.b=m.a[l].b=!0,m.a[l].a[0].b=!1,m.a[l].a[1].b=!1))));return g&&(i.b=!0,i.d=g.e,A!=g&&(T=new pue(A.d,A.e),sNn(e,v,g,T),L==g&&(L=T)),L.a[L.a[1]==A?1:0]=A.a[A.a[0]?0:1],--e.c),e.b=v.a[1],e.b&&(e.b.b=!1),i.b}function D8n(e,t,i){var a,u,l,g,m,v,S,T,A;for(!i&&(i=MTn(t.q.getTimezoneOffset())),u=(t.q.getTimezoneOffset()-i.a)*6e4,m=new v_e(Ua(Nc(t.q.getTime()),u)),v=m,m.q.getTimezoneOffset()!=t.q.getTimezoneOffset()&&(u>0?u-=864e5:u+=864e5,v=new v_e(Ua(Nc(t.q.getTime()),u))),T=new RN,S=e.a.length,l=0;l<S;)if(a=_a(e.a,l),a>=97&&a<=122||a>=65&&a<=90){for(g=l+1;g<S&&_a(e.a,g)==a;++g);R7n(T,a,g-l,m,v,i),l=g}else if(a==39){if(++l,l<S&&_a(e.a,l)==39){T.a+="'",++l;continue}for(A=!1;!A;){for(g=l;g<S&&_a(e.a,g)!=39;)++g;if(g>=S)throw ie(new dn("Missing trailing '"));g+1<S&&_a(e.a,g+1)==39?++g:A=!0,xr(T,uh(e.a,l,g)),l=g+1}}else T.a+=String.fromCharCode(a),++l;return T.a}function M8n(){Dr(wx,new as),Dr(Ir,new v5),Dr(kh,new Dy),Dr(Up,new Ok),Dr(m1e,new _5),Dr(Une,new $h),Dr(j1,new f9),Dr(BF,new rm),Dr(EV,new s9),Dr(l1e,new c9),Dr(Dw,new kk),Dr(Th,new u9),Dr(Bp,new QO),Dr(S_,new l9),Dr(yx,new HU),Dr(El,new m5),Dr(vx,new zc),Dr(xc,new Nu),Dr(gs,new $s),Dr(nl,new Hc),Dr(to,new nm),Dr(se($l,1),new Ry),Dr(BR,new Tk),Dr(D$,new b5),Dr(Iee,new Rv),Dr(fje,new ql),Dr(mo,new Py),Dr(C7e,new eS),Dr(O7e,new sb),Dr(Q7e,new xk),Dr(zne,new QC),Dr(eD,new fg),Dr(ba,new w5),Dr(d6e,new y5),Dr(i_,new Ck),Dr(f6e,new e3),Dr(Y7e,new Ly),Dr(hje,new ih),Dr(o_,new Ak),Dr(ft,new Ik),Dr(I7e,new E5),Dr(pje,new S5)}function nkt(e,t){var i,a,u,l,g,m,v,S,T;if(e==null)return bl;if(v=t.a.yc(e,t),v!=null)return"[...]";for(i=new uE(Fu,"[","]"),u=e,l=0,g=u.length;l<g;++l)a=u[l],a!=null&&(Fd(a).i&4)!=0?Array.isArray(a)&&(T=$H(a),!(T>=14&&T<=16))?t.a._b(a)?(i.a?xr(i.a,i.b):i.a=new ef(i.d),P7(i.a,"[...]")):(m=uR(a),S=new H7(t),b1(i,nkt(m,S))):me(a,183)?b1(i,FRn(f(a,183))):me(a,195)?b1(i,kOn(f(a,195))):me(a,201)?b1(i,PNn(f(a,201))):me(a,2057)?b1(i,TOn(f(a,2057))):me(a,54)?b1(i,$Rn(f(a,54))):me(a,579)?b1(i,ZRn(f(a,579))):me(a,825)?b1(i,jRn(f(a,825))):me(a,108)&&b1(i,MRn(f(a,108))):b1(i,a==null?bl:Qs(a));return i.a?i.e.length==0?i.a.a:i.a.a+(""+i.e):i.c}function _L(e,t){var i,a,u,l;l=e.F,t==null?(e.F=null,YP(e,null)):(e.F=(an(t),t),a=cm(t,gl(60)),a!=-1?(u=(ga(0,a,t.length),t.substr(0,a)),cm(t,gl(46))==-1&&!Mt(u,xR)&&!Mt(u,x$)&&!Mt(u,hee)&&!Mt(u,C$)&&!Mt(u,A$)&&!Mt(u,I$)&&!Mt(u,O$)&&!Mt(u,N$)&&(u=F3t),i=sX(t,gl(62)),i!=-1&&(u+=""+(gn(i+1,t.length+1),t.substr(i+1))),YP(e,u)):(u=t,cm(t,gl(46))==-1&&(a=cm(t,gl(91)),a!=-1&&(u=(ga(0,a,t.length),t.substr(0,a))),!Mt(u,xR)&&!Mt(u,x$)&&!Mt(u,hee)&&!Mt(u,C$)&&!Mt(u,A$)&&!Mt(u,I$)&&!Mt(u,O$)&&!Mt(u,N$)?(u=F3t,a!=-1&&(u+=""+(gn(a,t.length+1),t.substr(a)))):u=t),YP(e,u),u==t&&(e.F=e.D))),(e.Db&4)!=0&&(e.Db&1)==0&&Zr(e,new Jo(e,1,5,l,t))}function j8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;if(e.c=e.e,z=ut(V(t,(ot(),r6t))),F=z==null||(an(z),z),l=f(V(t,(Ge(),Cu)),22).Gc((us(),Vf)),u=f(V(t,Xi),100),i=!(u==(oa(),HE)||u==Lb||u==uu),F&&(i||!l)){for(A=new ne(t.a);A.a<A.c.c.length;)S=f(ee(A),9),S.p=0;for(L=new et,T=new ne(t.a);T.a<T.c.c.length;)if(S=f(ee(T),9),a=HSt(e,S,null),a){for(v=new Rxe,Rc(v,t),$e(v,R1,f(a.b,22)),rTe(v.d,t.d),$e(v,mD,null),m=f(a.a,15).Jc();m.Ob();)g=f(m.Pb(),9),Qe(v.a,g),g.a=v;L.Ec(v)}l&&(Ae(V(t,nx))===Ae((hE(),mge))?e.c=e.b:Ae(V(t,nx))===Ae(bge)?e.c=e.d:e.c=e.a)}else L=new bu(ce(se(UOt,1),KTt,36,0,[t]));return Ae(V(t,nx))!==Ae((hE(),XT))&&(Ht(),L.gd(new Y_)),L}function rkt(e,t){var i,a,u,l,g,m,v,S,T,A;if(v=t.length-1,m=(gn(v,t.length),t.charCodeAt(v)),m==93){if(g=cm(t,gl(91)),g>=0)return u=SCn(e,(ga(1,g,t.length),t.substr(1,g-1))),T=(ga(g+1,v,t.length),t.substr(g+1,v-(g+1))),u7n(e,T,u)}else{if(i=-1,S6e==null&&(S6e=new RegExp("\\d")),S6e.test(String.fromCharCode(m))&&(i=U_e(t,gl(46),v-1),i>=0)){a=f(sue(e,Bmt(e,(ga(1,i,t.length),t.substr(1,i-1))),!1),61),S=0;try{S=ff((gn(i+1,t.length+1),t.substr(i+1)),Ra,Qr)}catch(L){throw L=go(L),me(L,130)?(l=L,ie(new xJ(l))):ie(L)}if(S<a.gc())return A=a.Xb(S),me(A,74)&&(A=f(A,74).kd()),f(A,57)}if(i<0)return f(sue(e,Bmt(e,(gn(1,t.length+1),t.substr(1))),!1),57)}return null}function $8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(t.Sg("Label dummy insertions",1),A=new et,g=ke(Oe(V(e,(ot(),OA)))),S=ke(Oe(V(e,cx))),T=f(V(e,mf),86),F=new ne(e.a);F.a<F.c.c.length;)for(L=f(ee(F),9),l=new Tn(Cn(ao(L).a.Jc(),new B));Kn(l);)if(u=f(Dn(l),18),u.c.i!=u.d.i&&Ase(u.b,XOt)){for(W=ekn(u),z=lb(u.b.c.length),i=C6n(e,u,W,z),vn(A.c,i),a=i.o,m=new la(u.b,0);m.b<m.d.gc();)v=(Ln(m.b<m.d.gc()),f(m.d.Xb(m.c=m.b++),72)),Ae(V(v,Rm))===Ae((Sg(),PD))&&(T==(ko(),Wf)||T==jg?(a.a+=v.o.a+S,a.b=h.Math.max(a.b,v.o.b)):(a.a=h.Math.max(a.a,v.o.a),a.b+=v.o.b+S),vn(z.c,v),hd(m));T==(ko(),Wf)||T==jg?(a.a-=S,a.b+=g+W):a.b+=g-S+W}$o(e.a,A),t.Tg()}function kL(e,t,i){var a,u,l,g,m,v,S;if(S=e.c,!t&&(t=F7e),e.c=t,(e.Db&4)!=0&&(e.Db&1)==0&&(v=new Jo(e,1,2,S,e.c),i?i.kj(v):i=v),S!=t){if(me(e.Cb,290))e.Db>>16==-10?i=f(e.Cb,290).Vk(t,i):e.Db>>16==-15&&(!t&&(t=(Ut(),Ug)),!S&&(S=(Ut(),Ug)),e.Cb.Uh()&&(v=new v1(e.Cb,1,13,S,t,P0(gd(f(e.Cb,62)),e),!1),i?i.kj(v):i=v));else if(me(e.Cb,88))e.Db>>16==-23&&(me(t,88)||(t=(Ut(),Ch)),me(S,88)||(S=(Ut(),Ch)),e.Cb.Uh()&&(v=new v1(e.Cb,1,10,S,t,P0(iu(f(e.Cb,29)),e),!1),i?i.kj(v):i=v));else if(me(e.Cb,443))for(m=f(e.Cb,829),g=(!m.b&&(m.b=new iY(new bae)),m.b),l=(a=new yT(new Vi(g.a).a),new oY(a));l.a.b;)u=f(G3(l.a).jd(),87),i=kL(u,JZ(u,m),i)}return i}function F8n(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(g=lt(ut(Fe(e,(ot(),ix)))),L=f(Fe(e,ax),22),v=!1,S=!1,A=new Nn((!e.c&&(e.c=new Be(_d,e,9,9)),e.c));A.e!=A.i.gc()&&(!v||!S);){for(l=f(Mn(A),124),m=0,u=pb(Df(ce(se(Gf,1),Jt,20,0,[(!l.d&&(l.d=new Qt(vo,l,8,5)),l.d),(!l.e&&(l.e=new Qt(vo,l,7,4)),l.e)])));Kn(u)&&(a=f(Dn(u),85),T=g&&HS(a)&&lt(ut(Fe(a,PE))),i=M_t((!a.b&&(a.b=new Qt(Yn,a,4,7)),a.b),l)?e==so(Us(f(be((!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c),0),84))):e==so(Us(f(be((!a.b&&(a.b=new Qt(Yn,a,4,7)),a.b),0),84))),!((T||i)&&(++m,m>1))););(m>0||L.Gc((Xl(),M1))&&(!l.n&&(l.n=new Be(hc,l,1,7)),l.n).i>0)&&(v=!0),m>1&&(S=!0)}v&&t.Ec((us(),Vf)),S&&t.Ec((us(),V$))}function ikt(e){var t,i,a,u,l,g,m,v,S,T,A,L;if(L=f(Fe(e,(Rr(),UE)),22),L.dc())return null;if(m=0,g=0,L.Gc((Ud(),mV))){for(T=f(Fe(e,OF),100),a=2,i=2,u=2,l=2,t=so(e)?f(Fe(so(e),BE),86):f(Fe(e,BE),86),S=new Nn((!e.c&&(e.c=new Be(_d,e,9,9)),e.c));S.e!=S.i.gc();)if(v=f(Mn(S),124),A=f(Fe(v,qA),64),A==(tt(),fc)&&(A=QIe(v,t),pi(v,qA,A)),T==(oa(),uu))switch(A.g){case 1:a=h.Math.max(a,v.i+v.g);break;case 2:i=h.Math.max(i,v.j+v.f);break;case 3:u=h.Math.max(u,v.i+v.g);break;case 4:l=h.Math.max(l,v.j+v.f)}else switch(A.g){case 1:a+=v.g+2;break;case 2:i+=v.f+2;break;case 3:u+=v.g+2;break;case 4:l+=v.f+2}m=h.Math.max(a,u),g=h.Math.max(i,l)}return WS(e,m,g,!0,!0)}function B8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(u=null,a=new ne(t.a);a.a<a.c.c.length;)i=f(ee(a),9),Yde(i)?l=(m=mS(Oz(new qy,i),e.f),v=mS(Oz(new qy,i),e.f),S=new nTe(i,!0,m,v),T=i.o.b,A=(xT(),(i.q?i.q:(Ht(),Ht(),Cb))._b((ot(),g_))?L=f(V(i,g_),203):L=f(V(ra(i),nF),203),L),F=1e4,A==RA&&(F=1),z=Zh(Uh(Bh(Fh(zh(new ah,F),Aa(h.Math.ceil(T))),m),v)),A==PA&&Go(e.d,z),USt(e,Bd(lc(i,(tt(),pn))),S),USt(e,lc(i,_n),S),S):l=(W=mS(Oz(new qy,i),e.f),Zi(Jr(new Ft(null,new Bt(i.j,16)),new A8),new oat(e,W)),new nTe(i,!1,W,W)),e.i[i.p]=l,u&&(g=u.c.d.a+A3(e.n,u.c,i)+i.d.d,u.b||(g+=u.c.o.b),Zh(Uh(Bh(zh(Fh(new ah,Aa(h.Math.ceil(g))),0),u.d),l.a))),u=l}function tfe(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;for(pe=f($d(RJ(Jr(new Ft(null,new Bt(t.d,16)),new Get(i)),new qet(i)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),A=Qr,T=Ra,v=new ne(t.b.j);v.a<v.c.c.length;)m=f(ee(v),11),m.j==i&&(A=h.Math.min(A,m.p),T=h.Math.max(T,m.p));if(A==Qr)for(g=0;g<pe.gc();g++)cxe(f(pe.Xb(g),105),i,g);else for(ve=Pe(pr,Pr,30,u.length,15,1),Uvn(ve,ve.length),ue=pe.Jc();ue.Ob();){for(J=f(ue.Pb(),105),l=f(bn(e.b,J),183),S=0,W=A;W<=T;W++)l[W]&&(S=h.Math.max(S,a[W]));if(J.i){for(F=J.i.c,Ee=new Mo,L=0;L<u.length;L++)u[F][L]&&Go(Ee,rt(ve[L]));for(;xp(Ee,rt(S));)++S}for(cxe(J,i,S),z=A;z<=T;z++)l[z]&&(a[z]=S+1);J.i&&(ve[J.i.c]=S)}}function U8n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F;for(l=new _Et(t),A=TPn(e,t,l),F=h.Math.max(ke(Oe(V(t,(ot(),z0)))),1),T=new ne(A.a);T.a<T.c.c.length;)S=f(ee(T),47),v=yvt(f(S.a,8),f(S.b,8),F),b=!0,b=b&vS(i,new Ke(v.c,v.d)),b=b&vS(i,eT(new Ke(v.c,v.d),v.b,0)),b=b&vS(i,eT(new Ke(v.c,v.d),0,v.a)),b&vS(i,eT(new Ke(v.c,v.d),v.b,v.a));switch(L=l.d,m=yvt(f(A.b.a,8),f(A.b.b,8),F),L==(tt(),pn)||L==_n?(a.c[L.g]=h.Math.min(a.c[L.g],m.d),a.b[L.g]=h.Math.max(a.b[L.g],m.d+m.a)):(a.c[L.g]=h.Math.min(a.c[L.g],m.c),a.b[L.g]=h.Math.max(a.b[L.g],m.c+m.b)),u=qo,g=l.c.i.d,L.g){case 4:u=g.c;break;case 2:u=g.b;break;case 1:u=g.a;break;case 3:u=g.d}return a.a[L.g]=h.Math.max(a.a[L.g],u),l}function z8n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F;if(l=f(V(e,(Ge(),oi)),85),!!l){for(a=e.a,u=new ja(i),ii(u,qIn(e)),UP(e.d.i,e.c.i)?(L=e.c,A=uc(ce(se(Zo,1),Ye,8,0,[L.n,L.a])),Ho(A,i)):A=wg(e.c),Wi(a,A,a.a,a.a.a),F=wg(e.d),V(e,ime)!=null&&ii(F,f(V(e,ime),8)),Wi(a,F,a.c.b,a.c),wT(a,u),g=rQ(l),zJ(g,f(be((!l.b&&(l.b=new Qt(Yn,l,4,7)),l.b),0),84)),HJ(g,f(be((!l.c&&(l.c=new Qt(Yn,l,5,8)),l.c),0),84)),tOe(a,g),T=new ne(e.b);T.a<T.c.c.length;)S=f(ee(T),72),m=f(V(S,oi),153),RS(m,S.o.a),NS(m,S.o.b),Of(m,S.n.a+u.a,S.n.b+u.b),pi(m,(dR(),_ge),ut(V(S,_ge)));v=f(V(e,(ot(),Ls)),77),v?(wT(v,u),pi(l,Ls,v)):pi(l,Ls,null),t==(k1(),c4)?pi(l,P1,c4):pi(l,P1,null)}}function H8n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe;for(F=t.c.length,L=0,A=new ne(e.b);A.a<A.c.c.length;)if(T=f(ee(A),25),ue=T.a,ue.c.length!=0){for(J=new ne(ue),S=0,pe=null,u=f(ee(J),9),l=null;u;){if(l=f(ct(t,u.p),261),l.c>=0){for(v=null,m=new la(T.a,S+1);m.b<m.d.gc()&&(g=(Ln(m.b<m.d.gc()),f(m.d.Xb(m.c=m.b++),9)),v=f(ct(t,g.p),261),!(v.d==l.d&&v.c<l.c));)v=null;v&&(pe&&(af(a,u.p,rt(f(ct(a,u.p),17).a-1)),f(ct(i,pe.p),15).Kc(l)),l=OOn(l,u,F++),vn(t.c,l),Qe(i,new et),pe?(f(ct(i,pe.p),15).Ec(l),Qe(a,rt(1))):Qe(a,rt(0)))}z=null,J.a<J.c.c.length&&(z=f(ee(J),9),W=f(ct(t,z.p),261),f(ct(i,u.p),15).Ec(W),af(a,z.p,rt(f(ct(a,z.p),17).a+1))),l.d=L,l.c=S++,pe=u,u=z}++L}}function G8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;m=f(bn(t.c,e),454),pe=t.a.c,v=t.a.c+t.a.b,Ct=m.f,Kt=m.a,g=Ct<Kt,W=new Ke(pe,Ct),ve=new Ke(v,Kt),u=(pe+v)/2,J=new Ke(u,Ct),Ee=new Ke(u,Kt),l=J6n(e,Ct,Kt),He=wg(t.B),nt=new Ke(u,l),xt=wg(t.D),i=FTn(ce(se(Zo,1),Ye,8,0,[He,nt,xt])),F=!1,ue=t.B.i,ue&&ue.c&&m.d&&(S=g&&ue.p<ue.c.a.c.length-1||!g&&ue.p>0,S?S&&(L=ue.p,g?++L:--L,A=f(ct(ue.c.a,L),9),a=v1t(A),F=!(C2t(a,He,i[0])||Xlt(a,He,i[0]))):F=!0),z=!1,Le=t.D.i,Le&&Le.c&&m.e&&(T=g&&Le.p>0||!g&&Le.p<Le.c.a.c.length-1,T?(L=Le.p,g?--L:++L,A=f(ct(Le.c.a,L),9),a=v1t(A),z=!(C2t(a,i[0],xt)||Xlt(a,i[0],xt))):z=!0),F&&z&&Mr(e.a,nt),F||dj(e.a,ce(se(Zo,1),Ye,8,0,[W,J])),z||dj(e.a,ce(se(Zo,1),Ye,8,0,[Ee,ve]))}function okt(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt;if(i.c.length!=0){for(z=new et,F=new ne(i);F.a<F.c.c.length;)L=f(ee(F),27),Qe(z,new Ke(L.i,L.j));for(a.ah(t,"Before removing overlaps");vAe(e,i);)YZ(e,i,!1);if(a.ah(t,"After removing overlaps"),m=0,v=0,u=null,i.c.length!=0&&(u=(Gt(0,i.c.length),f(i.c[0],27)),m=u.i-(Gt(0,z.c.length),f(z.c[0],8)).a,v=u.j-(Gt(0,z.c.length),f(z.c[0],8)).b),g=h.Math.sqrt(m*m+v*v),A=xxn(i),l=1,A.a.gc()!=0){for(T=A.a.ec().Jc();T.Ob();)S=f(T.Pb(),27),W=e.f,J=W.i+W.g/2,ue=W.j+W.f/2,pe=S.i+S.g/2,ve=S.j+S.f/2,Ee=pe-J,Le=ve-ue,He=h.Math.sqrt(Ee*Ee+Le*Le),nt=Ee/He,xt=Le/He,md(S,S.i+nt*g),bd(S,S.j+xt*g);a.ah(t,"Child movement "+l),++l}e.a&&e.a.Eg(new Kl(A)),okt(e,t,new Kl(A),a)}}function nfe(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;return v=e,T=Ho(new Ke(t.a,t.b),e),S=i,A=Ho(new Ke(a.a,a.b),i),L=v.a,J=v.b,z=S.a,pe=S.b,F=T.a,ue=T.b,W=A.a,ve=A.b,u=W*ue-F*ve,gg(),Kh(Cm),h.Math.abs(0-u)<=Cm||u==0||isNaN(0)&&isNaN(u)?!1:(g=1/u*((L-z)*ue-(J-pe)*F),m=1/u*-(-(L-z)*ve+(J-pe)*W),l=(Kh(Cm),(h.Math.abs(0-g)<=Cm||g==0||isNaN(0)&&isNaN(g)?0:0<g?-1:0>g?1:pS(isNaN(0),isNaN(g)))<0&&(Kh(Cm),(h.Math.abs(g-1)<=Cm||g==1||isNaN(g)&&isNaN(1)?0:g<1?-1:g>1?1:pS(isNaN(g),isNaN(1)))<0)&&(Kh(Cm),(h.Math.abs(0-m)<=Cm||m==0||isNaN(0)&&isNaN(m)?0:0<m?-1:0>m?1:pS(isNaN(0),isNaN(m)))<0)&&(Kh(Cm),(h.Math.abs(m-1)<=Cm||m==1||isNaN(m)&&isNaN(1)?0:m<1?-1:m>1?1:pS(isNaN(m),isNaN(1)))<0)),l)}function q8n(e){var t,i,a,u;if(i=e.D!=null?e.D:e.B,t=cm(i,gl(91)),t!=-1){a=(ga(0,t,i.length),i.substr(0,t)),u=new p0;do u.a+="[";while((t=fP(i,91,++t))!=-1);Mt(a,xR)?u.a+="Z":Mt(a,x$)?u.a+="B":Mt(a,hee)?u.a+="C":Mt(a,C$)?u.a+="D":Mt(a,A$)?u.a+="F":Mt(a,I$)?u.a+="I":Mt(a,O$)?u.a+="J":Mt(a,N$)?u.a+="S":(u.a+="L",u.a+=""+a,u.a+=";");try{return null}catch(l){if(l=go(l),!me(l,63))throw ie(l)}}else if(cm(i,gl(46))==-1){if(Mt(i,xR))return kd;if(Mt(i,x$))return $l;if(Mt(i,hee))return Xf;if(Mt(i,C$))return sa;if(Mt(i,A$))return kx;if(Mt(i,I$))return pr;if(Mt(i,O$))return x_;if(Mt(i,N$))return JA}return null}function V8n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;for(A=new qke(new XK(e));A.c!=A.d.a.d;)for(T=amt(A),m=f(T.d,57),t=f(T.e,57),g=m.zh(),W=0,Ee=(g.i==null&&Em(g),g.i).length;W<Ee;++W)if(S=(l=(g.i==null&&Em(g),g.i),W>=0&&W<l.length?l[W]:null),S.ok()&&!S.pk()){if(me(S,101))v=f(S,19),(v.Bb&Dc)==0&&(He=os(v),!(He&&(He.Bb&Dc)!=0))&&PDn(e,v,m,t);else if(is(),f(S,69).uk()&&(i=(Le=S,f(Le?f(t,52).ci(Le):null,159)),i))for(F=f(m.Ih(S),159),a=i.gc(),J=0,z=F.gc();J<z;++J)if(L=F.Ql(J),me(L,101)){if(ve=F.Rl(J),u=Eg(e,ve),u==null&&ve!=null){if(pe=f(L,19),!e.b||(pe.Bb&Dc)!=0||os(pe))continue;u=ve}if(!i.Ll(L,u)){for(ue=0;ue<a;++ue)if(i.Ql(ue)==L&&Ae(i.Rl(ue))===Ae(u)){i.Qi(i.gc()-1,ue),--a;break}}}else i.Ll(F.Ql(J),F.Rl(J))}}function W8n(e,t,i){var a;if(i.Sg("StretchWidth layering",1),t.a.c.length==0){i.Tg();return}for(e.c=t,e.t=0,e.u=0,e.i=Gi,e.g=qo,e.d=ke(Oe(V(t,(ot(),op)))),H3n(e),M4n(e),D4n(e),BIn(e),XCn(e),e.i=h.Math.max(1,e.i),e.g=h.Math.max(1,e.g),e.d=e.d/e.i,e.f=e.g/e.i,e.s=sAn(e),a=new ru(e.c),Qe(e.c.b,a),e.r=rE(e.p),e.n=rJ(e.k,e.k.length);e.r.c.length!=0;)e.o=Sxn(e),!e.o||k1t(e)&&e.b.a.gc()!=0?(hOn(e,a),a=new ru(e.c),Qe(e.c.b,a),Ba(e.a,e.b),e.b.a.$b(),e.t=e.u,e.u=0):k1t(e)?(e.c.b.c.length=0,a=new ru(e.c),Qe(e.c.b,a),e.t=0,e.u=0,e.b.a.$b(),e.a.a.$b(),++e.f,e.r=rE(e.p),e.n=rJ(e.k,e.k.length)):(jo(e.o,a),hl(e.r,e.o),Go(e.b,e.o),e.t=e.t-e.k[e.o.p]*e.d+e.j[e.o.p],e.u+=e.e[e.o.p]*e.d);t.a.c.length=0,vG(t.b),i.Tg()}function K8n(e){var t,i,a,u,l,g,m,v,S,T,A;for(e.j=Pe(pr,Pr,30,e.g,15,1),e.o=new et,Zi(Zs(new Ft(null,new Bt(e.e.b,16)),new F2),new wtt(e)),e.a=Pe(kd,Tb,30,e.b,16,1),tG(new Ft(null,new Bt(e.e.b,16)),new Ett(e)),a=(A=new et,Zi(Jr(Zs(new Ft(null,new Bt(e.e.b,16)),new ky),new ytt(e)),new aat(e,A)),A),v=new ne(a);v.a<v.c.c.length;)if(m=f(ee(v),499),!(m.c.length<=1)){if(m.c.length==2){t4n(m),Yde((Gt(0,m.c.length),f(m.c[0],18)).d.i)||Qe(e.o,m);continue}if(!(SOn(m)||lRn(m,new gk)))for(S=new ne(m),u=null;S.a<S.c.c.length;)t=f(ee(S),18),i=e.c[t.p],!u||S.a>=S.c.c.length?T=DTe((cn(),Yi),uo):T=DTe((cn(),uo),uo),T*=2,l=i.a.g,i.a.g=h.Math.max(l,l+(T-l)),g=i.b.g,i.b.g=h.Math.max(g,g+(T-g)),u=t}}function gQ(e,t){var i;if(e.e)throw ie(new As((d1(tge),zfe+tge.k+Hfe)));if(!Sgn(e.a,t))throw ie(new oc(kTt+t+TTt));if(t==e.d)return e;switch(i=e.d,e.d=t,i.g){case 0:switch(t.g){case 2:BS(e);break;case 1:rw(e),BS(e);break;case 4:J3(e),BS(e);break;case 3:J3(e),rw(e),BS(e)}break;case 2:switch(t.g){case 1:rw(e),Rde(e);break;case 4:J3(e),BS(e);break;case 3:J3(e),rw(e),BS(e)}break;case 1:switch(t.g){case 2:rw(e),Rde(e);break;case 4:rw(e),J3(e),BS(e);break;case 3:rw(e),J3(e),rw(e),BS(e)}break;case 4:switch(t.g){case 2:J3(e),BS(e);break;case 1:J3(e),rw(e),BS(e);break;case 3:rw(e),Rde(e)}break;case 3:switch(t.g){case 2:rw(e),J3(e),BS(e);break;case 1:rw(e),J3(e),rw(e),BS(e);break;case 4:rw(e),Rde(e)}}return e}function oA(e,t){var i;if(e.d)throw ie(new As((d1(gge),zfe+gge.k+Hfe)));if(!Egn(e.a,t))throw ie(new oc(kTt+t+TTt));if(t==e.c)return e;switch(i=e.c,e.c=t,i.g){case 0:switch(t.g){case 2:dE(e);break;case 1:nw(e),dE(e);break;case 4:Z3(e),dE(e);break;case 3:Z3(e),nw(e),dE(e)}break;case 2:switch(t.g){case 1:nw(e),Pde(e);break;case 4:Z3(e),dE(e);break;case 3:Z3(e),nw(e),dE(e)}break;case 1:switch(t.g){case 2:nw(e),Pde(e);break;case 4:nw(e),Z3(e),dE(e);break;case 3:nw(e),Z3(e),nw(e),dE(e)}break;case 4:switch(t.g){case 2:Z3(e),dE(e);break;case 1:Z3(e),nw(e),dE(e);break;case 3:nw(e),Pde(e)}break;case 3:switch(t.g){case 2:nw(e),Z3(e),dE(e);break;case 1:nw(e),Z3(e),nw(e),dE(e);break;case 4:nw(e),Pde(e)}}return e}function Y8n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(A=e.b,T=new la(A,0),nT(T,new ru(e)),pe=!1,g=1;T.b<T.d.gc();){for(S=(Ln(T.b<T.d.gc()),f(T.d.Xb(T.c=T.b++),25)),W=(Gt(g,A.c.length),f(A.c[g],25)),J=rE(S.a),ue=J.c.length,z=new ne(J);z.a<z.c.c.length;)L=f(ee(z),9),jo(L,W);if(pe){for(F=Bd(J).Jc();F.Ob();)for(L=f(F.Pb(),9),l=new ne(rE(Fo(L)));l.a<l.c.c.length;)u=f(ee(l),18),VS(u,!0),$e(e,(Ge(),Nq),(mn(),!0)),a=$_t(e,u,ue),i=f(V(L,u_),314),ve=f(ct(a,a.c.length-1),18),i.k=ve.c.i,i.n=ve,i.b=u.d.i,i.c=u;pe=!1}else J.c.length!=0&&(t=(Gt(0,J.c.length),f(J.c[0],9)),t.k==(cn(),Dg)&&(pe=!0,g=-1));++g}for(m=new la(e.b,0);m.b<m.d.gc();)v=(Ln(m.b<m.d.gc()),f(m.d.Xb(m.c=m.b++),25)),v.a.c.length==0&&hd(m)}function mQ(e,t){var i,a,u,l,g,m,v,S;if(me(e.Ah(),173)?(mQ(f(e.Ah(),173),t),t.a+=" > "):t.a+="Root ",i=e.zh().zb,Mt(i.substr(0,3),"Elk")?xr(t,(gn(3,i.length+1),i.substr(3))):t.a+=""+i,u=e.hh(),u){xr((t.a+=" ",t),u);return}if(me(e,359)&&(S=f(e,153).a,S)){xr((t.a+=" ",t),S);return}for(g=new Nn(e.ih());g.e!=g.i.gc();)if(l=f(Mn(g),153),S=l.a,S){xr((t.a+=" ",t),S);return}if(me(e,268)&&(a=f(e,85),!a.b&&(a.b=new Qt(Yn,a,4,7)),a.b.i!=0&&(!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c.i!=0))){for(t.a+=" (",m=new HN((!a.b&&(a.b=new Qt(Yn,a,4,7)),a.b));m.e!=m.i.gc();)m.e>0&&(t.a+=Fu),mQ(f(Mn(m),173),t);for(t.a+=Xfe,v=new HN((!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c));v.e!=v.i.gc();)v.e>0&&(t.a+=Fu),mQ(f(Mn(v),173),t);t.a+=")"}}function X8n(e,t,i){var a,u,l,g,m,v,S,T;for(v=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));v.e!=v.i.gc();)for(m=f(Mn(v),27),u=new Tn(Cn(lw(m).a.Jc(),new B));Kn(u);){if(a=f(Dn(u),85),!a.b&&(a.b=new Qt(Yn,a,4,7)),!(a.b.i<=1&&(!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c.i<=1)))throw ie(new ON("Graph must not contain hyperedges."));if(!Uj(a)&&m!=Us(f(be((!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c),0),84)))for(S=new eut,Rc(S,a),$e(S,(ew(),qR),a),hz(S,f(ac(Is(i.f,m)),151)),W5(S,f(bn(i,Us(f(be((!a.c&&(a.c=new Qt(Yn,a,5,8)),a.c),0),84))),151)),Qe(t.c,S),g=new Nn((!a.n&&(a.n=new Be(hc,a,1,7)),a.n));g.e!=g.i.gc();)l=f(Mn(g),153),T=new rpt(S,l.a),Rc(T,l),$e(T,qR,l),T.e.a=h.Math.max(l.g,1),T.e.b=h.Math.max(l.f,1),oOe(T),Qe(t.d,T)}}function J8n(e,t,i){var a,u,l,g,m,v,S,T,A,L;switch(i.Sg("Node promotion heuristic",1),e.i=t,e.r=f(V(t,(ot(),xw)),242),e.r!=(gh(),w_)&&e.r!=ux?_9n(e):oPn(e),T=f(V(e.i,qLe),17).a,l=new uO,e.r.g){case 2:case 1:SL(e,l);break;case 3:for(e.r=Vte,SL(e,l),v=0,m=new ne(e.b);m.a<m.c.c.length;)g=f(ee(m),17),v=h.Math.max(v,g.a);v>e.k&&(e.r=Fq,SL(e,l));break;case 4:for(e.r=Vte,SL(e,l),S=0,u=new ne(e.c);u.a<u.c.c.length;)a=Oe(ee(u)),S=h.Math.max(S,(an(a),a));S>e.n&&(e.r=Bq,SL(e,l));break;case 6:L=Aa(h.Math.ceil(e.g.length*T/100)),SL(e,new wet(L));break;case 5:A=Aa(h.Math.ceil(e.e*T/100)),SL(e,new yet(A));break;case 8:Bkt(e,!0);break;case 9:Bkt(e,!1);break;default:SL(e,l)}e.r!=w_&&e.r!=ux?A5n(e,t):VPn(e,t),i.Tg()}function Z8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(A=new wOe(e),Zyn(A,!(t==(ko(),Wf)||t==jg)),T=A.a,L=new xN,u=(Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])),g=0,v=u.length;g<v;++g)i=u[g],S=Ose(T,_c,i),S&&(L.d=h.Math.max(L.d,S.ff()));for(a=ce(se(WT,1),Ve,236,0,[_c,Bu,kc]),l=0,m=a.length;l<m;++l)i=a[l],S=Ose(T,kc,i),S&&(L.a=h.Math.max(L.a,S.ff()));for(W=ce(se(WT,1),Ve,236,0,[_c,Bu,kc]),ue=0,ve=W.length;ue<ve;++ue)F=W[ue],S=Ose(T,F,_c),S&&(L.b=h.Math.max(L.b,S.gf()));for(z=ce(se(WT,1),Ve,236,0,[_c,Bu,kc]),J=0,pe=z.length;J<pe;++J)F=z[J],S=Ose(T,F,kc),S&&(L.c=h.Math.max(L.c,S.gf()));return L.d>0&&(L.d+=T.n.d,L.d+=T.d),L.a>0&&(L.a+=T.n.a,L.a+=T.d),L.b>0&&(L.b+=T.n.b,L.b+=T.d),L.c>0&&(L.c+=T.n.c,L.c+=T.d),L}function akt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z;for(L=i.d,A=i.c,l=new Ke(i.f.a+i.d.b+i.d.c,i.f.b+i.d.d+i.d.a),g=l.b,S=new ne(e.a);S.a<S.c.c.length;)if(m=f(ee(S),9),m.k==(cn(),bo)){switch(a=f(V(m,(Ge(),Tc)),64),u=f(V(m,FPe),8),T=m.n,a.g){case 2:T.a=i.f.a+L.c-A.a;break;case 4:T.a=-A.a-L.b}switch(z=0,a.g){case 2:case 4:t==(oa(),Lb)?(F=ke(Oe(V(m,d_))),T.b=l.b*F-f(V(m,(ot(),m_)),8).b,z=T.b+u.b,tZ(m,!1,!0)):t==uu&&(T.b=ke(Oe(V(m,d_)))-f(V(m,(ot(),m_)),8).b,z=T.b+u.b,tZ(m,!1,!0))}g=h.Math.max(g,z)}for(i.f.b+=g-l.b,v=new ne(e.a);v.a<v.c.c.length;)if(m=f(ee(v),9),m.k==(cn(),bo))switch(a=f(V(m,(Ge(),Tc)),64),T=m.n,a.g){case 1:T.b=-A.b-L.d;break;case 3:T.b=i.f.b+L.a-A.b}}function Q8n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;if(T=f(f(si(e.r,t),22),83),T.gc()<=2||t==(tt(),_n)||t==(tt(),pn)){xkt(e,t);return}for(W=e.u.Gc((Xl(),u4)),i=t==(tt(),hn)?(DS(),yq):(DS(),wq),ue=t==hn?(pl(),np):(pl(),jp),a=F2e(V_e(i),e.s),J=t==hn?Gi:qo,S=T.Jc();S.Ob();)m=f(S.Pb(),115),!(!m.c||m.c.d.c.length<=0)&&(z=m.b.Kf(),F=m.e,A=m.c,L=A.i,L.b=(l=A.n,A.e.a+l.b+l.c),L.a=(g=A.n,A.e.b+g.d+g.a),W?(L.c=F.a-(u=A.n,A.e.a+u.b+u.c)-e.s,W=!1):L.c=F.a+z.a+e.s,bH(ue,WOe),A.f=ue,Op(A,(Yl(),tp)),Qe(a.d,new cce(L,BCe(a,L))),J=t==hn?h.Math.min(J,F.b):h.Math.max(J,F.b+m.b.Kf().b));for(J+=t==hn?-e.t:e.t,r3e((a.e=J,a)),v=T.Jc();v.Ob();)m=f(v.Pb(),115),!(!m.c||m.c.d.c.length<=0)&&(L=m.c.i,L.c-=m.e.a,L.d-=m.e.b)}function e9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;for(gPn(e),a=1,T=new et,i=0;i<e.b.c.length;i++)if(t=oRn(f(ct(e.b,i),25)),W=PAn(f(ct(e.b,i),25)),f(ct(e.b,i),25).a.c.length>t){for(J=new et,Qe(J,f(ct(e.b,i),25)),l=0;l<t-1;l++)v=new ru(e),Qe(T,new Ya(v,rt(i+l+a))),vn(J.c,v);for(a+=t-1,z=(Gt(0,J.c.length),f(J.c[0],25)).a.c.length,u=0,L=0,ue=0;u<z;++u,++L,ue++)A=f(ct((Gt(0,J.c.length),f(J.c[0],25)).a,L),9),A.k!=(cn(),U$)?L+=q9n(e,J,ue%t,L):(u-=1,ue-=1),W&&A.k==uo&&(ue=-1)}for(S=new ne(T);S.a<S.c.c.length;)v=f(ee(S),47),Ky(e.b,f(v.b,17).a,f(v.a,25));for(m=new ne(e.b);m.a<m.c.c.length;)for(g=f(ee(m),25),F=new la(g.a,0);F.b<F.d.gc();)A=(Ln(F.b<F.d.gc()),f(F.d.Xb(F.c=F.b++),9)),(A.k==(cn(),_q)||A.k==U$)&&hd(F)}function skt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(u=new et,W=new ne(t.a);W.a<W.c.c.length;)if(z=f(ee(W),9),F=z.e,F&&(a=skt(e,F,z),$o(u,a),g8n(e,F,z),f(V(F,(Ge(),Cu)),22).Gc((us(),Vf))))for(pe=f(V(z,(ot(),Xi)),100),L=f(V(z,ax),181).Gc((Xl(),M1)),ue=new ne(z.j);ue.a<ue.c.c.length;)for(J=f(ee(ue),11),l=f(bn(e.b,J),9),l||(l=PG(J,pe,J.j,-(J.e.c.length-J.g.c.length),null,new ha,J.o,f(V(F,mf),86),F),$e(l,oi,J),qr(e.b,J,l),Qe(F.a,l)),g=f(ct(l.j,0),11),T=new ne(J.f);T.a<T.c.c.length;)S=f(ee(T),72),m=new mit,m.o.a=S.o.a,m.o.b=S.o.b,Qe(g.f,m),L||(ve=J.j,A=0,K7(f(V(z,ax),22))&&(A=HAe(S.n,S.o,J.o,0,ve)),pe==(oa(),Fg)||(tt(),tl).Gc(ve)?m.o.a=A:m.o.b=A);return v=new et,m8n(e,t,i,u,v),i&&LMn(e,t,i,v),v}function bOe(e,t,i){var a,u,l,g,m,v,S,T,A;if(!e.c[t.c.p][t.p].e){for(e.c[t.c.p][t.p].e=!0,e.c[t.c.p][t.p].b=0,e.c[t.c.p][t.p].d=0,e.c[t.c.p][t.p].a=null,T=new ne(t.j);T.a<T.c.c.length;)for(S=f(ee(T),11),A=i?new Z5(S):new _N(S),v=A.Jc();v.Ob();)m=f(v.Pb(),11),g=m.i,g.c==t.c?g!=t&&(bOe(e,g,i),e.c[t.c.p][t.p].b+=e.c[g.c.p][g.p].b,e.c[t.c.p][t.p].d+=e.c[g.c.p][g.p].d):(e.c[t.c.p][t.p].d+=e.g[m.p],++e.c[t.c.p][t.p].b);if(l=f(V(t,(Ge(),DPe)),15),l)for(u=l.Jc();u.Ob();)a=f(u.Pb(),9),t.c==a.c&&(bOe(e,a,i),e.c[t.c.p][t.p].b+=e.c[a.c.p][a.p].b,e.c[t.c.p][t.p].d+=e.c[a.c.p][a.p].d);e.c[t.c.p][t.p].b>0&&(e.c[t.c.p][t.p].d+=wd(e.i,24)*VG*.07000000029802322-.03500000014901161,e.c[t.c.p][t.p].a=e.c[t.c.p][t.p].d/e.c[t.c.p][t.p].b)}}function t9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(z=new ne(e);z.a<z.c.c.length;){for(F=f(ee(z),9),w0(F.n),w0(F.o),txe(F.f),_St(F),OPn(F),J=new ne(F.j);J.a<J.c.c.length;){for(W=f(ee(J),11),w0(W.n),w0(W.a),w0(W.o),No(W,Kvt(W.j)),l=f(V(W,(ot(),H0)),17),l&&$e(W,H0,rt(-l.a)),u=new ne(W.g);u.a<u.c.c.length;){for(a=f(ee(u),18),i=tr(a.a,0);i.b!=i.d.c;)t=f(Zn(i),8),w0(t);if(v=f(V(a,Ls),77),v)for(m=tr(v,0);m.b!=m.d.c;)g=f(Zn(m),8),w0(g);for(A=new ne(a.b);A.a<A.c.c.length;)S=f(ee(A),72),w0(S.n),w0(S.o)}for(L=new ne(W.f);L.a<L.c.c.length;)S=f(ee(L),72),w0(S.n),w0(S.o)}for(F.k==(cn(),bo)&&($e(F,(Ge(),Tc),Kvt(f(V(F,Tc),64))),$5n(F)),T=new ne(F.b);T.a<T.c.c.length;)S=f(ee(T),72),_St(S),w0(S.o),w0(S.n)}}function n9n(e){tT();var t,i,a,u,l,g,m;for(a=e.f.n,g=Bke(e.r).a.nc();g.Ob();){if(l=f(g.Pb(),115),u=0,l.b.nf((Rr(),Sd))&&(u=ke(Oe(l.b.mf(Sd))),u<0))switch(l.b.$f().g){case 1:a.d=h.Math.max(a.d,-u);break;case 3:a.a=h.Math.max(a.a,-u);break;case 2:a.c=h.Math.max(a.c,-u);break;case 4:a.b=h.Math.max(a.b,-u)}if(K7(e.u))switch(t=Jkn(l.b,u),m=!f(e.e.mf(Ow),181).Gc((yd(),bV)),i=!1,l.b.$f().g){case 1:i=t>a.d,a.d=h.Math.max(a.d,t),m&&i&&(a.d=h.Math.max(a.d,a.a),a.a=a.d+u);break;case 3:i=t>a.a,a.a=h.Math.max(a.a,t),m&&i&&(a.a=h.Math.max(a.a,a.d),a.d=a.a+u);break;case 2:i=t>a.c,a.c=h.Math.max(a.c,t),m&&i&&(a.c=h.Math.max(a.b,a.c),a.b=a.c+u);break;case 4:i=t>a.b,a.b=h.Math.max(a.b,t),m&&i&&(a.b=h.Math.max(a.b,a.c),a.c=a.b+u)}}}function ckt(e,t){var i,a,u,l,g,m,v,S,T;return S="",t.length==0?e.le(DOe,hfe,-1,-1):(T=IT(t),Mt(T.substr(0,3),"at ")&&(T=(gn(3,T.length+1),T.substr(3))),T=T.replace(/\[.*?\]/g,""),g=T.indexOf("("),g==-1?(g=T.indexOf("@"),g==-1?(S=T,T=""):(S=IT((gn(g+1,T.length+1),T.substr(g+1))),T=IT((ga(0,g,T.length),T.substr(0,g))))):(i=T.indexOf(")",g),S=(ga(g+1,i,T.length),T.substr(g+1,i-(g+1))),T=IT((ga(0,g,T.length),T.substr(0,g)))),g=cm(T,gl(46)),g!=-1&&(T=(gn(g+1,T.length+1),T.substr(g+1))),(T.length==0||Mt(T,"Anonymous function"))&&(T=hfe),m=sX(S,gl(58)),u=U_e(S,gl(58),m-1),v=-1,a=-1,l=DOe,m!=-1&&u!=-1&&(l=(ga(0,u,S.length),S.substr(0,u)),v=Sct((ga(u+1,m,S.length),S.substr(u+1,m-(u+1)))),a=Sct((gn(m+1,S.length+1),S.substr(m+1)))),e.le(l,T,v,a))}function r9n(e){var t,i,a,u,l,g,m,v,S,T,A;for(S=new ne(e);S.a<S.c.c.length;){switch(v=f(ee(S),9),g=f(V(v,(ot(),el)),171),l=null,g.g){case 1:case 2:l=(lE(),SA);break;case 3:case 4:l=(lE(),lD)}if(l)$e(v,(Ge(),Ste),(lE(),SA)),l==lD?nQ(v,g,(cs(),td)):l==SA&&nQ(v,g,(cs(),Uu));else if(T3(f(V(v,Xi),100))&&v.j.c.length!=0){for(t=!0,A=new ne(v.j);A.a<A.c.c.length;){if(T=f(ee(A),11),!(T.j==(tt(),_n)&&T.e.c.length-T.g.c.length>0||T.j==pn&&T.e.c.length-T.g.c.length<0)){t=!1;break}for(u=new ne(T.g);u.a<u.c.c.length;)if(i=f(ee(u),18),m=f(V(i.d.i,el),171),m==(Ff(),J$)||m==tx){t=!1;break}for(a=new ne(T.e);a.a<a.c.c.length;)if(i=f(ee(a),18),m=f(V(i.c.i,el),171),m==(Ff(),X$)||m==Tw){t=!1;break}}t&&nQ(v,g,(cs(),sF))}}}function i9n(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;for(He=0,F=0,A=new ne(t.e);A.a<A.c.c.length;){for(T=f(ee(A),9),L=0,m=0,v=i?f(V(T,Yte),17).a:Ra,ue=a?f(V(T,Xte),17).a:Ra,S=h.Math.max(v,ue),ve=new ne(T.j);ve.a<ve.c.c.length;){if(pe=f(ee(ve),11),Ee=T.n.b+pe.n.b+pe.a.b,a)for(g=new ne(pe.g);g.a<g.c.c.length;)l=f(ee(g),18),W=l.d,z=W.i,t!=e.a[z.p]&&(J=h.Math.max(f(V(z,Yte),17).a,f(V(z,Xte),17).a),Le=f(V(l,(ot(),e4)),17).a,Le>=S&&Le>=J&&(L+=z.n.b+W.n.b+W.a.b-Ee,++m));if(i)for(g=new ne(pe.e);g.a<g.c.c.length;)l=f(ee(g),18),W=l.c,z=W.i,t!=e.a[z.p]&&(J=h.Math.max(f(V(z,Yte),17).a,f(V(z,Xte),17).a),Le=f(V(l,(ot(),e4)),17).a,Le>=S&&Le>=J&&(L+=z.n.b+W.n.b+W.a.b-Ee,++m))}m>0&&(He+=L/m,++F)}F>0?(t.a=u*He/F,t.g=F):(t.a=0,t.g=0)}function ukt(e,t,i,a){var u,l,g,m,v;return m=new wOe(t),l5n(m,a),u=!0,e&&e.nf((Rr(),BE))&&(l=f(e.mf((Rr(),BE)),86),u=l==(ko(),$g)||l==Ds||l==Gs),lSt(m,!1),Mu(m.e.Pf(),new Y_e(m,!1,u)),Uce(m,m.f,(Np(),_c),(tt(),hn)),Uce(m,m.f,kc,Bn),Uce(m,m.g,_c,pn),Uce(m,m.g,kc,_n),Pvt(m,hn),Pvt(m,Bn),Ldt(m,_n),Ldt(m,pn),tT(),g=m.A.Gc((Ud(),mx))&&m.B.Gc((yd(),vV))?V0t(m):null,g&&Npn(m.a,g),n9n(m),jAn(m),$An(m),R8n(m),QLn(m),cIn(m),Ale(m,hn),Ale(m,Bn),GPn(m),_Mn(m),i&&(NCn(m),uIn(m),Ale(m,_n),Ale(m,pn),v=m.B.Gc((yd(),FF)),Zyt(m,v,hn),Zyt(m,v,Bn),Qyt(m,v,_n),Qyt(m,v,pn),Zi(new Ft(null,new Bt(new kp(m.i),0)),new hu),Zi(Jr(new Ft(null,Bke(m.r).a.oc()),new Qp),new Kb),yOn(m),m.e.Nf(m.o),Zi(new Ft(null,Bke(m.r).a.oc()),new S2)),m.o}function o9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(S=Gi,a=new ne(e.a.b);a.a<a.c.c.length;)t=f(ee(a),82),S=h.Math.min(S,t.d.f.g.c+t.e.a);for(F=new ki,g=new ne(e.a.a);g.a<g.c.c.length;)l=f(ee(g),194),l.i=S,l.e==0&&Wi(F,l,F.c.b,F.c);for(;F.b!=0;){for(l=f(F.b==0?null:(Ln(F.b!=0),lh(F,F.a.a)),194),u=l.f.g.c,L=l.a.a.ec().Jc();L.Ob();)T=f(L.Pb(),82),W=l.i+T.e.a,T.d.g||T.g.c<W?T.o=W:T.o=T.g.c;for(u-=l.f.o,l.b+=u,e.c==(ko(),Gs)||e.c==jg?l.c+=u:l.c-=u,A=l.a.a.ec().Jc();A.Ob();)for(T=f(A.Pb(),82),v=T.f.Jc();v.Ob();)m=f(v.Pb(),82),l1(e.c)?z=e.f.wf(T,m):z=e.f.xf(T,m),m.d.i=h.Math.max(m.d.i,T.o+T.g.b+z-m.e.a),m.k||(m.d.i=h.Math.max(m.d.i,m.g.c-m.e.a)),--m.d.e,m.d.e==0&&Mr(F,m.d)}for(i=new ne(e.a.b);i.a<i.c.c.length;)t=f(ee(i),82),t.g.c=t.o}function lkt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;for(i.Sg(Mxt,1),L=(_i(),Wme),e.a==(wj(),nne)&&(L=oMe),T=0,Ht(),t.gd(new sae(L)),l=t.gc(),m=t.dd(t.gc()),S=!0;S&&m.Sb();)pe=f(m.Ub(),39),f(V(pe,L),17).a==0?--l:S=!1;if(He=t.hd(0,l),g=new Hse(He),He=t.hd(l,t.gc()),v=new Hse(He),g.b==0)for(W=tr(v,0);W.b!=W.d.c;)z=f(Zn(W),39),$e(z,bF,rt(T++));else for(A=g.b,Le=tr(g,0);Le.b!=Le.d.c;){for(Ee=f(Zn(Le),39),$e(Ee,bF,rt(T++)),a=Wue(Ee),lkt(e,a,i.bh(1/A|0)),sE(a,Jke(new sae(bF))),F=new ki,ve=tr(a,0);ve.b!=ve.d.c;)for(pe=f(Zn(ve),39),ue=tr(Ee.d,0);ue.b!=ue.d.c;)J=f(Zn(ue),65),J.c==pe&&Wi(F,J,F.c.b,F.c);for(nf(Ee.d),Ba(Ee.d,F),m=tr(v,v.b),u=Ee.d.b,S=!0;0<u&&S&&m.Sb();)pe=f(m.Ub(),39),f(V(pe,L),17).a==0?($e(pe,bF,rt(T++)),--u,m.Qb()):S=!1}i.Tg()}function a9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(t.Sg("Inverted port preprocessing",1),T=e.b,S=new la(T,0),i=null,ve=new et;S.b<S.d.gc();){for(pe=i,i=(Ln(S.b<S.d.gc()),f(S.d.Xb(S.c=S.b++),25)),F=new ne(ve);F.a<F.c.c.length;)A=f(ee(F),9),jo(A,pe);for(ve.c.length=0,z=new ne(i.a);z.a<z.c.c.length;)if(A=f(ee(z),9),A.k==(cn(),Yi)&&T3(f(V(A,(ot(),Xi)),100))){for(ue=RIe(A,(cs(),td),(tt(),_n)).Jc();ue.Ob();)for(W=f(ue.Pb(),11),v=W.e,m=f(kg(v,Pe(vge,Jfe,18,v.c.length,0,1)),469),u=m,l=0,g=u.length;l<g;++l)a=u[l],MDn(e,W,a,ve);for(J=RIe(A,Uu,pn).Jc();J.Ob();)for(W=f(J.Pb(),11),v=W.g,m=f(kg(v,Pe(vge,Jfe,18,v.c.length,0,1)),469),u=m,l=0,g=u.length;l<g;++l)a=u[l],DDn(e,W,a,ve)}}for(L=new ne(ve);L.a<L.c.c.length;)A=f(ee(L),9),jo(A,i);t.Tg()}function s9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F;for(A=ke(Oe(Fe(e,(uw(),fbe)))),lt(ut(Fe(e,BMe)))&&(T=f(Fe(e,(R3(),MA)),27),l=f(be(Zke(f(be((!T.e&&(T.e=new Qt(vo,T,7,4)),T.e),(!T.e&&(T.e=new Qt(vo,T,7,4)),T.e).i-1),85)),0),27),a=f(be(Zke(f(be((!T.e&&(T.e=new Qt(vo,T,7,4)),T.e),0),85)),0),27),g=new Ke(l.i+l.g/2,l.j+l.f/2),u=new Ke(a.i+a.g/2,a.j+a.f/2),i=A,i<=0&&(i+=_E),L=h.Math.acos((g.a*u.a+g.b*u.b)/(h.Math.sqrt(g.a*g.a+g.b*g.b)*h.Math.sqrt(u.a*u.a+u.b*u.b))),L<=0&&(L+=_E),t=h.Math.atan2(g.b,g.a),t<=0&&(t+=_E),A=PRe-(t-i+L/2)),v=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));v.e!=v.i.gc();)m=f(Mn(v),27),S=new Ke(m.i+m.g/2,m.j+m.f/2),F=S.a*h.Math.cos(A)-S.b*h.Math.sin(A),S.b=S.a*h.Math.sin(A)+S.b*h.Math.cos(A),S.a=F,Of(m,S.a-m.g/2,S.b-m.f/2)}function c9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;if(W=e.n,J=e.o,L=e.d,A=ke(Oe(kT(e,(ot(),wme)))),t){for(T=A*(t.gc()-1),F=0,v=t.Jc();v.Ob();)g=f(v.Pb(),9),T+=g.o.a,F=h.Math.max(F,g.o.b);for(ue=W.a-(T-J.a)/2,l=W.b-L.d+F,a=J.a/(t.gc()+1),u=a,m=t.Jc();m.Ob();)g=f(m.Pb(),9),g.n.a=ue,g.n.b=l-g.o.b,ue+=g.o.a+A,S=S2t(g),S.n.a=g.o.a/2-S.a.a,S.n.b=g.o.b,z=f(V(g,(Ge(),yte)),11),z.e.c.length+z.g.c.length==1&&(z.n.a=u-z.a.a,z.n.b=0,cc(z,e)),u+=a}if(i){for(T=A*(i.gc()-1),F=0,v=i.Jc();v.Ob();)g=f(v.Pb(),9),T+=g.o.a,F=h.Math.max(F,g.o.b);for(ue=W.a-(T-J.a)/2,l=W.b+J.b+L.a-F,a=J.a/(i.gc()+1),u=a,m=i.Jc();m.Ob();)g=f(m.Pb(),9),g.n.a=ue,g.n.b=l,ue+=g.o.a+A,S=S2t(g),S.n.a=g.o.a/2-S.a.a,S.n.b=0,z=f(V(g,(Ge(),yte)),11),z.e.c.length+z.g.c.length==1&&(z.n.a=u-z.a.a,z.n.b=J.b,cc(z,e)),u+=a}}function rfe(e,t,i,a,u,l,g){var m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(L=null,a==(AS(),Aw)?L=t:a==LA&&(L=i),W=L.a.ec().Jc();W.Ob();){for(z=f(W.Pb(),11),J=uc(ce(se(Zo,1),Ye,8,0,[z.i.n,z.n,z.a])).b,ve=new Mo,m=new Mo,S=new yg(z.b);sc(S.a)||sc(S.b);)if(v=f(sc(S.a)?ee(S.a):ee(S.b),18),lt(ut(V(v,(Ge(),U0))))==u&&Oc(l,v,0)!=-1){if(v.d==z?ue=v.c:ue=v.d,pe=uc(ce(se(Zo,1),Ye,8,0,[ue.i.n,ue.n,ue.a])).b,h.Math.abs(pe-J)<.2)continue;pe<J?t.a._b(ue)?Go(ve,new Ya(Aw,v)):Go(ve,new Ya(LA,v)):t.a._b(ue)?Go(m,new Ya(Aw,v)):Go(m,new Ya(LA,v))}if(ve.a.gc()>1)for(F=new uOe(z,ve,a),Oa(ve,new gat(e,F)),vn(g.c,F),A=ve.a.ec().Jc();A.Ob();)T=f(A.Pb(),47),hl(l,T.b);if(m.a.gc()>1)for(F=new uOe(z,m,a),Oa(m,new pat(e,F)),vn(g.c,F),A=m.a.ec().Jc();A.Ob();)T=f(A.Pb(),47),hl(l,T.b)}}function u9n(e,t){var i,a,u,l,g,m;if(f(V(t,(Ge(),Cu)),22).Gc((us(),Vf))){for(m=new ne(t.a);m.a<m.c.c.length;)l=f(ee(m),9),l.k==(cn(),Yi)&&(u=f(V(l,(ot(),Dte)),139),e.c=h.Math.min(e.c,l.n.a-u.b),e.a=h.Math.max(e.a,l.n.a+l.o.a+u.c),e.d=h.Math.min(e.d,l.n.b-u.d),e.b=h.Math.max(e.b,l.n.b+l.o.b+u.a));for(g=new ne(t.a);g.a<g.c.c.length;)if(l=f(ee(g),9),l.k!=(cn(),Yi))switch(l.k.g){case 2:if(a=f(V(l,(ot(),el)),171),a==(Ff(),Tw)){l.n.a=e.c-10,fwt(l,new n8).Jb(new het(l));break}if(a==tx){l.n.a=e.a+10,fwt(l,new kU).Jb(new pet(l));break}if(i=f(V(l,OE),312),i==(w1(),kA)){D_t(l).Jb(new get(l)),l.n.b=e.d-10;break}if(i==WR){D_t(l).Jb(new met(l)),l.n.b=e.b+10;break}break;default:throw ie(new dn("The node type "+l.k+" is not supported by the "+ojn))}}}function l9n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J;for(v=new Ke(a.i+a.g/2,a.j+a.f/2),F=R_t(a),z=f(Fe(t,(ot(),Xi)),100),J=f(Fe(a,vD),64),Zat(Avt(a),ox)||(a.i==0&&a.j==0?W=0:W=tNn(a,J),pi(a,ox,W)),S=new Ke(t.g,t.f),u=PG(a,z,J,F,S,v,new Ke(a.g,a.f),f(V(i,mf),86),i),$e(u,(Ge(),oi),a),l=f(ct(u.j,0),11),vN(l,jLn(a)),$e(u,ax,(Xl(),At(Lw))),A=f(Fe(t,ax),181).Gc(M1),m=new Nn((!a.n&&(a.n=new Be(hc,a,1,7)),a.n));m.e!=m.i.gc();)if(g=f(Mn(m),153),!lt(ut(Fe(g,ME)))&&g.a&&(L=cle(g),Qe(l.f,L),!A))switch(T=0,K7(f(Fe(t,ax),22))&&(T=HAe(new Ke(g.i,g.j),new Ke(g.g,g.f),new Ke(a.g,a.f),0,J)),J.g){case 2:case 4:L.o.a=T;break;case 1:case 3:L.o.b=T}$e(u,yD,Oe(Fe(so(t),yD))),$e(u,ED,Oe(Fe(so(t),ED))),$e(u,cx,Oe(Fe(so(t),cx))),Qe(i.a,u),qr(e.a,a,u)}function d9n(e,t,i,a,u,l){var g,m,v,S,T,A;for(S=new ou,Rc(S,t),No(S,f(Fe(t,(ot(),vD)),64)),$e(S,(Ge(),oi),t),cc(S,i),A=S.o,A.a=t.g,A.b=t.f,T=S.n,T.a=t.i,T.b=t.j,qr(e.a,t,S),g=j3(Lu(Zs(new Ft(null,(!t.e&&(t.e=new Qt(vo,t,7,4)),new Bt(t.e,16))),new Xm),new A2),new ret(t)),g||(g=j3(Lu(Zs(new Ft(null,(!t.d&&(t.d=new Qt(vo,t,8,5)),new Bt(t.d,16))),new tO),new WM),new iet(t))),g||(g=j3(new Ft(null,(!t.e&&(t.e=new Qt(vo,t,7,4)),new Bt(t.e,16))),new l6)),$e(S,Rq,(mn(),!!g)),jMn(S,l,u,f(Fe(t,m_),8)),v=new Nn((!t.n&&(t.n=new Be(hc,t,1,7)),t.n));v.e!=v.i.gc();)m=f(Mn(v),153),!lt(ut(Fe(m,ME)))&&m.a&&Qe(S.f,cle(m));switch(u.g){case 2:case 1:(S.j==(tt(),hn)||S.j==Bn)&&a.Ec((us(),_A));break;case 4:case 3:(S.j==(tt(),_n)||S.j==pn)&&a.Ec((us(),_A))}return S}function f9n(e,t,i,a,u,l){var g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r;tn=bn(e.g,u),tn==null&&(tn=new AN,z=f(tn,147),ve=t+"_s",Ee=ve+l,F=new cT(Ee),Wh(z,Am,F)),Kt=f(tn,147),QN(i,Kt),$r=new AN,tE($r,"x",$Z(e,a,u.j)),tE($r,"y",FZ(e,a,u.k)),Wh(Kt,UCt,$r),xt=new AN,tE(xt,"x",$Z(e,a,u.b)),tE(xt,"y",FZ(e,a,u.c)),Wh(Kt,"endPoint",xt),L=Az((!u.a&&(u.a=new So(yf,u,5)),u.a)),W=!L,W&&(nt=new oh,g=new Fut(e,a,nt),Oa((!u.a&&(u.a=new So(yf,u,5)),u.a),g),Wh(Kt,see,nt)),S=Mle(u),Le=!!S,Le&&pAe(e.a,Kt,$4e,mde(e,Mle(u))),pe=jle(u),He=!!pe,He&&pAe(e.a,Kt,j4e,mde(e,jle(u))),T=(!u.e&&(u.e=new Qt(Ii,u,10,9)),u.e).i==0,J=!T,J&&(Ct=new oh,m=new Hat(e,Ct),Oa((!u.e&&(u.e=new Qt(Ii,u,10,9)),u.e),m),Wh(Kt,B4e,Ct)),A=(!u.g&&(u.g=new Qt(Ii,u,9,10)),u.g).i==0,ue=!A,ue&&(In=new oh,v=new Gat(e,In),Oa((!u.g&&(u.g=new Qt(Ii,u,9,10)),u.g),v),Wh(Kt,F4e,In))}function h9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;for(t.Sg("Comment pre-processing",1),i=0,v=new ne(e.a);v.a<v.c.c.length;)if(m=f(ee(v),9),lt(ut(V(m,(ot(),Z$))))){for(++i,u=0,a=null,S=null,z=new ne(m.j);z.a<z.c.c.length;)L=f(ee(z),11),u+=L.e.c.length+L.g.c.length,L.e.c.length==1&&(a=f(ct(L.e,0),18),S=a.c),L.g.c.length==1&&(a=f(ct(L.g,0),18),S=a.d);if(u==1&&S.e.c.length+S.g.c.length==1&&!lt(ut(V(S.i,Z$))))V9n(m,a,S,S.i),Y7(v);else{for(ue=new et,F=new ne(m.j);F.a<F.c.c.length;){for(L=f(ee(F),11),A=new ne(L.g);A.a<A.c.c.length;)T=f(ee(A),18),T.d.g.c.length==0||vn(ue.c,T);for(g=new ne(L.e);g.a<g.c.c.length;)l=f(ee(g),18),l.c.e.c.length==0||vn(ue.c,l)}for(J=new ne(ue);J.a<J.c.c.length;)W=f(ee(J),18),VS(W,!0)}}t.Zg()&&t._g("Found "+i+" comment boxes"),t.Tg()}function p9n(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;for(ve=0,z=0,F=0,L=1,pe=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));pe.e!=pe.i.gc();)J=f(Mn(pe),27),L+=g1(new Tn(Cn(lw(J).a.Jc(),new B))),xt=J.g,z=h.Math.max(z,xt),A=J.f,F=h.Math.max(F,A),ve+=xt*A;for(W=(!e.a&&(e.a=new Be(Ur,e,10,11)),e.a).i,g=ve+2*a*a*L*W,l=h.Math.sqrt(g),v=h.Math.max(l*i,z),m=h.Math.max(l/i,F),ue=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));ue.e!=ue.i.gc();)J=f(Mn(ue),27),Ct=u.b+(wd(t,26)*r$+wd(t,27)*i$)*(v-J.g),Kt=u.b+(wd(t,26)*r$+wd(t,27)*i$)*(m-J.f),md(J,Ct),bd(J,Kt);for(nt=v+(u.b+u.c),He=m+(u.d+u.a),Le=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));Le.e!=Le.i.gc();)for(Ee=f(Mn(Le),27),T=new Tn(Cn(lw(Ee).a.Jc(),new B));Kn(T);)S=f(Dn(T),85),Uj(S)||w7n(S,t,nt,He);nt+=u.b+u.c,He+=u.d+u.a,WS(e,nt,He,!1,!0)}function g9n(e){var t,i,a,u;for(Zi(Jr(new Ft(null,new Bt(e.a.b,16)),new O6),new O2),JIn(e),Zi(Jr(new Ft(null,new Bt(e.a.b,16)),new Xb),new uk),e.c==(k1(),c4)&&(Zi(Jr(Zs(new Ft(null,new Bt(new Ic(e.f),1)),new N2),new pC),new Det(e)),Zi(Jr(Lu(Zs(Zs(new Ft(null,new Bt(e.d.b,16)),new gC),new R2),new wO),new ng),new jet(e))),Zi(Jr(Zs(new Ft(null,new Bt(new Ic(e.f),1)),new P2),new vp),new $et(e)),u=new Ke(Gi,Gi),t=new Ke(qo,qo),a=new ne(e.a.b);a.a<a.c.c.length;)i=f(ee(a),60),u.a=h.Math.min(u.a,i.d.c),u.b=h.Math.min(u.b,i.d.d),t.a=h.Math.max(t.a,i.d.c+i.d.b),t.b=h.Math.max(t.b,i.d.d+i.d.a);ii(Cp(e.d.c),fX(new Ke(u.a,u.b))),ii(Cp(e.d.f),Ho(new Ke(t.a,t.b),u)),t5n(e,u,t),nu(e.f),nu(e.b),nu(e.g),nu(e.e),e.a.a.c.length=0,e.a.b.c.length=0,e.a=null,e.d=null}function vOe(e,t){olt();var i,a,u,l,g,m,v;if(this.a=new WSe(this),this.b=e,this.c=t,this.f=mce(Ns((Rl(),Ea),t)),this.f.dc())if((m=V3e(Ea,e))==t)for(this.e=!0,this.d=new et,this.f=new Yd,this.f.Ec(CE),f(pQ(LH(Ea,cf(e)),""),29)==e&&this.f.Ec(q7(Ea,cf(e))),u=Mde(Ea,e).Jc();u.Ob();)switch(a=f(u.Pb(),178),kS(Ns(Ea,a))){case 4:{this.d.Ec(a);break}case 5:{this.f.Fc(mce(Ns(Ea,a)));break}}else if(is(),f(t,69).uk())for(this.e=!0,this.f=null,this.d=new et,g=0,v=(e.i==null&&Em(e),e.i).length;g<v;++g)for(a=(i=(e.i==null&&Em(e),e.i),g>=0&&g<i.length?i[g]:null),l=rR(Ns(Ea,a));l;l=rR(Ns(Ea,l)))l==t&&this.d.Ec(a);else kS(Ns(Ea,t))==1&&m?(this.f=null,this.d=(_R(),EMt)):(this.f=null,this.e=!0,this.d=(Ht(),new mz(t)));else this.e=kS(Ns(Ea,t))==5,this.f.Fb(y1e)&&(this.f=y1e)}function dkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;for(i=0,a=CAn(e,t),L=e.s,F=e.t,S=f(f(si(e.r,t),22),83).Jc();S.Ob();)if(v=f(S.Pb(),115),!(!v.c||v.c.d.c.length<=0)){switch(z=v.b.Kf(),m=v.b.nf((Rr(),Sd))?ke(Oe(v.b.mf(Sd))):0,T=v.c,A=T.i,A.b=(g=T.n,T.e.a+g.b+g.c),A.a=(l=T.n,T.e.b+l.d+l.a),t.g){case 1:A.c=v.a?(z.a-A.b)/2:z.a+L,A.d=z.b+m+a,Op(T,(Yl(),Ew)),um(T,(pl(),jp));break;case 3:A.c=v.a?(z.a-A.b)/2:z.a+L,A.d=-m-a-A.a,Op(T,(Yl(),Ew)),um(T,(pl(),np));break;case 2:A.c=-m-a-A.b,v.a?(u=e.v?A.a:f(ct(T.d,0),187).Kf().b,A.d=(z.b-u)/2):A.d=z.b+F,Op(T,(Yl(),tp)),um(T,(pl(),Sw));break;case 4:A.c=z.a+m+a,v.a?(u=e.v?A.a:f(ct(T.d,0),187).Kf().b,A.d=(z.b-u)/2):A.d=z.b+F,Op(T,(Yl(),Om)),um(T,(pl(),Sw))}(t==(tt(),hn)||t==Bn)&&(i=h.Math.max(i,A.a))}i>0&&(f(Es(e.b,t),126).a.b=i)}function m9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn;for(He=new et,z=new ne(e.b);z.a<z.c.c.length;)for(F=f(ee(z),25),ue=new ne(F.a);ue.a<ue.c.c.length;)if(W=f(ee(ue),9),W.k==(cn(),bo)&&li(W,(Ge(),_te))){for(pe=null,Ee=null,ve=null,Ct=new ne(W.j);Ct.a<Ct.c.c.length;)switch(xt=f(ee(Ct),11),xt.j.g){case 4:pe=xt;break;case 2:Ee=xt;break;default:ve=xt}for(Le=f(ct(ve.g,0),18),T=new mY(Le.a),S=new ja(ve.n),ii(S,W.n),A=tr(T,0),UX(A,S),nt=vCe(Le.a),L=new ja(ve.n),ii(L,W.n),Wi(nt,L,nt.c.b,nt.c),Kt=f(V(W,_te),9),tn=f(ct(Kt.j,0),11),v=f(kg(pe.e,Pe(vge,Jfe,18,0,0,1)),469),a=v,l=0,m=a.length;l<m;++l)t=a[l],ua(t,tn),ale(t.a,t.a.b,T);for(v=pm(Ee.g),i=v,u=0,g=i.length;u<g;++u)t=i[u],Ia(t,tn),ale(t.a,0,nt);Ia(Le,null),ua(Le,null),vn(He.c,W)}for(J=new ne(He);J.a<J.c.c.length;)W=f(ee(J),9),jo(W,null)}function b9n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W;if(L=ke(Oe(V(e,(ot(),yD)))),F=ke(Oe(V(e,ED))),A=ke(Oe(V(e,cx))),m=e.o,l=f(ct(e.j,0),11),g=l.n,W=sRn(l,A),!!W){if(t.Gc((Xl(),M1)))switch(f(V(e,(Ge(),Tc)),64).g){case 1:W.c=(m.a-W.b)/2-g.a,W.d=F;break;case 3:W.c=(m.a-W.b)/2-g.a,W.d=-F-W.a;break;case 2:i&&l.e.c.length==0&&l.g.c.length==0?(T=a?W.a:f(ct(l.f,0),72).o.b,W.d=(m.b-T)/2-g.b):W.d=m.b+F-g.b,W.c=-L-W.b;break;case 4:i&&l.e.c.length==0&&l.g.c.length==0?(T=a?W.a:f(ct(l.f,0),72).o.b,W.d=(m.b-T)/2-g.b):W.d=m.b+F-g.b,W.c=L}else if(t.Gc(Lw))switch(f(V(e,(Ge(),Tc)),64).g){case 1:case 3:W.c=g.a+L;break;case 2:case 4:i&&!l.c?(T=a?W.a:f(ct(l.f,0),72).o.b,W.d=(m.b-T)/2-g.b):W.d=g.b+F}for(u=W.d,S=new ne(l.f);S.a<S.c.c.length;)v=f(ee(S),72),z=v.n,z.a=W.c,z.b=u,u+=v.o.b+A}}function wOe(e){var t;if(this.r=Qwn(new Jc,new Id),this.b=new Aj(f(or(Ja),295)),this.p=new Aj(f(or(Ja),295)),this.i=new Aj(f(or(BIt),295)),this.e=e,this.o=new ja(e.Kf()),this.D=lt(ut(e.mf((Rr(),RD)))),this.F=e.Wf()||lt(ut(e.mf(rV))),this.A=f(e.mf(UE),22),this.B=f(e.mf(Ow),22),this.q=f(e.mf(OF),100),this.u=f(e.mf(px),22),!lNn(this.u))throw ie(new h0("Invalid port label placement: "+this.u));if(this.v=lt(ut(e.mf(N9e))),this.j=f(e.mf(HA),22),!T6n(this.j))throw ie(new h0("Invalid node label placement: "+this.j));this.n=f(Pj(e,v9e),107),this.k=ke(Oe(Pj(e,Ene))),this.d=ke(Oe(Pj(e,D9e))),this.w=ke(Oe(Pj(e,B9e))),this.s=ke(Oe(Pj(e,M9e))),this.t=ke(Oe(Pj(e,j9e))),this.C=f(Pj(e,$9e),139),this.c=2*this.d,t=!this.B.Gc((yd(),bV)),this.f=new Oj(0,t,0),this.g=new Oj(1,t,0),Eae(this.f,(Np(),Bu),this.g)}function v9n(){Dr(xV,new C5),Dr(qne,new T9),Dr(CV,new iN),Dr(eje,new I5),Dr(ft,new XU),Dr(se($l,1),new O5),Dr(to,new N5),Dr(BR,new JU),Dr(ft,new eN),Dr(ft,new Nk),Dr(ft,new g9),Dr(mo,new m9),Dr(ft,new b9),Dr(qf,new T5),Dr(qf,new x5),Dr(ft,new qU),Dr(eD,new VU),Dr(ft,new v9),Dr(ft,new w9),Dr(ft,new Rk),Dr(ft,new tN),Dr(ft,new y9),Dr(se($l,1),new t3),Dr(ft,new E9),Dr(ft,new S9),Dr(qf,new _9),Dr(qf,new k9),Dr(ft,new nN),Dr(ba,new x9),Dr(ft,new gu),Dr(i_,new A5),Dr(ft,new WU),Dr(ft,new C9),Dr(ft,new rN),Dr(ft,new KU),Dr(qf,new A9),Dr(qf,new n3),Dr(ft,new jy),Dr(ft,new r3),Dr(ft,new oN),Dr(ft,new $y),Dr(ft,new l0),Dr(ft,new i3),Dr(o_,new YU),Dr(ft,new o3),Dr(ft,new a3),Dr(ft,new I9),Dr(o_,new O9),Dr(i_,new aN),Dr(ft,new N9),Dr(ba,new R9)}function fkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;switch(T=new fd,e.a.g){case 3:L=f(V(t.e,(Ge(),NE)),15),F=f(V(t.j,NE),15),z=f(V(t.f,NE),15),i=f(V(t.e,AA),15),a=f(V(t.j,AA),15),u=f(V(t.f,AA),15),g=new et,$o(g,L),F.Ic(new E8),$o(g,Bd(F)),$o(g,z),l=new et,$o(l,i),$o(l,Bd(a)),$o(l,u),$e(t.f,NE,g),$e(t.f,AA,l),$e(t.f,VPe,t.f),$e(t.e,NE,null),$e(t.e,AA,null),$e(t.j,NE,null),$e(t.j,AA,null);break;case 1:Ba(T,t.e.a),Mr(T,t.i.n),Ba(T,Bd(t.j.a)),Mr(T,t.a.n),Ba(T,t.f.a);break;default:Ba(T,t.e.a),Ba(T,Bd(t.j.a)),Ba(T,t.f.a)}nf(t.f.a),Ba(t.f.a,T),Ia(t.f,t.e.c),m=f(V(t.e,(ot(),Ls)),77),S=f(V(t.j,Ls),77),v=f(V(t.f,Ls),77),(m||S||v)&&(A=new fd,Dke(A,v),Dke(A,S),Dke(A,m),$e(t.f,Ls,A)),Ia(t.j,null),ua(t.j,null),Ia(t.e,null),ua(t.e,null),jo(t.a,null),jo(t.i,null),t.g&&fkt(e,t.g)}function hkt(){hkt=Z;var e,t,i;for(new nG(1,0),new nG(10,0),new nG(0,0),zAt=Pe(Vpe,Ye,245,11,0,1),IE=Pe(Xf,_m,30,100,15,1),O6e=ce(se(sa,1),_s,30,15,[1,5,25,125,625,3125,15625,78125,390625,1953125,9765625,48828125,244140625,1220703125,6103515625,30517578125,152587890625,762939453125,3814697265625,19073486328125,95367431640625,476837158203125,0x878678326eac9]),N6e=Pe(pr,Pr,30,O6e.length,15,1),R6e=ce(se(sa,1),_s,30,15,[1,10,100,M0,1e4,Lfe,1e6,1e7,1e8,SQ,1e10,1e11,1e12,1e13,1e14,1e15,1e16]),P6e=Pe(pr,Pr,30,R6e.length,15,1),L6e=Pe(Vpe,Ye,245,11,0,1),e=0;e<L6e.length;e++)zAt[e]=new nG(e,0),L6e[e]=new nG(0,e),IE[e]=48;for(;e<IE.length;e++)IE[e]=48;for(i=0;i<N6e.length;i++)N6e[i]=BAe(O6e[i]);for(t=0;t<P6e.length;t++)P6e[t]=BAe(R6e[t]);vL()}function w9n(){function e(){this.obj=this.createObject()}return e.prototype.createObject=function(t){return Object.create(null)},e.prototype.get=function(t){return this.obj[t]},e.prototype.set=function(t,i){this.obj[t]=i},e.prototype[jfe]=function(t){delete this.obj[t]},e.prototype.keys=function(){return Object.getOwnPropertyNames(this.obj)},e.prototype.entries=function(){var t=this.keys(),i=this,a=0;return{next:function(){if(a>=t.length)return{done:!0};var u=t[a++];return{value:[u,i.get(u)],done:!1}}}},kPn()||(e.prototype.createObject=function(){return{}},e.prototype.get=function(t){return this.obj[":"+t]},e.prototype.set=function(t,i){this.obj[":"+t]=i},e.prototype[jfe]=function(t){delete this.obj[":"+t]},e.prototype.keys=function(){var t=[];for(var i in this.obj)i.charCodeAt(0)==58&&t.push(i.substring(1));return t}),e}function _i(){_i=Z,mF=new Si(eNe),new Mi("DEPTH",rt(0)),Wme=new Mi("FAN",rt(0)),oMe=new Mi(Pxt,rt(0)),Iw=new Mi("ROOT",(mn(),!1)),Xme=new Mi("LEFTNEIGHBOR",null),T5t=new Mi("RIGHTNEIGHBOR",null),Qte=new Mi("LEFTSIBLING",null),Jme=new Mi("RIGHTSIBLING",null),Vme=new Mi("DUMMY",!1),new Mi("LEVEL",rt(0)),cMe=new Mi("REMOVABLE_EDGES",new ki),Vq=new Mi("XCOOR",rt(0)),Wq=new Mi("YCOOR",rt(0)),ene=new Mi("LEVELHEIGHT",0),Fp=new Mi("LEVELMIN",0),ap=new Mi("LEVELMAX",0),Kme=new Mi("GRAPH_XMIN",0),Yme=new Mi("GRAPH_YMIN",0),aMe=new Mi("GRAPH_XMAX",0),sMe=new Mi("GRAPH_YMAX",0),iMe=new Mi("COMPACT_LEVEL_ASCENSION",!1),qme=new Mi("COMPACT_CONSTRAINTS",new et),gF=new Mi("ID",""),bF=new Mi("POSITION",rt(0)),q0=new Mi("PRELIM",0),xD=new Mi("MODIFIER",0),TD=new Si(GTt),qq=new Si(qTt)}function y9n(e){ZIe();var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;if(e==null)return null;if(A=e.length*8,A==0)return"";for(m=A%24,F=A/24|0,L=m!=0?F+1:F,l=null,l=Pe(Xf,_m,30,L*4,15,1),S=0,T=0,t=0,i=0,a=0,g=0,u=0,v=0;v<F;v++)t=e[u++],i=e[u++],a=e[u++],T=(i&15)<<24>>24,S=(t&3)<<24>>24,z=(t&-128)==0?t>>2<<24>>24:(t>>2^192)<<24>>24,W=(i&-128)==0?i>>4<<24>>24:(i>>4^240)<<24>>24,J=(a&-128)==0?a>>6<<24>>24:(a>>6^252)<<24>>24,l[g++]=Q0[z],l[g++]=Q0[W|S<<4],l[g++]=Q0[T<<2|J],l[g++]=Q0[a&63];return m==8?(t=e[u],S=(t&3)<<24>>24,z=(t&-128)==0?t>>2<<24>>24:(t>>2^192)<<24>>24,l[g++]=Q0[z],l[g++]=Q0[S<<4],l[g++]=61,l[g++]=61):m==16&&(t=e[u],i=e[u+1],T=(i&15)<<24>>24,S=(t&3)<<24>>24,z=(t&-128)==0?t>>2<<24>>24:(t>>2^192)<<24>>24,W=(i&-128)==0?i>>4<<24>>24:(i>>4^240)<<24>>24,l[g++]=Q0[z],l[g++]=Q0[W|S<<4],l[g++]=Q0[T<<2],l[g++]=61),bm(l,0,l.length)}function E9n(e,t){var i,a,u,l,g,m,v;if(e.e==0&&e.p>0&&(e.p=-(e.p-1)),e.p>Ra&&MTe(t,e.p-hw),g=t.q.getDate(),wH(t,1),e.k>=0&&uEn(t,e.k),e.c>=0?wH(t,e.c):e.k>=0?(v=new gCe(t.q.getFullYear()-hw,t.q.getMonth(),35),a=35-v.q.getDate(),wH(t,h.Math.min(a,g))):wH(t,g),e.f<0&&(e.f=t.q.getHours()),e.b>0&&e.f<12&&(e.f+=12),vmn(t,e.f==24&&e.g?0:e.f),e.j>=0&&HSn(t,e.j),e.n>=0&&i_n(t,e.n),e.i>=0&&sst(t,Ua(Na(pG(Nc(t.q.getTime()),M0),M0),e.i)),e.a&&(u=new EY,MTe(u,u.q.getFullYear()-hw-80),$ae(Nc(t.q.getTime()),Nc(u.q.getTime()))&&MTe(t,u.q.getFullYear()-hw+100)),e.d>=0){if(e.c==-1)i=(7+e.d-t.q.getDay())%7,i>3&&(i-=7),m=t.q.getMonth(),wH(t,t.q.getDate()+i),t.q.getMonth()!=m&&wH(t,t.q.getDate()+(i>0?-7:7));else if(t.q.getDay()!=e.d)return!1}return e.o>Ra&&(l=t.q.getTimezoneOffset(),sst(t,Ua(Nc(t.q.getTime()),(e.o-l)*60*M0))),!0}function pkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;if(u=V(t,(Ge(),oi)),!!me(u,206)){for(z=f(u,27),W=t.e,L=new ja(t.c),l=t.d,L.a+=l.b,L.b+=l.d,Ee=f(Fe(z,(ot(),$te)),181),Cl(Ee,(yd(),xne))&&(F=f(Fe(z,JLe),107),Z9(F,l.a),Uy(F,l.d),V5(F,l.b),Q9(F,l.c)),i=new et,T=new ne(t.a);T.a<T.c.c.length;)for(v=f(ee(T),9),me(V(v,oi),206)?x9n(v,L):me(V(v,oi),193)&&!W&&(a=f(V(v,oi),124),pe=i_t(t,v,a.g,a.f),Of(a,pe.a,pe.b)),ue=new ne(v.j);ue.a<ue.c.c.length;)J=f(ee(ue),11),Zi(Jr(new Ft(null,new Bt(J.g,16)),new oet(v)),new aet(i));if(W)for(ue=new ne(W.j);ue.a<ue.c.c.length;)J=f(ee(ue),11),Zi(Jr(new Ft(null,new Bt(J.g,16)),new set(W)),new cet(i));for(ve=f(Fe(z,P1),220),m=new ne(i);m.a<m.c.c.length;)g=f(ee(m),18),z8n(g,ve,L);for($Ln(t),S=new ne(t.a);S.a<S.c.c.length;)v=f(ee(S),9),A=v.e,A&&pkt(e,A)}}function gkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F;if(!f(f(si(e.r,t),22),83).dc()){if(g=f(Es(e.b,t),126),v=g.i,m=g.n,T=Ade(e,t),a=v.b-m.b-m.c,u=g.a.a,l=v.c+m.b,F=e.w,(T==(K3(),Rw)||T==fV)&&f(f(si(e.r,t),22),83).gc()==1&&(u=T==Rw?u-2*e.w:u,T=LF),a<u&&!e.B.Gc((yd(),Cne)))T==Rw?(F+=(a-u)/(f(f(si(e.r,t),22),83).gc()+1),l+=F):F+=(a-u)/(f(f(si(e.r,t),22),83).gc()-1);else switch(a<u&&(u=T==Rw?u-2*e.w:u,T=LF),T.g){case 3:l+=(a-u)/2;break;case 4:l+=a-u;break;case 0:i=(a-u)/(f(f(si(e.r,t),22),83).gc()+1),F+=h.Math.max(0,i),l+=F;break;case 1:i=(a-u)/(f(f(si(e.r,t),22),83).gc()-1),F+=h.Math.max(0,i)}for(L=f(f(si(e.r,t),22),83).Jc();L.Ob();)A=f(L.Pb(),115),A.e.a=l+A.d.b,A.e.b=(S=A.b,S.nf((Rr(),Sd))?S.$f()==(tt(),hn)?-S.Kf().b-ke(Oe(S.mf(Sd))):ke(Oe(S.mf(Sd))):S.$f()==(tt(),hn)?-S.Kf().b:0),l+=A.d.b+A.b.Kf().a+A.d.c+F}}function mkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z;if(!f(f(si(e.r,t),22),83).dc()){if(g=f(Es(e.b,t),126),v=g.i,m=g.n,A=Ade(e,t),a=v.a-m.d-m.a,u=g.a.b,l=v.d+m.d,z=e.w,S=e.o.a,(A==(K3(),Rw)||A==fV)&&f(f(si(e.r,t),22),83).gc()==1&&(u=A==Rw?u-2*e.w:u,A=LF),a<u&&!e.B.Gc((yd(),Cne)))A==Rw?(z+=(a-u)/(f(f(si(e.r,t),22),83).gc()+1),l+=z):z+=(a-u)/(f(f(si(e.r,t),22),83).gc()-1);else switch(a<u&&(u=A==Rw?u-2*e.w:u,A=LF),A.g){case 3:l+=(a-u)/2;break;case 4:l+=a-u;break;case 0:i=(a-u)/(f(f(si(e.r,t),22),83).gc()+1),z+=h.Math.max(0,i),l+=z;break;case 1:i=(a-u)/(f(f(si(e.r,t),22),83).gc()-1),z+=h.Math.max(0,i)}for(F=f(f(si(e.r,t),22),83).Jc();F.Ob();)L=f(F.Pb(),115),L.e.a=(T=L.b,T.nf((Rr(),Sd))?T.$f()==(tt(),pn)?-T.Kf().a-ke(Oe(T.mf(Sd))):S+ke(Oe(T.mf(Sd))):T.$f()==(tt(),pn)?-T.Kf().a:S),L.e.b=l+L.d.d,l+=L.d.d+L.b.Kf().b+L.d.a+z}}function S9n(e,t){var i,a,u,l,g;for(t.Sg("Processor determine the coords for each level",1),a=new et,g=tr(e.b,0);g.b!=g.d.c;){for(u=f(Zn(g),39);f(V(u,(Ec(),Lm)),17).a>a.c.length-1;)Qe(a,new Ya($R,LRe));i=f(V(u,Lm),17).a,l1(f(V(e,E_),86))?(u.e.a<ke(Oe((Gt(i,a.c.length),f(a.c[i],47)).a))&&$k((Gt(i,a.c.length),f(a.c[i],47)),u.e.a),u.e.a+u.f.a>ke(Oe((Gt(i,a.c.length),f(a.c[i],47)).b))&&w((Gt(i,a.c.length),f(a.c[i],47)),u.e.a+u.f.a)):(u.e.b<ke(Oe((Gt(i,a.c.length),f(a.c[i],47)).a))&&$k((Gt(i,a.c.length),f(a.c[i],47)),u.e.b),u.e.b+u.f.b>ke(Oe((Gt(i,a.c.length),f(a.c[i],47)).b))&&w((Gt(i,a.c.length),f(a.c[i],47)),u.e.b+u.f.b))}for(l=tr(e.b,0);l.b!=l.d.c;)u=f(Zn(l),39),i=f(V(u,(Ec(),Lm)),17).a,$e(u,(_i(),Fp),Oe((Gt(i,a.c.length),f(a.c[i],47)).a)),$e(u,ap,Oe((Gt(i,a.c.length),f(a.c[i],47)).b));t.Tg()}function _9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(e.o=ke(Oe(V(e.i,(ot(),v_)))),e.f=ke(Oe(V(e.i,jE))),e.j=e.i.b.c.length,m=e.j-1,L=0,e.k=0,e.n=0,e.b=gm(Pe(ba,Ye,17,e.j,0,1)),e.c=gm(Pe(mo,Ye,341,e.j,7,1)),g=new ne(e.i.b);g.a<g.c.c.length;){for(u=f(ee(g),25),u.p=m,A=new ne(u.a);A.a<A.c.c.length;)T=f(ee(A),9),T.p=L,++L;--m}for(e.g=Pe(pr,Pr,30,L,15,1),e.d=iT(pr,[Ye,Pr],[54,30],15,[L,3],2),e.p=new et,e.q=new et,t=0,e.e=0,l=new ne(e.i.b);l.a<l.c.c.length;){for(u=f(ee(l),25),m=u.p,a=0,W=0,v=u.a.c.length,S=0,A=new ne(u.a);A.a<A.c.c.length;)T=f(ee(A),9),L=T.p,e.g[L]=T.c.p,S+=T.o.b+e.o,i=g1(new Tn(Cn(Fo(T).a.Jc(),new B))),z=g1(new Tn(Cn(ao(T).a.Jc(),new B))),e.d[L][0]=z-i,e.d[L][1]=i,e.d[L][2]=z,a+=i,W+=z,i>0&&Qe(e.q,T),Qe(e.p,T);t-=a,F=v+t,S+=t*e.f,af(e.b,m,rt(F)),af(e.c,m,S),e.k=h.Math.max(e.k,F),e.n=h.Math.max(e.n,S),e.e+=t,t+=W}}function bkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;if(t.b!=0){for(F=new ki,m=null,z=null,a=Aa(h.Math.floor(h.Math.log(t.b)*h.Math.LOG10E)+1),v=0,ve=tr(t,0);ve.b!=ve.d.c;)for(ue=f(Zn(ve),39),Ae(z)!==Ae(V(ue,(_i(),gF)))&&(z=dr(V(ue,gF)),v=0),z!=null?m=z+iht(v++,a):m=iht(v++,a),$e(ue,gF,m),J=(u=tr(new u1(ue).a.d,0),new m3(u));kz(J.a);)W=f(Zn(J.a),65).c,Wi(F,W,F.c.b,F.c),$e(W,gF,m);for(L=new Wn,g=0;g<m.length-a;g++)for(pe=tr(t,0);pe.b!=pe.d.c;)ue=f(Zn(pe),39),S=uh(dr(V(ue,(_i(),gF))),0,g+1),i=(S==null?ac(Is(L.f,null)):w7(L.i,S))!=null?f(S==null?ac(Is(L.f,null)):w7(L.i,S),17).a+1:1,Os(L,S,rt(i));for(A=new yT(new Vi(L).a);A.b;)T=G3(A),l=rt(bn(e.a,T.jd())!=null?f(bn(e.a,T.jd()),17).a:0),Os(e.a,dr(T.jd()),rt(f(T.kd(),17).a+l.a)),l=f(bn(e.b,T.jd()),17),(!l||l.a<f(T.kd(),17).a)&&Os(e.b,dr(T.jd()),f(T.kd(),17));bkt(e,F)}}function k9n(e){var t,i,a,u,l,g,m,v,S,T,A,L;for(i=null,v=null,u=f(V(e.b,(ot(),pme)),345),u==(Ej(),Uq)&&(i=new et,v=new et),m=new ne(e.d);m.a<m.c.c.length;)if(g=f(ee(m),105),l=g.i,!!l)switch(g.e.g){case 0:t=f(aj(new LN(g.b)),64),u==Uq&&t==(tt(),hn)?vn(i.c,g):u==Uq&&t==(tt(),Bn)?vn(v.c,g):RAn(g,t);break;case 1:S=g.a.d.j,T=g.c.d.j,S==(tt(),hn)?E0(g,hn,(_g(),rD),g.a):T==hn?E0(g,hn,(_g(),iD),g.c):S==Bn?E0(g,Bn,(_g(),iD),g.a):T==Bn&&E0(g,Bn,(_g(),rD),g.c);break;case 2:case 3:a=g.b,Cl(a,(tt(),hn))?Cl(a,Bn)?Cl(a,pn)?Cl(a,_n)||E0(g,hn,(_g(),iD),g.c):E0(g,hn,(_g(),rD),g.a):E0(g,hn,(_g(),VR),null):E0(g,Bn,(_g(),VR),null);break;case 4:A=g.a.d.j,L=g.a.d.j,A==(tt(),hn)||L==hn?E0(g,Bn,(_g(),VR),null):E0(g,hn,(_g(),VR),null)}i&&(i.c.length==0||h_t(i,(tt(),hn)),v.c.length==0||h_t(v,(tt(),Bn)))}function T9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;for(i.Sg("Breadth first model order layering",1),e.a=t,J=new et,W=new ne(e.a.a);W.a<W.c.c.length;)F=f(ee(W),9),F.k==(cn(),Yi)&&vn(J.c,F);for(Ht(),na(J,new L6),v=!0,u=new ru(e.a),a=null,Qe(e.a.b,u),z=new ne(J);z.a<z.c.c.length;)if(F=f(ee(z),9),v)jo(F,u),v=!1;else{for(m=new Tn(Cn(Fo(F).a.Jc(),new B));Kn(m);)l=f(Dn(m),18),(l.c.i.k==(cn(),Yi)&&l.c.i.c==u||l.c.i.k==Vc&&f(Dn(new Tn(Cn(Fo(l.c.i).a.Jc(),new B))),18).c.i.c==u)&&(a=new ru(e.a),Qe(e.a.b,a),u=new ru(e.a),Qe(e.a.b,u));for(g=new Tn(Cn(Fo(F).a.Jc(),new B));Kn(g);)l=f(Dn(g),18),l.c.i.k==(cn(),Vc)&&!l.c.i.c&&jo(l.c.i,a);jo(F,u)}for(e.a.a.c.length=0,ue=new et,A=new ne(e.a.b);A.a<A.c.c.length;)S=f(ee(A),25),S.a.c.length==0&&vn(ue.c,S);for(zCe(e.a.b,ue),L=0,T=new ne(e.a.b);T.a<T.c.c.length;)S=f(ee(T),25),S.p=L,++L;i.Tg()}function tt(){tt=Z;var e;fc=new Vz(s$,0),hn=new Vz("NORTH",1),_n=new Vz("EAST",2),Bn=new Vz("SOUTH",3),pn=new Vz("WEST",4),Bg=(Ht(),new tP((e=f(Tp(Ja),10),new Rf(e,f(Gh(e,e.length),10),0)))),dp=A0(xi(hn,ce(se(Ja,1),qc,64,0,[]))),Kf=A0(xi(_n,ce(se(Ja,1),qc,64,0,[]))),vf=A0(xi(Bn,ce(se(Ja,1),qc,64,0,[]))),_h=A0(xi(pn,ce(se(Ja,1),qc,64,0,[]))),wl=A0(xi(hn,ce(se(Ja,1),qc,64,0,[Bn]))),tl=A0(xi(_n,ce(se(Ja,1),qc,64,0,[pn]))),fp=A0(xi(hn,ce(se(Ja,1),qc,64,0,[pn]))),nd=A0(xi(hn,ce(se(Ja,1),qc,64,0,[_n]))),wf=A0(xi(Bn,ce(se(Ja,1),qc,64,0,[pn]))),Yf=A0(xi(_n,ce(se(Ja,1),qc,64,0,[Bn]))),rd=A0(xi(hn,ce(se(Ja,1),qc,64,0,[_n,pn]))),vl=A0(xi(_n,ce(se(Ja,1),qc,64,0,[Bn,pn]))),yl=A0(xi(hn,ce(se(Ja,1),qc,64,0,[Bn,pn]))),jl=A0(xi(hn,ce(se(Ja,1),qc,64,0,[_n,Bn]))),Au=A0(xi(hn,ce(se(Ja,1),qc,64,0,[_n,Bn,pn])))}function x9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W;for(a=f(V(e,(Ge(),oi)),27),z=f(V(e,(ot(),Ote)),17).a,l=f(V(e,Lte),17).a,pi(a,Ote,rt(z)),pi(a,Lte,rt(l)),md(a,e.n.a+t.a),bd(a,e.n.b+t.b),(f(Fe(a,DE),181).gc()!=0||e.e||Ae(V(ra(e),jte))===Ae((Nj(),aF))&&Est((xT(),(e.q?e.q:(Ht(),Ht(),Cb))._b(g_)?L=f(V(e,g_),203):L=f(V(ra(e),nF),203),L)))&&(RS(a,e.o.a),NS(a,e.o.b)),A=new ne(e.j);A.a<A.c.c.length;)S=f(ee(A),11),W=V(S,oi),me(W,193)&&(u=f(W,124),Of(u,S.n.a,S.n.b),pi(u,vD,S.j));for(F=f(V(e,p_),181).gc()!=0,v=new ne(e.b);v.a<v.c.c.length;)g=f(ee(v),72),(F||f(V(g,p_),181).gc()!=0)&&(i=f(V(g,oi),153),x3(i,g.o.a,g.o.b),Of(i,g.n.a,g.n.b));if(!K7(f(V(e,ax),22)))for(T=new ne(e.j);T.a<T.c.c.length;)for(S=f(ee(T),11),m=new ne(S.f);m.a<m.c.c.length;)g=f(ee(m),72),i=f(V(g,oi),153),RS(i,g.o.a),NS(i,g.o.b),Of(i,g.n.a,g.n.b)}function C9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct;for(t.Sg("Calculate Graph Size",1),t.ah(e,HRe),A=$R,L=$R,S=GRe,T=GRe,W=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));W.e!=W.i.gc();)F=f(Mn(W),27),pe=F.i,ve=F.j,Ct=F.g,m=F.f,v=f(Fe(F,(Rr(),OD)),139),A=h.Math.min(A,pe-v.b),L=h.Math.min(L,ve-v.d),S=h.Math.max(S,pe+Ct+v.c),T=h.Math.max(T,ve+m+v.a);for(ue=f(Fe(e,(Rr(),V0)),107),J=new Ke(A-ue.b,L-ue.d),xt=S-A+(ue.b+ue.c),g=T-L+(ue.d+ue.a),lt(ut(Fe(e,(uw(),MMe))))&&(Ee=f(Fe(e,(R3(),MA)),27),Le=f(Fe(Ee,OD),139),He=Ee.i+Ee.g/2+(Le.b+Le.c)/2-J.a,nt=Ee.j+Ee.f/2+(Le.d+Le.a)/2-J.b,u=xt-He,l=g-nt,u<xt/2?(i=u-He,xt+=i,J.a-=i):(i=He-u,xt+=i),l<g/2?(a=l-nt,g+=a,J.b-=a):(a=nt-l,g+=a)),z=new Nn((!e.a&&(e.a=new Be(Ur,e,10,11)),e.a));z.e!=z.i.gc();)F=f(Mn(z),27),md(F,F.i-J.a),bd(F,F.j-J.b);lt(ut(Fe(e,hx)))||(RS(e,xt),NS(e,g)),pi(e,BA,xt-(ue.b+ue.c)),pi(e,FA,g-(ue.d+ue.a)),t.ah(e,tee)}function A9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F;if(e.e.a.$b(),e.f.a.$b(),e.c.c.length=0,e.i.c.length=0,e.g.a.$b(),t)for(g=new ne(t.a);g.a<g.c.c.length;)for(l=f(ee(g),9),A=L0(l,(tt(),_n)).Jc();A.Ob();)for(T=f(A.Pb(),11),Go(e.e,T),u=new ne(T.g);u.a<u.c.c.length;)a=f(ee(u),18),!Fa(a)&&(Qe(e.c,a),Cvt(e,a),m=a.c.i.k,(m==(cn(),Yi)||m==xu||m==bo||m==Dg)&&Qe(e.j,a),F=a.d,L=F.i.c,L==i?Go(e.f,F):L==t?Go(e.e,F):hl(e.c,a));if(i)for(g=new ne(i.a);g.a<g.c.c.length;){for(l=f(ee(g),9),S=new ne(l.j);S.a<S.c.c.length;)for(v=f(ee(S),11),u=new ne(v.g);u.a<u.c.c.length;)a=f(ee(u),18),Fa(a)&&Go(e.g,a);for(A=L0(l,(tt(),pn)).Jc();A.Ob();)for(T=f(A.Pb(),11),Go(e.f,T),u=new ne(T.g);u.a<u.c.c.length;)a=f(ee(u),18),!Fa(a)&&(Qe(e.c,a),Cvt(e,a),m=a.c.i.k,(m==(cn(),Yi)||m==xu||m==bo||m==Dg)&&Qe(e.j,a),F=a.d,L=F.i.c,L==i?Go(e.f,F):L==t?Go(e.e,F):hl(e.c,a))}}function I9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;for(i.Sg("Polyline edge routing",1),J=ke(Oe(V(t,(ot(),PLe)))),F=ke(Oe(V(t,NA))),u=ke(Oe(V(t,IA))),a=h.Math.min(1,u/F),ve=0,v=0,t.b.c.length!=0&&(Ee=b2t(f(ct(t.b,0),25)),ve=.4*a*Ee),m=new la(t.b,0);m.b<m.d.gc();){for(g=(Ln(m.b<m.d.gc()),f(m.d.Xb(m.c=m.b++),25)),l=JY(g,Hq),l&&ve>0&&(ve-=F),lOe(g,ve),T=0,L=new ne(g.a);L.a<L.c.c.length;){for(A=f(ee(L),9),S=0,W=new Tn(Cn(ao(A).a.Jc(),new B));Kn(W);)z=f(Dn(W),18),ue=wg(z.c).b,pe=wg(z.d).b,g==z.d.i.c&&!Fa(z)&&(l4n(z,ve,.4*a*h.Math.abs(ue-pe)),z.c.j==(tt(),pn)&&(ue=0,pe=0)),S=h.Math.max(S,h.Math.abs(pe-ue));switch(A.k.g){case 0:case 4:case 1:case 3:case 5:XMn(e,A,ve,J)}T=h.Math.max(T,S)}m.b<m.d.gc()&&(Ee=b2t((Ln(m.b<m.d.gc()),f(m.d.Xb(m.c=m.b++),25))),T=h.Math.max(T,Ee),Ln(m.b>0),m.a.Xb(m.c=--m.b)),v=.4*a*T,!l&&m.b<m.d.gc()&&(v+=F),ve+=g.c.a+v}e.a.a.$b(),t.f.a=ve,i.Tg()}function O9n(e){var t,i,a,u,l;switch(bH(e,KCt),(!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b).i+(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c).i){case 0:throw ie(new dn("The edge must have at least one source or target."));case 1:return(!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b).i==0?so(Us(f(be((!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c),0),84))):so(Us(f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84)))}if((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b).i==1&&(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c).i==1){if(u=Us(f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84)),l=Us(f(be((!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c),0),84)),so(u)==so(l))return so(u);if(u==so(l))return u;if(l==so(u))return l}for(a=pb(Df(ce(se(Gf,1),Jt,20,0,[(!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),(!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c)]))),t=Us(f(Dn(a),84));Kn(a);)if(i=Us(f(Dn(a),84)),i!=t&&!bT(i,t)){if(so(i)==so(t))t=so(i);else if(t=v6n(t,i),!t)return null}return t}function yOe(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;if(L=t.length,L>0&&(v=(gn(0,t.length),t.charCodeAt(0)),v!=64)){if(v==37&&(A=t.lastIndexOf("%"),S=!1,A!=0&&(A==L-1||(S=(gn(A+1,t.length),t.charCodeAt(A+1)==46))))){if(g=(ga(1,A,t.length),t.substr(1,A-1)),ve=Mt("%",g)?null:EOe(g),a=0,S)try{a=ff((gn(A+2,t.length+1),t.substr(A+2)),Ra,Qr)}catch(Ee){throw Ee=go(Ee),me(Ee,130)?(m=Ee,ie(new xJ(m))):ie(Ee)}for(J=Gxe(e.Ch());J.Ob();)if(z=ZJ(J),me(z,503)&&(u=f(z,582),pe=u.d,(ve==null?pe==null:Mt(ve,pe))&&a--==0))return u;return null}if(T=t.lastIndexOf("."),F=T==-1?t:(ga(0,T,t.length),t.substr(0,T)),i=0,T!=-1)try{i=ff((gn(T+1,t.length+1),t.substr(T+1)),Ra,Qr)}catch(Ee){if(Ee=go(Ee),me(Ee,130))F=t;else throw ie(Ee)}for(F=Mt("%",F)?null:EOe(F),W=Gxe(e.Ch());W.Ob();)if(z=ZJ(W),me(z,197)&&(l=f(z,197),ue=l.ve(),(F==null?ue==null:Mt(F,ue))&&i--==0))return l;return null}return rkt(e,t)}function N9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe;for(T=new Wn,v=new CS,a=new ne(e.a.a.b);a.a<a.c.c.length;)if(t=f(ee(a),60),S=D3(t),S)ml(T.f,S,t);else if(pe=iR(t),pe)for(l=new ne(pe.k);l.a<l.c.c.length;)u=f(ee(l),18),jt(v,u,t);for(i=new ne(e.a.a.b);i.a<i.c.c.length;)if(t=f(ee(i),60),S=D3(t),S){for(m=new Tn(Cn(ao(S).a.Jc(),new B));Kn(m);)if(g=f(Dn(m),18),!Fa(g)&&(z=g.c,ue=g.d,!((tt(),wl).Gc(g.c.j)&&wl.Gc(g.d.j)))){if(W=f(bn(T,g.d.i),60),Zh(Uh(Bh(zh(Fh(new ah,0),100),e.c[t.a.d]),e.c[W.a.d])),z.j==pn&&Hlt((Nl(),z))){for(L=f(si(v,g),22).Jc();L.Ob();)if(A=f(L.Pb(),60),A.d.c<t.d.c){if(F=e.c[A.a.d],J=e.c[t.a.d],F==J)continue;Zh(Uh(Bh(zh(Fh(new ah,1),100),F),J))}}if(ue.j==_n&&Glt((Nl(),ue))){for(L=f(si(v,g),22).Jc();L.Ob();)if(A=f(L.Pb(),60),A.d.c>t.d.c){if(F=e.c[t.a.d],J=e.c[A.a.d],F==J)continue;Zh(Uh(Bh(zh(Fh(new ah,1),100),F),J))}}}}}function R9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;if(L=f(f(si(e.r,t),22),83),t==(tt(),_n)||t==pn){dkt(e,t);return}for(l=t==hn?(DS(),wq):(DS(),yq),Ee=t==hn?(pl(),jp):(pl(),np),i=f(Es(e.b,t),126),a=i.i,u=a.c+$3(ce(se(sa,1),_s,30,15,[i.n.b,e.C.b,e.k])),ue=a.c+a.b-$3(ce(se(sa,1),_s,30,15,[i.n.c,e.C.c,e.k])),g=F2e(V_e(l),e.t),pe=t==hn?qo:Gi,A=L.Jc();A.Ob();)S=f(A.Pb(),115),!(!S.c||S.c.d.c.length<=0)&&(J=S.b.Kf(),W=S.e,F=S.c,z=F.i,z.b=(v=F.n,F.e.a+v.b+v.c),z.a=(m=F.n,F.e.b+m.d+m.a),bH(Ee,WOe),F.f=Ee,Op(F,(Yl(),tp)),z.c=W.a-(z.b-J.a)/2,Le=h.Math.min(u,W.a),He=h.Math.max(ue,W.a+J.a),z.c<Le?z.c=Le:z.c+z.b>He&&(z.c=He-z.b),Qe(g.d,new cce(z,BCe(g,z))),pe=t==hn?h.Math.max(pe,W.b+S.b.Kf().b):h.Math.min(pe,W.b));for(pe+=t==hn?e.t:-e.t,ve=r3e((g.e=pe,g)),ve>0&&(f(Es(e.b,t),126).a.b=ve),T=L.Jc();T.Ob();)S=f(T.Pb(),115),!(!S.c||S.c.d.c.length<=0)&&(z=S.c.i,z.c-=S.e.a,z.d-=S.e.b)}function P9n(e,t){qde();var i,a,u,l,g,m,v,S,T,A,L,F,z,W;if(v=Eu(e,0)<0,v&&(e=T0(e)),Eu(e,0)==0)switch(t){case 0:return"0";case 1:return OL;case 2:return"0.00";case 3:return"0.000";case 4:return"0.0000";case 5:return"0.00000";case 6:return"0.000000";default:return F=new $v,t<0?F.a+="0E+":F.a+="0E",F.a+=t==Ra?"2147483648":""+-t,F.a}T=18,A=Pe(Xf,_m,30,T+1,15,1),i=T,W=e;do S=W,W=pG(W,10),A[--i]=mr(Ua(48,hh(S,Na(W,10))))&To;while(Eu(W,0)!=0);if(u=hh(hh(hh(T,i),t),1),t==0)return v&&(A[--i]=45),bm(A,i,T-i);if(t>0&&Eu(u,-6)>=0){if(Eu(u,0)>=0){for(l=i+mr(u),m=T-1;m>=l;m--)A[m+1]=A[m];return A[++l]=46,v&&(A[--i]=45),bm(A,i,T-i+1)}for(g=2;$ae(g,Ua(T0(u),1));g++)A[--i]=48;return A[--i]=46,A[--i]=48,v&&(A[--i]=45),bm(A,i,T-i)}return z=i+1,a=T,L=new RN,v&&(L.a+="-"),a-z>=1?(Qy(L,A[i]),L.a+=".",L.a+=bm(A,i+1,T-i-1)):L.a+=bm(A,i,T-i),L.a+="E",Eu(u,0)>0&&(L.a+="+"),L.a+=""+V7(u),L.a}function vkt(e){Wk(e,new zS(gY(qk(zk(Gk(Hk(new ug,Uf),"ELK Radial"),'A radial layout provider which is based on the algorithm of Peter Eades published in "Drawing free trees.", published by International Institute for Advanced Study of Social Information Science, Fujitsu Limited in 1991. The radial layouter takes a tree and places the nodes in radial order around the root. The nodes of the same tree level are placed on the same radius.'),new z8),Uf))),Je(e,Uf,YQ,st(EPt)),Je(e,Uf,UT,st(SPt)),Je(e,Uf,uA,st(bPt)),Je(e,Uf,DR,st(vPt)),Je(e,Uf,LR,st(wPt)),Je(e,Uf,$L,st(mPt)),Je(e,Uf,d$,st(jMe)),Je(e,Uf,FL,st(yPt)),Je(e,Uf,zhe,st(lbe)),Je(e,Uf,Uhe,st(dbe)),Je(e,Uf,nee,st(FMe)),Je(e,Uf,Hhe,st(fbe)),Je(e,Uf,Ghe,st(BMe)),Je(e,Uf,QRe,st(UMe)),Je(e,Uf,ZRe,st($Me)),Je(e,Uf,KRe,st(one)),Je(e,Uf,YRe,st(ane)),Je(e,Uf,XRe,st(Kq)),Je(e,Uf,JRe,st(zMe)),Je(e,Uf,WRe,st(MMe))}function WS(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;if(J=new Ke(e.g,e.f),W=RAe(e),W.a=h.Math.max(W.a,t),W.b=h.Math.max(W.b,i),He=W.a/J.a,T=W.b/J.b,Ee=W.a-J.a,v=W.b-J.b,a)for(g=so(e)?f(Fe(so(e),(Rr(),BE)),86):f(Fe(e,(Rr(),BE)),86),m=Ae(Fe(e,(Rr(),OF)))===Ae((oa(),uu)),pe=new Nn((!e.c&&(e.c=new Be(_d,e,9,9)),e.c));pe.e!=pe.i.gc();)switch(ue=f(Mn(pe),124),ve=f(Fe(ue,qA),64),ve==(tt(),fc)&&(ve=QIe(ue,g),pi(ue,qA,ve)),ve.g){case 1:m||md(ue,ue.i*He);break;case 2:md(ue,ue.i+Ee),m||bd(ue,ue.j*T);break;case 3:m||md(ue,ue.i*He),bd(ue,ue.j+v);break;case 4:m||bd(ue,ue.j*T)}if(x3(e,W.a,W.b),u)for(L=new Nn((!e.n&&(e.n=new Be(hc,e,1,7)),e.n));L.e!=L.i.gc();)A=f(Mn(L),153),F=A.i+A.g/2,z=A.j+A.f/2,Le=F/J.a,S=z/J.b,Le+S>=1&&(Le-S>0&&z>=0?(md(A,A.i+Ee),bd(A,A.j+v*S)):Le-S<0&&F>=0&&(md(A,A.i+Ee*Le),bd(A,A.j+v)));return pi(e,(Rr(),UE),(Ud(),l=f(Tp($F),10),new Rf(l,f(Gh(l,l.length),10),0))),new Ke(He,T)}function bQ(e){var t,i,a,u,l,g,m,v,S,T,A;if(e==null)throw ie(new sm(bl));if(S=e,l=e.length,v=!1,l>0&&(t=(gn(0,e.length),e.charCodeAt(0)),(t==45||t==43)&&(e=(gn(1,e.length+1),e.substr(1)),--l,v=t==45)),l==0)throw ie(new sm(XS+S+'"'));for(;e.length>0&&(gn(0,e.length),e.charCodeAt(0)==48);)e=(gn(1,e.length+1),e.substr(1)),--l;if(l>(ZSt(),BAt)[10])throw ie(new sm(XS+S+'"'));for(u=0;u<l;u++)if(dvt((gn(u,e.length),e.charCodeAt(u)))==-1)throw ie(new sm(XS+S+'"'));for(A=0,g=x6e[10],T=qpe[10],m=T0(C6e[10]),i=!0,a=l%g,a>0&&(A=-parseInt((ga(0,a,e.length),e.substr(0,a)),10),e=(gn(a,e.length+1),e.substr(a)),l-=a,i=!1);l>=g;){if(a=parseInt((ga(0,g,e.length),e.substr(0,g)),10),e=(gn(g,e.length+1),e.substr(g)),l-=g,i)i=!1;else{if(Eu(A,m)<0)throw ie(new sm(XS+S+'"'));A=Na(A,T)}A=hh(A,a)}if(Eu(A,0)>0)throw ie(new sm(XS+S+'"'));if(!v&&(A=T0(A),Eu(A,0)<0))throw ie(new sm(XS+S+'"'));return A}function EOe(e){Jde();var t,i,a,u,l,g,m,v;if(e==null)return null;if(u=cm(e,gl(37)),u<0)return e;for(v=new ef((ga(0,u,e.length),e.substr(0,u))),t=Pe($l,hA,30,4,15,1),m=0,a=0,g=e.length;u<g;u++)if(gn(u,e.length),e.charCodeAt(u)==37&&e.length>u+2&&Yue((gn(u+1,e.length),e.charCodeAt(u+1)),R7e,P7e)&&Yue((gn(u+2,e.length),e.charCodeAt(u+2)),R7e,P7e))if(i=wvn((gn(u+1,e.length),e.charCodeAt(u+1)),(gn(u+2,e.length),e.charCodeAt(u+2))),u+=2,a>0?(i&192)==128?t[m++]=i<<24>>24:a=0:i>=128&&((i&224)==192?(t[m++]=i<<24>>24,a=2):(i&240)==224?(t[m++]=i<<24>>24,a=3):(i&248)==240&&(t[m++]=i<<24>>24,a=4)),a>0){if(m==a){switch(m){case 2:{Qy(v,((t[0]&31)<<6|t[1]&63)&To);break}case 3:{Qy(v,((t[0]&15)<<12|(t[1]&63)<<6|t[2]&63)&To);break}}m=0,a=0}}else{for(l=0;l<m;++l)Qy(v,t[l]&To);m=0,v.a+=String.fromCharCode(i)}else{for(l=0;l<m;++l)Qy(v,t[l]&To);m=0,Qy(v,(gn(u,e.length),e.charCodeAt(u)))}return v.a}function wkt(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z;if(F=so(Us(f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84))),z=so(Us(f(be((!e.c&&(e.c=new Qt(Yn,e,5,8)),e.c),0),84))),A=F==z,m=new ha,t=f(Fe(e,(rZ(),Z9e)),77),t&&t.b>=2){if((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i==0)i=(Bv(),u=new em,u),Qn((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a),i);else if((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i>1)for(L=new HN((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a));L.e!=L.i.gc();)Mj(L);tOe(t,f(be((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a),0),168))}if(A)for(a=new Nn((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a));a.e!=a.i.gc();)for(i=f(Mn(a),168),S=new Nn((!i.a&&(i.a=new So(yf,i,5)),i.a));S.e!=S.i.gc();)v=f(Mn(S),369),m.a=h.Math.max(m.a,v.a),m.b=h.Math.max(m.b,v.b);for(g=new Nn((!e.n&&(e.n=new Be(hc,e,1,7)),e.n));g.e!=g.i.gc();)l=f(Mn(g),153),T=f(Fe(l,RF),8),T&&Of(l,T.a,T.b),A&&(m.a=h.Math.max(m.a,l.i+l.g),m.b=h.Math.max(m.b,l.j+l.f));return m}function ykt(e,t,i,a,u){var l,g,m;if(fmt(e,t),g=t[0],l=_a(i.c,0),m=-1,bCe(i))if(a>0){if(g+a>e.length)return!1;m=WZ((ga(0,g+a,e.length),e.substr(0,g+a)),t)}else m=WZ(e,t);switch(l){case 71:return m=Q3(e,g,ce(se(ft,1),Ye,2,6,[oTt,aTt]),t),u.e=m,!0;case 77:return cPn(e,t,u,m,g);case 76:return uPn(e,t,u,m,g);case 69:return bRn(e,t,g,u);case 99:return vRn(e,t,g,u);case 97:return m=Q3(e,g,ce(se(ft,1),Ye,2,6,["AM","PM"]),t),u.b=m,!0;case 121:return lPn(e,t,g,m,i,u);case 100:return m<=0?!1:(u.c=m,!0);case 83:return m<0?!1:MCn(m,g,t[0],u);case 104:m==12&&(m=0);case 75:case 72:return m<0?!1:(u.f=m,u.g=!1,!0);case 107:return m<0?!1:(u.f=m,u.g=!0,!0);case 109:return m<0?!1:(u.j=m,!0);case 115:return m<0?!1:(u.n=m,!0);case 90:if(g<e.length&&(gn(g,e.length),e.charCodeAt(g)==90))return++t[0],u.o=0,!0;case 122:case 118:return zOn(e,g,t,u);default:return!1}}function L9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt;for(ve=t.c.length,u=new nA(e.a,i,null,null),xt=Pe(sa,_s,30,ve,15,1),W=Pe(sa,_s,30,ve,15,1),z=Pe(sa,_s,30,ve,15,1),J=0,m=0;m<ve;m++)W[m]=Qr,z[m]=Ra;for(v=0;v<ve;v++)for(a=(Gt(v,t.c.length),f(t.c[v],185)),xt[v]=_de(a),xt[J]>xt[v]&&(J=v),A=new ne(e.a.b);A.a<A.c.c.length;)for(T=f(ee(A),25),pe=new ne(T.a);pe.a<pe.c.c.length;)ue=f(ee(pe),9),He=ke(a.p[ue.p])+ke(a.d[ue.p]),W[v]=h.Math.min(W[v],He),z[v]=h.Math.max(z[v],He+ue.o.b);for(nt=Pe(sa,_s,30,ve,15,1),S=0;S<ve;S++)(Gt(S,t.c.length),f(t.c[S],185)).o==(vg(),$E)?nt[S]=W[J]-W[S]:nt[S]=z[J]-z[S];for(l=Pe(sa,_s,30,ve,15,1),F=new ne(e.a.b);F.a<F.c.c.length;)for(L=f(ee(F),25),Le=new ne(L.a);Le.a<Le.c.c.length;){for(Ee=f(ee(Le),9),g=0;g<ve;g++)l[g]=ke((Gt(g,t.c.length),f(t.c[g],185)).p[Ee.p])+ke((Gt(g,t.c.length),f(t.c[g],185)).d[Ee.p])+nt[g];Vnt(l,Gmt(Re.prototype.Ke,Re,[])),u.p[Ee.p]=(l[1]+l[2])/2,u.d[Ee.p]=0}return u}function D9n(e,t,i){var a,u,l,g,m;switch(a=t.i,l=e.i.o,u=e.i.d,m=e.n,g=uc(ce(se(Zo,1),Ye,8,0,[m,e.a])),e.j.g){case 1:um(t,(pl(),np)),a.d=-u.d-i-a.a,f(f(ct(t.d,0),187).mf((Ge(),JR)),289)==(df(),Rb)?(Op(t,(Yl(),tp)),a.c=g.a-ke(Oe(V(e,TA)))-i-a.b):(Op(t,(Yl(),Om)),a.c=g.a+ke(Oe(V(e,TA)))+i);break;case 2:Op(t,(Yl(),Om)),a.c=l.a+u.c+i,f(f(ct(t.d,0),187).mf((Ge(),JR)),289)==(df(),Rb)?(um(t,(pl(),np)),a.d=g.b-ke(Oe(V(e,TA)))-i-a.a):(um(t,(pl(),jp)),a.d=g.b+ke(Oe(V(e,TA)))+i);break;case 3:um(t,(pl(),jp)),a.d=l.b+u.a+i,f(f(ct(t.d,0),187).mf((Ge(),JR)),289)==(df(),Rb)?(Op(t,(Yl(),tp)),a.c=g.a-ke(Oe(V(e,TA)))-i-a.b):(Op(t,(Yl(),Om)),a.c=g.a+ke(Oe(V(e,TA)))+i);break;case 4:Op(t,(Yl(),tp)),a.c=-u.b-i-a.b,f(f(ct(t.d,0),187).mf((Ge(),JR)),289)==(df(),Rb)?(um(t,(pl(),np)),a.d=g.b-ke(Oe(V(e,TA)))-i-a.a):(um(t,(pl(),jp)),a.d=g.b+ke(Oe(V(e,TA)))+i)}}function M9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe;for(t.Sg("Interactive node layering",1),i=new et,L=new ne(e.a);L.a<L.c.c.length;){for(T=f(ee(L),9),v=T.n.a,m=v+T.o.a,m=h.Math.max(v+1,m),pe=new la(i,0),a=null;pe.b<pe.d.gc();)if(J=(Ln(pe.b<pe.d.gc()),f(pe.d.Xb(pe.c=pe.b++),560)),J.c>=m){Ln(pe.b>0),pe.a.Xb(pe.c=--pe.b);break}else J.a>v&&(a?($o(a.b,J.b),a.a=h.Math.max(a.a,J.a),hd(pe)):(Qe(J.b,T),J.c=h.Math.min(J.c,v),J.a=h.Math.max(J.a,m),a=J));a||(a=new drt,a.c=v,a.a=m,nT(pe,a),Qe(a.b,T))}for(g=e.b,S=0,ue=new ne(i);ue.a<ue.c.c.length;)for(J=f(ee(ue),560),u=new ru(e),u.p=S++,vn(g.c,u),F=new ne(J.b);F.a<F.c.c.length;)T=f(ee(F),9),jo(T,u),T.p=0;for(A=new ne(e.a);A.a<A.c.c.length;)if(T=f(ee(A),9),T.p==0)for(W=dSt(T,e);W.a.gc()!=0;)z=f(W.a.ec().Jc().Pb(),9),W.a.Ac(z)!=null,Ba(W,dSt(z,e));for(l=new la(g,0);l.b<l.d.gc();)(Ln(l.b<l.d.gc()),f(l.d.Xb(l.c=l.b++),25)).a.c.length==0&&hd(l);e.a.c.length=0,t.Tg()}function j9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W;for(i.Sg(Bxt,1),!t.a&&(t.a=new Be(Ur,t,10,11)),a=ke(Oe(Fe(t,(Ag(),lne)))),T=ke(Oe(Fe(t,SF))),L=f(Fe(t,EF),107),F=new xSe(a,T),l=$kt(F,t,L),$mt(t,F),m=f(Fe(t,s8e),17).a;m>1;){if(u=r5n(t),A=l.g,z=f(Fe(t,EF),107),W=ke(Oe(Fe(t,lne))),(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i>1&&ke(Oe(Fe(t,(wb(),Ebe))))!=Gi&&(l.c+(z.b+z.c))/(l.b+(z.d+z.a))<W?pi(u,(wb(),fx),ke(Oe(Fe(t,fx)))+ke(Oe(Fe(t,Ebe)))):(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i>1&&ke(Oe(Fe(t,(wb(),ybe))))!=Gi&&(l.c+(z.b+z.c))/(l.b+(z.d+z.a))>W&&pi(u,(wb(),fx),h.Math.max(ke(Oe(Fe(t,yF))),ke(Oe(Fe(u,fx)))-ke(Oe(Fe(t,ybe))))),F=new xSe(a,T),v=$kt(F,u,L),S=v.g,S>=A&&S==S){for(g=0;g<(!u.a&&(u.a=new Be(Ur,u,10,11)),u.a).i;g++)hEt(e,f(be((!u.a&&(u.a=new Be(Ur,u,10,11)),u.a),g),27),f(be((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a),g),27));$mt(t,F),Xyn(l,v.c),Yyn(l,v.b)}--m}pi(t,(wb(),CD),l.b),pi(t,t4,l.c),i.Tg()}function $9n(e,t,i,a,u,l,g){var m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji;for(F=0,Kt=0,v=new ne(e);v.a<v.c.c.length;)m=f(ee(v),27),ikt(m),F=h.Math.max(F,m.g),Kt+=m.g*m.f;for(z=Kt/e.c.length,Ct=eIn(e,z),Kt+=e.c.length*Ct,F=h.Math.max(F,h.Math.sqrt(Kt*g))+i.b,$r=i.b,ji=i.d,L=0,T=i.b+i.c,xt=new ki,Mr(xt,rt(0)),He=new ki,S=new la(e,0);S.b<S.d.gc();)m=(Ln(S.b<S.d.gc()),f(S.d.Xb(S.c=S.b++),27)),In=m.g,A=m.f,$r+In>F&&(l&&(Ma(He,L),Ma(xt,rt(S.b-1))),$r=i.b,ji+=L+t,L=0,T=h.Math.max(T,i.b+i.c+In)),md(m,$r),bd(m,ji),T=h.Math.max(T,$r+In+i.c),L=h.Math.max(L,A),$r+=In+t;if(T=h.Math.max(T,a),tn=ji+L+i.a,tn<u&&(L+=u-tn,tn=u),l)for($r=i.b,S=new la(e,0),Ma(xt,rt(e.c.length)),nt=tr(xt,0),ue=f(Zn(nt),17).a,Ma(He,L),Le=tr(He,0),Ee=0;S.b<S.d.gc();)S.b==ue&&($r=i.b,Ee=ke(Oe(Zn(Le))),ue=f(Zn(nt),17).a),m=(Ln(S.b<S.d.gc()),f(S.d.Xb(S.c=S.b++),27)),pe=m.f,NS(m,Ee),W=Ee,S.b==ue&&(J=T-$r-i.c,ve=m.g,RS(m,J),LAe(m,new Ke(J,W),new Ke(ve,pe))),$r+=m.g+t;return new Ke(T,tn)}function F9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct;for(t.Sg("Compound graph postprocessor",1),i=lt(ut(V(e,(ot(),Sme)))),m=f(V(e,(Ge(),$Pe)),228),T=new Mo,ue=m.ec().Jc();ue.Ob();){for(J=f(ue.Pb(),18),g=new Kl(m.cc(J)),Ht(),na(g,new ZEe(e)),Le=Xkn((Gt(0,g.c.length),f(g.c[0],248))),nt=Mbt(f(ct(g,g.c.length-1),248)),ve=Le.i,UP(nt.i,ve)?pe=ve.e:pe=ra(ve),A=V3n(J,g),nf(J.a),L=null,l=new ne(g);l.a<l.c.c.length;)u=f(ee(l),248),W=new ha,IAe(W,u.a,pe),F=u.b,a=new fd,ale(a,0,F.a),wT(a,W),Ee=new ja(wg(F.c)),He=new ja(wg(F.d)),ii(Ee,W),ii(He,W),L&&(a.b==0?z=He:z=(Ln(a.b!=0),f(a.a.a.c,8)),xt=h.Math.abs(L.a-z.a)>Tm,Ct=h.Math.abs(L.b-z.b)>Tm,(!i&&xt&&Ct||i&&(xt||Ct))&&Mr(J.a,Ee)),Ba(J.a,a),a.b==0?L=Ee:L=(Ln(a.b!=0),f(a.c.b.c,8)),_Tn(F,A,W),Mbt(u)==nt&&(ra(nt.i)!=u.a&&(W=new ha,IAe(W,ra(nt.i),pe)),$e(J,ime,W)),DNn(F,J,pe),T.a.yc(F,T);Ia(J,Le),ua(J,nt)}for(S=T.a.ec().Jc();S.Ob();)v=f(S.Pb(),18),Ia(v,null),ua(v,null);t.Tg()}function B9n(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(u=f(V(e,(Ec(),E_)),86),T=u==(ko(),Ds)||u==Gs?jg:Gs,i=f($d(Jr(new Ft(null,new Bt(e.b,16)),new o0),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),v=f($d(Lu(i.Mc(),new Rtt(t)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[Zl]))),15),v.Fc(f($d(Lu(i.Mc(),new Ptt(t)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[Zl]))),16)),v.gd(new Ltt(T)),L=new g0(new Dtt(u)),a=new Wn,m=v.Jc();m.Ob();)g=f(m.Pb(),239),S=f(g.a,39),lt(ut(g.c))?(L.a.yc(S,(mn(),yw))==null,new g3(L.a.Xc(S,!1)).a.gc()>0&&qr(a,S,f(new g3(L.a.Xc(S,!1)).a.Tc(),39)),new g3(L.a.$c(S,!0)).a.gc()>1&&qr(a,G0t(L,S),S)):(new g3(L.a.Xc(S,!1)).a.gc()>0&&(l=f(new g3(L.a.Xc(S,!1)).a.Tc(),39),Ae(l)===Ae(ac(Is(a.f,S)))&&f(V(S,(_i(),qme)),15).Ec(l)),new g3(L.a.$c(S,!0)).a.gc()>1&&(A=G0t(L,S),Ae(ac(Is(a.f,A)))===Ae(S)&&f(V(A,(_i(),qme)),15).Ec(S)),L.a.Ac(S)!=null)}function Ekt(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;if(e.gc()==1)return f(e.Xb(0),234);if(e.gc()<=0)return new vJ;for(u=e.Jc();u.Ob();){for(i=f(u.Pb(),234),z=0,T=Qr,A=Qr,v=Ra,S=Ra,F=new ne(i.e);F.a<F.c.c.length;)L=f(ee(F),151),z+=f(V(L,(Qh(),GR)),17).a,T=h.Math.min(T,L.d.a-L.e.a/2),A=h.Math.min(A,L.d.b-L.e.b/2),v=h.Math.max(v,L.d.a+L.e.a/2),S=h.Math.max(S,L.d.b+L.e.b/2);$e(i,(Qh(),GR),rt(z)),$e(i,(ew(),B$),new Ke(T,A)),$e(i,Eq,new Ke(v,S))}for(Ht(),e.gd(new t0),W=new vJ,Rc(W,f(e.Xb(0),102)),m=0,pe=0,l=e.Jc();l.Ob();)i=f(l.Pb(),234),J=Ho($a(f(V(i,(ew(),Eq)),8)),f(V(i,B$),8)),m=h.Math.max(m,J.a),pe+=J.a*J.b;for(m=h.Math.max(m,h.Math.sqrt(pe)*ke(Oe(V(W,(Qh(),QIt))))),ue=ke(Oe(V(W,Fee))),ve=0,Ee=0,g=0,t=ue,a=e.Jc();a.Ob();)i=f(a.Pb(),234),J=Ho($a(f(V(i,(ew(),Eq)),8)),f(V(i,B$),8)),ve+J.a>m&&(ve=0,Ee+=g+ue,g=0),ELn(W,i,ve,Ee),t=h.Math.max(t,ve+J.a),g=h.Math.max(g,J.b),ve+=J.a+ue;return W}function U9n(e){ZIe();var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;if(e==null||(l=AJ(e),z=Kxn(l),z%4!=0))return null;if(W=z/4|0,W==0)return Pe($l,hA,30,0,15,1);for(A=null,t=0,i=0,a=0,u=0,g=0,m=0,v=0,S=0,F=0,L=0,T=0,A=Pe($l,hA,30,W*3,15,1);F<W-1;F++){if(!Cz(g=l[T++])||!Cz(m=l[T++])||!Cz(v=l[T++])||!Cz(S=l[T++]))return null;t=zg[g],i=zg[m],a=zg[v],u=zg[S],A[L++]=(t<<2|i>>4)<<24>>24,A[L++]=((i&15)<<4|a>>2&15)<<24>>24,A[L++]=(a<<6|u)<<24>>24}return!Cz(g=l[T++])||!Cz(m=l[T++])?null:(t=zg[g],i=zg[m],v=l[T++],S=l[T++],zg[v]==-1||zg[S]==-1?v==61&&S==61?(i&15)!=0?null:(J=Pe($l,hA,30,F*3+1,15,1),au(A,0,J,0,F*3),J[L]=(t<<2|i>>4)<<24>>24,J):v!=61&&S==61?(a=zg[v],(a&3)!=0?null:(J=Pe($l,hA,30,F*3+2,15,1),au(A,0,J,0,F*3),J[L++]=(t<<2|i>>4)<<24>>24,J[L]=((i&15)<<4|a>>2&15)<<24>>24,J)):null:(a=zg[v],u=zg[S],A[L++]=(t<<2|i>>4)<<24>>24,A[L++]=((i&15)<<4|a>>2&15)<<24>>24,A[L++]=(a<<6|u)<<24>>24,A))}function z9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le;for(t.Sg(dxt,1),z=f(V(e,(ot(),P1)),220),u=new ne(e.b);u.a<u.c.c.length;)for(a=f(ee(u),25),S=EH(a.a),g=S,m=0,v=g.length;m<v;++m)if(l=g[m],l.k==(cn(),xu)){if(z==(k1(),c4))for(A=new ne(l.j);A.a<A.c.c.length;)T=f(ee(A),11),T.e.c.length==0||m3n(T),T.g.c.length==0||b3n(T);else if(me(V(l,(Ge(),oi)),18))J=f(V(l,oi),18),ue=f(L0(l,(tt(),pn)).Jc().Pb(),11),pe=f(L0(l,_n).Jc().Pb(),11),ve=f(V(ue,oi),11),Ee=f(V(pe,oi),11),Ia(J,Ee),ua(J,ve),Le=new ja(pe.i.n),Le.a=uc(ce(se(Zo,1),Ye,8,0,[Ee.i.n,Ee.n,Ee.a])).a,Mr(J.a,Le),Le=new ja(ue.i.n),Le.a=uc(ce(se(Zo,1),Ye,8,0,[ve.i.n,ve.n,ve.a])).a,Mr(J.a,Le);else{if(l.j.c.length>=2){for(W=!0,L=new ne(l.j),i=f(ee(L),11),F=null;L.a<L.c.c.length;)if(F=i,i=f(ee(L),11),!ei(V(F,oi),V(i,oi))){W=!1;break}}else W=!1;for(A=new ne(l.j);A.a<A.c.c.length;)T=f(ee(A),11),T.e.c.length==0||N5n(T,W),T.g.c.length==0||R5n(T,W)}jo(l,null)}t.Tg()}function H9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le;for(m=new ne(e.a.b);m.a<m.c.c.length;)for(l=f(ee(m),25),ve=new ne(l.a);ve.a<ve.c.c.length;)pe=f(ee(ve),9),t.g[pe.p]=pe,t.a[pe.p]=pe,t.d[pe.p]=0;for(v=e.a.b,t.c==(fm(),G0)&&(v=Bd(v)),g=v.Jc();g.Ob();)for(l=f(g.Pb(),25),F=-1,L=l.a,t.o==(vg(),Mg)&&(F=Qr,L=Bd(L)),Le=L.Jc();Le.Ob();)if(Ee=f(Le.Pb(),9),A=null,t.c==G0?A=f(ct(e.b.f,Ee.p),15):A=f(ct(e.b.b,Ee.p),15),A.gc()>0)if(a=A.gc(),S=Aa(h.Math.floor((a+1)/2))-1,u=Aa(h.Math.ceil((a+1)/2))-1,t.o==Mg)for(T=u;T>=S;T--)t.a[Ee.p]==Ee&&(W=f(A.Xb(T),47),z=f(W.a,9),!xp(i,W.b)&&F>e.b.e[z.p]&&(t.a[z.p]=Ee,t.g[Ee.p]=t.g[z.p],t.a[Ee.p]=t.g[Ee.p],t.f[t.g[Ee.p].p]=(mn(),!!(lt(t.f[t.g[Ee.p].p])&Ee.k==(cn(),uo))),F=e.b.e[z.p]));else for(T=S;T<=u;T++)t.a[Ee.p]==Ee&&(ue=f(A.Xb(T),47),J=f(ue.a,9),!xp(i,ue.b)&&F<e.b.e[J.p]&&(t.a[J.p]=Ee,t.g[Ee.p]=t.g[J.p],t.a[Ee.p]=t.g[Ee.p],t.f[t.g[Ee.p].p]=(mn(),!!(lt(t.f[t.g[Ee.p].p])&Ee.k==(cn(),uo))),F=e.b.e[J.p]))}function Skt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt;return ve=e.c[(Gt(0,t.c.length),f(t.c[0],18)).p],nt=e.c[(Gt(1,t.c.length),f(t.c[1],18)).p],ve.a.e.e-ve.a.a-(ve.b.e.e-ve.b.a)==0&&nt.a.e.e-nt.a.a-(nt.b.e.e-nt.b.a)==0||(ue=ve.b.e.f,!me(ue,9))?!1:(J=f(ue,9),Le=e.i[J.p],He=J.c?Oc(J.c.a,J,0):-1,l=Gi,He>0&&(u=f(ct(J.c.a,He-1),9),g=e.i[u.p],xt=h.Math.ceil(A3(e.n,u,J)),l=Le.a.e-J.d.d-(g.a.e+u.o.b+u.d.a)-xt),S=Gi,He<J.c.a.c.length-1&&(v=f(ct(J.c.a,He+1),9),T=e.i[v.p],xt=h.Math.ceil(A3(e.n,v,J)),S=T.a.e-v.d.d-(Le.a.e+J.o.b+J.d.a)-xt),i&&(gg(),Kh(Cm),h.Math.abs(l-S)<=Cm||l==S||isNaN(l)&&isNaN(S))?!0:(a=wce(ve.a),m=-wce(ve.b),A=-wce(nt.a),pe=wce(nt.b),W=ve.a.e.e-ve.a.a-(ve.b.e.e-ve.b.a)>0&&nt.a.e.e-nt.a.a-(nt.b.e.e-nt.b.a)<0,z=ve.a.e.e-ve.a.a-(ve.b.e.e-ve.b.a)<0&&nt.a.e.e-nt.a.a-(nt.b.e.e-nt.b.a)>0,F=ve.a.e.e+ve.b.a<nt.b.e.e+nt.a.a,L=ve.a.e.e+ve.b.a>nt.b.e.e+nt.a.a,Ee=0,!W&&!z&&(L?l+A>0?Ee=A:S-a>0&&(Ee=a):F&&(l+m>0?Ee=m:S-pe>0&&(Ee=pe))),Le.a.e+=Ee,Le.b&&(Le.d.e+=Ee),!1))}function _kt(e,t,i){var a,u,l,g,m,v,S,T,A,L;if(a=new qh(t.Jf().a,t.Jf().b,t.Kf().a,t.Kf().b),u=new BN,e.c)for(g=new ne(t.Pf());g.a<g.c.c.length;)l=f(ee(g),187),u.c=l.Jf().a+t.Jf().a,u.d=l.Jf().b+t.Jf().b,u.b=l.Kf().a,u.a=l.Kf().b,jj(a,u);for(S=new ne(t.Vf());S.a<S.c.c.length;){if(v=f(ee(S),831),T=v.Jf().a+t.Jf().a,A=v.Jf().b+t.Jf().b,e.e&&(u.c=T,u.d=A,u.b=v.Kf().a,u.a=v.Kf().b,jj(a,u)),e.d)for(g=new ne(v.Pf());g.a<g.c.c.length;)l=f(ee(g),187),u.c=l.Jf().a+T,u.d=l.Jf().b+A,u.b=l.Kf().a,u.a=l.Kf().b,jj(a,u);if(e.b){if(L=new Ke(-i,-i),f(t.mf((Rr(),px)),181).Gc((Xl(),Lw)))for(g=new ne(v.Pf());g.a<g.c.c.length;)l=f(ee(g),187),L.a+=l.Kf().a+i,L.b+=l.Kf().b+i;L.a=h.Math.max(L.a,0),L.b=h.Math.max(L.b,0),WSt(a,v.Uf(),v.Sf(),t,v,L,i)}}e.b&&WSt(a,t.Uf(),t.Sf(),t,null,null,i),m=new Wse(t.Tf()),m.d=h.Math.max(0,t.Jf().b-a.d),m.a=h.Math.max(0,a.d+a.a-(t.Jf().b+t.Kf().b)),m.b=h.Math.max(0,t.Jf().a-a.c),m.c=h.Math.max(0,a.c+a.b-(t.Jf().a+t.Kf().a)),t.Xf(m)}function G9n(){var e=["\\u0000","\\u0001","\\u0002","\\u0003","\\u0004","\\u0005","\\u0006","\\u0007","\\b","\\t","\\n","\\u000B","\\f","\\r","\\u000E","\\u000F","\\u0010","\\u0011","\\u0012","\\u0013","\\u0014","\\u0015","\\u0016","\\u0017","\\u0018","\\u0019","\\u001A","\\u001B","\\u001C","\\u001D","\\u001E","\\u001F"];return e[34]='\\"',e[92]="\\\\",e[173]="\\u00ad",e[1536]="\\u0600",e[1537]="\\u0601",e[1538]="\\u0602",e[1539]="\\u0603",e[1757]="\\u06dd",e[1807]="\\u070f",e[6068]="\\u17b4",e[6069]="\\u17b5",e[8203]="\\u200b",e[8204]="\\u200c",e[8205]="\\u200d",e[8206]="\\u200e",e[8207]="\\u200f",e[8232]="\\u2028",e[8233]="\\u2029",e[8234]="\\u202a",e[8235]="\\u202b",e[8236]="\\u202c",e[8237]="\\u202d",e[8238]="\\u202e",e[8288]="\\u2060",e[8289]="\\u2061",e[8290]="\\u2062",e[8291]="\\u2063",e[8292]="\\u2064",e[8298]="\\u206a",e[8299]="\\u206b",e[8300]="\\u206c",e[8301]="\\u206d",e[8302]="\\u206e",e[8303]="\\u206f",e[65279]="\\ufeff",e[65529]="\\ufff9",e[65530]="\\ufffa",e[65531]="\\ufffb",e}function kkt(e){Wk(e,new zS(B2e(gY(qk(zk(Gk(Hk(new ug,Zu),"ELK Force"),"Force-based algorithm provided by the Eclipse Layout Kernel. Implements methods that follow physical analogies by simulating forces that move the nodes into a balanced distribution. Currently the original Eades model and the Fruchterman - Reingold model are supported."),new r6),Zu),xi((pL(),jne),ce(se($ne,1),Ve,243,0,[Dne]))))),Je(e,Zu,XG,rt(1)),Je(e,Zu,UT,80),Je(e,Zu,Kfe,5),Je(e,Zu,ML,DL),Je(e,Zu,AQ,rt(1)),Je(e,Zu,u$,(mn(),!0)),Je(e,Zu,zT,e5e),Je(e,Zu,l$,st(J6e)),Je(e,Zu,Yfe,st(t5e)),Je(e,Zu,IQ,!1),Je(e,Zu,d$,st(Q6e)),Je(e,Zu,jL,st(rOt)),Je(e,Zu,LR,st(iOt)),Je(e,Zu,uA,st(nOt)),Je(e,Zu,$L,st(tOt)),Je(e,Zu,FL,st(aOt)),Je(e,Zu,CQ,st(Z6e)),Je(e,Zu,Vfe,st(cge)),Je(e,Zu,ZOe,st($ee)),Je(e,Zu,Wfe,st(sge)),Je(e,Zu,QOe,st(n5e)),Je(e,Zu,JG,st(fOt)),Je(e,Zu,ZG,st(hOt)),Je(e,Zu,QG,st(dOt)),Je(e,Zu,eq,st(lOt)),Je(e,Zu,ZS,r5e)}function q9n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue;for(F=f(ct((Gt(0,t.c.length),f(t.c[0],25)).a,a),9),i>0&&jo(F,(Gt(i,t.c.length),f(t.c[i],25))),l=0,L=!0,ue=Bd(rE(Fo(F))),v=ue.Jc();v.Ob();){for(m=f(v.Pb(),18),L=!1,A=m,S=0;S<i;S++)u=f0t(e,A),a+l>(Gt(S,t.c.length),f(t.c[S],25)).a.c.length?jo(u,(Gt(S,t.c.length),f(t.c[S],25))):sw(u,a+l,(Gt(S,t.c.length),f(t.c[S],25))),A=Cde(A,u);i>0&&(l+=1)}if(L){for(S=0;S<i;S++)u=new Tg(e),Wl(u,(cn(),_q)),a+l>(Gt(S,t.c.length),f(t.c[S],25)).a.c.length?jo(u,(Gt(S,t.c.length),f(t.c[S],25))):sw(u,a+l,(Gt(S,t.c.length),f(t.c[S],25)));i>0&&(l+=1)}for(g=!1,W=new Tn(Cn(ao(F).a.Jc(),new B));Kn(W);){for(z=f(Dn(W),18),A=z,T=i+1;T<t.c.length;T++)u=f0t(e,A),jo(u,(Gt(T,t.c.length),f(t.c[T],25))),A=Cde(A,u);for(S=0;S<=i;S++)g&&(J=new Tg(e),Wl(J,(cn(),U$)),a+1>(Gt(S,t.c.length),f(t.c[S],25)).a.c.length?jo(J,(Gt(S,t.c.length),f(t.c[S],25))):sw(J,a+1,(Gt(S,t.c.length),f(t.c[S],25))));g&&(l+=1),g=!0}return l>0?l-1:0}function dw(e,t){Xr();var i,a,u,l,g,m,v,S,T,A,L,F,z;if(f7(BD)==0){for(A=Pe(wjn,Ye,121,BMt.length,0,1),g=0;g<A.length;g++)A[g]=new of(4);for(a=new l7,l=0;l<ije.length;l++){if(T=new of(4),l<84?(m=l*2,F=(gn(m,Ppe.length),Ppe.charCodeAt(m)),L=(gn(m+1,Ppe.length),Ppe.charCodeAt(m+1)),Su(T,F,L)):(m=(l-84)*2,Su(T,oje[m],oje[m+1])),v=ije[l],Mt(v,"Specials")&&Su(T,65520,65533),Mt(v,wAt)&&(Su(T,983040,1048573),Su(T,1048576,1114109)),Os(BD,v,T),Os(YF,v,LT(T)),S=a.a.length,0<S?a.a=uh(a.a,0,0):0>S&&(a.a+=Vst(Pe(Xf,_m,30,-S,15,1))),a.a+="Is",cm(v,gl(32))>=0)for(u=0;u<v.length;u++)gn(u,v.length),v.charCodeAt(u)!=32&&_P(a,(gn(u,v.length),v.charCodeAt(u)));else a.a+=""+v;d3e(a.a,v,!0)}d3e(Rpe,"Cn",!1),d3e(u6e,"Cn",!0),i=new of(4),Su(i,0,JL),Os(BD,"ALL",i),Os(YF,"ALL",LT(i)),!_x&&(_x=new Wn),Os(_x,Rpe,Rpe),!_x&&(_x=new Wn),Os(_x,u6e,u6e),!_x&&(_x=new Wn),Os(_x,"ALL","ALL")}return z=f(wu(t?BD:YF,e),136),z}function Tkt(e){Wk(e,new zS(B2e(gY(qk(zk(Gk(Hk(new ug,Qu),"ELK Mr. Tree"),"Tree-based algorithm provided by the Eclipse Layout Kernel. Computes a spanning tree of the input graph and arranges all nodes according to the resulting parent-children hierarchy. I pity the fool who doesn't use Mr. Tree Layout."),new LO),Dxt),At((pL(),d1e))))),Je(e,Qu,zT,vMe),Je(e,Qu,UT,20),Je(e,Qu,Ohe,3),Je(e,Qu,ML,DL),Je(e,Qu,XG,rt(1)),Je(e,Qu,u$,(mn(),!0)),Je(e,Qu,aq,st(hMe)),Je(e,Qu,Nhe,pMe),Je(e,Qu,l$,st(L5t)),Je(e,Qu,XQ,st(D5t)),Je(e,Qu,uA,st(j5t)),Je(e,Qu,jL,st($5t)),Je(e,Qu,DR,st(F5t)),Je(e,Qu,LR,st(B5t)),Je(e,Qu,$L,st(M5t)),Je(e,Qu,d$,st(mMe)),Je(e,Qu,FL,st(U5t)),Je(e,Qu,FRe,st(SMe)),Je(e,Qu,URe,st(wMe)),Je(e,Qu,JG,st(q5t)),Je(e,Qu,ZG,st(V5t)),Je(e,Qu,QG,st(G5t)),Je(e,Qu,eq,st(H5t)),Je(e,Qu,ZS,EMe),Je(e,Qu,$Re,st(DA)),Je(e,Qu,BRe,st(Zme)),Je(e,Qu,jRe,st(Lm)),Je(e,Qu,DRe,st(fMe)),Je(e,Qu,MRe,st(gMe))}function xkt(e,t){var i,a,u,l,g,m,v,S,T,A,L;for(S=f(f(si(e.r,t),22),83),g=eNn(e,t),i=e.u.Gc((Xl(),MF)),v=S.Jc();v.Ob();)if(m=f(v.Pb(),115),!(!m.c||m.c.d.c.length<=0)){switch(L=m.b.Kf(),T=m.c,A=T.i,A.b=(l=T.n,T.e.a+l.b+l.c),A.a=(u=T.n,T.e.b+u.d+u.a),t.g){case 1:m.a?(A.c=(L.a-A.b)/2,Op(T,(Yl(),Ew))):g||i?(A.c=-A.b-e.s,Op(T,(Yl(),tp))):(A.c=L.a+e.s,Op(T,(Yl(),Om))),A.d=-A.a-e.t,um(T,(pl(),np));break;case 3:m.a?(A.c=(L.a-A.b)/2,Op(T,(Yl(),Ew))):g||i?(A.c=-A.b-e.s,Op(T,(Yl(),tp))):(A.c=L.a+e.s,Op(T,(Yl(),Om))),A.d=L.b+e.t,um(T,(pl(),jp));break;case 2:m.a?(a=e.v?A.a:f(ct(T.d,0),187).Kf().b,A.d=(L.b-a)/2,um(T,(pl(),Sw))):g||i?(A.d=-A.a-e.t,um(T,(pl(),np))):(A.d=L.b+e.t,um(T,(pl(),jp))),A.c=L.a+e.s,Op(T,(Yl(),Om));break;case 4:m.a?(a=e.v?A.a:f(ct(T.d,0),187).Kf().b,A.d=(L.b-a)/2,um(T,(pl(),Sw))):g||i?(A.d=-A.a-e.t,um(T,(pl(),np))):(A.d=L.b+e.t,um(T,(pl(),jp))),A.c=-A.b-e.s,Op(T,(Yl(),tp))}g=!1}}function V9n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe;if(L=!1,A=!1,T3(f(V(a,(ot(),Xi)),100))){g=!1,m=!1;e:for(z=new ne(a.j);z.a<z.c.c.length;)for(F=f(ee(z),11),J=pb(Df(ce(se(Gf,1),Jt,20,0,[new Z5(F),new _N(F)])));Kn(J);)if(W=f(Dn(J),11),!lt(ut(V(W.i,Z$)))){if(F.j==(tt(),hn)){g=!0;break e}if(F.j==Bn){m=!0;break e}}L=m&&!g,A=g&&!m}if(!L&&!A&&a.b.c.length!=0){for(T=0,S=new ne(a.b);S.a<S.c.c.length;)v=f(ee(S),72),T+=v.n.b+v.o.b/2;T/=a.b.c.length,pe=T>=a.o.b/2}else pe=!A;pe?(ue=f(V(a,(Ge(),QR)),15),ue?L?l=ue:(u=f(V(a,KR),15),u?ue.gc()<=u.gc()?l=ue:l=u:(l=new et,$e(a,KR,l))):(l=new et,$e(a,QR,l))):(u=f(V(a,(Ge(),KR)),15),u?A?l=u:(ue=f(V(a,QR),15),ue?u.gc()<=ue.gc()?l=u:l=ue:(l=new et,$e(a,QR,l))):(l=new et,$e(a,KR,l))),l.Ec(e),$e(e,(Ge(),yte),i),t.d==i?(ua(t,null),i.e.c.length+i.g.c.length==0&&cc(i,null),XTn(i)):(Ia(t,null),i.e.c.length+i.g.c.length==0&&cc(i,null)),nf(t.a)}function W9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji;for(i.Sg("MinWidth layering",1),F=t.b,nt=t.a,ji=f(V(t,(ot(),HLe)),17).a,m=f(V(t,GLe),17).a,e.b=ke(Oe(V(t,op))),e.d=Gi,Ee=new ne(nt);Ee.a<Ee.c.c.length;)pe=f(ee(Ee),9),pe.k==(cn(),Yi)&&(Kt=pe.o.b,e.d=h.Math.min(e.d,Kt));for(e.d=h.Math.max(1,e.d),xt=nt.c.length,e.c=Pe(pr,Pr,30,xt,15,1),e.f=Pe(pr,Pr,30,xt,15,1),e.e=Pe(sa,_s,30,xt,15,1),S=0,e.a=0,Le=new ne(nt);Le.a<Le.c.c.length;)pe=f(ee(Le),9),pe.p=S++,e.c[pe.p]=Z1t(Fo(pe)),e.f[pe.p]=Z1t(ao(pe)),e.e[pe.p]=pe.o.b/e.d,e.a+=e.e[pe.p];for(e.b/=e.d,e.a/=xt,He=o4n(nt),na(nt,Jke(new Xet(e))),W=Gi,z=Qr,g=null,$r=ji,In=ji,l=m,u=m,ji<0&&($r=f(zDe.a.Gd(),17).a,In=f(zDe.b.Gd(),17).a),m<0&&(l=f(UDe.a.Gd(),17).a,u=f(UDe.b.Gd(),17).a),tn=$r;tn<=In;tn++)for(a=l;a<=u;a++)Ct=RDn(e,tn,a,nt,He),ue=ke(Oe(Ct.a)),L=f(Ct.b,15),J=L.gc(),(ue<W||ue==W&&J<z)&&(W=ue,z=J,g=L);for(A=g.Jc();A.Ob();){for(T=f(A.Pb(),15),v=new ru(t),ve=T.Jc();ve.Ob();)pe=f(ve.Pb(),9),jo(pe,v);vn(F.c,v)}vG(F),nt.c.length=0,i.Tg()}function K9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt;for(t.Sg(dxt,1),W=new et,He=new et,S=new ne(e.b);S.a<S.c.c.length;)for(v=f(ee(S),25),ue=-1,z=EH(v.a),A=z,L=0,F=A.length;L<F;++L)if(T=A[L],++ue,!!(T.k==(cn(),Yi)&&T3(f(V(T,(ot(),Xi)),100)))){for(!VN(f(V(T,(ot(),Xi)),100))&&Ae(V(ra(T),ip))===Ae((Pp(),Pm))&&w5n(T),$e(T,(Ge(),l_),T),W.c.length=0,He.c.length=0,i=new et,Ee=new ki,jue(Ee,L0(T,(tt(),hn))),Ae(V(ra(T),ip))!==Ae((Pp(),Pm))&&(Ee=Qwt(Ee)),Okt(e,Ee,W,He,i),m=ue,nt=T,l=new ne(W);l.a<l.c.c.length;)a=f(ee(l),9),sw(a,m,v),++ue,$e(a,l_,T),g=f(ct(a.j,0),11),J=f(V(g,oi),11),lt(ut(V(J,sme)))||f(V(a,XR),15).Ec(nt);for(nf(Ee),ve=L0(T,Bn).Jc();ve.Ob();)pe=f(ve.Pb(),11),Wi(Ee,pe,Ee.a,Ee.a.a);for(Ae(V(ra(T),ip))!==Ae(Pm)&&(Ee=Qwt(Ee)),Okt(e,Ee,He,null,i),Le=T,u=new ne(He);u.a<u.c.c.length;)a=f(ee(u),9),sw(a,++ue,v),$e(a,l_,T),g=f(ct(a.j,0),11),J=f(V(g,oi),11),lt(ut(V(J,sme)))||f(V(Le,XR),15).Ec(a);i.c.length==0||$e(T,DPe,i)}t.Tg()}function Y9n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In;if(i.Sg("Spline edge routing",1),t.b.c.length==0){t.f.a=0,i.Tg();return}pe=ke(Oe(V(t,(ot(),NA)))),m=ke(Oe(V(t,jE))),g=ke(Oe(V(t,IA))),ue=f(V(t,gme),346),xt=ue==(JP(),cF),nt=ke(Oe(V(t,DLe))),e.d=t,e.j.c.length=0,e.a.c.length=0,nu(e.k),v=f(ct(t.b,0),25),T=JY(v.a,(iQ(),Hq)),z=f(ct(t.b,t.b.c.length-1),25),A=JY(z.a,Hq),W=new ne(t.b),J=null,In=0;do{for(ve=W.a<W.c.c.length?f(ee(W),25):null,A9n(e,J,ve),sDn(e),Ct=Tpn(Ykn(WX(Jr(new Ft(null,new Bt(e.i,16)),new Cs),new xC))),tn=0,Ee=In,L=!J||T&&J==v,F=!ve||A&&ve==z,Ct>0?(S=0,J&&(S+=m),S+=(Ct-1)*g,ve&&(S+=m),xt&&ve&&(S=h.Math.max(S,k5n(ve,g,pe,nt))),S<pe&&!L&&!F&&(tn=(pe-S)/2,S=pe),Ee+=S):!L&&!F&&(Ee+=pe),ve&&lOe(ve,Ee),He=new ne(e.i);He.a<He.c.c.length;)Le=f(ee(He),132),Le.a.c=In,Le.a.b=Ee-In,Le.F=tn,Le.p=!J;$o(e.a,e.i),In=Ee,ve&&(In+=ve.c.a),J=ve,L=F}while(ve);for(u=new ne(e.j);u.a<u.c.c.length;)a=f(ee(u),18),l=Mkn(e,a),$e(a,(Ge(),AA),l),Kt=j5n(e,a),$e(a,NE,Kt);t.f.a=In,e.d=null,i.Tg()}function X9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;for(e.b=t,e.a=f(V(t,(ot(),jLe)),17).a,e.c=f(V(t,FLe),17).a,e.c==0&&(e.c=Qr),J=new la(t.b,0);J.b<J.d.gc();){for(W=(Ln(J.b<J.d.gc()),f(J.d.Xb(J.c=J.b++),25)),m=new et,T=-1,Ee=-1,ve=new ne(W.a);ve.a<ve.c.c.length;)pe=f(ee(ve),9),g1((aH(),new Tn(Cn(T1(pe).a.Jc(),new B))))>=e.a&&(a=LDn(e,pe),T=h.Math.max(T,a.b),Ee=h.Math.max(Ee,a.d),Qe(m,new Ya(pe,a)));for(xt=new et,S=0;S<T;++S)Ky(xt,0,(Ln(J.b>0),J.a.Xb(J.c=--J.b),Ct=new ru(e.b),nT(J,Ct),Ln(J.b<J.d.gc()),J.d.Xb(J.c=J.b++),Ct));for(g=new ne(m);g.a<g.c.c.length;)if(u=f(ee(g),47),F=f(u.b,562).a,!!F)for(L=new ne(F);L.a<L.c.c.length;)A=f(ee(L),9),D3e(e,A,zee,xt);for(i=new et,v=0;v<Ee;++v)Qe(i,(Kt=new ru(e.b),nT(J,Kt),Kt));for(l=new ne(m);l.a<l.c.c.length;)if(u=f(ee(l),47),nt=f(u.b,562).c,!!nt)for(He=new ne(nt);He.a<He.c.c.length;)Le=f(ee(He),9),D3e(e,Le,Hee,i)}for(ue=new la(t.b,0);ue.b<ue.d.gc();)z=(Ln(ue.b<ue.d.gc()),f(ue.d.Xb(ue.c=ue.b++),25)),z.a.c.length==0&&hd(ue)}function Ckt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;if(W=e.i!=0,ve=!1,ue=null,Ld(e.e)){if(T=t.gc(),T>0){for(L=T<100?null:new Fv(T),S=new Jxe(t),z=S.g,ue=Pe(pr,Pr,30,T,15,1),a=0,Ee=new OS(T),u=0;u<e.i;++u){m=e.g[u],F=m;e:for(pe=0;pe<2;++pe){for(v=T;--v>=0;)if(F!=null?ei(F,z[v]):Ae(F)===Ae(z[v])){ue.length<=a&&(J=ue,ue=Pe(pr,Pr,30,2*ue.length,15,1),au(J,0,ue,0,a)),ue[a++]=u,Qn(Ee,z[v]);break e}if(F=F,Ae(F)===Ae(m))break}}if(S=Ee,z=Ee.g,T=a,a>ue.length&&(J=ue,ue=Pe(pr,Pr,30,a,15,1),au(J,0,ue,0,a)),a>0){for(ve=!0,l=0;l<a;++l)F=z[l],L=Aut(e,f(F,74),L);for(g=a;--g>=0;)wR(e,ue[g]);if(a!=T){for(u=T;--u>=a;)wR(S,u);J=ue,ue=Pe(pr,Pr,30,a,15,1),au(J,0,ue,0,a)}t=S}}}else for(t=GAn(e,t),u=e.i;--u>=0;)t.Gc(e.g[u])&&(wR(e,u),ve=!0);if(ve){if(ue!=null){for(i=t.gc(),A=i==1?tj(e,4,t.Jc().Pb(),null,ue[0],W):tj(e,6,t,ue,ue[0],W),L=i<100?null:new Fv(i),u=t.Jc();u.Ob();)F=u.Pb(),L=q_e(e,f(F,74),L);L?(L.kj(A),L.lj()):Zr(e.e,A)}else{for(L=Bbn(t.gc()),u=t.Jc();u.Ob();)F=u.Pb(),L=q_e(e,f(F,74),L);L&&L.lj()}return!0}else return!1}function J9n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(i=new Mvt(t),i.a||RLn(t),S=PPn(t),v=new CS,J=new V2t,W=new ne(t.a);W.a<W.c.c.length;)for(z=f(ee(W),9),u=new Tn(Cn(ao(z).a.Jc(),new B));Kn(u);)a=f(Dn(u),18),(a.c.i.k==(cn(),bo)||a.d.i.k==bo)&&(T=U8n(e,a,S,J),jt(v,Nle(T.d),T.a));for(g=new et,ve=f(V(i.c,(Ge(),R1)),22).Jc();ve.Ob();){switch(pe=f(ve.Pb(),64),F=J.c[pe.g],L=J.b[pe.g],m=J.a[pe.g],l=null,ue=null,pe.g){case 4:l=new qh(e.d.a,F,S.b.a-e.d.a,L-F),ue=new qh(e.d.a,F,m,L-F),vS(S,new Ke(l.c+l.b,l.d)),vS(S,new Ke(l.c+l.b,l.d+l.a));break;case 2:l=new qh(S.a.a,F,e.c.a-S.a.a,L-F),ue=new qh(e.c.a-m,F,m,L-F),vS(S,new Ke(l.c,l.d)),vS(S,new Ke(l.c,l.d+l.a));break;case 1:l=new qh(F,e.d.b,L-F,S.b.b-e.d.b),ue=new qh(F,e.d.b,L-F,m),vS(S,new Ke(l.c,l.d+l.a)),vS(S,new Ke(l.c+l.b,l.d+l.a));break;case 3:l=new qh(F,S.a.b,L-F,e.c.b-S.a.b),ue=new qh(F,e.c.b-m,L-F,m),vS(S,new Ke(l.c,l.d)),vS(S,new Ke(l.c+l.b,l.d))}l&&(A=new crt,A.d=pe,A.b=l,A.c=ue,A.a=Lce(f(si(v,Nle(pe)),22)),vn(g.c,A))}return $o(i.b,g),i.d=vxn(n8n(S)),i}function Akt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W;if(i.p[t.p]==null){m=!0,i.p[t.p]=0,g=t,W=i.o==(vg(),$E)?qo:Gi;do u=e.b.e[g.p],l=g.c.a.c.length,i.o==$E&&u>0||i.o==Mg&&u<l-1?(v=null,S=null,i.o==Mg?v=f(ct(g.c.a,u+1),9):v=f(ct(g.c.a,u-1),9),S=i.g[v.p],Akt(e,S,i),W=e.e.ug(W,t,g),i.j[t.p]==t&&(i.j[t.p]=i.j[S.p]),i.j[t.p]==i.j[S.p]?(z=A3(e.d,g,v),i.o==Mg?(a=ke(i.p[t.p]),A=ke(i.p[S.p])+ke(i.d[v.p])-v.d.d-z-g.d.a-g.o.b-ke(i.d[g.p]),m?(m=!1,i.p[t.p]=h.Math.min(A,W)):i.p[t.p]=h.Math.min(a,h.Math.min(A,W))):(a=ke(i.p[t.p]),A=ke(i.p[S.p])+ke(i.d[v.p])+v.o.b+v.d.a+z+g.d.d-ke(i.d[g.p]),m?(m=!1,i.p[t.p]=h.Math.max(A,W)):i.p[t.p]=h.Math.max(a,h.Math.max(A,W)))):(z=ke(Oe(V(e.a,(ot(),v_)))),F=xbt(e,i.j[t.p]),T=xbt(e,i.j[S.p]),i.o==Mg?(L=ke(i.p[t.p])+ke(i.d[g.p])+g.o.b+g.d.a+z-(ke(i.p[S.p])+ke(i.d[v.p])-v.d.d),rht(F,T,L)):(L=ke(i.p[t.p])+ke(i.d[g.p])-g.d.d-ke(i.p[S.p])-ke(i.d[v.p])-v.o.b-v.d.a-z,rht(F,T,L)))):W=e.e.ug(W,t,g),g=i.a[g.p];while(g!=t);Zhn(e.e,t)}}function Z9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct;if(i=ke(Oe(V(e.a.j,(ot(),ILe)))),i<-1||!e.a.i||VN(f(V(e.a.o,Xi),100))||lc(e.a.o,(tt(),_n)).gc()<2&&lc(e.a.o,pn).gc()<2)return!0;if(e.a.c.ig())return!1;for(Le=0,Ee=0,ve=new et,v=e.a.e,S=0,T=v.length;S<T;++S){for(m=v[S],L=m,F=0,W=L.length;F<W;++F){if(A=L[F],A.k==(cn(),xu)){vn(ve.c,A);continue}for(a=e.b[A.c.p][A.p],A.k==bo?(a.b=1,f(V(A,(Ge(),oi)),11).j==(tt(),_n)&&(Ee+=a.a)):(Ct=lc(A,(tt(),pn)),Ct.dc()||!Ase(Ct,new wv)?a.c=1:(u=lc(A,_n),(u.dc()||!Ase(u,new $2))&&(Le+=a.a))),g=new Tn(Cn(ao(A).a.Jc(),new B));Kn(g);)l=f(Dn(g),18),Le+=a.c,Ee+=a.b,xt=l.d.i,uxe(e,a,xt);for(ue=Df(ce(se(Gf,1),Jt,20,0,[lc(A,(tt(),hn)),lc(A,Bn)])),nt=new Tn(new r_e(ue.a.length,ue.a));Kn(nt);)He=f(Dn(nt),11),pe=f(V(He,(Ge(),ed)),9),pe&&(Le+=a.c,Ee+=a.b,uxe(e,a,pe))}for(z=new ne(ve);z.a<z.c.c.length;)for(A=f(ee(z),9),a=e.b[A.c.p][A.p],g=new Tn(Cn(ao(A).a.Jc(),new B));Kn(g);)l=f(Dn(g),18),Le+=a.c,Ee+=a.b,xt=l.d.i,uxe(e,a,xt);ve.c.length=0}return t=Le+Ee,J=t==0?Gi:(Le-Ee)/t,J>=i}function Q9n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;for(ve=e.a,Ee=0,Le=ve.length;Ee<Le;++Ee){for(pe=ve[Ee],S=Qr,T=Qr,z=new ne(pe.e);z.a<z.c.c.length;)L=f(ee(z),9),g=L.c?Oc(L.c.a,L,0):-1,g>0?(A=f(ct(L.c.a,g-1),9),xt=A3(e.b,L,A),J=L.n.b-L.d.d-(A.n.b+A.o.b+A.d.a+xt)):J=L.n.b-L.d.d,S=h.Math.min(J,S),g<L.c.a.c.length-1?(A=f(ct(L.c.a,g+1),9),xt=A3(e.b,L,A),ue=A.n.b-A.d.d-(L.n.b+L.o.b+L.d.a+xt)):ue=2*L.n.b,T=h.Math.min(ue,T);for(v=Qr,l=!1,u=f(ct(pe.e,0),9),Kt=new ne(u.j);Kt.a<Kt.c.c.length;)for(Ct=f(ee(Kt),11),W=u.n.b+Ct.n.b+Ct.a.b,a=new ne(Ct.e);a.a<a.c.c.length;)i=f(ee(a),18),He=i.c,t=He.i.n.b+He.n.b+He.a.b-W,h.Math.abs(t)<h.Math.abs(v)&&h.Math.abs(t)<(t<0?S:T)&&(v=t,l=!0);for(m=f(ct(pe.e,pe.e.c.length-1),9),nt=new ne(m.j);nt.a<nt.c.c.length;)for(He=f(ee(nt),11),W=m.n.b+He.n.b+He.a.b,a=new ne(He.g);a.a<a.c.c.length;)i=f(ee(a),18),Ct=i.d,t=Ct.i.n.b+Ct.n.b+Ct.a.b-W,h.Math.abs(t)<h.Math.abs(v)&&h.Math.abs(t)<(t<0?S:T)&&(v=t,l=!0);if(l&&v!=0)for(F=new ne(pe.e);F.a<F.c.c.length;)L=f(ee(F),9),L.n.b+=v}}function e7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In;for(ve=t,pe=new CS,Ee=new CS,T=IS(ve,hpe),a=new Kdt(e,i,pe,Ee),xyt(a.a,a.b,a.c,a.d,T),v=(nt=pe.i,nt||(pe.i=new N3(pe,pe.c))),Ct=v.Jc();Ct.Ob();)for(xt=f(Ct.Pb(),168),u=f(si(pe,xt),22),W=u.Jc();W.Ob();)if(z=W.Pb(),Le=f(F3(e.f,z),168),Le)m=(!xt.e&&(xt.e=new Qt(Ii,xt,10,9)),xt.e),Qn(m,Le);else throw g=p1(ve,Am),L=HCt+z+GCt+g,F=L+YL,ie(new am(F));for(S=(He=Ee.i,He||(Ee.i=new N3(Ee,Ee.c))),tn=S.Jc();tn.Ob();)for(Kt=f(tn.Pb(),168),l=f(si(Ee,Kt),22),ue=l.Jc();ue.Ob();)if(J=ue.Pb(),Le=f(F3(e.f,J),168),Le)A=(!Kt.g&&(Kt.g=new Qt(Ii,Kt,9,10)),Kt.g),Qn(A,Le);else throw g=p1(ve,Am),L=HCt+J+GCt+g,F=L+YL,ie(new am(F));!i.b&&(i.b=new Qt(Yn,i,4,7)),i.b.i!=0&&(!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c.i!=0)&&(!i.b&&(i.b=new Qt(Yn,i,4,7)),i.b.i<=1&&(!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c.i<=1))&&(!i.a&&(i.a=new Be(Ii,i,6,6)),i.a).i==1&&(In=f(be((!i.a&&(i.a=new Be(Ii,i,6,6)),i.a),0),168),!Mle(In)&&!jle(In)&&(zJ(In,f(be((!i.b&&(i.b=new Qt(Yn,i,4,7)),i.b),0),84)),HJ(In,f(be((!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c),0),84))))}function t7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;if(a=new et,u=Qr,l=Qr,g=Qr,i)for(u=e.f.a,W=new ne(t.j);W.a<W.c.c.length;)for(z=f(ee(W),11),v=new ne(z.g);v.a<v.c.c.length;)m=f(ee(v),18),m.a.b!=0&&(T=f(hP(m.a),8),T.a<u&&(l=u-T.a,g=Qr,a.c.length=0,u=T.a),T.a<=u&&(vn(a.c,m),m.a.b>1&&(g=h.Math.min(g,h.Math.abs(f(Xh(m.a,1),8).b-T.b)))));else for(W=new ne(t.j);W.a<W.c.c.length;)for(z=f(ee(W),11),v=new ne(z.e);v.a<v.c.c.length;)m=f(ee(v),18),m.a.b!=0&&(L=f(Hh(m.a),8),L.a>u&&(l=L.a-u,g=Qr,a.c.length=0,u=L.a),L.a>=u&&(vn(a.c,m),m.a.b>1&&(g=h.Math.min(g,h.Math.abs(f(Xh(m.a,m.a.b-2),8).b-L.b)))));if(a.c.length!=0&&l>t.o.a/2&&g>t.o.b/2){for(F=new ou,cc(F,t),No(F,(tt(),hn)),F.n.a=t.o.a/2,ue=new ou,cc(ue,t),No(ue,Bn),ue.n.a=t.o.a/2,ue.n.b=t.o.b,v=new ne(a);v.a<v.c.c.length;)m=f(ee(v),18),i?(S=f(Zse(m.a),8),J=m.a.b==0?wg(m.d):f(hP(m.a),8),J.b>=S.b?Ia(m,ue):Ia(m,F)):(S=f(hvn(m.a),8),J=m.a.b==0?wg(m.c):f(Hh(m.a),8),J.b>=S.b?ua(m,ue):ua(m,F)),A=f(V(m,(ot(),Ls)),77),A&&_T(A,S,!0);t.n.a=u-t.o.a/2}}function n7n(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(m=tr(e.b,0);m.b!=m.d.c;)if(g=f(Zn(m),39),!Mt(g.c,QQ))for(S=q4n(g,e),t==(ko(),Ds)||t==Gs?na(S,new wk):na(S,new OC),v=S.c.length,a=0;a<v;a++)T=(Gt(a,S.c.length),f(S.c[a],65)).c,Mt(T.c,"n11"),!(lt(ut(V(g,(_i(),iMe))))&&!x1t((Gt(a,S.c.length),f(S.c[a],65)),e))&&(u=v==1?.5:(a+1)/(v+1),t==Ds?(l=ke(Oe(V(g,Fp))),L=g.e.b+g.f.b*u,S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(h.Math.min(l,g.e.a-i),L)),S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(g.e.a,L))):t==Gs?(l=ke(Oe(V(g,ap)))+i,L=g.e.b+g.f.b*u,S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(l,L)),S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(g.e.a+g.f.a,L))):t==Wf?(l=ke(Oe(V(g,Fp))),A=g.e.a+g.f.a*u,S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(A,h.Math.min(g.e.b-i,l))),S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(A,g.e.b))):(l=ke(Oe(V(g,ap)))+i,A=g.e.a+g.f.a*u,S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(A,l)),S3((Gt(a,S.c.length),f(S.c[a],65)).a,new Ke(A,g.e.b+g.f.b))))}function PG(e,t,i,a,u,l,g,m,v){var S,T,A,L,F,z,W;switch(F=i,T=new Tg(v),Wl(T,(cn(),bo)),$e(T,(Ge(),FPe),g),$e(T,(ot(),Xi),(oa(),uu)),W=ke(Oe(e.mf(ox))),$e(T,ox,W),A=new ou,cc(A,T),t!=Fg&&t!=Pw||(a>=0?F=mR(m):F=iG(mR(m)),e.of(vD,F)),S=new ha,L=!1,e.nf(m_)?(h_e(S,f(e.mf(m_),8)),L=!0):Rmn(S,g.a/2,g.b/2),F.g){case 4:$e(T,el,(Ff(),Tw)),$e(T,Ste,(lE(),SA)),T.o.b=g.b,W<0&&(T.o.a=-W),No(A,(tt(),_n)),L||(S.a=g.a),S.a-=g.a;break;case 2:$e(T,el,(Ff(),tx)),$e(T,Ste,(lE(),lD)),T.o.b=g.b,W<0&&(T.o.a=-W),No(A,(tt(),pn)),L||(S.a=0);break;case 1:$e(T,OE,(w1(),kA)),T.o.a=g.a,W<0&&(T.o.b=-W),No(A,(tt(),Bn)),L||(S.b=g.b),S.b-=g.b;break;case 3:$e(T,OE,(w1(),WR)),T.o.a=g.a,W<0&&(T.o.b=-W),No(A,(tt(),hn)),L||(S.b=0)}if(h_e(A.n,S),$e(T,m_,S),t==HE||t==Lb||t==uu){if(z=0,t==HE&&e.nf(H0))switch(F.g){case 1:case 2:z=f(e.mf(H0),17).a;break;case 3:case 4:z=-f(e.mf(H0),17).a}else switch(F.g){case 4:case 2:z=l.b,t==Lb&&(z/=u.b);break;case 1:case 3:z=l.a,t==Lb&&(z/=u.a)}$e(T,d_,z)}return $e(T,Tc,F),T}function r7n(){H2e();function e(a){var u=this;this.dispatch=function(l){var g=l.data;switch(g.cmd){case"algorithms":var m=i3e((Ht(),new h3(new kp(qE.b))));a.postMessage({id:g.id,data:m});break;case"categories":var v=i3e((Ht(),new h3(new kp(qE.c))));a.postMessage({id:g.id,data:v});break;case"options":var S=i3e((Ht(),new h3(new kp(qE.d))));a.postMessage({id:g.id,data:S});break;case"register":NMn(g.algorithms),a.postMessage({id:g.id});break;case"layout":TMn(g.graph,g.layoutOptions||{},g.options||{}),a.postMessage({id:g.id,data:g.graph});break}},this.saveDispatch=function(l){try{u.dispatch(l)}catch(g){a.postMessage({id:l.data.id,error:g})}}}function t(a){var u=this;this.dispatcher=new e({postMessage:function(l){u.onmessage({data:l})}}),this.postMessage=function(l){setTimeout(function(){u.dispatcher.saveDispatch({data:l})},0)}}if(typeof document===Ufe&&typeof self!==Ufe){var i=new e(self);self.onmessage=i.saveDispatch}else typeof s!==Ufe&&s.exports&&(Object.defineProperty(c,"__esModule",{value:!0}),s.exports={default:t,Worker:t})}function Ikt(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(T=new Tg(i),Rc(T,t),$e(T,(Ge(),oi),t),T.o.a=t.g,T.o.b=t.f,T.n.a=t.i,T.n.b=t.j,Qe(i.a,T),qr(e.a,t,T),((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i!=0||lt(ut(Fe(t,(ot(),ix)))))&&$e(T,MPe,(mn(),!0)),S=f(V(i,Cu),22),A=f(V(T,(ot(),Xi)),100),A==(oa(),Pw)?$e(T,Xi,Fg):A!=Fg&&S.Ec((us(),hD)),L=0,a=f(V(i,mf),86),v=new Nn((!t.c&&(t.c=new Be(_d,t,9,9)),t.c));v.e!=v.i.gc();)m=f(Mn(v),124),u=so(t),(Ae(Fe(u,ip))!==Ae((Pp(),Pm))||Ae(Fe(u,RE))===Ae((R0(),uD))||Ae(Fe(u,RE))===Ae((R0(),cD))||lt(ut(Fe(u,h_)))||Ae(Fe(u,nx))!==Ae((hE(),XT))||Ae(Fe(u,xw))===Ae((gh(),w_))||Ae(Fe(u,xw))===Ae((gh(),ux))||Ae(Fe(u,LE))===Ae((D0(),SD))||Ae(Fe(u,LE))===Ae((D0(),_D)))&&!lt(ut(Fe(t,Ite)))&&pi(m,hi,rt(L++)),lt(ut(Fe(m,ME)))||d9n(e,m,T,S,a,A);for(g=new Nn((!t.n&&(t.n=new Be(hc,t,1,7)),t.n));g.e!=g.i.gc();)l=f(Mn(g),153),!lt(ut(Fe(l,ME)))&&l.a&&Qe(T.b,cle(l));return lt(ut(V(T,Z$)))&&S.Ec((us(),bte)),lt(ut(V(T,Pte)))&&(S.Ec((us(),vte)),S.Ec(V$),$e(T,Xi,Fg)),T}function ife(e,t,i,a,u,l,g){var m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji;for(W=0,Kt=0,S=new ne(e.b);S.a<S.c.c.length;)v=f(ee(S),162),v.c&&ikt(v.c),W=h.Math.max(W,Al(v)),Kt+=Al(v)*jd(v);for(J=Kt/e.b.c.length,Ct=NIn(e.b,J),Kt+=e.b.c.length*Ct,W=h.Math.max(W,h.Math.sqrt(Kt*g))+i.b,$r=i.b,ji=i.d,F=0,A=i.b+i.c,xt=new ki,Mr(xt,rt(0)),He=new ki,T=new la(e.b,0),z=null,m=new et;T.b<T.d.gc();)v=(Ln(T.b<T.d.gc()),f(T.d.Xb(T.c=T.b++),162)),In=Al(v),L=jd(v),$r+In>W&&(l&&(Ma(He,F),Ma(xt,rt(T.b-1)),Qe(e.d,z),m.c.length=0),$r=i.b,ji+=F+t,F=0,A=h.Math.max(A,i.b+i.c+In)),vn(m.c,v),Ovt(v,$r,ji),A=h.Math.max(A,$r+In+i.c),F=h.Math.max(F,L),$r+=In+t,z=v;if($o(e.a,m),Qe(e.d,f(ct(m,m.c.length-1),162)),A=h.Math.max(A,a),tn=ji+F+i.a,tn<u&&(F+=u-tn,tn=u),l)for($r=i.b,T=new la(e.b,0),Ma(xt,rt(e.b.c.length)),nt=tr(xt,0),pe=f(Zn(nt),17).a,Ma(He,F),Le=tr(He,0),Ee=0;T.b<T.d.gc();)T.b==pe&&($r=i.b,Ee=ke(Oe(Zn(Le))),pe=f(Zn(nt),17).a),v=(Ln(T.b<T.d.gc()),f(T.d.Xb(T.c=T.b++),162)),uvt(v,Ee),T.b==pe&&(ue=A-$r-i.c,ve=Al(v),cvt(v,ue),R0t(v,(ue-ve)/2,0)),$r+=Al(v)+t;return new Ke(A,tn)}function i7n(e){e.N||(e.N=!0,e.b=yc(e,0),Ki(e.b,0),Ki(e.b,1),Ki(e.b,2),e.bb=yc(e,1),Ki(e.bb,0),Ki(e.bb,1),e.fb=yc(e,2),Ki(e.fb,3),Ki(e.fb,4),Ci(e.fb,5),e.qb=yc(e,3),Ki(e.qb,0),Ci(e.qb,1),Ci(e.qb,2),Ki(e.qb,3),Ki(e.qb,4),Ci(e.qb,5),Ki(e.qb,6),e.a=Br(e,4),e.c=Br(e,5),e.d=Br(e,6),e.e=Br(e,7),e.f=Br(e,8),e.g=Br(e,9),e.i=Br(e,10),e.j=Br(e,11),e.k=Br(e,12),e.n=Br(e,13),e.o=Br(e,14),e.p=Br(e,15),e.q=Br(e,16),e.s=Br(e,17),e.r=Br(e,18),e.t=Br(e,19),e.u=Br(e,20),e.v=Br(e,21),e.w=Br(e,22),e.B=Br(e,23),e.A=Br(e,24),e.C=Br(e,25),e.D=Br(e,26),e.F=Br(e,27),e.G=Br(e,28),e.H=Br(e,29),e.J=Br(e,30),e.I=Br(e,31),e.K=Br(e,32),e.M=Br(e,33),e.L=Br(e,34),e.P=Br(e,35),e.Q=Br(e,36),e.R=Br(e,37),e.S=Br(e,38),e.T=Br(e,39),e.U=Br(e,40),e.V=Br(e,41),e.X=Br(e,42),e.W=Br(e,43),e.Y=Br(e,44),e.Z=Br(e,45),e.$=Br(e,46),e._=Br(e,47),e.ab=Br(e,48),e.cb=Br(e,49),e.db=Br(e,50),e.eb=Br(e,51),e.gb=Br(e,52),e.hb=Br(e,53),e.ib=Br(e,54),e.jb=Br(e,55),e.kb=Br(e,56),e.lb=Br(e,57),e.mb=Br(e,58),e.nb=Br(e,59),e.ob=Br(e,60),e.pb=Br(e,61))}function o7n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;if(pe=0,t.f.a==0)for(J=new ne(e);J.a<J.c.c.length;)z=f(ee(J),9),pe=h.Math.max(pe,z.n.a+z.o.a+z.d.c);else pe=t.f.a-t.c.a;for(pe-=t.c.a,W=new ne(e);W.a<W.c.c.length;){switch(z=f(ee(W),9),r7(z.n,pe-z.o.a),ske(z.f),owt(z),(z.q?z.q:(Ht(),Ht(),Cb))._b((ot(),sx))&&r7(f(V(z,sx),8),pe-z.o.a),f(V(z,Nm),254).g){case 1:$e(z,Nm,(vb(),eV));break;case 2:$e(z,Nm,(vb(),Qq))}for(ue=z.o,Ee=new ne(z.j);Ee.a<Ee.c.c.length;){for(ve=f(ee(Ee),11),r7(ve.n,ue.a-ve.o.a),r7(ve.a,ve.o.a),No(ve,jbt(ve.j)),g=f(V(ve,H0),17),g&&$e(ve,H0,rt(-g.a)),l=new ne(ve.g);l.a<l.c.c.length;){for(u=f(ee(l),18),a=tr(u.a,0);a.b!=a.d.c;)i=f(Zn(a),8),i.a=pe-i.a;if(S=f(V(u,Ls),77),S)for(v=tr(S,0);v.b!=v.d.c;)m=f(Zn(v),8),m.a=pe-m.a;for(L=new ne(u.b);L.a<L.c.c.length;)T=f(ee(L),72),r7(T.n,pe-T.o.a)}for(F=new ne(ve.f);F.a<F.c.c.length;)T=f(ee(F),72),r7(T.n,ve.o.a-T.o.a)}for(z.k==(cn(),bo)&&($e(z,(Ge(),Tc),jbt(f(V(z,Tc),64))),_Nn(z)),A=new ne(z.b);A.a<A.c.c.length;)T=f(ee(A),72),owt(T),r7(T.n,ue.a-T.o.a)}}function a7n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;if(pe=0,t.f.b==0)for(J=new ne(e);J.a<J.c.c.length;)z=f(ee(J),9),pe=h.Math.max(pe,z.n.b+z.o.b+z.d.a);else pe=t.f.b-t.c.b;for(pe-=t.c.b,W=new ne(e);W.a<W.c.c.length;){switch(z=f(ee(W),9),n7(z.n,pe-z.o.b),cke(z.f),awt(z),(z.q?z.q:(Ht(),Ht(),Cb))._b((ot(),sx))&&n7(f(V(z,sx),8),pe-z.o.b),f(V(z,Nm),254).g){case 3:$e(z,Nm,(vb(),bne));break;case 4:$e(z,Nm,(vb(),wne))}for(ue=z.o,Ee=new ne(z.j);Ee.a<Ee.c.c.length;){for(ve=f(ee(Ee),11),n7(ve.n,ue.b-ve.o.b),n7(ve.a,ve.o.b),No(ve,$bt(ve.j)),g=f(V(ve,H0),17),g&&$e(ve,H0,rt(-g.a)),l=new ne(ve.g);l.a<l.c.c.length;){for(u=f(ee(l),18),a=tr(u.a,0);a.b!=a.d.c;)i=f(Zn(a),8),i.b=pe-i.b;if(S=f(V(u,Ls),77),S)for(v=tr(S,0);v.b!=v.d.c;)m=f(Zn(v),8),m.b=pe-m.b;for(L=new ne(u.b);L.a<L.c.c.length;)T=f(ee(L),72),n7(T.n,pe-T.o.b)}for(F=new ne(ve.f);F.a<F.c.c.length;)T=f(ee(F),72),n7(T.n,ve.o.b-T.o.b)}for(z.k==(cn(),bo)&&($e(z,(Ge(),Tc),$bt(f(V(z,Tc),64))),Qxn(z)),A=new ne(z.b);A.a<A.c.c.length;)T=f(ee(A),72),awt(T),n7(T.n,ue.b-T.o.b)}}function s7n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r;for(pe=new la(e.b,0),T=t.Jc(),z=0,S=f(T.Pb(),17).a,Le=0,i=new Mo,nt=new ub;pe.b<pe.d.gc();){for(ue=(Ln(pe.b<pe.d.gc()),f(pe.d.Xb(pe.c=pe.b++),25)),Ee=new ne(ue.a);Ee.a<Ee.c.c.length;){for(ve=f(ee(Ee),9),F=new Tn(Cn(ao(ve).a.Jc(),new B));Kn(F);)A=f(Dn(F),18),nt.a.yc(A,nt);for(L=new Tn(Cn(Fo(ve).a.Jc(),new B));Kn(L);)A=f(Dn(L),18),nt.a.Ac(A)!=null}if(z+1==S){for(u=new ru(e),nT(pe,u),l=new ru(e),nT(pe,l),Ct=nt.a.ec().Jc();Ct.Ob();)xt=f(Ct.Pb(),18),i.a._b(xt)||(++Le,i.a.yc(xt,i)),g=new Tg(e),$e(g,(ot(),Xi),(oa(),MD)),jo(g,u),Wl(g,(cn(),Dg)),W=new ou,cc(W,g),No(W,(tt(),pn)),Kt=new ou,cc(Kt,g),No(Kt,_n),a=new Tg(e),$e(a,Xi,MD),jo(a,l),Wl(a,Dg),J=new ou,cc(J,a),No(J,pn),tn=new ou,cc(tn,a),No(tn,_n),He=new xS,Ia(He,xt.c),ua(He,W),$e(He,(Ge(),hi),f(V(xt,hi),17)),$r=new xS,Ia($r,Kt),ua($r,J),$e($r,hi,f(V(xt,hi),17)),Ia(xt,tn),m=new PTe(g,a,He,$r,xt),$e(g,u_,m),$e(a,u_,m),In=He.c.i,In.k==Dg&&(v=f(V(In,u_),314),v.d=m,m.g=v);if(T.Ob())S=f(T.Pb(),17).a;else break}++z}return rt(Le)}function c7n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(W=new et,L=new ne(e.d.b);L.a<L.c.c.length;)for(A=f(ee(L),25),z=new ne(A.a);z.a<z.c.c.length;){for(F=f(ee(z),9),u=f(bn(e.f,F),60),v=new Tn(Cn(ao(F).a.Jc(),new B));Kn(v);)if(g=f(Dn(v),18),a=tr(g.a,0),S=!0,T=null,a.b!=a.d.c){for(t=f(Zn(a),8),i=null,g.c.j==(tt(),hn)&&(J=new Xj(t,new Ke(t.a,u.d.d),u,g),J.f.a=!0,J.a=g.c,vn(W.c,J)),g.c.j==Bn&&(J=new Xj(t,new Ke(t.a,u.d.d+u.d.a),u,g),J.f.d=!0,J.a=g.c,vn(W.c,J));a.b!=a.d.c;)i=f(Zn(a),8),Sxe(t.b,i.b)||(T=new Xj(t,i,null,g),vn(W.c,T),S&&(S=!1,i.b<u.d.d?T.f.a=!0:i.b>u.d.d+u.d.a?T.f.d=!0:(T.f.d=!0,T.f.a=!0))),a.b!=a.d.c&&(t=i);T&&(l=f(bn(e.f,g.d.i),60),t.b<l.d.d?T.f.a=!0:t.b>l.d.d+l.d.a?T.f.d=!0:(T.f.d=!0,T.f.a=!0))}for(m=new Tn(Cn(Fo(F).a.Jc(),new B));Kn(m);)g=f(Dn(m),18),g.a.b!=0&&(t=f(Hh(g.a),8),g.d.j==(tt(),hn)&&(J=new Xj(t,new Ke(t.a,u.d.d),u,g),J.f.a=!0,J.a=g.d,vn(W.c,J)),g.d.j==Bn&&(J=new Xj(t,new Ke(t.a,u.d.d+u.d.a),u,g),J.f.d=!0,J.a=g.d,vn(W.c,J)))}return W}function u7n(e,t,i){var a,u,l,g,m,v,S,T,A,L;for(v=new et,A=t.length,g=lCe(i),S=0;S<A;++S){switch(T=i_e(t,gl(61),S),a=Jxn(g,(ga(S,T,t.length),t.substr(S,T-S))),u=que(a),l=u.gk().si(),_a(t,++T)){case 39:{m=fP(t,39,++T),Qe(v,new KY(a,Nce((ga(T,m,t.length),t.substr(T,m-T)),l,u))),S=m+1;break}case 34:{m=fP(t,34,++T),Qe(v,new KY(a,Nce((ga(T,m,t.length),t.substr(T,m-T)),l,u))),S=m+1;break}case 91:{L=new et,Qe(v,new KY(a,L));e:for(;;){switch(_a(t,++T)){case 39:{m=fP(t,39,++T),Qe(L,Nce((ga(T,m,t.length),t.substr(T,m-T)),l,u)),T=m+1;break}case 34:{m=fP(t,34,++T),Qe(L,Nce((ga(T,m,t.length),t.substr(T,m-T)),l,u)),T=m+1;break}case 110:{if(++T,t.indexOf("ull",T)==T)L.c.push(null);else throw ie(new oc(LCt));T+=3;break}}if(T<A)switch(gn(T,t.length),t.charCodeAt(T)){case 44:break;case 93:break e;default:throw ie(new oc("Expecting , or ]"))}else break}S=T+1;break}case 110:{if(++T,t.indexOf("ull",T)==T)Qe(v,new KY(a,null));else throw ie(new oc(LCt));S=T+3;break}}if(S<A){if(gn(S,t.length),t.charCodeAt(S)!=44)throw ie(new oc("Expecting ,"))}else break}return rPn(e,v,i)}function l7n(e){var t,i,a,u,l;switch(t=e.c,l=null,t){case 6:return e.Bm();case 13:return e.Cm();case 23:return e.tm();case 22:return e.ym();case 18:return e.vm();case 8:Yr(e),l=(Xr(),aje);break;case 9:return e.bm(!0);case 19:return e.cm();case 10:switch(e.a){case 100:case 68:case 119:case 87:case 115:case 83:return l=e.am(e.a),Yr(e),l;case 101:case 102:case 110:case 114:case 116:case 117:case 118:case 120:i=e._l(),i<Xa?l=(Xr(),Xr(),new fb(0,i)):l=Ult(qCe(i));break;case 99:return e.lm();case 67:return e.gm();case 105:return e.om();case 73:return e.hm();case 103:return e.mm();case 88:return e.im();case 49:case 50:case 51:case 52:case 53:case 54:case 55:case 56:case 57:return e.dm();case 80:case 112:if(l=qAe(e,e.a),!l)throw ie(new vr(yr((lr(),vpe))));break;default:l=ilt(e.a)}Yr(e);break;case 0:if(e.a==93||e.a==123||e.a==125)throw ie(new vr(yr((lr(),G4e))));l=ilt(e.a),a=e.a,Yr(e),(a&64512)==GG&&e.c==0&&(e.a&64512)==56320&&(u=Pe(Xf,_m,30,2,15,1),u[0]=a&To,u[1]=e.a&To,l=hce(Ult(bm(u,0,u.length)),0),Yr(e));break;default:throw ie(new vr(yr((lr(),G4e))))}return l}function d7n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn;for(Ct=new ki,He=new ki,J=-1,v=new ne(e);v.a<v.c.c.length;){for(g=f(ee(v),132),g.s=J--,T=0,ve=0,l=new ne(g.t);l.a<l.c.c.length;)a=f(ee(l),270),ve+=a.c;for(u=new ne(g.i);u.a<u.c.c.length;)a=f(ee(u),270),T+=a.c;g.n=T,g.u=ve,ve==0?Wi(He,g,He.c.b,He.c):T==0&&Wi(Ct,g,Ct.c.b,Ct.c)}for(tn=fJ(e),A=e.c.length,W=A+1,ue=A-1,F=new et;tn.a.gc()!=0;){for(;He.b!=0;)Le=(Ln(He.b!=0),f(lh(He,He.a.a),132)),tn.a.Ac(Le)!=null,Le.s=ue--,fIe(Le,Ct,He);for(;Ct.b!=0;)nt=(Ln(Ct.b!=0),f(lh(Ct,Ct.a.a),132)),tn.a.Ac(nt)!=null,nt.s=W++,fIe(nt,Ct,He);for(z=Ra,S=tn.a.ec().Jc();S.Ob();)g=f(S.Pb(),132),pe=g.u-g.n,pe>=z&&(pe>z&&(F.c.length=0,z=pe),vn(F.c,g));F.c.length!=0&&(L=f(ct(F,IZ(t,F.c.length)),132),tn.a.Ac(L)!=null,L.s=W++,fIe(L,Ct,He),F.c.length=0)}for(Ee=e.c.length+1,m=new ne(e);m.a<m.c.c.length;)g=f(ee(m),132),g.s<A&&(g.s+=Ee);for(xt=new ne(e);xt.a<xt.c.c.length;)for(nt=f(ee(xt),132),i=new la(nt.t,0);i.b<i.d.gc();)a=(Ln(i.b<i.d.gc()),f(i.d.Xb(i.c=i.b++),270)),Kt=a.b,nt.s>Kt.s&&(hd(i),hl(Kt.i,a),a.c>0&&(a.a=Kt,Qe(Kt.t,a),a.b=nt,Qe(nt.i,a)))}function Okt(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn;for(W=new Du(t.b),Ee=new Du(t.b),L=new Du(t.b),xt=new Du(t.b),J=new Du(t.b),nt=tr(t,0);nt.b!=nt.d.c;)for(Le=f(Zn(nt),11),m=new ne(Le.g);m.a<m.c.c.length;)if(l=f(ee(m),18),l.c.i==l.d.i){if(Le.j==l.d.j){vn(xt.c,l);continue}else if(Le.j==(tt(),hn)&&l.d.j==Bn){vn(J.c,l);continue}}for(v=new ne(J);v.a<v.c.c.length;)l=f(ee(v),18),HLn(e,l,i,a,(tt(),_n));for(g=new ne(xt);g.a<g.c.c.length;)l=f(ee(g),18),Ct=new Tg(e),Wl(Ct,(cn(),xu)),$e(Ct,(ot(),Xi),(oa(),uu)),$e(Ct,(Ge(),oi),l),Kt=new ou,$e(Kt,oi,l.d),No(Kt,(tt(),pn)),cc(Kt,Ct),tn=new ou,$e(tn,oi,l.c),No(tn,_n),cc(tn,Ct),$e(l.c,ed,Ct),$e(l.d,ed,Ct),Ia(l,null),ua(l,null),vn(i.c,Ct),$e(Ct,Ete,rt(2));for(He=tr(t,0);He.b!=He.d.c;)Le=f(Zn(He),11),S=Le.e.c.length>0,ue=Le.g.c.length>0,S&&ue?vn(L.c,Le):S?vn(W.c,Le):ue&&vn(Ee.c,Le);for(z=new ne(W);z.a<z.c.c.length;)F=f(ee(z),11),Qe(u,iOe(e,F,null,i));for(ve=new ne(Ee);ve.a<ve.c.c.length;)pe=f(ee(ve),11),Qe(u,iOe(e,null,pe,i));for(A=new ne(L);A.a<A.c.c.length;)T=f(ee(A),11),Qe(u,iOe(e,T,T,i))}function SOe(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(L=Gi,F=Gi,T=0,A=0,v=new et,m=new Nn((!e.b&&(e.b=new Be(vo,e,12,3)),e.b));m.e!=m.i.gc();)l=f(Mn(m),85),v=Df(ce(se(Gf,1),Jt,20,0,[v,(!l.n&&(l.n=new Be(hc,l,1,7)),l.n)]));for(ve=pb(Df(ce(se(Gf,1),Jt,20,0,[(!e.n&&(e.n=new Be(hc,e,1,7)),e.n),(!e.a&&(e.a=new Be(Ur,e,10,11)),e.a),v])));Kn(ve);)pe=f(Dn(ve),273),S=f(pe.mf((Rr(),OD)),139),L>pe.lh()-S.b&&(L=pe.lh()-S.b),F>pe.mh()-S.d&&(F=pe.mh()-S.d),T<pe.lh()+pe.kh()+S.c&&(T=pe.lh()+pe.kh()+S.c),A<pe.mh()+pe.jh()+S.a&&(A=pe.mh()+pe.jh()+S.a);for(g=new Nn((!e.b&&(e.b=new Be(vo,e,12,3)),e.b));g.e!=g.i.gc();)for(l=f(Mn(g),85),ue=new Nn((!l.a&&(l.a=new Be(Ii,l,6,6)),l.a));ue.e!=ue.i.gc();)for(J=f(Mn(ue),168),z=J.j,a=J.b,W=J.k,u=J.c,L=h.Math.min(L,z),L=h.Math.min(L,a),T=h.Math.max(T,z),T=h.Math.max(T,a),F=h.Math.min(F,W),F=h.Math.min(F,u),A=h.Math.max(A,W),A=h.Math.max(A,u),i=new Nn((!J.a&&(J.a=new So(yf,J,5)),J.a));i.e!=i.i.gc();)t=f(Mn(i),369),L=h.Math.min(L,t.a),T=h.Math.max(T,t.a),F=h.Math.min(F,t.b),A=h.Math.max(A,t.b);pi(e,(Rr(),BA),T-L),pi(e,FA,A-F)}function Nkt(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji;for(F=f(V(e,(ew(),qR)),27),ve=Qr,Ee=Qr,ue=Ra,pe=Ra,He=new ne(e.e);He.a<He.c.c.length;)Le=f(ee(He),151),tn=Le.d,In=Le.e,ve=h.Math.min(ve,tn.a-In.a/2),Ee=h.Math.min(Ee,tn.b-In.b/2),ue=h.Math.max(ue,tn.a+In.a/2),pe=h.Math.max(pe,tn.b+In.b/2);for(i=new ne(e.b);i.a<i.c.c.length;)t=f(ee(i),249),tn=t.d,In=t.e,ve=h.Math.min(ve,tn.a-In.a/2),Ee=h.Math.min(Ee,tn.b-In.b/2),ue=h.Math.max(ue,tn.a+In.a/2),pe=h.Math.max(pe,tn.b+In.b/2);for(Kt=f(Fe(F,(Qh(),oOt)),107),Ct=new Ke(Kt.b-ve,Kt.d-Ee),S=new ne(e.e);S.a<S.c.c.length;)v=f(ee(S),151),xt=V(v,qR),me(xt,206)&&(W=f(xt,27),nt=ii(new ja(v.d),Ct),Of(W,nt.a-W.g/2,nt.b-W.f/2));for(l=new ne(e.c);l.a<l.c.c.length;)u=f(ee(l),288),A=f(V(u,qR),85),L=rQ(A),$r=new ja(Ixe(u)),ii($r,Ct),Yz(L,$r.a,$r.b),Mu(u.a,new zot(Ct,L)),a=new ja(Oxe(u)),ii(a,Ct),Kz(L,a.a,a.b);for(m=new ne(e.d);m.a<m.c.c.length;)g=f(ee(m),442),z=f(V(g,qR),153),J=ii(new ja(g.d),Ct),Of(z,J.a,J.b);ji=ue-ve+(Kt.b+Kt.c),T=pe-Ee+(Kt.d+Kt.a),lt(ut(Fe(F,(Rr(),hx))))||WS(F,ji,T,!1,!0),pi(F,BA,ji-(Kt.b+Kt.c)),pi(F,FA,T-(Kt.d+Kt.a))}function f7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve;for(i.Sg("Depth first model order layering",1),e.d=t,J=new et,W=new ne(e.d.a);W.a<W.c.c.length;)F=f(ee(W),9),F.k==(cn(),Yi)&&vn(J.c,F);for(Ht(),na(J,new hk),g=!0,e.b=new ru(e.d),e.a=null,Qe(e.d.b,e.b),e.b.p=0,e.c=0,e.f=new ki,z=new ne(J);z.a<z.c.c.length;)if(F=f(ee(z),9),g)jo(F,e.b),g=!1;else if(NDn(e,F))if(L=e.c,L=Lvt(L,F),a=L+2,T=L-e.c,e.f.b==0)CIe(e,a,F);else if(T>0){for(ve=tr(e.f,0);ve.b!=ve.d.c;)pe=f(Zn(ve),9),pe.p+=L-e.e;AAe(e),nf(e.f),CIe(e,a,F)}else{for(Mr(e.f,F),F.p=a,e.e=h.Math.max(e.e,a),l=new Tn(Cn(Fo(F).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),!u.c.i.c&&u.c.i.k==(cn(),Vc)&&(Mr(e.f,u.c.i),u.c.i.p=a-1);e.c=a}else AAe(e),nf(e.f),a=0,Kn(new Tn(Cn(Fo(F).a.Jc(),new B)))?(L=0,L=Lvt(L,F),a=L+2,CIe(e,a,F)):(Mr(e.f,F),F.p=0,e.e=h.Math.max(e.e,0),e.b=f(ct(e.d.b,0),25),e.c=0);for(e.f.b==0||AAe(e),e.d.a.c.length=0,ue=new et,S=new ne(e.d.b);S.a<S.c.c.length;)m=f(ee(S),25),m.a.c.length==0&&vn(ue.c,m);for(zCe(e.d.b,ue),A=0,v=new ne(e.d.b);v.a<v.c.c.length;)m=f(ee(v),25),m.p=A,++A;i.Tg()}function h7n(e,t,i){var a,u,l,g,m,v,S,T,A;if(i.Sg("Network simplex node placement",1),e.e=t,e.n=f(V(t,(Ge(),CA)),313),CMn(e),vOn(e),Zi(Zs(new Ft(null,new Bt(e.e.b,16)),new xO),new btt(e)),Zi(Jr(Zs(Jr(Zs(new Ft(null,new Bt(e.e.b,16)),new Qb),new Dh),new N8),new j6),new mtt(e)),lt(ut(V(e.e,(ot(),tF))))&&(g=i.bh(1),g.Sg("Straight Edges Pre-Processing",1),K8n(e),g.Tg()),aAn(e.f),l=f(V(t,iF),17).a*e.f.a.c.length,jde(j2e(M2e(Gse(e.f),l),!1),i.bh(1)),e.d.a.gc()!=0){for(g=i.bh(1),g.Sg("Flexible Where Space Processing",1),m=f(Dd(XN(Lu(new Ft(null,new Bt(e.f.a,16)),new M6),new x8)),17).a,v=f(Dd(oT(Lu(new Ft(null,new Bt(e.f.a,16)),new pk),new yp)),17).a,S=v-m,T=mS(new qy,e.f),A=mS(new qy,e.f),Zh(Uh(Bh(Fh(zh(new ah,2e4),S),T),A)),Zi(Jr(Jr(pce(e.i),new I8),new O8),new zdt(m,T,S,A)),u=e.d.a.ec().Jc();u.Ob();)a=f(u.Pb(),216),a.g=1;jde(j2e(M2e(Gse(e.f),l),!1),g.bh(1)),g.Tg()}lt(ut(V(t,tF)))&&(g=i.bh(1),g.Sg("Straight Edges Post-Processing",1),cRn(e),g.Tg()),N8n(e),e.e=null,e.f=null,e.i=null,e.c=null,nu(e.k),e.j=null,e.a=null,e.o=null,e.d.a.$b(),i.Tg()}function p7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji,id,Ms;if(Kt=null,In=t,tn=nbt(e,Xht(i),In),FP(tn,p1(In,Am)),$r=f(F3(e.n,SR(h1(In,ape))),27),L=h1(In,"sourcePort"),a=null,L&&(a=SR(L)),ji=f(F3(e.p,a),124),!$r)throw m=oL(In),z="An edge must have a source node (edge id: '"+m,W=z+YL,ie(new am(W));if(ji&&!dm(bg(ji),$r))throw v=p1(In,Am),J="The source port of an edge must be a port of the edge's source node (edge id: '"+v,ue=J+YL,ie(new am(ue));if(xt=(!tn.b&&(tn.b=new Qt(Yn,tn,4,7)),tn.b),l=null,ji?l=ji:l=$r,Qn(xt,l),id=f(F3(e.n,SR(h1(In,z4e))),27),F=h1(In,"targetPort"),u=null,F&&(u=SR(F)),Ms=f(F3(e.p,u),124),!id)throw A=oL(In),pe="An edge must have a target node (edge id: '"+A,ve=pe+YL,ie(new am(ve));if(Ms&&!dm(bg(Ms),id))throw S=p1(In,Am),Ee="The target port of an edge must be a port of the edge's target node (edge id: '"+S,Le=Ee+YL,ie(new am(Le));if(Ct=(!tn.c&&(tn.c=new Qt(Yn,tn,5,8)),tn.c),g=null,Ms?g=Ms:g=id,Qn(Ct,g),(!tn.b&&(tn.b=new Qt(Yn,tn,4,7)),tn.b).i==0||(!tn.c&&(tn.c=new Qt(Yn,tn,5,8)),tn.c).i==0)throw T=p1(In,Am),He=zCt+T,nt=He+YL,ie(new am(nt));return UZ(In,tn),t6n(In,tn),Kt=Zue(e,In,tn),Kt}function Rkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;return A=uDn(dl(e,(tt(),Bg)),t),z=X3(dl(e,dp),t),Ee=X3(dl(e,vf),t),xt=LZ(dl(e,_h),t),L=LZ(dl(e,Kf),t),pe=X3(dl(e,fp),t),W=X3(dl(e,nd),t),He=X3(dl(e,wf),t),Le=X3(dl(e,Yf),t),Ct=LZ(dl(e,tl),t),ue=X3(dl(e,wl),t),ve=X3(dl(e,rd),t),nt=X3(dl(e,vl),t),Kt=LZ(dl(e,yl),t),F=LZ(dl(e,jl),t),J=X3(dl(e,Au),t),i=$3(ce(se(sa,1),_s,30,15,[pe.a,xt.a,He.a,Kt.a])),a=$3(ce(se(sa,1),_s,30,15,[z.a,A.a,Ee.a,J.a])),u=ue.a,l=$3(ce(se(sa,1),_s,30,15,[W.a,L.a,Le.a,F.a])),S=$3(ce(se(sa,1),_s,30,15,[pe.b,z.b,W.b,ve.b])),v=$3(ce(se(sa,1),_s,30,15,[xt.b,A.b,L.b,J.b])),T=Ct.b,m=$3(ce(se(sa,1),_s,30,15,[He.b,Ee.b,Le.b,nt.b])),k0(dl(e,Bg),i+u,S+T),k0(dl(e,Au),i+u,S+T),k0(dl(e,dp),i+u,0),k0(dl(e,vf),i+u,S+T+v),k0(dl(e,_h),0,S+T),k0(dl(e,Kf),i+u+a,S+T),k0(dl(e,nd),i+u+a,0),k0(dl(e,wf),0,S+T+v),k0(dl(e,Yf),i+u+a,S+T+v),k0(dl(e,tl),0,S),k0(dl(e,wl),i,0),k0(dl(e,vl),0,S+T+v),k0(dl(e,jl),i+u+a,0),g=new ha,g.a=$3(ce(se(sa,1),_s,30,15,[i+a+u+l,Ct.a,ve.a,nt.a])),g.b=$3(ce(se(sa,1),_s,30,15,[S+v+T+m,ue.b,Kt.b,F.b])),g}function Pkt(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;for(pe=new Ke(Gi,Gi),t=new Ke(qo,qo),xt=new ne(e);xt.a<xt.c.c.length;)nt=f(ee(xt),8),pe.a=h.Math.min(pe.a,nt.a),pe.b=h.Math.min(pe.b,nt.b),t.a=h.Math.max(t.a,nt.a),t.b=h.Math.max(t.b,nt.b);for(L=new Ke(t.a-pe.a,t.b-pe.b),S=new Ke(pe.a-50,pe.b-L.a-50),T=new Ke(pe.a-50,t.b+L.a+50),A=new Ke(t.a+L.b/2+50,pe.b+L.b/2),F=new nOe(S,T,A),He=new Mo,l=new et,i=new et,He.a.yc(F,He),Kt=new ne(e);Kt.a<Kt.c.c.length;){for(Ct=f(ee(Kt),8),l.c.length=0,Le=He.a.ec().Jc();Le.Ob();)ve=f(Le.Pb(),318),a=ve.d,Qv(a,ve.a),jS(Qv(ve.d,Ct),Qv(ve.d,ve.a))<0&&vn(l.c,ve);for(i.c.length=0,Ee=new ne(l);Ee.a<Ee.c.c.length;)for(ve=f(ee(Ee),318),J=new ne(ve.e);J.a<J.c.c.length;){for(z=f(ee(J),176),g=!0,v=new ne(l);v.a<v.c.c.length;)m=f(ee(v),318),m!=ve&&(Pu(z,ct(m.e,0))||Pu(z,ct(m.e,1))||Pu(z,ct(m.e,2)))&&(g=!1);g&&vn(i.c,z)}for(oyt(He,l),Oa(He,new ld),W=new ne(i);W.a<W.c.c.length;)z=f(ee(W),176),Go(He,new nOe(Ct,z.a,z.b))}for(ue=new Mo,Oa(He,new Ji(ue)),u=ue.a.ec().Jc();u.Ob();)z=f(u.Pb(),176),(CJ(F,z.a)||CJ(F,z.b))&&u.Qb();return Oa(ue,new Yc),ue}function Gc(){Gc=Z,_it(),HDt=Qo.a,f(be(je(Qo.a),0),19),UDt=Qo.f,f(be(je(Qo.f),0),19),f(be(je(Qo.f),1),35),zDt=Qo.n,f(be(je(Qo.n),0),35),f(be(je(Qo.n),1),35),f(be(je(Qo.n),2),35),f(be(je(Qo.n),3),35),_7e=Qo.g,f(be(je(Qo.g),0),19),f(be(je(Qo.g),1),35),BDt=Qo.c,f(be(je(Qo.c),0),19),f(be(je(Qo.c),1),19),k7e=Qo.i,f(be(je(Qo.i),0),19),f(be(je(Qo.i),1),19),f(be(je(Qo.i),2),19),f(be(je(Qo.i),3),19),f(be(je(Qo.i),4),35),T7e=Qo.j,f(be(je(Qo.j),0),19),S7e=Qo.d,f(be(je(Qo.d),0),19),f(be(je(Qo.d),1),19),f(be(je(Qo.d),2),19),f(be(je(Qo.d),3),19),f(be(je(Qo.d),4),35),f(be(je(Qo.d),5),35),f(be(je(Qo.d),6),35),f(be(je(Qo.d),7),35),FDt=Qo.b,f(be(je(Qo.b),0),35),f(be(je(Qo.b),1),35),Rne=Qo.e,f(be(je(Qo.e),0),35),f(be(je(Qo.e),1),35),f(be(je(Qo.e),2),35),f(be(je(Qo.e),3),35),f(be(je(Qo.e),4),19),f(be(je(Qo.e),5),19),f(be(je(Qo.e),6),19),f(be(je(Qo.e),7),19),f(be(je(Qo.e),8),19),f(be(je(Qo.e),9),19),f(be(je(Qo.e),10),35),Db=Qo.k,f(be(je(Qo.k),0),35),f(be(je(Qo.k),1),35)}function _Oe(e){var t,i,a,u,l;switch(t=e.c,t){case 11:return e.sm();case 12:return e.um();case 14:return e.wm();case 15:return e.zm();case 16:return e.xm();case 17:return e.Am();case 21:return Yr(e),Xr(),Xr(),XF;case 10:switch(e.a){case 65:return e.em();case 90:return e.jm();case 122:return e.qm();case 98:return e.km();case 66:return e.fm();case 60:return e.pm();case 62:return e.nm()}}switch(l=l7n(e),t=e.c,t){case 3:return e.Fm(l);case 4:return e.Dm(l);case 5:return e.Em(l);case 0:if(e.a==123&&e.d<e.j){if(u=e.d,a=0,i=-1,(t=_a(e.i,u++))>=48&&t<=57){for(a=t-48;u<e.j&&(t=_a(e.i,u++))>=48&&t<=57;)if(a=a*10+t-48,a<0)throw ie(new vr(yr((lr(),K4e))))}else throw ie(new vr(yr((lr(),b3t))));if(i=a,t==44){if(u>=e.j)throw ie(new vr(yr((lr(),w3t))));if((t=_a(e.i,u++))>=48&&t<=57){for(i=t-48;u<e.j&&(t=_a(e.i,u++))>=48&&t<=57;)if(i=i*10+t-48,i<0)throw ie(new vr(yr((lr(),K4e))));if(a>i)throw ie(new vr(yr((lr(),y3t))))}else i=-1}if(t!=125)throw ie(new vr(yr((lr(),v3t))));e.$l(u)?(l=(Xr(),Xr(),new hT(9,l)),e.d=u+1):(l=(Xr(),Xr(),new hT(3,l)),e.d=u),l.Lm(a),l.Km(i),Yr(e)}}return l}function KS(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;if(i==null)return null;if(e.a!=t.gk())throw ie(new dn(VL+t.ve()+n_));if(me(t,455)){if(J=bPn(f(t,670),i),!J)throw ie(new dn(ipe+i+"' is not a valid enumerator of '"+t.ve()+"'"));return J}switch(I0((Rl(),Ea),t).Kl()){case 2:{i=_u(i,!1);break}case 3:{i=_u(i,!0);break}}if(a=I0(Ea,t).Gl(),a)return a.gk().si().pi(a,i);if(L=I0(Ea,t).Il(),L){for(J=new et,S=nle(i),T=0,A=S.length;T<A;++T)v=S[T],Qe(J,L.gk().si().pi(L,v));return J}if(W=I0(Ea,t).Jl(),!W.dc()){for(z=W.Jc();z.Ob();){F=f(z.Pb(),155);try{if(J=F.gk().si().pi(F,i),J!=null)return J}catch(ue){if(ue=go(ue),!me(ue,63))throw ie(ue)}}throw ie(new dn(ipe+i+"' does not match any member types of the union datatype '"+t.ve()+"'"))}if(f(t,826).lk(),u=yxn(t.hk()),!u)return null;if(u==D$){g=0;try{g=ff(i,Ra,Qr)&To}catch(ue){if(ue=go(ue),me(ue,130))l=AJ(i),g=l[0];else throw ie(ue)}return qH(g)}if(u==Iee){for(m=0;m<UF.length;++m)try{return Oot(UF[m],i)}catch(ue){if(ue=go(ue),!me(ue,32))throw ie(ue)}throw ie(new dn(ipe+i+"' is not a date formatted string of the form yyyy-MM-dd'T'HH:mm:ss'.'SSSZ or a valid subset thereof"))}throw ie(new dn(ipe+i+"' is invalid. "))}function ofe(){ofe=Z,gi=new CS,jt(gi,(tt(),Bg),Au),jt(gi,_h,Au),jt(gi,_h,yl),jt(gi,Kf,jl),jt(gi,Kf,Au),jt(gi,dp,Au),jt(gi,dp,rd),jt(gi,vf,vl),jt(gi,vf,Au),jt(gi,wl,tl),jt(gi,wl,Au),jt(gi,wl,rd),jt(gi,wl,vl),jt(gi,tl,wl),jt(gi,tl,yl),jt(gi,tl,jl),jt(gi,tl,Au),jt(gi,fp,fp),jt(gi,fp,rd),jt(gi,fp,yl),jt(gi,nd,nd),jt(gi,nd,rd),jt(gi,nd,jl),jt(gi,wf,wf),jt(gi,wf,vl),jt(gi,wf,yl),jt(gi,Yf,Yf),jt(gi,Yf,vl),jt(gi,Yf,jl),jt(gi,rd,dp),jt(gi,rd,wl),jt(gi,rd,fp),jt(gi,rd,nd),jt(gi,rd,Au),jt(gi,rd,rd),jt(gi,rd,yl),jt(gi,rd,jl),jt(gi,vl,vf),jt(gi,vl,wl),jt(gi,vl,wf),jt(gi,vl,Yf),jt(gi,vl,vl),jt(gi,vl,yl),jt(gi,vl,jl),jt(gi,vl,Au),jt(gi,yl,_h),jt(gi,yl,tl),jt(gi,yl,fp),jt(gi,yl,wf),jt(gi,yl,rd),jt(gi,yl,vl),jt(gi,yl,yl),jt(gi,yl,Au),jt(gi,jl,Kf),jt(gi,jl,tl),jt(gi,jl,nd),jt(gi,jl,Yf),jt(gi,jl,rd),jt(gi,jl,vl),jt(gi,jl,jl),jt(gi,jl,Au),jt(gi,Au,Bg),jt(gi,Au,_h),jt(gi,Au,Kf),jt(gi,Au,dp),jt(gi,Au,vf),jt(gi,Au,wl),jt(gi,Au,tl),jt(gi,Au,rd),jt(gi,Au,vl),jt(gi,Au,yl),jt(gi,Au,jl),jt(gi,Au,Au)}function kOe(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt;for(e.d=new Ke(Gi,Gi),e.c=new Ke(qo,qo),L=t.Jc();L.Ob();)for(T=f(L.Pb(),36),ve=new ne(T.a);ve.a<ve.c.c.length;)pe=f(ee(ve),9),e.d.a=h.Math.min(e.d.a,pe.n.a-pe.d.b),e.d.b=h.Math.min(e.d.b,pe.n.b-pe.d.d),e.c.a=h.Math.max(e.c.a,pe.n.a+pe.o.a+pe.d.c),e.c.b=h.Math.max(e.c.b,pe.n.b+pe.o.b+pe.d.a);for(m=new art,A=t.Jc();A.Ob();)T=f(A.Pb(),36),a=J9n(e,T),Qe(m.a,a),a.a=a.a|!f(V(a.c,(Ge(),R1)),22).dc();for(e.b=(ole(),xt=new cy,xt.f=new ubt(i),xt.b=s8n(xt.f,m),xt),y8n((z=e.b,new CN,z)),e.e=new ha,e.a=e.b.f.e,g=new ne(m.a);g.a<g.c.c.length;)for(u=f(ee(g),834),Ee=Qyn(e.b,u),eLn(u.c,Ee.a,Ee.b),J=new ne(u.c.a);J.a<J.c.c.length;)W=f(ee(J),9),W.k==(cn(),bo)&&(ue=hIe(e,W.n,f(V(W,(Ge(),Tc)),64)),ii(Cp(W.n),ue));for(l=new ne(m.a);l.a<l.c.c.length;)for(u=f(ee(l),834),S=new ne(nCn(u));S.a<S.c.c.length;)for(v=f(ee(S),18),nt=new mY(v.a),pP(nt,0,wg(v.c)),Mr(nt,wg(v.d)),F=null,He=tr(nt,0);He.b!=He.d.c;){if(Le=f(Zn(He),8),!F){F=Le;continue}_xe(F.a,Le.a)?(e.e.a=h.Math.min(e.e.a,F.a),e.a.a=h.Math.max(e.a.a,F.a)):_xe(F.b,Le.b)&&(e.e.b=h.Math.min(e.e.b,F.b),e.a.b=h.Math.max(e.a.b,F.b)),F=Le}fX(e.e),ii(e.a,e.e)}function g7n(e,t){var i,a,u,l,g,m,v,S;if(i=0,g=0,l=t.length,m=null,S=new RN,g<l&&(gn(g,t.length),t.charCodeAt(g)==43)&&(++g,++i,g<l&&(gn(g,t.length),t.charCodeAt(g)==43||(gn(g,t.length),t.charCodeAt(g)==45))))throw ie(new sm(XS+t+'"'));for(;g<l&&(gn(g,t.length),t.charCodeAt(g)!=46)&&(gn(g,t.length),t.charCodeAt(g)!=101)&&(gn(g,t.length),t.charCodeAt(g)!=69);)++g;if(S.a+=""+uh(t==null?bl:(an(t),t),i,g),g<l&&(gn(g,t.length),t.charCodeAt(g)==46)){for(++g,i=g;g<l&&(gn(g,t.length),t.charCodeAt(g)!=101)&&(gn(g,t.length),t.charCodeAt(g)!=69);)++g;e.e=g-i,S.a+=""+uh(t==null?bl:(an(t),t),i,g)}else e.e=0;if(g<l&&(gn(g,t.length),t.charCodeAt(g)==101||(gn(g,t.length),t.charCodeAt(g)==69))&&(++g,i=g,g<l&&(gn(g,t.length),t.charCodeAt(g)==43)&&(++g,g<l&&(gn(g,t.length),t.charCodeAt(g)!=45)&&++i),m=(ga(i,l,t.length),t.substr(i,l-i)),e.e=e.e-ff(m,Ra,Qr),e.e!=Aa(e.e)))throw ie(new sm("Scale out of range."));if(v=S.a,v.length<16){if(e.f=(D6e==null&&(D6e=new RegExp("^[+-]?\\d*$","i")),D6e.test(v)?parseInt(v,10):NaN),isNaN(e.f))throw ie(new sm(XS+t+'"'));e.a=BAe(e.f)}else $xn(e,new Gv(v));for(e.d=S.a.length,u=0;u<S.a.length&&(a=_a(S.a,u),!(a!=45&&a!=48));++u)--e.d;e.d==0&&(e.d=1)}function m7n(e){ir(e.b,bh,ce(se(ft,1),Ye,2,6,[r_,"ConsistentTransient"])),ir(e.a,bh,ce(se(ft,1),Ye,2,6,[r_,"WellFormedSourceURI"])),ir(e.o,bh,ce(se(ft,1),Ye,2,6,[r_,"InterfaceIsAbstract AtMostOneID UniqueFeatureNames UniqueOperationSignatures NoCircularSuperTypes WellFormedMapEntryClass ConsistentSuperTypes DisjointFeatureAndOperationSignatures"])),ir(e.p,bh,ce(se(ft,1),Ye,2,6,[r_,"WellFormedInstanceTypeName UniqueTypeParameterNames"])),ir(e.v,bh,ce(se(ft,1),Ye,2,6,[r_,"UniqueEnumeratorNames UniqueEnumeratorLiterals"])),ir(e.R,bh,ce(se(ft,1),Ye,2,6,[r_,"WellFormedName"])),ir(e.T,bh,ce(se(ft,1),Ye,2,6,[r_,"UniqueParameterNames UniqueTypeParameterNames NoRepeatingVoid"])),ir(e.U,bh,ce(se(ft,1),Ye,2,6,[r_,"WellFormedNsURI WellFormedNsPrefix UniqueSubpackageNames UniqueClassifierNames UniqueNsURIs"])),ir(e.W,bh,ce(se(ft,1),Ye,2,6,[r_,"ConsistentOpposite SingleContainer ConsistentKeys ConsistentUnique ConsistentContainer"])),ir(e.bb,bh,ce(se(ft,1),Ye,2,6,[r_,"ValidDefaultValueLiteral"])),ir(e.eb,bh,ce(se(ft,1),Ye,2,6,[r_,"ValidLowerBound ValidUpperBound ConsistentBounds ValidType"])),ir(e.H,bh,ce(se(ft,1),Ye,2,6,[r_,"ConsistentType ConsistentBounds ConsistentArguments"]))}function b7n(e){var t,i,a,u,l;switch(i=f(V(e,(Ge(),Cu)),22),t=nX(TOt),u=f(V(e,(ot(),rx)),342),u==(_1(),W0)&&lf(t,xOt),lt(ut(V(e,mme)))?Sr(t,(wa(),rp),(ka(),Sge)):Sr(t,(wa(),su),(ka(),Sge)),V(e,(bJ(),TF))!=null&&lf(t,COt),(lt(ut(V(e,BLe)))||lt(ut(V(e,MLe))))&&Il(t,(wa(),zs),(ka(),R5e)),f(V(e,mf),86).g){case 2:case 3:case 4:Il(Sr(t,(wa(),rp),(ka(),L5e)),zs,P5e)}switch(i.Gc((us(),bte))&&Il(Sr(Sr(t,(wa(),rp),(ka(),N5e)),cu,I5e),zs,O5e),Ae(V(e,xw))!==Ae((gh(),qte))&&Sr(t,(wa(),su),(ka(),Y5e)),i.Gc(wte)&&(Sr(t,(wa(),rp),(ka(),Q5e)),Sr(t,Ab,J5e),Sr(t,su,Z5e)),Ae(V(e,Cte))!==Ae((fL(),q$))&&Ae(V(e,P1))!==Ae((k1(),uV))&&Il(t,(wa(),zs),(ka(),H5e)),lt(ut(V(e,$Le)))&&Sr(t,(wa(),su),(ka(),z5e)),lt(ut(V(e,fme)))&&Sr(t,(wa(),su),(ka(),ePe)),H6n(e)&&(Ae(V(e,rx))===Ae(W0)?a=f(V(e,Dq),298):a=f(V(e,dme),298),l=a==(zP(),Wge)?(ka(),X5e):(ka(),rPe),Sr(t,(wa(),cu),l)),f(V(e,VLe),420).g==1&&Sr(t,(wa(),cu),(ka(),k5e)),f(V(e,hDe),380).g){case 1:Sr(t,(wa(),cu),(ka(),tPe));break;case 2:Il(Sr(Sr(t,(wa(),su),(ka(),T5e)),cu,x5e),zs,C5e)}return Ae(V(e,ip))!==Ae((Pp(),Pm))&&Sr(t,(wa(),su),(ka(),nPe)),t}function v7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct;if(!t.dc()){if(u=new fd,m=i||f(t.Xb(0),18),z=m.c,qj(),L=z.i.k,!(L==(cn(),Yi)||L==xu||L==bo||L==Dg))throw ie(new dn("The target node of the edge must be a normal node or a northSouthPort."));for(Ma(u,uc(ce(se(Zo,1),Ye,8,0,[z.i.n,z.n,z.a]))),(tt(),wl).Gc(z.j)&&(J=ke(Oe(V(z,(Ge(),pD)))),A=new Ke(uc(ce(se(Zo,1),Ye,8,0,[z.i.n,z.n,z.a])).a,J),Wi(u,A,u.c.b,u.c)),T=null,a=!1,v=t.Jc();v.Ob();)g=f(v.Pb(),18),l=g.a,l.b!=0&&(a?(S=qv(ii(T,(Ln(l.b!=0),f(l.a.a.c,8))),.5),Wi(u,S,u.c.b,u.c),a=!1):a=!0,T=$a((Ln(l.b!=0),f(l.c.b.c,8))),Ba(u,l),nf(l));W=m.d,wl.Gc(W.j)&&(J=ke(Oe(V(W,(Ge(),pD)))),A=new Ke(uc(ce(se(Zo,1),Ye,8,0,[W.i.n,W.n,W.a])).a,J),Wi(u,A,u.c.b,u.c)),Ma(u,uc(ce(se(Zo,1),Ye,8,0,[W.i.n,W.n,W.a]))),e.d==(JP(),Fme)&&(ue=(Ln(u.b!=0),f(u.a.a.c,8)),pe=f(Xh(u,1),8),ve=new Kce(wCe(z.j)),ve.a*=5,ve.b*=5,Ee=Ho(new Ke(pe.a,pe.b),ue),Le=new Ke(zce(ve.a,Ee.a),zce(ve.b,Ee.b)),ii(Le,ue),He=tr(u,1),UX(He,Le),nt=(Ln(u.b!=0),f(u.c.b.c,8)),xt=f(Xh(u,u.b-2),8),ve=new Kce(wCe(W.j)),ve.a*=5,ve.b*=5,Ee=Ho(new Ke(xt.a,xt.b),nt),Ct=new Ke(zce(ve.a,Ee.a),zce(ve.b,Ee.b)),ii(Ct,nt),pP(u,u.b-1,Ct)),F=new rOe(u),Ba(m.a,Vvt(F))}}function w7n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji,id,Ms,Ef,ZA,ev,hp,C_;if(ve=f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84),Le=ve.lh(),He=ve.mh(),Ee=ve.kh()/2,W=ve.jh()/2,me(ve,193)&&(pe=f(ve,124),Le+=bg(pe).i,Le+=bg(pe).i),Le+=Ee,He+=W,tn=f(be((!e.b&&(e.b=new Qt(Yn,e,4,7)),e.b),0),84),$r=tn.lh(),ji=tn.mh(),In=tn.kh()/2,nt=tn.jh()/2,me(tn,193)&&(Kt=f(tn,124),$r+=bg(Kt).i,$r+=bg(Kt).i),$r+=In,ji+=nt,(!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i==0)m=(Bv(),S=new em,S),Qn((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a),m);else if((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i>1)for(z=new HN((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a));z.e!=z.i.gc();)Mj(z);for(g=f(be((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a),0),168),J=$r,$r>Le+Ee?J=Le+Ee:$r<Le-Ee&&(J=Le-Ee),ue=ji,ji>He+W?ue=He+W:ji<He-W&&(ue=He-W),J>Le-Ee&&J<Le+Ee&&ue>He-W&&ue<He+W&&(J=Le+Ee),B3(g,J),H3(g,ue),xt=Le,Le>$r+In?xt=$r+In:Le<$r-In&&(xt=$r-In),Ct=He,He>ji+nt?Ct=ji+nt:He<ji-nt&&(Ct=ji-nt),xt>$r-In&&xt<$r+In&&Ct>ji-nt&&Ct<ji+nt&&(Ct=ji+nt),U3(g,xt),z3(g,Ct),Jn((!g.a&&(g.a=new So(yf,g,5)),g.a)),l=IZ(t,5),ve==tn&&++l,Ef=xt-J,hp=Ct-ue,id=h.Math.sqrt(Ef*Ef+hp*hp),A=id*.20000000298023224,ZA=Ef/(l+1),C_=hp/(l+1),Ms=J,ev=ue,T=0;T<l;T++)Ms+=ZA,ev+=C_,L=Ms+wd(t,24)*VG*A-A/2,L<0?L=1:L>i&&(L=i-1),F=ev+wd(t,24)*VG*A-A/2,F<0?F=1:F>a&&(F=a-1),u=(Bv(),v=new Iv,v),MJ(u,L),DJ(u,F),Qn((!g.a&&(g.a=new So(yf,g,5)),g.a),u)}function afe(e,t){qde();var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt;if(ve=e.e,T=e.d,u=e.a,ve==0)switch(t){case 0:return"0";case 1:return OL;case 2:return"0.00";case 3:return"0.000";case 4:return"0.0000";case 5:return"0.00000";case 6:return"0.000000";default:return ue=new $v,ue.a+="0E",ue.a+=-t,ue.a}if(W=T*10+1+7,J=Pe(Xf,_m,30,W+1,15,1),i=W,T==1)if(l=u[0],l<0){nt=ia(l,ls);do A=nt,nt=pG(nt,10),J[--i]=48+mr(hh(A,Na(nt,10)))&To;while(Eu(nt,0)!=0)}else{nt=l;do A=nt,nt=nt/10|0,J[--i]=48+(A-nt*10)&To;while(nt!=0)}else{Ee=Pe(pr,Pr,30,T,15,1),He=T,au(u,0,Ee,0,He);e:for(;;){for(pe=0,m=He-1;m>=0;m--)Le=Ua(hb(pe,32),ia(Ee[m],ls)),F=LOn(Le),Ee[m]=mr(F),pe=mr(wS(F,32));z=mr(pe),L=i;do J[--i]=48+z%10&To;while((z=z/10|0)!=0&&i!=0);for(a=9-L+i,g=0;g<a&&i>0;g++)J[--i]=48;for(v=He-1;Ee[v]==0;v--)if(v==0)break e;He=v+1}for(;J[i]==48;)++i}return S=ve<0,S&&(J[--i]=45),bm(J,i,W-i)}function Lkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;switch(e.c=t,e.g=new Wn,i=(fS(),new uS(e.c)),a=new rs(i),t3e(a),ve=dr(Fe(e.c,(mG(),P8e))),v=f(Fe(e.c,Mbe),325),Le=f(Fe(e.c,jbe),424),g=f(Fe(e.c,O8e),475),Ee=f(Fe(e.c,Dbe),425),e.j=ke(Oe(Fe(e.c,dLt))),m=e.a,v.g){case 0:m=e.a;break;case 1:m=e.b;break;case 2:m=e.i;break;case 3:m=e.e;break;case 4:m=e.f;break;default:throw ie(new dn(iee+(v.f!=null?v.f:""+v.g)))}if(e.d=new _ft(m,Le,g),$e(e.d,(HP(),F$),ut(Fe(e.c,uLt))),e.d.c=lt(ut(Fe(e.c,N8e))),XX(e.c).i==0)return e.d;for(A=new Nn(XX(e.c));A.e!=A.i.gc();){for(T=f(Mn(A),27),F=T.g/2,L=T.f/2,He=new Ke(T.i+F,T.j+L);Yu(e.g,He);)eT(He,(h.Math.random()-.5)*Tm,(h.Math.random()-.5)*Tm);W=f(Fe(T,(Rr(),OD)),139),J=new jft(He,new qh(He.a-F-e.j/2-W.b,He.b-L-e.j/2-W.d,T.g+e.j+(W.b+W.c),T.f+e.j+(W.d+W.a))),Qe(e.d.i,J),qr(e.g,He,new Ya(J,T))}switch(Ee.g){case 0:if(ve==null)e.d.d=f(ct(e.d.i,0),68);else for(pe=new ne(e.d.i);pe.a<pe.c.c.length;)J=f(ee(pe),68),z=f(f(bn(e.g,J.a),47).b,27).hh(),z!=null&&Mt(z,ve)&&(e.d.d=J);break;case 1:for(u=new Ke(e.c.g,e.c.f),u.a*=.5,u.b*=.5,eT(u,e.c.i,e.c.j),l=Gi,ue=new ne(e.d.i);ue.a<ue.c.c.length;)J=f(ee(ue),68),S=Qv(J.a,u),S<l&&(l=S,e.d.d=J);break;default:throw ie(new dn(iee+(Ee.f!=null?Ee.f:""+Ee.g)))}return e.d}function Dkt(e){Wk(e,new zS(qk(zk(Gk(Hk(new ug,Jl),"ELK Rectangle Packing"),"Algorithm for packing of unconnected boxes, i.e. graphs without edges. The given order of the boxes is always preserved and the main reading direction of the boxes is left to right. The algorithm is divided into two phases. One phase approximates the width in which the rectangles can be placed. The next phase places the rectangles in rows using the previously calculated width as bounding width and bundles rectangles with a similar height in blocks. A compaction step reduces the size of the drawing. Finally, the rectangles are expanded to fill their bounding box and eliminate empty unused spaces."),new cg))),Je(e,Jl,ML,1.3),Je(e,Jl,jL,(mn(),!1)),Je(e,Jl,zT,u8e),Je(e,Jl,UT,15),Je(e,Jl,KQ,st(GPt)),Je(e,Jl,uA,st(WPt)),Je(e,Jl,DR,st(YPt)),Je(e,Jl,LR,st(XPt)),Je(e,Jl,$L,st(VPt)),Je(e,Jl,d$,st(_be)),Je(e,Jl,FL,st(JPt)),Je(e,Jl,u4e,st(h8e)),Je(e,Jl,l4e,st(f8e)),Je(e,Jl,c4e,st(Tbe)),Je(e,Jl,s4e,st(p8e)),Je(e,Jl,d4e,st(c8e)),Je(e,Jl,f4e,st(kbe)),Je(e,Jl,h4e,st(s8e)),Je(e,Jl,p4e,st(d8e)),Je(e,Jl,l$,st(o8e)),Je(e,Jl,XQ,st(qPt)),Je(e,Jl,i4e,st(Yq)),Je(e,Jl,r4e,st(i8e)),Je(e,Jl,o4e,st(Xq)),Je(e,Jl,n4e,st(l8e)),Je(e,Jl,a4e,st(a8e))}function y7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji,id,Ms,Ef;for(i.Sg("Greedy cycle removal",1),ve=t.a,Ef=ve.c.length,e.a=Pe(pr,Pr,30,Ef,15,1),e.c=Pe(pr,Pr,30,Ef,15,1),e.b=Pe(pr,Pr,30,Ef,15,1),S=0,ue=new ne(ve);ue.a<ue.c.c.length;){for(W=f(ee(ue),9),W.p=S,Ct=new ne(W.j);Ct.a<Ct.c.c.length;){for(He=f(ee(Ct),11),m=new ne(He.e);m.a<m.c.c.length;)a=f(ee(m),18),a.c.i!=W&&(In=f(V(a,(ot(),wD)),17).a,e.a[S]+=In>0?In+1:1);for(g=new ne(He.g);g.a<g.c.c.length;)a=f(ee(g),18),a.d.i!=W&&(In=f(V(a,(ot(),wD)),17).a,e.c[S]+=In>0?In+1:1)}e.c[S]==0?Mr(e.e,W):e.a[S]==0&&Mr(e.f,W),++S}for(z=-1,F=1,A=new et,e.d=f(V(t,(Ge(),ZR)),233);Ef>0;){for(;e.e.b!=0;)ji=f(Zse(e.e),9),e.b[ji.p]=z--,GIe(e,ji),--Ef;for(;e.f.b!=0;)id=f(Zse(e.f),9),e.b[id.p]=F++,GIe(e,id),--Ef;if(Ef>0){for(L=Ra,pe=new ne(ve);pe.a<pe.c.c.length;)W=f(ee(pe),9),e.b[W.p]==0&&(Ee=e.c[W.p]-e.a[W.p],Ee>=L&&(Ee>L&&(A.c.length=0,L=Ee),vn(A.c,W)));T=e.qg(A),e.b[T.p]=F++,GIe(e,T),--Ef}}for($r=ve.c.length+1,S=0;S<ve.c.length;S++)e.b[S]<0&&(e.b[S]+=$r);for(J=new ne(ve);J.a<J.c.c.length;)for(W=f(ee(J),9),tn=$ht(W.j),nt=tn,xt=0,Kt=nt.length;xt<Kt;++xt)for(He=nt[xt],Le=pm(He.g),u=Le,l=0,v=u.length;l<v;++l)a=u[l],Ms=a.d.i.p,e.b[W.p]>e.b[Ms]&&(VS(a,!0),$e(t,Nq,(mn(),!0)));e.a=null,e.c=null,e.b=null,nf(e.f),nf(e.e),i.Tg()}function E7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(L=0,u=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));u.e!=u.i.gc();)a=f(Mn(u),27),lt(ut(Fe(a,(ot(),ME))))||(T=so(a),(Ae(Fe(T,ip))!==Ae((Pp(),Pm))||Ae(Fe(T,RE))===Ae((R0(),uD))||Ae(Fe(T,RE))===Ae((R0(),cD))||lt(ut(Fe(T,h_)))||Ae(Fe(T,nx))!==Ae((hE(),XT))||Ae(Fe(T,xw))===Ae((gh(),w_))||Ae(Fe(T,xw))===Ae((gh(),ux))||Ae(Fe(T,LE))===Ae((D0(),SD))||Ae(Fe(T,LE))===Ae((D0(),_D)))&&!lt(ut(Fe(a,Ite)))&&(pi(a,(Ge(),hi),rt(L)),++L),Ikt(e,a,i));for(L=0,S=new Nn((!t.b&&(t.b=new Be(vo,t,12,3)),t.b));S.e!=S.i.gc();)m=f(Mn(S),85),(Ae(Fe(t,(ot(),ip)))!==Ae((Pp(),Pm))||Ae(Fe(t,RE))===Ae((R0(),uD))||Ae(Fe(t,RE))===Ae((R0(),cD))||lt(ut(Fe(t,h_)))||Ae(Fe(t,nx))!==Ae((hE(),XT))||Ae(Fe(t,xw))===Ae((gh(),w_))||Ae(Fe(t,xw))===Ae((gh(),ux))||Ae(Fe(t,LE))===Ae((D0(),SD))||Ae(Fe(t,LE))===Ae((D0(),_D)))&&(pi(m,(Ge(),hi),rt(L)),++L),W=lde(m),J=lyt(m),A=lt(ut(Fe(W,ix))),z=!lt(ut(Fe(m,ME))),F=A&&HS(m)&&lt(ut(Fe(m,PE))),l=so(W)==t&&so(W)==so(J),g=(so(W)==t&&J==t)^(so(J)==t&&W==t),z&&!F&&(g||l)&&TOe(e,m,t,i);if(so(t))for(v=new Nn(qdt(so(t)));v.e!=v.i.gc();)m=f(Mn(v),85),W=lde(m),W==t&&HS(m)&&(F=lt(ut(Fe(W,(ot(),ix))))&&lt(ut(Fe(m,PE))),F&&TOe(e,m,t,i))}function Mkt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;for(Le=f(be((!e.a&&(e.a=new Be(Ii,e,6,6)),e.a),0),168),T=new fd,Ee=new Wn,He=YSt(Le),ml(Ee.f,Le,He),L=new Wn,a=new ki,z=pb(Df(ce(se(Gf,1),Jt,20,0,[(!t.d&&(t.d=new Qt(vo,t,8,5)),t.d),(!t.e&&(t.e=new Qt(vo,t,7,4)),t.e)])));Kn(z);){if(F=f(Dn(z),85),(!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i!=1)throw ie(new dn(TCt+(!e.a&&(e.a=new Be(Ii,e,6,6)),e.a).i));F!=e&&(J=f(be((!F.a&&(F.a=new Be(Ii,F,6,6)),F.a),0),168),Wi(a,J,a.c.b,a.c),W=f(ac(Is(Ee.f,J)),13),W||(W=YSt(J),ml(Ee.f,J,W)),A=i?Ho(new ja(f(ct(He,He.c.length-1),8)),f(ct(W,W.c.length-1),8)):Ho(new ja((Gt(0,He.c.length),f(He.c[0],8))),(Gt(0,W.c.length),f(W.c[0],8))),ml(L.f,J,A))}if(a.b!=0)for(ue=f(ct(He,i?He.c.length-1:0),8),S=1;S<He.c.length;S++){for(pe=f(ct(He,i?He.c.length-1-S:S),8),u=tr(a,0);u.b!=u.d.c;)J=f(Zn(u),168),W=f(ac(Is(Ee.f,J)),13),W.c.length<=S?Cue(u):(ve=ii(new ja(f(ct(W,i?W.c.length-1-S:S),8)),f(ac(Is(L.f,J)),8)),(pe.a!=ve.a||pe.b!=ve.b)&&(l=pe.a-ue.a,m=pe.b-ue.b,g=ve.a-ue.a,v=ve.b-ue.b,g*m==v*l&&(l==0||isNaN(l)?l:l<0?-1:1)==(g==0||isNaN(g)?g:g<0?-1:1)&&(m==0||isNaN(m)?m:m<0?-1:1)==(v==0||isNaN(v)?v:v<0?-1:1)?(h.Math.abs(l)<h.Math.abs(g)||h.Math.abs(m)<h.Math.abs(v))&&Wi(T,pe,T.c.b,T.c):S>1&&Wi(T,ue,T.c.b,T.c),Cue(u)));ue=pe}return T}function jkt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;for(i.Sg(Mxt,1),Kt=f($d(Jr(new Ft(null,new Bt(t,16)),new $O),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),T=f($d(Jr(new Ft(null,new Bt(t,16)),new jtt(t)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[Zl]))),15),z=f($d(Jr(new Ft(null,new Bt(t,16)),new Mtt(t)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[Zl]))),15),W=Pe(Zte,eee,39,t.gc(),0,1),g=0;g<T.gc();g++)u=f(T.Xb(g),39),Ct=f(V(u,(Ec(),DA)),17).a,Ct>=0&&Ct<T.gc()&&!W[Ct]&&(W[Ct]=u,T.ed(g),--g);for(m=0;m<T.gc();m++)for(u=f(T.Xb(m),39),Ct=f(V(u,(Ec(),DA)),17).a,L=0;;L++){if(F=Ct+L,F<W.length&&F>=0&&!W[F]){W[F]=u,T.ed(m),--m;break}if(F=Ct-L,F<W.length&&F>=0&&!W[F]){W[F]=u,T.ed(m),--m;break}}for(z.gd(new J6),v=W.length-1;v>=0;v--)!W[v]&&!z.dc()&&(W[v]=f(z.Xb(0),39),z.ed(0));for(S=0;S<W.length;S++)!W[S]&&!Kt.dc()&&(W[S]=f(Kt.Xb(0),39),Kt.ed(0));for(l=0;l<W.length;l++)$e(W[l],(_i(),bF),rt(l));for(A=f(HCn(Jr(new Ft(null,new Bt(t,16)),new FO)),519),He=A,nt=0,xt=He.length;nt<xt;++nt){for(Le=He[nt],a=Wue(Le),jkt(e,a,i.bh(1/A.length|0)),Ht(),sE(a,new sae((_i(),bF))),J=new ki,Ee=tr(a,0);Ee.b!=Ee.d.c;)for(ve=f(Zn(Ee),39),pe=tr(Le.d,0);pe.b!=pe.d.c;)ue=f(Zn(pe),65),ue.c==ve&&Wi(J,ue,J.c.b,J.c);nf(Le.d),Ba(Le.d,J)}i.Tg()}function $kt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;for(ve=ke(Oe(Fe(t,(wb(),fx)))),F=ke(Oe(Fe(t,yF))),L=ke(Oe(Fe(t,cne))),qxe((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a)),ue=N_t((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a),ve,e.b),J=0;J<ue.c.length;J++)if(v=(Gt(J,ue.c.length),f(ue.c[J],186)),J!=0&&(z=(Gt(J-1,ue.c.length),f(ue.c[J-1],186)),t0t(v,z.f+z.b+e.b)),W=k7n(J,ue,ve,e.b,lt(ut(Fe(t,(Ag(),kbe))))),lt(ut(W.b))){for(l=new ne(v.a);l.a<l.c.c.length;)u=f(ee(l),172),u.c=!1,u.k=!1,__t(u);v.d=new et,v.e=ve,--J}else if(yIn(e,v),J+1<ue.c.length&&(e.e=h.Math.max(v.e+e.b+f(ct((Gt(J+1,ue.c.length),f(ue.c[J+1],186)).a,0),172).r-ve,e.c),e.f=h.Math.min(v.e+e.b+f(ct((Gt(J+1,ue.c.length),f(ue.c[J+1],186)).a,0),172).r-ve,e.d),v.d.c.length!=0&&(e.c=h.Math.max(e.c,f(ct(v.d,v.d.c.length-1),316).d+(v.d.c.length<=1?0:e.b)),e.d=h.Math.min(e.c,f(ct(v.d,v.d.c.length-1),316).d+(v.d.c.length<=1?0:e.b)))),ue.c.length==1)for(A=f(ct(v.d,v.d.c.length-1),316),T=f(ct(A.a,A.a.c.length-1),172),m=new ne(T.n);m.a<m.c.c.length;)g=f(ee(m),208),e.c=h.Math.max(e.c,T.r-g.d),e.d=h.Math.min(e.d,T.r-g.d),e.e=h.Math.max(e.e,g.d+e.b),e.f=h.Math.min(e.f,g.d+e.b);return pe=Ewt(ue,e.b),Ee=h.Math.max(pe.a,F-(i.b+i.c)),S=h.Math.max(pe.b,L-(i.d+i.a)),a=S-pe.b,pi(t,wF,a),pi(t,une,ue),new pke(e.a,Ee,pe.b+a,(Hj(),Jq))}function S7n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In;if(nt=f(V(e,(ot(),Xi)),100),nt!=(oa(),Fg)&&nt!=Pw){for(z=e.b,F=z.c.length,T=new Du((uf(F+2,ffe),NJ(Ua(Ua(5,F+2),(F+2)/10|0)))),W=new Du((uf(F+2,ffe),NJ(Ua(Ua(5,F+2),(F+2)/10|0)))),Qe(T,new Wn),Qe(T,new Wn),Qe(W,new et),Qe(W,new et),He=new et,t=0;t<F;t++)for(i=(Gt(t,z.c.length),f(z.c[t],25)),xt=(Gt(t,T.c.length),f(T.c[t],91)),J=new Wn,vn(T.c,J),Kt=(Gt(t,W.c.length),f(W.c[t],15)),pe=new et,vn(W.c,pe),u=new ne(i.a);u.a<u.c.c.length;){if(a=f(ee(u),9),FCe(a)){vn(He.c,a);continue}for(S=new Tn(Cn(Fo(a).a.Jc(),new B));Kn(S);)m=f(Dn(S),18),tn=m.c.i,FCe(tn)&&(Ct=f(xt.xc(V(tn,(Ge(),oi))),9),Ct||(Ct=aEt(e,tn),xt.yc(V(tn,oi),Ct),Kt.Ec(Ct)),Ia(m,f(ct(Ct.j,1),11)));for(v=new Tn(Cn(ao(a).a.Jc(),new B));Kn(v);)m=f(Dn(v),18),In=m.d.i,FCe(In)&&(ue=f(bn(J,V(In,(Ge(),oi))),9),ue||(ue=aEt(e,In),qr(J,V(In,oi),ue),vn(pe.c,ue)),ua(m,f(ct(ue.j,0),11)))}for(A=0;A<W.c.length;A++)if(ve=(Gt(A,W.c.length),f(W.c[A],15)),!ve.dc())for(L=null,A==0?(L=new ru(e),fT(0,z.c.length),y7(z.c,0,L)):A==T.c.length-1?(L=new ru(e),vn(z.c,L)):L=(Gt(A-1,z.c.length),f(z.c[A-1],25)),g=ve.Jc();g.Ob();)l=f(g.Pb(),9),jo(l,L);for(Le=new ne(He);Le.a<Le.c.c.length;)Ee=f(ee(Le),9),jo(Ee,null);$e(e,(Ge(),Jge),He)}}function _7n(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji,id,Ms;for(ji=new et,z=new ne(t.b);z.a<z.c.c.length;)for(L=f(ee(z),25),Le=new ne(L.a);Le.a<Le.c.c.length;){for(Ee=f(ee(Le),9),Ee.p=-1,A=Ra,xt=Ra,Kt=new ne(Ee.j);Kt.a<Kt.c.c.length;){for(Ct=f(ee(Kt),11),u=new ne(Ct.e);u.a<u.c.c.length;)i=f(ee(u),18),tn=f(V(i,(ot(),e4)),17).a,A=h.Math.max(A,tn);for(a=new ne(Ct.g);a.a<a.c.c.length;)i=f(ee(a),18),tn=f(V(i,(ot(),e4)),17).a,xt=h.Math.max(xt,tn)}$e(Ee,Yte,rt(A)),$e(Ee,Xte,rt(xt))}for(ue=0,F=new ne(t.b);F.a<F.c.c.length;)for(L=f(ee(F),25),Le=new ne(L.a);Le.a<Le.c.c.length;)Ee=f(ee(Le),9),Ee.p<0&&($r=new g2e,$r.b=ue++,vSt(e,Ee,$r),vn(ji.c,$r));for(nt=lb(ji.c.length),T=lb(ji.c.length),g=0;g<ji.c.length;g++)Qe(nt,new et),Qe(T,rt(0));for(H8n(t,ji,nt,T),id=f(kg(ji,Pe(W6t,Axt,261,ji.c.length,0,1)),833),He=f(kg(nt,Pe(qf,BT,15,nt.c.length,0,1)),198),S=Pe(pr,Pr,30,T.c.length,15,1),m=0;m<S.length;m++)S[m]=(Gt(m,T.c.length),f(T.c[m],17)).a;for(pe=0,ve=new et,v=0;v<id.length;v++)S[v]==0&&vn(ve.c,id[v]);for(J=Pe(pr,Pr,30,id.length,15,1);ve.c.length!=0;)for($r=f(_0(ve,0),261),J[$r.b]=pe++;!He[$r.b].dc();)Ms=f(He[$r.b].ed(0),261),--S[Ms.b],S[Ms.b]==0&&vn(ve.c,Ms);for(e.a=Pe(W6t,Axt,261,id.length,0,1),l=0;l<id.length;l++)for(W=id[l],In=J[l],e.a[In]=W,W.b=In,Le=new ne(W.e);Le.a<Le.c.c.length;)Ee=f(ee(Le),9),Ee.p=In;return e.a}function k7n(e,t,i,a,u){var l,g,m,v,S,T,A,L,F,z,W,J,ue;for(W=!1,v=!1,L=e+1,z=(Gt(e,t.c.length),f(t.c[e],186)),m=z.a,S=null,g=0;g<z.a.c.length;g++)if(l=(Gt(g,m.c.length),f(m.c[g],172)),!l.c){if(l.b.c.length==0){b0(),fG(z,l),--g,W=!0;continue}if(l.k||(S&&PZ(S),S=new fTe(S?S.e+S.d+a:0,z.f,a),cG(l,S.e+S.d,z.f),Qe(z.d,S),Qxe(S,l),l.k=!0),T=null,T=(ue=null,g<z.a.c.length-1?ue=f(ct(z.a,g+1),172):L<t.c.length&&(Gt(L,t.c.length),f(t.c[L],186)).a.c.length!=0&&(ue=f(ct((Gt(L,t.c.length),f(t.c[L],186)).a,0),172)),ue),J=!1,T&&(J=!ei(T.j,z)),T){if(T.b.c.length!=0&&!lt(ut(f(ct(T.b,0),27).mf((Ag(),Xq)))))IH(l,i-l.s),PZ(l.q),W=W|TNn(z,l,T,i,a);else{fG(z,T);break}if(T.b.c.length==0)for(t.c.length>L&&fG((Gt(L,t.c.length),f(t.c[L],186)),T),T=null;t.c.length>L&&(Gt(L,t.c.length),f(t.c[L],186)).a.c.length==0;)hl(t,(Gt(L,t.c.length),t.c[L]));if(!T){--g;continue}if(!lt(ut(f(ct(T.b,0),27).mf((Ag(),Xq))))&&qLn(t,z,l,T,J,i,L,a)){W=!0;continue}if(J){if(F=z.b,A=T.f,!lt(ut(f(ct(T.b,0),27).mf(Xq)))&&a8n(t,z,l,T,i,L,a,u)){if(W=!0,F<A){v=!0,T.j=z;break}continue}else if(xCe(z,l)){l.c=!0,W=!0;continue}}else if(xCe(z,l)){l.c=!0,W=!0;continue}if(W)continue}if(xCe(z,l)){l.c=!0,W=!0,T&&(T.k=!1);continue}else PZ(l.q)}return new Ya((mn(),!!W),!!v)}function Yr(e){var t,i,a;if(e.d>=e.j){e.a=-1,e.c=1;return}if(t=_a(e.i,e.d++),e.a=t,e.b==1){switch(t){case 92:if(a=10,e.d>=e.j)throw ie(new vr(yr((lr(),dee))));e.a=_a(e.i,e.d++);break;case 45:(e.e&512)==512&&e.d<e.j&&_a(e.i,e.d)==91?(++e.d,a=24):a=0;break;case 91:if((e.e&512)!=512&&e.d<e.j&&_a(e.i,e.d)==58){++e.d,a=20;break}default:(t&64512)==GG&&e.d<e.j&&(i=_a(e.i,e.d),(i&64512)==56320&&(e.a=Xa+(t-GG<<10)+i-56320,++e.d)),a=0}e.c=a;return}switch(t){case 124:a=2;break;case 42:a=3;break;case 43:a=4;break;case 63:a=5;break;case 41:a=7;break;case 46:a=8;break;case 91:a=9;break;case 94:a=11;break;case 36:a=12;break;case 40:if(a=6,e.d>=e.j||_a(e.i,e.d)!=63)break;if(++e.d>=e.j)throw ie(new vr(yr((lr(),bpe))));switch(t=_a(e.i,e.d++),t){case 58:a=13;break;case 61:a=14;break;case 33:a=15;break;case 91:a=19;break;case 62:a=18;break;case 60:if(e.d>=e.j)throw ie(new vr(yr((lr(),bpe))));if(t=_a(e.i,e.d++),t==61)a=16;else if(t==33)a=17;else throw ie(new vr(yr((lr(),QCt))));break;case 35:for(;e.d<e.j&&(t=_a(e.i,e.d++),t!=41););if(t!=41)throw ie(new vr(yr((lr(),e3t))));a=21;break;default:if(t==45||97<=t&&t<=122||65<=t&&t<=90){--e.d,a=22;break}else if(t==40){a=23;break}throw ie(new vr(yr((lr(),bpe))))}break;case 92:if(a=10,e.d>=e.j)throw ie(new vr(yr((lr(),dee))));e.a=_a(e.i,e.d++);break;default:a=0}e.c=a}function T7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J;if(i.Sg("Process compaction",1),!!lt(ut(V(t,(Ec(),fMe))))){for(u=f(V(t,E_),86),F=ke(Oe(V(t,Qme))),eMn(e,t,u),B9n(t,F/2/2),z=t.b,sE(z,new Itt(u)),S=tr(z,0);S.b!=S.d.c;)if(v=f(Zn(S),39),!lt(ut(V(v,(_i(),Iw))))){if(a=LPn(v,u),W=_Dn(v,t),A=0,L=0,a)switch(J=a.e,u.g){case 2:A=J.a-F-v.f.a,W.e.a-F-v.f.a<A&&(A=W.e.a-F-v.f.a),L=A+v.f.a;break;case 1:A=J.a+a.f.a+F,W.e.a+F>A&&(A=W.e.a+W.f.a+F),L=A+v.f.a;break;case 4:A=J.b-F-v.f.b,W.e.b-F-v.f.b<A&&(A=W.e.b-F-v.f.b),L=A+v.f.b;break;case 3:A=J.b+a.f.b+F,W.e.b+F>A&&(A=W.e.b+W.f.b+F),L=A+v.f.b}else if(W)switch(u.g){case 2:A=W.e.a-F-v.f.a,L=A+v.f.a;break;case 1:A=W.e.a+W.f.a+F,L=A+v.f.a;break;case 4:A=W.e.b-F-v.f.b,L=A+v.f.b;break;case 3:A=W.e.b+W.f.b+F,L=A+v.f.b}Ae(V(t,Zme))===Ae((yj(),Gq))?(l=A,g=L,m=S1(Jr(new Ft(null,new Bt(e.a,16)),new bat(l,g))),m.a!=null?u==(ko(),Ds)||u==Gs?v.e.a=A:v.e.b=A:(u==(ko(),Ds)||u==Wf?m=S1(Jr(qmt(new Ft(null,new Bt(e.a,16))),new Ott(l))):m=S1(Jr(qmt(new Ft(null,new Bt(e.a,16))),new Ntt(l))),m.a!=null&&(u==Ds||u==Gs?v.e.a=ke(Oe((Ln(m.a!=null),f(m.a,47)).a)):v.e.b=ke(Oe((Ln(m.a!=null),f(m.a,47)).a)))),m.a!=null&&(T=Oc(e.a,(Ln(m.a!=null),m.a),0),T>0&&T!=f(V(v,Lm),17).a&&($e(v,iMe,(mn(),!0)),$e(v,Lm,rt(T))))):u==(ko(),Ds)||u==Gs?v.e.a=A:v.e.b=A}i.Tg()}}function ot(){ot=Z,wme=(Rr(),eDt),iDe=tDt,jq=nDt,op=rDt,OA=P9e,b_=L9e,cx=D9e,yD=M9e,ED=j9e,yme=Ene,v_=zE,Eme=iDt,rF=B9e,Fte=a4,Mq=(xOe(),h4t),IA=p4t,jE=g4t,NA=m4t,t6t=new pa(aV,rt(0)),wD=l4t,rDe=d4t,e4=f4t,hDe=$4t,aDe=w4t,sDe=S4t,_me=I4t,cDe=T4t,uDe=C4t,Bte=z4t,kme=F4t,dDe=L4t,lDe=R4t,fDe=M4t,VLe=VRt,gD=HRt,eF=GRt,g_=i4t,nF=o4t,gme=vRt,DLe=yRt,a6t=RD,s6t=Sne,o6t=ND,i6t=sV,oDe=(gR(),bx),new pa(s4,oDe),ZLe=new gS(12),JLe=new pa(V0,ZLe),RLe=(k1(),DD),P1=new pa(l9e,RLe),ox=new pa(Sd,0),n6t=new pa(t1e,rt(1)),Tte=new pa(ID,DL),ME=yne,Xi=OF,vD=qA,K4t=nV,Nm=ULt,rx=UA,r6t=new pa(n1e,(mn(),!0)),ix=rV,PE=Wbe,DE=UE,$te=Ow,vme=hx,NLe=(ko(),$g),mf=new pa(BE,NLe),p_=HA,Mte=v9e,ax=px,e6t=e1e,tDe=N9e,eDe=(K3(),fV),new pa(x9e,eDe),J4t=Xbe,Z4t=Jbe,Q4t=Zbe,X4t=Ybe,Sme=v4t,LE=zRt,xw=URt,iF=b4t,el=LRt,RE=lRt,Q$=uRt,h_=KNt,ALe=YNt,dme=QNt,Dq=XNt,fme=sRt,WLe=WRt,KLe=KRt,ULe=ARt,jte=c4t,bme=JRt,mme=_Rt,XLe=n4t,LLe=mRt,pme=bRt,lme=tV,YLe=YRt,Cte=BNt,xLe=FNt,xte=$Nt,$Le=xRt,jLe=TRt,FLe=CRt,mD=GA,Ls=zA,z0=qLt,Rm=Vbe,hme=qbe,ILe=tRt,H0=Qbe,Z$=GLt,Pte=WLt,m_=A9e,QLe=XLt,bD=JLt,HLe=MRt,GLe=$Rt,sx=o4,sme=jNt,qLe=BRt,Rte=hRt,Nte=fRt,Dte=OD,zLe=NRt,tF=QRt,$q=$9e,OLe=dRt,nDe=u4t,PLe=pRt,q4t=rRt,V4t=iRt,Y4t=PRt,W4t=oRt,BLe=Kbe,Lte=DRt,Ote=aRt,ip=WNt,cme=GNt,Ate=zNt,CLe=HNt,Ite=qNt,nx=UNt,ume=VNt,MLe=kRt}function x7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le;if(i.Sg("Coffman-Graham Layering",1),t.a.c.length==0){i.Tg();return}for(Le=f(V(t,(ot(),zLe)),17).a,v=0,g=0,L=new ne(t.a);L.a<L.c.c.length;)for(A=f(ee(L),9),A.p=v++,l=new Tn(Cn(ao(A).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),u.p=g++;for(e.d=Pe(kd,Tb,30,v,16,1),e.a=Pe(kd,Tb,30,g,16,1),e.b=Pe(pr,Pr,30,v,15,1),e.e=Pe(pr,Pr,30,v,15,1),e.f=Pe(pr,Pr,30,v,15,1),UJ(e.c),ZIn(e,t),z=new jX(new Ket(e)),Ee=new ne(t.a);Ee.a<Ee.c.c.length;){for(pe=f(ee(Ee),9),l=new Tn(Cn(Fo(pe).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),e.a[u.p]||++e.b[pe.p];e.b[pe.p]==0&&KN(dL(z,pe),NL)}for(m=0;z.b.c.length!=0;)for(pe=f(Tue(z),9),e.f[pe.p]=m++,l=new Tn(Cn(ao(pe).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),!e.a[u.p]&&(J=u.d.i,--e.b[J.p],jt(e.c,J,rt(e.f[pe.p])),e.b[J.p]==0&&KN(dL(z,J),NL));for(F=new jX(new Yet(e)),ve=new ne(t.a);ve.a<ve.c.c.length;){for(pe=f(ee(ve),9),l=new Tn(Cn(ao(pe).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),e.a[u.p]||++e.e[pe.p];e.e[pe.p]==0&&KN(dL(F,pe),NL)}for(T=new et,a=Ddt(t,T);F.b.c.length!=0;)for(ue=f(Tue(F),9),(a.a.c.length>=Le||!i3n(ue,a))&&(a=Ddt(t,T)),jo(ue,a),l=new Tn(Cn(Fo(ue).a.Jc(),new B));Kn(l);)u=f(Dn(l),18),!e.a[u.p]&&(W=u.c.i,--e.e[W.p],e.e[W.p]==0&&KN(dL(F,W),NL));for(S=T.c.length-1;S>=0;--S)Qe(t.b,(Gt(S,T.c.length),f(T.c[S],25)));t.a.c.length=0,i.Tg()}function Fkt(e){var t,i,a,u,l,g,m,v,S;for(e.b=1,Yr(e),t=null,e.c==0&&e.a==94?(Yr(e),t=(Xr(),Xr(),new of(4)),Su(t,0,JL),m=new of(4)):m=(Xr(),Xr(),new of(4)),u=!0;(S=e.c)!=1;){if(S==0&&e.a==93&&!u){t&&(t$(t,m),m=t);break}if(i=e.a,a=!1,S==10)switch(i){case 100:case 68:case 119:case 87:case 115:case 83:MT(m,yL(i)),a=!0;break;case 105:case 73:case 99:case 67:i=(MT(m,yL(i)),-1),i<0&&(a=!0);break;case 112:case 80:if(v=qAe(e,i),!v)throw ie(new vr(yr((lr(),vpe))));MT(m,v),a=!0;break;default:i=IIe(e)}else if(S==24&&!u){if(t&&(t$(t,m),m=t),l=Fkt(e),t$(m,l),e.c!=0||e.a!=93)throw ie(new vr(yr((lr(),u3t))));break}if(Yr(e),!a){if(S==0){if(i==91)throw ie(new vr(yr((lr(),V4e))));if(i==93)throw ie(new vr(yr((lr(),W4e))));if(i==45&&!u&&e.a!=93)throw ie(new vr(yr((lr(),wpe))))}if(e.c!=0||e.a!=45||i==45&&u)Su(m,i,i);else{if(Yr(e),(S=e.c)==1)throw ie(new vr(yr((lr(),fee))));if(S==0&&e.a==93)Su(m,i,i),Su(m,45,45);else{if(S==0&&e.a==93||S==24)throw ie(new vr(yr((lr(),wpe))));if(g=e.a,S==0){if(g==91)throw ie(new vr(yr((lr(),V4e))));if(g==93)throw ie(new vr(yr((lr(),W4e))));if(g==45)throw ie(new vr(yr((lr(),wpe))))}else S==10&&(g=IIe(e));if(Yr(e),i>g)throw ie(new vr(yr((lr(),f3t))));Su(m,i,g)}}}u=!1}if(e.c==1)throw ie(new vr(yr((lr(),fee))));return eA(m),Qj(m),e.b=0,Yr(e),m}function Bkt(e,t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;Ee=!1;do for(Ee=!1,l=t?new Ic(e.a.b).a.gc()-2:1;t?l>=0:l<new Ic(e.a.b).a.gc();l+=t?-1:1)for(u=RTe(e.a,rt(l)),F=0;F<u.b;F++)if(A=f(Xh(u,F),9),!!li(A,(Ge(),hi))&&!(iOn(e.a,rt(l))&&e.r==(gh(),w_)||oOn(e.a,rt(l))&&e.r==(gh(),ux))){for(ve=!0,ue=0;ue<u.b;ue++)J=f(Xh(u,ue),9),li(J,hi)&&(t&&f(V(A,hi),17).a<f(V(J,hi),17).a||!t&&f(V(A,hi),17).a>f(V(J,hi),17).a)&&(ve=!1);if(ve){for(v=t?l+1:l-1,m=RTe(e.a,rt(v)),g=!1,pe=!0,a=!1,T=tr(m,0);T.b!=T.d.c;)S=f(Zn(T),9),li(S,hi)?S.p!=A.p&&(g=g|(t?f(V(S,hi),17).a<f(V(A,hi),17).a:f(V(S,hi),17).a>f(V(A,hi),17).a),pe=!1):!g&&pe&&S.k==(cn(),Vc)&&(a=!0,t?L=f(Dn(new Tn(Cn(Fo(S).a.Jc(),new B))),18).c.i:L=f(Dn(new Tn(Cn(ao(S).a.Jc(),new B))),18).d.i,L==A&&(t?i=f(Dn(new Tn(Cn(ao(S).a.Jc(),new B))),18).d.i:i=f(Dn(new Tn(Cn(Fo(S).a.Jc(),new B))),18).c.i,(t?f(Zk(e.a,i),17).a-f(Zk(e.a,L),17).a:f(Zk(e.a,L),17).a-f(Zk(e.a,i),17).a)<=2&&(pe=!1)));if(a&&pe&&(t?i=f(Dn(new Tn(Cn(ao(A).a.Jc(),new B))),18).d.i:i=f(Dn(new Tn(Cn(Fo(A).a.Jc(),new B))),18).c.i,(t?f(Zk(e.a,i),17).a-f(Zk(e.a,A),17).a:f(Zk(e.a,A),17).a-f(Zk(e.a,i),17).a)<=2&&i.k==(cn(),Yi)&&(pe=!1)),g||pe){for(W=_2t(e,A,t);W.a.gc()!=0;)z=f(W.a.ec().Jc().Pb(),9),W.a.Ac(z)!=null,Ba(W,_2t(e,z,t));--F,Ee=!0}}}while(Ee)}function C7n(e){ir(e.c,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#decimal"])),ir(e.d,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#integer"])),ir(e.e,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#boolean"])),ir(e.f,kr,ce(se(ft,1),Ye,2,6,[Ta,"EBoolean",Gr,"EBoolean:Object"])),ir(e.i,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#byte"])),ir(e.g,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#hexBinary"])),ir(e.j,kr,ce(se(ft,1),Ye,2,6,[Ta,"EByte",Gr,"EByte:Object"])),ir(e.n,kr,ce(se(ft,1),Ye,2,6,[Ta,"EChar",Gr,"EChar:Object"])),ir(e.t,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#double"])),ir(e.u,kr,ce(se(ft,1),Ye,2,6,[Ta,"EDouble",Gr,"EDouble:Object"])),ir(e.F,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#float"])),ir(e.G,kr,ce(se(ft,1),Ye,2,6,[Ta,"EFloat",Gr,"EFloat:Object"])),ir(e.I,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#int"])),ir(e.J,kr,ce(se(ft,1),Ye,2,6,[Ta,"EInt",Gr,"EInt:Object"])),ir(e.N,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#long"])),ir(e.O,kr,ce(se(ft,1),Ye,2,6,[Ta,"ELong",Gr,"ELong:Object"])),ir(e.Z,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#short"])),ir(e.$,kr,ce(se(ft,1),Ye,2,6,[Ta,"EShort",Gr,"EShort:Object"])),ir(e._,kr,ce(se(ft,1),Ye,2,6,[Ta,"http://www.w3.org/2001/XMLSchema#string"]))}function A7n(e,t,i,a,u,l,g){var m,v,S,T,A,L,F,z;return L=f(a.a,17).a,F=f(a.b,17).a,A=e.b,z=e.c,m=0,T=0,t==(ko(),Ds)||t==Gs?(T=Fz(T0t(uT(Lu(new Ft(null,new Bt(i.b,16)),new Sv),new Y6))),A.e.b+A.f.b/2>T?(S=++F,m=ke(Oe(Dd(oT(Lu(new Ft(null,new Bt(i.b,16)),new yat(u,S)),new MO))))):(v=++L,m=ke(Oe(Dd(XN(Lu(new Ft(null,new Bt(i.b,16)),new Eat(u,v)),new $8)))))):(T=Fz(T0t(uT(Lu(new Ft(null,new Bt(i.b,16)),new jO),new j8))),A.e.a+A.f.a/2>T?(S=++F,m=ke(Oe(Dd(oT(Lu(new Ft(null,new Bt(i.b,16)),new vat(u,S)),new X6))))):(v=++L,m=ke(Oe(Dd(XN(Lu(new Ft(null,new Bt(i.b,16)),new wat(u,v)),new Da)))))),t==Ds?(Ma(e.a,new Ke(ke(Oe(V(A,(_i(),Fp))))-u,m)),Ma(e.a,new Ke(z.e.a+z.f.a+u+l,m)),Ma(e.a,new Ke(z.e.a+z.f.a+u+l,z.e.b+z.f.b/2)),Ma(e.a,new Ke(z.e.a+z.f.a,z.e.b+z.f.b/2))):t==Gs?(Ma(e.a,new Ke(ke(Oe(V(A,(_i(),ap))))+u,A.e.b+A.f.b/2)),Ma(e.a,new Ke(A.e.a+A.f.a+u,m)),Ma(e.a,new Ke(z.e.a-u-l,m)),Ma(e.a,new Ke(z.e.a-u-l,z.e.b+z.f.b/2)),Ma(e.a,new Ke(z.e.a,z.e.b+z.f.b/2))):t==Wf?(Ma(e.a,new Ke(m,ke(Oe(V(A,(_i(),Fp))))-u)),Ma(e.a,new Ke(m,z.e.b+z.f.b+u+l)),Ma(e.a,new Ke(z.e.a+z.f.a/2,z.e.b+z.f.b+u+l)),Ma(e.a,new Ke(z.e.a+z.f.a/2,z.e.b+z.f.b+u))):(e.a.b==0||(f(Hh(e.a),8).b=ke(Oe(V(A,(_i(),ap))))+u*f(g.b,17).a),Ma(e.a,new Ke(m,ke(Oe(V(A,(_i(),ap))))+u*f(g.b,17).a)),Ma(e.a,new Ke(m,z.e.b-u*f(g.a,17).a-l))),new Ya(rt(L),rt(F))}function LG(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;if(Yu(e.b,t)){if(xp(f(bn(e.b,t),50),i))return 1}else qr(e.b,t,new Mo);if(Yu(e.b,i)){if(xp(f(bn(e.b,i),50),t))return-1}else qr(e.b,i,new Mo);if(Yu(e.f,t)){if(xp(f(bn(e.f,t),50),i))return-1}else qr(e.f,t,new Mo);if(Yu(e.f,i)){if(xp(f(bn(e.b,i),50),t))return 1}else qr(e.f,i,new Mo);if(e.d==(Pp(),Rme)||!li(t,(Ge(),hi))||!li(i,(Ge(),hi))){for(L=null,T=new ne(t.j);T.a<T.c.c.length;)v=f(ee(T),11),v.e.c.length==0||f(ct(v.e,0),18).c.i.c.p==t.c.p-1&&(L=f(ct(v.e,0),18).c);for(z=null,S=new ne(i.j);S.a<S.c.c.length;)v=f(ee(S),11),v.e.c.length==0||f(ct(v.e,0),18).c.i.c.p==i.c.p-1&&(z=f(ct(v.e,0),18).c);if(L&&z){if(A=L.i,F=z.i,A&&A==F){for(J=new ne(A.j);J.a<J.c.c.length;){if(W=f(ee(J),11),W==L)return zd(e,i,t),-1;if(W==z)return zd(e,t,i),1}return u=nde(e,t),g=nde(e,i),u>g?(zd(e,t,i),1):(zd(e,i,t),-1)}for(pe=e.e,ve=0,Ee=pe.length;ve<Ee;++ve){if(ue=pe[ve],ue==A)return zd(e,i,t),-1;if(ue==F)return zd(e,t,i),1}}if(L&&!z||!L&&z){if(a=zkt(e,t,i),a!=0)return a>0?zd(e,t,i):zd(e,i,t),a;if(!li(t,(Ge(),hi))||!li(i,hi))return l=nde(e,t),m=nde(e,i),l>m?(zd(e,t,i),1):(zd(e,i,t),-1)}if(!L&&!z&&(a=zkt(e,t,i),a!=0))return a>0?zd(e,t,i):zd(e,i,t),a}return li(t,(Ge(),hi))&&li(i,hi)?(l=f(V(t,hi),17).a,m=f(V(i,hi),17).a,l>m?(zd(e,t,i),1):(zd(e,i,t),-1)):(zd(e,i,t),-1)}function I7n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F;if(g=!0,A=null,a=null,u=null,t=!1,F=eMt,S=null,l=null,m=0,v=Ile(e,m,L7e,D7e),v<e.length&&(gn(v,e.length),e.charCodeAt(v)==58)&&(A=(ga(m,v,e.length),e.substr(m,v-m)),m=v+1),i=A!=null&&Iz(Bne,A.toLowerCase()),i){if(v=e.lastIndexOf("!/"),v==-1)throw ie(new dn("no archive separator"));g=!0,a=uh(e,m,++v),m=v}else m>=0&&Mt(e.substr(m,2),"//")?(m+=2,v=Ile(e,m,zF,HF),a=(ga(m,v,e.length),e.substr(m,v-m)),m=v):A!=null&&(m==e.length||(gn(m,e.length),e.charCodeAt(m)!=47))&&(g=!1,v=i_e(e,gl(35),m),v==-1&&(v=e.length),a=(ga(m,v,e.length),e.substr(m,v-m)),m=v);if(!i&&m<e.length&&(gn(m,e.length),e.charCodeAt(m)==47)&&(v=Ile(e,m+1,zF,HF),T=(ga(m+1,v,e.length),e.substr(m+1,v-(m+1))),T.length>0&&_a(T,T.length-1)==58&&(u=T,m=v)),m<e.length&&(gn(m,e.length),e.charCodeAt(m)==47)&&(++m,t=!0),m<e.length&&(gn(m,e.length),e.charCodeAt(m)!=63)&&(gn(m,e.length),e.charCodeAt(m)!=35)){for(L=new et;m<e.length&&(gn(m,e.length),e.charCodeAt(m)!=63)&&(gn(m,e.length),e.charCodeAt(m)!=35);)v=Ile(e,m,zF,HF),Qe(L,(ga(m,v,e.length),e.substr(m,v-m))),m=v,m<e.length&&(gn(m,e.length),e.charCodeAt(m)==47)&&(wxn(e,++m)||L.c.push(""));F=Pe(ft,Ye,2,L.c.length,6,1),kg(L,F)}return m<e.length&&(gn(m,e.length),e.charCodeAt(m)==63)&&(v=fP(e,35,++m),v==-1&&(v=e.length),S=(ga(m,v,e.length),e.substr(m,v-m)),m=v),m<e.length&&(l=Gke(e,++m)),u8n(g,A,a,u,F,S),new Ode(g,A,a,u,t,F,S,l)}function Ukt(){Ukt=Z,ofe(),Ar=new CS,jt(Ar,(tt(),dp),Bg),jt(Ar,_h,Bg),jt(Ar,nd,Bg),jt(Ar,fp,Bg),jt(Ar,yl,Bg),jt(Ar,rd,Bg),jt(Ar,fp,dp),jt(Ar,Bg,Kf),jt(Ar,dp,Kf),jt(Ar,_h,Kf),jt(Ar,nd,Kf),jt(Ar,wl,Kf),jt(Ar,fp,Kf),jt(Ar,yl,Kf),jt(Ar,rd,Kf),jt(Ar,tl,Kf),jt(Ar,Bg,vf),jt(Ar,dp,vf),jt(Ar,Kf,vf),jt(Ar,_h,vf),jt(Ar,nd,vf),jt(Ar,wl,vf),jt(Ar,fp,vf),jt(Ar,tl,vf),jt(Ar,wf,vf),jt(Ar,yl,vf),jt(Ar,jl,vf),jt(Ar,rd,vf),jt(Ar,dp,_h),jt(Ar,nd,_h),jt(Ar,fp,_h),jt(Ar,rd,_h),jt(Ar,dp,nd),jt(Ar,_h,nd),jt(Ar,fp,nd),jt(Ar,nd,nd),jt(Ar,yl,nd),jt(Ar,Bg,Yf),jt(Ar,dp,Yf),jt(Ar,Kf,Yf),jt(Ar,vf,Yf),jt(Ar,_h,Yf),jt(Ar,nd,Yf),jt(Ar,wl,Yf),jt(Ar,fp,Yf),jt(Ar,wf,Yf),jt(Ar,tl,Yf),jt(Ar,rd,Yf),jt(Ar,yl,Yf),jt(Ar,Au,Yf),jt(Ar,Bg,wf),jt(Ar,dp,wf),jt(Ar,Kf,wf),jt(Ar,_h,wf),jt(Ar,nd,wf),jt(Ar,wl,wf),jt(Ar,fp,wf),jt(Ar,tl,wf),jt(Ar,rd,wf),jt(Ar,jl,wf),jt(Ar,Au,wf),jt(Ar,dp,tl),jt(Ar,_h,tl),jt(Ar,nd,tl),jt(Ar,fp,tl),jt(Ar,wf,tl),jt(Ar,rd,tl),jt(Ar,yl,tl),jt(Ar,Bg,vl),jt(Ar,dp,vl),jt(Ar,Kf,vl),jt(Ar,_h,vl),jt(Ar,nd,vl),jt(Ar,wl,vl),jt(Ar,fp,vl),jt(Ar,tl,vl),jt(Ar,rd,vl),jt(Ar,dp,yl),jt(Ar,Kf,yl),jt(Ar,vf,yl),jt(Ar,nd,yl),jt(Ar,Bg,jl),jt(Ar,dp,jl),jt(Ar,vf,jl),jt(Ar,_h,jl),jt(Ar,nd,jl),jt(Ar,wl,jl),jt(Ar,fp,jl),jt(Ar,fp,Au),jt(Ar,nd,Au),jt(Ar,tl,Bg),jt(Ar,tl,_h),jt(Ar,tl,Kf),jt(Ar,wl,Bg),jt(Ar,wl,dp),jt(Ar,wl,vf)}function O7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He;switch(i.Sg("Brandes & Koepf node placement",1),e.a=t,e.c=mDn(t),a=f(V(t,(ot(),bme)),279),F=lt(ut(V(t,tF))),e.d=a==(gG(),mte)&&!F||a==qge,o8n(e,t),Le=null,He=null,ue=null,pe=null,J=(uf(4,$T),new Du(4)),f(V(t,bme),279).g){case 3:ue=new nA(t,e.c.d,(vg(),$E),(fm(),G0)),vn(J.c,ue);break;case 1:pe=new nA(t,e.c.d,(vg(),Mg),(fm(),G0)),vn(J.c,pe);break;case 4:Le=new nA(t,e.c.d,(vg(),$E),(fm(),y_)),vn(J.c,Le);break;case 2:He=new nA(t,e.c.d,(vg(),Mg),(fm(),y_)),vn(J.c,He);break;default:ue=new nA(t,e.c.d,(vg(),$E),(fm(),G0)),pe=new nA(t,e.c.d,Mg,G0),Le=new nA(t,e.c.d,$E,y_),He=new nA(t,e.c.d,Mg,y_),vn(J.c,Le),vn(J.c,He),vn(J.c,ue),vn(J.c,pe)}for(u=new sat(t,e.c),m=new ne(J);m.a<m.c.c.length;)l=f(ee(m),185),H9n(u,l,e.b),GDn(l);for(L=new wwt(t,e.c),v=new ne(J);v.a<v.c.c.length;)l=f(ee(v),185),p8n(L,l);if(i.Zg())for(S=new ne(J);S.a<S.c.c.length;)l=f(ee(S),185),i._g(l+" size is "+_de(l));if(A=null,e.d&&(T=L9n(e,J,e.c.d),y_t(t,T,i)&&(A=T)),!A)for(S=new ne(J);S.a<S.c.c.length;)l=f(ee(S),185),y_t(t,l,i)&&(!A||_de(A)>_de(l))&&(A=l);for(!A&&(A=(Gt(0,J.c.length),f(J.c[0],185))),W=new ne(t.b);W.a<W.c.c.length;)for(z=f(ee(W),25),Ee=new ne(z.a);Ee.a<Ee.c.c.length;)ve=f(ee(Ee),9),ve.n.b=ke(A.p[ve.p])+ke(A.d[ve.p]);for(i.Zg()&&(i._g("Chosen node placement: "+A),i._g("Blocks: "+LEt(A)),i._g("Classes: "+z4n(A,i)),i._g("Marked edges: "+e.b)),g=new ne(J);g.a<g.c.c.length;)l=f(ee(g),185),l.g=null,l.b=null,l.a=null,l.d=null,l.j=null,l.i=null,l.p=null;H2n(e.c),e.b.a.$b(),i.Tg()}function zkt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;if(t.k==(cn(),uo)&&i.k==Yi)return l=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(t.j,16)),new Lh))),11).e,0),18).c,u=l.i,m=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(t.j,16)),new wp))),11).g,0),18).d,g=m.i,a=t.c.p,u.c.p!=a&&g.c.p!=a?0:u==i||g==i?(zd(e,t,i),1):LG(e,u,i);if(t.k==Yi&&i.k==uo)return l=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(i.j,16)),new Lh))),11).e,0),18).c,u=l.i,m=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(i.j,16)),new wp))),11).g,0),18).d,g=m.i,a=t.c.p,u.c.p!=a&&g.c.p!=a?0:u==t||g==t?(zd(e,i,t),-1):LG(e,t,u);if(t.k==uo&&i.k==uo){if(L=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(t.j,16)),new Lh))),11).e,0),18).c,F=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(t.j,16)),new wp))),11).g,0),18).d,z=L.i,W=F.i,v=t.c.p,T=!1,A=!1,Ee=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(i.j,16)),new Lh))),11).e,0),18).c,Le=f(ct(f(zN(S1(Jr(new Ft(null,new Bt(i.j,16)),new wp))),11).g,0),18).d,He=Ee.i,nt=Le.i,J=i.c.p,pe=!1,ve=!1,S=t,ue=i,z.c.p==v?(T=!0,S=z):W.c.p==v&&(A=!0,S=W),He.c.p==J?(pe=!0,ue=He):nt.c.p==J&&(ve=!0,ue=nt),S==ue)if(e.a){if(T&&pe)return Kt=Gkt(new ITe(e.e,e.d,null,ve),L,Ee),Kt>0?(zd(e,i,t),1):(zd(e,t,i),-1);if(T&&ve)return zd(e,i,t),1;if(A&&pe)return zd(e,t,i),-1;if(A&&ve)return 0}else for(Ct=new ne(S.j);Ct.a<Ct.c.c.length;){if(xt=f(ee(Ct),11),L==xt)return zd(e,i,t),-1;if(Ee==xt)return zd(e,t,i),1}return LG(e,S,ue)}else return 0}function N7n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In;if(e.c.length==1)return $vt((Gt(0,e.c.length),f(e.c[0],120))),Gt(0,e.c.length),f(e.c[0],120);if(e.c.length<=0)return new xH;for(v=new ne(e);v.a<v.c.c.length;){for(g=f(ee(v),120),pe=0,z=Qr,W=Qr,L=Ra,F=Ra,ue=tr(g.b,0);ue.b!=ue.d.c;)J=f(Zn(ue),39),pe+=f(V(J,(Ec(),tne)),17).a,z=h.Math.min(z,J.e.a),W=h.Math.min(W,J.e.b),L=h.Math.max(L,J.e.a+J.f.a),F=h.Math.max(F,J.e.b+J.f.b);$e(g,(Ec(),tne),rt(pe)),$e(g,(_i(),TD),new Ke(z,W)),$e(g,qq,new Ke(L,F))}for(Ht(),na(e,new L8),Le=new xH,Rc(Le,(Gt(0,e.c.length),f(e.c[0],102))),A=0,Kt=0,S=new ne(e);S.a<S.c.c.length;)g=f(ee(S),120),He=Ho($a(f(V(g,(_i(),qq)),8)),f(V(g,TD),8)),A=h.Math.max(A,He.a),Kt+=He.a*He.b;for(A=h.Math.max(A,h.Math.sqrt(Kt)*ke(Oe(V(Le,(Ec(),P5t))))),nt=ke(Oe(V(Le,Qme))),tn=0,In=0,T=0,t=nt,m=new ne(e);m.a<m.c.c.length;)g=f(ee(m),120),He=Ho($a(f(V(g,(_i(),qq)),8)),f(V(g,TD),8)),tn+He.a>A&&(tn=0,In+=T+nt,T=0),LSt(Le,g,tn,In),t=h.Math.max(t,tn+He.a),T=h.Math.max(T,He.b),tn+=He.a+nt;for(Ee=new Wn,i=new Wn,Ct=new ne(e);Ct.a<Ct.c.c.length;)for(xt=f(ee(Ct),120),a=lt(ut(V(xt,(Rr(),nV)))),ve=xt.q?xt.q:Cb,l=ve.vc().Jc();l.Ob();)u=f(l.Pb(),43),Yu(Ee,u.jd())?Ae(f(u.jd(),144).Qg())!==Ae(u.kd())&&(a&&Yu(i,u.jd())?(b0(),""+f(u.jd(),144).Ng()):(qr(Ee,f(u.jd(),144),u.kd()),$e(Le,f(u.jd(),144),u.kd()),a&&qr(i,f(u.jd(),144),u.kd()))):(qr(Ee,f(u.jd(),144),u.kd()),$e(Le,f(u.jd(),144),u.kd()));return $vt(Le),Le}function vQ(e,t){switch(e.e){case 0:case 2:case 4:case 6:case 42:case 44:case 46:case 48:case 8:case 10:case 12:case 14:case 16:case 18:case 20:case 22:case 24:case 26:case 28:case 30:case 32:case 34:case 36:case 38:return new dft(e.b,e.a,t,e.c);case 1:return new aX(e.a,t,Pi(t.zh(),e.c));case 43:return new Xst(e.a,t,Pi(t.zh(),e.c));case 3:return new So(e.a,t,Pi(t.zh(),e.c));case 45:return new xl(e.a,t,Pi(t.zh(),e.c));case 41:return new Ol(f(ph(e.c),29),e.a,t,Pi(t.zh(),e.c));case 50:return new lbt(f(ph(e.c),29),e.a,t,Pi(t.zh(),e.c));case 5:return new M_e(e.a,t,Pi(t.zh(),e.c),e.d.n);case 47:return new fut(e.a,t,Pi(t.zh(),e.c),e.d.n);case 7:return new Be(e.a,t,Pi(t.zh(),e.c),e.d.n);case 49:return new qN(e.a,t,Pi(t.zh(),e.c),e.d.n);case 9:return new Yst(e.a,t,Pi(t.zh(),e.c));case 11:return new Kst(e.a,t,Pi(t.zh(),e.c));case 13:return new YSe(e.a,t,Pi(t.zh(),e.c));case 15:return new yX(e.a,t,Pi(t.zh(),e.c));case 17:return new Jst(e.a,t,Pi(t.zh(),e.c));case 19:return new I3(e.a,t,Pi(t.zh(),e.c));case 21:return new XSe(e.a,t,Pi(t.zh(),e.c));case 23:return new lH(e.a,t,Pi(t.zh(),e.c));case 25:return new gut(e.a,t,Pi(t.zh(),e.c),e.d.n);case 27:return new Qt(e.a,t,Pi(t.zh(),e.c),e.d.n);case 29:return new put(e.a,t,Pi(t.zh(),e.c),e.d.n);case 31:return new hut(e.a,t,Pi(t.zh(),e.c),e.d.n);case 33:return new $_e(e.a,t,Pi(t.zh(),e.c),e.d.n);case 35:return new j_e(e.a,t,Pi(t.zh(),e.c),e.d.n);case 37:return new Use(e.a,t,Pi(t.zh(),e.c),e.d.n);case 39:return new FX(e.a,t,Pi(t.zh(),e.c),e.d.n);case 40:return new eo(t,Pi(t.zh(),e.c));default:throw ie(new oc("Unknown feature style: "+e.e))}}function Hkt(e){var t,i,a,u,l,g,m,v;for(l=0,u=e.a.b,v=tr(e.a,0);v.b!=v.d.c;){if(m=f(Zn(v),239),g=(l+1)/(u+1),!e.c&&!e.d)return;e.c&&!e.d?(e.g=!0,e.b==(ko(),Ds)?(a=e.c.e.b+e.c.f.b+e.e*(l+1),t=new Ke(ke(Oe(V(e.c,(_i(),ap))))+e.e,a),i=new Ke(ke(Oe(V(e.c,Fp)))-e.e,a)):e.b==Gs?(a=e.c.e.b+e.c.f.b+e.e*(l+1),t=new Ke(ke(Oe(V(e.c,(_i(),Fp))))-e.e,a),i=new Ke(ke(Oe(V(e.c,ap)))+e.e,a)):e.b==Wf?(a=e.c.e.a+e.c.f.a+e.e*(l+1),t=new Ke(a,ke(Oe(V(e.c,(_i(),ap))))+e.e),i=new Ke(a,ke(Oe(V(e.c,Fp)))-e.e)):(a=e.c.e.a+e.c.f.a+e.e*(l+1),t=new Ke(a,ke(Oe(V(e.c,(_i(),Fp))))-e.e),i=new Ke(a,ke(Oe(V(e.c,ap)))+e.e))):e.c&&e.d?e.b==(ko(),Ds)?(a=e.d.e.b*g+(e.c.e.b+e.c.f.b)*(1-g),t=new Ke(ke(Oe(V(e.c,(_i(),ap))))+e.e,a),i=new Ke(ke(Oe(V(e.c,Fp)))-e.e,a)):e.b==Gs?(a=e.d.e.b*g+(e.c.e.b+e.c.f.b)*(1-g),t=new Ke(ke(Oe(V(e.c,(_i(),Fp))))-e.e,a),i=new Ke(ke(Oe(V(e.c,ap)))+e.e,a)):e.b==Wf?(a=e.d.e.a*g+(e.c.e.a+e.c.f.a)*(1-g),t=new Ke(a,ke(Oe(V(e.c,(_i(),ap))))+e.e),i=new Ke(a,ke(Oe(V(e.c,Fp)))-e.e)):(a=e.d.e.a*g+(e.c.e.a+e.c.f.a)*(1-g),t=new Ke(a,ke(Oe(V(e.c,(_i(),Fp))))-e.e),i=new Ke(a,ke(Oe(V(e.c,ap)))+e.e)):(e.f=!0,e.b==(ko(),Ds)?(a=e.d.e.b-e.e*(l+1),t=new Ke(ke(Oe(V(e.d,(_i(),ap))))+e.e,a),i=new Ke(ke(Oe(V(e.d,Fp)))-e.e,a)):e.b==Gs?(a=e.d.e.b-e.e*(l+1),t=new Ke(ke(Oe(V(e.d,(_i(),Fp))))-e.e,a),i=new Ke(ke(Oe(V(e.d,ap)))+e.e,a)):e.b==Wf?(a=e.d.e.a-e.e*(l+1),t=new Ke(a,ke(Oe(V(e.d,(_i(),ap))))+e.e),i=new Ke(a,ke(Oe(V(e.d,Fp)))-e.e)):(a=e.d.e.a-e.e*(l+1),t=new Ke(a,ke(Oe(V(e.d,(_i(),Fp))))-e.e),i=new Ke(a,ke(Oe(V(e.d,ap)))+e.e))),f(m.a,8).a=t.a,f(m.a,8).b=t.b,m.b.a=i.a,m.b.b=i.b,++l}}function R7n(e,t,i,a,u,l){var g,m,v,S,T,A,L,F,z,W,J,ue;switch(t){case 71:m=a.q.getFullYear()-hw>=-1900?1:0,i>=4?xr(e,ce(se(ft,1),Ye,2,6,[oTt,aTt])[m]):xr(e,ce(se(ft,1),Ye,2,6,["BC","AD"])[m]);break;case 121:R3n(e,i,a);break;case 77:yLn(e,i,a);break;case 107:v=u.q.getHours(),v==0?bb(e,24,i):bb(e,v,i);break;case 83:z6n(e,i,u);break;case 69:T=a.q.getDay(),i==5?xr(e,ce(se(ft,1),Ye,2,6,["S","M","T","W","T","F","S"])[T]):i==4?xr(e,ce(se(ft,1),Ye,2,6,[xfe,Cfe,Afe,Ife,Ofe,Nfe,Rfe])[T]):xr(e,ce(se(ft,1),Ye,2,6,["Sun","Mon","Tue","Wed","Thu","Fri","Sat"])[T]);break;case 97:u.q.getHours()>=12&&u.q.getHours()<24?xr(e,ce(se(ft,1),Ye,2,6,["AM","PM"])[1]):xr(e,ce(se(ft,1),Ye,2,6,["AM","PM"])[0]);break;case 104:A=u.q.getHours()%12,A==0?bb(e,12,i):bb(e,A,i);break;case 75:L=u.q.getHours()%12,bb(e,L,i);break;case 72:F=u.q.getHours(),bb(e,F,i);break;case 99:z=a.q.getDay(),i==5?xr(e,ce(se(ft,1),Ye,2,6,["S","M","T","W","T","F","S"])[z]):i==4?xr(e,ce(se(ft,1),Ye,2,6,[xfe,Cfe,Afe,Ife,Ofe,Nfe,Rfe])[z]):i==3?xr(e,ce(se(ft,1),Ye,2,6,["Sun","Mon","Tue","Wed","Thu","Fri","Sat"])[z]):bb(e,z,1);break;case 76:W=a.q.getMonth(),i==5?xr(e,ce(se(ft,1),Ye,2,6,["J","F","M","A","M","J","J","A","S","O","N","D"])[W]):i==4?xr(e,ce(se(ft,1),Ye,2,6,[gfe,mfe,bfe,vfe,AR,wfe,yfe,Efe,Sfe,_fe,kfe,Tfe])[W]):i==3?xr(e,ce(se(ft,1),Ye,2,6,["Jan","Feb","Mar","Apr",AR,"Jun","Jul","Aug","Sep","Oct","Nov","Dec"])[W]):bb(e,W+1,i);break;case 81:J=a.q.getMonth()/3|0,i<4?xr(e,ce(se(ft,1),Ye,2,6,["Q1","Q2","Q3","Q4"])[J]):xr(e,ce(se(ft,1),Ye,2,6,["1st quarter","2nd quarter","3rd quarter","4th quarter"])[J]);break;case 100:ue=a.q.getDate(),bb(e,ue,i);break;case 109:S=u.q.getMinutes(),bb(e,S,i);break;case 115:g=u.q.getSeconds(),bb(e,g,i);break;case 122:i<4?xr(e,l.c[0]):xr(e,l.c[1]);break;case 118:xr(e,l.b);break;case 90:i<3?xr(e,VRn(l)):i==3?xr(e,XRn(l)):xr(e,JRn(l.a));break;default:return!1}return!0}function TOe(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r;if(SSt(t),v=f(be((!t.b&&(t.b=new Qt(Yn,t,4,7)),t.b),0),84),T=f(be((!t.c&&(t.c=new Qt(Yn,t,5,8)),t.c),0),84),m=Us(v),S=Us(T),g=(!t.a&&(t.a=new Be(Ii,t,6,6)),t.a).i==0?null:f(be((!t.a&&(t.a=new Be(Ii,t,6,6)),t.a),0),168),nt=f(bn(e.a,m),9),tn=f(bn(e.a,S),9),xt=null,In=null,me(v,193)&&(He=f(bn(e.a,v),306),me(He,11)?xt=f(He,11):me(He,9)&&(nt=f(He,9),xt=f(ct(nt.j,0),11))),me(T,193)&&(Kt=f(bn(e.a,T),306),me(Kt,11)?In=f(Kt,11):me(Kt,9)&&(tn=f(Kt,9),In=f(ct(tn.j,0),11))),!nt||!tn)throw ie(new ON("The source or the target of edge "+t+" could not be found. This usually happens when an edge connects a node laid out by ELK Layered to a node in another level of hierarchy laid out by either another instance of ELK Layered or another layout algorithm alltogether. The former can be solved by setting the hierarchyHandling option to INCLUDE_CHILDREN."));for(W=new xS,Rc(W,t),$e(W,(Ge(),oi),t),$e(W,(ot(),Ls),null),F=f(V(a,Cu),22),nt==tn&&F.Ec((us(),W$)),xt||(Le=(cs(),Uu),Ct=null,g&&T3(f(V(nt,Xi),100))&&(Ct=new Ke(g.j,g.k),tpt(Ct,lT(t)),Ipt(Ct,i),bT(S,m)&&(Le=td,ii(Ct,nt.n))),xt=T_t(nt,Ct,Le,a)),In||(Le=(cs(),td),$r=null,g&&T3(f(V(tn,Xi),100))&&($r=new Ke(g.b,g.c),tpt($r,lT(t)),Ipt($r,i)),In=T_t(tn,$r,Le,ra(tn))),Ia(W,xt),ua(W,In),(xt.e.c.length>1||xt.g.c.length>1||In.e.c.length>1||In.g.c.length>1)&&F.Ec((us(),V$)),L=new Nn((!t.n&&(t.n=new Be(hc,t,1,7)),t.n));L.e!=L.i.gc();)if(A=f(Mn(L),153),!lt(ut(Fe(A,ME)))&&A.a)switch(J=cle(A),Qe(W.b,J),f(V(J,Rm),276).g){case 1:case 2:F.Ec((us(),fD));break;case 0:F.Ec((us(),dD)),$e(J,Rm,(Sg(),PD))}if(l=f(V(a,Q$),323),ue=f(V(a,jte),324),u=l==(hj(),Aq)||ue==(Nj(),Ome),g&&(!g.a&&(g.a=new So(yf,g,5)),g.a).i!=0&&u){for(pe=WNn(g),z=new fd,Ee=tr(pe,0);Ee.b!=Ee.d.c;)ve=f(Zn(Ee),8),Mr(z,new ja(ve));$e(W,UPe,z)}return W}function P7n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji;for(Ct=0,Kt=0,nt=new Wn,Le=f(Dd(oT(Lu(new Ft(null,new Bt(e.b,16)),new vk),new DO)),17).a+1,xt=Pe(pr,Pr,30,Le,15,1),J=Pe(pr,Pr,30,Le,15,1),W=0;W<Le;W++)xt[W]=0,J[W]=0;for(v=f($d(mTe(new Ft(null,new Bt(e.a,16))),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),T=v.Jc();T.Ob();)if(S=f(T.Pb(),65),In=f(V(S.b,(Ec(),Lm)),17).a,ji=f(V(S.c,Lm),17).a,Ee=ji-In,Ee>1)for(m=In+1;m<ji;m++){if(A=m,He=f($d(Jr(new Ft(null,new Bt(e.b,16)),new $tt(A)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[Zl]))),15),z=0,t==(ko(),Ds)||t==Gs){for(He.gd(new Zg),z=0;z<He.gc()&&(ue=(m-In)/(ji-In),!(f(He.Xb(z),39).e.b>S.b.e.b*(1-ue)+S.c.e.b*ue));z++);if(He.gc()>0&&($r=S.a.b==0?$a(S.b.e):f(Hh(S.a),8),ve=ii($a(f(He.Xb(He.gc()-1),39).e),f(He.Xb(He.gc()-1),39).f),L=ii($a(f(He.Xb(0),39).e),f(He.Xb(0),39).f),z>=He.gc()-1&&$r.b>ve.b&&S.c.e.b>ve.b||z<=0&&$r.b<L.a&&S.c.e.b<L.b))continue}else{for(He.gd(new sg),z=0;z<He.gc()&&(ue=(m-In)/(ji-In),!(f(He.Xb(z),39).e.a>S.b.e.a*(1-ue)+S.c.e.a*ue));z++);if(He.gc()>0&&($r=S.a.b==0?$a(S.b.e):f(Hh(S.a),8),ve=ii($a(f(He.Xb(He.gc()-1),39).e),f(He.Xb(He.gc()-1),39).f),L=ii($a(f(He.Xb(0),39).e),f(He.Xb(0),39).f),z>=He.gc()-1&&$r.a>ve.a&&S.c.e.a>ve.a||z<=0&&$r.a<L.a&&S.c.e.a<L.a))continue}u=new ha,l=new ha,Mr(S.a,u),Mr(S.a,l),g=new Yse(u,l,S),pe=hm(hb(m,32),ia(z,ls)),Yu(nt,TT(pe))?(F=f(bn(nt,TT(pe)),660),Mr(F.a,g),l1(F.b)?sE(F.a,new PC):sE(F.a,new Z6),Hkt(F)):(F=new Zvt(z==0?null:f(He.Xb(z-1),39),z==He.gc()?null:f(He.Xb(z),39),g,e),qr(nt,TT(pe),F)),t==Ds||t==Gs?(F.f&&F.d.e.b<=ke(Oe(V(e,(_i(),Yme))))&&++Ct,F.g&&F.c.e.b+F.c.f.b>=ke(Oe(V(e,(_i(),sMe))))&&++Kt):(F.f&&F.d.e.a<=ke(Oe(V(e,(_i(),Kme))))&&++Ct,F.g&&F.c.e.a+F.c.f.a>=ke(Oe(V(e,(_i(),aMe))))&&++Kt)}else Ee==0?zAe(S):Ee<0&&(++xt[In],++J[ji],tn=A7n(S,t,e,new Ya(rt(Ct),rt(Kt)),i,a,new Ya(rt(J[ji]),rt(xt[In]))),Ct=f(tn.a,17).a,Kt=f(tn.b,17).a)}function L7n(e){e.gb||(e.gb=!0,e.b=yc(e,0),Ki(e.b,18),Ci(e.b,19),e.a=yc(e,1),Ki(e.a,1),Ci(e.a,2),Ci(e.a,3),Ci(e.a,4),Ci(e.a,5),e.o=yc(e,2),Ki(e.o,8),Ki(e.o,9),Ci(e.o,10),Ci(e.o,11),Ci(e.o,12),Ci(e.o,13),Ci(e.o,14),Ci(e.o,15),Ci(e.o,16),Ci(e.o,17),Ci(e.o,18),Ci(e.o,19),Ci(e.o,20),Ci(e.o,21),Ci(e.o,22),Ci(e.o,23),Rs(e.o),Rs(e.o),Rs(e.o),Rs(e.o),Rs(e.o),Rs(e.o),Rs(e.o),Rs(e.o),Rs(e.o),Rs(e.o),e.p=yc(e,3),Ki(e.p,2),Ki(e.p,3),Ki(e.p,4),Ki(e.p,5),Ci(e.p,6),Ci(e.p,7),Rs(e.p),Rs(e.p),e.q=yc(e,4),Ki(e.q,8),e.v=yc(e,5),Ci(e.v,9),Rs(e.v),Rs(e.v),Rs(e.v),e.w=yc(e,6),Ki(e.w,2),Ki(e.w,3),Ki(e.w,4),Ci(e.w,5),e.B=yc(e,7),Ci(e.B,1),Rs(e.B),Rs(e.B),Rs(e.B),e.Q=yc(e,8),Ci(e.Q,0),Rs(e.Q),e.R=yc(e,9),Ki(e.R,1),e.S=yc(e,10),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),Rs(e.S),e.T=yc(e,11),Ci(e.T,10),Ci(e.T,11),Ci(e.T,12),Ci(e.T,13),Ci(e.T,14),Rs(e.T),Rs(e.T),e.U=yc(e,12),Ki(e.U,2),Ki(e.U,3),Ci(e.U,4),Ci(e.U,5),Ci(e.U,6),Ci(e.U,7),Rs(e.U),e.V=yc(e,13),Ci(e.V,10),e.W=yc(e,14),Ki(e.W,18),Ki(e.W,19),Ki(e.W,20),Ci(e.W,21),Ci(e.W,22),Ci(e.W,23),e.bb=yc(e,15),Ki(e.bb,10),Ki(e.bb,11),Ki(e.bb,12),Ki(e.bb,13),Ki(e.bb,14),Ki(e.bb,15),Ki(e.bb,16),Ci(e.bb,17),Rs(e.bb),Rs(e.bb),e.eb=yc(e,16),Ki(e.eb,2),Ki(e.eb,3),Ki(e.eb,4),Ki(e.eb,5),Ki(e.eb,6),Ki(e.eb,7),Ci(e.eb,8),Ci(e.eb,9),e.ab=yc(e,17),Ki(e.ab,0),Ki(e.ab,1),e.H=yc(e,18),Ci(e.H,0),Ci(e.H,1),Ci(e.H,2),Ci(e.H,3),Ci(e.H,4),Ci(e.H,5),Rs(e.H),e.db=yc(e,19),Ci(e.db,2),e.c=Br(e,20),e.d=Br(e,21),e.e=Br(e,22),e.f=Br(e,23),e.i=Br(e,24),e.g=Br(e,25),e.j=Br(e,26),e.k=Br(e,27),e.n=Br(e,28),e.r=Br(e,29),e.s=Br(e,30),e.t=Br(e,31),e.u=Br(e,32),e.fb=Br(e,33),e.A=Br(e,34),e.C=Br(e,35),e.D=Br(e,36),e.F=Br(e,37),e.G=Br(e,38),e.I=Br(e,39),e.J=Br(e,40),e.L=Br(e,41),e.M=Br(e,42),e.N=Br(e,43),e.O=Br(e,44),e.P=Br(e,45),e.X=Br(e,46),e.Y=Br(e,47),e.Z=Br(e,48),e.$=Br(e,49),e._=Br(e,50),e.cb=Br(e,51),e.K=Br(e,52))}function D7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In;for(g=new ki,He=f(V(i,(ot(),mf)),86),W=0,Ba(g,(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));g.b!=0;)T=f(g.b==0?null:(Ln(g.b!=0),lh(g,g.a.a)),27),S=so(T),(Ae(Fe(S,ip))!==Ae((Pp(),Pm))||Ae(Fe(S,RE))===Ae((R0(),uD))||Ae(Fe(S,RE))===Ae((R0(),cD))||lt(ut(Fe(S,h_)))||Ae(Fe(S,nx))!==Ae((hE(),XT))||Ae(Fe(S,xw))===Ae((gh(),w_))||Ae(Fe(S,xw))===Ae((gh(),ux))||Ae(Fe(S,LE))===Ae((D0(),SD))||Ae(Fe(S,LE))===Ae((D0(),_D)))&&!lt(ut(Fe(T,Ite)))&&pi(T,(Ge(),hi),rt(W++)),ue=!lt(ut(Fe(T,ME))),ue&&(L=(!T.a&&(T.a=new Be(Ur,T,10,11)),T.a).i!=0,z=BOn(T),F=Ae(Fe(T,rx))===Ae((_1(),W0)),In=!mm(T,(Rr(),AD))||Lpt(dr(Fe(T,AD))),Ee=null,In&&F&&(L||z)&&(Ee=s_t(T),$e(Ee,mf,He),li(Ee,Mq)&&Xrt(new XCe(ke(Oe(V(Ee,Mq)))),Ee),f(Fe(T,DE),181).gc()!=0&&(A=Ee,Zi(new Ft(null,(!T.c&&(T.c=new Be(_d,T,9,9)),new Bt(T.c,16))),new net(A)),X2t(T,Ee))),nt=i,xt=f(bn(e.a,so(T)),9),xt&&(nt=xt.e),ve=Ikt(e,T,nt),Ee&&(ve.e=Ee,Ee.e=ve,Ba(g,(!T.a&&(T.a=new Be(Ur,T,10,11)),T.a))));for(W=0,Wi(g,t,g.c.b,g.c);g.b!=0;){for(l=f(g.b==0?null:(Ln(g.b!=0),lh(g,g.a.a)),27),v=new Nn((!l.b&&(l.b=new Be(vo,l,12,3)),l.b));v.e!=v.i.gc();)m=f(Mn(v),85),SSt(m),(Ae(Fe(t,ip))!==Ae((Pp(),Pm))||Ae(Fe(t,RE))===Ae((R0(),uD))||Ae(Fe(t,RE))===Ae((R0(),cD))||lt(ut(Fe(t,h_)))||Ae(Fe(t,nx))!==Ae((hE(),XT))||Ae(Fe(t,xw))===Ae((gh(),w_))||Ae(Fe(t,xw))===Ae((gh(),ux))||Ae(Fe(t,LE))===Ae((D0(),SD))||Ae(Fe(t,LE))===Ae((D0(),_D)))&&pi(m,(Ge(),hi),rt(W++)),Kt=Us(f(be((!m.b&&(m.b=new Qt(Yn,m,4,7)),m.b),0),84)),tn=Us(f(be((!m.c&&(m.c=new Qt(Yn,m,5,8)),m.c),0),84)),!(lt(ut(Fe(m,ME)))||lt(ut(Fe(Kt,ME)))||lt(ut(Fe(tn,ME))))&&(J=HS(m)&&lt(ut(Fe(Kt,ix)))&&lt(ut(Fe(m,PE))),Le=l,J||bT(tn,Kt)?Le=Kt:bT(Kt,tn)&&(Le=tn),nt=i,xt=f(bn(e.a,Le),9),xt&&(nt=xt.e),pe=TOe(e,m,Le,nt),$e(pe,(Ge(),jPe),V5n(e,m,t,i)));if(F=Ae(Fe(l,rx))===Ae((_1(),W0)),F)for(u=new Nn((!l.a&&(l.a=new Be(Ur,l,10,11)),l.a));u.e!=u.i.gc();)a=f(Mn(u),27),In=!mm(a,(Rr(),AD))||Lpt(dr(Fe(a,AD))),Ct=Ae(Fe(a,rx))===Ae(W0),In&&Ct&&Wi(g,a,g.c.b,g.c)}}function Ge(){Ge=Z;var e,t;oi=new Si(eNe),jPe=new Si("coordinateOrigin"),nme=new Si("processors"),MPe=new Mi("compoundNode",(mn(),!1)),Rq=new Mi("insideConnections",!1),UPe=new Si("originalBendpoints"),zPe=new Si("originalDummyNodePosition"),HPe=new Si("originalLabelEdge"),Y$=new Si("representedLabels"),K$=new Si("endLabels"),YR=new Si("endLabel.origin"),JR=new Mi("labelSide",(df(),dV)),TA=new Mi("maxEdgeThickness",0),U0=new Mi("reversed",!1),ZR=new Si(HTt),$p=new Mi("longEdgeSource",null),yh=new Mi("longEdgeTarget",null),ex=new Mi("longEdgeHasLabelDummies",!1),Pq=new Mi("longEdgeBeforeLabelDummy",!1),Ste=new Mi("edgeConstraint",(lE(),Fge)),l_=new Si("inLayerLayoutUnit"),OE=new Mi("inLayerConstraint",(w1(),Oq)),XR=new Mi("inLayerSuccessorConstraint",new et),BPe=new Mi("inLayerSuccessorConstraintBetweenNonDummies",!1),ed=new Si("portDummy"),Ete=new Mi("crossingHint",rt(0)),Cu=new Mi("graphProperties",(t=f(Tp(Vge),10),new Rf(t,f(Gh(t,t.length),10),0))),Tc=new Mi("externalPortSide",(tt(),fc)),FPe=new Mi("externalPortSize",new ha),Jge=new Si("externalPortReplacedDummies"),_te=new Si("externalPortReplacedDummy"),R1=new Mi("externalPortConnections",(e=f(Tp(Ja),10),new Rf(e,f(Gh(e,e.length),10),0))),d_=new Mi($Tt,0),DPe=new Si("barycenterAssociates"),QR=new Si("TopSideComments"),KR=new Si("BottomSideComments"),yte=new Si("CommentConnectionPort"),Qge=new Mi("inputCollect",!1),tme=new Mi("outputCollect",!1),Nq=new Mi("cyclic",!1),$Pe=new Si("crossHierarchyMap"),ime=new Si("targetOffset"),new Mi("splineLabelSize",new ha),CA=new Si("spacings"),kte=new Mi("partitionConstraint",!1),u_=new Si("breakingPoint.info"),VPe=new Si("splines.survivingEdge"),NE=new Si("splines.route.start"),AA=new Si("splines.edgeChain"),qPe=new Si("originalPortConstraints"),f_=new Si("selfLoopHolder"),pD=new Si("splines.nsPortY"),hi=new Si("modelOrder"),eme=new Si("longEdgeTargetNode"),kw=new Mi(gxt,!1),xA=new Mi(gxt,!1),Zge=new Si("layerConstraints.hiddenNodes"),GPe=new Si("layerConstraints.opposidePort"),rme=new Si("targetNode.modelOrder")}function Gkt(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le;if(l=t,A=i,Yu(e.a,l)){if(xp(f(bn(e.a,l),50),A))return 1}else qr(e.a,l,new Mo);if(Yu(e.a,A)){if(xp(f(bn(e.a,A),50),l))return-1}else qr(e.a,A,new Mo);if(Yu(e.d,l)){if(xp(f(bn(e.d,l),50),A))return-1}else qr(e.d,l,new Mo);if(Yu(e.d,A)){if(xp(f(bn(e.a,A),50),l))return 1}else qr(e.d,A,new Mo);if(l.j!=A.j)return Ee=Mgn(l.j,A.j),Ee>0?Bf(e,l,A,1):Bf(e,A,l,1),Ee;if(Le=1,l.e.c.length!=0&&A.e.c.length!=0){if((l.j==(tt(),pn)&&A.j==pn||l.j==hn&&A.j==hn||l.j==Bn&&A.j==Bn)&&(Le=-Le),S=f(ct(l.e,0),18).c,W=f(ct(A.e,0),18).c,m=S.i,F=W.i,m==F)for(pe=new ne(m.j);pe.a<pe.c.c.length;){if(ue=f(ee(pe),11),S==ue)return Bf(e,A,l,Le),-Le;if(W==ue)return Bf(e,l,A,Le),Le}if(S.i.k==(cn(),uo)&&W.i.k==uo&&m.c.p==F.c.p&&m.c.p==l.i.c.p&&(ve=m.c,u=d0t(ve,m,F),u!=0))return l.j==_n&&A.j==_n&&(Le=-Le),u>0?(Bf(e,l,A,Le),Le):(Bf(e,A,l,Le),-Le);if(a=d0t(f($d(pce(e.c),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),20),m,F),a!=0)return a>0?(Bf(e,l,A,Le),Le):(Bf(e,A,l,Le),-Le);if(e.b&&(Ee=Q1t(l,A),Ee!=0))return Ee>0?(Bf(e,l,A,Le),Le):(Bf(e,A,l,Le),-Le)}return l.g.c.length!=0&&A.g.c.length!=0?((l.j==(tt(),pn)&&A.j==pn||l.j==Bn&&A.j==Bn)&&(Le=-Le),T=f(V(l,(Ge(),eme)),9),J=f(V(A,eme),9),e.e==(Pp(),Pme)&&T&&J&&li(T,hi)&&li(J,hi)?(g=f(V(T,hi),17).a,L=f(V(J,hi),17).a,g>L?(Bf(e,l,A,Le),Le):(Bf(e,A,l,Le),-Le)):e.b&&(Ee=Q1t(l,A),Ee!=0)?Ee>0?(Bf(e,l,A,Le),Le):(Bf(e,A,l,Le),-Le):(v=0,z=0,li(f(ct(l.g,0),18),hi)&&(v=f(V(f(ct(l.g,0),18),hi),17).a),li(f(ct(A.g,0),18),hi)&&(z=f(V(f(ct(A.g,0),18),hi),17).a),T&&T==J||e.f&&(e.f._b(T)&&(v=f(e.f.xc(T),17).a),e.f._b(J)&&(z=f(e.f.xc(J),17).a)),v>z?(Bf(e,l,A,Le),Le):(Bf(e,A,l,Le),-Le))):l.e.c.length!=0&&A.g.c.length!=0?(Bf(e,l,A,Le),1):l.g.c.length!=0&&A.e.c.length!=0?(Bf(e,A,l,Le),-1):li(l,(Ge(),hi))&&li(A,hi)?(g=f(V(l,hi),17).a,L=f(V(A,hi),17).a,(l.j==(tt(),pn)&&A.j==pn||l.j==Bn&&A.j==Bn)&&(Le=-Le),g>L?(Bf(e,l,A,Le),Le):(Bf(e,A,l,Le),-Le)):(Bf(e,A,l,Le),-Le)}function M7n(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z;for(A=tr(e.b,0);A.b!=A.d.c;)if(T=f(Zn(A),39),!Mt(T.c,QQ))for(l=f($d(new Ft(null,new Bt(d4n(T,e),16)),sf(new Eo,new lo,new Yo,ce(se(Ql,1),Ve,131,0,[(fh(),Zl)]))),15),t==(ko(),Ds)||t==Gs?l.gd(new NC):l.gd(new F8),z=l.gc(),u=0;u<z;u++)g=z==1?.5:(1+u)/(z+1),t==Ds?(S=ke(Oe(V(T,(_i(),ap)))),T.e.a+T.f.a+a<S?Ma(f(l.Xb(u),65).a,new Ke(S+i,T.e.b+T.f.b*g)):f(l.Xb(u),65).a.b>0&&(m=f(Hh(f(l.Xb(u),65).a),8).a,L=T.e.a+T.f.a/2,v=f(Hh(f(l.Xb(u),65).a),8).b,F=T.e.b+T.f.b/2,a>0&&h.Math.abs(v-F)/(h.Math.abs(m-L)/40)>50&&(F>v?Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a+a/5.3,T.e.b+T.f.b*g-a/2)):Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a+a/5.3,T.e.b+T.f.b*g+a/2)))),Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a,T.e.b+T.f.b*g))):t==Gs?(S=ke(Oe(V(T,(_i(),Fp)))),T.e.a-a>S?Ma(f(l.Xb(u),65).a,new Ke(S-i,T.e.b+T.f.b*g)):f(l.Xb(u),65).a.b>0&&(m=f(Hh(f(l.Xb(u),65).a),8).a,L=T.e.a+T.f.a/2,v=f(Hh(f(l.Xb(u),65).a),8).b,F=T.e.b+T.f.b/2,a>0&&h.Math.abs(v-F)/(h.Math.abs(m-L)/40)>50&&(F>v?Ma(f(l.Xb(u),65).a,new Ke(T.e.a-a/5.3,T.e.b+T.f.b*g-a/2)):Ma(f(l.Xb(u),65).a,new Ke(T.e.a-a/5.3,T.e.b+T.f.b*g+a/2)))),Ma(f(l.Xb(u),65).a,new Ke(T.e.a,T.e.b+T.f.b*g))):t==Wf?(S=ke(Oe(V(T,(_i(),ap)))),T.e.b+T.f.b+a<S?Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g,S+i)):f(l.Xb(u),65).a.b>0&&(m=f(Hh(f(l.Xb(u),65).a),8).a,L=T.e.a+T.f.a/2,v=f(Hh(f(l.Xb(u),65).a),8).b,F=T.e.b+T.f.b/2,a>0&&h.Math.abs(m-L)/(h.Math.abs(v-F)/40)>50&&(L>m?Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g-a/2,T.e.b+a/5.3+T.f.b)):Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g+a/2,T.e.b+a/5.3+T.f.b)))),Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g,T.e.b+T.f.b))):(S=ke(Oe(V(T,(_i(),Fp)))),x1t(f(l.Xb(u),65),e)?Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g,f(Hh(f(l.Xb(u),65).a),8).b)):T.e.b-a>S?Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g,S-i)):f(l.Xb(u),65).a.b>0&&(m=f(Hh(f(l.Xb(u),65).a),8).a,L=T.e.a+T.f.a/2,v=f(Hh(f(l.Xb(u),65).a),8).b,F=T.e.b+T.f.b/2,a>0&&h.Math.abs(m-L)/(h.Math.abs(v-F)/40)>50&&(L>m?Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g-a/2,T.e.b-a/5.3)):Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g+a/2,T.e.b-a/5.3)))),Ma(f(l.Xb(u),65).a,new Ke(T.e.a+T.f.a*g,T.e.b)))}function Rr(){Rr=Z;var e,t;AD=new Si(nCt),VA=new Si(rCt),o9e=(vb(),Bbe),ULt=new Pt(lRe,o9e),ID=new Pt(ML,null),zLt=new Si(T4e),s9e=(mE(),xi(Hbe,ce(se(Gbe,1),Ve,296,0,[zbe]))),tV=new Pt(KQ,s9e),nV=new Pt(aq,(mn(),!1)),c9e=(ko(),$g),BE=new Pt(Nhe,c9e),d9e=(k1(),r1e),l9e=new Pt(oq,d9e),VLt=new Pt(_4e,!1),h9e=(_1(),kne),UA=new Pt(WQ,h9e),k9e=new gS(12),V0=new Pt(zT,k9e),iV=new Pt(l$,!1),Kbe=new Pt(XQ,!1),oV=new Pt(d$,!1),I9e=(oa(),Pw),OF=new Pt(Yfe,I9e),o4=new Si(YQ),aV=new Si(XG),t1e=new Si(AQ),n1e=new Si(u$),m9e=new fd,zA=new Pt(ERe,m9e),GLt=new Pt(TRe,!1),WLt=new Pt(xRe,!1),new Pt(iCt,0),b9e=new i7,OD=new Pt(ARe,b9e),yne=new Pt(cRe,!1),QLt=new Pt(oCt,1),BA=new Si(aCt),FA=new Si(sCt),RD=new Pt(JG,!1),new Pt(cCt,!0),rt(0),new Pt(uCt,rt(100)),new Pt(lCt,!1),rt(0),new Pt(dCt,rt(4e3)),rt(0),new Pt(fCt,rt(400)),new Pt(hCt,!1),new Pt(pCt,!1),new Pt(gCt,!0),new Pt(mCt,!1),a9e=(pZ(),c1e),HLt=new Pt(k4e,a9e),g9e=(lj(),pV),YLt=new Pt(bCt,g9e),p9e=(XP(),cV),KLt=new Pt(vCt,p9e),eDt=new Pt(XNe,10),tDt=new Pt(JNe,10),nDt=new Pt(ZNe,20),rDt=new Pt(QNe,10),P9e=new Pt(Kfe,2),L9e=new Pt(Ohe,10),D9e=new Pt(eRe,0),Ene=new Pt(rRe,5),M9e=new Pt(tRe,1),j9e=new Pt(nRe,1),zE=new Pt(UT,20),iDt=new Pt(iRe,10),B9e=new Pt(oRe,10),a4=new Si(aRe),F9e=new wst,$9e=new Pt(IRe,F9e),JLt=new Si(Phe),T9e=!1,XLt=new Pt(Rhe,T9e),w9e=new gS(5),v9e=new Pt(hRe,w9e),y9e=(NT(),t=f(Tp(ps),10),new Rf(t,f(Gh(t,t.length),10),0)),HA=new Pt($L,y9e),C9e=(K3(),Rw),x9e=new Pt(mRe,C9e),Xbe=new Si(bRe),Jbe=new Si(vRe),Zbe=new Si(wRe),Ybe=new Si(yRe),E9e=(e=f(Tp($F),10),new Rf(e,f(Gh(e,e.length),10),0)),UE=new Pt(uA,E9e),_9e=At((yd(),jD)),Ow=new Pt(LR,_9e),S9e=new Ke(0,0),GA=new Pt(DR,S9e),hx=new Pt(jL,!1),u9e=(Sg(),PD),Vbe=new Pt(_Re,u9e),qbe=new Pt(IQ,!1),rt(1),new Pt(wCt,null),A9e=new Si(CRe),Qbe=new Si(kRe),R9e=(tt(),fc),qA=new Pt(uRe,R9e),Sd=new Si(sRe),O9e=(Xl(),At(Lw)),px=new Pt(FL,O9e),e1e=new Pt(pRe,!1),N9e=new Pt(gRe,!0),Sne=new Pt(ZG,1),U9e=new Pt(x4e,null),ND=new Pt(QG,150),sV=new Pt(eq,1.414),s4=new Pt(ZS,null),oDt=new Pt(C4e,1),rV=new Pt(dRe,!1),Wbe=new Pt(fRe,!1),qLt=new Pt(SRe,1),f9e=(HZ(),o1e),new Pt(yCt,f9e),ZLt=!0,sDt=(gR(),bx),cDt=bx,aDt=bx}function ka(){ka=Z,L5e=new ho("DIRECTION_PREPROCESSOR",0),N5e=new ho("COMMENT_PREPROCESSOR",1),vA=new ho("EDGE_AND_LAYER_CONSTRAINT_EDGE_REVERSER",2),Ege=new ho("INTERACTIVE_EXTERNAL_PORT_POSITIONER",3),Q5e=new ho("PARTITION_PREPROCESSOR",4),Kee=new ho("LABEL_DUMMY_INSERTER",5),ite=new ho("SELF_LOOP_PREPROCESSOR",6),ZT=new ho("LAYER_CONSTRAINT_PREPROCESSOR",7),J5e=new ho("PARTITION_MIDPROCESSOR",8),z5e=new ho("HIGH_DEGREE_NODE_LAYER_PROCESSOR",9),Y5e=new ho("NODE_PROMOTION",10),JT=new ho("LAYER_CONSTRAINT_POSTPROCESSOR",11),Z5e=new ho("PARTITION_POSTPROCESSOR",12),F5e=new ho("HIERARCHICAL_PORT_CONSTRAINT_PROCESSOR",13),ePe=new ho("SEMI_INTERACTIVE_CROSSMIN_PROCESSOR",14),T5e=new ho("BREAKING_POINT_INSERTER",15),Zee=new ho("LONG_EDGE_SPLITTER",16),Sge=new ho("PORT_SIDE_PROCESSOR",17),Vee=new ho("INVERTED_PORT_PROCESSOR",18),tte=new ho("PORT_LIST_SORTER",19),nPe=new ho("SORT_BY_INPUT_ORDER_OF_MODEL",20),ete=new ho("NORTH_SOUTH_PORT_PREPROCESSOR",21),x5e=new ho("BREAKING_POINT_PROCESSOR",22),X5e=new ho(sxt,23),rPe=new ho(cxt,24),nte=new ho("SELF_LOOP_PORT_RESTORER",25),k5e=new ho("ALTERNATING_LAYER_UNZIPPER",26),tPe=new ho("SINGLE_EDGE_GRAPH_WRAPPER",27),Wee=new ho("IN_LAYER_CONSTRAINT_PROCESSOR",28),M5e=new ho("END_NODE_PORT_LABEL_MANAGEMENT_PROCESSOR",29),W5e=new ho("LABEL_AND_NODE_SIZE_PROCESSOR",30),V5e=new ho("INNERMOST_NODE_MARGIN_CALCULATOR",31),ote=new ho("SELF_LOOP_ROUTER",32),I5e=new ho("COMMENT_NODE_MARGIN_CALCULATOR",33),qee=new ho("END_LABEL_PREPROCESSOR",34),Xee=new ho("LABEL_DUMMY_SWITCHER",35),A5e=new ho("CENTER_LABEL_MANAGEMENT_PROCESSOR",36),nD=new ho("LABEL_SIDE_SELECTOR",37),G5e=new ho("HYPEREDGE_DUMMY_MERGER",38),B5e=new ho("HIERARCHICAL_PORT_DUMMY_SIZE_PROCESSOR",39),K5e=new ho("LAYER_SIZE_AND_GRAPH_HEIGHT_CALCULATOR",40),z$=new ho("HIERARCHICAL_PORT_POSITION_PROCESSOR",41),R5e=new ho("CONSTRAINTS_POSTPROCESSOR",42),O5e=new ho("COMMENT_POSTPROCESSOR",43),q5e=new ho("HYPERNODE_PROCESSOR",44),U5e=new ho("HIERARCHICAL_PORT_ORTHOGONAL_EDGE_ROUTER",45),Jee=new ho("LONG_EDGE_JOINER",46),rte=new ho("SELF_LOOP_POSTPROCESSOR",47),C5e=new ho("BREAKING_POINT_REMOVER",48),Qee=new ho("NORTH_SOUTH_PORT_POSTPROCESSOR",49),H5e=new ho("HORIZONTAL_COMPACTOR",50),Yee=new ho("LABEL_DUMMY_REMOVER",51),j5e=new ho("FINAL_SPLINE_BENDPOINTS_CALCULATOR",52),D5e=new ho("END_LABEL_SORTER",53),Tq=new ho("REVERSED_EDGE_RESTORER",54),Gee=new ho("END_LABEL_POSTPROCESSOR",55),$5e=new ho("HIERARCHICAL_NODE_RESIZER",56),P5e=new ho("DIRECTION_POSTPROCESSOR",57)}function j7n(e,t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji,id,Ms,Ef,ZA,ev,hp,C_,Jf,p4,JF,$1,zp,tv,WE,QA,g4,KE,YE,F1,Tx,gje,A_,ZF,T1e,m4,QF,xx,eB,x1e,JMt;for(gje=0,$r=t,Ms=0,ev=$r.length;Ms<ev;++Ms)for(tn=$r[Ms],tv=new ne(tn.j);tv.a<tv.c.c.length;){for(zp=f(ee(tv),11),QA=0,m=new ne(zp.g);m.a<m.c.c.length;)g=f(ee(m),18),tn.c!=g.d.i.c&&++QA;QA>0&&(e.a[zp.p]=gje++)}for(QF=0,ji=i,Ef=0,hp=ji.length;Ef<hp;++Ef){for(tn=ji[Ef],C_=0,tv=new ne(tn.j);tv.a<tv.c.c.length&&(zp=f(ee(tv),11),zp.j==(tt(),hn));)for(m=new ne(zp.e);m.a<m.c.c.length;)if(g=f(ee(m),18),tn.c!=g.c.i.c){++C_;break}for(p4=0,g4=new la(tn.j,tn.j.c.length);g4.b>0;){for(zp=(Ln(g4.b>0),f(g4.a.Xb(g4.c=--g4.b),11)),QA=0,m=new ne(zp.e);m.a<m.c.c.length;)g=f(ee(m),18),tn.c!=g.c.i.c&&++QA;QA>0&&(zp.j==(tt(),hn)?(e.a[zp.p]=QF,++QF):(e.a[zp.p]=QF+C_+p4,++p4))}QF+=p4}for(WE=new Wn,z=new ub,In=t,id=0,ZA=In.length;id<ZA;++id)for(tn=In[id],T1e=new ne(tn.j);T1e.a<T1e.c.c.length;)for(ZF=f(ee(T1e),11),m=new ne(ZF.g);m.a<m.c.c.length;)if(g=f(ee(m),18),eB=g.d,tn.c!=eB.i.c)if(A_=f(ac(Is(WE.f,ZF)),464),xx=f(ac(Is(WE.f,eB)),464),!A_&&!xx)F=new out,z.a.yc(F,z),Qe(F.a,g),Qe(F.d,ZF),ml(WE.f,ZF,F),Qe(F.d,eB),ml(WE.f,eB,F);else if(!A_)Qe(xx.a,g),Qe(xx.d,ZF),ml(WE.f,ZF,xx);else if(!xx)Qe(A_.a,g),Qe(A_.d,eB),ml(WE.f,eB,A_);else if(A_==xx)Qe(A_.a,g);else{for(Qe(A_.a,g),$1=new ne(xx.d);$1.a<$1.c.c.length;)JF=f(ee($1),11),ml(WE.f,JF,A_);$o(A_.a,xx.a),$o(A_.d,xx.d),z.a.Ac(xx)!=null}for(W=f(Rj(z,Pe(sjn,{3:1,4:1,5:1,1991:1},464,z.a.gc(),0,1)),1991),Kt=t[0].c,Tx=i[0].c,T=W,A=0,L=T.length;A<L;++A)for(S=T[A],S.e=gje,S.f=QF,tv=new ne(S.d);tv.a<tv.c.c.length;)zp=f(ee(tv),11),KE=e.a[zp.p],zp.i.c==Kt?(KE<S.e&&(S.e=KE),KE>S.b&&(S.b=KE)):zp.i.c==Tx&&(KE<S.f&&(S.f=KE),KE>S.c&&(S.c=KE));for(LP(W,0,W.length,null),m4=Pe(pr,Pr,30,W.length,15,1),a=Pe(pr,Pr,30,QF+1,15,1),ue=0;ue<W.length;ue++)m4[ue]=W[ue].f,a[m4[ue]]=1;for(l=0,pe=0;pe<a.length;pe++)a[pe]==1?a[pe]=l:--l;for(YE=0,ve=0;ve<m4.length;ve++)m4[ve]+=a[m4[ve]],YE=h.Math.max(YE,m4[ve]+1);for(v=1;v<YE;)v*=2;for(JMt=2*v-1,v-=1,x1e=Pe(pr,Pr,30,JMt,15,1),u=0,xt=0;xt<m4.length;xt++)for(nt=m4[xt]+v,++x1e[nt];nt>0;)nt%2>0&&(u+=x1e[nt+1]),nt=(nt-1)/2|0,++x1e[nt];for(Ct=Pe(z6t,Jt,367,W.length*2,0,1),Ee=0;Ee<W.length;Ee++)Ct[2*Ee]=new KX(W[Ee],W[Ee].e,W[Ee].b,(oj(),dF)),Ct[2*Ee+1]=new KX(W[Ee],W[Ee].b,W[Ee].e,lF);for(LP(Ct,0,Ct.length,null),Jf=0,Le=0;Le<Ct.length;Le++)switch(Ct[Le].d.g){case 0:++Jf;break;case 1:--Jf,u+=Jf}for(F1=Pe(z6t,Jt,367,W.length*2,0,1),He=0;He<W.length;He++)F1[2*He]=new KX(W[He],W[He].f,W[He].c,(oj(),dF)),F1[2*He+1]=new KX(W[He],W[He].c,W[He].f,lF);for(LP(F1,0,F1.length,null),Jf=0,J=0;J<F1.length;J++)switch(F1[J].d.g){case 0:++Jf;break;case 1:--Jf,u+=Jf}return u}function xOe(){xOe=Z,iLe=(FH(),hte),dRt=new Pt(uNe,iLe),_Rt=new Pt(lNe,(mn(),!1)),lLe=(lJ(),Xge),ARt=new Pt(LQ,lLe),WRt=new Pt(dNe,!1),KRt=new Pt(fNe,!0),jNt=new Pt(hNe,!1),ELe=(MH(),Lme),u4t=new Pt(pNe,ELe),rt(1),b4t=new Pt(gNe,rt(7)),v4t=new Pt(mNe,!1),kRt=new Pt(bNe,!1),rLe=(R0(),jge),lRt=new Pt(nhe,rLe),hLe=(D0(),xme),zRt=new Pt(iq,hLe),dLe=(Ff(),Lq),LRt=new Pt(vNe,dLe),rt(-1),PRt=new Pt(wNe,null),rt(-1),DRt=new Pt(yNe,rt(-1)),rt(-1),MRt=new Pt(rhe,rt(4)),rt(-1),$Rt=new Pt(ihe,rt(2)),fLe=(gh(),qte),URt=new Pt(ohe,fLe),rt(0),BRt=new Pt(ahe,rt(0)),NRt=new Pt(she,rt(Qr)),nLe=(hj(),G$),uRt=new Pt(p$,nLe),KNt=new Pt(ENe,!1),tRt=new Pt(che,.1),sRt=new Pt(uhe,!1),rRt=new Pt(SNe,null),iRt=new Pt(_Ne,null),rt(-1),oRt=new Pt(kNe,null),rt(-1),aRt=new Pt(TNe,rt(-1)),rt(0),YNt=new Pt(xNe,rt(40)),tLe=(zP(),Kge),QNt=new Pt(lhe,tLe),eLe=Iq,XNt=new Pt(DQ,eLe),yLe=(Nj(),oF),c4t=new Pt(lA,yLe),QRt=new Si(MQ),bLe=(RH(),gte),YRt=new Pt(dhe,bLe),vLe=(gG(),mte),JRt=new Pt(fhe,vLe),n4t=new Pt(hhe,.3),i4t=new Si(phe),wLe=(xT(),Gte),o4t=new Pt(ghe,wLe),sLe=(QJ(),Mme),mRt=new Pt(CNe,sLe),cLe=(Ej(),$me),bRt=new Pt(ANe,cLe),uLe=(JP(),cF),vRt=new Pt(jQ,uLe),yRt=new Pt($Q,.2),pRt=new Pt(mhe,2),h4t=new Pt(INe,null),g4t=new Pt(ONe,10),p4t=new Pt(NNe,10),m4t=new Pt(RNe,20),rt(0),l4t=new Pt(PNe,rt(0)),rt(0),d4t=new Pt(LNe,rt(0)),rt(0),f4t=new Pt(DNe,rt(0)),$Nt=new Pt(bhe,!1),XPe=(fL(),q$),BNt=new Pt(MNe,XPe),YPe=(SJ(),Dge),FNt=new Pt(jNe,YPe),xRt=new Pt(FQ,!1),rt(0),TRt=new Pt(vhe,rt(16)),rt(0),CRt=new Pt(whe,rt(5)),kLe=(eZ(),Ume),$4t=new Pt(I1,kLe),w4t=new Pt(BQ,10),S4t=new Pt(UQ,1),_Le=(FJ(),fte),I4t=new Pt(g$,_Le),T4t=new Si(yhe),SLe=rt(1),rt(0),C4t=new Pt(Ehe,SLe),TLe=($J(),Bme),z4t=new Pt(zQ,TLe),F4t=new Si(HQ),L4t=new Pt(GQ,!0),R4t=new Pt(qQ,2),M4t=new Pt(She,!0),mLe=(pJ(),ame),VRt=new Pt(_he,mLe),pLe=rt(2),rt(1),HRt=new Pt($Ne,pLe),gLe=!0,GRt=new Pt(khe,gLe),aLe=(qZ(),pte),hRt=new Pt(FNe,aLe),oLe=(ER(),aD),fRt=new Pt(BNe,oLe),QPe=(Pp(),Pm),WNt=new Pt(VQ,QPe),VNt=new Pt(UNe,!1),qNt=new Pt(zNe,!1),JPe=(hE(),XT),UNt=new Pt(The,JPe),ZPe=(Sj(),Cme),GNt=new Pt(HNe,ZPe),zNt=new Pt(xhe,0),HNt=new Pt(Che,0),ORt=$ge,IRt=Aq,jRt=zte,FRt=zte,RRt=Tme,nRt=(_1(),W0),cRt=G$,eRt=G$,JNt=G$,ZNt=W0,e4t=aF,t4t=oF,XRt=oF,ZRt=oF,r4t=Nme,s4t=aF,a4t=aF,wRt=(k1(),c4),ERt=c4,SRt=cF,gRt=uV,y4t=kD,E4t=lx,_4t=kD,k4t=lx,O4t=kD,N4t=lx,x4t=Mge,A4t=fte,H4t=kD,G4t=lx,B4t=kD,U4t=lx,D4t=lx,P4t=lx,j4t=lx,qRt=ome}function Xr(){Xr=Z,XF=new sS(7),sje=new fb(8,94),new fb(8,64),cje=new fb(8,36),zMt=new fb(8,65),HMt=new fb(8,122),GMt=new fb(8,90),VMt=new fb(8,98),UMt=new fb(8,66),qMt=new fb(8,60),WMt=new fb(8,62),aje=new sS(11),Wne=new of(4),Su(Wne,48,57),UD=new of(4),Su(UD,48,57),Su(UD,65,90),Su(UD,95,95),Su(UD,97,122),h4=new of(4),Su(h4,9,9),Su(h4,10,10),Su(h4,12,12),Su(h4,13,13),Su(h4,32,32),uje=LT(Wne),dje=LT(UD),lje=LT(h4),BD=new Wn,YF=new Wn,BMt=ce(se(ft,1),Ye,2,6,["Cn","Lu","Ll","Lt","Lm","Lo","Mn","Me","Mc","Nd","Nl","No","Zs","Zl","Zp","Cc","Cf",null,"Co","Cs","Pd","Ps","Pe","Pc","Po","Sm","Sc","Sk","So","Pi","Pf","L","M","N","Z","C","P","S"]),ije=ce(se(ft,1),Ye,2,6,["Basic Latin","Latin-1 Supplement","Latin Extended-A","Latin Extended-B","IPA Extensions","Spacing Modifier Letters","Combining Diacritical Marks","Greek","Cyrillic","Armenian","Hebrew","Arabic","Syriac","Thaana","Devanagari","Bengali","Gurmukhi","Gujarati","Oriya","Tamil","Telugu","Kannada","Malayalam","Sinhala","Thai","Lao","Tibetan","Myanmar","Georgian","Hangul Jamo","Ethiopic","Cherokee","Unified Canadian Aboriginal Syllabics","Ogham","Runic","Khmer","Mongolian","Latin Extended Additional","Greek Extended","General Punctuation","Superscripts and Subscripts","Currency Symbols","Combining Marks for Symbols","Letterlike Symbols","Number Forms","Arrows","Mathematical Operators","Miscellaneous Technical","Control Pictures","Optical Character Recognition","Enclosed Alphanumerics","Box Drawing","Block Elements","Geometric Shapes","Miscellaneous Symbols","Dingbats","Braille Patterns","CJK Radicals Supplement","Kangxi Radicals","Ideographic Description Characters","CJK Symbols and Punctuation","Hiragana","Katakana","Bopomofo","Hangul Compatibility Jamo","Kanbun","Bopomofo Extended","Enclosed CJK Letters and Months","CJK Compatibility","CJK Unified Ideographs Extension A","CJK Unified Ideographs","Yi Syllables","Yi Radicals","Hangul Syllables",wAt,"CJK Compatibility Ideographs","Alphabetic Presentation Forms","Arabic Presentation Forms-A","Combining Half Marks","CJK Compatibility Forms","Small Form Variants","Arabic Presentation Forms-B","Specials","Halfwidth and Fullwidth Forms","Old Italic","Gothic","Deseret","Byzantine Musical Symbols","Musical Symbols","Mathematical Alphanumeric Symbols","CJK Unified Ideographs Extension B","CJK Compatibility Ideographs Supplement","Tags"]),oje=ce(se(pr,1),Pr,30,15,[66304,66351,66352,66383,66560,66639,118784,119039,119040,119295,119808,120831,131072,173782,194560,195103,917504,917631])}function wQ(){wQ=Z,$It=new Mf("OUT_T_L",0,(Yl(),Om),(pl(),np),(Np(),_c),_c,ce(se(wh,1),Jt,22,0,[xi((NT(),up),ce(se(ps,1),Ve,94,0,[lp,sp]))])),jIt=new Mf("OUT_T_C",1,Ew,np,_c,Bu,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[lp,D1])),xi(up,ce(se(ps,1),Ve,94,0,[lp,D1,Eh]))])),FIt=new Mf("OUT_T_R",2,tp,np,_c,kc,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[lp,cp]))])),IIt=new Mf("OUT_B_L",3,Om,jp,kc,_c,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[Sh,sp]))])),AIt=new Mf("OUT_B_C",4,Ew,jp,kc,Bu,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[Sh,D1])),xi(up,ce(se(ps,1),Ve,94,0,[Sh,D1,Eh]))])),OIt=new Mf("OUT_B_R",5,tp,jp,kc,kc,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[Sh,cp]))])),PIt=new Mf("OUT_L_T",6,tp,jp,_c,_c,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[sp,lp,Eh]))])),RIt=new Mf("OUT_L_C",7,tp,Sw,Bu,_c,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[sp,Pb])),xi(up,ce(se(ps,1),Ve,94,0,[sp,Pb,Eh]))])),NIt=new Mf("OUT_L_B",8,tp,np,kc,_c,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[sp,Sh,Eh]))])),MIt=new Mf("OUT_R_T",9,Om,jp,_c,kc,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[cp,lp,Eh]))])),DIt=new Mf("OUT_R_C",10,Om,Sw,Bu,kc,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[cp,Pb])),xi(up,ce(se(ps,1),Ve,94,0,[cp,Pb,Eh]))])),LIt=new Mf("OUT_R_B",11,Om,np,kc,kc,ce(se(wh,1),Jt,22,0,[xi(up,ce(se(ps,1),Ve,94,0,[cp,Sh,Eh]))])),xIt=new Mf("IN_T_L",12,Om,jp,_c,_c,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[lp,sp])),xi(bf,ce(se(ps,1),Ve,94,0,[lp,sp,Eh]))])),TIt=new Mf("IN_T_C",13,Ew,jp,_c,Bu,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[lp,D1])),xi(bf,ce(se(ps,1),Ve,94,0,[lp,D1,Eh]))])),CIt=new Mf("IN_T_R",14,tp,jp,_c,kc,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[lp,cp])),xi(bf,ce(se(ps,1),Ve,94,0,[lp,cp,Eh]))])),_It=new Mf("IN_C_L",15,Om,Sw,Bu,_c,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[Pb,sp])),xi(bf,ce(se(ps,1),Ve,94,0,[Pb,sp,Eh]))])),SIt=new Mf("IN_C_C",16,Ew,Sw,Bu,Bu,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[Pb,D1])),xi(bf,ce(se(ps,1),Ve,94,0,[Pb,D1,Eh]))])),kIt=new Mf("IN_C_R",17,tp,Sw,Bu,kc,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[Pb,cp])),xi(bf,ce(se(ps,1),Ve,94,0,[Pb,cp,Eh]))])),yIt=new Mf("IN_B_L",18,Om,np,kc,_c,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[Sh,sp])),xi(bf,ce(se(ps,1),Ve,94,0,[Sh,sp,Eh]))])),wIt=new Mf("IN_B_C",19,Ew,np,kc,Bu,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[Sh,D1])),xi(bf,ce(se(ps,1),Ve,94,0,[Sh,D1,Eh]))])),EIt=new Mf("IN_B_R",20,tp,np,kc,kc,ce(se(wh,1),Jt,22,0,[xi(bf,ce(se(ps,1),Ve,94,0,[Sh,cp])),xi(bf,ce(se(ps,1),Ve,94,0,[Sh,cp,Eh]))])),rge=new Mf(s$,21,null,null,null,null,ce(se(wh,1),Jt,22,0,[]))}function Ut(){Ut=Z,Ex=(Wv(),sn).b,f(be(je(sn.b),0),35),f(be(je(sn.b),1),19),Y0=sn.a,f(be(je(sn.a),0),35),f(be(je(sn.a),1),19),f(be(je(sn.a),2),19),f(be(je(sn.a),3),19),f(be(je(sn.a),4),19),Mw=sn.o,f(be(je(sn.o),0),35),f(be(je(sn.o),1),35),aMt=f(be(je(sn.o),2),19),f(be(je(sn.o),3),19),f(be(je(sn.o),4),19),f(be(je(sn.o),5),19),f(be(je(sn.o),6),19),f(be(je(sn.o),7),19),f(be(je(sn.o),8),19),f(be(je(sn.o),9),19),f(be(je(sn.o),10),19),f(be(je(sn.o),11),19),f(be(je(sn.o),12),19),f(be(je(sn.o),13),19),f(be(je(sn.o),14),19),f(be(je(sn.o),15),19),f(be(ci(sn.o),0),62),f(be(ci(sn.o),1),62),f(be(ci(sn.o),2),62),f(be(ci(sn.o),3),62),f(be(ci(sn.o),4),62),f(be(ci(sn.o),5),62),f(be(ci(sn.o),6),62),f(be(ci(sn.o),7),62),f(be(ci(sn.o),8),62),f(be(ci(sn.o),9),62),oMt=sn.p,f(be(je(sn.p),0),35),f(be(je(sn.p),1),35),f(be(je(sn.p),2),35),f(be(je(sn.p),3),35),f(be(je(sn.p),4),19),f(be(je(sn.p),5),19),f(be(ci(sn.p),0),62),f(be(ci(sn.p),1),62),sMt=sn.q,f(be(je(sn.q),0),35),jw=sn.v,f(be(je(sn.v),0),19),f(be(ci(sn.v),0),62),f(be(ci(sn.v),1),62),f(be(ci(sn.v),2),62),X0=sn.w,f(be(je(sn.w),0),35),f(be(je(sn.w),1),35),f(be(je(sn.w),2),35),f(be(je(sn.w),3),19),$w=sn.B,f(be(je(sn.B),0),19),f(be(ci(sn.B),0),62),f(be(ci(sn.B),1),62),f(be(ci(sn.B),2),62),cMt=sn.Q,f(be(je(sn.Q),0),19),f(be(ci(sn.Q),0),62),uMt=sn.R,f(be(je(sn.R),0),35),Ch=sn.S,f(be(ci(sn.S),0),62),f(be(ci(sn.S),1),62),f(be(ci(sn.S),2),62),f(be(ci(sn.S),3),62),f(be(ci(sn.S),4),62),f(be(ci(sn.S),5),62),f(be(ci(sn.S),6),62),f(be(ci(sn.S),7),62),f(be(ci(sn.S),8),62),f(be(ci(sn.S),9),62),f(be(ci(sn.S),10),62),f(be(ci(sn.S),11),62),f(be(ci(sn.S),12),62),f(be(ci(sn.S),13),62),f(be(ci(sn.S),14),62),J0=sn.T,f(be(je(sn.T),0),19),f(be(je(sn.T),2),19),lMt=f(be(je(sn.T),3),19),f(be(je(sn.T),4),19),f(be(ci(sn.T),0),62),f(be(ci(sn.T),1),62),f(be(je(sn.T),1),19),Z0=sn.U,f(be(je(sn.U),0),35),f(be(je(sn.U),1),35),f(be(je(sn.U),2),19),f(be(je(sn.U),3),19),f(be(je(sn.U),4),19),f(be(je(sn.U),5),19),f(be(ci(sn.U),0),62),Sx=sn.V,f(be(je(sn.V),0),19),WA=sn.W,f(be(je(sn.W),0),35),f(be(je(sn.W),1),35),f(be(je(sn.W),2),35),f(be(je(sn.W),3),19),f(be(je(sn.W),4),19),f(be(je(sn.W),5),19),dMt=sn.bb,f(be(je(sn.bb),0),35),f(be(je(sn.bb),1),35),f(be(je(sn.bb),2),35),f(be(je(sn.bb),3),35),f(be(je(sn.bb),4),35),f(be(je(sn.bb),5),35),f(be(je(sn.bb),6),35),f(be(je(sn.bb),7),19),f(be(ci(sn.bb),0),62),f(be(ci(sn.bb),1),62),fMt=sn.eb,f(be(je(sn.eb),0),35),f(be(je(sn.eb),1),35),f(be(je(sn.eb),2),35),f(be(je(sn.eb),3),35),f(be(je(sn.eb),4),35),f(be(je(sn.eb),5),35),f(be(je(sn.eb),6),19),f(be(je(sn.eb),7),19),Za=sn.ab,f(be(je(sn.ab),0),35),f(be(je(sn.ab),1),35),__=sn.H,f(be(je(sn.H),0),19),f(be(je(sn.H),1),19),f(be(je(sn.H),2),19),f(be(je(sn.H),3),19),f(be(je(sn.H),4),19),f(be(je(sn.H),5),19),f(be(ci(sn.H),0),62),k_=sn.db,f(be(je(sn.db),0),19),Ug=sn.M}function $7n(e){var t;e.O||(e.O=!0,ju(e,"type"),WJ(e,"ecore.xml.type"),KJ(e,CE),t=f(gL((Uv(),xh),CE),1990),Qn(Bs(e.fb),e.b),wc(e.b,xV,"AnyType",!1,!1,!0),oo(f(be(je(e.b),0),35),e.wb.D,fq,null,0,-1,xV,!1,!1,!0,!1,!1,!1),oo(f(be(je(e.b),1),35),e.wb.D,"any",null,0,-1,xV,!0,!0,!0,!1,!1,!0),oo(f(be(je(e.b),2),35),e.wb.D,"anyAttribute",null,0,-1,xV,!1,!1,!0,!1,!1,!1),wc(e.bb,qne,eAt,!1,!1,!0),oo(f(be(je(e.bb),0),35),e.gb,"data",null,0,1,qne,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.bb),1),35),e.gb,z4e,null,1,1,qne,!1,!1,!0,!1,!0,!1),wc(e.fb,CV,tAt,!1,!1,!0),oo(f(be(je(e.fb),0),35),t.gb,"rawValue",null,0,1,CV,!0,!0,!0,!1,!0,!0),oo(f(be(je(e.fb),1),35),t.a,k$,null,0,1,CV,!0,!0,!0,!1,!0,!0),Ai(f(be(je(e.fb),2),19),e.wb.q,null,"instanceType",1,1,CV,!1,!1,!0,!1,!1,!1,!1),wc(e.qb,eje,nAt,!1,!1,!0),oo(f(be(je(e.qb),0),35),e.wb.D,fq,null,0,-1,null,!1,!1,!0,!1,!1,!1),Ai(f(be(je(e.qb),1),19),e.wb.ab,null,"xMLNSPrefixMap",0,-1,null,!0,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.qb),2),19),e.wb.ab,null,"xSISchemaLocation",0,-1,null,!0,!1,!0,!0,!1,!1,!1),oo(f(be(je(e.qb),3),35),e.gb,"cDATA",null,0,-2,null,!0,!0,!0,!1,!1,!0),oo(f(be(je(e.qb),4),35),e.gb,"comment",null,0,-2,null,!0,!0,!0,!1,!1,!0),Ai(f(be(je(e.qb),5),19),e.bb,null,gAt,0,-2,null,!0,!0,!0,!0,!1,!1,!0),oo(f(be(je(e.qb),6),35),e.gb,lpe,null,0,-2,null,!0,!0,!0,!1,!1,!0),Fr(e.a,Ro,"AnySimpleType",!0),Fr(e.c,ft,"AnyURI",!0),Fr(e.d,se($l,1),"Base64Binary",!0),Fr(e.e,kd,"Boolean",!0),Fr(e.f,to,"BooleanObject",!0),Fr(e.g,$l,"Byte",!0),Fr(e.i,BR,"ByteObject",!0),Fr(e.j,ft,"Date",!0),Fr(e.k,ft,"DateTime",!0),Fr(e.n,Vpe,"Decimal",!0),Fr(e.o,sa,"Double",!0),Fr(e.p,mo,"DoubleObject",!0),Fr(e.q,ft,"Duration",!0),Fr(e.s,qf,"ENTITIES",!0),Fr(e.r,qf,"ENTITIESBase",!0),Fr(e.t,ft,i6e,!0),Fr(e.u,kx,"Float",!0),Fr(e.v,eD,"FloatObject",!0),Fr(e.w,ft,"GDay",!0),Fr(e.B,ft,"GMonth",!0),Fr(e.A,ft,"GMonthDay",!0),Fr(e.C,ft,"GYear",!0),Fr(e.D,ft,"GYearMonth",!0),Fr(e.F,se($l,1),"HexBinary",!0),Fr(e.G,ft,"ID",!0),Fr(e.H,ft,"IDREF",!0),Fr(e.J,qf,"IDREFS",!0),Fr(e.I,qf,"IDREFSBase",!0),Fr(e.K,pr,"Int",!0),Fr(e.M,mA,"Integer",!0),Fr(e.L,ba,"IntObject",!0),Fr(e.P,ft,"Language",!0),Fr(e.Q,x_,"Long",!0),Fr(e.R,i_,"LongObject",!0),Fr(e.S,ft,"Name",!0),Fr(e.T,ft,kee,!0),Fr(e.U,mA,"NegativeInteger",!0),Fr(e.V,ft,s6e,!0),Fr(e.X,qf,"NMTOKENS",!0),Fr(e.W,qf,"NMTOKENSBase",!0),Fr(e.Y,mA,"NonNegativeInteger",!0),Fr(e.Z,mA,"NonPositiveInteger",!0),Fr(e.$,ft,"NormalizedString",!0),Fr(e._,ft,"NOTATION",!0),Fr(e.ab,ft,"PositiveInteger",!0),Fr(e.cb,ft,"QName",!0),Fr(e.db,JA,"Short",!0),Fr(e.eb,o_,"ShortObject",!0),Fr(e.gb,ft,POe,!0),Fr(e.hb,ft,"Time",!0),Fr(e.ib,ft,"Token",!0),Fr(e.jb,JA,"UnsignedByte",!0),Fr(e.kb,o_,"UnsignedByteObject",!0),Fr(e.lb,x_,"UnsignedInt",!0),Fr(e.mb,i_,"UnsignedIntObject",!0),Fr(e.nb,mA,"UnsignedLong",!0),Fr(e.ob,pr,"UnsignedShort",!0),Fr(e.pb,ba,"UnsignedShortObject",!0),CCe(e,CE),F7n(e))}function COe(e,t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt,tn,In,$r,ji,id,Ms,Ef,ZA,ev,hp,C_,Jf,p4,JF,$1,zp,tv,WE,QA,g4,KE,YE,F1,Tx;if(a.Yg()||lt(ut(Fe(t,(Rr(),yne)))))return Ht(),Ht(),Ga;if(nt=(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i!=0,Ct=IRn(t),xt=!Ct.dc(),nt||xt){if(u=f(Fe(t,VA),149),!u)throw ie(new h0("Resolved algorithm is not set; apply a LayoutAlgorithmResolver before computing layout."));if(KE=RSe(u,(pL(),Mne)),wvt(t),!nt&&xt&&!KE)return Ht(),Ht(),Ga;if(ve=new et,Ae(Fe(t,UA))===Ae((_1(),W0))&&(RSe(u,Lne)||RSe(u,Pne))){if(lt(ut(Fe(t,RD))))throw ie(new h0("Topdown layout cannot be used together with hierarchy handling."));for(ZA=gSt(e,t),ev=new ki,Ba(ev,(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));ev.b!=0;)Ms=f(ev.b==0?null:(Ln(ev.b!=0),lh(ev,ev.a.a)),27),wvt(Ms),g4=Ae(Fe(Ms,UA))===Ae(PF),g4||mm(Ms,AD)&&!zTe(u,Fe(Ms,VA))?(J=COe(e,Ms,i,a),$o(ve,J),pi(Ms,UA,PF),RSt(Ms)):Ba(ev,(!Ms.a&&(Ms.a=new Be(Ur,Ms,10,11)),Ms.a))}else{if(ZA=(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i,lt(ut(Fe(t,RD)))){if(YE=a.bh(1),YE.Sg(Jxt,1),Fe(t,s4)==null)throw ie(new h0(t.k+" has not been assigned a top-down node type."));if(f(Fe(t,s4),278)==(gR(),bx)||f(Fe(t,s4),278)==One)for(pe=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));pe.e!=pe.i.gc();)ue=f(Mn(pe),27),id=f(Fe(ue,VA),149),(!ue.a&&(ue.a=new Be(Ur,ue,10,11)),ue.a).i>0&&tJ(id.f),Fe(ue,U9e)!=null?(m=f(Fe(ue,U9e),343),QA=m.Rg(ue),C_=f(Fe(ue,V0),107),x3(ue,h.Math.max(ue.g,QA.a+C_.b+C_.c),h.Math.max(ue.f,QA.b+C_.d+C_.a))):(!ue.a&&(ue.a=new Be(Ur,ue,10,11)),ue.a).i!=0&&x3(ue,ke(Oe(Fe(ue,ND))),ke(Oe(Fe(ue,ND)))/ke(Oe(Fe(ue,sV))));if(hp=f(Fe(t,V0),107),F=t.g-(hp.b+hp.c),L=t.f-(hp.d+hp.a),YE._g("Available Child Area: ("+F+"|"+L+")"),pi(t,ID,F/L),Tvt(t,u,a.bh(ZA)),f(Fe(t,s4),278)==One&&(SOe(t),x3(t,hp.b+ke(Oe(Fe(t,BA)))+hp.c,hp.d+ke(Oe(Fe(t,FA)))+hp.a)),YE._g("Executed layout algorithm: "+dr(Fe(t,AD))+" on node "+t.k),f(Fe(t,s4),278)==bx){if(F<0||L<0)throw ie(new h0("The size defined by the parent parallel node is too small for the space provided by the paddings of the child hierarchical node. "+t.k));for(mm(t,BA)||mm(t,FA)||SOe(t),W=ke(Oe(Fe(t,BA))),z=ke(Oe(Fe(t,FA))),YE._g("Desired Child Area: ("+W+"|"+z+")"),p4=F/W,JF=L/z,Jf=h.Math.min(p4,h.Math.min(JF,ke(Oe(Fe(t,oDt))))),pi(t,Sne,Jf),YE._g(t.k+" -- Local Scale Factor (X|Y): ("+p4+"|"+JF+")"),Ee=f(Fe(t,tV),22),l=0,g=0,Jf<p4&&(Ee.Gc((mE(),xF))?l=(F/2-W*Jf/2)/Jf:Ee.Gc(CF)&&(l=(F-W*Jf)/Jf)),Jf<JF&&(Ee.Gc((mE(),IF))?g=(L/2-z*Jf/2)/Jf:Ee.Gc(AF)&&(g=(L-z*Jf)/Jf)),F1=l+(hp.b/Jf-hp.b),Tx=g+(hp.d/Jf-hp.d),YE._g("Shift: ("+F1+"|"+Tx+")"),Ef=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));Ef.e!=Ef.i.gc();)Ms=f(Mn(Ef),27),md(Ms,Ms.i+F1),bd(Ms,Ms.j+Tx);for(He=new Nn((!t.b&&(t.b=new Be(vo,t,12,3)),t.b));He.e!=He.i.gc();){for(Le=f(Mn(He),85),zp=new Nn((!Le.a&&(Le.a=new Be(Ii,Le,6,6)),Le.a));zp.e!=zp.i.gc();)for($1=f(Mn(zp),168),Yz($1,$1.j+F1,$1.k+Tx),Kz($1,$1.b+F1,$1.c+Tx),S=new Nn((!$1.a&&($1.a=new So(yf,$1,5)),$1.a));S.e!=S.i.gc();)v=f(Mn(S),369),Ese(v,v.a+F1,v.b+Tx);for(ji=new Nn((!Le.n&&(Le.n=new Be(hc,Le,1,7)),Le.n));ji.e!=ji.i.gc();)$r=f(Mn(ji),153),Of($r,$r.i+F1,$r.j+Tx);for(In=f(Fe(Le,zA),77),tn=tr(In,0);tn.b!=tn.d.c;)Kt=f(Zn(tn),8),Kt.a+=F1,Kt.b+=Tx;pi(Le,zA,In)}}YE.Tg()}for(A=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));A.e!=A.i.gc();)T=f(Mn(A),27),J=COe(e,T,i,a),$o(ve,J),RSt(T)}if(a.Yg())return Ht(),Ht(),Ga;for(WE=new ne(ve);WE.a<WE.c.c.length;)tv=f(ee(WE),85),pi(tv,yne,(mn(),!0));return lt(ut(Fe(t,RD)))||Tvt(t,u,a.bh(ZA)),NLn(ve),xt&&KE?Ct:(Ht(),Ht(),Ga)}else return Ht(),Ht(),Ga}function aA(e,t){var i,a;return XA||(XA=new Wn,FD=new Wn,a=(Xr(),Xr(),new of(4)),aG(a,`
\r\r  `),Os(XA,Ipe,a),Os(FD,Ipe,LT(a)),a=new of(4),aG(a,bAt),Os(XA,P$,a),Os(FD,P$,LT(a)),a=new of(4),aG(a,bAt),Os(XA,P$,a),Os(FD,P$,LT(a)),a=new of(4),aG(a,vAt),MT(a,f(wu(XA,P$),121)),Os(XA,Ape,a),Os(FD,Ape,LT(a)),a=new of(4),aG(a,"-.0:AZ__az··ÀÖØöøıĴľŁňŊžƀǃǍǰǴǵǺȗɐʨʻˁːˑ̀͠͡ͅΆΊΌΌΎΡΣώϐϖϚϚϜϜϞϞϠϠϢϳЁЌЎяёќўҁ҃҆ҐӄӇӈӋӌӐӫӮӵӸӹԱՖՙՙաֆֹֻֽֿֿׁׂ֑֣֡ׄׄאתװײءغـْ٠٩ٰڷںھۀێېۓە۪ۭۨ۰۹ँःअह़्॑॔क़ॣ०९ঁঃঅঌএঐওনপরললশহ়়াৄেৈো্ৗৗড়ঢ়য়ৣ০ৱਂਂਅਊਏਐਓਨਪਰਲਲ਼ਵਸ਼ਸਹ਼਼ਾੂੇੈੋ੍ਖ਼ੜਫ਼ਫ਼੦ੴઁઃઅઋઍઍએઑઓનપરલળવહ઼ૅેૉો્ૠૠ૦૯ଁଃଅଌଏଐଓନପରଲଳଶହ଼ୃେୈୋ୍ୖୗଡ଼ଢ଼ୟୡ୦୯ஂஃஅஊஎஐஒகஙசஜஜஞடணதநபமவஷஹாூெைொ்ௗௗ௧௯ఁఃఅఌఎఐఒనపళవహాౄెైొ్ౕౖౠౡ౦౯ಂಃಅಌಎಐಒನಪಳವಹಾೄೆೈೊ್ೕೖೞೞೠೡ೦೯ംഃഅഌഎഐഒനപഹാൃെൈൊ്ൗൗൠൡ൦൯กฮะฺเ๎๐๙ກຂຄຄງຈຊຊຍຍດທນຟມຣລລວວສຫອຮະູົຽເໄໆໆ່ໍ໐໙༘༙༠༩༹༹༵༵༷༷༾ཇཉཀྵ྄ཱ྆ྋྐྕྗྗྙྭྱྷྐྵྐྵႠჅაჶᄀᄀᄂᄃᄅᄇᄉᄉᄋᄌᄎᄒᄼᄼᄾᄾᅀᅀᅌᅌᅎᅎᅐᅐᅔᅕᅙᅙᅟᅡᅣᅣᅥᅥᅧᅧᅩᅩᅭᅮᅲᅳᅵᅵᆞᆞᆨᆨᆫᆫᆮᆯᆷᆸᆺᆺᆼᇂᇫᇫᇰᇰᇹᇹḀẛẠỹἀἕἘἝἠὅὈὍὐὗὙὙὛὛὝὝὟώᾀᾴᾶᾼιιῂῄῆῌῐΐῖΊῠῬῲῴῶῼ⃐⃜⃡⃡ΩΩKÅ℮℮ↀↂ々々〇〇〡〯〱〵ぁゔ゙゚ゝゞァヺーヾㄅㄬ一龥가힣"),Os(XA,Ope,a),Os(FD,Ope,LT(a)),a=new of(4),aG(a,vAt),Su(a,95,95),Su(a,58,58),Os(XA,Npe,a),Os(FD,Npe,LT(a))),i=f(wu(t?XA:FD,e),136),i}function qkt(e){Wk(e,new zS(B2e(gY(qk(zk(Gk(Hk(new ug,wn),"ELK Layered"),"Layer-based algorithm provided by the Eclipse Layout Kernel. Arranges as many edges as possible into one direction by placing nodes into subsequent layers. This implementation supports different routing styles (straight, orthogonal, splines); if orthogonal routing is selected, arbitrary port constraints are respected, thus enabling the layout of block diagrams such as actor-oriented models or circuit schematics. Furthermore, full layout of compound graphs with cross-hierarchy edges is supported when the respective option is activated on the top level."),new SC),wn),xi((pL(),h1e),ce(se($ne,1),Ve,243,0,[Mne,jne,Dne,f1e,Lne,Pne]))))),Je(e,wn,XNe,st(wme)),Je(e,wn,JNe,st(iDe)),Je(e,wn,ZNe,st(jq)),Je(e,wn,QNe,st(op)),Je(e,wn,Kfe,st(OA)),Je(e,wn,Ohe,st(b_)),Je(e,wn,eRe,st(cx)),Je(e,wn,tRe,st(yD)),Je(e,wn,nRe,st(ED)),Je(e,wn,rRe,st(yme)),Je(e,wn,UT,st(v_)),Je(e,wn,iRe,st(Eme)),Je(e,wn,oRe,st(rF)),Je(e,wn,aRe,st(Fte)),Je(e,wn,INe,st(Mq)),Je(e,wn,NNe,st(IA)),Je(e,wn,ONe,st(jE)),Je(e,wn,RNe,st(NA)),Je(e,wn,XG,rt(0)),Je(e,wn,PNe,st(wD)),Je(e,wn,LNe,st(rDe)),Je(e,wn,DNe,st(e4)),Je(e,wn,I1,st(hDe)),Je(e,wn,BQ,st(aDe)),Je(e,wn,UQ,st(sDe)),Je(e,wn,g$,st(_me)),Je(e,wn,yhe,st(cDe)),Je(e,wn,Ehe,st(uDe)),Je(e,wn,zQ,st(Bte)),Je(e,wn,HQ,st(kme)),Je(e,wn,GQ,st(dDe)),Je(e,wn,qQ,st(lDe)),Je(e,wn,She,st(fDe)),Je(e,wn,_he,st(VLe)),Je(e,wn,$Ne,st(gD)),Je(e,wn,khe,st(eF)),Je(e,wn,phe,st(g_)),Je(e,wn,ghe,st(nF)),Je(e,wn,jQ,st(gme)),Je(e,wn,$Q,st(DLe)),Je(e,wn,JG,st(a6t)),Je(e,wn,ZG,st(s6t)),Je(e,wn,QG,st(o6t)),Je(e,wn,eq,st(i6t)),Je(e,wn,ZS,oDe),Je(e,wn,zT,ZLe),Je(e,wn,oq,RLe),Je(e,wn,sRe,0),Je(e,wn,AQ,rt(1)),Je(e,wn,ML,DL),Je(e,wn,cRe,st(ME)),Je(e,wn,Yfe,st(Xi)),Je(e,wn,uRe,st(vD)),Je(e,wn,aq,st(K4t)),Je(e,wn,lRe,st(Nm)),Je(e,wn,WQ,st(rx)),Je(e,wn,u$,(mn(),!0)),Je(e,wn,dRe,st(ix)),Je(e,wn,fRe,st(PE)),Je(e,wn,uA,st(DE)),Je(e,wn,LR,st($te)),Je(e,wn,jL,st(vme)),Je(e,wn,Nhe,NLe),Je(e,wn,$L,st(p_)),Je(e,wn,hRe,st(Mte)),Je(e,wn,FL,st(ax)),Je(e,wn,pRe,st(e6t)),Je(e,wn,gRe,st(tDe)),Je(e,wn,mRe,eDe),Je(e,wn,bRe,st(J4t)),Je(e,wn,vRe,st(Z4t)),Je(e,wn,wRe,st(Q4t)),Je(e,wn,yRe,st(X4t)),Je(e,wn,mNe,st(Sme)),Je(e,wn,iq,st(LE)),Je(e,wn,ohe,st(xw)),Je(e,wn,gNe,st(iF)),Je(e,wn,vNe,st(el)),Je(e,wn,nhe,st(RE)),Je(e,wn,p$,st(Q$)),Je(e,wn,ENe,st(h_)),Je(e,wn,xNe,st(ALe)),Je(e,wn,lhe,st(dme)),Je(e,wn,DQ,st(Dq)),Je(e,wn,uhe,st(fme)),Je(e,wn,dNe,st(WLe)),Je(e,wn,fNe,st(KLe)),Je(e,wn,LQ,st(ULe)),Je(e,wn,lA,st(jte)),Je(e,wn,fhe,st(bme)),Je(e,wn,lNe,st(mme)),Je(e,wn,hhe,st(XLe)),Je(e,wn,CNe,st(LLe)),Je(e,wn,ANe,st(pme)),Je(e,wn,KQ,st(lme)),Je(e,wn,dhe,st(YLe)),Je(e,wn,MNe,st(Cte)),Je(e,wn,jNe,st(xLe)),Je(e,wn,bhe,st(xte)),Je(e,wn,FQ,st($Le)),Je(e,wn,vhe,st(jLe)),Je(e,wn,whe,st(FLe)),Je(e,wn,DR,st(mD)),Je(e,wn,ERe,st(Ls)),Je(e,wn,SRe,st(z0)),Je(e,wn,_Re,st(Rm)),Je(e,wn,IQ,st(hme)),Je(e,wn,che,st(ILe)),Je(e,wn,kRe,st(H0)),Je(e,wn,TRe,st(Z$)),Je(e,wn,xRe,st(Pte)),Je(e,wn,CRe,st(m_)),Je(e,wn,Rhe,st(QLe)),Je(e,wn,Phe,st(bD)),Je(e,wn,rhe,st(HLe)),Je(e,wn,ihe,st(GLe)),Je(e,wn,YQ,st(sx)),Je(e,wn,hNe,st(sme)),Je(e,wn,ahe,st(qLe)),Je(e,wn,FNe,st(Rte)),Je(e,wn,BNe,st(Nte)),Je(e,wn,ARe,st(Dte)),Je(e,wn,she,st(zLe)),Je(e,wn,MQ,st(tF)),Je(e,wn,IRe,st($q)),Je(e,wn,uNe,st(OLe)),Je(e,wn,pNe,st(nDe)),Je(e,wn,mhe,st(PLe)),Je(e,wn,SNe,st(q4t)),Je(e,wn,_Ne,st(V4t)),Je(e,wn,wNe,st(Y4t)),Je(e,wn,kNe,st(W4t)),Je(e,wn,XQ,st(BLe)),Je(e,wn,yNe,st(Lte)),Je(e,wn,TNe,st(Ote)),Je(e,wn,VQ,st(ip)),Je(e,wn,HNe,st(cme)),Je(e,wn,xhe,st(Ate)),Je(e,wn,Che,st(CLe)),Je(e,wn,zNe,st(Ite)),Je(e,wn,The,st(nx)),Je(e,wn,UNe,st(ume)),Je(e,wn,bNe,st(MLe))}function F7n(e){ir(e.a,kr,ce(se(ft,1),Ye,2,6,[Gr,"anySimpleType"])),ir(e.b,kr,ce(se(ft,1),Ye,2,6,[Gr,"anyType",mh,fq])),ir(f(be(je(e.b),0),35),kr,ce(se(ft,1),Ye,2,6,[mh,_pe,Gr,":mixed"])),ir(f(be(je(e.b),1),35),kr,ce(se(ft,1),Ye,2,6,[mh,_pe,n6e,Tpe,Gr,":1",rAt,"lax"])),ir(f(be(je(e.b),2),35),kr,ce(se(ft,1),Ye,2,6,[mh,Q3t,n6e,Tpe,Gr,":2",rAt,"lax"])),ir(e.c,kr,ce(se(ft,1),Ye,2,6,[Gr,"anyURI",vh,Mp])),ir(e.d,kr,ce(se(ft,1),Ye,2,6,[Gr,"base64Binary",vh,Mp])),ir(e.e,kr,ce(se(ft,1),Ye,2,6,[Gr,xR,vh,Mp])),ir(e.f,kr,ce(se(ft,1),Ye,2,6,[Gr,"boolean:Object",Ta,xR])),ir(e.g,kr,ce(se(ft,1),Ye,2,6,[Gr,x$])),ir(e.i,kr,ce(se(ft,1),Ye,2,6,[Gr,"byte:Object",Ta,x$])),ir(e.j,kr,ce(se(ft,1),Ye,2,6,[Gr,"date",vh,Mp])),ir(e.k,kr,ce(se(ft,1),Ye,2,6,[Gr,"dateTime",vh,Mp])),ir(e.n,kr,ce(se(ft,1),Ye,2,6,[Gr,"decimal",vh,Mp])),ir(e.o,kr,ce(se(ft,1),Ye,2,6,[Gr,C$,vh,Mp])),ir(e.p,kr,ce(se(ft,1),Ye,2,6,[Gr,"double:Object",Ta,C$])),ir(e.q,kr,ce(se(ft,1),Ye,2,6,[Gr,"duration",vh,Mp])),ir(e.s,kr,ce(se(ft,1),Ye,2,6,[Gr,"ENTITIES",Ta,iAt,r6e,"1"])),ir(e.r,kr,ce(se(ft,1),Ye,2,6,[Gr,iAt,kpe,i6e])),ir(e.t,kr,ce(se(ft,1),Ye,2,6,[Gr,i6e,Ta,kee])),ir(e.u,kr,ce(se(ft,1),Ye,2,6,[Gr,A$,vh,Mp])),ir(e.v,kr,ce(se(ft,1),Ye,2,6,[Gr,"float:Object",Ta,A$])),ir(e.w,kr,ce(se(ft,1),Ye,2,6,[Gr,"gDay",vh,Mp])),ir(e.B,kr,ce(se(ft,1),Ye,2,6,[Gr,"gMonth",vh,Mp])),ir(e.A,kr,ce(se(ft,1),Ye,2,6,[Gr,"gMonthDay",vh,Mp])),ir(e.C,kr,ce(se(ft,1),Ye,2,6,[Gr,"gYear",vh,Mp])),ir(e.D,kr,ce(se(ft,1),Ye,2,6,[Gr,"gYearMonth",vh,Mp])),ir(e.F,kr,ce(se(ft,1),Ye,2,6,[Gr,"hexBinary",vh,Mp])),ir(e.G,kr,ce(se(ft,1),Ye,2,6,[Gr,"ID",Ta,kee])),ir(e.H,kr,ce(se(ft,1),Ye,2,6,[Gr,"IDREF",Ta,kee])),ir(e.J,kr,ce(se(ft,1),Ye,2,6,[Gr,"IDREFS",Ta,oAt,r6e,"1"])),ir(e.I,kr,ce(se(ft,1),Ye,2,6,[Gr,oAt,kpe,"IDREF"])),ir(e.K,kr,ce(se(ft,1),Ye,2,6,[Gr,I$])),ir(e.M,kr,ce(se(ft,1),Ye,2,6,[Gr,o6e])),ir(e.L,kr,ce(se(ft,1),Ye,2,6,[Gr,"int:Object",Ta,I$])),ir(e.P,kr,ce(se(ft,1),Ye,2,6,[Gr,"language",Ta,xpe,Cpe,aAt])),ir(e.Q,kr,ce(se(ft,1),Ye,2,6,[Gr,O$])),ir(e.R,kr,ce(se(ft,1),Ye,2,6,[Gr,"long:Object",Ta,O$])),ir(e.S,kr,ce(se(ft,1),Ye,2,6,[Gr,"Name",Ta,xpe,Cpe,a6e])),ir(e.T,kr,ce(se(ft,1),Ye,2,6,[Gr,kee,Ta,"Name",Cpe,sAt])),ir(e.U,kr,ce(se(ft,1),Ye,2,6,[Gr,"negativeInteger",Ta,cAt,bq,"-1"])),ir(e.V,kr,ce(se(ft,1),Ye,2,6,[Gr,s6e,Ta,xpe,Cpe,"\\c+"])),ir(e.X,kr,ce(se(ft,1),Ye,2,6,[Gr,"NMTOKENS",Ta,uAt,r6e,"1"])),ir(e.W,kr,ce(se(ft,1),Ye,2,6,[Gr,uAt,kpe,s6e])),ir(e.Y,kr,ce(se(ft,1),Ye,2,6,[Gr,c6e,Ta,o6e,vq,"0"])),ir(e.Z,kr,ce(se(ft,1),Ye,2,6,[Gr,cAt,Ta,o6e,bq,"0"])),ir(e.$,kr,ce(se(ft,1),Ye,2,6,[Gr,lAt,Ta,sfe,vh,"replace"])),ir(e._,kr,ce(se(ft,1),Ye,2,6,[Gr,"NOTATION",vh,Mp])),ir(e.ab,kr,ce(se(ft,1),Ye,2,6,[Gr,"positiveInteger",Ta,c6e,vq,"1"])),ir(e.bb,kr,ce(se(ft,1),Ye,2,6,[Gr,"processingInstruction_._type",mh,"empty"])),ir(f(be(je(e.bb),0),35),kr,ce(se(ft,1),Ye,2,6,[mh,yee,Gr,"data"])),ir(f(be(je(e.bb),1),35),kr,ce(se(ft,1),Ye,2,6,[mh,yee,Gr,z4e])),ir(e.cb,kr,ce(se(ft,1),Ye,2,6,[Gr,"QName",vh,Mp])),ir(e.db,kr,ce(se(ft,1),Ye,2,6,[Gr,N$])),ir(e.eb,kr,ce(se(ft,1),Ye,2,6,[Gr,"short:Object",Ta,N$])),ir(e.fb,kr,ce(se(ft,1),Ye,2,6,[Gr,"simpleAnyType",mh,hq])),ir(f(be(je(e.fb),0),35),kr,ce(se(ft,1),Ye,2,6,[Gr,":3",mh,hq])),ir(f(be(je(e.fb),1),35),kr,ce(se(ft,1),Ye,2,6,[Gr,":4",mh,hq])),ir(f(be(je(e.fb),2),19),kr,ce(se(ft,1),Ye,2,6,[Gr,":5",mh,hq])),ir(e.gb,kr,ce(se(ft,1),Ye,2,6,[Gr,sfe,vh,"preserve"])),ir(e.hb,kr,ce(se(ft,1),Ye,2,6,[Gr,"time",vh,Mp])),ir(e.ib,kr,ce(se(ft,1),Ye,2,6,[Gr,xpe,Ta,lAt,vh,Mp])),ir(e.jb,kr,ce(se(ft,1),Ye,2,6,[Gr,dAt,bq,"255",vq,"0"])),ir(e.kb,kr,ce(se(ft,1),Ye,2,6,[Gr,"unsignedByte:Object",Ta,dAt])),ir(e.lb,kr,ce(se(ft,1),Ye,2,6,[Gr,fAt,bq,"4294967295",vq,"0"])),ir(e.mb,kr,ce(se(ft,1),Ye,2,6,[Gr,"unsignedInt:Object",Ta,fAt])),ir(e.nb,kr,ce(se(ft,1),Ye,2,6,[Gr,"unsignedLong",Ta,c6e,bq,hAt,vq,"0"])),ir(e.ob,kr,ce(se(ft,1),Ye,2,6,[Gr,pAt,bq,"65535",vq,"0"])),ir(e.pb,kr,ce(se(ft,1),Ye,2,6,[Gr,"unsignedShort:Object",Ta,pAt])),ir(e.qb,kr,ce(se(ft,1),Ye,2,6,[Gr,"",mh,fq])),ir(f(be(je(e.qb),0),35),kr,ce(se(ft,1),Ye,2,6,[mh,_pe,Gr,":mixed"])),ir(f(be(je(e.qb),1),19),kr,ce(se(ft,1),Ye,2,6,[mh,yee,Gr,"xmlns:prefix"])),ir(f(be(je(e.qb),2),19),kr,ce(se(ft,1),Ye,2,6,[mh,yee,Gr,"xsi:schemaLocation"])),ir(f(be(je(e.qb),3),35),kr,ce(se(ft,1),Ye,2,6,[mh,Eee,Gr,"cDATA",See,pq])),ir(f(be(je(e.qb),4),35),kr,ce(se(ft,1),Ye,2,6,[mh,Eee,Gr,"comment",See,pq])),ir(f(be(je(e.qb),5),19),kr,ce(se(ft,1),Ye,2,6,[mh,Eee,Gr,gAt,See,pq])),ir(f(be(je(e.qb),6),35),kr,ce(se(ft,1),Ye,2,6,[mh,Eee,Gr,lpe,See,pq]))}function yr(e){return Mt("_UI_EMFDiagnostic_marker",e)?"EMF Problem":Mt("_UI_CircularContainment_diagnostic",e)?"An object may not circularly contain itself":Mt(JCt,e)?"Wrong character.":Mt(ZCt,e)?"Invalid reference number.":Mt(dee,e)?"A character is required after \\.":Mt(bpe,e)?"'?' is not expected.  '(?:' or '(?=' or '(?!' or '(?<' or '(?#' or '(?>'?":Mt(QCt,e)?"'(?<' or '(?<!' is expected.":Mt(e3t,e)?"A comment is not terminated.":Mt(xE,e)?"')' is expected.":Mt(H4e,e)?"Unexpected end of the pattern in a modifier group.":Mt(t3t,e)?"':' is expected.":Mt(n3t,e)?"Unexpected end of the pattern in a conditional group.":Mt(r3t,e)?"A back reference or an anchor or a lookahead or a look-behind is expected in a conditional pattern.":Mt(i3t,e)?"There are more than three choices in a conditional group.":Mt(o3t,e)?"A character in U+0040-U+005f must follow \\c.":Mt(a3t,e)?"A '{' is required before a character category.":Mt(s3t,e)?"A property name is not closed by '}'.":Mt(G4e,e)?"Unexpected meta character.":Mt(vpe,e)?"Unknown property.":Mt(q4e,e)?"A POSIX character class must be closed by ':]'.":Mt(fee,e)?"Unexpected end of the pattern in a character class.":Mt(c3t,e)?"Unknown name for a POSIX character class.":Mt("parser.cc.4",e)?"'-' is invalid here.":Mt(u3t,e)?"']' is expected.":Mt(V4e,e)?"'[' is invalid in a character class.  Write '\\['.":Mt(W4e,e)?"']' is invalid in a character class.  Write '\\]'.":Mt(wpe,e)?"'-' is an invalid character range. Write '\\-'.":Mt(l3t,e)?"'[' is expected.":Mt(d3t,e)?"')' or '-[' or '+[' or '&[' is expected.":Mt(f3t,e)?"The range end code point is less than the start code point.":Mt(F0,e)?"Invalid Unicode hex notation.":Mt(h3t,e)?"Overflow in a hex notation.":Mt(p3t,e)?"'\\x{' must be closed by '}'.":Mt(g3t,e)?"Invalid Unicode code point.":Mt(m3t,e)?"An anchor must not be here.":Mt(Hf,e)?"This expression is not supported in the current option setting.":Mt(b3t,e)?"Invalid quantifier. A digit is expected.":Mt(v3t,e)?"Invalid quantifier. Invalid quantity or a '}' is missing.":Mt(w3t,e)?"Invalid quantifier. A digit or '}' is expected.":Mt(y3t,e)?"Invalid quantifier. A min quantity must be <= a max quantity.":Mt(K4e,e)?"Invalid quantifier. A quantity value overflow.":Mt("_UI_PackageRegistry_extensionpoint",e)?"Ecore Package Registry for Generated Packages":Mt("_UI_DynamicPackageRegistry_extensionpoint",e)?"Ecore Package Registry for Dynamic Packages":Mt("_UI_FactoryRegistry_extensionpoint",e)?"Ecore Factory Override Registry":Mt("_UI_URIExtensionParserRegistry_extensionpoint",e)?"URI Extension Parser Registry":Mt("_UI_URIProtocolParserRegistry_extensionpoint",e)?"URI Protocol Parser Registry":Mt("_UI_URIContentParserRegistry_extensionpoint",e)?"URI Content Parser Registry":Mt("_UI_ContentHandlerRegistry_extensionpoint",e)?"Content Handler Registry":Mt("_UI_URIMappingRegistry_extensionpoint",e)?"URI Converter Mapping Registry":Mt("_UI_PackageRegistryImplementation_extensionpoint",e)?"Ecore Package Registry Implementation":Mt("_UI_ValidationDelegateRegistry_extensionpoint",e)?"Validation Delegate Registry":Mt("_UI_SettingDelegateRegistry_extensionpoint",e)?"Feature Setting Delegate Factory Registry":Mt("_UI_InvocationDelegateRegistry_extensionpoint",e)?"Operation Invocation Delegate Factory Registry":Mt("_UI_EClassInterfaceNotAbstract_diagnostic",e)?"A class that is an interface must also be abstract":Mt("_UI_EClassNoCircularSuperTypes_diagnostic",e)?"A class may not be a super type of itself":Mt("_UI_EClassNotWellFormedMapEntryNoInstanceClassName_diagnostic",e)?"A class that inherits from a map entry class must have instance class name 'java.util.Map$Entry'":Mt("_UI_EReferenceOppositeOfOppositeInconsistent_diagnostic",e)?"The opposite of the opposite may not be a reference different from this one":Mt("_UI_EReferenceOppositeNotFeatureOfType_diagnostic",e)?"The opposite must be a feature of the reference's type":Mt("_UI_EReferenceTransientOppositeNotTransient_diagnostic",e)?"The opposite of a transient reference must be transient if it is proxy resolving":Mt("_UI_EReferenceOppositeBothContainment_diagnostic",e)?"The opposite of a containment reference must not be a containment reference":Mt("_UI_EReferenceConsistentUnique_diagnostic",e)?"A containment or bidirectional reference must be unique if its upper bound is different from 1":Mt("_UI_ETypedElementNoType_diagnostic",e)?"The typed element must have a type":Mt("_UI_EAttributeNoDataType_diagnostic",e)?"The generic attribute type must not refer to a class":Mt("_UI_EReferenceNoClass_diagnostic",e)?"The generic reference type must not refer to a data type":Mt("_UI_EGenericTypeNoTypeParameterAndClassifier_diagnostic",e)?"A generic type can't refer to both a type parameter and a classifier":Mt("_UI_EGenericTypeNoClass_diagnostic",e)?"A generic super type must refer to a class":Mt("_UI_EGenericTypeNoTypeParameterOrClassifier_diagnostic",e)?"A generic type in this context must refer to a classifier or a type parameter":Mt("_UI_EGenericTypeBoundsOnlyForTypeArgument_diagnostic",e)?"A generic type may have bounds only when used as a type argument":Mt("_UI_EGenericTypeNoUpperAndLowerBound_diagnostic",e)?"A generic type must not have both a lower and an upper bound":Mt("_UI_EGenericTypeNoTypeParameterOrClassifierAndBound_diagnostic",e)?"A generic type with bounds must not also refer to a type parameter or classifier":Mt("_UI_EGenericTypeNoArguments_diagnostic",e)?"A generic type may have arguments only if it refers to a classifier":Mt("_UI_EGenericTypeOutOfScopeTypeParameter_diagnostic",e)?"A generic type may only refer to a type parameter that is in scope":e}function B7n(e){var t,i,a,u,l,g,m,v,S,T,A,L,F,z,W;e.r||(e.r=!0,ju(e,"graph"),WJ(e,"graph"),KJ(e,WL),WH(e.o,"T"),Qn(Bs(e.a),e.p),Qn(Bs(e.f),e.a),Qn(Bs(e.n),e.f),Qn(Bs(e.g),e.n),Qn(Bs(e.c),e.n),Qn(Bs(e.i),e.c),Qn(Bs(e.j),e.c),Qn(Bs(e.d),e.f),Qn(Bs(e.e),e.a),wc(e.p,njn,UTt,!0,!0,!1),z=q3(e.p,e.p,"setProperty"),W=d1t(z),S=f1(e.o),T=(i=(a=new cS,a),i),Qn((!S.d&&(S.d=new So(gs,S,1)),S.d),T),A=Rce(W),eAe(T,A),MZ(z,S,D4e),S=Rce(W),MZ(z,S,k$),z=q3(e.p,null,"getProperty"),W=d1t(z),S=f1(e.o),T=Rce(W),Qn((!S.d&&(S.d=new So(gs,S,1)),S.d),T),MZ(z,S,D4e),S=Rce(W),F=xg(z,S,null),F&&F.lj(),z=q3(e.p,e.wb.e,"hasProperty"),S=f1(e.o),T=(u=(l=new cS,l),u),Qn((!S.d&&(S.d=new So(gs,S,1)),S.d),T),MZ(z,S,D4e),z=q3(e.p,e.p,"copyProperties"),Ps(z,e.p,ape),z=q3(e.p,null,"getAllProperties"),S=f1(e.wb.P),T=f1(e.o),Qn((!S.d&&(S.d=new So(gs,S,1)),S.d),T),A=(g=(m=new cS,m),g),Qn((!T.d&&(T.d=new So(gs,T,1)),T.d),A),T=f1(e.wb.M),Qn((!S.d&&(S.d=new So(gs,S,1)),S.d),T),L=xg(z,S,null),L&&L.lj(),wc(e.a,$D,xCt,!0,!1,!0),Ai(f(be(je(e.a),0),19),e.k,null,$Ct,0,-1,$D,!1,!1,!0,!0,!1,!1,!1),wc(e.f,wV,ACt,!0,!1,!0),Ai(f(be(je(e.f),0),19),e.g,f(be(je(e.g),0),19),"labels",0,-1,wV,!1,!1,!0,!0,!1,!1,!1),oo(f(be(je(e.f),1),35),e.wb._,FCt,null,0,1,wV,!1,!1,!0,!1,!0,!1),wc(e.n,yV,"ElkShape",!0,!1,!0),oo(f(be(je(e.n),0),35),e.wb.t,spe,OL,1,1,yV,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.n),1),35),e.wb.t,cpe,OL,1,1,yV,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.n),2),35),e.wb.t,"x",OL,1,1,yV,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.n),3),35),e.wb.t,"y",OL,1,1,yV,!1,!1,!0,!1,!0,!1),z=q3(e.n,null,"setDimensions"),Ps(z,e.wb.t,cpe),Ps(z,e.wb.t,spe),z=q3(e.n,null,"setLocation"),Ps(z,e.wb.t,"x"),Ps(z,e.wb.t,"y"),wc(e.g,hc,R4e,!1,!1,!0),Ai(f(be(je(e.g),0),19),e.f,f(be(je(e.f),0),19),upe,0,1,hc,!1,!1,!0,!1,!1,!1,!1),oo(f(be(je(e.g),1),35),e.wb._,lpe,"",0,1,hc,!1,!1,!0,!1,!0,!1),wc(e.c,Yn,ICt,!0,!1,!0),Ai(f(be(je(e.c),0),19),e.d,f(be(je(e.d),1),19),"outgoingEdges",0,-1,Yn,!1,!1,!0,!1,!0,!1,!1),Ai(f(be(je(e.c),1),19),e.d,f(be(je(e.d),2),19),"incomingEdges",0,-1,Yn,!1,!1,!0,!1,!0,!1,!1),wc(e.i,Ur,P4e,!1,!1,!0),Ai(f(be(je(e.i),0),19),e.j,f(be(je(e.j),0),19),"ports",0,-1,Ur,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.i),1),19),e.i,f(be(je(e.i),2),19),dpe,0,-1,Ur,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.i),2),19),e.i,f(be(je(e.i),1),19),upe,0,1,Ur,!1,!1,!0,!1,!1,!1,!1),Ai(f(be(je(e.i),3),19),e.d,f(be(je(e.d),0),19),"containedEdges",0,-1,Ur,!1,!1,!0,!0,!1,!1,!1),oo(f(be(je(e.i),4),35),e.wb.e,BCt,null,0,1,Ur,!0,!0,!1,!1,!0,!0),wc(e.j,_d,L4e,!1,!1,!0),Ai(f(be(je(e.j),0),19),e.i,f(be(je(e.i),0),19),upe,0,1,_d,!1,!1,!0,!1,!1,!1,!1),wc(e.d,vo,N4e,!1,!1,!0),Ai(f(be(je(e.d),0),19),e.i,f(be(je(e.i),3),19),"containingNode",0,1,vo,!1,!1,!0,!1,!1,!1,!1),Ai(f(be(je(e.d),1),19),e.c,f(be(je(e.c),0),19),M4e,0,-1,vo,!1,!1,!0,!1,!0,!1,!1),Ai(f(be(je(e.d),2),19),e.c,f(be(je(e.c),1),19),fpe,0,-1,vo,!1,!1,!0,!1,!0,!1,!1),Ai(f(be(je(e.d),3),19),e.e,f(be(je(e.e),5),19),hpe,0,-1,vo,!1,!1,!0,!0,!1,!1,!1),oo(f(be(je(e.d),4),35),e.wb.e,"hyperedge",null,0,1,vo,!0,!0,!1,!1,!0,!0),oo(f(be(je(e.d),5),35),e.wb.e,BCt,null,0,1,vo,!0,!0,!1,!1,!0,!0),oo(f(be(je(e.d),6),35),e.wb.e,"selfloop",null,0,1,vo,!0,!0,!1,!1,!0,!0),oo(f(be(je(e.d),7),35),e.wb.e,"connected",null,0,1,vo,!0,!0,!1,!1,!0,!0),wc(e.b,yf,CCt,!1,!1,!0),oo(f(be(je(e.b),0),35),e.wb.t,"x",OL,1,1,yf,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.b),1),35),e.wb.t,"y",OL,1,1,yf,!1,!1,!0,!1,!0,!1),z=q3(e.b,null,"set"),Ps(z,e.wb.t,"x"),Ps(z,e.wb.t,"y"),wc(e.e,Ii,OCt,!1,!1,!0),oo(f(be(je(e.e),0),35),e.wb.t,"startX",null,0,1,Ii,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.e),1),35),e.wb.t,"startY",null,0,1,Ii,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.e),2),35),e.wb.t,"endX",null,0,1,Ii,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.e),3),35),e.wb.t,"endY",null,0,1,Ii,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.e),4),19),e.b,null,see,0,-1,Ii,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.e),5),19),e.d,f(be(je(e.d),3),19),upe,0,1,Ii,!1,!1,!0,!1,!1,!1,!1),Ai(f(be(je(e.e),6),19),e.c,null,j4e,0,1,Ii,!1,!1,!0,!1,!0,!1,!1),Ai(f(be(je(e.e),7),19),e.c,null,$4e,0,1,Ii,!1,!1,!0,!1,!0,!1,!1),Ai(f(be(je(e.e),8),19),e.e,f(be(je(e.e),9),19),F4e,0,-1,Ii,!1,!1,!0,!1,!0,!1,!1),Ai(f(be(je(e.e),9),19),e.e,f(be(je(e.e),8),19),B4e,0,-1,Ii,!1,!1,!0,!1,!0,!1,!1),oo(f(be(je(e.e),10),35),e.wb._,FCt,null,0,1,Ii,!1,!1,!0,!1,!0,!1),z=q3(e.e,null,"setStartLocation"),Ps(z,e.wb.t,"x"),Ps(z,e.wb.t,"y"),z=q3(e.e,null,"setEndLocation"),Ps(z,e.wb.t,"x"),Ps(z,e.wb.t,"y"),wc(e.k,AE,"ElkPropertyToValueMapEntry",!1,!1,!1),S=f1(e.o),T=(v=(t=new cS,t),v),Qn((!S.d&&(S.d=new So(gs,S,1)),S.d),T),$Et(f(be(je(e.k),0),35),S,"key",AE,!1,!1,!0,!1),oo(f(be(je(e.k),1),35),e.s,k$,null,0,1,AE,!1,!1,!0,!1,!0,!1),Fr(e.o,Fbe,"IProperty",!0),Fr(e.s,Ro,"PropertyValue",!0),CCe(e,WL))}function Vkt(){Vkt=Z,le=Pe($l,hA,30,Xa,15,1),le[9]=35,le[10]=19,le[13]=19,le[32]=51,le[33]=49,le[34]=33,Ne(le,35,38,49),le[38]=1,Ne(le,39,45,49),Ne(le,45,47,-71),le[47]=49,Ne(le,48,58,-71),le[58]=61,le[59]=49,le[60]=1,le[61]=49,le[62]=33,Ne(le,63,65,49),Ne(le,65,91,-3),Ne(le,91,93,33),le[93]=1,le[94]=33,le[95]=-3,le[96]=33,Ne(le,97,123,-3),Ne(le,123,183,33),le[183]=-87,Ne(le,184,192,33),Ne(le,192,215,-19),le[215]=33,Ne(le,216,247,-19),le[247]=33,Ne(le,248,306,-19),Ne(le,306,308,33),Ne(le,308,319,-19),Ne(le,319,321,33),Ne(le,321,329,-19),le[329]=33,Ne(le,330,383,-19),le[383]=33,Ne(le,384,452,-19),Ne(le,452,461,33),Ne(le,461,497,-19),Ne(le,497,500,33),Ne(le,500,502,-19),Ne(le,502,506,33),Ne(le,506,536,-19),Ne(le,536,592,33),Ne(le,592,681,-19),Ne(le,681,699,33),Ne(le,699,706,-19),Ne(le,706,720,33),Ne(le,720,722,-87),Ne(le,722,768,33),Ne(le,768,838,-87),Ne(le,838,864,33),Ne(le,864,866,-87),Ne(le,866,902,33),le[902]=-19,le[903]=-87,Ne(le,904,907,-19),le[907]=33,le[908]=-19,le[909]=33,Ne(le,910,930,-19),le[930]=33,Ne(le,931,975,-19),le[975]=33,Ne(le,976,983,-19),Ne(le,983,986,33),le[986]=-19,le[987]=33,le[988]=-19,le[989]=33,le[990]=-19,le[991]=33,le[992]=-19,le[993]=33,Ne(le,994,1012,-19),Ne(le,1012,1025,33),Ne(le,1025,1037,-19),le[1037]=33,Ne(le,1038,1104,-19),le[1104]=33,Ne(le,1105,1117,-19),le[1117]=33,Ne(le,1118,1154,-19),le[1154]=33,Ne(le,1155,1159,-87),Ne(le,1159,1168,33),Ne(le,1168,1221,-19),Ne(le,1221,1223,33),Ne(le,1223,1225,-19),Ne(le,1225,1227,33),Ne(le,1227,1229,-19),Ne(le,1229,1232,33),Ne(le,1232,1260,-19),Ne(le,1260,1262,33),Ne(le,1262,1270,-19),Ne(le,1270,1272,33),Ne(le,1272,1274,-19),Ne(le,1274,1329,33),Ne(le,1329,1367,-19),Ne(le,1367,1369,33),le[1369]=-19,Ne(le,1370,1377,33),Ne(le,1377,1415,-19),Ne(le,1415,1425,33),Ne(le,1425,1442,-87),le[1442]=33,Ne(le,1443,1466,-87),le[1466]=33,Ne(le,1467,1470,-87),le[1470]=33,le[1471]=-87,le[1472]=33,Ne(le,1473,1475,-87),le[1475]=33,le[1476]=-87,Ne(le,1477,1488,33),Ne(le,1488,1515,-19),Ne(le,1515,1520,33),Ne(le,1520,1523,-19),Ne(le,1523,1569,33),Ne(le,1569,1595,-19),Ne(le,1595,1600,33),le[1600]=-87,Ne(le,1601,1611,-19),Ne(le,1611,1619,-87),Ne(le,1619,1632,33),Ne(le,1632,1642,-87),Ne(le,1642,1648,33),le[1648]=-87,Ne(le,1649,1720,-19),Ne(le,1720,1722,33),Ne(le,1722,1727,-19),le[1727]=33,Ne(le,1728,1743,-19),le[1743]=33,Ne(le,1744,1748,-19),le[1748]=33,le[1749]=-19,Ne(le,1750,1765,-87),Ne(le,1765,1767,-19),Ne(le,1767,1769,-87),le[1769]=33,Ne(le,1770,1774,-87),Ne(le,1774,1776,33),Ne(le,1776,1786,-87),Ne(le,1786,2305,33),Ne(le,2305,2308,-87),le[2308]=33,Ne(le,2309,2362,-19),Ne(le,2362,2364,33),le[2364]=-87,le[2365]=-19,Ne(le,2366,2382,-87),Ne(le,2382,2385,33),Ne(le,2385,2389,-87),Ne(le,2389,2392,33),Ne(le,2392,2402,-19),Ne(le,2402,2404,-87),Ne(le,2404,2406,33),Ne(le,2406,2416,-87),Ne(le,2416,2433,33),Ne(le,2433,2436,-87),le[2436]=33,Ne(le,2437,2445,-19),Ne(le,2445,2447,33),Ne(le,2447,2449,-19),Ne(le,2449,2451,33),Ne(le,2451,2473,-19),le[2473]=33,Ne(le,2474,2481,-19),le[2481]=33,le[2482]=-19,Ne(le,2483,2486,33),Ne(le,2486,2490,-19),Ne(le,2490,2492,33),le[2492]=-87,le[2493]=33,Ne(le,2494,2501,-87),Ne(le,2501,2503,33),Ne(le,2503,2505,-87),Ne(le,2505,2507,33),Ne(le,2507,2510,-87),Ne(le,2510,2519,33),le[2519]=-87,Ne(le,2520,2524,33),Ne(le,2524,2526,-19),le[2526]=33,Ne(le,2527,2530,-19),Ne(le,2530,2532,-87),Ne(le,2532,2534,33),Ne(le,2534,2544,-87),Ne(le,2544,2546,-19),Ne(le,2546,2562,33),le[2562]=-87,Ne(le,2563,2565,33),Ne(le,2565,2571,-19),Ne(le,2571,2575,33),Ne(le,2575,2577,-19),Ne(le,2577,2579,33),Ne(le,2579,2601,-19),le[2601]=33,Ne(le,2602,2609,-19),le[2609]=33,Ne(le,2610,2612,-19),le[2612]=33,Ne(le,2613,2615,-19),le[2615]=33,Ne(le,2616,2618,-19),Ne(le,2618,2620,33),le[2620]=-87,le[2621]=33,Ne(le,2622,2627,-87),Ne(le,2627,2631,33),Ne(le,2631,2633,-87),Ne(le,2633,2635,33),Ne(le,2635,2638,-87),Ne(le,2638,2649,33),Ne(le,2649,2653,-19),le[2653]=33,le[2654]=-19,Ne(le,2655,2662,33),Ne(le,2662,2674,-87),Ne(le,2674,2677,-19),Ne(le,2677,2689,33),Ne(le,2689,2692,-87),le[2692]=33,Ne(le,2693,2700,-19),le[2700]=33,le[2701]=-19,le[2702]=33,Ne(le,2703,2706,-19),le[2706]=33,Ne(le,2707,2729,-19),le[2729]=33,Ne(le,2730,2737,-19),le[2737]=33,Ne(le,2738,2740,-19),le[2740]=33,Ne(le,2741,2746,-19),Ne(le,2746,2748,33),le[2748]=-87,le[2749]=-19,Ne(le,2750,2758,-87),le[2758]=33,Ne(le,2759,2762,-87),le[2762]=33,Ne(le,2763,2766,-87),Ne(le,2766,2784,33),le[2784]=-19,Ne(le,2785,2790,33),Ne(le,2790,2800,-87),Ne(le,2800,2817,33),Ne(le,2817,2820,-87),le[2820]=33,Ne(le,2821,2829,-19),Ne(le,2829,2831,33),Ne(le,2831,2833,-19),Ne(le,2833,2835,33),Ne(le,2835,2857,-19),le[2857]=33,Ne(le,2858,2865,-19),le[2865]=33,Ne(le,2866,2868,-19),Ne(le,2868,2870,33),Ne(le,2870,2874,-19),Ne(le,2874,2876,33),le[2876]=-87,le[2877]=-19,Ne(le,2878,2884,-87),Ne(le,2884,2887,33),Ne(le,2887,2889,-87),Ne(le,2889,2891,33),Ne(le,2891,2894,-87),Ne(le,2894,2902,33),Ne(le,2902,2904,-87),Ne(le,2904,2908,33),Ne(le,2908,2910,-19),le[2910]=33,Ne(le,2911,2914,-19),Ne(le,2914,2918,33),Ne(le,2918,2928,-87),Ne(le,2928,2946,33),Ne(le,2946,2948,-87),le[2948]=33,Ne(le,2949,2955,-19),Ne(le,2955,2958,33),Ne(le,2958,2961,-19),le[2961]=33,Ne(le,2962,2966,-19),Ne(le,2966,2969,33),Ne(le,2969,2971,-19),le[2971]=33,le[2972]=-19,le[2973]=33,Ne(le,2974,2976,-19),Ne(le,2976,2979,33),Ne(le,2979,2981,-19),Ne(le,2981,2984,33),Ne(le,2984,2987,-19),Ne(le,2987,2990,33),Ne(le,2990,2998,-19),le[2998]=33,Ne(le,2999,3002,-19),Ne(le,3002,3006,33),Ne(le,3006,3011,-87),Ne(le,3011,3014,33),Ne(le,3014,3017,-87),le[3017]=33,Ne(le,3018,3022,-87),Ne(le,3022,3031,33),le[3031]=-87,Ne(le,3032,3047,33),Ne(le,3047,3056,-87),Ne(le,3056,3073,33),Ne(le,3073,3076,-87),le[3076]=33,Ne(le,3077,3085,-19),le[3085]=33,Ne(le,3086,3089,-19),le[3089]=33,Ne(le,3090,3113,-19),le[3113]=33,Ne(le,3114,3124,-19),le[3124]=33,Ne(le,3125,3130,-19),Ne(le,3130,3134,33),Ne(le,3134,3141,-87),le[3141]=33,Ne(le,3142,3145,-87),le[3145]=33,Ne(le,3146,3150,-87),Ne(le,3150,3157,33),Ne(le,3157,3159,-87),Ne(le,3159,3168,33),Ne(le,3168,3170,-19),Ne(le,3170,3174,33),Ne(le,3174,3184,-87),Ne(le,3184,3202,33),Ne(le,3202,3204,-87),le[3204]=33,Ne(le,3205,3213,-19),le[3213]=33,Ne(le,3214,3217,-19),le[3217]=33,Ne(le,3218,3241,-19),le[3241]=33,Ne(le,3242,3252,-19),le[3252]=33,Ne(le,3253,3258,-19),Ne(le,3258,3262,33),Ne(le,3262,3269,-87),le[3269]=33,Ne(le,3270,3273,-87),le[3273]=33,Ne(le,3274,3278,-87),Ne(le,3278,3285,33),Ne(le,3285,3287,-87),Ne(le,3287,3294,33),le[3294]=-19,le[3295]=33,Ne(le,3296,3298,-19),Ne(le,3298,3302,33),Ne(le,3302,3312,-87),Ne(le,3312,3330,33),Ne(le,3330,3332,-87),le[3332]=33,Ne(le,3333,3341,-19),le[3341]=33,Ne(le,3342,3345,-19),le[3345]=33,Ne(le,3346,3369,-19),le[3369]=33,Ne(le,3370,3386,-19),Ne(le,3386,3390,33),Ne(le,3390,3396,-87),Ne(le,3396,3398,33),Ne(le,3398,3401,-87),le[3401]=33,Ne(le,3402,3406,-87),Ne(le,3406,3415,33),le[3415]=-87,Ne(le,3416,3424,33),Ne(le,3424,3426,-19),Ne(le,3426,3430,33),Ne(le,3430,3440,-87),Ne(le,3440,3585,33),Ne(le,3585,3631,-19),le[3631]=33,le[3632]=-19,le[3633]=-87,Ne(le,3634,3636,-19),Ne(le,3636,3643,-87),Ne(le,3643,3648,33),Ne(le,3648,3654,-19),Ne(le,3654,3663,-87),le[3663]=33,Ne(le,3664,3674,-87),Ne(le,3674,3713,33),Ne(le,3713,3715,-19),le[3715]=33,le[3716]=-19,Ne(le,3717,3719,33),Ne(le,3719,3721,-19),le[3721]=33,le[3722]=-19,Ne(le,3723,3725,33),le[3725]=-19,Ne(le,3726,3732,33),Ne(le,3732,3736,-19),le[3736]=33,Ne(le,3737,3744,-19),le[3744]=33,Ne(le,3745,3748,-19),le[3748]=33,le[3749]=-19,le[3750]=33,le[3751]=-19,Ne(le,3752,3754,33),Ne(le,3754,3756,-19),le[3756]=33,Ne(le,3757,3759,-19),le[3759]=33,le[3760]=-19,le[3761]=-87,Ne(le,3762,3764,-19),Ne(le,3764,3770,-87),le[3770]=33,Ne(le,3771,3773,-87),le[3773]=-19,Ne(le,3774,3776,33),Ne(le,3776,3781,-19),le[3781]=33,le[3782]=-87,le[3783]=33,Ne(le,3784,3790,-87),Ne(le,3790,3792,33),Ne(le,3792,3802,-87),Ne(le,3802,3864,33),Ne(le,3864,3866,-87),Ne(le,3866,3872,33),Ne(le,3872,3882,-87),Ne(le,3882,3893,33),le[3893]=-87,le[3894]=33,le[3895]=-87,le[3896]=33,le[3897]=-87,Ne(le,3898,3902,33),Ne(le,3902,3904,-87),Ne(le,3904,3912,-19),le[3912]=33,Ne(le,3913,3946,-19),Ne(le,3946,3953,33),Ne(le,3953,3973,-87),le[3973]=33,Ne(le,3974,3980,-87),Ne(le,3980,3984,33),Ne(le,3984,3990,-87),le[3990]=33,le[3991]=-87,le[3992]=33,Ne(le,3993,4014,-87),Ne(le,4014,4017,33),Ne(le,4017,4024,-87),le[4024]=33,le[4025]=-87,Ne(le,4026,4256,33),Ne(le,4256,4294,-19),Ne(le,4294,4304,33),Ne(le,4304,4343,-19),Ne(le,4343,4352,33),le[4352]=-19,le[4353]=33,Ne(le,4354,4356,-19),le[4356]=33,Ne(le,4357,4360,-19),le[4360]=33,le[4361]=-19,le[4362]=33,Ne(le,4363,4365,-19),le[4365]=33,Ne(le,4366,4371,-19),Ne(le,4371,4412,33),le[4412]=-19,le[4413]=33,le[4414]=-19,le[4415]=33,le[4416]=-19,Ne(le,4417,4428,33),le[4428]=-19,le[4429]=33,le[4430]=-19,le[4431]=33,le[4432]=-19,Ne(le,4433,4436,33),Ne(le,4436,4438,-19),Ne(le,4438,4441,33),le[4441]=-19,Ne(le,4442,4447,33),Ne(le,4447,4450,-19),le[4450]=33,le[4451]=-19,le[4452]=33,le[4453]=-19,le[4454]=33,le[4455]=-19,le[4456]=33,le[4457]=-19,Ne(le,4458,4461,33),Ne(le,4461,4463,-19),Ne(le,4463,4466,33),Ne(le,4466,4468,-19),le[4468]=33,le[4469]=-19,Ne(le,4470,4510,33),le[4510]=-19,Ne(le,4511,4520,33),le[4520]=-19,Ne(le,4521,4523,33),le[4523]=-19,Ne(le,4524,4526,33),Ne(le,4526,4528,-19),Ne(le,4528,4535,33),Ne(le,4535,4537,-19),le[4537]=33,le[4538]=-19,le[4539]=33,Ne(le,4540,4547,-19),Ne(le,4547,4587,33),le[4587]=-19,Ne(le,4588,4592,33),le[4592]=-19,Ne(le,4593,4601,33),le[4601]=-19,Ne(le,4602,7680,33),Ne(le,7680,7836,-19),Ne(le,7836,7840,33),Ne(le,7840,7930,-19),Ne(le,7930,7936,33),Ne(le,7936,7958,-19),Ne(le,7958,7960,33),Ne(le,7960,7966,-19),Ne(le,7966,7968,33),Ne(le,7968,8006,-19),Ne(le,8006,8008,33),Ne(le,8008,8014,-19),Ne(le,8014,8016,33),Ne(le,8016,8024,-19),le[8024]=33,le[8025]=-19,le[8026]=33,le[8027]=-19,le[8028]=33,le[8029]=-19,le[8030]=33,Ne(le,8031,8062,-19),Ne(le,8062,8064,33),Ne(le,8064,8117,-19),le[8117]=33,Ne(le,8118,8125,-19),le[8125]=33,le[8126]=-19,Ne(le,8127,8130,33),Ne(le,8130,8133,-19),le[8133]=33,Ne(le,8134,8141,-19),Ne(le,8141,8144,33),Ne(le,8144,8148,-19),Ne(le,8148,8150,33),Ne(le,8150,8156,-19),Ne(le,8156,8160,33),Ne(le,8160,8173,-19),Ne(le,8173,8178,33),Ne(le,8178,8181,-19),le[8181]=33,Ne(le,8182,8189,-19),Ne(le,8189,8400,33),Ne(le,8400,8413,-87),Ne(le,8413,8417,33),le[8417]=-87,Ne(le,8418,8486,33),le[8486]=-19,Ne(le,8487,8490,33),Ne(le,8490,8492,-19),Ne(le,8492,8494,33),le[8494]=-19,Ne(le,8495,8576,33),Ne(le,8576,8579,-19),Ne(le,8579,12293,33),le[12293]=-87,le[12294]=33,le[12295]=-19,Ne(le,12296,12321,33),Ne(le,12321,12330,-19),Ne(le,12330,12336,-87),le[12336]=33,Ne(le,12337,12342,-87),Ne(le,12342,12353,33),Ne(le,12353,12437,-19),Ne(le,12437,12441,33),Ne(le,12441,12443,-87),Ne(le,12443,12445,33),Ne(le,12445,12447,-87),Ne(le,12447,12449,33),Ne(le,12449,12539,-19),le[12539]=33,Ne(le,12540,12543,-87),Ne(le,12543,12549,33),Ne(le,12549,12589,-19),Ne(le,12589,19968,33),Ne(le,19968,40870,-19),Ne(le,40870,44032,33),Ne(le,44032,55204,-19),Ne(le,55204,GG,33),Ne(le,57344,65534,33)}function U7n(e){var t,i,a,u,l,g,m;e.hb||(e.hb=!0,ju(e,"ecore"),WJ(e,"ecore"),KJ(e,bh),WH(e.fb,"E"),WH(e.L,"T"),WH(e.P,"K"),WH(e.P,"V"),WH(e.cb,"E"),Qn(Bs(e.b),e.bb),Qn(Bs(e.a),e.Q),Qn(Bs(e.o),e.p),Qn(Bs(e.p),e.R),Qn(Bs(e.q),e.p),Qn(Bs(e.v),e.q),Qn(Bs(e.w),e.R),Qn(Bs(e.B),e.Q),Qn(Bs(e.R),e.Q),Qn(Bs(e.T),e.eb),Qn(Bs(e.U),e.R),Qn(Bs(e.V),e.eb),Qn(Bs(e.W),e.bb),Qn(Bs(e.bb),e.eb),Qn(Bs(e.eb),e.R),Qn(Bs(e.db),e.R),wc(e.b,wx,O3t,!1,!1,!0),oo(f(be(je(e.b),0),35),e.e,"iD",null,0,1,wx,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.b),1),19),e.q,null,"eAttributeType",1,1,wx,!0,!0,!1,!1,!0,!1,!0),wc(e.a,Ir,C3t,!1,!1,!0),oo(f(be(je(e.a),0),35),e._,ape,null,0,1,Ir,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.a),1),19),e.ab,null,"details",0,-1,Ir,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.a),2),19),e.Q,f(be(je(e.Q),0),19),"eModelElement",0,1,Ir,!0,!1,!0,!1,!1,!1,!1),Ai(f(be(je(e.a),3),19),e.S,null,"contents",0,-1,Ir,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.a),4),19),e.S,null,"references",0,-1,Ir,!1,!1,!0,!1,!0,!1,!1),wc(e.o,kh,"EClass",!1,!1,!0),oo(f(be(je(e.o),0),35),e.e,"abstract",null,0,1,kh,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.o),1),35),e.e,"interface",null,0,1,kh,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.o),2),19),e.o,null,"eSuperTypes",0,-1,kh,!1,!1,!0,!1,!0,!0,!1),Ai(f(be(je(e.o),3),19),e.T,f(be(je(e.T),0),19),"eOperations",0,-1,kh,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.o),4),19),e.b,null,"eAllAttributes",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),5),19),e.W,null,"eAllReferences",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),6),19),e.W,null,"eReferences",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),7),19),e.b,null,"eAttributes",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),8),19),e.W,null,"eAllContainments",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),9),19),e.T,null,"eAllOperations",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),10),19),e.bb,null,"eAllStructuralFeatures",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),11),19),e.o,null,"eAllSuperTypes",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.o),12),19),e.b,null,"eIDAttribute",0,1,kh,!0,!0,!1,!1,!1,!1,!0),Ai(f(be(je(e.o),13),19),e.bb,f(be(je(e.bb),7),19),"eStructuralFeatures",0,-1,kh,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.o),14),19),e.H,null,"eGenericSuperTypes",0,-1,kh,!1,!1,!0,!0,!1,!0,!1),Ai(f(be(je(e.o),15),19),e.H,null,"eAllGenericSuperTypes",0,-1,kh,!0,!0,!1,!1,!0,!1,!0),m=Ss(f(be(ci(e.o),0),62),e.e,"isSuperTypeOf"),Ps(m,e.o,"someClass"),Ss(f(be(ci(e.o),1),62),e.I,"getFeatureCount"),m=Ss(f(be(ci(e.o),2),62),e.bb,z3t),Ps(m,e.I,"featureID"),m=Ss(f(be(ci(e.o),3),62),e.I,H3t),Ps(m,e.bb,R$),m=Ss(f(be(ci(e.o),4),62),e.bb,z3t),Ps(m,e._,"featureName"),Ss(f(be(ci(e.o),5),62),e.I,"getOperationCount"),m=Ss(f(be(ci(e.o),6),62),e.T,"getEOperation"),Ps(m,e.I,"operationID"),m=Ss(f(be(ci(e.o),7),62),e.I,G3t),Ps(m,e.T,e6e),m=Ss(f(be(ci(e.o),8),62),e.T,"getOverride"),Ps(m,e.T,e6e),m=Ss(f(be(ci(e.o),9),62),e.H,"getFeatureType"),Ps(m,e.bb,R$),wc(e.p,Up,N3t,!0,!1,!0),oo(f(be(je(e.p),0),35),e._,"instanceClassName",null,0,1,Up,!1,!0,!0,!0,!0,!1),t=f1(e.L),i=xTe(),Qn((!t.d&&(t.d=new So(gs,t,1)),t.d),i),$Et(f(be(je(e.p),1),35),t,"instanceClass",Up,!0,!0,!1,!0),oo(f(be(je(e.p),2),35),e.M,q3t,null,0,1,Up,!0,!0,!1,!1,!0,!0),oo(f(be(je(e.p),3),35),e._,"instanceTypeName",null,0,1,Up,!1,!0,!0,!0,!0,!1),Ai(f(be(je(e.p),4),19),e.U,f(be(je(e.U),3),19),"ePackage",0,1,Up,!0,!1,!1,!1,!0,!1,!1),Ai(f(be(je(e.p),5),19),e.db,null,V3t,0,-1,Up,!1,!1,!0,!0,!0,!1,!1),m=Ss(f(be(ci(e.p),0),62),e.e,W3t),Ps(m,e.M,DG),Ss(f(be(ci(e.p),1),62),e.I,"getClassifierID"),wc(e.q,m1e,"EDataType",!1,!1,!0),oo(f(be(je(e.q),0),35),e.e,"serializable",qL,0,1,m1e,!1,!1,!0,!1,!0,!1),wc(e.v,Une,"EEnum",!1,!1,!0),Ai(f(be(je(e.v),0),19),e.w,f(be(je(e.w),3),19),"eLiterals",0,-1,Une,!1,!1,!0,!0,!1,!1,!1),m=Ss(f(be(ci(e.v),0),62),e.w,K3t),Ps(m,e._,Gr),m=Ss(f(be(ci(e.v),1),62),e.w,K3t),Ps(m,e.I,k$),m=Ss(f(be(ci(e.v),2),62),e.w,"getEEnumLiteralByLiteral"),Ps(m,e._,"literal"),wc(e.w,j1,R3t,!1,!1,!0),oo(f(be(je(e.w),0),35),e.I,k$,null,0,1,j1,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.w),1),35),e.A,"instance",null,0,1,j1,!0,!1,!0,!1,!0,!1),oo(f(be(je(e.w),2),35),e._,"literal",null,0,1,j1,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.w),3),19),e.v,f(be(je(e.v),0),19),"eEnum",0,1,j1,!0,!1,!1,!1,!1,!1,!1),wc(e.B,BF,"EFactory",!1,!1,!0),Ai(f(be(je(e.B),0),19),e.U,f(be(je(e.U),2),19),"ePackage",1,1,BF,!0,!1,!0,!1,!1,!1,!1),m=Ss(f(be(ci(e.B),0),62),e.S,"create"),Ps(m,e.o,"eClass"),m=Ss(f(be(ci(e.B),1),62),e.M,"createFromString"),Ps(m,e.q,"eDataType"),Ps(m,e._,"literalValue"),m=Ss(f(be(ci(e.B),2),62),e._,"convertToString"),Ps(m,e.q,"eDataType"),Ps(m,e.M,"instanceValue"),wc(e.Q,EV,NCt,!0,!1,!0),Ai(f(be(je(e.Q),0),19),e.a,f(be(je(e.a),2),19),"eAnnotations",0,-1,EV,!1,!1,!0,!0,!1,!1,!1),m=Ss(f(be(ci(e.Q),0),62),e.a,"getEAnnotation"),Ps(m,e._,ape),wc(e.R,l1e,RCt,!0,!1,!0),oo(f(be(je(e.R),0),35),e._,Gr,null,0,1,l1e,!1,!1,!0,!1,!0,!1),wc(e.S,Dw,"EObject",!1,!1,!0),Ss(f(be(ci(e.S),0),62),e.o,"eClass"),Ss(f(be(ci(e.S),1),62),e.e,"eIsProxy"),Ss(f(be(ci(e.S),2),62),e.X,"eResource"),Ss(f(be(ci(e.S),3),62),e.S,"eContainer"),Ss(f(be(ci(e.S),4),62),e.bb,"eContainingFeature"),Ss(f(be(ci(e.S),5),62),e.W,"eContainmentFeature"),m=Ss(f(be(ci(e.S),6),62),null,"eContents"),t=f1(e.fb),i=f1(e.S),Qn((!t.d&&(t.d=new So(gs,t,1)),t.d),i),u=xg(m,t,null),u&&u.lj(),m=Ss(f(be(ci(e.S),7),62),null,"eAllContents"),t=f1(e.cb),i=f1(e.S),Qn((!t.d&&(t.d=new So(gs,t,1)),t.d),i),l=xg(m,t,null),l&&l.lj(),m=Ss(f(be(ci(e.S),8),62),null,"eCrossReferences"),t=f1(e.fb),i=f1(e.S),Qn((!t.d&&(t.d=new So(gs,t,1)),t.d),i),g=xg(m,t,null),g&&g.lj(),m=Ss(f(be(ci(e.S),9),62),e.M,"eGet"),Ps(m,e.bb,R$),m=Ss(f(be(ci(e.S),10),62),e.M,"eGet"),Ps(m,e.bb,R$),Ps(m,e.e,"resolve"),m=Ss(f(be(ci(e.S),11),62),null,"eSet"),Ps(m,e.bb,R$),Ps(m,e.M,"newValue"),m=Ss(f(be(ci(e.S),12),62),e.e,"eIsSet"),Ps(m,e.bb,R$),m=Ss(f(be(ci(e.S),13),62),null,"eUnset"),Ps(m,e.bb,R$),m=Ss(f(be(ci(e.S),14),62),e.M,"eInvoke"),Ps(m,e.T,e6e),t=f1(e.fb),i=xTe(),Qn((!t.d&&(t.d=new So(gs,t,1)),t.d),i),MZ(m,t,"arguments"),Mwn(m,e.K),wc(e.T,Th,L3t,!1,!1,!0),Ai(f(be(je(e.T),0),19),e.o,f(be(je(e.o),3),19),Y3t,0,1,Th,!0,!1,!1,!1,!1,!1,!1),Ai(f(be(je(e.T),1),19),e.db,null,V3t,0,-1,Th,!1,!1,!0,!0,!0,!1,!1),Ai(f(be(je(e.T),2),19),e.V,f(be(je(e.V),0),19),"eParameters",0,-1,Th,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.T),3),19),e.p,null,"eExceptions",0,-1,Th,!1,!1,!0,!1,!0,!0,!1),Ai(f(be(je(e.T),4),19),e.H,null,"eGenericExceptions",0,-1,Th,!1,!1,!0,!0,!1,!0,!1),Ss(f(be(ci(e.T),0),62),e.I,G3t),m=Ss(f(be(ci(e.T),1),62),e.e,"isOverrideOf"),Ps(m,e.T,"someOperation"),wc(e.U,Bp,"EPackage",!1,!1,!0),oo(f(be(je(e.U),0),35),e._,"nsURI",null,0,1,Bp,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.U),1),35),e._,"nsPrefix",null,0,1,Bp,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.U),2),19),e.B,f(be(je(e.B),0),19),"eFactoryInstance",1,1,Bp,!0,!1,!0,!1,!1,!1,!1),Ai(f(be(je(e.U),3),19),e.p,f(be(je(e.p),4),19),"eClassifiers",0,-1,Bp,!1,!1,!0,!0,!0,!1,!1),Ai(f(be(je(e.U),4),19),e.U,f(be(je(e.U),5),19),"eSubpackages",0,-1,Bp,!1,!1,!0,!0,!0,!1,!1),Ai(f(be(je(e.U),5),19),e.U,f(be(je(e.U),4),19),"eSuperPackage",0,1,Bp,!0,!1,!1,!1,!0,!1,!1),m=Ss(f(be(ci(e.U),0),62),e.p,"getEClassifier"),Ps(m,e._,Gr),wc(e.V,S_,D3t,!1,!1,!0),Ai(f(be(je(e.V),0),19),e.T,f(be(je(e.T),2),19),"eOperation",0,1,S_,!0,!1,!1,!1,!1,!1,!1),wc(e.W,yx,M3t,!1,!1,!0),oo(f(be(je(e.W),0),35),e.e,"containment",null,0,1,yx,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.W),1),35),e.e,"container",null,0,1,yx,!0,!0,!1,!1,!0,!0),oo(f(be(je(e.W),2),35),e.e,"resolveProxies",qL,0,1,yx,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.W),3),19),e.W,null,"eOpposite",0,1,yx,!1,!1,!0,!1,!0,!1,!1),Ai(f(be(je(e.W),4),19),e.o,null,"eReferenceType",1,1,yx,!0,!0,!1,!1,!0,!1,!0),Ai(f(be(je(e.W),5),19),e.b,null,"eKeys",0,-1,yx,!1,!1,!0,!1,!0,!1,!1),wc(e.bb,El,I3t,!0,!1,!0),oo(f(be(je(e.bb),0),35),e.e,"changeable",qL,0,1,El,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.bb),1),35),e.e,"volatile",null,0,1,El,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.bb),2),35),e.e,"transient",null,0,1,El,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.bb),3),35),e._,"defaultValueLiteral",null,0,1,El,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.bb),4),35),e.M,q3t,null,0,1,El,!0,!0,!1,!1,!0,!0),oo(f(be(je(e.bb),5),35),e.e,"unsettable",null,0,1,El,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.bb),6),35),e.e,"derived",null,0,1,El,!1,!1,!0,!1,!0,!1),Ai(f(be(je(e.bb),7),19),e.o,f(be(je(e.o),13),19),Y3t,0,1,El,!0,!1,!1,!1,!1,!1,!1),Ss(f(be(ci(e.bb),0),62),e.I,H3t),m=Ss(f(be(ci(e.bb),1),62),null,"getContainerClass"),t=f1(e.L),i=xTe(),Qn((!t.d&&(t.d=new So(gs,t,1)),t.d),i),a=xg(m,t,null),a&&a.lj(),wc(e.eb,vx,A3t,!0,!1,!0),oo(f(be(je(e.eb),0),35),e.e,"ordered",qL,0,1,vx,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.eb),1),35),e.e,"unique",qL,0,1,vx,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.eb),2),35),e.I,"lowerBound",null,0,1,vx,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.eb),3),35),e.I,"upperBound","1",0,1,vx,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.eb),4),35),e.e,"many",null,0,1,vx,!0,!0,!1,!1,!0,!0),oo(f(be(je(e.eb),5),35),e.e,"required",null,0,1,vx,!0,!0,!1,!1,!0,!0),Ai(f(be(je(e.eb),6),19),e.p,null,"eType",0,1,vx,!1,!0,!0,!1,!0,!0,!1),Ai(f(be(je(e.eb),7),19),e.H,null,"eGenericType",0,1,vx,!1,!0,!0,!0,!1,!0,!1),wc(e.ab,AE,"EStringToStringMapEntry",!1,!1,!1),oo(f(be(je(e.ab),0),35),e._,"key",null,0,1,AE,!1,!1,!0,!1,!0,!1),oo(f(be(je(e.ab),1),35),e._,k$,null,0,1,AE,!1,!1,!0,!1,!0,!1),wc(e.H,gs,P3t,!1,!1,!0),Ai(f(be(je(e.H),0),19),e.H,null,"eUpperBound",0,1,gs,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.H),1),19),e.H,null,"eTypeArguments",0,-1,gs,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.H),2),19),e.p,null,"eRawType",1,1,gs,!0,!1,!1,!1,!0,!1,!0),Ai(f(be(je(e.H),3),19),e.H,null,"eLowerBound",0,1,gs,!1,!1,!0,!0,!1,!1,!1),Ai(f(be(je(e.H),4),19),e.db,null,"eTypeParameter",0,1,gs,!1,!1,!0,!1,!1,!1,!1),Ai(f(be(je(e.H),5),19),e.p,null,"eClassifier",0,1,gs,!1,!1,!0,!1,!0,!1,!1),m=Ss(f(be(ci(e.H),0),62),e.e,W3t),Ps(m,e.M,DG),wc(e.db,nl,j3t,!1,!1,!0),Ai(f(be(je(e.db),0),19),e.H,null,"eBounds",0,-1,nl,!1,!1,!0,!0,!1,!1,!1),Fr(e.c,Vpe,"EBigDecimal",!0),Fr(e.d,mA,"EBigInteger",!0),Fr(e.e,kd,"EBoolean",!0),Fr(e.f,to,"EBooleanObject",!0),Fr(e.i,$l,"EByte",!0),Fr(e.g,se($l,1),"EByteArray",!0),Fr(e.j,BR,"EByteObject",!0),Fr(e.k,Xf,"EChar",!0),Fr(e.n,D$,"ECharacterObject",!0),Fr(e.r,Iee,"EDate",!0),Fr(e.s,fje,"EDiagnosticChain",!1),Fr(e.t,sa,"EDouble",!0),Fr(e.u,mo,"EDoubleObject",!0),Fr(e.fb,C7e,"EEList",!1),Fr(e.A,O7e,"EEnumerator",!1),Fr(e.C,Q7e,"EFeatureMap",!1),Fr(e.D,zne,"EFeatureMapEntry",!1),Fr(e.F,kx,"EFloat",!0),Fr(e.G,eD,"EFloatObject",!0),Fr(e.I,pr,"EInt",!0),Fr(e.J,ba,"EIntegerObject",!0),Fr(e.L,d6e,"EJavaClass",!0),Fr(e.M,Ro,"EJavaObject",!0),Fr(e.N,x_,"ELong",!0),Fr(e.O,i_,"ELongObject",!0),Fr(e.P,f6e,"EMap",!1),Fr(e.X,Y7e,"EResource",!1),Fr(e.Y,hje,"EResourceSet",!1),Fr(e.Z,JA,"EShort",!0),Fr(e.$,o_,"EShortObject",!0),Fr(e._,ft,"EString",!0),Fr(e.cb,I7e,"ETreeIterator",!1),Fr(e.K,pje,"EInvocationTargetException",!1),CCe(e,bh))}var DG="object",xR="boolean",AOe="number",sfe="string",cfe="function",Qr=2147483647,Sc="java.lang",MG={3:1},jG="com.google.common.base",Fu=", ",Wkt="%s (%s) must not be negative",Jt={3:1,4:1,5:1},Kkt="negative size: ",IOe="no calls to next() since the last call to remove()",Ykt="Optional.of(",bl="null",TL={204:1,49:1},Rt="com.google.common.collect",xL={204:1,49:1,127:1},wE={228:1,3:1},aa={49:1},Gn="java.util",YS={91:1},jT={20:1,31:1,16:1},Ig=2009,Pl={20:1,31:1,16:1,22:1},OOe={91:1,137:1,134:1},Xkt={20:1,31:1,16:1,22:1,83:1},NOe={20:1,31:1,16:1,274:1,22:1,83:1},yb={49:1,127:1},ufe={355:1,43:1},Jkt="AbstractMapEntry",Zkt="expectedValuesPerKey",Ye={3:1,6:1,4:1,5:1},Sm=16384,pf={158:1},xn={40:1},$G={202:1},FG={l:4194303,m:4194303,h:524287},lfe={252:1,3:1,34:1},Qkt="range unbounded on this side",Eb={20:1},eTt={20:1,16:1},ROe={3:1,20:1,31:1,16:1},CL={308:1,3:1,20:1,31:1,16:1,15:1,59:1},yQ={3:1,4:1,5:1,174:1},AL={3:1,91:1},dfe={20:1,16:1,22:1},CR={3:1,20:1,31:1,16:1,22:1},tTt={20:1,16:1,22:1,83:1},Sb=461845907,_b=-862048943,BG={3:1,6:1,4:1,5:1,174:1},nTt="expectedSize",n$=1073741824,$T="initialArraySize",Ve={3:1,6:1,4:1,10:1,5:1},IL={20:1,31:1,55:1,16:1,15:1},ffe="arraySize",rTt={20:1,31:1,55:1,16:1,15:1,59:1},_r={48:1},EQ={372:1},fw=1e-4,Ra=-2147483648,iTt="__noinit__",C1={3:1,103:1,63:1,80:1},UG="com.google.gwt.core.client.impl",POe="String",LOe="com.google.gwt.core.client",hfe="anonymous",pfe="fnStack",DOe="Unknown",_m={201:1,3:1,4:1},M0=1e3,To=65535,gfe="January",mfe="February",bfe="March",vfe="April",AR="May",wfe="June",yfe="July",Efe="August",Sfe="September",_fe="October",kfe="November",Tfe="December",hw=1900,Pr={54:1,3:1,4:1},oTt="Before Christ",aTt="Anno Domini",xfe="Sunday",Cfe="Monday",Afe="Tuesday",Ife="Wednesday",Ofe="Thursday",Nfe="Friday",Rfe="Saturday",MOe="com.google.gwt.i18n.shared",sTt="DateTimeFormat",Pfe="com.google.gwt.i18n.client",cTt="DefaultDateTimeFormatInfo",pw=2048,uTt={3:1,4:1,34:1,205:1},IR="com.google.gwt.json.client",Ed=4194303,A1=1048575,zG=524288,OR=4194304,yE=17592186044416,SQ=1e9,HG=-17592186044416,jOe="java.io",_Q={3:1,103:1,97:1,63:1,80:1},lTt={3:1,294:1,80:1},XS='For input string: "',Gi=1/0,qo=-1/0,FT=4096,dTt={3:1,4:1,579:1},wn="org.eclipse.elk.layered",Xa=65536,GG=55296,_s={108:1,3:1,4:1},Lfe=1e5,fTt=.3010299956639812,ls=4294967295,OL="0.0",Dfe={43:1},NL="Unable to add element to queue",hTt={3:1,4:1,20:1,31:1,55:1,13:1,16:1,15:1,59:1},pTt={3:1,20:1,31:1,55:1,16:1,15:1,59:1},gTt={20:1,16:1,15:1},Mfe={3:1,51:1},qG={189:1},sA={3:1,4:1,91:1},$Oe={3:1,4:1,20:1,31:1,16:1,50:1,22:1},jfe="delete",r$=14901161193847656e-24,i$=11102230246251565e-32,$fe=15525485,VG=5960464477539063e-23,FOe=16777216,kQ=16777215,BOe=", length: ",mTt={3:1,4:1,20:1,31:1,55:1,16:1,15:1,59:1},UOe="subMap: ",bTt=" less than ",Ffe={3:1,34:1,23:1,305:1},Bfe="java.util.function",o$="java.util.logging",vTt={3:1,4:1,5:1,835:1},Ufe="undefined",ds="java.util.stream",zOe={518:1,672:1},TQ="fromIndex: ",wTt=" > toIndex: ",HOe=", toIndex: ",GOe="Index: ",qOe=", Size: ",RL="org.eclipse.elk.alg.common",Cr={51:1},yTt="org.eclipse.elk.alg.common.compaction",ETt="Scanline/EventHandler",kb="org.eclipse.elk.alg.common.compaction.oned",STt="CNode belongs to another CGroup.",_Tt="ISpacingsHandler/1",zfe="The ",Hfe=" instance has been finished already.",kTt="The direction ",TTt=" is not supported by the CGraph instance.",xTt="OneDimensionalCompactor",CTt="OneDimensionalCompactor/lambda$0$Type",ATt="Quadruplet",ITt="ScanlineConstraintCalculator",OTt="ScanlineConstraintCalculator/ConstraintsScanlineHandler",NTt="ScanlineConstraintCalculator/ConstraintsScanlineHandler/lambda$0$Type",RTt="ScanlineConstraintCalculator/Timestamp",PTt="ScanlineConstraintCalculator/lambda$0$Type",km={177:1,48:1},a$="org.eclipse.elk.alg.common.networksimplex",Tb={183:1,3:1,4:1},LTt="org.eclipse.elk.alg.common.nodespacing",EE="org.eclipse.elk.alg.common.nodespacing.cellsystem",PL="CENTER",DTt={215:1,332:1},VOe={3:1,4:1,5:1,587:1},NR="LEFT",RR="RIGHT",WOe="Vertical alignment cannot be null",KOe="BOTTOM",xQ="org.eclipse.elk.alg.common.nodespacing.internal",s$="UNDEFINED",Og=.01,WG="org.eclipse.elk.alg.common.nodespacing.internal.algorithm",MTt="LabelPlacer/lambda$0$Type",jTt="LabelPlacer/lambda$1$Type",$Tt="portRatioOrPosition",LL="org.eclipse.elk.alg.common.overlaps",Gfe="DOWN",PR="org.eclipse.elk.alg.common.spore",BT={3:1,4:1,5:1,198:1},FTt={3:1,6:1,4:1,5:1,99:1,114:1},qfe="org.eclipse.elk.alg.force",YOe="ComponentsProcessor",BTt="ComponentsProcessor/1",XOe="ElkGraphImporter/lambda$0$Type",JS={222:1},cA="org.eclipse.elk.core",KG="org.eclipse.elk.graph.properties",UTt="IPropertyHolder",YG="org.eclipse.elk.alg.force.graph",zTt="Component Layout",JOe="org.eclipse.elk.alg.force.model",dc="org.eclipse.elk.core.data",CQ="org.eclipse.elk.force.model",ZOe="org.eclipse.elk.force.iterations",QOe="org.eclipse.elk.force.repulsivePower",Vfe="org.eclipse.elk.force.temperature",Tm=.001,Wfe="org.eclipse.elk.force.repulsion",Ng={146:1},c$="org.eclipse.elk.alg.force.options",DL=1.600000023841858,Zu="org.eclipse.elk.force",XG="org.eclipse.elk.priority",UT="org.eclipse.elk.spacing.nodeNode",Kfe="org.eclipse.elk.spacing.edgeLabel",ML="org.eclipse.elk.aspectRatio",AQ="org.eclipse.elk.randomSeed",u$="org.eclipse.elk.separateConnectedComponents",zT="org.eclipse.elk.padding",l$="org.eclipse.elk.interactive",Yfe="org.eclipse.elk.portConstraints",IQ="org.eclipse.elk.edgeLabels.inline",d$="org.eclipse.elk.omitNodeMicroLayout",jL="org.eclipse.elk.nodeSize.fixedGraphSize",LR="org.eclipse.elk.nodeSize.options",uA="org.eclipse.elk.nodeSize.constraints",$L="org.eclipse.elk.nodeLabels.placement",FL="org.eclipse.elk.portLabels.placement",JG="org.eclipse.elk.topdownLayout",ZG="org.eclipse.elk.topdown.scaleFactor",QG="org.eclipse.elk.topdown.hierarchicalNodeWidth",eq="org.eclipse.elk.topdown.hierarchicalNodeAspectRatio",ZS="org.eclipse.elk.topdown.nodeType",eNe="origin",HTt="random",GTt="boundingBox.upLeft",qTt="boundingBox.lowRight",tNe="org.eclipse.elk.stress.fixed",nNe="org.eclipse.elk.stress.desiredEdgeLength",rNe="org.eclipse.elk.stress.dimension",iNe="org.eclipse.elk.stress.epsilon",oNe="org.eclipse.elk.stress.iterationLimit",gw="org.eclipse.elk.stress",VTt="ELK Stress",DR="org.eclipse.elk.nodeSize.minimum",OQ="org.eclipse.elk.alg.force.stress",WTt="Layered layout",MR="org.eclipse.elk.alg.layered",tq="org.eclipse.elk.alg.layered.compaction.components",f$="org.eclipse.elk.alg.layered.compaction.oned",NQ="org.eclipse.elk.alg.layered.compaction.oned.algs",SE="org.eclipse.elk.alg.layered.compaction.recthull",Rg="org.eclipse.elk.alg.layered.components",Lp="NONE",aNe="MODEL_ORDER",qc={3:1,6:1,4:1,10:1,5:1,125:1},KTt={3:1,6:1,4:1,5:1,145:1,99:1,114:1},RQ="org.eclipse.elk.alg.layered.compound",vi={44:1},ku="org.eclipse.elk.alg.layered.graph",Xfe=" -> ",YTt="Not supported by LGraph",sNe="Port side is undefined",Jfe={3:1,6:1,4:1,5:1,469:1,145:1,99:1,114:1},j0={3:1,6:1,4:1,5:1,145:1,199:1,209:1,99:1,114:1},XTt={3:1,6:1,4:1,5:1,145:1,1988:1,209:1,99:1,114:1},JTt=`([{"' 	\r
`,ZTt=`)]}"' 	\r
`,QTt="The given string contains parts that cannot be parsed as numbers.",nq="org.eclipse.elk.core.math",ext={3:1,4:1,139:1,213:1,411:1},txt={3:1,4:1,107:1,213:1,411:1},$0="org.eclipse.elk.alg.layered.graph.transform",nxt="ElkGraphImporter",rxt="ElkGraphImporter/lambda$1$Type",ixt="ElkGraphImporter/lambda$2$Type",oxt="ElkGraphImporter/lambda$4$Type",yn="org.eclipse.elk.alg.layered.intermediate",axt="Node margin calculation",sxt="ONE_SIDED_GREEDY_SWITCH",cxt="TWO_SIDED_GREEDY_SWITCH",Zfe="No implementation is available for the layout processor ",Qfe="IntermediateProcessorStrategy",ehe="Node '",uxt="FIRST_SEPARATE",lxt="LAST_SEPARATE",dxt="Odd port side processing",co="org.eclipse.elk.alg.layered.intermediate.compaction",h$="org.eclipse.elk.alg.layered.intermediate.greedyswitch",xb="org.eclipse.elk.alg.layered.p3order.counting",rq={229:1},jR="org.eclipse.elk.alg.layered.intermediate.loops",gf="org.eclipse.elk.alg.layered.intermediate.loops.ordering",mw="org.eclipse.elk.alg.layered.intermediate.loops.routing",PQ="org.eclipse.elk.alg.layered.intermediate.preserveorder",xm="org.eclipse.elk.alg.layered.intermediate.wrapping",Pc="org.eclipse.elk.alg.layered.options",the="INTERACTIVE",cNe="GREEDY",fxt="DEPTH_FIRST",hxt="EDGE_LENGTH",pxt="SELF_LOOPS",gxt="firstTryWithInitialOrder",uNe="org.eclipse.elk.layered.directionCongruency",lNe="org.eclipse.elk.layered.feedbackEdges",LQ="org.eclipse.elk.layered.interactiveReferencePoint",dNe="org.eclipse.elk.layered.mergeEdges",fNe="org.eclipse.elk.layered.mergeHierarchyEdges",hNe="org.eclipse.elk.layered.allowNonFlowPortsToSwitchSides",pNe="org.eclipse.elk.layered.portSortingStrategy",gNe="org.eclipse.elk.layered.thoroughness",mNe="org.eclipse.elk.layered.unnecessaryBendpoints",bNe="org.eclipse.elk.layered.generatePositionAndLayerIds",nhe="org.eclipse.elk.layered.cycleBreaking.strategy",iq="org.eclipse.elk.layered.layering.strategy",vNe="org.eclipse.elk.layered.layering.layerConstraint",wNe="org.eclipse.elk.layered.layering.layerChoiceConstraint",yNe="org.eclipse.elk.layered.layering.layerId",rhe="org.eclipse.elk.layered.layering.minWidth.upperBoundOnWidth",ihe="org.eclipse.elk.layered.layering.minWidth.upperLayerEstimationScalingFactor",ohe="org.eclipse.elk.layered.layering.nodePromotion.strategy",ahe="org.eclipse.elk.layered.layering.nodePromotion.maxIterations",she="org.eclipse.elk.layered.layering.coffmanGraham.layerBound",p$="org.eclipse.elk.layered.crossingMinimization.strategy",ENe="org.eclipse.elk.layered.crossingMinimization.forceNodeModelOrder",che="org.eclipse.elk.layered.crossingMinimization.hierarchicalSweepiness",uhe="org.eclipse.elk.layered.crossingMinimization.semiInteractive",SNe="org.eclipse.elk.layered.crossingMinimization.inLayerPredOf",_Ne="org.eclipse.elk.layered.crossingMinimization.inLayerSuccOf",kNe="org.eclipse.elk.layered.crossingMinimization.positionChoiceConstraint",TNe="org.eclipse.elk.layered.crossingMinimization.positionId",xNe="org.eclipse.elk.layered.crossingMinimization.greedySwitch.activationThreshold",lhe="org.eclipse.elk.layered.crossingMinimization.greedySwitch.type",DQ="org.eclipse.elk.layered.crossingMinimization.greedySwitchHierarchical.type",lA="org.eclipse.elk.layered.nodePlacement.strategy",MQ="org.eclipse.elk.layered.nodePlacement.favorStraightEdges",dhe="org.eclipse.elk.layered.nodePlacement.bk.edgeStraightening",fhe="org.eclipse.elk.layered.nodePlacement.bk.fixedAlignment",hhe="org.eclipse.elk.layered.nodePlacement.linearSegments.deflectionDampening",phe="org.eclipse.elk.layered.nodePlacement.networkSimplex.nodeFlexibility",ghe="org.eclipse.elk.layered.nodePlacement.networkSimplex.nodeFlexibility.default",CNe="org.eclipse.elk.layered.edgeRouting.selfLoopDistribution",ANe="org.eclipse.elk.layered.edgeRouting.selfLoopOrdering",jQ="org.eclipse.elk.layered.edgeRouting.splines.mode",$Q="org.eclipse.elk.layered.edgeRouting.splines.sloppy.layerSpacingFactor",mhe="org.eclipse.elk.layered.edgeRouting.polyline.slopedEdgeZoneWidth",INe="org.eclipse.elk.layered.spacing.baseValue",ONe="org.eclipse.elk.layered.spacing.edgeNodeBetweenLayers",NNe="org.eclipse.elk.layered.spacing.edgeEdgeBetweenLayers",RNe="org.eclipse.elk.layered.spacing.nodeNodeBetweenLayers",PNe="org.eclipse.elk.layered.priority.direction",LNe="org.eclipse.elk.layered.priority.shortness",DNe="org.eclipse.elk.layered.priority.straightness",bhe="org.eclipse.elk.layered.compaction.connectedComponents",MNe="org.eclipse.elk.layered.compaction.postCompaction.strategy",jNe="org.eclipse.elk.layered.compaction.postCompaction.constraints",FQ="org.eclipse.elk.layered.highDegreeNodes.treatment",vhe="org.eclipse.elk.layered.highDegreeNodes.threshold",whe="org.eclipse.elk.layered.highDegreeNodes.treeHeight",I1="org.eclipse.elk.layered.wrapping.strategy",BQ="org.eclipse.elk.layered.wrapping.additionalEdgeSpacing",UQ="org.eclipse.elk.layered.wrapping.correctionFactor",g$="org.eclipse.elk.layered.wrapping.cutting.strategy",yhe="org.eclipse.elk.layered.wrapping.cutting.cuts",Ehe="org.eclipse.elk.layered.wrapping.cutting.msd.freedom",zQ="org.eclipse.elk.layered.wrapping.validify.strategy",HQ="org.eclipse.elk.layered.wrapping.validify.forbiddenIndices",GQ="org.eclipse.elk.layered.wrapping.multiEdge.improveCuts",qQ="org.eclipse.elk.layered.wrapping.multiEdge.distancePenalty",She="org.eclipse.elk.layered.wrapping.multiEdge.improveWrappedEdges",_he="org.eclipse.elk.layered.layerUnzipping.strategy",$Ne="org.eclipse.elk.layered.layerUnzipping.layerSplit",khe="org.eclipse.elk.layered.layerUnzipping.resetOnLongEdges",FNe="org.eclipse.elk.layered.edgeLabels.sideSelection",BNe="org.eclipse.elk.layered.edgeLabels.centerLabelPlacementStrategy",VQ="org.eclipse.elk.layered.considerModelOrder.strategy",UNe="org.eclipse.elk.layered.considerModelOrder.portModelOrder",zNe="org.eclipse.elk.layered.considerModelOrder.noModelOrder",The="org.eclipse.elk.layered.considerModelOrder.components",HNe="org.eclipse.elk.layered.considerModelOrder.longEdgeStrategy",xhe="org.eclipse.elk.layered.considerModelOrder.crossingCounterNodeInfluence",Che="org.eclipse.elk.layered.considerModelOrder.crossingCounterPortInfluence",Ahe="layering",mxt="layering.minWidth",bxt="layering.nodePromotion",BL="crossingMinimization",WQ="org.eclipse.elk.hierarchyHandling",vxt="crossingMinimization.greedySwitch",wxt="nodePlacement",yxt="nodePlacement.bk",Ext="edgeRouting",oq="org.eclipse.elk.edgeRouting",Pg="spacing",GNe="priority",qNe="compaction",Sxt="compaction.postCompaction",_xt="Specifies whether and how post-process compaction is applied.",VNe="highDegreeNodes",WNe="wrapping",kxt="wrapping.cutting",Txt="wrapping.validify",KNe="wrapping.multiEdge",YNe="layerUnzipping",Ihe="edgeLabels",m$="considerModelOrder",XNe="org.eclipse.elk.spacing.commentComment",JNe="org.eclipse.elk.spacing.commentNode",ZNe="org.eclipse.elk.spacing.componentComponent",QNe="org.eclipse.elk.spacing.edgeEdge",Ohe="org.eclipse.elk.spacing.edgeNode",eRe="org.eclipse.elk.spacing.labelLabel",tRe="org.eclipse.elk.spacing.labelPortHorizontal",nRe="org.eclipse.elk.spacing.labelPortVertical",rRe="org.eclipse.elk.spacing.labelNode",iRe="org.eclipse.elk.spacing.nodeSelfLoop",oRe="org.eclipse.elk.spacing.portPort",aRe="org.eclipse.elk.spacing.individual",sRe="org.eclipse.elk.port.borderOffset",cRe="org.eclipse.elk.noLayout",uRe="org.eclipse.elk.port.side",aq="org.eclipse.elk.debugMode",lRe="org.eclipse.elk.alignment",dRe="org.eclipse.elk.insideSelfLoops.activate",fRe="org.eclipse.elk.insideSelfLoops.yo",Nhe="org.eclipse.elk.direction",hRe="org.eclipse.elk.nodeLabels.padding",pRe="org.eclipse.elk.portLabels.nextToPortIfPossible",gRe="org.eclipse.elk.portLabels.treatAsGroup",mRe="org.eclipse.elk.portAlignment.default",bRe="org.eclipse.elk.portAlignment.north",vRe="org.eclipse.elk.portAlignment.south",wRe="org.eclipse.elk.portAlignment.west",yRe="org.eclipse.elk.portAlignment.east",KQ="org.eclipse.elk.contentAlignment",ERe="org.eclipse.elk.junctionPoints",SRe="org.eclipse.elk.edge.thickness",_Re="org.eclipse.elk.edgeLabels.placement",kRe="org.eclipse.elk.port.index",TRe="org.eclipse.elk.commentBox",xRe="org.eclipse.elk.hypernode",CRe="org.eclipse.elk.port.anchor",Rhe="org.eclipse.elk.partitioning.activate",Phe="org.eclipse.elk.partitioning.partition",YQ="org.eclipse.elk.position",ARe="org.eclipse.elk.margins",IRe="org.eclipse.elk.spacing.portsSurrounding",XQ="org.eclipse.elk.interactiveLayout",Lc="org.eclipse.elk.core.util",ORe={3:1,4:1,5:1,585:1},xxt="NETWORK_SIMPLEX",NRe="SIMPLE",fs={106:1,44:1},JQ="org.eclipse.elk.alg.layered.p1cycles",O1="org.eclipse.elk.alg.layered.p2layers",RRe={402:1,229:1},Cxt={825:1,3:1,4:1},Ll="org.eclipse.elk.alg.layered.p3order",$R=17976931348623157e292,hs="org.eclipse.elk.alg.layered.p4nodes",Axt={3:1,4:1,5:1,833:1},Cm=1e-5,bw="org.eclipse.elk.alg.layered.p4nodes.bk",Lhe="org.eclipse.elk.alg.layered.p5edges",Dp="org.eclipse.elk.alg.layered.p5edges.orthogonal",Dhe="org.eclipse.elk.alg.layered.p5edges.orthogonal.direction",Mhe=1e-6,HT="org.eclipse.elk.alg.layered.p5edges.splines",jhe=.09999999999999998,ZQ=1e-8,Ixt=4.71238898038469,Oxt=1.5707963267948966,PRe=3.141592653589793,N1="org.eclipse.elk.alg.mrtree",$he=.10000000149011612,QQ="SUPER_ROOT",b$="org.eclipse.elk.alg.mrtree.graph",LRe=-17976931348623157e292,Tu="org.eclipse.elk.alg.mrtree.intermediate",Nxt="Processor compute fanout",eee={3:1,6:1,4:1,5:1,519:1,99:1,114:1},Rxt="Set neighbors in level",sq="org.eclipse.elk.alg.mrtree.options",Pxt="DESCENDANTS",DRe="org.eclipse.elk.mrtree.compaction",MRe="org.eclipse.elk.mrtree.edgeEndTextureLength",jRe="org.eclipse.elk.mrtree.treeLevel",$Re="org.eclipse.elk.mrtree.positionConstraint",FRe="org.eclipse.elk.mrtree.weighting",BRe="org.eclipse.elk.mrtree.edgeRoutingMode",URe="org.eclipse.elk.mrtree.searchOrder",Lxt="Position Constraint",Qu="org.eclipse.elk.mrtree",Dxt="org.eclipse.elk.tree",Mxt="Processor arrange level",UL="org.eclipse.elk.alg.mrtree.p2order",Hd="org.eclipse.elk.alg.mrtree.p4route",zRe="org.eclipse.elk.alg.radial",_E=6.283185307179586,HRe="Before",GRe=5e-324,tee="After",qRe="org.eclipse.elk.alg.radial.intermediate",jxt="COMPACTION",Fhe="org.eclipse.elk.alg.radial.intermediate.compaction",$xt={3:1,4:1,5:1,99:1},VRe="org.eclipse.elk.alg.radial.intermediate.optimization",Bhe="No implementation is available for the layout option ",v$="org.eclipse.elk.alg.radial.options",Fxt="CompactionStrategy",WRe="org.eclipse.elk.radial.centerOnRoot",KRe="org.eclipse.elk.radial.orderId",YRe="org.eclipse.elk.radial.radius",nee="org.eclipse.elk.radial.rotate",Uhe="org.eclipse.elk.radial.compactor",zhe="org.eclipse.elk.radial.compactionStepSize",XRe="org.eclipse.elk.radial.sorter",JRe="org.eclipse.elk.radial.wedgeCriteria",ZRe="org.eclipse.elk.radial.optimizationCriteria",Hhe="org.eclipse.elk.radial.rotation.targetAngle",Ghe="org.eclipse.elk.radial.rotation.computeAdditionalWedgeSpace",QRe="org.eclipse.elk.radial.rotation.outgoingEdgeAngles",Bxt="Compaction",e4e="rotation",Uf="org.eclipse.elk.radial",Uxt="org.eclipse.elk.alg.radial.p1position.wedge",t4e="org.eclipse.elk.alg.radial.sorting",zxt=5.497787143782138,Hxt=3.9269908169872414,Gxt=2.356194490192345,qxt="org.eclipse.elk.alg.rectpacking",w$="org.eclipse.elk.alg.rectpacking.intermediate",qhe="org.eclipse.elk.alg.rectpacking.options",n4e="org.eclipse.elk.rectpacking.trybox",r4e="org.eclipse.elk.rectpacking.currentPosition",i4e="org.eclipse.elk.rectpacking.desiredPosition",o4e="org.eclipse.elk.rectpacking.inNewRow",a4e="org.eclipse.elk.rectpacking.orderBySize",s4e="org.eclipse.elk.rectpacking.widthApproximation.strategy",c4e="org.eclipse.elk.rectpacking.widthApproximation.targetWidth",u4e="org.eclipse.elk.rectpacking.widthApproximation.optimizationGoal",l4e="org.eclipse.elk.rectpacking.widthApproximation.lastPlaceShift",d4e="org.eclipse.elk.rectpacking.packing.strategy",f4e="org.eclipse.elk.rectpacking.packing.compaction.rowHeightReevaluation",h4e="org.eclipse.elk.rectpacking.packing.compaction.iterations",p4e="org.eclipse.elk.rectpacking.whiteSpaceElimination.strategy",Vhe="widthApproximation",Vxt="Compaction Strategy",Wxt="packing.compaction",Jl="org.eclipse.elk.rectpacking",zL="org.eclipse.elk.alg.rectpacking.p1widthapproximation",ree="org.eclipse.elk.alg.rectpacking.p2packing",Kxt="No Compaction",g4e="org.eclipse.elk.alg.rectpacking.p3whitespaceelimination",cq="org.eclipse.elk.alg.rectpacking.util",iee="No implementation available for ",GT="org.eclipse.elk.alg.spore",qT="org.eclipse.elk.alg.spore.options",QS="org.eclipse.elk.sporeCompaction",Whe="org.eclipse.elk.underlyingLayoutAlgorithm",m4e="org.eclipse.elk.processingOrder.treeConstruction",b4e="org.eclipse.elk.processingOrder.spanningTreeCostFunction",Khe="org.eclipse.elk.processingOrder.preferredRoot",Yhe="org.eclipse.elk.processingOrder.rootSelection",Xhe="org.eclipse.elk.structure.structureExtractionStrategy",v4e="org.eclipse.elk.compaction.compactionStrategy",w4e="org.eclipse.elk.compaction.orthogonal",y4e="org.eclipse.elk.overlapRemoval.maxIterations",E4e="org.eclipse.elk.overlapRemoval.runScanline",Jhe="processingOrder",Yxt="overlapRemoval",HL="org.eclipse.elk.sporeOverlap",Xxt="org.eclipse.elk.alg.spore.p1structure",Zhe="org.eclipse.elk.alg.spore.p2processingorder",Qhe="org.eclipse.elk.alg.spore.p3execution",Jxt="Topdown Layout",Zxt="Invalid index: ",GL="org.eclipse.elk.core.alg",dA={337:1},VT={293:1},Qxt="Make sure its type is registered with the ",S4e=" utility class.",qL="true",epe="false",eCt="Couldn't clone property '",e_=.05,Dl="org.eclipse.elk.core.options",tCt=1.2999999523162842,t_="org.eclipse.elk.box",_4e="org.eclipse.elk.expandNodes",k4e="org.eclipse.elk.box.packingMode",nCt="org.eclipse.elk.algorithm",rCt="org.eclipse.elk.resolvedAlgorithm",T4e="org.eclipse.elk.bendPoints",z7n="org.eclipse.elk.labelManager",iCt="org.eclipse.elk.softwrappingFuzziness",oCt="org.eclipse.elk.scaleFactor",aCt="org.eclipse.elk.childAreaWidth",sCt="org.eclipse.elk.childAreaHeight",cCt="org.eclipse.elk.animate",uCt="org.eclipse.elk.animTimeFactor",lCt="org.eclipse.elk.layoutAncestors",dCt="org.eclipse.elk.maxAnimTime",fCt="org.eclipse.elk.minAnimTime",hCt="org.eclipse.elk.progressBar",pCt="org.eclipse.elk.validateGraph",gCt="org.eclipse.elk.validateOptions",mCt="org.eclipse.elk.zoomToFit",bCt="org.eclipse.elk.json.shapeCoords",vCt="org.eclipse.elk.json.edgeCoords",H7n="org.eclipse.elk.font.name",wCt="org.eclipse.elk.font.size",x4e="org.eclipse.elk.topdown.sizeApproximator",C4e="org.eclipse.elk.topdown.scaleCap",yCt="org.eclipse.elk.edge.type",ECt="partitioning",SCt="nodeLabels",oee="portAlignment",tpe="nodeSize",npe="port",A4e="portLabels",uq="topdown",_Ct="insideSelfLoops",I4e="INHERIT",y$="org.eclipse.elk.fixed",aee="org.eclipse.elk.random",O4e={3:1,34:1,23:1,343:1},kCt="port must have a parent node to calculate the port side",TCt="The edge needs to have exactly one edge section. Found: ",E$="org.eclipse.elk.core.util.adapters",zf="org.eclipse.emf.ecore",fA="org.eclipse.elk.graph",xCt="EMapPropertyHolder",CCt="ElkBendPoint",ACt="ElkGraphElement",ICt="ElkConnectableShape",N4e="ElkEdge",OCt="ElkEdgeSection",NCt="EModelElement",RCt="ENamedElement",R4e="ElkLabel",P4e="ElkNode",L4e="ElkPort",PCt={93:1,92:1},FR="org.eclipse.emf.common.notify.impl",vw="The feature '",S$="' is not a valid changeable feature",LCt="Expecting null",rpe="' is not a valid feature",DCt="The feature ID",MCt=" is not a valid feature ID",Dc=32768,jCt={109:1,93:1,92:1,57:1,52:1,98:1},un="org.eclipse.emf.ecore.impl",kE="org.eclipse.elk.graph.impl",_$="Recursive containment not allowed for ",VL="The datatype '",n_="' is not a valid classifier",ipe="The value '",hA={195:1,3:1,4:1},ope="The class '",WL="http://www.eclipse.org/elk/ElkGraph",D4e="property",k$="value",ape="source",$Ct="properties",FCt="identifier",spe="height",cpe="width",upe="parent",lpe="text",dpe="children",BCt="hierarchical",M4e="sources",fpe="targets",hpe="sections",see="bendPoints",j4e="outgoingShape",$4e="incomingShape",F4e="outgoingSections",B4e="incomingSections",Ha="org.eclipse.emf.common.util",U4e="Severe implementation error in the Json to ElkGraph importer.",Am="id",ma="org.eclipse.elk.graph.json",KL="Unhandled parameter types: ",UCt="startPoint",zCt="An edge must have at least one source and one target (edge id: '",YL="').",HCt="Referenced edge section does not exist: ",GCt=" (edge id: '",z4e="target",qCt="sourcePoint",VCt="targetPoint",cee="group",Gr="name",WCt="connectableShape cannot be null",KCt="edge cannot be null",YCt="Passed edge is not 'simple'.",uee="org.eclipse.elk.graph.util",lq="The 'no duplicates' constraint is violated",ppe="targetIndex=",TE=", size=",gpe="sourceIndex=",Im={3:1,4:1,20:1,31:1,55:1,16:1,15:1,59:1,70:1,66:1,61:1},mpe={3:1,4:1,20:1,31:1,55:1,16:1,49:1,15:1,59:1,70:1,66:1,61:1,580:1},lee="logging",XCt="measureExecutionTime",JCt="parser.parse.1",ZCt="parser.parse.2",dee="parser.next.1",bpe="parser.next.2",QCt="parser.next.3",e3t="parser.next.4",xE="parser.factor.1",H4e="parser.factor.2",t3t="parser.factor.3",n3t="parser.factor.4",r3t="parser.factor.5",i3t="parser.factor.6",o3t="parser.atom.1",a3t="parser.atom.2",s3t="parser.atom.3",G4e="parser.atom.4",vpe="parser.atom.5",q4e="parser.cc.1",fee="parser.cc.2",c3t="parser.cc.3",u3t="parser.cc.5",V4e="parser.cc.6",W4e="parser.cc.7",wpe="parser.cc.8",l3t="parser.ope.1",d3t="parser.ope.2",f3t="parser.ope.3",F0="parser.descape.1",h3t="parser.descape.2",p3t="parser.descape.3",g3t="parser.descape.4",m3t="parser.descape.5",Hf="parser.process.1",b3t="parser.quantifier.1",v3t="parser.quantifier.2",w3t="parser.quantifier.3",y3t="parser.quantifier.4",K4e="parser.quantifier.5",E3t="org.eclipse.emf.common.notify",Y4e={412:1,671:1},S3t={3:1,4:1,20:1,31:1,55:1,16:1,15:1,70:1,61:1},dq={370:1,150:1},T$="index=",ype={3:1,4:1,5:1,128:1},_3t={3:1,4:1,20:1,31:1,55:1,16:1,15:1,59:1,70:1,61:1},X4e={3:1,6:1,4:1,5:1,198:1},k3t={3:1,4:1,5:1,174:1,371:1},ep=1024,T3t=";/?:@&=+$,",x3t="invalid authority: ",C3t="EAnnotation",A3t="ETypedElement",I3t="EStructuralFeature",O3t="EAttribute",N3t="EClassifier",R3t="EEnumLiteral",P3t="EGenericType",L3t="EOperation",D3t="EParameter",M3t="EReference",j3t="ETypeParameter",Oi="org.eclipse.emf.ecore.util",Epe={76:1},J4e={3:1,20:1,16:1,15:1,61:1,581:1,76:1,71:1,96:1},$3t="org.eclipse.emf.ecore.util.FeatureMap$Entry",Ml=8192,x$="byte",hee="char",C$="double",A$="float",I$="int",O$="long",N$="short",F3t="java.lang.Object",pA={3:1,4:1,5:1,253:1},Z4e={3:1,4:1,5:1,673:1},B3t={3:1,4:1,20:1,31:1,55:1,16:1,15:1,59:1,70:1,66:1,61:1,71:1},ec={3:1,4:1,20:1,31:1,55:1,16:1,15:1,59:1,70:1,66:1,61:1,76:1,71:1,96:1},fq="mixed",kr="http:///org/eclipse/emf/ecore/util/ExtendedMetaData",mh="kind",U3t={3:1,4:1,5:1,674:1},Q4e={3:1,4:1,20:1,31:1,55:1,16:1,15:1,70:1,61:1,76:1,71:1,96:1},pee={20:1,31:1,55:1,16:1,15:1,61:1,71:1},gee={49:1,127:1,284:1},mee={74:1,339:1},bee="The value of type '",vee="' must be of type '",gA=1291,bh="http://www.eclipse.org/emf/2002/Ecore",wee=-32768,r_="constraints",Ta="baseType",z3t="getEStructuralFeature",H3t="getFeatureID",R$="feature",G3t="getOperationID",e6e="operation",q3t="defaultValue",V3t="eTypeParameters",W3t="isInstance",K3t="getEEnumLiteral",Y3t="eContainingClass",jr={58:1},X3t={3:1,4:1,5:1,122:1},J3t="org.eclipse.emf.ecore.resource",Z3t={93:1,92:1,583:1,1980:1},Spe="org.eclipse.emf.ecore.resource.impl",t6e="unspecified",hq="simple",yee="attribute",Q3t="attributeWildcard",Eee="element",_pe="elementWildcard",Mp="collapse",kpe="itemType",See="namespace",pq="##targetNamespace",vh="whiteSpace",n6e="wildcards",CE="http://www.eclipse.org/emf/2003/XMLType",Tpe="##any",XL="uninitialized",gq="The multiplicity constraint is violated",_ee="org.eclipse.emf.ecore.xml.type",eAt="ProcessingInstruction",tAt="SimpleAnyType",nAt="XMLTypeDocumentRoot",xo="org.eclipse.emf.ecore.xml.type.impl",mq="INF",rAt="processing",iAt="ENTITIES_._base",r6e="minLength",i6e="ENTITY",kee="NCName",oAt="IDREFS_._base",o6e="integer",xpe="token",Cpe="pattern",aAt="[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*",a6e="\\i\\c*",sAt="[\\i-[:]][\\c-[:]]*",cAt="nonPositiveInteger",bq="maxInclusive",s6e="NMTOKEN",uAt="NMTOKENS_._base",c6e="nonNegativeInteger",vq="minInclusive",lAt="normalizedString",dAt="unsignedByte",fAt="unsignedInt",hAt="18446744073709551615",pAt="unsignedShort",gAt="processingInstruction",B0="org.eclipse.emf.ecore.xml.type.internal",JL=1114111,mAt="Internal Error: shorthands: \\u",P$="xml:isDigit",Ape="xml:isWord",Ipe="xml:isSpace",Ope="xml:isNameChar",Npe="xml:isInitialNameChar",bAt="09٠٩۰۹०९০৯੦੯૦૯୦୯௧௯౦౯೦೯൦൯๐๙໐໙༠༩",vAt="AZazÀÖØöøıĴľŁňŊžƀǃǍǰǴǵǺȗɐʨʻˁΆΆΈΊΌΌΎΡΣώϐϖϚϚϜϜϞϞϠϠϢϳЁЌЎяёќўҁҐӄӇӈӋӌӐӫӮӵӸӹԱՖՙՙաֆאתװײءغفيٱڷںھۀێېۓەەۥۦअहऽऽक़ॡঅঌএঐওনপরললশহড়ঢ়য়ৡৰৱਅਊਏਐਓਨਪਰਲਲ਼ਵਸ਼ਸਹਖ਼ੜਫ਼ਫ਼ੲੴઅઋઍઍએઑઓનપરલળવહઽઽૠૠଅଌଏଐଓନପରଲଳଶହଽଽଡ଼ଢ଼ୟୡஅஊஎஐஒகஙசஜஜஞடணதநபமவஷஹఅఌఎఐఒనపళవహౠౡಅಌಎಐಒನಪಳವಹೞೞೠೡഅഌഎഐഒനപഹൠൡกฮะะาำเๅກຂຄຄງຈຊຊຍຍດທນຟມຣລລວວສຫອຮະະາຳຽຽເໄཀཇཉཀྵႠჅაჶᄀᄀᄂᄃᄅᄇᄉᄉᄋᄌᄎᄒᄼᄼᄾᄾᅀᅀᅌᅌᅎᅎᅐᅐᅔᅕᅙᅙᅟᅡᅣᅣᅥᅥᅧᅧᅩᅩᅭᅮᅲᅳᅵᅵᆞᆞᆨᆨᆫᆫᆮᆯᆷᆸᆺᆺᆼᇂᇫᇫᇰᇰᇹᇹḀẛẠỹἀἕἘἝἠὅὈὍὐὗὙὙὛὛὝὝὟώᾀᾴᾶᾼιιῂῄῆῌῐΐῖΊῠῬῲῴῶῼΩΩKÅ℮℮ↀↂ〇〇〡〩ぁゔァヺㄅㄬ一龥가힣",wAt="Private Use",Rpe="ASSIGNED",Ppe="\0ÿĀſƀɏɐʯʰ˿̀ͯͰϿЀӿ԰֏֐׿؀ۿ܀ݏހ޿ऀॿঀ৿਀੿઀૿଀୿஀௿ఀ౿ಀ೿ഀൿ඀෿฀๿຀໿ༀ࿿က႟Ⴀჿᄀᇿሀ፿Ꭰ᏿᐀ᙿ ᚟ᚠ᛿ក៿᠀᢯Ḁỿἀ῿ ⁯⁰₟₠⃏⃐⃿℀⅏⅐↏←⇿∀⋿⌀⏿␀␿⑀⑟①⓿─╿▀▟■◿☀⛿✀➿⠀⣿⺀⻿⼀⿟⿰⿿　〿぀ゟ゠ヿ㄀ㄯ㄰㆏㆐㆟ㆠㆿ㈀㋿㌀㏿㐀䶵一鿿ꀀ꒏꒐꓏가힣豈﫿ﬀﭏﭐ﷿︠︯︰﹏﹐﹯ﹰ﻾\uFEFF\uFEFF＀￯",u6e="UNASSIGNED",ZL={3:1,121:1},yAt="org.eclipse.emf.ecore.xml.type.util",Tee={3:1,4:1,5:1,373:1},l6e="org.eclipse.xtext.xbase.lib",EAt="Cannot add elements to a Range",SAt="Cannot set elements in a Range",_At="Cannot remove elements from a Range",kAt="user.agent",p,xee,Lpe;h.goog=h.goog||{},h.goog.global=h.goog.global||h,xee={},N(1,null,{},y),p.Fb=function(t){return ust(this,t)},p.Gb=function(){return this.Om},p.Hb=function(){return bS(this)},p.Ib=function(){var t;return Vy(Fd(this))+"@"+(t=Ti(this)>>>0,t.toString(16))},p.equals=function(e){return this.Fb(e)},p.hashCode=function(){return this.Hb()},p.toString=function(){return this.Ib()};var TAt,xAt,CAt;N(295,1,{295:1,2070:1},uCe),p.te=function(t){var i;return i=new uCe,i.i=4,t>1?i.c=Lft(this,t-1):i.c=this,i},p.ue=function(){return d1(this),this.b},p.ve=function(){return Vy(this)},p.we=function(){return d1(this),this.k},p.xe=function(){return(this.i&4)!=0},p.ye=function(){return(this.i&1)!=0},p.Ib=function(){return yxe(this)},p.i=0;var Ro=R(Sc,"Object",1),d6e=R(Sc,"Class",295);N(2042,1,MG),R(jG,"Optional",2042),N(1155,2042,MG,E),p.Fb=function(t){return t===this},p.Hb=function(){return 2040732332},p.Ib=function(){return"Optional.absent()"},p.Jb=function(t){return or(t),a7(),Dpe};var Dpe;R(jG,"Absent",1155),N(621,1,{},Aae),R(jG,"Joiner",621);var G7n=Bi(jG,"Predicate");N(572,1,{177:1,572:1,3:1,48:1},G9),p.Mb=function(t){return N1t(this,t)},p.Lb=function(t){return N1t(this,t)},p.Fb=function(t){var i;return me(t,572)?(i=f(t,572),oIe(this.a,i.a)):!1},p.Hb=function(){return hCe(this.a)+306654252},p.Ib=function(){return aRn(this.a)},R(jG,"Predicates/AndPredicate",572),N(408,2042,{408:1,3:1},nS),p.Fb=function(t){var i;return me(t,408)?(i=f(t,408),ei(this.a,i.a)):!1},p.Hb=function(){return 1502476572+Ti(this.a)},p.Ib=function(){return Ykt+this.a+")"},p.Jb=function(t){return new nS(JX(t.Kb(this.a),"the Function passed to Optional.transform() must not return null."))},R(jG,"Present",408),N(204,1,TL),p.Nb=function(t){va(this,t)},p.Qb=function(){ait()},R(Rt,"UnmodifiableIterator",204),N(2022,204,xL),p.Qb=function(){ait()},p.Rb=function(t){throw ie(new cr)},p.Wb=function(t){throw ie(new cr)},R(Rt,"UnmodifiableListIterator",2022),N(390,2022,xL),p.Ob=function(){return this.c<this.d},p.Sb=function(){return this.c>0},p.Pb=function(){if(this.c>=this.d)throw ie(new ic);return this.Xb(this.c++)},p.Tb=function(){return this.c},p.Ub=function(){if(this.c<=0)throw ie(new ic);return this.Xb(--this.c)},p.Vb=function(){return this.c-1},p.c=0,p.d=0,R(Rt,"AbstractIndexedListIterator",390),N(697,204,TL),p.Ob=function(){return Mue(this)},p.Pb=function(){return pxe(this)},p.e=1,R(Rt,"AbstractIterator",697),N(2030,1,{228:1}),p.Zb=function(){var t;return t=this.f,t||(this.f=this.ac())},p.Fb=function(t){return tle(this,t)},p.Hb=function(){return Ti(this.Zb())},p.dc=function(){return this.gc()==0},p.ec=function(){return YN(this)},p.Ib=function(){return Qs(this.Zb())},R(Rt,"AbstractMultimap",2030),N(725,2030,wE),p.$b=function(){UJ(this)},p._b=function(t){return wit(this,t)},p.ac=function(){return new aP(this,this.c)},p.ic=function(t){return this.hc()},p.bc=function(){return new N3(this,this.c)},p.jc=function(){return this.mc(this.hc())},p.kc=function(){return new zrt(this)},p.lc=function(){return cde(this.c.vc().Lc(),new x,64,this.d)},p.cc=function(t){return si(this,t)},p.fc=function(t){return QH(this,t)},p.gc=function(){return this.d},p.mc=function(t){return Ht(),new h3(t)},p.nc=function(){return new Urt(this)},p.oc=function(){return cde(this.c.Bc().Lc(),new k,64,this.d)},p.pc=function(t,i){return new EJ(this,t,i,null)},p.d=0,R(Rt,"AbstractMapBasedMultimap",725),N(1637,725,wE),p.hc=function(){return new Du(this.a)},p.jc=function(){return Ht(),Ht(),Ga},p.cc=function(t){return f(si(this,t),15)},p.fc=function(t){return f(QH(this,t),15)},p.Zb=function(){return tR(this)},p.Fb=function(t){return tle(this,t)},p.qc=function(t){return f(si(this,t),15)},p.rc=function(t){return f(QH(this,t),15)},p.mc=function(t){return Dce(f(t,15))},p.pc=function(t,i){return Fht(this,t,f(i,15),null)},R(Rt,"AbstractListMultimap",1637),N(731,1,aa),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.c.Ob()||this.e.Ob()},p.Pb=function(){var t;return this.e.Ob()||(t=f(this.c.Pb(),43),this.b=t.jd(),this.a=f(t.kd(),16),this.e=this.a.Jc()),this.sc(this.b,this.e.Pb())},p.Qb=function(){this.e.Qb(),f(ch(this.a),16).dc()&&this.c.Qb(),--this.d.d},R(Rt,"AbstractMapBasedMultimap/Itr",731),N(1093,731,aa,Urt),p.sc=function(t,i){return i},R(Rt,"AbstractMapBasedMultimap/1",1093),N(1094,1,{},k),p.Kb=function(t){return f(t,16).Lc()},R(Rt,"AbstractMapBasedMultimap/1methodref$spliterator$Type",1094),N(1095,731,aa,zrt),p.sc=function(t,i){return new dS(t,i)},R(Rt,"AbstractMapBasedMultimap/2",1095);var f6e=Bi(Gn,"Map");N(2011,1,YS),p.wc=function(t){HH(this,t)},p.$b=function(){this.vc().$b()},p.tc=function(t){return Gle(this,t)},p._b=function(t){return!!rAe(this,t,!1)},p.uc=function(t){var i,a,u;for(a=this.vc().Jc();a.Ob();)if(i=f(a.Pb(),43),u=i.kd(),Ae(t)===Ae(u)||t!=null&&ei(t,u))return!0;return!1},p.Fb=function(t){var i,a,u;if(t===this)return!0;if(!me(t,91)||(u=f(t,91),this.gc()!=u.gc()))return!1;for(a=u.vc().Jc();a.Ob();)if(i=f(a.Pb(),43),!this.tc(i))return!1;return!0},p.xc=function(t){return ac(rAe(this,t,!1))},p.Hb=function(){return aCe(this.vc())},p.dc=function(){return this.gc()==0},p.ec=function(){return new Ic(this)},p.yc=function(t,i){throw ie(new f0("Put not supported on this map"))},p.zc=function(t){mj(this,t)},p.Ac=function(t){return ac(rAe(this,t,!0))},p.gc=function(){return this.vc().gc()},p.Ib=function(){return Xwt(this)},p.Bc=function(){return new kp(this)},R(Gn,"AbstractMap",2011),N(2031,2011,YS),p.bc=function(){return new wY(this)},p.vc=function(){return Blt(this)},p.ec=function(){var t;return t=this.g,t||(this.g=this.bc())},p.Bc=function(){var t;return t=this.i,t||(this.i=new uot(this))},R(Rt,"Maps/ViewCachingAbstractMap",2031),N(393,2031,YS,aP),p.xc=function(t){return Q_n(this,t)},p.Ac=function(t){return uxn(this,t)},p.$b=function(){this.d==this.e.c?this.e.$b():CX(new gke(this))},p._b=function(t){return E0t(this.d,t)},p.Dc=function(){return new iz(this)},p.Cc=function(){return this.Dc()},p.Fb=function(t){return this===t||ei(this.d,t)},p.Hb=function(){return Ti(this.d)},p.ec=function(){return this.e.ec()},p.gc=function(){return this.d.gc()},p.Ib=function(){return Qs(this.d)},R(Rt,"AbstractMapBasedMultimap/AsMap",393);var Gf=Bi(Sc,"Iterable");N(31,1,jT),p.Ic=function(t){Oa(this,t)},p.Lc=function(){return new Bt(this,0)},p.Mc=function(){return new Ft(null,this.Lc())},p.Ec=function(t){throw ie(new f0("Add not supported on this collection"))},p.Fc=function(t){return Ba(this,t)},p.$b=function(){tTe(this)},p.Gc=function(t){return _T(this,t,!1)},p.Hc=function(t){return KH(this,t)},p.dc=function(){return this.gc()==0},p.Kc=function(t){return _T(this,t,!0)},p.Nc=function(){return Cke(this)},p.Oc=function(t){return Rj(this,t)},p.Ib=function(){return Cg(this)},R(Gn,"AbstractCollection",31);var wh=Bi(Gn,"Set");N(Ig,31,Pl),p.Lc=function(){return new Bt(this,1)},p.Fb=function(t){return fvt(this,t)},p.Hb=function(){return aCe(this)},R(Gn,"AbstractSet",Ig),N(2014,Ig,Pl),R(Rt,"Sets/ImprovedAbstractSet",2014),N(2015,2014,Pl),p.$b=function(){this.Pc().$b()},p.Gc=function(t){return K0t(this,t)},p.dc=function(){return this.Pc().dc()},p.Kc=function(t){var i;return this.Gc(t)&&me(t,43)?(i=f(t,43),this.Pc().ec().Kc(i.jd())):!1},p.gc=function(){return this.Pc().gc()},R(Rt,"Maps/EntrySet",2015),N(1091,2015,Pl,iz),p.Gc=function(t){return DCe(this.a.d.vc(),t)},p.Jc=function(){return new gke(this.a)},p.Pc=function(){return this.a},p.Kc=function(t){var i;return DCe(this.a.d.vc(),t)?(i=f(ch(f(t,43)),43),$Sn(this.a.e,i.jd()),!0):!1},p.Lc=function(){return sH(this.a.d.vc().Lc(),new oz(this.a))},R(Rt,"AbstractMapBasedMultimap/AsMap/AsMapEntries",1091),N(1092,1,{},oz),p.Kb=function(t){return Cpt(this.a,f(t,43))},R(Rt,"AbstractMapBasedMultimap/AsMap/AsMapEntries/0methodref$wrapEntry$Type",1092),N(729,1,aa,gke),p.Nb=function(t){va(this,t)},p.Pb=function(){var t;return t=f(this.b.Pb(),43),this.a=f(t.kd(),16),Cpt(this.c,t)},p.Ob=function(){return this.b.Ob()},p.Qb=function(){mP(!!this.a),this.b.Qb(),this.c.e.d-=this.a.gc(),this.a.$b(),this.a=null},R(Rt,"AbstractMapBasedMultimap/AsMap/AsMapIterator",729),N(527,2014,Pl,wY),p.$b=function(){this.b.$b()},p.Gc=function(t){return this.b._b(t)},p.Ic=function(t){or(t),this.b.wc(new mN(t))},p.dc=function(){return this.b.dc()},p.Jc=function(){return new s7(this.b.vc().Jc())},p.Kc=function(t){return this.b._b(t)?(this.b.Ac(t),!0):!1},p.gc=function(){return this.b.gc()},R(Rt,"Maps/KeySet",527),N(327,527,Pl,N3),p.$b=function(){var t;CX((t=this.b.vc().Jc(),new K2e(this,t)))},p.Hc=function(t){return this.b.ec().Hc(t)},p.Fb=function(t){return this===t||ei(this.b.ec(),t)},p.Hb=function(){return Ti(this.b.ec())},p.Jc=function(){var t;return t=this.b.vc().Jc(),new K2e(this,t)},p.Kc=function(t){var i,a;return a=0,i=f(this.b.Ac(t),16),i&&(a=i.gc(),i.$b(),this.a.d-=a),a>0},p.Lc=function(){return this.b.ec().Lc()},R(Rt,"AbstractMapBasedMultimap/KeySet",327),N(730,1,aa,K2e),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.c.Ob()},p.Pb=function(){return this.a=f(this.c.Pb(),43),this.a.jd()},p.Qb=function(){var t;mP(!!this.a),t=f(this.a.kd(),16),this.c.Qb(),this.b.a.d-=t.gc(),t.$b(),this.a=null},R(Rt,"AbstractMapBasedMultimap/KeySet/1",730),N(487,393,{91:1,134:1},eH),p.bc=function(){return this.Qc()},p.ec=function(){return this.Sc()},p.Qc=function(){return new xz(this.c,this.Uc())},p.Rc=function(){return this.Uc().Rc()},p.Sc=function(){var t;return t=this.b,t||(this.b=this.Qc())},p.Tc=function(){return this.Uc().Tc()},p.Uc=function(){return f(this.d,134)},R(Rt,"AbstractMapBasedMultimap/SortedAsMap",487),N(434,487,OOe,B7),p.bc=function(){return new sP(this.a,f(f(this.d,134),137))},p.Qc=function(){return new sP(this.a,f(f(this.d,134),137))},p.ec=function(){var t;return t=this.b,f(t||(this.b=new sP(this.a,f(f(this.d,134),137))),274)},p.Sc=function(){var t;return t=this.b,f(t||(this.b=new sP(this.a,f(f(this.d,134),137))),274)},p.Uc=function(){return f(f(this.d,134),137)},p.Vc=function(t){return f(f(this.d,134),137).Vc(t)},p.Wc=function(t){return f(f(this.d,134),137).Wc(t)},p.Xc=function(t,i){return new B7(this.a,f(f(this.d,134),137).Xc(t,i))},p.Yc=function(t){return f(f(this.d,134),137).Yc(t)},p.Zc=function(t){return f(f(this.d,134),137).Zc(t)},p.$c=function(t,i){return new B7(this.a,f(f(this.d,134),137).$c(t,i))},R(Rt,"AbstractMapBasedMultimap/NavigableAsMap",434),N(486,327,Xkt,xz),p.Lc=function(){return this.b.ec().Lc()},R(Rt,"AbstractMapBasedMultimap/SortedKeySet",486),N(392,486,NOe,sP),R(Rt,"AbstractMapBasedMultimap/NavigableKeySet",392),N(536,31,jT,EJ),p.Ec=function(t){var i,a;return vd(this),a=this.d.dc(),i=this.d.Ec(t),i&&(++this.f.d,a&&rH(this)),i},p.Fc=function(t){var i,a,u;return t.dc()?!1:(u=(vd(this),this.d.gc()),i=this.d.Fc(t),i&&(a=this.d.gc(),this.f.d+=a-u,u==0&&rH(this)),i)},p.$b=function(){var t;t=(vd(this),this.d.gc()),t!=0&&(this.d.$b(),this.f.d-=t,PX(this))},p.Gc=function(t){return vd(this),this.d.Gc(t)},p.Hc=function(t){return vd(this),this.d.Hc(t)},p.Fb=function(t){return t===this?!0:(vd(this),ei(this.d,t))},p.Hb=function(){return vd(this),Ti(this.d)},p.Jc=function(){return vd(this),new tke(this)},p.Kc=function(t){var i;return vd(this),i=this.d.Kc(t),i&&(--this.f.d,PX(this)),i},p.gc=function(){return est(this)},p.Lc=function(){return vd(this),this.d.Lc()},p.Ib=function(){return vd(this),Qs(this.d)},R(Rt,"AbstractMapBasedMultimap/WrappedCollection",536);var qf=Bi(Gn,"List");N(727,536,{20:1,31:1,16:1,15:1},Ake),p.gd=function(t){sE(this,t)},p.Lc=function(){return vd(this),this.d.Lc()},p._c=function(t,i){var a;vd(this),a=this.d.dc(),f(this.d,15)._c(t,i),++this.a.d,a&&rH(this)},p.ad=function(t,i){var a,u,l;return i.dc()?!1:(l=(vd(this),this.d.gc()),a=f(this.d,15).ad(t,i),a&&(u=this.d.gc(),this.a.d+=u-l,l==0&&rH(this)),a)},p.Xb=function(t){return vd(this),f(this.d,15).Xb(t)},p.bd=function(t){return vd(this),f(this.d,15).bd(t)},p.cd=function(){return vd(this),new Ost(this)},p.dd=function(t){return vd(this),new Zdt(this,t)},p.ed=function(t){var i;return vd(this),i=f(this.d,15).ed(t),--this.a.d,PX(this),i},p.fd=function(t,i){return vd(this),f(this.d,15).fd(t,i)},p.hd=function(t,i){return vd(this),Fht(this.a,this.e,f(this.d,15).hd(t,i),this.b?this.b:this)},R(Rt,"AbstractMapBasedMultimap/WrappedList",727),N(1090,727,{20:1,31:1,16:1,15:1,59:1},vct),R(Rt,"AbstractMapBasedMultimap/RandomAccessWrappedList",1090),N(613,1,aa,tke),p.Nb=function(t){va(this,t)},p.Ob=function(){return CP(this),this.b.Ob()},p.Pb=function(){return CP(this),this.b.Pb()},p.Qb=function(){nct(this)},R(Rt,"AbstractMapBasedMultimap/WrappedCollection/WrappedIterator",613),N(728,613,yb,Ost,Zdt),p.Qb=function(){nct(this)},p.Rb=function(t){var i;i=est(this.a)==0,(CP(this),f(this.b,127)).Rb(t),++this.a.a.d,i&&rH(this.a)},p.Sb=function(){return(CP(this),f(this.b,127)).Sb()},p.Tb=function(){return(CP(this),f(this.b,127)).Tb()},p.Ub=function(){return(CP(this),f(this.b,127)).Ub()},p.Vb=function(){return(CP(this),f(this.b,127)).Vb()},p.Wb=function(t){(CP(this),f(this.b,127)).Wb(t)},R(Rt,"AbstractMapBasedMultimap/WrappedList/WrappedListIterator",728),N(726,536,Xkt,S_e),p.Lc=function(){return vd(this),this.d.Lc()},R(Rt,"AbstractMapBasedMultimap/WrappedSortedSet",726),N(1089,726,NOe,xst),R(Rt,"AbstractMapBasedMultimap/WrappedNavigableSet",1089),N(1088,536,Pl,Bct),p.Lc=function(){return vd(this),this.d.Lc()},R(Rt,"AbstractMapBasedMultimap/WrappedSet",1088),N(1097,1,{},x),p.Kb=function(t){return KSn(f(t,43))},R(Rt,"AbstractMapBasedMultimap/lambda$1$Type",1097),N(1096,1,{},az),p.Kb=function(t){return new dS(this.a,t)},R(Rt,"AbstractMapBasedMultimap/lambda$2$Type",1096);var AE=Bi(Gn,"Map/Entry");N(355,1,ufe),p.Fb=function(t){var i;return me(t,43)?(i=f(t,43),dm(this.jd(),i.jd())&&dm(this.kd(),i.kd())):!1},p.Hb=function(){var t,i;return t=this.jd(),i=this.kd(),(t==null?0:Ti(t))^(i==null?0:Ti(i))},p.ld=function(t){throw ie(new cr)},p.Ib=function(){return this.jd()+"="+this.kd()},R(Rt,Jkt,355),N(2032,31,jT),p.$b=function(){this.md().$b()},p.Gc=function(t){var i;return me(t,43)?(i=f(t,43),b2n(this.md(),i.jd(),i.kd())):!1},p.Kc=function(t){var i;return me(t,43)?(i=f(t,43),Iht(this.md(),i.jd(),i.kd())):!1},p.gc=function(){return this.md().d},R(Rt,"Multimaps/Entries",2032),N(732,2032,jT,q9),p.Jc=function(){return this.a.kc()},p.md=function(){return this.a},p.Lc=function(){return this.a.lc()},R(Rt,"AbstractMultimap/Entries",732),N(733,732,Pl,T2e),p.Lc=function(){return this.a.lc()},p.Fb=function(t){return _Ae(this,t)},p.Hb=function(){return Lbt(this)},R(Rt,"AbstractMultimap/EntrySet",733),N(734,31,jT,F5),p.$b=function(){this.a.$b()},p.Gc=function(t){return oxn(this.a,t)},p.Jc=function(){return this.a.nc()},p.gc=function(){return this.a.d},p.Lc=function(){return this.a.oc()},R(Rt,"AbstractMultimap/Values",734),N(2033,31,{828:1,20:1,31:1,16:1}),p.Ic=function(t){or(t),P3(this).Ic(new J9(t))},p.Lc=function(){var t;return t=P3(this).Lc(),cde(t,new D,64|t.wd()&1296,this.a.d)},p.Ec=function(t){return D2e(),!0},p.Fc=function(t){return or(this),or(t),me(t,537)?k2n(f(t,828)):!t.dc()&&kue(this,t.Jc())},p.Gc=function(t){var i;return i=f(ST(tR(this.a),t),16),(i?i.gc():0)>0},p.Fb=function(t){return n6n(this,t)},p.Hb=function(){return Ti(P3(this))},p.dc=function(){return P3(this).dc()},p.Kc=function(t){return fEt(this,t,1)>0},p.Ib=function(){return Qs(P3(this))},R(Rt,"AbstractMultiset",2033),N(2035,2014,Pl),p.$b=function(){UJ(this.a.a)},p.Gc=function(t){var i,a;return me(t,488)?(a=f(t,413),f(a.a.kd(),16).gc()<=0?!1:(i=Kft(this.a,a.a.jd()),i==f(a.a.kd(),16).gc())):!1},p.Kc=function(t){var i,a,u,l;return me(t,488)&&(a=f(t,413),i=a.a.jd(),u=f(a.a.kd(),16).gc(),u!=0)?(l=this.a,e4n(l,i,u)):!1},R(Rt,"Multisets/EntrySet",2035),N(1103,2035,Pl,Lk),p.Jc=function(){return new Wrt(Blt(tR(this.a.a)).Jc())},p.gc=function(){return tR(this.a.a).gc()},R(Rt,"AbstractMultiset/EntrySet",1103),N(612,725,wE),p.hc=function(){return this.nd()},p.jc=function(){return this.od()},p.cc=function(t){return this.pd(t)},p.fc=function(t){return this.qd(t)},p.Zb=function(){var t;return t=this.f,t||(this.f=this.ac())},p.od=function(){return Ht(),Ht(),Ree},p.Fb=function(t){return tle(this,t)},p.pd=function(t){return f(si(this,t),22)},p.qd=function(t){return f(QH(this,t),22)},p.mc=function(t){return Ht(),new tP(f(t,22))},p.pc=function(t,i){return new Bct(this,t,f(i,22))},R(Rt,"AbstractSetMultimap",612),N(1665,612,wE),p.hc=function(){return new g0(this.b)},p.nd=function(){return new g0(this.b)},p.jc=function(){return zke(new g0(this.b))},p.od=function(){return zke(new g0(this.b))},p.cc=function(t){return f(f(si(this,t),22),83)},p.pd=function(t){return f(f(si(this,t),22),83)},p.fc=function(t){return f(f(QH(this,t),22),83)},p.qd=function(t){return f(f(QH(this,t),22),83)},p.mc=function(t){return me(t,274)?zke(f(t,274)):(Ht(),new a_e(f(t,83)))},p.Zb=function(){var t;return t=this.f,t||(this.f=me(this.c,137)?new B7(this,f(this.c,137)):me(this.c,134)?new eH(this,f(this.c,134)):new aP(this,this.c))},p.pc=function(t,i){return me(i,274)?new xst(this,t,f(i,274)):new S_e(this,t,f(i,83))},R(Rt,"AbstractSortedSetMultimap",1665),N(1666,1665,wE),p.Zb=function(){var t;return t=this.f,f(f(t||(this.f=me(this.c,137)?new B7(this,f(this.c,137)):me(this.c,134)?new eH(this,f(this.c,134)):new aP(this,this.c)),134),137)},p.ec=function(){var t;return t=this.i,f(f(t||(this.i=me(this.c,137)?new sP(this,f(this.c,137)):me(this.c,134)?new xz(this,f(this.c,134)):new N3(this,this.c)),83),274)},p.bc=function(){return me(this.c,137)?new sP(this,f(this.c,137)):me(this.c,134)?new xz(this,f(this.c,134)):new N3(this,this.c)},R(Rt,"AbstractSortedKeySortedSetMultimap",1666),N(2055,1,{1992:1}),p.Fb=function(t){return HIn(this,t)},p.Hb=function(){var t;return aCe((t=this.g,t||(this.g=new rS(this))))},p.Ib=function(){var t;return Xwt((t=this.f,t||(this.f=new n_e(this))))},R(Rt,"AbstractTable",2055),N(664,Ig,Pl,rS),p.$b=function(){sit()},p.Gc=function(t){var i,a;return me(t,465)?(i=f(t,682),a=f(ST(udt(this.a),Hv(i.c.e,i.b)),91),!!a&&DCe(a.vc(),new dS(Hv(i.c.c,i.a),sR(i.c,i.b,i.a)))):!1},p.Jc=function(){return yyn(this.a)},p.Kc=function(t){var i,a;return me(t,465)?(i=f(t,682),a=f(ST(udt(this.a),Hv(i.c.e,i.b)),91),!!a&&qxn(a.vc(),new dS(Hv(i.c.c,i.a),sR(i.c,i.b,i.a)))):!1},p.gc=function(){return hlt(this.a)},p.Lc=function(){return I2n(this.a)},R(Rt,"AbstractTable/CellSet",664),N(1971,31,jT,V9),p.$b=function(){sit()},p.Gc=function(t){return POn(this.a,t)},p.Jc=function(){return Eyn(this.a)},p.gc=function(){return hlt(this.a)},p.Lc=function(){return Sht(this.a)},R(Rt,"AbstractTable/Values",1971),N(1638,1637,wE),R(Rt,"ArrayListMultimapGwtSerializationDependencies",1638),N(504,1638,wE,Cae,wTe),p.hc=function(){return new Du(this.a)},p.a=0,R(Rt,"ArrayListMultimap",504),N(663,2055,{663:1,1992:1,3:1},TEt),R(Rt,"ArrayTable",663),N(1967,390,xL,rct),p.Xb=function(t){return new sCe(this.a,t)},R(Rt,"ArrayTable/1",1967),N(1968,1,{},sz),p.rd=function(t){return new sCe(this.a,t)},R(Rt,"ArrayTable/1methodref$getCell$Type",1968),N(2056,1,{682:1}),p.Fb=function(t){var i;return t===this?!0:me(t,465)?(i=f(t,682),dm(Hv(this.c.e,this.b),Hv(i.c.e,i.b))&&dm(Hv(this.c.c,this.a),Hv(i.c.c,i.a))&&dm(sR(this.c,this.b,this.a),sR(i.c,i.b,i.a))):!1},p.Hb=function(){return oZ(ce(se(Ro,1),Jt,1,5,[Hv(this.c.e,this.b),Hv(this.c.c,this.a),sR(this.c,this.b,this.a)]))},p.Ib=function(){return"("+Hv(this.c.e,this.b)+","+Hv(this.c.c,this.a)+")="+sR(this.c,this.b,this.a)},R(Rt,"Tables/AbstractCell",2056),N(465,2056,{465:1,682:1},sCe),p.a=0,p.b=0,p.d=0,R(Rt,"ArrayTable/2",465),N(1970,1,{},Dk),p.rd=function(t){return Ngt(this.a,t)},R(Rt,"ArrayTable/2methodref$getValue$Type",1970),N(1969,390,xL,ict),p.Xb=function(t){return Ngt(this.a,t)},R(Rt,"ArrayTable/3",1969),N(2023,2011,YS),p.$b=function(){CX(this.kc())},p.vc=function(){return new l3(this)},p.lc=function(){return new Bdt(this.kc(),this.gc())},R(Rt,"Maps/IteratorBasedAbstractMap",2023),N(821,2023,YS),p.$b=function(){throw ie(new cr)},p._b=function(t){return yit(this.c,t)},p.kc=function(){return new oct(this,this.c.b.c.gc())},p.lc=function(){return tce(this.c.b.c.gc(),16,new u3(this))},p.xc=function(t){var i;return i=f(U7(this.c,t),17),i?this.td(i.a):null},p.dc=function(){return this.c.b.c.dc()},p.ec=function(){return uce(this.c)},p.yc=function(t,i){var a;if(a=f(U7(this.c,t),17),!a)throw ie(new dn(this.sd()+" "+t+" not in "+uce(this.c)));return this.ud(a.a,i)},p.Ac=function(t){throw ie(new cr)},p.gc=function(){return this.c.b.c.gc()},R(Rt,"ArrayTable/ArrayMap",821),N(1966,1,{},u3),p.rd=function(t){return ldt(this.a,t)},R(Rt,"ArrayTable/ArrayMap/0methodref$getEntry$Type",1966),N(1964,355,ufe,Kit),p.jd=function(){return Xmn(this.a,this.b)},p.kd=function(){return this.a.td(this.b)},p.ld=function(t){return this.a.ud(this.b,t)},p.b=0,R(Rt,"ArrayTable/ArrayMap/1",1964),N(1965,390,xL,oct),p.Xb=function(t){return ldt(this.a,t)},R(Rt,"ArrayTable/ArrayMap/2",1965),N(1963,821,YS,Jlt),p.sd=function(){return"Column"},p.td=function(t){return sR(this.b,this.a,t)},p.ud=function(t,i){return h1t(this.b,this.a,t,i)},p.a=0,R(Rt,"ArrayTable/Row",1963),N(822,821,YS,n_e),p.td=function(t){return new Jlt(this.a,t)},p.yc=function(t,i){return f(i,91),Spn()},p.ud=function(t,i){return f(i,91),_pn()},p.sd=function(){return"Row"},R(Rt,"ArrayTable/RowMap",822),N(1121,1,pf,Yit),p.yd=function(t){return(this.a.wd()&-262&t)!=0},p.wd=function(){return this.a.wd()&-262},p.xd=function(){return this.a.xd()},p.Nb=function(t){this.a.Nb(new Jit(t,this.b))},p.zd=function(t){return this.a.zd(new Xit(t,this.b))},R(Rt,"CollectSpliterators/1",1121),N(1122,1,xn,Xit),p.Ad=function(t){this.a.Ad(this.b.Kb(t))},R(Rt,"CollectSpliterators/1/lambda$0$Type",1122),N(1123,1,xn,Jit),p.Ad=function(t){this.a.Ad(this.b.Kb(t))},R(Rt,"CollectSpliterators/1/lambda$1$Type",1123),N(1118,1,pf,kut),p.yd=function(t){return((16464|this.b)&t)!=0},p.wd=function(){return 16464|this.b},p.xd=function(){return this.a.xd()},p.Nb=function(t){this.a.Oe(new Qit(t,this.c))},p.zd=function(t){return this.a.Pe(new Zit(t,this.c))},p.b=0,R(Rt,"CollectSpliterators/1WithCharacteristics",1118),N(1119,1,$G,Zit),p.Bd=function(t){this.a.Ad(this.b.rd(t))},R(Rt,"CollectSpliterators/1WithCharacteristics/lambda$0$Type",1119),N(1120,1,$G,Qit),p.Bd=function(t){this.a.Ad(this.b.rd(t))},R(Rt,"CollectSpliterators/1WithCharacteristics/lambda$1$Type",1120),N(1114,1,pf),p.yd=function(t){return(this.a&t)!=0},p.wd=function(){return this.a},p.xd=function(){return this.e&&(this.b=USe(this.b,this.e.xd())),USe(this.b,0)},p.Nb=function(t){this.e&&(this.e.Nb(t),this.e=null),this.c.Nb(new eot(this,t)),this.b=0},p.zd=function(t){for(;;){if(this.e&&this.e.zd(t))return N7(this.b,FG)&&(this.b=hh(this.b,1)),!0;if(this.e=null,!this.c.zd(new pN(this)))return!1}},p.a=0,p.b=0,R(Rt,"CollectSpliterators/FlatMapSpliterator",1114),N(1116,1,xn,pN),p.Ad=function(t){Vbn(this.a,t)},R(Rt,"CollectSpliterators/FlatMapSpliterator/lambda$0$Type",1116),N(1117,1,xn,eot),p.Ad=function(t){Zwn(this.a,this.b,t)},R(Rt,"CollectSpliterators/FlatMapSpliterator/lambda$1$Type",1117),N(1115,1114,pf,Yht),R(Rt,"CollectSpliterators/FlatMapSpliteratorOfObject",1115),N(252,1,lfe),p.Dd=function(t){return this.Cd(f(t,252))},p.Cd=function(t){var i;return t==(wae(),jpe)?1:t==(vae(),Mpe)?-1:(i=(kX(),zH(this.a,t.a)),i!=0?i:me(this,511)==me(t,511)?0:me(this,511)?1:-1)},p.Gd=function(){return this.a},p.Fb=function(t){return I3e(this,t)},R(Rt,"Cut",252),N(1769,252,lfe,Brt),p.Cd=function(t){return t==this?0:1},p.Ed=function(t){throw ie(new s2e)},p.Fd=function(t){t.a+="+∞)"},p.Gd=function(){throw ie(new As(Qkt))},p.Hb=function(){return b0(),m3e(this)},p.Hd=function(t){return!1},p.Ib=function(){return"+∞"};var Mpe;R(Rt,"Cut/AboveAll",1769),N(511,252,{252:1,511:1,3:1,34:1},dct),p.Ed=function(t){vu((t.a+="(",t),this.a)},p.Fd=function(t){Qy(vu(t,this.a),93)},p.Hb=function(){return~Ti(this.a)},p.Hd=function(t){return kX(),zH(this.a,t)<0},p.Ib=function(){return"/"+this.a+"\\"},R(Rt,"Cut/AboveValue",511),N(1768,252,lfe,Frt),p.Cd=function(t){return t==this?0:-1},p.Ed=function(t){t.a+="(-∞"},p.Fd=function(t){throw ie(new s2e)},p.Gd=function(){throw ie(new As(Qkt))},p.Hb=function(){return b0(),m3e(this)},p.Hd=function(t){return!0},p.Ib=function(){return"-∞"};var jpe;R(Rt,"Cut/BelowAll",1768),N(1770,252,lfe,fct),p.Ed=function(t){vu((t.a+="[",t),this.a)},p.Fd=function(t){Qy(vu(t,this.a),41)},p.Hb=function(){return Ti(this.a)},p.Hd=function(t){return kX(),zH(this.a,t)<=0},p.Ib=function(){return"\\"+this.a+"/"},R(Rt,"Cut/BelowValue",1770),N(532,1,Eb),p.Ic=function(t){Oa(this,t)},p.Ib=function(){return rCn(f(JX(this,"use Optional.orNull() instead of Optional.or(null)"),20).Jc())},R(Rt,"FluentIterable",532),N(430,532,Eb,L7),p.Jc=function(){return new Tn(Cn(this.a.Jc(),new B))},R(Rt,"FluentIterable/2",430),N(1034,532,Eb,gst),p.Jc=function(){return pb(this)},R(Rt,"FluentIterable/3",1034),N(708,390,xL,r_e),p.Xb=function(t){return this.a[t].Jc()},R(Rt,"FluentIterable/3/1",708),N(2016,1,{}),p.Ib=function(){return Qs(this.Id().b)},R(Rt,"ForwardingObject",2016),N(2017,2016,eTt),p.Id=function(){return this.Jd()},p.Ic=function(t){Oa(this,t)},p.Lc=function(){return new Bt(this,0)},p.Mc=function(){return new Ft(null,this.Lc())},p.Ec=function(t){return this.Jd(),kit()},p.Fc=function(t){return this.Jd(),Tit()},p.$b=function(){this.Jd(),xit()},p.Gc=function(t){return this.Jd().Gc(t)},p.Hc=function(t){return this.Jd().Hc(t)},p.dc=function(){return this.Jd().b.dc()},p.Jc=function(){return this.Jd().Jc()},p.Kc=function(t){return this.Jd(),Cit()},p.gc=function(){return this.Jd().b.gc()},p.Nc=function(){return this.Jd().Nc()},p.Oc=function(t){return this.Jd().Oc(t)},R(Rt,"ForwardingCollection",2017),N(2024,31,ROe),p.Jc=function(){return this.Md()},p.Ec=function(t){throw ie(new cr)},p.Fc=function(t){throw ie(new cr)},p.Kd=function(){var t;return t=this.c,t||(this.c=this.Ld())},p.$b=function(){throw ie(new cr)},p.Gc=function(t){return t!=null&&_T(this,t,!1)},p.Ld=function(){switch(this.gc()){case 0:return _S(),_S(),$pe;case 1:return _S(),new nce(or(this.Md().Pb()));default:return new Pke(this,this.Nc())}},p.Kc=function(t){throw ie(new cr)},R(Rt,"ImmutableCollection",2024),N(711,2024,ROe,o2e),p.Jc=function(){return cR(this.a.Jc())},p.Gc=function(t){return t!=null&&this.a.Gc(t)},p.Hc=function(t){return this.a.Hc(t)},p.dc=function(){return this.a.dc()},p.Md=function(){return cR(this.a.Jc())},p.gc=function(){return this.a.gc()},p.Nc=function(){return this.a.Nc()},p.Oc=function(t){return this.a.Oc(t)},p.Ib=function(){return Qs(this.a)},R(Rt,"ForwardingImmutableCollection",711),N(308,2024,CL),p.Jc=function(){return this.Md()},p.cd=function(){return this.Nd(0)},p.dd=function(t){return this.Nd(t)},p.gd=function(t){sE(this,t)},p.Lc=function(){return new Bt(this,16)},p.hd=function(t,i){return this.Od(t,i)},p._c=function(t,i){throw ie(new cr)},p.ad=function(t,i){throw ie(new cr)},p.Kd=function(){return this},p.Fb=function(t){return V4n(this,t)},p.Hb=function(){return bTn(this)},p.bd=function(t){return t==null?-1:TAn(this,t)},p.Md=function(){return this.Nd(0)},p.Nd=function(t){return Pse(this,t)},p.ed=function(t){throw ie(new cr)},p.fd=function(t,i){throw ie(new cr)},p.Od=function(t,i){var a;return cZ((a=new sot(this),new Xv(a,t,i)))};var $pe;R(Rt,"ImmutableList",308),N(2051,308,CL),p.Jc=function(){return cR(this.Pd().Jc())},p.hd=function(t,i){return cZ(this.Pd().hd(t,i))},p.Gc=function(t){return t!=null&&this.Pd().Gc(t)},p.Hc=function(t){return this.Pd().Hc(t)},p.Fb=function(t){return ei(this.Pd(),t)},p.Xb=function(t){return Hv(this,t)},p.Hb=function(){return Ti(this.Pd())},p.bd=function(t){return this.Pd().bd(t)},p.dc=function(){return this.Pd().dc()},p.Md=function(){return cR(this.Pd().Jc())},p.gc=function(){return this.Pd().gc()},p.Od=function(t,i){return cZ(this.Pd().hd(t,i))},p.Nc=function(){return this.Pd().Oc(Pe(Ro,Jt,1,this.Pd().gc(),5,1))},p.Oc=function(t){return this.Pd().Oc(t)},p.Ib=function(){return Qs(this.Pd())},R(Rt,"ForwardingImmutableList",2051),N(712,1,AL),p.vc=function(){return Yy(this)},p.wc=function(t){HH(this,t)},p.ec=function(){return uce(this)},p.Bc=function(){return this.Td()},p.$b=function(){throw ie(new cr)},p._b=function(t){return this.xc(t)!=null},p.uc=function(t){return this.Td().Gc(t)},p.Rd=function(){return new Wnt(this)},p.Sd=function(){return new Knt(this)},p.Fb=function(t){return axn(this,t)},p.Hb=function(){return Yy(this).Hb()},p.dc=function(){return this.gc()==0},p.yc=function(t,i){return kpn()},p.Ac=function(t){throw ie(new cr)},p.Ib=function(){return INn(this)},p.Td=function(){return this.e?this.e:this.e=this.Sd()},p.c=null,p.d=null,p.e=null;var AAt;R(Rt,"ImmutableMap",712),N(713,712,AL),p._b=function(t){return yit(this,t)},p.uc=function(t){return got(this.b,t)},p.Qd=function(){return p0t(new B5(this))},p.Rd=function(){return p0t(Mdt(this.b))},p.Sd=function(){return lm(),new o2e(jdt(this.b))},p.Fb=function(t){return mot(this.b,t)},p.xc=function(t){return U7(this,t)},p.Hb=function(){return Ti(this.b.c)},p.dc=function(){return this.b.c.dc()},p.gc=function(){return this.b.c.gc()},p.Ib=function(){return Qs(this.b.c)},R(Rt,"ForwardingImmutableMap",713),N(2018,2017,dfe),p.Id=function(){return this.Ud()},p.Jd=function(){return this.Ud()},p.Lc=function(){return new Bt(this,1)},p.Fb=function(t){return t===this||this.Ud().Fb(t)},p.Hb=function(){return this.Ud().Hb()},R(Rt,"ForwardingSet",2018),N(1050,2018,dfe,B5),p.Id=function(){return TP(this.a.b)},p.Jd=function(){return TP(this.a.b)},p.Gc=function(t){if(me(t,43)&&f(t,43).jd()==null)return!1;try{return bot(TP(this.a.b),t)}catch(i){if(i=go(i),me(i,211))return!1;throw ie(i)}},p.Ud=function(){return TP(this.a.b)},p.Oc=function(t){var i;return i=bft(TP(this.a.b),t),TP(this.a.b).b.gc()<i.length&&io(i,TP(this.a.b).b.gc(),null),i},R(Rt,"ForwardingImmutableMap/1",1050),N(2025,2024,CR),p.Jc=function(){return this.Md()},p.Lc=function(){return new Bt(this,1)},p.Fb=function(t){return _Ae(this,t)},p.Hb=function(){return Lbt(this)},R(Rt,"ImmutableSet",2025),N(703,2025,CR),p.Jc=function(){return cR(new p3(this.a.b.Jc()))},p.Gc=function(t){return t!=null&&Iz(this.a,t)},p.Hc=function(t){return hot(this.a,t)},p.Hb=function(){return Ti(this.a.b)},p.dc=function(){return this.a.b.dc()},p.Md=function(){return cR(new p3(this.a.b.Jc()))},p.gc=function(){return this.a.b.gc()},p.Nc=function(){return this.a.b.Nc()},p.Oc=function(t){return pot(this.a,t)},p.Ib=function(){return Qs(this.a.b)},R(Rt,"ForwardingImmutableSet",703),N(2019,2018,tTt),p.Id=function(){return this.b},p.Jd=function(){return this.b},p.Ud=function(){return this.b},p.Lc=function(){return new OY(this)},R(Rt,"ForwardingSortedSet",2019),N(528,2023,AL,NZ),p.zc=function(t){mj(this,t)},p.Bc=function(){var t;return t=this.d,new Tse(t||(this.d=new By(this)))},p.$b=function(){kH(this)},p._b=function(t){return!!pj(this,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))))},p.uc=function(t){return jmt(this,t)},p.kc=function(){return new act(this,this)},p.wc=function(t){Ift(this,t)},p.xc=function(t){return F3(this,t)},p.ec=function(){return new xse(this)},p.yc=function(t,i){return cQ(this,t,i)},p.Ac=function(t){var i;return i=pj(this,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15)))),i?(tA(this,i),i.e=null,i.c=null,i.i):null},p.gc=function(){return this.i},p.vd=function(){var t;return t=this.d,new Tse(t||(this.d=new By(this)))},p.f=0,p.g=0,p.i=0,R(Rt,"HashBiMap",528),N(529,1,aa),p.Nb=function(t){va(this,t)},p.Ob=function(){return Ght(this)},p.Pb=function(){var t;if(!Ght(this))throw ie(new ic);return t=f(ch(this.c),304),this.c=t.c,this.f=t,--this.d,this.Vd(t)},p.Qb=function(){if(this.e.g!=this.b)throw ie(new If);if(!this.f)throw ie(new As(IOe));tA(this.e,this.f),this.b=this.e.g,this.f=null},p.b=0,p.d=0,p.f=null,R(Rt,"HashBiMap/Itr",529),N(998,529,aa,act),p.Vd=function(t){return new tot(this,t)},R(Rt,"HashBiMap/1",998),N(999,355,ufe,tot),p.jd=function(){return this.a.g},p.kd=function(){return this.a.i},p.ld=function(t){var i,a,u;return a=this.a.i,u=mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))),u==this.a.f&&(Ae(t)===Ae(a)||t!=null&&ei(t,a))?t:(g1t(!gj(this.b.a,t,u),t),tA(this.b.a,this.a),i=new qX(this.a.g,this.a.a,t,u),yG(this.b.a,i,this.a),this.a.e=null,this.a.c=null,this.b.b=this.b.a.g,this.b.f==this.a&&(this.b.f=i),this.a=i,a)},R(Rt,"HashBiMap/1/MapEntry",999),N(244,355,{355:1,244:1,3:1,43:1},dS),p.jd=function(){return this.g},p.kd=function(){return this.i},p.ld=function(t){throw ie(new cr)},R(Rt,"ImmutableEntry",244),N(304,244,{355:1,304:1,244:1,3:1,43:1},qX),p.a=0,p.f=0;var Fpe=R(Rt,"HashBiMap/BiEntry",304);N(604,2023,AL,By),p.zc=function(t){mj(this,t)},p.Bc=function(){return new xse(this.a)},p.$b=function(){kH(this.a)},p._b=function(t){return jmt(this.a,t)},p.kc=function(){return new sct(this,this.a)},p.wc=function(t){or(t),Ift(this.a,new K9(t))},p.xc=function(t){return BJ(this,t)},p.ec=function(){return new Tse(this)},p.yc=function(t,i){return GLn(this.a,t,i)},p.Ac=function(t){var i;return i=gj(this.a,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15)))),i?(tA(this.a,i),i.e=null,i.c=null,i.g):null},p.gc=function(){return this.a.i},p.vd=function(){return new xse(this.a)},R(Rt,"HashBiMap/Inverse",604),N(995,529,aa,sct),p.Vd=function(t){return new not(this,t)},R(Rt,"HashBiMap/Inverse/1",995),N(996,355,ufe,not),p.jd=function(){return this.a.i},p.kd=function(){return this.a.g},p.ld=function(t){var i,a,u;return u=this.a.g,i=mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))),i==this.a.a&&(Ae(t)===Ae(u)||t!=null&&ei(t,u))?t:(g1t(!pj(this.b.a.a,t,i),t),tA(this.b.a.a,this.a),a=new qX(t,i,this.a.i,this.a.f),this.a=a,yG(this.b.a.a,a,null),this.b.b=this.b.a.a.g,u)},R(Rt,"HashBiMap/Inverse/1/InverseEntry",996),N(605,527,Pl,Tse),p.Jc=function(){return new Hrt(this.a.a)},p.Kc=function(t){var i;return i=gj(this.a.a,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15)))),i?(tA(this.a.a,i),!0):!1},R(Rt,"HashBiMap/Inverse/InverseKeySet",605),N(994,529,aa,Hrt),p.Vd=function(t){return t.i},R(Rt,"HashBiMap/Inverse/InverseKeySet/1",994),N(997,1,{},K9),p.Wd=function(t,i){Xhn(this.a,t,i)},R(Rt,"HashBiMap/Inverse/lambda$0$Type",997),N(603,527,Pl,xse),p.Jc=function(){return new Grt(this.a)},p.Kc=function(t){var i;return i=pj(this.a,t,mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15)))),i?(tA(this.a,i),i.e=null,i.c=null,!0):!1},R(Rt,"HashBiMap/KeySet",603),N(993,529,aa,Grt),p.Vd=function(t){return t.g},R(Rt,"HashBiMap/KeySet/1",993),N(1087,612,wE),R(Rt,"HashMultimapGwtSerializationDependencies",1087),N(269,1087,wE,CS),p.hc=function(){return new bY(mT(this.a))},p.nd=function(){return new bY(mT(this.a))},p.a=2,R(Rt,"HashMultimap",269),N(2043,308,CL),p.Gc=function(t){return this.Xd().Gc(t)},p.dc=function(){return this.Xd().dc()},p.gc=function(){return this.Xd().gc()},R(Rt,"ImmutableAsList",2043),N(1976,713,AL),p.Td=function(){return lm(),new NN(this.a)},p.Bc=function(){return lm(),new NN(this.a)},p.vd=function(){return lm(),new NN(this.a)},R(Rt,"ImmutableBiMap",1976),N(2021,1,{}),R(Rt,"ImmutableCollection/Builder",2021),N(1010,703,CR,qrt),R(Rt,"ImmutableEnumSet",1010),N(957,390,xL,Sut),p.Xb=function(t){return this.a.Xb(t)},R(Rt,"ImmutableList/1",957),N(956,2021,{},Xct),R(Rt,"ImmutableList/Builder",956),N(608,204,TL,gN),p.Ob=function(){return this.a.Ob()},p.Pb=function(){return f(this.a.Pb(),43).jd()},R(Rt,"ImmutableMap/1",608),N(1029,1,{},$),p.Kb=function(t){return f(t,43).jd()},R(Rt,"ImmutableMap/2methodref$getKey$Type",1029),N(1028,1,{},Jct),R(Rt,"ImmutableMap/Builder",1028),N(2044,2025,CR),p.Kd=function(){var t;return t=this.b,t||(this.b=new lae(this))},p.Ld=function(){return new Pke(this,Rj(this,Pe(Ro,Jt,1,this.gc(),5,1)))},R(Rt,"ImmutableSet/CachingAsList",2044),N(2045,2044,CR),p.Jc=function(){var t;return t=Yy(this.a).Md(),new gN(t)},p.Ld=function(){return new lae(this)},p.Ic=function(t){var i,a;for(or(t),a=this.gc(),i=0;i<a;i++)t.Ad(f(Yy(this.a).Kd().Xb(i),43).jd())},p.Md=function(){var t;return t=this.b,Pse(t||(this.b=new lae(this)),0)},p.Lc=function(){return tce(this.gc(),1296,new Zd(this))},R(Rt,"IndexedImmutableSet",2045),N(1190,2045,CR,Wnt),p.Jc=function(){var t;return t=Yy(this.a).Md(),new gN(t)},p.Gc=function(t){return this.a._b(t)},p.Ic=function(t){or(t),HH(this.a,new cz(t))},p.Md=function(){var t;return t=Yy(this.a).Md(),new gN(t)},p.gc=function(){return this.a.gc()},p.Lc=function(){return sH(Yy(this.a).Lc(),new $)},R(Rt,"ImmutableMapKeySet",1190),N(1191,1,{},cz),p.Wd=function(t,i){lm(),this.a.Ad(t)},R(Rt,"ImmutableMapKeySet/lambda$0$Type",1191),N(1187,2024,ROe,Knt),p.Jc=function(){return new Qse(this)},p.Kd=function(){var t;return t=Yy(this.a).Kd(),new Nst(this,t)},p.Gc=function(t){return t!=null&&zIn(new Qse(this),t)},p.Md=function(){return new Qse(this)},p.gc=function(){return this.a.gc()},p.Lc=function(){return sH(Yy(this.a).Lc(),new U)},R(Rt,"ImmutableMapValues",1187),N(1188,1,{},U),p.Kb=function(t){return f(t,43).kd()},R(Rt,"ImmutableMapValues/0methodref$getValue$Type",1188),N(622,204,TL,Qse),p.Ob=function(){return this.a.Ob()},p.Pb=function(){return f(this.a.Pb(),43).kd()},R(Rt,"ImmutableMapValues/1",622),N(1189,2043,CL,Nst),p.Xd=function(){return this.a},p.Xb=function(t){return f(this.b.Xb(t),43).kd()},R(Rt,"ImmutableMapValues/2",1189),N(1192,1,{},Zd),p.rd=function(t){return hdt(this.a,t)},R(Rt,"IndexedImmutableSet/0methodref$get$Type",1192),N(623,2043,CL,lae),p.Xd=function(){return this.a},p.Xb=function(t){return hdt(this.a,t)},p.gc=function(){return this.a.a.gc()},R(Rt,"IndexedImmutableSet/1",623),N(41,1,{},B),p.Kb=function(t){return f(t,20).Jc()},p.Fb=function(t){return this===t},R(Rt,"Iterables/10",41),N(1030,532,Eb,elt),p.Ic=function(t){or(t),this.b.Ic(new rot(this.a,t))},p.Jc=function(){return BSe(this)},R(Rt,"Iterables/4",1030),N(1031,1,xn,rot),p.Ad=function(t){bgn(this.b,this.a,t)},R(Rt,"Iterables/4/lambda$0$Type",1031),N(1032,532,Eb,tlt),p.Ic=function(t){or(t),Oa(this.a,new oot(t,this.b))},p.Jc=function(){return Cn(new Nn(this.a),this.b)},R(Rt,"Iterables/5",1032),N(1033,1,xn,oot),p.Ad=function(t){this.a.Ad(Sst(t))},R(Rt,"Iterables/5/lambda$0$Type",1033),N(1052,204,TL,Af),p.Ob=function(){return this.a.Ob()},p.Pb=function(){return this.a.Pb()},R(Rt,"Iterators/1",1052),N(1053,697,TL,iot),p.Yb=function(){for(var t;this.b.Ob();)if(t=this.b.Pb(),this.a.Lb(t))return t;return this.e=2,null},R(Rt,"Iterators/5",1053),N(481,1,aa),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.b.Ob()},p.Pb=function(){return this.Yd(this.b.Pb())},p.Qb=function(){this.b.Qb()},R(Rt,"TransformedIterator",481),N(1054,481,aa,cct),p.Yd=function(t){return this.a.Kb(t)},R(Rt,"Iterators/6",1054),N(715,204,TL,Y9),p.Ob=function(){return!this.a},p.Pb=function(){if(this.a)throw ie(new ic);return this.a=!0,this.b},p.a=!1,R(Rt,"Iterators/9",715),N(1051,390,xL,Nlt),p.Xb=function(t){return this.a[this.b+t]},p.b=0;var IAt;R(Rt,"Iterators/ArrayItr",1051),N(37,1,{37:1,49:1},Tn),p.Nb=function(t){va(this,t)},p.Ob=function(){return Kn(this)},p.Pb=function(){return Dn(this)},p.Qb=function(){if(!this.c)throw ie(new As(IOe));this.c.Qb(),this.c=null},R(Rt,"Iterators/ConcatenatedIterator",37),N(23,1,{3:1,34:1,23:1}),p.Dd=function(t){return Jrt(this,f(t,23))},p.Fb=function(t){return this===t},p.Hb=function(){return bS(this)},p.Ib=function(){return Lse(this)},p.g=0;var ar=R(Sc,"Enum",23);N(534,23,{534:1,3:1,34:1,23:1,49:1},Zst),p.Nb=function(t){va(this,t)},p.Ob=function(){return!1},p.Pb=function(){throw ie(new ic)},p.Qb=function(){mP(!1)};var Bpe,OAt=er(Rt,"Iterators/EmptyModifiableIterator",534,ar,_wn,Ymn),NAt;N(1853,612,wE),R(Rt,"LinkedHashMultimapGwtSerializationDependencies",1853),N(1854,1853,wE,K1t),p.hc=function(){return new Nse(mT(this.b))},p.$b=function(){UJ(this),Ez(this.a,this.a)},p.nd=function(){return new Nse(mT(this.b))},p.ic=function(t){return new F1t(this,t,this.b)},p.kc=function(){return new Hke(this)},p.lc=function(){var t;return new Bt((t=this.g,f(t||(this.g=new T2e(this)),22)),17)},p.ec=function(){var t;return t=this.i,t||(this.i=new N3(this,this.c))},p.nc=function(){return new C2e(new Hke(this))},p.oc=function(){var t;return sH(new Bt((t=this.g,f(t||(this.g=new T2e(this)),22)),17),new ye)},p.b=2,R(Rt,"LinkedHashMultimap",1854),N(1857,1,{},ye),p.Kb=function(t){return f(t,43).kd()},R(Rt,"LinkedHashMultimap/0methodref$getValue$Type",1857),N(813,1,aa,Hke),p.Nb=function(t){va(this,t)},p.Pb=function(){return pTn(this)},p.Ob=function(){return this.a!=this.b.a},p.Qb=function(){mP(!!this.c),Iht(this.b,this.c.g,this.c.i),this.c=null},R(Rt,"LinkedHashMultimap/1",813),N(226,244,{355:1,244:1,226:1,588:1,3:1,43:1},Xke),p.Zd=function(){return f(ch(this.f),588)},p.$d=function(t){this.c=t},p._d=function(t){this.f=t},p.d=0;var RAt=R(Rt,"LinkedHashMultimap/ValueEntry",226);N(1855,2014,{588:1,20:1,31:1,16:1,22:1},F1t),p.Ec=function(t){var i,a,u,l,g;for(g=mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))),i=g&this.b.length-1,l=this.b[i],a=l;a;a=a.a)if(a.d==g&&dm(a.i,t))return!1;return u=new Xke(this.c,t,g,l),rit(this.d,u),u.f=this,this.d=u,Ez(f(ch(this.g.a.b),226),u),Ez(u,this.g.a),this.b[i]=u,++this.f,++this.e,FIn(this),!0},p.$b=function(){var t,i;for(Pz(this.b,null),this.f=0,t=this.a;t!=this;t=t.Zd())i=f(t,226),Ez(f(ch(i.b),226),f(ch(i.e),226));this.a=this,this.d=this,++this.e},p.Gc=function(t){var i,a;for(a=mr(Na(Sb,gb(mr(Na(t==null?0:Ti(t),_b)),15))),i=this.b[a&this.b.length-1];i;i=i.a)if(i.d==a&&dm(i.i,t))return!0;return!1},p.Ic=function(t){var i;for(or(t),i=this.a;i!=this;i=i.Zd())t.Ad(f(i,226).i)},p.Zd=function(){return this.a},p.Jc=function(){return new ydt(this)},p.Kc=function(t){return D2t(this,t)},p.$d=function(t){this.d=t},p._d=function(t){this.a=t},p.gc=function(){return this.f},p.e=0,p.f=0,R(Rt,"LinkedHashMultimap/ValueSet",1855),N(1856,1,aa,ydt),p.Nb=function(t){va(this,t)},p.Ob=function(){return bke(this),this.b!=this.c},p.Pb=function(){var t,i;if(bke(this),this.b==this.c)throw ie(new ic);return t=f(this.b,226),i=t.i,this.d=t,this.b=f(ch(t.f),588),i},p.Qb=function(){bke(this),mP(!!this.d),D2t(this.c,this.d.i),this.a=this.c.e,this.d=null},p.a=0,R(Rt,"LinkedHashMultimap/ValueSet/1",1856),N(762,2030,wE,ast),p.Zb=function(){var t;return t=this.f,t||(this.f=new oSe(this))},p.Fb=function(t){return tle(this,t)},p.cc=function(t){return new Rae(this,t)},p.fc=function(t){return rxe(this,t)},p.$b=function(){edt(this)},p._b=function(t){return dot(this,t)},p.ac=function(){return new oSe(this)},p.bc=function(){return new uz(this)},p.qc=function(t){return new Rae(this,t)},p.dc=function(){return!this.a},p.rc=function(t){return rxe(this,t)},p.gc=function(){return this.d},p.c=0,p.d=0,R(Rt,"LinkedListMultimap",762),N(55,31,IL),p.gd=function(t){sE(this,t)},p.Lc=function(){return new Bt(this,16)},p._c=function(t,i){throw ie(new f0("Add not supported on this list"))},p.Ec=function(t){return this._c(this.gc(),t),!0},p.ad=function(t,i){var a,u,l;for(an(i),a=!1,l=i.Jc();l.Ob();)u=l.Pb(),this._c(t++,u),a=!0;return a},p.$b=function(){this.ae(0,this.gc())},p.Fb=function(t){return oIe(this,t)},p.Hb=function(){return hCe(this)},p.bd=function(t){return cbt(this,t)},p.Jc=function(){return new Fk(this)},p.cd=function(){return this.dd(0)},p.dd=function(t){return new la(this,t)},p.ed=function(t){throw ie(new f0("Remove not supported on this list"))},p.ae=function(t,i){var a,u;for(u=this.dd(t),a=t;a<i;++a)u.Pb(),u.Qb()},p.fd=function(t,i){throw ie(new f0("Set not supported on this list"))},p.hd=function(t,i){return new Xv(this,t,i)},p.j=0,R(Gn,"AbstractList",55),N(2008,55,IL),p._c=function(t,i){pP(this,t,i)},p.ad=function(t,i){return z1t(this,t,i)},p.Xb=function(t){return Xh(this,t)},p.Jc=function(){return this.dd(0)},p.ed=function(t){return Zle(this,t)},p.fd=function(t,i){var a,u;a=this.dd(t);try{return u=a.Pb(),a.Wb(i),u}catch(l){throw l=go(l),me(l,111)?ie(new Ru("Can't set element "+t)):ie(l)}},R(Gn,"AbstractSequentialList",2008),N(631,2008,IL,Rae),p.dd=function(t){return lct(this,t)},p.gc=function(){var t;return t=f(bn(this.a.b,this.b),260),t?t.a:0},R(Rt,"LinkedListMultimap/1",631),N(1269,2014,Pl,uz),p.Gc=function(t){return dot(this.a,t)},p.Jc=function(){return new Ibt(this.a)},p.Kc=function(t){return!rxe(this.a,t).a.dc()},p.gc=function(){return f7(this.a.b)},R(Rt,"LinkedListMultimap/1KeySetImpl",1269),N(1268,1,aa,Ibt),p.Nb=function(t){va(this,t)},p.Ob=function(){return vke(this),!!this.c},p.Pb=function(){if(vke(this),!this.c)throw ie(new ic);this.a=this.c,Go(this.d,this.a.a);do this.c=this.c.b;while(this.c&&!Go(this.d,this.c.a));return this.a.a},p.Qb=function(){vke(this),mP(!!this.a),CX(new wue(this.e,this.a.a)),this.a=null,this.b=this.e.c},p.b=0,R(Rt,"LinkedListMultimap/DistinctKeyIterator",1268),N(260,1,{260:1},cTe),p.a=0,R(Rt,"LinkedListMultimap/KeyList",260),N(495,355,{355:1,495:1,43:1},aot),p.jd=function(){return this.a},p.kd=function(){return this.f},p.ld=function(t){var i;return i=this.f,this.f=t,i},R(Rt,"LinkedListMultimap/Node",495),N(551,1,yb,wue,Byt),p.Nb=function(t){va(this,t)},p.Rb=function(t){this.e=zIe(this.f,this.b,t,this.c),++this.d,this.a=null},p.Ob=function(){return!!this.c},p.Sb=function(){return!!this.e},p.Pb=function(){return wxe(this)},p.Tb=function(){return this.d},p.Ub=function(){return gmt(this)},p.Vb=function(){return this.d-1},p.Qb=function(){mP(!!this.a),this.a!=this.c?(this.e=this.a.e,--this.d):this.c=this.a.c,U6n(this.f,this.a),this.a=null},p.Wb=function(t){e_e(!!this.a),this.a.f=t},p.d=0,R(Rt,"LinkedListMultimap/ValueForKeyIterator",551),N(1006,55,IL),p._c=function(t,i){this.a._c(t,i)},p.ad=function(t,i){return this.a.ad(t,i)},p.Gc=function(t){return this.a.Gc(t)},p.Xb=function(t){return this.a.Xb(t)},p.ed=function(t){return this.a.ed(t)},p.fd=function(t,i){return this.a.fd(t,i)},p.gc=function(){return this.a.gc()},R(Rt,"Lists/AbstractListWrapper",1006),N(1007,1006,rTt),R(Rt,"Lists/RandomAccessListWrapper",1007),N(1009,1007,rTt,sot),p.dd=function(t){return this.a.dd(t)},R(Rt,"Lists/1",1009),N(429,55,{429:1,20:1,31:1,55:1,16:1,15:1},Y2e),p._c=function(t,i){this.a._c(EP(this,t),i)},p.$b=function(){this.a.$b()},p.Xb=function(t){return this.a.Xb(Uke(this,t))},p.Jc=function(){return nht(this,0)},p.dd=function(t){return nht(this,t)},p.ed=function(t){return this.a.ed(Uke(this,t))},p.ae=function(t,i){(Pht(t,i,this.a.gc()),Bd(this.a.hd(EP(this,i),EP(this,t)))).$b()},p.fd=function(t,i){return this.a.fd(Uke(this,t),i)},p.gc=function(){return this.a.gc()},p.hd=function(t,i){return Pht(t,i,this.a.gc()),Bd(this.a.hd(EP(this,i),EP(this,t)))},R(Rt,"Lists/ReverseList",429),N(1005,429,{429:1,20:1,31:1,55:1,16:1,15:1,59:1},Vrt),R(Rt,"Lists/RandomAccessReverseList",1005),N(1008,1,yb,cot),p.Nb=function(t){va(this,t)},p.Rb=function(t){this.c.Rb(t),this.c.Ub(),this.a=!1},p.Ob=function(){return this.c.Sb()},p.Sb=function(){return this.c.Ob()},p.Pb=function(){if(!this.c.Sb())throw ie(new ic);return this.a=!0,this.c.Ub()},p.Tb=function(){return EP(this.b,this.c.Tb())},p.Ub=function(){if(!this.c.Ob())throw ie(new ic);return this.a=!0,this.c.Pb()},p.Vb=function(){return EP(this.b,this.c.Tb())-1},p.Qb=function(){mP(this.a),this.c.Qb(),this.a=!1},p.Wb=function(t){e_e(this.a),this.c.Wb(t)},p.a=!1,R(Rt,"Lists/ReverseList/1",1008),N(428,481,aa,s7),p.Yd=function(t){return IY(t)},R(Rt,"Maps/1",428),N(696,481,aa,C2e),p.Yd=function(t){return f(t,43).kd()},R(Rt,"Maps/2",696),N(953,481,aa,uct),p.Yd=function(t){return new dS(t,lst(this.a,t))},R(Rt,"Maps/3",953),N(950,2015,Pl,l3),p.Ic=function(t){tpn(this.a,t)},p.Jc=function(){return this.a.kc()},p.Pc=function(){return this.a},p.Lc=function(){return this.a.lc()},R(Rt,"Maps/IteratorBasedAbstractMap/1",950),N(951,1,{},mN),p.Wd=function(t,i){this.a.Ad(t)},R(Rt,"Maps/KeySet/lambda$0$Type",951),N(949,31,jT,uot),p.$b=function(){this.a.$b()},p.Gc=function(t){return this.a.uc(t)},p.Ic=function(t){or(t),this.a.wc(new lz(t))},p.dc=function(){return this.a.dc()},p.Jc=function(){return new C2e(this.a.vc().Jc())},p.Kc=function(t){var i,a;try{return _T(this,t,!0)}catch(u){if(u=go(u),me(u,46)){for(a=this.a.vc().Jc();a.Ob();)if(i=f(a.Pb(),43),dm(t,i.kd()))return this.a.Ac(i.jd()),!0;return!1}else throw ie(u)}},p.gc=function(){return this.a.gc()},R(Rt,"Maps/Values",949),N(952,1,{},lz),p.Wd=function(t,i){this.a.Ad(i)},R(Rt,"Maps/Values/lambda$0$Type",952),N(735,2031,YS,oSe),p.xc=function(t){return this.a._b(t)?this.a.cc(t):null},p.Ac=function(t){return this.a._b(t)?this.a.fc(t):null},p.$b=function(){this.a.$b()},p._b=function(t){return this.a._b(t)},p.Dc=function(){return new U5(this)},p.Cc=function(){return this.Dc()},p.dc=function(){return this.a.dc()},p.ec=function(){return this.a.ec()},p.gc=function(){return this.a.ec().gc()},R(Rt,"Multimaps/AsMap",735),N(1098,2015,Pl,U5),p.Jc=function(){return bmn(this.a.a.ec(),new dz(this))},p.Pc=function(){return this.a},p.Kc=function(t){var i;return K0t(this,t)?(i=f(ch(f(t,43)),43),dpn(this.a,i.jd()),!0):!1},R(Rt,"Multimaps/AsMap/EntrySet",1098),N(1102,1,{},dz),p.Kb=function(t){return lst(this,t)},p.Fb=function(t){return this===t},R(Rt,"Multimaps/AsMap/EntrySet/1",1102),N(537,2033,{537:1,828:1,20:1,31:1,16:1},Mk),p.$b=function(){UJ(this.a)},p.Gc=function(t){return wit(this.a,t)},p.Ic=function(t){or(t),Oa(J7(this.a),new X9(t))},p.Jc=function(){return new s7(J7(this.a).a.kc())},p.gc=function(){return this.a.d},p.Lc=function(){return sH(J7(this.a).Lc(),new we)},R(Rt,"Multimaps/Keys",537),N(1100,1,{},we),p.Kb=function(t){return f(t,43).jd()},R(Rt,"Multimaps/Keys/0methodref$getKey$Type",1100),N(1099,481,aa,Wrt),p.Yd=function(t){return new fz(f(t,43))},R(Rt,"Multimaps/Keys/1",1099),N(2034,1,{413:1}),p.Fb=function(t){var i;return me(t,488)?(i=f(t,413),f(this.a.kd(),16).gc()==f(i.a.kd(),16).gc()&&dm(this.a.jd(),i.a.jd())):!1},p.Hb=function(){var t;return t=this.a.jd(),(t==null?0:Ti(t))^f(this.a.kd(),16).gc()},p.Ib=function(){var t,i;return i=$7(this.a.jd()),t=f(this.a.kd(),16).gc(),t==1?i:i+" x "+t},R(Rt,"Multisets/AbstractEntry",2034),N(488,2034,{488:1,413:1},fz),R(Rt,"Multimaps/Keys/1/1",488),N(1101,1,xn,X9),p.Ad=function(t){this.a.Ad(f(t,43).jd())},R(Rt,"Multimaps/Keys/lambda$1$Type",1101),N(1104,1,xn,O),p.Ad=function(t){Rvn(f(t,413))},R(Rt,"Multiset/lambda$0$Type",1104),N(736,1,xn,J9),p.Ad=function(t){rkn(this.a,f(t,413))},R(Rt,"Multiset/lambda$1$Type",736),N(1105,1,{},P),R(Rt,"Multisets/0methodref$add$Type",1105),N(737,1,{},D),p.Kb=function(t){return U2n(f(t,413))},R(Rt,"Multisets/lambda$1$Type",737),N(2052,1,MG),R(Rt,"RangeGwtSerializationDependencies",2052),N(505,2052,{177:1,505:1,3:1,48:1},E3e),p.Lb=function(t){return Flt(this,f(t,34))},p.Mb=function(t){return Flt(this,f(t,34))},p.Fb=function(t){var i;return me(t,505)?(i=f(t,505),I3e(this.a,i.a)&&I3e(this.b,i.b)):!1},p.Hb=function(){return this.a.Hb()*31+this.b.Hb()},p.Ib=function(){return Jht(this.a,this.b)},R(Rt,"Range",505),N(637,2043,CL,Pke),p.dd=function(t){return Pse(this.b,t)},p.Xd=function(){return this.a},p.Xb=function(t){return Hv(this.b,t)},p.Nd=function(t){return Pse(this.b,t)},R(Rt,"RegularImmutableAsList",637),N(640,2051,CL,Fce),p.Pd=function(){return this.a},R(Rt,"RegularImmutableList",640),N(533,713,AL,A2e,I2e),R(Rt,"RegularImmutableMap",533),N(714,703,CR,nSe);var h6e;R(Rt,"RegularImmutableSet",714),N(2020,Ig,Pl),p.Jc=function(){return new eTe(this.a,this.b)},p.Ec=function(t){throw ie(new cr)},p.Fc=function(t){throw ie(new cr)},p.$b=function(){throw ie(new cr)},p.Kc=function(t){throw ie(new cr)},R(Rt,"Sets/SetView",2020),N(954,2020,Pl,lot),p.Jc=function(){return new eTe(this.a,this.b)},p.Gc=function(t){return oce(this.a,t)&&this.b.Gc(t)},p.Hc=function(t){return KH(this.a,t)&&this.b.Hc(t)},p.dc=function(){return Nvt(this.b,this.a)},p.gc=function(){return GH(this)},p.Mc=function(){return Jr(new Ft(null,new Bt(this.a,1)),new z5(this.b))},R(Rt,"Sets/2",954),N(955,1,_r,z5),p.Mb=function(t){return this.a.Gc(t)},R(Rt,"Sets/2/0methodref$contains$Type",955),N(698,697,TL,eTe),p.Yb=function(){for(var t;zSe(this.a);)if(t=aj(this.a),this.c.Gc(t))return t;return this.e=2,null},R(Rt,"Sets/2/1",698),N(601,2019,{601:1,3:1,20:1,16:1,274:1,22:1,83:1},Tft),p.Id=function(){return this.b},p.Jd=function(){return this.b},p.Ud=function(){return this.b},p.Ic=function(t){this.a.Ic(t)},p.Mc=function(){return this.a.Mc()},R(Rt,"Sets/UnmodifiableNavigableSet",601),N(1977,1976,AL,Idt),p.Td=function(){return lm(),new NN(this.a)},p.Bc=function(){return lm(),new NN(this.a)},p.vd=function(){return lm(),new NN(this.a)},R(Rt,"SingletonImmutableBiMap",1977),N(641,2051,CL,nce),p.Pd=function(){return this.a},R(Rt,"SingletonImmutableList",641),N(356,2025,CR,NN),p.Jc=function(){return new Y9(this.a)},p.Gc=function(t){return ei(this.a,t)},p.Md=function(){return new Y9(this.a)},p.gc=function(){return 1},R(Rt,"SingletonImmutableSet",356),N(1112,1,{},Te),p.Kb=function(t){return f(t,158)},R(Rt,"Streams/lambda$0$Type",1112),N(1113,1,EQ,H5),p.be=function(){FSn(this.a)},R(Rt,"Streams/lambda$1$Type",1113),N(1667,1666,wE,fft),p.Zb=function(){var t;return t=this.f,f(f(t||(this.f=me(this.c,137)?new B7(this,f(this.c,137)):me(this.c,134)?new eH(this,f(this.c,134)):new aP(this,this.c)),134),137)},p.hc=function(){return new g0(this.b)},p.nd=function(){return new g0(this.b)},p.ec=function(){var t;return t=this.i,f(f(t||(this.i=me(this.c,137)?new sP(this,f(this.c,137)):me(this.c,134)?new xz(this,f(this.c,134)):new N3(this,this.c)),83),274)},p.ac=function(){return me(this.c,137)?new B7(this,f(this.c,137)):me(this.c,134)?new eH(this,f(this.c,134)):new aP(this,this.c)},p.ic=function(t){return t==null&&this.a.Le(t,t),new g0(this.b)},R(Rt,"TreeMultimap",1667),N(80,1,{3:1,80:1}),p.ce=function(t){return new Error(t)},p.de=function(){return this.e},p.ee=function(){var t,i,a;for(a=(this.k==null&&(this.k=Pe(Upe,Ye,80,0,0,1)),this.k),i=Pe(Ro,Jt,1,a.length,5,1),t=0;t<a.length;t++)i[t]=a[t].e;return i},p.fe=function(){return this.f},p.ge=function(){return this.g},p.he=function(){apn(this,T2n(this.ce(QX(this,this.g)))),Ynt(this)},p.Ib=function(){return QX(this,this.ge())},p.e=iTt,p.i=!1,p.n=!0;var Upe=R(Sc,"Throwable",80);N(103,80,{3:1,103:1,80:1}),R(Sc,"Exception",103),N(63,103,C1,Gy,oc),R(Sc,"RuntimeException",63),N(591,63,C1),R(Sc,"JsException",591),N(851,591,C1),R(UG,"JavaScriptExceptionBase",851),N(472,851,{472:1,3:1,103:1,63:1,80:1},Y1t),p.ge=function(){return uOn(this),this.c},p.ie=function(){return Ae(this.b)===Ae(p6e)?null:this.b};var p6e;R(LOe,"JavaScriptException",472);var PAt=R(LOe,"JavaScriptObject$",0),zpe;N(1993,1,{}),R(LOe,"Scheduler",1993);var Cee=0,LAt=0,Aee=-1;N(878,1993,{},ge);var g6e;R(UG,"SchedulerImpl",878);var Hpe;N(2004,1,{}),R(UG,"StackTraceCreator/Collector",2004),N(852,2004,{},G),p.je=function(t){var i={},a=[];t[pfe]=a;for(var u=arguments.callee.caller;u;){var l=(AP(),u.name||(u.name=bkn(u.toString())));a.push(l);var g=":"+l,m=i[g];if(m){var v,S;for(v=0,S=m.length;v<S;v++)if(m[v]===u)return}(m||(i[g]=[])).push(u),u=u.caller}},p.ke=function(t){var i,a,u,l;for(u=(AP(),t&&t[pfe]?t[pfe]:[]),a=u.length,l=Pe(I6e,Ye,320,a,0,1),i=0;i<a;i++)l[i]=new Ace(u[i],null,-1);return l},R(UG,"StackTraceCreator/CollectorLegacy",852),N(2005,2004,{}),p.je=function(t){},p.le=function(t,i,a,u){return new Ace(i,t+"@"+u,a<0?-1:a)},p.ke=function(t){var i,a,u,l,g,m;if(l=FAn(t),g=Pe(I6e,Ye,320,0,0,1),i=0,u=l.length,u==0)return g;for(m=ckt(this,l[0]),Mt(m.d,hfe)||(g[i++]=m),a=1;a<u;a++)g[i++]=ckt(this,l[a]);return g},R(UG,"StackTraceCreator/CollectorModern",2005),N(853,2005,{},K),p.le=function(t,i,a,u){return new Ace(i,t,-1)},R(UG,"StackTraceCreator/CollectorModernNoSourceMap",853),N(1039,1,{}),R(MOe,sTt,1039),N(609,1039,{609:1},ddt);var m6e;R(Pfe,sTt,609),N(2047,1,{}),R(MOe,cTt,2047),N(pw,2047,{}),R(Pfe,cTt,pw),N(1084,1,{},X);var L$;R(Pfe,"LocaleInfo",1084),N(1973,1,{},ae),p.a=0,R(Pfe,"TimeZone",1973),N(1246,pw,{},he),R("com.google.gwt.i18n.client.impl.cldr","DateTimeFormatInfoImpl",1246),N(431,1,{431:1},Vut),p.a=!1,p.b=0,R(MOe,"DateTimeFormat/PatternPart",431),N(205,1,uTt,EY,gCe,v_e),p.Dd=function(t){return p2n(this,f(t,205))},p.Fb=function(t){return me(t,205)&&O7(Nc(this.q.getTime()),Nc(f(t,205).q.getTime()))},p.Hb=function(){var t;return t=Nc(this.q.getTime()),mr(Yce(t,Jy(t,32)))},p.Ib=function(){var t,i,a;return a=-this.q.getTimezoneOffset(),t=(a>=0?"+":"")+(a/60|0),i=tX(h.Math.abs(a)%60),(cyt(),WAt)[this.q.getDay()]+" "+KAt[this.q.getMonth()]+" "+tX(this.q.getDate())+" "+tX(this.q.getHours())+":"+tX(this.q.getMinutes())+":"+tX(this.q.getSeconds())+" GMT"+t+i+" "+this.q.getFullYear()};var Iee=R(Gn,"Date",205);N(1961,205,uTt,Iwt),p.a=!1,p.b=0,p.c=0,p.d=0,p.e=0,p.f=0,p.g=!1,p.i=0,p.j=0,p.k=0,p.n=0,p.o=0,p.p=0,R("com.google.gwt.i18n.shared.impl","DateRecord",1961),N(2010,1,{}),p.ne=function(){return null},p.oe=function(){return null},p.pe=function(){return null},p.qe=function(){return null},p.re=function(){return null},R(IR,"JSONValue",2010),N(138,2010,{138:1},oh,bN),p.Fb=function(t){return me(t,138)?STe(this.a,f(t,138).a):!1},p.me=function(){return Hhn},p.Hb=function(){return sTe(this.a)},p.ne=function(){return this},p.Ib=function(){var t,i,a;for(a=new ef("["),i=0,t=this.a.length;i<t;i++)i>0&&(a.a+=","),vu(a,gT(this,i));return a.a+="]",a.a},R(IR,"JSONArray",138),N(477,2010,{477:1},G5),p.me=function(){return Ghn},p.oe=function(){return this},p.Ib=function(){return mn(),""+this.a},p.a=!1;var DAt,MAt;R(IR,"JSONBoolean",477),N(974,63,C1,Krt),R(IR,"JSONException",974),N(1011,2010,{},oe),p.me=function(){return qhn},p.Ib=function(){return bl};var jAt;R(IR,"JSONNull",1011),N(263,2010,{263:1},oS),p.Fb=function(t){return me(t,263)?this.a==f(t,263).a:!1},p.me=function(){return Bhn},p.Hb=function(){return FN(this.a)},p.pe=function(){return this},p.Ib=function(){return this.a+""},p.a=0,R(IR,"JSONNumber",263),N(147,2010,{147:1},AN,d3),p.Fb=function(t){return me(t,147)?STe(this.a,f(t,147).a):!1},p.me=function(){return Uhn},p.Hb=function(){return sTe(this.a)},p.qe=function(){return this},p.Ib=function(){var t,i,a,u,l,g,m;for(m=new ef("{"),t=!0,g=Due(this,Pe(ft,Ye,2,0,6,1)),a=g,u=0,l=a.length;u<l;++u)i=a[u],t?t=!1:m.a+=Fu,xr(m,xEt(i)),m.a+=":",vu(m,h1(this,i));return m.a+="}",m.a},R(IR,"JSONObject",147),N(589,Ig,Pl,Pae),p.Gc=function(t){return ta(t)&&ypn(this.a,dr(t))},p.Jc=function(){return new Fk(new bu(this.b))},p.gc=function(){return this.b.length},R(IR,"JSONObject/1",589);var Gpe;N(210,2010,{210:1},cT),p.Fb=function(t){return me(t,210)?Mt(this.a,f(t,210).a):!1},p.me=function(){return zhn},p.Hb=function(){return C0(this.a)},p.re=function(){return this},p.Ib=function(){return xEt(this.a)},R(IR,"JSONString",210);var ww,b6e,$At,v6e,w6e;N(2006,1,{518:1}),R(jOe,"OutputStream",2006),N(2007,2006,{518:1}),R(jOe,"FilterOutputStream",2007),N(854,2007,{518:1},Ue),R(jOe,"PrintStream",854),N(415,1,{470:1}),p.Ib=function(){return this.a},R(Sc,"AbstractStringBuilder",415),N(523,63,C1,fY),R(Sc,"ArithmeticException",523),N(97,63,_Q,a2e,Ru),R(Sc,"IndexOutOfBoundsException",97),N(638,97,_Q,trt,P2e),R(Sc,"ArrayIndexOutOfBoundsException",638),N(522,63,C1,dae,tit),R(Sc,"ArrayStoreException",522),N(294,80,lTt,kae),R(Sc,"Error",294),N(200,294,lTt,s2e,UTe),R(Sc,"AssertionError",200),TAt={3:1,471:1,34:1};var yw,QL,to=R(Sc,"Boolean",471);N(241,1,{3:1,241:1});var y6e;R(Sc,"Number",241),N(219,241,{3:1,219:1,34:1,241:1},zn),p.Dd=function(t){return vpn(this,f(t,219))},p.se=function(){return this.a},p.Fb=function(t){return me(t,219)&&f(t,219).a==this.a},p.Hb=function(){return this.a},p.Ib=function(){return""+this.a},p.a=0;var BR=R(Sc,"Byte",219),E6e;N(179,1,{3:1,179:1,34:1},Ka),p.Dd=function(t){return wpn(this,f(t,179))},p.Fb=function(t){return me(t,179)&&f(t,179).a==this.a},p.Hb=function(){return this.a},p.Ib=function(){return String.fromCharCode(this.a)},p.a=0;var S6e,D$=R(Sc,"Character",179),_6e;N(211,63,{3:1,211:1,103:1,63:1,80:1},Xnt,eP),R(Sc,"ClassCastException",211),xAt={3:1,34:1,341:1,241:1};var mo=R(Sc,"Double",341);N(160,241,{3:1,34:1,160:1,241:1},f3,m2e),p.Dd=function(t){return Lgn(this,f(t,160))},p.se=function(){return this.a},p.Fb=function(t){return me(t,160)&&wut(this.a,f(t,160).a)},p.Hb=function(){return Aa(this.a)},p.Ib=function(){return""+this.a},p.a=0;var eD=R(Sc,"Float",160);N(32,63,{3:1,103:1,32:1,63:1,80:1},Sz,dn,W1t),R(Sc,"IllegalArgumentException",32),N(73,63,C1,Tl,As),R(Sc,"IllegalStateException",73),N(17,241,{3:1,34:1,17:1,241:1},KK),p.Dd=function(t){return DSe(this,f(t,17))},p.se=function(){return this.a},p.Fb=function(t){return me(t,17)&&f(t,17).a==this.a},p.Hb=function(){return this.a},p.Ib=function(){return""+this.a},p.a=0;var ba=R(Sc,"Integer",17),k6e,FAt;N(190,241,{3:1,34:1,190:1,241:1},Mv),p.Dd=function(t){return Dgn(this,f(t,190))},p.se=function(){return oE(this.a)},p.Fb=function(t){return me(t,190)&&O7(f(t,190).a,this.a)},p.Hb=function(){return Ibn(this.a)},p.Ib=function(){return""+V7(this.a)},p.a=0;var i_=R(Sc,"Long",190),T6e;N(2086,1,{}),N(1850,63,C1,nit),R(Sc,"NegativeArraySizeException",1850),N(169,591,{3:1,103:1,169:1,63:1,80:1},kN,IN),p.ce=function(t){return new TypeError(t)},R(Sc,"NullPointerException",169);var x6e,qpe,BAt,C6e;N(130,32,{3:1,103:1,32:1,130:1,63:1,80:1},sm),R(Sc,"NumberFormatException",130),N(191,241,{3:1,34:1,241:1,191:1},Fs),p.Dd=function(t){return Epn(this,f(t,191))},p.se=function(){return this.a},p.Fb=function(t){return me(t,191)&&f(t,191).a==this.a},p.Hb=function(){return this.a},p.Ib=function(){return""+this.a},p.a=0;var o_=R(Sc,"Short",191),A6e;N(320,1,{3:1,320:1},Ace),p.Fb=function(t){var i;return me(t,320)?(i=f(t,320),this.c==i.c&&this.d==i.d&&this.a==i.a&&this.b==i.b):!1},p.Hb=function(){return oZ(ce(se(Ro,1),Jt,1,5,[rt(this.c),this.a,this.d,this.b]))},p.Ib=function(){return this.a+"."+this.d+"("+(this.b!=null?this.b:"Unknown Source")+(this.c>=0?":"+this.c:"")+")"},p.c=0;var I6e=R(Sc,"StackTraceElement",320);CAt={3:1,470:1,34:1,2:1};var ft=R(Sc,POe,2);N(110,415,{470:1},p0,l7,sh),R(Sc,"StringBuffer",110),N(104,415,{470:1},$v,RN,ef),R(Sc,"StringBuilder",104),N(686,97,_Q,L2e),R(Sc,"StringIndexOutOfBoundsException",686),N(2091,1,{});var UAt;N(46,63,{3:1,103:1,63:1,80:1,46:1},cr,f0),R(Sc,"UnsupportedOperationException",46),N(245,241,{3:1,34:1,241:1,245:1},nG,G2e),p.Dd=function(t){return a_t(this,f(t,245))},p.se=function(){return AT(P_t(this))},p.Fb=function(t){var i;return this===t?!0:me(t,245)?(i=f(t,245),this.e==i.e&&a_t(this,i)==0):!1},p.Hb=function(){var t;return this.b!=0?this.b:this.a<54?(t=Nc(this.f),this.b=mr(ia(t,-1)),this.b=33*this.b+mr(ia(wS(t,32),-1)),this.b=17*this.b+Aa(this.e),this.b):(this.b=17*s0t(this.c)+Aa(this.e),this.b)},p.Ib=function(){return P_t(this)},p.a=0,p.b=0,p.d=0,p.e=0,p.f=0;var zAt,IE,O6e,N6e,R6e,P6e,L6e,D6e,Vpe=R("java.math","BigDecimal",245);N(90,241,{3:1,34:1,241:1,90:1},m1,Eht,Zy,hvt,Gv),p.Dd=function(t){return lvt(this,f(t,90))},p.se=function(){return AT(afe(this,0))},p.Fb=function(t){return e3e(this,t)},p.Hb=function(){return s0t(this)},p.Ib=function(){return afe(this,0)},p.b=-2,p.c=0,p.d=0,p.e=0;var HAt,Oee,GAt,Wpe,Nee,M$,mA=R("java.math","BigInteger",90),qAt,VAt,UR,j$;N(482,2011,YS),p.$b=function(){nu(this)},p._b=function(t){return Yu(this,t)},p.uc=function(t){return H1t(this,t,this.i)||H1t(this,t,this.f)},p.vc=function(){return new Vi(this)},p.xc=function(t){return bn(this,t)},p.yc=function(t,i){return qr(this,t,i)},p.Ac=function(t){return aR(this,t)},p.gc=function(){return f7(this)},p.g=0,R(Gn,"AbstractHashMap",482),N(302,Ig,Pl,Vi),p.$b=function(){this.a.$b()},p.Gc=function(t){return Lht(this,t)},p.Jc=function(){return new yT(this.a)},p.Kc=function(t){var i;return Lht(this,t)?(i=f(t,43).jd(),this.a.Ac(i),!0):!1},p.gc=function(){return this.a.gc()},R(Gn,"AbstractHashMap/EntrySet",302),N(303,1,aa,yT),p.Nb=function(t){va(this,t)},p.Pb=function(){return G3(this)},p.Ob=function(){return this.b},p.Qb=function(){rmt(this)},p.b=!1,p.d=0,R(Gn,"AbstractHashMap/EntrySetIterator",303),N(414,1,aa,Fk),p.Nb=function(t){va(this,t)},p.Ob=function(){return Fae(this)},p.Pb=function(){return iTe(this)},p.Qb=function(){hd(this)},p.b=0,p.c=-1,R(Gn,"AbstractList/IteratorImpl",414),N(95,414,yb,la),p.Qb=function(){hd(this)},p.Rb=function(t){nT(this,t)},p.Sb=function(){return this.b>0},p.Tb=function(){return this.b},p.Ub=function(){return Ln(this.b>0),this.a.Xb(this.c=--this.b)},p.Vb=function(){return this.b-1},p.Wb=function(t){Qk(this.c!=-1),this.a.fd(this.c,t)},R(Gn,"AbstractList/ListIteratorImpl",95),N(256,55,IL,Xv),p._c=function(t,i){fT(t,this.b),this.c._c(this.a+t,i),++this.b},p.Xb=function(t){return Gt(t,this.b),this.c.Xb(this.a+t)},p.ed=function(t){var i;return Gt(t,this.b),i=this.c.ed(this.a+t),--this.b,i},p.fd=function(t,i){return Gt(t,this.b),this.c.fd(this.a+t,i)},p.gc=function(){return this.b},p.a=0,p.b=0,R(Gn,"AbstractList/SubList",256),N(257,Ig,Pl,Ic),p.$b=function(){this.a.$b()},p.Gc=function(t){return this.a._b(t)},p.Jc=function(){var t;return t=this.a.vc().Jc(),new Hi(t)},p.Kc=function(t){return this.a._b(t)?(this.a.Ac(t),!0):!1},p.gc=function(){return this.a.gc()},R(Gn,"AbstractMap/1",257),N(526,1,aa,Hi),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.a.Ob()},p.Pb=function(){var t;return t=f(this.a.Pb(),43),t.jd()},p.Qb=function(){this.a.Qb()},R(Gn,"AbstractMap/1/1",526),N(230,31,jT,kp),p.$b=function(){this.a.$b()},p.Gc=function(t){return this.a.uc(t)},p.Jc=function(){var t;return t=this.a.vc().Jc(),new zy(t)},p.gc=function(){return this.a.gc()},R(Gn,"AbstractMap/2",230),N(300,1,aa,zy),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.a.Ob()},p.Pb=function(){var t;return t=f(this.a.Pb(),43),t.kd()},p.Qb=function(){this.a.Qb()},R(Gn,"AbstractMap/2/1",300),N(478,1,{478:1,43:1}),p.Fb=function(t){var i;return me(t,43)?(i=f(t,43),Pu(this.d,i.jd())&&Pu(this.e,i.kd())):!1},p.jd=function(){return this.d},p.kd=function(){return this.e},p.Hb=function(){return C3(this.d)^C3(this.e)},p.ld=function(t){return O_e(this,t)},p.Ib=function(){return this.d+"="+this.e},R(Gn,"AbstractMap/AbstractEntry",478),N(388,478,{478:1,388:1,43:1},AY),R(Gn,"AbstractMap/SimpleEntry",388),N(2028,1,Dfe),p.Fb=function(t){var i;return me(t,43)?(i=f(t,43),Pu(this.jd(),i.jd())&&Pu(this.kd(),i.kd())):!1},p.Hb=function(){return C3(this.jd())^C3(this.kd())},p.Ib=function(){return this.jd()+"="+this.kd()},R(Gn,Jkt,2028),N(2036,2011,OOe),p.Vc=function(t){return Oae(this.Ce(t))},p.tc=function(t){return Apt(this,t)},p._b=function(t){return N_e(this,t)},p.vc=function(){return new bz(this)},p.Rc=function(){return Qlt(this.Ee())},p.Wc=function(t){return Oae(this.Fe(t))},p.xc=function(t){var i;return i=t,ac(this.De(i))},p.Yc=function(t){return Oae(this.Ge(t))},p.ec=function(){return new nae(this)},p.Tc=function(){return Qlt(this.He())},p.Zc=function(t){return Oae(this.Ie(t))},R(Gn,"AbstractNavigableMap",2036),N(614,Ig,Pl,bz),p.Gc=function(t){return me(t,43)&&Apt(this.b,f(t,43))},p.Jc=function(){return this.b.Be()},p.Kc=function(t){var i;return me(t,43)?(i=f(t,43),this.b.Je(i)):!1},p.gc=function(){return this.b.gc()},R(Gn,"AbstractNavigableMap/EntrySet",614),N(1110,Ig,NOe,nae),p.Lc=function(){return new OY(this)},p.$b=function(){this.a.$b()},p.Gc=function(t){return N_e(this.a,t)},p.Jc=function(){var t;return t=this.a.vc().b.Be(),new rae(t)},p.Kc=function(t){return N_e(this.a,t)?(this.a.Ac(t),!0):!1},p.gc=function(){return this.a.gc()},R(Gn,"AbstractNavigableMap/NavigableKeySet",1110),N(1111,1,aa,rae),p.Nb=function(t){va(this,t)},p.Ob=function(){return Fae(this.a.a)},p.Pb=function(){var t;return t=Ect(this.a),t.jd()},p.Qb=function(){Iut(this.a)},R(Gn,"AbstractNavigableMap/NavigableKeySet/1",1111),N(2049,31,jT),p.Ec=function(t){return KN(dL(this,t),NL),!0},p.Fc=function(t){return an(t),uH(t!=this,"Can't add a queue to itself"),Ba(this,t)},p.$b=function(){for(;Tue(this)!=null;);},R(Gn,"AbstractQueue",2049),N(311,31,{4:1,20:1,31:1,16:1},O3,Tht),p.Ec=function(t){return NTe(this,t),!0},p.$b=function(){jTe(this)},p.Gc=function(t){return u1t(new ej(this),t)},p.dc=function(){return u7(this)},p.Jc=function(){return new ej(this)},p.Kc=function(t){return rEn(new ej(this),t)},p.gc=function(){return this.c-this.b&this.a.length-1},p.Lc=function(){return new Bt(this,272)},p.Oc=function(t){var i;return i=this.c-this.b&this.a.length-1,t.length<i&&(t=pY(new Array(i),t)),Cbt(this,t,i),t.length>i&&io(t,i,null),t},p.b=0,p.c=0,R(Gn,"ArrayDeque",311),N(445,1,aa,ej),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.a!=this.b},p.Pb=function(){return sZ(this)},p.Qb=function(){abt(this)},p.a=0,p.b=0,p.c=-1,R(Gn,"ArrayDeque/IteratorImpl",445),N(13,55,hTt,et,Du,Kl),p._c=function(t,i){Ky(this,t,i)},p.Ec=function(t){return Qe(this,t)},p.ad=function(t,i){return TCe(this,t,i)},p.Fc=function(t){return $o(this,t)},p.$b=function(){Bk(this.c,0)},p.Gc=function(t){return Oc(this,t,0)!=-1},p.Ic=function(t){Mu(this,t)},p.Xb=function(t){return ct(this,t)},p.bd=function(t){return Oc(this,t,0)},p.dc=function(){return this.c.length==0},p.Jc=function(){return new ne(this)},p.ed=function(t){return _0(this,t)},p.Kc=function(t){return hl(this,t)},p.ae=function(t,i){Zft(this,t,i)},p.fd=function(t,i){return af(this,t,i)},p.gc=function(){return this.c.length},p.gd=function(t){na(this,t)},p.Nc=function(){return SX(this.c)},p.Oc=function(t){return kg(this,t)};var q7n=R(Gn,"ArrayList",13);N(7,1,aa,ne),p.Nb=function(t){va(this,t)},p.Ob=function(){return sc(this)},p.Pb=function(){return ee(this)},p.Qb=function(){Y7(this)},p.a=0,p.b=-1,R(Gn,"ArrayList/1",7),N(2058,h.Function,{},Re),p.Ke=function(t,i){return di(t,i)},N(129,55,pTt,bu),p.Gc=function(t){return cbt(this,t)!=-1},p.Ic=function(t){var i,a,u,l;for(an(t),a=this.a,u=0,l=a.length;u<l;++u)i=a[u],t.Ad(i)},p.Xb=function(t){return glt(this,t)},p.fd=function(t,i){var a;return a=(Gt(t,this.a.length),this.a[t]),io(this.a,t,i),a},p.gc=function(){return this.a.length},p.gd=function(t){ace(this.a,this.a.length,t)},p.Nc=function(){return _vt(this,Pe(Ro,Jt,1,this.a.length,5,1))},p.Oc=function(t){return _vt(this,t)},R(Gn,"Arrays/ArrayList",129);var Ga,Cb,Ree;N(931,55,pTt,qe),p.Gc=function(t){return!1},p.Xb=function(t){return PSe(t)},p.Jc=function(){return Ht(),uP(),tD},p.cd=function(){return Ht(),uP(),tD},p.gc=function(){return 0},R(Gn,"Collections/EmptyList",931),N(932,1,yb,it),p.Nb=function(t){va(this,t)},p.Rb=function(t){throw ie(new cr)},p.Ob=function(){return!1},p.Sb=function(){return!1},p.Pb=function(){throw ie(new ic)},p.Tb=function(){return 0},p.Ub=function(){throw ie(new ic)},p.Vb=function(){return-1},p.Qb=function(){throw ie(new Tl)},p.Wb=function(t){throw ie(new Tl)};var tD;R(Gn,"Collections/EmptyListIterator",932),N(934,2011,AL,Xe),p._b=function(t){return!1},p.uc=function(t){return!1},p.vc=function(){return Ht(),Ree},p.xc=function(t){return null},p.ec=function(){return Ht(),Ree},p.gc=function(){return 0},p.Bc=function(){return Ht(),Ga},R(Gn,"Collections/EmptyMap",934),N(933,Ig,CR,_e),p.Gc=function(t){return!1},p.Jc=function(){return Ht(),uP(),tD},p.gc=function(){return 0},R(Gn,"Collections/EmptySet",933),N(592,55,{3:1,20:1,31:1,55:1,16:1,15:1},mz),p.Gc=function(t){return Pu(this.a,t)},p.Xb=function(t){return Gt(t,1),this.a},p.gc=function(){return 1},R(Gn,"Collections/SingletonList",592),N(376,1,eTt,h3),p.Ic=function(t){Oa(this,t)},p.Lc=function(){return new Bt(this,0)},p.Mc=function(){return new Ft(null,this.Lc())},p.Ec=function(t){return kit()},p.Fc=function(t){return Tit()},p.$b=function(){xit()},p.Gc=function(t){return Iz(this,t)},p.Hc=function(t){return hot(this,t)},p.dc=function(){return this.b.dc()},p.Jc=function(){return new p3(this.b.Jc())},p.Kc=function(t){return Cit()},p.gc=function(){return this.b.gc()},p.Nc=function(){return this.b.Nc()},p.Oc=function(t){return pot(this,t)},p.Ib=function(){return Qs(this.b)},R(Gn,"Collections/UnmodifiableCollection",376),N(375,1,aa,p3),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.b.Ob()},p.Pb=function(){return this.b.Pb()},p.Qb=function(){Ait()},R(Gn,"Collections/UnmodifiableCollectionIterator",375),N(525,376,gTt,cX),p.Lc=function(){return new Bt(this,16)},p._c=function(t,i){throw ie(new cr)},p.ad=function(t,i){throw ie(new cr)},p.Fb=function(t){return ei(this.a,t)},p.Xb=function(t){return this.a.Xb(t)},p.Hb=function(){return Ti(this.a)},p.bd=function(t){return this.a.bd(t)},p.dc=function(){return this.a.dc()},p.cd=function(){return new o_e(this.a.dd(0))},p.dd=function(t){return new o_e(this.a.dd(t))},p.ed=function(t){throw ie(new cr)},p.fd=function(t,i){throw ie(new cr)},p.gd=function(t){throw ie(new cr)},p.hd=function(t,i){return new cX(this.a.hd(t,i))},R(Gn,"Collections/UnmodifiableList",525),N(689,375,yb,o_e),p.Qb=function(){Ait()},p.Rb=function(t){throw ie(new cr)},p.Sb=function(){return this.a.Sb()},p.Tb=function(){return this.a.Tb()},p.Ub=function(){return this.a.Ub()},p.Vb=function(){return this.a.Vb()},p.Wb=function(t){throw ie(new cr)},R(Gn,"Collections/UnmodifiableListIterator",689),N(593,1,YS,vz),p.wc=function(t){HH(this,t)},p.$b=function(){throw ie(new cr)},p._b=function(t){return this.c._b(t)},p.uc=function(t){return got(this,t)},p.vc=function(){return TP(this)},p.Fb=function(t){return mot(this,t)},p.xc=function(t){return this.c.xc(t)},p.Hb=function(){return Ti(this.c)},p.dc=function(){return this.c.dc()},p.ec=function(){return Mdt(this)},p.yc=function(t,i){throw ie(new cr)},p.Ac=function(t){throw ie(new cr)},p.gc=function(){return this.c.gc()},p.Ib=function(){return Qs(this.c)},p.Bc=function(){return jdt(this)},R(Gn,"Collections/UnmodifiableMap",593),N(387,376,dfe,tP),p.Lc=function(){return new Bt(this,1)},p.Fb=function(t){return ei(this.b,t)},p.Hb=function(){return Ti(this.b)},R(Gn,"Collections/UnmodifiableSet",387),N(935,387,dfe,iit),p.Gc=function(t){return bot(this,t)},p.Hc=function(t){return this.b.Hc(t)},p.Jc=function(){var t;return t=this.b.Jc(),new iae(t)},p.Nc=function(){var t;return t=this.b.Nc(),npt(t,t.length),t},p.Oc=function(t){return bft(this,t)},R(Gn,"Collections/UnmodifiableMap/UnmodifiableEntrySet",935),N(936,1,aa,iae),p.Nb=function(t){va(this,t)},p.Pb=function(){return new YK(f(this.a.Pb(),43))},p.Ob=function(){return this.a.Ob()},p.Qb=function(){throw ie(new cr)},R(Gn,"Collections/UnmodifiableMap/UnmodifiableEntrySet/1",936),N(687,1,Dfe,YK),p.Fb=function(t){return this.a.Fb(t)},p.jd=function(){return this.a.jd()},p.kd=function(){return this.a.kd()},p.Hb=function(){return this.a.Hb()},p.ld=function(t){throw ie(new cr)},p.Ib=function(){return Qs(this.a)},R(Gn,"Collections/UnmodifiableMap/UnmodifiableEntrySet/UnmodifiableEntry",687),N(594,525,{20:1,16:1,15:1,59:1},Iae),R(Gn,"Collections/UnmodifiableRandomAccessList",594),N(688,387,tTt,a_e),p.Lc=function(){return new OY(this)},p.Fb=function(t){return ei(this.a,t)},p.Hb=function(){return Ti(this.a)},R(Gn,"Collections/UnmodifiableSortedSet",688),N(837,1,Mfe,de),p.Le=function(t,i){var a;return a=opt(f(t,11),f(i,11)),a!=0?a:XSt(f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Gn,"Comparator/lambda$0$Type",837);var M6e,Kpe,j6e;N(751,1,Mfe,We),p.Le=function(t,i){return Pvn(f(t,34),f(i,34))},p.Fb=function(t){return this===t},p.Me=function(){return TS(),j6e},R(Gn,"Comparators/NaturalOrderComparator",751),N(1186,1,Mfe,ht),p.Le=function(t,i){return Lvn(f(t,34),f(i,34))},p.Fb=function(t){return this===t},p.Me=function(){return TS(),Kpe},R(Gn,"Comparators/ReverseNaturalOrderComparator",1186),N(56,1,Mfe,wr),p.Fb=function(t){return this===t},p.Le=function(t,i){return this.a.Le(i,t)},p.Me=function(){return this.a},R(Gn,"Comparators/ReversedComparator",56),N(175,63,C1,If),R(Gn,"ConcurrentModificationException",175);var WAt,KAt;N(1894,1,qG,Ie),p.Ne=function(t){Z0t(this,t)},p.Ib=function(){return"DoubleSummaryStatistics[count = "+V7(this.a)+", avg = "+(jae(this.a,0)?JTe(this)/oE(this.a):0)+", min = "+this.c+", max = "+this.b+", sum = "+JTe(this)+"]"},p.a=0,p.b=qo,p.c=Gi,p.d=0,p.e=0,p.f=0,R(Gn,"DoubleSummaryStatistics",1894),N(1814,63,C1,Jnt),R(Gn,"EmptyStackException",1814),N(447,2011,YS,Aj),p.yc=function(t,i){return I_e(this,t,i)},p.$b=function(){Edt(this)},p._b=function(t){return Lot(this,t)},p.uc=function(t){var i,a;for(a=new LN(this.a);a.a<a.c.a.length;)if(i=aj(a),Pu(t,this.b[i.g]))return!0;return!1},p.vc=function(){return new oae(this)},p.xc=function(t){return Es(this,t)},p.Ac=function(t){return yTe(this,t)},p.gc=function(){return this.a.c},R(Gn,"EnumMap",447),N(1281,Ig,Pl,oae),p.$b=function(){Edt(this.a)},p.Gc=function(t){return Dht(this,t)},p.Jc=function(){return new mlt(this.a)},p.Kc=function(t){var i;return Dht(this,t)?(i=f(t,43).jd(),yTe(this.a,i),!0):!1},p.gc=function(){return this.a.a.c},R(Gn,"EnumMap/EntrySet",1281),N(1282,1,aa,mlt),p.Nb=function(t){va(this,t)},p.Pb=function(){return this.b=aj(this.a),new Mot(this.c,this.b)},p.Ob=function(){return zSe(this.a)},p.Qb=function(){Qk(!!this.b),yTe(this.c,this.b),this.b=null},R(Gn,"EnumMap/EntrySetIterator",1282),N(1283,2028,Dfe,Mot),p.jd=function(){return this.a},p.kd=function(){return this.b.b[this.a.g]},p.ld=function(t){return mke(this.b.b,this.a.g,t)},R(Gn,"EnumMap/MapEntry",1283),N(181,Ig,{20:1,31:1,16:1,181:1,22:1});var YAt=R(Gn,"EnumSet",181);N(161,181,{20:1,31:1,16:1,181:1,161:1,22:1},Rf),p.Ec=function(t){return Yh(this,f(t,23))},p.Gc=function(t){return oce(this,t)},p.Jc=function(){return new LN(this)},p.Kc=function(t){return Alt(this,t)},p.gc=function(){return this.c},p.c=0,R(Gn,"EnumSet/EnumSetImpl",161),N(353,1,aa,LN),p.Nb=function(t){va(this,t)},p.Pb=function(){return aj(this)},p.Ob=function(){return zSe(this)},p.Qb=function(){Qk(this.b!=-1),io(this.c.b,this.b,null),--this.c.c,this.b=-1},p.a=-1,p.b=-1,R(Gn,"EnumSet/EnumSetImpl/IteratorImpl",353),N(45,482,sA,Wn,PN,rst),p.ze=function(t,i){return Ae(t)===Ae(i)||t!=null&&ei(t,i)},p.Ae=function(t){var i;return t==null?0:(i=Ti(t),i|0)},R(Gn,"HashMap",45),N(50,Ig,$Oe,Mo,bY,H7),p.Ec=function(t){return Go(this,t)},p.$b=function(){this.a.$b()},p.Gc=function(t){return xp(this,t)},p.dc=function(){return this.a.gc()==0},p.Jc=function(){return this.a.ec().Jc()},p.Kc=function(t){return KSe(this,t)},p.gc=function(){return this.a.gc()};var V7n=R(Gn,"HashSet",50);N(1843,1,$G,$t),p.Bd=function(t){e1t(this,t)},p.Ib=function(){return"IntSummaryStatistics[count = "+V7(this.a)+", avg = "+(jae(this.a,0)?oE(this.d)/oE(this.a):0)+", min = "+this.c+", max = "+this.b+", sum = "+V7(this.d)+"]"},p.a=0,p.b=Ra,p.c=Qr,p.d=0,R(Gn,"IntSummaryStatistics",1843),N(1037,1,Eb,hst),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return new axe(this)},p.c=0,R(Gn,"InternalHashCodeMap",1037),N(710,1,aa,axe),p.Nb=function(t){va(this,t)},p.Pb=function(){return this.d=this.a[this.c++],this.d},p.Ob=function(){var t;return this.c<this.a.length?!0:(t=this.b.next(),t.done?!1:(this.a=t.value[1],this.c=0,!0))},p.Qb=function(){ZAe(this.e,this.d.jd()),this.c!=0&&--this.c},p.c=0,p.d=null,R(Gn,"InternalHashCodeMap/1",710);var XAt;N(1035,1,Eb,pst),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return new $Te(this)},p.c=0,p.d=0,R(Gn,"InternalStringMap",1035),N(709,1,aa,$Te),p.Nb=function(t){va(this,t)},p.Pb=function(){return this.c=this.a,this.a=this.b.next(),new Nut(this.d,this.c,this.d.d)},p.Ob=function(){return!this.a.done},p.Qb=function(){t1t(this.d,this.c.value[0])},R(Gn,"InternalStringMap/1",709),N(1036,2028,Dfe,Nut),p.jd=function(){return this.b.value[0]},p.kd=function(){return this.a.d!=this.c?w7(this.a,this.b.value[0]):this.b.value[1]},p.ld=function(t){return MS(this.a,this.b.value[0],t)},p.c=0,R(Gn,"InternalStringMap/2",1036),N(221,45,sA,Zv,oxe),p.$b=function(){Qct(this)},p._b=function(t){return Dot(this,t)},p.uc=function(t){var i;for(i=this.d.a;i!=this.d;){if(Pu(i.e,t))return!0;i=i.a}return!1},p.vc=function(){return new XK(this)},p.xc=function(t){return Eg(this,t)},p.yc=function(t,i){return pE(this,t,i)},p.Ac=function(t){return emt(this,t)},p.gc=function(){return f7(this.e)},p.c=!1,R(Gn,"LinkedHashMap",221),N(391,388,{478:1,388:1,391:1,43:1},wct,nke),R(Gn,"LinkedHashMap/ChainEntry",391),N(699,Ig,Pl,XK),p.$b=function(){Qct(this.a)},p.Gc=function(t){return Mht(this,t)},p.Jc=function(){return new qke(this)},p.Kc=function(t){var i;return Mht(this,t)?(i=f(t,43).jd(),emt(this.a,i),!0):!1},p.gc=function(){return f7(this.a.e)},R(Gn,"LinkedHashMap/EntrySet",699),N(700,1,aa,qke),p.Nb=function(t){va(this,t)},p.Pb=function(){return amt(this)},p.Ob=function(){return this.c!=this.d.a.d},p.Qb=function(){Qk(!!this.a),Vse(this.d.a.e.g,this.b),hke(this.a),aR(this.d.a.e,this.a.d),this.b=this.d.a.e.g,this.a=null},p.b=0,R(Gn,"LinkedHashMap/EntrySet/EntryIterator",700),N(180,50,$Oe,ub,Nse,wke);var W7n=R(Gn,"LinkedHashSet",180);N(67,2008,{3:1,4:1,20:1,31:1,55:1,16:1,67:1,15:1},ki,Hse),p.Ec=function(t){return Mr(this,t)},p.$b=function(){nf(this)},p.dd=function(t){return tr(this,t)},p.gc=function(){return this.b},p.b=0;var K7n=R(Gn,"LinkedList",67);N(958,1,yb,Rut),p.Nb=function(t){va(this,t)},p.Rb=function(t){UX(this,t)},p.Ob=function(){return kz(this)},p.Sb=function(){return this.b.b!=this.d.a},p.Pb=function(){return Zn(this)},p.Tb=function(){return this.a},p.Ub=function(){return _ht(this)},p.Vb=function(){return this.a-1},p.Qb=function(){Cue(this)},p.Wb=function(t){Qk(!!this.c),this.c.c=t},p.a=0,p.c=null,R(Gn,"LinkedList/ListIteratorImpl",958),N(602,1,{},An),R(Gn,"LinkedList/Node",602),N(2003,1,{});var $6e,JAt;R(Gn,"Locale",2003),N(849,2003,{},Yt),p.Ib=function(){return""},R(Gn,"Locale/1",849),N(850,2003,{},Or),p.Ib=function(){return"unknown"},R(Gn,"Locale/4",850),N(111,63,{3:1,103:1,63:1,80:1,111:1},ic,Udt),R(Gn,"NoSuchElementException",111),N(461,1,{461:1},yae),p.Fb=function(t){var i;return t===this?!0:me(t,461)?(i=f(t,461),Pu(this.a,i.a)):!1},p.Hb=function(){return C3(this.a)},p.Ib=function(){return this.a!=null?Ykt+$7(this.a)+")":"Optional.empty()"};var F6e;R(Gn,"Optional",461),N(403,1,{403:1},ist,Dse),p.Fb=function(t){var i;return t===this?!0:me(t,403)?(i=f(t,403),this.a==i.a&&di(this.b,i.b)==0):!1},p.Hb=function(){return this.a?Aa(this.b):0},p.Ib=function(){return this.a?"OptionalDouble.of("+(""+this.b)+")":"OptionalDouble.empty()"},p.a=!1,p.b=0;var Ype;R(Gn,"OptionalDouble",403),N(508,1,{508:1},ost,xct),p.Fb=function(t){var i;return t===this?!0:me(t,508)?(i=f(t,508),this.a==i.a&&tu(this.b,i.b)==0):!1},p.Hb=function(){return this.a?this.b:0},p.Ib=function(){return this.a?"OptionalInt.of("+(""+this.b)+")":"OptionalInt.empty()"},p.a=!1,p.b=0;var ZAt;R(Gn,"OptionalInt",508),N(494,2049,jT,jX),p.Fc=function(t){return L3e(this,t)},p.$b=function(){Bk(this.b.c,0)},p.Gc=function(t){return(t==null?-1:Oc(this.b,t,0))!=-1},p.Jc=function(){return new aae(this)},p.Kc=function(t){return _bt(this,t)},p.gc=function(){return this.b.c.length},p.Lc=function(){return new Bt(this,256)},p.Nc=function(){return SX(this.b.c)},p.Oc=function(t){return kg(this.b,t)},R(Gn,"PriorityQueue",494),N(1249,1,aa,aae),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.a<this.c.b.c.length},p.Pb=function(){return Ln(this.a<this.c.b.c.length),this.b=this.a++,ct(this.c.b,this.b)},p.Qb=function(){Qk(this.b!=-1),Pue(this.c,this.a=this.b),this.b=-1},p.a=0,p.b=-1,R(Gn,"PriorityQueue/1",1249),N(233,1,{233:1},vle,hJ),p.a=0,p.b=0;var B6e,U6e,Y7n=0;R(Gn,"Random",233),N(26,1,pf,Bt,sT,Bdt),p.yd=function(t){return(this.a&t)!=0},p.wd=function(){return this.a},p.xd=function(){return Ike(this),this.c},p.Nb=function(t){Ike(this),this.d.Nb(t)},p.zd=function(t){return Rmt(this,t)},p.a=0,p.c=0,R(Gn,"Spliterators/IteratorSpliterator",26),N(479,26,pf,OY),R(Gn,"SortedSet/1",479),N(595,1,qG,ZK),p.Ne=function(t){this.a.Ad(t)},R(Gn,"Spliterator/OfDouble/0methodref$accept$Type",595),N(596,1,qG,QK),p.Ne=function(t){this.a.Ad(t)},R(Gn,"Spliterator/OfDouble/1methodref$accept$Type",596),N(597,1,$G,eY),p.Bd=function(t){this.a.Ad(rt(t))},R(Gn,"Spliterator/OfInt/2methodref$accept$Type",597),N(598,1,$G,tY),p.Bd=function(t){this.a.Ad(rt(t))},R(Gn,"Spliterator/OfInt/3methodref$accept$Type",598),N(610,1,pf),p.Nb=function(t){q2e(this,t)},p.yd=function(t){return(this.d&t)!=0},p.wd=function(){return this.d},p.xd=function(){return this.e},p.d=0,p.e=0,R(Gn,"Spliterators/BaseSpliterator",610),N(719,610,pf),p.Oe=function(t){b3(this,t)},p.Nb=function(t){me(t,189)?b3(this,f(t,189)):b3(this,new QK(t))},p.zd=function(t){return me(t,189)?this.Pe(f(t,189)):this.Pe(new ZK(t))},R(Gn,"Spliterators/AbstractDoubleSpliterator",719),N(718,610,pf),p.Oe=function(t){b3(this,t)},p.Nb=function(t){me(t,202)?b3(this,f(t,202)):b3(this,new tY(t))},p.zd=function(t){return me(t,202)?this.Pe(f(t,202)):this.Pe(new eY(t))},R(Gn,"Spliterators/AbstractIntSpliterator",718),N(484,610,pf),R(Gn,"Spliterators/AbstractSpliterator",484),N(690,1,pf),p.Nb=function(t){q2e(this,t)},p.yd=function(t){return(this.b&t)!=0},p.wd=function(){return this.b},p.xd=function(){return this.d-this.c},p.b=0,p.c=0,p.d=0,R(Gn,"Spliterators/BaseArraySpliterator",690),N(938,690,pf,llt),p.Qe=function(t,i){Ppn(this,f(t,40),i)},p.Nb=function(t){kce(this,t)},p.zd=function(t){return yJ(this,t)},R(Gn,"Spliterators/ArraySpliterator",938),N(691,690,pf,but),p.Qe=function(t,i){Rpn(this,f(t,189),i)},p.Oe=function(t){kce(this,t)},p.Nb=function(t){me(t,189)?kce(this,f(t,189)):kce(this,new QK(t))},p.Pe=function(t){return yJ(this,t)},p.zd=function(t){return me(t,189)?yJ(this,f(t,189)):yJ(this,new ZK(t))},R(Gn,"Spliterators/DoubleArraySpliterator",691),N(2012,1,pf),p.Nb=function(t){q2e(this,t)},p.yd=function(t){return(16448&t)!=0},p.wd=function(){return 16448},p.xd=function(){return 0};var QAt;R(Gn,"Spliterators/EmptySpliterator",2012),N(937,2012,pf,Nt),p.Oe=function(t){H9(t)},p.Nb=function(t){me(t,202)?H9(f(t,202)):H9(new tY(t))},p.Pe=function(t){return iSe(t)},p.zd=function(t){return me(t,202)?iSe(f(t,202)):iSe(new eY(t))},R(Gn,"Spliterators/EmptySpliterator/OfInt",937),N(571,55,mTt,uY),p._c=function(t,i){yP(t,this.a.c.length+1),Ky(this.a,t,i)},p.Ec=function(t){return Qe(this.a,t)},p.ad=function(t,i){return yP(t,this.a.c.length+1),TCe(this.a,t,i)},p.Fc=function(t){return $o(this.a,t)},p.$b=function(){Bk(this.a.c,0)},p.Gc=function(t){return Oc(this.a,t,0)!=-1},p.Hc=function(t){return KH(this.a,t)},p.Ic=function(t){Mu(this.a,t)},p.Xb=function(t){return yP(t,this.a.c.length),ct(this.a,t)},p.bd=function(t){return Oc(this.a,t,0)},p.dc=function(){return this.a.c.length==0},p.Jc=function(){return new ne(this.a)},p.ed=function(t){return yP(t,this.a.c.length),_0(this.a,t)},p.ae=function(t,i){Zft(this.a,t,i)},p.fd=function(t,i){return yP(t,this.a.c.length),af(this.a,t,i)},p.gc=function(){return this.a.c.length},p.gd=function(t){na(this.a,t)},p.hd=function(t,i){return new Xv(this.a,t,i)},p.Nc=function(){return SX(this.a.c)},p.Oc=function(t){return kg(this.a,t)},p.Ib=function(){return Cg(this.a)},R(Gn,"Vector",571),N(803,571,mTt,v2e),R(Gn,"Stack",803),N(212,1,{212:1},uE),p.Ib=function(){return kht(this)},R(Gn,"StringJoiner",212),N(538,2036,{3:1,91:1,137:1,134:1},yot,dce),p.$b=function(){Yrt(this)},p.Be=function(){return new xht(this)},p.vc=function(){return new pct(this)},p.Ce=function(t){return iL(this,t,!0)},p.De=function(t){return $1t(this,t)},p.Ee=function(){return hxe(this)},p.Fe=function(t){return sG(this,t,!0)},p.Ge=function(t){return iL(this,t,!1)},p.He=function(){return zgt(this)},p.Ie=function(t){return sG(this,t,!1)},p.Xc=function(t,i){return Qft(this,t,i)},p.yc=function(t,i){return C1t(this,t,i)},p.Ac=function(t){return Sft(this,t)},p.Je=function(t){return nxe(this,t)},p.gc=function(){return this.c},p.$c=function(t,i){return eht(this,t,i)},p.c=0,R(Gn,"TreeMap",538),N(539,1,aa,xht,$ue),p.Nb=function(t){va(this,t)},p.Pb=function(){return Ect(this)},p.Ob=function(){return Fae(this.a)},p.Qb=function(){Iut(this)},R(Gn,"TreeMap/EntryIterator",539),N(1106,614,Pl,pct),p.$b=function(){Yrt(this.a)},R(Gn,"TreeMap/EntrySet",1106),N(435,388,{478:1,388:1,43:1,435:1},pue),p.b=!1;var X7n=R(Gn,"TreeMap/Node",435);N(615,1,{},on),p.Ib=function(){return"State: mv="+this.c+" value="+this.d+" done="+this.a+" found="+this.b},p.a=!1,p.b=!1,p.c=!1,R(Gn,"TreeMap/State",615),N(616,2036,OOe,hde),p.Be=function(){return new $ue(this.c,this.f,this.b,this.a,this.e,this.d)},p.vc=function(){return new bz(this)},p.Ce=function(t){return EX(this,iL(this.c,t,!0))},p.De=function(t){return EX(this,$1t(this.c,t))},p.Ee=function(){var t;return this.f.Re()?this.a?t=iL(this.c,this.b,!0):t=iL(this.c,this.b,!1):t=hxe(this.c),t&&GX(this,t.d)?t:null},p.Fe=function(t){return EX(this,sG(this.c,t,!0))},p.Ge=function(t){return EX(this,iL(this.c,t,!1))},p.He=function(){var t;return this.f.Se()?this.d?t=sG(this.c,this.e,!0):t=sG(this.c,this.e,!1):t=zgt(this.c),t&&GX(this,t.d)?t:null},p.Ie=function(t){return EX(this,sG(this.c,t,!1))},p.Xc=function(t,i){if(this.f.Se()&&this.c.a.Le(t,this.e)>0)throw ie(new dn(UOe+t+" greater than "+this.e));return this.f.Re()?Eft(this.c,this.b,this.a,t,i):Qft(this.c,t,i)},p.yc=function(t,i){if(!Jle(this.c,this.f,t,this.b,this.a,this.e,this.d))throw ie(new dn(t+" outside the range "+this.b+" to "+this.e));return C1t(this.c,t,i)},p.Ac=function(t){var i;return i=t,Jle(this.c,this.f,i,this.b,this.a,this.e,this.d)?Sft(this.c,i):null},p.Je=function(t){return GX(this,t.jd())&&nxe(this.c,t)},p.gc=function(){var t,i,a;if(this.f.Re()?this.a?i=iL(this.c,this.b,!0):i=iL(this.c,this.b,!1):i=hxe(this.c),!(i&&GX(this,i.d)&&i))return 0;for(t=0,a=new $ue(this.c,this.f,this.b,this.a,this.e,this.d);Fae(a.a);a.b=f(iTe(a.a),43))++t;return t},p.$c=function(t,i){if(this.f.Re()&&this.c.a.Le(t,this.b)<0)throw ie(new dn(UOe+t+bTt+this.b));return this.f.Se()?Eft(this.c,t,i,this.e,this.d):eht(this.c,t,i)},p.a=!1,p.d=!1,R(Gn,"TreeMap/SubMap",616),N(305,23,Ffe,NY),p.Re=function(){return!1},p.Se=function(){return!1};var Xpe,Jpe,Zpe,Qpe,Pee=er(Gn,"TreeMap/SubMapType",305,ar,R2n,s1n);N(1107,305,Ffe,kst),p.Se=function(){return!0},er(Gn,"TreeMap/SubMapType/1",1107,Pee,null,null),N(1108,305,Ffe,Lst),p.Re=function(){return!0},p.Se=function(){return!0},er(Gn,"TreeMap/SubMapType/2",1108,Pee,null,null),N(1109,305,Ffe,Tst),p.Re=function(){return!0},er(Gn,"TreeMap/SubMapType/3",1109,Pee,null,null);var eIt;N(156,Ig,{3:1,20:1,31:1,16:1,274:1,22:1,83:1,156:1},u2e,g0,g3),p.Lc=function(){return new OY(this)},p.Ec=function(t){return fH(this,t)},p.$b=function(){this.a.$b()},p.Gc=function(t){return this.a._b(t)},p.Jc=function(){return this.a.ec().Jc()},p.Kc=function(t){return Ise(this,t)},p.gc=function(){return this.a.gc()};var J7n=R(Gn,"TreeSet",156);N(1047,1,{},I),p.Te=function(t,i){return Sbn(this.a,t,i)},R(Bfe,"BinaryOperator/lambda$0$Type",1047),N(1048,1,{},M),p.Te=function(t,i){return _bn(this.a,t,i)},R(Bfe,"BinaryOperator/lambda$1$Type",1048),N(930,1,{},br),p.Kb=function(t){return t},R(Bfe,"Function/lambda$0$Type",930),N(386,1,_r,Y),p.Mb=function(t){return!this.a.Mb(t)},R(Bfe,"Predicate/lambda$2$Type",386),N(563,1,{563:1});var tIt=R(o$,"Handler",563);N(2053,1,MG),p.ve=function(){return"DUMMY"},p.Ib=function(){return this.ve()};var z6e;R(o$,"Level",2053),N(1648,2053,MG,wi),p.ve=function(){return"INFO"},R(o$,"Level/LevelInfo",1648),N(1789,1,{},nrt);var ege;R(o$,"LogManager",1789),N(1842,1,MG,Out),p.b=null,R(o$,"LogRecord",1842),N(509,1,{509:1},iue),p.e=!1;var nIt=!1,rIt=!1,Lg=!1,iIt=!1,oIt=!1;R(o$,"Logger",509),N(814,563,{563:1},Ca),R(o$,"SimpleConsoleLogHandler",814),N(131,23,{3:1,34:1,23:1,131:1},Bae);var H6e,Zl,G6e,Ql=er(ds,"Collector/Characteristics",131,ar,BEn,c1n),aIt;N(741,1,{},jke),R(ds,"CollectorImpl",741),N(1045,1,{},qi),p.Te=function(t,i){return Hxn(f(t,212),f(i,212))},R(ds,"Collectors/10methodref$merge$Type",1045),N(1046,1,{},yi),p.Kb=function(t){return kht(f(t,212))},R(ds,"Collectors/11methodref$toString$Type",1046),N(163,1,{},lo),p.Wd=function(t,i){f(t,16).Ec(i)},R(ds,"Collectors/20methodref$add$Type",163),N(165,1,{},Eo),p.Ve=function(){return new et},R(ds,"Collectors/21methodref$ctor$Type",165),N(1044,1,{},Fc),p.Wd=function(t,i){b1(f(t,212),f(i,470))},R(ds,"Collectors/9methodref$add$Type",1044),N(1043,1,{},Wut),p.Ve=function(){return new uE(this.a,this.b,this.c)},R(ds,"Collectors/lambda$15$Type",1043),N(164,1,{},Yo),p.Te=function(t,i){return rgn(f(t,16),f(i,16))},R(ds,"Collectors/lambda$45$Type",164),N(535,1,{}),p.Ye=function(){Q7(this)},p.d=!1,R(ds,"TerminatableStream",535),N(806,535,zOe,w_e),p.Ye=function(){Q7(this)},R(ds,"DoubleStreamImpl",806),N(1793,719,pf,Kut),p.Pe=function(t){return bAn(this,f(t,189))},p.a=null,R(ds,"DoubleStreamImpl/2",1793),N(1794,1,qG,fe),p.Ne=function(t){Wgn(this.a,t)},R(ds,"DoubleStreamImpl/2/lambda$0$Type",1794),N(1791,1,qG,Ce),p.Ne=function(t){Vgn(this.a,t)},R(ds,"DoubleStreamImpl/lambda$0$Type",1791),N(1792,1,qG,ze),p.Ne=function(t){Z0t(this.a,t)},R(ds,"DoubleStreamImpl/lambda$2$Type",1792),N(1336,718,pf,Rpt),p.Pe=function(t){return x2n(this,f(t,202))},p.a=0,p.b=0,p.c=0,R(ds,"IntStream/5",1336),N(786,535,zOe,y_e),p.Ye=function(){Q7(this)},p.Ze=function(){return Kv(this),this.a},R(ds,"IntStreamImpl",786),N(787,535,zOe,X2e),p.Ye=function(){Q7(this)},p.Ze=function(){return Kv(this),t_e(),QAt},R(ds,"IntStreamImpl/Empty",787),N(1628,1,$G,dt),p.Bd=function(t){e1t(this.a,t)},R(ds,"IntStreamImpl/lambda$4$Type",1628);var Z7n=Bi(ds,"Stream");N(28,535,{518:1,672:1,827:1},Ft),p.Ye=function(){Q7(this)};var zR;R(ds,"StreamImpl",28),N(1067,484,pf,_ut),p.zd=function(t){for(;E_n(this);){if(this.a.zd(t))return!0;Q7(this.b),this.b=null,this.a=null}return!1},R(ds,"StreamImpl/1",1067),N(1068,1,xn,It),p.Ad=function(t){uvn(this.a,f(t,827))},R(ds,"StreamImpl/1/lambda$0$Type",1068),N(1069,1,_r,qt),p.Mb=function(t){return Go(this.a,t)},R(ds,"StreamImpl/1methodref$add$Type",1069),N(1070,484,pf,Qdt),p.zd=function(t){var i;return this.a||(i=new et,this.b.a.Nb(new Hn(i)),Ht(),na(i,this.c),this.a=new Bt(i,16)),Rmt(this.a,t)},p.a=null,R(ds,"StreamImpl/5",1070),N(1071,1,xn,Hn),p.Ad=function(t){Qe(this.a,t)},R(ds,"StreamImpl/5/2methodref$add$Type",1071),N(720,484,pf,dxe),p.zd=function(t){for(this.b=!1;!this.b&&this.c.zd(new jot(this,t)););return this.b},p.b=!1,R(ds,"StreamImpl/FilterSpliterator",720),N(1061,1,xn,jot),p.Ad=function(t){iwn(this.a,this.b,t)},R(ds,"StreamImpl/FilterSpliterator/lambda$0$Type",1061),N(1056,719,pf,qpt),p.Pe=function(t){return Gbn(this,f(t,189))},R(ds,"StreamImpl/MapToDoubleSpliterator",1056),N(1060,1,xn,$ot),p.Ad=function(t){vgn(this.a,this.b,t)},R(ds,"StreamImpl/MapToDoubleSpliterator/lambda$0$Type",1060),N(1055,718,pf,Gpt),p.Pe=function(t){return qbn(this,f(t,202))},R(ds,"StreamImpl/MapToIntSpliterator",1055),N(1059,1,xn,Fot),p.Ad=function(t){wgn(this.a,this.b,t)},R(ds,"StreamImpl/MapToIntSpliterator/lambda$0$Type",1059),N(717,484,pf,ZTe),p.zd=function(t){return vut(this,t)},R(ds,"StreamImpl/MapToObjSpliterator",717),N(1058,1,xn,Bot),p.Ad=function(t){ygn(this.a,this.b,t)},R(ds,"StreamImpl/MapToObjSpliterator/lambda$0$Type",1058),N(1057,484,pf,fbt),p.zd=function(t){for(;jae(this.b,0);){if(!this.a.zd(new Ri))return!1;this.b=hh(this.b,1)}return this.a.zd(t)},p.b=0,R(ds,"StreamImpl/SkipSpliterator",1057),N(1062,1,xn,Ri),p.Ad=function(t){},R(ds,"StreamImpl/SkipSpliterator/lambda$0$Type",1062),N(611,1,xn,ro),p.Ad=function(t){q5(this,t)},R(ds,"StreamImpl/ValueConsumer",611),N(1063,1,xn,no),p.Ad=function(t){Wy()},R(ds,"StreamImpl/lambda$0$Type",1063),N(1064,1,xn,Ko),p.Ad=function(t){Wy()},R(ds,"StreamImpl/lambda$1$Type",1064),N(1065,1,{},qn),p.Te=function(t,i){return i1n(this.a,t,i)},R(ds,"StreamImpl/lambda$4$Type",1065),N(1066,1,xn,Uot),p.Ad=function(t){Lbn(this.b,this.a,t)},R(ds,"StreamImpl/lambda$5$Type",1066),N(1072,1,xn,sr),p.Ad=function(t){vTn(this.a,f(t,372))},R(ds,"TerminatableStream/lambda$0$Type",1072),N(2088,1,{}),N(1960,1,{},zl),R("javaemul.internal","ConsoleLogger",1960);var Q7n=0;N(2080,1,{}),N(1776,1,xn,ld),p.Ad=function(t){f(t,318)},R(RL,"BowyerWatsonTriangulation/lambda$0$Type",1776),N(1777,1,xn,Ji),p.Ad=function(t){Ba(this.a,f(t,318).e)},R(RL,"BowyerWatsonTriangulation/lambda$1$Type",1777),N(1778,1,xn,Yc),p.Ad=function(t){f(t,176)},R(RL,"BowyerWatsonTriangulation/lambda$2$Type",1778),N(1773,1,Cr,Uo),p.Le=function(t,i){return cSn(this.a,f(t,176),f(i,176))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(RL,"NaiveMinST/lambda$0$Type",1773),N(437,1,{},ri),R(RL,"NodeMicroLayout",437),N(176,1,{176:1},DN),p.Fb=function(t){var i;return me(t,176)?(i=f(t,176),Pu(this.a,i.a)&&Pu(this.b,i.b)||Pu(this.a,i.b)&&Pu(this.b,i.a)):!1},p.Hb=function(){return C3(this.a)+C3(this.b)};var ejn=R(RL,"TEdge",176);N(318,1,{318:1},nOe),p.Fb=function(t){var i;return me(t,318)?(i=f(t,318),CJ(this,i.a)&&CJ(this,i.b)&&CJ(this,i.c)):!1},p.Hb=function(){return C3(this.a)+C3(this.b)+C3(this.c)},R(RL,"TTriangle",318),N(224,1,{224:1},iX),R(RL,"Tree",224),N(1178,1,{},$ft),R(yTt,"Scanline",1178);var sIt=Bi(yTt,ETt);N(1704,1,{},Lmt),R(kb,"CGraph",1704),N(317,1,{317:1},Dft),p.b=0,p.c=0,p.d=0,p.g=0,p.i=0,p.k=qo,R(kb,"CGroup",317),N(809,1,{},d2e),R(kb,"CGroup/CGroupBuilder",809),N(60,1,{60:1},nut),p.Ib=function(){var t;return this.j?dr(this.j.Kb(this)):(d1(Lee),Lee.o+"@"+(t=bS(this)>>>0,t.toString(16)))},p.f=0,p.i=qo;var Lee=R(kb,"CNode",60);N(808,1,{},f2e),R(kb,"CNode/CNodeBuilder",808);var cIt;N(1530,1,{},Sn),p.df=function(t,i){return 0},p.ef=function(t,i){return 0},R(kb,_Tt,1530),N(1799,1,{},Pn),p.af=function(t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J;for(T=Gi,u=new ne(t.a.b);u.a<u.c.c.length;)i=f(ee(u),60),T=h.Math.min(T,i.a.j.d.c+i.b.a);for(z=new ki,m=new ne(t.a.a);m.a<m.c.c.length;)g=f(ee(m),317),g.k=T,g.g==0&&Wi(z,g,z.c.b,z.c);for(;z.b!=0;){for(g=f(z.b==0?null:(Ln(z.b!=0),lh(z,z.a.a)),317),l=g.j.d.c,F=g.a.a.ec().Jc();F.Ob();)A=f(F.Pb(),60),J=g.k+A.b.a,!Cxn(t,g,t.d)||A.d.c<J?A.i=J:A.i=A.d.c;for(l-=g.j.i,g.b+=l,t.d==(ko(),Gs)||t.d==jg?g.c+=l:g.c-=l,L=g.a.a.ec().Jc();L.Ob();)for(A=f(L.Pb(),60),S=A.c.Jc();S.Ob();)v=f(S.Pb(),60),l1(t.d)?W=t.g.df(A,v):W=t.g.ef(A,v),v.a.k=h.Math.max(v.a.k,A.i+A.d.b+W-v.b.a),gft(t,v,t.d)&&(v.a.k=h.Math.max(v.a.k,v.d.c-v.b.a)),--v.a.g,v.a.g==0&&Mr(z,v.a)}for(a=new ne(t.a.b);a.a<a.c.c.length;)i=f(ee(a),60),i.d.c=i.i},R(kb,"LongestPathCompaction",1799),N(1702,1,{},CEt),p.e=!1;var uIt,lIt,dIt,tge=R(kb,xTt,1702);N(1703,1,xn,ns),p.Ad=function(t){xTn(this.a,f(t,47))},R(kb,CTt,1703),N(1800,1,{},Kr),p.bf=function(t){var i,a,u,l,g,m,v;for(a=new ne(t.a.b);a.a<a.c.c.length;)i=f(ee(a),60),i.c.$b();for(l=new ne(t.a.b);l.a<l.c.c.length;)for(u=f(ee(l),60),m=new ne(t.a.b);m.a<m.c.c.length;)g=f(ee(m),60),u!=g&&(u.a&&u.a==g.a||(l1(t.d)?v=t.g.ef(u,g):v=t.g.df(u,g),(g.d.c>u.d.c||u.d.c==g.d.c&&u.d.b<g.d.b)&&n3n(g.d.d+g.d.a+v,u.d.d)&&f3e(g.d.d,u.d.d+u.d.a+v)&&u.c.Ec(g)))},R(kb,"QuadraticConstraintCalculation",1800),N(514,1,{514:1},uae),p.a=!1,p.b=!1,p.c=!1,p.d=!1,R(kb,ATt,514),N(797,1,{},Ske),p.bf=function(t){this.c=t,wG(this,new bs)},R(kb,ITt,797),N(1730,1,{678:1},cft),p._e=function(t){dRn(this,f(t,459))},R(kb,OTt,1730),N(1731,1,Cr,Ei),p.Le=function(t,i){return tyn(f(t,60),f(i,60))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(kb,NTt,1731),N(459,1,{459:1},aSe),p.a=!1,R(kb,RTt,459),N(1732,1,Cr,Sa),p.Le=function(t,i){return xIn(f(t,459),f(i,459))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(kb,PTt,1732),N(1733,1,km,bs),p.Lb=function(t){return f(t,60),!0},p.Fb=function(t){return this===t},p.Mb=function(t){return f(t,60),!0},R(kb,"ScanlineConstraintCalculator/lambda$1$Type",1733),N(216,1,{216:1},Cc),p.Ib=function(){return"NEdge[id="+this.b+" w="+this.g+" d="+this.a+"]"},p.a=1,p.b=0,p.c=0,p.f=!1,p.g=0;var fIt=R(a$,"NEdge",216);N(182,1,{},ah),R(a$,"NEdge/NEdgeBuilder",182),N(646,1,{},fae),R(a$,"NGraph",646),N(123,1,{123:1},zpt),p.c=-1,p.d=0,p.e=0,p.i=-1,p.j=!1;var q6e=R(a$,"NNode",123);N(788,1,gTt,l2e),p.Ic=function(t){Oa(this,t)},p.gd=function(t){sE(this,t)},p.Lc=function(){return new Bt(this,16)},p.Mc=function(){return new Ft(null,new Bt(this,16))},p._c=function(t,i){++this.b,Ky(this.a,t,i)},p.Ec=function(t){return uX(this,t)},p.ad=function(t,i){return++this.b,TCe(this.a,t,i)},p.Fc=function(t){return++this.b,$o(this.a,t)},p.$b=function(){++this.b,Bk(this.a.c,0)},p.Gc=function(t){return Oc(this.a,t,0)!=-1},p.Hc=function(t){return KH(this.a,t)},p.Xb=function(t){return ct(this.a,t)},p.bd=function(t){return Oc(this.a,t,0)},p.dc=function(){return this.a.c.length==0},p.Jc=function(){return cR(new ne(this.a))},p.cd=function(){throw ie(new cr)},p.dd=function(t){throw ie(new cr)},p.ed=function(t){return++this.b,_0(this.a,t)},p.Kc=function(t){return s_e(this,t)},p.fd=function(t,i){return++this.b,af(this.a,t,i)},p.gc=function(){return this.a.c.length},p.hd=function(t,i){return new Xv(this.a,t,i)},p.Nc=function(){return SX(this.a.c)},p.Oc=function(t){return kg(this.a,t)},p.b=0,R(a$,"NNode/ChangeAwareArrayList",788),N(271,1,{},qy),R(a$,"NNode/NNodeBuilder",271),N(1636,1,{},nh),p.a=!1,p.f=Qr,p.j=0,R(a$,"NetworkSimplex",1636),N(1267,1,xn,zo),p.Ad=function(t){ukt(this.a,f(t,680),!0,!1)},R(LTt,"NodeLabelAndSizeCalculator/lambda$0$Type",1267),N(550,1,{},rs),p.b=!0,p.c=!0,p.d=!0,p.e=!0,R(LTt,"NodeMarginCalculator",550),N(215,1,{215:1}),p.j=!1,p.k=!1;var hIt=R(EE,"Cell",215);N(126,215,{126:1,215:1},cut),p.ff=function(){return _X(this)},p.gf=function(){var t;return t=this.n,this.a.a+t.b+t.c},R(EE,"AtomicCell",126),N(236,23,{3:1,34:1,23:1,236:1},Uae);var _c,Bu,kc,WT=er(EE,"ContainerArea",236,ar,jEn,o1n),pIt;N(332,215,DTt),R(EE,"ContainerCell",332),N(1478,332,DTt,X0t),p.ff=function(){var t;return t=0,this.e?this.b?t=this.b.b:this.a[1][1]&&(t=this.a[1][1].ff()):t=ZCe(this,iwt(this,!0)),t>0?t+this.n.d+this.n.a:0},p.gf=function(){var t,i,a,u,l;if(l=0,this.e)this.b?l=this.b.a:this.a[1][1]&&(l=this.a[1][1].gf());else if(this.g)l=ZCe(this,Qle(this,null,!0));else for(i=(Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])),a=0,u=i.length;a<u;++a)t=i[a],l=h.Math.max(l,ZCe(this,Qle(this,t,!0)));return l>0?l+this.n.b+this.n.c:0},p.hf=function(){var t,i,a,u,l;if(this.g)for(t=Qle(this,null,!1),a=(Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])),u=0,l=a.length;u<l;++u)i=a[u],q2t(this,i,t);else for(a=(Np(),ce(se(WT,1),Ve,236,0,[_c,Bu,kc])),u=0,l=a.length;u<l;++u)i=a[u],t=Qle(this,i,!1),q2t(this,i,t)},p.jf=function(){var t,i,a,u;i=this.i,t=this.n,u=iwt(this,!1),KTe(this,(Np(),_c),i.d+t.d,u),KTe(this,kc,i.d+i.a-t.a-u[2],u),a=i.a-t.d-t.a,u[0]>0&&(u[0]+=this.d,a-=u[0]),u[2]>0&&(u[2]+=this.d,a-=u[2]),this.c.a=h.Math.max(0,a),this.c.d=i.d+t.d+(this.c.a-a)/2,u[1]=h.Math.max(u[1],a),KTe(this,Bu,i.d+t.d+u[0]-(u[1]-a)/2,u)},p.b=null,p.d=0,p.e=!1,p.f=!1,p.g=!1;var nge=0,Dee=0;R(EE,"GridContainerCell",1478),N(457,23,{3:1,34:1,23:1,457:1},zae);var Ew,Om,tp,gIt=er(EE,"HorizontalLabelAlignment",457,ar,$En,a1n),mIt;N(315,215,{215:1,315:1},kft,Dmt,vft),p.ff=function(){return rlt(this)},p.gf=function(){return dke(this)},p.a=0,p.c=!1;var tjn=R(EE,"LabelCell",315);N(251,332,{215:1,332:1,251:1},Oj),p.ff=function(){return $j(this)},p.gf=function(){return Fj(this)},p.hf=function(){Ude(this)},p.jf=function(){zde(this)},p.b=0,p.c=0,p.d=!1,R(EE,"StripContainerCell",251),N(1632,1,_r,Xc),p.Mb=function(t){return gpn(f(t,215))},R(EE,"StripContainerCell/lambda$0$Type",1632),N(1633,1,{},rc),p.We=function(t){return f(t,215).gf()},R(EE,"StripContainerCell/lambda$1$Type",1633),N(1634,1,_r,js),p.Mb=function(t){return mpn(f(t,215))},R(EE,"StripContainerCell/lambda$2$Type",1634),N(1635,1,{},Ac),p.We=function(t){return f(t,215).ff()},R(EE,"StripContainerCell/lambda$3$Type",1635),N(458,23,{3:1,34:1,23:1,458:1},Hae);var np,Sw,jp,bIt=er(EE,"VerticalLabelAlignment",458,ar,FEn,u1n),vIt;N(780,1,{},wOe),p.c=0,p.d=0,p.k=0,p.s=0,p.t=0,p.v=!1,p.w=0,p.D=!1,p.F=!1,R(xQ,"NodeContext",780),N(1476,1,Cr,Jc),p.Le=function(t,i){return yst(f(t,64),f(i,64))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(xQ,"NodeContext/0methodref$comparePortSides$Type",1476),N(1477,1,Cr,Id),p.Le=function(t,i){return JOn(f(t,115),f(i,115))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(xQ,"NodeContext/1methodref$comparePortContexts$Type",1477),N(166,23,{3:1,34:1,23:1,166:1},Mf);var wIt,yIt,EIt,SIt,_It,kIt,TIt,xIt,CIt,AIt,IIt,OIt,NIt,RIt,PIt,LIt,DIt,MIt,jIt,$It,FIt,rge,BIt=er(xQ,"NodeLabelLocation",166,ar,Ole,l1n),UIt;N(115,1,{115:1},IEt),p.a=!1,R(xQ,"PortContext",115),N(1481,1,xn,hu),p.Ad=function(t){Nit(f(t,315))},R(WG,MTt,1481),N(1482,1,_r,Qp),p.Mb=function(t){return!!f(t,115).c},R(WG,jTt,1482),N(1483,1,xn,Kb),p.Ad=function(t){Nit(f(t,115).c)},R(WG,"LabelPlacer/lambda$2$Type",1483);var V6e;N(1480,1,xn,S2),p.Ad=function(t){tT(),Khn(f(t,115))},R(WG,"NodeLabelAndSizeUtilities/lambda$0$Type",1480),N(781,1,xn,Y_e),p.Ad=function(t){cgn(this.b,this.c,this.a,f(t,187))},p.a=!1,p.c=!1,R(WG,"NodeLabelCellCreator/lambda$0$Type",781),N(1479,1,xn,Xs),p.Ad=function(t){Jhn(this.a,f(t,187))},R(WG,"PortContextCreator/lambda$0$Type",1479);var Mee;N(1848,1,{},_2),R(LL,"GreedyRectangleStripOverlapRemover",1848),N(1849,1,Cr,Km),p.Le=function(t,i){return $mn(f(t,225),f(i,225))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(LL,"GreedyRectangleStripOverlapRemover/0methodref$compareByYCoordinate$Type",1849),N(1795,1,{},srt),p.a=5,p.e=0,R(LL,"RectangleStripOverlapRemover",1795),N(1796,1,Cr,dv),p.Le=function(t,i){return Fmn(f(t,225),f(i,225))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(LL,"RectangleStripOverlapRemover/0methodref$compareLeftRectangleBorders$Type",1796),N(1798,1,Cr,K_),p.Le=function(t,i){return wwn(f(t,225),f(i,225))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(LL,"RectangleStripOverlapRemover/1methodref$compareRightRectangleBorders$Type",1798),N(406,23,{3:1,34:1,23:1,406:1},RY);var wq,ige,oge,yq,zIt=er(LL,"RectangleStripOverlapRemover/OverlapRemovalDirection",406,ar,O2n,d1n),HIt;N(225,1,{225:1},cce),R(LL,"RectangleStripOverlapRemover/RectangleNode",225),N(1797,1,xn,s1),p.Ad=function(t){AAn(this.a,f(t,225))},R(LL,"RectangleStripOverlapRemover/lambda$1$Type",1797);var GIt=!1,$$,W6e;N(1774,1,xn,YI),p.Ad=function(t){L_t(f(t,224))},R(PR,"DepthFirstCompaction/0methodref$compactTree$Type",1774),N(804,1,xn,pg),p.Ad=function(t){Fwn(this.a,f(t,224))},R(PR,"DepthFirstCompaction/lambda$1$Type",804),N(1775,1,xn,Put),p.Ad=function(t){g3n(this.a,this.b,this.c,f(t,224))},R(PR,"DepthFirstCompaction/lambda$2$Type",1775);var F$,K6e;N(68,1,{68:1},jft),R(PR,"Node",68),N(1174,1,{},Mst),R(PR,"ScanlineOverlapCheck",1174),N(1175,1,{678:1},sft),p._e=function(t){Abn(this,f(t,439))},R(PR,"ScanlineOverlapCheck/OverlapsScanlineHandler",1175),N(1176,1,Cr,n6),p.Le=function(t,i){return iCn(f(t,68),f(i,68))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(PR,"ScanlineOverlapCheck/OverlapsScanlineHandler/lambda$0$Type",1176),N(439,1,{439:1},sSe),p.a=!1,R(PR,"ScanlineOverlapCheck/Timestamp",439),N(1177,1,Cr,e0),p.Le=function(t,i){return CIn(f(t,439),f(i,439))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(PR,"ScanlineOverlapCheck/lambda$0$Type",1177),N(542,1,{},_l),R("org.eclipse.elk.alg.common.utils","SVGImage",542),N(743,1,{},rh),R(qfe,YOe,743),N(1159,1,Cr,t0),p.Le=function(t,i){return r4n(f(t,234),f(i,234))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(qfe,BTt,1159),N(1160,1,xn,zot),p.Ad=function(t){vEn(this.b,this.a,f(t,249))},R(qfe,XOe,1160),N(222,1,JS),R(cA,"AbstractLayoutProvider",222),N(721,222,JS,h2e),p.kf=function(t,i){v2t(this,t,i)},R(qfe,"ForceLayoutProvider",721);var njn=Bi(KG,UTt);N(148,1,{3:1,102:1,148:1},k2),p.of=function(t,i){return JH(this,t,i)},p.lf=function(){return vlt(this)},p.mf=function(t){return V(this,t)},p.nf=function(t){return li(this,t)},R(KG,"MapPropertyHolder",148),N(310,148,{3:1,310:1,102:1,148:1}),R(YG,"FParticle",310),N(249,310,{3:1,249:1,310:1,102:1,148:1},rdt),p.Ib=function(){var t;return this.a?(t=Oc(this.a.a,this,0),t>=0?"b"+t+"["+eue(this.a)+"]":"b["+eue(this.a)+"]"):"b_"+bS(this)},R(YG,"FBendpoint",249),N(288,148,{3:1,288:1,102:1,148:1},eut),p.Ib=function(){return eue(this)},R(YG,"FEdge",288),N(234,148,{3:1,234:1,102:1,148:1},vJ);var rjn=R(YG,"FGraph",234);N(442,310,{3:1,442:1,310:1,102:1,148:1},rpt),p.Ib=function(){return this.b==null||this.b.length==0?"l["+eue(this.a)+"]":"l_"+this.b},R(YG,"FLabel",442),N(151,310,{3:1,151:1,310:1,102:1,148:1},Dst),p.Ib=function(){return ETe(this)},p.a=0,R(YG,"FNode",151),N(2046,1,{}),p.qf=function(t){XIe(this,t)},p.rf=function(){nwt(this)},p.d=0,R(JOe,"AbstractForceModel",2046),N(626,2046,{626:1},Vbt),p.pf=function(t,i){var a,u,l,g,m;return j_t(this.f,t,i),l=Ho($a(i.d),t.d),m=h.Math.sqrt(l.a*l.a+l.b*l.b),u=h.Math.max(0,m-Z7(t.e)/2-Z7(i.e)/2),a=vEt(this.e,t,i),a>0?g=-dwn(u,this.c)*a:g=Qmn(u,this.b)*f(V(t,(Qh(),GR)),17).a,l.a*=g/m,l.b*=g/m,l},p.qf=function(t){XIe(this,t),this.a=f(V(t,(Qh(),$ee)),17).a,this.c=ke(Oe(V(t,Fee))),this.b=ke(Oe(V(t,sge)))},p.sf=function(t){return t<this.a},p.a=0,p.b=0,p.c=0,R(JOe,"EadesModel",626),N(627,2046,{627:1},Qut),p.pf=function(t,i){var a,u,l,g,m;return j_t(this.f,t,i),l=Ho($a(i.d),t.d),m=h.Math.sqrt(l.a*l.a+l.b*l.b),u=h.Math.max(0,m-Z7(t.e)/2-Z7(i.e)/2),g=ebn(u,this.a)*f(V(t,(Qh(),GR)),17).a,a=vEt(this.e,t,i),a>0&&(g-=lpn(u,this.a)*a),qv(l,g*this.b/m),l},p.qf=function(t){var i,a,u,l,g,m,v;for(XIe(this,t),this.b=ke(Oe(V(t,(Qh(),cge)))),this.c=this.b/f(V(t,$ee),17).a,u=t.e.c.length,g=0,l=0,v=new ne(t.e);v.a<v.c.c.length;)m=f(ee(v),151),g+=m.e.a,l+=m.e.b;i=g*l,a=ke(Oe(V(t,Fee)))*Og,this.a=h.Math.sqrt(i/(2*u))*a},p.rf=function(){nwt(this),this.b-=this.c},p.sf=function(t){return this.b>0},p.a=0,p.b=0,p.c=0,R(JOe,"FruchtermanReingoldModel",627);var HR=Bi(dc,"ILayoutMetaDataProvider");N(839,1,Ng,ss),p.tf=function(t){Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,CQ),""),"Force Model"),"Determines the model for force calculation."),Y6e),(bE(),Li)),X6e),At((wm(),en))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ZOe),""),"Iterations"),"The number of iterations on the force model."),rt(300)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,QOe),""),"Repulsive Power"),"Determines how many bend points are added to the edge; such bend points are regarded as repelling particles in the force model"),rt(0)),Hs),ba),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Vfe),""),"FR Temperature"),"The temperature is used as a scaling factor for particle displacements."),Tm),ya),mo),At(en)))),po(t,Vfe,CQ,JIt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Wfe),""),"Eades Repulsion"),"Factor for repulsive forces in Eades' model."),5),ya),mo),At(en)))),po(t,Wfe,CQ,KIt),kkt((new hg,t))};var qIt,VIt,Y6e,WIt,KIt,YIt,XIt,JIt;R(c$,"ForceMetaDataProvider",839),N(421,23,{3:1,34:1,23:1,421:1},cSe);var age,jee,X6e=er(c$,"ForceModelStrategy",421,ar,Cyn,h1n),ZIt;N(977,1,Ng,hg),p.tf=function(t){kkt(t)};var QIt,eOt,J6e,$ee,Z6e,tOt,nOt,rOt,iOt,Q6e,oOt,e5e,t5e,aOt,GR,sOt,sge,n5e,cOt,uOt,Fee,cge,lOt,dOt,fOt,r5e,hOt;R(c$,"ForceOptions",977),N(978,1,{},r6),p.uf=function(){var t;return t=new h2e,t},p.vf=function(t){},R(c$,"ForceOptions/ForceFactory",978);var Eq,B$,qR,Bee;N(840,1,Ng,P5),p.tf=function(t){Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,tNe),""),"Fixed Position"),"Prevent that the node is moved by the layout algorithm."),(mn(),!1)),(bE(),Po)),to),At((wm(),Ao))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,nNe),""),"Desired Edge Length"),"Either specified for parent nodes or for individual edges, where the latter takes higher precedence."),100),ya),mo),xi(en,ce(se(Ob,1),Ve,170,0,[Dm]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,rNe),""),"Layout Dimension"),"Dimensions that are permitted to be altered during layout."),i5e),Li),d5e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,iNe),""),"Stress Epsilon"),"Termination criterion for the iterative process."),Tm),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,oNe),""),"Iteration Limit"),"Maximum number of performed iterations. Takes higher precedence than 'epsilon'."),rt(Qr)),Hs),ba),At(en)))),Q_t((new L9,t))};var pOt,gOt,i5e,mOt,bOt,vOt;R(c$,"StressMetaDataProvider",840),N(981,1,Ng,L9),p.tf=function(t){Q_t(t)};var Uee,o5e,a5e,s5e,c5e,u5e,wOt,yOt,EOt,SOt,l5e,_Ot;R(c$,"StressOptions",981),N(982,1,{},i6),p.uf=function(){var t;return t=new tut,t},p.vf=function(t){},R(c$,"StressOptions/StressFactory",982),N(1075,222,JS,tut),p.kf=function(t,i){var a,u,l,g,m;for(i.Sg(VTt,1),lt(ut(Fe(t,(dG(),c5e))))?lt(ut(Fe(t,l5e)))||vH((a=new ri((fS(),new uS(t))),a)):v2t(new h2e,t,i.bh(1)),l=T1t(t),u=l_t(this.a,l),m=u.Jc();m.Ob();)g=f(m.Pb(),234),!(g.e.c.length<=1)&&(hMn(this.b,g),Z4n(this.b),Mu(g.d,new XI));l=Ekt(u),Nkt(l),i.Tg()},R(OQ,"StressLayoutProvider",1075),N(1076,1,xn,XI),p.Ad=function(t){oOe(f(t,442))},R(OQ,"StressLayoutProvider/lambda$0$Type",1076),N(979,1,{},ert),p.c=0,p.e=0,p.g=0,R(OQ,"StressMajorization",979),N(382,23,{3:1,34:1,23:1,382:1},Gae);var uge,lge,dge,d5e=er(OQ,"StressMajorization/Dimension",382,ar,UEn,p1n),kOt;N(980,1,Cr,c1),p.Le=function(t,i){return Wbn(this.a,f(t,151),f(i,151))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(OQ,"StressMajorization/lambda$0$Type",980),N(1156,1,{},aht),R(MR,"ElkLayered",1156),N(1157,1,xn,Qd),p.Ad=function(t){BRn(this.a,f(t,36))},R(MR,"ElkLayered/lambda$0$Type",1157),N(1158,1,xn,jv),p.Ad=function(t){Xbn(this.a,f(t,36))},R(MR,"ElkLayered/lambda$1$Type",1158),N(1241,1,{},jst);var TOt,xOt,COt;R(MR,"GraphConfigurator",1241),N(752,1,xn,Hy),p.Ad=function(t){wyt(this.a,f(t,9))},R(MR,"GraphConfigurator/lambda$0$Type",752),N(753,1,{},sy),p.Kb=function(t){return H3e(),new Ft(null,new Bt(f(t,25).a,16))},R(MR,"GraphConfigurator/lambda$1$Type",753),N(754,1,xn,wz),p.Ad=function(t){wyt(this.a,f(t,9))},R(MR,"GraphConfigurator/lambda$2$Type",754),N(1074,222,JS,rrt),p.kf=function(t,i){var a;a=KDn(new urt,t),Ae(Fe(t,(ot(),rx)))===Ae((_1(),W0))?lCn(this.a,a,i):K4n(this.a,a,i),i.Yg()||pkt(new R5,a)},R(MR,"LayeredLayoutProvider",1074),N(360,23,{3:1,34:1,23:1,360:1},Lz);var rp,Ab,su,cu,zs,f5e=er(MR,"LayeredPhases",360,ar,xSn,g1n),AOt;N(1659,1,{},ubt),p.i=0;var IOt;R(tq,"ComponentsToCGraphTransformer",1659);var OOt;N(1660,1,{},rC),p.wf=function(t,i){return h.Math.min(t.a!=null?ke(t.a):t.c.i,i.a!=null?ke(i.a):i.c.i)},p.xf=function(t,i){return h.Math.min(t.a!=null?ke(t.a):t.c.i,i.a!=null?ke(i.a):i.c.i)},R(tq,"ComponentsToCGraphTransformer/1",1660),N(82,1,{82:1}),p.i=0,p.k=!0,p.o=qo;var fge=R(f$,"CNode",82);N(456,82,{456:1,82:1},l_e,v3e),p.Ib=function(){return""},R(tq,"ComponentsToCGraphTransformer/CRectNode",456),N(1629,1,{},cy);var hge,pge;R(tq,"OneDimensionalComponentsCompaction",1629),N(1630,1,{},uy),p.Kb=function(t){return fEn(f(t,47))},p.Fb=function(t){return this===t},R(tq,"OneDimensionalComponentsCompaction/lambda$0$Type",1630),N(1631,1,{},ly),p.Kb=function(t){return yCn(f(t,47))},p.Fb=function(t){return this===t},R(tq,"OneDimensionalComponentsCompaction/lambda$1$Type",1631),N(1662,1,{},vdt),R(f$,"CGraph",1662),N(194,1,{194:1},Cle),p.b=0,p.c=0,p.e=0,p.g=!0,p.i=qo,R(f$,"CGroup",194),N(1661,1,{},Od),p.wf=function(t,i){return h.Math.max(t.a!=null?ke(t.a):t.c.i,i.a!=null?ke(i.a):i.c.i)},p.xf=function(t,i){return h.Math.max(t.a!=null?ke(t.a):t.c.i,i.a!=null?ke(i.a):i.c.i)},R(f$,_Tt,1661),N(1663,1,{},yEt),p.d=!1;var NOt,gge=R(f$,xTt,1663);N(1664,1,{},fv),p.Kb=function(t){return Q2e(),mn(),f(f(t,47).a,82).d.e!=0},p.Fb=function(t){return this===t},R(f$,CTt,1664),N(812,1,{},fke),p.a=!1,p.b=!1,p.c=!1,p.d=!1,R(f$,ATt,812),N(1844,1,{},Ilt),R(NQ,ITt,1844);var Sq=Bi(SE,ETt);N(1845,1,{374:1},uft),p._e=function(t){X5n(this,f(t,462))},R(NQ,OTt,1845),N(1846,1,Cr,JI),p.Le=function(t,i){return nyn(f(t,82),f(i,82))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(NQ,NTt,1846),N(462,1,{462:1},lSe),p.a=!1,R(NQ,RTt,462),N(1847,1,Cr,ZI),p.Le=function(t,i){return AIn(f(t,462),f(i,462))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(NQ,PTt,1847),N(143,1,{143:1},lP,ike),p.Fb=function(t){var i;return t==null||ijn!=Fd(t)?!1:(i=f(t,143),Pu(this.c,i.c)&&Pu(this.d,i.d))},p.Hb=function(){return oZ(ce(se(Ro,1),Jt,1,5,[this.c,this.d]))},p.Ib=function(){return"("+this.c+Fu+this.d+(this.a?"cx":"")+this.b+")"},p.a=!0,p.c=0,p.d=0;var ijn=R(SE,"Point",143);N(405,23,{3:1,34:1,23:1,405:1},PY);var a_,KT,bA,YT,ROt=er(SE,"Point/Quadrant",405,ar,N2n,f1n),POt;N(1650,1,{},irt),p.b=null,p.c=null,p.d=null,p.e=null,p.f=null;var LOt,DOt,MOt,jOt,$Ot;R(SE,"RectilinearConvexHull",1650),N(565,1,{374:1},xZ),p._e=function(t){w_n(this,f(t,143))},p.b=0;var h5e;R(SE,"RectilinearConvexHull/MaximalElementsEventHandler",565),N(1652,1,Cr,hv),p.Le=function(t,i){return eyn(Oe(t),Oe(i))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(SE,"RectilinearConvexHull/MaximalElementsEventHandler/lambda$0$Type",1652),N(1651,1,{374:1},Pmt),p._e=function(t){p5n(this,f(t,143))},p.a=0,p.b=null,p.c=null,p.d=null,p.e=null,R(SE,"RectilinearConvexHull/RectangleEventHandler",1651),N(1653,1,Cr,dy),p.Le=function(t,i){return i2n(f(t,143),f(i,143))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(SE,"RectilinearConvexHull/lambda$0$Type",1653),N(1654,1,Cr,T2),p.Le=function(t,i){return o2n(f(t,143),f(i,143))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(SE,"RectilinearConvexHull/lambda$1$Type",1654),N(1655,1,Cr,QI),p.Le=function(t,i){return a2n(f(t,143),f(i,143))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(SE,"RectilinearConvexHull/lambda$2$Type",1655),N(1656,1,Cr,x2),p.Le=function(t,i){return s2n(f(t,143),f(i,143))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(SE,"RectilinearConvexHull/lambda$3$Type",1656),N(1657,1,Cr,o6),p.Le=function(t,i){return fNn(f(t,143),f(i,143))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(SE,"RectilinearConvexHull/lambda$4$Type",1657),N(1658,1,{},Fft),R(SE,"Scanline",1658),N(2050,1,{}),R(Rg,"AbstractGraphPlacer",2050),N(331,1,{331:1},Cct),p.Df=function(t){return this.Ef(t)?(jt(this.b,f(V(t,(Ge(),R1)),22),t),!0):!1},p.Ef=function(t){var i,a,u,l;for(i=f(V(t,(Ge(),R1)),22),l=f(si(gi,i),22),u=l.Jc();u.Ob();)if(a=f(u.Pb(),22),!f(si(this.b,a),15).dc())return!1;return!0};var gi;R(Rg,"ComponentGroup",331),N(761,2050,{},p2e),p.Ff=function(t){var i,a;for(a=new ne(this.a);a.a<a.c.c.length;)if(i=f(ee(a),331),i.Df(t))return;Qe(this.a,new Cct(t))},p.Cf=function(t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W;if(this.a.c.length=0,i.a.c.length=0,t.dc()){i.f.a=0,i.f.b=0;return}for(m=f(t.Xb(0),36),Rc(i,m),l=t.Jc();l.Ob();)u=f(l.Pb(),36),this.Ff(u);for(W=new ha,g=ke(Oe(V(m,(ot(),jq)))),T=new ne(this.a);T.a<T.c.c.length;)v=f(ee(T),331),A=Rkt(v,g),k0(HX(v.b),W.a,W.b),W.a+=A.a,W.b+=A.b;if(i.f.a=W.a-g,i.f.b=W.b-g,lt(ut(V(m,xte)))&&Ae(V(m,P1))===Ae((k1(),DD))){for(z=t.Jc();z.Ob();)L=f(z.Pb(),36),wL(L,L.c.a,L.c.b);for(a=new fy,kOe(a,t,g),F=t.Jc();F.Ob();)L=f(F.Pb(),36),ii(Cp(L.c),a.e);ii(Cp(i.f),a.a)}for(S=new ne(this.a);S.a<S.c.c.length;)v=f(ee(S),331),lxe(i,HX(v.b))},R(Rg,"ComponentGroupGraphPlacer",761),N(1265,761,{},grt),p.Ff=function(t){b0t(this,t)},p.Cf=function(t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee;if(this.a.c.length=0,i.a.c.length=0,t.dc()){i.f.a=0,i.f.b=0;return}for(m=f(t.Xb(0),36),Rc(i,m),l=t.Jc();l.Ob();)u=f(l.Pb(),36),b0t(this,u);for(Ee=new ha,ve=new ha,J=new ha,W=new ha,g=ke(Oe(V(m,(ot(),jq)))),T=new ne(this.a);T.a<T.c.c.length;){if(v=f(ee(T),331),l1(f(V(i,(Rr(),BE)),86))){for(J.a=Ee.a,pe=new s7(J7(Ece(v.b).a).a.kc());pe.b.Ob();)if(ue=f(IY(pe.b.Pb()),22),ue.Gc((tt(),hn))){J.a=ve.a;break}}else if(k3(f(V(i,BE),86))){for(J.b=Ee.b,pe=new s7(J7(Ece(v.b).a).a.kc());pe.b.Ob();)if(ue=f(IY(pe.b.Pb()),22),ue.Gc((tt(),pn))){J.b=ve.b;break}}if(A=Rkt(f(v,561),g),k0(HX(v.b),J.a,J.b),l1(f(V(i,BE),86))){for(ve.a=J.a+A.a,W.a=h.Math.max(W.a,ve.a),pe=new s7(J7(Ece(v.b).a).a.kc());pe.b.Ob();)if(ue=f(IY(pe.b.Pb()),22),ue.Gc((tt(),Bn))){Ee.a=J.a+A.a;break}ve.b=J.b+A.b,J.b=ve.b,W.b=h.Math.max(W.b,J.b)}else if(k3(f(V(i,BE),86))){for(ve.b=J.b+A.b,W.b=h.Math.max(W.b,ve.b),pe=new s7(J7(Ece(v.b).a).a.kc());pe.b.Ob();)if(ue=f(IY(pe.b.Pb()),22),ue.Gc((tt(),_n))){Ee.b=J.b+A.b;break}ve.a=J.a+A.a,J.a=ve.a,W.a=h.Math.max(W.a,J.a)}}if(i.f.a=W.a-g,i.f.b=W.b-g,lt(ut(V(m,xte)))&&Ae(V(m,P1))===Ae((k1(),DD))){for(z=t.Jc();z.Ob();)L=f(z.Pb(),36),wL(L,L.c.a,L.c.b);for(a=new fy,kOe(a,t,g),F=t.Jc();F.Ob();)L=f(F.Pb(),36),ii(Cp(L.c),a.e);ii(Cp(i.f),a.a)}for(S=new ne(this.a);S.a<S.c.c.length;)v=f(ee(S),331),lxe(i,HX(v.b))},R(Rg,"ComponentGroupModelOrderGraphPlacer",1265),N(381,23,{3:1,34:1,23:1,381:1},LY);var mge,p5e,bge,XT,g5e=er(Rg,"ComponentOrderingStrategy",381,ar,P2n,m1n),FOt;N(643,1,{},fy),R(Rg,"ComponentsCompactor",643),N(1473,13,hTt,Npt),p.Ec=function(t){return Dj(this,f(t,143))},R(Rg,"ComponentsCompactor/Hullpoints",1473),N(1470,1,{834:1},Mvt),p.a=!1,R(Rg,"ComponentsCompactor/InternalComponent",1470),N(1469,1,Eb,art),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return new ne(this.a)},R(Rg,"ComponentsCompactor/InternalConnectedComponents",1469),N(1472,1,{586:1},_Et),p.zf=function(){return null},p.Af=function(){return this.a},p.yf=function(){return Nle(this.d)},p.Bf=function(){return this.b},R(Rg,"ComponentsCompactor/InternalExternalExtension",1472),N(1471,1,{586:1},crt),p.Af=function(){return this.a},p.yf=function(){return Nle(this.d)},p.zf=function(){return this.c},p.Bf=function(){return this.b},R(Rg,"ComponentsCompactor/InternalUnionExternalExtension",1471),N(1475,1,{},V2t),R(Rg,"ComponentsCompactor/OuterSegments",1475),N(1474,1,{},ort),R(Rg,"ComponentsCompactor/Segments",1474),N(1242,1,{},Hpt),R(Rg,YOe,1242),N(1243,1,Cr,Y_),p.Le=function(t,i){return d2n(f(t,36),f(i,36))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Rg,"ComponentsProcessor/lambda$0$Type",1243),N(561,331,{331:1,561:1},Opt),p.Df=function(t){return yCe(this,t)},p.Ef=function(t){return A2t(this,t)};var Ar;R(Rg,"ModelOrderComponentGroup",561),N(1263,2050,{},Ym),p.Cf=function(t,i){var a,u,l,g,m,v,S,T,A,L,F;if(t.gc()==1){L=f(t.Xb(0),36),L!=i&&(i.a.c.length=0,OSt(i,L,0,0),Rc(i,L),rTe(i.d,L.d),i.f.a=L.f.a,i.f.b=L.f.b);return}else if(t.dc()){i.a.c.length=0,i.f.a=0,i.f.b=0;return}for(this.Hf(t,i),l=f(t.Xb(0),36),i.a.c.length=0,Rc(i,l),T=0,F=0,m=t.Jc();m.Ob();)g=f(m.Pb(),36),A=g.f,T=h.Math.max(T,A.a),F+=A.a*A.b;if(T=h.Math.max(T,h.Math.sqrt(F)*ke(Oe(V(i,(ot(),Tte))))),u=ke(Oe(V(i,jq))),this.Gf(t,i,T,u),lt(ut(V(l,xte)))){for(a=new fy,kOe(a,t,u),S=t.Jc();S.Ob();)v=f(S.Pb(),36),ii(Cp(v.c),a.e);ii(Cp(i.f),a.a)}lxe(i,t)},p.Gf=function(t,i,a,u){var l,g,m,v,S,T,A,L;for(A=0,L=0,v=0,l=u,m=t.Jc();m.Ob();)g=f(m.Pb(),36),T=g.f,A+T.a>a&&(A=0,L+=v+u,v=0),S=g.c,wL(g,A+S.a,L+S.b),Cp(S),l=h.Math.max(l,A+T.a),v=h.Math.max(v,T.b),A+=T.a+u;i.f.a=l,i.f.b=L+v},p.Hf=function(t,i){var a,u,l,g,m;if(Ae(V(i,(ot(),nx)))===Ae((hE(),XT))){for(u=t.Jc();u.Ob();){for(a=f(u.Pb(),36),m=0,g=new ne(a.a);g.a<g.c.c.length;)l=f(ee(g),9),m+=f(V(l,t6t),17).a;a.p=m}Ht(),t.gd(new iC)}},R(Rg,"SimpleRowGraphPlacer",1263),N(1266,1263,{},C2),p.Gf=function(t,i,a,u){var l,g,m,v,S,T,A,L,F,z;for(F=0,z=0,v=0,l=u,S=null,L=0,m=t.Jc();m.Ob();)g=f(m.Pb(),36),A=g.f,(F+A.a>a&&!f(V(g,(Ge(),R1)),22).Gc((tt(),hn))||S&&f(V(S,(Ge(),R1)),22).Gc((tt(),_n))||f(V(g,(Ge(),R1)),22).Gc((tt(),pn)))&&(F=L,z+=v+u,v=0),T=g.c,f(V(g,(Ge(),R1)),22).Gc((tt(),hn))&&(F=l+u),wL(g,F+T.a,z+T.b),l=h.Math.max(l,F+A.a),f(V(g,R1),22).Gc(Bn)&&(L=h.Math.max(L,F+A.a+u)),Cp(T),v=h.Math.max(v,A.b),F+=A.a+u,S=g;i.f.a=l,i.f.b=z+v},p.Hf=function(t,i){},R(Rg,"ModelOrderRowGraphPlacer",1266),N(1264,1,Cr,iC),p.Le=function(t,i){return wTn(f(t,36),f(i,36))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Rg,"SimpleRowGraphPlacer/1",1264);var BOt;N(1240,1,km,hy),p.Lb=function(t){var i;return i=f(V(f(t,248).b,(ot(),Ls)),77),!!i&&i.b!=0},p.Fb=function(t){return this===t},p.Mb=function(t){var i;return i=f(V(f(t,248).b,(ot(),Ls)),77),!!i&&i.b!=0},R(RQ,"CompoundGraphPostprocessor/1",1240),N(1239,1,vi,lrt),p.If=function(t,i){Fvt(this,f(t,36),i)},R(RQ,"CompoundGraphPreprocessor",1239),N(441,1,{441:1},x0t),p.c=!1,R(RQ,"CompoundGraphPreprocessor/ExternalPort",441),N(248,1,{248:1},vX),p.Ib=function(){return Lse(this.c)+":"+pEt(this.b)},R(RQ,"CrossHierarchyEdge",248),N(759,1,Cr,ZEe),p.Le=function(t,i){return nIn(this,f(t,248),f(i,248))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(RQ,"CrossHierarchyEdgeComparator",759),N(306,148,{3:1,306:1,102:1,148:1}),p.p=0,R(ku,"LGraphElement",306),N(18,306,{3:1,18:1,306:1,102:1,148:1},xS),p.Ib=function(){return pEt(this)};var vge=R(ku,"LEdge",18);N(36,306,{3:1,20:1,36:1,306:1,102:1,148:1},Rxe),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return new ne(this.b)},p.Ib=function(){return this.b.c.length==0?"G-unlayered"+Cg(this.a):this.a.c.length==0?"G-layered"+Cg(this.b):"G[layerless"+Cg(this.a)+", layers"+Cg(this.b)+"]"};var UOt=R(ku,"LGraph",36),zOt;N(650,1,{}),p.Jf=function(){return this.e.n},p.mf=function(t){return V(this.e,t)},p.Kf=function(){return this.e.o},p.Lf=function(){return this.e.p},p.nf=function(t){return li(this.e,t)},p.Mf=function(t){this.e.n.a=t.a,this.e.n.b=t.b},p.Nf=function(t){this.e.o.a=t.a,this.e.o.b=t.b},p.Of=function(t){this.e.p=t},R(ku,"LGraphAdapters/AbstractLShapeAdapter",650),N(460,1,{832:1},t7),p.Pf=function(){var t,i;if(!this.b)for(this.b=lb(this.a.b.c.length),i=new ne(this.a.b);i.a<i.c.c.length;)t=f(ee(i),72),Qe(this.b,new nY(t));return this.b},p.b=null,R(ku,"LGraphAdapters/LEdgeAdapter",460),N(649,1,{},Oce),p.Qf=function(){var t,i,a,u,l,g;if(!this.b){for(this.b=new et,u=new ne(this.a.b);u.a<u.c.c.length;)for(a=f(ee(u),25),g=new ne(a.a);g.a<g.c.c.length;)if(l=f(ee(g),9),this.c.Mb(l)&&(Qe(this.b,new bX(this,l,this.e)),this.d)){if(li(l,(Ge(),QR)))for(i=f(V(l,QR),15).Jc();i.Ob();)t=f(i.Pb(),9),Qe(this.b,new bX(this,t,!1));if(li(l,KR))for(i=f(V(l,KR),15).Jc();i.Ob();)t=f(i.Pb(),9),Qe(this.b,new bX(this,t,!1))}}return this.b},p.Jf=function(){throw ie(new f0(YTt))},p.mf=function(t){return V(this.a,t)},p.Kf=function(){return this.a.f},p.Lf=function(){return this.a.p},p.nf=function(t){return li(this.a,t)},p.Mf=function(t){throw ie(new f0(YTt))},p.Nf=function(t){this.a.f.a=t.a,this.a.f.b=t.b},p.Of=function(t){this.a.p=t},p.b=null,p.d=!1,p.e=!1,R(ku,"LGraphAdapters/LGraphAdapter",649),N(567,650,{187:1},nY),R(ku,"LGraphAdapters/LLabelAdapter",567),N(566,650,{680:1},bX),p.Rf=function(){return this.b},p.Sf=function(){return Ht(),Ht(),Ga},p.Pf=function(){var t,i;if(!this.a)for(this.a=lb(f(this.e,9).b.c.length),i=new ne(f(this.e,9).b);i.a<i.c.c.length;)t=f(ee(i),72),Qe(this.a,new nY(t));return this.a},p.Tf=function(){var t;return t=f(this.e,9).d,new L_e(t.d,t.c,t.a,t.b)},p.Uf=function(){return Ht(),Ht(),Ga},p.Vf=function(){var t,i;if(!this.c)for(this.c=lb(f(this.e,9).j.c.length),i=new ne(f(this.e,9).j);i.a<i.c.c.length;)t=f(ee(i),11),Qe(this.c,new Hot(t,this.d));return this.c},p.Wf=function(){return lt(ut(V(f(this.e,9),(Ge(),MPe))))},p.Xf=function(t){f(this.e,9).d.b=t.b,f(this.e,9).d.d=t.d,f(this.e,9).d.c=t.c,f(this.e,9).d.a=t.a},p.Yf=function(t){f(this.e,9).f.b=t.b,f(this.e,9).f.d=t.d,f(this.e,9).f.c=t.c,f(this.e,9).f.a=t.a},p.Zf=function(){tkn(this,(h7(),zOt))},p.a=null,p.b=null,p.c=null,p.d=!1,R(ku,"LGraphAdapters/LNodeAdapter",566),N(1734,650,{831:1},Hot),p.Sf=function(){var t,i,a,u,l,g,m,v;if(this.d&&f(this.e,11).i.k==(cn(),xu))return Ht(),Ht(),Ga;if(!this.a){for(this.a=new et,a=new ne(f(this.e,11).e);a.a<a.c.c.length;)t=f(ee(a),18),Qe(this.a,new t7(t));if(this.d&&(u=f(V(f(this.e,11),(Ge(),ed)),9),u))for(i=new Tn(Cn(Fo(u).a.Jc(),new B));Kn(i);)t=f(Dn(i),18),Qe(this.a,new t7(t));if(li(f(this.e,11).i,(Ge(),f_))&&(m=f(V(f(this.e,11).i,f_),333),v=f(Eg(m.e,this.e),112),v))for(g=new ne(v.b);g.a<g.c.c.length;)l=f(ee(g),336),Qe(this.a,new t7(l.a))}return this.a},p.Pf=function(){var t,i;if(!this.b)for(this.b=lb(f(this.e,11).f.c.length),i=new ne(f(this.e,11).f);i.a<i.c.c.length;)t=f(ee(i),72),Qe(this.b,new nY(t));return this.b},p.Uf=function(){var t,i,a,u,l,g,m,v;if(this.d&&f(this.e,11).i.k==(cn(),xu))return Ht(),Ht(),Ga;if(!this.c){for(this.c=new et,a=new ne(f(this.e,11).g);a.a<a.c.c.length;)t=f(ee(a),18),Qe(this.c,new t7(t));if(this.d&&(u=f(V(f(this.e,11),(Ge(),ed)),9),u))for(i=new Tn(Cn(ao(u).a.Jc(),new B));Kn(i);)t=f(Dn(i),18),Qe(this.c,new t7(t));if(li(f(this.e,11).i,(Ge(),f_))&&(m=f(V(f(this.e,11).i,f_),333),v=f(Eg(m.e,this.e),112),v))for(g=new ne(v.e);g.a<g.c.c.length;)l=f(ee(g),336),Qe(this.c,new t7(l.a))}return this.c},p.$f=function(){return f(this.e,11).j},p._f=function(){return lt(ut(V(f(this.e,11),(Ge(),Rq))))},p.a=null,p.b=null,p.c=null,p.d=!1,R(ku,"LGraphAdapters/LPortAdapter",1734),N(1735,1,Cr,oC),p.Le=function(t,i){return xPn(f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(ku,"LGraphAdapters/PortComparator",1735),N(798,1,_r,py),p.Mb=function(t){return f(t,9),h7(),!0},R(ku,"LGraphAdapters/lambda$0$Type",798),N(395,306,{3:1,306:1,395:1,102:1,148:1}),R(ku,"LShape",395),N(72,395,{3:1,306:1,72:1,395:1,102:1,148:1},mit,HSe),p.Ib=function(){var t;return t=xvn(this),t==null?"label":"l_"+t},R(ku,"LLabel",72),N(213,1,{3:1,4:1,213:1,411:1}),p.Fb=function(t){var i;return me(t,213)?(i=f(t,213),this.d==i.d&&this.a==i.a&&this.b==i.b&&this.c==i.c):!1},p.Hb=function(){var t,i;return t=FN(this.b)<<16,t|=FN(this.a)&To,i=FN(this.c)<<16,i|=FN(this.d)&To,t^i},p.ag=function(t){var i,a,u,l,g,m,v,S,T,A,L;for(g=0;g<t.length&&n0t((gn(g,t.length),t.charCodeAt(g)),JTt);)++g;for(i=t.length;i>0&&n0t((gn(i-1,t.length),t.charCodeAt(i-1)),ZTt);)--i;if(g<i){A=DT((ga(g,i,t.length),t.substr(g,i-g)),",|;");try{for(v=A,S=0,T=v.length;S<T;++S){if(m=v[S],l=DT(m,"="),l.length!=2)throw ie(new dn("Expecting a list of key-value pairs."));u=IT(l[0]),L=AT(IT(l[1])),Mt(u,"top")?this.d=L:Mt(u,"left")?this.b=L:Mt(u,"bottom")?this.a=L:Mt(u,"right")&&(this.c=L)}}catch(F){throw F=go(F),me(F,130)?(a=F,ie(new dn(QTt+a))):ie(F)}}},p.Ib=function(){return"[top="+this.d+",left="+this.b+",bottom="+this.a+",right="+this.c+"]"},p.a=0,p.b=0,p.c=0,p.d=0,R(nq,"Spacing",213),N(139,213,ext,i7,wst,L_e,Wse);var m5e=R(nq,"ElkMargin",139);N(644,139,ext,hae),R(ku,"LMargin",644),N(9,395,{3:1,306:1,9:1,395:1,102:1,148:1},Tg),p.Ib=function(){return Rvt(this)},p.i=!1;var Ib=R(ku,"LNode",9);N(247,23,{3:1,34:1,23:1,247:1},MN);var Dg,bo,Vc,uo,U$,Yi,xu,_q,wge=er(ku,"LNode/NodeType",247,ar,Y_n,b1n),HOt;N(757,1,_r,VM),p.Mb=function(t){return lt(ut(V(f(t,72),(ot(),hme))))},R(ku,"LNode/lambda$0$Type",757),N(107,213,txt,xN,gS,W_e);var b5e=R(nq,"ElkPadding",107);N(760,107,txt,b2e),R(ku,"LPadding",760),N(11,395,{3:1,306:1,11:1,395:1,102:1,148:1},ou),p.Ib=function(){var t,i,a;return t=new $v,xr((t.a+="p_",t),DZ(this)),this.i&&xr(vu((t.a+="[",t),this.i),"]"),this.e.c.length==1&&this.g.c.length==0&&f(ct(this.e,0),18).c!=this&&(i=f(ct(this.e,0),18).c,xr((t.a+=" << ",t),DZ(i)),xr(vu((t.a+="[",t),i.i),"]")),this.e.c.length==0&&this.g.c.length==1&&f(ct(this.g,0),18).d!=this&&(a=f(ct(this.g,0),18).d,xr((t.a+=" >> ",t),DZ(a)),xr(vu((t.a+="[",t),a.i),"]")),t.a},p.c=!0,p.d=!1;var v5e,w5e,y5e,E5e,S5e,_5e,GOt=R(ku,"LPort",11);N(397,1,Eb,Z5),p.Ic=function(t){Oa(this,t)},p.Jc=function(){var t;return t=new ne(this.a.e),new eet(t)},R(ku,"LPort/1",397),N(1262,1,aa,eet),p.Nb=function(t){va(this,t)},p.Pb=function(){return f(ee(this.a),18).c},p.Ob=function(){return sc(this.a)},p.Qb=function(){Y7(this.a)},R(ku,"LPort/1/1",1262),N(362,1,Eb,_N),p.Ic=function(t){Oa(this,t)},p.Jc=function(){var t;return t=new ne(this.a.g),new QEe(t)},R(ku,"LPort/2",362),N(758,1,aa,QEe),p.Nb=function(t){va(this,t)},p.Pb=function(){return f(ee(this.a),18).d},p.Ob=function(){return sc(this.a)},p.Qb=function(){Y7(this.a)},R(ku,"LPort/2/1",758),N(1255,1,Eb,Got),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return new yg(this)},R(ku,"LPort/CombineIter",1255),N(207,1,aa,yg),p.Nb=function(t){va(this,t)},p.Qb=function(){Eit()},p.Ob=function(){return F7(this)},p.Pb=function(){return sc(this.a)?ee(this.a):ee(this.b)},R(ku,"LPort/CombineIter/1",207),N(1256,1,km,a6),p.Lb=function(t){return Hlt(t)},p.Fb=function(t){return this===t},p.Mb=function(t){return Nl(),f(t,11).g.c.length!=0},R(ku,"LPort/lambda$0$Type",1256),N(1257,1,km,Kg),p.Lb=function(t){return Glt(t)},p.Fb=function(t){return this===t},p.Mb=function(t){return Nl(),f(t,11).e.c.length!=0},R(ku,"LPort/lambda$1$Type",1257),N(1258,1,km,s6),p.Lb=function(t){return Nl(),f(t,11).j==(tt(),hn)},p.Fb=function(t){return this===t},p.Mb=function(t){return Nl(),f(t,11).j==(tt(),hn)},R(ku,"LPort/lambda$2$Type",1258),N(1259,1,km,c6),p.Lb=function(t){return Nl(),f(t,11).j==(tt(),_n)},p.Fb=function(t){return this===t},p.Mb=function(t){return Nl(),f(t,11).j==(tt(),_n)},R(ku,"LPort/lambda$3$Type",1259),N(1260,1,km,eO),p.Lb=function(t){return Nl(),f(t,11).j==(tt(),Bn)},p.Fb=function(t){return this===t},p.Mb=function(t){return Nl(),f(t,11).j==(tt(),Bn)},R(ku,"LPort/lambda$4$Type",1260),N(1261,1,km,u6),p.Lb=function(t){return Nl(),f(t,11).j==(tt(),pn)},p.Fb=function(t){return this===t},p.Mb=function(t){return Nl(),f(t,11).j==(tt(),pn)},R(ku,"LPort/lambda$5$Type",1261),N(25,306,{3:1,20:1,306:1,25:1,102:1,148:1},ru),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return new ne(this.a)},p.Ib=function(){return"L_"+Oc(this.b.b,this,0)+Cg(this.a)},R(ku,"Layer",25),N(1271,1,{},urt),R($0,nxt,1271),N(1275,1,{},A2),p.Kb=function(t){return Us(f(t,84))},R($0,"ElkGraphImporter/0methodref$connectableShapeToNode$Type",1275),N(1278,1,{},WM),p.Kb=function(t){return Us(f(t,84))},R($0,"ElkGraphImporter/1methodref$connectableShapeToNode$Type",1278),N(1272,1,xn,tet),p.Ad=function(t){REt(this.a,f(t,124))},R($0,XOe,1272),N(1273,1,xn,net),p.Ad=function(t){REt(this.a,f(t,124))},R($0,rxt,1273),N(1274,1,{},Xm),p.Kb=function(t){return new Ft(null,new Bt(Zke(f(t,85)),16))},R($0,ixt,1274),N(1276,1,_r,ret),p.Mb=function(t){return Qgn(this.a,f(t,27))},R($0,oxt,1276),N(1277,1,{},tO),p.Kb=function(t){return new Ft(null,new Bt(Kwn(f(t,85)),16))},R($0,"ElkGraphImporter/lambda$5$Type",1277),N(1279,1,_r,iet),p.Mb=function(t){return emn(this.a,f(t,27))},R($0,"ElkGraphImporter/lambda$7$Type",1279),N(1280,1,_r,l6),p.Mb=function(t){return lyn(f(t,85))},R($0,"ElkGraphImporter/lambda$8$Type",1280),N(1250,1,{},R5);var qOt;R($0,"ElkGraphLayoutTransferrer",1250),N(1251,1,_r,oet),p.Mb=function(t){return Dbn(this.a,f(t,18))},R($0,"ElkGraphLayoutTransferrer/lambda$0$Type",1251),N(1252,1,xn,aet),p.Ad=function(t){Rz(),Qe(this.a,f(t,18))},R($0,"ElkGraphLayoutTransferrer/lambda$1$Type",1252),N(1253,1,_r,set),p.Mb=function(t){return xbn(this.a,f(t,18))},R($0,"ElkGraphLayoutTransferrer/lambda$2$Type",1253),N(1254,1,xn,cet),p.Ad=function(t){Rz(),Qe(this.a,f(t,18))},R($0,"ElkGraphLayoutTransferrer/lambda$3$Type",1254),N(799,1,{},P_e),R(yn,"BiLinkedHashMultiMap",799),N(1490,1,vi,aC),p.If=function(t,i){jkn(f(t,36),i)},R(yn,"CommentNodeMarginCalculator",1490),N(1491,1,{},X_),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"CommentNodeMarginCalculator/lambda$0$Type",1491),N(1492,1,xn,d6),p.Ad=function(t){qDn(f(t,9))},R(yn,"CommentNodeMarginCalculator/lambda$1$Type",1492),N(1493,1,vi,nO),p.If=function(t,i){nPn(f(t,36),i)},R(yn,"CommentPostprocessor",1493),N(1494,1,vi,gy),p.If=function(t,i){h9n(f(t,36),i)},R(yn,"CommentPreprocessor",1494),N(1495,1,vi,J_),p.If=function(t,i){v5n(f(t,36),i)},R(yn,"ConstraintsPostprocessor",1495),N(1496,1,vi,f6),p.If=function(t,i){sTn(f(t,36),i)},R(yn,"EdgeAndLayerConstraintEdgeReverser",1496),N(1497,1,vi,h6),p.If=function(t,i){FCn(f(t,36),i)},R(yn,"EndLabelPostprocessor",1497),N(1498,1,{},p6),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"EndLabelPostprocessor/lambda$0$Type",1498),N(1499,1,_r,KM),p.Mb=function(t){return SSn(f(t,9))},R(yn,"EndLabelPostprocessor/lambda$1$Type",1499),N(1500,1,xn,YM),p.Ad=function(t){IIn(f(t,9))},R(yn,"EndLabelPostprocessor/lambda$2$Type",1500),N(1501,1,vi,sC),p.If=function(t,i){hRn(f(t,36),i)},R(yn,"EndLabelPreprocessor",1501),N(1502,1,{},rO),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"EndLabelPreprocessor/lambda$0$Type",1502),N(1503,1,xn,Lut),p.Ad=function(t){sgn(this.a,this.b,this.c,f(t,9))},p.a=0,p.b=0,p.c=!1,R(yn,"EndLabelPreprocessor/lambda$1$Type",1503),N(1504,1,_r,g6),p.Mb=function(t){return Ae(V(f(t,72),(ot(),Rm)))===Ae((Sg(),LD))},R(yn,"EndLabelPreprocessor/lambda$2$Type",1504),N(1505,1,xn,uet),p.Ad=function(t){Mr(this.a,f(t,72))},R(yn,"EndLabelPreprocessor/lambda$3$Type",1505),N(1506,1,_r,Z_),p.Mb=function(t){return Ae(V(f(t,72),(ot(),Rm)))===Ae((Sg(),gx))},R(yn,"EndLabelPreprocessor/lambda$4$Type",1506),N(1507,1,xn,det),p.Ad=function(t){Mr(this.a,f(t,72))},R(yn,"EndLabelPreprocessor/lambda$5$Type",1507),N(1555,1,vi,Pv),p.If=function(t,i){eCn(f(t,36),i)};var VOt;R(yn,"EndLabelSorter",1555),N(1556,1,Cr,Q_),p.Le=function(t,i){return y3n(f(t,452),f(i,452))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"EndLabelSorter/1",1556),N(452,1,{452:1},ift),R(yn,"EndLabelSorter/LabelGroup",452),N(1557,1,{},m6),p.Kb=function(t){return Nz(),new Ft(null,new Bt(f(t,25).a,16))},R(yn,"EndLabelSorter/lambda$0$Type",1557),N(1558,1,_r,b6),p.Mb=function(t){return Nz(),f(t,9).k==(cn(),Yi)},R(yn,"EndLabelSorter/lambda$1$Type",1558),N(1559,1,xn,XM),p.Ad=function(t){CNn(f(t,9))},R(yn,"EndLabelSorter/lambda$2$Type",1559),N(1560,1,_r,cC),p.Mb=function(t){return Nz(),Ae(V(f(t,72),(ot(),Rm)))===Ae((Sg(),gx))},R(yn,"EndLabelSorter/lambda$3$Type",1560),N(1561,1,_r,pv),p.Mb=function(t){return Nz(),Ae(V(f(t,72),(ot(),Rm)))===Ae((Sg(),LD))},R(yn,"EndLabelSorter/lambda$4$Type",1561),N(1508,1,vi,uC),p.If=function(t,i){oMn(this,f(t,36))},p.b=0,p.c=0,R(yn,"FinalSplineBendpointsCalculator",1508),N(1509,1,{},lC),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"FinalSplineBendpointsCalculator/lambda$0$Type",1509),N(1510,1,{},my),p.Kb=function(t){return new Ft(null,new sT(new Tn(Cn(ao(f(t,9)).a.Jc(),new B))))},R(yn,"FinalSplineBendpointsCalculator/lambda$1$Type",1510),N(1511,1,_r,v6),p.Mb=function(t){return!Fa(f(t,18))},R(yn,"FinalSplineBendpointsCalculator/lambda$2$Type",1511),N(1512,1,_r,iO),p.Mb=function(t){return li(f(t,18),(Ge(),NE))},R(yn,"FinalSplineBendpointsCalculator/lambda$3$Type",1512),N(1513,1,xn,fet),p.Ad=function(t){wLn(this.a,f(t,132))},R(yn,"FinalSplineBendpointsCalculator/lambda$4$Type",1513),N(1514,1,xn,JM),p.Ad=function(t){vG(f(t,18).a)},R(yn,"FinalSplineBendpointsCalculator/lambda$5$Type",1514),N(783,1,vi,e2e),p.If=function(t,i){e8n(this,f(t,36),i)},R(yn,"GraphTransformer",783),N(501,23,{3:1,34:1,23:1,501:1},uSe);var yge,kq,WOt=er(yn,"GraphTransformer/Mode",501,ar,Ayn,v1n),KOt;N(1515,1,vi,Zc),p.If=function(t,i){P6n(f(t,36),i)},R(yn,"HierarchicalNodeResizingProcessor",1515),N(1516,1,vi,Bc),p.If=function(t,i){Ekn(f(t,36),i)},R(yn,"HierarchicalPortConstraintProcessor",1516),N(1517,1,Cr,ZM),p.Le=function(t,i){return F3n(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"HierarchicalPortConstraintProcessor/NodeComparator",1517),N(1518,1,vi,eg),p.If=function(t,i){fDn(f(t,36),i)},R(yn,"HierarchicalPortDummySizeProcessor",1518),N(1519,1,vi,I2),p.If=function(t,i){APn(this,f(t,36),i)},p.a=0,R(yn,"HierarchicalPortOrthogonalEdgeRouter",1519),N(1520,1,Cr,oO),p.Le=function(t,i){return Bmn(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"HierarchicalPortOrthogonalEdgeRouter/1",1520),N(1521,1,Cr,dC),p.Le=function(t,i){return T_n(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"HierarchicalPortOrthogonalEdgeRouter/2",1521),N(1522,1,vi,SU),p.If=function(t,i){uNn(f(t,36),i)},R(yn,"HierarchicalPortPositionProcessor",1522),N(1523,1,vi,cN),p.If=function(t,i){X9n(this,f(t,36))},p.a=0,p.c=0;var zee,Hee;R(yn,"HighDegreeNodeLayeringProcessor",1523),N(562,1,{562:1},aO),p.b=-1,p.d=-1,R(yn,"HighDegreeNodeLayeringProcessor/HighDegreeNodeInformation",562),N(1524,1,{},w6),p.Kb=function(t){return aH(),Fo(f(t,9))},p.Fb=function(t){return this===t},R(yn,"HighDegreeNodeLayeringProcessor/lambda$0$Type",1524),N(1525,1,{},dd),p.Kb=function(t){return aH(),ao(f(t,9))},p.Fb=function(t){return this===t},R(yn,"HighDegreeNodeLayeringProcessor/lambda$1$Type",1525),N(1531,1,vi,ek),p.If=function(t,i){rDn(this,f(t,36),i)},R(yn,"HyperedgeDummyMerger",1531),N(784,1,{},K_e),p.a=!1,p.b=!1,p.c=!1,R(yn,"HyperedgeDummyMerger/MergeState",784),N(1532,1,{},by),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"HyperedgeDummyMerger/lambda$0$Type",1532),N(1533,1,{},QM),p.Kb=function(t){return new Ft(null,new Bt(f(t,9).j,16))},R(yn,"HyperedgeDummyMerger/lambda$1$Type",1533),N(1534,1,xn,y6),p.Ad=function(t){f(t,11).p=-1},R(yn,"HyperedgeDummyMerger/lambda$2$Type",1534),N(1535,1,vi,Jm),p.If=function(t,i){nDn(f(t,36),i)},R(yn,"HypernodesProcessor",1535),N(1536,1,vi,sO),p.If=function(t,i){dDn(f(t,36),i)},R(yn,"InLayerConstraintProcessor",1536),N(1537,1,vi,e8),p.If=function(t,i){Zkn(f(t,36),i)},R(yn,"InnermostNodeMarginCalculator",1537),N(1538,1,vi,t8),p.If=function(t,i){u9n(this,f(t,36))},p.a=qo,p.b=qo,p.c=Gi,p.d=Gi;var ojn=R(yn,"InteractiveExternalPortPositioner",1538);N(1539,1,{},n8),p.Kb=function(t){return f(t,18).d.i},p.Fb=function(t){return this===t},R(yn,"InteractiveExternalPortPositioner/lambda$0$Type",1539),N(1540,1,{},het),p.Kb=function(t){return Umn(this.a,Oe(t))},p.Fb=function(t){return this===t},R(yn,"InteractiveExternalPortPositioner/lambda$1$Type",1540),N(1541,1,{},kU),p.Kb=function(t){return f(t,18).c.i},p.Fb=function(t){return this===t},R(yn,"InteractiveExternalPortPositioner/lambda$2$Type",1541),N(1542,1,{},pet),p.Kb=function(t){return zmn(this.a,Oe(t))},p.Fb=function(t){return this===t},R(yn,"InteractiveExternalPortPositioner/lambda$3$Type",1542),N(1543,1,{},get),p.Kb=function(t){return jbn(this.a,Oe(t))},p.Fb=function(t){return this===t},R(yn,"InteractiveExternalPortPositioner/lambda$4$Type",1543),N(1544,1,{},met),p.Kb=function(t){return $bn(this.a,Oe(t))},p.Fb=function(t){return this===t},R(yn,"InteractiveExternalPortPositioner/lambda$5$Type",1544),N(79,23,{3:1,34:1,23:1,79:1,196:1},ho),p.bg=function(){switch(this.g){case 15:return new SO;case 22:return new j2;case 48:return new _O;case 29:case 36:return new gv;case 33:return new aC;case 43:return new nO;case 1:return new gy;case 42:return new J_;case 57:return new e2e((BP(),kq));case 0:return new e2e((BP(),yge));case 2:return new f6;case 55:return new h6;case 34:return new sC;case 52:return new uC;case 56:return new Zc;case 13:return new Bc;case 39:return new eg;case 45:return new I2;case 41:return new SU;case 9:return new cN;case 50:return new bct;case 38:return new ek;case 44:return new Jm;case 28:return new sO;case 31:return new e8;case 3:return new t8;case 18:return new i8;case 30:return new tg;case 5:return new Xd;case 51:return new o8;case 35:return new Vl;case 37:return new S6;case 53:return new Pv;case 11:return new TU;case 7:return new QU;case 40:return new _6;case 46:return new xU;case 16:return new CU;case 10:return new dat;case 49:return new tk;case 21:return new k6;case 23:return new Sae((PS(),uF));case 8:return new s8;case 12:return new dO;case 4:return new T6;case 19:return new Xo;case 17:return new ok;case 54:return new OU;case 6:return new RU;case 25:return new frt;case 26:return new ts;case 47:return new f8;case 32:return new rut;case 14:return new PU;case 27:return new EC;case 20:return new En;case 24:return new Sae((PS(),Wte));default:throw ie(new dn(Zfe+(this.f!=null?this.f:""+this.g)))}};var k5e,T5e,x5e,C5e,A5e,I5e,O5e,N5e,R5e,P5e,L5e,vA,Gee,qee,D5e,M5e,j5e,$5e,F5e,B5e,U5e,z$,z5e,H5e,G5e,q5e,V5e,Ege,Vee,Wee,W5e,Kee,Yee,Xee,nD,JT,ZT,K5e,Jee,Zee,Y5e,Qee,ete,X5e,J5e,Z5e,Q5e,tte,Sge,Tq,nte,rte,ite,ote,ePe,tPe,nPe,rPe,ajn=er(yn,Qfe,79,ar,O2t,w1n),YOt;N(1545,1,vi,i8),p.If=function(t,i){a9n(f(t,36),i)},R(yn,"InvertedPortProcessor",1545),N(1546,1,vi,tg),p.If=function(t,i){hLn(f(t,36),i)},R(yn,"LabelAndNodeSizeProcessor",1546),N(1547,1,_r,_U),p.Mb=function(t){return f(t,9).k==(cn(),Yi)},R(yn,"LabelAndNodeSizeProcessor/lambda$0$Type",1547),N(1548,1,_r,Zm),p.Mb=function(t){return f(t,9).k==(cn(),bo)},R(yn,"LabelAndNodeSizeProcessor/lambda$1$Type",1548),N(1549,1,xn,Dut),p.Ad=function(t){agn(this.b,this.a,this.c,f(t,9))},p.a=!1,p.c=!1,R(yn,"LabelAndNodeSizeProcessor/lambda$2$Type",1549),N(1550,1,vi,Xd),p.If=function(t,i){$8n(f(t,36),i)};var XOt;R(yn,"LabelDummyInserter",1550),N(1551,1,km,r8),p.Lb=function(t){return Ae(V(f(t,72),(ot(),Rm)))===Ae((Sg(),PD))},p.Fb=function(t){return this===t},p.Mb=function(t){return Ae(V(f(t,72),(ot(),Rm)))===Ae((Sg(),PD))},R(yn,"LabelDummyInserter/1",1551),N(1552,1,vi,o8),p.If=function(t,i){T8n(f(t,36),i)},R(yn,"LabelDummyRemover",1552),N(1553,1,_r,n0),p.Mb=function(t){return lt(ut(V(f(t,72),(ot(),hme))))},R(yn,"LabelDummyRemover/lambda$0$Type",1553),N(1317,1,vi,Vl),p.If=function(t,i){w8n(this,f(t,36),i)},p.a=null;var _ge;R(yn,"LabelDummySwitcher",1317),N(291,1,{291:1},kSt),p.c=0,p.d=null,p.f=0,R(yn,"LabelDummySwitcher/LabelDummyInfo",291),N(1318,1,{},cO),p.Kb=function(t){return dR(),new Ft(null,new Bt(f(t,25).a,16))},R(yn,"LabelDummySwitcher/lambda$0$Type",1318),N(1319,1,_r,E6),p.Mb=function(t){return dR(),f(t,9).k==(cn(),Vc)},R(yn,"LabelDummySwitcher/lambda$1$Type",1319),N(1320,1,{},bet),p.Kb=function(t){return Cbn(this.a,f(t,9))},R(yn,"LabelDummySwitcher/lambda$2$Type",1320),N(1321,1,xn,vet),p.Ad=function(t){xwn(this.a,f(t,291))},R(yn,"LabelDummySwitcher/lambda$3$Type",1321),N(1322,1,Cr,al),p.Le=function(t,i){return nwn(f(t,291),f(i,291))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"LabelDummySwitcher/lambda$4$Type",1322),N(782,1,vi,gv),p.If=function(t,i){XSn(f(t,36),i)},R(yn,"LabelManagementProcessor",782),N(1554,1,vi,S6),p.If=function(t,i){W5n(f(t,36),i)},R(yn,"LabelSideSelector",1554),N(1562,1,vi,TU),p.If=function(t,i){xDn(f(t,36),i)},R(yn,"LayerConstraintPostprocessor",1562),N(1563,1,vi,QU),p.If=function(t,i){L4n(f(t,36),i)};var iPe;R(yn,"LayerConstraintPreprocessor",1563),N(364,23,{3:1,34:1,23:1,364:1},DY);var xq,ate,ste,kge,JOt=er(yn,"LayerConstraintPreprocessor/HiddenNodeConnections",364,ar,L2n,y1n),ZOt;N(1564,1,vi,_6),p.If=function(t,i){zMn(f(t,36),i)},R(yn,"LayerSizeAndGraphHeightCalculator",1564),N(1565,1,vi,xU),p.If=function(t,i){L6n(f(t,36),i)},R(yn,"LongEdgeJoiner",1565),N(1566,1,vi,CU),p.If=function(t,i){mMn(f(t,36),i)},R(yn,"LongEdgeSplitter",1566),N(1567,1,vi,dat),p.If=function(t,i){J8n(this,f(t,36),i)},p.e=0,p.f=0,p.j=0,p.k=0,p.n=0,p.o=0;var QOt,eNt;R(yn,"NodePromotion",1567),N(1568,1,Cr,AU),p.Le=function(t,i){return txn(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"NodePromotion/1",1568),N(1569,1,Cr,a8),p.Le=function(t,i){return nxn(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"NodePromotion/2",1569),N(1570,1,{},uO),p.Kb=function(t){return f(t,47),mX(),mn(),!0},p.Fb=function(t){return this===t},R(yn,"NodePromotion/lambda$0$Type",1570),N(1571,1,{},wet),p.Kb=function(t){return aEn(this.a,f(t,47))},p.Fb=function(t){return this===t},p.a=0,R(yn,"NodePromotion/lambda$1$Type",1571),N(1572,1,{},yet),p.Kb=function(t){return oEn(this.a,f(t,47))},p.Fb=function(t){return this===t},p.a=0,R(yn,"NodePromotion/lambda$2$Type",1572),N(1573,1,vi,tk),p.If=function(t,i){z9n(f(t,36),i)},R(yn,"NorthSouthPortPostprocessor",1573),N(1574,1,vi,k6),p.If=function(t,i){K9n(f(t,36),i)},R(yn,"NorthSouthPortPreprocessor",1574),N(1575,1,Cr,nk),p.Le=function(t,i){return ETn(f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"NorthSouthPortPreprocessor/lambda$0$Type",1575),N(1576,1,vi,s8),p.If=function(t,i){VLn(f(t,36),i)},R(yn,"PartitionMidprocessor",1576),N(1577,1,_r,lO),p.Mb=function(t){return li(f(t,9),(ot(),bD))},R(yn,"PartitionMidprocessor/lambda$0$Type",1577),N(1578,1,xn,Eet),p.Ad=function(t){uyn(this.a,f(t,9))},R(yn,"PartitionMidprocessor/lambda$1$Type",1578),N(1579,1,vi,dO),p.If=function(t,i){Z6n(f(t,36),i)},R(yn,"PartitionPostprocessor",1579),N(1580,1,vi,T6),p.If=function(t,i){y4n(f(t,36),i)},R(yn,"PartitionPreprocessor",1580),N(1581,1,_r,c8),p.Mb=function(t){return li(f(t,9),(ot(),bD))},R(yn,"PartitionPreprocessor/lambda$0$Type",1581),N(1582,1,{},fO),p.Kb=function(t){return new Ft(null,new sT(new Tn(Cn(ao(f(t,9)).a.Jc(),new B))))},R(yn,"PartitionPreprocessor/lambda$1$Type",1582),N(1583,1,_r,vy),p.Mb=function(t){return T3n(f(t,18))},R(yn,"PartitionPreprocessor/lambda$2$Type",1583),N(1584,1,xn,IU),p.Ad=function(t){LTn(f(t,18))},R(yn,"PartitionPreprocessor/lambda$3$Type",1584),N(1585,1,vi,Xo),p.If=function(t,i){xLn(f(t,36),i)};var oPe,tNt,nNt,rNt,aPe,sPe;R(yn,"PortListSorter",1585),N(1586,1,{},u8),p.Kb=function(t){return WP(),f(t,11).e},R(yn,"PortListSorter/lambda$0$Type",1586),N(1587,1,{},Yg),p.Kb=function(t){return WP(),f(t,11).g},R(yn,"PortListSorter/lambda$1$Type",1587),N(1588,1,Cr,hO),p.Le=function(t,i){return opt(f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"PortListSorter/lambda$2$Type",1588),N(1589,1,Cr,rk),p.Le=function(t,i){return KAn(f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"PortListSorter/lambda$3$Type",1589),N(1590,1,Cr,ik),p.Le=function(t,i){return XSt(f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"PortListSorter/lambda$4$Type",1590),N(1591,1,vi,ok),p.If=function(t,i){U4n(f(t,36),i)},R(yn,"PortSideProcessor",1591),N(1592,1,vi,OU),p.If=function(t,i){UPn(f(t,36),i)},R(yn,"ReversedEdgeRestorer",1592),N(1597,1,vi,frt),p.If=function(t,i){NAn(this,f(t,36),i)},R(yn,"SelfLoopPortRestorer",1597),N(1598,1,{},pO),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"SelfLoopPortRestorer/lambda$0$Type",1598),N(1599,1,_r,l8),p.Mb=function(t){return f(t,9).k==(cn(),Yi)},R(yn,"SelfLoopPortRestorer/lambda$1$Type",1599),N(1600,1,_r,d8),p.Mb=function(t){return li(f(t,9),(Ge(),f_))},R(yn,"SelfLoopPortRestorer/lambda$2$Type",1600),N(1601,1,{},x6),p.Kb=function(t){return f(V(f(t,9),(Ge(),f_)),333)},R(yn,"SelfLoopPortRestorer/lambda$3$Type",1601),N(1602,1,xn,xet),p.Ad=function(t){UNn(this.a,f(t,333))},R(yn,"SelfLoopPortRestorer/lambda$4$Type",1602),N(785,1,xn,mv),p.Ad=function(t){QNn(f(t,105))},R(yn,"SelfLoopPortRestorer/lambda$5$Type",785),N(1604,1,vi,f8),p.If=function(t,i){z3n(f(t,36),i)},R(yn,"SelfLoopPostProcessor",1604),N(1605,1,{},h8),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"SelfLoopPostProcessor/lambda$0$Type",1605),N(1606,1,_r,gO),p.Mb=function(t){return f(t,9).k==(cn(),Yi)},R(yn,"SelfLoopPostProcessor/lambda$1$Type",1606),N(1607,1,_r,p8),p.Mb=function(t){return li(f(t,9),(Ge(),f_))},R(yn,"SelfLoopPostProcessor/lambda$2$Type",1607),N(1608,1,xn,wy),p.Ad=function(t){VIn(f(t,9))},R(yn,"SelfLoopPostProcessor/lambda$3$Type",1608),N(1609,1,{},mO),p.Kb=function(t){return new Ft(null,new Bt(f(t,105).f,1))},R(yn,"SelfLoopPostProcessor/lambda$4$Type",1609),N(1610,1,xn,Cet),p.Ad=function(t){D2n(this.a,f(t,336))},R(yn,"SelfLoopPostProcessor/lambda$5$Type",1610),N(1611,1,_r,NU),p.Mb=function(t){return!!f(t,105).i},R(yn,"SelfLoopPostProcessor/lambda$6$Type",1611),N(1612,1,xn,Aet),p.Ad=function(t){upn(this.a,f(t,105))},R(yn,"SelfLoopPostProcessor/lambda$7$Type",1612),N(1593,1,vi,RU),p.If=function(t,i){w6n(f(t,36),i)},R(yn,"SelfLoopPreProcessor",1593),N(1594,1,{},Xg),p.Kb=function(t){return new Ft(null,new Bt(f(t,105).f,1))},R(yn,"SelfLoopPreProcessor/lambda$0$Type",1594),N(1595,1,{},r0),p.Kb=function(t){return f(t,336).a},R(yn,"SelfLoopPreProcessor/lambda$1$Type",1595),N(1596,1,xn,Qm),p.Ad=function(t){gmn(f(t,18))},R(yn,"SelfLoopPreProcessor/lambda$2$Type",1596),N(1613,1,vi,rut),p.If=function(t,i){kNn(this,f(t,36),i)},R(yn,"SelfLoopRouter",1613),N(1614,1,{},C6),p.Kb=function(t){return new Ft(null,new Bt(f(t,25).a,16))},R(yn,"SelfLoopRouter/lambda$0$Type",1614),N(1615,1,_r,g8),p.Mb=function(t){return f(t,9).k==(cn(),Yi)},R(yn,"SelfLoopRouter/lambda$1$Type",1615),N(1616,1,_r,m8),p.Mb=function(t){return li(f(t,9),(Ge(),f_))},R(yn,"SelfLoopRouter/lambda$2$Type",1616),N(1617,1,{},A6),p.Kb=function(t){return f(V(f(t,9),(Ge(),f_)),333)},R(yn,"SelfLoopRouter/lambda$3$Type",1617),N(1618,1,xn,qot),p.Ad=function(t){syn(this.a,this.b,f(t,333))},R(yn,"SelfLoopRouter/lambda$4$Type",1618),N(1619,1,vi,PU),p.If=function(t,i){L5n(f(t,36),i)},R(yn,"SemiInteractiveCrossMinProcessor",1619),N(1620,1,_r,ak),p.Mb=function(t){return f(t,9).k==(cn(),Yi)},R(yn,"SemiInteractiveCrossMinProcessor/lambda$0$Type",1620),N(1621,1,_r,yy),p.Mb=function(t){return vlt(f(t,9))._b((ot(),sx))},R(yn,"SemiInteractiveCrossMinProcessor/lambda$1$Type",1621),N(1622,1,Cr,Dt),p.Le=function(t,i){return Dkn(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(yn,"SemiInteractiveCrossMinProcessor/lambda$2$Type",1622),N(1623,1,{},Rn),p.Te=function(t,i){return cyn(f(t,9),f(i,9))},R(yn,"SemiInteractiveCrossMinProcessor/lambda$3$Type",1623),N(1625,1,vi,En),p.If=function(t,i){$Mn(f(t,36),i)},R(yn,"SortByInputModelProcessor",1625),N(1626,1,_r,ui),p.Mb=function(t){return f(t,11).g.c.length!=0},R(yn,"SortByInputModelProcessor/lambda$0$Type",1626),N(1627,1,xn,Iet),p.Ad=function(t){iRn(this.a,f(t,11))},R(yn,"SortByInputModelProcessor/lambda$1$Type",1627),N(1705,797,{},Tbt),p.bf=function(t){var i,a,u,l;switch(this.c=t,this.a.g){case 2:i=new et,Zi(Jr(new Ft(null,new Bt(this.c.a.b,16)),new Yb),new Xot(this,i)),wG(this,new Nd),Mu(i,new Ks),i.c.length=0,Zi(Jr(new Ft(null,new Bt(this.c.a.b,16)),new bO),new ket(i)),wG(this,new b8),Mu(i,new Jg),i.c.length=0,a=Rst(Bue(uT(new Ft(null,new Bt(this.c.a.b,16)),new Tet(this))),new vO),Zi(new Ft(null,new Bt(this.c.a.a,16)),new Wot(a,i)),wG(this,new sk),Mu(i,new Lo),i.c.length=0;break;case 3:u=new et,wG(this,new Ws),l=Rst(Bue(uT(new Ft(null,new Bt(this.c.a.b,16)),new _et(this))),new LU),Zi(Jr(new Ft(null,new Bt(this.c.a.b,16)),new fC),new Yot(l,u)),wG(this,new hC),Mu(u,new bc),u.c.length=0;break;default:throw ie(new Znt)}},p.b=0,R(co,"EdgeAwareScanlineConstraintCalculation",1705),N(1706,1,km,Ws),p.Lb=function(t){return me(f(t,60).g,152)},p.Fb=function(t){return this===t},p.Mb=function(t){return me(f(t,60).g,152)},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$0$Type",1706),N(1707,1,{},_et),p.We=function(t){return ORn(this.a,f(t,60))},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$1$Type",1707),N(1715,1,EQ,Vot),p.be=function(){Lj(this.a,this.b,-1)},p.b=0,R(co,"EdgeAwareScanlineConstraintCalculation/lambda$10$Type",1715),N(1717,1,km,Nd),p.Lb=function(t){return me(f(t,60).g,152)},p.Fb=function(t){return this===t},p.Mb=function(t){return me(f(t,60).g,152)},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$11$Type",1717),N(1718,1,xn,Ks),p.Ad=function(t){f(t,372).be()},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$12$Type",1718),N(1719,1,_r,bO),p.Mb=function(t){return me(f(t,60).g,9)},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$13$Type",1719),N(1721,1,xn,ket),p.Ad=function(t){bCn(this.a,f(t,60))},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$14$Type",1721),N(1720,1,EQ,Qot),p.be=function(){Lj(this.b,this.a,-1)},p.a=0,R(co,"EdgeAwareScanlineConstraintCalculation/lambda$15$Type",1720),N(1722,1,km,b8),p.Lb=function(t){return me(f(t,60).g,9)},p.Fb=function(t){return this===t},p.Mb=function(t){return me(f(t,60).g,9)},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$16$Type",1722),N(1723,1,xn,Jg),p.Ad=function(t){f(t,372).be()},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$17$Type",1723),N(1724,1,{},Tet),p.We=function(t){return NRn(this.a,f(t,60))},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$18$Type",1724),N(1725,1,{},vO),p.Ue=function(){return 0},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$19$Type",1725),N(1708,1,{},LU),p.Ue=function(){return 0},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$2$Type",1708),N(1727,1,xn,Wot),p.Ad=function(t){Wvn(this.a,this.b,f(t,317))},p.a=0,R(co,"EdgeAwareScanlineConstraintCalculation/lambda$20$Type",1727),N(1726,1,EQ,Kot),p.be=function(){n2t(this.a,this.b,-1)},p.b=0,R(co,"EdgeAwareScanlineConstraintCalculation/lambda$21$Type",1726),N(1728,1,km,sk),p.Lb=function(t){return f(t,60),!0},p.Fb=function(t){return this===t},p.Mb=function(t){return f(t,60),!0},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$22$Type",1728),N(1729,1,xn,Lo),p.Ad=function(t){f(t,372).be()},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$23$Type",1729),N(1709,1,_r,fC),p.Mb=function(t){return me(f(t,60).g,9)},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$3$Type",1709),N(1711,1,xn,Yot),p.Ad=function(t){Vvn(this.a,this.b,f(t,60))},p.a=0,R(co,"EdgeAwareScanlineConstraintCalculation/lambda$4$Type",1711),N(1710,1,EQ,eat),p.be=function(){Lj(this.b,this.a,-1)},p.a=0,R(co,"EdgeAwareScanlineConstraintCalculation/lambda$5$Type",1710),N(1712,1,km,hC),p.Lb=function(t){return f(t,60),!0},p.Fb=function(t){return this===t},p.Mb=function(t){return f(t,60),!0},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$6$Type",1712),N(1713,1,xn,bc),p.Ad=function(t){f(t,372).be()},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$7$Type",1713),N(1714,1,_r,Yb),p.Mb=function(t){return me(f(t,60).g,152)},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$8$Type",1714),N(1716,1,xn,Xot),p.Ad=function(t){ikn(this.a,this.b,f(t,60))},R(co,"EdgeAwareScanlineConstraintCalculation/lambda$9$Type",1716),N(1526,1,vi,bct),p.If=function(t,i){EMn(this,f(t,36),i)};var iNt;R(co,"HorizontalGraphCompactor",1526),N(1527,1,{},Oet),p.df=function(t,i){var a,u,l;return gxe(t,i)||(a=D3(t),u=D3(i),a&&a.k==(cn(),bo)||u&&u.k==(cn(),bo))?0:(l=f(V(this.a.a,(Ge(),CA)),313),qmn(l,a?a.k:(cn(),uo),u?u.k:(cn(),uo)))},p.ef=function(t,i){var a,u,l;return gxe(t,i)?1:(a=D3(t),u=D3(i),l=f(V(this.a.a,(Ge(),CA)),313),c_e(l,a?a.k:(cn(),uo),u?u.k:(cn(),uo)))},R(co,"HorizontalGraphCompactor/1",1527),N(1528,1,{},I6),p.cf=function(t,i){return p7(),t.a.i==0},R(co,"HorizontalGraphCompactor/lambda$0$Type",1528),N(1529,1,{},Net),p.cf=function(t,i){return dyn(this.a,t,i)},R(co,"HorizontalGraphCompactor/lambda$1$Type",1529),N(1672,1,{},nmt);var oNt,aNt;R(co,"LGraphToCGraphTransformer",1672),N(1680,1,_r,bp),p.Mb=function(t){return t!=null},R(co,"LGraphToCGraphTransformer/0methodref$nonNull$Type",1680),N(1673,1,{},eb),p.Kb=function(t){return tf(),Qs(V(f(f(t,60).g,9),(Ge(),oi)))},R(co,"LGraphToCGraphTransformer/lambda$0$Type",1673),N(1674,1,{},ck),p.Kb=function(t){return tf(),v0t(f(f(t,60).g,152))},R(co,"LGraphToCGraphTransformer/lambda$1$Type",1674),N(1683,1,_r,O6),p.Mb=function(t){return tf(),me(f(t,60).g,9)},R(co,"LGraphToCGraphTransformer/lambda$10$Type",1683),N(1684,1,xn,O2),p.Ad=function(t){ryn(f(t,60))},R(co,"LGraphToCGraphTransformer/lambda$11$Type",1684),N(1685,1,_r,Xb),p.Mb=function(t){return tf(),me(f(t,60).g,152)},R(co,"LGraphToCGraphTransformer/lambda$12$Type",1685),N(1689,1,xn,uk),p.Ad=function(t){jxn(f(t,60))},R(co,"LGraphToCGraphTransformer/lambda$13$Type",1689),N(1686,1,xn,Ret),p.Ad=function(t){zgn(this.a,f(t,8))},p.a=0,R(co,"LGraphToCGraphTransformer/lambda$14$Type",1686),N(1687,1,xn,Pet),p.Ad=function(t){Ggn(this.a,f(t,119))},p.a=0,R(co,"LGraphToCGraphTransformer/lambda$15$Type",1687),N(1688,1,xn,Let),p.Ad=function(t){Hgn(this.a,f(t,8))},p.a=0,R(co,"LGraphToCGraphTransformer/lambda$16$Type",1688),N(1690,1,{},N2),p.Kb=function(t){return tf(),new Ft(null,new sT(new Tn(Cn(ao(f(t,9)).a.Jc(),new B))))},R(co,"LGraphToCGraphTransformer/lambda$17$Type",1690),N(1691,1,_r,pC),p.Mb=function(t){return tf(),Fa(f(t,18))},R(co,"LGraphToCGraphTransformer/lambda$18$Type",1691),N(1692,1,xn,Det),p.Ad=function(t){D_n(this.a,f(t,18))},R(co,"LGraphToCGraphTransformer/lambda$19$Type",1692),N(1676,1,xn,Met),p.Ad=function(t){u2n(this.a,f(t,152))},R(co,"LGraphToCGraphTransformer/lambda$2$Type",1676),N(1693,1,{},gC),p.Kb=function(t){return tf(),new Ft(null,new Bt(f(t,25).a,16))},R(co,"LGraphToCGraphTransformer/lambda$20$Type",1693),N(1694,1,{},R2),p.Kb=function(t){return tf(),new Ft(null,new sT(new Tn(Cn(ao(f(t,9)).a.Jc(),new B))))},R(co,"LGraphToCGraphTransformer/lambda$21$Type",1694),N(1695,1,{},wO),p.Kb=function(t){return tf(),f(V(f(t,18),(Ge(),NE)),15)},R(co,"LGraphToCGraphTransformer/lambda$22$Type",1695),N(1696,1,_r,ng),p.Mb=function(t){return Vmn(f(t,15))},R(co,"LGraphToCGraphTransformer/lambda$23$Type",1696),N(1697,1,xn,jet),p.Ad=function(t){RRn(this.a,f(t,15))},R(co,"LGraphToCGraphTransformer/lambda$24$Type",1697),N(1698,1,{},P2),p.Kb=function(t){return tf(),new Ft(null,new sT(new Tn(Cn(ao(f(t,9)).a.Jc(),new B))))},R(co,"LGraphToCGraphTransformer/lambda$25$Type",1698),N(1699,1,_r,vp),p.Mb=function(t){return tf(),Fa(f(t,18))},R(co,"LGraphToCGraphTransformer/lambda$26$Type",1699),N(1701,1,xn,$et),p.Ad=function(t){xkn(this.a,f(t,18))},R(co,"LGraphToCGraphTransformer/lambda$27$Type",1701),N(1700,1,xn,Fet),p.Ad=function(t){Lpn(this.a,f(t,72))},p.a=0,R(co,"LGraphToCGraphTransformer/lambda$28$Type",1700),N(1675,1,xn,Jot),p.Ad=function(t){sSn(this.a,this.b,f(t,152))},R(co,"LGraphToCGraphTransformer/lambda$3$Type",1675),N(1677,1,{},mC),p.Kb=function(t){return tf(),new Ft(null,new Bt(f(t,25).a,16))},R(co,"LGraphToCGraphTransformer/lambda$4$Type",1677),N(1678,1,{},yO),p.Kb=function(t){return tf(),new Ft(null,new sT(new Tn(Cn(ao(f(t,9)).a.Jc(),new B))))},R(co,"LGraphToCGraphTransformer/lambda$5$Type",1678),N(1679,1,{},L2),p.Kb=function(t){return tf(),f(V(f(t,18),(Ge(),NE)),15)},R(co,"LGraphToCGraphTransformer/lambda$6$Type",1679),N(1681,1,xn,Bet),p.Ad=function(t){HRn(this.a,f(t,15))},R(co,"LGraphToCGraphTransformer/lambda$8$Type",1681),N(1682,1,xn,Zot),p.Ad=function(t){fmn(this.a,this.b,f(t,152))},R(co,"LGraphToCGraphTransformer/lambda$9$Type",1682),N(1671,1,{},N6),p.af=function(t){var i,a,u,l,g;for(this.a=t,this.d=new fae,this.c=Pe(q6e,Jt,123,this.a.a.a.c.length,0,1),this.b=0,a=new ne(this.a.a.a);a.a<a.c.c.length;)i=f(ee(a),317),i.d=this.b,g=mS(Oz(new qy,i),this.d),this.c[this.b]=g,++this.b;for(L8n(this),N9n(this),X6n(this),jde(Gse(this.d),new CN),l=new ne(this.a.a.b);l.a<l.c.c.length;)u=f(ee(l),60),u.d.c=this.c[u.a.d].e+u.b.a},p.b=0,R(co,"NetworkSimplexCompaction",1671),N(152,1,{34:1,152:1},Xj),p.Dd=function(t){return z_n(this,f(t,152))},p.Ib=function(){return v0t(this)},R(co,"VerticalSegment",152),N(820,1,{},aAe),p.c=0,p.e=0,p.i=0,R(h$,"BetweenLayerEdgeTwoNodeCrossingsCounter",820),N(662,1,{662:1},Ubt),p.Ib=function(){return"AdjacencyList [node="+this.d+", adjacencies= "+this.a+"]"},p.b=0,p.c=0,p.f=0,R(h$,"BetweenLayerEdgeTwoNodeCrossingsCounter/AdjacencyList",662),N(292,1,{34:1,292:1},Tct),p.Dd=function(t){return yvn(this,f(t,292))},p.Ib=function(){return"Adjacency [position="+this.c+", cardinality="+this.a+", currentCardinality="+this.b+"]"},p.a=0,p.b=0,p.c=0,R(h$,"BetweenLayerEdgeTwoNodeCrossingsCounter/AdjacencyList/Adjacency",292),N(1972,1,{},Ayt),p.b=0,p.e=!1,R(h$,"CrossingMatrixFiller",1972);var sNt=Bi(xb,"IInitializable");N(1813,1,rq,nat),p.eg=function(t,i,a,u,l,g){},p.gg=function(t,i,a){},p.cg=function(){return this.c!=(PS(),uF)},p.dg=function(){this.e=Pe(pr,Pr,30,this.d,15,1)},p.fg=function(t,i){i[t][0].c.p=t},p.hg=function(t,i,a,u){++this.d},p.ig=function(){return!0},p.jg=function(t,i,a,u){return r0t(this,t,i,a),_Sn(this,i)},p.kg=function(t,i){var a;return a=Fpn(i,t.length),r0t(this,t,a,i),_1t(this,a)},p.d=0,R(h$,"GreedySwitchHeuristic",1813),N(1975,1,{},blt),p.b=0,p.d=0,R(h$,"NorthSouthEdgeNeighbouringNodeCrossingsCounter",1975),N(1962,1,{},uSt),p.a=!1,R(h$,"SwitchDecider",1962),N(105,1,{105:1},Fyt),p.a=null,p.c=null,p.i=null,R(jR,"SelfHyperLoop",105),N(1959,1,{},vvt),p.c=0,p.e=0,R(jR,"SelfHyperLoopLabels",1959),N(410,23,{3:1,34:1,23:1,410:1},MY);var VR,rD,iD,Tge,cNt=er(jR,"SelfHyperLoopLabels/Alignment",410,ar,j2n,_1n),uNt;N(336,1,{336:1},Upt),R(jR,"SelfLoopEdge",336),N(333,1,{333:1},bvt),p.a=!1,R(jR,"SelfLoopHolder",333),N(1736,1,_r,D2),p.Mb=function(t){return Fa(f(t,18))},R(jR,"SelfLoopHolder/lambda$0$Type",1736),N(112,1,{112:1},xvt),p.a=!1,p.c=!1,R(jR,"SelfLoopPort",112),N(1801,1,_r,R6),p.Mb=function(t){return Fa(f(t,18))},R(jR,"SelfLoopPort/lambda$0$Type",1801),N(368,23,{3:1,34:1,23:1,368:1},Dz);var cte,Cq,ute,lte,dte,lNt=er(jR,"SelfLoopType",368,ar,CSn,E1n),dNt;N(1744,1,{},Fy);var fNt,hNt,pNt,gNt;R(gf,"PortRestorer",1744),N(365,23,{3:1,34:1,23:1,365:1},qae);var s_,_w,c_,xge=er(gf,"PortRestorer/PortSideArea",365,ar,NEn,S1n),mNt;N(1745,1,{},cl),p.Kb=function(t){return cw(),f(t,15).Mc()},R(gf,"PortRestorer/lambda$0$Type",1745),N(1746,1,xn,vs),p.Ad=function(t){cw(),f(t,112).c=!1},R(gf,"PortRestorer/lambda$1$Type",1746),N(1755,1,_r,Jb),p.Mb=function(t){return cw(),f(t,11).j==(tt(),pn)},R(gf,"PortRestorer/lambda$10$Type",1755),N(1756,1,{},rg),p.Kb=function(t){return cw(),f(t,112).d},R(gf,"PortRestorer/lambda$11$Type",1756),N(1757,1,xn,Uet),p.Ad=function(t){Dpn(this.a,f(t,11))},R(gf,"PortRestorer/lambda$12$Type",1757),N(1747,1,xn,zet),p.Ad=function(t){tbn(this.a,f(t,105))},R(gf,"PortRestorer/lambda$2$Type",1747),N(1748,1,Cr,dk),p.Le=function(t,i){return hkn(f(t,112),f(i,112))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(gf,"PortRestorer/lambda$3$Type",1748),N(1749,1,_r,bv),p.Mb=function(t){return cw(),f(t,112).c},R(gf,"PortRestorer/lambda$4$Type",1749),N(1750,1,_r,Ey),p.Mb=function(t){return oTn(f(t,11))},R(gf,"PortRestorer/lambda$5$Type",1750),N(1751,1,_r,lk),p.Mb=function(t){return cw(),f(t,11).j==(tt(),hn)},R(gf,"PortRestorer/lambda$6$Type",1751),N(1752,1,_r,EO),p.Mb=function(t){return cw(),f(t,11).j==(tt(),_n)},R(gf,"PortRestorer/lambda$7$Type",1752),N(1753,1,_r,bC),p.Mb=function(t){return M2n(f(t,11))},R(gf,"PortRestorer/lambda$8$Type",1753),N(1754,1,_r,tb),p.Mb=function(t){return cw(),f(t,11).j==(tt(),Bn)},R(gf,"PortRestorer/lambda$9$Type",1754),N(272,23,{3:1,34:1,23:1,272:1},eR);var Cge,Age,Ige,Oge,Nge,Rge,Pge,Lge,cPe=er(gf,"PortSideAssigner/Target",272,ar,K_n,k1n),bNt;N(1737,1,{},nb),p.Kb=function(t){return Jr(new Ft(null,new Bt(f(t,105).j,16)),new Rd)},R(gf,"PortSideAssigner/lambda$1$Type",1737),N(1738,1,{},qu),p.Kb=function(t){return f(t,112).d},R(gf,"PortSideAssigner/lambda$2$Type",1738),N(1739,1,xn,sl),p.Ad=function(t){No(f(t,11),(tt(),hn))},R(gf,"PortSideAssigner/lambda$3$Type",1739),N(1740,1,{},vv),p.Kb=function(t){return f(t,112).d},R(gf,"PortSideAssigner/lambda$4$Type",1740),N(1741,1,xn,Het),p.Ad=function(t){Yhn(this.a,f(t,11))},R(gf,"PortSideAssigner/lambda$5$Type",1741),N(1742,1,Cr,P6),p.Le=function(t,i){return $wn(f(t,105),f(i,105))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(gf,"PortSideAssigner/lambda$6$Type",1742),N(1743,1,Cr,vC),p.Le=function(t,i){return gvn(f(t,112),f(i,112))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(gf,"PortSideAssigner/lambda$7$Type",1743),N(800,1,_r,Rd),p.Mb=function(t){return f(t,112).c},R(gf,"PortSideAssigner/lambda$8$Type",800),N(2054,1,{}),R(mw,"AbstractSelfLoopRouter",2054),N(1762,1,Cr,rb),p.Le=function(t,i){return Kbn(f(t,105),f(i,105))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(mw,MTt,1762),N(1763,1,Cr,v8),p.Le=function(t,i){return Ybn(f(t,105),f(i,105))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(mw,jTt,1763),N(1802,2054,{},w8),p.lg=function(t,i,a){return a},R(mw,"OrthogonalSelfLoopRouter",1802),N(1804,1,xn,tat),p.Ad=function(t){b3e(this.b,this.a,f(t,8))},R(mw,"OrthogonalSelfLoopRouter/lambda$0$Type",1804),N(1803,1802,{},wC),p.lg=function(t,i,a){var u,l;return u=t.c.d,pP(a,0,ii($a(u.n),u.a)),l=t.d.d,Mr(a,ii($a(l.n),l.a)),ULn(a)},R(mw,"PolylineSelfLoopRouter",1803),N(1758,1,{},L5),p.a=null;var wA;R(mw,"RoutingDirector",1758),N(1759,1,Cr,yC),p.Le=function(t,i){return lvn(f(t,112),f(i,112))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(mw,"RoutingDirector/lambda$0$Type",1759),N(1760,1,{},y8),p.Kb=function(t){return kY(),f(t,105).j},R(mw,"RoutingDirector/lambda$1$Type",1760),N(1761,1,xn,Sy),p.Ad=function(t){kY(),f(t,15).gd(wA)},R(mw,"RoutingDirector/lambda$2$Type",1761),N(1764,1,{},M2),R(mw,"RoutingSlotAssigner",1764),N(1765,1,_r,Get),p.Mb=function(t){return kgn(this.a,f(t,105))},R(mw,"RoutingSlotAssigner/lambda$0$Type",1765),N(1766,1,Cr,qet),p.Le=function(t,i){return Dvn(this.a,f(t,105),f(i,105))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(mw,"RoutingSlotAssigner/lambda$1$Type",1766),N(1805,1802,{},_y),p.lg=function(t,i,a){var u,l,g,m;return u=ke(Oe(TZ(t.b.g.b,(ot(),OA)))),m=new Act(ce(se(Zo,1),Ye,8,0,[(g=t.c.d,ii(new ja(g.n),g.a))])),k6n(t,i,a,m,u),Mr(m,(l=t.d.d,ii(new ja(l.n),l.a))),Vvt(new rOe(m))},R(mw,"SplineSelfLoopRouter",1805),N(510,1,Cr,Zxe,idt),p.Le=function(t,i){return LG(this,f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},p.a=!1,R(PQ,"ModelOrderNodeComparator",510),N(1767,1,_r,De),p.Mb=function(t){return f(t,11).e.c.length!=0},R(PQ,"ModelOrderNodeComparator/lambda$0$Type",1767),N(568,1,_r,Lh),p.Mb=function(t){return f(t,11).e.c.length!=0},R(PQ,"ModelOrderNodeComparator/lambda$1$Type",568),N(569,1,_r,wp),p.Mb=function(t){return f(t,11).g.c.length!=0},R(PQ,"ModelOrderNodeComparator/lambda$2$Type",569),N(651,1,Cr,X1t,ITe),p.Le=function(t,i){return qlt(this,t,i)},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},p.b=!1,R(PQ,"ModelOrderPortComparator",651),N(1603,1,vi,ts),p.If=function(t,i){e9n(f(t,36))},R("org.eclipse.elk.alg.layered.intermediate.unzipping","AlternatingLayerUnzipper",1603),N(795,1,{},fk),p.mg=function(t,i){var a,u,l,g;for(l=_yt(i),a=new et,g=i.f/l,u=1;u<l;++u)Qe(a,rt(mr(Nc(h.Math.round(u*g)))));return a},p.ng=function(){return!1},R(xm,"ARDCutIndexHeuristic",795),N(1484,1,vi,SO),p.If=function(t,i){aLn(f(t,36),i)},R(xm,"BreakingPointInserter",1484),N(314,1,{314:1},PTe),p.Ib=function(){var t;return t=new $v,t.a+="BPInfo[",t.a+=`
	start=`,vu(t,this.i),t.a+=`
	end=`,vu(t,this.a),t.a+=`
	nodeStartEdge=`,vu(t,this.e),t.a+=`
	startEndEdge=`,vu(t,this.j),t.a+=`
	originalEdge=`,vu(t,this.f),t.a+=`
	startInLayerDummy=`,vu(t,this.k),t.a+=`
	startInLayerEdge=`,vu(t,this.n),t.a+=`
	endInLayerDummy=`,vu(t,this.b),t.a+=`
	endInLayerEdge=`,vu(t,this.c),t.a},R(xm,"BreakingPointInserter/BPInfo",314),N(645,1,{645:1},ttt),p.a=!1,p.b=0,p.c=0,R(xm,"BreakingPointInserter/Cut",645),N(1485,1,vi,j2),p.If=function(t,i){A6n(f(t,36),i)},R(xm,"BreakingPointProcessor",1485),N(1486,1,_r,ig),p.Mb=function(t){return Omt(f(t,9))},R(xm,"BreakingPointProcessor/0methodref$isEnd$Type",1486),N(1487,1,_r,Hl),p.Mb=function(t){return Nmt(f(t,9))},R(xm,"BreakingPointProcessor/1methodref$isStart$Type",1487),N(1488,1,vi,_O),p.If=function(t,i){K6n(this,f(t,36),i)},R(xm,"BreakingPointRemover",1488),N(1489,1,xn,E8),p.Ad=function(t){f(t,132).k=!0},R(xm,"BreakingPointRemover/lambda$0$Type",1489),N(791,1,{},nIe),p.b=0,p.e=0,p.f=0,p.j=0,R(xm,"GraphStats",791),N(792,1,{},S8),p.Te=function(t,i){return h.Math.max(ke(Oe(t)),ke(Oe(i)))},R(xm,"GraphStats/0methodref$max$Type",792),N(793,1,{},_8),p.Te=function(t,i){return h.Math.max(ke(Oe(t)),ke(Oe(i)))},R(xm,"GraphStats/2methodref$max$Type",793),N(1668,1,{},DU),p.Te=function(t,i){return tvn(Oe(t),Oe(i))},R(xm,"GraphStats/lambda$1$Type",1668),N(1669,1,{},Vet),p.Kb=function(t){return Svt(this.a,f(t,25))},R(xm,"GraphStats/lambda$2$Type",1669),N(1670,1,{},Wet),p.Kb=function(t){return k2t(this.a,f(t,25))},R(xm,"GraphStats/lambda$6$Type",1670),N(794,1,{},Qc),p.mg=function(t,i){var a;return a=f(V(t,(ot(),cDe)),15),a||(Ht(),Ht(),Ga)},p.ng=function(){return!1},R(xm,"ICutIndexCalculator/ManualCutIndexCalculator",794),N(796,1,{},kO),p.mg=function(t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le;for(Le=(i.n==null&&Jvt(i),i.n),S=(i.d==null&&Jvt(i),i.d),Ee=Pe(sa,_s,30,Le.length,15,1),Ee[0]=Le[0],pe=Le[0],T=1;T<Le.length;T++)Ee[T]=Ee[T-1]+Le[T],pe+=Le[T];for(l=_yt(i)-1,m=f(V(t,(ot(),uDe)),17).a,u=qo,a=new et,F=h.Math.max(0,l-m);F<=h.Math.min(i.f-1,l+m);F++){if(J=pe/(F+1),ue=0,A=1,g=new et,ve=qo,L=0,v=0,W=S[0],F==0)ve=pe,v=(i.g==null&&(i.g=Pbt(i,new _8)),ke(i.g));else{for(;A<i.f;)Ee[A-1]-ue>=J&&(Qe(g,rt(A)),ve=h.Math.max(ve,Ee[A-1]-L),v+=W,ue+=Ee[A-1]-ue,L=Ee[A-1],W=S[A]),W=h.Math.max(W,S[A]),++A;v+=W}z=h.Math.min(1/ve,1/i.b/v),z>u&&(u=z,a=g)}return a},p.ng=function(){return!1},R(xm,"MSDCutIndexHeuristic",796),N(1624,1,vi,EC),p.If=function(t,i){ADn(f(t,36),i)},R(xm,"SingleEdgeGraphWrapper",1624),N(231,23,{3:1,34:1,23:1,231:1},E7);var yA,oD,aD,QT,H$,EA,sD=er(Pc,"CenterEdgeLabelPlacementStrategy",231,ar,u_n,T1n),vNt;N(419,23,{3:1,34:1,23:1,419:1},dSe);var uPe,Dge,lPe=er(Pc,"ConstraintCalculationStrategy",419,ar,Pyn,x1n),wNt;N(323,23,{3:1,34:1,23:1,323:1,188:1,196:1},Vae),p.bg=function(){return sEt(this)},p.og=function(){return sEt(this)};var Aq,G$,dPe,fPe=er(Pc,"CrossingMinimizationStrategy",323,ar,wEn,C1n),yNt;N(347,23,{3:1,34:1,23:1,347:1},Wae);var hPe,Mge,fte,pPe=er(Pc,"CuttingStrategy",347,ar,yEn,A1n),ENt;N(344,23,{3:1,34:1,23:1,344:1,188:1,196:1},Mz),p.bg=function(){return QEt(this)},p.og=function(){return QEt(this)};var gPe,jge,cD,$ge,uD,mPe=er(Pc,"CycleBreakingStrategy",344,ar,RSn,I1n),SNt;N(416,23,{3:1,34:1,23:1,416:1},fSe);var hte,bPe,vPe=er(Pc,"DirectionCongruency",416,ar,Lyn,O1n),_Nt;N(446,23,{3:1,34:1,23:1,446:1},Kae);var lD,Fge,SA,kNt=er(Pc,"EdgeConstraint",446,ar,EEn,N1n),TNt;N(281,23,{3:1,34:1,23:1,281:1},S7);var Bge,Uge,zge,Hge,pte,Gge,wPe=er(Pc,"EdgeLabelSideSelection",281,ar,c_n,R1n),xNt;N(474,23,{3:1,34:1,23:1,474:1},hSe);var gte,yPe,EPe=er(Pc,"EdgeStraighteningStrategy",474,ar,Myn,P1n),CNt;N(279,23,{3:1,34:1,23:1,279:1},_7);var qge,SPe,_Pe,mte,kPe,TPe,xPe=er(Pc,"FixedAlignment",279,ar,a_n,L1n),ANt;N(280,23,{3:1,34:1,23:1,280:1},k7);var CPe,APe,IPe,OPe,q$,NPe,RPe=er(Pc,"GraphCompactionStrategy",280,ar,s_n,D1n),INt;N(259,23,{3:1,34:1,23:1,259:1},Kk);var dD,bte,fD,Vf,V$,vte,hD,_A,wte,W$,Vge=er(Pc,"GraphProperties",259,ar,Ukn,M1n),ONt;N(298,23,{3:1,34:1,23:1,298:1},Yae);var Iq,Wge,Kge,Yge=er(Pc,"GreedySwitchType",298,ar,SEn,j1n),NNt;N(312,23,{3:1,34:1,23:1,312:1},Xae);var WR,Oq,kA,RNt=er(Pc,"InLayerConstraint",312,ar,_En,$1n),PNt;N(417,23,{3:1,34:1,23:1,417:1},pSe);var Xge,PPe,LPe=er(Pc,"InteractiveReferencePoint",417,ar,Dyn,F1n),LNt,DPe,KR,u_,yte,MPe,jPe,Ete,$Pe,Nq,Ste,K$,YR,R1,Jge,_te,Tc,FPe,kw,Cu,Zge,Qge,Rq,OE,l_,XR,BPe,JR,Pq,ex,$p,yh,eme,TA,hi,oi,UPe,zPe,HPe,GPe,qPe,tme,kte,ed,d_,nme,ZR,Y$,U0,xA,f_,CA,AA,pD,NE,VPe,rme,ime,QR;N(171,23,{3:1,34:1,23:1,171:1},jz);var X$,Tw,J$,tx,Lq,WPe=er(Pc,"LayerConstraint",171,ar,PSn,B1n),DNt;N(420,23,{3:1,34:1,23:1,420:1},gSe);var ome,ame,KPe=er(Pc,"LayerUnzippingStrategy",420,ar,$yn,U1n),MNt;N(838,1,Ng,Jd),p.tf=function(t){Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,uNe),""),"Direction Congruency"),"Specifies how drawings of the same graph with different layout directions compare to each other: either a natural reading direction is preserved or the drawings are rotated versions of each other."),iLe),(bE(),Li)),vPe),At((wm(),en))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,lNe),""),"Feedback Edges"),"Whether feedback edges should be highlighted by routing around the nodes."),(mn(),!1)),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,LQ),""),"Interactive Reference Point"),"Determines which point of a node is considered by interactive layout phases."),lLe),Li),LPe),At(en)))),po(t,LQ,nhe,ORt),po(t,LQ,p$,IRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,dNe),""),"Merge Edges"),"Edges that have no ports are merged so they touch the connected nodes at the same points. When this option is disabled, one port is created for each edge directly connected to a node. When it is enabled, all such incoming edges share an input port, and all outgoing edges share an output port."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,fNe),""),"Merge Hierarchy-Crossing Edges"),"If hierarchical layout is active, hierarchy-crossing edges use as few hierarchical ports as possible. They are broken by the algorithm, with hierarchical ports inserted as required. Usually, one such port is created for each edge at each hierarchy crossing point. With this option set to true, we try to create as few hierarchical ports as possible in the process. In particular, all edges that form a hyperedge can share a port."),!0),Po),to),At(en)))),Tt(t,new bt(zpn(_t(St(kt(Lt(vt(Et(wt(yt(new mt,hNe),""),"Allow Non-Flow Ports To Switch Sides"),"Specifies whether non-flow ports may switch sides if their node's port constraints are either FIXED_SIDE or FIXED_ORDER. A non-flow port is a port on a side that is not part of the currently configured layout flow. For instance, given a left-to-right layout direction, north and south ports would be considered non-flow ports. Further note that the underlying criterium whether to switch sides or not solely relies on the minimization of edge crossings. Hence, edge length and other aesthetics criteria are not addressed."),!1),Po),to),At(FE)),ce(se(ft,1),Ye,2,6,["org.eclipse.elk.layered.northOrSouthPort"])))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,pNe),""),"Port Sorting Strategy"),"Only relevant for nodes with FIXED_SIDE port constraints. Determines the way a node's ports are distributed on the sides of a node if their order is not prescribed. The option is set on parent nodes."),ELe),Li),ODe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,gNe),""),"Thoroughness"),"How much effort should be spent to produce a nice layout."),rt(7)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,mNe),""),"Add Unnecessary Bendpoints"),"Adds bend points even if an edge does not change direction. If true, each long edge dummy will contribute a bend point to its edges and hierarchy-crossing edges will always get a bend point where they cross hierarchy boundaries. By default, bend points are only added where an edge changes direction."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,bNe),""),"Generate Position and Layer IDs"),"If enabled position id and layer id are generated, which are usually only used internally when setting the interactiveLayout option. This option should be specified on the root node."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,nhe),"cycleBreaking"),"Cycle Breaking Strategy"),"Strategy for cycle breaking. Cycle breaking looks for cycles in the graph and determines which edges to reverse to break the cycles. Reversed edges will end up pointing to the opposite direction of regular edges (that is, reversed edges will point left if edges usually point right)."),rLe),Li),mPe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,iq),Ahe),"Node Layering Strategy"),"Strategy for node layering."),hLe),Li),vDe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,vNe),Ahe),"Layer Constraint"),"Determines a constraint on the placement of the node regarding the layering."),dLe),Li),WPe),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,wNe),Ahe),"Layer Choice Constraint"),"Allows to set a constraint regarding the layer placement of a node. Let i be the value of teh constraint. Assumed the drawing has n layers and i < n. If set to i, it expresses that the node should be placed in i-th layer. Should i>=n be true then the node is placed in the last layer of the drawing. Note that this option is not part of any of ELK Layered's default configurations but is only evaluated as part of the `InteractiveLayeredGraphVisitor`, which must be applied manually or used via the `DiagramLayoutEngine."),null),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,yNe),Ahe),"Layer ID"),"Layer identifier that was calculated by ELK Layered for a node. This is only generated if interactiveLayot or generatePositionAndLayerIds is set."),rt(-1)),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,rhe),mxt),"Upper Bound On Width [MinWidth Layerer]"),"Defines a loose upper bound on the width of the MinWidth layerer. If set to '-1' multiple values are tested and the best result is selected."),rt(4)),Hs),ba),At(en)))),po(t,rhe,iq,jRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ihe),mxt),"Upper Layer Estimation Scaling Factor [MinWidth Layerer]"),"Multiplied with Upper Bound On Width for defining an upper bound on the width of layers which haven't been determined yet, but whose maximum width had been (roughly) estimated by the MinWidth algorithm. Compensates for too high estimations. If set to '-1' multiple values are tested and the best result is selected."),rt(2)),Hs),ba),At(en)))),po(t,ihe,iq,FRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ohe),bxt),"Node Promotion Strategy"),"Reduces number of dummy nodes after layering phase (if possible)."),fLe),Li),CDe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ahe),bxt),"Max Node Promotion Iterations"),"Limits the number of iterations for node promotion."),rt(0)),Hs),ba),At(en)))),po(t,ahe,ohe,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,she),"layering.coffmanGraham"),"Layer Bound"),"The maximum number of nodes allowed per layer."),rt(Qr)),Hs),ba),At(en)))),po(t,she,iq,RRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,p$),BL),"Crossing Minimization Strategy"),"Strategy for crossing minimization."),nLe),Li),fPe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ENe),BL),"Force Node Model Order"),"The node order given by the model does not change to produce a better layout. E.g. if node A is before node B in the model this is not changed during crossing minimization. This assumes that the node model order is already respected before crossing minimization. This can be achieved by setting considerModelOrder.strategy to NODES_AND_EDGES."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,che),BL),"Hierarchical Sweepiness"),"How likely it is to use cross-hierarchy (1) vs bottom-up (-1)."),.1),ya),mo),At(en)))),po(t,che,WQ,nRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,uhe),BL),"Semi-Interactive Crossing Minimization"),"Preserves the order of nodes within a layer but still minimizes crossings between edges connecting long edge dummies. Derives the desired order from positions specified by the 'org.eclipse.elk.position' layout option. Requires a crossing minimization strategy that is able to process 'in-layer' constraints."),!1),Po),to),At(en)))),po(t,uhe,p$,cRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,SNe),BL),"In Layer Predecessor of"),"Allows to set a constraint which specifies of which node the current node is the predecessor. If set to 's' then the node is the predecessor of 's' and is in the same layer"),null),i4),ft),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,_Ne),BL),"In Layer Successor of"),"Allows to set a constraint which specifies of which node the current node is the successor. If set to 's' then the node is the successor of 's' and is in the same layer"),null),i4),ft),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,kNe),BL),"Position Choice Constraint"),"Allows to set a constraint regarding the position placement of a node in a layer. Assumed the layer in which the node placed includes n other nodes and i < n. If set to i, it expresses that the node should be placed at the i-th position. Should i>=n be true then the node is placed at the last position in the layer. Note that this option is not part of any of ELK Layered's default configurations but is only evaluated as part of the `InteractiveLayeredGraphVisitor`, which must be applied manually or used via the `DiagramLayoutEngine."),null),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,TNe),BL),"Position ID"),"Position within a layer that was determined by ELK Layered for a node. This is only generated if interactiveLayot or generatePositionAndLayerIds is set."),rt(-1)),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,xNe),vxt),"Greedy Switch Activation Threshold"),"By default it is decided automatically if the greedy switch is activated or not. The decision is based on whether the size of the input graph (without dummy nodes) is smaller than the value of this option. A '0' enforces the activation."),rt(40)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,lhe),vxt),"Greedy Switch Crossing Minimization"),"Greedy Switch strategy for crossing minimization. The greedy switch heuristic is executed after the regular crossing minimization as a post-processor. Note that if 'hierarchyHandling' is set to 'INCLUDE_CHILDREN', the 'greedySwitchHierarchical.type' option must be used."),tLe),Li),Yge),At(en)))),po(t,lhe,p$,eRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,DQ),"crossingMinimization.greedySwitchHierarchical"),"Greedy Switch Crossing Minimization (hierarchical)"),"Activates the greedy switch heuristic in case hierarchical layout is used. The differences to the non-hierarchical case (see 'greedySwitch.type') are: 1) greedy switch is inactive by default, 3) only the option value set on the node at which hierarchical layout starts is relevant, and 2) if it's activated by the user, it properly addresses hierarchy-crossing edges."),eLe),Li),Yge),At(en)))),po(t,DQ,p$,JNt),po(t,DQ,WQ,ZNt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,lA),wxt),"Node Placement Strategy"),"Strategy for node placement."),yLe),Li),SDe),At(en)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,MQ),wxt),"Favor Straight Edges Over Balancing"),"Favor straight edges over a balanced node placement. The default behavior is determined automatically based on the used 'edgeRouting'. For an orthogonal style it is set to true, for all other styles to false."),Po),to),At(en)))),po(t,MQ,lA,e4t),po(t,MQ,lA,t4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,dhe),yxt),"BK Edge Straightening"),"Specifies whether the Brandes Koepf node placer tries to increase the number of straight edges at the expense of diagram size. There is a subtle difference to the 'favorStraightEdges' option, which decides whether a balanced placement of the nodes is desired, or not. In bk terms this means combining the four alignments into a single balanced one, or not. This option on the other hand tries to straighten additional edges during the creation of each of the four alignments."),bLe),Li),EPe),At(en)))),po(t,dhe,lA,XRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,fhe),yxt),"BK Fixed Alignment"),"Tells the BK node placer to use a certain alignment (out of its four) instead of the one producing the smallest height, or the combination of all four."),vLe),Li),xPe),At(en)))),po(t,fhe,lA,ZRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,hhe),"nodePlacement.linearSegments"),"Linear Segments Deflection Dampening"),"Dampens the movement of nodes to keep the diagram from getting too large."),.3),ya),mo),At(en)))),po(t,hhe,lA,r4t),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,phe),"nodePlacement.networkSimplex"),"Node Flexibility"),"Aims at shorter and straighter edges. Two configurations are possible: (a) allow ports to move freely on the side they are assigned to (the order is always defined beforehand), (b) additionally allow to enlarge a node wherever it helps. If this option is not configured for a node, the 'nodeFlexibility.default' value is used, which is specified for the node's parent."),Li),Ime),At(Ao)))),po(t,phe,lA,s4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ghe),"nodePlacement.networkSimplex.nodeFlexibility"),"Node Flexibility Default"),"Default value of the 'nodeFlexibility' option for the children of a hierarchical node."),wLe),Li),Ime),At(en)))),po(t,ghe,lA,a4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,CNe),Ext),"Self-Loop Distribution"),"Alter the distribution of the loops around the node. It only takes effect for PortConstraints.FREE."),sLe),Li),PDe),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ANe),Ext),"Self-Loop Ordering"),"Alter the ordering of the loops they can either be stacked or sequenced. It only takes effect for PortConstraints.FREE."),cLe),Li),LDe),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,jQ),"edgeRouting.splines"),"Spline Routing Mode"),"Specifies the way control points are assembled for each individual edge. CONSERVATIVE ensures that edges are properly routed around the nodes but feels rather orthogonal at times. SLOPPY uses fewer control points to obtain curvier edge routes but may result in edges overlapping nodes."),uLe),Li),MDe),At(en)))),po(t,jQ,oq,wRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,$Q),"edgeRouting.splines.sloppy"),"Sloppy Spline Layer Spacing Factor"),"Spacing factor for routing area between layers when using sloppy spline routing."),.2),ya),mo),At(en)))),po(t,$Q,oq,ERt),po(t,$Q,jQ,SRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,mhe),"edgeRouting.polyline"),"Sloped Edge Zone Width"),"Width of the strip to the left and to the right of each layer where the polyline edge router is allowed to refrain from ensuring that edges are routed horizontally. This prevents awkward bend points for nodes that extent almost to the edge of their layer."),2),ya),mo),At(en)))),po(t,mhe,oq,gRt),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,INe),Pg),"Spacing Base Value"),"An optional base value for all other layout options of the 'spacing' group. It can be used to conveniently alter the overall 'spaciousness' of the drawing. Whenever an explicit value is set for the other layout options, this base value will have no effect. The base value is not inherited, i.e. it must be set for each hierarchical node."),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ONe),Pg),"Edge Node Between Layers Spacing"),"The spacing to be preserved between nodes and edges that are routed next to the node's layer. For the spacing between nodes and edges that cross the node's layer 'spacing.edgeNode' is used."),10),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,NNe),Pg),"Edge Edge Between Layer Spacing"),"Spacing to be preserved between pairs of edges that are routed between the same pair of layers. Note that 'spacing.edgeEdge' is used for the spacing between pairs of edges crossing the same layer."),10),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,RNe),Pg),"Node Node Between Layers Spacing"),"The spacing to be preserved between any pair of nodes of two adjacent layers. Note that 'spacing.nodeNode' is used for the spacing between nodes within the layer itself."),20),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,PNe),GNe),"Direction Priority"),"Defines how important it is to have a certain edge point into the direction of the overall layout. This option is evaluated during the cycle breaking phase."),rt(0)),Hs),ba),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,LNe),GNe),"Shortness Priority"),"Defines how important it is to keep an edge as short as possible. This option is evaluated during the layering phase."),rt(0)),Hs),ba),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,DNe),GNe),"Straightness Priority"),"Defines how important it is to keep an edge straight, i.e. aligned with one of the two axes. This option is evaluated during node placement."),rt(0)),Hs),ba),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,bhe),qNe),"Connected Components Compaction"),"Tries to further compact components (disconnected sub-graphs)."),!1),Po),to),At(en)))),po(t,bhe,u$,!0),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,MNe),Sxt),"Post Compaction Strategy"),_xt),XPe),Li),RPe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,jNe),Sxt),"Post Compaction Constraint Calculation"),_xt),YPe),Li),lPe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,FQ),VNe),"High Degree Node Treatment"),"Makes room around high degree nodes to place leafs and trees."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,vhe),VNe),"High Degree Node Threshold"),"Whether a node is considered to have a high degree."),rt(16)),Hs),ba),At(en)))),po(t,vhe,FQ,!0),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,whe),VNe),"High Degree Node Maximum Tree Height"),"Maximum height of a subtree connected to a high degree node to be moved to separate layers."),rt(5)),Hs),ba),At(en)))),po(t,whe,FQ,!0),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,I1),WNe),"Graph Wrapping Strategy"),"For certain graphs and certain prescribed drawing areas it may be desirable to split the laid out graph into chunks that are placed side by side. The edges that connect different chunks are 'wrapped' around from the end of one chunk to the start of the other chunk. The points between the chunks are referred to as 'cuts'."),kLe),Li),BDe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,BQ),WNe),"Additional Wrapped Edges Spacing"),"To visually separate edges that are wrapped from regularly routed edges an additional spacing value can be specified in form of this layout option. The spacing is added to the regular edgeNode spacing."),10),ya),mo),At(en)))),po(t,BQ,I1,y4t),po(t,BQ,I1,E4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,UQ),WNe),"Correction Factor for Wrapping"),"At times and for certain types of graphs the executed wrapping may produce results that are consistently biased in the same fashion: either wrapping to often or to rarely. This factor can be used to correct the bias. Internally, it is simply multiplied with the 'aspect ratio' layout option."),1),ya),mo),At(en)))),po(t,UQ,I1,_4t),po(t,UQ,I1,k4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,g$),kxt),"Cutting Strategy"),"The strategy by which the layer indexes are determined at which the layering crumbles into chunks."),_Le),Li),pPe),At(en)))),po(t,g$,I1,O4t),po(t,g$,I1,N4t),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,yhe),kxt),"Manually Specified Cuts"),"Allows the user to specify her own cuts for a certain graph."),Nb),qf),At(en)))),po(t,yhe,g$,x4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Ehe),"wrapping.cutting.msd"),"MSD Freedom"),"The MSD cutting strategy starts with an initial guess on the number of chunks the graph should be split into. The freedom specifies how much the strategy may deviate from this guess. E.g. if an initial number of 3 is computed, a freedom of 1 allows 2, 3, and 4 cuts."),SLe),Hs),ba),At(en)))),po(t,Ehe,g$,A4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,zQ),Txt),"Validification Strategy"),"When wrapping graphs, one can specify indices that are not allowed as split points. The validification strategy makes sure every computed split point is allowed."),TLe),Li),FDe),At(en)))),po(t,zQ,I1,H4t),po(t,zQ,I1,G4t),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,HQ),Txt),"Valid Indices for Wrapping"),null),Nb),qf),At(en)))),po(t,HQ,I1,B4t),po(t,HQ,I1,U4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,GQ),KNe),"Improve Cuts"),"For general graphs it is important that not too many edges wrap backwards. Thus a compromise between evenly-distributed cuts and the total number of cut edges is sought."),!0),Po),to),At(en)))),po(t,GQ,I1,D4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,qQ),KNe),"Distance Penalty When Improving Cuts"),null),2),ya),mo),At(en)))),po(t,qQ,I1,P4t),po(t,qQ,GQ,!0),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,She),KNe),"Improve Wrapped Edges"),"The initial wrapping is performed in a very simple way. As a consequence, edges that wrap from one chunk to another may be unnecessarily long. Activating this option tries to shorten such edges."),!0),Po),to),At(en)))),po(t,She,I1,j4t),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,_he),YNe),"Layer Unzipping Strategy"),"The strategy to use for unzipping a layer into multiple sublayers while maintaining the existing ordering of nodes and edges after crossing minimization. The default value is 'NONE'."),mLe),Li),KPe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,$Ne),YNe),"Unzipping Layer Split"),"Defines the number of sublayers to split a layer into. The property can be set to the nodes in a layer, which then applies the property for the layer. If multiple nodes set the value to different values, then the lowest value is chosen."),pLe),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,khe),YNe),"Reset Alternation on Long Edges"),"If set to true, nodes will always be placed in the first sublayer after a long edge when using the ALTERNATING strategy. Otherwise long edge dummies are treated the same as regular nodes. The default value is true. The property can be set to the nodes in a layer, which then applies the property for the layer. If any node sets the value to false, then the value is set to false for the entire layer."),gLe),Po),to),At(Ao)))),po(t,khe,_he,qRt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,FNe),Ihe),"Edge Label Side Selection"),"Method to decide on edge label sides."),aLe),Li),wPe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,BNe),Ihe),"Edge Center Label Placement Strategy"),"Determines in which layer center labels of long edges should be placed."),oLe),Li),sD),xi(en,ce(se(Ob,1),Ve,170,0,[L1]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,VQ),m$),"Consider Model Order"),"Preserves the order of nodes and edges in the model file if this does not lead to additional edge crossings. Depending on the strategy this is not always possible since the node and edge order might be conflicting."),QPe),Li),IDe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,UNe),m$),"Consider Port Order"),"If disabled the port order of output ports is derived from the edge order and input ports are ordered by their incoming connections. If enabled all ports are ordered by the port model order."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,zNe),m$),"No Model Order"),"Set on a node to not set a model order for this node even though it is a real node."),!1),Po),to),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,The),m$),"Consider Model Order for Components"),"If set to NONE the usual ordering strategy (by cumulative node priority and size of nodes) is used. INSIDE_PORT_SIDES orders the components with external ports only inside the groups with the same port side. FORCE_MODEL_ORDER enforces the mode order on components. This option might produce bad alignments and sub optimal drawings in terms of used area since the ordering should be respected."),JPe),Li),g5e),At(en)))),po(t,The,u$,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,HNe),m$),"Long Edge Ordering Strategy"),"Indicates whether long edges are sorted under, over, or equal to nodes that have no connection to a previous layer in a left-to-right or right-to-left layout. Under and over changes to right and left in a vertical layout."),ZPe),Li),yDe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,xhe),m$),"Crossing Counter Node Order Influence"),"Indicates with what percentage (1 for 100%) violations of the node model order are weighted against the crossings e.g. a value of 0.5 means two model order violations are as important as on edge crossing. This allows some edge crossings in favor of preserving the model order. It is advised to set this value to a very small positive value (e.g. 0.001) to have minimal crossing and a optimal node order. Defaults to no influence (0)."),0),ya),mo),At(en)))),po(t,xhe,VQ,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Che),m$),"Crossing Counter Port Order Influence"),"Indicates with what percentage (1 for 100%) violations of the port model order are weighted against the crossings e.g. a value of 0.5 means two model order violations are as important as on edge crossing. This allows some edge crossings in favor of preserving the model order. It is advised to set this value to a very small positive value (e.g. 0.001) to have minimal crossing and a optimal port order. Defaults to no influence (0)."),0),ya),mo),At(en)))),po(t,Che,VQ,null),qkt((new ez,t))};var jNt,$Nt,FNt,YPe,BNt,XPe,UNt,JPe,zNt,HNt,GNt,ZPe,qNt,VNt,WNt,QPe,KNt,YNt,XNt,eLe,JNt,ZNt,QNt,tLe,eRt,tRt,nRt,rRt,iRt,oRt,aRt,sRt,cRt,uRt,nLe,lRt,rLe,dRt,iLe,fRt,oLe,hRt,aLe,pRt,gRt,mRt,sLe,bRt,cLe,vRt,uLe,wRt,yRt,ERt,SRt,_Rt,kRt,TRt,xRt,CRt,ARt,lLe,IRt,ORt,NRt,RRt,PRt,LRt,dLe,DRt,MRt,jRt,$Rt,FRt,BRt,URt,fLe,zRt,hLe,HRt,pLe,GRt,gLe,qRt,VRt,mLe,WRt,KRt,YRt,bLe,XRt,JRt,vLe,ZRt,QRt,e4t,t4t,n4t,r4t,i4t,o4t,wLe,a4t,s4t,c4t,yLe,u4t,ELe,l4t,d4t,f4t,h4t,p4t,g4t,m4t,b4t,v4t,w4t,y4t,E4t,S4t,_4t,k4t,T4t,x4t,C4t,SLe,A4t,I4t,_Le,O4t,N4t,R4t,P4t,L4t,D4t,M4t,j4t,$4t,kLe,F4t,B4t,U4t,z4t,TLe,H4t,G4t;R(Pc,"LayeredMetaDataProvider",838),N(975,1,Ng,ez),p.tf=function(t){qkt(t)};var Nm,sme,Tte,Z$,xte,xLe,Cte,nx,Ate,CLe,cme,Ite,ume,ip,lme,h_,ALe,Dq,dme,ILe,q4t,V4t,W4t,Ote,fme,Q$,RE,K4t,mf,OLe,NLe,Nte,hme,Rm,Rte,P1,RLe,PLe,LLe,pme,gme,DLe,z0,mme,MLe,rx,jLe,$Le,FLe,Pte,ix,PE,BLe,ULe,Ls,zLe,Y4t,el,Lte,HLe,GLe,qLe,xw,LE,gD,eF,VLe,Dte,WLe,KLe,Mte,p_,YLe,bme,tF,XLe,g_,nF,jte,DE,vme,mD,$te,ME,JLe,ZLe,QLe,bD,eDe,X4t,J4t,Z4t,Q4t,m_,ox,Xi,H0,e6t,ax,tDe,vD,nDe,sx,t6t,wD,rDe,e4,n6t,r6t,Mq,wme,iDe,jq,op,IA,OA,b_,jE,Fte,cx,yme,yD,ED,v_,NA,Eme,$q,rF,iF,i6t,o6t,a6t,oDe,s6t,Sme,aDe,sDe,cDe,uDe,_me,lDe,dDe,fDe,hDe,kme,Bte;R(Pc,"LayeredOptions",975),N(976,1,{},SC),p.uf=function(){var t;return t=new rrt,t},p.vf=function(t){},R(Pc,"LayeredOptions/LayeredFactory",976),N(1330,1,{}),p.a=0;var c6t;R(Lc,"ElkSpacings/AbstractSpacingsBuilder",1330),N(772,1330,{},XCe);var Ute,u6t;R(Pc,"LayeredSpacings/LayeredSpacingsBuilder",772),N(265,23,{3:1,34:1,23:1,265:1,188:1,196:1},w3),p.bg=function(){return cSt(this)},p.og=function(){return cSt(this)};var SD,Tme,_D,pDe,gDe,mDe,zte,xme,bDe,vDe=er(Pc,"LayeringStrategy",265,ar,Skn,z1n),l6t;N(349,23,{3:1,34:1,23:1,349:1},Jae);var Cme,wDe,Hte,yDe=er(Pc,"LongEdgeOrderingStrategy",349,ar,kEn,H1n),d6t;N(203,23,{3:1,34:1,23:1,203:1},jY);var RA,PA,Gte,Ame,Ime=er(Pc,"NodeFlexibility",203,ar,$2n,G1n),f6t;N(324,23,{3:1,34:1,23:1,324:1,188:1,196:1},$z),p.bg=function(){return ZEt(this)},p.og=function(){return ZEt(this)};var oF,Ome,Nme,aF,EDe,SDe=er(Pc,"NodePlacementStrategy",324,ar,NSn,q1n),h6t;N(242,23,{3:1,34:1,23:1,242:1},Yk);var _De,w_,ux,Fq,kDe,TDe,Bq,xDe,qte,Vte,CDe=er(Pc,"NodePromotionStrategy",242,ar,zkn,V1n),p6t;N(266,23,{3:1,34:1,23:1,266:1},$Y);var ADe,Pm,Rme,Pme,IDe=er(Pc,"OrderingStrategy",266,ar,F2n,W1n),g6t;N(418,23,{3:1,34:1,23:1,418:1},mSe);var Lme,Dme,ODe=er(Pc,"PortSortingStrategy",418,ar,Ryn,K1n),m6t;N(449,23,{3:1,34:1,23:1,449:1},Qae);var td,Uu,sF,b6t=er(Pc,"PortType",449,ar,TEn,Y1n),v6t;N(379,23,{3:1,34:1,23:1,379:1},Zae);var NDe,Mme,RDe,PDe=er(Pc,"SelfLoopDistributionStrategy",379,ar,xEn,X1n),w6t;N(345,23,{3:1,34:1,23:1,345:1},ese);var jme,Uq,$me,LDe=er(Pc,"SelfLoopOrderingStrategy",345,ar,CEn,J1n),y6t;N(313,1,{313:1},H_t),R(Pc,"Spacings",313),N(346,23,{3:1,34:1,23:1,346:1},tse);var Fme,DDe,cF,MDe=er(Pc,"SplineRoutingMode",346,ar,AEn,Z1n),E6t;N(348,23,{3:1,34:1,23:1,348:1},nse);var Bme,jDe,$De,FDe=er(Pc,"ValidifyStrategy",348,ar,IEn,Q1n),S6t;N(380,23,{3:1,34:1,23:1,380:1},rse);var lx,Ume,kD,BDe=er(Pc,"WrappingStrategy",380,ar,OEn,e0n),_6t;N(1337,1,fs,i1),p.pg=function(t){return f(t,36),k6t},p.If=function(t,i){bMn(this,f(t,36),i)};var k6t;R(JQ,"DepthFirstCycleBreaker",1337),N(773,1,fs,_ke),p.pg=function(t){return f(t,36),T6t},p.If=function(t,i){y7n(this,f(t,36),i)},p.qg=function(t){return f(ct(t,IZ(this.d,t.c.length)),9)};var T6t;R(JQ,"GreedyCycleBreaker",773),N(1340,773,fs,Qat),p.qg=function(t){var i,a,u,l;for(l=null,i=Qr,u=new ne(t);u.a<u.c.c.length;)a=f(ee(u),9),li(a,(Ge(),hi))&&f(V(a,hi),17).a<i&&(i=f(V(a,hi),17).a,l=a);return l||f(ct(t,IZ(this.d,t.c.length)),9)},R(JQ,"GreedyModelOrderCycleBreaker",1340),N(1338,1,fs,M9),p.pg=function(t){return f(t,36),x6t},p.If=function(t,i){IMn(this,f(t,36),i)};var x6t;R(JQ,"InteractiveCycleBreaker",1338),N(1339,1,fs,D9),p.pg=function(t){return f(t,36),C6t},p.If=function(t,i){UMn(this,f(t,36),i)},p.a=0,p.b=0;var C6t;R(JQ,"ModelOrderCycleBreaker",1339),N(1352,1,fs,uN),p.pg=function(t){return f(t,36),A6t},p.If=function(t,i){T9n(this,f(t,36),i)};var A6t;R(O1,"BreadthFirstModelOrderLayerer",1352),N(1353,1,Cr,L6),p.Le=function(t,i){return TRn(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(O1,"BreadthFirstModelOrderLayerer/lambda$0$Type",1353),N(1343,1,fs,Aot),p.pg=function(t){return f(t,36),I6t},p.If=function(t,i){x7n(this,f(t,36),i)};var I6t;R(O1,"CoffmanGrahamLayerer",1343),N(1344,1,Cr,Ket),p.Le=function(t,i){return s4n(this.a,f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(O1,"CoffmanGrahamLayerer/0methodref$compareNodesInTopo$Type",1344),N(1345,1,Cr,Yet),p.Le=function(t,i){return Yvn(this.a,f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(O1,"CoffmanGrahamLayerer/lambda$1$Type",1345),N(1354,1,fs,eu),p.pg=function(t){return f(t,36),O6t},p.If=function(t,i){f7n(this,f(t,36),i)},p.c=0,p.e=0;var O6t;R(O1,"DepthFirstModelOrderLayerer",1354),N(1355,1,Cr,hk),p.Le=function(t,i){return xRn(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(O1,"DepthFirstModelOrderLayerer/lambda$0$Type",1355),N(1346,1,fs,og),p.pg=function(t){return f(t,36),Sr(Sr(Sr(new _o,(wa(),rp),(ka(),Ege)),Ab,ZT),su,JT)},p.If=function(t,i){M9n(f(t,36),i)},R(O1,"InteractiveLayerer",1346),N(560,1,{560:1},drt),p.a=0,p.c=0,R(O1,"InteractiveLayerer/LayerSpan",560),N(1342,1,fs,lN),p.pg=function(t){return f(t,36),N6t},p.If=function(t,i){_5n(this,f(t,36),i)};var N6t;R(O1,"LongestPathLayerer",1342),N(1351,1,fs,tS),p.pg=function(t){return f(t,36),R6t},p.If=function(t,i){H5n(this,f(t,36),i)};var R6t;R(O1,"LongestPathSourceLayerer",1351),N(1349,1,fs,ul),p.pg=function(t){return f(t,36),Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)},p.If=function(t,i){W9n(this,f(t,36),i)},p.a=0,p.b=0,p.d=0;var UDe,zDe;R(O1,"MinWidthLayerer",1349),N(1350,1,Cr,Xet),p.Le=function(t,i){return fTn(this,f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(O1,"MinWidthLayerer/MinOutgoingEdgesComparator",1350),N(1341,1,fs,kl),p.pg=function(t){return f(t,36),P6t},p.If=function(t,i){l8n(this,f(t,36),i)};var P6t;R(O1,"NetworkSimplexLayerer",1341),N(1347,1,fs,iut),p.pg=function(t){return f(t,36),Sr(Sr(Sr(new _o,(wa(),rp),(ka(),vA)),Ab,ZT),su,JT)},p.If=function(t,i){W8n(this,f(t,36),i)},p.d=0,p.f=0,p.g=0,p.i=0,p.s=0,p.t=0,p.u=0,R(O1,"StretchWidthLayerer",1347),N(1348,1,Cr,MU),p.Le=function(t,i){return WSn(f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(O1,"StretchWidthLayerer/1",1348),N(402,1,RRe),p.eg=function(t,i,a,u,l,g){},p.sg=function(t,i,a){return jSt(this,t,i,a)},p.dg=function(){this.g=Pe(kx,Cxt,30,this.d,15,1),this.f=Pe(kx,Cxt,30,this.d,15,1)},p.fg=function(t,i){this.e[t]=Pe(pr,Pr,30,i[t].length,15,1)},p.gg=function(t,i,a){var u;u=a[t][i],u.p=i,this.e[t][i]=i},p.hg=function(t,i,a,u){f(ct(u[t][i].j,a),11).p=this.d++},p.b=0,p.c=0,p.d=0,R(Ll,"AbstractBarycenterPortDistributor",402),N(1639,1,Cr,Jet),p.Le=function(t,i){return C3n(this.a,f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Ll,"AbstractBarycenterPortDistributor/lambda$0$Type",1639),N(811,1,rq,ATe),p.eg=function(t,i,a,u,l,g){},p.gg=function(t,i,a){},p.hg=function(t,i,a,u){},p.cg=function(){return!1},p.dg=function(){this.c=this.e.a,this.g=this.f.g},p.fg=function(t,i){i[t][0].c.p=t},p.ig=function(){return!1},p.tg=function(t,i,a,u){a?Owt(this,t):(Mwt(this,t,u),ekt(this,t,i)),t.c.length>1&&(lt(ut(V(ra((Gt(0,t.c.length),f(t.c[0],9))),(ot(),h_))))?u2t(t,this.d,f(this,654)):(Ht(),na(t,this.d)),r1t(this.e,t))},p.jg=function(t,i,a,u){var l,g,m,v,S,T,A;for(i!=ylt(a,t.length)&&(g=t[i-(a?1:-1)],exe(this.f,g,a?(cs(),Uu):(cs(),td))),l=t[i][0],A=!u||l.k==(cn(),bo),T=gm(t[i]),this.tg(T,A,!1,a),m=0,S=new ne(T);S.a<S.c.c.length;)v=f(ee(S),9),t[i][m++]=v;return!1},p.kg=function(t,i){var a,u,l,g,m;for(m=ylt(i,t.length),g=gm(t[m]),this.tg(g,!1,!0,i),a=0,l=new ne(g);l.a<l.c.c.length;)u=f(ee(l),9),t[m][a++]=u;return!1},R(Ll,"BarycenterHeuristic",811),N(652,1,{652:1},ett),p.Ib=function(){return"BarycenterState [node="+this.c+", summedWeight="+this.d+", degree="+this.b+", barycenter="+this.a+", visited="+this.e+"]"},p.b=0,p.d=0,p.e=!1;var L6t=R(Ll,"BarycenterHeuristic/BarycenterState",652);N(1811,1,Cr,Zet),p.Le=function(t,i){return pOn(this.a,f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Ll,"BarycenterHeuristic/lambda$0$Type",1811),N(810,1,rq,sAe),p.dg=function(){},p.eg=function(t,i,a,u,l,g){},p.hg=function(t,i,a,u){},p.fg=function(t,i){this.a[t]=Pe(L6t,{3:1,4:1,5:1,2063:1},652,i[t].length,0,1),this.b[t]=Pe(D6t,{3:1,4:1,5:1,2064:1},238,i[t].length,0,1)},p.gg=function(t,i,a){pvt(this,a[t][i],!0)},p.c=!1,R(Ll,"ForsterConstraintResolver",810),N(238,1,{238:1},rft,z_t),p.Ib=function(){var t,i;for(i=new $v,i.a+="[",t=0;t<this.d.length;t++)xr(i,Rvt(this.d[t])),mg(this.g,this.d[0]).a!=null&&xr(xr((i.a+="<",i),Cgn(mg(this.g,this.d[0]).a)),">"),t<this.d.length-1&&(i.a+=Fu);return(i.a+="]",i).a},p.a=0,p.c=0,p.f=0;var D6t=R(Ll,"ForsterConstraintResolver/ConstraintGroup",238);N(1806,1,xn,Qet),p.Ad=function(t){pvt(this.a,f(t,9),!1)},R(Ll,"ForsterConstraintResolver/lambda$0$Type",1806),N(217,1,{217:1,229:1},q_t),p.eg=function(t,i,a,u,l,g){},p.fg=function(t,i){},p.dg=function(){this.r=Pe(pr,Pr,30,this.n,15,1)},p.gg=function(t,i,a){var u,l;l=a[t][i],u=l.e,u&&Qe(this.b,u)},p.hg=function(t,i,a,u){++this.n},p.Ib=function(){return nkt(this.e,new Mo)},p.g=!1,p.i=!1,p.n=0,p.s=!1,R(Ll,"GraphInfoHolder",217),N(1851,1,rq,k8),p.eg=function(t,i,a,u,l,g){},p.fg=function(t,i){},p.hg=function(t,i,a,u){},p.sg=function(t,i,a){return a&&i>0?Hce(this.a,t[i-1],t[i]):!a&&i<t.length-1?Hce(this.a,t[i],t[i+1]):Lue(this.a,t[i],a?(tt(),pn):(tt(),_n)),O6n(this,t,i,a)},p.dg=function(){this.d=Pe(pr,Pr,30,this.c,15,1),this.a=new oH(this.d)},p.gg=function(t,i,a){var u;u=a[t][i],this.c+=u.j.c.length},p.c=0,R(Ll,"GreedyPortDistributor",1851),N(1360,1,fs,j9),p.pg=function(t){return Txn(f(t,36))},p.If=function(t,i){C8n(f(t,36),i)};var M6t;R(Ll,"InteractiveCrossingMinimizer",1360),N(1361,1,Cr,ntt),p.Le=function(t,i){return KIn(this,f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Ll,"InteractiveCrossingMinimizer/1",1361),N(498,1,{498:1,106:1,44:1},Sae),p.pg=function(t){var i;return f(t,36),i=nX(j6t),Sr(i,(wa(),su),(ka(),tte)),i},p.If=function(t,i){zLn(this,f(t,36),i)},p.e=0;var j6t;R(Ll,"LayerSweepCrossingMinimizer",498),N(1357,1,xn,rtt),p.Ad=function(t){IDn(this.a,f(t,217))},R(Ll,"LayerSweepCrossingMinimizer/0methodref$compareDifferentRandomizedLayouts$Type",1357),N(1358,1,xn,itt),p.Ad=function(t){Exn(this.a,f(t,217))},R(Ll,"LayerSweepCrossingMinimizer/1methodref$minimizeCrossingsNoCounter$Type",1358),N(1359,1,xn,ott),p.Ad=function(t){E_t(this.a,f(t,217))},R(Ll,"LayerSweepCrossingMinimizer/2methodref$minimizeCrossingsWithCounter$Type",1359),N(450,23,{3:1,34:1,23:1,450:1},ise);var zq,uF,Wte,$6t=er(Ll,"LayerSweepCrossingMinimizer/CrossMinType",450,ar,PEn,t0n),F6t;N(1356,1,_r,jU),p.Mb=function(t){return Xxe(),f(t,25).a.c.length==0},R(Ll,"LayerSweepCrossingMinimizer/lambda$0$Type",1356),N(1808,1,rq,Bft),p.dg=function(){},p.eg=function(t,i,a,u,l,g){},p.hg=function(t,i,a,u){},p.fg=function(t,i){i[t][0].c.p=t,this.b[t]=Pe(B6t,{3:1,4:1,5:1,1989:1},653,i[t].length,0,1)},p.gg=function(t,i,a){var u;u=a[t][i],u.p=i,io(this.b[t],i,new T8)},R(Ll,"LayerSweepTypeDecider",1808),N(653,1,{653:1},T8),p.Ib=function(){return"NodeInfo [connectedEdges="+this.a+", hierarchicalInfluence="+this.b+", randomInfluence="+this.c+"]"},p.a=0,p.b=0,p.c=0;var B6t=R(Ll,"LayerSweepTypeDecider/NodeInfo",653);N(1809,1,km,$2),p.Lb=function(t){return F7(new yg(f(t,11).b))},p.Fb=function(t){return this===t},p.Mb=function(t){return F7(new yg(f(t,11).b))},R(Ll,"LayerSweepTypeDecider/lambda$0$Type",1809),N(1810,1,km,wv),p.Lb=function(t){return F7(new yg(f(t,11).b))},p.Fb=function(t){return this===t},p.Mb=function(t){return F7(new yg(f(t,11).b))},R(Ll,"LayerSweepTypeDecider/lambda$1$Type",1810),N(1852,402,RRe,hit),p.rg=function(t,i,a){var u,l,g,m,v,S,T,A,L;switch(T=this.g,a.g){case 1:{for(u=0,l=0,S=new ne(t.j);S.a<S.c.c.length;)m=f(ee(S),11),m.e.c.length!=0&&(++u,m.j==(tt(),hn)&&++l);for(g=i+l,L=i+u,v=$S(t,(cs(),td)).Jc();v.Ob();)m=f(v.Pb(),11),m.j==(tt(),hn)?(T[m.p]=g,--g):(T[m.p]=L,--L);return u}case 2:{for(A=0,v=$S(t,(cs(),Uu)).Jc();v.Ob();)m=f(v.Pb(),11),++A,T[m.p]=i+A;return A}default:throw ie(new Sz)}},R(Ll,"LayerTotalPortDistributor",1852),N(654,811,{654:1,229:1},kbt),p.tg=function(t,i,a,u){a?Owt(this,t):(Mwt(this,t,u),ekt(this,t,i)),t.c.length>1&&(lt(ut(V(ra((Gt(0,t.c.length),f(t.c[0],9))),(ot(),h_))))?u2t(t,this.d,this):(Ht(),na(t,this.d)),lt(ut(V(ra((Gt(0,t.c.length),f(t.c[0],9))),h_)))||r1t(this.e,t))},R(Ll,"ModelOrderBarycenterHeuristic",654),N(1812,1,Cr,att),p.Le=function(t,i){return qNn(this.a,f(t,9),f(i,9))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Ll,"ModelOrderBarycenterHeuristic/lambda$0$Type",1812),N(1362,1,fs,$9),p.pg=function(t){var i;return f(t,36),i=nX(U6t),Sr(i,(wa(),su),(ka(),tte)),i},p.If=function(t,i){byn((f(t,36),i))};var U6t;R(Ll,"NoCrossingMinimizer",1362),N(789,402,RRe,U2e),p.rg=function(t,i,a){var u,l,g,m,v,S,T,A,L,F,z;switch(L=this.g,a.g){case 1:{for(l=0,g=0,A=new ne(t.j);A.a<A.c.c.length;)S=f(ee(A),11),S.e.c.length!=0&&(++l,S.j==(tt(),hn)&&++g);for(u=1/(l+1),m=i+g*u,z=i+1-u,T=$S(t,(cs(),td)).Jc();T.Ob();)S=f(T.Pb(),11),S.j==(tt(),hn)?(L[S.p]=m,m-=u):(L[S.p]=z,z-=u);break}case 2:{for(v=0,A=new ne(t.j);A.a<A.c.c.length;)S=f(ee(A),11),S.g.c.length==0||++v;for(u=1/(v+1),F=i+u,T=$S(t,(cs(),Uu)).Jc();T.Ob();)S=f(T.Pb(),11),L[S.p]=F,F+=u;break}default:throw ie(new dn("Port type is undefined"))}return 1},R(Ll,"NodeRelativePortDistributor",789),N(801,1,{},Vlt,hyt),R(Ll,"SweepCopy",801),N(1807,1,rq,J0t),p.fg=function(t,i){},p.dg=function(){var t;t=Pe(pr,Pr,30,this.f,15,1),this.d=new ptt(t),this.a=new oH(t)},p.eg=function(t,i,a,u,l,g){var m;m=f(ct(g[t][i].j,a),11),l.c==m&&l.c.i.c==l.d.i.c&&++this.e[t]},p.gg=function(t,i,a){var u;u=a[t][i],this.c[t]=this.c[t]|u.k==(cn(),xu)},p.hg=function(t,i,a,u){var l;l=f(ct(u[t][i].j,a),11),l.p=this.f++,l.g.c.length+l.e.c.length>1&&(l.j==(tt(),_n)?this.b[t]=!0:l.j==pn&&t>0&&(this.b[t-1]=!0))},p.f=0,R(xb,"AllCrossingsCounter",1807),N(578,1,{},YJ),p.b=0,p.d=0,R(xb,"BinaryIndexedTree",578),N(517,1,{},oH);var HDe,Kte;R(xb,"CrossingsCounter",517),N(1896,1,Cr,stt),p.Le=function(t,i){return Mvn(this.a,f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(xb,"CrossingsCounter/lambda$0$Type",1896),N(1897,1,Cr,ctt),p.Le=function(t,i){return jvn(this.a,f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(xb,"CrossingsCounter/lambda$1$Type",1897),N(1898,1,Cr,utt),p.Le=function(t,i){return $vn(this.a,f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(xb,"CrossingsCounter/lambda$2$Type",1898),N(1899,1,Cr,ltt),p.Le=function(t,i){return Fvn(this.a,f(t,11),f(i,11))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(xb,"CrossingsCounter/lambda$3$Type",1899),N(hw,1,xn,dtt),p.Ad=function(t){A_n(this.a,f(t,11))},R(xb,"CrossingsCounter/lambda$4$Type",hw),N(1901,1,_r,ftt),p.Mb=function(t){return _gn(this.a,f(t,11))},R(xb,"CrossingsCounter/lambda$5$Type",1901),N(1902,1,xn,htt),p.Ad=function(t){hat(this,t)},R(xb,"CrossingsCounter/lambda$6$Type",1902),N(1903,1,xn,rat),p.Ad=function(t){var i;vP(),Jv(this.b,(i=this.a,f(t,11),i))},R(xb,"CrossingsCounter/lambda$7$Type",1903),N(818,1,km,_C),p.Lb=function(t){return vP(),li(f(t,11),(Ge(),ed))},p.Fb=function(t){return this===t},p.Mb=function(t){return vP(),li(f(t,11),(Ge(),ed))},R(xb,"CrossingsCounter/lambda$8$Type",818),N(1895,1,{},ptt),R(xb,"HyperedgeCrossingsCounter",1895),N(464,1,{34:1,464:1},out),p.Dd=function(t){return l3n(this,f(t,464))},p.b=0,p.c=0,p.e=0,p.f=0;var sjn=R(xb,"HyperedgeCrossingsCounter/Hyperedge",464);N(367,1,{34:1,367:1},KX),p.Dd=function(t){return r6n(this,f(t,367))},p.b=0,p.c=0;var z6t=R(xb,"HyperedgeCrossingsCounter/HyperedgeCorner",367);N(516,23,{3:1,34:1,23:1,516:1},bSe);var lF,dF,H6t=er(xb,"HyperedgeCrossingsCounter/HyperedgeCorner/Type",516,ar,Nyn,n0n),G6t;N(1364,1,fs,F9),p.pg=function(t){return f(V(f(t,36),(Ge(),Cu)),22).Gc((us(),Vf))?q6t:null},p.If=function(t,i){RIn(this,f(t,36),i)};var q6t;R(hs,"InteractiveNodePlacer",1364),N(1365,1,fs,cb),p.pg=function(t){return f(V(f(t,36),(Ge(),Cu)),22).Gc((us(),Vf))?V6t:null},p.If=function(t,i){wAn(this,f(t,36),i)};var V6t,Yte,Xte;R(hs,"LinearSegmentsNodePlacer",1365),N(261,1,{34:1,261:1},g2e),p.Dd=function(t){return Mpn(this,f(t,261))},p.Fb=function(t){var i;return me(t,261)?(i=f(t,261),this.b==i.b):!1},p.Hb=function(){return this.b},p.Ib=function(){return"ls"+Cg(this.e)},p.a=0,p.b=0,p.c=-1,p.d=-1,p.g=0;var W6t=R(hs,"LinearSegmentsNodePlacer/LinearSegment",261);N(1367,1,fs,Olt),p.pg=function(t){return f(V(f(t,36),(Ge(),Cu)),22).Gc((us(),Vf))?K6t:null},p.If=function(t,i){h7n(this,f(t,36),i)},p.b=0,p.g=0;var K6t;R(hs,"NetworkSimplexPlacer",1367),N(1386,1,Cr,x8),p.Le=function(t,i){return tu(f(t,17).a,f(i,17).a)},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(hs,"NetworkSimplexPlacer/0methodref$compare$Type",1386),N(1388,1,Cr,yp),p.Le=function(t,i){return tu(f(t,17).a,f(i,17).a)},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(hs,"NetworkSimplexPlacer/1methodref$compare$Type",1388),N(639,1,{639:1},iat);var cjn=R(hs,"NetworkSimplexPlacer/EdgeRep",639);N(401,1,{401:1},nTe),p.b=!1;var ujn=R(hs,"NetworkSimplexPlacer/NodeRep",401);N(499,13,{3:1,4:1,20:1,31:1,55:1,13:1,16:1,15:1,59:1,499:1},mrt),R(hs,"NetworkSimplexPlacer/Path",499),N(1368,1,{},C8),p.Kb=function(t){return f(t,18).d.i.k},R(hs,"NetworkSimplexPlacer/Path/lambda$0$Type",1368),N(1369,1,_r,TO),p.Mb=function(t){return f(t,247)==(cn(),uo)},R(hs,"NetworkSimplexPlacer/Path/lambda$1$Type",1369),N(1370,1,{},D6),p.Kb=function(t){return f(t,18).d.i},R(hs,"NetworkSimplexPlacer/Path/lambda$2$Type",1370),N(1371,1,_r,gtt),p.Mb=function(t){return Hct(W0t(f(t,9)))},R(hs,"NetworkSimplexPlacer/Path/lambda$3$Type",1371),N(1372,1,_r,A8),p.Mb=function(t){return Tvn(f(t,11))},R(hs,"NetworkSimplexPlacer/lambda$0$Type",1372),N(1373,1,xn,oat),p.Ad=function(t){hmn(this.a,this.b,f(t,11))},R(hs,"NetworkSimplexPlacer/lambda$1$Type",1373),N(1382,1,xn,mtt),p.Ad=function(t){qRn(this.a,f(t,18))},R(hs,"NetworkSimplexPlacer/lambda$10$Type",1382),N(1383,1,{},xO),p.Kb=function(t){return rf(),new Ft(null,new Bt(f(t,25).a,16))},R(hs,"NetworkSimplexPlacer/lambda$11$Type",1383),N(1384,1,xn,btt),p.Ad=function(t){yPn(this.a,f(t,9))},R(hs,"NetworkSimplexPlacer/lambda$12$Type",1384),N(1385,1,{},M6),p.Kb=function(t){return rf(),rt(f(t,123).e)},R(hs,"NetworkSimplexPlacer/lambda$13$Type",1385),N(1387,1,{},pk),p.Kb=function(t){return rf(),rt(f(t,123).e)},R(hs,"NetworkSimplexPlacer/lambda$15$Type",1387),N(1389,1,_r,I8),p.Mb=function(t){return rf(),f(t,401).c.k==(cn(),Yi)},R(hs,"NetworkSimplexPlacer/lambda$17$Type",1389),N(1390,1,_r,O8),p.Mb=function(t){return rf(),f(t,401).c.j.c.length>1},R(hs,"NetworkSimplexPlacer/lambda$18$Type",1390),N(1391,1,xn,zdt),p.Ad=function(t){LCn(this.c,this.b,this.d,this.a,f(t,401))},p.c=0,p.d=0,R(hs,"NetworkSimplexPlacer/lambda$19$Type",1391),N(1374,1,{},i0),p.Kb=function(t){return rf(),new Ft(null,new Bt(f(t,25).a,16))},R(hs,"NetworkSimplexPlacer/lambda$2$Type",1374),N(1392,1,xn,vtt),p.Ad=function(t){wmn(this.a,f(t,11))},p.a=0,R(hs,"NetworkSimplexPlacer/lambda$20$Type",1392),N(1393,1,{},F2),p.Kb=function(t){return rf(),new Ft(null,new Bt(f(t,25).a,16))},R(hs,"NetworkSimplexPlacer/lambda$21$Type",1393),N(1394,1,xn,wtt),p.Ad=function(t){Tmn(this.a,f(t,9))},R(hs,"NetworkSimplexPlacer/lambda$22$Type",1394),N(1395,1,_r,gk),p.Mb=function(t){return Hct(t)},R(hs,"NetworkSimplexPlacer/lambda$23$Type",1395),N(1396,1,{},ky),p.Kb=function(t){return rf(),new Ft(null,new Bt(f(t,25).a,16))},R(hs,"NetworkSimplexPlacer/lambda$24$Type",1396),N(1397,1,_r,ytt),p.Mb=function(t){return Rgn(this.a,f(t,9))},R(hs,"NetworkSimplexPlacer/lambda$25$Type",1397),N(1398,1,xn,aat),p.Ad=function(t){YNn(this.a,this.b,f(t,9))},R(hs,"NetworkSimplexPlacer/lambda$26$Type",1398),N(1399,1,_r,yv),p.Mb=function(t){return rf(),!Fa(f(t,18))},R(hs,"NetworkSimplexPlacer/lambda$27$Type",1399),N(1400,1,_r,kC),p.Mb=function(t){return rf(),!Fa(f(t,18))},R(hs,"NetworkSimplexPlacer/lambda$28$Type",1400),N(1401,1,{},Ett),p.Te=function(t,i){return ymn(this.a,f(t,25),f(i,25))},R(hs,"NetworkSimplexPlacer/lambda$29$Type",1401),N(1375,1,{},Zb),p.Kb=function(t){return rf(),new Ft(null,new sT(new Tn(Cn(ao(f(t,9)).a.Jc(),new B))))},R(hs,"NetworkSimplexPlacer/lambda$3$Type",1375),N(1376,1,_r,TC),p.Mb=function(t){return rf(),E2n(f(t,18))},R(hs,"NetworkSimplexPlacer/lambda$4$Type",1376),N(1377,1,xn,Stt),p.Ad=function(t){kDn(this.a,f(t,18))},R(hs,"NetworkSimplexPlacer/lambda$5$Type",1377),N(1378,1,{},Qb),p.Kb=function(t){return rf(),new Ft(null,new Bt(f(t,25).a,16))},R(hs,"NetworkSimplexPlacer/lambda$6$Type",1378),N(1379,1,_r,Dh),p.Mb=function(t){return rf(),f(t,9).k==(cn(),Yi)},R(hs,"NetworkSimplexPlacer/lambda$7$Type",1379),N(1380,1,{},N8),p.Kb=function(t){return rf(),new Ft(null,new sT(new Tn(Cn(T1(f(t,9)).a.Jc(),new B))))},R(hs,"NetworkSimplexPlacer/lambda$8$Type",1380),N(1381,1,_r,j6),p.Mb=function(t){return rf(),Avn(f(t,18))},R(hs,"NetworkSimplexPlacer/lambda$9$Type",1381),N(1363,1,fs,fN),p.pg=function(t){return f(V(f(t,36),(Ge(),Cu)),22).Gc((us(),Vf))?Y6t:null},p.If=function(t,i){tMn(f(t,36),i)};var Y6t;R(hs,"SimpleNodePlacer",1363),N(185,1,{185:1},nA),p.Ib=function(){var t;return t="",this.c==(fm(),y_)?t+=RR:this.c==G0&&(t+=NR),this.o==(vg(),$E)?t+=Gfe:this.o==Mg?t+="UP":t+="BALANCED",t},R(bw,"BKAlignedLayout",185),N(507,23,{3:1,34:1,23:1,507:1},vSe);var G0,y_,X6t=er(bw,"BKAlignedLayout/HDirection",507,ar,Iyn,r0n),J6t;N(506,23,{3:1,34:1,23:1,506:1},wSe);var $E,Mg,Z6t=er(bw,"BKAlignedLayout/VDirection",506,ar,Oyn,i0n),Q6t;N(1640,1,{},sat),R(bw,"BKAligner",1640),N(1643,1,{},wwt),R(bw,"BKCompactor",1643),N(647,1,{647:1},B2),p.a=0,R(bw,"BKCompactor/ClassEdge",647),N(453,1,{453:1},hrt),p.a=null,p.b=0,R(bw,"BKCompactor/ClassNode",453),N(1366,1,fs,uat),p.pg=function(t){return f(V(f(t,36),(Ge(),Cu)),22).Gc((us(),Vf))?e5t:null},p.If=function(t,i){O7n(this,f(t,36),i)},p.d=!1;var e5t;R(bw,"BKNodePlacer",1366),N(1641,1,{},Ty),p.d=0,R(bw,"NeighborhoodInformation",1641),N(1642,1,Cr,_tt),p.Le=function(t,i){return V_n(this,f(t,47),f(i,47))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(bw,"NeighborhoodInformation/NeighborComparator",1642),N(802,1,{}),R(bw,"ThresholdStrategy",802),N(1771,802,{},brt),p.ug=function(t,i,a){return this.a.o==(vg(),Mg)?Gi:qo},p.vg=function(){},R(bw,"ThresholdStrategy/NullThresholdStrategy",1771),N(570,1,{570:1},cat),p.c=!1,p.d=!1,R(bw,"ThresholdStrategy/Postprocessable",570),N(1772,802,{},vrt),p.ug=function(t,i,a){var u,l,g;return l=i==a,u=this.a.a[a.p]==i,l||u?(g=t,this.a.c==(fm(),y_)?(l&&(g=Wde(this,i,!0)),!isNaN(g)&&!isFinite(g)&&u&&(g=Wde(this,a,!1))):(l&&(g=Wde(this,i,!0)),!isNaN(g)&&!isFinite(g)&&u&&(g=Wde(this,a,!1))),g):t},p.vg=function(){for(var t,i,a,u,l;this.d.b!=0;)l=f(Jyn(this.d),570),u=QSt(this,l),u.a&&(t=u.a,a=lt(this.a.f[this.a.g[l.b.p].p]),!(!a&&!Fa(t)&&t.c.i.c==t.d.i.c)&&(i=i2t(this,l),i||Bgn(this.e,l)));for(;this.e.a.c.length!=0;)i2t(this,f(G1t(this.e),570))},R(bw,"ThresholdStrategy/SimpleThresholdStrategy",1772),N(630,1,{630:1,188:1,196:1},Ev),p.bg=function(){return n1t(this)},p.og=function(){return n1t(this)};var zme;R(Lhe,"EdgeRouterFactory",630),N(1424,1,fs,M5),p.pg=function(t){return J5n(f(t,36))},p.If=function(t,i){cMn(f(t,36),i)};var t5t,n5t,r5t,i5t,o5t,GDe,a5t,s5t;R(Lhe,"OrthogonalEdgeRouter",1424),N(1417,1,fs,lat),p.pg=function(t){return UIn(f(t,36))},p.If=function(t,i){I9n(this,f(t,36),i)};var c5t,u5t,l5t,d5t,Hq,f5t;R(Lhe,"PolylineEdgeRouter",1417),N(1418,1,km,$6),p.Lb=function(t){return tCe(f(t,9))},p.Fb=function(t){return this===t},p.Mb=function(t){return tCe(f(t,9))},R(Lhe,"PolylineEdgeRouter/1",1418),N(1818,1,_r,e1),p.Mb=function(t){return f(t,133).c==(Ip(),Cw)},R(Dp,"HyperEdgeCycleDetector/lambda$0$Type",1818),N(1819,1,{},Wa),p.Xe=function(t){return f(t,133).d},R(Dp,"HyperEdgeCycleDetector/lambda$1$Type",1819),N(1820,1,_r,Uc),p.Mb=function(t){return f(t,133).c==(Ip(),Cw)},R(Dp,"HyperEdgeCycleDetector/lambda$2$Type",1820),N(1821,1,{},ws),p.Xe=function(t){return f(t,133).d},R(Dp,"HyperEdgeCycleDetector/lambda$3$Type",1821),N(1822,1,{},mk),p.Xe=function(t){return f(t,133).d},R(Dp,"HyperEdgeCycleDetector/lambda$4$Type",1822),N(1823,1,{},CO),p.Xe=function(t){return f(t,133).d},R(Dp,"HyperEdgeCycleDetector/lambda$5$Type",1823),N(116,1,{34:1,116:1},VH),p.Dd=function(t){return jpn(this,f(t,116))},p.Fb=function(t){var i;return me(t,116)?(i=f(t,116),this.g==i.g):!1},p.Hb=function(){return this.g},p.Ib=function(){var t,i,a,u;for(t=new ef("{"),u=new ne(this.n);u.a<u.c.c.length;)a=f(ee(u),11),i=lG(a.i),i==null&&(i="n"+yut(a.i)),t.a+=""+i,u.a<u.c.c.length&&(t.a+=",");return t.a+="}",t.a},p.a=0,p.b=0,p.c=NaN,p.d=0,p.g=0,p.i=0,p.o=0,p.s=NaN,R(Dp,"HyperEdgeSegment",116),N(133,1,{133:1},nE),p.Ib=function(){return this.a+"->"+this.b+" ("+Zmn(this.c)+")"},p.d=0,R(Dp,"HyperEdgeSegmentDependency",133),N(513,23,{3:1,34:1,23:1,513:1},ySe);var Cw,dx,h5t=er(Dp,"HyperEdgeSegmentDependency/DependencyType",513,ar,Gyn,o0n),p5t;N(1824,1,{},ktt),R(Dp,"HyperEdgeSegmentSplitter",1824),N(1825,1,{},bit),p.a=0,p.b=0,R(Dp,"HyperEdgeSegmentSplitter/AreaRating",1825),N(335,1,{335:1},Kse),p.a=0,p.b=0,p.c=0,R(Dp,"HyperEdgeSegmentSplitter/FreeArea",335),N(1826,1,Cr,U2),p.Le=function(t,i){return Qbn(f(t,116),f(i,116))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Dp,"HyperEdgeSegmentSplitter/lambda$0$Type",1826),N(1827,1,xn,Hdt),p.Ad=function(t){aSn(this.a,this.d,this.c,this.b,f(t,116))},p.b=0,R(Dp,"HyperEdgeSegmentSplitter/lambda$1$Type",1827),N(1828,1,{},Mh),p.Kb=function(t){return new Ft(null,new Bt(f(t,116).e,16))},R(Dp,"HyperEdgeSegmentSplitter/lambda$2$Type",1828),N(1829,1,{},bk),p.Kb=function(t){return new Ft(null,new Bt(f(t,116).j,16))},R(Dp,"HyperEdgeSegmentSplitter/lambda$3$Type",1829),N(1830,1,{},xy),p.We=function(t){return ke(Oe(t))},R(Dp,"HyperEdgeSegmentSplitter/lambda$4$Type",1830),N(648,1,{},vce),p.a=0,p.b=0,p.c=0,R(Dp,"OrthogonalRoutingGenerator",648),N(1644,1,{},ib),p.Kb=function(t){return new Ft(null,new Bt(f(t,116).e,16))},R(Dp,"OrthogonalRoutingGenerator/lambda$0$Type",1644),N(1645,1,{},R8),p.Kb=function(t){return new Ft(null,new Bt(f(t,116).j,16))},R(Dp,"OrthogonalRoutingGenerator/lambda$1$Type",1645),N(655,1,{}),R(Dhe,"BaseRoutingDirectionStrategy",655),N(1816,655,{},wrt),p.wg=function(t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J;if(!(t.r&&!t.q))for(A=i+t.o*a,T=new ne(t.n);T.a<T.c.c.length;)for(S=f(ee(T),11),L=uc(ce(se(Zo,1),Ye,8,0,[S.i.n,S.n,S.a])).a,v=new ne(S.g);v.a<v.c.c.length;)m=f(ee(v),18),Fa(m)||(W=m.d,J=uc(ce(se(Zo,1),Ye,8,0,[W.i.n,W.n,W.a])).a,h.Math.abs(L-J)>Tm&&(g=A,l=t,u=new Ke(L,g),Mr(m.a,u),qS(this,m,l,u,!1),F=t.r,F&&(z=ke(Oe(Xh(F.e,0))),u=new Ke(z,g),Mr(m.a,u),qS(this,m,l,u,!1),g=i+F.o*a,l=F,u=new Ke(z,g),Mr(m.a,u),qS(this,m,l,u,!1)),u=new Ke(J,g),Mr(m.a,u),qS(this,m,l,u,!1)))},p.xg=function(t){return t.i.n.a+t.n.a+t.a.a},p.yg=function(){return tt(),Bn},p.zg=function(){return tt(),hn},R(Dhe,"NorthToSouthRoutingStrategy",1816),N(1817,655,{},yrt),p.wg=function(t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J;if(!(t.r&&!t.q))for(A=i-t.o*a,T=new ne(t.n);T.a<T.c.c.length;)for(S=f(ee(T),11),L=uc(ce(se(Zo,1),Ye,8,0,[S.i.n,S.n,S.a])).a,v=new ne(S.g);v.a<v.c.c.length;)m=f(ee(v),18),Fa(m)||(W=m.d,J=uc(ce(se(Zo,1),Ye,8,0,[W.i.n,W.n,W.a])).a,h.Math.abs(L-J)>Tm&&(g=A,l=t,u=new Ke(L,g),Mr(m.a,u),qS(this,m,l,u,!1),F=t.r,F&&(z=ke(Oe(Xh(F.e,0))),u=new Ke(z,g),Mr(m.a,u),qS(this,m,l,u,!1),g=i-F.o*a,l=F,u=new Ke(z,g),Mr(m.a,u),qS(this,m,l,u,!1)),u=new Ke(J,g),Mr(m.a,u),qS(this,m,l,u,!1)))},p.xg=function(t){return t.i.n.a+t.n.a+t.a.a},p.yg=function(){return tt(),hn},p.zg=function(){return tt(),Bn},R(Dhe,"SouthToNorthRoutingStrategy",1817),N(1815,655,{},Ert),p.wg=function(t,i,a){var u,l,g,m,v,S,T,A,L,F,z,W,J;if(!(t.r&&!t.q))for(A=i+t.o*a,T=new ne(t.n);T.a<T.c.c.length;)for(S=f(ee(T),11),L=uc(ce(se(Zo,1),Ye,8,0,[S.i.n,S.n,S.a])).b,v=new ne(S.g);v.a<v.c.c.length;)m=f(ee(v),18),Fa(m)||(W=m.d,J=uc(ce(se(Zo,1),Ye,8,0,[W.i.n,W.n,W.a])).b,h.Math.abs(L-J)>Tm&&(g=A,l=t,u=new Ke(g,L),Mr(m.a,u),qS(this,m,l,u,!0),F=t.r,F&&(z=ke(Oe(Xh(F.e,0))),u=new Ke(g,z),Mr(m.a,u),qS(this,m,l,u,!0),g=i+F.o*a,l=F,u=new Ke(g,z),Mr(m.a,u),qS(this,m,l,u,!0)),u=new Ke(g,J),Mr(m.a,u),qS(this,m,l,u,!0)))},p.xg=function(t){return t.i.n.b+t.n.b+t.a.b},p.yg=function(){return tt(),_n},p.zg=function(){return tt(),pn},R(Dhe,"WestToEastRoutingStrategy",1815),N(807,1,{},rOe),p.Ib=function(){return Cg(this.a)},p.b=0,p.c=!1,p.d=!1,p.f=0,R(HT,"NubSpline",807),N(407,1,{407:1},$2t,wft),R(HT,"NubSpline/PolarCP",407),N(1419,1,fs,uwt),p.pg=function(t){return COn(f(t,36))},p.If=function(t,i){Y9n(this,f(t,36),i)};var g5t,m5t,b5t,v5t,w5t;R(HT,"SplineEdgeRouter",1419),N(270,1,{270:1},wJ),p.Ib=function(){return this.a+" ->("+this.c+") "+this.b},p.c=0,R(HT,"SplineEdgeRouter/Dependency",270),N(451,23,{3:1,34:1,23:1,451:1},ESe);var Aw,LA,y5t=er(HT,"SplineEdgeRouter/SideToProcess",451,ar,qyn,a0n),E5t;N(1420,1,_r,Cs),p.Mb=function(t){return qj(),!f(t,132).o},R(HT,"SplineEdgeRouter/lambda$0$Type",1420),N(1421,1,{},xC),p.Xe=function(t){return qj(),f(t,132).v+1},R(HT,"SplineEdgeRouter/lambda$1$Type",1421),N(1422,1,xn,gat),p.Ad=function(t){Evn(this.a,this.b,f(t,47))},R(HT,"SplineEdgeRouter/lambda$2$Type",1422),N(1423,1,xn,pat),p.Ad=function(t){Svn(this.a,this.b,f(t,47))},R(HT,"SplineEdgeRouter/lambda$3$Type",1423),N(132,1,{34:1,132:1},Vyt,uOe),p.Dd=function(t){return $pn(this,f(t,132))},p.b=0,p.e=!1,p.f=0,p.g=0,p.j=!1,p.k=!1,p.n=0,p.o=!1,p.p=!1,p.q=!1,p.s=0,p.u=0,p.v=0,p.F=0,R(HT,"SplineSegment",132),N(454,1,{454:1},F6),p.a=0,p.b=!1,p.c=!1,p.d=!1,p.e=!1,p.f=0,R(HT,"SplineSegment/EdgeInformation",454),N(1162,1,{},P8),R(N1,YOe,1162),N(1163,1,Cr,L8),p.Le=function(t,i){return i4n(f(t,120),f(i,120))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(N1,BTt,1163),N(1161,1,{},Mit),R(N1,"MrTree",1161),N(396,23,{3:1,34:1,23:1,396:1,188:1,196:1},FY),p.bg=function(){return gEt(this)},p.og=function(){return gEt(this)};var Jte,fF,hF,pF,qDe=er(N1,"TreeLayoutPhases",396,ar,B2n,s0n),S5t;N(1077,222,JS,aut),p.kf=function(t,i){var a,u,l,g,m,v,S,T;for(lt(ut(Fe(t,(Ec(),mMe))))||vH((a=new ri((fS(),new uS(t))),a)),m=i.bh($he),m.Sg("build tGraph",1),v=(S=new xH,Rc(S,t),$e(S,(_i(),mF),t),T=new Wn,PLn(t,S,T),XLn(t,S,T),S),m.Tg(),m=i.bh($he),m.Sg("Split graph",1),g=FLn(this.a,v),m.Tg(),l=new ne(g);l.a<l.c.c.length;)u=f(ee(l),120),gOn(this.b,u,i.bh(.5999999940395355/g.c.length));m=i.bh($he),m.Sg("Pack components",1),v=N7n(g),m.Tg(),m=i.bh($he),m.Sg("Apply layout results",1),b8n(v),m.Tg()},R(N1,"TreeLayoutProvider",1077),N(1840,1,Eb,U6),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return Ht(),uP(),tD},R(N1,"TreeUtil/1",1840),N(1841,1,Eb,z2),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return Ht(),uP(),tD},R(N1,"TreeUtil/2",1841),N(1831,1,_r,AO),p.Mb=function(t){return lt(ut(V(f(t,39),(_i(),Iw))))},R(N1,"TreeUtil/lambda$0$Type",1831),N(1837,1,_r,Ttt),p.Mb=function(t){return this.a.Gc(f(t,39))},R(N1,"TreeUtil/lambda$10$Type",1837),N(1838,1,{},xtt),p.Kb=function(t){return z2n(this.a,f(t,39))},R(N1,"TreeUtil/lambda$11$Type",1838),N(1839,1,_r,mat),p.Mb=function(t){return L_n(this.a,this.b,f(t,39))},R(N1,"TreeUtil/lambda$12$Type",1839),N(1832,1,_r,Ctt),p.Mb=function(t){return vCn(this.a,f(t,65))},R(N1,"TreeUtil/lambda$3$Type",1832),N(1833,1,Cr,B6),p.Le=function(t,i){return Jbn(f(t,65),f(i,65))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(N1,"TreeUtil/lambda$4$Type",1833),N(1834,1,_r,Att),p.Mb=function(t){return wCn(this.a,f(t,65))},R(N1,"TreeUtil/lambda$7$Type",1834),N(1835,1,Cr,ag),p.Le=function(t,i){return Zbn(f(t,65),f(i,65))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(N1,"TreeUtil/lambda$8$Type",1835),N(1836,1,{},Pd),p.Kb=function(t){return f(t,65).b},R(N1,"TreeUtil/lambda$9$Type",1836),N(492,148,{3:1,492:1,102:1,148:1}),p.g=0,R(b$,"TGraphElement",492),N(65,492,{3:1,65:1,492:1,102:1,148:1},oTe),p.Ib=function(){return this.b&&this.c?iE(this.b)+"->"+iE(this.c):"e_"+Ti(this)},R(b$,"TEdge",65),N(120,148,{3:1,120:1,102:1,148:1},xH),p.Ib=function(){var t,i,a,u,l;for(l=null,u=tr(this.b,0);u.b!=u.d.c;)a=f(Zn(u),39),l+=(a.c==null||a.c.length==0?"n_"+a.g:"n_"+a.c)+`
`;for(i=tr(this.a,0);i.b!=i.d.c;)t=f(Zn(i),65),l+=(t.b&&t.c?iE(t.b)+"->"+iE(t.c):"e_"+Ti(t))+`
`;return l};var ljn=R(b$,"TGraph",120);N(628,492,{3:1,492:1,628:1,102:1,148:1}),R(b$,"TShape",628),N(39,628,{3:1,492:1,39:1,628:1,102:1,148:1},Que),p.Ib=function(){return iE(this)};var Zte=R(b$,"TNode",39);N(235,1,Eb,u1),p.Ic=function(t){Oa(this,t)},p.Jc=function(){var t;return t=tr(this.a.d,0),new m3(t)},R(b$,"TNode/2",235),N(329,1,aa,m3),p.Nb=function(t){va(this,t)},p.Pb=function(){return f(Zn(this.a),65).c},p.Ob=function(){return kz(this.a)},p.Qb=function(){Cue(this.a)},R(b$,"TNode/2/1",329),N(1869,1,vi,H2),p.If=function(t,i){T7n(this,f(t,120),i)},R(Tu,"CompactionProcessor",1869),N(1870,1,Cr,Itt),p.Le=function(t,i){return lTn(this.a,f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Tu,"CompactionProcessor/lambda$0$Type",1870),N(1871,1,_r,bat),p.Mb=function(t){return kyn(this.b,this.a,f(t,47))},p.a=0,p.b=0,R(Tu,"CompactionProcessor/lambda$1$Type",1871),N(1880,1,Cr,CC),p.Le=function(t,i){return ywn(f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Tu,"CompactionProcessor/lambda$10$Type",1880),N(1881,1,Cr,IO),p.Le=function(t,i){return Hmn(f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Tu,"CompactionProcessor/lambda$11$Type",1881),N(1882,1,Cr,AC),p.Le=function(t,i){return Ewn(f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Tu,"CompactionProcessor/lambda$12$Type",1882),N(1872,1,_r,Ott),p.Mb=function(t){return xmn(this.a,f(t,47))},p.a=0,R(Tu,"CompactionProcessor/lambda$2$Type",1872),N(1873,1,_r,Ntt),p.Mb=function(t){return Cmn(this.a,f(t,47))},p.a=0,R(Tu,"CompactionProcessor/lambda$3$Type",1873),N(1874,1,_r,o0),p.Mb=function(t){return f(t,39).c.indexOf(QQ)==-1},R(Tu,"CompactionProcessor/lambda$4$Type",1874),N(1875,1,{},Rtt),p.Kb=function(t){return w2n(this.a,f(t,39))},p.a=0,R(Tu,"CompactionProcessor/lambda$5$Type",1875),N(1876,1,{},Ptt),p.Kb=function(t){return O_n(this.a,f(t,39))},p.a=0,R(Tu,"CompactionProcessor/lambda$6$Type",1876),N(1877,1,Cr,Ltt),p.Le=function(t,i){return BSn(this.a,f(t,239),f(i,239))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Tu,"CompactionProcessor/lambda$7$Type",1877),N(1878,1,Cr,Dtt),p.Le=function(t,i){return USn(this.a,f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Tu,"CompactionProcessor/lambda$8$Type",1878),N(1879,1,Cr,OO),p.Le=function(t,i){return Gmn(f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Tu,"CompactionProcessor/lambda$9$Type",1879),N(1867,1,vi,z6),p.If=function(t,i){RPn(f(t,120),i)},R(Tu,"DirectionProcessor",1867),N(1859,1,vi,sut),p.If=function(t,i){YLn(this,f(t,120),i)},R(Tu,"FanProcessor",1859),N(1883,1,vi,H6),p.If=function(t,i){EPn(f(t,120),i)},R(Tu,"GraphBoundsProcessor",1883),N(1884,1,{},G6),p.We=function(t){return f(t,39).e.a},R(Tu,"GraphBoundsProcessor/lambda$0$Type",1884),N(1885,1,{},NO),p.We=function(t){return f(t,39).e.b},R(Tu,"GraphBoundsProcessor/lambda$1$Type",1885),N(1886,1,{},D8),p.We=function(t){return lgn(f(t,39))},R(Tu,"GraphBoundsProcessor/lambda$2$Type",1886),N(1887,1,{},RO),p.We=function(t){return dgn(f(t,39))},R(Tu,"GraphBoundsProcessor/lambda$3$Type",1887),N(262,23,{3:1,34:1,23:1,262:1,196:1},hS),p.bg=function(){switch(this.g){case 0:return new Drt;case 1:return new sut;case 2:return new Lrt;case 3:return new $U;case 4:return new V6;case 8:return new q6;case 5:return new z6;case 6:return new K6;case 7:return new H2;case 9:return new H6;case 10:return new M8;default:throw ie(new dn(Zfe+(this.f!=null?this.f:""+this.g)))}};var VDe,WDe,KDe,YDe,XDe,JDe,ZDe,QDe,eMe,tMe,Hme,djn=er(Tu,Qfe,262,ar,Jbt,c0n),_5t;N(1866,1,vi,q6),p.If=function(t,i){S9n(f(t,120),i)},R(Tu,"LevelCoordinatesProcessor",1866),N(1864,1,vi,V6),p.If=function(t,i){n5n(this,f(t,120),i)},p.a=0,R(Tu,"LevelHeightProcessor",1864),N(1865,1,Eb,PO),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return Ht(),uP(),tD},R(Tu,"LevelHeightProcessor/1",1865),N(1860,1,vi,Lrt),p.If=function(t,i){hPn(this,f(t,120),i)},R(Tu,"LevelProcessor",1860),N(1861,1,_r,IC),p.Mb=function(t){return lt(ut(V(f(t,39),(_i(),Iw))))},R(Tu,"LevelProcessor/lambda$0$Type",1861),N(1862,1,vi,$U),p.If=function(t,i){pRn(this,f(t,120),i)},p.a=0,R(Tu,"NeighborsProcessor",1862),N(1863,1,Eb,W6),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return Ht(),uP(),tD},R(Tu,"NeighborsProcessor/1",1863),N(1868,1,vi,K6),p.If=function(t,i){WLn(this,f(t,120),i)},p.a=0,R(Tu,"NodePositionProcessor",1868),N(1858,1,vi,Drt),p.If=function(t,i){xMn(this,f(t,120),i)},R(Tu,"RootProcessor",1858),N(1888,1,vi,M8),p.If=function(t,i){tAn(f(t,120),i)},R(Tu,"Untreeifyer",1888),N(383,23,{3:1,34:1,23:1,383:1},ase);var Gq,Gme,nMe,rMe=er(sq,"EdgeRoutingMode",383,ar,REn,u0n),k5t,qq,TD,qme,iMe,oMe,Vme,Wme,aMe,Kme,sMe,Yme,gF,Xme,Qte,ene,ap,Fp,xD,mF,bF,q0,cMe,T5t,Jme,Iw,Vq,Wq;N(841,1,Ng,B9),p.tf=function(t){Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,DRe),""),Lxt),"Turns on Tree compaction which decreases the size of the whole tree by placing nodes of multiple levels in one large level"),(mn(),!1)),(bE(),Po)),to),At((wm(),en))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,MRe),""),"Edge End Texture Length"),"Should be set to the length of the texture at the end of an edge. This value can be used to improve the Edge Routing."),7),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,jRe),""),"Tree Level"),"The index for the tree level the node is in"),rt(0)),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,$Re),""),Lxt),"When set to a positive number this option will force the algorithm to place the node to the specified position within the trees layer if weighting is set to constraint"),rt(-1)),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,FRe),""),"Weighting of Nodes"),"Which weighting to use when computing a node order."),dMe),Li),_Me),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,BRe),""),"Edge Routing Mode"),"Chooses an Edge Routing algorithm."),uMe),Li),rMe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,URe),""),"Search Order"),"Which search order to use when computing a spanning tree."),lMe),Li),TMe),At(en)))),Tkt((new D5,t))};var x5t,C5t,A5t,uMe,I5t,O5t,lMe,N5t,R5t,dMe;R(sq,"MrTreeMetaDataProvider",841),N(983,1,Ng,D5),p.tf=function(t){Tkt(t)};var P5t,fMe,hMe,E_,pMe,gMe,Zme,L5t,D5t,M5t,j5t,$5t,F5t,B5t,mMe,bMe,vMe,U5t,DA,tne,wMe,z5t,yMe,Qme,H5t,G5t,q5t,EMe,V5t,Lm,SMe;R(sq,"MrTreeOptions",983),N(984,1,{},LO),p.uf=function(){var t;return t=new aut,t},p.vf=function(t){},R(sq,"MrTreeOptions/MrtreeFactory",984),N(350,23,{3:1,34:1,23:1,350:1},BY);var ebe,nne,tbe,nbe,_Me=er(sq,"OrderWeighting",350,ar,G2n,l0n),W5t;N(422,23,{3:1,34:1,23:1,422:1},SSe);var kMe,rbe,TMe=er(sq,"TreeifyingOrder",422,ar,Fyn,d0n),K5t;N(1425,1,fs,tz),p.pg=function(t){return f(t,120),Y5t},p.If=function(t,i){Wkn(this,f(t,120),i)};var Y5t;R("org.eclipse.elk.alg.mrtree.p1treeify","DFSTreeifyer",1425),N(1426,1,fs,dN),p.pg=function(t){return f(t,120),X5t},p.If=function(t,i){vPn(this,f(t,120),i)};var X5t;R(UL,"NodeOrderer",1426),N(1433,1,{},FU),p.rd=function(t){return ult(t)},R(UL,"NodeOrderer/0methodref$lambda$6$Type",1433),N(1427,1,_r,RC),p.Mb=function(t){return lR(),lt(ut(V(f(t,39),(_i(),Iw))))},R(UL,"NodeOrderer/lambda$0$Type",1427),N(1428,1,_r,$O),p.Mb=function(t){return lR(),f(V(f(t,39),(Ec(),DA)),17).a<0},R(UL,"NodeOrderer/lambda$1$Type",1428),N(1429,1,_r,jtt),p.Mb=function(t){return Akn(this.a,f(t,39))},R(UL,"NodeOrderer/lambda$2$Type",1429),N(1430,1,_r,Mtt),p.Mb=function(t){return y2n(this.a,f(t,39))},R(UL,"NodeOrderer/lambda$3$Type",1430),N(1431,1,Cr,J6),p.Le=function(t,i){return Z_n(f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(UL,"NodeOrderer/lambda$4$Type",1431),N(1432,1,_r,FO),p.Mb=function(t){return lR(),f(V(f(t,39),(_i(),Wme)),17).a!=0},R(UL,"NodeOrderer/lambda$5$Type",1432),N(1434,1,fs,_p),p.pg=function(t){return f(t,120),J5t},p.If=function(t,i){SLn(this,f(t,120),i)},p.b=0;var J5t;R("org.eclipse.elk.alg.mrtree.p3place","NodePlacer",1434),N(1435,1,fs,hN),p.pg=function(t){return f(t,120),Z5t},p.If=function(t,i){oLn(f(t,120),i)};var Z5t,fjn=R(Hd,"EdgeRouter",1435);N(1437,1,Cr,DO),p.Le=function(t,i){return tu(f(t,17).a,f(i,17).a)},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/0methodref$compare$Type",1437),N(1442,1,{},Y6),p.We=function(t){return ke(Oe(t))},R(Hd,"EdgeRouter/1methodref$doubleValue$Type",1442),N(1444,1,Cr,MO),p.Le=function(t,i){return di(ke(Oe(t)),ke(Oe(i)))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/2methodref$compare$Type",1444),N(1446,1,Cr,$8),p.Le=function(t,i){return di(ke(Oe(t)),ke(Oe(i)))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/3methodref$compare$Type",1446),N(1448,1,{},j8),p.We=function(t){return ke(Oe(t))},R(Hd,"EdgeRouter/4methodref$doubleValue$Type",1448),N(1450,1,Cr,X6),p.Le=function(t,i){return di(ke(Oe(t)),ke(Oe(i)))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/5methodref$compare$Type",1450),N(1452,1,Cr,Da),p.Le=function(t,i){return di(ke(Oe(t)),ke(Oe(i)))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/6methodref$compare$Type",1452),N(1436,1,{},vk),p.Kb=function(t){return y1(),f(V(f(t,39),(Ec(),Lm)),17)},R(Hd,"EdgeRouter/lambda$0$Type",1436),N(1447,1,{},jO),p.Kb=function(t){return rbn(f(t,39))},R(Hd,"EdgeRouter/lambda$11$Type",1447),N(1449,1,{},vat),p.Kb=function(t){return kvn(this.b,this.a,f(t,39))},p.a=0,p.b=0,R(Hd,"EdgeRouter/lambda$13$Type",1449),N(1451,1,{},wat),p.Kb=function(t){return ibn(this.b,this.a,f(t,39))},p.a=0,p.b=0,R(Hd,"EdgeRouter/lambda$15$Type",1451),N(1453,1,Cr,wk),p.Le=function(t,i){return P3n(f(t,65),f(i,65))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/lambda$17$Type",1453),N(1454,1,Cr,OC),p.Le=function(t,i){return L3n(f(t,65),f(i,65))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/lambda$18$Type",1454),N(1455,1,Cr,NC),p.Le=function(t,i){return M3n(f(t,65),f(i,65))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/lambda$19$Type",1455),N(1438,1,_r,$tt),p.Mb=function(t){return eEn(this.a,f(t,39))},p.a=0,R(Hd,"EdgeRouter/lambda$2$Type",1438),N(1456,1,Cr,F8),p.Le=function(t,i){return D3n(f(t,65),f(i,65))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/lambda$20$Type",1456),N(1439,1,Cr,Zg),p.Le=function(t,i){return dvn(f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/lambda$3$Type",1439),N(1440,1,Cr,sg),p.Le=function(t,i){return fvn(f(t,39),f(i,39))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"EdgeRouter/lambda$4$Type",1440),N(1441,1,{},Sv),p.Kb=function(t){return obn(f(t,39))},R(Hd,"EdgeRouter/lambda$5$Type",1441),N(1443,1,{},yat),p.Kb=function(t){return _vn(this.b,this.a,f(t,39))},p.a=0,p.b=0,R(Hd,"EdgeRouter/lambda$7$Type",1443),N(1445,1,{},Eat),p.Kb=function(t){return abn(this.b,this.a,f(t,39))},p.a=0,p.b=0,R(Hd,"EdgeRouter/lambda$9$Type",1445),N(660,1,{660:1},Zvt),p.e=0,p.f=!1,p.g=!1,R(Hd,"MultiLevelEdgeNodeNodeGap",660),N(1889,1,Cr,PC),p.Le=function(t,i){return hEn(f(t,239),f(i,239))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"MultiLevelEdgeNodeNodeGap/lambda$0$Type",1889),N(1890,1,Cr,Z6),p.Le=function(t,i){return pEn(f(t,239),f(i,239))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Hd,"MultiLevelEdgeNodeNodeGap/lambda$1$Type",1890);var MA;N(485,23,{3:1,34:1,23:1,485:1,188:1,196:1},_Se),p.bg=function(){return z0t(this)},p.og=function(){return z0t(this)};var rne,jA,xMe=er(zRe,"RadialLayoutPhases",485,ar,Byn,f0n),Q5t;N(1078,222,JS,Rit),p.kf=function(t,i){var a,u,l,g,m,v;if(a=L2t(this,t),i.Sg("Radial layout",a.c.length),lt(ut(Fe(t,(uw(),jMe))))||vH((u=new ri((fS(),new uS(t))),u)),v=NOn(t),pi(t,(R3(),MA),v),!v)throw ie(new dn("The given graph is not a tree!"));for(l=ke(Oe(Fe(t,ane))),l==0&&(l=uEt(t)),pi(t,ane,l),m=new ne(L2t(this,t));m.a<m.c.c.length;)g=f(ee(m),44),g.If(t,i.bh(1));i.Tg()},R(zRe,"RadialLayoutProvider",1078),N(541,1,Cr,vY),p.Le=function(t,i){return KPn(this.a,this.b,f(t,27),f(i,27))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},p.a=0,p.b=0,R(zRe,"RadialUtil/lambda$0$Type",541),N(1334,1,vi,LC),p.If=function(t,i){C9n(f(t,27),i)},R(qRe,"CalculateGraphSize",1334),N(1335,1,vi,B8),p.If=function(t,i){JDn(f(t,27))},R(qRe,"EdgeAngleCalculator",1335),N(361,23,{3:1,34:1,23:1,361:1,196:1},Bz),p.bg=function(){switch(this.g){case 0:return new U8;case 1:return new DC;case 2:return new Cy;case 3:return new LC;case 4:return new B8;default:throw ie(new dn(Zfe+(this.f!=null?this.f:""+this.g)))}};var ibe,obe,abe,sbe,cbe,ePt=er(qRe,Qfe,361,ar,ASn,h0n),tPt;N(636,1,{}),p.e=1,p.g=0,R(Fhe,"AbstractRadiusExtensionCompaction",636),N(1780,636,{},kct),p.Ag=function(t){var i,a,u,l,g,m,v,S,T;for(this.c=f(Fe(t,(R3(),MA)),27),EN(this,this.c),this.d=Dle(f(Fe(t,(uw(),Kq)),299)),S=f(Fe(t,lbe),17),S&&e7(this,S.a),v=Oe(Fe(t,(Rr(),zE))),SN(this,(an(v),v)),T=RT(this.c),this.d&&this.d.Eg(T),uLn(this,T),m=new bu(ce(se(Ur,1),$xt,27,0,[this.c])),a=0;a<2;a++)for(i=0;i<T.c.length;i++)l=new bu(ce(se(Ur,1),$xt,27,0,[(Gt(i,T.c.length),f(T.c[i],27))])),g=i<T.c.length-1?(Gt(i+1,T.c.length),f(T.c[i+1],27)):(Gt(0,T.c.length),f(T.c[0],27)),u=i==0?f(ct(T,T.c.length-1),27):(Gt(i-1,T.c.length),f(T.c[i-1],27)),ryt(this,(Gt(i,T.c.length),f(T.c[i],27),m),u,g,l)},R(Fhe,"AnnulusWedgeCompaction",1780),N(1332,1,vi,DC),p.If=function(t,i){Vkn(f(t,27),i)},R(Fhe,"GeneralCompactor",1332),N(1779,636,{},Q6),p.Ag=function(t){var i,a,u,l;a=f(Fe(t,(R3(),MA)),27),this.f=a,this.b=Dle(f(Fe(t,(uw(),Kq)),299)),l=f(Fe(t,lbe),17),l&&e7(this,l.a),u=Oe(Fe(t,(Rr(),zE))),SN(this,(an(u),u)),i=RT(a),this.b&&this.b.Eg(i),Myt(this,i)},p.a=0,R(Fhe,"RadialCompaction",1779),N(1788,1,{},BO),p.Bg=function(t){var i,a,u,l,g,m;for(this.a=t,i=0,m=RT(t),u=0,g=new ne(m);g.a<g.c.c.length;)for(l=f(ee(g),27),++u,a=u;a<m.c.length;a++)SDn(this,l,(Gt(a,m.c.length),f(m.c[a],27)))&&(i+=1);return i},R(VRe,"CrossingMinimizationPosition",1788),N(1786,1,{},MC),p.Bg=function(t){var i,a,u,l,g,m,v,S,T,A,L,F,z;for(u=0,a=new Tn(Cn(lw(t).a.Jc(),new B));Kn(a);)i=f(Dn(a),85),v=Us(f(be((!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c),0),84)),T=v.i+v.g/2,A=v.j+v.f/2,l=t.i+t.g/2,g=t.j+t.f/2,L=new ha,L.a=T-l,L.b=A-g,m=new Ke(L.a,L.b),uL(m,t.g,t.f),L.a-=m.a,L.b-=m.b,l=T-L.a,g=A-L.b,S=new Ke(L.a,L.b),uL(S,v.g,v.f),L.a-=S.a,L.b-=S.b,T=l+L.a,A=g+L.b,F=T-l,z=A-g,u+=h.Math.sqrt(F*F+z*z);return u},R(VRe,"EdgeLengthOptimization",1786),N(1787,1,{},_v),p.Bg=function(t){var i,a,u,l,g,m,v,S,T,A,L;for(u=0,a=new Tn(Cn(lw(t).a.Jc(),new B));Kn(a);)i=f(Dn(a),85),v=Us(f(be((!i.c&&(i.c=new Qt(Yn,i,5,8)),i.c),0),84)),S=v.i+v.g/2,T=v.j+v.f/2,l=f(Fe(v,(Rr(),o4)),8),g=t.i+l.a+t.g/2,m=t.j+l.b+t.f,A=S-g,L=T-m,u+=h.Math.sqrt(A*A+L*L);return u},R(VRe,"EdgeLengthPositionOptimization",1787),N(1331,636,vi,U8),p.If=function(t,i){ARn(this,f(t,27),i)},R("org.eclipse.elk.alg.radial.intermediate.overlaps","RadiusExtensionOverlapRemoval",1331),N(1333,1,vi,Cy),p.If=function(t,i){awn(f(t,27),i)},R("org.eclipse.elk.alg.radial.intermediate.rotation","GeneralRotator",1333),N(423,23,{3:1,34:1,23:1,423:1},kSe);var CMe,ube,AMe=er(v$,"AnnulusWedgeCriteria",423,ar,Hyn,g0n),nPt;N(384,23,{3:1,34:1,23:1,384:1},sse);var ine,IMe,OMe,NMe=er(v$,Fxt,384,ar,LEn,p0n),rPt;N(842,1,Ng,rz),p.tf=function(t){Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,WRe),""),"Center On Root"),"Centers the layout on the root of the tree i.e. so that the central node is also the center node of the final layout. This introduces additional whitespace."),(mn(),!1)),(bE(),Po)),to),At((wm(),en))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,KRe),""),"Order ID"),"The id can be used to define an order for nodes of one radius. This can be used to sort them in the layer accordingly."),rt(0)),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,YRe),""),"Radius"),"The radius option can be used to set the initial radius for the radial layouter."),0),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,nee),""),"Rotate"),"The rotate option determines whether a rotation of the layout should be performed."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Uhe),""),Bxt),"With the compacter option it can be determined how compaction on the graph is done. It can be chosen between none, the radial compaction or the compaction of wedges separately."),RMe),Li),NMe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,zhe),""),"Compaction Step Size"),"Determine the size of steps with which the compaction is done. Step size 1 correlates to a compaction of 1 pixel per Iteration."),rt(1)),Hs),ba),At(en)))),po(t,zhe,Uhe,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,XRe),""),"Sorter"),"Sort the nodes per radius according to the sorting algorithm. The strategies are none, by the given order id, or sorting them by polar coordinates."),LMe),Li),YMe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,JRe),""),"Annulus Wedge Criteria"),"Determine how the wedge for the node placement is calculated. It can be chosen between wedge determination by the number of leaves or by the maximum sum of diagonals."),DMe),Li),AMe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ZRe),""),"Translation Optimization"),"Find the optimal translation of the nodes of the first radii according to this criteria. For example edge crossings can be minimized."),PMe),Li),VMe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Hhe),e4e),"Target Angle"),"The angle in radians that the layout should be rotated to after layout."),0),ya),mo),At(en)))),po(t,Hhe,nee,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Ghe),e4e),"Additional Wedge Space"),"If set to true, modifies the target angle by rotating further such that space is left for an edge to pass in between the nodes. This option should only be used in conjunction with top-down layout."),!1),Po),to),At(en)))),po(t,Ghe,nee,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,QRe),e4e),"Outgoing Edge Angles"),"Calculate the required angle of connected nodes to leave space for an incoming edge. This option should only be used in conjunction with top-down layout."),!1),Po),to),At(en)))),vkt((new nz,t))};var iPt,oPt,aPt,RMe,sPt,PMe,cPt,uPt,lPt,dPt,fPt,hPt,pPt,LMe,gPt,DMe;R(v$,"RadialMetaDataProvider",842),N(985,1,Ng,nz),p.tf=function(t){vkt(t)};var MMe,lbe,dbe,mPt,bPt,vPt,wPt,jMe,$Me,one,yPt,EPt,ane,FMe,BMe,UMe,fbe,Kq,SPt,zMe;R(v$,"RadialOptions",985),N(986,1,{},z8),p.uf=function(){var t;return t=new Rit,t},p.vf=function(t){},R(v$,"RadialOptions/RadialFactory",986),N(351,23,{3:1,34:1,23:1,351:1},UY);var HMe,GMe,qMe,hbe,VMe=er(v$,"RadialTranslationStrategy",351,ar,q2n,m0n),_Pt;N(299,23,{3:1,34:1,23:1,299:1},cse);var WMe,pbe,KMe,YMe=er(v$,"SortingStrategy",299,ar,DEn,b0n),kPt;N(1415,1,fs,e5),p.pg=function(t){return f(t,27),null},p.If=function(t,i){KRn(this,f(t,27),i)},p.c=0,R("org.eclipse.elk.alg.radial.p1position","EadesRadial",1415),N(1784,1,{},G2),p.Cg=function(t){return Dvt(t)},R(Uxt,"AnnulusWedgeByLeafs",1784),N(1785,1,{},t5),p.Cg=function(t){return Ywt(this,t)},R(Uxt,"AnnulusWedgeByNodeSpace",1785),N(1416,1,fs,UO),p.pg=function(t){return f(t,27),null},p.If=function(t,i){mCn(this,f(t,27),i)},R("org.eclipse.elk.alg.radial.p2routing","StraightLineEdgeRouter",1416),N(805,1,{},k2e),p.Dg=function(t){},p.Eg=function(t){qnt(this,t)},R(t4e,"IDSorter",805),N(1783,1,Cr,H8),p.Le=function(t,i){return qkn(f(t,27),f(i,27))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(t4e,"IDSorter/lambda$0$Type",1783),N(1782,1,{},Obt),p.Dg=function(t){Nht(this,t)},p.Eg=function(t){var i;t.dc()||(this.e||(i=Elt(f(t.Xb(0),27)),Nht(this,i)),qnt(this.e,t))},R(t4e,"PolarCoordinateSorter",1782),N(433,23,{3:1,34:1,23:1,433:1},use);var vF,sne,gbe,XMe=er(qxt,"RectPackingLayoutPhases",433,ar,MEn,v0n),TPt;N(1082,222,JS,Pit),p.kf=function(t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;if(i.Sg("Rectangle Packing",1),L=f(Fe(t,(Ag(),EF)),107),S=lt(ut(Fe(t,KPt))),A=ke(Oe(Fe(t,SF))),Ct=lt(ut(Fe(t,l8e))),Ee=(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a),lt(ut(Fe(t,_be)))||vH((l=new ri((fS(),new uS(t))),l)),xt=!1,Ct&&Ee.i>=3)for(He=f(be(Ee,0),27),nt=f(be(Ee,1),27),g=0;g+2<Ee.i;)if(Le=He,He=nt,nt=f(be(Ee,g+2),27),Le.f>=He.f+nt.f+A||nt.f>=Le.f+He.f+A){xt=!0;break}else++g;else xt=!0;if(!xt){for(F=Ee.i,v=new Nn(Ee);v.e!=v.i.gc();)m=f(Mn(v),27),pi(m,(Rr(),aV),rt(F)),--F;c_t(t,new CN),i.Tg();return}for(a=(pH(this.a),Ap(this.a,(wZ(),vF),f(Fe(t,p8e),188)),Ap(this.a,sne,f(Fe(t,c8e),188)),Ap(this.a,gbe,f(Fe(t,d8e),188)),FSe(this.a,(Kt=new _o,Sr(Kt,vF,(zZ(),vbe)),Sr(Kt,sne,bbe),lt(ut(Fe(t,a8e)))&&Sr(Kt,vF,wbe),lt(ut(Fe(t,o8e)))&&Sr(Kt,vF,mbe),Kt)),NG(this.a,t)),T=1/a.c.length,W=new ne(a);W.a<W.c.c.length;){if(z=f(ee(W),44),i.Yg())return;z.If(t,i.bh(T))}for(ue=0,J=0,ve=new Nn(Ee);ve.e!=ve.i.gc();)pe=f(Mn(ve),27),ue=h.Math.max(ue,pe.i+pe.g),J=h.Math.max(J,pe.j+pe.f);LAe(t,new Ke(ke(Oe(Fe(t,(wb(),t4)))),ke(Oe(Fe(t,CD)))),new Ke(ue,J)),mTn(Ee,L),S||WS(t,ke(Oe(Fe(t,t4)))+(L.b+L.c),ke(Oe(Fe(t,CD)))+(L.d+L.a),!1,!0),lt(ut(Fe(t,_be)))||vH((u=new ri((fS(),new uS(t))),u)),i.Tg()},R(qxt,"RectPackingLayoutProvider",1082),N(1458,1,vi,n5),p.If=function(t,i){ZDn(f(t,27),i)},R(w$,"InteractiveNodeReorderer",1458),N(1459,1,Cr,jC),p.Le=function(t,i){return ZTn(f(t,27),f(i,27))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(w$,"InteractiveNodeReorderer/lambda$0$Type",1459),N(398,23,{3:1,34:1,23:1,398:1,196:1},zY),p.bg=function(){switch(this.g){case 0:return new Kd;case 1:return new n5;case 2:return new r5;case 3:return new G8}return null};var mbe,bbe,vbe,wbe,xPt=er(w$,Qfe,398,ar,V2n,w0n),CPt;N(1461,1,vi,G8),p.If=function(t,i){ZCn(f(t,27),i)},R(w$,"MinSizePostProcessor",1461),N(1460,1,vi,r5),p.If=function(t,i){lxn(f(t,27),i)},R(w$,"MinSizePreProcessor",1460),N(1647,1,Cr,q8),p.Le=function(t,i){return jgn(f(t,27),f(i,27))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(w$,"NodeSizeComparator",1647),N(1457,1,vi,Kd),p.If=function(t,i){S2n(f(t,27))},R(w$,"NodeSizeReorderer",1457);var wF,CD,t4,APt,IPt,cne,ybe,Ebe,yF,une,fx;N(385,23,{3:1,34:1,23:1,385:1},lse);var JMe,ZMe,Sbe,QMe=er(qhe,"OptimizationGoal",385,ar,zEn,y0n),OPt;N(844,1,Ng,Pk),p.tf=function(t){Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,n4e),""),"Try box layout first"),"Whether one should check whether the regions are stackable to see whether box layout would do the job. For example, nodes with the same height are not stackable inside a row. Therefore, box layout will perform better and faster."),(mn(),!1)),(bE(),Po)),to),At((wm(),en))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,r4e),""),"Current position of a node in the order of nodes"),"The rectangles are ordered. Normally according to their definition the the model. This option specifies the current position of a node."),rt(-1)),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,i4e),""),"Desired index of node"),"The rectangles are ordered. Normally according to their definition the the model. This option allows to specify a desired position that has preference over the original position."),rt(-1)),Hs),ba),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,o4e),""),"In new Row"),"If set to true this node begins in a new row. Consequently this node cannot be moved in a previous layer during compaction. Width approximation does does not take this into account."),!1),Po),to),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,a4e),""),"Order nodes by height"),"If set to true the nodes will be sorted by their height before computing the layout. The highest node will be in the first position."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,s4e),Vhe),"Width Approximation Strategy"),"Strategy for finding an initial width of the drawing."),r8e),Li),b8e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,c4e),Vhe),"Target Width"),"Option to place the rectangles in the given target width instead of approximating the width using the desired aspect ratio. The padding is not included in this. Meaning a drawing will have width of targetwidth + horizontal padding."),-1),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,u4e),Vhe),"Optimization Goal"),"Optimization goal for approximation of the bounding box given by the first iteration. Determines whether layout is sorted by the maximum scaling, aspect ratio, or area. Depending on the strategy the aspect ratio might be nearly ignored."),n8e),Li),QMe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,l4e),Vhe),"Shift Last Placed."),"When placing a rectangle behind or below the last placed rectangle in the first iteration, it is sometimes possible to shift the rectangle further to the left or right, resulting in less whitespace. True (default) enables the shift and false disables it. Disabling the shift produces a greater approximated area by the first iteration and a layout, when using ONLY the first iteration (default not the case), where it is sometimes impossible to implement a size transformation of rectangles that will fill the bounding box and eliminate empty spaces."),!0),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,d4e),"packing"),Vxt),"Strategy for finding an initial placement on nodes."),e8e),Li),y8e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,f4e),Wxt),"Row Height Reevaluation"),"During the compaction step the height of a row is normally not changed. If this options is set, the blocks of other rows might be added if they exceed the row height. If this is the case the whole row has to be packed again to be optimal regarding the new row height. This option should, therefore, be used with care since it might be computation heavy."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,h4e),Wxt),"Compaction iterations"),"Defines the number of compaction iterations. E.g. if set to 2 the width is initially approximated, then the drawing is compacted and based on the resulting drawing the target width is decreased or increased and a second compaction step is executed and the result compared to the first one. The best run is used based on the scale measure."),rt(1)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,p4e),"whiteSpaceElimination"),"White Space Approximation Strategy"),"Strategy for expanding nodes such that whitespace in the parent is eliminated."),t8e),Li),_8e),At(en)))),Dkt((new o1,t))};var NPt,RPt,PPt,LPt,DPt,MPt,jPt,e8e,$Pt,FPt,t8e,BPt,UPt,n8e,zPt,r8e,HPt;R(qhe,"RectPackingMetaDataProvider",844),N(991,1,Ng,o1),p.tf=function(t){Dkt(t)};var lne,GPt,i8e,Yq,o8e,qPt,Xq,VPt,WPt,KPt,YPt,XPt,_be,a8e,s8e,kbe,c8e,EF,u8e,JPt,SF,l8e,d8e,f8e,h8e,p8e,Tbe;R(qhe,"RectPackingOptions",991),N(992,1,{},cg),p.uf=function(){var t;return t=new Pit,t},p.vf=function(t){},R(qhe,"RectPackingOptions/RectpackingFactory",992),N(1646,1,{},Mut),p.a=0,p.c=!1,R(zL,"AreaApproximation",1646);var g8e=Bi(zL,"BestCandidateFilter");N(658,1,{520:1},yk),p.Fg=function(t,i,a){var u,l,g,m,v,S;for(S=new et,g=Gi,v=new ne(t);v.a<v.c.c.length;)m=f(ee(v),237),g=h.Math.min(g,(m.c+(a.b+a.c))*(m.b+(a.d+a.a)));for(l=new ne(t);l.a<l.c.c.length;)u=f(ee(l),237),(u.c+(a.b+a.c))*(u.b+(a.d+a.a))==g&&vn(S.c,u);return S},R(zL,"AreaFilter",658),N(659,1,{520:1},q2),p.Fg=function(t,i,a){var u,l,g,m,v,S;for(v=new et,S=Gi,m=new ne(t);m.a<m.c.c.length;)g=f(ee(m),237),S=h.Math.min(S,h.Math.abs((g.c+(a.b+a.c))/(g.b+(a.d+a.a))-i));for(l=new ne(t);l.a<l.c.c.length;)u=f(ee(l),237),h.Math.abs((u.c+(a.b+a.c))/(u.b+(a.d+a.a))-i)==S&&vn(v.c,u);return v},R(zL,"AspectRatioFilter",659),N(1408,1,fs,V2),p.pg=function(t){return f(t,27),null},p.If=function(t,i){K5n(f(t,27),i)},R(zL,"GreedyWidthApproximator",1408),N(657,1,{520:1},kv),p.Fg=function(t,i,a){var u,l,g,m,v,S;for(S=new et,g=qo,v=new ne(t);v.a<v.c.c.length;)m=f(ee(v),237),g=h.Math.max(g,AX(m.c+(a.b+a.c),m.b+(a.d+a.a),m.a));for(l=new ne(t);l.a<l.c.c.length;)u=f(ee(l),237),AX(u.c+(a.b+a.c),u.b+(a.d+a.a),u.a)==g&&vn(S.c,u);return S},R(zL,"ScaleMeasureFilter",657),N(1409,1,fs,i5),p.pg=function(t){return f(t,27),null},p.If=function(t,i){YRn(f(t,27),i)},R(zL,"TargetWidthWidthApproximator",1409),N(476,23,{3:1,34:1,23:1,476:1,188:1,196:1},TSe),p.bg=function(){return iyt(this)},p.og=function(){return iyt(this)};var xbe,m8e,b8e=er(zL,"WidthApproximationStrategy",476,ar,jyn,E0n),ZPt;N(1410,1,fs,W2),p.pg=function(t){return f(t,27),null},p.If=function(t,i){j9n(this,f(t,27),i)},R(ree,"Compactor",1410),N(1412,1,fs,V8),p.pg=function(t){return f(t,27),null},p.If=function(t,i){I5n(f(t,27),i)},R(ree,"NoPlacement",1412),N(426,23,{3:1,34:1,23:1,426:1,188:1,196:1},dse),p.bg=function(){return c0t(this)},p.og=function(){return c0t(this)};var Cbe,v8e,w8e,y8e=er(ree,"PackingStrategy",426,ar,HEn,S0n),QPt;N(790,1,{},xSe),p.a=0,p.b=0,p.c=0,p.d=Gi,p.e=0,p.f=Gi,R(ree,"RowFillingAndCompaction",790),N(1411,1,fs,Qg),p.pg=function(t){return f(t,27),null},p.If=function(t,i){r8n(f(t,27),i)},R(ree,"SimplePlacement",1411),N(1413,1,fs,Ep),p.pg=function(t){return f(t,27),null},p.If=function(t,i){this.Gg(f(t,27),i)},p.Gg=function(t,i){r2t(t,i)},R(g4e,"EqualWhitespaceEliminator",1413),N(1414,1413,fs,ob),p.Gg=function(t,i){var a,u,l,g,m;i.Sg("To Aspect Ratio Whitesapce Eliminator",1),m=ke(Oe(Fe(t,(wb(),t4)))),g=ke(Oe(Fe(t,CD))),l=ke(Oe(Fe(t,(Ag(),lne)))),a=ke(Oe(Fe(t,wF))),u=m/g,u<l?(m=g*l,pi(t,t4,m)):(a+=m/l-g,pi(t,wF,a),pi(t,CD,g+a)),r2t(t,i),i.Tg()},R(g4e,"ToAspectratioNodeExpander",1414),N(427,23,{3:1,34:1,23:1,427:1,188:1,196:1},fse),p.bg=function(){return c1t(this)},p.og=function(){return c1t(this)};var E8e,Abe,S8e,_8e=er(g4e,"WhiteSpaceEliminationStrategy",427,ar,GEn,_0n),eLt;N(172,1,{172:1},ele),p.a=0,p.c=!1,p.d=0,p.e=0,p.f=0,p.g=0,p.i=0,p.k=!1,p.o=Gi,p.p=Gi,p.r=0,p.s=0,p.t=0,R(cq,"Block",172),N(208,1,{208:1},iJ),p.a=0,p.b=0,p.d=0,p.e=0,p.f=0,R(cq,"BlockRow",208),N(316,1,{316:1},fTe),p.b=0,p.c=0,p.d=0,p.e=0,p.f=0,R(cq,"BlockStack",316),N(237,1,{237:1},pke,p3e),p.a=0,p.b=0,p.c=0,p.d=0,p.e=0,p.g=0;var hjn=R(cq,"DrawingData",237);N(366,23,{3:1,34:1,23:1,366:1},Uz);var n4,$A,_F,kF,Jq,tLt=er(cq,"DrawingDataDescriptor",366,ar,ISn,C0n),nLt;N(186,1,{186:1},_Te),p.b=0,p.c=0,p.e=0,p.f=0,R(cq,"RectRow",186),N(745,1,{},ACe),p.j=0,R(GT,nxt,745),N(1169,1,{},W8),p.$e=function(t){return Qv(t.a,t.b)},R(GT,XOe,1169),N(1170,1,{},Ftt),p.$e=function(t){return zSn(this.a,t)},R(GT,rxt,1170),N(1171,1,{},Btt),p.$e=function(t){return B3n(this.a,t)},R(GT,ixt,1171),N(1172,1,{},Utt),p.$e=function(t){return $Tn(this.a,t)},R(GT,"ElkGraphImporter/lambda$3$Type",1172),N(1173,1,{},ztt),p.$e=function(t){return f4n(this.a,t)},R(GT,oxt,1173),N(1079,222,JS,Lit),p.kf=function(t,i){var a,u,l,g,m,v,S,T,A,L,F,z;for(mm(t,(mG(),gne))&&(z=dr(Fe(t,(Ble(),G8e))),g=jIe(fj(),z),g&&(m=f(tJ(g.f),222),m.kf(t,i.bh(1)))),pi(t,Dbe,(CH(),pne)),pi(t,Mbe,(bG(),Lbe)),pi(t,jbe,(XH(),mne)),v=f(Fe(t,(Ble(),U8e)),17).a,i.Sg("Overlap removal",1),lt(ut(Fe(t,SLt))),S=new Mo,T=new Htt(S),u=new ACe,a=Lkt(u,t),A=!0,l=0;l<v&&A;){if(lt(ut(Fe(t,z8e)))){if(S.a.$b(),fRn(new Mst(T),a.i),S.a.gc()==0)break;a.e=S}for(pH(this.b),Ap(this.b,(rG(),dne),(g7(),Zq)),Ap(this.b,fne,a.g),Ap(this.b,hne,(xY(),Nbe)),this.a=NG(this.b,a),F=new ne(this.a);F.a<F.c.c.length;)L=f(ee(F),44),L.If(a,i.bh(1));LIn(u,a),A=lt(ut(V(a,(HP(),K6e)))),++l}K_t(u,a),i.Tg()},R(GT,"OverlapRemovalLayoutProvider",1079),N(1080,1,{},Htt),R(GT,"OverlapRemovalLayoutProvider/lambda$0$Type",1080),N(432,23,{3:1,34:1,23:1,432:1},hse);var dne,fne,hne,Ibe=er(GT,"SPOrEPhases",432,ar,qEn,T0n),rLt;N(1179,1,{},Dit),R(GT,"ShrinkTree",1179),N(1081,222,JS,jrt),p.kf=function(t,i){var a,u,l,g,m;mm(t,(mG(),gne))&&(m=dr(Fe(t,gne)),l=jIe(fj(),m),l&&(g=f(tJ(l.f),222),g.kf(t,i.bh(1)))),u=new ACe,a=Lkt(u,t),D6n(this.a,a,i.bh(1)),K_t(u,a)},R(GT,"ShrinkTreeLayoutProvider",1081),N(307,148,{3:1,307:1,102:1,148:1},_ft),p.c=!1,R("org.eclipse.elk.alg.spore.graph","Graph",307),N(475,23,{3:1,34:1,23:1,475:1,188:1,196:1},Iot),p.bg=function(){return D1t(this)},p.og=function(){return D1t(this)};var Obe,k8e=er(qT,Fxt,475,ar,Xwn,x0n),iLt;N(543,23,{3:1,34:1,23:1,543:1,188:1,196:1},Ict),p.bg=function(){return new K2},p.og=function(){return new K2};var Nbe,oLt=er(qT,"OverlapRemovalStrategy",543,ar,Ywn,k0n),aLt;N(425,23,{3:1,34:1,23:1,425:1},CSe);var pne,Rbe,T8e=er(qT,"RootSelection",425,ar,zyn,A0n),sLt;N(325,23,{3:1,34:1,23:1,325:1},zz);var x8e,Pbe,Lbe,C8e,A8e,I8e=er(qT,"SpanningTreeCostFunction",325,ar,OSn,I0n),cLt;N(989,1,Ng,j5),p.tf=function(t){C_t(t)};var O8e,N8e,uLt,lLt,R8e,P8e,Dbe,Mbe,jbe,dLt,fLt,gne;R(qT,"SporeCompactionOptions",989),N(990,1,{},Tv),p.uf=function(){var t;return t=new jrt,t},p.vf=function(t){},R(qT,"SporeCompactionOptions/SporeCompactionFactory",990),N(843,1,Ng,U9),p.tf=function(t){Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,Whe),""),"Underlying Layout Algorithm"),"A layout algorithm that is applied to the graph before it is compacted. If this is null, nothing is applied before compaction."),(bE(),i4)),ft),At((wm(),en))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Xhe),"structure"),"Structure Extraction Strategy"),"This option defines what kind of triangulation or other partitioning of the plane is applied to the vertices."),F8e),Li),q8e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,m4e),Jhe),"Tree Construction Strategy"),"Whether a minimum spanning tree or a maximum spanning tree should be constructed."),j8e),Li),W8e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,b4e),Jhe),"Cost Function for Spanning Tree"),"The cost function is used in the creation of the spanning tree."),M8e),Li),I8e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Khe),Jhe),"Root node for spanning tree construction"),"The identifier of the node that is preferred as the root of the spanning tree. If this is null, the first node is chosen."),null),i4),ft),At(en)))),po(t,Khe,Yhe,vLt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Yhe),Jhe),"Root selection for spanning tree"),"This sets the method used to select a root node for the construction of a spanning tree"),D8e),Li),T8e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,v4e),qNe),Vxt),"This option defines how the compaction is applied."),L8e),Li),k8e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,w4e),qNe),"Orthogonal Compaction"),"Restricts the translation of nodes to orthogonal directions in the compaction phase."),(mn(),!1)),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,y4e),Yxt),"Upper limit for iterations of overlap removal"),null),rt(64)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,E4e),Yxt),"Whether to run a supplementary scanline overlap check."),null),!0),Po),to),At(en)))),NSt((new z9,t)),C_t((new j5,t))};var hLt,L8e,pLt,gLt,mLt,bLt,vLt,wLt,D8e,yLt,M8e,ELt,j8e,$8e,F8e,B8e;R(qT,"SporeMetaDataProvider",843),N(987,1,Ng,z9),p.tf=function(t){NSt(t)};var SLt,U8e,z8e,H8e,_Lt,G8e;R(qT,"SporeOverlapRemovalOptions",987),N(988,1,{},a0),p.uf=function(){var t;return t=new Lit,t},p.vf=function(t){},R(qT,"SporeOverlapRemovalOptions/SporeOverlapFactory",988),N(524,23,{3:1,34:1,23:1,524:1,188:1,196:1},Rdt),p.bg=function(){return M1t(this)},p.og=function(){return M1t(this)};var Zq,q8e=er(qT,"StructureExtractionStrategy",524,ar,Jwn,O0n),kLt;N(424,23,{3:1,34:1,23:1,424:1,188:1,196:1},ASe),p.bg=function(){return H0t(this)},p.og=function(){return H0t(this)};var V8e,mne,W8e=er(qT,"TreeConstructionStrategy",424,ar,Uyn,N0n),TLt;N(1402,1,fs,Gl),p.pg=function(t){return f(t,307),new _o},p.If=function(t,i){jIn(f(t,307),i)},R(Xxt,"DelaunayTriangulationPhase",1402),N(1403,1,xn,Gtt),p.Ad=function(t){Qe(this.a,f(t,68).a)},R(Xxt,"DelaunayTriangulationPhase/lambda$0$Type",1403),N(774,1,fs,_2e),p.pg=function(t){return f(t,307),new _o},p.If=function(t,i){this.Hg(f(t,307),i)},p.Hg=function(t,i){var a,u,l;i.Sg("Minimum spanning tree construction",1),t.d?u=t.d.a:u=f(ct(t.i,0),68).a,lt(ut(V(t,(HP(),F$))))?l=Zde(t.e,u,(a=t.b,a)):l=Zde(t.e,u,t.b),m1t(this,l,t),i.Tg()},R(Zhe,"MinSTPhase",774),N(1405,774,fs,Srt),p.Hg=function(t,i){var a,u,l,g;i.Sg("Maximum spanning tree construction",1),a=new qtt(t),t.d?l=t.d.c:l=f(ct(t.i,0),68).c,lt(ut(V(t,(HP(),F$))))?g=Zde(t.e,l,(u=a,u)):g=Zde(t.e,l,a),m1t(this,g,t),i.Tg()},R(Zhe,"MaxSTPhase",1405),N(1406,1,{},qtt),p.$e=function(t){return xgn(this.a,t)},R(Zhe,"MaxSTPhase/lambda$0$Type",1406),N(1404,1,xn,Vtt),p.Ad=function(t){dmn(this.a,f(t,68))},R(Zhe,"MinSTPhase/lambda$0$Type",1404),N(776,1,fs,K2),p.pg=function(t){return f(t,307),new _o},p.If=function(t,i){YOn(this,f(t,307),i)},p.a=!1,R(Qhe,"GrowTreePhase",776),N(777,1,xn,X_e),p.Ad=function(t){TTn(this.a,this.b,this.c,f(t,224))},R(Qhe,"GrowTreePhase/lambda$0$Type",777),N(1407,1,fs,FC),p.pg=function(t){return f(t,307),new _o},p.If=function(t,i){lAn(this,f(t,307),i)},R(Qhe,"ShrinkTreeCompactionPhase",1407),N(775,1,xn,J_e),p.Ad=function(t){R4n(this.a,this.b,this.c,f(t,224))},R(Qhe,"ShrinkTreeCompactionPhase/lambda$0$Type",775);var K8e=Bi(Lc,"IGraphElementVisitor");N(848,1,{521:1},sht),p.Ig=function(t){var i;i=jPn(this,t),Rc(i,f(bn(this.b,t),102)),R6n(this,t,i)};var xLt,CLt;R(cA,"LayoutConfigurator",848);var pjn=Bi(cA,"LayoutConfigurator/IPropertyHolderOptionFilter");N(922,1,{1978:1},BC),p.Jg=function(t,i){return hR(),!t.nf(i)},R(cA,"LayoutConfigurator/lambda$0$Type",922),N(923,1,{1978:1},BU),p.Jg=function(t,i){return Oit(t,i)},R(cA,"LayoutConfigurator/lambda$1$Type",923),N(921,1,{824:1},$C),p.Kg=function(t,i){return hR(),!t.nf(i)},R(cA,"LayoutConfigurator/lambda$2$Type",921),N(924,1,_r,Sat),p.Mb=function(t){return jwn(this.a,this.b,f(t,1978))},R(cA,"LayoutConfigurator/lambda$3$Type",924),N(846,1,{},Ay),R(cA,"RecursiveGraphLayoutEngine",846),N(223,63,C1,Znt,h0),R(cA,"UnsupportedConfigurationException",223),N(363,63,C1,ON),R(cA,"UnsupportedGraphException",363),N(744,1,{}),R(Lc,"AbstractRandomListAccessor",744),N(438,744,{},Gj),p.Lg=function(){return null},p.d=!0,p.e=!0,p.f=0,R(GL,"AlgorithmAssembler",438),N(1164,1,_r,Y2),p.Mb=function(t){return!!f(t,106)},R(GL,"AlgorithmAssembler/lambda$0$Type",1164),N(1165,1,{},Wtt),p.Kb=function(t){return Upn(this.a,f(t,106))},R(GL,"AlgorithmAssembler/lambda$1$Type",1165),N(1166,1,_r,o5),p.Mb=function(t){return!!f(t,78)},R(GL,"AlgorithmAssembler/lambda$2$Type",1166),N(1167,1,xn,Ktt),p.Ad=function(t){lf(this.a,f(t,78))},R(GL,"AlgorithmAssembler/lambda$3$Type",1167),N(1168,1,xn,_at),p.Ad=function(t){ubn(this.a,this.b,f(t,196))},R(GL,"AlgorithmAssembler/lambda$4$Type",1168),N(1284,1,Cr,zO),p.Le=function(t,i){return myn(f(t,196),f(i,196))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(GL,"EnumBasedFactoryComparator",1284),N(78,744,{78:1},_o),p.Lg=function(){return new Mo},p.a=0,R(GL,"LayoutProcessorConfiguration",78),N(M0,1,{521:1},c3),p.Ig=function(t){HH(ILt,new Ytt(t))};var ALt,ILt,OLt;R(dc,"DeprecatedLayoutOptionReplacer",M0),N(1001,1,xn,K8),p.Ad=function(t){y_n(f(t,173))},R(dc,"DeprecatedLayoutOptionReplacer/lambda$0$Type",1001),N(1002,1,xn,UC),p.Ad=function(t){GCn(f(t,173))},R(dc,"DeprecatedLayoutOptionReplacer/lambda$1$Type",1002),N(1003,1,{},Ytt),p.Wd=function(t,i){cbn(this.a,f(t,144),f(i,40))},R(dc,"DeprecatedLayoutOptionReplacer/lambda$2$Type",1003),N(149,1,{685:1,149:1},zS),p.Fb=function(t){return zTe(this,t)},p.Mg=function(){return this.b},p.Ng=function(){return this.c},p.ve=function(){return this.e},p.Hb=function(){return C0(this.c)},p.Ib=function(){return"Layout Algorithm: "+this.c};var gjn=R(dc,"LayoutAlgorithmData",149);N(285,1,{},ug),R(dc,"LayoutAlgorithmData/Builder",285),N(1004,1,{521:1},a5),p.Ig=function(t){me(t,206)&&!lt(ut(t.mf((Rr(),yne))))&&bDn(f(t,27))},R(dc,"LayoutAlgorithmResolver",1004),N(232,1,{685:1,232:1},nR),p.Fb=function(t){return me(t,232)?Mt(this.b,f(t,232).b):!1},p.Mg=function(){return this.a},p.Ng=function(){return this.b},p.ve=function(){return this.d},p.Hb=function(){return C0(this.b)},p.Ib=function(){return"Layout Type: "+this.b},R(dc,"LayoutCategoryData",232),N(354,1,{},xv),R(dc,"LayoutCategoryData/Builder",354),N(855,1,{},wSt);var $be;R(dc,"LayoutMetaDataService",855),N(856,1,{},bdt),R(dc,"LayoutMetaDataService/Registry",856),N(473,1,{473:1},s5),R(dc,"LayoutMetaDataService/Registry/Triple",473),N(857,1,dA,s0),p.Og=function(){return new ha},R(dc,"LayoutMetaDataService/lambda$0$Type",857),N(858,1,VT,Y8),p.Pg=function(t){return $a(f(t,8))},R(dc,"LayoutMetaDataService/lambda$1$Type",858),N(867,1,dA,xf),p.Og=function(){return new et},R(dc,"LayoutMetaDataService/lambda$10$Type",867),N(868,1,VT,Vu),p.Pg=function(t){return new Kl(f(t,13))},R(dc,"LayoutMetaDataService/lambda$11$Type",868),N(869,1,dA,HO),p.Og=function(){return new ki},R(dc,"LayoutMetaDataService/lambda$12$Type",869),N(870,1,VT,GO),p.Pg=function(t){return cH(f(t,67))},R(dc,"LayoutMetaDataService/lambda$13$Type",870),N(871,1,dA,qO),p.Og=function(){return new Mo},R(dc,"LayoutMetaDataService/lambda$14$Type",871),N(872,1,VT,zC),p.Pg=function(t){return Lce(f(t,50))},R(dc,"LayoutMetaDataService/lambda$15$Type",872),N(873,1,dA,X2),p.Og=function(){return new ub},R(dc,"LayoutMetaDataService/lambda$16$Type",873),N(874,1,VT,fo),p.Pg=function(t){return fJ(f(t,50))},R(dc,"LayoutMetaDataService/lambda$17$Type",874),N(875,1,dA,Ek),p.Og=function(){return new u2e},R(dc,"LayoutMetaDataService/lambda$18$Type",875),N(876,1,VT,VO),p.Pg=function(t){return qut(f(t,156))},R(dc,"LayoutMetaDataService/lambda$19$Type",876),N(859,1,dA,Iy),p.Og=function(){return new fd},R(dc,"LayoutMetaDataService/lambda$2$Type",859),N(860,1,VT,X8),p.Pg=function(t){return new mY(f(t,77))},R(dc,"LayoutMetaDataService/lambda$3$Type",860),N(861,1,dA,HC),p.Og=function(){return new i7},R(dc,"LayoutMetaDataService/lambda$4$Type",861),N(862,1,VT,WO),p.Pg=function(t){return new Wse(f(t,139))},R(dc,"LayoutMetaDataService/lambda$5$Type",862),N(863,1,dA,Cf),p.Og=function(){return new xN},R(dc,"LayoutMetaDataService/lambda$6$Type",863),N(864,1,VT,J8),p.Pg=function(t){return new W_e(f(t,107))},R(dc,"LayoutMetaDataService/lambda$7$Type",864),N(865,1,dA,c5),p.Og=function(){return new JO},R(dc,"LayoutMetaDataService/lambda$8$Type",865),N(866,1,VT,GC),p.Pg=function(t){return new smt(f(t,377))},R(dc,"LayoutMetaDataService/lambda$9$Type",866);var Fbe=Bi(KG,"IProperty");N(21,1,{34:1,685:1,21:1,144:1},bt),p.Dd=function(t){return Amn(this,f(t,144))},p.Fb=function(t){return me(t,21)?Mt(this.f,f(t,21).f):me(t,144)&&Mt(this.f,f(t,144).Ng())},p.Qg=function(){var t;if(me(this.b,4)){if(t=g3e(this.b),t==null)throw ie(new As(eCt+this.f+"'. Make sure it's type is registered with the "+(d1(_V),_V.k)+S4e));return t}else return this.b},p.Mg=function(){return this.d},p.Ng=function(){return this.f},p.ve=function(){return this.i},p.Hb=function(){return C0(this.f)},p.Ib=function(){return"Layout Option: "+this.f},R(dc,"LayoutOptionData",21),N(24,1,{},mt),R(dc,"LayoutOptionData/Builder",24),N(170,23,{3:1,34:1,23:1,170:1},Hz);var Dm,L1,Ao,en,FE,Ob=er(dc,"LayoutOptionData/Target",170,ar,LSn,R0n),NLt;N(282,23,{3:1,34:1,23:1,282:1},jN);var Po,ya,Li,r4,Hs,Nb,i4,Y8e,RLt=er(dc,"LayoutOptionData/Type",282,ar,J_n,P0n),PLt,TF,X8e;N(119,1,{119:1},BN,qh,ZX),p.Fb=function(t){var i;return t==null||!me(t,119)?!1:(i=f(t,119),Pu(this.c,i.c)&&Pu(this.d,i.d)&&Pu(this.b,i.b)&&Pu(this.a,i.a))},p.Hb=function(){return oZ(ce(se(Ro,1),Jt,1,5,[this.c,this.d,this.b,this.a]))},p.Ib=function(){return"Rect[x="+this.c+",y="+this.d+",w="+this.b+",h="+this.a+"]"},p.a=0,p.b=0,p.c=0,p.d=0,R(nq,"ElkRectangle",119),N(8,1,{3:1,4:1,8:1,411:1},ha,Kce,Ke,ja),p.Fb=function(t){return Ebt(this,t)},p.Hb=function(){return FN(this.a)+BAn(FN(this.b))},p.ag=function(t){var i,a,u,l;for(u=0;u<t.length&&o0t((gn(u,t.length),t.charCodeAt(u)),JTt);)++u;for(i=t.length;i>0&&o0t((gn(i-1,t.length),t.charCodeAt(i-1)),ZTt);)--i;if(u>=i)throw ie(new dn("The given string does not contain any numbers."));if(l=DT((ga(u,i,t.length),t.substr(u,i-u)),`,|;|\r|
`),l.length!=2)throw ie(new dn("Exactly two numbers are expected, "+l.length+" were found."));try{this.a=AT(IT(l[0])),this.b=AT(IT(l[1]))}catch(g){throw g=go(g),me(g,130)?(a=g,ie(new dn(QTt+a))):ie(g)}},p.Ib=function(){return"("+this.a+","+this.b+")"},p.a=0,p.b=0;var Zo=R(nq,"KVector",8);N(77,67,{3:1,4:1,20:1,31:1,55:1,16:1,67:1,15:1,77:1,411:1},fd,mY,Act),p.Nc=function(){return cxn(this)},p.ag=function(t){var i,a,u,l,g,m;u=DT(t,`,|;|\\(|\\)|\\[|\\]|\\{|\\}| |	|
`),nf(this);try{for(a=0,g=0,l=0,m=0;a<u.length;)u[a]!=null&&IT(u[a]).length>0&&(g%2==0?l=AT(u[a]):m=AT(u[a]),g>0&&g%2!=0&&Mr(this,new Ke(l,m)),++g),++a}catch(v){throw v=go(v),me(v,130)?(i=v,ie(new dn("The given string does not match the expected format for vectors."+i))):ie(v)}},p.Ib=function(){var t,i,a;for(t=new ef("("),i=tr(this,0);i.b!=i.d.c;)a=f(Zn(i),8),xr(t,a.a+","+a.b),i.b!=i.d.c&&(t.a+="; ");return(t.a+=")",t).a};var J8e=R(nq,"KVectorChain",77);N(254,23,{3:1,34:1,23:1,254:1},T7);var Bbe,bne,vne,Qq,eV,wne,Z8e=er(Dl,"Alignment",254,ar,l_n,L0n),LLt;N(968,1,Ng,im),p.tf=function(t){qSt(t)};var Q8e,Ube,DLt,e9e,t9e,MLt,n9e,jLt,$Lt,r9e,i9e,FLt;R(Dl,"BoxLayouterOptions",968),N(969,1,{},u5),p.uf=function(){var t;return t=new qC,t},p.vf=function(t){},R(Dl,"BoxLayouterOptions/BoxFactory",969),N(296,23,{3:1,34:1,23:1,296:1},x7);var xF,zbe,CF,AF,IF,Hbe,Gbe=er(Dl,"ContentAlignment",296,ar,d_n,D0n),BLt;N(684,1,Ng,a1),p.tf=function(t){Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,nCt),""),"Layout Algorithm"),"Select a specific layout algorithm."),(bE(),i4)),ft),At((wm(),en))))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,rCt),""),"Resolved Layout Algorithm"),"Meta data associated with the selected algorithm."),Nb),gjn),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,lRe),""),"Alignment"),"Alignment of the selected node relative to other nodes; the exact meaning depends on the used algorithm."),o9e),Li),Z8e),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,ML),""),"Aspect Ratio"),"The desired aspect ratio of the drawing, that is the quotient of width by height."),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,T4e),""),"Bend Points"),"A fixed list of bend points for the edge. This is used by the 'Fixed Layout' algorithm to specify a pre-defined routing for an edge. The vector chain must include the source point, any bend points, and the target point, so it must have at least two points."),Nb),J8e),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,KQ),""),"Content Alignment"),"Specifies how the content of a node are aligned. Each node can individually control the alignment of its contents. I.e. if a node should be aligned top left in its parent node, the parent node should specify that option."),s9e),r4),Gbe),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,aq),""),"Debug Mode"),"Whether additional debug information shall be generated."),(mn(),!1)),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Nhe),""),"Direction"),"Overall direction of edges: horizontal (right / left) or vertical (down / up)."),c9e),Li),NF),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,oq),""),"Edge Routing"),"What kind of edge routing style should be applied for the content of a parent node. Algorithms may also set this option to single edges in order to mark them as splines. The bend point list of edges with this option set to SPLINES must be interpreted as control points for a piecewise cubic spline."),d9e),Li),i1e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,_4e),""),"Expand Nodes"),"If active, nodes are expanded to fill the area of their parent."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,WQ),""),"Hierarchy Handling"),"Determines whether separate layout runs are triggered for different compound nodes in a hierarchical graph. Setting a node's hierarchy handling to `INCLUDE_CHILDREN` will lay out that node and all of its descendants in a single layout run, until a descendant is encountered which has its hierarchy handling set to `SEPARATE_CHILDREN`. In general, `SEPARATE_CHILDREN` will ensure that a new layout run is triggered for a node with that setting. Including multiple levels of hierarchy in a single layout run may allow cross-hierarchical edges to be laid out properly. If the root node is set to `INHERIT` (or not set at all), the default behavior is `SEPARATE_CHILDREN`."),h9e),Li),r7e),xi(en,ce(se(Ob,1),Ve,170,0,[Ao]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,zT),""),"Padding"),"The padding to be left to a parent element's border when placing child elements. This can also serve as an output option of a layout algorithm if node size calculation is setup appropriately."),k9e),Nb),b5e),xi(en,ce(se(Ob,1),Ve,170,0,[Ao]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,l$),""),"Interactive"),"Whether the algorithm should be run in interactive mode for the content of a parent node. What this means exactly depends on how the specific algorithm interprets this option. Usually in the interactive mode algorithms try to modify the current layout as little as possible."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,XQ),""),"interactive Layout"),"Whether the graph should be changeable interactively and by setting constraints"),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,d$),""),"Omit Node Micro Layout"),"Node micro layout comprises the computation of node dimensions (if requested), the placement of ports and their labels, and the placement of node labels. The functionality is implemented independent of any specific layout algorithm and shouldn't have any negative impact on the layout algorithm's performance itself. Yet, if any unforeseen behavior occurs, this option allows to deactivate the micro layout."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Yfe),""),"Port Constraints"),"Defines constraints of the position of the ports of a node."),I9e),Li),a7e),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,YQ),""),"Position"),"The position of a node, port, or label. This is used by the 'Fixed Layout' algorithm to specify a pre-defined position."),Nb),Zo),xi(Ao,ce(se(Ob,1),Ve,170,0,[FE,L1]))))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,XG),""),"Priority"),"Defines the priority of an object; its meaning depends on the specific layout algorithm and the context where it is used."),Hs),ba),xi(Ao,ce(se(Ob,1),Ve,170,0,[Dm]))))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,AQ),""),"Randomization Seed"),"Seed used for pseudo-random number generators to control the layout algorithm. If the value is 0, the seed shall be determined pseudo-randomly (e.g. from the system time)."),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,u$),""),"Separate Connected Components"),"Whether each connected component should be processed separately."),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ERe),""),"Junction Points"),"This option is not used as option, but as output of the layout algorithms. It is attached to edges and determines the points where junction symbols should be drawn in order to represent hyperedges with orthogonal routing. Whether such points are computed depends on the chosen layout algorithm and edge routing style. The points are put into the vector chain with no specific order."),m9e),Nb),J8e),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,TRe),""),"Comment Box"),"Whether the node should be regarded as a comment box instead of a regular node. In that case its placement should be similar to how labels are handled. Any edges incident to a comment box specify to which graph elements the comment is related."),!1),Po),to),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,xRe),""),"Hypernode"),"Whether the node should be handled as a hypernode."),!1),Po),to),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,z7n),""),"Label Manager"),"Label managers can shorten labels upon a layout algorithm's request."),Nb),yjn),xi(en,ce(se(Ob,1),Ve,170,0,[L1]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,iCt),""),"Softwrapping Fuzziness"),"Determines the amount of fuzziness to be used when performing softwrapping on labels. The value expresses the percent of overhang that is permitted for each line. If the next line would take up less space than this threshold, it is appended to the current line instead of being placed in a new line."),0),ya),mo),At(L1)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ARe),""),"Margins"),"Margins define additional space around the actual bounds of a graph element. For instance, ports or labels being placed on the outside of a node's border might introduce such a margin. The margin is used to guarantee non-overlap of other graph elements with those ports or labels."),b9e),Nb),m5e),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,cRe),""),"No Layout"),"No layout is done for the associated element. This is used to mark parts of a diagram to avoid their inclusion in the layout graph, or to mark parts of the layout graph to prevent layout engines from processing them. If you wish to exclude the contents of a compound node from automatic layout, while the node itself is still considered on its own layer, use the 'Fixed Layout' algorithm for that node."),!1),Po),to),xi(Ao,ce(se(Ob,1),Ve,170,0,[Dm,FE,L1]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,oCt),""),"Scale Factor"),"The scaling factor to be applied to the corresponding node in recursive layout. It causes the corresponding node's size to be adjusted, and its ports and labels to be sized and placed accordingly after the layout of that node has been determined (and before the node itself and its siblings are arranged). The scaling is not reverted afterwards, so the resulting layout graph contains the adjusted size and position data. This option is currently not supported if 'Layout Hierarchy' is set."),1),ya),mo),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,aCt),""),"Child Area Width"),"The width of the area occupied by the laid out children of a node."),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,sCt),""),"Child Area Height"),"The height of the area occupied by the laid out children of a node."),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,JG),""),Jxt),"Turns topdown layout on and off. If this option is enabled, hierarchical layout will be computed first for the root node and then for its children recursively. Layouts are then scaled down to fit the area provided by their parents. Graphs must follow a certain structure for topdown layout to work properly. {@link TopdownNodeTypes.PARALLEL_NODE} nodes must have children of type {@link TopdownNodeTypes.HIERARCHICAL_NODE} and must define {@link topdown.hierarchicalNodeWidth} and {@link topdown.hierarchicalNodeAspectRatio} for their children. Furthermore they need to be laid out using an algorithm that is a {@link TopdownLayoutProvider}. Hierarchical nodes can also be parents of other hierarchical nodes and can optionally use a {@link TopdownSizeApproximator} to dynamically set sizes during topdown layout. In this case {@link topdown.hierarchicalNodeWidth} and {@link topdown.hierarchicalNodeAspectRatio} should be set on the node itself rather than the parent. The values are then used by the size approximator as base values. Hierarchical nodes require the layout option {@link nodeSize.fixedGraphSize} to be true to prevent the algorithm used there from resizing the hierarchical node. This option is not supported if 'Hierarchy Handling' is set to 'INCLUDE_CHILDREN'"),!1),Po),to),At(en)))),po(t,JG,ZS,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,cCt),""),"Animate"),"Whether the shift from the old layout to the new computed layout shall be animated."),!0),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,uCt),""),"Animation Time Factor"),"Factor for computation of animation time. The higher the value, the longer the animation time. If the value is 0, the resulting time is always equal to the minimum defined by 'Minimal Animation Time'."),rt(100)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,lCt),""),"Layout Ancestors"),"Whether the hierarchy levels on the path from the selected element to the root of the diagram shall be included in the layout process."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,dCt),""),"Maximal Animation Time"),"The maximal time for animations, in milliseconds."),rt(4e3)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,fCt),""),"Minimal Animation Time"),"The minimal time for animations, in milliseconds."),rt(400)),Hs),ba),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,hCt),""),"Progress Bar"),"Whether a progress bar shall be displayed during layout computations."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,pCt),""),"Validate Graph"),"Whether the graph shall be validated before any layout algorithm is applied. If this option is enabled and at least one error is found, the layout process is aborted and a message is shown to the user."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,gCt),""),"Validate Options"),"Whether layout options shall be validated before any layout algorithm is applied. If this option is enabled and at least one error is found, the layout process is aborted and a message is shown to the user."),!0),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,mCt),""),"Zoom to Fit"),"Whether the zoom level shall be set to view the whole diagram after layout."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,k4e),"box"),"Box Layout Mode"),"Configures the packing mode used by the {@link BoxLayoutProvider}. If SIMPLE is not required (neither priorities are used nor the interactive mode), GROUP_DEC can improve the packing and decrease the area. GROUP_MIXED and GROUP_INC may, in very specific scenarios, work better."),a9e),Li),v7e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,bCt),"json"),"Shape Coords"),"For layouts transferred into JSON graphs, specify the coordinate system to be used for nodes, ports, and labels of nodes and ports."),g9e),Li),u7e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,vCt),"json"),"Edge Coords"),"For layouts transferred into JSON graphs, specify the coordinate system to be used for edge route points and edge labels."),p9e),Li),G9e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,XNe),Pg),"Comment Comment Spacing"),"Spacing to be preserved between a comment box and other comment boxes connected to the same node. The space left between comment boxes of different nodes is controlled by the node-node spacing."),10),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,JNe),Pg),"Comment Node Spacing"),"Spacing to be preserved between a node and its connected comment boxes. The space left between a node and the comments of another node is controlled by the node-node spacing."),10),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ZNe),Pg),"Components Spacing"),"Spacing to be preserved between pairs of connected components. This option is only relevant if 'separateConnectedComponents' is activated."),20),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,QNe),Pg),"Edge Spacing"),"Spacing to be preserved between any two edges. Note that while this can somewhat easily be satisfied for the segments of orthogonally drawn edges, it is harder for general polylines or splines."),10),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Kfe),Pg),"Edge Label Spacing"),"The minimal distance to be preserved between a label and the edge it is associated with. Note that the placement of a label is influenced by the 'edgelabels.placement' option."),2),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Ohe),Pg),"Edge Node Spacing"),"Spacing to be preserved between nodes and edges."),10),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,eRe),Pg),"Label Spacing"),"Determines the amount of space to be left between two labels of the same graph element."),0),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,rRe),Pg),"Label Node Spacing"),"Spacing to be preserved between labels and the border of node they are associated with. Note that the placement of a label is influenced by the 'nodelabels.placement' option."),5),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,tRe),Pg),"Horizontal spacing between Label and Port"),"Horizontal spacing to be preserved between labels and the ports they are associated with. Note that the placement of a label is influenced by the 'portlabels.placement' option."),1),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,nRe),Pg),"Vertical spacing between Label and Port"),"Vertical spacing to be preserved between labels and the ports they are associated with. Note that the placement of a label is influenced by the 'portlabels.placement' option."),1),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,UT),Pg),"Node Spacing"),"The minimal distance to be preserved between each two nodes."),20),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,iRe),Pg),"Node Self Loop Spacing"),"Spacing to be preserved between a node and its self loops."),10),ya),mo),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,oRe),Pg),"Port Spacing"),"Spacing between pairs of ports of the same node."),10),ya),mo),xi(en,ce(se(Ob,1),Ve,170,0,[Ao]))))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,aRe),Pg),"Individual Spacing"),"Allows to specify individual spacing values for graph elements that shall be different from the value specified for the element's parent."),Nb),PDt),xi(Ao,ce(se(Ob,1),Ve,170,0,[Dm,FE,L1]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,IRe),Pg),"Additional Port Space"),"Additional space around the sets of ports on each node side. For each side of a node, this option can reserve additional space before and after the ports on each side. For example, a top spacing of 20 makes sure that the first port on the western and eastern side is 20 units away from the northern border."),F9e),Nb),m5e),At(en)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,Phe),ECt),"Layout Partition"),"Partition to which the node belongs. This requires Layout Partitioning to be active. Nodes with lower partition IDs will appear to the left of nodes with higher partition IDs (assuming a left-to-right layout direction)."),Hs),ba),xi(en,ce(se(Ob,1),Ve,170,0,[Ao]))))),po(t,Phe,Rhe,ZLt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,Rhe),ECt),"Layout Partitioning"),"Whether to activate partitioned layout. This will allow to group nodes through the Layout Partition option. a pair of nodes with different partition indices is then placed such that the node with lower index is placed to the left of the other node (with left-to-right layout direction). Depending on the layout algorithm, this may only be guaranteed to work if all nodes have a layout partition configured, or at least if edges that cross partitions are not part of a partition-crossing cycle."),T9e),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,hRe),SCt),"Node Label Padding"),"Define padding for node labels that are placed inside of a node."),w9e),Nb),b5e),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,$L),SCt),"Node Label Placement"),"Hints for where node labels are to be placed; if empty, the node label's position is not modified."),y9e),r4),ps),xi(Ao,ce(se(Ob,1),Ve,170,0,[L1]))))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,mRe),oee),"Port Alignment"),"Defines the default port distribution for a node. May be overridden for each side individually."),C9e),Li),DF),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,bRe),oee),"Port Alignment (North)"),"Defines how ports on the northern side are placed, overriding the node's general port alignment."),Li),DF),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,vRe),oee),"Port Alignment (South)"),"Defines how ports on the southern side are placed, overriding the node's general port alignment."),Li),DF),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,wRe),oee),"Port Alignment (West)"),"Defines how ports on the western side are placed, overriding the node's general port alignment."),Li),DF),At(Ao)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,yRe),oee),"Port Alignment (East)"),"Defines how ports on the eastern side are placed, overriding the node's general port alignment."),Li),DF),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,uA),tpe),"Node Size Constraints"),"What should be taken into account when calculating a node's size. Empty size constraints specify that a node's size is already fixed and should not be changed."),E9e),r4),$F),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,LR),tpe),"Node Size Options"),"Options modifying the behavior of the size constraints set on a node. Each member of the set specifies something that should be taken into account when calculating node sizes. The empty set corresponds to no further modifications."),_9e),r4),l7e),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,DR),tpe),"Node Size Minimum"),"The minimal size to which a node can be reduced."),S9e),Nb),Zo),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,jL),tpe),"Fixed Graph Size"),"By default, the fixed layout provider will enlarge a graph until it is large enough to contain its children. If this option is set, it won't do so."),!1),Po),to),At(en)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,_Re),Ihe),"Edge Label Placement"),"Gives a hint on where to put edge labels."),u9e),Li),q9e),At(L1)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,IQ),Ihe),"Inline Edge Labels"),"If true, an edge label is placed directly on its edge. May only apply to center edge labels. This kind of label placement is only advisable if the label's rendering is such that it is not crossed by its edge and thus stays legible."),!1),Po),to),At(L1)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,H7n),"font"),"Font Name"),"Font name used for a label."),i4),ft),At(L1)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,wCt),"font"),"Font Size"),"Font size used for a label."),Hs),ba),At(L1)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,CRe),npe),"Port Anchor Offset"),"The offset to the port position where connections shall be attached."),Nb),Zo),At(FE)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,kRe),npe),"Port Index"),"The index of a port in the fixed order around a node. The order is assumed as clockwise, starting with the leftmost port on the top side. This option must be set if 'Port Constraints' is set to FIXED_ORDER and no specific positions are given for the ports. Additionally, the option 'Port Side' must be defined in this case."),Hs),ba),At(FE)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,uRe),npe),"Port Side"),"The side of a node on which a port is situated. This option must be set if 'Port Constraints' is set to FIXED_SIDE or FIXED_ORDER and no specific positions are given for the ports."),R9e),Li),Ja),At(FE)))),Tt(t,new bt(_t(St(kt(vt(Et(wt(yt(new mt,sRe),npe),"Port Border Offset"),"The offset of ports on the node border. With a positive offset the port is moved outside of the node, while with a negative offset the port is moved towards the inside. An offset of 0 means that the port is placed directly on the node border, i.e. if the port side is north, the port's south border touches the nodes's north border; if the port side is east, the port's west border touches the nodes's east border; if the port side is south, the port's north border touches the node's south border; if the port side is west, the port's east border touches the node's west border."),ya),mo),At(FE)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,FL),A4e),"Port Label Placement"),"Decides on a placement method for port labels; if empty, the node label's position is not modified."),O9e),r4),Tne),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,pRe),A4e),"Port Labels Next to Port"),"Use 'portLabels.placement': NEXT_TO_PORT_OF_POSSIBLE."),!1),Po),to),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,gRe),A4e),"Treat Port Labels as Group"),"If this option is true (default), the labels of a port will be treated as a group when it comes to centering them next to their port. If this option is false, only the first label will be centered next to the port, with the others being placed below. This only applies to labels of eastern and western ports and will have no effect if labels are not placed next to their port."),!0),Po),to),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ZG),uq),"Topdown Scale Factor"),"The scaling factor to be applied to the nodes laid out within the node in recursive topdown layout. The difference to 'Scale Factor' is that the node itself is not scaled. This value has to be set on hierarchical nodes."),1),ya),mo),At(en)))),po(t,ZG,ZS,sDt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,x4e),uq),"Topdown Size Approximator"),"The size approximator to be used to set sizes of hierarchical nodes during topdown layout. The default value is null, which results in nodes keeping whatever size is defined for them e.g. through parent parallel node or by manually setting the size."),null),Li),Nne),At(Ao)))),po(t,x4e,ZS,cDt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,QG),uq),"Topdown Hierarchical Node Width"),"The fixed size of a hierarchical node when using topdown layout. If this value is set on a parallel node it applies to its children, when set on a hierarchical node it applies to the node itself."),150),ya),mo),xi(en,ce(se(Ob,1),Ve,170,0,[Ao]))))),po(t,QG,ZS,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,eq),uq),"Topdown Hierarchical Node Aspect Ratio"),"The fixed aspect ratio of a hierarchical node when using topdown layout. Default is 1/sqrt(2). If this value is set on a parallel node it applies to its children, when set on a hierarchical node it applies to the node itself."),1.414),ya),mo),xi(en,ce(se(Ob,1),Ve,170,0,[Ao]))))),po(t,eq,ZS,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,ZS),uq),"Topdown Node Type"),"The different node types used for topdown layout. If the node type is set to {@link TopdownNodeTypes.PARALLEL_NODE} the algorithm must be set to a {@link TopdownLayoutProvider} such as {@link TopdownPacking}. The {@link nodeSize.fixedGraphSize} option is technically only required for hierarchical nodes."),null),Li),f7e),At(Ao)))),po(t,ZS,jL,null),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,C4e),uq),"Topdown Scale Cap"),"Determines the upper limit for the topdown scale factor. The default value is 1.0 which ensures that nested children never end up appearing larger than their parents in terms of unit sizes such as the font size. If the limit is larger, nodes will fully utilize the available space, but it is counteriniuitive for inner nodes to have a larger scale than outer nodes."),1),ya),mo),At(en)))),po(t,C4e,ZS,aDt),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,dRe),_Ct),"Activate Inside Self Loops"),"Whether this node allows to route self loops inside of it instead of around it. If set to true, this will make the node a compound node if it isn't already, and will require the layout algorithm to support compound nodes with hierarchical ports."),!1),Po),to),At(Ao)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,fRe),_Ct),"Inside Self Loop"),"Whether a self loop should be routed inside a node instead of around that node."),!1),Po),to),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,SRe),"edge"),"Edge Thickness"),"The thickness of an edge. This is a hint on the line width used to draw an edge, possibly requiring more space to be reserved for it."),1),ya),mo),At(Dm)))),Tt(t,new bt(_t(St(kt(Lt(vt(Et(wt(yt(new mt,yCt),"edge"),"Edge Type"),"The type of an edge. This is usually used for UML class diagrams, where associations must be handled differently from generalizations."),f9e),Li),J9e),At(Dm)))),m7(t,new nR(d7(nP(rP(new xv,wn),"Layered"),'The layer-based method was introduced by Sugiyama, Tagawa and Toda in 1981. It emphasizes the direction of edges by pointing as many edges as possible into the same direction. The nodes are arranged in layers, which are sometimes called "hierarchies", and then reordered such that the number of edge crossings is minimized. Afterwards, concrete coordinates are computed for the nodes and edge bend points.'))),m7(t,new nR(d7(nP(rP(new xv,"org.eclipse.elk.orthogonal"),"Orthogonal"),`Orthogonal methods that follow the "topology-shape-metrics" approach by Batini, Nardelli and Tamassia '86. The first phase determines the topology of the drawing by applying a planarization technique, which results in a planar representation of the graph. The orthogonal shape is computed in the second phase, which aims at minimizing the number of edge bends, and is called orthogonalization. The third phase leads to concrete coordinates for nodes and edge bend points by applying a compaction method, thus defining the metrics.`))),m7(t,new nR(d7(nP(rP(new xv,Zu),"Force"),"Layout algorithms that follow physical analogies by simulating a system of attractive and repulsive forces. The first successful method of this kind was proposed by Eades in 1984."))),m7(t,new nR(d7(nP(rP(new xv,"org.eclipse.elk.circle"),"Circle"),"Circular layout algorithms emphasize cycles or biconnected components of a graph by arranging them in circles. This is useful if a drawing is desired where such components are clearly grouped, or where cycles are shown as prominent OPTIONS of the graph."))),m7(t,new nR(d7(nP(rP(new xv,Dxt),"Tree"),"Specialized layout methods for trees, i.e. acyclic graphs. The regular structure of graphs that have no undirected cycles can be emphasized using an algorithm of this type."))),m7(t,new nR(d7(nP(rP(new xv,"org.eclipse.elk.planar"),"Planar"),"Algorithms that require a planar or upward planar graph. Most of these algorithms are theoretically interesting, but not practically usable."))),m7(t,new nR(d7(nP(rP(new xv,Uf),"Radial"),"Radial layout algorithms usually position the nodes of the graph on concentric circles."))),ISt((new fa,t)),qSt((new im,t)),rSt((new d0,t))};var AD,ULt,o9e,ID,zLt,HLt,a9e,FA,BA,GLt,tV,s9e,nV,BE,c9e,qbe,Vbe,u9e,l9e,d9e,qLt,f9e,VLt,UA,h9e,WLt,rV,Wbe,iV,Kbe,KLt,p9e,YLt,g9e,zA,m9e,OD,b9e,v9e,w9e,HA,y9e,UE,E9e,hx,GA,S9e,Ow,_9e,yne,oV,V0,k9e,XLt,T9e,JLt,ZLt,x9e,C9e,Ybe,Xbe,Jbe,Zbe,A9e,Sd,OF,I9e,Qbe,e1e,px,O9e,N9e,qA,R9e,o4,aV,t1e,VA,QLt,n1e,eDt,tDt,nDt,rDt,P9e,L9e,a4,D9e,Ene,M9e,j9e,zE,iDt,$9e,F9e,B9e,sV,ND,RD,s4,oDt,aDt,Sne,sDt,U9e,cDt;R(Dl,"CoreOptions",684),N(86,23,{3:1,34:1,23:1,86:1},Gz);var jg,Ds,Gs,$g,Wf,NF=er(Dl,"Direction",86,ar,wSn,M0n),uDt;N(275,23,{3:1,34:1,23:1,275:1},HY);var _ne,cV,z9e,H9e,G9e=er(Dl,"EdgeCoords",275,ar,K2n,j0n),lDt;N(276,23,{3:1,34:1,23:1,276:1},pse);var PD,gx,LD,q9e=er(Dl,"EdgeLabelPlacement",276,ar,WEn,$0n),dDt;N(220,23,{3:1,34:1,23:1,220:1},GY);var DD,uV,c4,r1e,i1e=er(Dl,"EdgeRouting",220,ar,Y2n,F0n),fDt;N(322,23,{3:1,34:1,23:1,322:1},C7);var V9e,W9e,K9e,Y9e,o1e,X9e,J9e=er(Dl,"EdgeType",322,ar,f_n,B0n),hDt;N(966,1,Ng,fa),p.tf=function(t){ISt(t)};var Z9e,Q9e,e7e,t7e,pDt,n7e,RF;R(Dl,"FixedLayouterOptions",966),N(967,1,{},YO),p.uf=function(){var t;return t=new e9,t},p.vf=function(t){},R(Dl,"FixedLayouterOptions/FixedFactory",967),N(342,23,{3:1,34:1,23:1,342:1},gse);var W0,kne,PF,r7e=er(Dl,"HierarchyHandling",342,ar,VEn,U0n),gDt;N(289,23,{3:1,34:1,23:1,289:1},qY);var Rb,Nw,lV,dV,mDt=er(Dl,"LabelSide",289,ar,X2n,z0n),bDt;N(94,23,{3:1,34:1,23:1,94:1},y3);var D1,sp,Eh,cp,bf,up,Sh,Pb,lp,ps=er(Dl,"NodeLabelPlacement",94,ar,gkn,H0n),vDt;N(255,23,{3:1,34:1,23:1,255:1},qz);var i7e,LF,Rw,o7e,fV,DF=er(Dl,"PortAlignment",255,ar,DSn,G0n),wDt;N(100,23,{3:1,34:1,23:1,100:1},A7);var HE,uu,Lb,MD,Fg,Pw,a7e=er(Dl,"PortConstraints",100,ar,h_n,q0n),yDt;N(277,23,{3:1,34:1,23:1,277:1},I7);var MF,jF,M1,hV,Lw,u4,Tne=er(Dl,"PortLabelPlacement",277,ar,p_n,V0n),EDt;N(64,23,{3:1,34:1,23:1,64:1},Vz);var _n,hn,Kf,Yf,vl,tl,Bg,dp,nd,jl,Au,rd,wl,yl,fp,vf,wf,_h,Bn,fc,pn,Ja=er(Dl,"PortSide",64,ar,ySn,W0n),SDt;N(970,1,Ng,d0),p.tf=function(t){rSt(t)};var _Dt,kDt,s7e,TDt,xDt;R(Dl,"RandomLayouterOptions",970),N(971,1,{},XO),p.uf=function(){var t;return t=new KO,t},p.vf=function(t){},R(Dl,"RandomLayouterOptions/RandomFactory",971),N(297,23,{3:1,34:1,23:1,297:1},mse);var pV,a1e,c7e,u7e=er(Dl,"ShapeCoords",297,ar,YEn,K0n),CDt;N(378,23,{3:1,34:1,23:1,378:1},VY);var mx,gV,mV,GE,$F=er(Dl,"SizeConstraint",378,ar,W2n,Y0n),ADt;N(264,23,{3:1,34:1,23:1,264:1},E3);var bV,xne,jD,s1e,vV,FF,Cne,Ane,Ine,l7e=er(Dl,"SizeOptions",264,ar,_kn,X0n),IDt;N(278,23,{3:1,34:1,23:1,278:1},bse);var bx,d7e,One,f7e=er(Dl,"TopdownNodeTypes",278,ar,KEn,J0n),ODt;N(343,23,O4e);var h7e,p7e,Nne=er(Dl,"TopdownSizeApproximator",343,ar,Vyn,Z0n);N(964,343,O4e,flt),p.Rg=function(t){return Uvt(t)},er(Dl,"TopdownSizeApproximator/1",964,Nne,null,null),N(965,343,O4e,Rlt),p.Rg=function(t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct,Kt;for(i=f(Fe(t,(Rr(),VA)),149),nt=(Bv(),z=new o7,z),_G(nt,t),xt=new Wn,g=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));g.e!=g.i.gc();)u=f(Mn(g),27),ve=(F=new o7,F),eQ(ve,nt),_G(ve,u),Kt=Uvt(u),x3(ve,h.Math.max(u.g,Kt.a),h.Math.max(u.f,Kt.b)),ml(xt.f,u,ve);for(l=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));l.e!=l.i.gc();)for(u=f(Mn(l),27),A=new Nn((!u.e&&(u.e=new Qt(vo,u,7,4)),u.e));A.e!=A.i.gc();)T=f(Mn(A),85),Le=f(ac(Is(xt.f,u)),27),He=f(bn(xt,be((!T.c&&(T.c=new Qt(Yn,T,5,8)),T.c),0)),27),Ee=(L=new WC,L),Qn((!Ee.b&&(Ee.b=new Qt(Yn,Ee,4,7)),Ee.b),Le),Qn((!Ee.c&&(Ee.c=new Qt(Yn,Ee,5,8)),Ee.c),He),QZ(Ee,so(Le)),_G(Ee,T);J=f(tJ(i.f),222);try{J.kf(nt,new Av),xdt(i.f,J)}catch(tn){throw tn=go(tn),me(tn,103)?(W=tn,ie(W)):ie(tn)}return mm(nt,BA)||mm(nt,FA)||SOe(nt),S=ke(Oe(Fe(nt,BA))),v=ke(Oe(Fe(nt,FA))),m=S/v,a=ke(Oe(Fe(nt,ND)))*h.Math.sqrt((!nt.a&&(nt.a=new Be(Ur,nt,10,11)),nt.a).i),Ct=f(Fe(nt,V0),107),pe=Ct.b+Ct.c+1,ue=Ct.d+Ct.a+1,new Ke(h.Math.max(pe,a),h.Math.max(ue,a/m))},er(Dl,"TopdownSizeApproximator/2",965,Nne,null,null);var NDt;N(340,1,{847:1},CN),p.Sg=function(t,i){return tyt(this,t,i)},p.Tg=function(){Tyt(this)},p.Ug=function(){return this.q},p.Vg=function(){return this.f?Dce(this.f):null},p.Wg=function(){return Dce(this.a)},p.Xg=function(){return this.p},p.Yg=function(){return!1},p.Zg=function(){return this.n},p.$g=function(){return this.p!=null&&!this.b},p._g=function(t){var i;this.n&&(i=t,Qe(this.f,i))},p.ah=function(t,i){var a,u;this.n&&t&&f2n(this,(a=new fdt,u=$de(a,t),V8n(a),u),(iZ(),u1e))},p.bh=function(t){var i;return this.b?null:(i=W_n(this,this.g),Mr(this.a,i),i.i=this,this.d=t,i)},p.dh=function(t){t>0&&!this.b&&Hxe(this,t)},p.b=!1,p.c=0,p.d=-1,p.e=null,p.f=null,p.g=-1,p.j=!1,p.k=!1,p.n=!1,p.o=0,p.q=0,p.r=0,R(Lc,"BasicProgressMonitor",340),N(701,222,JS,qC),p.kf=function(t,i){c_t(t,i)},R(Lc,"BoxLayoutProvider",701),N(960,1,Cr,Xtt),p.Le=function(t,i){return i5n(this,f(t,27),f(i,27))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},p.a=!1,R(Lc,"BoxLayoutProvider/1",960),N(162,1,{162:1},LJ,Oct),p.Ib=function(){return this.c?MIe(this.c):Cg(this.b)},R(Lc,"BoxLayoutProvider/Group",162),N(321,23,{3:1,34:1,23:1,321:1},WY);var g7e,m7e,b7e,c1e,v7e=er(Lc,"BoxLayoutProvider/PackingMode",321,ar,J2n,Q0n),RDt;N(961,1,Cr,Z8),p.Le=function(t,i){return gyn(f(t,162),f(i,162))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Lc,"BoxLayoutProvider/lambda$0$Type",961),N(962,1,Cr,l5),p.Le=function(t,i){return oyn(f(t,162),f(i,162))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Lc,"BoxLayoutProvider/lambda$1$Type",962),N(963,1,Cr,Cv),p.Le=function(t,i){return ayn(f(t,162),f(i,162))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(Lc,"BoxLayoutProvider/lambda$2$Type",963),N(1323,1,{824:1},t1),p.Kg=function(t,i){return _Y(),!me(i,173)||Oit((hR(),f(t,173)),i)},R(Lc,"ElkSpacings/AbstractSpacingsBuilder/lambda$0$Type",1323),N(1324,1,xn,Jtt),p.Ad=function(t){dxn(this.a,f(t,144))},R(Lc,"ElkSpacings/AbstractSpacingsBuilder/lambda$1$Type",1324),N(1325,1,xn,Sp),p.Ad=function(t){f(t,102),_Y()},R(Lc,"ElkSpacings/AbstractSpacingsBuilder/lambda$2$Type",1325),N(1329,1,xn,Ztt),p.Ad=function(t){Lkn(this.a,f(t,102))},R(Lc,"ElkSpacings/AbstractSpacingsBuilder/lambda$3$Type",1329),N(1327,1,_r,Tat),p.Mb=function(t){return KTn(this.a,this.b,f(t,144))},R(Lc,"ElkSpacings/AbstractSpacingsBuilder/lambda$4$Type",1327),N(1326,1,_r,xat),p.Mb=function(t){return nbn(this.a,this.b,f(t,824))},R(Lc,"ElkSpacings/AbstractSpacingsBuilder/lambda$5$Type",1326),N(1328,1,xn,Cat),p.Ad=function(t){own(this.a,this.b,f(t,144))},R(Lc,"ElkSpacings/AbstractSpacingsBuilder/lambda$6$Type",1328),N(925,1,{},Q8),p.Kb=function(t){return Sst(t)},p.Fb=function(t){return this===t},R(Lc,"ElkUtil/lambda$0$Type",925),N(926,1,xn,Aat),p.Ad=function(t){h4n(this.a,this.b,f(t,85))},p.a=0,p.b=0,R(Lc,"ElkUtil/lambda$1$Type",926),N(927,1,xn,Iat),p.Ad=function(t){cpn(this.a,this.b,f(t,168))},p.a=0,p.b=0,R(Lc,"ElkUtil/lambda$2$Type",927),N(928,1,xn,Oat),p.Ad=function(t){Xgn(this.a,this.b,f(t,153))},p.a=0,p.b=0,R(Lc,"ElkUtil/lambda$3$Type",928),N(929,1,xn,Qtt),p.Ad=function(t){Cvn(this.a,f(t,369))},R(Lc,"ElkUtil/lambda$4$Type",929),N(326,1,{34:1,326:1},$hn),p.Dd=function(t){return Imn(this,f(t,241))},p.Fb=function(t){var i;return me(t,326)?(i=f(t,326),this.a==i.a):!1},p.Hb=function(){return Aa(this.a)},p.Ib=function(){return this.a+" (exclusive)"},p.a=0,R(Lc,"ExclusiveBounds/ExclusiveLowerBound",326),N(1083,222,JS,e9),p.kf=function(t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe,ve,Ee,Le,He,nt,xt,Ct;for(i.Sg("Fixed Layout",1),g=f(Fe(t,(Rr(),l9e)),220),L=0,F=0,ve=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));ve.e!=ve.i.gc();){for(ue=f(Mn(ve),27),Ct=f(Fe(ue,(rZ(),RF)),8),Ct&&(Of(ue,Ct.a,Ct.b),f(Fe(ue,Q9e),181).Gc((Ud(),mx))&&(z=f(Fe(ue,t7e),8),z.a>0&&z.b>0&&WS(ue,z.a,z.b,!0,!0))),L=h.Math.max(L,ue.i+ue.g),F=h.Math.max(F,ue.j+ue.f),T=new Nn((!ue.n&&(ue.n=new Be(hc,ue,1,7)),ue.n));T.e!=T.i.gc();)v=f(Mn(T),153),Ct=f(Fe(v,RF),8),Ct&&Of(v,Ct.a,Ct.b),L=h.Math.max(L,ue.i+v.i+v.g),F=h.Math.max(F,ue.j+v.j+v.f);for(He=new Nn((!ue.c&&(ue.c=new Be(_d,ue,9,9)),ue.c));He.e!=He.i.gc();)for(Le=f(Mn(He),124),Ct=f(Fe(Le,RF),8),Ct&&Of(Le,Ct.a,Ct.b),nt=ue.i+Le.i,xt=ue.j+Le.j,L=h.Math.max(L,nt+Le.g),F=h.Math.max(F,xt+Le.f),S=new Nn((!Le.n&&(Le.n=new Be(hc,Le,1,7)),Le.n));S.e!=S.i.gc();)v=f(Mn(S),153),Ct=f(Fe(v,RF),8),Ct&&Of(v,Ct.a,Ct.b),L=h.Math.max(L,nt+v.i+v.g),F=h.Math.max(F,xt+v.j+v.f);for(l=new Tn(Cn(lw(ue).a.Jc(),new B));Kn(l);)a=f(Dn(l),85),A=wkt(a),L=h.Math.max(L,A.a),F=h.Math.max(F,A.b);for(u=new Tn(Cn(kde(ue).a.Jc(),new B));Kn(u);)a=f(Dn(u),85),so(lde(a))!=t&&(A=wkt(a),L=h.Math.max(L,A.a),F=h.Math.max(F,A.b))}if(g==(k1(),DD))for(pe=new Nn((!t.a&&(t.a=new Be(Ur,t,10,11)),t.a));pe.e!=pe.i.gc();)for(ue=f(Mn(pe),27),u=new Tn(Cn(lw(ue).a.Jc(),new B));Kn(u);)a=f(Dn(u),85),m=tDn(a),m.b==0?pi(a,zA,null):pi(a,zA,m);lt(ut(Fe(t,(rZ(),e7e))))||(Ee=f(Fe(t,pDt),107),J=L+Ee.b+Ee.c,W=F+Ee.d+Ee.a,WS(t,J,W,!0,!0)),i.Tg()},R(Lc,"FixedLayoutProvider",1083),N(377,148,{3:1,411:1,377:1,102:1,148:1},JO,smt),p.ag=function(t){var i,a,u,l,g,m,v,S,T;if(t)try{for(S=DT(t,";,;"),g=S,m=0,v=g.length;m<v;++m){if(l=g[m],a=DT(l,"\\:"),u=pOe(fj(),a[0]),!u)throw ie(new dn("Invalid option id: "+a[0]));if(T=fOe(u,a[1]),T==null)throw ie(new dn("Invalid option value: "+a[1]));T==null?(!this.q&&(this.q=new Wn),aR(this.q,u)):(!this.q&&(this.q=new Wn),qr(this.q,u,T))}}catch(A){throw A=go(A),me(A,103)?(i=A,ie(new W1t(i))):ie(A)}},p.Ib=function(){var t;return t=dr($d(Lu((this.q?this.q:(Ht(),Ht(),Cb)).vc().Mc(),new pu),N_n(new Wut,new Fc,new qi,new yi,ce(se(Ql,1),Ve,131,0,[])))),t};var PDt=R(Lc,"IndividualSpacings",377);N(959,1,{},pu),p.Kb=function(t){return pyn(f(t,43))},R(Lc,"IndividualSpacings/lambda$0$Type",959),N(702,1,{},Slt),p.c=0,R(Lc,"InstancePool",702),N(1781,1,{},VC),R(Lc,"LoggedGraph",1781),N(404,23,{3:1,34:1,23:1,404:1},YY);var w7e,u1e,y7e,E7e,LDt=er(Lc,"LoggedGraph/Type",404,ar,Z2n,evn),DDt;N(1038,1,{847:1},Av),p.Sg=function(t,i){return!1},p.Tg=function(){},p.Ug=function(){return 0},p.Vg=function(){return null},p.Wg=function(){return null},p.Xg=function(){return null},p.Yg=function(){return!1},p.Zg=function(){return!1},p.$g=function(){return!1},p._g=function(t){},p.ah=function(t,i){},p.bh=function(t){return this},p.dh=function(t){},R(Lc,"NullElkProgressMonitor",1038),N(47,1,{20:1,47:1},Ya),p.Ic=function(t){Oa(this,t)},p.Fb=function(t){var i,a,u;return me(t,47)?(a=f(t,47),i=this.a==null?a.a==null:ei(this.a,a.a),u=this.b==null?a.b==null:ei(this.b,a.b),i&&u):!1},p.Hb=function(){var t,i,a,u,l,g;return a=this.a==null?0:Ti(this.a),t=a&To,i=a&-65536,g=this.b==null?0:Ti(this.b),u=g&To,l=g&-65536,t^l>>16&To|i^u<<16},p.Jc=function(){return new ent(this)},p.Ib=function(){return this.a==null&&this.b==null?"pair(null,null)":this.a==null?"pair(null,"+Qs(this.b)+")":this.b==null?"pair("+Qs(this.a)+",null)":"pair("+Qs(this.a)+","+Qs(this.b)+")"},R(Lc,"Pair",47),N(972,1,aa,ent),p.Nb=function(t){va(this,t)},p.Ob=function(){return!this.c&&(!this.b&&this.a.a!=null||this.a.b!=null)},p.Pb=function(){if(!this.c&&!this.b&&this.a.a!=null)return this.b=!0,this.a.a;if(!this.c&&this.a.b!=null)return this.c=!0,this.a.b;throw ie(new ic)},p.Qb=function(){throw this.c&&this.a.b!=null?this.a.b=null:this.b&&this.a.a!=null&&(this.a.a=null),ie(new Tl)},p.b=!1,p.c=!1,R(Lc,"Pair/1",972),N(1073,222,JS,KO),p.kf=function(t,i){var a,u,l,g,m;if(i.Sg("Random Layout",1),(!t.a&&(t.a=new Be(Ur,t,10,11)),t.a).i==0){i.Tg();return}g=f(Fe(t,(u3e(),TDt)),17),g&&g.a!=0?l=new hJ(g.a):l=new vle,a=_z(Oe(Fe(t,_Dt))),m=_z(Oe(Fe(t,xDt))),u=f(Fe(t,kDt),107),p9n(t,l,a,m,u),i.Tg()},R(Lc,"RandomLayoutProvider",1073),N(239,1,{239:1},Yse),p.Fb=function(t){return Pu(this.a,f(t,239).a)&&Pu(this.b,f(t,239).b)&&Pu(this.c,f(t,239).c)},p.Hb=function(){return oZ(ce(se(Ro,1),Jt,1,5,[this.a,this.b,this.c]))},p.Ib=function(){return"("+this.a+Fu+this.b+Fu+this.c+")"},R(Lc,"Triple",239);var MDt;N(547,1,{}),p.Jf=function(){return new Ke(this.f.i,this.f.j)},p.mf=function(t){return mft(t,(Rr(),Sd))?Fe(this.f,jDt):Fe(this.f,t)},p.Kf=function(){return new Ke(this.f.g,this.f.f)},p.Lf=function(){return this.g},p.nf=function(t){return mm(this.f,t)},p.Mf=function(t){md(this.f,t.a),bd(this.f,t.b)},p.Nf=function(t){RS(this.f,t.a),NS(this.f,t.b)},p.Of=function(t){this.g=t},p.g=0;var jDt;R(E$,"ElkGraphAdapters/AbstractElkGraphElementAdapter",547),N(548,1,{832:1},rY),p.Pf=function(){var t,i;if(!this.b)for(this.b=cJ(Cce(this.a).i),i=new Nn(Cce(this.a));i.e!=i.i.gc();)t=f(Mn(i),153),Qe(this.b,new _ae(t));return this.b},p.b=null,R(E$,"ElkGraphAdapters/ElkEdgeAdapter",548),N(287,547,{},uS),p.Qf=function(){return swt(this)},p.a=null,R(E$,"ElkGraphAdapters/ElkGraphAdapter",287),N(625,547,{187:1},_ae),R(E$,"ElkGraphAdapters/ElkLabelAdapter",625),N(624,547,{680:1},Rse),p.Pf=function(){return DAn(this)},p.Tf=function(){var t;return t=f(Fe(this.f,(Rr(),OD)),139),!t&&(t=new i7),t},p.Vf=function(){return MAn(this)},p.Xf=function(t){var i;i=new Wse(t),pi(this.f,(Rr(),OD),i)},p.Yf=function(t){pi(this.f,(Rr(),V0),new W_e(t))},p.Rf=function(){return this.d},p.Sf=function(){var t,i;if(!this.a)for(this.a=new et,i=new Tn(Cn(kde(f(this.f,27)).a.Jc(),new B));Kn(i);)t=f(Dn(i),85),Qe(this.a,new rY(t));return this.a},p.Uf=function(){var t,i;if(!this.c)for(this.c=new et,i=new Tn(Cn(lw(f(this.f,27)).a.Jc(),new B));Kn(i);)t=f(Dn(i),85),Qe(this.c,new rY(t));return this.c},p.Wf=function(){return XX(f(this.f,27)).i!=0||lt(ut(f(this.f,27).mf((Rr(),rV))))},p.Zf=function(){P_n(this,(fS(),MDt))},p.a=null,p.b=null,p.c=null,p.d=null,p.e=null,R(E$,"ElkGraphAdapters/ElkNodeAdapter",624),N(1244,547,{831:1},tnt),p.Pf=function(){return HAn(this)},p.Sf=function(){var t,i;if(!this.a)for(this.a=lb(f(this.f,124).fh().i),i=new Nn(f(this.f,124).fh());i.e!=i.i.gc();)t=f(Mn(i),85),Qe(this.a,new rY(t));return this.a},p.Uf=function(){var t,i;if(!this.c)for(this.c=lb(f(this.f,124).gh().i),i=new Nn(f(this.f,124).gh());i.e!=i.i.gc();)t=f(Mn(i),85),Qe(this.c,new rY(t));return this.c},p.$f=function(){return f(f(this.f,124).mf((Rr(),qA)),64)},p._f=function(){var t,i,a,u,l,g,m,v;for(u=bg(f(this.f,124)),a=new Nn(f(this.f,124).gh());a.e!=a.i.gc();)for(t=f(Mn(a),85),v=new Nn((!t.c&&(t.c=new Qt(Yn,t,5,8)),t.c));v.e!=v.i.gc();){if(m=f(Mn(v),84),bT(Us(m),u))return!0;if(Us(m)==u&&lt(ut(Fe(t,(Rr(),Wbe)))))return!0}for(i=new Nn(f(this.f,124).fh());i.e!=i.i.gc();)for(t=f(Mn(i),85),g=new Nn((!t.b&&(t.b=new Qt(Yn,t,4,7)),t.b));g.e!=g.i.gc();)if(l=f(Mn(g),84),bT(Us(l),u))return!0;return!1},p.a=null,p.b=null,p.c=null,R(E$,"ElkGraphAdapters/ElkPortAdapter",1244),N(1245,1,Cr,Oy),p.Le=function(t,i){return JPn(f(t,124),f(i,124))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(E$,"ElkGraphAdapters/PortComparator",1245);var Dw=Bi(zf,"EObject"),$D=Bi(fA,xCt),yf=Bi(fA,CCt),wV=Bi(fA,ACt),yV=Bi(fA,"ElkShape"),Yn=Bi(fA,ICt),vo=Bi(fA,N4e),Ii=Bi(fA,OCt),EV=Bi(zf,NCt),BF=Bi(zf,"EFactory"),$Dt,l1e=Bi(zf,RCt),Bp=Bi(zf,"EPackage"),Qo,FDt,BDt,S7e,Rne,UDt,_7e,k7e,T7e,Db,zDt,HDt,hc=Bi(fA,R4e),Ur=Bi(fA,P4e),_d=Bi(fA,L4e);N(92,1,PCt),p.ph=function(){return this.qh(),null},p.qh=function(){return null},p.rh=function(){return this.qh(),!1},p.sh=function(){return!1},p.th=function(t){Zr(this,t)},R(FR,"BasicNotifierImpl",92),N(98,92,jCt),p.Uh=function(){return Ld(this)},p.uh=function(t,i){return t},p.vh=function(){throw ie(new cr)},p.wh=function(t){var i;return i=os(f(Vt(this.zh(),this.Bh()),19)),this.Lh().Ph(this,i.n,i.f,t)},p.xh=function(t,i){throw ie(new cr)},p.yh=function(t,i,a){return hf(this,t,i,a)},p.zh=function(){var t;return this.vh()&&(t=this.vh().Kk(),t)?t:this.ei()},p.Ah=function(){return Sde(this)},p.Bh=function(){throw ie(new cr)},p.Ch=function(){var t,i;return i=this.Wh().Lk(),!i&&this.vh().Qk(i=(v7(),t=lTe(Em(this.zh())),t==null?v1e:new Zz(this,t))),i},p.Dh=function(t,i){return t},p.Eh=function(t){var i;return i=t.mk(),i?t.Ij():Pi(this.zh(),t)},p.Fh=function(){var t;return t=this.vh(),t?t.Nk():null},p.Gh=function(){return this.vh()?this.vh().Kk():null},p.Hh=function(t,i,a){return CZ(this,t,i,a)},p.Ih=function(t){return PP(this,t)},p.Jh=function(t,i){return sue(this,t,i)},p.Kh=function(){var t;return t=this.vh(),!!t&&t.Ok()},p.Lh=function(){throw ie(new cr)},p.Mh=function(){return EZ(this)},p.Nh=function(t,i,a,u){return vR(this,t,i,u)},p.Oh=function(t,i,a){var u;return u=f(Vt(this.zh(),i),69),u.tk().wk(this,this.di(),i-this.fi(),t,a)},p.Ph=function(t,i,a,u){return eJ(this,t,i,u)},p.Qh=function(t,i,a){var u;return u=f(Vt(this.zh(),i),69),u.tk().xk(this,this.di(),i-this.fi(),t,a)},p.Rh=function(){return!!this.vh()&&!!this.vh().Mk()},p.Sh=function(t){return Rle(this,t)},p.Th=function(t){return Oft(this,t)},p.Vh=function(t){return rkt(this,t)},p.Wh=function(){throw ie(new cr)},p.Xh=function(){return this.vh()?this.vh().Mk():null},p.Yh=function(){return EZ(this)},p.Zh=function(t,i){bde(this,t,i)},p.$h=function(t){this.Wh().Pk(t)},p._h=function(t){this.Wh().Sk(t)},p.ai=function(t){this.Wh().Rk(t)},p.bi=function(t,i){var a,u,l,g;return g=this.Fh(),g&&t&&(i=za(g.Bl(),this,i),g.Fl(this)),u=this.Lh(),u&&((Dde(this,this.Lh(),this.Bh()).Bb&Xa)!=0?(l=u.Mh(),l&&(t?!g&&l.Fl(this):l.El(this))):(i=(a=this.Bh(),a>=0?this.wh(i):this.Lh().Ph(this,-1-a,null,i)),i=this.yh(null,-1,i))),this._h(t),i},p.ci=function(t){var i,a,u,l,g,m,v,S;if(a=this.zh(),g=Pi(a,t),i=this.fi(),g>=i)return f(t,69).tk().Ak(this,this.di(),g-i);if(g<=-1)if(m=iA((Rl(),Ea),a,t),m){if(is(),f(m,69).uk()||(m=rR(Ns(Ea,m))),l=(u=this.Eh(m),f(u>=0?this.Hh(u,!0,!0):GS(this,m,!0),159)),S=m.Fk(),S>1||S==-1)return f(f(l,218).Pl(t,!1),76)}else throw ie(new dn(vw+t.ve()+rpe));else if(t.Gk())return u=this.Eh(t),f(u>=0?this.Hh(u,!1,!0):GS(this,t,!1),76);return v=new Vat(this,t),v},p.di=function(){return mxe(this)},p.ei=function(){return(Wv(),sn).S},p.fi=function(){return jn(this.ei())},p.gi=function(t){pde(this,t)},p.Ib=function(){return Jh(this)},R(un,"BasicEObjectImpl",98);var GDt;N(117,98,{109:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1}),p.hi=function(t){var i;return i=bxe(this),i[t]},p.ii=function(t,i){var a;a=bxe(this),io(a,t,i)},p.ji=function(t){var i;i=bxe(this),io(i,t,null)},p.ph=function(){return f(fn(this,4),128)},p.qh=function(){throw ie(new cr)},p.rh=function(){return(this.Db&4)!=0},p.vh=function(){throw ie(new cr)},p.ki=function(t){bR(this,2,t)},p.xh=function(t,i){this.Db=i<<16|this.Db&255,this.ki(t)},p.zh=function(){return fl(this)},p.Bh=function(){return this.Db>>16},p.Ch=function(){var t,i;return v7(),i=lTe(Em((t=f(fn(this,16),29),t||this.ei()))),i==null?v1e:new Zz(this,i)},p.sh=function(){return(this.Db&1)==0},p.Fh=function(){return f(fn(this,128),1980)},p.Gh=function(){return f(fn(this,16),29)},p.Kh=function(){return(this.Db&32)!=0},p.Lh=function(){return f(fn(this,2),52)},p.Rh=function(){return(this.Db&64)!=0},p.Wh=function(){throw ie(new cr)},p.Xh=function(){return f(fn(this,64),286)},p.$h=function(t){bR(this,16,t)},p._h=function(t){bR(this,128,t)},p.ai=function(t){bR(this,64,t)},p.di=function(){return Xu(this)},p.Db=0,R(un,"MinimalEObjectImpl",117),N(118,117,{109:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1}),p.ki=function(t){this.Cb=t},p.Lh=function(){return this.Cb},R(un,"MinimalEObjectImpl/Container",118),N(2029,118,{109:1,338:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1}),p.Hh=function(t,i,a){return N3e(this,t,i,a)},p.Qh=function(t,i,a){return yAe(this,t,i,a)},p.Sh=function(t){return TTe(this,t)},p.Zh=function(t,i){mCe(this,t,i)},p.ei=function(){return Gc(),HDt},p.gi=function(t){iCe(this,t)},p.lf=function(){return Avt(this)},p.eh=function(){return!this.o&&(this.o=new Ol((Gc(),Db),K0,this,0)),this.o},p.mf=function(t){return Fe(this,t)},p.nf=function(t){return mm(this,t)},p.of=function(t,i){return pi(this,t,i)},R(kE,"EMapPropertyHolderImpl",2029),N(555,118,{109:1,369:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1},Iv),p.Hh=function(t,i,a){switch(t){case 0:return this.a;case 1:return this.b}return CZ(this,t,i,a)},p.Sh=function(t){switch(t){case 0:return this.a!=0;case 1:return this.b!=0}return Rle(this,t)},p.Zh=function(t,i){switch(t){case 0:MJ(this,ke(Oe(i)));return;case 1:DJ(this,ke(Oe(i)));return}bde(this,t,i)},p.ei=function(){return Gc(),FDt},p.gi=function(t){switch(t){case 0:MJ(this,0);return;case 1:DJ(this,0);return}pde(this,t)},p.Ib=function(){var t;return(this.Db&64)!=0?Jh(this):(t=new sh(Jh(this)),t.a+=" (x: ",v3(t,this.a),t.a+=", y: ",v3(t,this.b),t.a+=")",t.a)},p.a=0,p.b=0,R(kE,"ElkBendPointImpl",555),N(722,2029,{109:1,338:1,173:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1}),p.Hh=function(t,i,a){return MCe(this,t,i,a)},p.Oh=function(t,i,a){return ade(this,t,i,a)},p.Qh=function(t,i,a){return Gue(this,t,i,a)},p.Sh=function(t){return eCe(this,t)},p.Zh=function(t,i){X3e(this,t,i)},p.ei=function(){return Gc(),UDt},p.gi=function(t){LCe(this,t)},p.hh=function(){return this.k},p.ih=function(){return Cce(this)},p.Ib=function(){return mle(this)},p.k=null,R(kE,"ElkGraphElementImpl",722),N(723,722,{109:1,338:1,173:1,273:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1}),p.Hh=function(t,i,a){return YCe(this,t,i,a)},p.Sh=function(t){return o3e(this,t)},p.Zh=function(t,i){J3e(this,t,i)},p.ei=function(){return Gc(),zDt},p.gi=function(t){a3e(this,t)},p.jh=function(){return this.f},p.kh=function(){return this.g},p.lh=function(){return this.i},p.mh=function(){return this.j},p.nh=function(t,i){x3(this,t,i)},p.oh=function(t,i){Of(this,t,i)},p.Ib=function(){return fde(this)},p.f=0,p.g=0,p.i=0,p.j=0,R(kE,"ElkShapeImpl",723),N(724,723,{109:1,338:1,84:1,173:1,273:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1}),p.Hh=function(t,i,a){return T3e(this,t,i,a)},p.Oh=function(t,i,a){return G3e(this,t,i,a)},p.Qh=function(t,i,a){return q3e(this,t,i,a)},p.Sh=function(t){return pCe(this,t)},p.Zh=function(t,i){iIe(this,t,i)},p.ei=function(){return Gc(),BDt},p.gi=function(t){y3e(this,t)},p.fh=function(){return!this.d&&(this.d=new Qt(vo,this,8,5)),this.d},p.gh=function(){return!this.e&&(this.e=new Qt(vo,this,7,4)),this.e},R(kE,"ElkConnectableShapeImpl",724),N(268,722,{109:1,338:1,85:1,173:1,268:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1},WC),p.wh=function(t){return B3e(this,t)},p.Hh=function(t,i,a){switch(t){case 3:return lT(this);case 4:return!this.b&&(this.b=new Qt(Yn,this,4,7)),this.b;case 5:return!this.c&&(this.c=new Qt(Yn,this,5,8)),this.c;case 6:return!this.a&&(this.a=new Be(Ii,this,6,6)),this.a;case 7:return mn(),!this.b&&(this.b=new Qt(Yn,this,4,7)),!(this.b.i<=1&&(!this.c&&(this.c=new Qt(Yn,this,5,8)),this.c.i<=1));case 8:return mn(),!!Uj(this);case 9:return mn(),!!HS(this);case 10:return mn(),!this.b&&(this.b=new Qt(Yn,this,4,7)),this.b.i!=0&&(!this.c&&(this.c=new Qt(Yn,this,5,8)),this.c.i!=0)}return MCe(this,t,i,a)},p.Oh=function(t,i,a){var u;switch(i){case 3:return this.Cb&&(a=(u=this.Db>>16,u>=0?B3e(this,a):this.Cb.Ph(this,-1-u,null,a))),__e(this,f(t,27),a);case 4:return!this.b&&(this.b=new Qt(Yn,this,4,7)),$u(this.b,t,a);case 5:return!this.c&&(this.c=new Qt(Yn,this,5,8)),$u(this.c,t,a);case 6:return!this.a&&(this.a=new Be(Ii,this,6,6)),$u(this.a,t,a)}return ade(this,t,i,a)},p.Qh=function(t,i,a){switch(i){case 3:return __e(this,null,a);case 4:return!this.b&&(this.b=new Qt(Yn,this,4,7)),za(this.b,t,a);case 5:return!this.c&&(this.c=new Qt(Yn,this,5,8)),za(this.c,t,a);case 6:return!this.a&&(this.a=new Be(Ii,this,6,6)),za(this.a,t,a)}return Gue(this,t,i,a)},p.Sh=function(t){switch(t){case 3:return!!lT(this);case 4:return!!this.b&&this.b.i!=0;case 5:return!!this.c&&this.c.i!=0;case 6:return!!this.a&&this.a.i!=0;case 7:return!this.b&&(this.b=new Qt(Yn,this,4,7)),!(this.b.i<=1&&(!this.c&&(this.c=new Qt(Yn,this,5,8)),this.c.i<=1));case 8:return Uj(this);case 9:return HS(this);case 10:return!this.b&&(this.b=new Qt(Yn,this,4,7)),this.b.i!=0&&(!this.c&&(this.c=new Qt(Yn,this,5,8)),this.c.i!=0)}return eCe(this,t)},p.Zh=function(t,i){switch(t){case 3:QZ(this,f(i,27));return;case 4:!this.b&&(this.b=new Qt(Yn,this,4,7)),Jn(this.b),!this.b&&(this.b=new Qt(Yn,this,4,7)),Qi(this.b,f(i,16));return;case 5:!this.c&&(this.c=new Qt(Yn,this,5,8)),Jn(this.c),!this.c&&(this.c=new Qt(Yn,this,5,8)),Qi(this.c,f(i,16));return;case 6:!this.a&&(this.a=new Be(Ii,this,6,6)),Jn(this.a),!this.a&&(this.a=new Be(Ii,this,6,6)),Qi(this.a,f(i,16));return}X3e(this,t,i)},p.ei=function(){return Gc(),S7e},p.gi=function(t){switch(t){case 3:QZ(this,null);return;case 4:!this.b&&(this.b=new Qt(Yn,this,4,7)),Jn(this.b);return;case 5:!this.c&&(this.c=new Qt(Yn,this,5,8)),Jn(this.c);return;case 6:!this.a&&(this.a=new Be(Ii,this,6,6)),Jn(this.a);return}LCe(this,t)},p.Ib=function(){return k_t(this)},R(kE,"ElkEdgeImpl",268),N(440,2029,{109:1,338:1,168:1,440:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1},em),p.wh=function(t){return M3e(this,t)},p.Hh=function(t,i,a){switch(t){case 1:return this.j;case 2:return this.k;case 3:return this.b;case 4:return this.c;case 5:return!this.a&&(this.a=new So(yf,this,5)),this.a;case 6:return xft(this);case 7:return i?jle(this):this.i;case 8:return i?Mle(this):this.f;case 9:return!this.g&&(this.g=new Qt(Ii,this,9,10)),this.g;case 10:return!this.e&&(this.e=new Qt(Ii,this,10,9)),this.e;case 11:return this.d}return N3e(this,t,i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 6:return this.Cb&&(a=(l=this.Db>>16,l>=0?M3e(this,a):this.Cb.Ph(this,-1-l,null,a))),k_e(this,f(t,85),a);case 9:return!this.g&&(this.g=new Qt(Ii,this,9,10)),$u(this.g,t,a);case 10:return!this.e&&(this.e=new Qt(Ii,this,10,9)),$u(this.e,t,a)}return g=f(Vt((u=f(fn(this,16),29),u||(Gc(),Rne)),i),69),g.tk().wk(this,Xu(this),i-jn((Gc(),Rne)),t,a)},p.Qh=function(t,i,a){switch(i){case 5:return!this.a&&(this.a=new So(yf,this,5)),za(this.a,t,a);case 6:return k_e(this,null,a);case 9:return!this.g&&(this.g=new Qt(Ii,this,9,10)),za(this.g,t,a);case 10:return!this.e&&(this.e=new Qt(Ii,this,10,9)),za(this.e,t,a)}return yAe(this,t,i,a)},p.Sh=function(t){switch(t){case 1:return this.j!=0;case 2:return this.k!=0;case 3:return this.b!=0;case 4:return this.c!=0;case 5:return!!this.a&&this.a.i!=0;case 6:return!!xft(this);case 7:return!!this.i;case 8:return!!this.f;case 9:return!!this.g&&this.g.i!=0;case 10:return!!this.e&&this.e.i!=0;case 11:return this.d!=null}return TTe(this,t)},p.Zh=function(t,i){switch(t){case 1:B3(this,ke(Oe(i)));return;case 2:H3(this,ke(Oe(i)));return;case 3:U3(this,ke(Oe(i)));return;case 4:z3(this,ke(Oe(i)));return;case 5:!this.a&&(this.a=new So(yf,this,5)),Jn(this.a),!this.a&&(this.a=new So(yf,this,5)),Qi(this.a,f(i,16));return;case 6:T2t(this,f(i,85));return;case 7:HJ(this,f(i,84));return;case 8:zJ(this,f(i,84));return;case 9:!this.g&&(this.g=new Qt(Ii,this,9,10)),Jn(this.g),!this.g&&(this.g=new Qt(Ii,this,9,10)),Qi(this.g,f(i,16));return;case 10:!this.e&&(this.e=new Qt(Ii,this,10,9)),Jn(this.e),!this.e&&(this.e=new Qt(Ii,this,10,9)),Qi(this.e,f(i,16));return;case 11:Uxe(this,dr(i));return}mCe(this,t,i)},p.ei=function(){return Gc(),Rne},p.gi=function(t){switch(t){case 1:B3(this,0);return;case 2:H3(this,0);return;case 3:U3(this,0);return;case 4:z3(this,0);return;case 5:!this.a&&(this.a=new So(yf,this,5)),Jn(this.a);return;case 6:T2t(this,null);return;case 7:HJ(this,null);return;case 8:zJ(this,null);return;case 9:!this.g&&(this.g=new Qt(Ii,this,9,10)),Jn(this.g);return;case 10:!this.e&&(this.e=new Qt(Ii,this,10,9)),Jn(this.e);return;case 11:Uxe(this,null);return}iCe(this,t)},p.Ib=function(){return jEt(this)},p.b=0,p.c=0,p.d=null,p.j=0,p.k=0,R(kE,"ElkEdgeSectionImpl",440),N(157,118,{109:1,93:1,92:1,154:1,57:1,113:1,52:1,98:1,157:1,117:1,118:1}),p.Hh=function(t,i,a){var u;return t==0?(!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab):Lf(this,t-jn(this.ei()),Vt((u=f(fn(this,16),29),u||this.ei()),t),i,a)},p.Oh=function(t,i,a){var u,l;return i==0?(!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a)):(l=f(Vt((u=f(fn(this,16),29),u||this.ei()),i),69),l.tk().wk(this,Xu(this),i-jn(this.ei()),t,a))},p.Qh=function(t,i,a){var u,l;return i==0?(!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a)):(l=f(Vt((u=f(fn(this,16),29),u||this.ei()),i),69),l.tk().xk(this,Xu(this),i-jn(this.ei()),t,a))},p.Sh=function(t){var i;return t==0?!!this.Ab&&this.Ab.i!=0:Pf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.Vh=function(t){return yOe(this,t)},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return}$f(this,t-jn(this.ei()),Vt((a=f(fn(this,16),29),a||this.ei()),t),i)},p._h=function(t){bR(this,128,t)},p.ei=function(){return Ut(),cMt},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return}jf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.li=function(){this.Bb|=1},p.mi=function(t){return Kj(this,t)},p.Bb=0,R(un,"EModelElementImpl",157),N(704,157,{109:1,93:1,92:1,466:1,154:1,57:1,113:1,52:1,98:1,157:1,117:1,118:1},om),p.ni=function(t,i){return J_t(this,t,i)},p.oi=function(t){var i,a,u,l,g;if(this.a!=cf(t)||(t.Bb&256)!=0)throw ie(new dn(ope+t.zb+n_));for(u=Bs(t);iu(u.a).i!=0;){if(a=f(RG(u,0,(i=f(be(iu(u.a),0),87),g=i.c,me(g,88)?f(g,29):(Ut(),Ch))),29),US(a))return l=cf(a).si().oi(a),f(l,52).$h(t),l;u=Bs(a)}return(t.D!=null?t.D:t.B)=="java.util.Map$Entry"?new dlt(t):new lke(t)},p.pi=function(t,i){return KS(this,t,i)},p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.a}return Lf(this,t-jn((Ut(),$w)),Vt((u=f(fn(this,16),29),u||$w),t),i,a)},p.Oh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 1:return this.a&&(a=f(this.a,52).Ph(this,4,Bp,a)),NCe(this,f(t,240),a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),$w)),i),69),l.tk().wk(this,Xu(this),i-jn((Ut(),$w)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 1:return NCe(this,null,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),$w)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),$w)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return!!this.a}return Pf(this,t-jn((Ut(),$w)),Vt((i=f(fn(this,16),29),i||$w),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:fyt(this,f(i,240));return}$f(this,t-jn((Ut(),$w)),Vt((a=f(fn(this,16),29),a||$w),t),i)},p.ei=function(){return Ut(),$w},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:fyt(this,null);return}jf(this,t-jn((Ut(),$w)),Vt((i=f(fn(this,16),29),i||$w),t))};var UF,x7e,qDt;R(un,"EFactoryImpl",704),N(1012,704,{109:1,2059:1,93:1,92:1,466:1,154:1,57:1,113:1,52:1,98:1,157:1,117:1,118:1},ZO),p.ni=function(t,i){switch(t.ek()){case 12:return f(i,144).Ng();case 13:return Qs(i);default:throw ie(new dn(VL+t.ve()+n_))}},p.oi=function(t){var i,a,u,l,g,m,v,S;switch(t.G==-1&&(t.G=(i=cf(t),i?P0(i.ri(),t):-1)),t.G){case 4:return g=new KC,g;case 6:return m=new o7,m;case 7:return v=new w2e,v;case 8:return u=new WC,u;case 9:return a=new Iv,a;case 10:return l=new em,l;case 11:return S=new Sk,S;default:throw ie(new dn(ope+t.zb+n_))}},p.pi=function(t,i){switch(t.ek()){case 13:case 12:return null;default:throw ie(new dn(VL+t.ve()+n_))}},R(kE,"ElkGraphFactoryImpl",1012),N(436,157,{109:1,93:1,92:1,154:1,197:1,57:1,113:1,52:1,98:1,157:1,117:1,118:1}),p.Ch=function(){var t,i;return i=(t=f(fn(this,16),29),lTe(Em(t||this.ei()))),i==null?(v7(),v7(),v1e):new Lct(this,i)},p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.ve()}return Lf(this,t-jn(this.ei()),Vt((u=f(fn(this,16),29),u||this.ei()),t),i,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null}return Pf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:this.qi(dr(i));return}$f(this,t-jn(this.ei()),Vt((a=f(fn(this,16),29),a||this.ei()),t),i)},p.ei=function(){return Ut(),uMt},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:this.qi(null);return}jf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.ve=function(){return this.zb},p.qi=function(t){ju(this,t)},p.Ib=function(){return _j(this)},p.zb=null,R(un,"ENamedElementImpl",436),N(184,436,{109:1,93:1,92:1,154:1,197:1,57:1,240:1,113:1,52:1,98:1,157:1,184:1,117:1,118:1,675:1},oft),p.wh=function(t){return _wt(this,t)},p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return this.yb;case 3:return this.xb;case 4:return this.sb;case 5:return!this.rb&&(this.rb=new aT(this,Up,this)),this.rb;case 6:return!this.vb&&(this.vb=new qN(Bp,this,6,7)),this.vb;case 7:return i?this.Db>>16==7?f(this.Cb,240):null:Rft(this)}return Lf(this,t-jn((Ut(),Z0)),Vt((u=f(fn(this,16),29),u||Z0),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 4:return this.sb&&(a=f(this.sb,52).Ph(this,1,BF,a)),RCe(this,f(t,466),a);case 5:return!this.rb&&(this.rb=new aT(this,Up,this)),$u(this.rb,t,a);case 6:return!this.vb&&(this.vb=new qN(Bp,this,6,7)),$u(this.vb,t,a);case 7:return this.Cb&&(a=(l=this.Db>>16,l>=0?_wt(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,7,a)}return g=f(Vt((u=f(fn(this,16),29),u||(Ut(),Z0)),i),69),g.tk().wk(this,Xu(this),i-jn((Ut(),Z0)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 4:return RCe(this,null,a);case 5:return!this.rb&&(this.rb=new aT(this,Up,this)),za(this.rb,t,a);case 6:return!this.vb&&(this.vb=new qN(Bp,this,6,7)),za(this.vb,t,a);case 7:return hf(this,null,7,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),Z0)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),Z0)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return this.yb!=null;case 3:return this.xb!=null;case 4:return!!this.sb;case 5:return!!this.rb&&this.rb.i!=0;case 6:return!!this.vb&&this.vb.i!=0;case 7:return!!Rft(this)}return Pf(this,t-jn((Ut(),Z0)),Vt((i=f(fn(this,16),29),i||Z0),t))},p.Vh=function(t){var i;return i=g5n(this,t),i||yOe(this,t)},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:ju(this,dr(i));return;case 2:KJ(this,dr(i));return;case 3:WJ(this,dr(i));return;case 4:dde(this,f(i,466));return;case 5:!this.rb&&(this.rb=new aT(this,Up,this)),Jn(this.rb),!this.rb&&(this.rb=new aT(this,Up,this)),Qi(this.rb,f(i,16));return;case 6:!this.vb&&(this.vb=new qN(Bp,this,6,7)),Jn(this.vb),!this.vb&&(this.vb=new qN(Bp,this,6,7)),Qi(this.vb,f(i,16));return}$f(this,t-jn((Ut(),Z0)),Vt((a=f(fn(this,16),29),a||Z0),t),i)},p.ai=function(t){var i,a;if(t&&this.rb)for(a=new Nn(this.rb);a.e!=a.i.gc();)i=Mn(a),me(i,357)&&(f(i,357).w=null);bR(this,64,t)},p.ei=function(){return Ut(),Z0},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:ju(this,null);return;case 2:KJ(this,null);return;case 3:WJ(this,null);return;case 4:dde(this,null);return;case 5:!this.rb&&(this.rb=new aT(this,Up,this)),Jn(this.rb);return;case 6:!this.vb&&(this.vb=new qN(Bp,this,6,7)),Jn(this.vb);return}jf(this,t-jn((Ut(),Z0)),Vt((i=f(fn(this,16),29),i||Z0),t))},p.li=function(){Xle(this)},p.ri=function(){return!this.rb&&(this.rb=new aT(this,Up,this)),this.rb},p.si=function(){return this.sb},p.ti=function(){return this.ub},p.ui=function(){return this.xb},p.vi=function(){return this.yb},p.wi=function(t){this.ub=t},p.Ib=function(){var t;return(this.Db&64)!=0?_j(this):(t=new sh(_j(this)),t.a+=" (nsURI: ",ys(t,this.yb),t.a+=", nsPrefix: ",ys(t,this.xb),t.a+=")",t.a)},p.xb=null,p.yb=null,R(un,"EPackageImpl",184),N(552,184,{109:1,2061:1,552:1,93:1,92:1,154:1,197:1,57:1,240:1,113:1,52:1,98:1,157:1,184:1,117:1,118:1,675:1},qEt),p.q=!1,p.r=!1;var VDt=!1;R(kE,"ElkGraphPackageImpl",552),N(359,723,{109:1,338:1,173:1,153:1,273:1,359:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1},KC),p.wh=function(t){return j3e(this,t)},p.Hh=function(t,i,a){switch(t){case 7:return gTe(this);case 8:return this.a}return YCe(this,t,i,a)},p.Oh=function(t,i,a){var u;switch(i){case 7:return this.Cb&&(a=(u=this.Db>>16,u>=0?j3e(this,a):this.Cb.Ph(this,-1-u,null,a))),Tke(this,f(t,173),a)}return ade(this,t,i,a)},p.Qh=function(t,i,a){return i==7?Tke(this,null,a):Gue(this,t,i,a)},p.Sh=function(t){switch(t){case 7:return!!gTe(this);case 8:return!Mt("",this.a)}return o3e(this,t)},p.Zh=function(t,i){switch(t){case 7:yIe(this,f(i,173));return;case 8:jxe(this,dr(i));return}J3e(this,t,i)},p.ei=function(){return Gc(),_7e},p.gi=function(t){switch(t){case 7:yIe(this,null);return;case 8:jxe(this,"");return}a3e(this,t)},p.Ib=function(){return Oyt(this)},p.a="",R(kE,"ElkLabelImpl",359),N(206,724,{109:1,338:1,84:1,173:1,27:1,273:1,206:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1},o7),p.wh=function(t){return U3e(this,t)},p.Hh=function(t,i,a){switch(t){case 9:return!this.c&&(this.c=new Be(_d,this,9,9)),this.c;case 10:return!this.a&&(this.a=new Be(Ur,this,10,11)),this.a;case 11:return so(this);case 12:return!this.b&&(this.b=new Be(vo,this,12,3)),this.b;case 13:return mn(),!this.a&&(this.a=new Be(Ur,this,10,11)),this.a.i>0}return T3e(this,t,i,a)},p.Oh=function(t,i,a){var u;switch(i){case 9:return!this.c&&(this.c=new Be(_d,this,9,9)),$u(this.c,t,a);case 10:return!this.a&&(this.a=new Be(Ur,this,10,11)),$u(this.a,t,a);case 11:return this.Cb&&(a=(u=this.Db>>16,u>=0?U3e(this,a):this.Cb.Ph(this,-1-u,null,a))),D_e(this,f(t,27),a);case 12:return!this.b&&(this.b=new Be(vo,this,12,3)),$u(this.b,t,a)}return G3e(this,t,i,a)},p.Qh=function(t,i,a){switch(i){case 9:return!this.c&&(this.c=new Be(_d,this,9,9)),za(this.c,t,a);case 10:return!this.a&&(this.a=new Be(Ur,this,10,11)),za(this.a,t,a);case 11:return D_e(this,null,a);case 12:return!this.b&&(this.b=new Be(vo,this,12,3)),za(this.b,t,a)}return q3e(this,t,i,a)},p.Sh=function(t){switch(t){case 9:return!!this.c&&this.c.i!=0;case 10:return!!this.a&&this.a.i!=0;case 11:return!!so(this);case 12:return!!this.b&&this.b.i!=0;case 13:return!this.a&&(this.a=new Be(Ur,this,10,11)),this.a.i>0}return pCe(this,t)},p.Zh=function(t,i){switch(t){case 9:!this.c&&(this.c=new Be(_d,this,9,9)),Jn(this.c),!this.c&&(this.c=new Be(_d,this,9,9)),Qi(this.c,f(i,16));return;case 10:!this.a&&(this.a=new Be(Ur,this,10,11)),Jn(this.a),!this.a&&(this.a=new Be(Ur,this,10,11)),Qi(this.a,f(i,16));return;case 11:eQ(this,f(i,27));return;case 12:!this.b&&(this.b=new Be(vo,this,12,3)),Jn(this.b),!this.b&&(this.b=new Be(vo,this,12,3)),Qi(this.b,f(i,16));return}iIe(this,t,i)},p.ei=function(){return Gc(),k7e},p.gi=function(t){switch(t){case 9:!this.c&&(this.c=new Be(_d,this,9,9)),Jn(this.c);return;case 10:!this.a&&(this.a=new Be(Ur,this,10,11)),Jn(this.a);return;case 11:eQ(this,null);return;case 12:!this.b&&(this.b=new Be(vo,this,12,3)),Jn(this.b);return}y3e(this,t)},p.Ib=function(){return MIe(this)},R(kE,"ElkNodeImpl",206),N(193,724,{109:1,338:1,84:1,173:1,124:1,273:1,193:1,102:1,93:1,92:1,57:1,113:1,52:1,98:1,117:1,118:1},w2e),p.wh=function(t){return $3e(this,t)},p.Hh=function(t,i,a){return t==9?bg(this):T3e(this,t,i,a)},p.Oh=function(t,i,a){var u;switch(i){case 9:return this.Cb&&(a=(u=this.Db>>16,u>=0?$3e(this,a):this.Cb.Ph(this,-1-u,null,a))),T_e(this,f(t,27),a)}return G3e(this,t,i,a)},p.Qh=function(t,i,a){return i==9?T_e(this,null,a):q3e(this,t,i,a)},p.Sh=function(t){return t==9?!!bg(this):pCe(this,t)},p.Zh=function(t,i){switch(t){case 9:gIe(this,f(i,27));return}iIe(this,t,i)},p.ei=function(){return Gc(),T7e},p.gi=function(t){switch(t){case 9:gIe(this,null);return}y3e(this,t)},p.Ib=function(){return ESt(this)},R(kE,"ElkPortImpl",193);var WDt=Bi(Ha,"BasicEMap/Entry");N(1086,118,{109:1,43:1,93:1,92:1,135:1,57:1,113:1,52:1,98:1,117:1,118:1},Sk),p.Fb=function(t){return this===t},p.jd=function(){return this.b},p.Hb=function(){return bS(this)},p.zi=function(t){Dxe(this,f(t,144))},p.Hh=function(t,i,a){switch(t){case 0:return this.b;case 1:return this.c}return CZ(this,t,i,a)},p.Sh=function(t){switch(t){case 0:return!!this.b;case 1:return this.c!=null}return Rle(this,t)},p.Zh=function(t,i){switch(t){case 0:Dxe(this,f(i,144));return;case 1:Mxe(this,i);return}bde(this,t,i)},p.ei=function(){return Gc(),Db},p.gi=function(t){switch(t){case 0:Dxe(this,null);return;case 1:Mxe(this,null);return}pde(this,t)},p.xi=function(){var t;return this.a==-1&&(t=this.b,this.a=t?Ti(t):0),this.a},p.kd=function(){return this.c},p.yi=function(t){this.a=t},p.ld=function(t){var i;return i=this.c,Mxe(this,t),i},p.Ib=function(){var t;return(this.Db&64)!=0?Jh(this):(t=new $v,xr(xr(xr(t,this.b?this.b.Ng():bl),Xfe),$7(this.c)),t.a)},p.a=-1,p.c=null;var K0=R(kE,"ElkPropertyToValueMapEntryImpl",1086);N(973,1,{},YC),R(ma,"JsonAdapter",973),N(214,63,C1,am),R(ma,"JsonImportException",214),N(845,1,{},BEt),R(ma,"JsonImporter",845),N(879,1,{},Nat),p.Ai=function(t){Pwt(this.a,this.b,f(t,138))},R(ma,"JsonImporter/lambda$0$Type",879),N(880,1,{},Rat),p.Ai=function(t){mEt(this.a,this.b,f(t,138))},R(ma,"JsonImporter/lambda$1$Type",880),N(888,1,{},nnt),p.Ai=function(t){$dt(this.a,f(t,147))},R(ma,"JsonImporter/lambda$10$Type",888),N(890,1,{},Pat),p.Ai=function(t){rEt(this.a,this.b,f(t,138))},R(ma,"JsonImporter/lambda$11$Type",890),N(891,1,{},Lat),p.Ai=function(t){iEt(this.a,this.b,f(t,138))},R(ma,"JsonImporter/lambda$12$Type",891),N(897,1,{},Kdt),p.Ai=function(t){xyt(this.a,this.b,this.c,this.d,f(t,138))},R(ma,"JsonImporter/lambda$13$Type",897),N(896,1,{},Ydt),p.Ai=function(t){zSt(this.a,this.b,this.c,this.d,f(t,147))},R(ma,"JsonImporter/lambda$14$Type",896),N(892,1,{},Dat),p.Ai=function(t){lut(this.a,this.b,dr(t))},R(ma,"JsonImporter/lambda$15$Type",892),N(893,1,{},Mat),p.Ai=function(t){dut(this.a,this.b,dr(t))},R(ma,"JsonImporter/lambda$16$Type",893),N(894,1,{},jat),p.Ai=function(t){pwt(this.b,this.a,f(t,138))},R(ma,"JsonImporter/lambda$17$Type",894),N(895,1,{},$at),p.Ai=function(t){gwt(this.b,this.a,f(t,138))},R(ma,"JsonImporter/lambda$18$Type",895),N(900,1,{},rnt),p.Ai=function(t){vyt(this.a,f(t,147))},R(ma,"JsonImporter/lambda$19$Type",900),N(881,1,{},int),p.Ai=function(t){xwt(this.a,f(t,138))},R(ma,"JsonImporter/lambda$2$Type",881),N(898,1,{},ont),p.Ai=function(t){B3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$20$Type",898),N(899,1,{},ant),p.Ai=function(t){H3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$21$Type",899),N(903,1,{},snt),p.Ai=function(t){byt(this.a,f(t,147))},R(ma,"JsonImporter/lambda$22$Type",903),N(901,1,{},cnt),p.Ai=function(t){U3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$23$Type",901),N(902,1,{},unt),p.Ai=function(t){z3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$24$Type",902),N(905,1,{},lnt),p.Ai=function(t){Gwt(this.a,f(t,138))},R(ma,"JsonImporter/lambda$25$Type",905),N(904,1,{},dnt),p.Ai=function(t){Fdt(this.a,f(t,147))},R(ma,"JsonImporter/lambda$26$Type",904),N(906,1,xn,Fat),p.Ad=function(t){b_n(this.b,this.a,dr(t))},R(ma,"JsonImporter/lambda$27$Type",906),N(907,1,xn,Bat),p.Ad=function(t){v_n(this.b,this.a,dr(t))},R(ma,"JsonImporter/lambda$28$Type",907),N(908,1,{},Uat),p.Ai=function(t){t2t(this.a,this.b,f(t,138))},R(ma,"JsonImporter/lambda$29$Type",908),N(884,1,{},fnt),p.Ai=function(t){$0t(this.a,f(t,147))},R(ma,"JsonImporter/lambda$3$Type",884),N(909,1,{},zat),p.Ai=function(t){E2t(this.a,this.b,f(t,138))},R(ma,"JsonImporter/lambda$30$Type",909),N(910,1,{},hnt),p.Ai=function(t){cmt(this.a,Oe(t))},R(ma,"JsonImporter/lambda$31$Type",910),N(911,1,{},pnt),p.Ai=function(t){umt(this.a,Oe(t))},R(ma,"JsonImporter/lambda$32$Type",911),N(912,1,{},gnt),p.Ai=function(t){lmt(this.a,Oe(t))},R(ma,"JsonImporter/lambda$33$Type",912),N(913,1,{},mnt),p.Ai=function(t){dmt(this.a,Oe(t))},R(ma,"JsonImporter/lambda$34$Type",913),N(914,1,{},bnt),p.Ai=function(t){pNn(this.a,f(t,57))},R(ma,"JsonImporter/lambda$35$Type",914),N(915,1,{},vnt),p.Ai=function(t){gNn(this.a,f(t,57))},R(ma,"JsonImporter/lambda$36$Type",915),N(919,1,{},Wdt),R(ma,"JsonImporter/lambda$37$Type",919),N(916,1,xn,Fut),p.Ad=function(t){Hkn(this.a,this.c,this.b,f(t,369))},R(ma,"JsonImporter/lambda$38$Type",916),N(917,1,xn,Hat),p.Ad=function(t){Ign(this.a,this.b,f(t,168))},R(ma,"JsonImporter/lambda$39$Type",917),N(882,1,{},wnt),p.Ai=function(t){B3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$4$Type",882),N(918,1,xn,Gat),p.Ad=function(t){Ogn(this.a,this.b,f(t,168))},R(ma,"JsonImporter/lambda$40$Type",918),N(920,1,xn,jut),p.Ad=function(t){Gkn(this.a,this.b,this.c,f(t,8))},R(ma,"JsonImporter/lambda$41$Type",920),N(883,1,{},ynt),p.Ai=function(t){H3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$5$Type",883),N(887,1,{},Ent),p.Ai=function(t){F0t(this.a,f(t,147))},R(ma,"JsonImporter/lambda$6$Type",887),N(885,1,{},Snt),p.Ai=function(t){U3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$7$Type",885),N(886,1,{},_nt),p.Ai=function(t){z3(this.a,ke(Oe(t)))},R(ma,"JsonImporter/lambda$8$Type",886),N(889,1,{},knt),p.Ai=function(t){qwt(this.a,f(t,138))},R(ma,"JsonImporter/lambda$9$Type",889),N(939,1,xn,Tnt),p.Ad=function(t){QN(this.a,new cT(dr(t)))},R(ma,"JsonMetaDataConverter/lambda$0$Type",939),N(940,1,xn,xnt),p.Ad=function(t){Iwn(this.a,f(t,243))},R(ma,"JsonMetaDataConverter/lambda$1$Type",940),N(941,1,xn,Cnt),p.Ad=function(t){mEn(this.a,f(t,149))},R(ma,"JsonMetaDataConverter/lambda$2$Type",941),N(942,1,xn,Ant),p.Ad=function(t){Own(this.a,f(t,170))},R(ma,"JsonMetaDataConverter/lambda$3$Type",942),N(243,23,{3:1,34:1,23:1,243:1},$N);var Pne,Lne,d1e,Dne,Mne,jne,f1e,h1e,$ne=er(KG,"GraphFeature",243,ar,X_n,nvn),KDt;N(12,1,{34:1,144:1},Si,Mi,Pt,pa),p.Dd=function(t){return Omn(this,f(t,144))},p.Fb=function(t){return mft(this,t)},p.Qg=function(){return st(this)},p.Ng=function(){return this.b},p.Hb=function(){return C0(this.b)},p.Ib=function(){return this.b},R(KG,"Property",12),N(656,1,Cr,sae),p.Le=function(t,i){return dCn(this,f(t,102),f(i,102))},p.Fb=function(t){return this===t},p.Me=function(){return new wr(this)},R(KG,"PropertyHolderComparator",656),N(693,1,aa,t2e),p.Nb=function(t){va(this,t)},p.Pb=function(){return S_n(this)},p.Qb=function(){Eit()},p.Ob=function(){return!!this.a},R(uee,"ElkGraphUtil/AncestorIterator",693);var C7e=Bi(Ha,"EList");N(70,55,{20:1,31:1,55:1,16:1,15:1,70:1,61:1}),p._c=function(t,i){xj(this,t,i)},p.Ec=function(t){return Qn(this,t)},p.ad=function(t,i){return cCe(this,t,i)},p.Fc=function(t){return Qi(this,t)},p.Fi=function(){return new GN(this)},p.Gi=function(){return new Qz(this)},p.Hi=function(t){return UH(this,t)},p.Ii=function(){return!0},p.Ji=function(t,i){},p.Ki=function(){},p.Li=function(t,i){fue(this,t,i)},p.Mi=function(t,i,a){},p.Ni=function(t,i){},p.Oi=function(t,i,a){},p.Fb=function(t){return oSt(this,t)},p.Hb=function(){return rCe(this)},p.Pi=function(){return!1},p.Jc=function(){return new Nn(this)},p.cd=function(){return new HN(this)},p.dd=function(t){var i;if(i=this.gc(),t<0||t>i)throw ie(new rT(t,i));return new bce(this,t)},p.Ri=function(t,i){this.Qi(t,this.bd(i))},p.Kc=function(t){return IJ(this,t)},p.Ti=function(t,i){return i},p.fd=function(t,i){return Y3(this,t,i)},p.Ib=function(){return QCe(this)},p.Vi=function(){return!0},p.Wi=function(t,i){return KP(this,i)},R(Ha,"AbstractEList",70),N(66,70,Im,J2,OS,Jxe),p.Bi=function(t,i){return sde(this,t,i)},p.Ci=function(t){return Yvt(this,t)},p.Di=function(t,i){eG(this,t,i)},p.Ei=function(t){TH(this,t)},p.Xi=function(t){return fxe(this,t)},p.$b=function(){sj(this)},p.Gc=function(t){return lL(this,t)},p.Xb=function(t){return be(this,t)},p.Yi=function(t){var i,a,u;++this.j,a=this.g==null?0:this.g.length,t>a&&(u=this.g,i=a+(a/2|0)+4,i<t&&(i=t),this.g=this.Zi(i),u!=null&&au(u,0,this.g,0,this.i))},p.bd=function(t){return vwt(this,t)},p.dc=function(){return this.i==0},p.Qi=function(t,i){return yde(this,t,i)},p.Zi=function(t){return Pe(Ro,Jt,1,t,5,1)},p.Si=function(t){return this.g[t]},p.ed=function(t){return wR(this,t)},p.Ui=function(t,i){return Sue(this,t,i)},p.gc=function(){return this.i},p.Nc=function(){return LTe(this)},p.Oc=function(t){return c3e(this,t)},p.i=0;var A7e=R(Ha,"BasicEList",66),I7e=Bi(Ha,"TreeIterator");N(692,66,mpe),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.g==null&&!this.c?pTe(this):this.g==null||this.i!=0&&f(this.g[this.i-1],49).Ob()},p.Pb=function(){return XZ(this)},p.Qb=function(){if(!this.e)throw ie(new As("There is no valid object to remove."));this.e.Qb()},p.c=!1,R(Ha,"AbstractTreeIterator",692),N(599,692,mpe,_se),p.$i=function(t){var i;return i=f(t,57).Ch().Jc(),me(i,284)&&f(i,284).tl(new XC),i},R(uee,"ElkGraphUtil/PropertiesSkippingTreeIterator",599),N(943,1,{},XC),R(uee,"ElkGraphUtil/PropertiesSkippingTreeIterator/1",943);var SV,p1e,_V=R(uee,"ElkReflect",null);N(877,1,VT,JC),p.Pg=function(t){return oJ(),n2n(f(t,181))},R(uee,"ElkReflect/lambda$0$Type",877);var qE;Bi(Ha,"ResourceLocator"),N(1040,1,{}),R(Ha,"DelegatingResourceLocator",1040),N(1041,1040,{}),R("org.eclipse.emf.common","EMFPlugin",1041);var g1e=Bi(E3t,"Adapter"),mjn=Bi(E3t,"Notification");N(1138,1,Y4e),p._i=function(){return this.d},p.aj=function(t){},p.bj=function(t){this.d=t},p.cj=function(t){this.d==t&&(this.d=null)},p.d=null,R(FR,"AdapterImpl",1138),N(2039,70,S3t),p.Bi=function(t,i){return n3e(this,t,i)},p.Ci=function(t){var i,a,u;if(++this.j,t.dc())return!1;for(i=this.Bj(),u=t.Jc();u.Ob();)a=u.Pb(),this.oj(this.Wi(i,a)),++i;return!0},p.Di=function(t,i){qct(this,t,i)},p.Ei=function(t){gdt(this,t)},p.mj=function(){return this.pj()},p.$b=function(){tH(this,this.Bj(),this.Cj())},p.Gc=function(t){return this.rj(t)},p.Hc=function(t){return this.sj(t)},p.nj=function(t,i){this.yj().Rm()},p.oj=function(t){this.yj().Rm()},p.pj=function(){return this.yj()},p.qj=function(){this.yj().Rm()},p.rj=function(t){return this.yj().Rm()},p.sj=function(t){return this.yj().Rm()},p.tj=function(t){return this.yj().Rm()},p.uj=function(t){return this.yj().Rm()},p.vj=function(){return this.yj().Rm()},p.wj=function(t){return this.yj().Rm()},p.xj=function(){return this.yj().Rm()},p.zj=function(t){return this.yj().Rm()},p.Aj=function(t,i){return this.yj().Rm()},p.Bj=function(){return this.yj().Rm()},p.Cj=function(){return this.yj().Rm()},p.Dj=function(t){return this.yj().Rm()},p.Ej=function(){return this.yj().Rm()},p.Fb=function(t){return this.tj(t)},p.Xb=function(t){return this.Ti(t,this.uj(t))},p.Hb=function(){return this.vj()},p.bd=function(t){return this.wj(t)},p.dc=function(){return this.xj()},p.Qi=function(t,i){return bAe(this,t,i)},p.Si=function(t){return this.uj(t)},p.ed=function(t){return xX(this,t)},p.Kc=function(t){var i;return i=this.bd(t),i>=0?(this.ed(i),!0):!1},p.Ui=function(t,i){return this.Aj(t,this.Wi(t,i))},p.gc=function(){return this.Bj()},p.Nc=function(){return this.Cj()},p.Oc=function(t){return this.Dj(t)},p.Ib=function(){return this.Ej()},R(Ha,"DelegatingEList",2039),N(2040,2039,S3t),p.Bi=function(t,i){return YIe(this,t,i)},p.Ci=function(t){return this.Bi(this.Bj(),t)},p.Di=function(t,i){VEt(this,t,i)},p.Ei=function(t){DEt(this,t)},p.Ii=function(){return!this.Jj()},p.$b=function(){e$(this)},p.Fj=function(t,i,a,u,l){return new hft(this,t,i,a,u,l)},p.Gj=function(t){Zr(this.gj(),t)},p.Hj=function(){return null},p.Ij=function(){return-1},p.gj=function(){return null},p.Jj=function(){return!1},p.Kj=function(t,i){return i},p.Lj=function(t,i){return i},p.Mj=function(){return!1},p.Nj=function(){return!this.xj()},p.Qi=function(t,i){var a,u;return this.Mj()?(u=this.Nj(),a=bAe(this,t,i),this.Gj(this.Fj(7,rt(i),a,t,u)),a):bAe(this,t,i)},p.ed=function(t){var i,a,u,l;return this.Mj()?(a=null,u=this.Nj(),i=this.Fj(4,l=xX(this,t),null,t,u),this.Jj()&&l?(a=this.Lj(l,a),a?(a.kj(i),a.lj()):this.Gj(i)):a?(a.kj(i),a.lj()):this.Gj(i),l):(l=xX(this,t),this.Jj()&&l&&(a=this.Lj(l,null),a&&a.lj()),l)},p.Ui=function(t,i){return e_t(this,t,i)},R(FR,"DelegatingNotifyingListImpl",2040),N(150,1,dq),p.kj=function(t){return nAe(this,t)},p.lj=function(){yue(this)},p.dj=function(){return this.d},p.Hj=function(){return null},p.Oj=function(){return null},p.ej=function(t){return-1},p.fj=function(){return U2t(this)},p.gj=function(){return null},p.hj=function(){return TIe(this)},p.ij=function(){return this.o<0?this.o<-2?-2-this.o-1:-1:this.o},p.Pj=function(){return!1},p.jj=function(t){var i,a,u,l,g,m,v,S,T,A,L;switch(this.d){case 1:case 2:switch(l=t.dj(),l){case 1:case 2:if(g=t.gj(),Ae(g)===Ae(this.gj())&&this.ej(null)==t.ej(null))return this.g=t.fj(),t.dj()==1&&(this.d=1),!0}case 4:{switch(l=t.dj(),l){case 4:{if(g=t.gj(),Ae(g)===Ae(this.gj())&&this.ej(null)==t.ej(null))return T=hOe(this),S=this.o<0?this.o<-2?-2-this.o-1:-1:this.o,m=t.ij(),this.d=6,L=new OS(2),S<=m?(Qn(L,this.n),Qn(L,t.hj()),this.g=ce(se(pr,1),Pr,30,15,[this.o=S,m+1])):(Qn(L,t.hj()),Qn(L,this.n),this.g=ce(se(pr,1),Pr,30,15,[this.o=m,S])),this.n=L,T||(this.o=-2-this.o-1),!0;break}}break}case 6:{switch(l=t.dj(),l){case 4:{if(g=t.gj(),Ae(g)===Ae(this.gj())&&this.ej(null)==t.ej(null)){for(T=hOe(this),m=t.ij(),A=f(this.g,54),u=Pe(pr,Pr,30,A.length+1,15,1),i=0;i<A.length&&(v=A[i],v<=m);)u[i++]=v,++m;for(a=f(this.n,15),a._c(i,t.hj()),u[i]=m;++i<u.length;)u[i]=A[i-1];return this.g=u,T||(this.o=-2-u[0]),!0}break}}break}}return!1},p.Ib=function(){var t,i,a,u;switch(u=new sh(Vy(this.Om)+"@"+(i=Ti(this)>>>0,i.toString(16))),u.a+=" (eventType: ",this.d){case 1:{u.a+="SET";break}case 2:{u.a+="UNSET";break}case 3:{u.a+="ADD";break}case 5:{u.a+="ADD_MANY";break}case 4:{u.a+="REMOVE";break}case 6:{u.a+="REMOVE_MANY";break}case 7:{u.a+="MOVE";break}case 8:{u.a+="REMOVING_ADAPTER";break}case 9:{u.a+="RESOLVE";break}default:{Nae(u,this.d);break}}if(CSt(this)&&(u.a+=", touch: true"),u.a+=", position: ",Nae(u,this.o<0?this.o<-2?-2-this.o-1:-1:this.o),u.a+=", notifier: ",R7(u,this.gj()),u.a+=", feature: ",R7(u,this.Hj()),u.a+=", oldValue: ",R7(u,TIe(this)),u.a+=", newValue: ",this.d==6&&me(this.g,54)){for(a=f(this.g,54),u.a+="[",t=0;t<a.length;)u.a+=a[t],++t<a.length&&(u.a+=Fu);u.a+="]"}else R7(u,U2t(this));return u.a+=", isTouch: ",m0(u,CSt(this)),u.a+=", wasSet: ",m0(u,hOe(this)),u.a+=")",u.a},p.d=0,p.e=0,p.f=0,p.j=0,p.k=0,p.o=0,p.p=0,R(FR,"NotificationImpl",150),N(1152,150,dq,hft),p.Hj=function(){return this.a.Hj()},p.ej=function(t){return this.a.Ij()},p.gj=function(){return this.a.gj()},R(FR,"DelegatingNotifyingListImpl/1",1152),N(250,66,Im,c0,Fv),p.Ec=function(t){return U0t(this,f(t,370))},p.kj=function(t){return U0t(this,t)},p.lj=function(){var t,i,a;for(t=0;t<this.i;++t)i=f(this.g[t],370),a=i.gj(),a!=null&&i.dj()!=-1&&f(a,93).th(i)},p.Zi=function(t){return Pe(mjn,Jt,370,t,0,1)},R(FR,"NotificationChainImpl",250),N(1464,92,PCt),p.qh=function(){return this.e},p.sh=function(){return(this.f&1)!=0},p.f=1,R(FR,"NotifierImpl",1464),N(2037,66,Im),p.Bi=function(t,i){return Bde(this,t,i)},p.Ci=function(t){return this.Bi(this.i,t)},p.Di=function(t,i){YAe(this,t,i)},p.Ei=function(t){Ede(this,t)},p.Ii=function(){return!this.Jj()},p.$b=function(){Jn(this)},p.Fj=function(t,i,a,u,l){return new pft(this,t,i,a,u,l)},p.Gj=function(t){Zr(this.gj(),t)},p.Hj=function(){return null},p.Ij=function(){return-1},p.gj=function(){return null},p.Jj=function(){return!1},p.Qj=function(){return!1},p.Kj=function(t,i){return i},p.Lj=function(t,i){return i},p.Mj=function(){return!1},p.Nj=function(){return this.i!=0},p.Qi=function(t,i){return Cj(this,t,i)},p.ed=function(t){return PT(this,t)},p.Ui=function(t,i){return b_t(this,t,i)},p.Rj=function(t,i){return i},p.Sj=function(t,i){return i},p.Tj=function(t,i,a){return a},R(FR,"NotifyingListImpl",2037),N(1151,150,dq,pft),p.Hj=function(){return this.a.Hj()},p.ej=function(t){return this.a.Ij()},p.gj=function(){return this.a.gj()},R(FR,"NotifyingListImpl/1",1151),N(944,66,Im,Kct),p.Gc=function(t){return this.i>10?((!this.b||this.c.j!=this.a)&&(this.b=new H7(this),this.a=this.j),xp(this.b,t)):lL(this,t)},p.Vi=function(){return!0},p.a=0,R(Ha,"AbstractEList/1",944),N(301,97,_Q,rT),R(Ha,"AbstractEList/BasicIndexOutOfBoundsException",301),N(42,1,aa,Nn),p.Nb=function(t){va(this,t)},p.Uj=function(){if(this.i.j!=this.f)throw ie(new If)},p.Vj=function(){return Mn(this)},p.Ob=function(){return this.e!=this.i.gc()},p.Pb=function(){return this.Vj()},p.Qb=function(){Mj(this)},p.e=0,p.f=0,p.g=-1,R(Ha,"AbstractEList/EIterator",42),N(283,42,yb,HN,bce),p.Qb=function(){Mj(this)},p.Rb=function(t){Q0t(this,t)},p.Wj=function(){var t;try{return t=this.d.Xb(--this.e),this.Uj(),this.g=this.e,t}catch(i){throw i=go(i),me(i,97)?(this.Uj(),ie(new ic)):ie(i)}},p.Xj=function(t){Xvt(this,t)},p.Sb=function(){return this.e!=0},p.Tb=function(){return this.e},p.Ub=function(){return this.Wj()},p.Vb=function(){return this.e-1},p.Wb=function(t){this.Xj(t)},R(Ha,"AbstractEList/EListIterator",283),N(352,42,aa,GN),p.Vj=function(){return Ple(this)},p.Qb=function(){throw ie(new cr)},R(Ha,"AbstractEList/NonResolvingEIterator",352),N(389,283,yb,Qz,z_e),p.Rb=function(t){throw ie(new cr)},p.Vj=function(){var t;try{return t=this.c.Si(this.e),this.Uj(),this.g=this.e++,t}catch(i){throw i=go(i),me(i,97)?(this.Uj(),ie(new ic)):ie(i)}},p.Wj=function(){var t;try{return t=this.c.Si(--this.e),this.Uj(),this.g=this.e,t}catch(i){throw i=go(i),me(i,97)?(this.Uj(),ie(new ic)):ie(i)}},p.Qb=function(){throw ie(new cr)},p.Wb=function(t){throw ie(new cr)},R(Ha,"AbstractEList/NonResolvingEListIterator",389),N(2026,70,_3t),p.Bi=function(t,i){var a,u,l,g,m,v,S,T,A,L,F;if(l=i.gc(),l!=0){for(T=f(fn(this.a,4),128),A=T==null?0:T.length,F=A+l,u=ile(this,F),L=A-t,L>0&&au(T,t,u,t+l,L),S=i.Jc(),m=0;m<l;++m)v=S.Pb(),a=t+m,Cse(u,a,KP(this,v));for(aL(this,u),g=0;g<l;++g)v=u[t],this.Ji(t,v),++t;return!0}else return++this.j,!1},p.Ci=function(t){var i,a,u,l,g,m,v,S,T;if(u=t.gc(),u!=0){for(S=(a=f(fn(this.a,4),128),a==null?0:a.length),T=S+u,i=ile(this,T),v=t.Jc(),g=S;g<T;++g)m=v.Pb(),Cse(i,g,KP(this,m));for(aL(this,i),l=S;l<T;++l)m=i[l],this.Ji(l,m);return!0}else return++this.j,!1},p.Di=function(t,i){var a,u,l,g;u=f(fn(this.a,4),128),l=u==null?0:u.length,a=ile(this,l+1),g=KP(this,i),t!=l&&au(u,t,a,t+1,l-t),io(a,t,g),aL(this,a),this.Ji(t,i)},p.Ei=function(t){var i,a,u;u=(a=f(fn(this.a,4),128),a==null?0:a.length),i=ile(this,u+1),Cse(i,u,KP(this,t)),aL(this,i),this.Ji(u,t)},p.Fi=function(){return new Dpt(this)},p.Gi=function(){return new Ylt(this)},p.Hi=function(t){var i,a;if(a=(i=f(fn(this.a,4),128),i==null?0:i.length),t<0||t>a)throw ie(new rT(t,a));return new Adt(this,t)},p.$b=function(){var t,i;++this.j,t=f(fn(this.a,4),128),i=t==null?0:t.length,aL(this,null),fue(this,i,t)},p.Gc=function(t){var i,a,u,l,g;if(i=f(fn(this.a,4),128),i!=null){if(t!=null){for(u=i,l=0,g=u.length;l<g;++l)if(a=u[l],ei(t,a))return!0}else for(u=i,l=0,g=u.length;l<g;++l)if(a=u[l],Ae(a)===Ae(t))return!0}return!1},p.Xb=function(t){var i,a;if(i=f(fn(this.a,4),128),a=i==null?0:i.length,t>=a)throw ie(new rT(t,a));return i[t]},p.bd=function(t){var i,a,u;if(i=f(fn(this.a,4),128),i!=null){if(t!=null){for(a=0,u=i.length;a<u;++a)if(ei(t,i[a]))return a}else for(a=0,u=i.length;a<u;++a)if(Ae(i[a])===Ae(t))return a}return-1},p.dc=function(){return f(fn(this.a,4),128)==null},p.Jc=function(){return new Mpt(this)},p.cd=function(){return new Klt(this)},p.dd=function(t){var i,a;if(a=(i=f(fn(this.a,4),128),i==null?0:i.length),t<0||t>a)throw ie(new rT(t,a));return new Cdt(this,t)},p.Qi=function(t,i){var a,u,l;if(a=ivt(this),l=a==null?0:a.length,t>=l)throw ie(new Ru(ppe+t+TE+l));if(i>=l)throw ie(new Ru(gpe+i+TE+l));return u=a[i],t!=i&&(t<i?au(a,t,a,t+1,i-t):au(a,i+1,a,i,t-i),io(a,t,u),aL(this,a)),u},p.Si=function(t){return f(fn(this.a,4),128)[t]},p.ed=function(t){return w4n(this,t)},p.Ui=function(t,i){var a,u;return a=ivt(this),u=a[t],Cse(a,t,KP(this,i)),aL(this,a),u},p.gc=function(){var t;return t=f(fn(this.a,4),128),t==null?0:t.length},p.Nc=function(){var t,i,a;return t=f(fn(this.a,4),128),a=t==null?0:t.length,i=Pe(g1e,ype,412,a,0,1),a>0&&au(t,0,i,0,a),i},p.Oc=function(t){var i,a,u;return i=f(fn(this.a,4),128),u=i==null?0:i.length,u>0&&(t.length<u&&(a=BH(Fd(t).c,u),t=a),au(i,0,t,0,u)),t.length>u&&io(t,u,null),t};var YDt;R(Ha,"ArrayDelegatingEList",2026),N(1026,42,aa,Mpt),p.Uj=function(){if(this.b.j!=this.f||Ae(f(fn(this.b.a,4),128))!==Ae(this.a))throw ie(new If)},p.Qb=function(){Mj(this),this.a=f(fn(this.b.a,4),128)},R(Ha,"ArrayDelegatingEList/EIterator",1026),N(706,283,yb,Klt,Cdt),p.Uj=function(){if(this.b.j!=this.f||Ae(f(fn(this.b.a,4),128))!==Ae(this.a))throw ie(new If)},p.Xj=function(t){Xvt(this,t),this.a=f(fn(this.b.a,4),128)},p.Qb=function(){Mj(this),this.a=f(fn(this.b.a,4),128)},R(Ha,"ArrayDelegatingEList/EListIterator",706),N(1027,352,aa,Dpt),p.Uj=function(){if(this.b.j!=this.f||Ae(f(fn(this.b.a,4),128))!==Ae(this.a))throw ie(new If)},R(Ha,"ArrayDelegatingEList/NonResolvingEIterator",1027),N(707,389,yb,Ylt,Adt),p.Uj=function(){if(this.b.j!=this.f||Ae(f(fn(this.b.a,4),128))!==Ae(this.a))throw ie(new If)},R(Ha,"ArrayDelegatingEList/NonResolvingEListIterator",707),N(600,301,_Q,vse),R(Ha,"BasicEList/BasicIndexOutOfBoundsException",600),N(694,66,Im,ISe),p._c=function(t,i){throw ie(new cr)},p.Ec=function(t){throw ie(new cr)},p.ad=function(t,i){throw ie(new cr)},p.Fc=function(t){throw ie(new cr)},p.$b=function(){throw ie(new cr)},p.Yi=function(t){throw ie(new cr)},p.Jc=function(){return this.Fi()},p.cd=function(){return this.Gi()},p.dd=function(t){return this.Hi(t)},p.Qi=function(t,i){throw ie(new cr)},p.Ri=function(t,i){throw ie(new cr)},p.ed=function(t){throw ie(new cr)},p.Kc=function(t){throw ie(new cr)},p.fd=function(t,i){throw ie(new cr)},R(Ha,"BasicEList/UnmodifiableEList",694),N(705,1,{3:1,20:1,16:1,15:1,61:1,581:1}),p._c=function(t,i){Emn(this,t,f(i,43))},p.Ec=function(t){return lbn(this,f(t,43))},p.Ic=function(t){Oa(this,t)},p.Xb=function(t){return f(be(this.c,t),135)},p.Qi=function(t,i){return f(this.c.Qi(t,i),43)},p.Ri=function(t,i){Smn(this,t,f(i,43))},p.ed=function(t){return f(this.c.ed(t),43)},p.fd=function(t,i){return Swn(this,t,f(i,43))},p.gd=function(t){sE(this,t)},p.Lc=function(){return new Bt(this,16)},p.Mc=function(){return new Ft(null,new Bt(this,16))},p.ad=function(t,i){return this.c.ad(t,i)},p.Fc=function(t){return this.c.Fc(t)},p.$b=function(){this.c.$b()},p.Gc=function(t){return this.c.Gc(t)},p.Hc=function(t){return KH(this.c,t)},p.Yj=function(){var t,i,a;if(this.d==null){for(this.d=Pe(A7e,X4e,66,2*this.f+1,0,1),a=this.e,this.f=0,i=this.c.Jc();i.e!=i.i.gc();)t=f(i.Vj(),135),OZ(this,t);this.e=a}},p.Fb=function(t){return Eut(this,t)},p.Hb=function(){return rCe(this.c)},p.bd=function(t){return this.c.bd(t)},p.Zj=function(){this.c=new Int(this)},p.dc=function(){return this.f==0},p.Jc=function(){return this.c.Jc()},p.cd=function(){return this.c.cd()},p.dd=function(t){return this.c.dd(t)},p.$j=function(){return AH(this)},p._j=function(t,i,a){return new $ut(t,i,a)},p.ak=function(){return new n9},p.Kc=function(t){return sbt(this,t)},p.gc=function(){return this.f},p.hd=function(t,i){return new Xv(this.c,t,i)},p.Nc=function(){return this.c.Nc()},p.Oc=function(t){return this.c.Oc(t)},p.Ib=function(){return QCe(this.c)},p.e=0,p.f=0,R(Ha,"BasicEMap",705),N(1021,66,Im,Int),p.Ji=function(t,i){Qhn(this,f(i,135))},p.Mi=function(t,i,a){var u;++(u=this,f(i,135),u).a.e},p.Ni=function(t,i){epn(this,f(i,135))},p.Oi=function(t,i,a){Wmn(this,f(i,135),f(a,135))},p.Li=function(t,i){Zbt(this.a)},R(Ha,"BasicEMap/1",1021),N(1022,66,Im,n9),p.Zi=function(t){return Pe(bjn,k3t,606,t,0,1)},R(Ha,"BasicEMap/2",1022),N(1023,Ig,Pl,Ont),p.$b=function(){this.a.c.$b()},p.Gc=function(t){return Ele(this.a,t)},p.Jc=function(){return this.a.f==0?(gP(),kV.a):new pit(this.a)},p.Kc=function(t){var i;return i=this.a.f,yZ(this.a,t),this.a.f!=i},p.gc=function(){return this.a.f},R(Ha,"BasicEMap/3",1023),N(ep,31,jT,Nnt),p.$b=function(){this.a.c.$b()},p.Gc=function(t){return aSt(this.a,t)},p.Jc=function(){return this.a.f==0?(gP(),kV.a):new git(this.a)},p.gc=function(){return this.a.f},R(Ha,"BasicEMap/4",ep),N(1025,Ig,Pl,Rnt),p.$b=function(){this.a.c.$b()},p.Gc=function(t){var i,a,u,l,g,m,v,S,T;if(this.a.f>0&&me(t,43)&&(this.a.Yj(),S=f(t,43),v=S.jd(),l=v==null?0:Ti(v),g=C_e(this.a,l),i=this.a.d[g],i)){for(a=f(i.g,371),T=i.i,m=0;m<T;++m)if(u=a[m],u.xi()==l&&u.Fb(S))return!0}return!1},p.Jc=function(){return this.a.f==0?(gP(),kV.a):new Pce(this.a)},p.Kc=function(t){return JEt(this,t)},p.gc=function(){return this.a.f},R(Ha,"BasicEMap/5",1025),N(607,1,aa,Pce),p.Nb=function(t){va(this,t)},p.Ob=function(){return this.b!=-1},p.Pb=function(){var t;if(this.f.e!=this.c)throw ie(new If);if(this.b==-1)throw ie(new ic);return this.d=this.a,this.e=this.b,Nwt(this),t=f(this.f.d[this.d].g[this.e],135),this.bk(t)},p.Qb=function(){if(this.f.e!=this.c)throw ie(new If);if(this.e==-1)throw ie(new Tl);this.f.c.Kc(be(this.f.d[this.d],this.e)),this.c=this.f.e,this.e=-1,this.a==this.d&&this.b!=-1&&--this.b},p.bk=function(t){return t},p.a=0,p.b=-1,p.c=0,p.d=0,p.e=0,R(Ha,"BasicEMap/BasicEMapIterator",607),N(1019,607,aa,pit),p.bk=function(t){return t.jd()},R(Ha,"BasicEMap/BasicEMapKeyIterator",1019),N(1020,607,aa,git),p.bk=function(t){return t.kd()},R(Ha,"BasicEMap/BasicEMapValueIterator",1020),N(1018,1,YS,Pnt),p.wc=function(t){HH(this,t)},p.$b=function(){this.a.c.$b()},p._b=function(t){return Zat(this,t)},p.uc=function(t){return aSt(this.a,t)},p.vc=function(){return QSn(this.a)},p.Fb=function(t){return Eut(this.a,t)},p.xc=function(t){return Rp(this.a,t)},p.Hb=function(){return rCe(this.a.c)},p.dc=function(){return this.a.f==0},p.ec=function(){return e_n(this.a)},p.yc=function(t,i){return hG(this.a,t,i)},p.Ac=function(t){return yZ(this.a,t)},p.gc=function(){return this.a.f},p.Ib=function(){return QCe(this.a.c)},p.Bc=function(){return ZSn(this.a)},R(Ha,"BasicEMap/DelegatingMap",1018),N(606,1,{43:1,135:1,606:1},$ut),p.Fb=function(t){var i;return me(t,43)?(i=f(t,43),(this.b!=null?ei(this.b,i.jd()):Ae(this.b)===Ae(i.jd()))&&(this.c!=null?ei(this.c,i.kd()):Ae(this.c)===Ae(i.kd()))):!1},p.xi=function(){return this.a},p.jd=function(){return this.b},p.kd=function(){return this.c},p.Hb=function(){return this.a^(this.c==null?0:Ti(this.c))},p.yi=function(t){this.a=t},p.zi=function(t){throw ie(new Gy)},p.ld=function(t){var i;return i=this.c,this.c=t,i},p.Ib=function(){return this.b+"->"+this.c},p.a=0;var bjn=R(Ha,"BasicEMap/EntryImpl",606);N(531,1,{},Ys),R(Ha,"BasicEMap/View",531);var kV;N(763,1,{}),p.Fb=function(t){return oIe((Ht(),Ga),t)},p.Hb=function(){return hCe((Ht(),Ga))},p.Ib=function(){return Cg((Ht(),Ga))},R(Ha,"ECollections/BasicEmptyUnmodifiableEList",763),N(1287,1,yb,Ov),p.Nb=function(t){va(this,t)},p.Rb=function(t){throw ie(new cr)},p.Ob=function(){return!1},p.Sb=function(){return!1},p.Pb=function(){throw ie(new ic)},p.Tb=function(){return 0},p.Ub=function(){throw ie(new ic)},p.Vb=function(){return-1},p.Qb=function(){throw ie(new cr)},p.Wb=function(t){throw ie(new cr)},R(Ha,"ECollections/BasicEmptyUnmodifiableEList/1",1287),N(1285,763,{20:1,16:1,15:1,61:1},_rt),p._c=function(t,i){jit()},p.Ec=function(t){return $it()},p.ad=function(t,i){return Fit()},p.Fc=function(t){return Bit()},p.$b=function(){Uit()},p.Gc=function(t){return!1},p.Hc=function(t){return!1},p.Ic=function(t){Oa(this,t)},p.Xb=function(t){return PSe((Ht(),t)),null},p.bd=function(t){return-1},p.dc=function(){return!0},p.Jc=function(){return this.a},p.cd=function(){return this.a},p.dd=function(t){return this.a},p.Qi=function(t,i){return zit()},p.Ri=function(t,i){Hit()},p.ed=function(t){return Git()},p.Kc=function(t){return qit()},p.fd=function(t,i){return Vit()},p.gc=function(){return 0},p.gd=function(t){sE(this,t)},p.Lc=function(){return new Bt(this,16)},p.Mc=function(){return new Ft(null,new Bt(this,16))},p.hd=function(t,i){return Ht(),new Xv(Ga,t,i)},p.Nc=function(){return Cke((Ht(),Ga))},p.Oc=function(t){return Ht(),Rj(Ga,t)},R(Ha,"ECollections/EmptyUnmodifiableEList",1285),N(1286,763,{20:1,16:1,15:1,61:1,581:1},krt),p._c=function(t,i){jit()},p.Ec=function(t){return $it()},p.ad=function(t,i){return Fit()},p.Fc=function(t){return Bit()},p.$b=function(){Uit()},p.Gc=function(t){return!1},p.Hc=function(t){return!1},p.Ic=function(t){Oa(this,t)},p.Xb=function(t){return PSe((Ht(),t)),null},p.bd=function(t){return-1},p.dc=function(){return!0},p.Jc=function(){return this.a},p.cd=function(){return this.a},p.dd=function(t){return this.a},p.Qi=function(t,i){return zit()},p.Ri=function(t,i){Hit()},p.ed=function(t){return Git()},p.Kc=function(t){return qit()},p.fd=function(t,i){return Vit()},p.gc=function(){return 0},p.gd=function(t){sE(this,t)},p.Lc=function(){return new Bt(this,16)},p.Mc=function(){return new Ft(null,new Bt(this,16))},p.hd=function(t,i){return Ht(),new Xv(Ga,t,i)},p.Nc=function(){return Cke((Ht(),Ga))},p.Oc=function(t){return Ht(),Rj(Ga,t)},p.$j=function(){return Ht(),Ht(),Cb},R(Ha,"ECollections/EmptyUnmodifiableEMap",1286);var O7e=Bi(Ha,"Enumerator"),Fne;N(286,1,{286:1},Ode),p.Fb=function(t){var i;return this===t?!0:me(t,286)?(i=f(t,286),this.f==i.f&&zvn(this.i,i.i)&&ice(this.a,(this.f&256)!=0?(i.f&256)!=0?i.a:null:(i.f&256)!=0?null:i.a)&&ice(this.d,i.d)&&ice(this.g,i.g)&&ice(this.e,i.e)&&G3n(this,i)):!1},p.Hb=function(){return this.f},p.Ib=function(){return FSt(this)},p.f=0;var XDt=0,JDt=0,ZDt=0,QDt=0,N7e=0,R7e=0,P7e=0,L7e=0,D7e=0,eMt,zF=0,HF=0,tMt=0,nMt=0,Bne,M7e;R(Ha,"URI",286),N(1085,45,sA,Trt),p.yc=function(t,i){return f(Os(this,dr(t),f(i,286)),286)},R(Ha,"URI/URICache",1085),N(490,66,Im,Yd,OX),p.Pi=function(){return!0},R(Ha,"UniqueEList",490),N(573,63,C1,xJ),R(Ha,"WrappedException",573);var Ir=Bi(zf,C3t),vx=Bi(zf,A3t),El=Bi(zf,I3t),wx=Bi(zf,O3t),Up=Bi(zf,N3t),kh=Bi(zf,"EClass"),m1e=Bi(zf,"EDataType"),rMt;N(1193,45,sA,xrt),p.xc=function(t){return ta(t)?wu(this,t):ac(Is(this.f,t))},R(zf,"EDataType/Internal/ConversionDelegate/Factory/Registry/Impl",1193);var Une=Bi(zf,"EEnum"),j1=Bi(zf,R3t),gs=Bi(zf,P3t),Th=Bi(zf,L3t),xh,S_=Bi(zf,D3t),yx=Bi(zf,M3t);N(1017,1,{},UU),p.Ib=function(){return"NIL"},R(zf,"EStructuralFeature/Internal/DynamicValueHolder/1",1017);var iMt;N(1016,45,sA,Crt),p.xc=function(t){return ta(t)?wu(this,t):ac(Is(this.f,t))},R(zf,"EStructuralFeature/Internal/SettingDelegate/Factory/Registry/Impl",1016);var nl=Bi(zf,j3t),l4=Bi(zf,"EValidator/PatternMatcher"),j7e,$7e,sn,Y0,Ex,Mw,oMt,aMt,sMt,jw,X0,$w,__,Ug,cMt,uMt,Ch,J0,lMt,Z0,Sx,WA,Za,dMt,fMt,k_,zne=Bi(Oi,"FeatureMap/Entry");N(530,1,{74:1},KY),p.Ik=function(){return this.a},p.kd=function(){return this.b},R(un,"BasicEObjectImpl/1",530),N(1015,1,Epe,Vat),p.Ck=function(t){return sue(this.a,this.b,t)},p.Nj=function(){return Oft(this.a,this.b)},p.Wb=function(t){hTe(this.a,this.b,t)},p.Dk=function(){Dwn(this.a,this.b)},R(un,"BasicEObjectImpl/4",1015),N(2027,1,{113:1}),p.Jk=function(t){this.e=t==0?hMt:Pe(Ro,Jt,1,t,5,1)},p.hi=function(t){return this.e[t]},p.ii=function(t,i){this.e[t]=i},p.ji=function(t){this.e[t]=null},p.Kk=function(){return this.c},p.Lk=function(){throw ie(new cr)},p.Mk=function(){throw ie(new cr)},p.Nk=function(){return this.d},p.Ok=function(){return this.e!=null},p.Pk=function(t){this.c=t},p.Qk=function(t){throw ie(new cr)},p.Rk=function(t){throw ie(new cr)},p.Sk=function(t){this.d=t};var hMt;R(un,"BasicEObjectImpl/EPropertiesHolderBaseImpl",2027),N(192,2027,{113:1},ll),p.Lk=function(){return this.a},p.Mk=function(){return this.b},p.Qk=function(t){this.a=t},p.Rk=function(t){this.b=t},R(un,"BasicEObjectImpl/EPropertiesHolderImpl",192),N(500,98,jCt,ZC),p.qh=function(){return this.f},p.vh=function(){return this.k},p.xh=function(t,i){this.g=t,this.i=i},p.zh=function(){return(this.j&2)==0?this.ei():this.Wh().Kk()},p.Bh=function(){return this.i},p.sh=function(){return(this.j&1)!=0},p.Lh=function(){return this.g},p.Rh=function(){return(this.j&4)!=0},p.Wh=function(){return!this.k&&(this.k=new ll),this.k},p.$h=function(t){this.Wh().Pk(t),t?this.j|=2:this.j&=-3},p.ai=function(t){this.Wh().Rk(t),t?this.j|=4:this.j&=-5},p.ei=function(){return(Wv(),sn).S},p.i=0,p.j=1,R(un,"EObjectImpl",500),N(778,500,{109:1,93:1,92:1,57:1,113:1,52:1,98:1},lke),p.hi=function(t){return this.e[t]},p.ii=function(t,i){this.e[t]=i},p.ji=function(t){this.e[t]=null},p.zh=function(){return this.d},p.Eh=function(t){return Pi(this.d,t)},p.Gh=function(){return this.d},p.Kh=function(){return this.e!=null},p.Wh=function(){return!this.k&&(this.k=new r9),this.k},p.$h=function(t){this.d=t},p.di=function(){var t;return this.e==null&&(t=jn(this.d),this.e=t==0?pMt:Pe(Ro,Jt,1,t,5,1)),this},p.fi=function(){return 0};var pMt;R(un,"DynamicEObjectImpl",778),N(1462,778,{109:1,43:1,93:1,92:1,135:1,57:1,113:1,52:1,98:1},dlt),p.Fb=function(t){return this===t},p.Hb=function(){return bS(this)},p.$h=function(t){this.d=t,this.b=kG(t,"key"),this.c=kG(t,k$)},p.xi=function(){var t;return this.a==-1&&(t=Eue(this,this.b),this.a=t==null?0:Ti(t)),this.a},p.jd=function(){return Eue(this,this.b)},p.kd=function(){return Eue(this,this.c)},p.yi=function(t){this.a=t},p.zi=function(t){hTe(this,this.b,t)},p.ld=function(t){var i;return i=Eue(this,this.c),hTe(this,this.c,t),i},p.a=0,R(un,"DynamicEObjectImpl/BasicEMapEntry",1462),N(1463,1,{113:1},r9),p.Jk=function(t){throw ie(new cr)},p.hi=function(t){throw ie(new cr)},p.ii=function(t,i){throw ie(new cr)},p.ji=function(t){throw ie(new cr)},p.Kk=function(){throw ie(new cr)},p.Lk=function(){return this.a},p.Mk=function(){return this.b},p.Nk=function(){return this.c},p.Ok=function(){throw ie(new cr)},p.Pk=function(t){throw ie(new cr)},p.Qk=function(t){this.a=t},p.Rk=function(t){this.b=t},p.Sk=function(t){this.c=t},R(un,"DynamicEObjectImpl/DynamicEPropertiesHolderImpl",1463),N(503,157,{109:1,93:1,92:1,582:1,154:1,57:1,113:1,52:1,98:1,503:1,157:1,117:1,118:1},Wu),p.wh=function(t){return F3e(this,t)},p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.d;case 2:return a?(!this.b&&(this.b=new Md((Ut(),Za),xc,this)),this.b):(!this.b&&(this.b=new Md((Ut(),Za),xc,this)),AH(this.b));case 3:return Pft(this);case 4:return!this.a&&(this.a=new So(Dw,this,4)),this.a;case 5:return!this.c&&(this.c=new I3(Dw,this,5)),this.c}return Lf(this,t-jn((Ut(),Y0)),Vt((u=f(fn(this,16),29),u||Y0),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 3:return this.Cb&&(a=(l=this.Db>>16,l>=0?F3e(this,a):this.Cb.Ph(this,-1-l,null,a))),kke(this,f(t,154),a)}return g=f(Vt((u=f(fn(this,16),29),u||(Ut(),Y0)),i),69),g.tk().wk(this,Xu(this),i-jn((Ut(),Y0)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 2:return!this.b&&(this.b=new Md((Ut(),Za),xc,this)),pX(this.b,t,a);case 3:return kke(this,null,a);case 4:return!this.a&&(this.a=new So(Dw,this,4)),za(this.a,t,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),Y0)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),Y0)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.d!=null;case 2:return!!this.b&&this.b.f!=0;case 3:return!!Pft(this);case 4:return!!this.a&&this.a.i!=0;case 5:return!!this.c&&this.c.i!=0}return Pf(this,t-jn((Ut(),Y0)),Vt((i=f(fn(this,16),29),i||Y0),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:Ivn(this,dr(i));return;case 2:!this.b&&(this.b=new Md((Ut(),Za),xc,this)),XJ(this.b,i);return;case 3:I2t(this,f(i,154));return;case 4:!this.a&&(this.a=new So(Dw,this,4)),Jn(this.a),!this.a&&(this.a=new So(Dw,this,4)),Qi(this.a,f(i,16));return;case 5:!this.c&&(this.c=new I3(Dw,this,5)),Jn(this.c),!this.c&&(this.c=new I3(Dw,this,5)),Qi(this.c,f(i,16));return}$f(this,t-jn((Ut(),Y0)),Vt((a=f(fn(this,16),29),a||Y0),t),i)},p.ei=function(){return Ut(),Y0},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:$xe(this,null);return;case 2:!this.b&&(this.b=new Md((Ut(),Za),xc,this)),this.b.c.$b();return;case 3:I2t(this,null);return;case 4:!this.a&&(this.a=new So(Dw,this,4)),Jn(this.a);return;case 5:!this.c&&(this.c=new I3(Dw,this,5)),Jn(this.c);return}jf(this,t-jn((Ut(),Y0)),Vt((i=f(fn(this,16),29),i||Y0),t))},p.Ib=function(){return _0t(this)},p.d=null,R(un,"EAnnotationImpl",503),N(140,705,J4e,Ol),p.Di=function(t,i){Ygn(this,t,f(i,43))},p.Tk=function(t,i){return t1n(this,f(t,43),i)},p.Xi=function(t){return f(f(this.c,71).Xi(t),135)},p.Fi=function(){return f(this.c,71).Fi()},p.Gi=function(){return f(this.c,71).Gi()},p.Hi=function(t){return f(this.c,71).Hi(t)},p.Uk=function(t,i){return pX(this,t,i)},p.Ck=function(t){return f(this.c,76).Ck(t)},p.Zj=function(){},p.Nj=function(){return f(this.c,76).Nj()},p._j=function(t,i,a){var u;return u=f(cf(this.b).si().oi(this.b),135),u.yi(t),u.zi(i),u.ld(a),u},p.ak=function(){return new r2e(this)},p.Wb=function(t){XJ(this,t)},p.Dk=function(){f(this.c,76).Dk()},R(Oi,"EcoreEMap",140),N(167,140,J4e,Md),p.Yj=function(){var t,i,a,u,l,g;if(this.d==null){for(g=Pe(A7e,X4e,66,2*this.f+1,0,1),a=this.c.Jc();a.e!=a.i.gc();)i=f(a.Vj(),135),u=i.xi(),l=(u&Qr)%g.length,t=g[l],!t&&(t=g[l]=new r2e(this)),t.Ec(i);this.d=g}},R(un,"EAnnotationImpl/1",167),N(290,436,{109:1,93:1,92:1,154:1,197:1,57:1,113:1,467:1,52:1,98:1,157:1,290:1,117:1,118:1}),p.Hh=function(t,i,a){var u,l;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return mn(),(this.Bb&256)!=0;case 3:return mn(),(this.Bb&512)!=0;case 4:return rt(this.s);case 5:return rt(this.t);case 6:return mn(),!!this.Gk();case 7:return mn(),l=this.s,l>=1;case 8:return i?ph(this):this.r;case 9:return this.q}return Lf(this,t-jn(this.ei()),Vt((u=f(fn(this,16),29),u||this.ei()),t),i,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 9:return yce(this,a)}return l=f(Vt((u=f(fn(this,16),29),u||this.ei()),i),69),l.tk().xk(this,Xu(this),i-jn(this.ei()),t,a)},p.Sh=function(t){var i,a;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return(this.Bb&256)==0;case 3:return(this.Bb&512)==0;case 4:return this.s!=0;case 5:return this.t!=1;case 6:return this.Gk();case 7:return a=this.s,a>=1;case 8:return!!this.r&&!this.q.e&&ES(this.q).i==0;case 9:return!!this.q&&!(this.r&&!this.q.e&&ES(this.q).i==0)}return Pf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.Zh=function(t,i){var a,u;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:this.qi(dr(i));return;case 2:O0(this,lt(ut(i)));return;case 3:N0(this,lt(ut(i)));return;case 4:x0(this,f(i,17).a);return;case 5:this.Wk(f(i,17).a);return;case 8:fE(this,f(i,141));return;case 9:u=xg(this,f(i,87),null),u&&u.lj();return}$f(this,t-jn(this.ei()),Vt((a=f(fn(this,16),29),a||this.ei()),t),i)},p.ei=function(){return Ut(),fMt},p.gi=function(t){var i,a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:this.qi(null);return;case 2:O0(this,!0);return;case 3:N0(this,!0);return;case 4:x0(this,0);return;case 5:this.Wk(1);return;case 8:fE(this,null);return;case 9:a=xg(this,null,null),a&&a.lj();return}jf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.li=function(){ph(this),this.Bb|=1},p.Ek=function(){return ph(this)},p.Fk=function(){return this.t},p.Gk=function(){var t;return t=this.t,t>1||t==-1},p.Pi=function(){return(this.Bb&512)!=0},p.Vk=function(t,i){return PCe(this,t,i)},p.Wk=function(t){vT(this,t)},p.Ib=function(){return XAe(this)},p.s=0,p.t=1,R(un,"ETypedElementImpl",290),N(448,290,{109:1,93:1,92:1,154:1,197:1,57:1,178:1,69:1,113:1,467:1,52:1,98:1,157:1,448:1,290:1,117:1,118:1,677:1}),p.wh=function(t){return dwt(this,t)},p.Hh=function(t,i,a){var u,l;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return mn(),(this.Bb&256)!=0;case 3:return mn(),(this.Bb&512)!=0;case 4:return rt(this.s);case 5:return rt(this.t);case 6:return mn(),!!this.Gk();case 7:return mn(),l=this.s,l>=1;case 8:return i?ph(this):this.r;case 9:return this.q;case 10:return mn(),(this.Bb&ep)!=0;case 11:return mn(),(this.Bb&pw)!=0;case 12:return mn(),(this.Bb&FT)!=0;case 13:return this.j;case 14:return mL(this);case 15:return mn(),(this.Bb&Ml)!=0;case 16:return mn(),(this.Bb&Sm)!=0;case 17:return dT(this)}return Lf(this,t-jn(this.ei()),Vt((u=f(fn(this,16),29),u||this.ei()),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 17:return this.Cb&&(a=(l=this.Db>>16,l>=0?dwt(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,17,a)}return g=f(Vt((u=f(fn(this,16),29),u||this.ei()),i),69),g.tk().wk(this,Xu(this),i-jn(this.ei()),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 9:return yce(this,a);case 17:return hf(this,null,17,a)}return l=f(Vt((u=f(fn(this,16),29),u||this.ei()),i),69),l.tk().xk(this,Xu(this),i-jn(this.ei()),t,a)},p.Sh=function(t){var i,a;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return(this.Bb&256)==0;case 3:return(this.Bb&512)==0;case 4:return this.s!=0;case 5:return this.t!=1;case 6:return this.Gk();case 7:return a=this.s,a>=1;case 8:return!!this.r&&!this.q.e&&ES(this.q).i==0;case 9:return!!this.q&&!(this.r&&!this.q.e&&ES(this.q).i==0);case 10:return(this.Bb&ep)==0;case 11:return(this.Bb&pw)!=0;case 12:return(this.Bb&FT)!=0;case 13:return this.j!=null;case 14:return mL(this)!=null;case 15:return(this.Bb&Ml)!=0;case 16:return(this.Bb&Sm)!=0;case 17:return!!dT(this)}return Pf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.Zh=function(t,i){var a,u;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:Wce(this,dr(i));return;case 2:O0(this,lt(ut(i)));return;case 3:N0(this,lt(ut(i)));return;case 4:x0(this,f(i,17).a);return;case 5:this.Wk(f(i,17).a);return;case 8:fE(this,f(i,141));return;case 9:u=xg(this,f(i,87),null),u&&u.lj();return;case 10:QP(this,lt(ut(i)));return;case 11:tL(this,lt(ut(i)));return;case 12:eL(this,lt(ut(i)));return;case 13:NSe(this,dr(i));return;case 15:nL(this,lt(ut(i)));return;case 16:rL(this,lt(ut(i)));return}$f(this,t-jn(this.ei()),Vt((a=f(fn(this,16),29),a||this.ei()),t),i)},p.ei=function(){return Ut(),dMt},p.gi=function(t){var i,a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:me(this.Cb,88)&&OT(pd(f(this.Cb,88)),4),ju(this,null);return;case 2:O0(this,!0);return;case 3:N0(this,!0);return;case 4:x0(this,0);return;case 5:this.Wk(1);return;case 8:fE(this,null);return;case 9:a=xg(this,null,null),a&&a.lj();return;case 10:QP(this,!0);return;case 11:tL(this,!1);return;case 12:eL(this,!1);return;case 13:this.i=null,qJ(this,null);return;case 15:nL(this,!1);return;case 16:rL(this,!1);return}jf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.li=function(){xP(Ns((Rl(),Ea),this)),ph(this),this.Bb|=1},p.mk=function(){return this.f},p.fk=function(){return mL(this)},p.nk=function(){return dT(this)},p.rk=function(){return null},p.Xk=function(){return this.k},p.Ij=function(){return this.n},p.sk=function(){return jZ(this)},p.tk=function(){var t,i,a,u,l,g,m,v,S;return this.p||(a=dT(this),(a.i==null&&Em(a),a.i).length,u=this.rk(),u&&jn(dT(u)),l=ph(this),m=l.hk(),t=m?(m.i&1)!=0?m==kd?to:m==pr?ba:m==kx?eD:m==sa?mo:m==x_?i_:m==JA?o_:m==$l?BR:D$:m:null,i=mL(this),v=l.fk(),hCn(this),(this.Bb&Sm)!=0&&((g=V3e((Rl(),Ea),a))&&g!=this||(g=rR(Ns(Ea,this))))?this.p=new Kat(this,g):this.Gk()?this.Zk()?u?(this.Bb&Ml)!=0?t?this.$k()?this.p=new eE(47,t,this,u):this.p=new eE(5,t,this,u):this.$k()?this.p=new aE(46,this,u):this.p=new aE(4,this,u):t?this.$k()?this.p=new eE(49,t,this,u):this.p=new eE(7,t,this,u):this.$k()?this.p=new aE(48,this,u):this.p=new aE(6,this,u):(this.Bb&Ml)!=0?t?t==AE?this.p=new y0(50,WDt,this):this.$k()?this.p=new y0(43,t,this):this.p=new y0(1,t,this):this.$k()?this.p=new S0(42,this):this.p=new S0(0,this):t?t==AE?this.p=new y0(41,WDt,this):this.$k()?this.p=new y0(45,t,this):this.p=new y0(3,t,this):this.$k()?this.p=new S0(44,this):this.p=new S0(2,this):me(l,155)?t==zne?this.p=new S0(40,this):(this.Bb&512)!=0?(this.Bb&Ml)!=0?t?this.p=new y0(9,t,this):this.p=new S0(8,this):t?this.p=new y0(11,t,this):this.p=new S0(10,this):(this.Bb&Ml)!=0?t?this.p=new y0(13,t,this):this.p=new S0(12,this):t?this.p=new y0(15,t,this):this.p=new S0(14,this):u?(S=u.t,S>1||S==-1?this.$k()?(this.Bb&Ml)!=0?t?this.p=new eE(25,t,this,u):this.p=new aE(24,this,u):t?this.p=new eE(27,t,this,u):this.p=new aE(26,this,u):(this.Bb&Ml)!=0?t?this.p=new eE(29,t,this,u):this.p=new aE(28,this,u):t?this.p=new eE(31,t,this,u):this.p=new aE(30,this,u):this.$k()?(this.Bb&Ml)!=0?t?this.p=new eE(33,t,this,u):this.p=new aE(32,this,u):t?this.p=new eE(35,t,this,u):this.p=new aE(34,this,u):(this.Bb&Ml)!=0?t?this.p=new eE(37,t,this,u):this.p=new aE(36,this,u):t?this.p=new eE(39,t,this,u):this.p=new aE(38,this,u)):this.$k()?(this.Bb&Ml)!=0?t?this.p=new y0(17,t,this):this.p=new S0(16,this):t?this.p=new y0(19,t,this):this.p=new S0(18,this):(this.Bb&Ml)!=0?t?this.p=new y0(21,t,this):this.p=new S0(20,this):t?this.p=new y0(23,t,this):this.p=new S0(22,this):this.Yk()?this.$k()?this.p=new Gut(f(l,29),this,u):this.p=new dTe(f(l,29),this,u):me(l,155)?t==zne?this.p=new S0(40,this):(this.Bb&Ml)!=0?t?this.p=new jlt(i,v,this,(_le(),m==pr?q7e:m==kd?B7e:m==x_?V7e:m==kx?G7e:m==sa?H7e:m==JA?W7e:m==$l?U7e:m==Xf?z7e:w1e)):this.p=new Jdt(f(l,155),i,v,this):t?this.p=new Mlt(i,v,this,(_le(),m==pr?q7e:m==kd?B7e:m==x_?V7e:m==kx?G7e:m==sa?H7e:m==JA?W7e:m==$l?U7e:m==Xf?z7e:w1e)):this.p=new Xdt(f(l,155),i,v,this):this.Zk()?u?(this.Bb&Ml)!=0?this.$k()?this.p=new Uut(f(l,29),this,u):this.p=new Q_e(f(l,29),this,u):this.$k()?this.p=new But(f(l,29),this,u):this.p=new Xse(f(l,29),this,u):(this.Bb&Ml)!=0?this.$k()?this.p=new jct(f(l,29),this):this.p=new m_e(f(l,29),this):this.$k()?this.p=new Mct(f(l,29),this):this.p=new Mse(f(l,29),this):this.$k()?u?(this.Bb&Ml)!=0?this.p=new zut(f(l,29),this,u):this.p=new Z_e(f(l,29),this,u):(this.Bb&Ml)!=0?this.p=new $ct(f(l,29),this):this.p=new g_e(f(l,29),this):u?(this.Bb&Ml)!=0?this.p=new Hut(f(l,29),this,u):this.p=new eke(f(l,29),this,u):(this.Bb&Ml)!=0?this.p=new Dct(f(l,29),this):this.p=new NX(f(l,29),this)),this.p},p.ok=function(){return(this.Bb&ep)!=0},p.Yk=function(){return!1},p.Zk=function(){return!1},p.pk=function(){return(this.Bb&Sm)!=0},p.uk=function(){return _ue(this)},p.$k=function(){return!1},p.qk=function(){return(this.Bb&Ml)!=0},p._k=function(t){this.k=t},p.qi=function(t){Wce(this,t)},p.Ib=function(){return sQ(this)},p.e=!1,p.n=0,R(un,"EStructuralFeatureImpl",448),N(330,448,{109:1,93:1,92:1,35:1,154:1,197:1,57:1,178:1,69:1,113:1,467:1,52:1,98:1,330:1,157:1,448:1,290:1,117:1,118:1,677:1},pae),p.Hh=function(t,i,a){var u,l;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return mn(),(this.Bb&256)!=0;case 3:return mn(),(this.Bb&512)!=0;case 4:return rt(this.s);case 5:return rt(this.t);case 6:return mn(),!!GAe(this);case 7:return mn(),l=this.s,l>=1;case 8:return i?ph(this):this.r;case 9:return this.q;case 10:return mn(),(this.Bb&ep)!=0;case 11:return mn(),(this.Bb&pw)!=0;case 12:return mn(),(this.Bb&FT)!=0;case 13:return this.j;case 14:return mL(this);case 15:return mn(),(this.Bb&Ml)!=0;case 16:return mn(),(this.Bb&Sm)!=0;case 17:return dT(this);case 18:return mn(),(this.Bb&Dc)!=0;case 19:return i?que(this):Vpt(this)}return Lf(this,t-jn((Ut(),Ex)),Vt((u=f(fn(this,16),29),u||Ex),t),i,a)},p.Sh=function(t){var i,a;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return(this.Bb&256)==0;case 3:return(this.Bb&512)==0;case 4:return this.s!=0;case 5:return this.t!=1;case 6:return GAe(this);case 7:return a=this.s,a>=1;case 8:return!!this.r&&!this.q.e&&ES(this.q).i==0;case 9:return!!this.q&&!(this.r&&!this.q.e&&ES(this.q).i==0);case 10:return(this.Bb&ep)==0;case 11:return(this.Bb&pw)!=0;case 12:return(this.Bb&FT)!=0;case 13:return this.j!=null;case 14:return mL(this)!=null;case 15:return(this.Bb&Ml)!=0;case 16:return(this.Bb&Sm)!=0;case 17:return!!dT(this);case 18:return(this.Bb&Dc)!=0;case 19:return!!Vpt(this)}return Pf(this,t-jn((Ut(),Ex)),Vt((i=f(fn(this,16),29),i||Ex),t))},p.Zh=function(t,i){var a,u;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:Wce(this,dr(i));return;case 2:O0(this,lt(ut(i)));return;case 3:N0(this,lt(ut(i)));return;case 4:x0(this,f(i,17).a);return;case 5:vit(this,f(i,17).a);return;case 8:fE(this,f(i,141));return;case 9:u=xg(this,f(i,87),null),u&&u.lj();return;case 10:QP(this,lt(ut(i)));return;case 11:tL(this,lt(ut(i)));return;case 12:eL(this,lt(ut(i)));return;case 13:NSe(this,dr(i));return;case 15:nL(this,lt(ut(i)));return;case 16:rL(this,lt(ut(i)));return;case 18:gle(this,lt(ut(i)));return}$f(this,t-jn((Ut(),Ex)),Vt((a=f(fn(this,16),29),a||Ex),t),i)},p.ei=function(){return Ut(),Ex},p.gi=function(t){var i,a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:me(this.Cb,88)&&OT(pd(f(this.Cb,88)),4),ju(this,null);return;case 2:O0(this,!0);return;case 3:N0(this,!0);return;case 4:x0(this,0);return;case 5:this.b=0,vT(this,1);return;case 8:fE(this,null);return;case 9:a=xg(this,null,null),a&&a.lj();return;case 10:QP(this,!0);return;case 11:tL(this,!1);return;case 12:eL(this,!1);return;case 13:this.i=null,qJ(this,null);return;case 15:nL(this,!1);return;case 16:rL(this,!1);return;case 18:gle(this,!1);return}jf(this,t-jn((Ut(),Ex)),Vt((i=f(fn(this,16),29),i||Ex),t))},p.li=function(){que(this),xP(Ns((Rl(),Ea),this)),ph(this),this.Bb|=1},p.Gk=function(){return GAe(this)},p.Vk=function(t,i){return this.b=0,this.a=null,PCe(this,t,i)},p.Wk=function(t){vit(this,t)},p.Ib=function(){var t;return(this.Db&64)!=0?sQ(this):(t=new sh(sQ(this)),t.a+=" (iD: ",m0(t,(this.Bb&Dc)!=0),t.a+=")",t.a)},p.b=0,R(un,"EAttributeImpl",330),N(357,436,{109:1,93:1,92:1,141:1,154:1,197:1,57:1,113:1,52:1,98:1,357:1,157:1,117:1,118:1,676:1}),p.al=function(t){return t.zh()==this},p.wh=function(t){return Yle(this,t)},p.xh=function(t,i){this.w=null,this.Db=i<<16|this.Db&255,this.Cb=t},p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return this.D!=null?this.D:this.B;case 3:return US(this);case 4:return this.fk();case 5:return this.F;case 6:return i?cf(this):IP(this);case 7:return!this.A&&(this.A=new xl(nl,this,7)),this.A}return Lf(this,t-jn(this.ei()),Vt((u=f(fn(this,16),29),u||this.ei()),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 6:return this.Cb&&(a=(l=this.Db>>16,l>=0?Yle(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,6,a)}return g=f(Vt((u=f(fn(this,16),29),u||this.ei()),i),69),g.tk().wk(this,Xu(this),i-jn(this.ei()),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 6:return hf(this,null,6,a);case 7:return!this.A&&(this.A=new xl(nl,this,7)),za(this.A,t,a)}return l=f(Vt((u=f(fn(this,16),29),u||this.ei()),i),69),l.tk().xk(this,Xu(this),i-jn(this.ei()),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return this.D!=null&&this.D==this.F;case 3:return!!US(this);case 4:return this.fk()!=null;case 5:return this.F!=null&&this.F!=this.D&&this.F!=this.B;case 6:return!!IP(this);case 7:return!!this.A&&this.A.i!=0}return Pf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:aJ(this,dr(i));return;case 2:Sse(this,dr(i));return;case 5:_L(this,dr(i));return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A),!this.A&&(this.A=new xl(nl,this,7)),Qi(this.A,f(i,16));return}$f(this,t-jn(this.ei()),Vt((a=f(fn(this,16),29),a||this.ei()),t),i)},p.ei=function(){return Ut(),oMt},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:me(this.Cb,184)&&(f(this.Cb,184).tb=null),ju(this,null);return;case 2:YP(this,null),jP(this,this.D);return;case 5:_L(this,null);return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A);return}jf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.ek=function(){var t;return this.G==-1&&(this.G=(t=cf(this),t?P0(t.ri(),this):-1)),this.G},p.fk=function(){return null},p.gk=function(){return cf(this)},p.bl=function(){return this.v},p.hk=function(){return US(this)},p.ik=function(){return this.D!=null?this.D:this.B},p.jk=function(){return this.F},p.ck=function(t){return Fde(this,t)},p.cl=function(t){this.v=t},p.dl=function(t){Nbt(this,t)},p.el=function(t){this.C=t},p.qi=function(t){aJ(this,t)},p.Ib=function(){return mZ(this)},p.C=null,p.D=null,p.G=-1,R(un,"EClassifierImpl",357),N(88,357,{109:1,93:1,92:1,29:1,141:1,154:1,197:1,57:1,113:1,52:1,98:1,88:1,357:1,157:1,468:1,117:1,118:1,676:1},$5),p.al=function(t){return Fbn(this,t.zh())},p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return this.D!=null?this.D:this.B;case 3:return US(this);case 4:return null;case 5:return this.F;case 6:return i?cf(this):IP(this);case 7:return!this.A&&(this.A=new xl(nl,this,7)),this.A;case 8:return mn(),(this.Bb&256)!=0;case 9:return mn(),(this.Bb&512)!=0;case 10:return Bs(this);case 11:return!this.q&&(this.q=new Be(Th,this,11,10)),this.q;case 12:return rA(this);case 13:return Jj(this);case 14:return Jj(this),this.r;case 15:return rA(this),this.k;case 16:return PAe(this);case 17:return Hde(this);case 18:return Em(this);case 19:return ZZ(this);case 20:return rA(this),this.o;case 21:return!this.s&&(this.s=new Be(El,this,21,17)),this.s;case 22:return iu(this);case 23:return Ide(this)}return Lf(this,t-jn((Ut(),Mw)),Vt((u=f(fn(this,16),29),u||Mw),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 6:return this.Cb&&(a=(l=this.Db>>16,l>=0?Yle(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,6,a);case 11:return!this.q&&(this.q=new Be(Th,this,11,10)),$u(this.q,t,a);case 21:return!this.s&&(this.s=new Be(El,this,21,17)),$u(this.s,t,a)}return g=f(Vt((u=f(fn(this,16),29),u||(Ut(),Mw)),i),69),g.tk().wk(this,Xu(this),i-jn((Ut(),Mw)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 6:return hf(this,null,6,a);case 7:return!this.A&&(this.A=new xl(nl,this,7)),za(this.A,t,a);case 11:return!this.q&&(this.q=new Be(Th,this,11,10)),za(this.q,t,a);case 21:return!this.s&&(this.s=new Be(El,this,21,17)),za(this.s,t,a);case 22:return za(iu(this),t,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),Mw)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),Mw)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return this.D!=null&&this.D==this.F;case 3:return!!US(this);case 4:return!1;case 5:return this.F!=null&&this.F!=this.D&&this.F!=this.B;case 6:return!!IP(this);case 7:return!!this.A&&this.A.i!=0;case 8:return(this.Bb&256)!=0;case 9:return(this.Bb&512)!=0;case 10:return!!this.u&&iu(this.u.a).i!=0&&!(this.n&&$le(this.n));case 11:return!!this.q&&this.q.i!=0;case 12:return rA(this).i!=0;case 13:return Jj(this).i!=0;case 14:return Jj(this),this.r.i!=0;case 15:return rA(this),this.k.i!=0;case 16:return PAe(this).i!=0;case 17:return Hde(this).i!=0;case 18:return Em(this).i!=0;case 19:return ZZ(this).i!=0;case 20:return rA(this),!!this.o;case 21:return!!this.s&&this.s.i!=0;case 22:return!!this.n&&$le(this.n);case 23:return Ide(this).i!=0}return Pf(this,t-jn((Ut(),Mw)),Vt((i=f(fn(this,16),29),i||Mw),t))},p.Vh=function(t){var i;return i=this.i==null||this.q&&this.q.i!=0?null:kG(this,t),i||yOe(this,t)},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:aJ(this,dr(i));return;case 2:Sse(this,dr(i));return;case 5:_L(this,dr(i));return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A),!this.A&&(this.A=new xl(nl,this,7)),Qi(this.A,f(i,16));return;case 8:jCe(this,lt(ut(i)));return;case 9:$Ce(this,lt(ut(i)));return;case 10:e$(Bs(this)),Qi(Bs(this),f(i,16));return;case 11:!this.q&&(this.q=new Be(Th,this,11,10)),Jn(this.q),!this.q&&(this.q=new Be(Th,this,11,10)),Qi(this.q,f(i,16));return;case 21:!this.s&&(this.s=new Be(El,this,21,17)),Jn(this.s),!this.s&&(this.s=new Be(El,this,21,17)),Qi(this.s,f(i,16));return;case 22:Jn(iu(this)),Qi(iu(this),f(i,16));return}$f(this,t-jn((Ut(),Mw)),Vt((a=f(fn(this,16),29),a||Mw),t),i)},p.ei=function(){return Ut(),Mw},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:me(this.Cb,184)&&(f(this.Cb,184).tb=null),ju(this,null);return;case 2:YP(this,null),jP(this,this.D);return;case 5:_L(this,null);return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A);return;case 8:jCe(this,!1);return;case 9:$Ce(this,!1);return;case 10:this.u&&e$(this.u);return;case 11:!this.q&&(this.q=new Be(Th,this,11,10)),Jn(this.q);return;case 21:!this.s&&(this.s=new Be(El,this,21,17)),Jn(this.s);return;case 22:this.n&&Jn(this.n);return}jf(this,t-jn((Ut(),Mw)),Vt((i=f(fn(this,16),29),i||Mw),t))},p.li=function(){var t,i;if(rA(this),Jj(this),PAe(this),Hde(this),Em(this),ZZ(this),Ide(this),sj(svn(pd(this))),this.s)for(t=0,i=this.s.i;t<i;++t)Jz(be(this.s,t));if(this.q)for(t=0,i=this.q.i;t<i;++t)Jz(be(this.q,t));I0((Rl(),Ea),this).ve(),this.Bb|=1},p.Ib=function(){return Q3e(this)},p.k=null,p.r=null;var GF,gMt,b1e;R(un,"EClassImpl",88),N(2038,2037,B3t),p.Bi=function(t,i){return Bde(this,t,i)},p.Ci=function(t){return Bde(this,this.i,t)},p.Di=function(t,i){YAe(this,t,i)},p.Ei=function(t){Ede(this,t)},p.Tk=function(t,i){return $u(this,t,i)},p.Xi=function(t){return fxe(this,t)},p.Uk=function(t,i){return za(this,t,i)},p.Ui=function(t,i){return b_t(this,t,i)},p.Fi=function(){return new GN(this)},p.Gi=function(){return new Qz(this)},p.Hi=function(t){return UH(this,t)},R(Oi,"NotifyingInternalEListImpl",2038),N(617,2038,ec),p.Gc=function(t){return M_t(this,t)},p.Fj=function(t,i,a,u,l){return tj(this,t,i,a,u,l)},p.Gj=function(t){Q5(this,t)},p.Ck=function(t){return this},p.Ik=function(){return Vt(this.e.zh(),this.Ij())},p.Hj=function(){return this.Ik()},p.Ij=function(){return Pi(this.e.zh(),this.Ik())},p.fl=function(){return f(this.Ik().Ek(),29).hk()},p.gl=function(){return os(f(this.Ik(),19)).n},p.gj=function(){return this.e},p.hl=function(){return!0},p.il=function(){return!1},p.jl=function(){return!1},p.kl=function(){return!1},p.bd=function(t){return P0(this,t)},p.Kj=function(t,i){var a;return a=f(t,52),this.jl()?this.hl()?a.Nh(this.e,this.gl(),this.fl(),i):a.Nh(this.e,Pi(a.zh(),os(f(this.Ik(),19))),null,i):a.Nh(this.e,-1-this.Ij(),null,i)},p.Lj=function(t,i){var a;return a=f(t,52),this.jl()?this.hl()?a.Ph(this.e,this.gl(),this.fl(),i):a.Ph(this.e,Pi(a.zh(),os(f(this.Ik(),19))),null,i):a.Ph(this.e,-1-this.Ij(),null,i)},p.Zk=function(){return!1},p.ll=function(){return!0},p.ck=function(t){return Zht(this.d,t)},p.Mj=function(){return Ld(this.e)},p.Nj=function(){return this.i!=0},p.Zi=function(t){return BH(this.d,t)},p.Ti=function(t,i){return this.ll()&&this.kl()?kR(this,t,f(i,57)):i},p.ml=function(t){return t.Rh()?iw(this.e,f(t,52)):t},p.Wb=function(t){qst(this,t)},p.Nc=function(){return Ugt(this)},p.Oc=function(t){var i;if(this.kl())for(i=this.i-1;i>=0;--i)be(this,i);return c3e(this,t)},p.Dk=function(){Jn(this)},p.Wi=function(t,i){return ibt(this,t,i)},R(Oi,"EcoreEList",617),N(489,617,ec,lH),p.Ii=function(){return!1},p.Ij=function(){return this.c},p.Jj=function(){return!1},p.ll=function(){return!0},p.Pi=function(){return!0},p.Ti=function(t,i){return i},p.Vi=function(){return!1},p.c=0,R(Oi,"EObjectEList",489),N(81,489,ec,So),p.Jj=function(){return!0},p.jl=function(){return!1},p.Zk=function(){return!0},R(Oi,"EObjectContainmentEList",81),N(540,81,ec,aX),p.Ki=function(){this.b=!0},p.Nj=function(){return this.b},p.Dk=function(){var t;Jn(this),Ld(this.e)?(t=this.b,this.b=!1,Zr(this.e,new Vh(this.e,2,this.c,t,!1))):this.b=!1},p.b=!1,R(Oi,"EObjectContainmentEList/Unsettable",540),N(1125,540,ec,Llt),p.Qi=function(t,i){var a,u;return a=f(Cj(this,t,i),87),Ld(this.e)&&Q5(this,new NH(this.a,7,(Ut(),aMt),rt(i),(u=a.c,me(u,88)?f(u,29):Ch),t)),a},p.Rj=function(t,i){return KCn(this,f(t,87),i)},p.Sj=function(t,i){return YCn(this,f(t,87),i)},p.Tj=function(t,i,a){return WIn(this,f(t,87),f(i,87),a)},p.Fj=function(t,i,a,u,l){switch(t){case 3:return tj(this,t,i,a,u,this.i>1);case 5:return tj(this,t,i,a,u,this.i-f(a,15).gc()>0);default:return new v1(this.e,t,this.c,i,a,u,!0)}},p.Qj=function(){return!0},p.Nj=function(){return $le(this)},p.Dk=function(){Jn(this)},R(un,"EClassImpl/1",1125),N(1139,1138,Y4e),p.aj=function(t){var i,a,u,l,g,m,v;if(a=t.dj(),a!=8){if(u=N3n(t),u==0)switch(a){case 1:case 9:{v=t.hj(),v!=null&&(i=pd(f(v,468)),!i.c&&(i.c=new ab),IJ(i.c,t.gj())),m=t.fj(),m!=null&&(l=f(m,468),(l.Bb&1)==0&&(i=pd(l),!i.c&&(i.c=new ab),Qn(i.c,f(t.gj(),29))));break}case 3:{m=t.fj(),m!=null&&(l=f(m,468),(l.Bb&1)==0&&(i=pd(l),!i.c&&(i.c=new ab),Qn(i.c,f(t.gj(),29))));break}case 5:{if(m=t.fj(),m!=null)for(g=f(m,16).Jc();g.Ob();)l=f(g.Pb(),468),(l.Bb&1)==0&&(i=pd(l),!i.c&&(i.c=new ab),Qn(i.c,f(t.gj(),29)));break}case 4:{v=t.hj(),v!=null&&(l=f(v,468),(l.Bb&1)==0&&(i=pd(l),!i.c&&(i.c=new ab),IJ(i.c,t.gj())));break}case 6:{if(v=t.hj(),v!=null)for(g=f(v,16).Jc();g.Ob();)l=f(g.Pb(),468),(l.Bb&1)==0&&(i=pd(l),!i.c&&(i.c=new ab),IJ(i.c,t.gj()));break}}this.nl(u)}},p.nl=function(t){hSt(this,t)},p.b=63,R(un,"ESuperAdapter",1139),N(1140,1139,Y4e,Lnt),p.nl=function(t){OT(this,t)},R(un,"EClassImpl/10",1140),N(1129,694,ec),p.Bi=function(t,i){return sde(this,t,i)},p.Ci=function(t){return Yvt(this,t)},p.Di=function(t,i){eG(this,t,i)},p.Ei=function(t){TH(this,t)},p.Xi=function(t){return fxe(this,t)},p.Ui=function(t,i){return Sue(this,t,i)},p.Tk=function(t,i){throw ie(new cr)},p.Fi=function(){return new GN(this)},p.Gi=function(){return new Qz(this)},p.Hi=function(t){return UH(this,t)},p.Uk=function(t,i){throw ie(new cr)},p.Ck=function(t){return this},p.Nj=function(){return this.i!=0},p.Wb=function(t){throw ie(new cr)},p.Dk=function(){throw ie(new cr)},R(Oi,"EcoreEList/UnmodifiableEList",1129),N(328,1129,ec,_3),p.Vi=function(){return!1},R(Oi,"EcoreEList/UnmodifiableEList/FastCompare",328),N(1132,328,ec,L1t),p.bd=function(t){var i,a,u;if(me(t,178)&&(i=f(t,178),a=i.Ij(),a!=-1)){for(u=this.i;a<u;++a)if(Ae(this.g[a])===Ae(t))return a}return-1},R(un,"EClassImpl/1EAllStructuralFeaturesList",1132),N(1126,490,Im,lg),p.Zi=function(t){return Pe(gs,U3t,87,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/1EGenericSuperTypeEList",1126),N(618,490,Im,n1),p.Zi=function(t){return Pe(El,pA,178,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/1EStructuralFeatureUniqueEList",618),N(738,490,Im,d5),p.Zi=function(t){return Pe(yx,pA,19,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/1ReferenceList",738),N(1127,490,Im,Dnt),p.Ji=function(t,i){Xvn(this,f(i,35))},p.Zi=function(t){return Pe(wx,pA,35,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/2",1127),N(1128,490,Im,i9),p.Zi=function(t){return Pe(wx,pA,35,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/3",1128),N(1130,328,ec,Xut),p.Ec=function(t){return hSn(this,f(t,35))},p.Ei=function(t){Zpn(this,f(t,35))},R(un,"EClassImpl/4",1130),N(1131,328,ec,Jut),p.Ec=function(t){return pSn(this,f(t,19))},p.Ei=function(t){Qpn(this,f(t,19))},R(un,"EClassImpl/5",1131),N(1133,490,Im,zU),p.Zi=function(t){return Pe(Th,Z4e,62,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/6",1133),N(1134,490,Im,f5),p.Zi=function(t){return Pe(yx,pA,19,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/7",1134),N(2041,2040,{3:1,4:1,20:1,31:1,55:1,16:1,15:1,70:1,61:1,71:1}),p.Bi=function(t,i){return YIe(this,t,i)},p.Ci=function(t){return YIe(this,this.Bj(),t)},p.Di=function(t,i){VEt(this,t,i)},p.Ei=function(t){DEt(this,t)},p.Tk=function(t,i){return EAn(this,t,i)},p.Uk=function(t,i){return X3n(this,t,i)},p.Ui=function(t,i){return e_t(this,t,i)},p.Xi=function(t){return this.uj(t)},p.Fi=function(){return new GN(this)},p.mj=function(){return this.pj()},p.Gi=function(){return new Qz(this)},p.Hi=function(t){return UH(this,t)},R(Oi,"DelegatingNotifyingInternalEListImpl",2041),N(739,2041,Q4e),p.Ii=function(){var t;return t=Vt(fl(this.b),this.Ij()).Ek(),me(t,155)&&!me(t,455)&&(t.hk().i&1)==0},p.Gc=function(t){var i,a,u,l,g,m,v,S;if(this.ll()){if(S=this.Bj(),S>4)if(this.ck(t)){if(this.Zk()){if(u=f(t,52),a=u.Ah(),v=a==this.b&&(this.jl()?u.uh(u.Bh(),f(Vt(fl(this.b),this.Ij()).Ek(),29).hk())==os(f(Vt(fl(this.b),this.Ij()),19)).n:-1-u.Bh()==this.Ij()),this.kl()&&!v&&!a&&u.Fh()){for(l=0;l<S;++l)if(i=fce(this,this.uj(l)),Ae(i)===Ae(t))return!0}return v}else if(this.jl()&&!this.il()){if(g=f(t,57).Ih(os(f(Vt(fl(this.b),this.Ij()),19))),Ae(g)===Ae(this.b))return!0;if(g==null||!f(g,57).Rh())return!1}}else return!1;if(m=this.rj(t),this.kl()&&!m){for(l=0;l<S;++l)if(u=fce(this,this.uj(l)),Ae(u)===Ae(t))return!0}return m}else return this.rj(t)},p.Fj=function(t,i,a,u,l){return new v1(this.b,t,this.Ij(),i,a,u,l)},p.Gj=function(t){Zr(this.b,t)},p.Ck=function(t){return this},p.Hj=function(){return Vt(fl(this.b),this.Ij())},p.Ij=function(){return Pi(fl(this.b),Vt(fl(this.b),this.Ij()))},p.gj=function(){return this.b},p.hl=function(){return!!Vt(fl(this.b),this.Ij()).Ek().hk()},p.Jj=function(){var t,i;return i=Vt(fl(this.b),this.Ij()),me(i,101)?(t=f(i,19),(t.Bb&Dc)!=0||!!os(f(i,19))):!1},p.il=function(){var t,i,a,u;return i=Vt(fl(this.b),this.Ij()),me(i,101)?(t=f(i,19),a=os(t),!!a&&(u=a.t,u>1||u==-1)):!1},p.jl=function(){var t,i,a;return i=Vt(fl(this.b),this.Ij()),me(i,101)?(t=f(i,19),a=os(t),!!a):!1},p.kl=function(){var t,i;return i=Vt(fl(this.b),this.Ij()),me(i,101)?(t=f(i,19),(t.Bb&Xa)!=0):!1},p.bd=function(t){var i,a,u,l;if(u=this.wj(t),u>=0)return u;if(this.ll()){for(a=0,l=this.Bj();a<l;++a)if(i=fce(this,this.uj(a)),Ae(i)===Ae(t))return a}return-1},p.Kj=function(t,i){var a;return a=f(t,52),this.jl()?this.hl()?a.Nh(this.b,os(f(Vt(fl(this.b),this.Ij()),19)).n,f(Vt(fl(this.b),this.Ij()).Ek(),29).hk(),i):a.Nh(this.b,Pi(a.zh(),os(f(Vt(fl(this.b),this.Ij()),19))),null,i):a.Nh(this.b,-1-this.Ij(),null,i)},p.Lj=function(t,i){var a;return a=f(t,52),this.jl()?this.hl()?a.Ph(this.b,os(f(Vt(fl(this.b),this.Ij()),19)).n,f(Vt(fl(this.b),this.Ij()).Ek(),29).hk(),i):a.Ph(this.b,Pi(a.zh(),os(f(Vt(fl(this.b),this.Ij()),19))),null,i):a.Ph(this.b,-1-this.Ij(),null,i)},p.Zk=function(){var t,i;return i=Vt(fl(this.b),this.Ij()),me(i,101)?(t=f(i,19),(t.Bb&Dc)!=0):!1},p.ll=function(){return me(Vt(fl(this.b),this.Ij()).Ek(),88)},p.ck=function(t){return Vt(fl(this.b),this.Ij()).Ek().ck(t)},p.Mj=function(){return Ld(this.b)},p.Nj=function(){return!this.xj()},p.Pi=function(){return Vt(fl(this.b),this.Ij()).Pi()},p.Ti=function(t,i){return RG(this,t,i)},p.Wb=function(t){e$(this),Qi(this,f(t,15))},p.Nc=function(){var t;if(this.kl())for(t=this.Bj()-1;t>=0;--t)RG(this,t,this.uj(t));return this.Cj()},p.Oc=function(t){var i;if(this.kl())for(i=this.Bj()-1;i>=0;--i)RG(this,i,this.uj(i));return this.Dj(t)},p.Dk=function(){e$(this)},p.Wi=function(t,i){return Ogt(this,t,i)},R(Oi,"DelegatingEcoreEList",739),N(1135,739,Q4e,Yct),p.nj=function(t,i){hbn(this,t,f(i,29))},p.oj=function(t){Jgn(this,f(t,29))},p.uj=function(t){var i,a;return i=f(be(iu(this.a),t),87),a=i.c,me(a,88)?f(a,29):(Ut(),Ch)},p.zj=function(t){var i,a;return i=f(PT(iu(this.a),t),87),a=i.c,me(a,88)?f(a,29):(Ut(),Ch)},p.Aj=function(t,i){return SAn(this,t,f(i,29))},p.Ii=function(){return!1},p.Fj=function(t,i,a,u,l){return null},p.pj=function(){return new Mnt(this)},p.qj=function(){Jn(iu(this.a))},p.rj=function(t){return k0t(this,t)},p.sj=function(t){var i,a;for(a=t.Jc();a.Ob();)if(i=a.Pb(),!k0t(this,i))return!1;return!0},p.tj=function(t){var i,a,u;if(me(t,15)&&(u=f(t,15),u.gc()==iu(this.a).i)){for(i=u.Jc(),a=new Nn(this);i.Ob();)if(Ae(i.Pb())!==Ae(Mn(a)))return!1;return!0}return!1},p.vj=function(){var t,i,a,u,l;for(a=1,i=new Nn(iu(this.a));i.e!=i.i.gc();)t=f(Mn(i),87),u=(l=t.c,me(l,88)?f(l,29):(Ut(),Ch)),a=31*a+(u?bS(u):0);return a},p.wj=function(t){var i,a,u,l;for(u=0,a=new Nn(iu(this.a));a.e!=a.i.gc();){if(i=f(Mn(a),87),Ae(t)===Ae((l=i.c,me(l,88)?f(l,29):(Ut(),Ch))))return u;++u}return-1},p.xj=function(){return iu(this.a).i==0},p.yj=function(){return null},p.Bj=function(){return iu(this.a).i},p.Cj=function(){var t,i,a,u,l,g;for(g=iu(this.a).i,l=Pe(Ro,Jt,1,g,5,1),a=0,i=new Nn(iu(this.a));i.e!=i.i.gc();)t=f(Mn(i),87),l[a++]=(u=t.c,me(u,88)?f(u,29):(Ut(),Ch));return l},p.Dj=function(t){var i,a,u,l,g,m,v;for(v=iu(this.a).i,t.length<v&&(l=BH(Fd(t).c,v),t=l),t.length>v&&io(t,v,null),u=0,a=new Nn(iu(this.a));a.e!=a.i.gc();)i=f(Mn(a),87),g=(m=i.c,me(m,88)?f(m,29):(Ut(),Ch)),io(t,u++,g);return t},p.Ej=function(){var t,i,a,u,l;for(l=new p0,l.a+="[",t=iu(this.a),i=0,u=iu(this.a).i;i<u;)ys(l,$7((a=f(be(t,i),87).c,me(a,88)?f(a,29):(Ut(),Ch)))),++i<u&&(l.a+=Fu);return l.a+="]",l.a},p.Gj=function(t){},p.Ij=function(){return 10},p.hl=function(){return!0},p.Jj=function(){return!1},p.il=function(){return!1},p.jl=function(){return!1},p.kl=function(){return!0},p.Zk=function(){return!1},p.ll=function(){return!0},p.ck=function(t){return me(t,88)},p.Nj=function(){return g2n(this.a)},p.Pi=function(){return!0},p.Vi=function(){return!0},R(un,"EClassImpl/8",1135),N(1136,2008,IL,Mnt),p.dd=function(t){return UH(this.a,t)},p.gc=function(){return iu(this.a.a).i},R(un,"EClassImpl/8/1",1136),N(1137,490,Im,tm),p.Zi=function(t){return Pe(Up,Jt,141,t,0,1)},p.Vi=function(){return!1},R(un,"EClassImpl/9",1137),N(1124,50,$Oe,Art),R(un,"EClassImpl/MyHashSet",1124),N(559,357,{109:1,93:1,92:1,141:1,155:1,826:1,154:1,197:1,57:1,113:1,52:1,98:1,357:1,157:1,117:1,118:1,676:1},sY),p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return this.D!=null?this.D:this.B;case 3:return US(this);case 4:return this.fk();case 5:return this.F;case 6:return i?cf(this):IP(this);case 7:return!this.A&&(this.A=new xl(nl,this,7)),this.A;case 8:return mn(),(this.Bb&256)!=0}return Lf(this,t-jn(this.ei()),Vt((u=f(fn(this,16),29),u||this.ei()),t),i,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return this.D!=null&&this.D==this.F;case 3:return!!US(this);case 4:return this.fk()!=null;case 5:return this.F!=null&&this.F!=this.D&&this.F!=this.B;case 6:return!!IP(this);case 7:return!!this.A&&this.A.i!=0;case 8:return(this.Bb&256)==0}return Pf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:aJ(this,dr(i));return;case 2:Sse(this,dr(i));return;case 5:_L(this,dr(i));return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A),!this.A&&(this.A=new xl(nl,this,7)),Qi(this.A,f(i,16));return;case 8:uZ(this,lt(ut(i)));return}$f(this,t-jn(this.ei()),Vt((a=f(fn(this,16),29),a||this.ei()),t),i)},p.ei=function(){return Ut(),sMt},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:me(this.Cb,184)&&(f(this.Cb,184).tb=null),ju(this,null);return;case 2:YP(this,null),jP(this,this.D);return;case 5:_L(this,null);return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A);return;case 8:uZ(this,!0);return}jf(this,t-jn(this.ei()),Vt((i=f(fn(this,16),29),i||this.ei()),t))},p.li=function(){I0((Rl(),Ea),this).ve(),this.Bb|=1},p.lk=function(){var t,i,a;if(!this.c&&(t=a2t(cf(this)),!t.dc()))for(a=t.Jc();a.Ob();)i=dr(a.Pb()),Kj(this,i)&&tCn(this);return this.b},p.fk=function(){var t;if(!this.e){t=null;try{t=US(this)}catch(i){if(i=go(i),!me(i,103))throw ie(i)}this.d=null,t&&(t.i&1)!=0&&(t==kd?this.d=(mn(),yw):t==pr?this.d=rt(0):t==kx?this.d=new f3(0):t==sa?this.d=0:t==x_?this.d=TT(0):t==JA?this.d=ZP(0):t==$l?this.d=jH(0):this.d=qH(0)),this.e=!0}return this.d},p.kk=function(){return(this.Bb&256)!=0},p.ol=function(t){t&&(this.D="org.eclipse.emf.common.util.AbstractEnumerator")},p.dl=function(t){Nbt(this,t),this.ol(t)},p.el=function(t){this.C=t,this.e=!1},p.Ib=function(){var t;return(this.Db&64)!=0?mZ(this):(t=new sh(mZ(this)),t.a+=" (serializable: ",m0(t,(this.Bb&256)!=0),t.a+=")",t.a)},p.c=!1,p.d=null,p.e=!1,R(un,"EDataTypeImpl",559),N(455,559,{109:1,93:1,92:1,141:1,155:1,826:1,670:1,154:1,197:1,57:1,113:1,52:1,98:1,357:1,455:1,157:1,117:1,118:1,676:1},Irt),p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return this.D!=null?this.D:this.B;case 3:return US(this);case 4:return kCe(this);case 5:return this.F;case 6:return i?cf(this):IP(this);case 7:return!this.A&&(this.A=new xl(nl,this,7)),this.A;case 8:return mn(),(this.Bb&256)!=0;case 9:return!this.a&&(this.a=new Be(j1,this,9,5)),this.a}return Lf(this,t-jn((Ut(),jw)),Vt((u=f(fn(this,16),29),u||jw),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 6:return this.Cb&&(a=(l=this.Db>>16,l>=0?Yle(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,6,a);case 9:return!this.a&&(this.a=new Be(j1,this,9,5)),$u(this.a,t,a)}return g=f(Vt((u=f(fn(this,16),29),u||(Ut(),jw)),i),69),g.tk().wk(this,Xu(this),i-jn((Ut(),jw)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 6:return hf(this,null,6,a);case 7:return!this.A&&(this.A=new xl(nl,this,7)),za(this.A,t,a);case 9:return!this.a&&(this.a=new Be(j1,this,9,5)),za(this.a,t,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),jw)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),jw)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return this.D!=null&&this.D==this.F;case 3:return!!US(this);case 4:return!!kCe(this);case 5:return this.F!=null&&this.F!=this.D&&this.F!=this.B;case 6:return!!IP(this);case 7:return!!this.A&&this.A.i!=0;case 8:return(this.Bb&256)==0;case 9:return!!this.a&&this.a.i!=0}return Pf(this,t-jn((Ut(),jw)),Vt((i=f(fn(this,16),29),i||jw),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:aJ(this,dr(i));return;case 2:Sse(this,dr(i));return;case 5:_L(this,dr(i));return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A),!this.A&&(this.A=new xl(nl,this,7)),Qi(this.A,f(i,16));return;case 8:uZ(this,lt(ut(i)));return;case 9:!this.a&&(this.a=new Be(j1,this,9,5)),Jn(this.a),!this.a&&(this.a=new Be(j1,this,9,5)),Qi(this.a,f(i,16));return}$f(this,t-jn((Ut(),jw)),Vt((a=f(fn(this,16),29),a||jw),t),i)},p.ei=function(){return Ut(),jw},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:me(this.Cb,184)&&(f(this.Cb,184).tb=null),ju(this,null);return;case 2:YP(this,null),jP(this,this.D);return;case 5:_L(this,null);return;case 7:!this.A&&(this.A=new xl(nl,this,7)),Jn(this.A);return;case 8:uZ(this,!0);return;case 9:!this.a&&(this.a=new Be(j1,this,9,5)),Jn(this.a);return}jf(this,t-jn((Ut(),jw)),Vt((i=f(fn(this,16),29),i||jw),t))},p.li=function(){var t,i;if(this.a)for(t=0,i=this.a.i;t<i;++t)Jz(be(this.a,t));I0((Rl(),Ea),this).ve(),this.Bb|=1},p.fk=function(){return kCe(this)},p.ck=function(t){return t!=null},p.ol=function(t){},R(un,"EEnumImpl",455),N(564,436,{109:1,93:1,92:1,1985:1,679:1,154:1,197:1,57:1,113:1,52:1,98:1,564:1,157:1,117:1,118:1},Qnt),p.ve=function(){return this.zb},p.wh=function(t){return Swt(this,t)},p.Hh=function(t,i,a){var u,l;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return rt(this.d);case 3:return this.b?this.b:this.a;case 4:return l=this.c,l??this.zb;case 5:return this.Db>>16==5?f(this.Cb,670):null}return Lf(this,t-jn((Ut(),X0)),Vt((u=f(fn(this,16),29),u||X0),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 5:return this.Cb&&(a=(l=this.Db>>16,l>=0?Swt(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,5,a)}return g=f(Vt((u=f(fn(this,16),29),u||(Ut(),X0)),i),69),g.tk().wk(this,Xu(this),i-jn((Ut(),X0)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 5:return hf(this,null,5,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),X0)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),X0)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return this.d!=0;case 3:return!!this.b;case 4:return this.c!=null;case 5:return!!(this.Db>>16==5&&f(this.Cb,670))}return Pf(this,t-jn((Ut(),X0)),Vt((i=f(fn(this,16),29),i||X0),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:ju(this,dr(i));return;case 2:Iue(this,f(i,17).a);return;case 3:kEt(this,f(i,1985));return;case 4:Nue(this,dr(i));return}$f(this,t-jn((Ut(),X0)),Vt((a=f(fn(this,16),29),a||X0),t),i)},p.ei=function(){return Ut(),X0},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:ju(this,null);return;case 2:Iue(this,0);return;case 3:kEt(this,null);return;case 4:Nue(this,null);return}jf(this,t-jn((Ut(),X0)),Vt((i=f(fn(this,16),29),i||X0),t))},p.Ib=function(){var t;return t=this.c,t??this.zb},p.b=null,p.c=null,p.d=0,R(un,"EEnumLiteralImpl",564);var vjn=Bi(un,"EFactoryImpl/InternalEDateTimeFormat");N(483,1,{2060:1},yz),R(un,"EFactoryImpl/1ClientInternalEDateTimeFormat",483),N(246,118,{109:1,93:1,92:1,87:1,57:1,113:1,52:1,98:1,246:1,117:1,118:1},cS),p.yh=function(t,i,a){var u;return a=hf(this,t,i,a),this.e&&me(t,178)&&(u=JZ(this,this.e),u!=this.c&&(a=kL(this,u,a))),a},p.Hh=function(t,i,a){var u;switch(t){case 0:return this.f;case 1:return!this.d&&(this.d=new So(gs,this,1)),this.d;case 2:return i?uQ(this):this.c;case 3:return this.b;case 4:return this.e;case 5:return i?Ule(this):this.a}return Lf(this,t-jn((Ut(),__)),Vt((u=f(fn(this,16),29),u||__),t),i,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return l0t(this,null,a);case 1:return!this.d&&(this.d=new So(gs,this,1)),za(this.d,t,a);case 3:return u0t(this,null,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),__)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),__)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.f;case 1:return!!this.d&&this.d.i!=0;case 2:return!!this.c;case 3:return!!this.b;case 4:return!!this.e;case 5:return!!this.a}return Pf(this,t-jn((Ut(),__)),Vt((i=f(fn(this,16),29),i||__),t))},p.Zh=function(t,i){var a;switch(t){case 0:Uwt(this,f(i,87));return;case 1:!this.d&&(this.d=new So(gs,this,1)),Jn(this.d),!this.d&&(this.d=new So(gs,this,1)),Qi(this.d,f(i,16));return;case 3:eAe(this,f(i,87));return;case 4:wAe(this,f(i,829));return;case 5:$P(this,f(i,141));return}$f(this,t-jn((Ut(),__)),Vt((a=f(fn(this,16),29),a||__),t),i)},p.ei=function(){return Ut(),__},p.gi=function(t){var i;switch(t){case 0:Uwt(this,null);return;case 1:!this.d&&(this.d=new So(gs,this,1)),Jn(this.d);return;case 3:eAe(this,null);return;case 4:wAe(this,null);return;case 5:$P(this,null);return}jf(this,t-jn((Ut(),__)),Vt((i=f(fn(this,16),29),i||__),t))},p.Ib=function(){var t;return t=new ef(Jh(this)),t.a+=" (expression: ",Kde(this,t),t.a+=")",t.a};var F7e;R(un,"EGenericTypeImpl",246),N(2013,2008,pee),p.Di=function(t,i){Vct(this,t,i)},p.Tk=function(t,i){return Vct(this,this.gc(),t),i},p.Xi=function(t){return Xh(this.mj(),t)},p.Fi=function(){return this.Gi()},p.mj=function(){return new Unt(this)},p.Gi=function(){return this.Hi(0)},p.Hi=function(t){return this.mj().dd(t)},p.Uk=function(t,i){return _T(this,t,!0),i},p.Qi=function(t,i){var a,u;return u=Zle(this,i),a=this.dd(t),a.Rb(u),u},p.Ri=function(t,i){var a;_T(this,i,!0),a=this.dd(t),a.Rb(i)},R(Oi,"AbstractSequentialInternalEList",2013),N(480,2013,pee,Zz),p.Xi=function(t){return Xh(this.mj(),t)},p.Fi=function(){return this.b==null?(v0(),v0(),TV):this.pl()},p.mj=function(){return new cst(this.a,this.b)},p.Gi=function(){return this.b==null?(v0(),v0(),TV):this.pl()},p.Hi=function(t){var i,a;if(this.b==null){if(t<0||t>1)throw ie(new Ru(T$+t+", size=0"));return v0(),v0(),TV}for(a=this.pl(),i=0;i<t;++i)ZJ(a);return a},p.dc=function(){var t,i,a,u,l,g;if(this.b!=null){for(a=0;a<this.b.length;++a)if(t=this.b[a],!this.sl()||this.a.Th(t)){if(g=this.a.Jh(t,!1),is(),f(t,69).uk()){for(i=f(g,159),u=0,l=i.gc();u<l;++u)if(mdt(i.Ql(u))&&i.Rl(u)!=null)return!1}else if(t.Gk()){if(!f(g,16).dc())return!1}else if(g!=null)return!1}}return!0},p.Jc=function(){return Gxe(this)},p.dd=function(t){var i,a;if(this.b==null){if(t!=0)throw ie(new Ru(T$+t+", size=0"));return v0(),v0(),TV}for(a=this.rl()?this.ql():this.pl(),i=0;i<t;++i)ZJ(a);return a},p.Qi=function(t,i){throw ie(new cr)},p.Ri=function(t,i){throw ie(new cr)},p.pl=function(){return new lX(this.a,this.b)},p.ql=function(){return new p_e(this.a,this.b)},p.rl=function(){return!0},p.gc=function(){var t,i,a,u,l,g,m;if(l=0,this.b!=null){for(a=0;a<this.b.length;++a)if(t=this.b[a],!this.sl()||this.a.Th(t))if(m=this.a.Jh(t,!1),is(),f(t,69).uk())for(i=f(m,159),u=0,g=i.gc();u<g;++u)mdt(i.Ql(u))&&i.Rl(u)!=null&&++l;else t.Gk()?l+=f(m,16).gc():m!=null&&++l}return l},p.sl=function(){return!0};var v1e;R(Oi,"EContentsEList",480),N(1141,480,pee,Lct),p.pl=function(){return new Pct(this.a,this.b)},p.ql=function(){return new Rct(this.a,this.b)},p.sl=function(){return!1},R(un,"ENamedElementImpl/1",1141),N(284,1,gee,lX),p.Nb=function(t){va(this,t)},p.Rb=function(t){throw ie(new cr)},p.tl=function(t){if(this.g!=0||this.e)throw ie(new As("Iterator already in use or already filtered"));this.e=t},p.Ob=function(){var t,i,a,u,l,g;switch(this.g){case 3:case 2:return!0;case 1:return!1;case-3:this.p?this.p.Pb():++this.n;default:if(!this.k||(this.p?!jyt(this,this.p):!AEt(this))){for(;this.d<this.c.length;)if(i=this.c[this.d++],(!this.e||i.mk()!=$D||i.Ij()!=0)&&(!this.sl()||this.b.Th(i))){if(g=this.b.Jh(i,this.rl()),this.f=(is(),f(i,69).uk()),this.f||i.Gk()){if(this.rl()?(u=f(g,15),this.k=u):(u=f(g,71),this.k=this.j=u),me(this.k,59)?(this.p=null,this.o=this.k.gc(),this.n=0):this.p=this.j?this.j.Gi():this.k.cd(),this.p?jyt(this,this.p):AEt(this))return l=this.p?this.p.Pb():this.j?this.j.Xi(this.n++):this.k.Xb(this.n++),this.f?(t=f(l,74),t.Ik(),a=t.kd(),this.i=a):(a=l,this.i=a),this.g=3,!0}else if(g!=null)return this.k=null,this.p=null,a=g,this.i=a,this.g=2,!0}return this.k=null,this.p=null,this.f=!1,this.g=1,!1}else return l=this.p?this.p.Pb():this.j?this.j.Xi(this.n++):this.k.Xb(this.n++),this.f?(t=f(l,74),t.Ik(),a=t.kd(),this.i=a):(a=l,this.i=a),this.g=3,!0}},p.Sb=function(){var t,i,a,u,l,g;switch(this.g){case-3:case-2:return!0;case-1:return!1;case 3:this.p?this.p.Ub():--this.n;default:if(!this.k||(this.p?!$yt(this,this.p):!Yyt(this))){for(;this.d>0;)if(i=this.c[--this.d],(!this.e||i.mk()!=$D||i.Ij()!=0)&&(!this.sl()||this.b.Th(i))){if(g=this.b.Jh(i,this.rl()),this.f=(is(),f(i,69).uk()),this.f||i.Gk()){if(this.rl()?(u=f(g,15),this.k=u):(u=f(g,71),this.k=this.j=u),me(this.k,59)?(this.o=this.k.gc(),this.n=this.o):this.p=this.j?this.j.Hi(this.k.gc()):this.k.dd(this.k.gc()),this.p?$yt(this,this.p):Yyt(this))return l=this.p?this.p.Ub():this.j?this.j.Xi(--this.n):this.k.Xb(--this.n),this.f?(t=f(l,74),t.Ik(),a=t.kd(),this.i=a):(a=l,this.i=a),this.g=-3,!0}else if(g!=null)return this.k=null,this.p=null,a=g,this.i=a,this.g=-2,!0}return this.k=null,this.p=null,this.g=-1,!1}else return l=this.p?this.p.Ub():this.j?this.j.Xi(--this.n):this.k.Xb(--this.n),this.f?(t=f(l,74),t.Ik(),a=t.kd(),this.i=a):(a=l,this.i=a),this.g=-3,!0}},p.Pb=function(){return ZJ(this)},p.Tb=function(){return this.a},p.Ub=function(){var t;if(this.g<-1||this.Sb())return--this.a,this.g=0,t=this.i,this.Sb(),t;throw ie(new ic)},p.Vb=function(){return this.a-1},p.Qb=function(){throw ie(new cr)},p.rl=function(){return!1},p.Wb=function(t){throw ie(new cr)},p.sl=function(){return!0},p.a=0,p.d=0,p.f=!1,p.g=0,p.n=0,p.o=0;var TV;R(Oi,"EContentsEList/FeatureIteratorImpl",284),N(695,284,gee,p_e),p.rl=function(){return!0},R(Oi,"EContentsEList/ResolvingFeatureIteratorImpl",695),N(1142,695,gee,Rct),p.sl=function(){return!1},R(un,"ENamedElementImpl/1/1",1142),N(1143,284,gee,Pct),p.sl=function(){return!1},R(un,"ENamedElementImpl/1/2",1143),N(38,150,dq,pT,tue,Jo,gue,v1,Vh,kxe,dht,Txe,fht,qTe,hht,Axe,pht,VTe,lht,xxe,ght,W7,NH,jce,Cxe,mht,GTe,bht),p.Hj=function(){return sxe(this)},p.Oj=function(){var t;return t=sxe(this),t?t.fk():null},p.ej=function(t){return this.b==-1&&this.a&&(this.b=this.c.Dh(this.a.Ij(),this.a.mk())),this.c.uh(this.b,t)},p.gj=function(){return this.c},p.Pj=function(){var t;return t=sxe(this),t?t.qk():!1},p.b=-1,R(un,"ENotificationImpl",38),N(400,290,{109:1,93:1,92:1,154:1,197:1,57:1,62:1,113:1,467:1,52:1,98:1,157:1,400:1,290:1,117:1,118:1},gae),p.wh=function(t){return kwt(this,t)},p.Hh=function(t,i,a){var u,l,g;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return mn(),(this.Bb&256)!=0;case 3:return mn(),(this.Bb&512)!=0;case 4:return rt(this.s);case 5:return rt(this.t);case 6:return mn(),g=this.t,g>1||g==-1;case 7:return mn(),l=this.s,l>=1;case 8:return i?ph(this):this.r;case 9:return this.q;case 10:return this.Db>>16==10?f(this.Cb,29):null;case 11:return!this.d&&(this.d=new xl(nl,this,11)),this.d;case 12:return!this.c&&(this.c=new Be(S_,this,12,10)),this.c;case 13:return!this.a&&(this.a=new nH(this,this)),this.a;case 14:return gd(this)}return Lf(this,t-jn((Ut(),J0)),Vt((u=f(fn(this,16),29),u||J0),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 10:return this.Cb&&(a=(l=this.Db>>16,l>=0?kwt(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,10,a);case 12:return!this.c&&(this.c=new Be(S_,this,12,10)),$u(this.c,t,a)}return g=f(Vt((u=f(fn(this,16),29),u||(Ut(),J0)),i),69),g.tk().wk(this,Xu(this),i-jn((Ut(),J0)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 9:return yce(this,a);case 10:return hf(this,null,10,a);case 11:return!this.d&&(this.d=new xl(nl,this,11)),za(this.d,t,a);case 12:return!this.c&&(this.c=new Be(S_,this,12,10)),za(this.c,t,a);case 14:return za(gd(this),t,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),J0)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),J0)),t,a)},p.Sh=function(t){var i,a,u;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return(this.Bb&256)==0;case 3:return(this.Bb&512)==0;case 4:return this.s!=0;case 5:return this.t!=1;case 6:return u=this.t,u>1||u==-1;case 7:return a=this.s,a>=1;case 8:return!!this.r&&!this.q.e&&ES(this.q).i==0;case 9:return!!this.q&&!(this.r&&!this.q.e&&ES(this.q).i==0);case 10:return!!(this.Db>>16==10&&f(this.Cb,29));case 11:return!!this.d&&this.d.i!=0;case 12:return!!this.c&&this.c.i!=0;case 13:return!!this.a&&gd(this.a.a).i!=0&&!(this.b&&Fle(this.b));case 14:return!!this.b&&Fle(this.b)}return Pf(this,t-jn((Ut(),J0)),Vt((i=f(fn(this,16),29),i||J0),t))},p.Zh=function(t,i){var a,u;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:ju(this,dr(i));return;case 2:O0(this,lt(ut(i)));return;case 3:N0(this,lt(ut(i)));return;case 4:x0(this,f(i,17).a);return;case 5:vT(this,f(i,17).a);return;case 8:fE(this,f(i,141));return;case 9:u=xg(this,f(i,87),null),u&&u.lj();return;case 11:!this.d&&(this.d=new xl(nl,this,11)),Jn(this.d),!this.d&&(this.d=new xl(nl,this,11)),Qi(this.d,f(i,16));return;case 12:!this.c&&(this.c=new Be(S_,this,12,10)),Jn(this.c),!this.c&&(this.c=new Be(S_,this,12,10)),Qi(this.c,f(i,16));return;case 13:!this.a&&(this.a=new nH(this,this)),e$(this.a),!this.a&&(this.a=new nH(this,this)),Qi(this.a,f(i,16));return;case 14:Jn(gd(this)),Qi(gd(this),f(i,16));return}$f(this,t-jn((Ut(),J0)),Vt((a=f(fn(this,16),29),a||J0),t),i)},p.ei=function(){return Ut(),J0},p.gi=function(t){var i,a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:ju(this,null);return;case 2:O0(this,!0);return;case 3:N0(this,!0);return;case 4:x0(this,0);return;case 5:vT(this,1);return;case 8:fE(this,null);return;case 9:a=xg(this,null,null),a&&a.lj();return;case 11:!this.d&&(this.d=new xl(nl,this,11)),Jn(this.d);return;case 12:!this.c&&(this.c=new Be(S_,this,12,10)),Jn(this.c);return;case 13:this.a&&e$(this.a);return;case 14:this.b&&Jn(this.b);return}jf(this,t-jn((Ut(),J0)),Vt((i=f(fn(this,16),29),i||J0),t))},p.li=function(){var t,i;if(this.c)for(t=0,i=this.c.i;t<i;++t)Jz(be(this.c,t));ph(this),this.Bb|=1},R(un,"EOperationImpl",400),N(497,739,Q4e,nH),p.nj=function(t,i){pbn(this,t,f(i,141))},p.oj=function(t){Zgn(this,f(t,141))},p.uj=function(t){var i,a;return i=f(be(gd(this.a),t),87),a=i.c,a||(Ut(),Ug)},p.zj=function(t){var i,a;return i=f(PT(gd(this.a),t),87),a=i.c,a||(Ut(),Ug)},p.Aj=function(t,i){return u3n(this,t,f(i,141))},p.Ii=function(){return!1},p.Fj=function(t,i,a,u,l){return null},p.pj=function(){return new jnt(this)},p.qj=function(){Jn(gd(this.a))},p.rj=function(t){return C0t(this,t)},p.sj=function(t){var i,a;for(a=t.Jc();a.Ob();)if(i=a.Pb(),!C0t(this,i))return!1;return!0},p.tj=function(t){var i,a,u;if(me(t,15)&&(u=f(t,15),u.gc()==gd(this.a).i)){for(i=u.Jc(),a=new Nn(this);i.Ob();)if(Ae(i.Pb())!==Ae(Mn(a)))return!1;return!0}return!1},p.vj=function(){var t,i,a,u,l;for(a=1,i=new Nn(gd(this.a));i.e!=i.i.gc();)t=f(Mn(i),87),u=(l=t.c,l||(Ut(),Ug)),a=31*a+(u?Ti(u):0);return a},p.wj=function(t){var i,a,u,l;for(u=0,a=new Nn(gd(this.a));a.e!=a.i.gc();){if(i=f(Mn(a),87),Ae(t)===Ae((l=i.c,l||(Ut(),Ug))))return u;++u}return-1},p.xj=function(){return gd(this.a).i==0},p.yj=function(){return null},p.Bj=function(){return gd(this.a).i},p.Cj=function(){var t,i,a,u,l,g;for(g=gd(this.a).i,l=Pe(Ro,Jt,1,g,5,1),a=0,i=new Nn(gd(this.a));i.e!=i.i.gc();)t=f(Mn(i),87),l[a++]=(u=t.c,u||(Ut(),Ug));return l},p.Dj=function(t){var i,a,u,l,g,m,v;for(v=gd(this.a).i,t.length<v&&(l=BH(Fd(t).c,v),t=l),t.length>v&&io(t,v,null),u=0,a=new Nn(gd(this.a));a.e!=a.i.gc();)i=f(Mn(a),87),g=(m=i.c,m||(Ut(),Ug)),io(t,u++,g);return t},p.Ej=function(){var t,i,a,u,l;for(l=new p0,l.a+="[",t=gd(this.a),i=0,u=gd(this.a).i;i<u;)ys(l,$7((a=f(be(t,i),87).c,a||(Ut(),Ug)))),++i<u&&(l.a+=Fu);return l.a+="]",l.a},p.Gj=function(t){},p.Ij=function(){return 13},p.hl=function(){return!0},p.Jj=function(){return!1},p.il=function(){return!1},p.jl=function(){return!1},p.kl=function(){return!0},p.Zk=function(){return!1},p.ll=function(){return!0},p.ck=function(t){return me(t,141)},p.Nj=function(){return m2n(this.a)},p.Pi=function(){return!0},p.Vi=function(){return!0},R(un,"EOperationImpl/1",497),N(1315,2008,IL,jnt),p.dd=function(t){return UH(this.a,t)},p.gc=function(){return gd(this.a.a).i},R(un,"EOperationImpl/1/1",1315),N(1316,540,ec,Dlt),p.Qi=function(t,i){var a,u;return a=f(Cj(this,t,i),87),Ld(this.e)&&Q5(this,new NH(this.a,7,(Ut(),lMt),rt(i),(u=a.c,u||Ug),t)),a},p.Rj=function(t,i){return Bxn(this,f(t,87),i)},p.Sj=function(t,i){return Uxn(this,f(t,87),i)},p.Tj=function(t,i,a){return A3n(this,f(t,87),f(i,87),a)},p.Fj=function(t,i,a,u,l){switch(t){case 3:return tj(this,t,i,a,u,this.i>1);case 5:return tj(this,t,i,a,u,this.i-f(a,15).gc()>0);default:return new v1(this.e,t,this.c,i,a,u,!0)}},p.Qj=function(){return!0},p.Nj=function(){return Fle(this)},p.Dk=function(){Jn(this)},R(un,"EOperationImpl/2",1316),N(491,1,{1983:1,491:1},Wat),R(un,"EPackageImpl/1",491),N(14,81,ec,Be),p.fl=function(){return this.d},p.gl=function(){return this.b},p.jl=function(){return!0},p.b=0,R(Oi,"EObjectContainmentWithInverseEList",14),N(358,14,ec,qN),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectContainmentWithInverseEList/Resolving",358),N(309,358,ec,aT),p.Ki=function(){this.a.tb=null},R(un,"EPackageImpl/2",309),N(1238,1,{},r1),R(un,"EPackageImpl/3",1238),N(716,45,sA,y2e),p._b=function(t){return ta(t)?Mce(this,t):!!Is(this.f,t)},R(un,"EPackageRegistryImpl",716),N(502,290,{109:1,93:1,92:1,154:1,197:1,57:1,2062:1,113:1,467:1,52:1,98:1,157:1,502:1,290:1,117:1,118:1},mae),p.wh=function(t){return Twt(this,t)},p.Hh=function(t,i,a){var u,l,g;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return mn(),(this.Bb&256)!=0;case 3:return mn(),(this.Bb&512)!=0;case 4:return rt(this.s);case 5:return rt(this.t);case 6:return mn(),g=this.t,g>1||g==-1;case 7:return mn(),l=this.s,l>=1;case 8:return i?ph(this):this.r;case 9:return this.q;case 10:return this.Db>>16==10?f(this.Cb,62):null}return Lf(this,t-jn((Ut(),Sx)),Vt((u=f(fn(this,16),29),u||Sx),t),i,a)},p.Oh=function(t,i,a){var u,l,g;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),$u(this.Ab,t,a);case 10:return this.Cb&&(a=(l=this.Db>>16,l>=0?Twt(this,a):this.Cb.Ph(this,-1-l,null,a))),hf(this,t,10,a)}return g=f(Vt((u=f(fn(this,16),29),u||(Ut(),Sx)),i),69),g.tk().wk(this,Xu(this),i-jn((Ut(),Sx)),t,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 9:return yce(this,a);case 10:return hf(this,null,10,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),Sx)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),Sx)),t,a)},p.Sh=function(t){var i,a,u;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return(this.Bb&256)==0;case 3:return(this.Bb&512)==0;case 4:return this.s!=0;case 5:return this.t!=1;case 6:return u=this.t,u>1||u==-1;case 7:return a=this.s,a>=1;case 8:return!!this.r&&!this.q.e&&ES(this.q).i==0;case 9:return!!this.q&&!(this.r&&!this.q.e&&ES(this.q).i==0);case 10:return!!(this.Db>>16==10&&f(this.Cb,62))}return Pf(this,t-jn((Ut(),Sx)),Vt((i=f(fn(this,16),29),i||Sx),t))},p.ei=function(){return Ut(),Sx},R(un,"EParameterImpl",502),N(101,448,{109:1,93:1,92:1,154:1,197:1,57:1,19:1,178:1,69:1,113:1,467:1,52:1,98:1,157:1,101:1,448:1,290:1,117:1,118:1,677:1},E_e),p.Hh=function(t,i,a){var u,l,g,m;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return mn(),(this.Bb&256)!=0;case 3:return mn(),(this.Bb&512)!=0;case 4:return rt(this.s);case 5:return rt(this.t);case 6:return mn(),m=this.t,m>1||m==-1;case 7:return mn(),l=this.s,l>=1;case 8:return i?ph(this):this.r;case 9:return this.q;case 10:return mn(),(this.Bb&ep)!=0;case 11:return mn(),(this.Bb&pw)!=0;case 12:return mn(),(this.Bb&FT)!=0;case 13:return this.j;case 14:return mL(this);case 15:return mn(),(this.Bb&Ml)!=0;case 16:return mn(),(this.Bb&Sm)!=0;case 17:return dT(this);case 18:return mn(),(this.Bb&Dc)!=0;case 19:return mn(),g=os(this),!!(g&&(g.Bb&Dc)!=0);case 20:return mn(),(this.Bb&Xa)!=0;case 21:return i?os(this):this.b;case 22:return i?lCe(this):Ppt(this);case 23:return!this.a&&(this.a=new I3(wx,this,23)),this.a}return Lf(this,t-jn((Ut(),WA)),Vt((u=f(fn(this,16),29),u||WA),t),i,a)},p.Sh=function(t){var i,a,u,l;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return(this.Bb&256)==0;case 3:return(this.Bb&512)==0;case 4:return this.s!=0;case 5:return this.t!=1;case 6:return l=this.t,l>1||l==-1;case 7:return a=this.s,a>=1;case 8:return!!this.r&&!this.q.e&&ES(this.q).i==0;case 9:return!!this.q&&!(this.r&&!this.q.e&&ES(this.q).i==0);case 10:return(this.Bb&ep)==0;case 11:return(this.Bb&pw)!=0;case 12:return(this.Bb&FT)!=0;case 13:return this.j!=null;case 14:return mL(this)!=null;case 15:return(this.Bb&Ml)!=0;case 16:return(this.Bb&Sm)!=0;case 17:return!!dT(this);case 18:return(this.Bb&Dc)!=0;case 19:return u=os(this),!!u&&(u.Bb&Dc)!=0;case 20:return(this.Bb&Xa)==0;case 21:return!!this.b;case 22:return!!Ppt(this);case 23:return!!this.a&&this.a.i!=0}return Pf(this,t-jn((Ut(),WA)),Vt((i=f(fn(this,16),29),i||WA),t))},p.Zh=function(t,i){var a,u;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:Wce(this,dr(i));return;case 2:O0(this,lt(ut(i)));return;case 3:N0(this,lt(ut(i)));return;case 4:x0(this,f(i,17).a);return;case 5:vT(this,f(i,17).a);return;case 8:fE(this,f(i,141));return;case 9:u=xg(this,f(i,87),null),u&&u.lj();return;case 10:QP(this,lt(ut(i)));return;case 11:tL(this,lt(ut(i)));return;case 12:eL(this,lt(ut(i)));return;case 13:NSe(this,dr(i));return;case 15:nL(this,lt(ut(i)));return;case 16:rL(this,lt(ut(i)));return;case 18:gEn(this,lt(ut(i)));return;case 20:WCe(this,lt(ut(i)));return;case 21:zxe(this,f(i,19));return;case 23:!this.a&&(this.a=new I3(wx,this,23)),Jn(this.a),!this.a&&(this.a=new I3(wx,this,23)),Qi(this.a,f(i,16));return}$f(this,t-jn((Ut(),WA)),Vt((a=f(fn(this,16),29),a||WA),t),i)},p.ei=function(){return Ut(),WA},p.gi=function(t){var i,a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:me(this.Cb,88)&&OT(pd(f(this.Cb,88)),4),ju(this,null);return;case 2:O0(this,!0);return;case 3:N0(this,!0);return;case 4:x0(this,0);return;case 5:vT(this,1);return;case 8:fE(this,null);return;case 9:a=xg(this,null,null),a&&a.lj();return;case 10:QP(this,!0);return;case 11:tL(this,!1);return;case 12:eL(this,!1);return;case 13:this.i=null,qJ(this,null);return;case 15:nL(this,!1);return;case 16:rL(this,!1);return;case 18:VCe(this,!1),me(this.Cb,88)&&OT(pd(f(this.Cb,88)),2);return;case 20:WCe(this,!0);return;case 21:zxe(this,null);return;case 23:!this.a&&(this.a=new I3(wx,this,23)),Jn(this.a);return}jf(this,t-jn((Ut(),WA)),Vt((i=f(fn(this,16),29),i||WA),t))},p.li=function(){lCe(this),xP(Ns((Rl(),Ea),this)),ph(this),this.Bb|=1},p.rk=function(){return os(this)},p.Yk=function(){var t;return t=os(this),!!t&&(t.Bb&Dc)!=0},p.Zk=function(){return(this.Bb&Dc)!=0},p.$k=function(){return(this.Bb&Xa)!=0},p.Vk=function(t,i){return this.c=null,PCe(this,t,i)},p.Ib=function(){var t;return(this.Db&64)!=0?sQ(this):(t=new sh(sQ(this)),t.a+=" (containment: ",m0(t,(this.Bb&Dc)!=0),t.a+=", resolveProxies: ",m0(t,(this.Bb&Xa)!=0),t.a+=")",t.a)},R(un,"EReferenceImpl",101),N(546,118,{109:1,43:1,93:1,92:1,135:1,57:1,113:1,52:1,98:1,546:1,117:1,118:1},h5),p.Fb=function(t){return this===t},p.jd=function(){return this.b},p.kd=function(){return this.c},p.Hb=function(){return bS(this)},p.zi=function(t){Ovn(this,dr(t))},p.ld=function(t){return pvn(this,dr(t))},p.Hh=function(t,i,a){var u;switch(t){case 0:return this.b;case 1:return this.c}return Lf(this,t-jn((Ut(),Za)),Vt((u=f(fn(this,16),29),u||Za),t),i,a)},p.Sh=function(t){var i;switch(t){case 0:return this.b!=null;case 1:return this.c!=null}return Pf(this,t-jn((Ut(),Za)),Vt((i=f(fn(this,16),29),i||Za),t))},p.Zh=function(t,i){var a;switch(t){case 0:Nvn(this,dr(i));return;case 1:Fxe(this,dr(i));return}$f(this,t-jn((Ut(),Za)),Vt((a=f(fn(this,16),29),a||Za),t),i)},p.ei=function(){return Ut(),Za},p.gi=function(t){var i;switch(t){case 0:Bxe(this,null);return;case 1:Fxe(this,null);return}jf(this,t-jn((Ut(),Za)),Vt((i=f(fn(this,16),29),i||Za),t))},p.xi=function(){var t;return this.a==-1&&(t=this.b,this.a=t==null?0:C0(t)),this.a},p.yi=function(t){this.a=t},p.Ib=function(){var t;return(this.Db&64)!=0?Jh(this):(t=new sh(Jh(this)),t.a+=" (key: ",ys(t,this.b),t.a+=", value: ",ys(t,this.c),t.a+=")",t.a)},p.a=-1,p.b=null,p.c=null;var xc=R(un,"EStringToStringMapEntryImpl",546),mMt=Bi(Oi,"FeatureMap/Entry/Internal");N(558,1,mee),p.ul=function(t){return this.vl(f(t,52))},p.vl=function(t){return this.ul(t)},p.Fb=function(t){var i,a;return this===t?!0:me(t,74)?(i=f(t,74),i.Ik()==this.c?(a=this.kd(),a==null?i.kd()==null:ei(a,i.kd())):!1):!1},p.Ik=function(){return this.c},p.Hb=function(){var t;return t=this.kd(),Ti(this.c)^(t==null?0:Ti(t))},p.Ib=function(){var t,i;return t=this.c,i=cf(t.nk()).ui(),t.ve(),(i!=null&&i.length!=0?i+":"+t.ve():t.ve())+"="+this.kd()},R(un,"EStructuralFeatureImpl/BasicFeatureMapEntry",558),N(771,558,mee,A_e),p.vl=function(t){return new A_e(this.c,t)},p.kd=function(){return this.a},p.wl=function(t,i,a){return tTn(this,t,this.a,i,a)},p.xl=function(t,i,a){return nTn(this,t,this.a,i,a)},R(un,"EStructuralFeatureImpl/ContainmentUpdatingFeatureMapEntry",771),N(1289,1,{},Kat),p.vk=function(t,i,a,u,l){var g;return g=f(PP(t,this.b),218),g.Vl(this.a).Ck(u)},p.wk=function(t,i,a,u,l){var g;return g=f(PP(t,this.b),218),g.Ml(this.a,u,l)},p.xk=function(t,i,a,u,l){var g;return g=f(PP(t,this.b),218),g.Nl(this.a,u,l)},p.yk=function(t,i,a){var u;return u=f(PP(t,this.b),218),u.Vl(this.a).Nj()},p.zk=function(t,i,a,u){var l;l=f(PP(t,this.b),218),l.Vl(this.a).Wb(u)},p.Ak=function(t,i,a){return f(PP(t,this.b),218).Vl(this.a)},p.Bk=function(t,i,a){var u;u=f(PP(t,this.b),218),u.Vl(this.a).Dk()},R(un,"EStructuralFeatureImpl/InternalSettingDelegateFeatureMapDelegator",1289),N(89,1,{},y0,eE,S0,aE),p.vk=function(t,i,a,u,l){var g;if(g=i.hi(a),g==null&&i.ii(a,g=vQ(this,t)),!l)switch(this.e){case 50:case 41:return f(g,581).$j();case 40:return f(g,218).Sl()}return g},p.wk=function(t,i,a,u,l){var g,m;return m=i.hi(a),m==null&&i.ii(a,m=vQ(this,t)),g=f(m,71).Tk(u,l),g},p.xk=function(t,i,a,u,l){var g;return g=i.hi(a),g!=null&&(l=f(g,71).Uk(u,l)),l},p.yk=function(t,i,a){var u;return u=i.hi(a),u!=null&&f(u,76).Nj()},p.zk=function(t,i,a,u){var l;l=f(i.hi(a),76),!l&&i.ii(a,l=vQ(this,t)),l.Wb(u)},p.Ak=function(t,i,a){var u,l;return l=i.hi(a),l==null&&i.ii(a,l=vQ(this,t)),me(l,76)?f(l,76):(u=f(i.hi(a),15),new $nt(u))},p.Bk=function(t,i,a){var u;u=f(i.hi(a),76),!u&&i.ii(a,u=vQ(this,t)),u.Dk()},p.b=0,p.e=0,R(un,"EStructuralFeatureImpl/InternalSettingDelegateMany",89),N(496,1,{}),p.wk=function(t,i,a,u,l){throw ie(new cr)},p.xk=function(t,i,a,u,l){throw ie(new cr)},p.Ak=function(t,i,a){return new Vdt(this,t,i,a)};var Mb;R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingle",496),N(1306,1,Epe,Vdt),p.Ck=function(t){return this.a.vk(this.c,this.d,this.b,t,!0)},p.Nj=function(){return this.a.yk(this.c,this.d,this.b)},p.Wb=function(t){this.a.zk(this.c,this.d,this.b,t)},p.Dk=function(){this.a.Bk(this.c,this.d,this.b)},p.b=0,R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingle/1",1306),N(764,496,{},dTe),p.vk=function(t,i,a,u,l){return Dde(t,t.Lh(),t.Bh())==this.b?this.$k()&&u?Sde(t):t.Lh():null},p.wk=function(t,i,a,u,l){var g,m;return t.Lh()&&(l=(g=t.Bh(),g>=0?t.wh(l):t.Lh().Ph(t,-1-g,null,l))),m=Pi(t.zh(),this.e),t.yh(u,m,l)},p.xk=function(t,i,a,u,l){var g;return g=Pi(t.zh(),this.e),t.yh(null,g,l)},p.yk=function(t,i,a){var u;return u=Pi(t.zh(),this.e),!!t.Lh()&&t.Bh()==u},p.zk=function(t,i,a,u){var l,g,m,v,S;if(u!=null&&!Fde(this.a,u))throw ie(new eP(bee+(me(u,57)?Q3e(f(u,57).zh()):yxe(Fd(u)))+vee+this.a+"'"));if(l=t.Lh(),m=Pi(t.zh(),this.e),Ae(u)!==Ae(l)||t.Bh()!=m&&u!=null){if(cL(t,f(u,57)))throw ie(new dn(_$+t.Ib()));S=null,l&&(S=(g=t.Bh(),g>=0?t.wh(S):t.Lh().Ph(t,-1-g,null,S))),v=f(u,52),v&&(S=v.Nh(t,Pi(v.zh(),this.b),null,S)),S=t.yh(v,m,S),S&&S.lj()}else t.rh()&&t.sh()&&Zr(t,new Jo(t,1,m,u,u))},p.Bk=function(t,i,a){var u,l,g,m;u=t.Lh(),u?(m=(l=t.Bh(),l>=0?t.wh(null):t.Lh().Ph(t,-1-l,null,null)),g=Pi(t.zh(),this.e),m=t.yh(null,g,m),m&&m.lj()):t.rh()&&t.sh()&&Zr(t,new W7(t,1,this.e,null,null))},p.$k=function(){return!1},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleContainer",764),N(1290,764,{},Gut),p.$k=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleContainerResolving",1290),N(556,496,{}),p.vk=function(t,i,a,u,l){var g;return g=i.hi(a),g==null?this.b:Ae(g)===Ae(Mb)?null:g},p.yk=function(t,i,a){var u;return u=i.hi(a),u!=null&&(Ae(u)===Ae(Mb)||!ei(u,this.b))},p.zk=function(t,i,a,u){var l,g;t.rh()&&t.sh()?(l=(g=i.hi(a),g==null?this.b:Ae(g)===Ae(Mb)?null:g),u==null?this.c!=null?(i.ii(a,null),u=this.b):this.b!=null?i.ii(a,Mb):i.ii(a,null):(this.yl(u),i.ii(a,u)),Zr(t,this.d.zl(t,1,this.e,l,u))):u==null?this.c!=null?i.ii(a,null):this.b!=null?i.ii(a,Mb):i.ii(a,null):(this.yl(u),i.ii(a,u))},p.Bk=function(t,i,a){var u,l;t.rh()&&t.sh()?(u=(l=i.hi(a),l==null?this.b:Ae(l)===Ae(Mb)?null:l),i.ji(a),Zr(t,this.d.zl(t,1,this.e,u,this.b))):i.ji(a)},p.yl=function(t){throw ie(new Xnt)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData",556),N(gA,1,{},dg),p.zl=function(t,i,a,u,l){return new W7(t,i,a,u,l)},p.Al=function(t,i,a,u,l,g){return new jce(t,i,a,u,l,g)};var B7e,U7e,z7e,H7e,G7e,q7e,V7e,w1e,W7e;R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator",gA),N(1307,gA,{},o9),p.zl=function(t,i,a,u,l){return new GTe(t,i,a,lt(ut(u)),lt(ut(l)))},p.Al=function(t,i,a,u,l,g){return new bht(t,i,a,lt(ut(u)),lt(ut(l)),g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/1",1307),N(1308,gA,{},Ny),p.zl=function(t,i,a,u,l){return new kxe(t,i,a,f(u,219).a,f(l,219).a)},p.Al=function(t,i,a,u,l,g){return new dht(t,i,a,f(u,219).a,f(l,219).a,g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/2",1308),N(1309,gA,{},Z2),p.zl=function(t,i,a,u,l){return new Txe(t,i,a,f(u,179).a,f(l,179).a)},p.Al=function(t,i,a,u,l,g){return new fht(t,i,a,f(u,179).a,f(l,179).a,g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/3",1309),N(1310,gA,{},Nv),p.zl=function(t,i,a,u,l){return new qTe(t,i,a,ke(Oe(u)),ke(Oe(l)))},p.Al=function(t,i,a,u,l,g){return new hht(t,i,a,ke(Oe(u)),ke(Oe(l)),g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/4",1310),N(1311,gA,{},p5),p.zl=function(t,i,a,u,l){return new Axe(t,i,a,f(u,160).a,f(l,160).a)},p.Al=function(t,i,a,u,l,g){return new pht(t,i,a,f(u,160).a,f(l,160).a,g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/5",1311),N(1312,gA,{},g5),p.zl=function(t,i,a,u,l){return new VTe(t,i,a,f(u,17).a,f(l,17).a)},p.Al=function(t,i,a,u,l,g){return new lht(t,i,a,f(u,17).a,f(l,17).a,g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/6",1312),N(1313,gA,{},u0),p.zl=function(t,i,a,u,l){return new xxe(t,i,a,f(u,190).a,f(l,190).a)},p.Al=function(t,i,a,u,l,g){return new ght(t,i,a,f(u,190).a,f(l,190).a,g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/7",1313),N(1314,gA,{},a9),p.zl=function(t,i,a,u,l){return new Cxe(t,i,a,f(u,191).a,f(l,191).a)},p.Al=function(t,i,a,u,l,g){return new mht(t,i,a,f(u,191).a,f(l,191).a,g)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleData/NotificationCreator/8",1314),N(1292,556,{},Xdt),p.yl=function(t){if(!this.a.ck(t))throw ie(new eP(bee+Fd(t)+vee+this.a+"'"))},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleDataDynamic",1292),N(1293,556,{},Mlt),p.yl=function(t){},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleDataStatic",1293),N(765,556,{}),p.yk=function(t,i,a){var u;return u=i.hi(a),u!=null},p.zk=function(t,i,a,u){var l,g;t.rh()&&t.sh()?(l=!0,g=i.hi(a),g==null?(l=!1,g=this.b):Ae(g)===Ae(Mb)&&(g=null),u==null?this.c!=null?(i.ii(a,null),u=this.b):i.ii(a,Mb):(this.yl(u),i.ii(a,u)),Zr(t,this.d.Al(t,1,this.e,g,u,!l))):u==null?this.c!=null?i.ii(a,null):i.ii(a,Mb):(this.yl(u),i.ii(a,u))},p.Bk=function(t,i,a){var u,l;t.rh()&&t.sh()?(u=!0,l=i.hi(a),l==null?(u=!1,l=this.b):Ae(l)===Ae(Mb)&&(l=null),i.ji(a),Zr(t,this.d.Al(t,2,this.e,l,this.b,u))):i.ji(a)},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleDataUnsettable",765),N(1294,765,{},Jdt),p.yl=function(t){if(!this.a.ck(t))throw ie(new eP(bee+Fd(t)+vee+this.a+"'"))},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleDataUnsettableDynamic",1294),N(1295,765,{},jlt),p.yl=function(t){},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleDataUnsettableStatic",1295),N(399,496,{},NX),p.vk=function(t,i,a,u,l){var g,m,v,S,T;if(T=i.hi(a),this.qk()&&Ae(T)===Ae(Mb))return null;if(this.$k()&&u&&T!=null){if(v=f(T,52),v.Rh()&&(S=iw(t,v),v!=S)){if(!Fde(this.a,S))throw ie(new eP(bee+Fd(S)+vee+this.a+"'"));i.ii(a,T=S),this.Zk()&&(g=f(S,52),m=v.Ph(t,this.b?Pi(v.zh(),this.b):-1-Pi(t.zh(),this.e),null,null),!g.Lh()&&(m=g.Nh(t,this.b?Pi(g.zh(),this.b):-1-Pi(t.zh(),this.e),null,m)),m&&m.lj()),t.rh()&&t.sh()&&Zr(t,new W7(t,9,this.e,v,S))}return T}else return T},p.wk=function(t,i,a,u,l){var g,m;return m=i.hi(a),Ae(m)===Ae(Mb)&&(m=null),i.ii(a,u),this.Jj()?Ae(m)!==Ae(u)&&m!=null&&(g=f(m,52),l=g.Ph(t,Pi(g.zh(),this.b),null,l)):this.Zk()&&m!=null&&(l=f(m,52).Ph(t,-1-Pi(t.zh(),this.e),null,l)),t.rh()&&t.sh()&&(!l&&(l=new Fv(4)),l.kj(new W7(t,1,this.e,m,u))),l},p.xk=function(t,i,a,u,l){var g;return g=i.hi(a),Ae(g)===Ae(Mb)&&(g=null),i.ji(a),t.rh()&&t.sh()&&(!l&&(l=new Fv(4)),this.qk()?l.kj(new W7(t,2,this.e,g,null)):l.kj(new W7(t,1,this.e,g,null))),l},p.yk=function(t,i,a){var u;return u=i.hi(a),u!=null},p.zk=function(t,i,a,u){var l,g,m,v,S;if(u!=null&&!Fde(this.a,u))throw ie(new eP(bee+(me(u,57)?Q3e(f(u,57).zh()):yxe(Fd(u)))+vee+this.a+"'"));S=i.hi(a),v=S!=null,this.qk()&&Ae(S)===Ae(Mb)&&(S=null),m=null,this.Jj()?Ae(S)!==Ae(u)&&(S!=null&&(l=f(S,52),m=l.Ph(t,Pi(l.zh(),this.b),null,m)),u!=null&&(l=f(u,52),m=l.Nh(t,Pi(l.zh(),this.b),null,m))):this.Zk()&&Ae(S)!==Ae(u)&&(S!=null&&(m=f(S,52).Ph(t,-1-Pi(t.zh(),this.e),null,m)),u!=null&&(m=f(u,52).Nh(t,-1-Pi(t.zh(),this.e),null,m))),u==null&&this.qk()?i.ii(a,Mb):i.ii(a,u),t.rh()&&t.sh()?(g=new jce(t,1,this.e,S,u,this.qk()&&!v),m?(m.kj(g),m.lj()):Zr(t,g)):m&&m.lj()},p.Bk=function(t,i,a){var u,l,g,m,v;v=i.hi(a),m=v!=null,this.qk()&&Ae(v)===Ae(Mb)&&(v=null),g=null,v!=null&&(this.Jj()?(u=f(v,52),g=u.Ph(t,Pi(u.zh(),this.b),null,g)):this.Zk()&&(g=f(v,52).Ph(t,-1-Pi(t.zh(),this.e),null,g))),i.ji(a),t.rh()&&t.sh()?(l=new jce(t,this.qk()?2:1,this.e,v,null,m),g?(g.kj(l),g.lj()):Zr(t,l)):g&&g.lj()},p.Jj=function(){return!1},p.Zk=function(){return!1},p.$k=function(){return!1},p.qk=function(){return!1},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObject",399),N(557,399,{},Mse),p.Zk=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainment",557),N(1298,557,{},Mct),p.$k=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainmentResolving",1298),N(767,557,{},m_e),p.qk=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainmentUnsettable",767),N(1300,767,{},jct),p.$k=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainmentUnsettableResolving",1300),N(633,557,{},Xse),p.Jj=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainmentWithInverse",633),N(1299,633,{},But),p.$k=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainmentWithInverseResolving",1299),N(768,633,{},Q_e),p.qk=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainmentWithInverseUnsettable",768),N(1301,768,{},Uut),p.$k=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectContainmentWithInverseUnsettableResolving",1301),N(634,399,{},g_e),p.$k=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectResolving",634),N(1302,634,{},$ct),p.qk=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectResolvingUnsettable",1302),N(769,634,{},Z_e),p.Jj=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectResolvingWithInverse",769),N(1303,769,{},zut),p.qk=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectResolvingWithInverseUnsettable",1303),N(1296,399,{},Dct),p.qk=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectUnsettable",1296),N(766,399,{},eke),p.Jj=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectWithInverse",766),N(1297,766,{},Hut),p.qk=function(){return!0},R(un,"EStructuralFeatureImpl/InternalSettingDelegateSingleEObjectWithInverseUnsettable",1297),N(770,558,mee,Vke),p.vl=function(t){return new Vke(this.a,this.c,t)},p.kd=function(){return this.b},p.wl=function(t,i,a){return t_n(this,t,this.b,a)},p.xl=function(t,i,a){return n_n(this,t,this.b,a)},R(un,"EStructuralFeatureImpl/InverseUpdatingFeatureMapEntry",770),N(1304,1,Epe,$nt),p.Ck=function(t){return this.a},p.Nj=function(){return me(this.a,96)?f(this.a,96).Nj():!this.a.dc()},p.Wb=function(t){this.a.$b(),this.a.Fc(f(t,15))},p.Dk=function(){me(this.a,96)?f(this.a,96).Dk():this.a.$b()},R(un,"EStructuralFeatureImpl/SettingMany",1304),N(1305,558,mee,apt),p.ul=function(t){return new Fse((fi(),WF),this.b.ni(this.a,t))},p.kd=function(){return null},p.wl=function(t,i,a){return a},p.xl=function(t,i,a){return a},R(un,"EStructuralFeatureImpl/SimpleContentFeatureMapEntry",1305),N(635,558,mee,Fse),p.ul=function(t){return new Fse(this.c,t)},p.kd=function(){return this.a},p.wl=function(t,i,a){return a},p.xl=function(t,i,a){return a},R(un,"EStructuralFeatureImpl/SimpleFeatureMapEntry",635),N(394,490,Im,ab),p.Zi=function(t){return Pe(kh,Jt,29,t,0,1)},p.Vi=function(){return!1},R(un,"ESuperAdapter/1",394),N(443,436,{109:1,93:1,92:1,154:1,197:1,57:1,113:1,829:1,52:1,98:1,157:1,443:1,117:1,118:1},Q2),p.Hh=function(t,i,a){var u;switch(t){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),this.Ab;case 1:return this.zb;case 2:return!this.a&&(this.a=new G7(this,gs,this)),this.a}return Lf(this,t-jn((Ut(),k_)),Vt((u=f(fn(this,16),29),u||k_),t),i,a)},p.Qh=function(t,i,a){var u,l;switch(i){case 0:return!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),za(this.Ab,t,a);case 2:return!this.a&&(this.a=new G7(this,gs,this)),za(this.a,t,a)}return l=f(Vt((u=f(fn(this,16),29),u||(Ut(),k_)),i),69),l.tk().xk(this,Xu(this),i-jn((Ut(),k_)),t,a)},p.Sh=function(t){var i;switch(t){case 0:return!!this.Ab&&this.Ab.i!=0;case 1:return this.zb!=null;case 2:return!!this.a&&this.a.i!=0}return Pf(this,t-jn((Ut(),k_)),Vt((i=f(fn(this,16),29),i||k_),t))},p.Zh=function(t,i){var a;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab),!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Qi(this.Ab,f(i,16));return;case 1:ju(this,dr(i));return;case 2:!this.a&&(this.a=new G7(this,gs,this)),Jn(this.a),!this.a&&(this.a=new G7(this,gs,this)),Qi(this.a,f(i,16));return}$f(this,t-jn((Ut(),k_)),Vt((a=f(fn(this,16),29),a||k_),t),i)},p.ei=function(){return Ut(),k_},p.gi=function(t){var i;switch(t){case 0:!this.Ab&&(this.Ab=new Be(Ir,this,0,3)),Jn(this.Ab);return;case 1:ju(this,null);return;case 2:!this.a&&(this.a=new G7(this,gs,this)),Jn(this.a);return}jf(this,t-jn((Ut(),k_)),Vt((i=f(fn(this,16),29),i||k_),t))},R(un,"ETypeParameterImpl",443),N(444,81,ec,G7),p.Kj=function(t,i){return GOn(this,f(t,87),i)},p.Lj=function(t,i){return qOn(this,f(t,87),i)},R(un,"ETypeParameterImpl/1",444),N(632,45,sA,bae),p.ec=function(){return new iY(this)},R(un,"ETypeParameterImpl/2",632),N(553,Ig,Pl,iY),p.Ec=function(t){return uut(this,f(t,87))},p.Fc=function(t){var i,a,u;for(u=!1,a=t.Jc();a.Ob();)i=f(a.Pb(),87),qr(this.a,i,"")==null&&(u=!0);return u},p.$b=function(){nu(this.a)},p.Gc=function(t){return Yu(this.a,t)},p.Jc=function(){var t;return t=new yT(new Vi(this.a).a),new oY(t)},p.Kc=function(t){return Wpt(this,t)},p.gc=function(){return f7(this.a)},R(un,"ETypeParameterImpl/2/1",553),N(554,1,aa,oY),p.Nb=function(t){va(this,t)},p.Pb=function(){return f(G3(this.a).jd(),87)},p.Ob=function(){return this.a.b},p.Qb=function(){rmt(this.a)},R(un,"ETypeParameterImpl/2/1/1",554),N(1270,45,sA,Ort),p._b=function(t){return ta(t)?Mce(this,t):!!Is(this.f,t)},p.xc=function(t){var i,a;return i=ta(t)?wu(this,t):ac(Is(this.f,t)),me(i,830)?(a=f(i,830),i=a.Hk(),qr(this,f(t,240),i),i):i??(t==null?(Mae(),vMt):null)},R(un,"EValidatorRegistryImpl",1270),N(1288,704,{109:1,93:1,92:1,466:1,154:1,57:1,113:1,1986:1,52:1,98:1,157:1,117:1,118:1},jh),p.ni=function(t,i){switch(t.ek()){case 21:case 22:case 23:case 24:case 26:case 31:case 32:case 37:case 38:case 39:case 40:case 43:case 44:case 48:case 49:case 20:return i==null?null:Qs(i);case 25:return dkn(i);case 27:return x_n(i);case 28:return C_n(i);case 29:return i==null?null:Bst(UF[0],f(i,205));case 41:return i==null?"":Vy(f(i,295));case 42:return Qs(i);case 50:return dr(i);default:throw ie(new dn(VL+t.ve()+n_))}},p.oi=function(t){var i,a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue;switch(t.G==-1&&(t.G=(F=cf(t),F?P0(F.ri(),t):-1)),t.G){case 0:return a=new pae,a;case 1:return i=new Wu,i;case 2:return u=new $5,u;case 4:return l=new sY,l;case 5:return g=new Irt,g;case 6:return m=new Qnt,m;case 7:return v=new om,v;case 10:return T=new ZC,T;case 11:return A=new gae,A;case 12:return L=new oft,L;case 13:return z=new mae,z;case 14:return W=new E_e,W;case 17:return J=new h5,J;case 18:return S=new cS,S;case 19:return ue=new Q2,ue;default:throw ie(new dn(ope+t.zb+n_))}},p.pi=function(t,i){switch(t.ek()){case 20:return i==null?null:new G2e(i);case 21:return i==null?null:new Gv(i);case 23:case 22:return i==null?null:d3n(i);case 26:case 24:return i==null?null:jH(ff(i,-128,127)<<24>>24);case 25:return o6n(i);case 27:return qAn(i);case 28:return VAn(i);case 29:return dNn(i);case 32:case 31:return i==null?null:AT(i);case 38:case 37:return i==null?null:new m2e(i);case 40:case 39:return i==null?null:rt(ff(i,Ra,Qr));case 41:return null;case 42:return i==null,null;case 44:case 43:return i==null?null:TT(bQ(i));case 49:case 48:return i==null?null:ZP(ff(i,wee,32767)<<16>>16);case 50:return i;default:throw ie(new dn(VL+t.ve()+n_))}},R(un,"EcoreFactoryImpl",1288),N(545,184,{109:1,93:1,92:1,154:1,197:1,57:1,240:1,113:1,1984:1,52:1,98:1,157:1,184:1,545:1,117:1,118:1,675:1},Ndt),p.gb=!1,p.hb=!1;var K7e,bMt=!1;R(un,"EcorePackageImpl",545),N(1194,1,{830:1},_k),p.Hk=function(){return Qst(),wMt},R(un,"EcorePackageImpl/1",1194),N(1203,1,jr,s9),p.ck=function(t){return me(t,154)},p.dk=function(t){return Pe(EV,Jt,154,t,0,1)},R(un,"EcorePackageImpl/10",1203),N(1204,1,jr,c9),p.ck=function(t){return me(t,197)},p.dk=function(t){return Pe(l1e,Jt,197,t,0,1)},R(un,"EcorePackageImpl/11",1204),N(1205,1,jr,kk),p.ck=function(t){return me(t,57)},p.dk=function(t){return Pe(Dw,Jt,57,t,0,1)},R(un,"EcorePackageImpl/12",1205),N(1206,1,jr,u9),p.ck=function(t){return me(t,400)},p.dk=function(t){return Pe(Th,Z4e,62,t,0,1)},R(un,"EcorePackageImpl/13",1206),N(1207,1,jr,QO),p.ck=function(t){return me(t,240)},p.dk=function(t){return Pe(Bp,Jt,240,t,0,1)},R(un,"EcorePackageImpl/14",1207),N(1208,1,jr,l9),p.ck=function(t){return me(t,502)},p.dk=function(t){return Pe(S_,Jt,2062,t,0,1)},R(un,"EcorePackageImpl/15",1208),N(1209,1,jr,HU),p.ck=function(t){return me(t,101)},p.dk=function(t){return Pe(yx,pA,19,t,0,1)},R(un,"EcorePackageImpl/16",1209),N(1210,1,jr,m5),p.ck=function(t){return me(t,178)},p.dk=function(t){return Pe(El,pA,178,t,0,1)},R(un,"EcorePackageImpl/17",1210),N(1211,1,jr,zc),p.ck=function(t){return me(t,467)},p.dk=function(t){return Pe(vx,Jt,467,t,0,1)},R(un,"EcorePackageImpl/18",1211),N(1212,1,jr,Nu),p.ck=function(t){return me(t,546)},p.dk=function(t){return Pe(xc,k3t,546,t,0,1)},R(un,"EcorePackageImpl/19",1212),N(1195,1,jr,as),p.ck=function(t){return me(t,330)},p.dk=function(t){return Pe(wx,pA,35,t,0,1)},R(un,"EcorePackageImpl/2",1195),N(1213,1,jr,$s),p.ck=function(t){return me(t,246)},p.dk=function(t){return Pe(gs,U3t,87,t,0,1)},R(un,"EcorePackageImpl/20",1213),N(1214,1,jr,Hc),p.ck=function(t){return me(t,443)},p.dk=function(t){return Pe(nl,Jt,829,t,0,1)},R(un,"EcorePackageImpl/21",1214),N(1215,1,jr,nm),p.ck=function(t){return Xk(t)},p.dk=function(t){return Pe(to,Ye,471,t,8,1)},R(un,"EcorePackageImpl/22",1215),N(1216,1,jr,Ry),p.ck=function(t){return me(t,195)},p.dk=function(t){return Pe($l,Ye,195,t,0,2)},R(un,"EcorePackageImpl/23",1216),N(1217,1,jr,Tk),p.ck=function(t){return me(t,219)},p.dk=function(t){return Pe(BR,Ye,219,t,0,1)},R(un,"EcorePackageImpl/24",1217),N(1218,1,jr,b5),p.ck=function(t){return me(t,179)},p.dk=function(t){return Pe(D$,Ye,179,t,0,1)},R(un,"EcorePackageImpl/25",1218),N(1219,1,jr,Rv),p.ck=function(t){return me(t,205)},p.dk=function(t){return Pe(Iee,Ye,205,t,0,1)},R(un,"EcorePackageImpl/26",1219),N(1220,1,jr,ql),p.ck=function(t){return!1},p.dk=function(t){return Pe(fje,Jt,2159,t,0,1)},R(un,"EcorePackageImpl/27",1220),N(1221,1,jr,Py),p.ck=function(t){return Jk(t)},p.dk=function(t){return Pe(mo,Ye,341,t,7,1)},R(un,"EcorePackageImpl/28",1221),N(1222,1,jr,eS),p.ck=function(t){return me(t,61)},p.dk=function(t){return Pe(C7e,BT,61,t,0,1)},R(un,"EcorePackageImpl/29",1222),N(1196,1,jr,v5),p.ck=function(t){return me(t,503)},p.dk=function(t){return Pe(Ir,{3:1,4:1,5:1,1979:1},582,t,0,1)},R(un,"EcorePackageImpl/3",1196),N(1223,1,jr,sb),p.ck=function(t){return me(t,564)},p.dk=function(t){return Pe(O7e,Jt,1985,t,0,1)},R(un,"EcorePackageImpl/30",1223),N(1224,1,jr,xk),p.ck=function(t){return me(t,159)},p.dk=function(t){return Pe(Q7e,BT,159,t,0,1)},R(un,"EcorePackageImpl/31",1224),N(1225,1,jr,QC),p.ck=function(t){return me(t,74)},p.dk=function(t){return Pe(zne,X3t,74,t,0,1)},R(un,"EcorePackageImpl/32",1225),N(1226,1,jr,fg),p.ck=function(t){return me(t,160)},p.dk=function(t){return Pe(eD,Ye,160,t,0,1)},R(un,"EcorePackageImpl/33",1226),N(1227,1,jr,w5),p.ck=function(t){return me(t,17)},p.dk=function(t){return Pe(ba,Ye,17,t,0,1)},R(un,"EcorePackageImpl/34",1227),N(1228,1,jr,y5),p.ck=function(t){return me(t,295)},p.dk=function(t){return Pe(d6e,Jt,295,t,0,1)},R(un,"EcorePackageImpl/35",1228),N(1229,1,jr,Ck),p.ck=function(t){return me(t,190)},p.dk=function(t){return Pe(i_,Ye,190,t,0,1)},R(un,"EcorePackageImpl/36",1229),N(1230,1,jr,e3),p.ck=function(t){return me(t,91)},p.dk=function(t){return Pe(f6e,Jt,91,t,0,1)},R(un,"EcorePackageImpl/37",1230),N(1231,1,jr,Ly),p.ck=function(t){return me(t,583)},p.dk=function(t){return Pe(Y7e,Jt,583,t,0,1)},R(un,"EcorePackageImpl/38",1231),N(1232,1,jr,ih),p.ck=function(t){return!1},p.dk=function(t){return Pe(hje,Jt,2160,t,0,1)},R(un,"EcorePackageImpl/39",1232),N(1197,1,jr,Dy),p.ck=function(t){return me(t,88)},p.dk=function(t){return Pe(kh,Jt,29,t,0,1)},R(un,"EcorePackageImpl/4",1197),N(1233,1,jr,Ak),p.ck=function(t){return me(t,191)},p.dk=function(t){return Pe(o_,Ye,191,t,0,1)},R(un,"EcorePackageImpl/40",1233),N(1234,1,jr,Ik),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(un,"EcorePackageImpl/41",1234),N(1235,1,jr,E5),p.ck=function(t){return me(t,580)},p.dk=function(t){return Pe(I7e,Jt,580,t,0,1)},R(un,"EcorePackageImpl/42",1235),N(1236,1,jr,S5),p.ck=function(t){return!1},p.dk=function(t){return Pe(pje,Ye,2161,t,0,1)},R(un,"EcorePackageImpl/43",1236),N(1237,1,jr,d9),p.ck=function(t){return me(t,43)},p.dk=function(t){return Pe(AE,yQ,43,t,0,1)},R(un,"EcorePackageImpl/44",1237),N(1198,1,jr,Ok),p.ck=function(t){return me(t,141)},p.dk=function(t){return Pe(Up,Jt,141,t,0,1)},R(un,"EcorePackageImpl/5",1198),N(1199,1,jr,_5),p.ck=function(t){return me(t,155)},p.dk=function(t){return Pe(m1e,Jt,155,t,0,1)},R(un,"EcorePackageImpl/6",1199),N(1200,1,jr,$h),p.ck=function(t){return me(t,455)},p.dk=function(t){return Pe(Une,Jt,670,t,0,1)},R(un,"EcorePackageImpl/7",1200),N(1201,1,jr,f9),p.ck=function(t){return me(t,564)},p.dk=function(t){return Pe(j1,Jt,679,t,0,1)},R(un,"EcorePackageImpl/8",1201),N(1202,1,jr,rm),p.ck=function(t){return me(t,466)},p.dk=function(t){return Pe(BF,Jt,466,t,0,1)},R(un,"EcorePackageImpl/9",1202),N(1013,2026,_3t,eit),p.Ji=function(t,i){zxn(this,f(i,412))},p.Ni=function(t,i){Wyt(this,t,f(i,412))},R(un,"MinimalEObjectImpl/1ArrayDelegatingAdapterList",1013),N(1014,150,dq,wdt),p.gj=function(){return this.a.a},R(un,"MinimalEObjectImpl/1ArrayDelegatingAdapterList/1",1014),N(1042,1041,{},Pst),R("org.eclipse.emf.ecore.plugin","EcorePlugin",1042);var Y7e=Bi(J3t,"Resource");N(779,1464,Z3t),p.El=function(t){},p.Fl=function(t){},p.Bl=function(){return!this.a&&(this.a=new cae(this)),this.a},p.Cl=function(t){var i,a,u,l,g;if(u=t.length,u>0)if(gn(0,t.length),t.charCodeAt(0)==47){for(g=new Du(4),l=1,i=1;i<u;++i)gn(i,t.length),t.charCodeAt(i)==47&&(Qe(g,l==i?"":(ga(l,i,t.length),t.substr(l,i-l))),l=i+1);return Qe(g,(gn(l,t.length+1),t.substr(l))),MIn(this,g)}else gn(u-1,t.length),t.charCodeAt(u-1)==63&&(a=U_e(t,gl(63),u-2),a>0&&(t=(ga(0,a,t.length),t.substr(0,a))));return QRn(this,t)},p.Dl=function(){return this.c},p.Ib=function(){var t;return Vy(this.Om)+"@"+(t=Ti(this)>>>0,t.toString(16))+" uri='"+this.d+"'"},p.b=!1,R(Spe,"ResourceImpl",779),N(1465,779,Z3t,Bnt),R(Spe,"BinaryResourceImpl",1465),N(1154,692,mpe),p.$i=function(t){return me(t,57)?_yn(this,f(t,57)):me(t,583)?new Nn(f(t,583).Bl()):Ae(t)===Ae(this.f)?f(t,16).Jc():(gP(),kV.a)},p.Ob=function(){return WAe(this)},p.a=!1,R(Oi,"EcoreUtil/ContentTreeIterator",1154),N(1466,1154,mpe,Zlt),p.$i=function(t){return Ae(t)===Ae(this.f)?f(t,15).Jc():new jht(f(t,57))},R(Spe,"ResourceImpl/5",1466),N(642,2038,B3t,cae),p.Gc=function(t){return this.i<=4?lL(this,t):me(t,52)&&f(t,52).Fh()==this.a},p.Ji=function(t,i){t==this.i-1&&(this.a.b||(this.a.b=!0))},p.Li=function(t,i){t==0?this.a.b||(this.a.b=!0):fue(this,t,i)},p.Ni=function(t,i){},p.Oi=function(t,i,a){},p.Ij=function(){return 2},p.gj=function(){return this.a},p.Jj=function(){return!0},p.Kj=function(t,i){var a;return a=f(t,52),i=a.bi(this.a,i),i},p.Lj=function(t,i){var a;return a=f(t,52),a.bi(null,i)},p.Mj=function(){return!1},p.Pi=function(){return!0},p.Zi=function(t){return Pe(Dw,Jt,57,t,0,1)},p.Vi=function(){return!1},R(Spe,"ResourceImpl/ContentsEList",642),N(948,2008,IL,Unt),p.dd=function(t){return this.a.Hi(t)},p.gc=function(){return this.a.gc()},R(Oi,"AbstractSequentialInternalEList/1",948);var X7e,J7e,Ea,Z7e;N(619,1,{},Zut);var Hne,Gne;R(Oi,"BasicExtendedMetaData",619),N(1145,1,{},Yat),p.Gl=function(){return null},p.Hl=function(){return this.a==-2&&C(this,iNn(this.d,this.b)),this.a},p.Il=function(){return null},p.Jl=function(){return Ht(),Ht(),Ga},p.ve=function(){return this.c==XL&&te(this,mvt(this.d,this.b)),this.c},p.Kl=function(){return 0},p.a=-2,p.c=XL,R(Oi,"BasicExtendedMetaData/EClassExtendedMetaDataImpl",1145),N(1146,1,{},wht),p.Gl=function(){return this.a==(RP(),Hne)&&j(this,FPn(this.f,this.b)),this.a},p.Hl=function(){return 0},p.Il=function(){return this.c==(RP(),Hne)&&Se(this,BPn(this.f,this.b)),this.c},p.Jl=function(){return!this.d&&pt(this,wDn(this.f,this.b)),this.d},p.ve=function(){return this.e==XL&&$n(this,mvt(this.f,this.b)),this.e},p.Kl=function(){return this.g==-2&&ln(this,_On(this.f,this.b)),this.g},p.e=XL,p.g=-2,R(Oi,"BasicExtendedMetaData/EDataTypeExtendedMetaDataImpl",1146),N(1144,1,{},Xat),p.b=!1,p.c=!1,R(Oi,"BasicExtendedMetaData/EPackageExtendedMetaDataImpl",1144),N(1147,1,{},yht),p.c=-2,p.e=XL,p.f=XL,R(Oi,"BasicExtendedMetaData/EStructuralFeatureExtendedMetaDataImpl",1147),N(576,617,ec,yX),p.Ij=function(){return this.c},p.ll=function(){return!1},p.Ti=function(t,i){return i},p.c=0,R(Oi,"EDataTypeEList",576);var Q7e=Bi(Oi,"FeatureMap");N(75,576,{3:1,4:1,20:1,31:1,55:1,16:1,15:1,59:1,70:1,66:1,61:1,76:1,159:1,218:1,1982:1,71:1,96:1},eo),p._c=function(t,i){a5n(this,t,f(i,74))},p.Ec=function(t){return x6n(this,f(t,74))},p.Ei=function(t){kwn(this,f(t,74))},p.Kj=function(t,i){return n1n(this,f(t,74),i)},p.Lj=function(t,i){return q_e(this,f(t,74),i)},p.Qi=function(t,i){return CLn(this,t,i)},p.Ti=function(t,i){return c8n(this,t,f(i,74))},p.fd=function(t,i){return q5n(this,t,f(i,74))},p.Rj=function(t,i){return r1n(this,f(t,74),i)},p.Sj=function(t,i){return Aut(this,f(t,74),i)},p.Tj=function(t,i,a){return fOn(this,f(t,74),f(i,74),a)},p.Wi=function(t,i){return ide(this,t,f(i,74))},p.Ll=function(t,i){return UIe(this,t,i)},p.ad=function(t,i){var a,u,l,g,m,v,S,T,A;for(T=new OS(i.gc()),l=i.Jc();l.Ob();)if(u=f(l.Pb(),74),g=u.Ik(),x1(this.e,g))(!g.Pi()||!dJ(this,g,u.kd())&&!lL(T,u))&&Qn(T,u);else{for(A=Ju(this.e.zh(),g),a=f(this.g,122),m=!0,v=0;v<this.i;++v)if(S=a[v],A.Zl(S.Ik())){f(Y3(this,v,u),74),m=!1;break}m&&Qn(T,u)}return cCe(this,t,T)},p.Fc=function(t){var i,a,u,l,g,m,v,S,T;for(S=new OS(t.gc()),u=t.Jc();u.Ob();)if(a=f(u.Pb(),74),l=a.Ik(),x1(this.e,l))(!l.Pi()||!dJ(this,l,a.kd())&&!lL(S,a))&&Qn(S,a);else{for(T=Ju(this.e.zh(),l),i=f(this.g,122),g=!0,m=0;m<this.i;++m)if(v=i[m],T.Zl(v.Ik())){f(Y3(this,m,a),74),g=!1;break}g&&Qn(S,a)}return Qi(this,S)},p.Ci=function(t){return this.j=-1,Bde(this,this.i,t)},p.Ml=function(t,i,a){return PIe(this,t,i,a)},p.Uk=function(t,i){return CG(this,t,i)},p.Nl=function(t,i,a){return sOe(this,t,i,a)},p.Ol=function(){return this},p.Pl=function(t,i){return OG(this,t,i)},p.Ql=function(t){return f(be(this,t),74).Ik()},p.Rl=function(t){return f(be(this,t),74).kd()},p.Sl=function(){return this.b},p.Jj=function(){return!0},p.Qj=function(){return!0},p.Tl=function(t){return!oG(this,t)},p.Zi=function(t){return Pe(mMt,X3t,339,t,0,1)},p.ml=function(t){return jse(this,t)},p.Wb=function(t){hH(this,t)},p.Ul=function(t,i){hQ(this,t,i)},p.Vl=function(t){return wbt(this,t)},p.Wl=function(t){Qvt(this,t)},R(Oi,"BasicFeatureMap",75),N(1906,1,yb),p.Nb=function(t){va(this,t)},p.Rb=function(t){if(this.g==-1)throw ie(new Tl);MX(this);try{z2t(this.e,this.b,this.a,t),this.d=this.e.j,AZ(this)}catch(i){throw i=go(i),me(i,97)?ie(new If):ie(i)}},p.Ob=function(){return rle(this)},p.Sb=function(){return a0t(this)},p.Pb=function(){return AZ(this)},p.Tb=function(){return this.a},p.Ub=function(){var t;if(a0t(this))return MX(this),this.g=--this.a,this.rl()&&(t=Bj(this.e,this.b,this.c,this.a,this.j),this.j=t),this.i=0,this.j;throw ie(new ic)},p.Vb=function(){return this.a-1},p.Qb=function(){if(this.g==-1)throw ie(new Tl);MX(this);try{Nyt(this.e,this.b,this.g),this.d=this.e.j,this.g<this.a&&(--this.a,--this.c),--this.g}catch(t){throw t=go(t),me(t,97)?ie(new If):ie(t)}},p.rl=function(){return!1},p.Wb=function(t){if(this.g==-1)throw ie(new Tl);MX(this);try{f_t(this.e,this.b,this.g,t),this.d=this.e.j}catch(i){throw i=go(i),me(i,97)?ie(new If):ie(i)}},p.a=0,p.c=0,p.d=0,p.f=!1,p.g=0,p.i=0,R(Oi,"FeatureMapUtil/BasicFeatureEIterator",1906),N(409,1906,yb,GP),p.Xl=function(){var t,i,a;for(a=this.e.i,t=f(this.e.g,122);this.c<a;){if(i=t[this.c],this.k.Zl(i.Ik()))return this.j=this.f?i:i.kd(),this.i=2,!0;++this.c}return this.i=1,this.g=-1,!1},p.Yl=function(){var t,i;for(t=f(this.e.g,122);--this.c>=0;)if(i=t[this.c],this.k.Zl(i.Ik()))return this.j=this.f?i:i.kd(),this.i=-2,!0;return this.i=-1,this.g=-1,!1},R(Oi,"BasicFeatureMap/FeatureEIterator",409),N(661,409,yb,wse),p.rl=function(){return!0},R(Oi,"BasicFeatureMap/ResolvingFeatureEIterator",661),N(946,480,pee,zst),p.mj=function(){return this},R(Oi,"EContentsEList/1",946),N(947,480,pee,cst),p.rl=function(){return!1},R(Oi,"EContentsEList/2",947),N(945,284,gee,Hst),p.tl=function(t){},p.Ob=function(){return!1},p.Sb=function(){return!1},R(Oi,"EContentsEList/FeatureIteratorImpl/1",945),N(819,576,ec,YSe),p.Ki=function(){this.a=!0},p.Nj=function(){return this.a},p.Dk=function(){var t;Jn(this),Ld(this.e)?(t=this.a,this.a=!1,Zr(this.e,new Vh(this.e,2,this.c,t,!1))):this.a=!1},p.a=!1,R(Oi,"EDataTypeEList/Unsettable",819),N(1904,576,ec,Kst),p.Pi=function(){return!0},R(Oi,"EDataTypeUniqueEList",1904),N(1905,819,ec,Yst),p.Pi=function(){return!0},R(Oi,"EDataTypeUniqueEList/Unsettable",1905),N(142,81,ec,xl),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectContainmentEList/Resolving",142),N(1148,540,ec,Xst),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectContainmentEList/Unsettable/Resolving",1148),N(748,14,ec,M_e),p.Ki=function(){this.a=!0},p.Nj=function(){return this.a},p.Dk=function(){var t;Jn(this),Ld(this.e)?(t=this.a,this.a=!1,Zr(this.e,new Vh(this.e,2,this.c,t,!1))):this.a=!1},p.a=!1,R(Oi,"EObjectContainmentWithInverseEList/Unsettable",748),N(1182,748,ec,fut),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectContainmentWithInverseEList/Unsettable/Resolving",1182),N(740,489,ec,XSe),p.Ki=function(){this.a=!0},p.Nj=function(){return this.a},p.Dk=function(){var t;Jn(this),Ld(this.e)?(t=this.a,this.a=!1,Zr(this.e,new Vh(this.e,2,this.c,t,!1))):this.a=!1},p.a=!1,R(Oi,"EObjectEList/Unsettable",740),N(334,489,ec,I3),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectResolvingEList",334),N(1790,740,ec,Jst),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectResolvingEList/Unsettable",1790),N(1467,1,{},h9);var vMt;R(Oi,"EObjectValidator",1467),N(544,489,ec,FX),p.fl=function(){return this.d},p.gl=function(){return this.b},p.Jj=function(){return!0},p.jl=function(){return!0},p.b=0,R(Oi,"EObjectWithInverseEList",544),N(1185,544,ec,hut),p.il=function(){return!0},R(Oi,"EObjectWithInverseEList/ManyInverse",1185),N(620,544,ec,Use),p.Ki=function(){this.a=!0},p.Nj=function(){return this.a},p.Dk=function(){var t;Jn(this),Ld(this.e)?(t=this.a,this.a=!1,Zr(this.e,new Vh(this.e,2,this.c,t,!1))):this.a=!1},p.a=!1,R(Oi,"EObjectWithInverseEList/Unsettable",620),N(1184,620,ec,put),p.il=function(){return!0},R(Oi,"EObjectWithInverseEList/Unsettable/ManyInverse",1184),N(749,544,ec,j_e),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectWithInverseResolvingEList",749),N(33,749,ec,Qt),p.il=function(){return!0},R(Oi,"EObjectWithInverseResolvingEList/ManyInverse",33),N(750,620,ec,$_e),p.kl=function(){return!0},p.Ti=function(t,i){return kR(this,t,f(i,57))},R(Oi,"EObjectWithInverseResolvingEList/Unsettable",750),N(1183,750,ec,gut),p.il=function(){return!0},R(Oi,"EObjectWithInverseResolvingEList/Unsettable/ManyInverse",1183),N(1149,617,ec),p.Ii=function(){return(this.b&1792)==0},p.Ki=function(){this.b|=1},p.hl=function(){return(this.b&4)!=0},p.Jj=function(){return(this.b&40)!=0},p.il=function(){return(this.b&16)!=0},p.jl=function(){return(this.b&8)!=0},p.kl=function(){return(this.b&pw)!=0},p.Zk=function(){return(this.b&32)!=0},p.ll=function(){return(this.b&ep)!=0},p.ck=function(t){return this.d?Zht(this.d,t):this.Ik().Ek().ck(t)},p.Nj=function(){return(this.b&2)!=0?(this.b&1)!=0:this.i!=0},p.Pi=function(){return(this.b&128)!=0},p.Dk=function(){var t;Jn(this),(this.b&2)!=0&&(Ld(this.e)?(t=(this.b&1)!=0,this.b&=-2,Q5(this,new Vh(this.e,2,Pi(this.e.zh(),this.Ik()),t,!1))):this.b&=-2)},p.Vi=function(){return(this.b&1536)==0},p.b=0,R(Oi,"EcoreEList/Generic",1149),N(1150,1149,ec,dft),p.Ik=function(){return this.a},R(Oi,"EcoreEList/Dynamic",1150),N(747,66,Im,r2e),p.Zi=function(t){return BH(this.a.a,t)},R(Oi,"EcoreEMap/1",747),N(746,81,ec,Rke),p.Ji=function(t,i){OZ(this.b,f(i,135))},p.Li=function(t,i){Zbt(this.b)},p.Mi=function(t,i,a){var u;++(u=this.b,f(i,135),u).e},p.Ni=function(t,i){hle(this.b,f(i,135))},p.Oi=function(t,i,a){hle(this.b,f(a,135)),Ae(a)===Ae(i)&&f(a,135).yi(imn(f(i,135).jd())),OZ(this.b,f(i,135))},R(Oi,"EcoreEMap/DelegateEObjectContainmentEList",746),N(1180,140,J4e,lbt),R(Oi,"EcoreEMap/Unsettable",1180),N(1181,746,ec,mut),p.Ki=function(){this.a=!0},p.Nj=function(){return this.a},p.Dk=function(){var t;Jn(this),Ld(this.e)?(t=this.a,this.a=!1,Zr(this.e,new Vh(this.e,2,this.c,t,!1))):this.a=!1},p.a=!1,R(Oi,"EcoreEMap/Unsettable/UnsettableDelegateEObjectContainmentEList",1181),N(1153,221,sA,fdt),p.a=!1,p.b=!1,R(Oi,"EcoreUtil/Copier",1153),N(742,1,aa,jht),p.Nb=function(t){va(this,t)},p.Ob=function(){return Y0t(this)},p.Pb=function(){var t;return Y0t(this),t=this.b,this.b=null,t},p.Qb=function(){this.a.Qb()},R(Oi,"EcoreUtil/ProperContentIterator",742),N(1468,1467,{},vc);var wMt;R(Oi,"EcoreValidator",1468);var yMt;Bi(Oi,"FeatureMapUtil/Validator"),N(1248,1,{1987:1},p9),p.Zl=function(t){return!0},R(Oi,"FeatureMapUtil/1",1248),N(755,1,{1987:1},vOe),p.Zl=function(t){var i;return this.c==t?!0:(i=ut(bn(this.a,t)),i==null?qPn(this,t)?(jpt(this.a,t,(mn(),QL)),!0):(jpt(this.a,t,(mn(),yw)),!1):i==(mn(),QL))},p.e=!1;var y1e;R(Oi,"FeatureMapUtil/BasicValidator",755),N(756,45,sA,WSe),R(Oi,"FeatureMapUtil/BasicValidator/Cache",756),N(493,55,{20:1,31:1,55:1,16:1,15:1,61:1,76:1,71:1,96:1},Wz),p._c=function(t,i){z2t(this.c,this.b,t,i)},p.Ec=function(t){return UIe(this.c,this.b,t)},p.ad=function(t,i){return rMn(this.c,this.b,t,i)},p.Fc=function(t){return M7(this,t)},p.Di=function(t,i){okn(this.c,this.b,t,i)},p.Tk=function(t,i){return PIe(this.c,this.b,t,i)},p.Xi=function(t){return fQ(this.c,this.b,t,!1)},p.Fi=function(){return _st(this.c,this.b)},p.Gi=function(){return qgn(this.c,this.b)},p.Hi=function(t){return r_n(this.c,this.b,t)},p.Uk=function(t,i){return Wct(this,t,i)},p.$b=function(){TN(this)},p.Gc=function(t){return dJ(this.c,this.b,t)},p.Hc=function(t){return Qkn(this.c,this.b,t)},p.Xb=function(t){return fQ(this.c,this.b,t,!0)},p.Ck=function(t){return this},p.bd=function(t){return lSn(this.c,this.b,t)},p.dc=function(){return XY(this)},p.Nj=function(){return!oG(this.c,this.b)},p.Jc=function(){return H_n(this.c,this.b)},p.cd=function(){return G_n(this.c,this.b)},p.dd=function(t){return uCn(this.c,this.b,t)},p.Qi=function(t,i){return r_t(this.c,this.b,t,i)},p.Ri=function(t,i){o_n(this.c,this.b,t,i)},p.ed=function(t){return Nyt(this.c,this.b,t)},p.Kc=function(t){return fLn(this.c,this.b,t)},p.fd=function(t,i){return f_t(this.c,this.b,t,i)},p.Wb=function(t){KZ(this.c,this.b),M7(this,f(t,15))},p.gc=function(){return sCn(this.c,this.b)},p.Nc=function(){return h2n(this.c,this.b)},p.Oc=function(t){return dSn(this.c,this.b,t)},p.Ib=function(){var t,i;for(i=new p0,i.a+="[",t=_st(this.c,this.b);rle(t);)ys(i,$7(AZ(t))),rle(t)&&(i.a+=Fu);return i.a+="]",i.a},p.Dk=function(){KZ(this.c,this.b)},R(Oi,"FeatureMapUtil/FeatureEList",493),N(629,38,dq,nue),p.ej=function(t){return Tj(this,t)},p.jj=function(t){var i,a,u,l,g,m,v;switch(this.d){case 1:case 2:{if(g=t.gj(),Ae(g)===Ae(this.c)&&Tj(this,null)==t.ej(null))return this.g=t.fj(),t.dj()==1&&(this.d=1),!0;break}case 3:{switch(l=t.dj(),l){case 3:{if(g=t.gj(),Ae(g)===Ae(this.c)&&Tj(this,null)==t.ej(null))return this.d=5,i=new OS(2),Qn(i,this.g),Qn(i,t.fj()),this.g=i,!0;break}}break}case 5:{switch(l=t.dj(),l){case 3:{if(g=t.gj(),Ae(g)===Ae(this.c)&&Tj(this,null)==t.ej(null))return a=f(this.g,16),a.Ec(t.fj()),!0;break}}break}case 4:{switch(l=t.dj(),l){case 3:{if(g=t.gj(),Ae(g)===Ae(this.c)&&Tj(this,null)==t.ej(null))return this.d=1,this.g=t.fj(),!0;break}case 4:{if(g=t.gj(),Ae(g)===Ae(this.c)&&Tj(this,null)==t.ej(null))return this.d=6,v=new OS(2),Qn(v,this.n),Qn(v,t.hj()),this.n=v,m=ce(se(pr,1),Pr,30,15,[this.o,t.ij()]),this.g=m,!0;break}}break}case 6:{switch(l=t.dj(),l){case 4:{if(g=t.gj(),Ae(g)===Ae(this.c)&&Tj(this,null)==t.ej(null))return a=f(this.n,16),a.Ec(t.hj()),m=f(this.g,54),u=Pe(pr,Pr,30,m.length+1,15,1),au(m,0,u,0,m.length),u[m.length]=t.ij(),this.g=u,!0;break}}break}}return!1},R(Oi,"FeatureMapUtil/FeatureENotificationImpl",629),N(549,493,{20:1,31:1,55:1,16:1,15:1,61:1,76:1,159:1,218:1,1982:1,71:1,96:1},TX),p.Ll=function(t,i){return UIe(this.c,t,i)},p.Ml=function(t,i,a){return PIe(this.c,t,i,a)},p.Nl=function(t,i,a){return sOe(this.c,t,i,a)},p.Ol=function(){return this},p.Pl=function(t,i){return OG(this.c,t,i)},p.Ql=function(t){return f(fQ(this.c,this.b,t,!1),74).Ik()},p.Rl=function(t){return f(fQ(this.c,this.b,t,!1),74).kd()},p.Sl=function(){return this.a},p.Tl=function(t){return!oG(this.c,t)},p.Ul=function(t,i){hQ(this.c,t,i)},p.Vl=function(t){return wbt(this.c,t)},p.Wl=function(t){Qvt(this.c,t)},R(Oi,"FeatureMapUtil/FeatureFeatureMap",549),N(1247,1,Epe,Jat),p.Ck=function(t){return fQ(this.b,this.a,-1,t)},p.Nj=function(){return!oG(this.b,this.a)},p.Wb=function(t){hQ(this.b,this.a,t)},p.Dk=function(){KZ(this.b,this.a)},R(Oi,"FeatureMapUtil/FeatureValue",1247);var d4,E1e,S1e,f4,EMt,xV=Bi(_ee,"AnyType");N(665,63,C1,Tae),R(_ee,"InvalidDatatypeValueException",665);var qne=Bi(_ee,eAt),CV=Bi(_ee,tAt),eje=Bi(_ee,nAt),SMt,Mc,tje,VE,_Mt,kMt,TMt,xMt,CMt,AMt,IMt,OMt,NMt,RMt,PMt,KA,LMt,YA,qF,DMt,T_,AV,IV,MMt,VF,WF;N(823,500,{109:1,93:1,92:1,57:1,52:1,98:1,836:1},E2e),p.Hh=function(t,i,a){switch(t){case 0:return a?(!this.c&&(this.c=new eo(this,0)),this.c):(!this.c&&(this.c=new eo(this,0)),this.c.b);case 1:return a?(!this.c&&(this.c=new eo(this,0)),f(yu(this.c,(fi(),VE)),159)):(!this.c&&(this.c=new eo(this,0)),f(f(yu(this.c,(fi(),VE)),159),218)).Sl();case 2:return a?(!this.b&&(this.b=new eo(this,2)),this.b):(!this.b&&(this.b=new eo(this,2)),this.b.b)}return Lf(this,t-jn(this.ei()),Vt((this.j&2)==0?this.ei():(!this.k&&(this.k=new ll),this.k).Kk(),t),i,a)},p.Qh=function(t,i,a){var u;switch(i){case 0:return!this.c&&(this.c=new eo(this,0)),CG(this.c,t,a);case 1:return(!this.c&&(this.c=new eo(this,0)),f(f(yu(this.c,(fi(),VE)),159),71)).Uk(t,a);case 2:return!this.b&&(this.b=new eo(this,2)),CG(this.b,t,a)}return u=f(Vt((this.j&2)==0?this.ei():(!this.k&&(this.k=new ll),this.k).Kk(),i),69),u.tk().xk(this,mxe(this),i-jn(this.ei()),t,a)},p.Sh=function(t){switch(t){case 0:return!!this.c&&this.c.i!=0;case 1:return!(!this.c&&(this.c=new eo(this,0)),f(yu(this.c,(fi(),VE)),159)).dc();case 2:return!!this.b&&this.b.i!=0}return Pf(this,t-jn(this.ei()),Vt((this.j&2)==0?this.ei():(!this.k&&(this.k=new ll),this.k).Kk(),t))},p.Zh=function(t,i){switch(t){case 0:!this.c&&(this.c=new eo(this,0)),hH(this.c,i);return;case 1:(!this.c&&(this.c=new eo(this,0)),f(f(yu(this.c,(fi(),VE)),159),218)).Wb(i);return;case 2:!this.b&&(this.b=new eo(this,2)),hH(this.b,i);return}$f(this,t-jn(this.ei()),Vt((this.j&2)==0?this.ei():(!this.k&&(this.k=new ll),this.k).Kk(),t),i)},p.ei=function(){return fi(),tje},p.gi=function(t){switch(t){case 0:!this.c&&(this.c=new eo(this,0)),Jn(this.c);return;case 1:(!this.c&&(this.c=new eo(this,0)),f(yu(this.c,(fi(),VE)),159)).$b();return;case 2:!this.b&&(this.b=new eo(this,2)),Jn(this.b);return}jf(this,t-jn(this.ei()),Vt((this.j&2)==0?this.ei():(!this.k&&(this.k=new ll),this.k).Kk(),t))},p.Ib=function(){var t;return(this.j&4)!=0?Jh(this):(t=new sh(Jh(this)),t.a+=" (mixed: ",R7(t,this.c),t.a+=", anyAttribute: ",R7(t,this.b),t.a+=")",t.a)},R(xo,"AnyTypeImpl",823),N(666,500,{109:1,93:1,92:1,57:1,52:1,98:1,2065:1,666:1},k5),p.Hh=function(t,i,a){switch(t){case 0:return this.a;case 1:return this.b}return Lf(this,t-jn((fi(),KA)),Vt((this.j&2)==0?KA:(!this.k&&(this.k=new ll),this.k).Kk(),t),i,a)},p.Sh=function(t){switch(t){case 0:return this.a!=null;case 1:return this.b!=null}return Pf(this,t-jn((fi(),KA)),Vt((this.j&2)==0?KA:(!this.k&&(this.k=new ll),this.k).Kk(),t))},p.Zh=function(t,i){switch(t){case 0:Oo(this,dr(i));return;case 1:zt(this,dr(i));return}$f(this,t-jn((fi(),KA)),Vt((this.j&2)==0?KA:(!this.k&&(this.k=new ll),this.k).Kk(),t),i)},p.ei=function(){return fi(),KA},p.gi=function(t){switch(t){case 0:this.a=null;return;case 1:this.b=null;return}jf(this,t-jn((fi(),KA)),Vt((this.j&2)==0?KA:(!this.k&&(this.k=new ll),this.k).Kk(),t))},p.Ib=function(){var t;return(this.j&4)!=0?Jh(this):(t=new sh(Jh(this)),t.a+=" (data: ",ys(t,this.a),t.a+=", target: ",ys(t,this.b),t.a+=")",t.a)},p.a=null,p.b=null,R(xo,"ProcessingInstructionImpl",666),N(667,823,{109:1,93:1,92:1,57:1,52:1,98:1,836:1,2066:1,667:1},Nrt),p.Hh=function(t,i,a){switch(t){case 0:return a?(!this.c&&(this.c=new eo(this,0)),this.c):(!this.c&&(this.c=new eo(this,0)),this.c.b);case 1:return a?(!this.c&&(this.c=new eo(this,0)),f(yu(this.c,(fi(),VE)),159)):(!this.c&&(this.c=new eo(this,0)),f(f(yu(this.c,(fi(),VE)),159),218)).Sl();case 2:return a?(!this.b&&(this.b=new eo(this,2)),this.b):(!this.b&&(this.b=new eo(this,2)),this.b.b);case 3:return!this.c&&(this.c=new eo(this,0)),dr(OG(this.c,(fi(),qF),!0));case 4:return B_e(this.a,(!this.c&&(this.c=new eo(this,0)),dr(OG(this.c,(fi(),qF),!0))));case 5:return this.a}return Lf(this,t-jn((fi(),YA)),Vt((this.j&2)==0?YA:(!this.k&&(this.k=new ll),this.k).Kk(),t),i,a)},p.Sh=function(t){switch(t){case 0:return!!this.c&&this.c.i!=0;case 1:return!(!this.c&&(this.c=new eo(this,0)),f(yu(this.c,(fi(),VE)),159)).dc();case 2:return!!this.b&&this.b.i!=0;case 3:return!this.c&&(this.c=new eo(this,0)),dr(OG(this.c,(fi(),qF),!0))!=null;case 4:return B_e(this.a,(!this.c&&(this.c=new eo(this,0)),dr(OG(this.c,(fi(),qF),!0))))!=null;case 5:return!!this.a}return Pf(this,t-jn((fi(),YA)),Vt((this.j&2)==0?YA:(!this.k&&(this.k=new ll),this.k).Kk(),t))},p.Zh=function(t,i){switch(t){case 0:!this.c&&(this.c=new eo(this,0)),hH(this.c,i);return;case 1:(!this.c&&(this.c=new eo(this,0)),f(f(yu(this.c,(fi(),VE)),159),218)).Wb(i);return;case 2:!this.b&&(this.b=new eo(this,2)),hH(this.b,i);return;case 3:kTe(this,dr(i));return;case 4:kTe(this,F_e(this.a,i));return;case 5:mu(this,f(i,155));return}$f(this,t-jn((fi(),YA)),Vt((this.j&2)==0?YA:(!this.k&&(this.k=new ll),this.k).Kk(),t),i)},p.ei=function(){return fi(),YA},p.gi=function(t){switch(t){case 0:!this.c&&(this.c=new eo(this,0)),Jn(this.c);return;case 1:(!this.c&&(this.c=new eo(this,0)),f(yu(this.c,(fi(),VE)),159)).$b();return;case 2:!this.b&&(this.b=new eo(this,2)),Jn(this.b);return;case 3:!this.c&&(this.c=new eo(this,0)),hQ(this.c,(fi(),qF),null);return;case 4:kTe(this,F_e(this.a,null));return;case 5:this.a=null;return}jf(this,t-jn((fi(),YA)),Vt((this.j&2)==0?YA:(!this.k&&(this.k=new ll),this.k).Kk(),t))},R(xo,"SimpleAnyTypeImpl",667),N(668,500,{109:1,93:1,92:1,57:1,52:1,98:1,2067:1,668:1},Rrt),p.Hh=function(t,i,a){switch(t){case 0:return a?(!this.a&&(this.a=new eo(this,0)),this.a):(!this.a&&(this.a=new eo(this,0)),this.a.b);case 1:return a?(!this.b&&(this.b=new Ol((Ut(),Za),xc,this,1)),this.b):(!this.b&&(this.b=new Ol((Ut(),Za),xc,this,1)),AH(this.b));case 2:return a?(!this.c&&(this.c=new Ol((Ut(),Za),xc,this,2)),this.c):(!this.c&&(this.c=new Ol((Ut(),Za),xc,this,2)),AH(this.c));case 3:return!this.a&&(this.a=new eo(this,0)),yu(this.a,(fi(),AV));case 4:return!this.a&&(this.a=new eo(this,0)),yu(this.a,(fi(),IV));case 5:return!this.a&&(this.a=new eo(this,0)),yu(this.a,(fi(),VF));case 6:return!this.a&&(this.a=new eo(this,0)),yu(this.a,(fi(),WF))}return Lf(this,t-jn((fi(),T_)),Vt((this.j&2)==0?T_:(!this.k&&(this.k=new ll),this.k).Kk(),t),i,a)},p.Qh=function(t,i,a){var u;switch(i){case 0:return!this.a&&(this.a=new eo(this,0)),CG(this.a,t,a);case 1:return!this.b&&(this.b=new Ol((Ut(),Za),xc,this,1)),pX(this.b,t,a);case 2:return!this.c&&(this.c=new Ol((Ut(),Za),xc,this,2)),pX(this.c,t,a);case 5:return!this.a&&(this.a=new eo(this,0)),Wct(yu(this.a,(fi(),VF)),t,a)}return u=f(Vt((this.j&2)==0?(fi(),T_):(!this.k&&(this.k=new ll),this.k).Kk(),i),69),u.tk().xk(this,mxe(this),i-jn((fi(),T_)),t,a)},p.Sh=function(t){switch(t){case 0:return!!this.a&&this.a.i!=0;case 1:return!!this.b&&this.b.f!=0;case 2:return!!this.c&&this.c.f!=0;case 3:return!this.a&&(this.a=new eo(this,0)),!XY(yu(this.a,(fi(),AV)));case 4:return!this.a&&(this.a=new eo(this,0)),!XY(yu(this.a,(fi(),IV)));case 5:return!this.a&&(this.a=new eo(this,0)),!XY(yu(this.a,(fi(),VF)));case 6:return!this.a&&(this.a=new eo(this,0)),!XY(yu(this.a,(fi(),WF)))}return Pf(this,t-jn((fi(),T_)),Vt((this.j&2)==0?T_:(!this.k&&(this.k=new ll),this.k).Kk(),t))},p.Zh=function(t,i){switch(t){case 0:!this.a&&(this.a=new eo(this,0)),hH(this.a,i);return;case 1:!this.b&&(this.b=new Ol((Ut(),Za),xc,this,1)),XJ(this.b,i);return;case 2:!this.c&&(this.c=new Ol((Ut(),Za),xc,this,2)),XJ(this.c,i);return;case 3:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),AV))),!this.a&&(this.a=new eo(this,0)),M7(yu(this.a,AV),f(i,16));return;case 4:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),IV))),!this.a&&(this.a=new eo(this,0)),M7(yu(this.a,IV),f(i,16));return;case 5:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),VF))),!this.a&&(this.a=new eo(this,0)),M7(yu(this.a,VF),f(i,16));return;case 6:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),WF))),!this.a&&(this.a=new eo(this,0)),M7(yu(this.a,WF),f(i,16));return}$f(this,t-jn((fi(),T_)),Vt((this.j&2)==0?T_:(!this.k&&(this.k=new ll),this.k).Kk(),t),i)},p.ei=function(){return fi(),T_},p.gi=function(t){switch(t){case 0:!this.a&&(this.a=new eo(this,0)),Jn(this.a);return;case 1:!this.b&&(this.b=new Ol((Ut(),Za),xc,this,1)),this.b.c.$b();return;case 2:!this.c&&(this.c=new Ol((Ut(),Za),xc,this,2)),this.c.c.$b();return;case 3:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),AV)));return;case 4:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),IV)));return;case 5:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),VF)));return;case 6:!this.a&&(this.a=new eo(this,0)),TN(yu(this.a,(fi(),WF)));return}jf(this,t-jn((fi(),T_)),Vt((this.j&2)==0?T_:(!this.k&&(this.k=new ll),this.k).Kk(),t))},p.Ib=function(){var t;return(this.j&4)!=0?Jh(this):(t=new sh(Jh(this)),t.a+=" (mixed: ",R7(t,this.a),t.a+=")",t.a)},R(xo,"XMLTypeDocumentRootImpl",668),N(1974,704,{109:1,93:1,92:1,466:1,154:1,57:1,113:1,52:1,98:1,157:1,117:1,118:1,2068:1},GU),p.ni=function(t,i){switch(t.ek()){case 7:case 8:case 9:case 10:case 16:case 22:case 23:case 24:case 25:case 26:case 32:case 33:case 34:case 36:case 37:case 44:case 45:case 50:case 51:case 53:case 55:case 56:case 57:case 58:case 60:case 61:case 4:return i==null?null:Qs(i);case 19:case 28:case 29:case 35:case 38:case 39:case 41:case 46:case 52:case 54:case 5:return dr(i);case 6:return wbn(f(i,195));case 12:case 47:case 49:case 11:return J_t(this,t,i);case 13:return i==null?null:sMn(f(i,245));case 15:case 14:return i==null?null:mwn(ke(Oe(i)));case 17:return zwt((fi(),i));case 18:return zwt(i);case 21:case 20:return i==null?null:bwn(f(i,160).a);case 27:return ybn(f(i,195));case 30:return ewt((fi(),f(i,15)));case 31:return ewt(f(i,15));case 40:return vbn((fi(),i));case 42:return Hwt((fi(),i));case 43:return Hwt(i);case 59:case 48:return bbn((fi(),i));default:throw ie(new dn(VL+t.ve()+n_))}},p.oi=function(t){var i,a,u,l,g;switch(t.G==-1&&(t.G=(a=cf(t),a?P0(a.ri(),t):-1)),t.G){case 0:return i=new E2e,i;case 1:return u=new k5,u;case 2:return l=new Nrt,l;case 3:return g=new Rrt,g;default:throw ie(new dn(ope+t.zb+n_))}},p.pi=function(t,i){var a,u,l,g,m,v,S,T,A,L,F,z,W,J,ue,pe;switch(t.ek()){case 5:case 52:case 4:return i;case 6:return j3n(i);case 8:case 7:return i==null?null:wOn(i);case 9:return i==null?null:jH(ff((u=_u(i,!0),u.length>0&&(gn(0,u.length),u.charCodeAt(0)==43)?(gn(1,u.length+1),u.substr(1)):u),-128,127)<<24>>24);case 10:return i==null?null:jH(ff((l=_u(i,!0),l.length>0&&(gn(0,l.length),l.charCodeAt(0)==43)?(gn(1,l.length+1),l.substr(1)):l),-128,127)<<24>>24);case 11:return dr(KS(this,(fi(),TMt),i));case 12:return dr(KS(this,(fi(),xMt),i));case 13:return i==null?null:new G2e(_u(i,!0));case 15:case 14:return I6n(i);case 16:return dr(KS(this,(fi(),CMt),i));case 17:return ovt((fi(),i));case 18:return ovt(i);case 28:case 29:case 35:case 38:case 39:case 41:case 54:case 19:return _u(i,!0);case 21:case 20:return F6n(i);case 22:return dr(KS(this,(fi(),AMt),i));case 23:return dr(KS(this,(fi(),IMt),i));case 24:return dr(KS(this,(fi(),OMt),i));case 25:return dr(KS(this,(fi(),NMt),i));case 26:return dr(KS(this,(fi(),RMt),i));case 27:return O3n(i);case 30:return avt((fi(),i));case 31:return avt(i);case 32:return i==null?null:rt(ff((A=_u(i,!0),A.length>0&&(gn(0,A.length),A.charCodeAt(0)==43)?(gn(1,A.length+1),A.substr(1)):A),Ra,Qr));case 33:return i==null?null:new Gv((L=_u(i,!0),L.length>0&&(gn(0,L.length),L.charCodeAt(0)==43)?(gn(1,L.length+1),L.substr(1)):L));case 34:return i==null?null:rt(ff((F=_u(i,!0),F.length>0&&(gn(0,F.length),F.charCodeAt(0)==43)?(gn(1,F.length+1),F.substr(1)):F),Ra,Qr));case 36:return i==null?null:TT(bQ((z=_u(i,!0),z.length>0&&(gn(0,z.length),z.charCodeAt(0)==43)?(gn(1,z.length+1),z.substr(1)):z)));case 37:return i==null?null:TT(bQ((W=_u(i,!0),W.length>0&&(gn(0,W.length),W.charCodeAt(0)==43)?(gn(1,W.length+1),W.substr(1)):W)));case 40:return kAn((fi(),i));case 42:return svt((fi(),i));case 43:return svt(i);case 44:return i==null?null:new Gv((J=_u(i,!0),J.length>0&&(gn(0,J.length),J.charCodeAt(0)==43)?(gn(1,J.length+1),J.substr(1)):J));case 45:return i==null?null:new Gv((ue=_u(i,!0),ue.length>0&&(gn(0,ue.length),ue.charCodeAt(0)==43)?(gn(1,ue.length+1),ue.substr(1)):ue));case 46:return _u(i,!1);case 47:return dr(KS(this,(fi(),PMt),i));case 59:case 48:return _An((fi(),i));case 49:return dr(KS(this,(fi(),LMt),i));case 50:return i==null?null:ZP(ff((pe=_u(i,!0),pe.length>0&&(gn(0,pe.length),pe.charCodeAt(0)==43)?(gn(1,pe.length+1),pe.substr(1)):pe),wee,32767)<<16>>16);case 51:return i==null?null:ZP(ff((g=_u(i,!0),g.length>0&&(gn(0,g.length),g.charCodeAt(0)==43)?(gn(1,g.length+1),g.substr(1)):g),wee,32767)<<16>>16);case 53:return dr(KS(this,(fi(),DMt),i));case 55:return i==null?null:ZP(ff((m=_u(i,!0),m.length>0&&(gn(0,m.length),m.charCodeAt(0)==43)?(gn(1,m.length+1),m.substr(1)):m),wee,32767)<<16>>16);case 56:return i==null?null:ZP(ff((v=_u(i,!0),v.length>0&&(gn(0,v.length),v.charCodeAt(0)==43)?(gn(1,v.length+1),v.substr(1)):v),wee,32767)<<16>>16);case 57:return i==null?null:TT(bQ((S=_u(i,!0),S.length>0&&(gn(0,S.length),S.charCodeAt(0)==43)?(gn(1,S.length+1),S.substr(1)):S)));case 58:return i==null?null:TT(bQ((T=_u(i,!0),T.length>0&&(gn(0,T.length),T.charCodeAt(0)==43)?(gn(1,T.length+1),T.substr(1)):T)));case 60:return i==null?null:rt(ff((a=_u(i,!0),a.length>0&&(gn(0,a.length),a.charCodeAt(0)==43)?(gn(1,a.length+1),a.substr(1)):a),Ra,Qr));case 61:return i==null?null:rt(ff(_u(i,!0),Ra,Qr));default:throw ie(new dn(VL+t.ve()+n_))}};var jMt,nje,$Mt,rje;R(xo,"XMLTypeFactoryImpl",1974),N(577,184,{109:1,93:1,92:1,154:1,197:1,57:1,240:1,113:1,52:1,98:1,157:1,184:1,117:1,118:1,675:1,1990:1,577:1},Odt),p.N=!1,p.O=!1;var FMt=!1;R(xo,"XMLTypePackageImpl",577),N(1907,1,{830:1},My),p.Hk=function(){return JIe(),KMt},R(xo,"XMLTypePackageImpl/1",1907),N(1916,1,jr,eN),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/10",1916),N(1917,1,jr,Nk),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/11",1917),N(1918,1,jr,g9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/12",1918),N(1919,1,jr,m9),p.ck=function(t){return Jk(t)},p.dk=function(t){return Pe(mo,Ye,341,t,7,1)},R(xo,"XMLTypePackageImpl/13",1919),N(1920,1,jr,b9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/14",1920),N(1921,1,jr,T5),p.ck=function(t){return me(t,15)},p.dk=function(t){return Pe(qf,BT,15,t,0,1)},R(xo,"XMLTypePackageImpl/15",1921),N(1922,1,jr,x5),p.ck=function(t){return me(t,15)},p.dk=function(t){return Pe(qf,BT,15,t,0,1)},R(xo,"XMLTypePackageImpl/16",1922),N(1923,1,jr,qU),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/17",1923),N(1924,1,jr,VU),p.ck=function(t){return me(t,160)},p.dk=function(t){return Pe(eD,Ye,160,t,0,1)},R(xo,"XMLTypePackageImpl/18",1924),N(1925,1,jr,v9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/19",1925),N(1908,1,jr,C5),p.ck=function(t){return me(t,836)},p.dk=function(t){return Pe(xV,Jt,836,t,0,1)},R(xo,"XMLTypePackageImpl/2",1908),N(1926,1,jr,w9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/20",1926),N(1927,1,jr,Rk),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/21",1927),N(1928,1,jr,tN),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/22",1928),N(1929,1,jr,y9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/23",1929),N(1930,1,jr,t3),p.ck=function(t){return me(t,195)},p.dk=function(t){return Pe($l,Ye,195,t,0,2)},R(xo,"XMLTypePackageImpl/24",1930),N(1931,1,jr,E9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/25",1931),N(1932,1,jr,S9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/26",1932),N(1933,1,jr,_9),p.ck=function(t){return me(t,15)},p.dk=function(t){return Pe(qf,BT,15,t,0,1)},R(xo,"XMLTypePackageImpl/27",1933),N(1934,1,jr,k9),p.ck=function(t){return me(t,15)},p.dk=function(t){return Pe(qf,BT,15,t,0,1)},R(xo,"XMLTypePackageImpl/28",1934),N(1935,1,jr,nN),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/29",1935),N(1909,1,jr,T9),p.ck=function(t){return me(t,666)},p.dk=function(t){return Pe(qne,Jt,2065,t,0,1)},R(xo,"XMLTypePackageImpl/3",1909),N(1936,1,jr,x9),p.ck=function(t){return me(t,17)},p.dk=function(t){return Pe(ba,Ye,17,t,0,1)},R(xo,"XMLTypePackageImpl/30",1936),N(1937,1,jr,gu),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/31",1937),N(1938,1,jr,A5),p.ck=function(t){return me(t,190)},p.dk=function(t){return Pe(i_,Ye,190,t,0,1)},R(xo,"XMLTypePackageImpl/32",1938),N(1939,1,jr,WU),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/33",1939),N(1940,1,jr,C9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/34",1940),N(1941,1,jr,rN),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/35",1941),N(1942,1,jr,KU),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/36",1942),N(1943,1,jr,A9),p.ck=function(t){return me(t,15)},p.dk=function(t){return Pe(qf,BT,15,t,0,1)},R(xo,"XMLTypePackageImpl/37",1943),N(1944,1,jr,n3),p.ck=function(t){return me(t,15)},p.dk=function(t){return Pe(qf,BT,15,t,0,1)},R(xo,"XMLTypePackageImpl/38",1944),N(1945,1,jr,jy),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/39",1945),N(1910,1,jr,iN),p.ck=function(t){return me(t,667)},p.dk=function(t){return Pe(CV,Jt,2066,t,0,1)},R(xo,"XMLTypePackageImpl/4",1910),N(1946,1,jr,r3),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/40",1946),N(1947,1,jr,oN),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/41",1947),N(1948,1,jr,$y),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/42",1948),N(1949,1,jr,l0),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/43",1949),N(1950,1,jr,i3),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/44",1950),N(1951,1,jr,YU),p.ck=function(t){return me(t,191)},p.dk=function(t){return Pe(o_,Ye,191,t,0,1)},R(xo,"XMLTypePackageImpl/45",1951),N(1952,1,jr,o3),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/46",1952),N(1953,1,jr,a3),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/47",1953),N(1954,1,jr,I9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/48",1954),N(1955,1,jr,O9),p.ck=function(t){return me(t,191)},p.dk=function(t){return Pe(o_,Ye,191,t,0,1)},R(xo,"XMLTypePackageImpl/49",1955),N(1911,1,jr,I5),p.ck=function(t){return me(t,668)},p.dk=function(t){return Pe(eje,Jt,2067,t,0,1)},R(xo,"XMLTypePackageImpl/5",1911),N(1956,1,jr,aN),p.ck=function(t){return me(t,190)},p.dk=function(t){return Pe(i_,Ye,190,t,0,1)},R(xo,"XMLTypePackageImpl/50",1956),N(1957,1,jr,N9),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/51",1957),N(1958,1,jr,R9),p.ck=function(t){return me(t,17)},p.dk=function(t){return Pe(ba,Ye,17,t,0,1)},R(xo,"XMLTypePackageImpl/52",1958),N(1912,1,jr,XU),p.ck=function(t){return ta(t)},p.dk=function(t){return Pe(ft,Ye,2,t,6,1)},R(xo,"XMLTypePackageImpl/6",1912),N(1913,1,jr,O5),p.ck=function(t){return me(t,195)},p.dk=function(t){return Pe($l,Ye,195,t,0,2)},R(xo,"XMLTypePackageImpl/7",1913),N(1914,1,jr,N5),p.ck=function(t){return Xk(t)},p.dk=function(t){return Pe(to,Ye,471,t,8,1)},R(xo,"XMLTypePackageImpl/8",1914),N(1915,1,jr,JU),p.ck=function(t){return me(t,219)},p.dk=function(t){return Pe(BR,Ye,219,t,0,1)},R(xo,"XMLTypePackageImpl/9",1915);var zg,Q0,KF,Vne,le;N(53,63,C1,vr),R(B0,"RegEx/ParseException",53),N(815,1,{},P9),p.$l=function(t){return t<this.j&&_a(this.i,t)==63},p._l=function(){var t,i,a,u,l;if(this.c!=10)throw ie(new vr(yr((lr(),dee))));switch(t=this.a,t){case 101:t=27;break;case 102:t=12;break;case 110:t=10;break;case 114:t=13;break;case 116:t=9;break;case 120:if(Yr(this),this.c!=0)throw ie(new vr(yr((lr(),F0))));if(this.a==123){l=0,a=0;do{if(Yr(this),this.c!=0)throw ie(new vr(yr((lr(),F0))));if((l=gE(this.a))<0)break;if(a>a*16)throw ie(new vr(yr((lr(),h3t))));a=a*16+l}while(!0);if(this.a!=125)throw ie(new vr(yr((lr(),p3t))));if(a>JL)throw ie(new vr(yr((lr(),g3t))));t=a}else{if(l=0,this.c!=0||(l=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(a=l,Yr(this),this.c!=0||(l=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));a=a*16+l,t=a}break;case 117:if(u=0,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=i*16+u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=i*16+u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));i=i*16+u,t=i;break;case 118:if(Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=i*16+u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=i*16+u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=i*16+u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=i*16+u,Yr(this),this.c!=0||(u=gE(this.a))<0)throw ie(new vr(yr((lr(),F0))));if(i=i*16+u,i>JL)throw ie(new vr(yr((lr(),"parser.descappe.4"))));t=i;break;case 65:case 90:case 122:throw ie(new vr(yr((lr(),m3t))))}return t},p.am=function(t){var i,a;switch(t){case 100:a=(this.e&32)==32?dw("Nd",!0):(Xr(),Wne);break;case 68:a=(this.e&32)==32?dw("Nd",!1):(Xr(),uje);break;case 119:a=(this.e&32)==32?dw("IsWord",!0):(Xr(),UD);break;case 87:a=(this.e&32)==32?dw("IsWord",!1):(Xr(),dje);break;case 115:a=(this.e&32)==32?dw("IsSpace",!0):(Xr(),h4);break;case 83:a=(this.e&32)==32?dw("IsSpace",!1):(Xr(),lje);break;default:throw ie(new oc((i=t,mAt+i.toString(16))))}return a},p.bm=function(t){var i,a,u,l,g,m,v,S,T,A,L,F;for(this.b=1,Yr(this),i=null,this.c==0&&this.a==94?(Yr(this),t?A=(Xr(),Xr(),new of(5)):(i=(Xr(),Xr(),new of(4)),Su(i,0,JL),A=new of(4))):A=(Xr(),Xr(),new of(4)),l=!0;(F=this.c)!=1&&!(F==0&&this.a==93&&!l);){if(l=!1,a=this.a,u=!1,F==10)switch(a){case 100:case 68:case 119:case 87:case 115:case 83:MT(A,this.am(a)),u=!0;break;case 105:case 73:case 99:case 67:a=this.rm(A,a),a<0&&(u=!0);break;case 112:case 80:if(L=qAe(this,a),!L)throw ie(new vr(yr((lr(),vpe))));MT(A,L),u=!0;break;default:a=this._l()}else if(F==20){if(m=fP(this.i,58,this.d),m<0)throw ie(new vr(yr((lr(),q4e))));if(v=!0,_a(this.i,this.d)==94&&(++this.d,v=!1),g=uh(this.i,this.d,m),S=Agt(g,v,(this.e&512)==512),!S)throw ie(new vr(yr((lr(),c3t))));if(MT(A,S),u=!0,m+1>=this.j||_a(this.i,m+1)!=93)throw ie(new vr(yr((lr(),q4e))));this.d=m+2}if(Yr(this),!u)if(this.c!=0||this.a!=45)Su(A,a,a);else{if(Yr(this),(F=this.c)==1)throw ie(new vr(yr((lr(),fee))));F==0&&this.a==93?(Su(A,a,a),Su(A,45,45)):(T=this.a,F==10&&(T=this._l()),Yr(this),Su(A,a,T))}(this.e&ep)==ep&&this.c==0&&this.a==44&&Yr(this)}if(this.c==1)throw ie(new vr(yr((lr(),fee))));return i&&(t$(i,A),A=i),eA(A),Qj(A),this.b=0,Yr(this),A},p.cm=function(){var t,i,a,u;for(a=this.bm(!1);(u=this.c)!=7;)if(t=this.a,u==0&&(t==45||t==38)||u==4){if(Yr(this),this.c!=9)throw ie(new vr(yr((lr(),l3t))));if(i=this.bm(!1),u==4)MT(a,i);else if(t==45)t$(a,i);else if(t==38)W_t(a,i);else throw ie(new oc("ASSERT"))}else throw ie(new vr(yr((lr(),d3t))));return Yr(this),a},p.dm=function(){var t,i;return t=this.a-48,i=(Xr(),Xr(),new Bce(12,null,t)),!this.g&&(this.g=new uY),cY(this.g,new i2e(t)),Yr(this),i},p.em=function(){return Yr(this),Xr(),zMt},p.fm=function(){return Yr(this),Xr(),UMt},p.gm=function(){throw ie(new vr(yr((lr(),Hf))))},p.hm=function(){throw ie(new vr(yr((lr(),Hf))))},p.im=function(){return Yr(this),YTn()},p.jm=function(){return Yr(this),Xr(),GMt},p.km=function(){return Yr(this),Xr(),VMt},p.lm=function(){var t;if(this.d>=this.j||((t=_a(this.i,this.d++))&65504)!=64)throw ie(new vr(yr((lr(),o3t))));return Yr(this),Xr(),Xr(),new fb(0,t-64)},p.mm=function(){return Yr(this),gDn()},p.nm=function(){return Yr(this),Xr(),WMt},p.om=function(){var t;return t=(Xr(),Xr(),new fb(0,105)),Yr(this),t},p.pm=function(){return Yr(this),Xr(),qMt},p.qm=function(){return Yr(this),Xr(),HMt},p.rm=function(t,i){return this._l()},p.sm=function(){return Yr(this),Xr(),sje},p.tm=function(){var t,i,a,u,l;if(this.d+1>=this.j)throw ie(new vr(yr((lr(),n3t))));if(u=-1,i=null,t=_a(this.i,this.d),49<=t&&t<=57){if(u=t-48,!this.g&&(this.g=new uY),cY(this.g,new i2e(u)),++this.d,_a(this.i,this.d)!=41)throw ie(new vr(yr((lr(),xE))));++this.d}else switch(t==63&&--this.d,Yr(this),i=_Oe(this),i.e){case 20:case 21:case 22:case 23:break;case 8:if(this.c!=7)throw ie(new vr(yr((lr(),xE))));break;default:throw ie(new vr(yr((lr(),r3t))))}if(Yr(this),l=FS(this),a=null,l.e==2){if(l.Mm()!=2)throw ie(new vr(yr((lr(),i3t))));a=l.Im(1),l=l.Im(0)}if(this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),Xr(),Xr(),new hmt(u,i,l,a)},p.um=function(){return Yr(this),Xr(),cje},p.vm=function(){var t;if(Yr(this),t=BX(24,FS(this)),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),t},p.wm=function(){var t;if(Yr(this),t=BX(20,FS(this)),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),t},p.xm=function(){var t;if(Yr(this),t=BX(22,FS(this)),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),t},p.ym=function(){var t,i,a,u,l;for(t=0,a=0,i=-1;this.d<this.j&&(i=_a(this.i,this.d),l=JAe(i),l!=0);)t|=l,++this.d;if(this.d>=this.j)throw ie(new vr(yr((lr(),H4e))));if(i==45){for(++this.d;this.d<this.j&&(i=_a(this.i,this.d),l=JAe(i),l!=0);)a|=l,++this.d;if(this.d>=this.j)throw ie(new vr(yr((lr(),H4e))))}if(i==58){if(++this.d,Yr(this),u=cdt(FS(this),t,a),this.c!=7)throw ie(new vr(yr((lr(),xE))));Yr(this)}else if(i==41)++this.d,Yr(this),u=cdt(FS(this),t,a);else throw ie(new vr(yr((lr(),t3t))));return u},p.zm=function(){var t;if(Yr(this),t=BX(21,FS(this)),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),t},p.Am=function(){var t;if(Yr(this),t=BX(23,FS(this)),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),t},p.Bm=function(){var t,i;if(Yr(this),t=this.f++,i=hce(FS(this),t),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),i},p.Cm=function(){var t;if(Yr(this),t=hce(FS(this),0),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),t},p.Dm=function(t){return Yr(this),this.c==5?(Yr(this),RX(t,(Xr(),Xr(),new hT(9,t)))):RX(t,(Xr(),Xr(),new hT(3,t)))},p.Em=function(t){var i;return Yr(this),i=(Xr(),Xr(),new D7(2)),this.c==5?(Yr(this),vE(i,XF),vE(i,t)):(vE(i,t),vE(i,XF)),i},p.Fm=function(t){return Yr(this),this.c==5?(Yr(this),Xr(),Xr(),new hT(9,t)):(Xr(),Xr(),new hT(3,t))},p.a=0,p.b=0,p.c=0,p.d=0,p.e=0,p.f=1,p.g=null,p.j=0,R(B0,"RegEx/RegexParser",815),N(1893,815,{},Prt),p.$l=function(t){return!1},p._l=function(){return IIe(this)},p.am=function(t){return yL(t)},p.bm=function(t){return Fkt(this)},p.cm=function(){throw ie(new vr(yr((lr(),Hf))))},p.dm=function(){throw ie(new vr(yr((lr(),Hf))))},p.em=function(){throw ie(new vr(yr((lr(),Hf))))},p.fm=function(){throw ie(new vr(yr((lr(),Hf))))},p.gm=function(){return Yr(this),yL(67)},p.hm=function(){return Yr(this),yL(73)},p.im=function(){throw ie(new vr(yr((lr(),Hf))))},p.jm=function(){throw ie(new vr(yr((lr(),Hf))))},p.km=function(){throw ie(new vr(yr((lr(),Hf))))},p.lm=function(){return Yr(this),yL(99)},p.mm=function(){throw ie(new vr(yr((lr(),Hf))))},p.nm=function(){throw ie(new vr(yr((lr(),Hf))))},p.om=function(){return Yr(this),yL(105)},p.pm=function(){throw ie(new vr(yr((lr(),Hf))))},p.qm=function(){throw ie(new vr(yr((lr(),Hf))))},p.rm=function(t,i){return MT(t,yL(i)),-1},p.sm=function(){return Yr(this),Xr(),Xr(),new fb(0,94)},p.tm=function(){throw ie(new vr(yr((lr(),Hf))))},p.um=function(){return Yr(this),Xr(),Xr(),new fb(0,36)},p.vm=function(){throw ie(new vr(yr((lr(),Hf))))},p.wm=function(){throw ie(new vr(yr((lr(),Hf))))},p.xm=function(){throw ie(new vr(yr((lr(),Hf))))},p.ym=function(){throw ie(new vr(yr((lr(),Hf))))},p.zm=function(){throw ie(new vr(yr((lr(),Hf))))},p.Am=function(){throw ie(new vr(yr((lr(),Hf))))},p.Bm=function(){var t;if(Yr(this),t=hce(FS(this),0),this.c!=7)throw ie(new vr(yr((lr(),xE))));return Yr(this),t},p.Cm=function(){throw ie(new vr(yr((lr(),Hf))))},p.Dm=function(t){return Yr(this),RX(t,(Xr(),Xr(),new hT(3,t)))},p.Em=function(t){var i;return Yr(this),i=(Xr(),Xr(),new D7(2)),vE(i,t),vE(i,XF),i},p.Fm=function(t){return Yr(this),Xr(),Xr(),new hT(3,t)};var XA=null,FD=null;R(B0,"RegEx/ParserForXMLSchema",1893),N(121,1,ZL,sS),p.Gm=function(t){throw ie(new oc("Not supported."))},p.Hm=function(){return-1},p.Im=function(t){return null},p.Jm=function(){return null},p.Km=function(t){},p.Lm=function(t){},p.Mm=function(){return 0},p.Ib=function(){return this.Nm(0)},p.Nm=function(t){return this.e==11?".":""},p.e=0;var ije,BD,YF,BMt,oje,_x=null,Wne,_1e=null,aje,XF,k1e=null,sje,cje,uje,lje,dje,UMt,h4,zMt,HMt,GMt,qMt,UD,VMt,WMt,wjn=R(B0,"RegEx/Token",121);N(136,121,{3:1,136:1,121:1},of),p.Nm=function(t){var i,a,u;if(this.e==4)if(this==aje)a=".";else if(this==Wne)a="\\d";else if(this==UD)a="\\w";else if(this==h4)a="\\s";else{for(u=new p0,u.a+="[",i=0;i<this.b.length;i+=2)(t&ep)!=0&&i>0&&(u.a+=","),this.b[i]===this.b[i+1]?ys(u,IG(this.b[i])):(ys(u,IG(this.b[i])),u.a+="-",ys(u,IG(this.b[i+1])));u.a+="]",a=u.a}else if(this==uje)a="\\D";else if(this==dje)a="\\W";else if(this==lje)a="\\S";else{for(u=new p0,u.a+="[^",i=0;i<this.b.length;i+=2)(t&ep)!=0&&i>0&&(u.a+=","),this.b[i]===this.b[i+1]?ys(u,IG(this.b[i])):(ys(u,IG(this.b[i])),u.a+="-",ys(u,IG(this.b[i+1])));u.a+="]",a=u.a}return a},p.a=!1,p.c=!1,R(B0,"RegEx/RangeToken",136),N(575,1,{575:1},i2e),p.a=0,R(B0,"RegEx/RegexParser/ReferencePosition",575),N(574,1,{3:1,574:1},vot),p.Fb=function(t){var i;return t==null||!me(t,574)?!1:(i=f(t,574),Mt(this.b,i.b)&&this.a==i.a)},p.Hb=function(){return C0(this.b+"/"+SIe(this.a))},p.Ib=function(){return this.c.Nm(this.a)},p.a=0,R(B0,"RegEx/RegularExpression",574),N(227,121,ZL,fb),p.Hm=function(){return this.a},p.Nm=function(t){var i,a,u;switch(this.e){case 0:switch(this.a){case 124:case 42:case 43:case 63:case 40:case 41:case 46:case 91:case 123:case 92:u="\\"+Bse(this.a&To);break;case 12:u="\\f";break;case 10:u="\\n";break;case 13:u="\\r";break;case 9:u="\\t";break;case 27:u="\\e";break;default:this.a>=Xa?(a=(i=this.a>>>0,"0"+i.toString(16)),u="\\v"+uh(a,a.length-6,a.length)):u=""+Bse(this.a&To)}break;case 8:this==sje||this==cje?u=""+Bse(this.a&To):u="\\"+Bse(this.a&To);break;default:u=null}return u},p.a=0,R(B0,"RegEx/Token/CharToken",227),N(319,121,ZL,hT),p.Im=function(t){return this.a},p.Km=function(t){this.b=t},p.Lm=function(t){this.c=t},p.Mm=function(){return 1},p.Nm=function(t){var i;if(this.e==3)if(this.c<0&&this.b<0)i=this.a.Nm(t)+"*";else if(this.c==this.b)i=this.a.Nm(t)+"{"+this.c+"}";else if(this.c>=0&&this.b>=0)i=this.a.Nm(t)+"{"+this.c+","+this.b+"}";else if(this.c>=0&&this.b<0)i=this.a.Nm(t)+"{"+this.c+",}";else throw ie(new oc("Token#toString(): CLOSURE "+this.c+Fu+this.b));else if(this.c<0&&this.b<0)i=this.a.Nm(t)+"*?";else if(this.c==this.b)i=this.a.Nm(t)+"{"+this.c+"}?";else if(this.c>=0&&this.b>=0)i=this.a.Nm(t)+"{"+this.c+","+this.b+"}?";else if(this.c>=0&&this.b<0)i=this.a.Nm(t)+"{"+this.c+",}?";else throw ie(new oc("Token#toString(): NONGREEDYCLOSURE "+this.c+Fu+this.b));return i},p.b=0,p.c=0,R(B0,"RegEx/Token/ClosureToken",319),N(816,121,ZL,$ke),p.Im=function(t){return t==0?this.a:this.b},p.Mm=function(){return 2},p.Nm=function(t){var i;return this.b.e==3&&this.b.Im(0)==this.a?i=this.a.Nm(t)+"+":this.b.e==9&&this.b.Im(0)==this.a?i=this.a.Nm(t)+"+?":i=this.a.Nm(t)+(""+this.b.Nm(t)),i},R(B0,"RegEx/Token/ConcatToken",816),N(1891,121,ZL,hmt),p.Im=function(t){if(t==0)return this.d;if(t==1)return this.b;throw ie(new oc("Internal Error: "+t))},p.Mm=function(){return this.b?2:1},p.Nm=function(t){var i;return this.c>0?i="(?("+this.c+")":this.a.e==8?i="(?("+this.a+")":i="(?"+this.a,this.b?i+=this.d+"|"+this.b+")":i+=this.d+")",i},p.c=0,R(B0,"RegEx/Token/ConditionToken",1891),N(1892,121,ZL,tht),p.Im=function(t){return this.b},p.Mm=function(){return 1},p.Nm=function(t){return"(?"+(this.a==0?"":SIe(this.a))+(this.c==0?"":SIe(this.c))+":"+this.b.Nm(t)+")"},p.a=0,p.c=0,R(B0,"RegEx/Token/ModifierToken",1892),N(817,121,ZL,Wke),p.Im=function(t){return this.a},p.Mm=function(){return 1},p.Nm=function(t){var i;switch(i=null,this.e){case 6:this.b==0?i="(?:"+this.a.Nm(t)+")":i="("+this.a.Nm(t)+")";break;case 20:i="(?="+this.a.Nm(t)+")";break;case 21:i="(?!"+this.a.Nm(t)+")";break;case 22:i="(?<="+this.a.Nm(t)+")";break;case 23:i="(?<!"+this.a.Nm(t)+")";break;case 24:i="(?>"+this.a.Nm(t)+")"}return i},p.b=0,R(B0,"RegEx/Token/ParenToken",817),N(515,121,{3:1,121:1,515:1},Bce),p.Jm=function(){return this.b},p.Nm=function(t){return this.e==12?"\\"+this.a:m6n(this.b)},p.a=0,R(B0,"RegEx/Token/StringToken",515),N(463,121,ZL,D7),p.Gm=function(t){vE(this,t)},p.Im=function(t){return f(SS(this.a,t),121)},p.Mm=function(){return this.a?this.a.a.c.length:0},p.Nm=function(t){var i,a,u,l,g;if(this.e==1){if(this.a.a.c.length==2)i=f(SS(this.a,0),121),a=f(SS(this.a,1),121),a.e==3&&a.Im(0)==i?l=i.Nm(t)+"+":a.e==9&&a.Im(0)==i?l=i.Nm(t)+"+?":l=i.Nm(t)+(""+a.Nm(t));else{for(g=new p0,u=0;u<this.a.a.c.length;u++)ys(g,f(SS(this.a,u),121).Nm(t));l=g.a}return l}if(this.a.a.c.length==2&&f(SS(this.a,1),121).e==7)l=f(SS(this.a,0),121).Nm(t)+"?";else if(this.a.a.c.length==2&&f(SS(this.a,0),121).e==7)l=f(SS(this.a,1),121).Nm(t)+"??";else{for(g=new p0,ys(g,f(SS(this.a,0),121).Nm(t)),u=1;u<this.a.a.c.length;u++)g.a+="|",ys(g,f(SS(this.a,u),121).Nm(t));l=g.a}return l},R(B0,"RegEx/Token/UnionToken",463),N(512,1,{584:1},Tz),p.Ib=function(){return this.a.b},R(yAt,"XMLTypeUtil/PatternMatcherImpl",512),N(1649,1467,{},s3);var KMt;R(yAt,"XMLTypeValidator",1649),N(267,1,Eb,Xy),p.Ic=function(t){Oa(this,t)},p.Jc=function(){return(this.b-this.a)*this.c<0?Fw:new Vv(this)},p.a=0,p.b=0,p.c=0;var Fw;R(l6e,"ExclusiveRange",267),N(1049,1,yb,ZU),p.Rb=function(t){f(t,17),smn()},p.Nb=function(t){va(this,t)},p.Pb=function(){return Kpn()},p.Ub=function(){return Ypn()},p.Wb=function(t){f(t,17),cmn()},p.Ob=function(){return!1},p.Sb=function(){return!1},p.Tb=function(){return-1},p.Vb=function(){return-1},p.Qb=function(){throw ie(new f0(_At))},R(l6e,"ExclusiveRange/1",1049),N(258,1,yb,Vv),p.Rb=function(t){f(t,17),lmn()},p.Nb=function(t){va(this,t)},p.Pb=function(){return kxn(this)},p.Ub=function(){return __n(this)},p.Wb=function(t){f(t,17),umn()},p.Ob=function(){return this.c.c<0?this.a>=this.c.b:this.a<=this.c.b},p.Sb=function(){return this.b>0},p.Tb=function(){return this.b},p.Vb=function(){return this.b-1},p.Qb=function(){throw ie(new f0(_At))},p.a=0,p.b=0,R(l6e,"ExclusiveRange/RangeIterator",258);var Xf=kP(hee,"C"),pr=kP(I$,"I"),kd=kP(xR,"Z"),x_=kP(O$,"J"),$l=kP(x$,"B"),sa=kP(C$,"D"),kx=kP(A$,"F"),JA=kP(N$,"S"),yjn=Bi("org.eclipse.elk.core.labels","ILabelManager"),fje=Bi(Ha,"DiagnosticChain"),hje=Bi(J3t,"ResourceSet"),pje=R(Ha,"InvocationTargetException",null),YMt=(hY(),MSn),XMt=XMt=rOn;ykn(Whn),$kn("permProps",[[["locale","default"],[kAt,"gecko1_8"]],[["locale","default"],[kAt,"safari"]]]),XMt(null,"elk",null)}).call(this)}).call(this,typeof pI<"u"?pI:typeof self<"u"?self:typeof window<"u"?window:{})},{}],3:[function(o,s,c){function d(B){"@babel/helpers - typeof";return d=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(G){return typeof G}:function(G){return G&&typeof Symbol=="function"&&G.constructor===Symbol&&G!==Symbol.prototype?"symbol":typeof G},d(B)}function h(B,G,K){return Object.defineProperty(B,"prototype",{writable:!1}),B}function b(B,G){if(!(B instanceof G))throw new TypeError("Cannot call a class as a function")}function y(B,G,K){return G=O(G),E(B,x()?Reflect.construct(G,K||[],O(B).constructor):G.apply(B,K))}function E(B,G){if(G&&(d(G)=="object"||typeof G=="function"))return G;if(G!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return k(B)}function k(B){if(B===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return B}function x(){try{var B=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(x=function(){return!!B})()}function O(B){return O=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(G){return G.__proto__||Object.getPrototypeOf(G)},O(B)}function P(B,G){if(typeof G!="function"&&G!==null)throw new TypeError("Super expression must either be null or a function");B.prototype=Object.create(G&&G.prototype,{constructor:{value:B,writable:!0,configurable:!0}}),Object.defineProperty(B,"prototype",{writable:!1}),G&&D(B,G)}function D(B,G){return D=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(K,X){return K.__proto__=X,K},D(B,G)}var $=o("./elk-api.js").default,U=function(B){function G(){var K=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};b(this,G);var X=Object.assign({},K),ae=!1;try{o.resolve("web-worker"),ae=!0}catch{}if(K.workerUrl)if(ae){var oe=o("web-worker");X.workerFactory=function(we){return new oe(we)}}else console.warn(`Web worker requested but 'web-worker' package not installed.
Consider installing the package or pass your own 'workerFactory' to ELK's constructor.
... Falling back to non-web worker version.`);if(!X.workerFactory){var he=o("./elk-worker.min.js"),ye=he.Worker;X.workerFactory=function(we){return new ye(we)}}return y(this,G,[X])}return P(G,B),h(G)}($);Object.defineProperty(s.exports,"__esModule",{value:!0}),s.exports=U,U.default=U},{"./elk-api.js":1,"./elk-worker.min.js":2,"web-worker":4}],4:[function(o,s,c){var d=typeof Worker<"u"?Worker:void 0;s.exports=d},{}]},{},[3])(3)})}($$e)),$$e.exports}var MIr=DIr();const jIr=Y4(MIr);function q_(n){if(typeof n=="string"||typeof n=="number")return""+n;let r="";if(Array.isArray(n))for(let o=0,s;o<n.length;o++)(s=q_(n[o]))!==""&&(r+=(r&&" ")+s);else for(let o in n)n[o]&&(r+=(r&&" ")+o);return r}const $Ir={},L$t=n=>{let r;const o=new Set,s=(k,x)=>{const O=typeof k=="function"?k(r):k;if(!Object.is(O,r)){const P=r;r=x??(typeof O!="object"||O===null)?O:Object.assign({},r,O),o.forEach(D=>D(r,P))}},c=()=>r,y={setState:s,getState:c,getInitialState:()=>E,subscribe:k=>(o.add(k),()=>o.delete(k)),destroy:()=>{($Ir?"production":void 0)!=="production"&&console.warn("[DEPRECATED] The `destroy` method will be unsupported in a future version. Instead use unsubscribe function returned by subscribe. Everything will be garbage-collected if store is garbage-collected."),o.clear()}},E=r=n(s,c,y);return y},FIr=n=>n?L$t(n):L$t,{useDebugValue:BIr}=Un,{useSyncExternalStoreWithSelector:UIr}=Yor,zIr=n=>n;function isn(n,r=zIr,o){const s=UIr(n.subscribe,n.getState,n.getServerState||n.getInitialState,r,o);return BIr(s),s}const D$t=(n,r)=>{const o=FIr(n),s=(c,d=r)=>isn(o,c,d);return Object.assign(s,o),s},HIr=(n,r)=>n?D$t(n,r):D$t;function ey(n,r){if(Object.is(n,r))return!0;if(typeof n!="object"||n===null||typeof r!="object"||r===null)return!1;if(n instanceof Map&&r instanceof Map){if(n.size!==r.size)return!1;for(const[s,c]of n)if(!Object.is(c,r.get(s)))return!1;return!0}if(n instanceof Set&&r instanceof Set){if(n.size!==r.size)return!1;for(const s of n)if(!r.has(s))return!1;return!0}const o=Object.keys(n);if(o.length!==Object.keys(r).length)return!1;for(const s of o)if(!Object.prototype.hasOwnProperty.call(r,s)||!Object.is(n[s],r[s]))return!1;return!0}var GIr={value:()=>{}};function uEe(){for(var n=0,r=arguments.length,o={},s;n<r;++n){if(!(s=arguments[n]+"")||s in o||/[\s.]/.test(s))throw new Error("illegal type: "+s);o[s]=[]}return new vve(o)}function vve(n){this._=n}function qIr(n,r){return n.trim().split(/^|\s+/).map(function(o){var s="",c=o.indexOf(".");if(c>=0&&(s=o.slice(c+1),o=o.slice(0,c)),o&&!r.hasOwnProperty(o))throw new Error("unknown type: "+o);return{type:o,name:s}})}vve.prototype=uEe.prototype={constructor:vve,on:function(n,r){var o=this._,s=qIr(n+"",o),c,d=-1,h=s.length;if(arguments.length<2){for(;++d<h;)if((c=(n=s[d]).type)&&(c=VIr(o[c],n.name)))return c;return}if(r!=null&&typeof r!="function")throw new Error("invalid callback: "+r);for(;++d<h;)if(c=(n=s[d]).type)o[c]=M$t(o[c],n.name,r);else if(r==null)for(c in o)o[c]=M$t(o[c],n.name,null);return this},copy:function(){var n={},r=this._;for(var o in r)n[o]=r[o].slice();return new vve(n)},call:function(n,r){if((c=arguments.length-2)>0)for(var o=new Array(c),s=0,c,d;s<c;++s)o[s]=arguments[s+2];if(!this._.hasOwnProperty(n))throw new Error("unknown type: "+n);for(d=this._[n],s=0,c=d.length;s<c;++s)d[s].value.apply(r,o)},apply:function(n,r,o){if(!this._.hasOwnProperty(n))throw new Error("unknown type: "+n);for(var s=this._[n],c=0,d=s.length;c<d;++c)s[c].value.apply(r,o)}};function VIr(n,r){for(var o=0,s=n.length,c;o<s;++o)if((c=n[o]).name===r)return c.value}function M$t(n,r,o){for(var s=0,c=n.length;s<c;++s)if(n[s].name===r){n[s]=GIr,n=n.slice(0,s).concat(n.slice(s+1));break}return o!=null&&n.push({name:r,value:o}),n}var rWe="http://www.w3.org/1999/xhtml";const j$t={svg:"http://www.w3.org/2000/svg",xhtml:rWe,xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"};function lEe(n){var r=n+="",o=r.indexOf(":");return o>=0&&(r=n.slice(0,o))!=="xmlns"&&(n=n.slice(o+1)),j$t.hasOwnProperty(r)?{space:j$t[r],local:n}:n}function WIr(n){return function(){var r=this.ownerDocument,o=this.namespaceURI;return o===rWe&&r.documentElement.namespaceURI===rWe?r.createElement(n):r.createElementNS(o,n)}}function KIr(n){return function(){return this.ownerDocument.createElementNS(n.space,n.local)}}function osn(n){var r=lEe(n);return(r.local?KIr:WIr)(r)}function YIr(){}function SZe(n){return n==null?YIr:function(){return this.querySelector(n)}}function XIr(n){typeof n!="function"&&(n=SZe(n));for(var r=this._groups,o=r.length,s=new Array(o),c=0;c<o;++c)for(var d=r[c],h=d.length,b=s[c]=new Array(h),y,E,k=0;k<h;++k)(y=d[k])&&(E=n.call(y,y.__data__,k,d))&&("__data__"in y&&(E.__data__=y.__data__),b[k]=E);return new m2(s,this._parents)}function JIr(n){return n==null?[]:Array.isArray(n)?n:Array.from(n)}function ZIr(){return[]}function asn(n){return n==null?ZIr:function(){return this.querySelectorAll(n)}}function QIr(n){return function(){return JIr(n.apply(this,arguments))}}function eOr(n){typeof n=="function"?n=QIr(n):n=asn(n);for(var r=this._groups,o=r.length,s=[],c=[],d=0;d<o;++d)for(var h=r[d],b=h.length,y,E=0;E<b;++E)(y=h[E])&&(s.push(n.call(y,y.__data__,E,h)),c.push(y));return new m2(s,c)}function ssn(n){return function(){return this.matches(n)}}function csn(n){return function(r){return r.matches(n)}}var tOr=Array.prototype.find;function nOr(n){return function(){return tOr.call(this.children,n)}}function rOr(){return this.firstElementChild}function iOr(n){return this.select(n==null?rOr:nOr(typeof n=="function"?n:csn(n)))}var oOr=Array.prototype.filter;function aOr(){return Array.from(this.children)}function sOr(n){return function(){return oOr.call(this.children,n)}}function cOr(n){return this.selectAll(n==null?aOr:sOr(typeof n=="function"?n:csn(n)))}function uOr(n){typeof n!="function"&&(n=ssn(n));for(var r=this._groups,o=r.length,s=new Array(o),c=0;c<o;++c)for(var d=r[c],h=d.length,b=s[c]=[],y,E=0;E<h;++E)(y=d[E])&&n.call(y,y.__data__,E,d)&&b.push(y);return new m2(s,this._parents)}function usn(n){return new Array(n.length)}function lOr(){return new m2(this._enter||this._groups.map(usn),this._parents)}function mwe(n,r){this.ownerDocument=n.ownerDocument,this.namespaceURI=n.namespaceURI,this._next=null,this._parent=n,this.__data__=r}mwe.prototype={constructor:mwe,appendChild:function(n){return this._parent.insertBefore(n,this._next)},insertBefore:function(n,r){return this._parent.insertBefore(n,r)},querySelector:function(n){return this._parent.querySelector(n)},querySelectorAll:function(n){return this._parent.querySelectorAll(n)}};function dOr(n){return function(){return n}}function fOr(n,r,o,s,c,d){for(var h=0,b,y=r.length,E=d.length;h<E;++h)(b=r[h])?(b.__data__=d[h],s[h]=b):o[h]=new mwe(n,d[h]);for(;h<y;++h)(b=r[h])&&(c[h]=b)}function hOr(n,r,o,s,c,d,h){var b,y,E=new Map,k=r.length,x=d.length,O=new Array(k),P;for(b=0;b<k;++b)(y=r[b])&&(O[b]=P=h.call(y,y.__data__,b,r)+"",E.has(P)?c[b]=y:E.set(P,y));for(b=0;b<x;++b)P=h.call(n,d[b],b,d)+"",(y=E.get(P))?(s[b]=y,y.__data__=d[b],E.delete(P)):o[b]=new mwe(n,d[b]);for(b=0;b<k;++b)(y=r[b])&&E.get(O[b])===y&&(c[b]=y)}function pOr(n){return n.__data__}function gOr(n,r){if(!arguments.length)return Array.from(this,pOr);var o=r?hOr:fOr,s=this._parents,c=this._groups;typeof n!="function"&&(n=dOr(n));for(var d=c.length,h=new Array(d),b=new Array(d),y=new Array(d),E=0;E<d;++E){var k=s[E],x=c[E],O=x.length,P=mOr(n.call(k,k&&k.__data__,E,s)),D=P.length,$=b[E]=new Array(D),U=h[E]=new Array(D),B=y[E]=new Array(O);o(k,x,$,U,B,P,r);for(var G=0,K=0,X,ae;G<D;++G)if(X=$[G]){for(G>=K&&(K=G+1);!(ae=U[K])&&++K<D;);X._next=ae||null}}return h=new m2(h,s),h._enter=b,h._exit=y,h}function mOr(n){return typeof n=="object"&&"length"in n?n:Array.from(n)}function bOr(){return new m2(this._exit||this._groups.map(usn),this._parents)}function vOr(n,r,o){var s=this.enter(),c=this,d=this.exit();return typeof n=="function"?(s=n(s),s&&(s=s.selection())):s=s.append(n+""),r!=null&&(c=r(c),c&&(c=c.selection())),o==null?d.remove():o(d),s&&c?s.merge(c).order():c}function wOr(n){for(var r=n.selection?n.selection():n,o=this._groups,s=r._groups,c=o.length,d=s.length,h=Math.min(c,d),b=new Array(c),y=0;y<h;++y)for(var E=o[y],k=s[y],x=E.length,O=b[y]=new Array(x),P,D=0;D<x;++D)(P=E[D]||k[D])&&(O[D]=P);for(;y<c;++y)b[y]=o[y];return new m2(b,this._parents)}function yOr(){for(var n=this._groups,r=-1,o=n.length;++r<o;)for(var s=n[r],c=s.length-1,d=s[c],h;--c>=0;)(h=s[c])&&(d&&h.compareDocumentPosition(d)^4&&d.parentNode.insertBefore(h,d),d=h);return this}function EOr(n){n||(n=SOr);function r(x,O){return x&&O?n(x.__data__,O.__data__):!x-!O}for(var o=this._groups,s=o.length,c=new Array(s),d=0;d<s;++d){for(var h=o[d],b=h.length,y=c[d]=new Array(b),E,k=0;k<b;++k)(E=h[k])&&(y[k]=E);y.sort(r)}return new m2(c,this._parents).order()}function SOr(n,r){return n<r?-1:n>r?1:n>=r?0:NaN}function _Or(){var n=arguments[0];return arguments[0]=this,n.apply(null,arguments),this}function kOr(){return Array.from(this)}function TOr(){for(var n=this._groups,r=0,o=n.length;r<o;++r)for(var s=n[r],c=0,d=s.length;c<d;++c){var h=s[c];if(h)return h}return null}function xOr(){let n=0;for(const r of this)++n;return n}function COr(){return!this.node()}function AOr(n){for(var r=this._groups,o=0,s=r.length;o<s;++o)for(var c=r[o],d=0,h=c.length,b;d<h;++d)(b=c[d])&&n.call(b,b.__data__,d,c);return this}function IOr(n){return function(){this.removeAttribute(n)}}function OOr(n){return function(){this.removeAttributeNS(n.space,n.local)}}function NOr(n,r){return function(){this.setAttribute(n,r)}}function ROr(n,r){return function(){this.setAttributeNS(n.space,n.local,r)}}function POr(n,r){return function(){var o=r.apply(this,arguments);o==null?this.removeAttribute(n):this.setAttribute(n,o)}}function LOr(n,r){return function(){var o=r.apply(this,arguments);o==null?this.removeAttributeNS(n.space,n.local):this.setAttributeNS(n.space,n.local,o)}}function DOr(n,r){var o=lEe(n);if(arguments.length<2){var s=this.node();return o.local?s.getAttributeNS(o.space,o.local):s.getAttribute(o)}return this.each((r==null?o.local?OOr:IOr:typeof r=="function"?o.local?LOr:POr:o.local?ROr:NOr)(o,r))}function lsn(n){return n.ownerDocument&&n.ownerDocument.defaultView||n.document&&n||n.defaultView}function MOr(n){return function(){this.style.removeProperty(n)}}function jOr(n,r,o){return function(){this.style.setProperty(n,r,o)}}function $Or(n,r,o){return function(){var s=r.apply(this,arguments);s==null?this.style.removeProperty(n):this.style.setProperty(n,s,o)}}function FOr(n,r,o){return arguments.length>1?this.each((r==null?MOr:typeof r=="function"?$Or:jOr)(n,r,o??"")):rK(this.node(),n)}function rK(n,r){return n.style.getPropertyValue(r)||lsn(n).getComputedStyle(n,null).getPropertyValue(r)}function BOr(n){return function(){delete this[n]}}function UOr(n,r){return function(){this[n]=r}}function zOr(n,r){return function(){var o=r.apply(this,arguments);o==null?delete this[n]:this[n]=o}}function HOr(n,r){return arguments.length>1?this.each((r==null?BOr:typeof r=="function"?zOr:UOr)(n,r)):this.node()[n]}function dsn(n){return n.trim().split(/^|\s+/)}function _Ze(n){return n.classList||new fsn(n)}function fsn(n){this._node=n,this._names=dsn(n.getAttribute("class")||"")}fsn.prototype={add:function(n){var r=this._names.indexOf(n);r<0&&(this._names.push(n),this._node.setAttribute("class",this._names.join(" ")))},remove:function(n){var r=this._names.indexOf(n);r>=0&&(this._names.splice(r,1),this._node.setAttribute("class",this._names.join(" ")))},contains:function(n){return this._names.indexOf(n)>=0}};function hsn(n,r){for(var o=_Ze(n),s=-1,c=r.length;++s<c;)o.add(r[s])}function psn(n,r){for(var o=_Ze(n),s=-1,c=r.length;++s<c;)o.remove(r[s])}function GOr(n){return function(){hsn(this,n)}}function qOr(n){return function(){psn(this,n)}}function VOr(n,r){return function(){(r.apply(this,arguments)?hsn:psn)(this,n)}}function WOr(n,r){var o=dsn(n+"");if(arguments.length<2){for(var s=_Ze(this.node()),c=-1,d=o.length;++c<d;)if(!s.contains(o[c]))return!1;return!0}return this.each((typeof r=="function"?VOr:r?GOr:qOr)(o,r))}function KOr(){this.textContent=""}function YOr(n){return function(){this.textContent=n}}function XOr(n){return function(){var r=n.apply(this,arguments);this.textContent=r??""}}function JOr(n){return arguments.length?this.each(n==null?KOr:(typeof n=="function"?XOr:YOr)(n)):this.node().textContent}function ZOr(){this.innerHTML=""}function QOr(n){return function(){this.innerHTML=n}}function eNr(n){return function(){var r=n.apply(this,arguments);this.innerHTML=r??""}}function tNr(n){return arguments.length?this.each(n==null?ZOr:(typeof n=="function"?eNr:QOr)(n)):this.node().innerHTML}function nNr(){this.nextSibling&&this.parentNode.appendChild(this)}function rNr(){return this.each(nNr)}function iNr(){this.previousSibling&&this.parentNode.insertBefore(this,this.parentNode.firstChild)}function oNr(){return this.each(iNr)}function aNr(n){var r=typeof n=="function"?n:osn(n);return this.select(function(){return this.appendChild(r.apply(this,arguments))})}function sNr(){return null}function cNr(n,r){var o=typeof n=="function"?n:osn(n),s=r==null?sNr:typeof r=="function"?r:SZe(r);return this.select(function(){return this.insertBefore(o.apply(this,arguments),s.apply(this,arguments)||null)})}function uNr(){var n=this.parentNode;n&&n.removeChild(this)}function lNr(){return this.each(uNr)}function dNr(){var n=this.cloneNode(!1),r=this.parentNode;return r?r.insertBefore(n,this.nextSibling):n}function fNr(){var n=this.cloneNode(!0),r=this.parentNode;return r?r.insertBefore(n,this.nextSibling):n}function hNr(n){return this.select(n?fNr:dNr)}function pNr(n){return arguments.length?this.property("__data__",n):this.node().__data__}function gNr(n){return function(r){n.call(this,r,this.__data__)}}function mNr(n){return n.trim().split(/^|\s+/).map(function(r){var o="",s=r.indexOf(".");return s>=0&&(o=r.slice(s+1),r=r.slice(0,s)),{type:r,name:o}})}function bNr(n){return function(){var r=this.__on;if(r){for(var o=0,s=-1,c=r.length,d;o<c;++o)d=r[o],(!n.type||d.type===n.type)&&d.name===n.name?this.removeEventListener(d.type,d.listener,d.options):r[++s]=d;++s?r.length=s:delete this.__on}}}function vNr(n,r,o){return function(){var s=this.__on,c,d=gNr(r);if(s){for(var h=0,b=s.length;h<b;++h)if((c=s[h]).type===n.type&&c.name===n.name){this.removeEventListener(c.type,c.listener,c.options),this.addEventListener(c.type,c.listener=d,c.options=o),c.value=r;return}}this.addEventListener(n.type,d,o),c={type:n.type,name:n.name,value:r,listener:d,options:o},s?s.push(c):this.__on=[c]}}function wNr(n,r,o){var s=mNr(n+""),c,d=s.length,h;if(arguments.length<2){var b=this.node().__on;if(b){for(var y=0,E=b.length,k;y<E;++y)for(c=0,k=b[y];c<d;++c)if((h=s[c]).type===k.type&&h.name===k.name)return k.value}return}for(b=r?vNr:bNr,c=0;c<d;++c)this.each(b(s[c],r,o));return this}function gsn(n,r,o){var s=lsn(n),c=s.CustomEvent;typeof c=="function"?c=new c(r,o):(c=s.document.createEvent("Event"),o?(c.initEvent(r,o.bubbles,o.cancelable),c.detail=o.detail):c.initEvent(r,!1,!1)),n.dispatchEvent(c)}function yNr(n,r){return function(){return gsn(this,n,r)}}function ENr(n,r){return function(){return gsn(this,n,r.apply(this,arguments))}}function SNr(n,r){return this.each((typeof r=="function"?ENr:yNr)(n,r))}function*_Nr(){for(var n=this._groups,r=0,o=n.length;r<o;++r)for(var s=n[r],c=0,d=s.length,h;c<d;++c)(h=s[c])&&(yield h)}var msn=[null];function m2(n,r){this._groups=n,this._parents=r}function Hoe(){return new m2([[document.documentElement]],msn)}function kNr(){return this}m2.prototype=Hoe.prototype={constructor:m2,select:XIr,selectAll:eOr,selectChild:iOr,selectChildren:cOr,filter:uOr,data:gOr,enter:lOr,exit:bOr,join:vOr,merge:wOr,selection:kNr,order:yOr,sort:EOr,call:_Or,nodes:kOr,node:TOr,size:xOr,empty:COr,each:AOr,attr:DOr,style:FOr,property:HOr,classed:WOr,text:JOr,html:tNr,raise:rNr,lower:oNr,append:aNr,insert:cNr,remove:lNr,clone:hNr,datum:pNr,on:wNr,dispatch:SNr,[Symbol.iterator]:_Nr};function Rx(n){return typeof n=="string"?new m2([[document.querySelector(n)]],[document.documentElement]):new m2([[n]],msn)}function TNr(n){let r;for(;r=n.sourceEvent;)n=r;return n}function sI(n,r){if(n=TNr(n),r===void 0&&(r=n.currentTarget),r){var o=r.ownerSVGElement||r;if(o.createSVGPoint){var s=o.createSVGPoint();return s.x=n.clientX,s.y=n.clientY,s=s.matrixTransform(r.getScreenCTM().inverse()),[s.x,s.y]}if(r.getBoundingClientRect){var c=r.getBoundingClientRect();return[n.clientX-c.left-r.clientLeft,n.clientY-c.top-r.clientTop]}}return[n.pageX,n.pageY]}const xNr={passive:!1},Jie={capture:!0,passive:!1};function F$e(n){n.stopImmediatePropagation()}function RW(n){n.preventDefault(),n.stopImmediatePropagation()}function bsn(n){var r=n.document.documentElement,o=Rx(n).on("dragstart.drag",RW,Jie);"onselectstart"in r?o.on("selectstart.drag",RW,Jie):(r.__noselect=r.style.MozUserSelect,r.style.MozUserSelect="none")}function vsn(n,r){var o=n.document.documentElement,s=Rx(n).on("dragstart.drag",null);r&&(s.on("click.drag",RW,Jie),setTimeout(function(){s.on("click.drag",null)},0)),"onselectstart"in o?s.on("selectstart.drag",null):(o.style.MozUserSelect=o.__noselect,delete o.__noselect)}const p0e=n=>()=>n;function iWe(n,{sourceEvent:r,subject:o,target:s,identifier:c,active:d,x:h,y:b,dx:y,dy:E,dispatch:k}){Object.defineProperties(this,{type:{value:n,enumerable:!0,configurable:!0},sourceEvent:{value:r,enumerable:!0,configurable:!0},subject:{value:o,enumerable:!0,configurable:!0},target:{value:s,enumerable:!0,configurable:!0},identifier:{value:c,enumerable:!0,configurable:!0},active:{value:d,enumerable:!0,configurable:!0},x:{value:h,enumerable:!0,configurable:!0},y:{value:b,enumerable:!0,configurable:!0},dx:{value:y,enumerable:!0,configurable:!0},dy:{value:E,enumerable:!0,configurable:!0},_:{value:k}})}iWe.prototype.on=function(){var n=this._.on.apply(this._,arguments);return n===this._?this:n};function CNr(n){return!n.ctrlKey&&!n.button}function ANr(){return this.parentNode}function INr(n,r){return r??{x:n.x,y:n.y}}function ONr(){return navigator.maxTouchPoints||"ontouchstart"in this}function NNr(){var n=CNr,r=ANr,o=INr,s=ONr,c={},d=uEe("start","drag","end"),h=0,b,y,E,k,x=0;function O(X){X.on("mousedown.drag",P).filter(s).on("touchstart.drag",U).on("touchmove.drag",B,xNr).on("touchend.drag touchcancel.drag",G).style("touch-action","none").style("-webkit-tap-highlight-color","rgba(0,0,0,0)")}function P(X,ae){if(!(k||!n.call(this,X,ae))){var oe=K(this,r.call(this,X,ae),X,ae,"mouse");oe&&(Rx(X.view).on("mousemove.drag",D,Jie).on("mouseup.drag",$,Jie),bsn(X.view),F$e(X),E=!1,b=X.clientX,y=X.clientY,oe("start",X))}}function D(X){if(RW(X),!E){var ae=X.clientX-b,oe=X.clientY-y;E=ae*ae+oe*oe>x}c.mouse("drag",X)}function $(X){Rx(X.view).on("mousemove.drag mouseup.drag",null),vsn(X.view,E),RW(X),c.mouse("end",X)}function U(X,ae){if(n.call(this,X,ae)){var oe=X.changedTouches,he=r.call(this,X,ae),ye=oe.length,we,ge;for(we=0;we<ye;++we)(ge=K(this,he,X,ae,oe[we].identifier,oe[we]))&&(F$e(X),ge("start",X,oe[we]))}}function B(X){var ae=X.changedTouches,oe=ae.length,he,ye;for(he=0;he<oe;++he)(ye=c[ae[he].identifier])&&(RW(X),ye("drag",X,ae[he]))}function G(X){var ae=X.changedTouches,oe=ae.length,he,ye;for(k&&clearTimeout(k),k=setTimeout(function(){k=null},500),he=0;he<oe;++he)(ye=c[ae[he].identifier])&&(F$e(X),ye("end",X,ae[he]))}function K(X,ae,oe,he,ye,we){var ge=d.copy(),Te=sI(we||oe,ae),xe,Ue,Z;if((Z=o.call(X,new iWe("beforestart",{sourceEvent:oe,target:O,identifier:ye,active:h,x:Te[0],y:Te[1],dx:0,dy:0,dispatch:ge}),he))!=null)return xe=Z.x-Te[0]||0,Ue=Z.y-Te[1]||0,function Re(_e,qe,it){var Xe=Te,de;switch(_e){case"start":c[ye]=Re,de=h++;break;case"end":delete c[ye],--h;case"drag":Te=sI(it||qe,ae),de=h;break}ge.call(_e,X,new iWe(_e,{sourceEvent:qe,subject:Z,target:O,identifier:ye,active:de,x:Te[0]+xe,y:Te[1]+Ue,dx:Te[0]-Xe[0],dy:Te[1]-Xe[1],dispatch:ge}),he)}}return O.filter=function(X){return arguments.length?(n=typeof X=="function"?X:p0e(!!X),O):n},O.container=function(X){return arguments.length?(r=typeof X=="function"?X:p0e(X),O):r},O.subject=function(X){return arguments.length?(o=typeof X=="function"?X:p0e(X),O):o},O.touchable=function(X){return arguments.length?(s=typeof X=="function"?X:p0e(!!X),O):s},O.on=function(){var X=d.on.apply(d,arguments);return X===d?O:X},O.clickDistance=function(X){return arguments.length?(x=(X=+X)*X,O):Math.sqrt(x)},O}function kZe(n,r,o){n.prototype=r.prototype=o,o.constructor=n}function wsn(n,r){var o=Object.create(n.prototype);for(var s in r)o[s]=r[s];return o}function Goe(){}var Zie=.7,bwe=1/Zie,PW="\\s*([+-]?\\d+)\\s*",Qie="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)\\s*",xI="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)%\\s*",RNr=/^#([0-9a-f]{3,8})$/,PNr=new RegExp(`^rgb\\(${PW},${PW},${PW}\\)$`),LNr=new RegExp(`^rgb\\(${xI},${xI},${xI}\\)$`),DNr=new RegExp(`^rgba\\(${PW},${PW},${PW},${Qie}\\)$`),MNr=new RegExp(`^rgba\\(${xI},${xI},${xI},${Qie}\\)$`),jNr=new RegExp(`^hsl\\(${Qie},${xI},${xI}\\)$`),$Nr=new RegExp(`^hsla\\(${Qie},${xI},${xI},${Qie}\\)$`),$$t={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};kZe(Goe,eoe,{copy(n){return Object.assign(new this.constructor,this,n)},displayable(){return this.rgb().displayable()},hex:F$t,formatHex:F$t,formatHex8:FNr,formatHsl:BNr,formatRgb:B$t,toString:B$t});function F$t(){return this.rgb().formatHex()}function FNr(){return this.rgb().formatHex8()}function BNr(){return ysn(this).formatHsl()}function B$t(){return this.rgb().formatRgb()}function eoe(n){var r,o;return n=(n+"").trim().toLowerCase(),(r=RNr.exec(n))?(o=r[1].length,r=parseInt(r[1],16),o===6?U$t(r):o===3?new Xw(r>>8&15|r>>4&240,r>>4&15|r&240,(r&15)<<4|r&15,1):o===8?g0e(r>>24&255,r>>16&255,r>>8&255,(r&255)/255):o===4?g0e(r>>12&15|r>>8&240,r>>8&15|r>>4&240,r>>4&15|r&240,((r&15)<<4|r&15)/255):null):(r=PNr.exec(n))?new Xw(r[1],r[2],r[3],1):(r=LNr.exec(n))?new Xw(r[1]*255/100,r[2]*255/100,r[3]*255/100,1):(r=DNr.exec(n))?g0e(r[1],r[2],r[3],r[4]):(r=MNr.exec(n))?g0e(r[1]*255/100,r[2]*255/100,r[3]*255/100,r[4]):(r=jNr.exec(n))?G$t(r[1],r[2]/100,r[3]/100,1):(r=$Nr.exec(n))?G$t(r[1],r[2]/100,r[3]/100,r[4]):$$t.hasOwnProperty(n)?U$t($$t[n]):n==="transparent"?new Xw(NaN,NaN,NaN,0):null}function U$t(n){return new Xw(n>>16&255,n>>8&255,n&255,1)}function g0e(n,r,o,s){return s<=0&&(n=r=o=NaN),new Xw(n,r,o,s)}function UNr(n){return n instanceof Goe||(n=eoe(n)),n?(n=n.rgb(),new Xw(n.r,n.g,n.b,n.opacity)):new Xw}function oWe(n,r,o,s){return arguments.length===1?UNr(n):new Xw(n,r,o,s??1)}function Xw(n,r,o,s){this.r=+n,this.g=+r,this.b=+o,this.opacity=+s}kZe(Xw,oWe,wsn(Goe,{brighter(n){return n=n==null?bwe:Math.pow(bwe,n),new Xw(this.r*n,this.g*n,this.b*n,this.opacity)},darker(n){return n=n==null?Zie:Math.pow(Zie,n),new Xw(this.r*n,this.g*n,this.b*n,this.opacity)},rgb(){return this},clamp(){return new Xw(zB(this.r),zB(this.g),zB(this.b),vwe(this.opacity))},displayable(){return-.5<=this.r&&this.r<255.5&&-.5<=this.g&&this.g<255.5&&-.5<=this.b&&this.b<255.5&&0<=this.opacity&&this.opacity<=1},hex:z$t,formatHex:z$t,formatHex8:zNr,formatRgb:H$t,toString:H$t}));function z$t(){return`#${DB(this.r)}${DB(this.g)}${DB(this.b)}`}function zNr(){return`#${DB(this.r)}${DB(this.g)}${DB(this.b)}${DB((isNaN(this.opacity)?1:this.opacity)*255)}`}function H$t(){const n=vwe(this.opacity);return`${n===1?"rgb(":"rgba("}${zB(this.r)}, ${zB(this.g)}, ${zB(this.b)}${n===1?")":`, ${n})`}`}function vwe(n){return isNaN(n)?1:Math.max(0,Math.min(1,n))}function zB(n){return Math.max(0,Math.min(255,Math.round(n)||0))}function DB(n){return n=zB(n),(n<16?"0":"")+n.toString(16)}function G$t(n,r,o,s){return s<=0?n=r=o=NaN:o<=0||o>=1?n=r=NaN:r<=0&&(n=NaN),new Px(n,r,o,s)}function ysn(n){if(n instanceof Px)return new Px(n.h,n.s,n.l,n.opacity);if(n instanceof Goe||(n=eoe(n)),!n)return new Px;if(n instanceof Px)return n;n=n.rgb();var r=n.r/255,o=n.g/255,s=n.b/255,c=Math.min(r,o,s),d=Math.max(r,o,s),h=NaN,b=d-c,y=(d+c)/2;return b?(r===d?h=(o-s)/b+(o<s)*6:o===d?h=(s-r)/b+2:h=(r-o)/b+4,b/=y<.5?d+c:2-d-c,h*=60):b=y>0&&y<1?0:h,new Px(h,b,y,n.opacity)}function HNr(n,r,o,s){return arguments.length===1?ysn(n):new Px(n,r,o,s??1)}function Px(n,r,o,s){this.h=+n,this.s=+r,this.l=+o,this.opacity=+s}kZe(Px,HNr,wsn(Goe,{brighter(n){return n=n==null?bwe:Math.pow(bwe,n),new Px(this.h,this.s,this.l*n,this.opacity)},darker(n){return n=n==null?Zie:Math.pow(Zie,n),new Px(this.h,this.s,this.l*n,this.opacity)},rgb(){var n=this.h%360+(this.h<0)*360,r=isNaN(n)||isNaN(this.s)?0:this.s,o=this.l,s=o+(o<.5?o:1-o)*r,c=2*o-s;return new Xw(B$e(n>=240?n-240:n+120,c,s),B$e(n,c,s),B$e(n<120?n+240:n-120,c,s),this.opacity)},clamp(){return new Px(q$t(this.h),m0e(this.s),m0e(this.l),vwe(this.opacity))},displayable(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1},formatHsl(){const n=vwe(this.opacity);return`${n===1?"hsl(":"hsla("}${q$t(this.h)}, ${m0e(this.s)*100}%, ${m0e(this.l)*100}%${n===1?")":`, ${n})`}`}}));function q$t(n){return n=(n||0)%360,n<0?n+360:n}function m0e(n){return Math.max(0,Math.min(1,n||0))}function B$e(n,r,o){return(n<60?r+(o-r)*n/60:n<180?o:n<240?r+(o-r)*(240-n)/60:r)*255}const Esn=n=>()=>n;function GNr(n,r){return function(o){return n+o*r}}function qNr(n,r,o){return n=Math.pow(n,o),r=Math.pow(r,o)-n,o=1/o,function(s){return Math.pow(n+s*r,o)}}function VNr(n){return(n=+n)==1?Ssn:function(r,o){return o-r?qNr(r,o,n):Esn(isNaN(r)?o:r)}}function Ssn(n,r){var o=r-n;return o?GNr(n,o):Esn(isNaN(n)?r:n)}const V$t=function n(r){var o=VNr(r);function s(c,d){var h=o((c=oWe(c)).r,(d=oWe(d)).r),b=o(c.g,d.g),y=o(c.b,d.b),E=Ssn(c.opacity,d.opacity);return function(k){return c.r=h(k),c.g=b(k),c.b=y(k),c.opacity=E(k),c+""}}return s.gamma=n,s}(1);function iM(n,r){return n=+n,r=+r,function(o){return n*(1-o)+r*o}}var aWe=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,U$e=new RegExp(aWe.source,"g");function WNr(n){return function(){return n}}function KNr(n){return function(r){return n(r)+""}}function YNr(n,r){var o=aWe.lastIndex=U$e.lastIndex=0,s,c,d,h=-1,b=[],y=[];for(n=n+"",r=r+"";(s=aWe.exec(n))&&(c=U$e.exec(r));)(d=c.index)>o&&(d=r.slice(o,d),b[h]?b[h]+=d:b[++h]=d),(s=s[0])===(c=c[0])?b[h]?b[h]+=c:b[++h]=c:(b[++h]=null,y.push({i:h,x:iM(s,c)})),o=U$e.lastIndex;return o<r.length&&(d=r.slice(o),b[h]?b[h]+=d:b[++h]=d),b.length<2?y[0]?KNr(y[0].x):WNr(r):(r=y.length,function(E){for(var k=0,x;k<r;++k)b[(x=y[k]).i]=x.x(E);return b.join("")})}var W$t=180/Math.PI,sWe={translateX:0,translateY:0,rotate:0,skewX:0,scaleX:1,scaleY:1};function _sn(n,r,o,s,c,d){var h,b,y;return(h=Math.sqrt(n*n+r*r))&&(n/=h,r/=h),(y=n*o+r*s)&&(o-=n*y,s-=r*y),(b=Math.sqrt(o*o+s*s))&&(o/=b,s/=b,y/=b),n*s<r*o&&(n=-n,r=-r,y=-y,h=-h),{translateX:c,translateY:d,rotate:Math.atan2(r,n)*W$t,skewX:Math.atan(y)*W$t,scaleX:h,scaleY:b}}var b0e;function XNr(n){const r=new(typeof DOMMatrix=="function"?DOMMatrix:WebKitCSSMatrix)(n+"");return r.isIdentity?sWe:_sn(r.a,r.b,r.c,r.d,r.e,r.f)}function JNr(n){return n==null||(b0e||(b0e=document.createElementNS("http://www.w3.org/2000/svg","g")),b0e.setAttribute("transform",n),!(n=b0e.transform.baseVal.consolidate()))?sWe:(n=n.matrix,_sn(n.a,n.b,n.c,n.d,n.e,n.f))}function ksn(n,r,o,s){function c(E){return E.length?E.pop()+" ":""}function d(E,k,x,O,P,D){if(E!==x||k!==O){var $=P.push("translate(",null,r,null,o);D.push({i:$-4,x:iM(E,x)},{i:$-2,x:iM(k,O)})}else(x||O)&&P.push("translate("+x+r+O+o)}function h(E,k,x,O){E!==k?(E-k>180?k+=360:k-E>180&&(E+=360),O.push({i:x.push(c(x)+"rotate(",null,s)-2,x:iM(E,k)})):k&&x.push(c(x)+"rotate("+k+s)}function b(E,k,x,O){E!==k?O.push({i:x.push(c(x)+"skewX(",null,s)-2,x:iM(E,k)}):k&&x.push(c(x)+"skewX("+k+s)}function y(E,k,x,O,P,D){if(E!==x||k!==O){var $=P.push(c(P)+"scale(",null,",",null,")");D.push({i:$-4,x:iM(E,x)},{i:$-2,x:iM(k,O)})}else(x!==1||O!==1)&&P.push(c(P)+"scale("+x+","+O+")")}return function(E,k){var x=[],O=[];return E=n(E),k=n(k),d(E.translateX,E.translateY,k.translateX,k.translateY,x,O),h(E.rotate,k.rotate,x,O),b(E.skewX,k.skewX,x,O),y(E.scaleX,E.scaleY,k.scaleX,k.scaleY,x,O),E=k=null,function(P){for(var D=-1,$=O.length,U;++D<$;)x[(U=O[D]).i]=U.x(P);return x.join("")}}}var ZNr=ksn(XNr,"px, ","px)","deg)"),QNr=ksn(JNr,", ",")",")"),eRr=1e-12;function K$t(n){return((n=Math.exp(n))+1/n)/2}function tRr(n){return((n=Math.exp(n))-1/n)/2}function nRr(n){return((n=Math.exp(2*n))-1)/(n+1)}const rRr=function n(r,o,s){function c(d,h){var b=d[0],y=d[1],E=d[2],k=h[0],x=h[1],O=h[2],P=k-b,D=x-y,$=P*P+D*D,U,B;if($<eRr)B=Math.log(O/E)/r,U=function(he){return[b+he*P,y+he*D,E*Math.exp(r*he*B)]};else{var G=Math.sqrt($),K=(O*O-E*E+s*$)/(2*E*o*G),X=(O*O-E*E-s*$)/(2*O*o*G),ae=Math.log(Math.sqrt(K*K+1)-K),oe=Math.log(Math.sqrt(X*X+1)-X);B=(oe-ae)/r,U=function(he){var ye=he*B,we=K$t(ae),ge=E/(o*G)*(we*nRr(r*ye+ae)-tRr(ae));return[b+ge*P,y+ge*D,E*we/K$t(r*ye+ae)]}}return U.duration=B*1e3*r/Math.SQRT2,U}return c.rho=function(d){var h=Math.max(.001,+d),b=h*h,y=b*b;return n(h,b,y)},c}(Math.SQRT2,2,4);var iK=0,$re=0,pre=0,Tsn=1e3,wwe,Fre,ywe=0,tU=0,dEe=0,toe=typeof performance=="object"&&performance.now?performance:Date,xsn=typeof window=="object"&&window.requestAnimationFrame?window.requestAnimationFrame.bind(window):function(n){setTimeout(n,17)};function TZe(){return tU||(xsn(iRr),tU=toe.now()+dEe)}function iRr(){tU=0}function Ewe(){this._call=this._time=this._next=null}Ewe.prototype=Csn.prototype={constructor:Ewe,restart:function(n,r,o){if(typeof n!="function")throw new TypeError("callback is not a function");o=(o==null?TZe():+o)+(r==null?0:+r),!this._next&&Fre!==this&&(Fre?Fre._next=this:wwe=this,Fre=this),this._call=n,this._time=o,cWe()},stop:function(){this._call&&(this._call=null,this._time=1/0,cWe())}};function Csn(n,r,o){var s=new Ewe;return s.restart(n,r,o),s}function oRr(){TZe(),++iK;for(var n=wwe,r;n;)(r=tU-n._time)>=0&&n._call.call(void 0,r),n=n._next;--iK}function Y$t(){tU=(ywe=toe.now())+dEe,iK=$re=0;try{oRr()}finally{iK=0,sRr(),tU=0}}function aRr(){var n=toe.now(),r=n-ywe;r>Tsn&&(dEe-=r,ywe=n)}function sRr(){for(var n,r=wwe,o,s=1/0;r;)r._call?(s>r._time&&(s=r._time),n=r,r=r._next):(o=r._next,r._next=null,r=n?n._next=o:wwe=o);Fre=n,cWe(s)}function cWe(n){if(!iK){$re&&($re=clearTimeout($re));var r=n-tU;r>24?(n<1/0&&($re=setTimeout(Y$t,n-toe.now()-dEe)),pre&&(pre=clearInterval(pre))):(pre||(ywe=toe.now(),pre=setInterval(aRr,Tsn)),iK=1,xsn(Y$t))}}function X$t(n,r,o){var s=new Ewe;return r=r==null?0:+r,s.restart(c=>{s.stop(),n(c+r)},r,o),s}var cRr=uEe("start","end","cancel","interrupt"),uRr=[],Asn=0,J$t=1,uWe=2,wve=3,Z$t=4,lWe=5,yve=6;function fEe(n,r,o,s,c,d){var h=n.__transition;if(!h)n.__transition={};else if(o in h)return;lRr(n,o,{name:r,index:s,group:c,on:cRr,tween:uRr,time:d.time,delay:d.delay,duration:d.duration,ease:d.ease,timer:null,state:Asn})}function xZe(n,r){var o=tC(n,r);if(o.state>Asn)throw new Error("too late; already scheduled");return o}function VI(n,r){var o=tC(n,r);if(o.state>wve)throw new Error("too late; already running");return o}function tC(n,r){var o=n.__transition;if(!o||!(o=o[r]))throw new Error("transition not found");return o}function lRr(n,r,o){var s=n.__transition,c;s[r]=o,o.timer=Csn(d,0,o.time);function d(E){o.state=J$t,o.timer.restart(h,o.delay,o.time),o.delay<=E&&h(E-o.delay)}function h(E){var k,x,O,P;if(o.state!==J$t)return y();for(k in s)if(P=s[k],P.name===o.name){if(P.state===wve)return X$t(h);P.state===Z$t?(P.state=yve,P.timer.stop(),P.on.call("interrupt",n,n.__data__,P.index,P.group),delete s[k]):+k<r&&(P.state=yve,P.timer.stop(),P.on.call("cancel",n,n.__data__,P.index,P.group),delete s[k])}if(X$t(function(){o.state===wve&&(o.state=Z$t,o.timer.restart(b,o.delay,o.time),b(E))}),o.state=uWe,o.on.call("start",n,n.__data__,o.index,o.group),o.state===uWe){for(o.state=wve,c=new Array(O=o.tween.length),k=0,x=-1;k<O;++k)(P=o.tween[k].value.call(n,n.__data__,o.index,o.group))&&(c[++x]=P);c.length=x+1}}function b(E){for(var k=E<o.duration?o.ease.call(null,E/o.duration):(o.timer.restart(y),o.state=lWe,1),x=-1,O=c.length;++x<O;)c[x].call(n,k);o.state===lWe&&(o.on.call("end",n,n.__data__,o.index,o.group),y())}function y(){o.state=yve,o.timer.stop(),delete s[r];for(var E in s)return;delete n.__transition}}function Eve(n,r){var o=n.__transition,s,c,d=!0,h;if(o){r=r==null?null:r+"";for(h in o){if((s=o[h]).name!==r){d=!1;continue}c=s.state>uWe&&s.state<lWe,s.state=yve,s.timer.stop(),s.on.call(c?"interrupt":"cancel",n,n.__data__,s.index,s.group),delete o[h]}d&&delete n.__transition}}function dRr(n){return this.each(function(){Eve(this,n)})}function fRr(n,r){var o,s;return function(){var c=VI(this,n),d=c.tween;if(d!==o){s=o=d;for(var h=0,b=s.length;h<b;++h)if(s[h].name===r){s=s.slice(),s.splice(h,1);break}}c.tween=s}}function hRr(n,r,o){var s,c;if(typeof o!="function")throw new Error;return function(){var d=VI(this,n),h=d.tween;if(h!==s){c=(s=h).slice();for(var b={name:r,value:o},y=0,E=c.length;y<E;++y)if(c[y].name===r){c[y]=b;break}y===E&&c.push(b)}d.tween=c}}function pRr(n,r){var o=this._id;if(n+="",arguments.length<2){for(var s=tC(this.node(),o).tween,c=0,d=s.length,h;c<d;++c)if((h=s[c]).name===n)return h.value;return null}return this.each((r==null?fRr:hRr)(o,n,r))}function CZe(n,r,o){var s=n._id;return n.each(function(){var c=VI(this,s);(c.value||(c.value={}))[r]=o.apply(this,arguments)}),function(c){return tC(c,s).value[r]}}function Isn(n,r){var o;return(typeof r=="number"?iM:r instanceof eoe?V$t:(o=eoe(r))?(r=o,V$t):YNr)(n,r)}function gRr(n){return function(){this.removeAttribute(n)}}function mRr(n){return function(){this.removeAttributeNS(n.space,n.local)}}function bRr(n,r,o){var s,c=o+"",d;return function(){var h=this.getAttribute(n);return h===c?null:h===s?d:d=r(s=h,o)}}function vRr(n,r,o){var s,c=o+"",d;return function(){var h=this.getAttributeNS(n.space,n.local);return h===c?null:h===s?d:d=r(s=h,o)}}function wRr(n,r,o){var s,c,d;return function(){var h,b=o(this),y;return b==null?void this.removeAttribute(n):(h=this.getAttribute(n),y=b+"",h===y?null:h===s&&y===c?d:(c=y,d=r(s=h,b)))}}function yRr(n,r,o){var s,c,d;return function(){var h,b=o(this),y;return b==null?void this.removeAttributeNS(n.space,n.local):(h=this.getAttributeNS(n.space,n.local),y=b+"",h===y?null:h===s&&y===c?d:(c=y,d=r(s=h,b)))}}function ERr(n,r){var o=lEe(n),s=o==="transform"?QNr:Isn;return this.attrTween(n,typeof r=="function"?(o.local?yRr:wRr)(o,s,CZe(this,"attr."+n,r)):r==null?(o.local?mRr:gRr)(o):(o.local?vRr:bRr)(o,s,r))}function SRr(n,r){return function(o){this.setAttribute(n,r.call(this,o))}}function _Rr(n,r){return function(o){this.setAttributeNS(n.space,n.local,r.call(this,o))}}function kRr(n,r){var o,s;function c(){var d=r.apply(this,arguments);return d!==s&&(o=(s=d)&&_Rr(n,d)),o}return c._value=r,c}function TRr(n,r){var o,s;function c(){var d=r.apply(this,arguments);return d!==s&&(o=(s=d)&&SRr(n,d)),o}return c._value=r,c}function xRr(n,r){var o="attr."+n;if(arguments.length<2)return(o=this.tween(o))&&o._value;if(r==null)return this.tween(o,null);if(typeof r!="function")throw new Error;var s=lEe(n);return this.tween(o,(s.local?kRr:TRr)(s,r))}function CRr(n,r){return function(){xZe(this,n).delay=+r.apply(this,arguments)}}function ARr(n,r){return r=+r,function(){xZe(this,n).delay=r}}function IRr(n){var r=this._id;return arguments.length?this.each((typeof n=="function"?CRr:ARr)(r,n)):tC(this.node(),r).delay}function ORr(n,r){return function(){VI(this,n).duration=+r.apply(this,arguments)}}function NRr(n,r){return r=+r,function(){VI(this,n).duration=r}}function RRr(n){var r=this._id;return arguments.length?this.each((typeof n=="function"?ORr:NRr)(r,n)):tC(this.node(),r).duration}function PRr(n,r){if(typeof r!="function")throw new Error;return function(){VI(this,n).ease=r}}function LRr(n){var r=this._id;return arguments.length?this.each(PRr(r,n)):tC(this.node(),r).ease}function DRr(n,r){return function(){var o=r.apply(this,arguments);if(typeof o!="function")throw new Error;VI(this,n).ease=o}}function MRr(n){if(typeof n!="function")throw new Error;return this.each(DRr(this._id,n))}function jRr(n){typeof n!="function"&&(n=ssn(n));for(var r=this._groups,o=r.length,s=new Array(o),c=0;c<o;++c)for(var d=r[c],h=d.length,b=s[c]=[],y,E=0;E<h;++E)(y=d[E])&&n.call(y,y.__data__,E,d)&&b.push(y);return new V4(s,this._parents,this._name,this._id)}function $Rr(n){if(n._id!==this._id)throw new Error;for(var r=this._groups,o=n._groups,s=r.length,c=o.length,d=Math.min(s,c),h=new Array(s),b=0;b<d;++b)for(var y=r[b],E=o[b],k=y.length,x=h[b]=new Array(k),O,P=0;P<k;++P)(O=y[P]||E[P])&&(x[P]=O);for(;b<s;++b)h[b]=r[b];return new V4(h,this._parents,this._name,this._id)}function FRr(n){return(n+"").trim().split(/^|\s+/).every(function(r){var o=r.indexOf(".");return o>=0&&(r=r.slice(0,o)),!r||r==="start"})}function BRr(n,r,o){var s,c,d=FRr(r)?xZe:VI;return function(){var h=d(this,n),b=h.on;b!==s&&(c=(s=b).copy()).on(r,o),h.on=c}}function URr(n,r){var o=this._id;return arguments.length<2?tC(this.node(),o).on.on(n):this.each(BRr(o,n,r))}function zRr(n){return function(){var r=this.parentNode;for(var o in this.__transition)if(+o!==n)return;r&&r.removeChild(this)}}function HRr(){return this.on("end.remove",zRr(this._id))}function GRr(n){var r=this._name,o=this._id;typeof n!="function"&&(n=SZe(n));for(var s=this._groups,c=s.length,d=new Array(c),h=0;h<c;++h)for(var b=s[h],y=b.length,E=d[h]=new Array(y),k,x,O=0;O<y;++O)(k=b[O])&&(x=n.call(k,k.__data__,O,b))&&("__data__"in k&&(x.__data__=k.__data__),E[O]=x,fEe(E[O],r,o,O,E,tC(k,o)));return new V4(d,this._parents,r,o)}function qRr(n){var r=this._name,o=this._id;typeof n!="function"&&(n=asn(n));for(var s=this._groups,c=s.length,d=[],h=[],b=0;b<c;++b)for(var y=s[b],E=y.length,k,x=0;x<E;++x)if(k=y[x]){for(var O=n.call(k,k.__data__,x,y),P,D=tC(k,o),$=0,U=O.length;$<U;++$)(P=O[$])&&fEe(P,r,o,$,O,D);d.push(O),h.push(k)}return new V4(d,h,r,o)}var VRr=Hoe.prototype.constructor;function WRr(){return new VRr(this._groups,this._parents)}function KRr(n,r){var o,s,c;return function(){var d=rK(this,n),h=(this.style.removeProperty(n),rK(this,n));return d===h?null:d===o&&h===s?c:c=r(o=d,s=h)}}function Osn(n){return function(){this.style.removeProperty(n)}}function YRr(n,r,o){var s,c=o+"",d;return function(){var h=rK(this,n);return h===c?null:h===s?d:d=r(s=h,o)}}function XRr(n,r,o){var s,c,d;return function(){var h=rK(this,n),b=o(this),y=b+"";return b==null&&(y=b=(this.style.removeProperty(n),rK(this,n))),h===y?null:h===s&&y===c?d:(c=y,d=r(s=h,b))}}function JRr(n,r){var o,s,c,d="style."+r,h="end."+d,b;return function(){var y=VI(this,n),E=y.on,k=y.value[d]==null?b||(b=Osn(r)):void 0;(E!==o||c!==k)&&(s=(o=E).copy()).on(h,c=k),y.on=s}}function ZRr(n,r,o){var s=(n+="")=="transform"?ZNr:Isn;return r==null?this.styleTween(n,KRr(n,s)).on("end.style."+n,Osn(n)):typeof r=="function"?this.styleTween(n,XRr(n,s,CZe(this,"style."+n,r))).each(JRr(this._id,n)):this.styleTween(n,YRr(n,s,r),o).on("end.style."+n,null)}function QRr(n,r,o){return function(s){this.style.setProperty(n,r.call(this,s),o)}}function e4r(n,r,o){var s,c;function d(){var h=r.apply(this,arguments);return h!==c&&(s=(c=h)&&QRr(n,h,o)),s}return d._value=r,d}function t4r(n,r,o){var s="style."+(n+="");if(arguments.length<2)return(s=this.tween(s))&&s._value;if(r==null)return this.tween(s,null);if(typeof r!="function")throw new Error;return this.tween(s,e4r(n,r,o??""))}function n4r(n){return function(){this.textContent=n}}function r4r(n){return function(){var r=n(this);this.textContent=r??""}}function i4r(n){return this.tween("text",typeof n=="function"?r4r(CZe(this,"text",n)):n4r(n==null?"":n+""))}function o4r(n){return function(r){this.textContent=n.call(this,r)}}function a4r(n){var r,o;function s(){var c=n.apply(this,arguments);return c!==o&&(r=(o=c)&&o4r(c)),r}return s._value=n,s}function s4r(n){var r="text";if(arguments.length<1)return(r=this.tween(r))&&r._value;if(n==null)return this.tween(r,null);if(typeof n!="function")throw new Error;return this.tween(r,a4r(n))}function c4r(){for(var n=this._name,r=this._id,o=Nsn(),s=this._groups,c=s.length,d=0;d<c;++d)for(var h=s[d],b=h.length,y,E=0;E<b;++E)if(y=h[E]){var k=tC(y,r);fEe(y,n,o,E,h,{time:k.time+k.delay+k.duration,delay:0,duration:k.duration,ease:k.ease})}return new V4(s,this._parents,n,o)}function u4r(){var n,r,o=this,s=o._id,c=o.size();return new Promise(function(d,h){var b={value:h},y={value:function(){--c===0&&d()}};o.each(function(){var E=VI(this,s),k=E.on;k!==n&&(r=(n=k).copy(),r._.cancel.push(b),r._.interrupt.push(b),r._.end.push(y)),E.on=r}),c===0&&d()})}var l4r=0;function V4(n,r,o,s){this._groups=n,this._parents=r,this._name=o,this._id=s}function Nsn(){return++l4r}var y4=Hoe.prototype;V4.prototype={constructor:V4,select:GRr,selectAll:qRr,selectChild:y4.selectChild,selectChildren:y4.selectChildren,filter:jRr,merge:$Rr,selection:WRr,transition:c4r,call:y4.call,nodes:y4.nodes,node:y4.node,size:y4.size,empty:y4.empty,each:y4.each,on:URr,attr:ERr,attrTween:xRr,style:ZRr,styleTween:t4r,text:i4r,textTween:s4r,remove:HRr,tween:pRr,delay:IRr,duration:RRr,ease:LRr,easeVarying:MRr,end:u4r,[Symbol.iterator]:y4[Symbol.iterator]};function d4r(n){return((n*=2)<=1?n*n*n:(n-=2)*n*n+2)/2}var f4r={time:null,delay:0,duration:250,ease:d4r};function h4r(n,r){for(var o;!(o=n.__transition)||!(o=o[r]);)if(!(n=n.parentNode))throw new Error(`transition ${r} not found`);return o}function p4r(n){var r,o;n instanceof V4?(r=n._id,n=n._name):(r=Nsn(),(o=f4r).time=TZe(),n=n==null?null:n+"");for(var s=this._groups,c=s.length,d=0;d<c;++d)for(var h=s[d],b=h.length,y,E=0;E<b;++E)(y=h[E])&&fEe(y,n,r,E,h,o||h4r(y,r));return new V4(s,this._parents,n,r)}Hoe.prototype.interrupt=dRr;Hoe.prototype.transition=p4r;const v0e=n=>()=>n;function g4r(n,{sourceEvent:r,target:o,transform:s,dispatch:c}){Object.defineProperties(this,{type:{value:n,enumerable:!0,configurable:!0},sourceEvent:{value:r,enumerable:!0,configurable:!0},target:{value:o,enumerable:!0,configurable:!0},transform:{value:s,enumerable:!0,configurable:!0},_:{value:c}})}function M4(n,r,o){this.k=n,this.x=r,this.y=o}M4.prototype={constructor:M4,scale:function(n){return n===1?this:new M4(this.k*n,this.x,this.y)},translate:function(n,r){return n===0&r===0?this:new M4(this.k,this.x+this.k*n,this.y+this.k*r)},apply:function(n){return[n[0]*this.k+this.x,n[1]*this.k+this.y]},applyX:function(n){return n*this.k+this.x},applyY:function(n){return n*this.k+this.y},invert:function(n){return[(n[0]-this.x)/this.k,(n[1]-this.y)/this.k]},invertX:function(n){return(n-this.x)/this.k},invertY:function(n){return(n-this.y)/this.k},rescaleX:function(n){return n.copy().domain(n.range().map(this.invertX,this).map(n.invert,n))},rescaleY:function(n){return n.copy().domain(n.range().map(this.invertY,this).map(n.invert,n))},toString:function(){return"translate("+this.x+","+this.y+") scale("+this.k+")"}};var mM=new M4(1,0,0);M4.prototype;function z$e(n){n.stopImmediatePropagation()}function gre(n){n.preventDefault(),n.stopImmediatePropagation()}function m4r(n){return(!n.ctrlKey||n.type==="wheel")&&!n.button}function b4r(){var n=this;return n instanceof SVGElement?(n=n.ownerSVGElement||n,n.hasAttribute("viewBox")?(n=n.viewBox.baseVal,[[n.x,n.y],[n.x+n.width,n.y+n.height]]):[[0,0],[n.width.baseVal.value,n.height.baseVal.value]]):[[0,0],[n.clientWidth,n.clientHeight]]}function Q$t(){return this.__zoom||mM}function v4r(n){return-n.deltaY*(n.deltaMode===1?.05:n.deltaMode?1:.002)*(n.ctrlKey?10:1)}function w4r(){return navigator.maxTouchPoints||"ontouchstart"in this}function y4r(n,r,o){var s=n.invertX(r[0][0])-o[0][0],c=n.invertX(r[1][0])-o[1][0],d=n.invertY(r[0][1])-o[0][1],h=n.invertY(r[1][1])-o[1][1];return n.translate(c>s?(s+c)/2:Math.min(0,s)||Math.max(0,c),h>d?(d+h)/2:Math.min(0,d)||Math.max(0,h))}function E4r(){var n=m4r,r=b4r,o=y4r,s=v4r,c=w4r,d=[0,1/0],h=[[-1/0,-1/0],[1/0,1/0]],b=250,y=rRr,E=uEe("start","zoom","end"),k,x,O,P=500,D=150,$=0,U=10;function B(Z){Z.property("__zoom",Q$t).on("wheel.zoom",ye,{passive:!1}).on("mousedown.zoom",we).on("dblclick.zoom",ge).filter(c).on("touchstart.zoom",Te).on("touchmove.zoom",xe).on("touchend.zoom touchcancel.zoom",Ue).style("-webkit-tap-highlight-color","rgba(0,0,0,0)")}B.transform=function(Z,Re,_e,qe){var it=Z.selection?Z.selection():Z;it.property("__zoom",Q$t),Z!==it?ae(Z,Re,_e,qe):it.interrupt().each(function(){oe(this,arguments).event(qe).start().zoom(null,typeof Re=="function"?Re.apply(this,arguments):Re).end()})},B.scaleBy=function(Z,Re,_e,qe){B.scaleTo(Z,function(){var it=this.__zoom.k,Xe=typeof Re=="function"?Re.apply(this,arguments):Re;return it*Xe},_e,qe)},B.scaleTo=function(Z,Re,_e,qe){B.transform(Z,function(){var it=r.apply(this,arguments),Xe=this.__zoom,de=_e==null?X(it):typeof _e=="function"?_e.apply(this,arguments):_e,We=Xe.invert(de),ht=typeof Re=="function"?Re.apply(this,arguments):Re;return o(K(G(Xe,ht),de,We),it,h)},_e,qe)},B.translateBy=function(Z,Re,_e,qe){B.transform(Z,function(){return o(this.__zoom.translate(typeof Re=="function"?Re.apply(this,arguments):Re,typeof _e=="function"?_e.apply(this,arguments):_e),r.apply(this,arguments),h)},null,qe)},B.translateTo=function(Z,Re,_e,qe,it){B.transform(Z,function(){var Xe=r.apply(this,arguments),de=this.__zoom,We=qe==null?X(Xe):typeof qe=="function"?qe.apply(this,arguments):qe;return o(mM.translate(We[0],We[1]).scale(de.k).translate(typeof Re=="function"?-Re.apply(this,arguments):-Re,typeof _e=="function"?-_e.apply(this,arguments):-_e),Xe,h)},qe,it)};function G(Z,Re){return Re=Math.max(d[0],Math.min(d[1],Re)),Re===Z.k?Z:new M4(Re,Z.x,Z.y)}function K(Z,Re,_e){var qe=Re[0]-_e[0]*Z.k,it=Re[1]-_e[1]*Z.k;return qe===Z.x&&it===Z.y?Z:new M4(Z.k,qe,it)}function X(Z){return[(+Z[0][0]+ +Z[1][0])/2,(+Z[0][1]+ +Z[1][1])/2]}function ae(Z,Re,_e,qe){Z.on("start.zoom",function(){oe(this,arguments).event(qe).start()}).on("interrupt.zoom end.zoom",function(){oe(this,arguments).event(qe).end()}).tween("zoom",function(){var it=this,Xe=arguments,de=oe(it,Xe).event(qe),We=r.apply(it,Xe),ht=_e==null?X(We):typeof _e=="function"?_e.apply(it,Xe):_e,Ie=Math.max(We[1][0]-We[0][0],We[1][1]-We[0][1]),$t=it.__zoom,An=typeof Re=="function"?Re.apply(it,Xe):Re,Yt=y($t.invert(ht).concat(Ie/$t.k),An.invert(ht).concat(Ie/An.k));return function(Or){if(Or===1)Or=An;else{var Nt=Yt(Or),on=Ie/Nt[2];Or=new M4(on,ht[0]-Nt[0]*on,ht[1]-Nt[1]*on)}de.zoom(null,Or)}})}function oe(Z,Re,_e){return!_e&&Z.__zooming||new he(Z,Re)}function he(Z,Re){this.that=Z,this.args=Re,this.active=0,this.sourceEvent=null,this.extent=r.apply(Z,Re),this.taps=0}he.prototype={event:function(Z){return Z&&(this.sourceEvent=Z),this},start:function(){return++this.active===1&&(this.that.__zooming=this,this.emit("start")),this},zoom:function(Z,Re){return this.mouse&&Z!=="mouse"&&(this.mouse[1]=Re.invert(this.mouse[0])),this.touch0&&Z!=="touch"&&(this.touch0[1]=Re.invert(this.touch0[0])),this.touch1&&Z!=="touch"&&(this.touch1[1]=Re.invert(this.touch1[0])),this.that.__zoom=Re,this.emit("zoom"),this},end:function(){return--this.active===0&&(delete this.that.__zooming,this.emit("end")),this},emit:function(Z){var Re=Rx(this.that).datum();E.call(Z,this.that,new g4r(Z,{sourceEvent:this.sourceEvent,target:B,transform:this.that.__zoom,dispatch:E}),Re)}};function ye(Z,...Re){if(!n.apply(this,arguments))return;var _e=oe(this,Re).event(Z),qe=this.__zoom,it=Math.max(d[0],Math.min(d[1],qe.k*Math.pow(2,s.apply(this,arguments)))),Xe=sI(Z);if(_e.wheel)(_e.mouse[0][0]!==Xe[0]||_e.mouse[0][1]!==Xe[1])&&(_e.mouse[1]=qe.invert(_e.mouse[0]=Xe)),clearTimeout(_e.wheel);else{if(qe.k===it)return;_e.mouse=[Xe,qe.invert(Xe)],Eve(this),_e.start()}gre(Z),_e.wheel=setTimeout(de,D),_e.zoom("mouse",o(K(G(qe,it),_e.mouse[0],_e.mouse[1]),_e.extent,h));function de(){_e.wheel=null,_e.end()}}function we(Z,...Re){if(O||!n.apply(this,arguments))return;var _e=Z.currentTarget,qe=oe(this,Re,!0).event(Z),it=Rx(Z.view).on("mousemove.zoom",ht,!0).on("mouseup.zoom",Ie,!0),Xe=sI(Z,_e),de=Z.clientX,We=Z.clientY;bsn(Z.view),z$e(Z),qe.mouse=[Xe,this.__zoom.invert(Xe)],Eve(this),qe.start();function ht($t){if(gre($t),!qe.moved){var An=$t.clientX-de,Yt=$t.clientY-We;qe.moved=An*An+Yt*Yt>$}qe.event($t).zoom("mouse",o(K(qe.that.__zoom,qe.mouse[0]=sI($t,_e),qe.mouse[1]),qe.extent,h))}function Ie($t){it.on("mousemove.zoom mouseup.zoom",null),vsn($t.view,qe.moved),gre($t),qe.event($t).end()}}function ge(Z,...Re){if(n.apply(this,arguments)){var _e=this.__zoom,qe=sI(Z.changedTouches?Z.changedTouches[0]:Z,this),it=_e.invert(qe),Xe=_e.k*(Z.shiftKey?.5:2),de=o(K(G(_e,Xe),qe,it),r.apply(this,Re),h);gre(Z),b>0?Rx(this).transition().duration(b).call(ae,de,qe,Z):Rx(this).call(B.transform,de,qe,Z)}}function Te(Z,...Re){if(n.apply(this,arguments)){var _e=Z.touches,qe=_e.length,it=oe(this,Re,Z.changedTouches.length===qe).event(Z),Xe,de,We,ht;for(z$e(Z),de=0;de<qe;++de)We=_e[de],ht=sI(We,this),ht=[ht,this.__zoom.invert(ht),We.identifier],it.touch0?!it.touch1&&it.touch0[2]!==ht[2]&&(it.touch1=ht,it.taps=0):(it.touch0=ht,Xe=!0,it.taps=1+!!k);k&&(k=clearTimeout(k)),Xe&&(it.taps<2&&(x=ht[0],k=setTimeout(function(){k=null},P)),Eve(this),it.start())}}function xe(Z,...Re){if(this.__zooming){var _e=oe(this,Re).event(Z),qe=Z.changedTouches,it=qe.length,Xe,de,We,ht;for(gre(Z),Xe=0;Xe<it;++Xe)de=qe[Xe],We=sI(de,this),_e.touch0&&_e.touch0[2]===de.identifier?_e.touch0[0]=We:_e.touch1&&_e.touch1[2]===de.identifier&&(_e.touch1[0]=We);if(de=_e.that.__zoom,_e.touch1){var Ie=_e.touch0[0],$t=_e.touch0[1],An=_e.touch1[0],Yt=_e.touch1[1],Or=(Or=An[0]-Ie[0])*Or+(Or=An[1]-Ie[1])*Or,Nt=(Nt=Yt[0]-$t[0])*Nt+(Nt=Yt[1]-$t[1])*Nt;de=G(de,Math.sqrt(Or/Nt)),We=[(Ie[0]+An[0])/2,(Ie[1]+An[1])/2],ht=[($t[0]+Yt[0])/2,($t[1]+Yt[1])/2]}else if(_e.touch0)We=_e.touch0[0],ht=_e.touch0[1];else return;_e.zoom("touch",o(K(de,We,ht),_e.extent,h))}}function Ue(Z,...Re){if(this.__zooming){var _e=oe(this,Re).event(Z),qe=Z.changedTouches,it=qe.length,Xe,de;for(z$e(Z),O&&clearTimeout(O),O=setTimeout(function(){O=null},P),Xe=0;Xe<it;++Xe)de=qe[Xe],_e.touch0&&_e.touch0[2]===de.identifier?delete _e.touch0:_e.touch1&&_e.touch1[2]===de.identifier&&delete _e.touch1;if(_e.touch1&&!_e.touch0&&(_e.touch0=_e.touch1,delete _e.touch1),_e.touch0)_e.touch0[1]=this.__zoom.invert(_e.touch0[0]);else if(_e.end(),_e.taps===2&&(de=sI(de,this),Math.hypot(x[0]-de[0],x[1]-de[1])<U)){var We=Rx(this).on("dblclick.zoom");We&&We.apply(this,arguments)}}}return B.wheelDelta=function(Z){return arguments.length?(s=typeof Z=="function"?Z:v0e(+Z),B):s},B.filter=function(Z){return arguments.length?(n=typeof Z=="function"?Z:v0e(!!Z),B):n},B.touchable=function(Z){return arguments.length?(c=typeof Z=="function"?Z:v0e(!!Z),B):c},B.extent=function(Z){return arguments.length?(r=typeof Z=="function"?Z:v0e([[+Z[0][0],+Z[0][1]],[+Z[1][0],+Z[1][1]]]),B):r},B.scaleExtent=function(Z){return arguments.length?(d[0]=+Z[0],d[1]=+Z[1],B):[d[0],d[1]]},B.translateExtent=function(Z){return arguments.length?(h[0][0]=+Z[0][0],h[1][0]=+Z[1][0],h[0][1]=+Z[0][1],h[1][1]=+Z[1][1],B):[[h[0][0],h[0][1]],[h[1][0],h[1][1]]]},B.constrain=function(Z){return arguments.length?(o=Z,B):o},B.duration=function(Z){return arguments.length?(b=+Z,B):b},B.interpolate=function(Z){return arguments.length?(y=Z,B):y},B.on=function(){var Z=E.on.apply(E,arguments);return Z===E?B:Z},B.clickDistance=function(Z){return arguments.length?($=(Z=+Z)*Z,B):Math.sqrt($)},B.tapDistance=function(Z){return arguments.length?(U=+Z,B):U},B}const hEe=Q.createContext(null),S4r=hEe.Provider,W4={error001:()=>"[React Flow]: Seems like you have not used zustand provider as an ancestor. Help: https://reactflow.dev/error#001",error002:()=>"It looks like you've created a new nodeTypes or edgeTypes object. If this wasn't on purpose please define the nodeTypes/edgeTypes outside of the component or memoize them.",error003:n=>`Node type "${n}" not found. Using fallback type "default".`,error004:()=>"The React Flow parent container needs a width and a height to render the graph.",error005:()=>"Only child nodes can use a parent extent.",error006:()=>"Can't create edge. An edge needs a source and a target.",error007:n=>`The old edge with id=${n} does not exist.`,error009:n=>`Marker type "${n}" doesn't exist.`,error008:(n,r)=>`Couldn't create edge for ${n?"target":"source"} handle id: "${n?r.targetHandle:r.sourceHandle}", edge id: ${r.id}.`,error010:()=>"Handle: No node id found. Make sure to only use a Handle inside a custom Node.",error011:n=>`Edge type "${n}" not found. Using fallback type "default".`,error012:n=>`Node with id "${n}" does not exist, it may have been removed. This can happen when a node is deleted before the "onNodeClick" handler is called.`},Rsn=W4.error001();function Nh(n,r){const o=Q.useContext(hEe);if(o===null)throw new Error(Rsn);return isn(o,n,r)}const Wb=()=>{const n=Q.useContext(hEe);if(n===null)throw new Error(Rsn);return Q.useMemo(()=>({getState:n.getState,setState:n.setState,subscribe:n.subscribe,destroy:n.destroy}),[n])},_4r=n=>n.userSelectionActive?"none":"all";function k4r({position:n,children:r,className:o,style:s,...c}){const d=Nh(_4r),h=`${n}`.split("-");return Un.createElement("div",{className:q_(["react-flow__panel",o,...h]),style:{...s,pointerEvents:d},...c},r)}function T4r({proOptions:n,position:r="bottom-right"}){return n?.hideAttribution?null:Un.createElement(k4r,{position:r,className:"react-flow__attribution","data-message":"Please only hide this attribution when you are subscribed to React Flow Pro: https://reactflow.dev/pro"},Un.createElement("a",{href:"https://reactflow.dev",target:"_blank",rel:"noopener noreferrer","aria-label":"React Flow attribution"},"React Flow"))}const x4r=({x:n,y:r,label:o,labelStyle:s={},labelShowBg:c=!0,labelBgStyle:d={},labelBgPadding:h=[2,4],labelBgBorderRadius:b=2,children:y,className:E,...k})=>{const x=Q.useRef(null),[O,P]=Q.useState({x:0,y:0,width:0,height:0}),D=q_(["react-flow__edge-textwrapper",E]);return Q.useEffect(()=>{if(x.current){const $=x.current.getBBox();P({x:$.x,y:$.y,width:$.width,height:$.height})}},[o]),typeof o>"u"||!o?null:Un.createElement("g",{transform:`translate(${n-O.width/2} ${r-O.height/2})`,className:D,visibility:O.width?"visible":"hidden",...k},c&&Un.createElement("rect",{width:O.width+2*h[0],x:-h[0],y:-h[1],height:O.height+2*h[1],className:"react-flow__edge-textbg",style:d,rx:b,ry:b}),Un.createElement("text",{className:"react-flow__edge-text",y:O.height/2,dy:"0.3em",ref:x,style:s},o),y)};var C4r=Q.memo(x4r);const AZe=n=>({width:n.offsetWidth,height:n.offsetHeight}),oK=(n,r=0,o=1)=>Math.min(Math.max(n,r),o),IZe=(n={x:0,y:0},r)=>({x:oK(n.x,r[0][0],r[1][0]),y:oK(n.y,r[0][1],r[1][1])}),eFt=(n,r,o)=>n<r?oK(Math.abs(n-r),1,50)/50:n>o?-oK(Math.abs(n-o),1,50)/50:0,Psn=(n,r)=>{const o=eFt(n.x,35,r.width-35)*20,s=eFt(n.y,35,r.height-35)*20;return[o,s]},Lsn=n=>n.getRootNode?.()||window?.document,A4r=(n,r)=>({x:Math.min(n.x,r.x),y:Math.min(n.y,r.y),x2:Math.max(n.x2,r.x2),y2:Math.max(n.y2,r.y2)}),OZe=({x:n,y:r,width:o,height:s})=>({x:n,y:r,x2:n+o,y2:r+s}),I4r=({x:n,y:r,x2:o,y2:s})=>({x:n,y:r,width:o-n,height:s-r}),tFt=n=>({...n.positionAbsolute||{x:0,y:0},width:n.width||0,height:n.height||0}),dWe=(n,r)=>{const o=Math.max(0,Math.min(n.x+n.width,r.x+r.width)-Math.max(n.x,r.x)),s=Math.max(0,Math.min(n.y+n.height,r.y+r.height)-Math.max(n.y,r.y));return Math.ceil(o*s)},O4r=n=>M_(n.width)&&M_(n.height)&&M_(n.x)&&M_(n.y),M_=n=>!isNaN(n)&&isFinite(n),Oh=Symbol.for("internals"),Dsn=["Enter"," ","Escape"],N4r=(n,r)=>{},R4r=n=>"nativeEvent"in n;function fWe(n){const o=(R4r(n)?n.nativeEvent:n).composedPath?.()?.[0]||n.target;return["INPUT","SELECT","TEXTAREA"].includes(o?.nodeName)||o?.hasAttribute("contenteditable")||!!o?.closest(".nokey")}const Msn=n=>"clientX"in n,bM=(n,r)=>{const o=Msn(n),s=o?n.clientX:n.touches?.[0].clientX,c=o?n.clientY:n.touches?.[0].clientY;return{x:s-(r?.left??0),y:c-(r?.top??0)}},Swe=()=>typeof navigator<"u"&&navigator?.userAgent?.indexOf("Mac")>=0,mU=({id:n,path:r,labelX:o,labelY:s,label:c,labelStyle:d,labelShowBg:h,labelBgStyle:b,labelBgPadding:y,labelBgBorderRadius:E,style:k,markerEnd:x,markerStart:O,interactionWidth:P=20})=>Un.createElement(Un.Fragment,null,Un.createElement("path",{id:n,style:k,d:r,fill:"none",className:"react-flow__edge-path",markerEnd:x,markerStart:O}),P&&Un.createElement("path",{d:r,fill:"none",strokeOpacity:0,strokeWidth:P,className:"react-flow__edge-interaction"}),c&&M_(o)&&M_(s)?Un.createElement(C4r,{x:o,y:s,label:c,labelStyle:d,labelShowBg:h,labelBgStyle:b,labelBgPadding:y,labelBgBorderRadius:E}):null);mU.displayName="BaseEdge";function mre(n,r,o){return o===void 0?o:s=>{const c=r().edges.find(d=>d.id===n);c&&o(s,{...c})}}function jsn({sourceX:n,sourceY:r,targetX:o,targetY:s}){const c=Math.abs(o-n)/2,d=o<n?o+c:o-c,h=Math.abs(s-r)/2,b=s<r?s+h:s-h;return[d,b,c,h]}function $sn({sourceX:n,sourceY:r,targetX:o,targetY:s,sourceControlX:c,sourceControlY:d,targetControlX:h,targetControlY:b}){const y=n*.125+c*.375+h*.375+o*.125,E=r*.125+d*.375+b*.375+s*.125,k=Math.abs(y-n),x=Math.abs(E-r);return[y,E,k,x]}var nU;(function(n){n.Strict="strict",n.Loose="loose"})(nU||(nU={}));var MB;(function(n){n.Free="free",n.Vertical="vertical",n.Horizontal="horizontal"})(MB||(MB={}));var noe;(function(n){n.Partial="partial",n.Full="full"})(noe||(noe={}));var aM;(function(n){n.Bezier="default",n.Straight="straight",n.Step="step",n.SmoothStep="smoothstep",n.SimpleBezier="simplebezier"})(aM||(aM={}));var roe;(function(n){n.Arrow="arrow",n.ArrowClosed="arrowclosed"})(roe||(roe={}));var ni;(function(n){n.Left="left",n.Top="top",n.Right="right",n.Bottom="bottom"})(ni||(ni={}));function nFt({pos:n,x1:r,y1:o,x2:s,y2:c}){return n===ni.Left||n===ni.Right?[.5*(r+s),o]:[r,.5*(o+c)]}function Fsn({sourceX:n,sourceY:r,sourcePosition:o=ni.Bottom,targetX:s,targetY:c,targetPosition:d=ni.Top}){const[h,b]=nFt({pos:o,x1:n,y1:r,x2:s,y2:c}),[y,E]=nFt({pos:d,x1:s,y1:c,x2:n,y2:r}),[k,x,O,P]=$sn({sourceX:n,sourceY:r,targetX:s,targetY:c,sourceControlX:h,sourceControlY:b,targetControlX:y,targetControlY:E});return[`M${n},${r} C${h},${b} ${y},${E} ${s},${c}`,k,x,O,P]}const NZe=Q.memo(({sourceX:n,sourceY:r,targetX:o,targetY:s,sourcePosition:c=ni.Bottom,targetPosition:d=ni.Top,label:h,labelStyle:b,labelShowBg:y,labelBgStyle:E,labelBgPadding:k,labelBgBorderRadius:x,style:O,markerEnd:P,markerStart:D,interactionWidth:$})=>{const[U,B,G]=Fsn({sourceX:n,sourceY:r,sourcePosition:c,targetX:o,targetY:s,targetPosition:d});return Un.createElement(mU,{path:U,labelX:B,labelY:G,label:h,labelStyle:b,labelShowBg:y,labelBgStyle:E,labelBgPadding:k,labelBgBorderRadius:x,style:O,markerEnd:P,markerStart:D,interactionWidth:$})});NZe.displayName="SimpleBezierEdge";const rFt={[ni.Left]:{x:-1,y:0},[ni.Right]:{x:1,y:0},[ni.Top]:{x:0,y:-1},[ni.Bottom]:{x:0,y:1}},P4r=({source:n,sourcePosition:r=ni.Bottom,target:o})=>r===ni.Left||r===ni.Right?n.x<o.x?{x:1,y:0}:{x:-1,y:0}:n.y<o.y?{x:0,y:1}:{x:0,y:-1},iFt=(n,r)=>Math.sqrt(Math.pow(r.x-n.x,2)+Math.pow(r.y-n.y,2));function L4r({source:n,sourcePosition:r=ni.Bottom,target:o,targetPosition:s=ni.Top,center:c,offset:d}){const h=rFt[r],b=rFt[s],y={x:n.x+h.x*d,y:n.y+h.y*d},E={x:o.x+b.x*d,y:o.y+b.y*d},k=P4r({source:y,sourcePosition:r,target:E}),x=k.x!==0?"x":"y",O=k[x];let P=[],D,$;const U={x:0,y:0},B={x:0,y:0},[G,K,X,ae]=jsn({sourceX:n.x,sourceY:n.y,targetX:o.x,targetY:o.y});if(h[x]*b[x]===-1){D=c.x??G,$=c.y??K;const he=[{x:D,y:y.y},{x:D,y:E.y}],ye=[{x:y.x,y:$},{x:E.x,y:$}];h[x]===O?P=x==="x"?he:ye:P=x==="x"?ye:he}else{const he=[{x:y.x,y:E.y}],ye=[{x:E.x,y:y.y}];if(x==="x"?P=h.x===O?ye:he:P=h.y===O?he:ye,r===s){const Ue=Math.abs(n[x]-o[x]);if(Ue<=d){const Z=Math.min(d-1,d-Ue);h[x]===O?U[x]=(y[x]>n[x]?-1:1)*Z:B[x]=(E[x]>o[x]?-1:1)*Z}}if(r!==s){const Ue=x==="x"?"y":"x",Z=h[x]===b[Ue],Re=y[Ue]>E[Ue],_e=y[Ue]<E[Ue];(h[x]===1&&(!Z&&Re||Z&&_e)||h[x]!==1&&(!Z&&_e||Z&&Re))&&(P=x==="x"?he:ye)}const we={x:y.x+U.x,y:y.y+U.y},ge={x:E.x+B.x,y:E.y+B.y},Te=Math.max(Math.abs(we.x-P[0].x),Math.abs(ge.x-P[0].x)),xe=Math.max(Math.abs(we.y-P[0].y),Math.abs(ge.y-P[0].y));Te>=xe?(D=(we.x+ge.x)/2,$=P[0].y):(D=P[0].x,$=(we.y+ge.y)/2)}return[[n,{x:y.x+U.x,y:y.y+U.y},...P,{x:E.x+B.x,y:E.y+B.y},o],D,$,X,ae]}function D4r(n,r,o,s){const c=Math.min(iFt(n,r)/2,iFt(r,o)/2,s),{x:d,y:h}=r;if(n.x===d&&d===o.x||n.y===h&&h===o.y)return`L${d} ${h}`;if(n.y===h){const E=n.x<o.x?-1:1,k=n.y<o.y?1:-1;return`L ${d+c*E},${h}Q ${d},${h} ${d},${h+c*k}`}const b=n.x<o.x?1:-1,y=n.y<o.y?-1:1;return`L ${d},${h+c*y}Q ${d},${h} ${d+c*b},${h}`}function ioe({sourceX:n,sourceY:r,sourcePosition:o=ni.Bottom,targetX:s,targetY:c,targetPosition:d=ni.Top,borderRadius:h=5,centerX:b,centerY:y,offset:E=20}){const[k,x,O,P,D]=L4r({source:{x:n,y:r},sourcePosition:o,target:{x:s,y:c},targetPosition:d,center:{x:b,y},offset:E});return[k.reduce((U,B,G)=>{let K="";return G>0&&G<k.length-1?K=D4r(k[G-1],B,k[G+1],h):K=`${G===0?"M":"L"}${B.x} ${B.y}`,U+=K,U},""),x,O,P,D]}const pEe=Q.memo(({sourceX:n,sourceY:r,targetX:o,targetY:s,label:c,labelStyle:d,labelShowBg:h,labelBgStyle:b,labelBgPadding:y,labelBgBorderRadius:E,style:k,sourcePosition:x=ni.Bottom,targetPosition:O=ni.Top,markerEnd:P,markerStart:D,pathOptions:$,interactionWidth:U})=>{const[B,G,K]=ioe({sourceX:n,sourceY:r,sourcePosition:x,targetX:o,targetY:s,targetPosition:O,borderRadius:$?.borderRadius,offset:$?.offset});return Un.createElement(mU,{path:B,labelX:G,labelY:K,label:c,labelStyle:d,labelShowBg:h,labelBgStyle:b,labelBgPadding:y,labelBgBorderRadius:E,style:k,markerEnd:P,markerStart:D,interactionWidth:U})});pEe.displayName="SmoothStepEdge";const RZe=Q.memo(n=>Un.createElement(pEe,{...n,pathOptions:Q.useMemo(()=>({borderRadius:0,offset:n.pathOptions?.offset}),[n.pathOptions?.offset])}));RZe.displayName="StepEdge";function M4r({sourceX:n,sourceY:r,targetX:o,targetY:s}){const[c,d,h,b]=jsn({sourceX:n,sourceY:r,targetX:o,targetY:s});return[`M ${n},${r}L ${o},${s}`,c,d,h,b]}const PZe=Q.memo(({sourceX:n,sourceY:r,targetX:o,targetY:s,label:c,labelStyle:d,labelShowBg:h,labelBgStyle:b,labelBgPadding:y,labelBgBorderRadius:E,style:k,markerEnd:x,markerStart:O,interactionWidth:P})=>{const[D,$,U]=M4r({sourceX:n,sourceY:r,targetX:o,targetY:s});return Un.createElement(mU,{path:D,labelX:$,labelY:U,label:c,labelStyle:d,labelShowBg:h,labelBgStyle:b,labelBgPadding:y,labelBgBorderRadius:E,style:k,markerEnd:x,markerStart:O,interactionWidth:P})});PZe.displayName="StraightEdge";function w0e(n,r){return n>=0?.5*n:r*25*Math.sqrt(-n)}function oFt({pos:n,x1:r,y1:o,x2:s,y2:c,c:d}){switch(n){case ni.Left:return[r-w0e(r-s,d),o];case ni.Right:return[r+w0e(s-r,d),o];case ni.Top:return[r,o-w0e(o-c,d)];case ni.Bottom:return[r,o+w0e(c-o,d)]}}function Bsn({sourceX:n,sourceY:r,sourcePosition:o=ni.Bottom,targetX:s,targetY:c,targetPosition:d=ni.Top,curvature:h=.25}){const[b,y]=oFt({pos:o,x1:n,y1:r,x2:s,y2:c,c:h}),[E,k]=oFt({pos:d,x1:s,y1:c,x2:n,y2:r,c:h}),[x,O,P,D]=$sn({sourceX:n,sourceY:r,targetX:s,targetY:c,sourceControlX:b,sourceControlY:y,targetControlX:E,targetControlY:k});return[`M${n},${r} C${b},${y} ${E},${k} ${s},${c}`,x,O,P,D]}const _we=Q.memo(({sourceX:n,sourceY:r,targetX:o,targetY:s,sourcePosition:c=ni.Bottom,targetPosition:d=ni.Top,label:h,labelStyle:b,labelShowBg:y,labelBgStyle:E,labelBgPadding:k,labelBgBorderRadius:x,style:O,markerEnd:P,markerStart:D,pathOptions:$,interactionWidth:U})=>{const[B,G,K]=Bsn({sourceX:n,sourceY:r,sourcePosition:c,targetX:o,targetY:s,targetPosition:d,curvature:$?.curvature});return Un.createElement(mU,{path:B,labelX:G,labelY:K,label:h,labelStyle:b,labelShowBg:y,labelBgStyle:E,labelBgPadding:k,labelBgBorderRadius:x,style:O,markerEnd:P,markerStart:D,interactionWidth:U})});_we.displayName="BezierEdge";const LZe=Q.createContext(null),j4r=LZe.Provider;LZe.Consumer;const $4r=()=>Q.useContext(LZe),F4r=n=>"id"in n&&"source"in n&&"target"in n,B4r=({source:n,sourceHandle:r,target:o,targetHandle:s})=>`reactflow__edge-${n}${r||""}-${o}${s||""}`,hWe=(n,r)=>typeof n>"u"?"":typeof n=="string"?n:`${r?`${r}__`:""}${Object.keys(n).sort().map(s=>`${s}=${n[s]}`).join("&")}`,U4r=(n,r)=>r.some(o=>o.source===n.source&&o.target===n.target&&(o.sourceHandle===n.sourceHandle||!o.sourceHandle&&!n.sourceHandle)&&(o.targetHandle===n.targetHandle||!o.targetHandle&&!n.targetHandle)),z4r=(n,r)=>{if(!n.source||!n.target)return r;let o;return F4r(n)?o={...n}:o={...n,id:B4r(n)},U4r(o,r)?r:r.concat(o)},pWe=({x:n,y:r},[o,s,c],d,[h,b])=>{const y={x:(n-o)/c,y:(r-s)/c};return d?{x:h*Math.round(y.x/h),y:b*Math.round(y.y/b)}:y},Usn=({x:n,y:r},[o,s,c])=>({x:n*c+o,y:r*c+s}),LW=(n,r=[0,0])=>{if(!n)return{x:0,y:0,positionAbsolute:{x:0,y:0}};const o=(n.width??0)*r[0],s=(n.height??0)*r[1],c={x:n.position.x-o,y:n.position.y-s};return{...c,positionAbsolute:n.positionAbsolute?{x:n.positionAbsolute.x-o,y:n.positionAbsolute.y-s}:c}},DZe=(n,r=[0,0])=>{if(n.length===0)return{x:0,y:0,width:0,height:0};const o=n.reduce((s,c)=>{const{x:d,y:h}=LW(c,r).positionAbsolute;return A4r(s,OZe({x:d,y:h,width:c.width||0,height:c.height||0}))},{x:1/0,y:1/0,x2:-1/0,y2:-1/0});return I4r(o)},zsn=(n,r,[o,s,c]=[0,0,1],d=!1,h=!1,b=[0,0])=>{const y={x:(r.x-o)/c,y:(r.y-s)/c,width:r.width/c,height:r.height/c},E=[];return n.forEach(k=>{const{width:x,height:O,selectable:P=!0,hidden:D=!1}=k;if(h&&!P||D)return!1;const{positionAbsolute:$}=LW(k,b),U={x:$.x,y:$.y,width:x||0,height:O||0},B=dWe(y,U),G=typeof x>"u"||typeof O>"u"||x===null||O===null,K=d&&B>0,X=(x||0)*(O||0);(G||K||B>=X||k.dragging)&&E.push(k)}),E},Hsn=(n,r)=>{const o=n.map(s=>s.id);return r.filter(s=>o.includes(s.source)||o.includes(s.target))},Gsn=(n,r,o,s,c,d=.1)=>{const h=r/(n.width*(1+d)),b=o/(n.height*(1+d)),y=Math.min(h,b),E=oK(y,s,c),k=n.x+n.width/2,x=n.y+n.height/2,O=r/2-k*E,P=o/2-x*E;return{x:O,y:P,zoom:E}},_B=(n,r=0)=>n.transition().duration(r);function aFt(n,r,o,s){return(r[o]||[]).reduce((c,d)=>(`${n.id}-${d.id}-${o}`!==s&&c.push({id:d.id||null,type:o,nodeId:n.id,x:(n.positionAbsolute?.x??0)+d.x+d.width/2,y:(n.positionAbsolute?.y??0)+d.y+d.height/2}),c),[])}function H4r(n,r,o,s,c,d){const{x:h,y:b}=bM(n),E=r.elementsFromPoint(h,b).find(D=>D.classList.contains("react-flow__handle"));if(E){const D=E.getAttribute("data-nodeid");if(D){const $=MZe(void 0,E),U=E.getAttribute("data-handleid"),B=d({nodeId:D,id:U,type:$});if(B){const G=c.find(K=>K.nodeId===D&&K.type===$&&K.id===U);return{handle:{id:U,type:$,nodeId:D,x:G?.x||o.x,y:G?.y||o.y},validHandleResult:B}}}}let k=[],x=1/0;if(c.forEach(D=>{const $=Math.sqrt((D.x-o.x)**2+(D.y-o.y)**2);if($<=s){const U=d(D);$<=x&&($<x?k=[{handle:D,validHandleResult:U}]:$===x&&k.push({handle:D,validHandleResult:U}),x=$)}}),!k.length)return{handle:null,validHandleResult:qsn()};if(k.length===1)return k[0];const O=k.some(({validHandleResult:D})=>D.isValid),P=k.some(({handle:D})=>D.type==="target");return k.find(({handle:D,validHandleResult:$})=>P?D.type==="target":O?$.isValid:!0)||k[0]}const G4r={source:null,target:null,sourceHandle:null,targetHandle:null},qsn=()=>({handleDomNode:null,isValid:!1,connection:G4r,endHandle:null});function Vsn(n,r,o,s,c,d,h){const b=c==="target",y=h.querySelector(`.react-flow__handle[data-id="${n?.nodeId}-${n?.id}-${n?.type}"]`),E={...qsn(),handleDomNode:y};if(y){const k=MZe(void 0,y),x=y.getAttribute("data-nodeid"),O=y.getAttribute("data-handleid"),P=y.classList.contains("connectable"),D=y.classList.contains("connectableend"),$={source:b?x:o,sourceHandle:b?O:s,target:b?o:x,targetHandle:b?s:O};E.connection=$,P&&D&&(r===nU.Strict?b&&k==="source"||!b&&k==="target":x!==o||O!==s)&&(E.endHandle={nodeId:x,handleId:O,type:k},E.isValid=d($))}return E}function q4r({nodes:n,nodeId:r,handleId:o,handleType:s}){return n.reduce((c,d)=>{if(d[Oh]){const{handleBounds:h}=d[Oh];let b=[],y=[];h&&(b=aFt(d,h,"source",`${r}-${o}-${s}`),y=aFt(d,h,"target",`${r}-${o}-${s}`)),c.push(...b,...y)}return c},[])}function MZe(n,r){return n||(r?.classList.contains("target")?"target":r?.classList.contains("source")?"source":null)}function H$e(n){n?.classList.remove("valid","connecting","react-flow__handle-valid","react-flow__handle-connecting")}function V4r(n,r){let o=null;return r?o="valid":n&&!r&&(o="invalid"),o}function Wsn({event:n,handleId:r,nodeId:o,onConnect:s,isTarget:c,getState:d,setState:h,isValidConnection:b,edgeUpdaterType:y,onReconnectEnd:E}){const k=Lsn(n.target),{connectionMode:x,domNode:O,autoPanOnConnect:P,connectionRadius:D,onConnectStart:$,panBy:U,getNodes:B,cancelConnection:G}=d();let K=0,X;const{x:ae,y:oe}=bM(n),he=k?.elementFromPoint(ae,oe),ye=MZe(y,he),we=O?.getBoundingClientRect();if(!we||!ye)return;let ge,Te=bM(n,we),xe=!1,Ue=null,Z=!1,Re=null;const _e=q4r({nodes:B(),nodeId:o,handleId:r,handleType:ye}),qe=()=>{if(!P)return;const[de,We]=Psn(Te,we);U({x:de,y:We}),K=requestAnimationFrame(qe)};h({connectionPosition:Te,connectionStatus:null,connectionNodeId:o,connectionHandleId:r,connectionHandleType:ye,connectionStartHandle:{nodeId:o,handleId:r,type:ye},connectionEndHandle:null}),$?.(n,{nodeId:o,handleId:r,handleType:ye});function it(de){const{transform:We}=d();Te=bM(de,we);const{handle:ht,validHandleResult:Ie}=H4r(de,k,pWe(Te,We,!1,[1,1]),D,_e,$t=>Vsn($t,x,o,r,c?"target":"source",b,k));if(X=ht,xe||(qe(),xe=!0),Re=Ie.handleDomNode,Ue=Ie.connection,Z=Ie.isValid,h({connectionPosition:X&&Z?Usn({x:X.x,y:X.y},We):Te,connectionStatus:V4r(!!X,Z),connectionEndHandle:Ie.endHandle}),!X&&!Z&&!Re)return H$e(ge);Ue.source!==Ue.target&&Re&&(H$e(ge),ge=Re,Re.classList.add("connecting","react-flow__handle-connecting"),Re.classList.toggle("valid",Z),Re.classList.toggle("react-flow__handle-valid",Z))}function Xe(de){(X||Re)&&Ue&&Z&&s?.(Ue),d().onConnectEnd?.(de),y&&E?.(de),H$e(ge),G(),cancelAnimationFrame(K),xe=!1,Z=!1,Ue=null,Re=null,k.removeEventListener("mousemove",it),k.removeEventListener("mouseup",Xe),k.removeEventListener("touchmove",it),k.removeEventListener("touchend",Xe)}k.addEventListener("mousemove",it),k.addEventListener("mouseup",Xe),k.addEventListener("touchmove",it),k.addEventListener("touchend",Xe)}const sFt=()=>!0,W4r=n=>({connectionStartHandle:n.connectionStartHandle,connectOnClick:n.connectOnClick,noPanClassName:n.noPanClassName}),K4r=(n,r,o)=>s=>{const{connectionStartHandle:c,connectionEndHandle:d,connectionClickStartHandle:h}=s;return{connecting:c?.nodeId===n&&c?.handleId===r&&c?.type===o||d?.nodeId===n&&d?.handleId===r&&d?.type===o,clickConnecting:h?.nodeId===n&&h?.handleId===r&&h?.type===o}},Ksn=Q.forwardRef(({type:n="source",position:r=ni.Top,isValidConnection:o,isConnectable:s=!0,isConnectableStart:c=!0,isConnectableEnd:d=!0,id:h,onConnect:b,children:y,className:E,onMouseDown:k,onTouchStart:x,...O},P)=>{const D=h||null,$=n==="target",U=Wb(),B=$4r(),{connectOnClick:G,noPanClassName:K}=Nh(W4r,ey),{connecting:X,clickConnecting:ae}=Nh(K4r(B,D,n),ey);B||U.getState().onError?.("010",W4.error010());const oe=we=>{const{defaultEdgeOptions:ge,onConnect:Te,hasDefaultEdges:xe}=U.getState(),Ue={...ge,...we};if(xe){const{edges:Z,setEdges:Re}=U.getState();Re(z4r(Ue,Z))}Te?.(Ue),b?.(Ue)},he=we=>{if(!B)return;const ge=Msn(we);c&&(ge&&we.button===0||!ge)&&Wsn({event:we,handleId:D,nodeId:B,onConnect:oe,isTarget:$,getState:U.getState,setState:U.setState,isValidConnection:o||U.getState().isValidConnection||sFt}),ge?k?.(we):x?.(we)},ye=we=>{const{onClickConnectStart:ge,onClickConnectEnd:Te,connectionClickStartHandle:xe,connectionMode:Ue,isValidConnection:Z}=U.getState();if(!B||!xe&&!c)return;if(!xe){ge?.(we,{nodeId:B,handleId:D,handleType:n}),U.setState({connectionClickStartHandle:{nodeId:B,type:n,handleId:D}});return}const Re=Lsn(we.target),_e=o||Z||sFt,{connection:qe,isValid:it}=Vsn({nodeId:B,id:D,type:n},Ue,xe.nodeId,xe.handleId||null,xe.type,_e,Re);it&&oe(qe),Te?.(we),U.setState({connectionClickStartHandle:null})};return Un.createElement("div",{"data-handleid":D,"data-nodeid":B,"data-handlepos":r,"data-id":`${B}-${D}-${n}`,className:q_(["react-flow__handle",`react-flow__handle-${r}`,"nodrag",K,E,{source:!$,target:$,connectable:s,connectablestart:c,connectableend:d,connecting:ae,connectionindicator:s&&(c&&!X||d&&X)}]),onMouseDown:he,onTouchStart:he,onClick:G?ye:void 0,ref:P,...O},y)});Ksn.displayName="Handle";var Um=Q.memo(Ksn);const Ysn=({data:n,isConnectable:r,targetPosition:o=ni.Top,sourcePosition:s=ni.Bottom})=>Un.createElement(Un.Fragment,null,Un.createElement(Um,{type:"target",position:o,isConnectable:r}),n?.label,Un.createElement(Um,{type:"source",position:s,isConnectable:r}));Ysn.displayName="DefaultNode";var gWe=Q.memo(Ysn);const Xsn=({data:n,isConnectable:r,sourcePosition:o=ni.Bottom})=>Un.createElement(Un.Fragment,null,n?.label,Un.createElement(Um,{type:"source",position:o,isConnectable:r}));Xsn.displayName="InputNode";var Jsn=Q.memo(Xsn);const Zsn=({data:n,isConnectable:r,targetPosition:o=ni.Top})=>Un.createElement(Un.Fragment,null,Un.createElement(Um,{type:"target",position:o,isConnectable:r}),n?.label);Zsn.displayName="OutputNode";var Qsn=Q.memo(Zsn);const jZe=()=>null;jZe.displayName="GroupNode";const Y4r=n=>({selectedNodes:n.getNodes().filter(r=>r.selected),selectedEdges:n.edges.filter(r=>r.selected).map(r=>({...r}))}),y0e=n=>n.id;function X4r(n,r){return ey(n.selectedNodes.map(y0e),r.selectedNodes.map(y0e))&&ey(n.selectedEdges.map(y0e),r.selectedEdges.map(y0e))}const ecn=Q.memo(({onSelectionChange:n})=>{const r=Wb(),{selectedNodes:o,selectedEdges:s}=Nh(Y4r,X4r);return Q.useEffect(()=>{const c={nodes:o,edges:s};n?.(c),r.getState().onSelectionChange.forEach(d=>d(c))},[o,s,n]),null});ecn.displayName="SelectionListener";const J4r=n=>!!n.onSelectionChange;function Z4r({onSelectionChange:n}){const r=Nh(J4r);return n||r?Un.createElement(ecn,{onSelectionChange:n}):null}const Q4r=n=>({setNodes:n.setNodes,setEdges:n.setEdges,setDefaultNodesAndEdges:n.setDefaultNodesAndEdges,setMinZoom:n.setMinZoom,setMaxZoom:n.setMaxZoom,setTranslateExtent:n.setTranslateExtent,setNodeExtent:n.setNodeExtent,reset:n.reset});function GV(n,r){Q.useEffect(()=>{typeof n<"u"&&r(n)},[n])}function tc(n,r,o){Q.useEffect(()=>{typeof r<"u"&&o({[n]:r})},[r])}const e6r=({nodes:n,edges:r,defaultNodes:o,defaultEdges:s,onConnect:c,onConnectStart:d,onConnectEnd:h,onClickConnectStart:b,onClickConnectEnd:y,nodesDraggable:E,nodesConnectable:k,nodesFocusable:x,edgesFocusable:O,edgesUpdatable:P,elevateNodesOnSelect:D,minZoom:$,maxZoom:U,nodeExtent:B,onNodesChange:G,onEdgesChange:K,elementsSelectable:X,connectionMode:ae,snapGrid:oe,snapToGrid:he,translateExtent:ye,connectOnClick:we,defaultEdgeOptions:ge,fitView:Te,fitViewOptions:xe,onNodesDelete:Ue,onEdgesDelete:Z,onNodeDrag:Re,onNodeDragStart:_e,onNodeDragStop:qe,onSelectionDrag:it,onSelectionDragStart:Xe,onSelectionDragStop:de,noPanClassName:We,nodeOrigin:ht,rfId:Ie,autoPanOnConnect:$t,autoPanOnNodeDrag:An,onError:Yt,connectionRadius:Or,isValidConnection:Nt,nodeDragThreshold:on})=>{const{setNodes:br,setEdges:wi,setDefaultNodesAndEdges:no,setMinZoom:Ri,setMaxZoom:ro,setTranslateExtent:Ko,setNodeExtent:Ca,reset:qi}=Nh(Q4r,ey),yi=Wb();return Q.useEffect(()=>{const lo=s?.map(Eo=>({...Eo,...ge}));return no(o,lo),()=>{qi()}},[]),tc("defaultEdgeOptions",ge,yi.setState),tc("connectionMode",ae,yi.setState),tc("onConnect",c,yi.setState),tc("onConnectStart",d,yi.setState),tc("onConnectEnd",h,yi.setState),tc("onClickConnectStart",b,yi.setState),tc("onClickConnectEnd",y,yi.setState),tc("nodesDraggable",E,yi.setState),tc("nodesConnectable",k,yi.setState),tc("nodesFocusable",x,yi.setState),tc("edgesFocusable",O,yi.setState),tc("edgesUpdatable",P,yi.setState),tc("elementsSelectable",X,yi.setState),tc("elevateNodesOnSelect",D,yi.setState),tc("snapToGrid",he,yi.setState),tc("snapGrid",oe,yi.setState),tc("onNodesChange",G,yi.setState),tc("onEdgesChange",K,yi.setState),tc("connectOnClick",we,yi.setState),tc("fitViewOnInit",Te,yi.setState),tc("fitViewOnInitOptions",xe,yi.setState),tc("onNodesDelete",Ue,yi.setState),tc("onEdgesDelete",Z,yi.setState),tc("onNodeDrag",Re,yi.setState),tc("onNodeDragStart",_e,yi.setState),tc("onNodeDragStop",qe,yi.setState),tc("onSelectionDrag",it,yi.setState),tc("onSelectionDragStart",Xe,yi.setState),tc("onSelectionDragStop",de,yi.setState),tc("noPanClassName",We,yi.setState),tc("nodeOrigin",ht,yi.setState),tc("rfId",Ie,yi.setState),tc("autoPanOnConnect",$t,yi.setState),tc("autoPanOnNodeDrag",An,yi.setState),tc("onError",Yt,yi.setState),tc("connectionRadius",Or,yi.setState),tc("isValidConnection",Nt,yi.setState),tc("nodeDragThreshold",on,yi.setState),GV(n,br),GV(r,wi),GV($,Ri),GV(U,ro),GV(ye,Ko),GV(B,Ca),null},cFt={display:"none"},t6r={position:"absolute",width:1,height:1,margin:-1,border:0,padding:0,overflow:"hidden",clip:"rect(0px, 0px, 0px, 0px)",clipPath:"inset(100%)"},tcn="react-flow__node-desc",ncn="react-flow__edge-desc",n6r="react-flow__aria-live",r6r=n=>n.ariaLiveMessage;function i6r({rfId:n}){const r=Nh(r6r);return Un.createElement("div",{id:`${n6r}-${n}`,"aria-live":"assertive","aria-atomic":"true",style:t6r},r)}function o6r({rfId:n,disableKeyboardA11y:r}){return Un.createElement(Un.Fragment,null,Un.createElement("div",{id:`${tcn}-${n}`,style:cFt},"Press enter or space to select a node.",!r&&"You can then use the arrow keys to move the node around."," Press delete to remove it and escape to cancel."," "),Un.createElement("div",{id:`${ncn}-${n}`,style:cFt},"Press enter or space to select an edge. You can then press delete to remove it or escape to cancel."),!r&&Un.createElement(i6r,{rfId:n}))}var ooe=(n=null,r={actInsideInputWithModifier:!0})=>{const[o,s]=Q.useState(!1),c=Q.useRef(!1),d=Q.useRef(new Set([])),[h,b]=Q.useMemo(()=>{if(n!==null){const E=(Array.isArray(n)?n:[n]).filter(x=>typeof x=="string").map(x=>x.split("+")),k=E.reduce((x,O)=>x.concat(...O),[]);return[E,k]}return[[],[]]},[n]);return Q.useEffect(()=>{const y=typeof document<"u"?document:null,E=r?.target||y;if(n!==null){const k=P=>{if(c.current=P.ctrlKey||P.metaKey||P.shiftKey,(!c.current||c.current&&!r.actInsideInputWithModifier)&&fWe(P))return!1;const $=lFt(P.code,b);d.current.add(P[$]),uFt(h,d.current,!1)&&(P.preventDefault(),s(!0))},x=P=>{if((!c.current||c.current&&!r.actInsideInputWithModifier)&&fWe(P))return!1;const $=lFt(P.code,b);uFt(h,d.current,!0)?(s(!1),d.current.clear()):d.current.delete(P[$]),P.key==="Meta"&&d.current.clear(),c.current=!1},O=()=>{d.current.clear(),s(!1)};return E?.addEventListener("keydown",k),E?.addEventListener("keyup",x),window.addEventListener("blur",O),()=>{E?.removeEventListener("keydown",k),E?.removeEventListener("keyup",x),window.removeEventListener("blur",O)}}},[n,s]),o};function uFt(n,r,o){return n.filter(s=>o||s.length===r.size).some(s=>s.every(c=>r.has(c)))}function lFt(n,r){return r.includes(n)?"code":"key"}function rcn(n,r,o,s){const c=n.parentNode||n.parentId;if(!c)return o;const d=r.get(c),h=LW(d,s);return rcn(d,r,{x:(o.x??0)+h.x,y:(o.y??0)+h.y,z:(d[Oh]?.z??0)>(o.z??0)?d[Oh]?.z??0:o.z??0},s)}function icn(n,r,o){n.forEach(s=>{const c=s.parentNode||s.parentId;if(c&&!n.has(c))throw new Error(`Parent node ${c} not found`);if(c||o?.[s.id]){const{x:d,y:h,z:b}=rcn(s,n,{...s.position,z:s[Oh]?.z??0},r);s.positionAbsolute={x:d,y:h},s[Oh].z=b,o?.[s.id]&&(s[Oh].isParent=!0)}})}function G$e(n,r,o,s){const c=new Map,d={},h=s?1e3:0;return n.forEach(b=>{const y=(M_(b.zIndex)?b.zIndex:0)+(b.selected?h:0),E=r.get(b.id),k={...b,positionAbsolute:{x:b.position.x,y:b.position.y}},x=b.parentNode||b.parentId;x&&(d[x]=!0);const O=E?.type&&E?.type!==b.type;Object.defineProperty(k,Oh,{enumerable:!1,value:{handleBounds:O?void 0:E?.[Oh]?.handleBounds,z:y}}),c.set(b.id,k)}),icn(c,o,d),c}function ocn(n,r={}){const{getNodes:o,width:s,height:c,minZoom:d,maxZoom:h,d3Zoom:b,d3Selection:y,fitViewOnInitDone:E,fitViewOnInit:k,nodeOrigin:x}=n(),O=r.initial&&!E&&k;if(b&&y&&(O||!r.initial)){const D=o().filter(U=>{const B=r.includeHiddenNodes?U.width&&U.height:!U.hidden;return r.nodes?.length?B&&r.nodes.some(G=>G.id===U.id):B}),$=D.every(U=>U.width&&U.height);if(D.length>0&&$){const U=DZe(D,x),{x:B,y:G,zoom:K}=Gsn(U,s,c,r.minZoom??d,r.maxZoom??h,r.padding??.1),X=mM.translate(B,G).scale(K);return typeof r.duration=="number"&&r.duration>0?b.transform(_B(y,r.duration),X):b.transform(y,X),!0}}return!1}function a6r(n,r){return n.forEach(o=>{const s=r.get(o.id);s&&r.set(s.id,{...s,[Oh]:s[Oh],selected:o.selected})}),new Map(r)}function s6r(n,r){return r.map(o=>{const s=n.find(c=>c.id===o.id);return s&&(o.selected=s.selected),o})}function E0e({changedNodes:n,changedEdges:r,get:o,set:s}){const{nodeInternals:c,edges:d,onNodesChange:h,onEdgesChange:b,hasDefaultNodes:y,hasDefaultEdges:E}=o();n?.length&&(y&&s({nodeInternals:a6r(n,c)}),h?.(n)),r?.length&&(E&&s({edges:s6r(r,d)}),b?.(r))}const qV=()=>{},c6r={zoomIn:qV,zoomOut:qV,zoomTo:qV,getZoom:()=>1,setViewport:qV,getViewport:()=>({x:0,y:0,zoom:1}),fitView:()=>!1,setCenter:qV,fitBounds:qV,project:n=>n,screenToFlowPosition:n=>n,flowToScreenPosition:n=>n,viewportInitialized:!1},u6r=n=>({d3Zoom:n.d3Zoom,d3Selection:n.d3Selection}),l6r=()=>{const n=Wb(),{d3Zoom:r,d3Selection:o}=Nh(u6r,ey);return Q.useMemo(()=>o&&r?{zoomIn:c=>r.scaleBy(_B(o,c?.duration),1.2),zoomOut:c=>r.scaleBy(_B(o,c?.duration),1/1.2),zoomTo:(c,d)=>r.scaleTo(_B(o,d?.duration),c),getZoom:()=>n.getState().transform[2],setViewport:(c,d)=>{const[h,b,y]=n.getState().transform,E=mM.translate(c.x??h,c.y??b).scale(c.zoom??y);r.transform(_B(o,d?.duration),E)},getViewport:()=>{const[c,d,h]=n.getState().transform;return{x:c,y:d,zoom:h}},fitView:c=>ocn(n.getState,c),setCenter:(c,d,h)=>{const{width:b,height:y,maxZoom:E}=n.getState(),k=typeof h?.zoom<"u"?h.zoom:E,x=b/2-c*k,O=y/2-d*k,P=mM.translate(x,O).scale(k);r.transform(_B(o,h?.duration),P)},fitBounds:(c,d)=>{const{width:h,height:b,minZoom:y,maxZoom:E}=n.getState(),{x:k,y:x,zoom:O}=Gsn(c,h,b,y,E,d?.padding??.1),P=mM.translate(k,x).scale(O);r.transform(_B(o,d?.duration),P)},project:c=>{const{transform:d,snapToGrid:h,snapGrid:b}=n.getState();return console.warn("[DEPRECATED] `project` is deprecated. Instead use `screenToFlowPosition`. There is no need to subtract the react flow bounds anymore! https://reactflow.dev/api-reference/types/react-flow-instance#screen-to-flow-position"),pWe(c,d,h,b)},screenToFlowPosition:c=>{const{transform:d,snapToGrid:h,snapGrid:b,domNode:y}=n.getState();if(!y)return c;const{x:E,y:k}=y.getBoundingClientRect(),x={x:c.x-E,y:c.y-k};return pWe(x,d,h,b)},flowToScreenPosition:c=>{const{transform:d,domNode:h}=n.getState();if(!h)return c;const{x:b,y}=h.getBoundingClientRect(),E=Usn(c,d);return{x:E.x+b,y:E.y+y}},viewportInitialized:!0}:c6r,[r,o])};function $Ze(){const n=l6r(),r=Wb(),o=Q.useCallback(()=>r.getState().getNodes().map($=>({...$})),[]),s=Q.useCallback($=>r.getState().nodeInternals.get($),[]),c=Q.useCallback(()=>{const{edges:$=[]}=r.getState();return $.map(U=>({...U}))},[]),d=Q.useCallback($=>{const{edges:U=[]}=r.getState();return U.find(B=>B.id===$)},[]),h=Q.useCallback($=>{const{getNodes:U,setNodes:B,hasDefaultNodes:G,onNodesChange:K}=r.getState(),X=U(),ae=typeof $=="function"?$(X):$;if(G)B(ae);else if(K){const oe=ae.length===0?X.map(he=>({type:"remove",id:he.id})):ae.map(he=>({item:he,type:"reset"}));K(oe)}},[]),b=Q.useCallback($=>{const{edges:U=[],setEdges:B,hasDefaultEdges:G,onEdgesChange:K}=r.getState(),X=typeof $=="function"?$(U):$;if(G)B(X);else if(K){const ae=X.length===0?U.map(oe=>({type:"remove",id:oe.id})):X.map(oe=>({item:oe,type:"reset"}));K(ae)}},[]),y=Q.useCallback($=>{const U=Array.isArray($)?$:[$],{getNodes:B,setNodes:G,hasDefaultNodes:K,onNodesChange:X}=r.getState();if(K){const oe=[...B(),...U];G(oe)}else if(X){const ae=U.map(oe=>({item:oe,type:"add"}));X(ae)}},[]),E=Q.useCallback($=>{const U=Array.isArray($)?$:[$],{edges:B=[],setEdges:G,hasDefaultEdges:K,onEdgesChange:X}=r.getState();if(K)G([...B,...U]);else if(X){const ae=U.map(oe=>({item:oe,type:"add"}));X(ae)}},[]),k=Q.useCallback(()=>{const{getNodes:$,edges:U=[],transform:B}=r.getState(),[G,K,X]=B;return{nodes:$().map(ae=>({...ae})),edges:U.map(ae=>({...ae})),viewport:{x:G,y:K,zoom:X}}},[]),x=Q.useCallback(({nodes:$,edges:U})=>{const{nodeInternals:B,getNodes:G,edges:K,hasDefaultNodes:X,hasDefaultEdges:ae,onNodesDelete:oe,onEdgesDelete:he,onNodesChange:ye,onEdgesChange:we}=r.getState(),ge=($||[]).map(Re=>Re.id),Te=(U||[]).map(Re=>Re.id),xe=G().reduce((Re,_e)=>{const qe=_e.parentNode||_e.parentId,it=!ge.includes(_e.id)&&qe&&Re.find(de=>de.id===qe);return(typeof _e.deletable=="boolean"?_e.deletable:!0)&&(ge.includes(_e.id)||it)&&Re.push(_e),Re},[]),Ue=K.filter(Re=>typeof Re.deletable=="boolean"?Re.deletable:!0),Z=Ue.filter(Re=>Te.includes(Re.id));if(xe||Z){const Re=Hsn(xe,Ue),_e=[...Z,...Re],qe=_e.reduce((it,Xe)=>(it.includes(Xe.id)||it.push(Xe.id),it),[]);if((ae||X)&&(ae&&r.setState({edges:K.filter(it=>!qe.includes(it.id))}),X&&(xe.forEach(it=>{B.delete(it.id)}),r.setState({nodeInternals:new Map(B)}))),qe.length>0&&(he?.(_e),we&&we(qe.map(it=>({id:it,type:"remove"})))),xe.length>0&&(oe?.(xe),ye)){const it=xe.map(Xe=>({id:Xe.id,type:"remove"}));ye(it)}}},[]),O=Q.useCallback($=>{const U=O4r($),B=U?null:r.getState().nodeInternals.get($.id);return!U&&!B?[null,null,U]:[U?$:tFt(B),B,U]},[]),P=Q.useCallback(($,U=!0,B)=>{const[G,K,X]=O($);return G?(B||r.getState().getNodes()).filter(ae=>{if(!X&&(ae.id===K.id||!ae.positionAbsolute))return!1;const oe=tFt(ae),he=dWe(oe,G);return U&&he>0||he>=G.width*G.height}):[]},[]),D=Q.useCallback(($,U,B=!0)=>{const[G]=O($);if(!G)return!1;const K=dWe(G,U);return B&&K>0||K>=G.width*G.height},[]);return Q.useMemo(()=>({...n,getNodes:o,getNode:s,getEdges:c,getEdge:d,setNodes:h,setEdges:b,addNodes:y,addEdges:E,toObject:k,deleteElements:x,getIntersectingNodes:P,isNodeIntersecting:D}),[n,o,s,c,d,h,b,y,E,k,x,P,D])}const d6r={actInsideInputWithModifier:!1};var f6r=({deleteKeyCode:n,multiSelectionKeyCode:r})=>{const o=Wb(),{deleteElements:s}=$Ze(),c=ooe(n,d6r),d=ooe(r);Q.useEffect(()=>{if(c){const{edges:h,getNodes:b}=o.getState(),y=b().filter(k=>k.selected),E=h.filter(k=>k.selected);s({nodes:y,edges:E}),o.setState({nodesSelectionActive:!1})}},[c]),Q.useEffect(()=>{o.setState({multiSelectionActive:d})},[d])};function h6r(n){const r=Wb();Q.useEffect(()=>{let o;const s=()=>{if(!n.current)return;const c=AZe(n.current);(c.height===0||c.width===0)&&r.getState().onError?.("004",W4.error004()),r.setState({width:c.width||500,height:c.height||500})};return s(),window.addEventListener("resize",s),n.current&&(o=new ResizeObserver(()=>s()),o.observe(n.current)),()=>{window.removeEventListener("resize",s),o&&n.current&&o.unobserve(n.current)}},[])}const FZe={position:"absolute",width:"100%",height:"100%",top:0,left:0},p6r=(n,r)=>n.x!==r.x||n.y!==r.y||n.zoom!==r.k,S0e=n=>({x:n.x,y:n.y,zoom:n.k}),VV=(n,r)=>n.target.closest(`.${r}`),dFt=(n,r)=>r===2&&Array.isArray(n)&&n.includes(2),fFt=n=>{const r=n.ctrlKey&&Swe()?10:1;return-n.deltaY*(n.deltaMode===1?.05:n.deltaMode?1:.002)*r},g6r=n=>({d3Zoom:n.d3Zoom,d3Selection:n.d3Selection,d3ZoomHandler:n.d3ZoomHandler,userSelectionActive:n.userSelectionActive}),m6r=({onMove:n,onMoveStart:r,onMoveEnd:o,onPaneContextMenu:s,zoomOnScroll:c=!0,zoomOnPinch:d=!0,panOnScroll:h=!1,panOnScrollSpeed:b=.5,panOnScrollMode:y=MB.Free,zoomOnDoubleClick:E=!0,elementsSelectable:k,panOnDrag:x=!0,defaultViewport:O,translateExtent:P,minZoom:D,maxZoom:$,zoomActivationKeyCode:U,preventScrolling:B=!0,children:G,noWheelClassName:K,noPanClassName:X})=>{const ae=Q.useRef(),oe=Wb(),he=Q.useRef(!1),ye=Q.useRef(!1),we=Q.useRef(null),ge=Q.useRef({x:0,y:0,zoom:0}),{d3Zoom:Te,d3Selection:xe,d3ZoomHandler:Ue,userSelectionActive:Z}=Nh(g6r,ey),Re=ooe(U),_e=Q.useRef(0),qe=Q.useRef(!1),it=Q.useRef();return h6r(we),Q.useEffect(()=>{if(we.current){const Xe=we.current.getBoundingClientRect(),de=E4r().scaleExtent([D,$]).translateExtent(P),We=Rx(we.current).call(de),ht=mM.translate(O.x,O.y).scale(oK(O.zoom,D,$)),Ie=[[0,0],[Xe.width,Xe.height]],$t=de.constrain()(ht,Ie,P);de.transform(We,$t),de.wheelDelta(fFt),oe.setState({d3Zoom:de,d3Selection:We,d3ZoomHandler:We.on("wheel.zoom"),transform:[$t.x,$t.y,$t.k],domNode:we.current.closest(".react-flow")})}},[]),Q.useEffect(()=>{xe&&Te&&(h&&!Re&&!Z?xe.on("wheel.zoom",Xe=>{if(VV(Xe,K))return!1;Xe.preventDefault(),Xe.stopImmediatePropagation();const de=xe.property("__zoom").k||1;if(Xe.ctrlKey&&d){const Nt=sI(Xe),on=fFt(Xe),br=de*Math.pow(2,on);Te.scaleTo(xe,br,Nt,Xe);return}const We=Xe.deltaMode===1?20:1;let ht=y===MB.Vertical?0:Xe.deltaX*We,Ie=y===MB.Horizontal?0:Xe.deltaY*We;!Swe()&&Xe.shiftKey&&y!==MB.Vertical&&(ht=Xe.deltaY*We,Ie=0),Te.translateBy(xe,-(ht/de)*b,-(Ie/de)*b,{internal:!0});const $t=S0e(xe.property("__zoom")),{onViewportChangeStart:An,onViewportChange:Yt,onViewportChangeEnd:Or}=oe.getState();clearTimeout(it.current),qe.current||(qe.current=!0,r?.(Xe,$t),An?.($t)),qe.current&&(n?.(Xe,$t),Yt?.($t),it.current=setTimeout(()=>{o?.(Xe,$t),Or?.($t),qe.current=!1},150))},{passive:!1}):typeof Ue<"u"&&xe.on("wheel.zoom",function(Xe,de){if(!B&&Xe.type==="wheel"&&!Xe.ctrlKey||VV(Xe,K))return null;Xe.preventDefault(),Ue.call(this,Xe,de)},{passive:!1}))},[Z,h,y,xe,Te,Ue,Re,d,B,K,r,n,o]),Q.useEffect(()=>{Te&&Te.on("start",Xe=>{if(!Xe.sourceEvent||Xe.sourceEvent.internal)return null;_e.current=Xe.sourceEvent?.button;const{onViewportChangeStart:de}=oe.getState(),We=S0e(Xe.transform);he.current=!0,ge.current=We,Xe.sourceEvent?.type==="mousedown"&&oe.setState({paneDragging:!0}),de?.(We),r?.(Xe.sourceEvent,We)})},[Te,r]),Q.useEffect(()=>{Te&&(Z&&!he.current?Te.on("zoom",null):Z||Te.on("zoom",Xe=>{const{onViewportChange:de}=oe.getState();if(oe.setState({transform:[Xe.transform.x,Xe.transform.y,Xe.transform.k]}),ye.current=!!(s&&dFt(x,_e.current??0)),(n||de)&&!Xe.sourceEvent?.internal){const We=S0e(Xe.transform);de?.(We),n?.(Xe.sourceEvent,We)}}))},[Z,Te,n,x,s]),Q.useEffect(()=>{Te&&Te.on("end",Xe=>{if(!Xe.sourceEvent||Xe.sourceEvent.internal)return null;const{onViewportChangeEnd:de}=oe.getState();if(he.current=!1,oe.setState({paneDragging:!1}),s&&dFt(x,_e.current??0)&&!ye.current&&s(Xe.sourceEvent),ye.current=!1,(o||de)&&p6r(ge.current,Xe.transform)){const We=S0e(Xe.transform);ge.current=We,clearTimeout(ae.current),ae.current=setTimeout(()=>{de?.(We),o?.(Xe.sourceEvent,We)},h?150:0)}})},[Te,h,x,o,s]),Q.useEffect(()=>{Te&&Te.filter(Xe=>{const de=Re||c,We=d&&Xe.ctrlKey;if((x===!0||Array.isArray(x)&&x.includes(1))&&Xe.button===1&&Xe.type==="mousedown"&&(VV(Xe,"react-flow__node")||VV(Xe,"react-flow__edge")))return!0;if(!x&&!de&&!h&&!E&&!d||Z||!E&&Xe.type==="dblclick"||VV(Xe,K)&&Xe.type==="wheel"||VV(Xe,X)&&(Xe.type!=="wheel"||h&&Xe.type==="wheel"&&!Re)||!d&&Xe.ctrlKey&&Xe.type==="wheel"||!de&&!h&&!We&&Xe.type==="wheel"||!x&&(Xe.type==="mousedown"||Xe.type==="touchstart")||Array.isArray(x)&&!x.includes(Xe.button)&&Xe.type==="mousedown")return!1;const ht=Array.isArray(x)&&x.includes(Xe.button)||!Xe.button||Xe.button<=1;return(!Xe.ctrlKey||Xe.type==="wheel")&&ht})},[Z,Te,c,d,h,E,x,k,Re]),Un.createElement("div",{className:"react-flow__renderer",ref:we,style:FZe},G)},b6r=n=>({userSelectionActive:n.userSelectionActive,userSelectionRect:n.userSelectionRect});function v6r(){const{userSelectionActive:n,userSelectionRect:r}=Nh(b6r,ey);return n&&r?Un.createElement("div",{className:"react-flow__selection react-flow__container",style:{width:r.width,height:r.height,transform:`translate(${r.x}px, ${r.y}px)`}}):null}function hFt(n,r){const o=r.parentNode||r.parentId,s=n.find(c=>c.id===o);if(s){const c=r.position.x+r.width-s.width,d=r.position.y+r.height-s.height;if(c>0||d>0||r.position.x<0||r.position.y<0){if(s.style={...s.style},s.style.width=s.style.width??s.width,s.style.height=s.style.height??s.height,c>0&&(s.style.width+=c),d>0&&(s.style.height+=d),r.position.x<0){const h=Math.abs(r.position.x);s.position.x=s.position.x-h,s.style.width+=h,r.position.x=0}if(r.position.y<0){const h=Math.abs(r.position.y);s.position.y=s.position.y-h,s.style.height+=h,r.position.y=0}s.width=s.style.width,s.height=s.style.height}}}function acn(n,r){if(n.some(s=>s.type==="reset"))return n.filter(s=>s.type==="reset").map(s=>s.item);const o=n.filter(s=>s.type==="add").map(s=>s.item);return r.reduce((s,c)=>{const d=n.filter(b=>b.id===c.id);if(d.length===0)return s.push(c),s;const h={...c};for(const b of d)if(b)switch(b.type){case"select":{h.selected=b.selected;break}case"position":{typeof b.position<"u"&&(h.position=b.position),typeof b.positionAbsolute<"u"&&(h.positionAbsolute=b.positionAbsolute),typeof b.dragging<"u"&&(h.dragging=b.dragging),h.expandParent&&hFt(s,h);break}case"dimensions":{typeof b.dimensions<"u"&&(h.width=b.dimensions.width,h.height=b.dimensions.height),typeof b.updateStyle<"u"&&(h.style={...h.style||{},...b.dimensions}),typeof b.resizing=="boolean"&&(h.resizing=b.resizing),h.expandParent&&hFt(s,h);break}case"remove":return s}return s.push(h),s},o)}function scn(n,r){return acn(n,r)}function w6r(n,r){return acn(n,r)}const oM=(n,r)=>({id:n,type:"select",selected:r});function dW(n,r){return n.reduce((o,s)=>{const c=r.includes(s.id);return!s.selected&&c?(s.selected=!0,o.push(oM(s.id,!0))):s.selected&&!c&&(s.selected=!1,o.push(oM(s.id,!1))),o},[])}const q$e=(n,r)=>o=>{o.target===r.current&&n?.(o)},y6r=n=>({userSelectionActive:n.userSelectionActive,elementsSelectable:n.elementsSelectable,dragging:n.paneDragging}),ccn=Q.memo(({isSelecting:n,selectionMode:r=noe.Full,panOnDrag:o,onSelectionStart:s,onSelectionEnd:c,onPaneClick:d,onPaneContextMenu:h,onPaneScroll:b,onPaneMouseEnter:y,onPaneMouseMove:E,onPaneMouseLeave:k,children:x})=>{const O=Q.useRef(null),P=Wb(),D=Q.useRef(0),$=Q.useRef(0),U=Q.useRef(),{userSelectionActive:B,elementsSelectable:G,dragging:K}=Nh(y6r,ey),X=()=>{P.setState({userSelectionActive:!1,userSelectionRect:null}),D.current=0,$.current=0},ae=Ue=>{d?.(Ue),P.getState().resetSelectedElements(),P.setState({nodesSelectionActive:!1})},oe=Ue=>{if(Array.isArray(o)&&o?.includes(2)){Ue.preventDefault();return}h?.(Ue)},he=b?Ue=>b(Ue):void 0,ye=Ue=>{const{resetSelectedElements:Z,domNode:Re}=P.getState();if(U.current=Re?.getBoundingClientRect(),!G||!n||Ue.button!==0||Ue.target!==O.current||!U.current)return;const{x:_e,y:qe}=bM(Ue,U.current);Z(),P.setState({userSelectionRect:{width:0,height:0,startX:_e,startY:qe,x:_e,y:qe}}),s?.(Ue)},we=Ue=>{const{userSelectionRect:Z,nodeInternals:Re,edges:_e,transform:qe,onNodesChange:it,onEdgesChange:Xe,nodeOrigin:de,getNodes:We}=P.getState();if(!n||!U.current||!Z)return;P.setState({userSelectionActive:!0,nodesSelectionActive:!1});const ht=bM(Ue,U.current),Ie=Z.startX??0,$t=Z.startY??0,An={...Z,x:ht.x<Ie?ht.x:Ie,y:ht.y<$t?ht.y:$t,width:Math.abs(ht.x-Ie),height:Math.abs(ht.y-$t)},Yt=We(),Or=zsn(Re,An,qe,r===noe.Partial,!0,de),Nt=Hsn(Or,_e).map(br=>br.id),on=Or.map(br=>br.id);if(D.current!==on.length){D.current=on.length;const br=dW(Yt,on);br.length&&it?.(br)}if($.current!==Nt.length){$.current=Nt.length;const br=dW(_e,Nt);br.length&&Xe?.(br)}P.setState({userSelectionRect:An})},ge=Ue=>{if(Ue.button!==0)return;const{userSelectionRect:Z}=P.getState();!B&&Z&&Ue.target===O.current&&ae?.(Ue),P.setState({nodesSelectionActive:D.current>0}),X(),c?.(Ue)},Te=Ue=>{B&&(P.setState({nodesSelectionActive:D.current>0}),c?.(Ue)),X()},xe=G&&(n||B);return Un.createElement("div",{className:q_(["react-flow__pane",{dragging:K,selection:n}]),onClick:xe?void 0:q$e(ae,O),onContextMenu:q$e(oe,O),onWheel:q$e(he,O),onMouseEnter:xe?void 0:y,onMouseDown:xe?ye:void 0,onMouseMove:xe?we:E,onMouseUp:xe?ge:void 0,onMouseLeave:xe?Te:k,ref:O,style:FZe},x,Un.createElement(v6r,null))});ccn.displayName="Pane";function ucn(n,r){const o=n.parentNode||n.parentId;if(!o)return!1;const s=r.get(o);return s?s.selected?!0:ucn(s,r):!1}function pFt(n,r,o){let s=n;do{if(s?.matches(r))return!0;if(s===o.current)return!1;s=s.parentElement}while(s);return!1}function E6r(n,r,o,s){return Array.from(n.values()).filter(c=>(c.selected||c.id===s)&&(!c.parentNode||c.parentId||!ucn(c,n))&&(c.draggable||r&&typeof c.draggable>"u")).map(c=>({id:c.id,position:c.position||{x:0,y:0},positionAbsolute:c.positionAbsolute||{x:0,y:0},distance:{x:o.x-(c.positionAbsolute?.x??0),y:o.y-(c.positionAbsolute?.y??0)},delta:{x:0,y:0},extent:c.extent,parentNode:c.parentNode||c.parentId,parentId:c.parentNode||c.parentId,width:c.width,height:c.height,expandParent:c.expandParent}))}function S6r(n,r){return!r||r==="parent"?r:[r[0],[r[1][0]-(n.width||0),r[1][1]-(n.height||0)]]}function lcn(n,r,o,s,c=[0,0],d){const h=S6r(n,n.extent||s);let b=h;const y=n.parentNode||n.parentId;if(n.extent==="parent"&&!n.expandParent)if(y&&n.width&&n.height){const x=o.get(y),{x:O,y:P}=LW(x,c).positionAbsolute;b=x&&M_(O)&&M_(P)&&M_(x.width)&&M_(x.height)?[[O+n.width*c[0],P+n.height*c[1]],[O+x.width-n.width+n.width*c[0],P+x.height-n.height+n.height*c[1]]]:b}else d?.("005",W4.error005()),b=h;else if(n.extent&&y&&n.extent!=="parent"){const x=o.get(y),{x:O,y:P}=LW(x,c).positionAbsolute;b=[[n.extent[0][0]+O,n.extent[0][1]+P],[n.extent[1][0]+O,n.extent[1][1]+P]]}let E={x:0,y:0};if(y){const x=o.get(y);E=LW(x,c).positionAbsolute}const k=b&&b!=="parent"?IZe(r,b):r;return{position:{x:k.x-E.x,y:k.y-E.y},positionAbsolute:k}}function V$e({nodeId:n,dragItems:r,nodeInternals:o}){const s=r.map(c=>({...o.get(c.id),position:c.position,positionAbsolute:c.positionAbsolute}));return[n?s.find(c=>c.id===n):s[0],s]}const gFt=(n,r,o,s)=>{const c=r.querySelectorAll(n);if(!c||!c.length)return null;const d=Array.from(c),h=r.getBoundingClientRect(),b={x:h.width*s[0],y:h.height*s[1]};return d.map(y=>{const E=y.getBoundingClientRect();return{id:y.getAttribute("data-handleid"),position:y.getAttribute("data-handlepos"),x:(E.left-h.left-b.x)/o,y:(E.top-h.top-b.y)/o,...AZe(y)}})};function bre(n,r,o){return o===void 0?o:s=>{const c=r().nodeInternals.get(n);c&&o(s,{...c})}}function mWe({id:n,store:r,unselect:o=!1,nodeRef:s}){const{addSelectedNodes:c,unselectNodesAndEdges:d,multiSelectionActive:h,nodeInternals:b,onError:y}=r.getState(),E=b.get(n);if(!E){y?.("012",W4.error012(n));return}r.setState({nodesSelectionActive:!1}),E.selected?(o||E.selected&&h)&&(d({nodes:[E],edges:[]}),requestAnimationFrame(()=>s?.current?.blur())):c([n])}function _6r(){const n=Wb();return Q.useCallback(({sourceEvent:o})=>{const{transform:s,snapGrid:c,snapToGrid:d}=n.getState(),h=o.touches?o.touches[0].clientX:o.clientX,b=o.touches?o.touches[0].clientY:o.clientY,y={x:(h-s[0])/s[2],y:(b-s[1])/s[2]};return{xSnapped:d?c[0]*Math.round(y.x/c[0]):y.x,ySnapped:d?c[1]*Math.round(y.y/c[1]):y.y,...y}},[])}function W$e(n){return(r,o,s)=>n?.(r,s)}function dcn({nodeRef:n,disabled:r=!1,noDragClassName:o,handleSelector:s,nodeId:c,isSelectable:d,selectNodesOnDrag:h}){const b=Wb(),[y,E]=Q.useState(!1),k=Q.useRef([]),x=Q.useRef({x:null,y:null}),O=Q.useRef(0),P=Q.useRef(null),D=Q.useRef({x:0,y:0}),$=Q.useRef(null),U=Q.useRef(!1),B=Q.useRef(!1),G=Q.useRef(!1),K=_6r();return Q.useEffect(()=>{if(n?.current){const X=Rx(n.current),ae=({x:ye,y:we})=>{const{nodeInternals:ge,onNodeDrag:Te,onSelectionDrag:xe,updateNodePositions:Ue,nodeExtent:Z,snapGrid:Re,snapToGrid:_e,nodeOrigin:qe,onError:it}=b.getState();x.current={x:ye,y:we};let Xe=!1,de={x:0,y:0,x2:0,y2:0};if(k.current.length>1&&Z){const ht=DZe(k.current,qe);de=OZe(ht)}if(k.current=k.current.map(ht=>{const Ie={x:ye-ht.distance.x,y:we-ht.distance.y};_e&&(Ie.x=Re[0]*Math.round(Ie.x/Re[0]),Ie.y=Re[1]*Math.round(Ie.y/Re[1]));const $t=[[Z[0][0],Z[0][1]],[Z[1][0],Z[1][1]]];k.current.length>1&&Z&&!ht.extent&&($t[0][0]=ht.positionAbsolute.x-de.x+Z[0][0],$t[1][0]=ht.positionAbsolute.x+(ht.width??0)-de.x2+Z[1][0],$t[0][1]=ht.positionAbsolute.y-de.y+Z[0][1],$t[1][1]=ht.positionAbsolute.y+(ht.height??0)-de.y2+Z[1][1]);const An=lcn(ht,Ie,ge,$t,qe,it);return Xe=Xe||ht.position.x!==An.position.x||ht.position.y!==An.position.y,ht.position=An.position,ht.positionAbsolute=An.positionAbsolute,ht}),!Xe)return;Ue(k.current,!0,!0),E(!0);const We=c?Te:W$e(xe);if(We&&$.current){const[ht,Ie]=V$e({nodeId:c,dragItems:k.current,nodeInternals:ge});We($.current,ht,Ie)}},oe=()=>{if(!P.current)return;const[ye,we]=Psn(D.current,P.current);if(ye!==0||we!==0){const{transform:ge,panBy:Te}=b.getState();x.current.x=(x.current.x??0)-ye/ge[2],x.current.y=(x.current.y??0)-we/ge[2],Te({x:ye,y:we})&&ae(x.current)}O.current=requestAnimationFrame(oe)},he=ye=>{const{nodeInternals:we,multiSelectionActive:ge,nodesDraggable:Te,unselectNodesAndEdges:xe,onNodeDragStart:Ue,onSelectionDragStart:Z}=b.getState();B.current=!0;const Re=c?Ue:W$e(Z);(!h||!d)&&!ge&&c&&(we.get(c)?.selected||xe()),c&&d&&h&&mWe({id:c,store:b,nodeRef:n});const _e=K(ye);if(x.current=_e,k.current=E6r(we,Te,_e,c),Re&&k.current){const[qe,it]=V$e({nodeId:c,dragItems:k.current,nodeInternals:we});Re(ye.sourceEvent,qe,it)}};if(r)X.on(".drag",null);else{const ye=NNr().on("start",we=>{const{domNode:ge,nodeDragThreshold:Te}=b.getState();Te===0&&he(we),G.current=!1;const xe=K(we);x.current=xe,P.current=ge?.getBoundingClientRect()||null,D.current=bM(we.sourceEvent,P.current)}).on("drag",we=>{const ge=K(we),{autoPanOnNodeDrag:Te,nodeDragThreshold:xe}=b.getState();if(we.sourceEvent.type==="touchmove"&&we.sourceEvent.touches.length>1&&(G.current=!0),!G.current){if(!U.current&&B.current&&Te&&(U.current=!0,oe()),!B.current){const Ue=ge.xSnapped-(x?.current?.x??0),Z=ge.ySnapped-(x?.current?.y??0);Math.sqrt(Ue*Ue+Z*Z)>xe&&he(we)}(x.current.x!==ge.xSnapped||x.current.y!==ge.ySnapped)&&k.current&&B.current&&($.current=we.sourceEvent,D.current=bM(we.sourceEvent,P.current),ae(ge))}}).on("end",we=>{if(!(!B.current||G.current)&&(E(!1),U.current=!1,B.current=!1,cancelAnimationFrame(O.current),k.current)){const{updateNodePositions:ge,nodeInternals:Te,onNodeDragStop:xe,onSelectionDragStop:Ue}=b.getState(),Z=c?xe:W$e(Ue);if(ge(k.current,!1,!1),Z){const[Re,_e]=V$e({nodeId:c,dragItems:k.current,nodeInternals:Te});Z(we.sourceEvent,Re,_e)}}}).filter(we=>{const ge=we.target;return!we.button&&(!o||!pFt(ge,`.${o}`,n))&&(!s||pFt(ge,s,n))});return X.call(ye),()=>{X.on(".drag",null)}}}},[n,r,o,s,d,b,c,h,K]),y}function fcn(){const n=Wb();return Q.useCallback(o=>{const{nodeInternals:s,nodeExtent:c,updateNodePositions:d,getNodes:h,snapToGrid:b,snapGrid:y,onError:E,nodesDraggable:k}=n.getState(),x=h().filter(G=>G.selected&&(G.draggable||k&&typeof G.draggable>"u")),O=b?y[0]:5,P=b?y[1]:5,D=o.isShiftPressed?4:1,$=o.x*O*D,U=o.y*P*D,B=x.map(G=>{if(G.positionAbsolute){const K={x:G.positionAbsolute.x+$,y:G.positionAbsolute.y+U};b&&(K.x=y[0]*Math.round(K.x/y[0]),K.y=y[1]*Math.round(K.y/y[1]));const{positionAbsolute:X,position:ae}=lcn(G,K,s,c,void 0,E);G.position=ae,G.positionAbsolute=X}return G});d(B,!0,!1)},[])}const DW={ArrowUp:{x:0,y:-1},ArrowDown:{x:0,y:1},ArrowLeft:{x:-1,y:0},ArrowRight:{x:1,y:0}};var vre=n=>{const r=({id:o,type:s,data:c,xPos:d,yPos:h,xPosOrigin:b,yPosOrigin:y,selected:E,onClick:k,onMouseEnter:x,onMouseMove:O,onMouseLeave:P,onContextMenu:D,onDoubleClick:$,style:U,className:B,isDraggable:G,isSelectable:K,isConnectable:X,isFocusable:ae,selectNodesOnDrag:oe,sourcePosition:he,targetPosition:ye,hidden:we,resizeObserver:ge,dragHandle:Te,zIndex:xe,isParent:Ue,noDragClassName:Z,noPanClassName:Re,initialized:_e,disableKeyboardA11y:qe,ariaLabel:it,rfId:Xe,hasHandleBounds:de})=>{const We=Wb(),ht=Q.useRef(null),Ie=Q.useRef(null),$t=Q.useRef(he),An=Q.useRef(ye),Yt=Q.useRef(s),Or=K||G||k||x||O||P,Nt=fcn(),on=bre(o,We.getState,x),br=bre(o,We.getState,O),wi=bre(o,We.getState,P),no=bre(o,We.getState,D),Ri=bre(o,We.getState,$),ro=qi=>{const{nodeDragThreshold:yi}=We.getState();if(K&&(!oe||!G||yi>0)&&mWe({id:o,store:We,nodeRef:ht}),k){const lo=We.getState().nodeInternals.get(o);lo&&k(qi,{...lo})}},Ko=qi=>{if(!fWe(qi)&&!qe)if(Dsn.includes(qi.key)&&K){const yi=qi.key==="Escape";mWe({id:o,store:We,unselect:yi,nodeRef:ht})}else G&&E&&Object.prototype.hasOwnProperty.call(DW,qi.key)&&(We.setState({ariaLiveMessage:`Moved selected node ${qi.key.replace("Arrow","").toLowerCase()}. New position, x: ${~~d}, y: ${~~h}`}),Nt({x:DW[qi.key].x,y:DW[qi.key].y,isShiftPressed:qi.shiftKey}))};Q.useEffect(()=>()=>{Ie.current&&(ge?.unobserve(Ie.current),Ie.current=null)},[]),Q.useEffect(()=>{if(ht.current&&!we){const qi=ht.current;(!_e||!de||Ie.current!==qi)&&(Ie.current&&ge?.unobserve(Ie.current),ge?.observe(qi),Ie.current=qi)}},[we,_e,de]),Q.useEffect(()=>{const qi=Yt.current!==s,yi=$t.current!==he,lo=An.current!==ye;ht.current&&(qi||yi||lo)&&(qi&&(Yt.current=s),yi&&($t.current=he),lo&&(An.current=ye),We.getState().updateNodeDimensions([{id:o,nodeElement:ht.current,forceUpdate:!0}]))},[o,s,he,ye]);const Ca=dcn({nodeRef:ht,disabled:we||!G,noDragClassName:Z,handleSelector:Te,nodeId:o,isSelectable:K,selectNodesOnDrag:oe});return we?null:Un.createElement("div",{className:q_(["react-flow__node",`react-flow__node-${s}`,{[Re]:G},B,{selected:E,selectable:K,parent:Ue,dragging:Ca}]),ref:ht,style:{zIndex:xe,transform:`translate(${b}px,${y}px)`,pointerEvents:Or?"all":"none",visibility:_e?"visible":"hidden",...U},"data-id":o,"data-testid":`rf__node-${o}`,onMouseEnter:on,onMouseMove:br,onMouseLeave:wi,onContextMenu:no,onClick:ro,onDoubleClick:Ri,onKeyDown:ae?Ko:void 0,tabIndex:ae?0:void 0,role:ae?"button":void 0,"aria-describedby":qe?void 0:`${tcn}-${Xe}`,"aria-label":it},Un.createElement(j4r,{value:o},Un.createElement(n,{id:o,data:c,type:s,xPos:d,yPos:h,selected:E,isConnectable:X,sourcePosition:he,targetPosition:ye,dragging:Ca,dragHandle:Te,zIndex:xe})))};return r.displayName="NodeWrapper",Q.memo(r)};const k6r=n=>{const r=n.getNodes().filter(o=>o.selected);return{...DZe(r,n.nodeOrigin),transformString:`translate(${n.transform[0]}px,${n.transform[1]}px) scale(${n.transform[2]})`,userSelectionActive:n.userSelectionActive}};function T6r({onSelectionContextMenu:n,noPanClassName:r,disableKeyboardA11y:o}){const s=Wb(),{width:c,height:d,x:h,y:b,transformString:y,userSelectionActive:E}=Nh(k6r,ey),k=fcn(),x=Q.useRef(null);if(Q.useEffect(()=>{o||x.current?.focus({preventScroll:!0})},[o]),dcn({nodeRef:x}),E||!c||!d)return null;const O=n?D=>{const $=s.getState().getNodes().filter(U=>U.selected);n(D,$)}:void 0,P=D=>{Object.prototype.hasOwnProperty.call(DW,D.key)&&k({x:DW[D.key].x,y:DW[D.key].y,isShiftPressed:D.shiftKey})};return Un.createElement("div",{className:q_(["react-flow__nodesselection","react-flow__container",r]),style:{transform:y}},Un.createElement("div",{ref:x,className:"react-flow__nodesselection-rect",onContextMenu:O,tabIndex:o?void 0:-1,onKeyDown:o?void 0:P,style:{width:c,height:d,top:b,left:h}}))}var x6r=Q.memo(T6r);const C6r=n=>n.nodesSelectionActive,hcn=({children:n,onPaneClick:r,onPaneMouseEnter:o,onPaneMouseMove:s,onPaneMouseLeave:c,onPaneContextMenu:d,onPaneScroll:h,deleteKeyCode:b,onMove:y,onMoveStart:E,onMoveEnd:k,selectionKeyCode:x,selectionOnDrag:O,selectionMode:P,onSelectionStart:D,onSelectionEnd:$,multiSelectionKeyCode:U,panActivationKeyCode:B,zoomActivationKeyCode:G,elementsSelectable:K,zoomOnScroll:X,zoomOnPinch:ae,panOnScroll:oe,panOnScrollSpeed:he,panOnScrollMode:ye,zoomOnDoubleClick:we,panOnDrag:ge,defaultViewport:Te,translateExtent:xe,minZoom:Ue,maxZoom:Z,preventScrolling:Re,onSelectionContextMenu:_e,noWheelClassName:qe,noPanClassName:it,disableKeyboardA11y:Xe})=>{const de=Nh(C6r),We=ooe(x),ht=ooe(B),Ie=ht||ge,$t=ht||oe,An=We||O&&Ie!==!0;return f6r({deleteKeyCode:b,multiSelectionKeyCode:U}),Un.createElement(m6r,{onMove:y,onMoveStart:E,onMoveEnd:k,onPaneContextMenu:d,elementsSelectable:K,zoomOnScroll:X,zoomOnPinch:ae,panOnScroll:$t,panOnScrollSpeed:he,panOnScrollMode:ye,zoomOnDoubleClick:we,panOnDrag:!We&&Ie,defaultViewport:Te,translateExtent:xe,minZoom:Ue,maxZoom:Z,zoomActivationKeyCode:G,preventScrolling:Re,noWheelClassName:qe,noPanClassName:it},Un.createElement(ccn,{onSelectionStart:D,onSelectionEnd:$,onPaneClick:r,onPaneMouseEnter:o,onPaneMouseMove:s,onPaneMouseLeave:c,onPaneContextMenu:d,onPaneScroll:h,panOnDrag:Ie,isSelecting:!!An,selectionMode:P},n,de&&Un.createElement(x6r,{onSelectionContextMenu:_e,noPanClassName:it,disableKeyboardA11y:Xe})))};hcn.displayName="FlowRenderer";var A6r=Q.memo(hcn);function I6r(n){return Nh(Q.useCallback(o=>n?zsn(o.nodeInternals,{x:0,y:0,width:o.width,height:o.height},o.transform,!0):o.getNodes(),[n]))}function O6r(n){const r={input:vre(n.input||Jsn),default:vre(n.default||gWe),output:vre(n.output||Qsn),group:vre(n.group||jZe)},o={},s=Object.keys(n).filter(c=>!["input","default","output","group"].includes(c)).reduce((c,d)=>(c[d]=vre(n[d]||gWe),c),o);return{...r,...s}}const N6r=({x:n,y:r,width:o,height:s,origin:c})=>!o||!s?{x:n,y:r}:c[0]<0||c[1]<0||c[0]>1||c[1]>1?{x:n,y:r}:{x:n-o*c[0],y:r-s*c[1]},R6r=n=>({nodesDraggable:n.nodesDraggable,nodesConnectable:n.nodesConnectable,nodesFocusable:n.nodesFocusable,elementsSelectable:n.elementsSelectable,updateNodeDimensions:n.updateNodeDimensions,onError:n.onError}),pcn=n=>{const{nodesDraggable:r,nodesConnectable:o,nodesFocusable:s,elementsSelectable:c,updateNodeDimensions:d,onError:h}=Nh(R6r,ey),b=I6r(n.onlyRenderVisibleElements),y=Q.useRef(),E=Q.useMemo(()=>{if(typeof ResizeObserver>"u")return null;const k=new ResizeObserver(x=>{const O=x.map(P=>({id:P.target.getAttribute("data-id"),nodeElement:P.target,forceUpdate:!0}));d(O)});return y.current=k,k},[]);return Q.useEffect(()=>()=>{y?.current?.disconnect()},[]),Un.createElement("div",{className:"react-flow__nodes",style:FZe},b.map(k=>{let x=k.type||"default";n.nodeTypes[x]||(h?.("003",W4.error003(x)),x="default");const O=n.nodeTypes[x]||n.nodeTypes.default,P=!!(k.draggable||r&&typeof k.draggable>"u"),D=!!(k.selectable||c&&typeof k.selectable>"u"),$=!!(k.connectable||o&&typeof k.connectable>"u"),U=!!(k.focusable||s&&typeof k.focusable>"u"),B=n.nodeExtent?IZe(k.positionAbsolute,n.nodeExtent):k.positionAbsolute,G=B?.x??0,K=B?.y??0,X=N6r({x:G,y:K,width:k.width??0,height:k.height??0,origin:n.nodeOrigin});return Un.createElement(O,{key:k.id,id:k.id,className:k.className,style:k.style,type:x,data:k.data,sourcePosition:k.sourcePosition||ni.Bottom,targetPosition:k.targetPosition||ni.Top,hidden:k.hidden,xPos:G,yPos:K,xPosOrigin:X.x,yPosOrigin:X.y,selectNodesOnDrag:n.selectNodesOnDrag,onClick:n.onNodeClick,onMouseEnter:n.onNodeMouseEnter,onMouseMove:n.onNodeMouseMove,onMouseLeave:n.onNodeMouseLeave,onContextMenu:n.onNodeContextMenu,onDoubleClick:n.onNodeDoubleClick,selected:!!k.selected,isDraggable:P,isSelectable:D,isConnectable:$,isFocusable:U,resizeObserver:E,dragHandle:k.dragHandle,zIndex:k[Oh]?.z??0,isParent:!!k[Oh]?.isParent,noDragClassName:n.noDragClassName,noPanClassName:n.noPanClassName,initialized:!!k.width&&!!k.height,rfId:n.rfId,disableKeyboardA11y:n.disableKeyboardA11y,ariaLabel:k.ariaLabel,hasHandleBounds:!!k[Oh]?.handleBounds})}))};pcn.displayName="NodeRenderer";var P6r=Q.memo(pcn);const L6r=(n,r,o)=>o===ni.Left?n-r:o===ni.Right?n+r:n,D6r=(n,r,o)=>o===ni.Top?n-r:o===ni.Bottom?n+r:n,mFt="react-flow__edgeupdater",bFt=({position:n,centerX:r,centerY:o,radius:s=10,onMouseDown:c,onMouseEnter:d,onMouseOut:h,type:b})=>Un.createElement("circle",{onMouseDown:c,onMouseEnter:d,onMouseOut:h,className:q_([mFt,`${mFt}-${b}`]),cx:L6r(r,s,n),cy:D6r(o,s,n),r:s,stroke:"transparent",fill:"transparent"}),M6r=()=>!0;var WV=n=>{const r=({id:o,className:s,type:c,data:d,onClick:h,onEdgeDoubleClick:b,selected:y,animated:E,label:k,labelStyle:x,labelShowBg:O,labelBgStyle:P,labelBgPadding:D,labelBgBorderRadius:$,style:U,source:B,target:G,sourceX:K,sourceY:X,targetX:ae,targetY:oe,sourcePosition:he,targetPosition:ye,elementsSelectable:we,hidden:ge,sourceHandleId:Te,targetHandleId:xe,onContextMenu:Ue,onMouseEnter:Z,onMouseMove:Re,onMouseLeave:_e,reconnectRadius:qe,onReconnect:it,onReconnectStart:Xe,onReconnectEnd:de,markerEnd:We,markerStart:ht,rfId:Ie,ariaLabel:$t,isFocusable:An,isReconnectable:Yt,pathOptions:Or,interactionWidth:Nt,disableKeyboardA11y:on})=>{const br=Q.useRef(null),[wi,no]=Q.useState(!1),[Ri,ro]=Q.useState(!1),Ko=Wb(),Ca=Q.useMemo(()=>`url('#${hWe(ht,Ie)}')`,[ht,Ie]),qi=Q.useMemo(()=>`url('#${hWe(We,Ie)}')`,[We,Ie]);if(ge)return null;const yi=bs=>{const{edges:Cc,addSelectedEdges:Ac,unselectNodesAndEdges:Xc,multiSelectionActive:rc}=Ko.getState(),js=Cc.find(Jc=>Jc.id===o);js&&(we&&(Ko.setState({nodesSelectionActive:!1}),js.selected&&rc?(Xc({nodes:[],edges:[js]}),br.current?.blur()):Ac([o])),h&&h(bs,js))},lo=mre(o,Ko.getState,b),Eo=mre(o,Ko.getState,Ue),Fc=mre(o,Ko.getState,Z),Yo=mre(o,Ko.getState,Re),zl=mre(o,Ko.getState,_e),ld=(bs,Cc)=>{if(bs.button!==0)return;const{edges:Ac,isValidConnection:Xc}=Ko.getState(),rc=Cc?G:B,js=(Cc?xe:Te)||null,Jc=Cc?"target":"source",Id=Xc||M6r,nh=Cc,hu=Ac.find(Km=>Km.id===o);ro(!0),Xe?.(bs,hu,Jc);const Qp=Km=>{ro(!1),de?.(Km,hu,Jc)};Wsn({event:bs,handleId:js,nodeId:rc,onConnect:Km=>it?.(hu,Km),isTarget:nh,getState:Ko.getState,setState:Ko.setState,isValidConnection:Id,edgeUpdaterType:Jc,onReconnectEnd:Qp})},Yc=bs=>ld(bs,!0),Sn=bs=>ld(bs,!1),Pn=()=>no(!0),Kr=()=>no(!1),Ei=!we&&!h,Sa=bs=>{if(!on&&Dsn.includes(bs.key)&&we){const{unselectNodesAndEdges:Cc,addSelectedEdges:Ac,edges:Xc}=Ko.getState();bs.key==="Escape"?(br.current?.blur(),Cc({edges:[Xc.find(js=>js.id===o)]})):Ac([o])}};return Un.createElement("g",{className:q_(["react-flow__edge",`react-flow__edge-${c}`,s,{selected:y,animated:E,inactive:Ei,updating:wi}]),onClick:yi,onDoubleClick:lo,onContextMenu:Eo,onMouseEnter:Fc,onMouseMove:Yo,onMouseLeave:zl,onKeyDown:An?Sa:void 0,tabIndex:An?0:void 0,role:An?"button":"img","data-testid":`rf__edge-${o}`,"aria-label":$t===null?void 0:$t||`Edge from ${B} to ${G}`,"aria-describedby":An?`${ncn}-${Ie}`:void 0,ref:br},!Ri&&Un.createElement(n,{id:o,source:B,target:G,selected:y,animated:E,label:k,labelStyle:x,labelShowBg:O,labelBgStyle:P,labelBgPadding:D,labelBgBorderRadius:$,data:d,style:U,sourceX:K,sourceY:X,targetX:ae,targetY:oe,sourcePosition:he,targetPosition:ye,sourceHandleId:Te,targetHandleId:xe,markerStart:Ca,markerEnd:qi,pathOptions:Or,interactionWidth:Nt}),Yt&&Un.createElement(Un.Fragment,null,(Yt==="source"||Yt===!0)&&Un.createElement(bFt,{position:he,centerX:K,centerY:X,radius:qe,onMouseDown:Yc,onMouseEnter:Pn,onMouseOut:Kr,type:"source"}),(Yt==="target"||Yt===!0)&&Un.createElement(bFt,{position:ye,centerX:ae,centerY:oe,radius:qe,onMouseDown:Sn,onMouseEnter:Pn,onMouseOut:Kr,type:"target"})))};return r.displayName="EdgeWrapper",Q.memo(r)};function j6r(n){const r={default:WV(n.default||_we),straight:WV(n.bezier||PZe),step:WV(n.step||RZe),smoothstep:WV(n.step||pEe),simplebezier:WV(n.simplebezier||NZe)},o={},s=Object.keys(n).filter(c=>!["default","bezier"].includes(c)).reduce((c,d)=>(c[d]=WV(n[d]||_we),c),o);return{...r,...s}}function vFt(n,r,o=null){const s=(o?.x||0)+r.x,c=(o?.y||0)+r.y,d=o?.width||r.width,h=o?.height||r.height;switch(n){case ni.Top:return{x:s+d/2,y:c};case ni.Right:return{x:s+d,y:c+h/2};case ni.Bottom:return{x:s+d/2,y:c+h};case ni.Left:return{x:s,y:c+h/2}}}function wFt(n,r){return n?n.length===1||!r?n[0]:r&&n.find(o=>o.id===r)||null:null}const $6r=(n,r,o,s,c,d)=>{const h=vFt(o,n,r),b=vFt(d,s,c);return{sourceX:h.x,sourceY:h.y,targetX:b.x,targetY:b.y}};function F6r({sourcePos:n,targetPos:r,sourceWidth:o,sourceHeight:s,targetWidth:c,targetHeight:d,width:h,height:b,transform:y}){const E={x:Math.min(n.x,r.x),y:Math.min(n.y,r.y),x2:Math.max(n.x+o,r.x+c),y2:Math.max(n.y+s,r.y+d)};E.x===E.x2&&(E.x2+=1),E.y===E.y2&&(E.y2+=1);const k=OZe({x:(0-y[0])/y[2],y:(0-y[1])/y[2],width:h/y[2],height:b/y[2]}),x=Math.max(0,Math.min(k.x2,E.x2)-Math.max(k.x,E.x)),O=Math.max(0,Math.min(k.y2,E.y2)-Math.max(k.y,E.y));return Math.ceil(x*O)>0}function yFt(n){const r=n?.[Oh]?.handleBounds||null,o=r&&n?.width&&n?.height&&typeof n?.positionAbsolute?.x<"u"&&typeof n?.positionAbsolute?.y<"u";return[{x:n?.positionAbsolute?.x||0,y:n?.positionAbsolute?.y||0,width:n?.width||0,height:n?.height||0},r,!!o]}const B6r=[{level:0,isMaxLevel:!0,edges:[]}];function U6r(n,r,o=!1){let s=-1;const c=n.reduce((h,b)=>{const y=M_(b.zIndex);let E=y?b.zIndex:0;if(o){const k=r.get(b.target),x=r.get(b.source),O=b.selected||k?.selected||x?.selected,P=Math.max(x?.[Oh]?.z||0,k?.[Oh]?.z||0,1e3);E=(y?b.zIndex:0)+(O?P:0)}return h[E]?h[E].push(b):h[E]=[b],s=E>s?E:s,h},{}),d=Object.entries(c).map(([h,b])=>{const y=+h;return{edges:b,level:y,isMaxLevel:y===s}});return d.length===0?B6r:d}function z6r(n,r,o){const s=Nh(Q.useCallback(c=>n?c.edges.filter(d=>{const h=r.get(d.source),b=r.get(d.target);return h?.width&&h?.height&&b?.width&&b?.height&&F6r({sourcePos:h.positionAbsolute||{x:0,y:0},targetPos:b.positionAbsolute||{x:0,y:0},sourceWidth:h.width,sourceHeight:h.height,targetWidth:b.width,targetHeight:b.height,width:c.width,height:c.height,transform:c.transform})}):c.edges,[n,r]));return U6r(s,r,o)}const H6r=({color:n="none",strokeWidth:r=1})=>Un.createElement("polyline",{style:{stroke:n,strokeWidth:r},strokeLinecap:"round",strokeLinejoin:"round",fill:"none",points:"-5,-4 0,0 -5,4"}),G6r=({color:n="none",strokeWidth:r=1})=>Un.createElement("polyline",{style:{stroke:n,fill:n,strokeWidth:r},strokeLinecap:"round",strokeLinejoin:"round",points:"-5,-4 0,0 -5,4 -5,-4"}),EFt={[roe.Arrow]:H6r,[roe.ArrowClosed]:G6r};function q6r(n){const r=Wb();return Q.useMemo(()=>Object.prototype.hasOwnProperty.call(EFt,n)?EFt[n]:(r.getState().onError?.("009",W4.error009(n)),null),[n])}const V6r=({id:n,type:r,color:o,width:s=12.5,height:c=12.5,markerUnits:d="strokeWidth",strokeWidth:h,orient:b="auto-start-reverse"})=>{const y=q6r(r);return y?Un.createElement("marker",{className:"react-flow__arrowhead",id:n,markerWidth:`${s}`,markerHeight:`${c}`,viewBox:"-10 -10 20 20",markerUnits:d,orient:b,refX:"0",refY:"0"},Un.createElement(y,{color:o,strokeWidth:h})):null},W6r=({defaultColor:n,rfId:r})=>o=>{const s=[];return o.edges.reduce((c,d)=>([d.markerStart,d.markerEnd].forEach(h=>{if(h&&typeof h=="object"){const b=hWe(h,r);s.includes(b)||(c.push({id:b,color:h.color||n,...h}),s.push(b))}}),c),[]).sort((c,d)=>c.id.localeCompare(d.id))},gcn=({defaultColor:n,rfId:r})=>{const o=Nh(Q.useCallback(W6r({defaultColor:n,rfId:r}),[n,r]),(s,c)=>!(s.length!==c.length||s.some((d,h)=>d.id!==c[h].id)));return Un.createElement("defs",null,o.map(s=>Un.createElement(V6r,{id:s.id,key:s.id,type:s.type,color:s.color,width:s.width,height:s.height,markerUnits:s.markerUnits,strokeWidth:s.strokeWidth,orient:s.orient})))};gcn.displayName="MarkerDefinitions";var K6r=Q.memo(gcn);const Y6r=n=>({nodesConnectable:n.nodesConnectable,edgesFocusable:n.edgesFocusable,edgesUpdatable:n.edgesUpdatable,elementsSelectable:n.elementsSelectable,width:n.width,height:n.height,connectionMode:n.connectionMode,nodeInternals:n.nodeInternals,onError:n.onError}),mcn=({defaultMarkerColor:n,onlyRenderVisibleElements:r,elevateEdgesOnSelect:o,rfId:s,edgeTypes:c,noPanClassName:d,onEdgeContextMenu:h,onEdgeMouseEnter:b,onEdgeMouseMove:y,onEdgeMouseLeave:E,onEdgeClick:k,onEdgeDoubleClick:x,onReconnect:O,onReconnectStart:P,onReconnectEnd:D,reconnectRadius:$,children:U,disableKeyboardA11y:B})=>{const{edgesFocusable:G,edgesUpdatable:K,elementsSelectable:X,width:ae,height:oe,connectionMode:he,nodeInternals:ye,onError:we}=Nh(Y6r,ey),ge=z6r(r,ye,o);return ae?Un.createElement(Un.Fragment,null,ge.map(({level:Te,edges:xe,isMaxLevel:Ue})=>Un.createElement("svg",{key:Te,style:{zIndex:Te},width:ae,height:oe,className:"react-flow__edges react-flow__container"},Ue&&Un.createElement(K6r,{defaultColor:n,rfId:s}),Un.createElement("g",null,xe.map(Z=>{const[Re,_e,qe]=yFt(ye.get(Z.source)),[it,Xe,de]=yFt(ye.get(Z.target));if(!qe||!de)return null;let We=Z.type||"default";c[We]||(we?.("011",W4.error011(We)),We="default");const ht=c[We]||c.default,Ie=he===nU.Strict?Xe.target:(Xe.target??[]).concat(Xe.source??[]),$t=wFt(_e.source,Z.sourceHandle),An=wFt(Ie,Z.targetHandle),Yt=$t?.position||ni.Bottom,Or=An?.position||ni.Top,Nt=!!(Z.focusable||G&&typeof Z.focusable>"u"),on=Z.reconnectable||Z.updatable,br=typeof O<"u"&&(on||K&&typeof on>"u");if(!$t||!An)return we?.("008",W4.error008($t,Z)),null;const{sourceX:wi,sourceY:no,targetX:Ri,targetY:ro}=$6r(Re,$t,Yt,it,An,Or);return Un.createElement(ht,{key:Z.id,id:Z.id,className:q_([Z.className,d]),type:We,data:Z.data,selected:!!Z.selected,animated:!!Z.animated,hidden:!!Z.hidden,label:Z.label,labelStyle:Z.labelStyle,labelShowBg:Z.labelShowBg,labelBgStyle:Z.labelBgStyle,labelBgPadding:Z.labelBgPadding,labelBgBorderRadius:Z.labelBgBorderRadius,style:Z.style,source:Z.source,target:Z.target,sourceHandleId:Z.sourceHandle,targetHandleId:Z.targetHandle,markerEnd:Z.markerEnd,markerStart:Z.markerStart,sourceX:wi,sourceY:no,targetX:Ri,targetY:ro,sourcePosition:Yt,targetPosition:Or,elementsSelectable:X,onContextMenu:h,onMouseEnter:b,onMouseMove:y,onMouseLeave:E,onClick:k,onEdgeDoubleClick:x,onReconnect:O,onReconnectStart:P,onReconnectEnd:D,reconnectRadius:$,rfId:s,ariaLabel:Z.ariaLabel,isFocusable:Nt,isReconnectable:br,pathOptions:"pathOptions"in Z?Z.pathOptions:void 0,interactionWidth:Z.interactionWidth,disableKeyboardA11y:B})})))),U):null};mcn.displayName="EdgeRenderer";var X6r=Q.memo(mcn);const J6r=n=>`translate(${n.transform[0]}px,${n.transform[1]}px) scale(${n.transform[2]})`;function Z6r({children:n}){const r=Nh(J6r);return Un.createElement("div",{className:"react-flow__viewport react-flow__container",style:{transform:r}},n)}function Q6r(n){const r=$Ze(),o=Q.useRef(!1);Q.useEffect(()=>{!o.current&&r.viewportInitialized&&n&&(setTimeout(()=>n(r),1),o.current=!0)},[n,r.viewportInitialized])}const e5r={[ni.Left]:ni.Right,[ni.Right]:ni.Left,[ni.Top]:ni.Bottom,[ni.Bottom]:ni.Top},bcn=({nodeId:n,handleType:r,style:o,type:s=aM.Bezier,CustomComponent:c,connectionStatus:d})=>{const{fromNode:h,handleId:b,toX:y,toY:E,connectionMode:k}=Nh(Q.useCallback(oe=>({fromNode:oe.nodeInternals.get(n),handleId:oe.connectionHandleId,toX:(oe.connectionPosition.x-oe.transform[0])/oe.transform[2],toY:(oe.connectionPosition.y-oe.transform[1])/oe.transform[2],connectionMode:oe.connectionMode}),[n]),ey),x=h?.[Oh]?.handleBounds;let O=x?.[r];if(k===nU.Loose&&(O=O||x?.[r==="source"?"target":"source"]),!h||!O)return null;const P=b?O.find(oe=>oe.id===b):O[0],D=P?P.x+P.width/2:(h.width??0)/2,$=P?P.y+P.height/2:h.height??0,U=(h.positionAbsolute?.x??0)+D,B=(h.positionAbsolute?.y??0)+$,G=P?.position,K=G?e5r[G]:null;if(!G||!K)return null;if(c)return Un.createElement(c,{connectionLineType:s,connectionLineStyle:o,fromNode:h,fromHandle:P,fromX:U,fromY:B,toX:y,toY:E,fromPosition:G,toPosition:K,connectionStatus:d});let X="";const ae={sourceX:U,sourceY:B,sourcePosition:G,targetX:y,targetY:E,targetPosition:K};return s===aM.Bezier?[X]=Bsn(ae):s===aM.Step?[X]=ioe({...ae,borderRadius:0}):s===aM.SmoothStep?[X]=ioe(ae):s===aM.SimpleBezier?[X]=Fsn(ae):X=`M${U},${B} ${y},${E}`,Un.createElement("path",{d:X,fill:"none",className:"react-flow__connection-path",style:o})};bcn.displayName="ConnectionLine";const t5r=n=>({nodeId:n.connectionNodeId,handleType:n.connectionHandleType,nodesConnectable:n.nodesConnectable,connectionStatus:n.connectionStatus,width:n.width,height:n.height});function n5r({containerStyle:n,style:r,type:o,component:s}){const{nodeId:c,handleType:d,nodesConnectable:h,width:b,height:y,connectionStatus:E}=Nh(t5r,ey);return!(c&&d&&b&&h)?null:Un.createElement("svg",{style:n,width:b,height:y,className:"react-flow__edges react-flow__connectionline react-flow__container"},Un.createElement("g",{className:q_(["react-flow__connection",E])},Un.createElement(bcn,{nodeId:c,handleType:d,style:r,type:o,CustomComponent:s,connectionStatus:E})))}function SFt(n,r){return Q.useRef(null),Wb(),Q.useMemo(()=>r(n),[n])}const vcn=({nodeTypes:n,edgeTypes:r,onMove:o,onMoveStart:s,onMoveEnd:c,onInit:d,onNodeClick:h,onEdgeClick:b,onNodeDoubleClick:y,onEdgeDoubleClick:E,onNodeMouseEnter:k,onNodeMouseMove:x,onNodeMouseLeave:O,onNodeContextMenu:P,onSelectionContextMenu:D,onSelectionStart:$,onSelectionEnd:U,connectionLineType:B,connectionLineStyle:G,connectionLineComponent:K,connectionLineContainerStyle:X,selectionKeyCode:ae,selectionOnDrag:oe,selectionMode:he,multiSelectionKeyCode:ye,panActivationKeyCode:we,zoomActivationKeyCode:ge,deleteKeyCode:Te,onlyRenderVisibleElements:xe,elementsSelectable:Ue,selectNodesOnDrag:Z,defaultViewport:Re,translateExtent:_e,minZoom:qe,maxZoom:it,preventScrolling:Xe,defaultMarkerColor:de,zoomOnScroll:We,zoomOnPinch:ht,panOnScroll:Ie,panOnScrollSpeed:$t,panOnScrollMode:An,zoomOnDoubleClick:Yt,panOnDrag:Or,onPaneClick:Nt,onPaneMouseEnter:on,onPaneMouseMove:br,onPaneMouseLeave:wi,onPaneScroll:no,onPaneContextMenu:Ri,onEdgeContextMenu:ro,onEdgeMouseEnter:Ko,onEdgeMouseMove:Ca,onEdgeMouseLeave:qi,onReconnect:yi,onReconnectStart:lo,onReconnectEnd:Eo,reconnectRadius:Fc,noDragClassName:Yo,noWheelClassName:zl,noPanClassName:ld,elevateEdgesOnSelect:Yc,disableKeyboardA11y:Sn,nodeOrigin:Pn,nodeExtent:Kr,rfId:Ei})=>{const Sa=SFt(n,O6r),bs=SFt(r,j6r);return Q6r(d),Un.createElement(A6r,{onPaneClick:Nt,onPaneMouseEnter:on,onPaneMouseMove:br,onPaneMouseLeave:wi,onPaneContextMenu:Ri,onPaneScroll:no,deleteKeyCode:Te,selectionKeyCode:ae,selectionOnDrag:oe,selectionMode:he,onSelectionStart:$,onSelectionEnd:U,multiSelectionKeyCode:ye,panActivationKeyCode:we,zoomActivationKeyCode:ge,elementsSelectable:Ue,onMove:o,onMoveStart:s,onMoveEnd:c,zoomOnScroll:We,zoomOnPinch:ht,zoomOnDoubleClick:Yt,panOnScroll:Ie,panOnScrollSpeed:$t,panOnScrollMode:An,panOnDrag:Or,defaultViewport:Re,translateExtent:_e,minZoom:qe,maxZoom:it,onSelectionContextMenu:D,preventScrolling:Xe,noDragClassName:Yo,noWheelClassName:zl,noPanClassName:ld,disableKeyboardA11y:Sn},Un.createElement(Z6r,null,Un.createElement(X6r,{edgeTypes:bs,onEdgeClick:b,onEdgeDoubleClick:E,onlyRenderVisibleElements:xe,onEdgeContextMenu:ro,onEdgeMouseEnter:Ko,onEdgeMouseMove:Ca,onEdgeMouseLeave:qi,onReconnect:yi,onReconnectStart:lo,onReconnectEnd:Eo,reconnectRadius:Fc,defaultMarkerColor:de,noPanClassName:ld,elevateEdgesOnSelect:!!Yc,disableKeyboardA11y:Sn,rfId:Ei},Un.createElement(n5r,{style:G,type:B,component:K,containerStyle:X})),Un.createElement("div",{className:"react-flow__edgelabel-renderer"}),Un.createElement(P6r,{nodeTypes:Sa,onNodeClick:h,onNodeDoubleClick:y,onNodeMouseEnter:k,onNodeMouseMove:x,onNodeMouseLeave:O,onNodeContextMenu:P,selectNodesOnDrag:Z,onlyRenderVisibleElements:xe,noPanClassName:ld,noDragClassName:Yo,disableKeyboardA11y:Sn,nodeOrigin:Pn,nodeExtent:Kr,rfId:Ei})))};vcn.displayName="GraphView";var r5r=Q.memo(vcn);const bWe=[[Number.NEGATIVE_INFINITY,Number.NEGATIVE_INFINITY],[Number.POSITIVE_INFINITY,Number.POSITIVE_INFINITY]],qD={rfId:"1",width:0,height:0,transform:[0,0,1],nodeInternals:new Map,edges:[],onNodesChange:null,onEdgesChange:null,hasDefaultNodes:!1,hasDefaultEdges:!1,d3Zoom:null,d3Selection:null,d3ZoomHandler:void 0,minZoom:.5,maxZoom:2,translateExtent:bWe,nodeExtent:bWe,nodesSelectionActive:!1,userSelectionActive:!1,userSelectionRect:null,connectionNodeId:null,connectionHandleId:null,connectionHandleType:"source",connectionPosition:{x:0,y:0},connectionStatus:null,connectionMode:nU.Strict,domNode:null,paneDragging:!1,noPanClassName:"nopan",nodeOrigin:[0,0],nodeDragThreshold:0,snapGrid:[15,15],snapToGrid:!1,nodesDraggable:!0,nodesConnectable:!0,nodesFocusable:!0,edgesFocusable:!0,edgesUpdatable:!0,elementsSelectable:!0,elevateNodesOnSelect:!0,fitViewOnInit:!1,fitViewOnInitDone:!1,fitViewOnInitOptions:void 0,onSelectionChange:[],multiSelectionActive:!1,connectionStartHandle:null,connectionEndHandle:null,connectionClickStartHandle:null,connectOnClick:!0,ariaLiveMessage:"",autoPanOnConnect:!0,autoPanOnNodeDrag:!0,connectionRadius:20,onError:N4r,isValidConnection:void 0},i5r=()=>HIr((n,r)=>({...qD,setNodes:o=>{const{nodeInternals:s,nodeOrigin:c,elevateNodesOnSelect:d}=r();n({nodeInternals:G$e(o,s,c,d)})},getNodes:()=>Array.from(r().nodeInternals.values()),setEdges:o=>{const{defaultEdgeOptions:s={}}=r();n({edges:o.map(c=>({...s,...c}))})},setDefaultNodesAndEdges:(o,s)=>{const c=typeof o<"u",d=typeof s<"u",h=c?G$e(o,new Map,r().nodeOrigin,r().elevateNodesOnSelect):new Map;n({nodeInternals:h,edges:d?s:[],hasDefaultNodes:c,hasDefaultEdges:d})},updateNodeDimensions:o=>{const{onNodesChange:s,nodeInternals:c,fitViewOnInit:d,fitViewOnInitDone:h,fitViewOnInitOptions:b,domNode:y,nodeOrigin:E}=r(),k=y?.querySelector(".react-flow__viewport");if(!k)return;const x=window.getComputedStyle(k),{m22:O}=new window.DOMMatrixReadOnly(x.transform),P=o.reduce(($,U)=>{const B=c.get(U.id);if(B?.hidden)c.set(B.id,{...B,[Oh]:{...B[Oh],handleBounds:void 0}});else if(B){const G=AZe(U.nodeElement);!!(G.width&&G.height&&(B.width!==G.width||B.height!==G.height||U.forceUpdate))&&(c.set(B.id,{...B,[Oh]:{...B[Oh],handleBounds:{source:gFt(".source",U.nodeElement,O,E),target:gFt(".target",U.nodeElement,O,E)}},...G}),$.push({id:B.id,type:"dimensions",dimensions:G}))}return $},[]);icn(c,E);const D=h||d&&!h&&ocn(r,{initial:!0,...b});n({nodeInternals:new Map(c),fitViewOnInitDone:D}),P?.length>0&&s?.(P)},updateNodePositions:(o,s=!0,c=!1)=>{const{triggerNodeChanges:d}=r(),h=o.map(b=>{const y={id:b.id,type:"position",dragging:c};return s&&(y.positionAbsolute=b.positionAbsolute,y.position=b.position),y});d(h)},triggerNodeChanges:o=>{const{onNodesChange:s,nodeInternals:c,hasDefaultNodes:d,nodeOrigin:h,getNodes:b,elevateNodesOnSelect:y}=r();if(o?.length){if(d){const E=scn(o,b()),k=G$e(E,c,h,y);n({nodeInternals:k})}s?.(o)}},addSelectedNodes:o=>{const{multiSelectionActive:s,edges:c,getNodes:d}=r();let h,b=null;s?h=o.map(y=>oM(y,!0)):(h=dW(d(),o),b=dW(c,[])),E0e({changedNodes:h,changedEdges:b,get:r,set:n})},addSelectedEdges:o=>{const{multiSelectionActive:s,edges:c,getNodes:d}=r();let h,b=null;s?h=o.map(y=>oM(y,!0)):(h=dW(c,o),b=dW(d(),[])),E0e({changedNodes:b,changedEdges:h,get:r,set:n})},unselectNodesAndEdges:({nodes:o,edges:s}={})=>{const{edges:c,getNodes:d}=r(),h=o||d(),b=s||c,y=h.map(k=>(k.selected=!1,oM(k.id,!1))),E=b.map(k=>oM(k.id,!1));E0e({changedNodes:y,changedEdges:E,get:r,set:n})},setMinZoom:o=>{const{d3Zoom:s,maxZoom:c}=r();s?.scaleExtent([o,c]),n({minZoom:o})},setMaxZoom:o=>{const{d3Zoom:s,minZoom:c}=r();s?.scaleExtent([c,o]),n({maxZoom:o})},setTranslateExtent:o=>{r().d3Zoom?.translateExtent(o),n({translateExtent:o})},resetSelectedElements:()=>{const{edges:o,getNodes:s}=r(),d=s().filter(b=>b.selected).map(b=>oM(b.id,!1)),h=o.filter(b=>b.selected).map(b=>oM(b.id,!1));E0e({changedNodes:d,changedEdges:h,get:r,set:n})},setNodeExtent:o=>{const{nodeInternals:s}=r();s.forEach(c=>{c.positionAbsolute=IZe(c.position,o)}),n({nodeExtent:o,nodeInternals:new Map(s)})},panBy:o=>{const{transform:s,width:c,height:d,d3Zoom:h,d3Selection:b,translateExtent:y}=r();if(!h||!b||!o.x&&!o.y)return!1;const E=mM.translate(s[0]+o.x,s[1]+o.y).scale(s[2]),k=[[0,0],[c,d]],x=h?.constrain()(E,k,y);return h.transform(b,x),s[0]!==x.x||s[1]!==x.y||s[2]!==x.k},cancelConnection:()=>n({connectionNodeId:qD.connectionNodeId,connectionHandleId:qD.connectionHandleId,connectionHandleType:qD.connectionHandleType,connectionStatus:qD.connectionStatus,connectionStartHandle:qD.connectionStartHandle,connectionEndHandle:qD.connectionEndHandle}),reset:()=>n({...qD})}),Object.is),BZe=({children:n})=>{const r=Q.useRef(null);return r.current||(r.current=i5r()),Un.createElement(S4r,{value:r.current},n)};BZe.displayName="ReactFlowProvider";const wcn=({children:n})=>Q.useContext(hEe)?Un.createElement(Un.Fragment,null,n):Un.createElement(BZe,null,n);wcn.displayName="ReactFlowWrapper";const o5r={input:Jsn,default:gWe,output:Qsn,group:jZe},a5r={default:_we,straight:PZe,step:RZe,smoothstep:pEe,simplebezier:NZe},s5r=[0,0],c5r=[15,15],u5r={x:0,y:0,zoom:1},l5r={width:"100%",height:"100%",overflow:"hidden",position:"relative",zIndex:0},ycn=Q.forwardRef(({nodes:n,edges:r,defaultNodes:o,defaultEdges:s,className:c,nodeTypes:d=o5r,edgeTypes:h=a5r,onNodeClick:b,onEdgeClick:y,onInit:E,onMove:k,onMoveStart:x,onMoveEnd:O,onConnect:P,onConnectStart:D,onConnectEnd:$,onClickConnectStart:U,onClickConnectEnd:B,onNodeMouseEnter:G,onNodeMouseMove:K,onNodeMouseLeave:X,onNodeContextMenu:ae,onNodeDoubleClick:oe,onNodeDragStart:he,onNodeDrag:ye,onNodeDragStop:we,onNodesDelete:ge,onEdgesDelete:Te,onSelectionChange:xe,onSelectionDragStart:Ue,onSelectionDrag:Z,onSelectionDragStop:Re,onSelectionContextMenu:_e,onSelectionStart:qe,onSelectionEnd:it,connectionMode:Xe=nU.Strict,connectionLineType:de=aM.Bezier,connectionLineStyle:We,connectionLineComponent:ht,connectionLineContainerStyle:Ie,deleteKeyCode:$t="Backspace",selectionKeyCode:An="Shift",selectionOnDrag:Yt=!1,selectionMode:Or=noe.Full,panActivationKeyCode:Nt="Space",multiSelectionKeyCode:on=Swe()?"Meta":"Control",zoomActivationKeyCode:br=Swe()?"Meta":"Control",snapToGrid:wi=!1,snapGrid:no=c5r,onlyRenderVisibleElements:Ri=!1,selectNodesOnDrag:ro=!0,nodesDraggable:Ko,nodesConnectable:Ca,nodesFocusable:qi,nodeOrigin:yi=s5r,edgesFocusable:lo,edgesUpdatable:Eo,elementsSelectable:Fc,defaultViewport:Yo=u5r,minZoom:zl=.5,maxZoom:ld=2,translateExtent:Yc=bWe,preventScrolling:Sn=!0,nodeExtent:Pn,defaultMarkerColor:Kr="#b1b1b7",zoomOnScroll:Ei=!0,zoomOnPinch:Sa=!0,panOnScroll:bs=!1,panOnScrollSpeed:Cc=.5,panOnScrollMode:Ac=MB.Free,zoomOnDoubleClick:Xc=!0,panOnDrag:rc=!0,onPaneClick:js,onPaneMouseEnter:Jc,onPaneMouseMove:Id,onPaneMouseLeave:nh,onPaneScroll:hu,onPaneContextMenu:Qp,children:Kb,onEdgeContextMenu:Km,onEdgeDoubleClick:S2,onEdgeMouseEnter:_2,onEdgeMouseMove:dv,onEdgeMouseLeave:K_,onEdgeUpdate:YI,onEdgeUpdateStart:n6,onEdgeUpdateEnd:e0,onReconnect:_l,onReconnectStart:rh,onReconnectEnd:t0,reconnectRadius:k2=10,edgeUpdaterRadius:r6=10,onNodesChange:i6,onEdgesChange:XI,noDragClassName:sy="nodrag",noWheelClassName:rC="nowheel",noPanClassName:fv="nopan",fitView:cy=!1,fitViewOptions:uy,connectOnClick:ly=!0,attributionPosition:Od,proOptions:JI,defaultEdgeOptions:ZI,elevateNodesOnSelect:hv=!0,elevateEdgesOnSelect:dy=!1,disableKeyboardA11y:T2=!1,autoPanOnConnect:QI=!0,autoPanOnNodeDrag:x2=!0,connectionRadius:o6=20,isValidConnection:fy,onError:Y_,style:Ym,id:C2,nodeDragThreshold:iC,...hy},oC)=>{const py=C2||"1";return Un.createElement("div",{...hy,style:{...Ym,...l5r},ref:oC,className:q_(["react-flow",c]),"data-testid":"rf__wrapper",id:C2},Un.createElement(wcn,null,Un.createElement(r5r,{onInit:E,onMove:k,onMoveStart:x,onMoveEnd:O,onNodeClick:b,onEdgeClick:y,onNodeMouseEnter:G,onNodeMouseMove:K,onNodeMouseLeave:X,onNodeContextMenu:ae,onNodeDoubleClick:oe,nodeTypes:d,edgeTypes:h,connectionLineType:de,connectionLineStyle:We,connectionLineComponent:ht,connectionLineContainerStyle:Ie,selectionKeyCode:An,selectionOnDrag:Yt,selectionMode:Or,deleteKeyCode:$t,multiSelectionKeyCode:on,panActivationKeyCode:Nt,zoomActivationKeyCode:br,onlyRenderVisibleElements:Ri,selectNodesOnDrag:ro,defaultViewport:Yo,translateExtent:Yc,minZoom:zl,maxZoom:ld,preventScrolling:Sn,zoomOnScroll:Ei,zoomOnPinch:Sa,zoomOnDoubleClick:Xc,panOnScroll:bs,panOnScrollSpeed:Cc,panOnScrollMode:Ac,panOnDrag:rc,onPaneClick:js,onPaneMouseEnter:Jc,onPaneMouseMove:Id,onPaneMouseLeave:nh,onPaneScroll:hu,onPaneContextMenu:Qp,onSelectionContextMenu:_e,onSelectionStart:qe,onSelectionEnd:it,onEdgeContextMenu:Km,onEdgeDoubleClick:S2,onEdgeMouseEnter:_2,onEdgeMouseMove:dv,onEdgeMouseLeave:K_,onReconnect:_l??YI,onReconnectStart:rh??n6,onReconnectEnd:t0??e0,reconnectRadius:k2??r6,defaultMarkerColor:Kr,noDragClassName:sy,noWheelClassName:rC,noPanClassName:fv,elevateEdgesOnSelect:dy,rfId:py,disableKeyboardA11y:T2,nodeOrigin:yi,nodeExtent:Pn}),Un.createElement(e6r,{nodes:n,edges:r,defaultNodes:o,defaultEdges:s,onConnect:P,onConnectStart:D,onConnectEnd:$,onClickConnectStart:U,onClickConnectEnd:B,nodesDraggable:Ko,nodesConnectable:Ca,nodesFocusable:qi,edgesFocusable:lo,edgesUpdatable:Eo,elementsSelectable:Fc,elevateNodesOnSelect:hv,minZoom:zl,maxZoom:ld,nodeExtent:Pn,onNodesChange:i6,onEdgesChange:XI,snapToGrid:wi,snapGrid:no,connectionMode:Xe,translateExtent:Yc,connectOnClick:ly,defaultEdgeOptions:ZI,fitView:cy,fitViewOptions:uy,onNodesDelete:ge,onEdgesDelete:Te,onNodeDragStart:he,onNodeDrag:ye,onNodeDragStop:we,onSelectionDrag:Z,onSelectionDragStart:Ue,onSelectionDragStop:Re,noPanClassName:fv,nodeOrigin:yi,rfId:py,autoPanOnConnect:QI,autoPanOnNodeDrag:x2,onError:Y_,connectionRadius:o6,isValidConnection:fy,nodeDragThreshold:iC}),Un.createElement(Z4r,{onSelectionChange:xe}),Kb,Un.createElement(T4r,{proOptions:JI,position:Od}),Un.createElement(o6r,{rfId:py,disableKeyboardA11y:T2})))});ycn.displayName="ReactFlow";function Ecn(n){return r=>{const[o,s]=Q.useState(r),c=Q.useCallback(d=>s(h=>n(d,h)),[]);return[o,s,c]}}const d5r=Ecn(scn),f5r=Ecn(w6r);var Qf=(n=>(n.Loading="loading",n.Hidden="hidden",n.Step1="step1",n.Step2="step2",n.Step3="step3",n.Finished="finished",n))(Qf||{});const UZe=180,zZe=50,Scn=3.25,h5r=448,p5r=360,g5r=620,_Ft="action_listen",m5r="rasa_session_history",_cn="do",kcn="project",kFt="download",TFt="feedback",b5r="/tya-bot.png",v5r="/tya-user.png",w5r="OnboardingState",y5r="https://forum.rasa.com/",xFt=3,CFt=10*1e3,E5r=1e3*60*5,K$e=`👋 Welcome to Hello Rasa!

I’m your **Copilot** — here to help you explore Rasa and start customizing your agent.
You can ask me how your agent works, how to add new skills, or how to connect integrations.`;FB.Finance+"",`${K$e}`,FB.Telco+"",`${K$e}`,FB.Basic+"",`${K$e}`;FB.Finance,FB.Telco,FB.Basic;const _0e={positive:{title:"👏 Thanks for sharing your feedback",description:"Your feedback helps us make Rasa Copilot better!"},negative:{title:"💡 Thanks for letting us know.",description:"Your feedback helps us make Rasa Copilot better!"}},S5r={[Qf.Step1]:{title:"💬 Try your agent",description:"Start by sending a message to your new agent. This is the live conversation preview of your project, so you can try it out and test your changes.",action:"Next",counter:"1/2"},[Qf.Step2]:{title:"🤔 Have a question?",description:"Copilot is here to help. Ask it for help with building new skills or answer your Rasa questions.",action:"Got it",counter:"2/2"},[Qf.Step3]:{title:"🔍 Debug your agent",description:"Toggle inspector mode to see how your agent makes decisions and remembers what the user has said. Think of it like dev tools for Rasa-made to help you debug and improve conversations.",action:"Finish"}},Tcn={fontFamily:"IBM Plex Mono",minimap:{enabled:!1},scrollbar:{vertical:"hidden",horizontal:"auto"}},_5r=new jIr({defaultLayoutOptions:{"elk.algorithm":"org.eclipse.elk.layered","elk.direction":"DOWN","elk.layered.spacing.edgeNodeBetweenLayers":"50","org.eclipse.elk.layered.nodePlacement.bk.fixedAlignment":"BALANCED","org.eclipse.elk.alignment":"TOP","spacing.nodeNodeBetweenLayers":"40","org.eclipse.elk.spacing.nodeNode":"50","org.eclipse.elk.layered.considerModelOrder.strategy":"PREFER_EDGES"}}),xcn=n=>{const{data:r,id:o,type:s,xPos:c,yPos:d,selected:h}=n;return{data:r,id:o,type:s,position:{x:c,y:d},selected:h}},HZe=n=>n.label;function k5r(n){return{x:Math.floor((n.x||1)-(n.width||1)/2),y:Math.floor((n.y||1)-(n.height||1)/2)}}async function T5r(n,r){const o=n.map(k=>({id:k.id,width:UZe,height:zZe})),s=n.reduce((k,x)=>Object.assign(k,{[x.id]:x}),{}),h=r.filter(k=>k.type===AW.Auto).sort((k,x)=>{const O=s[k.targetId],P=s[x.targetId];return O.type===Pa.Condition&&O.subType===gM.ELSE?1:P.type===Pa.Condition&&P.subType===gM.ELSE?-1:O.type===Pa.Condition&&P.type===Pa.Condition?O.order-P.order:1}).map(k=>({id:`${k.targetId}-${k.sourceId}`,targets:[k.targetId],sources:[k.sourceId]})),b={id:"root",layoutOptions:{"elk.algorithm":"org.eclipse.elk.layered","elk.direction":"DOWN"},children:o,edges:h},y=await _5r.layout(b);if(!y?.children)throw new Error("Invalid graph");const E=Object.fromEntries(y.children.map(k=>[k.id,k]));return n.map(k=>{const x=E[k.id],{x:O,y:P}=k5r(x);return{...k,metadata:{x:O,y:P}}})}const x5r=n=>/^\b((mailto|tel|sms):|[a-z]+:\/\/)/i.test(n),Ccn=n=>{const r=n?.split(".").pop();return r?nIr.languages(r)?.[0]:"plain text"},Acn=async n=>V_(`${onn.PROJECT_SERVICE_URL}${ann.GET_PROJECT(n)}`,{method:"GET"},void 0,rlr),gEe=xs({status:Tr(),copilot_welcome_job_id:Tr().optional(),copilot_training_success_job_id:Tr().optional(),copilot_template_prompt_job_id:Tr().optional(),copilot_rollback_message_job_id:Tr().optional(),copilot_revert_success_job_id:Tr().optional(),message:Tr().optional()}),mEe=xs({status:Tr(),message:Tr(),copilot_job_id:Tr().optional()}),C5r=()=>({loading:!1,loadingProject:!1,isRestoring:!1,inactive:!1,error:null,projectId:null,projectUrl:null,jobId:null,jobStatus:null,templateName:null,assistantName:null,availableUntil:null,showInfrastructureNotice:!1,restoreJobId:null}),da=new IM(C5r());var a2=(n=>(n.Received="received",n.Generating="generating",n.GenerationSuccess="generation_success",n.Training="training",n.TrainSuccess="train_success",n.InitialTrainError="initial_train_error",n.Heartbeat="heartbeat",n.Done="done",n))(a2||{});const Icn=n=>{da.setState(r=>({...r,projectUrl:n,projectId:"standalone"}))},GZe=async n=>{try{const r=await Shr({projectUrl:n});r.assistant_id&&da.setState(o=>({...o,assistantName:r.assistant_id}))}catch(r){Va(r,{tags:{component:"project-actions",action:"updateAssistantNameFromProject"},extra:{projectUrl:n}})}},A5r=(n,r)=>{const o=`${onn.PROJECT_SERVICE_URL}${ann.JOB_EVENTS(n)}`,s=new EventSource(o);s.onmessage=c=>{try{const d=JSON.parse(c.data),h=CI(d,gEe);h.status==="progress"&&h.message==="Restore complete"&&(s.close(),da.setState(b=>({...b,isRestoring:!1,restoreJobId:null,loading:!0,error:null})),setTimeout(()=>{N5r(r,3)},1e3))}catch(d){console.error("Error parsing restore progress event:",d),da.setState(h=>({...h,isRestoring:!1,restoreJobId:null,error:`Error parsing restore progress event: ${d instanceof Error?d.message:String(d)}`,loading:!1})),s.close()}},s.addEventListener("error",c=>{try{const d=JSON.parse(c.data),h=CI(d,mEe),b=`Restore failed: ${h.message}`;Va(new Error(b),{tags:{component:"project-actions",action:"createRestoreEventSource"},extra:{jobId:n,projectId:r,status:h.status,message:h.message}}),da.setState(y=>({...y,isRestoring:!1,restoreJobId:null,error:b,loading:!1})),s.close()}catch(d){da.setState(h=>({...h,isRestoring:!1,restoreJobId:null,error:`Error parsing restore error event: ${d instanceof Error?d.message:String(d)}`,loading:!1})),s.close()}})},Ocn=async n=>{da.setState(r=>({...r,loading:!0,error:null}));try{const r=await Acn(n);if(r.status==="restoring"&&r.job_id){da.setState(o=>({...o,isRestoring:!0,restoreJobId:r.job_id||null,projectId:n,loading:!1})),A5r(r.job_id,n);return}da.setState(o=>({...o,projectId:r.project_id,projectUrl:r.url,availableUntil:r.available_until,loading:!1})),await GZe(r.url)}catch(r){console.error("loadProject:",r),da.setState(o=>({...o,error:r instanceof Error?r.message:String(r),loading:!1}))}},I5r=async n=>{da.setState(r=>({...r,projectId:n.project_id,projectUrl:n.url,availableUntil:n.available_until,loading:!1,error:null})),await GZe(n.url)},O5r=(n,r,o,s)=>{Va(new Error("Project load retry failed"),{tags:{component:"project-actions",action:"loadProjectWithRetry"},extra:{projectId:n,attempt:r.toString(),maxRetries:o.toString(),error:s instanceof Error?s.message:String(s)}}),da.setState(c=>({...c,error:"Failed to load project after restore. Please refresh the page.",loading:!1}))},N5r=async(n,r)=>{for(let o=1;o<=r;o++)try{const s=await Acn(n),c=s.status==="restoring",d=o>=r;if(c&&!d){await new Promise(h=>setTimeout(h,2e3));continue}if(c&&d)throw new Error("Project restore may have failed - still getting restore response after retries");await I5r(s);return}catch(s){console.warn(`Attempt ${o} failed:`,s),o<r?await new Promise(c=>setTimeout(c,2e3)):O5r(n,o,r,s)}},R5r=()=>da.state.projectId,Ncn=()=>da.state.projectUrl,AFt=n=>{da.setState(r=>({...r,jobStatus:n}))},P5r=(n,r)=>{da.setState(o=>({...o,jobId:n,jobStatus:r}))},Bre=n=>{da.setState(r=>({...r,inactive:n}))},IFt="{{SEGMENT_WRITE_KEY}}",L5r=n=>!(n.startsWith("{{")&&n.endsWith("}}")),DI=L5r(IFt)?Tfr.load({writeKey:IFt}):null,e6=async()=>{if(!DI)return`anonymous_${n2()}`;try{return(await DI.user()).anonymousId()}catch(n){return Va(n,{tags:{component:"analytics",action:"getSegmentUserId"}}),`anonymous_${n2()}`}},z_=async(n,r={})=>{try{if(!DI)return;const o=await e6(),s=R5r();await DI.track(n,{...r,projectId:s,userId:o})}catch(o){Va(o,{tags:{component:"analytics",action:"track"}})}},OFt=async(n,r)=>{try{if(!DI)return;await DI.identify(n,{...r})}catch(o){Va(o,{tags:{component:"analytics",action:"identify"}})}},Rcn=async(n,r={})=>{try{if(!DI)return;const o=await e6();await DI.page(n,{...r,userId:o})}catch(o){Va(o,{tags:{component:"analytics",action:"page"}})}},Pcn=async n=>{if(!uPr()||!n?.sub)return await OFt();await OFt(n.sub,{email:n.email,name:n.name})},D5r=async()=>{if(DI)try{await DI.reset()}catch(n){Va(n,{tags:{component:"analytics",action:"deidentify"}})}},H_={copilotFeedback:"Copilot Feedback",userLoggedIn:"User Logged In",projectCreated:"Project Created",projectEdited:"Project Edited",projectDownloaded:"Project Downloaded",connectClicked:"Connect Clicked",userMessageSent:"User Message Sent",generalFeedback:"General Feedback",visitForumClicked:"Visit Forum Clicked",shareCopyLinkClicked:"Share Copy Link Clicked",fileLinkClicked:"File Link Clicked"},M5r={utm_source:"hello_rasa",utm_medium:"connect_popup",utm_campaign:"community_growth",utm_content:"join_community"};async function Lcn(n,r={},o){let s;try{const c=await e6();s=await fetch(n,{...r,headers:{"Content-Type":"application/json","X-User-Id":String(c),...r.headers||{}},body:o?JSON.stringify(o):void 0})}catch(c){let d="Network error: Unable to reach the server.";throw c instanceof Error&&c.message&&(d+=`
Details: ${c.message}`),new Error(d)}if(!s.ok){const c=await s.json().catch(()=>({}));let d="An unknown API error occurred.";throw c&&typeof c=="object"&&"error"in c&&typeof c.error=="string"&&(d=c.error||d),new Error(d)}return s}async function j5r(n,r={},o){return await(await Lcn(n,r,o)).blob()}async function V_(n,r={},o,s){const d=await(await Lcn(n,r,o)).json();return s?CI(d,s):d}function CI(n,r){const o=r.safeParse(n);if(!o.success)throw new Error(`Response data validation failed: ${JSON.stringify(o.error.issues)}`);return o.data}function Dcn(n,r,o=!1){let s=null,c=null;return function(...h){c=h;const b=()=>{s=null,!o&&c&&n(...c)},y=o&&!s;s&&clearTimeout(s),s=setTimeout(b,r),y&&n(...h)}}const $5r=n=>{il.setState(r=>({...r,sessionId:n}))},kwe=n=>{il.setState(r=>{const o=[...r.messages,n];return{...r,messages:o}})},F5r=n=>{il.setState(r=>{const o=[...r.messages];return o.length===0?r:(o[o.length-1]=n,{...r,messages:o})})},Twe=n=>{il.setState(r=>({...r,isLoading:n}))},xwe=n=>{il.setState(r=>({...r,streamingMessage:n}))},vWe=n=>{il.setState(r=>({...r,shouldScrollOnStreamStart:n}))},Cwe=(n,r)=>{il.setState(o=>{const s=[...o.streamingMessage?.content||[]],c=s.findIndex(d=>d.type==="text");return c===-1?s.push({type:"text",text:n}):s[c]={type:"text",text:n},{...o,streamingMessage:o.streamingMessage?{...o.streamingMessage,response_category:r,content:s}:null}})},B5r=n=>{il.setState(r=>r.attachments[0]?.id===n.id?r:{...r,attachments:[n]})},U5r=n=>{il.setState(r=>({...r,attachments:r.attachments.filter(o=>o.id!==n)}))},z5r=()=>{il.setState(n=>({...n,attachments:[]}))},H5r=(n,r)=>{const o=n.trim();if(!o)throw new Error("Message cannot be empty");const s=Ncn();if(!s)throw new Error("Project URL is not set");if(!r)throw new Error("Session ID is not set");return{trimmedMessage:o,projectUrl:s}},G5r=(n,r,o)=>{const s=[{type:"text",text:n}];if(r&&r.length>0)for(const c of r)c.metadata.rawEvent&&s.push({type:"event",...c.metadata.rawEvent});return{role:o,content:s,timestamp:new Date}},Mcn=()=>({role:kI.Copilot,content:[{type:"text",text:""}],timestamp:new Date}),q5r=n=>{const r=il.state.streamingMessage;if(!r)return;const s=[...r.content].filter(d=>d.type!=="references"),c={type:"references",references:n};s.push(c),il.setState(d=>({...d,streamingMessage:{...d.streamingMessage,content:s}}))},V5r=n=>{if(n.length>1)Va(new Error("Received more than one training error log"),{tags:{component:"CopilotActions",action:"handleTrainingErrorLog"},extra:{logs:JSON.stringify(n)}});else if(n.length===0){Va(new Error("Received none training error log"),{tags:{component:"CopilotActions",action:"handleTrainingErrorLog"}});return}const r=il.state.streamingMessage;if(!r)return;const o=[...r.content];o.push({type:"logs",logs:[{type:han.Log,content:n[0].content,context:n[0].context}]}),il.setState(s=>({...s,streamingMessage:{...s.streamingMessage,content:o}}))},W5r=(n,r)=>{const o=il.state.streamingMessage;if(!o)return;const s=[...o.content];s.push({type:"commit",commit:n}),il.setState(c=>({...c,streamingMessage:{...c.streamingMessage,content:s}})),r?.(n)},K5r=n=>{const r=il.state.streamingMessage;if(!r)return;const o=[...r.content],s=o.findIndex(d=>d.type==="plan"),c={type:"plan",tasks:n};s===-1?o.unshift(c):o[s]=c,il.setState(d=>({...d,streamingMessage:{...d.streamingMessage,content:o}}))},Y5r=({content:n,completeness:r,response_category:o,accumulatedText:s,isFirstChunk:c,setLoading:d,updateStreamingMessage:h})=>{let b=s;return r===Yie.Token?b+=n:r===Yie.Complete&&(b=n),c&&d(!1),h(b,o),{newText:b,newIsFirstChunk:!1,isPolicyViolation:o===$b.GuardrailsPolicyViolation}},jcn=(n,r,o,s,c,d)=>{switch(n.response_category){case $b.Copilot:case $b.CopilotTextContentPartStart:case $b.CopilotTextContentPartDelta:case $b.CopilotTextContentPartEnd:case $b.Commit:case $b.GuardrailsPolicyViolation:case $b.GuardrailsBlocked:case $b.OutOfScopeDetection:case $b.ErrorFallback:case $b.RoleplayDetection:case $b.KnowledgeBaseAccessRequested:{if(n.commit&&W5r(n.commit,d),n.content){const h=Y5r({content:n.content,completeness:n.completeness,response_category:n.response_category,accumulatedText:r,isFirstChunk:o,setLoading:s,updateStreamingMessage:c});return{newText:h.newText,newIsFirstChunk:h.newIsFirstChunk,isPolicyViolation:h.isPolicyViolation}}return{newText:r,newIsFirstChunk:o}}case $b.Reference:return n.references&&q5r(n.references),{newText:r,newIsFirstChunk:o};case $b.TrainingErrorLog:return n.logs&&V5r(n.logs),{newText:r,newIsFirstChunk:o};case $b.TaskPlanning:return n.tasks&&K5r(n.tasks),{newText:r,newIsFirstChunk:o};default:return{newText:r,newIsFirstChunk:o}}},X5r=async(n,r)=>{const o=n.getReader(),s=new TextDecoder;let c="",d=!0,h="",b=!1;const y=Dcn(Cwe,16);try{for(;;){const{done:E,value:k}=await o.read();if(E)break;const x=s.decode(k,{stream:!0});h+=x;const{events:O,remainingBuffer:P}=NIr(h);h=P;for(const D of O){const $=D.completeness===Yie.Complete?Cwe:y,U=jcn(D,c,d,Twe,$,r);c=U.newText,d=U.newIsFirstChunk,b=U.isPolicyViolation||b}}}finally{o.releaseLock()}return{isPolicyViolation:b}},J5r=n=>{const r={role:kI.Copilot,content:[{type:"text",text:"Sorry, I encountered an error. Please try again."}],timestamp:new Date};Va(n,{tags:{component:"CopilotActions",action:"sendMessage"}}),kwe(r)},NFt=async({message:n,role:r=kI.User,additionalResponseContent:o,token:s,onCommit:c})=>{const{attachments:d,sessionId:h}=il.state;try{const{trimmedMessage:b,projectUrl:y}=H5r(n,h),E=G5r(b,d,r);kwe(E),Twe(!0),z5r();const k={...E,timestamp:LIr(E.timestamp)},x=await hhr({projectUrl:y,requestBody:{session_id:h,message:k},token:s}),O=Mcn();xwe(O);const{isPolicyViolation:P}=await X5r(x,c);let D=il.state.streamingMessage;P&&F5r({...E,response_category:$b.GuardrailsPolicyViolation}),D&&o&&(D={...D,content:[...D.content,o]}),D&&(kwe(D),xwe(null))}catch(b){J5r(b)}finally{Twe(!1)}},wWe=({jobId:n,projectUrl:r,contentStatusKey:o,onSuccess:s,onFirstEvent:c,onError:d,onCommit:h})=>{const b=new EventSource(`${r}${Vb.JOB_EVENTS}/${n}`);let y="",E=!0,k=!1,x=!1;const O=Dcn(Cwe,16);b.addEventListener("progress",P=>{try{const D=JSON.parse(P.data),$=CI(D,gEe);if($.status===o){const U=CI(D,man);if(!x&&U.response_category){c?.();const K=Mcn();xwe(K),x=!0}const B=U.completeness===Yie.Complete?Cwe:O,G=jcn(U,y,E,Twe,B,h);y=G.newText,E=G.newIsFirstChunk,k=G.isPolicyViolation||k}else if($.status==="done"){b.close();let U=Z5r();U&&(U={...U,content:[...U.content]},kwe(U),xwe(null),s?.())}}catch(D){Va(D,{tags:{component:"copilot",action:"createEventSourceForStandaloneCopilotMessage"}}),d?.(`Error parsing copilot event data: ${D instanceof Error?D.message:String(D)}`),b.close()}}),b.addEventListener("error",P=>{try{const D=JSON.parse(P.data),$=CI(D,mEe),U=`Copilot job failed with status: ${$.status}: ${$.message}`;Va(new Error(U),{tags:{component:"copilot",action:"createEventSourceForStandaloneCopilotMessage"}}),d?.(U),b.close()}catch(D){Va(D,{tags:{component:"copilot",action:"createEventSourceForStandaloneCopilotMessage"}}),d?.(`Error parsing copilot event data: ${D instanceof Error?D.message:String(D)}`),b.close()}})},Z5r=()=>il.state.streamingMessage,Q5r=()=>({files:null,getFilesLoading:!1,postFilesLoading:!1,trainingLoading:!1,error:null,filesInitialized:!1,editorStatus:null}),gc=new IM(Q5r()),ePr=n=>{gc.setState(r=>({...r,files:n}))},tPr=n=>{gc.setState(r=>({...r,files:n,filesInitialized:!0}))},$cn=n=>{gc.setState(r=>({...r,filesInitialized:n}))},hie=n=>{gc.setState(r=>({...r,error:n}))},PK=n=>n.getFilesLoading||n.postFilesLoading||n.trainingLoading,nPr=n=>{gc.setState(r=>({...r,getFilesLoading:n}))},Y$e=n=>{gc.setState(r=>({...r,postFilesLoading:n}))},rPr=(n,r)=>{gc.setState(o=>{if(!o.files)return o;const s=o.files.findIndex(h=>h.path===n.path);if(s===-1)return o;const c={...o.files[s],path:r,edited:!0},d=[...o.files];return d[s]=c,{...o,files:d}})},iPr=(n,r)=>{gc.setState(o=>{if(!o.files)return o;const s=o.files.map(c=>c.path.startsWith(n)?{...c,path:c.path.replace(n,r),edited:!0}:c);return{...o,files:s}})},Awe=n=>(gc.setState(r=>{if(!r.files)return r;const o=r.files.findIndex(c=>c.path===n.path);if(o===-1)return{...r,files:[...r.files,n]};const s=[...r.files];return s[o]=n,{...r,files:s}}),n),oPr=n=>{const r=Ncn();if(!r)throw new Error("No project URL available");const o=new EventSource(`${r}${Vb.JOB_EVENTS}/${n}`);gc.setState(s=>({...s,trainingLoading:!0})),o.addEventListener("error",s=>{try{const c=JSON.parse(s.data),d=CI(c,mEe),h=`Job failed with status: ${d.status}: ${d.message}`;AFt(d.status),gc.setState(b=>({...b,error:h,trainingLoading:!1,editorStatus:d.status})),gc.setState(b=>({...b,files:b.files?b.files.map(y=>y.edited||y.hasErrors?{...y,edited:!1,hasErrors:!0}:{...y,edited:!1,hasErrors:!1}):null})),z_(H_.projectEdited,{status:"error"}),o.close(),d.copilot_job_id&&wWe({jobId:d.copilot_job_id,projectUrl:r,contentStatusKey:"copilot_analyzing",onFirstEvent:()=>{gc.setState(b=>({...b,loading:!1,editorStatus:"done"}))},onError:b=>{gc.setState(y=>({...y,error:b,loading:!1,editorStatus:"error"}))}})}catch(c){Va(c,{tags:{component:"editor store",action:"createTrainingEventSource"}}),gc.setState(d=>({...d,error:`Error parsing error event data: ${c instanceof Error?c.message:String(c)}`,trainingLoading:!1,editorStatus:null})),o.close()}}),o.addEventListener("progress",s=>{try{const c=JSON.parse(s.data),d=CI(c,gEe);AFt(d.status),Bcn(d.status),d.status==="done"&&(o.close(),gc.setState(h=>({...h,files:h.files?h.files.map(b=>({...b,edited:!1,hasErrors:!1})):null,trainingLoading:!1})),z_(H_.projectEdited,{status:"success"}),d.copilot_training_success_job_id&&wWe({jobId:d.copilot_training_success_job_id,projectUrl:r,contentStatusKey:"train_success_message",onFirstEvent:()=>{gc.setState(h=>({...h,loading:!1,editorStatus:"done"}))},onError:h=>{gc.setState(b=>({...b,error:h,loading:!1,editorStatus:"error"}))}}))}catch(c){gc.setState(d=>({...d,error:`Error parsing progress event data: ${c instanceof Error?c.message:String(c)}`,trainingLoading:!1})),o.close()}})},aPr=n=>{gc.setState(r=>{if(!r.files)return r;const o=r.files.findIndex(d=>d.path===n);if(o===-1)return r;const s={...r.files[o],edited:!0,deleted:!0},c=[...r.files];return c[o]=s,{...r,files:c}})},sPr=(n,r=!1)=>{gc.setState(o=>{if(!o.files)return o;const s=[];for(const[d,h]of o.files.entries())h.path.startsWith(`${n}/`)&&s.push(d),h.path===n&&s.push(d);if(s.length===0)return o;if(r){const d=o.files.filter((h,b)=>!s.includes(b));return{...o,files:d}}const c=[...o.files];for(const d of s)c[d]={...c[d],edited:!0,deleted:!0};return{...o,files:c}})},Fcn=(n,r)=>{const o=gc.state;return o.files?r?o.files.some(s=>s.path.startsWith(n+"/")&&!s.deleted):o.files.some(s=>s.path===n&&!s.deleted):!1},Bcn=n=>{gc.setState(r=>({...r,editorStatus:n}))},cPr=()=>({necessary:!0,marketing:!1,analytics:!1}),Ucn=new IM(cPr()),RFt=n=>{Ucn.setState(r=>({...r,...n}))},uPr=()=>Ucn.state.analytics,lPr=()=>({isEnabled:!0}),zcn=new IM(lPr()),PFt=n=>{zcn.setState(r=>({...r,isEnabled:n}))},dPr=()=>({loading:!1,error:null,jobId:null,jobStatus:null}),bEe=new IM(dPr()),fPr=({jobId:n,projectUrl:r,contentStatusKey:o="rollback_message",onSuccess:s,onFirstEvent:c,onError:d=h=>k0e(h)})=>{const h=new EventSource(`${r}${Vb.JOB_EVENTS}/${n}`);h.addEventListener("error",b=>{try{const y=JSON.parse(b.data),E=CI(y,mEe),k=`Commit Job failed with status: ${E.status}: ${E.message}`;k0e(k),d?.(k),h.close()}catch(y){k0e(`Error parsing error event data: ${y instanceof Error?y.message:String(y)}`),d?.(`Error parsing commit error event data: ${y instanceof Error?y.message:String(y)}`),h.close()}}),h.addEventListener("progress",b=>{try{const y=JSON.parse(b.data),E=CI(y,gEe);bEe.setState(k=>({...k,jobStatus:E.status})),E.status==="done"&&(h.close(),E.copilot_rollback_message_job_id?wWe({jobId:E.copilot_rollback_message_job_id,projectUrl:r,contentStatusKey:o,onFirstEvent:()=>{c?.()},onSuccess:()=>{s?.()},onError:k=>{d(k)}}):(d?.("No copilot job ID found"),h.close(),Va("No copilot job ID found",{tags:{component:"store-commit-actions",action:"createCommitEventSource"}})))}catch(y){k0e(`Error parsing progress event data: ${y instanceof Error?y.message:String(y)}`),d?.(`Error parsing commit event data: ${y instanceof Error?y.message:String(y)}`),h.close()}})},k0e=n=>{Va(n,{tags:{component:"store-commit-actions",action:"setError"}}),bEe.setState(r=>({...r,jobStatus:null,jobId:null,error:n,loading:!1}))},T0e=n=>{bEe.setState(r=>({...r,loading:n}))},Hcn=Q.createContext(void 0),hPr=()=>!0,pPr=({children:n})=>{const[r]=Dye(),o=Q.useMemo(()=>hPr()?{mode:IW.StudioEmbed,isStudioEmbed:!0,isStandalone:!1,isCloud:!1,showCopilot:!0,showRasaLogo:!1,showProjectActions:!1}:{mode:IW.Standalone,isStudioEmbed:!1,isStandalone:!0,isCloud:!1,showCopilot:!0,showRasaLogo:!0,showProjectActions:!1},[r]);return Q.useEffect(()=>{(o.isStandalone||o.isStudioEmbed)&&Icn(window.location.origin)},[o.isStandalone,o.isStudioEmbed]),H.jsx(Hcn.Provider,{value:o,children:n})};function gPr(n,r,o={}){if(typeof document>"u")return;let s=`${encodeURIComponent(n)}=${encodeURIComponent(r)}`;if(o.days!==void 0){const c=new Date(Date.now()+o.days*864e5).toUTCString();s+=`; expires=${c}`}s+=`; path=${o.path??"/"}`,o.domain&&(s+=`; domain=${o.domain}`),o.secure&&(s+="; Secure"),o.sameSite&&(s+=`; SameSite=${o.sameSite}`),document.cookie=s}function Gcn(n){if(typeof document>"u"||!document.cookie)return;const r=n.replaceAll(/[.*+?^${}()|[\]\\]/g,"\\$&"),s=new RegExp(`(?:^|; )${r}=([^;]*)`).exec(document.cookie);return s?decodeURIComponent(s[1]):void 0}const mPr="hubspotutk",LFt="rasa_hubspot_signup_set",bPr="https://api.hsforms.com/submissions/v3/integration/submit",vPr="{{HUBSPOT_PORTAL_ID}}",wPr="{{HUBSPOT_FORM_ID}}",yPr=EPr("{{HUBSPOT_ENABLED}}");function EPr(n){let r=!1;try{const o=JSON.parse(n);if(typeof o=="boolean")r=o;else throw new TypeError("VITE_HUBSPOT_ENABLED is not a boolean, defaulting to false")}catch(o){Va(o,{tags:{component:"Hubspot",action:"parse VITE_HUBSPOT_ENABLED"}}),r=!1}return r}async function SPr({email:n,pageUri:r=globalThis.window===void 0?"":globalThis.location.href,pageName:o=globalThis.document===void 0?"":globalThis.document.title,portalId:s=vPr,formId:c=wPr}){if(!yPr||!s||!c)throw new Error("HubSpot is not configured or enabled");const d=Gcn(mPr),h={fields:[{name:"email",value:n}],context:{hutk:d,pageUri:r,pageName:o}},b=`${bPr}/${s}/${c}`;await _Pr({url:b,requestBody:h})}async function _Pr({url:n,requestBody:r}){if(Gcn(LFt))return;const o=await fetch(n,{method:"POST",keepalive:!0,headers:{"Content-Type":"application/json"},body:JSON.stringify(r)});if(o.ok)gPr(LFt,"true");else{const s=await o.text();let c;try{c=JSON.parse(s)}catch{console.error("JSON parse error",s)}Va(new Error("HubSpot submission failed"),{extra:{...c},tags:{component:"Hubspot",action:"submitHubSpotSignup"}})}}var kn=typeof window<"u"?window:void 0,Jw=typeof globalThis<"u"?globalThis:kn,qcn=Array.prototype,DFt=qcn.forEach,MFt=qcn.indexOf,c2=Jw?.navigator,Di=Jw?.document,Kw=Jw?.location,yWe=Jw?.fetch,EWe=Jw!=null&&Jw.XMLHttpRequest&&"withCredentials"in new Jw.XMLHttpRequest?Jw.XMLHttpRequest:void 0,jFt=Jw?.AbortController,Gw=c2?.userAgent,La=kn??{},I4={DEBUG:!1,LIB_VERSION:"1.257.2"},X$e="$copy_autocapture",kPr=["$snapshot","$pageview","$pageleave","$set","survey dismissed","survey sent","survey shown","$identify","$groupidentify","$create_alias","$$client_ingestion_warning","$web_experiment_applied","$feature_enrollment_update","$feature_flag_called"],N4=function(n){return n.GZipJS="gzip-js",n.Base64="base64",n}({}),TPr=["fatal","error","warning","log","info","debug"];function pc(n,r){return n.indexOf(r)!==-1}var vEe=function(n){return n.trim()},SWe=function(n){return n.replace(/^\$/,"")},xPr=Array.isArray,Vcn=Object.prototype,Wcn=Vcn.hasOwnProperty,wEe=Vcn.toString,kf=xPr||function(n){return wEe.call(n)==="[object Array]"},j_=n=>typeof n=="function",Ah=n=>n===Object(n)&&!kf(n),fW=n=>{if(Ah(n)){for(var r in n)if(Wcn.call(n,r))return!1;return!0}return!1},mi=n=>n===void 0,Ih=n=>wEe.call(n)=="[object String]",_We=n=>Ih(n)&&n.trim().length===0,HM=n=>n===null,Gu=n=>mi(n)||HM(n),G1=n=>wEe.call(n)=="[object Number]",wI=n=>wEe.call(n)==="[object Boolean]",CPr=n=>n instanceof FormData,APr=n=>pc(kPr,n),Kcn=n=>{var r={t:function(o){if(kn&&(I4.DEBUG||La.POSTHOG_DEBUG)&&!mi(kn.console)&&kn.console){for(var s=("__rrweb_original__"in kn.console[o])?kn.console[o].__rrweb_original__:kn.console[o],c=arguments.length,d=new Array(c>1?c-1:0),h=1;h<c;h++)d[h-1]=arguments[h];s(n,...d)}},info:function(){for(var o=arguments.length,s=new Array(o),c=0;c<o;c++)s[c]=arguments[c];r.t("log",...s)},warn:function(){for(var o=arguments.length,s=new Array(o),c=0;c<o;c++)s[c]=arguments[c];r.t("warn",...s)},error:function(){for(var o=arguments.length,s=new Array(o),c=0;c<o;c++)s[c]=arguments[c];r.t("error",...s)},critical:function(){for(var o=arguments.length,s=new Array(o),c=0;c<o;c++)s[c]=arguments[c];console.error(n,...s)},uninitializedWarning:o=>{r.error("You must initialize PostHog before calling "+o)},createLogger:o=>Kcn(n+" "+o)};return r},zi=Kcn("[PostHog.js]"),Wm=zi.createLogger,IPr=Wm("[ExternalScriptsLoader]"),$Ft=(n,r,o)=>{if(n.config.disable_external_dependency_loading)return IPr.warn(r+" was requested but loading of external scripts is disabled."),o("Loading of external scripts is disabled");var s=Di?.querySelectorAll("script");if(s){for(var c=0;c<s.length;c++)if(s[c].src===r)return o()}var d=()=>{if(!Di)return o("document not found");var h=Di.createElement("script");if(h.type="text/javascript",h.crossOrigin="anonymous",h.src=r,h.onload=E=>o(void 0,E),h.onerror=E=>o(E),n.config.prepare_external_dependency_script&&(h=n.config.prepare_external_dependency_script(h)),!h)return o("prepare_external_dependency_script returned null");var b,y=Di.querySelectorAll("body > script");y.length>0?(b=y[0].parentNode)==null||b.insertBefore(h,y[0]):Di.body.appendChild(h)};Di!=null&&Di.body?d():Di?.addEventListener("DOMContentLoaded",d)};function Ni(){return Ni=Object.assign?Object.assign.bind():function(n){for(var r=1;r<arguments.length;r++){var o=arguments[r];for(var s in o)({}).hasOwnProperty.call(o,s)&&(n[s]=o[s])}return n},Ni.apply(null,arguments)}function Ycn(n,r){if(n==null)return{};var o={};for(var s in n)if({}.hasOwnProperty.call(n,s)){if(r.indexOf(s)!==-1)continue;o[s]=n[s]}return o}La.__PosthogExtensions__=La.__PosthogExtensions__||{},La.__PosthogExtensions__.loadExternalDependency=(n,r,o)=>{var s="/static/"+r+".js?v="+n.version;if(r==="remote-config"&&(s="/array/"+n.config.token+"/config.js"),r==="toolbar"){var c=3e5;s=s+"&t="+Math.floor(Date.now()/c)*c}var d=n.requestRouter.endpointFor("assets",s);$Ft(n,d,o)},La.__PosthogExtensions__.loadSiteApp=(n,r,o)=>{var s=n.requestRouter.endpointFor("api",r);$Ft(n,s,o)};var Iwe={};function vM(n,r,o){if(kf(n)){if(DFt&&n.forEach===DFt)n.forEach(r,o);else if("length"in n&&n.length===+n.length){for(var s=0,c=n.length;s<c;s++)if(s in n&&r.call(o,n[s],s)===Iwe)return}}}function Bl(n,r,o){if(!Gu(n)){if(kf(n))return vM(n,r,o);if(CPr(n)){for(var s of n.entries())if(r.call(o,s[1],s[0])===Iwe)return}else for(var c in n)if(Wcn.call(n,c)&&r.call(o,n[c],c)===Iwe)return}}var pp=function(n){for(var r=arguments.length,o=new Array(r>1?r-1:0),s=1;s<r;s++)o[s-1]=arguments[s];return vM(o,function(c){for(var d in c)c[d]!==void 0&&(n[d]=c[d])}),n},yEe=function(n){for(var r=arguments.length,o=new Array(r>1?r-1:0),s=1;s<r;s++)o[s-1]=arguments[s];return vM(o,function(c){vM(c,function(d){n.push(d)})}),n};function Sve(n){for(var r=Object.keys(n),o=r.length,s=new Array(o);o--;)s[o]=[r[o],n[r[o]]];return s}var FFt=function(n){try{return n()}catch{return}},OPr=function(n){return function(){try{for(var r=arguments.length,o=new Array(r),s=0;s<r;s++)o[s]=arguments[s];return n.apply(this,o)}catch(c){zi.critical("Implementation error. Please turn on debug mode and open a ticket on https://app.posthog.com/home#panel=support%3Asupport%3A."),zi.critical(c)}}},qZe=function(n){var r={};return Bl(n,function(o,s){(Ih(o)&&o.length>0||G1(o))&&(r[s]=o)}),r};function NPr(n,r){return o=n,s=d=>Ih(d)&&!HM(r)?d.slice(0,r):d,c=new Set,function d(h,b){return h!==Object(h)?s?s(h,b):h:c.has(h)?void 0:(c.add(h),kf(h)?(y=[],vM(h,E=>{y.push(d(E))})):(y={},Bl(h,(E,k)=>{c.has(E)||(y[k]=d(E,k))})),y);var y}(o);var o,s,c}var RPr=["herokuapp.com","vercel.app","netlify.app"];function PPr(n){var r=n?.hostname;if(!Ih(r))return!1;var o=r.split(".").slice(-2).join(".");for(var s of RPr)if(o===s)return!1;return!0}function Xcn(n,r){for(var o=0;o<n.length;o++)if(r(n[o]))return n[o]}function th(n,r,o,s){var{capture:c=!1,passive:d=!0}=s??{};n?.addEventListener(r,o,{capture:c,passive:d})}var Jcn="$people_distinct_id",Ure="__alias",zre="__timers",BFt="$autocapture_disabled_server_side",kWe="$heatmaps_enabled_server_side",UFt="$exception_capture_enabled_server_side",TWe="$error_tracking_suppression_rules",zFt="$web_vitals_enabled_server_side",Zcn="$dead_clicks_enabled_server_side",HFt="$web_vitals_allowed_metrics",xWe="$session_recording_enabled_server_side",GFt="$console_log_recording_enabled_server_side",qFt="$session_recording_network_payload_capture",VFt="$session_recording_masking",WFt="$session_recording_canvas_recording",KFt="$replay_sample_rate",YFt="$replay_minimum_duration",XFt="$replay_script_config",Owe="$sesid",Hre="$session_is_sampled",CWe="$session_recording_url_trigger_activated_session",AWe="$session_recording_event_trigger_activated_session",hW="$enabled_feature_flags",_ve="$early_access_features",IWe="$feature_flag_details",Gre="$stored_person_properties",kB="$stored_group_properties",OWe="$surveys",x0e="$surveys_activated",Nwe="$flag_call_reported",C4="$user_state",NWe="$client_session_props",RWe="$capture_rate_limit",PWe="$initial_campaign_params",LWe="$initial_referrer_info",Rwe="$initial_person_info",Pwe="$epp",Qcn="__POSTHOG_TOOLBAR__",C0e="$posthog_cookieless",LPr=[Jcn,Ure,"__cmpns",zre,xWe,kWe,Owe,hW,TWe,C4,_ve,IWe,kB,Gre,OWe,Nwe,NWe,RWe,PWe,LWe,Pwe,Rwe];function JFt(n){return n instanceof Element&&(n.id===Qcn||!(n.closest==null||!n.closest(".toolbar-global-fade-container")))}function EEe(n){return!!n&&n.nodeType===1}function wM(n,r){return!!n&&!!n.tagName&&n.tagName.toLowerCase()===r.toLowerCase()}function eun(n){return!!n&&n.nodeType===3}function tun(n){return!!n&&n.nodeType===11}function VZe(n){return n?vEe(n).split(/\s+/):[]}function ZFt(n){var r=kn?.location.href;return!!(r&&n&&n.some(o=>r.match(o)))}function Lwe(n){var r="";switch(typeof n.className){case"string":r=n.className;break;case"object":r=(n.className&&"baseVal"in n.className?n.className.baseVal:null)||n.getAttribute("class")||"";break;default:r=""}return VZe(r)}function nun(n){return Gu(n)?null:vEe(n).split(/(\s+)/).filter(r=>aK(r)).join("").replace(/[\r\n]/g," ").replace(/[ ]+/g," ").substring(0,255)}function SEe(n){var r="";return MWe(n)&&!oun(n)&&n.childNodes&&n.childNodes.length&&Bl(n.childNodes,function(o){var s;eun(o)&&o.textContent&&(r+=(s=nun(o.textContent))!==null&&s!==void 0?s:"")}),vEe(r)}function run(n){return mi(n.target)?n.srcElement||null:(r=n.target)!=null&&r.shadowRoot?n.composedPath()[0]||null:n.target||null;var r}var DWe=["a","button","form","input","select","textarea","label"];function iun(n){var r=n.parentNode;return!(!r||!EEe(r))&&r}function DPr(n,r,o,s,c){var d,h,b;if(o===void 0&&(o=void 0),!kn||!n||wM(n,"html")||!EEe(n)||(d=o)!=null&&d.url_allowlist&&!ZFt(o.url_allowlist)||(h=o)!=null&&h.url_ignorelist&&ZFt(o.url_ignorelist))return!1;if((b=o)!=null&&b.dom_event_allowlist){var y=o.dom_event_allowlist;if(y&&!y.some(U=>r.type===U))return!1}for(var E=!1,k=[n],x=!0,O=n;O.parentNode&&!wM(O,"body");)if(tun(O.parentNode))k.push(O.parentNode.host),O=O.parentNode.host;else{if(!(x=iun(O)))break;if(s||DWe.indexOf(x.tagName.toLowerCase())>-1)E=!0;else{var P=kn.getComputedStyle(x);P&&P.getPropertyValue("cursor")==="pointer"&&(E=!0)}k.push(x),O=x}if(!function(U,B){var G=B?.element_allowlist;if(mi(G))return!0;var K,X=function(oe){if(G.some(he=>oe.tagName.toLowerCase()===he))return{v:!0}};for(var ae of U)if(K=X(ae))return K.v;return!1}(k,o)||!function(U,B){var G=B?.css_selector_allowlist;if(mi(G))return!0;var K,X=function(oe){if(G.some(he=>oe.matches(he)))return{v:!0}};for(var ae of U)if(K=X(ae))return K.v;return!1}(k,o))return!1;var D=kn.getComputedStyle(n);if(D&&D.getPropertyValue("cursor")==="pointer"&&r.type==="click")return!0;var $=n.tagName.toLowerCase();switch($){case"html":return!1;case"form":return(c||["submit"]).indexOf(r.type)>=0;case"input":case"select":case"textarea":return(c||["change","click"]).indexOf(r.type)>=0;default:return E?(c||["click"]).indexOf(r.type)>=0:(c||["click"]).indexOf(r.type)>=0&&(DWe.indexOf($)>-1||n.getAttribute("contenteditable")==="true")}}function MWe(n){for(var r=n;r.parentNode&&!wM(r,"body");r=r.parentNode){var o=Lwe(r);if(pc(o,"ph-sensitive")||pc(o,"ph-no-capture"))return!1}if(pc(Lwe(n),"ph-include"))return!0;var s=n.type||"";if(Ih(s))switch(s.toLowerCase()){case"hidden":case"password":return!1}var c=n.name||n.id||"";return!(Ih(c)&&/^cc|cardnum|ccnum|creditcard|csc|cvc|cvv|exp|pass|pwd|routing|seccode|securitycode|securitynum|socialsec|socsec|ssn/i.test(c.replace(/[^a-zA-Z0-9]/g,"")))}function oun(n){return!!(wM(n,"input")&&!["button","checkbox","submit","reset"].includes(n.type)||wM(n,"select")||wM(n,"textarea")||n.getAttribute("contenteditable")==="true")}var aun="(4[0-9]{12}(?:[0-9]{3})?)|(5[1-5][0-9]{14})|(6(?:011|5[0-9]{2})[0-9]{12})|(3[47][0-9]{13})|(3(?:0[0-5]|[68][0-9])[0-9]{11})|((?:2131|1800|35[0-9]{3})[0-9]{11})",MPr=new RegExp("^(?:"+aun+")$"),jPr=new RegExp(aun),sun="\\d{3}-?\\d{2}-?\\d{4}",$Pr=new RegExp("^("+sun+")$"),FPr=new RegExp("("+sun+")");function aK(n,r){return r===void 0&&(r=!0),!(Gu(n)||Ih(n)&&(n=vEe(n),(r?MPr:jPr).test((n||"").replace(/[- ]/g,""))||(r?$Pr:FPr).test(n)))}function cun(n){var r=SEe(n);return aK(r=(r+" "+uun(n)).trim())?r:""}function uun(n){var r="";return n&&n.childNodes&&n.childNodes.length&&Bl(n.childNodes,function(o){var s;if(o&&((s=o.tagName)==null?void 0:s.toLowerCase())==="span")try{var c=SEe(o);r=(r+" "+c).trim(),o.childNodes&&o.childNodes.length&&(r=(r+" "+uun(o)).trim())}catch(d){zi.error("[AutoCapture]",d)}}),r}function BPr(n){return function(r){var o=r.map(s=>{var c,d,h="";if(s.tag_name&&(h+=s.tag_name),s.attr_class)for(var b of(s.attr_class.sort(),s.attr_class))h+="."+b.replace(/"/g,"");var y=Ni({},s.text?{text:s.text}:{},{"nth-child":(c=s.nth_child)!==null&&c!==void 0?c:0,"nth-of-type":(d=s.nth_of_type)!==null&&d!==void 0?d:0},s.href?{href:s.href}:{},s.attr_id?{attr_id:s.attr_id}:{},s.attributes),E={};return Sve(y).sort((k,x)=>{var[O]=k,[P]=x;return O.localeCompare(P)}).forEach(k=>{var[x,O]=k;return E[QFt(x.toString())]=QFt(O.toString())}),h+=":",h+=Sve(E).map(k=>{var[x,O]=k;return x+'="'+O+'"'}).join("")});return o.join(";")}(function(r){return r.map(o=>{var s,c,d={text:(s=o.$el_text)==null?void 0:s.slice(0,400),tag_name:o.tag_name,href:(c=o.attr__href)==null?void 0:c.slice(0,2048),attr_class:UPr(o),attr_id:o.attr__id,nth_child:o.nth_child,nth_of_type:o.nth_of_type,attributes:{}};return Sve(o).filter(h=>{var[b]=h;return b.indexOf("attr__")===0}).forEach(h=>{var[b,y]=h;return d.attributes[b]=y}),d})}(n))}function QFt(n){return n.replace(/"|\\"/g,'\\"')}function UPr(n){var r=n.attr__class;return r?kf(r)?r:VZe(r):void 0}class lun{constructor(){this.clicks=[]}isRageClick(r,o,s){var c=this.clicks[this.clicks.length-1];if(c&&Math.abs(r-c.x)+Math.abs(o-c.y)<30&&s-c.timestamp<1e3){if(this.clicks.push({x:r,y:o,timestamp:s}),this.clicks.length===3)return!0}else this.clicks=[{x:r,y:o,timestamp:s}];return!1}}var zPr=["localhost","127.0.0.1"],sK=n=>{var r=Di?.createElement("a");return mi(r)?null:(r.href=n,r)},HPr=function(n,r){var o,s;r===void 0&&(r="&");var c=[];return Bl(n,function(d,h){mi(d)||mi(h)||h==="undefined"||(o=encodeURIComponent((b=>b instanceof File)(d)?d.name:d.toString()),s=encodeURIComponent(h),c[c.length]=s+"="+o)}),c.join(r)},Dwe=function(n,r){for(var o,s=((n.split("#")[0]||"").split(/\?(.*)/)[1]||"").replace(/^\?+/g,"").split("&"),c=0;c<s.length;c++){var d=s[c].split("=");if(d[0]===r){o=d;break}}if(!kf(o)||o.length<2)return"";var h=o[1];try{h=decodeURIComponent(h)}catch{zi.error("Skipping decoding for malformed query param: "+h)}return h.replace(/\+/g," ")},WZe=function(n,r,o){if(!n||!r||!r.length)return n;for(var s=n.split("#"),c=s[0]||"",d=s[1],h=c.split("?"),b=h[1],y=h[0],E=(b||"").split("&"),k=[],x=0;x<E.length;x++){var O=E[x].split("=");kf(O)&&(r.includes(O[0])?k.push(O[0]+"="+o):k.push(E[x]))}var P=y;return b!=null&&(P+="?"+k.join("&")),d!=null&&(P+="#"+d),P},Mwe=function(n,r){var o=n.match(new RegExp(r+"=([^&]*)"));return o?o[1]:null},eBt=Wm("[AutoCapture]");function J$e(n,r){return r.length>n?r.slice(0,n)+"...":r}function GPr(n){if(n.previousElementSibling)return n.previousElementSibling;var r=n;do r=r.previousSibling;while(r&&!EEe(r));return r}function qPr(n,r,o,s){var c=n.tagName.toLowerCase(),d={tag_name:c};DWe.indexOf(c)>-1&&!o&&(c.toLowerCase()==="a"||c.toLowerCase()==="button"?d.$el_text=J$e(1024,cun(n)):d.$el_text=J$e(1024,SEe(n)));var h=Lwe(n);h.length>0&&(d.classes=h.filter(function(k){return k!==""})),Bl(n.attributes,function(k){var x;if((!oun(n)||["name","id","class","aria-label"].indexOf(k.name)!==-1)&&(s==null||!s.includes(k.name))&&!r&&aK(k.value)&&(x=k.name,!Ih(x)||x.substring(0,10)!=="_ngcontent"&&x.substring(0,7)!=="_nghost")){var O=k.value;k.name==="class"&&(O=VZe(O).join(" ")),d["attr__"+k.name]=J$e(1024,O)}});for(var b=1,y=1,E=n;E=GPr(E);)b++,E.tagName===n.tagName&&y++;return d.nth_child=b,d.nth_of_type=y,d}function VPr(n,r){for(var o,s,{e:c,maskAllElementAttributes:d,maskAllText:h,elementAttributeIgnoreList:b,elementsChainAsString:y}=r,E=[n],k=n;k.parentNode&&!wM(k,"body");)tun(k.parentNode)?(E.push(k.parentNode.host),k=k.parentNode.host):(E.push(k.parentNode),k=k.parentNode);var x,O=[],P={},D=!1,$=!1;if(Bl(E,X=>{var ae=MWe(X);X.tagName.toLowerCase()==="a"&&(D=X.getAttribute("href"),D=ae&&D&&aK(D)&&D),pc(Lwe(X),"ph-no-capture")&&($=!0),O.push(qPr(X,d,h,b));var oe=function(he){if(!MWe(he))return{};var ye={};return Bl(he.attributes,function(we){if(we.name&&we.name.indexOf("data-ph-capture-attribute")===0){var ge=we.name.replace("data-ph-capture-attribute-",""),Te=we.value;ge&&Te&&aK(Te)&&(ye[ge]=Te)}}),ye}(X);pp(P,oe)}),$)return{props:{},explicitNoCapture:$};if(h||(n.tagName.toLowerCase()==="a"||n.tagName.toLowerCase()==="button"?O[0].$el_text=cun(n):O[0].$el_text=SEe(n)),D){var U,B;O[0].attr__href=D;var G=(U=sK(D))==null?void 0:U.host,K=kn==null||(B=kn.location)==null?void 0:B.host;G&&K&&G!==K&&(x=D)}return{props:pp({$event_type:c.type,$ce_version:1},y?{}:{$elements:O},{$elements_chain:BPr(O)},(o=O[0])!=null&&o.$el_text?{$el_text:(s=O[0])==null?void 0:s.$el_text}:{},x&&c.type==="click"?{$external_click_url:x}:{},P)}}class WPr{constructor(r){this.i=!1,this.o=null,this.rageclicks=new lun,this.h=!1,this.instance=r,this.m=null}get S(){var r,o,s=Ah(this.instance.config.autocapture)?this.instance.config.autocapture:{};return s.url_allowlist=(r=s.url_allowlist)==null?void 0:r.map(c=>new RegExp(c)),s.url_ignorelist=(o=s.url_ignorelist)==null?void 0:o.map(c=>new RegExp(c)),s}$(){if(this.isBrowserSupported()){if(kn&&Di){var r=s=>{s=s||kn?.event;try{this.k(s)}catch(c){eBt.error("Failed to capture event",c)}};if(th(Di,"submit",r,{capture:!0}),th(Di,"change",r,{capture:!0}),th(Di,"click",r,{capture:!0}),this.S.capture_copied_text){var o=s=>{s=s||kn?.event,this.k(s,X$e)};th(Di,"copy",o,{capture:!0}),th(Di,"cut",o,{capture:!0})}}}else eBt.info("Disabling Automatic Event Collection because this browser is not supported")}startIfEnabled(){this.isEnabled&&!this.i&&(this.$(),this.i=!0)}onRemoteConfig(r){r.elementsChainAsString&&(this.h=r.elementsChainAsString),this.instance.persistence&&this.instance.persistence.register({[BFt]:!!r.autocapture_opt_out}),this.o=!!r.autocapture_opt_out,this.startIfEnabled()}setElementSelectors(r){this.m=r}getElementSelectors(r){var o,s=[];return(o=this.m)==null||o.forEach(c=>{var d=Di?.querySelectorAll(c);d?.forEach(h=>{r===h&&s.push(c)})}),s}get isEnabled(){var r,o,s=(r=this.instance.persistence)==null?void 0:r.props[BFt],c=this.o;if(HM(c)&&!wI(s)&&!this.instance.I())return!1;var d=(o=this.o)!==null&&o!==void 0?o:!!s;return!!this.instance.config.autocapture&&!d}k(r,o){if(o===void 0&&(o="$autocapture"),this.isEnabled){var s,c=run(r);eun(c)&&(c=c.parentNode||null),o==="$autocapture"&&r.type==="click"&&r instanceof MouseEvent&&this.instance.config.rageclick&&(s=this.rageclicks)!=null&&s.isRageClick(r.clientX,r.clientY,new Date().getTime())&&this.k(r,"$rageclick");var d=o===X$e;if(c&&DPr(c,r,this.S,d,d?["copy","cut"]:void 0)){var{props:h,explicitNoCapture:b}=VPr(c,{e:r,maskAllElementAttributes:this.instance.config.mask_all_element_attributes,maskAllText:this.instance.config.mask_all_text,elementAttributeIgnoreList:this.S.element_attribute_ignorelist,elementsChainAsString:this.h});if(b)return!1;var y=this.getElementSelectors(c);if(y&&y.length>0&&(h.$element_selectors=y),o===X$e){var E,k=nun(kn==null||(E=kn.getSelection())==null?void 0:E.toString()),x=r.type||"clipboard";if(!k)return!1;h.$selected_content=k,h.$copy_type=x}return this.instance.capture(o,h),!0}}}isBrowserSupported(){return j_(Di?.querySelectorAll)}}Math.trunc||(Math.trunc=function(n){return n<0?Math.ceil(n):Math.floor(n)}),Number.isInteger||(Number.isInteger=function(n){return G1(n)&&isFinite(n)&&Math.floor(n)===n});var tBt="0123456789abcdef";class jwe{constructor(r){if(this.bytes=r,r.length!==16)throw new TypeError("not 128-bit length")}static fromFieldsV7(r,o,s,c){if(!Number.isInteger(r)||!Number.isInteger(o)||!Number.isInteger(s)||!Number.isInteger(c)||r<0||o<0||s<0||c<0||r>0xffffffffffff||o>4095||s>1073741823||c>4294967295)throw new RangeError("invalid field value");var d=new Uint8Array(16);return d[0]=r/Math.pow(2,40),d[1]=r/Math.pow(2,32),d[2]=r/Math.pow(2,24),d[3]=r/Math.pow(2,16),d[4]=r/Math.pow(2,8),d[5]=r,d[6]=112|o>>>8,d[7]=o,d[8]=128|s>>>24,d[9]=s>>>16,d[10]=s>>>8,d[11]=s,d[12]=c>>>24,d[13]=c>>>16,d[14]=c>>>8,d[15]=c,new jwe(d)}toString(){for(var r="",o=0;o<this.bytes.length;o++)r=r+tBt.charAt(this.bytes[o]>>>4)+tBt.charAt(15&this.bytes[o]),o!==3&&o!==5&&o!==7&&o!==9||(r+="-");if(r.length!==36)throw new Error("Invalid UUIDv7 was generated");return r}clone(){return new jwe(this.bytes.slice(0))}equals(r){return this.compareTo(r)===0}compareTo(r){for(var o=0;o<16;o++){var s=this.bytes[o]-r.bytes[o];if(s!==0)return Math.sign(s)}return 0}}class KPr{constructor(){this.P=0,this.R=0,this.T=new YPr}generate(){var r=this.generateOrAbort();if(mi(r)){this.P=0;var o=this.generateOrAbort();if(mi(o))throw new Error("Could not generate UUID after timestamp reset");return o}return r}generateOrAbort(){var r=Date.now();if(r>this.P)this.P=r,this.M();else{if(!(r+1e4>this.P))return;this.R++,this.R>4398046511103&&(this.P++,this.M())}return jwe.fromFieldsV7(this.P,Math.trunc(this.R/Math.pow(2,30)),this.R&Math.pow(2,30)-1,this.T.nextUint32())}M(){this.R=1024*this.T.nextUint32()+(1023&this.T.nextUint32())}}var nBt,dun=n=>{if(typeof UUIDV7_DENY_WEAK_RNG<"u"&&UUIDV7_DENY_WEAK_RNG)throw new Error("no cryptographically strong RNG available");for(var r=0;r<n.length;r++)n[r]=65536*Math.trunc(65536*Math.random())+Math.trunc(65536*Math.random());return n};kn&&!mi(kn.crypto)&&crypto.getRandomValues&&(dun=n=>crypto.getRandomValues(n));class YPr{constructor(){this.C=new Uint32Array(8),this.F=1/0}nextUint32(){return this.F>=this.C.length&&(dun(this.C),this.F=0),this.C[this.F++]}}var uM=()=>XPr().toString(),XPr=()=>(nBt||(nBt=new KPr)).generate(),wre="",JPr=/[a-z0-9][a-z0-9-]+\.[a-z]{2,}$/i;function ZPr(n,r){if(r){var o=function(c,d){if(d===void 0&&(d=Di),wre)return wre;if(!d||["localhost","127.0.0.1"].includes(c))return"";for(var h=c.split("."),b=Math.min(h.length,8),y="dmn_chk_"+uM();!wre&&b--;){var E=h.slice(b).join("."),k=y+"=1;domain=."+E+";path=/";d.cookie=k+";max-age=3",d.cookie.includes(y)&&(d.cookie=k+";max-age=0",wre=E)}return wre}(n);if(!o){var s=(c=>{var d=c.match(JPr);return d?d[0]:""})(n);s!==o&&zi.info("Warning: cookie subdomain discovery mismatch",s,o),o=s}return o?"; domain=."+o:""}return""}var AI={O:()=>!!Di,A:function(n){zi.error("cookieStore error: "+n)},D:function(n){if(Di){try{for(var r=n+"=",o=Di.cookie.split(";").filter(d=>d.length),s=0;s<o.length;s++){for(var c=o[s];c.charAt(0)==" ";)c=c.substring(1,c.length);if(c.indexOf(r)===0)return decodeURIComponent(c.substring(r.length,c.length))}}catch{}return null}},L:function(n){var r;try{r=JSON.parse(AI.D(n))||{}}catch{}return r},j:function(n,r,o,s,c){if(Di)try{var d="",h="",b=ZPr(Di.location.hostname,s);if(o){var y=new Date;y.setTime(y.getTime()+24*o*60*60*1e3),d="; expires="+y.toUTCString()}c&&(h="; secure");var E=n+"="+encodeURIComponent(JSON.stringify(r))+d+"; SameSite=Lax; path=/"+b+h;return E.length>3686.4&&zi.warn("cookieStore warning: large cookie, len="+E.length),Di.cookie=E,E}catch{return}},N:function(n,r){try{AI.j(n,"",-1,r)}catch{return}}},Z$e=null,Gg={O:function(){if(!HM(Z$e))return Z$e;var n=!0;if(mi(kn))n=!1;else try{var r="__mplssupport__";Gg.j(r,"xyz"),Gg.D(r)!=='"xyz"'&&(n=!1),Gg.N(r)}catch{n=!1}return n||zi.error("localStorage unsupported; falling back to cookie store"),Z$e=n,n},A:function(n){zi.error("localStorage error: "+n)},D:function(n){try{return kn?.localStorage.getItem(n)}catch(r){Gg.A(r)}return null},L:function(n){try{return JSON.parse(Gg.D(n))||{}}catch{}return null},j:function(n,r){try{kn?.localStorage.setItem(n,JSON.stringify(r))}catch(o){Gg.A(o)}},N:function(n){try{kn?.localStorage.removeItem(n)}catch(r){Gg.A(r)}}},QPr=["distinct_id",Owe,Hre,Pwe,Rwe],A0e=Ni({},Gg,{L:function(n){try{var r={};try{r=AI.L(n)||{}}catch{}var o=pp(r,JSON.parse(Gg.D(n)||"{}"));return Gg.j(n,o),o}catch{}return null},j:function(n,r,o,s,c,d){try{Gg.j(n,r,void 0,void 0,d);var h={};QPr.forEach(b=>{r[b]&&(h[b]=r[b])}),Object.keys(h).length&&AI.j(n,h,o,s,c,d)}catch(b){Gg.A(b)}},N:function(n,r){try{kn?.localStorage.removeItem(n),AI.N(n,r)}catch(o){Gg.A(o)}}}),I0e={},eLr={O:function(){return!0},A:function(n){zi.error("memoryStorage error: "+n)},D:function(n){return I0e[n]||null},L:function(n){return I0e[n]||null},j:function(n,r){I0e[n]=r},N:function(n){delete I0e[n]}},mB=null,Ub={O:function(){if(!HM(mB))return mB;if(mB=!0,mi(kn))mB=!1;else try{var n="__support__";Ub.j(n,"xyz"),Ub.D(n)!=='"xyz"'&&(mB=!1),Ub.N(n)}catch{mB=!1}return mB},A:function(n){zi.error("sessionStorage error: ",n)},D:function(n){try{return kn?.sessionStorage.getItem(n)}catch(r){Ub.A(r)}return null},L:function(n){try{return JSON.parse(Ub.D(n))||null}catch{}return null},j:function(n,r){try{kn?.sessionStorage.setItem(n,JSON.stringify(r))}catch(o){Ub.A(o)}},N:function(n){try{kn?.sessionStorage.removeItem(n)}catch(r){Ub.A(r)}}},KV=function(n){return n[n.PENDING=-1]="PENDING",n[n.DENIED=0]="DENIED",n[n.GRANTED=1]="GRANTED",n}({});class tLr{constructor(r){this._instance=r}get S(){return this._instance.config}get consent(){return this.U()?KV.DENIED:this.q}isOptedOut(){return this.consent===KV.DENIED||this.consent===KV.PENDING&&this.S.opt_out_capturing_by_default}isOptedIn(){return!this.isOptedOut()}optInOut(r){this.B.j(this.H,r?1:0,this.S.cookie_expiration,this.S.cross_subdomain_cookie,this.S.secure_cookie)}reset(){this.B.N(this.H,this.S.cross_subdomain_cookie)}get H(){var{token:r,opt_out_capturing_cookie_prefix:o}=this._instance.config;return(o||"__ph_opt_in_out_")+r}get q(){var r=this.B.D(this.H);return r==="1"?KV.GRANTED:r==="0"?KV.DENIED:KV.PENDING}get B(){if(!this.W){var r=this.S.opt_out_capturing_persistence_type;this.W=r==="localStorage"?Gg:AI;var o=r==="localStorage"?AI:Gg;o.D(this.H)&&(this.W.D(this.H)||this.optInOut(o.D(this.H)==="1"),o.N(this.H,this.S.cross_subdomain_cookie))}return this.W}U(){return!!this.S.respect_dnt&&!!Xcn([c2?.doNotTrack,c2?.msDoNotTrack,La.doNotTrack],r=>pc([!0,1,"1","yes"],r))}}var O0e=Wm("[Dead Clicks]"),nLr=()=>!0,rLr=n=>{var r,o=!((r=n.instance.persistence)==null||!r.get_property(Zcn)),s=n.instance.config.capture_dead_clicks;return wI(s)?s:o};let fun=class{get lazyLoadedDeadClicksAutocapture(){return this.G}constructor(r,o,s){this.instance=r,this.isEnabled=o,this.onCapture=s,this.startIfEnabled()}onRemoteConfig(r){this.instance.persistence&&this.instance.persistence.register({[Zcn]:r?.captureDeadClicks}),this.startIfEnabled()}startIfEnabled(){this.isEnabled(this)&&this.J(()=>{this.V()})}J(r){var o,s;(o=La.__PosthogExtensions__)!=null&&o.initDeadClicksAutocapture&&r(),(s=La.__PosthogExtensions__)==null||s.loadExternalDependency==null||s.loadExternalDependency(this.instance,"dead-clicks-autocapture",c=>{c?O0e.error("failed to load script",c):r()})}V(){var r;if(Di){if(!this.G&&(r=La.__PosthogExtensions__)!=null&&r.initDeadClicksAutocapture){var o=Ah(this.instance.config.capture_dead_clicks)?this.instance.config.capture_dead_clicks:{};o.__onCapture=this.onCapture,this.G=La.__PosthogExtensions__.initDeadClicksAutocapture(this.instance,o),this.G.start(Di),O0e.info("starting...")}}else O0e.error("`document` not found. Cannot start.")}stop(){this.G&&(this.G.stop(),this.G=void 0,O0e.info("stopping..."))}};function u2(n,r,o,s,c){return r>o&&(zi.warn("min cannot be greater than max."),r=o),G1(n)?n>o?(s&&zi.warn(s+" cannot be  greater than max: "+o+". Using max value instead."),o):n<r?(s&&zi.warn(s+" cannot be less than min: "+r+". Using min value instead."),r):n:(s&&zi.warn(s+" must be a number. using max or fallback. max: "+o+", fallback: "+c),u2(c||o,r,o,s))}let hun=class{constructor(r){this.K={},this.Y=()=>{Object.keys(this.K).forEach(o=>{var s=this.X(o)+this.Z;s>=this.tt?delete this.K[o]:this.it(o,s)})},this.X=o=>this.K[String(o)],this.it=(o,s)=>{this.K[String(o)]=s},this.consumeRateLimit=o=>{var s,c=(s=this.X(o))!==null&&s!==void 0?s:this.tt;if((c=Math.max(c-1,0))===0)return!0;this.it(o,c);var d,h=c===0;return h&&((d=this.et)==null||d.call(this,o)),h},this.rt=r,this.et=this.rt.et,this.tt=u2(this.rt.bucketSize,0,100,"rate limiter bucket size"),this.Z=u2(this.rt.refillRate,0,this.tt,"rate limiter refill rate"),this.st=u2(this.rt.refillInterval,0,864e5,"rate limiter refill interval"),setInterval(()=>{this.Y()},this.st)}};var N0e=Wm("[ExceptionAutocapture]");let iLr=class{constructor(r){var o,s,c;this.nt=()=>{var d;if(kn&&this.isEnabled&&(d=La.__PosthogExtensions__)!=null&&d.errorWrappingFunctions){var h=La.__PosthogExtensions__.errorWrappingFunctions.wrapOnError,b=La.__PosthogExtensions__.errorWrappingFunctions.wrapUnhandledRejection,y=La.__PosthogExtensions__.errorWrappingFunctions.wrapConsoleError;try{!this.ot&&this.S.capture_unhandled_errors&&(this.ot=h(this.captureException.bind(this))),!this.lt&&this.S.capture_unhandled_rejections&&(this.lt=b(this.captureException.bind(this))),!this.ut&&this.S.capture_console_errors&&(this.ut=y(this.captureException.bind(this)))}catch(E){N0e.error("failed to start",E),this.ht()}}},this._instance=r,this.dt=!((o=this._instance.persistence)==null||!o.props[UFt]),this.S=this.vt(),this.ct=new hun({refillRate:(s=this._instance.config.error_tracking.__exceptionRateLimiterRefillRate)!==null&&s!==void 0?s:1,bucketSize:(c=this._instance.config.error_tracking.__exceptionRateLimiterBucketSize)!==null&&c!==void 0?c:10,refillInterval:1e4}),this.startIfEnabled()}vt(){var r=this._instance.config.capture_exceptions,o={capture_unhandled_errors:!1,capture_unhandled_rejections:!1,capture_console_errors:!1};return Ah(r)?o=Ni({},o,r):(mi(r)?this.dt:r)&&(o=Ni({},o,{capture_unhandled_errors:!0,capture_unhandled_rejections:!0})),o}get isEnabled(){return this.S.capture_console_errors||this.S.capture_unhandled_errors||this.S.capture_unhandled_rejections}startIfEnabled(){this.isEnabled&&(N0e.info("enabled"),this.J(this.nt))}J(r){var o,s;(o=La.__PosthogExtensions__)!=null&&o.errorWrappingFunctions&&r(),(s=La.__PosthogExtensions__)==null||s.loadExternalDependency==null||s.loadExternalDependency(this._instance,"exception-autocapture",c=>{if(c)return N0e.error("failed to load script",c);r()})}ht(){var r,o,s;(r=this.ot)==null||r.call(this),this.ot=void 0,(o=this.lt)==null||o.call(this),this.lt=void 0,(s=this.ut)==null||s.call(this),this.ut=void 0}onRemoteConfig(r){var o=r.autocaptureExceptions;this.dt=!!o||!1,this.S=this.vt(),this._instance.persistence&&this._instance.persistence.register({[UFt]:this.dt}),this.startIfEnabled()}captureException(r){var o,s=this._instance.requestRouter.endpointFor("ui");r.$exception_personURL=s+"/project/"+this._instance.config.token+"/person/"+this._instance.get_distinct_id();var c=(o=r.$exception_list[0].type)!==null&&o!==void 0?o:"Exception";this.ct.consumeRateLimit(c)?N0e.info("Skipping exception capture because of client rate limiting.",{exception:r.$exception_list[0].type}):this._instance.exceptions.sendExceptionEvent(r)}};function pun(n){return!mi(Event)&&gun(n,Event)}function gun(n,r){try{return n instanceof r}catch{return!1}}function mun(n){switch(Object.prototype.toString.call(n)){case"[object Error]":case"[object Exception]":case"[object DOMException]":case"[object DOMError]":return!0;default:return gun(n,Error)}}function kve(n,r){return Object.prototype.toString.call(n)==="[object "+r+"]"}function Q$e(n){return kve(n,"DOMError")}var rBt=/\(error: (.*)\)/,iBt=50,pW="?";function eFe(n,r,o,s){var c={platform:"web:javascript",filename:n,function:r==="<anonymous>"?pW:r,in_app:!0};return mi(o)||(c.lineno=o),mi(s)||(c.colno=s),c}var oLr=/^\s*at (\S+?)(?::(\d+))(?::(\d+))\s*$/i,aLr=/^\s*at (?:(.+?\)(?: \[.+\])?|.*?) ?\((?:address at )?)?(?:async )?((?:<anonymous>|[-a-z]+:|.*bundle|\/)?.*?)(?::(\d+))?(?::(\d+))?\)?\s*$/i,sLr=/\((\S*)(?::(\d+))(?::(\d+))\)/,cLr=/^\s*(.*?)(?:\((.*?)\))?(?:^|@)?((?:[-a-z]+)?:\/.*?|\[native code\]|[^@]*(?:bundle|\d+\.js)|\/[\w\-. /=]+)(?::(\d+))?(?::(\d+))?\s*$/i,uLr=/(\S+) line (\d+)(?: > eval line \d+)* > eval/i,lLr=function(){for(var n=arguments.length,r=new Array(n),o=0;o<n;o++)r[o]=arguments[o];var s=r.sort((c,d)=>c[0]-d[0]).map(c=>c[1]);return function(c,d){d===void 0&&(d=0);for(var h=[],b=c.split(`
`),y=d;y<b.length;y++){var E=b[y];if(!(E.length>1024)){var k=rBt.test(E)?E.replace(rBt,"$1"):E;if(!k.match(/\S*Error: /)){for(var x of s){var O=x(k);if(O){h.push(O);break}}if(h.length>=iBt)break}}}return function(P){if(!P.length)return[];var D=Array.from(P);return D.reverse(),D.slice(0,iBt).map($=>Ni({},$,{filename:$.filename||dLr(D).filename,function:$.function||pW}))}(h)}}([30,n=>{var r=oLr.exec(n);if(r){var[,o,s,c]=r;return eFe(o,pW,+s,+c)}var d=aLr.exec(n);if(d){if(d[2]&&d[2].indexOf("eval")===0){var h=sLr.exec(d[2]);h&&(d[2]=h[1],d[3]=h[2],d[4]=h[3])}var[b,y]=aBt(d[1]||pW,d[2]);return eFe(y,b,d[3]?+d[3]:void 0,d[4]?+d[4]:void 0)}}],[50,n=>{var r=cLr.exec(n);if(r){if(r[3]&&r[3].indexOf(" > eval")>-1){var o=uLr.exec(r[3]);o&&(r[1]=r[1]||"eval",r[3]=o[1],r[4]=o[2],r[5]="")}var s=r[3],c=r[1]||pW;return[c,s]=aBt(c,s),eFe(s,c,r[4]?+r[4]:void 0,r[5]?+r[5]:void 0)}}]);function dLr(n){return n[n.length-1]||{}}var R0e,oBt,tFe,aBt=(n,r)=>{var o=n.indexOf("safari-extension")!==-1,s=n.indexOf("safari-web-extension")!==-1;return o||s?[n.indexOf("@")!==-1?n.split("@")[0]:pW,o?"safari-extension:"+r:"safari-web-extension:"+r]:[n,r]},fLr=/^(?:[Uu]ncaught (?:exception: )?)?(?:((?:Eval|Internal|Range|Reference|Syntax|Type|URI|)Error): )?(.*)$/i;function KZe(n,r){r===void 0&&(r=0);var o=n.stacktrace||n.stack||"",s=function(h){return h&&hLr.test(h.message)?1:0}(n);try{var c=lLr,d=function(h,b){var y=function(E){var k=globalThis._posthogChunkIds;if(!k)return{};var x=Object.keys(k);return tFe&&x.length===oBt||(oBt=x.length,tFe=x.reduce((O,P)=>{R0e||(R0e={});var D=R0e[P];if(D)O[D[0]]=D[1];else for(var $=E(P),U=$.length-1;U>=0;U--){var B=$[U],G=B?.filename,K=k[P];if(G&&K){O[G]=K,R0e[P]=[G,K];break}}return O},{})),tFe}(b);return h.forEach(E=>{E.filename&&(E.chunk_id=y[E.filename])}),h}(c(o,s),c);return d.slice(0,d.length-r)}catch{}return[]}var hLr=/Minified React error #\d+;/i;function pLr(n,r){var o,s,c=KZe(n),d=(o=r?.handled)===null||o===void 0||o,h=(s=r?.synthetic)!==null&&s!==void 0&&s;return{type:r!=null&&r.overrideExceptionType?r.overrideExceptionType:n.name,value:function(b){var y=b.message;return y.error&&typeof y.error.message=="string"?String(y.error.message):String(y)}(n),stacktrace:{frames:c,type:"raw"},mechanism:{handled:d,synthetic:h}}}function bun(n,r){var o=pLr(n,r);return n.cause&&mun(n.cause)&&n.cause!==n?[o,...bun(n.cause,{handled:r?.handled,synthetic:r?.synthetic})]:[o]}function nFe(n,r){return{$exception_list:bun(n,r),$exception_level:"error"}}function rFe(n,r){var o,s,c,d=(o=r?.handled)===null||o===void 0||o,h=(s=r?.synthetic)===null||s===void 0||s,b={type:r!=null&&r.overrideExceptionType?r.overrideExceptionType:(c=r?.defaultExceptionType)!==null&&c!==void 0?c:"Error",value:n||r?.defaultExceptionMessage,mechanism:{handled:d,synthetic:h}};if(r!=null&&r.syntheticException){var y=KZe(r.syntheticException,1);y.length&&(b.stacktrace={frames:y,type:"raw"})}return{$exception_list:[b],$exception_level:"error"}}function gLr(n){return Ih(n)&&!_We(n)&&TPr.indexOf(n)>=0}function mLr(n,r){var o,s,c=(o=r?.handled)===null||o===void 0||o,d=(s=r?.synthetic)===null||s===void 0||s,h=r!=null&&r.overrideExceptionType?r.overrideExceptionType:pun(n)?n.constructor.name:"Error",b="Non-Error 'exception' captured with keys: "+function(k,x){x===void 0&&(x=40);var O=Object.keys(k);if(O.sort(),!O.length)return"[object has no keys]";for(var P=O.length;P>0;P--){var D=O.slice(0,P).join(", ");if(!(D.length>x))return P===O.length||D.length<=x?D:D.slice(0,x)+"..."}return""}(n),y={type:h,value:b,mechanism:{handled:c,synthetic:d}};if(r!=null&&r.syntheticException){var E=KZe(r?.syntheticException,1);E.length&&(y.stacktrace={frames:E,type:"raw"})}return{$exception_list:[y],$exception_level:gLr(n.level)?n.level:"error"}}function bLr(n,r){var{error:o,event:s}=n,c={$exception_list:[]},d=o||s;if(Q$e(d)||function(O){return kve(O,"DOMException")}(d)){var h=d;if(function(O){return"stack"in O}(d))c=nFe(d,r);else{var b=h.name||(Q$e(h)?"DOMError":"DOMException"),y=h.message?b+": "+h.message:b;c=rFe(y,Ni({},r,{overrideExceptionType:Q$e(h)?"DOMError":"DOMException",defaultExceptionMessage:y}))}return"code"in h&&(c.$exception_DOMException_code=""+h.code),c}if(function(O){return kve(O,"ErrorEvent")}(d)&&d.error)return nFe(d.error,r);if(mun(d))return nFe(d,r);if(function(O){return kve(O,"Object")}(d)||pun(d))return mLr(d,r);if(mi(o)&&Ih(s)){var E="Error",k=s,x=s.match(fLr);return x&&(E=x[1],k=x[2]),rFe(k,Ni({},r,{overrideExceptionType:E,defaultExceptionMessage:k}))}return rFe(d,r)}function sBt(n,r,o){try{if(!(r in n))return()=>{};var s=n[r],c=o(s);return j_(c)&&(c.prototype=c.prototype||{},Object.defineProperties(c,{__posthog_wrapped__:{enumerable:!1,value:!0}})),n[r]=c,()=>{n[r]=s}}catch{return()=>{}}}let vLr=class{constructor(r){var o;this._instance=r,this.ft=(kn==null||(o=kn.location)==null?void 0:o.pathname)||""}get isEnabled(){return this._instance.config.capture_pageview==="history_change"}startIfEnabled(){this.isEnabled&&(zi.info("History API monitoring enabled, starting..."),this.monitorHistoryChanges())}stop(){this.gt&&this.gt(),this.gt=void 0,zi.info("History API monitoring stopped")}monitorHistoryChanges(){var r,o;if(kn&&kn.history){var s=this;(r=kn.history.pushState)!=null&&r.__posthog_wrapped__||sBt(kn.history,"pushState",c=>function(d,h,b){c.call(this,d,h,b),s._t("pushState")}),(o=kn.history.replaceState)!=null&&o.__posthog_wrapped__||sBt(kn.history,"replaceState",c=>function(d,h,b){c.call(this,d,h,b),s._t("replaceState")}),this.bt()}}_t(r){try{var o,s=kn==null||(o=kn.location)==null?void 0:o.pathname;if(!s)return;s!==this.ft&&this.isEnabled&&this._instance.capture("$pageview",{navigation_type:r}),this.ft=s}catch(c){zi.error("Error capturing "+r+" pageview",c)}}bt(){if(!this.gt){var r=()=>{this._t("popstate")};th(kn,"popstate",r),this.gt=()=>{kn&&kn.removeEventListener("popstate",r)}}}};function $we(n){var r,o;return((r=JSON.stringify(n,(o=[],function(s,c){if(Ah(c)){for(;o.length>0&&o[o.length-1]!==this;)o.pop();return o.includes(c)?"[Circular]":(o.push(c),c)}return c})))==null?void 0:r.length)||0}function jWe(n,r){if(r===void 0&&(r=66060288e-1),n.size>=r&&n.data.length>1){var o=Math.floor(n.data.length/2),s=n.data.slice(0,o),c=n.data.slice(o);return[jWe({size:$we(s),data:s,sessionId:n.sessionId,windowId:n.windowId}),jWe({size:$we(c),data:c,sessionId:n.sessionId,windowId:n.windowId})].flatMap(d=>d)}return[n]}var A4=(n=>(n[n.DomContentLoaded=0]="DomContentLoaded",n[n.Load=1]="Load",n[n.FullSnapshot=2]="FullSnapshot",n[n.IncrementalSnapshot=3]="IncrementalSnapshot",n[n.Meta=4]="Meta",n[n.Custom=5]="Custom",n[n.Plugin=6]="Plugin",n))(A4||{}),Cx=(n=>(n[n.Mutation=0]="Mutation",n[n.MouseMove=1]="MouseMove",n[n.MouseInteraction=2]="MouseInteraction",n[n.Scroll=3]="Scroll",n[n.ViewportResize=4]="ViewportResize",n[n.Input=5]="Input",n[n.TouchMove=6]="TouchMove",n[n.MediaInteraction=7]="MediaInteraction",n[n.StyleSheetRule=8]="StyleSheetRule",n[n.CanvasMutation=9]="CanvasMutation",n[n.Font=10]="Font",n[n.Log=11]="Log",n[n.Drag=12]="Drag",n[n.StyleDeclaration=13]="StyleDeclaration",n[n.Selection=14]="Selection",n[n.AdoptedStyleSheet=15]="AdoptedStyleSheet",n[n.CustomElement=16]="CustomElement",n))(Cx||{}),$We="[SessionRecording]",FWe="redacted",P0e={initiatorTypes:["audio","beacon","body","css","early-hint","embed","fetch","frame","iframe","icon","image","img","input","link","navigation","object","ping","script","track","video","xmlhttprequest"],maskRequestFn:n=>n,recordHeaders:!1,recordBody:!1,recordInitialRequests:!1,recordPerformance:!1,performanceEntryTypeToObserve:["first-input","navigation","paint","resource"],payloadSizeLimitBytes:1e6,payloadHostDenyList:[".lr-ingest.io",".ingest.sentry.io",".clarity.ms","analytics.google.com","bam.nr-data.net"]},wLr=["authorization","x-forwarded-for","authorization","cookie","set-cookie","x-api-key","x-real-ip","remote-addr","forwarded","proxy-authorization","x-csrf-token","x-csrftoken","x-xsrf-token"],yLr=["password","secret","passwd","api_key","apikey","auth","credentials","mysql_pwd","privatekey","private_key","token"],ELr=["/s/","/e/","/i/"];function cBt(n,r,o,s){if(Gu(n))return n;var c=r?.["content-length"]||function(d){return new Blob([d]).size}(n);return Ih(c)&&(c=parseInt(c)),c>o?$We+" "+s+" body too large to record ("+c+" bytes)":n}function uBt(n,r){if(Gu(n))return n;var o=n;return aK(o,!1)||(o=$We+" "+r+" body "+FWe),Bl(yLr,s=>{var c,d;(c=o)!=null&&c.length&&((d=o)==null?void 0:d.indexOf(s))!==-1&&(o=$We+" "+r+" body "+FWe+" as might contain: "+s)}),o}var SLr=(n,r)=>{var o,s,c,d={payloadSizeLimitBytes:P0e.payloadSizeLimitBytes,performanceEntryTypeToObserve:[...P0e.performanceEntryTypeToObserve],payloadHostDenyList:[...r.payloadHostDenyList||[],...P0e.payloadHostDenyList]},h=n.session_recording.recordHeaders!==!1&&r.recordHeaders,b=n.session_recording.recordBody!==!1&&r.recordBody,y=n.capture_performance!==!1&&r.recordPerformance,E=(o=d,c=Math.min(1e6,(s=o.payloadSizeLimitBytes)!==null&&s!==void 0?s:1e6),O=>(O!=null&&O.requestBody&&(O.requestBody=cBt(O.requestBody,O.requestHeaders,c,"Request")),O!=null&&O.responseBody&&(O.responseBody=cBt(O.responseBody,O.responseHeaders,c,"Response")),O)),k=O=>{return E((($,U)=>{var B,G=sK($.name),K=U.indexOf("http")===0?(B=sK(U))==null?void 0:B.pathname:U;K==="/"&&(K="");var X=G?.pathname.replace(K||"","");if(!(G&&X&&ELr.some(ae=>X.indexOf(ae)===0)))return $})((D=(P=O).requestHeaders,Gu(D)||Bl(Object.keys(D??{}),$=>{wLr.includes($.toLowerCase())&&(D[$]=FWe)}),P),n.api_host));var P,D},x=j_(n.session_recording.maskNetworkRequestFn);return x&&j_(n.session_recording.maskCapturedNetworkRequestFn)&&zi.warn("Both `maskNetworkRequestFn` and `maskCapturedNetworkRequestFn` are defined. `maskNetworkRequestFn` will be ignored."),x&&(n.session_recording.maskCapturedNetworkRequestFn=O=>{var P=n.session_recording.maskNetworkRequestFn({url:O.name});return Ni({},O,{name:P?.url})}),d.maskRequestFn=j_(n.session_recording.maskCapturedNetworkRequestFn)?O=>{var P,D=k(O);return D&&(P=n.session_recording.maskCapturedNetworkRequestFn==null?void 0:n.session_recording.maskCapturedNetworkRequestFn(D))!==null&&P!==void 0?P:void 0}:O=>function(P){if(!mi(P))return P.requestBody=uBt(P.requestBody,"Request"),P.responseBody=uBt(P.responseBody,"Response"),P}(k(O)),Ni({},P0e,d,{recordHeaders:h,recordBody:b,recordPerformance:y,recordInitialRequests:y})};class _Lr{constructor(r,o){var s,c;o===void 0&&(o={}),this.yt={},this.wt=d=>{if(!this.yt[d]){var h,b;this.yt[d]=!0;var y=this.St(d);(h=(b=this.rt).onBlockedNode)==null||h.call(b,d,y)}},this.$t=d=>{var h=this.St(d);if(h?.nodeName!=="svg"&&h instanceof Element){var b=h.closest("svg");if(b)return[this._rrweb.mirror.getId(b),b]}return[d,h]},this.St=d=>this._rrweb.mirror.getNode(d),this.kt=d=>{var h,b,y,E,k,x,O,P;return((h=(b=d.removes)==null?void 0:b.length)!==null&&h!==void 0?h:0)+((y=(E=d.attributes)==null?void 0:E.length)!==null&&y!==void 0?y:0)+((k=(x=d.texts)==null?void 0:x.length)!==null&&k!==void 0?k:0)+((O=(P=d.adds)==null?void 0:P.length)!==null&&O!==void 0?O:0)},this.throttleMutations=d=>{if(d.type!==3||d.data.source!==0)return d;var h=d.data,b=this.kt(h);h.attributes&&(h.attributes=h.attributes.filter(E=>{var[k]=this.$t(E.id);return!this.ct.consumeRateLimit(k)&&E}));var y=this.kt(h);return y!==0||b===y?d:void 0},this._rrweb=r,this.rt=o,this.ct=new hun({bucketSize:(s=this.rt.bucketSize)!==null&&s!==void 0?s:100,refillRate:(c=this.rt.refillRate)!==null&&c!==void 0?c:10,refillInterval:1e3,et:this.wt})}}var $_=Uint8Array,Zw=Uint16Array,cK=Uint32Array,YZe=new $_([0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,0,0,0,0]),XZe=new $_([0,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,0,0]),lBt=new $_([16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15]),vun=function(n,r){for(var o=new Zw(31),s=0;s<31;++s)o[s]=r+=1<<n[s-1];var c=new cK(o[30]);for(s=1;s<30;++s)for(var d=o[s];d<o[s+1];++d)c[d]=d-o[s]<<5|s;return[o,c]},wun=vun(YZe,2),kLr=wun[0],BWe=wun[1];kLr[28]=258,BWe[258]=28;for(var dBt=vun(XZe,0)[1],yun=new Zw(32768),Gd=0;Gd<32768;++Gd){var bB=(43690&Gd)>>>1|(21845&Gd)<<1;bB=(61680&(bB=(52428&bB)>>>2|(13107&bB)<<2))>>>4|(3855&bB)<<4,yun[Gd]=((65280&bB)>>>8|(255&bB)<<8)>>>1}var pie=function(n,r,o){for(var s=n.length,c=0,d=new Zw(r);c<s;++c)++d[n[c]-1];var h,b=new Zw(r);for(c=0;c<r;++c)b[c]=b[c-1]+d[c-1]<<1;for(h=new Zw(s),c=0;c<s;++c)h[c]=yun[b[n[c]-1]++]>>>15-n[c];return h},rU=new $_(288);for(Gd=0;Gd<144;++Gd)rU[Gd]=8;for(Gd=144;Gd<256;++Gd)rU[Gd]=9;for(Gd=256;Gd<280;++Gd)rU[Gd]=7;for(Gd=280;Gd<288;++Gd)rU[Gd]=8;var Fwe=new $_(32);for(Gd=0;Gd<32;++Gd)Fwe[Gd]=5;var TLr=pie(rU,9),xLr=pie(Fwe,5),Eun=function(n){return(n/8>>0)+(7&n&&1)},Sun=function(n,r,o){(o==null||o>n.length)&&(o=n.length);var s=new(n instanceof Zw?Zw:n instanceof cK?cK:$_)(o-r);return s.set(n.subarray(r,o)),s},E4=function(n,r,o){o<<=7&r;var s=r/8>>0;n[s]|=o,n[s+1]|=o>>>8},yre=function(n,r,o){o<<=7&r;var s=r/8>>0;n[s]|=o,n[s+1]|=o>>>8,n[s+2]|=o>>>16},iFe=function(n,r){for(var o=[],s=0;s<n.length;++s)n[s]&&o.push({s,f:n[s]});var c=o.length,d=o.slice();if(!c)return[new $_(0),0];if(c==1){var h=new $_(o[0].s+1);return h[o[0].s]=1,[h,1]}o.sort(function(ae,oe){return ae.f-oe.f}),o.push({s:-1,f:25001});var b=o[0],y=o[1],E=0,k=1,x=2;for(o[0]={s:-1,f:b.f+y.f,l:b,r:y};k!=c-1;)b=o[o[E].f<o[x].f?E++:x++],y=o[E!=k&&o[E].f<o[x].f?E++:x++],o[k++]={s:-1,f:b.f+y.f,l:b,r:y};var O=d[0].s;for(s=1;s<c;++s)d[s].s>O&&(O=d[s].s);var P=new Zw(O+1),D=UWe(o[k-1],P,0);if(D>r){s=0;var $=0,U=D-r,B=1<<U;for(d.sort(function(ae,oe){return P[oe.s]-P[ae.s]||ae.f-oe.f});s<c;++s){var G=d[s].s;if(!(P[G]>r))break;$+=B-(1<<D-P[G]),P[G]=r}for($>>>=U;$>0;){var K=d[s].s;P[K]<r?$-=1<<r-P[K]++-1:++s}for(;s>=0&&$;--s){var X=d[s].s;P[X]==r&&(--P[X],++$)}D=r}return[new $_(P),D]},UWe=function(n,r,o){return n.s==-1?Math.max(UWe(n.l,r,o+1),UWe(n.r,r,o+1)):r[n.s]=o},fBt=function(n){for(var r=n.length;r&&!n[--r];);for(var o=new Zw(++r),s=0,c=n[0],d=1,h=function(y){o[s++]=y},b=1;b<=r;++b)if(n[b]==c&&b!=r)++d;else{if(!c&&d>2){for(;d>138;d-=138)h(32754);d>2&&(h(d>10?d-11<<5|28690:d-3<<5|12305),d=0)}else if(d>3){for(h(c),--d;d>6;d-=6)h(8304);d>2&&(h(d-3<<5|8208),d=0)}for(;d--;)h(c);d=1,c=n[b]}return[o.subarray(0,s),r]},Ere=function(n,r){for(var o=0,s=0;s<r.length;++s)o+=n[s]*r[s];return o},zWe=function(n,r,o){var s=o.length,c=Eun(r+2);n[c]=255&s,n[c+1]=s>>>8,n[c+2]=255^n[c],n[c+3]=255^n[c+1];for(var d=0;d<s;++d)n[c+d+4]=o[d];return 8*(c+4+s)},hBt=function(n,r,o,s,c,d,h,b,y,E,k){E4(r,k++,o),++c[256];for(var x=iFe(c,15),O=x[0],P=x[1],D=iFe(d,15),$=D[0],U=D[1],B=fBt(O),G=B[0],K=B[1],X=fBt($),ae=X[0],oe=X[1],he=new Zw(19),ye=0;ye<G.length;++ye)he[31&G[ye]]++;for(ye=0;ye<ae.length;++ye)he[31&ae[ye]]++;for(var we=iFe(he,7),ge=we[0],Te=we[1],xe=19;xe>4&&!ge[lBt[xe-1]];--xe);var Ue,Z,Re,_e,qe=E+5<<3,it=Ere(c,rU)+Ere(d,Fwe)+h,Xe=Ere(c,O)+Ere(d,$)+h+14+3*xe+Ere(he,ge)+(2*he[16]+3*he[17]+7*he[18]);if(qe<=it&&qe<=Xe)return zWe(r,k,n.subarray(y,y+E));if(E4(r,k,1+(Xe<it)),k+=2,Xe<it){Ue=pie(O,P),Z=O,Re=pie($,U),_e=$;var de=pie(ge,Te);for(E4(r,k,K-257),E4(r,k+5,oe-1),E4(r,k+10,xe-4),k+=14,ye=0;ye<xe;++ye)E4(r,k+3*ye,ge[lBt[ye]]);k+=3*xe;for(var We=[G,ae],ht=0;ht<2;++ht){var Ie=We[ht];for(ye=0;ye<Ie.length;++ye){var $t=31&Ie[ye];E4(r,k,de[$t]),k+=ge[$t],$t>15&&(E4(r,k,Ie[ye]>>>5&127),k+=Ie[ye]>>>12)}}}else Ue=TLr,Z=rU,Re=xLr,_e=Fwe;for(ye=0;ye<b;++ye)if(s[ye]>255){$t=s[ye]>>>18&31,yre(r,k,Ue[$t+257]),k+=Z[$t+257],$t>7&&(E4(r,k,s[ye]>>>23&31),k+=YZe[$t]);var An=31&s[ye];yre(r,k,Re[An]),k+=_e[An],An>3&&(yre(r,k,s[ye]>>>5&8191),k+=XZe[An])}else yre(r,k,Ue[s[ye]]),k+=Z[s[ye]];return yre(r,k,Ue[256]),k+Z[256]},CLr=new cK([65540,131080,131088,131104,262176,1048704,1048832,2114560,2117632]),ALr=function(){for(var n=new cK(256),r=0;r<256;++r){for(var o=r,s=9;--s;)o=(1&o&&3988292384)^o>>>1;n[r]=o}return n}(),ILr=function(){var n=4294967295;return{p:function(r){for(var o=n,s=0;s<r.length;++s)o=ALr[255&o^r[s]]^o>>>8;n=o},d:function(){return 4294967295^n}}},OLr=function(n,r,o,s,c){return function(d,h,b,y,E,k){var x=d.length,O=new $_(y+x+5*(1+Math.floor(x/7e3))+E),P=O.subarray(y,O.length-E),D=0;if(!h||x<8)for(var $=0;$<=x;$+=65535){var U=$+65535;U<x?D=zWe(P,D,d.subarray($,U)):(P[$]=k,D=zWe(P,D,d.subarray($,x)))}else{for(var B=CLr[h-1],G=B>>>13,K=8191&B,X=(1<<b)-1,ae=new Zw(32768),oe=new Zw(X+1),he=Math.ceil(b/3),ye=2*he,we=function(qi){return(d[qi]^d[qi+1]<<he^d[qi+2]<<ye)&X},ge=new cK(25e3),Te=new Zw(288),xe=new Zw(32),Ue=0,Z=0,Re=($=0,0),_e=0,qe=0;$<x;++$){var it=we($),Xe=32767&$,de=oe[it];if(ae[Xe]=de,oe[it]=Xe,_e<=$){var We=x-$;if((Ue>7e3||Re>24576)&&We>423){D=hBt(d,P,0,ge,Te,xe,Z,Re,qe,$-qe,D),Re=Ue=Z=0,qe=$;for(var ht=0;ht<286;++ht)Te[ht]=0;for(ht=0;ht<30;++ht)xe[ht]=0}var Ie=2,$t=0,An=K,Yt=Xe-de&32767;if(We>2&&it==we($-Yt))for(var Or=Math.min(G,We)-1,Nt=Math.min(32767,$),on=Math.min(258,We);Yt<=Nt&&--An&&Xe!=de;){if(d[$+Ie]==d[$+Ie-Yt]){for(var br=0;br<on&&d[$+br]==d[$+br-Yt];++br);if(br>Ie){if(Ie=br,$t=Yt,br>Or)break;var wi=Math.min(Yt,br-2),no=0;for(ht=0;ht<wi;++ht){var Ri=$-Yt+ht+32768&32767,ro=Ri-ae[Ri]+32768&32767;ro>no&&(no=ro,de=Ri)}}}Yt+=(Xe=de)-(de=ae[Xe])+32768&32767}if($t){ge[Re++]=268435456|BWe[Ie]<<18|dBt[$t];var Ko=31&BWe[Ie],Ca=31&dBt[$t];Z+=YZe[Ko]+XZe[Ca],++Te[257+Ko],++xe[Ca],_e=$+Ie,++Ue}else ge[Re++]=d[$],++Te[d[$]]}}D=hBt(d,P,k,ge,Te,xe,Z,Re,qe,$-qe,D)}return Sun(O,0,y+Eun(D)+E)}(n,r.level==null?6:r.level,r.mem==null?Math.ceil(1.5*Math.max(8,Math.min(13,Math.log(n.length)))):12+r.mem,o,s,!0)},HWe=function(n,r,o){for(;o;++r)n[r]=o,o>>>=8},NLr=function(n,r){var o=r.filename;if(n[0]=31,n[1]=139,n[2]=8,n[8]=r.level<2?4:r.level==9?2:0,n[9]=3,r.mtime!=0&&HWe(n,4,Math.floor(new Date(r.mtime||Date.now())/1e3)),o){n[3]=8;for(var s=0;s<=o.length;++s)n[s+10]=o.charCodeAt(s)}},RLr=function(n){return 10+(n.filename&&n.filename.length+1||0)};function _un(n,r){r===void 0&&(r={});var o=ILr(),s=n.length;o.p(n);var c=OLr(n,r,RLr(r),8),d=c.length;return NLr(c,r),HWe(c,d-8,o.d()),HWe(c,d-4,s),c}function kun(n,r){var o=n.length;if(typeof TextEncoder<"u")return new TextEncoder().encode(n);for(var s=new $_(n.length+(n.length>>>1)),c=0,d=function(E){s[c++]=E},h=0;h<o;++h){if(c+5>s.length){var b=new $_(c+8+(o-h<<1));b.set(s),s=b}var y=n.charCodeAt(h);y<128||r?d(y):y<2048?(d(192|y>>>6),d(128|63&y)):y>55295&&y<57344?(d(240|(y=65536+(1047552&y)|1023&n.charCodeAt(++h))>>>18),d(128|y>>>12&63),d(128|y>>>6&63),d(128|63&y)):(d(224|y>>>12),d(128|y>>>6&63),d(128|63&y))}return Sun(s,0,c)}function PLr(n,r){return function(o){for(var s=0,c=0;c<o.length;c++)s=(s<<5)-s+o.charCodeAt(c),s|=0;return Math.abs(s)}(n)%100<u2(100*r,0,100)}var LM="disabled",JZe="sampled",Bwe="active",iU="buffering",ZZe="paused",QZe="trigger",II=QZe+"_activated",Y1=QZe+"_pending",MI=QZe+"_"+LM;function pBt(n,r){return r.some(o=>o.matching==="regex"&&new RegExp(o.url).test(n))}class Tun{constructor(r){this.xt=r}triggerStatus(r){var o=this.xt.map(s=>s.triggerStatus(r));return o.includes(II)?II:o.includes(Y1)?Y1:MI}stop(){this.xt.forEach(r=>r.stop())}}class xun{constructor(r){this.xt=r}triggerStatus(r){var o=new Set;for(var s of this.xt)o.add(s.triggerStatus(r));switch(o.delete(MI),o.size){case 0:return MI;case 1:return Array.from(o)[0];default:return Y1}}stop(){this.xt.forEach(r=>r.stop())}}class LLr{triggerStatus(){return Y1}stop(){}}class DLr{constructor(r){this.Et=[],this.It=[],this.urlBlocked=!1,this._instance=r}onRemoteConfig(r){var o,s;this.Et=((o=r.sessionRecording)==null?void 0:o.urlTriggers)||[],this.It=((s=r.sessionRecording)==null?void 0:s.urlBlocklist)||[]}Pt(r){var o;return this.Et.length===0?MI:((o=this._instance)==null?void 0:o.get_property(CWe))===r?II:Y1}triggerStatus(r){var o=this.Pt(r),s=o===II?II:o===Y1?Y1:MI;return this._instance.register_for_session({$sdk_debug_replay_url_trigger_status:s}),s}checkUrlTriggerConditions(r,o,s){if(kn!==void 0&&kn.location.href){var c=kn.location.href,d=this.urlBlocked,h=pBt(c,this.It);d&&h||(h&&!d?r():!h&&d&&o(),pBt(c,this.Et)&&s("url"))}}stop(){}}class MLr{constructor(r){this.linkedFlag=null,this.linkedFlagSeen=!1,this.Rt=()=>{},this._instance=r}triggerStatus(){var r=Y1;return Gu(this.linkedFlag)&&(r=MI),this.linkedFlagSeen&&(r=II),this._instance.register_for_session({$sdk_debug_replay_linked_flag_trigger_status:r}),r}onRemoteConfig(r,o){var s;if(this.linkedFlag=((s=r.sessionRecording)==null?void 0:s.linkedFlag)||null,!Gu(this.linkedFlag)&&!this.linkedFlagSeen){var c=Ih(this.linkedFlag)?this.linkedFlag:this.linkedFlag.flag,d=Ih(this.linkedFlag)?null:this.linkedFlag.variant;this.Rt=this._instance.onFeatureFlags((h,b)=>{var y=!1;if(Ah(b)&&c in b){var E=b[c];y=wI(E)?E===!0:d?E===d:!!E}this.linkedFlagSeen=y,y&&o(c,d)})}}stop(){this.Rt()}}class jLr{constructor(r){this.Tt=[],this._instance=r}onRemoteConfig(r){var o;this.Tt=((o=r.sessionRecording)==null?void 0:o.eventTriggers)||[]}Mt(r){var o;return this.Tt.length===0?MI:((o=this._instance)==null?void 0:o.get_property(AWe))===r?II:Y1}triggerStatus(r){var o=this.Mt(r),s=o===II?II:o===Y1?Y1:MI;return this._instance.register_for_session({$sdk_debug_replay_event_trigger_status:s}),s}stop(){}}function $Lr(n){return n.isRecordingEnabled?iU:LM}function FLr(n){if(!n.receivedFlags)return iU;if(!n.isRecordingEnabled)return LM;if(n.urlTriggerMatching.urlBlocked)return ZZe;var r=n.isSampled===!0,o=new Tun([n.eventTriggerMatching,n.urlTriggerMatching,n.linkedFlagMatching]).triggerStatus(n.sessionId);return r?JZe:o===II?Bwe:o===Y1?iU:n.isSampled===!1?LM:Bwe}function BLr(n){if(!n.receivedFlags)return iU;if(!n.isRecordingEnabled)return LM;if(n.urlTriggerMatching.urlBlocked)return ZZe;var r=new xun([n.eventTriggerMatching,n.urlTriggerMatching,n.linkedFlagMatching]).triggerStatus(n.sessionId),o=r!==MI,s=wI(n.isSampled);return o&&r===Y1?iU:o&&r===MI||s&&!n.isSampled?LM:n.isSampled===!0?JZe:Bwe}var qre="[SessionRecording]",U1=Wm(qre);function L0e(){var n;return La==null||(n=La.__PosthogExtensions__)==null||(n=n.rrweb)==null?void 0:n.record}var ULr=3e5,zLr=[Cx.MouseMove,Cx.MouseInteraction,Cx.Scroll,Cx.ViewportResize,Cx.Input,Cx.TouchMove,Cx.MediaInteraction,Cx.Drag],gBt=n=>({rrwebMethod:n,enqueuedAt:Date.now(),attempt:1});function vB(n){return function(r,o){for(var s="",c=0;c<r.length;){var d=r[c++];s+=String.fromCharCode(d)}return s}(_un(kun(JSON.stringify(n))))}function mBt(n){return n.type===A4.Custom&&n.data.tag==="sessionIdle"}class HLr{get sessionId(){return this.Ct}get Ft(){return this._instance.config.session_recording.session_idle_threshold_ms||3e5}get started(){return this.Ot}get At(){if(!this._instance.sessionManager)throw new Error(qre+" must be started with a valid sessionManager.");return this._instance.sessionManager}get Dt(){var r,o;return this.Lt.triggerStatus(this.sessionId)===Y1?6e4:(r=(o=this._instance.config.session_recording)==null?void 0:o.full_snapshot_interval_millis)!==null&&r!==void 0?r:ULr}get jt(){var r=this._instance.get_property(Hre);return wI(r)?r:null}get Nt(){var r,o,s=(r=this.C)==null?void 0:r.data[((o=this.C)==null?void 0:o.data.length)-1],{sessionStartTimestamp:c}=this.At.checkAndGetSessionAndWindowId(!0);return s?s.timestamp-c:null}get zt(){var r=!!this._instance.get_property(xWe),o=!this._instance.config.disable_session_recording;return kn&&r&&o}get Ut(){var r=!!this._instance.get_property(GFt),o=this._instance.config.enable_recording_console_log;return o??r}get qt(){var r,o,s,c,d,h,b=this._instance.config.session_recording.captureCanvas,y=this._instance.get_property(WFt),E=(r=(o=b?.recordCanvas)!==null&&o!==void 0?o:y?.enabled)!==null&&r!==void 0&&r,k=(s=(c=b?.canvasFps)!==null&&c!==void 0?c:y?.fps)!==null&&s!==void 0?s:4,x=(d=(h=b?.canvasQuality)!==null&&h!==void 0?h:y?.quality)!==null&&d!==void 0?d:.4;if(typeof x=="string"){var O=parseFloat(x);x=isNaN(O)?.4:O}return{enabled:E,fps:u2(k,0,12,"canvas recording fps",4),quality:u2(x,0,1,"canvas recording quality",.4)}}get Bt(){var r,o,s=this._instance.get_property(qFt),c={recordHeaders:(r=this._instance.config.session_recording)==null?void 0:r.recordHeaders,recordBody:(o=this._instance.config.session_recording)==null?void 0:o.recordBody},d=c?.recordHeaders||s?.recordHeaders,h=c?.recordBody||s?.recordBody,b=Ah(this._instance.config.capture_performance)?this._instance.config.capture_performance.network_timing:this._instance.config.capture_performance,y=!!(wI(b)?b:s?.capturePerformance);return d||h||y?{recordHeaders:d,recordBody:h,recordPerformance:y}:void 0}get Ht(){var r,o,s,c,d,h,b=this._instance.get_property(VFt),y={maskAllInputs:(r=this._instance.config.session_recording)==null?void 0:r.maskAllInputs,maskTextSelector:(o=this._instance.config.session_recording)==null?void 0:o.maskTextSelector,blockSelector:(s=this._instance.config.session_recording)==null?void 0:s.blockSelector},E=(c=y?.maskAllInputs)!==null&&c!==void 0?c:b?.maskAllInputs,k=(d=y?.maskTextSelector)!==null&&d!==void 0?d:b?.maskTextSelector,x=(h=y?.blockSelector)!==null&&h!==void 0?h:b?.blockSelector;return mi(E)&&mi(k)&&mi(x)?void 0:{maskAllInputs:E==null||E,maskTextSelector:k,blockSelector:x}}get Wt(){var r=this._instance.get_property(KFt);return G1(r)?r:null}get Gt(){var r=this._instance.get_property(YFt);return G1(r)?r:null}get status(){return this.Jt?this.Vt({receivedFlags:this.Jt,isRecordingEnabled:this.zt,isSampled:this.jt,urlTriggerMatching:this.Kt,eventTriggerMatching:this.Yt,linkedFlagMatching:this.Xt,sessionId:this.sessionId}):iU}constructor(r){if(this.Vt=$Lr,this.Jt=!1,this.Qt=[],this.Zt="unknown",this.ti=Date.now(),this.Lt=new LLr,this.ii=void 0,this.ei=void 0,this.ri=void 0,this.si=void 0,this.ni=void 0,this._forceAllowLocalhostNetworkCapture=!1,this.oi=()=>{this.ai()},this.li=()=>{this.ui("browser offline",{})},this.hi=()=>{this.ui("browser online",{})},this.di=()=>{if(Di!=null&&Di.visibilityState){var c="window "+Di.visibilityState;this.ui(c,{})}},this._instance=r,this.Ot=!1,this.vi="/s/",this.ci=void 0,this.Jt=!1,!this._instance.sessionManager)throw U1.error("started without valid sessionManager"),new Error(qre+" started without valid sessionManager. This is a bug.");if(this._instance.config.__preview_experimental_cookieless_mode)throw new Error(qre+" cannot be used with __preview_experimental_cookieless_mode.");this.Xt=new MLr(this._instance),this.Kt=new DLr(this._instance),this.Yt=new jLr(this._instance);var{sessionId:o,windowId:s}=this.At.checkAndGetSessionAndWindowId();this.Ct=o,this.fi=s,this.C=this.pi(),this.Ft>=this.At.sessionTimeoutMs&&U1.warn("session_idle_threshold_ms ("+this.Ft+") is greater than the session timeout ("+this.At.sessionTimeoutMs+"). Session will never be detected as idle")}startIfEnabledOrStop(r){this.zt?(this.gi(r),th(kn,"beforeunload",this.oi),th(kn,"offline",this.li),th(kn,"online",this.hi),th(kn,"visibilitychange",this.di),this.mi(),this.bi(),Gu(this.ii)&&(this.ii=this._instance.on("eventCaptured",o=>{try{if(o.event==="$pageview"){var s=o!=null&&o.properties.$current_url?this.yi(o?.properties.$current_url):"";if(!s)return;this.ui("$pageview",{href:s})}}catch(c){U1.error("Could not add $pageview to rrweb session",c)}})),this.ei||(this.ei=this.At.onSessionId((o,s,c)=>{var d,h;c&&(this.ui("$session_id_change",{sessionId:o,windowId:s,changeReason:c}),(d=this._instance)==null||(d=d.persistence)==null||d.unregister(AWe),(h=this._instance)==null||(h=h.persistence)==null||h.unregister(CWe))}))):this.stopRecording()}stopRecording(){var r,o,s,c;this.Ot&&this.ci&&(this.ci(),this.ci=void 0,this.Ot=!1,kn?.removeEventListener("beforeunload",this.oi),kn?.removeEventListener("offline",this.li),kn?.removeEventListener("online",this.hi),kn?.removeEventListener("visibilitychange",this.di),this.pi(),clearInterval(this.wi),(r=this.ii)==null||r.call(this),this.ii=void 0,(o=this.ni)==null||o.call(this),this.ni=void 0,(s=this.ei)==null||s.call(this),this.ei=void 0,(c=this.si)==null||c.call(this),this.si=void 0,this.Yt.stop(),this.Kt.stop(),this.Xt.stop(),U1.info("stopped"))}Si(){var r;(r=this._instance.persistence)==null||r.unregister(Hre)}$i(r){var o,s=this.Ct!==r,c=this.Wt;if(G1(c)){var d=this.jt,h=s||!wI(d),b=h?PLr(r,c):d;h&&(b?this.ki(JZe):U1.warn("Sample rate ("+c+") has determined that this sessionId ("+r+") will not be sent to the server."),this.ui("samplingDecisionMade",{sampleRate:c,isSampled:b})),(o=this._instance.persistence)==null||o.register({[Hre]:b})}else this.Si()}onRemoteConfig(r){var o,s,c,d;this.ui("$remote_config_received",r),this.xi(r),(o=r.sessionRecording)!=null&&o.endpoint&&(this.vi=(d=r.sessionRecording)==null?void 0:d.endpoint),this.mi(),((s=r.sessionRecording)==null?void 0:s.triggerMatchType)==="any"?(this.Vt=FLr,this.Lt=new Tun([this.Yt,this.Kt])):(this.Vt=BLr,this.Lt=new xun([this.Yt,this.Kt])),this._instance.register_for_session({$sdk_debug_replay_remote_trigger_matching_config:(c=r.sessionRecording)==null?void 0:c.triggerMatchType}),this.Kt.onRemoteConfig(r),this.Yt.onRemoteConfig(r),this.Xt.onRemoteConfig(r,(h,b)=>{this.ki("linked_flag_matched",{flag:h,variant:b})}),this.Jt=!0,this.startIfEnabledOrStop()}mi(){G1(this.Wt)&&Gu(this.si)&&(this.si=this.At.onSessionId(r=>{this.$i(r)}))}xi(r){if(this._instance.persistence){var o,s=this._instance.persistence,c=()=>{var d,h,b,y,E,k,x,O,P,D=(d=r.sessionRecording)==null?void 0:d.sampleRate,$=Gu(D)?null:parseFloat(D);Gu($)&&this.Si();var U=(h=r.sessionRecording)==null?void 0:h.minimumDurationMilliseconds;s.register({[xWe]:!!r.sessionRecording,[GFt]:(b=r.sessionRecording)==null?void 0:b.consoleLogRecordingEnabled,[qFt]:Ni({capturePerformance:r.capturePerformance},(y=r.sessionRecording)==null?void 0:y.networkPayloadCapture),[VFt]:(E=r.sessionRecording)==null?void 0:E.masking,[WFt]:{enabled:(k=r.sessionRecording)==null?void 0:k.recordCanvas,fps:(x=r.sessionRecording)==null?void 0:x.canvasFps,quality:(O=r.sessionRecording)==null?void 0:O.canvasQuality},[KFt]:$,[YFt]:mi(U)?null:U,[XFt]:(P=r.sessionRecording)==null?void 0:P.scriptConfig})};c(),(o=this.ri)==null||o.call(this),this.ri=this.At.onSessionId(c)}}log(r,o){var s;o===void 0&&(o="log"),(s=this._instance.sessionRecording)==null||s.onRRwebEmit({type:6,data:{plugin:"rrweb/console@1",payload:{level:o,trace:[],payload:[JSON.stringify(r)]}},timestamp:Date.now()})}gi(r){if(!mi(Object.assign)&&!mi(Array.from)&&!(this.Ot||this._instance.config.disable_session_recording||this._instance.consent.isOptedOut())){var o;this.Ot=!0,this.At.checkAndGetSessionAndWindowId(),L0e()?this.Ei():(o=La.__PosthogExtensions__)==null||o.loadExternalDependency==null||o.loadExternalDependency(this._instance,this.Ii,s=>{if(s)return U1.error("could not load recorder",s);this.Ei()}),U1.info("starting"),this.status===Bwe&&this.ki(r||"recording_initialized")}}get Ii(){var r;return((r=this._instance)==null||(r=r.persistence)==null||(r=r.get_property(XFt))==null?void 0:r.script)||"recorder"}Pi(r){var o;return r.type===3&&zLr.indexOf((o=r.data)==null?void 0:o.source)!==-1}Ri(r){var o=this.Pi(r);o||this.Zt||r.timestamp-this.ti>this.Ft&&(this.Zt=!0,clearInterval(this.wi),this.ui("sessionIdle",{eventTimestamp:r.timestamp,lastActivityTimestamp:this.ti,threshold:this.Ft,bufferLength:this.C.data.length,bufferSize:this.C.size}),this.ai());var s=!1;if(o&&(this.ti=r.timestamp,this.Zt)){var c=this.Zt==="unknown";this.Zt=!1,c||(this.ui("sessionNoLongerIdle",{reason:"user activity",type:r.type}),s=!0)}if(!this.Zt){var{windowId:d,sessionId:h}=this.At.checkAndGetSessionAndWindowId(!o,r.timestamp),b=this.Ct!==h,y=this.fi!==d;this.fi=d,this.Ct=h,b||y?(this.stopRecording(),this.startIfEnabledOrStop("session_id_changed")):s&&this.Ti()}}Mi(r){try{return r.rrwebMethod(),!0}catch(o){return this.Qt.length<10?this.Qt.push({enqueuedAt:r.enqueuedAt||Date.now(),attempt:r.attempt++,rrwebMethod:r.rrwebMethod}):U1.warn("could not emit queued rrweb event.",o,r),!1}}ui(r,o){return this.Mi(gBt(()=>L0e().addCustomEvent(r,o)))}Ci(){return this.Mi(gBt(()=>L0e().takeFullSnapshot()))}Ei(){var r,o,s,c,d={blockClass:"ph-no-capture",blockSelector:void 0,ignoreClass:"ph-ignore-input",maskTextClass:"ph-mask",maskTextSelector:void 0,maskTextFn:void 0,maskAllInputs:!0,maskInputOptions:{password:!0},maskInputFn:void 0,slimDOMOptions:{},collectFonts:!1,inlineStylesheet:!0,recordCrossOriginIframes:!1},h=this._instance.config.session_recording;for(var[b,y]of Object.entries(h||{}))b in d&&(b==="maskInputOptions"?d.maskInputOptions=Ni({password:!0},y):d[b]=y);this.qt&&this.qt.enabled&&(d.recordCanvas=!0,d.sampling={canvas:this.qt.fps},d.dataURLOptions={type:"image/webp",quality:this.qt.quality}),this.Ht&&(d.maskAllInputs=(o=this.Ht.maskAllInputs)===null||o===void 0||o,d.maskTextSelector=(s=this.Ht.maskTextSelector)!==null&&s!==void 0?s:void 0,d.blockSelector=(c=this.Ht.blockSelector)!==null&&c!==void 0?c:void 0);var E=L0e();if(E){this.Fi=(r=this.Fi)!==null&&r!==void 0?r:new _Lr(E,{refillRate:this._instance.config.session_recording.__mutationThrottlerRefillRate,bucketSize:this._instance.config.session_recording.__mutationThrottlerBucketSize,onBlockedNode:(x,O)=>{var P="Too many mutations on node '"+x+"'. Rate limiting. This could be due to SVG animations or something similar";U1.info(P,{node:O}),this.log(qre+" "+P,"warn")}});var k=this.Oi();this.ci=E(Ni({emit:x=>{this.onRRwebEmit(x)},plugins:k},d)),this.ti=Date.now(),this.Zt=wI(this.Zt)?this.Zt:"unknown",this.ui("$session_options",{sessionRecordingOptions:d,activePlugins:k.map(x=>x?.name)}),this.ui("$posthog_config",{config:this._instance.config})}else U1.error("onScriptLoaded was called but rrwebRecord is not available. This indicates something has gone wrong.")}Ti(){if(this.wi&&clearInterval(this.wi),this.Zt!==!0){var r=this.Dt;r&&(this.wi=setInterval(()=>{this.Ci()},r))}}Oi(){var r,o,s=[],c=(r=La.__PosthogExtensions__)==null||(r=r.rrwebPlugins)==null?void 0:r.getRecordConsolePlugin;c&&this.Ut&&s.push(c());var d=(o=La.__PosthogExtensions__)==null||(o=o.rrwebPlugins)==null?void 0:o.getRecordNetworkPlugin;return this.Bt&&j_(d)&&(!zPr.includes(location.hostname)||this._forceAllowLocalhostNetworkCapture?s.push(d(SLr(this._instance.config,this.Bt))):U1.info("NetworkCapture not started because we are on localhost.")),s}onRRwebEmit(r){var o;if(this.Ai(),r&&Ah(r)){if(r.type===A4.Meta){var s=this.yi(r.data.href);if(this.Di=s,!s)return;r.data.href=s}else this.Li();if(this.Kt.checkUrlTriggerConditions(()=>this.ji(),()=>this.Ni(),O=>this.zi(O)),!this.Kt.urlBlocked||(c=r).type===A4.Custom&&c.data.tag==="recording paused"){var c;r.type===A4.FullSnapshot&&this.Ti(),r.type===A4.FullSnapshot&&this.Jt&&this.Lt.triggerStatus(this.sessionId)===Y1&&this.pi();var d=this.Fi?this.Fi.throttleMutations(r):r;if(d){var h=function(O){var P=O;if(P&&Ah(P)&&P.type===6&&Ah(P.data)&&P.data.plugin==="rrweb/console@1"){P.data.payload.payload.length>10&&(P.data.payload.payload=P.data.payload.payload.slice(0,10),P.data.payload.payload.push("...[truncated]"));for(var D=[],$=0;$<P.data.payload.payload.length;$++)P.data.payload.payload[$]&&P.data.payload.payload[$].length>2e3?D.push(P.data.payload.payload[$].slice(0,2e3)+"...[truncated]"):D.push(P.data.payload.payload[$]);return P.data.payload.payload=D,O}return O}(d);if(this.Ri(h),this.Zt!==!0||mBt(h)){if(mBt(h)){var b=h.data.payload;if(b){var y=b.lastActivityTimestamp,E=b.threshold;h.timestamp=y+E}}var k=(o=this._instance.config.session_recording.compress_events)===null||o===void 0||o?function(O){if($we(O)<1024)return O;try{if(O.type===A4.FullSnapshot)return Ni({},O,{data:vB(O.data),cv:"2024-10"});if(O.type===A4.IncrementalSnapshot&&O.data.source===Cx.Mutation)return Ni({},O,{cv:"2024-10",data:Ni({},O.data,{texts:vB(O.data.texts),attributes:vB(O.data.attributes),removes:vB(O.data.removes),adds:vB(O.data.adds)})});if(O.type===A4.IncrementalSnapshot&&O.data.source===Cx.StyleSheetRule)return Ni({},O,{cv:"2024-10",data:Ni({},O.data,{adds:O.data.adds?vB(O.data.adds):void 0,removes:O.data.removes?vB(O.data.removes):void 0})})}catch(P){U1.error("could not compress event - will use uncompressed event",P)}return O}(h):h,x={$snapshot_bytes:$we(k),$snapshot_data:k,$session_id:this.Ct,$window_id:this.fi};this.status!==LM?this.Ui(x):this.pi()}}}}}Li(){if(!this._instance.config.capture_pageview&&kn){var r=this.yi(kn.location.href);this.Di!==r&&(this.ui("$url_changed",{href:r}),this.Di=r)}}Ai(){if(this.Qt.length){var r=[...this.Qt];this.Qt=[],r.forEach(o=>{Date.now()-o.enqueuedAt<=2e3&&this.Mi(o)})}}yi(r){var o=this._instance.config.session_recording;if(o.maskNetworkRequestFn){var s,c={url:r};return(s=c=o.maskNetworkRequestFn(c))==null?void 0:s.url}return r}pi(){return this.C={size:0,data:[],sessionId:this.Ct,windowId:this.fi},this.C}ai(){this.qi&&(clearTimeout(this.qi),this.qi=void 0);var r=this.Gt,o=this.Nt,s=G1(o)&&o>=0,c=G1(r)&&s&&o<r;return this.status===iU||this.status===ZZe||this.status===LM||c?(this.qi=setTimeout(()=>{this.ai()},2e3),this.C):(this.C.data.length>0&&jWe(this.C).forEach(d=>{this.Bi({$snapshot_bytes:d.size,$snapshot_data:d.data,$session_id:d.sessionId,$window_id:d.windowId,$lib:"web",$lib_version:I4.LIB_VERSION})}),this.pi())}Ui(r){var o,s=2+(((o=this.C)==null?void 0:o.data.length)||0);!this.Zt&&(this.C.size+r.$snapshot_bytes+s>943718.4||this.C.sessionId!==this.Ct)&&(this.C=this.ai()),this.C.size+=r.$snapshot_bytes,this.C.data.push(r.$snapshot_data),this.qi||this.Zt||(this.qi=setTimeout(()=>{this.ai()},2e3))}Bi(r){this._instance.capture("$snapshot",r,{_url:this._instance.requestRouter.endpointFor("api",this.vi),_noTruncate:!0,_batchKey:"recordings",skip_client_rate_limiting:!0})}zi(r){var o;this.Lt.triggerStatus(this.sessionId)===Y1&&((o=this._instance)==null||(o=o.persistence)==null||o.register({[r==="url"?CWe:AWe]:this.Ct}),this.ai(),this.ki(r+"_trigger_matched"))}ji(){this.Kt.urlBlocked||(this.Kt.urlBlocked=!0,clearInterval(this.wi),U1.info("recording paused due to URL blocker"),this.ui("recording paused",{reason:"url blocker"}))}Ni(){this.Kt.urlBlocked&&(this.Kt.urlBlocked=!1,this.Ci(),this.Ti(),this.ui("recording resumed",{reason:"left blocked url"}),U1.info("recording resumed"))}bi(){this.Yt.Tt.length!==0&&Gu(this.ni)&&(this.ni=this._instance.on("eventCaptured",r=>{try{this.Yt.Tt.includes(r.event)&&this.zi("event")}catch(o){U1.error("Could not activate event trigger",o)}}))}overrideLinkedFlag(){this.Xt.linkedFlagSeen=!0,this.Ci(),this.ki("linked_flag_overridden")}overrideSampling(){var r;(r=this._instance.persistence)==null||r.register({[Hre]:!0}),this.Ci(),this.ki("sampling_overridden")}overrideTrigger(r){this.zi(r)}ki(r,o){this._instance.register_for_session({$session_recording_start_reason:r}),U1.info(r.replace("_"," "),o),pc(["recording_initialized","session_id_changed"],r)||this.ui(r,o)}get sdkDebugProperties(){var{sessionStartTimestamp:r}=this.At.checkAndGetSessionAndWindowId(!0);return{$recording_status:this.status,$sdk_debug_replay_internal_buffer_length:this.C.data.length,$sdk_debug_replay_internal_buffer_size:this.C.size,$sdk_debug_current_session_duration:this.Nt,$sdk_debug_session_start:r}}}var oFe=Wm("[SegmentIntegration]");function GLr(n,r){var o=n.config.segment;if(!o)return r();(function(s,c){var d=s.config.segment;if(!d)return c();var h=y=>{var E=()=>y.anonymousId()||uM();s.config.get_device_id=E,y.id()&&(s.register({distinct_id:y.id(),$device_id:E()}),s.persistence.set_property(C4,"identified")),c()},b=d.user();"then"in b&&j_(b.then)?b.then(y=>h(y)):h(b)})(n,()=>{o.register((s=>{Promise&&Promise.resolve||oFe.warn("This browser does not have Promise support, and can not use the segment integration");var c=(d,h)=>{if(!h)return d;d.event.userId||d.event.anonymousId===s.get_distinct_id()||(oFe.info("No userId set, resetting PostHog"),s.reset()),d.event.userId&&d.event.userId!==s.get_distinct_id()&&(oFe.info("UserId set, identifying with PostHog"),s.identify(d.event.userId));var b=s.calculateEventProperties(h,d.event.properties);return d.event.properties=Object.assign({},b,d.event.properties),d};return{name:"PostHog JS",type:"enrichment",version:"1.0.0",isLoaded:()=>!0,load:()=>Promise.resolve(),track:d=>c(d,d.event.event),page:d=>c(d,"$pageview"),identify:d=>c(d,"$identify"),screen:d=>c(d,"$screen")}})(n)).then(()=>{r()})})}var Cun="posthog-js";function Aun(n,r){var{organization:o,projectId:s,prefix:c,severityAllowList:d=["error"]}=r===void 0?{}:r;return h=>{var b,y,E,k,x;if(!(d==="*"||d.includes(h.level))||!n.__loaded)return h;h.tags||(h.tags={});var O=n.requestRouter.endpointFor("ui","/project/"+n.config.token+"/person/"+n.get_distinct_id());h.tags["PostHog Person URL"]=O,n.sessionRecordingStarted()&&(h.tags["PostHog Recording URL"]=n.get_session_replay_url({withTimestamp:!0}));var P=((b=h.exception)==null?void 0:b.values)||[],D=P.map(U=>Ni({},U,{stacktrace:U.stacktrace?Ni({},U.stacktrace,{type:"raw",frames:(U.stacktrace.frames||[]).map(B=>Ni({},B,{platform:"web:javascript"}))}):void 0})),$={$exception_message:((y=P[0])==null?void 0:y.value)||h.message,$exception_type:(E=P[0])==null?void 0:E.type,$exception_personURL:O,$exception_level:h.level,$exception_list:D,$sentry_event_id:h.event_id,$sentry_exception:h.exception,$sentry_exception_message:((k=P[0])==null?void 0:k.value)||h.message,$sentry_exception_type:(x=P[0])==null?void 0:x.type,$sentry_tags:h.tags};return o&&s&&($.$sentry_url=(c||"https://sentry.io/organizations/")+o+"/issues/?project="+s+"&query="+h.event_id),n.exceptions.sendExceptionEvent($),h}}class qLr{constructor(r,o,s,c,d){this.name=Cun,this.setupOnce=function(h){h(Aun(r,{organization:o,projectId:s,prefix:c,severityAllowList:d}))}}}var VLr=kn!=null&&kn.location?Mwe(kn.location.hash,"__posthog")||Mwe(location.hash,"state"):null,bBt="_postHogToolbarParams",vBt=Wm("[Toolbar]"),QD=function(n){return n[n.UNINITIALIZED=0]="UNINITIALIZED",n[n.LOADING=1]="LOADING",n[n.LOADED=2]="LOADED",n}(QD||{});class WLr{constructor(r){this.instance=r}Hi(r){La.ph_toolbar_state=r}Wi(){var r;return(r=La.ph_toolbar_state)!==null&&r!==void 0?r:QD.UNINITIALIZED}maybeLoadToolbar(r,o,s){if(r===void 0&&(r=void 0),o===void 0&&(o=void 0),s===void 0&&(s=void 0),!kn||!Di)return!1;r=r??kn.location,s=s??kn.history;try{if(!o){try{kn.localStorage.setItem("test","test"),kn.localStorage.removeItem("test")}catch{return!1}o=kn?.localStorage}var c,d=VLr||Mwe(r.hash,"__posthog")||Mwe(r.hash,"state"),h=d?FFt(()=>JSON.parse(atob(decodeURIComponent(d))))||FFt(()=>JSON.parse(decodeURIComponent(d))):null;return h&&h.action==="ph_authorize"?((c=h).source="url",c&&Object.keys(c).length>0&&(h.desiredHash?r.hash=h.desiredHash:s?s.replaceState(s.state,"",r.pathname+r.search):r.hash="")):((c=JSON.parse(o.getItem(bBt)||"{}")).source="localstorage",delete c.userIntent),!(!c.token||this.instance.config.token!==c.token)&&(this.loadToolbar(c),!0)}catch{return!1}}Gi(r){var o=La.ph_load_toolbar||La.ph_load_editor;!Gu(o)&&j_(o)?o(r,this.instance):vBt.warn("No toolbar load function found")}loadToolbar(r){var o=!(Di==null||!Di.getElementById(Qcn));if(!kn||o)return!1;var s=this.instance.requestRouter.region==="custom"&&this.instance.config.advanced_disable_toolbar_metrics,c=Ni({token:this.instance.config.token},r,{apiURL:this.instance.requestRouter.endpointFor("ui")},s?{instrument:!1}:{});if(kn.localStorage.setItem(bBt,JSON.stringify(Ni({},c,{source:void 0}))),this.Wi()===QD.LOADED)this.Gi(c);else if(this.Wi()===QD.UNINITIALIZED){var d;this.Hi(QD.LOADING),(d=La.__PosthogExtensions__)==null||d.loadExternalDependency==null||d.loadExternalDependency(this.instance,"toolbar",h=>{if(h)return vBt.error("[Toolbar] Failed to load",h),void this.Hi(QD.UNINITIALIZED);this.Hi(QD.LOADED),this.Gi(c)}),th(kn,"turbolinks:load",()=>{this.Hi(QD.UNINITIALIZED),this.loadToolbar(c)})}return!0}Ji(r){return this.loadToolbar(r)}maybeLoadEditor(r,o,s){return r===void 0&&(r=void 0),o===void 0&&(o=void 0),s===void 0&&(s=void 0),this.maybeLoadToolbar(r,o,s)}}var KLr=Wm("[TracingHeaders]");class YLr{constructor(r){this.Vi=void 0,this.Ki=void 0,this.nt=()=>{var o,s;mi(this.Vi)&&((o=La.__PosthogExtensions__)==null||(o=o.tracingHeadersPatchFns)==null||o._patchXHR(this._instance.config.__add_tracing_headers||[],this._instance.get_distinct_id(),this._instance.sessionManager)),mi(this.Ki)&&((s=La.__PosthogExtensions__)==null||(s=s.tracingHeadersPatchFns)==null||s._patchFetch(this._instance.config.__add_tracing_headers||[],this._instance.get_distinct_id(),this._instance.sessionManager))},this._instance=r}J(r){var o,s;(o=La.__PosthogExtensions__)!=null&&o.tracingHeadersPatchFns&&r(),(s=La.__PosthogExtensions__)==null||s.loadExternalDependency==null||s.loadExternalDependency(this._instance,"tracing-headers",c=>{if(c)return KLr.error("failed to load script",c);r()})}startIfEnabledOrStop(){var r,o;this._instance.config.__add_tracing_headers?this.J(this.nt):((r=this.Vi)==null||r.call(this),(o=this.Ki)==null||o.call(this),this.Vi=void 0,this.Ki=void 0)}}var VD=Wm("[Web Vitals]"),wBt=9e5;class XLr{constructor(r){var o;this.Yi=!1,this.i=!1,this.C={url:void 0,metrics:[],firstMetricTimestamp:void 0},this.Xi=()=>{clearTimeout(this.Qi),this.C.metrics.length!==0&&(this._instance.capture("$web_vitals",this.C.metrics.reduce((s,c)=>Ni({},s,{["$web_vitals_"+c.name+"_event"]:Ni({},c),["$web_vitals_"+c.name+"_value"]:c.value}),{})),this.C={url:void 0,metrics:[],firstMetricTimestamp:void 0})},this.Zi=s=>{var c,d=(c=this._instance.sessionManager)==null?void 0:c.checkAndGetSessionAndWindowId(!0);if(mi(d))VD.error("Could not read session ID. Dropping metrics!");else{this.C=this.C||{url:void 0,metrics:[],firstMetricTimestamp:void 0};var h=this.te();mi(h)||(Gu(s?.name)||Gu(s?.value)?VD.error("Invalid metric received",s):this.ie&&s.value>=this.ie?VD.error("Ignoring metric with value >= "+this.ie,s):(this.C.url!==h&&(this.Xi(),this.Qi=setTimeout(this.Xi,this.flushToCaptureTimeoutMs)),mi(this.C.url)&&(this.C.url=h),this.C.firstMetricTimestamp=mi(this.C.firstMetricTimestamp)?Date.now():this.C.firstMetricTimestamp,s.attribution&&s.attribution.interactionTargetElement&&(s.attribution.interactionTargetElement=void 0),this.C.metrics.push(Ni({},s,{$current_url:h,$session_id:d.sessionId,$window_id:d.windowId,timestamp:Date.now()})),this.C.metrics.length===this.allowedMetrics.length&&this.Xi()))}},this.nt=()=>{var s,c,d,h,b=La.__PosthogExtensions__;mi(b)||mi(b.postHogWebVitalsCallbacks)||({onLCP:s,onCLS:c,onFCP:d,onINP:h}=b.postHogWebVitalsCallbacks),s&&c&&d&&h?(this.allowedMetrics.indexOf("LCP")>-1&&s(this.Zi.bind(this)),this.allowedMetrics.indexOf("CLS")>-1&&c(this.Zi.bind(this)),this.allowedMetrics.indexOf("FCP")>-1&&d(this.Zi.bind(this)),this.allowedMetrics.indexOf("INP")>-1&&h(this.Zi.bind(this)),this.i=!0):VD.error("web vitals callbacks not loaded - not starting")},this._instance=r,this.Yi=!((o=this._instance.persistence)==null||!o.props[zFt]),this.startIfEnabled()}get allowedMetrics(){var r,o,s=Ah(this._instance.config.capture_performance)?(r=this._instance.config.capture_performance)==null?void 0:r.web_vitals_allowed_metrics:void 0;return mi(s)?((o=this._instance.persistence)==null?void 0:o.props[HFt])||["CLS","FCP","INP","LCP"]:s}get flushToCaptureTimeoutMs(){return(Ah(this._instance.config.capture_performance)?this._instance.config.capture_performance.web_vitals_delayed_flush_ms:void 0)||5e3}get ie(){var r=Ah(this._instance.config.capture_performance)&&G1(this._instance.config.capture_performance.__web_vitals_max_value)?this._instance.config.capture_performance.__web_vitals_max_value:wBt;return 0<r&&r<=6e4?wBt:r}get isEnabled(){var r=Kw?.protocol;if(r!=="http:"&&r!=="https:")return VD.info("Web Vitals are disabled on non-http/https protocols"),!1;var o=Ah(this._instance.config.capture_performance)?this._instance.config.capture_performance.web_vitals:wI(this._instance.config.capture_performance)?this._instance.config.capture_performance:void 0;return wI(o)?o:this.Yi}startIfEnabled(){this.isEnabled&&!this.i&&(VD.info("enabled, starting..."),this.J(this.nt))}onRemoteConfig(r){var o=Ah(r.capturePerformance)&&!!r.capturePerformance.web_vitals,s=Ah(r.capturePerformance)?r.capturePerformance.web_vitals_allowed_metrics:void 0;this._instance.persistence&&(this._instance.persistence.register({[zFt]:o}),this._instance.persistence.register({[HFt]:s})),this.Yi=o,this.startIfEnabled()}J(r){var o,s;(o=La.__PosthogExtensions__)!=null&&o.postHogWebVitalsCallbacks&&r(),(s=La.__PosthogExtensions__)==null||s.loadExternalDependency==null||s.loadExternalDependency(this._instance,"web-vitals",c=>{c?VD.error("failed to load script",c):r()})}te(){var r=kn?kn.location.href:void 0;return r||VD.error("Could not determine current URL"),r}}var JLr=Wm("[Heatmaps]");function yBt(n){return Ah(n)&&"clientX"in n&&"clientY"in n&&G1(n.clientX)&&G1(n.clientY)}class ZLr{constructor(r){var o;this.rageclicks=new lun,this.Yi=!1,this.i=!1,this.ee=null,this.instance=r,this.Yi=!((o=this.instance.persistence)==null||!o.props[kWe])}get flushIntervalMilliseconds(){var r=5e3;return Ah(this.instance.config.capture_heatmaps)&&this.instance.config.capture_heatmaps.flush_interval_milliseconds&&(r=this.instance.config.capture_heatmaps.flush_interval_milliseconds),r}get isEnabled(){return mi(this.instance.config.capture_heatmaps)?mi(this.instance.config.enable_heatmaps)?this.Yi:this.instance.config.enable_heatmaps:this.instance.config.capture_heatmaps!==!1}startIfEnabled(){if(this.isEnabled){if(this.i)return;JLr.info("starting..."),this.re(),this.ee=setInterval(this.se.bind(this),this.flushIntervalMilliseconds)}else{var r,o;clearInterval((r=this.ee)!==null&&r!==void 0?r:void 0),(o=this.ne)==null||o.stop(),this.getAndClearBuffer()}}onRemoteConfig(r){var o=!!r.heatmaps;this.instance.persistence&&this.instance.persistence.register({[kWe]:o}),this.Yi=o,this.startIfEnabled()}getAndClearBuffer(){var r=this.C;return this.C=void 0,r}oe(r){this.ae(r.originalEvent,"deadclick")}re(){kn&&Di&&(th(kn,"beforeunload",this.se.bind(this)),th(Di,"click",r=>this.ae(r||kn?.event),{capture:!0}),th(Di,"mousemove",r=>this.le(r||kn?.event),{capture:!0}),this.ne=new fun(this.instance,nLr,this.oe.bind(this)),this.ne.startIfEnabled(),this.i=!0)}ue(r,o){var s=this.instance.scrollManager.scrollY(),c=this.instance.scrollManager.scrollX(),d=this.instance.scrollManager.scrollElement(),h=function(b,y,E){for(var k=b;k&&EEe(k)&&!wM(k,"body");){if(k===E)return!1;if(pc(y,kn?.getComputedStyle(k).position))return!0;k=iun(k)}return!1}(run(r),["fixed","sticky"],d);return{x:r.clientX+(h?0:c),y:r.clientY+(h?0:s),target_fixed:h,type:o}}ae(r,o){var s;if(o===void 0&&(o="click"),!JFt(r.target)&&yBt(r)){var c=this.ue(r,o);(s=this.rageclicks)!=null&&s.isRageClick(r.clientX,r.clientY,new Date().getTime())&&this.he(Ni({},c,{type:"rageclick"})),this.he(c)}}le(r){!JFt(r.target)&&yBt(r)&&(clearTimeout(this.de),this.de=setTimeout(()=>{this.he(this.ue(r,"mousemove"))},500))}he(r){if(kn){var o=kn.location.href;this.C=this.C||{},this.C[o]||(this.C[o]=[]),this.C[o].push(r)}}se(){this.C&&!fW(this.C)&&this.instance.capture("$$heatmap",{$heatmap_data:this.getAndClearBuffer()})}}class QLr{constructor(r){this._instance=r}doPageView(r,o){var s,c=this.ve(r,o);return this.ce={pathname:(s=kn?.location.pathname)!==null&&s!==void 0?s:"",pageViewId:o,timestamp:r},this._instance.scrollManager.resetContext(),c}doPageLeave(r){var o;return this.ve(r,(o=this.ce)==null?void 0:o.pageViewId)}doEvent(){var r;return{$pageview_id:(r=this.ce)==null?void 0:r.pageViewId}}ve(r,o){var s=this.ce;if(!s)return{$pageview_id:o};var c={$pageview_id:o,$prev_pageview_id:s.pageViewId},d=this._instance.scrollManager.getContext();if(d&&!this._instance.config.disable_scroll_properties){var{maxScrollHeight:h,lastScrollY:b,maxScrollY:y,maxContentHeight:E,lastContentY:k,maxContentY:x}=d;if(!(mi(h)||mi(b)||mi(y)||mi(E)||mi(k)||mi(x))){h=Math.ceil(h),b=Math.ceil(b),y=Math.ceil(y),E=Math.ceil(E),k=Math.ceil(k),x=Math.ceil(x);var O=h<=1?1:u2(b/h,0,1),P=h<=1?1:u2(y/h,0,1),D=E<=1?1:u2(k/E,0,1),$=E<=1?1:u2(x/E,0,1);c=pp(c,{$prev_pageview_last_scroll:b,$prev_pageview_last_scroll_percentage:O,$prev_pageview_max_scroll:y,$prev_pageview_max_scroll_percentage:P,$prev_pageview_last_content:k,$prev_pageview_last_content_percentage:D,$prev_pageview_max_content:x,$prev_pageview_max_content_percentage:$})}}return s.pathname&&(c.$prev_pageview_pathname=s.pathname),s.timestamp&&(c.$prev_pageview_duration=(r.getTime()-s.timestamp.getTime())/1e3),c}}var eDr=function(n){var r,o,s,c,d="";for(r=o=0,s=(n=(n+"").replace(/\r\n/g,`
`).replace(/\r/g,`
`)).length,c=0;c<s;c++){var h=n.charCodeAt(c),b=null;h<128?o++:b=h>127&&h<2048?String.fromCharCode(h>>6|192,63&h|128):String.fromCharCode(h>>12|224,h>>6&63|128,63&h|128),HM(b)||(o>r&&(d+=n.substring(r,o)),d+=b,r=o=c+1)}return o>r&&(d+=n.substring(r,n.length)),d},tDr=!!EWe||!!yWe,EBt="text/plain",Uwe=(n,r)=>{var[o,s]=n.split("?"),c=Ni({},r);s?.split("&").forEach(h=>{var[b]=h.split("=");delete c[b]});var d=HPr(c);return o+"?"+(d=d?(s?s+"&":"")+d:s)},Vre=(n,r)=>JSON.stringify(n,(o,s)=>typeof s=="bigint"?s.toString():s,r),aFe=n=>{var{data:r,compression:o}=n;if(r){if(o===N4.GZipJS){var s=_un(kun(Vre(r)),{mtime:0}),c=new Blob([s],{type:EBt});return{contentType:EBt,body:c,estimatedSize:c.size}}if(o===N4.Base64){var d=function(y){var E,k,x,O,P,D="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=",$=0,U=0,B="",G=[];if(!y)return y;y=eDr(y);do E=(P=y.charCodeAt($++)<<16|y.charCodeAt($++)<<8|y.charCodeAt($++))>>18&63,k=P>>12&63,x=P>>6&63,O=63&P,G[U++]=D.charAt(E)+D.charAt(k)+D.charAt(x)+D.charAt(O);while($<y.length);switch(B=G.join(""),y.length%3){case 1:B=B.slice(0,-2)+"==";break;case 2:B=B.slice(0,-1)+"="}return B}(Vre(r)),h=(y=>"data="+encodeURIComponent(typeof y=="string"?y:Vre(y)))(d);return{contentType:"application/x-www-form-urlencoded",body:h,estimatedSize:new Blob([h]).size}}var b=Vre(r);return{contentType:"application/json",body:b,estimatedSize:new Blob([b]).size}}},gie=[];yWe&&gie.push({transport:"fetch",method:n=>{var r,o,{contentType:s,body:c,estimatedSize:d}=(r=aFe(n))!==null&&r!==void 0?r:{},h=new Headers;Bl(n.headers,function(k,x){h.append(x,k)}),s&&h.append("Content-Type",s);var b=n.url,y=null;if(jFt){var E=new jFt;y={signal:E.signal,timeout:setTimeout(()=>E.abort(),n.timeout)}}yWe(b,Ni({method:n?.method||"GET",headers:h,keepalive:n.method==="POST"&&(d||0)<52428.8,body:c,signal:(o=y)==null?void 0:o.signal},n.fetchOptions)).then(k=>k.text().then(x=>{var O={statusCode:k.status,text:x};if(k.status===200)try{O.json=JSON.parse(x)}catch(P){zi.error(P)}n.callback==null||n.callback(O)})).catch(k=>{zi.error(k),n.callback==null||n.callback({statusCode:0,text:k})}).finally(()=>y?clearTimeout(y.timeout):null)}}),EWe&&gie.push({transport:"XHR",method:n=>{var r,o=new EWe;o.open(n.method||"GET",n.url,!0);var{contentType:s,body:c}=(r=aFe(n))!==null&&r!==void 0?r:{};Bl(n.headers,function(d,h){o.setRequestHeader(h,d)}),s&&o.setRequestHeader("Content-Type",s),n.timeout&&(o.timeout=n.timeout),o.withCredentials=!0,o.onreadystatechange=()=>{if(o.readyState===4){var d={statusCode:o.status,text:o.responseText};if(o.status===200)try{d.json=JSON.parse(o.responseText)}catch{}n.callback==null||n.callback(d)}},o.send(c)}}),c2!=null&&c2.sendBeacon&&gie.push({transport:"sendBeacon",method:n=>{var r=Uwe(n.url,{beacon:"1"});try{var o,{contentType:s,body:c}=(o=aFe(n))!==null&&o!==void 0?o:{},d=typeof c=="string"?new Blob([c],{type:s}):c;c2.sendBeacon(r,d)}catch{}}});var uK=function(n,r){if(!function(o){try{new RegExp(o)}catch{return!1}return!0}(r))return!1;try{return new RegExp(r).test(n)}catch{return!1}};function SBt(n,r,o){return Vre({distinct_id:n,userPropertiesToSet:r,userPropertiesToSetOnce:o})}var nDr={exact:(n,r)=>r.some(o=>n.some(s=>o===s)),is_not:(n,r)=>r.every(o=>n.every(s=>o!==s)),regex:(n,r)=>r.some(o=>n.some(s=>uK(o,s))),not_regex:(n,r)=>r.every(o=>n.every(s=>!uK(o,s))),icontains:(n,r)=>r.map(D0e).some(o=>n.map(D0e).some(s=>o.includes(s))),not_icontains:(n,r)=>r.map(D0e).every(o=>n.map(D0e).every(s=>!o.includes(s)))},D0e=n=>n.toLowerCase(),rDr=Wm("[Error tracking]");class iDr{constructor(r){var o,s;this.fe=[],this._instance=r,this.fe=(o=(s=this._instance.persistence)==null?void 0:s.get_property(TWe))!==null&&o!==void 0?o:[]}onRemoteConfig(r){var o,s,c=(o=(s=r.errorTracking)==null?void 0:s.suppressionRules)!==null&&o!==void 0?o:[];this.fe=c,this._instance.persistence&&this._instance.persistence.register({[TWe]:this.fe})}sendExceptionEvent(r){if(!this.pe(r))return this._instance.capture("$exception",r,{_noTruncate:!0,_batchKey:"exceptionEvent"});rDr.info("Skipping exception capture because a suppression rule matched")}pe(r){var o=r.$exception_list;if(!o||!kf(o)||o.length===0)return!1;var s=o.reduce((c,d)=>{var{type:h,value:b}=d;return Ih(h)&&h.length>0&&c.$exception_types.push(h),Ih(b)&&b.length>0&&c.$exception_values.push(b),c},{$exception_types:[],$exception_values:[]});return this.fe.some(c=>{var d=c.values.map(h=>{var b,y=nDr[h.operator],E=kf(h.value)?h.value:[h.value],k=(b=s[h.key])!==null&&b!==void 0?b:[];return E.length>0&&y(E,k)});return c.type==="OR"?d.some(Boolean):d.every(Boolean)})}}var F_="Mobile",zwe="iOS",yI="Android",aoe="Tablet",Iun=yI+" "+aoe,Oun="iPad",Nun="Apple",Run=Nun+" Watch",soe="Safari",lK="BlackBerry",Pun="Samsung",Lun=Pun+"Browser",Dun=Pun+" Internet",oU="Chrome",oDr=oU+" OS",Mun=oU+" "+zwe,eQe="Internet Explorer",jun=eQe+" "+F_,tQe="Opera",aDr=tQe+" Mini",nQe="Edge",$un="Microsoft "+nQe,MW="Firefox",Fun=MW+" "+zwe,coe="Nintendo",uoe="PlayStation",jW="Xbox",Bun=yI+" "+F_,Uun=F_+" "+soe,Wre="Windows",GWe=Wre+" Phone",_Bt="Nokia",qWe="Ouya",zun="Generic",sDr=zun+" "+F_.toLowerCase(),Hun=zun+" "+aoe.toLowerCase(),VWe="Konqueror",iv="(\\d+(\\.\\d+)?)",sFe=new RegExp("Version/"+iv),cDr=new RegExp(jW,"i"),uDr=new RegExp(uoe+" \\w+","i"),lDr=new RegExp(coe+" \\w+","i"),rQe=new RegExp(lK+"|PlayBook|BB10","i"),dDr={"NT3.51":"NT 3.11","NT4.0":"NT 4.0","5.0":"2000",5.1:"XP",5.2:"XP","6.0":"Vista",6.1:"7",6.2:"8",6.3:"8.1",6.4:"10","10.0":"10"},fDr=(n,r)=>r&&pc(r,Nun)||function(o){return pc(o,soe)&&!pc(o,oU)&&!pc(o,yI)}(n),Gun=function(n,r){return r=r||"",pc(n," OPR/")&&pc(n,"Mini")?aDr:pc(n," OPR/")?tQe:rQe.test(n)?lK:pc(n,"IE"+F_)||pc(n,"WPDesktop")?jun:pc(n,Lun)?Dun:pc(n,nQe)||pc(n,"Edg/")?$un:pc(n,"FBIOS")?"Facebook "+F_:pc(n,"UCWEB")||pc(n,"UCBrowser")?"UC Browser":pc(n,"CriOS")?Mun:pc(n,"CrMo")||pc(n,oU)?oU:pc(n,yI)&&pc(n,soe)?Bun:pc(n,"FxiOS")?Fun:pc(n.toLowerCase(),VWe.toLowerCase())?VWe:fDr(n,r)?pc(n,F_)?Uun:soe:pc(n,MW)?MW:pc(n,"MSIE")||pc(n,"Trident/")?eQe:pc(n,"Gecko")?MW:""},hDr={[jun]:[new RegExp("rv:"+iv)],[$un]:[new RegExp(nQe+"?\\/"+iv)],[oU]:[new RegExp("("+oU+"|CrMo)\\/"+iv)],[Mun]:[new RegExp("CriOS\\/"+iv)],"UC Browser":[new RegExp("(UCBrowser|UCWEB)\\/"+iv)],[soe]:[sFe],[Uun]:[sFe],[tQe]:[new RegExp("(Opera|OPR)\\/"+iv)],[MW]:[new RegExp(MW+"\\/"+iv)],[Fun]:[new RegExp("FxiOS\\/"+iv)],[VWe]:[new RegExp("Konqueror[:/]?"+iv,"i")],[lK]:[new RegExp(lK+" "+iv),sFe],[Bun]:[new RegExp("android\\s"+iv,"i")],[Dun]:[new RegExp(Lun+"\\/"+iv)],[eQe]:[new RegExp("(rv:|MSIE )"+iv)],Mozilla:[new RegExp("rv:"+iv)]},pDr=function(n,r){var o=Gun(n,r),s=hDr[o];if(mi(s))return null;for(var c=0;c<s.length;c++){var d=s[c],h=n.match(d);if(h)return parseFloat(h[h.length-2])}return null},kBt=[[new RegExp(jW+"; "+jW+" (.*?)[);]","i"),n=>[jW,n&&n[1]||""]],[new RegExp(coe,"i"),[coe,""]],[new RegExp(uoe,"i"),[uoe,""]],[rQe,[lK,""]],[new RegExp(Wre,"i"),(n,r)=>{if(/Phone/.test(r)||/WPDesktop/.test(r))return[GWe,""];if(new RegExp(F_).test(r)&&!/IEMobile\b/.test(r))return[Wre+" "+F_,""];var o=/Windows NT ([0-9.]+)/i.exec(r);if(o&&o[1]){var s=o[1],c=dDr[s]||"";return/arm/i.test(r)&&(c="RT"),[Wre,c]}return[Wre,""]}],[/((iPhone|iPad|iPod).*?OS (\d+)_(\d+)_?(\d+)?|iPhone)/,n=>{if(n&&n[3]){var r=[n[3],n[4],n[5]||"0"];return[zwe,r.join(".")]}return[zwe,""]}],[/(watch.*\/(\d+\.\d+\.\d+)|watch os,(\d+\.\d+),)/i,n=>{var r="";return n&&n.length>=3&&(r=mi(n[2])?n[3]:n[2]),["watchOS",r]}],[new RegExp("("+yI+" (\\d+)\\.(\\d+)\\.?(\\d+)?|"+yI+")","i"),n=>{if(n&&n[2]){var r=[n[2],n[3],n[4]||"0"];return[yI,r.join(".")]}return[yI,""]}],[/Mac OS X (\d+)[_.](\d+)[_.]?(\d+)?/i,n=>{var r=["Mac OS X",""];if(n&&n[1]){var o=[n[1],n[2],n[3]||"0"];r[1]=o.join(".")}return r}],[/Mac/i,["Mac OS X",""]],[/CrOS/,[oDr,""]],[/Linux|debian/i,["Linux",""]]],TBt=function(n){return lDr.test(n)?coe:uDr.test(n)?uoe:cDr.test(n)?jW:new RegExp(qWe,"i").test(n)?qWe:new RegExp("("+GWe+"|WPDesktop)","i").test(n)?GWe:/iPad/.test(n)?Oun:/iPod/.test(n)?"iPod Touch":/iPhone/.test(n)?"iPhone":/(watch)(?: ?os[,/]|\d,\d\/)[\d.]+/i.test(n)?Run:rQe.test(n)?lK:/(kobo)\s(ereader|touch)/i.test(n)?"Kobo":new RegExp(_Bt,"i").test(n)?_Bt:/(kf[a-z]{2}wi|aeo[c-r]{2})( bui|\))/i.test(n)||/(kf[a-z]+)( bui|\)).+silk\//i.test(n)?"Kindle Fire":/(Android|ZTE)/i.test(n)?!new RegExp(F_).test(n)||/(9138B|TB782B|Nexus [97]|pixel c|HUAWEISHT|BTV|noble nook|smart ultra 6)/i.test(n)?/pixel[\daxl ]{1,6}/i.test(n)&&!/pixel c/i.test(n)||/(huaweimed-al00|tah-|APA|SM-G92|i980|zte|U304AA)/i.test(n)||/lmy47v/i.test(n)&&!/QTAQZ3/i.test(n)?yI:Iun:yI:new RegExp("(pda|"+F_+")","i").test(n)?sDr:new RegExp(aoe,"i").test(n)&&!new RegExp(aoe+" pc","i").test(n)?Hun:""},M0e="https?://(.*)",_Ee=["gclid","gclsrc","dclid","gbraid","wbraid","fbclid","msclkid","twclid","li_fat_id","igshid","ttclid","rdt_cid","epik","qclid","sccid","irclid","_kx"],gDr=yEe(["utm_source","utm_medium","utm_campaign","utm_content","utm_term","gad_source","mc_cid"],_Ee),iQe="<masked>",mDr=["li_fat_id"];function qun(n,r,o){if(!Di)return{};var s,c=r?yEe([],_Ee,o||[]):[],d=Vun(WZe(Di.URL,c,iQe),n),h=(s={},Bl(mDr,function(b){var y=AI.D(b);s[b]=y||null}),s);return pp(h,d)}function Vun(n,r){var o=gDr.concat(r||[]),s={};return Bl(o,function(c){var d=Dwe(n,c);s[c]=d||null}),s}function Wun(n){var r=function(d){return d?d.search(M0e+"google.([^/?]*)")===0?"google":d.search(M0e+"bing.com")===0?"bing":d.search(M0e+"yahoo.com")===0?"yahoo":d.search(M0e+"duckduckgo.com")===0?"duckduckgo":null:null}(n),o=r!="yahoo"?"q":"p",s={};if(!HM(r)){s.$search_engine=r;var c=Di?Dwe(Di.referrer,o):"";c.length&&(s.ph_keyword=c)}return s}function xBt(){return navigator.language||navigator.userLanguage}function Kun(){return Di?.referrer||"$direct"}function Yun(n,r){var o=n?yEe([],_Ee,r||[]):[],s=Kw?.href.substring(0,1e3);return{r:Kun().substring(0,1e3),u:s?WZe(s,o,iQe):void 0}}function Xun(n){var r,{r:o,u:s}=n,c={$referrer:o,$referring_domain:o==null?void 0:o=="$direct"?"$direct":(r=sK(o))==null?void 0:r.host};if(s){c.$current_url=s;var d=sK(s);c.$host=d?.host,c.$pathname=d?.pathname;var h=Vun(s);pp(c,h)}if(o){var b=Wun(o);pp(c,b)}return c}function Jun(){try{return Intl.DateTimeFormat().resolvedOptions().timeZone}catch{return}}function bDr(){try{return new Date().getTimezoneOffset()}catch{return}}function vDr(n,r){if(!Gw)return{};var o,s,c,d=n?yEe([],_Ee,r||[]):[],[h,b]=function(y){for(var E=0;E<kBt.length;E++){var[k,x]=kBt[E],O=k.exec(y),P=O&&(j_(x)?x(O,y):x);if(P)return P}return["",""]}(Gw);return pp(qZe({$os:h,$os_version:b,$browser:Gun(Gw,navigator.vendor),$device:TBt(Gw),$device_type:(s=Gw,c=TBt(s),c===Oun||c===Iun||c==="Kobo"||c==="Kindle Fire"||c===Hun?aoe:c===coe||c===jW||c===uoe||c===qWe?"Console":c===Run?"Wearable":c?F_:"Desktop"),$timezone:Jun(),$timezone_offset:bDr()}),{$current_url:WZe(Kw?.href,d,iQe),$host:Kw?.host,$pathname:Kw?.pathname,$raw_user_agent:Gw.length>1e3?Gw.substring(0,997)+"...":Gw,$browser_version:pDr(Gw,navigator.vendor),$browser_language:xBt(),$browser_language_prefix:(o=xBt(),typeof o=="string"?o.split("-")[0]:void 0),$screen_height:kn?.screen.height,$screen_width:kn?.screen.width,$viewport_height:kn?.innerHeight,$viewport_width:kn?.innerWidth,$lib:"web",$lib_version:I4.LIB_VERSION,$insert_id:Math.random().toString(36).substring(2,10)+Math.random().toString(36).substring(2,10),$time:Date.now()/1e3})}var iI=Wm("[FeatureFlags]"),cFe="$active_feature_flags",YV="$override_feature_flags",CBt="$feature_flag_payloads",Sre="$override_feature_flag_payloads",ABt="$feature_flag_request_id",IBt=n=>{var r={};for(var[o,s]of Sve(n||{}))s&&(r[o]=s);return r},wDr=n=>{var r=n.flags;return r?(n.featureFlags=Object.fromEntries(Object.keys(r).map(o=>{var s;return[o,(s=r[o].variant)!==null&&s!==void 0?s:r[o].enabled]})),n.featureFlagPayloads=Object.fromEntries(Object.keys(r).filter(o=>r[o].enabled).filter(o=>{var s;return(s=r[o].metadata)==null?void 0:s.payload}).map(o=>{var s;return[o,(s=r[o].metadata)==null?void 0:s.payload]}))):iI.warn("Using an older version of the feature flags endpoint. Please upgrade your PostHog server to the latest version"),n},yDr=function(n){return n.FeatureFlags="feature_flags",n.Recordings="recordings",n}({});class EDr{constructor(r){this.ge=!1,this._e=!1,this.me=!1,this.be=!1,this.ye=!1,this.we=!1,this.Se=!1,this._instance=r,this.featureFlagEventHandlers=[]}flags(){if(this._instance.config.__preview_remote_config)this.we=!0;else{var r=!this.$e&&(this._instance.config.advanced_disable_feature_flags||this._instance.config.advanced_disable_feature_flags_on_first_load);this.ke({disableFlags:r})}}get hasLoadedFlags(){return this._e}getFlags(){return Object.keys(this.getFlagVariants())}getFlagsWithDetails(){var r=this._instance.get_property(IWe),o=this._instance.get_property(YV),s=this._instance.get_property(Sre);if(!s&&!o)return r||{};var c=pp({},r||{}),d=[...new Set([...Object.keys(s||{}),...Object.keys(o||{})])];for(var h of d){var b,y,E=c[h],k=o?.[h],x=mi(k)?(b=E?.enabled)!==null&&b!==void 0&&b:!!k,O=mi(k)?E.variant:typeof k=="string"?k:void 0,P=s?.[h],D=Ni({},E,{enabled:x,variant:x?O??E?.variant:void 0});x!==E?.enabled&&(D.original_enabled=E?.enabled),O!==E?.variant&&(D.original_variant=E?.variant),P&&(D.metadata=Ni({},E?.metadata,{payload:P,original_payload:E==null||(y=E.metadata)==null?void 0:y.payload})),c[h]=D}return this.ge||(iI.warn(" Overriding feature flag details!",{flagDetails:r,overriddenPayloads:s,finalDetails:c}),this.ge=!0),c}getFlagVariants(){var r=this._instance.get_property(hW),o=this._instance.get_property(YV);if(!o)return r||{};for(var s=pp({},r),c=Object.keys(o),d=0;d<c.length;d++)s[c[d]]=o[c[d]];return this.ge||(iI.warn(" Overriding feature flags!",{enabledFlags:r,overriddenFlags:o,finalFlags:s}),this.ge=!0),s}getFlagPayloads(){var r=this._instance.get_property(CBt),o=this._instance.get_property(Sre);if(!o)return r||{};for(var s=pp({},r||{}),c=Object.keys(o),d=0;d<c.length;d++)s[c[d]]=o[c[d]];return this.ge||(iI.warn(" Overriding feature flag payloads!",{flagPayloads:r,overriddenPayloads:o,finalPayloads:s}),this.ge=!0),s}reloadFeatureFlags(){this.be||this._instance.config.advanced_disable_feature_flags||this.$e||(this.$e=setTimeout(()=>{this.ke()},5))}xe(){clearTimeout(this.$e),this.$e=void 0}ensureFlagsLoaded(){this._e||this.me||this.$e||this.reloadFeatureFlags()}setAnonymousDistinctId(r){this.$anon_distinct_id=r}setReloadingPaused(r){this.be=r}ke(r){var o;if(this.xe(),!this._instance.I())if(this.me)this.ye=!0;else{var s={token:this._instance.config.token,distinct_id:this._instance.get_distinct_id(),groups:this._instance.getGroups(),$anon_distinct_id:this.$anon_distinct_id,person_properties:Ni({},((o=this._instance.persistence)==null?void 0:o.get_initial_props())||{},this._instance.get_property(Gre)||{}),group_properties:this._instance.get_property(kB)};(r!=null&&r.disableFlags||this._instance.config.advanced_disable_feature_flags)&&(s.disable_flags=!0);var c=this._instance.config.__preview_remote_config,d=c?"/flags/?v=2":"/flags/?v=2&config=true",h=this._instance.config.advanced_only_evaluate_survey_feature_flags?"&only_evaluate_survey_feature_flags=true":"",b=this._instance.requestRouter.endpointFor("api",d+h);c&&(s.timezone=Jun()),this.me=!0,this._instance.Ee({method:"POST",url:b,data:s,compression:this._instance.config.disable_compression?void 0:N4.Base64,timeout:this._instance.config.feature_flag_request_timeout_ms,callback:y=>{var E,k,x=!0;if(y.statusCode===200&&(this.ye||(this.$anon_distinct_id=void 0),x=!1),this.me=!1,this.we||(this.we=!0,this._instance.Ie((k=y.json)!==null&&k!==void 0?k:{})),!s.disable_flags||this.ye)if(this.Se=!x,y.json&&(E=y.json.quotaLimited)!=null&&E.includes(yDr.FeatureFlags))iI.warn("You have hit your feature flags quota limit, and will not be able to load feature flags until the quota is reset.  Please visit https://posthog.com/docs/billing/limits-alerts to learn more.");else{var O;s.disable_flags||this.receivedFeatureFlags((O=y.json)!==null&&O!==void 0?O:{},x),this.ye&&(this.ye=!1,this.ke())}}})}}getFeatureFlag(r,o){if(o===void 0&&(o={}),this._e||this.getFlags()&&this.getFlags().length>0){var s=this.getFlagVariants()[r],c=""+s,d=this._instance.get_property(ABt)||void 0,h=this._instance.get_property(Nwe)||{};if((o.send_event||!("send_event"in o))&&(!(r in h)||!h[r].includes(c))){var b,y,E,k,x,O,P,D,$;kf(h[r])?h[r].push(c):h[r]=[c],(b=this._instance.persistence)==null||b.register({[Nwe]:h});var U=this.getFeatureFlagDetails(r),B={$feature_flag:r,$feature_flag_response:s,$feature_flag_payload:this.getFeatureFlagPayload(r)||null,$feature_flag_request_id:d,$feature_flag_bootstrapped_response:((y=this._instance.config.bootstrap)==null||(y=y.featureFlags)==null?void 0:y[r])||null,$feature_flag_bootstrapped_payload:((E=this._instance.config.bootstrap)==null||(E=E.featureFlagPayloads)==null?void 0:E[r])||null,$used_bootstrap_value:!this.Se};mi(U==null||(k=U.metadata)==null?void 0:k.version)||(B.$feature_flag_version=U.metadata.version);var G,K=(x=U==null||(O=U.reason)==null?void 0:O.description)!==null&&x!==void 0?x:U==null||(P=U.reason)==null?void 0:P.code;K&&(B.$feature_flag_reason=K),U!=null&&(D=U.metadata)!=null&&D.id&&(B.$feature_flag_id=U.metadata.id),mi(U?.original_variant)&&mi(U?.original_enabled)||(B.$feature_flag_original_response=mi(U.original_variant)?U.original_enabled:U.original_variant),U!=null&&($=U.metadata)!=null&&$.original_payload&&(B.$feature_flag_original_payload=U==null||(G=U.metadata)==null?void 0:G.original_payload),this._instance.capture("$feature_flag_called",B)}return s}iI.warn('getFeatureFlag for key "'+r+`" failed. Feature flags didn't load in time.`)}getFeatureFlagDetails(r){return this.getFlagsWithDetails()[r]}getFeatureFlagPayload(r){return this.getFlagPayloads()[r]}getRemoteConfigPayload(r,o){var s=this._instance.config.token;this._instance.Ee({method:"POST",url:this._instance.requestRouter.endpointFor("api","/flags/?v=2&config=true"),data:{distinct_id:this._instance.get_distinct_id(),token:s},compression:this._instance.config.disable_compression?void 0:N4.Base64,timeout:this._instance.config.feature_flag_request_timeout_ms,callback:c=>{var d,h=(d=c.json)==null?void 0:d.featureFlagPayloads;o(h?.[r]||void 0)}})}isFeatureEnabled(r,o){if(o===void 0&&(o={}),this._e||this.getFlags()&&this.getFlags().length>0)return!!this.getFeatureFlag(r,o);iI.warn('isFeatureEnabled for key "'+r+`" failed. Feature flags didn't load in time.`)}addFeatureFlagsHandler(r){this.featureFlagEventHandlers.push(r)}removeFeatureFlagsHandler(r){this.featureFlagEventHandlers=this.featureFlagEventHandlers.filter(o=>o!==r)}receivedFeatureFlags(r,o){if(this._instance.persistence){this._e=!0;var s=this.getFlagVariants(),c=this.getFlagPayloads(),d=this.getFlagsWithDetails();(function(h,b,y,E,k){y===void 0&&(y={}),E===void 0&&(E={}),k===void 0&&(k={});var x=wDr(h),O=x.flags,P=x.featureFlags,D=x.featureFlagPayloads;if(P){var $=h.requestId;if(kf(P)){iI.warn("v1 of the feature flags endpoint is deprecated. Please use the latest version.");var U={};if(P)for(var B=0;B<P.length;B++)U[P[B]]=!0;b&&b.register({[cFe]:P,[hW]:U})}else{var G=P,K=D,X=O;h.errorsWhileComputingFlags&&(G=Ni({},y,G),K=Ni({},E,K),X=Ni({},k,X)),b&&b.register(Ni({[cFe]:Object.keys(IBt(G)),[hW]:G||{},[CBt]:K||{},[IWe]:X||{}},$?{[ABt]:$}:{}))}}})(r,this._instance.persistence,s,c,d),this.Pe(o)}}override(r,o){o===void 0&&(o=!1),iI.warn("override is deprecated. Please use overrideFeatureFlags instead."),this.overrideFeatureFlags({flags:r,suppressWarning:o})}overrideFeatureFlags(r){if(!this._instance.__loaded||!this._instance.persistence)return iI.uninitializedWarning("posthog.featureFlags.overrideFeatureFlags");if(r===!1)return this._instance.persistence.unregister(YV),this._instance.persistence.unregister(Sre),void this.Pe();if(r&&typeof r=="object"&&("flags"in r||"payloads"in r)){var o,s=r;if(this.ge=!!((o=s.suppressWarning)!==null&&o!==void 0&&o),"flags"in s){if(s.flags===!1)this._instance.persistence.unregister(YV);else if(s.flags)if(kf(s.flags)){for(var c={},d=0;d<s.flags.length;d++)c[s.flags[d]]=!0;this._instance.persistence.register({[YV]:c})}else this._instance.persistence.register({[YV]:s.flags})}return"payloads"in s&&(s.payloads===!1?this._instance.persistence.unregister(Sre):s.payloads&&this._instance.persistence.register({[Sre]:s.payloads})),void this.Pe()}this.Pe()}onFeatureFlags(r){if(this.addFeatureFlagsHandler(r),this._e){var{flags:o,flagVariants:s}=this.Re();r(o,s)}return()=>this.removeFeatureFlagsHandler(r)}updateEarlyAccessFeatureEnrollment(r,o){var s,c=(this._instance.get_property(_ve)||[]).find(y=>y.flagKey===r),d={["$feature_enrollment/"+r]:o},h={$feature_flag:r,$feature_enrollment:o,$set:d};c&&(h.$early_access_feature_name=c.name),this._instance.capture("$feature_enrollment_update",h),this.setPersonPropertiesForFlags(d,!1);var b=Ni({},this.getFlagVariants(),{[r]:o});(s=this._instance.persistence)==null||s.register({[cFe]:Object.keys(IBt(b)),[hW]:b}),this.Pe()}getEarlyAccessFeatures(r,o,s){o===void 0&&(o=!1);var c=this._instance.get_property(_ve),d=s?"&"+s.map(h=>"stage="+h).join("&"):"";if(c&&!o)return r(c);this._instance.Ee({url:this._instance.requestRouter.endpointFor("api","/api/early_access_features/?token="+this._instance.config.token+d),method:"GET",callback:h=>{var b;if(h.json){var y=h.json.earlyAccessFeatures;return(b=this._instance.persistence)==null||b.register({[_ve]:y}),r(y)}}})}Re(){var r=this.getFlags(),o=this.getFlagVariants();return{flags:r.filter(s=>o[s]),flagVariants:Object.keys(o).filter(s=>o[s]).reduce((s,c)=>(s[c]=o[c],s),{})}}Pe(r){var{flags:o,flagVariants:s}=this.Re();this.featureFlagEventHandlers.forEach(c=>c(o,s,{errorsLoading:r}))}setPersonPropertiesForFlags(r,o){o===void 0&&(o=!0);var s=this._instance.get_property(Gre)||{};this._instance.register({[Gre]:Ni({},s,r)}),o&&this._instance.reloadFeatureFlags()}resetPersonPropertiesForFlags(){this._instance.unregister(Gre)}setGroupPropertiesForFlags(r,o){o===void 0&&(o=!0);var s=this._instance.get_property(kB)||{};Object.keys(s).length!==0&&Object.keys(s).forEach(c=>{s[c]=Ni({},s[c],r[c]),delete r[c]}),this._instance.register({[kB]:Ni({},s,r)}),o&&this._instance.reloadFeatureFlags()}resetGroupPropertiesForFlags(r){if(r){var o=this._instance.get_property(kB)||{};this._instance.register({[kB]:Ni({},o,{[r]:{}})})}else this._instance.unregister(kB)}reset(){this._e=!1,this.me=!1,this.be=!1,this.ye=!1,this.we=!1,this.Se=!1,this.$anon_distinct_id=void 0,this.xe(),this.ge=!1}}var SDr=["cookie","localstorage","localstorage+cookie","sessionstorage","memory"];class uFe{constructor(r,o){this.S=r,this.props={},this.Te=!1,this.Me=(s=>{var c="";return s.token&&(c=s.token.replace(/\+/g,"PL").replace(/\//g,"SL").replace(/=/g,"EQ")),s.persistence_name?"ph_"+s.persistence_name:"ph_"+c+"_posthog"})(r),this.B=this.Ce(r),this.load(),r.debug&&zi.info("Persistence loaded",r.persistence,Ni({},this.props)),this.update_config(r,r,o),this.save()}isDisabled(){return!!this.Fe}Ce(r){SDr.indexOf(r.persistence.toLowerCase())===-1&&(zi.critical("Unknown persistence type "+r.persistence+"; falling back to localStorage+cookie"),r.persistence="localStorage+cookie");var o=r.persistence.toLowerCase();return o==="localstorage"&&Gg.O()?Gg:o==="localstorage+cookie"&&A0e.O()?A0e:o==="sessionstorage"&&Ub.O()?Ub:o==="memory"?eLr:o==="cookie"?AI:A0e.O()?A0e:AI}properties(){var r={};return Bl(this.props,function(o,s){if(s===hW&&Ah(o))for(var c=Object.keys(o),d=0;d<c.length;d++)r["$feature/"+c[d]]=o[c[d]];else b=s,y=!1,(HM(h=LPr)?y:MFt&&h.indexOf===MFt?h.indexOf(b)!=-1:(Bl(h,function(E){if(y||(y=E===b))return Iwe}),y))||(r[s]=o);var h,b,y}),r}load(){if(!this.Fe){var r=this.B.L(this.Me);r&&(this.props=pp({},r))}}save(){this.Fe||this.B.j(this.Me,this.props,this.Oe,this.Ae,this.De,this.S.debug)}remove(){this.B.N(this.Me,!1),this.B.N(this.Me,!0)}clear(){this.remove(),this.props={}}register_once(r,o,s){if(Ah(r)){mi(o)&&(o="None"),this.Oe=mi(s)?this.Le:s;var c=!1;if(Bl(r,(d,h)=>{this.props.hasOwnProperty(h)&&this.props[h]!==o||(this.props[h]=d,c=!0)}),c)return this.save(),!0}return!1}register(r,o){if(Ah(r)){this.Oe=mi(o)?this.Le:o;var s=!1;if(Bl(r,(c,d)=>{r.hasOwnProperty(d)&&this.props[d]!==c&&(this.props[d]=c,s=!0)}),s)return this.save(),!0}return!1}unregister(r){r in this.props&&(delete this.props[r],this.save())}update_campaign_params(){if(!this.Te){var r=qun(this.S.custom_campaign_params,this.S.mask_personal_data_properties,this.S.custom_personal_data_properties);fW(qZe(r))||this.register(r),this.Te=!0}}update_search_keyword(){var r;this.register((r=Di?.referrer)?Wun(r):{})}update_referrer_info(){var r;this.register_once({$referrer:Kun(),$referring_domain:Di!=null&&Di.referrer&&((r=sK(Di.referrer))==null?void 0:r.host)||"$direct"},void 0)}set_initial_person_info(){this.props[PWe]||this.props[LWe]||this.register_once({[Rwe]:Yun(this.S.mask_personal_data_properties,this.S.custom_personal_data_properties)},void 0)}get_initial_props(){var r={};Bl([LWe,PWe],h=>{var b=this.props[h];b&&Bl(b,function(y,E){r["$initial_"+SWe(E)]=y})});var o,s,c=this.props[Rwe];if(c){var d=(o=Xun(c),s={},Bl(o,function(h,b){s["$initial_"+SWe(b)]=h}),s);pp(r,d)}return r}safe_merge(r){return Bl(this.props,function(o,s){s in r||(r[s]=o)}),r}update_config(r,o,s){if(this.Le=this.Oe=r.cookie_expiration,this.set_disabled(r.disable_persistence||!!s),this.set_cross_subdomain(r.cross_subdomain_cookie),this.set_secure(r.secure_cookie),r.persistence!==o.persistence){var c=this.Ce(r),d=this.props;this.clear(),this.B=c,this.props=d,this.save()}}set_disabled(r){this.Fe=r,this.Fe?this.remove():this.save()}set_cross_subdomain(r){r!==this.Ae&&(this.Ae=r,this.remove(),this.save())}set_secure(r){r!==this.De&&(this.De=r,this.remove(),this.save())}set_event_timer(r,o){var s=this.props[zre]||{};s[r]=o,this.props[zre]=s,this.save()}remove_event_timer(r){var o=(this.props[zre]||{})[r];return mi(o)||(delete this.props[zre][r],this.save()),o}get_property(r){return this.props[r]}set_property(r,o){this.props[r]=o,this.save()}}class Zun{constructor(){this.je={},this.je={}}on(r,o){return this.je[r]||(this.je[r]=[]),this.je[r].push(o),()=>{this.je[r]=this.je[r].filter(s=>s!==o)}}emit(r,o){for(var s of this.je[r]||[])s(o);for(var c of this.je["*"]||[])c(r,o)}}class xB{constructor(r){this.Ne=new Zun,this.ze=(o,s)=>this.Ue(o,s)&&this.qe(o,s)&&this.Be(o,s),this.Ue=(o,s)=>s==null||!s.event||o?.event===s?.event,this._instance=r,this.He=new Set,this.We=new Set}init(){var r;if(!mi((r=this._instance)==null?void 0:r.Ge)){var o;(o=this._instance)==null||o.Ge((s,c)=>{this.on(s,c)})}}register(r){var o,s;if(!mi((o=this._instance)==null?void 0:o.Ge)&&(r.forEach(h=>{var b,y;(b=this.We)==null||b.add(h),(y=h.steps)==null||y.forEach(E=>{var k;(k=this.He)==null||k.add(E?.event||"")})}),(s=this._instance)!=null&&s.autocapture)){var c,d=new Set;r.forEach(h=>{var b;(b=h.steps)==null||b.forEach(y=>{y!=null&&y.selector&&d.add(y?.selector)})}),(c=this._instance)==null||c.autocapture.setElementSelectors(d)}}on(r,o){var s;o!=null&&r.length!=0&&(this.He.has(r)||this.He.has(o?.event))&&this.We&&((s=this.We)==null?void 0:s.size)>0&&this.We.forEach(c=>{this.Je(o,c)&&this.Ne.emit("actionCaptured",c.name)})}Ve(r){this.onAction("actionCaptured",o=>r(o))}Je(r,o){if(o?.steps==null)return!1;for(var s of o.steps)if(this.ze(r,s))return!0;return!1}onAction(r,o){return this.Ne.on(r,o)}qe(r,o){if(o!=null&&o.url){var s,c=r==null||(s=r.properties)==null?void 0:s.$current_url;if(!c||typeof c!="string"||!xB.Ke(c,o?.url,o?.url_matching||"contains"))return!1}return!0}static Ke(r,o,s){switch(s){case"regex":return!!kn&&uK(r,o);case"exact":return o===r;case"contains":var c=xB.Ye(o).replace(/_/g,".").replace(/%/g,".*");return uK(r,c);default:return!1}}static Ye(r){return r.replace(/[|\\{}()[\]^$+*?.]/g,"\\$&").replace(/-/g,"\\x2d")}Be(r,o){if((o!=null&&o.href||o!=null&&o.tag_name||o!=null&&o.text)&&!this.Xe(r).some(d=>!(o!=null&&o.href&&!xB.Ke(d.href||"",o?.href,o?.href_matching||"exact"))&&(o==null||!o.tag_name||d.tag_name===o?.tag_name)&&!(o!=null&&o.text&&!xB.Ke(d.text||"",o?.text,o?.text_matching||"exact")&&!xB.Ke(d.$el_text||"",o?.text,o?.text_matching||"exact"))))return!1;if(o!=null&&o.selector){var s,c=r==null||(s=r.properties)==null?void 0:s.$element_selectors;if(!c||!c.includes(o?.selector))return!1}return!0}Xe(r){return r?.properties.$elements==null?[]:r?.properties.$elements}}var $m=Wm("[Surveys]"),WWe="seenSurvey_",_Dr=(n,r)=>{var o="$survey_"+r+"/"+n.id;return n.current_iteration&&n.current_iteration>0&&(o="$survey_"+r+"/"+n.id+"/"+n.current_iteration),o};class kDr{constructor(r){this._instance=r,this.Qe=new Map,this.Ze=new Map}register(r){var o;mi((o=this._instance)==null?void 0:o.Ge)||(this.tr(r),this.ir(r))}ir(r){var o=r.filter(s=>{var c,d;return((c=s.conditions)==null?void 0:c.actions)&&((d=s.conditions)==null||(d=d.actions)==null||(d=d.values)==null?void 0:d.length)>0});o.length!==0&&(this.er==null&&(this.er=new xB(this._instance),this.er.init(),this.er.Ve(s=>{this.onAction(s)})),o.forEach(s=>{var c,d,h,b,y;s.conditions&&(c=s.conditions)!=null&&c.actions&&(d=s.conditions)!=null&&(d=d.actions)!=null&&d.values&&((h=s.conditions)==null||(h=h.actions)==null||(h=h.values)==null?void 0:h.length)>0&&((b=this.er)==null||b.register(s.conditions.actions.values),(y=s.conditions)==null||(y=y.actions)==null||(y=y.values)==null||y.forEach(E=>{if(E&&E.name){var k=this.Ze.get(E.name);k&&k.push(s.id),this.Ze.set(E.name,k||[s.id])}}))}))}tr(r){var o;r.filter(s=>{var c,d;return((c=s.conditions)==null?void 0:c.events)&&((d=s.conditions)==null||(d=d.events)==null||(d=d.values)==null?void 0:d.length)>0}).length!==0&&((o=this._instance)==null||o.Ge((s,c)=>{this.onEvent(s,c)}),r.forEach(s=>{var c;(c=s.conditions)==null||(c=c.events)==null||(c=c.values)==null||c.forEach(d=>{if(d&&d.name){var h=this.Qe.get(d.name);h&&h.push(s.id),this.Qe.set(d.name,h||[s.id])}})}))}onEvent(r,o){var s,c=((s=this._instance)==null||(s=s.persistence)==null?void 0:s.props[x0e])||[];if(r==="survey shown"&&o&&c.length>0){var d;$m.info("survey event matched, removing survey from activated surveys",{event:r,eventPayload:o,existingActivatedSurveys:c});var h=o==null||(d=o.properties)==null?void 0:d.$survey_id;if(h){var b=c.indexOf(h);b>=0&&(c.splice(b,1),this.rr(c))}}else this.Qe.has(r)&&($m.info("survey event matched, updating activated surveys",{event:r,surveys:this.Qe.get(r)}),this.rr(c.concat(this.Qe.get(r)||[])))}onAction(r){var o,s=((o=this._instance)==null||(o=o.persistence)==null?void 0:o.props[x0e])||[];this.Ze.has(r)&&this.rr(s.concat(this.Ze.get(r)||[]))}rr(r){var o;(o=this._instance)==null||(o=o.persistence)==null||o.register({[x0e]:[...new Set(r)]})}getSurveys(){var r,o=(r=this._instance)==null||(r=r.persistence)==null?void 0:r.props[x0e];return o||[]}getEventToSurveys(){return this.Qe}sr(){return this.er}}class TDr{constructor(r){this.nr=void 0,this.ar=null,this.lr=!1,this.ur=!1,this.hr=[],this._instance=r,this._surveyEventReceiver=null}onRemoteConfig(r){var o=r.surveys;if(Gu(o))return $m.warn("Flags not loaded yet. Not loading surveys.");var s=kf(o);this.nr=s?o.length>0:o,$m.info("flags response received, isSurveysEnabled: "+this.nr),this.loadIfEnabled()}reset(){localStorage.removeItem("lastSeenSurveyDate");for(var r=[],o=0;o<localStorage.length;o++){var s=localStorage.key(o);(s!=null&&s.startsWith(WWe)||s!=null&&s.startsWith("inProgressSurvey_"))&&r.push(s)}r.forEach(c=>localStorage.removeItem(c))}loadIfEnabled(){if(!this.ar)if(this.ur)$m.info("Already initializing surveys, skipping...");else if(this._instance.config.disable_surveys)$m.info("Disabled. Not loading surveys.");else{var r=La?.__PosthogExtensions__;if(r){var o=this.nr;if(!mi(o)){this.ur=!0;try{var s=r.generateSurveys;if(s)return void this.dr(s,o);var c=r.loadExternalDependency;if(!c)return void this.vr("PostHog loadExternalDependency extension not found.");c(this._instance,"surveys",d=>{d||!r.generateSurveys?this.vr("Could not load surveys script",d):this.dr(r.generateSurveys,o)})}catch(d){throw this.vr("Error initializing surveys",d),d}finally{this.ur=!1}}}else $m.error("PostHog Extensions not found.")}}dr(r,o){this.ar=r(this._instance,o),this._surveyEventReceiver=new kDr(this._instance),$m.info("Surveys loaded successfully"),this.cr({isLoaded:!0})}vr(r,o){$m.error(r,o),this.cr({isLoaded:!1,error:r})}onSurveysLoaded(r){return this.hr.push(r),this.ar&&this.cr({isLoaded:!0}),()=>{this.hr=this.hr.filter(o=>o!==r)}}getSurveys(r,o){if(o===void 0&&(o=!1),this._instance.config.disable_surveys)return $m.info("Disabled. Not loading surveys."),r([]);var s=this._instance.get_property(OWe);if(s&&!o)return r(s,{isLoaded:!0});if(this.lr)return r([],{isLoaded:!1,error:"Surveys are already being loaded"});try{this.lr=!0,this._instance.Ee({url:this._instance.requestRouter.endpointFor("api","/api/surveys/?token="+this._instance.config.token),method:"GET",timeout:this._instance.config.surveys_request_timeout_ms,callback:c=>{var d;this.lr=!1;var h=c.statusCode;if(h!==200||!c.json){var b="Surveys API could not be loaded, status: "+h;return $m.error(b),r([],{isLoaded:!1,error:b})}var y,E=c.json.surveys||[],k=E.filter(x=>function(O){return!(!O.start_date||O.end_date)}(x)&&(function(O){var P;return!((P=O.conditions)==null||(P=P.events)==null||(P=P.values)==null||!P.length)}(x)||function(O){var P;return!((P=O.conditions)==null||(P=P.actions)==null||(P=P.values)==null||!P.length)}(x)));return k.length>0&&((y=this._surveyEventReceiver)==null||y.register(k)),(d=this._instance.persistence)==null||d.register({[OWe]:E}),r(E,{isLoaded:!0})}})}catch(c){throw this.lr=!1,c}}cr(r){for(var o of this.hr)try{r.isLoaded?this.getSurveys(o):o([],r)}catch(s){$m.error("Error in survey callback",s)}}getActiveMatchingSurveys(r,o){if(o===void 0&&(o=!1),!Gu(this.ar))return this.ar.getActiveMatchingSurveys(r,o);$m.warn("init was not called")}pr(r){var o=null;return this.getSurveys(s=>{var c;o=(c=s.find(d=>d.id===r))!==null&&c!==void 0?c:null}),o}gr(r){if(Gu(this.ar))return{eligible:!1,reason:"SDK is not enabled or survey functionality is not yet loaded"};var o=typeof r=="string"?this.pr(r):r;return o?this.ar.checkSurveyEligibility(o):{eligible:!1,reason:"Survey not found"}}canRenderSurvey(r){if(Gu(this.ar))return $m.warn("init was not called"),{visible:!1,disabledReason:"SDK is not enabled or survey functionality is not yet loaded"};var o=this.gr(r);return{visible:o.eligible,disabledReason:o.reason}}canRenderSurveyAsync(r,o){return Gu(this.ar)?($m.warn("init was not called"),Promise.resolve({visible:!1,disabledReason:"SDK is not enabled or survey functionality is not yet loaded"})):new Promise(s=>{this.getSurveys(c=>{var d,h=(d=c.find(y=>y.id===r))!==null&&d!==void 0?d:null;if(h){var b=this.gr(h);s({visible:b.eligible,disabledReason:b.reason})}else s({visible:!1,disabledReason:"Survey not found"})},o)})}renderSurvey(r,o){if(Gu(this.ar))$m.warn("init was not called");else{var s=this.pr(r),c=Di?.querySelector(o);s?c?this.ar.renderSurvey(s,c):$m.warn("Survey element not found"):$m.warn("Survey not found")}}}(function(n){return n.Button="button",n.Tab="tab",n.Selector="selector",n})({});(function(n){return n.TopLeft="top_left",n.TopRight="top_right",n.TopCenter="top_center",n.MiddleLeft="middle_left",n.MiddleRight="middle_right",n.MiddleCenter="middle_center",n.Left="left",n.Center="center",n.Right="right",n.NextToTrigger="next_to_trigger",n})({});(function(n){return n.Popover="popover",n.API="api",n.Widget="widget",n})({});(function(n){return n.Open="open",n.MultipleChoice="multiple_choice",n.SingleChoice="single_choice",n.Rating="rating",n.Link="link",n})({});(function(n){return n.NextQuestion="next_question",n.End="end",n.ResponseBased="response_based",n.SpecificQuestion="specific_question",n})({});(function(n){return n.Once="once",n.Recurring="recurring",n.Always="always",n})({});var lFe=function(n){return n.SHOWN="survey shown",n.DISMISSED="survey dismissed",n.SENT="survey sent",n}({}),OBt=function(n){return n.SURVEY_ID="$survey_id",n.SURVEY_NAME="$survey_name",n.SURVEY_RESPONSE="$survey_response",n.SURVEY_ITERATION="$survey_iteration",n.SURVEY_ITERATION_START_DATE="$survey_iteration_start_date",n.SURVEY_PARTIALLY_COMPLETED="$survey_partially_completed",n.SURVEY_SUBMISSION_ID="$survey_submission_id",n.SURVEY_QUESTIONS="$survey_questions",n.SURVEY_COMPLETED="$survey_completed",n}({}),NBt=Wm("[RateLimiter]");class xDr{constructor(r){var o,s;this.serverLimits={},this.lastEventRateLimited=!1,this.checkForLimiting=c=>{var d=c.text;if(d&&d.length)try{(JSON.parse(d).quota_limited||[]).forEach(h=>{NBt.info((h||"events")+" is quota limited."),this.serverLimits[h]=new Date().getTime()+6e4})}catch(h){return void NBt.warn('could not rate limit - continuing. Error: "'+h?.message+'"',{text:d})}},this.instance=r,this.captureEventsPerSecond=((o=r.config.rate_limiting)==null?void 0:o.events_per_second)||10,this.captureEventsBurstLimit=Math.max(((s=r.config.rate_limiting)==null?void 0:s.events_burst_limit)||10*this.captureEventsPerSecond,this.captureEventsPerSecond),this.lastEventRateLimited=this.clientRateLimitContext(!0).isRateLimited}clientRateLimitContext(r){var o,s,c;r===void 0&&(r=!1);var d=new Date().getTime(),h=(o=(s=this.instance.persistence)==null?void 0:s.get_property(RWe))!==null&&o!==void 0?o:{tokens:this.captureEventsBurstLimit,last:d};h.tokens+=(d-h.last)/1e3*this.captureEventsPerSecond,h.last=d,h.tokens>this.captureEventsBurstLimit&&(h.tokens=this.captureEventsBurstLimit);var b=h.tokens<1;return b||r||(h.tokens=Math.max(0,h.tokens-1)),!b||this.lastEventRateLimited||r||this.instance.capture("$$client_ingestion_warning",{$$client_ingestion_warning_message:"posthog-js client rate limited. Config is set to "+this.captureEventsPerSecond+" events per second and "+this.captureEventsBurstLimit+" events burst limit."},{skip_client_rate_limiting:!0}),this.lastEventRateLimited=b,(c=this.instance.persistence)==null||c.set_property(RWe,h),{isRateLimited:b,remainingTokens:h.tokens}}isServerRateLimited(r){var o=this.serverLimits[r||"events"]||!1;return o!==!1&&new Date().getTime()<o}}var wB=Wm("[RemoteConfig]");class CDr{constructor(r){this._instance=r}get remoteConfig(){var r;return(r=La._POSTHOG_REMOTE_CONFIG)==null||(r=r[this._instance.config.token])==null?void 0:r.config}_r(r){var o,s;(o=La.__PosthogExtensions__)!=null&&o.loadExternalDependency?(s=La.__PosthogExtensions__)==null||s.loadExternalDependency==null||s.loadExternalDependency(this._instance,"remote-config",()=>r(this.remoteConfig)):(wB.error("PostHog Extensions not found. Cannot load remote config."),r())}mr(r){this._instance.Ee({method:"GET",url:this._instance.requestRouter.endpointFor("assets","/array/"+this._instance.config.token+"/config"),callback:o=>{r(o.json)}})}load(){try{if(this.remoteConfig)return wB.info("Using preloaded remote config",this.remoteConfig),void this.Ie(this.remoteConfig);if(this._instance.I())return void wB.warn("Remote config is disabled. Falling back to local config.");this._r(r=>{if(!r)return wB.info("No config found after loading remote JS config. Falling back to JSON."),void this.mr(o=>{this.Ie(o)});this.Ie(r)})}catch(r){wB.error("Error loading remote config",r)}}Ie(r){r?this._instance.config.__preview_remote_config?(this._instance.Ie(r),r.hasFeatureFlags!==!1&&this._instance.featureFlags.ensureFlagsLoaded()):wB.info("__preview_remote_config is disabled. Logging config instead",r):wB.error("Failed to fetch remote config from PostHog.")}}var KWe=3e3;class ADr{constructor(r,o){this.br=!0,this.yr=[],this.wr=u2(o?.flush_interval_ms||KWe,250,5e3,"flush interval",KWe),this.Sr=r}enqueue(r){this.yr.push(r),this.$r||this.kr()}unload(){this.Er();var r=this.yr.length>0?this.Ir():{},o=Object.values(r);[...o.filter(s=>s.url.indexOf("/e")===0),...o.filter(s=>s.url.indexOf("/e")!==0)].map(s=>{this.Sr(Ni({},s,{transport:"sendBeacon"}))})}enable(){this.br=!1,this.kr()}kr(){var r=this;this.br||(this.$r=setTimeout(()=>{if(this.Er(),this.yr.length>0){var o=this.Ir(),s=function(){var d=o[c],h=new Date().getTime();d.data&&kf(d.data)&&Bl(d.data,b=>{b.offset=Math.abs(b.timestamp-h),delete b.timestamp}),r.Sr(d)};for(var c in o)s()}},this.wr))}Er(){clearTimeout(this.$r),this.$r=void 0}Ir(){var r={};return Bl(this.yr,o=>{var s,c=o,d=(c?c.batchKey:null)||c.url;mi(r[d])&&(r[d]=Ni({},c,{data:[]})),(s=r[d].data)==null||s.push(c.data)}),this.yr=[],r}}var IDr=["retriesPerformedSoFar"];class ODr{constructor(r){this.Pr=!1,this.Rr=3e3,this.yr=[],this._instance=r,this.yr=[],this.Tr=!0,!mi(kn)&&"onLine"in kn.navigator&&(this.Tr=kn.navigator.onLine,th(kn,"online",()=>{this.Tr=!0,this.se()}),th(kn,"offline",()=>{this.Tr=!1}))}get length(){return this.yr.length}retriableRequest(r){var{retriesPerformedSoFar:o}=r,s=Ycn(r,IDr);G1(o)&&o>0&&(s.url=Uwe(s.url,{retry_count:o})),this._instance.Ee(Ni({},s,{callback:c=>{c.statusCode!==200&&(c.statusCode<400||c.statusCode>=500)&&(o??0)<10?this.Mr(Ni({retriesPerformedSoFar:o},s)):s.callback==null||s.callback(c)}}))}Mr(r){var o=r.retriesPerformedSoFar||0;r.retriesPerformedSoFar=o+1;var s=function(h){var b=3e3*Math.pow(2,h),y=b/2,E=Math.min(18e5,b),k=(Math.random()-.5)*(E-y);return Math.ceil(E+k)}(o),c=Date.now()+s;this.yr.push({retryAt:c,requestOptions:r});var d="Enqueued failed request for retry in "+s;navigator.onLine||(d+=" (Browser is offline)"),zi.warn(d),this.Pr||(this.Pr=!0,this.Cr())}Cr(){this.Fr&&clearTimeout(this.Fr),this.Fr=setTimeout(()=>{this.Tr&&this.yr.length>0&&this.se(),this.Cr()},this.Rr)}se(){var r=Date.now(),o=[],s=this.yr.filter(d=>d.retryAt<r||(o.push(d),!1));if(this.yr=o,s.length>0)for(var{requestOptions:c}of s)this.retriableRequest(c)}unload(){for(var{requestOptions:r}of(this.Fr&&(clearTimeout(this.Fr),this.Fr=void 0),this.yr))try{this._instance.Ee(Ni({},r,{transport:"sendBeacon"}))}catch(o){zi.error(o)}this.yr=[]}}class NDr{constructor(r){this.Or=()=>{var o,s,c,d;this.Ar||(this.Ar={});var h=this.scrollElement(),b=this.scrollY(),y=h?Math.max(0,h.scrollHeight-h.clientHeight):0,E=b+(h?.clientHeight||0),k=h?.scrollHeight||0;this.Ar.lastScrollY=Math.ceil(b),this.Ar.maxScrollY=Math.max(b,(o=this.Ar.maxScrollY)!==null&&o!==void 0?o:0),this.Ar.maxScrollHeight=Math.max(y,(s=this.Ar.maxScrollHeight)!==null&&s!==void 0?s:0),this.Ar.lastContentY=E,this.Ar.maxContentY=Math.max(E,(c=this.Ar.maxContentY)!==null&&c!==void 0?c:0),this.Ar.maxContentHeight=Math.max(k,(d=this.Ar.maxContentHeight)!==null&&d!==void 0?d:0)},this._instance=r}getContext(){return this.Ar}resetContext(){var r=this.Ar;return setTimeout(this.Or,0),r}startMeasuringScrollPosition(){th(kn,"scroll",this.Or,{capture:!0}),th(kn,"scrollend",this.Or,{capture:!0}),th(kn,"resize",this.Or)}scrollElement(){if(!this._instance.config.scroll_root_selector)return kn?.document.documentElement;var r=kf(this._instance.config.scroll_root_selector)?this._instance.config.scroll_root_selector:[this._instance.config.scroll_root_selector];for(var o of r){var s=kn?.document.querySelector(o);if(s)return s}}scrollY(){if(this._instance.config.scroll_root_selector){var r=this.scrollElement();return r&&r.scrollTop||0}return kn&&(kn.scrollY||kn.pageYOffset||kn.document.documentElement.scrollTop)||0}scrollX(){if(this._instance.config.scroll_root_selector){var r=this.scrollElement();return r&&r.scrollLeft||0}return kn&&(kn.scrollX||kn.pageXOffset||kn.document.documentElement.scrollLeft)||0}}var RDr=n=>Yun(n?.config.mask_personal_data_properties,n?.config.custom_personal_data_properties);class PDr{constructor(r,o,s,c){this.Dr=d=>{var h=this.Lr();if(!h||h.sessionId!==d){var b={sessionId:d,props:this.jr(this._instance)};this.Nr.register({[NWe]:b})}},this._instance=r,this.zr=o,this.Nr=s,this.jr=c||RDr,this.zr.onSessionId(this.Dr)}Lr(){return this.Nr.props[NWe]}getSetOnceProps(){var r,o=(r=this.Lr())==null?void 0:r.props;return o?"r"in o?Xun(o):{$referring_domain:o.referringDomain,$pathname:o.initialPathName,utm_source:o.utm_source,utm_campaign:o.utm_campaign,utm_medium:o.utm_medium,utm_content:o.utm_content,utm_term:o.utm_term}:{}}getSessionProps(){var r={};return Bl(qZe(this.getSetOnceProps()),(o,s)=>{s==="$current_url"&&(s="url"),r["$session_entry_"+SWe(s)]=o}),r}}var RBt=Wm("[SessionId]");class LDr{constructor(r,o,s){var c;if(this.Ur=[],!r.persistence)throw new Error("SessionIdManager requires a PostHogPersistence instance");if(r.config.__preview_experimental_cookieless_mode)throw new Error("SessionIdManager cannot be used with __preview_experimental_cookieless_mode");this.S=r.config,this.Nr=r.persistence,this.fi=void 0,this.Ct=void 0,this._sessionStartTimestamp=null,this._sessionActivityTimestamp=null,this.qr=o||uM,this.Br=s||uM;var d=this.S.persistence_name||this.S.token,h=this.S.session_idle_timeout_seconds||1800;if(this._sessionTimeoutMs=1e3*u2(h,60,36e3,"session_idle_timeout_seconds",1800),r.register({$configured_session_timeout_ms:this._sessionTimeoutMs}),this.Hr(),this.Wr="ph_"+d+"_window_id",this.Gr="ph_"+d+"_primary_window_exists",this.Jr()){var b=Ub.L(this.Wr),y=Ub.L(this.Gr);b&&!y?this.fi=b:Ub.N(this.Wr),Ub.j(this.Gr,!0)}if((c=this.S.bootstrap)!=null&&c.sessionID)try{var E=(k=>{var x=k.replace(/-/g,"");if(x.length!==32)throw new Error("Not a valid UUID");if(x[12]!=="7")throw new Error("Not a UUIDv7");return parseInt(x.substring(0,12),16)})(this.S.bootstrap.sessionID);this.Vr(this.S.bootstrap.sessionID,new Date().getTime(),E)}catch(k){RBt.error("Invalid sessionID in bootstrap",k)}this.Kr()}get sessionTimeoutMs(){return this._sessionTimeoutMs}onSessionId(r){return mi(this.Ur)&&(this.Ur=[]),this.Ur.push(r),this.Ct&&r(this.Ct,this.fi),()=>{this.Ur=this.Ur.filter(o=>o!==r)}}Jr(){return this.S.persistence!=="memory"&&!this.Nr.Fe&&Ub.O()}Yr(r){r!==this.fi&&(this.fi=r,this.Jr()&&Ub.j(this.Wr,r))}Xr(){return this.fi?this.fi:this.Jr()?Ub.L(this.Wr):null}Vr(r,o,s){r===this.Ct&&o===this._sessionActivityTimestamp&&s===this._sessionStartTimestamp||(this._sessionStartTimestamp=s,this._sessionActivityTimestamp=o,this.Ct=r,this.Nr.register({[Owe]:[o,r,s]}))}Qr(){if(this.Ct&&this._sessionActivityTimestamp&&this._sessionStartTimestamp)return[this._sessionActivityTimestamp,this.Ct,this._sessionStartTimestamp];var r=this.Nr.props[Owe];return kf(r)&&r.length===2&&r.push(r[0]),r||[0,null,0]}resetSessionId(){this.Vr(null,null,null)}Kr(){th(kn,"beforeunload",()=>{this.Jr()&&Ub.N(this.Gr)},{capture:!1})}checkAndGetSessionAndWindowId(r,o){if(r===void 0&&(r=!1),o===void 0&&(o=null),this.S.__preview_experimental_cookieless_mode)throw new Error("checkAndGetSessionAndWindowId should not be called in __preview_experimental_cookieless_mode");var s=o||new Date().getTime(),[c,d,h]=this.Qr(),b=this.Xr(),y=G1(h)&&h>0&&Math.abs(s-h)>864e5,E=!1,k=!d,x=!r&&Math.abs(s-c)>this.sessionTimeoutMs;k||x||y?(d=this.qr(),b=this.Br(),RBt.info("new session ID generated",{sessionId:d,windowId:b,changeReason:{noSessionId:k,activityTimeout:x,sessionPastMaximumLength:y}}),h=s,E=!0):b||(b=this.Br(),E=!0);var O=c===0||!r||y?s:c,P=h===0?new Date().getTime():h;return this.Yr(b),this.Vr(d,O,P),r||this.Hr(),E&&this.Ur.forEach(D=>D(d,b,E?{noSessionId:k,activityTimeout:x,sessionPastMaximumLength:y}:void 0)),{sessionId:d,windowId:b,sessionStartTimestamp:P,changeReason:E?{noSessionId:k,activityTimeout:x,sessionPastMaximumLength:y}:void 0,lastActivityTimestamp:c}}Hr(){clearTimeout(this.Zr),this.Zr=setTimeout(()=>{this.resetSessionId()},1.1*this.sessionTimeoutMs)}}var DDr=["$set_once","$set"],WD=Wm("[SiteApps]");class MDr{constructor(r){this._instance=r,this.ts=[],this.apps={}}get isEnabled(){return!!this._instance.config.opt_in_site_apps}es(r,o){if(o){var s=this.globalsForEvent(o);this.ts.push(s),this.ts.length>1e3&&(this.ts=this.ts.slice(10))}}get siteAppLoaders(){var r;return(r=La._POSTHOG_REMOTE_CONFIG)==null||(r=r[this._instance.config.token])==null?void 0:r.siteApps}init(){if(this.isEnabled){var r=this._instance.Ge(this.es.bind(this));this.rs=()=>{r(),this.ts=[],this.rs=void 0}}}globalsForEvent(r){var o,s,c,d,h,b,y;if(!r)throw new Error("Event payload is required");var E={},k=this._instance.get_property("$groups")||[],x=this._instance.get_property("$stored_group_properties")||{};for(var[O,P]of Object.entries(x))E[O]={id:k[O],type:O,properties:P};var{$set_once:D,$set:$}=r;return{event:Ni({},Ycn(r,DDr),{properties:Ni({},r.properties,$?{$set:Ni({},(o=(s=r.properties)==null?void 0:s.$set)!==null&&o!==void 0?o:{},$)}:{},D?{$set_once:Ni({},(c=(d=r.properties)==null?void 0:d.$set_once)!==null&&c!==void 0?c:{},D)}:{}),elements_chain:(h=(b=r.properties)==null?void 0:b.$elements_chain)!==null&&h!==void 0?h:"",distinct_id:(y=r.properties)==null?void 0:y.distinct_id}),person:{properties:this._instance.get_property("$stored_person_properties")},groups:E}}setupSiteApp(r){var o=this.apps[r.id],s=()=>{var b;!o.errored&&this.ts.length&&(WD.info("Processing "+this.ts.length+" events for site app with id "+r.id),this.ts.forEach(y=>o.processEvent==null?void 0:o.processEvent(y)),o.processedBuffer=!0),Object.values(this.apps).every(y=>y.processedBuffer||y.errored)&&((b=this.rs)==null||b.call(this))},c=!1,d=b=>{o.errored=!b,o.loaded=!0,WD.info("Site app with id "+r.id+" "+(b?"loaded":"errored")),c&&s()};try{var{processEvent:h}=r.init({posthog:this._instance,callback:b=>{d(b)}});h&&(o.processEvent=h),c=!0}catch(b){WD.error("Error while initializing PostHog app with config id "+r.id,b),d(!1)}if(c&&o.loaded)try{s()}catch(b){WD.error("Error while processing buffered events PostHog app with config id "+r.id,b),o.errored=!0}}ss(){var r=this.siteAppLoaders||[];for(var o of r)this.apps[o.id]={id:o.id,loaded:!1,errored:!1,processedBuffer:!1};for(var s of r)this.setupSiteApp(s)}ns(r){if(Object.keys(this.apps).length!==0){var o=this.globalsForEvent(r);for(var s of Object.values(this.apps))try{s.processEvent==null||s.processEvent(o)}catch(c){WD.error("Error while processing event "+r.event+" for site app "+s.id,c)}}}onRemoteConfig(r){var o,s,c,d=this;if((o=this.siteAppLoaders)!=null&&o.length)return this.isEnabled?(this.ss(),void this._instance.on("eventCaptured",E=>this.ns(E))):void WD.error('PostHog site apps are disabled. Enable the "opt_in_site_apps" config to proceed.');if((s=this.rs)==null||s.call(this),(c=r.siteApps)!=null&&c.length)if(this.isEnabled){var h=function(E){var k;La["__$$ph_site_app_"+E]=d._instance,(k=La.__PosthogExtensions__)==null||k.loadSiteApp==null||k.loadSiteApp(d._instance,y,x=>{if(x)return WD.error("Error while initializing PostHog app with config id "+E,x)})};for(var{id:b,url:y}of r.siteApps)h(b)}else WD.error('PostHog site apps are disabled. Enable the "opt_in_site_apps" config to proceed.')}}var jDr=["amazonbot","amazonproductbot","app.hypefactors.com","applebot","archive.org_bot","awariobot","backlinksextendedbot","baiduspider","bingbot","bingpreview","chrome-lighthouse","dataforseobot","deepscan","duckduckbot","facebookexternal","facebookcatalog","http://yandex.com/bots","hubspot","ia_archiver","leikibot","linkedinbot","meta-externalagent","mj12bot","msnbot","nessus","petalbot","pinterest","prerender","rogerbot","screaming frog","sebot-wa","sitebulb","slackbot","slurp","trendictionbot","turnitin","twitterbot","vercelbot","yahoo! slurp","yandexbot","zoombot","bot.htm","bot.php","(bot;","bot/","crawler","ahrefsbot","ahrefssiteaudit","semrushbot","siteauditbot","splitsignalbot","gptbot","oai-searchbot","chatgpt-user","perplexitybot","better uptime bot","sentryuptimebot","uptimerobot","headlesschrome","cypress","google-hoteladsverifier","adsbot-google","apis-google","duplexweb-google","feedfetcher-google","google favicon","google web preview","google-read-aloud","googlebot","googleother","google-cloudvertexbot","googleweblight","mediapartners-google","storebot-google","google-inspectiontool","bytespider"],PBt=function(n,r){if(!n)return!1;var o=n.toLowerCase();return jDr.concat(r||[]).some(s=>{var c=s.toLowerCase();return o.indexOf(c)!==-1})},Qun=function(n,r){if(!n)return!1;var o=n.userAgent;if(o&&PBt(o,r))return!0;try{var s=n?.userAgentData;if(s!=null&&s.brands&&s.brands.some(c=>PBt(c?.brand,r)))return!0}catch{}return!!n.webdriver},Kre=function(n){return n.US="us",n.EU="eu",n.CUSTOM="custom",n}({}),LBt="i.posthog.com";class $Dr{constructor(r){this.os={},this.instance=r}get apiHost(){var r=this.instance.config.api_host.trim().replace(/\/$/,"");return r==="https://app.posthog.com"?"https://us.i.posthog.com":r}get uiHost(){var r,o=(r=this.instance.config.ui_host)==null?void 0:r.replace(/\/$/,"");return o||(o=this.apiHost.replace("."+LBt,".posthog.com")),o==="https://app.posthog.com"?"https://us.posthog.com":o}get region(){return this.os[this.apiHost]||(/https:\/\/(app|us|us-assets)(\.i)?\.posthog\.com/i.test(this.apiHost)?this.os[this.apiHost]=Kre.US:/https:\/\/(eu|eu-assets)(\.i)?\.posthog\.com/i.test(this.apiHost)?this.os[this.apiHost]=Kre.EU:this.os[this.apiHost]=Kre.CUSTOM),this.os[this.apiHost]}endpointFor(r,o){if(o===void 0&&(o=""),o&&(o=o[0]==="/"?o:"/"+o),r==="ui")return this.uiHost+o;if(this.region===Kre.CUSTOM)return this.apiHost+o;var s=LBt+o;switch(r){case"assets":return"https://"+this.region+"-assets."+s;case"api":return"https://"+this.region+"."+s}}}var FDr={icontains:(n,r)=>!!kn&&r.href.toLowerCase().indexOf(n.toLowerCase())>-1,not_icontains:(n,r)=>!!kn&&r.href.toLowerCase().indexOf(n.toLowerCase())===-1,regex:(n,r)=>!!kn&&uK(r.href,n),not_regex:(n,r)=>!!kn&&!uK(r.href,n),exact:(n,r)=>r.href===n,is_not:(n,r)=>r.href!==n};class H1{constructor(r){var o=this;this.getWebExperimentsAndEvaluateDisplayLogic=function(s){s===void 0&&(s=!1),o.getWebExperiments(c=>{H1.ls("retrieved web experiments from the server"),o.us=new Map,c.forEach(d=>{if(d.feature_flag_key){var h;o.us&&(H1.ls("setting flag key ",d.feature_flag_key," to web experiment ",d),(h=o.us)==null||h.set(d.feature_flag_key,d));var b=o._instance.getFeatureFlag(d.feature_flag_key);Ih(b)&&d.variants[b]&&o.hs(d.name,b,d.variants[b].transforms)}else if(d.variants)for(var y in d.variants){var E=d.variants[y];H1.ds(E)&&o.hs(d.name,y,E.transforms)}})},s)},this._instance=r,this._instance.onFeatureFlags(s=>{this.onFeatureFlags(s)})}onFeatureFlags(r){if(this._is_bot())H1.ls("Refusing to render web experiment since the viewer is a likely bot");else if(!this._instance.config.disable_web_experiments){if(Gu(this.us))return this.us=new Map,this.loadIfEnabled(),void this.previewWebExperiment();H1.ls("applying feature flags",r),r.forEach(o=>{var s;if(this.us&&(s=this.us)!=null&&s.has(o)){var c,d=this._instance.getFeatureFlag(o),h=(c=this.us)==null?void 0:c.get(o);d&&h!=null&&h.variants[d]&&this.hs(h.name,d,h.variants[d].transforms)}})}}previewWebExperiment(){var r=H1.getWindowLocation();if(r!=null&&r.search){var o=Dwe(r?.search,"__experiment_id"),s=Dwe(r?.search,"__experiment_variant");o&&s&&(H1.ls("previewing web experiments "+o+" && "+s),this.getWebExperiments(c=>{this.vs(parseInt(o),s,c)},!1,!0))}}loadIfEnabled(){this._instance.config.disable_web_experiments||this.getWebExperimentsAndEvaluateDisplayLogic()}getWebExperiments(r,o,s){if(this._instance.config.disable_web_experiments&&!s)return r([]);var c=this._instance.get_property("$web_experiments");if(c&&!o)return r(c);this._instance.Ee({url:this._instance.requestRouter.endpointFor("api","/api/web_experiments/?token="+this._instance.config.token),method:"GET",callback:d=>{if(d.statusCode!==200||!d.json)return r([]);var h=d.json.experiments||[];return r(h)}})}vs(r,o,s){var c=s.filter(d=>d.id===r);c&&c.length>0&&(H1.ls("Previewing web experiment ["+c[0].name+"] with variant ["+o+"]"),this.hs(c[0].name,o,c[0].variants[o].transforms))}static ds(r){return!Gu(r.conditions)&&H1.cs(r)&&H1.fs(r)}static cs(r){var o;if(Gu(r.conditions)||Gu((o=r.conditions)==null?void 0:o.url))return!0;var s,c,d,h=H1.getWindowLocation();return!!h&&((s=r.conditions)==null||!s.url||FDr[(c=(d=r.conditions)==null?void 0:d.urlMatchType)!==null&&c!==void 0?c:"icontains"](r.conditions.url,h))}static getWindowLocation(){return kn?.location}static fs(r){var o;if(Gu(r.conditions)||Gu((o=r.conditions)==null?void 0:o.utm))return!0;var s=qun();if(s.utm_source){var c,d,h,b,y,E,k,x,O=(c=r.conditions)==null||(c=c.utm)==null||!c.utm_campaign||((d=r.conditions)==null||(d=d.utm)==null?void 0:d.utm_campaign)==s.utm_campaign,P=(h=r.conditions)==null||(h=h.utm)==null||!h.utm_source||((b=r.conditions)==null||(b=b.utm)==null?void 0:b.utm_source)==s.utm_source,D=(y=r.conditions)==null||(y=y.utm)==null||!y.utm_medium||((E=r.conditions)==null||(E=E.utm)==null?void 0:E.utm_medium)==s.utm_medium,$=(k=r.conditions)==null||(k=k.utm)==null||!k.utm_term||((x=r.conditions)==null||(x=x.utm)==null?void 0:x.utm_term)==s.utm_term;return O&&D&&$&&P}return!1}static ls(r){for(var o=arguments.length,s=new Array(o>1?o-1:0),c=1;c<o;c++)s[c-1]=arguments[c];zi.info("[WebExperiments] "+r,s)}hs(r,o,s){this._is_bot()?H1.ls("Refusing to render web experiment since the viewer is a likely bot"):o!=="control"?s.forEach(c=>{if(c.selector){var d;H1.ls("applying transform of variant "+o+" for experiment "+r+" ",c);var h=(d=document)==null?void 0:d.querySelectorAll(c.selector);h?.forEach(b=>{var y=b;c.html&&(y.innerHTML=c.html),c.css&&y.setAttribute("style",c.css)})}}):H1.ls("Control variants leave the page unmodified.")}_is_bot(){return c2&&this._instance?Qun(c2,this._instance.config.custom_blocked_useragents):void 0}}var BDr=Wm("[PostHog ExternalIntegrations]"),UDr={intercom:"intercom-integration",crispChat:"crisp-chat-integration"};class zDr{constructor(r){this._instance=r}J(r,o){var s;(s=La.__PosthogExtensions__)==null||s.loadExternalDependency==null||s.loadExternalDependency(this._instance,r,c=>{if(c)return BDr.error("failed to load script",c);o()})}startIfEnabledOrStop(){var r=this,o=function(h){var b,y,E;!c||(b=La.__PosthogExtensions__)!=null&&(b=b.integrations)!=null&&b[h]||r.J(UDr[h],()=>{var k;(k=La.__PosthogExtensions__)==null||(k=k.integrations)==null||(k=k[h])==null||k.start(r._instance)}),!c&&(y=La.__PosthogExtensions__)!=null&&(y=y.integrations)!=null&&y[h]&&((E=La.__PosthogExtensions__)==null||(E=E.integrations)==null||(E=E[h])==null||E.stop())};for(var[s,c]of Object.entries((d=this._instance.config.integrations)!==null&&d!==void 0?d:{})){var d;o(s)}}}var mie={},YWe=()=>{},aW="posthog",eln=!tDr&&Gw?.indexOf("MSIE")===-1&&Gw?.indexOf("Mozilla")===-1,DBt=n=>{var r;return{api_host:"https://us.i.posthog.com",ui_host:null,token:"",autocapture:!0,rageclick:!0,cross_subdomain_cookie:PPr(Di?.location),persistence:"localStorage+cookie",persistence_name:"",loaded:YWe,save_campaign_params:!0,custom_campaign_params:[],custom_blocked_useragents:[],save_referrer:!0,capture_pageview:n!=="2025-05-24"||"history_change",capture_pageleave:"if_capture_pageview",defaults:n??"unset",debug:Kw&&Ih(Kw?.search)&&Kw.search.indexOf("__posthog_debug=true")!==-1||!1,cookie_expiration:365,upgrade:!1,disable_session_recording:!1,disable_persistence:!1,disable_web_experiments:!0,disable_surveys:!1,disable_surveys_automatic_display:!1,disable_external_dependency_loading:!1,enable_recording_console_log:void 0,secure_cookie:(kn==null||(r=kn.location)==null?void 0:r.protocol)==="https:",ip:!1,opt_out_capturing_by_default:!1,opt_out_persistence_by_default:!1,opt_out_useragent_filter:!1,opt_out_capturing_persistence_type:"localStorage",opt_out_capturing_cookie_prefix:null,opt_in_site_apps:!1,property_denylist:[],respect_dnt:!1,sanitize_properties:null,request_headers:{},request_batching:!0,properties_string_max_length:65535,session_recording:{},mask_all_element_attributes:!1,mask_all_text:!1,mask_personal_data_properties:!1,custom_personal_data_properties:[],advanced_disable_flags:!1,advanced_disable_decide:!1,advanced_disable_feature_flags:!1,advanced_disable_feature_flags_on_first_load:!1,advanced_only_evaluate_survey_feature_flags:!1,advanced_disable_toolbar_metrics:!1,feature_flag_request_timeout_ms:3e3,surveys_request_timeout_ms:1e4,on_request_error:o=>{var s="Bad HTTP status: "+o.statusCode+" "+o.text;zi.error(s)},get_device_id:o=>o,capture_performance:void 0,name:"posthog",bootstrap:{},disable_compression:!1,session_idle_timeout_seconds:1800,person_profiles:"identified_only",before_send:void 0,request_queue_config:{flush_interval_ms:KWe},error_tracking:{},_onCapture:YWe}},MBt=n=>{var r={};mi(n.process_person)||(r.person_profiles=n.process_person),mi(n.xhr_headers)||(r.request_headers=n.xhr_headers),mi(n.cookie_name)||(r.persistence_name=n.cookie_name),mi(n.disable_cookie)||(r.disable_persistence=n.disable_cookie),mi(n.store_google)||(r.save_campaign_params=n.store_google),mi(n.verbose)||(r.debug=n.verbose);var o=pp({},r,n);return kf(n.property_blacklist)&&(mi(n.property_denylist)?o.property_denylist=n.property_blacklist:kf(n.property_denylist)?o.property_denylist=[...n.property_blacklist,...n.property_denylist]:zi.error("Invalid value for property_denylist config: "+n.property_denylist)),o};class HDr{constructor(){this.__forceAllowLocalhost=!1}get ps(){return this.__forceAllowLocalhost}set ps(r){zi.error("WebPerformanceObserver is deprecated and has no impact on network capture. Use `_forceAllowLocalhostNetworkCapture` on `posthog.sessionRecording`"),this.__forceAllowLocalhost=r}}class kEe{get decideEndpointWasHit(){var r,o;return(r=(o=this.featureFlags)==null?void 0:o.hasLoadedFlags)!==null&&r!==void 0&&r}get flagsEndpointWasHit(){var r,o;return(r=(o=this.featureFlags)==null?void 0:o.hasLoadedFlags)!==null&&r!==void 0&&r}constructor(){this.webPerformance=new HDr,this.gs=!1,this.version=I4.LIB_VERSION,this._s=new Zun,this._calculate_event_properties=this.calculateEventProperties.bind(this),this.config=DBt(),this.SentryIntegration=qLr,this.sentryIntegration=r=>function(o,s){var c=Aun(o,s);return{name:Cun,processEvent:d=>c(d)}}(this,r),this.__request_queue=[],this.__loaded=!1,this.analyticsDefaultEndpoint="/e/",this.bs=!1,this.ys=null,this.ws=null,this.Ss=null,this.featureFlags=new EDr(this),this.toolbar=new WLr(this),this.scrollManager=new NDr(this),this.pageViewManager=new QLr(this),this.surveys=new TDr(this),this.experiments=new H1(this),this.exceptions=new iDr(this),this.rateLimiter=new xDr(this),this.requestRouter=new $Dr(this),this.consent=new tLr(this),this.externalIntegrations=new zDr(this),this.people={set:(r,o,s)=>{var c=Ih(r)?{[r]:o}:r;this.setPersonProperties(c),s?.({})},set_once:(r,o,s)=>{var c=Ih(r)?{[r]:o}:r;this.setPersonProperties(void 0,c),s?.({})}},this.on("eventCaptured",r=>zi.info('send "'+r?.event+'"',r))}init(r,o,s){if(s&&s!==aW){var c,d=(c=mie[s])!==null&&c!==void 0?c:new kEe;return d._init(r,o,s),mie[s]=d,mie[aW][s]=d,d}return this._init(r,o,s)}_init(r,o,s){var c,d;if(o===void 0&&(o={}),mi(r)||_We(r))return zi.critical("PostHog was initialized without a token. This likely indicates a misconfiguration. Please check the first argument passed to posthog.init()"),this;if(this.__loaded)return zi.warn("You have already initialized PostHog! Re-initializing is a no-op"),this;this.__loaded=!0,this.config={},this.$s=o,this.ks=[],o.person_profiles&&(this.ws=o.person_profiles),this.set_config(pp({},DBt(o.defaults),MBt(o),{name:s,token:r})),this.config.on_xhr_error&&zi.error("on_xhr_error is deprecated. Use on_request_error instead"),this.compression=o.disable_compression?void 0:N4.GZipJS;var h=this.xs();this.persistence=new uFe(this.config,h),this.sessionPersistence=this.config.persistence==="sessionStorage"||this.config.persistence==="memory"?this.persistence:new uFe(Ni({},this.config,{persistence:"sessionStorage"}),h);var b=Ni({},this.persistence.props),y=Ni({},this.sessionPersistence.props);if(this.register({$initialization_time:new Date().toISOString()}),this.Es=new ADr(G=>this.Is(G),this.config.request_queue_config),this.Ps=new ODr(this),this.__request_queue=[],this.config.__preview_experimental_cookieless_mode||(this.sessionManager=new LDr(this),this.sessionPropsManager=new PDr(this,this.sessionManager,this.persistence)),new YLr(this).startIfEnabledOrStop(),this.siteApps=new MDr(this),(c=this.siteApps)==null||c.init(),this.config.__preview_experimental_cookieless_mode||(this.sessionRecording=new HLr(this),this.sessionRecording.startIfEnabledOrStop()),this.config.disable_scroll_properties||this.scrollManager.startMeasuringScrollPosition(),this.autocapture=new WPr(this),this.autocapture.startIfEnabled(),this.surveys.loadIfEnabled(),this.heatmaps=new ZLr(this),this.heatmaps.startIfEnabled(),this.webVitalsAutocapture=new XLr(this),this.exceptionObserver=new iLr(this),this.exceptionObserver.startIfEnabled(),this.deadClicksAutocapture=new fun(this,rLr),this.deadClicksAutocapture.startIfEnabled(),this.historyAutocapture=new vLr(this),this.historyAutocapture.startIfEnabled(),I4.DEBUG=I4.DEBUG||this.config.debug,I4.DEBUG&&zi.info("Starting in debug mode",{this:this,config:o,thisC:Ni({},this.config),p:b,s:y}),((d=o.bootstrap)==null?void 0:d.distinctID)!==void 0){var E,k,x=this.config.get_device_id(uM()),O=(E=o.bootstrap)!=null&&E.isIdentifiedID?x:o.bootstrap.distinctID;this.persistence.set_property(C4,(k=o.bootstrap)!=null&&k.isIdentifiedID?"identified":"anonymous"),this.register({distinct_id:o.bootstrap.distinctID,$device_id:O})}if(this.Rs()){var P,D,$=Object.keys(((P=o.bootstrap)==null?void 0:P.featureFlags)||{}).filter(G=>{var K;return!((K=o.bootstrap)==null||(K=K.featureFlags)==null||!K[G])}).reduce((G,K)=>{var X;return G[K]=((X=o.bootstrap)==null||(X=X.featureFlags)==null?void 0:X[K])||!1,G},{}),U=Object.keys(((D=o.bootstrap)==null?void 0:D.featureFlagPayloads)||{}).filter(G=>$[G]).reduce((G,K)=>{var X,ae;return(X=o.bootstrap)!=null&&(X=X.featureFlagPayloads)!=null&&X[K]&&(G[K]=(ae=o.bootstrap)==null||(ae=ae.featureFlagPayloads)==null?void 0:ae[K]),G},{});this.featureFlags.receivedFeatureFlags({featureFlags:$,featureFlagPayloads:U})}if(this.config.__preview_experimental_cookieless_mode)this.register_once({distinct_id:C0e,$device_id:null},"");else if(!this.get_distinct_id()){var B=this.config.get_device_id(uM());this.register_once({distinct_id:B,$device_id:B},""),this.persistence.set_property(C4,"anonymous")}return th(kn,"onpagehide"in self?"pagehide":"unload",this._handle_unload.bind(this),{passive:!1}),this.toolbar.maybeLoadToolbar(),o.segment?GLr(this,()=>this.Ts()):this.Ts(),j_(this.config._onCapture)&&this.config._onCapture!==YWe&&(zi.warn("onCapture is deprecated. Please use `before_send` instead"),this.on("eventCaptured",G=>this.config._onCapture(G.event,G))),this.config.ip&&zi.warn('The `ip` config option has NO EFFECT AT ALL and has been deprecated. Use a custom transformation or "Discard IP data" project setting instead. See https://posthog.com/tutorials/web-redact-properties#hiding-customer-ip-address for more information.'),this}Ie(r){var o,s,c,d,h,b,y,E;if(!Di||!Di.body)return zi.info("document not ready yet, trying again in 500 milliseconds..."),void setTimeout(()=>{this.Ie(r)},500);this.compression=void 0,r.supportedCompression&&!this.config.disable_compression&&(this.compression=pc(r.supportedCompression,N4.GZipJS)?N4.GZipJS:pc(r.supportedCompression,N4.Base64)?N4.Base64:void 0),(o=r.analytics)!=null&&o.endpoint&&(this.analyticsDefaultEndpoint=r.analytics.endpoint),this.set_config({person_profiles:this.ws?this.ws:"identified_only"}),(s=this.siteApps)==null||s.onRemoteConfig(r),(c=this.sessionRecording)==null||c.onRemoteConfig(r),(d=this.autocapture)==null||d.onRemoteConfig(r),(h=this.heatmaps)==null||h.onRemoteConfig(r),this.surveys.onRemoteConfig(r),(b=this.webVitalsAutocapture)==null||b.onRemoteConfig(r),(y=this.exceptionObserver)==null||y.onRemoteConfig(r),this.exceptions.onRemoteConfig(r),(E=this.deadClicksAutocapture)==null||E.onRemoteConfig(r)}Ts(){try{this.config.loaded(this)}catch(r){zi.critical("`loaded` function failed",r)}this.Ms(),this.config.capture_pageview&&setTimeout(()=>{this.consent.isOptedIn()&&this.Cs()},1),new CDr(this).load(),this.featureFlags.flags()}Ms(){var r;this.has_opted_out_capturing()||this.config.request_batching&&((r=this.Es)==null||r.enable())}_dom_loaded(){this.has_opted_out_capturing()||vM(this.__request_queue,r=>this.Is(r)),this.__request_queue=[],this.Ms()}_handle_unload(){var r,o;this.config.request_batching?(this.Fs()&&this.capture("$pageleave"),(r=this.Es)==null||r.unload(),(o=this.Ps)==null||o.unload()):this.Fs()&&this.capture("$pageleave",null,{transport:"sendBeacon"})}Ee(r){this.__loaded&&(eln?this.__request_queue.push(r):this.rateLimiter.isServerRateLimited(r.batchKey)||(r.transport=r.transport||this.config.api_transport,r.url=Uwe(r.url,{ip:this.config.ip?1:0}),r.headers=Ni({},this.config.request_headers),r.compression=r.compression==="best-available"?this.compression:r.compression,r.fetchOptions=r.fetchOptions||this.config.fetch_options,(o=>{var s,c,d,h=Ni({},o);h.timeout=h.timeout||6e4,h.url=Uwe(h.url,{_:new Date().getTime().toString(),ver:I4.LIB_VERSION,compression:h.compression});var b=(s=h.transport)!==null&&s!==void 0?s:"fetch",y=(c=(d=Xcn(gie,E=>E.transport===b))==null?void 0:d.method)!==null&&c!==void 0?c:gie[0].method;if(!y)throw new Error("No available transport method");y(h)})(Ni({},r,{callback:o=>{var s,c;this.rateLimiter.checkForLimiting(o),o.statusCode>=400&&((s=(c=this.config).on_request_error)==null||s.call(c,o)),r.callback==null||r.callback(o)}}))))}Is(r){this.Ps?this.Ps.retriableRequest(r):this.Ee(r)}_execute_array(r){var o,s=[],c=[],d=[];vM(r,b=>{b&&(o=b[0],kf(o)?d.push(b):j_(b)?b.call(this):kf(b)&&o==="alias"?s.push(b):kf(b)&&o.indexOf("capture")!==-1&&j_(this[o])?d.push(b):c.push(b))});var h=function(b,y){vM(b,function(E){if(kf(E[0])){var k=y;Bl(E,function(x){k=k[x[0]].apply(k,x.slice(1))})}else this[E[0]].apply(this,E.slice(1))},y)};h(s,this),h(c,this),h(d,this)}Rs(){var r,o;return((r=this.config.bootstrap)==null?void 0:r.featureFlags)&&Object.keys((o=this.config.bootstrap)==null?void 0:o.featureFlags).length>0||!1}push(r){this._execute_array([r])}capture(r,o,s){var c;if(this.__loaded&&this.persistence&&this.sessionPersistence&&this.Es){if(!this.consent.isOptedOut())if(!mi(r)&&Ih(r)){if(this.config.opt_out_useragent_filter||!this._is_bot()){var d=s!=null&&s.skip_client_rate_limiting?void 0:this.rateLimiter.clientRateLimitContext();if(d==null||!d.isRateLimited){o!=null&&o.$current_url&&!Ih(o?.$current_url)&&(zi.error("Invalid `$current_url` property provided to `posthog.capture`. Input must be a string. Ignoring provided value."),o==null||delete o.$current_url),this.sessionPersistence.update_search_keyword(),this.config.save_campaign_params&&this.sessionPersistence.update_campaign_params(),this.config.save_referrer&&this.sessionPersistence.update_referrer_info(),(this.config.save_campaign_params||this.config.save_referrer)&&this.persistence.set_initial_person_info();var h=new Date,b=s?.timestamp||h,y=uM(),E={uuid:y,event:r,properties:this.calculateEventProperties(r,o||{},b,y)};d&&(E.properties.$lib_rate_limit_remaining_tokens=d.remainingTokens),s?.$set&&(E.$set=s?.$set);var k,x,O=this.Os(s?.$set_once);if(O&&(E.$set_once=O),(E=NPr(E,s!=null&&s._noTruncate?null:this.config.properties_string_max_length)).timestamp=b,mi(s?.timestamp)||(E.properties.$event_time_override_provided=!0,E.properties.$event_time_override_system_time=h),r===lFe.DISMISSED||r===lFe.SENT){var P=o?.[OBt.SURVEY_ID],D=o?.[OBt.SURVEY_ITERATION];localStorage.setItem((x=""+WWe+(k={id:P,current_iteration:D}).id,k.current_iteration&&k.current_iteration>0&&(x=""+WWe+k.id+"_"+k.current_iteration),x),"true"),E.$set=Ni({},E.$set,{[_Dr({id:P,current_iteration:D},r===lFe.SENT?"responded":"dismissed")]:!0})}var $=Ni({},E.properties.$set,E.$set);if(fW($)||this.setPersonPropertiesForFlags($),!Gu(this.config.before_send)){var U=this.As(E);if(!U)return;E=U}this._s.emit("eventCaptured",E);var B={method:"POST",url:(c=s?._url)!==null&&c!==void 0?c:this.requestRouter.endpointFor("api",this.analyticsDefaultEndpoint),data:E,compression:"best-available",batchKey:s?._batchKey};return!this.config.request_batching||s&&(s==null||!s._batchKey)||s!=null&&s.send_instantly?this.Is(B):this.Es.enqueue(B),E}zi.critical("This capture call is ignored due to client rate limiting.")}}else zi.error("No event name provided to posthog.capture")}else zi.uninitializedWarning("posthog.capture")}Ge(r){return this.on("eventCaptured",o=>r(o.event,o))}calculateEventProperties(r,o,s,c,d){if(s=s||new Date,!this.persistence||!this.sessionPersistence)return o;var h=d?void 0:this.persistence.remove_event_timer(r),b=Ni({},o);if(b.token=this.config.token,b.$config_defaults=this.config.defaults,this.config.__preview_experimental_cookieless_mode&&(b.$cookieless_mode=!0),r==="$snapshot"){var y=Ni({},this.persistence.properties(),this.sessionPersistence.properties());return b.distinct_id=y.distinct_id,(!Ih(b.distinct_id)&&!G1(b.distinct_id)||_We(b.distinct_id))&&zi.error("Invalid distinct_id for replay event. This indicates a bug in your implementation"),b}var E,k=vDr(this.config.mask_personal_data_properties,this.config.custom_personal_data_properties);if(this.sessionManager){var{sessionId:x,windowId:O}=this.sessionManager.checkAndGetSessionAndWindowId(d,s.getTime());b.$session_id=x,b.$window_id=O}this.sessionPropsManager&&pp(b,this.sessionPropsManager.getSessionProps());try{var P;this.sessionRecording&&pp(b,this.sessionRecording.sdkDebugProperties),b.$sdk_debug_retry_queue_size=(P=this.Ps)==null?void 0:P.length}catch(B){b.$sdk_debug_error_capturing_properties=String(B)}if(this.requestRouter.region===Kre.CUSTOM&&(b.$lib_custom_api_host=this.config.api_host),E=r!=="$pageview"||d?r!=="$pageleave"||d?this.pageViewManager.doEvent():this.pageViewManager.doPageLeave(s):this.pageViewManager.doPageView(s,c),b=pp(b,E),r==="$pageview"&&Di&&(b.title=Di.title),!mi(h)){var D=s.getTime()-h;b.$duration=parseFloat((D/1e3).toFixed(3))}Gw&&this.config.opt_out_useragent_filter&&(b.$browser_type=this._is_bot()?"bot":"browser"),(b=pp({},k,this.persistence.properties(),this.sessionPersistence.properties(),b)).$is_identified=this._isIdentified(),kf(this.config.property_denylist)?Bl(this.config.property_denylist,function(B){delete b[B]}):zi.error("Invalid value for property_denylist config: "+this.config.property_denylist+" or property_blacklist config: "+this.config.property_blacklist);var $=this.config.sanitize_properties;$&&(zi.error("sanitize_properties is deprecated. Use before_send instead"),b=$(b,r));var U=this.Ds();return b.$process_person_profile=U,U&&!d&&this.Ls("_calculate_event_properties"),b}Os(r){var o;if(!this.persistence||!this.Ds()||this.gs)return r;var s=this.persistence.get_initial_props(),c=(o=this.sessionPropsManager)==null?void 0:o.getSetOnceProps(),d=pp({},s,c||{},r||{}),h=this.config.sanitize_properties;return h&&(zi.error("sanitize_properties is deprecated. Use before_send instead"),d=h(d,"$set_once")),this.gs=!0,fW(d)?void 0:d}register(r,o){var s;(s=this.persistence)==null||s.register(r,o)}register_once(r,o,s){var c;(c=this.persistence)==null||c.register_once(r,o,s)}register_for_session(r){var o;(o=this.sessionPersistence)==null||o.register(r)}unregister(r){var o;(o=this.persistence)==null||o.unregister(r)}unregister_for_session(r){var o;(o=this.sessionPersistence)==null||o.unregister(r)}js(r,o){this.register({[r]:o})}getFeatureFlag(r,o){return this.featureFlags.getFeatureFlag(r,o)}getFeatureFlagPayload(r){var o=this.featureFlags.getFeatureFlagPayload(r);try{return JSON.parse(o)}catch{return o}}isFeatureEnabled(r,o){return this.featureFlags.isFeatureEnabled(r,o)}reloadFeatureFlags(){this.featureFlags.reloadFeatureFlags()}updateEarlyAccessFeatureEnrollment(r,o){this.featureFlags.updateEarlyAccessFeatureEnrollment(r,o)}getEarlyAccessFeatures(r,o,s){return o===void 0&&(o=!1),this.featureFlags.getEarlyAccessFeatures(r,o,s)}on(r,o){return this._s.on(r,o)}onFeatureFlags(r){return this.featureFlags.onFeatureFlags(r)}onSurveysLoaded(r){return this.surveys.onSurveysLoaded(r)}onSessionId(r){var o,s;return(o=(s=this.sessionManager)==null?void 0:s.onSessionId(r))!==null&&o!==void 0?o:()=>{}}getSurveys(r,o){o===void 0&&(o=!1),this.surveys.getSurveys(r,o)}getActiveMatchingSurveys(r,o){o===void 0&&(o=!1),this.surveys.getActiveMatchingSurveys(r,o)}renderSurvey(r,o){this.surveys.renderSurvey(r,o)}canRenderSurvey(r){return this.surveys.canRenderSurvey(r)}canRenderSurveyAsync(r,o){return o===void 0&&(o=!1),this.surveys.canRenderSurveyAsync(r,o)}identify(r,o,s){if(!this.__loaded||!this.persistence)return zi.uninitializedWarning("posthog.identify");if(G1(r)&&(r=r.toString(),zi.warn("The first argument to posthog.identify was a number, but it should be a string. It has been converted to a string.")),r)if(["distinct_id","distinctid"].includes(r.toLowerCase()))zi.critical('The string "'+r+'" was set in posthog.identify which indicates an error. This ID should be unique to the user and not a hardcoded string.');else if(r!==C0e){if(this.Ls("posthog.identify")){var c=this.get_distinct_id();if(this.register({$user_id:r}),!this.get_property("$device_id")){var d=c;this.register_once({$had_persisted_distinct_id:!0,$device_id:d},"")}r!==c&&r!==this.get_property(Ure)&&(this.unregister(Ure),this.register({distinct_id:r}));var h=(this.persistence.get_property(C4)||"anonymous")==="anonymous";r!==c&&h?(this.persistence.set_property(C4,"identified"),this.setPersonPropertiesForFlags(Ni({},s||{},o||{}),!1),this.capture("$identify",{distinct_id:r,$anon_distinct_id:c},{$set:o||{},$set_once:s||{}}),this.Ss=SBt(r,o,s),this.featureFlags.setAnonymousDistinctId(c)):(o||s)&&this.setPersonProperties(o,s),r!==c&&(this.reloadFeatureFlags(),this.unregister(Nwe))}}else zi.critical('The string "'+C0e+'" was set in posthog.identify which indicates an error. This ID is only used as a sentinel value.');else zi.error("Unique user id has not been set in posthog.identify")}setPersonProperties(r,o){if((r||o)&&this.Ls("posthog.setPersonProperties")){var s=SBt(this.get_distinct_id(),r,o);this.Ss!==s?(this.setPersonPropertiesForFlags(Ni({},o||{},r||{})),this.capture("$set",{$set:r||{},$set_once:o||{}}),this.Ss=s):zi.info("A duplicate setPersonProperties call was made with the same properties. It has been ignored.")}}group(r,o,s){if(r&&o){if(this.Ls("posthog.group")){var c=this.getGroups();c[r]!==o&&this.resetGroupPropertiesForFlags(r),this.register({$groups:Ni({},c,{[r]:o})}),s&&(this.capture("$groupidentify",{$group_type:r,$group_key:o,$group_set:s}),this.setGroupPropertiesForFlags({[r]:s})),c[r]===o||s||this.reloadFeatureFlags()}}else zi.error("posthog.group requires a group type and group key")}resetGroups(){this.register({$groups:{}}),this.resetGroupPropertiesForFlags(),this.reloadFeatureFlags()}setPersonPropertiesForFlags(r,o){o===void 0&&(o=!0),this.featureFlags.setPersonPropertiesForFlags(r,o)}resetPersonPropertiesForFlags(){this.featureFlags.resetPersonPropertiesForFlags()}setGroupPropertiesForFlags(r,o){o===void 0&&(o=!0),this.Ls("posthog.setGroupPropertiesForFlags")&&this.featureFlags.setGroupPropertiesForFlags(r,o)}resetGroupPropertiesForFlags(r){this.featureFlags.resetGroupPropertiesForFlags(r)}reset(r){var o,s,c,d;if(zi.info("reset"),!this.__loaded)return zi.uninitializedWarning("posthog.reset");var h=this.get_property("$device_id");if(this.consent.reset(),(o=this.persistence)==null||o.clear(),(s=this.sessionPersistence)==null||s.clear(),this.surveys.reset(),this.featureFlags.reset(),(c=this.persistence)==null||c.set_property(C4,"anonymous"),(d=this.sessionManager)==null||d.resetSessionId(),this.Ss=null,this.config.__preview_experimental_cookieless_mode)this.register_once({distinct_id:C0e,$device_id:null},"");else{var b=this.config.get_device_id(uM());this.register_once({distinct_id:b,$device_id:r?b:h},"")}this.register({$last_posthog_reset:new Date().toISOString()},1)}get_distinct_id(){return this.get_property("distinct_id")}getGroups(){return this.get_property("$groups")||{}}get_session_id(){var r,o;return(r=(o=this.sessionManager)==null?void 0:o.checkAndGetSessionAndWindowId(!0).sessionId)!==null&&r!==void 0?r:""}get_session_replay_url(r){if(!this.sessionManager)return"";var{sessionId:o,sessionStartTimestamp:s}=this.sessionManager.checkAndGetSessionAndWindowId(!0),c=this.requestRouter.endpointFor("ui","/project/"+this.config.token+"/replay/"+o);if(r!=null&&r.withTimestamp&&s){var d,h=(d=r.timestampLookBack)!==null&&d!==void 0?d:10;if(!s)return c;c+="?t="+Math.max(Math.floor((new Date().getTime()-s)/1e3)-h,0)}return c}alias(r,o){return r===this.get_property(Jcn)?(zi.critical("Attempting to create alias for existing People user - aborting."),-2):this.Ls("posthog.alias")?(mi(o)&&(o=this.get_distinct_id()),r!==o?(this.js(Ure,r),this.capture("$create_alias",{alias:r,distinct_id:o})):(zi.warn("alias matches current distinct_id - skipping api call."),this.identify(r),-1)):void 0}set_config(r){var o=Ni({},this.config);if(Ah(r)){var s,c,d,h,b;pp(this.config,MBt(r));var y=this.xs();(s=this.persistence)==null||s.update_config(this.config,o,y),this.sessionPersistence=this.config.persistence==="sessionStorage"||this.config.persistence==="memory"?this.persistence:new uFe(Ni({},this.config,{persistence:"sessionStorage"}),y),Gg.O()&&Gg.D("ph_debug")==="true"&&(this.config.debug=!0),this.config.debug&&(I4.DEBUG=!0,zi.info("set_config",{config:r,oldConfig:o,newConfig:Ni({},this.config)})),(c=this.sessionRecording)==null||c.startIfEnabledOrStop(),(d=this.autocapture)==null||d.startIfEnabled(),(h=this.heatmaps)==null||h.startIfEnabled(),this.surveys.loadIfEnabled(),this.Ns(),(b=this.externalIntegrations)==null||b.startIfEnabledOrStop()}}startSessionRecording(r){var o=r===!0,s={sampling:o||!(r==null||!r.sampling),linked_flag:o||!(r==null||!r.linked_flag),url_trigger:o||!(r==null||!r.url_trigger),event_trigger:o||!(r==null||!r.event_trigger)};if(Object.values(s).some(Boolean)){var c,d,h,b,y;(c=this.sessionManager)==null||c.checkAndGetSessionAndWindowId(),s.sampling&&((d=this.sessionRecording)==null||d.overrideSampling()),s.linked_flag&&((h=this.sessionRecording)==null||h.overrideLinkedFlag()),s.url_trigger&&((b=this.sessionRecording)==null||b.overrideTrigger("url")),s.event_trigger&&((y=this.sessionRecording)==null||y.overrideTrigger("event"))}this.set_config({disable_session_recording:!1})}stopSessionRecording(){this.set_config({disable_session_recording:!0})}sessionRecordingStarted(){var r;return!((r=this.sessionRecording)==null||!r.started)}captureException(r,o){var s=new Error("PostHog syntheticException");return this.exceptions.sendExceptionEvent(Ni({},bLr((c=>c instanceof Error)(r)?{error:r,event:r.message}:{event:r},{syntheticException:s}),o))}loadToolbar(r){return this.toolbar.loadToolbar(r)}get_property(r){var o;return(o=this.persistence)==null?void 0:o.props[r]}getSessionProperty(r){var o;return(o=this.sessionPersistence)==null?void 0:o.props[r]}toString(){var r,o=(r=this.config.name)!==null&&r!==void 0?r:aW;return o!==aW&&(o=aW+"."+o),o}_isIdentified(){var r,o;return((r=this.persistence)==null?void 0:r.get_property(C4))==="identified"||((o=this.sessionPersistence)==null?void 0:o.get_property(C4))==="identified"}Ds(){var r,o;return!(this.config.person_profiles==="never"||this.config.person_profiles==="identified_only"&&!this._isIdentified()&&fW(this.getGroups())&&((r=this.persistence)==null||(r=r.props)==null||!r[Ure])&&((o=this.persistence)==null||(o=o.props)==null||!o[Pwe]))}Fs(){return this.config.capture_pageleave===!0||this.config.capture_pageleave==="if_capture_pageview"&&(this.config.capture_pageview===!0||this.config.capture_pageview==="history_change")}createPersonProfile(){this.Ds()||this.Ls("posthog.createPersonProfile")&&this.setPersonProperties({},{})}Ls(r){return this.config.person_profiles==="never"?(zi.error(r+' was called, but process_person is set to "never". This call will be ignored.'),!1):(this.js(Pwe,!0),!0)}xs(){var r=this.consent.isOptedOut(),o=this.config.opt_out_persistence_by_default;return this.config.disable_persistence||r&&!!o}Ns(){var r,o,s,c,d=this.xs();return((r=this.persistence)==null?void 0:r.Fe)!==d&&((s=this.persistence)==null||s.set_disabled(d)),((o=this.sessionPersistence)==null?void 0:o.Fe)!==d&&((c=this.sessionPersistence)==null||c.set_disabled(d)),d}opt_in_capturing(r){var o;this.consent.optInOut(!0),this.Ns(),(mi(r?.captureEventName)||r!=null&&r.captureEventName)&&this.capture((o=r?.captureEventName)!==null&&o!==void 0?o:"$opt_in",r?.captureProperties,{send_instantly:!0}),this.config.capture_pageview&&this.Cs()}opt_out_capturing(){this.consent.optInOut(!1),this.Ns()}has_opted_in_capturing(){return this.consent.isOptedIn()}has_opted_out_capturing(){return this.consent.isOptedOut()}clear_opt_in_out_capturing(){this.consent.reset(),this.Ns()}_is_bot(){return c2?Qun(c2,this.config.custom_blocked_useragents):void 0}Cs(){Di&&(Di.visibilityState==="visible"?this.bs||(this.bs=!0,this.capture("$pageview",{title:Di.title},{send_instantly:!0}),this.ys&&(Di.removeEventListener("visibilitychange",this.ys),this.ys=null)):this.ys||(this.ys=this.Cs.bind(this),th(Di,"visibilitychange",this.ys)))}debug(r){r===!1?(kn?.console.log("You've disabled debug mode."),localStorage&&localStorage.removeItem("ph_debug"),this.set_config({debug:!1})):(kn?.console.log("You're now in debug mode. All calls to PostHog will be logged in your console.\nYou can disable this with `posthog.debug(false)`."),localStorage&&localStorage.setItem("ph_debug","true"),this.set_config({debug:!0}))}I(){var r,o,s,c,d,h,b,y=this.$s||{};return"advanced_disable_flags"in y?!!y.advanced_disable_flags:this.config.advanced_disable_flags!==!1?!!this.config.advanced_disable_flags:this.config.advanced_disable_decide===!0?(zi.warn("Config field 'advanced_disable_decide' is deprecated. Please use 'advanced_disable_flags' instead. The old field will be removed in a future major version."),!0):(s="advanced_disable_decide",c=!1,d=zi,h=(o="advanced_disable_flags")in(r=y)&&!mi(r[o]),b=s in r&&!mi(r[s]),h?r[o]:b?(d&&d.warn("Config field '"+s+"' is deprecated. Please use '"+o+"' instead. The old field will be removed in a future major version."),r[s]):c)}As(r){if(Gu(this.config.before_send))return r;var o=kf(this.config.before_send)?this.config.before_send:[this.config.before_send],s=r;for(var c of o){if(s=c(s),Gu(s)){var d="Event '"+r.event+"' was rejected in beforeSend function";return APr(r.event)?zi.warn(d+". This can cause unexpected behavior."):zi.info(d),null}s.properties&&!fW(s.properties)||zi.warn("Event '"+r.event+"' has no properties after beforeSend function, this is likely an error.")}return s}getPageViewId(){var r;return(r=this.pageViewManager.ce)==null?void 0:r.pageViewId}captureTraceFeedback(r,o){this.capture("$ai_feedback",{$ai_trace_id:String(r),$ai_feedback_text:o})}captureTraceMetric(r,o,s){this.capture("$ai_metric",{$ai_trace_id:String(r),$ai_metric_name:o,$ai_metric_value:String(s)})}}(function(n,r){for(var o=0;o<r.length;o++)n.prototype[r[o]]=OPr(n.prototype[r[o]])})(kEe,["identify"]);var jBt,qw=(jBt=mie[aW]=new kEe,function(){function n(){n.done||(n.done=!0,eln=!1,Bl(mie,function(r){r._dom_loaded()}))}Di!=null&&Di.addEventListener?Di.readyState==="complete"?n():th(Di,"DOMContentLoaded",n,{capture:!1}):kn&&zi.error("Browser doesn't support `document.addEventListener` so PostHog couldn't be initialized")}(),jBt),tln=Q.createContext({client:qw});function nln(n,r,o){if(o===void 0&&(o=new WeakMap),n===r)return!0;if(typeof n!="object"||n===null||typeof r!="object"||r===null)return!1;if(o.has(n)&&o.get(n)===r)return!0;o.set(n,r);var s=Object.keys(n),c=Object.keys(r);if(s.length!==c.length)return!1;for(var d=0,h=s;d<h.length;d++){var b=h[d];if(!c.includes(b)||!nln(n[b],r[b],o))return!1}return!0}function GDr(n){var r=n.children,o=n.client,s=n.apiKey,c=n.options,d=Q.useRef(null),h=Q.useMemo(function(){return o?(s&&console.warn("[PostHog.js] You have provided both `client` and `apiKey` to `PostHogProvider`. `apiKey` will be ignored in favour of `client`."),c&&console.warn("[PostHog.js] You have provided both `client` and `options` to `PostHogProvider`. `options` will be ignored in favour of `client`."),o):(s||console.warn("[PostHog.js] No `apiKey` or `client` were provided to `PostHogProvider`. Using default global `window.posthog` instance. You must initialize it manually. This is not recommended behavior."),qw)},[o,s,JSON.stringify(c)]);return Q.useEffect(function(){if(!o){var b=d.current;b?(s!==b.apiKey&&console.warn("[PostHog.js] You have provided a different `apiKey` to `PostHogProvider` than the one that was already initialized. This is not supported by our provider and we'll keep using the previous key. If you need to toggle between API Keys you need to control the `client` yourself and pass it in as a prop rather than an `apiKey` prop."),c&&!nln(c,b.options)&&qw.set_config(c),d.current={apiKey:s,options:c??{}}):(qw.__loaded&&console.warn("[PostHog.js] `posthog` was already loaded elsewhere. This may cause issues."),qw.init(s,c),d.current={apiKey:s,options:c??{}})}},[o,s,JSON.stringify(c)]),Un.createElement(tln.Provider,{value:{client:h}},r)}var XWe=function(n,r){return XWe=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(o,s){o.__proto__=s}||function(o,s){for(var c in s)Object.prototype.hasOwnProperty.call(s,c)&&(o[c]=s[c])},XWe(n,r)};function qDr(n,r){if(typeof r!="function"&&r!==null)throw new TypeError("Class extends value "+String(r)+" is not a constructor or null");XWe(n,r);function o(){this.constructor=n}n.prototype=r===null?Object.create(r):(o.prototype=r.prototype,new o)}var dFe=function(n){return typeof n=="function"},VDr={componentStack:null,error:null},WDr={INVALID_FALLBACK:"[PostHog.js][PostHogErrorBoundary] Invalid fallback prop, provide a valid React element or a function that returns a valid React element."};(function(n){qDr(r,n);function r(o){var s=n.call(this,o)||this;return s.state=VDr,s}return r.prototype.componentDidCatch=function(o,s){var c=s.componentStack,d=this.props.additionalProperties;this.setState({error:o,componentStack:c});var h;dFe(d)?h=d(o):typeof d=="object"&&(h=d);var b=this.context.client;b.captureException(o,h)},r.prototype.render=function(){var o=this.props,s=o.children,c=o.fallback,d=this.state;if(d.componentStack==null)return dFe(s)?s():s;var h=dFe(c)?Un.createElement(c,{error:d.error,componentStack:d.componentStack}):c;return Un.isValidElement(h)?h:(console.warn(WDr.INVALID_FALLBACK),Un.createElement(Un.Fragment,null))},r.contextType=tln,r})(Un.Component);const KDr="https://eu.i.posthog.com",rln=()=>{const n=Q.useRef(!1),r=Q.useCallback(()=>{n.current||(n.current=!0,qw.init("{{POSTHOG_KEY}}",{api_host:KDr,debug:!1,capture_pageview:!1,autocapture:!1,disable_session_recording:!0,disable_cookie:!0,disable_persistence:!0,persistence:"memory",respect_dnt:!0}))},[]),o=Q.useCallback(s=>{s?(qw.stopSessionRecording?.(),qw.set_config({capture_pageview:!1,autocapture:!1,disable_session_recording:!0,disable_cookie:!0,disable_persistence:!0,persistence:"memory"}),qw.reset()):(qw.set_config({capture_pageview:!0,disable_cookie:!1,disable_persistence:!1,persistence:"localStorage+cookie",autocapture:!0}),qw.startSessionRecording?.(),qw.capture("$pageview"))},[]);return{posthog:qw,configure:o,init:r}};function YDr({children:n}){const{posthog:r,init:o}=rln();return Q.useEffect(()=>{o()},[o]),H.jsx(GDr,{client:r,children:n})}/*!
* CookieConsent 3.1.0
* https://github.com/orestbida/cookieconsent
* Author Orest Bida
* Released under the MIT License
*/const iln="opt-in",Yre="opt-out",oln="show--consent",aln="show--preferences",$Bt="disable--interaction",Hwe="data-category",lu="div",Lx="button",Dx="aria-hidden",bie="btn-group",q1="click",$W="data-role",oQe="consentModal",aQe="preferencesModal";class XDr{constructor(){this.t={mode:iln,revision:0,autoShow:!0,lazyHtmlGeneration:!0,autoClearCookies:!0,manageScriptTags:!0,hideFromBots:!0,cookie:{name:"cc_cookie",expiresAfterDays:182,domain:"",path:"/",secure:!0,sameSite:"Lax"}},this.o={i:{},l:"",_:{},u:{},p:{},m:[],v:!1,h:null,C:null,S:null,M:"",D:!0,T:!1,k:!1,A:!1,N:!1,H:[],V:!1,I:!0,L:[],j:!1,F:"",P:!1,O:[],R:[],B:[],$:[],G:!1,J:!1,U:!1,q:[],K:[],W:[],X:{},Y:{},Z:{},ee:{},te:{},oe:[]},this.ne={ae:{},se:{}},this.ce={},this.re={ie:"cc:onFirstConsent",le:"cc:onConsent",de:"cc:onChange",fe:"cc:onModalShow",_e:"cc:onModalHide",ue:"cc:onModalReady"}}}const rr=new XDr,sln=(n,r)=>n.indexOf(r),Fl=(n,r)=>sln(n,r)!==-1,TEe=n=>Array.isArray(n),LK=n=>typeof n=="string",JWe=n=>!!n&&typeof n=="object"&&!TEe(n),fI=n=>typeof n=="function",HB=n=>Object.keys(n),cln=n=>Array.from(new Set(n)),Gwe=()=>document.activeElement,gW=n=>n.preventDefault(),sQe=(n,r)=>n.querySelectorAll(r),Ui=n=>{const r=document.createElement(n);return n===Lx&&(r.type=n),r},qa=(n,r,o)=>n.setAttribute(r,o),fFe=(n,r,o)=>{n.removeAttribute(o?"data-"+r:r)},Tve=(n,r,o)=>n.getAttribute(o?"data-"+r:r),$i=(n,r)=>n.appendChild(r),sd=(n,r)=>n.classList.add(r),rv=(n,r)=>sd(n,"cm__"+r),Wc=(n,r)=>sd(n,"pm__"+r),yM=(n,r)=>n.classList.remove(r),lM=n=>{if(typeof n!="object")return n;if(n instanceof Date)return new Date(n.getTime());let r=Array.isArray(n)?[]:{};for(let o in n){let s=n[o];r[o]=lM(s)}return r},FBt=(n,r)=>dispatchEvent(new CustomEvent(n,{detail:r})),eh=(n,r,o,s)=>{n.addEventListener(r,o),s&&rr.o.m.push({pe:n,ge:r,me:o})},JDr=()=>{const n=rr.t.cookie.expiresAfterDays;return fI(n)?n(rr.o.F):n},hFe=(n,r)=>{const o=n||[],s=r||[];return o.filter(c=>!Fl(s,c)).concat(s.filter(c=>!Fl(o,c)))},uln=n=>{rr.o.R=cln(n),rr.o.F=(()=>{let r="custom";const{R:o,O:s,B:c}=rr.o,d=o.length;return d===s.length?r="all":d===c.length&&(r="necessary"),r})()},lln=(n,r,o,s)=>{const c="accept-",{show:d,showPreferences:h,hide:b,hidePreferences:y,acceptCategory:E}=r,k=n||document,x=K=>sQe(k,`[data-cc="${K}"]`),O=(K,X)=>{gW(K),E(X),y(),b()},P=x("show-preferencesModal"),D=x("show-consentModal"),$=x(c+"all"),U=x(c+"necessary"),B=x(c+"custom"),G=rr.t.lazyHtmlGeneration;for(const K of P)qa(K,"aria-haspopup","dialog"),eh(K,q1,X=>{gW(X),h()}),G&&(eh(K,"mouseenter",X=>{gW(X),rr.o.N||o(r,s)},!0),eh(K,"focus",()=>{rr.o.N||o(r,s)}));for(let K of D)qa(K,"aria-haspopup","dialog"),eh(K,q1,X=>{gW(X),d(!0)},!0);for(let K of $)eh(K,q1,X=>{O(X,"all")},!0);for(let K of B)eh(K,q1,X=>{O(X)},!0);for(let K of U)eh(K,q1,X=>{O(X,[])},!0)},OI=(n,r)=>{n&&(r&&(n.tabIndex=-1),n.focus(),r&&n.removeAttribute("tabindex"))},dln=(n,r)=>{const o=s=>{s.target.removeEventListener("transitionend",o),s.propertyName==="opacity"&&getComputedStyle(n).opacity==="1"&&OI((c=>c===1?rr.ne.be:rr.ne.ve)(r))};eh(n,"transitionend",o)};let BBt;const fln=n=>{clearTimeout(BBt),n?sd(rr.ne.ye,$Bt):BBt=setTimeout(()=>{yM(rr.ne.ye,$Bt)},500)},ZDr=["M 19.5 4.5 L 4.5 19.5 M 4.5 4.501 L 19.5 19.5","M 3.572 13.406 L 8.281 18.115 L 20.428 5.885","M 21.999 6.94 L 11.639 17.18 L 2.001 6.82 "],loe=(n=0,r=1.5)=>`<svg viewBox="0 0 24 24" stroke-width="${r}"><path d="${ZDr[n]}"/></svg>`,hln=n=>{const r=rr.ne,o=rr.o;(s=>{const c=s===r.he,d=o.i.disablePageInteraction?r.ye:c?r.Ce:r.ye;eh(d,"keydown",h=>{if(h.key!=="Tab"||!(c?o.k&&!o.A:o.A))return;const b=Gwe(),y=c?o.q:o.K;y.length!==0&&(h.shiftKey?b!==y[0]&&s.contains(b)||(gW(h),OI(y[1])):b!==y[1]&&s.contains(b)||(gW(h),OI(y[0])))},!0)})(n)},QDr=["[href]",Lx,"input","details","[tabindex]"].map(n=>n+':not([tabindex="-1"])').join(","),pln=n=>{const{o:r,ne:o}=rr,s=(c,d)=>{const h=sQe(c,QDr);d[0]=h[0],d[1]=h[h.length-1]};n===1&&r.T&&s(o.he,r.q),n===2&&r.N&&s(o.we,r.K)},NI=(n,r,o)=>{const{de:s,le:c,ie:d,_e:h,ue:b,fe:y}=rr.ce,E=rr.re;if(r){const x={modalName:r};return n===E.fe?fI(y)&&y(x):n===E._e?fI(h)&&h(x):(x.modal=o,fI(b)&&b(x)),FBt(n,x)}const k={cookie:rr.o.p};n===E.ie?fI(d)&&d(lM(k)):n===E.le?fI(c)&&c(lM(k)):(k.changedCategories=rr.o.L,k.changedServices=rr.o.ee,fI(s)&&s(lM(k))),FBt(n,lM(k))},qwe=(n,r)=>{try{return n()}catch(o){return!r&&console.warn("CookieConsent:",o),!1}},ZWe=n=>{const{Y:r,ee:o,O:s,X:c,oe:d,p:h,L:b}=rr.o;for(const x of s){const O=o[x]||r[x]||[];for(const P of O){const D=c[x][P];if(!D)continue;const{onAccept:$,onReject:U}=D;!D.Se&&Fl(r[x],P)?(D.Se=!0,fI($)&&$()):D.Se&&!Fl(r[x],P)&&(D.Se=!1,fI(U)&&U())}}if(!rr.t.manageScriptTags)return;const y=d,E=n||h.categories||[],k=(x,O)=>{if(O>=x.length)return;const P=d[O];if(P.xe)return k(x,O+1);const D=P.Me,$=P.De,U=P.Te,B=Fl(E,$),G=!!U&&Fl(r[$],U);if(!U&&!P.ke&&B||!U&&P.ke&&!B&&Fl(b,$)||U&&!P.ke&&G||U&&P.ke&&!G&&Fl(o[$]||[],U)){P.xe=!0;const K=Tve(D,"type",!0);fFe(D,"type",!!K),fFe(D,Hwe);let X=Tve(D,"src",!0);X&&fFe(D,"src",!0);const ae=Ui("script");ae.textContent=D.innerHTML;for(const{nodeName:he}of D.attributes)qa(ae,he,D[he]||Tve(D,he));K&&(ae.type=K),X?ae.src=X:X=D.src;const oe=!!X&&(!K||["text/javascript","module"].includes(K));if(oe&&(ae.onload=ae.onerror=()=>{k(x,++O)}),D.replaceWith(ae),oe)return}k(x,++O)};k(y,0)},xve="bottom",QWe="left",gln="center",cQe="right",pFe="inline",mln="wide",bln="pm--",gFe=["middle","top",xve],UBt=[QWe,gln,cQe],eMr={box:{Ee:[mln,pFe],Ae:gFe,Ne:UBt,He:xve,Ve:cQe},cloud:{Ee:[pFe],Ae:gFe,Ne:UBt,He:xve,Ve:gln},bar:{Ee:[pFe],Ae:gFe.slice(1),Ne:[],He:xve,Ve:""}},tMr={box:{Ee:[],Ae:[],Ne:[],He:"",Ve:""},bar:{Ee:[mln],Ae:[],Ne:[QWe,cQe],He:"",Ve:QWe}},vln=n=>{const r=rr.o.i.guiOptions,o=r&&r.consentModal,s=r&&r.preferencesModal;n===0&&zBt(rr.ne.he,eMr,o,"cm--","box","cm"),n===1&&zBt(rr.ne.we,tMr,s,bln,"box","pm")},zBt=(n,r,o,s,c,d)=>{n.className=d;const h=o&&o.layout,b=o&&o.position,y=o&&o.flipButtons,E=!o||o.equalWeightButtons!==!1,k=h&&h.split(" ")||[],x=k[0],O=k[1],P=x in r?x:c,D=r[P],$=Fl(D.Ee,O)&&O,U=b&&b.split(" ")||[],B=U[0],G=s===bln?U[0]:U[1],K=Fl(D.Ae,B)?B:D.He,X=Fl(D.Ne,G)?G:D.Ve,ae=he=>{he&&sd(n,s+he)};ae(P),ae($),ae(K),ae(X),y&&ae("flip");const oe=d+"__btn--secondary";if(d==="cm"){const{Ie:he,Le:ye}=rr.ne;he&&(E?yM(he,oe):sd(he,oe)),ye&&(E?yM(ye,oe):sd(ye,oe))}else{const{je:he}=rr.ne;he&&(E?yM(he,oe):sd(he,oe))}},doe=(n,r)=>{const o=rr.o,s=rr.ne,{hide:c,hidePreferences:d,acceptCategory:h}=n,b=X=>{h(X),d(),c()},y=o.u&&o.u.preferencesModal;if(!y)return;const E=y.title,k=y.closeIconLabel,x=y.acceptAllBtn,O=y.acceptNecessaryBtn,P=y.savePreferencesBtn,D=y.sections||[],$=x||O||P;if(s.Fe)s.Pe=Ui(lu),Wc(s.Pe,"body");else{s.Fe=Ui(lu),sd(s.Fe,"pm-wrapper");const X=Ui("div");sd(X,"pm-overlay"),$i(s.Fe,X),eh(X,q1,d),s.we=Ui(lu),sd(s.we,"pm"),qa(s.we,"role","dialog"),qa(s.we,Dx,!0),qa(s.we,"aria-modal",!0),qa(s.we,"aria-labelledby","pm__title"),eh(s.ye,"keydown",ae=>{ae.keyCode===27&&d()},!0),s.Oe=Ui(lu),Wc(s.Oe,"header"),s.Re=Ui("h2"),Wc(s.Re,"title"),s.Re.id="pm__title",s.Be=Ui(Lx),Wc(s.Be,"close-btn"),qa(s.Be,"aria-label",y.closeIconLabel||""),eh(s.Be,q1,d),s.$e=Ui("span"),s.$e.innerHTML=loe(),$i(s.Be,s.$e),s.Ge=Ui(lu),Wc(s.Ge,"body"),s.Je=Ui(lu),Wc(s.Je,"footer");var U=Ui(lu);sd(U,"btns");var B=Ui(lu),G=Ui(lu);Wc(B,bie),Wc(G,bie),$i(s.Je,B),$i(s.Je,G),$i(s.Oe,s.Re),$i(s.Oe,s.Be),s.ve=Ui(lu),qa(s.ve,"tabIndex",-1),$i(s.we,s.ve),$i(s.we,s.Oe),$i(s.we,s.Ge),$&&$i(s.we,s.Je),$i(s.Fe,s.we)}let K;E&&(s.Re.innerHTML=E,k&&qa(s.Be,"aria-label",k)),D.forEach((X,ae)=>{const oe=X.title,he=X.description,ye=X.linkedCategory,we=ye&&o.P[ye],ge=X.cookieTable,Te=ge&&ge.body,xe=ge&&ge.caption,Ue=Te&&Te.length>0,Z=!!we,Re=Z&&o.X[ye],_e=JWe(Re)&&HB(Re)||[],qe=Z&&(!!he||!!Ue||HB(Re).length>0);var it=Ui(lu);if(Wc(it,"section"),qe||he){var Xe=Ui(lu);Wc(Xe,"section-desc-wrapper")}let de=_e.length;if(qe&&de>0){const Yt=Ui(lu);Wc(Yt,"section-services");for(const Or of _e){const Nt=Re[Or],on=Nt&&Nt.label||Or,br=Ui(lu),wi=Ui(lu),no=Ui(lu),Ri=Ui(lu);Wc(br,"service"),Wc(Ri,"service-title"),Wc(wi,"service-header"),Wc(no,"service-icon");const ro=HBt(on,Or,we,!0,ye);Ri.innerHTML=on,$i(wi,no),$i(wi,Ri),$i(br,wi),$i(br,ro),$i(Yt,br)}$i(Xe,Yt)}if(oe){var We=Ui(lu),ht=Ui(Z?Lx:lu);if(Wc(We,"section-title-wrapper"),Wc(ht,"section-title"),ht.innerHTML=oe,$i(We,ht),Z){const Yt=Ui("span");Yt.innerHTML=loe(2,3.5),Wc(Yt,"section-arrow"),$i(We,Yt),it.className+="--toggle";const Or=HBt(oe,ye,we);let Nt=y.serviceCounterLabel;if(de>0&&LK(Nt)){let on=Ui("span");Wc(on,"badge"),Wc(on,"service-counter"),qa(on,Dx,!0),qa(on,"data-servicecounter",de),Nt&&(Nt=Nt.split("|"),Nt=Nt.length>1&&de>1?Nt[1]:Nt[0],qa(on,"data-counterlabel",Nt)),on.innerHTML=de+(Nt?" "+Nt:""),$i(ht,on)}if(qe){Wc(it,"section--expandable");var Ie=ye+"-desc";qa(ht,"aria-expanded",!1),qa(ht,"aria-controls",Ie)}$i(We,Or)}else qa(ht,"role","heading"),qa(ht,"aria-level","3");$i(it,We)}if(he){var $t=Ui("p");Wc($t,"section-desc"),$t.innerHTML=he,$i(Xe,$t)}if(qe&&(qa(Xe,Dx,"true"),Xe.id=Ie,((Yt,Or,Nt)=>{eh(ht,q1,()=>{Or.classList.contains("is-expanded")?(yM(Or,"is-expanded"),qa(Nt,"aria-expanded","false"),qa(Yt,Dx,"true")):(sd(Or,"is-expanded"),qa(Nt,"aria-expanded","true"),qa(Yt,Dx,"false"))})})(Xe,it,ht),Ue)){const Yt=Ui("table"),Or=Ui("thead"),Nt=Ui("tbody");if(xe){const ro=Ui("caption");Wc(ro,"table-caption"),ro.innerHTML=xe,Yt.appendChild(ro)}Wc(Yt,"section-table"),Wc(Or,"table-head"),Wc(Nt,"table-body");const on=ge.headers,br=HB(on),wi=s.Ue.createDocumentFragment(),no=Ui("tr");for(const ro of br){const Ko=on[ro],Ca=Ui("th");Ca.id="cc__row-"+Ko+ae,qa(Ca,"scope","col"),Wc(Ca,"table-th"),Ca.innerHTML=Ko,$i(wi,Ca)}$i(no,wi),$i(Or,no);const Ri=s.Ue.createDocumentFragment();for(const ro of Te){const Ko=Ui("tr");Wc(Ko,"table-tr");for(const Ca of br){const qi=on[Ca],yi=ro[Ca],lo=Ui("td"),Eo=Ui(lu);Wc(lo,"table-td"),qa(lo,"data-column",qi),qa(lo,"headers","cc__row-"+qi+ae),Eo.insertAdjacentHTML("beforeend",yi),$i(lo,Eo),$i(Ko,lo)}$i(Ri,Ko)}$i(Nt,Ri),$i(Yt,Or),$i(Yt,Nt),$i(Xe,Yt)}(qe||he)&&$i(it,Xe);const An=s.Pe||s.Ge;Z?(K||(K=Ui(lu),Wc(K,"section-toggles")),K.appendChild(it)):K=null,$i(An,K||it)}),x&&(s.ze||(s.ze=Ui(Lx),Wc(s.ze,"btn"),qa(s.ze,$W,"all"),$i(B,s.ze),eh(s.ze,q1,()=>b("all"))),s.ze.innerHTML=x),O&&(s.je||(s.je=Ui(Lx),Wc(s.je,"btn"),qa(s.je,$W,"necessary"),$i(B,s.je),eh(s.je,q1,()=>b([]))),s.je.innerHTML=O),P&&(s.qe||(s.qe=Ui(Lx),Wc(s.qe,"btn"),Wc(s.qe,"btn--secondary"),qa(s.qe,$W,"save"),$i(G,s.qe),eh(s.qe,q1,()=>b())),s.qe.innerHTML=P),s.Pe&&(s.we.replaceChild(s.Pe,s.Ge),s.Ge=s.Pe),vln(1),o.N||(o.N=!0,NI(rr.re.ue,aQe,s.we),r(n),$i(s.Ce,s.Fe),hln(s.we),setTimeout(()=>sd(s.Fe,"cc--anim"),100)),pln(2)};function HBt(n,r,o,s,c){const d=rr.o,h=rr.ne,b=Ui("label"),y=Ui("input"),E=Ui("span"),k=Ui("span"),x=Ui("span"),O=Ui("span"),P=Ui("span");if(O.innerHTML=loe(1,3),P.innerHTML=loe(0,3),y.type="checkbox",sd(b,"section__toggle-wrapper"),sd(y,"section__toggle"),sd(O,"toggle__icon-on"),sd(P,"toggle__icon-off"),sd(E,"toggle__icon"),sd(k,"toggle__icon-circle"),sd(x,"toggle__label"),qa(E,Dx,"true"),s?(sd(b,"toggle-service"),qa(y,Hwe,c),h.se[c][r]=y):h.ae[r]=y,s?(D=>{eh(y,"change",()=>{const $=h.se[D],U=h.ae[D];d.Z[D]=[];for(let B in $){const G=$[B];G.checked&&d.Z[D].push(G.value)}U.checked=d.Z[D].length>0})})(c):(D=>{eh(y,q1,()=>{const $=h.se[D],U=y.checked;d.Z[D]=[];for(let B in $)$[B].checked=U,U&&d.Z[D].push(B)})})(r),y.value=r,x.textContent=n.replace(/<.*>.*<\/.*>/gm,""),$i(k,P),$i(k,O),$i(E,k),d.D)(o.readOnly||o.enabled)&&(y.checked=!0);else if(s){const D=d.Y[c];y.checked=o.readOnly||Fl(D,r)}else Fl(d.R,r)&&(y.checked=!0);return o.readOnly&&(y.disabled=!0),$i(b,y),$i(b,E),$i(b,x),b}const mFe=()=>{const n=Ui("span");return rr.ne.Ke||(rr.ne.Ke=n),n},wln=(n,r)=>{const o=rr.o,s=rr.ne,{hide:c,showPreferences:d,acceptCategory:h}=n,b=o.u&&o.u.consentModal;if(!b)return;const y=b.acceptAllBtn,E=b.acceptNecessaryBtn,k=b.showPreferencesBtn,x=b.closeIconLabel,O=b.footer,P=b.label,D=b.title,$=B=>{c(),h(B)};if(!s.Qe){s.Qe=Ui(lu),s.he=Ui(lu),s.We=Ui(lu),s.Xe=Ui(lu),s.Ye=Ui(lu),sd(s.Qe,"cm-wrapper"),sd(s.he,"cm"),rv(s.We,"body"),rv(s.Xe,"texts"),rv(s.Ye,"btns"),qa(s.he,"role","dialog"),qa(s.he,"aria-modal","true"),qa(s.he,Dx,"false"),qa(s.he,"aria-describedby","cm__desc"),P?qa(s.he,"aria-label",P):D&&qa(s.he,"aria-labelledby","cm__title");const B="box",G=o.i.guiOptions,K=G&&G.consentModal,X=(K&&K.layout||B).split(" ")[0]===B;D&&x&&X&&(s.Le||(s.Le=Ui(Lx),s.Le.innerHTML=loe(),rv(s.Le,"btn"),rv(s.Le,"btn--close"),eh(s.Le,q1,()=>{$([])}),$i(s.We,s.Le)),qa(s.Le,"aria-label",x)),$i(s.We,s.Xe),(y||E||k)&&$i(s.We,s.Ye),s.be=Ui(lu),qa(s.be,"tabIndex",-1),$i(s.he,s.be),$i(s.he,s.We),$i(s.Qe,s.he)}D&&(s.Ze||(s.Ze=Ui("h2"),s.Ze.className=s.Ze.id="cm__title",$i(s.Xe,s.Ze)),s.Ze.innerHTML=D);let U=b.description;if(U&&(o.V&&(U=U.replace("{{revisionMessage}}",o.I?"":b.revisionMessage||"")),s.et||(s.et=Ui("p"),s.et.className=s.et.id="cm__desc",$i(s.Xe,s.et)),s.et.innerHTML=U),y&&(s.tt||(s.tt=Ui(Lx),$i(s.tt,mFe()),rv(s.tt,"btn"),qa(s.tt,$W,"all"),eh(s.tt,q1,()=>{$("all")})),s.tt.firstElementChild.innerHTML=y),E&&(s.Ie||(s.Ie=Ui(Lx),$i(s.Ie,mFe()),rv(s.Ie,"btn"),qa(s.Ie,$W,"necessary"),eh(s.Ie,q1,()=>{$([])})),s.Ie.firstElementChild.innerHTML=E),k&&(s.ot||(s.ot=Ui(Lx),$i(s.ot,mFe()),rv(s.ot,"btn"),rv(s.ot,"btn--secondary"),qa(s.ot,$W,"show"),eh(s.ot,"mouseenter",()=>{o.N||doe(n,r)}),eh(s.ot,q1,d)),s.ot.firstElementChild.innerHTML=k),s.nt||(s.nt=Ui(lu),rv(s.nt,bie),y&&$i(s.nt,s.tt),E&&$i(s.nt,s.Ie),(y||E)&&$i(s.We,s.nt),$i(s.Ye,s.nt)),s.ot&&!s.st&&(s.st=Ui(lu),s.Ie&&s.tt?(rv(s.st,bie),$i(s.st,s.ot),$i(s.Ye,s.st)):($i(s.nt,s.ot),rv(s.nt,bie+"--uneven"))),O){if(!s.ct){let B=Ui(lu),G=Ui(lu);s.ct=Ui(lu),rv(B,"footer"),rv(G,"links"),rv(s.ct,"link-group"),$i(G,s.ct),$i(B,G),$i(s.he,B)}s.ct.innerHTML=O}vln(0),o.T||(o.T=!0,NI(rr.re.ue,oQe,s.he),r(n),$i(s.Ce,s.Qe),hln(s.he),setTimeout(()=>sd(s.Qe,"cc--anim"),100)),pln(1),lln(s.We,n,doe,r)},yln=n=>{if(!LK(n))return null;if(n in rr.o._)return n;let r=n.slice(0,2);return r in rr.o._?r:null},Eln=()=>rr.o.l||rr.o.i.language.default,Sln=n=>{n&&(rr.o.l=n)},nMr=async n=>{const r=rr.o;let o=yln(n)?n:Eln(),s=r._[o];if(LK(s)?s=await(async c=>{try{return await(await fetch(c)).json()}catch(d){return console.error(d),!1}})(s):fI(s)&&(s=await s()),!s)throw`Could not load translation for the '${o}' language`;return r.u=s,Sln(o),!0},rMr=()=>{let n=rr.o.i.language.rtl,r=rr.ne.Ce;n&&r&&(TEe(n)||(n=[n]),Fl(n,rr.o.l)?sd(r,"cc--rtl"):yM(r,"cc--rtl"))},vie=()=>{const n=rr.ne;if(n.Ce)return;n.Ce=Ui(lu),n.Ce.id="cc-main",n.Ce.setAttribute("data-nosnippet",""),rMr();let r=rr.o.i.root;r&&LK(r)&&(r=document.querySelector(r)),(r||n.Ue.body).appendChild(n.Ce)},iMr=n=>qwe(()=>localStorage.removeItem(n)),GBt=(n,r)=>{if(r instanceof RegExp)return n.filter(o=>r.test(o));{const o=sln(n,r);return o>-1?[n[o]]:[]}},oMr=n=>{const{hostname:r,protocol:o}=location,{name:s,path:c,domain:d,sameSite:h,useLocalStorage:b,secure:y}=rr.t.cookie,E=864e5*JDr(),k=new Date;k.setTime(k.getTime()+E),rr.o.p.expirationTime=k.getTime();const x=JSON.stringify(rr.o.p);let O=s+"="+encodeURIComponent(x)+(E!==0?"; expires="+k.toUTCString():"")+"; Path="+c+"; SameSite="+h;Fl(r,".")&&(O+="; Domain="+d),y&&o==="https:"&&(O+="; Secure"),b?((P,D)=>{qwe(()=>localStorage.setItem(P,D))})(s,x):document.cookie=O,rr.o.p},qBt=(n,r,o)=>{if(n.length===0)return;const s=o||rr.t.cookie.domain,c=r||rr.t.cookie.path,d=s.slice(0,4)==="www.",h=d&&s.substring(4),b=(y,E)=>{E&&E.slice(0,1)!=="."&&(E="."+E),document.cookie=y+"=; path="+c+(E?"; domain="+E:"")+"; expires=Thu, 01 Jan 1970 00:00:01 GMT;"};for(const y of n)b(y,o),o||b(y,s),d&&b(y,h)},_ln=n=>{const r=rr.t.cookie.name,o=rr.t.cookie.useLocalStorage;return((c,d)=>{let h;return h=qwe(()=>JSON.parse(d?c:decodeURIComponent(c)),!0)||{},h})(o?(s=r,qwe(()=>localStorage.getItem(s))||""):aMr(r),o);var s},aMr=(n,r)=>{const o=document.cookie.match("(^|;)\\s*"+n+"\\s*=\\s*([^;]+)");return o?o.pop():""},sMr=n=>{const r=document.cookie.split(/;\s*/),o=[];for(const s of r){let c=s.split("=")[0];o.push(c)}return o},cMr=(n,r=[])=>{((o,s)=>{const{O:c,R:d,B:h,N:b,Z:y,$:E,X:k}=rr.o;let x=[];if(o){TEe(o)?x.push(...o):LK(o)&&(x=o==="all"?c:[o]);for(const O of c)y[O]=Fl(x,O)?HB(k[O]):[]}else x=[...d,...E],b&&(x=(()=>{const O=rr.ne.ae;if(!O)return[];let P=[];for(let D in O)O[D].checked&&P.push(D);return P})());x=x.filter(O=>!Fl(c,O)||!Fl(s,O)),x.push(...h),uln(x)})(n,r),(()=>{const o=rr.o,{Z:s,B:c,Y:d,X:h,O:b}=o,y=b;o.te=lM(d);for(const E of y){const k=h[E],x=HB(k),O=s[E]&&s[E].length>0,P=Fl(c,E);if(x.length!==0){if(d[E]=[],P)d[E].push(...x);else if(O){const D=s[E];d[E].push(...D)}else d[E]=o.Z[E];d[E]=cln(d[E])}}})(),(()=>{const o=rr.o;o.L=rr.t.mode===Yre&&o.D?hFe(o.$,o.R):hFe(o.R,o.p.categories);let s=o.L.length>0,c=!1;for(const y of o.O)o.ee[y]=hFe(o.Y[y],o.te[y]),o.ee[y].length>0&&(c=!0);const d=rr.ne.ae;for(const y in d)d[y].checked=Fl(o.R,y);for(const y of o.O){const E=rr.ne.se[y],k=o.Y[y];for(const x in E)E[x].checked=Fl(k,x)}o.C||(o.C=new Date),o.M||(o.M=("10000000-1000-4000-8000"+-1e11).replace(/[018]/g,y=>(y^crypto.getRandomValues(new Uint8Array(1))[0]&15>>y/4).toString(16))),o.p={categories:lM(o.R),revision:rr.t.revision,data:o.h,consentTimestamp:o.C.toISOString(),consentId:o.M,services:lM(o.Y),languageCode:rr.o.l},o.S&&(o.p.lastConsentTimestamp=o.S.toISOString());let h=!1;const b=s||c;(o.D||b)&&(o.D&&(o.D=!1,h=!0),o.S=o.S?new Date:o.C,o.p.lastConsentTimestamp=o.S.toISOString(),oMr(),rr.t.autoClearCookies&&(h||b)&&(y=>{const E=rr.o,k=sMr(),x=(O=>{const P=rr.o;return(O?P.O:P.L).filter(D=>{const $=P.P[D];return!!$&&!$.readOnly&&!!$.autoClear})})(y);for(const O in E.ee)for(const P of E.ee[O]){const D=E.X[O][P].cookies;if(!Fl(E.Y[O],P)&&D)for(const $ of D){const U=GBt(k,$.name);qBt(U,$.path,$.domain)}}for(const O of x){const P=E.P[O].autoClear,D=P&&P.cookies||[],$=Fl(E.L,O),U=!Fl(E.R,O),B=$&&U;if(y?U:B){P.reloadPage&&B&&(E.j=!0);for(const G of D){const K=GBt(k,G.name);qBt(K,G.path,G.domain)}}}})(h),ZWe()),h&&(NI(rr.re.ie),NI(rr.re.le),rr.t.mode===iln)||(b&&NI(rr.re.de),o.j&&(o.j=!1,location.reload()))})()},uMr=n=>{const r=rr.o.D?[]:rr.o.R;return Fl(r,n)},lMr=(n,r)=>{const o=rr.o.D?[]:rr.o.Y[r]||[];return Fl(o,n)},kln=n=>{const{ne:r,o}=rr;if(!o.k){if(!o.T){if(!n)return;wln(uQe,vie)}o.k=!0,o.J=Gwe(),o.v&&fln(!0),dln(r.he,1),sd(r.ye,oln),qa(r.he,Dx,"false"),setTimeout(()=>{OI(rr.ne.be)},100),NI(rr.re.fe,oQe)}},dMr=()=>{const{ne:n,o:r,re:o}=rr;r.k&&(r.k=!1,r.v&&fln(),OI(n.Ke,!0),yM(n.ye,oln),qa(n.he,Dx,"true"),OI(r.J),r.J=null,NI(o._e,oQe))},fMr=()=>{const n=rr.o;n.A||(n.N||doe(uQe,vie),n.A=!0,n.k?n.U=Gwe():n.J=Gwe(),dln(rr.ne.we,2),sd(rr.ne.ye,aln),qa(rr.ne.we,Dx,"false"),setTimeout(()=>{OI(rr.ne.ve)},100),NI(rr.re.fe,aQe))},hMr=()=>{const n=rr.o;n.A&&(n.A=!1,(()=>{const r=Tln(),o=rr.o.P,s=rr.ne.ae,c=rr.ne.se,d=h=>Fl(rr.o.$,h);for(const h in s){const b=!!o[h].readOnly;s[h].checked=b||(r?uMr(h):d(h));for(const y in c[h])c[h][y].checked=b||(r?lMr(y,h):d(h))}})(),OI(rr.ne.$e,!0),yM(rr.ne.ye,aln),qa(rr.ne.we,Dx,"true"),n.k?(OI(n.U),n.U=null):(OI(n.J),n.J=null),NI(rr.re._e,aQe))};var uQe={show:kln,hide:dMr,showPreferences:fMr,hidePreferences:hMr,acceptCategory:cMr};const pMr=(n,r)=>_ln(),Tln=()=>!rr.o.D,gMr=async n=>{const{o:r,t:o,re:s}=rr,c=window;if(!c._ccRun){if(c._ccRun=!0,(b=>{const{ne:y,t:E,o:k}=rr,x=E,O=k,{cookie:P}=x,D=rr.ce,$=b.cookie,U=b.categories,B=HB(U)||[],G=navigator,K=document;y.Ue=K,y.ye=K.documentElement,P.domain=location.hostname,O.i=b,O.P=U,O.O=B,O._=b.language.translations,O.v=!!b.disablePageInteraction,D.ie=b.onFirstConsent,D.le=b.onConsent,D.de=b.onChange,D._e=b.onModalHide,D.fe=b.onModalShow,D.ue=b.onModalReady;const{mode:X,autoShow:ae,lazyHtmlGeneration:oe,autoClearCookies:he,revision:ye,manageScriptTags:we,hideFromBots:ge}=b;X===Yre&&(x.mode=X),typeof he=="boolean"&&(x.autoClearCookies=he),typeof we=="boolean"&&(x.manageScriptTags=we),typeof ye=="number"&&ye>=0&&(x.revision=ye,O.V=!0),typeof ae=="boolean"&&(x.autoShow=ae),typeof oe=="boolean"&&(x.lazyHtmlGeneration=oe),ge===!1&&(x.hideFromBots=!1),x.hideFromBots===!0&&G&&(O.G=G.userAgent&&/bot|crawl|spider|slurp|teoma/i.test(G.userAgent)||G.webdriver),JWe($)&&(x.cookie={...P,...$}),x.autoClearCookies,O.V,x.manageScriptTags,(Te=>{const{P:xe,X:Ue,Y:Z,Z:Re,B:_e}=rr.o;for(let qe of Te){const it=xe[qe],Xe=it.services||{},de=JWe(Xe)&&HB(Xe)||[];Ue[qe]={},Z[qe]=[],Re[qe]=[],it.readOnly&&(_e.push(qe),Z[qe]=de),rr.ne.se[qe]={};for(let We of de){const ht=Xe[We];ht.Se=!1,Ue[qe][We]=ht}}})(B),(()=>{if(!rr.t.manageScriptTags)return;const Te=rr.o,xe=sQe(document,"script["+Hwe+"]");for(const Ue of xe){let Z=Tve(Ue,Hwe),Re=Ue.dataset.service||"",_e=!1;if(Z&&Z.charAt(0)==="!"&&(Z=Z.slice(1),_e=!0),Re.charAt(0)==="!"&&(Re=Re.slice(1),_e=!0),Fl(Te.O,Z)&&(Te.oe.push({Me:Ue,xe:!1,ke:_e,De:Z,Te:Re}),Re)){const qe=Te.X[Z];qe[Re]||(qe[Re]={Se:!1})}}})(),Sln((()=>{const Te=rr.o.i.language.autoDetect;if(Te){const xe={browser:navigator.language,document:document.documentElement.lang},Ue=yln(xe[Te]);if(Ue)return Ue}return Eln()})())})(n),r.G)return;(()=>{const b=rr.o,y=rr.t,E=_ln(),{categories:k,services:x,consentId:O,consentTimestamp:P,lastConsentTimestamp:D,data:$,revision:U}=E,B=TEe(k);b.p=E,b.M=O;const G=!!O&&LK(O);b.C=P,b.C&&(b.C=new Date(P)),b.S=D,b.S&&(b.S=new Date(D)),b.h=$!==void 0?$:null,b.V&&G&&U!==y.revision&&(b.I=!1),b.D=!(G&&b.I&&b.C&&b.S&&B),y.cookie.useLocalStorage&&!b.D&&(b.D=new Date().getTime()>(E.expirationTime||0),b.D&&iMr(y.cookie.name)),b.D,(()=>{const K=rr.o;for(const X of K.O){const ae=K.P[X];if(ae.readOnly||ae.enabled){K.$.push(X);const oe=K.X[X]||{};for(let he in oe)K.Z[X].push(he),K.i.mode===Yre&&K.Y[X].push(he)}}})(),b.D?y.mode===Yre&&(b.R=[...b.$]):(b.Y={...b.Y,...x},b.Z={...b.Y},uln([...b.B,...k]))})();const h=Tln();if(!await nMr())return!1;if(lln(null,d=uQe,doe,vie),rr.o.D&&wln(d,vie),rr.t.lazyHtmlGeneration||doe(d,vie),o.autoShow&&!h&&kln(!0),h)return ZWe(),NI(s.le);o.mode===Yre&&ZWe(r.$)}var d};function mMr(){return"{{COOKIE_CONSENT_DOMAIN}}"}function bMr({onChange:n,onConsent:r}={}){const o=mMr(),s={onChange:n,onConsent:r,guiOptions:{consentModal:{layout:"box inline",equalWeightButtons:!1},preferencesModal:{layout:"box"}},disablePageInteraction:!1,categories:{necessary:{enabled:!0,readOnly:!0},analytics:{enabled:!0,readOnly:!1},marketing:{enabled:!0,readOnly:!1}},language:{default:"en",translations:{en:{consentModal:{title:"🍪 Cookie settings",description:"We use cookies to keep things running smoothly and improve your experience.<br/><a role='button' data-cc='show-preferencesModal'>Customize settings</a>",acceptAllBtn:"Accept all",acceptNecessaryBtn:"Reject all"},preferencesModal:{title:"Cookie settings",savePreferencesBtn:"Save preferences",closeIconLabel:"Close modal",sections:[{description:"We use cookies to keep things running smoothly and improve your experience. You can change your preferences anytime."},{title:"Strictly necessary cookies",description:"These cookies are essential for the proper functioning of the website and cannot be disabled.",linkedCategory:"necessary"},{title:"Performance & analytics",description:"These cookies collect information about how you use our website.",linkedCategory:"analytics"},{title:"Marketing",description:"These cookies collect marketing information.",linkedCategory:"marketing"}]}}}}};return s.cookie={domain:o},s}function vMr({onConsentUpdate:n}={}){const r=Q.useRef(!1);Q.useEffect(()=>{const c=pMr();RFt(VBt(c?.categories))},[]);const o=Q.useCallback(({cookie:c})=>{const d=VBt(c.categories);RFt(d),n?.(d)},[n]);return{init:Q.useCallback(async()=>{r.current||(await gMr(bMr({onChange:o,onConsent:o})),r.current=!0)},[o])}}const VBt=n=>{const r=n?.includes("marketing")??!1,o=n?.includes("analytics")??!1,s=n?.includes("necessary")??!0;return{marketing:r,analytics:o,necessary:s}};function wMr(){const{getIdTokenClaims:n,isAuthenticated:r,user:o}=ty(),{configure:s}=rln(),{init:c}=vMr({onConsentUpdate:h=>{d(h)}}),d=async h=>{if(s(!h.analytics),h.marketing){const b=await n();!b?.email||typeof b.email!="string"?r&&Va(new Error("Email is not available or invalid while user is authenticated. Skipping hubspot submission."),{tags:{component:"useConsentOrchestration",action:"handleConsentChange"}}):SPr({email:b.email})}h.analytics?Pcn(o):D5r()};return Q.useEffect(()=>{(async()=>{await c()})().catch(h=>{Va(h,{tags:{component:"useConsentOrchestration",action:"init"}})})},[c]),{initConsent:c}}const GM=()=>{const n=Q.useContext(Hcn);if(n===void 0)throw new Error("useLayoutMode must be used within a LayoutProvider");return n},xln=Q.createContext(void 0),WBt=n=>{const r=n?.toUpperCase();return!r||!Object.values(q4).includes(r)?q4.Chat:r},yMr=(n,r,o)=>{switch(n){case IW.StudioEmbed:case IW.Standalone:return`/${r}`;case IW.Cloud:default:return`/project/${o}/${r}`}},Cln=({children:n})=>{const r=y2(),{view:o,pathToFile:s,projectId:c}=G_(),{mode:d,isCloud:h}=GM(),[b,y]=Q.useState(WBt(o)),[E,k]=Q.useState(!h),[x,O]=Q.useState(),[P,D]=Q.useState(0);Q.useEffect(()=>{if(!s)return;O(decodeURIComponent(s));const U=yMr(d,o,c);r(U,{replace:!0})},[s,c,o,r,d]);const $=Q.useMemo(()=>({selectedView:b,inspectEnabled:E,openFile:x,version:P,setInspectEnabled:k,setOpenFile:O,setVersion:D}),[b,E,x,P]);return Q.useEffect(()=>{y(WBt(o))},[o]),H.jsx(xln.Provider,{value:$,children:n})},EMr="/assets/no-flow-DpEYMX3c.svg",SMr="/assets/mobile-kE92MLqG.svg",_Mr="/assets/cubes-CUD1JLED.svg",kMr="/assets/cubes-error-DeEIGykw.svg",TMr="data:image/svg+xml,%3csvg%20width='17'%20height='16'%20viewBox='0%200%2017%2016'%20fill='none'%20xmlns='http://www.w3.org/2000/svg'%3e%3ccircle%20cx='8.5'%20cy='8'%20r='8'%20fill='url(%23paint0_radial_1390_20842)'/%3e%3cdefs%3e%3cradialGradient%20id='paint0_radial_1390_20842'%20cx='0'%20cy='0'%20r='1'%20gradientUnits='userSpaceOnUse'%20gradientTransform='translate(6%203)%20rotate(79.1145)%20scale(13.2382)'%3e%3cstop%20stop-color='%232B1B44'/%3e%3cstop%20offset='0.591346'%20stop-color='%235A17EE'/%3e%3cstop%20offset='1'%20stop-color='%23A855F7'/%3e%3c/radialGradient%3e%3c/defs%3e%3c/svg%3e",xMr="/assets/cubes-pause-BbB_lALP.svg",CMr={[zm.NoFlow]:EMr,[zm.Mobile]:SMr,[zm.Cubes]:_Mr,[zm.CubesError]:kMr,[zm.CubesPause]:xMr},Bx=({shortLabel:n="No results",longLabel:r="We couldn't find any results to display.",image:o=zm.NoFlow,children:s,...c})=>{const d={display:"grid",rowGap:"0.5rem",maxWidth:"21.25rem",mb:"1rem"};return H.jsxs(kYe,{height:"100%",flexDirection:"column",...c,children:[H.jsx(Iye,{src:CMr[o],alt:n,height:"14rem",mb:6}),H.jsxs(ai,{css:d,children:[H.jsx(Kx,{fontSize:"1rem",textAlign:"center",children:n}),r&&H.jsx(Fi,{color:"rasaNeutral.700",textAlign:"center",children:r})]}),s]})},AMr=()=>{const n=Io(da,b=>b.jobStatus),[r,o]=Q.useState(0),s={borderColor:"rasaNeutral.500","&[data-current]":{bg:"rasaNeutral.50",borderColor:"rasaNeutral.500"}},c={"&[data-current]":{bg:"rasaNeutral.500"}},d={animation:"fadeInOut 2s infinite ease-in-out"};Q.useEffect(()=>{n!==a2.Heartbeat&&(n===a2.Generating||n===a2.GenerationSuccess?o(1):n===a2.Training||n===a2.TrainSuccess?o(2):o(0))},[n]);const h=[{id:"analyzing-prompt",label:"Analyzing prompt"},{id:"building-agent",label:"Building agent"},{id:"finalizing-agent",label:"Finalizing agent"}];return H.jsx(Bx,{shortLabel:"Creating your AI agent...",longLabel:"",image:zm.Cubes,children:H.jsxs(Vd,{children:[H.jsx(JQn,{size:"sm",pt:"0.5rem",width:"95%",count:3,defaultStep:0,step:r,children:H.jsx(ZQn,{children:h.map((b,y)=>H.jsxs(QQn,{index:y,gap:"0",children:[H.jsx(ner,{css:s,children:H.jsx(Den,{incomplete:"",complete:H.jsx(xa,{icon:uXe,size:"xl","data-testid":`step-${b.id}-complete`}),current:H.jsx(Iye,{src:TMr,css:d,"data-testid":`step-${b.id}-loading`})})}),H.jsx(ter,{m:"0",css:c})]},b.id))})}),H.jsx(Iu,{justifyContent:"space-between",width:"123%",children:h.map(b=>H.jsx(Fi,{size:"md","data-testid":`step-${b.id}-label`,children:b.label},`step-title-${b.id}`))})]})})},bU=()=>{const n=Q.useContext(xln);if(!n)throw new Error("useViewContext must be used within a ViewContextProvider");return n},jI=Object.create(null);jI.open="0";jI.close="1";jI.ping="2";jI.pong="3";jI.message="4";jI.upgrade="5";jI.noop="6";const Cve=Object.create(null);Object.keys(jI).forEach(n=>{Cve[jI[n]]=n});const eKe={type:"error",data:"parser error"},Aln=typeof Blob=="function"||typeof Blob<"u"&&Object.prototype.toString.call(Blob)==="[object BlobConstructor]",Iln=typeof ArrayBuffer=="function",Oln=n=>typeof ArrayBuffer.isView=="function"?ArrayBuffer.isView(n):n&&n.buffer instanceof ArrayBuffer,lQe=({type:n,data:r},o,s)=>Aln&&r instanceof Blob?o?s(r):KBt(r,s):Iln&&(r instanceof ArrayBuffer||Oln(r))?o?s(r):KBt(new Blob([r]),s):s(jI[n]+(r||"")),KBt=(n,r)=>{const o=new FileReader;return o.onload=function(){const s=o.result.split(",")[1];r("b"+(s||""))},o.readAsDataURL(n)};function YBt(n){return n instanceof Uint8Array?n:n instanceof ArrayBuffer?new Uint8Array(n):new Uint8Array(n.buffer,n.byteOffset,n.byteLength)}let bFe;function IMr(n,r){if(Aln&&n.data instanceof Blob)return n.data.arrayBuffer().then(YBt).then(r);if(Iln&&(n.data instanceof ArrayBuffer||Oln(n.data)))return r(YBt(n.data));lQe(n,!1,o=>{bFe||(bFe=new TextEncoder),r(bFe.encode(o))})}const XBt="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",Xre=typeof Uint8Array>"u"?[]:new Uint8Array(256);for(let n=0;n<XBt.length;n++)Xre[XBt.charCodeAt(n)]=n;const OMr=n=>{let r=n.length*.75,o=n.length,s,c=0,d,h,b,y;n[n.length-1]==="="&&(r--,n[n.length-2]==="="&&r--);const E=new ArrayBuffer(r),k=new Uint8Array(E);for(s=0;s<o;s+=4)d=Xre[n.charCodeAt(s)],h=Xre[n.charCodeAt(s+1)],b=Xre[n.charCodeAt(s+2)],y=Xre[n.charCodeAt(s+3)],k[c++]=d<<2|h>>4,k[c++]=(h&15)<<4|b>>2,k[c++]=(b&3)<<6|y&63;return E},NMr=typeof ArrayBuffer=="function",dQe=(n,r)=>{if(typeof n!="string")return{type:"message",data:Nln(n,r)};const o=n.charAt(0);return o==="b"?{type:"message",data:RMr(n.substring(1),r)}:Cve[o]?n.length>1?{type:Cve[o],data:n.substring(1)}:{type:Cve[o]}:eKe},RMr=(n,r)=>{if(NMr){const o=OMr(n);return Nln(o,r)}else return{base64:!0,data:n}},Nln=(n,r)=>{switch(r){case"blob":return n instanceof Blob?n:new Blob([n]);case"arraybuffer":default:return n instanceof ArrayBuffer?n:n.buffer}},Rln="",PMr=(n,r)=>{const o=n.length,s=new Array(o);let c=0;n.forEach((d,h)=>{lQe(d,!1,b=>{s[h]=b,++c===o&&r(s.join(Rln))})})},LMr=(n,r)=>{const o=n.split(Rln),s=[];for(let c=0;c<o.length;c++){const d=dQe(o[c],r);if(s.push(d),d.type==="error")break}return s};function DMr(){return new TransformStream({transform(n,r){IMr(n,o=>{const s=o.length;let c;if(s<126)c=new Uint8Array(1),new DataView(c.buffer).setUint8(0,s);else if(s<65536){c=new Uint8Array(3);const d=new DataView(c.buffer);d.setUint8(0,126),d.setUint16(1,s)}else{c=new Uint8Array(9);const d=new DataView(c.buffer);d.setUint8(0,127),d.setBigUint64(1,BigInt(s))}n.data&&typeof n.data!="string"&&(c[0]|=128),r.enqueue(c),r.enqueue(o)})}})}let vFe;function j0e(n){return n.reduce((r,o)=>r+o.length,0)}function $0e(n,r){if(n[0].length===r)return n.shift();const o=new Uint8Array(r);let s=0;for(let c=0;c<r;c++)o[c]=n[0][s++],s===n[0].length&&(n.shift(),s=0);return n.length&&s<n[0].length&&(n[0]=n[0].slice(s)),o}function MMr(n,r){vFe||(vFe=new TextDecoder);const o=[];let s=0,c=-1,d=!1;return new TransformStream({transform(h,b){for(o.push(h);;){if(s===0){if(j0e(o)<1)break;const y=$0e(o,1);d=(y[0]&128)===128,c=y[0]&127,c<126?s=3:c===126?s=1:s=2}else if(s===1){if(j0e(o)<2)break;const y=$0e(o,2);c=new DataView(y.buffer,y.byteOffset,y.length).getUint16(0),s=3}else if(s===2){if(j0e(o)<8)break;const y=$0e(o,8),E=new DataView(y.buffer,y.byteOffset,y.length),k=E.getUint32(0);if(k>Math.pow(2,21)-1){b.enqueue(eKe);break}c=k*Math.pow(2,32)+E.getUint32(4),s=3}else{if(j0e(o)<c)break;const y=$0e(o,c);b.enqueue(dQe(d?y:vFe.decode(y),r)),s=0}if(c===0||c>n){b.enqueue(eKe);break}}}})}const Pln=4;function Xp(n){if(n)return jMr(n)}function jMr(n){for(var r in Xp.prototype)n[r]=Xp.prototype[r];return n}Xp.prototype.on=Xp.prototype.addEventListener=function(n,r){return this._callbacks=this._callbacks||{},(this._callbacks["$"+n]=this._callbacks["$"+n]||[]).push(r),this};Xp.prototype.once=function(n,r){function o(){this.off(n,o),r.apply(this,arguments)}return o.fn=r,this.on(n,o),this};Xp.prototype.off=Xp.prototype.removeListener=Xp.prototype.removeAllListeners=Xp.prototype.removeEventListener=function(n,r){if(this._callbacks=this._callbacks||{},arguments.length==0)return this._callbacks={},this;var o=this._callbacks["$"+n];if(!o)return this;if(arguments.length==1)return delete this._callbacks["$"+n],this;for(var s,c=0;c<o.length;c++)if(s=o[c],s===r||s.fn===r){o.splice(c,1);break}return o.length===0&&delete this._callbacks["$"+n],this};Xp.prototype.emit=function(n){this._callbacks=this._callbacks||{};for(var r=new Array(arguments.length-1),o=this._callbacks["$"+n],s=1;s<arguments.length;s++)r[s-1]=arguments[s];if(o){o=o.slice(0);for(var s=0,c=o.length;s<c;++s)o[s].apply(this,r)}return this};Xp.prototype.emitReserved=Xp.prototype.emit;Xp.prototype.listeners=function(n){return this._callbacks=this._callbacks||{},this._callbacks["$"+n]||[]};Xp.prototype.hasListeners=function(n){return!!this.listeners(n).length};const xEe=typeof Promise=="function"&&typeof Promise.resolve=="function"?r=>Promise.resolve().then(r):(r,o)=>o(r,0),P_=typeof self<"u"?self:typeof window<"u"?window:Function("return this")(),$Mr="arraybuffer";function Lln(n,...r){return r.reduce((o,s)=>(n.hasOwnProperty(s)&&(o[s]=n[s]),o),{})}const FMr=P_.setTimeout,BMr=P_.clearTimeout;function CEe(n,r){r.useNativeTimers?(n.setTimeoutFn=FMr.bind(P_),n.clearTimeoutFn=BMr.bind(P_)):(n.setTimeoutFn=P_.setTimeout.bind(P_),n.clearTimeoutFn=P_.clearTimeout.bind(P_))}const UMr=1.33;function zMr(n){return typeof n=="string"?HMr(n):Math.ceil((n.byteLength||n.size)*UMr)}function HMr(n){let r=0,o=0;for(let s=0,c=n.length;s<c;s++)r=n.charCodeAt(s),r<128?o+=1:r<2048?o+=2:r<55296||r>=57344?o+=3:(s++,o+=4);return o}function Dln(){return Date.now().toString(36).substring(3)+Math.random().toString(36).substring(2,5)}function GMr(n){let r="";for(let o in n)n.hasOwnProperty(o)&&(r.length&&(r+="&"),r+=encodeURIComponent(o)+"="+encodeURIComponent(n[o]));return r}function qMr(n){let r={},o=n.split("&");for(let s=0,c=o.length;s<c;s++){let d=o[s].split("=");r[decodeURIComponent(d[0])]=decodeURIComponent(d[1])}return r}class VMr extends Error{constructor(r,o,s){super(r),this.description=o,this.context=s,this.type="TransportError"}}class fQe extends Xp{constructor(r){super(),this.writable=!1,CEe(this,r),this.opts=r,this.query=r.query,this.socket=r.socket,this.supportsBinary=!r.forceBase64}onError(r,o,s){return super.emitReserved("error",new VMr(r,o,s)),this}open(){return this.readyState="opening",this.doOpen(),this}close(){return(this.readyState==="opening"||this.readyState==="open")&&(this.doClose(),this.onClose()),this}send(r){this.readyState==="open"&&this.write(r)}onOpen(){this.readyState="open",this.writable=!0,super.emitReserved("open")}onData(r){const o=dQe(r,this.socket.binaryType);this.onPacket(o)}onPacket(r){super.emitReserved("packet",r)}onClose(r){this.readyState="closed",super.emitReserved("close",r)}pause(r){}createUri(r,o={}){return r+"://"+this._hostname()+this._port()+this.opts.path+this._query(o)}_hostname(){const r=this.opts.hostname;return r.indexOf(":")===-1?r:"["+r+"]"}_port(){return this.opts.port&&(this.opts.secure&&+(this.opts.port!==443)||!this.opts.secure&&Number(this.opts.port)!==80)?":"+this.opts.port:""}_query(r){const o=GMr(r);return o.length?"?"+o:""}}class WMr extends fQe{constructor(){super(...arguments),this._polling=!1}get name(){return"polling"}doOpen(){this._poll()}pause(r){this.readyState="pausing";const o=()=>{this.readyState="paused",r()};if(this._polling||!this.writable){let s=0;this._polling&&(s++,this.once("pollComplete",function(){--s||o()})),this.writable||(s++,this.once("drain",function(){--s||o()}))}else o()}_poll(){this._polling=!0,this.doPoll(),this.emitReserved("poll")}onData(r){const o=s=>{if(this.readyState==="opening"&&s.type==="open"&&this.onOpen(),s.type==="close")return this.onClose({description:"transport closed by the server"}),!1;this.onPacket(s)};LMr(r,this.socket.binaryType).forEach(o),this.readyState!=="closed"&&(this._polling=!1,this.emitReserved("pollComplete"),this.readyState==="open"&&this._poll())}doClose(){const r=()=>{this.write([{type:"close"}])};this.readyState==="open"?r():this.once("open",r)}write(r){this.writable=!1,PMr(r,o=>{this.doWrite(o,()=>{this.writable=!0,this.emitReserved("drain")})})}uri(){const r=this.opts.secure?"https":"http",o=this.query||{};return this.opts.timestampRequests!==!1&&(o[this.opts.timestampParam]=Dln()),!this.supportsBinary&&!o.sid&&(o.b64=1),this.createUri(r,o)}}let Mln=!1;try{Mln=typeof XMLHttpRequest<"u"&&"withCredentials"in new XMLHttpRequest}catch{}const KMr=Mln;function YMr(){}class XMr extends WMr{constructor(r){if(super(r),typeof location<"u"){const o=location.protocol==="https:";let s=location.port;s||(s=o?"443":"80"),this.xd=typeof location<"u"&&r.hostname!==location.hostname||s!==r.port}}doWrite(r,o){const s=this.request({method:"POST",data:r});s.on("success",o),s.on("error",(c,d)=>{this.onError("xhr post error",c,d)})}doPoll(){const r=this.request();r.on("data",this.onData.bind(this)),r.on("error",(o,s)=>{this.onError("xhr poll error",o,s)}),this.pollXhr=r}}class RI extends Xp{constructor(r,o,s){super(),this.createRequest=r,CEe(this,s),this._opts=s,this._method=s.method||"GET",this._uri=o,this._data=s.data!==void 0?s.data:null,this._create()}_create(){var r;const o=Lln(this._opts,"agent","pfx","key","passphrase","cert","ca","ciphers","rejectUnauthorized","autoUnref");o.xdomain=!!this._opts.xd;const s=this._xhr=this.createRequest(o);try{s.open(this._method,this._uri,!0);try{if(this._opts.extraHeaders){s.setDisableHeaderCheck&&s.setDisableHeaderCheck(!0);for(let c in this._opts.extraHeaders)this._opts.extraHeaders.hasOwnProperty(c)&&s.setRequestHeader(c,this._opts.extraHeaders[c])}}catch{}if(this._method==="POST")try{s.setRequestHeader("Content-type","text/plain;charset=UTF-8")}catch{}try{s.setRequestHeader("Accept","*/*")}catch{}(r=this._opts.cookieJar)===null||r===void 0||r.addCookies(s),"withCredentials"in s&&(s.withCredentials=this._opts.withCredentials),this._opts.requestTimeout&&(s.timeout=this._opts.requestTimeout),s.onreadystatechange=()=>{var c;s.readyState===3&&((c=this._opts.cookieJar)===null||c===void 0||c.parseCookies(s.getResponseHeader("set-cookie"))),s.readyState===4&&(s.status===200||s.status===1223?this._onLoad():this.setTimeoutFn(()=>{this._onError(typeof s.status=="number"?s.status:0)},0))},s.send(this._data)}catch(c){this.setTimeoutFn(()=>{this._onError(c)},0);return}typeof document<"u"&&(this._index=RI.requestsCount++,RI.requests[this._index]=this)}_onError(r){this.emitReserved("error",r,this._xhr),this._cleanup(!0)}_cleanup(r){if(!(typeof this._xhr>"u"||this._xhr===null)){if(this._xhr.onreadystatechange=YMr,r)try{this._xhr.abort()}catch{}typeof document<"u"&&delete RI.requests[this._index],this._xhr=null}}_onLoad(){const r=this._xhr.responseText;r!==null&&(this.emitReserved("data",r),this.emitReserved("success"),this._cleanup())}abort(){this._cleanup()}}RI.requestsCount=0;RI.requests={};if(typeof document<"u"){if(typeof attachEvent=="function")attachEvent("onunload",JBt);else if(typeof addEventListener=="function"){const n="onpagehide"in P_?"pagehide":"unload";addEventListener(n,JBt,!1)}}function JBt(){for(let n in RI.requests)RI.requests.hasOwnProperty(n)&&RI.requests[n].abort()}const JMr=function(){const n=jln({xdomain:!1});return n&&n.responseType!==null}();class ZMr extends XMr{constructor(r){super(r);const o=r&&r.forceBase64;this.supportsBinary=JMr&&!o}request(r={}){return Object.assign(r,{xd:this.xd},this.opts),new RI(jln,this.uri(),r)}}function jln(n){const r=n.xdomain;try{if(typeof XMLHttpRequest<"u"&&(!r||KMr))return new XMLHttpRequest}catch{}if(!r)try{return new P_[["Active"].concat("Object").join("X")]("Microsoft.XMLHTTP")}catch{}}const $ln=typeof navigator<"u"&&typeof navigator.product=="string"&&navigator.product.toLowerCase()==="reactnative";class QMr extends fQe{get name(){return"websocket"}doOpen(){const r=this.uri(),o=this.opts.protocols,s=$ln?{}:Lln(this.opts,"agent","perMessageDeflate","pfx","key","passphrase","cert","ca","ciphers","rejectUnauthorized","localAddress","protocolVersion","origin","maxPayload","family","checkServerIdentity");this.opts.extraHeaders&&(s.headers=this.opts.extraHeaders);try{this.ws=this.createSocket(r,o,s)}catch(c){return this.emitReserved("error",c)}this.ws.binaryType=this.socket.binaryType,this.addEventListeners()}addEventListeners(){this.ws.onopen=()=>{this.opts.autoUnref&&this.ws._socket.unref(),this.onOpen()},this.ws.onclose=r=>this.onClose({description:"websocket connection closed",context:r}),this.ws.onmessage=r=>this.onData(r.data),this.ws.onerror=r=>this.onError("websocket error",r)}write(r){this.writable=!1;for(let o=0;o<r.length;o++){const s=r[o],c=o===r.length-1;lQe(s,this.supportsBinary,d=>{try{this.doWrite(s,d)}catch{}c&&xEe(()=>{this.writable=!0,this.emitReserved("drain")},this.setTimeoutFn)})}}doClose(){typeof this.ws<"u"&&(this.ws.onerror=()=>{},this.ws.close(),this.ws=null)}uri(){const r=this.opts.secure?"wss":"ws",o=this.query||{};return this.opts.timestampRequests&&(o[this.opts.timestampParam]=Dln()),this.supportsBinary||(o.b64=1),this.createUri(r,o)}}const wFe=P_.WebSocket||P_.MozWebSocket;class e8r extends QMr{createSocket(r,o,s){return $ln?new wFe(r,o,s):o?new wFe(r,o):new wFe(r)}doWrite(r,o){this.ws.send(o)}}class t8r extends fQe{get name(){return"webtransport"}doOpen(){try{this._transport=new WebTransport(this.createUri("https"),this.opts.transportOptions[this.name])}catch(r){return this.emitReserved("error",r)}this._transport.closed.then(()=>{this.onClose()}).catch(r=>{this.onError("webtransport error",r)}),this._transport.ready.then(()=>{this._transport.createBidirectionalStream().then(r=>{const o=MMr(Number.MAX_SAFE_INTEGER,this.socket.binaryType),s=r.readable.pipeThrough(o).getReader(),c=DMr();c.readable.pipeTo(r.writable),this._writer=c.writable.getWriter();const d=()=>{s.read().then(({done:b,value:y})=>{b||(this.onPacket(y),d())}).catch(b=>{})};d();const h={type:"open"};this.query.sid&&(h.data=`{"sid":"${this.query.sid}"}`),this._writer.write(h).then(()=>this.onOpen())})})}write(r){this.writable=!1;for(let o=0;o<r.length;o++){const s=r[o],c=o===r.length-1;this._writer.write(s).then(()=>{c&&xEe(()=>{this.writable=!0,this.emitReserved("drain")},this.setTimeoutFn)})}}doClose(){var r;(r=this._transport)===null||r===void 0||r.close()}}const n8r={websocket:e8r,webtransport:t8r,polling:ZMr},r8r=/^(?:(?![^:@\/?#]+:[^:@\/]*@)(http|https|ws|wss):\/\/)?((?:(([^:@\/?#]*)(?::([^:@\/?#]*))?)?@)?((?:[a-f0-9]{0,4}:){2,7}[a-f0-9]{0,4}|[^:\/?#]*)(?::(\d*))?)(((\/(?:[^?#](?![^?#\/]*\.[^?#\/.]+(?:[?#]|$)))*\/?)?([^?#\/]*))(?:\?([^#]*))?(?:#(.*))?)/,i8r=["source","protocol","authority","userInfo","user","password","host","port","relative","path","directory","file","query","anchor"];function tKe(n){if(n.length>8e3)throw"URI too long";const r=n,o=n.indexOf("["),s=n.indexOf("]");o!=-1&&s!=-1&&(n=n.substring(0,o)+n.substring(o,s).replace(/:/g,";")+n.substring(s,n.length));let c=r8r.exec(n||""),d={},h=14;for(;h--;)d[i8r[h]]=c[h]||"";return o!=-1&&s!=-1&&(d.source=r,d.host=d.host.substring(1,d.host.length-1).replace(/;/g,":"),d.authority=d.authority.replace("[","").replace("]","").replace(/;/g,":"),d.ipv6uri=!0),d.pathNames=o8r(d,d.path),d.queryKey=a8r(d,d.query),d}function o8r(n,r){const o=/\/{2,9}/g,s=r.replace(o,"/").split("/");return(r.slice(0,1)=="/"||r.length===0)&&s.splice(0,1),r.slice(-1)=="/"&&s.splice(s.length-1,1),s}function a8r(n,r){const o={};return r.replace(/(?:^|&)([^&=]*)=?([^&]*)/g,function(s,c,d){c&&(o[c]=d)}),o}const nKe=typeof addEventListener=="function"&&typeof removeEventListener=="function",Ave=[];nKe&&addEventListener("offline",()=>{Ave.forEach(n=>n())},!1);class EM extends Xp{constructor(r,o){if(super(),this.binaryType=$Mr,this.writeBuffer=[],this._prevBufferLen=0,this._pingInterval=-1,this._pingTimeout=-1,this._maxPayload=-1,this._pingTimeoutTime=1/0,r&&typeof r=="object"&&(o=r,r=null),r){const s=tKe(r);o.hostname=s.host,o.secure=s.protocol==="https"||s.protocol==="wss",o.port=s.port,s.query&&(o.query=s.query)}else o.host&&(o.hostname=tKe(o.host).host);CEe(this,o),this.secure=o.secure!=null?o.secure:typeof location<"u"&&location.protocol==="https:",o.hostname&&!o.port&&(o.port=this.secure?"443":"80"),this.hostname=o.hostname||(typeof location<"u"?location.hostname:"localhost"),this.port=o.port||(typeof location<"u"&&location.port?location.port:this.secure?"443":"80"),this.transports=[],this._transportsByName={},o.transports.forEach(s=>{const c=s.prototype.name;this.transports.push(c),this._transportsByName[c]=s}),this.opts=Object.assign({path:"/engine.io",agent:!1,withCredentials:!1,upgrade:!0,timestampParam:"t",rememberUpgrade:!1,addTrailingSlash:!0,rejectUnauthorized:!0,perMessageDeflate:{threshold:1024},transportOptions:{},closeOnBeforeunload:!1},o),this.opts.path=this.opts.path.replace(/\/$/,"")+(this.opts.addTrailingSlash?"/":""),typeof this.opts.query=="string"&&(this.opts.query=qMr(this.opts.query)),nKe&&(this.opts.closeOnBeforeunload&&(this._beforeunloadEventListener=()=>{this.transport&&(this.transport.removeAllListeners(),this.transport.close())},addEventListener("beforeunload",this._beforeunloadEventListener,!1)),this.hostname!=="localhost"&&(this._offlineEventListener=()=>{this._onClose("transport close",{description:"network connection lost"})},Ave.push(this._offlineEventListener))),this.opts.withCredentials&&(this._cookieJar=void 0),this._open()}createTransport(r){const o=Object.assign({},this.opts.query);o.EIO=Pln,o.transport=r,this.id&&(o.sid=this.id);const s=Object.assign({},this.opts,{query:o,socket:this,hostname:this.hostname,secure:this.secure,port:this.port},this.opts.transportOptions[r]);return new this._transportsByName[r](s)}_open(){if(this.transports.length===0){this.setTimeoutFn(()=>{this.emitReserved("error","No transports available")},0);return}const r=this.opts.rememberUpgrade&&EM.priorWebsocketSuccess&&this.transports.indexOf("websocket")!==-1?"websocket":this.transports[0];this.readyState="opening";const o=this.createTransport(r);o.open(),this.setTransport(o)}setTransport(r){this.transport&&this.transport.removeAllListeners(),this.transport=r,r.on("drain",this._onDrain.bind(this)).on("packet",this._onPacket.bind(this)).on("error",this._onError.bind(this)).on("close",o=>this._onClose("transport close",o))}onOpen(){this.readyState="open",EM.priorWebsocketSuccess=this.transport.name==="websocket",this.emitReserved("open"),this.flush()}_onPacket(r){if(this.readyState==="opening"||this.readyState==="open"||this.readyState==="closing")switch(this.emitReserved("packet",r),this.emitReserved("heartbeat"),r.type){case"open":this.onHandshake(JSON.parse(r.data));break;case"ping":this._sendPacket("pong"),this.emitReserved("ping"),this.emitReserved("pong"),this._resetPingTimeout();break;case"error":const o=new Error("server error");o.code=r.data,this._onError(o);break;case"message":this.emitReserved("data",r.data),this.emitReserved("message",r.data);break}}onHandshake(r){this.emitReserved("handshake",r),this.id=r.sid,this.transport.query.sid=r.sid,this._pingInterval=r.pingInterval,this._pingTimeout=r.pingTimeout,this._maxPayload=r.maxPayload,this.onOpen(),this.readyState!=="closed"&&this._resetPingTimeout()}_resetPingTimeout(){this.clearTimeoutFn(this._pingTimeoutTimer);const r=this._pingInterval+this._pingTimeout;this._pingTimeoutTime=Date.now()+r,this._pingTimeoutTimer=this.setTimeoutFn(()=>{this._onClose("ping timeout")},r),this.opts.autoUnref&&this._pingTimeoutTimer.unref()}_onDrain(){this.writeBuffer.splice(0,this._prevBufferLen),this._prevBufferLen=0,this.writeBuffer.length===0?this.emitReserved("drain"):this.flush()}flush(){if(this.readyState!=="closed"&&this.transport.writable&&!this.upgrading&&this.writeBuffer.length){const r=this._getWritablePackets();this.transport.send(r),this._prevBufferLen=r.length,this.emitReserved("flush")}}_getWritablePackets(){if(!(this._maxPayload&&this.transport.name==="polling"&&this.writeBuffer.length>1))return this.writeBuffer;let o=1;for(let s=0;s<this.writeBuffer.length;s++){const c=this.writeBuffer[s].data;if(c&&(o+=zMr(c)),s>0&&o>this._maxPayload)return this.writeBuffer.slice(0,s);o+=2}return this.writeBuffer}_hasPingExpired(){if(!this._pingTimeoutTime)return!0;const r=Date.now()>this._pingTimeoutTime;return r&&(this._pingTimeoutTime=0,xEe(()=>{this._onClose("ping timeout")},this.setTimeoutFn)),r}write(r,o,s){return this._sendPacket("message",r,o,s),this}send(r,o,s){return this._sendPacket("message",r,o,s),this}_sendPacket(r,o,s,c){if(typeof o=="function"&&(c=o,o=void 0),typeof s=="function"&&(c=s,s=null),this.readyState==="closing"||this.readyState==="closed")return;s=s||{},s.compress=s.compress!==!1;const d={type:r,data:o,options:s};this.emitReserved("packetCreate",d),this.writeBuffer.push(d),c&&this.once("flush",c),this.flush()}close(){const r=()=>{this._onClose("forced close"),this.transport.close()},o=()=>{this.off("upgrade",o),this.off("upgradeError",o),r()},s=()=>{this.once("upgrade",o),this.once("upgradeError",o)};return(this.readyState==="opening"||this.readyState==="open")&&(this.readyState="closing",this.writeBuffer.length?this.once("drain",()=>{this.upgrading?s():r()}):this.upgrading?s():r()),this}_onError(r){if(EM.priorWebsocketSuccess=!1,this.opts.tryAllTransports&&this.transports.length>1&&this.readyState==="opening")return this.transports.shift(),this._open();this.emitReserved("error",r),this._onClose("transport error",r)}_onClose(r,o){if(this.readyState==="opening"||this.readyState==="open"||this.readyState==="closing"){if(this.clearTimeoutFn(this._pingTimeoutTimer),this.transport.removeAllListeners("close"),this.transport.close(),this.transport.removeAllListeners(),nKe&&(this._beforeunloadEventListener&&removeEventListener("beforeunload",this._beforeunloadEventListener,!1),this._offlineEventListener)){const s=Ave.indexOf(this._offlineEventListener);s!==-1&&Ave.splice(s,1)}this.readyState="closed",this.id=null,this.emitReserved("close",r,o),this.writeBuffer=[],this._prevBufferLen=0}}}EM.protocol=Pln;class s8r extends EM{constructor(){super(...arguments),this._upgrades=[]}onOpen(){if(super.onOpen(),this.readyState==="open"&&this.opts.upgrade)for(let r=0;r<this._upgrades.length;r++)this._probe(this._upgrades[r])}_probe(r){let o=this.createTransport(r),s=!1;EM.priorWebsocketSuccess=!1;const c=()=>{s||(o.send([{type:"ping",data:"probe"}]),o.once("packet",x=>{if(!s)if(x.type==="pong"&&x.data==="probe"){if(this.upgrading=!0,this.emitReserved("upgrading",o),!o)return;EM.priorWebsocketSuccess=o.name==="websocket",this.transport.pause(()=>{s||this.readyState!=="closed"&&(k(),this.setTransport(o),o.send([{type:"upgrade"}]),this.emitReserved("upgrade",o),o=null,this.upgrading=!1,this.flush())})}else{const O=new Error("probe error");O.transport=o.name,this.emitReserved("upgradeError",O)}}))};function d(){s||(s=!0,k(),o.close(),o=null)}const h=x=>{const O=new Error("probe error: "+x);O.transport=o.name,d(),this.emitReserved("upgradeError",O)};function b(){h("transport closed")}function y(){h("socket closed")}function E(x){o&&x.name!==o.name&&d()}const k=()=>{o.removeListener("open",c),o.removeListener("error",h),o.removeListener("close",b),this.off("close",y),this.off("upgrading",E)};o.once("open",c),o.once("error",h),o.once("close",b),this.once("close",y),this.once("upgrading",E),this._upgrades.indexOf("webtransport")!==-1&&r!=="webtransport"?this.setTimeoutFn(()=>{s||o.open()},200):o.open()}onHandshake(r){this._upgrades=this._filterUpgrades(r.upgrades),super.onHandshake(r)}_filterUpgrades(r){const o=[];for(let s=0;s<r.length;s++)~this.transports.indexOf(r[s])&&o.push(r[s]);return o}}let c8r=class extends s8r{constructor(r,o={}){const s=typeof r=="object"?r:o;(!s.transports||s.transports&&typeof s.transports[0]=="string")&&(s.transports=(s.transports||["polling","websocket","webtransport"]).map(c=>n8r[c]).filter(c=>!!c)),super(r,s)}};function u8r(n,r="",o){let s=n;o=o||typeof location<"u"&&location,n==null&&(n=o.protocol+"//"+o.host),typeof n=="string"&&(n.charAt(0)==="/"&&(n.charAt(1)==="/"?n=o.protocol+n:n=o.host+n),/^(https?|wss?):\/\//.test(n)||(typeof o<"u"?n=o.protocol+"//"+n:n="https://"+n),s=tKe(n)),s.port||(/^(http|ws)$/.test(s.protocol)?s.port="80":/^(http|ws)s$/.test(s.protocol)&&(s.port="443")),s.path=s.path||"/";const d=s.host.indexOf(":")!==-1?"["+s.host+"]":s.host;return s.id=s.protocol+"://"+d+":"+s.port+r,s.href=s.protocol+"://"+d+(o&&o.port===s.port?"":":"+s.port),s}const l8r=typeof ArrayBuffer=="function",d8r=n=>typeof ArrayBuffer.isView=="function"?ArrayBuffer.isView(n):n.buffer instanceof ArrayBuffer,Fln=Object.prototype.toString,f8r=typeof Blob=="function"||typeof Blob<"u"&&Fln.call(Blob)==="[object BlobConstructor]",h8r=typeof File=="function"||typeof File<"u"&&Fln.call(File)==="[object FileConstructor]";function hQe(n){return l8r&&(n instanceof ArrayBuffer||d8r(n))||f8r&&n instanceof Blob||h8r&&n instanceof File}function Ive(n,r){if(!n||typeof n!="object")return!1;if(Array.isArray(n)){for(let o=0,s=n.length;o<s;o++)if(Ive(n[o]))return!0;return!1}if(hQe(n))return!0;if(n.toJSON&&typeof n.toJSON=="function"&&arguments.length===1)return Ive(n.toJSON(),!0);for(const o in n)if(Object.prototype.hasOwnProperty.call(n,o)&&Ive(n[o]))return!0;return!1}function p8r(n){const r=[],o=n.data,s=n;return s.data=rKe(o,r),s.attachments=r.length,{packet:s,buffers:r}}function rKe(n,r){if(!n)return n;if(hQe(n)){const o={_placeholder:!0,num:r.length};return r.push(n),o}else if(Array.isArray(n)){const o=new Array(n.length);for(let s=0;s<n.length;s++)o[s]=rKe(n[s],r);return o}else if(typeof n=="object"&&!(n instanceof Date)){const o={};for(const s in n)Object.prototype.hasOwnProperty.call(n,s)&&(o[s]=rKe(n[s],r));return o}return n}function g8r(n,r){return n.data=iKe(n.data,r),delete n.attachments,n}function iKe(n,r){if(!n)return n;if(n&&n._placeholder===!0){if(typeof n.num=="number"&&n.num>=0&&n.num<r.length)return r[n.num];throw new Error("illegal attachments")}else if(Array.isArray(n))for(let o=0;o<n.length;o++)n[o]=iKe(n[o],r);else if(typeof n=="object")for(const o in n)Object.prototype.hasOwnProperty.call(n,o)&&(n[o]=iKe(n[o],r));return n}const m8r=["connect","connect_error","disconnect","disconnecting","newListener","removeListener"],b8r=5;var jc;(function(n){n[n.CONNECT=0]="CONNECT",n[n.DISCONNECT=1]="DISCONNECT",n[n.EVENT=2]="EVENT",n[n.ACK=3]="ACK",n[n.CONNECT_ERROR=4]="CONNECT_ERROR",n[n.BINARY_EVENT=5]="BINARY_EVENT",n[n.BINARY_ACK=6]="BINARY_ACK"})(jc||(jc={}));class v8r{constructor(r){this.replacer=r}encode(r){return(r.type===jc.EVENT||r.type===jc.ACK)&&Ive(r)?this.encodeAsBinary({type:r.type===jc.EVENT?jc.BINARY_EVENT:jc.BINARY_ACK,nsp:r.nsp,data:r.data,id:r.id}):[this.encodeAsString(r)]}encodeAsString(r){let o=""+r.type;return(r.type===jc.BINARY_EVENT||r.type===jc.BINARY_ACK)&&(o+=r.attachments+"-"),r.nsp&&r.nsp!=="/"&&(o+=r.nsp+","),r.id!=null&&(o+=r.id),r.data!=null&&(o+=JSON.stringify(r.data,this.replacer)),o}encodeAsBinary(r){const o=p8r(r),s=this.encodeAsString(o.packet),c=o.buffers;return c.unshift(s),c}}function ZBt(n){return Object.prototype.toString.call(n)==="[object Object]"}class pQe extends Xp{constructor(r){super(),this.reviver=r}add(r){let o;if(typeof r=="string"){if(this.reconstructor)throw new Error("got plaintext data when reconstructing a packet");o=this.decodeString(r);const s=o.type===jc.BINARY_EVENT;s||o.type===jc.BINARY_ACK?(o.type=s?jc.EVENT:jc.ACK,this.reconstructor=new w8r(o),o.attachments===0&&super.emitReserved("decoded",o)):super.emitReserved("decoded",o)}else if(hQe(r)||r.base64)if(this.reconstructor)o=this.reconstructor.takeBinaryData(r),o&&(this.reconstructor=null,super.emitReserved("decoded",o));else throw new Error("got binary data when not reconstructing a packet");else throw new Error("Unknown type: "+r)}decodeString(r){let o=0;const s={type:Number(r.charAt(0))};if(jc[s.type]===void 0)throw new Error("unknown packet type "+s.type);if(s.type===jc.BINARY_EVENT||s.type===jc.BINARY_ACK){const d=o+1;for(;r.charAt(++o)!=="-"&&o!=r.length;);const h=r.substring(d,o);if(h!=Number(h)||r.charAt(o)!=="-")throw new Error("Illegal attachments");s.attachments=Number(h)}if(r.charAt(o+1)==="/"){const d=o+1;for(;++o&&!(r.charAt(o)===","||o===r.length););s.nsp=r.substring(d,o)}else s.nsp="/";const c=r.charAt(o+1);if(c!==""&&Number(c)==c){const d=o+1;for(;++o;){const h=r.charAt(o);if(h==null||Number(h)!=h){--o;break}if(o===r.length)break}s.id=Number(r.substring(d,o+1))}if(r.charAt(++o)){const d=this.tryParse(r.substr(o));if(pQe.isPayloadValid(s.type,d))s.data=d;else throw new Error("invalid payload")}return s}tryParse(r){try{return JSON.parse(r,this.reviver)}catch{return!1}}static isPayloadValid(r,o){switch(r){case jc.CONNECT:return ZBt(o);case jc.DISCONNECT:return o===void 0;case jc.CONNECT_ERROR:return typeof o=="string"||ZBt(o);case jc.EVENT:case jc.BINARY_EVENT:return Array.isArray(o)&&(typeof o[0]=="number"||typeof o[0]=="string"&&m8r.indexOf(o[0])===-1);case jc.ACK:case jc.BINARY_ACK:return Array.isArray(o)}}destroy(){this.reconstructor&&(this.reconstructor.finishedReconstruction(),this.reconstructor=null)}}class w8r{constructor(r){this.packet=r,this.buffers=[],this.reconPack=r}takeBinaryData(r){if(this.buffers.push(r),this.buffers.length===this.reconPack.attachments){const o=g8r(this.reconPack,this.buffers);return this.finishedReconstruction(),o}return null}finishedReconstruction(){this.reconPack=null,this.buffers=[]}}const y8r=Object.freeze(Object.defineProperty({__proto__:null,Decoder:pQe,Encoder:v8r,get PacketType(){return jc},protocol:b8r},Symbol.toStringTag,{value:"Module"}));function Ax(n,r,o){return n.on(r,o),function(){n.off(r,o)}}const E8r=Object.freeze({connect:1,connect_error:1,disconnect:1,disconnecting:1,newListener:1,removeListener:1});class Bln extends Xp{constructor(r,o,s){super(),this.connected=!1,this.recovered=!1,this.receiveBuffer=[],this.sendBuffer=[],this._queue=[],this._queueSeq=0,this.ids=0,this.acks={},this.flags={},this.io=r,this.nsp=o,s&&s.auth&&(this.auth=s.auth),this._opts=Object.assign({},s),this.io._autoConnect&&this.open()}get disconnected(){return!this.connected}subEvents(){if(this.subs)return;const r=this.io;this.subs=[Ax(r,"open",this.onopen.bind(this)),Ax(r,"packet",this.onpacket.bind(this)),Ax(r,"error",this.onerror.bind(this)),Ax(r,"close",this.onclose.bind(this))]}get active(){return!!this.subs}connect(){return this.connected?this:(this.subEvents(),this.io._reconnecting||this.io.open(),this.io._readyState==="open"&&this.onopen(),this)}open(){return this.connect()}send(...r){return r.unshift("message"),this.emit.apply(this,r),this}emit(r,...o){var s,c,d;if(E8r.hasOwnProperty(r))throw new Error('"'+r.toString()+'" is a reserved event name');if(o.unshift(r),this._opts.retries&&!this.flags.fromQueue&&!this.flags.volatile)return this._addToQueue(o),this;const h={type:jc.EVENT,data:o};if(h.options={},h.options.compress=this.flags.compress!==!1,typeof o[o.length-1]=="function"){const k=this.ids++,x=o.pop();this._registerAckCallback(k,x),h.id=k}const b=(c=(s=this.io.engine)===null||s===void 0?void 0:s.transport)===null||c===void 0?void 0:c.writable,y=this.connected&&!(!((d=this.io.engine)===null||d===void 0)&&d._hasPingExpired());return this.flags.volatile&&!b||(y?(this.notifyOutgoingListeners(h),this.packet(h)):this.sendBuffer.push(h)),this.flags={},this}_registerAckCallback(r,o){var s;const c=(s=this.flags.timeout)!==null&&s!==void 0?s:this._opts.ackTimeout;if(c===void 0){this.acks[r]=o;return}const d=this.io.setTimeoutFn(()=>{delete this.acks[r];for(let b=0;b<this.sendBuffer.length;b++)this.sendBuffer[b].id===r&&this.sendBuffer.splice(b,1);o.call(this,new Error("operation has timed out"))},c),h=(...b)=>{this.io.clearTimeoutFn(d),o.apply(this,b)};h.withError=!0,this.acks[r]=h}emitWithAck(r,...o){return new Promise((s,c)=>{const d=(h,b)=>h?c(h):s(b);d.withError=!0,o.push(d),this.emit(r,...o)})}_addToQueue(r){let o;typeof r[r.length-1]=="function"&&(o=r.pop());const s={id:this._queueSeq++,tryCount:0,pending:!1,args:r,flags:Object.assign({fromQueue:!0},this.flags)};r.push((c,...d)=>s!==this._queue[0]?void 0:(c!==null?s.tryCount>this._opts.retries&&(this._queue.shift(),o&&o(c)):(this._queue.shift(),o&&o(null,...d)),s.pending=!1,this._drainQueue())),this._queue.push(s),this._drainQueue()}_drainQueue(r=!1){if(!this.connected||this._queue.length===0)return;const o=this._queue[0];o.pending&&!r||(o.pending=!0,o.tryCount++,this.flags=o.flags,this.emit.apply(this,o.args))}packet(r){r.nsp=this.nsp,this.io._packet(r)}onopen(){typeof this.auth=="function"?this.auth(r=>{this._sendConnectPacket(r)}):this._sendConnectPacket(this.auth)}_sendConnectPacket(r){this.packet({type:jc.CONNECT,data:this._pid?Object.assign({pid:this._pid,offset:this._lastOffset},r):r})}onerror(r){this.connected||this.emitReserved("connect_error",r)}onclose(r,o){this.connected=!1,delete this.id,this.emitReserved("disconnect",r,o),this._clearAcks()}_clearAcks(){Object.keys(this.acks).forEach(r=>{if(!this.sendBuffer.some(s=>String(s.id)===r)){const s=this.acks[r];delete this.acks[r],s.withError&&s.call(this,new Error("socket has been disconnected"))}})}onpacket(r){if(r.nsp===this.nsp)switch(r.type){case jc.CONNECT:r.data&&r.data.sid?this.onconnect(r.data.sid,r.data.pid):this.emitReserved("connect_error",new Error("It seems you are trying to reach a Socket.IO server in v2.x with a v3.x client, but they are not compatible (more information here: https://socket.io/docs/v3/migrating-from-2-x-to-3-0/)"));break;case jc.EVENT:case jc.BINARY_EVENT:this.onevent(r);break;case jc.ACK:case jc.BINARY_ACK:this.onack(r);break;case jc.DISCONNECT:this.ondisconnect();break;case jc.CONNECT_ERROR:this.destroy();const s=new Error(r.data.message);s.data=r.data.data,this.emitReserved("connect_error",s);break}}onevent(r){const o=r.data||[];r.id!=null&&o.push(this.ack(r.id)),this.connected?this.emitEvent(o):this.receiveBuffer.push(Object.freeze(o))}emitEvent(r){if(this._anyListeners&&this._anyListeners.length){const o=this._anyListeners.slice();for(const s of o)s.apply(this,r)}super.emit.apply(this,r),this._pid&&r.length&&typeof r[r.length-1]=="string"&&(this._lastOffset=r[r.length-1])}ack(r){const o=this;let s=!1;return function(...c){s||(s=!0,o.packet({type:jc.ACK,id:r,data:c}))}}onack(r){const o=this.acks[r.id];typeof o=="function"&&(delete this.acks[r.id],o.withError&&r.data.unshift(null),o.apply(this,r.data))}onconnect(r,o){this.id=r,this.recovered=o&&this._pid===o,this._pid=o,this.connected=!0,this.emitBuffered(),this.emitReserved("connect"),this._drainQueue(!0)}emitBuffered(){this.receiveBuffer.forEach(r=>this.emitEvent(r)),this.receiveBuffer=[],this.sendBuffer.forEach(r=>{this.notifyOutgoingListeners(r),this.packet(r)}),this.sendBuffer=[]}ondisconnect(){this.destroy(),this.onclose("io server disconnect")}destroy(){this.subs&&(this.subs.forEach(r=>r()),this.subs=void 0),this.io._destroy(this)}disconnect(){return this.connected&&this.packet({type:jc.DISCONNECT}),this.destroy(),this.connected&&this.onclose("io client disconnect"),this}close(){return this.disconnect()}compress(r){return this.flags.compress=r,this}get volatile(){return this.flags.volatile=!0,this}timeout(r){return this.flags.timeout=r,this}onAny(r){return this._anyListeners=this._anyListeners||[],this._anyListeners.push(r),this}prependAny(r){return this._anyListeners=this._anyListeners||[],this._anyListeners.unshift(r),this}offAny(r){if(!this._anyListeners)return this;if(r){const o=this._anyListeners;for(let s=0;s<o.length;s++)if(r===o[s])return o.splice(s,1),this}else this._anyListeners=[];return this}listenersAny(){return this._anyListeners||[]}onAnyOutgoing(r){return this._anyOutgoingListeners=this._anyOutgoingListeners||[],this._anyOutgoingListeners.push(r),this}prependAnyOutgoing(r){return this._anyOutgoingListeners=this._anyOutgoingListeners||[],this._anyOutgoingListeners.unshift(r),this}offAnyOutgoing(r){if(!this._anyOutgoingListeners)return this;if(r){const o=this._anyOutgoingListeners;for(let s=0;s<o.length;s++)if(r===o[s])return o.splice(s,1),this}else this._anyOutgoingListeners=[];return this}listenersAnyOutgoing(){return this._anyOutgoingListeners||[]}notifyOutgoingListeners(r){if(this._anyOutgoingListeners&&this._anyOutgoingListeners.length){const o=this._anyOutgoingListeners.slice();for(const s of o)s.apply(this,r.data)}}}function DK(n){n=n||{},this.ms=n.min||100,this.max=n.max||1e4,this.factor=n.factor||2,this.jitter=n.jitter>0&&n.jitter<=1?n.jitter:0,this.attempts=0}DK.prototype.duration=function(){var n=this.ms*Math.pow(this.factor,this.attempts++);if(this.jitter){var r=Math.random(),o=Math.floor(r*this.jitter*n);n=(Math.floor(r*10)&1)==0?n-o:n+o}return Math.min(n,this.max)|0};DK.prototype.reset=function(){this.attempts=0};DK.prototype.setMin=function(n){this.ms=n};DK.prototype.setMax=function(n){this.max=n};DK.prototype.setJitter=function(n){this.jitter=n};class oKe extends Xp{constructor(r,o){var s;super(),this.nsps={},this.subs=[],r&&typeof r=="object"&&(o=r,r=void 0),o=o||{},o.path=o.path||"/socket.io",this.opts=o,CEe(this,o),this.reconnection(o.reconnection!==!1),this.reconnectionAttempts(o.reconnectionAttempts||1/0),this.reconnectionDelay(o.reconnectionDelay||1e3),this.reconnectionDelayMax(o.reconnectionDelayMax||5e3),this.randomizationFactor((s=o.randomizationFactor)!==null&&s!==void 0?s:.5),this.backoff=new DK({min:this.reconnectionDelay(),max:this.reconnectionDelayMax(),jitter:this.randomizationFactor()}),this.timeout(o.timeout==null?2e4:o.timeout),this._readyState="closed",this.uri=r;const c=o.parser||y8r;this.encoder=new c.Encoder,this.decoder=new c.Decoder,this._autoConnect=o.autoConnect!==!1,this._autoConnect&&this.open()}reconnection(r){return arguments.length?(this._reconnection=!!r,r||(this.skipReconnect=!0),this):this._reconnection}reconnectionAttempts(r){return r===void 0?this._reconnectionAttempts:(this._reconnectionAttempts=r,this)}reconnectionDelay(r){var o;return r===void 0?this._reconnectionDelay:(this._reconnectionDelay=r,(o=this.backoff)===null||o===void 0||o.setMin(r),this)}randomizationFactor(r){var o;return r===void 0?this._randomizationFactor:(this._randomizationFactor=r,(o=this.backoff)===null||o===void 0||o.setJitter(r),this)}reconnectionDelayMax(r){var o;return r===void 0?this._reconnectionDelayMax:(this._reconnectionDelayMax=r,(o=this.backoff)===null||o===void 0||o.setMax(r),this)}timeout(r){return arguments.length?(this._timeout=r,this):this._timeout}maybeReconnectOnOpen(){!this._reconnecting&&this._reconnection&&this.backoff.attempts===0&&this.reconnect()}open(r){if(~this._readyState.indexOf("open"))return this;this.engine=new c8r(this.uri,this.opts);const o=this.engine,s=this;this._readyState="opening",this.skipReconnect=!1;const c=Ax(o,"open",function(){s.onopen(),r&&r()}),d=b=>{this.cleanup(),this._readyState="closed",this.emitReserved("error",b),r?r(b):this.maybeReconnectOnOpen()},h=Ax(o,"error",d);if(this._timeout!==!1){const b=this._timeout,y=this.setTimeoutFn(()=>{c(),d(new Error("timeout")),o.close()},b);this.opts.autoUnref&&y.unref(),this.subs.push(()=>{this.clearTimeoutFn(y)})}return this.subs.push(c),this.subs.push(h),this}connect(r){return this.open(r)}onopen(){this.cleanup(),this._readyState="open",this.emitReserved("open");const r=this.engine;this.subs.push(Ax(r,"ping",this.onping.bind(this)),Ax(r,"data",this.ondata.bind(this)),Ax(r,"error",this.onerror.bind(this)),Ax(r,"close",this.onclose.bind(this)),Ax(this.decoder,"decoded",this.ondecoded.bind(this)))}onping(){this.emitReserved("ping")}ondata(r){try{this.decoder.add(r)}catch(o){this.onclose("parse error",o)}}ondecoded(r){xEe(()=>{this.emitReserved("packet",r)},this.setTimeoutFn)}onerror(r){this.emitReserved("error",r)}socket(r,o){let s=this.nsps[r];return s?this._autoConnect&&!s.active&&s.connect():(s=new Bln(this,r,o),this.nsps[r]=s),s}_destroy(r){const o=Object.keys(this.nsps);for(const s of o)if(this.nsps[s].active)return;this._close()}_packet(r){const o=this.encoder.encode(r);for(let s=0;s<o.length;s++)this.engine.write(o[s],r.options)}cleanup(){this.subs.forEach(r=>r()),this.subs.length=0,this.decoder.destroy()}_close(){this.skipReconnect=!0,this._reconnecting=!1,this.onclose("forced close")}disconnect(){return this._close()}onclose(r,o){var s;this.cleanup(),(s=this.engine)===null||s===void 0||s.close(),this.backoff.reset(),this._readyState="closed",this.emitReserved("close",r,o),this._reconnection&&!this.skipReconnect&&this.reconnect()}reconnect(){if(this._reconnecting||this.skipReconnect)return this;const r=this;if(this.backoff.attempts>=this._reconnectionAttempts)this.backoff.reset(),this.emitReserved("reconnect_failed"),this._reconnecting=!1;else{const o=this.backoff.duration();this._reconnecting=!0;const s=this.setTimeoutFn(()=>{r.skipReconnect||(this.emitReserved("reconnect_attempt",r.backoff.attempts),!r.skipReconnect&&r.open(c=>{c?(r._reconnecting=!1,r.reconnect(),this.emitReserved("reconnect_error",c)):r.onreconnect()}))},o);this.opts.autoUnref&&s.unref(),this.subs.push(()=>{this.clearTimeoutFn(s)})}}onreconnect(){const r=this.backoff.attempts;this._reconnecting=!1,this.backoff.reset(),this.emitReserved("reconnect",r)}}const _re={};function Ove(n,r){typeof n=="object"&&(r=n,n=void 0),r=r||{};const o=u8r(n,r.path||"/socket.io"),s=o.source,c=o.id,d=o.path,h=_re[c]&&d in _re[c].nsps,b=r.forceNew||r["force new connection"]||r.multiplex===!1||h;let y;return b?y=new oKe(s,r):(_re[c]||(_re[c]=new oKe(s,r)),y=_re[c]),o.query&&!r.query&&(r.query=o.queryKey),y.socket(o.path,r)}Object.assign(Ove,{Manager:oKe,Socket:Bln,io:Ove,connect:Ove});function AEe(n,r){const[o,s]=Q.useState(()=>S8r(n,r)),c=Q.useRef(n),d=Q.useRef(r);c.current=n,d.current=r,Q.useEffect(()=>{const y=E=>{if(E.detail.key!==c.current)return;const k=E.detail.newValue?JSON.parse(E.detail.newValue):d.current;s(k)};return globalThis.addEventListener("local-storage-changed",y),()=>{globalThis.removeEventListener("local-storage-changed",y)}},[]);const h=Q.useCallback(y=>{const E=typeof y=="function"?y(o):y;try{const k=localStorage.getItem(c.current),x=JSON.stringify(E);localStorage.setItem(c.current,x),globalThis.dispatchEvent(new CustomEvent("local-storage-changed",{detail:{key:c.current,newValue:x,oldValue:k}}))}catch(k){console.warn(`Failed to set localStorage key "${c.current}":`,k)}},[o]),b=Q.useCallback(()=>{const y=localStorage.getItem(c.current);localStorage.removeItem(c.current),globalThis.dispatchEvent(new CustomEvent("local-storage-changed",{detail:{key:c.current,newValue:null,oldValue:y}}))},[]);return[o,h,b]}function S8r(n,r){try{const o=localStorage.getItem(n);return o===null?r:o==="undefined"?void 0:JSON.parse(o)}catch(o){return Va(o,{tags:{component:"useLocalStorage",action:"readFromStorage"},extra:{key:n}}),r}}function _8r(n){const r=new Map;for(const o of n)if(o.event==="flow_started"&&o.flow_id)r.set(o.flow_id,{flowId:o.flow_id,startedAt:o.timestamp,tokens:0,latencies:[]});else if(o.event==="flow_completed"&&o.flow_id){const s=r.get(o.flow_id);s&&(s.completedAt=o.timestamp)}else if(o.event==="user"&&o.metadata?.prompts)for(const s of o.metadata.prompts){const c=s.llm_response_metadata?.usage?.total_tokens??0,d=s.llm_response_metadata?.latency??0,h=Array.from(r.values()).filter(b=>!b.completedAt);if(h.length>0){const b=h[h.length-1];b.tokens+=c,d>0&&b.latencies.push(d*1e3)}}else if(o.event==="bot"&&o.metadata?.prompts)for(const s of o.metadata.prompts){const c=s.llm_response_metadata?.usage?.total_tokens??0,d=s.llm_response_metadata?.latency??0,h=Array.from(r.values()).filter(b=>!b.completedAt);if(h.length>0){const b=h[h.length-1];b.tokens+=c,d>0&&b.latencies.push(d*1e3)}}return Array.from(r.values()).map(o=>({flowId:o.flowId,flowName:o.flowId,status:o.completedAt?"completed":"active",startedAt:o.startedAt,completedAt:o.completedAt,tokens:o.tokens,latencyMs:o.latencies.length>0?o.latencies.reduce((s,c)=>s+c,0)/o.latencies.length:0}))}const QBt="/session_start";function k8r(n){return n?Object.entries(n).filter(r=>r[1]!=null).map(r=>({name:r[0],value:r[1]})):[]}function T8r(n){const r=n[n.length-1];return r&&!UB(r)&&r.name==="action_listen"}function x8r(){return n2().replaceAll("-","")}function C8r(n){return{id:x8r(),text:n,tokens:[],timestamp:`${Date.now()}`,type:Vx.User,__typename:"Utterance",rephrase:!1,rephrasePrompt:null,metadata:{parseData:{}}}}function Uln({useMemoryOnly:n=!1}={}){const{projectId:r}=G_(),[o,s]=AEe(`${m5r}_${r}`,{}),[c,d]=Q.useState({}),[h,b]=n?[c,d]:[o,s],y=Yt=>({events:[],id:Yt,startDate:new Date().toISOString(),tags:[],reviewed:!1,totalNumberOfUserMessages:0}),[E,k]=Q.useState(n2()),[x,O]=Q.useState(""),[P,D]=Q.useState(void 0),$=Q.useRef(void 0),[U,B]=Q.useState(y(E)),[G,K]=Q.useState([]),[X,ae]=Q.useState(void 0),[oe,he]=Q.useState(!0),[ye,we]=Q.useState([]),ge=Q.useRef(X),[Te,xe]=Q.useState([]),[Ue,Z]=Q.useState(!1),[Re,_e]=Q.useState(!1),[qe,it]=Q.useState([]),Xe=()=>{$.current?.removeAllListeners(),$.current?.io.removeAllListeners(),$.current?.disconnect()},de=()=>{he(!0),Z(!1),_e(!1)},We=Q.useCallback(Yt=>{if($.current?.emit("user_message",{message:Yt,session_id:E}),Yt===QBt)return;z_(H_.userMessageSent,{target:"rasa_agent"});const Or=C8r(Yt);B(Nt=>({...Nt,events:[...Nt.events,Or]})),Z(!1),_e(!1)},[E]);Q.useEffect(()=>{b({...h,[U.id]:U})},[U]),Q.useEffect(()=>{ge.current=X},[X]),Q.useEffect(()=>{if(!$.current&&x){const Yt=new URL(x),Or=Yt.pathname==="/"?void 0:Yt.pathname;return $.current=Ove(Yt.toString(),{transports:["websocket","polling"],path:Or,reconnectionAttempts:xFt,reconnectionDelayMax:2e3}),$.current?.on("connect",()=>{$.current?.emit("session_request",{session_id:E,is_voice:!1})}),$.current?.on("session_confirm",()=>{if(he(!1),D(void 0),Bre(!1),ge.current){ae(void 0),$.current?.emit("update_tracker",ge.current);return}We(QBt)}),$.current?.on("error",Nt=>{D(Nt),de()}),$.current?.on("disconnect",(Nt,on)=>{$.current?.active||(de(),Va(Nt,{tags:{component:"useBotConnection",action:"disconnect"},extra:on}),Nt!=="io client disconnect"&&Bre(!0))}),$.current?.io.on("reconnect_error",Nt=>{de(),Bre(!0),Va(Nt,{tags:{component:"useBotConnection",action:"reconnect_error"}})}),$.current?.io.on("reconnect_failed",()=>{de(),Bre(!0),Va(`websocket wasn't able to reconnect within ${xFt} attempts`,{tags:{component:"useBotConnection",action:"reconnect_failed"}})}),$.current?.on("tracker",Nt=>{if(!Nt)return;if(Nt.sender_id===E){const br=nsn(Nt.events);xe(xIr(br)),we(k8r(Nt.slots));const wi=_8r(Nt.events);it(wi);const no=Nt.stack.map(Ri=>({frameId:Ri.frame_id,flowId:Ri.flow_id,stepId:Ri.step_id,collect:Ri.collect,utter:Ri.utter,ended:!1}));no.length>0&&K(no),B(Ri=>({...Ri,events:br,totalNumberOfUserMessages:br.filter(ro=>UB(ro)).length})),_e(!1),Z(T8r(br))}}),()=>{$.current&&Xe()}}},[x,E,We]),Q.useEffect(()=>{$5r(E)},[E]);const ht=()=>{O(""),Ie()},Ie=Q.useCallback(()=>{he(!0),Xe(),$.current=void 0;const Yt=n2();return B(y(Yt)),k(Yt),d({}),it([]),Yt},[]),$t=Q.useCallback(Yt=>{_e(!0);const Or=Ie();ae({sender_id:Or,events:Yt.map(Nt=>Nt.metadata.rawEvent)})},[_e,Ie]);return{conversationList:Q.useMemo(()=>Object.values(h).sort((Yt,Or)=>new Date(Yt.startDate).getTime()-new Date(Or.startDate).getTime()),[h]),stack:G,inputDisabled:oe,sessionId:E,slots:ye,slotRelatedEvents:Te,flowHistory:qe,replayingConversation:Re,waitingForUserInput:Ue,error:P,setUrl:O,reset:ht,sendMessage:We,replayConversation:$t,setInputDisabled:he,startNewConversation:Ie,cleanup:Xe}}const IEe=(n={})=>{const{autoFetch:r=!0,interval:o}=n,{projectUrl:s,projectId:c,projectInactive:d}=Io(da,P=>({projectUrl:P.projectUrl,projectId:P.projectId,projectInactive:P.inactive})),[h,b]=Q.useState(()=>!document.hidden&&document.visibilityState==="visible");Q.useEffect(()=>{const P=()=>{b(!document.hidden&&document.visibilityState==="visible")};return document.addEventListener("visibilitychange",P),()=>{document.removeEventListener("visibilitychange",P)}},[]);const{data:y,isLoading:E,error:k,refetch:x}=xK({queryKey:["botBuilderStatus"],queryFn:()=>mhr({projectUrl:s||""}),enabled:r&&!!c&&!!s&&h&&!d,refetchInterval:o&&o>0&&h?o:void 0,retry:2});let O=null;return k&&(k instanceof Error?O=k.message:O="Bot Builder: An unknown error occurred"),{data:y??null,loading:E,error:O,refetch:x}},A8r=()=>{const{data:n}=IEe({interval:5e3}),r=Io(da,d=>d.projectUrl),o=Roe(),s=Q.useRef(null),{isStandalone:c}=GM();Q.useEffect(()=>{const d=n?.model_trained_at;if(!c||!d)return;const h=d;if(s.current===null){s.current=h;return}s.current!==h&&(s.current=h,$cn(!1),o.invalidateQueries({queryKey:["files",r]}),o.invalidateQueries({queryKey:["botData",r]}))},[n?.model_trained_at,r,o,c])};var Jre={exports:{}};/**
 * @license
 * Lodash <https://lodash.com/>
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */var I8r=Jre.exports,eUt;function O8r(){return eUt||(eUt=1,function(n,r){(function(){var o,s="4.17.21",c=200,d="Unsupported core-js use. Try https://npms.io/search?q=ponyfill.",h="Expected a function",b="Invalid `variable` option passed into `_.template`",y="__lodash_hash_undefined__",E=500,k="__lodash_placeholder__",x=1,O=2,P=4,D=1,$=2,U=1,B=2,G=4,K=8,X=16,ae=32,oe=64,he=128,ye=256,we=512,ge=30,Te="...",xe=800,Ue=16,Z=1,Re=2,_e=3,qe=1/0,it=9007199254740991,Xe=17976931348623157e292,de=NaN,We=4294967295,ht=We-1,Ie=We>>>1,$t=[["ary",he],["bind",U],["bindKey",B],["curry",K],["curryRight",X],["flip",we],["partial",ae],["partialRight",oe],["rearg",ye]],An="[object Arguments]",Yt="[object Array]",Or="[object AsyncFunction]",Nt="[object Boolean]",on="[object Date]",br="[object DOMException]",wi="[object Error]",no="[object Function]",Ri="[object GeneratorFunction]",ro="[object Map]",Ko="[object Number]",Ca="[object Null]",qi="[object Object]",yi="[object Promise]",lo="[object Proxy]",Eo="[object RegExp]",Fc="[object Set]",Yo="[object String]",zl="[object Symbol]",ld="[object Undefined]",Yc="[object WeakMap]",Sn="[object WeakSet]",Pn="[object ArrayBuffer]",Kr="[object DataView]",Ei="[object Float32Array]",Sa="[object Float64Array]",bs="[object Int8Array]",Cc="[object Int16Array]",Ac="[object Int32Array]",Xc="[object Uint8Array]",rc="[object Uint8ClampedArray]",js="[object Uint16Array]",Jc="[object Uint32Array]",Id=/\b__p \+= '';/g,nh=/\b(__p \+=) '' \+/g,hu=/(__e\(.*?\)|\b__t\)) \+\n'';/g,Qp=/&(?:amp|lt|gt|quot|#39);/g,Kb=/[&<>"']/g,Km=RegExp(Qp.source),S2=RegExp(Kb.source),_2=/<%-([\s\S]+?)%>/g,dv=/<%([\s\S]+?)%>/g,K_=/<%=([\s\S]+?)%>/g,YI=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,n6=/^\w*$/,e0=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,_l=/[\\^$.*+?()[\]{}|]/g,rh=RegExp(_l.source),t0=/^\s+/,k2=/\s/,r6=/\{(?:\n\/\* \[wrapped with .+\] \*\/)?\n?/,i6=/\{\n\/\* \[wrapped with (.+)\] \*/,XI=/,? & /,sy=/[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g,rC=/[()=,{}\[\]\/\s]/,fv=/\\(\\)?/g,cy=/\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g,uy=/\w*$/,ly=/^[-+]0x[0-9a-f]+$/i,Od=/^0b[01]+$/i,JI=/^\[object .+?Constructor\]$/,ZI=/^0o[0-7]+$/i,hv=/^(?:0|[1-9]\d*)$/,dy=/[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g,T2=/($^)/,QI=/['\n\r\u2028\u2029\\]/g,x2="\\ud800-\\udfff",o6="\\u0300-\\u036f",fy="\\ufe20-\\ufe2f",Y_="\\u20d0-\\u20ff",Ym=o6+fy+Y_,C2="\\u2700-\\u27bf",iC="a-z\\xdf-\\xf6\\xf8-\\xff",hy="\\xac\\xb1\\xd7\\xf7",oC="\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf",py="\\u2000-\\u206f",VM=" \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000",a6="A-Z\\xc0-\\xd6\\xd8-\\xde",Kg="\\ufe0e\\ufe0f",s6=hy+oC+py+VM,c6="['’]",eO="["+x2+"]",u6="["+s6+"]",A2="["+Ym+"]",WM="\\d+",Xm="["+C2+"]",tO="["+iC+"]",l6="[^"+x2+s6+WM+C2+iC+a6+"]",aC="\\ud83c[\\udffb-\\udfff]",X_="(?:"+A2+"|"+aC+")",d6="[^"+x2+"]",nO="(?:\\ud83c[\\udde6-\\uddff]){2}",gy="[\\ud800-\\udbff][\\udc00-\\udfff]",J_="["+a6+"]",f6="\\u200d",h6="(?:"+tO+"|"+l6+")",p6="(?:"+J_+"|"+l6+")",KM="(?:"+c6+"(?:d|ll|m|re|s|t|ve))?",YM="(?:"+c6+"(?:D|LL|M|RE|S|T|VE))?",sC=X_+"?",rO="["+Kg+"]?",g6="(?:"+f6+"(?:"+[d6,nO,gy].join("|")+")"+rO+sC+")*",Z_="\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])",Q_="\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])",m6=rO+sC+g6,b6="(?:"+[Xm,nO,gy].join("|")+")"+m6,XM="(?:"+[d6+A2+"?",A2,nO,gy,eO].join("|")+")",cC=RegExp(c6,"g"),pv=RegExp(A2,"g"),uC=RegExp(aC+"(?="+aC+")|"+XM+m6,"g"),lC=RegExp([J_+"?"+tO+"+"+KM+"(?="+[u6,J_,"$"].join("|")+")",p6+"+"+YM+"(?="+[u6,J_+h6,"$"].join("|")+")",J_+"?"+h6+"+"+KM,J_+"+"+YM,Q_,Z_,WM,b6].join("|"),"g"),my=RegExp("["+f6+x2+Ym+Kg+"]"),v6=/[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/,iO=["Array","Buffer","DataView","Date","Error","Float32Array","Float64Array","Function","Int8Array","Int16Array","Int32Array","Map","Math","Object","Promise","RegExp","Set","String","Symbol","TypeError","Uint8Array","Uint8ClampedArray","Uint16Array","Uint32Array","WeakMap","_","clearTimeout","isFinite","parseInt","setTimeout"],JM=-1,Zc={};Zc[Ei]=Zc[Sa]=Zc[bs]=Zc[Cc]=Zc[Ac]=Zc[Xc]=Zc[rc]=Zc[js]=Zc[Jc]=!0,Zc[An]=Zc[Yt]=Zc[Pn]=Zc[Nt]=Zc[Kr]=Zc[on]=Zc[wi]=Zc[no]=Zc[ro]=Zc[Ko]=Zc[qi]=Zc[Eo]=Zc[Fc]=Zc[Yo]=Zc[Yc]=!1;var Bc={};Bc[An]=Bc[Yt]=Bc[Pn]=Bc[Kr]=Bc[Nt]=Bc[on]=Bc[Ei]=Bc[Sa]=Bc[bs]=Bc[Cc]=Bc[Ac]=Bc[ro]=Bc[Ko]=Bc[qi]=Bc[Eo]=Bc[Fc]=Bc[Yo]=Bc[zl]=Bc[Xc]=Bc[rc]=Bc[js]=Bc[Jc]=!0,Bc[wi]=Bc[no]=Bc[Yc]=!1;var ZM={À:"A",Á:"A",Â:"A",Ã:"A",Ä:"A",Å:"A",à:"a",á:"a",â:"a",ã:"a",ä:"a",å:"a",Ç:"C",ç:"c",Ð:"D",ð:"d",È:"E",É:"E",Ê:"E",Ë:"E",è:"e",é:"e",ê:"e",ë:"e",Ì:"I",Í:"I",Î:"I",Ï:"I",ì:"i",í:"i",î:"i",ï:"i",Ñ:"N",ñ:"n",Ò:"O",Ó:"O",Ô:"O",Õ:"O",Ö:"O",Ø:"O",ò:"o",ó:"o",ô:"o",õ:"o",ö:"o",ø:"o",Ù:"U",Ú:"U",Û:"U",Ü:"U",ù:"u",ú:"u",û:"u",ü:"u",Ý:"Y",ý:"y",ÿ:"y",Æ:"Ae",æ:"ae",Þ:"Th",þ:"th",ß:"ss",Ā:"A",Ă:"A",Ą:"A",ā:"a",ă:"a",ą:"a",Ć:"C",Ĉ:"C",Ċ:"C",Č:"C",ć:"c",ĉ:"c",ċ:"c",č:"c",Ď:"D",Đ:"D",ď:"d",đ:"d",Ē:"E",Ĕ:"E",Ė:"E",Ę:"E",Ě:"E",ē:"e",ĕ:"e",ė:"e",ę:"e",ě:"e",Ĝ:"G",Ğ:"G",Ġ:"G",Ģ:"G",ĝ:"g",ğ:"g",ġ:"g",ģ:"g",Ĥ:"H",Ħ:"H",ĥ:"h",ħ:"h",Ĩ:"I",Ī:"I",Ĭ:"I",Į:"I",İ:"I",ĩ:"i",ī:"i",ĭ:"i",į:"i",ı:"i",Ĵ:"J",ĵ:"j",Ķ:"K",ķ:"k",ĸ:"k",Ĺ:"L",Ļ:"L",Ľ:"L",Ŀ:"L",Ł:"L",ĺ:"l",ļ:"l",ľ:"l",ŀ:"l",ł:"l",Ń:"N",Ņ:"N",Ň:"N",Ŋ:"N",ń:"n",ņ:"n",ň:"n",ŋ:"n",Ō:"O",Ŏ:"O",Ő:"O",ō:"o",ŏ:"o",ő:"o",Ŕ:"R",Ŗ:"R",Ř:"R",ŕ:"r",ŗ:"r",ř:"r",Ś:"S",Ŝ:"S",Ş:"S",Š:"S",ś:"s",ŝ:"s",ş:"s",š:"s",Ţ:"T",Ť:"T",Ŧ:"T",ţ:"t",ť:"t",ŧ:"t",Ũ:"U",Ū:"U",Ŭ:"U",Ů:"U",Ű:"U",Ų:"U",ũ:"u",ū:"u",ŭ:"u",ů:"u",ű:"u",ų:"u",Ŵ:"W",ŵ:"w",Ŷ:"Y",ŷ:"y",Ÿ:"Y",Ź:"Z",Ż:"Z",Ž:"Z",ź:"z",ż:"z",ž:"z",Ĳ:"IJ",ĳ:"ij",Œ:"Oe",œ:"oe",ŉ:"'n",ſ:"s"},eg={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"},I2={"&amp;":"&","&lt;":"<","&gt;":">","&quot;":'"',"&#39;":"'"},oO={"\\":"\\","'":"'","\n":"n","\r":"r","\u2028":"u2028","\u2029":"u2029"},dC=parseFloat,SU=parseInt,aO=typeof pI=="object"&&pI&&pI.Object===Object&&pI,w6=typeof self=="object"&&self&&self.Object===Object&&self,dd=aO||w6||Function("return this")(),ek=r&&!r.nodeType&&r,by=ek&&!0&&n&&!n.nodeType&&n,QM=by&&by.exports===ek,y6=QM&&aO.process,Jm=function(){try{var Dt=by&&by.require&&by.require("util").types;return Dt||y6&&y6.binding&&y6.binding("util")}catch{}}(),sO=Jm&&Jm.isArrayBuffer,e8=Jm&&Jm.isDate,t8=Jm&&Jm.isMap,n8=Jm&&Jm.isRegExp,r8=Jm&&Jm.isSet,i8=Jm&&Jm.isTypedArray;function tg(Dt,Rn,En){switch(En.length){case 0:return Dt.call(Rn);case 1:return Dt.call(Rn,En[0]);case 2:return Dt.call(Rn,En[0],En[1]);case 3:return Dt.call(Rn,En[0],En[1],En[2])}return Dt.apply(Rn,En)}function _U(Dt,Rn,En,ui){for(var Lo=-1,Ws=Dt==null?0:Dt.length;++Lo<Ws;){var Nd=Dt[Lo];Rn(ui,Nd,En(Nd),Dt)}return ui}function Zm(Dt,Rn){for(var En=-1,ui=Dt==null?0:Dt.length;++En<ui&&Rn(Dt[En],En,Dt)!==!1;);return Dt}function kU(Dt,Rn){for(var En=Dt==null?0:Dt.length;En--&&Rn(Dt[En],En,Dt)!==!1;);return Dt}function o8(Dt,Rn){for(var En=-1,ui=Dt==null?0:Dt.length;++En<ui;)if(!Rn(Dt[En],En,Dt))return!1;return!0}function n0(Dt,Rn){for(var En=-1,ui=Dt==null?0:Dt.length,Lo=0,Ws=[];++En<ui;){var Nd=Dt[En];Rn(Nd,En,Dt)&&(Ws[Lo++]=Nd)}return Ws}function cO(Dt,Rn){var En=Dt==null?0:Dt.length;return!!En&&tk(Dt,Rn,0)>-1}function E6(Dt,Rn,En){for(var ui=-1,Lo=Dt==null?0:Dt.length;++ui<Lo;)if(En(Rn,Dt[ui]))return!0;return!1}function al(Dt,Rn){for(var En=-1,ui=Dt==null?0:Dt.length,Lo=Array(ui);++En<ui;)Lo[En]=Rn(Dt[En],En,Dt);return Lo}function gv(Dt,Rn){for(var En=-1,ui=Rn.length,Lo=Dt.length;++En<ui;)Dt[Lo+En]=Rn[En];return Dt}function S6(Dt,Rn,En,ui){var Lo=-1,Ws=Dt==null?0:Dt.length;for(ui&&Ws&&(En=Dt[++Lo]);++Lo<Ws;)En=Rn(En,Dt[Lo],Lo,Dt);return En}function TU(Dt,Rn,En,ui){var Lo=Dt==null?0:Dt.length;for(ui&&Lo&&(En=Dt[--Lo]);Lo--;)En=Rn(En,Dt[Lo],Lo,Dt);return En}function _6(Dt,Rn){for(var En=-1,ui=Dt==null?0:Dt.length;++En<ui;)if(Rn(Dt[En],En,Dt))return!0;return!1}var xU=lO("length");function CU(Dt){return Dt.split("")}function AU(Dt){return Dt.match(sy)||[]}function a8(Dt,Rn,En){var ui;return En(Dt,function(Lo,Ws,Nd){if(Rn(Lo,Ws,Nd))return ui=Ws,!1}),ui}function uO(Dt,Rn,En,ui){for(var Lo=Dt.length,Ws=En+(ui?1:-1);ui?Ws--:++Ws<Lo;)if(Rn(Dt[Ws],Ws,Dt))return Ws;return-1}function tk(Dt,Rn,En){return Rn===Rn?RU(Dt,Rn,En):uO(Dt,nk,En)}function k6(Dt,Rn,En,ui){for(var Lo=En-1,Ws=Dt.length;++Lo<Ws;)if(ui(Dt[Lo],Rn))return Lo;return-1}function nk(Dt){return Dt!==Dt}function s8(Dt,Rn){var En=Dt==null?0:Dt.length;return En?fO(Dt,Rn)/En:de}function lO(Dt){return function(Rn){return Rn==null?o:Rn[Dt]}}function dO(Dt){return function(Rn){return Dt==null?o:Dt[Rn]}}function T6(Dt,Rn,En,ui,Lo){return Lo(Dt,function(Ws,Nd,Ks){En=ui?(ui=!1,Ws):Rn(En,Ws,Nd,Ks)}),En}function c8(Dt,Rn){var En=Dt.length;for(Dt.sort(Rn);En--;)Dt[En]=Dt[En].value;return Dt}function fO(Dt,Rn){for(var En,ui=-1,Lo=Dt.length;++ui<Lo;){var Ws=Rn(Dt[ui]);Ws!==o&&(En=En===o?Ws:En+Ws)}return En}function vy(Dt,Rn){for(var En=-1,ui=Array(Dt);++En<Dt;)ui[En]=Rn(En);return ui}function IU(Dt,Rn){return al(Rn,function(En){return[En,Dt[En]]})}function u8(Dt){return Dt&&Dt.slice(0,C6(Dt)+1).replace(t0,"")}function Yg(Dt){return function(Rn){return Dt(Rn)}}function hO(Dt,Rn){return al(Rn,function(En){return Dt[En]})}function rk(Dt,Rn){return Dt.has(Rn)}function ik(Dt,Rn){for(var En=-1,ui=Dt.length;++En<ui&&tk(Rn,Dt[En],0)>-1;);return En}function ok(Dt,Rn){for(var En=Dt.length;En--&&tk(Rn,Dt[En],0)>-1;);return En}function OU(Dt,Rn){for(var En=Dt.length,ui=0;En--;)Dt[En]===Rn&&++ui;return ui}var pO=dO(ZM),l8=dO(eg);function d8(Dt){return"\\"+oO[Dt]}function x6(Dt,Rn){return Dt==null?o:Dt[Rn]}function mv(Dt){return my.test(Dt)}function f8(Dt){return v6.test(Dt)}function h8(Dt){for(var Rn,En=[];!(Rn=Dt.next()).done;)En.push(Rn.value);return En}function gO(Dt){var Rn=-1,En=Array(Dt.size);return Dt.forEach(function(ui,Lo){En[++Rn]=[Lo,ui]}),En}function p8(Dt,Rn){return function(En){return Dt(Rn(En))}}function wy(Dt,Rn){for(var En=-1,ui=Dt.length,Lo=0,Ws=[];++En<ui;){var Nd=Dt[En];(Nd===Rn||Nd===k)&&(Dt[En]=k,Ws[Lo++]=En)}return Ws}function mO(Dt){var Rn=-1,En=Array(Dt.size);return Dt.forEach(function(ui){En[++Rn]=ui}),En}function NU(Dt){var Rn=-1,En=Array(Dt.size);return Dt.forEach(function(ui){En[++Rn]=[ui,ui]}),En}function RU(Dt,Rn,En){for(var ui=En-1,Lo=Dt.length;++ui<Lo;)if(Dt[ui]===Rn)return ui;return-1}function Xg(Dt,Rn,En){for(var ui=En+1;ui--;)if(Dt[ui]===Rn)return ui;return ui}function r0(Dt){return mv(Dt)?m8(Dt):xU(Dt)}function Qm(Dt){return mv(Dt)?A6(Dt):CU(Dt)}function C6(Dt){for(var Rn=Dt.length;Rn--&&k2.test(Dt.charAt(Rn)););return Rn}var g8=dO(I2);function m8(Dt){for(var Rn=uC.lastIndex=0;uC.test(Dt);)++Rn;return Rn}function A6(Dt){return Dt.match(uC)||[]}function PU(Dt){return Dt.match(lC)||[]}var ak=function Dt(Rn){Rn=Rn==null?dd:yy.defaults(dd.Object(),Rn,yy.pick(dd,iO));var En=Rn.Array,ui=Rn.Date,Lo=Rn.Error,Ws=Rn.Function,Nd=Rn.Math,Ks=Rn.Object,bO=Rn.RegExp,b8=Rn.String,Jg=Rn.TypeError,vO=En.prototype,LU=Ws.prototype,sk=Ks.prototype,fC=Rn["__core-js_shared__"],hC=LU.toString,bc=sk.hasOwnProperty,Yb=0,I6=function(){var I=/[^.]+$/.exec(fC&&fC.keys&&fC.keys.IE_PROTO||"");return I?"Symbol(src)_1."+I:""}(),bp=sk.toString,eb=hC.call(Ks),ck=dd._,O6=bO("^"+hC.call(bc).replace(_l,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$"),O2=QM?Rn.Buffer:o,Xb=Rn.Symbol,uk=Rn.Uint8Array,N2=O2?O2.allocUnsafe:o,pC=p8(Ks.getPrototypeOf,Ks),gC=Ks.create,R2=sk.propertyIsEnumerable,wO=vO.splice,ng=Xb?Xb.isConcatSpreadable:o,P2=Xb?Xb.iterator:o,vp=Xb?Xb.toStringTag:o,mC=function(){try{var I=Iy(Ks,"defineProperty");return I({},"",{}),I}catch{}}(),yO=Rn.clearTimeout!==dd.clearTimeout&&Rn.clearTimeout,L2=ui&&ui.now!==dd.Date.now&&ui.now,N6=Rn.setTimeout!==dd.setTimeout&&Rn.setTimeout,lk=Nd.ceil,bv=Nd.floor,Ey=Ks.getOwnPropertySymbols,EO=O2?O2.isBuffer:o,bC=Rn.isFinite,tb=vO.join,nb=p8(Ks.keys,Ks),qu=Nd.max,sl=Nd.min,vv=ui.now,D2=Rn.parseInt,R6=Nd.random,P6=vO.reverse,vC=Iy(Rn,"DataView"),Rd=Iy(Rn,"Map"),cl=Iy(Rn,"Promise"),vs=Iy(Rn,"Set"),Jb=Iy(Rn,"WeakMap"),rg=Iy(Ks,"create"),dk=Jb&&new Jb,rb={},v8=Ov(vC),w8=Ov(Rd),wC=Ov(cl),yC=Ov(vs),y8=Ov(Jb),Sy=Xb?Xb.prototype:o,M2=Sy?Sy.valueOf:o,_y=Sy?Sy.toString:o;function De(I){if(kl(I)&&!Xo(I)&&!(I instanceof ts)){if(I instanceof ig)return I;if(bc.call(I,"__wrapped__"))return Yd(I)}return new ig(I)}var Lh=function(){function I(){}return function(M){if(!ul(M))return{};if(gC)return gC(M);I.prototype=M;var Y=new I;return I.prototype=o,Y}}();function wp(){}function ig(I,M){this.__wrapped__=I,this.__actions__=[],this.__chain__=!!M,this.__index__=0,this.__values__=o}De.templateSettings={escape:_2,evaluate:dv,interpolate:K_,variable:"",imports:{_:De}},De.prototype=wp.prototype,De.prototype.constructor=De,ig.prototype=Lh(wp.prototype),ig.prototype.constructor=ig;function ts(I){this.__wrapped__=I,this.__actions__=[],this.__dir__=1,this.__filtered__=!1,this.__iteratees__=[],this.__takeCount__=We,this.__views__=[]}function fk(){var I=new ts(this.__wrapped__);return I.__actions__=Kd(this.__actions__),I.__dir__=this.__dir__,I.__filtered__=this.__filtered__,I.__iteratees__=Kd(this.__iteratees__),I.__takeCount__=this.__takeCount__,I.__views__=Kd(this.__views__),I}function SO(){if(this.__filtered__){var I=new ts(this);I.__dir__=-1,I.__filtered__=!0}else I=this.clone(),I.__dir__*=-1;return I}function j2(){var I=this.__wrapped__.value(),M=this.__dir__,Y=Xo(I),fe=M<0,Ce=Y?I.length:0,ze=J8(0,Ce,this.__views__),dt=ze.start,It=ze.end,qt=It-dt,Hn=fe?It:dt-1,qn=this.__iteratees__,sr=qn.length,ri=0,Ji=sl(qt,this.__takeCount__);if(!Y||!fe&&Ce==qt&&Ji==qt)return B8(I,this.__actions__);var Uo=[];e:for(;qt--&&ri<Ji;){Hn+=M;for(var ns=-1,zo=I[Hn];++ns<sr;){var rs=qn[ns],Xs=rs.iteratee,s1=rs.type,pg=Xs(zo);if(s1==Re)zo=pg;else if(!pg){if(s1==Z)continue e;break e}}Uo[ri++]=zo}return Uo}ts.prototype=Lh(wp.prototype),ts.prototype.constructor=ts;function Hl(I){var M=-1,Y=I==null?0:I.length;for(this.clear();++M<Y;){var fe=I[M];this.set(fe[0],fe[1])}}function _O(){this.__data__=rg?rg(null):{},this.size=0}function E8(I){var M=this.has(I)&&delete this.__data__[I];return this.size-=M?1:0,M}function S8(I){var M=this.__data__;if(rg){var Y=M[I];return Y===y?o:Y}return bc.call(M,I)?M[I]:o}function _8(I){var M=this.__data__;return rg?M[I]!==o:bc.call(M,I)}function DU(I,M){var Y=this.__data__;return this.size+=this.has(I)?0:1,Y[I]=rg&&M===o?y:M,this}Hl.prototype.clear=_O,Hl.prototype.delete=E8,Hl.prototype.get=S8,Hl.prototype.has=_8,Hl.prototype.set=DU;function Qc(I){var M=-1,Y=I==null?0:I.length;for(this.clear();++M<Y;){var fe=I[M];this.set(fe[0],fe[1])}}function kO(){this.__data__=[],this.size=0}function EC(I){var M=this.__data__,Y=gk(M,I);if(Y<0)return!1;var fe=M.length-1;return Y==fe?M.pop():wO.call(M,Y,1),--this.size,!0}function SC(I){var M=this.__data__,Y=gk(M,I);return Y<0?o:M[Y][1]}function L6(I){return gk(this.__data__,I)>-1}function hk(I,M){var Y=this.__data__,fe=gk(Y,I);return fe<0?(++this.size,Y.push([I,M])):Y[fe][1]=M,this}Qc.prototype.clear=kO,Qc.prototype.delete=EC,Qc.prototype.get=SC,Qc.prototype.has=L6,Qc.prototype.set=hk;function og(I){var M=-1,Y=I==null?0:I.length;for(this.clear();++M<Y;){var fe=I[M];this.set(fe[0],fe[1])}}function MU(){this.size=0,this.__data__={hash:new Hl,map:new(Rd||Qc),string:new Hl}}function k8(I){var M=Ek(this,I).delete(I);return this.size-=M?1:0,M}function jU(I){return Ek(this,I).get(I)}function T8(I){return Ek(this,I).has(I)}function $2(I,M){var Y=Ek(this,I),fe=Y.size;return Y.set(I,M),this.size+=Y.size==fe?0:1,this}og.prototype.clear=MU,og.prototype.delete=k8,og.prototype.get=jU,og.prototype.has=T8,og.prototype.set=$2;function wv(I){var M=-1,Y=I==null?0:I.length;for(this.__data__=new og;++M<Y;)this.add(I[M])}function _C(I){return this.__data__.set(I,y),this}function x8(I){return this.__data__.has(I)}wv.prototype.add=wv.prototype.push=_C,wv.prototype.has=x8;function yp(I){var M=this.__data__=new Qc(I);this.size=M.size}function C8(){this.__data__=new Qc,this.size=0}function TO(I){var M=this.__data__,Y=M.delete(I);return this.size=M.size,Y}function D6(I){return this.__data__.get(I)}function A8(I){return this.__data__.has(I)}function xO(I,M){var Y=this.__data__;if(Y instanceof Qc){var fe=Y.__data__;if(!Rd||fe.length<c-1)return fe.push([I,M]),this.size=++Y.size,this;Y=this.__data__=new og(fe)}return Y.set(I,M),this.size=Y.size,this}yp.prototype.clear=C8,yp.prototype.delete=TO,yp.prototype.get=D6,yp.prototype.has=A8,yp.prototype.set=xO;function M6(I,M){var Y=Xo(I),fe=!Y&&Pv(I),Ce=!Y&&!fe&&Fy(I),ze=!Y&&!fe&&!Ce&&Pk(I),dt=Y||fe||Ce||ze,It=dt?vy(I.length,b8):[],qt=It.length;for(var Hn in I)(M||bc.call(I,Hn))&&!(dt&&(Hn=="length"||Ce&&(Hn=="offset"||Hn=="parent")||ze&&(Hn=="buffer"||Hn=="byteLength"||Hn=="byteOffset")||t1(Hn,qt)))&&It.push(Hn);return It}function pk(I){var M=I.length;return M?I[MO(0,M-1)]:o}function I8(I,M){return JC(Kd(I),Qb(M,0,I.length))}function O8(I){return JC(Kd(I))}function i0(I,M,Y){(Y!==o&&!hg(I[M],Y)||Y===o&&!(M in I))&&Zb(I,M,Y)}function F2(I,M,Y){var fe=I[M];(!(bc.call(I,M)&&hg(fe,Y))||Y===o&&!(M in I))&&Zb(I,M,Y)}function gk(I,M){for(var Y=I.length;Y--;)if(hg(I[Y][0],M))return Y;return-1}function ky(I,M,Y,fe){return Ev(I,function(Ce,ze,dt){M(fe,Ce,Y(Ce),dt)}),fe}function yv(I,M){return I&&cg(M,Zd(M),I)}function kC(I,M){return I&&cg(M,Af(M),I)}function Zb(I,M,Y){M=="__proto__"&&mC?mC(I,M,{configurable:!0,enumerable:!0,value:Y,writable:!0}):I[M]=Y}function TC(I,M){for(var Y=-1,fe=M.length,Ce=En(fe),ze=I==null;++Y<fe;)Ce[Y]=ze?o:iS(I,M[Y]);return Ce}function Qb(I,M,Y){return I===I&&(Y!==o&&(I=I<=Y?I:Y),M!==o&&(I=I>=M?I:M)),I}function Dh(I,M,Y,fe,Ce,ze){var dt,It=M&x,qt=M&O,Hn=M&P;if(Y&&(dt=Ce?Y(I,fe,Ce,ze):Y(I)),dt!==o)return dt;if(!ul(I))return I;var qn=Xo(I);if(qn){if(dt=mt(I),!It)return Kd(I,dt)}else{var sr=Cf(I),ri=sr==no||sr==Ri;if(Fy(I))return e5(I,It);if(sr==qi||sr==An||ri&&!Ce){if(dt=qt||ri?{}:u5(I),!It)return qt?q2(I,kC(dt,I)):yk(I,yv(dt,I))}else{if(!Bc[sr])return Ce?I:{};dt=l5(I,sr,It)}}ze||(ze=new yp);var Ji=ze.get(I);if(Ji)return Ji;ze.set(I,dt),B9(I)?I.forEach(function(zo){dt.add(Dh(zo,M,Y,zo,I,ze))}):dN(I)&&I.forEach(function(zo,rs){dt.set(rs,Dh(zo,M,Y,rs,I,ze))});var Uo=Hn?qt?GO:HO:qt?Af:Zd,ns=qn?o:Uo(I);return Zm(ns||I,function(zo,rs){ns&&(rs=zo,zo=I[rs]),F2(dt,rs,Dh(zo,M,Y,rs,I,ze))}),dt}function N8(I){var M=Zd(I);return function(Y){return j6(Y,I,M)}}function j6(I,M,Y){var fe=Y.length;if(I==null)return!fe;for(I=Ks(I);fe--;){var Ce=Y[fe],ze=M[Ce],dt=I[Ce];if(dt===o&&!(Ce in I)||!ze(dt))return!1}return!0}function B2(I,M,Y){if(typeof I!="function")throw new Jg(h);return Sk(function(){I.apply(o,Y)},M)}function Ty(I,M,Y,fe){var Ce=-1,ze=cO,dt=!0,It=I.length,qt=[],Hn=M.length;if(!It)return qt;Y&&(M=al(M,Yg(Y))),fe?(ze=E6,dt=!1):M.length>=c&&(ze=rk,dt=!1,M=new wv(M));e:for(;++Ce<It;){var qn=I[Ce],sr=Y==null?qn:Y(qn);if(qn=fe||qn!==0?qn:0,dt&&sr===sr){for(var ri=Hn;ri--;)if(M[ri]===sr)continue e;qt.push(qn)}else ze(M,sr,fe)||qt.push(qn)}return qt}var Ev=i5(Mh),CO=i5(bk,!0);function $6(I,M){var Y=!0;return Ev(I,function(fe,Ce,ze){return Y=!!M(fe,Ce,ze),Y}),Y}function e1(I,M,Y){for(var fe=-1,Ce=I.length;++fe<Ce;){var ze=I[fe],dt=M(ze);if(dt!=null&&(It===o?dt===dt&&!_p(dt):Y(dt,It)))var It=dt,qt=ze}return qt}function Wa(I,M,Y,fe){var Ce=I.length;for(Y=fa(Y),Y<0&&(Y=-Y>Ce?0:Ce+Y),fe=fe===o||fe>Ce?Ce:fa(fe),fe<0&&(fe+=Ce),fe=Y>fe?0:d0(fe);Y<fe;)I[Y++]=M;return I}function Uc(I,M){var Y=[];return Ev(I,function(fe,Ce,ze){M(fe,Ce,ze)&&Y.push(fe)}),Y}function ws(I,M,Y,fe,Ce){var ze=-1,dt=I.length;for(Y||(Y=Cv),Ce||(Ce=[]);++ze<dt;){var It=I[ze];M>0&&Y(It)?M>1?ws(It,M-1,Y,fe,Ce):gv(Ce,It):fe||(Ce[Ce.length]=It)}return Ce}var mk=W2(),U2=W2(!0);function Mh(I,M){return I&&mk(I,M,Zd)}function bk(I,M){return I&&U2(I,M,Zd)}function xy(I,M){return n0(M,function(Y){return i1(I[Y])})}function ib(I,M){M=_v(M,I);for(var Y=0,fe=M.length;I!=null&&Y<fe;)I=I[Ys(M[Y++])];return Y&&Y==fe?I:o}function R8(I,M,Y){var fe=M(I);return Xo(I)?fe:gv(fe,Y(I))}function Cs(I){return I==null?I===o?ld:Ca:vp&&vp in Ks(I)?X8(I):t9(I)}function xC(I,M){return I>M}function F6(I,M){return I!=null&&bc.call(I,M)}function P8(I,M){return I!=null&&M in Ks(I)}function L8(I,M,Y){return I>=sl(M,Y)&&I<qu(M,Y)}function B6(I,M,Y){for(var fe=Y?E6:cO,Ce=I[0].length,ze=I.length,dt=ze,It=En(ze),qt=1/0,Hn=[];dt--;){var qn=I[dt];dt&&M&&(qn=al(qn,Yg(M))),qt=sl(qn.length,qt),It[dt]=!Y&&(M||Ce>=120&&qn.length>=120)?new wv(dt&&qn):o}qn=I[0];var sr=-1,ri=It[0];e:for(;++sr<Ce&&Hn.length<qt;){var Ji=qn[sr],Uo=M?M(Ji):Ji;if(Ji=Y||Ji!==0?Ji:0,!(ri?rk(ri,Uo):fe(Hn,Uo,Y))){for(dt=ze;--dt;){var ns=It[dt];if(!(ns?rk(ns,Uo):fe(I[dt],Uo,Y)))continue e}ri&&ri.push(Uo),Hn.push(Ji)}}return Hn}function U6(I,M,Y,fe){return Mh(I,function(Ce,ze,dt){M(fe,Y(Ce),ze,dt)}),fe}function z2(I,M,Y){M=_v(M,I),I=Iv(I,M);var fe=I==null?I:I[Ys(jh(M))];return fe==null?o:tg(fe,I,Y)}function AO(I){return kl(I)&&Cs(I)==An}function ag(I){return kl(I)&&Cs(I)==Pn}function Pd(I){return kl(I)&&Cs(I)==on}function H2(I,M,Y,fe,Ce){return I===M?!0:I==null||M==null||!kl(I)&&!kl(M)?I!==I&&M!==M:CC(I,M,Y,fe,H2,Ce)}function CC(I,M,Y,fe,Ce,ze){var dt=Xo(I),It=Xo(M),qt=dt?Yt:Cf(I),Hn=It?Yt:Cf(M);qt=qt==An?qi:qt,Hn=Hn==An?qi:Hn;var qn=qt==qi,sr=Hn==qi,ri=qt==Hn;if(ri&&Fy(I)){if(!Fy(M))return!1;dt=!0,qn=!1}if(ri&&!qn)return ze||(ze=new yp),dt||Pk(I)?s5(I,M,Y,fe,Ce,ze):Y8(I,M,qt,Y,fe,Ce,ze);if(!(Y&D)){var Ji=qn&&bc.call(I,"__wrapped__"),Uo=sr&&bc.call(M,"__wrapped__");if(Ji||Uo){var ns=Ji?I.value():I,zo=Uo?M.value():M;return ze||(ze=new yp),Ce(ns,zo,Y,fe,ze)}}return ri?(ze||(ze=new yp),xf(I,M,Y,fe,Ce,ze)):!1}function IO(I){return kl(I)&&Cf(I)==ro}function AC(I,M,Y,fe){var Ce=Y.length,ze=Ce,dt=!fe;if(I==null)return!ze;for(I=Ks(I);Ce--;){var It=Y[Ce];if(dt&&It[2]?It[1]!==I[It[0]]:!(It[0]in I))return!1}for(;++Ce<ze;){It=Y[Ce];var qt=It[0],Hn=I[qt],qn=It[1];if(dt&&It[2]){if(Hn===o&&!(qt in I))return!1}else{var sr=new yp;if(fe)var ri=fe(Hn,qn,qt,I,M,sr);if(!(ri===o?H2(qn,Hn,D|$,fe,sr):ri))return!1}}return!0}function o0(I){if(!ul(I)||XO(I))return!1;var M=i1(I)?O6:JI;return M.test(Ov(I))}function OO(I){return kl(I)&&Cs(I)==Eo}function z6(I){return kl(I)&&Cf(I)==Fc}function H6(I){return kl(I)&&tS(I.length)&&!!Zc[Cs(I)]}function G6(I){return typeof I=="function"?I:I==null?Zt:typeof I=="object"?Xo(I)?PO(I[0],I[1]):V6(I):Fs(I)}function NO(I){if(!pu(I))return nb(I);var M=[];for(var Y in Ks(I))bc.call(I,Y)&&Y!="constructor"&&M.push(Y);return M}function D8(I){if(!ul(I))return Oy(I);var M=pu(I),Y=[];for(var fe in I)fe=="constructor"&&(M||!bc.call(I,fe))||Y.push(fe);return Y}function RO(I,M){return I<M}function q6(I,M){var Y=-1,fe=Xd(I)?En(I.length):[];return Ev(I,function(Ce,ze,dt){fe[++Y]=M(Ce,ze,dt)}),fe}function V6(I){var M=VO(I);return M.length==1&&M[0][2]?e9(M[0][0],M[0][1]):function(Y){return Y===I||AC(Y,I,M)}}function PO(I,M){return KO(I)&&Q8(M)?e9(Ys(I),M):function(Y){var fe=iS(Y,I);return fe===o&&fe===M?By(Y,I):H2(M,fe,D|$)}}function IC(I,M,Y,fe,Ce){I!==M&&mk(M,function(ze,dt){if(Ce||(Ce=new yp),ul(ze))$U(I,M,dt,Y,IC,fe,Ce);else{var It=fe?fe(ZO(I,dt),ze,dt+"",I,M,Ce):o;It===o&&(It=ze),i0(I,dt,It)}},Af)}function $U(I,M,Y,fe,Ce,ze,dt){var It=ZO(I,Y),qt=ZO(M,Y),Hn=dt.get(qt);if(Hn){i0(I,Y,Hn);return}var qn=ze?ze(It,qt,Y+"",I,M,dt):o,sr=qn===o;if(sr){var ri=Xo(qt),Ji=!ri&&Fy(qt),Uo=!ri&&!Ji&&Pk(qt);qn=qt,ri||Ji||Uo?Xo(It)?qn=It:Vl(It)?qn=Kd(It):Ji?(sr=!1,qn=e5(qt,!0)):Uo?(sr=!1,qn=n5(qt,!0)):qn=[]:cb(qt)||Pv(qt)?(qn=It,Pv(It)?qn=$5(It):(!ul(It)||i1(It))&&(qn=u5(qt))):sr=!1}sr&&(dt.set(qt,qn),Ce(qn,qt,fe,ze,dt),dt.delete(qt)),i0(I,Y,qn)}function W6(I,M){var Y=I.length;if(Y)return M+=M<0?Y:0,t1(M,Y)?I[M]:o}function K6(I,M,Y){M.length?M=al(M,function(ze){return Xo(ze)?function(dt){return ib(dt,ze.length===1?ze[0]:ze)}:ze}):M=[Zt];var fe=-1;M=al(M,Yg(fo()));var Ce=q6(I,function(ze,dt,It){var qt=al(M,function(Hn){return Hn(ze)});return{criteria:qt,index:++fe,value:ze}});return c8(Ce,function(ze,dt){return G8(ze,dt,Y)})}function M8(I,M){return LO(I,M,function(Y,fe){return By(I,fe)})}function LO(I,M,Y){for(var fe=-1,Ce=M.length,ze={};++fe<Ce;){var dt=M[fe],It=ib(I,dt);Y(It,dt)&&wk(ze,_v(dt,I),It)}return ze}function j8(I){return function(M){return ib(M,I)}}function DO(I,M,Y,fe){var Ce=fe?k6:tk,ze=-1,dt=M.length,It=I;for(I===M&&(M=Kd(M)),Y&&(It=al(I,Yg(Y)));++ze<dt;)for(var qt=0,Hn=M[ze],qn=Y?Y(Hn):Hn;(qt=Ce(It,qn,qt,fe))>-1;)It!==I&&wO.call(It,qt,1),wO.call(I,qt,1);return I}function Y6(I,M){for(var Y=I?M.length:0,fe=Y-1;Y--;){var Ce=M[Y];if(Y==fe||Ce!==ze){var ze=Ce;t1(Ce)?wO.call(I,Ce,1):PC(I,Ce)}}return I}function MO(I,M){return I+bv(R6()*(M-I+1))}function $8(I,M,Y,fe){for(var Ce=-1,ze=qu(lk((M-I)/(Y||1)),0),dt=En(ze);ze--;)dt[fe?ze:++Ce]=I,I+=Y;return dt}function X6(I,M){var Y="";if(!I||M<1||M>it)return Y;do M%2&&(Y+=I),M=bv(M/2),M&&(I+=I);while(M);return Y}function Da(I,M){return YC(WC(I,M,Zt),I+"")}function vk(I){return pk(oh(I))}function jO(I,M){var Y=oh(I);return JC(Y,Qb(M,0,Y.length))}function wk(I,M,Y,fe){if(!ul(I))return I;M=_v(M,I);for(var Ce=-1,ze=M.length,dt=ze-1,It=I;It!=null&&++Ce<ze;){var qt=Ys(M[Ce]),Hn=Y;if(qt==="__proto__"||qt==="constructor"||qt==="prototype")return I;if(Ce!=dt){var qn=It[qt];Hn=fe?fe(qn,qt,It):o,Hn===o&&(Hn=ul(qn)?qn:t1(M[Ce+1])?[]:{})}F2(It,qt,Hn),It=It[qt]}return I}var OC=dk?function(I,M){return dk.set(I,M),I}:Zt,NC=mC?function(I,M){return mC(I,"toString",{configurable:!0,enumerable:!1,value:Xt(M),writable:!0})}:Zt;function F8(I){return JC(oh(I))}function Zg(I,M,Y){var fe=-1,Ce=I.length;M<0&&(M=-M>Ce?0:Ce+M),Y=Y>Ce?Ce:Y,Y<0&&(Y+=Ce),Ce=M>Y?0:Y-M>>>0,M>>>=0;for(var ze=En(Ce);++fe<Ce;)ze[fe]=I[fe+M];return ze}function FU(I,M){var Y;return Ev(I,function(fe,Ce,ze){return Y=M(fe,Ce,ze),!Y}),!!Y}function RC(I,M,Y){var fe=0,Ce=I==null?fe:I.length;if(typeof M=="number"&&M===M&&Ce<=Ie){for(;fe<Ce;){var ze=fe+Ce>>>1,dt=I[ze];dt!==null&&!_p(dt)&&(Y?dt<=M:dt<M)?fe=ze+1:Ce=ze}return Ce}return $O(I,M,Zt,Y)}function $O(I,M,Y,fe){var Ce=0,ze=I==null?0:I.length;if(ze===0)return 0;M=Y(M);for(var dt=M!==M,It=M===null,qt=_p(M),Hn=M===o;Ce<ze;){var qn=bv((Ce+ze)/2),sr=Y(I[qn]),ri=sr!==o,Ji=sr===null,Uo=sr===sr,ns=_p(sr);if(dt)var zo=fe||Uo;else Hn?zo=Uo&&(fe||ri):It?zo=Uo&&ri&&(fe||!Ji):qt?zo=Uo&&ri&&!Ji&&(fe||!ns):Ji||ns?zo=!1:zo=fe?sr<=M:sr<M;zo?Ce=qn+1:ze=qn}return sl(ze,ht)}function J6(I,M){for(var Y=-1,fe=I.length,Ce=0,ze=[];++Y<fe;){var dt=I[Y],It=M?M(dt):dt;if(!Y||!hg(It,qt)){var qt=It;ze[Ce++]=dt===0?0:dt}}return ze}function FO(I){return typeof I=="number"?I:_p(I)?de:+I}function sg(I){if(typeof I=="string")return I;if(Xo(I))return al(I,sg)+"";if(_p(I))return _y?_y.call(I):"";var M=I+"";return M=="0"&&1/I==-qe?"-0":M}function Sv(I,M,Y){var fe=-1,Ce=cO,ze=I.length,dt=!0,It=[],qt=It;if(Y)dt=!1,Ce=E6;else if(ze>=c){var Hn=M?null:K8(I);if(Hn)return mO(Hn);dt=!1,Ce=rk,qt=new wv}else qt=M?[]:It;e:for(;++fe<ze;){var qn=I[fe],sr=M?M(qn):qn;if(qn=Y||qn!==0?qn:0,dt&&sr===sr){for(var ri=qt.length;ri--;)if(qt[ri]===sr)continue e;M&&qt.push(sr),It.push(qn)}else Ce(qt,sr,Y)||(qt!==It&&qt.push(sr),It.push(qn))}return It}function PC(I,M){return M=_v(M,I),I=Iv(I,M),I==null||delete I[Ys(jh(M))]}function Z6(I,M,Y,fe){return wk(I,M,Y(ib(I,M)),fe)}function LC(I,M,Y,fe){for(var Ce=I.length,ze=fe?Ce:-1;(fe?ze--:++ze<Ce)&&M(I[ze],ze,I););return Y?Zg(I,fe?0:ze,fe?ze+1:Ce):Zg(I,fe?ze+1:0,fe?Ce:ze)}function B8(I,M){var Y=I;return Y instanceof ts&&(Y=Y.value()),S6(M,function(fe,Ce){return Ce.func.apply(Ce.thisArg,gv([fe],Ce.args))},Y)}function DC(I,M,Y){var fe=I.length;if(fe<2)return fe?Sv(I[0]):[];for(var Ce=-1,ze=En(fe);++Ce<fe;)for(var dt=I[Ce],It=-1;++It<fe;)It!=Ce&&(ze[Ce]=Ty(ze[Ce]||dt,I[It],M,Y));return Sv(ws(ze,1),M,Y)}function Q6(I,M,Y){for(var fe=-1,Ce=I.length,ze=M.length,dt={};++fe<Ce;){var It=fe<ze?M[fe]:o;Y(dt,I[fe],It)}return dt}function BO(I){return Vl(I)?I:[]}function MC(I){return typeof I=="function"?I:Zt}function _v(I,M){return Xo(I)?I:KO(I,M)?[I]:c0(vc(I))}var U8=Da;function Cy(I,M,Y){var fe=I.length;return Y=Y===o?fe:Y,!M&&Y>=fe?I:Zg(I,M,Y)}var z8=yO||function(I){return dd.clearTimeout(I)};function e5(I,M){if(M)return I.slice();var Y=I.length,fe=N2?N2(Y):new I.constructor(Y);return I.copy(fe),fe}function G2(I){var M=new I.constructor(I.byteLength);return new uk(M).set(new uk(I)),M}function t5(I,M){var Y=M?G2(I.buffer):I.buffer;return new I.constructor(Y,I.byteOffset,I.byteLength)}function UO(I){var M=new I.constructor(I.source,uy.exec(I));return M.lastIndex=I.lastIndex,M}function H8(I){return M2?Ks(M2.call(I)):{}}function n5(I,M){var Y=M?G2(I.buffer):I.buffer;return new I.constructor(Y,I.byteOffset,I.length)}function jC(I,M){if(I!==M){var Y=I!==o,fe=I===null,Ce=I===I,ze=_p(I),dt=M!==o,It=M===null,qt=M===M,Hn=_p(M);if(!It&&!Hn&&!ze&&I>M||ze&&dt&&qt&&!It&&!Hn||fe&&dt&&qt||!Y&&qt||!Ce)return 1;if(!fe&&!ze&&!Hn&&I<M||Hn&&Y&&Ce&&!fe&&!ze||It&&Y&&Ce||!dt&&Ce||!qt)return-1}return 0}function G8(I,M,Y){for(var fe=-1,Ce=I.criteria,ze=M.criteria,dt=Ce.length,It=Y.length;++fe<dt;){var qt=jC(Ce[fe],ze[fe]);if(qt){if(fe>=It)return qt;var Hn=Y[fe];return qt*(Hn=="desc"?-1:1)}}return I.index-M.index}function r5(I,M,Y,fe){for(var Ce=-1,ze=I.length,dt=Y.length,It=-1,qt=M.length,Hn=qu(ze-dt,0),qn=En(qt+Hn),sr=!fe;++It<qt;)qn[It]=M[It];for(;++Ce<dt;)(sr||Ce<ze)&&(qn[Y[Ce]]=I[Ce]);for(;Hn--;)qn[It++]=I[Ce++];return qn}function q8(I,M,Y,fe){for(var Ce=-1,ze=I.length,dt=-1,It=Y.length,qt=-1,Hn=M.length,qn=qu(ze-It,0),sr=En(qn+Hn),ri=!fe;++Ce<qn;)sr[Ce]=I[Ce];for(var Ji=Ce;++qt<Hn;)sr[Ji+qt]=M[qt];for(;++dt<It;)(ri||Ce<ze)&&(sr[Ji+Y[dt]]=I[Ce++]);return sr}function Kd(I,M){var Y=-1,fe=I.length;for(M||(M=En(fe));++Y<fe;)M[Y]=I[Y];return M}function cg(I,M,Y,fe){var Ce=!Y;Y||(Y={});for(var ze=-1,dt=M.length;++ze<dt;){var It=M[ze],qt=fe?fe(Y[It],I[It],It,Y,I):o;qt===o&&(qt=I[It]),Ce?Zb(Y,It,qt):F2(Y,It,qt)}return Y}function yk(I,M){return cg(I,HC(I),M)}function q2(I,M){return cg(I,WO(I),M)}function V2(I,M){return function(Y,fe){var Ce=Xo(Y)?_U:ky,ze=M?M():{};return Ce(Y,I,fo(fe,2),ze)}}function kv(I){return Da(function(M,Y){var fe=-1,Ce=Y.length,ze=Ce>1?Y[Ce-1]:o,dt=Ce>2?Y[2]:o;for(ze=I.length>3&&typeof ze=="function"?(Ce--,ze):o,dt&&Sp(Y[0],Y[1],dt)&&(ze=Ce<3?o:ze,Ce=1),M=Ks(M);++fe<Ce;){var It=Y[fe];It&&I(M,It,fe,ze)}return M})}function i5(I,M){return function(Y,fe){if(Y==null)return Y;if(!Xd(Y))return I(Y,fe);for(var Ce=Y.length,ze=M?Ce:-1,dt=Ks(Y);(M?ze--:++ze<Ce)&&fe(dt[ze],ze,dt)!==!1;);return Y}}function W2(I){return function(M,Y,fe){for(var Ce=-1,ze=Ks(M),dt=fe(M),It=dt.length;It--;){var qt=dt[I?It:++Ce];if(Y(ze[qt],qt,ze)===!1)break}return M}}function V8(I,M,Y){var fe=M&U,Ce=ob(I);function ze(){var dt=this&&this!==dd&&this instanceof ze?Ce:I;return dt.apply(fe?Y:this,arguments)}return ze}function Qg(I){return function(M){M=vc(M);var Y=mv(M)?Qm(M):o,fe=Y?Y[0]:M.charAt(0),Ce=Y?Cy(Y,1).join(""):M.slice(1);return fe[I]()+Ce}}function Ep(I){return function(M){return S6(q(Uy(M).replace(cC,"")),I,"")}}function ob(I){return function(){var M=arguments;switch(M.length){case 0:return new I;case 1:return new I(M[0]);case 2:return new I(M[0],M[1]);case 3:return new I(M[0],M[1],M[2]);case 4:return new I(M[0],M[1],M[2],M[3]);case 5:return new I(M[0],M[1],M[2],M[3],M[4]);case 6:return new I(M[0],M[1],M[2],M[3],M[4],M[5]);case 7:return new I(M[0],M[1],M[2],M[3],M[4],M[5],M[6])}var Y=Lh(I.prototype),fe=I.apply(Y,M);return ul(fe)?fe:Y}}function W8(I,M,Y){var fe=ob(I);function Ce(){for(var ze=arguments.length,dt=En(ze),It=ze,qt=X2(Ce);It--;)dt[It]=arguments[It];var Hn=ze<3&&dt[0]!==qt&&dt[ze-1]!==qt?[]:wy(dt,qt);if(ze-=Hn.length,ze<Y)return o5(I,M,Gl,Ce.placeholder,o,dt,Hn,o,o,Y-ze);var qn=this&&this!==dd&&this instanceof Ce?fe:I;return tg(qn,this,dt)}return Ce}function Tv(I){return function(M,Y,fe){var Ce=Ks(M);if(!Xd(M)){var ze=fo(Y,3);M=Zd(M),Y=function(It){return ze(Ce[It],It,Ce)}}var dt=I(M,Y,fe);return dt>-1?Ce[ze?M[dt]:dt]:o}}function a0(I){return Vu(function(M){var Y=M.length,fe=Y,Ce=ig.prototype.thru;for(I&&M.reverse();fe--;){var ze=M[fe];if(typeof ze!="function")throw new Jg(h);if(Ce&&!dt&&zC(ze)=="wrapper")var dt=new ig([],!0)}for(fe=dt?fe:Y;++fe<Y;){ze=M[fe];var It=zC(ze),qt=It=="wrapper"?qO(ze):o;qt&&YO(qt[0])&&qt[1]==(he|K|ae|ye)&&!qt[4].length&&qt[9]==1?dt=dt[zC(qt[0])].apply(dt,qt[3]):dt=ze.length==1&&YO(ze)?dt[It]():dt.thru(ze)}return function(){var Hn=arguments,qn=Hn[0];if(dt&&Hn.length==1&&Xo(qn))return dt.plant(qn).value();for(var sr=0,ri=Y?M[sr].apply(this,Hn):qn;++sr<Y;)ri=M[sr].call(this,ri);return ri}})}function Gl(I,M,Y,fe,Ce,ze,dt,It,qt,Hn){var qn=M&he,sr=M&U,ri=M&B,Ji=M&(K|X),Uo=M&we,ns=ri?o:ob(I);function zo(){for(var rs=arguments.length,Xs=En(rs),s1=rs;s1--;)Xs[s1]=arguments[s1];if(Ji)var pg=X2(zo),c1=OU(Xs,pg);if(fe&&(Xs=r5(Xs,fe,Ce,Ji)),ze&&(Xs=q8(Xs,ze,dt,Ji)),rs-=c1,Ji&&rs<Hn){var Qd=wy(Xs,pg);return o5(I,M,Gl,zo.placeholder,Y,Xs,Qd,It,qt,Hn-rs)}var jv=sr?Y:this,Hy=ri?jv[I]:I;return rs=Xs.length,It?Xs=em(Xs,It):Uo&&rs>1&&Xs.reverse(),qn&&qt<rs&&(Xs.length=qt),this&&this!==dd&&this instanceof zo&&(Hy=ns||ob(Hy)),Hy.apply(jv,Xs)}return zo}function $C(I,M){return function(Y,fe){return U6(Y,I,M(fe),{})}}function K2(I,M){return function(Y,fe){var Ce;if(Y===o&&fe===o)return M;if(Y!==o&&(Ce=Y),fe!==o){if(Ce===o)return fe;typeof Y=="string"||typeof fe=="string"?(Y=sg(Y),fe=sg(fe)):(Y=FO(Y),fe=FO(fe)),Ce=I(Y,fe)}return Ce}}function FC(I){return Vu(function(M){return M=al(M,Yg(fo())),Da(function(Y){var fe=this;return I(M,function(Ce){return tg(Ce,fe,Y)})})})}function BC(I,M){M=M===o?" ":sg(M);var Y=M.length;if(Y<2)return Y?X6(M,I):M;var fe=X6(M,lk(I/r0(M)));return mv(M)?Cy(Qm(fe),0,I).join(""):fe.slice(0,I)}function BU(I,M,Y,fe){var Ce=M&U,ze=ob(I);function dt(){for(var It=-1,qt=arguments.length,Hn=-1,qn=fe.length,sr=En(qn+qt),ri=this&&this!==dd&&this instanceof dt?ze:I;++Hn<qn;)sr[Hn]=fe[Hn];for(;qt--;)sr[Hn++]=arguments[++It];return tg(ri,Ce?Y:this,sr)}return dt}function Ay(I){return function(M,Y,fe){return fe&&typeof fe!="number"&&Sp(M,Y,fe)&&(Y=fe=o),M=a1(M),Y===o?(Y=M,M=0):Y=a1(Y),fe=fe===o?M<Y?1:-1:a1(fe),$8(M,Y,fe,I)}}function Y2(I){return function(M,Y){return typeof M=="string"&&typeof Y=="string"||(M=om(M),Y=om(Y)),I(M,Y)}}function o5(I,M,Y,fe,Ce,ze,dt,It,qt,Hn){var qn=M&K,sr=qn?dt:o,ri=qn?o:dt,Ji=qn?ze:o,Uo=qn?o:ze;M|=qn?ae:oe,M&=~(qn?oe:ae),M&G||(M&=-4);var ns=[I,M,Ce,Ji,sr,Uo,ri,It,qt,Hn],zo=Y.apply(o,ns);return YO(I)&&KC(zo,ns),zo.placeholder=fe,J2(zo,I,M)}function zO(I){var M=Nd[I];return function(Y,fe){if(Y=om(Y),fe=fe==null?0:sl(fa(fe),292),fe&&bC(Y)){var Ce=(vc(Y)+"e").split("e"),ze=M(Ce[0]+"e"+(+Ce[1]+fe));return Ce=(vc(ze)+"e").split("e"),+(Ce[0]+"e"+(+Ce[1]-fe))}return M(Y)}}var K8=vs&&1/mO(new vs([,-0]))[1]==qe?function(I){return new vs(I)}:bi;function UC(I){return function(M){var Y=Cf(M);return Y==ro?gO(M):Y==Fc?NU(M):IU(M,I(M))}}function s0(I,M,Y,fe,Ce,ze,dt,It){var qt=M&B;if(!qt&&typeof I!="function")throw new Jg(h);var Hn=fe?fe.length:0;if(Hn||(M&=-97,fe=Ce=o),dt=dt===o?dt:qu(fa(dt),0),It=It===o?It:fa(It),Hn-=Ce?Ce.length:0,M&oe){var qn=fe,sr=Ce;fe=Ce=o}var ri=qt?o:qO(I),Ji=[I,M,Y,fe,Ce,qn,sr,ze,dt,It];if(ri&&VC(Ji,ri),I=Ji[0],M=Ji[1],Y=Ji[2],fe=Ji[3],Ce=Ji[4],It=Ji[9]=Ji[9]===o?qt?0:I.length:qu(Ji[9]-Hn,0),!It&&M&(K|X)&&(M&=-25),!M||M==U)var Uo=V8(I,M,Y);else M==K||M==X?Uo=W8(I,M,It):(M==ae||M==(U|ae))&&!Ce.length?Uo=BU(I,M,Y,fe):Uo=Gl.apply(o,Ji);var ns=ri?OC:KC;return J2(ns(Uo,Ji),I,M)}function ug(I,M,Y,fe){return I===o||hg(I,sk[Y])&&!bc.call(fe,Y)?M:I}function a5(I,M,Y,fe,Ce,ze){return ul(I)&&ul(M)&&(ze.set(M,I),IC(I,M,o,a5,ze),ze.delete(M)),I}function xv(I){return cb(I)?o:I}function s5(I,M,Y,fe,Ce,ze){var dt=Y&D,It=I.length,qt=M.length;if(It!=qt&&!(dt&&qt>It))return!1;var Hn=ze.get(I),qn=ze.get(M);if(Hn&&qn)return Hn==M&&qn==I;var sr=-1,ri=!0,Ji=Y&$?new wv:o;for(ze.set(I,M),ze.set(M,I);++sr<It;){var Uo=I[sr],ns=M[sr];if(fe)var zo=dt?fe(ns,Uo,sr,M,I,ze):fe(Uo,ns,sr,I,M,ze);if(zo!==o){if(zo)continue;ri=!1;break}if(Ji){if(!_6(M,function(rs,Xs){if(!rk(Ji,Xs)&&(Uo===rs||Ce(Uo,rs,Y,fe,ze)))return Ji.push(Xs)})){ri=!1;break}}else if(!(Uo===ns||Ce(Uo,ns,Y,fe,ze))){ri=!1;break}}return ze.delete(I),ze.delete(M),ri}function Y8(I,M,Y,fe,Ce,ze,dt){switch(Y){case Kr:if(I.byteLength!=M.byteLength||I.byteOffset!=M.byteOffset)return!1;I=I.buffer,M=M.buffer;case Pn:return!(I.byteLength!=M.byteLength||!ze(new uk(I),new uk(M)));case Nt:case on:case Ko:return hg(+I,+M);case wi:return I.name==M.name&&I.message==M.message;case Eo:case Yo:return I==M+"";case ro:var It=gO;case Fc:var qt=fe&D;if(It||(It=mO),I.size!=M.size&&!qt)return!1;var Hn=dt.get(I);if(Hn)return Hn==M;fe|=$,dt.set(I,M);var qn=s5(It(I),It(M),fe,Ce,ze,dt);return dt.delete(I),qn;case zl:if(M2)return M2.call(I)==M2.call(M)}return!1}function xf(I,M,Y,fe,Ce,ze){var dt=Y&D,It=HO(I),qt=It.length,Hn=HO(M),qn=Hn.length;if(qt!=qn&&!dt)return!1;for(var sr=qt;sr--;){var ri=It[sr];if(!(dt?ri in M:bc.call(M,ri)))return!1}var Ji=ze.get(I),Uo=ze.get(M);if(Ji&&Uo)return Ji==M&&Uo==I;var ns=!0;ze.set(I,M),ze.set(M,I);for(var zo=dt;++sr<qt;){ri=It[sr];var rs=I[ri],Xs=M[ri];if(fe)var s1=dt?fe(Xs,rs,ri,M,I,ze):fe(rs,Xs,ri,I,M,ze);if(!(s1===o?rs===Xs||Ce(rs,Xs,Y,fe,ze):s1)){ns=!1;break}zo||(zo=ri=="constructor")}if(ns&&!zo){var pg=I.constructor,c1=M.constructor;pg!=c1&&"constructor"in I&&"constructor"in M&&!(typeof pg=="function"&&pg instanceof pg&&typeof c1=="function"&&c1 instanceof c1)&&(ns=!1)}return ze.delete(I),ze.delete(M),ns}function Vu(I){return YC(WC(I,o,dg),I+"")}function HO(I){return R8(I,Zd,HC)}function GO(I){return R8(I,Af,WO)}var qO=dk?function(I){return dk.get(I)}:bi;function zC(I){for(var M=I.name+"",Y=rb[M],fe=bc.call(rb,M)?Y.length:0;fe--;){var Ce=Y[fe],ze=Ce.func;if(ze==null||ze==I)return Ce.name}return M}function X2(I){var M=bc.call(De,"placeholder")?De:I;return M.placeholder}function fo(){var I=De.iteratee||Do;return I=I===Do?G6:I,arguments.length?I(arguments[0],arguments[1]):I}function Ek(I,M){var Y=I.__data__;return Av(M)?Y[typeof M=="string"?"string":"hash"]:Y.map}function VO(I){for(var M=Zd(I),Y=M.length;Y--;){var fe=M[Y],Ce=I[fe];M[Y]=[fe,Ce,Q8(Ce)]}return M}function Iy(I,M){var Y=x6(I,M);return o0(Y)?Y:o}function X8(I){var M=bc.call(I,vp),Y=I[vp];try{I[vp]=o;var fe=!0}catch{}var Ce=bp.call(I);return fe&&(M?I[vp]=Y:delete I[vp]),Ce}var HC=Ey?function(I){return I==null?[]:(I=Ks(I),n0(Ey(I),function(M){return R2.call(I,M)}))}:kp,WO=Ey?function(I){for(var M=[];I;)gv(M,HC(I)),I=pC(I);return M}:kp,Cf=Cs;(vC&&Cf(new vC(new ArrayBuffer(1)))!=Kr||Rd&&Cf(new Rd)!=ro||cl&&Cf(cl.resolve())!=yi||vs&&Cf(new vs)!=Fc||Jb&&Cf(new Jb)!=Yc)&&(Cf=function(I){var M=Cs(I),Y=M==qi?I.constructor:o,fe=Y?Ov(Y):"";if(fe)switch(fe){case v8:return Kr;case w8:return ro;case wC:return yi;case yC:return Fc;case y8:return Yc}return M});function J8(I,M,Y){for(var fe=-1,Ce=Y.length;++fe<Ce;){var ze=Y[fe],dt=ze.size;switch(ze.type){case"drop":I+=dt;break;case"dropRight":M-=dt;break;case"take":M=sl(M,I+dt);break;case"takeRight":I=qu(I,M-dt);break}}return{start:I,end:M}}function c5(I){var M=I.match(i6);return M?M[1].split(XI):[]}function GC(I,M,Y){M=_v(M,I);for(var fe=-1,Ce=M.length,ze=!1;++fe<Ce;){var dt=Ys(M[fe]);if(!(ze=I!=null&&Y(I,dt)))break;I=I[dt]}return ze||++fe!=Ce?ze:(Ce=I==null?0:I.length,!!Ce&&tS(Ce)&&t1(dt,Ce)&&(Xo(I)||Pv(I)))}function mt(I){var M=I.length,Y=new I.constructor(M);return M&&typeof I[0]=="string"&&bc.call(I,"index")&&(Y.index=I.index,Y.input=I.input),Y}function u5(I){return typeof I.constructor=="function"&&!pu(I)?Lh(pC(I)):{}}function l5(I,M,Y){var fe=I.constructor;switch(M){case Pn:return G2(I);case Nt:case on:return new fe(+I);case Kr:return t5(I,Y);case Ei:case Sa:case bs:case Cc:case Ac:case Xc:case rc:case js:case Jc:return n5(I,Y);case ro:return new fe;case Ko:case Yo:return new fe(I);case Eo:return UO(I);case Fc:return new fe;case zl:return H8(I)}}function Z8(I,M){var Y=M.length;if(!Y)return I;var fe=Y-1;return M[fe]=(Y>1?"& ":"")+M[fe],M=M.join(Y>2?", ":" "),I.replace(r6,`{
/* [wrapped with `+M+`] */
`)}function Cv(I){return Xo(I)||Pv(I)||!!(ng&&I&&I[ng])}function t1(I,M){var Y=typeof I;return M=M??it,!!M&&(Y=="number"||Y!="symbol"&&hv.test(I))&&I>-1&&I%1==0&&I<M}function Sp(I,M,Y){if(!ul(Y))return!1;var fe=typeof M;return(fe=="number"?Xd(Y)&&t1(M,Y.length):fe=="string"&&M in Y)?hg(Y[M],I):!1}function KO(I,M){if(Xo(I))return!1;var Y=typeof I;return Y=="number"||Y=="symbol"||Y=="boolean"||I==null||_p(I)?!0:n6.test(I)||!YI.test(I)||M!=null&&I in Ks(M)}function Av(I){var M=typeof I;return M=="string"||M=="number"||M=="symbol"||M=="boolean"?I!=="__proto__":I===null}function YO(I){var M=zC(I),Y=De[M];if(typeof Y!="function"||!(M in ts.prototype))return!1;if(I===Y)return!0;var fe=qO(Y);return!!fe&&I===fe[0]}function XO(I){return!!I6&&I6 in I}var qC=fC?i1:Fk;function pu(I){var M=I&&I.constructor,Y=typeof M=="function"&&M.prototype||sk;return I===Y}function Q8(I){return I===I&&!ul(I)}function e9(I,M){return function(Y){return Y==null?!1:Y[I]===M&&(M!==o||I in Ks(Y))}}function JO(I){var M=o3(I,function(fe){return Y.size===E&&Y.clear(),fe}),Y=M.cache;return M}function VC(I,M){var Y=I[1],fe=M[1],Ce=Y|fe,ze=Ce<(U|B|he),dt=fe==he&&Y==K||fe==he&&Y==ye&&I[7].length<=M[8]||fe==(he|ye)&&M[7].length<=M[8]&&Y==K;if(!(ze||dt))return I;fe&U&&(I[2]=M[2],Ce|=Y&U?0:G);var It=M[3];if(It){var qt=I[3];I[3]=qt?r5(qt,It,M[4]):It,I[4]=qt?wy(I[3],k):M[4]}return It=M[5],It&&(qt=I[5],I[5]=qt?q8(qt,It,M[6]):It,I[6]=qt?wy(I[5],k):M[6]),It=M[7],It&&(I[7]=It),fe&he&&(I[8]=I[8]==null?M[8]:sl(I[8],M[8])),I[9]==null&&(I[9]=M[9]),I[0]=M[0],I[1]=Ce,I}function Oy(I){var M=[];if(I!=null)for(var Y in Ks(I))M.push(Y);return M}function t9(I){return bp.call(I)}function WC(I,M,Y){return M=qu(M===o?I.length-1:M,0),function(){for(var fe=arguments,Ce=-1,ze=qu(fe.length-M,0),dt=En(ze);++Ce<ze;)dt[Ce]=fe[M+Ce];Ce=-1;for(var It=En(M+1);++Ce<M;)It[Ce]=fe[Ce];return It[M]=Y(dt),tg(I,this,It)}}function Iv(I,M){return M.length<2?I:ib(I,Zg(M,0,-1))}function em(I,M){for(var Y=I.length,fe=sl(M.length,Y),Ce=Kd(I);fe--;){var ze=M[fe];I[fe]=t1(ze,Y)?Ce[ze]:o}return I}function ZO(I,M){if(!(M==="constructor"&&typeof I[M]=="function")&&M!="__proto__")return I[M]}var KC=XC(OC),Sk=N6||function(I,M){return dd.setTimeout(I,M)},YC=XC(NC);function J2(I,M,Y){var fe=M+"";return YC(I,Z8(fe,n9(c5(fe),Y)))}function XC(I){var M=0,Y=0;return function(){var fe=vv(),Ce=Ue-(fe-Y);if(Y=fe,Ce>0){if(++M>=xe)return arguments[0]}else M=0;return I.apply(o,arguments)}}function JC(I,M){var Y=-1,fe=I.length,Ce=fe-1;for(M=M===o?fe:M;++Y<M;){var ze=MO(Y,Ce),dt=I[ze];I[ze]=I[Y],I[Y]=dt}return I.length=M,I}var c0=JO(function(I){var M=[];return I.charCodeAt(0)===46&&M.push(""),I.replace(e0,function(Y,fe,Ce,ze){M.push(Ce?ze.replace(fv,"$1"):fe||Y)}),M});function Ys(I){if(typeof I=="string"||_p(I))return I;var M=I+"";return M=="0"&&1/I==-qe?"-0":M}function Ov(I){if(I!=null){try{return hC.call(I)}catch{}try{return I+""}catch{}}return""}function n9(I,M){return Zm($t,function(Y){var fe="_."+Y[0];M&Y[1]&&!cO(I,fe)&&I.push(fe)}),I.sort()}function Yd(I){if(I instanceof ts)return I.clone();var M=new ig(I.__wrapped__,I.__chain__);return M.__actions__=Kd(I.__actions__),M.__index__=I.__index__,M.__values__=I.__values__,M}function UU(I,M,Y){(Y?Sp(I,M,Y):M===o)?M=1:M=qu(fa(M),0);var fe=I==null?0:I.length;if(!fe||M<1)return[];for(var Ce=0,ze=0,dt=En(lk(fe/M));Ce<fe;)dt[ze++]=Zg(I,Ce,Ce+=M);return dt}function ZC(I){for(var M=-1,Y=I==null?0:I.length,fe=0,Ce=[];++M<Y;){var ze=I[M];ze&&(Ce[fe++]=ze)}return Ce}function r9(){var I=arguments.length;if(!I)return[];for(var M=En(I-1),Y=arguments[0],fe=I;fe--;)M[fe-1]=arguments[fe];return gv(Xo(Y)?Kd(Y):[Y],ws(M,1))}var Wu=Da(function(I,M){return Vl(I)?Ty(I,ws(M,1,Vl,!0)):[]}),lg=Da(function(I,M){var Y=jh(M);return Vl(Y)&&(Y=o),Vl(I)?Ty(I,ws(M,1,Vl,!0),fo(Y,2)):[]}),n1=Da(function(I,M){var Y=jh(M);return Vl(Y)&&(Y=o),Vl(I)?Ty(I,ws(M,1,Vl,!0),o,Y):[]});function d5(I,M,Y){var fe=I==null?0:I.length;return fe?(M=Y||M===o?1:fa(M),Zg(I,M<0?0:M,fe)):[]}function i9(I,M,Y){var fe=I==null?0:I.length;return fe?(M=Y||M===o?1:fa(M),M=fe-M,Zg(I,0,M<0?0:M)):[]}function zU(I,M){return I&&I.length?LC(I,fo(M,3),!0,!0):[]}function f5(I,M){return I&&I.length?LC(I,fo(M,3),!0):[]}function tm(I,M,Y,fe){var Ce=I==null?0:I.length;return Ce?(Y&&typeof Y!="number"&&Sp(I,M,Y)&&(Y=0,fe=Ce),Wa(I,M,Y,fe)):[]}function r1(I,M,Y){var fe=I==null?0:I.length;if(!fe)return-1;var Ce=Y==null?0:fa(Y);return Ce<0&&(Ce=qu(fe+Ce,0)),uO(I,fo(M,3),Ce)}function h5(I,M,Y){var fe=I==null?0:I.length;if(!fe)return-1;var Ce=fe-1;return Y!==o&&(Ce=fa(Y),Ce=Y<0?qu(fe+Ce,0):sl(Ce,fe-1)),uO(I,fo(M,3),Ce,!0)}function dg(I){var M=I==null?0:I.length;return M?ws(I,1):[]}function o9(I){var M=I==null?0:I.length;return M?ws(I,qe):[]}function Ny(I,M){var Y=I==null?0:I.length;return Y?(M=M===o?1:fa(M),ws(I,M)):[]}function Z2(I){for(var M=-1,Y=I==null?0:I.length,fe={};++M<Y;){var Ce=I[M];fe[Ce[0]]=Ce[1]}return fe}function Nv(I){return I&&I.length?I[0]:o}function p5(I,M,Y){var fe=I==null?0:I.length;if(!fe)return-1;var Ce=Y==null?0:fa(Y);return Ce<0&&(Ce=qu(fe+Ce,0)),tk(I,M,Ce)}function g5(I){var M=I==null?0:I.length;return M?Zg(I,0,-1):[]}var u0=Da(function(I){var M=al(I,BO);return M.length&&M[0]===I[0]?B6(M):[]}),a9=Da(function(I){var M=jh(I),Y=al(I,BO);return M===jh(Y)?M=o:Y.pop(),Y.length&&Y[0]===I[0]?B6(Y,fo(M,2)):[]}),ab=Da(function(I){var M=jh(I),Y=al(I,BO);return M=typeof M=="function"?M:o,M&&Y.pop(),Y.length&&Y[0]===I[0]?B6(Y,o,M):[]});function Q2(I,M){return I==null?"":tb.call(I,M)}function jh(I){var M=I==null?0:I.length;return M?I[M-1]:o}function _k(I,M,Y){var fe=I==null?0:I.length;if(!fe)return-1;var Ce=fe;return Y!==o&&(Ce=fa(Y),Ce=Ce<0?qu(fe+Ce,0):sl(Ce,fe-1)),M===M?Xg(I,M,Ce):uO(I,nk,Ce,!0)}function s9(I,M){return I&&I.length?W6(I,fa(M)):o}var c9=Da(kk);function kk(I,M){return I&&I.length&&M&&M.length?DO(I,M):I}function u9(I,M,Y){return I&&I.length&&M&&M.length?DO(I,M,fo(Y,2)):I}function QO(I,M,Y){return I&&I.length&&M&&M.length?DO(I,M,o,Y):I}var l9=Vu(function(I,M){var Y=I==null?0:I.length,fe=TC(I,M);return Y6(I,al(M,function(Ce){return t1(Ce,Y)?+Ce:Ce}).sort(jC)),fe});function HU(I,M){var Y=[];if(!(I&&I.length))return Y;var fe=-1,Ce=[],ze=I.length;for(M=fo(M,3);++fe<ze;){var dt=I[fe];M(dt,fe,I)&&(Y.push(dt),Ce.push(fe))}return Y6(I,Ce),Y}function m5(I){return I==null?I:P6.call(I)}function zc(I,M,Y){var fe=I==null?0:I.length;return fe?(Y&&typeof Y!="number"&&Sp(I,M,Y)?(M=0,Y=fe):(M=M==null?0:fa(M),Y=Y===o?fe:fa(Y)),Zg(I,M,Y)):[]}function Nu(I,M){return RC(I,M)}function as(I,M,Y){return $O(I,M,fo(Y,2))}function $s(I,M){var Y=I==null?0:I.length;if(Y){var fe=RC(I,M);if(fe<Y&&hg(I[fe],M))return fe}return-1}function Hc(I,M){return RC(I,M,!0)}function nm(I,M,Y){return $O(I,M,fo(Y,2),!0)}function Ry(I,M){var Y=I==null?0:I.length;if(Y){var fe=RC(I,M,!0)-1;if(hg(I[fe],M))return fe}return-1}function Tk(I){return I&&I.length?J6(I):[]}function b5(I,M){return I&&I.length?J6(I,fo(M,2)):[]}function Rv(I){var M=I==null?0:I.length;return M?Zg(I,1,M):[]}function ql(I,M,Y){return I&&I.length?(M=Y||M===o?1:fa(M),Zg(I,0,M<0?0:M)):[]}function Py(I,M,Y){var fe=I==null?0:I.length;return fe?(M=Y||M===o?1:fa(M),M=fe-M,Zg(I,M<0?0:M,fe)):[]}function eS(I,M){return I&&I.length?LC(I,fo(M,3),!1,!0):[]}function v5(I,M){return I&&I.length?LC(I,fo(M,3)):[]}var sb=Da(function(I){return Sv(ws(I,1,Vl,!0))}),xk=Da(function(I){var M=jh(I);return Vl(M)&&(M=o),Sv(ws(I,1,Vl,!0),fo(M,2))}),QC=Da(function(I){var M=jh(I);return M=typeof M=="function"?M:o,Sv(ws(I,1,Vl,!0),o,M)});function fg(I){return I&&I.length?Sv(I):[]}function w5(I,M){return I&&I.length?Sv(I,fo(M,2)):[]}function y5(I,M){return M=typeof M=="function"?M:o,I&&I.length?Sv(I,o,M):[]}function Ck(I){if(!(I&&I.length))return[];var M=0;return I=n0(I,function(Y){if(Vl(Y))return M=qu(Y.length,M),!0}),vy(M,function(Y){return al(I,lO(Y))})}function e3(I,M){if(!(I&&I.length))return[];var Y=Ck(I);return M==null?Y:al(Y,function(fe){return tg(M,o,fe)})}var Ly=Da(function(I,M){return Vl(I)?Ty(I,M):[]}),ih=Da(function(I){return DC(n0(I,Vl))}),Dy=Da(function(I){var M=jh(I);return Vl(M)&&(M=o),DC(n0(I,Vl),fo(M,2))}),Ak=Da(function(I){var M=jh(I);return M=typeof M=="function"?M:o,DC(n0(I,Vl),o,M)}),Ik=Da(Ck);function E5(I,M){return Q6(I||[],M||[],F2)}function S5(I,M){return Q6(I||[],M||[],wk)}var d9=Da(function(I){var M=I.length,Y=M>1?I[M-1]:o;return Y=typeof Y=="function"?(I.pop(),Y):o,e3(I,Y)});function Ok(I){var M=De(I);return M.__chain__=!0,M}function _5(I,M){return M(I),I}function $h(I,M){return M(I)}var f9=Vu(function(I){var M=I.length,Y=M?I[0]:0,fe=this.__wrapped__,Ce=function(ze){return TC(ze,I)};return M>1||this.__actions__.length||!(fe instanceof ts)||!t1(Y)?this.thru(Ce):(fe=fe.slice(Y,+Y+(M?1:0)),fe.__actions__.push({func:$h,args:[Ce],thisArg:o}),new ig(fe,this.__chain__).thru(function(ze){return M&&!ze.length&&ze.push(o),ze}))});function rm(){return Ok(this)}function h9(){return new ig(this.value(),this.__chain__)}function p9(){this.__values__===o&&(this.__values__=im(this.value()));var I=this.__index__>=this.__values__.length,M=I?o:this.__values__[this.__index__++];return{done:I,value:M}}function GU(){return this}function My(I){for(var M,Y=this;Y instanceof wp;){var fe=Yd(Y);fe.__index__=0,fe.__values__=o,M?Ce.__wrapped__=fe:M=fe;var Ce=fe;Y=Y.__wrapped__}return Ce.__wrapped__=I,M}function eN(){var I=this.__wrapped__;if(I instanceof ts){var M=I;return this.__actions__.length&&(M=new ts(this)),M=M.reverse(),M.__actions__.push({func:$h,args:[m5],thisArg:o}),new ig(M,this.__chain__)}return this.thru(m5)}function k5(){return B8(this.__wrapped__,this.__actions__)}var Nk=V2(function(I,M,Y){bc.call(I,Y)?++I[Y]:Zb(I,Y,1)});function g9(I,M,Y){var fe=Xo(I)?o8:$6;return Y&&Sp(I,M,Y)&&(M=o),fe(I,fo(M,3))}function m9(I,M){var Y=Xo(I)?n0:Uc;return Y(I,fo(M,3))}var b9=Tv(r1),T5=Tv(h5);function x5(I,M){return ws(t3(I,M),1)}function qU(I,M){return ws(t3(I,M),qe)}function VU(I,M,Y){return Y=Y===o?1:fa(Y),ws(t3(I,M),Y)}function v9(I,M){var Y=Xo(I)?Zm:Ev;return Y(I,fo(M,3))}function C5(I,M){var Y=Xo(I)?kU:CO;return Y(I,fo(M,3))}var w9=V2(function(I,M,Y){bc.call(I,Y)?I[Y].push(M):Zb(I,Y,[M])});function Rk(I,M,Y,fe){I=Xd(I)?I:oh(I),Y=Y&&!fe?fa(Y):0;var Ce=I.length;return Y<0&&(Y=qu(Ce+Y,0)),hN(I)?Y<=Ce&&I.indexOf(M,Y)>-1:!!Ce&&tk(I,M,Y)>-1}var tN=Da(function(I,M,Y){var fe=-1,Ce=typeof M=="function",ze=Xd(I)?En(I.length):[];return Ev(I,function(dt){ze[++fe]=Ce?tg(M,dt,Y):z2(dt,M,Y)}),ze}),y9=V2(function(I,M,Y){Zb(I,Y,M)});function t3(I,M){var Y=Xo(I)?al:q6;return Y(I,fo(M,3))}function E9(I,M,Y,fe){return I==null?[]:(Xo(M)||(M=M==null?[]:[M]),Y=fe?o:Y,Xo(Y)||(Y=Y==null?[]:[Y]),K6(I,M,Y))}var S9=V2(function(I,M,Y){I[Y?0:1].push(M)},function(){return[[],[]]});function _9(I,M,Y){var fe=Xo(I)?S6:T6,Ce=arguments.length<3;return fe(I,fo(M,4),Y,Ce,Ev)}function k9(I,M,Y){var fe=Xo(I)?TU:T6,Ce=arguments.length<3;return fe(I,fo(M,4),Y,Ce,CO)}function nN(I,M){var Y=Xo(I)?n0:Uc;return Y(I,a3(fo(M,3)))}function T9(I){var M=Xo(I)?pk:vk;return M(I)}function x9(I,M,Y){(Y?Sp(I,M,Y):M===o)?M=1:M=fa(M);var fe=Xo(I)?I8:jO;return fe(I,M)}function gu(I){var M=Xo(I)?O8:F8;return M(I)}function A5(I){if(I==null)return 0;if(Xd(I))return hN(I)?r0(I):I.length;var M=Cf(I);return M==ro||M==Fc?I.size:NO(I).length}function WU(I,M,Y){var fe=Xo(I)?_6:FU;return Y&&Sp(I,M,Y)&&(M=o),fe(I,fo(M,3))}var C9=Da(function(I,M){if(I==null)return[];var Y=M.length;return Y>1&&Sp(I,M[0],M[1])?M=[]:Y>2&&Sp(M[0],M[1],M[2])&&(M=[M[0]]),K6(I,ws(M,1),[])}),rN=L2||function(){return dd.Date.now()};function KU(I,M){if(typeof M!="function")throw new Jg(h);return I=fa(I),function(){if(--I<1)return M.apply(this,arguments)}}function A9(I,M,Y){return M=Y?o:M,M=I&&M==null?I.length:M,s0(I,he,o,o,o,o,M)}function n3(I,M){var Y;if(typeof M!="function")throw new Jg(h);return I=fa(I),function(){return--I>0&&(Y=M.apply(this,arguments)),I<=1&&(M=o),Y}}var jy=Da(function(I,M,Y){var fe=U;if(Y.length){var Ce=wy(Y,X2(jy));fe|=ae}return s0(I,fe,M,Y,Ce)}),iN=Da(function(I,M,Y){var fe=U|B;if(Y.length){var Ce=wy(Y,X2(iN));fe|=ae}return s0(M,fe,I,Y,Ce)});function r3(I,M,Y){M=Y?o:M;var fe=s0(I,K,o,o,o,o,o,M);return fe.placeholder=r3.placeholder,fe}function oN(I,M,Y){M=Y?o:M;var fe=s0(I,X,o,o,o,o,o,M);return fe.placeholder=oN.placeholder,fe}function $y(I,M,Y){var fe,Ce,ze,dt,It,qt,Hn=0,qn=!1,sr=!1,ri=!0;if(typeof I!="function")throw new Jg(h);M=om(M)||0,ul(Y)&&(qn=!!Y.leading,sr="maxWait"in Y,ze=sr?qu(om(Y.maxWait)||0,M):ze,ri="trailing"in Y?!!Y.trailing:ri);function Ji(Qd){var jv=fe,Hy=Ce;return fe=Ce=o,Hn=Qd,dt=I.apply(Hy,jv),dt}function Uo(Qd){return Hn=Qd,It=Sk(rs,M),qn?Ji(Qd):dt}function ns(Qd){var jv=Qd-qt,Hy=Qd-Hn,wz=M-jv;return sr?sl(wz,ze-Hy):wz}function zo(Qd){var jv=Qd-qt,Hy=Qd-Hn;return qt===o||jv>=M||jv<0||sr&&Hy>=ze}function rs(){var Qd=rN();if(zo(Qd))return Xs(Qd);It=Sk(rs,ns(Qd))}function Xs(Qd){return It=o,ri&&fe?Ji(Qd):(fe=Ce=o,dt)}function s1(){It!==o&&z8(It),Hn=0,fe=qt=Ce=It=o}function pg(){return It===o?dt:Xs(rN())}function c1(){var Qd=rN(),jv=zo(Qd);if(fe=arguments,Ce=this,qt=Qd,jv){if(It===o)return Uo(qt);if(sr)return z8(It),It=Sk(rs,M),Ji(qt)}return It===o&&(It=Sk(rs,M)),dt}return c1.cancel=s1,c1.flush=pg,c1}var l0=Da(function(I,M){return B2(I,1,M)}),i3=Da(function(I,M,Y){return B2(I,om(M)||0,Y)});function YU(I){return s0(I,we)}function o3(I,M){if(typeof I!="function"||M!=null&&typeof M!="function")throw new Jg(h);var Y=function(){var fe=arguments,Ce=M?M.apply(this,fe):fe[0],ze=Y.cache;if(ze.has(Ce))return ze.get(Ce);var dt=I.apply(this,fe);return Y.cache=ze.set(Ce,dt)||ze,dt};return Y.cache=new(o3.Cache||og),Y}o3.Cache=og;function a3(I){if(typeof I!="function")throw new Jg(h);return function(){var M=arguments;switch(M.length){case 0:return!I.call(this);case 1:return!I.call(this,M[0]);case 2:return!I.call(this,M[0],M[1]);case 3:return!I.call(this,M[0],M[1],M[2])}return!I.apply(this,M)}}function I9(I){return n3(2,I)}var O9=U8(function(I,M){M=M.length==1&&Xo(M[0])?al(M[0],Yg(fo())):al(ws(M,1),Yg(fo()));var Y=M.length;return Da(function(fe){for(var Ce=-1,ze=sl(fe.length,Y);++Ce<ze;)fe[Ce]=M[Ce].call(this,fe[Ce]);return tg(I,this,fe)})}),I5=Da(function(I,M){var Y=wy(M,X2(I5));return s0(I,ae,o,M,Y)}),aN=Da(function(I,M){var Y=wy(M,X2(aN));return s0(I,oe,o,M,Y)}),N9=Vu(function(I,M){return s0(I,ye,o,o,o,M)});function R9(I,M){if(typeof I!="function")throw new Jg(h);return M=M===o?M:fa(M),Da(I,M)}function XU(I,M){if(typeof I!="function")throw new Jg(h);return M=M==null?0:qu(fa(M),0),Da(function(Y){var fe=Y[M],Ce=Cy(Y,0,M);return fe&&gv(Ce,fe),tg(I,this,Ce)})}function O5(I,M,Y){var fe=!0,Ce=!0;if(typeof I!="function")throw new Jg(h);return ul(Y)&&(fe="leading"in Y?!!Y.leading:fe,Ce="trailing"in Y?!!Y.trailing:Ce),$y(I,M,{leading:fe,maxWait:M,trailing:Ce})}function N5(I){return A9(I,1)}function JU(I,M){return I5(MC(M),I)}function P9(){if(!arguments.length)return[];var I=arguments[0];return Xo(I)?I:[I]}function s3(I){return Dh(I,P)}function ZU(I,M){return M=typeof M=="function"?M:o,Dh(I,P,M)}function sN(I){return Dh(I,x|P)}function ss(I,M){return M=typeof M=="function"?M:o,Dh(I,x|P,M)}function R5(I,M){return M==null||j6(I,M,Zd(M))}function hg(I,M){return I===M||I!==I&&M!==M}var P5=Y2(xC),L9=Y2(function(I,M){return I>=M}),Pv=AO(function(){return arguments}())?AO:function(I){return kl(I)&&bc.call(I,"callee")&&!R2.call(I,"callee")},Xo=En.isArray,cN=sO?Yg(sO):ag;function Xd(I){return I!=null&&tS(I.length)&&!i1(I)}function Vl(I){return kl(I)&&Xd(I)}function QU(I){return I===!0||I===!1||kl(I)&&Cs(I)==Nt}var Fy=EO||Fk,D9=e8?Yg(e8):Pd;function M9(I){return kl(I)&&I.nodeType===1&&!cb(I)}function uN(I){if(I==null)return!0;if(Xd(I)&&(Xo(I)||typeof I=="string"||typeof I.splice=="function"||Fy(I)||Pk(I)||Pv(I)))return!I.length;var M=Cf(I);if(M==ro||M==Fc)return!I.size;if(pu(I))return!NO(I).length;for(var Y in I)if(bc.call(I,Y))return!1;return!0}function eu(I,M){return H2(I,M)}function L5(I,M,Y){Y=typeof Y=="function"?Y:o;var fe=Y?Y(I,M):o;return fe===o?H2(I,M,o,Y):!!fe}function Jd(I){if(!kl(I))return!1;var M=Cs(I);return M==wi||M==br||typeof I.message=="string"&&typeof I.name=="string"&&!cb(I)}function ez(I){return typeof I=="number"&&bC(I)}function i1(I){if(!ul(I))return!1;var M=Cs(I);return M==no||M==Ri||M==Or||M==lo}function lN(I){return typeof I=="number"&&I==fa(I)}function tS(I){return typeof I=="number"&&I>-1&&I%1==0&&I<=it}function ul(I){var M=typeof I;return I!=null&&(M=="object"||M=="function")}function kl(I){return I!=null&&typeof I=="object"}var dN=t8?Yg(t8):IO;function D5(I,M){return I===M||AC(I,M,VO(M))}function tz(I,M,Y){return Y=typeof Y=="function"?Y:o,AC(I,M,VO(M),Y)}function j9(I){return F9(I)&&I!=+I}function nz(I){if(qC(I))throw new Lo(d);return o0(I)}function $9(I){return I===null}function rz(I){return I==null}function F9(I){return typeof I=="number"||kl(I)&&Cs(I)==Ko}function cb(I){if(!kl(I)||Cs(I)!=qi)return!1;var M=pC(I);if(M===null)return!0;var Y=bc.call(M,"constructor")&&M.constructor;return typeof Y=="function"&&Y instanceof Y&&hC.call(Y)==eb}var fN=n8?Yg(n8):OO;function M5(I){return lN(I)&&I>=-it&&I<=it}var B9=r8?Yg(r8):z6;function hN(I){return typeof I=="string"||!Xo(I)&&kl(I)&&Cs(I)==Yo}function _p(I){return typeof I=="symbol"||kl(I)&&Cs(I)==zl}var Pk=i8?Yg(i8):H6;function o1(I){return I===o}function j5(I){return kl(I)&&Cf(I)==Yc}function U9(I){return kl(I)&&Cs(I)==Sn}var z9=Y2(RO),c3=Y2(function(I,M){return I<=M});function im(I){if(!I)return[];if(Xd(I))return hN(I)?Qm(I):Kd(I);if(P2&&I[P2])return h8(I[P2]());var M=Cf(I),Y=M==ro?gO:M==Fc?mO:oh;return Y(I)}function a1(I){if(!I)return I===0?I:0;if(I=om(I),I===qe||I===-qe){var M=I<0?-1:1;return M*Xe}return I===I?I:0}function fa(I){var M=a1(I),Y=M%1;return M===M?Y?M-Y:M:0}function d0(I){return I?Qb(fa(I),0,We):0}function om(I){if(typeof I=="number")return I;if(_p(I))return de;if(ul(I)){var M=typeof I.valueOf=="function"?I.valueOf():I;I=ul(M)?M+"":M}if(typeof I!="string")return I===0?I:+I;I=u8(I);var Y=Od.test(I);return Y||ZI.test(I)?SU(I.slice(2),Y?2:8):ly.test(I)?de:+I}function $5(I){return cg(I,Af(I))}function ll(I){return I?Qb(fa(I),-it,it):I===0?I:0}function vc(I){return I==null?"":sg(I)}var H9=kv(function(I,M){if(pu(M)||Xd(M)){cg(M,Zd(M),I);return}for(var Y in M)bc.call(M,Y)&&F2(I,Y,M[Y])}),G9=kv(function(I,M){cg(M,Af(M),I)}),nS=kv(function(I,M,Y,fe){cg(M,Af(M),I,fe)}),iz=kv(function(I,M,Y,fe){cg(M,Zd(M),I,fe)}),oz=Vu(TC);function az(I,M){var Y=Lh(I);return M==null?Y:yv(Y,M)}var q9=Da(function(I,M){I=Ks(I);var Y=-1,fe=M.length,Ce=fe>2?M[2]:o;for(Ce&&Sp(M[0],M[1],Ce)&&(fe=1);++Y<fe;)for(var ze=M[Y],dt=Af(ze),It=-1,qt=dt.length;++It<qt;){var Hn=dt[It],qn=I[Hn];(qn===o||hg(qn,sk[Hn])&&!bc.call(I,Hn))&&(I[Hn]=ze[Hn])}return I}),F5=Da(function(I){return I.push(o,a5),tg(l3,o,I)});function Lk(I,M){return a8(I,fo(M,3),Mh)}function rS(I,M){return a8(I,fo(M,3),bk)}function V9(I,M){return I==null?I:mk(I,fo(M,3),Af)}function sz(I,M){return I==null?I:U2(I,fo(M,3),Af)}function Dk(I,M){return I&&Mh(I,fo(M,3))}function u3(I,M){return I&&bk(I,fo(M,3))}function W9(I){return I==null?[]:xy(I,Zd(I))}function pN(I){return I==null?[]:xy(I,Af(I))}function iS(I,M,Y){var fe=I==null?o:ib(I,M);return fe===o?Y:fe}function B5(I,M){return I!=null&&GC(I,M,F6)}function By(I,M){return I!=null&&GC(I,M,P8)}var K9=$C(function(I,M,Y){M!=null&&typeof M.toString!="function"&&(M=bp.call(M)),I[M]=Y},Xt(Zt)),gN=$C(function(I,M,Y){M!=null&&typeof M.toString!="function"&&(M=bp.call(M)),bc.call(I,M)?I[M].push(Y):I[M]=[Y]},fo),cz=Da(z2);function Zd(I){return Xd(I)?M6(I):NO(I)}function Af(I){return Xd(I)?M6(I,!0):D8(I)}function Y9(I,M){var Y={};return M=fo(M,3),Mh(I,function(fe,Ce,ze){Zb(Y,M(fe,Ce,ze),fe)}),Y}function uz(I,M){var Y={};return M=fo(M,3),Mh(I,function(fe,Ce,ze){Zb(Y,Ce,M(fe,Ce,ze))}),Y}var lz=kv(function(I,M,Y){IC(I,M,Y)}),l3=kv(function(I,M,Y,fe){IC(I,M,Y,fe)}),mN=Vu(function(I,M){var Y={};if(I==null)return Y;var fe=!1;M=al(M,function(ze){return ze=_v(ze,I),fe||(fe=ze.length>1),ze}),cg(I,GO(I),Y),fe&&(Y=Dh(Y,x|O|P,xv));for(var Ce=M.length;Ce--;)PC(Y,M[Ce]);return Y});function U5(I,M){return Mk(I,a3(fo(M)))}var dz=Vu(function(I,M){return I==null?{}:M8(I,M)});function Mk(I,M){if(I==null)return{};var Y=al(GO(I),function(fe){return[fe]});return M=fo(M),LO(I,Y,function(fe,Ce){return M(fe,Ce[0])})}function fz(I,M,Y){M=_v(M,I);var fe=-1,Ce=M.length;for(Ce||(Ce=1,I=o);++fe<Ce;){var ze=I==null?o:I[Ys(M[fe])];ze===o&&(fe=Ce,ze=Y),I=i1(ze)?ze.call(I):ze}return I}function X9(I,M,Y){return I==null?I:wk(I,M,Y)}function J9(I,M,Y,fe){return fe=typeof fe=="function"?fe:o,I==null?I:wk(I,M,Y,fe)}var z5=UC(Zd),H5=UC(Af);function bN(I,M,Y){var fe=Xo(I),Ce=fe||Fy(I)||Pk(I);if(M=fo(M,4),Y==null){var ze=I&&I.constructor;Ce?Y=fe?new ze:[]:ul(I)?Y=i1(ze)?Lh(pC(I)):{}:Y={}}return(Ce?Zm:Mh)(I,function(dt,It,qt){return M(Y,dt,It,qt)}),Y}function G5(I,M){return I==null?!0:PC(I,M)}function oS(I,M,Y){return I==null?I:Z6(I,M,MC(Y))}function d3(I,M,Y,fe){return fe=typeof fe=="function"?fe:o,I==null?I:Z6(I,M,MC(Y),fe)}function oh(I){return I==null?[]:hO(I,Zd(I))}function q5(I){return I==null?[]:hO(I,Af(I))}function Z9(I,M,Y){return Y===o&&(Y=M,M=o),Y!==o&&(Y=om(Y),Y=Y===Y?Y:0),M!==o&&(M=om(M),M=M===M?M:0),Qb(om(I),M,Y)}function V5(I,M,Y){return M=a1(M),Y===o?(Y=M,M=0):Y=a1(Y),I=om(I),L8(I,M,Y)}function Q9(I,M,Y){if(Y&&typeof Y!="boolean"&&Sp(I,M,Y)&&(M=Y=o),Y===o&&(typeof M=="boolean"?(Y=M,M=o):typeof I=="boolean"&&(Y=I,I=o)),I===o&&M===o?(I=0,M=1):(I=a1(I),M===o?(M=I,I=0):M=a1(M)),I>M){var fe=I;I=M,M=fe}if(Y||I%1||M%1){var Ce=R6();return sl(I+Ce*(M-I+dC("1e-"+((Ce+"").length-1))),M)}return MO(I,M)}var hz=Ep(function(I,M,Y){return M=M.toLowerCase(),I+(Y?W5(M):M)});function W5(I){return j(vc(I).toLowerCase())}function Uy(I){return I=vc(I),I&&I.replace(dy,pO).replace(pv,"")}function Lv(I,M,Y){I=vc(I),M=sg(M);var fe=I.length;Y=Y===o?fe:Qb(fa(Y),0,fe);var Ce=Y;return Y-=M.length,Y>=0&&I.slice(Y,Ce)==M}function Wl(I){return I=vc(I),I&&S2.test(I)?I.replace(Kb,l8):I}function vN(I){return I=vc(I),I&&rh.test(I)?I.replace(_l,"\\$&"):I}var jk=Ep(function(I,M,Y){return I+(Y?"-":"")+M.toLowerCase()}),Dv=Ep(function(I,M,Y){return I+(Y?" ":"")+M.toLowerCase()}),GK=Qg("toLowerCase");function pz(I,M,Y){I=vc(I),M=fa(M);var fe=M?r0(I):0;if(!M||fe>=M)return I;var Ce=(M-fe)/2;return BC(bv(Ce),Y)+I+BC(lk(Ce),Y)}function wN(I,M,Y){I=vc(I),M=fa(M);var fe=M?r0(I):0;return M&&fe<M?I+BC(M-fe,Y):I}function qK(I,M,Y){I=vc(I),M=fa(M);var fe=M?r0(I):0;return M&&fe<M?BC(M-fe,Y)+I:I}function yN(I,M,Y){return Y||M==null?M=0:M&&(M=+M),D2(vc(I).replace(t0,""),M||0)}function aS(I,M,Y){return(Y?Sp(I,M,Y):M===o)?M=1:M=fa(M),X6(vc(I),M)}function K5(){var I=arguments,M=vc(I[0]);return I.length<3?M:M.replace(I[1],I[2])}var VK=Ep(function(I,M,Y){return I+(Y?"_":"")+M.toLowerCase()});function Y5(I,M,Y){return Y&&typeof Y!="number"&&Sp(I,M,Y)&&(M=Y=o),Y=Y===o?We:Y>>>0,Y?(I=vc(I),I&&(typeof M=="string"||M!=null&&!fN(M))&&(M=sg(M),!M&&mv(I))?Cy(Qm(I),0,Y):I.split(M,Y)):[]}var X5=Ep(function(I,M,Y){return I+(Y?" ":"")+j(M)});function e7(I,M,Y){return I=vc(I),Y=Y==null?0:Qb(fa(Y),0,I.length),M=sg(M),I.slice(Y,Y+M.length)==M}function EN(I,M,Y){var fe=De.templateSettings;Y&&Sp(I,M,Y)&&(M=o),I=vc(I),M=nS({},M,fe,ug);var Ce=nS({},M.imports,fe.imports,ug),ze=Zd(Ce),dt=hO(Ce,ze),It,qt,Hn=0,qn=M.interpolate||T2,sr="__p += '",ri=bO((M.escape||T2).source+"|"+qn.source+"|"+(qn===K_?cy:T2).source+"|"+(M.evaluate||T2).source+"|$","g"),Ji="//# sourceURL="+(bc.call(M,"sourceURL")?(M.sourceURL+"").replace(/\s/g," "):"lodash.templateSources["+ ++JM+"]")+`
`;I.replace(ri,function(zo,rs,Xs,s1,pg,c1){return Xs||(Xs=s1),sr+=I.slice(Hn,c1).replace(QI,d8),rs&&(It=!0,sr+=`' +
__e(`+rs+`) +
'`),pg&&(qt=!0,sr+=`';
`+pg+`;
__p += '`),Xs&&(sr+=`' +
((__t = (`+Xs+`)) == null ? '' : __t) +
'`),Hn=c1+zo.length,zo}),sr+=`';
`;var Uo=bc.call(M,"variable")&&M.variable;if(!Uo)sr=`with (obj) {
`+sr+`
}
`;else if(rC.test(Uo))throw new Lo(b);sr=(qt?sr.replace(Id,""):sr).replace(nh,"$1").replace(hu,"$1;"),sr="function("+(Uo||"obj")+`) {
`+(Uo?"":`obj || (obj = {});
`)+"var __t, __p = ''"+(It?", __e = _.escape":"")+(qt?`, __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
`:`;
`)+sr+`return __p
}`;var ns=te(function(){return Ws(ze,Ji+"return "+sr).apply(o,dt)});if(ns.source=sr,Jd(ns))throw ns;return ns}function SN(I){return vc(I).toLowerCase()}function J5(I){return vc(I).toUpperCase()}function gz(I,M,Y){if(I=vc(I),I&&(Y||M===o))return u8(I);if(!I||!(M=sg(M)))return I;var fe=Qm(I),Ce=Qm(M),ze=ik(fe,Ce),dt=ok(fe,Ce)+1;return Cy(fe,ze,dt).join("")}function WK(I,M,Y){if(I=vc(I),I&&(Y||M===o))return I.slice(0,C6(I)+1);if(!I||!(M=sg(M)))return I;var fe=Qm(I),Ce=ok(fe,Qm(M))+1;return Cy(fe,0,Ce).join("")}function $k(I,M,Y){if(I=vc(I),I&&(Y||M===o))return I.replace(t0,"");if(!I||!(M=sg(M)))return I;var fe=Qm(I),Ce=ik(fe,Qm(M));return Cy(fe,Ce).join("")}function w(I,M){var Y=ge,fe=Te;if(ul(M)){var Ce="separator"in M?M.separator:Ce;Y="length"in M?fa(M.length):Y,fe="omission"in M?sg(M.omission):fe}I=vc(I);var ze=I.length;if(mv(I)){var dt=Qm(I);ze=dt.length}if(Y>=ze)return I;var It=Y-r0(fe);if(It<1)return fe;var qt=dt?Cy(dt,0,It).join(""):I.slice(0,It);if(Ce===o)return qt+fe;if(dt&&(It+=qt.length-It),fN(Ce)){if(I.slice(It).search(Ce)){var Hn,qn=qt;for(Ce.global||(Ce=bO(Ce.source,vc(uy.exec(Ce))+"g")),Ce.lastIndex=0;Hn=Ce.exec(qn);)var sr=Hn.index;qt=qt.slice(0,sr===o?It:sr)}}else if(I.indexOf(sg(Ce),It)!=It){var ri=qt.lastIndexOf(Ce);ri>-1&&(qt=qt.slice(0,ri))}return qt+fe}function _(I){return I=vc(I),I&&Km.test(I)?I.replace(Qp,g8):I}var C=Ep(function(I,M,Y){return I+(Y?" ":"")+M.toUpperCase()}),j=Qg("toUpperCase");function q(I,M,Y){return I=vc(I),M=Y?o:M,M===o?f8(I)?PU(I):AU(I):I.match(M)||[]}var te=Da(function(I,M){try{return tg(I,o,M)}catch(Y){return Jd(Y)?Y:new Lo(Y)}}),Se=Vu(function(I,M){return Zm(M,function(Y){Y=Ys(Y),Zb(I,Y,jy(I[Y],I))}),I});function Me(I){var M=I==null?0:I.length,Y=fo();return I=M?al(I,function(fe){if(typeof fe[1]!="function")throw new Jg(h);return[Y(fe[0]),fe[1]]}):[],Da(function(fe){for(var Ce=-1;++Ce<M;){var ze=I[Ce];if(tg(ze[0],this,fe))return tg(ze[1],this,fe)}})}function pt(I){return N8(Dh(I,x))}function Xt(I){return function(){return I}}function $n(I,M){return I==null||I!==I?M:I}var Vn=a0(),ln=a0(!0);function Zt(I){return I}function Do(I){return G6(typeof I=="function"?I:Dh(I,x))}function Oo(I){return V6(Dh(I,x))}function mu(I,M){return PO(I,Dh(M,x))}var zt=Da(function(I,M){return function(Y){return z2(Y,I,M)}}),Ot=Da(function(I,M){return function(Y){return z2(I,Y,M)}});function Wt(I,M,Y){var fe=Zd(M),Ce=xy(M,fe);Y==null&&!(ul(M)&&(Ce.length||!fe.length))&&(Y=M,M=I,I=this,Ce=xy(M,Zd(M)));var ze=!(ul(Y)&&"chain"in Y)||!!Y.chain,dt=i1(I);return Zm(Ce,function(It){var qt=M[It];I[It]=qt,dt&&(I.prototype[It]=function(){var Hn=this.__chain__;if(ze||Hn){var qn=I(this.__wrapped__),sr=qn.__actions__=Kd(this.__actions__);return sr.push({func:qt,args:arguments,thisArg:I}),qn.__chain__=Hn,qn}return qt.apply(I,gv([this.value()],arguments))})}),I}function zn(){return dd._===this&&(dd._=ck),this}function bi(){}function Ka(I){return I=fa(I),Da(function(M){return W6(M,I)})}var Hi=FC(al),Vi=FC(o8),Ic=FC(_6);function Fs(I){return KO(I)?lO(Ys(I)):j8(I)}function Mv(I){return function(M){return I==null?o:ib(I,M)}}var f3=Ay(),KK=Ay(!0);function kp(){return[]}function Fk(){return!1}function zy(){return{}}function nae(){return""}function rae(){return!0}function YK(I,M){if(I=fa(I),I<1||I>it)return[];var Y=We,fe=sl(I,We);M=fo(M),I-=We;for(var Ce=vy(fe,M);++Y<I;)M(Y);return Ce}function iae(I){return Xo(I)?al(I,Ys):_p(I)?[I]:Kd(c0(vc(I)))}function mz(I){var M=++Yb;return vc(I)+M}var h3=K2(function(I,M){return I+M},0),p3=zO("ceil"),bz=K2(function(I,M){return I/M},1),ne=zO("floor");function vz(I){return I&&I.length?e1(I,Zt,xC):o}function XK(I,M){return I&&I.length?e1(I,fo(M,2),xC):o}function wr(I){return s8(I,Zt)}function oae(I,M){return s8(I,fo(M,2))}function JK(I){return I&&I.length?e1(I,Zt,RO):o}function ZK(I,M){return I&&I.length?e1(I,fo(M,2),RO):o}var QK=K2(function(I,M){return I*M},1),eY=zO("round"),tY=K2(function(I,M){return I-M},0);function aae(I){return I&&I.length?fO(I,Zt):0}function g3(I,M){return I&&I.length?fO(I,fo(M,2)):0}return De.after=KU,De.ary=A9,De.assign=H9,De.assignIn=G9,De.assignInWith=nS,De.assignWith=iz,De.at=oz,De.before=n3,De.bind=jy,De.bindAll=Se,De.bindKey=iN,De.castArray=P9,De.chain=Ok,De.chunk=UU,De.compact=ZC,De.concat=r9,De.cond=Me,De.conforms=pt,De.constant=Xt,De.countBy=Nk,De.create=az,De.curry=r3,De.curryRight=oN,De.debounce=$y,De.defaults=q9,De.defaultsDeep=F5,De.defer=l0,De.delay=i3,De.difference=Wu,De.differenceBy=lg,De.differenceWith=n1,De.drop=d5,De.dropRight=i9,De.dropRightWhile=zU,De.dropWhile=f5,De.fill=tm,De.filter=m9,De.flatMap=x5,De.flatMapDeep=qU,De.flatMapDepth=VU,De.flatten=dg,De.flattenDeep=o9,De.flattenDepth=Ny,De.flip=YU,De.flow=Vn,De.flowRight=ln,De.fromPairs=Z2,De.functions=W9,De.functionsIn=pN,De.groupBy=w9,De.initial=g5,De.intersection=u0,De.intersectionBy=a9,De.intersectionWith=ab,De.invert=K9,De.invertBy=gN,De.invokeMap=tN,De.iteratee=Do,De.keyBy=y9,De.keys=Zd,De.keysIn=Af,De.map=t3,De.mapKeys=Y9,De.mapValues=uz,De.matches=Oo,De.matchesProperty=mu,De.memoize=o3,De.merge=lz,De.mergeWith=l3,De.method=zt,De.methodOf=Ot,De.mixin=Wt,De.negate=a3,De.nthArg=Ka,De.omit=mN,De.omitBy=U5,De.once=I9,De.orderBy=E9,De.over=Hi,De.overArgs=O9,De.overEvery=Vi,De.overSome=Ic,De.partial=I5,De.partialRight=aN,De.partition=S9,De.pick=dz,De.pickBy=Mk,De.property=Fs,De.propertyOf=Mv,De.pull=c9,De.pullAll=kk,De.pullAllBy=u9,De.pullAllWith=QO,De.pullAt=l9,De.range=f3,De.rangeRight=KK,De.rearg=N9,De.reject=nN,De.remove=HU,De.rest=R9,De.reverse=m5,De.sampleSize=x9,De.set=X9,De.setWith=J9,De.shuffle=gu,De.slice=zc,De.sortBy=C9,De.sortedUniq=Tk,De.sortedUniqBy=b5,De.split=Y5,De.spread=XU,De.tail=Rv,De.take=ql,De.takeRight=Py,De.takeRightWhile=eS,De.takeWhile=v5,De.tap=_5,De.throttle=O5,De.thru=$h,De.toArray=im,De.toPairs=z5,De.toPairsIn=H5,De.toPath=iae,De.toPlainObject=$5,De.transform=bN,De.unary=N5,De.union=sb,De.unionBy=xk,De.unionWith=QC,De.uniq=fg,De.uniqBy=w5,De.uniqWith=y5,De.unset=G5,De.unzip=Ck,De.unzipWith=e3,De.update=oS,De.updateWith=d3,De.values=oh,De.valuesIn=q5,De.without=Ly,De.words=q,De.wrap=JU,De.xor=ih,De.xorBy=Dy,De.xorWith=Ak,De.zip=Ik,De.zipObject=E5,De.zipObjectDeep=S5,De.zipWith=d9,De.entries=z5,De.entriesIn=H5,De.extend=G9,De.extendWith=nS,Wt(De,De),De.add=h3,De.attempt=te,De.camelCase=hz,De.capitalize=W5,De.ceil=p3,De.clamp=Z9,De.clone=s3,De.cloneDeep=sN,De.cloneDeepWith=ss,De.cloneWith=ZU,De.conformsTo=R5,De.deburr=Uy,De.defaultTo=$n,De.divide=bz,De.endsWith=Lv,De.eq=hg,De.escape=Wl,De.escapeRegExp=vN,De.every=g9,De.find=b9,De.findIndex=r1,De.findKey=Lk,De.findLast=T5,De.findLastIndex=h5,De.findLastKey=rS,De.floor=ne,De.forEach=v9,De.forEachRight=C5,De.forIn=V9,De.forInRight=sz,De.forOwn=Dk,De.forOwnRight=u3,De.get=iS,De.gt=P5,De.gte=L9,De.has=B5,De.hasIn=By,De.head=Nv,De.identity=Zt,De.includes=Rk,De.indexOf=p5,De.inRange=V5,De.invoke=cz,De.isArguments=Pv,De.isArray=Xo,De.isArrayBuffer=cN,De.isArrayLike=Xd,De.isArrayLikeObject=Vl,De.isBoolean=QU,De.isBuffer=Fy,De.isDate=D9,De.isElement=M9,De.isEmpty=uN,De.isEqual=eu,De.isEqualWith=L5,De.isError=Jd,De.isFinite=ez,De.isFunction=i1,De.isInteger=lN,De.isLength=tS,De.isMap=dN,De.isMatch=D5,De.isMatchWith=tz,De.isNaN=j9,De.isNative=nz,De.isNil=rz,De.isNull=$9,De.isNumber=F9,De.isObject=ul,De.isObjectLike=kl,De.isPlainObject=cb,De.isRegExp=fN,De.isSafeInteger=M5,De.isSet=B9,De.isString=hN,De.isSymbol=_p,De.isTypedArray=Pk,De.isUndefined=o1,De.isWeakMap=j5,De.isWeakSet=U9,De.join=Q2,De.kebabCase=jk,De.last=jh,De.lastIndexOf=_k,De.lowerCase=Dv,De.lowerFirst=GK,De.lt=z9,De.lte=c3,De.max=vz,De.maxBy=XK,De.mean=wr,De.meanBy=oae,De.min=JK,De.minBy=ZK,De.stubArray=kp,De.stubFalse=Fk,De.stubObject=zy,De.stubString=nae,De.stubTrue=rae,De.multiply=QK,De.nth=s9,De.noConflict=zn,De.noop=bi,De.now=rN,De.pad=pz,De.padEnd=wN,De.padStart=qK,De.parseInt=yN,De.random=Q9,De.reduce=_9,De.reduceRight=k9,De.repeat=aS,De.replace=K5,De.result=fz,De.round=eY,De.runInContext=Dt,De.sample=T9,De.size=A5,De.snakeCase=VK,De.some=WU,De.sortedIndex=Nu,De.sortedIndexBy=as,De.sortedIndexOf=$s,De.sortedLastIndex=Hc,De.sortedLastIndexBy=nm,De.sortedLastIndexOf=Ry,De.startCase=X5,De.startsWith=e7,De.subtract=tY,De.sum=aae,De.sumBy=g3,De.template=EN,De.times=YK,De.toFinite=a1,De.toInteger=fa,De.toLength=d0,De.toLower=SN,De.toNumber=om,De.toSafeInteger=ll,De.toString=vc,De.toUpper=J5,De.trim=gz,De.trimEnd=WK,De.trimStart=$k,De.truncate=w,De.unescape=_,De.uniqueId=mz,De.upperCase=C,De.upperFirst=j,De.each=v9,De.eachRight=C5,De.first=Nv,Wt(De,function(){var I={};return Mh(De,function(M,Y){bc.call(De.prototype,Y)||(I[Y]=M)}),I}(),{chain:!1}),De.VERSION=s,Zm(["bind","bindKey","curry","curryRight","partial","partialRight"],function(I){De[I].placeholder=De}),Zm(["drop","take"],function(I,M){ts.prototype[I]=function(Y){Y=Y===o?1:qu(fa(Y),0);var fe=this.__filtered__&&!M?new ts(this):this.clone();return fe.__filtered__?fe.__takeCount__=sl(Y,fe.__takeCount__):fe.__views__.push({size:sl(Y,We),type:I+(fe.__dir__<0?"Right":"")}),fe},ts.prototype[I+"Right"]=function(Y){return this.reverse()[I](Y).reverse()}}),Zm(["filter","map","takeWhile"],function(I,M){var Y=M+1,fe=Y==Z||Y==_e;ts.prototype[I]=function(Ce){var ze=this.clone();return ze.__iteratees__.push({iteratee:fo(Ce,3),type:Y}),ze.__filtered__=ze.__filtered__||fe,ze}}),Zm(["head","last"],function(I,M){var Y="take"+(M?"Right":"");ts.prototype[I]=function(){return this[Y](1).value()[0]}}),Zm(["initial","tail"],function(I,M){var Y="drop"+(M?"":"Right");ts.prototype[I]=function(){return this.__filtered__?new ts(this):this[Y](1)}}),ts.prototype.compact=function(){return this.filter(Zt)},ts.prototype.find=function(I){return this.filter(I).head()},ts.prototype.findLast=function(I){return this.reverse().find(I)},ts.prototype.invokeMap=Da(function(I,M){return typeof I=="function"?new ts(this):this.map(function(Y){return z2(Y,I,M)})}),ts.prototype.reject=function(I){return this.filter(a3(fo(I)))},ts.prototype.slice=function(I,M){I=fa(I);var Y=this;return Y.__filtered__&&(I>0||M<0)?new ts(Y):(I<0?Y=Y.takeRight(-I):I&&(Y=Y.drop(I)),M!==o&&(M=fa(M),Y=M<0?Y.dropRight(-M):Y.take(M-I)),Y)},ts.prototype.takeRightWhile=function(I){return this.reverse().takeWhile(I).reverse()},ts.prototype.toArray=function(){return this.take(We)},Mh(ts.prototype,function(I,M){var Y=/^(?:filter|find|map|reject)|While$/.test(M),fe=/^(?:head|last)$/.test(M),Ce=De[fe?"take"+(M=="last"?"Right":""):M],ze=fe||/^find/.test(M);Ce&&(De.prototype[M]=function(){var dt=this.__wrapped__,It=fe?[1]:arguments,qt=dt instanceof ts,Hn=It[0],qn=qt||Xo(dt),sr=function(rs){var Xs=Ce.apply(De,gv([rs],It));return fe&&ri?Xs[0]:Xs};qn&&Y&&typeof Hn=="function"&&Hn.length!=1&&(qt=qn=!1);var ri=this.__chain__,Ji=!!this.__actions__.length,Uo=ze&&!ri,ns=qt&&!Ji;if(!ze&&qn){dt=ns?dt:new ts(this);var zo=I.apply(dt,It);return zo.__actions__.push({func:$h,args:[sr],thisArg:o}),new ig(zo,ri)}return Uo&&ns?I.apply(this,It):(zo=this.thru(sr),Uo?fe?zo.value()[0]:zo.value():zo)})}),Zm(["pop","push","shift","sort","splice","unshift"],function(I){var M=vO[I],Y=/^(?:push|sort|unshift)$/.test(I)?"tap":"thru",fe=/^(?:pop|shift)$/.test(I);De.prototype[I]=function(){var Ce=arguments;if(fe&&!this.__chain__){var ze=this.value();return M.apply(Xo(ze)?ze:[],Ce)}return this[Y](function(dt){return M.apply(Xo(dt)?dt:[],Ce)})}}),Mh(ts.prototype,function(I,M){var Y=De[M];if(Y){var fe=Y.name+"";bc.call(rb,fe)||(rb[fe]=[]),rb[fe].push({name:M,func:Y})}}),rb[Gl(o,B).name]=[{name:"wrapper",func:o}],ts.prototype.clone=fk,ts.prototype.reverse=SO,ts.prototype.value=j2,De.prototype.at=f9,De.prototype.chain=rm,De.prototype.commit=h9,De.prototype.next=p9,De.prototype.plant=My,De.prototype.reverse=eN,De.prototype.toJSON=De.prototype.valueOf=De.prototype.value=k5,De.prototype.first=De.prototype.head,P2&&(De.prototype[P2]=GU),De},yy=ak();by?((by.exports=yy)._=yy,ek._=yy):dd._=yy}).call(I8r)}(Jre,Jre.exports)),Jre.exports}var OEe=O8r();const N8r=(n=1e3*60)=>{const[r,o]=Q.useState(!1);return Q.useEffect(()=>{let s;const c=()=>{o(!0)},d=OEe.throttle(()=>{o(!1),clearTimeout(s),s=globalThis.setTimeout(c,n)},500);return s=globalThis.setTimeout(c,n),globalThis.addEventListener("mousemove",d),globalThis.addEventListener("mousedown",d),globalThis.addEventListener("resize",d),globalThis.addEventListener("keydown",d),globalThis.addEventListener("touchstart",d),globalThis.addEventListener("wheel",d),()=>{globalThis.removeEventListener("mousemove",d),globalThis.removeEventListener("mousedown",d),globalThis.removeEventListener("resize",d),globalThis.removeEventListener("keydown",d),globalThis.removeEventListener("touchstart",d),globalThis.removeEventListener("wheel",d),clearTimeout(s),d.cancel()}},[n]),r},zln=()=>{const n=N8r(E5r),r=Q.useRef(null),{inactive:o}=Io(da,c=>({inactive:c.inactive})),s=()=>{r.current&&clearTimeout(r.current),queueMicrotask(()=>{Kp.dismiss("idle-notification")})};Q.useEffect(()=>{if(!o){if(n)return queueMicrotask(()=>{Kp.create({id:"idle-notification",title:"Still around?",description:"We’ll pause your project soon. Just move your mouse or press any key to stay connected.",type:"warning",duration:CFt,closable:!0,meta:{icon:Aor}})}),r.current&&clearTimeout(r.current),r.current=setTimeout(()=>{Bre(!0)},CFt),s;s()}},[n,o])},XV="retrain-notification",R8r={received:"Starting training...",validating:"Validating configuration...",validation_success:"Validation complete",training:"Training model...",train_success:"Training complete",done:"Training finished successfully!"},P8r=()=>{const{data:n}=IEe({interval:5e3}),r=Io(da,O=>O.projectUrl),o=Roe(),s=Q.useRef(!1),c=Q.useRef(null),{isStandalone:d}=GM(),h=Q.useCallback(()=>{c.current&&(c.current.close(),c.current=null)},[]),b=Q.useCallback(O=>{const P=R8r[O]||`Status: ${O}`;Kp.update(XV,{title:"Retraining",description:P,type:"loading",duration:1/0,closable:!1,action:void 0})},[]),y=Q.useCallback(()=>{Kp.update(XV,{title:"Training complete",description:"Your changes are now live.",type:"success",duration:5e3,closable:!0,action:void 0,meta:{icon:sXe}})},[]),E=Q.useCallback(O=>{Kp.update(XV,{title:"Training failed",description:O,type:"error",duration:1e4,closable:!0,meta:{icon:rnn}})},[]),k=Q.useCallback(O=>{if(!r)return;h();const P=new EventSource(`${r}${Vb.JOB_EVENTS}/${O}`);c.current=P,P.addEventListener("progress",D=>{try{const $=JSON.parse(D.data);b($.status),$.status==="done"&&(h(),y(),s.current=!1,o.invalidateQueries({queryKey:["botBuilderStatus"]}))}catch{}}),P.addEventListener("error",D=>{try{const $=JSON.parse(D.data);h(),E($.message||`Training failed: ${$.status}`),s.current=!1}catch{h(),E("Training failed unexpectedly"),s.current=!1}}),P.onerror=()=>{h(),E("Connection to server lost"),s.current=!1}},[r,h,b,y,E,o]),{mutate:x}=BYe({mutationFn:async()=>{if(!r)throw new Error("No project URL available");return whr({projectUrl:r})},onMutate:()=>{Kp.update(XV,{title:"Retraining",description:"Starting training...",type:"loading",duration:1/0,closable:!1,action:void 0})},onSuccess:O=>{O.job_id&&k(O.job_id)},onError:O=>{E(O instanceof Error?O.message:"Failed to start training"),s.current=!1}});Q.useEffect(()=>{if(!d)return;const O=n?.files_changed_since_training===!0;O&&!s.current?(s.current=!0,queueMicrotask(()=>{Kp.create({id:XV,title:"Files changed",description:"Files have changed since last training. Retrain to see the latest changes.",type:"warning",duration:1/0,closable:!0,meta:{icon:Jtn},action:{label:"Retrain",onClick:()=>x()}})})):!O&&s.current&&(s.current=!1,queueMicrotask(()=>{Kp.dismiss(XV)}))},[n?.files_changed_since_training,x,d]),Q.useEffect(()=>()=>{h()},[h])};function gQe(n,r){const o=`${window.location.origin}/login-result`,s=new URLSearchParams;if(n&&s.set(kcn,n),r)for(const[c,d]of Object.entries(r))d&&s.set(c,d);return s.toString()?`${o}?${s.toString()}`:o}function L8r(n){const r=`${window.location.origin}/logout-result`;return n?`${r}?${kcn}=${n}`:r}const Hln=({message:n,redirectParams:r})=>{const{loginWithRedirect:o}=ty(),{projectId:s}=G_(),c=d=>{o({authorizationParams:{screen_hint:d,redirect_uri:gQe(s,r)}}).catch(h=>{console.error("Login redirect failed",h)})};return H.jsxs(Vd,{align:"center",gap:4,py:2,px:6,children:[H.jsx(Fi,{size:"md",textAlign:"center",pb:4,children:n}),H.jsx(ud,{colorPalette:"dark",w:"full",onClick:()=>c(),children:"Continue to sign in"}),H.jsxs($c,{pt:2,children:[H.jsx(Fi,{size:"sm",children:"Don't have an account?"}),H.jsx(_I,{variant:"underline",onClick:()=>c("signup"),children:H.jsx(Fi,{size:"sm",children:"Create an account"})})]})]})},Gln=({onClick:n,...r})=>{const o=s=>{r.disabled?s.preventDefault():n?.(s)};return H.jsx(ud,{as:KYe,tabIndex:0,...r,onClick:o})},qln=({size:n="xl",color:r,showText:o,...s})=>{const c=o??(n==="lg"||n==="xl");return H.jsxs(kYe,{height:"100%",flexDirection:"column",...s,children:[H.jsx(kK,{"data-testid":"loading-spinner",animationDuration:"1s",color:r??"#574AE2",size:n,mb:c?"1rem":0}),c?H.jsx(Fi,{children:"Loading"}):null]})},D8r=(n,r)=>(Q.useEffect(()=>{const o=r?.current;if(!o)return;const s=new ResizeObserver(c=>{for(const d of c){const{width:h,height:b}=d.contentRect;h>0&&b>0&&n(r?.current)}});return s.observe(o),()=>{s.disconnect()}},[r,n]),r),Wg=Un.forwardRef(function(r,o){const{showArrow:s,children:c,disabled:d,portalled:h=!0,content:b,contentProps:y,portalRef:E,contentRef:k,onlyShowIfTruncated:x=!1,bgColor:O="rasawebDeepPurple.900",...P}=r,[D,$]=Q.useState(!1),[U]=pen("colors",[O]);return D8r(B=>{!B||!x||B.scrollWidth>B.clientWidth&&$(!0)},k),d||x&&k&&!D?c:H.jsxs(ver,{...P,openDelay:500,children:[H.jsx(wer,{asChild:!0,children:c}),H.jsx(lU,{disabled:!h,container:E,children:H.jsx(yer,{children:H.jsxs(Eer,{ref:k,bg:O,fontSize:"sm",px:4,py:3,borderRadius:"md",boxShadow:"lg",fontWeight:"normal",...y,children:[s&&H.jsx(Ser,{children:H.jsx(Fen,{style:{borderColor:U,backgroundColor:U}})}),b]})})})]})}),K4=n=>{const{open:r,button:o,onClose:s,onOpen:c,title:d,children:h,actions:b,type:y="div",displayOverlay:E=!0,footer:k,isLoading:x=!1,canBeClosed:O=!0,tooltipContent:P,...D}=n,$=b?.length;O||(D.closeOnEscape=!1,D.closeOnInteractOutside=!1);const U={display:"flex",flexDirection:"column",maxHeight:"100%",overflow:"hidden",flexGrow:1};return H.jsxs(cQn,{placement:"center",open:r,onOpenChange:()=>r?s():c(),...D,children:[o&&(P?H.jsx(Wg,{content:P,showArrow:!0,children:H.jsx(O7t,{asChild:!0,children:o()})}):H.jsx(O7t,{asChild:!0,children:o()})),H.jsxs(lU,{children:[E&&H.jsx(fQn,{}),H.jsx(uQn,{children:H.jsxs(lQn,{children:[H.jsx(gQn,{pt:"1rem",children:H.jsx(Kx,{size:"lg",children:d})}),x?H.jsx(Iu,{flexGrow:1,justifyContent:"center",alignItems:"center",children:H.jsx(qln,{})}):H.jsxs(ai,{as:y,css:U,children:[H.jsx(hQn,{p:"2rem",pt:"0.5rem",children:h}),$||k?H.jsxs(pQn,{children:[k,$?H.jsx(M8r,{actions:b}):null]}):null,O&&H.jsx(dQn,{asChild:!0,children:H.jsx(Gm,{size:"lg",fontSize:"lg",variant:"subtle",colorPalette:"dark","aria-label":"Close",children:H.jsx(xa,{icon:tnn})})})]})]})})]})]})},M8r=({actions:n})=>H.jsx($c,{gap:"0.5rem",flexGrow:1,justifyContent:"flex-end",children:n.map(r=>{const{id:o,title:s,onClick:c,type:d="button",linksTo:h,...b}=r,y=E=>{E.preventDefault(),c?.(E)};return h?H.jsx(Gln,{to:h,variant:b.variant,children:s},o):H.jsx(ud,{type:d,onClick:y,borderRadius:"0.5rem",...b,children:s},o)})}),mQe=({isRequired:n,onClose:r,title:o="Login to continue",message:s="Sign in or create an account to continue where you left off.",redirectParams:c})=>{const{open:d,onOpen:h,onClose:b}=_M(),{isAuthenticated:y}=ty(),E=Q.useCallback(()=>{b(),r&&r()},[b,r]);return Q.useEffect(()=>{n&&!y?h():(y||!n)&&d&&E()},[n,y,h,d,E]),y?null:H.jsx(K4,{open:d,onClose:E,onOpen:h,title:o,canBeClosed:!0,children:H.jsx(Hln,{message:s,redirectParams:c})})},j8r=()=>{const{data:n,loading:r,error:o}=IEe({interval:5e3}),c=Q.useCallback(()=>r?{text:"Checking connection to Bot Builder...",bg:"rasaNeutral.600"}:o?{text:`Connection to Bot Builder failed: ${o}`,bg:"rasaRed.800"}:n?.status==="ok"?{text:"Connection to Bot Builder is healthy",bg:"rasaGreen.800"}:{text:`Status: ${n?.status||"Unknown"} to Bot Builder`,bg:"rasaYellow.400"},[n,r,o])();return H.jsx(Wg,{content:c.text,showArrow:!0,children:H.jsx(Iu,{alignItems:"center",justifyContent:"center",width:"1.5rem",height:"1.5rem",pt:"1px",children:H.jsx(ai,{bg:c.bg,width:"0.5rem",height:"0.5rem",borderRadius:"50%"})})})},NEe=Q.forwardRef(function(r,o){const{name:s,src:c,srcSet:d,loading:h,icon:b,fallback:y,children:E,objectFit:k,...x}=r;return H.jsxs(VZn,{ref:o,...x,children:[H.jsx(XZn,{name:s,children:y||b}),H.jsx(JZn,{src:c,srcSet:d,objectFit:k,loading:h}),E]})}),$8r=()=>{const n=Q.useId(),{user:r,logout:o,loginWithRedirect:s,isLoading:c,isAuthenticated:d}=ty(),{projectId:h}=G_(),b=Io(da,E=>E.loading),y=()=>{b||o({logoutParams:{returnTo:L8r(h)}})};return!d||c?H.jsx(Wg,{content:"Log in",showArrow:!0,children:H.jsx(Gm,{rounded:"full",variant:"solid",colorPalette:"light",size:"sm",backgroundColor:"rasaNeutral.300",loading:c,disabled:b,"aria-label":"Log in",onClick:()=>void s({authorizationParams:{redirect_uri:gQe(h)}}),children:H.jsx(xa,{icon:hor})})}):H.jsxs(AYe,{positioning:{placement:"bottom-end"},ids:{trigger:n},children:[H.jsx(Wg,{content:"Log out",showArrow:!0,ids:{trigger:n},children:H.jsx(LQn,{rounded:"full",_hover:{cursor:"pointer"},"data-testid":"logged-in-avatar",children:H.jsx(NEe,{size:"xs",bg:"#D178F6",color:"rasaNeutral.50",name:r?.name||r?.email||"User"})})}),H.jsx(lU,{children:H.jsx(IYe,{children:H.jsx(OYe,{zIndex:9999,children:H.jsx(NB,{value:"logout",disabled:b,onClick:y,children:"Log out"})})})})]})},MK=()=>{const{projectId:n}=G_(),{isStudioEmbed:r,isStandalone:o}=GM();return{getPath:Q.useCallback(c=>r||o?c?`/${c}`:"/":c?`/project/${n}/${c}`:`/project/${n}`,[n,r,o])}},F8r=[{label:"Preview",icon:lor,value:q4.Chat},{label:"Code",icon:NK,value:q4.Code}],B8r=()=>{const{selectedView:n}=bU(),r=Io(da,k=>k.loading),o=Io(da,k=>k.loadingProject),s=Io(da,k=>k.isRestoring),c=Io(gc,PK),d=Io(da,k=>k.inactive),h=y2(),{getPath:b}=MK();if(r||o||c||d||s)return null;const y=F8r.map(({value:k,icon:x,label:O})=>({value:k,label:H.jsxs($c,{children:[H.jsx(xa,{icon:x}),O]})})),E={xl:{position:"absolute !important",left:"50%",translate:"-50%"}};return H.jsx(ai,{css:E,children:H.jsxs(HQn,{"data-testid":"view-control",size:"sm",value:n,onValueChange:k=>{if(k.value===q4.Chat||k.value===q4.Code){const x=b(k.value.toLowerCase()),O=h(x,{replace:!0});O instanceof Promise&&O.catch(P=>{console.error("Navigation error:",P)})}},children:[H.jsx(VQn,{}),H.jsx(KQn,{items:y})]})})},Vln=({item:n,isActive:r,onActivate:o,children:s,onClick:c,...d})=>{const h={borderColor:"rasaNeutral.300",fontWeight:"400",_hover:{bg:"#ECEFF7CC",color:"rasawebDeepPurple.800"}},b={color:"rasaNeutral.50",bg:"rasawebDeepPurple.800",borderColor:"rasawebDeepPurple.800",_hover:{bg:"rasawebDeepPurple.800",color:"rasaNeutral.50"}},y=E=>{c?.(E),o(n.value)};return H.jsx(ud,{rounded:"full",variant:"outline",colorPalette:"dark",css:r?{...h,...b}:h,type:"button",onClick:y,...d,children:s||n.label})},U8r=[{value:"1",emoji:"😞",label:"Very dissatisfied"},{value:"2",emoji:"🙁",label:"Dissatisfied"},{value:"3",emoji:"😐",label:"Neutral"},{value:"4",emoji:"😊",label:"Satisfied"},{value:"5",emoji:"😍",label:"Very satisfied"}],z8r=[{value:"yes",label:"Yes"},{value:"no",label:"No"}],H8r={size:"lg",variant:"ghost",fontSize:"xl",borderRadius:"md",border:"2px solid",transition:"all 0.15s ease-out",_active:{transform:"scale(0.95)"},css:{"@media (prefers-reduced-motion: reduce)":{transition:"none",transform:"none !important"}}},G8r=n=>({...H8r,bg:n?"rasawebDeepPurple.100":"transparent",color:n?"rasawebDeepPurple.800":"rasaNeutral.600",borderColor:n?"rasawebDeepPurple.400":"transparent",_hover:{bg:n?"rasawebDeepPurple.200":"rasaNeutral.100",transform:"scale(1.1)"}}),q8r=async n=>{const r={...n,source:"connect_popup"};await z_(H_.generalFeedback,r)},V8r=({onSuccess:n,onCancel:r})=>{const[o,s]=Q.useState(""),[c,d]=Q.useState(""),[h,b]=Q.useState("yes"),[y,E]=Q.useState(!1),{user:k}=ty(),x=!!o&&!y,O=async D=>{if(x){E(!0);try{const $={rating:Number.parseInt(o),feedback:c.trim()||void 0,allowFollowUp:h==="yes",userEmail:k?.email||"",timestamp:new Date().toISOString()};await q8r($),Kp.create({title:"🙏 Thank you for your feedback!",description:"Your input helps us improve the Hello Rasa experience.",type:"success",duration:4e3}),n?.()}catch($){const U=$ instanceof Error?$.message:"Unknown error occurred";Kp.create({title:"Failed to submit feedback",description:`Please try again or contact support if the issue persists. ${U}`,type:"error",duration:5e3})}finally{E(!1)}}},P=()=>{r?.()};return H.jsx(Sen,{children:H.jsxs(Vd,{gap:"6",align:"start",children:[H.jsxs(sie,{required:!0,children:[H.jsxs(a$e,{fontSize:"sm",fontWeight:"medium",color:"rasawebDeepPurple.800",mb:"3",htmlFor:"rating-label",children:["How would you rate your experience?",H.jsx(xQn,{})]}),H.jsx($c,{gap:"2",justify:"center",role:"radiogroup","aria-required":"true",children:U8r.map(D=>H.jsx(Gm,{"aria-label":D.label,role:"radio","aria-checked":o===D.value,id:`rating-button-${D.value}`,onClick:()=>s(D.value),...G8r(o===D.value),children:D.emoji},D.value))})]}),H.jsxs(sie,{children:[H.jsx(a$e,{fontSize:"sm",fontWeight:"medium",color:"rasawebDeepPurple.800",mb:"2",children:"Tell us in a sentence or two — what's on your mind?"}),H.jsx(DYe,{value:c,onChange:D=>d(D.target.value),placeholder:"Share your thoughts, suggestions, or any issues you've encountered...",rows:3,resize:"none",bg:"rasaNeutral.50",borderColor:"rasaNeutral.300",fontSize:"sm",color:"rasawebDeepPurple.800",_placeholder:{color:"rasaNeutral.600"},_focus:{borderColor:"rasawebPurple.800",boxShadow:"focus",bg:"white"}})]}),H.jsxs(sie,{children:[H.jsx(a$e,{htmlFor:"pill:followup-label",fontSize:"sm",fontWeight:"medium",color:"rasawebDeepPurple.800",mb:"3",children:"👋 Would you be open to a follow-up about your feedback?"}),H.jsx(NYe,{value:h,onValueChange:D=>b(D.value||"yes"),defaultValue:"yes",children:H.jsx($c,{gap:"4",children:z8r.map(D=>H.jsx(RYe,{value:String(D.value),asChild:!0,children:H.jsx(Vln,{item:{value:D.value,label:D.label},isActive:h===D.value,onActivate:b,size:"sm",id:"pill:followup-label"})},String(D.value)))})})]}),H.jsxs($c,{gap:"3",justify:"flex-end",w:"100%",pt:"4",children:[H.jsx(ud,{variant:"ghost",onClick:P,disabled:y,children:"Cancel"}),H.jsx(ud,{bg:"rasawebDeepPurple.800",color:"white",onClick:()=>void O(),loading:y,disabled:!x,_hover:{bg:"rasawebDeepPurple.900"},_disabled:{bg:"rasaNeutral.300",color:"rasaNeutral.600",cursor:"not-allowed"},children:"Share feedback"})]})]})})},W8r={sm:{minW:"72",maxW:"80"},md:{minW:"80",maxW:"100"},lg:{minW:"100",maxW:"150"}},K8r=({title:n,trigger:r,tooltipContent:o,children:s,isOpen:c,onOpenChange:d,size:h="md"})=>{const[b,y]=Q.useState(!1),E=c??b,k=P=>{d?d(P):y(P.open)},x=W8r[h],O=n.toLowerCase().replaceAll(/[^a-z0-9]+/g,"-").replaceAll(/(^-+|-+$)/g,"");return H.jsxs(MQn,{open:E,onOpenChange:k,children:[o?H.jsx(Wg,{content:o,showArrow:!0,children:H.jsx("div",{children:H.jsx(P7t,{asChild:!0,children:r})})}):H.jsx(P7t,{asChild:!0,children:r}),H.jsx(lU,{children:H.jsx(jQn,{children:H.jsxs($Qn,{borderRadius:"xl",boxShadow:"popup",bg:"rasaNeutral.50",...x,px:"6","data-testid":`popup-${O}`,css:{animation:"popupEnter 0.2s ease-out","@keyframes popupEnter":{"0%":{opacity:0,transform:"scale(0.95) translateY(0.5rem)"},"100%":{opacity:1,transform:"scale(1) translateY(0)"}}},children:[H.jsxs(ai,{py:"6",pb:"4",position:"relative",children:[H.jsx(Kx,{size:"lg",children:n}),H.jsx(BQn,{asChild:!0,children:H.jsx(Gm,{size:"lg",fontSize:"lg",variant:"subtle",colorPalette:"dark","aria-label":"Close",position:"absolute",top:"4",right:"-2",children:H.jsx(xa,{icon:tnn})})})]}),H.jsx(ai,{p:"0",pb:"6",children:s})]})})})]})},Wln=K8r,Nve=({icon:n,label:r,onClick:o,href:s,external:c=!1,showArrow:d=!0,buttonProps:h={}})=>{const b={variant:"ghost",w:"100%",h:"auto",p:"0.75rem",borderRadius:"2xl",justifyContent:"flex-start",bg:"transparent",transition:"all 0.2s ease-out",cursor:"pointer",_hover:{bg:"rasaNeutral.100",transform:"translateX(0.1rem)"},_active:{transform:"translateX(0.25rem)",bg:"rasaNeutral.200"},...h},y=H.jsxs($c,{w:"100%",justify:"space-between",align:"center",children:[H.jsxs($c,{gap:"1rem",align:"center",children:[H.jsx(ai,{bg:"rasawebDeepPurple.800",borderRadius:"full",w:"2rem",h:"2rem",display:"flex",alignItems:"center",justifyContent:"center",transition:"all 0.2s ease-out",children:H.jsx(xa,{icon:n,color:"white",fontSize:"lg"})}),H.jsx(Kx,{fontSize:"sm",fontWeight:"medium",color:"rasawebDeepPurple.800",children:r})]}),d&&H.jsx(ai,{transition:"all 0.2s ease-out",children:H.jsx(xa,{icon:nnn,color:"rasaNeutral.700",fontSize:"sm"})})]});if(s){const k=c?{target:"_blank",rel:"noopener noreferrer"}:{};return H.jsx(ud,{...b,asChild:!0,children:H.jsx("a",{href:s,...k,onClick:o,children:y})})}return H.jsx(ud,{...b,onClick:o,children:y})},Kln=({children:n,listProps:r={}})=>H.jsx(ai,{as:"ul",listStyleType:"none",m:0,p:0,...r,children:Un.Children.map(n,o=>H.jsx(ai,{as:"li",children:o}))}),Y8r={step:Qf.Loading,lastUpdated:new Date().toISOString()},qoe=()=>{const[n,r]=AEe(w5r,Y8r),o=b=>{const y={step:b,lastUpdated:new Date().toISOString()};r(y)};return{onboardingState:n,restartOnboarding:()=>{o(Qf.Step1)},updateOnboardingStatus:o,initializeIfNeeded:()=>n===null?(o(Qf.Loading),!0):!1,progressOnboarding:b=>{switch(b){case Qf.Step1:o(Qf.Step2);break;case Qf.Step2:o(Qf.Hidden);break;default:o(Qf.Finished);break}},dismissOnboarding:()=>{o(Qf.Finished)}}},X8r=`${y5r}?${new URLSearchParams(M5r).toString()}`,J8r=({children:n})=>{const[r,o]=Q.useState(!1),[s,c]=Q.useState(!1),[d]=Dye(),h=y2(),{getPath:b}=MK(),{isAuthenticated:y}=ty(),{restartOnboarding:E}=qoe();Q.useEffect(()=>{d.get(_cn)===TFt&&(o(!0),c(!1),h(b(),{replace:!0}))},[d,h,b]);const k=()=>{E(),c(!1)},x=()=>{z_(H_.visitForumClicked),c(!1)},O=()=>{o(!0),c(!1)},P=()=>{o(!1)},D=()=>{P()},$=()=>{P()},U=B=>{c(B.open)};return H.jsxs(H.Fragment,{children:[H.jsx(Wln,{isOpen:s,onOpenChange:U,title:"Help & resources",trigger:n,tooltipContent:"Help & resources",children:H.jsxs(Kln,{children:[H.jsx(Nve,{icon:Bor,label:"Show me around",onClick:k}),H.jsx(Nve,{href:X8r,icon:Eor,label:"Join the community",onClick:x,external:!0}),H.jsx(Nve,{icon:inn,label:"Give us feedback",onClick:O})]})}),H.jsx(mQe,{isRequired:r&&!y,onClose:P,message:"Please sign in so we can connect your feedback to your account — we'll only follow up if you'd like us to.",redirectParams:{do:TFt}}),y&&H.jsx(K4,{open:r,onClose:P,onOpen:()=>o(!0),title:"Let us know what you think",canBeClosed:!0,type:"div",children:H.jsx(V8r,{onSuccess:D,onCancel:$})})]})},Z8r=({getIdTokenClaims:n,projectId:r,projectUrl:o,isAuthenticated:s})=>{const[c,d]=Q.useState(void 0);return{...BYe({mutationFn:async()=>{if(!r||!s)throw new Error("Missing project or auth");const y=(await n())?.__raw;if(!y)throw new Error("Failed to get authentication token");if(!o)throw new Error("Project URL is not available. Please try again later.");return _hr({projectUrl:o,projectId:r,token:y})},onSuccess:async b=>{const y=`${r}.tar.gz`,E=globalThis.URL.createObjectURL(b),k=document.createElement("a");k.href=E,k.download=y,document.body.appendChild(k),k.click(),document.body.removeChild(k),globalThis.URL.revokeObjectURL(E),await z_(H_.projectDownloaded)},onError:b=>{d(b instanceof Error?b.message:"Unknown error")}}),downloadError:c}};function Q8r(n,r){if(n==null)return{};var o={};for(var s in n)if({}.hasOwnProperty.call(n,s)){if(r.indexOf(s)!==-1)continue;o[s]=n[s]}return o}function e9r(n,r){if(n==null)return{};var o,s,c=Q8r(n,r);if(Object.getOwnPropertySymbols){var d=Object.getOwnPropertySymbols(n);for(s=0;s<d.length;s++)o=d[s],r.indexOf(o)===-1&&{}.propertyIsEnumerable.call(n,o)&&(c[o]=n[o])}return c}function aKe(n,r){(r==null||r>n.length)&&(r=n.length);for(var o=0,s=Array(r);o<r;o++)s[o]=n[o];return s}function t9r(n){if(Array.isArray(n))return aKe(n)}function n9r(n){if(typeof Symbol<"u"&&n[Symbol.iterator]!=null||n["@@iterator"]!=null)return Array.from(n)}function r9r(n,r){if(n){if(typeof n=="string")return aKe(n,r);var o={}.toString.call(n).slice(8,-1);return o==="Object"&&n.constructor&&(o=n.constructor.name),o==="Map"||o==="Set"?Array.from(n):o==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(o)?aKe(n,r):void 0}}function i9r(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function sKe(n){return t9r(n)||n9r(n)||r9r(n)||i9r()}function foe(n){"@babel/helpers - typeof";return foe=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(r){return typeof r}:function(r){return r&&typeof Symbol=="function"&&r.constructor===Symbol&&r!==Symbol.prototype?"symbol":typeof r},foe(n)}function o9r(n,r){if(foe(n)!="object"||!n)return n;var o=n[Symbol.toPrimitive];if(o!==void 0){var s=o.call(n,r);if(foe(s)!="object")return s;throw new TypeError("@@toPrimitive must return a primitive value.")}return(r==="string"?String:Number)(n)}function a9r(n){var r=o9r(n,"string");return foe(r)=="symbol"?r:r+""}function Yln(n,r,o){return(r=a9r(r))in n?Object.defineProperty(n,r,{value:o,enumerable:!0,configurable:!0,writable:!0}):n[r]=o,n}function tUt(n,r){var o=Object.keys(n);if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(n);r&&(s=s.filter(function(c){return Object.getOwnPropertyDescriptor(n,c).enumerable})),o.push.apply(o,s)}return o}function mW(n){for(var r=1;r<arguments.length;r++){var o=arguments[r]!=null?arguments[r]:{};r%2?tUt(Object(o),!0).forEach(function(s){Yln(n,s,o[s])}):Object.getOwnPropertyDescriptors?Object.defineProperties(n,Object.getOwnPropertyDescriptors(o)):tUt(Object(o)).forEach(function(s){Object.defineProperty(n,s,Object.getOwnPropertyDescriptor(o,s))})}return n}function s9r(n){var r=n.length;if(r===0||r===1)return n;if(r===2)return[n[0],n[1],"".concat(n[0],".").concat(n[1]),"".concat(n[1],".").concat(n[0])];if(r===3)return[n[0],n[1],n[2],"".concat(n[0],".").concat(n[1]),"".concat(n[0],".").concat(n[2]),"".concat(n[1],".").concat(n[0]),"".concat(n[1],".").concat(n[2]),"".concat(n[2],".").concat(n[0]),"".concat(n[2],".").concat(n[1]),"".concat(n[0],".").concat(n[1],".").concat(n[2]),"".concat(n[0],".").concat(n[2],".").concat(n[1]),"".concat(n[1],".").concat(n[0],".").concat(n[2]),"".concat(n[1],".").concat(n[2],".").concat(n[0]),"".concat(n[2],".").concat(n[0],".").concat(n[1]),"".concat(n[2],".").concat(n[1],".").concat(n[0])];if(r>=4)return[n[0],n[1],n[2],n[3],"".concat(n[0],".").concat(n[1]),"".concat(n[0],".").concat(n[2]),"".concat(n[0],".").concat(n[3]),"".concat(n[1],".").concat(n[0]),"".concat(n[1],".").concat(n[2]),"".concat(n[1],".").concat(n[3]),"".concat(n[2],".").concat(n[0]),"".concat(n[2],".").concat(n[1]),"".concat(n[2],".").concat(n[3]),"".concat(n[3],".").concat(n[0]),"".concat(n[3],".").concat(n[1]),"".concat(n[3],".").concat(n[2]),"".concat(n[0],".").concat(n[1],".").concat(n[2]),"".concat(n[0],".").concat(n[1],".").concat(n[3]),"".concat(n[0],".").concat(n[2],".").concat(n[1]),"".concat(n[0],".").concat(n[2],".").concat(n[3]),"".concat(n[0],".").concat(n[3],".").concat(n[1]),"".concat(n[0],".").concat(n[3],".").concat(n[2]),"".concat(n[1],".").concat(n[0],".").concat(n[2]),"".concat(n[1],".").concat(n[0],".").concat(n[3]),"".concat(n[1],".").concat(n[2],".").concat(n[0]),"".concat(n[1],".").concat(n[2],".").concat(n[3]),"".concat(n[1],".").concat(n[3],".").concat(n[0]),"".concat(n[1],".").concat(n[3],".").concat(n[2]),"".concat(n[2],".").concat(n[0],".").concat(n[1]),"".concat(n[2],".").concat(n[0],".").concat(n[3]),"".concat(n[2],".").concat(n[1],".").concat(n[0]),"".concat(n[2],".").concat(n[1],".").concat(n[3]),"".concat(n[2],".").concat(n[3],".").concat(n[0]),"".concat(n[2],".").concat(n[3],".").concat(n[1]),"".concat(n[3],".").concat(n[0],".").concat(n[1]),"".concat(n[3],".").concat(n[0],".").concat(n[2]),"".concat(n[3],".").concat(n[1],".").concat(n[0]),"".concat(n[3],".").concat(n[1],".").concat(n[2]),"".concat(n[3],".").concat(n[2],".").concat(n[0]),"".concat(n[3],".").concat(n[2],".").concat(n[1]),"".concat(n[0],".").concat(n[1],".").concat(n[2],".").concat(n[3]),"".concat(n[0],".").concat(n[1],".").concat(n[3],".").concat(n[2]),"".concat(n[0],".").concat(n[2],".").concat(n[1],".").concat(n[3]),"".concat(n[0],".").concat(n[2],".").concat(n[3],".").concat(n[1]),"".concat(n[0],".").concat(n[3],".").concat(n[1],".").concat(n[2]),"".concat(n[0],".").concat(n[3],".").concat(n[2],".").concat(n[1]),"".concat(n[1],".").concat(n[0],".").concat(n[2],".").concat(n[3]),"".concat(n[1],".").concat(n[0],".").concat(n[3],".").concat(n[2]),"".concat(n[1],".").concat(n[2],".").concat(n[0],".").concat(n[3]),"".concat(n[1],".").concat(n[2],".").concat(n[3],".").concat(n[0]),"".concat(n[1],".").concat(n[3],".").concat(n[0],".").concat(n[2]),"".concat(n[1],".").concat(n[3],".").concat(n[2],".").concat(n[0]),"".concat(n[2],".").concat(n[0],".").concat(n[1],".").concat(n[3]),"".concat(n[2],".").concat(n[0],".").concat(n[3],".").concat(n[1]),"".concat(n[2],".").concat(n[1],".").concat(n[0],".").concat(n[3]),"".concat(n[2],".").concat(n[1],".").concat(n[3],".").concat(n[0]),"".concat(n[2],".").concat(n[3],".").concat(n[0],".").concat(n[1]),"".concat(n[2],".").concat(n[3],".").concat(n[1],".").concat(n[0]),"".concat(n[3],".").concat(n[0],".").concat(n[1],".").concat(n[2]),"".concat(n[3],".").concat(n[0],".").concat(n[2],".").concat(n[1]),"".concat(n[3],".").concat(n[1],".").concat(n[0],".").concat(n[2]),"".concat(n[3],".").concat(n[1],".").concat(n[2],".").concat(n[0]),"".concat(n[3],".").concat(n[2],".").concat(n[0],".").concat(n[1]),"".concat(n[3],".").concat(n[2],".").concat(n[1],".").concat(n[0])]}var yFe={};function c9r(n){if(n.length===0||n.length===1)return n;var r=n.join(".");return yFe[r]||(yFe[r]=s9r(n)),yFe[r]}function u9r(n){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},o=arguments.length>2?arguments[2]:void 0,s=n.filter(function(d){return d!=="token"}),c=c9r(s);return c.reduce(function(d,h){return mW(mW({},d),o[h])},r)}function nUt(n){return n.join(" ")}function l9r(n,r){var o=0;return function(s){return o+=1,s.map(function(c,d){return Xln({node:c,stylesheet:n,useInlineStyles:r,key:"code-segment-".concat(o,"-").concat(d)})})}}function Xln(n){var r=n.node,o=n.stylesheet,s=n.style,c=s===void 0?{}:s,d=n.useInlineStyles,h=n.key,b=r.properties,y=r.type,E=r.tagName,k=r.value;if(y==="text")return k;if(E){var x=l9r(o,d),O;if(!d)O=mW(mW({},b),{},{className:nUt(b.className)});else{var P=Object.keys(o).reduce(function(B,G){return G.split(".").forEach(function(K){B.includes(K)||B.push(K)}),B},[]),D=b.className&&b.className.includes("token")?["token"]:[],$=b.className&&D.concat(b.className.filter(function(B){return!P.includes(B)}));O=mW(mW({},b),{},{className:nUt($)||void 0,style:u9r(b.className,Object.assign({},b.style,c),o)})}var U=x(r.children);return Un.createElement(E,Dqe({key:h},O),U)}}const d9r=function(n,r){var o=n.listLanguages();return o.indexOf(r)!==-1};var f9r=["language","children","style","customStyle","codeTagProps","useInlineStyles","showLineNumbers","showInlineLineNumbers","startingLineNumber","lineNumberContainerStyle","lineNumberStyle","wrapLines","wrapLongLines","lineProps","renderer","PreTag","CodeTag","code","astGenerator"];function rUt(n,r){var o=Object.keys(n);if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(n);r&&(s=s.filter(function(c){return Object.getOwnPropertyDescriptor(n,c).enumerable})),o.push.apply(o,s)}return o}function dM(n){for(var r=1;r<arguments.length;r++){var o=arguments[r]!=null?arguments[r]:{};r%2?rUt(Object(o),!0).forEach(function(s){Yln(n,s,o[s])}):Object.getOwnPropertyDescriptors?Object.defineProperties(n,Object.getOwnPropertyDescriptors(o)):rUt(Object(o)).forEach(function(s){Object.defineProperty(n,s,Object.getOwnPropertyDescriptor(o,s))})}return n}var h9r=/\n/g;function p9r(n){return n.match(h9r)}function g9r(n){var r=n.lines,o=n.startingLineNumber,s=n.style;return r.map(function(c,d){var h=d+o;return Un.createElement("span",{key:"line-".concat(d),className:"react-syntax-highlighter-line-number",style:typeof s=="function"?s(h):s},"".concat(h,`
`))})}function m9r(n){var r=n.codeString,o=n.codeStyle,s=n.containerStyle,c=s===void 0?{float:"left",paddingRight:"10px"}:s,d=n.numberStyle,h=d===void 0?{}:d,b=n.startingLineNumber;return Un.createElement("code",{style:Object.assign({},o,c)},g9r({lines:r.replace(/\n$/,"").split(`
`),style:h,startingLineNumber:b}))}function b9r(n){return"".concat(n.toString().length,".25em")}function Jln(n,r){return{type:"element",tagName:"span",properties:{key:"line-number--".concat(n),className:["comment","linenumber","react-syntax-highlighter-line-number"],style:r},children:[{type:"text",value:n}]}}function Zln(n,r,o){var s={display:"inline-block",minWidth:b9r(o),paddingRight:"1em",textAlign:"right",userSelect:"none"},c=typeof n=="function"?n(r):n,d=dM(dM({},s),c);return d}function Rve(n){var r=n.children,o=n.lineNumber,s=n.lineNumberStyle,c=n.largestLineNumber,d=n.showInlineLineNumbers,h=n.lineProps,b=h===void 0?{}:h,y=n.className,E=y===void 0?[]:y,k=n.showLineNumbers,x=n.wrapLongLines,O=n.wrapLines,P=O===void 0?!1:O,D=P?dM({},typeof b=="function"?b(o):b):{};if(D.className=D.className?[].concat(sKe(D.className.trim().split(/\s+/)),sKe(E)):E,o&&d){var $=Zln(s,o,c);r.unshift(Jln(o,$))}return x&k&&(D.style=dM({display:"flex"},D.style)),{type:"element",tagName:"span",properties:D,children:r}}function Qln(n){for(var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:[],o=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[],s=0;s<n.length;s++){var c=n[s];if(c.type==="text")o.push(Rve({children:[c],className:sKe(new Set(r))}));else if(c.children){var d=r.concat(c.properties.className);Qln(c.children,d).forEach(function(h){return o.push(h)})}}return o}function v9r(n,r,o,s,c,d,h,b,y){var E,k=Qln(n.value),x=[],O=-1,P=0;function D(ae,oe){var he=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[];return Rve({children:ae,lineNumber:oe,lineNumberStyle:b,largestLineNumber:h,showInlineLineNumbers:c,lineProps:o,className:he,showLineNumbers:s,wrapLongLines:y,wrapLines:r})}function $(ae,oe){if(s&&oe&&c){var he=Zln(b,oe,h);ae.unshift(Jln(oe,he))}return ae}function U(ae,oe){var he=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[];return r||he.length>0?D(ae,oe,he):$(ae,oe)}for(var B=function(){var oe=k[P],he=oe.children[0].value,ye=p9r(he);if(ye){var we=he.split(`
`);we.forEach(function(ge,Te){var xe=s&&x.length+d,Ue={type:"text",value:"".concat(ge,`
`)};if(Te===0){var Z=k.slice(O+1,P).concat(Rve({children:[Ue],className:oe.properties.className})),Re=U(Z,xe);x.push(Re)}else if(Te===we.length-1){var _e=k[P+1]&&k[P+1].children&&k[P+1].children[0],qe={type:"text",value:"".concat(ge)};if(_e){var it=Rve({children:[qe],className:oe.properties.className});k.splice(P+1,0,it)}else{var Xe=[qe],de=U(Xe,xe,oe.properties.className);x.push(de)}}else{var We=[Ue],ht=U(We,xe,oe.properties.className);x.push(ht)}}),O=P}P++};P<k.length;)B();if(O!==k.length-1){var G=k.slice(O+1,k.length);if(G&&G.length){var K=s&&x.length+d,X=U(G,K);x.push(X)}}return r?x:(E=[]).concat.apply(E,x)}function w9r(n){var r=n.rows,o=n.stylesheet,s=n.useInlineStyles;return r.map(function(c,d){return Xln({node:c,stylesheet:o,useInlineStyles:s,key:"code-segement".concat(d)})})}function edn(n){return n&&typeof n.highlightAuto<"u"}function y9r(n){var r=n.astGenerator,o=n.language,s=n.code,c=n.defaultCodeValue;if(edn(r)){var d=d9r(r,o);return o==="text"?{value:c,language:"text"}:d?r.highlight(o,s):r.highlightAuto(s)}try{return o&&o!=="text"?{value:r.highlight(s,o)}:{value:c}}catch{return{value:c}}}function E9r(n,r){return function(s){var c=s.language,d=s.children,h=s.style,b=h===void 0?r:h,y=s.customStyle,E=y===void 0?{}:y,k=s.codeTagProps,x=k===void 0?{className:c?"language-".concat(c):void 0,style:dM(dM({},b['code[class*="language-"]']),b['code[class*="language-'.concat(c,'"]')])}:k,O=s.useInlineStyles,P=O===void 0?!0:O,D=s.showLineNumbers,$=D===void 0?!1:D,U=s.showInlineLineNumbers,B=U===void 0?!0:U,G=s.startingLineNumber,K=G===void 0?1:G,X=s.lineNumberContainerStyle,ae=s.lineNumberStyle,oe=ae===void 0?{}:ae,he=s.wrapLines,ye=s.wrapLongLines,we=ye===void 0?!1:ye,ge=s.lineProps,Te=ge===void 0?{}:ge,xe=s.renderer,Ue=s.PreTag,Z=Ue===void 0?"pre":Ue,Re=s.CodeTag,_e=Re===void 0?"code":Re,qe=s.code,it=qe===void 0?(Array.isArray(d)?d[0]:d)||"":qe,Xe=s.astGenerator,de=e9r(s,f9r);Xe=Xe||n;var We=$?Un.createElement(m9r,{containerStyle:X,codeStyle:x.style||{},numberStyle:oe,startingLineNumber:K,codeString:it}):null,ht=b.hljs||b['pre[class*="language-"]']||{backgroundColor:"#fff"},Ie=edn(Xe)?"hljs":"prismjs",$t=P?Object.assign({},de,{style:Object.assign({},ht,E)}):Object.assign({},de,{className:de.className?"".concat(Ie," ").concat(de.className):Ie,style:Object.assign({},E)});if(we?x.style=dM({whiteSpace:"pre-wrap"},x.style):x.style=dM({whiteSpace:"pre"},x.style),!Xe)return Un.createElement(Z,$t,We,Un.createElement(_e,x,it));(he===void 0&&xe||we)&&(he=!0),xe=xe||w9r;var An=[{type:"text",value:it}],Yt=y9r({astGenerator:Xe,language:c,code:it,defaultCodeValue:An});Yt.language===null&&(Yt.value=An);var Or=Yt.value.length;Or===1&&Yt.value[0].type==="text"&&(Or=Yt.value[0].value.split(`
`).length);var Nt=Or+K,on=v9r(Yt,he,Te,$,B,K,Nt,oe,we);return Un.createElement(Z,$t,Un.createElement(_e,x,!B&&We,xe({rows:on,stylesheet:b,useInlineStyles:P})))}}const S9r=["abap","abnf","actionscript","ada","agda","al","antlr4","apacheconf","apex","apl","applescript","aql","arduino","arff","asciidoc","asm6502","asmatmel","aspnet","autohotkey","autoit","avisynth","avro-idl","bash","basic","batch","bbcode","bicep","birb","bison","bnf","brainfuck","brightscript","bro","bsl","c","cfscript","chaiscript","cil","clike","clojure","cmake","cobol","coffeescript","concurnas","coq","cpp","crystal","csharp","cshtml","csp","css-extras","css","csv","cypher","d","dart","dataweave","dax","dhall","diff","django","dns-zone-file","docker","dot","ebnf","editorconfig","eiffel","ejs","elixir","elm","erb","erlang","etlua","excel-formula","factor","false","firestore-security-rules","flow","fortran","fsharp","ftl","gap","gcode","gdscript","gedcom","gherkin","git","glsl","gml","gn","go-module","go","graphql","groovy","haml","handlebars","haskell","haxe","hcl","hlsl","hoon","hpkp","hsts","http","ichigojam","icon","icu-message-format","idris","iecst","ignore","inform7","ini","io","j","java","javadoc","javadoclike","javascript","javastacktrace","jexl","jolie","jq","js-extras","js-templates","jsdoc","json","json5","jsonp","jsstacktrace","jsx","julia","keepalived","keyman","kotlin","kumir","kusto","latex","latte","less","lilypond","liquid","lisp","livescript","llvm","log","lolcode","lua","magma","makefile","markdown","markup-templating","markup","matlab","maxscript","mel","mermaid","mizar","mongodb","monkey","moonscript","n1ql","n4js","nand2tetris-hdl","naniscript","nasm","neon","nevod","nginx","nim","nix","nsis","objectivec","ocaml","opencl","openqasm","oz","parigp","parser","pascal","pascaligo","pcaxis","peoplecode","perl","php-extras","php","phpdoc","plsql","powerquery","powershell","processing","prolog","promql","properties","protobuf","psl","pug","puppet","pure","purebasic","purescript","python","q","qml","qore","qsharp","r","racket","reason","regex","rego","renpy","rest","rip","roboconf","robotframework","ruby","rust","sas","sass","scala","scheme","scss","shell-session","smali","smalltalk","smarty","sml","solidity","solution-file","soy","sparql","splunk-spl","sqf","sql","squirrel","stan","stylus","swift","systemd","t4-cs","t4-templating","t4-vb","tap","tcl","textile","toml","tremor","tsx","tt2","turtle","twig","typescript","typoscript","unrealscript","uorazor","uri","v","vala","vbnet","velocity","verilog","vhdl","vim","visual-basic","warpscript","wasm","web-idl","wiki","wolfram","wren","xeora","xml-doc","xojo","xquery","yaml","yang","zig"];var EFe,iUt;function _9r(){if(iUt)return EFe;iUt=1,EFe=r;var n=Object.prototype.hasOwnProperty;function r(){for(var o={},s=0;s<arguments.length;s++){var c=arguments[s];for(var d in c)n.call(c,d)&&(o[d]=c[d])}return o}return EFe}var SFe,oUt;function tdn(){if(oUt)return SFe;oUt=1,SFe=r;var n=r.prototype;n.space=null,n.normal={},n.property={};function r(o,s,c){this.property=o,this.normal=s,c&&(this.space=c)}return SFe}var _Fe,aUt;function k9r(){if(aUt)return _Fe;aUt=1;var n=_9r(),r=tdn();_Fe=o;function o(s){for(var c=s.length,d=[],h=[],b=-1,y,E;++b<c;)y=s[b],d.push(y.property),h.push(y.normal),E=y.space;return new r(n.apply(null,d),n.apply(null,h),E)}return _Fe}var kFe,sUt;function bQe(){if(sUt)return kFe;sUt=1,kFe=n;function n(r){return r.toLowerCase()}return kFe}var TFe,cUt;function ndn(){if(cUt)return TFe;cUt=1,TFe=r;var n=r.prototype;n.space=null,n.attribute=null,n.property=null,n.boolean=!1,n.booleanish=!1,n.overloadedBoolean=!1,n.number=!1,n.commaSeparated=!1,n.spaceSeparated=!1,n.commaOrSpaceSeparated=!1,n.mustUseProperty=!1,n.defined=!1;function r(o,s){this.property=o,this.attribute=s}return TFe}var S4={},uUt;function vQe(){if(uUt)return S4;uUt=1;var n=0;S4.boolean=r(),S4.booleanish=r(),S4.overloadedBoolean=r(),S4.number=r(),S4.spaceSeparated=r(),S4.commaSeparated=r(),S4.commaOrSpaceSeparated=r();function r(){return Math.pow(2,++n)}return S4}var xFe,lUt;function rdn(){if(lUt)return xFe;lUt=1;var n=ndn(),r=vQe();xFe=c,c.prototype=new n,c.prototype.defined=!0;var o=["boolean","booleanish","overloadedBoolean","number","commaSeparated","spaceSeparated","commaOrSpaceSeparated"],s=o.length;function c(h,b,y,E){var k=-1,x;for(d(this,"space",E),n.call(this,h,b);++k<s;)x=o[k],d(this,x,(y&r[x])===r[x])}function d(h,b,y){y&&(h[b]=y)}return xFe}var CFe,dUt;function Voe(){if(dUt)return CFe;dUt=1;var n=bQe(),r=tdn(),o=rdn();CFe=s;function s(c){var d=c.space,h=c.mustUseProperty||[],b=c.attributes||{},y=c.properties,E=c.transform,k={},x={},O,P;for(O in y)P=new o(O,E(b,O),y[O],d),h.indexOf(O)!==-1&&(P.mustUseProperty=!0),k[O]=P,x[n(O)]=O,x[n(P.attribute)]=O;return new r(k,x,d)}return CFe}var AFe,fUt;function T9r(){if(fUt)return AFe;fUt=1;var n=Voe();AFe=n({space:"xlink",transform:r,properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null}});function r(o,s){return"xlink:"+s.slice(5).toLowerCase()}return AFe}var IFe,hUt;function x9r(){if(hUt)return IFe;hUt=1;var n=Voe();IFe=n({space:"xml",transform:r,properties:{xmlLang:null,xmlBase:null,xmlSpace:null}});function r(o,s){return"xml:"+s.slice(3).toLowerCase()}return IFe}var OFe,pUt;function C9r(){if(pUt)return OFe;pUt=1,OFe=n;function n(r,o){return o in r?r[o]:o}return OFe}var NFe,gUt;function idn(){if(gUt)return NFe;gUt=1;var n=C9r();NFe=r;function r(o,s){return n(o,s.toLowerCase())}return NFe}var RFe,mUt;function A9r(){if(mUt)return RFe;mUt=1;var n=Voe(),r=idn();return RFe=n({space:"xmlns",attributes:{xmlnsxlink:"xmlns:xlink"},transform:r,properties:{xmlns:null,xmlnsXLink:null}}),RFe}var PFe,bUt;function I9r(){if(bUt)return PFe;bUt=1;var n=vQe(),r=Voe(),o=n.booleanish,s=n.number,c=n.spaceSeparated;PFe=r({transform:d,properties:{ariaActiveDescendant:null,ariaAtomic:o,ariaAutoComplete:null,ariaBusy:o,ariaChecked:o,ariaColCount:s,ariaColIndex:s,ariaColSpan:s,ariaControls:c,ariaCurrent:null,ariaDescribedBy:c,ariaDetails:null,ariaDisabled:o,ariaDropEffect:c,ariaErrorMessage:null,ariaExpanded:o,ariaFlowTo:c,ariaGrabbed:o,ariaHasPopup:null,ariaHidden:o,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:c,ariaLevel:s,ariaLive:null,ariaModal:o,ariaMultiLine:o,ariaMultiSelectable:o,ariaOrientation:null,ariaOwns:c,ariaPlaceholder:null,ariaPosInSet:s,ariaPressed:o,ariaReadOnly:o,ariaRelevant:null,ariaRequired:o,ariaRoleDescription:c,ariaRowCount:s,ariaRowIndex:s,ariaRowSpan:s,ariaSelected:o,ariaSetSize:s,ariaSort:null,ariaValueMax:s,ariaValueMin:s,ariaValueNow:s,ariaValueText:null,role:null}});function d(h,b){return b==="role"?b:"aria-"+b.slice(4).toLowerCase()}return PFe}var LFe,vUt;function O9r(){if(vUt)return LFe;vUt=1;var n=vQe(),r=Voe(),o=idn(),s=n.boolean,c=n.overloadedBoolean,d=n.booleanish,h=n.number,b=n.spaceSeparated,y=n.commaSeparated;return LFe=r({space:"html",attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},transform:o,mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:y,acceptCharset:b,accessKey:b,action:null,allow:null,allowFullScreen:s,allowPaymentRequest:s,allowUserMedia:s,alt:null,as:null,async:s,autoCapitalize:null,autoComplete:b,autoFocus:s,autoPlay:s,capture:s,charSet:null,checked:s,cite:null,className:b,cols:h,colSpan:null,content:null,contentEditable:d,controls:s,controlsList:b,coords:h|y,crossOrigin:null,data:null,dateTime:null,decoding:null,default:s,defer:s,dir:null,dirName:null,disabled:s,download:c,draggable:d,encType:null,enterKeyHint:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:s,formTarget:null,headers:b,height:h,hidden:s,high:h,href:null,hrefLang:null,htmlFor:b,httpEquiv:b,id:null,imageSizes:null,imageSrcSet:y,inputMode:null,integrity:null,is:null,isMap:s,itemId:null,itemProp:b,itemRef:b,itemScope:s,itemType:b,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:s,low:h,manifest:null,max:null,maxLength:h,media:null,method:null,min:null,minLength:h,multiple:s,muted:s,name:null,nonce:null,noModule:s,noValidate:s,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforePrint:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextMenu:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:s,optimum:h,pattern:null,ping:b,placeholder:null,playsInline:s,poster:null,preload:null,readOnly:s,referrerPolicy:null,rel:b,required:s,reversed:s,rows:h,rowSpan:h,sandbox:b,scope:null,scoped:s,seamless:s,selected:s,shape:null,size:h,sizes:null,slot:null,span:h,spellCheck:d,src:null,srcDoc:null,srcLang:null,srcSet:y,start:h,step:null,style:null,tabIndex:h,target:null,title:null,translate:null,type:null,typeMustMatch:s,useMap:null,value:d,width:h,wrap:null,align:null,aLink:null,archive:b,axis:null,background:null,bgColor:null,border:h,borderColor:null,bottomMargin:h,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:s,declare:s,event:null,face:null,frame:null,frameBorder:null,hSpace:h,leftMargin:h,link:null,longDesc:null,lowSrc:null,marginHeight:h,marginWidth:h,noResize:s,noHref:s,noShade:s,noWrap:s,object:null,profile:null,prompt:null,rev:null,rightMargin:h,rules:null,scheme:null,scrolling:d,standby:null,summary:null,text:null,topMargin:h,valueType:null,version:null,vAlign:null,vLink:null,vSpace:h,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:s,disableRemotePlayback:s,prefix:null,property:null,results:h,security:null,unselectable:null}}),LFe}var DFe,wUt;function N9r(){if(wUt)return DFe;wUt=1;var n=k9r(),r=T9r(),o=x9r(),s=A9r(),c=I9r(),d=O9r();return DFe=n([o,r,s,c,d]),DFe}var MFe,yUt;function R9r(){if(yUt)return MFe;yUt=1;var n=bQe(),r=rdn(),o=ndn(),s="data";MFe=b;var c=/^data[-\w.:]+$/i,d=/-[a-z]/g,h=/[A-Z]/g;function b(O,P){var D=n(P),$=P,U=o;return D in O.normal?O.property[O.normal[D]]:(D.length>4&&D.slice(0,4)===s&&c.test(P)&&(P.charAt(4)==="-"?$=y(P):P=E(P),U=r),new U($,P))}function y(O){var P=O.slice(5).replace(d,x);return s+P.charAt(0).toUpperCase()+P.slice(1)}function E(O){var P=O.slice(4);return d.test(P)?O:(P=P.replace(h,k),P.charAt(0)!=="-"&&(P="-"+P),s+P)}function k(O){return"-"+O.toLowerCase()}function x(O){return O.charAt(1).toUpperCase()}return MFe}var jFe,EUt;function P9r(){if(EUt)return jFe;EUt=1,jFe=r;var n=/[#.]/g;function r(o,s){for(var c=o||"",d=s||"div",h={},b=0,y,E,k;b<c.length;)n.lastIndex=b,k=n.exec(c),y=c.slice(b,k?k.index:c.length),y&&(E?E==="#"?h.id=y:h.className?h.className.push(y):h.className=[y]:d=y,b+=y.length),k&&(E=k[0],b++);return{type:"element",tagName:d,properties:h,children:[]}}return jFe}var F0e={},SUt;function L9r(){if(SUt)return F0e;SUt=1,F0e.parse=s,F0e.stringify=c;var n="",r=" ",o=/[ \t\n\r\f]+/g;function s(d){var h=String(d||n).trim();return h===n?[]:h.split(o)}function c(d){return d.join(r).trim()}return F0e}var B0e={},_Ut;function D9r(){if(_Ut)return B0e;_Ut=1,B0e.parse=s,B0e.stringify=c;var n=",",r=" ",o="";function s(d){for(var h=[],b=String(d||o),y=b.indexOf(n),E=0,k=!1,x;!k;)y===-1&&(y=b.length,k=!0),x=b.slice(E,y).trim(),(x||!k)&&h.push(x),E=y+1,y=b.indexOf(n,E);return h}function c(d,h){var b=h||{},y=b.padLeft===!1?o:r,E=b.padRight?r:o;return d[d.length-1]===o&&(d=d.concat(o)),d.join(E+n+y).trim()}return B0e}var $Fe,kUt;function M9r(){if(kUt)return $Fe;kUt=1;var n=R9r(),r=bQe(),o=P9r(),s=L9r().parse,c=D9r().parse;$Fe=h;var d={}.hasOwnProperty;function h(D,$,U){var B=U?P(U):null;return G;function G(X,ae){var oe=o(X,$),he=Array.prototype.slice.call(arguments,2),ye=oe.tagName.toLowerCase(),we;if(oe.tagName=B&&d.call(B,ye)?B[ye]:ye,ae&&b(ae,oe)&&(he.unshift(ae),ae=null),ae)for(we in ae)K(oe.properties,we,ae[we]);return E(oe.children,he),oe.tagName==="template"&&(oe.content={type:"root",children:oe.children},oe.children=[]),oe}function K(X,ae,oe){var he,ye,we;oe==null||oe!==oe||(he=n(D,ae),ye=he.property,we=oe,typeof we=="string"&&(he.spaceSeparated?we=s(we):he.commaSeparated?we=c(we):he.commaOrSpaceSeparated&&(we=s(c(we).join(" ")))),ye==="style"&&typeof oe!="string"&&(we=O(we)),ye==="className"&&X.className&&(we=X.className.concat(we)),X[ye]=k(he,ye,we))}}function b(D,$){return typeof D=="string"||"length"in D||y($.tagName,D)}function y(D,$){var U=$.type;return D==="input"||!U||typeof U!="string"?!1:typeof $.children=="object"&&"length"in $.children?!0:(U=U.toLowerCase(),D==="button"?U!=="menu"&&U!=="submit"&&U!=="reset"&&U!=="button":"value"in $)}function E(D,$){var U,B;if(typeof $=="string"||typeof $=="number"){D.push({type:"text",value:String($)});return}if(typeof $=="object"&&"length"in $){for(U=-1,B=$.length;++U<B;)E(D,$[U]);return}if(typeof $!="object"||!("type"in $))throw new Error("Expected node, nodes, or string, got `"+$+"`");D.push($)}function k(D,$,U){var B,G,K;if(typeof U!="object"||!("length"in U))return x(D,$,U);for(G=U.length,B=-1,K=[];++B<G;)K[B]=x(D,$,U[B]);return K}function x(D,$,U){var B=U;return D.number||D.positiveNumber?!isNaN(B)&&B!==""&&(B=Number(B)):(D.boolean||D.overloadedBoolean)&&typeof B=="string"&&(B===""||r(U)===r($))&&(B=!0),B}function O(D){var $=[],U;for(U in D)$.push([U,D[U]].join(": "));return $.join("; ")}function P(D){for(var $=D.length,U=-1,B={},G;++U<$;)G=D[U],B[G.toLowerCase()]=G;return B}return $Fe}var FFe,TUt;function j9r(){if(TUt)return FFe;TUt=1;var n=N9r(),r=M9r(),o=r(n,"div");return o.displayName="html",FFe=o,FFe}var BFe,xUt;function $9r(){return xUt||(xUt=1,BFe=j9r()),BFe}const F9r="Æ",B9r="&",U9r="Á",z9r="Â",H9r="À",G9r="Å",q9r="Ã",V9r="Ä",W9r="©",K9r="Ç",Y9r="Ð",X9r="É",J9r="Ê",Z9r="È",Q9r="Ë",e7r=">",t7r="Í",n7r="Î",r7r="Ì",i7r="Ï",o7r="<",a7r="Ñ",s7r="Ó",c7r="Ô",u7r="Ò",l7r="Ø",d7r="Õ",f7r="Ö",h7r='"',p7r="®",g7r="Þ",m7r="Ú",b7r="Û",v7r="Ù",w7r="Ü",y7r="Ý",E7r="á",S7r="â",_7r="´",k7r="æ",T7r="à",x7r="&",C7r="å",A7r="ã",I7r="ä",O7r="¦",N7r="ç",R7r="¸",P7r="¢",L7r="©",D7r="¤",M7r="°",j7r="÷",$7r="é",F7r="ê",B7r="è",U7r="ð",z7r="ë",H7r="½",G7r="¼",q7r="¾",V7r=">",W7r="í",K7r="î",Y7r="¡",X7r="ì",J7r="¿",Z7r="ï",Q7r="«",ejr="<",tjr="¯",njr="µ",rjr="·",ijr=" ",ojr="¬",ajr="ñ",sjr="ó",cjr="ô",ujr="ò",ljr="ª",djr="º",fjr="ø",hjr="õ",pjr="ö",gjr="¶",mjr="±",bjr="£",vjr='"',wjr="»",yjr="®",Ejr="§",Sjr="­",_jr="¹",kjr="²",Tjr="³",xjr="ß",Cjr="þ",Ajr="×",Ijr="ú",Ojr="û",Njr="ù",Rjr="¨",Pjr="ü",Ljr="ý",Djr="¥",Mjr="ÿ",jjr={AElig:F9r,AMP:B9r,Aacute:U9r,Acirc:z9r,Agrave:H9r,Aring:G9r,Atilde:q9r,Auml:V9r,COPY:W9r,Ccedil:K9r,ETH:Y9r,Eacute:X9r,Ecirc:J9r,Egrave:Z9r,Euml:Q9r,GT:e7r,Iacute:t7r,Icirc:n7r,Igrave:r7r,Iuml:i7r,LT:o7r,Ntilde:a7r,Oacute:s7r,Ocirc:c7r,Ograve:u7r,Oslash:l7r,Otilde:d7r,Ouml:f7r,QUOT:h7r,REG:p7r,THORN:g7r,Uacute:m7r,Ucirc:b7r,Ugrave:v7r,Uuml:w7r,Yacute:y7r,aacute:E7r,acirc:S7r,acute:_7r,aelig:k7r,agrave:T7r,amp:x7r,aring:C7r,atilde:A7r,auml:I7r,brvbar:O7r,ccedil:N7r,cedil:R7r,cent:P7r,copy:L7r,curren:D7r,deg:M7r,divide:j7r,eacute:$7r,ecirc:F7r,egrave:B7r,eth:U7r,euml:z7r,frac12:H7r,frac14:G7r,frac34:q7r,gt:V7r,iacute:W7r,icirc:K7r,iexcl:Y7r,igrave:X7r,iquest:J7r,iuml:Z7r,laquo:Q7r,lt:ejr,macr:tjr,micro:njr,middot:rjr,nbsp:ijr,not:ojr,ntilde:ajr,oacute:sjr,ocirc:cjr,ograve:ujr,ordf:ljr,ordm:djr,oslash:fjr,otilde:hjr,ouml:pjr,para:gjr,plusmn:mjr,pound:bjr,quot:vjr,raquo:wjr,reg:yjr,sect:Ejr,shy:Sjr,sup1:_jr,sup2:kjr,sup3:Tjr,szlig:xjr,thorn:Cjr,times:Ajr,uacute:Ijr,ucirc:Ojr,ugrave:Njr,uml:Rjr,uuml:Pjr,yacute:Ljr,yen:Djr,yuml:Mjr},$jr={0:"�",128:"€",130:"‚",131:"ƒ",132:"„",133:"…",134:"†",135:"‡",136:"ˆ",137:"‰",138:"Š",139:"‹",140:"Œ",142:"Ž",145:"‘",146:"’",147:"“",148:"”",149:"•",150:"–",151:"—",152:"˜",153:"™",154:"š",155:"›",156:"œ",158:"ž",159:"Ÿ"};var UFe,CUt;function odn(){if(CUt)return UFe;CUt=1,UFe=n;function n(r){var o=typeof r=="string"?r.charCodeAt(0):r;return o>=48&&o<=57}return UFe}var zFe,AUt;function Fjr(){if(AUt)return zFe;AUt=1,zFe=n;function n(r){var o=typeof r=="string"?r.charCodeAt(0):r;return o>=97&&o<=102||o>=65&&o<=70||o>=48&&o<=57}return zFe}var HFe,IUt;function Bjr(){if(IUt)return HFe;IUt=1,HFe=n;function n(r){var o=typeof r=="string"?r.charCodeAt(0):r;return o>=97&&o<=122||o>=65&&o<=90}return HFe}var GFe,OUt;function Ujr(){if(OUt)return GFe;OUt=1;var n=Bjr(),r=odn();GFe=o;function o(s){return n(s)||r(s)}return GFe}var qFe,NUt;function zjr(){if(NUt)return qFe;NUt=1;var n,r=59;qFe=o;function o(s){var c="&"+s+";",d;return n=n||document.createElement("i"),n.innerHTML=c,d=n.textContent,d.charCodeAt(d.length-1)===r&&s!=="semi"||d===c?!1:d}return qFe}var VFe,RUt;function Hjr(){if(RUt)return VFe;RUt=1;var n=jjr,r=$jr,o=odn(),s=Fjr(),c=Ujr(),d=zjr();VFe=Xe;var h={}.hasOwnProperty,b=String.fromCharCode,y=Function.prototype,E={warning:null,reference:null,text:null,warningContext:null,referenceContext:null,textContext:null,position:{},additional:null,attribute:!1,nonTerminated:!0},k=9,x=10,O=12,P=32,D=38,$=59,U=60,B=61,G=35,K=88,X=120,ae=65533,oe="named",he="hexadecimal",ye="decimal",we={};we[he]=16,we[ye]=10;var ge={};ge[oe]=c,ge[ye]=o,ge[he]=s;var Te=1,xe=2,Ue=3,Z=4,Re=5,_e=6,qe=7,it={};it[Te]="Named character references must be terminated by a semicolon",it[xe]="Numeric character references must be terminated by a semicolon",it[Ue]="Named character references cannot be empty",it[Z]="Numeric character references cannot be empty",it[Re]="Named character references must be known",it[_e]="Numeric character references cannot be disallowed",it[qe]="Numeric character references cannot be outside the permissible Unicode range";function Xe(Ie,$t){var An={},Yt,Or;$t||($t={});for(Or in E)Yt=$t[Or],An[Or]=Yt??E[Or];return(An.position.indent||An.position.start)&&(An.indent=An.position.indent||[],An.position=An.position.start),de(Ie,An)}function de(Ie,$t){var An=$t.additional,Yt=$t.nonTerminated,Or=$t.text,Nt=$t.reference,on=$t.warning,br=$t.textContext,wi=$t.referenceContext,no=$t.warningContext,Ri=$t.position,ro=$t.indent||[],Ko=Ie.length,Ca=0,qi=-1,yi=Ri.column||1,lo=Ri.line||1,Eo="",Fc=[],Yo,zl,ld,Yc,Sn,Pn,Kr,Ei,Sa,bs,Cc,Ac,Xc,rc,js,Jc,Id,nh,hu;for(typeof An=="string"&&(An=An.charCodeAt(0)),Jc=Qp(),Ei=on?Kb:y,Ca--,Ko++;++Ca<Ko;)if(Sn===x&&(yi=ro[qi]||1),Sn=Ie.charCodeAt(Ca),Sn===D){if(Kr=Ie.charCodeAt(Ca+1),Kr===k||Kr===x||Kr===O||Kr===P||Kr===D||Kr===U||Kr!==Kr||An&&Kr===An){Eo+=b(Sn),yi++;continue}for(Xc=Ca+1,Ac=Xc,hu=Xc,Kr===G?(hu=++Ac,Kr=Ie.charCodeAt(hu),Kr===K||Kr===X?(rc=he,hu=++Ac):rc=ye):rc=oe,Yo="",Cc="",Yc="",js=ge[rc],hu--;++hu<Ko&&(Kr=Ie.charCodeAt(hu),!!js(Kr));)Yc+=b(Kr),rc===oe&&h.call(n,Yc)&&(Yo=Yc,Cc=n[Yc]);ld=Ie.charCodeAt(hu)===$,ld&&(hu++,zl=rc===oe?d(Yc):!1,zl&&(Yo=Yc,Cc=zl)),nh=1+hu-Xc,!ld&&!Yt||(Yc?rc===oe?(ld&&!Cc?Ei(Re,1):(Yo!==Yc&&(hu=Ac+Yo.length,nh=1+hu-Ac,ld=!1),ld||(Sa=Yo?Te:Ue,$t.attribute?(Kr=Ie.charCodeAt(hu),Kr===B?(Ei(Sa,nh),Cc=null):c(Kr)?Cc=null:Ei(Sa,nh)):Ei(Sa,nh))),Pn=Cc):(ld||Ei(xe,nh),Pn=parseInt(Yc,we[rc]),We(Pn)?(Ei(qe,nh),Pn=b(ae)):Pn in r?(Ei(_e,nh),Pn=r[Pn]):(bs="",ht(Pn)&&Ei(_e,nh),Pn>65535&&(Pn-=65536,bs+=b(Pn>>>10|55296),Pn=56320|Pn&1023),Pn=bs+b(Pn))):rc!==oe&&Ei(Z,nh)),Pn?(Km(),Jc=Qp(),Ca=hu-1,yi+=hu-Xc+1,Fc.push(Pn),Id=Qp(),Id.offset++,Nt&&Nt.call(wi,Pn,{start:Jc,end:Id},Ie.slice(Xc-1,hu)),Jc=Id):(Yc=Ie.slice(Xc-1,hu),Eo+=Yc,yi+=Yc.length,Ca=hu-1)}else Sn===10&&(lo++,qi++,yi=0),Sn===Sn?(Eo+=b(Sn),yi++):Km();return Fc.join("");function Qp(){return{line:lo,column:yi,offset:Ca+(Ri.offset||0)}}function Kb(S2,_2){var dv=Qp();dv.column+=_2,dv.offset+=_2,on.call(no,it[S2],dv,S2)}function Km(){Eo&&(Fc.push(Eo),Or&&Or.call(br,Eo,{start:Jc,end:Qp()}),Eo="")}}function We(Ie){return Ie>=55296&&Ie<=57343||Ie>1114111}function ht(Ie){return Ie>=1&&Ie<=8||Ie===11||Ie>=13&&Ie<=31||Ie>=127&&Ie<=159||Ie>=64976&&Ie<=65007||(Ie&65535)===65535||(Ie&65535)===65534}return VFe}var WFe={exports:{}},PUt;function Gjr(){return PUt||(PUt=1,function(n){var r=typeof window<"u"?window:typeof WorkerGlobalScope<"u"&&self instanceof WorkerGlobalScope?self:{};/**
 * Prism: Lightweight, robust, elegant syntax highlighting
 *
 * @license MIT <https://opensource.org/licenses/MIT>
 * @author Lea Verou <https://lea.verou.me>
 * @namespace
 * @public
 */var o=function(s){var c=/(?:^|\s)lang(?:uage)?-([\w-]+)(?=\s|$)/i,d=0,h={},b={manual:s.Prism&&s.Prism.manual,disableWorkerMessageHandler:s.Prism&&s.Prism.disableWorkerMessageHandler,util:{encode:function G(K){return K instanceof y?new y(K.type,G(K.content),K.alias):Array.isArray(K)?K.map(G):K.replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/\u00a0/g," ")},type:function(G){return Object.prototype.toString.call(G).slice(8,-1)},objId:function(G){return G.__id||Object.defineProperty(G,"__id",{value:++d}),G.__id},clone:function G(K,X){X=X||{};var ae,oe;switch(b.util.type(K)){case"Object":if(oe=b.util.objId(K),X[oe])return X[oe];ae={},X[oe]=ae;for(var he in K)K.hasOwnProperty(he)&&(ae[he]=G(K[he],X));return ae;case"Array":return oe=b.util.objId(K),X[oe]?X[oe]:(ae=[],X[oe]=ae,K.forEach(function(ye,we){ae[we]=G(ye,X)}),ae);default:return K}},getLanguage:function(G){for(;G;){var K=c.exec(G.className);if(K)return K[1].toLowerCase();G=G.parentElement}return"none"},setLanguage:function(G,K){G.className=G.className.replace(RegExp(c,"gi"),""),G.classList.add("language-"+K)},currentScript:function(){if(typeof document>"u")return null;if("currentScript"in document)return document.currentScript;try{throw new Error}catch(ae){var G=(/at [^(\r\n]*\((.*):[^:]+:[^:]+\)$/i.exec(ae.stack)||[])[1];if(G){var K=document.getElementsByTagName("script");for(var X in K)if(K[X].src==G)return K[X]}return null}},isActive:function(G,K,X){for(var ae="no-"+K;G;){var oe=G.classList;if(oe.contains(K))return!0;if(oe.contains(ae))return!1;G=G.parentElement}return!!X}},languages:{plain:h,plaintext:h,text:h,txt:h,extend:function(G,K){var X=b.util.clone(b.languages[G]);for(var ae in K)X[ae]=K[ae];return X},insertBefore:function(G,K,X,ae){ae=ae||b.languages;var oe=ae[G],he={};for(var ye in oe)if(oe.hasOwnProperty(ye)){if(ye==K)for(var we in X)X.hasOwnProperty(we)&&(he[we]=X[we]);X.hasOwnProperty(ye)||(he[ye]=oe[ye])}var ge=ae[G];return ae[G]=he,b.languages.DFS(b.languages,function(Te,xe){xe===ge&&Te!=G&&(this[Te]=he)}),he},DFS:function G(K,X,ae,oe){oe=oe||{};var he=b.util.objId;for(var ye in K)if(K.hasOwnProperty(ye)){X.call(K,ye,K[ye],ae||ye);var we=K[ye],ge=b.util.type(we);ge==="Object"&&!oe[he(we)]?(oe[he(we)]=!0,G(we,X,null,oe)):ge==="Array"&&!oe[he(we)]&&(oe[he(we)]=!0,G(we,X,ye,oe))}}},plugins:{},highlightAll:function(G,K){b.highlightAllUnder(document,G,K)},highlightAllUnder:function(G,K,X){var ae={callback:X,container:G,selector:'code[class*="language-"], [class*="language-"] code, code[class*="lang-"], [class*="lang-"] code'};b.hooks.run("before-highlightall",ae),ae.elements=Array.prototype.slice.apply(ae.container.querySelectorAll(ae.selector)),b.hooks.run("before-all-elements-highlight",ae);for(var oe=0,he;he=ae.elements[oe++];)b.highlightElement(he,K===!0,ae.callback)},highlightElement:function(G,K,X){var ae=b.util.getLanguage(G),oe=b.languages[ae];b.util.setLanguage(G,ae);var he=G.parentElement;he&&he.nodeName.toLowerCase()==="pre"&&b.util.setLanguage(he,ae);var ye=G.textContent,we={element:G,language:ae,grammar:oe,code:ye};function ge(xe){we.highlightedCode=xe,b.hooks.run("before-insert",we),we.element.innerHTML=we.highlightedCode,b.hooks.run("after-highlight",we),b.hooks.run("complete",we),X&&X.call(we.element)}if(b.hooks.run("before-sanity-check",we),he=we.element.parentElement,he&&he.nodeName.toLowerCase()==="pre"&&!he.hasAttribute("tabindex")&&he.setAttribute("tabindex","0"),!we.code){b.hooks.run("complete",we),X&&X.call(we.element);return}if(b.hooks.run("before-highlight",we),!we.grammar){ge(b.util.encode(we.code));return}if(K&&s.Worker){var Te=new Worker(b.filename);Te.onmessage=function(xe){ge(xe.data)},Te.postMessage(JSON.stringify({language:we.language,code:we.code,immediateClose:!0}))}else ge(b.highlight(we.code,we.grammar,we.language))},highlight:function(G,K,X){var ae={code:G,grammar:K,language:X};if(b.hooks.run("before-tokenize",ae),!ae.grammar)throw new Error('The language "'+ae.language+'" has no grammar.');return ae.tokens=b.tokenize(ae.code,ae.grammar),b.hooks.run("after-tokenize",ae),y.stringify(b.util.encode(ae.tokens),ae.language)},tokenize:function(G,K){var X=K.rest;if(X){for(var ae in X)K[ae]=X[ae];delete K.rest}var oe=new x;return O(oe,oe.head,G),k(G,oe,K,oe.head,0),D(oe)},hooks:{all:{},add:function(G,K){var X=b.hooks.all;X[G]=X[G]||[],X[G].push(K)},run:function(G,K){var X=b.hooks.all[G];if(!(!X||!X.length))for(var ae=0,oe;oe=X[ae++];)oe(K)}},Token:y};s.Prism=b;function y(G,K,X,ae){this.type=G,this.content=K,this.alias=X,this.length=(ae||"").length|0}y.stringify=function G(K,X){if(typeof K=="string")return K;if(Array.isArray(K)){var ae="";return K.forEach(function(ge){ae+=G(ge,X)}),ae}var oe={type:K.type,content:G(K.content,X),tag:"span",classes:["token",K.type],attributes:{},language:X},he=K.alias;he&&(Array.isArray(he)?Array.prototype.push.apply(oe.classes,he):oe.classes.push(he)),b.hooks.run("wrap",oe);var ye="";for(var we in oe.attributes)ye+=" "+we+'="'+(oe.attributes[we]||"").replace(/"/g,"&quot;")+'"';return"<"+oe.tag+' class="'+oe.classes.join(" ")+'"'+ye+">"+oe.content+"</"+oe.tag+">"};function E(G,K,X,ae){G.lastIndex=K;var oe=G.exec(X);if(oe&&ae&&oe[1]){var he=oe[1].length;oe.index+=he,oe[0]=oe[0].slice(he)}return oe}function k(G,K,X,ae,oe,he){for(var ye in X)if(!(!X.hasOwnProperty(ye)||!X[ye])){var we=X[ye];we=Array.isArray(we)?we:[we];for(var ge=0;ge<we.length;++ge){if(he&&he.cause==ye+","+ge)return;var Te=we[ge],xe=Te.inside,Ue=!!Te.lookbehind,Z=!!Te.greedy,Re=Te.alias;if(Z&&!Te.pattern.global){var _e=Te.pattern.toString().match(/[imsuy]*$/)[0];Te.pattern=RegExp(Te.pattern.source,_e+"g")}for(var qe=Te.pattern||Te,it=ae.next,Xe=oe;it!==K.tail&&!(he&&Xe>=he.reach);Xe+=it.value.length,it=it.next){var de=it.value;if(K.length>G.length)return;if(!(de instanceof y)){var We=1,ht;if(Z){if(ht=E(qe,Xe,G,Ue),!ht||ht.index>=G.length)break;var Yt=ht.index,Ie=ht.index+ht[0].length,$t=Xe;for($t+=it.value.length;Yt>=$t;)it=it.next,$t+=it.value.length;if($t-=it.value.length,Xe=$t,it.value instanceof y)continue;for(var An=it;An!==K.tail&&($t<Ie||typeof An.value=="string");An=An.next)We++,$t+=An.value.length;We--,de=G.slice(Xe,$t),ht.index-=Xe}else if(ht=E(qe,0,de,Ue),!ht)continue;var Yt=ht.index,Or=ht[0],Nt=de.slice(0,Yt),on=de.slice(Yt+Or.length),br=Xe+de.length;he&&br>he.reach&&(he.reach=br);var wi=it.prev;Nt&&(wi=O(K,wi,Nt),Xe+=Nt.length),P(K,wi,We);var no=new y(ye,xe?b.tokenize(Or,xe):Or,Re,Or);if(it=O(K,wi,no),on&&O(K,it,on),We>1){var Ri={cause:ye+","+ge,reach:br};k(G,K,X,it.prev,Xe,Ri),he&&Ri.reach>he.reach&&(he.reach=Ri.reach)}}}}}}function x(){var G={value:null,prev:null,next:null},K={value:null,prev:G,next:null};G.next=K,this.head=G,this.tail=K,this.length=0}function O(G,K,X){var ae=K.next,oe={value:X,prev:K,next:ae};return K.next=oe,ae.prev=oe,G.length++,oe}function P(G,K,X){for(var ae=K.next,oe=0;oe<X&&ae!==G.tail;oe++)ae=ae.next;K.next=ae,ae.prev=K,G.length-=oe}function D(G){for(var K=[],X=G.head.next;X!==G.tail;)K.push(X.value),X=X.next;return K}if(!s.document)return s.addEventListener&&(b.disableWorkerMessageHandler||s.addEventListener("message",function(G){var K=JSON.parse(G.data),X=K.language,ae=K.code,oe=K.immediateClose;s.postMessage(b.highlight(ae,b.languages[X],X)),oe&&s.close()},!1)),b;var $=b.util.currentScript();$&&(b.filename=$.src,$.hasAttribute("data-manual")&&(b.manual=!0));function U(){b.manual||b.highlightAll()}if(!b.manual){var B=document.readyState;B==="loading"||B==="interactive"&&$&&$.defer?document.addEventListener("DOMContentLoaded",U):window.requestAnimationFrame?window.requestAnimationFrame(U):window.setTimeout(U,16)}return b}(r);n.exports&&(n.exports=o),typeof pI<"u"&&(pI.Prism=o)}(WFe)),WFe.exports}var KFe,LUt;function qjr(){if(LUt)return KFe;LUt=1,KFe=n,n.displayName="markup",n.aliases=["html","mathml","svg","xml","ssml","atom","rss"];function n(r){r.languages.markup={comment:{pattern:/<!--(?:(?!<!--)[\s\S])*?-->/,greedy:!0},prolog:{pattern:/<\?[\s\S]+?\?>/,greedy:!0},doctype:{pattern:/<!DOCTYPE(?:[^>"'[\]]|"[^"]*"|'[^']*')+(?:\[(?:[^<"'\]]|"[^"]*"|'[^']*'|<(?!!--)|<!--(?:[^-]|-(?!->))*-->)*\]\s*)?>/i,greedy:!0,inside:{"internal-subset":{pattern:/(^[^\[]*\[)[\s\S]+(?=\]>$)/,lookbehind:!0,greedy:!0,inside:null},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},punctuation:/^<!|>$|[[\]]/,"doctype-tag":/^DOCTYPE/i,name:/[^\s<>'"]+/}},cdata:{pattern:/<!\[CDATA\[[\s\S]*?\]\]>/i,greedy:!0},tag:{pattern:/<\/?(?!\d)[^\s>\/=$<%]+(?:\s(?:\s*[^\s>\/=]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+(?=[\s>]))|(?=[\s/>])))+)?\s*\/?>/,greedy:!0,inside:{tag:{pattern:/^<\/?[^\s>\/]+/,inside:{punctuation:/^<\/?/,namespace:/^[^\s>\/:]+:/}},"special-attr":[],"attr-value":{pattern:/=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+)/,inside:{punctuation:[{pattern:/^=/,alias:"attr-equals"},/"|'/]}},punctuation:/\/?>/,"attr-name":{pattern:/[^\s>\/]+/,inside:{namespace:/^[^\s>\/:]+:/}}}},entity:[{pattern:/&[\da-z]{1,8};/i,alias:"named-entity"},/&#x?[\da-f]{1,8};/i]},r.languages.markup.tag.inside["attr-value"].inside.entity=r.languages.markup.entity,r.languages.markup.doctype.inside["internal-subset"].inside=r.languages.markup,r.hooks.add("wrap",function(o){o.type==="entity"&&(o.attributes.title=o.content.value.replace(/&amp;/,"&"))}),Object.defineProperty(r.languages.markup.tag,"addInlined",{value:function(s,c){var d={};d["language-"+c]={pattern:/(^<!\[CDATA\[)[\s\S]+?(?=\]\]>$)/i,lookbehind:!0,inside:r.languages[c]},d.cdata=/^<!\[CDATA\[|\]\]>$/i;var h={"included-cdata":{pattern:/<!\[CDATA\[[\s\S]*?\]\]>/i,inside:d}};h["language-"+c]={pattern:/[\s\S]+/,inside:r.languages[c]};var b={};b[s]={pattern:RegExp(/(<__[^>]*>)(?:<!\[CDATA\[(?:[^\]]|\](?!\]>))*\]\]>|(?!<!\[CDATA\[)[\s\S])*?(?=<\/__>)/.source.replace(/__/g,function(){return s}),"i"),lookbehind:!0,greedy:!0,inside:h},r.languages.insertBefore("markup","cdata",b)}}),Object.defineProperty(r.languages.markup.tag,"addAttribute",{value:function(o,s){r.languages.markup.tag.inside["special-attr"].push({pattern:RegExp(/(^|["'\s])/.source+"(?:"+o+")"+/\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+(?=[\s>]))/.source,"i"),lookbehind:!0,inside:{"attr-name":/^[^\s=]+/,"attr-value":{pattern:/=[\s\S]+/,inside:{value:{pattern:/(^=\s*(["']|(?!["'])))\S[\s\S]*(?=\2$)/,lookbehind:!0,alias:[s,"language-"+s],inside:r.languages[s]},punctuation:[{pattern:/^=/,alias:"attr-equals"},/"|'/]}}}})}}),r.languages.html=r.languages.markup,r.languages.mathml=r.languages.markup,r.languages.svg=r.languages.markup,r.languages.xml=r.languages.extend("markup",{}),r.languages.ssml=r.languages.xml,r.languages.atom=r.languages.xml,r.languages.rss=r.languages.xml}return KFe}var YFe,DUt;function Vjr(){if(DUt)return YFe;DUt=1,YFe=n,n.displayName="css",n.aliases=[];function n(r){(function(o){var s=/(?:"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"|'(?:\\(?:\r\n|[\s\S])|[^'\\\r\n])*')/;o.languages.css={comment:/\/\*[\s\S]*?\*\//,atrule:{pattern:/@[\w-](?:[^;{\s]|\s+(?![\s{]))*(?:;|(?=\s*\{))/,inside:{rule:/^@[\w-]+/,"selector-function-argument":{pattern:/(\bselector\s*\(\s*(?![\s)]))(?:[^()\s]|\s+(?![\s)])|\((?:[^()]|\([^()]*\))*\))+(?=\s*\))/,lookbehind:!0,alias:"selector"},keyword:{pattern:/(^|[^\w-])(?:and|not|only|or)(?![\w-])/,lookbehind:!0}}},url:{pattern:RegExp("\\burl\\((?:"+s.source+"|"+/(?:[^\\\r\n()"']|\\[\s\S])*/.source+")\\)","i"),greedy:!0,inside:{function:/^url/i,punctuation:/^\(|\)$/,string:{pattern:RegExp("^"+s.source+"$"),alias:"url"}}},selector:{pattern:RegExp(`(^|[{}\\s])[^{}\\s](?:[^{};"'\\s]|\\s+(?![\\s{])|`+s.source+")*(?=\\s*\\{)"),lookbehind:!0},string:{pattern:s,greedy:!0},property:{pattern:/(^|[^-\w\xA0-\uFFFF])(?!\s)[-_a-z\xA0-\uFFFF](?:(?!\s)[-\w\xA0-\uFFFF])*(?=\s*:)/i,lookbehind:!0},important:/!important\b/i,function:{pattern:/(^|[^-a-z0-9])[-a-z0-9]+(?=\()/i,lookbehind:!0},punctuation:/[(){};:,]/},o.languages.css.atrule.inside.rest=o.languages.css;var c=o.languages.markup;c&&(c.tag.addInlined("style","css"),c.tag.addAttribute("style","css"))})(r)}return YFe}var XFe,MUt;function Wjr(){if(MUt)return XFe;MUt=1,XFe=n,n.displayName="clike",n.aliases=[];function n(r){r.languages.clike={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},"class-name":{pattern:/(\b(?:class|extends|implements|instanceof|interface|new|trait)\s+|\bcatch\s+\()[\w.\\]+/i,lookbehind:!0,inside:{punctuation:/[.\\]/}},keyword:/\b(?:break|catch|continue|do|else|finally|for|function|if|in|instanceof|new|null|return|throw|try|while)\b/,boolean:/\b(?:false|true)\b/,function:/\b\w+(?=\()/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,operator:/[<>]=?|[!=]=?=?|--?|\+\+?|&&?|\|\|?|[?*/~^%]/,punctuation:/[{}[\];(),.:]/}}return XFe}var JFe,jUt;function Kjr(){if(jUt)return JFe;jUt=1,JFe=n,n.displayName="javascript",n.aliases=["js"];function n(r){r.languages.javascript=r.languages.extend("clike",{"class-name":[r.languages.clike["class-name"],{pattern:/(^|[^$\w\xA0-\uFFFF])(?!\s)[_$A-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\.(?:constructor|prototype))/,lookbehind:!0}],keyword:[{pattern:/((?:^|\})\s*)catch\b/,lookbehind:!0},{pattern:/(^|[^.]|\.\.\.\s*)\b(?:as|assert(?=\s*\{)|async(?=\s*(?:function\b|\(|[$\w\xA0-\uFFFF]|$))|await|break|case|class|const|continue|debugger|default|delete|do|else|enum|export|extends|finally(?=\s*(?:\{|$))|for|from(?=\s*(?:['"]|$))|function|(?:get|set)(?=\s*(?:[#\[$\w\xA0-\uFFFF]|$))|if|implements|import|in|instanceof|interface|let|new|null|of|package|private|protected|public|return|static|super|switch|this|throw|try|typeof|undefined|var|void|while|with|yield)\b/,lookbehind:!0}],function:/#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*(?:\.\s*(?:apply|bind|call)\s*)?\()/,number:{pattern:RegExp(/(^|[^\w$])/.source+"(?:"+(/NaN|Infinity/.source+"|"+/0[bB][01]+(?:_[01]+)*n?/.source+"|"+/0[oO][0-7]+(?:_[0-7]+)*n?/.source+"|"+/0[xX][\dA-Fa-f]+(?:_[\dA-Fa-f]+)*n?/.source+"|"+/\d+(?:_\d+)*n/.source+"|"+/(?:\d+(?:_\d+)*(?:\.(?:\d+(?:_\d+)*)?)?|\.\d+(?:_\d+)*)(?:[Ee][+-]?\d+(?:_\d+)*)?/.source)+")"+/(?![\w$])/.source),lookbehind:!0},operator:/--|\+\+|\*\*=?|=>|&&=?|\|\|=?|[!=]==|<<=?|>>>?=?|[-+*/%&|^!=<>]=?|\.{3}|\?\?=?|\?\.?|[~:]/}),r.languages.javascript["class-name"][0].pattern=/(\b(?:class|extends|implements|instanceof|interface|new)\s+)[\w.\\]+/,r.languages.insertBefore("javascript","keyword",{regex:{pattern:/((?:^|[^$\w\xA0-\uFFFF."'\])\s]|\b(?:return|yield))\s*)\/(?:\[(?:[^\]\\\r\n]|\\.)*\]|\\.|[^/\\\[\r\n])+\/[dgimyus]{0,7}(?=(?:\s|\/\*(?:[^*]|\*(?!\/))*\*\/)*(?:$|[\r\n,.;:})\]]|\/\/))/,lookbehind:!0,greedy:!0,inside:{"regex-source":{pattern:/^(\/)[\s\S]+(?=\/[a-z]*$)/,lookbehind:!0,alias:"language-regex",inside:r.languages.regex},"regex-delimiter":/^\/|\/$/,"regex-flags":/^[a-z]+$/}},"function-variable":{pattern:/#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*[=:]\s*(?:async\s*)?(?:\bfunction\b|(?:\((?:[^()]|\([^()]*\))*\)|(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)\s*=>))/,alias:"function"},parameter:[{pattern:/(function(?:\s+(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)?\s*\(\s*)(?!\s)(?:[^()\s]|\s+(?![\s)])|\([^()]*\))+(?=\s*\))/,lookbehind:!0,inside:r.languages.javascript},{pattern:/(^|[^$\w\xA0-\uFFFF])(?!\s)[_$a-z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*=>)/i,lookbehind:!0,inside:r.languages.javascript},{pattern:/(\(\s*)(?!\s)(?:[^()\s]|\s+(?![\s)])|\([^()]*\))+(?=\s*\)\s*=>)/,lookbehind:!0,inside:r.languages.javascript},{pattern:/((?:\b|\s|^)(?!(?:as|async|await|break|case|catch|class|const|continue|debugger|default|delete|do|else|enum|export|extends|finally|for|from|function|get|if|implements|import|in|instanceof|interface|let|new|null|of|package|private|protected|public|return|set|static|super|switch|this|throw|try|typeof|undefined|var|void|while|with|yield)(?![$\w\xA0-\uFFFF]))(?:(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*\s*)\(\s*|\]\s*\(\s*)(?!\s)(?:[^()\s]|\s+(?![\s)])|\([^()]*\))+(?=\s*\)\s*\{)/,lookbehind:!0,inside:r.languages.javascript}],constant:/\b[A-Z](?:[A-Z_]|\dx?)*\b/}),r.languages.insertBefore("javascript","string",{hashbang:{pattern:/^#!.*/,greedy:!0,alias:"comment"},"template-string":{pattern:/`(?:\\[\s\S]|\$\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}|(?!\$\{)[^\\`])*`/,greedy:!0,inside:{"template-punctuation":{pattern:/^`|`$/,alias:"string"},interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}/,lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},rest:r.languages.javascript}},string:/[\s\S]+/}},"string-property":{pattern:/((?:^|[,{])[ \t]*)(["'])(?:\\(?:\r\n|[\s\S])|(?!\2)[^\\\r\n])*\2(?=\s*:)/m,lookbehind:!0,greedy:!0,alias:"property"}}),r.languages.insertBefore("javascript","operator",{"literal-property":{pattern:/((?:^|[,{])[ \t]*)(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*:)/m,lookbehind:!0,alias:"property"}}),r.languages.markup&&(r.languages.markup.tag.addInlined("script","javascript"),r.languages.markup.tag.addAttribute(/on(?:abort|blur|change|click|composition(?:end|start|update)|dblclick|error|focus(?:in|out)?|key(?:down|up)|load|mouse(?:down|enter|leave|move|out|over|up)|reset|resize|scroll|select|slotchange|submit|unload|wheel)/.source,"javascript")),r.languages.js=r.languages.javascript}return JFe}var ZFe,$Ut;function Yjr(){if($Ut)return ZFe;$Ut=1;var n=typeof globalThis=="object"?globalThis:typeof self=="object"?self:typeof window=="object"?window:typeof pI=="object"?pI:{},r=ae();n.Prism={manual:!0,disableWorkerMessageHandler:!0};var o=$9r(),s=Hjr(),c=Gjr(),d=qjr(),h=Vjr(),b=Wjr(),y=Kjr();r();var E={}.hasOwnProperty;function k(){}k.prototype=c;var x=new k;ZFe=x,x.highlight=D,x.register=O,x.alias=P,x.registered=$,x.listLanguages=U,O(d),O(h),O(b),O(y),x.util.encode=K,x.Token.stringify=B;function O(oe){if(typeof oe!="function"||!oe.displayName)throw new Error("Expected `function` for `grammar`, got `"+oe+"`");x.languages[oe.displayName]===void 0&&oe(x)}function P(oe,he){var ye=x.languages,we=oe,ge,Te,xe,Ue;he&&(we={},we[oe]=he);for(ge in we)for(Te=we[ge],Te=typeof Te=="string"?[Te]:Te,xe=Te.length,Ue=-1;++Ue<xe;)ye[Te[Ue]]=ye[ge]}function D(oe,he){var ye=c.highlight,we;if(typeof oe!="string")throw new Error("Expected `string` for `value`, got `"+oe+"`");if(x.util.type(he)==="Object")we=he,he=null;else{if(typeof he!="string")throw new Error("Expected `string` for `name`, got `"+he+"`");if(E.call(x.languages,he))we=x.languages[he];else throw new Error("Unknown language: `"+he+"` is not registered")}return ye.call(this,oe,we,he)}function $(oe){if(typeof oe!="string")throw new Error("Expected `string` for `language`, got `"+oe+"`");return E.call(x.languages,oe)}function U(){var oe=x.languages,he=[],ye;for(ye in oe)E.call(oe,ye)&&typeof oe[ye]=="object"&&he.push(ye);return he}function B(oe,he,ye){var we;return typeof oe=="string"?{type:"text",value:oe}:x.util.type(oe)==="Array"?G(oe,he):(we={type:oe.type,content:x.Token.stringify(oe.content,he,ye),tag:"span",classes:["token",oe.type],attributes:{},language:he,parent:ye},oe.alias&&(we.classes=we.classes.concat(oe.alias)),x.hooks.run("wrap",we),o(we.tag+"."+we.classes.join("."),X(we.attributes),we.content))}function G(oe,he){for(var ye=[],we=oe.length,ge=-1,Te;++ge<we;)Te=oe[ge],Te!==""&&Te!==null&&Te!==void 0&&ye.push(Te);for(ge=-1,we=ye.length;++ge<we;)Te=ye[ge],ye[ge]=x.Token.stringify(Te,he,ye);return ye}function K(oe){return oe}function X(oe){var he;for(he in oe)oe[he]=s(oe[he]);return oe}function ae(){var oe="Prism"in n,he=oe?n.Prism:void 0;return ye;function ye(){oe?n.Prism=he:delete n.Prism,oe=void 0,he=void 0}}return ZFe}const Xjr={'code[class*="language-"]':{color:"black",background:"none",textShadow:"0 1px white",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"black",background:"#f5f2f0",textShadow:"0 1px white",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"]::selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#b3d4fc"},':not(pre) > code[class*="language-"]':{background:"#f5f2f0",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"slategray"},prolog:{color:"slategray"},doctype:{color:"slategray"},cdata:{color:"slategray"},punctuation:{color:"#999"},namespace:{Opacity:".7"},property:{color:"#905"},tag:{color:"#905"},boolean:{color:"#905"},number:{color:"#905"},constant:{color:"#905"},symbol:{color:"#905"},deleted:{color:"#905"},selector:{color:"#690"},"attr-name":{color:"#690"},string:{color:"#690"},char:{color:"#690"},builtin:{color:"#690"},inserted:{color:"#690"},operator:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},entity:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)",cursor:"help"},url:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},".language-css .token.string":{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},".style .token.string":{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},atrule:{color:"#07a"},"attr-value":{color:"#07a"},keyword:{color:"#07a"},function:{color:"#DD4A68"},"class-name":{color:"#DD4A68"},regex:{color:"#e90"},important:{color:"#e90",fontWeight:"bold"},variable:{color:"#e90"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}};var QFe,FUt;function Jjr(){if(FUt)return QFe;FUt=1,QFe=n,n.displayName="abap",n.aliases=[];function n(r){r.languages.abap={comment:/^\*.*/m,string:/(`|')(?:\\.|(?!\1)[^\\\r\n])*\1/,"string-template":{pattern:/([|}])(?:\\.|[^\\|{\r\n])*(?=[|{])/,lookbehind:!0,alias:"string"},"eol-comment":{pattern:/(^|\s)".*/m,lookbehind:!0,alias:"comment"},keyword:{pattern:/(\s|\.|^)(?:SCIENTIFIC_WITH_LEADING_ZERO|SCALE_PRESERVING_SCIENTIFIC|RMC_COMMUNICATION_FAILURE|END-ENHANCEMENT-SECTION|MULTIPLY-CORRESPONDING|SUBTRACT-CORRESPONDING|VERIFICATION-MESSAGE|DIVIDE-CORRESPONDING|ENHANCEMENT-SECTION|CURRENCY_CONVERSION|RMC_SYSTEM_FAILURE|START-OF-SELECTION|MOVE-CORRESPONDING|RMC_INVALID_STATUS|CUSTOMER-FUNCTION|END-OF-DEFINITION|ENHANCEMENT-POINT|SYSTEM-EXCEPTIONS|ADD-CORRESPONDING|SCALE_PRESERVING|SELECTION-SCREEN|CURSOR-SELECTION|END-OF-SELECTION|LOAD-OF-PROGRAM|SCROLL-BOUNDARY|SELECTION-TABLE|EXCEPTION-TABLE|IMPLEMENTATIONS|PARAMETER-TABLE|RIGHT-JUSTIFIED|UNIT_CONVERSION|AUTHORITY-CHECK|LIST-PROCESSING|SIGN_AS_POSTFIX|COL_BACKGROUND|IMPLEMENTATION|INTERFACE-POOL|TRANSFORMATION|IDENTIFICATION|ENDENHANCEMENT|LINE-SELECTION|INITIALIZATION|LEFT-JUSTIFIED|SELECT-OPTIONS|SELECTION-SETS|COMMUNICATION|CORRESPONDING|DECIMAL_SHIFT|PRINT-CONTROL|VALUE-REQUEST|CHAIN-REQUEST|FUNCTION-POOL|FIELD-SYMBOLS|FUNCTIONALITY|INVERTED-DATE|SELECTION-SET|CLASS-METHODS|OUTPUT-LENGTH|CLASS-CODING|COL_NEGATIVE|ERRORMESSAGE|FIELD-GROUPS|HELP-REQUEST|NO-EXTENSION|NO-TOPOFPAGE|REDEFINITION|DISPLAY-MODE|ENDINTERFACE|EXIT-COMMAND|FIELD-SYMBOL|NO-SCROLLING|SHORTDUMP-ID|ACCESSPOLICY|CLASS-EVENTS|COL_POSITIVE|DECLARATIONS|ENHANCEMENTS|FILTER-TABLE|SWITCHSTATES|SYNTAX-CHECK|TRANSPORTING|ASYNCHRONOUS|SYNTAX-TRACE|TOKENIZATION|USER-COMMAND|WITH-HEADING|ABAP-SOURCE|BREAK-POINT|CHAIN-INPUT|COMPRESSION|FIXED-POINT|NEW-SECTION|NON-UNICODE|OCCURRENCES|RESPONSIBLE|SYSTEM-CALL|TRACE-TABLE|ABBREVIATED|CHAR-TO-HEX|END-OF-FILE|ENDFUNCTION|ENVIRONMENT|ASSOCIATION|COL_HEADING|EDITOR-CALL|END-OF-PAGE|ENGINEERING|IMPLEMENTED|INTENSIFIED|RADIOBUTTON|SYSTEM-EXIT|TOP-OF-PAGE|TRANSACTION|APPLICATION|CONCATENATE|DESTINATION|ENHANCEMENT|IMMEDIATELY|NO-GROUPING|PRECOMPILED|REPLACEMENT|TITLE-LINES|ACTIVATION|BYTE-ORDER|CLASS-POOL|CONNECTION|CONVERSION|DEFINITION|DEPARTMENT|EXPIRATION|INHERITING|MESSAGE-ID|NO-HEADING|PERFORMING|QUEUE-ONLY|RIGHTSPACE|SCIENTIFIC|STATUSINFO|STRUCTURES|SYNCPOINTS|WITH-TITLE|ATTRIBUTES|BOUNDARIES|CLASS-DATA|COL_NORMAL|DD\/MM\/YYYY|DESCENDING|INTERFACES|LINE-COUNT|MM\/DD\/YYYY|NON-UNIQUE|PRESERVING|SELECTIONS|STATEMENTS|SUBROUTINE|TRUNCATION|TYPE-POOLS|ARITHMETIC|BACKGROUND|ENDPROVIDE|EXCEPTIONS|IDENTIFIER|INDEX-LINE|OBLIGATORY|PARAMETERS|PERCENTAGE|PUSHBUTTON|RESOLUTION|COMPONENTS|DEALLOCATE|DISCONNECT|DUPLICATES|FIRST-LINE|HEAD-LINES|NO-DISPLAY|OCCURRENCE|RESPECTING|RETURNCODE|SUBMATCHES|TRACE-FILE|ASCENDING|BYPASSING|ENDMODULE|EXCEPTION|EXCLUDING|EXPORTING|INCREMENT|MATCHCODE|PARAMETER|PARTIALLY|PREFERRED|REFERENCE|REPLACING|RETURNING|SELECTION|SEPARATED|SPECIFIED|STATEMENT|TIMESTAMP|TYPE-POOL|ACCEPTING|APPENDAGE|ASSIGNING|COL_GROUP|COMPARING|CONSTANTS|DANGEROUS|IMPORTING|INSTANCES|LEFTSPACE|LOG-POINT|QUICKINFO|READ-ONLY|SCROLLING|SQLSCRIPT|STEP-LOOP|TOP-LINES|TRANSLATE|APPENDING|AUTHORITY|CHARACTER|COMPONENT|CONDITION|DIRECTORY|DUPLICATE|MESSAGING|RECEIVING|SUBSCREEN|ACCORDING|COL_TOTAL|END-LINES|ENDMETHOD|ENDSELECT|EXPANDING|EXTENSION|INCLUDING|INFOTYPES|INTERFACE|INTERVALS|LINE-SIZE|PF-STATUS|PROCEDURE|PROTECTED|REQUESTED|RESUMABLE|RIGHTPLUS|SAP-SPOOL|SECONDARY|STRUCTURE|SUBSTRING|TABLEVIEW|NUMOFCHAR|ADJACENT|ANALYSIS|ASSIGNED|BACKWARD|CHANNELS|CHECKBOX|CONTINUE|CRITICAL|DATAINFO|DD\/MM\/YY|DURATION|ENCODING|ENDCLASS|FUNCTION|LEFTPLUS|LINEFEED|MM\/DD\/YY|OVERFLOW|RECEIVED|SKIPPING|SORTABLE|STANDARD|SUBTRACT|SUPPRESS|TABSTRIP|TITLEBAR|TRUNCATE|UNASSIGN|WHENEVER|ANALYZER|COALESCE|COMMENTS|CONDENSE|DECIMALS|DEFERRED|ENDWHILE|EXPLICIT|KEYWORDS|MESSAGES|POSITION|PRIORITY|RECEIVER|RENAMING|TIMEZONE|TRAILING|ALLOCATE|CENTERED|CIRCULAR|CONTROLS|CURRENCY|DELETING|DESCRIBE|DISTANCE|ENDCATCH|EXPONENT|EXTENDED|GENERATE|IGNORING|INCLUDES|INTERNAL|MAJOR-ID|MODIFIER|NEW-LINE|OPTIONAL|PROPERTY|ROLLBACK|STARTING|SUPPLIED|ABSTRACT|CHANGING|CONTEXTS|CREATING|CUSTOMER|DATABASE|DAYLIGHT|DEFINING|DISTINCT|DIVISION|ENABLING|ENDCHAIN|ESCAPING|HARMLESS|IMPLICIT|INACTIVE|LANGUAGE|MINOR-ID|MULTIPLY|NEW-PAGE|NO-TITLE|POS_HIGH|SEPARATE|TEXTPOOL|TRANSFER|SELECTOR|DBMAXLEN|ITERATOR|ARCHIVE|BIT-XOR|BYTE-CO|COLLECT|COMMENT|CURRENT|DEFAULT|DISPLAY|ENDFORM|EXTRACT|LEADING|LISTBOX|LOCATOR|MEMBERS|METHODS|NESTING|POS_LOW|PROCESS|PROVIDE|RAISING|RESERVE|SECONDS|SUMMARY|VISIBLE|BETWEEN|BIT-AND|BYTE-CS|CLEANUP|COMPUTE|CONTROL|CONVERT|DATASET|ENDCASE|FORWARD|HEADERS|HOTSPOT|INCLUDE|INVERSE|KEEPING|NO-ZERO|OBJECTS|OVERLAY|PADDING|PATTERN|PROGRAM|REFRESH|SECTION|SUMMING|TESTING|VERSION|WINDOWS|WITHOUT|BIT-NOT|BYTE-CA|BYTE-NA|CASTING|CONTEXT|COUNTRY|DYNAMIC|ENABLED|ENDLOOP|EXECUTE|FRIENDS|HANDLER|HEADING|INITIAL|\*-INPUT|LOGFILE|MAXIMUM|MINIMUM|NO-GAPS|NO-SIGN|PRAGMAS|PRIMARY|PRIVATE|REDUCED|REPLACE|REQUEST|RESULTS|UNICODE|WARNING|ALIASES|BYTE-CN|BYTE-NS|CALLING|COL_KEY|COLUMNS|CONNECT|ENDEXEC|ENTRIES|EXCLUDE|FILTERS|FURTHER|HELP-ID|LOGICAL|MAPPING|MESSAGE|NAMETAB|OPTIONS|PACKAGE|PERFORM|RECEIVE|STATICS|VARYING|BINDING|CHARLEN|GREATER|XSTRLEN|ACCEPT|APPEND|DETAIL|ELSEIF|ENDING|ENDTRY|FORMAT|FRAMES|GIVING|HASHED|HEADER|IMPORT|INSERT|MARGIN|MODULE|NATIVE|OBJECT|OFFSET|REMOTE|RESUME|SAVING|SIMPLE|SUBMIT|TABBED|TOKENS|UNIQUE|UNPACK|UPDATE|WINDOW|YELLOW|ACTUAL|ASPECT|CENTER|CURSOR|DELETE|DIALOG|DIVIDE|DURING|ERRORS|EVENTS|EXTEND|FILTER|HANDLE|HAVING|IGNORE|LITTLE|MEMORY|NO-GAP|OCCURS|OPTION|PERSON|PLACES|PUBLIC|REDUCE|REPORT|RESULT|SINGLE|SORTED|SWITCH|SYNTAX|TARGET|VALUES|WRITER|ASSERT|BLOCKS|BOUNDS|BUFFER|CHANGE|COLUMN|COMMIT|CONCAT|COPIES|CREATE|DDMMYY|DEFINE|ENDIAN|ESCAPE|EXPAND|KERNEL|LAYOUT|LEGACY|LEVELS|MMDDYY|NUMBER|OUTPUT|RANGES|READER|RETURN|SCREEN|SEARCH|SELECT|SHARED|SOURCE|STABLE|STATIC|SUBKEY|SUFFIX|TABLES|UNWIND|YYMMDD|ASSIGN|BACKUP|BEFORE|BINARY|BIT-OR|BLANKS|CLIENT|CODING|COMMON|DEMAND|DYNPRO|EXCEPT|EXISTS|EXPORT|FIELDS|GLOBAL|GROUPS|LENGTH|LOCALE|MEDIUM|METHOD|MODIFY|NESTED|OTHERS|REJECT|SCROLL|SUPPLY|SYMBOL|ENDFOR|STRLEN|ALIGN|BEGIN|BOUND|ENDAT|ENTRY|EVENT|FINAL|FLUSH|GRANT|INNER|SHORT|USING|WRITE|AFTER|BLACK|BLOCK|CLOCK|COLOR|COUNT|DUMMY|EMPTY|ENDDO|ENDON|GREEN|INDEX|INOUT|LEAVE|LEVEL|LINES|MODIF|ORDER|OUTER|RANGE|RESET|RETRY|RIGHT|SMART|SPLIT|STYLE|TABLE|THROW|UNDER|UNTIL|UPPER|UTF-8|WHERE|ALIAS|BLANK|CLEAR|CLOSE|EXACT|FETCH|FIRST|FOUND|GROUP|LLANG|LOCAL|OTHER|REGEX|SPOOL|TITLE|TYPES|VALID|WHILE|ALPHA|BOXED|CATCH|CHAIN|CHECK|CLASS|COVER|ENDIF|EQUIV|FIELD|FLOOR|FRAME|INPUT|LOWER|MATCH|NODES|PAGES|PRINT|RAISE|ROUND|SHIFT|SPACE|SPOTS|STAMP|STATE|TASKS|TIMES|TRMAC|ULINE|UNION|VALUE|WIDTH|EQUAL|LOG10|TRUNC|BLOB|CASE|CEIL|CLOB|COND|EXIT|FILE|GAPS|HOLD|INCL|INTO|KEEP|KEYS|LAST|LINE|LONG|LPAD|MAIL|MODE|OPEN|PINK|READ|ROWS|TEST|THEN|ZERO|AREA|BACK|BADI|BYTE|CAST|EDIT|EXEC|FAIL|FIND|FKEQ|FONT|FREE|GKEQ|HIDE|INIT|ITNO|LATE|LOOP|MAIN|MARK|MOVE|NEXT|NULL|RISK|ROLE|UNIT|WAIT|ZONE|BASE|CALL|CODE|DATA|DATE|FKGE|GKGE|HIGH|KIND|LEFT|LIST|MASK|MESH|NAME|NODE|PACK|PAGE|POOL|SEND|SIGN|SIZE|SOME|STOP|TASK|TEXT|TIME|USER|VARY|WITH|WORD|BLUE|CONV|COPY|DEEP|ELSE|FORM|FROM|HINT|ICON|JOIN|LIKE|LOAD|ONLY|PART|SCAN|SKIP|SORT|TYPE|UNIX|VIEW|WHEN|WORK|ACOS|ASIN|ATAN|COSH|EACH|FRAC|LESS|RTTI|SINH|SQRT|TANH|AVG|BIT|DIV|ISO|LET|OUT|PAD|SQL|ALL|CI_|CPI|END|LOB|LPI|MAX|MIN|NEW|OLE|RUN|SET|\?TO|YES|ABS|ADD|AND|BIG|FOR|HDB|JOB|LOW|NOT|SAP|TRY|VIA|XML|ANY|GET|IDS|KEY|MOD|OFF|PUT|RAW|RED|REF|SUM|TAB|XSD|CNT|COS|EXP|LOG|SIN|TAN|XOR|AT|CO|CP|DO|GT|ID|IF|NS|OR|BT|CA|CS|GE|NA|NB|EQ|IN|LT|NE|NO|OF|ON|PF|TO|AS|BY|CN|IS|LE|NP|UP|E|I|M|O|Z|C|X)\b/i,lookbehind:!0},number:/\b\d+\b/,operator:{pattern:/(\s)(?:\*\*?|<[=>]?|>=?|\?=|[-+\/=])(?=\s)/,lookbehind:!0},"string-operator":{pattern:/(\s)&&?(?=\s)/,lookbehind:!0,alias:"keyword"},"token-operator":[{pattern:/(\w)(?:->?|=>|[~|{}])(?=\w)/,lookbehind:!0,alias:"punctuation"},{pattern:/[|{}]/,alias:"punctuation"}],punctuation:/[,.:()]/}}return QFe}var eBe,BUt;function Zjr(){if(BUt)return eBe;BUt=1,eBe=n,n.displayName="abnf",n.aliases=[];function n(r){(function(o){var s="(?:ALPHA|BIT|CHAR|CR|CRLF|CTL|DIGIT|DQUOTE|HEXDIG|HTAB|LF|LWSP|OCTET|SP|VCHAR|WSP)";o.languages.abnf={comment:/;.*/,string:{pattern:/(?:%[is])?"[^"\n\r]*"/,greedy:!0,inside:{punctuation:/^%[is]/}},range:{pattern:/%(?:b[01]+-[01]+|d\d+-\d+|x[A-F\d]+-[A-F\d]+)/i,alias:"number"},terminal:{pattern:/%(?:b[01]+(?:\.[01]+)*|d\d+(?:\.\d+)*|x[A-F\d]+(?:\.[A-F\d]+)*)/i,alias:"number"},repetition:{pattern:/(^|[^\w-])(?:\d*\*\d*|\d+)/,lookbehind:!0,alias:"operator"},definition:{pattern:/(^[ \t]*)(?:[a-z][\w-]*|<[^<>\r\n]*>)(?=\s*=)/m,lookbehind:!0,alias:"keyword",inside:{punctuation:/<|>/}},"core-rule":{pattern:RegExp("(?:(^|[^<\\w-])"+s+"|<"+s+">)(?![\\w-])","i"),lookbehind:!0,alias:["rule","constant"],inside:{punctuation:/<|>/}},rule:{pattern:/(^|[^<\w-])[a-z][\w-]*|<[^<>\r\n]*>/i,lookbehind:!0,inside:{punctuation:/<|>/}},operator:/=\/?|\//,punctuation:/[()\[\]]/}})(r)}return eBe}var tBe,UUt;function Qjr(){if(UUt)return tBe;UUt=1,tBe=n,n.displayName="actionscript",n.aliases=[];function n(r){r.languages.actionscript=r.languages.extend("javascript",{keyword:/\b(?:as|break|case|catch|class|const|default|delete|do|dynamic|each|else|extends|final|finally|for|function|get|if|implements|import|in|include|instanceof|interface|internal|is|namespace|native|new|null|override|package|private|protected|public|return|set|static|super|switch|this|throw|try|typeof|use|var|void|while|with)\b/,operator:/\+\+|--|(?:[+\-*\/%^]|&&?|\|\|?|<<?|>>?>?|[!=]=?)=?|[~?@]/}),r.languages.actionscript["class-name"].alias="function",delete r.languages.actionscript.parameter,delete r.languages.actionscript["literal-property"],r.languages.markup&&r.languages.insertBefore("actionscript","string",{xml:{pattern:/(^|[^.])<\/?\w+(?:\s+[^\s>\/=]+=("|')(?:\\[\s\S]|(?!\2)[^\\])*\2)*\s*\/?>/,lookbehind:!0,inside:r.languages.markup}})}return tBe}var nBe,zUt;function e$r(){if(zUt)return nBe;zUt=1,nBe=n,n.displayName="ada",n.aliases=[];function n(r){r.languages.ada={comment:/--.*/,string:/"(?:""|[^"\r\f\n])*"/,number:[{pattern:/\b\d(?:_?\d)*#[\dA-F](?:_?[\dA-F])*(?:\.[\dA-F](?:_?[\dA-F])*)?#(?:E[+-]?\d(?:_?\d)*)?/i},{pattern:/\b\d(?:_?\d)*(?:\.\d(?:_?\d)*)?(?:E[+-]?\d(?:_?\d)*)?\b/i}],"attr-name":/\b'\w+/,keyword:/\b(?:abort|abs|abstract|accept|access|aliased|all|and|array|at|begin|body|case|constant|declare|delay|delta|digits|do|else|elsif|end|entry|exception|exit|for|function|generic|goto|if|in|interface|is|limited|loop|mod|new|not|null|of|others|out|overriding|package|pragma|private|procedure|protected|raise|range|record|rem|renames|requeue|return|reverse|select|separate|some|subtype|synchronized|tagged|task|terminate|then|type|until|use|when|while|with|xor)\b/i,boolean:/\b(?:false|true)\b/i,operator:/<[=>]?|>=?|=>?|:=|\/=?|\*\*?|[&+-]/,punctuation:/\.\.?|[,;():]/,char:/'.'/,variable:/\b[a-z](?:\w)*\b/i}}return nBe}var rBe,HUt;function t$r(){if(HUt)return rBe;HUt=1,rBe=n,n.displayName="agda",n.aliases=[];function n(r){(function(o){o.languages.agda={comment:/\{-[\s\S]*?(?:-\}|$)|--.*/,string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^\\\r\n"])*"/,greedy:!0},punctuation:/[(){}⦃⦄.;@]/,"class-name":{pattern:/((?:data|record) +)\S+/,lookbehind:!0},function:{pattern:/(^[ \t]*)(?!\s)[^:\r\n]+(?=:)/m,lookbehind:!0},operator:{pattern:/(^\s*|\s)(?:[=|:∀→λ\\?_]|->)(?=\s)/,lookbehind:!0},keyword:/\b(?:Set|abstract|constructor|data|eta-equality|field|forall|hiding|import|in|inductive|infix|infixl|infixr|instance|let|macro|module|mutual|no-eta-equality|open|overlap|pattern|postulate|primitive|private|public|quote|quoteContext|quoteGoal|quoteTerm|record|renaming|rewrite|syntax|tactic|unquote|unquoteDecl|unquoteDef|using|variable|where|with)\b/}})(r)}return rBe}var iBe,GUt;function n$r(){if(GUt)return iBe;GUt=1,iBe=n,n.displayName="al",n.aliases=[];function n(r){r.languages.al={comment:/\/\/.*|\/\*[\s\S]*?\*\//,string:{pattern:/'(?:''|[^'\r\n])*'(?!')|"(?:""|[^"\r\n])*"(?!")/,greedy:!0},function:{pattern:/(\b(?:event|procedure|trigger)\s+|(?:^|[^.])\.\s*)[a-z_]\w*(?=\s*\()/i,lookbehind:!0},keyword:[/\b(?:array|asserterror|begin|break|case|do|downto|else|end|event|exit|for|foreach|function|if|implements|in|indataset|interface|internal|local|of|procedure|program|protected|repeat|runonclient|securityfiltering|suppressdispose|temporary|then|to|trigger|until|var|while|with|withevents)\b/i,/\b(?:action|actions|addafter|addbefore|addfirst|addlast|area|assembly|chartpart|codeunit|column|controladdin|cuegroup|customizes|dataitem|dataset|dotnet|elements|enum|enumextension|extends|field|fieldattribute|fieldelement|fieldgroup|fieldgroups|fields|filter|fixed|grid|group|key|keys|label|labels|layout|modify|moveafter|movebefore|movefirst|movelast|page|pagecustomization|pageextension|part|profile|query|repeater|report|requestpage|schema|separator|systempart|table|tableelement|tableextension|textattribute|textelement|type|usercontrol|value|xmlport)\b/i],number:/\b(?:0x[\da-f]+|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?)(?:F|LL?|U(?:LL?)?)?\b/i,boolean:/\b(?:false|true)\b/i,variable:/\b(?:Curr(?:FieldNo|Page|Report)|x?Rec|RequestOptionsPage)\b/,"class-name":/\b(?:automation|biginteger|bigtext|blob|boolean|byte|char|clienttype|code|completiontriggererrorlevel|connectiontype|database|dataclassification|datascope|date|dateformula|datetime|decimal|defaultlayout|dialog|dictionary|dotnetassembly|dotnettypedeclaration|duration|errorinfo|errortype|executioncontext|executionmode|fieldclass|fieldref|fieldtype|file|filterpagebuilder|guid|httpclient|httpcontent|httpheaders|httprequestmessage|httpresponsemessage|instream|integer|joker|jsonarray|jsonobject|jsontoken|jsonvalue|keyref|list|moduledependencyinfo|moduleinfo|none|notification|notificationscope|objecttype|option|outstream|pageresult|record|recordid|recordref|reportformat|securityfilter|sessionsettings|tableconnectiontype|tablefilter|testaction|testfield|testfilterfield|testpage|testpermissions|testrequestpage|text|textbuilder|textconst|textencoding|time|transactionmodel|transactiontype|variant|verbosity|version|view|views|webserviceactioncontext|webserviceactionresultcode|xmlattribute|xmlattributecollection|xmlcdata|xmlcomment|xmldeclaration|xmldocument|xmldocumenttype|xmlelement|xmlnamespacemanager|xmlnametable|xmlnode|xmlnodelist|xmlprocessinginstruction|xmlreadoptions|xmltext|xmlwriteoptions)\b/i,operator:/\.\.|:[=:]|[-+*/]=?|<>|[<>]=?|=|\b(?:and|div|mod|not|or|xor)\b/i,punctuation:/[()\[\]{}:.;,]/}}return iBe}var oBe,qUt;function r$r(){if(qUt)return oBe;qUt=1,oBe=n,n.displayName="antlr4",n.aliases=["g4"];function n(r){r.languages.antlr4={comment:/\/\/.*|\/\*[\s\S]*?(?:\*\/|$)/,string:{pattern:/'(?:\\.|[^\\'\r\n])*'/,greedy:!0},"character-class":{pattern:/\[(?:\\.|[^\\\]\r\n])*\]/,greedy:!0,alias:"regex",inside:{range:{pattern:/([^[]|(?:^|[^\\])(?:\\\\)*\\\[)-(?!\])/,lookbehind:!0,alias:"punctuation"},escape:/\\(?:u(?:[a-fA-F\d]{4}|\{[a-fA-F\d]+\})|[pP]\{[=\w-]+\}|[^\r\nupP])/,punctuation:/[\[\]]/}},action:{pattern:/\{(?:[^{}]|\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\})*\}/,greedy:!0,inside:{content:{pattern:/(\{)[\s\S]+(?=\})/,lookbehind:!0},punctuation:/[{}]/}},command:{pattern:/(->\s*(?!\s))(?:\s*(?:,\s*)?\b[a-z]\w*(?:\s*\([^()\r\n]*\))?)+(?=\s*;)/i,lookbehind:!0,inside:{function:/\b\w+(?=\s*(?:[,(]|$))/,punctuation:/[,()]/}},annotation:{pattern:/@\w+(?:::\w+)*/,alias:"keyword"},label:{pattern:/#[ \t]*\w+/,alias:"punctuation"},keyword:/\b(?:catch|channels|finally|fragment|grammar|import|lexer|locals|mode|options|parser|returns|throws|tokens)\b/,definition:[{pattern:/\b[a-z]\w*(?=\s*:)/,alias:["rule","class-name"]},{pattern:/\b[A-Z]\w*(?=\s*:)/,alias:["token","constant"]}],constant:/\b[A-Z][A-Z_]*\b/,operator:/\.\.|->|[|~]|[*+?]\??/,punctuation:/[;:()=]/},r.languages.g4=r.languages.antlr4}return oBe}var aBe,VUt;function i$r(){if(VUt)return aBe;VUt=1,aBe=n,n.displayName="apacheconf",n.aliases=[];function n(r){r.languages.apacheconf={comment:/#.*/,"directive-inline":{pattern:/(^[\t ]*)\b(?:AcceptFilter|AcceptPathInfo|AccessFileName|Action|Add(?:Alt|AltByEncoding|AltByType|Charset|DefaultCharset|Description|Encoding|Handler|Icon|IconByEncoding|IconByType|InputFilter|Language|ModuleInfo|OutputFilter|OutputFilterByType|Type)|Alias|AliasMatch|Allow(?:CONNECT|EncodedSlashes|Methods|Override|OverrideList)?|Anonymous(?:_LogEmail|_MustGiveEmail|_NoUserID|_VerifyEmail)?|AsyncRequestWorkerFactor|Auth(?:BasicAuthoritative|BasicFake|BasicProvider|BasicUseDigestAlgorithm|DBDUserPWQuery|DBDUserRealmQuery|DBMGroupFile|DBMType|DBMUserFile|Digest(?:Algorithm|Domain|NonceLifetime|Provider|Qop|ShmemSize)|Form(?:Authoritative|Body|DisableNoStore|FakeBasicAuth|Location|LoginRequiredLocation|LoginSuccessLocation|LogoutLocation|Method|Mimetype|Password|Provider|SitePassphrase|Size|Username)|GroupFile|LDAP(?:AuthorizePrefix|BindAuthoritative|BindDN|BindPassword|CharsetConfig|CompareAsUser|CompareDNOnServer|DereferenceAliases|GroupAttribute|GroupAttributeIsDN|InitialBindAsUser|InitialBindPattern|MaxSubGroupDepth|RemoteUserAttribute|RemoteUserIsDN|SearchAsUser|SubGroupAttribute|SubGroupClass|Url)|Merging|Name|nCache(?:Context|Enable|ProvideFor|SOCache|Timeout)|nzFcgiCheckAuthnProvider|nzFcgiDefineProvider|Type|UserFile|zDBDLoginToReferer|zDBDQuery|zDBDRedirectQuery|zDBMType|zSendForbiddenOnFailure)|BalancerGrowth|BalancerInherit|BalancerMember|BalancerPersist|BrowserMatch|BrowserMatchNoCase|BufferedLogs|BufferSize|Cache(?:DefaultExpire|DetailHeader|DirLength|DirLevels|Disable|Enable|File|Header|IgnoreCacheControl|IgnoreHeaders|IgnoreNoLastMod|IgnoreQueryString|IgnoreURLSessionIdentifiers|KeyBaseURL|LastModifiedFactor|Lock|LockMaxAge|LockPath|MaxExpire|MaxFileSize|MinExpire|MinFileSize|NegotiatedDocs|QuickHandler|ReadSize|ReadTime|Root|Socache(?:MaxSize|MaxTime|MinTime|ReadSize|ReadTime)?|StaleOnError|StoreExpired|StoreNoStore|StorePrivate)|CGIDScriptTimeout|CGIMapExtension|CharsetDefault|CharsetOptions|CharsetSourceEnc|CheckCaseOnly|CheckSpelling|ChrootDir|ContentDigest|CookieDomain|CookieExpires|CookieName|CookieStyle|CookieTracking|CoreDumpDirectory|CustomLog|Dav|DavDepthInfinity|DavGenericLockDB|DavLockDB|DavMinTimeout|DBDExptime|DBDInitSQL|DBDKeep|DBDMax|DBDMin|DBDParams|DBDPersist|DBDPrepareSQL|DBDriver|DefaultIcon|DefaultLanguage|DefaultRuntimeDir|DefaultType|Define|Deflate(?:BufferSize|CompressionLevel|FilterNote|InflateLimitRequestBody|InflateRatio(?:Burst|Limit)|MemLevel|WindowSize)|Deny|DirectoryCheckHandler|DirectoryIndex|DirectoryIndexRedirect|DirectorySlash|DocumentRoot|DTracePrivileges|DumpIOInput|DumpIOOutput|EnableExceptionHook|EnableMMAP|EnableSendfile|Error|ErrorDocument|ErrorLog|ErrorLogFormat|Example|ExpiresActive|ExpiresByType|ExpiresDefault|ExtendedStatus|ExtFilterDefine|ExtFilterOptions|FallbackResource|FileETag|FilterChain|FilterDeclare|FilterProtocol|FilterProvider|FilterTrace|ForceLanguagePriority|ForceType|ForensicLog|GprofDir|GracefulShutdownTimeout|Group|Header|HeaderName|Heartbeat(?:Address|Listen|MaxServers|Storage)|HostnameLookups|IdentityCheck|IdentityCheckTimeout|ImapBase|ImapDefault|ImapMenu|Include|IncludeOptional|Index(?:HeadInsert|Ignore|IgnoreReset|Options|OrderDefault|StyleSheet)|InputSed|ISAPI(?:AppendLogToErrors|AppendLogToQuery|CacheFile|FakeAsync|LogNotSupported|ReadAheadBuffer)|KeepAlive|KeepAliveTimeout|KeptBodySize|LanguagePriority|LDAP(?:CacheEntries|CacheTTL|ConnectionPoolTTL|ConnectionTimeout|LibraryDebug|OpCacheEntries|OpCacheTTL|ReferralHopLimit|Referrals|Retries|RetryDelay|SharedCacheFile|SharedCacheSize|Timeout|TrustedClientCert|TrustedGlobalCert|TrustedMode|VerifyServerCert)|Limit(?:InternalRecursion|Request(?:Body|Fields|FieldSize|Line)|XMLRequestBody)|Listen|ListenBackLog|LoadFile|LoadModule|LogFormat|LogLevel|LogMessage|LuaAuthzProvider|LuaCodeCache|Lua(?:Hook(?:AccessChecker|AuthChecker|CheckUserID|Fixups|InsertFilter|Log|MapToStorage|TranslateName|TypeChecker)|Inherit|InputFilter|MapHandler|OutputFilter|PackageCPath|PackagePath|QuickHandler|Root|Scope)|Max(?:ConnectionsPerChild|KeepAliveRequests|MemFree|RangeOverlaps|RangeReversals|Ranges|RequestWorkers|SpareServers|SpareThreads|Threads)|MergeTrailers|MetaDir|MetaFiles|MetaSuffix|MimeMagicFile|MinSpareServers|MinSpareThreads|MMapFile|ModemStandard|ModMimeUsePathInfo|MultiviewsMatch|Mutex|NameVirtualHost|NoProxy|NWSSLTrustedCerts|NWSSLUpgradeable|Options|Order|OutputSed|PassEnv|PidFile|PrivilegesMode|Protocol|ProtocolEcho|Proxy(?:AddHeaders|BadHeader|Block|Domain|ErrorOverride|ExpressDBMFile|ExpressDBMType|ExpressEnable|FtpDirCharset|FtpEscapeWildcards|FtpListOnWildcard|HTML(?:BufSize|CharsetOut|DocType|Enable|Events|Extended|Fixups|Interp|Links|Meta|StripComments|URLMap)|IOBufferSize|MaxForwards|Pass(?:Inherit|InterpolateEnv|Match|Reverse|ReverseCookieDomain|ReverseCookiePath)?|PreserveHost|ReceiveBufferSize|Remote|RemoteMatch|Requests|SCGIInternalRedirect|SCGISendfile|Set|SourceAddress|Status|Timeout|Via)|ReadmeName|ReceiveBufferSize|Redirect|RedirectMatch|RedirectPermanent|RedirectTemp|ReflectorHeader|RemoteIP(?:Header|InternalProxy|InternalProxyList|ProxiesHeader|TrustedProxy|TrustedProxyList)|RemoveCharset|RemoveEncoding|RemoveHandler|RemoveInputFilter|RemoveLanguage|RemoveOutputFilter|RemoveType|RequestHeader|RequestReadTimeout|Require|Rewrite(?:Base|Cond|Engine|Map|Options|Rule)|RLimitCPU|RLimitMEM|RLimitNPROC|Satisfy|ScoreBoardFile|Script(?:Alias|AliasMatch|InterpreterSource|Log|LogBuffer|LogLength|Sock)?|SecureListen|SeeRequestTail|SendBufferSize|Server(?:Admin|Alias|Limit|Name|Path|Root|Signature|Tokens)|Session(?:Cookie(?:Name|Name2|Remove)|Crypto(?:Cipher|Driver|Passphrase|PassphraseFile)|DBD(?:CookieName|CookieName2|CookieRemove|DeleteLabel|InsertLabel|PerUser|SelectLabel|UpdateLabel)|Env|Exclude|Header|Include|MaxAge)?|SetEnv|SetEnvIf|SetEnvIfExpr|SetEnvIfNoCase|SetHandler|SetInputFilter|SetOutputFilter|SSIEndTag|SSIErrorMsg|SSIETag|SSILastModified|SSILegacyExprParser|SSIStartTag|SSITimeFormat|SSIUndefinedEcho|SSL(?:CACertificateFile|CACertificatePath|CADNRequestFile|CADNRequestPath|CARevocationCheck|CARevocationFile|CARevocationPath|CertificateChainFile|CertificateFile|CertificateKeyFile|CipherSuite|Compression|CryptoDevice|Engine|FIPS|HonorCipherOrder|InsecureRenegotiation|OCSP(?:DefaultResponder|Enable|OverrideResponder|ResponderTimeout|ResponseMaxAge|ResponseTimeSkew|UseRequestNonce)|OpenSSLConfCmd|Options|PassPhraseDialog|Protocol|Proxy(?:CACertificateFile|CACertificatePath|CARevocation(?:Check|File|Path)|CheckPeer(?:CN|Expire|Name)|CipherSuite|Engine|MachineCertificate(?:ChainFile|File|Path)|Protocol|Verify|VerifyDepth)|RandomSeed|RenegBufferSize|Require|RequireSSL|Session(?:Cache|CacheTimeout|TicketKeyFile|Tickets)|SRPUnknownUserSeed|SRPVerifierFile|Stapling(?:Cache|ErrorCacheTimeout|FakeTryLater|ForceURL|ResponderTimeout|ResponseMaxAge|ResponseTimeSkew|ReturnResponderErrors|StandardCacheTimeout)|StrictSNIVHostCheck|UserName|UseStapling|VerifyClient|VerifyDepth)|StartServers|StartThreads|Substitute|Suexec|SuexecUserGroup|ThreadLimit|ThreadsPerChild|ThreadStackSize|TimeOut|TraceEnable|TransferLog|TypesConfig|UnDefine|UndefMacro|UnsetEnv|Use|UseCanonicalName|UseCanonicalPhysicalPort|User|UserDir|VHostCGIMode|VHostCGIPrivs|VHostGroup|VHostPrivs|VHostSecure|VHostUser|Virtual(?:DocumentRoot|ScriptAlias)(?:IP)?|WatchdogInterval|XBitHack|xml2EncAlias|xml2EncDefault|xml2StartParse)\b/im,lookbehind:!0,alias:"property"},"directive-block":{pattern:/<\/?\b(?:Auth[nz]ProviderAlias|Directory|DirectoryMatch|Else|ElseIf|Files|FilesMatch|If|IfDefine|IfModule|IfVersion|Limit|LimitExcept|Location|LocationMatch|Macro|Proxy|Require(?:All|Any|None)|VirtualHost)\b.*>/i,inside:{"directive-block":{pattern:/^<\/?\w+/,inside:{punctuation:/^<\/?/},alias:"tag"},"directive-block-parameter":{pattern:/.*[^>]/,inside:{punctuation:/:/,string:{pattern:/("|').*\1/,inside:{variable:/[$%]\{?(?:\w\.?[-+:]?)+\}?/}}},alias:"attr-value"},punctuation:/>/},alias:"tag"},"directive-flags":{pattern:/\[(?:[\w=],?)+\]/,alias:"keyword"},string:{pattern:/("|').*\1/,inside:{variable:/[$%]\{?(?:\w\.?[-+:]?)+\}?/}},variable:/[$%]\{?(?:\w\.?[-+:]?)+\}?/,regex:/\^?.*\$|\^.*\$?/}}return aBe}var sBe,WUt;function wQe(){if(WUt)return sBe;WUt=1,sBe=n,n.displayName="sql",n.aliases=[];function n(r){r.languages.sql={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:--|\/\/|#).*)/,lookbehind:!0},variable:[{pattern:/@(["'`])(?:\\[\s\S]|(?!\1)[^\\])+\1/,greedy:!0},/@[\w.$]+/],string:{pattern:/(^|[^@\\])("|')(?:\\[\s\S]|(?!\2)[^\\]|\2\2)*\2/,greedy:!0,lookbehind:!0},identifier:{pattern:/(^|[^@\\])`(?:\\[\s\S]|[^`\\]|``)*`/,greedy:!0,lookbehind:!0,inside:{punctuation:/^`|`$/}},function:/\b(?:AVG|COUNT|FIRST|FORMAT|LAST|LCASE|LEN|MAX|MID|MIN|MOD|NOW|ROUND|SUM|UCASE)(?=\s*\()/i,keyword:/\b(?:ACTION|ADD|AFTER|ALGORITHM|ALL|ALTER|ANALYZE|ANY|APPLY|AS|ASC|AUTHORIZATION|AUTO_INCREMENT|BACKUP|BDB|BEGIN|BERKELEYDB|BIGINT|BINARY|BIT|BLOB|BOOL|BOOLEAN|BREAK|BROWSE|BTREE|BULK|BY|CALL|CASCADED?|CASE|CHAIN|CHAR(?:ACTER|SET)?|CHECK(?:POINT)?|CLOSE|CLUSTERED|COALESCE|COLLATE|COLUMNS?|COMMENT|COMMIT(?:TED)?|COMPUTE|CONNECT|CONSISTENT|CONSTRAINT|CONTAINS(?:TABLE)?|CONTINUE|CONVERT|CREATE|CROSS|CURRENT(?:_DATE|_TIME|_TIMESTAMP|_USER)?|CURSOR|CYCLE|DATA(?:BASES?)?|DATE(?:TIME)?|DAY|DBCC|DEALLOCATE|DEC|DECIMAL|DECLARE|DEFAULT|DEFINER|DELAYED|DELETE|DELIMITERS?|DENY|DESC|DESCRIBE|DETERMINISTIC|DISABLE|DISCARD|DISK|DISTINCT|DISTINCTROW|DISTRIBUTED|DO|DOUBLE|DROP|DUMMY|DUMP(?:FILE)?|DUPLICATE|ELSE(?:IF)?|ENABLE|ENCLOSED|END|ENGINE|ENUM|ERRLVL|ERRORS|ESCAPED?|EXCEPT|EXEC(?:UTE)?|EXISTS|EXIT|EXPLAIN|EXTENDED|FETCH|FIELDS|FILE|FILLFACTOR|FIRST|FIXED|FLOAT|FOLLOWING|FOR(?: EACH ROW)?|FORCE|FOREIGN|FREETEXT(?:TABLE)?|FROM|FULL|FUNCTION|GEOMETRY(?:COLLECTION)?|GLOBAL|GOTO|GRANT|GROUP|HANDLER|HASH|HAVING|HOLDLOCK|HOUR|IDENTITY(?:COL|_INSERT)?|IF|IGNORE|IMPORT|INDEX|INFILE|INNER|INNODB|INOUT|INSERT|INT|INTEGER|INTERSECT|INTERVAL|INTO|INVOKER|ISOLATION|ITERATE|JOIN|KEYS?|KILL|LANGUAGE|LAST|LEAVE|LEFT|LEVEL|LIMIT|LINENO|LINES|LINESTRING|LOAD|LOCAL|LOCK|LONG(?:BLOB|TEXT)|LOOP|MATCH(?:ED)?|MEDIUM(?:BLOB|INT|TEXT)|MERGE|MIDDLEINT|MINUTE|MODE|MODIFIES|MODIFY|MONTH|MULTI(?:LINESTRING|POINT|POLYGON)|NATIONAL|NATURAL|NCHAR|NEXT|NO|NONCLUSTERED|NULLIF|NUMERIC|OFF?|OFFSETS?|ON|OPEN(?:DATASOURCE|QUERY|ROWSET)?|OPTIMIZE|OPTION(?:ALLY)?|ORDER|OUT(?:ER|FILE)?|OVER|PARTIAL|PARTITION|PERCENT|PIVOT|PLAN|POINT|POLYGON|PRECEDING|PRECISION|PREPARE|PREV|PRIMARY|PRINT|PRIVILEGES|PROC(?:EDURE)?|PUBLIC|PURGE|QUICK|RAISERROR|READS?|REAL|RECONFIGURE|REFERENCES|RELEASE|RENAME|REPEAT(?:ABLE)?|REPLACE|REPLICATION|REQUIRE|RESIGNAL|RESTORE|RESTRICT|RETURN(?:ING|S)?|REVOKE|RIGHT|ROLLBACK|ROUTINE|ROW(?:COUNT|GUIDCOL|S)?|RTREE|RULE|SAVE(?:POINT)?|SCHEMA|SECOND|SELECT|SERIAL(?:IZABLE)?|SESSION(?:_USER)?|SET(?:USER)?|SHARE|SHOW|SHUTDOWN|SIMPLE|SMALLINT|SNAPSHOT|SOME|SONAME|SQL|START(?:ING)?|STATISTICS|STATUS|STRIPED|SYSTEM_USER|TABLES?|TABLESPACE|TEMP(?:ORARY|TABLE)?|TERMINATED|TEXT(?:SIZE)?|THEN|TIME(?:STAMP)?|TINY(?:BLOB|INT|TEXT)|TOP?|TRAN(?:SACTIONS?)?|TRIGGER|TRUNCATE|TSEQUAL|TYPES?|UNBOUNDED|UNCOMMITTED|UNDEFINED|UNION|UNIQUE|UNLOCK|UNPIVOT|UNSIGNED|UPDATE(?:TEXT)?|USAGE|USE|USER|USING|VALUES?|VAR(?:BINARY|CHAR|CHARACTER|YING)|VIEW|WAITFOR|WARNINGS|WHEN|WHERE|WHILE|WITH(?: ROLLUP|IN)?|WORK|WRITE(?:TEXT)?|YEAR)\b/i,boolean:/\b(?:FALSE|NULL|TRUE)\b/i,number:/\b0x[\da-f]+\b|\b\d+(?:\.\d*)?|\B\.\d+\b/i,operator:/[-+*\/=%^~]|&&?|\|\|?|!=?|<(?:=>?|<|>)?|>[>=]?|\b(?:AND|BETWEEN|DIV|ILIKE|IN|IS|LIKE|NOT|OR|REGEXP|RLIKE|SOUNDS LIKE|XOR)\b/i,punctuation:/[;[\]()`,.]/}}return sBe}var cBe,KUt;function o$r(){if(KUt)return cBe;KUt=1;var n=wQe();cBe=r,r.displayName="apex",r.aliases=[];function r(o){o.register(n),function(s){var c=/\b(?:(?:after|before)(?=\s+[a-z])|abstract|activate|and|any|array|as|asc|autonomous|begin|bigdecimal|blob|boolean|break|bulk|by|byte|case|cast|catch|char|class|collect|commit|const|continue|currency|date|datetime|decimal|default|delete|desc|do|double|else|end|enum|exception|exit|export|extends|final|finally|float|for|from|get(?=\s*[{};])|global|goto|group|having|hint|if|implements|import|in|inner|insert|instanceof|int|integer|interface|into|join|like|limit|list|long|loop|map|merge|new|not|null|nulls|number|object|of|on|or|outer|override|package|parallel|pragma|private|protected|public|retrieve|return|rollback|select|set|short|sObject|sort|static|string|super|switch|synchronized|system|testmethod|then|this|throw|time|transaction|transient|trigger|try|undelete|update|upsert|using|virtual|void|webservice|when|where|while|(?:inherited|with|without)\s+sharing)\b/i,d=/\b(?:(?=[a-z_]\w*\s*[<\[])|(?!<keyword>))[A-Z_]\w*(?:\s*\.\s*[A-Z_]\w*)*\b(?:\s*(?:\[\s*\]|<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>))*/.source.replace(/<keyword>/g,function(){return c.source});function h(y){return RegExp(y.replace(/<CLASS-NAME>/g,function(){return d}),"i")}var b={keyword:c,punctuation:/[()\[\]{};,:.<>]/};s.languages.apex={comment:s.languages.clike.comment,string:s.languages.clike.string,sql:{pattern:/((?:[=,({:]|\breturn)\s*)\[[^\[\]]*\]/i,lookbehind:!0,greedy:!0,alias:"language-sql",inside:s.languages.sql},annotation:{pattern:/@\w+\b/,alias:"punctuation"},"class-name":[{pattern:h(/(\b(?:class|enum|extends|implements|instanceof|interface|new|trigger\s+\w+\s+on)\s+)<CLASS-NAME>/.source),lookbehind:!0,inside:b},{pattern:h(/(\(\s*)<CLASS-NAME>(?=\s*\)\s*[\w(])/.source),lookbehind:!0,inside:b},{pattern:h(/<CLASS-NAME>(?=\s*\w+\s*[;=,(){:])/.source),inside:b}],trigger:{pattern:/(\btrigger\s+)\w+\b/i,lookbehind:!0,alias:"class-name"},keyword:c,function:/\b[a-z_]\w*(?=\s*\()/i,boolean:/\b(?:false|true)\b/i,number:/(?:\B\.\d+|\b\d+(?:\.\d+|L)?)\b/i,operator:/[!=](?:==?)?|\?\.?|&&|\|\||--|\+\+|[-+*/^&|]=?|:|<<?=?|>{1,3}=?/,punctuation:/[()\[\]{};,.]/}}(o)}return cBe}var uBe,YUt;function a$r(){if(YUt)return uBe;YUt=1,uBe=n,n.displayName="apl",n.aliases=[];function n(r){r.languages.apl={comment:/(?:⍝|#[! ]).*$/m,string:{pattern:/'(?:[^'\r\n]|'')*'/,greedy:!0},number:/¯?(?:\d*\.?\b\d+(?:e[+¯]?\d+)?|¯|∞)(?:j¯?(?:(?:\d+(?:\.\d+)?|\.\d+)(?:e[+¯]?\d+)?|¯|∞))?/i,statement:/:[A-Z][a-z][A-Za-z]*\b/,"system-function":{pattern:/⎕[A-Z]+/i,alias:"function"},constant:/[⍬⌾#⎕⍞]/,function:/[-+×÷⌈⌊∣|⍳⍸?*⍟○!⌹<≤=>≥≠≡≢∊⍷∪∩~∨∧⍱⍲⍴,⍪⌽⊖⍉↑↓⊂⊃⊆⊇⌷⍋⍒⊤⊥⍕⍎⊣⊢⍁⍂≈⍯↗¤→]/,"monadic-operator":{pattern:/[\\\/⌿⍀¨⍨⌶&∥]/,alias:"operator"},"dyadic-operator":{pattern:/[.⍣⍠⍤∘⌸@⌺⍥]/,alias:"operator"},assignment:{pattern:/←/,alias:"keyword"},punctuation:/[\[;\]()◇⋄]/,dfn:{pattern:/[{}⍺⍵⍶⍹∇⍫:]/,alias:"builtin"}}}return uBe}var lBe,XUt;function s$r(){if(XUt)return lBe;XUt=1,lBe=n,n.displayName="applescript",n.aliases=[];function n(r){r.languages.applescript={comment:[/\(\*(?:\(\*(?:[^*]|\*(?!\)))*\*\)|(?!\(\*)[\s\S])*?\*\)/,/--.+/,/#.+/],string:/"(?:\\.|[^"\\\r\n])*"/,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e-?\d+)?\b/i,operator:[/[&=≠≤≥*+\-\/÷^]|[<>]=?/,/\b(?:(?:begin|end|start)s? with|(?:contains?|(?:does not|doesn't) contain)|(?:is|isn't|is not) (?:contained by|in)|(?:(?:is|isn't|is not) )?(?:greater|less) than(?: or equal)?(?: to)?|(?:comes|(?:does not|doesn't) come) (?:after|before)|(?:is|isn't|is not) equal(?: to)?|(?:(?:does not|doesn't) equal|equal to|equals|is not|isn't)|(?:a )?(?:ref(?: to)?|reference to)|(?:and|as|div|mod|not|or))\b/],keyword:/\b(?:about|above|after|against|apart from|around|aside from|at|back|before|beginning|behind|below|beneath|beside|between|but|by|considering|continue|copy|does|eighth|else|end|equal|error|every|exit|false|fifth|first|for|fourth|from|front|get|given|global|if|ignoring|in|instead of|into|is|it|its|last|local|me|middle|my|ninth|of|on|onto|out of|over|prop|property|put|repeat|return|returning|second|set|seventh|since|sixth|some|tell|tenth|that|the|then|third|through|thru|timeout|times|to|transaction|true|try|until|where|while|whose|with|without)\b/,"class-name":/\b(?:POSIX file|RGB color|alias|application|boolean|centimeters|centimetres|class|constant|cubic centimeters|cubic centimetres|cubic feet|cubic inches|cubic meters|cubic metres|cubic yards|date|degrees Celsius|degrees Fahrenheit|degrees Kelvin|feet|file|gallons|grams|inches|integer|kilograms|kilometers|kilometres|list|liters|litres|meters|metres|miles|number|ounces|pounds|quarts|real|record|reference|script|square feet|square kilometers|square kilometres|square meters|square metres|square miles|square yards|text|yards)\b/,punctuation:/[{}():,¬«»《》]/}}return lBe}var dBe,JUt;function c$r(){if(JUt)return dBe;JUt=1,dBe=n,n.displayName="aql",n.aliases=[];function n(r){r.languages.aql={comment:/\/\/.*|\/\*[\s\S]*?\*\//,property:{pattern:/([{,]\s*)(?:(?!\d)\w+|(["'´`])(?:(?!\2)[^\\\r\n]|\\.)*\2)(?=\s*:)/,lookbehind:!0,greedy:!0},string:{pattern:/(["'])(?:(?!\1)[^\\\r\n]|\\.)*\1/,greedy:!0},identifier:{pattern:/([´`])(?:(?!\1)[^\\\r\n]|\\.)*\1/,greedy:!0},variable:/@@?\w+/,keyword:[{pattern:/(\bWITH\s+)COUNT(?=\s+INTO\b)/i,lookbehind:!0},/\b(?:AGGREGATE|ALL|AND|ANY|ASC|COLLECT|DESC|DISTINCT|FILTER|FOR|GRAPH|IN|INBOUND|INSERT|INTO|K_PATHS|K_SHORTEST_PATHS|LET|LIKE|LIMIT|NONE|NOT|NULL|OR|OUTBOUND|REMOVE|REPLACE|RETURN|SHORTEST_PATH|SORT|UPDATE|UPSERT|WINDOW|WITH)\b/i,{pattern:/(^|[^\w.[])(?:KEEP|PRUNE|SEARCH|TO)\b/i,lookbehind:!0},{pattern:/(^|[^\w.[])(?:CURRENT|NEW|OLD)\b/,lookbehind:!0},{pattern:/\bOPTIONS(?=\s*\{)/i}],function:/\b(?!\d)\w+(?=\s*\()/,boolean:/\b(?:false|true)\b/i,range:{pattern:/\.\./,alias:"operator"},number:[/\b0b[01]+/i,/\b0x[0-9a-f]+/i,/(?:\B\.\d+|\b(?:0|[1-9]\d*)(?:\.\d+)?)(?:e[+-]?\d+)?/i],operator:/\*{2,}|[=!]~|[!=<>]=?|&&|\|\||[-+*/%]/,punctuation:/::|[?.:,;()[\]{}]/}}return dBe}var fBe,ZUt;function vU(){if(ZUt)return fBe;ZUt=1,fBe=n,n.displayName="c",n.aliases=[];function n(r){r.languages.c=r.languages.extend("clike",{comment:{pattern:/\/\/(?:[^\r\n\\]|\\(?:\r\n?|\n|(?![\r\n])))*|\/\*[\s\S]*?(?:\*\/|$)/,greedy:!0},string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0},"class-name":{pattern:/(\b(?:enum|struct)\s+(?:__attribute__\s*\(\([\s\S]*?\)\)\s*)?)\w+|\b[a-z]\w*_t\b/,lookbehind:!0},keyword:/\b(?:_Alignas|_Alignof|_Atomic|_Bool|_Complex|_Generic|_Imaginary|_Noreturn|_Static_assert|_Thread_local|__attribute__|asm|auto|break|case|char|const|continue|default|do|double|else|enum|extern|float|for|goto|if|inline|int|long|register|return|short|signed|sizeof|static|struct|switch|typedef|typeof|union|unsigned|void|volatile|while)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,number:/(?:\b0x(?:[\da-f]+(?:\.[\da-f]*)?|\.[\da-f]+)(?:p[+-]?\d+)?|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)[ful]{0,4}/i,operator:/>>=?|<<=?|->|([-+&|:])\1|[?:~]|[-+*/%&|^!=<>]=?/}),r.languages.insertBefore("c","string",{char:{pattern:/'(?:\\(?:\r\n|[\s\S])|[^'\\\r\n]){0,32}'/,greedy:!0}}),r.languages.insertBefore("c","string",{macro:{pattern:/(^[\t ]*)#\s*[a-z](?:[^\r\n\\/]|\/(?!\*)|\/\*(?:[^*]|\*(?!\/))*\*\/|\\(?:\r\n|[\s\S]))*/im,lookbehind:!0,greedy:!0,alias:"property",inside:{string:[{pattern:/^(#\s*include\s*)<[^>]+>/,lookbehind:!0},r.languages.c.string],char:r.languages.c.char,comment:r.languages.c.comment,"macro-name":[{pattern:/(^#\s*define\s+)\w+\b(?!\()/i,lookbehind:!0},{pattern:/(^#\s*define\s+)\w+\b(?=\()/i,lookbehind:!0,alias:"function"}],directive:{pattern:/^(#\s*)[a-z]+/,lookbehind:!0,alias:"keyword"},"directive-hash":/^#/,punctuation:/##|\\(?=[\r\n])/,expression:{pattern:/\S[\s\S]*/,inside:r.languages.c}}}}),r.languages.insertBefore("c","function",{constant:/\b(?:EOF|NULL|SEEK_CUR|SEEK_END|SEEK_SET|__DATE__|__FILE__|__LINE__|__TIMESTAMP__|__TIME__|__func__|stderr|stdin|stdout)\b/}),delete r.languages.c.boolean}return fBe}var hBe,QUt;function yQe(){if(QUt)return hBe;QUt=1;var n=vU();hBe=r,r.displayName="cpp",r.aliases=[];function r(o){o.register(n),function(s){var c=/\b(?:alignas|alignof|asm|auto|bool|break|case|catch|char|char16_t|char32_t|char8_t|class|co_await|co_return|co_yield|compl|concept|const|const_cast|consteval|constexpr|constinit|continue|decltype|default|delete|do|double|dynamic_cast|else|enum|explicit|export|extern|final|float|for|friend|goto|if|import|inline|int|int16_t|int32_t|int64_t|int8_t|long|module|mutable|namespace|new|noexcept|nullptr|operator|override|private|protected|public|register|reinterpret_cast|requires|return|short|signed|sizeof|static|static_assert|static_cast|struct|switch|template|this|thread_local|throw|try|typedef|typeid|typename|uint16_t|uint32_t|uint64_t|uint8_t|union|unsigned|using|virtual|void|volatile|wchar_t|while)\b/,d=/\b(?!<keyword>)\w+(?:\s*\.\s*\w+)*\b/.source.replace(/<keyword>/g,function(){return c.source});s.languages.cpp=s.languages.extend("c",{"class-name":[{pattern:RegExp(/(\b(?:class|concept|enum|struct|typename)\s+)(?!<keyword>)\w+/.source.replace(/<keyword>/g,function(){return c.source})),lookbehind:!0},/\b[A-Z]\w*(?=\s*::\s*\w+\s*\()/,/\b[A-Z_]\w*(?=\s*::\s*~\w+\s*\()/i,/\b\w+(?=\s*<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>\s*::\s*\w+\s*\()/],keyword:c,number:{pattern:/(?:\b0b[01']+|\b0x(?:[\da-f']+(?:\.[\da-f']*)?|\.[\da-f']+)(?:p[+-]?[\d']+)?|(?:\b[\d']+(?:\.[\d']*)?|\B\.[\d']+)(?:e[+-]?[\d']+)?)[ful]{0,4}/i,greedy:!0},operator:/>>=?|<<=?|->|--|\+\+|&&|\|\||[?:~]|<=>|[-+*/%&|^!=<>]=?|\b(?:and|and_eq|bitand|bitor|not|not_eq|or|or_eq|xor|xor_eq)\b/,boolean:/\b(?:false|true)\b/}),s.languages.insertBefore("cpp","string",{module:{pattern:RegExp(/(\b(?:import|module)\s+)/.source+"(?:"+/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"|<[^<>\r\n]*>/.source+"|"+/<mod-name>(?:\s*:\s*<mod-name>)?|:\s*<mod-name>/.source.replace(/<mod-name>/g,function(){return d})+")"),lookbehind:!0,greedy:!0,inside:{string:/^[<"][\s\S]+/,operator:/:/,punctuation:/\./}},"raw-string":{pattern:/R"([^()\\ ]{0,16})\([\s\S]*?\)\1"/,alias:"string",greedy:!0}}),s.languages.insertBefore("cpp","keyword",{"generic-function":{pattern:/\b(?!operator\b)[a-z_]\w*\s*<(?:[^<>]|<[^<>]*>)*>(?=\s*\()/i,inside:{function:/^\w+/,generic:{pattern:/<[\s\S]+/,alias:"class-name",inside:s.languages.cpp}}}}),s.languages.insertBefore("cpp","operator",{"double-colon":{pattern:/::/,alias:"punctuation"}}),s.languages.insertBefore("cpp","class-name",{"base-clause":{pattern:/(\b(?:class|struct)\s+\w+\s*:\s*)[^;{}"'\s]+(?:\s+[^;{}"'\s]+)*(?=\s*[;{])/,lookbehind:!0,greedy:!0,inside:s.languages.extend("cpp",{})}}),s.languages.insertBefore("inside","double-colon",{"class-name":/\b[a-z_]\w*\b(?!\s*::)/i},s.languages.cpp["base-clause"])}(o)}return hBe}var pBe,ezt;function u$r(){if(ezt)return pBe;ezt=1;var n=yQe();pBe=r,r.displayName="arduino",r.aliases=["ino"];function r(o){o.register(n),o.languages.arduino=o.languages.extend("cpp",{keyword:/\b(?:String|array|bool|boolean|break|byte|case|catch|continue|default|do|double|else|finally|for|function|goto|if|in|instanceof|int|integer|long|loop|new|null|return|setup|string|switch|throw|try|void|while|word)\b/,constant:/\b(?:ANALOG_MESSAGE|DEFAULT|DIGITAL_MESSAGE|EXTERNAL|FIRMATA_STRING|HIGH|INPUT|INPUT_PULLUP|INTERNAL|INTERNAL1V1|INTERNAL2V56|LED_BUILTIN|LOW|OUTPUT|REPORT_ANALOG|REPORT_DIGITAL|SET_PIN_MODE|SYSEX_START|SYSTEM_RESET)\b/,builtin:/\b(?:Audio|BSSID|Bridge|Client|Console|EEPROM|Esplora|EsploraTFT|Ethernet|EthernetClient|EthernetServer|EthernetUDP|File|FileIO|FileSystem|Firmata|GPRS|GSM|GSMBand|GSMClient|GSMModem|GSMPIN|GSMScanner|GSMServer|GSMVoiceCall|GSM_SMS|HttpClient|IPAddress|IRread|Keyboard|KeyboardController|LiquidCrystal|LiquidCrystal_I2C|Mailbox|Mouse|MouseController|PImage|Process|RSSI|RobotControl|RobotMotor|SD|SPI|SSID|Scheduler|Serial|Server|Servo|SoftwareSerial|Stepper|Stream|TFT|Task|USBHost|WiFi|WiFiClient|WiFiServer|WiFiUDP|Wire|YunClient|YunServer|abs|addParameter|analogRead|analogReadResolution|analogReference|analogWrite|analogWriteResolution|answerCall|attach|attachGPRS|attachInterrupt|attached|autoscroll|available|background|beep|begin|beginPacket|beginSD|beginSMS|beginSpeaker|beginTFT|beginTransmission|beginWrite|bit|bitClear|bitRead|bitSet|bitWrite|blink|blinkVersion|buffer|changePIN|checkPIN|checkPUK|checkReg|circle|cityNameRead|cityNameWrite|clear|clearScreen|click|close|compassRead|config|connect|connected|constrain|cos|countryNameRead|countryNameWrite|createChar|cursor|debugPrint|delay|delayMicroseconds|detach|detachInterrupt|digitalRead|digitalWrite|disconnect|display|displayLogos|drawBMP|drawCompass|encryptionType|end|endPacket|endSMS|endTransmission|endWrite|exists|exitValue|fill|find|findUntil|flush|gatewayIP|get|getAsynchronously|getBand|getButton|getCurrentCarrier|getIMEI|getKey|getModifiers|getOemKey|getPINUsed|getResult|getSignalStrength|getSocket|getVoiceCallStatus|getXChange|getYChange|hangCall|height|highByte|home|image|interrupts|isActionDone|isDirectory|isListening|isPIN|isPressed|isValid|keyPressed|keyReleased|keyboardRead|knobRead|leftToRight|line|lineFollowConfig|listen|listenOnLocalhost|loadImage|localIP|lowByte|macAddress|maintain|map|max|messageAvailable|micros|millis|min|mkdir|motorsStop|motorsWrite|mouseDragged|mouseMoved|mousePressed|mouseReleased|move|noAutoscroll|noBlink|noBuffer|noCursor|noDisplay|noFill|noInterrupts|noListenOnLocalhost|noStroke|noTone|onReceive|onRequest|open|openNextFile|overflow|parseCommand|parseFloat|parseInt|parsePacket|pauseMode|peek|pinMode|playFile|playMelody|point|pointTo|position|pow|prepare|press|print|printFirmwareVersion|printVersion|println|process|processInput|pulseIn|put|random|randomSeed|read|readAccelerometer|readBlue|readButton|readBytes|readBytesUntil|readGreen|readJoystickButton|readJoystickSwitch|readJoystickX|readJoystickY|readLightSensor|readMessage|readMicrophone|readNetworks|readRed|readSlider|readString|readStringUntil|readTemperature|ready|rect|release|releaseAll|remoteIP|remoteNumber|remotePort|remove|requestFrom|retrieveCallingNumber|rewindDirectory|rightToLeft|rmdir|robotNameRead|robotNameWrite|run|runAsynchronously|runShellCommand|runShellCommandAsynchronously|running|scanNetworks|scrollDisplayLeft|scrollDisplayRight|seek|sendAnalog|sendDigitalPortPair|sendDigitalPorts|sendString|sendSysex|serialEvent|setBand|setBitOrder|setClockDivider|setCursor|setDNS|setDataMode|setFirmwareVersion|setMode|setPINUsed|setSpeed|setTextSize|setTimeout|shiftIn|shiftOut|shutdown|sin|size|sqrt|startLoop|step|stop|stroke|subnetMask|switchPIN|tan|tempoWrite|text|tone|transfer|tuneWrite|turn|updateIR|userNameRead|userNameWrite|voiceCall|waitContinue|width|write|writeBlue|writeGreen|writeJSON|writeMessage|writeMicroseconds|writeRGB|writeRed|yield)\b/}),o.languages.ino=o.languages.arduino}return pBe}var gBe,tzt;function l$r(){if(tzt)return gBe;tzt=1,gBe=n,n.displayName="arff",n.aliases=[];function n(r){r.languages.arff={comment:/%.*/,string:{pattern:/(["'])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:/@(?:attribute|data|end|relation)\b/i,number:/\b\d+(?:\.\d+)?\b/,punctuation:/[{},]/}}return gBe}var mBe,nzt;function d$r(){if(nzt)return mBe;nzt=1,mBe=n,n.displayName="asciidoc",n.aliases=["adoc"];function n(r){(function(o){var s={pattern:/(^[ \t]*)\[(?!\[)(?:(["'$`])(?:(?!\2)[^\\]|\\.)*\2|\[(?:[^\[\]\\]|\\.)*\]|[^\[\]\\"'$`]|\\.)*\]/m,lookbehind:!0,inside:{quoted:{pattern:/([$`])(?:(?!\1)[^\\]|\\.)*\1/,inside:{punctuation:/^[$`]|[$`]$/}},interpreted:{pattern:/'(?:[^'\\]|\\.)*'/,inside:{punctuation:/^'|'$/}},string:/"(?:[^"\\]|\\.)*"/,variable:/\w+(?==)/,punctuation:/^\[|\]$|,/,operator:/=/,"attr-value":/(?!^\s+$).+/}},c=o.languages.asciidoc={"comment-block":{pattern:/^(\/{4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1/m,alias:"comment"},table:{pattern:/^\|={3,}(?:(?:\r?\n|\r(?!\n)).*)*?(?:\r?\n|\r)\|={3,}$/m,inside:{specifiers:{pattern:/(?:(?:(?:\d+(?:\.\d+)?|\.\d+)[+*](?:[<^>](?:\.[<^>])?|\.[<^>])?|[<^>](?:\.[<^>])?|\.[<^>])[a-z]*|[a-z]+)(?=\|)/,alias:"attr-value"},punctuation:{pattern:/(^|[^\\])[|!]=*/,lookbehind:!0}}},"passthrough-block":{pattern:/^(\+{4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1$/m,inside:{punctuation:/^\++|\++$/}},"literal-block":{pattern:/^(-{4,}|\.{4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1$/m,inside:{punctuation:/^(?:-+|\.+)|(?:-+|\.+)$/}},"other-block":{pattern:/^(--|\*{4,}|_{4,}|={4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1$/m,inside:{punctuation:/^(?:-+|\*+|_+|=+)|(?:-+|\*+|_+|=+)$/}},"list-punctuation":{pattern:/(^[ \t]*)(?:-|\*{1,5}|\.{1,5}|(?:[a-z]|\d+)\.|[xvi]+\))(?= )/im,lookbehind:!0,alias:"punctuation"},"list-label":{pattern:/(^[ \t]*)[a-z\d].+(?::{2,4}|;;)(?=\s)/im,lookbehind:!0,alias:"symbol"},"indented-block":{pattern:/((\r?\n|\r)\2)([ \t]+)\S.*(?:(?:\r?\n|\r)\3.+)*(?=\2{2}|$)/,lookbehind:!0},comment:/^\/\/.*/m,title:{pattern:/^.+(?:\r?\n|\r)(?:={3,}|-{3,}|~{3,}|\^{3,}|\+{3,})$|^={1,5} .+|^\.(?![\s.]).*/m,alias:"important",inside:{punctuation:/^(?:\.|=+)|(?:=+|-+|~+|\^+|\++)$/}},"attribute-entry":{pattern:/^:[^:\r\n]+:(?: .*?(?: \+(?:\r?\n|\r).*?)*)?$/m,alias:"tag"},attributes:s,hr:{pattern:/^'{3,}$/m,alias:"punctuation"},"page-break":{pattern:/^<{3,}$/m,alias:"punctuation"},admonition:{pattern:/^(?:CAUTION|IMPORTANT|NOTE|TIP|WARNING):/m,alias:"keyword"},callout:[{pattern:/(^[ \t]*)<?\d*>/m,lookbehind:!0,alias:"symbol"},{pattern:/<\d+>/,alias:"symbol"}],macro:{pattern:/\b[a-z\d][a-z\d-]*::?(?:[^\s\[\]]*\[(?:[^\]\\"']|(["'])(?:(?!\1)[^\\]|\\.)*\1|\\.)*\])/,inside:{function:/^[a-z\d-]+(?=:)/,punctuation:/^::?/,attributes:{pattern:/(?:\[(?:[^\]\\"']|(["'])(?:(?!\1)[^\\]|\\.)*\1|\\.)*\])/,inside:s.inside}}},inline:{pattern:/(^|[^\\])(?:(?:\B\[(?:[^\]\\"']|(["'])(?:(?!\2)[^\\]|\\.)*\2|\\.)*\])?(?:\b_(?!\s)(?: _|[^_\\\r\n]|\\.)+(?:(?:\r?\n|\r)(?: _|[^_\\\r\n]|\\.)+)*_\b|\B``(?!\s).+?(?:(?:\r?\n|\r).+?)*''\B|\B`(?!\s)(?:[^`'\s]|\s+\S)+['`]\B|\B(['*+#])(?!\s)(?: \3|(?!\3)[^\\\r\n]|\\.)+(?:(?:\r?\n|\r)(?: \3|(?!\3)[^\\\r\n]|\\.)+)*\3\B)|(?:\[(?:[^\]\\"']|(["'])(?:(?!\4)[^\\]|\\.)*\4|\\.)*\])?(?:(__|\*\*|\+\+\+?|##|\$\$|[~^]).+?(?:(?:\r?\n|\r).+?)*\5|\{[^}\r\n]+\}|\[\[\[?.+?(?:(?:\r?\n|\r).+?)*\]?\]\]|<<.+?(?:(?:\r?\n|\r).+?)*>>|\(\(\(?.+?(?:(?:\r?\n|\r).+?)*\)?\)\)))/m,lookbehind:!0,inside:{attributes:s,url:{pattern:/^(?:\[\[\[?.+?\]?\]\]|<<.+?>>)$/,inside:{punctuation:/^(?:\[\[\[?|<<)|(?:\]\]\]?|>>)$/}},"attribute-ref":{pattern:/^\{.+\}$/,inside:{variable:{pattern:/(^\{)[a-z\d,+_-]+/,lookbehind:!0},operator:/^[=?!#%@$]|!(?=[:}])/,punctuation:/^\{|\}$|::?/}},italic:{pattern:/^(['_])[\s\S]+\1$/,inside:{punctuation:/^(?:''?|__?)|(?:''?|__?)$/}},bold:{pattern:/^\*[\s\S]+\*$/,inside:{punctuation:/^\*\*?|\*\*?$/}},punctuation:/^(?:``?|\+{1,3}|##?|\$\$|[~^]|\(\(\(?)|(?:''?|\+{1,3}|##?|\$\$|[~^`]|\)?\)\))$/}},replacement:{pattern:/\((?:C|R|TM)\)/,alias:"builtin"},entity:/&#?[\da-z]{1,8};/i,"line-continuation":{pattern:/(^| )\+$/m,lookbehind:!0,alias:"punctuation"}};function d(h){h=h.split(" ");for(var b={},y=0,E=h.length;y<E;y++)b[h[y]]=c[h[y]];return b}s.inside.interpreted.inside.rest=d("macro inline replacement entity"),c["passthrough-block"].inside.rest=d("macro"),c["literal-block"].inside.rest=d("callout"),c.table.inside.rest=d("comment-block passthrough-block literal-block other-block list-punctuation indented-block comment title attribute-entry attributes hr page-break admonition list-label callout macro inline replacement entity line-continuation"),c["other-block"].inside.rest=d("table list-punctuation indented-block comment attribute-entry attributes hr page-break admonition list-label macro inline replacement entity line-continuation"),c.title.inside.rest=d("macro inline replacement entity"),o.hooks.add("wrap",function(h){h.type==="entity"&&(h.attributes.title=h.content.value.replace(/&amp;/,"&"))}),o.languages.adoc=o.languages.asciidoc})(r)}return mBe}var bBe,rzt;function f$r(){if(rzt)return bBe;rzt=1,bBe=n,n.displayName="asm6502",n.aliases=[];function n(r){r.languages.asm6502={comment:/;.*/,directive:{pattern:/\.\w+(?= )/,alias:"property"},string:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,"op-code":{pattern:/\b(?:ADC|AND|ASL|BCC|BCS|BEQ|BIT|BMI|BNE|BPL|BRK|BVC|BVS|CLC|CLD|CLI|CLV|CMP|CPX|CPY|DEC|DEX|DEY|EOR|INC|INX|INY|JMP|JSR|LDA|LDX|LDY|LSR|NOP|ORA|PHA|PHP|PLA|PLP|ROL|ROR|RTI|RTS|SBC|SEC|SED|SEI|STA|STX|STY|TAX|TAY|TSX|TXA|TXS|TYA|adc|and|asl|bcc|bcs|beq|bit|bmi|bne|bpl|brk|bvc|bvs|clc|cld|cli|clv|cmp|cpx|cpy|dec|dex|dey|eor|inc|inx|iny|jmp|jsr|lda|ldx|ldy|lsr|nop|ora|pha|php|pla|plp|rol|ror|rti|rts|sbc|sec|sed|sei|sta|stx|sty|tax|tay|tsx|txa|txs|tya)\b/,alias:"keyword"},"hex-number":{pattern:/#?\$[\da-f]{1,4}\b/i,alias:"number"},"binary-number":{pattern:/#?%[01]+\b/,alias:"number"},"decimal-number":{pattern:/#?\b\d+\b/,alias:"number"},register:{pattern:/\b[xya]\b/i,alias:"variable"},punctuation:/[(),:]/}}return bBe}var vBe,izt;function h$r(){if(izt)return vBe;izt=1,vBe=n,n.displayName="asmatmel",n.aliases=[];function n(r){r.languages.asmatmel={comment:{pattern:/;.*/,greedy:!0},string:{pattern:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},constant:/\b(?:PORT[A-Z]|DDR[A-Z]|(?:DD|P)[A-Z](?:\d|[0-2]\d|3[01]))\b/,directive:{pattern:/\.\w+(?= )/,alias:"property"},"r-register":{pattern:/\br(?:\d|[12]\d|3[01])\b/,alias:"variable"},"op-code":{pattern:/\b(?:ADC|ADD|ADIW|AND|ANDI|ASR|BCLR|BLD|BRBC|BRBS|BRCC|BRCS|BREAK|BREQ|BRGE|BRHC|BRHS|BRID|BRIE|BRLO|BRLT|BRMI|BRNE|BRPL|BRSH|BRTC|BRTS|BRVC|BRVS|BSET|BST|CALL|CBI|CBR|CLC|CLH|CLI|CLN|CLR|CLS|CLT|CLV|CLZ|COM|CP|CPC|CPI|CPSE|DEC|DES|EICALL|EIJMP|ELPM|EOR|FMUL|FMULS|FMULSU|ICALL|IJMP|IN|INC|JMP|LAC|LAS|LAT|LD|LD[A-Za-z0-9]|LPM|LSL|LSR|MOV|MOVW|MUL|MULS|MULSU|NEG|NOP|OR|ORI|OUT|POP|PUSH|RCALL|RET|RETI|RJMP|ROL|ROR|SBC|SBCI|SBI|SBIC|SBIS|SBIW|SBR|SBRC|SBRS|SEC|SEH|SEI|SEN|SER|SES|SET|SEV|SEZ|SLEEP|SPM|ST|ST[A-Z0-9]|SUB|SUBI|SWAP|TST|WDR|XCH|adc|add|adiw|and|andi|asr|bclr|bld|brbc|brbs|brcc|brcs|break|breq|brge|brhc|brhs|brid|brie|brlo|brlt|brmi|brne|brpl|brsh|brtc|brts|brvc|brvs|bset|bst|call|cbi|cbr|clc|clh|cli|cln|clr|cls|clt|clv|clz|com|cp|cpc|cpi|cpse|dec|des|eicall|eijmp|elpm|eor|fmul|fmuls|fmulsu|icall|ijmp|in|inc|jmp|lac|las|lat|ld|ld[a-z0-9]|lpm|lsl|lsr|mov|movw|mul|muls|mulsu|neg|nop|or|ori|out|pop|push|rcall|ret|reti|rjmp|rol|ror|sbc|sbci|sbi|sbic|sbis|sbiw|sbr|sbrc|sbrs|sec|seh|sei|sen|ser|ses|set|sev|sez|sleep|spm|st|st[a-zA-Z0-9]|sub|subi|swap|tst|wdr|xch)\b/,alias:"keyword"},"hex-number":{pattern:/#?\$[\da-f]{2,4}\b/i,alias:"number"},"binary-number":{pattern:/#?%[01]+\b/,alias:"number"},"decimal-number":{pattern:/#?\b\d+\b/,alias:"number"},register:{pattern:/\b[acznvshtixy]\b/i,alias:"variable"},operator:/>>=?|<<=?|&&?|\|\|?|[-+*/%&|^!=<>?]=?/,punctuation:/[(),:]/}}return vBe}var wBe,ozt;function REe(){if(ozt)return wBe;ozt=1,wBe=n,n.displayName="csharp",n.aliases=["dotnet","cs"];function n(r){(function(o){function s(We,ht){return We.replace(/<<(\d+)>>/g,function(Ie,$t){return"(?:"+ht[+$t]+")"})}function c(We,ht,Ie){return RegExp(s(We,ht),"")}function d(We,ht){for(var Ie=0;Ie<ht;Ie++)We=We.replace(/<<self>>/g,function(){return"(?:"+We+")"});return We.replace(/<<self>>/g,"[^\\s\\S]")}var h={type:"bool byte char decimal double dynamic float int long object sbyte short string uint ulong ushort var void",typeDeclaration:"class enum interface record struct",contextual:"add alias and ascending async await by descending from(?=\\s*(?:\\w|$)) get global group into init(?=\\s*;) join let nameof not notnull on or orderby partial remove select set unmanaged value when where with(?=\\s*{)",other:"abstract as base break case catch checked const continue default delegate do else event explicit extern finally fixed for foreach goto if implicit in internal is lock namespace new null operator out override params private protected public readonly ref return sealed sizeof stackalloc static switch this throw try typeof unchecked unsafe using virtual volatile while yield"};function b(We){return"\\b(?:"+We.trim().replace(/ /g,"|")+")\\b"}var y=b(h.typeDeclaration),E=RegExp(b(h.type+" "+h.typeDeclaration+" "+h.contextual+" "+h.other)),k=b(h.typeDeclaration+" "+h.contextual+" "+h.other),x=b(h.type+" "+h.typeDeclaration+" "+h.other),O=d(/<(?:[^<>;=+\-*/%&|^]|<<self>>)*>/.source,2),P=d(/\((?:[^()]|<<self>>)*\)/.source,2),D=/@?\b[A-Za-z_]\w*\b/.source,$=s(/<<0>>(?:\s*<<1>>)?/.source,[D,O]),U=s(/(?!<<0>>)<<1>>(?:\s*\.\s*<<1>>)*/.source,[k,$]),B=/\[\s*(?:,\s*)*\]/.source,G=s(/<<0>>(?:\s*(?:\?\s*)?<<1>>)*(?:\s*\?)?/.source,[U,B]),K=s(/[^,()<>[\];=+\-*/%&|^]|<<0>>|<<1>>|<<2>>/.source,[O,P,B]),X=s(/\(<<0>>+(?:,<<0>>+)+\)/.source,[K]),ae=s(/(?:<<0>>|<<1>>)(?:\s*(?:\?\s*)?<<2>>)*(?:\s*\?)?/.source,[X,U,B]),oe={keyword:E,punctuation:/[<>()?,.:[\]]/},he=/'(?:[^\r\n'\\]|\\.|\\[Uux][\da-fA-F]{1,8})'/.source,ye=/"(?:\\.|[^\\"\r\n])*"/.source,we=/@"(?:""|\\[\s\S]|[^\\"])*"(?!")/.source;o.languages.csharp=o.languages.extend("clike",{string:[{pattern:c(/(^|[^$\\])<<0>>/.source,[we]),lookbehind:!0,greedy:!0},{pattern:c(/(^|[^@$\\])<<0>>/.source,[ye]),lookbehind:!0,greedy:!0}],"class-name":[{pattern:c(/(\busing\s+static\s+)<<0>>(?=\s*;)/.source,[U]),lookbehind:!0,inside:oe},{pattern:c(/(\busing\s+<<0>>\s*=\s*)<<1>>(?=\s*;)/.source,[D,ae]),lookbehind:!0,inside:oe},{pattern:c(/(\busing\s+)<<0>>(?=\s*=)/.source,[D]),lookbehind:!0},{pattern:c(/(\b<<0>>\s+)<<1>>/.source,[y,$]),lookbehind:!0,inside:oe},{pattern:c(/(\bcatch\s*\(\s*)<<0>>/.source,[U]),lookbehind:!0,inside:oe},{pattern:c(/(\bwhere\s+)<<0>>/.source,[D]),lookbehind:!0},{pattern:c(/(\b(?:is(?:\s+not)?|as)\s+)<<0>>/.source,[G]),lookbehind:!0,inside:oe},{pattern:c(/\b<<0>>(?=\s+(?!<<1>>|with\s*\{)<<2>>(?:\s*[=,;:{)\]]|\s+(?:in|when)\b))/.source,[ae,x,D]),inside:oe}],keyword:E,number:/(?:\b0(?:x[\da-f_]*[\da-f]|b[01_]*[01])|(?:\B\.\d+(?:_+\d+)*|\b\d+(?:_+\d+)*(?:\.\d+(?:_+\d+)*)?)(?:e[-+]?\d+(?:_+\d+)*)?)(?:[dflmu]|lu|ul)?\b/i,operator:/>>=?|<<=?|[-=]>|([-+&|])\1|~|\?\?=?|[-+*/%&|^!=<>]=?/,punctuation:/\?\.?|::|[{}[\];(),.:]/}),o.languages.insertBefore("csharp","number",{range:{pattern:/\.\./,alias:"operator"}}),o.languages.insertBefore("csharp","punctuation",{"named-parameter":{pattern:c(/([(,]\s*)<<0>>(?=\s*:)/.source,[D]),lookbehind:!0,alias:"punctuation"}}),o.languages.insertBefore("csharp","class-name",{namespace:{pattern:c(/(\b(?:namespace|using)\s+)<<0>>(?:\s*\.\s*<<0>>)*(?=\s*[;{])/.source,[D]),lookbehind:!0,inside:{punctuation:/\./}},"type-expression":{pattern:c(/(\b(?:default|sizeof|typeof)\s*\(\s*(?!\s))(?:[^()\s]|\s(?!\s)|<<0>>)*(?=\s*\))/.source,[P]),lookbehind:!0,alias:"class-name",inside:oe},"return-type":{pattern:c(/<<0>>(?=\s+(?:<<1>>\s*(?:=>|[({]|\.\s*this\s*\[)|this\s*\[))/.source,[ae,U]),inside:oe,alias:"class-name"},"constructor-invocation":{pattern:c(/(\bnew\s+)<<0>>(?=\s*[[({])/.source,[ae]),lookbehind:!0,inside:oe,alias:"class-name"},"generic-method":{pattern:c(/<<0>>\s*<<1>>(?=\s*\()/.source,[D,O]),inside:{function:c(/^<<0>>/.source,[D]),generic:{pattern:RegExp(O),alias:"class-name",inside:oe}}},"type-list":{pattern:c(/\b((?:<<0>>\s+<<1>>|record\s+<<1>>\s*<<5>>|where\s+<<2>>)\s*:\s*)(?:<<3>>|<<4>>|<<1>>\s*<<5>>|<<6>>)(?:\s*,\s*(?:<<3>>|<<4>>|<<6>>))*(?=\s*(?:where|[{;]|=>|$))/.source,[y,$,D,ae,E.source,P,/\bnew\s*\(\s*\)/.source]),lookbehind:!0,inside:{"record-arguments":{pattern:c(/(^(?!new\s*\()<<0>>\s*)<<1>>/.source,[$,P]),lookbehind:!0,greedy:!0,inside:o.languages.csharp},keyword:E,"class-name":{pattern:RegExp(ae),greedy:!0,inside:oe},punctuation:/[,()]/}},preprocessor:{pattern:/(^[\t ]*)#.*/m,lookbehind:!0,alias:"property",inside:{directive:{pattern:/(#)\b(?:define|elif|else|endif|endregion|error|if|line|nullable|pragma|region|undef|warning)\b/,lookbehind:!0,alias:"keyword"}}}});var ge=ye+"|"+he,Te=s(/\/(?![*/])|\/\/[^\r\n]*[\r\n]|\/\*(?:[^*]|\*(?!\/))*\*\/|<<0>>/.source,[ge]),xe=d(s(/[^"'/()]|<<0>>|\(<<self>>*\)/.source,[Te]),2),Ue=/\b(?:assembly|event|field|method|module|param|property|return|type)\b/.source,Z=s(/<<0>>(?:\s*\(<<1>>*\))?/.source,[U,xe]);o.languages.insertBefore("csharp","class-name",{attribute:{pattern:c(/((?:^|[^\s\w>)?])\s*\[\s*)(?:<<0>>\s*:\s*)?<<1>>(?:\s*,\s*<<1>>)*(?=\s*\])/.source,[Ue,Z]),lookbehind:!0,greedy:!0,inside:{target:{pattern:c(/^<<0>>(?=\s*:)/.source,[Ue]),alias:"keyword"},"attribute-arguments":{pattern:c(/\(<<0>>*\)/.source,[xe]),inside:o.languages.csharp},"class-name":{pattern:RegExp(U),inside:{punctuation:/\./}},punctuation:/[:,]/}}});var Re=/:[^}\r\n]+/.source,_e=d(s(/[^"'/()]|<<0>>|\(<<self>>*\)/.source,[Te]),2),qe=s(/\{(?!\{)(?:(?![}:])<<0>>)*<<1>>?\}/.source,[_e,Re]),it=d(s(/[^"'/()]|\/(?!\*)|\/\*(?:[^*]|\*(?!\/))*\*\/|<<0>>|\(<<self>>*\)/.source,[ge]),2),Xe=s(/\{(?!\{)(?:(?![}:])<<0>>)*<<1>>?\}/.source,[it,Re]);function de(We,ht){return{interpolation:{pattern:c(/((?:^|[^{])(?:\{\{)*)<<0>>/.source,[We]),lookbehind:!0,inside:{"format-string":{pattern:c(/(^\{(?:(?![}:])<<0>>)*)<<1>>(?=\}$)/.source,[ht,Re]),lookbehind:!0,inside:{punctuation:/^:/}},punctuation:/^\{|\}$/,expression:{pattern:/[\s\S]+/,alias:"language-csharp",inside:o.languages.csharp}}},string:/[\s\S]+/}}o.languages.insertBefore("csharp","string",{"interpolation-string":[{pattern:c(/(^|[^\\])(?:\$@|@\$)"(?:""|\\[\s\S]|\{\{|<<0>>|[^\\{"])*"/.source,[qe]),lookbehind:!0,greedy:!0,inside:de(qe,_e)},{pattern:c(/(^|[^@\\])\$"(?:\\.|\{\{|<<0>>|[^\\"{])*"/.source,[Xe]),lookbehind:!0,greedy:!0,inside:de(Xe,it)}],char:{pattern:RegExp(he),greedy:!0}}),o.languages.dotnet=o.languages.cs=o.languages.csharp})(r)}return wBe}var yBe,azt;function p$r(){if(azt)return yBe;azt=1;var n=REe();yBe=r,r.displayName="aspnet",r.aliases=[];function r(o){o.register(n),o.languages.aspnet=o.languages.extend("markup",{"page-directive":{pattern:/<%\s*@.*%>/,alias:"tag",inside:{"page-directive":{pattern:/<%\s*@\s*(?:Assembly|Control|Implements|Import|Master(?:Type)?|OutputCache|Page|PreviousPageType|Reference|Register)?|%>/i,alias:"tag"},rest:o.languages.markup.tag.inside}},directive:{pattern:/<%.*%>/,alias:"tag",inside:{directive:{pattern:/<%\s*?[$=%#:]{0,2}|%>/,alias:"tag"},rest:o.languages.csharp}}}),o.languages.aspnet.tag.pattern=/<(?!%)\/?[^\s>\/]+(?:\s+[^\s>\/=]+(?:=(?:("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|[^\s'">=]+))?)*\s*\/?>/,o.languages.insertBefore("inside","punctuation",{directive:o.languages.aspnet.directive},o.languages.aspnet.tag.inside["attr-value"]),o.languages.insertBefore("aspnet","comment",{"asp-comment":{pattern:/<%--[\s\S]*?--%>/,alias:["asp","comment"]}}),o.languages.insertBefore("aspnet",o.languages.javascript?"script":"tag",{"asp-script":{pattern:/(<script(?=.*runat=['"]?server\b)[^>]*>)[\s\S]*?(?=<\/script>)/i,lookbehind:!0,alias:["asp","script"],inside:o.languages.csharp||{}}})}return yBe}var EBe,szt;function g$r(){if(szt)return EBe;szt=1,EBe=n,n.displayName="autohotkey",n.aliases=[];function n(r){r.languages.autohotkey={comment:[{pattern:/(^|\s);.*/,lookbehind:!0},{pattern:/(^[\t ]*)\/\*(?:[\r\n](?![ \t]*\*\/)|[^\r\n])*(?:[\r\n][ \t]*\*\/)?/m,lookbehind:!0,greedy:!0}],tag:{pattern:/^([ \t]*)[^\s,`":]+(?=:[ \t]*$)/m,lookbehind:!0},string:/"(?:[^"\n\r]|"")*"/,variable:/%\w+%/,number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee]-?\d+)?/,operator:/\?|\/\/?=?|:=|\|[=|]?|&[=&]?|\+[=+]?|-[=-]?|\*[=*]?|<(?:<=?|>|=)?|>>?=?|[.^!=~]=?|\b(?:AND|NOT|OR)\b/,boolean:/\b(?:false|true)\b/,selector:/\b(?:AutoTrim|BlockInput|Break|Click|ClipWait|Continue|Control|ControlClick|ControlFocus|ControlGet|ControlGetFocus|ControlGetPos|ControlGetText|ControlMove|ControlSend|ControlSendRaw|ControlSetText|CoordMode|Critical|DetectHiddenText|DetectHiddenWindows|Drive|DriveGet|DriveSpaceFree|EnvAdd|EnvDiv|EnvGet|EnvMult|EnvSet|EnvSub|EnvUpdate|Exit|ExitApp|FileAppend|FileCopy|FileCopyDir|FileCreateDir|FileCreateShortcut|FileDelete|FileEncoding|FileGetAttrib|FileGetShortcut|FileGetSize|FileGetTime|FileGetVersion|FileInstall|FileMove|FileMoveDir|FileRead|FileReadLine|FileRecycle|FileRecycleEmpty|FileRemoveDir|FileSelectFile|FileSelectFolder|FileSetAttrib|FileSetTime|FormatTime|GetKeyState|Gosub|Goto|GroupActivate|GroupAdd|GroupClose|GroupDeactivate|Gui|GuiControl|GuiControlGet|Hotkey|ImageSearch|IniDelete|IniRead|IniWrite|Input|InputBox|KeyWait|ListHotkeys|ListLines|ListVars|Loop|Menu|MouseClick|MouseClickDrag|MouseGetPos|MouseMove|MsgBox|OnExit|OutputDebug|Pause|PixelGetColor|PixelSearch|PostMessage|Process|Progress|Random|RegDelete|RegRead|RegWrite|Reload|Repeat|Return|Run|RunAs|RunWait|Send|SendEvent|SendInput|SendMessage|SendMode|SendPlay|SendRaw|SetBatchLines|SetCapslockState|SetControlDelay|SetDefaultMouseSpeed|SetEnv|SetFormat|SetKeyDelay|SetMouseDelay|SetNumlockState|SetRegView|SetScrollLockState|SetStoreCapslockMode|SetTimer|SetTitleMatchMode|SetWinDelay|SetWorkingDir|Shutdown|Sleep|Sort|SoundBeep|SoundGet|SoundGetWaveVolume|SoundPlay|SoundSet|SoundSetWaveVolume|SplashImage|SplashTextOff|SplashTextOn|SplitPath|StatusBarGetText|StatusBarWait|StringCaseSense|StringGetPos|StringLeft|StringLen|StringLower|StringMid|StringReplace|StringRight|StringSplit|StringTrimLeft|StringTrimRight|StringUpper|Suspend|SysGet|Thread|ToolTip|Transform|TrayTip|URLDownloadToFile|WinActivate|WinActivateBottom|WinClose|WinGet|WinGetActiveStats|WinGetActiveTitle|WinGetClass|WinGetPos|WinGetText|WinGetTitle|WinHide|WinKill|WinMaximize|WinMenuSelectItem|WinMinimize|WinMinimizeAll|WinMinimizeAllUndo|WinMove|WinRestore|WinSet|WinSetTitle|WinShow|WinWait|WinWaitActive|WinWaitClose|WinWaitNotActive)\b/i,constant:/\b(?:a_ahkpath|a_ahkversion|a_appdata|a_appdatacommon|a_autotrim|a_batchlines|a_caretx|a_carety|a_computername|a_controldelay|a_cursor|a_dd|a_ddd|a_dddd|a_defaultmousespeed|a_desktop|a_desktopcommon|a_detecthiddentext|a_detecthiddenwindows|a_endchar|a_eventinfo|a_exitreason|a_fileencoding|a_formatfloat|a_formatinteger|a_gui|a_guicontrol|a_guicontrolevent|a_guievent|a_guiheight|a_guiwidth|a_guix|a_guiy|a_hour|a_iconfile|a_iconhidden|a_iconnumber|a_icontip|a_index|a_ipaddress1|a_ipaddress2|a_ipaddress3|a_ipaddress4|a_is64bitos|a_isadmin|a_iscompiled|a_iscritical|a_ispaused|a_issuspended|a_isunicode|a_keydelay|a_language|a_lasterror|a_linefile|a_linenumber|a_loopfield|a_loopfileattrib|a_loopfiledir|a_loopfileext|a_loopfilefullpath|a_loopfilelongpath|a_loopfilename|a_loopfileshortname|a_loopfileshortpath|a_loopfilesize|a_loopfilesizekb|a_loopfilesizemb|a_loopfiletimeaccessed|a_loopfiletimecreated|a_loopfiletimemodified|a_loopreadline|a_loopregkey|a_loopregname|a_loopregsubkey|a_loopregtimemodified|a_loopregtype|a_mday|a_min|a_mm|a_mmm|a_mmmm|a_mon|a_mousedelay|a_msec|a_mydocuments|a_now|a_nowutc|a_numbatchlines|a_ostype|a_osversion|a_priorhotkey|a_priorkey|a_programfiles|a_programs|a_programscommon|a_ptrsize|a_regview|a_screendpi|a_screenheight|a_screenwidth|a_scriptdir|a_scriptfullpath|a_scripthwnd|a_scriptname|a_sec|a_space|a_startmenu|a_startmenucommon|a_startup|a_startupcommon|a_stringcasesense|a_tab|a_temp|a_thisfunc|a_thishotkey|a_thislabel|a_thismenu|a_thismenuitem|a_thismenuitempos|a_tickcount|a_timeidle|a_timeidlephysical|a_timesincepriorhotkey|a_timesincethishotkey|a_titlematchmode|a_titlematchmodespeed|a_username|a_wday|a_windelay|a_windir|a_workingdir|a_yday|a_year|a_yweek|a_yyyy|clipboard|clipboardall|comspec|errorlevel|programfiles)\b/i,builtin:/\b(?:abs|acos|asc|asin|atan|ceil|chr|class|comobjactive|comobjarray|comobjconnect|comobjcreate|comobjerror|comobjflags|comobjget|comobjquery|comobjtype|comobjvalue|cos|dllcall|exp|fileexist|Fileopen|floor|format|il_add|il_create|il_destroy|instr|isfunc|islabel|IsObject|ln|log|ltrim|lv_add|lv_delete|lv_deletecol|lv_getcount|lv_getnext|lv_gettext|lv_insert|lv_insertcol|lv_modify|lv_modifycol|lv_setimagelist|mod|numget|numput|onmessage|regexmatch|regexreplace|registercallback|round|rtrim|sb_seticon|sb_setparts|sb_settext|sin|sqrt|strlen|strreplace|strsplit|substr|tan|tv_add|tv_delete|tv_get|tv_getchild|tv_getcount|tv_getnext|tv_getparent|tv_getprev|tv_getselection|tv_gettext|tv_modify|varsetcapacity|winactive|winexist|__Call|__Get|__New|__Set)\b/i,symbol:/\b(?:alt|altdown|altup|appskey|backspace|browser_back|browser_favorites|browser_forward|browser_home|browser_refresh|browser_search|browser_stop|bs|capslock|ctrl|ctrlbreak|ctrldown|ctrlup|del|delete|down|end|enter|esc|escape|f1|f10|f11|f12|f13|f14|f15|f16|f17|f18|f19|f2|f20|f21|f22|f23|f24|f3|f4|f5|f6|f7|f8|f9|home|ins|insert|joy1|joy10|joy11|joy12|joy13|joy14|joy15|joy16|joy17|joy18|joy19|joy2|joy20|joy21|joy22|joy23|joy24|joy25|joy26|joy27|joy28|joy29|joy3|joy30|joy31|joy32|joy4|joy5|joy6|joy7|joy8|joy9|joyaxes|joybuttons|joyinfo|joyname|joypov|joyr|joyu|joyv|joyx|joyy|joyz|lalt|launch_app1|launch_app2|launch_mail|launch_media|lbutton|lcontrol|lctrl|left|lshift|lwin|lwindown|lwinup|mbutton|media_next|media_play_pause|media_prev|media_stop|numlock|numpad0|numpad1|numpad2|numpad3|numpad4|numpad5|numpad6|numpad7|numpad8|numpad9|numpadadd|numpadclear|numpaddel|numpaddiv|numpaddot|numpaddown|numpadend|numpadenter|numpadhome|numpadins|numpadleft|numpadmult|numpadpgdn|numpadpgup|numpadright|numpadsub|numpadup|pgdn|pgup|printscreen|ralt|rbutton|rcontrol|rctrl|right|rshift|rwin|rwindown|rwinup|scrolllock|shift|shiftdown|shiftup|space|tab|up|volume_down|volume_mute|volume_up|wheeldown|wheelleft|wheelright|wheelup|xbutton1|xbutton2)\b/i,important:/#\b(?:AllowSameLineComments|ClipboardTimeout|CommentFlag|DerefChar|ErrorStdOut|EscapeChar|HotkeyInterval|HotkeyModifierTimeout|Hotstring|If|IfTimeout|IfWinActive|IfWinExist|IfWinNotActive|IfWinNotExist|Include|IncludeAgain|InputLevel|InstallKeybdHook|InstallMouseHook|KeyHistory|MaxHotkeysPerInterval|MaxMem|MaxThreads|MaxThreadsBuffer|MaxThreadsPerHotkey|MenuMaskKey|NoEnv|NoTrayIcon|Persistent|SingleInstance|UseHook|Warn|WinActivateForce)\b/i,keyword:/\b(?:Abort|AboveNormal|Add|ahk_class|ahk_exe|ahk_group|ahk_id|ahk_pid|All|Alnum|Alpha|AltSubmit|AltTab|AltTabAndMenu|AltTabMenu|AltTabMenuDismiss|AlwaysOnTop|AutoSize|Background|BackgroundTrans|BelowNormal|between|BitAnd|BitNot|BitOr|BitShiftLeft|BitShiftRight|BitXOr|Bold|Border|Button|ByRef|Catch|Checkbox|Checked|CheckedGray|Choose|ChooseString|Close|Color|ComboBox|Contains|ControlList|Count|Date|DateTime|Days|DDL|Default|DeleteAll|Delimiter|Deref|Destroy|Digit|Disable|Disabled|DropDownList|Edit|Eject|Else|Enable|Enabled|Error|Exist|Expand|ExStyle|FileSystem|Finally|First|Flash|Float|FloatFast|Focus|Font|for|global|Grid|Group|GroupBox|GuiClose|GuiContextMenu|GuiDropFiles|GuiEscape|GuiSize|Hdr|Hidden|Hide|High|HKCC|HKCR|HKCU|HKEY_CLASSES_ROOT|HKEY_CURRENT_CONFIG|HKEY_CURRENT_USER|HKEY_LOCAL_MACHINE|HKEY_USERS|HKLM|HKU|Hours|HScroll|Icon|IconSmall|ID|IDLast|If|IfEqual|IfExist|IfGreater|IfGreaterOrEqual|IfInString|IfLess|IfLessOrEqual|IfMsgBox|IfNotEqual|IfNotExist|IfNotInString|IfWinActive|IfWinExist|IfWinNotActive|IfWinNotExist|Ignore|ImageList|in|Integer|IntegerFast|Interrupt|is|italic|Join|Label|LastFound|LastFoundExist|Limit|Lines|List|ListBox|ListView|local|Lock|Logoff|Low|Lower|Lowercase|MainWindow|Margin|Maximize|MaximizeBox|MaxSize|Minimize|MinimizeBox|MinMax|MinSize|Minutes|MonthCal|Mouse|Move|Multi|NA|No|NoActivate|NoDefault|NoHide|NoIcon|NoMainWindow|norm|Normal|NoSort|NoSortHdr|NoStandard|Not|NoTab|NoTimers|Number|Off|Ok|On|OwnDialogs|Owner|Parse|Password|Picture|Pixel|Pos|Pow|Priority|ProcessName|Radio|Range|Read|ReadOnly|Realtime|Redraw|Region|REG_BINARY|REG_DWORD|REG_EXPAND_SZ|REG_MULTI_SZ|REG_SZ|Relative|Rename|Report|Resize|Restore|Retry|RGB|Screen|Seconds|Section|Serial|SetLabel|ShiftAltTab|Show|Single|Slider|SortDesc|Standard|static|Status|StatusBar|StatusCD|strike|Style|Submit|SysMenu|Tab2|TabStop|Text|Theme|Throw|Tile|ToggleCheck|ToggleEnable|ToolWindow|Top|Topmost|TransColor|Transparent|Tray|TreeView|Try|TryAgain|Type|UnCheck|underline|Unicode|Unlock|Until|UpDown|Upper|Uppercase|UseErrorLevel|Vis|VisFirst|Visible|VScroll|Wait|WaitClose|WantCtrlA|WantF2|WantReturn|While|Wrap|Xdigit|xm|xp|xs|Yes|ym|yp|ys)\b/i,function:/[^(); \t,\n+*\-=?>:\\\/<&%\[\]]+(?=\()/,punctuation:/[{}[\]():,]/}}return EBe}var SBe,czt;function m$r(){if(czt)return SBe;czt=1,SBe=n,n.displayName="autoit",n.aliases=[];function n(r){r.languages.autoit={comment:[/;.*/,{pattern:/(^[\t ]*)#(?:comments-start|cs)[\s\S]*?^[ \t]*#(?:ce|comments-end)/m,lookbehind:!0}],url:{pattern:/(^[\t ]*#include\s+)(?:<[^\r\n>]+>|"[^\r\n"]+")/m,lookbehind:!0},string:{pattern:/(["'])(?:\1\1|(?!\1)[^\r\n])*\1/,greedy:!0,inside:{variable:/([%$@])\w+\1/}},directive:{pattern:/(^[\t ]*)#[\w-]+/m,lookbehind:!0,alias:"keyword"},function:/\b\w+(?=\()/,variable:/[$@]\w+/,keyword:/\b(?:Case|Const|Continue(?:Case|Loop)|Default|Dim|Do|Else(?:If)?|End(?:Func|If|Select|Switch|With)|Enum|Exit(?:Loop)?|For|Func|Global|If|In|Local|Next|Null|ReDim|Select|Static|Step|Switch|Then|To|Until|Volatile|WEnd|While|With)\b/i,number:/\b(?:0x[\da-f]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/i,boolean:/\b(?:False|True)\b/i,operator:/<[=>]?|[-+*\/=&>]=?|[?^]|\b(?:And|Not|Or)\b/i,punctuation:/[\[\]().,:]/}}return SBe}var _Be,uzt;function b$r(){if(uzt)return _Be;uzt=1,_Be=n,n.displayName="avisynth",n.aliases=["avs"];function n(r){(function(o){function s(k,x){return k.replace(/<<(\d+)>>/g,function(O,P){return x[+P]})}function c(k,x,O){return RegExp(s(k,x),O)}var d=/bool|clip|float|int|string|val/.source,h=[/is(?:bool|clip|float|int|string)|defined|(?:(?:internal)?function|var)?exists?/.source,/apply|assert|default|eval|import|nop|select|undefined/.source,/opt_(?:allowfloataudio|avipadscanlines|dwchannelmask|enable_(?:b64a|planartopackedrgb|v210|y3_10_10|y3_10_16)|usewaveextensible|vdubplanarhack)|set(?:cachemode|maxcpu|memorymax|planarlegacyalignment|workingdir)/.source,/hex(?:value)?|value/.source,/abs|ceil|continued(?:denominator|numerator)?|exp|floor|fmod|frac|log(?:10)?|max|min|muldiv|pi|pow|rand|round|sign|spline|sqrt/.source,/a?sinh?|a?cosh?|a?tan[2h]?/.source,/(?:bit(?:and|not|x?or|[lr]?shift[aslu]?|sh[lr]|sa[lr]|[lr]rotatel?|ro[rl]|te?st|set(?:count)?|cl(?:ea)?r|ch(?:an)?ge?))/.source,/average(?:[bgr]|chroma[uv]|luma)|(?:[rgb]|chroma[uv]|luma|rgb|[yuv](?=difference(?:fromprevious|tonext)))difference(?:fromprevious|tonext)?|[yuvrgb]plane(?:median|min|max|minmaxdifference)/.source,/getprocessinfo|logmsg|script(?:dir(?:utf8)?|file(?:utf8)?|name(?:utf8)?)|setlogparams/.source,/chr|(?:fill|find|left|mid|replace|rev|right)str|format|[lu]case|ord|str(?:cmpi?|fromutf8|len|toutf8)|time|trim(?:all|left|right)/.source,/isversionorgreater|version(?:number|string)/.source,/buildpixeltype|colorspacenametopixeltype/.source,/addautoloaddir|on(?:cpu|cuda)|prefetch|setfiltermtmode/.source].join("|"),b=[/has(?:audio|video)/.source,/height|width/.source,/frame(?:count|rate)|framerate(?:denominator|numerator)/.source,/getparity|is(?:field|frame)based/.source,/bitspercomponent|componentsize|hasalpha|is(?:planar(?:rgba?)?|interleaved|rgb(?:24|32|48|64)?|y(?:8|u(?:va?|y2))?|yv(?:12|16|24|411)|420|422|444|packedrgb)|numcomponents|pixeltype/.source,/audio(?:bits|channels|duration|length(?:[fs]|hi|lo)?|rate)|isaudio(?:float|int)/.source].join("|"),y=[/avi(?:file)?source|directshowsource|image(?:reader|source|sourceanim)|opendmlsource|segmented(?:avisource|directshowsource)|wavsource/.source,/coloryuv|convertbacktoyuy2|convertto(?:RGB(?:24|32|48|64)|(?:planar)?RGBA?|Y8?|YV(?:12|16|24|411)|YUVA?(?:411|420|422|444)|YUY2)|fixluminance|gr[ae]yscale|invert|levels|limiter|mergea?rgb|merge(?:chroma|luma)|rgbadjust|show(?:alpha|blue|green|red)|swapuv|tweak|[uv]toy8?|ytouv/.source,/(?:colorkey|reset)mask|layer|mask(?:hs)?|merge|overlay|subtract/.source,/addborders|(?:bicubic|bilinear|blackman|gauss|lanczos4|lanczos|point|sinc|spline(?:16|36|64))resize|crop(?:bottom)?|flip(?:horizontal|vertical)|(?:horizontal|vertical)?reduceby2|letterbox|skewrows|turn(?:180|left|right)/.source,/blur|fixbrokenchromaupsampling|generalconvolution|(?:spatial|temporal)soften|sharpen/.source,/trim|(?:un)?alignedsplice|(?:assume|assumescaled|change|convert)FPS|(?:delete|duplicate)frame|dissolve|fade(?:in|io|out)[02]?|freezeframe|interleave|loop|reverse|select(?:even|odd|(?:range)?every)/.source,/assume[bt]ff|assume(?:field|frame)based|bob|complementparity|doubleweave|peculiarblend|pulldown|separate(?:columns|fields|rows)|swapfields|weave(?:columns|rows)?/.source,/amplify(?:db)?|assumesamplerate|audiodub(?:ex)?|audiotrim|convertaudioto(?:(?:8|16|24|32)bit|float)|converttomono|delayaudio|ensurevbrmp3sync|get(?:left|right)?channel|kill(?:audio|video)|mergechannels|mixaudio|monotostereo|normalize|resampleaudio|ssrc|supereq|timestretch/.source,/animate|applyrange|conditional(?:filter|reader|select)|frameevaluate|scriptclip|tcp(?:server|source)|writefile(?:end|if|start)?/.source,/imagewriter/.source,/blackness|blankclip|colorbars(?:hd)?|compare|dumpfiltergraph|echo|histogram|info|messageclip|preroll|setgraphanalysis|show(?:framenumber|smpte|time)|showfiveversions|stack(?:horizontal|vertical)|subtitle|tone|version/.source].join("|"),E=[h,b,y].join("|");o.languages.avisynth={comment:[{pattern:/(^|[^\\])\[\*(?:[^\[*]|\[(?!\*)|\*(?!\])|\[\*(?:[^\[*]|\[(?!\*)|\*(?!\]))*\*\])*\*\]/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\$])#.*/,lookbehind:!0,greedy:!0}],argument:{pattern:c(/\b(?:<<0>>)\s+("?)\w+\1/.source,[d],"i"),inside:{keyword:/^\w+/}},"argument-label":{pattern:/([,(][\s\\]*)\w+\s*=(?!=)/,lookbehind:!0,inside:{"argument-name":{pattern:/^\w+/,alias:"punctuation"},punctuation:/=$/}},string:[{pattern:/"""[\s\S]*?"""/,greedy:!0},{pattern:/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0,inside:{constant:{pattern:/\b(?:DEFAULT_MT_MODE|(?:MAINSCRIPT|PROGRAM|SCRIPT)DIR|(?:MACHINE|USER)_(?:CLASSIC|PLUS)_PLUGINS)\b/}}}],variable:/\b(?:last)\b/i,boolean:/\b(?:false|no|true|yes)\b/i,keyword:/\b(?:catch|else|for|function|global|if|return|try|while|__END__)\b/i,constant:/\bMT_(?:MULTI_INSTANCE|NICE_FILTER|SERIALIZED|SPECIAL_MT)\b/,"builtin-function":{pattern:c(/\b(?:<<0>>)\b/.source,[E],"i"),alias:"function"},"type-cast":{pattern:c(/\b(?:<<0>>)(?=\s*\()/.source,[d],"i"),alias:"keyword"},function:{pattern:/\b[a-z_]\w*(?=\s*\()|(\.)[a-z_]\w*\b/i,lookbehind:!0},"line-continuation":{pattern:/(^[ \t]*)\\|\\(?=[ \t]*$)/m,lookbehind:!0,alias:"punctuation"},number:/\B\$(?:[\da-f]{6}|[\da-f]{8})\b|(?:(?:\b|\B-)\d+(?:\.\d*)?\b|\B\.\d+\b)/i,operator:/\+\+?|[!=<>]=?|&&|\|\||[?:*/%-]/,punctuation:/[{}\[\]();,.]/},o.languages.avs=o.languages.avisynth})(r)}return _Be}var kBe,lzt;function v$r(){if(lzt)return kBe;lzt=1,kBe=n,n.displayName="avroIdl",n.aliases=[];function n(r){r.languages["avro-idl"]={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/(^|[^\\])"(?:[^\r\n"\\]|\\.)*"/,lookbehind:!0,greedy:!0},annotation:{pattern:/@(?:[$\w.-]|`[^\r\n`]+`)+/,greedy:!0,alias:"function"},"function-identifier":{pattern:/`[^\r\n`]+`(?=\s*\()/,greedy:!0,alias:"function"},identifier:{pattern:/`[^\r\n`]+`/,greedy:!0},"class-name":{pattern:/(\b(?:enum|error|protocol|record|throws)\b\s+)[$\w]+/,lookbehind:!0,greedy:!0},keyword:/\b(?:array|boolean|bytes|date|decimal|double|enum|error|false|fixed|float|idl|import|int|local_timestamp_ms|long|map|null|oneway|protocol|record|schema|string|throws|time_ms|timestamp_ms|true|union|uuid|void)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,number:[{pattern:/(^|[^\w.])-?(?:(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|0x(?:[a-f0-9]+(?:\.[a-f0-9]*)?|\.[a-f0-9]+)(?:p[+-]?\d+)?)[dfl]?(?![\w.])/i,lookbehind:!0},/-?\b(?:Infinity|NaN)\b/],operator:/=/,punctuation:/[()\[\]{}<>.:,;-]/},r.languages.avdl=r.languages["avro-idl"]}return kBe}var TBe,dzt;function adn(){if(dzt)return TBe;dzt=1,TBe=n,n.displayName="bash",n.aliases=["shell"];function n(r){(function(o){var s="\\b(?:BASH|BASHOPTS|BASH_ALIASES|BASH_ARGC|BASH_ARGV|BASH_CMDS|BASH_COMPLETION_COMPAT_DIR|BASH_LINENO|BASH_REMATCH|BASH_SOURCE|BASH_VERSINFO|BASH_VERSION|COLORTERM|COLUMNS|COMP_WORDBREAKS|DBUS_SESSION_BUS_ADDRESS|DEFAULTS_PATH|DESKTOP_SESSION|DIRSTACK|DISPLAY|EUID|GDMSESSION|GDM_LANG|GNOME_KEYRING_CONTROL|GNOME_KEYRING_PID|GPG_AGENT_INFO|GROUPS|HISTCONTROL|HISTFILE|HISTFILESIZE|HISTSIZE|HOME|HOSTNAME|HOSTTYPE|IFS|INSTANCE|JOB|LANG|LANGUAGE|LC_ADDRESS|LC_ALL|LC_IDENTIFICATION|LC_MEASUREMENT|LC_MONETARY|LC_NAME|LC_NUMERIC|LC_PAPER|LC_TELEPHONE|LC_TIME|LESSCLOSE|LESSOPEN|LINES|LOGNAME|LS_COLORS|MACHTYPE|MAILCHECK|MANDATORY_PATH|NO_AT_BRIDGE|OLDPWD|OPTERR|OPTIND|ORBIT_SOCKETDIR|OSTYPE|PAPERSIZE|PATH|PIPESTATUS|PPID|PS1|PS2|PS3|PS4|PWD|RANDOM|REPLY|SECONDS|SELINUX_INIT|SESSION|SESSIONTYPE|SESSION_MANAGER|SHELL|SHELLOPTS|SHLVL|SSH_AUTH_SOCK|TERM|UID|UPSTART_EVENTS|UPSTART_INSTANCE|UPSTART_JOB|UPSTART_SESSION|USER|WINDOWID|XAUTHORITY|XDG_CONFIG_DIRS|XDG_CURRENT_DESKTOP|XDG_DATA_DIRS|XDG_GREETER_DATA_DIR|XDG_MENU_PREFIX|XDG_RUNTIME_DIR|XDG_SEAT|XDG_SEAT_PATH|XDG_SESSION_DESKTOP|XDG_SESSION_ID|XDG_SESSION_PATH|XDG_SESSION_TYPE|XDG_VTNR|XMODIFIERS)\\b",c={pattern:/(^(["']?)\w+\2)[ \t]+\S.*/,lookbehind:!0,alias:"punctuation",inside:null},d={bash:c,environment:{pattern:RegExp("\\$"+s),alias:"constant"},variable:[{pattern:/\$?\(\([\s\S]+?\)\)/,greedy:!0,inside:{variable:[{pattern:/(^\$\(\([\s\S]+)\)\)/,lookbehind:!0},/^\$\(\(/],number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee]-?\d+)?/,operator:/--|\+\+|\*\*=?|<<=?|>>=?|&&|\|\||[=!+\-*/%<>^&|]=?|[?~:]/,punctuation:/\(\(?|\)\)?|,|;/}},{pattern:/\$\((?:\([^)]+\)|[^()])+\)|`[^`]+`/,greedy:!0,inside:{variable:/^\$\(|^`|\)$|`$/}},{pattern:/\$\{[^}]+\}/,greedy:!0,inside:{operator:/:[-=?+]?|[!\/]|##?|%%?|\^\^?|,,?/,punctuation:/[\[\]]/,environment:{pattern:RegExp("(\\{)"+s),lookbehind:!0,alias:"constant"}}},/\$(?:\w+|[#?*!@$])/],entity:/\\(?:[abceEfnrtv\\"]|O?[0-7]{1,3}|U[0-9a-fA-F]{8}|u[0-9a-fA-F]{4}|x[0-9a-fA-F]{1,2})/};o.languages.bash={shebang:{pattern:/^#!\s*\/.*/,alias:"important"},comment:{pattern:/(^|[^"{\\$])#.*/,lookbehind:!0},"function-name":[{pattern:/(\bfunction\s+)[\w-]+(?=(?:\s*\(?:\s*\))?\s*\{)/,lookbehind:!0,alias:"function"},{pattern:/\b[\w-]+(?=\s*\(\s*\)\s*\{)/,alias:"function"}],"for-or-select":{pattern:/(\b(?:for|select)\s+)\w+(?=\s+in\s)/,alias:"variable",lookbehind:!0},"assign-left":{pattern:/(^|[\s;|&]|[<>]\()\w+(?=\+?=)/,inside:{environment:{pattern:RegExp("(^|[\\s;|&]|[<>]\\()"+s),lookbehind:!0,alias:"constant"}},alias:"variable",lookbehind:!0},string:[{pattern:/((?:^|[^<])<<-?\s*)(\w+)\s[\s\S]*?(?:\r?\n|\r)\2/,lookbehind:!0,greedy:!0,inside:d},{pattern:/((?:^|[^<])<<-?\s*)(["'])(\w+)\2\s[\s\S]*?(?:\r?\n|\r)\3/,lookbehind:!0,greedy:!0,inside:{bash:c}},{pattern:/(^|[^\\](?:\\\\)*)"(?:\\[\s\S]|\$\([^)]+\)|\$(?!\()|`[^`]+`|[^"\\`$])*"/,lookbehind:!0,greedy:!0,inside:d},{pattern:/(^|[^$\\])'[^']*'/,lookbehind:!0,greedy:!0},{pattern:/\$'(?:[^'\\]|\\[\s\S])*'/,greedy:!0,inside:{entity:d.entity}}],environment:{pattern:RegExp("\\$?"+s),alias:"constant"},variable:d.variable,function:{pattern:/(^|[\s;|&]|[<>]\()(?:add|apropos|apt|apt-cache|apt-get|aptitude|aspell|automysqlbackup|awk|basename|bash|bc|bconsole|bg|bzip2|cal|cat|cfdisk|chgrp|chkconfig|chmod|chown|chroot|cksum|clear|cmp|column|comm|composer|cp|cron|crontab|csplit|curl|cut|date|dc|dd|ddrescue|debootstrap|df|diff|diff3|dig|dir|dircolors|dirname|dirs|dmesg|docker|docker-compose|du|egrep|eject|env|ethtool|expand|expect|expr|fdformat|fdisk|fg|fgrep|file|find|fmt|fold|format|free|fsck|ftp|fuser|gawk|git|gparted|grep|groupadd|groupdel|groupmod|groups|grub-mkconfig|gzip|halt|head|hg|history|host|hostname|htop|iconv|id|ifconfig|ifdown|ifup|import|install|ip|jobs|join|kill|killall|less|link|ln|locate|logname|logrotate|look|lpc|lpr|lprint|lprintd|lprintq|lprm|ls|lsof|lynx|make|man|mc|mdadm|mkconfig|mkdir|mke2fs|mkfifo|mkfs|mkisofs|mknod|mkswap|mmv|more|most|mount|mtools|mtr|mutt|mv|nano|nc|netstat|nice|nl|node|nohup|notify-send|npm|nslookup|op|open|parted|passwd|paste|pathchk|ping|pkill|pnpm|podman|podman-compose|popd|pr|printcap|printenv|ps|pushd|pv|quota|quotacheck|quotactl|ram|rar|rcp|reboot|remsync|rename|renice|rev|rm|rmdir|rpm|rsync|scp|screen|sdiff|sed|sendmail|seq|service|sftp|sh|shellcheck|shuf|shutdown|sleep|slocate|sort|split|ssh|stat|strace|su|sudo|sum|suspend|swapon|sync|tac|tail|tar|tee|time|timeout|top|touch|tr|traceroute|tsort|tty|umount|uname|unexpand|uniq|units|unrar|unshar|unzip|update-grub|uptime|useradd|userdel|usermod|users|uudecode|uuencode|v|vcpkg|vdir|vi|vim|virsh|vmstat|wait|watch|wc|wget|whereis|which|who|whoami|write|xargs|xdg-open|yarn|yes|zenity|zip|zsh|zypper)(?=$|[)\s;|&])/,lookbehind:!0},keyword:{pattern:/(^|[\s;|&]|[<>]\()(?:case|do|done|elif|else|esac|fi|for|function|if|in|select|then|until|while)(?=$|[)\s;|&])/,lookbehind:!0},builtin:{pattern:/(^|[\s;|&]|[<>]\()(?:\.|:|alias|bind|break|builtin|caller|cd|command|continue|declare|echo|enable|eval|exec|exit|export|getopts|hash|help|let|local|logout|mapfile|printf|pwd|read|readarray|readonly|return|set|shift|shopt|source|test|times|trap|type|typeset|ulimit|umask|unalias|unset)(?=$|[)\s;|&])/,lookbehind:!0,alias:"class-name"},boolean:{pattern:/(^|[\s;|&]|[<>]\()(?:false|true)(?=$|[)\s;|&])/,lookbehind:!0},"file-descriptor":{pattern:/\B&\d\b/,alias:"important"},operator:{pattern:/\d?<>|>\||\+=|=[=~]?|!=?|<<[<-]?|[&\d]?>>|\d[<>]&?|[<>][&=]?|&[>&]?|\|[&|]?/,inside:{"file-descriptor":{pattern:/^\d/,alias:"important"}}},punctuation:/\$?\(\(?|\)\)?|\.\.|[{}[\];\\]/,number:{pattern:/(^|\s)(?:[1-9]\d*|0)(?:[.,]\d+)?\b/,lookbehind:!0}},c.inside=o.languages.bash;for(var h=["comment","function-name","for-or-select","assign-left","string","environment","function","keyword","builtin","boolean","file-descriptor","operator","punctuation","number"],b=d.variable[1].inside,y=0;y<h.length;y++)b[h[y]]=o.languages.bash[h[y]];o.languages.shell=o.languages.bash})(r)}return TBe}var xBe,fzt;function sdn(){if(fzt)return xBe;fzt=1,xBe=n,n.displayName="basic",n.aliases=[];function n(r){r.languages.basic={comment:{pattern:/(?:!|REM\b).+/i,inside:{keyword:/^REM/i}},string:{pattern:/"(?:""|[!#$%&'()*,\/:;<=>?^\w +\-.])*"/,greedy:!0},number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,keyword:/\b(?:AS|BEEP|BLOAD|BSAVE|CALL(?: ABSOLUTE)?|CASE|CHAIN|CHDIR|CLEAR|CLOSE|CLS|COM|COMMON|CONST|DATA|DECLARE|DEF(?: FN| SEG|DBL|INT|LNG|SNG|STR)|DIM|DO|DOUBLE|ELSE|ELSEIF|END|ENVIRON|ERASE|ERROR|EXIT|FIELD|FILES|FOR|FUNCTION|GET|GOSUB|GOTO|IF|INPUT|INTEGER|IOCTL|KEY|KILL|LINE INPUT|LOCATE|LOCK|LONG|LOOP|LSET|MKDIR|NAME|NEXT|OFF|ON(?: COM| ERROR| KEY| TIMER)?|OPEN|OPTION BASE|OUT|POKE|PUT|READ|REDIM|REM|RESTORE|RESUME|RETURN|RMDIR|RSET|RUN|SELECT CASE|SHARED|SHELL|SINGLE|SLEEP|STATIC|STEP|STOP|STRING|SUB|SWAP|SYSTEM|THEN|TIMER|TO|TROFF|TRON|TYPE|UNLOCK|UNTIL|USING|VIEW PRINT|WAIT|WEND|WHILE|WRITE)(?:\$|\b)/i,function:/\b(?:ABS|ACCESS|ACOS|ANGLE|AREA|ARITHMETIC|ARRAY|ASIN|ASK|AT|ATN|BASE|BEGIN|BREAK|CAUSE|CEIL|CHR|CLIP|COLLATE|COLOR|CON|COS|COSH|COT|CSC|DATE|DATUM|DEBUG|DECIMAL|DEF|DEG|DEGREES|DELETE|DET|DEVICE|DISPLAY|DOT|ELAPSED|EPS|ERASABLE|EXLINE|EXP|EXTERNAL|EXTYPE|FILETYPE|FIXED|FP|GO|GRAPH|HANDLER|IDN|IMAGE|IN|INT|INTERNAL|IP|IS|KEYED|LBOUND|LCASE|LEFT|LEN|LENGTH|LET|LINE|LINES|LOG|LOG10|LOG2|LTRIM|MARGIN|MAT|MAX|MAXNUM|MID|MIN|MISSING|MOD|NATIVE|NUL|NUMERIC|OF|OPTION|ORD|ORGANIZATION|OUTIN|OUTPUT|PI|POINT|POINTER|POINTS|POS|PRINT|PROGRAM|PROMPT|RAD|RADIANS|RANDOMIZE|RECORD|RECSIZE|RECTYPE|RELATIVE|REMAINDER|REPEAT|REST|RETRY|REWRITE|RIGHT|RND|ROUND|RTRIM|SAME|SEC|SELECT|SEQUENTIAL|SET|SETTER|SGN|SIN|SINH|SIZE|SKIP|SQR|STANDARD|STATUS|STR|STREAM|STYLE|TAB|TAN|TANH|TEMPLATE|TEXT|THERE|TIME|TIMEOUT|TRACE|TRANSFORM|TRUNCATE|UBOUND|UCASE|USE|VAL|VARIABLE|VIEWPORT|WHEN|WINDOW|WITH|ZER|ZONEWIDTH)(?:\$|\b)/i,operator:/<[=>]?|>=?|[+\-*\/^=&]|\b(?:AND|EQV|IMP|NOT|OR|XOR)\b/i,punctuation:/[,;:()]/}}return xBe}var CBe,hzt;function w$r(){if(hzt)return CBe;hzt=1,CBe=n,n.displayName="batch",n.aliases=[];function n(r){(function(o){var s=/%%?[~:\w]+%?|!\S+!/,c={pattern:/\/[a-z?]+(?=[ :]|$):?|-[a-z]\b|--[a-z-]+\b/im,alias:"attr-name",inside:{punctuation:/:/}},d=/"(?:[\\"]"|[^"])*"(?!")/,h=/(?:\b|-)\d+\b/;o.languages.batch={comment:[/^::.*/m,{pattern:/((?:^|[&(])[ \t]*)rem\b(?:[^^&)\r\n]|\^(?:\r\n|[\s\S]))*/im,lookbehind:!0}],label:{pattern:/^:.*/m,alias:"property"},command:[{pattern:/((?:^|[&(])[ \t]*)for(?: \/[a-z?](?:[ :](?:"[^"]*"|[^\s"/]\S*))?)* \S+ in \([^)]+\) do/im,lookbehind:!0,inside:{keyword:/\b(?:do|in)\b|^for\b/i,string:d,parameter:c,variable:s,number:h,punctuation:/[()',]/}},{pattern:/((?:^|[&(])[ \t]*)if(?: \/[a-z?](?:[ :](?:"[^"]*"|[^\s"/]\S*))?)* (?:not )?(?:cmdextversion \d+|defined \w+|errorlevel \d+|exist \S+|(?:"[^"]*"|(?!")(?:(?!==)\S)+)?(?:==| (?:equ|geq|gtr|leq|lss|neq) )(?:"[^"]*"|[^\s"]\S*))/im,lookbehind:!0,inside:{keyword:/\b(?:cmdextversion|defined|errorlevel|exist|not)\b|^if\b/i,string:d,parameter:c,variable:s,number:h,operator:/\^|==|\b(?:equ|geq|gtr|leq|lss|neq)\b/i}},{pattern:/((?:^|[&()])[ \t]*)else\b/im,lookbehind:!0,inside:{keyword:/^else\b/i}},{pattern:/((?:^|[&(])[ \t]*)set(?: \/[a-z](?:[ :](?:"[^"]*"|[^\s"/]\S*))?)* (?:[^^&)\r\n]|\^(?:\r\n|[\s\S]))*/im,lookbehind:!0,inside:{keyword:/^set\b/i,string:d,parameter:c,variable:[s,/\w+(?=(?:[*\/%+\-&^|]|<<|>>)?=)/],number:h,operator:/[*\/%+\-&^|]=?|<<=?|>>=?|[!~_=]/,punctuation:/[()',]/}},{pattern:/((?:^|[&(])[ \t]*@?)\w+\b(?:"(?:[\\"]"|[^"])*"(?!")|[^"^&)\r\n]|\^(?:\r\n|[\s\S]))*/m,lookbehind:!0,inside:{keyword:/^\w+\b/,string:d,parameter:c,label:{pattern:/(^\s*):\S+/m,lookbehind:!0,alias:"property"},variable:s,number:h,operator:/\^/}}],operator:/[&@]/,punctuation:/[()']/}})(r)}return CBe}var ABe,pzt;function y$r(){if(pzt)return ABe;pzt=1,ABe=n,n.displayName="bbcode",n.aliases=["shortcode"];function n(r){r.languages.bbcode={tag:{pattern:/\[\/?[^\s=\]]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'"\]=]+))?(?:\s+[^\s=\]]+\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'"\]=]+))*\s*\]/,inside:{tag:{pattern:/^\[\/?[^\s=\]]+/,inside:{punctuation:/^\[\/?/}},"attr-value":{pattern:/=\s*(?:"[^"]*"|'[^']*'|[^\s'"\]=]+)/,inside:{punctuation:[/^=/,{pattern:/^(\s*)["']|["']$/,lookbehind:!0}]}},punctuation:/\]/,"attr-name":/[^\s=\]]+/}}},r.languages.shortcode=r.languages.bbcode}return ABe}var IBe,gzt;function E$r(){if(gzt)return IBe;gzt=1,IBe=n,n.displayName="bicep",n.aliases=[];function n(r){r.languages.bicep={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],property:[{pattern:/([\r\n][ \t]*)[a-z_]\w*(?=[ \t]*:)/i,lookbehind:!0},{pattern:/([\r\n][ \t]*)'(?:\\.|\$(?!\{)|[^'\\\r\n$])*'(?=[ \t]*:)/,lookbehind:!0,greedy:!0}],string:[{pattern:/'''[^'][\s\S]*?'''/,greedy:!0},{pattern:/(^|[^\\'])'(?:\\.|\$(?!\{)|[^'\\\r\n$])*'/,lookbehind:!0,greedy:!0}],"interpolated-string":{pattern:/(^|[^\\'])'(?:\\.|\$(?:(?!\{)|\{[^{}\r\n]*\})|[^'\\\r\n$])*'/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/\$\{[^{}\r\n]*\}/,inside:{expression:{pattern:/(^\$\{)[\s\S]+(?=\}$)/,lookbehind:!0},punctuation:/^\$\{|\}$/}},string:/[\s\S]+/}},datatype:{pattern:/(\b(?:output|param)\b[ \t]+\w+[ \t]+)\w+\b/,lookbehind:!0,alias:"class-name"},boolean:/\b(?:false|true)\b/,keyword:/\b(?:existing|for|if|in|module|null|output|param|resource|targetScope|var)\b/,decorator:/@\w+\b/,function:/\b[a-z_]\w*(?=[ \t]*\()/i,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,operator:/--|\+\+|\*\*=?|=>|&&=?|\|\|=?|[!=]==|<<=?|>>>?=?|[-+*/%&|^!=<>]=?|\.{3}|\?\?=?|\?\.?|[~:]/,punctuation:/[{}[\];(),.:]/},r.languages.bicep["interpolated-string"].inside.interpolation.inside.expression.inside=r.languages.bicep}return IBe}var OBe,mzt;function S$r(){if(mzt)return OBe;mzt=1,OBe=n,n.displayName="birb",n.aliases=[];function n(r){r.languages.birb=r.languages.extend("clike",{string:{pattern:/r?("|')(?:\\.|(?!\1)[^\\])*\1/,greedy:!0},"class-name":[/\b[A-Z](?:[\d_]*[a-zA-Z]\w*)?\b/,/\b(?:[A-Z]\w*|(?!(?:var|void)\b)[a-z]\w*)(?=\s+\w+\s*[;,=()])/],keyword:/\b(?:assert|break|case|class|const|default|else|enum|final|follows|for|grab|if|nest|new|next|noSeeb|return|static|switch|throw|var|void|while)\b/,operator:/\+\+|--|&&|\|\||<<=?|>>=?|~(?:\/=?)?|[+\-*\/%&^|=!<>]=?|\?|:/,variable:/\b[a-z_]\w*\b/}),r.languages.insertBefore("birb","function",{metadata:{pattern:/<\w+>/,greedy:!0,alias:"symbol"}})}return OBe}var NBe,bzt;function _$r(){if(bzt)return NBe;bzt=1;var n=vU();NBe=r,r.displayName="bison",r.aliases=[];function r(o){o.register(n),o.languages.bison=o.languages.extend("c",{}),o.languages.insertBefore("bison","comment",{bison:{pattern:/^(?:[^%]|%(?!%))*%%[\s\S]*?%%/,inside:{c:{pattern:/%\{[\s\S]*?%\}|\{(?:\{[^}]*\}|[^{}])*\}/,inside:{delimiter:{pattern:/^%?\{|%?\}$/,alias:"punctuation"},"bison-variable":{pattern:/[$@](?:<[^\s>]+>)?[\w$]+/,alias:"variable",inside:{punctuation:/<|>/}},rest:o.languages.c}},comment:o.languages.c.comment,string:o.languages.c.string,property:/\S+(?=:)/,keyword:/%\w+/,number:{pattern:/(^|[^@])\b(?:0x[\da-f]+|\d+)/i,lookbehind:!0},punctuation:/%[%?]|[|:;\[\]<>]/}}})}return NBe}var RBe,vzt;function k$r(){if(vzt)return RBe;vzt=1,RBe=n,n.displayName="bnf",n.aliases=["rbnf"];function n(r){r.languages.bnf={string:{pattern:/"[^\r\n"]*"|'[^\r\n']*'/},definition:{pattern:/<[^<>\r\n\t]+>(?=\s*::=)/,alias:["rule","keyword"],inside:{punctuation:/^<|>$/}},rule:{pattern:/<[^<>\r\n\t]+>/,inside:{punctuation:/^<|>$/}},operator:/::=|[|()[\]{}*+?]|\.{3}/},r.languages.rbnf=r.languages.bnf}return RBe}var PBe,wzt;function T$r(){if(wzt)return PBe;wzt=1,PBe=n,n.displayName="brainfuck",n.aliases=[];function n(r){r.languages.brainfuck={pointer:{pattern:/<|>/,alias:"keyword"},increment:{pattern:/\+/,alias:"inserted"},decrement:{pattern:/-/,alias:"deleted"},branching:{pattern:/\[|\]/,alias:"important"},operator:/[.,]/,comment:/\S+/}}return PBe}var LBe,yzt;function x$r(){if(yzt)return LBe;yzt=1,LBe=n,n.displayName="brightscript",n.aliases=[];function n(r){r.languages.brightscript={comment:/(?:\brem|').*/i,"directive-statement":{pattern:/(^[\t ]*)#(?:const|else(?:[\t ]+if)?|end[\t ]+if|error|if).*/im,lookbehind:!0,alias:"property",inside:{"error-message":{pattern:/(^#error).+/,lookbehind:!0},directive:{pattern:/^#(?:const|else(?:[\t ]+if)?|end[\t ]+if|error|if)/,alias:"keyword"},expression:{pattern:/[\s\S]+/,inside:null}}},property:{pattern:/([\r\n{,][\t ]*)(?:(?!\d)\w+|"(?:[^"\r\n]|"")*"(?!"))(?=[ \t]*:)/,lookbehind:!0,greedy:!0},string:{pattern:/"(?:[^"\r\n]|"")*"(?!")/,greedy:!0},"class-name":{pattern:/(\bAs[\t ]+)\w+/i,lookbehind:!0},keyword:/\b(?:As|Dim|Each|Else|Elseif|End|Exit|For|Function|Goto|If|In|Print|Return|Step|Stop|Sub|Then|To|While)\b/i,boolean:/\b(?:false|true)\b/i,function:/\b(?!\d)\w+(?=[\t ]*\()/,number:/(?:\b\d+(?:\.\d+)?(?:[ed][+-]\d+)?|&h[a-f\d]+)\b[%&!#]?/i,operator:/--|\+\+|>>=?|<<=?|<>|[-+*/\\<>]=?|[:^=?]|\b(?:and|mod|not|or)\b/i,punctuation:/[.,;()[\]{}]/,constant:/\b(?:LINE_NUM)\b/i},r.languages.brightscript["directive-statement"].inside.expression.inside=r.languages.brightscript}return LBe}var DBe,Ezt;function C$r(){if(Ezt)return DBe;Ezt=1,DBe=n,n.displayName="bro",n.aliases=[];function n(r){r.languages.bro={comment:{pattern:/(^|[^\\$])#.*/,lookbehind:!0,inside:{italic:/\b(?:FIXME|TODO|XXX)\b/}},string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},boolean:/\b[TF]\b/,function:{pattern:/(\b(?:event|function|hook)[ \t]+)\w+(?:::\w+)?/,lookbehind:!0},builtin:/(?:@(?:load(?:-(?:plugin|sigs))?|unload|prefixes|ifn?def|else|(?:end)?if|DIR|FILENAME))|(?:&?(?:add_func|create_expire|default|delete_func|encrypt|error_handler|expire_func|group|log|mergeable|optional|persistent|priority|raw_output|read_expire|redef|rotate_interval|rotate_size|synchronized|type_column|write_expire))/,constant:{pattern:/(\bconst[ \t]+)\w+/i,lookbehind:!0},keyword:/\b(?:add|addr|alarm|any|bool|break|const|continue|count|delete|double|else|enum|event|export|file|for|function|global|hook|if|in|int|interval|local|module|next|of|opaque|pattern|port|print|record|return|schedule|set|string|subnet|table|time|timeout|using|vector|when)\b/,operator:/--?|\+\+?|!=?=?|<=?|>=?|==?=?|&&|\|\|?|\?|\*|\/|~|\^|%/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,punctuation:/[{}[\];(),.:]/}}return DBe}var MBe,Szt;function A$r(){if(Szt)return MBe;Szt=1,MBe=n,n.displayName="bsl",n.aliases=[];function n(r){r.languages.bsl={comment:/\/\/.*/,string:[{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},{pattern:/'(?:[^'\r\n\\]|\\.)*'/}],keyword:[{pattern:/(^|[^\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])(?:пока|для|новый|прервать|попытка|исключение|вызватьисключение|иначе|конецпопытки|неопределено|функция|перем|возврат|конецфункции|если|иначеесли|процедура|конецпроцедуры|тогда|знач|экспорт|конецесли|из|каждого|истина|ложь|по|цикл|конеццикла|выполнить)(?![\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])/i,lookbehind:!0},{pattern:/\b(?:break|do|each|else|elseif|enddo|endfunction|endif|endprocedure|endtry|except|execute|export|false|for|function|if|in|new|null|procedure|raise|return|then|to|true|try|undefined|val|var|while)\b/i}],number:{pattern:/(^(?=\d)|[^\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])(?:\d+(?:\.\d*)?|\.\d+)(?:E[+-]?\d+)?/i,lookbehind:!0},operator:[/[<>+\-*/]=?|[%=]/,{pattern:/(^|[^\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])(?:и|или|не)(?![\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])/i,lookbehind:!0},{pattern:/\b(?:and|not|or)\b/i}],punctuation:/\(\.|\.\)|[()\[\]:;,.]/,directive:[{pattern:/^([ \t]*)&.*/m,lookbehind:!0,greedy:!0,alias:"important"},{pattern:/^([ \t]*)#.*/gm,lookbehind:!0,greedy:!0,alias:"important"}]},r.languages.oscript=r.languages.bsl}return MBe}var jBe,_zt;function I$r(){if(_zt)return jBe;_zt=1,jBe=n,n.displayName="cfscript",n.aliases=[];function n(r){r.languages.cfscript=r.languages.extend("clike",{comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,inside:{annotation:{pattern:/(?:^|[^.])@[\w\.]+/,alias:"punctuation"}}},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],keyword:/\b(?:abstract|break|catch|component|continue|default|do|else|extends|final|finally|for|function|if|in|include|package|private|property|public|remote|required|rethrow|return|static|switch|throw|try|var|while|xml)\b(?!\s*=)/,operator:[/\+\+|--|&&|\|\||::|=>|[!=]==|<=?|>=?|[-+*/%&|^!=<>]=?|\?(?:\.|:)?|[?:]/,/\b(?:and|contains|eq|equal|eqv|gt|gte|imp|is|lt|lte|mod|not|or|xor)\b/],scope:{pattern:/\b(?:application|arguments|cgi|client|cookie|local|session|super|this|variables)\b/,alias:"global"},type:{pattern:/\b(?:any|array|binary|boolean|date|guid|numeric|query|string|struct|uuid|void|xml)\b/,alias:"builtin"}}),r.languages.insertBefore("cfscript","keyword",{"function-variable":{pattern:/[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*[=:]\s*(?:\bfunction\b|(?:\((?:[^()]|\([^()]*\))*\)|(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)\s*=>))/,alias:"function"}}),delete r.languages.cfscript["class-name"],r.languages.cfc=r.languages.cfscript}return jBe}var $Be,kzt;function O$r(){if(kzt)return $Be;kzt=1;var n=yQe();$Be=r,r.displayName="chaiscript",r.aliases=[];function r(o){o.register(n),o.languages.chaiscript=o.languages.extend("clike",{string:{pattern:/(^|[^\\])'(?:[^'\\]|\\[\s\S])*'/,lookbehind:!0,greedy:!0},"class-name":[{pattern:/(\bclass\s+)\w+/,lookbehind:!0},{pattern:/(\b(?:attr|def)\s+)\w+(?=\s*::)/,lookbehind:!0}],keyword:/\b(?:attr|auto|break|case|catch|class|continue|def|default|else|finally|for|fun|global|if|return|switch|this|try|var|while)\b/,number:[o.languages.cpp.number,/\b(?:Infinity|NaN)\b/],operator:/>>=?|<<=?|\|\||&&|:[:=]?|--|\+\+|[=!<>+\-*/%|&^]=?|[?~]|`[^`\r\n]{1,4}`/}),o.languages.insertBefore("chaiscript","operator",{"parameter-type":{pattern:/([,(]\s*)\w+(?=\s+\w)/,lookbehind:!0,alias:"class-name"}}),o.languages.insertBefore("chaiscript","string",{"string-interpolation":{pattern:/(^|[^\\])"(?:[^"$\\]|\\[\s\S]|\$(?!\{)|\$\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\})*"/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\}/,lookbehind:!0,inside:{"interpolation-expression":{pattern:/(^\$\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:o.languages.chaiscript},"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"}}},string:/[\s\S]+/}}})}return $Be}var FBe,Tzt;function N$r(){if(Tzt)return FBe;Tzt=1,FBe=n,n.displayName="cil",n.aliases=[];function n(r){r.languages.cil={comment:/\/\/.*/,string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},directive:{pattern:/(^|\W)\.[a-z]+(?=\s)/,lookbehind:!0,alias:"class-name"},variable:/\[[\w\.]+\]/,keyword:/\b(?:abstract|ansi|assembly|auto|autochar|beforefieldinit|bool|bstr|byvalstr|catch|char|cil|class|currency|date|decimal|default|enum|error|explicit|extends|extern|famandassem|family|famorassem|final(?:ly)?|float32|float64|hidebysig|u?int(?:8|16|32|64)?|iant|idispatch|implements|import|initonly|instance|interface|iunknown|literal|lpstr|lpstruct|lptstr|lpwstr|managed|method|native(?:Type)?|nested|newslot|object(?:ref)?|pinvokeimpl|private|privatescope|public|reqsecobj|rtspecialname|runtime|sealed|sequential|serializable|specialname|static|string|struct|syschar|tbstr|unicode|unmanagedexp|unsigned|value(?:type)?|variant|virtual|void)\b/,function:/\b(?:(?:constrained|no|readonly|tail|unaligned|volatile)\.)?(?:conv\.(?:[iu][1248]?|ovf\.[iu][1248]?(?:\.un)?|r\.un|r4|r8)|ldc\.(?:i4(?:\.\d+|\.[mM]1|\.s)?|i8|r4|r8)|ldelem(?:\.[iu][1248]?|\.r[48]|\.ref|a)?|ldind\.(?:[iu][1248]?|r[48]|ref)|stelem\.?(?:i[1248]?|r[48]|ref)?|stind\.(?:i[1248]?|r[48]|ref)?|end(?:fault|filter|finally)|ldarg(?:\.[0-3s]|a(?:\.s)?)?|ldloc(?:\.\d+|\.s)?|sub(?:\.ovf(?:\.un)?)?|mul(?:\.ovf(?:\.un)?)?|add(?:\.ovf(?:\.un)?)?|stloc(?:\.[0-3s])?|refany(?:type|val)|blt(?:\.un)?(?:\.s)?|ble(?:\.un)?(?:\.s)?|bgt(?:\.un)?(?:\.s)?|bge(?:\.un)?(?:\.s)?|unbox(?:\.any)?|init(?:blk|obj)|call(?:i|virt)?|brfalse(?:\.s)?|bne\.un(?:\.s)?|ldloca(?:\.s)?|brzero(?:\.s)?|brtrue(?:\.s)?|brnull(?:\.s)?|brinst(?:\.s)?|starg(?:\.s)?|leave(?:\.s)?|shr(?:\.un)?|rem(?:\.un)?|div(?:\.un)?|clt(?:\.un)?|alignment|castclass|ldvirtftn|beq(?:\.s)?|ckfinite|ldsflda|ldtoken|localloc|mkrefany|rethrow|cgt\.un|arglist|switch|stsfld|sizeof|newobj|newarr|ldsfld|ldnull|ldflda|isinst|throw|stobj|stfld|ldstr|ldobj|ldlen|ldftn|ldfld|cpobj|cpblk|break|br\.s|xor|shl|ret|pop|not|nop|neg|jmp|dup|cgt|ceq|box|and|or|br)\b/,boolean:/\b(?:false|true)\b/,number:/\b-?(?:0x[0-9a-f]+|\d+)(?:\.[0-9a-f]+)?\b/i,punctuation:/[{}[\];(),:=]|IL_[0-9A-Za-z]+/}}return FBe}var BBe,xzt;function R$r(){if(xzt)return BBe;xzt=1,BBe=n,n.displayName="clojure",n.aliases=[];function n(r){r.languages.clojure={comment:{pattern:/;.*/,greedy:!0},string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0},char:/\\\w+/,symbol:{pattern:/(^|[\s()\[\]{},])::?[\w*+!?'<>=/.-]+/,lookbehind:!0},keyword:{pattern:/(\()(?:-|->|->>|\.|\.\.|\*|\/|\+|<|<=|=|==|>|>=|accessor|agent|agent-errors|aget|alength|all-ns|alter|and|append-child|apply|array-map|aset|aset-boolean|aset-byte|aset-char|aset-double|aset-float|aset-int|aset-long|aset-short|assert|assoc|await|await-for|bean|binding|bit-and|bit-not|bit-or|bit-shift-left|bit-shift-right|bit-xor|boolean|branch\?|butlast|byte|cast|char|children|class|clear-agent-errors|comment|commute|comp|comparator|complement|concat|cond|conj|cons|constantly|construct-proxy|contains\?|count|create-ns|create-struct|cycle|dec|declare|def|def-|definline|definterface|defmacro|defmethod|defmulti|defn|defn-|defonce|defproject|defprotocol|defrecord|defstruct|deftype|deref|difference|disj|dissoc|distinct|do|doall|doc|dorun|doseq|dosync|dotimes|doto|double|down|drop|drop-while|edit|end\?|ensure|eval|every\?|false\?|ffirst|file-seq|filter|find|find-doc|find-ns|find-var|first|float|flush|fn|fnseq|for|frest|gensym|get|get-proxy-class|hash-map|hash-set|identical\?|identity|if|if-let|if-not|import|in-ns|inc|index|insert-child|insert-left|insert-right|inspect-table|inspect-tree|instance\?|int|interleave|intersection|into|into-array|iterate|join|key|keys|keyword|keyword\?|last|lazy-cat|lazy-cons|left|lefts|let|line-seq|list|list\*|load|load-file|locking|long|loop|macroexpand|macroexpand-1|make-array|make-node|map|map-invert|map\?|mapcat|max|max-key|memfn|merge|merge-with|meta|min|min-key|monitor-enter|name|namespace|neg\?|new|newline|next|nil\?|node|not|not-any\?|not-every\?|not=|ns|ns-imports|ns-interns|ns-map|ns-name|ns-publics|ns-refers|ns-resolve|ns-unmap|nth|nthrest|or|parse|partial|path|peek|pop|pos\?|pr|pr-str|print|print-str|println|println-str|prn|prn-str|project|proxy|proxy-mappings|quot|quote|rand|rand-int|range|re-find|re-groups|re-matcher|re-matches|re-pattern|re-seq|read|read-line|recur|reduce|ref|ref-set|refer|rem|remove|remove-method|remove-ns|rename|rename-keys|repeat|replace|replicate|resolve|rest|resultset-seq|reverse|rfirst|right|rights|root|rrest|rseq|second|select|select-keys|send|send-off|seq|seq-zip|seq\?|set|set!|short|slurp|some|sort|sort-by|sorted-map|sorted-map-by|sorted-set|special-symbol\?|split-at|split-with|str|string\?|struct|struct-map|subs|subvec|symbol|symbol\?|sync|take|take-nth|take-while|test|throw|time|to-array|to-array-2d|tree-seq|true\?|try|union|up|update-proxy|val|vals|var|var-get|var-set|var\?|vector|vector-zip|vector\?|when|when-first|when-let|when-not|with-local-vars|with-meta|with-open|with-out-str|xml-seq|xml-zip|zero\?|zipmap|zipper)(?=[\s)]|$)/,lookbehind:!0},boolean:/\b(?:false|nil|true)\b/,number:{pattern:/(^|[^\w$@])(?:\d+(?:[/.]\d+)?(?:e[+-]?\d+)?|0x[a-f0-9]+|[1-9]\d?r[a-z0-9]+)[lmn]?(?![\w$@])/i,lookbehind:!0},function:{pattern:/((?:^|[^'])\()[\w*+!?'<>=/.-]+(?=[\s)]|$)/,lookbehind:!0},operator:/[#@^`~]/,punctuation:/[{}\[\](),]/}}return BBe}var UBe,Czt;function P$r(){if(Czt)return UBe;Czt=1,UBe=n,n.displayName="cmake",n.aliases=[];function n(r){r.languages.cmake={comment:/#.*/,string:{pattern:/"(?:[^\\"]|\\.)*"/,greedy:!0,inside:{interpolation:{pattern:/\$\{(?:[^{}$]|\$\{[^{}$]*\})*\}/,inside:{punctuation:/\$\{|\}/,variable:/\w+/}}}},variable:/\b(?:CMAKE_\w+|\w+_(?:(?:BINARY|SOURCE)_DIR|DESCRIPTION|HOMEPAGE_URL|ROOT|VERSION(?:_MAJOR|_MINOR|_PATCH|_TWEAK)?)|(?:ANDROID|APPLE|BORLAND|BUILD_SHARED_LIBS|CACHE|CPACK_(?:ABSOLUTE_DESTINATION_FILES|COMPONENT_INCLUDE_TOPLEVEL_DIRECTORY|ERROR_ON_ABSOLUTE_INSTALL_DESTINATION|INCLUDE_TOPLEVEL_DIRECTORY|INSTALL_DEFAULT_DIRECTORY_PERMISSIONS|INSTALL_SCRIPT|PACKAGING_INSTALL_PREFIX|SET_DESTDIR|WARN_ON_ABSOLUTE_INSTALL_DESTINATION)|CTEST_(?:BINARY_DIRECTORY|BUILD_COMMAND|BUILD_NAME|BZR_COMMAND|BZR_UPDATE_OPTIONS|CHANGE_ID|CHECKOUT_COMMAND|CONFIGURATION_TYPE|CONFIGURE_COMMAND|COVERAGE_COMMAND|COVERAGE_EXTRA_FLAGS|CURL_OPTIONS|CUSTOM_(?:COVERAGE_EXCLUDE|ERROR_EXCEPTION|ERROR_MATCH|ERROR_POST_CONTEXT|ERROR_PRE_CONTEXT|MAXIMUM_FAILED_TEST_OUTPUT_SIZE|MAXIMUM_NUMBER_OF_(?:ERRORS|WARNINGS)|MAXIMUM_PASSED_TEST_OUTPUT_SIZE|MEMCHECK_IGNORE|POST_MEMCHECK|POST_TEST|PRE_MEMCHECK|PRE_TEST|TESTS_IGNORE|WARNING_EXCEPTION|WARNING_MATCH)|CVS_CHECKOUT|CVS_COMMAND|CVS_UPDATE_OPTIONS|DROP_LOCATION|DROP_METHOD|DROP_SITE|DROP_SITE_CDASH|DROP_SITE_PASSWORD|DROP_SITE_USER|EXTRA_COVERAGE_GLOB|GIT_COMMAND|GIT_INIT_SUBMODULES|GIT_UPDATE_CUSTOM|GIT_UPDATE_OPTIONS|HG_COMMAND|HG_UPDATE_OPTIONS|LABELS_FOR_SUBPROJECTS|MEMORYCHECK_(?:COMMAND|COMMAND_OPTIONS|SANITIZER_OPTIONS|SUPPRESSIONS_FILE|TYPE)|NIGHTLY_START_TIME|P4_CLIENT|P4_COMMAND|P4_OPTIONS|P4_UPDATE_OPTIONS|RUN_CURRENT_SCRIPT|SCP_COMMAND|SITE|SOURCE_DIRECTORY|SUBMIT_URL|SVN_COMMAND|SVN_OPTIONS|SVN_UPDATE_OPTIONS|TEST_LOAD|TEST_TIMEOUT|TRIGGER_SITE|UPDATE_COMMAND|UPDATE_OPTIONS|UPDATE_VERSION_ONLY|USE_LAUNCHERS)|CYGWIN|ENV|EXECUTABLE_OUTPUT_PATH|GHS-MULTI|IOS|LIBRARY_OUTPUT_PATH|MINGW|MSVC(?:10|11|12|14|60|70|71|80|90|_IDE|_TOOLSET_VERSION|_VERSION)?|MSYS|PROJECT_(?:BINARY_DIR|DESCRIPTION|HOMEPAGE_URL|NAME|SOURCE_DIR|VERSION|VERSION_(?:MAJOR|MINOR|PATCH|TWEAK))|UNIX|WIN32|WINCE|WINDOWS_PHONE|WINDOWS_STORE|XCODE|XCODE_VERSION))\b/,property:/\b(?:cxx_\w+|(?:ARCHIVE_OUTPUT_(?:DIRECTORY|NAME)|COMPILE_DEFINITIONS|COMPILE_PDB_NAME|COMPILE_PDB_OUTPUT_DIRECTORY|EXCLUDE_FROM_DEFAULT_BUILD|IMPORTED_(?:IMPLIB|LIBNAME|LINK_DEPENDENT_LIBRARIES|LINK_INTERFACE_LANGUAGES|LINK_INTERFACE_LIBRARIES|LINK_INTERFACE_MULTIPLICITY|LOCATION|NO_SONAME|OBJECTS|SONAME)|INTERPROCEDURAL_OPTIMIZATION|LIBRARY_OUTPUT_DIRECTORY|LIBRARY_OUTPUT_NAME|LINK_FLAGS|LINK_INTERFACE_LIBRARIES|LINK_INTERFACE_MULTIPLICITY|LOCATION|MAP_IMPORTED_CONFIG|OSX_ARCHITECTURES|OUTPUT_NAME|PDB_NAME|PDB_OUTPUT_DIRECTORY|RUNTIME_OUTPUT_DIRECTORY|RUNTIME_OUTPUT_NAME|STATIC_LIBRARY_FLAGS|VS_CSHARP|VS_DOTNET_REFERENCEPROP|VS_DOTNET_REFERENCE|VS_GLOBAL_SECTION_POST|VS_GLOBAL_SECTION_PRE|VS_GLOBAL|XCODE_ATTRIBUTE)_\w+|\w+_(?:CLANG_TIDY|COMPILER_LAUNCHER|CPPCHECK|CPPLINT|INCLUDE_WHAT_YOU_USE|OUTPUT_NAME|POSTFIX|VISIBILITY_PRESET)|ABSTRACT|ADDITIONAL_MAKE_CLEAN_FILES|ADVANCED|ALIASED_TARGET|ALLOW_DUPLICATE_CUSTOM_TARGETS|ANDROID_(?:ANT_ADDITIONAL_OPTIONS|API|API_MIN|ARCH|ASSETS_DIRECTORIES|GUI|JAR_DEPENDENCIES|NATIVE_LIB_DEPENDENCIES|NATIVE_LIB_DIRECTORIES|PROCESS_MAX|PROGUARD|PROGUARD_CONFIG_PATH|SECURE_PROPS_PATH|SKIP_ANT_STEP|STL_TYPE)|ARCHIVE_OUTPUT_DIRECTORY|ATTACHED_FILES|ATTACHED_FILES_ON_FAIL|AUTOGEN_(?:BUILD_DIR|ORIGIN_DEPENDS|PARALLEL|SOURCE_GROUP|TARGETS_FOLDER|TARGET_DEPENDS)|AUTOMOC|AUTOMOC_(?:COMPILER_PREDEFINES|DEPEND_FILTERS|EXECUTABLE|MACRO_NAMES|MOC_OPTIONS|SOURCE_GROUP|TARGETS_FOLDER)|AUTORCC|AUTORCC_EXECUTABLE|AUTORCC_OPTIONS|AUTORCC_SOURCE_GROUP|AUTOUIC|AUTOUIC_EXECUTABLE|AUTOUIC_OPTIONS|AUTOUIC_SEARCH_PATHS|BINARY_DIR|BUILDSYSTEM_TARGETS|BUILD_RPATH|BUILD_RPATH_USE_ORIGIN|BUILD_WITH_INSTALL_NAME_DIR|BUILD_WITH_INSTALL_RPATH|BUNDLE|BUNDLE_EXTENSION|CACHE_VARIABLES|CLEAN_NO_CUSTOM|COMMON_LANGUAGE_RUNTIME|COMPATIBLE_INTERFACE_(?:BOOL|NUMBER_MAX|NUMBER_MIN|STRING)|COMPILE_(?:DEFINITIONS|FEATURES|FLAGS|OPTIONS|PDB_NAME|PDB_OUTPUT_DIRECTORY)|COST|CPACK_DESKTOP_SHORTCUTS|CPACK_NEVER_OVERWRITE|CPACK_PERMANENT|CPACK_STARTUP_SHORTCUTS|CPACK_START_MENU_SHORTCUTS|CPACK_WIX_ACL|CROSSCOMPILING_EMULATOR|CUDA_EXTENSIONS|CUDA_PTX_COMPILATION|CUDA_RESOLVE_DEVICE_SYMBOLS|CUDA_SEPARABLE_COMPILATION|CUDA_STANDARD|CUDA_STANDARD_REQUIRED|CXX_EXTENSIONS|CXX_STANDARD|CXX_STANDARD_REQUIRED|C_EXTENSIONS|C_STANDARD|C_STANDARD_REQUIRED|DEBUG_CONFIGURATIONS|DEFINE_SYMBOL|DEFINITIONS|DEPENDS|DEPLOYMENT_ADDITIONAL_FILES|DEPLOYMENT_REMOTE_DIRECTORY|DISABLED|DISABLED_FEATURES|ECLIPSE_EXTRA_CPROJECT_CONTENTS|ECLIPSE_EXTRA_NATURES|ENABLED_FEATURES|ENABLED_LANGUAGES|ENABLE_EXPORTS|ENVIRONMENT|EXCLUDE_FROM_ALL|EXCLUDE_FROM_DEFAULT_BUILD|EXPORT_NAME|EXPORT_PROPERTIES|EXTERNAL_OBJECT|EchoString|FAIL_REGULAR_EXPRESSION|FIND_LIBRARY_USE_LIB32_PATHS|FIND_LIBRARY_USE_LIB64_PATHS|FIND_LIBRARY_USE_LIBX32_PATHS|FIND_LIBRARY_USE_OPENBSD_VERSIONING|FIXTURES_CLEANUP|FIXTURES_REQUIRED|FIXTURES_SETUP|FOLDER|FRAMEWORK|Fortran_FORMAT|Fortran_MODULE_DIRECTORY|GENERATED|GENERATOR_FILE_NAME|GENERATOR_IS_MULTI_CONFIG|GHS_INTEGRITY_APP|GHS_NO_SOURCE_GROUP_FILE|GLOBAL_DEPENDS_DEBUG_MODE|GLOBAL_DEPENDS_NO_CYCLES|GNUtoMS|HAS_CXX|HEADER_FILE_ONLY|HELPSTRING|IMPLICIT_DEPENDS_INCLUDE_TRANSFORM|IMPORTED|IMPORTED_(?:COMMON_LANGUAGE_RUNTIME|CONFIGURATIONS|GLOBAL|IMPLIB|LIBNAME|LINK_DEPENDENT_LIBRARIES|LINK_INTERFACE_(?:LANGUAGES|LIBRARIES|MULTIPLICITY)|LOCATION|NO_SONAME|OBJECTS|SONAME)|IMPORT_PREFIX|IMPORT_SUFFIX|INCLUDE_DIRECTORIES|INCLUDE_REGULAR_EXPRESSION|INSTALL_NAME_DIR|INSTALL_RPATH|INSTALL_RPATH_USE_LINK_PATH|INTERFACE_(?:AUTOUIC_OPTIONS|COMPILE_DEFINITIONS|COMPILE_FEATURES|COMPILE_OPTIONS|INCLUDE_DIRECTORIES|LINK_DEPENDS|LINK_DIRECTORIES|LINK_LIBRARIES|LINK_OPTIONS|POSITION_INDEPENDENT_CODE|SOURCES|SYSTEM_INCLUDE_DIRECTORIES)|INTERPROCEDURAL_OPTIMIZATION|IN_TRY_COMPILE|IOS_INSTALL_COMBINED|JOB_POOLS|JOB_POOL_COMPILE|JOB_POOL_LINK|KEEP_EXTENSION|LABELS|LANGUAGE|LIBRARY_OUTPUT_DIRECTORY|LINKER_LANGUAGE|LINK_(?:DEPENDS|DEPENDS_NO_SHARED|DIRECTORIES|FLAGS|INTERFACE_LIBRARIES|INTERFACE_MULTIPLICITY|LIBRARIES|OPTIONS|SEARCH_END_STATIC|SEARCH_START_STATIC|WHAT_YOU_USE)|LISTFILE_STACK|LOCATION|MACOSX_BUNDLE|MACOSX_BUNDLE_INFO_PLIST|MACOSX_FRAMEWORK_INFO_PLIST|MACOSX_PACKAGE_LOCATION|MACOSX_RPATH|MACROS|MANUALLY_ADDED_DEPENDENCIES|MEASUREMENT|MODIFIED|NAME|NO_SONAME|NO_SYSTEM_FROM_IMPORTED|OBJECT_DEPENDS|OBJECT_OUTPUTS|OSX_ARCHITECTURES|OUTPUT_NAME|PACKAGES_FOUND|PACKAGES_NOT_FOUND|PARENT_DIRECTORY|PASS_REGULAR_EXPRESSION|PDB_NAME|PDB_OUTPUT_DIRECTORY|POSITION_INDEPENDENT_CODE|POST_INSTALL_SCRIPT|PREDEFINED_TARGETS_FOLDER|PREFIX|PRE_INSTALL_SCRIPT|PRIVATE_HEADER|PROCESSORS|PROCESSOR_AFFINITY|PROJECT_LABEL|PUBLIC_HEADER|REPORT_UNDEFINED_PROPERTIES|REQUIRED_FILES|RESOURCE|RESOURCE_LOCK|RULE_LAUNCH_COMPILE|RULE_LAUNCH_CUSTOM|RULE_LAUNCH_LINK|RULE_MESSAGES|RUNTIME_OUTPUT_DIRECTORY|RUN_SERIAL|SKIP_AUTOGEN|SKIP_AUTOMOC|SKIP_AUTORCC|SKIP_AUTOUIC|SKIP_BUILD_RPATH|SKIP_RETURN_CODE|SOURCES|SOURCE_DIR|SOVERSION|STATIC_LIBRARY_FLAGS|STATIC_LIBRARY_OPTIONS|STRINGS|SUBDIRECTORIES|SUFFIX|SYMBOLIC|TARGET_ARCHIVES_MAY_BE_SHARED_LIBS|TARGET_MESSAGES|TARGET_SUPPORTS_SHARED_LIBS|TESTS|TEST_INCLUDE_FILE|TEST_INCLUDE_FILES|TIMEOUT|TIMEOUT_AFTER_MATCH|TYPE|USE_FOLDERS|VALUE|VARIABLES|VERSION|VISIBILITY_INLINES_HIDDEN|VS_(?:CONFIGURATION_TYPE|COPY_TO_OUT_DIR|DEBUGGER_(?:COMMAND|COMMAND_ARGUMENTS|ENVIRONMENT|WORKING_DIRECTORY)|DEPLOYMENT_CONTENT|DEPLOYMENT_LOCATION|DOTNET_REFERENCES|DOTNET_REFERENCES_COPY_LOCAL|GLOBAL_KEYWORD|GLOBAL_PROJECT_TYPES|GLOBAL_ROOTNAMESPACE|INCLUDE_IN_VSIX|IOT_STARTUP_TASK|KEYWORD|RESOURCE_GENERATOR|SCC_AUXPATH|SCC_LOCALPATH|SCC_PROJECTNAME|SCC_PROVIDER|SDK_REFERENCES|SHADER_(?:DISABLE_OPTIMIZATIONS|ENABLE_DEBUG|ENTRYPOINT|FLAGS|MODEL|OBJECT_FILE_NAME|OUTPUT_HEADER_FILE|TYPE|VARIABLE_NAME)|STARTUP_PROJECT|TOOL_OVERRIDE|USER_PROPS|WINRT_COMPONENT|WINRT_EXTENSIONS|WINRT_REFERENCES|XAML_TYPE)|WILL_FAIL|WIN32_EXECUTABLE|WINDOWS_EXPORT_ALL_SYMBOLS|WORKING_DIRECTORY|WRAP_EXCLUDE|XCODE_(?:EMIT_EFFECTIVE_PLATFORM_NAME|EXPLICIT_FILE_TYPE|FILE_ATTRIBUTES|LAST_KNOWN_FILE_TYPE|PRODUCT_TYPE|SCHEME_(?:ADDRESS_SANITIZER|ADDRESS_SANITIZER_USE_AFTER_RETURN|ARGUMENTS|DISABLE_MAIN_THREAD_CHECKER|DYNAMIC_LIBRARY_LOADS|DYNAMIC_LINKER_API_USAGE|ENVIRONMENT|EXECUTABLE|GUARD_MALLOC|MAIN_THREAD_CHECKER_STOP|MALLOC_GUARD_EDGES|MALLOC_SCRIBBLE|MALLOC_STACK|THREAD_SANITIZER(?:_STOP)?|UNDEFINED_BEHAVIOUR_SANITIZER(?:_STOP)?|ZOMBIE_OBJECTS))|XCTEST)\b/,keyword:/\b(?:add_compile_definitions|add_compile_options|add_custom_command|add_custom_target|add_definitions|add_dependencies|add_executable|add_library|add_link_options|add_subdirectory|add_test|aux_source_directory|break|build_command|build_name|cmake_host_system_information|cmake_minimum_required|cmake_parse_arguments|cmake_policy|configure_file|continue|create_test_sourcelist|ctest_build|ctest_configure|ctest_coverage|ctest_empty_binary_directory|ctest_memcheck|ctest_read_custom_files|ctest_run_script|ctest_sleep|ctest_start|ctest_submit|ctest_test|ctest_update|ctest_upload|define_property|else|elseif|enable_language|enable_testing|endforeach|endfunction|endif|endmacro|endwhile|exec_program|execute_process|export|export_library_dependencies|file|find_file|find_library|find_package|find_path|find_program|fltk_wrap_ui|foreach|function|get_cmake_property|get_directory_property|get_filename_component|get_property|get_source_file_property|get_target_property|get_test_property|if|include|include_directories|include_external_msproject|include_guard|include_regular_expression|install|install_files|install_programs|install_targets|link_directories|link_libraries|list|load_cache|load_command|macro|make_directory|mark_as_advanced|math|message|option|output_required_files|project|qt_wrap_cpp|qt_wrap_ui|remove|remove_definitions|return|separate_arguments|set|set_directory_properties|set_property|set_source_files_properties|set_target_properties|set_tests_properties|site_name|source_group|string|subdir_depends|subdirs|target_compile_definitions|target_compile_features|target_compile_options|target_include_directories|target_link_directories|target_link_libraries|target_link_options|target_sources|try_compile|try_run|unset|use_mangled_mesa|utility_source|variable_requires|variable_watch|while|write_file)(?=\s*\()\b/,boolean:/\b(?:FALSE|OFF|ON|TRUE)\b/,namespace:/\b(?:INTERFACE|PRIVATE|PROPERTIES|PUBLIC|SHARED|STATIC|TARGET_OBJECTS)\b/,operator:/\b(?:AND|DEFINED|EQUAL|GREATER|LESS|MATCHES|NOT|OR|STREQUAL|STRGREATER|STRLESS|VERSION_EQUAL|VERSION_GREATER|VERSION_LESS)\b/,inserted:{pattern:/\b\w+::\w+\b/,alias:"class-name"},number:/\b\d+(?:\.\d+)*\b/,function:/\b[a-z_]\w*(?=\s*\()\b/i,punctuation:/[()>}]|\$[<{]/}}return UBe}var zBe,Azt;function L$r(){if(Azt)return zBe;Azt=1,zBe=n,n.displayName="cobol",n.aliases=[];function n(r){r.languages.cobol={comment:{pattern:/\*>.*|(^[ \t]*)\*.*/m,lookbehind:!0,greedy:!0},string:{pattern:/[xzgn]?(?:"(?:[^\r\n"]|"")*"(?!")|'(?:[^\r\n']|'')*'(?!'))/i,greedy:!0},level:{pattern:/(^[ \t]*)\d+\b/m,lookbehind:!0,greedy:!0,alias:"number"},"class-name":{pattern:/(\bpic(?:ture)?\s+)(?:(?:[-\w$/,:*+<>]|\.(?!\s|$))(?:\(\d+\))?)+/i,lookbehind:!0,inside:{number:{pattern:/(\()\d+/,lookbehind:!0},punctuation:/[()]/}},keyword:{pattern:/(^|[^\w-])(?:ABORT|ACCEPT|ACCESS|ADD|ADDRESS|ADVANCING|AFTER|ALIGNED|ALL|ALPHABET|ALPHABETIC|ALPHABETIC-LOWER|ALPHABETIC-UPPER|ALPHANUMERIC|ALPHANUMERIC-EDITED|ALSO|ALTER|ALTERNATE|ANY|ARE|AREA|AREAS|AS|ASCENDING|ASCII|ASSIGN|ASSOCIATED-DATA|ASSOCIATED-DATA-LENGTH|AT|ATTRIBUTE|AUTHOR|AUTO|AUTO-SKIP|BACKGROUND-COLOR|BACKGROUND-COLOUR|BASIS|BEEP|BEFORE|BEGINNING|BELL|BINARY|BIT|BLANK|BLINK|BLOCK|BOTTOM|BOUNDS|BY|BYFUNCTION|BYTITLE|CALL|CANCEL|CAPABLE|CCSVERSION|CD|CF|CH|CHAINING|CHANGED|CHANNEL|CHARACTER|CHARACTERS|CLASS|CLASS-ID|CLOCK-UNITS|CLOSE|CLOSE-DISPOSITION|COBOL|CODE|CODE-SET|COL|COLLATING|COLUMN|COM-REG|COMMA|COMMITMENT|COMMON|COMMUNICATION|COMP|COMP-1|COMP-2|COMP-3|COMP-4|COMP-5|COMPUTATIONAL|COMPUTATIONAL-1|COMPUTATIONAL-2|COMPUTATIONAL-3|COMPUTATIONAL-4|COMPUTATIONAL-5|COMPUTE|CONFIGURATION|CONTAINS|CONTENT|CONTINUE|CONTROL|CONTROL-POINT|CONTROLS|CONVENTION|CONVERTING|COPY|CORR|CORRESPONDING|COUNT|CRUNCH|CURRENCY|CURSOR|DATA|DATA-BASE|DATE|DATE-COMPILED|DATE-WRITTEN|DAY|DAY-OF-WEEK|DBCS|DE|DEBUG-CONTENTS|DEBUG-ITEM|DEBUG-LINE|DEBUG-NAME|DEBUG-SUB-1|DEBUG-SUB-2|DEBUG-SUB-3|DEBUGGING|DECIMAL-POINT|DECLARATIVES|DEFAULT|DEFAULT-DISPLAY|DEFINITION|DELETE|DELIMITED|DELIMITER|DEPENDING|DESCENDING|DESTINATION|DETAIL|DFHRESP|DFHVALUE|DISABLE|DISK|DISPLAY|DISPLAY-1|DIVIDE|DIVISION|DONTCARE|DOUBLE|DOWN|DUPLICATES|DYNAMIC|EBCDIC|EGCS|EGI|ELSE|EMI|EMPTY-CHECK|ENABLE|END|END-ACCEPT|END-ADD|END-CALL|END-COMPUTE|END-DELETE|END-DIVIDE|END-EVALUATE|END-IF|END-MULTIPLY|END-OF-PAGE|END-PERFORM|END-READ|END-RECEIVE|END-RETURN|END-REWRITE|END-SEARCH|END-START|END-STRING|END-SUBTRACT|END-UNSTRING|END-WRITE|ENDING|ENTER|ENTRY|ENTRY-PROCEDURE|ENVIRONMENT|EOL|EOP|EOS|ERASE|ERROR|ESCAPE|ESI|EVALUATE|EVENT|EVERY|EXCEPTION|EXCLUSIVE|EXHIBIT|EXIT|EXPORT|EXTEND|EXTENDED|EXTERNAL|FD|FILE|FILE-CONTROL|FILLER|FINAL|FIRST|FOOTING|FOR|FOREGROUND-COLOR|FOREGROUND-COLOUR|FROM|FULL|FUNCTION|FUNCTION-POINTER|FUNCTIONNAME|GENERATE|GIVING|GLOBAL|GO|GOBACK|GRID|GROUP|HEADING|HIGH-VALUE|HIGH-VALUES|HIGHLIGHT|I-O|I-O-CONTROL|ID|IDENTIFICATION|IF|IMPLICIT|IMPORT|IN|INDEX|INDEXED|INDICATE|INITIAL|INITIALIZE|INITIATE|INPUT|INPUT-OUTPUT|INSPECT|INSTALLATION|INTEGER|INTO|INVALID|INVOKE|IS|JUST|JUSTIFIED|KANJI|KEPT|KEY|KEYBOARD|LABEL|LANGUAGE|LAST|LB|LD|LEADING|LEFT|LEFTLINE|LENGTH|LENGTH-CHECK|LIBACCESS|LIBPARAMETER|LIBRARY|LIMIT|LIMITS|LINAGE|LINAGE-COUNTER|LINE|LINE-COUNTER|LINES|LINKAGE|LIST|LOCAL|LOCAL-STORAGE|LOCK|LONG-DATE|LONG-TIME|LOW-VALUE|LOW-VALUES|LOWER|LOWLIGHT|MEMORY|MERGE|MESSAGE|MMDDYYYY|MODE|MODULES|MORE-LABELS|MOVE|MULTIPLE|MULTIPLY|NAMED|NATIONAL|NATIONAL-EDITED|NATIVE|NEGATIVE|NETWORK|NEXT|NO|NO-ECHO|NULL|NULLS|NUMBER|NUMERIC|NUMERIC-DATE|NUMERIC-EDITED|NUMERIC-TIME|OBJECT-COMPUTER|OCCURS|ODT|OF|OFF|OMITTED|ON|OPEN|OPTIONAL|ORDER|ORDERLY|ORGANIZATION|OTHER|OUTPUT|OVERFLOW|OVERLINE|OWN|PACKED-DECIMAL|PADDING|PAGE|PAGE-COUNTER|PASSWORD|PERFORM|PF|PH|PIC|PICTURE|PLUS|POINTER|PORT|POSITION|POSITIVE|PRINTER|PRINTING|PRIVATE|PROCEDURE|PROCEDURE-POINTER|PROCEDURES|PROCEED|PROCESS|PROGRAM|PROGRAM-ID|PROGRAM-LIBRARY|PROMPT|PURGE|QUEUE|QUOTE|QUOTES|RANDOM|RD|READ|READER|REAL|RECEIVE|RECEIVED|RECORD|RECORDING|RECORDS|RECURSIVE|REDEFINES|REEL|REF|REFERENCE|REFERENCES|RELATIVE|RELEASE|REMAINDER|REMARKS|REMOTE|REMOVAL|REMOVE|RENAMES|REPLACE|REPLACING|REPORT|REPORTING|REPORTS|REQUIRED|RERUN|RESERVE|RESET|RETURN|RETURN-CODE|RETURNING|REVERSE-VIDEO|REVERSED|REWIND|REWRITE|RF|RH|RIGHT|ROUNDED|RUN|SAME|SAVE|SCREEN|SD|SEARCH|SECTION|SECURE|SECURITY|SEGMENT|SEGMENT-LIMIT|SELECT|SEND|SENTENCE|SEPARATE|SEQUENCE|SEQUENTIAL|SET|SHARED|SHAREDBYALL|SHAREDBYRUNUNIT|SHARING|SHIFT-IN|SHIFT-OUT|SHORT-DATE|SIGN|SIZE|SORT|SORT-CONTROL|SORT-CORE-SIZE|SORT-FILE-SIZE|SORT-MERGE|SORT-MESSAGE|SORT-MODE-SIZE|SORT-RETURN|SOURCE|SOURCE-COMPUTER|SPACE|SPACES|SPECIAL-NAMES|STANDARD|STANDARD-1|STANDARD-2|START|STATUS|STOP|STRING|SUB-QUEUE-1|SUB-QUEUE-2|SUB-QUEUE-3|SUBTRACT|SUM|SUPPRESS|SYMBOL|SYMBOLIC|SYNC|SYNCHRONIZED|TABLE|TALLY|TALLYING|TAPE|TASK|TERMINAL|TERMINATE|TEST|TEXT|THEN|THREAD|THREAD-LOCAL|THROUGH|THRU|TIME|TIMER|TIMES|TITLE|TO|TODAYS-DATE|TODAYS-NAME|TOP|TRAILING|TRUNCATED|TYPE|TYPEDEF|UNDERLINE|UNIT|UNSTRING|UNTIL|UP|UPON|USAGE|USE|USING|VALUE|VALUES|VARYING|VIRTUAL|WAIT|WHEN|WHEN-COMPILED|WITH|WORDS|WORKING-STORAGE|WRITE|YEAR|YYYYDDD|YYYYMMDD|ZERO-FILL|ZEROES|ZEROS)(?![\w-])/i,lookbehind:!0},boolean:{pattern:/(^|[^\w-])(?:false|true)(?![\w-])/i,lookbehind:!0},number:{pattern:/(^|[^\w-])(?:[+-]?(?:(?:\d+(?:[.,]\d+)?|[.,]\d+)(?:e[+-]?\d+)?|zero))(?![\w-])/i,lookbehind:!0},operator:[/<>|[<>]=?|[=+*/&]/,{pattern:/(^|[^\w-])(?:-|and|equal|greater|less|not|or|than)(?![\w-])/i,lookbehind:!0}],punctuation:/[.:,()]/}}return zBe}var HBe,Izt;function D$r(){if(Izt)return HBe;Izt=1,HBe=n,n.displayName="coffeescript",n.aliases=["coffee"];function n(r){(function(o){var s=/#(?!\{).+/,c={pattern:/#\{[^}]+\}/,alias:"variable"};o.languages.coffeescript=o.languages.extend("javascript",{comment:s,string:[{pattern:/'(?:\\[\s\S]|[^\\'])*'/,greedy:!0},{pattern:/"(?:\\[\s\S]|[^\\"])*"/,greedy:!0,inside:{interpolation:c}}],keyword:/\b(?:and|break|by|catch|class|continue|debugger|delete|do|each|else|extend|extends|false|finally|for|if|in|instanceof|is|isnt|let|loop|namespace|new|no|not|null|of|off|on|or|own|return|super|switch|then|this|throw|true|try|typeof|undefined|unless|until|when|while|window|with|yes|yield)\b/,"class-member":{pattern:/@(?!\d)\w+/,alias:"variable"}}),o.languages.insertBefore("coffeescript","comment",{"multiline-comment":{pattern:/###[\s\S]+?###/,alias:"comment"},"block-regex":{pattern:/\/{3}[\s\S]*?\/{3}/,alias:"regex",inside:{comment:s,interpolation:c}}}),o.languages.insertBefore("coffeescript","string",{"inline-javascript":{pattern:/`(?:\\[\s\S]|[^\\`])*`/,inside:{delimiter:{pattern:/^`|`$/,alias:"punctuation"},script:{pattern:/[\s\S]+/,alias:"language-javascript",inside:o.languages.javascript}}},"multiline-string":[{pattern:/'''[\s\S]*?'''/,greedy:!0,alias:"string"},{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string",inside:{interpolation:c}}]}),o.languages.insertBefore("coffeescript","keyword",{property:/(?!\d)\w+(?=\s*:(?!:))/}),delete o.languages.coffeescript["template-string"],o.languages.coffee=o.languages.coffeescript})(r)}return HBe}var GBe,Ozt;function M$r(){if(Ozt)return GBe;Ozt=1,GBe=n,n.displayName="concurnas",n.aliases=["conc"];function n(r){r.languages.concurnas={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?(?:\*\/|$)|\/\/.*)/,lookbehind:!0,greedy:!0},langext:{pattern:/\b\w+\s*\|\|[\s\S]+?\|\|/,greedy:!0,inside:{"class-name":/^\w+/,string:{pattern:/(^\s*\|\|)[\s\S]+(?=\|\|$)/,lookbehind:!0},punctuation:/\|\|/}},function:{pattern:/((?:^|\s)def[ \t]+)[a-zA-Z_]\w*(?=\s*\()/,lookbehind:!0},keyword:/\b(?:abstract|actor|also|annotation|assert|async|await|bool|boolean|break|byte|case|catch|changed|char|class|closed|constant|continue|def|default|del|double|elif|else|enum|every|extends|false|finally|float|for|from|global|gpudef|gpukernel|if|import|in|init|inject|int|lambda|local|long|loop|match|new|nodefault|null|of|onchange|open|out|override|package|parfor|parforsync|post|pre|private|protected|provide|provider|public|return|shared|short|single|size_t|sizeof|super|sync|this|throw|trait|trans|transient|true|try|typedef|unchecked|using|val|var|void|while|with)\b/,boolean:/\b(?:false|true)\b/,number:/\b0b[01][01_]*L?\b|\b0x(?:[\da-f_]*\.)?[\da-f_p+-]+\b|(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.\d[\d_]*)(?:e[+-]?\d[\d_]*)?[dfls]?/i,punctuation:/[{}[\];(),.:]/,operator:/<==|>==|=>|->|<-|<>|&==|&<>|\?:?|\.\?|\+\+|--|[-+*/=<>]=?|[!^~]|\b(?:and|as|band|bor|bxor|comp|is|isnot|mod|or)\b=?/,annotation:{pattern:/@(?:\w+:)?(?:\w+|\[[^\]]+\])?/,alias:"builtin"}},r.languages.insertBefore("concurnas","langext",{"regex-literal":{pattern:/\br("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}/,lookbehind:!0,inside:r.languages.concurnas},regex:/[\s\S]+/}},"string-literal":{pattern:/(?:\B|\bs)("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}/,lookbehind:!0,inside:r.languages.concurnas},string:/[\s\S]+/}}}),r.languages.conc=r.languages.concurnas}return GBe}var qBe,Nzt;function j$r(){if(Nzt)return qBe;Nzt=1,qBe=n,n.displayName="coq",n.aliases=[];function n(r){(function(o){for(var s=/\(\*(?:[^(*]|\((?!\*)|\*(?!\))|<self>)*\*\)/.source,c=0;c<2;c++)s=s.replace(/<self>/g,function(){return s});s=s.replace(/<self>/g,"[]"),o.languages.coq={comment:RegExp(s),string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},attribute:[{pattern:RegExp(/#\[(?:[^\[\]("]|"(?:[^"]|"")*"(?!")|\((?!\*)|<comment>)*\]/.source.replace(/<comment>/g,function(){return s})),greedy:!0,alias:"attr-name",inside:{comment:RegExp(s),string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},operator:/=/,punctuation:/^#\[|\]$|[,()]/}},{pattern:/\b(?:Cumulative|Global|Local|Monomorphic|NonCumulative|Polymorphic|Private|Program)\b/,alias:"attr-name"}],keyword:/\b(?:Abort|About|Add|Admit|Admitted|All|Arguments|As|Assumptions|Axiom|Axioms|Back|BackTo|Backtrace|BinOp|BinOpSpec|BinRel|Bind|Blacklist|Canonical|Case|Cd|Check|Class|Classes|Close|CoFixpoint|CoInductive|Coercion|Coercions|Collection|Combined|Compute|Conjecture|Conjectures|Constant|Constants|Constraint|Constructors|Context|Corollary|Create|CstOp|Custom|Cut|Debug|Declare|Defined|Definition|Delimit|Dependencies|Dependent|Derive|Diffs|Drop|Elimination|End|Entry|Equality|Eval|Example|Existential|Existentials|Existing|Export|Extern|Extraction|Fact|Fail|Field|File|Firstorder|Fixpoint|Flags|Focus|From|Funclass|Function|Functional|GC|Generalizable|Goal|Grab|Grammar|Graph|Guarded|Haskell|Heap|Hide|Hint|HintDb|Hints|Hypotheses|Hypothesis|IF|Identity|Immediate|Implicit|Implicits|Import|Include|Induction|Inductive|Infix|Info|Initial|InjTyp|Inline|Inspect|Instance|Instances|Intro|Intros|Inversion|Inversion_clear|JSON|Language|Left|Lemma|Let|Lia|Libraries|Library|Load|LoadPath|Locate|Ltac|Ltac2|ML|Match|Method|Minimality|Module|Modules|Morphism|Next|NoInline|Notation|Number|OCaml|Obligation|Obligations|Opaque|Open|Optimize|Parameter|Parameters|Parametric|Path|Paths|Prenex|Preterm|Primitive|Print|Profile|Projections|Proof|Prop|PropBinOp|PropOp|PropUOp|Property|Proposition|Pwd|Qed|Quit|Rec|Record|Recursive|Redirect|Reduction|Register|Relation|Remark|Remove|Require|Reserved|Reset|Resolve|Restart|Rewrite|Right|Ring|Rings|SProp|Saturate|Save|Scheme|Scope|Scopes|Search|SearchHead|SearchPattern|SearchRewrite|Section|Separate|Set|Setoid|Show|Signatures|Solve|Solver|Sort|Sortclass|Sorted|Spec|Step|Strategies|Strategy|String|Structure|SubClass|Subgraph|SuchThat|Tactic|Term|TestCompile|Theorem|Time|Timeout|To|Transparent|Type|Typeclasses|Types|Typing|UnOp|UnOpSpec|Undelimit|Undo|Unfocus|Unfocused|Unfold|Universe|Universes|Unshelve|Variable|Variables|Variant|Verbose|View|Visibility|Zify|_|apply|as|at|by|cofix|else|end|exists|exists2|fix|for|forall|fun|if|in|let|match|measure|move|removed|return|struct|then|using|wf|where|with)\b/,number:/\b(?:0x[a-f0-9][a-f0-9_]*(?:\.[a-f0-9_]+)?(?:p[+-]?\d[\d_]*)?|\d[\d_]*(?:\.[\d_]+)?(?:e[+-]?\d[\d_]*)?)\b/i,punct:{pattern:/@\{|\{\||\[=|:>/,alias:"punctuation"},operator:/\/\\|\\\/|\.{2,3}|:{1,2}=|\*\*|[-=]>|<(?:->?|[+:=>]|<:)|>(?:=|->)|\|[-|]?|[-!%&*+/<=>?@^~']/,punctuation:/\.\(|`\(|@\{|`\{|\{\||\[=|:>|[:.,;(){}\[\]]/}})(r)}return qBe}var VBe,Rzt;function PEe(){if(Rzt)return VBe;Rzt=1,VBe=n,n.displayName="ruby",n.aliases=["rb"];function n(r){(function(o){o.languages.ruby=o.languages.extend("clike",{comment:{pattern:/#.*|^=begin\s[\s\S]*?^=end/m,greedy:!0},"class-name":{pattern:/(\b(?:class|module)\s+|\bcatch\s+\()[\w.\\]+|\b[A-Z_]\w*(?=\s*\.\s*new\b)/,lookbehind:!0,inside:{punctuation:/[.\\]/}},keyword:/\b(?:BEGIN|END|alias|and|begin|break|case|class|def|define_method|defined|do|each|else|elsif|end|ensure|extend|for|if|in|include|module|new|next|nil|not|or|prepend|private|protected|public|raise|redo|require|rescue|retry|return|self|super|then|throw|undef|unless|until|when|while|yield)\b/,operator:/\.{2,3}|&\.|===|<?=>|[!=]?~|(?:&&|\|\||<<|>>|\*\*|[+\-*/%<>!^&|=])=?|[?:]/,punctuation:/[(){}[\].,;]/}),o.languages.insertBefore("ruby","operator",{"double-colon":{pattern:/::/,alias:"punctuation"}});var s={pattern:/((?:^|[^\\])(?:\\{2})*)#\{(?:[^{}]|\{[^{}]*\})*\}/,lookbehind:!0,inside:{content:{pattern:/^(#\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:o.languages.ruby},delimiter:{pattern:/^#\{|\}$/,alias:"punctuation"}}};delete o.languages.ruby.function;var c="(?:"+[/([^a-zA-Z0-9\s{(\[<=])(?:(?!\1)[^\\]|\\[\s\S])*\1/.source,/\((?:[^()\\]|\\[\s\S]|\((?:[^()\\]|\\[\s\S])*\))*\)/.source,/\{(?:[^{}\\]|\\[\s\S]|\{(?:[^{}\\]|\\[\s\S])*\})*\}/.source,/\[(?:[^\[\]\\]|\\[\s\S]|\[(?:[^\[\]\\]|\\[\s\S])*\])*\]/.source,/<(?:[^<>\\]|\\[\s\S]|<(?:[^<>\\]|\\[\s\S])*>)*>/.source].join("|")+")",d=/(?:"(?:\\.|[^"\\\r\n])*"|(?:\b[a-zA-Z_]\w*|[^\s\0-\x7F]+)[?!]?|\$.)/.source;o.languages.insertBefore("ruby","keyword",{"regex-literal":[{pattern:RegExp(/%r/.source+c+/[egimnosux]{0,6}/.source),greedy:!0,inside:{interpolation:s,regex:/[\s\S]+/}},{pattern:/(^|[^/])\/(?!\/)(?:\[[^\r\n\]]+\]|\\.|[^[/\\\r\n])+\/[egimnosux]{0,6}(?=\s*(?:$|[\r\n,.;})#]))/,lookbehind:!0,greedy:!0,inside:{interpolation:s,regex:/[\s\S]+/}}],variable:/[@$]+[a-zA-Z_]\w*(?:[?!]|\b)/,symbol:[{pattern:RegExp(/(^|[^:]):/.source+d),lookbehind:!0,greedy:!0},{pattern:RegExp(/([\r\n{(,][ \t]*)/.source+d+/(?=:(?!:))/.source),lookbehind:!0,greedy:!0}],"method-definition":{pattern:/(\bdef\s+)\w+(?:\s*\.\s*\w+)?/,lookbehind:!0,inside:{function:/\b\w+$/,keyword:/^self\b/,"class-name":/^\w+/,punctuation:/\./}}}),o.languages.insertBefore("ruby","string",{"string-literal":[{pattern:RegExp(/%[qQiIwWs]?/.source+c),greedy:!0,inside:{interpolation:s,string:/[\s\S]+/}},{pattern:/("|')(?:#\{[^}]+\}|#(?!\{)|\\(?:\r\n|[\s\S])|(?!\1)[^\\#\r\n])*\1/,greedy:!0,inside:{interpolation:s,string:/[\s\S]+/}},{pattern:/<<[-~]?([a-z_]\w*)[\r\n](?:.*[\r\n])*?[\t ]*\1/i,alias:"heredoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<[-~]?[a-z_]\w*|\b[a-z_]\w*$/i,inside:{symbol:/\b\w+/,punctuation:/^<<[-~]?/}},interpolation:s,string:/[\s\S]+/}},{pattern:/<<[-~]?'([a-z_]\w*)'[\r\n](?:.*[\r\n])*?[\t ]*\1/i,alias:"heredoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<[-~]?'[a-z_]\w*'|\b[a-z_]\w*$/i,inside:{symbol:/\b\w+/,punctuation:/^<<[-~]?'|'$/}},string:/[\s\S]+/}}],"command-literal":[{pattern:RegExp(/%x/.source+c),greedy:!0,inside:{interpolation:s,command:{pattern:/[\s\S]+/,alias:"string"}}},{pattern:/`(?:#\{[^}]+\}|#(?!\{)|\\(?:\r\n|[\s\S])|[^\\`#\r\n])*`/,greedy:!0,inside:{interpolation:s,command:{pattern:/[\s\S]+/,alias:"string"}}}]}),delete o.languages.ruby.string,o.languages.insertBefore("ruby","number",{builtin:/\b(?:Array|Bignum|Binding|Class|Continuation|Dir|Exception|FalseClass|File|Fixnum|Float|Hash|IO|Integer|MatchData|Method|Module|NilClass|Numeric|Object|Proc|Range|Regexp|Stat|String|Struct|Symbol|TMS|Thread|ThreadGroup|Time|TrueClass)\b/,constant:/\b[A-Z][A-Z0-9_]*(?:[?!]|\b)/}),o.languages.rb=o.languages.ruby})(r)}return VBe}var WBe,Pzt;function $$r(){if(Pzt)return WBe;Pzt=1;var n=PEe();WBe=r,r.displayName="crystal",r.aliases=[];function r(o){o.register(n),function(s){s.languages.crystal=s.languages.extend("ruby",{keyword:[/\b(?:__DIR__|__END_LINE__|__FILE__|__LINE__|abstract|alias|annotation|as|asm|begin|break|case|class|def|do|else|elsif|end|ensure|enum|extend|for|fun|if|ifdef|include|instance_sizeof|lib|macro|module|next|of|out|pointerof|private|protected|ptr|require|rescue|return|select|self|sizeof|struct|super|then|type|typeof|undef|uninitialized|union|unless|until|when|while|with|yield)\b/,{pattern:/(\.\s*)(?:is_a|responds_to)\?/,lookbehind:!0}],number:/\b(?:0b[01_]*[01]|0o[0-7_]*[0-7]|0x[\da-fA-F_]*[\da-fA-F]|(?:\d(?:[\d_]*\d)?)(?:\.[\d_]*\d)?(?:[eE][+-]?[\d_]*\d)?)(?:_(?:[uif](?:8|16|32|64))?)?\b/,operator:[/->/,s.languages.ruby.operator],punctuation:/[(){}[\].,;\\]/}),s.languages.insertBefore("crystal","string-literal",{attribute:{pattern:/@\[.*?\]/,inside:{delimiter:{pattern:/^@\[|\]$/,alias:"punctuation"},attribute:{pattern:/^(\s*)\w+/,lookbehind:!0,alias:"class-name"},args:{pattern:/\S(?:[\s\S]*\S)?/,inside:s.languages.crystal}}},expansion:{pattern:/\{(?:\{.*?\}|%.*?%)\}/,inside:{content:{pattern:/^(\{.)[\s\S]+(?=.\}$)/,lookbehind:!0,inside:s.languages.crystal},delimiter:{pattern:/^\{[\{%]|[\}%]\}$/,alias:"operator"}}},char:{pattern:/'(?:[^\\\r\n]{1,2}|\\(?:.|u(?:[A-Fa-f0-9]{1,4}|\{[A-Fa-f0-9]{1,6}\})))'/,greedy:!0}})}(o)}return WBe}var KBe,Lzt;function F$r(){if(Lzt)return KBe;Lzt=1;var n=REe();KBe=r,r.displayName="cshtml",r.aliases=["razor"];function r(o){o.register(n),function(s){var c=/\/(?![/*])|\/\/.*[\r\n]|\/\*[^*]*(?:\*(?!\/)[^*]*)*\*\//.source,d=/@(?!")|"(?:[^\r\n\\"]|\\.)*"|@"(?:[^\\"]|""|\\[\s\S])*"(?!")/.source+"|"+/'(?:(?:[^\r\n'\\]|\\.|\\[Uux][\da-fA-F]{1,8})'|(?=[^\\](?!')))/.source;function h(U,B){for(var G=0;G<B;G++)U=U.replace(/<self>/g,function(){return"(?:"+U+")"});return U.replace(/<self>/g,"[^\\s\\S]").replace(/<str>/g,"(?:"+d+")").replace(/<comment>/g,"(?:"+c+")")}var b=h(/\((?:[^()'"@/]|<str>|<comment>|<self>)*\)/.source,2),y=h(/\[(?:[^\[\]'"@/]|<str>|<comment>|<self>)*\]/.source,2),E=h(/\{(?:[^{}'"@/]|<str>|<comment>|<self>)*\}/.source,2),k=h(/<(?:[^<>'"@/]|<str>|<comment>|<self>)*>/.source,2),x=/(?:\s(?:\s*[^\s>\/=]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+(?=[\s>]))|(?=[\s/>])))+)?/.source,O=/(?!\d)[^\s>\/=$<%]+/.source+x+/\s*\/?>/.source,P=/\B@?/.source+"(?:"+/<([a-zA-Z][\w:]*)/.source+x+/\s*>/.source+"(?:"+(/[^<]/.source+"|"+/<\/?(?!\1\b)/.source+O+"|"+h(/<\1/.source+x+/\s*>/.source+"(?:"+(/[^<]/.source+"|"+/<\/?(?!\1\b)/.source+O+"|<self>")+")*"+/<\/\1\s*>/.source,2))+")*"+/<\/\1\s*>/.source+"|"+/</.source+O+")";s.languages.cshtml=s.languages.extend("markup",{});var D=s.languages.insertBefore("csharp","string",{html:{pattern:RegExp(P),greedy:!0,inside:s.languages.cshtml}},{csharp:s.languages.extend("csharp",{})}),$={pattern:/\S[\s\S]*/,alias:"language-csharp",inside:D};s.languages.insertBefore("cshtml","prolog",{"razor-comment":{pattern:/@\*[\s\S]*?\*@/,greedy:!0,alias:"comment"},block:{pattern:RegExp(/(^|[^@])@/.source+"(?:"+[E,/(?:code|functions)\s*/.source+E,/(?:for|foreach|lock|switch|using|while)\s*/.source+b+/\s*/.source+E,/do\s*/.source+E+/\s*while\s*/.source+b+/(?:\s*;)?/.source,/try\s*/.source+E+/\s*catch\s*/.source+b+/\s*/.source+E+/\s*finally\s*/.source+E,/if\s*/.source+b+/\s*/.source+E+"(?:"+/\s*else/.source+"(?:"+/\s+if\s*/.source+b+")?"+/\s*/.source+E+")*"].join("|")+")"),lookbehind:!0,greedy:!0,inside:{keyword:/^@\w*/,csharp:$}},directive:{pattern:/^([ \t]*)@(?:addTagHelper|attribute|implements|inherits|inject|layout|model|namespace|page|preservewhitespace|removeTagHelper|section|tagHelperPrefix|using)(?=\s).*/m,lookbehind:!0,greedy:!0,inside:{keyword:/^@\w+/,csharp:$}},value:{pattern:RegExp(/(^|[^@])@/.source+/(?:await\b\s*)?/.source+"(?:"+/\w+\b/.source+"|"+b+")(?:"+/[?!]?\.\w+\b/.source+"|"+b+"|"+y+"|"+k+b+")*"),lookbehind:!0,greedy:!0,alias:"variable",inside:{keyword:/^@/,csharp:$}},"delegate-operator":{pattern:/(^|[^@])@(?=<)/,lookbehind:!0,alias:"operator"}}),s.languages.razor=s.languages.cshtml}(o)}return KBe}var YBe,Dzt;function B$r(){if(Dzt)return YBe;Dzt=1,YBe=n,n.displayName="csp",n.aliases=[];function n(r){(function(o){function s(c){return RegExp(/([ \t])/.source+"(?:"+c+")"+/(?=[\s;]|$)/.source,"i")}o.languages.csp={directive:{pattern:/(^|[\s;])(?:base-uri|block-all-mixed-content|(?:child|connect|default|font|frame|img|manifest|media|object|prefetch|script|style|worker)-src|disown-opener|form-action|frame-(?:ancestors|options)|input-protection(?:-(?:clip|selectors))?|navigate-to|plugin-types|policy-uri|referrer|reflected-xss|report-(?:to|uri)|require-sri-for|sandbox|(?:script|style)-src-(?:attr|elem)|upgrade-insecure-requests)(?=[\s;]|$)/i,lookbehind:!0,alias:"property"},scheme:{pattern:s(/[a-z][a-z0-9.+-]*:/.source),lookbehind:!0},none:{pattern:s(/'none'/.source),lookbehind:!0,alias:"keyword"},nonce:{pattern:s(/'nonce-[-+/\w=]+'/.source),lookbehind:!0,alias:"number"},hash:{pattern:s(/'sha(?:256|384|512)-[-+/\w=]+'/.source),lookbehind:!0,alias:"number"},host:{pattern:s(/[a-z][a-z0-9.+-]*:\/\/[^\s;,']*/.source+"|"+/\*[^\s;,']*/.source+"|"+/[a-z0-9-]+(?:\.[a-z0-9-]+)+(?::[\d*]+)?(?:\/[^\s;,']*)?/.source),lookbehind:!0,alias:"url",inside:{important:/\*/}},keyword:[{pattern:s(/'unsafe-[a-z-]+'/.source),lookbehind:!0,alias:"unsafe"},{pattern:s(/'[a-z-]+'/.source),lookbehind:!0,alias:"safe"}],punctuation:/;/}})(r)}return YBe}var XBe,Mzt;function U$r(){if(Mzt)return XBe;Mzt=1,XBe=n,n.displayName="cssExtras",n.aliases=[];function n(r){(function(o){var s=/("|')(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,c;o.languages.css.selector={pattern:o.languages.css.selector.pattern,lookbehind:!0,inside:c={"pseudo-element":/:(?:after|before|first-letter|first-line|selection)|::[-\w]+/,"pseudo-class":/:[-\w]+/,class:/\.[-\w]+/,id:/#[-\w]+/,attribute:{pattern:RegExp(`\\[(?:[^[\\]"']|`+s.source+")*\\]"),greedy:!0,inside:{punctuation:/^\[|\]$/,"case-sensitivity":{pattern:/(\s)[si]$/i,lookbehind:!0,alias:"keyword"},namespace:{pattern:/^(\s*)(?:(?!\s)[-*\w\xA0-\uFFFF])*\|(?!=)/,lookbehind:!0,inside:{punctuation:/\|$/}},"attr-name":{pattern:/^(\s*)(?:(?!\s)[-\w\xA0-\uFFFF])+/,lookbehind:!0},"attr-value":[s,{pattern:/(=\s*)(?:(?!\s)[-\w\xA0-\uFFFF])+(?=\s*$)/,lookbehind:!0}],operator:/[|~*^$]?=/}},"n-th":[{pattern:/(\(\s*)[+-]?\d*[\dn](?:\s*[+-]\s*\d+)?(?=\s*\))/,lookbehind:!0,inside:{number:/[\dn]+/,operator:/[+-]/}},{pattern:/(\(\s*)(?:even|odd)(?=\s*\))/i,lookbehind:!0}],combinator:/>|\+|~|\|\|/,punctuation:/[(),]/}},o.languages.css.atrule.inside["selector-function-argument"].inside=c,o.languages.insertBefore("css","property",{variable:{pattern:/(^|[^-\w\xA0-\uFFFF])--(?!\s)[-_a-z\xA0-\uFFFF](?:(?!\s)[-\w\xA0-\uFFFF])*/i,lookbehind:!0}});var d={pattern:/(\b\d+)(?:%|[a-z]+(?![\w-]))/,lookbehind:!0},h={pattern:/(^|[^\w.-])-?(?:\d+(?:\.\d+)?|\.\d+)/,lookbehind:!0};o.languages.insertBefore("css","function",{operator:{pattern:/(\s)[+\-*\/](?=\s)/,lookbehind:!0},hexcode:{pattern:/\B#[\da-f]{3,8}\b/i,alias:"color"},color:[{pattern:/(^|[^\w-])(?:AliceBlue|AntiqueWhite|Aqua|Aquamarine|Azure|Beige|Bisque|Black|BlanchedAlmond|Blue|BlueViolet|Brown|BurlyWood|CadetBlue|Chartreuse|Chocolate|Coral|CornflowerBlue|Cornsilk|Crimson|Cyan|DarkBlue|DarkCyan|DarkGoldenRod|DarkGr[ae]y|DarkGreen|DarkKhaki|DarkMagenta|DarkOliveGreen|DarkOrange|DarkOrchid|DarkRed|DarkSalmon|DarkSeaGreen|DarkSlateBlue|DarkSlateGr[ae]y|DarkTurquoise|DarkViolet|DeepPink|DeepSkyBlue|DimGr[ae]y|DodgerBlue|FireBrick|FloralWhite|ForestGreen|Fuchsia|Gainsboro|GhostWhite|Gold|GoldenRod|Gr[ae]y|Green|GreenYellow|HoneyDew|HotPink|IndianRed|Indigo|Ivory|Khaki|Lavender|LavenderBlush|LawnGreen|LemonChiffon|LightBlue|LightCoral|LightCyan|LightGoldenRodYellow|LightGr[ae]y|LightGreen|LightPink|LightSalmon|LightSeaGreen|LightSkyBlue|LightSlateGr[ae]y|LightSteelBlue|LightYellow|Lime|LimeGreen|Linen|Magenta|Maroon|MediumAquaMarine|MediumBlue|MediumOrchid|MediumPurple|MediumSeaGreen|MediumSlateBlue|MediumSpringGreen|MediumTurquoise|MediumVioletRed|MidnightBlue|MintCream|MistyRose|Moccasin|NavajoWhite|Navy|OldLace|Olive|OliveDrab|Orange|OrangeRed|Orchid|PaleGoldenRod|PaleGreen|PaleTurquoise|PaleVioletRed|PapayaWhip|PeachPuff|Peru|Pink|Plum|PowderBlue|Purple|Red|RosyBrown|RoyalBlue|SaddleBrown|Salmon|SandyBrown|SeaGreen|SeaShell|Sienna|Silver|SkyBlue|SlateBlue|SlateGr[ae]y|Snow|SpringGreen|SteelBlue|Tan|Teal|Thistle|Tomato|Transparent|Turquoise|Violet|Wheat|White|WhiteSmoke|Yellow|YellowGreen)(?![\w-])/i,lookbehind:!0},{pattern:/\b(?:hsl|rgb)\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*\)\B|\b(?:hsl|rgb)a\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*,\s*(?:0|0?\.\d+|1)\s*\)\B/i,inside:{unit:d,number:h,function:/[\w-]+(?=\()/,punctuation:/[(),]/}}],entity:/\\[\da-f]{1,8}/i,unit:d,number:h})})(r)}return XBe}var JBe,jzt;function z$r(){if(jzt)return JBe;jzt=1,JBe=n,n.displayName="csv",n.aliases=[];function n(r){r.languages.csv={value:/[^\r\n,"]+|"(?:[^"]|"")*"(?!")/,punctuation:/,/}}return JBe}var ZBe,$zt;function H$r(){if($zt)return ZBe;$zt=1,ZBe=n,n.displayName="cypher",n.aliases=[];function n(r){r.languages.cypher={comment:/\/\/.*/,string:{pattern:/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\\\r\n]|\\.)*'/,greedy:!0},"class-name":{pattern:/(:\s*)(?:\w+|`(?:[^`\\\r\n])*`)(?=\s*[{):])/,lookbehind:!0,greedy:!0},relationship:{pattern:/(-\[\s*(?:\w+\s*|`(?:[^`\\\r\n])*`\s*)?:\s*|\|\s*:\s*)(?:\w+|`(?:[^`\\\r\n])*`)/,lookbehind:!0,greedy:!0,alias:"property"},identifier:{pattern:/`(?:[^`\\\r\n])*`/,greedy:!0},variable:/\$\w+/,keyword:/\b(?:ADD|ALL|AND|AS|ASC|ASCENDING|ASSERT|BY|CALL|CASE|COMMIT|CONSTRAINT|CONTAINS|CREATE|CSV|DELETE|DESC|DESCENDING|DETACH|DISTINCT|DO|DROP|ELSE|END|ENDS|EXISTS|FOR|FOREACH|IN|INDEX|IS|JOIN|KEY|LIMIT|LOAD|MANDATORY|MATCH|MERGE|NODE|NOT|OF|ON|OPTIONAL|OR|ORDER(?=\s+BY)|PERIODIC|REMOVE|REQUIRE|RETURN|SCALAR|SCAN|SET|SKIP|START|STARTS|THEN|UNION|UNIQUE|UNWIND|USING|WHEN|WHERE|WITH|XOR|YIELD)\b/i,function:/\b\w+\b(?=\s*\()/,boolean:/\b(?:false|null|true)\b/i,number:/\b(?:0x[\da-fA-F]+|\d+(?:\.\d+)?(?:[eE][+-]?\d+)?)\b/,operator:/:|<--?|--?>?|<>|=~?|[<>]=?|[+*/%^|]|\.\.\.?/,punctuation:/[()[\]{},;.]/}}return ZBe}var QBe,Fzt;function G$r(){if(Fzt)return QBe;Fzt=1,QBe=n,n.displayName="d",n.aliases=[];function n(r){r.languages.d=r.languages.extend("clike",{comment:[{pattern:/^\s*#!.+/,greedy:!0},{pattern:RegExp(/(^|[^\\])/.source+"(?:"+[/\/\+(?:\/\+(?:[^+]|\+(?!\/))*\+\/|(?!\/\+)[\s\S])*?\+\//.source,/\/\/.*/.source,/\/\*[\s\S]*?\*\//.source].join("|")+")"),lookbehind:!0,greedy:!0}],string:[{pattern:RegExp([/\b[rx]"(?:\\[\s\S]|[^\\"])*"[cwd]?/.source,/\bq"(?:\[[\s\S]*?\]|\([\s\S]*?\)|<[\s\S]*?>|\{[\s\S]*?\})"/.source,/\bq"((?!\d)\w+)$[\s\S]*?^\1"/.source,/\bq"(.)[\s\S]*?\2"/.source,/(["`])(?:\\[\s\S]|(?!\3)[^\\])*\3[cwd]?/.source].join("|"),"m"),greedy:!0},{pattern:/\bq\{(?:\{[^{}]*\}|[^{}])*\}/,greedy:!0,alias:"token-string"}],keyword:/\$|\b(?:__(?:(?:DATE|EOF|FILE|FUNCTION|LINE|MODULE|PRETTY_FUNCTION|TIMESTAMP|TIME|VENDOR|VERSION)__|gshared|parameters|traits|vector)|abstract|alias|align|asm|assert|auto|body|bool|break|byte|case|cast|catch|cdouble|cent|cfloat|char|class|const|continue|creal|dchar|debug|default|delegate|delete|deprecated|do|double|dstring|else|enum|export|extern|false|final|finally|float|for|foreach|foreach_reverse|function|goto|idouble|if|ifloat|immutable|import|inout|int|interface|invariant|ireal|lazy|long|macro|mixin|module|new|nothrow|null|out|override|package|pragma|private|protected|ptrdiff_t|public|pure|real|ref|return|scope|shared|short|size_t|static|string|struct|super|switch|synchronized|template|this|throw|true|try|typedef|typeid|typeof|ubyte|ucent|uint|ulong|union|unittest|ushort|version|void|volatile|wchar|while|with|wstring)\b/,number:[/\b0x\.?[a-f\d_]+(?:(?!\.\.)\.[a-f\d_]*)?(?:p[+-]?[a-f\d_]+)?[ulfi]{0,4}/i,{pattern:/((?:\.\.)?)(?:\b0b\.?|\b|\.)\d[\d_]*(?:(?!\.\.)\.[\d_]*)?(?:e[+-]?\d[\d_]*)?[ulfi]{0,4}/i,lookbehind:!0}],operator:/\|[|=]?|&[&=]?|\+[+=]?|-[-=]?|\.?\.\.|=[>=]?|!(?:i[ns]\b|<>?=?|>=?|=)?|\bi[ns]\b|(?:<[<>]?|>>?>?|\^\^|[*\/%^~])=?/}),r.languages.insertBefore("d","string",{char:/'(?:\\(?:\W|\w+)|[^\\])'/}),r.languages.insertBefore("d","keyword",{property:/\B@\w*/}),r.languages.insertBefore("d","function",{register:{pattern:/\b(?:[ABCD][LHX]|E?(?:BP|DI|SI|SP)|[BS]PL|[ECSDGF]S|CR[0234]|[DS]IL|DR[012367]|E[ABCD]X|X?MM[0-7]|R(?:1[0-5]|[89])[BWD]?|R[ABCD]X|R[BS]P|R[DS]I|TR[3-7]|XMM(?:1[0-5]|[89])|YMM(?:1[0-5]|\d))\b|\bST(?:\([0-7]\)|\b)/,alias:"variable"}})}return QBe}var eUe,Bzt;function q$r(){if(Bzt)return eUe;Bzt=1,eUe=n,n.displayName="dart",n.aliases=[];function n(r){(function(o){var s=[/\b(?:async|sync|yield)\*/,/\b(?:abstract|assert|async|await|break|case|catch|class|const|continue|covariant|default|deferred|do|dynamic|else|enum|export|extends|extension|external|factory|final|finally|for|get|hide|if|implements|import|in|interface|library|mixin|new|null|on|operator|part|rethrow|return|set|show|static|super|switch|sync|this|throw|try|typedef|var|void|while|with|yield)\b/],c=/(^|[^\w.])(?:[a-z]\w*\s*\.\s*)*(?:[A-Z]\w*\s*\.\s*)*/.source,d={pattern:RegExp(c+/[A-Z](?:[\d_A-Z]*[a-z]\w*)?\b/.source),lookbehind:!0,inside:{namespace:{pattern:/^[a-z]\w*(?:\s*\.\s*[a-z]\w*)*(?:\s*\.)?/,inside:{punctuation:/\./}}}};o.languages.dart=o.languages.extend("clike",{"class-name":[d,{pattern:RegExp(c+/[A-Z]\w*(?=\s+\w+\s*[;,=()])/.source),lookbehind:!0,inside:d.inside}],keyword:s,operator:/\bis!|\b(?:as|is)\b|\+\+|--|&&|\|\||<<=?|>>=?|~(?:\/=?)?|[+\-*\/%&^|=!<>]=?|\?/}),o.languages.insertBefore("dart","string",{"string-literal":{pattern:/r?(?:("""|''')[\s\S]*?\1|(["'])(?:\\.|(?!\2)[^\\\r\n])*\2(?!\2))/,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:\w+|\{(?:[^{}]|\{[^{}]*\})*\})/,lookbehind:!0,inside:{punctuation:/^\$\{?|\}$/,expression:{pattern:/[\s\S]+/,inside:o.languages.dart}}},string:/[\s\S]+/}},string:void 0}),o.languages.insertBefore("dart","class-name",{metadata:{pattern:/@\w+/,alias:"function"}}),o.languages.insertBefore("dart","class-name",{generics:{pattern:/<(?:[\w\s,.&?]|<(?:[\w\s,.&?]|<(?:[\w\s,.&?]|<[\w\s,.&?]*>)*>)*>)*>/,inside:{"class-name":d,keyword:s,punctuation:/[<>(),.:]/,operator:/[?&|]/}}})})(r)}return eUe}var tUe,Uzt;function V$r(){if(Uzt)return tUe;Uzt=1,tUe=n,n.displayName="dataweave",n.aliases=[];function n(r){(function(o){o.languages.dataweave={url:/\b[A-Za-z]+:\/\/[\w/:.?=&-]+|\burn:[\w:.?=&-]+/,property:{pattern:/(?:\b\w+#)?(?:"(?:\\.|[^\\"\r\n])*"|\b\w+)(?=\s*[:@])/,greedy:!0},string:{pattern:/(["'`])(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0},"mime-type":/\b(?:application|audio|image|multipart|text|video)\/[\w+-]+/,date:{pattern:/\|[\w:+-]+\|/,greedy:!0},comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],regex:{pattern:/\/(?:[^\\\/\r\n]|\\[^\r\n])+\//,greedy:!0},keyword:/\b(?:and|as|at|case|do|else|fun|if|input|is|match|not|ns|null|or|output|type|unless|update|using|var)\b/,function:/\b[A-Z_]\w*(?=\s*\()/i,number:/-?\b\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,punctuation:/[{}[\];(),.:@]/,operator:/<<|>>|->|[<>~=]=?|!=|--?-?|\+\+?|!|\?/,boolean:/\b(?:false|true)\b/}})(r)}return tUe}var nUe,zzt;function W$r(){if(zzt)return nUe;zzt=1,nUe=n,n.displayName="dax",n.aliases=[];function n(r){r.languages.dax={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:--|\/\/).*)/,lookbehind:!0},"data-field":{pattern:/'(?:[^']|'')*'(?!')(?:\[[ \w\xA0-\uFFFF]+\])?|\w+\[[ \w\xA0-\uFFFF]+\]/,alias:"symbol"},measure:{pattern:/\[[ \w\xA0-\uFFFF]+\]/,alias:"constant"},string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},function:/\b(?:ABS|ACOS|ACOSH|ACOT|ACOTH|ADDCOLUMNS|ADDMISSINGITEMS|ALL|ALLCROSSFILTERED|ALLEXCEPT|ALLNOBLANKROW|ALLSELECTED|AND|APPROXIMATEDISTINCTCOUNT|ASIN|ASINH|ATAN|ATANH|AVERAGE|AVERAGEA|AVERAGEX|BETA\.DIST|BETA\.INV|BLANK|CALCULATE|CALCULATETABLE|CALENDAR|CALENDARAUTO|CEILING|CHISQ\.DIST|CHISQ\.DIST\.RT|CHISQ\.INV|CHISQ\.INV\.RT|CLOSINGBALANCEMONTH|CLOSINGBALANCEQUARTER|CLOSINGBALANCEYEAR|COALESCE|COMBIN|COMBINA|COMBINEVALUES|CONCATENATE|CONCATENATEX|CONFIDENCE\.NORM|CONFIDENCE\.T|CONTAINS|CONTAINSROW|CONTAINSSTRING|CONTAINSSTRINGEXACT|CONVERT|COS|COSH|COT|COTH|COUNT|COUNTA|COUNTAX|COUNTBLANK|COUNTROWS|COUNTX|CROSSFILTER|CROSSJOIN|CURRENCY|CURRENTGROUP|CUSTOMDATA|DATATABLE|DATE|DATEADD|DATEDIFF|DATESBETWEEN|DATESINPERIOD|DATESMTD|DATESQTD|DATESYTD|DATEVALUE|DAY|DEGREES|DETAILROWS|DISTINCT|DISTINCTCOUNT|DISTINCTCOUNTNOBLANK|DIVIDE|EARLIER|EARLIEST|EDATE|ENDOFMONTH|ENDOFQUARTER|ENDOFYEAR|EOMONTH|ERROR|EVEN|EXACT|EXCEPT|EXP|EXPON\.DIST|FACT|FALSE|FILTER|FILTERS|FIND|FIRSTDATE|FIRSTNONBLANK|FIRSTNONBLANKVALUE|FIXED|FLOOR|FORMAT|GCD|GENERATE|GENERATEALL|GENERATESERIES|GEOMEAN|GEOMEANX|GROUPBY|HASONEFILTER|HASONEVALUE|HOUR|IF|IF\.EAGER|IFERROR|IGNORE|INT|INTERSECT|ISBLANK|ISCROSSFILTERED|ISEMPTY|ISERROR|ISEVEN|ISFILTERED|ISINSCOPE|ISLOGICAL|ISNONTEXT|ISNUMBER|ISO\.CEILING|ISODD|ISONORAFTER|ISSELECTEDMEASURE|ISSUBTOTAL|ISTEXT|KEEPFILTERS|KEYWORDMATCH|LASTDATE|LASTNONBLANK|LASTNONBLANKVALUE|LCM|LEFT|LEN|LN|LOG|LOG10|LOOKUPVALUE|LOWER|MAX|MAXA|MAXX|MEDIAN|MEDIANX|MID|MIN|MINA|MINUTE|MINX|MOD|MONTH|MROUND|NATURALINNERJOIN|NATURALLEFTOUTERJOIN|NEXTDAY|NEXTMONTH|NEXTQUARTER|NEXTYEAR|NONVISUAL|NORM\.DIST|NORM\.INV|NORM\.S\.DIST|NORM\.S\.INV|NOT|NOW|ODD|OPENINGBALANCEMONTH|OPENINGBALANCEQUARTER|OPENINGBALANCEYEAR|OR|PARALLELPERIOD|PATH|PATHCONTAINS|PATHITEM|PATHITEMREVERSE|PATHLENGTH|PERCENTILE\.EXC|PERCENTILE\.INC|PERCENTILEX\.EXC|PERCENTILEX\.INC|PERMUT|PI|POISSON\.DIST|POWER|PREVIOUSDAY|PREVIOUSMONTH|PREVIOUSQUARTER|PREVIOUSYEAR|PRODUCT|PRODUCTX|QUARTER|QUOTIENT|RADIANS|RAND|RANDBETWEEN|RANK\.EQ|RANKX|RELATED|RELATEDTABLE|REMOVEFILTERS|REPLACE|REPT|RIGHT|ROLLUP|ROLLUPADDISSUBTOTAL|ROLLUPGROUP|ROLLUPISSUBTOTAL|ROUND|ROUNDDOWN|ROUNDUP|ROW|SAMEPERIODLASTYEAR|SAMPLE|SEARCH|SECOND|SELECTCOLUMNS|SELECTEDMEASURE|SELECTEDMEASUREFORMATSTRING|SELECTEDMEASURENAME|SELECTEDVALUE|SIGN|SIN|SINH|SQRT|SQRTPI|STARTOFMONTH|STARTOFQUARTER|STARTOFYEAR|STDEV\.P|STDEV\.S|STDEVX\.P|STDEVX\.S|SUBSTITUTE|SUBSTITUTEWITHINDEX|SUM|SUMMARIZE|SUMMARIZECOLUMNS|SUMX|SWITCH|T\.DIST|T\.DIST\.2T|T\.DIST\.RT|T\.INV|T\.INV\.2T|TAN|TANH|TIME|TIMEVALUE|TODAY|TOPN|TOPNPERLEVEL|TOPNSKIP|TOTALMTD|TOTALQTD|TOTALYTD|TREATAS|TRIM|TRUE|TRUNC|UNICHAR|UNICODE|UNION|UPPER|USERELATIONSHIP|USERNAME|USEROBJECTID|USERPRINCIPALNAME|UTCNOW|UTCTODAY|VALUE|VALUES|VAR\.P|VAR\.S|VARX\.P|VARX\.S|WEEKDAY|WEEKNUM|XIRR|XNPV|YEAR|YEARFRAC)(?=\s*\()/i,keyword:/\b(?:DEFINE|EVALUATE|MEASURE|ORDER\s+BY|RETURN|VAR|START\s+AT|ASC|DESC)\b/i,boolean:{pattern:/\b(?:FALSE|NULL|TRUE)\b/i,alias:"constant"},number:/\b\d+(?:\.\d*)?|\B\.\d+\b/,operator:/:=|[-+*\/=^]|&&?|\|\||<(?:=>?|<|>)?|>[>=]?|\b(?:IN|NOT)\b/i,punctuation:/[;\[\](){}`,.]/}}return nUe}var rUe,Hzt;function K$r(){if(Hzt)return rUe;Hzt=1,rUe=n,n.displayName="dhall",n.aliases=[];function n(r){r.languages.dhall={comment:/--.*|\{-(?:[^-{]|-(?!\})|\{(?!-)|\{-(?:[^-{]|-(?!\})|\{(?!-))*-\})*-\}/,string:{pattern:/"(?:[^"\\]|\\.)*"|''(?:[^']|'(?!')|'''|''\$\{)*''(?!'|\$)/,greedy:!0,inside:{interpolation:{pattern:/\$\{[^{}]*\}/,inside:{expression:{pattern:/(^\$\{)[\s\S]+(?=\}$)/,lookbehind:!0,alias:"language-dhall",inside:null},punctuation:/\$\{|\}/}}}},label:{pattern:/`[^`]*`/,greedy:!0},url:{pattern:/\bhttps?:\/\/[\w.:%!$&'*+;=@~-]+(?:\/[\w.:%!$&'*+;=@~-]*)*(?:\?[/?\w.:%!$&'*+;=@~-]*)?/,greedy:!0},env:{pattern:/\benv:(?:(?!\d)\w+|"(?:[^"\\=]|\\.)*")/,greedy:!0,inside:{function:/^env/,operator:/^:/,variable:/[\s\S]+/}},hash:{pattern:/\bsha256:[\da-fA-F]{64}\b/,inside:{function:/sha256/,operator:/:/,number:/[\da-fA-F]{64}/}},keyword:/\b(?:as|assert|else|forall|if|in|let|merge|missing|then|toMap|using|with)\b|\u2200/,builtin:/\b(?:None|Some)\b/,boolean:/\b(?:False|True)\b/,number:/\bNaN\b|-?\bInfinity\b|[+-]?\b(?:0x[\da-fA-F]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/,operator:/\/\\|\/\/\\\\|&&|\|\||===|[!=]=|\/\/|->|\+\+|::|[+*#@=:?<>|\\\u2227\u2a53\u2261\u2afd\u03bb\u2192]/,punctuation:/\.\.|[{}\[\](),./]/,"class-name":/\b[A-Z]\w*\b/},r.languages.dhall.string.inside.interpolation.inside.expression.inside=r.languages.dhall}return rUe}var iUe,Gzt;function Y$r(){if(Gzt)return iUe;Gzt=1,iUe=n,n.displayName="diff",n.aliases=[];function n(r){(function(o){o.languages.diff={coord:[/^(?:\*{3}|-{3}|\+{3}).*$/m,/^@@.*@@$/m,/^\d.*$/m]};var s={"deleted-sign":"-","deleted-arrow":"<","inserted-sign":"+","inserted-arrow":">",unchanged:" ",diff:"!"};Object.keys(s).forEach(function(c){var d=s[c],h=[];/^\w+$/.test(c)||h.push(/\w+/.exec(c)[0]),c==="diff"&&h.push("bold"),o.languages.diff[c]={pattern:RegExp("^(?:["+d+`].*(?:\r
?|
|(?![\\s\\S])))+`,"m"),alias:h,inside:{line:{pattern:/(.)(?=[\s\S]).*(?:\r\n?|\n)?/,lookbehind:!0},prefix:{pattern:/[\s\S]/,alias:/\w+/.exec(c)[0]}}}}),Object.defineProperty(o.languages.diff,"PREFIXES",{value:s})})(r)}return iUe}var oUe,qzt;function oy(){if(qzt)return oUe;qzt=1,oUe=n,n.displayName="markupTemplating",n.aliases=[];function n(r){(function(o){function s(c,d){return"___"+c.toUpperCase()+d+"___"}Object.defineProperties(o.languages["markup-templating"]={},{buildPlaceholders:{value:function(c,d,h,b){if(c.language===d){var y=c.tokenStack=[];c.code=c.code.replace(h,function(E){if(typeof b=="function"&&!b(E))return E;for(var k=y.length,x;c.code.indexOf(x=s(d,k))!==-1;)++k;return y[k]=E,x}),c.grammar=o.languages.markup}}},tokenizePlaceholders:{value:function(c,d){if(c.language!==d||!c.tokenStack)return;c.grammar=o.languages[d];var h=0,b=Object.keys(c.tokenStack);function y(E){for(var k=0;k<E.length&&!(h>=b.length);k++){var x=E[k];if(typeof x=="string"||x.content&&typeof x.content=="string"){var O=b[h],P=c.tokenStack[O],D=typeof x=="string"?x:x.content,$=s(d,O),U=D.indexOf($);if(U>-1){++h;var B=D.substring(0,U),G=new o.Token(d,o.tokenize(P,c.grammar),"language-"+d,P),K=D.substring(U+$.length),X=[];B&&X.push.apply(X,y([B])),X.push(G),K&&X.push.apply(X,y([K])),typeof x=="string"?E.splice.apply(E,[k,1].concat(X)):x.content=X}}else x.content&&y(x.content)}return E}y(c.tokens)}}})})(r)}return oUe}var aUe,Vzt;function X$r(){if(Vzt)return aUe;Vzt=1;var n=oy();aUe=r,r.displayName="django",r.aliases=["jinja2"];function r(o){o.register(n),function(s){s.languages.django={comment:/^\{#[\s\S]*?#\}$/,tag:{pattern:/(^\{%[+-]?\s*)\w+/,lookbehind:!0,alias:"keyword"},delimiter:{pattern:/^\{[{%][+-]?|[+-]?[}%]\}$/,alias:"punctuation"},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},filter:{pattern:/(\|)\w+/,lookbehind:!0,alias:"function"},test:{pattern:/(\bis\s+(?:not\s+)?)(?!not\b)\w+/,lookbehind:!0,alias:"function"},function:/\b[a-z_]\w+(?=\s*\()/i,keyword:/\b(?:and|as|by|else|for|if|import|in|is|loop|not|or|recursive|with|without)\b/,operator:/[-+%=]=?|!=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]/,number:/\b\d+(?:\.\d+)?\b/,boolean:/[Ff]alse|[Nn]one|[Tt]rue/,variable:/\b\w+\b/,punctuation:/[{}[\](),.:;]/};var c=/\{\{[\s\S]*?\}\}|\{%[\s\S]*?%\}|\{#[\s\S]*?#\}/g,d=s.languages["markup-templating"];s.hooks.add("before-tokenize",function(h){d.buildPlaceholders(h,"django",c)}),s.hooks.add("after-tokenize",function(h){d.tokenizePlaceholders(h,"django")}),s.languages.jinja2=s.languages.django,s.hooks.add("before-tokenize",function(h){d.buildPlaceholders(h,"jinja2",c)}),s.hooks.add("after-tokenize",function(h){d.tokenizePlaceholders(h,"jinja2")})}(o)}return aUe}var sUe,Wzt;function J$r(){if(Wzt)return sUe;Wzt=1,sUe=n,n.displayName="dnsZoneFile",n.aliases=[];function n(r){r.languages["dns-zone-file"]={comment:/;.*/,string:{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0},variable:[{pattern:/(^\$ORIGIN[ \t]+)\S+/m,lookbehind:!0},{pattern:/(^|\s)@(?=\s|$)/,lookbehind:!0}],keyword:/^\$(?:INCLUDE|ORIGIN|TTL)(?=\s|$)/m,class:{pattern:/(^|\s)(?:CH|CS|HS|IN)(?=\s|$)/,lookbehind:!0,alias:"keyword"},type:{pattern:/(^|\s)(?:A|A6|AAAA|AFSDB|APL|ATMA|CAA|CDNSKEY|CDS|CERT|CNAME|DHCID|DLV|DNAME|DNSKEY|DS|EID|GID|GPOS|HINFO|HIP|IPSECKEY|ISDN|KEY|KX|LOC|MAILA|MAILB|MB|MD|MF|MG|MINFO|MR|MX|NAPTR|NB|NBSTAT|NIMLOC|NINFO|NS|NSAP|NSAP-PTR|NSEC|NSEC3|NSEC3PARAM|NULL|NXT|OPENPGPKEY|PTR|PX|RKEY|RP|RRSIG|RT|SIG|SINK|SMIMEA|SOA|SPF|SRV|SSHFP|TA|TKEY|TLSA|TSIG|TXT|UID|UINFO|UNSPEC|URI|WKS|X25)(?=\s|$)/,lookbehind:!0,alias:"keyword"},punctuation:/[()]/},r.languages["dns-zone"]=r.languages["dns-zone-file"]}return sUe}var cUe,Kzt;function Z$r(){if(Kzt)return cUe;Kzt=1,cUe=n,n.displayName="docker",n.aliases=["dockerfile"];function n(r){(function(o){var s=/\\[\r\n](?:\s|\\[\r\n]|#.*(?!.))*(?![\s#]|\\[\r\n])/.source,c=/(?:[ \t]+(?![ \t])(?:<SP_BS>)?|<SP_BS>)/.source.replace(/<SP_BS>/g,function(){return s}),d=/"(?:[^"\\\r\n]|\\(?:\r\n|[\s\S]))*"|'(?:[^'\\\r\n]|\\(?:\r\n|[\s\S]))*'/.source,h=/--[\w-]+=(?:<STR>|(?!["'])(?:[^\s\\]|\\.)+)/.source.replace(/<STR>/g,function(){return d}),b={pattern:RegExp(d),greedy:!0},y={pattern:/(^[ \t]*)#.*/m,lookbehind:!0,greedy:!0};function E(k,x){return k=k.replace(/<OPT>/g,function(){return h}).replace(/<SP>/g,function(){return c}),RegExp(k,x)}o.languages.docker={instruction:{pattern:/(^[ \t]*)(?:ADD|ARG|CMD|COPY|ENTRYPOINT|ENV|EXPOSE|FROM|HEALTHCHECK|LABEL|MAINTAINER|ONBUILD|RUN|SHELL|STOPSIGNAL|USER|VOLUME|WORKDIR)(?=\s)(?:\\.|[^\r\n\\])*(?:\\$(?:\s|#.*$)*(?![\s#])(?:\\.|[^\r\n\\])*)*/im,lookbehind:!0,greedy:!0,inside:{options:{pattern:E(/(^(?:ONBUILD<SP>)?\w+<SP>)<OPT>(?:<SP><OPT>)*/.source,"i"),lookbehind:!0,greedy:!0,inside:{property:{pattern:/(^|\s)--[\w-]+/,lookbehind:!0},string:[b,{pattern:/(=)(?!["'])(?:[^\s\\]|\\.)+/,lookbehind:!0}],operator:/\\$/m,punctuation:/=/}},keyword:[{pattern:E(/(^(?:ONBUILD<SP>)?HEALTHCHECK<SP>(?:<OPT><SP>)*)(?:CMD|NONE)\b/.source,"i"),lookbehind:!0,greedy:!0},{pattern:E(/(^(?:ONBUILD<SP>)?FROM<SP>(?:<OPT><SP>)*(?!--)[^ \t\\]+<SP>)AS/.source,"i"),lookbehind:!0,greedy:!0},{pattern:E(/(^ONBUILD<SP>)\w+/.source,"i"),lookbehind:!0,greedy:!0},{pattern:/^\w+/,greedy:!0}],comment:y,string:b,variable:/\$(?:\w+|\{[^{}"'\\]*\})/,operator:/\\$/m}},comment:y},o.languages.dockerfile=o.languages.docker})(r)}return cUe}var uUe,Yzt;function Q$r(){if(Yzt)return uUe;Yzt=1,uUe=n,n.displayName="dot",n.aliases=["gv"];function n(r){(function(o){var s="(?:"+[/[a-zA-Z_\x80-\uFFFF][\w\x80-\uFFFF]*/.source,/-?(?:\.\d+|\d+(?:\.\d*)?)/.source,/"[^"\\]*(?:\\[\s\S][^"\\]*)*"/.source,/<(?:[^<>]|(?!<!--)<(?:[^<>"']|"[^"]*"|'[^']*')+>|<!--(?:[^-]|-(?!->))*-->)*>/.source].join("|")+")",c={markup:{pattern:/(^<)[\s\S]+(?=>$)/,lookbehind:!0,alias:["language-markup","language-html","language-xml"],inside:o.languages.markup}};function d(h,b){return RegExp(h.replace(/<ID>/g,function(){return s}),b)}o.languages.dot={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\/|^#.*/m,greedy:!0},"graph-name":{pattern:d(/(\b(?:digraph|graph|subgraph)[ \t\r\n]+)<ID>/.source,"i"),lookbehind:!0,greedy:!0,alias:"class-name",inside:c},"attr-value":{pattern:d(/(=[ \t\r\n]*)<ID>/.source),lookbehind:!0,greedy:!0,inside:c},"attr-name":{pattern:d(/([\[;, \t\r\n])<ID>(?=[ \t\r\n]*=)/.source),lookbehind:!0,greedy:!0,inside:c},keyword:/\b(?:digraph|edge|graph|node|strict|subgraph)\b/i,"compass-point":{pattern:/(:[ \t\r\n]*)(?:[ewc_]|[ns][ew]?)(?![\w\x80-\uFFFF])/,lookbehind:!0,alias:"builtin"},node:{pattern:d(/(^|[^-.\w\x80-\uFFFF\\])<ID>/.source),lookbehind:!0,greedy:!0,inside:c},operator:/[=:]|-[->]/,punctuation:/[\[\]{};,]/},o.languages.gv=o.languages.dot})(r)}return uUe}var lUe,Xzt;function eFr(){if(Xzt)return lUe;Xzt=1,lUe=n,n.displayName="ebnf",n.aliases=[];function n(r){r.languages.ebnf={comment:/\(\*[\s\S]*?\*\)/,string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,greedy:!0},special:{pattern:/\?[^?\r\n]*\?/,greedy:!0,alias:"class-name"},definition:{pattern:/^([\t ]*)[a-z]\w*(?:[ \t]+[a-z]\w*)*(?=\s*=)/im,lookbehind:!0,alias:["rule","keyword"]},rule:/\b[a-z]\w*(?:[ \t]+[a-z]\w*)*\b/i,punctuation:/\([:/]|[:/]\)|[.,;()[\]{}]/,operator:/[-=|*/!]/}}return lUe}var dUe,Jzt;function tFr(){if(Jzt)return dUe;Jzt=1,dUe=n,n.displayName="editorconfig",n.aliases=[];function n(r){r.languages.editorconfig={comment:/[;#].*/,section:{pattern:/(^[ \t]*)\[.+\]/m,lookbehind:!0,alias:"selector",inside:{regex:/\\\\[\[\]{},!?.*]/,operator:/[!?]|\.\.|\*{1,2}/,punctuation:/[\[\]{},]/}},key:{pattern:/(^[ \t]*)[^\s=]+(?=[ \t]*=)/m,lookbehind:!0,alias:"attr-name"},value:{pattern:/=.*/,alias:"attr-value",inside:{punctuation:/^=/}}}}return dUe}var fUe,Zzt;function nFr(){if(Zzt)return fUe;Zzt=1,fUe=n,n.displayName="eiffel",n.aliases=[];function n(r){r.languages.eiffel={comment:/--.*/,string:[{pattern:/"([^[]*)\[[\s\S]*?\]\1"/,greedy:!0},{pattern:/"([^{]*)\{[\s\S]*?\}\1"/,greedy:!0},{pattern:/"(?:%(?:(?!\n)\s)*\n\s*%|%\S|[^%"\r\n])*"/,greedy:!0}],char:/'(?:%.|[^%'\r\n])+'/,keyword:/\b(?:across|agent|alias|all|and|as|assign|attached|attribute|check|class|convert|create|Current|debug|deferred|detachable|do|else|elseif|end|ensure|expanded|export|external|feature|from|frozen|if|implies|inherit|inspect|invariant|like|local|loop|not|note|obsolete|old|once|or|Precursor|redefine|rename|require|rescue|Result|retry|select|separate|some|then|undefine|until|variant|Void|when|xor)\b/i,boolean:/\b(?:False|True)\b/i,"class-name":/\b[A-Z][\dA-Z_]*\b/,number:[/\b0[xcb][\da-f](?:_*[\da-f])*\b/i,/(?:\b\d(?:_*\d)*)?\.(?:(?:\d(?:_*\d)*)?e[+-]?)?\d(?:_*\d)*\b|\b\d(?:_*\d)*\b\.?/i],punctuation:/:=|<<|>>|\(\||\|\)|->|\.(?=\w)|[{}[\];(),:?]/,operator:/\\\\|\|\.\.\||\.\.|\/[~\/=]?|[><]=?|[-+*^=~]/}}return fUe}var hUe,Qzt;function rFr(){if(Qzt)return hUe;Qzt=1;var n=oy();hUe=r,r.displayName="ejs",r.aliases=["eta"];function r(o){o.register(n),function(s){s.languages.ejs={delimiter:{pattern:/^<%[-_=]?|[-_]?%>$/,alias:"punctuation"},comment:/^#[\s\S]*/,"language-javascript":{pattern:/[\s\S]+/,inside:s.languages.javascript}},s.hooks.add("before-tokenize",function(c){var d=/<%(?!%)[\s\S]+?%>/g;s.languages["markup-templating"].buildPlaceholders(c,"ejs",d)}),s.hooks.add("after-tokenize",function(c){s.languages["markup-templating"].tokenizePlaceholders(c,"ejs")}),s.languages.eta=s.languages.ejs}(o)}return hUe}var pUe,eHt;function iFr(){if(eHt)return pUe;eHt=1,pUe=n,n.displayName="elixir",n.aliases=[];function n(r){r.languages.elixir={doc:{pattern:/@(?:doc|moduledoc)\s+(?:("""|''')[\s\S]*?\1|("|')(?:\\(?:\r\n|[\s\S])|(?!\2)[^\\\r\n])*\2)/,inside:{attribute:/^@\w+/,string:/['"][\s\S]+/}},comment:{pattern:/#.*/,greedy:!0},regex:{pattern:/~[rR](?:("""|''')(?:\\[\s\S]|(?!\1)[^\\])+\1|([\/|"'])(?:\\.|(?!\2)[^\\\r\n])+\2|\((?:\\.|[^\\)\r\n])+\)|\[(?:\\.|[^\\\]\r\n])+\]|\{(?:\\.|[^\\}\r\n])+\}|<(?:\\.|[^\\>\r\n])+>)[uismxfr]*/,greedy:!0},string:[{pattern:/~[cCsSwW](?:("""|''')(?:\\[\s\S]|(?!\1)[^\\])+\1|([\/|"'])(?:\\.|(?!\2)[^\\\r\n])+\2|\((?:\\.|[^\\)\r\n])+\)|\[(?:\\.|[^\\\]\r\n])+\]|\{(?:\\.|#\{[^}]+\}|#(?!\{)|[^#\\}\r\n])+\}|<(?:\\.|[^\\>\r\n])+>)[csa]?/,greedy:!0,inside:{}},{pattern:/("""|''')[\s\S]*?\1/,greedy:!0,inside:{}},{pattern:/("|')(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0,inside:{}}],atom:{pattern:/(^|[^:]):\w+/,lookbehind:!0,alias:"symbol"},module:{pattern:/\b[A-Z]\w*\b/,alias:"class-name"},"attr-name":/\b\w+\??:(?!:)/,argument:{pattern:/(^|[^&])&\d+/,lookbehind:!0,alias:"variable"},attribute:{pattern:/@\w+/,alias:"variable"},function:/\b[_a-zA-Z]\w*[?!]?(?:(?=\s*(?:\.\s*)?\()|(?=\/\d))/,number:/\b(?:0[box][a-f\d_]+|\d[\d_]*)(?:\.[\d_]+)?(?:e[+-]?[\d_]+)?\b/i,keyword:/\b(?:after|alias|and|case|catch|cond|def(?:callback|delegate|exception|impl|macro|module|n|np|p|protocol|struct)?|do|else|end|fn|for|if|import|not|or|quote|raise|require|rescue|try|unless|unquote|use|when)\b/,boolean:/\b(?:false|nil|true)\b/,operator:[/\bin\b|&&?|\|[|>]?|\\\\|::|\.\.\.?|\+\+?|-[->]?|<[-=>]|>=|!==?|\B!|=(?:==?|[>~])?|[*\/^]/,{pattern:/([^<])<(?!<)/,lookbehind:!0},{pattern:/([^>])>(?!>)/,lookbehind:!0}],punctuation:/<<|>>|[.,%\[\]{}()]/},r.languages.elixir.string.forEach(function(o){o.inside={interpolation:{pattern:/#\{[^}]+\}/,inside:{delimiter:{pattern:/^#\{|\}$/,alias:"punctuation"},rest:r.languages.elixir}}}})}return pUe}var gUe,tHt;function oFr(){if(tHt)return gUe;tHt=1,gUe=n,n.displayName="elm",n.aliases=[];function n(r){r.languages.elm={comment:/--.*|\{-[\s\S]*?-\}/,char:{pattern:/'(?:[^\\'\r\n]|\\(?:[abfnrtv\\']|\d+|x[0-9a-fA-F]+|u\{[0-9a-fA-F]+\}))'/,greedy:!0},string:[{pattern:/"""[\s\S]*?"""/,greedy:!0},{pattern:/"(?:[^\\"\r\n]|\\.)*"/,greedy:!0}],"import-statement":{pattern:/(^[\t ]*)import\s+[A-Z]\w*(?:\.[A-Z]\w*)*(?:\s+as\s+(?:[A-Z]\w*)(?:\.[A-Z]\w*)*)?(?:\s+exposing\s+)?/m,lookbehind:!0,inside:{keyword:/\b(?:as|exposing|import)\b/}},keyword:/\b(?:alias|as|case|else|exposing|if|in|infixl|infixr|let|module|of|then|type)\b/,builtin:/\b(?:abs|acos|always|asin|atan|atan2|ceiling|clamp|compare|cos|curry|degrees|e|flip|floor|fromPolar|identity|isInfinite|isNaN|logBase|max|min|negate|never|not|pi|radians|rem|round|sin|sqrt|tan|toFloat|toPolar|toString|truncate|turns|uncurry|xor)\b/,number:/\b(?:\d+(?:\.\d+)?(?:e[+-]?\d+)?|0x[0-9a-f]+)\b/i,operator:/\s\.\s|[+\-/*=.$<>:&|^?%#@~!]{2,}|[+\-/*=$<>:&|^?%#@~!]/,hvariable:/\b(?:[A-Z]\w*\.)*[a-z]\w*\b/,constant:/\b(?:[A-Z]\w*\.)*[A-Z]\w*\b/,punctuation:/[{}[\]|(),.:]/}}return gUe}var mUe,nHt;function aFr(){if(nHt)return mUe;nHt=1;var n=PEe(),r=oy();mUe=o,o.displayName="erb",o.aliases=[];function o(s){s.register(n),s.register(r),function(c){c.languages.erb={delimiter:{pattern:/^(\s*)<%=?|%>(?=\s*$)/,lookbehind:!0,alias:"punctuation"},ruby:{pattern:/\s*\S[\s\S]*/,alias:"language-ruby",inside:c.languages.ruby}},c.hooks.add("before-tokenize",function(d){var h=/<%=?(?:[^\r\n]|[\r\n](?!=begin)|[\r\n]=begin\s(?:[^\r\n]|[\r\n](?!=end))*[\r\n]=end)+?%>/g;c.languages["markup-templating"].buildPlaceholders(d,"erb",h)}),c.hooks.add("after-tokenize",function(d){c.languages["markup-templating"].tokenizePlaceholders(d,"erb")})}(s)}return mUe}var bUe,rHt;function sFr(){if(rHt)return bUe;rHt=1,bUe=n,n.displayName="erlang",n.aliases=[];function n(r){r.languages.erlang={comment:/%.+/,string:{pattern:/"(?:\\.|[^\\"\r\n])*"/,greedy:!0},"quoted-function":{pattern:/'(?:\\.|[^\\'\r\n])+'(?=\()/,alias:"function"},"quoted-atom":{pattern:/'(?:\\.|[^\\'\r\n])+'/,alias:"atom"},boolean:/\b(?:false|true)\b/,keyword:/\b(?:after|case|catch|end|fun|if|of|receive|try|when)\b/,number:[/\$\\?./,/\b\d+#[a-z0-9]+/i,/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i],function:/\b[a-z][\w@]*(?=\()/,variable:{pattern:/(^|[^@])(?:\b|\?)[A-Z_][\w@]*/,lookbehind:!0},operator:[/[=\/<>:]=|=[:\/]=|\+\+?|--?|[=*\/!]|\b(?:and|andalso|band|bnot|bor|bsl|bsr|bxor|div|not|or|orelse|rem|xor)\b/,{pattern:/(^|[^<])<(?!<)/,lookbehind:!0},{pattern:/(^|[^>])>(?!>)/,lookbehind:!0}],atom:/\b[a-z][\w@]*/,punctuation:/[()[\]{}:;,.#|]|<<|>>/}}return bUe}var vUe,iHt;function cdn(){if(iHt)return vUe;iHt=1,vUe=n,n.displayName="lua",n.aliases=[];function n(r){r.languages.lua={comment:/^#!.+|--(?:\[(=*)\[[\s\S]*?\]\1\]|.*)/m,string:{pattern:/(["'])(?:(?!\1)[^\\\r\n]|\\z(?:\r\n|\s)|\\(?:\r\n|[^z]))*\1|\[(=*)\[[\s\S]*?\]\2\]/,greedy:!0},number:/\b0x[a-f\d]+(?:\.[a-f\d]*)?(?:p[+-]?\d+)?\b|\b\d+(?:\.\B|(?:\.\d*)?(?:e[+-]?\d+)?\b)|\B\.\d+(?:e[+-]?\d+)?\b/i,keyword:/\b(?:and|break|do|else|elseif|end|false|for|function|goto|if|in|local|nil|not|or|repeat|return|then|true|until|while)\b/,function:/(?!\d)\w+(?=\s*(?:[({]))/,operator:[/[-+*%^&|#]|\/\/?|<[<=]?|>[>=]?|[=~]=?/,{pattern:/(^|[^.])\.\.(?!\.)/,lookbehind:!0}],punctuation:/[\[\](){},;]|\.+|:+/}}return vUe}var wUe,oHt;function cFr(){if(oHt)return wUe;oHt=1;var n=cdn(),r=oy();wUe=o,o.displayName="etlua",o.aliases=[];function o(s){s.register(n),s.register(r),function(c){c.languages.etlua={delimiter:{pattern:/^<%[-=]?|-?%>$/,alias:"punctuation"},"language-lua":{pattern:/[\s\S]+/,inside:c.languages.lua}},c.hooks.add("before-tokenize",function(d){var h=/<%[\s\S]+?%>/g;c.languages["markup-templating"].buildPlaceholders(d,"etlua",h)}),c.hooks.add("after-tokenize",function(d){c.languages["markup-templating"].tokenizePlaceholders(d,"etlua")})}(s)}return wUe}var yUe,aHt;function uFr(){if(aHt)return yUe;aHt=1,yUe=n,n.displayName="excelFormula",n.aliases=[];function n(r){r.languages["excel-formula"]={comment:{pattern:/(\bN\(\s*)"(?:[^"]|"")*"(?=\s*\))/i,lookbehind:!0,greedy:!0},string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},reference:{pattern:/(?:'[^']*'|(?:[^\s()[\]{}<>*?"';,$&]*\[[^^\s()[\]{}<>*?"']+\])?\w+)!/,greedy:!0,alias:"string",inside:{operator:/!$/,punctuation:/'/,sheet:{pattern:/[^[\]]+$/,alias:"function"},file:{pattern:/\[[^[\]]+\]$/,inside:{punctuation:/[[\]]/}},path:/[\s\S]+/}},"function-name":{pattern:/\b[A-Z]\w*(?=\()/i,alias:"keyword"},range:{pattern:/\$?\b(?:[A-Z]+\$?\d+:\$?[A-Z]+\$?\d+|[A-Z]+:\$?[A-Z]+|\d+:\$?\d+)\b/i,alias:"property",inside:{operator:/:/,cell:/\$?[A-Z]+\$?\d+/i,column:/\$?[A-Z]+/i,row:/\$?\d+/}},cell:{pattern:/\b[A-Z]+\d+\b|\$[A-Za-z]+\$?\d+\b|\b[A-Za-z]+\$\d+\b/,alias:"property"},number:/(?:\b\d+(?:\.\d+)?|\B\.\d+)(?:e[+-]?\d+)?\b/i,boolean:/\b(?:FALSE|TRUE)\b/i,operator:/[-+*/^%=&,]|<[=>]?|>=?/,punctuation:/[[\]();{}|]/},r.languages.xlsx=r.languages.xls=r.languages["excel-formula"]}return yUe}var EUe,sHt;function lFr(){if(sHt)return EUe;sHt=1,EUe=n,n.displayName="factor",n.aliases=[];function n(r){(function(o){var s={function:/\b(?:BUGS?|FIX(?:MES?)?|NOTES?|TODOS?|XX+|HACKS?|WARN(?:ING)?|\?{2,}|!{2,})\b/},c={number:/\\[^\s']|%\w/},d={comment:[{pattern:/(^|\s)(?:! .*|!$)/,lookbehind:!0,inside:s},{pattern:/(^|\s)\/\*\s[\s\S]*?\*\/(?=\s|$)/,lookbehind:!0,greedy:!0,inside:s},{pattern:/(^|\s)!\[(={0,6})\[\s[\s\S]*?\]\2\](?=\s|$)/,lookbehind:!0,greedy:!0,inside:s}],number:[{pattern:/(^|\s)[+-]?\d+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)[+-]?0(?:b[01]+|o[0-7]+|d\d+|x[\dA-F]+)(?=\s|$)/i,lookbehind:!0},{pattern:/(^|\s)[+-]?\d+\/\d+\.?(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)\+?\d+\+\d+\/\d+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)-\d+-\d+\/\d+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)[+-]?(?:\d*\.\d+|\d+\.\d*|\d+)(?:e[+-]?\d+)?(?=\s|$)/i,lookbehind:!0},{pattern:/(^|\s)NAN:\s+[\da-fA-F]+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)[+-]?0(?:b1\.[01]*|o1\.[0-7]*|d1\.\d*|x1\.[\dA-F]*)p\d+(?=\s|$)/i,lookbehind:!0}],regexp:{pattern:/(^|\s)R\/\s(?:\\\S|[^\\/])*\/(?:[idmsr]*|[idmsr]+-[idmsr]+)(?=\s|$)/,lookbehind:!0,alias:"number",inside:{variable:/\\\S/,keyword:/[+?*\[\]^$(){}.|]/,operator:{pattern:/(\/)[idmsr]+(?:-[idmsr]+)?/,lookbehind:!0}}},boolean:{pattern:/(^|\s)[tf](?=\s|$)/,lookbehind:!0},"custom-string":{pattern:/(^|\s)[A-Z0-9\-]+"\s(?:\\\S|[^"\\])*"/,lookbehind:!0,greedy:!0,alias:"string",inside:{number:/\\\S|%\w|\//}},"multiline-string":[{pattern:/(^|\s)STRING:\s+\S+(?:\n|\r\n).*(?:\n|\r\n)\s*;(?=\s|$)/,lookbehind:!0,greedy:!0,alias:"string",inside:{number:c.number,"semicolon-or-setlocal":{pattern:/([\r\n][ \t]*);(?=\s|$)/,lookbehind:!0,alias:"function"}}},{pattern:/(^|\s)HEREDOC:\s+\S+(?:\n|\r\n).*(?:\n|\r\n)\s*\S+(?=\s|$)/,lookbehind:!0,greedy:!0,alias:"string",inside:c},{pattern:/(^|\s)\[(={0,6})\[\s[\s\S]*?\]\2\](?=\s|$)/,lookbehind:!0,greedy:!0,alias:"string",inside:c}],"special-using":{pattern:/(^|\s)USING:(?:\s\S+)*(?=\s+;(?:\s|$))/,lookbehind:!0,alias:"function",inside:{string:{pattern:/(\s)[^:\s]+/,lookbehind:!0}}},"stack-effect-delimiter":[{pattern:/(^|\s)(?:call|eval|execute)?\((?=\s)/,lookbehind:!0,alias:"operator"},{pattern:/(\s)--(?=\s)/,lookbehind:!0,alias:"operator"},{pattern:/(\s)\)(?=\s|$)/,lookbehind:!0,alias:"operator"}],combinators:{pattern:null,lookbehind:!0,alias:"keyword"},"kernel-builtin":{pattern:null,lookbehind:!0,alias:"variable"},"sequences-builtin":{pattern:null,lookbehind:!0,alias:"variable"},"math-builtin":{pattern:null,lookbehind:!0,alias:"variable"},"constructor-word":{pattern:/(^|\s)<(?!=+>|-+>)\S+>(?=\s|$)/,lookbehind:!0,alias:"keyword"},"other-builtin-syntax":{pattern:null,lookbehind:!0,alias:"operator"},"conventionally-named-word":{pattern:/(^|\s)(?!")(?:(?:change|new|set|with)-\S+|\$\S+|>[^>\s]+|[^:>\s]+>|[^>\s]+>[^>\s]+|\+[^+\s]+\+|[^?\s]+\?|\?[^?\s]+|[^>\s]+>>|>>[^>\s]+|[^<\s]+<<|\([^()\s]+\)|[^!\s]+!|[^*\s]\S*\*|[^.\s]\S*\.)(?=\s|$)/,lookbehind:!0,alias:"keyword"},"colon-syntax":{pattern:/(^|\s)(?:[A-Z0-9\-]+#?)?:{1,2}\s+(?:;\S+|(?!;)\S+)(?=\s|$)/,lookbehind:!0,greedy:!0,alias:"function"},"semicolon-or-setlocal":{pattern:/(\s)(?:;|:>)(?=\s|$)/,lookbehind:!0,alias:"function"},"curly-brace-literal-delimiter":[{pattern:/(^|\s)[a-z]*\{(?=\s)/i,lookbehind:!0,alias:"operator"},{pattern:/(\s)\}(?=\s|$)/,lookbehind:!0,alias:"operator"}],"quotation-delimiter":[{pattern:/(^|\s)\[(?=\s)/,lookbehind:!0,alias:"operator"},{pattern:/(\s)\](?=\s|$)/,lookbehind:!0,alias:"operator"}],"normal-word":{pattern:/(^|\s)[^"\s]\S*(?=\s|$)/,lookbehind:!0},string:{pattern:/"(?:\\\S|[^"\\])*"/,greedy:!0,inside:c}},h=function(k){return(k+"").replace(/([.?*+\^$\[\]\\(){}|\-])/g,"\\$1")},b=function(k){return new RegExp("(^|\\s)(?:"+k.map(h).join("|")+")(?=\\s|$)")},y={"kernel-builtin":["or","2nipd","4drop","tuck","wrapper","nip","wrapper?","callstack>array","die","dupd","callstack","callstack?","3dup","hashcode","pick","4nip","build",">boolean","nipd","clone","5nip","eq?","?","=","swapd","2over","clear","2dup","get-retainstack","not","tuple?","dup","3nipd","call","-rotd","object","drop","assert=","assert?","-rot","execute","boa","get-callstack","curried?","3drop","pickd","overd","over","roll","3nip","swap","and","2nip","rotd","throw","(clone)","hashcode*","spin","reach","4dup","equal?","get-datastack","assert","2drop","<wrapper>","boolean?","identity-hashcode","identity-tuple?","null","composed?","new","5drop","rot","-roll","xor","identity-tuple","boolean"],"other-builtin-syntax":["=======","recursive","flushable",">>","<<<<<<","M\\","B","PRIVATE>","\\","======","final","inline","delimiter","deprecated","<PRIVATE",">>>>>>","<<<<<<<","parse-complex","malformed-complex","read-only",">>>>>>>","call-next-method","<<","foldable","$","$[","${"],"sequences-builtin":["member-eq?","mismatch","append","assert-sequence=","longer","repetition","clone-like","3sequence","assert-sequence?","last-index-from","reversed","index-from","cut*","pad-tail","join-as","remove-eq!","concat-as","but-last","snip","nths","nth","sequence","longest","slice?","<slice>","remove-nth","tail-slice","empty?","tail*","member?","virtual-sequence?","set-length","drop-prefix","iota","unclip","bounds-error?","unclip-last-slice","non-negative-integer-expected","non-negative-integer-expected?","midpoint@","longer?","?set-nth","?first","rest-slice","prepend-as","prepend","fourth","sift","subseq-start","new-sequence","?last","like","first4","1sequence","reverse","slice","virtual@","repetition?","set-last","index","4sequence","max-length","set-second","immutable-sequence","first2","first3","supremum","unclip-slice","suffix!","insert-nth","tail","3append","short","suffix","concat","flip","immutable?","reverse!","2sequence","sum","delete-all","indices","snip-slice","<iota>","check-slice","sequence?","head","append-as","halves","sequence=","collapse-slice","?second","slice-error?","product","bounds-check?","bounds-check","immutable","virtual-exemplar","harvest","remove","pad-head","last","set-fourth","cartesian-product","remove-eq","shorten","shorter","reversed?","shorter?","shortest","head-slice","pop*","tail-slice*","but-last-slice","iota?","append!","cut-slice","new-resizable","head-slice*","sequence-hashcode","pop","set-nth","?nth","second","join","immutable-sequence?","<reversed>","3append-as","virtual-sequence","subseq?","remove-nth!","length","last-index","lengthen","assert-sequence","copy","move","third","first","tail?","set-first","prefix","bounds-error","<repetition>","exchange","surround","cut","min-length","set-third","push-all","head?","subseq-start-from","delete-slice","rest","sum-lengths","head*","infimum","remove!","glue","slice-error","subseq","push","replace-slice","subseq-as","unclip-last"],"math-builtin":["number=","next-power-of-2","?1+","fp-special?","imaginary-part","float>bits","number?","fp-infinity?","bignum?","fp-snan?","denominator","gcd","*","+","fp-bitwise=","-","u>=","/",">=","bitand","power-of-2?","log2-expects-positive","neg?","<","log2",">","integer?","number","bits>double","2/","zero?","bits>float","float?","shift","ratio?","rect>","even?","ratio","fp-sign","bitnot",">fixnum","complex?","/i","integer>fixnum","/f","sgn",">bignum","next-float","u<","u>","mod","recip","rational",">float","2^","integer","fixnum?","neg","fixnum","sq","bignum",">rect","bit?","fp-qnan?","simple-gcd","complex","<fp-nan>","real",">fraction","double>bits","bitor","rem","fp-nan-payload","real-part","log2-expects-positive?","prev-float","align","unordered?","float","fp-nan?","abs","bitxor","integer>fixnum-strict","u<=","odd?","<=","/mod",">integer","real?","rational?","numerator"]};Object.keys(y).forEach(function(k){d[k].pattern=b(y[k])});var E=["2bi","while","2tri","bi*","4dip","both?","same?","tri@","curry","prepose","3bi","?if","tri*","2keep","3keep","curried","2keepd","when","2bi*","2tri*","4keep","bi@","keepdd","do","unless*","tri-curry","if*","loop","bi-curry*","when*","2bi@","2tri@","with","2with","either?","bi","until","3dip","3curry","tri-curry*","tri-curry@","bi-curry","keepd","compose","2dip","if","3tri","unless","tuple","keep","2curry","tri","most","while*","dip","composed","bi-curry@","find-last-from","trim-head-slice","map-as","each-from","none?","trim-tail","partition","if-empty","accumulate*","reject!","find-from","accumulate-as","collector-for-as","reject","map","map-sum","accumulate!","2each-from","follow","supremum-by","map!","unless-empty","collector","padding","reduce-index","replicate-as","infimum-by","trim-tail-slice","count","find-index","filter","accumulate*!","reject-as","map-integers","map-find","reduce","selector","interleave","2map","filter-as","binary-reduce","map-index-as","find","produce","filter!","replicate","cartesian-map","cartesian-each","find-index-from","map-find-last","3map-as","3map","find-last","selector-as","2map-as","2map-reduce","accumulate","each","each-index","accumulate*-as","when-empty","all?","collector-as","push-either","new-like","collector-for","2selector","push-if","2all?","map-reduce","3each","any?","trim-slice","2reduce","change-nth","produce-as","2each","trim","trim-head","cartesian-find","map-index","if-zero","each-integer","unless-zero","(find-integer)","when-zero","find-last-integer","(all-integers?)","times","(each-integer)","find-integer","all-integers?","unless-negative","if-positive","when-positive","when-negative","unless-positive","if-negative","case","2cleave","cond>quot","case>quot","3cleave","wrong-values","to-fixed-point","alist>quot","cond","cleave","call-effect","recursive-hashcode","spread","deep-spread>quot","2||","0||","n||","0&&","2&&","3||","1||","1&&","n&&","3&&","smart-unless*","keep-inputs","reduce-outputs","smart-when*","cleave>array","smart-with","smart-apply","smart-if","inputs/outputs","output>sequence-n","map-outputs","map-reduce-outputs","dropping","output>array","smart-map-reduce","smart-2map-reduce","output>array-n","nullary","input<sequence","append-outputs","drop-inputs","inputs","smart-2reduce","drop-outputs","smart-reduce","preserving","smart-when","outputs","append-outputs-as","smart-unless","smart-if*","sum-outputs","input<sequence-unsafe","output>sequence"];d.combinators.pattern=b(E),o.languages.factor=d})(r)}return EUe}var SUe,cHt;function dFr(){if(cHt)return SUe;cHt=1,SUe=n,n.displayName="$false",n.aliases=[];function n(r){(function(o){o.languages.false={comment:{pattern:/\{[^}]*\}/},string:{pattern:/"[^"]*"/,greedy:!0},"character-code":{pattern:/'(?:[^\r]|\r\n?)/,alias:"number"},"assembler-code":{pattern:/\d+`/,alias:"important"},number:/\d+/,operator:/[-!#$%&'*+,./:;=>?@\\^_`|~ßø]/,punctuation:/\[|\]/,variable:/[a-z]/,"non-standard":{pattern:/[()<BDO®]/,alias:"bold"}}})(r)}return SUe}var _Ue,uHt;function fFr(){if(uHt)return _Ue;uHt=1,_Ue=n,n.displayName="firestoreSecurityRules",n.aliases=[];function n(r){r.languages["firestore-security-rules"]=r.languages.extend("clike",{comment:/\/\/.*/,keyword:/\b(?:allow|function|if|match|null|return|rules_version|service)\b/,operator:/&&|\|\||[<>!=]=?|[-+*/%]|\b(?:in|is)\b/}),delete r.languages["firestore-security-rules"]["class-name"],r.languages.insertBefore("firestore-security-rules","keyword",{path:{pattern:/(^|[\s(),])(?:\/(?:[\w\xA0-\uFFFF]+|\{[\w\xA0-\uFFFF]+(?:=\*\*)?\}|\$\([\w\xA0-\uFFFF.]+\)))+/,lookbehind:!0,greedy:!0,inside:{variable:{pattern:/\{[\w\xA0-\uFFFF]+(?:=\*\*)?\}|\$\([\w\xA0-\uFFFF.]+\)/,inside:{operator:/=/,keyword:/\*\*/,punctuation:/[.$(){}]/}},punctuation:/\//}},method:{pattern:/(\ballow\s+)[a-z]+(?:\s*,\s*[a-z]+)*(?=\s*[:;])/,lookbehind:!0,alias:"builtin",inside:{punctuation:/,/}}})}return _Ue}var kUe,lHt;function hFr(){if(lHt)return kUe;lHt=1,kUe=n,n.displayName="flow",n.aliases=[];function n(r){(function(o){o.languages.flow=o.languages.extend("javascript",{}),o.languages.insertBefore("flow","keyword",{type:[{pattern:/\b(?:[Bb]oolean|Function|[Nn]umber|[Ss]tring|any|mixed|null|void)\b/,alias:"tag"}]}),o.languages.flow["function-variable"].pattern=/(?!\s)[_$a-z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*=\s*(?:function\b|(?:\([^()]*\)(?:\s*:\s*\w+)?|(?!\s)[_$a-z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)\s*=>))/i,delete o.languages.flow.parameter,o.languages.insertBefore("flow","operator",{"flow-punctuation":{pattern:/\{\||\|\}/,alias:"punctuation"}}),Array.isArray(o.languages.flow.keyword)||(o.languages.flow.keyword=[o.languages.flow.keyword]),o.languages.flow.keyword.unshift({pattern:/(^|[^$]\b)(?:Class|declare|opaque|type)\b(?!\$)/,lookbehind:!0},{pattern:/(^|[^$]\B)\$(?:Diff|Enum|Exact|Keys|ObjMap|PropertyType|Record|Shape|Subtype|Supertype|await)\b(?!\$)/,lookbehind:!0})})(r)}return kUe}var TUe,dHt;function pFr(){if(dHt)return TUe;dHt=1,TUe=n,n.displayName="fortran",n.aliases=[];function n(r){r.languages.fortran={"quoted-number":{pattern:/[BOZ](['"])[A-F0-9]+\1/i,alias:"number"},string:{pattern:/(?:\b\w+_)?(['"])(?:\1\1|&(?:\r\n?|\n)(?:[ \t]*!.*(?:\r\n?|\n)|(?![ \t]*!))|(?!\1).)*(?:\1|&)/,inside:{comment:{pattern:/(&(?:\r\n?|\n)\s*)!.*/,lookbehind:!0}}},comment:{pattern:/!.*/,greedy:!0},boolean:/\.(?:FALSE|TRUE)\.(?:_\w+)?/i,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[ED][+-]?\d+)?(?:_\w+)?/i,keyword:[/\b(?:CHARACTER|COMPLEX|DOUBLE ?PRECISION|INTEGER|LOGICAL|REAL)\b/i,/\b(?:END ?)?(?:BLOCK ?DATA|DO|FILE|FORALL|FUNCTION|IF|INTERFACE|MODULE(?! PROCEDURE)|PROGRAM|SELECT|SUBROUTINE|TYPE|WHERE)\b/i,/\b(?:ALLOCATABLE|ALLOCATE|BACKSPACE|CALL|CASE|CLOSE|COMMON|CONTAINS|CONTINUE|CYCLE|DATA|DEALLOCATE|DIMENSION|DO|END|EQUIVALENCE|EXIT|EXTERNAL|FORMAT|GO ?TO|IMPLICIT(?: NONE)?|INQUIRE|INTENT|INTRINSIC|MODULE PROCEDURE|NAMELIST|NULLIFY|OPEN|OPTIONAL|PARAMETER|POINTER|PRINT|PRIVATE|PUBLIC|READ|RETURN|REWIND|SAVE|SELECT|STOP|TARGET|WHILE|WRITE)\b/i,/\b(?:ASSIGNMENT|DEFAULT|ELEMENTAL|ELSE|ELSEIF|ELSEWHERE|ENTRY|IN|INCLUDE|INOUT|KIND|NULL|ONLY|OPERATOR|OUT|PURE|RECURSIVE|RESULT|SEQUENCE|STAT|THEN|USE)\b/i],operator:[/\*\*|\/\/|=>|[=\/]=|[<>]=?|::|[+\-*=%]|\.[A-Z]+\./i,{pattern:/(^|(?!\().)\/(?!\))/,lookbehind:!0}],punctuation:/\(\/|\/\)|[(),;:&]/}}return TUe}var xUe,fHt;function gFr(){if(fHt)return xUe;fHt=1,xUe=n,n.displayName="fsharp",n.aliases=[];function n(r){r.languages.fsharp=r.languages.extend("clike",{comment:[{pattern:/(^|[^\\])\(\*(?!\))[\s\S]*?\*\)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(?:"""[\s\S]*?"""|@"(?:""|[^"])*"|"(?:\\[\s\S]|[^\\"])*")B?/,greedy:!0},"class-name":{pattern:/(\b(?:exception|inherit|interface|new|of|type)\s+|\w\s*:\s*|\s:\??>\s*)[.\w]+\b(?:\s*(?:->|\*)\s*[.\w]+\b)*(?!\s*[:.])/,lookbehind:!0,inside:{operator:/->|\*/,punctuation:/\./}},keyword:/\b(?:let|return|use|yield)(?:!\B|\b)|\b(?:abstract|and|as|asr|assert|atomic|base|begin|break|checked|class|component|const|constraint|constructor|continue|default|delegate|do|done|downcast|downto|eager|elif|else|end|event|exception|extern|external|false|finally|fixed|for|fun|function|functor|global|if|in|include|inherit|inline|interface|internal|land|lazy|lor|lsl|lsr|lxor|match|member|method|mixin|mod|module|mutable|namespace|new|not|null|object|of|open|or|override|parallel|private|process|protected|public|pure|rec|sealed|select|sig|static|struct|tailcall|then|to|trait|true|try|type|upcast|val|virtual|void|volatile|when|while|with)\b/,number:[/\b0x[\da-fA-F]+(?:LF|lf|un)?\b/,/\b0b[01]+(?:uy|y)?\b/,/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[fm]|e[+-]?\d+)?\b/i,/\b\d+(?:[IlLsy]|UL|u[lsy]?)?\b/],operator:/([<>~&^])\1\1|([*.:<>&])\2|<-|->|[!=:]=|<?\|{1,3}>?|\??(?:<=|>=|<>|[-+*/%=<>])\??|[!?^&]|~[+~-]|:>|:\?>?/}),r.languages.insertBefore("fsharp","keyword",{preprocessor:{pattern:/(^[\t ]*)#.*/m,lookbehind:!0,alias:"property",inside:{directive:{pattern:/(^#)\b(?:else|endif|if|light|line|nowarn)\b/,lookbehind:!0,alias:"keyword"}}}}),r.languages.insertBefore("fsharp","punctuation",{"computation-expression":{pattern:/\b[_a-z]\w*(?=\s*\{)/i,alias:"keyword"}}),r.languages.insertBefore("fsharp","string",{annotation:{pattern:/\[<.+?>\]/,greedy:!0,inside:{punctuation:/^\[<|>\]$/,"class-name":{pattern:/^\w+$|(^|;\s*)[A-Z]\w*(?=\()/,lookbehind:!0},"annotation-content":{pattern:/[\s\S]+/,inside:r.languages.fsharp}}},char:{pattern:/'(?:[^\\']|\\(?:.|\d{3}|x[a-fA-F\d]{2}|u[a-fA-F\d]{4}|U[a-fA-F\d]{8}))'B?/,greedy:!0}})}return xUe}var CUe,hHt;function mFr(){if(hHt)return CUe;hHt=1;var n=oy();CUe=r,r.displayName="ftl",r.aliases=[];function r(o){o.register(n),function(s){for(var c=/[^<()"']|\((?:<expr>)*\)|<(?!#--)|<#--(?:[^-]|-(?!->))*-->|"(?:[^\\"]|\\.)*"|'(?:[^\\']|\\.)*'/.source,d=0;d<2;d++)c=c.replace(/<expr>/g,function(){return c});c=c.replace(/<expr>/g,/[^\s\S]/.source);var h={comment:/<#--[\s\S]*?-->/,string:[{pattern:/\br("|')(?:(?!\1)[^\\]|\\.)*\1/,greedy:!0},{pattern:RegExp(/("|')(?:(?!\1|\$\{)[^\\]|\\.|\$\{(?:(?!\})(?:<expr>))*\})*\1/.source.replace(/<expr>/g,function(){return c})),greedy:!0,inside:{interpolation:{pattern:RegExp(/((?:^|[^\\])(?:\\\\)*)\$\{(?:(?!\})(?:<expr>))*\}/.source.replace(/<expr>/g,function(){return c})),lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},rest:null}}}}],keyword:/\b(?:as)\b/,boolean:/\b(?:false|true)\b/,"builtin-function":{pattern:/((?:^|[^?])\?\s*)\w+/,lookbehind:!0,alias:"function"},function:/\b\w+(?=\s*\()/,number:/\b\d+(?:\.\d+)?\b/,operator:/\.\.[<*!]?|->|--|\+\+|&&|\|\||\?{1,2}|[-+*/%!=<>]=?|\b(?:gt|gte|lt|lte)\b/,punctuation:/[,;.:()[\]{}]/};h.string[1].inside.interpolation.inside.rest=h,s.languages.ftl={"ftl-comment":{pattern:/^<#--[\s\S]*/,alias:"comment"},"ftl-directive":{pattern:/^<[\s\S]+>$/,inside:{directive:{pattern:/(^<\/?)[#@][a-z]\w*/i,lookbehind:!0,alias:"keyword"},punctuation:/^<\/?|\/?>$/,content:{pattern:/\s*\S[\s\S]*/,alias:"ftl",inside:h}}},"ftl-interpolation":{pattern:/^\$\{[\s\S]*\}$/,inside:{punctuation:/^\$\{|\}$/,content:{pattern:/\s*\S[\s\S]*/,alias:"ftl",inside:h}}}},s.hooks.add("before-tokenize",function(b){var y=RegExp(/<#--[\s\S]*?-->|<\/?[#@][a-zA-Z](?:<expr>)*?>|\$\{(?:<expr>)*?\}/.source.replace(/<expr>/g,function(){return c}),"gi");s.languages["markup-templating"].buildPlaceholders(b,"ftl",y)}),s.hooks.add("after-tokenize",function(b){s.languages["markup-templating"].tokenizePlaceholders(b,"ftl")})}(o)}return CUe}var AUe,pHt;function bFr(){if(pHt)return AUe;pHt=1,AUe=n,n.displayName="gap",n.aliases=[];function n(r){r.languages.gap={shell:{pattern:/^gap>[\s\S]*?(?=^gap>|$(?![\s\S]))/m,greedy:!0,inside:{gap:{pattern:/^(gap>).+(?:(?:\r(?:\n|(?!\n))|\n)>.*)*/,lookbehind:!0,inside:null},punctuation:/^gap>/}},comment:{pattern:/#.*/,greedy:!0},string:{pattern:/(^|[^\\'"])(?:'(?:[^\r\n\\']|\\.){1,10}'|"(?:[^\r\n\\"]|\\.)*"(?!")|"""[\s\S]*?""")/,lookbehind:!0,greedy:!0,inside:{continuation:{pattern:/([\r\n])>/,lookbehind:!0,alias:"punctuation"}}},keyword:/\b(?:Assert|Info|IsBound|QUIT|TryNextMethod|Unbind|and|atomic|break|continue|do|elif|else|end|fi|for|function|if|in|local|mod|not|od|or|quit|readonly|readwrite|rec|repeat|return|then|until|while)\b/,boolean:/\b(?:false|true)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,number:{pattern:/(^|[^\w.]|\.\.)(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?(?:_[a-z]?)?(?=$|[^\w.]|\.\.)/,lookbehind:!0},continuation:{pattern:/([\r\n])>/,lookbehind:!0,alias:"punctuation"},operator:/->|[-+*/^~=!]|<>|[<>]=?|:=|\.\./,punctuation:/[()[\]{},;.:]/},r.languages.gap.shell.inside.gap.inside=r.languages.gap}return AUe}var IUe,gHt;function vFr(){if(gHt)return IUe;gHt=1,IUe=n,n.displayName="gcode",n.aliases=[];function n(r){r.languages.gcode={comment:/;.*|\B\(.*?\)\B/,string:{pattern:/"(?:""|[^"])*"/,greedy:!0},keyword:/\b[GM]\d+(?:\.\d+)?\b/,property:/\b[A-Z]/,checksum:{pattern:/(\*)\d+/,lookbehind:!0,alias:"number"},punctuation:/[:*]/}}return IUe}var OUe,mHt;function wFr(){if(mHt)return OUe;mHt=1,OUe=n,n.displayName="gdscript",n.aliases=[];function n(r){r.languages.gdscript={comment:/#.*/,string:{pattern:/@?(?:("|')(?:(?!\1)[^\n\\]|\\[\s\S])*\1(?!"|')|"""(?:[^\\]|\\[\s\S])*?""")/,greedy:!0},"class-name":{pattern:/(^(?:class|class_name|extends)[ \t]+|^export\([ \t]*|\bas[ \t]+|(?:\b(?:const|var)[ \t]|[,(])[ \t]*\w+[ \t]*:[ \t]*|->[ \t]*)[a-zA-Z_]\w*/m,lookbehind:!0},keyword:/\b(?:and|as|assert|break|breakpoint|class|class_name|const|continue|elif|else|enum|export|extends|for|func|if|in|is|master|mastersync|match|not|null|onready|or|pass|preload|puppet|puppetsync|remote|remotesync|return|self|setget|signal|static|tool|var|while|yield)\b/,function:/\b[a-z_]\w*(?=[ \t]*\()/i,variable:/\$\w+/,number:[/\b0b[01_]+\b|\b0x[\da-fA-F_]+\b|(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.[\d_]+)(?:e[+-]?[\d_]+)?\b/,/\b(?:INF|NAN|PI|TAU)\b/],constant:/\b[A-Z][A-Z_\d]*\b/,boolean:/\b(?:false|true)\b/,operator:/->|:=|&&|\|\||<<|>>|[-+*/%&|!<>=]=?|[~^]/,punctuation:/[.:,;()[\]{}]/}}return OUe}var NUe,bHt;function yFr(){if(bHt)return NUe;bHt=1,NUe=n,n.displayName="gedcom",n.aliases=[];function n(r){r.languages.gedcom={"line-value":{pattern:/(^[\t ]*\d+ +(?:@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@ +)?\w+ ).+/m,lookbehind:!0,inside:{pointer:{pattern:/^@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@$/,alias:"variable"}}},tag:{pattern:/(^[\t ]*\d+ +(?:@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@ +)?)\w+/m,lookbehind:!0,alias:"string"},level:{pattern:/(^[\t ]*)\d+/m,lookbehind:!0,alias:"number"},pointer:{pattern:/@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@/,alias:"variable"}}}return NUe}var RUe,vHt;function EFr(){if(vHt)return RUe;vHt=1,RUe=n,n.displayName="gherkin",n.aliases=[];function n(r){(function(o){var s=/(?:\r?\n|\r)[ \t]*\|.+\|(?:(?!\|).)*/.source;o.languages.gherkin={pystring:{pattern:/("""|''')[\s\S]+?\1/,alias:"string"},comment:{pattern:/(^[ \t]*)#.*/m,lookbehind:!0},tag:{pattern:/(^[ \t]*)@\S*/m,lookbehind:!0},feature:{pattern:/((?:^|\r?\n|\r)[ \t]*)(?:Ability|Ahoy matey!|Arwedd|Aspekt|Besigheid Behoefte|Business Need|Caracteristica|Característica|Egenskab|Egenskap|Eiginleiki|Feature|Fīča|Fitur|Fonctionnalité|Fonksyonalite|Funcionalidade|Funcionalitat|Functionalitate|Funcţionalitate|Funcționalitate|Functionaliteit|Fungsi|Funkcia|Funkcija|Funkcionalitāte|Funkcionalnost|Funkcja|Funksie|Funktionalität|Funktionalitéit|Funzionalità|Hwaet|Hwæt|Jellemző|Karakteristik|Lastnost|Mak|Mogucnost|laH|Mogućnost|Moznosti|Možnosti|OH HAI|Omadus|Ominaisuus|Osobina|Özellik|Potrzeba biznesowa|perbogh|poQbogh malja'|Požadavek|Požiadavka|Pretty much|Qap|Qu'meH 'ut|Savybė|Tính năng|Trajto|Vermoë|Vlastnosť|Właściwość|Značilnost|Δυνατότητα|Λειτουργία|Могућност|Мөмкинлек|Особина|Свойство|Үзенчәлеклелек|Функционал|Функционалност|Функция|Функціонал|תכונה|خاصية|خصوصیت|صلاحیت|کاروبار کی ضرورت|وِیژگی|रूप लेख|ਖਾਸੀਅਤ|ਨਕਸ਼ ਨੁਹਾਰ|ਮੁਹਾਂਦਰਾ|గుణము|ಹೆಚ್ಚಳ|ความต้องการทางธุรกิจ|ความสามารถ|โครงหลัก|기능|フィーチャ|功能|機能):(?:[^:\r\n]+(?:\r?\n|\r|$))*/,lookbehind:!0,inside:{important:{pattern:/(:)[^\r\n]+/,lookbehind:!0},keyword:/[^:\r\n]+:/}},scenario:{pattern:/(^[ \t]*)(?:Abstract Scenario|Abstrakt Scenario|Achtergrond|Aer|Ær|Agtergrond|All y'all|Antecedentes|Antecedents|Atburðarás|Atburðarásir|Awww, look mate|B4|Background|Baggrund|Bakgrund|Bakgrunn|Bakgrunnur|Beispiele|Beispiller|Bối cảnh|Cefndir|Cenario|Cenário|Cenario de Fundo|Cenário de Fundo|Cenarios|Cenários|Contesto|Context|Contexte|Contexto|Conto|Contoh|Contone|Dæmi|Dasar|Dead men tell no tales|Delineacao do Cenario|Delineação do Cenário|Dis is what went down|Dữ liệu|Dyagram Senaryo|Dyagram senaryo|Egzanp|Ejemplos|Eksempler|Ekzemploj|Enghreifftiau|Esbozo do escenario|Escenari|Escenario|Esempi|Esquema de l'escenari|Esquema del escenario|Esquema do Cenario|Esquema do Cenário|EXAMPLZ|Examples|Exempel|Exemple|Exemples|Exemplos|First off|Fono|Forgatókönyv|Forgatókönyv vázlat|Fundo|Geçmiş|Grundlage|Hannergrond|ghantoH|Háttér|Heave to|Istorik|Juhtumid|Keadaan|Khung kịch bản|Khung tình huống|Kịch bản|Koncept|Konsep skenario|Kontèks|Kontekst|Kontekstas|Konteksts|Kontext|Konturo de la scenaro|Latar Belakang|lut chovnatlh|lut|lutmey|Lýsing Atburðarásar|Lýsing Dæma|MISHUN SRSLY|MISHUN|Menggariskan Senario|mo'|Náčrt Scenára|Náčrt Scénáře|Náčrt Scenáru|Oris scenarija|Örnekler|Osnova|Osnova Scenára|Osnova scénáře|Osnutek|Ozadje|Paraugs|Pavyzdžiai|Példák|Piemēri|Plan du scénario|Plan du Scénario|Plan Senaryo|Plan senaryo|Plang vum Szenario|Pozadí|Pozadie|Pozadina|Príklady|Příklady|Primer|Primeri|Primjeri|Przykłady|Raamstsenaarium|Reckon it's like|Rerefons|Scenár|Scénář|Scenarie|Scenarij|Scenarijai|Scenarijaus šablonas|Scenariji|Scenārijs|Scenārijs pēc parauga|Scenarijus|Scenario|Scénario|Scenario Amlinellol|Scenario Outline|Scenario Template|Scenariomal|Scenariomall|Scenarios|Scenariu|Scenariusz|Scenaro|Schema dello scenario|Se ðe|Se the|Se þe|Senario|Senaryo Deskripsyon|Senaryo deskripsyon|Senaryo|Senaryo taslağı|Shiver me timbers|Situācija|Situai|Situasie Uiteensetting|Situasie|Skenario konsep|Skenario|Skica|Structura scenariu|Structură scenariu|Struktura scenarija|Stsenaarium|Swa hwaer swa|Swa|Swa hwær swa|Szablon scenariusza|Szenario|Szenariogrundriss|Tapaukset|Tapaus|Tapausaihio|Taust|Tausta|Template Keadaan|Template Senario|Template Situai|The thing of it is|Tình huống|Variantai|Voorbeelde|Voorbeelden|Wharrimean is|Yo-ho-ho|You'll wanna|Założenia|Παραδείγματα|Περιγραφή Σεναρίου|Σενάρια|Σενάριο|Υπόβαθρο|Кереш|Контекст|Концепт|Мисаллар|Мисоллар|Основа|Передумова|Позадина|Предистория|Предыстория|Приклади|Пример|Примери|Примеры|Рамка на сценарий|Скица|Структура сценарија|Структура сценария|Структура сценарію|Сценарий|Сценарий структураси|Сценарийның төзелеше|Сценарији|Сценарио|Сценарій|Тарих|Үрнәкләр|דוגמאות|רקע|תבנית תרחיש|תרחיש|الخلفية|الگوی سناریو|امثلة|پس منظر|زمینه|سناریو|سيناريو|سيناريو مخطط|مثالیں|منظر نامے کا خاکہ|منظرنامہ|نمونه ها|उदाहरण|परिदृश्य|परिदृश्य रूपरेखा|पृष्ठभूमि|ਉਦਾਹਰਨਾਂ|ਪਟਕਥਾ|ਪਟਕਥਾ ਢਾਂਚਾ|ਪਟਕਥਾ ਰੂਪ ਰੇਖਾ|ਪਿਛੋਕੜ|ఉదాహరణలు|కథనం|నేపథ్యం|సన్నివేశం|ಉದಾಹರಣೆಗಳು|ಕಥಾಸಾರಾಂಶ|ವಿವರಣೆ|ಹಿನ್ನೆಲೆ|โครงสร้างของเหตุการณ์|ชุดของตัวอย่าง|ชุดของเหตุการณ์|แนวคิด|สรุปเหตุการณ์|เหตุการณ์|배경|시나리오|시나리오 개요|예|サンプル|シナリオ|シナリオアウトライン|シナリオテンプレ|シナリオテンプレート|テンプレ|例|例子|剧本|剧本大纲|劇本|劇本大綱|场景|场景大纲|場景|場景大綱|背景):[^:\r\n]*/m,lookbehind:!0,inside:{important:{pattern:/(:)[^\r\n]*/,lookbehind:!0},keyword:/[^:\r\n]+:/}},"table-body":{pattern:RegExp("("+s+")(?:"+s+")+"),lookbehind:!0,inside:{outline:{pattern:/<[^>]+>/,alias:"variable"},td:{pattern:/\s*[^\s|][^|]*/,alias:"string"},punctuation:/\|/}},"table-head":{pattern:RegExp(s),inside:{th:{pattern:/\s*[^\s|][^|]*/,alias:"variable"},punctuation:/\|/}},atrule:{pattern:/(^[ \t]+)(?:'a|'ach|'ej|7|a|A také|A taktiež|A tiež|A zároveň|Aber|Ac|Adott|Akkor|Ak|Aleshores|Ale|Ali|Allora|Alors|Als|Ama|Amennyiben|Amikor|Ampak|an|AN|Ananging|And y'all|And|Angenommen|Anrhegedig a|An|Apabila|Atès|Atesa|Atunci|Avast!|Aye|A|awer|Bagi|Banjur|Bet|Biết|Blimey!|Buh|But at the end of the day I reckon|But y'all|But|BUT|Cal|Când|Cand|Cando|Ce|Cuando|Če|Ða ðe|Ða|Dadas|Dada|Dados|Dado|DaH ghu' bejlu'|dann|Dann|Dano|Dan|Dar|Dat fiind|Data|Date fiind|Date|Dati fiind|Dati|Daţi fiind|Dați fiind|DEN|Dato|De|Den youse gotta|Dengan|Diberi|Diyelim ki|Donada|Donat|Donitaĵo|Do|Dun|Duota|Ðurh|Eeldades|Ef|Eğer ki|Entao|Então|Entón|E|En|Entonces|Epi|És|Etant donnée|Etant donné|Et|Étant données|Étant donnée|Étant donné|Etant données|Etant donnés|Étant donnés|Fakat|Gangway!|Gdy|Gegeben seien|Gegeben sei|Gegeven|Gegewe|ghu' noblu'|Gitt|Given y'all|Given|Givet|Givun|Ha|Cho|I CAN HAZ|In|Ir|It's just unbelievable|I|Ja|Jeśli|Jeżeli|Kad|Kada|Kadar|Kai|Kaj|Když|Keď|Kemudian|Ketika|Khi|Kiedy|Ko|Kuid|Kui|Kun|Lan|latlh|Le sa a|Let go and haul|Le|Lè sa a|Lè|Logo|Lorsqu'<|Lorsque|mä|Maar|Mais|Mając|Ma|Majd|Maka|Manawa|Mas|Men|Menawa|Mutta|Nalika|Nalikaning|Nanging|Når|När|Nato|Nhưng|Niin|Njuk|O zaman|Och|Og|Oletetaan|Ond|Onda|Oraz|Pak|Pero|Però|Podano|Pokiaľ|Pokud|Potem|Potom|Privzeto|Pryd|Quan|Quand|Quando|qaSDI'|Så|Sed|Se|Siis|Sipoze ke|Sipoze Ke|Sipoze|Si|Şi|Și|Soit|Stel|Tada|Tad|Takrat|Tak|Tapi|Ter|Tetapi|Tha the|Tha|Then y'all|Then|Thì|Thurh|Toda|Too right|Un|Und|ugeholl|Và|vaj|Vendar|Ve|wann|Wanneer|WEN|Wenn|When y'all|When|Wtedy|Wun|Y'know|Yeah nah|Yna|Youse know like when|Youse know when youse got|Y|Za predpokladu|Za předpokladu|Zadan|Zadani|Zadano|Zadate|Zadato|Zakładając|Zaradi|Zatati|Þa þe|Þa|Þá|Þegar|Þurh|Αλλά|Δεδομένου|Και|Όταν|Τότε|А також|Агар|Але|Али|Аммо|А|Әгәр|Әйтик|Әмма|Бирок|Ва|Вә|Дадено|Дано|Допустим|Если|Задате|Задати|Задато|И|І|К тому же|Када|Кад|Когато|Когда|Коли|Ләкин|Лекин|Нәтиҗәдә|Нехай|Но|Онда|Припустимо, що|Припустимо|Пусть|Также|Та|Тогда|Тоді|То|Унда|Һәм|Якщо|אבל|אזי|אז|בהינתן|וגם|כאשר|آنگاه|اذاً|اگر|اما|اور|با فرض|بالفرض|بفرض|پھر|تب|ثم|جب|عندما|فرض کیا|لكن|لیکن|متى|هنگامی|و|अगर|और|कदा|किन्तु|चूंकि|जब|तथा|तदा|तब|परन्तु|पर|यदि|ਅਤੇ|ਜਦੋਂ|ਜਿਵੇਂ ਕਿ|ਜੇਕਰ|ਤਦ|ਪਰ|అప్పుడు|ఈ పరిస్థితిలో|కాని|చెప్పబడినది|మరియు|ಆದರೆ|ನಂತರ|ನೀಡಿದ|ಮತ್ತು|ಸ್ಥಿತಿಯನ್ನು|กำหนดให้|ดังนั้น|แต่|เมื่อ|และ|그러면<|그리고<|단<|만약<|만일<|먼저<|조건<|하지만<|かつ<|しかし<|ただし<|ならば<|もし<|並且<|但し<|但是<|假如<|假定<|假設<|假设<|前提<|同时<|同時<|并且<|当<|當<|而且<|那么<|那麼<)(?=[ \t])/m,lookbehind:!0},string:{pattern:/"(?:\\.|[^"\\\r\n])*"|'(?:\\.|[^'\\\r\n])*'/,inside:{outline:{pattern:/<[^>]+>/,alias:"variable"}}},outline:{pattern:/<[^>]+>/,alias:"variable"}}})(r)}return RUe}var PUe,wHt;function SFr(){if(wHt)return PUe;wHt=1,PUe=n,n.displayName="git",n.aliases=[];function n(r){r.languages.git={comment:/^#.*/m,deleted:/^[-–].*/m,inserted:/^\+.*/m,string:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,command:{pattern:/^.*\$ git .*$/m,inside:{parameter:/\s--?\w+/}},coord:/^@@.*@@$/m,"commit-sha1":/^commit \w{40}$/m}}return PUe}var LUe,yHt;function _Fr(){if(yHt)return LUe;yHt=1;var n=vU();LUe=r,r.displayName="glsl",r.aliases=[];function r(o){o.register(n),o.languages.glsl=o.languages.extend("c",{keyword:/\b(?:active|asm|atomic_uint|attribute|[ibdu]?vec[234]|bool|break|buffer|case|cast|centroid|class|coherent|common|const|continue|d?mat[234](?:x[234])?|default|discard|do|double|else|enum|extern|external|false|filter|fixed|flat|float|for|fvec[234]|goto|half|highp|hvec[234]|[iu]?sampler2DMS(?:Array)?|[iu]?sampler2DRect|[iu]?samplerBuffer|[iu]?samplerCube|[iu]?samplerCubeArray|[iu]?sampler[123]D|[iu]?sampler[12]DArray|[iu]?image2DMS(?:Array)?|[iu]?image2DRect|[iu]?imageBuffer|[iu]?imageCube|[iu]?imageCubeArray|[iu]?image[123]D|[iu]?image[12]DArray|if|in|inline|inout|input|int|interface|invariant|layout|long|lowp|mediump|namespace|noinline|noperspective|out|output|partition|patch|precise|precision|public|readonly|resource|restrict|return|sample|sampler[12]DArrayShadow|sampler[12]DShadow|sampler2DRectShadow|sampler3DRect|samplerCubeArrayShadow|samplerCubeShadow|shared|short|sizeof|smooth|static|struct|subroutine|superp|switch|template|this|true|typedef|uint|uniform|union|unsigned|using|varying|void|volatile|while|writeonly)\b/})}return LUe}var DUe,EHt;function kFr(){if(EHt)return DUe;EHt=1,DUe=n,n.displayName="gml",n.aliases=[];function n(r){r.languages.gamemakerlanguage=r.languages.gml=r.languages.extend("clike",{keyword:/\b(?:break|case|continue|default|do|else|enum|exit|for|globalvar|if|repeat|return|switch|until|var|while)\b/,number:/(?:\b0x[\da-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)[ulf]{0,4}/i,operator:/--|\+\+|[-+%/=]=?|!=|\*\*?=?|<[<=>]?|>[=>]?|&&?|\^\^?|\|\|?|~|\b(?:and|at|not|or|with|xor)\b/,constant:/\b(?:GM_build_date|GM_version|action_(?:continue|restart|reverse|stop)|all|gamespeed_(?:fps|microseconds)|global|local|noone|other|pi|pointer_(?:invalid|null)|self|timezone_(?:local|utc)|undefined|ev_(?:create|destroy|step|alarm|keyboard|mouse|collision|other|draw|draw_(?:begin|end|post|pre)|keypress|keyrelease|trigger|(?:left|middle|no|right)_button|(?:left|middle|right)_press|(?:left|middle|right)_release|mouse_(?:enter|leave|wheel_down|wheel_up)|global_(?:left|middle|right)_button|global_(?:left|middle|right)_press|global_(?:left|middle|right)_release|joystick(?:1|2)_(?:button1|button2|button3|button4|button5|button6|button7|button8|down|left|right|up)|outside|boundary|game_start|game_end|room_start|room_end|no_more_lives|animation_end|end_of_path|no_more_health|user\d|gui|gui_begin|gui_end|step_(?:begin|end|normal))|vk_(?:alt|anykey|backspace|control|delete|down|end|enter|escape|home|insert|left|nokey|pagedown|pageup|pause|printscreen|return|right|shift|space|tab|up|f\d|numpad\d|add|decimal|divide|lalt|lcontrol|lshift|multiply|ralt|rcontrol|rshift|subtract)|achievement_(?:filter_(?:all_players|favorites_only|friends_only)|friends_info|info|leaderboard_info|our_info|pic_loaded|show_(?:achievement|bank|friend_picker|leaderboard|profile|purchase_prompt|ui)|type_challenge|type_score_challenge)|asset_(?:font|object|path|room|script|shader|sound|sprite|tiles|timeline|unknown)|audio_(?:3d|falloff_(?:exponent_distance|exponent_distance_clamped|inverse_distance|inverse_distance_clamped|linear_distance|linear_distance_clamped|none)|mono|new_system|old_system|stereo)|bm_(?:add|complex|dest_alpha|dest_color|dest_colour|inv_dest_alpha|inv_dest_color|inv_dest_colour|inv_src_alpha|inv_src_color|inv_src_colour|max|normal|one|src_alpha|src_alpha_sat|src_color|src_colour|subtract|zero)|browser_(?:chrome|firefox|ie|ie_mobile|not_a_browser|opera|safari|safari_mobile|tizen|unknown|windows_store)|buffer_(?:bool|f16|f32|f64|fast|fixed|generalerror|grow|invalidtype|network|outofbounds|outofspace|s16|s32|s8|seek_end|seek_relative|seek_start|string|text|u16|u32|u64|u8|vbuffer|wrap)|c_(?:aqua|black|blue|dkgray|fuchsia|gray|green|lime|ltgray|maroon|navy|olive|orange|purple|red|silver|teal|white|yellow)|cmpfunc_(?:always|equal|greater|greaterequal|less|lessequal|never|notequal)|cr_(?:appstart|arrow|beam|cross|default|drag|handpoint|hourglass|none|size_all|size_nesw|size_ns|size_nwse|size_we|uparrow)|cull_(?:clockwise|counterclockwise|noculling)|device_(?:emulator|tablet)|device_ios_(?:ipad|ipad_retina|iphone|iphone5|iphone6|iphone6plus|iphone_retina|unknown)|display_(?:landscape|landscape_flipped|portrait|portrait_flipped)|dll_(?:cdecl|cdel|stdcall)|ds_type_(?:grid|list|map|priority|queue|stack)|ef_(?:cloud|ellipse|explosion|firework|flare|rain|ring|smoke|smokeup|snow|spark|star)|fa_(?:archive|bottom|center|directory|hidden|left|middle|readonly|right|sysfile|top|volumeid)|fb_login_(?:default|fallback_to_webview|forcing_safari|forcing_webview|no_fallback_to_webview|use_system_account)|iap_(?:available|canceled|ev_consume|ev_product|ev_purchase|ev_restore|ev_storeload|failed|purchased|refunded|status_available|status_loading|status_processing|status_restoring|status_unavailable|status_uninitialised|storeload_failed|storeload_ok|unavailable)|leaderboard_type_(?:number|time_mins_secs)|lighttype_(?:dir|point)|matrix_(?:projection|view|world)|mb_(?:any|left|middle|none|right)|network_(?:config_(?:connect_timeout|disable_reliable_udp|enable_reliable_udp|use_non_blocking_socket)|socket_(?:bluetooth|tcp|udp)|type_(?:connect|data|disconnect|non_blocking_connect))|of_challenge_(?:lose|tie|win)|os_(?:android|ios|linux|macosx|ps3|ps4|psvita|unknown|uwp|win32|win8native|windows|winphone|xboxone)|phy_debug_render_(?:aabb|collision_pairs|coms|core_shapes|joints|obb|shapes)|phy_joint_(?:anchor_1_x|anchor_1_y|anchor_2_x|anchor_2_y|angle|angle_limits|damping_ratio|frequency|length_1|length_2|lower_angle_limit|max_force|max_length|max_motor_force|max_motor_torque|max_torque|motor_force|motor_speed|motor_torque|reaction_force_x|reaction_force_y|reaction_torque|speed|translation|upper_angle_limit)|phy_particle_data_flag_(?:category|color|colour|position|typeflags|velocity)|phy_particle_flag_(?:colormixing|colourmixing|elastic|powder|spring|tensile|viscous|wall|water|zombie)|phy_particle_group_flag_(?:rigid|solid)|pr_(?:linelist|linestrip|pointlist|trianglefan|trianglelist|trianglestrip)|ps_(?:distr|shape)_(?:diamond|ellipse|gaussian|invgaussian|line|linear|rectangle)|pt_shape_(?:circle|cloud|disk|explosion|flare|line|pixel|ring|smoke|snow|spark|sphere|square|star)|ty_(?:real|string)|gp_(?:face\d|axislh|axislv|axisrh|axisrv|padd|padl|padr|padu|select|shoulderl|shoulderlb|shoulderr|shoulderrb|start|stickl|stickr)|lb_disp_(?:none|numeric|time_ms|time_sec)|lb_sort_(?:ascending|descending|none)|ov_(?:achievements|community|friends|gamegroup|players|settings)|ugc_(?:filetype_(?:community|microtrans)|list_(?:Favorited|Followed|Published|Subscribed|UsedOrPlayed|VotedDown|VotedOn|VotedUp|WillVoteLater)|match_(?:AllGuides|Artwork|Collections|ControllerBindings|IntegratedGuides|Items|Items_Mtx|Items_ReadyToUse|Screenshots|UsableInGame|Videos|WebGuides)|query_(?:AcceptedForGameRankedByAcceptanceDate|CreatedByFriendsRankedByPublicationDate|FavoritedByFriendsRankedByPublicationDate|NotYetRated)|query_RankedBy(?:NumTimesReported|PublicationDate|TextSearch|TotalVotesAsc|Trend|Vote|VotesUp)|result_success|sortorder_CreationOrder(?:Asc|Desc)|sortorder_(?:ForModeration|LastUpdatedDesc|SubscriptionDateDesc|TitleAsc|VoteScoreDesc)|visibility_(?:friends_only|private|public))|vertex_usage_(?:binormal|blendindices|blendweight|color|colour|depth|fog|normal|position|psize|sample|tangent|texcoord|textcoord)|vertex_type_(?:float\d|color|colour|ubyte4)|input_type|layerelementtype_(?:background|instance|oldtilemap|particlesystem|sprite|tile|tilemap|undefined)|se_(?:chorus|compressor|echo|equalizer|flanger|gargle|none|reverb)|text_type|tile_(?:flip|index_mask|mirror|rotate)|(?:obj|rm|scr|spr)\w+)\b/,variable:/\b(?:alarm|application_surface|async_load|background_(?:alpha|blend|color|colour|foreground|height|hspeed|htiled|index|showcolor|showcolour|visible|vspeed|vtiled|width|x|xscale|y|yscale)|bbox_(?:bottom|left|right|top)|browser_(?:height|width)|caption_(?:health|lives|score)|current_(?:day|hour|minute|month|second|time|weekday|year)|cursor_sprite|debug_mode|delta_time|direction|display_aa|error_(?:last|occurred)|event_(?:action|number|object|type)|fps|fps_real|friction|game_(?:display|project|save)_(?:id|name)|gamemaker_(?:pro|registered|version)|gravity|gravity_direction|(?:h|v)speed|health|iap_data|id|image_(?:alpha|angle|blend|depth|index|number|speed|xscale|yscale)|instance_(?:count|id)|keyboard_(?:key|lastchar|lastkey|string)|layer|lives|mask_index|mouse_(?:button|lastbutton|x|y)|object_index|os_(?:browser|device|type|version)|path_(?:endaction|index|orientation|position|positionprevious|scale|speed)|persistent|phy_(?:rotation|(?:col_normal|collision|com|linear_velocity|position|speed)_(?:x|y)|angular_(?:damping|velocity)|position_(?:x|y)previous|speed|linear_damping|bullet|fixed_rotation|active|mass|inertia|dynamic|kinematic|sleeping|collision_points)|pointer_(?:invalid|null)|room|room_(?:caption|first|height|last|persistent|speed|width)|score|secure_mode|show_(?:health|lives|score)|solid|speed|sprite_(?:height|index|width|xoffset|yoffset)|temp_directory|timeline_(?:index|loop|position|running|speed)|transition_(?:color|kind|steps)|undefined|view_(?:angle|current|enabled|(?:h|v)(?:border|speed)|(?:h|w|x|y)port|(?:h|w|x|y)view|object|surface_id|visible)|visible|webgl_enabled|working_directory|(?:x|y)(?:previous|start)|x|y|argument(?:_relitive|_count|\d)|argument|global|local|other|self)\b/})}return DUe}var MUe,SHt;function TFr(){if(SHt)return MUe;SHt=1,MUe=n,n.displayName="gn",n.aliases=["gni"];function n(r){r.languages.gn={comment:{pattern:/#.*/,greedy:!0},"string-literal":{pattern:/(^|[^\\"])"(?:[^\r\n"\\]|\\.)*"/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:\{[\s\S]*?\}|[a-zA-Z_]\w*|0x[a-fA-F0-9]{2})/,lookbehind:!0,inside:{number:/^\$0x[\s\S]{2}$/,variable:/^\$\w+$/,"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:null}}},string:/[\s\S]+/}},keyword:/\b(?:else|if)\b/,boolean:/\b(?:false|true)\b/,"builtin-function":{pattern:/\b(?:assert|defined|foreach|import|pool|print|template|tool|toolchain)(?=\s*\()/i,alias:"keyword"},function:/\b[a-z_]\w*(?=\s*\()/i,constant:/\b(?:current_cpu|current_os|current_toolchain|default_toolchain|host_cpu|host_os|root_build_dir|root_gen_dir|root_out_dir|target_cpu|target_gen_dir|target_os|target_out_dir)\b/,number:/-?\b\d+\b/,operator:/[-+!=<>]=?|&&|\|\|/,punctuation:/[(){}[\],.]/},r.languages.gn["string-literal"].inside.interpolation.inside.expression.inside=r.languages.gn,r.languages.gni=r.languages.gn}return MUe}var jUe,_Ht;function xFr(){if(_Ht)return jUe;_Ht=1,jUe=n,n.displayName="goModule",n.aliases=[];function n(r){r.languages["go-mod"]=r.languages["go-module"]={comment:{pattern:/\/\/.*/,greedy:!0},version:{pattern:/(^|[\s()[\],])v\d+\.\d+\.\d+(?:[+-][-+.\w]*)?(?![^\s()[\],])/,lookbehind:!0,alias:"number"},"go-version":{pattern:/((?:^|\s)go\s+)\d+(?:\.\d+){1,2}/,lookbehind:!0,alias:"number"},keyword:{pattern:/^([ \t]*)(?:exclude|go|module|replace|require|retract)\b/m,lookbehind:!0},operator:/=>/,punctuation:/[()[\],]/}}return jUe}var $Ue,kHt;function CFr(){if(kHt)return $Ue;kHt=1,$Ue=n,n.displayName="go",n.aliases=[];function n(r){r.languages.go=r.languages.extend("clike",{string:{pattern:/(^|[^\\])"(?:\\.|[^"\\\r\n])*"|`[^`]*`/,lookbehind:!0,greedy:!0},keyword:/\b(?:break|case|chan|const|continue|default|defer|else|fallthrough|for|func|go(?:to)?|if|import|interface|map|package|range|return|select|struct|switch|type|var)\b/,boolean:/\b(?:_|false|iota|nil|true)\b/,number:[/\b0(?:b[01_]+|o[0-7_]+)i?\b/i,/\b0x(?:[a-f\d_]+(?:\.[a-f\d_]*)?|\.[a-f\d_]+)(?:p[+-]?\d+(?:_\d+)*)?i?(?!\w)/i,/(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.\d[\d_]*)(?:e[+-]?[\d_]+)?i?(?!\w)/i],operator:/[*\/%^!=]=?|\+[=+]?|-[=-]?|\|[=|]?|&(?:=|&|\^=?)?|>(?:>=?|=)?|<(?:<=?|=|-)?|:=|\.\.\./,builtin:/\b(?:append|bool|byte|cap|close|complex|complex(?:64|128)|copy|delete|error|float(?:32|64)|u?int(?:8|16|32|64)?|imag|len|make|new|panic|print(?:ln)?|real|recover|rune|string|uintptr)\b/}),r.languages.insertBefore("go","string",{char:{pattern:/'(?:\\.|[^'\\\r\n]){0,10}'/,greedy:!0}}),delete r.languages.go["class-name"]}return $Ue}var FUe,THt;function AFr(){if(THt)return FUe;THt=1,FUe=n,n.displayName="graphql",n.aliases=[];function n(r){r.languages.graphql={comment:/#.*/,description:{pattern:/(?:"""(?:[^"]|(?!""")")*"""|"(?:\\.|[^\\"\r\n])*")(?=\s*[a-z_])/i,greedy:!0,alias:"string",inside:{"language-markdown":{pattern:/(^"(?:"")?)(?!\1)[\s\S]+(?=\1$)/,lookbehind:!0,inside:r.languages.markdown}}},string:{pattern:/"""(?:[^"]|(?!""")")*"""|"(?:\\.|[^\\"\r\n])*"/,greedy:!0},number:/(?:\B-|\b)\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,boolean:/\b(?:false|true)\b/,variable:/\$[a-z_]\w*/i,directive:{pattern:/@[a-z_]\w*/i,alias:"function"},"attr-name":{pattern:/\b[a-z_]\w*(?=\s*(?:\((?:[^()"]|"(?:\\.|[^\\"\r\n])*")*\))?:)/i,greedy:!0},"atom-input":{pattern:/\b[A-Z]\w*Input\b/,alias:"class-name"},scalar:/\b(?:Boolean|Float|ID|Int|String)\b/,constant:/\b[A-Z][A-Z_\d]*\b/,"class-name":{pattern:/(\b(?:enum|implements|interface|on|scalar|type|union)\s+|&\s*|:\s*|\[)[A-Z_]\w*/,lookbehind:!0},fragment:{pattern:/(\bfragment\s+|\.{3}\s*(?!on\b))[a-zA-Z_]\w*/,lookbehind:!0,alias:"function"},"definition-mutation":{pattern:/(\bmutation\s+)[a-zA-Z_]\w*/,lookbehind:!0,alias:"function"},"definition-query":{pattern:/(\bquery\s+)[a-zA-Z_]\w*/,lookbehind:!0,alias:"function"},keyword:/\b(?:directive|enum|extend|fragment|implements|input|interface|mutation|on|query|repeatable|scalar|schema|subscription|type|union)\b/,operator:/[!=|&]|\.{3}/,"property-query":/\w+(?=\s*\()/,object:/\w+(?=\s*\{)/,punctuation:/[!(){}\[\]:=,]/,property:/\w+/},r.hooks.add("after-tokenize",function(s){if(s.language!=="graphql")return;var c=s.tokens.filter(function(B){return typeof B!="string"&&B.type!=="comment"&&B.type!=="scalar"}),d=0;function h(B){return c[d+B]}function b(B,G){G=G||0;for(var K=0;K<B.length;K++){var X=h(K+G);if(!X||X.type!==B[K])return!1}return!0}function y(B,G){for(var K=1,X=d;X<c.length;X++){var ae=c[X],oe=ae.content;if(ae.type==="punctuation"&&typeof oe=="string"){if(B.test(oe))K++;else if(G.test(oe)&&(K--,K===0))return X}}return-1}function E(B,G){var K=B.alias;K?Array.isArray(K)||(B.alias=K=[K]):B.alias=K=[],K.push(G)}for(;d<c.length;){var k=c[d++];if(k.type==="keyword"&&k.content==="mutation"){var x=[];if(b(["definition-mutation","punctuation"])&&h(1).content==="("){d+=2;var O=y(/^\($/,/^\)$/);if(O===-1)continue;for(;d<O;d++){var P=h(0);P.type==="variable"&&(E(P,"variable-input"),x.push(P.content))}d=O+1}if(b(["punctuation","property-query"])&&h(0).content==="{"&&(d++,E(h(0),"property-mutation"),x.length>0)){var D=y(/^\{$/,/^\}$/);if(D===-1)continue;for(var $=d;$<D;$++){var U=c[$];U.type==="variable"&&x.indexOf(U.content)>=0&&E(U,"variable-input")}}}}})}return FUe}var BUe,xHt;function IFr(){if(xHt)return BUe;xHt=1,BUe=n,n.displayName="groovy",n.aliases=[];function n(r){r.languages.groovy=r.languages.extend("clike",{string:[{pattern:/("""|''')(?:[^\\]|\\[\s\S])*?\1|\$\/(?:[^/$]|\$(?:[/$]|(?![/$]))|\/(?!\$))*\/\$/,greedy:!0},{pattern:/(["'/])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0}],keyword:/\b(?:abstract|as|assert|boolean|break|byte|case|catch|char|class|const|continue|def|default|do|double|else|enum|extends|final|finally|float|for|goto|if|implements|import|in|instanceof|int|interface|long|native|new|package|private|protected|public|return|short|static|strictfp|super|switch|synchronized|this|throw|throws|trait|transient|try|void|volatile|while)\b/,number:/\b(?:0b[01_]+|0x[\da-f_]+(?:\.[\da-f_p\-]+)?|[\d_]+(?:\.[\d_]+)?(?:e[+-]?\d+)?)[glidf]?\b/i,operator:{pattern:/(^|[^.])(?:~|==?~?|\?[.:]?|\*(?:[.=]|\*=?)?|\.[@&]|\.\.<|\.\.(?!\.)|-[-=>]?|\+[+=]?|!=?|<(?:<=?|=>?)?|>(?:>>?=?|=)?|&[&=]?|\|[|=]?|\/=?|\^=?|%=?)/,lookbehind:!0},punctuation:/\.+|[{}[\];(),:$]/}),r.languages.insertBefore("groovy","string",{shebang:{pattern:/#!.+/,alias:"comment"}}),r.languages.insertBefore("groovy","punctuation",{"spock-block":/\b(?:and|cleanup|expect|given|setup|then|when|where):/}),r.languages.insertBefore("groovy","function",{annotation:{pattern:/(^|[^.])@\w+/,lookbehind:!0,alias:"punctuation"}}),r.hooks.add("wrap",function(o){if(o.language==="groovy"&&o.type==="string"){var s=o.content.value[0];if(s!="'"){var c=/([^\\])(?:\$(?:\{.*?\}|[\w.]+))/;s==="$"&&(c=/([^\$])(?:\$(?:\{.*?\}|[\w.]+))/),o.content.value=o.content.value.replace(/&lt;/g,"<").replace(/&amp;/g,"&"),o.content=r.highlight(o.content.value,{expression:{pattern:c,lookbehind:!0,inside:r.languages.groovy}}),o.classes.push(s==="/"?"regex":"gstring")}}})}return BUe}var UUe,CHt;function OFr(){if(CHt)return UUe;CHt=1;var n=PEe();UUe=r,r.displayName="haml",r.aliases=[];function r(o){o.register(n),function(s){s.languages.haml={"multiline-comment":{pattern:/((?:^|\r?\n|\r)([\t ]*))(?:\/|-#).*(?:(?:\r?\n|\r)\2[\t ].+)*/,lookbehind:!0,alias:"comment"},"multiline-code":[{pattern:/((?:^|\r?\n|\r)([\t ]*)(?:[~-]|[&!]?=)).*,[\t ]*(?:(?:\r?\n|\r)\2[\t ].*,[\t ]*)*(?:(?:\r?\n|\r)\2[\t ].+)/,lookbehind:!0,inside:s.languages.ruby},{pattern:/((?:^|\r?\n|\r)([\t ]*)(?:[~-]|[&!]?=)).*\|[\t ]*(?:(?:\r?\n|\r)\2[\t ].*\|[\t ]*)*/,lookbehind:!0,inside:s.languages.ruby}],filter:{pattern:/((?:^|\r?\n|\r)([\t ]*)):[\w-]+(?:(?:\r?\n|\r)(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/,lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"symbol"}}},markup:{pattern:/((?:^|\r?\n|\r)[\t ]*)<.+/,lookbehind:!0,inside:s.languages.markup},doctype:{pattern:/((?:^|\r?\n|\r)[\t ]*)!!!(?: .+)?/,lookbehind:!0},tag:{pattern:/((?:^|\r?\n|\r)[\t ]*)[%.#][\w\-#.]*[\w\-](?:\([^)]+\)|\{(?:\{[^}]+\}|[^{}])+\}|\[[^\]]+\])*[\/<>]*/,lookbehind:!0,inside:{attributes:[{pattern:/(^|[^#])\{(?:\{[^}]+\}|[^{}])+\}/,lookbehind:!0,inside:s.languages.ruby},{pattern:/\([^)]+\)/,inside:{"attr-value":{pattern:/(=\s*)(?:"(?:\\.|[^\\"\r\n])*"|[^)\s]+)/,lookbehind:!0},"attr-name":/[\w:-]+(?=\s*!?=|\s*[,)])/,punctuation:/[=(),]/}},{pattern:/\[[^\]]+\]/,inside:s.languages.ruby}],punctuation:/[<>]/}},code:{pattern:/((?:^|\r?\n|\r)[\t ]*(?:[~-]|[&!]?=)).+/,lookbehind:!0,inside:s.languages.ruby},interpolation:{pattern:/#\{[^}]+\}/,inside:{delimiter:{pattern:/^#\{|\}$/,alias:"punctuation"},ruby:{pattern:/[\s\S]+/,inside:s.languages.ruby}}},punctuation:{pattern:/((?:^|\r?\n|\r)[\t ]*)[~=\-&!]+/,lookbehind:!0}};for(var c="((?:^|\\r?\\n|\\r)([\\t ]*)):{{filter_name}}(?:(?:\\r?\\n|\\r)(?:\\2[\\t ].+|\\s*?(?=\\r?\\n|\\r)))+",d=["css",{filter:"coffee",language:"coffeescript"},"erb","javascript","less","markdown","ruby","scss","textile"],h={},b=0,y=d.length;b<y;b++){var E=d[b];E=typeof E=="string"?{filter:E,language:E}:E,s.languages[E.language]&&(h["filter-"+E.filter]={pattern:RegExp(c.replace("{{filter_name}}",function(){return E.filter})),lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"symbol"},text:{pattern:/[\s\S]+/,alias:[E.language,"language-"+E.language],inside:s.languages[E.language]}}})}s.languages.insertBefore("haml","filter",h)}(o)}return UUe}var zUe,AHt;function NFr(){if(AHt)return zUe;AHt=1;var n=oy();zUe=r,r.displayName="handlebars",r.aliases=["hbs"];function r(o){o.register(n),function(s){s.languages.handlebars={comment:/\{\{![\s\S]*?\}\}/,delimiter:{pattern:/^\{\{\{?|\}\}\}?$/,alias:"punctuation"},string:/(["'])(?:\\.|(?!\1)[^\\\r\n])*\1/,number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][+-]?\d+)?/,boolean:/\b(?:false|true)\b/,block:{pattern:/^(\s*(?:~\s*)?)[#\/]\S+?(?=\s*(?:~\s*)?$|\s)/,lookbehind:!0,alias:"keyword"},brackets:{pattern:/\[[^\]]+\]/,inside:{punctuation:/\[|\]/,variable:/[\s\S]+/}},punctuation:/[!"#%&':()*+,.\/;<=>@\[\\\]^`{|}~]/,variable:/[^!"#%&'()*+,\/;<=>@\[\\\]^`{|}~\s]+/},s.hooks.add("before-tokenize",function(c){var d=/\{\{\{[\s\S]+?\}\}\}|\{\{[\s\S]+?\}\}/g;s.languages["markup-templating"].buildPlaceholders(c,"handlebars",d)}),s.hooks.add("after-tokenize",function(c){s.languages["markup-templating"].tokenizePlaceholders(c,"handlebars")}),s.languages.hbs=s.languages.handlebars}(o)}return zUe}var HUe,IHt;function EQe(){if(IHt)return HUe;IHt=1,HUe=n,n.displayName="haskell",n.aliases=["hs"];function n(r){r.languages.haskell={comment:{pattern:/(^|[^-!#$%*+=?&@|~.:<>^\\\/])(?:--(?:(?=.)[^-!#$%*+=?&@|~.:<>^\\\/].*|$)|\{-[\s\S]*?-\})/m,lookbehind:!0},char:{pattern:/'(?:[^\\']|\\(?:[abfnrtv\\"'&]|\^[A-Z@[\]^_]|ACK|BEL|BS|CAN|CR|DC1|DC2|DC3|DC4|DEL|DLE|EM|ENQ|EOT|ESC|ETB|ETX|FF|FS|GS|HT|LF|NAK|NUL|RS|SI|SO|SOH|SP|STX|SUB|SYN|US|VT|\d+|o[0-7]+|x[0-9a-fA-F]+))'/,alias:"string"},string:{pattern:/"(?:[^\\"]|\\(?:\S|\s+\\))*"/,greedy:!0},keyword:/\b(?:case|class|data|deriving|do|else|if|in|infixl|infixr|instance|let|module|newtype|of|primitive|then|type|where)\b/,"import-statement":{pattern:/(^[\t ]*)import\s+(?:qualified\s+)?(?:[A-Z][\w']*)(?:\.[A-Z][\w']*)*(?:\s+as\s+(?:[A-Z][\w']*)(?:\.[A-Z][\w']*)*)?(?:\s+hiding\b)?/m,lookbehind:!0,inside:{keyword:/\b(?:as|hiding|import|qualified)\b/,punctuation:/\./}},builtin:/\b(?:abs|acos|acosh|all|and|any|appendFile|approxRational|asTypeOf|asin|asinh|atan|atan2|atanh|basicIORun|break|catch|ceiling|chr|compare|concat|concatMap|const|cos|cosh|curry|cycle|decodeFloat|denominator|digitToInt|div|divMod|drop|dropWhile|either|elem|encodeFloat|enumFrom|enumFromThen|enumFromThenTo|enumFromTo|error|even|exp|exponent|fail|filter|flip|floatDigits|floatRadix|floatRange|floor|fmap|foldl|foldl1|foldr|foldr1|fromDouble|fromEnum|fromInt|fromInteger|fromIntegral|fromRational|fst|gcd|getChar|getContents|getLine|group|head|id|inRange|index|init|intToDigit|interact|ioError|isAlpha|isAlphaNum|isAscii|isControl|isDenormalized|isDigit|isHexDigit|isIEEE|isInfinite|isLower|isNaN|isNegativeZero|isOctDigit|isPrint|isSpace|isUpper|iterate|last|lcm|length|lex|lexDigits|lexLitChar|lines|log|logBase|lookup|map|mapM|mapM_|max|maxBound|maximum|maybe|min|minBound|minimum|mod|negate|not|notElem|null|numerator|odd|or|ord|otherwise|pack|pi|pred|primExitWith|print|product|properFraction|putChar|putStr|putStrLn|quot|quotRem|range|rangeSize|read|readDec|readFile|readFloat|readHex|readIO|readInt|readList|readLitChar|readLn|readOct|readParen|readSigned|reads|readsPrec|realToFrac|recip|rem|repeat|replicate|return|reverse|round|scaleFloat|scanl|scanl1|scanr|scanr1|seq|sequence|sequence_|show|showChar|showInt|showList|showLitChar|showParen|showSigned|showString|shows|showsPrec|significand|signum|sin|sinh|snd|sort|span|splitAt|sqrt|subtract|succ|sum|tail|take|takeWhile|tan|tanh|threadToIOResult|toEnum|toInt|toInteger|toLower|toRational|toUpper|truncate|uncurry|undefined|unlines|until|unwords|unzip|unzip3|userError|words|writeFile|zip|zip3|zipWith|zipWith3)\b/,number:/\b(?:\d+(?:\.\d+)?(?:e[+-]?\d+)?|0o[0-7]+|0x[0-9a-f]+)\b/i,operator:[{pattern:/`(?:[A-Z][\w']*\.)*[_a-z][\w']*`/,greedy:!0},{pattern:/(\s)\.(?=\s)/,lookbehind:!0},/[-!#$%*+=?&@|~:<>^\\\/][-!#$%*+=?&@|~.:<>^\\\/]*|\.[-!#$%*+=?&@|~.:<>^\\\/]+/],hvariable:{pattern:/\b(?:[A-Z][\w']*\.)*[_a-z][\w']*/,inside:{punctuation:/\./}},constant:{pattern:/\b(?:[A-Z][\w']*\.)*[A-Z][\w']*/,inside:{punctuation:/\./}},punctuation:/[{}[\];(),.:]/},r.languages.hs=r.languages.haskell}return HUe}var GUe,OHt;function RFr(){if(OHt)return GUe;OHt=1,GUe=n,n.displayName="haxe",n.aliases=[];function n(r){r.languages.haxe=r.languages.extend("clike",{string:{pattern:/"(?:[^"\\]|\\[\s\S])*"/,greedy:!0},"class-name":[{pattern:/(\b(?:abstract|class|enum|extends|implements|interface|new|typedef)\s+)[A-Z_]\w*/,lookbehind:!0},/\b[A-Z]\w*/],keyword:/\bthis\b|\b(?:abstract|as|break|case|cast|catch|class|continue|default|do|dynamic|else|enum|extends|extern|final|for|from|function|if|implements|import|in|inline|interface|macro|new|null|operator|overload|override|package|private|public|return|static|super|switch|throw|to|try|typedef|untyped|using|var|while)(?!\.)\b/,function:{pattern:/\b[a-z_]\w*(?=\s*(?:<[^<>]*>\s*)?\()/i,greedy:!0},operator:/\.{3}|\+\+|--|&&|\|\||->|=>|(?:<<?|>{1,3}|[-+*/%!=&|^])=?|[?:~]/}),r.languages.insertBefore("haxe","string",{"string-interpolation":{pattern:/'(?:[^'\\]|\\[\s\S])*'/,greedy:!0,inside:{interpolation:{pattern:/(^|[^\\])\$(?:\w+|\{[^{}]+\})/,lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^\$\{?|\}$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:r.languages.haxe}}},string:/[\s\S]+/}}}),r.languages.insertBefore("haxe","class-name",{regex:{pattern:/~\/(?:[^\/\\\r\n]|\\.)+\/[a-z]*/,greedy:!0,inside:{"regex-flags":/\b[a-z]+$/,"regex-source":{pattern:/^(~\/)[\s\S]+(?=\/$)/,lookbehind:!0,alias:"language-regex",inside:r.languages.regex},"regex-delimiter":/^~\/|\/$/}}}),r.languages.insertBefore("haxe","keyword",{preprocessor:{pattern:/#(?:else|elseif|end|if)\b.*/,alias:"property"},metadata:{pattern:/@:?[\w.]+/,alias:"symbol"},reification:{pattern:/\$(?:\w+|(?=\{))/,alias:"important"}})}return GUe}var qUe,NHt;function PFr(){if(NHt)return qUe;NHt=1,qUe=n,n.displayName="hcl",n.aliases=[];function n(r){r.languages.hcl={comment:/(?:\/\/|#).*|\/\*[\s\S]*?(?:\*\/|$)/,heredoc:{pattern:/<<-?(\w+\b)[\s\S]*?^[ \t]*\1/m,greedy:!0,alias:"string"},keyword:[{pattern:/(?:data|resource)\s+(?:"(?:\\[\s\S]|[^\\"])*")(?=\s+"[\w-]+"\s+\{)/i,inside:{type:{pattern:/(resource|data|\s+)(?:"(?:\\[\s\S]|[^\\"])*")/i,lookbehind:!0,alias:"variable"}}},{pattern:/(?:backend|module|output|provider|provisioner|variable)\s+(?:[\w-]+|"(?:\\[\s\S]|[^\\"])*")\s+(?=\{)/i,inside:{type:{pattern:/(backend|module|output|provider|provisioner|variable)\s+(?:[\w-]+|"(?:\\[\s\S]|[^\\"])*")\s+/i,lookbehind:!0,alias:"variable"}}},/[\w-]+(?=\s+\{)/],property:[/[-\w\.]+(?=\s*=(?!=))/,/"(?:\\[\s\S]|[^\\"])+"(?=\s*[:=])/],string:{pattern:/"(?:[^\\$"]|\\[\s\S]|\$(?:(?=")|\$+(?!\$)|[^"${])|\$\{(?:[^{}"]|"(?:[^\\"]|\\[\s\S])*")*\})*"/,greedy:!0,inside:{interpolation:{pattern:/(^|[^$])\$\{(?:[^{}"]|"(?:[^\\"]|\\[\s\S])*")*\}/,lookbehind:!0,inside:{type:{pattern:/(\b(?:count|data|local|module|path|self|terraform|var)\b\.)[\w\*]+/i,lookbehind:!0,alias:"variable"},keyword:/\b(?:count|data|local|module|path|self|terraform|var)\b/i,function:/\w+(?=\()/,string:{pattern:/"(?:\\[\s\S]|[^\\"])*"/,greedy:!0},number:/\b0x[\da-f]+\b|\b\d+(?:\.\d*)?(?:e[+-]?\d+)?/i,punctuation:/[!\$#%&'()*+,.\/;<=>@\[\\\]^`{|}~?:]/}}}},number:/\b0x[\da-f]+\b|\b\d+(?:\.\d*)?(?:e[+-]?\d+)?/i,boolean:/\b(?:false|true)\b/i,punctuation:/[=\[\]{}]/}}return qUe}var VUe,RHt;function LFr(){if(RHt)return VUe;RHt=1;var n=vU();VUe=r,r.displayName="hlsl",r.aliases=[];function r(o){o.register(n),o.languages.hlsl=o.languages.extend("c",{"class-name":[o.languages.c["class-name"],/\b(?:AppendStructuredBuffer|BlendState|Buffer|ByteAddressBuffer|CompileShader|ComputeShader|ConsumeStructuredBuffer|DepthStencilState|DepthStencilView|DomainShader|GeometryShader|Hullshader|InputPatch|LineStream|OutputPatch|PixelShader|PointStream|RWBuffer|RWByteAddressBuffer|RWStructuredBuffer|RWTexture(?:1D|1DArray|2D|2DArray|3D)|RasterizerState|RenderTargetView|SamplerComparisonState|SamplerState|StructuredBuffer|Texture(?:1D|1DArray|2D|2DArray|2DMS|2DMSArray|3D|Cube|CubeArray)|TriangleStream|VertexShader)\b/],keyword:[/\b(?:asm|asm_fragment|auto|break|case|catch|cbuffer|centroid|char|class|column_major|compile|compile_fragment|const|const_cast|continue|default|delete|discard|do|dynamic_cast|else|enum|explicit|export|extern|for|friend|fxgroup|goto|groupshared|if|in|inline|inout|interface|line|lineadj|linear|long|matrix|mutable|namespace|new|nointerpolation|noperspective|operator|out|packoffset|pass|pixelfragment|point|precise|private|protected|public|register|reinterpret_cast|return|row_major|sample|sampler|shared|short|signed|sizeof|snorm|stateblock|stateblock_state|static|static_cast|string|struct|switch|tbuffer|technique|technique10|technique11|template|texture|this|throw|triangle|triangleadj|try|typedef|typename|uniform|union|unorm|unsigned|using|vector|vertexfragment|virtual|void|volatile|while)\b/,/\b(?:bool|double|dword|float|half|int|min(?:10float|12int|16(?:float|int|uint))|uint)(?:[1-4](?:x[1-4])?)?\b/],number:/(?:(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[eE][+-]?\d+)?|\b0x[\da-fA-F]+)[fFhHlLuU]?\b/,boolean:/\b(?:false|true)\b/})}return VUe}var WUe,PHt;function DFr(){if(PHt)return WUe;PHt=1,WUe=n,n.displayName="hoon",n.aliases=[];function n(r){r.languages.hoon={comment:{pattern:/::.*/,greedy:!0},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},constant:/%(?:\.[ny]|[\w-]+)/,"class-name":/@(?:[a-z0-9-]*[a-z0-9])?|\*/i,function:/(?:\+[-+] {2})?(?:[a-z](?:[a-z0-9-]*[a-z0-9])?)/,keyword:/\.[\^\+\*=\?]|![><:\.=\?!]|=[>|:,\.\-\^<+;/~\*\?]|\?[>|:\.\-\^<\+&~=@!]|\|[\$_%:\.\-\^~\*=@\?]|\+[|\$\+\*]|:[_\-\^\+~\*]|%[_:\.\-\^\+~\*=]|\^[|:\.\-\+&~\*=\?]|\$[|_%:<>\-\^&~@=\?]|;[:<\+;\/~\*=]|~[>|\$_%<\+\/&=\?!]|--|==/}}return WUe}var KUe,LHt;function MFr(){if(LHt)return KUe;LHt=1,KUe=n,n.displayName="hpkp",n.aliases=[];function n(r){r.languages.hpkp={directive:{pattern:/\b(?:includeSubDomains|max-age|pin-sha256|preload|report-to|report-uri|strict)(?=[\s;=]|$)/i,alias:"property"},operator:/=/,punctuation:/;/}}return KUe}var YUe,DHt;function jFr(){if(DHt)return YUe;DHt=1,YUe=n,n.displayName="hsts",n.aliases=[];function n(r){r.languages.hsts={directive:{pattern:/\b(?:includeSubDomains|max-age|preload)(?=[\s;=]|$)/i,alias:"property"},operator:/=/,punctuation:/;/}}return YUe}var XUe,MHt;function $Fr(){if(MHt)return XUe;MHt=1,XUe=n,n.displayName="http",n.aliases=[];function n(r){(function(o){function s(x){return RegExp("(^(?:"+x+"):[ 	]*(?![ 	]))[^]+","i")}o.languages.http={"request-line":{pattern:/^(?:CONNECT|DELETE|GET|HEAD|OPTIONS|PATCH|POST|PRI|PUT|SEARCH|TRACE)\s(?:https?:\/\/|\/)\S*\sHTTP\/[\d.]+/m,inside:{method:{pattern:/^[A-Z]+\b/,alias:"property"},"request-target":{pattern:/^(\s)(?:https?:\/\/|\/)\S*(?=\s)/,lookbehind:!0,alias:"url",inside:o.languages.uri},"http-version":{pattern:/^(\s)HTTP\/[\d.]+/,lookbehind:!0,alias:"property"}}},"response-status":{pattern:/^HTTP\/[\d.]+ \d+ .+/m,inside:{"http-version":{pattern:/^HTTP\/[\d.]+/,alias:"property"},"status-code":{pattern:/^(\s)\d+(?=\s)/,lookbehind:!0,alias:"number"},"reason-phrase":{pattern:/^(\s).+/,lookbehind:!0,alias:"string"}}},header:{pattern:/^[\w-]+:.+(?:(?:\r\n?|\n)[ \t].+)*/m,inside:{"header-value":[{pattern:s(/Content-Security-Policy/.source),lookbehind:!0,alias:["csp","languages-csp"],inside:o.languages.csp},{pattern:s(/Public-Key-Pins(?:-Report-Only)?/.source),lookbehind:!0,alias:["hpkp","languages-hpkp"],inside:o.languages.hpkp},{pattern:s(/Strict-Transport-Security/.source),lookbehind:!0,alias:["hsts","languages-hsts"],inside:o.languages.hsts},{pattern:s(/[^:]+/.source),lookbehind:!0}],"header-name":{pattern:/^[^:]+/,alias:"keyword"},punctuation:/^:/}}};var c=o.languages,d={"application/javascript":c.javascript,"application/json":c.json||c.javascript,"application/xml":c.xml,"text/xml":c.xml,"text/html":c.html,"text/css":c.css,"text/plain":c.plain},h={"application/json":!0,"application/xml":!0};function b(x){var O=x.replace(/^[a-z]+\//,""),P="\\w+/(?:[\\w.-]+\\+)+"+O+"(?![+\\w.-])";return"(?:"+x+"|"+P+")"}var y;for(var E in d)if(d[E]){y=y||{};var k=h[E]?b(E):E;y[E.replace(/\//g,"-")]={pattern:RegExp("("+/content-type:\s*/.source+k+/(?:(?:\r\n?|\n)[\w-].*)*(?:\r(?:\n|(?!\n))|\n)/.source+")"+/[^ \t\w-][\s\S]*/.source,"i"),lookbehind:!0,inside:d[E]}}y&&o.languages.insertBefore("http","header",y)})(r)}return XUe}var JUe,jHt;function FFr(){if(jHt)return JUe;jHt=1,JUe=n,n.displayName="ichigojam",n.aliases=[];function n(r){r.languages.ichigojam={comment:/(?:\B'|REM)(?:[^\n\r]*)/i,string:{pattern:/"(?:""|[!#$%&'()*,\/:;<=>?^\w +\-.])*"/,greedy:!0},number:/\B#[0-9A-F]+|\B`[01]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,keyword:/\b(?:BEEP|BPS|CASE|CLEAR|CLK|CLO|CLP|CLS|CLT|CLV|CONT|COPY|ELSE|END|FILE|FILES|FOR|GOSUB|GOTO|GSB|IF|INPUT|KBD|LED|LET|LIST|LOAD|LOCATE|LRUN|NEW|NEXT|OUT|PLAY|POKE|PRINT|PWM|REM|RENUM|RESET|RETURN|RIGHT|RTN|RUN|SAVE|SCROLL|SLEEP|SRND|STEP|STOP|SUB|TEMPO|THEN|TO|UART|VIDEO|WAIT)(?:\$|\b)/i,function:/\b(?:ABS|ANA|ASC|BIN|BTN|DEC|END|FREE|HELP|HEX|I2CR|I2CW|IN|INKEY|LEN|LINE|PEEK|RND|SCR|SOUND|STR|TICK|USR|VER|VPEEK|ZER)(?:\$|\b)/i,label:/(?:\B@\S+)/,operator:/<[=>]?|>=?|\|\||&&|[+\-*\/=|&^~!]|\b(?:AND|NOT|OR)\b/i,punctuation:/[\[,;:()\]]/}}return JUe}var ZUe,$Ht;function BFr(){if($Ht)return ZUe;$Ht=1,ZUe=n,n.displayName="icon",n.aliases=[];function n(r){r.languages.icon={comment:/#.*/,string:{pattern:/(["'])(?:(?!\1)[^\\\r\n_]|\\.|_(?!\1)(?:\r\n|[\s\S]))*\1/,greedy:!0},number:/\b(?:\d+r[a-z\d]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b|\.\d+\b/i,"builtin-keyword":{pattern:/&(?:allocated|ascii|clock|collections|cset|current|date|dateline|digits|dump|e|error(?:number|text|value)?|errout|fail|features|file|host|input|lcase|letters|level|line|main|null|output|phi|pi|pos|progname|random|regions|source|storage|subject|time|trace|ucase|version)\b/,alias:"variable"},directive:{pattern:/\$\w+/,alias:"builtin"},keyword:/\b(?:break|by|case|create|default|do|else|end|every|fail|global|if|initial|invocable|link|local|next|not|of|procedure|record|repeat|return|static|suspend|then|to|until|while)\b/,function:/\b(?!\d)\w+(?=\s*[({]|\s*!\s*\[)/,operator:/[+-]:(?!=)|(?:[\/?@^%&]|\+\+?|--?|==?=?|~==?=?|\*\*?|\|\|\|?|<(?:->?|<?=?)|>>?=?)(?::=)?|:(?:=:?)?|[!.\\|~]/,punctuation:/[\[\](){},;]/}}return ZUe}var QUe,FHt;function UFr(){if(FHt)return QUe;FHt=1,QUe=n,n.displayName="icuMessageFormat",n.aliases=[];function n(r){(function(o){function s(E,k){return k<=0?/[]/.source:E.replace(/<SELF>/g,function(){return s(E,k-1)})}var c=/'[{}:=,](?:[^']|'')*'(?!')/,d={pattern:/''/,greedy:!0,alias:"operator"},h={pattern:c,greedy:!0,inside:{escape:d}},b=s(/\{(?:[^{}']|'(?![{},'])|''|<STR>|<SELF>)*\}/.source.replace(/<STR>/g,function(){return c.source}),8),y={pattern:RegExp(b),inside:{message:{pattern:/^(\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:null},"message-delimiter":{pattern:/./,alias:"punctuation"}}};o.languages["icu-message-format"]={argument:{pattern:RegExp(b),greedy:!0,inside:{content:{pattern:/^(\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:{"argument-name":{pattern:/^(\s*)[^{}:=,\s]+/,lookbehind:!0},"choice-style":{pattern:/^(\s*,\s*choice\s*,\s*)\S(?:[\s\S]*\S)?/,lookbehind:!0,inside:{punctuation:/\|/,range:{pattern:/^(\s*)[+-]?(?:\d+(?:\.\d*)?|\u221e)\s*[<#\u2264]/,lookbehind:!0,inside:{operator:/[<#\u2264]/,number:/\S+/}},rest:null}},"plural-style":{pattern:/^(\s*,\s*(?:plural|selectordinal)\s*,\s*)\S(?:[\s\S]*\S)?/,lookbehind:!0,inside:{offset:/^offset:\s*\d+/,"nested-message":y,selector:{pattern:/=\d+|[^{}:=,\s]+/,inside:{keyword:/^(?:few|many|one|other|two|zero)$/}}}},"select-style":{pattern:/^(\s*,\s*select\s*,\s*)\S(?:[\s\S]*\S)?/,lookbehind:!0,inside:{"nested-message":y,selector:{pattern:/[^{}:=,\s]+/,inside:{keyword:/^other$/}}}},keyword:/\b(?:choice|plural|select|selectordinal)\b/,"arg-type":{pattern:/\b(?:date|duration|number|ordinal|spellout|time)\b/,alias:"keyword"},"arg-skeleton":{pattern:/(,\s*)::[^{}:=,\s]+/,lookbehind:!0},"arg-style":{pattern:/(,\s*)(?:currency|full|integer|long|medium|percent|short)(?=\s*$)/,lookbehind:!0},"arg-style-text":{pattern:RegExp(/(^\s*,\s*(?=\S))/.source+s(/(?:[^{}']|'[^']*'|\{(?:<SELF>)?\})+/.source,8)+"$"),lookbehind:!0,alias:"string"},punctuation:/,/}},"argument-delimiter":{pattern:/./,alias:"operator"}}},escape:d,string:h},y.inside.message.inside=o.languages["icu-message-format"],o.languages["icu-message-format"].argument.inside.content.inside["choice-style"].inside.rest=o.languages["icu-message-format"]})(r)}return QUe}var eze,BHt;function zFr(){if(BHt)return eze;BHt=1;var n=EQe();eze=r,r.displayName="idris",r.aliases=["idr"];function r(o){o.register(n),o.languages.idris=o.languages.extend("haskell",{comment:{pattern:/(?:(?:--|\|\|\|).*$|\{-[\s\S]*?-\})/m},keyword:/\b(?:Type|case|class|codata|constructor|corecord|data|do|dsl|else|export|if|implementation|implicit|import|impossible|in|infix|infixl|infixr|instance|interface|let|module|mutual|namespace|of|parameters|partial|postulate|private|proof|public|quoteGoal|record|rewrite|syntax|then|total|using|where|with)\b/,builtin:void 0}),o.languages.insertBefore("idris","keyword",{"import-statement":{pattern:/(^\s*import\s+)(?:[A-Z][\w']*)(?:\.[A-Z][\w']*)*/m,lookbehind:!0,inside:{punctuation:/\./}}}),o.languages.idr=o.languages.idris}return eze}var tze,UHt;function HFr(){if(UHt)return tze;UHt=1,tze=n,n.displayName="iecst",n.aliases=[];function n(r){r.languages.iecst={comment:[{pattern:/(^|[^\\])(?:\/\*[\s\S]*?(?:\*\/|$)|\(\*[\s\S]*?(?:\*\)|$)|\{[\s\S]*?(?:\}|$))/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:[/\b(?:END_)?(?:PROGRAM|CONFIGURATION|INTERFACE|FUNCTION_BLOCK|FUNCTION|ACTION|TRANSITION|TYPE|STRUCT|(?:INITIAL_)?STEP|NAMESPACE|LIBRARY|CHANNEL|FOLDER|RESOURCE|VAR_(?:ACCESS|CONFIG|EXTERNAL|GLOBAL|INPUT|IN_OUT|OUTPUT|TEMP)|VAR|METHOD|PROPERTY)\b/i,/\b(?:AT|BY|(?:END_)?(?:CASE|FOR|IF|REPEAT|WHILE)|CONSTANT|CONTINUE|DO|ELSE|ELSIF|EXIT|EXTENDS|FROM|GET|GOTO|IMPLEMENTS|JMP|NON_RETAIN|OF|PRIVATE|PROTECTED|PUBLIC|RETAIN|RETURN|SET|TASK|THEN|TO|UNTIL|USING|WITH|__CATCH|__ENDTRY|__FINALLY|__TRY)\b/],"class-name":/\b(?:ANY|ARRAY|BOOL|BYTE|U?(?:D|L|S)?INT|(?:D|L)?WORD|DATE(?:_AND_TIME)?|DT|L?REAL|POINTER|STRING|TIME(?:_OF_DAY)?|TOD)\b/,address:{pattern:/%[IQM][XBWDL][\d.]*|%[IQ][\d.]*/,alias:"symbol"},number:/\b(?:16#[\da-f]+|2#[01_]+|0x[\da-f]+)\b|\b(?:D|DT|T|TOD)#[\d_shmd:]*|\b[A-Z]*#[\d.,_]*|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,boolean:/\b(?:FALSE|NULL|TRUE)\b/,operator:/S?R?:?=>?|&&?|\*\*?|<[=>]?|>=?|[-:^/+#]|\b(?:AND|EQ|EXPT|GE|GT|LE|LT|MOD|NE|NOT|OR|XOR)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,punctuation:/[()[\].,;]/}}return tze}var nze,zHt;function GFr(){if(zHt)return nze;zHt=1,nze=n,n.displayName="ignore",n.aliases=["gitignore","hgignore","npmignore"];function n(r){(function(o){o.languages.ignore={comment:/^#.*/m,entry:{pattern:/\S(?:.*(?:(?:\\ )|\S))?/,alias:"string",inside:{operator:/^!|\*\*?|\?/,regex:{pattern:/(^|[^\\])\[[^\[\]]*\]/,lookbehind:!0},punctuation:/\//}}},o.languages.gitignore=o.languages.ignore,o.languages.hgignore=o.languages.ignore,o.languages.npmignore=o.languages.ignore})(r)}return nze}var rze,HHt;function qFr(){if(HHt)return rze;HHt=1,rze=n,n.displayName="inform7",n.aliases=[];function n(r){r.languages.inform7={string:{pattern:/"[^"]*"/,inside:{substitution:{pattern:/\[[^\[\]]+\]/,inside:{delimiter:{pattern:/\[|\]/,alias:"punctuation"}}}}},comment:{pattern:/\[[^\[\]]+\]/,greedy:!0},title:{pattern:/^[ \t]*(?:book|chapter|part(?! of)|section|table|volume)\b.+/im,alias:"important"},number:{pattern:/(^|[^-])(?:\b\d+(?:\.\d+)?(?:\^\d+)?(?:(?!\d)\w+)?|\b(?:eight|eleven|five|four|nine|one|seven|six|ten|three|twelve|two))\b(?!-)/i,lookbehind:!0},verb:{pattern:/(^|[^-])\b(?:answering|applying to|are|asking|attacking|be(?:ing)?|burning|buying|called|carries|carry(?! out)|carrying|climbing|closing|conceal(?:ing|s)?|consulting|contain(?:ing|s)?|cutting|drinking|dropping|eating|enclos(?:es?|ing)|entering|examining|exiting|getting|giving|going|ha(?:s|ve|ving)|hold(?:ing|s)?|impl(?:ies|y)|incorporat(?:es?|ing)|inserting|is|jumping|kissing|listening|locking|looking|mean(?:ing|s)?|opening|provid(?:es?|ing)|pulling|pushing|putting|relat(?:es?|ing)|removing|searching|see(?:ing|s)?|setting|showing|singing|sleeping|smelling|squeezing|support(?:ing|s)?|swearing|switching|taking|tasting|telling|thinking|throwing|touching|turning|tying|unlock(?:ing|s)?|var(?:ies|y|ying)|waiting|waking|waving|wear(?:ing|s)?)\b(?!-)/i,lookbehind:!0,alias:"operator"},keyword:{pattern:/(^|[^-])\b(?:after|before|carry out|check|continue the action|definition(?= *:)|do nothing|else|end (?:if|the story|unless)|every turn|if|include|instead(?: of)?|let|move|no|now|otherwise|repeat|report|resume the story|rule for|running through|say(?:ing)?|stop the action|test|try(?:ing)?|understand|unless|use|when|while|yes)\b(?!-)/i,lookbehind:!0},property:{pattern:/(^|[^-])\b(?:adjacent(?! to)|carried|closed|concealed|contained|dark|described|edible|empty|enclosed|enterable|even|female|fixed in place|full|handled|held|improper-named|incorporated|inedible|invisible|lighted|lit|lock(?:able|ed)|male|marked for listing|mentioned|negative|neuter|non-(?:empty|full|recurring)|odd|opaque|open(?:able)?|plural-named|portable|positive|privately-named|proper-named|provided|publically-named|pushable between rooms|recurring|related|rubbing|scenery|seen|singular-named|supported|swinging|switch(?:able|ed(?: off| on)?)|touch(?:able|ed)|transparent|unconcealed|undescribed|unlit|unlocked|unmarked for listing|unmentioned|unopenable|untouchable|unvisited|variable|visible|visited|wearable|worn)\b(?!-)/i,lookbehind:!0,alias:"symbol"},position:{pattern:/(^|[^-])\b(?:above|adjacent to|back side of|below|between|down|east|everywhere|front side|here|in|inside(?: from)?|north(?:east|west)?|nowhere|on(?: top of)?|other side|outside(?: from)?|parts? of|regionally in|south(?:east|west)?|through|up|west|within)\b(?!-)/i,lookbehind:!0,alias:"keyword"},type:{pattern:/(^|[^-])\b(?:actions?|activit(?:ies|y)|actors?|animals?|backdrops?|containers?|devices?|directions?|doors?|holders?|kinds?|lists?|m[ae]n|nobody|nothing|nouns?|numbers?|objects?|people|persons?|player(?:'s holdall)?|regions?|relations?|rooms?|rule(?:book)?s?|scenes?|someone|something|supporters?|tables?|texts?|things?|time|vehicles?|wom[ae]n)\b(?!-)/i,lookbehind:!0,alias:"variable"},punctuation:/[.,:;(){}]/},r.languages.inform7.string.inside.substitution.inside.rest=r.languages.inform7,r.languages.inform7.string.inside.substitution.inside.rest.text={pattern:/\S(?:\s*\S)*/,alias:"comment"}}return rze}var ize,GHt;function VFr(){if(GHt)return ize;GHt=1,ize=n,n.displayName="ini",n.aliases=[];function n(r){r.languages.ini={comment:{pattern:/(^[ \f\t\v]*)[#;][^\n\r]*/m,lookbehind:!0},section:{pattern:/(^[ \f\t\v]*)\[[^\n\r\]]*\]?/m,lookbehind:!0,inside:{"section-name":{pattern:/(^\[[ \f\t\v]*)[^ \f\t\v\]]+(?:[ \f\t\v]+[^ \f\t\v\]]+)*/,lookbehind:!0,alias:"selector"},punctuation:/\[|\]/}},key:{pattern:/(^[ \f\t\v]*)[^ \f\n\r\t\v=]+(?:[ \f\t\v]+[^ \f\n\r\t\v=]+)*(?=[ \f\t\v]*=)/m,lookbehind:!0,alias:"attr-name"},value:{pattern:/(=[ \f\t\v]*)[^ \f\n\r\t\v]+(?:[ \f\t\v]+[^ \f\n\r\t\v]+)*/,lookbehind:!0,alias:"attr-value",inside:{"inner-value":{pattern:/^("|').+(?=\1$)/,lookbehind:!0}}},punctuation:/=/}}return ize}var oze,qHt;function WFr(){if(qHt)return oze;qHt=1,oze=n,n.displayName="io",n.aliases=[];function n(r){r.languages.io={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?(?:\*\/|$)|\/\/.*|#.*)/,lookbehind:!0,greedy:!0},"triple-quoted-string":{pattern:/"""(?:\\[\s\S]|(?!""")[^\\])*"""/,greedy:!0,alias:"string"},string:{pattern:/"(?:\\.|[^\\\r\n"])*"/,greedy:!0},keyword:/\b(?:activate|activeCoroCount|asString|block|break|call|catch|clone|collectGarbage|compileString|continue|do|doFile|doMessage|doString|else|elseif|exit|for|foreach|forward|getEnvironmentVariable|getSlot|hasSlot|if|ifFalse|ifNil|ifNilEval|ifTrue|isActive|isNil|isResumable|list|message|method|parent|pass|pause|perform|performWithArgList|print|println|proto|raise|raiseResumable|removeSlot|resend|resume|schedulerSleepSeconds|self|sender|setSchedulerSleepSeconds|setSlot|shallowCopy|slotNames|super|system|then|thisBlock|thisContext|try|type|uniqueId|updateSlot|wait|while|write|yield)\b/,builtin:/\b(?:Array|AudioDevice|AudioMixer|BigNum|Block|Box|Buffer|CFunction|CGI|Color|Curses|DBM|DNSResolver|DOConnection|DOProxy|DOServer|Date|Directory|Duration|DynLib|Error|Exception|FFT|File|Fnmatch|Font|Future|GL|GLE|GLScissor|GLU|GLUCylinder|GLUQuadric|GLUSphere|GLUT|Host|Image|Importer|LinkList|List|Lobby|Locals|MD5|MP3Decoder|MP3Encoder|Map|Message|Movie|Notification|Number|Object|OpenGL|Point|Protos|Random|Regex|SGML|SGMLElement|SGMLParser|SQLite|Sequence|Server|ShowMessage|SleepyCat|SleepyCatCursor|Socket|SocketManager|Sound|Soup|Store|String|Tree|UDPSender|UPDReceiver|URL|User|Warning|WeakLink)\b/,boolean:/\b(?:false|nil|true)\b/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e-?\d+)?/i,operator:/[=!*/%+\-^&|]=|>>?=?|<<?=?|:?:?=|\+\+?|--?|\*\*?|\/\/?|%|\|\|?|&&?|\b(?:and|not|or|return)\b|@@?|\?\??|\.\./,punctuation:/[{}[\];(),.:]/}}return oze}var aze,VHt;function KFr(){if(VHt)return aze;VHt=1,aze=n,n.displayName="j",n.aliases=[];function n(r){r.languages.j={comment:{pattern:/\bNB\..*/,greedy:!0},string:{pattern:/'(?:''|[^'\r\n])*'/,greedy:!0},keyword:/\b(?:(?:CR|LF|adverb|conjunction|def|define|dyad|monad|noun|verb)\b|(?:assert|break|case|catch[dt]?|continue|do|else|elseif|end|fcase|for|for_\w+|goto_\w+|if|label_\w+|return|select|throw|try|while|whilst)\.)/,verb:{pattern:/(?!\^:|;\.|[=!][.:])(?:\{(?:\.|::?)?|p(?:\.\.?|:)|[=!\]]|[<>+*\-%$|,#][.:]?|[?^]\.?|[;\[]:?|[~}"i][.:]|[ACeEIjLor]\.|(?:[_\/\\qsux]|_?\d):)/,alias:"keyword"},number:/\b_?(?:(?!\d:)\d+(?:\.\d+)?(?:(?:ad|ar|[ejpx])_?\d+(?:\.\d+)?)*(?:b_?[\da-z]+(?:\.[\da-z]+)?)?|_\b(?!\.))/,adverb:{pattern:/[~}]|[\/\\]\.?|[bfM]\.|t[.:]/,alias:"builtin"},operator:/[=a][.:]|_\./,conjunction:{pattern:/&(?:\.:?|:)?|[.:@][.:]?|[!D][.:]|[;dHT]\.|`:?|[\^LS]:|"/,alias:"variable"},punctuation:/[()]/}}return aze}var sze,WHt;function SQe(){if(WHt)return sze;WHt=1,sze=n,n.displayName="java",n.aliases=[];function n(r){(function(o){var s=/\b(?:abstract|assert|boolean|break|byte|case|catch|char|class|const|continue|default|do|double|else|enum|exports|extends|final|finally|float|for|goto|if|implements|import|instanceof|int|interface|long|module|native|new|non-sealed|null|open|opens|package|permits|private|protected|provides|public|record|requires|return|sealed|short|static|strictfp|super|switch|synchronized|this|throw|throws|to|transient|transitive|try|uses|var|void|volatile|while|with|yield)\b/,c=/(^|[^\w.])(?:[a-z]\w*\s*\.\s*)*(?:[A-Z]\w*\s*\.\s*)*/.source,d={pattern:RegExp(c+/[A-Z](?:[\d_A-Z]*[a-z]\w*)?\b/.source),lookbehind:!0,inside:{namespace:{pattern:/^[a-z]\w*(?:\s*\.\s*[a-z]\w*)*(?:\s*\.)?/,inside:{punctuation:/\./}},punctuation:/\./}};o.languages.java=o.languages.extend("clike",{string:{pattern:/(^|[^\\])"(?:\\.|[^"\\\r\n])*"/,lookbehind:!0,greedy:!0},"class-name":[d,{pattern:RegExp(c+/[A-Z]\w*(?=\s+\w+\s*[;,=()])/.source),lookbehind:!0,inside:d.inside}],keyword:s,function:[o.languages.clike.function,{pattern:/(::\s*)[a-z_]\w*/,lookbehind:!0}],number:/\b0b[01][01_]*L?\b|\b0x(?:\.[\da-f_p+-]+|[\da-f_]+(?:\.[\da-f_p+-]+)?)\b|(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.\d[\d_]*)(?:e[+-]?\d[\d_]*)?[dfl]?/i,operator:{pattern:/(^|[^.])(?:<<=?|>>>?=?|->|--|\+\+|&&|\|\||::|[?:~]|[-+*/%&|^!=<>]=?)/m,lookbehind:!0}}),o.languages.insertBefore("java","string",{"triple-quoted-string":{pattern:/"""[ \t]*[\r\n](?:(?:"|"")?(?:\\.|[^"\\]))*"""/,greedy:!0,alias:"string"},char:{pattern:/'(?:\\.|[^'\\\r\n]){1,6}'/,greedy:!0}}),o.languages.insertBefore("java","class-name",{annotation:{pattern:/(^|[^.])@\w+(?:\s*\.\s*\w+)*/,lookbehind:!0,alias:"punctuation"},generics:{pattern:/<(?:[\w\s,.?]|&(?!&)|<(?:[\w\s,.?]|&(?!&)|<(?:[\w\s,.?]|&(?!&)|<(?:[\w\s,.?]|&(?!&))*>)*>)*>)*>/,inside:{"class-name":d,keyword:s,punctuation:/[<>(),.:]/,operator:/[?&|]/}},namespace:{pattern:RegExp(/(\b(?:exports|import(?:\s+static)?|module|open|opens|package|provides|requires|to|transitive|uses|with)\s+)(?!<keyword>)[a-z]\w*(?:\.[a-z]\w*)*\.?/.source.replace(/<keyword>/g,function(){return s.source})),lookbehind:!0,inside:{punctuation:/\./}}})})(r)}return sze}var cze,KHt;function LEe(){if(KHt)return cze;KHt=1,cze=n,n.displayName="javadoclike",n.aliases=[];function n(r){(function(o){var s=o.languages.javadoclike={parameter:{pattern:/(^[\t ]*(?:\/{3}|\*|\/\*\*)\s*@(?:arg|arguments|param)\s+)\w+/m,lookbehind:!0},keyword:{pattern:/(^[\t ]*(?:\/{3}|\*|\/\*\*)\s*|\{)@[a-z][a-zA-Z-]+\b/m,lookbehind:!0},punctuation:/[{}]/};function c(h,b){var y="doc-comment",E=o.languages[h];if(E){var k=E[y];if(!k){var x={};x[y]={pattern:/(^|[^\\])\/\*\*[^/][\s\S]*?(?:\*\/|$)/,lookbehind:!0,alias:"comment"},E=o.languages.insertBefore(h,"comment",x),k=E[y]}if(k instanceof RegExp&&(k=E[y]={pattern:k}),Array.isArray(k))for(var O=0,P=k.length;O<P;O++)k[O]instanceof RegExp&&(k[O]={pattern:k[O]}),b(k[O]);else b(k)}}function d(h,b){typeof h=="string"&&(h=[h]),h.forEach(function(y){c(y,function(E){E.inside||(E.inside={}),E.inside.rest=b})})}Object.defineProperty(s,"addSupport",{value:d}),s.addSupport(["java","javascript","php"],s)})(r)}return cze}var uze,YHt;function YFr(){if(YHt)return uze;YHt=1;var n=SQe(),r=LEe();uze=o,o.displayName="javadoc",o.aliases=[];function o(s){s.register(n),s.register(r),function(c){var d=/(^(?:[\t ]*(?:\*\s*)*))[^*\s].*$/m,h=/#\s*\w+(?:\s*\([^()]*\))?/.source,b=/(?:\b[a-zA-Z]\w+\s*\.\s*)*\b[A-Z]\w*(?:\s*<mem>)?|<mem>/.source.replace(/<mem>/g,function(){return h});c.languages.javadoc=c.languages.extend("javadoclike",{}),c.languages.insertBefore("javadoc","keyword",{reference:{pattern:RegExp(/(@(?:exception|link|linkplain|see|throws|value)\s+(?:\*\s*)?)/.source+"(?:"+b+")"),lookbehind:!0,inside:{function:{pattern:/(#\s*)\w+(?=\s*\()/,lookbehind:!0},field:{pattern:/(#\s*)\w+/,lookbehind:!0},namespace:{pattern:/\b(?:[a-z]\w*\s*\.\s*)+/,inside:{punctuation:/\./}},"class-name":/\b[A-Z]\w*/,keyword:c.languages.java.keyword,punctuation:/[#()[\],.]/}},"class-name":{pattern:/(@param\s+)<[A-Z]\w*>/,lookbehind:!0,inside:{punctuation:/[.<>]/}},"code-section":[{pattern:/(\{@code\s+(?!\s))(?:[^\s{}]|\s+(?![\s}])|\{(?:[^{}]|\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\})*\})+(?=\s*\})/,lookbehind:!0,inside:{code:{pattern:d,lookbehind:!0,inside:c.languages.java,alias:"language-java"}}},{pattern:/(<(code|pre|tt)>(?!<code>)\s*)\S(?:\S|\s+\S)*?(?=\s*<\/\2>)/,lookbehind:!0,inside:{line:{pattern:d,lookbehind:!0,inside:{tag:c.languages.markup.tag,entity:c.languages.markup.entity,code:{pattern:/.+/,inside:c.languages.java,alias:"language-java"}}}}}],tag:c.languages.markup.tag,entity:c.languages.markup.entity}),c.languages.javadoclike.addSupport("java",c.languages.javadoc)}(s)}return uze}var lze,XHt;function XFr(){if(XHt)return lze;XHt=1,lze=n,n.displayName="javastacktrace",n.aliases=[];function n(r){r.languages.javastacktrace={summary:{pattern:/^([\t ]*)(?:(?:Caused by:|Suppressed:|Exception in thread "[^"]*")[\t ]+)?[\w$.]+(?::.*)?$/m,lookbehind:!0,inside:{keyword:{pattern:/^([\t ]*)(?:(?:Caused by|Suppressed)(?=:)|Exception in thread)/m,lookbehind:!0},string:{pattern:/^(\s*)"[^"]*"/,lookbehind:!0},exceptions:{pattern:/^(:?\s*)[\w$.]+(?=:|$)/,lookbehind:!0,inside:{"class-name":/[\w$]+$/,namespace:/\b[a-z]\w*\b/,punctuation:/\./}},message:{pattern:/(:\s*)\S.*/,lookbehind:!0,alias:"string"},punctuation:/:/}},"stack-frame":{pattern:/^([\t ]*)at (?:[\w$./]|@[\w$.+-]*\/)+(?:<init>)?\([^()]*\)/m,lookbehind:!0,inside:{keyword:{pattern:/^(\s*)at(?= )/,lookbehind:!0},source:[{pattern:/(\()\w+\.\w+:\d+(?=\))/,lookbehind:!0,inside:{file:/^\w+\.\w+/,punctuation:/:/,"line-number":{pattern:/\b\d+\b/,alias:"number"}}},{pattern:/(\()[^()]*(?=\))/,lookbehind:!0,inside:{keyword:/^(?:Native Method|Unknown Source)$/}}],"class-name":/[\w$]+(?=\.(?:<init>|[\w$]+)\()/,function:/(?:<init>|[\w$]+)(?=\()/,"class-loader":{pattern:/(\s)[a-z]\w*(?:\.[a-z]\w*)*(?=\/[\w@$.]*\/)/,lookbehind:!0,alias:"namespace",inside:{punctuation:/\./}},module:{pattern:/([\s/])[a-z]\w*(?:\.[a-z]\w*)*(?:@[\w$.+-]*)?(?=\/)/,lookbehind:!0,inside:{version:{pattern:/(@)[\s\S]+/,lookbehind:!0,alias:"number"},punctuation:/[@.]/}},namespace:{pattern:/(?:\b[a-z]\w*\.)+/,inside:{punctuation:/\./}},punctuation:/[()/.]/}},more:{pattern:/^([\t ]*)\.{3} \d+ [a-z]+(?: [a-z]+)*/m,lookbehind:!0,inside:{punctuation:/\.{3}/,number:/\d+/,keyword:/\b[a-z]+(?: [a-z]+)*\b/}}}}return lze}var dze,JHt;function JFr(){if(JHt)return dze;JHt=1,dze=n,n.displayName="jexl",n.aliases=[];function n(r){r.languages.jexl={string:/(["'])(?:\\[\s\S]|(?!\1)[^\\])*\1/,transform:{pattern:/(\|\s*)[a-zA-Zа-яА-Я_\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$][\wа-яА-Я\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$]*/,alias:"function",lookbehind:!0},function:/[a-zA-Zа-яА-Я_\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$][\wа-яА-Я\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$]*\s*(?=\()/,number:/\b\d+(?:\.\d+)?\b|\B\.\d+\b/,operator:/[<>!]=?|-|\+|&&|==|\|\|?|\/\/?|[?:*^%]/,boolean:/\b(?:false|true)\b/,keyword:/\bin\b/,punctuation:/[{}[\](),.]/}}return dze}var fze,ZHt;function ZFr(){if(ZHt)return fze;ZHt=1,fze=n,n.displayName="jolie",n.aliases=[];function n(r){r.languages.jolie=r.languages.extend("clike",{string:{pattern:/(^|[^\\])"(?:\\[\s\S]|[^"\\])*"/,lookbehind:!0,greedy:!0},"class-name":{pattern:/((?:\b(?:as|courier|embed|in|inputPort|outputPort|service)\b|@)[ \t]*)\w+/,lookbehind:!0},keyword:/\b(?:as|cH|comp|concurrent|constants|courier|cset|csets|default|define|else|embed|embedded|execution|exit|extender|for|foreach|forward|from|global|if|import|in|include|init|inputPort|install|instanceof|interface|is_defined|linkIn|linkOut|main|new|nullProcess|outputPort|over|private|provide|public|scope|sequential|service|single|spawn|synchronized|this|throw|throws|type|undef|until|while|with)\b/,function:/\b[a-z_]\w*(?=[ \t]*[@(])/i,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?l?/i,operator:/-[-=>]?|\+[+=]?|<[<=]?|[>=*!]=?|&&|\|\||[?\/%^@|]/,punctuation:/[()[\]{},;.:]/,builtin:/\b(?:Byte|any|bool|char|double|enum|float|int|length|long|ranges|regex|string|undefined|void)\b/}),r.languages.insertBefore("jolie","keyword",{aggregates:{pattern:/(\bAggregates\s*:\s*)(?:\w+(?:\s+with\s+\w+)?\s*,\s*)*\w+(?:\s+with\s+\w+)?/,lookbehind:!0,inside:{keyword:/\bwith\b/,"class-name":/\w+/,punctuation:/,/}},redirects:{pattern:/(\bRedirects\s*:\s*)(?:\w+\s*=>\s*\w+\s*,\s*)*(?:\w+\s*=>\s*\w+)/,lookbehind:!0,inside:{punctuation:/,/,"class-name":/\w+/,operator:/=>/}},property:{pattern:/\b(?:Aggregates|[Ii]nterfaces|Java|Javascript|Jolie|[Ll]ocation|OneWay|[Pp]rotocol|Redirects|RequestResponse)\b(?=[ \t]*:)/}})}return fze}var hze,QHt;function QFr(){if(QHt)return hze;QHt=1,hze=n,n.displayName="jq",n.aliases=[];function n(r){(function(o){var s=/\\\((?:[^()]|\([^()]*\))*\)/.source,c=RegExp(/(^|[^\\])"(?:[^"\r\n\\]|\\[^\r\n(]|__)*"/.source.replace(/__/g,function(){return s})),d={interpolation:{pattern:RegExp(/((?:^|[^\\])(?:\\{2})*)/.source+s),lookbehind:!0,inside:{content:{pattern:/^(\\\()[\s\S]+(?=\)$)/,lookbehind:!0,inside:null},punctuation:/^\\\(|\)$/}}},h=o.languages.jq={comment:/#.*/,property:{pattern:RegExp(c.source+/(?=\s*:(?!:))/.source),lookbehind:!0,greedy:!0,inside:d},string:{pattern:c,lookbehind:!0,greedy:!0,inside:d},function:{pattern:/(\bdef\s+)[a-z_]\w+/i,lookbehind:!0},variable:/\B\$\w+/,"property-literal":{pattern:/\b[a-z_]\w*(?=\s*:(?!:))/i,alias:"property"},keyword:/\b(?:as|break|catch|def|elif|else|end|foreach|if|import|include|label|module|modulemeta|null|reduce|then|try|while)\b/,boolean:/\b(?:false|true)\b/,number:/(?:\b\d+\.|\B\.)?\b\d+(?:[eE][+-]?\d+)?\b/,operator:[{pattern:/\|=?/,alias:"pipe"},/\.\.|[!=<>]?=|\?\/\/|\/\/=?|[-+*/%]=?|[<>?]|\b(?:and|not|or)\b/],"c-style-function":{pattern:/\b[a-z_]\w*(?=\s*\()/i,alias:"function"},punctuation:/::|[()\[\]{},:;]|\.(?=\s*[\[\w$])/,dot:{pattern:/\./,alias:"important"}};d.interpolation.inside.content.inside=h})(r)}return hze}var pze,eGt;function eBr(){if(eGt)return pze;eGt=1,pze=n,n.displayName="jsExtras",n.aliases=[];function n(r){(function(o){o.languages.insertBefore("javascript","function-variable",{"method-variable":{pattern:RegExp("(\\.\\s*)"+o.languages.javascript["function-variable"].pattern.source),lookbehind:!0,alias:["function-variable","method","function","property-access"]}}),o.languages.insertBefore("javascript","function",{method:{pattern:RegExp("(\\.\\s*)"+o.languages.javascript.function.source),lookbehind:!0,alias:["function","property-access"]}}),o.languages.insertBefore("javascript","constant",{"known-class-name":[{pattern:/\b(?:(?:Float(?:32|64)|(?:Int|Uint)(?:8|16|32)|Uint8Clamped)?Array|ArrayBuffer|BigInt|Boolean|DataView|Date|Error|Function|Intl|JSON|(?:Weak)?(?:Map|Set)|Math|Number|Object|Promise|Proxy|Reflect|RegExp|String|Symbol|WebAssembly)\b/,alias:"class-name"},{pattern:/\b(?:[A-Z]\w*)Error\b/,alias:"class-name"}]});function s(E,k){return RegExp(E.replace(/<ID>/g,function(){return/(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*/.source}),k)}o.languages.insertBefore("javascript","keyword",{imports:{pattern:s(/(\bimport\b\s*)(?:<ID>(?:\s*,\s*(?:\*\s*as\s+<ID>|\{[^{}]*\}))?|\*\s*as\s+<ID>|\{[^{}]*\})(?=\s*\bfrom\b)/.source),lookbehind:!0,inside:o.languages.javascript},exports:{pattern:s(/(\bexport\b\s*)(?:\*(?:\s*as\s+<ID>)?(?=\s*\bfrom\b)|\{[^{}]*\})/.source),lookbehind:!0,inside:o.languages.javascript}}),o.languages.javascript.keyword.unshift({pattern:/\b(?:as|default|export|from|import)\b/,alias:"module"},{pattern:/\b(?:await|break|catch|continue|do|else|finally|for|if|return|switch|throw|try|while|yield)\b/,alias:"control-flow"},{pattern:/\bnull\b/,alias:["null","nil"]},{pattern:/\bundefined\b/,alias:"nil"}),o.languages.insertBefore("javascript","operator",{spread:{pattern:/\.{3}/,alias:"operator"},arrow:{pattern:/=>/,alias:"operator"}}),o.languages.insertBefore("javascript","punctuation",{"property-access":{pattern:s(/(\.\s*)#?<ID>/.source),lookbehind:!0},"maybe-class-name":{pattern:/(^|[^$\w\xA0-\uFFFF])[A-Z][$\w\xA0-\uFFFF]+/,lookbehind:!0},dom:{pattern:/\b(?:document|(?:local|session)Storage|location|navigator|performance|window)\b/,alias:"variable"},console:{pattern:/\bconsole(?=\s*\.)/,alias:"class-name"}});for(var c=["function","function-variable","method","method-variable","property-access"],d=0;d<c.length;d++){var h=c[d],b=o.languages.javascript[h];o.util.type(b)==="RegExp"&&(b=o.languages.javascript[h]={pattern:b});var y=b.inside||{};b.inside=y,y["maybe-class-name"]=/^[A-Z][\s\S]*/}})(r)}return pze}var gze,tGt;function tBr(){if(tGt)return gze;tGt=1,gze=n,n.displayName="jsTemplates",n.aliases=[];function n(r){(function(o){var s=o.languages.javascript["template-string"],c=s.pattern.source,d=s.inside.interpolation,h=d.inside["interpolation-punctuation"],b=d.pattern.source;function y($,U){if(o.languages[$])return{pattern:RegExp("((?:"+U+")\\s*)"+c),lookbehind:!0,greedy:!0,inside:{"template-punctuation":{pattern:/^`|`$/,alias:"string"},"embedded-code":{pattern:/[\s\S]+/,alias:$}}}}o.languages.javascript["template-string"]=[y("css",/\b(?:styled(?:\([^)]*\))?(?:\s*\.\s*\w+(?:\([^)]*\))*)*|css(?:\s*\.\s*(?:global|resolve))?|createGlobalStyle|keyframes)/.source),y("html",/\bhtml|\.\s*(?:inner|outer)HTML\s*\+?=/.source),y("svg",/\bsvg/.source),y("markdown",/\b(?:markdown|md)/.source),y("graphql",/\b(?:gql|graphql(?:\s*\.\s*experimental)?)/.source),y("sql",/\bsql/.source),s].filter(Boolean);function E($,U){return"___"+U.toUpperCase()+"_"+$+"___"}function k($,U,B){var G={code:$,grammar:U,language:B};return o.hooks.run("before-tokenize",G),G.tokens=o.tokenize(G.code,G.grammar),o.hooks.run("after-tokenize",G),G.tokens}function x($){var U={};U["interpolation-punctuation"]=h;var B=o.tokenize($,U);if(B.length===3){var G=[1,1];G.push.apply(G,k(B[1],o.languages.javascript,"javascript")),B.splice.apply(B,G)}return new o.Token("interpolation",B,d.alias,$)}function O($,U,B){var G=o.tokenize($,{interpolation:{pattern:RegExp(b),lookbehind:!0}}),K=0,X={},ae=G.map(function(we){if(typeof we=="string")return we;for(var ge=we.content,Te;$.indexOf(Te=E(K++,B))!==-1;);return X[Te]=ge,Te}).join(""),oe=k(ae,U,B),he=Object.keys(X);K=0;function ye(we){for(var ge=0;ge<we.length;ge++){if(K>=he.length)return;var Te=we[ge];if(typeof Te=="string"||typeof Te.content=="string"){var xe=he[K],Ue=typeof Te=="string"?Te:Te.content,Z=Ue.indexOf(xe);if(Z!==-1){++K;var Re=Ue.substring(0,Z),_e=x(X[xe]),qe=Ue.substring(Z+xe.length),it=[];if(Re&&it.push(Re),it.push(_e),qe){var Xe=[qe];ye(Xe),it.push.apply(it,Xe)}typeof Te=="string"?(we.splice.apply(we,[ge,1].concat(it)),ge+=it.length-1):Te.content=it}}else{var de=Te.content;Array.isArray(de)?ye(de):ye([de])}}}return ye(oe),new o.Token(B,oe,"language-"+B,$)}var P={javascript:!0,js:!0,typescript:!0,ts:!0,jsx:!0,tsx:!0};o.hooks.add("after-tokenize",function($){if(!($.language in P))return;function U(B){for(var G=0,K=B.length;G<K;G++){var X=B[G];if(typeof X!="string"){var ae=X.content;if(!Array.isArray(ae)){typeof ae!="string"&&U([ae]);continue}if(X.type==="template-string"){var oe=ae[1];if(ae.length===3&&typeof oe!="string"&&oe.type==="embedded-code"){var he=D(oe),ye=oe.alias,we=Array.isArray(ye)?ye[0]:ye,ge=o.languages[we];if(!ge)continue;ae[1]=O(he,ge,we)}}else U(ae)}}}U($.tokens)});function D($){return typeof $=="string"?$:Array.isArray($)?$.map(D).join(""):D($.content)}})(r)}return gze}var mze,nGt;function _Qe(){if(nGt)return mze;nGt=1,mze=n,n.displayName="typescript",n.aliases=["ts"];function n(r){(function(o){o.languages.typescript=o.languages.extend("javascript",{"class-name":{pattern:/(\b(?:class|extends|implements|instanceof|interface|new|type)\s+)(?!keyof\b)(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?:\s*<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>)?/,lookbehind:!0,greedy:!0,inside:null},builtin:/\b(?:Array|Function|Promise|any|boolean|console|never|number|string|symbol|unknown)\b/}),o.languages.typescript.keyword.push(/\b(?:abstract|declare|is|keyof|readonly|require)\b/,/\b(?:asserts|infer|interface|module|namespace|type)\b(?=\s*(?:[{_$a-zA-Z\xA0-\uFFFF]|$))/,/\btype\b(?=\s*(?:[\{*]|$))/),delete o.languages.typescript.parameter,delete o.languages.typescript["literal-property"];var s=o.languages.extend("typescript",{});delete s["class-name"],o.languages.typescript["class-name"].inside=s,o.languages.insertBefore("typescript","function",{decorator:{pattern:/@[$\w\xA0-\uFFFF]+/,inside:{at:{pattern:/^@/,alias:"operator"},function:/^[\s\S]+/}},"generic-function":{pattern:/#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*\s*<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>(?=\s*\()/,greedy:!0,inside:{function:/^#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*/,generic:{pattern:/<[\s\S]+/,alias:"class-name",inside:s}}}}),o.languages.ts=o.languages.typescript})(r)}return mze}var bze,rGt;function nBr(){if(rGt)return bze;rGt=1;var n=LEe(),r=_Qe();bze=o,o.displayName="jsdoc",o.aliases=[];function o(s){s.register(n),s.register(r),function(c){var d=c.languages.javascript,h=/\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})+\}/.source,b="(@(?:arg|argument|param|property)\\s+(?:"+h+"\\s+)?)";c.languages.jsdoc=c.languages.extend("javadoclike",{parameter:{pattern:RegExp(b+/(?:(?!\s)[$\w\xA0-\uFFFF.])+(?=\s|$)/.source),lookbehind:!0,inside:{punctuation:/\./}}}),c.languages.insertBefore("jsdoc","keyword",{"optional-parameter":{pattern:RegExp(b+/\[(?:(?!\s)[$\w\xA0-\uFFFF.])+(?:=[^[\]]+)?\](?=\s|$)/.source),lookbehind:!0,inside:{parameter:{pattern:/(^\[)[$\w\xA0-\uFFFF\.]+/,lookbehind:!0,inside:{punctuation:/\./}},code:{pattern:/(=)[\s\S]*(?=\]$)/,lookbehind:!0,inside:d,alias:"language-javascript"},punctuation:/[=[\]]/}},"class-name":[{pattern:RegExp(/(@(?:augments|class|extends|interface|memberof!?|template|this|typedef)\s+(?:<TYPE>\s+)?)[A-Z]\w*(?:\.[A-Z]\w*)*/.source.replace(/<TYPE>/g,function(){return h})),lookbehind:!0,inside:{punctuation:/\./}},{pattern:RegExp("(@[a-z]+\\s+)"+h),lookbehind:!0,inside:{string:d.string,number:d.number,boolean:d.boolean,keyword:c.languages.typescript.keyword,operator:/=>|\.\.\.|[&|?:*]/,punctuation:/[.,;=<>{}()[\]]/}}],example:{pattern:/(@example\s+(?!\s))(?:[^@\s]|\s+(?!\s))+?(?=\s*(?:\*\s*)?(?:@\w|\*\/))/,lookbehind:!0,inside:{code:{pattern:/^([\t ]*(?:\*\s*)?)\S.*$/m,lookbehind:!0,inside:d,alias:"language-javascript"}}}}),c.languages.javadoclike.addSupport("javascript",c.languages.jsdoc)}(s)}return bze}var vze,iGt;function kQe(){if(iGt)return vze;iGt=1,vze=n,n.displayName="json",n.aliases=["webmanifest"];function n(r){r.languages.json={property:{pattern:/(^|[^\\])"(?:\\.|[^\\"\r\n])*"(?=\s*:)/,lookbehind:!0,greedy:!0},string:{pattern:/(^|[^\\])"(?:\\.|[^\\"\r\n])*"(?!\s*:)/,lookbehind:!0,greedy:!0},comment:{pattern:/\/\/.*|\/\*[\s\S]*?(?:\*\/|$)/,greedy:!0},number:/-?\b\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,punctuation:/[{}[\],]/,operator:/:/,boolean:/\b(?:false|true)\b/,null:{pattern:/\bnull\b/,alias:"keyword"}},r.languages.webmanifest=r.languages.json}return vze}var wze,oGt;function rBr(){if(oGt)return wze;oGt=1;var n=kQe();wze=r,r.displayName="json5",r.aliases=[];function r(o){o.register(n),function(s){var c=/("|')(?:\\(?:\r\n?|\n|.)|(?!\1)[^\\\r\n])*\1/;s.languages.json5=s.languages.extend("json",{property:[{pattern:RegExp(c.source+"(?=\\s*:)"),greedy:!0},{pattern:/(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*:)/,alias:"unquoted"}],string:{pattern:c,greedy:!0},number:/[+-]?\b(?:NaN|Infinity|0x[a-fA-F\d]+)\b|[+-]?(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[eE][+-]?\d+\b)?/})}(o)}return wze}var yze,aGt;function iBr(){if(aGt)return yze;aGt=1;var n=kQe();yze=r,r.displayName="jsonp",r.aliases=[];function r(o){o.register(n),o.languages.jsonp=o.languages.extend("json",{punctuation:/[{}[\]();,.]/}),o.languages.insertBefore("jsonp","punctuation",{function:/(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*\()/})}return yze}var Eze,sGt;function oBr(){if(sGt)return Eze;sGt=1,Eze=n,n.displayName="jsstacktrace",n.aliases=[];function n(r){r.languages.jsstacktrace={"error-message":{pattern:/^\S.*/m,alias:"string"},"stack-frame":{pattern:/(^[ \t]+)at[ \t].*/m,lookbehind:!0,inside:{"not-my-code":{pattern:/^at[ \t]+(?!\s)(?:node\.js|<unknown>|.*(?:node_modules|\(<anonymous>\)|\(<unknown>|<anonymous>$|\(internal\/|\(node\.js)).*/m,alias:"comment"},filename:{pattern:/(\bat\s+(?!\s)|\()(?:[a-zA-Z]:)?[^():]+(?=:)/,lookbehind:!0,alias:"url"},function:{pattern:/(\bat\s+(?:new\s+)?)(?!\s)[_$a-zA-Z\xA0-\uFFFF<][.$\w\xA0-\uFFFF<>]*/,lookbehind:!0,inside:{punctuation:/\./}},punctuation:/[()]/,keyword:/\b(?:at|new)\b/,alias:{pattern:/\[(?:as\s+)?(?!\s)[_$a-zA-Z\xA0-\uFFFF][$\w\xA0-\uFFFF]*\]/,alias:"variable"},"line-number":{pattern:/:\d+(?::\d+)?\b/,alias:"number",inside:{punctuation:/:/}}}}}}return Eze}var Sze,cGt;function udn(){if(cGt)return Sze;cGt=1,Sze=n,n.displayName="jsx",n.aliases=[];function n(r){(function(o){var s=o.util.clone(o.languages.javascript),c=/(?:\s|\/\/.*(?!.)|\/\*(?:[^*]|\*(?!\/))\*\/)/.source,d=/(?:\{(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])*\})/.source,h=/(?:\{<S>*\.{3}(?:[^{}]|<BRACES>)*\})/.source;function b(k,x){return k=k.replace(/<S>/g,function(){return c}).replace(/<BRACES>/g,function(){return d}).replace(/<SPREAD>/g,function(){return h}),RegExp(k,x)}h=b(h).source,o.languages.jsx=o.languages.extend("markup",s),o.languages.jsx.tag.pattern=b(/<\/?(?:[\w.:-]+(?:<S>+(?:[\w.:$-]+(?:=(?:"(?:\\[\s\S]|[^\\"])*"|'(?:\\[\s\S]|[^\\'])*'|[^\s{'"/>=]+|<BRACES>))?|<SPREAD>))*<S>*\/?)?>/.source),o.languages.jsx.tag.inside.tag.pattern=/^<\/?[^\s>\/]*/,o.languages.jsx.tag.inside["attr-value"].pattern=/=(?!\{)(?:"(?:\\[\s\S]|[^\\"])*"|'(?:\\[\s\S]|[^\\'])*'|[^\s'">]+)/,o.languages.jsx.tag.inside.tag.inside["class-name"]=/^[A-Z]\w*(?:\.[A-Z]\w*)*$/,o.languages.jsx.tag.inside.comment=s.comment,o.languages.insertBefore("inside","attr-name",{spread:{pattern:b(/<SPREAD>/.source),inside:o.languages.jsx}},o.languages.jsx.tag),o.languages.insertBefore("inside","special-attr",{script:{pattern:b(/=<BRACES>/.source),alias:"language-javascript",inside:{"script-punctuation":{pattern:/^=(?=\{)/,alias:"punctuation"},rest:o.languages.jsx}}},o.languages.jsx.tag);var y=function(k){return k?typeof k=="string"?k:typeof k.content=="string"?k.content:k.content.map(y).join(""):""},E=function(k){for(var x=[],O=0;O<k.length;O++){var P=k[O],D=!1;if(typeof P!="string"&&(P.type==="tag"&&P.content[0]&&P.content[0].type==="tag"?P.content[0].content[0].content==="</"?x.length>0&&x[x.length-1].tagName===y(P.content[0].content[1])&&x.pop():P.content[P.content.length-1].content==="/>"||x.push({tagName:y(P.content[0].content[1]),openedBraces:0}):x.length>0&&P.type==="punctuation"&&P.content==="{"?x[x.length-1].openedBraces++:x.length>0&&x[x.length-1].openedBraces>0&&P.type==="punctuation"&&P.content==="}"?x[x.length-1].openedBraces--:D=!0),(D||typeof P=="string")&&x.length>0&&x[x.length-1].openedBraces===0){var $=y(P);O<k.length-1&&(typeof k[O+1]=="string"||k[O+1].type==="plain-text")&&($+=y(k[O+1]),k.splice(O+1,1)),O>0&&(typeof k[O-1]=="string"||k[O-1].type==="plain-text")&&($=y(k[O-1])+$,k.splice(O-1,1),O--),k[O]=new o.Token("plain-text",$,null,$)}P.content&&typeof P.content!="string"&&E(P.content)}};o.hooks.add("after-tokenize",function(k){k.language!=="jsx"&&k.language!=="tsx"||E(k.tokens)})})(r)}return Sze}var _ze,uGt;function aBr(){if(uGt)return _ze;uGt=1,_ze=n,n.displayName="julia",n.aliases=[];function n(r){r.languages.julia={comment:{pattern:/(^|[^\\])(?:#=(?:[^#=]|=(?!#)|#(?!=)|#=(?:[^#=]|=(?!#)|#(?!=))*=#)*=#|#.*)/,lookbehind:!0},regex:{pattern:/r"(?:\\.|[^"\\\r\n])*"[imsx]{0,4}/,greedy:!0},string:{pattern:/"""[\s\S]+?"""|(?:\b\w+)?"(?:\\.|[^"\\\r\n])*"|`(?:[^\\`\r\n]|\\.)*`/,greedy:!0},char:{pattern:/(^|[^\w'])'(?:\\[^\r\n][^'\r\n]*|[^\\\r\n])'/,lookbehind:!0,greedy:!0},keyword:/\b(?:abstract|baremodule|begin|bitstype|break|catch|ccall|const|continue|do|else|elseif|end|export|finally|for|function|global|if|immutable|import|importall|in|let|local|macro|module|print|println|quote|return|struct|try|type|typealias|using|while)\b/,boolean:/\b(?:false|true)\b/,number:/(?:\b(?=\d)|\B(?=\.))(?:0[box])?(?:[\da-f]+(?:_[\da-f]+)*(?:\.(?:\d+(?:_\d+)*)?)?|\.\d+(?:_\d+)*)(?:[efp][+-]?\d+(?:_\d+)*)?j?/i,operator:/&&|\|\||[-+*^%÷⊻&$\\]=?|\/[\/=]?|!=?=?|\|[=>]?|<(?:<=?|[=:|])?|>(?:=|>>?=?)?|==?=?|[~≠≤≥'√∛]/,punctuation:/::?|[{}[\]();,.?]/,constant:/\b(?:(?:Inf|NaN)(?:16|32|64)?|im|pi)\b|[πℯ]/}}return _ze}var kze,lGt;function sBr(){if(lGt)return kze;lGt=1,kze=n,n.displayName="keepalived",n.aliases=[];function n(r){r.languages.keepalived={comment:{pattern:/[#!].*/,greedy:!0},string:{pattern:/(^|[^\\])(?:"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"|'(?:\\(?:\r\n|[\s\S])|[^'\\\r\n])*')/,lookbehind:!0,greedy:!0},ip:{pattern:RegExp(/\b(?:(?:(?:[\da-f]{1,4}:){7}[\da-f]{1,4}|(?:[\da-f]{1,4}:){6}:[\da-f]{1,4}|(?:[\da-f]{1,4}:){5}:(?:[\da-f]{1,4}:)?[\da-f]{1,4}|(?:[\da-f]{1,4}:){4}:(?:[\da-f]{1,4}:){0,2}[\da-f]{1,4}|(?:[\da-f]{1,4}:){3}:(?:[\da-f]{1,4}:){0,3}[\da-f]{1,4}|(?:[\da-f]{1,4}:){2}:(?:[\da-f]{1,4}:){0,4}[\da-f]{1,4}|(?:[\da-f]{1,4}:){6}<ipv4>|(?:[\da-f]{1,4}:){0,5}:<ipv4>|::(?:[\da-f]{1,4}:){0,5}<ipv4>|[\da-f]{1,4}::(?:[\da-f]{1,4}:){0,5}[\da-f]{1,4}|::(?:[\da-f]{1,4}:){0,6}[\da-f]{1,4}|(?:[\da-f]{1,4}:){1,7}:)(?:\/\d{1,3})?|<ipv4>(?:\/\d{1,2})?)\b/.source.replace(/<ipv4>/g,function(){return/(?:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)\.){3}(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d))/.source}),"i"),alias:"number"},path:{pattern:/(\s)\/(?:[^\/\s]+\/)*[^\/\s]*|\b[a-zA-Z]:\\(?:[^\\\s]+\\)*[^\\\s]*/,lookbehind:!0,alias:"string"},variable:/\$\{?\w+\}?/,email:{pattern:/[\w-]+@[\w-]+(?:\.[\w-]{2,3}){1,2}/,alias:"string"},"conditional-configuration":{pattern:/@\^?[\w-]+/,alias:"variable"},operator:/=/,property:/\b(?:BFD_CHECK|DNS_CHECK|FILE_CHECK|HTTP_GET|MISC_CHECK|NAME|PING_CHECK|SCRIPTS|SMTP_CHECK|SSL|SSL_GET|TCP_CHECK|UDP_CHECK|accept|advert_int|alpha|auth_pass|auth_type|authentication|bfd_cpu_affinity|bfd_instance|bfd_no_swap|bfd_priority|bfd_process_name|bfd_rlimit_rttime|bfd_rt_priority|bind_if|bind_port|bindto|ca|certificate|check_unicast_src|checker|checker_cpu_affinity|checker_log_all_failures|checker_no_swap|checker_priority|checker_rlimit_rttime|checker_rt_priority|child_wait_time|connect_ip|connect_port|connect_timeout|dbus_service_name|debug|default_interface|delay|delay_before_retry|delay_loop|digest|dont_track_primary|dynamic|dynamic_interfaces|enable_(?:dbus|script_security|sni|snmp_checker|snmp_rfc|snmp_rfcv2|snmp_rfcv3|snmp_vrrp|traps)|end|fall|fast_recovery|file|flag-[123]|fork_delay|full_command|fwmark|garp_group|garp_interval|garp_lower_prio_delay|garp_lower_prio_repeat|garp_master_delay|garp_master_refresh|garp_master_refresh_repeat|garp_master_repeat|global_defs|global_tracking|gna_interval|group|ha_suspend|hashed|helo_name|higher_prio_send_advert|hoplimit|http_protocol|hysteresis|idle_tx|include|inhibit_on_failure|init_fail|init_file|instance|interface|interfaces|interval|ip_family|ipvs_process_name|keepalived.conf|kernel_rx_buf_size|key|linkbeat_interfaces|linkbeat_use_polling|log_all_failures|log_unknown_vrids|lower_prio_no_advert|lthreshold|lvs_flush|lvs_flush_onstop|lvs_method|lvs_netlink_cmd_rcv_bufs|lvs_netlink_cmd_rcv_bufs_force|lvs_netlink_monitor_rcv_bufs|lvs_netlink_monitor_rcv_bufs_force|lvs_notify_fifo|lvs_notify_fifo_script|lvs_sched|lvs_sync_daemon|max_auto_priority|max_hops|mcast_src_ip|mh-fallback|mh-port|min_auto_priority_delay|min_rx|min_tx|misc_dynamic|misc_path|misc_timeout|multiplier|name|namespace_with_ipsets|native_ipv6|neighbor_ip|net_namespace|net_namespace_ipvs|nftables|nftables_counters|nftables_ifindex|nftables_priority|no_accept|no_checker_emails|no_email_faults|nopreempt|notification_email|notification_email_from|notify|notify_backup|notify_deleted|notify_down|notify_fault|notify_fifo|notify_fifo_script|notify_master|notify_master_rx_lower_pri|notify_priority_changes|notify_stop|notify_up|old_unicast_checksum|omega|ops|param_match|passive|password|path|persistence_engine|persistence_granularity|persistence_timeout|preempt|preempt_delay|priority|process|process_monitor_rcv_bufs|process_monitor_rcv_bufs_force|process_name|process_names|promote_secondaries|protocol|proxy_arp|proxy_arp_pvlan|quorum|quorum_down|quorum_max|quorum_up|random_seed|real_server|regex|regex_max_offset|regex_min_offset|regex_no_match|regex_options|regex_stack|reload_repeat|reload_time_file|require_reply|retry|rise|router_id|rs_init_notifies|script|script_user|sh-fallback|sh-port|shutdown_script|shutdown_script_timeout|skip_check_adv_addr|smtp_alert|smtp_alert_checker|smtp_alert_vrrp|smtp_connect_timeout|smtp_helo_name|smtp_server|snmp_socket|sorry_server|sorry_server_inhibit|sorry_server_lvs_method|source_ip|start|startup_script|startup_script_timeout|state|static_ipaddress|static_routes|static_rules|status_code|step|strict_mode|sync_group_tracking_weight|terminate_delay|timeout|track_bfd|track_file|track_group|track_interface|track_process|track_script|track_src_ip|ttl|type|umask|unicast_peer|unicast_src_ip|unicast_ttl|url|use_ipvlan|use_pid_dir|use_vmac|user|uthreshold|val[123]|version|virtual_ipaddress|virtual_ipaddress_excluded|virtual_router_id|virtual_routes|virtual_rules|virtual_server|virtual_server_group|virtualhost|vmac_xmit_base|vrrp|vrrp_(?:check_unicast_src|cpu_affinity|garp_interval|garp_lower_prio_delay|garp_lower_prio_repeat|garp_master_delay|garp_master_refresh|garp_master_refresh_repeat|garp_master_repeat|gna_interval|higher_prio_send_advert|instance|ipsets|iptables|lower_prio_no_advert|mcast_group4|mcast_group6|min_garp|netlink_cmd_rcv_bufs|netlink_cmd_rcv_bufs_force|netlink_monitor_rcv_bufs|netlink_monitor_rcv_bufs_force|no_swap|notify_fifo|notify_fifo_script|notify_priority_changes|priority|process_name|rlimit_rttime|rt_priority|rx_bufs_multiplier|rx_bufs_policy|script|skip_check_adv_addr|startup_delay|strict|sync_group|track_process|version)|warmup|weight)\b/,constant:/\b(?:A|AAAA|AH|BACKUP|CNAME|DR|MASTER|MX|NAT|NS|PASS|SCTP|SOA|TCP|TUN|TXT|UDP|dh|fo|lblc|lblcr|lc|mh|nq|ovf|rr|sed|sh|wlc|wrr)\b/,number:{pattern:/(^|[^\w.-])-?\d+(?:\.\d+)?/,lookbehind:!0},boolean:/\b(?:false|no|off|on|true|yes)\b/,punctuation:/[\{\}]/}}return kze}var Tze,dGt;function cBr(){if(dGt)return Tze;dGt=1,Tze=n,n.displayName="keyman",n.aliases=[];function n(r){r.languages.keyman={comment:{pattern:/\bc .*/i,greedy:!0},string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,greedy:!0},"virtual-key":{pattern:/\[\s*(?:(?:ALT|CAPS|CTRL|LALT|LCTRL|NCAPS|RALT|RCTRL|SHIFT)\s+)*(?:[TKU]_[\w?]+|[A-E]\d\d?|"[^"\r\n]*"|'[^'\r\n]*')\s*\]/i,greedy:!0,alias:"function"},"header-keyword":{pattern:/&\w+/,alias:"bold"},"header-statement":{pattern:/\b(?:bitmap|bitmaps|caps always off|caps on only|copyright|hotkey|language|layout|message|name|shift frees caps|version)\b/i,alias:"bold"},"rule-keyword":{pattern:/\b(?:any|baselayout|beep|call|context|deadkey|dk|if|index|layer|notany|nul|outs|platform|reset|return|save|set|store|use)\b/i,alias:"keyword"},"structural-keyword":{pattern:/\b(?:ansi|begin|group|match|nomatch|unicode|using keys)\b/i,alias:"keyword"},"compile-target":{pattern:/\$(?:keyman|keymanonly|keymanweb|kmfl|weaver):/i,alias:"property"},number:/\b(?:U\+[\dA-F]+|d\d+|x[\da-f]+|\d+)\b/i,operator:/[+>\\$]|\.\./,punctuation:/[()=,]/}}return Tze}var xze,fGt;function uBr(){if(fGt)return xze;fGt=1,xze=n,n.displayName="kotlin",n.aliases=["kt","kts"];function n(r){(function(o){o.languages.kotlin=o.languages.extend("clike",{keyword:{pattern:/(^|[^.])\b(?:abstract|actual|annotation|as|break|by|catch|class|companion|const|constructor|continue|crossinline|data|do|dynamic|else|enum|expect|external|final|finally|for|fun|get|if|import|in|infix|init|inline|inner|interface|internal|is|lateinit|noinline|null|object|open|operator|out|override|package|private|protected|public|reified|return|sealed|set|super|suspend|tailrec|this|throw|to|try|typealias|val|var|vararg|when|where|while)\b/,lookbehind:!0},function:[{pattern:/(?:`[^\r\n`]+`|\b\w+)(?=\s*\()/,greedy:!0},{pattern:/(\.)(?:`[^\r\n`]+`|\w+)(?=\s*\{)/,lookbehind:!0,greedy:!0}],number:/\b(?:0[xX][\da-fA-F]+(?:_[\da-fA-F]+)*|0[bB][01]+(?:_[01]+)*|\d+(?:_\d+)*(?:\.\d+(?:_\d+)*)?(?:[eE][+-]?\d+(?:_\d+)*)?[fFL]?)\b/,operator:/\+[+=]?|-[-=>]?|==?=?|!(?:!|==?)?|[\/*%<>]=?|[?:]:?|\.\.|&&|\|\||\b(?:and|inv|or|shl|shr|ushr|xor)\b/}),delete o.languages.kotlin["class-name"];var s={"interpolation-punctuation":{pattern:/^\$\{?|\}$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:o.languages.kotlin}};o.languages.insertBefore("kotlin","string",{"string-literal":[{pattern:/"""(?:[^$]|\$(?:(?!\{)|\{[^{}]*\}))*?"""/,alias:"multiline",inside:{interpolation:{pattern:/\$(?:[a-z_]\w*|\{[^{}]*\})/i,inside:s},string:/[\s\S]+/}},{pattern:/"(?:[^"\\\r\n$]|\\.|\$(?:(?!\{)|\{[^{}]*\}))*"/,alias:"singleline",inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:[a-z_]\w*|\{[^{}]*\})/i,lookbehind:!0,inside:s},string:/[\s\S]+/}}],char:{pattern:/'(?:[^'\\\r\n]|\\(?:.|u[a-fA-F0-9]{0,4}))'/,greedy:!0}}),delete o.languages.kotlin.string,o.languages.insertBefore("kotlin","keyword",{annotation:{pattern:/\B@(?:\w+:)?(?:[A-Z]\w*|\[[^\]]+\])/,alias:"builtin"}}),o.languages.insertBefore("kotlin","function",{label:{pattern:/\b\w+@|@\w+\b/,alias:"symbol"}}),o.languages.kt=o.languages.kotlin,o.languages.kts=o.languages.kotlin})(r)}return xze}var Cze,hGt;function lBr(){if(hGt)return Cze;hGt=1,Cze=n,n.displayName="kumir",n.aliases=["kum"];function n(r){(function(o){var s=/\s\x00-\x1f\x22-\x2f\x3a-\x3f\x5b-\x5e\x60\x7b-\x7e/.source;function c(d,h){return RegExp(d.replace(/<nonId>/g,s),h)}o.languages.kumir={comment:{pattern:/\|.*/},prolog:{pattern:/#.*/,greedy:!0},string:{pattern:/"[^\n\r"]*"|'[^\n\r']*'/,greedy:!0},boolean:{pattern:c(/(^|[<nonId>])(?:да|нет)(?=[<nonId>]|$)/.source),lookbehind:!0},"operator-word":{pattern:c(/(^|[<nonId>])(?:и|или|не)(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"keyword"},"system-variable":{pattern:c(/(^|[<nonId>])знач(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"keyword"},type:[{pattern:c(/(^|[<nonId>])(?:вещ|лит|лог|сим|цел)(?:\x20*таб)?(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"builtin"},{pattern:c(/(^|[<nonId>])(?:компл|сканкод|файл|цвет)(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"important"}],keyword:{pattern:c(/(^|[<nonId>])(?:алг|арг(?:\x20*рез)?|ввод|ВКЛЮЧИТЬ|вс[её]|выбор|вывод|выход|дано|для|до|дс|если|иначе|исп|использовать|кон(?:(?:\x20+|_)исп)?|кц(?:(?:\x20+|_)при)?|надо|нач|нс|нц|от|пауза|пока|при|раза?|рез|стоп|таб|то|утв|шаг)(?=[<nonId>]|$)/.source),lookbehind:!0},name:{pattern:c(/(^|[<nonId>])[^\d<nonId>][^<nonId>]*(?:\x20+[^<nonId>]+)*(?=[<nonId>]|$)/.source),lookbehind:!0},number:{pattern:c(/(^|[<nonId>])(?:\B\$[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)(?=[<nonId>]|$)/.source,"i"),lookbehind:!0},punctuation:/:=|[(),:;\[\]]/,"operator-char":{pattern:/\*\*?|<[=>]?|>=?|[-+/=]/,alias:"operator"}},o.languages.kum=o.languages.kumir})(r)}return Cze}var Aze,pGt;function dBr(){if(pGt)return Aze;pGt=1,Aze=n,n.displayName="kusto",n.aliases=[];function n(r){r.languages.kusto={comment:{pattern:/\/\/.*/,greedy:!0},string:{pattern:/```[\s\S]*?```|[hH]?(?:"(?:[^\r\n\\"]|\\.)*"|'(?:[^\r\n\\']|\\.)*'|@(?:"[^\r\n"]*"|'[^\r\n']*'))/,greedy:!0},verb:{pattern:/(\|\s*)[a-z][\w-]*/i,lookbehind:!0,alias:"keyword"},command:{pattern:/\.[a-z][a-z\d-]*\b/,alias:"keyword"},"class-name":/\b(?:bool|datetime|decimal|dynamic|guid|int|long|real|string|timespan)\b/,keyword:/\b(?:access|alias|and|anti|as|asc|auto|between|by|(?:contains|(?:ends|starts)with|has(?:perfix|suffix)?)(?:_cs)?|database|declare|desc|external|from|fullouter|has_all|in|ingestion|inline|inner|innerunique|into|(?:left|right)(?:anti(?:semi)?|inner|outer|semi)?|let|like|local|not|of|on|or|pattern|print|query_parameters|range|restrict|schema|set|step|table|tables|to|view|where|with|matches\s+regex|nulls\s+(?:first|last))(?![\w-])/,boolean:/\b(?:false|null|true)\b/,function:/\b[a-z_]\w*(?=\s*\()/,datetime:[{pattern:/\b(?:(?:Fri|Friday|Mon|Monday|Sat|Saturday|Sun|Sunday|Thu|Thursday|Tue|Tuesday|Wed|Wednesday)\s*,\s*)?\d{1,2}(?:\s+|-)(?:Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep)(?:\s+|-)\d{2}\s+\d{2}:\d{2}(?::\d{2})?(?:\s*(?:\b(?:[A-Z]|(?:[ECMT][DS]|GM|U)T)|[+-]\d{4}))?\b/,alias:"number"},{pattern:/[+-]?\b(?:\d{4}-\d{2}-\d{2}(?:[ T]\d{2}:\d{2}(?::\d{2}(?:\.\d+)?)?)?|\d{2}:\d{2}(?::\d{2}(?:\.\d+)?)?)Z?/,alias:"number"}],number:/\b(?:0x[0-9A-Fa-f]+|\d+(?:\.\d+)?(?:[Ee][+-]?\d+)?)(?:(?:min|sec|[mnµ]s|[dhms]|microsecond|tick)\b)?|[+-]?\binf\b/,operator:/=>|[!=]~|[!=<>]=?|[-+*/%|]|\.\./,punctuation:/[()\[\]{},;.:]/}}return Aze}var Ize,gGt;function fBr(){if(gGt)return Ize;gGt=1,Ize=n,n.displayName="latex",n.aliases=["tex","context"];function n(r){(function(o){var s=/\\(?:[^a-z()[\]]|[a-z*]+)/i,c={"equation-command":{pattern:s,alias:"regex"}};o.languages.latex={comment:/%.*/,cdata:{pattern:/(\\begin\{((?:lstlisting|verbatim)\*?)\})[\s\S]*?(?=\\end\{\2\})/,lookbehind:!0},equation:[{pattern:/\$\$(?:\\[\s\S]|[^\\$])+\$\$|\$(?:\\[\s\S]|[^\\$])+\$|\\\([\s\S]*?\\\)|\\\[[\s\S]*?\\\]/,inside:c,alias:"string"},{pattern:/(\\begin\{((?:align|eqnarray|equation|gather|math|multline)\*?)\})[\s\S]*?(?=\\end\{\2\})/,lookbehind:!0,inside:c,alias:"string"}],keyword:{pattern:/(\\(?:begin|cite|documentclass|end|label|ref|usepackage)(?:\[[^\]]+\])?\{)[^}]+(?=\})/,lookbehind:!0},url:{pattern:/(\\url\{)[^}]+(?=\})/,lookbehind:!0},headline:{pattern:/(\\(?:chapter|frametitle|paragraph|part|section|subparagraph|subsection|subsubparagraph|subsubsection|subsubsubparagraph)\*?(?:\[[^\]]+\])?\{)[^}]+(?=\})/,lookbehind:!0,alias:"class-name"},function:{pattern:s,alias:"selector"},punctuation:/[[\]{}&]/},o.languages.tex=o.languages.latex,o.languages.context=o.languages.latex})(r)}return Ize}var Oze,mGt;function DEe(){if(mGt)return Oze;mGt=1;var n=oy();Oze=r,r.displayName="php",r.aliases=[];function r(o){o.register(n),function(s){var c=/\/\*[\s\S]*?\*\/|\/\/.*|#(?!\[).*/,d=[{pattern:/\b(?:false|true)\b/i,alias:"boolean"},{pattern:/(::\s*)\b[a-z_]\w*\b(?!\s*\()/i,greedy:!0,lookbehind:!0},{pattern:/(\b(?:case|const)\s+)\b[a-z_]\w*(?=\s*[;=])/i,greedy:!0,lookbehind:!0},/\b(?:null)\b/i,/\b[A-Z_][A-Z0-9_]*\b(?!\s*\()/],h=/\b0b[01]+(?:_[01]+)*\b|\b0o[0-7]+(?:_[0-7]+)*\b|\b0x[\da-f]+(?:_[\da-f]+)*\b|(?:\b\d+(?:_\d+)*\.?(?:\d+(?:_\d+)*)?|\B\.\d+)(?:e[+-]?\d+)?/i,b=/<?=>|\?\?=?|\.{3}|\??->|[!=]=?=?|::|\*\*=?|--|\+\+|&&|\|\||<<|>>|[?~]|[/^|%*&<>.+-]=?/,y=/[{}\[\](),:;]/;s.languages.php={delimiter:{pattern:/\?>$|^<\?(?:php(?=\s)|=)?/i,alias:"important"},comment:c,variable:/\$+(?:\w+\b|(?=\{))/,package:{pattern:/(namespace\s+|use\s+(?:function\s+)?)(?:\\?\b[a-z_]\w*)+\b(?!\\)/i,lookbehind:!0,inside:{punctuation:/\\/}},"class-name-definition":{pattern:/(\b(?:class|enum|interface|trait)\s+)\b[a-z_]\w*(?!\\)\b/i,lookbehind:!0,alias:"class-name"},"function-definition":{pattern:/(\bfunction\s+)[a-z_]\w*(?=\s*\()/i,lookbehind:!0,alias:"function"},keyword:[{pattern:/(\(\s*)\b(?:array|bool|boolean|float|int|integer|object|string)\b(?=\s*\))/i,alias:"type-casting",greedy:!0,lookbehind:!0},{pattern:/([(,?]\s*)\b(?:array(?!\s*\()|bool|callable|(?:false|null)(?=\s*\|)|float|int|iterable|mixed|object|self|static|string)\b(?=\s*\$)/i,alias:"type-hint",greedy:!0,lookbehind:!0},{pattern:/(\)\s*:\s*(?:\?\s*)?)\b(?:array(?!\s*\()|bool|callable|(?:false|null)(?=\s*\|)|float|int|iterable|mixed|object|self|static|string|void)\b/i,alias:"return-type",greedy:!0,lookbehind:!0},{pattern:/\b(?:array(?!\s*\()|bool|float|int|iterable|mixed|object|string|void)\b/i,alias:"type-declaration",greedy:!0},{pattern:/(\|\s*)(?:false|null)\b|\b(?:false|null)(?=\s*\|)/i,alias:"type-declaration",greedy:!0,lookbehind:!0},{pattern:/\b(?:parent|self|static)(?=\s*::)/i,alias:"static-context",greedy:!0},{pattern:/(\byield\s+)from\b/i,lookbehind:!0},/\bclass\b/i,{pattern:/((?:^|[^\s>:]|(?:^|[^-])>|(?:^|[^:]):)\s*)\b(?:abstract|and|array|as|break|callable|case|catch|clone|const|continue|declare|default|die|do|echo|else|elseif|empty|enddeclare|endfor|endforeach|endif|endswitch|endwhile|enum|eval|exit|extends|final|finally|fn|for|foreach|function|global|goto|if|implements|include|include_once|instanceof|insteadof|interface|isset|list|match|namespace|new|or|parent|print|private|protected|public|require|require_once|return|self|static|switch|throw|trait|try|unset|use|var|while|xor|yield|__halt_compiler)\b/i,lookbehind:!0}],"argument-name":{pattern:/([(,]\s+)\b[a-z_]\w*(?=\s*:(?!:))/i,lookbehind:!0},"class-name":[{pattern:/(\b(?:extends|implements|instanceof|new(?!\s+self|\s+static))\s+|\bcatch\s*\()\b[a-z_]\w*(?!\\)\b/i,greedy:!0,lookbehind:!0},{pattern:/(\|\s*)\b[a-z_]\w*(?!\\)\b/i,greedy:!0,lookbehind:!0},{pattern:/\b[a-z_]\w*(?!\\)\b(?=\s*\|)/i,greedy:!0},{pattern:/(\|\s*)(?:\\?\b[a-z_]\w*)+\b/i,alias:"class-name-fully-qualified",greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}},{pattern:/(?:\\?\b[a-z_]\w*)+\b(?=\s*\|)/i,alias:"class-name-fully-qualified",greedy:!0,inside:{punctuation:/\\/}},{pattern:/(\b(?:extends|implements|instanceof|new(?!\s+self\b|\s+static\b))\s+|\bcatch\s*\()(?:\\?\b[a-z_]\w*)+\b(?!\\)/i,alias:"class-name-fully-qualified",greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}},{pattern:/\b[a-z_]\w*(?=\s*\$)/i,alias:"type-declaration",greedy:!0},{pattern:/(?:\\?\b[a-z_]\w*)+(?=\s*\$)/i,alias:["class-name-fully-qualified","type-declaration"],greedy:!0,inside:{punctuation:/\\/}},{pattern:/\b[a-z_]\w*(?=\s*::)/i,alias:"static-context",greedy:!0},{pattern:/(?:\\?\b[a-z_]\w*)+(?=\s*::)/i,alias:["class-name-fully-qualified","static-context"],greedy:!0,inside:{punctuation:/\\/}},{pattern:/([(,?]\s*)[a-z_]\w*(?=\s*\$)/i,alias:"type-hint",greedy:!0,lookbehind:!0},{pattern:/([(,?]\s*)(?:\\?\b[a-z_]\w*)+(?=\s*\$)/i,alias:["class-name-fully-qualified","type-hint"],greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}},{pattern:/(\)\s*:\s*(?:\?\s*)?)\b[a-z_]\w*(?!\\)\b/i,alias:"return-type",greedy:!0,lookbehind:!0},{pattern:/(\)\s*:\s*(?:\?\s*)?)(?:\\?\b[a-z_]\w*)+\b(?!\\)/i,alias:["class-name-fully-qualified","return-type"],greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}}],constant:d,function:{pattern:/(^|[^\\\w])\\?[a-z_](?:[\w\\]*\w)?(?=\s*\()/i,lookbehind:!0,inside:{punctuation:/\\/}},property:{pattern:/(->\s*)\w+/,lookbehind:!0},number:h,operator:b,punctuation:y};var E={pattern:/\{\$(?:\{(?:\{[^{}]+\}|[^{}]+)\}|[^{}])+\}|(^|[^\\{])\$+(?:\w+(?:\[[^\r\n\[\]]+\]|->\w+)?)/,lookbehind:!0,inside:s.languages.php},k=[{pattern:/<<<'([^']+)'[\r\n](?:.*[\r\n])*?\1;/,alias:"nowdoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<<'[^']+'|[a-z_]\w*;$/i,alias:"symbol",inside:{punctuation:/^<<<'?|[';]$/}}}},{pattern:/<<<(?:"([^"]+)"[\r\n](?:.*[\r\n])*?\1;|([a-z_]\w*)[\r\n](?:.*[\r\n])*?\2;)/i,alias:"heredoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<<(?:"[^"]+"|[a-z_]\w*)|[a-z_]\w*;$/i,alias:"symbol",inside:{punctuation:/^<<<"?|[";]$/}},interpolation:E}},{pattern:/`(?:\\[\s\S]|[^\\`])*`/,alias:"backtick-quoted-string",greedy:!0},{pattern:/'(?:\\[\s\S]|[^\\'])*'/,alias:"single-quoted-string",greedy:!0},{pattern:/"(?:\\[\s\S]|[^\\"])*"/,alias:"double-quoted-string",greedy:!0,inside:{interpolation:E}}];s.languages.insertBefore("php","variable",{string:k,attribute:{pattern:/#\[(?:[^"'\/#]|\/(?![*/])|\/\/.*$|#(?!\[).*$|\/\*(?:[^*]|\*(?!\/))*\*\/|"(?:\\[\s\S]|[^\\"])*"|'(?:\\[\s\S]|[^\\'])*')+\](?=\s*[a-z$#])/im,greedy:!0,inside:{"attribute-content":{pattern:/^(#\[)[\s\S]+(?=\]$)/,lookbehind:!0,inside:{comment:c,string:k,"attribute-class-name":[{pattern:/([^:]|^)\b[a-z_]\w*(?!\\)\b/i,alias:"class-name",greedy:!0,lookbehind:!0},{pattern:/([^:]|^)(?:\\?\b[a-z_]\w*)+/i,alias:["class-name","class-name-fully-qualified"],greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}}],constant:d,number:h,operator:b,punctuation:y}},delimiter:{pattern:/^#\[|\]$/,alias:"punctuation"}}}}),s.hooks.add("before-tokenize",function(x){if(/<\?/.test(x.code)){var O=/<\?(?:[^"'/#]|\/(?![*/])|("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|(?:\/\/|#(?!\[))(?:[^?\n\r]|\?(?!>))*(?=$|\?>|[\r\n])|#\[|\/\*(?:[^*]|\*(?!\/))*(?:\*\/|$))*?(?:\?>|$)/g;s.languages["markup-templating"].buildPlaceholders(x,"php",O)}}),s.hooks.add("after-tokenize",function(x){s.languages["markup-templating"].tokenizePlaceholders(x,"php")})}(o)}return Oze}var Nze,bGt;function hBr(){if(bGt)return Nze;bGt=1;var n=oy(),r=DEe();Nze=o,o.displayName="latte",o.aliases=[];function o(s){s.register(n),s.register(r),function(c){c.languages.latte={comment:/^\{\*[\s\S]*/,"latte-tag":{pattern:/(^\{(?:\/(?=[a-z]))?)(?:[=_]|[a-z]\w*\b(?!\())/i,lookbehind:!0,alias:"important"},delimiter:{pattern:/^\{\/?|\}$/,alias:"punctuation"},php:{pattern:/\S(?:[\s\S]*\S)?/,alias:"language-php",inside:c.languages.php}};var d=c.languages.extend("markup",{});c.languages.insertBefore("inside","attr-value",{"n-attr":{pattern:/n:[\w-]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+))?/,inside:{"attr-name":{pattern:/^[^\s=]+/,alias:"important"},"attr-value":{pattern:/=[\s\S]+/,inside:{punctuation:[/^=/,{pattern:/^(\s*)["']|["']$/,lookbehind:!0}],php:{pattern:/\S(?:[\s\S]*\S)?/,inside:c.languages.php}}}}}},d.tag),c.hooks.add("before-tokenize",function(h){if(h.language==="latte"){var b=/\{\*[\s\S]*?\*\}|\{[^'"\s{}*](?:[^"'/{}]|\/(?![*/])|("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|\/\*(?:[^*]|\*(?!\/))*\*\/)*\}/g;c.languages["markup-templating"].buildPlaceholders(h,"latte",b),h.grammar=d}}),c.hooks.add("after-tokenize",function(h){c.languages["markup-templating"].tokenizePlaceholders(h,"latte")})}(s)}return Nze}var Rze,vGt;function pBr(){if(vGt)return Rze;vGt=1,Rze=n,n.displayName="less",n.aliases=[];function n(r){r.languages.less=r.languages.extend("css",{comment:[/\/\*[\s\S]*?\*\//,{pattern:/(^|[^\\])\/\/.*/,lookbehind:!0}],atrule:{pattern:/@[\w-](?:\((?:[^(){}]|\([^(){}]*\))*\)|[^(){};\s]|\s+(?!\s))*?(?=\s*\{)/,inside:{punctuation:/[:()]/}},selector:{pattern:/(?:@\{[\w-]+\}|[^{};\s@])(?:@\{[\w-]+\}|\((?:[^(){}]|\([^(){}]*\))*\)|[^(){};@\s]|\s+(?!\s))*?(?=\s*\{)/,inside:{variable:/@+[\w-]+/}},property:/(?:@\{[\w-]+\}|[\w-])+(?:\+_?)?(?=\s*:)/,operator:/[+\-*\/]/}),r.languages.insertBefore("less","property",{variable:[{pattern:/@[\w-]+\s*:/,inside:{punctuation:/:/}},/@@?[\w-]+/],"mixin-usage":{pattern:/([{;]\s*)[.#](?!\d)[\w-].*?(?=[(;])/,lookbehind:!0,alias:"function"}})}return Rze}var Pze,wGt;function TQe(){if(wGt)return Pze;wGt=1,Pze=n,n.displayName="scheme",n.aliases=[];function n(r){(function(o){o.languages.scheme={comment:/;.*|#;\s*(?:\((?:[^()]|\([^()]*\))*\)|\[(?:[^\[\]]|\[[^\[\]]*\])*\])|#\|(?:[^#|]|#(?!\|)|\|(?!#)|#\|(?:[^#|]|#(?!\|)|\|(?!#))*\|#)*\|#/,string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0},symbol:{pattern:/'[^()\[\]#'\s]+/,greedy:!0},char:{pattern:/#\\(?:[ux][a-fA-F\d]+\b|[-a-zA-Z]+\b|[\uD800-\uDBFF][\uDC00-\uDFFF]|\S)/,greedy:!0},"lambda-parameter":[{pattern:/((?:^|[^'`#])[(\[]lambda\s+)(?:[^|()\[\]'\s]+|\|(?:[^\\|]|\\.)*\|)/,lookbehind:!0},{pattern:/((?:^|[^'`#])[(\[]lambda\s+[(\[])[^()\[\]']+/,lookbehind:!0}],keyword:{pattern:/((?:^|[^'`#])[(\[])(?:begin|case(?:-lambda)?|cond(?:-expand)?|define(?:-library|-macro|-record-type|-syntax|-values)?|defmacro|delay(?:-force)?|do|else|except|export|guard|if|import|include(?:-ci|-library-declarations)?|lambda|let(?:rec)?(?:-syntax|-values|\*)?|let\*-values|only|parameterize|prefix|(?:quasi-?)?quote|rename|set!|syntax-(?:case|rules)|unless|unquote(?:-splicing)?|when)(?=[()\[\]\s]|$)/,lookbehind:!0},builtin:{pattern:/((?:^|[^'`#])[(\[])(?:abs|and|append|apply|assoc|ass[qv]|binary-port\?|boolean=?\?|bytevector(?:-append|-copy|-copy!|-length|-u8-ref|-u8-set!|\?)?|caar|cadr|call-with-(?:current-continuation|port|values)|call\/cc|car|cdar|cddr|cdr|ceiling|char(?:->integer|-ready\?|\?|<\?|<=\?|=\?|>\?|>=\?)|close-(?:input-port|output-port|port)|complex\?|cons|current-(?:error|input|output)-port|denominator|dynamic-wind|eof-object\??|eq\?|equal\?|eqv\?|error|error-object(?:-irritants|-message|\?)|eval|even\?|exact(?:-integer-sqrt|-integer\?|\?)?|expt|features|file-error\?|floor(?:-quotient|-remainder|\/)?|flush-output-port|for-each|gcd|get-output-(?:bytevector|string)|inexact\??|input-port(?:-open\?|\?)|integer(?:->char|\?)|lcm|length|list(?:->string|->vector|-copy|-ref|-set!|-tail|\?)?|make-(?:bytevector|list|parameter|string|vector)|map|max|member|memq|memv|min|modulo|negative\?|newline|not|null\?|number(?:->string|\?)|numerator|odd\?|open-(?:input|output)-(?:bytevector|string)|or|output-port(?:-open\?|\?)|pair\?|peek-char|peek-u8|port\?|positive\?|procedure\?|quotient|raise|raise-continuable|rational\?|rationalize|read-(?:bytevector|bytevector!|char|error\?|line|string|u8)|real\?|remainder|reverse|round|set-c[ad]r!|square|string(?:->list|->number|->symbol|->utf8|->vector|-append|-copy|-copy!|-fill!|-for-each|-length|-map|-ref|-set!|\?|<\?|<=\?|=\?|>\?|>=\?)?|substring|symbol(?:->string|\?|=\?)|syntax-error|textual-port\?|truncate(?:-quotient|-remainder|\/)?|u8-ready\?|utf8->string|values|vector(?:->list|->string|-append|-copy|-copy!|-fill!|-for-each|-length|-map|-ref|-set!|\?)?|with-exception-handler|write-(?:bytevector|char|string|u8)|zero\?)(?=[()\[\]\s]|$)/,lookbehind:!0},operator:{pattern:/((?:^|[^'`#])[(\[])(?:[-+*%/]|[<>]=?|=>?)(?=[()\[\]\s]|$)/,lookbehind:!0},number:{pattern:RegExp(s({"<ureal dec>":/\d+(?:\/\d+)|(?:\d+(?:\.\d*)?|\.\d+)(?:[esfdl][+-]?\d+)?/.source,"<real dec>":/[+-]?<ureal dec>|[+-](?:inf|nan)\.0/.source,"<imaginary dec>":/[+-](?:<ureal dec>|(?:inf|nan)\.0)?i/.source,"<complex dec>":/<real dec>(?:@<real dec>|<imaginary dec>)?|<imaginary dec>/.source,"<num dec>":/(?:#d(?:#[ei])?|#[ei](?:#d)?)?<complex dec>/.source,"<ureal box>":/[0-9a-f]+(?:\/[0-9a-f]+)?/.source,"<real box>":/[+-]?<ureal box>|[+-](?:inf|nan)\.0/.source,"<imaginary box>":/[+-](?:<ureal box>|(?:inf|nan)\.0)?i/.source,"<complex box>":/<real box>(?:@<real box>|<imaginary box>)?|<imaginary box>/.source,"<num box>":/#[box](?:#[ei])?|(?:#[ei])?#[box]<complex box>/.source,"<number>":/(^|[()\[\]\s])(?:<num dec>|<num box>)(?=[()\[\]\s]|$)/.source}),"i"),lookbehind:!0},boolean:{pattern:/(^|[()\[\]\s])#(?:[ft]|false|true)(?=[()\[\]\s]|$)/,lookbehind:!0},function:{pattern:/((?:^|[^'`#])[(\[])(?:[^|()\[\]'\s]+|\|(?:[^\\|]|\\.)*\|)(?=[()\[\]\s]|$)/,lookbehind:!0},identifier:{pattern:/(^|[()\[\]\s])\|(?:[^\\|]|\\.)*\|(?=[()\[\]\s]|$)/,lookbehind:!0,greedy:!0},punctuation:/[()\[\]']/};function s(c){for(var d in c)c[d]=c[d].replace(/<[\w\s]+>/g,function(h){return"(?:"+c[h].trim()+")"});return c[d]}})(r)}return Pze}var Lze,yGt;function gBr(){if(yGt)return Lze;yGt=1;var n=TQe();Lze=r,r.displayName="lilypond",r.aliases=[];function r(o){o.register(n),function(s){for(var c=/\((?:[^();"#\\]|\\[\s\S]|;.*(?!.)|"(?:[^"\\]|\\.)*"|#(?:\{(?:(?!#\})[\s\S])*#\}|[^{])|<expr>)*\)/.source,d=5,h=0;h<d;h++)c=c.replace(/<expr>/g,function(){return c});c=c.replace(/<expr>/g,/[^\s\S]/.source);var b=s.languages.lilypond={comment:/%(?:(?!\{).*|\{[\s\S]*?%\})/,"embedded-scheme":{pattern:RegExp(/(^|[=\s])#(?:"(?:[^"\\]|\\.)*"|[^\s()"]*(?:[^\s()]|<expr>))/.source.replace(/<expr>/g,function(){return c}),"m"),lookbehind:!0,greedy:!0,inside:{scheme:{pattern:/^(#)[\s\S]+$/,lookbehind:!0,alias:"language-scheme",inside:{"embedded-lilypond":{pattern:/#\{[\s\S]*?#\}/,greedy:!0,inside:{punctuation:/^#\{|#\}$/,lilypond:{pattern:/[\s\S]+/,alias:"language-lilypond",inside:null}}},rest:s.languages.scheme}},punctuation:/#/}},string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0},"class-name":{pattern:/(\\new\s+)[\w-]+/,lookbehind:!0},keyword:{pattern:/\\[a-z][-\w]*/i,inside:{punctuation:/^\\/}},operator:/[=|]|<<|>>/,punctuation:{pattern:/(^|[a-z\d])(?:'+|,+|[_^]?-[_^]?(?:[-+^!>._]|(?=\d))|[_^]\.?|[.!])|[{}()[\]<>^~]|\\[()[\]<>\\!]|--|__/,lookbehind:!0},number:/\b\d+(?:\/\d+)?\b/};b["embedded-scheme"].inside.scheme.inside["embedded-lilypond"].inside.lilypond.inside=b,s.languages.ly=b}(o)}return Lze}var Dze,EGt;function mBr(){if(EGt)return Dze;EGt=1;var n=oy();Dze=r,r.displayName="liquid",r.aliases=[];function r(o){o.register(n),o.languages.liquid={comment:{pattern:/(^\{%\s*comment\s*%\})[\s\S]+(?=\{%\s*endcomment\s*%\}$)/,lookbehind:!0},delimiter:{pattern:/^\{(?:\{\{|[%\{])-?|-?(?:\}\}|[%\}])\}$/,alias:"punctuation"},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},keyword:/\b(?:as|assign|break|(?:end)?(?:capture|case|comment|for|form|if|paginate|raw|style|tablerow|unless)|continue|cycle|decrement|echo|else|elsif|in|include|increment|limit|liquid|offset|range|render|reversed|section|when|with)\b/,object:/\b(?:address|all_country_option_tags|article|block|blog|cart|checkout|collection|color|country|country_option_tags|currency|current_page|current_tags|customer|customer_address|date|discount_allocation|discount_application|external_video|filter|filter_value|font|forloop|fulfillment|generic_file|gift_card|group|handle|image|line_item|link|linklist|localization|location|measurement|media|metafield|model|model_source|order|page|page_description|page_image|page_title|part|policy|product|product_option|recommendations|request|robots|routes|rule|script|search|selling_plan|selling_plan_allocation|selling_plan_group|shipping_method|shop|shop_locale|sitemap|store_availability|tax_line|template|theme|transaction|unit_price_measurement|user_agent|variant|video|video_source)\b/,function:[{pattern:/(\|\s*)\w+/,lookbehind:!0,alias:"filter"},{pattern:/(\.\s*)(?:first|last|size)/,lookbehind:!0}],boolean:/\b(?:false|nil|true)\b/,range:{pattern:/\.\./,alias:"operator"},number:/\b\d+(?:\.\d+)?\b/,operator:/[!=]=|<>|[<>]=?|[|?:=-]|\b(?:and|contains(?=\s)|or)\b/,punctuation:/[.,\[\]()]/,empty:{pattern:/\bempty\b/,alias:"keyword"}},o.hooks.add("before-tokenize",function(s){var c=/\{%\s*comment\s*%\}[\s\S]*?\{%\s*endcomment\s*%\}|\{(?:%[\s\S]*?%|\{\{[\s\S]*?\}\}|\{[\s\S]*?\})\}/g,d=!1;o.languages["markup-templating"].buildPlaceholders(s,"liquid",c,function(h){var b=/^\{%-?\s*(\w+)/.exec(h);if(b){var y=b[1];if(y==="raw"&&!d)return d=!0,!0;if(y==="endraw")return d=!1,!0}return!d})}),o.hooks.add("after-tokenize",function(s){o.languages["markup-templating"].tokenizePlaceholders(s,"liquid")})}return Dze}var Mze,SGt;function bBr(){if(SGt)return Mze;SGt=1,Mze=n,n.displayName="lisp",n.aliases=[];function n(r){(function(o){function s($){return RegExp(/(\()/.source+"(?:"+$+")"+/(?=[\s\)])/.source)}function c($){return RegExp(/([\s([])/.source+"(?:"+$+")"+/(?=[\s)])/.source)}var d=/(?!\d)[-+*/~!@$%^=<>{}\w]+/.source,h="&"+d,b="(\\()",y="(?=\\))",E="(?=\\s)",k=/(?:[^()]|\((?:[^()]|\((?:[^()]|\((?:[^()]|\((?:[^()]|\([^()]*\))*\))*\))*\))*\))*/.source,x={heading:{pattern:/;;;.*/,alias:["comment","title"]},comment:/;.*/,string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0,inside:{argument:/[-A-Z]+(?=[.,\s])/,symbol:RegExp("`"+d+"'")}},"quoted-symbol":{pattern:RegExp("#?'"+d),alias:["variable","symbol"]},"lisp-property":{pattern:RegExp(":"+d),alias:"property"},splice:{pattern:RegExp(",@?"+d),alias:["symbol","variable"]},keyword:[{pattern:RegExp(b+"(?:and|(?:cl-)?letf|cl-loop|cond|cons|error|if|(?:lexical-)?let\\*?|message|not|null|or|provide|require|setq|unless|use-package|when|while)"+E),lookbehind:!0},{pattern:RegExp(b+"(?:append|by|collect|concat|do|finally|for|in|return)"+E),lookbehind:!0}],declare:{pattern:s(/declare/.source),lookbehind:!0,alias:"keyword"},interactive:{pattern:s(/interactive/.source),lookbehind:!0,alias:"keyword"},boolean:{pattern:c(/nil|t/.source),lookbehind:!0},number:{pattern:c(/[-+]?\d+(?:\.\d*)?/.source),lookbehind:!0},defvar:{pattern:RegExp(b+"def(?:const|custom|group|var)\\s+"+d),lookbehind:!0,inside:{keyword:/^def[a-z]+/,variable:RegExp(d)}},defun:{pattern:RegExp(b+/(?:cl-)?(?:defmacro|defun\*?)\s+/.source+d+/\s+\(/.source+k+/\)/.source),lookbehind:!0,greedy:!0,inside:{keyword:/^(?:cl-)?def\S+/,arguments:null,function:{pattern:RegExp("(^\\s)"+d),lookbehind:!0},punctuation:/[()]/}},lambda:{pattern:RegExp(b+"lambda\\s+\\(\\s*(?:&?"+d+"(?:\\s+&?"+d+")*\\s*)?\\)"),lookbehind:!0,greedy:!0,inside:{keyword:/^lambda/,arguments:null,punctuation:/[()]/}},car:{pattern:RegExp(b+d),lookbehind:!0},punctuation:[/(?:['`,]?\(|[)\[\]])/,{pattern:/(\s)\.(?=\s)/,lookbehind:!0}]},O={"lisp-marker":RegExp(h),varform:{pattern:RegExp(/\(/.source+d+/\s+(?=\S)/.source+k+/\)/.source),inside:x},argument:{pattern:RegExp(/(^|[\s(])/.source+d),lookbehind:!0,alias:"variable"},rest:x},P="\\S+(?:\\s+\\S+)*",D={pattern:RegExp(b+k+y),lookbehind:!0,inside:{"rest-vars":{pattern:RegExp("&(?:body|rest)\\s+"+P),inside:O},"other-marker-vars":{pattern:RegExp("&(?:aux|optional)\\s+"+P),inside:O},keys:{pattern:RegExp("&key\\s+"+P+"(?:\\s+&allow-other-keys)?"),inside:O},argument:{pattern:RegExp(d),alias:"variable"},punctuation:/[()]/}};x.lambda.inside.arguments=D,x.defun.inside.arguments=o.util.clone(D),x.defun.inside.arguments.inside.sublist=D,o.languages.lisp=x,o.languages.elisp=x,o.languages.emacs=x,o.languages["emacs-lisp"]=x})(r)}return Mze}var jze,_Gt;function vBr(){if(_Gt)return jze;_Gt=1,jze=n,n.displayName="livescript",n.aliases=[];function n(r){r.languages.livescript={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0},{pattern:/(^|[^\\])#.*/,lookbehind:!0}],"interpolated-string":{pattern:/(^|[^"])("""|")(?:\\[\s\S]|(?!\2)[^\\])*\2(?!")/,lookbehind:!0,greedy:!0,inside:{variable:{pattern:/(^|[^\\])#[a-z_](?:-?[a-z]|[\d_])*/m,lookbehind:!0},interpolation:{pattern:/(^|[^\\])#\{[^}]+\}/m,lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^#\{|\}$/,alias:"variable"}}},string:/[\s\S]+/}},string:[{pattern:/('''|')(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0},{pattern:/<\[[\s\S]*?\]>/,greedy:!0},/\\[^\s,;\])}]+/],regex:[{pattern:/\/\/(?:\[[^\r\n\]]*\]|\\.|(?!\/\/)[^\\\[])+\/\/[gimyu]{0,5}/,greedy:!0,inside:{comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0}}},{pattern:/\/(?:\[[^\r\n\]]*\]|\\.|[^/\\\r\n\[])+\/[gimyu]{0,5}/,greedy:!0}],keyword:{pattern:/(^|(?!-).)\b(?:break|case|catch|class|const|continue|default|do|else|extends|fallthrough|finally|for(?: ever)?|function|if|implements|it|let|loop|new|null|otherwise|own|return|super|switch|that|then|this|throw|try|unless|until|var|void|when|while|yield)(?!-)\b/m,lookbehind:!0},"keyword-operator":{pattern:/(^|[^-])\b(?:(?:delete|require|typeof)!|(?:and|by|delete|export|from|import(?: all)?|in|instanceof|is(?: not|nt)?|not|of|or|til|to|typeof|with|xor)(?!-)\b)/m,lookbehind:!0,alias:"operator"},boolean:{pattern:/(^|[^-])\b(?:false|no|off|on|true|yes)(?!-)\b/m,lookbehind:!0},argument:{pattern:/(^|(?!\.&\.)[^&])&(?!&)\d*/m,lookbehind:!0,alias:"variable"},number:/\b(?:\d+~[\da-z]+|\d[\d_]*(?:\.\d[\d_]*)?(?:[a-z]\w*)?)/i,identifier:/[a-z_](?:-?[a-z]|[\d_])*/i,operator:[{pattern:/( )\.(?= )/,lookbehind:!0},/\.(?:[=~]|\.\.?)|\.(?:[&|^]|<<|>>>?)\.|:(?:=|:=?)|&&|\|[|>]|<(?:<<?<?|--?!?|~~?!?|[|=?])?|>[>=?]?|-(?:->?|>)?|\+\+?|@@?|%%?|\*\*?|!(?:~?=|--?>|~?~>)?|~(?:~?>|=)?|==?|\^\^?|[\/?]/],punctuation:/[(){}\[\]|.,:;`]/},r.languages.livescript["interpolated-string"].inside.interpolation.inside.rest=r.languages.livescript}return jze}var $ze,kGt;function wBr(){if(kGt)return $ze;kGt=1,$ze=n,n.displayName="llvm",n.aliases=[];function n(r){(function(o){o.languages.llvm={comment:/;.*/,string:{pattern:/"[^"]*"/,greedy:!0},boolean:/\b(?:false|true)\b/,variable:/[%@!#](?:(?!\d)(?:[-$.\w]|\\[a-f\d]{2})+|\d+)/i,label:/(?!\d)(?:[-$.\w]|\\[a-f\d]{2})+:/i,type:{pattern:/\b(?:double|float|fp128|half|i[1-9]\d*|label|metadata|ppc_fp128|token|void|x86_fp80|x86_mmx)\b/,alias:"class-name"},keyword:/\b[a-z_][a-z_0-9]*\b/,number:/[+-]?\b\d+(?:\.\d+)?(?:[eE][+-]?\d+)?\b|\b0x[\dA-Fa-f]+\b|\b0xK[\dA-Fa-f]{20}\b|\b0x[ML][\dA-Fa-f]{32}\b|\b0xH[\dA-Fa-f]{4}\b/,punctuation:/[{}[\];(),.!*=<>]/}})(r)}return $ze}var Fze,TGt;function yBr(){if(TGt)return Fze;TGt=1,Fze=n,n.displayName="log",n.aliases=[];function n(r){r.languages.log={string:{pattern:/"(?:[^"\\\r\n]|\\.)*"|'(?![st] | \w)(?:[^'\\\r\n]|\\.)*'/,greedy:!0},exception:{pattern:/(^|[^\w.])[a-z][\w.]*(?:Error|Exception):.*(?:(?:\r\n?|\n)[ \t]*(?:at[ \t].+|\.{3}.*|Caused by:.*))+(?:(?:\r\n?|\n)[ \t]*\.\.\. .*)?/,lookbehind:!0,greedy:!0,alias:["javastacktrace","language-javastacktrace"],inside:r.languages.javastacktrace||{keyword:/\bat\b/,function:/[a-z_][\w$]*(?=\()/,punctuation:/[.:()]/}},level:[{pattern:/\b(?:ALERT|CRIT|CRITICAL|EMERG|EMERGENCY|ERR|ERROR|FAILURE|FATAL|SEVERE)\b/,alias:["error","important"]},{pattern:/\b(?:WARN|WARNING|WRN)\b/,alias:["warning","important"]},{pattern:/\b(?:DISPLAY|INF|INFO|NOTICE|STATUS)\b/,alias:["info","keyword"]},{pattern:/\b(?:DBG|DEBUG|FINE)\b/,alias:["debug","keyword"]},{pattern:/\b(?:FINER|FINEST|TRACE|TRC|VERBOSE|VRB)\b/,alias:["trace","comment"]}],property:{pattern:/((?:^|[\]|])[ \t]*)[a-z_](?:[\w-]|\b\/\b)*(?:[. ]\(?\w(?:[\w-]|\b\/\b)*\)?)*:(?=\s)/im,lookbehind:!0},separator:{pattern:/(^|[^-+])-{3,}|={3,}|\*{3,}|- - /m,lookbehind:!0,alias:"comment"},url:/\b(?:file|ftp|https?):\/\/[^\s|,;'"]*[^\s|,;'">.]/,email:{pattern:/(^|\s)[-\w+.]+@[a-z][a-z0-9-]*(?:\.[a-z][a-z0-9-]*)+(?=\s)/,lookbehind:!0,alias:"url"},"ip-address":{pattern:/\b(?:\d{1,3}(?:\.\d{1,3}){3})\b/,alias:"constant"},"mac-address":{pattern:/\b[a-f0-9]{2}(?::[a-f0-9]{2}){5}\b/i,alias:"constant"},domain:{pattern:/(^|\s)[a-z][a-z0-9-]*(?:\.[a-z][a-z0-9-]*)*\.[a-z][a-z0-9-]+(?=\s)/,lookbehind:!0,alias:"constant"},uuid:{pattern:/\b[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\b/i,alias:"constant"},hash:{pattern:/\b(?:[a-f0-9]{32}){1,2}\b/i,alias:"constant"},"file-path":{pattern:/\b[a-z]:[\\/][^\s|,;:(){}\[\]"']+|(^|[\s:\[\](>|])\.{0,2}\/\w[^\s|,;:(){}\[\]"']*/i,lookbehind:!0,greedy:!0,alias:"string"},date:{pattern:RegExp(/\b\d{4}[-/]\d{2}[-/]\d{2}(?:T(?=\d{1,2}:)|(?=\s\d{1,2}:))/.source+"|"+/\b\d{1,4}[-/ ](?:\d{1,2}|Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep)[-/ ]\d{2,4}T?\b/.source+"|"+/\b(?:(?:Fri|Mon|Sat|Sun|Thu|Tue|Wed)(?:\s{1,2}(?:Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep))?|Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep)\s{1,2}\d{1,2}\b/.source,"i"),alias:"number"},time:{pattern:/\b\d{1,2}:\d{1,2}:\d{1,2}(?:[.,:]\d+)?(?:\s?[+-]\d{2}:?\d{2}|Z)?\b/,alias:"number"},boolean:/\b(?:false|null|true)\b/i,number:{pattern:/(^|[^.\w])(?:0x[a-f0-9]+|0o[0-7]+|0b[01]+|v?\d[\da-f]*(?:\.\d+)*(?:e[+-]?\d+)?[a-z]{0,3}\b)\b(?!\.\w)/i,lookbehind:!0},operator:/[;:?<=>~/@!$%&+\-|^(){}*#]/,punctuation:/[\[\].,]/}}return Fze}var Bze,xGt;function EBr(){if(xGt)return Bze;xGt=1,Bze=n,n.displayName="lolcode",n.aliases=[];function n(r){r.languages.lolcode={comment:[/\bOBTW\s[\s\S]*?\sTLDR\b/,/\bBTW.+/],string:{pattern:/"(?::.|[^":])*"/,inside:{variable:/:\{[^}]+\}/,symbol:[/:\([a-f\d]+\)/i,/:\[[^\]]+\]/,/:[)>o":]/]},greedy:!0},number:/(?:\B-)?(?:\b\d+(?:\.\d*)?|\B\.\d+)/,symbol:{pattern:/(^|\s)(?:A )?(?:BUKKIT|NOOB|NUMBAR|NUMBR|TROOF|YARN)(?=\s|,|$)/,lookbehind:!0,inside:{keyword:/A(?=\s)/}},label:{pattern:/((?:^|\s)(?:IM IN YR|IM OUTTA YR) )[a-zA-Z]\w*/,lookbehind:!0,alias:"string"},function:{pattern:/((?:^|\s)(?:HOW IZ I|I IZ|IZ) )[a-zA-Z]\w*/,lookbehind:!0},keyword:[{pattern:/(^|\s)(?:AN|FOUND YR|GIMMEH|GTFO|HAI|HAS A|HOW IZ I|I HAS A|I IZ|IF U SAY SO|IM IN YR|IM OUTTA YR|IS NOW(?: A)?|ITZ(?: A)?|IZ|KTHX|KTHXBYE|LIEK(?: A)?|MAEK|MEBBE|MKAY|NERFIN|NO WAI|O HAI IM|O RLY\?|OIC|OMG|OMGWTF|R|SMOOSH|SRS|TIL|UPPIN|VISIBLE|WILE|WTF\?|YA RLY|YR)(?=\s|,|$)/,lookbehind:!0},/'Z(?=\s|,|$)/],boolean:{pattern:/(^|\s)(?:FAIL|WIN)(?=\s|,|$)/,lookbehind:!0},variable:{pattern:/(^|\s)IT(?=\s|,|$)/,lookbehind:!0},operator:{pattern:/(^|\s)(?:NOT|BOTH SAEM|DIFFRINT|(?:ALL|ANY|BIGGR|BOTH|DIFF|EITHER|MOD|PRODUKT|QUOSHUNT|SMALLR|SUM|WON) OF)(?=\s|,|$)/,lookbehind:!0},punctuation:/\.{3}|…|,|!/}}return Bze}var Uze,CGt;function SBr(){if(CGt)return Uze;CGt=1,Uze=n,n.displayName="magma",n.aliases=[];function n(r){r.languages.magma={output:{pattern:/^(>.*(?:\r(?:\n|(?!\n))|\n))(?!>)(?:.+|(?:\r(?:\n|(?!\n))|\n)(?!>).*)(?:(?:\r(?:\n|(?!\n))|\n)(?!>).*)*/m,lookbehind:!0,greedy:!0},comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/(^|[^\\"])"(?:[^\r\n\\"]|\\.)*"/,lookbehind:!0,greedy:!0},keyword:/\b(?:_|adj|and|assert|assert2|assert3|assigned|break|by|case|cat|catch|clear|cmpeq|cmpne|continue|declare|default|delete|diff|div|do|elif|else|end|eq|error|eval|exists|exit|for|forall|forward|fprintf|freeze|function|ge|gt|if|iload|import|in|intrinsic|is|join|le|load|local|lt|meet|mod|ne|not|notadj|notin|notsubset|or|print|printf|procedure|quit|random|read|readi|repeat|require|requirege|requirerange|restore|return|save|sdiff|select|subset|then|time|to|try|until|vprint|vprintf|vtime|when|where|while|xor)\b/,boolean:/\b(?:false|true)\b/,generator:{pattern:/\b[a-z_]\w*(?=\s*<)/i,alias:"class-name"},function:/\b[a-z_]\w*(?=\s*\()/i,number:{pattern:/(^|[^\w.]|\.\.)(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?(?:_[a-z]?)?(?=$|[^\w.]|\.\.)/,lookbehind:!0},operator:/->|[-+*/^~!|#=]|:=|\.\./,punctuation:/[()[\]{}<>,;.:]/}}return Uze}var zze,AGt;function _Br(){if(AGt)return zze;AGt=1,zze=n,n.displayName="makefile",n.aliases=[];function n(r){r.languages.makefile={comment:{pattern:/(^|[^\\])#(?:\\(?:\r\n|[\s\S])|[^\\\r\n])*/,lookbehind:!0},string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},"builtin-target":{pattern:/\.[A-Z][^:#=\s]+(?=\s*:(?!=))/,alias:"builtin"},target:{pattern:/^(?:[^:=\s]|[ \t]+(?![\s:]))+(?=\s*:(?!=))/m,alias:"symbol",inside:{variable:/\$+(?:(?!\$)[^(){}:#=\s]+|(?=[({]))/}},variable:/\$+(?:(?!\$)[^(){}:#=\s]+|\([@*%<^+?][DF]\)|(?=[({]))/,keyword:/-include\b|\b(?:define|else|endef|endif|export|ifn?def|ifn?eq|include|override|private|sinclude|undefine|unexport|vpath)\b/,function:{pattern:/(\()(?:abspath|addsuffix|and|basename|call|dir|error|eval|file|filter(?:-out)?|findstring|firstword|flavor|foreach|guile|if|info|join|lastword|load|notdir|or|origin|patsubst|realpath|shell|sort|strip|subst|suffix|value|warning|wildcard|word(?:list|s)?)(?=[ \t])/,lookbehind:!0},operator:/(?:::|[?:+!])?=|[|@]/,punctuation:/[:;(){}]/}}return zze}var Hze,IGt;function kBr(){if(IGt)return Hze;IGt=1,Hze=n,n.displayName="markdown",n.aliases=["md"];function n(r){(function(o){var s=/(?:\\.|[^\\\n\r]|(?:\n|\r\n?)(?![\r\n]))/.source;function c(O){return O=O.replace(/<inner>/g,function(){return s}),RegExp(/((?:^|[^\\])(?:\\{2})*)/.source+"(?:"+O+")")}var d=/(?:\\.|``(?:[^`\r\n]|`(?!`))+``|`[^`\r\n]+`|[^\\|\r\n`])+/.source,h=/\|?__(?:\|__)+\|?(?:(?:\n|\r\n?)|(?![\s\S]))/.source.replace(/__/g,function(){return d}),b=/\|?[ \t]*:?-{3,}:?[ \t]*(?:\|[ \t]*:?-{3,}:?[ \t]*)+\|?(?:\n|\r\n?)/.source;o.languages.markdown=o.languages.extend("markup",{}),o.languages.insertBefore("markdown","prolog",{"front-matter-block":{pattern:/(^(?:\s*[\r\n])?)---(?!.)[\s\S]*?[\r\n]---(?!.)/,lookbehind:!0,greedy:!0,inside:{punctuation:/^---|---$/,"front-matter":{pattern:/\S+(?:\s+\S+)*/,alias:["yaml","language-yaml"],inside:o.languages.yaml}}},blockquote:{pattern:/^>(?:[\t ]*>)*/m,alias:"punctuation"},table:{pattern:RegExp("^"+h+b+"(?:"+h+")*","m"),inside:{"table-data-rows":{pattern:RegExp("^("+h+b+")(?:"+h+")*$"),lookbehind:!0,inside:{"table-data":{pattern:RegExp(d),inside:o.languages.markdown},punctuation:/\|/}},"table-line":{pattern:RegExp("^("+h+")"+b+"$"),lookbehind:!0,inside:{punctuation:/\||:?-{3,}:?/}},"table-header-row":{pattern:RegExp("^"+h+"$"),inside:{"table-header":{pattern:RegExp(d),alias:"important",inside:o.languages.markdown},punctuation:/\|/}}}},code:[{pattern:/((?:^|\n)[ \t]*\n|(?:^|\r\n?)[ \t]*\r\n?)(?: {4}|\t).+(?:(?:\n|\r\n?)(?: {4}|\t).+)*/,lookbehind:!0,alias:"keyword"},{pattern:/^```[\s\S]*?^```$/m,greedy:!0,inside:{"code-block":{pattern:/^(```.*(?:\n|\r\n?))[\s\S]+?(?=(?:\n|\r\n?)^```$)/m,lookbehind:!0},"code-language":{pattern:/^(```).+/,lookbehind:!0},punctuation:/```/}}],title:[{pattern:/\S.*(?:\n|\r\n?)(?:==+|--+)(?=[ \t]*$)/m,alias:"important",inside:{punctuation:/==+$|--+$/}},{pattern:/(^\s*)#.+/m,lookbehind:!0,alias:"important",inside:{punctuation:/^#+|#+$/}}],hr:{pattern:/(^\s*)([*-])(?:[\t ]*\2){2,}(?=\s*$)/m,lookbehind:!0,alias:"punctuation"},list:{pattern:/(^\s*)(?:[*+-]|\d+\.)(?=[\t ].)/m,lookbehind:!0,alias:"punctuation"},"url-reference":{pattern:/!?\[[^\]]+\]:[\t ]+(?:\S+|<(?:\\.|[^>\\])+>)(?:[\t ]+(?:"(?:\\.|[^"\\])*"|'(?:\\.|[^'\\])*'|\((?:\\.|[^)\\])*\)))?/,inside:{variable:{pattern:/^(!?\[)[^\]]+/,lookbehind:!0},string:/(?:"(?:\\.|[^"\\])*"|'(?:\\.|[^'\\])*'|\((?:\\.|[^)\\])*\))$/,punctuation:/^[\[\]!:]|[<>]/},alias:"url"},bold:{pattern:c(/\b__(?:(?!_)<inner>|_(?:(?!_)<inner>)+_)+__\b|\*\*(?:(?!\*)<inner>|\*(?:(?!\*)<inner>)+\*)+\*\*/.source),lookbehind:!0,greedy:!0,inside:{content:{pattern:/(^..)[\s\S]+(?=..$)/,lookbehind:!0,inside:{}},punctuation:/\*\*|__/}},italic:{pattern:c(/\b_(?:(?!_)<inner>|__(?:(?!_)<inner>)+__)+_\b|\*(?:(?!\*)<inner>|\*\*(?:(?!\*)<inner>)+\*\*)+\*/.source),lookbehind:!0,greedy:!0,inside:{content:{pattern:/(^.)[\s\S]+(?=.$)/,lookbehind:!0,inside:{}},punctuation:/[*_]/}},strike:{pattern:c(/(~~?)(?:(?!~)<inner>)+\2/.source),lookbehind:!0,greedy:!0,inside:{content:{pattern:/(^~~?)[\s\S]+(?=\1$)/,lookbehind:!0,inside:{}},punctuation:/~~?/}},"code-snippet":{pattern:/(^|[^\\`])(?:``[^`\r\n]+(?:`[^`\r\n]+)*``(?!`)|`[^`\r\n]+`(?!`))/,lookbehind:!0,greedy:!0,alias:["code","keyword"]},url:{pattern:c(/!?\[(?:(?!\])<inner>)+\](?:\([^\s)]+(?:[\t ]+"(?:\\.|[^"\\])*")?\)|[ \t]?\[(?:(?!\])<inner>)+\])/.source),lookbehind:!0,greedy:!0,inside:{operator:/^!/,content:{pattern:/(^\[)[^\]]+(?=\])/,lookbehind:!0,inside:{}},variable:{pattern:/(^\][ \t]?\[)[^\]]+(?=\]$)/,lookbehind:!0},url:{pattern:/(^\]\()[^\s)]+/,lookbehind:!0},string:{pattern:/(^[ \t]+)"(?:\\.|[^"\\])*"(?=\)$)/,lookbehind:!0}}}}),["url","bold","italic","strike"].forEach(function(O){["url","bold","italic","strike","code-snippet"].forEach(function(P){O!==P&&(o.languages.markdown[O].inside.content.inside[P]=o.languages.markdown[P])})}),o.hooks.add("after-tokenize",function(O){if(O.language!=="markdown"&&O.language!=="md")return;function P(D){if(!(!D||typeof D=="string"))for(var $=0,U=D.length;$<U;$++){var B=D[$];if(B.type!=="code"){P(B.content);continue}var G=B.content[1],K=B.content[3];if(G&&K&&G.type==="code-language"&&K.type==="code-block"&&typeof G.content=="string"){var X=G.content.replace(/\b#/g,"sharp").replace(/\b\+\+/g,"pp");X=(/[a-z][\w-]*/i.exec(X)||[""])[0].toLowerCase();var ae="language-"+X;K.alias?typeof K.alias=="string"?K.alias=[K.alias,ae]:K.alias.push(ae):K.alias=[ae]}}}P(O.tokens)}),o.hooks.add("wrap",function(O){if(O.type==="code-block"){for(var P="",D=0,$=O.classes.length;D<$;D++){var U=O.classes[D],B=/language-(.+)/.exec(U);if(B){P=B[1];break}}var G=o.languages[P];if(G)O.content=o.highlight(x(O.content.value),G,P);else if(P&&P!=="none"&&o.plugins.autoloader){var K="md-"+new Date().valueOf()+"-"+Math.floor(Math.random()*1e16);O.attributes.id=K,o.plugins.autoloader.loadLanguages(P,function(){var X=document.getElementById(K);X&&(X.innerHTML=o.highlight(X.textContent,o.languages[P],P))})}}});var y=RegExp(o.languages.markup.tag.pattern.source,"gi"),E={amp:"&",lt:"<",gt:">",quot:'"'},k=String.fromCodePoint||String.fromCharCode;function x(O){var P=O.replace(y,"");return P=P.replace(/&(\w{1,8}|#x?[\da-f]{1,8});/gi,function(D,$){if($=$.toLowerCase(),$[0]==="#"){var U;return $[1]==="x"?U=parseInt($.slice(2),16):U=Number($.slice(1)),k(U)}else{var B=E[$];return B||D}}),P}o.languages.md=o.languages.markdown})(r)}return Hze}var Gze,OGt;function TBr(){if(OGt)return Gze;OGt=1,Gze=n,n.displayName="matlab",n.aliases=[];function n(r){r.languages.matlab={comment:[/%\{[\s\S]*?\}%/,/%.+/],string:{pattern:/\B'(?:''|[^'\r\n])*'/,greedy:!0},number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[eE][+-]?\d+)?(?:[ij])?|\b[ij]\b/,keyword:/\b(?:NaN|break|case|catch|continue|else|elseif|end|for|function|if|inf|otherwise|parfor|pause|pi|return|switch|try|while)\b/,function:/\b(?!\d)\w+(?=\s*\()/,operator:/\.?[*^\/\\']|[+\-:@]|[<>=~]=?|&&?|\|\|?/,punctuation:/\.{3}|[.,;\[\](){}!]/}}return Gze}var qze,NGt;function xBr(){if(NGt)return qze;NGt=1,qze=n,n.displayName="maxscript",n.aliases=[];function n(r){(function(o){var s=/\b(?:about|and|animate|as|at|attributes|by|case|catch|collect|continue|coordsys|do|else|exit|fn|for|from|function|global|if|in|local|macroscript|mapped|max|not|of|off|on|or|parameters|persistent|plugin|rcmenu|return|rollout|set|struct|then|throw|to|tool|try|undo|utility|when|where|while|with)\b/i;o.languages.maxscript={comment:{pattern:/\/\*[\s\S]*?(?:\*\/|$)|--.*/,greedy:!0},string:{pattern:/(^|[^"\\@])(?:"(?:[^"\\]|\\[\s\S])*"|@"[^"]*")/,lookbehind:!0,greedy:!0},path:{pattern:/\$(?:[\w/\\.*?]|'[^']*')*/,greedy:!0,alias:"string"},"function-call":{pattern:RegExp("((?:"+(/^/.source+"|"+/[;=<>+\-*/^({\[]/.source+"|"+/\b(?:and|by|case|catch|collect|do|else|if|in|not|or|return|then|to|try|where|while|with)\b/.source)+")[ 	]*)(?!"+s.source+")"+/[a-z_]\w*\b/.source+"(?=[ 	]*(?:"+("(?!"+s.source+")"+/[a-z_]/.source+"|"+/\d|-\.?\d/.source+"|"+/[({'"$@#?]/.source)+"))","im"),lookbehind:!0,greedy:!0,alias:"function"},"function-definition":{pattern:/(\b(?:fn|function)\s+)\w+\b/i,lookbehind:!0,alias:"function"},argument:{pattern:/\b[a-z_]\w*(?=:)/i,alias:"attr-name"},keyword:s,boolean:/\b(?:false|true)\b/,time:{pattern:/(^|[^\w.])(?:(?:(?:\d+(?:\.\d*)?|\.\d+)(?:[eEdD][+-]\d+|[LP])?[msft])+|\d+:\d+(?:\.\d*)?)(?![\w.:])/,lookbehind:!0,alias:"number"},number:[{pattern:/(^|[^\w.])(?:(?:\d+(?:\.\d*)?|\.\d+)(?:[eEdD][+-]\d+|[LP])?|0x[a-fA-F0-9]+)(?![\w.:])/,lookbehind:!0},/\b(?:e|pi)\b/],constant:/\b(?:dontcollect|ok|silentValue|undefined|unsupplied)\b/,color:{pattern:/\b(?:black|blue|brown|gray|green|orange|red|white|yellow)\b/i,alias:"constant"},operator:/[-+*/<>=!]=?|[&^?]|#(?!\()/,punctuation:/[()\[\]{}.:,;]|#(?=\()|\\$/m}})(r)}return qze}var Vze,RGt;function CBr(){if(RGt)return Vze;RGt=1,Vze=n,n.displayName="mel",n.aliases=[];function n(r){r.languages.mel={comment:/\/\/.*/,code:{pattern:/`(?:\\.|[^\\`\r\n])*`/,greedy:!0,alias:"italic",inside:{delimiter:{pattern:/^`|`$/,alias:"punctuation"}}},string:{pattern:/"(?:\\.|[^\\"\r\n])*"/,greedy:!0},variable:/\$\w+/,number:/\b0x[\da-fA-F]+\b|\b\d+(?:\.\d*)?|\B\.\d+/,flag:{pattern:/-[^\d\W]\w*/,alias:"operator"},keyword:/\b(?:break|case|continue|default|do|else|float|for|global|if|in|int|matrix|proc|return|string|switch|vector|while)\b/,function:/\b\w+(?=\()|\b(?:CBG|HfAddAttractorToAS|HfAssignAS|HfBuildEqualMap|HfBuildFurFiles|HfBuildFurImages|HfCancelAFR|HfConnectASToHF|HfCreateAttractor|HfDeleteAS|HfEditAS|HfPerformCreateAS|HfRemoveAttractorFromAS|HfSelectAttached|HfSelectAttractors|HfUnAssignAS|Mayatomr|about|abs|addAttr|addAttributeEditorNodeHelp|addDynamic|addNewShelfTab|addPP|addPanelCategory|addPrefixToName|advanceToNextDrivenKey|affectedNet|affects|aimConstraint|air|alias|aliasAttr|align|alignCtx|alignCurve|alignSurface|allViewFit|ambientLight|angle|angleBetween|animCone|animCurveEditor|animDisplay|animView|annotate|appendStringArray|applicationName|applyAttrPreset|applyTake|arcLenDimContext|arcLengthDimension|arclen|arrayMapper|art3dPaintCtx|artAttrCtx|artAttrPaintVertexCtx|artAttrSkinPaintCtx|artAttrTool|artBuildPaintMenu|artFluidAttrCtx|artPuttyCtx|artSelectCtx|artSetPaintCtx|artUserPaintCtx|assignCommand|assignInputDevice|assignViewportFactories|attachCurve|attachDeviceAttr|attachSurface|attrColorSliderGrp|attrCompatibility|attrControlGrp|attrEnumOptionMenu|attrEnumOptionMenuGrp|attrFieldGrp|attrFieldSliderGrp|attrNavigationControlGrp|attrPresetEditWin|attributeExists|attributeInfo|attributeMenu|attributeQuery|autoKeyframe|autoPlace|bakeClip|bakeFluidShading|bakePartialHistory|bakeResults|bakeSimulation|basename|basenameEx|batchRender|bessel|bevel|bevelPlus|binMembership|bindSkin|blend2|blendShape|blendShapeEditor|blendShapePanel|blendTwoAttr|blindDataType|boneLattice|boundary|boxDollyCtx|boxZoomCtx|bufferCurve|buildBookmarkMenu|buildKeyframeMenu|button|buttonManip|cacheFile|cacheFileCombine|cacheFileMerge|cacheFileTrack|camera|cameraView|canCreateManip|canvas|capitalizeString|catch|catchQuiet|ceil|changeSubdivComponentDisplayLevel|changeSubdivRegion|channelBox|character|characterMap|characterOutlineEditor|characterize|chdir|checkBox|checkBoxGrp|checkDefaultRenderGlobals|choice|circle|circularFillet|clamp|clear|clearCache|clip|clipEditor|clipEditorCurrentTimeCtx|clipSchedule|clipSchedulerOutliner|clipTrimBefore|closeCurve|closeSurface|cluster|cmdFileOutput|cmdScrollFieldExecuter|cmdScrollFieldReporter|cmdShell|coarsenSubdivSelectionList|collision|color|colorAtPoint|colorEditor|colorIndex|colorIndexSliderGrp|colorSliderButtonGrp|colorSliderGrp|columnLayout|commandEcho|commandLine|commandPort|compactHairSystem|componentEditor|compositingInterop|computePolysetVolume|condition|cone|confirmDialog|connectAttr|connectControl|connectDynamic|connectJoint|connectionInfo|constrain|constrainValue|constructionHistory|container|containsMultibyte|contextInfo|control|convertFromOldLayers|convertIffToPsd|convertLightmap|convertSolidTx|convertTessellation|convertUnit|copyArray|copyFlexor|copyKey|copySkinWeights|cos|cpButton|cpCache|cpClothSet|cpCollision|cpConstraint|cpConvClothToMesh|cpForces|cpGetSolverAttr|cpPanel|cpProperty|cpRigidCollisionFilter|cpSeam|cpSetEdit|cpSetSolverAttr|cpSolver|cpSolverTypes|cpTool|cpUpdateClothUVs|createDisplayLayer|createDrawCtx|createEditor|createLayeredPsdFile|createMotionField|createNewShelf|createNode|createRenderLayer|createSubdivRegion|cross|crossProduct|ctxAbort|ctxCompletion|ctxEditMode|ctxTraverse|currentCtx|currentTime|currentTimeCtx|currentUnit|curve|curveAddPtCtx|curveCVCtx|curveEPCtx|curveEditorCtx|curveIntersect|curveMoveEPCtx|curveOnSurface|curveSketchCtx|cutKey|cycleCheck|cylinder|dagPose|date|defaultLightListCheckBox|defaultNavigation|defineDataServer|defineVirtualDevice|deformer|deg_to_rad|delete|deleteAttr|deleteShadingGroupsAndMaterials|deleteShelfTab|deleteUI|deleteUnusedBrushes|delrandstr|detachCurve|detachDeviceAttr|detachSurface|deviceEditor|devicePanel|dgInfo|dgdirty|dgeval|dgtimer|dimWhen|directKeyCtx|directionalLight|dirmap|dirname|disable|disconnectAttr|disconnectJoint|diskCache|displacementToPoly|displayAffected|displayColor|displayCull|displayLevelOfDetail|displayPref|displayRGBColor|displaySmoothness|displayStats|displayString|displaySurface|distanceDimContext|distanceDimension|doBlur|dolly|dollyCtx|dopeSheetEditor|dot|dotProduct|doubleProfileBirailSurface|drag|dragAttrContext|draggerContext|dropoffLocator|duplicate|duplicateCurve|duplicateSurface|dynCache|dynControl|dynExport|dynExpression|dynGlobals|dynPaintEditor|dynParticleCtx|dynPref|dynRelEdPanel|dynRelEditor|dynamicLoad|editAttrLimits|editDisplayLayerGlobals|editDisplayLayerMembers|editRenderLayerAdjustment|editRenderLayerGlobals|editRenderLayerMembers|editor|editorTemplate|effector|emit|emitter|enableDevice|encodeString|endString|endsWith|env|equivalent|equivalentTol|erf|error|eval|evalDeferred|evalEcho|event|exactWorldBoundingBox|exclusiveLightCheckBox|exec|executeForEachObject|exists|exp|expression|expressionEditorListen|extendCurve|extendSurface|extrude|fcheck|fclose|feof|fflush|fgetline|fgetword|file|fileBrowserDialog|fileDialog|fileExtension|fileInfo|filetest|filletCurve|filter|filterCurve|filterExpand|filterStudioImport|findAllIntersections|findAnimCurves|findKeyframe|findMenuItem|findRelatedSkinCluster|finder|firstParentOf|fitBspline|flexor|floatEq|floatField|floatFieldGrp|floatScrollBar|floatSlider|floatSlider2|floatSliderButtonGrp|floatSliderGrp|floor|flow|fluidCacheInfo|fluidEmitter|fluidVoxelInfo|flushUndo|fmod|fontDialog|fopen|formLayout|format|fprint|frameLayout|fread|freeFormFillet|frewind|fromNativePath|fwrite|gamma|gauss|geometryConstraint|getApplicationVersionAsFloat|getAttr|getClassification|getDefaultBrush|getFileList|getFluidAttr|getInputDeviceRange|getMayaPanelTypes|getModifiers|getPanel|getParticleAttr|getPluginResource|getenv|getpid|glRender|glRenderEditor|globalStitch|gmatch|goal|gotoBindPose|grabColor|gradientControl|gradientControlNoAttr|graphDollyCtx|graphSelectContext|graphTrackCtx|gravity|grid|gridLayout|group|groupObjectsByName|hardenPointCurve|hardware|hardwareRenderPanel|headsUpDisplay|headsUpMessage|help|helpLine|hermite|hide|hilite|hitTest|hotBox|hotkey|hotkeyCheck|hsv_to_rgb|hudButton|hudSlider|hudSliderButton|hwReflectionMap|hwRender|hwRenderLoad|hyperGraph|hyperPanel|hyperShade|hypot|iconTextButton|iconTextCheckBox|iconTextRadioButton|iconTextRadioCollection|iconTextScrollList|iconTextStaticLabel|ikHandle|ikHandleCtx|ikHandleDisplayScale|ikSolver|ikSplineHandleCtx|ikSystem|ikSystemInfo|ikfkDisplayMethod|illustratorCurves|image|imfPlugins|inheritTransform|insertJoint|insertJointCtx|insertKeyCtx|insertKnotCurve|insertKnotSurface|instance|instanceable|instancer|intField|intFieldGrp|intScrollBar|intSlider|intSliderGrp|interToUI|internalVar|intersect|iprEngine|isAnimCurve|isConnected|isDirty|isParentOf|isSameObject|isTrue|isValidObjectName|isValidString|isValidUiName|isolateSelect|itemFilter|itemFilterAttr|itemFilterRender|itemFilterType|joint|jointCluster|jointCtx|jointDisplayScale|jointLattice|keyTangent|keyframe|keyframeOutliner|keyframeRegionCurrentTimeCtx|keyframeRegionDirectKeyCtx|keyframeRegionDollyCtx|keyframeRegionInsertKeyCtx|keyframeRegionMoveKeyCtx|keyframeRegionScaleKeyCtx|keyframeRegionSelectKeyCtx|keyframeRegionSetKeyCtx|keyframeRegionTrackCtx|keyframeStats|lassoContext|lattice|latticeDeformKeyCtx|launch|launchImageEditor|layerButton|layeredShaderPort|layeredTexturePort|layout|layoutDialog|lightList|lightListEditor|lightListPanel|lightlink|lineIntersection|linearPrecision|linstep|listAnimatable|listAttr|listCameras|listConnections|listDeviceAttachments|listHistory|listInputDeviceAxes|listInputDeviceButtons|listInputDevices|listMenuAnnotation|listNodeTypes|listPanelCategories|listRelatives|listSets|listTransforms|listUnselected|listerEditor|loadFluid|loadNewShelf|loadPlugin|loadPluginLanguageResources|loadPrefObjects|localizedPanelLabel|lockNode|loft|log|longNameOf|lookThru|ls|lsThroughFilter|lsType|lsUI|mag|makeIdentity|makeLive|makePaintable|makeRoll|makeSingleSurface|makeTubeOn|makebot|manipMoveContext|manipMoveLimitsCtx|manipOptions|manipRotateContext|manipRotateLimitsCtx|manipScaleContext|manipScaleLimitsCtx|marker|match|max|memory|menu|menuBarLayout|menuEditor|menuItem|menuItemToShelf|menuSet|menuSetPref|messageLine|min|minimizeApp|mirrorJoint|modelCurrentTimeCtx|modelEditor|modelPanel|mouse|movIn|movOut|move|moveIKtoFK|moveKeyCtx|moveVertexAlongDirection|multiProfileBirailSurface|mute|nParticle|nameCommand|nameField|namespace|namespaceInfo|newPanelItems|newton|nodeCast|nodeIconButton|nodeOutliner|nodePreset|nodeType|noise|nonLinear|normalConstraint|normalize|nurbsBoolean|nurbsCopyUVSet|nurbsCube|nurbsEditUV|nurbsPlane|nurbsSelect|nurbsSquare|nurbsToPoly|nurbsToPolygonsPref|nurbsToSubdiv|nurbsToSubdivPref|nurbsUVSet|nurbsViewDirectionVector|objExists|objectCenter|objectLayer|objectType|objectTypeUI|obsoleteProc|oceanNurbsPreviewPlane|offsetCurve|offsetCurveOnSurface|offsetSurface|openGLExtension|openMayaPref|optionMenu|optionMenuGrp|optionVar|orbit|orbitCtx|orientConstraint|outlinerEditor|outlinerPanel|overrideModifier|paintEffectsDisplay|pairBlend|palettePort|paneLayout|panel|panelConfiguration|panelHistory|paramDimContext|paramDimension|paramLocator|parent|parentConstraint|particle|particleExists|particleInstancer|particleRenderInfo|partition|pasteKey|pathAnimation|pause|pclose|percent|performanceOptions|pfxstrokes|pickWalk|picture|pixelMove|planarSrf|plane|play|playbackOptions|playblast|plugAttr|plugNode|pluginInfo|pluginResourceUtil|pointConstraint|pointCurveConstraint|pointLight|pointMatrixMult|pointOnCurve|pointOnSurface|pointPosition|poleVectorConstraint|polyAppend|polyAppendFacetCtx|polyAppendVertex|polyAutoProjection|polyAverageNormal|polyAverageVertex|polyBevel|polyBlendColor|polyBlindData|polyBoolOp|polyBridgeEdge|polyCacheMonitor|polyCheck|polyChipOff|polyClipboard|polyCloseBorder|polyCollapseEdge|polyCollapseFacet|polyColorBlindData|polyColorDel|polyColorPerVertex|polyColorSet|polyCompare|polyCone|polyCopyUV|polyCrease|polyCreaseCtx|polyCreateFacet|polyCreateFacetCtx|polyCube|polyCut|polyCutCtx|polyCylinder|polyCylindricalProjection|polyDelEdge|polyDelFacet|polyDelVertex|polyDuplicateAndConnect|polyDuplicateEdge|polyEditUV|polyEditUVShell|polyEvaluate|polyExtrudeEdge|polyExtrudeFacet|polyExtrudeVertex|polyFlipEdge|polyFlipUV|polyForceUV|polyGeoSampler|polyHelix|polyInfo|polyInstallAction|polyLayoutUV|polyListComponentConversion|polyMapCut|polyMapDel|polyMapSew|polyMapSewMove|polyMergeEdge|polyMergeEdgeCtx|polyMergeFacet|polyMergeFacetCtx|polyMergeUV|polyMergeVertex|polyMirrorFace|polyMoveEdge|polyMoveFacet|polyMoveFacetUV|polyMoveUV|polyMoveVertex|polyNormal|polyNormalPerVertex|polyNormalizeUV|polyOptUvs|polyOptions|polyOutput|polyPipe|polyPlanarProjection|polyPlane|polyPlatonicSolid|polyPoke|polyPrimitive|polyPrism|polyProjection|polyPyramid|polyQuad|polyQueryBlindData|polyReduce|polySelect|polySelectConstraint|polySelectConstraintMonitor|polySelectCtx|polySelectEditCtx|polySeparate|polySetToFaceNormal|polySewEdge|polyShortestPathCtx|polySmooth|polySoftEdge|polySphere|polySphericalProjection|polySplit|polySplitCtx|polySplitEdge|polySplitRing|polySplitVertex|polyStraightenUVBorder|polySubdivideEdge|polySubdivideFacet|polyToSubdiv|polyTorus|polyTransfer|polyTriangulate|polyUVSet|polyUnite|polyWedgeFace|popen|popupMenu|pose|pow|preloadRefEd|print|progressBar|progressWindow|projFileViewer|projectCurve|projectTangent|projectionContext|projectionManip|promptDialog|propModCtx|propMove|psdChannelOutliner|psdEditTextureFile|psdExport|psdTextureFile|putenv|pwd|python|querySubdiv|quit|rad_to_deg|radial|radioButton|radioButtonGrp|radioCollection|radioMenuItemCollection|rampColorPort|rand|randomizeFollicles|randstate|rangeControl|readTake|rebuildCurve|rebuildSurface|recordAttr|recordDevice|redo|reference|referenceEdit|referenceQuery|refineSubdivSelectionList|refresh|refreshAE|registerPluginResource|rehash|reloadImage|removeJoint|removeMultiInstance|removePanelCategory|rename|renameAttr|renameSelectionList|renameUI|render|renderGlobalsNode|renderInfo|renderLayerButton|renderLayerParent|renderLayerPostProcess|renderLayerUnparent|renderManip|renderPartition|renderQualityNode|renderSettings|renderThumbnailUpdate|renderWindowEditor|renderWindowSelectContext|renderer|reorder|reorderDeformers|requires|reroot|resampleFluid|resetAE|resetPfxToPolyCamera|resetTool|resolutionNode|retarget|reverseCurve|reverseSurface|revolve|rgb_to_hsv|rigidBody|rigidSolver|roll|rollCtx|rootOf|rot|rotate|rotationInterpolation|roundConstantRadius|rowColumnLayout|rowLayout|runTimeCommand|runup|sampleImage|saveAllShelves|saveAttrPreset|saveFluid|saveImage|saveInitialState|saveMenu|savePrefObjects|savePrefs|saveShelf|saveToolSettings|scale|scaleBrushBrightness|scaleComponents|scaleConstraint|scaleKey|scaleKeyCtx|sceneEditor|sceneUIReplacement|scmh|scriptCtx|scriptEditorInfo|scriptJob|scriptNode|scriptTable|scriptToShelf|scriptedPanel|scriptedPanelType|scrollField|scrollLayout|sculpt|searchPathArray|seed|selLoadSettings|select|selectContext|selectCurveCV|selectKey|selectKeyCtx|selectKeyframeRegionCtx|selectMode|selectPref|selectPriority|selectType|selectedNodes|selectionConnection|separator|setAttr|setAttrEnumResource|setAttrMapping|setAttrNiceNameResource|setConstraintRestPosition|setDefaultShadingGroup|setDrivenKeyframe|setDynamic|setEditCtx|setEditor|setFluidAttr|setFocus|setInfinity|setInputDeviceMapping|setKeyCtx|setKeyPath|setKeyframe|setKeyframeBlendshapeTargetWts|setMenuMode|setNodeNiceNameResource|setNodeTypeFlag|setParent|setParticleAttr|setPfxToPolyCamera|setPluginResource|setProject|setStampDensity|setStartupMessage|setState|setToolTo|setUITemplate|setXformManip|sets|shadingConnection|shadingGeometryRelCtx|shadingLightRelCtx|shadingNetworkCompare|shadingNode|shapeCompare|shelfButton|shelfLayout|shelfTabLayout|shellField|shortNameOf|showHelp|showHidden|showManipCtx|showSelectionInTitle|showShadingGroupAttrEditor|showWindow|sign|simplify|sin|singleProfileBirailSurface|size|sizeBytes|skinCluster|skinPercent|smoothCurve|smoothTangentSurface|smoothstep|snap2to2|snapKey|snapMode|snapTogetherCtx|snapshot|soft|softMod|softModCtx|sort|sound|soundControl|source|spaceLocator|sphere|sphrand|spotLight|spotLightPreviewPort|spreadSheetEditor|spring|sqrt|squareSurface|srtContext|stackTrace|startString|startsWith|stitchAndExplodeShell|stitchSurface|stitchSurfacePoints|strcmp|stringArrayCatenate|stringArrayContains|stringArrayCount|stringArrayInsertAtIndex|stringArrayIntersector|stringArrayRemove|stringArrayRemoveAtIndex|stringArrayRemoveDuplicates|stringArrayRemoveExact|stringArrayToString|stringToStringArray|strip|stripPrefixFromName|stroke|subdAutoProjection|subdCleanTopology|subdCollapse|subdDuplicateAndConnect|subdEditUV|subdListComponentConversion|subdMapCut|subdMapSewMove|subdMatchTopology|subdMirror|subdToBlind|subdToPoly|subdTransferUVsToCache|subdiv|subdivCrease|subdivDisplaySmoothness|substitute|substituteAllString|substituteGeometry|substring|surface|surfaceSampler|surfaceShaderList|swatchDisplayPort|switchTable|symbolButton|symbolCheckBox|sysFile|system|tabLayout|tan|tangentConstraint|texLatticeDeformContext|texManipContext|texMoveContext|texMoveUVShellContext|texRotateContext|texScaleContext|texSelectContext|texSelectShortestPathCtx|texSmudgeUVContext|texWinToolCtx|text|textCurves|textField|textFieldButtonGrp|textFieldGrp|textManip|textScrollList|textToShelf|textureDisplacePlane|textureHairColor|texturePlacementContext|textureWindow|threadCount|threePointArcCtx|timeControl|timePort|timerX|toNativePath|toggle|toggleAxis|toggleWindowVisibility|tokenize|tokenizeList|tolerance|tolower|toolButton|toolCollection|toolDropped|toolHasOptions|toolPropertyWindow|torus|toupper|trace|track|trackCtx|transferAttributes|transformCompare|transformLimits|translator|trim|trunc|truncateFluidCache|truncateHairCache|tumble|tumbleCtx|turbulence|twoPointArcCtx|uiRes|uiTemplate|unassignInputDevice|undo|undoInfo|ungroup|uniform|unit|unloadPlugin|untangleUV|untitledFileName|untrim|upAxis|updateAE|userCtx|uvLink|uvSnapshot|validateShelfName|vectorize|view2dToolCtx|viewCamera|viewClipPlane|viewFit|viewHeadOn|viewLookAt|viewManip|viewPlace|viewSet|visor|volumeAxis|vortex|waitCursor|warning|webBrowser|webBrowserPrefs|whatIs|window|windowPref|wire|wireContext|workspace|wrinkle|wrinkleContext|writeTake|xbmLangPathList|xform)\b/,operator:[/\+[+=]?|-[-=]?|&&|\|\||[<>]=|[*\/!=]=?|[%^]/,{pattern:/(^|[^<])<(?!<)/,lookbehind:!0},{pattern:/(^|[^>])>(?!>)/,lookbehind:!0}],punctuation:/<<|>>|[.,:;?\[\](){}]/},r.languages.mel.code.inside.rest=r.languages.mel}return Vze}var Wze,PGt;function ABr(){if(PGt)return Wze;PGt=1,Wze=n,n.displayName="mermaid",n.aliases=[];function n(r){r.languages.mermaid={comment:{pattern:/%%.*/,greedy:!0},style:{pattern:/^([ \t]*(?:classDef|linkStyle|style)[ \t]+[\w$-]+[ \t]+)\w.*[^\s;]/m,lookbehind:!0,inside:{property:/\b\w[\w-]*(?=[ \t]*:)/,operator:/:/,punctuation:/,/}},"inter-arrow-label":{pattern:/([^<>ox.=-])(?:-[-.]|==)(?![<>ox.=-])[ \t]*(?:"[^"\r\n]*"|[^\s".=-](?:[^\r\n.=-]*[^\s.=-])?)[ \t]*(?:\.+->?|--+[->]|==+[=>])(?![<>ox.=-])/,lookbehind:!0,greedy:!0,inside:{arrow:{pattern:/(?:\.+->?|--+[->]|==+[=>])$/,alias:"operator"},label:{pattern:/^([\s\S]{2}[ \t]*)\S(?:[\s\S]*\S)?/,lookbehind:!0,alias:"property"},"arrow-head":{pattern:/^\S+/,alias:["arrow","operator"]}}},arrow:[{pattern:/(^|[^{}|o.-])[|}][|o](?:--|\.\.)[|o][|{](?![{}|o.-])/,lookbehind:!0,alias:"operator"},{pattern:/(^|[^<>ox.=-])(?:[<ox](?:==+|--+|-\.*-)[>ox]?|(?:==+|--+|-\.*-)[>ox]|===+|---+|-\.+-)(?![<>ox.=-])/,lookbehind:!0,alias:"operator"},{pattern:/(^|[^<>()x-])(?:--?(?:>>|[x>)])(?![<>()x])|(?:<<|[x<(])--?(?!-))/,lookbehind:!0,alias:"operator"},{pattern:/(^|[^<>|*o.-])(?:[*o]--|--[*o]|<\|?(?:--|\.\.)|(?:--|\.\.)\|?>|--|\.\.)(?![<>|*o.-])/,lookbehind:!0,alias:"operator"}],label:{pattern:/(^|[^|<])\|(?:[^\r\n"|]|"[^"\r\n]*")+\|/,lookbehind:!0,greedy:!0,alias:"property"},text:{pattern:/(?:[(\[{]+|\b>)(?:[^\r\n"()\[\]{}]|"[^"\r\n]*")+(?:[)\]}]+|>)/,alias:"string"},string:{pattern:/"[^"\r\n]*"/,greedy:!0},annotation:{pattern:/<<(?:abstract|choice|enumeration|fork|interface|join|service)>>|\[\[(?:choice|fork|join)\]\]/i,alias:"important"},keyword:[{pattern:/(^[ \t]*)(?:action|callback|class|classDef|classDiagram|click|direction|erDiagram|flowchart|gantt|gitGraph|graph|journey|link|linkStyle|pie|requirementDiagram|sequenceDiagram|stateDiagram|stateDiagram-v2|style|subgraph)(?![\w$-])/m,lookbehind:!0,greedy:!0},{pattern:/(^[ \t]*)(?:activate|alt|and|as|autonumber|deactivate|else|end(?:[ \t]+note)?|loop|opt|par|participant|rect|state|note[ \t]+(?:over|(?:left|right)[ \t]+of))(?![\w$-])/im,lookbehind:!0,greedy:!0}],entity:/#[a-z0-9]+;/,operator:{pattern:/(\w[ \t]*)&(?=[ \t]*\w)|:::|:/,lookbehind:!0},punctuation:/[(){};]/}}return Wze}var Kze,LGt;function IBr(){if(LGt)return Kze;LGt=1,Kze=n,n.displayName="mizar",n.aliases=[];function n(r){r.languages.mizar={comment:/::.+/,keyword:/@proof\b|\b(?:according|aggregate|all|and|antonym|are|as|associativity|assume|asymmetry|attr|be|begin|being|by|canceled|case|cases|clusters?|coherence|commutativity|compatibility|connectedness|consider|consistency|constructors|contradiction|correctness|def|deffunc|define|definitions?|defpred|do|does|end|environ|equals|ex|exactly|existence|for|from|func|given|hence|hereby|holds|idempotence|identity|iff?|implies|involutiveness|irreflexivity|is|it|let|means|mode|non|not|notations?|now|of|or|otherwise|over|per|pred|prefix|projectivity|proof|provided|qua|reconsider|redefine|reduce|reducibility|reflexivity|registrations?|requirements|reserve|sch|schemes?|section|selector|set|sethood|st|struct|such|suppose|symmetry|synonym|take|that|the|then|theorems?|thesis|thus|to|transitivity|uniqueness|vocabular(?:ies|y)|when|where|with|wrt)\b/,parameter:{pattern:/\$(?:10|\d)/,alias:"variable"},variable:/\b\w+(?=:)/,number:/(?:\b|-)\d+\b/,operator:/\.\.\.|->|&|\.?=/,punctuation:/\(#|#\)|[,:;\[\](){}]/}}return Kze}var Yze,DGt;function OBr(){if(DGt)return Yze;DGt=1,Yze=n,n.displayName="mongodb",n.aliases=[];function n(r){(function(o){var s=["$eq","$gt","$gte","$in","$lt","$lte","$ne","$nin","$and","$not","$nor","$or","$exists","$type","$expr","$jsonSchema","$mod","$regex","$text","$where","$geoIntersects","$geoWithin","$near","$nearSphere","$all","$elemMatch","$size","$bitsAllClear","$bitsAllSet","$bitsAnyClear","$bitsAnySet","$comment","$elemMatch","$meta","$slice","$currentDate","$inc","$min","$max","$mul","$rename","$set","$setOnInsert","$unset","$addToSet","$pop","$pull","$push","$pullAll","$each","$position","$slice","$sort","$bit","$addFields","$bucket","$bucketAuto","$collStats","$count","$currentOp","$facet","$geoNear","$graphLookup","$group","$indexStats","$limit","$listLocalSessions","$listSessions","$lookup","$match","$merge","$out","$planCacheStats","$project","$redact","$replaceRoot","$replaceWith","$sample","$set","$skip","$sort","$sortByCount","$unionWith","$unset","$unwind","$setWindowFields","$abs","$accumulator","$acos","$acosh","$add","$addToSet","$allElementsTrue","$and","$anyElementTrue","$arrayElemAt","$arrayToObject","$asin","$asinh","$atan","$atan2","$atanh","$avg","$binarySize","$bsonSize","$ceil","$cmp","$concat","$concatArrays","$cond","$convert","$cos","$dateFromParts","$dateToParts","$dateFromString","$dateToString","$dayOfMonth","$dayOfWeek","$dayOfYear","$degreesToRadians","$divide","$eq","$exp","$filter","$first","$floor","$function","$gt","$gte","$hour","$ifNull","$in","$indexOfArray","$indexOfBytes","$indexOfCP","$isArray","$isNumber","$isoDayOfWeek","$isoWeek","$isoWeekYear","$last","$last","$let","$literal","$ln","$log","$log10","$lt","$lte","$ltrim","$map","$max","$mergeObjects","$meta","$min","$millisecond","$minute","$mod","$month","$multiply","$ne","$not","$objectToArray","$or","$pow","$push","$radiansToDegrees","$range","$reduce","$regexFind","$regexFindAll","$regexMatch","$replaceOne","$replaceAll","$reverseArray","$round","$rtrim","$second","$setDifference","$setEquals","$setIntersection","$setIsSubset","$setUnion","$size","$sin","$slice","$split","$sqrt","$stdDevPop","$stdDevSamp","$strcasecmp","$strLenBytes","$strLenCP","$substr","$substrBytes","$substrCP","$subtract","$sum","$switch","$tan","$toBool","$toDate","$toDecimal","$toDouble","$toInt","$toLong","$toObjectId","$toString","$toLower","$toUpper","$trim","$trunc","$type","$week","$year","$zip","$count","$dateAdd","$dateDiff","$dateSubtract","$dateTrunc","$getField","$rand","$sampleRate","$setField","$unsetField","$comment","$explain","$hint","$max","$maxTimeMS","$min","$orderby","$query","$returnKey","$showDiskLoc","$natural"],c=["ObjectId","Code","BinData","DBRef","Timestamp","NumberLong","NumberDecimal","MaxKey","MinKey","RegExp","ISODate","UUID"];s=s.map(function(h){return h.replace("$","\\$")});var d="(?:"+s.join("|")+")\\b";o.languages.mongodb=o.languages.extend("javascript",{}),o.languages.insertBefore("mongodb","string",{property:{pattern:/(?:(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1|(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)(?=\s*:)/,greedy:!0,inside:{keyword:RegExp(`^(['"])?`+d+"(?:\\1)?$")}}}),o.languages.mongodb.string.inside={url:{pattern:/https?:\/\/[-\w@:%.+~#=]{1,256}\.[a-z0-9()]{1,6}\b[-\w()@:%+.~#?&/=]*/i,greedy:!0},entity:{pattern:/\b(?:(?:[01]?\d\d?|2[0-4]\d|25[0-5])\.){3}(?:[01]?\d\d?|2[0-4]\d|25[0-5])\b/,greedy:!0}},o.languages.insertBefore("mongodb","constant",{builtin:{pattern:RegExp("\\b(?:"+c.join("|")+")\\b"),alias:"keyword"}})})(r)}return Yze}var Xze,MGt;function NBr(){if(MGt)return Xze;MGt=1,Xze=n,n.displayName="monkey",n.aliases=[];function n(r){r.languages.monkey={comment:{pattern:/^#Rem\s[\s\S]*?^#End|'.+/im,greedy:!0},string:{pattern:/"[^"\r\n]*"/,greedy:!0},preprocessor:{pattern:/(^[ \t]*)#.+/m,lookbehind:!0,greedy:!0,alias:"property"},function:/\b\w+(?=\()/,"type-char":{pattern:/\b[?%#$]/,alias:"class-name"},number:{pattern:/((?:\.\.)?)(?:(?:\b|\B-\.?|\B\.)\d+(?:(?!\.\.)\.\d*)?|\$[\da-f]+)/i,lookbehind:!0},keyword:/\b(?:Abstract|Array|Bool|Case|Catch|Class|Const|Continue|Default|Eachin|Else|ElseIf|End|EndIf|Exit|Extends|Extern|False|Field|Final|Float|For|Forever|Function|Global|If|Implements|Import|Inline|Int|Interface|Local|Method|Module|New|Next|Null|Object|Private|Property|Public|Repeat|Return|Select|Self|Step|Strict|String|Super|Then|Throw|To|True|Try|Until|Void|Wend|While)\b/i,operator:/\.\.|<[=>]?|>=?|:?=|(?:[+\-*\/&~|]|\b(?:Mod|Shl|Shr)\b)=?|\b(?:And|Not|Or)\b/i,punctuation:/[.,:;()\[\]]/}}return Xze}var Jze,jGt;function RBr(){if(jGt)return Jze;jGt=1,Jze=n,n.displayName="moonscript",n.aliases=["moon"];function n(r){r.languages.moonscript={comment:/--.*/,string:[{pattern:/'[^']*'|\[(=*)\[[\s\S]*?\]\1\]/,greedy:!0},{pattern:/"[^"]*"/,greedy:!0,inside:{interpolation:{pattern:/#\{[^{}]*\}/,inside:{moonscript:{pattern:/(^#\{)[\s\S]+(?=\})/,lookbehind:!0,inside:null},"interpolation-punctuation":{pattern:/#\{|\}/,alias:"punctuation"}}}}}],"class-name":[{pattern:/(\b(?:class|extends)[ \t]+)\w+/,lookbehind:!0},/\b[A-Z]\w*/],keyword:/\b(?:class|continue|do|else|elseif|export|extends|for|from|if|import|in|local|nil|return|self|super|switch|then|unless|using|when|while|with)\b/,variable:/@@?\w*/,property:{pattern:/\b(?!\d)\w+(?=:)|(:)(?!\d)\w+/,lookbehind:!0},function:{pattern:/\b(?:_G|_VERSION|assert|collectgarbage|coroutine\.(?:create|resume|running|status|wrap|yield)|debug\.(?:debug|getfenv|gethook|getinfo|getlocal|getmetatable|getregistry|getupvalue|setfenv|sethook|setlocal|setmetatable|setupvalue|traceback)|dofile|error|getfenv|getmetatable|io\.(?:close|flush|input|lines|open|output|popen|read|stderr|stdin|stdout|tmpfile|type|write)|ipairs|load|loadfile|loadstring|math\.(?:abs|acos|asin|atan|atan2|ceil|cos|cosh|deg|exp|floor|fmod|frexp|ldexp|log|log10|max|min|modf|pi|pow|rad|random|randomseed|sin|sinh|sqrt|tan|tanh)|module|next|os\.(?:clock|date|difftime|execute|exit|getenv|remove|rename|setlocale|time|tmpname)|package\.(?:cpath|loaded|loadlib|path|preload|seeall)|pairs|pcall|print|rawequal|rawget|rawset|require|select|setfenv|setmetatable|string\.(?:byte|char|dump|find|format|gmatch|gsub|len|lower|match|rep|reverse|sub|upper)|table\.(?:concat|insert|maxn|remove|sort)|tonumber|tostring|type|unpack|xpcall)\b/,inside:{punctuation:/\./}},boolean:/\b(?:false|true)\b/,number:/(?:\B\.\d+|\b\d+\.\d+|\b\d+(?=[eE]))(?:[eE][-+]?\d+)?\b|\b(?:0x[a-fA-F\d]+|\d+)(?:U?LL)?\b/,operator:/\.{3}|[-=]>|~=|(?:[-+*/%<>!=]|\.\.)=?|[:#^]|\b(?:and|or)\b=?|\b(?:not)\b/,punctuation:/[.,()[\]{}\\]/},r.languages.moonscript.string[1].inside.interpolation.inside.moonscript.inside=r.languages.moonscript,r.languages.moon=r.languages.moonscript}return Jze}var Zze,$Gt;function PBr(){if($Gt)return Zze;$Gt=1,Zze=n,n.displayName="n1ql",n.aliases=[];function n(r){r.languages.n1ql={comment:{pattern:/\/\*[\s\S]*?(?:$|\*\/)|--.*/,greedy:!0},string:{pattern:/(["'])(?:\\[\s\S]|(?!\1)[^\\]|\1\1)*\1/,greedy:!0},identifier:{pattern:/`(?:\\[\s\S]|[^\\`]|``)*`/,greedy:!0},parameter:/\$[\w.]+/,keyword:/\b(?:ADVISE|ALL|ALTER|ANALYZE|AS|ASC|AT|BEGIN|BINARY|BOOLEAN|BREAK|BUCKET|BUILD|BY|CALL|CAST|CLUSTER|COLLATE|COLLECTION|COMMIT|COMMITTED|CONNECT|CONTINUE|CORRELATE|CORRELATED|COVER|CREATE|CURRENT|DATABASE|DATASET|DATASTORE|DECLARE|DECREMENT|DELETE|DERIVED|DESC|DESCRIBE|DISTINCT|DO|DROP|EACH|ELEMENT|EXCEPT|EXCLUDE|EXECUTE|EXPLAIN|FETCH|FILTER|FLATTEN|FLUSH|FOLLOWING|FOR|FORCE|FROM|FTS|FUNCTION|GOLANG|GRANT|GROUP|GROUPS|GSI|HASH|HAVING|IF|IGNORE|ILIKE|INCLUDE|INCREMENT|INDEX|INFER|INLINE|INNER|INSERT|INTERSECT|INTO|IS|ISOLATION|JAVASCRIPT|JOIN|KEY|KEYS|KEYSPACE|KNOWN|LANGUAGE|LAST|LEFT|LET|LETTING|LEVEL|LIMIT|LSM|MAP|MAPPING|MATCHED|MATERIALIZED|MERGE|MINUS|MISSING|NAMESPACE|NEST|NL|NO|NTH_VALUE|NULL|NULLS|NUMBER|OBJECT|OFFSET|ON|OPTION|OPTIONS|ORDER|OTHERS|OUTER|OVER|PARSE|PARTITION|PASSWORD|PATH|POOL|PRECEDING|PREPARE|PRIMARY|PRIVATE|PRIVILEGE|PROBE|PROCEDURE|PUBLIC|RANGE|RAW|REALM|REDUCE|RENAME|RESPECT|RETURN|RETURNING|REVOKE|RIGHT|ROLE|ROLLBACK|ROW|ROWS|SATISFIES|SAVEPOINT|SCHEMA|SCOPE|SELECT|SELF|SEMI|SET|SHOW|SOME|START|STATISTICS|STRING|SYSTEM|TIES|TO|TRAN|TRANSACTION|TRIGGER|TRUNCATE|UNBOUNDED|UNDER|UNION|UNIQUE|UNKNOWN|UNNEST|UNSET|UPDATE|UPSERT|USE|USER|USING|VALIDATE|VALUE|VALUES|VIA|VIEW|WHERE|WHILE|WINDOW|WITH|WORK|XOR)\b/i,function:/\b[a-z_]\w*(?=\s*\()/i,boolean:/\b(?:FALSE|TRUE)\b/i,number:/(?:\b\d+\.|\B\.)\d+e[+\-]?\d+\b|\b\d+(?:\.\d*)?|\B\.\d+\b/i,operator:/[-+*\/%]|!=|==?|\|\||<[>=]?|>=?|\b(?:AND|ANY|ARRAY|BETWEEN|CASE|ELSE|END|EVERY|EXISTS|FIRST|IN|LIKE|NOT|OR|THEN|VALUED|WHEN|WITHIN)\b/i,punctuation:/[;[\](),.{}:]/}}return Zze}var Qze,FGt;function LBr(){if(FGt)return Qze;FGt=1,Qze=n,n.displayName="n4js",n.aliases=["n4jsd"];function n(r){r.languages.n4js=r.languages.extend("javascript",{keyword:/\b(?:Array|any|boolean|break|case|catch|class|const|constructor|continue|debugger|declare|default|delete|do|else|enum|export|extends|false|finally|for|from|function|get|if|implements|import|in|instanceof|interface|let|module|new|null|number|package|private|protected|public|return|set|static|string|super|switch|this|throw|true|try|typeof|var|void|while|with|yield)\b/}),r.languages.insertBefore("n4js","constant",{annotation:{pattern:/@+\w+/,alias:"operator"}}),r.languages.n4jsd=r.languages.n4js}return Qze}var eHe,BGt;function DBr(){if(BGt)return eHe;BGt=1,eHe=n,n.displayName="nand2tetrisHdl",n.aliases=[];function n(r){r.languages["nand2tetris-hdl"]={comment:/\/\/.*|\/\*[\s\S]*?(?:\*\/|$)/,keyword:/\b(?:BUILTIN|CHIP|CLOCKED|IN|OUT|PARTS)\b/,boolean:/\b(?:false|true)\b/,function:/\b[A-Za-z][A-Za-z0-9]*(?=\()/,number:/\b\d+\b/,operator:/=|\.\./,punctuation:/[{}[\];(),:]/}}return eHe}var tHe,UGt;function MBr(){if(UGt)return tHe;UGt=1,tHe=n,n.displayName="naniscript",n.aliases=[];function n(r){(function(o){var s=/\{[^\r\n\[\]{}]*\}/,c={"quoted-string":{pattern:/"(?:[^"\\]|\\.)*"/,alias:"operator"},"command-param-id":{pattern:/(\s)\w+:/,lookbehind:!0,alias:"property"},"command-param-value":[{pattern:s,alias:"selector"},{pattern:/([\t ])\S+/,lookbehind:!0,greedy:!0,alias:"operator"},{pattern:/\S(?:.*\S)?/,alias:"operator"}]};o.languages.naniscript={comment:{pattern:/^([\t ]*);.*/m,lookbehind:!0},define:{pattern:/^>.+/m,alias:"tag",inside:{value:{pattern:/(^>\w+[\t ]+)(?!\s)[^{}\r\n]+/,lookbehind:!0,alias:"operator"},key:{pattern:/(^>)\w+/,lookbehind:!0}}},label:{pattern:/^([\t ]*)#[\t ]*\w+[\t ]*$/m,lookbehind:!0,alias:"regex"},command:{pattern:/^([\t ]*)@\w+(?=[\t ]|$).*/m,lookbehind:!0,alias:"function",inside:{"command-name":/^@\w+/,expression:{pattern:s,greedy:!0,alias:"selector"},"command-params":{pattern:/\s*\S[\s\S]*/,inside:c}}},"generic-text":{pattern:/(^[ \t]*)[^#@>;\s].*/m,lookbehind:!0,alias:"punctuation",inside:{"escaped-char":/\\[{}\[\]"]/,expression:{pattern:s,greedy:!0,alias:"selector"},"inline-command":{pattern:/\[[\t ]*\w[^\r\n\[\]]*\]/,greedy:!0,alias:"function",inside:{"command-params":{pattern:/(^\[[\t ]*\w+\b)[\s\S]+(?=\]$)/,lookbehind:!0,inside:c},"command-param-name":{pattern:/^(\[[\t ]*)\w+/,lookbehind:!0,alias:"name"},"start-stop-char":/[\[\]]/}}}}},o.languages.nani=o.languages.naniscript,o.hooks.add("after-tokenize",function(b){var y=b.tokens;y.forEach(function(E){if(typeof E!="string"&&E.type==="generic-text"){var k=h(E);d(k)||(E.type="bad-line",E.content=k)}})});function d(b){for(var y="[]{}",E=[],k=0;k<b.length;k++){var x=b[k],O=y.indexOf(x);if(O!==-1){if(O%2===0)E.push(O+1);else if(E.pop()!==O)return!1}}return E.length===0}function h(b){return typeof b=="string"?b:Array.isArray(b)?b.map(h).join(""):h(b.content)}})(r)}return tHe}var nHe,zGt;function jBr(){if(zGt)return nHe;zGt=1,nHe=n,n.displayName="nasm",n.aliases=[];function n(r){r.languages.nasm={comment:/;.*$/m,string:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,label:{pattern:/(^\s*)[A-Za-z._?$][\w.?$@~#]*:/m,lookbehind:!0,alias:"function"},keyword:[/\[?BITS (?:16|32|64)\]?/,{pattern:/(^\s*)section\s*[a-z.]+:?/im,lookbehind:!0},/(?:extern|global)[^;\r\n]*/i,/(?:CPU|DEFAULT|FLOAT).*$/m],register:{pattern:/\b(?:st\d|[xyz]mm\d\d?|[cdt]r\d|r\d\d?[bwd]?|[er]?[abcd]x|[abcd][hl]|[er]?(?:bp|di|si|sp)|[cdefgs]s)\b/i,alias:"variable"},number:/(?:\b|(?=\$))(?:0[hx](?:\.[\da-f]+|[\da-f]+(?:\.[\da-f]+)?)(?:p[+-]?\d+)?|\d[\da-f]+[hx]|\$\d[\da-f]*|0[oq][0-7]+|[0-7]+[oq]|0[by][01]+|[01]+[by]|0[dt]\d+|(?:\d+(?:\.\d+)?|\.\d+)(?:\.?e[+-]?\d+)?[dt]?)\b/i,operator:/[\[\]*+\-\/%<>=&|$!]/}}return nHe}var rHe,HGt;function $Br(){if(HGt)return rHe;HGt=1,rHe=n,n.displayName="neon",n.aliases=[];function n(r){r.languages.neon={comment:{pattern:/#.*/,greedy:!0},datetime:{pattern:/(^|[[{(=:,\s])\d\d\d\d-\d\d?-\d\d?(?:(?:[Tt]| +)\d\d?:\d\d:\d\d(?:\.\d*)? *(?:Z|[-+]\d\d?(?::?\d\d)?)?)?(?=$|[\]}),\s])/,lookbehind:!0,alias:"number"},key:{pattern:/(^|[[{(,\s])[^,:=[\]{}()'"\s]+(?=\s*:(?:$|[\]}),\s])|\s*=)/,lookbehind:!0,alias:"atrule"},number:{pattern:/(^|[[{(=:,\s])[+-]?(?:0x[\da-fA-F]+|0o[0-7]+|0b[01]+|(?:\d+(?:\.\d*)?|\.?\d+)(?:[eE][+-]?\d+)?)(?=$|[\]}),:=\s])/,lookbehind:!0},boolean:{pattern:/(^|[[{(=:,\s])(?:false|no|true|yes)(?=$|[\]}),:=\s])/i,lookbehind:!0},null:{pattern:/(^|[[{(=:,\s])(?:null)(?=$|[\]}),:=\s])/i,lookbehind:!0,alias:"keyword"},string:{pattern:/(^|[[{(=:,\s])(?:('''|""")\r?\n(?:(?:[^\r\n]|\r?\n(?![\t ]*\2))*\r?\n)?[\t ]*\2|'[^'\r\n]*'|"(?:\\.|[^\\"\r\n])*")/,lookbehind:!0,greedy:!0},literal:{pattern:/(^|[[{(=:,\s])(?:[^#"',:=[\]{}()\s`-]|[:-][^"',=[\]{}()\s])(?:[^,:=\]})(\s]|:(?![\s,\]})]|$)|[ \t]+[^#,:=\]})(\s])*/,lookbehind:!0,alias:"string"},punctuation:/[,:=[\]{}()-]/}}return rHe}var iHe,GGt;function FBr(){if(GGt)return iHe;GGt=1,iHe=n,n.displayName="nevod",n.aliases=[];function n(r){r.languages.nevod={comment:/\/\/.*|(?:\/\*[\s\S]*?(?:\*\/|$))/,string:{pattern:/(?:"(?:""|[^"])*"(?!")|'(?:''|[^'])*'(?!'))!?\*?/,greedy:!0,inside:{"string-attrs":/!$|!\*$|\*$/}},namespace:{pattern:/(@namespace\s+)[a-zA-Z0-9\-.]+(?=\s*\{)/,lookbehind:!0},pattern:{pattern:/(@pattern\s+)?#?[a-zA-Z0-9\-.]+(?:\s*\(\s*(?:~\s*)?[a-zA-Z0-9\-.]+\s*(?:,\s*(?:~\s*)?[a-zA-Z0-9\-.]*)*\))?(?=\s*=)/,lookbehind:!0,inside:{"pattern-name":{pattern:/^#?[a-zA-Z0-9\-.]+/,alias:"class-name"},fields:{pattern:/\(.*\)/,inside:{"field-name":{pattern:/[a-zA-Z0-9\-.]+/,alias:"variable"},punctuation:/[,()]/,operator:{pattern:/~/,alias:"field-hidden-mark"}}}}},search:{pattern:/(@search\s+|#)[a-zA-Z0-9\-.]+(?:\.\*)?(?=\s*;)/,alias:"function",lookbehind:!0},keyword:/@(?:having|inside|namespace|outside|pattern|require|search|where)\b/,"standard-pattern":{pattern:/\b(?:Alpha|AlphaNum|Any|Blank|End|LineBreak|Num|NumAlpha|Punct|Space|Start|Symbol|Word|WordBreak)\b(?:\([a-zA-Z0-9\-.,\s+]*\))?/,inside:{"standard-pattern-name":{pattern:/^[a-zA-Z0-9\-.]+/,alias:"builtin"},quantifier:{pattern:/\b\d+(?:\s*\+|\s*-\s*\d+)?(?!\w)/,alias:"number"},"standard-pattern-attr":{pattern:/[a-zA-Z0-9\-.]+/,alias:"builtin"},punctuation:/[,()]/}},quantifier:{pattern:/\b\d+(?:\s*\+|\s*-\s*\d+)?(?!\w)/,alias:"number"},operator:[{pattern:/=/,alias:"pattern-def"},{pattern:/&/,alias:"conjunction"},{pattern:/~/,alias:"exception"},{pattern:/\?/,alias:"optionality"},{pattern:/[[\]]/,alias:"repetition"},{pattern:/[{}]/,alias:"variation"},{pattern:/[+_]/,alias:"sequence"},{pattern:/\.{2,3}/,alias:"span"}],"field-capture":[{pattern:/([a-zA-Z0-9\-.]+\s*\()\s*[a-zA-Z0-9\-.]+\s*:\s*[a-zA-Z0-9\-.]+(?:\s*,\s*[a-zA-Z0-9\-.]+\s*:\s*[a-zA-Z0-9\-.]+)*(?=\s*\))/,lookbehind:!0,inside:{"field-name":{pattern:/[a-zA-Z0-9\-.]+/,alias:"variable"},colon:/:/}},{pattern:/[a-zA-Z0-9\-.]+\s*:/,inside:{"field-name":{pattern:/[a-zA-Z0-9\-.]+/,alias:"variable"},colon:/:/}}],punctuation:/[:;,()]/,name:/[a-zA-Z0-9\-.]+/}}return iHe}var oHe,qGt;function BBr(){if(qGt)return oHe;qGt=1,oHe=n,n.displayName="nginx",n.aliases=[];function n(r){(function(o){var s=/\$(?:\w[a-z\d]*(?:_[^\x00-\x1F\s"'\\()$]*)?|\{[^}\s"'\\]+\})/i;o.languages.nginx={comment:{pattern:/(^|[\s{};])#.*/,lookbehind:!0,greedy:!0},directive:{pattern:/(^|\s)\w(?:[^;{}"'\\\s]|\\.|"(?:[^"\\]|\\.)*"|'(?:[^'\\]|\\.)*'|\s+(?:#.*(?!.)|(?![#\s])))*?(?=\s*[;{])/,lookbehind:!0,greedy:!0,inside:{string:{pattern:/((?:^|[^\\])(?:\\\\)*)(?:"(?:[^"\\]|\\.)*"|'(?:[^'\\]|\\.)*')/,lookbehind:!0,greedy:!0,inside:{escape:{pattern:/\\["'\\nrt]/,alias:"entity"},variable:s}},comment:{pattern:/(\s)#.*/,lookbehind:!0,greedy:!0},keyword:{pattern:/^\S+/,greedy:!0},boolean:{pattern:/(\s)(?:off|on)(?!\S)/,lookbehind:!0},number:{pattern:/(\s)\d+[a-z]*(?!\S)/i,lookbehind:!0},variable:s}},punctuation:/[{};]/}})(r)}return oHe}var aHe,VGt;function UBr(){if(VGt)return aHe;VGt=1,aHe=n,n.displayName="nim",n.aliases=[];function n(r){r.languages.nim={comment:{pattern:/#.*/,greedy:!0},string:{pattern:/(?:\b(?!\d)(?:\w|\\x[89a-fA-F][0-9a-fA-F])+)?(?:"""[\s\S]*?"""(?!")|"(?:\\[\s\S]|""|[^"\\])*")/,greedy:!0},char:{pattern:/'(?:\\(?:\d+|x[\da-fA-F]{0,2}|.)|[^'])'/,greedy:!0},function:{pattern:/(?:(?!\d)(?:\w|\\x[89a-fA-F][0-9a-fA-F])+|`[^`\r\n]+`)\*?(?:\[[^\]]+\])?(?=\s*\()/,greedy:!0,inside:{operator:/\*$/}},identifier:{pattern:/`[^`\r\n]+`/,greedy:!0,inside:{punctuation:/`/}},number:/\b(?:0[xXoObB][\da-fA-F_]+|\d[\d_]*(?:(?!\.\.)\.[\d_]*)?(?:[eE][+-]?\d[\d_]*)?)(?:'?[iuf]\d*)?/,keyword:/\b(?:addr|as|asm|atomic|bind|block|break|case|cast|concept|const|continue|converter|defer|discard|distinct|do|elif|else|end|enum|except|export|finally|for|from|func|generic|if|import|include|interface|iterator|let|macro|method|mixin|nil|object|out|proc|ptr|raise|ref|return|static|template|try|tuple|type|using|var|when|while|with|without|yield)\b/,operator:{pattern:/(^|[({\[](?=\.\.)|(?![({\[]\.).)(?:(?:[=+\-*\/<>@$~&%|!?^:\\]|\.\.|\.(?![)}\]]))+|\b(?:and|div|in|is|isnot|mod|not|notin|of|or|shl|shr|xor)\b)/m,lookbehind:!0},punctuation:/[({\[]\.|\.[)}\]]|[`(){}\[\],:]/}}return aHe}var sHe,WGt;function zBr(){if(WGt)return sHe;WGt=1,sHe=n,n.displayName="nix",n.aliases=[];function n(r){r.languages.nix={comment:{pattern:/\/\*[\s\S]*?\*\/|#.*/,greedy:!0},string:{pattern:/"(?:[^"\\]|\\[\s\S])*"|''(?:(?!'')[\s\S]|''(?:'|\\|\$\{))*''/,greedy:!0,inside:{interpolation:{pattern:/(^|(?:^|(?!'').)[^\\])\$\{(?:[^{}]|\{[^}]*\})*\}/,lookbehind:!0,inside:null}}},url:[/\b(?:[a-z]{3,7}:\/\/)[\w\-+%~\/.:#=?&]+/,{pattern:/([^\/])(?:[\w\-+%~.:#=?&]*(?!\/\/)[\w\-+%~\/.:#=?&])?(?!\/\/)\/[\w\-+%~\/.:#=?&]*/,lookbehind:!0}],antiquotation:{pattern:/\$(?=\{)/,alias:"important"},number:/\b\d+\b/,keyword:/\b(?:assert|builtins|else|if|in|inherit|let|null|or|then|with)\b/,function:/\b(?:abort|add|all|any|attrNames|attrValues|baseNameOf|compareVersions|concatLists|currentSystem|deepSeq|derivation|dirOf|div|elem(?:At)?|fetch(?:Tarball|url)|filter(?:Source)?|fromJSON|genList|getAttr|getEnv|hasAttr|hashString|head|import|intersectAttrs|is(?:Attrs|Bool|Function|Int|List|Null|String)|length|lessThan|listToAttrs|map|mul|parseDrvName|pathExists|read(?:Dir|File)|removeAttrs|replaceStrings|seq|sort|stringLength|sub(?:string)?|tail|throw|to(?:File|JSON|Path|String|XML)|trace|typeOf)\b|\bfoldl'\B/,boolean:/\b(?:false|true)\b/,operator:/[=!<>]=?|\+\+?|\|\||&&|\/\/|->?|[?@]/,punctuation:/[{}()[\].,:;]/},r.languages.nix.string.inside.interpolation.inside=r.languages.nix}return sHe}var cHe,KGt;function HBr(){if(KGt)return cHe;KGt=1,cHe=n,n.displayName="nsis",n.aliases=[];function n(r){r.languages.nsis={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|[#;].*)/,lookbehind:!0,greedy:!0},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:{pattern:/(^[\t ]*)(?:Abort|Add(?:BrandingImage|Size)|AdvSplash|Allow(?:RootDirInstall|SkipFiles)|AutoCloseWindow|BG(?:Font|Gradient|Image)|Banner|BrandingText|BringToFront|CRCCheck|Call(?:InstDLL)?|Caption|ChangeUI|CheckBitmap|ClearErrors|CompletedText|ComponentText|CopyFiles|Create(?:Directory|Font|ShortCut)|Delete(?:INISec|INIStr|RegKey|RegValue)?|Detail(?:Print|sButtonText)|Dialer|Dir(?:Text|Var|Verify)|EnableWindow|Enum(?:RegKey|RegValue)|Exch|Exec(?:Shell(?:Wait)?|Wait)?|ExpandEnvStrings|File(?:BufSize|Close|ErrorText|Open|Read|ReadByte|ReadUTF16LE|ReadWord|Seek|Write|WriteByte|WriteUTF16LE|WriteWord)?|Find(?:Close|First|Next|Window)|FlushINI|Get(?:CurInstType|CurrentAddress|DLLVersion(?:Local)?|DlgItem|ErrorLevel|FileTime(?:Local)?|FullPathName|Function(?:Address|End)?|InstDirError|LabelAddress|TempFileName)|Goto|HideWindow|Icon|If(?:Abort|Errors|FileExists|RebootFlag|Silent)|InitPluginsDir|InstProgressFlags|Inst(?:Type(?:GetText|SetText)?)|Install(?:ButtonText|Colors|Dir(?:RegKey)?)|Int(?:64|Ptr)?CmpU?|Int(?:64)?Fmt|Int(?:Ptr)?Op|IsWindow|Lang(?:DLL|String)|License(?:BkColor|Data|ForceSelection|LangString|Text)|LoadLanguageFile|LockWindow|Log(?:Set|Text)|Manifest(?:DPIAware|SupportedOS)|Math|MessageBox|MiscButtonText|NSISdl|Name|Nop|OutFile|PE(?:DllCharacteristics|SubsysVer)|Page(?:Callbacks)?|Pop|Push|Quit|RMDir|Read(?:EnvStr|INIStr|RegDWORD|RegStr)|Reboot|RegDLL|Rename|RequestExecutionLevel|ReserveFile|Return|SearchPath|Section(?:End|GetFlags|GetInstTypes|GetSize|GetText|Group|In|SetFlags|SetInstTypes|SetSize|SetText)?|SendMessage|Set(?:AutoClose|BrandingImage|Compress|Compressor(?:DictSize)?|CtlColors|CurInstType|DatablockOptimize|DateSave|Details(?:Print|View)|ErrorLevel|Errors|FileAttributes|Font|OutPath|Overwrite|PluginUnload|RebootFlag|RegView|ShellVarContext|Silent)|Show(?:InstDetails|UninstDetails|Window)|Silent(?:Install|UnInstall)|Sleep|SpaceTexts|Splash|StartMenu|Str(?:CmpS?|Cpy|Len)|SubCaption|System|UnRegDLL|Unicode|UninstPage|Uninstall(?:ButtonText|Caption|Icon|SubCaption|Text)|UserInfo|VI(?:AddVersionKey|FileVersion|ProductVersion)|VPatch|Var|WindowIcon|Write(?:INIStr|Reg(?:Bin|DWORD|ExpandStr|MultiStr|None|Str)|Uninstaller)|XPStyle|ns(?:Dialogs|Exec))\b/m,lookbehind:!0},property:/\b(?:ARCHIVE|FILE_(?:ATTRIBUTE_ARCHIVE|ATTRIBUTE_NORMAL|ATTRIBUTE_OFFLINE|ATTRIBUTE_READONLY|ATTRIBUTE_SYSTEM|ATTRIBUTE_TEMPORARY)|HK(?:(?:CR|CU|LM)(?:32|64)?|DD|PD|U)|HKEY_(?:CLASSES_ROOT|CURRENT_CONFIG|CURRENT_USER|DYN_DATA|LOCAL_MACHINE|PERFORMANCE_DATA|USERS)|ID(?:ABORT|CANCEL|IGNORE|NO|OK|RETRY|YES)|MB_(?:ABORTRETRYIGNORE|DEFBUTTON1|DEFBUTTON2|DEFBUTTON3|DEFBUTTON4|ICONEXCLAMATION|ICONINFORMATION|ICONQUESTION|ICONSTOP|OK|OKCANCEL|RETRYCANCEL|RIGHT|RTLREADING|SETFOREGROUND|TOPMOST|USERICON|YESNO)|NORMAL|OFFLINE|READONLY|SHCTX|SHELL_CONTEXT|SYSTEM|TEMPORARY|admin|all|auto|both|colored|false|force|hide|highest|lastused|leave|listonly|none|normal|notset|off|on|open|print|show|silent|silentlog|smooth|textonly|true|user)\b/,constant:/\$\{[!\w\.:\^-]+\}|\$\([!\w\.:\^-]+\)/,variable:/\$\w[\w\.]*/,number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee]-?\d+)?/,operator:/--?|\+\+?|<=?|>=?|==?=?|&&?|\|\|?|[?*\/~^%]/,punctuation:/[{}[\];(),.:]/,important:{pattern:/(^[\t ]*)!(?:addincludedir|addplugindir|appendfile|cd|define|delfile|echo|else|endif|error|execute|finalize|getdllversion|gettlbversion|if|ifdef|ifmacrodef|ifmacrondef|ifndef|include|insertmacro|macro|macroend|makensis|packhdr|pragma|searchparse|searchreplace|system|tempfile|undef|verbose|warning)\b/im,lookbehind:!0}}}return cHe}var uHe,YGt;function GBr(){if(YGt)return uHe;YGt=1;var n=vU();uHe=r,r.displayName="objectivec",r.aliases=["objc"];function r(o){o.register(n),o.languages.objectivec=o.languages.extend("c",{string:{pattern:/@?"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0},keyword:/\b(?:asm|auto|break|case|char|const|continue|default|do|double|else|enum|extern|float|for|goto|if|in|inline|int|long|register|return|self|short|signed|sizeof|static|struct|super|switch|typedef|typeof|union|unsigned|void|volatile|while)\b|(?:@interface|@end|@implementation|@protocol|@class|@public|@protected|@private|@property|@try|@catch|@finally|@throw|@synthesize|@dynamic|@selector)\b/,operator:/-[->]?|\+\+?|!=?|<<?=?|>>?=?|==?|&&?|\|\|?|[~^%?*\/@]/}),delete o.languages.objectivec["class-name"],o.languages.objc=o.languages.objectivec}return uHe}var lHe,XGt;function qBr(){if(XGt)return lHe;XGt=1,lHe=n,n.displayName="ocaml",n.aliases=[];function n(r){r.languages.ocaml={comment:{pattern:/\(\*[\s\S]*?\*\)/,greedy:!0},char:{pattern:/'(?:[^\\\r\n']|\\(?:.|[ox]?[0-9a-f]{1,3}))'/i,greedy:!0},string:[{pattern:/"(?:\\(?:[\s\S]|\r\n)|[^\\\r\n"])*"/,greedy:!0},{pattern:/\{([a-z_]*)\|[\s\S]*?\|\1\}/,greedy:!0}],number:[/\b(?:0b[01][01_]*|0o[0-7][0-7_]*)\b/i,/\b0x[a-f0-9][a-f0-9_]*(?:\.[a-f0-9_]*)?(?:p[+-]?\d[\d_]*)?(?!\w)/i,/\b\d[\d_]*(?:\.[\d_]*)?(?:e[+-]?\d[\d_]*)?(?!\w)/i],directive:{pattern:/\B#\w+/,alias:"property"},label:{pattern:/\B~\w+/,alias:"property"},"type-variable":{pattern:/\B'\w+/,alias:"function"},variant:{pattern:/`\w+/,alias:"symbol"},keyword:/\b(?:as|assert|begin|class|constraint|do|done|downto|else|end|exception|external|for|fun|function|functor|if|in|include|inherit|initializer|lazy|let|match|method|module|mutable|new|nonrec|object|of|open|private|rec|sig|struct|then|to|try|type|val|value|virtual|when|where|while|with)\b/,boolean:/\b(?:false|true)\b/,"operator-like-punctuation":{pattern:/\[[<>|]|[>|]\]|\{<|>\}/,alias:"punctuation"},operator:/\.[.~]|:[=>]|[=<>@^|&+\-*\/$%!?~][!$%&*+\-.\/:<=>?@^|~]*|\b(?:and|asr|land|lor|lsl|lsr|lxor|mod|or)\b/,punctuation:/;;|::|[(){}\[\].,:;#]|\b_\b/}}return lHe}var dHe,JGt;function VBr(){if(JGt)return dHe;JGt=1;var n=vU();dHe=r,r.displayName="opencl",r.aliases=[];function r(o){o.register(n),function(s){s.languages.opencl=s.languages.extend("c",{keyword:/\b(?:(?:__)?(?:constant|global|kernel|local|private|read_only|read_write|write_only)|__attribute__|auto|(?:bool|u?(?:char|int|long|short)|half|quad)(?:2|3|4|8|16)?|break|case|complex|const|continue|(?:double|float)(?:16(?:x(?:1|2|4|8|16))?|1x(?:1|2|4|8|16)|2(?:x(?:1|2|4|8|16))?|3|4(?:x(?:1|2|4|8|16))?|8(?:x(?:1|2|4|8|16))?)?|default|do|else|enum|extern|for|goto|if|imaginary|inline|packed|pipe|register|restrict|return|signed|sizeof|static|struct|switch|typedef|uniform|union|unsigned|void|volatile|while)\b/,number:/(?:\b0x(?:[\da-f]+(?:\.[\da-f]*)?|\.[\da-f]+)(?:p[+-]?\d+)?|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)[fuhl]{0,4}/i,boolean:/\b(?:false|true)\b/,"constant-opencl-kernel":{pattern:/\b(?:CHAR_(?:BIT|MAX|MIN)|CLK_(?:ADDRESS_(?:CLAMP(?:_TO_EDGE)?|NONE|REPEAT)|FILTER_(?:LINEAR|NEAREST)|(?:GLOBAL|LOCAL)_MEM_FENCE|NORMALIZED_COORDS_(?:FALSE|TRUE))|CL_(?:BGRA|(?:HALF_)?FLOAT|INTENSITY|LUMINANCE|A?R?G?B?[Ax]?|(?:(?:UN)?SIGNED|[US]NORM)_(?:INT(?:8|16|32))|UNORM_(?:INT_101010|SHORT_(?:555|565)))|(?:DBL|FLT|HALF)_(?:DIG|EPSILON|(?:MAX|MIN)(?:(?:_10)?_EXP)?|MANT_DIG)|FLT_RADIX|HUGE_VALF?|(?:INT|LONG|SCHAR|SHRT)_(?:MAX|MIN)|INFINITY|MAXFLOAT|M_(?:[12]_PI|2_SQRTPI|E|LN(?:2|10)|LOG(?:2|10)E?|PI(?:_[24])?|SQRT(?:1_2|2))(?:_F|_H)?|NAN|(?:UCHAR|UINT|ULONG|USHRT)_MAX)\b/,alias:"constant"}}),s.languages.insertBefore("opencl","class-name",{"builtin-type":{pattern:/\b(?:_cl_(?:command_queue|context|device_id|event|kernel|mem|platform_id|program|sampler)|cl_(?:image_format|mem_fence_flags)|clk_event_t|event_t|image(?:1d_(?:array_|buffer_)?t|2d_(?:array_(?:depth_|msaa_depth_|msaa_)?|depth_|msaa_depth_|msaa_)?t|3d_t)|intptr_t|ndrange_t|ptrdiff_t|queue_t|reserve_id_t|sampler_t|size_t|uintptr_t)\b/,alias:"keyword"}});var c={"type-opencl-host":{pattern:/\b(?:cl_(?:GLenum|GLint|GLuin|addressing_mode|bitfield|bool|buffer_create_type|build_status|channel_(?:order|type)|(?:u?(?:char|int|long|short)|double|float)(?:2|3|4|8|16)?|command_(?:queue(?:_info|_properties)?|type)|context(?:_info|_properties)?|device_(?:exec_capabilities|fp_config|id|info|local_mem_type|mem_cache_type|type)|(?:event|sampler)(?:_info)?|filter_mode|half|image_info|kernel(?:_info|_work_group_info)?|map_flags|mem(?:_flags|_info|_object_type)?|platform_(?:id|info)|profiling_info|program(?:_build_info|_info)?))\b/,alias:"keyword"},"boolean-opencl-host":{pattern:/\bCL_(?:FALSE|TRUE)\b/,alias:"boolean"},"constant-opencl-host":{pattern:/\bCL_(?:A|ABGR|ADDRESS_(?:CLAMP(?:_TO_EDGE)?|MIRRORED_REPEAT|NONE|REPEAT)|ARGB|BGRA|BLOCKING|BUFFER_CREATE_TYPE_REGION|BUILD_(?:ERROR|IN_PROGRESS|NONE|PROGRAM_FAILURE|SUCCESS)|COMMAND_(?:ACQUIRE_GL_OBJECTS|BARRIER|COPY_(?:BUFFER(?:_RECT|_TO_IMAGE)?|IMAGE(?:_TO_BUFFER)?)|FILL_(?:BUFFER|IMAGE)|MAP(?:_BUFFER|_IMAGE)|MARKER|MIGRATE(?:_SVM)?_MEM_OBJECTS|NATIVE_KERNEL|NDRANGE_KERNEL|READ_(?:BUFFER(?:_RECT)?|IMAGE)|RELEASE_GL_OBJECTS|SVM_(?:FREE|MAP|MEMCPY|MEMFILL|UNMAP)|TASK|UNMAP_MEM_OBJECT|USER|WRITE_(?:BUFFER(?:_RECT)?|IMAGE))|COMPILER_NOT_AVAILABLE|COMPILE_PROGRAM_FAILURE|COMPLETE|CONTEXT_(?:DEVICES|INTEROP_USER_SYNC|NUM_DEVICES|PLATFORM|PROPERTIES|REFERENCE_COUNT)|DEPTH(?:_STENCIL)?|DEVICE_(?:ADDRESS_BITS|AFFINITY_DOMAIN_(?:L[1-4]_CACHE|NEXT_PARTITIONABLE|NUMA)|AVAILABLE|BUILT_IN_KERNELS|COMPILER_AVAILABLE|DOUBLE_FP_CONFIG|ENDIAN_LITTLE|ERROR_CORRECTION_SUPPORT|EXECUTION_CAPABILITIES|EXTENSIONS|GLOBAL_(?:MEM_(?:CACHELINE_SIZE|CACHE_SIZE|CACHE_TYPE|SIZE)|VARIABLE_PREFERRED_TOTAL_SIZE)|HOST_UNIFIED_MEMORY|IL_VERSION|IMAGE(?:2D_MAX_(?:HEIGHT|WIDTH)|3D_MAX_(?:DEPTH|HEIGHT|WIDTH)|_BASE_ADDRESS_ALIGNMENT|_MAX_ARRAY_SIZE|_MAX_BUFFER_SIZE|_PITCH_ALIGNMENT|_SUPPORT)|LINKER_AVAILABLE|LOCAL_MEM_SIZE|LOCAL_MEM_TYPE|MAX_(?:CLOCK_FREQUENCY|COMPUTE_UNITS|CONSTANT_ARGS|CONSTANT_BUFFER_SIZE|GLOBAL_VARIABLE_SIZE|MEM_ALLOC_SIZE|NUM_SUB_GROUPS|ON_DEVICE_(?:EVENTS|QUEUES)|PARAMETER_SIZE|PIPE_ARGS|READ_IMAGE_ARGS|READ_WRITE_IMAGE_ARGS|SAMPLERS|WORK_GROUP_SIZE|WORK_ITEM_DIMENSIONS|WORK_ITEM_SIZES|WRITE_IMAGE_ARGS)|MEM_BASE_ADDR_ALIGN|MIN_DATA_TYPE_ALIGN_SIZE|NAME|NATIVE_VECTOR_WIDTH_(?:CHAR|DOUBLE|FLOAT|HALF|INT|LONG|SHORT)|NOT_(?:AVAILABLE|FOUND)|OPENCL_C_VERSION|PARENT_DEVICE|PARTITION_(?:AFFINITY_DOMAIN|BY_AFFINITY_DOMAIN|BY_COUNTS|BY_COUNTS_LIST_END|EQUALLY|FAILED|MAX_SUB_DEVICES|PROPERTIES|TYPE)|PIPE_MAX_(?:ACTIVE_RESERVATIONS|PACKET_SIZE)|PLATFORM|PREFERRED_(?:GLOBAL_ATOMIC_ALIGNMENT|INTEROP_USER_SYNC|LOCAL_ATOMIC_ALIGNMENT|PLATFORM_ATOMIC_ALIGNMENT|VECTOR_WIDTH_(?:CHAR|DOUBLE|FLOAT|HALF|INT|LONG|SHORT))|PRINTF_BUFFER_SIZE|PROFILE|PROFILING_TIMER_RESOLUTION|QUEUE_(?:ON_(?:DEVICE_(?:MAX_SIZE|PREFERRED_SIZE|PROPERTIES)|HOST_PROPERTIES)|PROPERTIES)|REFERENCE_COUNT|SINGLE_FP_CONFIG|SUB_GROUP_INDEPENDENT_FORWARD_PROGRESS|SVM_(?:ATOMICS|CAPABILITIES|COARSE_GRAIN_BUFFER|FINE_GRAIN_BUFFER|FINE_GRAIN_SYSTEM)|TYPE(?:_ACCELERATOR|_ALL|_CPU|_CUSTOM|_DEFAULT|_GPU)?|VENDOR(?:_ID)?|VERSION)|DRIVER_VERSION|EVENT_(?:COMMAND_(?:EXECUTION_STATUS|QUEUE|TYPE)|CONTEXT|REFERENCE_COUNT)|EXEC_(?:KERNEL|NATIVE_KERNEL|STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST)|FILTER_(?:LINEAR|NEAREST)|FLOAT|FP_(?:CORRECTLY_ROUNDED_DIVIDE_SQRT|DENORM|FMA|INF_NAN|ROUND_TO_INF|ROUND_TO_NEAREST|ROUND_TO_ZERO|SOFT_FLOAT)|GLOBAL|HALF_FLOAT|IMAGE_(?:ARRAY_SIZE|BUFFER|DEPTH|ELEMENT_SIZE|FORMAT|FORMAT_MISMATCH|FORMAT_NOT_SUPPORTED|HEIGHT|NUM_MIP_LEVELS|NUM_SAMPLES|ROW_PITCH|SLICE_PITCH|WIDTH)|INTENSITY|INVALID_(?:ARG_INDEX|ARG_SIZE|ARG_VALUE|BINARY|BUFFER_SIZE|BUILD_OPTIONS|COMMAND_QUEUE|COMPILER_OPTIONS|CONTEXT|DEVICE|DEVICE_PARTITION_COUNT|DEVICE_QUEUE|DEVICE_TYPE|EVENT|EVENT_WAIT_LIST|GLOBAL_OFFSET|GLOBAL_WORK_SIZE|GL_OBJECT|HOST_PTR|IMAGE_DESCRIPTOR|IMAGE_FORMAT_DESCRIPTOR|IMAGE_SIZE|KERNEL|KERNEL_ARGS|KERNEL_DEFINITION|KERNEL_NAME|LINKER_OPTIONS|MEM_OBJECT|MIP_LEVEL|OPERATION|PIPE_SIZE|PLATFORM|PROGRAM|PROGRAM_EXECUTABLE|PROPERTY|QUEUE_PROPERTIES|SAMPLER|VALUE|WORK_DIMENSION|WORK_GROUP_SIZE|WORK_ITEM_SIZE)|KERNEL_(?:ARG_(?:ACCESS_(?:NONE|QUALIFIER|READ_ONLY|READ_WRITE|WRITE_ONLY)|ADDRESS_(?:CONSTANT|GLOBAL|LOCAL|PRIVATE|QUALIFIER)|INFO_NOT_AVAILABLE|NAME|TYPE_(?:CONST|NAME|NONE|PIPE|QUALIFIER|RESTRICT|VOLATILE))|ATTRIBUTES|COMPILE_NUM_SUB_GROUPS|COMPILE_WORK_GROUP_SIZE|CONTEXT|EXEC_INFO_SVM_FINE_GRAIN_SYSTEM|EXEC_INFO_SVM_PTRS|FUNCTION_NAME|GLOBAL_WORK_SIZE|LOCAL_MEM_SIZE|LOCAL_SIZE_FOR_SUB_GROUP_COUNT|MAX_NUM_SUB_GROUPS|MAX_SUB_GROUP_SIZE_FOR_NDRANGE|NUM_ARGS|PREFERRED_WORK_GROUP_SIZE_MULTIPLE|PRIVATE_MEM_SIZE|PROGRAM|REFERENCE_COUNT|SUB_GROUP_COUNT_FOR_NDRANGE|WORK_GROUP_SIZE)|LINKER_NOT_AVAILABLE|LINK_PROGRAM_FAILURE|LOCAL|LUMINANCE|MAP_(?:FAILURE|READ|WRITE|WRITE_INVALIDATE_REGION)|MEM_(?:ALLOC_HOST_PTR|ASSOCIATED_MEMOBJECT|CONTEXT|COPY_HOST_PTR|COPY_OVERLAP|FLAGS|HOST_NO_ACCESS|HOST_PTR|HOST_READ_ONLY|HOST_WRITE_ONLY|KERNEL_READ_AND_WRITE|MAP_COUNT|OBJECT_(?:ALLOCATION_FAILURE|BUFFER|IMAGE1D|IMAGE1D_ARRAY|IMAGE1D_BUFFER|IMAGE2D|IMAGE2D_ARRAY|IMAGE3D|PIPE)|OFFSET|READ_ONLY|READ_WRITE|REFERENCE_COUNT|SIZE|SVM_ATOMICS|SVM_FINE_GRAIN_BUFFER|TYPE|USES_SVM_POINTER|USE_HOST_PTR|WRITE_ONLY)|MIGRATE_MEM_OBJECT_(?:CONTENT_UNDEFINED|HOST)|MISALIGNED_SUB_BUFFER_OFFSET|NONE|NON_BLOCKING|OUT_OF_(?:HOST_MEMORY|RESOURCES)|PIPE_(?:MAX_PACKETS|PACKET_SIZE)|PLATFORM_(?:EXTENSIONS|HOST_TIMER_RESOLUTION|NAME|PROFILE|VENDOR|VERSION)|PROFILING_(?:COMMAND_(?:COMPLETE|END|QUEUED|START|SUBMIT)|INFO_NOT_AVAILABLE)|PROGRAM_(?:BINARIES|BINARY_SIZES|BINARY_TYPE(?:_COMPILED_OBJECT|_EXECUTABLE|_LIBRARY|_NONE)?|BUILD_(?:GLOBAL_VARIABLE_TOTAL_SIZE|LOG|OPTIONS|STATUS)|CONTEXT|DEVICES|IL|KERNEL_NAMES|NUM_DEVICES|NUM_KERNELS|REFERENCE_COUNT|SOURCE)|QUEUED|QUEUE_(?:CONTEXT|DEVICE|DEVICE_DEFAULT|ON_DEVICE|ON_DEVICE_DEFAULT|OUT_OF_ORDER_EXEC_MODE_ENABLE|PROFILING_ENABLE|PROPERTIES|REFERENCE_COUNT|SIZE)|R|RA|READ_(?:ONLY|WRITE)_CACHE|RG|RGB|RGBA|RGBx|RGx|RUNNING|Rx|SAMPLER_(?:ADDRESSING_MODE|CONTEXT|FILTER_MODE|LOD_MAX|LOD_MIN|MIP_FILTER_MODE|NORMALIZED_COORDS|REFERENCE_COUNT)|(?:UN)?SIGNED_INT(?:8|16|32)|SNORM_INT(?:8|16)|SUBMITTED|SUCCESS|UNORM_INT(?:8|16|24|_101010|_101010_2)|UNORM_SHORT_(?:555|565)|VERSION_(?:1_0|1_1|1_2|2_0|2_1)|sBGRA|sRGB|sRGBA|sRGBx)\b/,alias:"constant"},"function-opencl-host":{pattern:/\bcl(?:BuildProgram|CloneKernel|CompileProgram|Create(?:Buffer|CommandQueue(?:WithProperties)?|Context|ContextFromType|Image|Image2D|Image3D|Kernel|KernelsInProgram|Pipe|ProgramWith(?:Binary|BuiltInKernels|IL|Source)|Sampler|SamplerWithProperties|SubBuffer|SubDevices|UserEvent)|Enqueue(?:(?:Barrier|Marker)(?:WithWaitList)?|Copy(?:Buffer(?:Rect|ToImage)?|Image(?:ToBuffer)?)|(?:Fill|Map)(?:Buffer|Image)|MigrateMemObjects|NDRangeKernel|NativeKernel|(?:Read|Write)(?:Buffer(?:Rect)?|Image)|SVM(?:Free|Map|MemFill|Memcpy|MigrateMem|Unmap)|Task|UnmapMemObject|WaitForEvents)|Finish|Flush|Get(?:CommandQueueInfo|ContextInfo|Device(?:AndHostTimer|IDs|Info)|Event(?:Profiling)?Info|ExtensionFunctionAddress(?:ForPlatform)?|HostTimer|ImageInfo|Kernel(?:ArgInfo|Info|SubGroupInfo|WorkGroupInfo)|MemObjectInfo|PipeInfo|Platform(?:IDs|Info)|Program(?:Build)?Info|SamplerInfo|SupportedImageFormats)|LinkProgram|(?:Release|Retain)(?:CommandQueue|Context|Device|Event|Kernel|MemObject|Program|Sampler)|SVM(?:Alloc|Free)|Set(?:CommandQueueProperty|DefaultDeviceCommandQueue|EventCallback|Kernel|Kernel(?:Arg(?:SVMPointer)?|ExecInfo)|MemObjectDestructorCallback|UserEventStatus)|Unload(?:Platform)?Compiler|WaitForEvents)\b/,alias:"function"}};s.languages.insertBefore("c","keyword",c),s.languages.cpp&&(c["type-opencl-host-cpp"]={pattern:/\b(?:Buffer|BufferGL|BufferRenderGL|CommandQueue|Context|Device|DeviceCommandQueue|EnqueueArgs|Event|Image|Image1D|Image1DArray|Image1DBuffer|Image2D|Image2DArray|Image2DGL|Image3D|Image3DGL|ImageFormat|ImageGL|Kernel|KernelFunctor|LocalSpaceArg|Memory|NDRange|Pipe|Platform|Program|SVMAllocator|SVMTraitAtomic|SVMTraitCoarse|SVMTraitFine|SVMTraitReadOnly|SVMTraitReadWrite|SVMTraitWriteOnly|Sampler|UserEvent)\b/,alias:"keyword"},s.languages.insertBefore("cpp","keyword",c))}(o)}return dHe}var fHe,ZGt;function WBr(){if(ZGt)return fHe;ZGt=1,fHe=n,n.displayName="openqasm",n.aliases=["qasm"];function n(r){r.languages.openqasm={comment:/\/\*[\s\S]*?\*\/|\/\/.*/,string:{pattern:/"[^"\r\n\t]*"|'[^'\r\n\t]*'/,greedy:!0},keyword:/\b(?:CX|OPENQASM|U|barrier|boxas|boxto|break|const|continue|ctrl|def|defcal|defcalgrammar|delay|else|end|for|gate|gphase|if|in|include|inv|kernel|lengthof|let|measure|pow|reset|return|rotary|stretchinf|while)\b|#pragma\b/,"class-name":/\b(?:angle|bit|bool|creg|fixed|float|int|length|qreg|qubit|stretch|uint)\b/,function:/\b(?:cos|exp|ln|popcount|rotl|rotr|sin|sqrt|tan)\b(?=\s*\()/,constant:/\b(?:euler|pi|tau)\b|π|𝜏|ℇ/,number:{pattern:/(^|[^.\w$])(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?(?:dt|ns|us|µs|ms|s)?/i,lookbehind:!0},operator:/->|>>=?|<<=?|&&|\|\||\+\+|--|[!=<>&|~^+\-*/%]=?|@/,punctuation:/[(){}\[\];,:.]/},r.languages.qasm=r.languages.openqasm}return fHe}var hHe,QGt;function KBr(){if(QGt)return hHe;QGt=1,hHe=n,n.displayName="oz",n.aliases=[];function n(r){r.languages.oz={comment:{pattern:/\/\*[\s\S]*?\*\/|%.*/,greedy:!0},string:{pattern:/"(?:[^"\\]|\\[\s\S])*"/,greedy:!0},atom:{pattern:/'(?:[^'\\]|\\[\s\S])*'/,greedy:!0,alias:"builtin"},keyword:/\$|\[\]|\b(?:_|at|attr|case|catch|choice|class|cond|declare|define|dis|else(?:case|if)?|end|export|fail|false|feat|finally|from|fun|functor|if|import|in|local|lock|meth|nil|not|of|or|prepare|proc|prop|raise|require|self|skip|then|thread|true|try|unit)\b/,function:[/\b[a-z][A-Za-z\d]*(?=\()/,{pattern:/(\{)[A-Z][A-Za-z\d]*\b/,lookbehind:!0}],number:/\b(?:0[bx][\da-f]+|\d+(?:\.\d*)?(?:e~?\d+)?)\b|&(?:[^\\]|\\(?:\d{3}|.))/i,variable:/`(?:[^`\\]|\\.)+`/,"attr-name":/\b\w+(?=[ \t]*:(?![:=]))/,operator:/:(?:=|::?)|<[-:=]?|=(?:=|<?:?)|>=?:?|\\=:?|!!?|[|#+\-*\/,~^@]|\b(?:andthen|div|mod|orelse)\b/,punctuation:/[\[\](){}.:;?]/}}return hHe}var pHe,eqt;function YBr(){if(eqt)return pHe;eqt=1,pHe=n,n.displayName="parigp",n.aliases=[];function n(r){r.languages.parigp={comment:/\/\*[\s\S]*?\*\/|\\\\.*/,string:{pattern:/"(?:[^"\\\r\n]|\\.)*"/,greedy:!0},keyword:function(){var o=["breakpoint","break","dbg_down","dbg_err","dbg_up","dbg_x","forcomposite","fordiv","forell","forpart","forprime","forstep","forsubgroup","forvec","for","iferr","if","local","my","next","return","until","while"];return o=o.map(function(s){return s.split("").join(" *")}).join("|"),RegExp("\\b(?:"+o+")\\b")}(),function:/\b\w(?:[\w ]*\w)?(?= *\()/,number:{pattern:/((?:\. *\. *)?)(?:\b\d(?: *\d)*(?: *(?!\. *\.)\.(?: *\d)*)?|\. *\d(?: *\d)*)(?: *e *(?:[+-] *)?\d(?: *\d)*)?/i,lookbehind:!0},operator:/\. *\.|[*\/!](?: *=)?|%(?: *=|(?: *#)?(?: *')*)?|\+(?: *[+=])?|-(?: *[-=>])?|<(?: *>|(?: *<)?(?: *=)?)?|>(?: *>)?(?: *=)?|=(?: *=){0,2}|\\(?: *\/)?(?: *=)?|&(?: *&)?|\| *\||['#~^]/,punctuation:/[\[\]{}().,:;|]/}}return pHe}var gHe,tqt;function XBr(){if(tqt)return gHe;tqt=1,gHe=n,n.displayName="parser",n.aliases=[];function n(r){(function(o){var s=o.languages.parser=o.languages.extend("markup",{keyword:{pattern:/(^|[^^])(?:\^(?:case|eval|for|if|switch|throw)\b|@(?:BASE|CLASS|GET(?:_DEFAULT)?|OPTIONS|SET_DEFAULT|USE)\b)/,lookbehind:!0},variable:{pattern:/(^|[^^])\B\$(?:\w+|(?=[.{]))(?:(?:\.|::?)\w+)*(?:\.|::?)?/,lookbehind:!0,inside:{punctuation:/\.|:+/}},function:{pattern:/(^|[^^])\B[@^]\w+(?:(?:\.|::?)\w+)*(?:\.|::?)?/,lookbehind:!0,inside:{keyword:{pattern:/(^@)(?:GET_|SET_)/,lookbehind:!0},punctuation:/\.|:+/}},escape:{pattern:/\^(?:[$^;@()\[\]{}"':]|#[a-f\d]*)/i,alias:"builtin"},punctuation:/[\[\](){};]/});s=o.languages.insertBefore("parser","keyword",{"parser-comment":{pattern:/(\s)#.*/,lookbehind:!0,alias:"comment"},expression:{pattern:/(^|[^^])\((?:[^()]|\((?:[^()]|\((?:[^()])*\))*\))*\)/,greedy:!0,lookbehind:!0,inside:{string:{pattern:/(^|[^^])(["'])(?:(?!\2)[^^]|\^[\s\S])*\2/,lookbehind:!0},keyword:s.keyword,variable:s.variable,function:s.function,boolean:/\b(?:false|true)\b/,number:/\b(?:0x[a-f\d]+|\d+(?:\.\d*)?(?:e[+-]?\d+)?)\b/i,escape:s.escape,operator:/[~+*\/\\%]|!(?:\|\|?|=)?|&&?|\|\|?|==|<[<=]?|>[>=]?|-[fd]?|\b(?:def|eq|ge|gt|in|is|le|lt|ne)\b/,punctuation:s.punctuation}}}),o.languages.insertBefore("inside","punctuation",{expression:s.expression,keyword:s.keyword,variable:s.variable,function:s.function,escape:s.escape,"parser-punctuation":{pattern:s.punctuation,alias:"punctuation"}},s.tag.inside["attr-value"])})(r)}return gHe}var mHe,nqt;function JBr(){if(nqt)return mHe;nqt=1,mHe=n,n.displayName="pascal",n.aliases=["objectpascal"];function n(r){r.languages.pascal={directive:{pattern:/\{\$[\s\S]*?\}/,greedy:!0,alias:["marco","property"]},comment:{pattern:/\(\*[\s\S]*?\*\)|\{[\s\S]*?\}|\/\/.*/,greedy:!0},string:{pattern:/(?:'(?:''|[^'\r\n])*'(?!')|#[&$%]?[a-f\d]+)+|\^[a-z]/i,greedy:!0},asm:{pattern:/(\basm\b)[\s\S]+?(?=\bend\s*[;[])/i,lookbehind:!0,greedy:!0,inside:null},keyword:[{pattern:/(^|[^&])\b(?:absolute|array|asm|begin|case|const|constructor|destructor|do|downto|else|end|file|for|function|goto|if|implementation|inherited|inline|interface|label|nil|object|of|operator|packed|procedure|program|record|reintroduce|repeat|self|set|string|then|to|type|unit|until|uses|var|while|with)\b/i,lookbehind:!0},{pattern:/(^|[^&])\b(?:dispose|exit|false|new|true)\b/i,lookbehind:!0},{pattern:/(^|[^&])\b(?:class|dispinterface|except|exports|finalization|finally|initialization|inline|library|on|out|packed|property|raise|resourcestring|threadvar|try)\b/i,lookbehind:!0},{pattern:/(^|[^&])\b(?:absolute|abstract|alias|assembler|bitpacked|break|cdecl|continue|cppdecl|cvar|default|deprecated|dynamic|enumerator|experimental|export|external|far|far16|forward|generic|helper|implements|index|interrupt|iochecks|local|message|name|near|nodefault|noreturn|nostackframe|oldfpccall|otherwise|overload|override|pascal|platform|private|protected|public|published|read|register|reintroduce|result|safecall|saveregisters|softfloat|specialize|static|stdcall|stored|strict|unaligned|unimplemented|varargs|virtual|write)\b/i,lookbehind:!0}],number:[/(?:[&%]\d+|\$[a-f\d]+)/i,/\b\d+(?:\.\d+)?(?:e[+-]?\d+)?/i],operator:[/\.\.|\*\*|:=|<[<=>]?|>[>=]?|[+\-*\/]=?|[@^=]/,{pattern:/(^|[^&])\b(?:and|as|div|exclude|in|include|is|mod|not|or|shl|shr|xor)\b/,lookbehind:!0}],punctuation:/\(\.|\.\)|[()\[\]:;,.]/},r.languages.pascal.asm.inside=r.languages.extend("pascal",{asm:void 0,keyword:void 0,operator:void 0}),r.languages.objectpascal=r.languages.pascal}return mHe}var bHe,rqt;function ZBr(){if(rqt)return bHe;rqt=1,bHe=n,n.displayName="pascaligo",n.aliases=[];function n(r){(function(o){var s=/\((?:[^()]|\((?:[^()]|\([^()]*\))*\))*\)/.source,c=/(?:\b\w+(?:<braces>)?|<braces>)/.source.replace(/<braces>/g,function(){return s}),d=o.languages.pascaligo={comment:/\(\*[\s\S]+?\*\)|\/\/.*/,string:{pattern:/(["'`])(?:\\[\s\S]|(?!\1)[^\\])*\1|\^[a-z]/i,greedy:!0},"class-name":[{pattern:RegExp(/(\btype\s+\w+\s+is\s+)<type>/.source.replace(/<type>/g,function(){return c}),"i"),lookbehind:!0,inside:null},{pattern:RegExp(/<type>(?=\s+is\b)/.source.replace(/<type>/g,function(){return c}),"i"),inside:null},{pattern:RegExp(/(:\s*)<type>/.source.replace(/<type>/g,function(){return c})),lookbehind:!0,inside:null}],keyword:{pattern:/(^|[^&])\b(?:begin|block|case|const|else|end|fail|for|from|function|if|is|nil|of|remove|return|skip|then|type|var|while|with)\b/i,lookbehind:!0},boolean:{pattern:/(^|[^&])\b(?:False|True)\b/i,lookbehind:!0},builtin:{pattern:/(^|[^&])\b(?:bool|int|list|map|nat|record|string|unit)\b/i,lookbehind:!0},function:/\b\w+(?=\s*\()/,number:[/%[01]+|&[0-7]+|\$[a-f\d]+/i,/\b\d+(?:\.\d+)?(?:e[+-]?\d+)?(?:mtz|n)?/i],operator:/->|=\/=|\.\.|\*\*|:=|<[<=>]?|>[>=]?|[+\-*\/]=?|[@^=|]|\b(?:and|mod|or)\b/,punctuation:/\(\.|\.\)|[()\[\]:;,.{}]/},h=["comment","keyword","builtin","operator","punctuation"].reduce(function(b,y){return b[y]=d[y],b},{});d["class-name"].forEach(function(b){b.inside=h})})(r)}return bHe}var vHe,iqt;function QBr(){if(iqt)return vHe;iqt=1,vHe=n,n.displayName="pcaxis",n.aliases=["px"];function n(r){r.languages.pcaxis={string:/"[^"]*"/,keyword:{pattern:/((?:^|;)\s*)[-A-Z\d]+(?:\s*\[[-\w]+\])?(?:\s*\("[^"]*"(?:,\s*"[^"]*")*\))?(?=\s*=)/,lookbehind:!0,greedy:!0,inside:{keyword:/^[-A-Z\d]+/,language:{pattern:/^(\s*)\[[-\w]+\]/,lookbehind:!0,inside:{punctuation:/^\[|\]$/,property:/[-\w]+/}},"sub-key":{pattern:/^(\s*)\S[\s\S]*/,lookbehind:!0,inside:{parameter:{pattern:/"[^"]*"/,alias:"property"},punctuation:/^\(|\)$|,/}}}},operator:/=/,tlist:{pattern:/TLIST\s*\(\s*\w+(?:(?:\s*,\s*"[^"]*")+|\s*,\s*"[^"]*"-"[^"]*")?\s*\)/,greedy:!0,inside:{function:/^TLIST/,property:{pattern:/^(\s*\(\s*)\w+/,lookbehind:!0},string:/"[^"]*"/,punctuation:/[(),]/,operator:/-/}},punctuation:/[;,]/,number:{pattern:/(^|\s)\d+(?:\.\d+)?(?!\S)/,lookbehind:!0},boolean:/NO|YES/},r.languages.px=r.languages.pcaxis}return vHe}var wHe,oqt;function eUr(){if(oqt)return wHe;oqt=1,wHe=n,n.displayName="peoplecode",n.aliases=["pcode"];function n(r){r.languages.peoplecode={comment:RegExp([/\/\*[\s\S]*?\*\//.source,/\bREM[^;]*;/.source,/<\*(?:[^<*]|\*(?!>)|<(?!\*)|<\*(?:(?!\*>)[\s\S])*\*>)*\*>/.source,/\/\+[\s\S]*?\+\//.source].join("|")),string:{pattern:/'(?:''|[^'\r\n])*'(?!')|"(?:""|[^"\r\n])*"(?!")/,greedy:!0},variable:/%\w+/,"function-definition":{pattern:/((?:^|[^\w-])(?:function|method)\s+)\w+/i,lookbehind:!0,alias:"function"},"class-name":{pattern:/((?:^|[^-\w])(?:as|catch|class|component|create|extends|global|implements|instance|local|of|property|returns)\s+)\w+(?::\w+)*/i,lookbehind:!0,inside:{punctuation:/:/}},keyword:/\b(?:abstract|alias|as|catch|class|component|constant|create|declare|else|end-(?:class|evaluate|for|function|get|if|method|set|try|while)|evaluate|extends|for|function|get|global|if|implements|import|instance|library|local|method|null|of|out|peopleCode|private|program|property|protected|readonly|ref|repeat|returns?|set|step|then|throw|to|try|until|value|when(?:-other)?|while)\b/i,"operator-keyword":{pattern:/\b(?:and|not|or)\b/i,alias:"operator"},function:/[_a-z]\w*(?=\s*\()/i,boolean:/\b(?:false|true)\b/i,number:/\b\d+(?:\.\d+)?\b/,operator:/<>|[<>]=?|!=|\*\*|[-+*/|=@]/,punctuation:/[:.;,()[\]]/},r.languages.pcode=r.languages.peoplecode}return wHe}var yHe,aqt;function tUr(){if(aqt)return yHe;aqt=1,yHe=n,n.displayName="perl",n.aliases=[];function n(r){(function(o){var s=/(?:\((?:[^()\\]|\\[\s\S])*\)|\{(?:[^{}\\]|\\[\s\S])*\}|\[(?:[^[\]\\]|\\[\s\S])*\]|<(?:[^<>\\]|\\[\s\S])*>)/.source;o.languages.perl={comment:[{pattern:/(^\s*)=\w[\s\S]*?=cut.*/m,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\$])#.*/,lookbehind:!0,greedy:!0}],string:[{pattern:RegExp(/\b(?:q|qq|qw|qx)(?![a-zA-Z0-9])\s*/.source+"(?:"+[/([^a-zA-Z0-9\s{(\[<])(?:(?!\1)[^\\]|\\[\s\S])*\1/.source,/([a-zA-Z0-9])(?:(?!\2)[^\\]|\\[\s\S])*\2/.source,s].join("|")+")"),greedy:!0},{pattern:/("|`)(?:(?!\1)[^\\]|\\[\s\S])*\1/,greedy:!0},{pattern:/'(?:[^'\\\r\n]|\\.)*'/,greedy:!0}],regex:[{pattern:RegExp(/\b(?:m|qr)(?![a-zA-Z0-9])\s*/.source+"(?:"+[/([^a-zA-Z0-9\s{(\[<])(?:(?!\1)[^\\]|\\[\s\S])*\1/.source,/([a-zA-Z0-9])(?:(?!\2)[^\\]|\\[\s\S])*\2/.source,s].join("|")+")"+/[msixpodualngc]*/.source),greedy:!0},{pattern:RegExp(/(^|[^-])\b(?:s|tr|y)(?![a-zA-Z0-9])\s*/.source+"(?:"+[/([^a-zA-Z0-9\s{(\[<])(?:(?!\2)[^\\]|\\[\s\S])*\2(?:(?!\2)[^\\]|\\[\s\S])*\2/.source,/([a-zA-Z0-9])(?:(?!\3)[^\\]|\\[\s\S])*\3(?:(?!\3)[^\\]|\\[\s\S])*\3/.source,s+/\s*/.source+s].join("|")+")"+/[msixpodualngcer]*/.source),lookbehind:!0,greedy:!0},{pattern:/\/(?:[^\/\\\r\n]|\\.)*\/[msixpodualngc]*(?=\s*(?:$|[\r\n,.;})&|\-+*~<>!?^]|(?:and|cmp|eq|ge|gt|le|lt|ne|not|or|x|xor)\b))/,greedy:!0}],variable:[/[&*$@%]\{\^[A-Z]+\}/,/[&*$@%]\^[A-Z_]/,/[&*$@%]#?(?=\{)/,/[&*$@%]#?(?:(?:::)*'?(?!\d)[\w$]+(?![\w$]))+(?:::)*/,/[&*$@%]\d+/,/(?!%=)[$@%][!"#$%&'()*+,\-.\/:;<=>?@[\\\]^_`{|}~]/],filehandle:{pattern:/<(?![<=])\S*?>|\b_\b/,alias:"symbol"},"v-string":{pattern:/v\d+(?:\.\d+)*|\d+(?:\.\d+){2,}/,alias:"string"},function:{pattern:/(\bsub[ \t]+)\w+/,lookbehind:!0},keyword:/\b(?:any|break|continue|default|delete|die|do|else|elsif|eval|for|foreach|given|goto|if|last|local|my|next|our|package|print|redo|require|return|say|state|sub|switch|undef|unless|until|use|when|while)\b/,number:/\b(?:0x[\dA-Fa-f](?:_?[\dA-Fa-f])*|0b[01](?:_?[01])*|(?:(?:\d(?:_?\d)*)?\.)?\d(?:_?\d)*(?:[Ee][+-]?\d+)?)\b/,operator:/-[rwxoRWXOezsfdlpSbctugkTBMAC]\b|\+[+=]?|-[-=>]?|\*\*?=?|\/\/?=?|=[=~>]?|~[~=]?|\|\|?=?|&&?=?|<(?:=>?|<=?)?|>>?=?|![~=]?|[%^]=?|\.(?:=|\.\.?)?|[\\?]|\bx(?:=|\b)|\b(?:and|cmp|eq|ge|gt|le|lt|ne|not|or|xor)\b/,punctuation:/[{}[\];(),:]/}})(r)}return yHe}var EHe,sqt;function nUr(){if(sqt)return EHe;sqt=1;var n=DEe();EHe=r,r.displayName="phpExtras",r.aliases=[];function r(o){o.register(n),o.languages.insertBefore("php","variable",{this:{pattern:/\$this\b/,alias:"keyword"},global:/\$(?:GLOBALS|HTTP_RAW_POST_DATA|_(?:COOKIE|ENV|FILES|GET|POST|REQUEST|SERVER|SESSION)|argc|argv|http_response_header|php_errormsg)\b/,scope:{pattern:/\b[\w\\]+::/,inside:{keyword:/\b(?:parent|self|static)\b/,punctuation:/::|\\/}}})}return EHe}var SHe,cqt;function rUr(){if(cqt)return SHe;cqt=1;var n=DEe(),r=LEe();SHe=o,o.displayName="phpdoc",o.aliases=[];function o(s){s.register(n),s.register(r),function(c){var d=/(?:\b[a-zA-Z]\w*|[|\\[\]])+/.source;c.languages.phpdoc=c.languages.extend("javadoclike",{parameter:{pattern:RegExp("(@(?:global|param|property(?:-read|-write)?|var)\\s+(?:"+d+"\\s+)?)\\$\\w+"),lookbehind:!0}}),c.languages.insertBefore("phpdoc","keyword",{"class-name":[{pattern:RegExp("(@(?:global|package|param|property(?:-read|-write)?|return|subpackage|throws|var)\\s+)"+d),lookbehind:!0,inside:{keyword:/\b(?:array|bool|boolean|callback|double|false|float|int|integer|mixed|null|object|resource|self|string|true|void)\b/,punctuation:/[|\\[\]()]/}}]}),c.languages.javadoclike.addSupport("php",c.languages.phpdoc)}(s)}return SHe}var _He,uqt;function iUr(){if(uqt)return _He;uqt=1;var n=wQe();_He=r,r.displayName="plsql",r.aliases=[];function r(o){o.register(n),o.languages.plsql=o.languages.extend("sql",{comment:{pattern:/\/\*[\s\S]*?\*\/|--.*/,greedy:!0},keyword:/\b(?:A|ACCESSIBLE|ADD|AGENT|AGGREGATE|ALL|ALTER|AND|ANY|ARRAY|AS|ASC|AT|ATTRIBUTE|AUTHID|AVG|BEGIN|BETWEEN|BFILE_BASE|BINARY|BLOB_BASE|BLOCK|BODY|BOTH|BOUND|BULK|BY|BYTE|C|CALL|CALLING|CASCADE|CASE|CHAR|CHARACTER|CHARSET|CHARSETFORM|CHARSETID|CHAR_BASE|CHECK|CLOB_BASE|CLONE|CLOSE|CLUSTER|CLUSTERS|COLAUTH|COLLECT|COLUMNS|COMMENT|COMMIT|COMMITTED|COMPILED|COMPRESS|CONNECT|CONSTANT|CONSTRUCTOR|CONTEXT|CONTINUE|CONVERT|COUNT|CRASH|CREATE|CREDENTIAL|CURRENT|CURSOR|CUSTOMDATUM|DANGLING|DATA|DATE|DATE_BASE|DAY|DECLARE|DEFAULT|DEFINE|DELETE|DESC|DETERMINISTIC|DIRECTORY|DISTINCT|DOUBLE|DROP|DURATION|ELEMENT|ELSE|ELSIF|EMPTY|END|ESCAPE|EXCEPT|EXCEPTION|EXCEPTIONS|EXCLUSIVE|EXECUTE|EXISTS|EXIT|EXTERNAL|FETCH|FINAL|FIRST|FIXED|FLOAT|FOR|FORALL|FORCE|FROM|FUNCTION|GENERAL|GOTO|GRANT|GROUP|HASH|HAVING|HEAP|HIDDEN|HOUR|IDENTIFIED|IF|IMMEDIATE|IMMUTABLE|IN|INCLUDING|INDEX|INDEXES|INDICATOR|INDICES|INFINITE|INSERT|INSTANTIABLE|INT|INTERFACE|INTERSECT|INTERVAL|INTO|INVALIDATE|IS|ISOLATION|JAVA|LANGUAGE|LARGE|LEADING|LENGTH|LEVEL|LIBRARY|LIKE|LIKE2|LIKE4|LIKEC|LIMIT|LIMITED|LOCAL|LOCK|LONG|LOOP|MAP|MAX|MAXLEN|MEMBER|MERGE|MIN|MINUS|MINUTE|MOD|MODE|MODIFY|MONTH|MULTISET|MUTABLE|NAME|NAN|NATIONAL|NATIVE|NCHAR|NEW|NOCOMPRESS|NOCOPY|NOT|NOWAIT|NULL|NUMBER_BASE|OBJECT|OCICOLL|OCIDATE|OCIDATETIME|OCIDURATION|OCIINTERVAL|OCILOBLOCATOR|OCINUMBER|OCIRAW|OCIREF|OCIREFCURSOR|OCIROWID|OCISTRING|OCITYPE|OF|OLD|ON|ONLY|OPAQUE|OPEN|OPERATOR|OPTION|OR|ORACLE|ORADATA|ORDER|ORGANIZATION|ORLANY|ORLVARY|OTHERS|OUT|OVERLAPS|OVERRIDING|PACKAGE|PARALLEL_ENABLE|PARAMETER|PARAMETERS|PARENT|PARTITION|PASCAL|PERSISTABLE|PIPE|PIPELINED|PLUGGABLE|POLYMORPHIC|PRAGMA|PRECISION|PRIOR|PRIVATE|PROCEDURE|PUBLIC|RAISE|RANGE|RAW|READ|RECORD|REF|REFERENCE|RELIES_ON|REM|REMAINDER|RENAME|RESOURCE|RESULT|RESULT_CACHE|RETURN|RETURNING|REVERSE|REVOKE|ROLLBACK|ROW|SAMPLE|SAVE|SAVEPOINT|SB1|SB2|SB4|SECOND|SEGMENT|SELECT|SELF|SEPARATE|SEQUENCE|SERIALIZABLE|SET|SHARE|SHORT|SIZE|SIZE_T|SOME|SPARSE|SQL|SQLCODE|SQLDATA|SQLNAME|SQLSTATE|STANDARD|START|STATIC|STDDEV|STORED|STRING|STRUCT|STYLE|SUBMULTISET|SUBPARTITION|SUBSTITUTABLE|SUBTYPE|SUM|SYNONYM|TABAUTH|TABLE|TDO|THE|THEN|TIME|TIMESTAMP|TIMEZONE_ABBR|TIMEZONE_HOUR|TIMEZONE_MINUTE|TIMEZONE_REGION|TO|TRAILING|TRANSACTION|TRANSACTIONAL|TRUSTED|TYPE|UB1|UB2|UB4|UNDER|UNION|UNIQUE|UNPLUG|UNSIGNED|UNTRUSTED|UPDATE|USE|USING|VALIST|VALUE|VALUES|VARIABLE|VARIANCE|VARRAY|VARYING|VIEW|VIEWS|VOID|WHEN|WHERE|WHILE|WITH|WORK|WRAPPED|WRITE|YEAR|ZONE)\b/i,operator:/:=?|=>|[<>^~!]=|\.\.|\|\||\*\*|[-+*/%<>=@]/}),o.languages.insertBefore("plsql","operator",{label:{pattern:/<<\s*\w+\s*>>/,alias:"symbol"}})}return _He}var kHe,lqt;function oUr(){if(lqt)return kHe;lqt=1,kHe=n,n.displayName="powerquery",n.aliases=[];function n(r){r.languages.powerquery={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0,greedy:!0},"quoted-identifier":{pattern:/#"(?:[^"\r\n]|"")*"(?!")/,greedy:!0},string:{pattern:/(?:#!)?"(?:[^"\r\n]|"")*"(?!")/,greedy:!0},constant:[/\bDay\.(?:Friday|Monday|Saturday|Sunday|Thursday|Tuesday|Wednesday)\b/,/\bTraceLevel\.(?:Critical|Error|Information|Verbose|Warning)\b/,/\bOccurrence\.(?:All|First|Last)\b/,/\bOrder\.(?:Ascending|Descending)\b/,/\bRoundingMode\.(?:AwayFromZero|Down|ToEven|TowardZero|Up)\b/,/\bMissingField\.(?:Error|Ignore|UseNull)\b/,/\bQuoteStyle\.(?:Csv|None)\b/,/\bJoinKind\.(?:FullOuter|Inner|LeftAnti|LeftOuter|RightAnti|RightOuter)\b/,/\bGroupKind\.(?:Global|Local)\b/,/\bExtraValues\.(?:Error|Ignore|List)\b/,/\bJoinAlgorithm\.(?:Dynamic|LeftHash|LeftIndex|PairwiseHash|RightHash|RightIndex|SortMerge)\b/,/\bJoinSide\.(?:Left|Right)\b/,/\bPrecision\.(?:Decimal|Double)\b/,/\bRelativePosition\.From(?:End|Start)\b/,/\bTextEncoding\.(?:Ascii|BigEndianUnicode|Unicode|Utf16|Utf8|Windows)\b/,/\b(?:Any|Binary|Date|DateTime|DateTimeZone|Duration|Function|Int16|Int32|Int64|Int8|List|Logical|None|Number|Record|Table|Text|Time)\.Type\b/,/\bnull\b/],boolean:/\b(?:false|true)\b/,keyword:/\b(?:and|as|each|else|error|if|in|is|let|meta|not|nullable|optional|or|otherwise|section|shared|then|try|type)\b|#(?:binary|date|datetime|datetimezone|duration|infinity|nan|sections|shared|table|time)\b/,function:{pattern:/(^|[^#\w.])[a-z_][\w.]*(?=\s*\()/i,lookbehind:!0},"data-type":{pattern:/\b(?:any|anynonnull|binary|date|datetime|datetimezone|duration|function|list|logical|none|number|record|table|text|time)\b/,alias:"class-name"},number:{pattern:/\b0x[\da-f]+\b|(?:[+-]?(?:\b\d+\.)?\b\d+|[+-]\.\d+|(^|[^.])\B\.\d+)(?:e[+-]?\d+)?\b/i,lookbehind:!0},operator:/[-+*\/&?@^]|<(?:=>?|>)?|>=?|=>?|\.\.\.?/,punctuation:/[,;\[\](){}]/},r.languages.pq=r.languages.powerquery,r.languages.mscript=r.languages.powerquery}return kHe}var THe,dqt;function aUr(){if(dqt)return THe;dqt=1,THe=n,n.displayName="powershell",n.aliases=[];function n(r){(function(o){var s=o.languages.powershell={comment:[{pattern:/(^|[^`])<#[\s\S]*?#>/,lookbehind:!0},{pattern:/(^|[^`])#.*/,lookbehind:!0}],string:[{pattern:/"(?:`[\s\S]|[^`"])*"/,greedy:!0,inside:null},{pattern:/'(?:[^']|'')*'/,greedy:!0}],namespace:/\[[a-z](?:\[(?:\[[^\]]*\]|[^\[\]])*\]|[^\[\]])*\]/i,boolean:/\$(?:false|true)\b/i,variable:/\$\w+\b/,function:[/\b(?:Add|Approve|Assert|Backup|Block|Checkpoint|Clear|Close|Compare|Complete|Compress|Confirm|Connect|Convert|ConvertFrom|ConvertTo|Copy|Debug|Deny|Disable|Disconnect|Dismount|Edit|Enable|Enter|Exit|Expand|Export|Find|ForEach|Format|Get|Grant|Group|Hide|Import|Initialize|Install|Invoke|Join|Limit|Lock|Measure|Merge|Move|New|Open|Optimize|Out|Ping|Pop|Protect|Publish|Push|Read|Receive|Redo|Register|Remove|Rename|Repair|Request|Reset|Resize|Resolve|Restart|Restore|Resume|Revoke|Save|Search|Select|Send|Set|Show|Skip|Sort|Split|Start|Step|Stop|Submit|Suspend|Switch|Sync|Tee|Test|Trace|Unblock|Undo|Uninstall|Unlock|Unprotect|Unpublish|Unregister|Update|Use|Wait|Watch|Where|Write)-[a-z]+\b/i,/\b(?:ac|cat|chdir|clc|cli|clp|clv|compare|copy|cp|cpi|cpp|cvpa|dbp|del|diff|dir|ebp|echo|epal|epcsv|epsn|erase|fc|fl|ft|fw|gal|gbp|gc|gci|gcs|gdr|gi|gl|gm|gp|gps|group|gsv|gu|gv|gwmi|iex|ii|ipal|ipcsv|ipsn|irm|iwmi|iwr|kill|lp|ls|measure|mi|mount|move|mp|mv|nal|ndr|ni|nv|ogv|popd|ps|pushd|pwd|rbp|rd|rdr|ren|ri|rm|rmdir|rni|rnp|rp|rv|rvpa|rwmi|sal|saps|sasv|sbp|sc|select|set|shcm|si|sl|sleep|sls|sort|sp|spps|spsv|start|sv|swmi|tee|trcm|type|write)\b/i],keyword:/\b(?:Begin|Break|Catch|Class|Continue|Data|Define|Do|DynamicParam|Else|ElseIf|End|Exit|Filter|Finally|For|ForEach|From|Function|If|InlineScript|Parallel|Param|Process|Return|Sequence|Switch|Throw|Trap|Try|Until|Using|Var|While|Workflow)\b/i,operator:{pattern:/(^|\W)(?:!|-(?:b?(?:and|x?or)|as|(?:Not)?(?:Contains|In|Like|Match)|eq|ge|gt|is(?:Not)?|Join|le|lt|ne|not|Replace|sh[lr])\b|-[-=]?|\+[+=]?|[*\/%]=?)/i,lookbehind:!0},punctuation:/[|{}[\];(),.]/};s.string[0].inside={function:{pattern:/(^|[^`])\$\((?:\$\([^\r\n()]*\)|(?!\$\()[^\r\n)])*\)/,lookbehind:!0,inside:s},boolean:s.boolean,variable:s.variable}})(r)}return THe}var xHe,fqt;function sUr(){if(fqt)return xHe;fqt=1,xHe=n,n.displayName="processing",n.aliases=[];function n(r){r.languages.processing=r.languages.extend("clike",{keyword:/\b(?:break|case|catch|class|continue|default|else|extends|final|for|if|implements|import|new|null|private|public|return|static|super|switch|this|try|void|while)\b/,function:/\b\w+(?=\s*\()/,operator:/<[<=]?|>[>=]?|&&?|\|\|?|[%?]|[!=+\-*\/]=?/}),r.languages.insertBefore("processing","number",{constant:/\b(?!XML\b)[A-Z][A-Z\d_]+\b/,type:{pattern:/\b(?:boolean|byte|char|color|double|float|int|[A-Z]\w*)\b/,alias:"class-name"}})}return xHe}var CHe,hqt;function cUr(){if(hqt)return CHe;hqt=1,CHe=n,n.displayName="prolog",n.aliases=[];function n(r){r.languages.prolog={comment:{pattern:/\/\*[\s\S]*?\*\/|%.*/,greedy:!0},string:{pattern:/(["'])(?:\1\1|\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1(?!\1)/,greedy:!0},builtin:/\b(?:fx|fy|xf[xy]?|yfx?)\b/,function:/\b[a-z]\w*(?:(?=\()|\/\d+)/,number:/\b\d+(?:\.\d*)?/,operator:/[:\\=><\-?*@\/;+^|!$.]+|\b(?:is|mod|not|xor)\b/,punctuation:/[(){}\[\],]/}}return CHe}var AHe,pqt;function uUr(){if(pqt)return AHe;pqt=1,AHe=n,n.displayName="promql",n.aliases=[];function n(r){(function(o){var s=["sum","min","max","avg","group","stddev","stdvar","count","count_values","bottomk","topk","quantile"],c=["on","ignoring","group_right","group_left","by","without"],d=["offset"],h=s.concat(c,d);o.languages.promql={comment:{pattern:/(^[ \t]*)#.*/m,lookbehind:!0},"vector-match":{pattern:new RegExp("((?:"+c.join("|")+")\\s*)\\([^)]*\\)"),lookbehind:!0,inside:{"label-key":{pattern:/\b[^,]+\b/,alias:"attr-name"},punctuation:/[(),]/}},"context-labels":{pattern:/\{[^{}]*\}/,inside:{"label-key":{pattern:/\b[a-z_]\w*(?=\s*(?:=|![=~]))/,alias:"attr-name"},"label-value":{pattern:/(["'`])(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0,alias:"attr-value"},punctuation:/\{|\}|=~?|![=~]|,/}},"context-range":[{pattern:/\[[\w\s:]+\]/,inside:{punctuation:/\[|\]|:/,"range-duration":{pattern:/\b(?:\d+(?:[smhdwy]|ms))+\b/i,alias:"number"}}},{pattern:/(\boffset\s+)\w+/,lookbehind:!0,inside:{"range-duration":{pattern:/\b(?:\d+(?:[smhdwy]|ms))+\b/i,alias:"number"}}}],keyword:new RegExp("\\b(?:"+h.join("|")+")\\b","i"),function:/\b[a-z_]\w*(?=\s*\()/i,number:/[-+]?(?:(?:\b\d+(?:\.\d+)?|\B\.\d+)(?:e[-+]?\d+)?\b|\b(?:0x[0-9a-f]+|nan|inf)\b)/i,operator:/[\^*/%+-]|==|!=|<=|<|>=|>|\b(?:and|or|unless)\b/i,punctuation:/[{};()`,.[\]]/}})(r)}return AHe}var IHe,gqt;function lUr(){if(gqt)return IHe;gqt=1,IHe=n,n.displayName="properties",n.aliases=[];function n(r){r.languages.properties={comment:/^[ \t]*[#!].*$/m,"attr-value":{pattern:/(^[ \t]*(?:\\(?:\r\n|[\s\S])|[^\\\s:=])+(?: *[=:] *(?! )| ))(?:\\(?:\r\n|[\s\S])|[^\\\r\n])+/m,lookbehind:!0},"attr-name":/^[ \t]*(?:\\(?:\r\n|[\s\S])|[^\\\s:=])+(?= *[=:]| )/m,punctuation:/[=:]/}}return IHe}var OHe,mqt;function dUr(){if(mqt)return OHe;mqt=1,OHe=n,n.displayName="protobuf",n.aliases=[];function n(r){(function(o){var s=/\b(?:bool|bytes|double|s?fixed(?:32|64)|float|[su]?int(?:32|64)|string)\b/;o.languages.protobuf=o.languages.extend("clike",{"class-name":[{pattern:/(\b(?:enum|extend|message|service)\s+)[A-Za-z_]\w*(?=\s*\{)/,lookbehind:!0},{pattern:/(\b(?:rpc\s+\w+|returns)\s*\(\s*(?:stream\s+)?)\.?[A-Za-z_]\w*(?:\.[A-Za-z_]\w*)*(?=\s*\))/,lookbehind:!0}],keyword:/\b(?:enum|extend|extensions|import|message|oneof|option|optional|package|public|repeated|required|reserved|returns|rpc(?=\s+\w)|service|stream|syntax|to)\b(?!\s*=\s*\d)/,function:/\b[a-z_]\w*(?=\s*\()/i}),o.languages.insertBefore("protobuf","operator",{map:{pattern:/\bmap<\s*[\w.]+\s*,\s*[\w.]+\s*>(?=\s+[a-z_]\w*\s*[=;])/i,alias:"class-name",inside:{punctuation:/[<>.,]/,builtin:s}},builtin:s,"positional-class-name":{pattern:/(?:\b|\B\.)[a-z_]\w*(?:\.[a-z_]\w*)*(?=\s+[a-z_]\w*\s*[=;])/i,alias:"class-name",inside:{punctuation:/\./}},annotation:{pattern:/(\[\s*)[a-z_]\w*(?=\s*=)/i,lookbehind:!0}})})(r)}return OHe}var NHe,bqt;function fUr(){if(bqt)return NHe;bqt=1,NHe=n,n.displayName="psl",n.aliases=[];function n(r){r.languages.psl={comment:{pattern:/#.*/,greedy:!0},string:{pattern:/"(?:\\.|[^\\"])*"/,greedy:!0,inside:{symbol:/\\[ntrbA-Z"\\]/}},"heredoc-string":{pattern:/<<<([a-zA-Z_]\w*)[\r\n](?:.*[\r\n])*?\1\b/,alias:"string",greedy:!0},keyword:/\b(?:__multi|__single|case|default|do|else|elsif|exit|export|for|foreach|function|if|last|line|local|next|requires|return|switch|until|while|word)\b/,constant:/\b(?:ALARM|CHART_ADD_GRAPH|CHART_DELETE_GRAPH|CHART_DESTROY|CHART_LOAD|CHART_PRINT|EOF|OFFLINE|OK|PSL_PROF_LOG|R_CHECK_HORIZ|R_CHECK_VERT|R_CLICKER|R_COLUMN|R_FRAME|R_ICON|R_LABEL|R_LABEL_CENTER|R_LIST_MULTIPLE|R_LIST_MULTIPLE_ND|R_LIST_SINGLE|R_LIST_SINGLE_ND|R_MENU|R_POPUP|R_POPUP_SCROLLED|R_RADIO_HORIZ|R_RADIO_VERT|R_ROW|R_SCALE_HORIZ|R_SCALE_VERT|R_SEP_HORIZ|R_SEP_VERT|R_SPINNER|R_TEXT_FIELD|R_TEXT_FIELD_LABEL|R_TOGGLE|TRIM_LEADING|TRIM_LEADING_AND_TRAILING|TRIM_REDUNDANT|TRIM_TRAILING|VOID|WARN)\b/,boolean:/\b(?:FALSE|False|NO|No|TRUE|True|YES|Yes|false|no|true|yes)\b/,variable:/\b(?:PslDebug|errno|exit_status)\b/,builtin:{pattern:/\b(?:PslExecute|PslFunctionCall|PslFunctionExists|PslSetOptions|_snmp_debug|acos|add_diary|annotate|annotate_get|ascii_to_ebcdic|asctime|asin|atan|atexit|batch_set|blackout|cat|ceil|chan_exists|change_state|close|code_cvt|cond_signal|cond_wait|console_type|convert_base|convert_date|convert_locale_date|cos|cosh|create|date|dcget_text|destroy|destroy_lock|dget_text|difference|dump_hist|ebcdic_to_ascii|encrypt|event_archive|event_catalog_get|event_check|event_query|event_range_manage|event_range_query|event_report|event_schedule|event_trigger|event_trigger2|execute|exists|exp|fabs|file|floor|fmod|fopen|fseek|ftell|full_discovery|get|get_chan_info|get_ranges|get_text|get_vars|getenv|gethostinfo|getpid|getpname|grep|history|history_get_retention|in_transition|index|int|internal|intersection|is_var|isnumber|join|kill|length|lines|lock|lock_info|log|log10|loge|matchline|msg_check|msg_get_format|msg_get_severity|msg_printf|msg_sprintf|ntharg|nthargf|nthline|nthlinef|num_bytes|num_consoles|pconfig|popen|poplines|pow|print|printf|proc_exists|process|random|read|readln|refresh_parameters|remote_check|remote_close|remote_event_query|remote_event_trigger|remote_file_send|remote_open|remove|replace|rindex|sec_check_priv|sec_store_get|sec_store_set|set|set_alarm_ranges|set_locale|share|sin|sinh|sleep|snmp_agent_config|snmp_agent_start|snmp_agent_stop|snmp_close|snmp_config|snmp_get|snmp_get_next|snmp_h_get|snmp_h_get_next|snmp_h_set|snmp_open|snmp_set|snmp_trap_ignore|snmp_trap_listen|snmp_trap_raise_std_trap|snmp_trap_receive|snmp_trap_register_im|snmp_trap_send|snmp_walk|sopen|sort|splitline|sprintf|sqrt|srandom|str_repeat|strcasecmp|subset|substr|system|tail|tan|tanh|text_domain|time|tmpnam|tolower|toupper|trace_psl_process|trim|union|unique|unlock|unset|va_arg|va_start|write)\b/,alias:"builtin-function"},"foreach-variable":{pattern:/(\bforeach\s+(?:(?:\w+\b|"(?:\\.|[^\\"])*")\s+){0,2})[_a-zA-Z]\w*(?=\s*\()/,lookbehind:!0,greedy:!0},function:/\b[_a-z]\w*\b(?=\s*\()/i,number:/\b(?:0x[0-9a-f]+|\d+(?:\.\d+)?)\b/i,operator:/--|\+\+|&&=?|\|\|=?|<<=?|>>=?|[=!]~|[-+*/%&|^!=<>]=?|\.|[:?]/,punctuation:/[(){}\[\];,]/}}return NHe}var RHe,vqt;function hUr(){if(vqt)return RHe;vqt=1,RHe=n,n.displayName="pug",n.aliases=[];function n(r){(function(o){o.languages.pug={comment:{pattern:/(^([\t ]*))\/\/.*(?:(?:\r?\n|\r)\2[\t ].+)*/m,lookbehind:!0},"multiline-script":{pattern:/(^([\t ]*)script\b.*\.[\t ]*)(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/m,lookbehind:!0,inside:o.languages.javascript},filter:{pattern:/(^([\t ]*)):.+(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/m,lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"variable"},text:/\S[\s\S]*/}},"multiline-plain-text":{pattern:/(^([\t ]*)[\w\-#.]+\.[\t ]*)(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/m,lookbehind:!0},markup:{pattern:/(^[\t ]*)<.+/m,lookbehind:!0,inside:o.languages.markup},doctype:{pattern:/((?:^|\n)[\t ]*)doctype(?: .+)?/,lookbehind:!0},"flow-control":{pattern:/(^[\t ]*)(?:case|default|each|else|if|unless|when|while)\b(?: .+)?/m,lookbehind:!0,inside:{each:{pattern:/^each .+? in\b/,inside:{keyword:/\b(?:each|in)\b/,punctuation:/,/}},branch:{pattern:/^(?:case|default|else|if|unless|when|while)\b/,alias:"keyword"},rest:o.languages.javascript}},keyword:{pattern:/(^[\t ]*)(?:append|block|extends|include|prepend)\b.+/m,lookbehind:!0},mixin:[{pattern:/(^[\t ]*)mixin .+/m,lookbehind:!0,inside:{keyword:/^mixin/,function:/\w+(?=\s*\(|\s*$)/,punctuation:/[(),.]/}},{pattern:/(^[\t ]*)\+.+/m,lookbehind:!0,inside:{name:{pattern:/^\+\w+/,alias:"function"},rest:o.languages.javascript}}],script:{pattern:/(^[\t ]*script(?:(?:&[^(]+)?\([^)]+\))*[\t ]).+/m,lookbehind:!0,inside:o.languages.javascript},"plain-text":{pattern:/(^[\t ]*(?!-)[\w\-#.]*[\w\-](?:(?:&[^(]+)?\([^)]+\))*\/?[\t ]).+/m,lookbehind:!0},tag:{pattern:/(^[\t ]*)(?!-)[\w\-#.]*[\w\-](?:(?:&[^(]+)?\([^)]+\))*\/?:?/m,lookbehind:!0,inside:{attributes:[{pattern:/&[^(]+\([^)]+\)/,inside:o.languages.javascript},{pattern:/\([^)]+\)/,inside:{"attr-value":{pattern:/(=\s*(?!\s))(?:\{[^}]*\}|[^,)\r\n]+)/,lookbehind:!0,inside:o.languages.javascript},"attr-name":/[\w-]+(?=\s*!?=|\s*[,)])/,punctuation:/[!=(),]+/}}],punctuation:/:/,"attr-id":/#[\w\-]+/,"attr-class":/\.[\w\-]+/}},code:[{pattern:/(^[\t ]*(?:-|!?=)).+/m,lookbehind:!0,inside:o.languages.javascript}],punctuation:/[.\-!=|]+/};for(var s=/(^([\t ]*)):<filter_name>(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/.source,c=[{filter:"atpl",language:"twig"},{filter:"coffee",language:"coffeescript"},"ejs","handlebars","less","livescript","markdown",{filter:"sass",language:"scss"},"stylus"],d={},h=0,b=c.length;h<b;h++){var y=c[h];y=typeof y=="string"?{filter:y,language:y}:y,o.languages[y.language]&&(d["filter-"+y.filter]={pattern:RegExp(s.replace("<filter_name>",function(){return y.filter}),"m"),lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"variable"},text:{pattern:/\S[\s\S]*/,alias:[y.language,"language-"+y.language],inside:o.languages[y.language]}}})}o.languages.insertBefore("pug","filter",d)})(r)}return RHe}var PHe,wqt;function pUr(){if(wqt)return PHe;wqt=1,PHe=n,n.displayName="puppet",n.aliases=[];function n(r){(function(o){o.languages.puppet={heredoc:[{pattern:/(@\("([^"\r\n\/):]+)"(?:\/[nrts$uL]*)?\).*(?:\r?\n|\r))(?:.*(?:\r?\n|\r(?!\n)))*?[ \t]*(?:\|[ \t]*)?(?:-[ \t]*)?\2/,lookbehind:!0,alias:"string",inside:{punctuation:/(?=\S).*\S(?= *$)/}},{pattern:/(@\(([^"\r\n\/):]+)(?:\/[nrts$uL]*)?\).*(?:\r?\n|\r))(?:.*(?:\r?\n|\r(?!\n)))*?[ \t]*(?:\|[ \t]*)?(?:-[ \t]*)?\2/,lookbehind:!0,greedy:!0,alias:"string",inside:{punctuation:/(?=\S).*\S(?= *$)/}},{pattern:/@\("?(?:[^"\r\n\/):]+)"?(?:\/[nrts$uL]*)?\)/,alias:"string",inside:{punctuation:{pattern:/(\().+?(?=\))/,lookbehind:!0}}}],"multiline-comment":{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0,greedy:!0,alias:"comment"},regex:{pattern:/((?:\bnode\s+|[~=\(\[\{,]\s*|[=+]>\s*|^\s*))\/(?:[^\/\\]|\\[\s\S])+\/(?:[imx]+\b|\B)/,lookbehind:!0,greedy:!0,inside:{"extended-regex":{pattern:/^\/(?:[^\/\\]|\\[\s\S])+\/[im]*x[im]*$/,inside:{comment:/#.*/}}}},comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0,greedy:!0},string:{pattern:/(["'])(?:\$\{(?:[^'"}]|(["'])(?:(?!\2)[^\\]|\\[\s\S])*\2)+\}|\$(?!\{)|(?!\1)[^\\$]|\\[\s\S])*\1/,greedy:!0,inside:{"double-quoted":{pattern:/^"[\s\S]*"$/,inside:{}}}},variable:{pattern:/\$(?:::)?\w+(?:::\w+)*/,inside:{punctuation:/::/}},"attr-name":/(?:\b\w+|\*)(?=\s*=>)/,function:[{pattern:/(\.)(?!\d)\w+/,lookbehind:!0},/\b(?:contain|debug|err|fail|include|info|notice|realize|require|tag|warning)\b|\b(?!\d)\w+(?=\()/],number:/\b(?:0x[a-f\d]+|\d+(?:\.\d+)?(?:e-?\d+)?)\b/i,boolean:/\b(?:false|true)\b/,keyword:/\b(?:application|attr|case|class|consumes|default|define|else|elsif|function|if|import|inherits|node|private|produces|type|undef|unless)\b/,datatype:{pattern:/\b(?:Any|Array|Boolean|Callable|Catalogentry|Class|Collection|Data|Default|Enum|Float|Hash|Integer|NotUndef|Numeric|Optional|Pattern|Regexp|Resource|Runtime|Scalar|String|Struct|Tuple|Type|Undef|Variant)\b/,alias:"symbol"},operator:/=[=~>]?|![=~]?|<(?:<\|?|[=~|-])?|>[>=]?|->?|~>|\|>?>?|[*\/%+?]|\b(?:and|in|or)\b/,punctuation:/[\[\]{}().,;]|:+/};var s=[{pattern:/(^|[^\\])\$\{(?:[^'"{}]|\{[^}]*\}|(["'])(?:(?!\2)[^\\]|\\[\s\S])*\2)+\}/,lookbehind:!0,inside:{"short-variable":{pattern:/(^\$\{)(?!\w+\()(?:::)?\w+(?:::\w+)*/,lookbehind:!0,alias:"variable",inside:{punctuation:/::/}},delimiter:{pattern:/^\$/,alias:"variable"},rest:o.languages.puppet}},{pattern:/(^|[^\\])\$(?:::)?\w+(?:::\w+)*/,lookbehind:!0,alias:"variable",inside:{punctuation:/::/}}];o.languages.puppet.heredoc[0].inside.interpolation=s,o.languages.puppet.string.inside["double-quoted"].inside.interpolation=s})(r)}return PHe}var LHe,yqt;function gUr(){if(yqt)return LHe;yqt=1,LHe=n,n.displayName="pure",n.aliases=[];function n(r){(function(o){o.languages.pure={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0},/#!.+/],"inline-lang":{pattern:/%<[\s\S]+?%>/,greedy:!0,inside:{lang:{pattern:/(^%< *)-\*-.+?-\*-/,lookbehind:!0,alias:"comment"},delimiter:{pattern:/^%<.*|%>$/,alias:"punctuation"}}},string:{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0},number:{pattern:/((?:\.\.)?)(?:\b(?:inf|nan)\b|\b0x[\da-f]+|(?:\b(?:0b)?\d+(?:\.\d+)?|\B\.\d+)(?:e[+-]?\d+)?L?)/i,lookbehind:!0},keyword:/\b(?:NULL|ans|break|bt|case|catch|cd|clear|const|def|del|dump|else|end|exit|extern|false|force|help|if|infix[lr]?|interface|let|ls|mem|namespace|nonfix|of|otherwise|outfix|override|postfix|prefix|private|public|pwd|quit|run|save|show|stats|then|throw|trace|true|type|underride|using|when|with)\b/,function:/\b(?:abs|add_(?:addr|constdef|(?:fundef|interface|macdef|typedef)(?:_at)?|vardef)|all|any|applp?|arity|bigintp?|blob(?:_crc|_size|p)?|boolp?|byte_c?string(?:_pointer)?|byte_(?:matrix|pointer)|calloc|cat|catmap|ceil|char[ps]?|check_ptrtag|chr|clear_sentry|clearsym|closurep?|cmatrixp?|cols?|colcat(?:map)?|colmap|colrev|colvector(?:p|seq)?|complex(?:_float_(?:matrix|pointer)|_matrix(?:_view)?|_pointer|p)?|conj|cookedp?|cst|cstring(?:_(?:dup|list|vector))?|curry3?|cyclen?|del_(?:constdef|fundef|interface|macdef|typedef|vardef)|delete|diag(?:mat)?|dim|dmatrixp?|do|double(?:_matrix(?:_view)?|_pointer|p)?|dowith3?|drop|dropwhile|eval(?:cmd)?|exactp|filter|fix|fixity|flip|float(?:_matrix|_pointer)|floor|fold[lr]1?|frac|free|funp?|functionp?|gcd|get(?:_(?:byte|constdef|double|float|fundef|int(?:64)?|interface(?:_typedef)?|long|macdef|pointer|ptrtag|sentry|short|string|typedef|vardef))?|globsym|hash|head|id|im|imatrixp?|index|inexactp|infp|init|insert|int(?:_matrix(?:_view)?|_pointer|p)?|int64_(?:matrix|pointer)|integerp?|iteraten?|iterwhile|join|keys?|lambdap?|last(?:err(?:pos)?)?|lcd|list[2p]?|listmap|make_ptrtag|malloc|map|matcat|matrixp?|max|member|min|nanp|nargs|nmatrixp?|null|numberp?|ord|pack(?:ed)?|pointer(?:_cast|_tag|_type|p)?|pow|pred|ptrtag|put(?:_(?:byte|double|float|int(?:64)?|long|pointer|short|string))?|rationalp?|re|realp?|realloc|recordp?|redim|reduce(?:_with)?|refp?|repeatn?|reverse|rlistp?|round|rows?|rowcat(?:map)?|rowmap|rowrev|rowvector(?:p|seq)?|same|scan[lr]1?|sentry|sgn|short_(?:matrix|pointer)|slice|smatrixp?|sort|split|str|strcat|stream|stride|string(?:_(?:dup|list|vector)|p)?|subdiag(?:mat)?|submat|subseq2?|substr|succ|supdiag(?:mat)?|symbolp?|tail|take|takewhile|thunkp?|transpose|trunc|tuplep?|typep|ubyte|uint(?:64)?|ulong|uncurry3?|unref|unzip3?|update|ushort|vals?|varp?|vector(?:p|seq)?|void|zip3?|zipwith3?)\b/,special:{pattern:/\b__[a-z]+__\b/i,alias:"builtin"},operator:/(?:[!"#$%&'*+,\-.\/:<=>?@\\^`|~\u00a1-\u00bf\u00d7-\u00f7\u20d0-\u2bff]|\b_+\b)+|\b(?:and|div|mod|not|or)\b/,punctuation:/[(){}\[\];,|]/};var s=["c",{lang:"c++",alias:"cpp"},"fortran"],c=/%< *-\*- *<lang>\d* *-\*-[\s\S]+?%>/.source;s.forEach(function(d){var h=d;if(typeof d!="string"&&(h=d.alias,d=d.lang),o.languages[h]){var b={};b["inline-lang-"+h]={pattern:RegExp(c.replace("<lang>",d.replace(/([.+*?\/\\(){}\[\]])/g,"\\$1")),"i"),inside:o.util.clone(o.languages.pure["inline-lang"].inside)},b["inline-lang-"+h].inside.rest=o.util.clone(o.languages[h]),o.languages.insertBefore("pure","inline-lang",b)}}),o.languages.c&&(o.languages.pure["inline-lang"].inside.rest=o.util.clone(o.languages.c))})(r)}return LHe}var DHe,Eqt;function mUr(){if(Eqt)return DHe;Eqt=1,DHe=n,n.displayName="purebasic",n.aliases=[];function n(r){r.languages.purebasic=r.languages.extend("clike",{comment:/;.*/,keyword:/\b(?:align|and|as|break|calldebugger|case|compilercase|compilerdefault|compilerelse|compilerelseif|compilerendif|compilerendselect|compilererror|compilerif|compilerselect|continue|data|datasection|debug|debuglevel|declare|declarec|declarecdll|declaredll|declaremodule|default|define|dim|disableasm|disabledebugger|disableexplicit|else|elseif|enableasm|enabledebugger|enableexplicit|end|enddatasection|enddeclaremodule|endenumeration|endif|endimport|endinterface|endmacro|endmodule|endprocedure|endselect|endstructure|endstructureunion|endwith|enumeration|extends|fakereturn|for|foreach|forever|global|gosub|goto|if|import|importc|includebinary|includefile|includepath|interface|macro|module|newlist|newmap|next|not|or|procedure|procedurec|procedurecdll|proceduredll|procedurereturn|protected|prototype|prototypec|read|redim|repeat|restore|return|runtime|select|shared|static|step|structure|structureunion|swap|threaded|to|until|wend|while|with|xincludefile|xor)\b/i,function:/\b\w+(?:\.\w+)?\s*(?=\()/,number:/(?:\$[\da-f]+|\b-?(?:\d+(?:\.\d+)?|\.\d+)(?:e[+-]?\d+)?)\b/i,operator:/(?:@\*?|\?|\*)\w+|-[>-]?|\+\+?|!=?|<<?=?|>>?=?|==?|&&?|\|?\||[~^%?*/@]/}),r.languages.insertBefore("purebasic","keyword",{tag:/#\w+\$?/,asm:{pattern:/(^[\t ]*)!.*/m,lookbehind:!0,alias:"tag",inside:{comment:/;.*/,string:{pattern:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},"label-reference-anonymous":{pattern:/(!\s*j[a-z]+\s+)@[fb]/i,lookbehind:!0,alias:"fasm-label"},"label-reference-addressed":{pattern:/(!\s*j[a-z]+\s+)[A-Z._?$@][\w.?$@~#]*/i,lookbehind:!0,alias:"fasm-label"},keyword:[/\b(?:extern|global)\b[^;\r\n]*/i,/\b(?:CPU|DEFAULT|FLOAT)\b.*/],function:{pattern:/^([\t ]*!\s*)[\da-z]+(?=\s|$)/im,lookbehind:!0},"function-inline":{pattern:/(:\s*)[\da-z]+(?=\s)/i,lookbehind:!0,alias:"function"},label:{pattern:/^([\t ]*!\s*)[A-Za-z._?$@][\w.?$@~#]*(?=:)/m,lookbehind:!0,alias:"fasm-label"},register:/\b(?:st\d|[xyz]mm\d\d?|[cdt]r\d|r\d\d?[bwd]?|[er]?[abcd]x|[abcd][hl]|[er]?(?:bp|di|si|sp)|[cdefgs]s|mm\d+)\b/i,number:/(?:\b|-|(?=\$))(?:0[hx](?:[\da-f]*\.)?[\da-f]+(?:p[+-]?\d+)?|\d[\da-f]+[hx]|\$\d[\da-f]*|0[oq][0-7]+|[0-7]+[oq]|0[by][01]+|[01]+[by]|0[dt]\d+|(?:\d+(?:\.\d+)?|\.\d+)(?:\.?e[+-]?\d+)?[dt]?)\b/i,operator:/[\[\]*+\-/%<>=&|$!,.:]/}}}),delete r.languages.purebasic["class-name"],delete r.languages.purebasic.boolean,r.languages.pbfasm=r.languages.purebasic}return DHe}var MHe,Sqt;function bUr(){if(Sqt)return MHe;Sqt=1;var n=EQe();MHe=r,r.displayName="purescript",r.aliases=["purs"];function r(o){o.register(n),o.languages.purescript=o.languages.extend("haskell",{keyword:/\b(?:ado|case|class|data|derive|do|else|forall|if|in|infixl|infixr|instance|let|module|newtype|of|primitive|then|type|where)\b|∀/,"import-statement":{pattern:/(^[\t ]*)import\s+[A-Z][\w']*(?:\.[A-Z][\w']*)*(?:\s+as\s+[A-Z][\w']*(?:\.[A-Z][\w']*)*)?(?:\s+hiding\b)?/m,lookbehind:!0,inside:{keyword:/\b(?:as|hiding|import)\b/,punctuation:/\./}},builtin:/\b(?:absurd|add|ap|append|apply|between|bind|bottom|clamp|compare|comparing|compose|conj|const|degree|discard|disj|div|eq|flap|flip|gcd|identity|ifM|join|lcm|liftA1|liftM1|map|max|mempty|min|mod|mul|negate|not|notEq|one|otherwise|recip|show|sub|top|unit|unless|unlessM|void|when|whenM|zero)\b/,operator:[o.languages.haskell.operator[0],o.languages.haskell.operator[2],/[\xa2-\xa6\xa8\xa9\xac\xae-\xb1\xb4\xb8\xd7\xf7\u02c2-\u02c5\u02d2-\u02df\u02e5-\u02eb\u02ed\u02ef-\u02ff\u0375\u0384\u0385\u03f6\u0482\u058d-\u058f\u0606-\u0608\u060b\u060e\u060f\u06de\u06e9\u06fd\u06fe\u07f6\u07fe\u07ff\u09f2\u09f3\u09fa\u09fb\u0af1\u0b70\u0bf3-\u0bfa\u0c7f\u0d4f\u0d79\u0e3f\u0f01-\u0f03\u0f13\u0f15-\u0f17\u0f1a-\u0f1f\u0f34\u0f36\u0f38\u0fbe-\u0fc5\u0fc7-\u0fcc\u0fce\u0fcf\u0fd5-\u0fd8\u109e\u109f\u1390-\u1399\u166d\u17db\u1940\u19de-\u19ff\u1b61-\u1b6a\u1b74-\u1b7c\u1fbd\u1fbf-\u1fc1\u1fcd-\u1fcf\u1fdd-\u1fdf\u1fed-\u1fef\u1ffd\u1ffe\u2044\u2052\u207a-\u207c\u208a-\u208c\u20a0-\u20bf\u2100\u2101\u2103-\u2106\u2108\u2109\u2114\u2116-\u2118\u211e-\u2123\u2125\u2127\u2129\u212e\u213a\u213b\u2140-\u2144\u214a-\u214d\u214f\u218a\u218b\u2190-\u2307\u230c-\u2328\u232b-\u2426\u2440-\u244a\u249c-\u24e9\u2500-\u2767\u2794-\u27c4\u27c7-\u27e5\u27f0-\u2982\u2999-\u29d7\u29dc-\u29fb\u29fe-\u2b73\u2b76-\u2b95\u2b97-\u2bff\u2ce5-\u2cea\u2e50\u2e51\u2e80-\u2e99\u2e9b-\u2ef3\u2f00-\u2fd5\u2ff0-\u2ffb\u3004\u3012\u3013\u3020\u3036\u3037\u303e\u303f\u309b\u309c\u3190\u3191\u3196-\u319f\u31c0-\u31e3\u3200-\u321e\u322a-\u3247\u3250\u3260-\u327f\u328a-\u32b0\u32c0-\u33ff\u4dc0-\u4dff\ua490-\ua4c6\ua700-\ua716\ua720\ua721\ua789\ua78a\ua828-\ua82b\ua836-\ua839\uaa77-\uaa79\uab5b\uab6a\uab6b\ufb29\ufbb2-\ufbc1\ufdfc\ufdfd\ufe62\ufe64-\ufe66\ufe69\uff04\uff0b\uff1c-\uff1e\uff3e\uff40\uff5c\uff5e\uffe0-\uffe6\uffe8-\uffee\ufffc\ufffd]/]}),o.languages.purs=o.languages.purescript}return MHe}var jHe,_qt;function vUr(){if(_qt)return jHe;_qt=1,jHe=n,n.displayName="python",n.aliases=["py"];function n(r){r.languages.python={comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0,greedy:!0},"string-interpolation":{pattern:/(?:f|fr|rf)(?:("""|''')[\s\S]*?\1|("|')(?:\\.|(?!\2)[^\\\r\n])*\2)/i,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^{])(?:\{\{)*)\{(?!\{)(?:[^{}]|\{(?!\{)(?:[^{}]|\{(?!\{)(?:[^{}])+\})+\})+\}/,lookbehind:!0,inside:{"format-spec":{pattern:/(:)[^:(){}]+(?=\}$)/,lookbehind:!0},"conversion-option":{pattern:/![sra](?=[:}]$)/,alias:"punctuation"},rest:null}},string:/[\s\S]+/}},"triple-quoted-string":{pattern:/(?:[rub]|br|rb)?("""|''')[\s\S]*?\1/i,greedy:!0,alias:"string"},string:{pattern:/(?:[rub]|br|rb)?("|')(?:\\.|(?!\1)[^\\\r\n])*\1/i,greedy:!0},function:{pattern:/((?:^|\s)def[ \t]+)[a-zA-Z_]\w*(?=\s*\()/g,lookbehind:!0},"class-name":{pattern:/(\bclass\s+)\w+/i,lookbehind:!0},decorator:{pattern:/(^[\t ]*)@\w+(?:\.\w+)*/m,lookbehind:!0,alias:["annotation","punctuation"],inside:{punctuation:/\./}},keyword:/\b(?:_(?=\s*:)|and|as|assert|async|await|break|case|class|continue|def|del|elif|else|except|exec|finally|for|from|global|if|import|in|is|lambda|match|nonlocal|not|or|pass|print|raise|return|try|while|with|yield)\b/,builtin:/\b(?:__import__|abs|all|any|apply|ascii|basestring|bin|bool|buffer|bytearray|bytes|callable|chr|classmethod|cmp|coerce|compile|complex|delattr|dict|dir|divmod|enumerate|eval|execfile|file|filter|float|format|frozenset|getattr|globals|hasattr|hash|help|hex|id|input|int|intern|isinstance|issubclass|iter|len|list|locals|long|map|max|memoryview|min|next|object|oct|open|ord|pow|property|range|raw_input|reduce|reload|repr|reversed|round|set|setattr|slice|sorted|staticmethod|str|sum|super|tuple|type|unichr|unicode|vars|xrange|zip)\b/,boolean:/\b(?:False|None|True)\b/,number:/\b0(?:b(?:_?[01])+|o(?:_?[0-7])+|x(?:_?[a-f0-9])+)\b|(?:\b\d+(?:_\d+)*(?:\.(?:\d+(?:_\d+)*)?)?|\B\.\d+(?:_\d+)*)(?:e[+-]?\d+(?:_\d+)*)?j?(?!\w)/i,operator:/[-+%=]=?|!=|:=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]/,punctuation:/[{}[\];(),.:]/},r.languages.python["string-interpolation"].inside.interpolation.inside.rest=r.languages.python,r.languages.py=r.languages.python}return jHe}var $He,kqt;function wUr(){if(kqt)return $He;kqt=1,$He=n,n.displayName="q",n.aliases=[];function n(r){r.languages.q={string:/"(?:\\.|[^"\\\r\n])*"/,comment:[{pattern:/([\t )\]}])\/.*/,lookbehind:!0,greedy:!0},{pattern:/(^|\r?\n|\r)\/[\t ]*(?:(?:\r?\n|\r)(?:.*(?:\r?\n|\r(?!\n)))*?(?:\\(?=[\t ]*(?:\r?\n|\r))|$)|\S.*)/,lookbehind:!0,greedy:!0},{pattern:/^\\[\t ]*(?:\r?\n|\r)[\s\S]+/m,greedy:!0},{pattern:/^#!.+/m,greedy:!0}],symbol:/`(?::\S+|[\w.]*)/,datetime:{pattern:/0N[mdzuvt]|0W[dtz]|\d{4}\.\d\d(?:m|\.\d\d(?:T(?:\d\d(?::\d\d(?::\d\d(?:[.:]\d\d\d)?)?)?)?)?[dz]?)|\d\d:\d\d(?::\d\d(?:[.:]\d\d\d)?)?[uvt]?/,alias:"number"},number:/\b(?![01]:)(?:0N[hje]?|0W[hj]?|0[wn]|0x[\da-fA-F]+|\d+(?:\.\d*)?(?:e[+-]?\d+)?[hjfeb]?)/,keyword:/\\\w+\b|\b(?:abs|acos|aj0?|all|and|any|asc|asin|asof|atan|attr|avgs?|binr?|by|ceiling|cols|cor|cos|count|cov|cross|csv|cut|delete|deltas|desc|dev|differ|distinct|div|do|dsave|ej|enlist|eval|except|exec|exit|exp|fby|fills|first|fkeys|flip|floor|from|get|getenv|group|gtime|hclose|hcount|hdel|hopen|hsym|iasc|identity|idesc|if|ij|in|insert|inter|inv|keys?|last|like|list|ljf?|load|log|lower|lsq|ltime|ltrim|mavg|maxs?|mcount|md5|mdev|med|meta|mins?|mmax|mmin|mmu|mod|msum|neg|next|not|null|or|over|parse|peach|pj|plist|prds?|prev|prior|rand|rank|ratios|raze|read0|read1|reciprocal|reval|reverse|rload|rotate|rsave|rtrim|save|scan|scov|sdev|select|set|setenv|show|signum|sin|sqrt|ssr?|string|sublist|sums?|sv|svar|system|tables|tan|til|trim|txf|type|uj|ungroup|union|update|upper|upsert|value|var|views?|vs|wavg|where|while|within|wj1?|wsum|ww|xasc|xbar|xcols?|xdesc|xexp|xgroup|xkey|xlog|xprev|xrank)\b/,adverb:{pattern:/['\/\\]:?|\beach\b/,alias:"function"},verb:{pattern:/(?:\B\.\B|\b[01]:|<[=>]?|>=?|[:+\-*%,!?~=|$&#@^]):?|\b_\b:?/,alias:"operator"},punctuation:/[(){}\[\];.]/}}return $He}var FHe,Tqt;function yUr(){if(Tqt)return FHe;Tqt=1,FHe=n,n.displayName="qml",n.aliases=[];function n(r){(function(o){for(var s=/"(?:\\.|[^\\"\r\n])*"|'(?:\\.|[^\\'\r\n])*'/.source,c=/\/\/.*(?!.)|\/\*(?:[^*]|\*(?!\/))*\*\//.source,d=/(?:[^\\()[\]{}"'/]|<string>|\/(?![*/])|<comment>|\(<expr>*\)|\[<expr>*\]|\{<expr>*\}|\\[\s\S])/.source.replace(/<string>/g,function(){return s}).replace(/<comment>/g,function(){return c}),h=0;h<2;h++)d=d.replace(/<expr>/g,function(){return d});d=d.replace(/<expr>/g,"[^\\s\\S]"),o.languages.qml={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},"javascript-function":{pattern:RegExp(/((?:^|;)[ \t]*)function\s+(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*\s*\(<js>*\)\s*\{<js>*\}/.source.replace(/<js>/g,function(){return d}),"m"),lookbehind:!0,greedy:!0,alias:"language-javascript",inside:o.languages.javascript},"class-name":{pattern:/((?:^|[:;])[ \t]*)(?!\d)\w+(?=[ \t]*\{|[ \t]+on\b)/m,lookbehind:!0},property:[{pattern:/((?:^|[;{])[ \t]*)(?!\d)\w+(?:\.\w+)*(?=[ \t]*:)/m,lookbehind:!0},{pattern:/((?:^|[;{])[ \t]*)property[ \t]+(?!\d)\w+(?:\.\w+)*[ \t]+(?!\d)\w+(?:\.\w+)*(?=[ \t]*:)/m,lookbehind:!0,inside:{keyword:/^property/,property:/\w+(?:\.\w+)*/}}],"javascript-expression":{pattern:RegExp(/(:[ \t]*)(?![\s;}[])(?:(?!$|[;}])<js>)+/.source.replace(/<js>/g,function(){return d}),"m"),lookbehind:!0,greedy:!0,alias:"language-javascript",inside:o.languages.javascript},string:{pattern:/"(?:\\.|[^\\"\r\n])*"/,greedy:!0},keyword:/\b(?:as|import|on)\b/,punctuation:/[{}[\]:;,]/}})(r)}return FHe}var BHe,xqt;function EUr(){if(xqt)return BHe;xqt=1,BHe=n,n.displayName="qore",n.aliases=[];function n(r){r.languages.qore=r.languages.extend("clike",{comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:\/\/|#).*)/,lookbehind:!0},string:{pattern:/("|')(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0},keyword:/\b(?:abstract|any|assert|binary|bool|boolean|break|byte|case|catch|char|class|code|const|continue|data|default|do|double|else|enum|extends|final|finally|float|for|goto|hash|if|implements|import|inherits|instanceof|int|interface|long|my|native|new|nothing|null|object|our|own|private|reference|rethrow|return|short|soft(?:bool|date|float|int|list|number|string)|static|strictfp|string|sub|super|switch|synchronized|this|throw|throws|transient|try|void|volatile|while)\b/,boolean:/\b(?:false|true)\b/i,function:/\$?\b(?!\d)\w+(?=\()/,number:/\b(?:0b[01]+|0x(?:[\da-f]*\.)?[\da-fp\-]+|(?:\d+(?:\.\d+)?|\.\d+)(?:e\d+)?[df]|(?:\d+(?:\.\d+)?|\.\d+))\b/i,operator:{pattern:/(^|[^.])(?:\+[+=]?|-[-=]?|[!=](?:==?|~)?|>>?=?|<(?:=>?|<=?)?|&[&=]?|\|[|=]?|[*\/%^]=?|[~?])/,lookbehind:!0},variable:/\$(?!\d)\w+\b/})}return BHe}var UHe,Cqt;function SUr(){if(Cqt)return UHe;Cqt=1,UHe=n,n.displayName="qsharp",n.aliases=["qs"];function n(r){(function(o){function s(D,$){return D.replace(/<<(\d+)>>/g,function(U,B){return"(?:"+$[+B]+")"})}function c(D,$,U){return RegExp(s(D,$),"")}function d(D,$){for(var U=0;U<$;U++)D=D.replace(/<<self>>/g,function(){return"(?:"+D+")"});return D.replace(/<<self>>/g,"[^\\s\\S]")}var h={type:"Adj BigInt Bool Ctl Double false Int One Pauli PauliI PauliX PauliY PauliZ Qubit Range Result String true Unit Zero",other:"Adjoint adjoint apply as auto body borrow borrowing Controlled controlled distribute elif else fail fixup for function if in internal intrinsic invert is let mutable namespace new newtype open operation repeat return self set until use using while within"};function b(D){return"\\b(?:"+D.trim().replace(/ /g,"|")+")\\b"}var y=RegExp(b(h.type+" "+h.other)),E=/\b[A-Za-z_]\w*\b/.source,k=s(/<<0>>(?:\s*\.\s*<<0>>)*/.source,[E]),x={keyword:y,punctuation:/[<>()?,.:[\]]/},O=/"(?:\\.|[^\\"])*"/.source;o.languages.qsharp=o.languages.extend("clike",{comment:/\/\/.*/,string:[{pattern:c(/(^|[^$\\])<<0>>/.source,[O]),lookbehind:!0,greedy:!0}],"class-name":[{pattern:c(/(\b(?:as|open)\s+)<<0>>(?=\s*(?:;|as\b))/.source,[k]),lookbehind:!0,inside:x},{pattern:c(/(\bnamespace\s+)<<0>>(?=\s*\{)/.source,[k]),lookbehind:!0,inside:x}],keyword:y,number:/(?:\b0(?:x[\da-f]+|b[01]+|o[0-7]+)|(?:\B\.\d+|\b\d+(?:\.\d*)?)(?:e[-+]?\d+)?)l?\b/i,operator:/\band=|\bor=|\band\b|\bnot\b|\bor\b|<[-=]|[-=]>|>>>=?|<<<=?|\^\^\^=?|\|\|\|=?|&&&=?|w\/=?|~~~|[*\/+\-^=!%]=?/,punctuation:/::|[{}[\];(),.:]/}),o.languages.insertBefore("qsharp","number",{range:{pattern:/\.\./,alias:"operator"}});var P=d(s(/\{(?:[^"{}]|<<0>>|<<self>>)*\}/.source,[O]),2);o.languages.insertBefore("qsharp","string",{"interpolation-string":{pattern:c(/\$"(?:\\.|<<0>>|[^\\"{])*"/.source,[P]),greedy:!0,inside:{interpolation:{pattern:c(/((?:^|[^\\])(?:\\\\)*)<<0>>/.source,[P]),lookbehind:!0,inside:{punctuation:/^\{|\}$/,expression:{pattern:/[\s\S]+/,alias:"language-qsharp",inside:o.languages.qsharp}}},string:/[\s\S]+/}}})})(r),r.languages.qs=r.languages.qsharp}return UHe}var zHe,Aqt;function _Ur(){if(Aqt)return zHe;Aqt=1,zHe=n,n.displayName="r",n.aliases=[];function n(r){r.languages.r={comment:/#.*/,string:{pattern:/(['"])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},"percent-operator":{pattern:/%[^%\s]*%/,alias:"operator"},boolean:/\b(?:FALSE|TRUE)\b/,ellipsis:/\.\.(?:\.|\d+)/,number:[/\b(?:Inf|NaN)\b/,/(?:\b0x[\dA-Fa-f]+(?:\.\d*)?|\b\d+(?:\.\d*)?|\B\.\d+)(?:[EePp][+-]?\d+)?[iL]?/],keyword:/\b(?:NA|NA_character_|NA_complex_|NA_integer_|NA_real_|NULL|break|else|for|function|if|in|next|repeat|while)\b/,operator:/->?>?|<(?:=|<?-)?|[>=!]=?|::?|&&?|\|\|?|[+*\/^$@~]/,punctuation:/[(){}\[\],;]/}}return zHe}var HHe,Iqt;function kUr(){if(Iqt)return HHe;Iqt=1;var n=TQe();HHe=r,r.displayName="racket",r.aliases=["rkt"];function r(o){o.register(n),o.languages.racket=o.languages.extend("scheme",{"lambda-parameter":{pattern:/([(\[]lambda\s+[(\[])[^()\[\]'\s]+/,lookbehind:!0}}),o.languages.insertBefore("racket","string",{lang:{pattern:/^#lang.+/m,greedy:!0,alias:"keyword"}}),o.languages.rkt=o.languages.racket}return HHe}var GHe,Oqt;function TUr(){if(Oqt)return GHe;Oqt=1,GHe=n,n.displayName="reason",n.aliases=[];function n(r){r.languages.reason=r.languages.extend("clike",{string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^\\\r\n"])*"/,greedy:!0},"class-name":/\b[A-Z]\w*/,keyword:/\b(?:and|as|assert|begin|class|constraint|do|done|downto|else|end|exception|external|for|fun|function|functor|if|in|include|inherit|initializer|lazy|let|method|module|mutable|new|nonrec|object|of|open|or|private|rec|sig|struct|switch|then|to|try|type|val|virtual|when|while|with)\b/,operator:/\.{3}|:[:=]|\|>|->|=(?:==?|>)?|<=?|>=?|[|^?'#!~`]|[+\-*\/]\.?|\b(?:asr|land|lor|lsl|lsr|lxor|mod)\b/}),r.languages.insertBefore("reason","class-name",{char:{pattern:/'(?:\\x[\da-f]{2}|\\o[0-3][0-7][0-7]|\\\d{3}|\\.|[^'\\\r\n])'/,greedy:!0},constructor:/\b[A-Z]\w*\b(?!\s*\.)/,label:{pattern:/\b[a-z]\w*(?=::)/,alias:"symbol"}}),delete r.languages.reason.function}return GHe}var qHe,Nqt;function xUr(){if(Nqt)return qHe;Nqt=1,qHe=n,n.displayName="regex",n.aliases=[];function n(r){(function(o){var s={pattern:/\\[\\(){}[\]^$+*?|.]/,alias:"escape"},c=/\\(?:x[\da-fA-F]{2}|u[\da-fA-F]{4}|u\{[\da-fA-F]+\}|0[0-7]{0,2}|[123][0-7]{2}|c[a-zA-Z]|.)/,d={pattern:/\.|\\[wsd]|\\p\{[^{}]+\}/i,alias:"class-name"},h={pattern:/\\[wsd]|\\p\{[^{}]+\}/i,alias:"class-name"},b="(?:[^\\\\-]|"+c.source+")",y=RegExp(b+"-"+b),E={pattern:/(<|')[^<>']+(?=[>']$)/,lookbehind:!0,alias:"variable"};o.languages.regex={"char-class":{pattern:/((?:^|[^\\])(?:\\\\)*)\[(?:[^\\\]]|\\[\s\S])*\]/,lookbehind:!0,inside:{"char-class-negation":{pattern:/(^\[)\^/,lookbehind:!0,alias:"operator"},"char-class-punctuation":{pattern:/^\[|\]$/,alias:"punctuation"},range:{pattern:y,inside:{escape:c,"range-punctuation":{pattern:/-/,alias:"operator"}}},"special-escape":s,"char-set":h,escape:c}},"special-escape":s,"char-set":d,backreference:[{pattern:/\\(?![123][0-7]{2})[1-9]/,alias:"keyword"},{pattern:/\\k<[^<>']+>/,alias:"keyword",inside:{"group-name":E}}],anchor:{pattern:/[$^]|\\[ABbGZz]/,alias:"function"},escape:c,group:[{pattern:/\((?:\?(?:<[^<>']+>|'[^<>']+'|[>:]|<?[=!]|[idmnsuxU]+(?:-[idmnsuxU]+)?:?))?/,alias:"punctuation",inside:{"group-name":E}},{pattern:/\)/,alias:"punctuation"}],quantifier:{pattern:/(?:[+*?]|\{\d+(?:,\d*)?\})[?+]?/,alias:"number"},alternation:{pattern:/\|/,alias:"keyword"}}})(r)}return qHe}var VHe,Rqt;function CUr(){if(Rqt)return VHe;Rqt=1,VHe=n,n.displayName="rego",n.aliases=[];function n(r){r.languages.rego={comment:/#.*/,property:{pattern:/(^|[^\\.])(?:"(?:\\.|[^\\"\r\n])*"|`[^`]*`|\b[a-z_]\w*\b)(?=\s*:(?!=))/i,lookbehind:!0,greedy:!0},string:{pattern:/(^|[^\\])"(?:\\.|[^\\"\r\n])*"|`[^`]*`/,lookbehind:!0,greedy:!0},keyword:/\b(?:as|default|else|import|not|null|package|set(?=\s*\()|some|with)\b/,boolean:/\b(?:false|true)\b/,function:{pattern:/\b[a-z_]\w*\b(?:\s*\.\s*\b[a-z_]\w*\b)*(?=\s*\()/i,inside:{namespace:/\b\w+\b(?=\s*\.)/,punctuation:/\./}},number:/-?\b\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,operator:/[-+*/%|&]|[<>:=]=?|!=|\b_\b/,punctuation:/[,;.\[\]{}()]/}}return VHe}var WHe,Pqt;function AUr(){if(Pqt)return WHe;Pqt=1,WHe=n,n.displayName="renpy",n.aliases=["rpy"];function n(r){r.languages.renpy={comment:{pattern:/(^|[^\\])#.+/,lookbehind:!0},string:{pattern:/("""|''')[\s\S]+?\1|("|')(?:\\.|(?!\2)[^\\])*\2|(?:^#?(?:(?:[0-9a-fA-F]){3}|[0-9a-fA-F]{6})$)/m,greedy:!0},function:/\b[a-z_]\w*(?=\()/i,property:/\b(?:Update|UpdateVersion|action|activate_sound|adv_nvl_transition|after_load_transition|align|alpha|alt|anchor|antialias|area|auto|background|bar_invert|bar_resizing|bar_vertical|black_color|bold|bottom_bar|bottom_gutter|bottom_margin|bottom_padding|box_reverse|box_wrap|can_update|caret|child|color|crop|default_afm_enable|default_afm_time|default_fullscreen|default_text_cps|developer|directory_name|drag_handle|drag_joined|drag_name|drag_raise|draggable|dragged|drop_shadow|drop_shadow_color|droppable|dropped|easein|easeout|edgescroll|end_game_transition|end_splash_transition|enter_replay_transition|enter_sound|enter_transition|enter_yesno_transition|executable_name|exit_replay_transition|exit_sound|exit_transition|exit_yesno_transition|fadein|fadeout|first_indent|first_spacing|fit_first|focus|focus_mask|font|foreground|game_main_transition|get_installed_packages|google_play_key|google_play_salt|ground|has_music|has_sound|has_voice|height|help|hinting|hover|hover_background|hover_color|hover_sound|hovered|hyperlink_functions|idle|idle_color|image_style|include_update|insensitive|insensitive_background|insensitive_color|inside|intra_transition|italic|justify|kerning|keyboard_focus|language|layer_clipping|layers|layout|left_bar|left_gutter|left_margin|left_padding|length|line_leading|line_overlap_split|line_spacing|linear|main_game_transition|main_menu_music|maximum|min_width|minimum|minwidth|modal|mouse|mousewheel|name|narrator_menu|newline_indent|nvl_adv_transition|offset|order_reverse|outlines|overlay_functions|pos|position|prefix|radius|range|rest_indent|right_bar|right_gutter|right_margin|right_padding|rotate|rotate_pad|ruby_style|sample_sound|save_directory|say_attribute_transition|screen_height|screen_width|scrollbars|selected_hover|selected_hover_color|selected_idle|selected_idle_color|selected_insensitive|show_side_image|show_two_window|side_spacing|side_xpos|side_ypos|size|size_group|slow_cps|slow_cps_multiplier|spacing|strikethrough|subpixel|text_align|text_style|text_xpos|text_y_fudge|text_ypos|thumb|thumb_offset|thumb_shadow|thumbnail_height|thumbnail_width|time|top_bar|top_gutter|top_margin|top_padding|translations|underline|unscrollable|update|value|version|version_name|version_tuple|vertical|width|window_hide_transition|window_icon|window_left_padding|window_show_transition|window_title|windows_icon|xadjustment|xalign|xanchor|xanchoraround|xaround|xcenter|xfill|xinitial|xmargin|xmaximum|xminimum|xoffset|xofsset|xpadding|xpos|xsize|xzoom|yadjustment|yalign|yanchor|yanchoraround|yaround|ycenter|yfill|yinitial|ymargin|ymaximum|yminimum|yoffset|ypadding|ypos|ysize|ysizexysize|yzoom|zoom|zorder)\b/,tag:/\b(?:bar|block|button|buttoscreenn|drag|draggroup|fixed|frame|grid|[hv]box|hotbar|hotspot|image|imagebutton|imagemap|input|key|label|menu|mm_menu_frame|mousearea|nvl|parallel|screen|self|side|tag|text|textbutton|timer|vbar|viewport|window)\b|\$/,keyword:/\b(?:None|add|adjustment|alignaround|allow|angle|animation|around|as|assert|behind|box_layout|break|build|cache|call|center|changed|child_size|choice|circles|class|clear|clicked|clipping|clockwise|config|contains|continue|corner1|corner2|counterclockwise|def|default|define|del|delay|disabled|disabled_text|dissolve|elif|else|event|except|exclude|exec|expression|fade|finally|for|from|function|global|gm_root|has|hide|id|if|import|in|init|is|jump|knot|lambda|left|less_rounded|mm_root|movie|music|null|on|onlayer|pass|pause|persistent|play|print|python|queue|raise|random|renpy|repeat|return|right|rounded_window|scene|scope|set|show|slow|slow_abortable|slow_done|sound|stop|store|style|style_group|substitute|suffix|theme|transform|transform_anchor|transpose|try|ui|unhovered|updater|use|voice|while|widget|widget_hover|widget_selected|widget_text|yield)\b/,boolean:/\b(?:[Ff]alse|[Tt]rue)\b/,number:/(?:\b(?:0[bo])?(?:(?:\d|0x[\da-f])[\da-f]*(?:\.\d*)?)|\B\.\d+)(?:e[+-]?\d+)?j?/i,operator:/[-+%=]=?|!=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]|\b(?:and|at|not|or|with)\b/,punctuation:/[{}[\];(),.:]/},r.languages.rpy=r.languages.renpy}return WHe}var KHe,Lqt;function IUr(){if(Lqt)return KHe;Lqt=1,KHe=n,n.displayName="rest",n.aliases=[];function n(r){r.languages.rest={table:[{pattern:/(^[\t ]*)(?:\+[=-]+)+\+(?:\r?\n|\r)(?:\1[+|].+[+|](?:\r?\n|\r))+\1(?:\+[=-]+)+\+/m,lookbehind:!0,inside:{punctuation:/\||(?:\+[=-]+)+\+/}},{pattern:/(^[\t ]*)=+ [ =]*=(?:(?:\r?\n|\r)\1.+)+(?:\r?\n|\r)\1=+ [ =]*=(?=(?:\r?\n|\r){2}|\s*$)/m,lookbehind:!0,inside:{punctuation:/[=-]+/}}],"substitution-def":{pattern:/(^[\t ]*\.\. )\|(?:[^|\s](?:[^|]*[^|\s])?)\| [^:]+::/m,lookbehind:!0,inside:{substitution:{pattern:/^\|(?:[^|\s]|[^|\s][^|]*[^|\s])\|/,alias:"attr-value",inside:{punctuation:/^\||\|$/}},directive:{pattern:/( )(?! )[^:]+::/,lookbehind:!0,alias:"function",inside:{punctuation:/::$/}}}},"link-target":[{pattern:/(^[\t ]*\.\. )\[[^\]]+\]/m,lookbehind:!0,alias:"string",inside:{punctuation:/^\[|\]$/}},{pattern:/(^[\t ]*\.\. )_(?:`[^`]+`|(?:[^:\\]|\\.)+):/m,lookbehind:!0,alias:"string",inside:{punctuation:/^_|:$/}}],directive:{pattern:/(^[\t ]*\.\. )[^:]+::/m,lookbehind:!0,alias:"function",inside:{punctuation:/::$/}},comment:{pattern:/(^[\t ]*\.\.)(?:(?: .+)?(?:(?:\r?\n|\r).+)+| .+)(?=(?:\r?\n|\r){2}|$)/m,lookbehind:!0},title:[{pattern:/^(([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\2+)(?:\r?\n|\r).+(?:\r?\n|\r)\1$/m,inside:{punctuation:/^[!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]+|[!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]+$/,important:/.+/}},{pattern:/(^|(?:\r?\n|\r){2}).+(?:\r?\n|\r)([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\2+(?=\r?\n|\r|$)/,lookbehind:!0,inside:{punctuation:/[!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]+$/,important:/.+/}}],hr:{pattern:/((?:\r?\n|\r){2})([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\2{3,}(?=(?:\r?\n|\r){2})/,lookbehind:!0,alias:"punctuation"},field:{pattern:/(^[\t ]*):[^:\r\n]+:(?= )/m,lookbehind:!0,alias:"attr-name"},"command-line-option":{pattern:/(^[\t ]*)(?:[+-][a-z\d]|(?:--|\/)[a-z\d-]+)(?:[ =](?:[a-z][\w-]*|<[^<>]+>))?(?:, (?:[+-][a-z\d]|(?:--|\/)[a-z\d-]+)(?:[ =](?:[a-z][\w-]*|<[^<>]+>))?)*(?=(?:\r?\n|\r)? {2,}\S)/im,lookbehind:!0,alias:"symbol"},"literal-block":{pattern:/::(?:\r?\n|\r){2}([ \t]+)(?![ \t]).+(?:(?:\r?\n|\r)\1.+)*/,inside:{"literal-block-punctuation":{pattern:/^::/,alias:"punctuation"}}},"quoted-literal-block":{pattern:/::(?:\r?\n|\r){2}([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]).*(?:(?:\r?\n|\r)\1.*)*/,inside:{"literal-block-punctuation":{pattern:/^(?:::|([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\1*)/m,alias:"punctuation"}}},"list-bullet":{pattern:/(^[\t ]*)(?:[*+\-•‣⁃]|\(?(?:\d+|[a-z]|[ivxdclm]+)\)|(?:\d+|[a-z]|[ivxdclm]+)\.)(?= )/im,lookbehind:!0,alias:"punctuation"},"doctest-block":{pattern:/(^[\t ]*)>>> .+(?:(?:\r?\n|\r).+)*/m,lookbehind:!0,inside:{punctuation:/^>>>/}},inline:[{pattern:/(^|[\s\-:\/'"<(\[{])(?::[^:]+:`.*?`|`.*?`:[^:]+:|(\*\*?|``?|\|)(?!\s)(?:(?!\2).)*\S\2(?=[\s\-.,:;!?\\\/'")\]}]|$))/m,lookbehind:!0,inside:{bold:{pattern:/(^\*\*).+(?=\*\*$)/,lookbehind:!0},italic:{pattern:/(^\*).+(?=\*$)/,lookbehind:!0},"inline-literal":{pattern:/(^``).+(?=``$)/,lookbehind:!0,alias:"symbol"},role:{pattern:/^:[^:]+:|:[^:]+:$/,alias:"function",inside:{punctuation:/^:|:$/}},"interpreted-text":{pattern:/(^`).+(?=`$)/,lookbehind:!0,alias:"attr-value"},substitution:{pattern:/(^\|).+(?=\|$)/,lookbehind:!0,alias:"attr-value"},punctuation:/\*\*?|``?|\|/}}],link:[{pattern:/\[[^\[\]]+\]_(?=[\s\-.,:;!?\\\/'")\]}]|$)/,alias:"string",inside:{punctuation:/^\[|\]_$/}},{pattern:/(?:\b[a-z\d]+(?:[_.:+][a-z\d]+)*_?_|`[^`]+`_?_|_`[^`]+`)(?=[\s\-.,:;!?\\\/'")\]}]|$)/i,alias:"string",inside:{punctuation:/^_?`|`$|`?_?_$/}}],punctuation:{pattern:/(^[\t ]*)(?:\|(?= |$)|(?:---?|—|\.\.|__)(?= )|\.\.$)/m,lookbehind:!0}}}return KHe}var YHe,Dqt;function OUr(){if(Dqt)return YHe;Dqt=1,YHe=n,n.displayName="rip",n.aliases=[];function n(r){r.languages.rip={comment:{pattern:/#.*/,greedy:!0},char:{pattern:/\B`[^\s`'",.:;#\/\\()<>\[\]{}]\b/,greedy:!0},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},regex:{pattern:/(^|[^/])\/(?!\/)(?:\[[^\n\r\]]*\]|\\.|[^/\\\r\n\[])+\/(?=\s*(?:$|[\r\n,.;})]))/,lookbehind:!0,greedy:!0},keyword:/(?:=>|->)|\b(?:case|catch|class|else|exit|finally|if|raise|return|switch|try)\b/,builtin:/@|\bSystem\b/,boolean:/\b(?:false|true)\b/,date:/\b\d{4}-\d{2}-\d{2}\b/,time:/\b\d{2}:\d{2}:\d{2}\b/,datetime:/\b\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\b/,symbol:/:[^\d\s`'",.:;#\/\\()<>\[\]{}][^\s`'",.:;#\/\\()<>\[\]{}]*/,number:/[+-]?\b(?:\d+\.\d+|\d+)\b/,punctuation:/(?:\.{2,3})|[`,.:;=\/\\()<>\[\]{}]/,reference:/[^\d\s`'",.:;#\/\\()<>\[\]{}][^\s`'",.:;#\/\\()<>\[\]{}]*/}}return YHe}var XHe,Mqt;function NUr(){if(Mqt)return XHe;Mqt=1,XHe=n,n.displayName="roboconf",n.aliases=[];function n(r){r.languages.roboconf={comment:/#.*/,keyword:{pattern:/(^|\s)(?:(?:external|import)\b|(?:facet|instance of)(?=[ \t]+[\w-]+[ \t]*\{))/,lookbehind:!0},component:{pattern:/[\w-]+(?=[ \t]*\{)/,alias:"variable"},property:/[\w.-]+(?=[ \t]*:)/,value:{pattern:/(=[ \t]*(?![ \t]))[^,;]+/,lookbehind:!0,alias:"attr-value"},optional:{pattern:/\(optional\)/,alias:"builtin"},wildcard:{pattern:/(\.)\*/,lookbehind:!0,alias:"operator"},punctuation:/[{},.;:=]/}}return XHe}var JHe,jqt;function RUr(){if(jqt)return JHe;jqt=1,JHe=n,n.displayName="robotframework",n.aliases=[];function n(r){(function(o){var s={pattern:/(^[ \t]*| {2}|\t)#.*/m,lookbehind:!0,greedy:!0},c={pattern:/((?:^|[^\\])(?:\\{2})*)[$@&%]\{(?:[^{}\r\n]|\{[^{}\r\n]*\})*\}/,lookbehind:!0,inside:{punctuation:/^[$@&%]\{|\}$/}};function d(E,k){var x={};x["section-header"]={pattern:/^ ?\*{3}.+?\*{3}/,alias:"keyword"};for(var O in k)x[O]=k[O];return x.tag={pattern:/([\r\n](?: {2}|\t)[ \t]*)\[[-\w]+\]/,lookbehind:!0,inside:{punctuation:/\[|\]/}},x.variable=c,x.comment=s,{pattern:RegExp(/^ ?\*{3}[ \t]*<name>[ \t]*\*{3}(?:.|[\r\n](?!\*{3}))*/.source.replace(/<name>/g,function(){return E}),"im"),alias:"section",inside:x}}var h={pattern:/(\[Documentation\](?: {2}|\t)[ \t]*)(?![ \t]|#)(?:.|(?:\r\n?|\n)[ \t]*\.{3})+/,lookbehind:!0,alias:"string"},b={pattern:/([\r\n] ?)(?!#)(?:\S(?:[ \t]\S)*)+/,lookbehind:!0,alias:"function",inside:{variable:c}},y={pattern:/([\r\n](?: {2}|\t)[ \t]*)(?!\[|\.{3}|#)(?:\S(?:[ \t]\S)*)+/,lookbehind:!0,inside:{variable:c}};o.languages.robotframework={settings:d("Settings",{documentation:{pattern:/([\r\n] ?Documentation(?: {2}|\t)[ \t]*)(?![ \t]|#)(?:.|(?:\r\n?|\n)[ \t]*\.{3})+/,lookbehind:!0,alias:"string"},property:{pattern:/([\r\n] ?)(?!\.{3}|#)(?:\S(?:[ \t]\S)*)+/,lookbehind:!0}}),variables:d("Variables"),"test-cases":d("Test Cases",{"test-name":b,documentation:h,property:y}),keywords:d("Keywords",{"keyword-name":b,documentation:h,property:y}),tasks:d("Tasks",{"task-name":b,documentation:h,property:y}),comment:s},o.languages.robot=o.languages.robotframework})(r)}return JHe}var ZHe,$qt;function PUr(){if($qt)return ZHe;$qt=1,ZHe=n,n.displayName="rust",n.aliases=[];function n(r){(function(o){for(var s=/\/\*(?:[^*/]|\*(?!\/)|\/(?!\*)|<self>)*\*\//.source,c=0;c<2;c++)s=s.replace(/<self>/g,function(){return s});s=s.replace(/<self>/g,function(){return/[^\s\S]/.source}),o.languages.rust={comment:[{pattern:RegExp(/(^|[^\\])/.source+s),lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/b?"(?:\\[\s\S]|[^\\"])*"|b?r(#*)"(?:[^"]|"(?!\1))*"\1/,greedy:!0},char:{pattern:/b?'(?:\\(?:x[0-7][\da-fA-F]|u\{(?:[\da-fA-F]_*){1,6}\}|.)|[^\\\r\n\t'])'/,greedy:!0},attribute:{pattern:/#!?\[(?:[^\[\]"]|"(?:\\[\s\S]|[^\\"])*")*\]/,greedy:!0,alias:"attr-name",inside:{string:null}},"closure-params":{pattern:/([=(,:]\s*|\bmove\s*)\|[^|]*\||\|[^|]*\|(?=\s*(?:\{|->))/,lookbehind:!0,greedy:!0,inside:{"closure-punctuation":{pattern:/^\||\|$/,alias:"punctuation"},rest:null}},"lifetime-annotation":{pattern:/'\w+/,alias:"symbol"},"fragment-specifier":{pattern:/(\$\w+:)[a-z]+/,lookbehind:!0,alias:"punctuation"},variable:/\$\w+/,"function-definition":{pattern:/(\bfn\s+)\w+/,lookbehind:!0,alias:"function"},"type-definition":{pattern:/(\b(?:enum|struct|trait|type|union)\s+)\w+/,lookbehind:!0,alias:"class-name"},"module-declaration":[{pattern:/(\b(?:crate|mod)\s+)[a-z][a-z_\d]*/,lookbehind:!0,alias:"namespace"},{pattern:/(\b(?:crate|self|super)\s*)::\s*[a-z][a-z_\d]*\b(?:\s*::(?:\s*[a-z][a-z_\d]*\s*::)*)?/,lookbehind:!0,alias:"namespace",inside:{punctuation:/::/}}],keyword:[/\b(?:Self|abstract|as|async|await|become|box|break|const|continue|crate|do|dyn|else|enum|extern|final|fn|for|if|impl|in|let|loop|macro|match|mod|move|mut|override|priv|pub|ref|return|self|static|struct|super|trait|try|type|typeof|union|unsafe|unsized|use|virtual|where|while|yield)\b/,/\b(?:bool|char|f(?:32|64)|[ui](?:8|16|32|64|128|size)|str)\b/],function:/\b[a-z_]\w*(?=\s*(?:::\s*<|\())/,macro:{pattern:/\b\w+!/,alias:"property"},constant:/\b[A-Z_][A-Z_\d]+\b/,"class-name":/\b[A-Z]\w*\b/,namespace:{pattern:/(?:\b[a-z][a-z_\d]*\s*::\s*)*\b[a-z][a-z_\d]*\s*::(?!\s*<)/,inside:{punctuation:/::/}},number:/\b(?:0x[\dA-Fa-f](?:_?[\dA-Fa-f])*|0o[0-7](?:_?[0-7])*|0b[01](?:_?[01])*|(?:(?:\d(?:_?\d)*)?\.)?\d(?:_?\d)*(?:[Ee][+-]?\d+)?)(?:_?(?:f32|f64|[iu](?:8|16|32|64|size)?))?\b/,boolean:/\b(?:false|true)\b/,punctuation:/->|\.\.=|\.{1,3}|::|[{}[\];(),:]/,operator:/[-+*\/%!^]=?|=[=>]?|&[&=]?|\|[|=]?|<<?=?|>>?=?|[@?]/},o.languages.rust["closure-params"].inside.rest=o.languages.rust,o.languages.rust.attribute.inside.string=o.languages.rust.string})(r)}return ZHe}var QHe,Fqt;function LUr(){if(Fqt)return QHe;Fqt=1,QHe=n,n.displayName="sas",n.aliases=[];function n(r){(function(o){var s=/(?:"(?:""|[^"])*"(?!")|'(?:''|[^'])*'(?!'))/.source,c=/\b(?:\d[\da-f]*x|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/i,d={pattern:RegExp(s+"[bx]"),alias:"number"},h={pattern:/&[a-z_]\w*/i},b={pattern:/((?:^|\s|=|\())%(?:ABORT|BY|CMS|COPY|DISPLAY|DO|ELSE|END|EVAL|GLOBAL|GO|GOTO|IF|INC|INCLUDE|INDEX|INPUT|KTRIM|LENGTH|LET|LIST|LOCAL|PUT|QKTRIM|QSCAN|QSUBSTR|QSYSFUNC|QUPCASE|RETURN|RUN|SCAN|SUBSTR|SUPERQ|SYMDEL|SYMEXIST|SYMGLOBL|SYMLOCAL|SYSCALL|SYSEVALF|SYSEXEC|SYSFUNC|SYSGET|SYSRPUT|THEN|TO|TSO|UNQUOTE|UNTIL|UPCASE|WHILE|WINDOW)\b/i,lookbehind:!0,alias:"keyword"},y={pattern:/(^|\s)(?:proc\s+\w+|data(?!=)|quit|run)\b/i,alias:"keyword",lookbehind:!0},E=[/\/\*[\s\S]*?\*\//,{pattern:/(^[ \t]*|;\s*)\*[^;]*;/m,lookbehind:!0}],k={pattern:RegExp(s),greedy:!0},x=/[$%@.(){}\[\];,\\]/,O={pattern:/%?\b\w+(?=\()/,alias:"keyword"},P={function:O,"arg-value":{pattern:/(=\s*)[A-Z\.]+/i,lookbehind:!0},operator:/=/,"macro-variable":h,arg:{pattern:/[A-Z]+/i,alias:"keyword"},number:c,"numeric-constant":d,punctuation:x,string:k},D={pattern:/\b(?:format|put)\b=?[\w'$.]+/i,inside:{keyword:/^(?:format|put)(?==)/i,equals:/=/,format:{pattern:/(?:\w|\$\d)+\.\d?/,alias:"number"}}},$={pattern:/\b(?:format|put)\s+[\w']+(?:\s+[$.\w]+)+(?=;)/i,inside:{keyword:/^(?:format|put)/i,format:{pattern:/[\w$]+\.\d?/,alias:"number"}}},U={pattern:/((?:^|\s)=?)(?:catname|checkpoint execute_always|dm|endsas|filename|footnote|%include|libname|%list|lock|missing|options|page|resetline|%run|sasfile|skip|sysecho|title\d?)\b/i,lookbehind:!0,alias:"keyword"},B={pattern:/(^|\s)(?:submit(?:\s+(?:load|norun|parseonly))?|endsubmit)\b/i,lookbehind:!0,alias:"keyword"},G=/aStore|accessControl|aggregation|audio|autotune|bayesianNetClassifier|bioMedImage|boolRule|builtins|cardinality|cdm|clustering|conditionalRandomFields|configuration|copula|countreg|dataDiscovery|dataPreprocess|dataSciencePilot|dataStep|decisionTree|deduplication|deepLearn|deepNeural|deepRnn|ds2|ecm|entityRes|espCluster|explainModel|factmac|fastKnn|fcmpact|fedSql|freqTab|gVarCluster|gam|gleam|graphSemiSupLearn|hiddenMarkovModel|hyperGroup|ica|image|iml|kernalPca|langModel|ldaTopic|loadStreams|mbc|mixed|mlTools|modelPublishing|network|neuralNet|nmf|nonParametricBayes|nonlinear|optNetwork|optimization|panel|pca|percentile|phreg|pls|qkb|qlim|quantreg|recommend|regression|reinforcementLearn|robustPca|ruleMining|sampling|sandwich|sccasl|search(?:Analytics)?|sentimentAnalysis|sequence|session(?:Prop)?|severity|simSystem|simple|smartData|sparkEmbeddedProcess|sparseML|spatialreg|spc|stabilityMonitoring|svDataDescription|svm|table|text(?:Filters|Frequency|Mining|Parse|Rule(?:Develop|Score)|Topic|Util)|timeData|transpose|tsInfo|tsReconcile|uniTimeSeries|varReduce/.source,K={pattern:RegExp(/(^|\s)(?:action\s+)?(?:<act>)\.[a-z]+\b[^;]+/.source.replace(/<act>/g,function(){return G}),"i"),lookbehind:!0,inside:{keyword:RegExp(/(?:<act>)\.[a-z]+\b/.source.replace(/<act>/g,function(){return G}),"i"),action:{pattern:/(?:action)/i,alias:"keyword"},comment:E,function:O,"arg-value":P["arg-value"],operator:P.operator,argument:P.arg,number:c,"numeric-constant":d,punctuation:x,string:k}},X={pattern:/((?:^|\s)=?)(?:after|analysis|and|array|barchart|barwidth|begingraph|by|call|cas|cbarline|cfill|class(?:lev)?|close|column|computed?|contains|continue|data(?==)|define|delete|describe|document|do\s+over|do|dol|drop|dul|else|end(?:comp|source)?|entryTitle|eval(?:uate)?|exec(?:ute)?|exit|file(?:name)?|fill(?:attrs)?|flist|fnc|function(?:list)?|global|goto|group(?:by)?|headline|headskip|histogram|if|infile|keep|keylabel|keyword|label|layout|leave|legendlabel|length|libname|loadactionset|merge|midpoints|_?null_|name|noobs|nowd|ods|options|or|otherwise|out(?:put)?|over(?:lay)?|plot|print|put|raise|ranexp|rannor|rbreak|retain|return|select|session|sessref|set|source|statgraph|sum|summarize|table|temp|terminate|then\s+do|then|title\d?|to|var|when|where|xaxisopts|y2axisopts|yaxisopts)\b/i,lookbehind:!0};o.languages.sas={datalines:{pattern:/^([ \t]*)(?:cards|(?:data)?lines);[\s\S]+?^[ \t]*;/im,lookbehind:!0,alias:"string",inside:{keyword:{pattern:/^(?:cards|(?:data)?lines)/i},punctuation:/;/}},"proc-sql":{pattern:/(^proc\s+(?:fed)?sql(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|data|quit|run);|(?![\s\S]))/im,lookbehind:!0,inside:{sql:{pattern:RegExp(/^[ \t]*(?:select|alter\s+table|(?:create|describe|drop)\s+(?:index|table(?:\s+constraints)?|view)|create\s+unique\s+index|insert\s+into|update)(?:<str>|[^;"'])+;/.source.replace(/<str>/g,function(){return s}),"im"),alias:"language-sql",inside:o.languages.sql},"global-statements":U,"sql-statements":{pattern:/(^|\s)(?:disconnect\s+from|begin|commit|exec(?:ute)?|reset|rollback|validate)\b/i,lookbehind:!0,alias:"keyword"},number:c,"numeric-constant":d,punctuation:x,string:k}},"proc-groovy":{pattern:/(^proc\s+groovy(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|data|quit|run);|(?![\s\S]))/im,lookbehind:!0,inside:{comment:E,groovy:{pattern:RegExp(/(^[ \t]*submit(?:\s+(?:load|norun|parseonly))?)(?:<str>|[^"'])+?(?=endsubmit;)/.source.replace(/<str>/g,function(){return s}),"im"),lookbehind:!0,alias:"language-groovy",inside:o.languages.groovy},keyword:X,"submit-statement":B,"global-statements":U,number:c,"numeric-constant":d,punctuation:x,string:k}},"proc-lua":{pattern:/(^proc\s+lua(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|data|quit|run);|(?![\s\S]))/im,lookbehind:!0,inside:{comment:E,lua:{pattern:RegExp(/(^[ \t]*submit(?:\s+(?:load|norun|parseonly))?)(?:<str>|[^"'])+?(?=endsubmit;)/.source.replace(/<str>/g,function(){return s}),"im"),lookbehind:!0,alias:"language-lua",inside:o.languages.lua},keyword:X,"submit-statement":B,"global-statements":U,number:c,"numeric-constant":d,punctuation:x,string:k}},"proc-cas":{pattern:/(^proc\s+cas(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|quit|data);|(?![\s\S]))/im,lookbehind:!0,inside:{comment:E,"statement-var":{pattern:/((?:^|\s)=?)saveresult\s[^;]+/im,lookbehind:!0,inside:{statement:{pattern:/^saveresult\s+\S+/i,inside:{keyword:/^(?:saveresult)/i}},rest:P}},"cas-actions":K,statement:{pattern:/((?:^|\s)=?)(?:default|(?:un)?set|on|output|upload)[^;]+/im,lookbehind:!0,inside:P},step:y,keyword:X,function:O,format:D,altformat:$,"global-statements":U,number:c,"numeric-constant":d,punctuation:x,string:k}},"proc-args":{pattern:RegExp(/(^proc\s+\w+\s+)(?!\s)(?:[^;"']|<str>)+;/.source.replace(/<str>/g,function(){return s}),"im"),lookbehind:!0,inside:P},"macro-keyword":b,"macro-variable":h,"macro-string-functions":{pattern:/((?:^|\s|=))%(?:BQUOTE|NRBQUOTE|NRQUOTE|NRSTR|QUOTE|STR)\(.*?(?:[^%]\))/i,lookbehind:!0,inside:{function:{pattern:/%(?:BQUOTE|NRBQUOTE|NRQUOTE|NRSTR|QUOTE|STR)/i,alias:"keyword"},"macro-keyword":b,"macro-variable":h,"escaped-char":{pattern:/%['"()<>=¬^~;,#]/},punctuation:x}},"macro-declaration":{pattern:/^%macro[^;]+(?=;)/im,inside:{keyword:/%macro/i}},"macro-end":{pattern:/^%mend[^;]+(?=;)/im,inside:{keyword:/%mend/i}},macro:{pattern:/%_\w+(?=\()/,alias:"keyword"},input:{pattern:/\binput\s[-\w\s/*.$&]+;/i,inside:{input:{alias:"keyword",pattern:/^input/i},comment:E,number:c,"numeric-constant":d}},"options-args":{pattern:/(^options)[-'"|/\\<>*+=:()\w\s]*(?=;)/im,lookbehind:!0,inside:P},"cas-actions":K,comment:E,function:O,format:D,altformat:$,"numeric-constant":d,datetime:{pattern:RegExp(s+"(?:dt?|t)"),alias:"number"},string:k,step:y,keyword:X,"operator-keyword":{pattern:/\b(?:eq|ge|gt|in|le|lt|ne|not)\b/i,alias:"operator"},number:c,operator:/\*\*?|\|\|?|!!?|¦¦?|<[>=]?|>[<=]?|[-+\/=&]|[~¬^]=?/,punctuation:x}})(r)}return QHe}var eGe,Bqt;function DUr(){if(Bqt)return eGe;Bqt=1,eGe=n,n.displayName="sass",n.aliases=[];function n(r){(function(o){o.languages.sass=o.languages.extend("css",{comment:{pattern:/^([ \t]*)\/[\/*].*(?:(?:\r?\n|\r)\1[ \t].+)*/m,lookbehind:!0,greedy:!0}}),o.languages.insertBefore("sass","atrule",{"atrule-line":{pattern:/^(?:[ \t]*)[@+=].+/m,greedy:!0,inside:{atrule:/(?:@[\w-]+|[+=])/}}}),delete o.languages.sass.atrule;var s=/\$[-\w]+|#\{\$[-\w]+\}/,c=[/[+*\/%]|[=!]=|<=?|>=?|\b(?:and|not|or)\b/,{pattern:/(\s)-(?=\s)/,lookbehind:!0}];o.languages.insertBefore("sass","property",{"variable-line":{pattern:/^[ \t]*\$.+/m,greedy:!0,inside:{punctuation:/:/,variable:s,operator:c}},"property-line":{pattern:/^[ \t]*(?:[^:\s]+ *:.*|:[^:\s].*)/m,greedy:!0,inside:{property:[/[^:\s]+(?=\s*:)/,{pattern:/(:)[^:\s]+/,lookbehind:!0}],punctuation:/:/,variable:s,operator:c,important:o.languages.sass.important}}}),delete o.languages.sass.property,delete o.languages.sass.important,o.languages.insertBefore("sass","punctuation",{selector:{pattern:/^([ \t]*)\S(?:,[^,\r\n]+|[^,\r\n]*)(?:,[^,\r\n]+)*(?:,(?:\r?\n|\r)\1[ \t]+\S(?:,[^,\r\n]+|[^,\r\n]*)(?:,[^,\r\n]+)*)*/m,lookbehind:!0,greedy:!0}})})(r)}return eGe}var tGe,Uqt;function MUr(){if(Uqt)return tGe;Uqt=1;var n=SQe();tGe=r,r.displayName="scala",r.aliases=[];function r(o){o.register(n),o.languages.scala=o.languages.extend("java",{"triple-quoted-string":{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string"},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:/<-|=>|\b(?:abstract|case|catch|class|def|do|else|extends|final|finally|for|forSome|if|implicit|import|lazy|match|new|null|object|override|package|private|protected|return|sealed|self|super|this|throw|trait|try|type|val|var|while|with|yield)\b/,number:/\b0x(?:[\da-f]*\.)?[\da-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e\d+)?[dfl]?/i,builtin:/\b(?:Any|AnyRef|AnyVal|Boolean|Byte|Char|Double|Float|Int|Long|Nothing|Short|String|Unit)\b/,symbol:/'[^\d\s\\]\w*/}),o.languages.insertBefore("scala","triple-quoted-string",{"string-interpolation":{pattern:/\b[a-z]\w*(?:"""(?:[^$]|\$(?:[^{]|\{(?:[^{}]|\{[^{}]*\})*\}))*?"""|"(?:[^$"\r\n]|\$(?:[^{]|\{(?:[^{}]|\{[^{}]*\})*\}))*")/i,greedy:!0,inside:{id:{pattern:/^\w+/,greedy:!0,alias:"function"},escape:{pattern:/\\\$"|\$[$"]/,greedy:!0,alias:"symbol"},interpolation:{pattern:/\$(?:\w+|\{(?:[^{}]|\{[^{}]*\})*\})/,greedy:!0,inside:{punctuation:/^\$\{?|\}$/,expression:{pattern:/[\s\S]+/,inside:o.languages.scala}}},string:/[\s\S]+/}}}),delete o.languages.scala["class-name"],delete o.languages.scala.function}return tGe}var nGe,zqt;function jUr(){if(zqt)return nGe;zqt=1,nGe=n,n.displayName="scss",n.aliases=[];function n(r){r.languages.scss=r.languages.extend("css",{comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0},atrule:{pattern:/@[\w-](?:\([^()]+\)|[^()\s]|\s+(?!\s))*?(?=\s+[{;])/,inside:{rule:/@[\w-]+/}},url:/(?:[-a-z]+-)?url(?=\()/i,selector:{pattern:/(?=\S)[^@;{}()]?(?:[^@;{}()\s]|\s+(?!\s)|#\{\$[-\w]+\})+(?=\s*\{(?:\}|\s|[^}][^:{}]*[:{][^}]))/,inside:{parent:{pattern:/&/,alias:"important"},placeholder:/%[-\w]+/,variable:/\$[-\w]+|#\{\$[-\w]+\}/}},property:{pattern:/(?:[-\w]|\$[-\w]|#\{\$[-\w]+\})+(?=\s*:)/,inside:{variable:/\$[-\w]+|#\{\$[-\w]+\}/}}}),r.languages.insertBefore("scss","atrule",{keyword:[/@(?:content|debug|each|else(?: if)?|extend|for|forward|function|if|import|include|mixin|return|use|warn|while)\b/i,{pattern:/( )(?:from|through)(?= )/,lookbehind:!0}]}),r.languages.insertBefore("scss","important",{variable:/\$[-\w]+|#\{\$[-\w]+\}/}),r.languages.insertBefore("scss","function",{"module-modifier":{pattern:/\b(?:as|hide|show|with)\b/i,alias:"keyword"},placeholder:{pattern:/%[-\w]+/,alias:"selector"},statement:{pattern:/\B!(?:default|optional)\b/i,alias:"keyword"},boolean:/\b(?:false|true)\b/,null:{pattern:/\bnull\b/,alias:"keyword"},operator:{pattern:/(\s)(?:[-+*\/%]|[=!]=|<=?|>=?|and|not|or)(?=\s)/,lookbehind:!0}}),r.languages.scss.atrule.inside.rest=r.languages.scss}return nGe}var rGe,Hqt;function $Ur(){if(Hqt)return rGe;Hqt=1;var n=adn();rGe=r,r.displayName="shellSession",r.aliases=[];function r(o){o.register(n),function(s){var c=[/"(?:\\[\s\S]|\$\([^)]+\)|\$(?!\()|`[^`]+`|[^"\\`$])*"/.source,/'[^']*'/.source,/\$'(?:[^'\\]|\\[\s\S])*'/.source,/<<-?\s*(["']?)(\w+)\1\s[\s\S]*?[\r\n]\2/.source].join("|");s.languages["shell-session"]={command:{pattern:RegExp(/^/.source+"(?:"+(/[^\s@:$#%*!/\\]+@[^\r\n@:$#%*!/\\]+(?::[^\0-\x1F$#%*?"<>:;|]+)?/.source+"|"+/[/~.][^\0-\x1F$#%*?"<>@:;|]*/.source)+")?"+/[$#%](?=\s)/.source+/(?:[^\\\r\n \t'"<$]|[ \t](?:(?!#)|#.*$)|\\(?:[^\r]|\r\n?)|\$(?!')|<(?!<)|<<str>>)+/.source.replace(/<<str>>/g,function(){return c}),"m"),greedy:!0,inside:{info:{pattern:/^[^#$%]+/,alias:"punctuation",inside:{user:/^[^\s@:$#%*!/\\]+@[^\r\n@:$#%*!/\\]+/,punctuation:/:/,path:/[\s\S]+/}},bash:{pattern:/(^[$#%]\s*)\S[\s\S]*/,lookbehind:!0,alias:"language-bash",inside:s.languages.bash},"shell-symbol":{pattern:/^[$#%]/,alias:"important"}}},output:/.(?:.*(?:[\r\n]|.$))*/},s.languages["sh-session"]=s.languages.shellsession=s.languages["shell-session"]}(o)}return rGe}var iGe,Gqt;function FUr(){if(Gqt)return iGe;Gqt=1,iGe=n,n.displayName="smali",n.aliases=[];function n(r){r.languages.smali={comment:/#.*/,string:{pattern:/"(?:[^\r\n\\"]|\\.)*"|'(?:[^\r\n\\']|\\(?:.|u[\da-fA-F]{4}))'/,greedy:!0},"class-name":{pattern:/(^|[^L])L(?:(?:\w+|`[^`\r\n]*`)\/)*(?:[\w$]+|`[^`\r\n]*`)(?=\s*;)/,lookbehind:!0,inside:{"class-name":{pattern:/(^L|\/)(?:[\w$]+|`[^`\r\n]*`)$/,lookbehind:!0},namespace:{pattern:/^(L)(?:(?:\w+|`[^`\r\n]*`)\/)+/,lookbehind:!0,inside:{punctuation:/\//}},builtin:/^L/}},builtin:[{pattern:/([();\[])[BCDFIJSVZ]+/,lookbehind:!0},{pattern:/([\w$>]:)[BCDFIJSVZ]/,lookbehind:!0}],keyword:[{pattern:/(\.end\s+)[\w-]+/,lookbehind:!0},{pattern:/(^|[^\w.-])\.(?!\d)[\w-]+/,lookbehind:!0},{pattern:/(^|[^\w.-])(?:abstract|annotation|bridge|constructor|enum|final|interface|private|protected|public|runtime|static|synthetic|system|transient)(?![\w.-])/,lookbehind:!0}],function:{pattern:/(^|[^\w.-])(?:\w+|<[\w$-]+>)(?=\()/,lookbehind:!0},field:{pattern:/[\w$]+(?=:)/,alias:"variable"},register:{pattern:/(^|[^\w.-])[vp]\d(?![\w.-])/,lookbehind:!0,alias:"variable"},boolean:{pattern:/(^|[^\w.-])(?:false|true)(?![\w.-])/,lookbehind:!0},number:{pattern:/(^|[^/\w.-])-?(?:NAN|INFINITY|0x(?:[\dA-F]+(?:\.[\dA-F]*)?|\.[\dA-F]+)(?:p[+-]?[\dA-F]+)?|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?)[dflst]?(?![\w.-])/i,lookbehind:!0},label:{pattern:/(:)\w+/,lookbehind:!0,alias:"property"},operator:/->|\.\.|[\[=]/,punctuation:/[{}(),;:]/}}return iGe}var oGe,qqt;function BUr(){if(qqt)return oGe;qqt=1,oGe=n,n.displayName="smalltalk",n.aliases=[];function n(r){r.languages.smalltalk={comment:{pattern:/"(?:""|[^"])*"/,greedy:!0},char:{pattern:/\$./,greedy:!0},string:{pattern:/'(?:''|[^'])*'/,greedy:!0},symbol:/#[\da-z]+|#(?:-|([+\/\\*~<>=@%|&?!])\1?)|#(?=\()/i,"block-arguments":{pattern:/(\[\s*):[^\[|]*\|/,lookbehind:!0,inside:{variable:/:[\da-z]+/i,punctuation:/\|/}},"temporary-variables":{pattern:/\|[^|]+\|/,inside:{variable:/[\da-z]+/i,punctuation:/\|/}},keyword:/\b(?:new|nil|self|super)\b/,boolean:/\b(?:false|true)\b/,number:[/\d+r-?[\dA-Z]+(?:\.[\dA-Z]+)?(?:e-?\d+)?/,/\b\d+(?:\.\d+)?(?:e-?\d+)?/],operator:/[<=]=?|:=|~[~=]|\/\/?|\\\\|>[>=]?|[!^+\-*&|,@]/,punctuation:/[.;:?\[\](){}]/}}return oGe}var aGe,Vqt;function UUr(){if(Vqt)return aGe;Vqt=1;var n=oy();aGe=r,r.displayName="smarty",r.aliases=[];function r(o){o.register(n),function(s){s.languages.smarty={comment:{pattern:/^\{\*[\s\S]*?\*\}/,greedy:!0},"embedded-php":{pattern:/^\{php\}[\s\S]*?\{\/php\}/,greedy:!0,inside:{smarty:{pattern:/^\{php\}|\{\/php\}$/,inside:null},php:{pattern:/[\s\S]+/,alias:"language-php",inside:s.languages.php}}},string:[{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0,inside:{interpolation:{pattern:/\{[^{}]*\}|`[^`]*`/,inside:{"interpolation-punctuation":{pattern:/^[{`]|[`}]$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:null}}},variable:/\$\w+/}},{pattern:/'(?:\\.|[^'\\\r\n])*'/,greedy:!0}],keyword:{pattern:/(^\{\/?)[a-z_]\w*\b(?!\()/i,lookbehind:!0,greedy:!0},delimiter:{pattern:/^\{\/?|\}$/,greedy:!0,alias:"punctuation"},number:/\b0x[\dA-Fa-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][-+]?\d+)?/,variable:[/\$(?!\d)\w+/,/#(?!\d)\w+#/,{pattern:/(\.|->|\w\s*=)(?!\d)\w+\b(?!\()/,lookbehind:!0},{pattern:/(\[)(?!\d)\w+(?=\])/,lookbehind:!0}],function:{pattern:/(\|\s*)@?[a-z_]\w*|\b[a-z_]\w*(?=\()/i,lookbehind:!0},"attr-name":/\b[a-z_]\w*(?=\s*=)/i,boolean:/\b(?:false|no|off|on|true|yes)\b/,punctuation:/[\[\](){}.,:`]|->/,operator:[/[+\-*\/%]|==?=?|[!<>]=?|&&|\|\|?/,/\bis\s+(?:not\s+)?(?:div|even|odd)(?:\s+by)?\b/,/\b(?:and|eq|gt?e|gt|lt?e|lt|mod|neq?|not|or)\b/]},s.languages.smarty["embedded-php"].inside.smarty.inside=s.languages.smarty,s.languages.smarty.string[0].inside.interpolation.inside.expression.inside=s.languages.smarty;var c=/"(?:\\.|[^"\\\r\n])*"|'(?:\\.|[^'\\\r\n])*'/,d=RegExp(/\{\*[\s\S]*?\*\}/.source+"|"+/\{php\}[\s\S]*?\{\/php\}/.source+"|"+/\{(?:[^{}"']|<str>|\{(?:[^{}"']|<str>|\{(?:[^{}"']|<str>)*\})*\})*\}/.source.replace(/<str>/g,function(){return c.source}),"g");s.hooks.add("before-tokenize",function(h){var b="{literal}",y="{/literal}",E=!1;s.languages["markup-templating"].buildPlaceholders(h,"smarty",d,function(k){return k===y&&(E=!1),E?!1:(k===b&&(E=!0),!0)})}),s.hooks.add("after-tokenize",function(h){s.languages["markup-templating"].tokenizePlaceholders(h,"smarty")})}(o)}return aGe}var sGe,Wqt;function zUr(){if(Wqt)return sGe;Wqt=1,sGe=n,n.displayName="sml",n.aliases=["smlnj"];function n(r){(function(o){var s=/\b(?:abstype|and|andalso|as|case|datatype|do|else|end|eqtype|exception|fn|fun|functor|handle|if|in|include|infix|infixr|let|local|nonfix|of|op|open|orelse|raise|rec|sharing|sig|signature|struct|structure|then|type|val|where|while|with|withtype)\b/i;o.languages.sml={comment:/\(\*(?:[^*(]|\*(?!\))|\((?!\*)|\(\*(?:[^*(]|\*(?!\))|\((?!\*))*\*\))*\*\)/,string:{pattern:/#?"(?:[^"\\]|\\.)*"/,greedy:!0},"class-name":[{pattern:RegExp(/((?:^|[^:]):\s*)<TERMINAL>(?:\s*(?:(?:\*|->)\s*<TERMINAL>|,\s*<TERMINAL>(?:(?=<NOT-LAST>)|(?!<NOT-LAST>)\s+<LONG-ID>)))*/.source.replace(/<NOT-LAST>/g,function(){return/\s*(?:[*,]|->)/.source}).replace(/<TERMINAL>/g,function(){return/(?:'[\w']*|<LONG-ID>|\((?:[^()]|\([^()]*\))*\)|\{(?:[^{}]|\{[^{}]*\})*\})(?:\s+<LONG-ID>)*/.source}).replace(/<LONG-ID>/g,function(){return/(?!<KEYWORD>)[a-z\d_][\w'.]*/.source}).replace(/<KEYWORD>/g,function(){return s.source}),"i"),lookbehind:!0,greedy:!0,inside:null},{pattern:/((?:^|[^\w'])(?:datatype|exception|functor|signature|structure|type)\s+)[a-z_][\w'.]*/i,lookbehind:!0}],function:{pattern:/((?:^|[^\w'])fun\s+)[a-z_][\w'.]*/i,lookbehind:!0},keyword:s,variable:{pattern:/(^|[^\w'])'[\w']*/,lookbehind:!0},number:/~?\b(?:\d+(?:\.\d+)?(?:e~?\d+)?|0x[\da-f]+)\b/i,word:{pattern:/\b0w(?:\d+|x[\da-f]+)\b/i,alias:"constant"},boolean:/\b(?:false|true)\b/i,operator:/\.\.\.|:[>=:]|=>?|->|[<>]=?|[!+\-*/^#|@~]/,punctuation:/[(){}\[\].:,;]/},o.languages.sml["class-name"][0].inside=o.languages.sml,o.languages.smlnj=o.languages.sml})(r)}return sGe}var cGe,Kqt;function HUr(){if(Kqt)return cGe;Kqt=1,cGe=n,n.displayName="solidity",n.aliases=["sol"];function n(r){r.languages.solidity=r.languages.extend("clike",{"class-name":{pattern:/(\b(?:contract|enum|interface|library|new|struct|using)\s+)(?!\d)[\w$]+/,lookbehind:!0},keyword:/\b(?:_|anonymous|as|assembly|assert|break|calldata|case|constant|constructor|continue|contract|default|delete|do|else|emit|enum|event|external|for|from|function|if|import|indexed|inherited|interface|internal|is|let|library|mapping|memory|modifier|new|payable|pragma|private|public|pure|require|returns?|revert|selfdestruct|solidity|storage|struct|suicide|switch|this|throw|using|var|view|while)\b/,operator:/=>|->|:=|=:|\*\*|\+\+|--|\|\||&&|<<=?|>>=?|[-+*/%^&|<>!=]=?|[~?]/}),r.languages.insertBefore("solidity","keyword",{builtin:/\b(?:address|bool|byte|u?int(?:8|16|24|32|40|48|56|64|72|80|88|96|104|112|120|128|136|144|152|160|168|176|184|192|200|208|216|224|232|240|248|256)?|string|bytes(?:[1-9]|[12]\d|3[0-2])?)\b/}),r.languages.insertBefore("solidity","number",{version:{pattern:/([<>]=?|\^)\d+\.\d+\.\d+\b/,lookbehind:!0,alias:"number"}}),r.languages.sol=r.languages.solidity}return cGe}var uGe,Yqt;function GUr(){if(Yqt)return uGe;Yqt=1,uGe=n,n.displayName="solutionFile",n.aliases=[];function n(r){(function(o){var s={pattern:/\{[\da-f]{8}-[\da-f]{4}-[\da-f]{4}-[\da-f]{4}-[\da-f]{12}\}/i,alias:"constant",inside:{punctuation:/[{}]/}};o.languages["solution-file"]={comment:{pattern:/#.*/,greedy:!0},string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,greedy:!0,inside:{guid:s}},object:{pattern:/^([ \t]*)(?:([A-Z]\w*)\b(?=.*(?:\r\n?|\n)(?:\1[ \t].*(?:\r\n?|\n))*\1End\2(?=[ \t]*$))|End[A-Z]\w*(?=[ \t]*$))/m,lookbehind:!0,greedy:!0,alias:"keyword"},property:{pattern:/^([ \t]*)(?!\s)[^\r\n"#=()]*[^\s"#=()](?=\s*=)/m,lookbehind:!0,inside:{guid:s}},guid:s,number:/\b\d+(?:\.\d+)*\b/,boolean:/\b(?:FALSE|TRUE)\b/,operator:/=/,punctuation:/[(),]/},o.languages.sln=o.languages["solution-file"]})(r)}return uGe}var lGe,Xqt;function qUr(){if(Xqt)return lGe;Xqt=1;var n=oy();lGe=r,r.displayName="soy",r.aliases=[];function r(o){o.register(n),function(s){var c=/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,d=/\b\d+(?:\.\d+)?(?:[eE][+-]?\d+)?\b|\b0x[\dA-F]+\b/;s.languages.soy={comment:[/\/\*[\s\S]*?\*\//,{pattern:/(\s)\/\/.*/,lookbehind:!0,greedy:!0}],"command-arg":{pattern:/(\{+\/?\s*(?:alias|call|delcall|delpackage|deltemplate|namespace|template)\s+)\.?[\w.]+/,lookbehind:!0,alias:"string",inside:{punctuation:/\./}},parameter:{pattern:/(\{+\/?\s*@?param\??\s+)\.?[\w.]+/,lookbehind:!0,alias:"variable"},keyword:[{pattern:/(\{+\/?[^\S\r\n]*)(?:\\[nrt]|alias|call|case|css|default|delcall|delpackage|deltemplate|else(?:if)?|fallbackmsg|for(?:each)?|if(?:empty)?|lb|let|literal|msg|namespace|nil|@?param\??|rb|sp|switch|template|xid)/,lookbehind:!0},/\b(?:any|as|attributes|bool|css|float|html|in|int|js|list|map|null|number|string|uri)\b/],delimiter:{pattern:/^\{+\/?|\/?\}+$/,alias:"punctuation"},property:/\w+(?==)/,variable:{pattern:/\$[^\W\d]\w*(?:\??(?:\.\w+|\[[^\]]+\]))*/,inside:{string:{pattern:c,greedy:!0},number:d,punctuation:/[\[\].?]/}},string:{pattern:c,greedy:!0},function:[/\w+(?=\()/,{pattern:/(\|[^\S\r\n]*)\w+/,lookbehind:!0}],boolean:/\b(?:false|true)\b/,number:d,operator:/\?:?|<=?|>=?|==?|!=|[+*/%-]|\b(?:and|not|or)\b/,punctuation:/[{}()\[\]|.,:]/},s.hooks.add("before-tokenize",function(h){var b=/\{\{.+?\}\}|\{.+?\}|\s\/\/.*|\/\*[\s\S]*?\*\//g,y="{literal}",E="{/literal}",k=!1;s.languages["markup-templating"].buildPlaceholders(h,"soy",b,function(x){return x===E&&(k=!1),k?!1:(x===y&&(k=!0),!0)})}),s.hooks.add("after-tokenize",function(h){s.languages["markup-templating"].tokenizePlaceholders(h,"soy")})}(o)}return lGe}var dGe,Jqt;function ldn(){if(Jqt)return dGe;Jqt=1,dGe=n,n.displayName="turtle",n.aliases=[];function n(r){r.languages.turtle={comment:{pattern:/#.*/,greedy:!0},"multiline-string":{pattern:/"""(?:(?:""?)?(?:[^"\\]|\\.))*"""|'''(?:(?:''?)?(?:[^'\\]|\\.))*'''/,greedy:!0,alias:"string",inside:{comment:/#.*/}},string:{pattern:/"(?:[^\\"\r\n]|\\.)*"|'(?:[^\\'\r\n]|\\.)*'/,greedy:!0},url:{pattern:/<(?:[^\x00-\x20<>"{}|^`\\]|\\(?:u[\da-fA-F]{4}|U[\da-fA-F]{8}))*>/,greedy:!0,inside:{punctuation:/[<>]/}},function:{pattern:/(?:(?![-.\d\xB7])[-.\w\xB7\xC0-\uFFFD]+)?:(?:(?![-.])(?:[-.:\w\xC0-\uFFFD]|%[\da-f]{2}|\\.)+)?/i,inside:{"local-name":{pattern:/([^:]*:)[\s\S]+/,lookbehind:!0},prefix:{pattern:/[\s\S]+/,inside:{punctuation:/:/}}}},number:/[+-]?\b\d+(?:\.\d*)?(?:e[+-]?\d+)?/i,punctuation:/[{}.,;()[\]]|\^\^/,boolean:/\b(?:false|true)\b/,keyword:[/(?:\ba|@prefix|@base)\b|=/,/\b(?:base|graph|prefix)\b/i],tag:{pattern:/@[a-z]+(?:-[a-z\d]+)*/i,inside:{punctuation:/@/}}},r.languages.trig=r.languages.turtle}return dGe}var fGe,Zqt;function VUr(){if(Zqt)return fGe;Zqt=1;var n=ldn();fGe=r,r.displayName="sparql",r.aliases=["rq"];function r(o){o.register(n),o.languages.sparql=o.languages.extend("turtle",{boolean:/\b(?:false|true)\b/i,variable:{pattern:/[?$]\w+/,greedy:!0}}),o.languages.insertBefore("sparql","punctuation",{keyword:[/\b(?:A|ADD|ALL|AS|ASC|ASK|BNODE|BY|CLEAR|CONSTRUCT|COPY|CREATE|DATA|DEFAULT|DELETE|DESC|DESCRIBE|DISTINCT|DROP|EXISTS|FILTER|FROM|GROUP|HAVING|INSERT|INTO|LIMIT|LOAD|MINUS|MOVE|NAMED|NOT|NOW|OFFSET|OPTIONAL|ORDER|RAND|REDUCED|SELECT|SEPARATOR|SERVICE|SILENT|STRUUID|UNION|USING|UUID|VALUES|WHERE)\b/i,/\b(?:ABS|AVG|BIND|BOUND|CEIL|COALESCE|CONCAT|CONTAINS|COUNT|DATATYPE|DAY|ENCODE_FOR_URI|FLOOR|GROUP_CONCAT|HOURS|IF|IRI|isBLANK|isIRI|isLITERAL|isNUMERIC|isURI|LANG|LANGMATCHES|LCASE|MAX|MD5|MIN|MINUTES|MONTH|REGEX|REPLACE|ROUND|sameTerm|SAMPLE|SECONDS|SHA1|SHA256|SHA384|SHA512|STR|STRAFTER|STRBEFORE|STRDT|STRENDS|STRLANG|STRLEN|STRSTARTS|SUBSTR|SUM|TIMEZONE|TZ|UCASE|URI|YEAR)\b(?=\s*\()/i,/\b(?:BASE|GRAPH|PREFIX)\b/i]}),o.languages.rq=o.languages.sparql}return fGe}var hGe,Qqt;function WUr(){if(Qqt)return hGe;Qqt=1,hGe=n,n.displayName="splunkSpl",n.aliases=[];function n(r){r.languages["splunk-spl"]={comment:/`comment\("(?:\\.|[^\\"])*"\)`/,string:{pattern:/"(?:\\.|[^\\"])*"/,greedy:!0},keyword:/\b(?:abstract|accum|addcoltotals|addinfo|addtotals|analyzefields|anomalies|anomalousvalue|anomalydetection|append|appendcols|appendcsv|appendlookup|appendpipe|arules|associate|audit|autoregress|bin|bucket|bucketdir|chart|cluster|cofilter|collect|concurrency|contingency|convert|correlate|datamodel|dbinspect|dedup|delete|delta|diff|erex|eval|eventcount|eventstats|extract|fieldformat|fields|fieldsummary|filldown|fillnull|findtypes|folderize|foreach|format|from|gauge|gentimes|geom|geomfilter|geostats|head|highlight|history|iconify|input|inputcsv|inputlookup|iplocation|join|kmeans|kv|kvform|loadjob|localize|localop|lookup|makecontinuous|makemv|makeresults|map|mcollect|metadata|metasearch|meventcollect|mstats|multikv|multisearch|mvcombine|mvexpand|nomv|outlier|outputcsv|outputlookup|outputtext|overlap|pivot|predict|rangemap|rare|regex|relevancy|reltime|rename|replace|rest|return|reverse|rex|rtorder|run|savedsearch|script|scrub|search|searchtxn|selfjoin|sendemail|set|setfields|sichart|sirare|sistats|sitimechart|sitop|sort|spath|stats|strcat|streamstats|table|tags|tail|timechart|timewrap|top|transaction|transpose|trendline|tscollect|tstats|typeahead|typelearner|typer|union|uniq|untable|where|x11|xmlkv|xmlunescape|xpath|xyseries)\b/i,"operator-word":{pattern:/\b(?:and|as|by|not|or|xor)\b/i,alias:"operator"},function:/\b\w+(?=\s*\()/,property:/\b\w+(?=\s*=(?!=))/,date:{pattern:/\b\d{1,2}\/\d{1,2}\/\d{1,4}(?:(?::\d{1,2}){3})?\b/,alias:"number"},number:/\b\d+(?:\.\d+)?\b/,boolean:/\b(?:f|false|t|true)\b/i,operator:/[<>=]=?|[-+*/%|]/,punctuation:/[()[\],]/}}return hGe}var pGe,eVt;function KUr(){if(eVt)return pGe;eVt=1,pGe=n,n.displayName="sqf",n.aliases=[];function n(r){r.languages.sqf=r.languages.extend("clike",{string:{pattern:/"(?:(?:"")?[^"])*"(?!")|'(?:[^'])*'/,greedy:!0},keyword:/\b(?:breakOut|breakTo|call|case|catch|default|do|echo|else|execFSM|execVM|exitWith|for|forEach|forEachMember|forEachMemberAgent|forEachMemberTeam|from|goto|if|nil|preprocessFile|preprocessFileLineNumbers|private|scopeName|spawn|step|switch|then|throw|to|try|while|with)\b/i,boolean:/\b(?:false|true)\b/i,function:/\b(?:abs|accTime|acos|action|actionIDs|actionKeys|actionKeysImages|actionKeysNames|actionKeysNamesArray|actionName|actionParams|activateAddons|activatedAddons|activateKey|add3DENConnection|add3DENEventHandler|add3DENLayer|addAction|addBackpack|addBackpackCargo|addBackpackCargoGlobal|addBackpackGlobal|addCamShake|addCuratorAddons|addCuratorCameraArea|addCuratorEditableObjects|addCuratorEditingArea|addCuratorPoints|addEditorObject|addEventHandler|addForce|addForceGeneratorRTD|addGoggles|addGroupIcon|addHandgunItem|addHeadgear|addItem|addItemCargo|addItemCargoGlobal|addItemPool|addItemToBackpack|addItemToUniform|addItemToVest|addLiveStats|addMagazine|addMagazineAmmoCargo|addMagazineCargo|addMagazineCargoGlobal|addMagazineGlobal|addMagazinePool|addMagazines|addMagazineTurret|addMenu|addMenuItem|addMissionEventHandler|addMPEventHandler|addMusicEventHandler|addOwnedMine|addPlayerScores|addPrimaryWeaponItem|addPublicVariableEventHandler|addRating|addResources|addScore|addScoreSide|addSecondaryWeaponItem|addSwitchableUnit|addTeamMember|addToRemainsCollector|addTorque|addUniform|addVehicle|addVest|addWaypoint|addWeapon|addWeaponCargo|addWeaponCargoGlobal|addWeaponGlobal|addWeaponItem|addWeaponPool|addWeaponTurret|admin|agent|agents|AGLToASL|aimedAtTarget|aimPos|airDensityCurveRTD|airDensityRTD|airplaneThrottle|airportSide|AISFinishHeal|alive|all3DENEntities|allAirports|allControls|allCurators|allCutLayers|allDead|allDeadMen|allDisplays|allGroups|allMapMarkers|allMines|allMissionObjects|allow3DMode|allowCrewInImmobile|allowCuratorLogicIgnoreAreas|allowDamage|allowDammage|allowFileOperations|allowFleeing|allowGetIn|allowSprint|allPlayers|allSimpleObjects|allSites|allTurrets|allUnits|allUnitsUAV|allVariables|ammo|ammoOnPylon|animate|animateBay|animateDoor|animatePylon|animateSource|animationNames|animationPhase|animationSourcePhase|animationState|append|apply|armoryPoints|arrayIntersect|asin|ASLToAGL|ASLToATL|assert|assignAsCargo|assignAsCargoIndex|assignAsCommander|assignAsDriver|assignAsGunner|assignAsTurret|assignCurator|assignedCargo|assignedCommander|assignedDriver|assignedGunner|assignedItems|assignedTarget|assignedTeam|assignedVehicle|assignedVehicleRole|assignItem|assignTeam|assignToAirport|atan|atan2|atg|ATLToASL|attachedObject|attachedObjects|attachedTo|attachObject|attachTo|attackEnabled|backpack|backpackCargo|backpackContainer|backpackItems|backpackMagazines|backpackSpaceFor|behaviour|benchmark|binocular|blufor|boundingBox|boundingBoxReal|boundingCenter|briefingName|buildingExit|buildingPos|buldozer_EnableRoadDiag|buldozer_IsEnabledRoadDiag|buldozer_LoadNewRoads|buldozer_reloadOperMap|buttonAction|buttonSetAction|cadetMode|callExtension|camCommand|camCommit|camCommitPrepared|camCommitted|camConstuctionSetParams|camCreate|camDestroy|cameraEffect|cameraEffectEnableHUD|cameraInterest|cameraOn|cameraView|campaignConfigFile|camPreload|camPreloaded|camPrepareBank|camPrepareDir|camPrepareDive|camPrepareFocus|camPrepareFov|camPrepareFovRange|camPreparePos|camPrepareRelPos|camPrepareTarget|camSetBank|camSetDir|camSetDive|camSetFocus|camSetFov|camSetFovRange|camSetPos|camSetRelPos|camSetTarget|camTarget|camUseNVG|canAdd|canAddItemToBackpack|canAddItemToUniform|canAddItemToVest|cancelSimpleTaskDestination|canFire|canMove|canSlingLoad|canStand|canSuspend|canTriggerDynamicSimulation|canUnloadInCombat|canVehicleCargo|captive|captiveNum|cbChecked|cbSetChecked|ceil|channelEnabled|cheatsEnabled|checkAIFeature|checkVisibility|civilian|className|clear3DENAttribute|clear3DENInventory|clearAllItemsFromBackpack|clearBackpackCargo|clearBackpackCargoGlobal|clearForcesRTD|clearGroupIcons|clearItemCargo|clearItemCargoGlobal|clearItemPool|clearMagazineCargo|clearMagazineCargoGlobal|clearMagazinePool|clearOverlay|clearRadio|clearVehicleInit|clearWeaponCargo|clearWeaponCargoGlobal|clearWeaponPool|clientOwner|closeDialog|closeDisplay|closeOverlay|collapseObjectTree|collect3DENHistory|collectiveRTD|combatMode|commandArtilleryFire|commandChat|commander|commandFire|commandFollow|commandFSM|commandGetOut|commandingMenu|commandMove|commandRadio|commandStop|commandSuppressiveFire|commandTarget|commandWatch|comment|commitOverlay|compile|compileFinal|completedFSM|composeText|configClasses|configFile|configHierarchy|configName|configNull|configProperties|configSourceAddonList|configSourceMod|configSourceModList|confirmSensorTarget|connectTerminalToUAV|controlNull|controlsGroupCtrl|copyFromClipboard|copyToClipboard|copyWaypoints|cos|count|countEnemy|countFriendly|countSide|countType|countUnknown|create3DENComposition|create3DENEntity|createAgent|createCenter|createDialog|createDiaryLink|createDiaryRecord|createDiarySubject|createDisplay|createGearDialog|createGroup|createGuardedPoint|createLocation|createMarker|createMarkerLocal|createMenu|createMine|createMissionDisplay|createMPCampaignDisplay|createSimpleObject|createSimpleTask|createSite|createSoundSource|createTask|createTeam|createTrigger|createUnit|createVehicle|createVehicleCrew|createVehicleLocal|crew|ctAddHeader|ctAddRow|ctClear|ctCurSel|ctData|ctFindHeaderRows|ctFindRowHeader|ctHeaderControls|ctHeaderCount|ctRemoveHeaders|ctRemoveRows|ctrlActivate|ctrlAddEventHandler|ctrlAngle|ctrlAutoScrollDelay|ctrlAutoScrollRewind|ctrlAutoScrollSpeed|ctrlChecked|ctrlClassName|ctrlCommit|ctrlCommitted|ctrlCreate|ctrlDelete|ctrlEnable|ctrlEnabled|ctrlFade|ctrlHTMLLoaded|ctrlIDC|ctrlIDD|ctrlMapAnimAdd|ctrlMapAnimClear|ctrlMapAnimCommit|ctrlMapAnimDone|ctrlMapCursor|ctrlMapMouseOver|ctrlMapScale|ctrlMapScreenToWorld|ctrlMapWorldToScreen|ctrlModel|ctrlModelDirAndUp|ctrlModelScale|ctrlParent|ctrlParentControlsGroup|ctrlPosition|ctrlRemoveAllEventHandlers|ctrlRemoveEventHandler|ctrlScale|ctrlSetActiveColor|ctrlSetAngle|ctrlSetAutoScrollDelay|ctrlSetAutoScrollRewind|ctrlSetAutoScrollSpeed|ctrlSetBackgroundColor|ctrlSetChecked|ctrlSetDisabledColor|ctrlSetEventHandler|ctrlSetFade|ctrlSetFocus|ctrlSetFont|ctrlSetFontH1|ctrlSetFontH1B|ctrlSetFontH2|ctrlSetFontH2B|ctrlSetFontH3|ctrlSetFontH3B|ctrlSetFontH4|ctrlSetFontH4B|ctrlSetFontH5|ctrlSetFontH5B|ctrlSetFontH6|ctrlSetFontH6B|ctrlSetFontHeight|ctrlSetFontHeightH1|ctrlSetFontHeightH2|ctrlSetFontHeightH3|ctrlSetFontHeightH4|ctrlSetFontHeightH5|ctrlSetFontHeightH6|ctrlSetFontHeightSecondary|ctrlSetFontP|ctrlSetFontPB|ctrlSetFontSecondary|ctrlSetForegroundColor|ctrlSetModel|ctrlSetModelDirAndUp|ctrlSetModelScale|ctrlSetPixelPrecision|ctrlSetPosition|ctrlSetScale|ctrlSetStructuredText|ctrlSetText|ctrlSetTextColor|ctrlSetTextColorSecondary|ctrlSetTextSecondary|ctrlSetTooltip|ctrlSetTooltipColorBox|ctrlSetTooltipColorShade|ctrlSetTooltipColorText|ctrlShow|ctrlShown|ctrlText|ctrlTextHeight|ctrlTextSecondary|ctrlTextWidth|ctrlType|ctrlVisible|ctRowControls|ctRowCount|ctSetCurSel|ctSetData|ctSetHeaderTemplate|ctSetRowTemplate|ctSetValue|ctValue|curatorAddons|curatorCamera|curatorCameraArea|curatorCameraAreaCeiling|curatorCoef|curatorEditableObjects|curatorEditingArea|curatorEditingAreaType|curatorMouseOver|curatorPoints|curatorRegisteredObjects|curatorSelected|curatorWaypointCost|current3DENOperation|currentChannel|currentCommand|currentMagazine|currentMagazineDetail|currentMagazineDetailTurret|currentMagazineTurret|currentMuzzle|currentNamespace|currentTask|currentTasks|currentThrowable|currentVisionMode|currentWaypoint|currentWeapon|currentWeaponMode|currentWeaponTurret|currentZeroing|cursorObject|cursorTarget|customChat|customRadio|cutFadeOut|cutObj|cutRsc|cutText|damage|date|dateToNumber|daytime|deActivateKey|debriefingText|debugFSM|debugLog|deg|delete3DENEntities|deleteAt|deleteCenter|deleteCollection|deleteEditorObject|deleteGroup|deleteGroupWhenEmpty|deleteIdentity|deleteLocation|deleteMarker|deleteMarkerLocal|deleteRange|deleteResources|deleteSite|deleteStatus|deleteTeam|deleteVehicle|deleteVehicleCrew|deleteWaypoint|detach|detectedMines|diag_activeMissionFSMs|diag_activeScripts|diag_activeSQFScripts|diag_activeSQSScripts|diag_captureFrame|diag_captureFrameToFile|diag_captureSlowFrame|diag_codePerformance|diag_drawMode|diag_dynamicSimulationEnd|diag_enable|diag_enabled|diag_fps|diag_fpsMin|diag_frameNo|diag_lightNewLoad|diag_list|diag_log|diag_logSlowFrame|diag_mergeConfigFile|diag_recordTurretLimits|diag_setLightNew|diag_tickTime|diag_toggle|dialog|diarySubjectExists|didJIP|didJIPOwner|difficulty|difficultyEnabled|difficultyEnabledRTD|difficultyOption|direction|directSay|disableAI|disableCollisionWith|disableConversation|disableDebriefingStats|disableMapIndicators|disableNVGEquipment|disableRemoteSensors|disableSerialization|disableTIEquipment|disableUAVConnectability|disableUserInput|displayAddEventHandler|displayCtrl|displayNull|displayParent|displayRemoveAllEventHandlers|displayRemoveEventHandler|displaySetEventHandler|dissolveTeam|distance|distance2D|distanceSqr|distributionRegion|do3DENAction|doArtilleryFire|doFire|doFollow|doFSM|doGetOut|doMove|doorPhase|doStop|doSuppressiveFire|doTarget|doWatch|drawArrow|drawEllipse|drawIcon|drawIcon3D|drawLine|drawLine3D|drawLink|drawLocation|drawPolygon|drawRectangle|drawTriangle|driver|drop|dynamicSimulationDistance|dynamicSimulationDistanceCoef|dynamicSimulationEnabled|dynamicSimulationSystemEnabled|east|edit3DENMissionAttributes|editObject|editorSetEventHandler|effectiveCommander|emptyPositions|enableAI|enableAIFeature|enableAimPrecision|enableAttack|enableAudioFeature|enableAutoStartUpRTD|enableAutoTrimRTD|enableCamShake|enableCaustics|enableChannel|enableCollisionWith|enableCopilot|enableDebriefingStats|enableDiagLegend|enableDynamicSimulation|enableDynamicSimulationSystem|enableEndDialog|enableEngineArtillery|enableEnvironment|enableFatigue|enableGunLights|enableInfoPanelComponent|enableIRLasers|enableMimics|enablePersonTurret|enableRadio|enableReload|enableRopeAttach|enableSatNormalOnDetail|enableSaving|enableSentences|enableSimulation|enableSimulationGlobal|enableStamina|enableStressDamage|enableTeamSwitch|enableTraffic|enableUAVConnectability|enableUAVWaypoints|enableVehicleCargo|enableVehicleSensor|enableWeaponDisassembly|endl|endLoadingScreen|endMission|engineOn|enginesIsOnRTD|enginesPowerRTD|enginesRpmRTD|enginesTorqueRTD|entities|environmentEnabled|estimatedEndServerTime|estimatedTimeLeft|evalObjectArgument|everyBackpack|everyContainer|exec|execEditorScript|exp|expectedDestination|exportJIPMessages|eyeDirection|eyePos|face|faction|fadeMusic|fadeRadio|fadeSound|fadeSpeech|failMission|fillWeaponsFromPool|find|findCover|findDisplay|findEditorObject|findEmptyPosition|findEmptyPositionReady|findIf|findNearestEnemy|finishMissionInit|finite|fire|fireAtTarget|firstBackpack|flag|flagAnimationPhase|flagOwner|flagSide|flagTexture|fleeing|floor|flyInHeight|flyInHeightASL|fog|fogForecast|fogParams|forceAddUniform|forceAtPositionRTD|forcedMap|forceEnd|forceFlagTexture|forceFollowRoad|forceGeneratorRTD|forceMap|forceRespawn|forceSpeed|forceWalk|forceWeaponFire|forceWeatherChange|forgetTarget|format|formation|formationDirection|formationLeader|formationMembers|formationPosition|formationTask|formatText|formLeader|freeLook|fromEditor|fuel|fullCrew|gearIDCAmmoCount|gearSlotAmmoCount|gearSlotData|get3DENActionState|get3DENAttribute|get3DENCamera|get3DENConnections|get3DENEntity|get3DENEntityID|get3DENGrid|get3DENIconsVisible|get3DENLayerEntities|get3DENLinesVisible|get3DENMissionAttribute|get3DENMouseOver|get3DENSelected|getAimingCoef|getAllEnvSoundControllers|getAllHitPointsDamage|getAllOwnedMines|getAllSoundControllers|getAmmoCargo|getAnimAimPrecision|getAnimSpeedCoef|getArray|getArtilleryAmmo|getArtilleryComputerSettings|getArtilleryETA|getAssignedCuratorLogic|getAssignedCuratorUnit|getBackpackCargo|getBleedingRemaining|getBurningValue|getCameraViewDirection|getCargoIndex|getCenterOfMass|getClientState|getClientStateNumber|getCompatiblePylonMagazines|getConnectedUAV|getContainerMaxLoad|getCursorObjectParams|getCustomAimCoef|getDammage|getDescription|getDir|getDirVisual|getDLCAssetsUsage|getDLCAssetsUsageByName|getDLCs|getDLCUsageTime|getEditorCamera|getEditorMode|getEditorObjectScope|getElevationOffset|getEngineTargetRpmRTD|getEnvSoundController|getFatigue|getFieldManualStartPage|getForcedFlagTexture|getFriend|getFSMVariable|getFuelCargo|getGroupIcon|getGroupIconParams|getGroupIcons|getHideFrom|getHit|getHitIndex|getHitPointDamage|getItemCargo|getMagazineCargo|getMarkerColor|getMarkerPos|getMarkerSize|getMarkerType|getMass|getMissionConfig|getMissionConfigValue|getMissionDLCs|getMissionLayerEntities|getMissionLayers|getModelInfo|getMousePosition|getMusicPlayedTime|getNumber|getObjectArgument|getObjectChildren|getObjectDLC|getObjectMaterials|getObjectProxy|getObjectTextures|getObjectType|getObjectViewDistance|getOxygenRemaining|getPersonUsedDLCs|getPilotCameraDirection|getPilotCameraPosition|getPilotCameraRotation|getPilotCameraTarget|getPlateNumber|getPlayerChannel|getPlayerScores|getPlayerUID|getPlayerUIDOld|getPos|getPosASL|getPosASLVisual|getPosASLW|getPosATL|getPosATLVisual|getPosVisual|getPosWorld|getPylonMagazines|getRelDir|getRelPos|getRemoteSensorsDisabled|getRepairCargo|getResolution|getRotorBrakeRTD|getShadowDistance|getShotParents|getSlingLoad|getSoundController|getSoundControllerResult|getSpeed|getStamina|getStatValue|getSuppression|getTerrainGrid|getTerrainHeightASL|getText|getTotalDLCUsageTime|getTrimOffsetRTD|getUnitLoadout|getUnitTrait|getUserMFDText|getUserMFDValue|getVariable|getVehicleCargo|getWeaponCargo|getWeaponSway|getWingsOrientationRTD|getWingsPositionRTD|getWPPos|glanceAt|globalChat|globalRadio|goggles|group|groupChat|groupFromNetId|groupIconSelectable|groupIconsVisible|groupId|groupOwner|groupRadio|groupSelectedUnits|groupSelectUnit|grpNull|gunner|gusts|halt|handgunItems|handgunMagazine|handgunWeapon|handsHit|hasInterface|hasPilotCamera|hasWeapon|hcAllGroups|hcGroupParams|hcLeader|hcRemoveAllGroups|hcRemoveGroup|hcSelected|hcSelectGroup|hcSetGroup|hcShowBar|hcShownBar|headgear|hideBody|hideObject|hideObjectGlobal|hideSelection|hint|hintC|hintCadet|hintSilent|hmd|hostMission|htmlLoad|HUDMovementLevels|humidity|image|importAllGroups|importance|in|inArea|inAreaArray|incapacitatedState|independent|inflame|inflamed|infoPanel|infoPanelComponentEnabled|infoPanelComponents|infoPanels|inGameUISetEventHandler|inheritsFrom|initAmbientLife|inPolygon|inputAction|inRangeOfArtillery|insertEditorObject|intersect|is3DEN|is3DENMultiplayer|isAbleToBreathe|isAgent|isAimPrecisionEnabled|isArray|isAutoHoverOn|isAutonomous|isAutoStartUpEnabledRTD|isAutotest|isAutoTrimOnRTD|isBleeding|isBurning|isClass|isCollisionLightOn|isCopilotEnabled|isDamageAllowed|isDedicated|isDLCAvailable|isEngineOn|isEqualTo|isEqualType|isEqualTypeAll|isEqualTypeAny|isEqualTypeArray|isEqualTypeParams|isFilePatchingEnabled|isFlashlightOn|isFlatEmpty|isForcedWalk|isFormationLeader|isGroupDeletedWhenEmpty|isHidden|isInRemainsCollector|isInstructorFigureEnabled|isIRLaserOn|isKeyActive|isKindOf|isLaserOn|isLightOn|isLocalized|isManualFire|isMarkedForCollection|isMultiplayer|isMultiplayerSolo|isNil|isNull|isNumber|isObjectHidden|isObjectRTD|isOnRoad|isPipEnabled|isPlayer|isRealTime|isRemoteExecuted|isRemoteExecutedJIP|isServer|isShowing3DIcons|isSimpleObject|isSprintAllowed|isStaminaEnabled|isSteamMission|isStreamFriendlyUIEnabled|isStressDamageEnabled|isText|isTouchingGround|isTurnedOut|isTutHintsEnabled|isUAVConnectable|isUAVConnected|isUIContext|isUniformAllowed|isVehicleCargo|isVehicleRadarOn|isVehicleSensorEnabled|isWalking|isWeaponDeployed|isWeaponRested|itemCargo|items|itemsWithMagazines|join|joinAs|joinAsSilent|joinSilent|joinString|kbAddDatabase|kbAddDatabaseTargets|kbAddTopic|kbHasTopic|kbReact|kbRemoveTopic|kbTell|kbWasSaid|keyImage|keyName|knowsAbout|land|landAt|landResult|language|laserTarget|lbAdd|lbClear|lbColor|lbColorRight|lbCurSel|lbData|lbDelete|lbIsSelected|lbPicture|lbPictureRight|lbSelection|lbSetColor|lbSetColorRight|lbSetCurSel|lbSetData|lbSetPicture|lbSetPictureColor|lbSetPictureColorDisabled|lbSetPictureColorSelected|lbSetPictureRight|lbSetPictureRightColor|lbSetPictureRightColorDisabled|lbSetPictureRightColorSelected|lbSetSelectColor|lbSetSelectColorRight|lbSetSelected|lbSetText|lbSetTextRight|lbSetTooltip|lbSetValue|lbSize|lbSort|lbSortByValue|lbText|lbTextRight|lbValue|leader|leaderboardDeInit|leaderboardGetRows|leaderboardInit|leaderboardRequestRowsFriends|leaderboardRequestRowsGlobal|leaderboardRequestRowsGlobalAroundUser|leaderboardsRequestUploadScore|leaderboardsRequestUploadScoreKeepBest|leaderboardState|leaveVehicle|libraryCredits|libraryDisclaimers|lifeState|lightAttachObject|lightDetachObject|lightIsOn|lightnings|limitSpeed|linearConversion|lineBreak|lineIntersects|lineIntersectsObjs|lineIntersectsSurfaces|lineIntersectsWith|linkItem|list|listObjects|listRemoteTargets|listVehicleSensors|ln|lnbAddArray|lnbAddColumn|lnbAddRow|lnbClear|lnbColor|lnbColorRight|lnbCurSelRow|lnbData|lnbDeleteColumn|lnbDeleteRow|lnbGetColumnsPosition|lnbPicture|lnbPictureRight|lnbSetColor|lnbSetColorRight|lnbSetColumnsPos|lnbSetCurSelRow|lnbSetData|lnbSetPicture|lnbSetPictureColor|lnbSetPictureColorRight|lnbSetPictureColorSelected|lnbSetPictureColorSelectedRight|lnbSetPictureRight|lnbSetText|lnbSetTextRight|lnbSetValue|lnbSize|lnbSort|lnbSortByValue|lnbText|lnbTextRight|lnbValue|load|loadAbs|loadBackpack|loadFile|loadGame|loadIdentity|loadMagazine|loadOverlay|loadStatus|loadUniform|loadVest|local|localize|locationNull|locationPosition|lock|lockCameraTo|lockCargo|lockDriver|locked|lockedCargo|lockedDriver|lockedTurret|lockIdentity|lockTurret|lockWP|log|logEntities|logNetwork|logNetworkTerminate|lookAt|lookAtPos|magazineCargo|magazines|magazinesAllTurrets|magazinesAmmo|magazinesAmmoCargo|magazinesAmmoFull|magazinesDetail|magazinesDetailBackpack|magazinesDetailUniform|magazinesDetailVest|magazinesTurret|magazineTurretAmmo|mapAnimAdd|mapAnimClear|mapAnimCommit|mapAnimDone|mapCenterOnCamera|mapGridPosition|markAsFinishedOnSteam|markerAlpha|markerBrush|markerColor|markerDir|markerPos|markerShape|markerSize|markerText|markerType|max|members|menuAction|menuAdd|menuChecked|menuClear|menuCollapse|menuData|menuDelete|menuEnable|menuEnabled|menuExpand|menuHover|menuPicture|menuSetAction|menuSetCheck|menuSetData|menuSetPicture|menuSetValue|menuShortcut|menuShortcutText|menuSize|menuSort|menuText|menuURL|menuValue|min|mineActive|mineDetectedBy|missionConfigFile|missionDifficulty|missionName|missionNamespace|missionStart|missionVersion|modelToWorld|modelToWorldVisual|modelToWorldVisualWorld|modelToWorldWorld|modParams|moonIntensity|moonPhase|morale|move|move3DENCamera|moveInAny|moveInCargo|moveInCommander|moveInDriver|moveInGunner|moveInTurret|moveObjectToEnd|moveOut|moveTime|moveTo|moveToCompleted|moveToFailed|musicVolume|name|nameSound|nearEntities|nearestBuilding|nearestLocation|nearestLocations|nearestLocationWithDubbing|nearestObject|nearestObjects|nearestTerrainObjects|nearObjects|nearObjectsReady|nearRoads|nearSupplies|nearTargets|needReload|netId|netObjNull|newOverlay|nextMenuItemIndex|nextWeatherChange|nMenuItems|numberOfEnginesRTD|numberToDate|objectCurators|objectFromNetId|objectParent|objNull|objStatus|onBriefingGear|onBriefingGroup|onBriefingNotes|onBriefingPlan|onBriefingTeamSwitch|onCommandModeChanged|onDoubleClick|onEachFrame|onGroupIconClick|onGroupIconOverEnter|onGroupIconOverLeave|onHCGroupSelectionChanged|onMapSingleClick|onPlayerConnected|onPlayerDisconnected|onPreloadFinished|onPreloadStarted|onShowNewObject|onTeamSwitch|openCuratorInterface|openDLCPage|openDSInterface|openMap|openSteamApp|openYoutubeVideo|opfor|orderGetIn|overcast|overcastForecast|owner|param|params|parseNumber|parseSimpleArray|parseText|parsingNamespace|particlesQuality|pi|pickWeaponPool|pitch|pixelGrid|pixelGridBase|pixelGridNoUIScale|pixelH|pixelW|playableSlotsNumber|playableUnits|playAction|playActionNow|player|playerRespawnTime|playerSide|playersNumber|playGesture|playMission|playMove|playMoveNow|playMusic|playScriptedMission|playSound|playSound3D|position|positionCameraToWorld|posScreenToWorld|posWorldToScreen|ppEffectAdjust|ppEffectCommit|ppEffectCommitted|ppEffectCreate|ppEffectDestroy|ppEffectEnable|ppEffectEnabled|ppEffectForceInNVG|precision|preloadCamera|preloadObject|preloadSound|preloadTitleObj|preloadTitleRsc|primaryWeapon|primaryWeaponItems|primaryWeaponMagazine|priority|processDiaryLink|processInitCommands|productVersion|profileName|profileNamespace|profileNameSteam|progressLoadingScreen|progressPosition|progressSetPosition|publicVariable|publicVariableClient|publicVariableServer|pushBack|pushBackUnique|putWeaponPool|queryItemsPool|queryMagazinePool|queryWeaponPool|rad|radioChannelAdd|radioChannelCreate|radioChannelRemove|radioChannelSetCallSign|radioChannelSetLabel|radioVolume|rain|rainbow|random|rank|rankId|rating|rectangular|registeredTasks|registerTask|reload|reloadEnabled|remoteControl|remoteExec|remoteExecCall|remoteExecutedOwner|remove3DENConnection|remove3DENEventHandler|remove3DENLayer|removeAction|removeAll3DENEventHandlers|removeAllActions|removeAllAssignedItems|removeAllContainers|removeAllCuratorAddons|removeAllCuratorCameraAreas|removeAllCuratorEditingAreas|removeAllEventHandlers|removeAllHandgunItems|removeAllItems|removeAllItemsWithMagazines|removeAllMissionEventHandlers|removeAllMPEventHandlers|removeAllMusicEventHandlers|removeAllOwnedMines|removeAllPrimaryWeaponItems|removeAllWeapons|removeBackpack|removeBackpackGlobal|removeCuratorAddons|removeCuratorCameraArea|removeCuratorEditableObjects|removeCuratorEditingArea|removeDrawIcon|removeDrawLinks|removeEventHandler|removeFromRemainsCollector|removeGoggles|removeGroupIcon|removeHandgunItem|removeHeadgear|removeItem|removeItemFromBackpack|removeItemFromUniform|removeItemFromVest|removeItems|removeMagazine|removeMagazineGlobal|removeMagazines|removeMagazinesTurret|removeMagazineTurret|removeMenuItem|removeMissionEventHandler|removeMPEventHandler|removeMusicEventHandler|removeOwnedMine|removePrimaryWeaponItem|removeSecondaryWeaponItem|removeSimpleTask|removeSwitchableUnit|removeTeamMember|removeUniform|removeVest|removeWeapon|removeWeaponAttachmentCargo|removeWeaponCargo|removeWeaponGlobal|removeWeaponTurret|reportRemoteTarget|requiredVersion|resetCamShake|resetSubgroupDirection|resistance|resize|resources|respawnVehicle|restartEditorCamera|reveal|revealMine|reverse|reversedMouseY|roadAt|roadsConnectedTo|roleDescription|ropeAttachedObjects|ropeAttachedTo|ropeAttachEnabled|ropeAttachTo|ropeCreate|ropeCut|ropeDestroy|ropeDetach|ropeEndPosition|ropeLength|ropes|ropeUnwind|ropeUnwound|rotorsForcesRTD|rotorsRpmRTD|round|runInitScript|safeZoneH|safeZoneW|safeZoneWAbs|safeZoneX|safeZoneXAbs|safeZoneY|save3DENInventory|saveGame|saveIdentity|saveJoysticks|saveOverlay|saveProfileNamespace|saveStatus|saveVar|savingEnabled|say|say2D|say3D|score|scoreSide|screenshot|screenToWorld|scriptDone|scriptName|scriptNull|scudState|secondaryWeapon|secondaryWeaponItems|secondaryWeaponMagazine|select|selectBestPlaces|selectDiarySubject|selectedEditorObjects|selectEditorObject|selectionNames|selectionPosition|selectLeader|selectMax|selectMin|selectNoPlayer|selectPlayer|selectRandom|selectRandomWeighted|selectWeapon|selectWeaponTurret|sendAUMessage|sendSimpleCommand|sendTask|sendTaskResult|sendUDPMessage|serverCommand|serverCommandAvailable|serverCommandExecutable|serverName|serverTime|set|set3DENAttribute|set3DENAttributes|set3DENGrid|set3DENIconsVisible|set3DENLayer|set3DENLinesVisible|set3DENLogicType|set3DENMissionAttribute|set3DENMissionAttributes|set3DENModelsVisible|set3DENObjectType|set3DENSelected|setAccTime|setActualCollectiveRTD|setAirplaneThrottle|setAirportSide|setAmmo|setAmmoCargo|setAmmoOnPylon|setAnimSpeedCoef|setAperture|setApertureNew|setArmoryPoints|setAttributes|setAutonomous|setBehaviour|setBleedingRemaining|setBrakesRTD|setCameraInterest|setCamShakeDefParams|setCamShakeParams|setCamUseTI|setCaptive|setCenterOfMass|setCollisionLight|setCombatMode|setCompassOscillation|setConvoySeparation|setCuratorCameraAreaCeiling|setCuratorCoef|setCuratorEditingAreaType|setCuratorWaypointCost|setCurrentChannel|setCurrentTask|setCurrentWaypoint|setCustomAimCoef|setCustomWeightRTD|setDamage|setDammage|setDate|setDebriefingText|setDefaultCamera|setDestination|setDetailMapBlendPars|setDir|setDirection|setDrawIcon|setDriveOnPath|setDropInterval|setDynamicSimulationDistance|setDynamicSimulationDistanceCoef|setEditorMode|setEditorObjectScope|setEffectCondition|setEngineRpmRTD|setFace|setFaceAnimation|setFatigue|setFeatureType|setFlagAnimationPhase|setFlagOwner|setFlagSide|setFlagTexture|setFog|setForceGeneratorRTD|setFormation|setFormationTask|setFormDir|setFriend|setFromEditor|setFSMVariable|setFuel|setFuelCargo|setGroupIcon|setGroupIconParams|setGroupIconsSelectable|setGroupIconsVisible|setGroupId|setGroupIdGlobal|setGroupOwner|setGusts|setHideBehind|setHit|setHitIndex|setHitPointDamage|setHorizonParallaxCoef|setHUDMovementLevels|setIdentity|setImportance|setInfoPanel|setLeader|setLightAmbient|setLightAttenuation|setLightBrightness|setLightColor|setLightDayLight|setLightFlareMaxDistance|setLightFlareSize|setLightIntensity|setLightnings|setLightUseFlare|setLocalWindParams|setMagazineTurretAmmo|setMarkerAlpha|setMarkerAlphaLocal|setMarkerBrush|setMarkerBrushLocal|setMarkerColor|setMarkerColorLocal|setMarkerDir|setMarkerDirLocal|setMarkerPos|setMarkerPosLocal|setMarkerShape|setMarkerShapeLocal|setMarkerSize|setMarkerSizeLocal|setMarkerText|setMarkerTextLocal|setMarkerType|setMarkerTypeLocal|setMass|setMimic|setMousePosition|setMusicEffect|setMusicEventHandler|setName|setNameSound|setObjectArguments|setObjectMaterial|setObjectMaterialGlobal|setObjectProxy|setObjectTexture|setObjectTextureGlobal|setObjectViewDistance|setOvercast|setOwner|setOxygenRemaining|setParticleCircle|setParticleClass|setParticleFire|setParticleParams|setParticleRandom|setPilotCameraDirection|setPilotCameraRotation|setPilotCameraTarget|setPilotLight|setPiPEffect|setPitch|setPlateNumber|setPlayable|setPlayerRespawnTime|setPos|setPosASL|setPosASL2|setPosASLW|setPosATL|setPosition|setPosWorld|setPylonLoadOut|setPylonsPriority|setRadioMsg|setRain|setRainbow|setRandomLip|setRank|setRectangular|setRepairCargo|setRotorBrakeRTD|setShadowDistance|setShotParents|setSide|setSimpleTaskAlwaysVisible|setSimpleTaskCustomData|setSimpleTaskDescription|setSimpleTaskDestination|setSimpleTaskTarget|setSimpleTaskType|setSimulWeatherLayers|setSize|setSkill|setSlingLoad|setSoundEffect|setSpeaker|setSpeech|setSpeedMode|setStamina|setStaminaScheme|setStatValue|setSuppression|setSystemOfUnits|setTargetAge|setTaskMarkerOffset|setTaskResult|setTaskState|setTerrainGrid|setText|setTimeMultiplier|setTitleEffect|setToneMapping|setToneMappingParams|setTrafficDensity|setTrafficDistance|setTrafficGap|setTrafficSpeed|setTriggerActivation|setTriggerArea|setTriggerStatements|setTriggerText|setTriggerTimeout|setTriggerType|setType|setUnconscious|setUnitAbility|setUnitLoadout|setUnitPos|setUnitPosWeak|setUnitRank|setUnitRecoilCoefficient|setUnitTrait|setUnloadInCombat|setUserActionText|setUserMFDText|setUserMFDValue|setVariable|setVectorDir|setVectorDirAndUp|setVectorUp|setVehicleAmmo|setVehicleAmmoDef|setVehicleArmor|setVehicleCargo|setVehicleId|setVehicleInit|setVehicleLock|setVehiclePosition|setVehicleRadar|setVehicleReceiveRemoteTargets|setVehicleReportOwnPosition|setVehicleReportRemoteTargets|setVehicleTIPars|setVehicleVarName|setVelocity|setVelocityModelSpace|setVelocityTransformation|setViewDistance|setVisibleIfTreeCollapsed|setWantedRpmRTD|setWaves|setWaypointBehaviour|setWaypointCombatMode|setWaypointCompletionRadius|setWaypointDescription|setWaypointForceBehaviour|setWaypointFormation|setWaypointHousePosition|setWaypointLoiterRadius|setWaypointLoiterType|setWaypointName|setWaypointPosition|setWaypointScript|setWaypointSpeed|setWaypointStatements|setWaypointTimeout|setWaypointType|setWaypointVisible|setWeaponReloadingTime|setWind|setWindDir|setWindForce|setWindStr|setWingForceScaleRTD|setWPPos|show3DIcons|showChat|showCinemaBorder|showCommandingMenu|showCompass|showCuratorCompass|showGPS|showHUD|showLegend|showMap|shownArtilleryComputer|shownChat|shownCompass|shownCuratorCompass|showNewEditorObject|shownGPS|shownHUD|shownMap|shownPad|shownRadio|shownScoretable|shownUAVFeed|shownWarrant|shownWatch|showPad|showRadio|showScoretable|showSubtitles|showUAVFeed|showWarrant|showWatch|showWaypoint|showWaypoints|side|sideAmbientLife|sideChat|sideEmpty|sideEnemy|sideFriendly|sideLogic|sideRadio|sideUnknown|simpleTasks|simulationEnabled|simulCloudDensity|simulCloudOcclusion|simulInClouds|simulWeatherSync|sin|size|sizeOf|skill|skillFinal|skipTime|sleep|sliderPosition|sliderRange|sliderSetPosition|sliderSetRange|sliderSetSpeed|sliderSpeed|slingLoadAssistantShown|soldierMagazines|someAmmo|sort|soundVolume|speaker|speed|speedMode|splitString|sqrt|squadParams|stance|startLoadingScreen|stop|stopEngineRTD|stopped|str|sunOrMoon|supportInfo|suppressFor|surfaceIsWater|surfaceNormal|surfaceType|swimInDepth|switchableUnits|switchAction|switchCamera|switchGesture|switchLight|switchMove|synchronizedObjects|synchronizedTriggers|synchronizedWaypoints|synchronizeObjectsAdd|synchronizeObjectsRemove|synchronizeTrigger|synchronizeWaypoint|systemChat|systemOfUnits|tan|targetKnowledge|targets|targetsAggregate|targetsQuery|taskAlwaysVisible|taskChildren|taskCompleted|taskCustomData|taskDescription|taskDestination|taskHint|taskMarkerOffset|taskNull|taskParent|taskResult|taskState|taskType|teamMember|teamMemberNull|teamName|teams|teamSwitch|teamSwitchEnabled|teamType|terminate|terrainIntersect|terrainIntersectASL|terrainIntersectAtASL|text|textLog|textLogFormat|tg|time|timeMultiplier|titleCut|titleFadeOut|titleObj|titleRsc|titleText|toArray|toFixed|toLower|toString|toUpper|triggerActivated|triggerActivation|triggerArea|triggerAttachedVehicle|triggerAttachObject|triggerAttachVehicle|triggerDynamicSimulation|triggerStatements|triggerText|triggerTimeout|triggerTimeoutCurrent|triggerType|turretLocal|turretOwner|turretUnit|tvAdd|tvClear|tvCollapse|tvCollapseAll|tvCount|tvCurSel|tvData|tvDelete|tvExpand|tvExpandAll|tvPicture|tvPictureRight|tvSetColor|tvSetCurSel|tvSetData|tvSetPicture|tvSetPictureColor|tvSetPictureColorDisabled|tvSetPictureColorSelected|tvSetPictureRight|tvSetPictureRightColor|tvSetPictureRightColorDisabled|tvSetPictureRightColorSelected|tvSetSelectColor|tvSetText|tvSetTooltip|tvSetValue|tvSort|tvSortByValue|tvText|tvTooltip|tvValue|type|typeName|typeOf|UAVControl|uiNamespace|uiSleep|unassignCurator|unassignItem|unassignTeam|unassignVehicle|underwater|uniform|uniformContainer|uniformItems|uniformMagazines|unitAddons|unitAimPosition|unitAimPositionVisual|unitBackpack|unitIsUAV|unitPos|unitReady|unitRecoilCoefficient|units|unitsBelowHeight|unlinkItem|unlockAchievement|unregisterTask|updateDrawIcon|updateMenuItem|updateObjectTree|useAIOperMapObstructionTest|useAISteeringComponent|useAudioTimeForMoves|userInputDisabled|vectorAdd|vectorCos|vectorCrossProduct|vectorDiff|vectorDir|vectorDirVisual|vectorDistance|vectorDistanceSqr|vectorDotProduct|vectorFromTo|vectorMagnitude|vectorMagnitudeSqr|vectorModelToWorld|vectorModelToWorldVisual|vectorMultiply|vectorNormalized|vectorUp|vectorUpVisual|vectorWorldToModel|vectorWorldToModelVisual|vehicle|vehicleCargoEnabled|vehicleChat|vehicleRadio|vehicleReceiveRemoteTargets|vehicleReportOwnPosition|vehicleReportRemoteTargets|vehicles|vehicleVarName|velocity|velocityModelSpace|verifySignature|vest|vestContainer|vestItems|vestMagazines|viewDistance|visibleCompass|visibleGPS|visibleMap|visiblePosition|visiblePositionASL|visibleScoretable|visibleWatch|waitUntil|waves|waypointAttachedObject|waypointAttachedVehicle|waypointAttachObject|waypointAttachVehicle|waypointBehaviour|waypointCombatMode|waypointCompletionRadius|waypointDescription|waypointForceBehaviour|waypointFormation|waypointHousePosition|waypointLoiterRadius|waypointLoiterType|waypointName|waypointPosition|waypoints|waypointScript|waypointsEnabledUAV|waypointShow|waypointSpeed|waypointStatements|waypointTimeout|waypointTimeoutCurrent|waypointType|waypointVisible|weaponAccessories|weaponAccessoriesCargo|weaponCargo|weaponDirection|weaponInertia|weaponLowered|weapons|weaponsItems|weaponsItemsCargo|weaponState|weaponsTurret|weightRTD|west|WFSideText|wind|windDir|windRTD|windStr|wingsForcesRTD|worldName|worldSize|worldToModel|worldToModelVisual|worldToScreen)\b/i,number:/(?:\$|\b0x)[\da-f]+\b|(?:\B\.\d+|\b\d+(?:\.\d+)?)(?:e[+-]?\d+)?\b/i,operator:/##|>>|&&|\|\||[!=<>]=?|[-+*/%#^]|\b(?:and|mod|not|or)\b/i,"magic-variable":{pattern:/\b(?:this|thisList|thisTrigger|_exception|_fnc_scriptName|_fnc_scriptNameParent|_forEachIndex|_this|_thisEventHandler|_thisFSM|_thisScript|_x)\b/i,alias:"keyword"},constant:/\bDIK(?:_[a-z\d]+)+\b/i}),r.languages.insertBefore("sqf","string",{macro:{pattern:/(^[ \t]*)#[a-z](?:[^\r\n\\]|\\(?:\r\n|[\s\S]))*/im,lookbehind:!0,greedy:!0,alias:"property",inside:{directive:{pattern:/#[a-z]+\b/i,alias:"keyword"},comment:r.languages.sqf.comment}}}),delete r.languages.sqf["class-name"]}return pGe}var gGe,tVt;function YUr(){if(tVt)return gGe;tVt=1,gGe=n,n.displayName="squirrel",n.aliases=[];function n(r){r.languages.squirrel=r.languages.extend("clike",{comment:[r.languages.clike.comment[0],{pattern:/(^|[^\\:])(?:\/\/|#).*/,lookbehind:!0,greedy:!0}],string:{pattern:/(^|[^\\"'@])(?:@"(?:[^"]|"")*"(?!")|"(?:[^\\\r\n"]|\\.)*")/,lookbehind:!0,greedy:!0},"class-name":{pattern:/(\b(?:class|enum|extends|instanceof)\s+)\w+(?:\.\w+)*/,lookbehind:!0,inside:{punctuation:/\./}},keyword:/\b(?:__FILE__|__LINE__|base|break|case|catch|class|clone|const|constructor|continue|default|delete|else|enum|extends|for|foreach|function|if|in|instanceof|local|null|resume|return|static|switch|this|throw|try|typeof|while|yield)\b/,number:/\b(?:0x[0-9a-fA-F]+|\d+(?:\.(?:\d+|[eE][+-]?\d+))?)\b/,operator:/\+\+|--|<=>|<[-<]|>>>?|&&?|\|\|?|[-+*/%!=<>]=?|[~^]|::?/,punctuation:/[(){}\[\],;.]/}),r.languages.insertBefore("squirrel","string",{char:{pattern:/(^|[^\\"'])'(?:[^\\']|\\(?:[xuU][0-9a-fA-F]{0,8}|[\s\S]))'/,lookbehind:!0,greedy:!0}}),r.languages.insertBefore("squirrel","operator",{"attribute-punctuation":{pattern:/<\/|\/>/,alias:"important"},lambda:{pattern:/@(?=\()/,alias:"operator"}})}return gGe}var mGe,nVt;function XUr(){if(nVt)return mGe;nVt=1,mGe=n,n.displayName="stan",n.aliases=[];function n(r){(function(o){var s=/\b(?:algebra_solver|algebra_solver_newton|integrate_1d|integrate_ode|integrate_ode_bdf|integrate_ode_rk45|map_rect|ode_(?:adams|bdf|ckrk|rk45)(?:_tol)?|ode_adjoint_tol_ctl|reduce_sum|reduce_sum_static)\b/;o.languages.stan={comment:/\/\/.*|\/\*[\s\S]*?\*\/|#(?!include).*/,string:{pattern:/"[\x20\x21\x23-\x5B\x5D-\x7E]*"/,greedy:!0},directive:{pattern:/^([ \t]*)#include\b.*/m,lookbehind:!0,alias:"property"},"function-arg":{pattern:RegExp("("+s.source+/\s*\(\s*/.source+")"+/[a-zA-Z]\w*/.source),lookbehind:!0,alias:"function"},constraint:{pattern:/(\b(?:int|matrix|real|row_vector|vector)\s*)<[^<>]*>/,lookbehind:!0,inside:{expression:{pattern:/(=\s*)\S(?:\S|\s+(?!\s))*?(?=\s*(?:>$|,\s*\w+\s*=))/,lookbehind:!0,inside:null},property:/\b[a-z]\w*(?=\s*=)/i,operator:/=/,punctuation:/^<|>$|,/}},keyword:[{pattern:/\bdata(?=\s*\{)|\b(?:functions|generated|model|parameters|quantities|transformed)\b/,alias:"program-block"},/\b(?:array|break|cholesky_factor_corr|cholesky_factor_cov|complex|continue|corr_matrix|cov_matrix|data|else|for|if|in|increment_log_prob|int|matrix|ordered|positive_ordered|print|real|reject|return|row_vector|simplex|target|unit_vector|vector|void|while)\b/,s],function:/\b[a-z]\w*(?=\s*\()/i,number:/(?:\b\d+(?:_\d+)*(?:\.(?:\d+(?:_\d+)*)?)?|\B\.\d+(?:_\d+)*)(?:E[+-]?\d+(?:_\d+)*)?i?(?!\w)/i,boolean:/\b(?:false|true)\b/,operator:/<-|\.[*/]=?|\|\|?|&&|[!=<>+\-*/]=?|['^%~?:]/,punctuation:/[()\[\]{},;]/},o.languages.stan.constraint.inside.expression.inside=o.languages.stan})(r)}return mGe}var bGe,rVt;function JUr(){if(rVt)return bGe;rVt=1,bGe=n,n.displayName="stylus",n.aliases=[];function n(r){(function(o){var s={pattern:/(\b\d+)(?:%|[a-z]+)/,lookbehind:!0},c={pattern:/(^|[^\w.-])-?(?:\d+(?:\.\d+)?|\.\d+)/,lookbehind:!0},d={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0},url:{pattern:/\burl\((["']?).*?\1\)/i,greedy:!0},string:{pattern:/("|')(?:(?!\1)[^\\\r\n]|\\(?:\r\n|[\s\S]))*\1/,greedy:!0},interpolation:null,func:null,important:/\B!(?:important|optional)\b/i,keyword:{pattern:/(^|\s+)(?:(?:else|for|if|return|unless)(?=\s|$)|@[\w-]+)/,lookbehind:!0},hexcode:/#[\da-f]{3,6}/i,color:[/\b(?:AliceBlue|AntiqueWhite|Aqua|Aquamarine|Azure|Beige|Bisque|Black|BlanchedAlmond|Blue|BlueViolet|Brown|BurlyWood|CadetBlue|Chartreuse|Chocolate|Coral|CornflowerBlue|Cornsilk|Crimson|Cyan|DarkBlue|DarkCyan|DarkGoldenRod|DarkGr[ae]y|DarkGreen|DarkKhaki|DarkMagenta|DarkOliveGreen|DarkOrange|DarkOrchid|DarkRed|DarkSalmon|DarkSeaGreen|DarkSlateBlue|DarkSlateGr[ae]y|DarkTurquoise|DarkViolet|DeepPink|DeepSkyBlue|DimGr[ae]y|DodgerBlue|FireBrick|FloralWhite|ForestGreen|Fuchsia|Gainsboro|GhostWhite|Gold|GoldenRod|Gr[ae]y|Green|GreenYellow|HoneyDew|HotPink|IndianRed|Indigo|Ivory|Khaki|Lavender|LavenderBlush|LawnGreen|LemonChiffon|LightBlue|LightCoral|LightCyan|LightGoldenRodYellow|LightGr[ae]y|LightGreen|LightPink|LightSalmon|LightSeaGreen|LightSkyBlue|LightSlateGr[ae]y|LightSteelBlue|LightYellow|Lime|LimeGreen|Linen|Magenta|Maroon|MediumAquaMarine|MediumBlue|MediumOrchid|MediumPurple|MediumSeaGreen|MediumSlateBlue|MediumSpringGreen|MediumTurquoise|MediumVioletRed|MidnightBlue|MintCream|MistyRose|Moccasin|NavajoWhite|Navy|OldLace|Olive|OliveDrab|Orange|OrangeRed|Orchid|PaleGoldenRod|PaleGreen|PaleTurquoise|PaleVioletRed|PapayaWhip|PeachPuff|Peru|Pink|Plum|PowderBlue|Purple|Red|RosyBrown|RoyalBlue|SaddleBrown|Salmon|SandyBrown|SeaGreen|SeaShell|Sienna|Silver|SkyBlue|SlateBlue|SlateGr[ae]y|Snow|SpringGreen|SteelBlue|Tan|Teal|Thistle|Tomato|Transparent|Turquoise|Violet|Wheat|White|WhiteSmoke|Yellow|YellowGreen)\b/i,{pattern:/\b(?:hsl|rgb)\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*\)\B|\b(?:hsl|rgb)a\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*,\s*(?:0|0?\.\d+|1)\s*\)\B/i,inside:{unit:s,number:c,function:/[\w-]+(?=\()/,punctuation:/[(),]/}}],entity:/\\[\da-f]{1,8}/i,unit:s,boolean:/\b(?:false|true)\b/,operator:[/~|[+!\/%<>?=]=?|[-:]=|\*[*=]?|\.{2,3}|&&|\|\||\B-\B|\b(?:and|in|is(?: a| defined| not|nt)?|not|or)\b/],number:c,punctuation:/[{}()\[\];:,]/};d.interpolation={pattern:/\{[^\r\n}:]+\}/,alias:"variable",inside:{delimiter:{pattern:/^\{|\}$/,alias:"punctuation"},rest:d}},d.func={pattern:/[\w-]+\([^)]*\).*/,inside:{function:/^[^(]+/,rest:d}},o.languages.stylus={"atrule-declaration":{pattern:/(^[ \t]*)@.+/m,lookbehind:!0,inside:{atrule:/^@[\w-]+/,rest:d}},"variable-declaration":{pattern:/(^[ \t]*)[\w$-]+\s*.?=[ \t]*(?:\{[^{}]*\}|\S.*|$)/m,lookbehind:!0,inside:{variable:/^\S+/,rest:d}},statement:{pattern:/(^[ \t]*)(?:else|for|if|return|unless)[ \t].+/m,lookbehind:!0,inside:{keyword:/^\S+/,rest:d}},"property-declaration":{pattern:/((?:^|\{)([ \t]*))(?:[\w-]|\{[^}\r\n]+\})+(?:\s*:\s*|[ \t]+)(?!\s)[^{\r\n]*(?:;|[^{\r\n,]$(?!(?:\r?\n|\r)(?:\{|\2[ \t])))/m,lookbehind:!0,inside:{property:{pattern:/^[^\s:]+/,inside:{interpolation:d.interpolation}},rest:d}},selector:{pattern:/(^[ \t]*)(?:(?=\S)(?:[^{}\r\n:()]|::?[\w-]+(?:\([^)\r\n]*\)|(?![\w-]))|\{[^}\r\n]+\})+)(?:(?:\r?\n|\r)(?:\1(?:(?=\S)(?:[^{}\r\n:()]|::?[\w-]+(?:\([^)\r\n]*\)|(?![\w-]))|\{[^}\r\n]+\})+)))*(?:,$|\{|(?=(?:\r?\n|\r)(?:\{|\1[ \t])))/m,lookbehind:!0,inside:{interpolation:d.interpolation,comment:d.comment,punctuation:/[{},]/}},func:d.func,string:d.string,comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0,greedy:!0},interpolation:d.interpolation,punctuation:/[{}()\[\];:.]/}})(r)}return bGe}var vGe,iVt;function ZUr(){if(iVt)return vGe;iVt=1,vGe=n,n.displayName="swift",n.aliases=[];function n(r){r.languages.swift={comment:{pattern:/(^|[^\\:])(?:\/\/.*|\/\*(?:[^/*]|\/(?!\*)|\*(?!\/)|\/\*(?:[^*]|\*(?!\/))*\*\/)*\*\/)/,lookbehind:!0,greedy:!0},"string-literal":[{pattern:RegExp(/(^|[^"#])/.source+"(?:"+/"(?:\\(?:\((?:[^()]|\([^()]*\))*\)|\r\n|[^(])|[^\\\r\n"])*"/.source+"|"+/"""(?:\\(?:\((?:[^()]|\([^()]*\))*\)|[^(])|[^\\"]|"(?!""))*"""/.source+")"+/(?!["#])/.source),lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/(\\\()(?:[^()]|\([^()]*\))*(?=\))/,lookbehind:!0,inside:null},"interpolation-punctuation":{pattern:/^\)|\\\($/,alias:"punctuation"},punctuation:/\\(?=[\r\n])/,string:/[\s\S]+/}},{pattern:RegExp(/(^|[^"#])(#+)/.source+"(?:"+/"(?:\\(?:#+\((?:[^()]|\([^()]*\))*\)|\r\n|[^#])|[^\\\r\n])*?"/.source+"|"+/"""(?:\\(?:#+\((?:[^()]|\([^()]*\))*\)|[^#])|[^\\])*?"""/.source+")\\2"),lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/(\\#+\()(?:[^()]|\([^()]*\))*(?=\))/,lookbehind:!0,inside:null},"interpolation-punctuation":{pattern:/^\)|\\#+\($/,alias:"punctuation"},string:/[\s\S]+/}}],directive:{pattern:RegExp(/#/.source+"(?:"+(/(?:elseif|if)\b/.source+"(?:[ 	]*"+/(?:![ \t]*)?(?:\b\w+\b(?:[ \t]*\((?:[^()]|\([^()]*\))*\))?|\((?:[^()]|\([^()]*\))*\))(?:[ \t]*(?:&&|\|\|))?/.source+")+")+"|"+/(?:else|endif)\b/.source+")"),alias:"property",inside:{"directive-name":/^#\w+/,boolean:/\b(?:false|true)\b/,number:/\b\d+(?:\.\d+)*\b/,operator:/!|&&|\|\||[<>]=?/,punctuation:/[(),]/}},literal:{pattern:/#(?:colorLiteral|column|dsohandle|file(?:ID|Literal|Path)?|function|imageLiteral|line)\b/,alias:"constant"},"other-directive":{pattern:/#\w+\b/,alias:"property"},attribute:{pattern:/@\w+/,alias:"atrule"},"function-definition":{pattern:/(\bfunc\s+)\w+/,lookbehind:!0,alias:"function"},label:{pattern:/\b(break|continue)\s+\w+|\b[a-zA-Z_]\w*(?=\s*:\s*(?:for|repeat|while)\b)/,lookbehind:!0,alias:"important"},keyword:/\b(?:Any|Protocol|Self|Type|actor|as|assignment|associatedtype|associativity|async|await|break|case|catch|class|continue|convenience|default|defer|deinit|didSet|do|dynamic|else|enum|extension|fallthrough|fileprivate|final|for|func|get|guard|higherThan|if|import|in|indirect|infix|init|inout|internal|is|isolated|lazy|left|let|lowerThan|mutating|none|nonisolated|nonmutating|open|operator|optional|override|postfix|precedencegroup|prefix|private|protocol|public|repeat|required|rethrows|return|right|safe|self|set|some|static|struct|subscript|super|switch|throw|throws|try|typealias|unowned|unsafe|var|weak|where|while|willSet)\b/,boolean:/\b(?:false|true)\b/,nil:{pattern:/\bnil\b/,alias:"constant"},"short-argument":/\$\d+\b/,omit:{pattern:/\b_\b/,alias:"keyword"},number:/\b(?:[\d_]+(?:\.[\de_]+)?|0x[a-f0-9_]+(?:\.[a-f0-9p_]+)?|0b[01_]+|0o[0-7_]+)\b/i,"class-name":/\b[A-Z](?:[A-Z_\d]*[a-z]\w*)?\b/,function:/\b[a-z_]\w*(?=\s*\()/i,constant:/\b(?:[A-Z_]{2,}|k[A-Z][A-Za-z_]+)\b/,operator:/[-+*/%=!<>&|^~?]+|\.[.\-+*/%=!<>&|^~?]+/,punctuation:/[{}[\]();,.:\\]/},r.languages.swift["string-literal"].forEach(function(o){o.inside.interpolation.inside=r.languages.swift})}return vGe}var wGe,oVt;function QUr(){if(oVt)return wGe;oVt=1,wGe=n,n.displayName="systemd",n.aliases=[];function n(r){(function(o){var s={pattern:/^[;#].*/m,greedy:!0},c=/"(?:[^\r\n"\\]|\\(?:[^\r]|\r\n?))*"(?!\S)/.source;o.languages.systemd={comment:s,section:{pattern:/^\[[^\n\r\[\]]*\](?=[ \t]*$)/m,greedy:!0,inside:{punctuation:/^\[|\]$/,"section-name":{pattern:/[\s\S]+/,alias:"selector"}}},key:{pattern:/^[^\s=]+(?=[ \t]*=)/m,greedy:!0,alias:"attr-name"},value:{pattern:RegExp(/(=[ \t]*(?!\s))/.source+"(?:"+c+`|(?=[^"\r
]))(?:`+(/[^\s\\]/.source+'|[ 	]+(?:(?![ 	"])|'+c+")|"+/\\[\r\n]+(?:[#;].*[\r\n]+)*(?![#;])/.source)+")*"),lookbehind:!0,greedy:!0,alias:"attr-value",inside:{comment:s,quoted:{pattern:RegExp(/(^|\s)/.source+c),lookbehind:!0,greedy:!0},punctuation:/\\$/m,boolean:{pattern:/^(?:false|no|off|on|true|yes)$/,greedy:!0}}},punctuation:/=/}})(r)}return wGe}var yGe,aVt;function xQe(){if(aVt)return yGe;aVt=1,yGe=n,n.displayName="t4Templating",n.aliases=[];function n(r){(function(o){function s(d,h,b){return{pattern:RegExp("<#"+d+"[\\s\\S]*?#>"),alias:"block",inside:{delimiter:{pattern:RegExp("^<#"+d+"|#>$"),alias:"important"},content:{pattern:/[\s\S]+/,inside:h,alias:b}}}}function c(d){var h=o.languages[d],b="language-"+d;return{block:{pattern:/<#[\s\S]+?#>/,inside:{directive:s("@",{"attr-value":{pattern:/=(?:("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|[^\s'">=]+)/,inside:{punctuation:/^=|^["']|["']$/}},keyword:/\b\w+(?=\s)/,"attr-name":/\b\w+/}),expression:s("=",h,b),"class-feature":s("\\+",h,b),standard:s("",h,b)}}}}o.languages["t4-templating"]=Object.defineProperty({},"createT4",{value:c})})(r)}return yGe}var EGe,sVt;function ezr(){if(sVt)return EGe;sVt=1;var n=xQe(),r=REe();EGe=o,o.displayName="t4Cs",o.aliases=[];function o(s){s.register(n),s.register(r),s.languages.t4=s.languages["t4-cs"]=s.languages["t4-templating"].createT4("csharp")}return EGe}var SGe,cVt;function ddn(){if(cVt)return SGe;cVt=1;var n=sdn();SGe=r,r.displayName="vbnet",r.aliases=[];function r(o){o.register(n),o.languages.vbnet=o.languages.extend("basic",{comment:[{pattern:/(?:!|REM\b).+/i,inside:{keyword:/^REM/i}},{pattern:/(^|[^\\:])'.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(^|[^"])"(?:""|[^"])*"(?!")/,lookbehind:!0,greedy:!0},keyword:/(?:\b(?:ADDHANDLER|ADDRESSOF|ALIAS|AND|ANDALSO|AS|BEEP|BLOAD|BOOLEAN|BSAVE|BYREF|BYTE|BYVAL|CALL(?: ABSOLUTE)?|CASE|CATCH|CBOOL|CBYTE|CCHAR|CDATE|CDBL|CDEC|CHAIN|CHAR|CHDIR|CINT|CLASS|CLEAR|CLNG|CLOSE|CLS|COBJ|COM|COMMON|CONST|CONTINUE|CSBYTE|CSHORT|CSNG|CSTR|CTYPE|CUINT|CULNG|CUSHORT|DATA|DATE|DECIMAL|DECLARE|DEF(?: FN| SEG|DBL|INT|LNG|SNG|STR)|DEFAULT|DELEGATE|DIM|DIRECTCAST|DO|DOUBLE|ELSE|ELSEIF|END|ENUM|ENVIRON|ERASE|ERROR|EVENT|EXIT|FALSE|FIELD|FILES|FINALLY|FOR(?: EACH)?|FRIEND|FUNCTION|GET|GETTYPE|GETXMLNAMESPACE|GLOBAL|GOSUB|GOTO|HANDLES|IF|IMPLEMENTS|IMPORTS|IN|INHERITS|INPUT|INTEGER|INTERFACE|IOCTL|IS|ISNOT|KEY|KILL|LET|LIB|LIKE|LINE INPUT|LOCATE|LOCK|LONG|LOOP|LSET|ME|MKDIR|MOD|MODULE|MUSTINHERIT|MUSTOVERRIDE|MYBASE|MYCLASS|NAME|NAMESPACE|NARROWING|NEW|NEXT|NOT|NOTHING|NOTINHERITABLE|NOTOVERRIDABLE|OBJECT|OF|OFF|ON(?: COM| ERROR| KEY| TIMER)?|OPEN|OPERATOR|OPTION(?: BASE)?|OPTIONAL|OR|ORELSE|OUT|OVERLOADS|OVERRIDABLE|OVERRIDES|PARAMARRAY|PARTIAL|POKE|PRIVATE|PROPERTY|PROTECTED|PUBLIC|PUT|RAISEEVENT|READ|READONLY|REDIM|REM|REMOVEHANDLER|RESTORE|RESUME|RETURN|RMDIR|RSET|RUN|SBYTE|SELECT(?: CASE)?|SET|SHADOWS|SHARED|SHELL|SHORT|SINGLE|SLEEP|STATIC|STEP|STOP|STRING|STRUCTURE|SUB|SWAP|SYNCLOCK|SYSTEM|THEN|THROW|TIMER|TO|TROFF|TRON|TRUE|TRY|TRYCAST|TYPE|TYPEOF|UINTEGER|ULONG|UNLOCK|UNTIL|USHORT|USING|VIEW PRINT|WAIT|WEND|WHEN|WHILE|WIDENING|WITH|WITHEVENTS|WRITE|WRITEONLY|XOR)|\B(?:#CONST|#ELSE|#ELSEIF|#END|#IF))(?:\$|\b)/i,punctuation:/[,;:(){}]/})}return SGe}var _Ge,uVt;function tzr(){if(uVt)return _Ge;uVt=1;var n=xQe(),r=ddn();_Ge=o,o.displayName="t4Vb",o.aliases=[];function o(s){s.register(n),s.register(r),s.languages["t4-vb"]=s.languages["t4-templating"].createT4("vbnet")}return _Ge}var kGe,lVt;function fdn(){if(lVt)return kGe;lVt=1,kGe=n,n.displayName="yaml",n.aliases=["yml"];function n(r){(function(o){var s=/[*&][^\s[\]{},]+/,c=/!(?:<[\w\-%#;/?:@&=+$,.!~*'()[\]]+>|(?:[a-zA-Z\d-]*!)?[\w\-%#;/?:@&=+$.~*'()]+)?/,d="(?:"+c.source+"(?:[ 	]+"+s.source+")?|"+s.source+"(?:[ 	]+"+c.source+")?)",h=/(?:[^\s\x00-\x08\x0e-\x1f!"#%&'*,\-:>?@[\]`{|}\x7f-\x84\x86-\x9f\ud800-\udfff\ufffe\uffff]|[?:-]<PLAIN>)(?:[ \t]*(?:(?![#:])<PLAIN>|:<PLAIN>))*/.source.replace(/<PLAIN>/g,function(){return/[^\s\x00-\x08\x0e-\x1f,[\]{}\x7f-\x84\x86-\x9f\ud800-\udfff\ufffe\uffff]/.source}),b=/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\\\r\n]|\\.)*'/.source;function y(E,k){k=(k||"").replace(/m/g,"")+"m";var x=/([:\-,[{]\s*(?:\s<<prop>>[ \t]+)?)(?:<<value>>)(?=[ \t]*(?:$|,|\]|\}|(?:[\r\n]\s*)?#))/.source.replace(/<<prop>>/g,function(){return d}).replace(/<<value>>/g,function(){return E});return RegExp(x,k)}o.languages.yaml={scalar:{pattern:RegExp(/([\-:]\s*(?:\s<<prop>>[ \t]+)?[|>])[ \t]*(?:((?:\r?\n|\r)[ \t]+)\S[^\r\n]*(?:\2[^\r\n]+)*)/.source.replace(/<<prop>>/g,function(){return d})),lookbehind:!0,alias:"string"},comment:/#.*/,key:{pattern:RegExp(/((?:^|[:\-,[{\r\n?])[ \t]*(?:<<prop>>[ \t]+)?)<<key>>(?=\s*:\s)/.source.replace(/<<prop>>/g,function(){return d}).replace(/<<key>>/g,function(){return"(?:"+h+"|"+b+")"})),lookbehind:!0,greedy:!0,alias:"atrule"},directive:{pattern:/(^[ \t]*)%.+/m,lookbehind:!0,alias:"important"},datetime:{pattern:y(/\d{4}-\d\d?-\d\d?(?:[tT]|[ \t]+)\d\d?:\d{2}:\d{2}(?:\.\d*)?(?:[ \t]*(?:Z|[-+]\d\d?(?::\d{2})?))?|\d{4}-\d{2}-\d{2}|\d\d?:\d{2}(?::\d{2}(?:\.\d*)?)?/.source),lookbehind:!0,alias:"number"},boolean:{pattern:y(/false|true/.source,"i"),lookbehind:!0,alias:"important"},null:{pattern:y(/null|~/.source,"i"),lookbehind:!0,alias:"important"},string:{pattern:y(b),lookbehind:!0,greedy:!0},number:{pattern:y(/[+-]?(?:0x[\da-f]+|0o[0-7]+|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|\.inf|\.nan)/.source,"i"),lookbehind:!0},tag:c,important:s,punctuation:/---|[:[\]{}\-,|>?]|\.\.\./},o.languages.yml=o.languages.yaml})(r)}return kGe}var TGe,dVt;function nzr(){if(dVt)return TGe;dVt=1;var n=fdn();TGe=r,r.displayName="tap",r.aliases=[];function r(o){o.register(n),o.languages.tap={fail:/not ok[^#{\n\r]*/,pass:/ok[^#{\n\r]*/,pragma:/pragma [+-][a-z]+/,bailout:/bail out!.*/i,version:/TAP version \d+/i,plan:/\b\d+\.\.\d+(?: +#.*)?/,subtest:{pattern:/# Subtest(?:: .*)?/,greedy:!0},punctuation:/[{}]/,directive:/#.*/,yamlish:{pattern:/(^[ \t]*)---[\s\S]*?[\r\n][ \t]*\.\.\.$/m,lookbehind:!0,inside:o.languages.yaml,alias:"language-yaml"}}}return TGe}var xGe,fVt;function rzr(){if(fVt)return xGe;fVt=1,xGe=n,n.displayName="tcl",n.aliases=[];function n(r){r.languages.tcl={comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0},string:{pattern:/"(?:[^"\\\r\n]|\\(?:\r\n|[\s\S]))*"/,greedy:!0},variable:[{pattern:/(\$)(?:::)?(?:[a-zA-Z0-9]+::)*\w+/,lookbehind:!0},{pattern:/(\$)\{[^}]+\}/,lookbehind:!0},{pattern:/(^[\t ]*set[ \t]+)(?:::)?(?:[a-zA-Z0-9]+::)*\w+/m,lookbehind:!0}],function:{pattern:/(^[\t ]*proc[ \t]+)\S+/m,lookbehind:!0},builtin:[{pattern:/(^[\t ]*)(?:break|class|continue|error|eval|exit|for|foreach|if|proc|return|switch|while)\b/m,lookbehind:!0},/\b(?:else|elseif)\b/],scope:{pattern:/(^[\t ]*)(?:global|upvar|variable)\b/m,lookbehind:!0,alias:"constant"},keyword:{pattern:/(^[\t ]*|\[)(?:Safe_Base|Tcl|after|append|apply|array|auto_(?:execok|import|load|mkindex|qualify|reset)|automkindex_old|bgerror|binary|catch|cd|chan|clock|close|concat|dde|dict|encoding|eof|exec|expr|fblocked|fconfigure|fcopy|file(?:event|name)?|flush|gets|glob|history|http|incr|info|interp|join|lappend|lassign|lindex|linsert|list|llength|load|lrange|lrepeat|lreplace|lreverse|lsearch|lset|lsort|math(?:func|op)|memory|msgcat|namespace|open|package|parray|pid|pkg_mkIndex|platform|puts|pwd|re_syntax|read|refchan|regexp|registry|regsub|rename|scan|seek|set|socket|source|split|string|subst|tcl(?:_endOfWord|_findLibrary|startOf(?:Next|Previous)Word|test|vars|wordBreak(?:After|Before))|tell|time|tm|trace|unknown|unload|unset|update|uplevel|vwait)\b/m,lookbehind:!0},operator:/!=?|\*\*?|==|&&?|\|\|?|<[=<]?|>[=>]?|[-+~\/%?^]|\b(?:eq|in|ne|ni)\b/,punctuation:/[{}()\[\]]/}}return xGe}var CGe,hVt;function izr(){if(hVt)return CGe;hVt=1,CGe=n,n.displayName="textile",n.aliases=[];function n(r){(function(o){var s=/\([^|()\n]+\)|\[[^\]\n]+\]|\{[^}\n]+\}/.source,c=/\)|\((?![^|()\n]+\))/.source;function d(O,P){return RegExp(O.replace(/<MOD>/g,function(){return"(?:"+s+")"}).replace(/<PAR>/g,function(){return"(?:"+c+")"}),P||"")}var h={css:{pattern:/\{[^{}]+\}/,inside:{rest:o.languages.css}},"class-id":{pattern:/(\()[^()]+(?=\))/,lookbehind:!0,alias:"attr-value"},lang:{pattern:/(\[)[^\[\]]+(?=\])/,lookbehind:!0,alias:"attr-value"},punctuation:/[\\\/]\d+|\S/},b=o.languages.textile=o.languages.extend("markup",{phrase:{pattern:/(^|\r|\n)\S[\s\S]*?(?=$|\r?\n\r?\n|\r\r)/,lookbehind:!0,inside:{"block-tag":{pattern:d(/^[a-z]\w*(?:<MOD>|<PAR>|[<>=])*\./.source),inside:{modifier:{pattern:d(/(^[a-z]\w*)(?:<MOD>|<PAR>|[<>=])+(?=\.)/.source),lookbehind:!0,inside:h},tag:/^[a-z]\w*/,punctuation:/\.$/}},list:{pattern:d(/^[*#]+<MOD>*\s+\S.*/.source,"m"),inside:{modifier:{pattern:d(/(^[*#]+)<MOD>+/.source),lookbehind:!0,inside:h},punctuation:/^[*#]+/}},table:{pattern:d(/^(?:(?:<MOD>|<PAR>|[<>=^~])+\.\s*)?(?:\|(?:(?:<MOD>|<PAR>|[<>=^~_]|[\\/]\d+)+\.|(?!(?:<MOD>|<PAR>|[<>=^~_]|[\\/]\d+)+\.))[^|]*)+\|/.source,"m"),inside:{modifier:{pattern:d(/(^|\|(?:\r?\n|\r)?)(?:<MOD>|<PAR>|[<>=^~_]|[\\/]\d+)+(?=\.)/.source),lookbehind:!0,inside:h},punctuation:/\||^\./}},inline:{pattern:d(/(^|[^a-zA-Z\d])(\*\*|__|\?\?|[*_%@+\-^~])<MOD>*.+?\2(?![a-zA-Z\d])/.source),lookbehind:!0,inside:{bold:{pattern:d(/(^(\*\*?)<MOD>*).+?(?=\2)/.source),lookbehind:!0},italic:{pattern:d(/(^(__?)<MOD>*).+?(?=\2)/.source),lookbehind:!0},cite:{pattern:d(/(^\?\?<MOD>*).+?(?=\?\?)/.source),lookbehind:!0,alias:"string"},code:{pattern:d(/(^@<MOD>*).+?(?=@)/.source),lookbehind:!0,alias:"keyword"},inserted:{pattern:d(/(^\+<MOD>*).+?(?=\+)/.source),lookbehind:!0},deleted:{pattern:d(/(^-<MOD>*).+?(?=-)/.source),lookbehind:!0},span:{pattern:d(/(^%<MOD>*).+?(?=%)/.source),lookbehind:!0},modifier:{pattern:d(/(^\*\*|__|\?\?|[*_%@+\-^~])<MOD>+/.source),lookbehind:!0,inside:h},punctuation:/[*_%?@+\-^~]+/}},"link-ref":{pattern:/^\[[^\]]+\]\S+$/m,inside:{string:{pattern:/(^\[)[^\]]+(?=\])/,lookbehind:!0},url:{pattern:/(^\])\S+$/,lookbehind:!0},punctuation:/[\[\]]/}},link:{pattern:d(/"<MOD>*[^"]+":.+?(?=[^\w/]?(?:\s|$))/.source),inside:{text:{pattern:d(/(^"<MOD>*)[^"]+(?=")/.source),lookbehind:!0},modifier:{pattern:d(/(^")<MOD>+/.source),lookbehind:!0,inside:h},url:{pattern:/(:).+/,lookbehind:!0},punctuation:/[":]/}},image:{pattern:d(/!(?:<MOD>|<PAR>|[<>=])*(?![<>=])[^!\s()]+(?:\([^)]+\))?!(?::.+?(?=[^\w/]?(?:\s|$)))?/.source),inside:{source:{pattern:d(/(^!(?:<MOD>|<PAR>|[<>=])*)(?![<>=])[^!\s()]+(?:\([^)]+\))?(?=!)/.source),lookbehind:!0,alias:"url"},modifier:{pattern:d(/(^!)(?:<MOD>|<PAR>|[<>=])+/.source),lookbehind:!0,inside:h},url:{pattern:/(:).+/,lookbehind:!0},punctuation:/[!:]/}},footnote:{pattern:/\b\[\d+\]/,alias:"comment",inside:{punctuation:/\[|\]/}},acronym:{pattern:/\b[A-Z\d]+\([^)]+\)/,inside:{comment:{pattern:/(\()[^()]+(?=\))/,lookbehind:!0},punctuation:/[()]/}},mark:{pattern:/\b\((?:C|R|TM)\)/,alias:"comment",inside:{punctuation:/[()]/}}}}}),y=b.phrase.inside,E={inline:y.inline,link:y.link,image:y.image,footnote:y.footnote,acronym:y.acronym,mark:y.mark};b.tag.pattern=/<\/?(?!\d)[a-z0-9]+(?:\s+[^\s>\/=]+(?:=(?:("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|[^\s'">=]+))?)*\s*\/?>/i;var k=y.inline.inside;k.bold.inside=E,k.italic.inside=E,k.inserted.inside=E,k.deleted.inside=E,k.span.inside=E;var x=y.table.inside;x.inline=E.inline,x.link=E.link,x.image=E.image,x.footnote=E.footnote,x.acronym=E.acronym,x.mark=E.mark})(r)}return CGe}var AGe,pVt;function ozr(){if(pVt)return AGe;pVt=1,AGe=n,n.displayName="toml",n.aliases=[];function n(r){(function(o){var s=/(?:[\w-]+|'[^'\n\r]*'|"(?:\\.|[^\\"\r\n])*")/.source;function c(d){return d.replace(/__/g,function(){return s})}o.languages.toml={comment:{pattern:/#.*/,greedy:!0},table:{pattern:RegExp(c(/(^[\t ]*\[\s*(?:\[\s*)?)__(?:\s*\.\s*__)*(?=\s*\])/.source),"m"),lookbehind:!0,greedy:!0,alias:"class-name"},key:{pattern:RegExp(c(/(^[\t ]*|[{,]\s*)__(?:\s*\.\s*__)*(?=\s*=)/.source),"m"),lookbehind:!0,greedy:!0,alias:"property"},string:{pattern:/"""(?:\\[\s\S]|[^\\])*?"""|'''[\s\S]*?'''|'[^'\n\r]*'|"(?:\\.|[^\\"\r\n])*"/,greedy:!0},date:[{pattern:/\b\d{4}-\d{2}-\d{2}(?:[T\s]\d{2}:\d{2}:\d{2}(?:\.\d+)?(?:Z|[+-]\d{2}:\d{2})?)?\b/i,alias:"number"},{pattern:/\b\d{2}:\d{2}:\d{2}(?:\.\d+)?\b/,alias:"number"}],number:/(?:\b0(?:x[\da-zA-Z]+(?:_[\da-zA-Z]+)*|o[0-7]+(?:_[0-7]+)*|b[10]+(?:_[10]+)*))\b|[-+]?\b\d+(?:_\d+)*(?:\.\d+(?:_\d+)*)?(?:[eE][+-]?\d+(?:_\d+)*)?\b|[-+]?\b(?:inf|nan)\b/,boolean:/\b(?:false|true)\b/,punctuation:/[.,=[\]{}]/}})(r)}return AGe}var IGe,gVt;function azr(){if(gVt)return IGe;gVt=1,IGe=n,n.displayName="tremor",n.aliases=[];function n(r){(function(o){o.languages.tremor={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:--|\/\/|#).*)/,lookbehind:!0},"interpolated-string":null,extractor:{pattern:/\b[a-z_]\w*\|(?:[^\r\n\\|]|\\(?:\r\n|[\s\S]))*\|/i,greedy:!0,inside:{regex:{pattern:/(^re)\|[\s\S]+/,lookbehind:!0},function:/^\w+/,value:/\|[\s\S]+/}},identifier:{pattern:/`[^`]*`/,greedy:!0},function:/\b[a-z_]\w*(?=\s*(?:::\s*<|\())\b/,keyword:/\b(?:args|as|by|case|config|connect|connector|const|copy|create|default|define|deploy|drop|each|emit|end|erase|event|flow|fn|for|from|group|having|insert|into|intrinsic|let|links|match|merge|mod|move|of|operator|patch|pipeline|recur|script|select|set|sliding|state|stream|to|tumbling|update|use|when|where|window|with)\b/,boolean:/\b(?:false|null|true)\b/i,number:/\b(?:0b[01_]*|0x[0-9a-fA-F_]*|\d[\d_]*(?:\.\d[\d_]*)?(?:[Ee][+-]?[\d_]+)?)\b/,"pattern-punctuation":{pattern:/%(?=[({[])/,alias:"punctuation"},operator:/[-+*\/%~!^]=?|=[=>]?|&[&=]?|\|[|=]?|<<?=?|>>?>?=?|(?:absent|and|not|or|present|xor)\b/,punctuation:/::|[;\[\]()\{\},.:]/};var s=/#\{(?:[^"{}]|\{[^{}]*\}|"(?:[^"\\\r\n]|\\(?:\r\n|[\s\S]))*")*\}/.source;o.languages.tremor["interpolated-string"]={pattern:RegExp(/(^|[^\\])/.source+'(?:"""(?:'+/[^"\\#]|\\[\s\S]|"(?!"")|#(?!\{)/.source+"|"+s+')*"""|"(?:'+/[^"\\\r\n#]|\\(?:\r\n|[\s\S])|#(?!\{)/.source+"|"+s+')*")'),lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:RegExp(s),inside:{punctuation:/^#\{|\}$/,expression:{pattern:/[\s\S]+/,inside:o.languages.tremor}}},string:/[\s\S]+/}},o.languages.troy=o.languages.tremor,o.languages.trickle=o.languages.tremor})(r)}return IGe}var OGe,mVt;function szr(){if(mVt)return OGe;mVt=1;var n=udn(),r=_Qe();OGe=o,o.displayName="tsx",o.aliases=[];function o(s){s.register(n),s.register(r),function(c){var d=c.util.clone(c.languages.typescript);c.languages.tsx=c.languages.extend("jsx",d),delete c.languages.tsx.parameter,delete c.languages.tsx["literal-property"];var h=c.languages.tsx.tag;h.pattern=RegExp(/(^|[^\w$]|(?=<\/))/.source+"(?:"+h.pattern.source+")",h.pattern.flags),h.lookbehind=!0}(s)}return OGe}var NGe,bVt;function czr(){if(bVt)return NGe;bVt=1;var n=oy();NGe=r,r.displayName="tt2",r.aliases=[];function r(o){o.register(n),function(s){s.languages.tt2=s.languages.extend("clike",{comment:/#.*|\[%#[\s\S]*?%\]/,keyword:/\b(?:BLOCK|CALL|CASE|CATCH|CLEAR|DEBUG|DEFAULT|ELSE|ELSIF|END|FILTER|FINAL|FOREACH|GET|IF|IN|INCLUDE|INSERT|LAST|MACRO|META|NEXT|PERL|PROCESS|RAWPERL|RETURN|SET|STOP|SWITCH|TAGS|THROW|TRY|UNLESS|USE|WHILE|WRAPPER)\b/,punctuation:/[[\]{},()]/}),s.languages.insertBefore("tt2","number",{operator:/=[>=]?|!=?|<=?|>=?|&&|\|\|?|\b(?:and|not|or)\b/,variable:{pattern:/\b[a-z]\w*(?:\s*\.\s*(?:\d+|\$?[a-z]\w*))*\b/i}}),s.languages.insertBefore("tt2","keyword",{delimiter:{pattern:/^(?:\[%|%%)-?|-?%\]$/,alias:"punctuation"}}),s.languages.insertBefore("tt2","string",{"single-quoted-string":{pattern:/'[^\\']*(?:\\[\s\S][^\\']*)*'/,greedy:!0,alias:"string"},"double-quoted-string":{pattern:/"[^\\"]*(?:\\[\s\S][^\\"]*)*"/,greedy:!0,alias:"string",inside:{variable:{pattern:/\$(?:[a-z]\w*(?:\.(?:\d+|\$?[a-z]\w*))*)/i}}}}),delete s.languages.tt2.string,s.hooks.add("before-tokenize",function(c){var d=/\[%[\s\S]+?%\]/g;s.languages["markup-templating"].buildPlaceholders(c,"tt2",d)}),s.hooks.add("after-tokenize",function(c){s.languages["markup-templating"].tokenizePlaceholders(c,"tt2")})}(o)}return NGe}var RGe,vVt;function uzr(){if(vVt)return RGe;vVt=1;var n=oy();RGe=r,r.displayName="twig",r.aliases=[];function r(o){o.register(n),o.languages.twig={comment:/^\{#[\s\S]*?#\}$/,"tag-name":{pattern:/(^\{%-?\s*)\w+/,lookbehind:!0,alias:"keyword"},delimiter:{pattern:/^\{[{%]-?|-?[%}]\}$/,alias:"punctuation"},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,inside:{punctuation:/^['"]|['"]$/}},keyword:/\b(?:even|if|odd)\b/,boolean:/\b(?:false|null|true)\b/,number:/\b0x[\dA-Fa-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][-+]?\d+)?/,operator:[{pattern:/(\s)(?:and|b-and|b-or|b-xor|ends with|in|is|matches|not|or|same as|starts with)(?=\s)/,lookbehind:!0},/[=<>]=?|!=|\*\*?|\/\/?|\?:?|[-+~%|]/],punctuation:/[()\[\]{}:.,]/},o.hooks.add("before-tokenize",function(s){if(s.language==="twig"){var c=/\{(?:#[\s\S]*?#|%[\s\S]*?%|\{[\s\S]*?\})\}/g;o.languages["markup-templating"].buildPlaceholders(s,"twig",c)}}),o.hooks.add("after-tokenize",function(s){o.languages["markup-templating"].tokenizePlaceholders(s,"twig")})}return RGe}var PGe,wVt;function lzr(){if(wVt)return PGe;wVt=1,PGe=n,n.displayName="typoscript",n.aliases=["tsconfig"];function n(r){(function(o){var s=/\b(?:ACT|ACTIFSUB|CARRAY|CASE|CLEARGIF|COA|COA_INT|CONSTANTS|CONTENT|CUR|EDITPANEL|EFFECT|EXT|FILE|FLUIDTEMPLATE|FORM|FRAME|FRAMESET|GIFBUILDER|GMENU|GMENU_FOLDOUT|GMENU_LAYERS|GP|HMENU|HRULER|HTML|IENV|IFSUB|IMAGE|IMGMENU|IMGMENUITEM|IMGTEXT|IMG_RESOURCE|INCLUDE_TYPOSCRIPT|JSMENU|JSMENUITEM|LLL|LOAD_REGISTER|NO|PAGE|RECORDS|RESTORE_REGISTER|TEMPLATE|TEXT|TMENU|TMENUITEM|TMENU_LAYERS|USER|USER_INT|_GIFBUILDER|global|globalString|globalVar)\b/;o.languages.typoscript={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0},{pattern:/(^|[^\\:= \t]|(?:^|[^= \t])[ \t]+)\/\/.*/,lookbehind:!0,greedy:!0},{pattern:/(^|[^"'])#.*/,lookbehind:!0,greedy:!0}],function:[{pattern:/<INCLUDE_TYPOSCRIPT:\s*source\s*=\s*(?:"[^"\r\n]*"|'[^'\r\n]*')\s*>/,inside:{string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,inside:{keyword:s}},keyword:{pattern:/INCLUDE_TYPOSCRIPT/}}},{pattern:/@import\s*(?:"[^"\r\n]*"|'[^'\r\n]*')/,inside:{string:/"[^"\r\n]*"|'[^'\r\n]*'/}}],string:{pattern:/^([^=]*=[< ]?)(?:(?!\]\n).)*/,lookbehind:!0,inside:{function:/\{\$.*\}/,keyword:s,number:/^\d+$/,punctuation:/[,|:]/}},keyword:s,number:{pattern:/\b\d+\s*[.{=]/,inside:{operator:/[.{=]/}},tag:{pattern:/\.?[-\w\\]+\.?/,inside:{punctuation:/\./}},punctuation:/[{}[\];(),.:|]/,operator:/[<>]=?|[!=]=?=?|--?|\+\+?|&&?|\|\|?|[?*/~^%]/},o.languages.tsconfig=o.languages.typoscript})(r)}return PGe}var LGe,yVt;function dzr(){if(yVt)return LGe;yVt=1,LGe=n,n.displayName="unrealscript",n.aliases=["uc","uscript"];function n(r){r.languages.unrealscript={comment:/\/\/.*|\/\*[\s\S]*?\*\//,string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},category:{pattern:/(\b(?:(?:autoexpand|hide|show)categories|var)\s*\()[^()]+(?=\))/,lookbehind:!0,greedy:!0,alias:"property"},metadata:{pattern:/(\w\s*)<\s*\w+\s*=[^<>|=\r\n]+(?:\|\s*\w+\s*=[^<>|=\r\n]+)*>/,lookbehind:!0,greedy:!0,inside:{property:/\b\w+(?=\s*=)/,operator:/=/,punctuation:/[<>|]/}},macro:{pattern:/`\w+/,alias:"property"},"class-name":{pattern:/(\b(?:class|enum|extends|interface|state(?:\(\))?|struct|within)\s+)\w+/,lookbehind:!0},keyword:/\b(?:abstract|actor|array|auto|autoexpandcategories|bool|break|byte|case|class|classgroup|client|coerce|collapsecategories|config|const|continue|default|defaultproperties|delegate|dependson|deprecated|do|dontcollapsecategories|editconst|editinlinenew|else|enum|event|exec|export|extends|final|float|for|forcescriptorder|foreach|function|goto|guid|hidecategories|hidedropdown|if|ignores|implements|inherits|input|int|interface|iterator|latent|local|material|name|native|nativereplication|noexport|nontransient|noteditinlinenew|notplaceable|operator|optional|out|pawn|perobjectconfig|perobjectlocalized|placeable|postoperator|preoperator|private|protected|reliable|replication|return|server|showcategories|simulated|singular|state|static|string|struct|structdefault|structdefaultproperties|switch|texture|transient|travel|unreliable|until|var|vector|while|within)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,boolean:/\b(?:false|true)\b/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,operator:/>>|<<|--|\+\+|\*\*|[-+*/~!=<>$@]=?|&&?|\|\|?|\^\^?|[?:%]|\b(?:ClockwiseFrom|Cross|Dot)\b/,punctuation:/[()[\]{};,.]/},r.languages.uc=r.languages.uscript=r.languages.unrealscript}return LGe}var DGe,EVt;function fzr(){if(EVt)return DGe;EVt=1,DGe=n,n.displayName="uorazor",n.aliases=[];function n(r){r.languages.uorazor={"comment-hash":{pattern:/#.*/,alias:"comment",greedy:!0},"comment-slash":{pattern:/\/\/.*/,alias:"comment",greedy:!0},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,inside:{punctuation:/^['"]|['"]$/},greedy:!0},"source-layers":{pattern:/\b(?:arms|backpack|blue|bracelet|cancel|clear|cloak|criminal|earrings|enemy|facialhair|friend|friendly|gloves|gray|grey|ground|hair|head|innerlegs|innertorso|innocent|lefthand|middletorso|murderer|neck|nonfriendly|onehandedsecondary|outerlegs|outertorso|pants|red|righthand|ring|self|shirt|shoes|talisman|waist)\b/i,alias:"function"},"source-commands":{pattern:/\b(?:alliance|attack|cast|clearall|clearignore|clearjournal|clearlist|clearsysmsg|createlist|createtimer|dclick|dclicktype|dclickvar|dress|dressconfig|drop|droprelloc|emote|getlabel|guild|gumpclose|gumpresponse|hotkey|ignore|lasttarget|lift|lifttype|menu|menuresponse|msg|org|organize|organizer|overhead|pause|poplist|potion|promptresponse|pushlist|removelist|removetimer|rename|restock|say|scav|scavenger|script|setability|setlasttarget|setskill|settimer|setvar|sysmsg|target|targetloc|targetrelloc|targettype|undress|unignore|unsetvar|useobject|useonce|useskill|usetype|virtue|wait|waitforgump|waitformenu|waitforprompt|waitforstat|waitforsysmsg|waitfortarget|walk|wfsysmsg|wft|whisper|yell)\b/,alias:"function"},"tag-name":{pattern:/(^\{%-?\s*)\w+/,lookbehind:!0,alias:"keyword"},delimiter:{pattern:/^\{[{%]-?|-?[%}]\}$/,alias:"punctuation"},function:/\b(?:atlist|close|closest|count|counter|counttype|dead|dex|diffhits|diffmana|diffstam|diffweight|find|findbuff|finddebuff|findlayer|findtype|findtypelist|followers|gumpexists|hidden|hits|hp|hue|human|humanoid|ingump|inlist|insysmessage|insysmsg|int|invul|lhandempty|list|listexists|mana|maxhits|maxhp|maxmana|maxstam|maxweight|monster|mounted|name|next|noto|paralyzed|poisoned|position|prev|previous|queued|rand|random|rhandempty|skill|stam|str|targetexists|timer|timerexists|varexist|warmode|weight)\b/,keyword:/\b(?:and|as|break|continue|else|elseif|endfor|endif|endwhile|for|if|loop|not|or|replay|stop|while)\b/,boolean:/\b(?:false|null|true)\b/,number:/\b0x[\dA-Fa-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][-+]?\d+)?/,operator:[{pattern:/(\s)(?:and|b-and|b-or|b-xor|ends with|in|is|matches|not|or|same as|starts with)(?=\s)/,lookbehind:!0},/[=<>]=?|!=|\*\*?|\/\/?|\?:?|[-+~%|]/],punctuation:/[()\[\]{}:.,]/}}return DGe}var MGe,SVt;function hzr(){if(SVt)return MGe;SVt=1,MGe=n,n.displayName="uri",n.aliases=["url"];function n(r){r.languages.uri={scheme:{pattern:/^[a-z][a-z0-9+.-]*:/im,greedy:!0,inside:{"scheme-delimiter":/:$/}},fragment:{pattern:/#[\w\-.~!$&'()*+,;=%:@/?]*/,inside:{"fragment-delimiter":/^#/}},query:{pattern:/\?[\w\-.~!$&'()*+,;=%:@/?]*/,inside:{"query-delimiter":{pattern:/^\?/,greedy:!0},"pair-delimiter":/[&;]/,pair:{pattern:/^[^=][\s\S]*/,inside:{key:/^[^=]+/,value:{pattern:/(^=)[\s\S]+/,lookbehind:!0}}}}},authority:{pattern:RegExp(/^\/\//.source+/(?:[\w\-.~!$&'()*+,;=%:]*@)?/.source+("(?:"+/\[(?:[0-9a-fA-F:.]{2,48}|v[0-9a-fA-F]+\.[\w\-.~!$&'()*+,;=]+)\]/.source+"|"+/[\w\-.~!$&'()*+,;=%]*/.source+")")+/(?::\d*)?/.source,"m"),inside:{"authority-delimiter":/^\/\//,"user-info-segment":{pattern:/^[\w\-.~!$&'()*+,;=%:]*@/,inside:{"user-info-delimiter":/@$/,"user-info":/^[\w\-.~!$&'()*+,;=%:]+/}},"port-segment":{pattern:/:\d*$/,inside:{"port-delimiter":/^:/,port:/^\d+/}},host:{pattern:/[\s\S]+/,inside:{"ip-literal":{pattern:/^\[[\s\S]+\]$/,inside:{"ip-literal-delimiter":/^\[|\]$/,"ipv-future":/^v[\s\S]+/,"ipv6-address":/^[\s\S]+/}},"ipv4-address":/^(?:(?:[03-9]\d?|[12]\d{0,2})\.){3}(?:[03-9]\d?|[12]\d{0,2})$/}}}},path:{pattern:/^[\w\-.~!$&'()*+,;=%:@/]+/m,inside:{"path-separator":/\//}}},r.languages.url=r.languages.uri}return MGe}var jGe,_Vt;function pzr(){if(_Vt)return jGe;_Vt=1,jGe=n,n.displayName="v",n.aliases=[];function n(r){(function(o){var s={pattern:/[\s\S]+/,inside:null};o.languages.v=o.languages.extend("clike",{string:{pattern:/r?(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,alias:"quoted-string",greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:\{[^{}]*\}|\w+(?:\.\w+(?:\([^\(\)]*\))?|\[[^\[\]]+\])*)/,lookbehind:!0,inside:{"interpolation-variable":{pattern:/^\$\w[\s\S]*$/,alias:"variable"},"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},"interpolation-expression":s}}}},"class-name":{pattern:/(\b(?:enum|interface|struct|type)\s+)(?:C\.)?\w+/,lookbehind:!0},keyword:/(?:\b(?:__global|as|asm|assert|atomic|break|chan|const|continue|defer|else|embed|enum|fn|for|go(?:to)?|if|import|in|interface|is|lock|match|module|mut|none|or|pub|return|rlock|select|shared|sizeof|static|struct|type(?:of)?|union|unsafe)|\$(?:else|for|if)|#(?:flag|include))\b/,number:/\b(?:0x[a-f\d]+(?:_[a-f\d]+)*|0b[01]+(?:_[01]+)*|0o[0-7]+(?:_[0-7]+)*|\d+(?:_\d+)*(?:\.\d+(?:_\d+)*)?)\b/i,operator:/~|\?|[*\/%^!=]=?|\+[=+]?|-[=-]?|\|[=|]?|&(?:=|&|\^=?)?|>(?:>=?|=)?|<(?:<=?|=|-)?|:=|\.\.\.?/,builtin:/\b(?:any(?:_float|_int)?|bool|byte(?:ptr)?|charptr|f(?:32|64)|i(?:8|16|64|128|nt)|rune|size_t|string|u(?:16|32|64|128)|voidptr)\b/}),s.inside=o.languages.v,o.languages.insertBefore("v","string",{char:{pattern:/`(?:\\`|\\?[^`]{1,2})`/,alias:"rune"}}),o.languages.insertBefore("v","operator",{attribute:{pattern:/(^[\t ]*)\[(?:deprecated|direct_array_access|flag|inline|live|ref_only|typedef|unsafe_fn|windows_stdcall)\]/m,lookbehind:!0,alias:"annotation",inside:{punctuation:/[\[\]]/,keyword:/\w+/}},generic:{pattern:/<\w+>(?=\s*[\)\{])/,inside:{punctuation:/[<>]/,"class-name":/\w+/}}}),o.languages.insertBefore("v","function",{"generic-function":{pattern:/\b\w+\s*<\w+>(?=\()/,inside:{function:/^\w+/,generic:{pattern:/<\w+>/,inside:o.languages.v.generic.inside}}}})})(r)}return jGe}var $Ge,kVt;function gzr(){if(kVt)return $Ge;kVt=1,$Ge=n,n.displayName="vala",n.aliases=[];function n(r){r.languages.vala=r.languages.extend("clike",{"class-name":[{pattern:/\b[A-Z]\w*(?:\.\w+)*\b(?=(?:\?\s+|\*?\s+\*?)\w)/,inside:{punctuation:/\./}},{pattern:/(\[)[A-Z]\w*(?:\.\w+)*\b/,lookbehind:!0,inside:{punctuation:/\./}},{pattern:/(\b(?:class|interface)\s+[A-Z]\w*(?:\.\w+)*\s*:\s*)[A-Z]\w*(?:\.\w+)*\b/,lookbehind:!0,inside:{punctuation:/\./}},{pattern:/((?:\b(?:class|enum|interface|new|struct)\s+)|(?:catch\s+\())[A-Z]\w*(?:\.\w+)*\b/,lookbehind:!0,inside:{punctuation:/\./}}],keyword:/\b(?:abstract|as|assert|async|base|bool|break|case|catch|char|class|const|construct|continue|default|delegate|delete|do|double|dynamic|else|ensures|enum|errordomain|extern|finally|float|for|foreach|get|if|in|inline|int|int16|int32|int64|int8|interface|internal|is|lock|long|namespace|new|null|out|override|owned|params|private|protected|public|ref|requires|return|set|short|signal|sizeof|size_t|ssize_t|static|string|struct|switch|this|throw|throws|try|typeof|uchar|uint|uint16|uint32|uint64|uint8|ulong|unichar|unowned|ushort|using|value|var|virtual|void|volatile|weak|while|yield)\b/i,function:/\b\w+(?=\s*\()/,number:/(?:\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)(?:f|u?l?)?/i,operator:/\+\+|--|&&|\|\||<<=?|>>=?|=>|->|~|[+\-*\/%&^|=!<>]=?|\?\??|\.\.\./,punctuation:/[{}[\];(),.:]/,constant:/\b[A-Z0-9_]+\b/}),r.languages.insertBefore("vala","string",{"raw-string":{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string"},"template-string":{pattern:/@"[\s\S]*?"/,greedy:!0,inside:{interpolation:{pattern:/\$(?:\([^)]*\)|[a-zA-Z]\w*)/,inside:{delimiter:{pattern:/^\$\(?|\)$/,alias:"punctuation"},rest:r.languages.vala}},string:/[\s\S]+/}}}),r.languages.insertBefore("vala","keyword",{regex:{pattern:/\/(?:\[(?:[^\]\\\r\n]|\\.)*\]|\\.|[^/\\\[\r\n])+\/[imsx]{0,4}(?=\s*(?:$|[\r\n,.;})\]]))/,greedy:!0,inside:{"regex-source":{pattern:/^(\/)[\s\S]+(?=\/[a-z]*$)/,lookbehind:!0,alias:"language-regex",inside:r.languages.regex},"regex-delimiter":/^\//,"regex-flags":/^[a-z]+$/}}})}return $Ge}var FGe,TVt;function mzr(){if(TVt)return FGe;TVt=1,FGe=n,n.displayName="velocity",n.aliases=[];function n(r){(function(o){o.languages.velocity=o.languages.extend("markup",{});var s={variable:{pattern:/(^|[^\\](?:\\\\)*)\$!?(?:[a-z][\w-]*(?:\([^)]*\))?(?:\.[a-z][\w-]*(?:\([^)]*\))?|\[[^\]]+\])*|\{[^}]+\})/i,lookbehind:!0,inside:{}},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},number:/\b\d+\b/,boolean:/\b(?:false|true)\b/,operator:/[=!<>]=?|[+*/%-]|&&|\|\||\.\.|\b(?:eq|g[et]|l[et]|n(?:e|ot))\b/,punctuation:/[(){}[\]:,.]/};s.variable.inside={string:s.string,function:{pattern:/([^\w-])[a-z][\w-]*(?=\()/,lookbehind:!0},number:s.number,boolean:s.boolean,punctuation:s.punctuation},o.languages.insertBefore("velocity","comment",{unparsed:{pattern:/(^|[^\\])#\[\[[\s\S]*?\]\]#/,lookbehind:!0,greedy:!0,inside:{punctuation:/^#\[\[|\]\]#$/}},"velocity-comment":[{pattern:/(^|[^\\])#\*[\s\S]*?\*#/,lookbehind:!0,greedy:!0,alias:"comment"},{pattern:/(^|[^\\])##.*/,lookbehind:!0,greedy:!0,alias:"comment"}],directive:{pattern:/(^|[^\\](?:\\\\)*)#@?(?:[a-z][\w-]*|\{[a-z][\w-]*\})(?:\s*\((?:[^()]|\([^()]*\))*\))?/i,lookbehind:!0,inside:{keyword:{pattern:/^#@?(?:[a-z][\w-]*|\{[a-z][\w-]*\})|\bin\b/,inside:{punctuation:/[{}]/}},rest:s}},variable:s.variable}),o.languages.velocity.tag.inside["attr-value"].inside.rest=o.languages.velocity})(r)}return FGe}var BGe,xVt;function bzr(){if(xVt)return BGe;xVt=1,BGe=n,n.displayName="verilog",n.aliases=[];function n(r){r.languages.verilog={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0},"kernel-function":{pattern:/\B\$\w+\b/,alias:"property"},constant:/\B`\w+\b/,function:/\b\w+(?=\()/,keyword:/\b(?:alias|and|assert|assign|assume|automatic|before|begin|bind|bins|binsof|bit|break|buf|bufif0|bufif1|byte|case|casex|casez|cell|chandle|class|clocking|cmos|config|const|constraint|context|continue|cover|covergroup|coverpoint|cross|deassign|default|defparam|design|disable|dist|do|edge|else|end|endcase|endclass|endclocking|endconfig|endfunction|endgenerate|endgroup|endinterface|endmodule|endpackage|endprimitive|endprogram|endproperty|endsequence|endspecify|endtable|endtask|enum|event|expect|export|extends|extern|final|first_match|for|force|foreach|forever|fork|forkjoin|function|generate|genvar|highz0|highz1|if|iff|ifnone|ignore_bins|illegal_bins|import|incdir|include|initial|inout|input|inside|instance|int|integer|interface|intersect|join|join_any|join_none|large|liblist|library|local|localparam|logic|longint|macromodule|matches|medium|modport|module|nand|negedge|new|nmos|nor|noshowcancelled|not|notif0|notif1|null|or|output|package|packed|parameter|pmos|posedge|primitive|priority|program|property|protected|pull0|pull1|pulldown|pullup|pulsestyle_ondetect|pulsestyle_onevent|pure|rand|randc|randcase|randsequence|rcmos|real|realtime|ref|reg|release|repeat|return|rnmos|rpmos|rtran|rtranif0|rtranif1|scalared|sequence|shortint|shortreal|showcancelled|signed|small|solve|specify|specparam|static|string|strong0|strong1|struct|super|supply0|supply1|table|tagged|task|this|throughout|time|timeprecision|timeunit|tran|tranif0|tranif1|tri|tri0|tri1|triand|trior|trireg|type|typedef|union|unique|unsigned|use|uwire|var|vectored|virtual|void|wait|wait_order|wand|weak0|weak1|while|wildcard|wire|with|within|wor|xnor|xor)\b/,important:/\b(?:always|always_comb|always_ff|always_latch)\b(?: *@)?/,number:/\B##?\d+|(?:\b\d+)?'[odbh] ?[\da-fzx_?]+|\b(?:\d*[._])?\d+(?:e[-+]?\d+)?/i,operator:/[-+{}^~%*\/?=!<>&|]+/,punctuation:/[[\];(),.:]/}}return BGe}var UGe,CVt;function vzr(){if(CVt)return UGe;CVt=1,UGe=n,n.displayName="vhdl",n.aliases=[];function n(r){r.languages.vhdl={comment:/--.+/,"vhdl-vectors":{pattern:/\b[oxb]"[\da-f_]+"|"[01uxzwlh-]+"/i,alias:"number"},"quoted-function":{pattern:/"\S+?"(?=\()/,alias:"function"},string:/"(?:[^\\"\r\n]|\\(?:\r\n|[\s\S]))*"/,constant:/\b(?:library|use)\b/i,keyword:/\b(?:'active|'ascending|'base|'delayed|'driving|'driving_value|'event|'high|'image|'instance_name|'last_active|'last_event|'last_value|'left|'leftof|'length|'low|'path_name|'pos|'pred|'quiet|'range|'reverse_range|'right|'rightof|'simple_name|'stable|'succ|'transaction|'val|'value|access|after|alias|all|architecture|array|assert|attribute|begin|block|body|buffer|bus|case|component|configuration|constant|disconnect|downto|else|elsif|end|entity|exit|file|for|function|generate|generic|group|guarded|if|impure|in|inertial|inout|is|label|library|linkage|literal|loop|map|new|next|null|of|on|open|others|out|package|port|postponed|procedure|process|pure|range|record|register|reject|report|return|select|severity|shared|signal|subtype|then|to|transport|type|unaffected|units|until|use|variable|wait|when|while|with)\b/i,boolean:/\b(?:false|true)\b/i,function:/\w+(?=\()/,number:/'[01uxzwlh-]'|\b(?:\d+#[\da-f_.]+#|\d[\d_.]*)(?:e[-+]?\d+)?/i,operator:/[<>]=?|:=|[-+*/&=]|\b(?:abs|and|mod|nand|nor|not|or|rem|rol|ror|sla|sll|sra|srl|xnor|xor)\b/i,punctuation:/[{}[\];(),.:]/}}return UGe}var zGe,AVt;function wzr(){if(AVt)return zGe;AVt=1,zGe=n,n.displayName="vim",n.aliases=[];function n(r){r.languages.vim={string:/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\r\n]|'')*'/,comment:/".*/,function:/\b\w+(?=\()/,keyword:/\b(?:N|Next|P|Print|X|XMLent|XMLns|ab|abbreviate|abc|abclear|abo|aboveleft|al|all|ar|arga|argadd|argd|argdelete|argdo|arge|argedit|argg|argglobal|argl|arglocal|args|argu|argument|as|ascii|b|bN|bNext|ba|bad|badd|ball|bd|bdelete|be|bel|belowright|bf|bfirst|bl|blast|bm|bmodified|bn|bnext|bo|botright|bp|bprevious|br|brea|break|breaka|breakadd|breakd|breakdel|breakl|breaklist|brewind|bro|browse|bufdo|buffer|buffers|bun|bunload|bw|bwipeout|c|cN|cNext|cNfcNfile|ca|cabbrev|cabc|cabclear|cad|caddb|caddbuffer|caddexpr|caddf|caddfile|cal|call|cat|catch|cb|cbuffer|cc|ccl|cclose|cd|ce|center|cex|cexpr|cf|cfile|cfir|cfirst|cg|cgetb|cgetbuffer|cgete|cgetexpr|cgetfile|change|changes|chd|chdir|che|checkpath|checkt|checktime|cl|cla|clast|clist|clo|close|cmapc|cmapclear|cn|cnew|cnewer|cnext|cnf|cnfile|cnorea|cnoreabbrev|co|col|colder|colo|colorscheme|comc|comclear|comp|compiler|con|conf|confirm|continue|cope|copen|copy|cp|cpf|cpfile|cprevious|cq|cquit|cr|crewind|cu|cuna|cunabbrev|cunmap|cw|cwindow|d|debugg|debuggreedy|delc|delcommand|delete|delf|delfunction|delm|delmarks|di|diffg|diffget|diffoff|diffpatch|diffpu|diffput|diffsplit|diffthis|diffu|diffupdate|dig|digraphs|display|dj|djump|dl|dlist|dr|drop|ds|dsearch|dsp|dsplit|e|earlier|echoe|echoerr|echom|echomsg|echon|edit|el|else|elsei|elseif|em|emenu|en|endf|endfo|endfor|endfun|endfunction|endif|endt|endtry|endw|endwhile|ene|enew|ex|exi|exit|exu|exusage|f|file|files|filetype|fin|fina|finally|find|fini|finish|fir|first|fix|fixdel|fo|fold|foldc|foldclose|foldd|folddoc|folddoclosed|folddoopen|foldo|foldopen|for|fu|fun|function|go|goto|gr|grep|grepa|grepadd|h|ha|hardcopy|help|helpf|helpfind|helpg|helpgrep|helpt|helptags|hid|hide|his|history|ia|iabbrev|iabc|iabclear|if|ij|ijump|il|ilist|imapc|imapclear|in|inorea|inoreabbrev|isearch|isp|isplit|iu|iuna|iunabbrev|iunmap|j|join|ju|jumps|k|kee|keepalt|keepj|keepjumps|keepmarks|l|lN|lNext|lNf|lNfile|la|lad|laddb|laddbuffer|laddexpr|laddf|laddfile|lan|language|last|later|lb|lbuffer|lc|lcd|lch|lchdir|lcl|lclose|left|lefta|leftabove|let|lex|lexpr|lf|lfile|lfir|lfirst|lg|lgetb|lgetbuffer|lgete|lgetexpr|lgetfile|lgr|lgrep|lgrepa|lgrepadd|lh|lhelpgrep|list|ll|lla|llast|lli|llist|lm|lmak|lmake|lmap|lmapc|lmapclear|ln|lne|lnew|lnewer|lnext|lnf|lnfile|lnoremap|lo|loadview|loc|lockmarks|lockv|lockvar|lol|lolder|lop|lopen|lp|lpf|lpfile|lprevious|lr|lrewind|ls|lt|ltag|lu|lunmap|lv|lvimgrep|lvimgrepa|lvimgrepadd|lw|lwindow|m|ma|mak|make|mark|marks|mat|match|menut|menutranslate|mk|mkexrc|mks|mksession|mksp|mkspell|mkv|mkvie|mkview|mkvimrc|mod|mode|move|mz|mzf|mzfile|mzscheme|n|nbkey|new|next|nmapc|nmapclear|noh|nohlsearch|norea|noreabbrev|nu|number|nun|nunmap|o|omapc|omapclear|on|only|open|opt|options|ou|ounmap|p|pc|pclose|pe|ped|pedit|perl|perld|perldo|po|pop|popu|popup|pp|ppop|pre|preserve|prev|previous|print|prof|profd|profdel|profile|promptf|promptfind|promptr|promptrepl|ps|psearch|ptN|ptNext|pta|ptag|ptf|ptfirst|ptj|ptjump|ptl|ptlast|ptn|ptnext|ptp|ptprevious|ptr|ptrewind|pts|ptselect|pu|put|pw|pwd|py|pyf|pyfile|python|q|qa|qall|quit|quita|quitall|r|read|rec|recover|red|redi|redir|redo|redr|redraw|redraws|redrawstatus|reg|registers|res|resize|ret|retab|retu|return|rew|rewind|ri|right|rightb|rightbelow|ru|rub|ruby|rubyd|rubydo|rubyf|rubyfile|runtime|rv|rviminfo|sN|sNext|sa|sal|sall|san|sandbox|sargument|sav|saveas|sb|sbN|sbNext|sba|sball|sbf|sbfirst|sbl|sblast|sbm|sbmodified|sbn|sbnext|sbp|sbprevious|sbr|sbrewind|sbuffer|scrip|scripte|scriptencoding|scriptnames|se|set|setf|setfiletype|setg|setglobal|setl|setlocal|sf|sfind|sfir|sfirst|sh|shell|sign|sil|silent|sim|simalt|sl|sla|slast|sleep|sm|smagic|smap|smapc|smapclear|sme|smenu|sn|snext|sni|sniff|sno|snomagic|snor|snoremap|snoreme|snoremenu|so|sor|sort|source|sp|spe|spelld|spelldump|spellgood|spelli|spellinfo|spellr|spellrepall|spellu|spellundo|spellw|spellwrong|split|spr|sprevious|sre|srewind|st|sta|stag|star|startg|startgreplace|startinsert|startr|startreplace|stj|stjump|stop|stopi|stopinsert|sts|stselect|sun|sunhide|sunm|sunmap|sus|suspend|sv|sview|syncbind|t|tN|tNext|ta|tab|tabN|tabNext|tabc|tabclose|tabd|tabdo|tabe|tabedit|tabf|tabfind|tabfir|tabfirst|tabl|tablast|tabm|tabmove|tabn|tabnew|tabnext|tabo|tabonly|tabp|tabprevious|tabr|tabrewind|tabs|tag|tags|tc|tcl|tcld|tcldo|tclf|tclfile|te|tearoff|tf|tfirst|th|throw|tj|tjump|tl|tlast|tm|tmenu|tn|tnext|to|topleft|tp|tprevious|tr|trewind|try|ts|tselect|tu|tunmenu|u|una|unabbreviate|undo|undoj|undojoin|undol|undolist|unh|unhide|unlet|unlo|unlockvar|unm|unmap|up|update|ve|verb|verbose|version|vert|vertical|vi|vie|view|vim|vimgrep|vimgrepa|vimgrepadd|visual|viu|viusage|vmapc|vmapclear|vne|vnew|vs|vsplit|vu|vunmap|w|wN|wNext|wa|wall|wh|while|win|winc|wincmd|windo|winp|winpos|winsize|wn|wnext|wp|wprevious|wq|wqa|wqall|write|ws|wsverb|wv|wviminfo|x|xa|xall|xit|xm|xmap|xmapc|xmapclear|xme|xmenu|xn|xnoremap|xnoreme|xnoremenu|xu|xunmap|y|yank)\b/,builtin:/\b(?:acd|ai|akm|aleph|allowrevins|altkeymap|ambiwidth|ambw|anti|antialias|arab|arabic|arabicshape|ari|arshape|autochdir|autocmd|autoindent|autoread|autowrite|autowriteall|aw|awa|background|backspace|backup|backupcopy|backupdir|backupext|backupskip|balloondelay|ballooneval|balloonexpr|bdir|bdlay|beval|bex|bexpr|bg|bh|bin|binary|biosk|bioskey|bk|bkc|bomb|breakat|brk|browsedir|bs|bsdir|bsk|bt|bufhidden|buflisted|buftype|casemap|ccv|cdpath|cedit|cfu|ch|charconvert|ci|cin|cindent|cink|cinkeys|cino|cinoptions|cinw|cinwords|clipboard|cmdheight|cmdwinheight|cmp|cms|columns|com|comments|commentstring|compatible|complete|completefunc|completeopt|consk|conskey|copyindent|cot|cpo|cpoptions|cpt|cscopepathcomp|cscopeprg|cscopequickfix|cscopetag|cscopetagorder|cscopeverbose|cspc|csprg|csqf|cst|csto|csverb|cuc|cul|cursorcolumn|cursorline|cwh|debug|deco|def|define|delcombine|dex|dg|dict|dictionary|diff|diffexpr|diffopt|digraph|dip|dir|directory|dy|ea|ead|eadirection|eb|ed|edcompatible|ef|efm|ei|ek|enc|encoding|endofline|eol|ep|equalalways|equalprg|errorbells|errorfile|errorformat|esckeys|et|eventignore|expandtab|exrc|fcl|fcs|fdc|fde|fdi|fdl|fdls|fdm|fdn|fdo|fdt|fen|fenc|fencs|fex|ff|ffs|fileencoding|fileencodings|fileformat|fileformats|fillchars|fk|fkmap|flp|fml|fmr|foldcolumn|foldenable|foldexpr|foldignore|foldlevel|foldlevelstart|foldmarker|foldmethod|foldminlines|foldnestmax|foldtext|formatexpr|formatlistpat|formatoptions|formatprg|fp|fs|fsync|ft|gcr|gd|gdefault|gfm|gfn|gfs|gfw|ghr|gp|grepformat|grepprg|gtl|gtt|guicursor|guifont|guifontset|guifontwide|guiheadroom|guioptions|guipty|guitablabel|guitabtooltip|helpfile|helpheight|helplang|hf|hh|hi|hidden|highlight|hk|hkmap|hkmapp|hkp|hl|hlg|hls|hlsearch|ic|icon|iconstring|ignorecase|im|imactivatekey|imak|imc|imcmdline|imd|imdisable|imi|iminsert|ims|imsearch|inc|include|includeexpr|incsearch|inde|indentexpr|indentkeys|indk|inex|inf|infercase|insertmode|invacd|invai|invakm|invallowrevins|invaltkeymap|invanti|invantialias|invar|invarab|invarabic|invarabicshape|invari|invarshape|invautochdir|invautoindent|invautoread|invautowrite|invautowriteall|invaw|invawa|invbackup|invballooneval|invbeval|invbin|invbinary|invbiosk|invbioskey|invbk|invbl|invbomb|invbuflisted|invcf|invci|invcin|invcindent|invcompatible|invconfirm|invconsk|invconskey|invcopyindent|invcp|invcscopetag|invcscopeverbose|invcst|invcsverb|invcuc|invcul|invcursorcolumn|invcursorline|invdeco|invdelcombine|invdg|invdiff|invdigraph|invdisable|invea|inveb|inved|invedcompatible|invek|invendofline|inveol|invequalalways|inverrorbells|invesckeys|invet|invex|invexpandtab|invexrc|invfen|invfk|invfkmap|invfoldenable|invgd|invgdefault|invguipty|invhid|invhidden|invhk|invhkmap|invhkmapp|invhkp|invhls|invhlsearch|invic|invicon|invignorecase|invim|invimc|invimcmdline|invimd|invincsearch|invinf|invinfercase|invinsertmode|invis|invjoinspaces|invjs|invlazyredraw|invlbr|invlinebreak|invlisp|invlist|invloadplugins|invlpl|invlz|invma|invmacatsui|invmagic|invmh|invml|invmod|invmodeline|invmodifiable|invmodified|invmore|invmousef|invmousefocus|invmousehide|invnu|invnumber|invodev|invopendevice|invpaste|invpi|invpreserveindent|invpreviewwindow|invprompt|invpvw|invreadonly|invremap|invrestorescreen|invrevins|invri|invrightleft|invrightleftcmd|invrl|invrlc|invro|invrs|invru|invruler|invsb|invsc|invscb|invscrollbind|invscs|invsecure|invsft|invshellslash|invshelltemp|invshiftround|invshortname|invshowcmd|invshowfulltag|invshowmatch|invshowmode|invsi|invsm|invsmartcase|invsmartindent|invsmarttab|invsmd|invsn|invsol|invspell|invsplitbelow|invsplitright|invspr|invsr|invssl|invsta|invstartofline|invstmp|invswapfile|invswf|invta|invtagbsearch|invtagrelative|invtagstack|invtbi|invtbidi|invtbs|invtermbidi|invterse|invtextauto|invtextmode|invtf|invtgst|invtildeop|invtimeout|invtitle|invto|invtop|invtr|invttimeout|invttybuiltin|invttyfast|invtx|invvb|invvisualbell|invwa|invwarn|invwb|invweirdinvert|invwfh|invwfw|invwildmenu|invwinfixheight|invwinfixwidth|invwiv|invwmnu|invwrap|invwrapscan|invwrite|invwriteany|invwritebackup|invws|isf|isfname|isi|isident|isk|iskeyword|isprint|joinspaces|js|key|keymap|keymodel|keywordprg|km|kmp|kp|langmap|langmenu|laststatus|lazyredraw|lbr|lcs|linebreak|lines|linespace|lisp|lispwords|listchars|loadplugins|lpl|lsp|lz|macatsui|magic|makeef|makeprg|matchpairs|matchtime|maxcombine|maxfuncdepth|maxmapdepth|maxmem|maxmempattern|maxmemtot|mco|mef|menuitems|mfd|mh|mis|mkspellmem|ml|mls|mm|mmd|mmp|mmt|modeline|modelines|modifiable|modified|more|mouse|mousef|mousefocus|mousehide|mousem|mousemodel|mouses|mouseshape|mouset|mousetime|mp|mps|msm|mzq|mzquantum|nf|noacd|noai|noakm|noallowrevins|noaltkeymap|noanti|noantialias|noar|noarab|noarabic|noarabicshape|noari|noarshape|noautochdir|noautoindent|noautoread|noautowrite|noautowriteall|noaw|noawa|nobackup|noballooneval|nobeval|nobin|nobinary|nobiosk|nobioskey|nobk|nobl|nobomb|nobuflisted|nocf|noci|nocin|nocindent|nocompatible|noconfirm|noconsk|noconskey|nocopyindent|nocp|nocscopetag|nocscopeverbose|nocst|nocsverb|nocuc|nocul|nocursorcolumn|nocursorline|nodeco|nodelcombine|nodg|nodiff|nodigraph|nodisable|noea|noeb|noed|noedcompatible|noek|noendofline|noeol|noequalalways|noerrorbells|noesckeys|noet|noex|noexpandtab|noexrc|nofen|nofk|nofkmap|nofoldenable|nogd|nogdefault|noguipty|nohid|nohidden|nohk|nohkmap|nohkmapp|nohkp|nohls|noic|noicon|noignorecase|noim|noimc|noimcmdline|noimd|noincsearch|noinf|noinfercase|noinsertmode|nois|nojoinspaces|nojs|nolazyredraw|nolbr|nolinebreak|nolisp|nolist|noloadplugins|nolpl|nolz|noma|nomacatsui|nomagic|nomh|noml|nomod|nomodeline|nomodifiable|nomodified|nomore|nomousef|nomousefocus|nomousehide|nonu|nonumber|noodev|noopendevice|nopaste|nopi|nopreserveindent|nopreviewwindow|noprompt|nopvw|noreadonly|noremap|norestorescreen|norevins|nori|norightleft|norightleftcmd|norl|norlc|noro|nors|noru|noruler|nosb|nosc|noscb|noscrollbind|noscs|nosecure|nosft|noshellslash|noshelltemp|noshiftround|noshortname|noshowcmd|noshowfulltag|noshowmatch|noshowmode|nosi|nosm|nosmartcase|nosmartindent|nosmarttab|nosmd|nosn|nosol|nospell|nosplitbelow|nosplitright|nospr|nosr|nossl|nosta|nostartofline|nostmp|noswapfile|noswf|nota|notagbsearch|notagrelative|notagstack|notbi|notbidi|notbs|notermbidi|noterse|notextauto|notextmode|notf|notgst|notildeop|notimeout|notitle|noto|notop|notr|nottimeout|nottybuiltin|nottyfast|notx|novb|novisualbell|nowa|nowarn|nowb|noweirdinvert|nowfh|nowfw|nowildmenu|nowinfixheight|nowinfixwidth|nowiv|nowmnu|nowrap|nowrapscan|nowrite|nowriteany|nowritebackup|nows|nrformats|numberwidth|nuw|odev|oft|ofu|omnifunc|opendevice|operatorfunc|opfunc|osfiletype|pa|para|paragraphs|paste|pastetoggle|patchexpr|patchmode|path|pdev|penc|pex|pexpr|pfn|ph|pheader|pi|pm|pmbcs|pmbfn|popt|preserveindent|previewheight|previewwindow|printdevice|printencoding|printexpr|printfont|printheader|printmbcharset|printmbfont|printoptions|prompt|pt|pumheight|pvh|pvw|qe|quoteescape|readonly|remap|report|restorescreen|revins|rightleft|rightleftcmd|rl|rlc|ro|rs|rtp|ruf|ruler|rulerformat|runtimepath|sbo|sc|scb|scr|scroll|scrollbind|scrolljump|scrolloff|scrollopt|scs|sect|sections|secure|sel|selection|selectmode|sessionoptions|sft|shcf|shellcmdflag|shellpipe|shellquote|shellredir|shellslash|shelltemp|shelltype|shellxquote|shiftround|shiftwidth|shm|shortmess|shortname|showbreak|showcmd|showfulltag|showmatch|showmode|showtabline|shq|si|sidescroll|sidescrolloff|siso|sj|slm|smartcase|smartindent|smarttab|smc|smd|softtabstop|sol|spc|spell|spellcapcheck|spellfile|spelllang|spellsuggest|spf|spl|splitbelow|splitright|sps|sr|srr|ss|ssl|ssop|stal|startofline|statusline|stl|stmp|su|sua|suffixes|suffixesadd|sw|swapfile|swapsync|swb|swf|switchbuf|sws|sxq|syn|synmaxcol|syntax|t_AB|t_AF|t_AL|t_CS|t_CV|t_Ce|t_Co|t_Cs|t_DL|t_EI|t_F1|t_F2|t_F3|t_F4|t_F5|t_F6|t_F7|t_F8|t_F9|t_IE|t_IS|t_K1|t_K3|t_K4|t_K5|t_K6|t_K7|t_K8|t_K9|t_KA|t_KB|t_KC|t_KD|t_KE|t_KF|t_KG|t_KH|t_KI|t_KJ|t_KK|t_KL|t_RI|t_RV|t_SI|t_Sb|t_Sf|t_WP|t_WS|t_ZH|t_ZR|t_al|t_bc|t_cd|t_ce|t_cl|t_cm|t_cs|t_da|t_db|t_dl|t_fs|t_k1|t_k2|t_k3|t_k4|t_k5|t_k6|t_k7|t_k8|t_k9|t_kB|t_kD|t_kI|t_kN|t_kP|t_kb|t_kd|t_ke|t_kh|t_kl|t_kr|t_ks|t_ku|t_le|t_mb|t_md|t_me|t_mr|t_ms|t_nd|t_op|t_se|t_so|t_sr|t_te|t_ti|t_ts|t_ue|t_us|t_ut|t_vb|t_ve|t_vi|t_vs|t_xs|tabline|tabpagemax|tabstop|tagbsearch|taglength|tagrelative|tagstack|tal|tb|tbi|tbidi|tbis|tbs|tenc|term|termbidi|termencoding|terse|textauto|textmode|textwidth|tgst|thesaurus|tildeop|timeout|timeoutlen|title|titlelen|titleold|titlestring|toolbar|toolbariconsize|top|tpm|tsl|tsr|ttimeout|ttimeoutlen|ttm|tty|ttybuiltin|ttyfast|ttym|ttymouse|ttyscroll|ttytype|tw|tx|uc|ul|undolevels|updatecount|updatetime|ut|vb|vbs|vdir|verbosefile|vfile|viewdir|viewoptions|viminfo|virtualedit|visualbell|vop|wak|warn|wb|wc|wcm|wd|weirdinvert|wfh|wfw|whichwrap|wi|wig|wildchar|wildcharm|wildignore|wildmenu|wildmode|wildoptions|wim|winaltkeys|window|winfixheight|winfixwidth|winheight|winminheight|winminwidth|winwidth|wiv|wiw|wm|wmh|wmnu|wmw|wop|wrap|wrapmargin|wrapscan|writeany|writebackup|writedelay|ww)\b/,number:/\b(?:0x[\da-f]+|\d+(?:\.\d+)?)\b/i,operator:/\|\||&&|[-+.]=?|[=!](?:[=~][#?]?)?|[<>]=?[#?]?|[*\/%?]|\b(?:is(?:not)?)\b/,punctuation:/[{}[\](),;:]/}}return zGe}var HGe,IVt;function yzr(){if(IVt)return HGe;IVt=1,HGe=n,n.displayName="visualBasic",n.aliases=[];function n(r){r.languages["visual-basic"]={comment:{pattern:/(?:['‘’]|REM\b)(?:[^\r\n_]|_(?:\r\n?|\n)?)*/i,inside:{keyword:/^REM/i}},directive:{pattern:/#(?:Const|Else|ElseIf|End|ExternalChecksum|ExternalSource|If|Region)(?:\b_[ \t]*(?:\r\n?|\n)|.)+/i,alias:"property",greedy:!0},string:{pattern:/\$?["“”](?:["“”]{2}|[^"“”])*["“”]C?/i,greedy:!0},date:{pattern:/#[ \t]*(?:\d+([/-])\d+\1\d+(?:[ \t]+(?:\d+[ \t]*(?:AM|PM)|\d+:\d+(?::\d+)?(?:[ \t]*(?:AM|PM))?))?|\d+[ \t]*(?:AM|PM)|\d+:\d+(?::\d+)?(?:[ \t]*(?:AM|PM))?)[ \t]*#/i,alias:"number"},number:/(?:(?:\b\d+(?:\.\d+)?|\.\d+)(?:E[+-]?\d+)?|&[HO][\dA-F]+)(?:[FRD]|U?[ILS])?/i,boolean:/\b(?:False|Nothing|True)\b/i,keyword:/\b(?:AddHandler|AddressOf|Alias|And(?:Also)?|As|Boolean|ByRef|Byte|ByVal|Call|Case|Catch|C(?:Bool|Byte|Char|Date|Dbl|Dec|Int|Lng|Obj|SByte|Short|Sng|Str|Type|UInt|ULng|UShort)|Char|Class|Const|Continue|Currency|Date|Decimal|Declare|Default|Delegate|Dim|DirectCast|Do|Double|Each|Else(?:If)?|End(?:If)?|Enum|Erase|Error|Event|Exit|Finally|For|Friend|Function|Get(?:Type|XMLNamespace)?|Global|GoSub|GoTo|Handles|If|Implements|Imports|In|Inherits|Integer|Interface|Is|IsNot|Let|Lib|Like|Long|Loop|Me|Mod|Module|Must(?:Inherit|Override)|My(?:Base|Class)|Namespace|Narrowing|New|Next|Not(?:Inheritable|Overridable)?|Object|Of|On|Operator|Option(?:al)?|Or(?:Else)?|Out|Overloads|Overridable|Overrides|ParamArray|Partial|Private|Property|Protected|Public|RaiseEvent|ReadOnly|ReDim|RemoveHandler|Resume|Return|SByte|Select|Set|Shadows|Shared|short|Single|Static|Step|Stop|String|Structure|Sub|SyncLock|Then|Throw|To|Try|TryCast|Type|TypeOf|U(?:Integer|Long|Short)|Until|Using|Variant|Wend|When|While|Widening|With(?:Events)?|WriteOnly|Xor)\b/i,operator:/[+\-*/\\^<=>&#@$%!]|\b_(?=[ \t]*[\r\n])/,punctuation:/[{}().,:?]/},r.languages.vb=r.languages["visual-basic"],r.languages.vba=r.languages["visual-basic"]}return HGe}var GGe,OVt;function Ezr(){if(OVt)return GGe;OVt=1,GGe=n,n.displayName="warpscript",n.aliases=[];function n(r){r.languages.warpscript={comment:/#.*|\/\/.*|\/\*[\s\S]*?\*\//,string:{pattern:/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\\\r\n]|\\.)*'|<'(?:[^\\']|'(?!>)|\\.)*'>/,greedy:!0},variable:/\$\S+/,macro:{pattern:/@\S+/,alias:"property"},keyword:/\b(?:BREAK|CHECKMACRO|CONTINUE|CUDF|DEFINED|DEFINEDMACRO|EVAL|FAIL|FOR|FOREACH|FORSTEP|IFT|IFTE|MSGFAIL|NRETURN|RETHROW|RETURN|SWITCH|TRY|UDF|UNTIL|WHILE)\b/,number:/[+-]?\b(?:NaN|Infinity|\d+(?:\.\d*)?(?:[Ee][+-]?\d+)?|0x[\da-fA-F]+|0b[01]+)\b/,boolean:/\b(?:F|T|false|true)\b/,punctuation:/<%|%>|[{}[\]()]/,operator:/==|&&?|\|\|?|\*\*?|>>>?|<<|[<>!~]=?|[-/%^]|\+!?|\b(?:AND|NOT|OR)\b/}}return GGe}var qGe,NVt;function Szr(){if(NVt)return qGe;NVt=1,qGe=n,n.displayName="wasm",n.aliases=[];function n(r){r.languages.wasm={comment:[/\(;[\s\S]*?;\)/,{pattern:/;;.*/,greedy:!0}],string:{pattern:/"(?:\\[\s\S]|[^"\\])*"/,greedy:!0},keyword:[{pattern:/\b(?:align|offset)=/,inside:{operator:/=/}},{pattern:/\b(?:(?:f32|f64|i32|i64)(?:\.(?:abs|add|and|ceil|clz|const|convert_[su]\/i(?:32|64)|copysign|ctz|demote\/f64|div(?:_[su])?|eqz?|extend_[su]\/i32|floor|ge(?:_[su])?|gt(?:_[su])?|le(?:_[su])?|load(?:(?:8|16|32)_[su])?|lt(?:_[su])?|max|min|mul|neg?|nearest|or|popcnt|promote\/f32|reinterpret\/[fi](?:32|64)|rem_[su]|rot[lr]|shl|shr_[su]|sqrt|store(?:8|16|32)?|sub|trunc(?:_[su]\/f(?:32|64))?|wrap\/i64|xor))?|memory\.(?:grow|size))\b/,inside:{punctuation:/\./}},/\b(?:anyfunc|block|br(?:_if|_table)?|call(?:_indirect)?|data|drop|elem|else|end|export|func|get_(?:global|local)|global|if|import|local|loop|memory|module|mut|nop|offset|param|result|return|select|set_(?:global|local)|start|table|tee_local|then|type|unreachable)\b/],variable:/\$[\w!#$%&'*+\-./:<=>?@\\^`|~]+/,number:/[+-]?\b(?:\d(?:_?\d)*(?:\.\d(?:_?\d)*)?(?:[eE][+-]?\d(?:_?\d)*)?|0x[\da-fA-F](?:_?[\da-fA-F])*(?:\.[\da-fA-F](?:_?[\da-fA-D])*)?(?:[pP][+-]?\d(?:_?\d)*)?)\b|\binf\b|\bnan(?::0x[\da-fA-F](?:_?[\da-fA-D])*)?\b/,punctuation:/[()]/}}return qGe}var VGe,RVt;function _zr(){if(RVt)return VGe;RVt=1,VGe=n,n.displayName="webIdl",n.aliases=[];function n(r){(function(o){var s=/(?:\B-|\b_|\b)[A-Za-z][\w-]*(?![\w-])/.source,c="(?:"+/\b(?:unsigned\s+)?long\s+long(?![\w-])/.source+"|"+/\b(?:unrestricted|unsigned)\s+[a-z]+(?![\w-])/.source+"|"+/(?!(?:unrestricted|unsigned)\b)/.source+s+/(?:\s*<(?:[^<>]|<[^<>]*>)*>)?/.source+")"+/(?:\s*\?)?/.source,d={};o.languages["web-idl"]={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/"[^"]*"/,greedy:!0},namespace:{pattern:RegExp(/(\bnamespace\s+)/.source+s),lookbehind:!0},"class-name":[{pattern:/(^|[^\w-])(?:iterable|maplike|setlike)\s*<(?:[^<>]|<[^<>]*>)*>/,lookbehind:!0,inside:d},{pattern:RegExp(/(\b(?:attribute|const|deleter|getter|optional|setter)\s+)/.source+c),lookbehind:!0,inside:d},{pattern:RegExp("("+/\bcallback\s+/.source+s+/\s*=\s*/.source+")"+c),lookbehind:!0,inside:d},{pattern:RegExp(/(\btypedef\b\s*)/.source+c),lookbehind:!0,inside:d},{pattern:RegExp(/(\b(?:callback|dictionary|enum|interface(?:\s+mixin)?)\s+)(?!(?:interface|mixin)\b)/.source+s),lookbehind:!0},{pattern:RegExp(/(:\s*)/.source+s),lookbehind:!0},RegExp(s+/(?=\s+(?:implements|includes)\b)/.source),{pattern:RegExp(/(\b(?:implements|includes)\s+)/.source+s),lookbehind:!0},{pattern:RegExp(c+"(?="+/\s*(?:\.{3}\s*)?/.source+s+/\s*[(),;=]/.source+")"),inside:d}],builtin:/\b(?:ArrayBuffer|BigInt64Array|BigUint64Array|ByteString|DOMString|DataView|Float32Array|Float64Array|FrozenArray|Int16Array|Int32Array|Int8Array|ObservableArray|Promise|USVString|Uint16Array|Uint32Array|Uint8Array|Uint8ClampedArray)\b/,keyword:[/\b(?:async|attribute|callback|const|constructor|deleter|dictionary|enum|getter|implements|includes|inherit|interface|mixin|namespace|null|optional|or|partial|readonly|required|setter|static|stringifier|typedef|unrestricted)\b/,/\b(?:any|bigint|boolean|byte|double|float|iterable|long|maplike|object|octet|record|sequence|setlike|short|symbol|undefined|unsigned|void)\b/],boolean:/\b(?:false|true)\b/,number:{pattern:/(^|[^\w-])-?(?:0x[0-9a-f]+|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|NaN|Infinity)(?![\w-])/i,lookbehind:!0},operator:/\.{3}|[=:?<>-]/,punctuation:/[(){}[\].,;]/};for(var h in o.languages["web-idl"])h!=="class-name"&&(d[h]=o.languages["web-idl"][h]);o.languages.webidl=o.languages["web-idl"]})(r)}return VGe}var WGe,PVt;function kzr(){if(PVt)return WGe;PVt=1,WGe=n,n.displayName="wiki",n.aliases=[];function n(r){r.languages.wiki=r.languages.extend("markup",{"block-comment":{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0,alias:"comment"},heading:{pattern:/^(=+)[^=\r\n].*?\1/m,inside:{punctuation:/^=+|=+$/,important:/.+/}},emphasis:{pattern:/('{2,5}).+?\1/,inside:{"bold-italic":{pattern:/(''''').+?(?=\1)/,lookbehind:!0,alias:["bold","italic"]},bold:{pattern:/(''')[^'](?:.*?[^'])?(?=\1)/,lookbehind:!0},italic:{pattern:/('')[^'](?:.*?[^'])?(?=\1)/,lookbehind:!0},punctuation:/^''+|''+$/}},hr:{pattern:/^-{4,}/m,alias:"punctuation"},url:[/ISBN +(?:97[89][ -]?)?(?:\d[ -]?){9}[\dx]\b|(?:PMID|RFC) +\d+/i,/\[\[.+?\]\]|\[.+?\]/],variable:[/__[A-Z]+__/,/\{{3}.+?\}{3}/,/\{\{.+?\}\}/],symbol:[/^#redirect/im,/~{3,5}/],"table-tag":{pattern:/((?:^|[|!])[|!])[^|\r\n]+\|(?!\|)/m,lookbehind:!0,inside:{"table-bar":{pattern:/\|$/,alias:"punctuation"},rest:r.languages.markup.tag.inside}},punctuation:/^(?:\{\||\|\}|\|-|[*#:;!|])|\|\||!!/m}),r.languages.insertBefore("wiki","tag",{nowiki:{pattern:/<(nowiki|pre|source)\b[^>]*>[\s\S]*?<\/\1>/i,inside:{tag:{pattern:/<(?:nowiki|pre|source)\b[^>]*>|<\/(?:nowiki|pre|source)>/i,inside:r.languages.markup.tag.inside}}}})}return WGe}var KGe,LVt;function Tzr(){if(LVt)return KGe;LVt=1,KGe=n,n.displayName="wolfram",n.aliases=["mathematica","wl","nb"];function n(r){r.languages.wolfram={comment:/\(\*(?:\(\*(?:[^*]|\*(?!\)))*\*\)|(?!\(\*)[\s\S])*?\*\)/,string:{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0},keyword:/\b(?:Abs|AbsArg|Accuracy|Block|Do|For|Function|If|Manipulate|Module|Nest|NestList|None|Return|Switch|Table|Which|While)\b/,context:{pattern:/\b\w+`+\w*/,alias:"class-name"},blank:{pattern:/\b\w+_\b/,alias:"regex"},"global-variable":{pattern:/\$\w+/,alias:"variable"},boolean:/\b(?:False|True)\b/,number:/(?:\b(?=\d)|\B(?=\.))(?:0[bo])?(?:(?:\d|0x[\da-f])[\da-f]*(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?j?\b/i,operator:/\/\.|;|=\.|\^=|\^:=|:=|<<|>>|<\||\|>|:>|\|->|->|<-|@@@|@@|@|\/@|=!=|===|==|=|\+|-|\^|\[\/-+%=\]=?|!=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]/,punctuation:/[{}[\];(),.:]/},r.languages.mathematica=r.languages.wolfram,r.languages.wl=r.languages.wolfram,r.languages.nb=r.languages.wolfram}return KGe}var YGe,DVt;function xzr(){if(DVt)return YGe;DVt=1,YGe=n,n.displayName="wren",n.aliases=[];function n(r){r.languages.wren={comment:[{pattern:/\/\*(?:[^*/]|\*(?!\/)|\/(?!\*)|\/\*(?:[^*/]|\*(?!\/)|\/(?!\*)|\/\*(?:[^*/]|\*(?!\/)|\/(?!\*))*\*\/)*\*\/)*\*\//,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],"triple-quoted-string":{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string"},"string-literal":null,hashbang:{pattern:/^#!\/.+/,greedy:!0,alias:"comment"},attribute:{pattern:/#!?[ \t\u3000]*\w+/,alias:"keyword"},"class-name":[{pattern:/(\bclass\s+)\w+/,lookbehind:!0},/\b[A-Z][a-z\d_]*\b/],constant:/\b[A-Z][A-Z\d_]*\b/,null:{pattern:/\bnull\b/,alias:"keyword"},keyword:/\b(?:as|break|class|construct|continue|else|for|foreign|if|import|in|is|return|static|super|this|var|while)\b/,boolean:/\b(?:false|true)\b/,number:/\b(?:0x[\da-f]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/i,function:/\b[a-z_]\w*(?=\s*[({])/i,operator:/<<|>>|[=!<>]=?|&&|\|\||[-+*/%~^&|?:]|\.{2,3}/,punctuation:/[\[\](){}.,;]/},r.languages.wren["string-literal"]={pattern:/(^|[^\\"])"(?:[^\\"%]|\\[\s\S]|%(?!\()|%\((?:[^()]|\((?:[^()]|\([^)]*\))*\))*\))*"/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)%\((?:[^()]|\((?:[^()]|\([^)]*\))*\))*\)/,lookbehind:!0,inside:{expression:{pattern:/^(%\()[\s\S]+(?=\)$)/,lookbehind:!0,inside:r.languages.wren},"interpolation-punctuation":{pattern:/^%\(|\)$/,alias:"punctuation"}}},string:/[\s\S]+/}}}return YGe}var XGe,MVt;function Czr(){if(MVt)return XGe;MVt=1,XGe=n,n.displayName="xeora",n.aliases=["xeoracube"];function n(r){(function(o){o.languages.xeora=o.languages.extend("markup",{constant:{pattern:/\$(?:DomainContents|PageRenderDuration)\$/,inside:{punctuation:{pattern:/\$/}}},variable:{pattern:/\$@?(?:#+|[-+*~=^])?[\w.]+\$/,inside:{punctuation:{pattern:/[$.]/},operator:{pattern:/#+|[-+*~=^@]/}}},"function-inline":{pattern:/\$F:[-\w.]+\?[-\w.]+(?:,(?:(?:@[-#]*\w+\.[\w+.]\.*)*\|)*(?:(?:[\w+]|[-#*.~^]+[\w+]|=\S)(?:[^$=]|=+[^=])*=*|(?:@[-#]*\w+\.[\w+.]\.*)+(?:(?:[\w+]|[-#*~^][-#*.~^]*[\w+]|=\S)(?:[^$=]|=+[^=])*=*)?)?)?\$/,inside:{variable:{pattern:/(?:[,|])@?(?:#+|[-+*~=^])?[\w.]+/,inside:{punctuation:{pattern:/[,.|]/},operator:{pattern:/#+|[-+*~=^@]/}}},punctuation:{pattern:/\$\w:|[$:?.,|]/}},alias:"function"},"function-block":{pattern:/\$XF:\{[-\w.]+\?[-\w.]+(?:,(?:(?:@[-#]*\w+\.[\w+.]\.*)*\|)*(?:(?:[\w+]|[-#*.~^]+[\w+]|=\S)(?:[^$=]|=+[^=])*=*|(?:@[-#]*\w+\.[\w+.]\.*)+(?:(?:[\w+]|[-#*~^][-#*.~^]*[\w+]|=\S)(?:[^$=]|=+[^=])*=*)?)?)?\}:XF\$/,inside:{punctuation:{pattern:/[$:{}?.,|]/}},alias:"function"},"directive-inline":{pattern:/\$\w(?:#\d+\+?)?(?:\[[-\w.]+\])?:[-\/\w.]+\$/,inside:{punctuation:{pattern:/\$(?:\w:|C(?:\[|#\d))?|[:{[\]]/,inside:{tag:{pattern:/#\d/}}}},alias:"function"},"directive-block-open":{pattern:/\$\w+:\{|\$\w(?:#\d+\+?)?(?:\[[-\w.]+\])?:[-\w.]+:\{(?:![A-Z]+)?/,inside:{punctuation:{pattern:/\$(?:\w:|C(?:\[|#\d))?|[:{[\]]/,inside:{tag:{pattern:/#\d/}}},attribute:{pattern:/![A-Z]+$/,inside:{punctuation:{pattern:/!/}},alias:"keyword"}},alias:"function"},"directive-block-separator":{pattern:/\}:[-\w.]+:\{/,inside:{punctuation:{pattern:/[:{}]/}},alias:"function"},"directive-block-close":{pattern:/\}:[-\w.]+\$/,inside:{punctuation:{pattern:/[:{}$]/}},alias:"function"}}),o.languages.insertBefore("inside","punctuation",{variable:o.languages.xeora["function-inline"].inside.variable},o.languages.xeora["function-block"]),o.languages.xeoracube=o.languages.xeora})(r)}return XGe}var JGe,jVt;function Azr(){if(jVt)return JGe;jVt=1,JGe=n,n.displayName="xmlDoc",n.aliases=[];function n(r){(function(o){function s(b,y){o.languages[b]&&o.languages.insertBefore(b,"comment",{"doc-comment":y})}var c=o.languages.markup.tag,d={pattern:/\/\/\/.*/,greedy:!0,alias:"comment",inside:{tag:c}},h={pattern:/'''.*/,greedy:!0,alias:"comment",inside:{tag:c}};s("csharp",d),s("fsharp",d),s("vbnet",h)})(r)}return JGe}var ZGe,$Vt;function Izr(){if($Vt)return ZGe;$Vt=1,ZGe=n,n.displayName="xojo",n.aliases=[];function n(r){r.languages.xojo={comment:{pattern:/(?:'|\/\/|Rem\b).+/i,greedy:!0},string:{pattern:/"(?:""|[^"])*"/,greedy:!0},number:[/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,/&[bchou][a-z\d]+/i],directive:{pattern:/#(?:Else|ElseIf|Endif|If|Pragma)\b/i,alias:"property"},keyword:/\b(?:AddHandler|App|Array|As(?:signs)?|Auto|Boolean|Break|By(?:Ref|Val)|Byte|Call|Case|Catch|CFStringRef|CGFloat|Class|Color|Const|Continue|CString|Currency|CurrentMethodName|Declare|Delegate|Dim|Do(?:uble|wnTo)?|Each|Else(?:If)?|End|Enumeration|Event|Exception|Exit|Extends|False|Finally|For|Function|Get|GetTypeInfo|Global|GOTO|If|Implements|In|Inherits|Int(?:8|16|32|64|eger|erface)?|Lib|Loop|Me|Module|Next|Nil|Object|Optional|OSType|ParamArray|Private|Property|Protected|PString|Ptr|Raise(?:Event)?|ReDim|RemoveHandler|Return|Select(?:or)?|Self|Set|Shared|Short|Single|Soft|Static|Step|String|Sub|Super|Text|Then|To|True|Try|Ubound|UInt(?:8|16|32|64|eger)?|Until|Using|Var(?:iant)?|Wend|While|WindowPtr|WString)\b/i,operator:/<[=>]?|>=?|[+\-*\/\\^=]|\b(?:AddressOf|And|Ctype|IsA?|Mod|New|Not|Or|WeakAddressOf|Xor)\b/i,punctuation:/[.,;:()]/}}return ZGe}var QGe,FVt;function Ozr(){if(FVt)return QGe;FVt=1,QGe=n,n.displayName="xquery",n.aliases=[];function n(r){(function(o){o.languages.xquery=o.languages.extend("markup",{"xquery-comment":{pattern:/\(:[\s\S]*?:\)/,greedy:!0,alias:"comment"},string:{pattern:/(["'])(?:\1\1|(?!\1)[\s\S])*\1/,greedy:!0},extension:{pattern:/\(#.+?#\)/,alias:"symbol"},variable:/\$[-\w:]+/,axis:{pattern:/(^|[^-])(?:ancestor(?:-or-self)?|attribute|child|descendant(?:-or-self)?|following(?:-sibling)?|parent|preceding(?:-sibling)?|self)(?=::)/,lookbehind:!0,alias:"operator"},"keyword-operator":{pattern:/(^|[^:-])\b(?:and|castable as|div|eq|except|ge|gt|idiv|instance of|intersect|is|le|lt|mod|ne|or|union)\b(?=$|[^:-])/,lookbehind:!0,alias:"operator"},keyword:{pattern:/(^|[^:-])\b(?:as|ascending|at|base-uri|boundary-space|case|cast as|collation|construction|copy-namespaces|declare|default|descending|else|empty (?:greatest|least)|encoding|every|external|for|function|if|import|in|inherit|lax|let|map|module|namespace|no-inherit|no-preserve|option|order(?: by|ed|ing)?|preserve|return|satisfies|schema|some|stable|strict|strip|then|to|treat as|typeswitch|unordered|validate|variable|version|where|xquery)\b(?=$|[^:-])/,lookbehind:!0},function:/[\w-]+(?::[\w-]+)*(?=\s*\()/,"xquery-element":{pattern:/(element\s+)[\w-]+(?::[\w-]+)*/,lookbehind:!0,alias:"tag"},"xquery-attribute":{pattern:/(attribute\s+)[\w-]+(?::[\w-]+)*/,lookbehind:!0,alias:"attr-name"},builtin:{pattern:/(^|[^:-])\b(?:attribute|comment|document|element|processing-instruction|text|xs:(?:ENTITIES|ENTITY|ID|IDREFS?|NCName|NMTOKENS?|NOTATION|Name|QName|anyAtomicType|anyType|anyURI|base64Binary|boolean|byte|date|dateTime|dayTimeDuration|decimal|double|duration|float|gDay|gMonth|gMonthDay|gYear|gYearMonth|hexBinary|int|integer|language|long|negativeInteger|nonNegativeInteger|nonPositiveInteger|normalizedString|positiveInteger|short|string|time|token|unsigned(?:Byte|Int|Long|Short)|untyped(?:Atomic)?|yearMonthDuration))\b(?=$|[^:-])/,lookbehind:!0},number:/\b\d+(?:\.\d+)?(?:E[+-]?\d+)?/,operator:[/[+*=?|@]|\.\.?|:=|!=|<[=<]?|>[=>]?/,{pattern:/(\s)-(?=\s)/,lookbehind:!0}],punctuation:/[[\](){},;:/]/}),o.languages.xquery.tag.pattern=/<\/?(?!\d)[^\s>\/=$<%]+(?:\s+[^\s>\/=]+(?:=(?:("|')(?:\\[\s\S]|\{(?!\{)(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])+\}|(?!\1)[^\\])*\1|[^\s'">=]+))?)*\s*\/?>/,o.languages.xquery.tag.inside["attr-value"].pattern=/=(?:("|')(?:\\[\s\S]|\{(?!\{)(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])+\}|(?!\1)[^\\])*\1|[^\s'">=]+)/,o.languages.xquery.tag.inside["attr-value"].inside.punctuation=/^="|"$/,o.languages.xquery.tag.inside["attr-value"].inside.expression={pattern:/\{(?!\{)(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])+\}/,inside:o.languages.xquery,alias:"language-xquery"};var s=function(d){return typeof d=="string"?d:typeof d.content=="string"?d.content:d.content.map(s).join("")},c=function(d){for(var h=[],b=0;b<d.length;b++){var y=d[b],E=!1;if(typeof y!="string"&&(y.type==="tag"&&y.content[0]&&y.content[0].type==="tag"?y.content[0].content[0].content==="</"?h.length>0&&h[h.length-1].tagName===s(y.content[0].content[1])&&h.pop():y.content[y.content.length-1].content==="/>"||h.push({tagName:s(y.content[0].content[1]),openedBraces:0}):h.length>0&&y.type==="punctuation"&&y.content==="{"&&(!d[b+1]||d[b+1].type!=="punctuation"||d[b+1].content!=="{")&&(!d[b-1]||d[b-1].type!=="plain-text"||d[b-1].content!=="{")?h[h.length-1].openedBraces++:h.length>0&&h[h.length-1].openedBraces>0&&y.type==="punctuation"&&y.content==="}"?h[h.length-1].openedBraces--:y.type!=="comment"&&(E=!0)),(E||typeof y=="string")&&h.length>0&&h[h.length-1].openedBraces===0){var k=s(y);b<d.length-1&&(typeof d[b+1]=="string"||d[b+1].type==="plain-text")&&(k+=s(d[b+1]),d.splice(b+1,1)),b>0&&(typeof d[b-1]=="string"||d[b-1].type==="plain-text")&&(k=s(d[b-1])+k,d.splice(b-1,1),b--),/^\s+$/.test(k)?d[b]=k:d[b]=new o.Token("plain-text",k,null,k)}y.content&&typeof y.content!="string"&&c(y.content)}};o.hooks.add("after-tokenize",function(d){d.language==="xquery"&&c(d.tokens)})})(r)}return QGe}var eqe,BVt;function Nzr(){if(BVt)return eqe;BVt=1,eqe=n,n.displayName="yang",n.aliases=[];function n(r){r.languages.yang={comment:/\/\*[\s\S]*?\*\/|\/\/.*/,string:{pattern:/"(?:[^\\"]|\\.)*"|'[^']*'/,greedy:!0},keyword:{pattern:/(^|[{};\r\n][ \t]*)[a-z_][\w.-]*/i,lookbehind:!0},namespace:{pattern:/(\s)[a-z_][\w.-]*(?=:)/i,lookbehind:!0},boolean:/\b(?:false|true)\b/,operator:/\+/,punctuation:/[{};:]/}}return eqe}var tqe,UVt;function Rzr(){if(UVt)return tqe;UVt=1,tqe=n,n.displayName="zig",n.aliases=[];function n(r){(function(o){function s(k){return function(){return k}}var c=/\b(?:align|allowzero|and|anyframe|anytype|asm|async|await|break|cancel|catch|comptime|const|continue|defer|else|enum|errdefer|error|export|extern|fn|for|if|inline|linksection|nakedcc|noalias|nosuspend|null|or|orelse|packed|promise|pub|resume|return|stdcallcc|struct|suspend|switch|test|threadlocal|try|undefined|union|unreachable|usingnamespace|var|volatile|while)\b/,d="\\b(?!"+c.source+")(?!\\d)\\w+\\b",h=/align\s*\((?:[^()]|\([^()]*\))*\)/.source,b=/(?:\?|\bpromise->|(?:\[[^[\]]*\]|\*(?!\*)|\*\*)(?:\s*<ALIGN>|\s*const\b|\s*volatile\b|\s*allowzero\b)*)/.source.replace(/<ALIGN>/g,s(h)),y=/(?:\bpromise\b|(?:\berror\.)?<ID>(?:\.<ID>)*(?!\s+<ID>))/.source.replace(/<ID>/g,s(d)),E="(?!\\s)(?:!?\\s*(?:"+b+"\\s*)*"+y+")+";o.languages.zig={comment:[{pattern:/\/\/[/!].*/,alias:"doc-comment"},/\/{2}.*/],string:[{pattern:/(^|[^\\@])c?"(?:[^"\\\r\n]|\\.)*"/,lookbehind:!0,greedy:!0},{pattern:/([\r\n])([ \t]+c?\\{2}).*(?:(?:\r\n?|\n)\2.*)*/,lookbehind:!0,greedy:!0}],char:{pattern:/(^|[^\\])'(?:[^'\\\r\n]|[\uD800-\uDFFF]{2}|\\(?:.|x[a-fA-F\d]{2}|u\{[a-fA-F\d]{1,6}\}))'/,lookbehind:!0,greedy:!0},builtin:/\B@(?!\d)\w+(?=\s*\()/,label:{pattern:/(\b(?:break|continue)\s*:\s*)\w+\b|\b(?!\d)\w+\b(?=\s*:\s*(?:\{|while\b))/,lookbehind:!0},"class-name":[/\b(?!\d)\w+(?=\s*=\s*(?:(?:extern|packed)\s+)?(?:enum|struct|union)\s*[({])/,{pattern:RegExp(/(:\s*)<TYPE>(?=\s*(?:<ALIGN>\s*)?[=;,)])|<TYPE>(?=\s*(?:<ALIGN>\s*)?\{)/.source.replace(/<TYPE>/g,s(E)).replace(/<ALIGN>/g,s(h))),lookbehind:!0,inside:null},{pattern:RegExp(/(\)\s*)<TYPE>(?=\s*(?:<ALIGN>\s*)?;)/.source.replace(/<TYPE>/g,s(E)).replace(/<ALIGN>/g,s(h))),lookbehind:!0,inside:null}],"builtin-type":{pattern:/\b(?:anyerror|bool|c_u?(?:int|long|longlong|short)|c_longdouble|c_void|comptime_(?:float|int)|f(?:16|32|64|128)|[iu](?:8|16|32|64|128|size)|noreturn|type|void)\b/,alias:"keyword"},keyword:c,function:/\b(?!\d)\w+(?=\s*\()/,number:/\b(?:0b[01]+|0o[0-7]+|0x[a-fA-F\d]+(?:\.[a-fA-F\d]*)?(?:[pP][+-]?[a-fA-F\d]+)?|\d+(?:\.\d*)?(?:[eE][+-]?\d+)?)\b/,boolean:/\b(?:false|true)\b/,operator:/\.[*?]|\.{2,3}|[-=]>|\*\*|\+\+|\|\||(?:<<|>>|[-+*]%|[-+*/%^&|<>!=])=?|[?~]/,punctuation:/[.:,;(){}[\]]/},o.languages.zig["class-name"].forEach(function(k){k.inside===null&&(k.inside=o.languages.zig)})})(r)}return tqe}var nqe,zVt;function Pzr(){if(zVt)return nqe;zVt=1;var n=Yjr();return nqe=n,n.register(Jjr()),n.register(Zjr()),n.register(Qjr()),n.register(e$r()),n.register(t$r()),n.register(n$r()),n.register(r$r()),n.register(i$r()),n.register(o$r()),n.register(a$r()),n.register(s$r()),n.register(c$r()),n.register(u$r()),n.register(l$r()),n.register(d$r()),n.register(f$r()),n.register(h$r()),n.register(p$r()),n.register(g$r()),n.register(m$r()),n.register(b$r()),n.register(v$r()),n.register(adn()),n.register(sdn()),n.register(w$r()),n.register(y$r()),n.register(E$r()),n.register(S$r()),n.register(_$r()),n.register(k$r()),n.register(T$r()),n.register(x$r()),n.register(C$r()),n.register(A$r()),n.register(vU()),n.register(I$r()),n.register(O$r()),n.register(N$r()),n.register(R$r()),n.register(P$r()),n.register(L$r()),n.register(D$r()),n.register(M$r()),n.register(j$r()),n.register(yQe()),n.register($$r()),n.register(REe()),n.register(F$r()),n.register(B$r()),n.register(U$r()),n.register(z$r()),n.register(H$r()),n.register(G$r()),n.register(q$r()),n.register(V$r()),n.register(W$r()),n.register(K$r()),n.register(Y$r()),n.register(X$r()),n.register(J$r()),n.register(Z$r()),n.register(Q$r()),n.register(eFr()),n.register(tFr()),n.register(nFr()),n.register(rFr()),n.register(iFr()),n.register(oFr()),n.register(aFr()),n.register(sFr()),n.register(cFr()),n.register(uFr()),n.register(lFr()),n.register(dFr()),n.register(fFr()),n.register(hFr()),n.register(pFr()),n.register(gFr()),n.register(mFr()),n.register(bFr()),n.register(vFr()),n.register(wFr()),n.register(yFr()),n.register(EFr()),n.register(SFr()),n.register(_Fr()),n.register(kFr()),n.register(TFr()),n.register(xFr()),n.register(CFr()),n.register(AFr()),n.register(IFr()),n.register(OFr()),n.register(NFr()),n.register(EQe()),n.register(RFr()),n.register(PFr()),n.register(LFr()),n.register(DFr()),n.register(MFr()),n.register(jFr()),n.register($Fr()),n.register(FFr()),n.register(BFr()),n.register(UFr()),n.register(zFr()),n.register(HFr()),n.register(GFr()),n.register(qFr()),n.register(VFr()),n.register(WFr()),n.register(KFr()),n.register(SQe()),n.register(YFr()),n.register(LEe()),n.register(XFr()),n.register(JFr()),n.register(ZFr()),n.register(QFr()),n.register(eBr()),n.register(tBr()),n.register(nBr()),n.register(kQe()),n.register(rBr()),n.register(iBr()),n.register(oBr()),n.register(udn()),n.register(aBr()),n.register(sBr()),n.register(cBr()),n.register(uBr()),n.register(lBr()),n.register(dBr()),n.register(fBr()),n.register(hBr()),n.register(pBr()),n.register(gBr()),n.register(mBr()),n.register(bBr()),n.register(vBr()),n.register(wBr()),n.register(yBr()),n.register(EBr()),n.register(cdn()),n.register(SBr()),n.register(_Br()),n.register(kBr()),n.register(oy()),n.register(TBr()),n.register(xBr()),n.register(CBr()),n.register(ABr()),n.register(IBr()),n.register(OBr()),n.register(NBr()),n.register(RBr()),n.register(PBr()),n.register(LBr()),n.register(DBr()),n.register(MBr()),n.register(jBr()),n.register($Br()),n.register(FBr()),n.register(BBr()),n.register(UBr()),n.register(zBr()),n.register(HBr()),n.register(GBr()),n.register(qBr()),n.register(VBr()),n.register(WBr()),n.register(KBr()),n.register(YBr()),n.register(XBr()),n.register(JBr()),n.register(ZBr()),n.register(QBr()),n.register(eUr()),n.register(tUr()),n.register(nUr()),n.register(DEe()),n.register(rUr()),n.register(iUr()),n.register(oUr()),n.register(aUr()),n.register(sUr()),n.register(cUr()),n.register(uUr()),n.register(lUr()),n.register(dUr()),n.register(fUr()),n.register(hUr()),n.register(pUr()),n.register(gUr()),n.register(mUr()),n.register(bUr()),n.register(vUr()),n.register(wUr()),n.register(yUr()),n.register(EUr()),n.register(SUr()),n.register(_Ur()),n.register(kUr()),n.register(TUr()),n.register(xUr()),n.register(CUr()),n.register(AUr()),n.register(IUr()),n.register(OUr()),n.register(NUr()),n.register(RUr()),n.register(PEe()),n.register(PUr()),n.register(LUr()),n.register(DUr()),n.register(MUr()),n.register(TQe()),n.register(jUr()),n.register($Ur()),n.register(FUr()),n.register(BUr()),n.register(UUr()),n.register(zUr()),n.register(HUr()),n.register(GUr()),n.register(qUr()),n.register(VUr()),n.register(WUr()),n.register(KUr()),n.register(wQe()),n.register(YUr()),n.register(XUr()),n.register(JUr()),n.register(ZUr()),n.register(QUr()),n.register(ezr()),n.register(xQe()),n.register(tzr()),n.register(nzr()),n.register(rzr()),n.register(izr()),n.register(ozr()),n.register(azr()),n.register(szr()),n.register(czr()),n.register(ldn()),n.register(uzr()),n.register(_Qe()),n.register(lzr()),n.register(dzr()),n.register(fzr()),n.register(hzr()),n.register(pzr()),n.register(gzr()),n.register(ddn()),n.register(mzr()),n.register(bzr()),n.register(vzr()),n.register(wzr()),n.register(yzr()),n.register(Ezr()),n.register(Szr()),n.register(_zr()),n.register(kzr()),n.register(Tzr()),n.register(xzr()),n.register(Czr()),n.register(Azr()),n.register(Izr()),n.register(Ozr()),n.register(fdn()),n.register(Nzr()),n.register(Rzr()),nqe}var Lzr=Pzr();const Dzr=Y4(Lzr);var hdn=E9r(Dzr,Xjr);hdn.supportedLanguages=S9r;const Mzr={'pre[class*="language-"]':{color:"#d4d4d4",fontSize:"13px",textShadow:"none",fontFamily:'Menlo, Monaco, Consolas, "Andale Mono", "Ubuntu Mono", "Courier New", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",background:"#1e1e1e"},'code[class*="language-"]':{color:"#d4d4d4",fontSize:"13px",textShadow:"none",fontFamily:'Menlo, Monaco, Consolas, "Andale Mono", "Ubuntu Mono", "Courier New", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#264F78"},'code[class*="language-"]::selection':{textShadow:"none",background:"#264F78"},'pre[class*="language-"] *::selection':{textShadow:"none",background:"#264F78"},'code[class*="language-"] *::selection':{textShadow:"none",background:"#264F78"},':not(pre) > code[class*="language-"]':{padding:".1em .3em",borderRadius:".3em",color:"#db4c69",background:"#1e1e1e"},".namespace":{Opacity:".7"},"doctype.doctype-tag":{color:"#569CD6"},"doctype.name":{color:"#9cdcfe"},comment:{color:"#6a9955"},prolog:{color:"#6a9955"},punctuation:{color:"#d4d4d4"},".language-html .language-css .token.punctuation":{color:"#d4d4d4"},".language-html .language-javascript .token.punctuation":{color:"#d4d4d4"},property:{color:"#9cdcfe"},tag:{color:"#569cd6"},boolean:{color:"#569cd6"},number:{color:"#b5cea8"},constant:{color:"#9cdcfe"},symbol:{color:"#b5cea8"},inserted:{color:"#b5cea8"},unit:{color:"#b5cea8"},selector:{color:"#d7ba7d"},"attr-name":{color:"#9cdcfe"},string:{color:"#ce9178"},char:{color:"#ce9178"},builtin:{color:"#ce9178"},deleted:{color:"#ce9178"},".language-css .token.string.url":{textDecoration:"underline"},operator:{color:"#d4d4d4"},entity:{color:"#569cd6"},"operator.arrow":{color:"#569CD6"},atrule:{color:"#ce9178"},"atrule.rule":{color:"#c586c0"},"atrule.url":{color:"#9cdcfe"},"atrule.url.function":{color:"#dcdcaa"},"atrule.url.punctuation":{color:"#d4d4d4"},keyword:{color:"#569CD6"},"keyword.module":{color:"#c586c0"},"keyword.control-flow":{color:"#c586c0"},function:{color:"#dcdcaa"},"function.maybe-class-name":{color:"#dcdcaa"},regex:{color:"#d16969"},important:{color:"#569cd6"},italic:{fontStyle:"italic"},"class-name":{color:"#4ec9b0"},"maybe-class-name":{color:"#4ec9b0"},console:{color:"#9cdcfe"},parameter:{color:"#9cdcfe"},interpolation:{color:"#9cdcfe"},"punctuation.interpolation-punctuation":{color:"#569cd6"},variable:{color:"#9cdcfe"},"imports.maybe-class-name":{color:"#9cdcfe"},"exports.maybe-class-name":{color:"#9cdcfe"},escape:{color:"#d7ba7d"},"tag.punctuation":{color:"#808080"},cdata:{color:"#808080"},"attr-value":{color:"#ce9178"},"attr-value.punctuation":{color:"#ce9178"},"attr-value.punctuation.attr-equals":{color:"#d4d4d4"},namespace:{color:"#4ec9b0"},'pre[class*="language-javascript"]':{color:"#9cdcfe"},'code[class*="language-javascript"]':{color:"#9cdcfe"},'pre[class*="language-jsx"]':{color:"#9cdcfe"},'code[class*="language-jsx"]':{color:"#9cdcfe"},'pre[class*="language-typescript"]':{color:"#9cdcfe"},'code[class*="language-typescript"]':{color:"#9cdcfe"},'pre[class*="language-tsx"]':{color:"#9cdcfe"},'code[class*="language-tsx"]':{color:"#9cdcfe"},'pre[class*="language-css"]':{color:"#ce9178"},'code[class*="language-css"]':{color:"#ce9178"},'pre[class*="language-html"]':{color:"#d4d4d4"},'code[class*="language-html"]':{color:"#d4d4d4"},".language-regex .token.anchor":{color:"#dcdcaa"},".language-html .token.punctuation":{color:"#808080"},'pre[class*="language-"] > code[class*="language-"]':{position:"relative",zIndex:"1"},".line-highlight.line-highlight":{background:"#f7ebc6",boxShadow:"inset 5px 0 0 #f7d87c",zIndex:"0"}},FW=({language:n,code:r,hideHeader:o=!1})=>{const[s,c]=Q.useState(!1),d=Q.useRef(void 0);Q.useEffect(()=>()=>{d.current&&clearTimeout(d.current)},[]);const h=async()=>{try{await navigator.clipboard.writeText(r),c(!0),d.current&&clearTimeout(d.current),d.current=globalThis.setTimeout(()=>{c(!1)},2e3)}catch(b){Va(b,{tags:{component:"CodeBlock",action:"handleCopy"}})}};return H.jsxs(ai,{bg:"rasawebNeutral.900",borderRadius:"0.5rem",overflow:"hidden",my:"1rem",children:[o?null:H.jsxs($c,{bg:"rasawebNeutral.700",px:"1rem",py:"0.5rem",justify:"space-between",children:[H.jsx(Fi,{size:"xs",fontFamily:"IBM Plex Mono, monospace",color:"rasaNeutral.100",children:n||"code"}),H.jsx(Wg,{content:"Copy",showArrow:!0,children:H.jsxs(ud,{size:"sm",onClick:()=>void h(),variant:"ghost",color:"rasaNeutral.50","aria-label":"Copy",_hover:{bg:s?"none":"whiteAlpha.200",cursor:s?"default":"pointer"},height:"1.5rem",width:s?"auto":"1.5rem",gap:"0.25rem",px:s?"0.5rem":"0",bg:"transparent",children:[s?H.jsx(xa,{icon:uXe}):H.jsx(xa,{icon:jor}),s?H.jsx(Fi,{size:"sm",color:"rasaNeutral.100",children:"Copied!"}):null]})})]}),H.jsx(hdn,{language:n,style:Mzr,customStyle:{margin:0,padding:"1rem",fontFamily:"IBM Plex Mono, monospace",fontSize:"0.813rem",backgroundColor:"inherit"},codeTagProps:{style:{fontFamily:"IBM Plex Mono, monospace"}},children:r})]})},jzr=({triggerButton:n})=>{const{open:r,onOpen:o,onClose:s}=_M(),{isAuthenticated:c,isLoading:d,getIdTokenClaims:h}=ty(),{projectId:b}=G_(),{loading:y,loadingProject:E,isRestoring:k}=Io(da,ae=>({loading:ae.loading,isRestoring:ae.isRestoring,loadingProject:ae.loadingProject})),[x]=Dye(),O=y2(),{getPath:P}=MK(),D=Q.useRef(null),$=Io(da,ae=>ae.projectUrl);Q.useEffect(()=>{x.get(_cn)===kFt&&b&&(o(),O(P(),{replace:!0}))},[x,o,O,b,P]);const U=()=>{o()},B=Z8r({getIdTokenClaims:h,projectId:b,projectUrl:$||"",isAuthenticated:c}),G=b?`cd ${b}`:"cd <project-directory>",K=`uv run rasa train
uv run rasa inspect`,X=Q.useCallback(()=>d?H.jsx(Vd,{align:"center",gap:4,py:8,children:H.jsx(Fi,{textAlign:"center",children:"Loading..."})}):c?H.jsxs(Vd,{align:"stretch",gap:4,children:[H.jsx(ai,{ref:D,tabIndex:-1,position:"absolute",left:"-9999px",width:"1px",height:"1px",overflow:"hidden","aria-hidden":"true"}),H.jsxs(Fi,{fontSize:"sm",color:"gray.600",children:["This guide uses"," ",H.jsx(Fi,{as:"span",fontWeight:"bold",children:"UV"})," ","for package management. If you don't have it installed you can"," ",H.jsx(_I,{href:"https://docs.astral.sh/uv/getting-started/installation/",target:"_blank",variant:"underline",children:"get it here."})]}),H.jsxs(ai,{children:[H.jsxs($c,{align:"center",gap:3,mb:3,children:[H.jsx(Fi,{color:"neutral.700",children:"1. Download the project:"}),H.jsx(ud,{colorPalette:"dark",variant:"outline",size:"sm",disabled:B.isPending,onClick:()=>B.mutate(),children:"Download"})]}),B.downloadError!=null&&H.jsxs(Fi,{fontSize:"sm",color:"red.500",mt:2,children:["Download failed: ",B.downloadError]})]}),H.jsxs(ai,{children:[H.jsx(Fi,{color:"neutral.700",mb:2,children:"2. Open your project in your favorite IDE"}),H.jsx(FW,{code:G,hideHeader:!0})]}),H.jsxs(ai,{children:[H.jsxs(Fi,{color:"neutral.700",mb:2,children:["3."," ",H.jsx(_I,{href:"https://platform.openai.com/settings/organization/api-keys",target:"_blank",variant:"underline",children:"Configure OpenAI"})," ","by setting an ",H.jsx(lVe,{children:"OPENAI_API_KEY"})]}),H.jsx(FW,{code:'echo "OPENAI_API_KEY=<your-api-key>" >> .env',hideHeader:!0})]}),H.jsxs(ai,{children:[H.jsx(Fi,{color:"neutral.700",mb:2,children:"4. Install and run Rasa Pro"}),H.jsx(FW,{code:K,hideHeader:!0})]}),H.jsxs(Fi,{fontSize:"sm",color:"gray.600",children:["If you run into any issues ask Rasa Copilot or check out"," ",H.jsx(_I,{href:"https://rasa.com/docs/pro/installation/troubleshooting/",target:"_blank",rel:"noopener noreferrer",variant:"underline",children:"Troubleshooting in our Docs"})]})]}):H.jsx(Hln,{message:"To download your new project, you'll need to sign in first.",redirectParams:{do:kFt}}),[d,c,K,G,B]);return H.jsx(K4,{open:r,onOpen:U,onClose:s,button:n,title:"Continue building locally",size:"md",tooltipContent:y||E||k?void 0:"Download project",initialFocusEl:()=>D.current,children:X()})},$zr=({children:n})=>{const[r,o]=Q.useState(!1),[s,c]=Q.useState(!1),{projectId:d,loading:h,loadingProject:b,isRestoring:y}=Io(da,O=>({projectId:O.projectId,loading:O.loading,isRestoring:O.isRestoring,loadingProject:O.loadingProject})),{isAuthenticated:E}=ty();Q.useEffect(()=>{let O;return s&&(O=setTimeout(()=>c(!1),2e3)),()=>{O&&clearTimeout(O)}},[s]);const k=()=>{z_(H_.shareCopyLinkClicked);const O=`${globalThis.location.origin}/agent/${d}`;navigator.clipboard.writeText(O).then(()=>c(!0)).catch(P=>{Va(P,{tags:{component:"Navbar",action:"copyShareLink"}})})},x=O=>{o(O.open)};return H.jsxs(Wln,{isOpen:E&&r,onOpenChange:x,title:"Share your agent",tooltipContent:h||b||y?void 0:"Share preview of your agent",trigger:n,children:[H.jsx(Fi,{pt:"0.75rem",pb:"1rem",lineHeight:"170%",children:"Show what your agent can do — share a live preview."}),H.jsx(Kln,{children:H.jsx(Nve,{icon:Ror,label:s?"Copied!":"Copy link",showArrow:!1,onClick:k})}),H.jsx(Fi,{size:"sm",variant:"muted",pt:"0.75rem",children:"Private link. Updates to your agent sync automatically."})]})},pdn=()=>{const{isAuthenticated:n,isLoading:r,error:o}=ty(),{assistantName:s,loading:c,loadingProject:d,isRestoring:h}=Io(da,D=>({assistantName:D.assistantName,loading:D.loading,isRestoring:D.isRestoring,loadingProject:D.loadingProject})),b=Io(gc,PK),y=c||d||h||b,[E,k]=Q.useState(!1),{showProjectActions:x}=GM(),O={position:"sticky",borderBottom:"1px solid",borderColor:"rasaNeutral.400",height:`${Scn}rem`,w:"100%",px:"0"};Q.useEffect(()=>{o&&!r&&Kp.create({id:"login-failed",title:"Login Failed",description:o.message,type:"error",duration:1e4,closable:!0})},[o,r]);const P=()=>{c||n||k(!0)};return H.jsx($c,{css:O,children:H.jsxs($c,{gap:3,px:4,justifyContent:"space-between",alignItems:"center",flexGrow:"1",height:"100%",children:[H.jsx($c,{gap:"0",children:x&&H.jsxs(H.Fragment,{children:[H.jsx(Fi,{size:"sm","data-testid":"assistant-name",children:s||"Agent"}),H.jsx(j8r,{})]})}),H.jsx(B8r,{}),x&&H.jsxs($c,{children:[H.jsx(mQe,{isRequired:E,onClose:()=>{k(!1)},message:"Sign in to get a link to a preview page where others can chat with your agent directly."}),y||H.jsxs(H.Fragment,{children:[H.jsx($8r,{}),H.jsx(J8r,{children:H.jsx(Gm,{"data-testid":"connect-button",colorPalette:"light",size:"sm",disabled:c||d||h,"aria-label":"Help & resources",onClick:()=>{z_(H_.connectClicked).catch(D=>Va(D,{tags:{component:"navbar",action:"connectClicked"}}))},children:H.jsx(xa,{icon:Qtn})})}),H.jsx($zr,{children:H.jsx(Gm,{"aria-label":"Share project",variant:"solid",size:"sm",colorPalette:"light",disabled:c||d||h,onClick:P,children:H.jsx(xa,{icon:gor})})}),H.jsx(jzr,{triggerButton:()=>H.jsx(ud,{"aria-label":"Download project",variant:"solid",size:"sm",colorPalette:"dark",disabled:c||d||h,children:"Download"})})]})]})]})})},Fzr=()=>{const n=y2();return H.jsx(Bx,{shortLabel:"Oops! Failed to build the assistant",longLabel:"Something went wrong and we couldn't build the assistant",image:zm.CubesError,children:H.jsxs(Vd,{children:[H.jsx(ud,{onClick:()=>n("/"),variant:"solid",colorPalette:"dark",mb:"0.5rem",children:"Try again"}),H.jsx(ud,{onClick:()=>n("/"),colorPalette:"dark",variant:"outline",children:"Go back"})]})})},Bzr=()=>{const n=y2();return H.jsx(Bx,{shortLabel:"Your agent couldn't be trained",longLabel:"Generating from prompt is still in Alpha and sometimes things go wrong.",image:zm.CubesError,children:H.jsxs(Vd,{children:[H.jsx(ud,{onClick:()=>n("code"),variant:"solid",colorPalette:"dark",mb:"0.5rem",children:"Go to code"}),H.jsx(ud,{onClick:()=>n("/"),colorPalette:"dark",variant:"outline",children:"Go back"})]})})},Uzr=({message:n="Loading your project..."})=>H.jsx(Bx,{shortLabel:n,longLabel:"",image:zm.CubesPause,children:H.jsx(Vd,{children:H.jsx(kK,{size:"md",colorPalette:"dark"})})}),gdn=()=>H.jsx(Bx,{shortLabel:"We paused your project while you were away",longLabel:"",image:zm.CubesPause,children:H.jsxs($c,{children:[H.jsx(kK,{size:"md",colorPalette:"dark"}),H.jsx(Fi,{fontSize:"md",color:"gray.600",children:"Restoring project..."})]})}),mdn=()=>H.jsx(Bx,{shortLabel:"We paused your project while you were away",longLabel:"Refresh the page to jump back in",image:zm.CubesPause,children:H.jsx(Vd,{children:H.jsx(ud,{onClick:()=>globalThis.location.reload(),variant:"solid",colorPalette:"dark",mb:"0.5rem",children:"Refresh page"})})}),bdn=n=>{const{children:r,css:o,...s}=n,c={height:"100%",width:"100%",gridTemplateRows:"auto 1fr auto",gridTemplateAreas:`"header"
                        "content"
                        "footer"`,...o};return H.jsx(IFn,{css:c,...s,children:r})},zzr=n=>{const{children:r,...o}=n;return H.jsx(xYe,{area:"header",...o,children:r})},vdn=n=>{const{children:r,withSpacing:o=!0,css:s,...c}=n,d={...o&&{m:"1rem",mr:"0.5rem",pr:"1rem"},...s,overflow:"auto"};return H.jsx(xYe,{area:"content",css:d,...c,children:H.jsxs(Ien,{children:[H.jsx(Oen,{children:r}),H.jsx(Nen,{orientation:"vertical",children:H.jsx(PYe,{})})]})})},Hzr=n=>{const{children:r,...o}=n;return H.jsx(xYe,{area:"footer",...o,children:r})},CQe=({children:n,step:r})=>{const o=Q.useRef(null),{progressOnboarding:s,dismissOnboarding:c,onboardingState:d}=qoe(),h=S5r[r],y=Io(zcn,k=>k.isEnabled)&&!!d.step&&d.step===r&&r!==Qf.Hidden&&!!h;if(!y)return H.jsx(H.Fragment,{children:n});const E=H.jsxs(Vd,{align:"start",gap:3,maxW:"320px",children:[H.jsxs(Vd,{align:"start",gap:2,children:[H.jsx(Kx,{size:"lg",color:"rasawebNeutral.50",children:h.title}),H.jsx(Fi,{fontSize:"xs",color:"rasawebNeutral.50",lineHeight:"1.4",children:h.description})]}),H.jsxs($c,{justify:"space-between",width:"100%",children:[H.jsxs($c,{gap:2,children:[h.counter&&H.jsx(Fi,{fontSize:"xs",color:"rasawebNeutral.50",children:h.counter}),H.jsx(ud,{size:"xs",variant:"ghost",color:"rasawebNeutral.50",_hover:{bg:"whiteAlpha.200"},onClick:()=>c(),children:"Skip tour"})]}),H.jsx(ud,{variant:"outline",colorPalette:"dark",size:"xs",onClick:()=>s(r),children:h.action})]})]});return H.jsx(Wg,{content:E,showArrow:!0,open:y,closeDelay:0,portalled:!0,interactive:!0,contentProps:{px:6,py:6},contentRef:o,children:n})},Gzr=({item:n,isActive:r,onActivate:o,...s})=>{const c={borderColor:"rasaNeutral.300",color:"rasawebDeepPurple.800",_hover:{bg:"rasaNeutral.50"}},d={color:"rasaNeutral.50",bg:"rasawebPurple.800",borderColor:"rasawebPurple.800",_hover:{bg:"rasawebPurple.800",color:"rasaNeutral.50"}};return H.jsxs(ud,{...s,rounded:"full",variant:"outline",size:"sm",height:"1.5rem",fontSize:"13px",px:2.5,css:r?{...c,...d}:c,onClick:()=>o(n.value),children:[H.jsx(xa,{icon:Sor}),n.label]})},qzr=({onNewConversation:n})=>{const{inspectEnabled:r,setInspectEnabled:o}=bU(),{onboardingState:s,progressOnboarding:c}=qoe(),d=()=>{n()};return H.jsxs(Iu,{justifyContent:"space-between",flexShrink:1,alignItems:"center",children:[H.jsx(CQe,{step:Qf.Step3,children:H.jsx(Gzr,{item:{value:"inspect",label:"Inspect"},onActivate:()=>{s?.step===Qf.Step3&&c(s?.step),o(!r)},isActive:r,"aria-label":"Toggle inspect","data-testid":"inspect-toggle"})}),H.jsx(ai,{display:"flex",gap:"0.25rem",alignItems:"center",children:H.jsx(Wg,{content:"Restart conversation",showArrow:!0,children:H.jsx(Gm,{"aria-label":"Restart conversation","data-testid":"restart-conversation",variant:"solid",colorPalette:"light",onClick:d,size:"sm",children:H.jsx(xa,{icon:Jtn})})})})]})},MEe=n=>{const r="#F6F7FB",o="#FFFFFF",s="rasaNeutral.100",c={flexDirection:"row",textAlign:"left",ml:0,mr:0,_first:{mt:0},p:"0.5rem",pl:"1.5rem",pr:"1.5rem",bg:n?r:o},d={cursor:"pointer",_hover:{bg:s,"& .message-bubble":{bg:"rasaNeutral.50"}}},h={...c,...d};return{containerSx:c,hoverableSx:d,hoverableContainerSx:h,baseMessageSx:{display:"flex",flexDirection:"row",alignItems:"center",borderRadius:"0.5rem"},iconSx:{marginRight:"0.5rem"}}},Woe=({event:n})=>zI().pathname.startsWith("/project/")?H.jsx(ai,{position:"absolute",top:"-1.25rem",right:"5",zIndex:1,bg:"rasaNeutral.50",boxShadow:"0px 2px 8px 0px #00000026",borderRadius:"lg",children:H.jsxs(ud,{variant:"subtle",colorPalette:"dark",size:"sm",onClick:()=>B5r({...n}),"aria-label":"Ask copilot about this message",children:[H.jsx(xa,{icon:Qtn}),"Ask Copilot"]})}):null,Vzr={[Sl.FlowStarted]:BVe,[Sl.FlowResumed]:BVe,[Sl.FlowCompleted]:sXe,[Sl.FlowCancelled]:rnn,[Sl.FlowInterrupted]:dor};function Wzr(n){const r=n.flowId?.startsWith("pattern_")?"System flow":"Flow",o=bIr[n.conversationEventType]??"";return H.jsxs(Fi,{size:"sm",variant:"muted",lineClamp:2,wordBreak:"break-all",children:[r," ",H.jsx(Fi,{as:"span",fontWeight:"600",variant:"muted",children:n.flowId})," ",o]})}const Kzr=Q.forwardRef((n,r)=>{const{event:o,isSelected:s=!1,...c}=n,{hoverableContainerSx:d,baseMessageSx:h,iconSx:b}=MEe(s),[y,E]=Un.useState(!1),k={...h,color:"rasaNeutral.700"},x=Vzr[o.conversationEventType]||null;return SIr(o.flowId||"")?null:H.jsxs(Iu,{css:d,ref:r,tabIndex:0,position:"relative",onMouseEnter:()=>E(!0),onMouseLeave:()=>E(!1),...c,children:[y&&H.jsx(Woe,{event:o}),H.jsxs(ai,{css:k,children:[x?H.jsx(xa,{icon:x,style:b}):null,Wzr(o)]})]})});function Yzr(n){if(n.conversationEventType===Sl.ResetSlots)return H.jsx(Fi,{variant:"muted",children:"Slots reset"});const o=n.conversationEventType===Sl.Slot&&n.metadata?.reset||n.slotValue===null?"is cleared":"set",s=n.metadata?.was_modified_by_studio?" (modified)":"";return H.jsxs(Fi,{size:"sm",lineClamp:2,wordBreak:"break-all",variant:"muted",children:["Slot"," ",H.jsx(Fi,{as:"span",fontWeight:"600",variant:"muted",children:n.name})," ",o,s]})}const Xzr=Q.forwardRef((n,r)=>{const{event:o,isSelected:s=!1,...c}=n,{hoverableContainerSx:d,baseMessageSx:h,iconSx:b}=MEe(s),[y,E]=Un.useState(!1),k={...h,color:"rasaNeutral.700"};return vIr(o.name||"")?null:H.jsxs(Iu,{css:d,ref:r,tabIndex:0,position:"relative",onMouseEnter:()=>E(!0),onMouseLeave:()=>E(!1),...c,children:[y&&H.jsx(Woe,{event:o}),H.jsxs(ai,{css:k,children:[H.jsx(xa,{icon:cXe,style:b}),Yzr(o)]})]})}),Jzr=Q.forwardRef((n,r)=>{const{event:o,isSelected:s=!1,replayConversation:c,...d}=n,{hoverableContainerSx:h,containerSx:b,baseMessageSx:y,iconSx:E}=MEe(s),[k,x]=Un.useState(!1),O=o.metadata?.execution_success===!1,P={...b,pb:0},D={...h,":hover .replayIcon":{display:"inline-block"},pt:void 0,pb:void 0},$={mr:"0.5rem",color:"#A72E2C",ml:"0.5rem"},U={...y,color:"rasaNeutral.700"},B={...U,ml:"2.75rem"},G=o.name===_Ft?mor:NK,K=o.name===_Ft?"Waiting for user input":o.name;return wIr(o.name||"")?null:o.name?.startsWith("utter_")?H.jsx(Iu,{css:P,ref:r,children:H.jsx(Fi,{size:"sm",css:B,children:o.name})}):H.jsxs(Iu,{css:D,alignItems:"center",ref:r,tabIndex:0,position:"relative",onMouseEnter:()=>x(!0),onMouseLeave:()=>x(!1),...d,children:[k&&H.jsx(Woe,{event:o}),H.jsxs(ai,{css:U,children:[H.jsx(xa,{icon:G,style:E}),H.jsx(Fi,{size:"sm",variant:"muted",lineClamp:2,wordBreak:"break-all",children:K})]}),O?H.jsx(xa,{icon:Xtn,style:$,"box-size":"1rem","aria-label":"Action failed"}):null,null]})}),Zzr=Q.forwardRef((n,r)=>{const{event:o}=n,[s,c]=Un.useState(!1),d={fontSize:"0.75rem",mb:"0.5rem",ml:"1.5rem",_last:{mb:0}},h={padding:"0.5rem",pr:"3.5rem",_first:{mt:0}};return yIr(o.conversationEventType.toLowerCase())?null:H.jsxs(Iu,{css:h,ref:r,position:"relative",onMouseEnter:()=>c(!0),onMouseLeave:()=>c(!1),children:[s&&H.jsx(Woe,{event:o}),H.jsx(ai,{ml:"2rem",children:H.jsx(Fi,{size:"sm",variant:"muted",css:d,children:o.conversationEventType.toLowerCase()})})]})}),AQe=[Sl.FlowCancelled,Sl.FlowCompleted,Sl.FlowStarted,Sl.FlowInterrupted,Sl.FlowResumed],IQe=[Sl.Slot,Sl.ResetSlots],Qzr=Q.forwardRef((n,r)=>{const{event:o,selectable:s=!1,onEventSelect:c,isSelected:d=!1,replayConversation:h,isLastActiveEvent:b=!1,...y}=n,E=()=>{s&&c&&c(o)},k={...y,"data-testid":b?"conversation-events-last-active-event":void 0},x=O=>{O.key==="Enter"&&s&&c&&c(o)};return AQe.includes(o.conversationEventType)?H.jsx(Kzr,{ref:r,event:o,onClick:E,onKeyDown:x,isSelected:s&&d,...k}):IQe.includes(o.conversationEventType)?H.jsx(Xzr,{ref:r,event:o,onClick:E,onKeyDown:x,isSelected:s&&d,...k}):o.conversationEventType===Sl.Action?H.jsx(Jzr,{ref:r,event:o,onClick:E,onKeyDown:x,isSelected:s&&d,replayConversation:h,...k}):H.jsx(Zzr,{event:o,ref:r,...k})}),OQe=Q.forwardRef((n,r)=>{const{utterance:o,isSelected:s,isHighlighted:c,previousUtterance:d,topIntentName:h,children:b,containerSx:y,messageSx:E,messageBreakout:k,isInteractive:x=!1,isUser:O=!1,topOverlay:P,...D}=n,{hoverableSx:$}=MEe(s??!1),U="rasaNeutral.100",B="rasawebDeepPurple.800",G="rasawebDeepPurple.800",K="rasaNeutral.50",X="rasaNeutral.50",oe=(o?.type||Vx.Bot)!==d?.type,he={_first:{mt:0},px:"1rem",py:"0.5rem",pr:O?"1.5rem":"3.5rem",pl:O?"3.5rem":"1.5rem",bg:s?"rasaNeutral.200":"transparent",...$,...y};let ye;O?ye=B:ye=s?X:U;const ge={borderRadius:O?"1rem 0.25rem 1rem 1rem":"0.25rem 1rem 1rem 1rem",p:o?"0":"1rem",bg:ye,color:O||!x&&c?K:G,fontSize:"0.813rem",overflow:"hidden",mb:"0.25rem",wordBreak:"break-word",maxWidth:"20rem",...E},Te={color:"#6B7694",fontSize:"0.75rem",mb:"0.5rem",mr:"1rem",_last:{mb:0}},xe=()=>o?c?"conversation-message-highlighted":O?"assistant-user-message":"assistant-response":"agent-loading-spinner";return H.jsxs(Iu,{css:he,...D,ref:r,position:"relative",children:[oe?H.jsx(NEe,{size:"sm",mr:O?0:"0.5rem",ml:O?"0.5rem":0,src:O?v5r:b5r}):H.jsx(ai,{width:"2.5rem"}),H.jsxs(ai,{display:"inline-block",children:[P,H.jsx(ai,{css:ge,"data-testid":xe(),className:O?"":"message-bubble",children:b}),k,!!h&&H.jsx(ai,{css:Te,children:h})]})]})}),eHr=({items:n,onQuickReply:r,isDisabled:o=!1})=>{const s={mt:"0.5rem",gap:"0.5rem",flexWrap:"wrap"},c={whiteSpace:"unset",height:"auto",minHeight:"2rem"},d=(b,y)=>{o||(r?.(y),b.stopPropagation())},h=n.map(b=>{const y=b.type==="link",E=b.type==="button",k=b.type==="attachment";return y||k?H.jsxs(Gln,{variant:"outline",colorPalette:"dark",to:b.payload,target:"_blank",size:"sm",css:c,onClick:x=>x.stopPropagation(),children:[b.title,b.icon&&H.jsx(xa,{icon:b.icon})]},b.title):E?H.jsx(ud,{variant:"outline",colorPalette:"dark",onClick:x=>d(x,b.payload),disabled:o,size:"sm",css:c,children:b.title},b.title):null});return H.jsx(Iu,{css:s,children:h})},tHr=Q.forwardRef((n,r)=>{const{utterance:o,onQuickReply:s,isSelected:c,isInteractive:d,...h}=n,{open:b,onOpen:y,onClose:E}=_M(),[k,x]=Q.useState(!1),O=o.responseData?.image,P=nHr(o.responseData),D=o.text,$=o.responseData?.custom,U=[{id:"Cancel",title:"Cancel",onClick:K=>{K.stopPropagation(),E()},"aria-label":"Cancel",variant:"outline",colorPalette:"dark"}],B={borderTopRadius:"0.25rem",borderTopRightRadius:"1rem",objectFit:"cover",objectPosition:"center",width:"100%"},G={borderColor:"#DDE2EF",my:"0.5rem"};return H.jsxs(OQe,{onMouseEnter:()=>x(!0),onMouseLeave:()=>x(!1),utterance:o,topOverlay:k?H.jsx(Woe,{event:o}):void 0,isSelected:c,...h,ref:r,messageBreakout:P.length?H.jsx(eHr,{items:P,onQuickReply:s,isDisabled:!d}):null,children:[O?H.jsx(Iye,{src:O,css:B,alt:"Assistant image"}):null,H.jsxs(ai,{p:"1rem",_empty:{display:"none"},children:[D?H.jsx(Fi,{size:"md",whiteSpace:"pre-wrap",children:D}):null,D&&$?H.jsx(dVe,{orientation:"horizontal",css:G}):null,$?H.jsxs(ai,{children:[H.jsxs(Iu,{alignItems:"center",children:[H.jsx(xa,{icon:NK,fontSize:"1rem",style:{marginRight:"7"}}),"JSON"]}),H.jsx(K4,{button:()=>H.jsx(ud,{size:"sm",mt:"3",colorPalette:"dark",children:"Click to view code"}),title:o.metadata?.utter_action,actions:U,open:b,onClose:E,onOpen:y,children:H.jsx(FW,{language:"json",code:JSON.stringify($,null,2)||"-"})})]}):null]})]})}),nHr=n=>{if(!n)return[];const r=[],o=[...n.buttons,...n.quickReplies].filter(Boolean);for(const s of o){const c=x5r(s.payload);r.push({title:s.title,payload:s.payload,type:c?"link":"button",...c?{icon:Uye}:{}})}return n.attachment&&r.push({title:n.attachment,payload:n.attachment,icon:_or,type:"attachment"}),r},rHr=Q.forwardRef((n,r)=>{const{utterance:o,...s}=n,c={flexDirection:"row-reverse",textAlign:"right"},d={p:"1rem",cursor:n.selectable?"pointer":"auto",whiteSpace:"pre-wrap"};return EIr(o.text||"")?null:H.jsx(OQe,{utterance:o,containerSx:c,messageSx:d,...s,isUser:!0,ref:r,children:o.text})}),iHr=Q.forwardRef((n,r)=>{const{utterance:o,selectable:s=!1,isSelected:c=!1,isInteractive:d=!1,onMessageSelect:h,onQuickReply:b=void 0,...y}=n,E=o.type===Vx.User,k=Q.useCallback(P=>{d&&b?.(P)},[d,b]),x=()=>{s&&h&&h({...o,entities:[]})},O=P=>{P.key==="Enter"&&s&&h&&h({...o,entities:[]})};return E?H.jsx(rHr,{utterance:o,...y,isInteractive:d,ref:r,onClick:x,onKeyDown:O,tabIndex:0,isSelected:s&&c}):H.jsx(tHr,{utterance:o,...y,isInteractive:d,ref:r,onClick:x,tabIndex:0,isSelected:s&&c,onQuickReply:k,onKeyDown:O})}),oHr=()=>{const n={borderRadius:"50%",width:"0.375rem",height:"0.375rem",bg:"rasaNeutral.500",animationName:"blink",animationDuration:"1.25s",animationTimingFunction:"ease-in-out",animationIterationCount:"infinite"};return H.jsx(OQe,{"aria-label":"Agent is typing",children:H.jsxs(Iu,{gap:"0.25rem","data-testid":"loading-dots",children:[H.jsx(ai,{css:n}),H.jsx(ai,{css:n,animationDelay:"0.25s"}),H.jsx(ai,{css:n,animationDelay:"0.5s"})]})})},aHr=({conversation:n,onQuickReply:r,selectable:o=!1,onSelect:s,selectedElementId:c,inspectorMode:d,replayConversation:h,interactive:b=!1,waitingForResponse:y=!1,replayingConversation:E=!1,assistantVersion:k})=>{const{events:x=[]}=n,O=rsn(new Date(n.startDate)),P={color:"rasaNeutral.700",fontSize:"0.75rem",textAlign:"center",maxWidth:"80%"},D={overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap"},$=x.filter(K=>UB(K)||nK(K)),U=$.findLast(K=>UB(K)),B=U?.type===Vx.Bot?U.id:void 0,G=Q.useCallback(K=>{d&&s?.(K)},[s,d]);return H.jsxs(ai,{mt:"1rem",mb:"1rem",children:[H.jsxs($c,{padding:"0.5rem",width:"100%",justifyContent:"center",children:[H.jsx(dVe,{orientation:"horizontal",marginLeft:"1rem"}),H.jsxs(ai,{css:P,flex:"0 1 auto",children:[H.jsxs(Fi,{size:"sm",variant:"muted",children:["Session started on ",O]}),k?H.jsx(Fi,{css:D,children:k}):null]}),H.jsx(dVe,{orientation:"horizontal",marginRight:"1rem"})]}),$.map((K,X)=>{if(UB(K))return H.jsx(iHr,{utterance:K,isInteractive:b&&K.id===B,onQuickReply:r,onMessageSelect:G,isSelected:K.id===c,selectable:o,inspectorMode:d},K.id);if(nK(K))return d?H.jsx(Qzr,{event:K,selectable:o,onEventSelect:G,isSelected:K.id===c,replayConversation:h,isLastActiveEvent:X===$.length-1&&b},K.id):null}),y||E?H.jsx(oHr,{}):null]})},sHr=({conversationList:n,onQuickReply:r,conversationAssistant:o={},onSelect:s,selectedElementId:c,inspectorMode:d,replayConversation:h,waitingForResponse:b=!1,replayingConversation:y=!1})=>{const E=Q.useRef(null),k=n[n.length-1],x=k?.events[k.events.length-1]?.id;Q.useEffect(()=>{setTimeout(()=>{E.current?.scrollIntoView?.({behavior:"smooth",block:"center"})},0)},[x,d]);const O=n.filter((P,D)=>P.totalNumberOfUserMessages>0||D===n.length-1);return H.jsx(bdn,{"data-testid":"assistant-chat",children:H.jsxs(vdn,{withSpacing:!1,children:[O.map((P,D)=>H.jsx(aHr,{interactive:D===O.length-1,conversation:P,onQuickReply:r,assistantVersion:o[P.id],selectable:!0,onSelect:s,replayConversation:h,selectedElementId:c,inspectorMode:d,waitingForResponse:D===O.length-1&&b,replayingConversation:D===O.length-1&&y},P.id)),H.jsx("div",{ref:E})]})})},cHr=Q.forwardRef(({onSubmit:n,isDisabled:r=!1},o)=>{const[s,c]=Q.useState(""),{updateOnboardingStatus:d,onboardingState:h}=qoe(),b=()=>{s&&!r&&(h.step!==Qf.Finished&&h.step!==Qf.Loading&&d(Qf.Step3),n(s),c(""))},y=x=>{c(x.target.value)},E=x=>{x.key==="Enter"&&!r&&b()},k={bg:"rasaNeutral.100",borderRadius:"3xl",display:"flex",justifyContent:"center"};return H.jsx(NQn,{onKeyDown:E,css:k,m:"1.5rem",px:"1rem",width:"calc(100% - 3rem)",height:"3.5rem","data-testid":"assistant-input",children:H.jsxs(H.Fragment,{children:[H.jsx(OQn,{type:"text",placeholder:"Type your message",value:s,onChange:y,disabled:r,ref:o,css:{borderWidth:"0",outline:"none","&:hover":{outline:"none"}}}),H.jsx(Gm,{variant:"solid",colorPalette:"dark",rounded:"full","aria-label":"Send message",size:"2xs",fontSize:"0.6rem",onClick:b,disabled:s.length===0||r,children:H.jsx(xa,{icon:Uor})})]})})}),uHr=({replayConversation:n,conversationAssistants:r,conversationList:o,handleMessageSubmit:s,handleSelect:c,selectedElement:d,inputDisabled:h,inspectorMode:b,waitingForResponse:y,replayingConversation:E,onNewConversation:k})=>{const x=Q.useRef(null),{view:O}=G_(),P="#FFFFFF",D={borderBottom:"1px solid",borderColor:"rasaNeutral.200",textAlign:"center",px:"1.5rem",py:"0.5rem"};return Q.useEffect(()=>{O==="chat"&&x.current&&!h&&x.current.focus()},[O,h]),H.jsxs(bdn,{children:[H.jsx(zzr,{css:D,children:H.jsx(qzr,{onNewConversation:k})}),H.jsx(vdn,{withSpacing:!1,bg:P,mb:"1rem",children:H.jsx(sHr,{replayConversation:n,conversationAssistant:r,conversationList:o,onQuickReply:s,onSelect:c,selectedElementId:d?.id,inspectorMode:b,waitingForResponse:y,replayingConversation:E})}),H.jsx(Hzr,{children:H.jsx(CQe,{step:Qf.Step1,children:H.jsx(ai,{children:H.jsx(cHr,{ref:x,onSubmit:s,isDisabled:h})})})})]})};function lHr(n){switch(n){case o2.Collect:return av.COLLECT_INFORMATION;case o2.Link:return av.LINK;case o2.Action:return av.CUSTOM_ACTION;case o2.SetSlots:return av.SET_SLOTS;case o2.Call:return av.CALL;case o2.Noop:return;default:return}}function wdn(n){let r=n.stepType;if(r===void 0)switch(Object.keys(n).find(s=>["collect","action","set_slots","link","call","noop"].includes(s))){case"collect":r=o2.Collect;break;case"action":r=o2.Action;break;case"set_slots":r=o2.SetSlots;break;case"link":r=o2.Link;break;case"call":r=o2.Call;break;case"noop":r=o2.Noop;break;default:throw new Error(`Cannot determine step type for step: ${JSON.stringify(n,null,2)}`)}return{...n,nodeType:lHr(r)}}function dHr(n){return typeof n=="object"&&"if"in n[0]}function fHr(n){return n.noop&&n.id?n.id:crypto.randomUUID()}function ydn({steps:n,processedSteps:r,stepSources:o,conditionStepSources:s}){const[c]=n;if(!c)return{steps:r,edges:o,conditionStepSources:s};let d=n.slice(1);const h={...c,id:c.id||crypto.randomUUID()},b={...o},y={...s};if(r.length||(b[h.id]=[null]),!c.id||!b[c.id]){const E=r[r.length-1];b[h.id]=[E]}if(c.next&&c.next!=="END")if(typeof c.next=="string"){b[c.next]=[...b[c.next]||[],h];const[E,k]=OEe.partition(d,{id:c.next});E?.[0]?.next&&(d=[...E,...k])}else if(dHr(c.next)){const E=fHr(c),x=c.next.map((D,$)=>({...D,nodeType:av.CONDITION,condition:D$e(D)?D.if:void 0,conditionType:D$e(D)?gM.IF:gM.ELSE,conditionOrder:$+1,next:D$e(D)?D.then:D.else,id:crypto.randomUUID()})),O={id:E,nodeType:av.LOGIC},P=[O,...x];for(const D of x)b[D.id]=[O],y[D.id]=O.id;d=[...P,...d]}else d=[...c.next.map(wdn),...d];return ydn({steps:d,processedSteps:[...r,h],stepSources:b,conditionStepSources:y})}function hHr(n,r){const o={id:n.id,metadata:{x:0,y:0}};switch(n.nodeType){case av.COLLECT_INFORMATION:return{...o,type:Pa.CollectInformation,label:n.collect??"Collect Information"};case av.LINK:return{...o,type:Pa.Link,label:n.link??"Link"};case av.CUSTOM_ACTION:return OEe.startsWith(n.action,"utter_")?{...o,type:Pa.Message,label:n.action??"Message"}:{...o,type:Pa.CustomAction,label:n.action||"Custom Action"};case av.SET_SLOTS:return{...o,type:Pa.SetSlots,label:"Set Slots"};case av.CALL:return{...o,type:Pa.Call,label:"Call a flow and return"};case av.LOGIC:return{...o,type:Pa.Logic,label:"If..."};case av.CONDITION:{const s=n.conditionType===gM.IF,c=r[n.id];return{...o,type:Pa.Condition,logicId:c,order:n.conditionOrder??1,subType:s?gM.IF:gM.ELSE,label:s?`If ${n.condition||"Condition"}`:`Else ${n.condition||"Condition"}`}}default:console.warn(`Unknown step type: ${JSON.stringify(n)}`);return}}function pHr(n,r){const{steps:o,edges:s,conditionStepSources:c}=ydn({steps:n.map(wdn),processedSteps:[],stepSources:{},conditionStepSources:{}}),d={},h=new Map;for(const O of o){if(O.noop)continue;const P=hHr(O,d);if(P&&P.type===Pa.Logic){const D=Object.keys(c).filter($=>c[$]===O.id);for(const $ of D)d[$]=P.id}if(!O.id)throw new Error(`Step should always have an ID, flow: ${r}`);P&&h.set(O.id,P)}const b=[];for(const O of Object.keys(s)){if(!o.find(B=>B.id===O&&!B.noop))continue;const D=h.get(O);if(!D)throw new Error(`Target node for ${O} not found, flow: ${r}`);const U=s[O].map(B=>{if(B&&!B.id)throw new Error(`Source step should always have an ID, flow: ${r}`);return B?h.get(B.id):null});for(const[B,G]of U.entries())G&&b.push({sourceId:G.id,targetId:D.id,type:B===0?AW.Auto:AW.Custom})}const y={id:"START",type:Pa.Start,label:"Start",metadata:{x:0,y:0}};if(o.every(O=>O.noop))return{nodes:[y],edges:[]};const k=o.find(O=>!O.noop);if(!k)throw new Error(`No valid target step found for start edge, flow: ${r}`);const x={sourceId:y.id,targetId:k.id,type:AW.Auto};return{nodes:[y,...Array.from(h.values())],edges:[x,...b]}}const gHr=()=>{const n=Io(da,h=>h.projectUrl),{data:r,isLoading:o,error:s,refetch:c}=xK({queryKey:["botData",n],queryFn:()=>bhr({projectUrl:n}),enabled:!!n,refetchOnWindowFocus:!1});return{flows:Q.useMemo(()=>{if(!r?.flows)return[];const h=[];for(const[b,y]of Object.entries(r.flows)){const{nodes:E,edges:k}=pHr(y.steps,b);h.push({id:b,name:y.name||b,nodes:E,edges:k})}return h},[r?.flows]),isLoading:o,error:s,refetch:c}},Koe={rasaNeutral:{50:{value:"#FFFFFF"},100:{value:"#F6F7FB"},200:{value:"#F2F4FA"},300:{value:"#ECEFF7"},400:{value:"#DDE2EF"},500:{value:"#CED5E7"},600:{value:"#9DA6BE"},700:{value:"#6B7694"}},rasawebNeutral:{50:{value:"#FDF9F3"},600:{value:"#656565"},700:{value:"#3A3939"},900:{value:"#292727"},"900D-20":{value:"rgba(41, 39, 39, 0.2)"},"900D-60":{value:"rgba(41, 39, 39, 0.6)"}},rasawebPurple:{50:{value:"#F5F3F9"},800:{value:"#5A17EE"},900:{value:"#380D98"}},rasawebLavender:{200:{value:"#BBA8FF"},700:{value:"#A855F7"}},rasawebYellow:{400:{value:"#EAB308"},900:{value:"#806908"}},rasawebDeepPurple:{800:{value:"#2B1B44"},900:{value:"#0E0817"}},rasaCyan:{800:{value:"#2C9793"}},rasawebCyan:{400:{value:"#1EC9C6"}},rasaOrange:{600:{value:"#EB7D15"}},rasawebOrange:{600:{value:"#F97316"}},rasaBlue:{900:{value:"#294BA0"}},rasawebBlue:{600:{value:"#2563EB"}},rasaGreen:{800:{value:"#2B8F5A"}},rasaPink:{900:{value:"#DA6BE0"}},rasaRed:{800:{value:"#A72E2C"}}},mHr=Bo({slots:["root","backdrop","positioner","content","header","body","footer","closeTrigger","trigger"],base:{content:{borderRadius:"1rem"}},variants:{size:{sm:{content:{maxW:"448px"}},md:{content:{maxW:"684px"}},lg:{content:{maxW:"894px"}}}},defaultVariants:{size:"sm"}}),bHr={px:{value:"1px"},0:{value:"0"},1:{value:"0.25rem"},2:{value:"0.5rem"},3:{value:"0.75rem"},4:{value:"1rem"},5:{value:"1.25rem"},6:{value:"1.5rem"}},vHr={xs:{value:"0px 1px 2px rgba(0, 0, 0, 0.05)"},sm:{value:"0px 1px 3px rgba(0, 0, 0, 0.1)"},md:{value:"0px 4px 8px rgba(0, 0, 0, 0.1)"},lg:{value:"0px 8px 16px rgba(0, 0, 0, 0.1)"},xl:{value:"0px 16px 32px rgba(0, 0, 0, 0.12)"},"2xl":{value:"0px 24px 48px rgba(0, 0, 0, 0.15)"},popup:{value:"0px 8px 32px rgba(0, 0, 0, 0.12)"},card:{value:"0px 4px 16px rgba(0, 0, 0, 0.08)"},tooltip:{value:"0px 2px 8px rgba(0, 0, 0, 0.1)"},modal:{value:"0px 12px 48px rgba(0, 0, 0, 0.15)"},focus:{value:"0 0 0 1px rgba(37, 99, 235, 0.3)"}},HVt=Rh({base:{_disabled:{opacity:"0.4"},borderRadius:"lg"},variants:{size:{sm:{height:"8",minWidth:"2rem"}},variant:{solid:{bg:"colorPalette.solid",color:"colorPalette.contrast",_hover:{bg:"colorPalette.fg"},_expanded:{bg:"colorPalette.fg"},_disabled:{bg:"colorPalette.solid"}},outline:{bg:"colorPalette.contrast",color:"colorPalette.solid",borderColor:"colorPalette.solid",_hover:{color:"colorPalette.fg",bg:"colorPalette.muted"},_expanded:{color:"colorPalette.emphasized",bg:"colorPalette.muted"},_disabled:{color:"colorPalette.fg"}},ghost:{bg:"transparent",color:"rasawebDeepPurple.800",_hover:{color:"rasawebPurple.900",bg:"transparent"},_disabled:{color:"rasawebDeepPurple.800"}},subtle:{bg:"colorPalette.contrast",color:"colorPalette.subtle",_hover:{color:"colorPalette.fg",bg:"none"},_disabled:{color:"colorPalette.fg"}}}}}),wHr=Rh({base:{fontWeight:"500",color:"rasawebDeepPurple.800",letterSpacing:"0.4px"},variants:{size:{sm:{fontSize:"13px"},md:{fontSize:"14px"},lg:{fontSize:"16px"},xxl:{fontSize:"42px"}}}}),yHr=Rh({base:{fontWeight:"400",color:"rasawebDeepPurple.800",letterSpacing:"0.4px"},variants:{size:{xs:{fontSize:"11px",lineHeight:"16px"},sm:{fontSize:"13px",lineHeight:"20px"},md:{fontSize:"14px",lineHeight:"20px"},xxl:{fontSize:"20px",lineHeight:"28px"}},variant:{primary:{color:"rasawebDeepPurple.800"},muted:{color:"rasaNeutral.700"}}},defaultVariants:{variant:"primary"}}),EHr=Bo({slots:["root","item","itemText","indicator","label"],base:{root:{bg:"rasaNeutral.50",height:"2rem",borderRadius:"0.75rem",color:"rasaNeutral.400",boxShadow:"inset 0 0 0 1px",gap:"0.25rem",p:"1",divideStyle:"none","& label":{padding:"0.6rem",height:"1.5rem",lineHeight:"1.5rem",borderRadius:"0.5rem","&:not(:hover)":{backgroundColor:"transparent !important"},"&:hover:not([data-state='checked'])":{bg:"rasaNeutral.300",color:"rasawebDeepPurple.800"},"&:hover[data-state='checked']":{bg:"rasawebDeepPurple.900"}}},item:{fontWeight:"500",color:"rasaNeutral.700",lineHeight:"1.5rem",cursor:"pointer"},indicator:{bg:"rasawebDeepPurple.800",color:"rasaNeutral.50",height:"1.5rem",borderRadius:"0.5rem"},itemText:{height:"1.5rem",fontSize:"13px","&[data-state='checked']":{color:"rasaNeutral.50"}}},variants:{}}),SHr=Bo({slots:["root","label"],base:{root:{bg:"colorPalette.subtle",borderRadius:"0.5rem"},label:{color:"colorPalette.solid",fontWeight:"500",fontSize:"13px"}},variants:{}}),_Hr=Bo({slots:["root"],base:{root:{borderRadius:"2xl","&[data-type=warning]":{bg:"rasawebPurple.800",color:"rasaNeutral.50"}}}}),kHr=EYe({theme:{tokens:{colors:Koe,spacing:bHr,shadows:vHr,fonts:{heading:{value:'Soehne, -apple-system, BlinkMacSystemFont, "Segoe UI", sans-serif'},body:{value:'Soehne, -apple-system, BlinkMacSystemFont, "Segoe UI", sans-serif'}},fontWeights:{normal:{value:400},medium:{value:500}},animations:{blink:{value:"blink 1s ease-in-out infinite"}}},textStyles:{heading:{value:{fontFamily:"{fonts.heading}",fontWeight:"{fontWeights.medium}"}},body:{value:{fontFamily:"{fonts.body}",fontWeight:"{fontWeights.normal}"}}},semanticTokens:{colors:{purple:{solid:{value:"{colors.rasawebPurple.800}"},fg:{value:"{colors.rasawebPurple.900}"},contrast:{value:"{colors.rasaNeutral.50}"},subtle:{value:"{colors.rasawebPurple.50}"},muted:{value:"{colors.rasaNeutral.200}"},emphasized:{value:"{colors.rasaNeutral.500}"}},dark:{solid:{value:"{colors.rasawebDeepPurple.800}"},fg:{value:"{colors.rasawebDeepPurple.900}"},contrast:{value:"{colors.rasaNeutral.50}"},subtle:{value:"{colors.rasaNeutral.700}"},muted:{value:"{colors.rasaNeutral.200}"},emphasized:{value:"{colors.rasaNeutral.500}"}},light:{solid:{value:"{colors.rasaNeutral.50}"},fg:{value:"{colors.rasaNeutral.300}"},contrast:{value:"{colors.rasaNeutral.700}"}},copilot:{solid:{value:"{colors.rasawebDeepPurple.800}"},fg:{value:"{colors.rasawebDeepPurple.800}"},contrast:{value:"{colors.rasaNeutral.50}"},subtle:{value:"{colors.rasaNeutral.700}"},muted:{value:"{colors.rasaNeutral.200}"},emphasized:{value:"{colors.rasaNeutral.600}"}}}},recipes:{button:HVt,iconButton:HVt,heading:wHr,text:yHr},slotRecipes:{segmentGroup:EHr,tag:SHr,toast:_Hr,dialog:mHr},breakpoints:{sm:"480px",md:"768px",lg:"1024px",xl:"1280px","2xl":"1536px","3xl":"1920px"},keyframes:{blink:{"0%, 100%":{},"50%":{backgroundColor:"#6B7694",transform:"scale(1.3)"}}}},cssVarsPrefix:"app",preflight:!0}),Edn=sen(fen,kHr),THr=()=>({getToken:Edn.token}),xHr={flowName:"",nodes:[],edges:[],handleInit:()=>null,focusOnNode:()=>null,handleNodesChange:()=>null,handleZoomInClick:()=>null,handleZoomOutClick:()=>null,handleZoomTo100:()=>null,handleFitToCanvasClick:()=>null},Sdn=Q.createContext(xHr),jEe=()=>Q.useContext(Sdn),CHr=({children:n,flowName:r,initNodes:o=[],initEdges:s=[],highlightedEdges:c=[],highlightedNodes:d=[]})=>{const{getToken:h}=THr(),b=h("colors.rasaNeutral.500"),y=h("colors.rasawebDeepPurple.800"),{zoomIn:E,zoomOut:k,fitView:x,zoomTo:O}=$Ze(),[P,D]=Q.useState([]),[$,U]=Q.useState(null),[B,G,K]=d5r(GVt(P,s,d)),[X,ae]=f5r(qVt(s,c,b,y)),oe=Q.useMemo(()=>AHr(B),[B]);Q.useEffect(()=>{ae(qVt(s,c,b,y))},[s,c,ae,b,y]);const he=Q.useCallback((_e,qe)=>{T5r(_e,qe).then(it=>D(it)).catch(it=>{Va(it,{tags:{component:"CanvasContextProvider",action:"initializeNodes"}})})},[]);Q.useEffect(()=>{he(o,s)},[o,s,he]),Q.useEffect(()=>{if(P.length>0){const _e=GVt(P,s,d);G(_e)}},[P,s,G,d]);const ye=Q.useCallback((_e=null,qe)=>{if(!$||!_e)return;const{zoom:it}=$.getViewport(),Xe=qe?200:0,de=_e.position.x+UZe/2,We=_e.position.y+zZe/2;$.setCenter(de,We,{zoom:it,duration:Xe})},[$]);Q.useEffect(()=>{const _e=OEe.debounce(function(){ye(oe??B[0],!0)},500);return globalThis.addEventListener("resize",_e),()=>{globalThis.removeEventListener("resize",_e)}},[oe,B,ye]);const we=Q.useRef(null);Q.useEffect(()=>(we.current=setTimeout(()=>oe?ye(oe,!0):null,0),()=>{we.current&&clearTimeout(we.current)}),[ye,oe]);const ge=()=>{E({duration:200})},Te=()=>{k({duration:200})},xe=()=>{O(1,{duration:200})},Ue=()=>{x({minZoom:.1})},Z=Q.useCallback(_e=>{U(_e),ye(_e.getNodes()?.[0],!1)},[U,ye]),Re={flowName:r,nodes:B,edges:X,handleInit:Z,focusOnNode:ye,handleNodesChange:K,handleZoomInClick:ge,handleZoomOutClick:Te,handleZoomTo100:xe,handleFitToCanvasClick:Ue};return H.jsx(Sdn.Provider,{value:Re,children:n})};function AHr(n){return n.findLast(o=>o.selected)??null}function GVt(n,r,o,s=!0){const c=n.length===1?"startAlone":"start";return n.map(d=>{const b=r.filter(({sourceId:x})=>x===d.id).length===0,y=d.type===Pa.Start,E=o?.some(x=>x.id===d.id);let k;switch(d.type){case Pa.Start:k={...d};break;case Pa.CollectInformation:k={...d};break;case Pa.CustomAction:k={...d};break;case Pa.Logic:k={...d};break;case Pa.Link:k={...d};break;case Pa.Message:k={...d};break;case Pa.Condition:k={...d};break;case Pa.SetSlots:k={...d};break;case Pa.Call:k={...d};break;default:k={...d}}return{id:d.id,position:{x:d.metadata.x,y:d.metadata.y},type:y?c:b?"end":"standard",data:{...k,isReadOnly:s},selected:!!E,selectable:!1}})}function qVt(n,r,o,s){return n.map(c=>{let d="add";c.type===AW.Custom&&(d="custom");const h=r?.includes(c);let b;return h?b=s:b=o,{id:`${c.sourceId}-${c.targetId}`,source:c.sourceId,target:c.targetId,type:d,selected:h??!1,markerEnd:{type:roe.Arrow,width:20,height:20,color:b},data:{isReadOnly:!0},pathOptions:{borderRadius:8}}})}const IHr=n=>{const{data:r,isConnectable:o,sourcePosition:s}=n,c={bg:"#2C9793",color:"#FFFFFF",borderRadius:"9999px",border:"none",borderColor:"transparent",px:"1.5rem",py:"0.5rem",pointerEvents:"all",height:"auto"};return H.jsxs(H.Fragment,{children:[H.jsx(Fi,{size:"xs",tabIndex:0,css:c,"data-testid":"node",children:HZe(r)}),H.jsx(Um,{type:"source",position:s||ni.Bottom,isConnectable:o,style:{bottom:0}})]})},OHr=n=>{const{data:r,isConnectable:o,sourcePosition:s}=n,c={bg:"#2C9793",color:"#FFFFFF",borderRadius:"9999px",border:"none",borderColor:"transparent",boxShadow:"0 0.125rem 0.5rem 0 rgba(0, 0, 0, 0.15)",px:"1.5rem",py:"0.5rem",pointerEvents:"all",height:"auto"};return H.jsxs(H.Fragment,{children:[H.jsx(Fi,{size:"xs",tabIndex:0,css:c,"data-testid":"node",children:HZe(r)}),H.jsx(Um,{type:"source",position:s||ni.Bottom,isConnectable:o})]})},_dn=n=>{const r={bg:"#FFFFFF",color:"#2C3951",borderRadius:"1rem",overflow:"hidden",height:"100%",alignItems:"center",justifyContent:"center",flexDirection:"column",border:"1px solid","@supports (font: -apple-system-body) and (-webkit-appearance: none)":{boxShadow:"none"},...n.css};return H.jsx(Iu,{...n,css:r,children:n.children})},NHr=({type:n})=>{const r=PHr(n),s={borderRadius:"0.5rem",color:RHr(n)[0],p:"0.25rem",bg:r[0],width:"20px",justifyContent:"center"};return H.jsx(Iu,{css:s,children:H.jsx(xa,{icon:LHr(n)})})},RHr=n=>n===Pa.Start?["rasaCyan.800"]:["rasaNeutral.50"],PHr=n=>{switch(n){case Pa.CollectInformation:return["rasaGreen.800"];case Pa.Message:return["rasaBlue.900"];case Pa.CustomAction:return["rasaOrange.600"];case Pa.SetSlots:case Pa.Start:return["rasaCyan.800"];case Pa.Condition:case Pa.Logic:return["rasaPink.900"];case"invalid":return["rasaRed.800"];case Pa.Link:case Pa.Call:default:return["rasawebDeepPurple.800"]}},LHr=n=>{switch(n){case Pa.CollectInformation:return yor;case Pa.Message:return UVe;case Pa.CustomAction:return NK;case Pa.Condition:case Pa.Logic:return Cor;case Pa.SetSlots:return cXe;case Pa.Call:return xor;case Pa.Start:return zor;case"invalid":return For;case Pa.Link:default:return Ztn}},kdn=({node:n})=>{const{data:r}=n,o={flexGrow:1,px:"1rem",alignItems:"center",justifyContent:"flex-start",width:"100%",position:"relative",minWidth:r.type!==Pa.Logic?"11.5rem":void 0},s={ml:"0.5rem",wordBreak:"break-word",display:"-webkit-box","-webkit-line-clamp":"3","-webkit-box-orient":"vertical",overflow:"hidden"};return H.jsxs(Iu,{css:o,children:[H.jsx(NHr,{type:r.type}),H.jsx(Fi,{size:"xs",css:s,children:HZe(r)})]})},DHr=n=>{const{isConnectable:r,targetPosition:o,sourcePosition:s,selected:c}=n,d=xcn(n),h={pointerEvents:"all",borderColor:c?"rasawebDeepPurple.800":"rasaNeutral.500"};return H.jsxs(H.Fragment,{children:[H.jsx(Um,{type:"target",position:o||ni.Top,isConnectable:!1,style:{top:-.25}}),H.jsx(_dn,{tabIndex:0,css:h,"data-testid":"node",children:H.jsx(kdn,{node:d})}),H.jsx(Um,{type:"source",position:s||ni.Bottom,isConnectable:r,style:{bottom:0}}),H.jsx(Um,{type:"source",position:ni.Left,isConnectable:r}),H.jsx(Um,{type:"target",position:ni.Right,isConnectable:r}),H.jsx(Um,{type:"target",position:ni.Left,isConnectable:r}),H.jsx(Um,{type:"source",position:ni.Right,isConnectable:r})]})},MHr=n=>{const{targetPosition:r,sourcePosition:o,selected:s}=n,c=xcn(n),d={pointerEvents:"all",overflow:"visible",borderColor:s?"rasawebDeepPurple.800":"rasaNeutral.500"};return H.jsxs(H.Fragment,{children:[H.jsx(Um,{type:"target",position:r||ni.Top,isConnectable:!1,style:{top:-.25}}),H.jsx(Um,{type:"source",position:o||ni.Bottom,isConnectable:!1,style:{bottom:0}}),H.jsx(Um,{type:"source",position:ni.Left,isConnectable:!1}),H.jsx(Um,{type:"target",position:ni.Right,isConnectable:!1}),H.jsx(Um,{type:"target",position:ni.Left,isConnectable:!1}),H.jsx(Um,{type:"source",position:ni.Right,isConnectable:!1}),H.jsx(_dn,{tabIndex:0,css:d,"data-testid":"node",children:H.jsx(kdn,{node:c})})]})},jHr=n=>{const{source:r,sourceX:o,sourceY:s,targetX:c,targetY:d,sourcePosition:h,targetPosition:b,style:y={},markerEnd:E}=n,{nodes:k}=jEe(),[x]=ioe({sourceX:o,sourceY:s,targetX:c,targetY:d,sourcePosition:h,targetPosition:b,borderRadius:0});return k.find(P=>P.id===r)?H.jsx(mU,{path:x,markerEnd:E,style:y}):null},$Hr=({flowName:n})=>{const r={position:"absolute",top:"1rem",left:"1.5rem",height:"auto",flexDirection:"row",maxWidth:"calc(100% - 3rem)"};return H.jsxs($c,{css:r,fontSize:"0.813rem",children:[H.jsx(Kx,{size:"md",children:"Current flow: "}),H.jsx(Fi,{size:"md",overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap",children:n})]})};function VVt(n){return n===Pa.Logic?76.586:184}function FHr({sourceNode:n,targetNode:r,sourceX:o,sourceY:s,targetX:c,targetY:d}){let h=ni.Bottom,b=ni.Bottom,y=o,E=s,k=c,x=d;const O=56,P=VVt(n.type),D=VVt(r.type);return s>d&&o>c&&(h=ni.Right,b=ni.Right,x=d+O/2,k=c+D/2,E=s-O/2,y=o+P/2),s<d&&o>c&&(h=ni.Bottom,b=ni.Right,E=s,x=d+O/2,k=c+D/2),s>d&&o<c&&(h=ni.Bottom,b=ni.Right,x=d+O/2,k=c+D/2),s<d&&o<c&&(h=ni.Bottom,b=ni.Left,x=d+O/2,k=c-D/2),n.metadata.y===r.metadata.y&&y<k&&(h=ni.Right,b=ni.Left,y=o+P/2,E=s-O/2,k=c-D/2,x=d+O/2),n.metadata.y===r.metadata.y&&y>k&&(h=ni.Left,b=ni.Right,y=o-P/2,E=s-O/2,k=c+D/2,x=d+O/2),n.metadata.x===r.metadata.x&&E>x&&(h=ni.Left,b=ni.Left,y=o-P/2,E=s-O/2,k=c-D/2,x=d+O/2),{calculatedSourceX:y,calculatedSourceY:E,calculatedTargetX:k,calculatedTargetY:x,calculatedTargetPosition:b,calculatedSourcePosition:h}}const BHr=n=>{const{source:r,sourceX:o,sourceY:s,target:c,targetX:d,targetY:h,markerEnd:b,selected:y}=n,{nodes:E}=jEe(),O=y?"rasawebDeepPurple.800":"rasaNeutral.500",P=E.find(oe=>oe.id===r),D=E.find(oe=>oe.id===c);if(!P||!D)return null;const{calculatedSourceX:$,calculatedSourceY:U,calculatedTargetX:B,calculatedTargetY:G,calculatedSourcePosition:K,calculatedTargetPosition:X}=FHr({sourceNode:P.data,targetNode:D.data,sourceX:o,sourceY:s,targetX:d,targetY:h}),[ae]=ioe({sourceX:$,sourceY:U,targetX:B,targetY:G,sourcePosition:K,targetPosition:X,borderRadius:8});return H.jsx(mU,{path:ae,markerEnd:b,style:{stroke:O}})},UHr=()=>{const{handleZoomInClick:n,handleZoomOutClick:r,handleZoomTo100:o,handleFitToCanvasClick:s}=jEe(),c=Nh(y=>y.transform[2]),d=new Intl.NumberFormat("en-US",{style:"percent"}),h={position:"absolute",bottom:"1rem",left:"2rem",height:"auto"},b={colorPalette:"dark",variant:"subtle"};return H.jsx(Iu,{css:h,children:H.jsxs(ai,{zIndex:1,bg:"rasaNeutral.50",boxShadow:"0px 2px 8px 0px #00000026",borderRadius:"lg",display:"flex",children:[H.jsx(Wg,{content:"Fit to canvas",showArrow:!0,children:H.jsx(Gm,{...b,"aria-label":"Fit to canvas",onClick:s,children:H.jsx(xa,{icon:Lor})})}),H.jsx(Wg,{content:"Zoom in",showArrow:!0,children:H.jsx(Gm,{...b,"aria-label":"Zoom in",onClick:n,children:H.jsx(xa,{icon:Por})})}),H.jsx(Wg,{content:"Zoom out",showArrow:!0,children:H.jsx(Gm,{...b,"aria-label":"Zoom out",onClick:r,children:H.jsx(xa,{icon:enn})})}),H.jsx(Wg,{content:"Zoom to 100%",showArrow:!0,children:H.jsx(ud,{...b,pr:"0.5rem",pl:"0.25rem","aria-label":"Zoom to 100%",onClick:o,children:d.format(c)})})]})})},zHr={standard:DHr,start:IHr,startAlone:OHr,end:MHr},HHr={add:jHr,custom:BHr},GHr=()=>{const n={position:"relative",width:"100%",height:"100%",fontSize:"0.75rem","& .react-flow__panel":{margin:0},"& .react-flow__node-start":{alignItems:"flex-end"},"& .react-flow__node-end":{justifyContent:"flex-start"},"& .react-flow__node":{width:`${UZe}px`,height:`${zZe}px`,display:"flex",justifyContent:"center",zIndex:"-1 !important"},"& .react-flow__node-startAlone":{height:"auto"},"& .react-flow__edges":{zIndex:"-1 !important"},"& .react-flow__edge-add":{stroke:"rasaNeutral.500"},"& .react-flow__edge-path":{stroke:"rasaNeutral.500"},"& .react-flow__handle":{visibility:"hidden"}},{flowName:r,nodes:o,edges:s,handleInit:c,handleNodesChange:d}=jEe();return H.jsxs(ai,{css:n,"data-testid":"canvas",children:[H.jsx(ycn,{nodes:o,edges:s,onInit:c,onNodesChange:d,nodesDraggable:!1,nodesConnectable:!1,nodeTypes:zHr,edgeTypes:HHr,proOptions:{hideAttribution:!0},deleteKeyCode:null,selectionKeyCode:null,multiSelectionKeyCode:null,minZoom:.1,maxZoom:5}),H.jsx($Hr,{flowName:r}),H.jsx(UHr,{})]})};function WVt(n,r,o){const s=[[r]],c=new Set;for(;s.length>0;){const d=s.shift(),h=d[d.length-1];if(h===o)return d;if(!c.has(h)){c.add(h);const b=n.filter(y=>y.sourceId===h).map(y=>y.targetId);for(const y of b)s.push([...d,y])}}return null}function KVt(n,r){return n.find(o=>o.id===r.sourceId)}function YVt(n,r){return n.find(o=>o.id===r.targetId)}function qHr(n,r){const o=TIr(n)[r.id||""]||[],s=new Set,c=[],d=[];let h=o.length>0?o[0]:null;for(let b=0;b<o.length-1;b++){const y=o[b],E=o[b+1],k=`${y}->${E}`;if(s.has(k))continue;const x=r.edges?.find(P=>P.sourceId===y&&P.targetId===E);if(x){s.add(k),c.push(x);const P=KVt(r.nodes||[],x);P&&d.push(P);const D=YVt(r.nodes||[],x);D&&d.push(D),h=E;continue}let O=WVt(r.edges||[],y,E);if(!(!O&&(!h||(O=WVt(r.edges||[],h,E),!O)))){s.add(k);for(let P=0;P<O.length-1;P++){const D=r.edges?.find($=>$.sourceId===O[P]&&$.targetId===O[P+1]);if(D){const $=`${D.sourceId}->${D.targetId}`;if(!s.has($)){s.add($),c.push(D);const U=KVt(r.nodes||[],D);U&&d.push(U);const B=YVt(r.nodes||[],D);B&&d.push(B)}h=O[P+1]}}}}return{highlightedEdges:c,highlightedNodes:d}}function VHr({flow:n,conversation:r}){const{highlightedEdges:o,highlightedNodes:s}=Q.useMemo(()=>n?qHr(r?.events??[],n):{highlightedEdges:[],highlightedNodes:[]},[r?.events,n]);return H.jsx(BZe,{children:H.jsx(CHr,{flowName:n.name||n.id,initNodes:n?.nodes,initEdges:n?.edges,highlightedEdges:o,highlightedNodes:s,children:H.jsx(GHr,{})})})}const WHr=({stackToShow:n,conversationList:r,conversationForSelectedElement:o})=>{const{flows:s,isLoading:c,error:d}=gHr(),h=n?.flowId?s?.find(b=>b.id===n?.flowId):void 0;return c?H.jsx(Bx,{image:zm.NoFlow,shortLabel:"Loading flows...",longLabel:"Please wait while we load the conversation flows"}):d?(Va(d,{tags:{component:"TryAssistant",action:"getConversationData"}}),H.jsx(Bx,{image:zm.NoFlow,shortLabel:"Failed to load flows",longLabel:"Error while obtaining trained flow data"})):n?.flowId?h?H.jsx(VHr,{flow:h,selectedNodeId:n?.stepId,conversation:o||r[r.length-1]}):H.jsx(Bx,{image:zm.NoFlow,shortLabel:"This flow isn’t available anymore",longLabel:`"${n?.flowId}" was removed, so it can’t be previewed in this version`}):H.jsx(Bx,{image:zm.NoFlow,shortLabel:"No flow is currently active",longLabel:"Type a message to talk to the latest trained version of your assistant"})},KHr=new Set(["flow_hashes","session_started_metadata","mocked_datetime","interrupted_flow_to_continue","continue_interrupted_flow_confirmation","max_clarification_options","continue_conversation","confirm_correction","example_questions","agent_name","consecutive_silence_timeouts","silence_timeout"]),YHr=n=>n.startsWith("pattern_")&&n!=="pattern_session_start"&&n!=="pattern_completed",XHr=({slots:n,flowHistory:r})=>{const o=Q.useMemo(()=>n.filter(c=>!KHr.has(c.name)),[n]),s=Q.useMemo(()=>r.filter(c=>!YHr(c.flowId)),[r]);return H.jsxs(Vd,{height:"100%",width:"100%",gap:0,alignItems:"stretch",overflow:"hidden",children:[H.jsxs(ai,{flex:1,overflow:"auto",borderBottom:"1px solid",borderColor:"rasaNeutral.300",children:[H.jsx(ai,{px:4,py:3,borderBottom:"1px solid",borderColor:"rasaNeutral.200",children:H.jsx(Fi,{fontWeight:"semibold",fontSize:"md",color:"rasaNeutral.900",children:"History"})}),H.jsx(Vd,{gap:0,px:2,py:2,alignItems:"stretch",children:s.length===0?H.jsx(Fi,{fontSize:"sm",color:"rasaNeutral.600",px:2,py:4,children:"No flow history yet. Start a conversation to see flow execution."}):s.map((c,d)=>H.jsx(JHr,{entry:c},`${c.flowId}-${d}`))})]}),H.jsxs(ai,{flex:1,overflow:"auto",children:[H.jsx(ai,{px:4,py:3,borderBottom:"1px solid",borderColor:"rasaNeutral.200",children:H.jsx(Fi,{fontWeight:"semibold",fontSize:"md",color:"rasaNeutral.900",children:"Context"})}),H.jsx(Vd,{gap:0,alignItems:"stretch",px:4,py:2,children:o.map(c=>H.jsx(ZHr,{name:c.name,value:c.value},c.name))})]})]})},JHr=({entry:n})=>{const r=n.status==="active",o=r?"rasaGreen.800":"rasaNeutral.500",c=new Date(n.startedAt*1e3).toLocaleTimeString([],{hour:"numeric",minute:"2-digit"}),d=h=>h>=1e3?`${(h/1e3).toFixed(2)}s`:`${Math.round(h)}ms`;return H.jsxs($c,{px:2,py:2,gap:2,_hover:{bg:"rasaNeutral.50"},children:[H.jsx(ai,{width:"8px",height:"8px",borderRadius:"full",bg:o,flexShrink:0,mt:1,alignSelf:"flex-start"}),H.jsxs(Vd,{alignItems:"flex-start",gap:0,flex:1,children:[H.jsx(Fi,{fontSize:"sm",fontWeight:"medium",color:"rasaNeutral.900",children:QHr(n.flowName)}),H.jsxs(Fi,{fontSize:"xs",color:"rasaNeutral.600",children:[r?"Active":"Completed",!r&&n.startedAt&&` · Started at ${c}`,n.latencyMs>0&&H.jsxs(H.Fragment,{children:[" · Avg latency: ",H.jsx(Fi,{as:"span",color:"rasaOrange.700",fontWeight:"medium",children:d(n.latencyMs)})]}),n.tokens>0&&` · Tokens: ${n.tokens}`]})]})]})},ZHr=({name:n,value:r})=>{const o=tGr(r);return H.jsxs(Iu,{py:2,gap:4,borderBottom:"1px solid",borderColor:"rasaNeutral.100",children:[H.jsx(Fi,{fontSize:"sm",color:"rasaNeutral.600",minWidth:"140px",flexShrink:0,children:eGr(n)}),H.jsx(Fi,{fontSize:"sm",color:"rasaNeutral.900",wordBreak:"break-word",children:o})]})};function QHr(n){return n.replace(/_/g," ").replace(/\b\w/g,r=>r.toUpperCase())}function eGr(n){return n.replace(/_/g," ").replace(/\b\w/g,r=>r.toUpperCase())}function tGr(n){if(n==null)return"—";if(typeof n=="boolean")return n?"Yes":"No";if(typeof n=="number")return String(n);if(typeof n=="string")return n.includes(`
`)?n.split(`
`).map(r=>r.replace(/^-\s*/,"").trim()).filter(Boolean).join(`
`):n;if(typeof n=="object")try{return JSON.stringify(n,null,2)}catch{return"[Object]"}return typeof n=="symbol"||typeof n=="bigint"?n.toString():"[Unknown]"}function Tdn({flowView:n,sendMessage:r,setUrl:o,startNewConversation:s,conversationList:c,inputDisabled:d,sessionId:h,stack:b,slots:y,flowHistory:E,replayingConversation:k,waitingForUserInput:x,replayConversation:O}){const[P,D]=Q.useState(),$=Io(da,xe=>xe.projectUrl),U=["pattern_session_start","pattern_completed"],B=xe=>!xe.flowId?.startsWith("pattern_")||U.includes(xe.flowId),G=b.slice().reverse().find(B),[K,X]=Q.useState(),ae=K?.id?c.find(xe=>xe.events.some(Ue=>Ue?.id===K.id)):void 0,oe=xe=>{X(void 0),r(xe)},he=xe=>{xe.id===K?.id?X(void 0):X(xe)},ye=xe=>{const Ue=c.find(Re=>Re.events.some(_e=>_e?.id===xe));if(!Ue)return;const Z=Ue?.events.slice(0,Ue.events.findIndex(Re=>Re.id===xe)+1);O(Z)},we=Q.useMemo(()=>d,[d]),Te={borderLeft:"1px solid",borderColor:"rasaNeutral.400"};return Q.useEffect(()=>{$&&o($)},[$,o]),Q.useEffect(()=>{G&&D(G)},[G]),Q.useEffect(()=>{if(K?.metadata?.flow_id&&K?.metadata?.step_id){const xe={frameId:K.id,flowId:K.metadata.flow_id,stepId:K.metadata.step_id,ended:!1};D(xe)}else K===void 0&&G&&D(G)},[K,G]),we?H.jsx(qln,{}):H.jsxs(Iu,{flexGrow:"1","data-testid":"try-assistant-container",children:[H.jsx(ai,{flexBasis:n?"35%":"100%",minWidth:0,children:H.jsx(uHr,{inspectorMode:n,sessionId:h,replayConversation:ye,conversationAssistants:{},conversationList:c,handleMessageSubmit:oe,handleSelect:he,selectedElement:K,inputDisabled:d,waitingForResponse:!x&&!d,replayingConversation:k,onNewConversation:s})}),n&&H.jsxs(H.Fragment,{children:[H.jsx(ai,{css:Te,flexBasis:"40%",minWidth:0,"data-testid":"inspector-canvas",children:H.jsx(WHr,{stackToShow:P,conversationList:c,conversationForSelectedElement:ae})}),H.jsx(ai,{css:Te,flexBasis:"25%",minWidth:0,"data-testid":"tracker-panel",children:H.jsx(XHr,{slots:y,flowHistory:E})})]})]})}function nGr(n,r,o){return r in n?Object.defineProperty(n,r,{value:o,enumerable:!0,configurable:!0,writable:!0}):n[r]=o,n}function XVt(n,r){var o=Object.keys(n);if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(n);r&&(s=s.filter(function(c){return Object.getOwnPropertyDescriptor(n,c).enumerable})),o.push.apply(o,s)}return o}function JVt(n){for(var r=1;r<arguments.length;r++){var o=arguments[r]!=null?arguments[r]:{};r%2?XVt(Object(o),!0).forEach(function(s){nGr(n,s,o[s])}):Object.getOwnPropertyDescriptors?Object.defineProperties(n,Object.getOwnPropertyDescriptors(o)):XVt(Object(o)).forEach(function(s){Object.defineProperty(n,s,Object.getOwnPropertyDescriptor(o,s))})}return n}function rGr(n,r){if(n==null)return{};var o={},s=Object.keys(n),c,d;for(d=0;d<s.length;d++)c=s[d],!(r.indexOf(c)>=0)&&(o[c]=n[c]);return o}function iGr(n,r){if(n==null)return{};var o=rGr(n,r),s,c;if(Object.getOwnPropertySymbols){var d=Object.getOwnPropertySymbols(n);for(c=0;c<d.length;c++)s=d[c],!(r.indexOf(s)>=0)&&Object.prototype.propertyIsEnumerable.call(n,s)&&(o[s]=n[s])}return o}function oGr(n,r){return aGr(n)||sGr(n,r)||cGr(n,r)||uGr()}function aGr(n){if(Array.isArray(n))return n}function sGr(n,r){if(!(typeof Symbol>"u"||!(Symbol.iterator in Object(n)))){var o=[],s=!0,c=!1,d=void 0;try{for(var h=n[Symbol.iterator](),b;!(s=(b=h.next()).done)&&(o.push(b.value),!(r&&o.length===r));s=!0);}catch(y){c=!0,d=y}finally{try{!s&&h.return!=null&&h.return()}finally{if(c)throw d}}return o}}function cGr(n,r){if(n){if(typeof n=="string")return ZVt(n,r);var o=Object.prototype.toString.call(n).slice(8,-1);if(o==="Object"&&n.constructor&&(o=n.constructor.name),o==="Map"||o==="Set")return Array.from(n);if(o==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(o))return ZVt(n,r)}}function ZVt(n,r){(r==null||r>n.length)&&(r=n.length);for(var o=0,s=new Array(r);o<r;o++)s[o]=n[o];return s}function uGr(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function lGr(n,r,o){return r in n?Object.defineProperty(n,r,{value:o,enumerable:!0,configurable:!0,writable:!0}):n[r]=o,n}function QVt(n,r){var o=Object.keys(n);if(Object.getOwnPropertySymbols){var s=Object.getOwnPropertySymbols(n);r&&(s=s.filter(function(c){return Object.getOwnPropertyDescriptor(n,c).enumerable})),o.push.apply(o,s)}return o}function eWt(n){for(var r=1;r<arguments.length;r++){var o=arguments[r]!=null?arguments[r]:{};r%2?QVt(Object(o),!0).forEach(function(s){lGr(n,s,o[s])}):Object.getOwnPropertyDescriptors?Object.defineProperties(n,Object.getOwnPropertyDescriptors(o)):QVt(Object(o)).forEach(function(s){Object.defineProperty(n,s,Object.getOwnPropertyDescriptor(o,s))})}return n}function dGr(){for(var n=arguments.length,r=new Array(n),o=0;o<n;o++)r[o]=arguments[o];return function(s){return r.reduceRight(function(c,d){return d(c)},s)}}function Zre(n){return function r(){for(var o=this,s=arguments.length,c=new Array(s),d=0;d<s;d++)c[d]=arguments[d];return c.length>=n.length?n.apply(this,c):function(){for(var h=arguments.length,b=new Array(h),y=0;y<h;y++)b[y]=arguments[y];return r.apply(o,[].concat(c,b))}}}function Vwe(n){return{}.toString.call(n).includes("Object")}function fGr(n){return!Object.keys(n).length}function hoe(n){return typeof n=="function"}function hGr(n,r){return Object.prototype.hasOwnProperty.call(n,r)}function pGr(n,r){return Vwe(r)||SM("changeType"),Object.keys(r).some(function(o){return!hGr(n,o)})&&SM("changeField"),r}function gGr(n){hoe(n)||SM("selectorType")}function mGr(n){hoe(n)||Vwe(n)||SM("handlerType"),Vwe(n)&&Object.values(n).some(function(r){return!hoe(r)})&&SM("handlersType")}function bGr(n){n||SM("initialIsRequired"),Vwe(n)||SM("initialType"),fGr(n)&&SM("initialContent")}function vGr(n,r){throw new Error(n[r]||n.default)}var wGr={initialIsRequired:"initial state is required",initialType:"initial state should be an object",initialContent:"initial state shouldn't be an empty object",handlerType:"handler should be an object or a function",handlersType:"all handlers should be a functions",selectorType:"selector should be a function",changeType:"provided value of changes should be an object",changeField:'it seams you want to change a field in the state which is not specified in the "initial" state',default:"an unknown error accured in `state-local` package"},SM=Zre(vGr)(wGr),U0e={changes:pGr,selector:gGr,handler:mGr,initial:bGr};function yGr(n){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};U0e.initial(n),U0e.handler(r);var o={current:n},s=Zre(_Gr)(o,r),c=Zre(SGr)(o),d=Zre(U0e.changes)(n),h=Zre(EGr)(o);function b(){var E=arguments.length>0&&arguments[0]!==void 0?arguments[0]:function(k){return k};return U0e.selector(E),E(o.current)}function y(E){dGr(s,c,d,h)(E)}return[b,y]}function EGr(n,r){return hoe(r)?r(n.current):r}function SGr(n,r){return n.current=eWt(eWt({},n.current),r),r}function _Gr(n,r,o){return hoe(r)?r(n.current):Object.keys(o).forEach(function(s){var c;return(c=r[s])===null||c===void 0?void 0:c.call(r,n.current[s])}),o}var kGr={create:yGr},TGr={paths:{vs:"https://cdn.jsdelivr.net/npm/monaco-editor@0.52.2/min/vs"}};function xGr(n){return function r(){for(var o=this,s=arguments.length,c=new Array(s),d=0;d<s;d++)c[d]=arguments[d];return c.length>=n.length?n.apply(this,c):function(){for(var h=arguments.length,b=new Array(h),y=0;y<h;y++)b[y]=arguments[y];return r.apply(o,[].concat(c,b))}}}function CGr(n){return{}.toString.call(n).includes("Object")}function AGr(n){return n||tWt("configIsRequired"),CGr(n)||tWt("configType"),n.urls?(IGr(),{paths:{vs:n.urls.monacoBase}}):n}function IGr(){console.warn(xdn.deprecation)}function OGr(n,r){throw new Error(n[r]||n.default)}var xdn={configIsRequired:"the configuration object is required",configType:"the configuration object should be an object",default:"an unknown error accured in `@monaco-editor/loader` package",deprecation:`Deprecation warning!
    You are using deprecated way of configuration.

    Instead of using
      monaco.config({ urls: { monacoBase: '...' } })
    use
      monaco.config({ paths: { vs: '...' } })

    For more please check the link https://github.com/suren-atoyan/monaco-loader#config
  `},tWt=xGr(OGr)(xdn),NGr={config:AGr},RGr=function(){for(var r=arguments.length,o=new Array(r),s=0;s<r;s++)o[s]=arguments[s];return function(c){return o.reduceRight(function(d,h){return h(d)},c)}};function Cdn(n,r){return Object.keys(r).forEach(function(o){r[o]instanceof Object&&n[o]&&Object.assign(r[o],Cdn(n[o],r[o]))}),JVt(JVt({},n),r)}var PGr={type:"cancelation",msg:"operation is manually canceled"};function rqe(n){var r=!1,o=new Promise(function(s,c){n.then(function(d){return r?c(PGr):s(d)}),n.catch(c)});return o.cancel=function(){return r=!0},o}var LGr=kGr.create({config:TGr,isInitialized:!1,resolve:null,reject:null,monaco:null}),Adn=oGr(LGr,2),Yoe=Adn[0],$Ee=Adn[1];function DGr(n){var r=NGr.config(n),o=r.monaco,s=iGr(r,["monaco"]);$Ee(function(c){return{config:Cdn(c.config,s),monaco:o}})}function MGr(){var n=Yoe(function(r){var o=r.monaco,s=r.isInitialized,c=r.resolve;return{monaco:o,isInitialized:s,resolve:c}});if(!n.isInitialized){if($Ee({isInitialized:!0}),n.monaco)return n.resolve(n.monaco),rqe(iqe);if(window.monaco&&window.monaco.editor)return Idn(window.monaco),n.resolve(window.monaco),rqe(iqe);RGr(jGr,FGr)(BGr)}return rqe(iqe)}function jGr(n){return document.body.appendChild(n)}function $Gr(n){var r=document.createElement("script");return n&&(r.src=n),r}function FGr(n){var r=Yoe(function(s){var c=s.config,d=s.reject;return{config:c,reject:d}}),o=$Gr("".concat(r.config.paths.vs,"/loader.js"));return o.onload=function(){return n()},o.onerror=r.reject,o}function BGr(){var n=Yoe(function(o){var s=o.config,c=o.resolve,d=o.reject;return{config:s,resolve:c,reject:d}}),r=window.require;r.config(n.config),r(["vs/editor/editor.main"],function(o){Idn(o),n.resolve(o)},function(o){n.reject(o)})}function Idn(n){Yoe().monaco||$Ee({monaco:n})}function UGr(){return Yoe(function(n){var r=n.monaco;return r})}var iqe=new Promise(function(n,r){return $Ee({resolve:n,reject:r})}),Odn={config:DGr,init:MGr,__getMonacoInstance:UGr},zGr={wrapper:{display:"flex",position:"relative",textAlign:"initial"},fullWidth:{width:"100%"},hide:{display:"none"}},oqe=zGr,HGr={container:{display:"flex",height:"100%",width:"100%",justifyContent:"center",alignItems:"center"}},GGr=HGr;function qGr({children:n}){return Un.createElement("div",{style:GGr.container},n)}var VGr=qGr,WGr=VGr;function KGr({width:n,height:r,isEditorReady:o,loading:s,_ref:c,className:d,wrapperProps:h}){return Un.createElement("section",{style:{...oqe.wrapper,width:n,height:r},...h},!o&&Un.createElement(WGr,null,s),Un.createElement("div",{ref:c,style:{...oqe.fullWidth,...!o&&oqe.hide},className:d}))}var YGr=KGr,Ndn=Q.memo(YGr);function XGr(n){Q.useEffect(n,[])}var Rdn=XGr;function JGr(n,r,o=!0){let s=Q.useRef(!0);Q.useEffect(s.current||!o?()=>{s.current=!1}:n,r)}var r2=JGr;function wie(){}function bW(n,r,o,s){return ZGr(n,s)||QGr(n,r,o,s)}function ZGr(n,r){return n.editor.getModel(Pdn(n,r))}function QGr(n,r,o,s){return n.editor.createModel(r,o,s?Pdn(n,s):void 0)}function Pdn(n,r){return n.Uri.parse(r)}function eqr({original:n,modified:r,language:o,originalLanguage:s,modifiedLanguage:c,originalModelPath:d,modifiedModelPath:h,keepCurrentOriginalModel:b=!1,keepCurrentModifiedModel:y=!1,theme:E="light",loading:k="Loading...",options:x={},height:O="100%",width:P="100%",className:D,wrapperProps:$={},beforeMount:U=wie,onMount:B=wie}){let[G,K]=Q.useState(!1),[X,ae]=Q.useState(!0),oe=Q.useRef(null),he=Q.useRef(null),ye=Q.useRef(null),we=Q.useRef(B),ge=Q.useRef(U),Te=Q.useRef(!1);Rdn(()=>{let Re=Odn.init();return Re.then(_e=>(he.current=_e)&&ae(!1)).catch(_e=>_e?.type!=="cancelation"&&console.error("Monaco initialization: error:",_e)),()=>oe.current?Z():Re.cancel()}),r2(()=>{if(oe.current&&he.current){let Re=oe.current.getOriginalEditor(),_e=bW(he.current,n||"",s||o||"text",d||"");_e!==Re.getModel()&&Re.setModel(_e)}},[d],G),r2(()=>{if(oe.current&&he.current){let Re=oe.current.getModifiedEditor(),_e=bW(he.current,r||"",c||o||"text",h||"");_e!==Re.getModel()&&Re.setModel(_e)}},[h],G),r2(()=>{let Re=oe.current.getModifiedEditor();Re.getOption(he.current.editor.EditorOption.readOnly)?Re.setValue(r||""):r!==Re.getValue()&&(Re.executeEdits("",[{range:Re.getModel().getFullModelRange(),text:r||"",forceMoveMarkers:!0}]),Re.pushUndoStop())},[r],G),r2(()=>{oe.current?.getModel()?.original.setValue(n||"")},[n],G),r2(()=>{let{original:Re,modified:_e}=oe.current.getModel();he.current.editor.setModelLanguage(Re,s||o||"text"),he.current.editor.setModelLanguage(_e,c||o||"text")},[o,s,c],G),r2(()=>{he.current?.editor.setTheme(E)},[E],G),r2(()=>{oe.current?.updateOptions(x)},[x],G);let xe=Q.useCallback(()=>{if(!he.current)return;ge.current(he.current);let Re=bW(he.current,n||"",s||o||"text",d||""),_e=bW(he.current,r||"",c||o||"text",h||"");oe.current?.setModel({original:Re,modified:_e})},[o,r,c,n,s,d,h]),Ue=Q.useCallback(()=>{!Te.current&&ye.current&&(oe.current=he.current.editor.createDiffEditor(ye.current,{automaticLayout:!0,...x}),xe(),he.current?.editor.setTheme(E),K(!0),Te.current=!0)},[x,E,xe]);Q.useEffect(()=>{G&&we.current(oe.current,he.current)},[G]),Q.useEffect(()=>{!X&&!G&&Ue()},[X,G,Ue]);function Z(){let Re=oe.current?.getModel();b||Re?.original?.dispose(),y||Re?.modified?.dispose(),oe.current?.dispose()}return Un.createElement(Ndn,{width:P,height:O,isEditorReady:G,loading:k,_ref:ye,className:D,wrapperProps:$})}var tqr=eqr,nqr=Q.memo(tqr);function rqr(n){let r=Q.useRef();return Q.useEffect(()=>{r.current=n},[n]),r.current}var iqr=rqr,z0e=new Map;function oqr({defaultValue:n,defaultLanguage:r,defaultPath:o,value:s,language:c,path:d,theme:h="light",line:b,loading:y="Loading...",options:E={},overrideServices:k={},saveViewState:x=!0,keepCurrentModel:O=!1,width:P="100%",height:D="100%",className:$,wrapperProps:U={},beforeMount:B=wie,onMount:G=wie,onChange:K,onValidate:X=wie}){let[ae,oe]=Q.useState(!1),[he,ye]=Q.useState(!0),we=Q.useRef(null),ge=Q.useRef(null),Te=Q.useRef(null),xe=Q.useRef(G),Ue=Q.useRef(B),Z=Q.useRef(),Re=Q.useRef(s),_e=iqr(d),qe=Q.useRef(!1),it=Q.useRef(!1);Rdn(()=>{let We=Odn.init();return We.then(ht=>(we.current=ht)&&ye(!1)).catch(ht=>ht?.type!=="cancelation"&&console.error("Monaco initialization: error:",ht)),()=>ge.current?de():We.cancel()}),r2(()=>{let We=bW(we.current,n||s||"",r||c||"",d||o||"");We!==ge.current?.getModel()&&(x&&z0e.set(_e,ge.current?.saveViewState()),ge.current?.setModel(We),x&&ge.current?.restoreViewState(z0e.get(d)))},[d],ae),r2(()=>{ge.current?.updateOptions(E)},[E],ae),r2(()=>{!ge.current||s===void 0||(ge.current.getOption(we.current.editor.EditorOption.readOnly)?ge.current.setValue(s):s!==ge.current.getValue()&&(it.current=!0,ge.current.executeEdits("",[{range:ge.current.getModel().getFullModelRange(),text:s,forceMoveMarkers:!0}]),ge.current.pushUndoStop(),it.current=!1))},[s],ae),r2(()=>{let We=ge.current?.getModel();We&&c&&we.current?.editor.setModelLanguage(We,c)},[c],ae),r2(()=>{b!==void 0&&ge.current?.revealLine(b)},[b],ae),r2(()=>{we.current?.editor.setTheme(h)},[h],ae);let Xe=Q.useCallback(()=>{if(!(!Te.current||!we.current)&&!qe.current){Ue.current(we.current);let We=d||o,ht=bW(we.current,s||n||"",r||c||"",We||"");ge.current=we.current?.editor.create(Te.current,{model:ht,automaticLayout:!0,...E},k),x&&ge.current.restoreViewState(z0e.get(We)),we.current.editor.setTheme(h),b!==void 0&&ge.current.revealLine(b),oe(!0),qe.current=!0}},[n,r,o,s,c,d,E,k,x,h,b]);Q.useEffect(()=>{ae&&xe.current(ge.current,we.current)},[ae]),Q.useEffect(()=>{!he&&!ae&&Xe()},[he,ae,Xe]),Re.current=s,Q.useEffect(()=>{ae&&K&&(Z.current?.dispose(),Z.current=ge.current?.onDidChangeModelContent(We=>{it.current||K(ge.current.getValue(),We)}))},[ae,K]),Q.useEffect(()=>{if(ae){let We=we.current.editor.onDidChangeMarkers(ht=>{let Ie=ge.current.getModel()?.uri;if(Ie&&ht.find($t=>$t.path===Ie.path)){let $t=we.current.editor.getModelMarkers({resource:Ie});X?.($t)}});return()=>{We?.dispose()}}return()=>{}},[ae,X]);function de(){Z.current?.dispose(),O?x&&z0e.set(d,ge.current.saveViewState()):ge.current.getModel()?.dispose(),ge.current.dispose()}return Un.createElement(Ndn,{width:P,height:D,isEditorReady:ae,loading:y,_ref:Te,className:$,wrapperProps:U})}var aqr=oqr,sqr=Q.memo(aqr);function cqr(n,r="",o=!1){return{id:r?`${r}/${n}`:n,name:n,isDirectory:o,isTemporary:!0,isEditable:!0,parentPath:r,children:o?[]:void 0}}function uqr(n){const r=!n.isDirectory,o=gqr(n.path);return{id:n.path,name:o,file:r?n:void 0,isDirectory:!r,children:r?void 0:[],isEditable:!1,deleted:n.deleted}}function lqr(n){const r=new Map;for(const o of n)r.set(o.path,uqr(o));return r}function dqr(n){const r=[];for(const[o,s]of n){const c=o.split("/");if(c.length===1)r.push({...s,isRoot:!0});else{const d=c.slice(0,-1).join("/"),h=n.get(d);h?.children&&h.children.push(s)}}return r}function Ldn(n){return n.toSorted((r,o)=>r.isDirectory&&!o.isDirectory?-1:!r.isDirectory&&o.isDirectory?1:r.name.localeCompare(o.name)).map(r=>({...r,children:r.children?Ldn(r.children):void 0}))}function fqr(n,r=[],o){const s=lqr(n);for(const d of r){const h=d.split("/");let b="";for(const y of h)if(b=b?`${b}/${y}`:y,!s.has(b)){const E=n.some(x=>x.path.startsWith(b+"/")||x.path===b)&&n.every(x=>x.path.startsWith(b+"/")||x.path===b?x.deleted:!1),k={id:b,name:y,isDirectory:!0,children:[],isEditable:!1,deleted:E};s.set(b,k)}}o&&s.set(o.id,o);const c=dqr(s);return Ldn(c)}function dK(n){if(!n)return[];const r=n.split("/").slice(0,-1);return r.map((o,s)=>r.slice(0,s+1).join("/"))}function hqr(n){const r=n.split("/");return r.pop(),r.join("/")}function pqr(n){return n?n.isDirectory?n.id:hqr(n.id):""}function gqr(n){const r=n.split("/");return r.length===0?"":r.length===1?r[0]:r.pop()}const mqr=()=>{const n=Io(da,h=>h.projectUrl),{getIdTokenClaims:r,isAuthenticated:o}=ty(),{mutate:s}=BYe({mutationFn:async({filesSnapshot:h})=>{if(!n)throw new Error("No project URL available");if(!h||h.length===0)throw new Error("No files to apply");Y$e(!0);const b=await vqr({getIdTokenClaims:r,isAuthenticated:o}),y=h.filter(k=>!k.deleted)||[],E={};for(const k of y)k.isDirectory||(E[k.path]=k.content);return await Ehr({projectUrl:n,files:E,token:b})},onSuccess:(h,{filesSnapshot:b})=>{const{job_id:y,status:E}=h;if(!y)throw new Error("Job ID is missing in the response");if(!E)throw new Error("Job status is missing in the response");const k=bqr(b);ePr(k),P5r(y,E),Y$e(!1)},onError:h=>{Va(h,{tags:{component:"useFileOperations",action:"postFiles"}}),hie(h instanceof Error?h.message:"Failed to post files"),Y$e(!1)}}),c=Io(gc,h=>h.error),d=Io(gc,PK);return{postFiles:s,loading:d,error:c}};function bqr(n){const r=n?.filter(d=>!d.deleted)||[],o={};for(const d of r){if(d.isDirectory)continue;const h=dK(d.path);for(const b of h)o[b]=!0}const s=new Set(r.map(d=>d.path)),c=Object.keys(o).filter(d=>!s.has(d)).map(d=>({path:d,content:"",edited:!1,hasErrors:!1,isDirectory:!0,isTemporary:!1}));return[...r,...c]}async function vqr({getIdTokenClaims:n,isAuthenticated:r}){let o;try{r&&(o=(await n())?.__raw)}catch(s){r&&Va(s,{tags:{component:"useFileOperations",action:"getAuthToken"}})}return o}const FEe=()=>{const n=Io(da,x=>x.projectUrl),r=Io(da,x=>x.loadingProject||x.loading||x.isRestoring),o=Io(gc,x=>x.filesInitialized),{data:s,error:c,isFetching:d,refetch:h}=xK({queryKey:["files",n],queryFn:()=>vhr({projectUrl:n}),select:wqr,enabled:!!n&&!o&&!r,refetchOnWindowFocus:!1,retry:3});Q.useEffect(()=>{d&&hie(null),nPr(d)},[d]),Q.useEffect(()=>{s&&!o&&!d&&tPr(s)},[s,o,d]);const b=Q.useCallback(async()=>($cn(!1),h()),[h]);Q.useEffect(()=>{if(!c){hie(null);return}const x=c instanceof Error?c.message:"Failed to fetch files";hie(x),Va(c,{tags:{component:"useFiles",action:"useFiles"}})},[c]);const y=Io(gc,x=>x.files),E=Io(gc,x=>x.error),k=Io(gc,PK);return{files:y,loading:k,error:E,getFiles:b}};function wqr(n){if(!n)return null;const r={},o=[];for(const[s,c]of Object.entries(n)){o.push({path:s,content:c||"",edited:!1,hasErrors:!1,isDirectory:!1,isTemporary:!1});const d=dK(s);for(const h of d)r[h]||(r[h]={path:h,content:"",edited:!1,hasErrors:!1,isDirectory:!0,isTemporary:!1})}for(const[s,c]of Object.entries(r))o.push(c);return o}function yqr(n,r){if(n.isDirectory){const o=n.id.substring(0,n.id.lastIndexOf("/"));return o?`${o}/${r}`:r}else{const o=n.id.split("/").slice(0,-1).join("/");return o?`${o}/${r}`:r}}function Qre(n){Kp.create({title:"Error",description:n,type:"error",duration:4e3})}function Ddn(n){if(n.trim()==="")return"File or folder name cannot be empty.";if(n.startsWith("/"))return"File or folder name cannot start with a slash (/).";if(n.endsWith("/"))return"File or folder name cannot end with a slash (/).";if(n.includes("/"))return"File or folder name cannot contain forward slashes (/).";if(n.includes("\\"))return String.raw`File or folder name cannot contain backslashes (\).`;const r=/[<>:"|?*]/;return r.test(n)?`File or folder name cannot contain the character: ${new RegExp(r).exec(n)?.[0]}`:/[\x00-\x1F]/.test(n)?"File or folder name cannot contain control characters.":null}function Eqr(n,r,o){const s=[".rasa","models",".env",".git"];if(n.trim()==="")return Qre("File or folder name cannot be empty."),!1;if(o&&n.includes("/")){const c=Mdn(n);if(c)return Qre(c),!1}else{const c=Ddn(n);if(c)return Qre(c),!1}return s.includes(n)||Fcn(r,o)?(Qre(`A file or folder ${n} already exists.`),!1):!0}function Sqr(n,r,o,s,c=[]){let d=n,h=1;for(;;){const b=r?`${r}/${d}`:d,y=o.some(x=>x.path===b),E=c.includes(b),k=Fcn(b,s);if(!y&&!E&&!k)return d;h++,d=`${n} ${h}`}}function Mdn(n){if(n.startsWith("/"))return"File path cannot start with a slash (/).";if(n.endsWith("/"))return"File path cannot end with a slash (/).";if(n.includes("//"))return"File path cannot contain consecutive slashes (//).";const r=n.split("/");if(r.includes(""))return"File path cannot contain empty directory or file names.";const o=r.filter(s=>s!=="");if(o.length===0)return"File path cannot be empty.";for(const s of o){const c=Ddn(s);if(c)return c}return null}function _qr(n,r){const o=Mdn(n);if(o)throw Qre(o),new Error(o);const s=dK(n);for(const d of s)r?.some(b=>b.path===d)||Awe({path:d,content:"",edited:!1,hasErrors:!1,deleted:!1,isDirectory:!0,isTemporary:!1});Awe({path:n,content:"",edited:!0,hasErrors:!1,deleted:!1})}const kqr=({onClick:n,disabled:r=!1,loading:o=!1,editorStatus:s=null})=>{const[c,d]=Q.useState(s==="done"||s==="validation_error"||s==="training_error"||s==="error");return Q.useEffect(()=>{d(s==="done"||s==="validation_error"||s==="training_error"||s==="error")},[s]),H.jsx(ud,{colorPalette:"purple",onClick:n,variant:"solid",size:"md",mt:"0.5rem",disabled:r,loading:o,minW:"152px","data-testid":"apply-changes-button",children:c?H.jsxs(H.Fragment,{children:[H.jsx(xa,{icon:uXe}),H.jsx("span",{children:" Applied "})]}):H.jsx("span",{children:" Apply changes "})})},Tqr=({onConfirm:n,isDisabled:r=!1,isLoading:o=!1})=>{const{open:s,onOpen:c,onClose:d}=_M(),h="rasawebDeepPurple.800",b=[{id:"Cancel",title:"Cancel",onClick:()=>d(),"aria-label":"Cancel",variant:"outline",colorPalette:"dark"},{id:"Confirm",title:"Discard changes",onClick:()=>{n(),d()},"aria-label":"Confirm",variant:"solid",colorPalette:"dark"}];return H.jsxs(H.Fragment,{children:[H.jsx(Wg,{positioning:{placement:"top-start"},content:"Discard all changes",showArrow:!0,children:H.jsx(Gm,{"aria-label":"Discard all changes",onClick:c,colorPalette:"purple",variant:"solid",size:"md",mt:"0.5rem",disabled:r,loading:o,children:H.jsx(xa,{icon:Tor})})}),H.jsx(K4,{title:"Discard changes",actions:b,open:s,onClose:d,onOpen:c,children:H.jsx(Fi,{color:h,fontSize:"sm",children:"Undo the changes and revert to the last successfully saved version. This action cannot be undone."})})]})},xqr=({onAddFile:n,onAddFolder:r,children:o,disabled:s})=>H.jsxs(AYe,{children:[H.jsx(Ten,{asChild:!0,children:o}),H.jsx(lU,{children:H.jsx(IYe,{children:H.jsxs(OYe,{children:[H.jsx(NB,{value:"new-file",onClick:()=>!s&&n(),disabled:s,children:"New file"}),H.jsx(NB,{value:"new-folder",onClick:()=>!s&&r(),disabled:s,children:"New folder"})]})})})]}),jdn=({node:n,onSave:r,onRename:o,onCancel:s,isEditing:c})=>{const d="rasaNeutral.200",h="rasawebNeutral.900",b=Q.useRef(null),[y,E]=Q.useState(n.name),k=Q.useCallback(()=>{const O=yqr(n,y);if(!(!n.isTemporary&&n.name===O)){if(!Eqr(y,O,n.isDirectory)){E(n.name),s();return}n.isTemporary?r(y):(n.file?rPr(n.file,O):n.isDirectory&&iPr(n.id,O),o(O))}},[y,n,r,o,s]),x=()=>{E(n.name),s?.()};return H.jsxs(mQn,{fontSize:"sm",color:d,value:y,onValueChange:O=>{E(O.value)},onValueCommit:k,onValueRevert:x,placeholder:"Enter name",flex:1,minWidth:0,maxWidth:"100%",defaultEdit:c,children:[H.jsx(bQn,{ref:b,p:0,_hover:{bg:h},border:0,lineHeight:"1.2",minH:"1.2em",overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap",width:"100%",display:"block"}),H.jsx(vQn,{p:0,border:0,outline:"none",lineHeight:"1.2",minH:"1.2em",_focus:{border:0,outline:"none"},overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap",width:"100%",display:"block"})]})},Cqr=({node:n,isCurrentlyEditing:r,isEditable:o,onSave:s,onCompleteRename:c,onCancel:d,...h})=>{const b="rasawebNeutral.900",y="rasaNeutral.200",E=Q.useRef(null);return H.jsx(xer,{minH:"2rem",bg:"transparent",_hover:{bg:b},_selected:{bg:"transparent"},_focus:{bg:"transparent"},transition:"background-color 0.2s",role:"none",...h,children:H.jsxs(Cer,{minW:"100%",children:[H.jsx(Aer,{asChild:!0,color:y,children:H.jsx(xa,{icon:nnn,color:y,fontSize:"sm"})}),H.jsx(Wg,{content:n.name,showArrow:!0,onlyShowIfTruncated:!0,contentRef:E,openDelay:3e3,positioning:{placement:"top-end"},children:H.jsx(Ier,{ref:E,color:y,_hover:{bg:b},fontSize:"sm",ml:"0.5rem",overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap",children:o||r?H.jsx(jdn,{node:n,onSave:s,onRename:c,onCancel:d,isEditing:r}):H.jsx(H.Fragment,{children:n.name})})})]})})},Aqr=({node:n,isCurrentlySelected:r=!1,isCurrentlyEditing:o,isEditable:s,onSave:c,onCompleteRename:d,onCancel:h,...b})=>{const y=Q.useRef(null),E="rasawebOrange.600",k="rasaRed.800",x="rasaNeutral.200",O="rasawebNeutral.900",P=r?O:"transparent",D=n.file?.edited,$=n.file?.hasErrors,U=n.isRoot?H.jsx(xa,{icon:enn,fontSize:"sm",color:x}):null,B=n.file?.deleted?{"&::after":{content:'""',position:"absolute",left:0,right:0,top:"50%",height:"1px",backgroundColor:"currentColor",pointerEvents:"none"}}:void 0;return H.jsx(Ner,{minH:"2rem",bg:P,_hover:{bg:O},transition:"background-color 0.2s",role:"none",...b,children:H.jsxs(Iu,{align:"center",h:"100%",w:"100%",gap:"0.5rem",children:[U,s||o?H.jsx(jdn,{node:n,onSave:c,onRename:d,onCancel:h,isEditing:o}):H.jsx(Wg,{content:n.name,showArrow:!0,onlyShowIfTruncated:!0,contentRef:y,openDelay:3e3,positioning:{placement:"top-end"},children:H.jsx(Fi,{ref:y,size:"md",color:x,overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap",position:"relative",css:B,children:n.name})}),D||$?H.jsx(ai,{bg:$?k:E,width:"0.5rem",height:"0.5rem",borderRadius:"50%",flexShrink:0}):null]})})},Iqr=({open:n,onClose:r,onOpen:o,onConfirm:s,isDirectory:c,name:d})=>{const h="rasawebDeepPurple.800",b=[{id:"Cancel",title:"Cancel",onClick:()=>r(),"aria-label":"Cancel",variant:"outline",colorPalette:"dark"},{id:"Delete",title:"Delete",onClick:()=>{s(),r()},"aria-label":"Confirm",variant:"solid",colorPalette:"dark"}];return H.jsx(K4,{title:c?"Delete folder":"Delete file",actions:b,open:n,onClose:r,onOpen:o,children:c?H.jsxs(Fi,{color:h,fontSize:"sm",children:["Are you sure you want to delete all files in ",d,"?"]}):H.jsxs(Fi,{color:h,fontSize:"sm",children:["Are you sure you want to delete ",d,"?"]})})},Oqr=({node:n,onAddFile:r,onAddFolder:o,onDeleteNode:s,expandedValue:c,setExpandedValue:d,onRename:h,children:b,disabled:y})=>{const{open:E,onOpen:k,onClose:x}=_M(),O=(P,D)=>{if(P.isDirectory&&!c.includes(P.id))d([...c,P.id]);else{const $=P.id.substring(0,P.id.lastIndexOf("/"));$&&!c.includes($)&&d([...c,$])}D==="file"?r(P):o(P)};return H.jsxs(H.Fragment,{children:[H.jsxs(AYe,{children:[H.jsx(Ten,{asChild:!0,children:b}),H.jsx(lU,{children:H.jsx(IYe,{children:H.jsxs(OYe,{children:[H.jsx(NB,{value:"new-file",onClick:()=>!y&&O(n,"file"),disabled:y,children:"New file"}),H.jsx(NB,{value:"new-folder",onClick:()=>!y&&O(n,"folder"),disabled:y,children:"New folder"}),H.jsx(NB,{value:"rename",onClick:()=>!y&&h(n),disabled:y,children:"Rename"}),H.jsx(NB,{value:"delete",onClick:()=>!y&&k(),disabled:y,children:"Delete"})]})})})]}),H.jsx(Iqr,{onConfirm:()=>s(n),open:E,onOpen:k,onClose:x,isDirectory:n?.isDirectory??!1,name:n.name})]})},Nqr=({node:n,onAddFile:r,onAddFolder:o,currentFile:s,expandedValue:c,setExpandedValue:d,editingNodeId:h,onEditingComplete:b,setEditingNodeId:y,onTemporaryNodeCommit:E,onTemporaryNodeCancel:k,onDeleteNode:x,renderDeletedFiles:O=!1,fileTreeNodeActionsEnabled:P=!0})=>{const D=h===n.id,$=n.deleted||n.file?.deleted,[U,B]=Q.useState(n.isEditable||!1),G=Io(gc,PK),K=he=>{B(!1),E(n.id,he)},X=()=>{B(!1),k(n.id)},ae=he=>{B(!1),b(he),n.isDirectory&&d([...c,n.id])};if($&&!O)return null;const oe=n.isDirectory?H.jsx(Cqr,{node:n,isCurrentlyEditing:D,isEditable:U,onSave:K,onCompleteRename:ae,onCancel:X}):H.jsx(Aqr,{node:n,isCurrentlySelected:s?.path===n.id,isCurrentlyEditing:D,onSave:K,onCompleteRename:ae,onCancel:X,isEditable:U});return P?H.jsx(Oqr,{node:n,onAddFile:r,onAddFolder:o,onDeleteNode:x,onRename:()=>{y(n.id),B(!0)},expandedValue:c,setExpandedValue:d,disabled:G,children:oe}):oe},$dn=({files:n,isEdited:r,loading:o,editorStatus:s,showDeletedFiles:c=!1,fileTreeActionsEnabled:d=!0,title:h,currentFile:b,onSelectFile:y,onCurrentFileChange:E,onApplyChanges:k,onDiscardChanges:x})=>{const O=Io(da,Re=>Re.assistantName),P="rasawebNeutral.700",D="rasaNeutral.200",[$,U]=Q.useState([]),[B,G]=Q.useState(null),[K,X]=Q.useState(null),[ae,oe]=Q.useState([]);Q.useEffect(()=>{if(!b?.path||!n.find(qe=>qe.path===b.path))return;const _e=dK(b.path);U(qe=>Array.from(new Set([...qe,..._e])))},[b?.path,n]);const he=Q.useCallback(Re=>{const _e=Re.selectedValue[0],qe=n.find(it=>it.path===_e);qe&&y(qe)},[n,y]),ye=Q.useMemo(()=>{const Re=fqr(n,ae,K||void 0);return KGn({nodeToValue:_e=>_e.id,nodeToString:_e=>_e.name,nodeToChildrenCount:_e=>_e.isDirectory&&(!_e.children||_e.children?.length===0)?1:_e.children?.length||0,rootNode:{id:"ROOT",name:"",isDirectory:!0,children:Re,isEditable:!1}})},[n,ae,K]);function we(Re,_e=!1){X(null);const qe=pqr(Re),Xe=Sqr(_e?"New folder":"New file",qe,n,_e,ae),de=cqr(Xe,qe,_e);X(de),G(de.id)}function ge(Re){if(Re.isDirectory){const _e=ae.includes(Re.id);sPr(Re.id,_e)}else aPr(Re.id);oe(_e=>_e.filter(qe=>qe!==Re.id&&!qe.startsWith(`${Re.id}/`)))}function Te(Re,_e){if(K?.id===Re){if(_e.trim()===""){xe(Re);return}if(K.isDirectory){const qe=K.parentPath?`${K.parentPath}/${_e}`:_e;if(_e.includes("/")){const it=dK(qe);oe(Xe=>{const de=new Set(Xe);for(const We of it)de.add(We);return de.add(qe),Array.from(de)})}else oe(it=>[...it,qe]);X(null)}else{const it={path:K.parentPath?`${K.parentPath}/${_e}`:_e,content:"# New File Content",edited:!0,hasErrors:!1};Awe(it),E?.(it.path),X(null)}G(null)}}function xe(Re){K?.id===Re&&X(null),G(null)}async function Ue(){d&&(await k?.(),oe([]))}const Z=Re=>{if(B&&b?.path.startsWith(B)){const _e=b.path.replace(B,Re);E?.(_e)}G(null)};return H.jsxs(Iu,{direction:"column",bg:P,w:"15.5rem",h:"100%",p:"1rem","data-testid":"file-tree",children:[H.jsx(ai,{overflowY:"auto",style:{scrollbarWidth:"none",msOverflowStyle:"none"},children:H.jsxs(_er,{collection:ye,selectedValue:b?[b.path]:[],onSelectionChange:he,expandedValue:$,onExpandedChange:Re=>U(Re.expandedValue),size:"sm",children:[H.jsx(Kx,{size:"md",color:D,mb:"0.5rem",children:h||O||"Agent"}),H.jsx(Rer,{children:H.jsx(Per,{indentGuide:H.jsx(Oer,{bg:"rasawebNeutral.600"}),render:({node:Re})=>H.jsx(Nqr,{node:Re,currentFile:b,expandedValue:$,setExpandedValue:U,onAddFile:_e=>we(_e,!1),onAddFolder:_e=>we(_e,!0),onDeleteNode:ge,editingNodeId:B,onEditingComplete:Z,onTemporaryNodeCommit:Te,onTemporaryNodeCancel:xe,setEditingNodeId:G,renderDeletedFiles:c,fileTreeNodeActionsEnabled:d})})})]})}),d&&H.jsxs(H.Fragment,{children:[H.jsx(xqr,{onAddFile:()=>we(void 0,!1),onAddFolder:()=>we(void 0,!0),disabled:o,children:H.jsx(ai,{flex:"1","data-testid":"empty-space-menu"})}),H.jsxs(Iu,{px:"0.5rem",mt:"0.5rem",justify:"flex-end",gap:"0.5rem",children:[x&&H.jsx(Tqr,{onConfirm:x,isDisabled:!r||o}),H.jsx(kqr,{onClick:()=>void Ue(),disabled:!r,loading:o,editorStatus:s})]})]})]})},nWt=Koe.rasawebNeutral[900].value,Rqr=()=>{const{openFile:n,setOpenFile:r,version:o,setVersion:s}=bU(),{files:c,loading:d,getFiles:h}=FEe(),{postFiles:b}=mqr(),{projectUrl:y,jobStatus:E,jobId:k}=Io(da,Te=>({projectUrl:Te.projectUrl,jobStatus:Te.jobStatus,jobId:Te.jobId})),[x,O]=Q.useState(!1),P=Io(gc,Te=>Te.editorStatus),D=Io(gc,Te=>Te.error),{getIdTokenClaims:$}=ty(),U=Q.useRef(null),B=Q.useMemo(()=>!n||!c?.length?void 0:c.find(xe=>xe.path===n)||null,[n,c]);Q.useEffect(()=>{(async()=>{!c&&!d&&y&&await h()})()},[d,c,y,h]),Q.useEffect(()=>{if(c&&c.length>0){const Te=c.some(xe=>xe.edited);O(Te),Te&&Bcn(null)}else O(!1)},[c]);const G=Q.useMemo(()=>P!==null&&P!=="done"||d,[P,d]);Q.useEffect(()=>{!D||G||Kp.create({title:"Code editor error",description:D,type:"error",duration:4e3})},[D,G]),Q.useEffect(()=>{k&&E===a2.Received&&oPr(k)},[k,E,$]);const K=Te=>{if(r(Te.path),U.current){const xe=U.current.getModel();xe&&xe.pushStackElement(),setTimeout(()=>{U?.current?.focus()},0)}},X=async()=>{await h(),r(void 0),s(Te=>Te+1)},ae=Te=>{if(B&&Te!==void 0){const xe=Awe({...B,content:Te,edited:!0});r(xe.path)}},oe=(Te,xe)=>{U.current=Te,we(Te,xe)};let he="Select a file to edit";d?he="Loading files...":(!c||c.length===0)&&(he="No files found!");const ye=()=>new Promise((Te,xe)=>{vWe(!0),b({filesSnapshot:c},{onSuccess:()=>Te(),onError:()=>{vWe(!1),Kp.create({title:"Error applying changes",description:"Failed to apply changes",type:"error",duration:4e3}),xe(new Error("Failed to apply changes"))}})}),we=Q.useCallback((Te,xe)=>{xe.editor.defineTheme("my-vs-dark-diff",{base:"vs-dark",inherit:!0,rules:[],colors:{"editor.background":nWt}}),xe.editor.setTheme("my-vs-dark-diff")},[]),ge=B?`${B.path}/${o}`:n2();return H.jsxs(H.Fragment,{children:[H.jsx($dn,{files:c??[],onSelectFile:K,onCurrentFileChange:r,onApplyChanges:ye,onDiscardChanges:()=>void X(),currentFile:B,isEdited:x,loading:G,editorStatus:P}),H.jsx(ai,{flex:1,height:"100%",py:"1rem",bg:nWt,overflowY:"auto",children:H.jsx(sqr,{saveViewState:!0,path:ge,defaultValue:B?.content||he,defaultLanguage:Ccn(B?.path),theme:"vs-dark",options:{...Tcn,readOnly:!B||d},onChange:ae,onMount:oe})})]})};function Pqr(n,r,o){return o===YJe.Renamed&&n&&r?Lqr(n,r):r}function Lqr(n,r){if(!r||!n)return"";const o=n.split("/"),s=r.split("/");if(o.length!==s.length)return r;let c="";for(let d=0;d<o.length;d++)o[d]===s[d]?c+=o[d]:c+=o[d]+" >> "+s[d],d!==o.length-1&&(c+="/");return c}function Dqr(n){if(!n?.files)return null;const r={},o=[];for(const[s,c]of Object.entries(n.files)){if(!c)continue;const d=Pqr(c.path_original,s,c.status),h=c.status===YJe.Deleted;o.push({path:d,content:c.content_modified,status:c.status,originalContent:c.content_original,edited:!1,deleted:h,hasErrors:!1,isDirectory:!1,isTemporary:!1});const b=dK(d);for(const y of b)r[y]||(r[y]={path:y,content:"",edited:!1,hasErrors:!1,isDirectory:!0,isTemporary:!1})}for(const[s,c]of Object.entries(r))o.push(c);return o}const Mqr=({commitSha:n})=>{const r=Io(da,h=>h.projectUrl),{data:o,error:s,isFetching:c,refetch:d}=xK({queryKey:["fileDiffWithContents",r,n],queryFn:()=>yhr({projectUrl:r,commitSha:n}),select:Dqr,enabled:!!r&&!!n&&n!=="",staleTime:1/0,gcTime:1/0,refetchOnMount:!1,refetchOnWindowFocus:!1,refetchOnReconnect:!1,refetchInterval:!1,refetchIntervalInBackground:!1,retryDelay:1e3,retry:3});return{getDiff:d,loading:c,error:s,files:o}},jqr=({handleBackToEditing:n,title:r,disabled:o=!1})=>{const s=Q.useCallback(()=>{o||n()},[o,n]);return H.jsx(ai,{as:"button",onClick:s,cursor:o?"not-allowed":"pointer",bg:"rasawebPurple.800",height:"52px",display:"flex",alignItems:"center",pl:5,children:H.jsxs($c,{gap:"8px",children:[H.jsx(xa,{icon:Dor,width:"8px",height:"8px",color:"rasaNeutral.50"}),H.jsx(Fi,{color:"rasaNeutral.50",size:"sm",children:r})]})})},H0e=`${Koe.rasaRed[800].value}33`,aqe=`${Koe.rasaGreen[800].value}33`,G0e=Koe.rasawebNeutral[900].value,$qr=()=>{const n=y2(),{getPath:r}=MK(),{commitSha:o}=G_(),[s,c]=Q.useState(null),{loading:d,error:h,files:b}=Mqr({commitSha:o});Q.useEffect(()=>{if(b&&b.length>0){const x=b.find(O=>!O.isDirectory);if(!x)return;c({original:x,modified:x})}},[b,o]),Q.useEffect(()=>{h&&(setTimeout(()=>{Kp.create({title:"Error loading the diff",description:h.message,type:"error",duration:4e3})},0),Va(h,{tags:{component:"DiffFile",action:"getDiff"}}))},[h]);const y=Q.useCallback((x,O)=>{O.editor.defineTheme("my-vs-dark-diff",{base:"vs-dark",inherit:!0,rules:[{token:"",foreground:"D4D4D4"}],colors:{"editor.background":G0e,"diffEditor.removedLineBackground":H0e,"diffEditor.removedTextBackground":H0e,"diffEditorGutter.removedBackground":H0e,"diffEditor.diagonalFill":H0e,"diffEditor.insertedLineBackground":aqe,"diffEditor.insertedTextBackground":aqe,"diffEditorGutter.insertedBackground":aqe,"diffEditor.unchangedCodeBackground":G0e}}),O.editor.setTheme("my-vs-dark-diff")},[]);if(h||!o||o==="")return null;const E=async()=>{await n(r("code"))},k=x=>{if(x.isDirectory)return;const O=b?.find(P=>P.path===x.path);if(!O){Va(new Error(`Original file not found for ${x.path}`),{tags:{component:"CodeDiffViewer",action:"handleSelectFile"}});return}c({original:O,modified:x})};return H.jsxs(Iu,{direction:"column",height:"100%",width:"100%",children:[H.jsx(jqr,{handleBackToEditing:E,title:"Back to editing"}),H.jsxs(Iu,{flex:"1",overflow:"hidden",children:[H.jsx($dn,{title:"Changed files (read-only)",files:b??[],onSelectFile:k,currentFile:s?.original??void 0,isEdited:!1,loading:d,editorStatus:null,showDeletedFiles:!0,fileTreeActionsEnabled:!1}),H.jsx(ai,{"data-testid":"diff-editor",flex:"1",height:"100%",bg:G0e,py:"1rem",children:H.jsx(nqr,{onMount:y,keepCurrentOriginalModel:!0,keepCurrentModifiedModel:!0,height:"100%",loading:H.jsx(ai,{height:"100%",bg:G0e}),original:d?"Loading files...":s?.original?.originalContent??"",modified:d?"Loading files...":s?.modified?.content??"",language:Ccn(s?.original?.path),options:{...Tcn,readOnly:!0,experimental:{useTrueInlineView:!0},renderOverviewRuler:!1,renderSideBySide:!1,diffCodeLens:!1,ignoreTrimWhitespace:!0,renderIndicators:!1,hideUnchangedRegions:{enabled:!0}}},o)})]})]})},Fqr=()=>{const{commitSha:n}=G_();return n?H.jsx($qr,{}):H.jsx(Rqr,{})},rWt=()=>{zln(),P8r(),A8r();const{selectedView:n,inspectEnabled:r}=bU(),{sendMessage:o,setUrl:s,startNewConversation:c,conversationList:d,inputDisabled:h,sessionId:b,stack:y,slots:E,flowHistory:k,replayingConversation:x,waitingForUserInput:O,replayConversation:P,cleanup:D}=Uln(),{isLoading:$,projectError:U,isInactive:B,jobStatus:G,isLoadingProject:K,isRestoring:X}=Io(da,we=>({isLoading:we.loading,isLoadingProject:we.loadingProject,projectError:we.error,isInactive:we.inactive,jobStatus:we.jobStatus,isRestoring:we.isRestoring})),ae=AEe("useAlphaGenerate",!1)[0],oe=Q.useRef(null);Q.useEffect(()=>{const we=oe.current;we&&(we===a2.Training||we===a2.Heartbeat)&&G===a2.TrainSuccess&&c(),oe.current=G},[G,c]),Q.useEffect(()=>{B&&D()},[B,D]);const he=Q.useCallback((we,ge=!0)=>{const Te={flex:"1",height:"100%",px:ge?"3.75rem":"2rem",py:"3.75rem",transition:"all 0.3s ease-in-out",background:"url(/bg.svg)",backgroundSize:"cover",justifyContent:"center"},xe={h:"100%",w:"100%",maxWidth:ge?{base:"38rem",xl:"40rem","2xl":"52.5rem","3xl":"62.5rem"}:"100%",transition:"max-width 0.3s ease-in-out",justifyContent:"center",borderColor:"gray.200",borderRadius:"1rem",backgroundColor:"white"};return H.jsx(ai,{flex:1,children:H.jsx(Iu,{css:Te,children:H.jsx(Iu,{css:xe,children:we})})})},[]),ye=Q.useCallback(()=>{const we=G===a2.Done||G===a2.InitialTrainError&&ae;return B?he(H.jsx(mdn,{})):U?he(H.jsx(Fzr,{})):K&&G&&!we?he(H.jsx(AMr,{})):$?he(H.jsx(Uzr,{message:"Loading..."})):X?he(H.jsx(gdn,{})):n===q4.Code?H.jsx(Fqr,{}):G===a2.InitialTrainError&&ae?he(H.jsx(Bzr,{})):he(H.jsx(Tdn,{flowView:r,sendMessage:o,setUrl:s,startNewConversation:c,conversationList:d,inputDisabled:h,sessionId:b,stack:y,slots:E,flowHistory:k,replayingConversation:x,waitingForUserInput:O,replayConversation:P}),!r)},[B,U,K,G,$,X,n,r,he,o,s,c,d,h,E,k,b,y,x,O,P,ae]);return H.jsxs(Vd,{height:"100%",flexGrow:1,gap:0,overflow:"auto",children:[H.jsx(pdn,{}),H.jsx(Iu,{w:"100%",flex:1,minHeight:0,children:ye()})]})},Bqr={slow:"2s",normal:"1.5s",fast:"1s"},Fdn=({children:n,size:r="md",colorPalette:o="copilot",speed:s="slow"})=>{const c=Bqr[s],d=`shimmer-${s}`;return H.jsxs(H.Fragment,{children:[H.jsx("style",{children:`
        @keyframes ${d} {
          0% {
            background-position: -200% 0;
          }
          100% {
            background-position: 200% 0;
          }
        }
      `}),H.jsxs(ai,{position:"relative",display:"inline-block",colorPalette:o,children:[H.jsx(Fi,{fontSize:r,color:"colorPalette.subtle",position:"relative",zIndex:"docked",children:n}),H.jsx(ai,{position:"absolute",inset:"0",pointerEvents:"none",zIndex:"overlay",style:{background:`linear-gradient(
              90deg,
              transparent 0%,
              rgba(255, 255, 255, 0.9) 50%,
              transparent 100%
            )`,backgroundSize:"200% 100%",animation:`${d} ${c} linear infinite`},_motionReduce:{animation:"none !important"}})]})]})},QE="0.875rem",Wwe="0.75rem",aU="170%",Uqr=[{value:Ox.WrongAnswer,label:"Wrong or incomplete answer"},{value:Ox.TooLong,label:"Too long"},{value:Ox.OffTopic,label:"Off topic"},{value:Ox.BrokenLinks,label:"Used broken or incorrect links"},{value:Ox.WrongFlag,label:"Flagged my answer incorrectly"},{value:Ox.Other,label:"Other"}],Bdn=({attachment:n,removable:r=!1})=>{const o=zqr(n);return H.jsxs(Men,{size:"md",variant:"solid",borderRadius:"full",bg:"rasaNeutral.200",color:"rasawebDeepPurple.800",px:"0.75rem",py:"0.25rem",alignItems:"center",gap:"0.5rem","data-testid":"copilot-attachment",children:[H.jsx(oer,{display:"flex",alignItems:"center",children:Hqr(n)}),H.jsx(jen,{fontSize:Wwe,display:"flex",alignItems:"center",children:o}),r?H.jsx(aer,{display:"flex",alignItems:"center",children:H.jsx(ier,{_hover:{cursor:"pointer"},onClick:()=>U5r(n.id),color:"rasaNeutral.700","data-testid":"remove-attachment-button"})}):null]},n.id)},zqr=n=>nK(n)?AQe.includes(n.conversationEventType)?`Flow (${n.flowId})`:IQe.includes(n.conversationEventType)?`Slot (${n.name})`:n.conversationEventType===Sl.Action?`Action (${n.name})`:n.name:UB(n)&&n.type===Vx.Bot?`Response (${n.metadata.utter_action||n.text})`:"",Hqr=n=>{if(UB(n)&&n.type===Vx.Bot)return H.jsx(xa,{color:"rasawebDeepPurple.800",icon:UVe});if(nK(n))return AQe.includes(n.conversationEventType)?H.jsx(xa,{color:"rasawebDeepPurple.800",icon:BVe}):IQe.includes(n.conversationEventType)?H.jsx(xa,{color:"rasawebDeepPurple.800",icon:cXe}):n.conversationEventType===Sl.Action?H.jsx(xa,{color:"rasawebDeepPurple.800",icon:NK}):H.jsx(xa,{color:"rasawebDeepPurple.800",icon:UVe})};function iWt(n){const r=[],o=String(n||"");let s=o.indexOf(","),c=0,d=!1;for(;!d;){s===-1&&(s=o.length,d=!0);const h=o.slice(c,s).trim();(h||!d)&&r.push(h),c=s+1,s=o.indexOf(",",c)}return r}function Udn(n,r){const o={};return(n[n.length-1]===""?[...n,""]:n).join((o.padRight?" ":"")+","+(o.padLeft===!1?"":" ")).trim()}const Gqr=/^[$_\p{ID_Start}][$_\u{200C}\u{200D}\p{ID_Continue}]*$/u,qqr=/^[$_\p{ID_Start}][-$_\u{200C}\u{200D}\p{ID_Continue}]*$/u,Vqr={};function oWt(n,r){return(Vqr.jsx?qqr:Gqr).test(n)}const Wqr=/[ \t\n\f\r]/g;function Kqr(n){return typeof n=="object"?n.type==="text"?aWt(n.value):!1:aWt(n)}function aWt(n){return n.replace(Wqr,"")===""}let Xoe=class{constructor(r,o,s){this.normal=o,this.property=r,s&&(this.space=s)}};Xoe.prototype.normal={};Xoe.prototype.property={};Xoe.prototype.space=void 0;function zdn(n,r){const o={},s={};for(const c of n)Object.assign(o,c.property),Object.assign(s,c.normal);return new Xoe(o,s,r)}function poe(n){return n.toLowerCase()}let ay=class{constructor(r,o){this.attribute=o,this.property=r}};ay.prototype.attribute="";ay.prototype.booleanish=!1;ay.prototype.boolean=!1;ay.prototype.commaOrSpaceSeparated=!1;ay.prototype.commaSeparated=!1;ay.prototype.defined=!1;ay.prototype.mustUseProperty=!1;ay.prototype.number=!1;ay.prototype.overloadedBoolean=!1;ay.prototype.property="";ay.prototype.spaceSeparated=!1;ay.prototype.space=void 0;let Yqr=0;const Ts=wU(),Hp=wU(),cKe=wU(),Vr=wU(),Td=wU(),BW=wU(),JE=wU();function wU(){return 2**++Yqr}const uKe=Object.freeze(Object.defineProperty({__proto__:null,boolean:Ts,booleanish:Hp,commaOrSpaceSeparated:JE,commaSeparated:BW,number:Vr,overloadedBoolean:cKe,spaceSeparated:Td},Symbol.toStringTag,{value:"Module"})),sqe=Object.keys(uKe);let NQe=class extends ay{constructor(r,o,s,c){let d=-1;if(super(r,o),sWt(this,"space",c),typeof s=="number")for(;++d<sqe.length;){const h=sqe[d];sWt(this,sqe[d],(s&uKe[h])===uKe[h])}}};NQe.prototype.defined=!0;function sWt(n,r,o){o&&(n[r]=o)}function jK(n){const r={},o={};for(const[s,c]of Object.entries(n.properties)){const d=new NQe(s,n.transform(n.attributes||{},s),c,n.space);n.mustUseProperty&&n.mustUseProperty.includes(s)&&(d.mustUseProperty=!0),r[s]=d,o[poe(s)]=s,o[poe(d.attribute)]=s}return new Xoe(r,o,n.space)}const Hdn=jK({properties:{ariaActiveDescendant:null,ariaAtomic:Hp,ariaAutoComplete:null,ariaBusy:Hp,ariaChecked:Hp,ariaColCount:Vr,ariaColIndex:Vr,ariaColSpan:Vr,ariaControls:Td,ariaCurrent:null,ariaDescribedBy:Td,ariaDetails:null,ariaDisabled:Hp,ariaDropEffect:Td,ariaErrorMessage:null,ariaExpanded:Hp,ariaFlowTo:Td,ariaGrabbed:Hp,ariaHasPopup:null,ariaHidden:Hp,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:Td,ariaLevel:Vr,ariaLive:null,ariaModal:Hp,ariaMultiLine:Hp,ariaMultiSelectable:Hp,ariaOrientation:null,ariaOwns:Td,ariaPlaceholder:null,ariaPosInSet:Vr,ariaPressed:Hp,ariaReadOnly:Hp,ariaRelevant:null,ariaRequired:Hp,ariaRoleDescription:Td,ariaRowCount:Vr,ariaRowIndex:Vr,ariaRowSpan:Vr,ariaSelected:Hp,ariaSetSize:Vr,ariaSort:null,ariaValueMax:Vr,ariaValueMin:Vr,ariaValueNow:Vr,ariaValueText:null,role:null},transform(n,r){return r==="role"?r:"aria-"+r.slice(4).toLowerCase()}});function Gdn(n,r){return r in n?n[r]:r}function qdn(n,r){return Gdn(n,r.toLowerCase())}const Xqr=jK({attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:BW,acceptCharset:Td,accessKey:Td,action:null,allow:null,allowFullScreen:Ts,allowPaymentRequest:Ts,allowUserMedia:Ts,alt:null,as:null,async:Ts,autoCapitalize:null,autoComplete:Td,autoFocus:Ts,autoPlay:Ts,blocking:Td,capture:null,charSet:null,checked:Ts,cite:null,className:Td,cols:Vr,colSpan:null,content:null,contentEditable:Hp,controls:Ts,controlsList:Td,coords:Vr|BW,crossOrigin:null,data:null,dateTime:null,decoding:null,default:Ts,defer:Ts,dir:null,dirName:null,disabled:Ts,download:cKe,draggable:Hp,encType:null,enterKeyHint:null,fetchPriority:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:Ts,formTarget:null,headers:Td,height:Vr,hidden:cKe,high:Vr,href:null,hrefLang:null,htmlFor:Td,httpEquiv:Td,id:null,imageSizes:null,imageSrcSet:null,inert:Ts,inputMode:null,integrity:null,is:null,isMap:Ts,itemId:null,itemProp:Td,itemRef:Td,itemScope:Ts,itemType:Td,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:Ts,low:Vr,manifest:null,max:null,maxLength:Vr,media:null,method:null,min:null,minLength:Vr,multiple:Ts,muted:Ts,name:null,nonce:null,noModule:Ts,noValidate:Ts,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforeMatch:null,onBeforePrint:null,onBeforeToggle:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextLost:null,onContextMenu:null,onContextRestored:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onScrollEnd:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:Ts,optimum:Vr,pattern:null,ping:Td,placeholder:null,playsInline:Ts,popover:null,popoverTarget:null,popoverTargetAction:null,poster:null,preload:null,readOnly:Ts,referrerPolicy:null,rel:Td,required:Ts,reversed:Ts,rows:Vr,rowSpan:Vr,sandbox:Td,scope:null,scoped:Ts,seamless:Ts,selected:Ts,shadowRootClonable:Ts,shadowRootDelegatesFocus:Ts,shadowRootMode:null,shape:null,size:Vr,sizes:null,slot:null,span:Vr,spellCheck:Hp,src:null,srcDoc:null,srcLang:null,srcSet:null,start:Vr,step:null,style:null,tabIndex:Vr,target:null,title:null,translate:null,type:null,typeMustMatch:Ts,useMap:null,value:Hp,width:Vr,wrap:null,writingSuggestions:null,align:null,aLink:null,archive:Td,axis:null,background:null,bgColor:null,border:Vr,borderColor:null,bottomMargin:Vr,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:Ts,declare:Ts,event:null,face:null,frame:null,frameBorder:null,hSpace:Vr,leftMargin:Vr,link:null,longDesc:null,lowSrc:null,marginHeight:Vr,marginWidth:Vr,noResize:Ts,noHref:Ts,noShade:Ts,noWrap:Ts,object:null,profile:null,prompt:null,rev:null,rightMargin:Vr,rules:null,scheme:null,scrolling:Hp,standby:null,summary:null,text:null,topMargin:Vr,valueType:null,version:null,vAlign:null,vLink:null,vSpace:Vr,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:Ts,disableRemotePlayback:Ts,prefix:null,property:null,results:Vr,security:null,unselectable:null},space:"html",transform:qdn}),Jqr=jK({attributes:{accentHeight:"accent-height",alignmentBaseline:"alignment-baseline",arabicForm:"arabic-form",baselineShift:"baseline-shift",capHeight:"cap-height",className:"class",clipPath:"clip-path",clipRule:"clip-rule",colorInterpolation:"color-interpolation",colorInterpolationFilters:"color-interpolation-filters",colorProfile:"color-profile",colorRendering:"color-rendering",crossOrigin:"crossorigin",dataType:"datatype",dominantBaseline:"dominant-baseline",enableBackground:"enable-background",fillOpacity:"fill-opacity",fillRule:"fill-rule",floodColor:"flood-color",floodOpacity:"flood-opacity",fontFamily:"font-family",fontSize:"font-size",fontSizeAdjust:"font-size-adjust",fontStretch:"font-stretch",fontStyle:"font-style",fontVariant:"font-variant",fontWeight:"font-weight",glyphName:"glyph-name",glyphOrientationHorizontal:"glyph-orientation-horizontal",glyphOrientationVertical:"glyph-orientation-vertical",hrefLang:"hreflang",horizAdvX:"horiz-adv-x",horizOriginX:"horiz-origin-x",horizOriginY:"horiz-origin-y",imageRendering:"image-rendering",letterSpacing:"letter-spacing",lightingColor:"lighting-color",markerEnd:"marker-end",markerMid:"marker-mid",markerStart:"marker-start",navDown:"nav-down",navDownLeft:"nav-down-left",navDownRight:"nav-down-right",navLeft:"nav-left",navNext:"nav-next",navPrev:"nav-prev",navRight:"nav-right",navUp:"nav-up",navUpLeft:"nav-up-left",navUpRight:"nav-up-right",onAbort:"onabort",onActivate:"onactivate",onAfterPrint:"onafterprint",onBeforePrint:"onbeforeprint",onBegin:"onbegin",onCancel:"oncancel",onCanPlay:"oncanplay",onCanPlayThrough:"oncanplaythrough",onChange:"onchange",onClick:"onclick",onClose:"onclose",onCopy:"oncopy",onCueChange:"oncuechange",onCut:"oncut",onDblClick:"ondblclick",onDrag:"ondrag",onDragEnd:"ondragend",onDragEnter:"ondragenter",onDragExit:"ondragexit",onDragLeave:"ondragleave",onDragOver:"ondragover",onDragStart:"ondragstart",onDrop:"ondrop",onDurationChange:"ondurationchange",onEmptied:"onemptied",onEnd:"onend",onEnded:"onended",onError:"onerror",onFocus:"onfocus",onFocusIn:"onfocusin",onFocusOut:"onfocusout",onHashChange:"onhashchange",onInput:"oninput",onInvalid:"oninvalid",onKeyDown:"onkeydown",onKeyPress:"onkeypress",onKeyUp:"onkeyup",onLoad:"onload",onLoadedData:"onloadeddata",onLoadedMetadata:"onloadedmetadata",onLoadStart:"onloadstart",onMessage:"onmessage",onMouseDown:"onmousedown",onMouseEnter:"onmouseenter",onMouseLeave:"onmouseleave",onMouseMove:"onmousemove",onMouseOut:"onmouseout",onMouseOver:"onmouseover",onMouseUp:"onmouseup",onMouseWheel:"onmousewheel",onOffline:"onoffline",onOnline:"ononline",onPageHide:"onpagehide",onPageShow:"onpageshow",onPaste:"onpaste",onPause:"onpause",onPlay:"onplay",onPlaying:"onplaying",onPopState:"onpopstate",onProgress:"onprogress",onRateChange:"onratechange",onRepeat:"onrepeat",onReset:"onreset",onResize:"onresize",onScroll:"onscroll",onSeeked:"onseeked",onSeeking:"onseeking",onSelect:"onselect",onShow:"onshow",onStalled:"onstalled",onStorage:"onstorage",onSubmit:"onsubmit",onSuspend:"onsuspend",onTimeUpdate:"ontimeupdate",onToggle:"ontoggle",onUnload:"onunload",onVolumeChange:"onvolumechange",onWaiting:"onwaiting",onZoom:"onzoom",overlinePosition:"overline-position",overlineThickness:"overline-thickness",paintOrder:"paint-order",panose1:"panose-1",pointerEvents:"pointer-events",referrerPolicy:"referrerpolicy",renderingIntent:"rendering-intent",shapeRendering:"shape-rendering",stopColor:"stop-color",stopOpacity:"stop-opacity",strikethroughPosition:"strikethrough-position",strikethroughThickness:"strikethrough-thickness",strokeDashArray:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeLineCap:"stroke-linecap",strokeLineJoin:"stroke-linejoin",strokeMiterLimit:"stroke-miterlimit",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",tabIndex:"tabindex",textAnchor:"text-anchor",textDecoration:"text-decoration",textRendering:"text-rendering",transformOrigin:"transform-origin",typeOf:"typeof",underlinePosition:"underline-position",underlineThickness:"underline-thickness",unicodeBidi:"unicode-bidi",unicodeRange:"unicode-range",unitsPerEm:"units-per-em",vAlphabetic:"v-alphabetic",vHanging:"v-hanging",vIdeographic:"v-ideographic",vMathematical:"v-mathematical",vectorEffect:"vector-effect",vertAdvY:"vert-adv-y",vertOriginX:"vert-origin-x",vertOriginY:"vert-origin-y",wordSpacing:"word-spacing",writingMode:"writing-mode",xHeight:"x-height",playbackOrder:"playbackorder",timelineBegin:"timelinebegin"},properties:{about:JE,accentHeight:Vr,accumulate:null,additive:null,alignmentBaseline:null,alphabetic:Vr,amplitude:Vr,arabicForm:null,ascent:Vr,attributeName:null,attributeType:null,azimuth:Vr,bandwidth:null,baselineShift:null,baseFrequency:null,baseProfile:null,bbox:null,begin:null,bias:Vr,by:null,calcMode:null,capHeight:Vr,className:Td,clip:null,clipPath:null,clipPathUnits:null,clipRule:null,color:null,colorInterpolation:null,colorInterpolationFilters:null,colorProfile:null,colorRendering:null,content:null,contentScriptType:null,contentStyleType:null,crossOrigin:null,cursor:null,cx:null,cy:null,d:null,dataType:null,defaultAction:null,descent:Vr,diffuseConstant:Vr,direction:null,display:null,dur:null,divisor:Vr,dominantBaseline:null,download:Ts,dx:null,dy:null,edgeMode:null,editable:null,elevation:Vr,enableBackground:null,end:null,event:null,exponent:Vr,externalResourcesRequired:null,fill:null,fillOpacity:Vr,fillRule:null,filter:null,filterRes:null,filterUnits:null,floodColor:null,floodOpacity:null,focusable:null,focusHighlight:null,fontFamily:null,fontSize:null,fontSizeAdjust:null,fontStretch:null,fontStyle:null,fontVariant:null,fontWeight:null,format:null,fr:null,from:null,fx:null,fy:null,g1:BW,g2:BW,glyphName:BW,glyphOrientationHorizontal:null,glyphOrientationVertical:null,glyphRef:null,gradientTransform:null,gradientUnits:null,handler:null,hanging:Vr,hatchContentUnits:null,hatchUnits:null,height:null,href:null,hrefLang:null,horizAdvX:Vr,horizOriginX:Vr,horizOriginY:Vr,id:null,ideographic:Vr,imageRendering:null,initialVisibility:null,in:null,in2:null,intercept:Vr,k:Vr,k1:Vr,k2:Vr,k3:Vr,k4:Vr,kernelMatrix:JE,kernelUnitLength:null,keyPoints:null,keySplines:null,keyTimes:null,kerning:null,lang:null,lengthAdjust:null,letterSpacing:null,lightingColor:null,limitingConeAngle:Vr,local:null,markerEnd:null,markerMid:null,markerStart:null,markerHeight:null,markerUnits:null,markerWidth:null,mask:null,maskContentUnits:null,maskUnits:null,mathematical:null,max:null,media:null,mediaCharacterEncoding:null,mediaContentEncodings:null,mediaSize:Vr,mediaTime:null,method:null,min:null,mode:null,name:null,navDown:null,navDownLeft:null,navDownRight:null,navLeft:null,navNext:null,navPrev:null,navRight:null,navUp:null,navUpLeft:null,navUpRight:null,numOctaves:null,observer:null,offset:null,onAbort:null,onActivate:null,onAfterPrint:null,onBeforePrint:null,onBegin:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnd:null,onEnded:null,onError:null,onFocus:null,onFocusIn:null,onFocusOut:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadStart:null,onMessage:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onMouseWheel:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRepeat:null,onReset:null,onResize:null,onScroll:null,onSeeked:null,onSeeking:null,onSelect:null,onShow:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnload:null,onVolumeChange:null,onWaiting:null,onZoom:null,opacity:null,operator:null,order:null,orient:null,orientation:null,origin:null,overflow:null,overlay:null,overlinePosition:Vr,overlineThickness:Vr,paintOrder:null,panose1:null,path:null,pathLength:Vr,patternContentUnits:null,patternTransform:null,patternUnits:null,phase:null,ping:Td,pitch:null,playbackOrder:null,pointerEvents:null,points:null,pointsAtX:Vr,pointsAtY:Vr,pointsAtZ:Vr,preserveAlpha:null,preserveAspectRatio:null,primitiveUnits:null,propagate:null,property:JE,r:null,radius:null,referrerPolicy:null,refX:null,refY:null,rel:JE,rev:JE,renderingIntent:null,repeatCount:null,repeatDur:null,requiredExtensions:JE,requiredFeatures:JE,requiredFonts:JE,requiredFormats:JE,resource:null,restart:null,result:null,rotate:null,rx:null,ry:null,scale:null,seed:null,shapeRendering:null,side:null,slope:null,snapshotTime:null,specularConstant:Vr,specularExponent:Vr,spreadMethod:null,spacing:null,startOffset:null,stdDeviation:null,stemh:null,stemv:null,stitchTiles:null,stopColor:null,stopOpacity:null,strikethroughPosition:Vr,strikethroughThickness:Vr,string:null,stroke:null,strokeDashArray:JE,strokeDashOffset:null,strokeLineCap:null,strokeLineJoin:null,strokeMiterLimit:Vr,strokeOpacity:Vr,strokeWidth:null,style:null,surfaceScale:Vr,syncBehavior:null,syncBehaviorDefault:null,syncMaster:null,syncTolerance:null,syncToleranceDefault:null,systemLanguage:JE,tabIndex:Vr,tableValues:null,target:null,targetX:Vr,targetY:Vr,textAnchor:null,textDecoration:null,textRendering:null,textLength:null,timelineBegin:null,title:null,transformBehavior:null,type:null,typeOf:JE,to:null,transform:null,transformOrigin:null,u1:null,u2:null,underlinePosition:Vr,underlineThickness:Vr,unicode:null,unicodeBidi:null,unicodeRange:null,unitsPerEm:Vr,values:null,vAlphabetic:Vr,vMathematical:Vr,vectorEffect:null,vHanging:Vr,vIdeographic:Vr,version:null,vertAdvY:Vr,vertOriginX:Vr,vertOriginY:Vr,viewBox:null,viewTarget:null,visibility:null,width:null,widths:null,wordSpacing:null,writingMode:null,x:null,x1:null,x2:null,xChannelSelector:null,xHeight:Vr,y:null,y1:null,y2:null,yChannelSelector:null,z:null,zoomAndPan:null},space:"svg",transform:Gdn}),Vdn=jK({properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null},space:"xlink",transform(n,r){return"xlink:"+r.slice(5).toLowerCase()}}),Wdn=jK({attributes:{xmlnsxlink:"xmlns:xlink"},properties:{xmlnsXLink:null,xmlns:null},space:"xmlns",transform:qdn}),Kdn=jK({properties:{xmlBase:null,xmlLang:null,xmlSpace:null},space:"xml",transform(n,r){return"xml:"+r.slice(3).toLowerCase()}}),Zqr={classId:"classID",dataType:"datatype",itemId:"itemID",strokeDashArray:"strokeDasharray",strokeDashOffset:"strokeDashoffset",strokeLineCap:"strokeLinecap",strokeLineJoin:"strokeLinejoin",strokeMiterLimit:"strokeMiterlimit",typeOf:"typeof",xLinkActuate:"xlinkActuate",xLinkArcRole:"xlinkArcrole",xLinkHref:"xlinkHref",xLinkRole:"xlinkRole",xLinkShow:"xlinkShow",xLinkTitle:"xlinkTitle",xLinkType:"xlinkType",xmlnsXLink:"xmlnsXlink"},Qqr=/[A-Z]/g,cWt=/-[a-z]/g,eVr=/^data[-\w.:]+$/i;function RQe(n,r){const o=poe(r);let s=r,c=ay;if(o in n.normal)return n.property[n.normal[o]];if(o.length>4&&o.slice(0,4)==="data"&&eVr.test(r)){if(r.charAt(4)==="-"){const d=r.slice(5).replace(cWt,nVr);s="data"+d.charAt(0).toUpperCase()+d.slice(1)}else{const d=r.slice(4);if(!cWt.test(d)){let h=d.replace(Qqr,tVr);h.charAt(0)!=="-"&&(h="-"+h),r="data"+h}}c=NQe}return new c(s,r)}function tVr(n){return"-"+n.toLowerCase()}function nVr(n){return n.charAt(1).toUpperCase()}const BEe=zdn([Hdn,Xqr,Vdn,Wdn,Kdn],"html"),$K=zdn([Hdn,Jqr,Vdn,Wdn,Kdn],"svg");function uWt(n){const r=String(n||"").trim();return r?r.split(/[ \t\n\r\f]+/g):[]}function Ydn(n){return n.join(" ").trim()}var JV={},cqe,lWt;function rVr(){if(lWt)return cqe;lWt=1;var n=/\/\*[^*]*\*+([^/*][^*]*\*+)*\//g,r=/\n/g,o=/^\s*/,s=/^(\*?[-#/*\\\w]+(\[[0-9a-z_-]+\])?)\s*/,c=/^:\s*/,d=/^((?:'(?:\\'|.)*?'|"(?:\\"|.)*?"|\([^)]*?\)|[^};])+)/,h=/^[;\s]*/,b=/^\s+|\s+$/g,y=`
`,E="/",k="*",x="",O="comment",P="declaration";cqe=function($,U){if(typeof $!="string")throw new TypeError("First argument must be a string");if(!$)return[];U=U||{};var B=1,G=1;function K(Ue){var Z=Ue.match(r);Z&&(B+=Z.length);var Re=Ue.lastIndexOf(y);G=~Re?Ue.length-Re:G+Ue.length}function X(){var Ue={line:B,column:G};return function(Z){return Z.position=new ae(Ue),ye(),Z}}function ae(Ue){this.start=Ue,this.end={line:B,column:G},this.source=U.source}ae.prototype.content=$;function oe(Ue){var Z=new Error(U.source+":"+B+":"+G+": "+Ue);if(Z.reason=Ue,Z.filename=U.source,Z.line=B,Z.column=G,Z.source=$,!U.silent)throw Z}function he(Ue){var Z=Ue.exec($);if(Z){var Re=Z[0];return K(Re),$=$.slice(Re.length),Z}}function ye(){he(o)}function we(Ue){var Z;for(Ue=Ue||[];Z=ge();)Z!==!1&&Ue.push(Z);return Ue}function ge(){var Ue=X();if(!(E!=$.charAt(0)||k!=$.charAt(1))){for(var Z=2;x!=$.charAt(Z)&&(k!=$.charAt(Z)||E!=$.charAt(Z+1));)++Z;if(Z+=2,x===$.charAt(Z-1))return oe("End of comment missing");var Re=$.slice(2,Z-2);return G+=2,K(Re),$=$.slice(Z),G+=2,Ue({type:O,comment:Re})}}function Te(){var Ue=X(),Z=he(s);if(Z){if(ge(),!he(c))return oe("property missing ':'");var Re=he(d),_e=Ue({type:P,property:D(Z[0].replace(n,x)),value:Re?D(Re[0].replace(n,x)):x});return he(h),_e}}function xe(){var Ue=[];we(Ue);for(var Z;Z=Te();)Z!==!1&&(Ue.push(Z),we(Ue));return Ue}return ye(),xe()};function D($){return $?$.replace(b,x):x}return cqe}var dWt;function iVr(){if(dWt)return JV;dWt=1;var n=JV&&JV.__importDefault||function(s){return s&&s.__esModule?s:{default:s}};Object.defineProperty(JV,"__esModule",{value:!0}),JV.default=o;var r=n(rVr());function o(s,c){var d=null;if(!s||typeof s!="string")return d;var h=(0,r.default)(s),b=typeof c=="function";return h.forEach(function(y){if(y.type==="declaration"){var E=y.property,k=y.value;b?c(E,k,y):k&&(d=d||{},d[E]=k)}}),d}return JV}var kre={},fWt;function oVr(){if(fWt)return kre;fWt=1,Object.defineProperty(kre,"__esModule",{value:!0}),kre.camelCase=void 0;var n=/^--[a-zA-Z0-9_-]+$/,r=/-([a-z])/g,o=/^[^-]+$/,s=/^-(webkit|moz|ms|o|khtml)-/,c=/^-(ms)-/,d=function(E){return!E||o.test(E)||n.test(E)},h=function(E,k){return k.toUpperCase()},b=function(E,k){return"".concat(k,"-")},y=function(E,k){return k===void 0&&(k={}),d(E)?E:(E=E.toLowerCase(),k.reactCompat?E=E.replace(c,b):E=E.replace(s,b),E.replace(r,h))};return kre.camelCase=y,kre}var Tre,hWt;function aVr(){if(hWt)return Tre;hWt=1;var n=Tre&&Tre.__importDefault||function(c){return c&&c.__esModule?c:{default:c}},r=n(iVr()),o=oVr();function s(c,d){var h={};return!c||typeof c!="string"||(0,r.default)(c,function(b,y){b&&y&&(h[(0,o.camelCase)(b,d)]=y)}),h}return s.default=s,Tre=s,Tre}var sVr=aVr();const cVr=Y4(sVr),UEe=Xdn("end"),WI=Xdn("start");function Xdn(n){return r;function r(o){const s=o&&o.position&&o.position[n]||{};if(typeof s.line=="number"&&s.line>0&&typeof s.column=="number"&&s.column>0)return{line:s.line,column:s.column,offset:typeof s.offset=="number"&&s.offset>-1?s.offset:void 0}}}function uVr(n){const r=WI(n),o=UEe(n);if(r&&o)return{start:r,end:o}}function yie(n){return!n||typeof n!="object"?"":"position"in n||"type"in n?pWt(n.position):"start"in n||"end"in n?pWt(n):"line"in n||"column"in n?lKe(n):""}function lKe(n){return gWt(n&&n.line)+":"+gWt(n&&n.column)}function pWt(n){return lKe(n&&n.start)+"-"+lKe(n&&n.end)}function gWt(n){return n&&typeof n=="number"?n:1}class Z1 extends Error{constructor(r,o,s){super(),typeof o=="string"&&(s=o,o=void 0);let c="",d={},h=!1;if(o&&("line"in o&&"column"in o?d={place:o}:"start"in o&&"end"in o?d={place:o}:"type"in o?d={ancestors:[o],place:o.position}:d={...o}),typeof r=="string"?c=r:!d.cause&&r&&(h=!0,c=r.message,d.cause=r),!d.ruleId&&!d.source&&typeof s=="string"){const y=s.indexOf(":");y===-1?d.ruleId=s:(d.source=s.slice(0,y),d.ruleId=s.slice(y+1))}if(!d.place&&d.ancestors&&d.ancestors){const y=d.ancestors[d.ancestors.length-1];y&&(d.place=y.position)}const b=d.place&&"start"in d.place?d.place.start:d.place;this.ancestors=d.ancestors||void 0,this.cause=d.cause||void 0,this.column=b?b.column:void 0,this.fatal=void 0,this.file="",this.message=c,this.line=b?b.line:void 0,this.name=yie(d.place)||"1:1",this.place=d.place||void 0,this.reason=this.message,this.ruleId=d.ruleId||void 0,this.source=d.source||void 0,this.stack=h&&d.cause&&typeof d.cause.stack=="string"?d.cause.stack:"",this.actual=void 0,this.expected=void 0,this.note=void 0,this.url=void 0}}Z1.prototype.file="";Z1.prototype.name="";Z1.prototype.reason="";Z1.prototype.message="";Z1.prototype.stack="";Z1.prototype.column=void 0;Z1.prototype.line=void 0;Z1.prototype.ancestors=void 0;Z1.prototype.cause=void 0;Z1.prototype.fatal=void 0;Z1.prototype.place=void 0;Z1.prototype.ruleId=void 0;Z1.prototype.source=void 0;const PQe={}.hasOwnProperty,lVr=new Map,dVr=/[A-Z]/g,fVr=new Set(["table","tbody","thead","tfoot","tr"]),hVr=new Set(["td","th"]),Jdn="https://github.com/syntax-tree/hast-util-to-jsx-runtime";function pVr(n,r){if(!r||r.Fragment===void 0)throw new TypeError("Expected `Fragment` in options");const o=r.filePath||void 0;let s;if(r.development){if(typeof r.jsxDEV!="function")throw new TypeError("Expected `jsxDEV` in options when `development: true`");s=SVr(o,r.jsxDEV)}else{if(typeof r.jsx!="function")throw new TypeError("Expected `jsx` in production options");if(typeof r.jsxs!="function")throw new TypeError("Expected `jsxs` in production options");s=EVr(o,r.jsx,r.jsxs)}const c={Fragment:r.Fragment,ancestors:[],components:r.components||{},create:s,elementAttributeNameCase:r.elementAttributeNameCase||"react",evaluater:r.createEvaluater?r.createEvaluater():void 0,filePath:o,ignoreInvalidStyle:r.ignoreInvalidStyle||!1,passKeys:r.passKeys!==!1,passNode:r.passNode||!1,schema:r.space==="svg"?$K:BEe,stylePropertyNameCase:r.stylePropertyNameCase||"dom",tableCellAlignToStyle:r.tableCellAlignToStyle!==!1},d=Zdn(c,n,void 0);return d&&typeof d!="string"?d:c.create(n,c.Fragment,{children:d||void 0},void 0)}function Zdn(n,r,o){if(r.type==="element")return gVr(n,r,o);if(r.type==="mdxFlowExpression"||r.type==="mdxTextExpression")return mVr(n,r);if(r.type==="mdxJsxFlowElement"||r.type==="mdxJsxTextElement")return vVr(n,r,o);if(r.type==="mdxjsEsm")return bVr(n,r);if(r.type==="root")return wVr(n,r,o);if(r.type==="text")return yVr(n,r)}function gVr(n,r,o){const s=n.schema;let c=s;r.tagName.toLowerCase()==="svg"&&s.space==="html"&&(c=$K,n.schema=c),n.ancestors.push(r);const d=efn(n,r.tagName,!1),h=_Vr(n,r);let b=DQe(n,r);return fVr.has(r.tagName)&&(b=b.filter(function(y){return typeof y=="string"?!Kqr(y):!0})),Qdn(n,h,d,r),LQe(h,b),n.ancestors.pop(),n.schema=s,n.create(r,d,h,o)}function mVr(n,r){if(r.data&&r.data.estree&&n.evaluater){const s=r.data.estree.body[0];return s.type,n.evaluater.evaluateExpression(s.expression)}goe(n,r.position)}function bVr(n,r){if(r.data&&r.data.estree&&n.evaluater)return n.evaluater.evaluateProgram(r.data.estree);goe(n,r.position)}function vVr(n,r,o){const s=n.schema;let c=s;r.name==="svg"&&s.space==="html"&&(c=$K,n.schema=c),n.ancestors.push(r);const d=r.name===null?n.Fragment:efn(n,r.name,!0),h=kVr(n,r),b=DQe(n,r);return Qdn(n,h,d,r),LQe(h,b),n.ancestors.pop(),n.schema=s,n.create(r,d,h,o)}function wVr(n,r,o){const s={};return LQe(s,DQe(n,r)),n.create(r,n.Fragment,s,o)}function yVr(n,r){return r.value}function Qdn(n,r,o,s){typeof o!="string"&&o!==n.Fragment&&n.passNode&&(r.node=s)}function LQe(n,r){if(r.length>0){const o=r.length>1?r:r[0];o&&(n.children=o)}}function EVr(n,r,o){return s;function s(c,d,h,b){const E=Array.isArray(h.children)?o:r;return b?E(d,h,b):E(d,h)}}function SVr(n,r){return o;function o(s,c,d,h){const b=Array.isArray(d.children),y=WI(s);return r(c,d,h,b,{columnNumber:y?y.column-1:void 0,fileName:n,lineNumber:y?y.line:void 0},void 0)}}function _Vr(n,r){const o={};let s,c;for(c in r.properties)if(c!=="children"&&PQe.call(r.properties,c)){const d=TVr(n,c,r.properties[c]);if(d){const[h,b]=d;n.tableCellAlignToStyle&&h==="align"&&typeof b=="string"&&hVr.has(r.tagName)?s=b:o[h]=b}}if(s){const d=o.style||(o.style={});d[n.stylePropertyNameCase==="css"?"text-align":"textAlign"]=s}return o}function kVr(n,r){const o={};for(const s of r.attributes)if(s.type==="mdxJsxExpressionAttribute")if(s.data&&s.data.estree&&n.evaluater){const d=s.data.estree.body[0];d.type;const h=d.expression;h.type;const b=h.properties[0];b.type,Object.assign(o,n.evaluater.evaluateExpression(b.argument))}else goe(n,r.position);else{const c=s.name;let d;if(s.value&&typeof s.value=="object")if(s.value.data&&s.value.data.estree&&n.evaluater){const b=s.value.data.estree.body[0];b.type,d=n.evaluater.evaluateExpression(b.expression)}else goe(n,r.position);else d=s.value===null?!0:s.value;o[c]=d}return o}function DQe(n,r){const o=[];let s=-1;const c=n.passKeys?new Map:lVr;for(;++s<r.children.length;){const d=r.children[s];let h;if(n.passKeys){const y=d.type==="element"?d.tagName:d.type==="mdxJsxFlowElement"||d.type==="mdxJsxTextElement"?d.name:void 0;if(y){const E=c.get(y)||0;h=y+"-"+E,c.set(y,E+1)}}const b=Zdn(n,d,h);b!==void 0&&o.push(b)}return o}function TVr(n,r,o){const s=RQe(n.schema,r);if(!(o==null||typeof o=="number"&&Number.isNaN(o))){if(Array.isArray(o)&&(o=s.commaSeparated?Udn(o):Ydn(o)),s.property==="style"){let c=typeof o=="object"?o:xVr(n,String(o));return n.stylePropertyNameCase==="css"&&(c=CVr(c)),["style",c]}return[n.elementAttributeNameCase==="react"&&s.space?Zqr[s.property]||s.property:s.attribute,o]}}function xVr(n,r){try{return cVr(r,{reactCompat:!0})}catch(o){if(n.ignoreInvalidStyle)return{};const s=o,c=new Z1("Cannot parse `style` attribute",{ancestors:n.ancestors,cause:s,ruleId:"style",source:"hast-util-to-jsx-runtime"});throw c.file=n.filePath||void 0,c.url=Jdn+"#cannot-parse-style-attribute",c}}function efn(n,r,o){let s;if(!o)s={type:"Literal",value:r};else if(r.includes(".")){const c=r.split(".");let d=-1,h;for(;++d<c.length;){const b=oWt(c[d])?{type:"Identifier",name:c[d]}:{type:"Literal",value:c[d]};h=h?{type:"MemberExpression",object:h,property:b,computed:!!(d&&b.type==="Literal"),optional:!1}:b}s=h}else s=oWt(r)&&!/^[a-z]/.test(r)?{type:"Identifier",name:r}:{type:"Literal",value:r};if(s.type==="Literal"){const c=s.value;return PQe.call(n.components,c)?n.components[c]:c}if(n.evaluater)return n.evaluater.evaluateExpression(s);goe(n)}function goe(n,r){const o=new Z1("Cannot handle MDX estrees without `createEvaluater`",{ancestors:n.ancestors,place:r,ruleId:"mdx-estree",source:"hast-util-to-jsx-runtime"});throw o.file=n.filePath||void 0,o.url=Jdn+"#cannot-handle-mdx-estrees-without-createevaluater",o}function CVr(n){const r={};let o;for(o in n)PQe.call(n,o)&&(r[AVr(o)]=n[o]);return r}function AVr(n){let r=n.replace(dVr,IVr);return r.slice(0,3)==="ms-"&&(r="-"+r),r}function IVr(n){return"-"+n.toLowerCase()}const uqe={action:["form"],cite:["blockquote","del","ins","q"],data:["object"],formAction:["button","input"],href:["a","area","base","link"],icon:["menuitem"],itemId:null,manifest:["html"],ping:["a","area"],poster:["video"],src:["audio","embed","iframe","img","input","script","source","track","video"]},OVr={};function MQe(n,r){const o=OVr,s=typeof o.includeImageAlt=="boolean"?o.includeImageAlt:!0,c=typeof o.includeHtml=="boolean"?o.includeHtml:!0;return tfn(n,s,c)}function tfn(n,r,o){if(NVr(n)){if("value"in n)return n.type==="html"&&!o?"":n.value;if(r&&"alt"in n&&n.alt)return n.alt;if("children"in n)return mWt(n.children,r,o)}return Array.isArray(n)?mWt(n,r,o):""}function mWt(n,r,o){const s=[];let c=-1;for(;++c<n.length;)s[c]=tfn(n[c],r,o);return s.join("")}function NVr(n){return!!(n&&typeof n=="object")}const bWt=document.createElement("i");function jQe(n){const r="&"+n+";";bWt.innerHTML=r;const o=bWt.textContent;return o.charCodeAt(o.length-1)===59&&n!=="semi"||o===r?!1:o}function p2(n,r,o,s){const c=n.length;let d=0,h;if(r<0?r=-r>c?0:c+r:r=r>c?c:r,o=o>0?o:0,s.length<1e4)h=Array.from(s),h.unshift(r,o),n.splice(...h);else for(o&&n.splice(r,o);d<s.length;)h=s.slice(d,d+1e4),h.unshift(r,0),n.splice(...h),d+=1e4,r+=1e4}function L_(n,r){return n.length>0?(p2(n,n.length,0,r),n):r}const vWt={}.hasOwnProperty;function nfn(n){const r={};let o=-1;for(;++o<n.length;)RVr(r,n[o]);return r}function RVr(n,r){let o;for(o in r){const c=(vWt.call(n,o)?n[o]:void 0)||(n[o]={}),d=r[o];let h;if(d)for(h in d){vWt.call(c,h)||(c[h]=[]);const b=d[h];PVr(c[h],Array.isArray(b)?b:b?[b]:[])}}}function PVr(n,r){let o=-1;const s=[];for(;++o<r.length;)(r[o].add==="after"?n:s).push(r[o]);p2(n,0,0,s)}function rfn(n,r){const o=Number.parseInt(n,r);return o<9||o===11||o>13&&o<32||o>126&&o<160||o>55295&&o<57344||o>64975&&o<65008||(o&65535)===65535||(o&65535)===65534||o>1114111?"�":String.fromCodePoint(o)}function zx(n){return n.replace(/[\t\n\r ]+/g," ").replace(/^ | $/g,"").toLowerCase().toUpperCase()}const sv=qM(/[A-Za-z]/),K1=qM(/[\dA-Za-z]/),LVr=qM(/[#-'*+\--9=?A-Z^-~]/);function Kwe(n){return n!==null&&(n<32||n===127)}const dKe=qM(/\d/),DVr=qM(/[\dA-Fa-f]/),MVr=qM(/[!-/:-@[-`{-~]/);function ea(n){return n!==null&&n<-2}function cd(n){return n!==null&&(n<0||n===32)}function mc(n){return n===-2||n===-1||n===32}const zEe=qM(new RegExp("\\p{P}|\\p{S}","u")),sU=qM(/\s/);function qM(n){return r;function r(o){return o!==null&&o>-1&&n.test(String.fromCharCode(o))}}function FK(n){const r=[];let o=-1,s=0,c=0;for(;++o<n.length;){const d=n.charCodeAt(o);let h="";if(d===37&&K1(n.charCodeAt(o+1))&&K1(n.charCodeAt(o+2)))c=2;else if(d<128)/[!#$&-;=?-Z_a-z~]/.test(String.fromCharCode(d))||(h=String.fromCharCode(d));else if(d>55295&&d<57344){const b=n.charCodeAt(o+1);d<56320&&b>56319&&b<57344?(h=String.fromCharCode(d,b),c=1):h="�"}else h=String.fromCharCode(d);h&&(r.push(n.slice(s,o),encodeURIComponent(h)),s=o+c+1,h=""),c&&(o+=c,c=0)}return r.join("")+n.slice(s)}function Kc(n,r,o,s){const c=s?s-1:Number.POSITIVE_INFINITY;let d=0;return h;function h(y){return mc(y)?(n.enter(o),b(y)):r(y)}function b(y){return mc(y)&&d++<c?(n.consume(y),b):(n.exit(o),r(y))}}const jVr={tokenize:$Vr};function $Vr(n){const r=n.attempt(this.parser.constructs.contentInitial,s,c);let o;return r;function s(b){if(b===null){n.consume(b);return}return n.enter("lineEnding"),n.consume(b),n.exit("lineEnding"),Kc(n,r,"linePrefix")}function c(b){return n.enter("paragraph"),d(b)}function d(b){const y=n.enter("chunkText",{contentType:"text",previous:o});return o&&(o.next=y),o=y,h(b)}function h(b){if(b===null){n.exit("chunkText"),n.exit("paragraph"),n.consume(b);return}return ea(b)?(n.consume(b),n.exit("chunkText"),d):(n.consume(b),h)}}const FVr={tokenize:BVr},wWt={tokenize:UVr};function BVr(n){const r=this,o=[];let s=0,c,d,h;return b;function b(K){if(s<o.length){const X=o[s];return r.containerState=X[1],n.attempt(X[0].continuation,y,E)(K)}return E(K)}function y(K){if(s++,r.containerState._closeFlow){r.containerState._closeFlow=void 0,c&&G();const X=r.events.length;let ae=X,oe;for(;ae--;)if(r.events[ae][0]==="exit"&&r.events[ae][1].type==="chunkFlow"){oe=r.events[ae][1].end;break}B(s);let he=X;for(;he<r.events.length;)r.events[he][1].end={...oe},he++;return p2(r.events,ae+1,0,r.events.slice(X)),r.events.length=he,E(K)}return b(K)}function E(K){if(s===o.length){if(!c)return O(K);if(c.currentConstruct&&c.currentConstruct.concrete)return D(K);r.interrupt=!!(c.currentConstruct&&!c._gfmTableDynamicInterruptHack)}return r.containerState={},n.check(wWt,k,x)(K)}function k(K){return c&&G(),B(s),O(K)}function x(K){return r.parser.lazy[r.now().line]=s!==o.length,h=r.now().offset,D(K)}function O(K){return r.containerState={},n.attempt(wWt,P,D)(K)}function P(K){return s++,o.push([r.currentConstruct,r.containerState]),O(K)}function D(K){if(K===null){c&&G(),B(0),n.consume(K);return}return c=c||r.parser.flow(r.now()),n.enter("chunkFlow",{_tokenizer:c,contentType:"flow",previous:d}),$(K)}function $(K){if(K===null){U(n.exit("chunkFlow"),!0),B(0),n.consume(K);return}return ea(K)?(n.consume(K),U(n.exit("chunkFlow")),s=0,r.interrupt=void 0,b):(n.consume(K),$)}function U(K,X){const ae=r.sliceStream(K);if(X&&ae.push(null),K.previous=d,d&&(d.next=K),d=K,c.defineSkip(K.start),c.write(ae),r.parser.lazy[K.start.line]){let oe=c.events.length;for(;oe--;)if(c.events[oe][1].start.offset<h&&(!c.events[oe][1].end||c.events[oe][1].end.offset>h))return;const he=r.events.length;let ye=he,we,ge;for(;ye--;)if(r.events[ye][0]==="exit"&&r.events[ye][1].type==="chunkFlow"){if(we){ge=r.events[ye][1].end;break}we=!0}for(B(s),oe=he;oe<r.events.length;)r.events[oe][1].end={...ge},oe++;p2(r.events,ye+1,0,r.events.slice(he)),r.events.length=oe}}function B(K){let X=o.length;for(;X-- >K;){const ae=o[X];r.containerState=ae[1],ae[0].exit.call(r,n)}o.length=K}function G(){c.write([null]),d=void 0,c=void 0,r.containerState._closeFlow=void 0}}function UVr(n,r,o){return Kc(n,n.attempt(this.parser.constructs.document,r,o),"linePrefix",this.parser.constructs.disable.null.includes("codeIndented")?void 0:4)}function fK(n){if(n===null||cd(n)||sU(n))return 1;if(zEe(n))return 2}function HEe(n,r,o){const s=[];let c=-1;for(;++c<n.length;){const d=n[c].resolveAll;d&&!s.includes(d)&&(r=d(r,o),s.push(d))}return r}const fKe={name:"attention",resolveAll:zVr,tokenize:HVr};function zVr(n,r){let o=-1,s,c,d,h,b,y,E,k;for(;++o<n.length;)if(n[o][0]==="enter"&&n[o][1].type==="attentionSequence"&&n[o][1]._close){for(s=o;s--;)if(n[s][0]==="exit"&&n[s][1].type==="attentionSequence"&&n[s][1]._open&&r.sliceSerialize(n[s][1]).charCodeAt(0)===r.sliceSerialize(n[o][1]).charCodeAt(0)){if((n[s][1]._close||n[o][1]._open)&&(n[o][1].end.offset-n[o][1].start.offset)%3&&!((n[s][1].end.offset-n[s][1].start.offset+n[o][1].end.offset-n[o][1].start.offset)%3))continue;y=n[s][1].end.offset-n[s][1].start.offset>1&&n[o][1].end.offset-n[o][1].start.offset>1?2:1;const x={...n[s][1].end},O={...n[o][1].start};yWt(x,-y),yWt(O,y),h={type:y>1?"strongSequence":"emphasisSequence",start:x,end:{...n[s][1].end}},b={type:y>1?"strongSequence":"emphasisSequence",start:{...n[o][1].start},end:O},d={type:y>1?"strongText":"emphasisText",start:{...n[s][1].end},end:{...n[o][1].start}},c={type:y>1?"strong":"emphasis",start:{...h.start},end:{...b.end}},n[s][1].end={...h.start},n[o][1].start={...b.end},E=[],n[s][1].end.offset-n[s][1].start.offset&&(E=L_(E,[["enter",n[s][1],r],["exit",n[s][1],r]])),E=L_(E,[["enter",c,r],["enter",h,r],["exit",h,r],["enter",d,r]]),E=L_(E,HEe(r.parser.constructs.insideSpan.null,n.slice(s+1,o),r)),E=L_(E,[["exit",d,r],["enter",b,r],["exit",b,r],["exit",c,r]]),n[o][1].end.offset-n[o][1].start.offset?(k=2,E=L_(E,[["enter",n[o][1],r],["exit",n[o][1],r]])):k=0,p2(n,s-1,o-s+3,E),o=s+E.length-k-2;break}}for(o=-1;++o<n.length;)n[o][1].type==="attentionSequence"&&(n[o][1].type="data");return n}function HVr(n,r){const o=this.parser.constructs.attentionMarkers.null,s=this.previous,c=fK(s);let d;return h;function h(y){return d=y,n.enter("attentionSequence"),b(y)}function b(y){if(y===d)return n.consume(y),b;const E=n.exit("attentionSequence"),k=fK(y),x=!k||k===2&&c||o.includes(y),O=!c||c===2&&k||o.includes(s);return E._open=!!(d===42?x:x&&(c||!O)),E._close=!!(d===42?O:O&&(k||!x)),r(y)}}function yWt(n,r){n.column+=r,n.offset+=r,n._bufferIndex+=r}const GVr={name:"autolink",tokenize:qVr};function qVr(n,r,o){let s=0;return c;function c(P){return n.enter("autolink"),n.enter("autolinkMarker"),n.consume(P),n.exit("autolinkMarker"),n.enter("autolinkProtocol"),d}function d(P){return sv(P)?(n.consume(P),h):P===64?o(P):E(P)}function h(P){return P===43||P===45||P===46||K1(P)?(s=1,b(P)):E(P)}function b(P){return P===58?(n.consume(P),s=0,y):(P===43||P===45||P===46||K1(P))&&s++<32?(n.consume(P),b):(s=0,E(P))}function y(P){return P===62?(n.exit("autolinkProtocol"),n.enter("autolinkMarker"),n.consume(P),n.exit("autolinkMarker"),n.exit("autolink"),r):P===null||P===32||P===60||Kwe(P)?o(P):(n.consume(P),y)}function E(P){return P===64?(n.consume(P),k):LVr(P)?(n.consume(P),E):o(P)}function k(P){return K1(P)?x(P):o(P)}function x(P){return P===46?(n.consume(P),s=0,k):P===62?(n.exit("autolinkProtocol").type="autolinkEmail",n.enter("autolinkMarker"),n.consume(P),n.exit("autolinkMarker"),n.exit("autolink"),r):O(P)}function O(P){if((P===45||K1(P))&&s++<63){const D=P===45?O:x;return n.consume(P),D}return o(P)}}const Joe={partial:!0,tokenize:VVr};function VVr(n,r,o){return s;function s(d){return mc(d)?Kc(n,c,"linePrefix")(d):c(d)}function c(d){return d===null||ea(d)?r(d):o(d)}}const ifn={continuation:{tokenize:KVr},exit:YVr,name:"blockQuote",tokenize:WVr};function WVr(n,r,o){const s=this;return c;function c(h){if(h===62){const b=s.containerState;return b.open||(n.enter("blockQuote",{_container:!0}),b.open=!0),n.enter("blockQuotePrefix"),n.enter("blockQuoteMarker"),n.consume(h),n.exit("blockQuoteMarker"),d}return o(h)}function d(h){return mc(h)?(n.enter("blockQuotePrefixWhitespace"),n.consume(h),n.exit("blockQuotePrefixWhitespace"),n.exit("blockQuotePrefix"),r):(n.exit("blockQuotePrefix"),r(h))}}function KVr(n,r,o){const s=this;return c;function c(h){return mc(h)?Kc(n,d,"linePrefix",s.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(h):d(h)}function d(h){return n.attempt(ifn,r,o)(h)}}function YVr(n){n.exit("blockQuote")}const ofn={name:"characterEscape",tokenize:XVr};function XVr(n,r,o){return s;function s(d){return n.enter("characterEscape"),n.enter("escapeMarker"),n.consume(d),n.exit("escapeMarker"),c}function c(d){return MVr(d)?(n.enter("characterEscapeValue"),n.consume(d),n.exit("characterEscapeValue"),n.exit("characterEscape"),r):o(d)}}const afn={name:"characterReference",tokenize:JVr};function JVr(n,r,o){const s=this;let c=0,d,h;return b;function b(x){return n.enter("characterReference"),n.enter("characterReferenceMarker"),n.consume(x),n.exit("characterReferenceMarker"),y}function y(x){return x===35?(n.enter("characterReferenceMarkerNumeric"),n.consume(x),n.exit("characterReferenceMarkerNumeric"),E):(n.enter("characterReferenceValue"),d=31,h=K1,k(x))}function E(x){return x===88||x===120?(n.enter("characterReferenceMarkerHexadecimal"),n.consume(x),n.exit("characterReferenceMarkerHexadecimal"),n.enter("characterReferenceValue"),d=6,h=DVr,k):(n.enter("characterReferenceValue"),d=7,h=dKe,k(x))}function k(x){if(x===59&&c){const O=n.exit("characterReferenceValue");return h===K1&&!jQe(s.sliceSerialize(O))?o(x):(n.enter("characterReferenceMarker"),n.consume(x),n.exit("characterReferenceMarker"),n.exit("characterReference"),r)}return h(x)&&c++<d?(n.consume(x),k):o(x)}}const EWt={partial:!0,tokenize:QVr},SWt={concrete:!0,name:"codeFenced",tokenize:ZVr};function ZVr(n,r,o){const s=this,c={partial:!0,tokenize:ae};let d=0,h=0,b;return y;function y(oe){return E(oe)}function E(oe){const he=s.events[s.events.length-1];return d=he&&he[1].type==="linePrefix"?he[2].sliceSerialize(he[1],!0).length:0,b=oe,n.enter("codeFenced"),n.enter("codeFencedFence"),n.enter("codeFencedFenceSequence"),k(oe)}function k(oe){return oe===b?(h++,n.consume(oe),k):h<3?o(oe):(n.exit("codeFencedFenceSequence"),mc(oe)?Kc(n,x,"whitespace")(oe):x(oe))}function x(oe){return oe===null||ea(oe)?(n.exit("codeFencedFence"),s.interrupt?r(oe):n.check(EWt,$,X)(oe)):(n.enter("codeFencedFenceInfo"),n.enter("chunkString",{contentType:"string"}),O(oe))}function O(oe){return oe===null||ea(oe)?(n.exit("chunkString"),n.exit("codeFencedFenceInfo"),x(oe)):mc(oe)?(n.exit("chunkString"),n.exit("codeFencedFenceInfo"),Kc(n,P,"whitespace")(oe)):oe===96&&oe===b?o(oe):(n.consume(oe),O)}function P(oe){return oe===null||ea(oe)?x(oe):(n.enter("codeFencedFenceMeta"),n.enter("chunkString",{contentType:"string"}),D(oe))}function D(oe){return oe===null||ea(oe)?(n.exit("chunkString"),n.exit("codeFencedFenceMeta"),x(oe)):oe===96&&oe===b?o(oe):(n.consume(oe),D)}function $(oe){return n.attempt(c,X,U)(oe)}function U(oe){return n.enter("lineEnding"),n.consume(oe),n.exit("lineEnding"),B}function B(oe){return d>0&&mc(oe)?Kc(n,G,"linePrefix",d+1)(oe):G(oe)}function G(oe){return oe===null||ea(oe)?n.check(EWt,$,X)(oe):(n.enter("codeFlowValue"),K(oe))}function K(oe){return oe===null||ea(oe)?(n.exit("codeFlowValue"),G(oe)):(n.consume(oe),K)}function X(oe){return n.exit("codeFenced"),r(oe)}function ae(oe,he,ye){let we=0;return ge;function ge(Re){return oe.enter("lineEnding"),oe.consume(Re),oe.exit("lineEnding"),Te}function Te(Re){return oe.enter("codeFencedFence"),mc(Re)?Kc(oe,xe,"linePrefix",s.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(Re):xe(Re)}function xe(Re){return Re===b?(oe.enter("codeFencedFenceSequence"),Ue(Re)):ye(Re)}function Ue(Re){return Re===b?(we++,oe.consume(Re),Ue):we>=h?(oe.exit("codeFencedFenceSequence"),mc(Re)?Kc(oe,Z,"whitespace")(Re):Z(Re)):ye(Re)}function Z(Re){return Re===null||ea(Re)?(oe.exit("codeFencedFence"),he(Re)):ye(Re)}}}function QVr(n,r,o){const s=this;return c;function c(h){return h===null?o(h):(n.enter("lineEnding"),n.consume(h),n.exit("lineEnding"),d)}function d(h){return s.parser.lazy[s.now().line]?o(h):r(h)}}const lqe={name:"codeIndented",tokenize:tWr},eWr={partial:!0,tokenize:nWr};function tWr(n,r,o){const s=this;return c;function c(E){return n.enter("codeIndented"),Kc(n,d,"linePrefix",5)(E)}function d(E){const k=s.events[s.events.length-1];return k&&k[1].type==="linePrefix"&&k[2].sliceSerialize(k[1],!0).length>=4?h(E):o(E)}function h(E){return E===null?y(E):ea(E)?n.attempt(eWr,h,y)(E):(n.enter("codeFlowValue"),b(E))}function b(E){return E===null||ea(E)?(n.exit("codeFlowValue"),h(E)):(n.consume(E),b)}function y(E){return n.exit("codeIndented"),r(E)}}function nWr(n,r,o){const s=this;return c;function c(h){return s.parser.lazy[s.now().line]?o(h):ea(h)?(n.enter("lineEnding"),n.consume(h),n.exit("lineEnding"),c):Kc(n,d,"linePrefix",5)(h)}function d(h){const b=s.events[s.events.length-1];return b&&b[1].type==="linePrefix"&&b[2].sliceSerialize(b[1],!0).length>=4?r(h):ea(h)?c(h):o(h)}}const rWr={name:"codeText",previous:oWr,resolve:iWr,tokenize:aWr};function iWr(n){let r=n.length-4,o=3,s,c;if((n[o][1].type==="lineEnding"||n[o][1].type==="space")&&(n[r][1].type==="lineEnding"||n[r][1].type==="space")){for(s=o;++s<r;)if(n[s][1].type==="codeTextData"){n[o][1].type="codeTextPadding",n[r][1].type="codeTextPadding",o+=2,r-=2;break}}for(s=o-1,r++;++s<=r;)c===void 0?s!==r&&n[s][1].type!=="lineEnding"&&(c=s):(s===r||n[s][1].type==="lineEnding")&&(n[c][1].type="codeTextData",s!==c+2&&(n[c][1].end=n[s-1][1].end,n.splice(c+2,s-c-2),r-=s-c-2,s=c+2),c=void 0);return n}function oWr(n){return n!==96||this.events[this.events.length-1][1].type==="characterEscape"}function aWr(n,r,o){let s=0,c,d;return h;function h(x){return n.enter("codeText"),n.enter("codeTextSequence"),b(x)}function b(x){return x===96?(n.consume(x),s++,b):(n.exit("codeTextSequence"),y(x))}function y(x){return x===null?o(x):x===32?(n.enter("space"),n.consume(x),n.exit("space"),y):x===96?(d=n.enter("codeTextSequence"),c=0,k(x)):ea(x)?(n.enter("lineEnding"),n.consume(x),n.exit("lineEnding"),y):(n.enter("codeTextData"),E(x))}function E(x){return x===null||x===32||x===96||ea(x)?(n.exit("codeTextData"),y(x)):(n.consume(x),E)}function k(x){return x===96?(n.consume(x),c++,k):c===s?(n.exit("codeTextSequence"),n.exit("codeText"),r(x)):(d.type="codeTextData",E(x))}}class sWr{constructor(r){this.left=r?[...r]:[],this.right=[]}get(r){if(r<0||r>=this.left.length+this.right.length)throw new RangeError("Cannot access index `"+r+"` in a splice buffer of size `"+(this.left.length+this.right.length)+"`");return r<this.left.length?this.left[r]:this.right[this.right.length-r+this.left.length-1]}get length(){return this.left.length+this.right.length}shift(){return this.setCursor(0),this.right.pop()}slice(r,o){const s=o??Number.POSITIVE_INFINITY;return s<this.left.length?this.left.slice(r,s):r>this.left.length?this.right.slice(this.right.length-s+this.left.length,this.right.length-r+this.left.length).reverse():this.left.slice(r).concat(this.right.slice(this.right.length-s+this.left.length).reverse())}splice(r,o,s){const c=o||0;this.setCursor(Math.trunc(r));const d=this.right.splice(this.right.length-c,Number.POSITIVE_INFINITY);return s&&xre(this.left,s),d.reverse()}pop(){return this.setCursor(Number.POSITIVE_INFINITY),this.left.pop()}push(r){this.setCursor(Number.POSITIVE_INFINITY),this.left.push(r)}pushMany(r){this.setCursor(Number.POSITIVE_INFINITY),xre(this.left,r)}unshift(r){this.setCursor(0),this.right.push(r)}unshiftMany(r){this.setCursor(0),xre(this.right,r.reverse())}setCursor(r){if(!(r===this.left.length||r>this.left.length&&this.right.length===0||r<0&&this.left.length===0))if(r<this.left.length){const o=this.left.splice(r,Number.POSITIVE_INFINITY);xre(this.right,o.reverse())}else{const o=this.right.splice(this.left.length+this.right.length-r,Number.POSITIVE_INFINITY);xre(this.left,o.reverse())}}}function xre(n,r){let o=0;if(r.length<1e4)n.push(...r);else for(;o<r.length;)n.push(...r.slice(o,o+1e4)),o+=1e4}function sfn(n){const r={};let o=-1,s,c,d,h,b,y,E;const k=new sWr(n);for(;++o<k.length;){for(;o in r;)o=r[o];if(s=k.get(o),o&&s[1].type==="chunkFlow"&&k.get(o-1)[1].type==="listItemPrefix"&&(y=s[1]._tokenizer.events,d=0,d<y.length&&y[d][1].type==="lineEndingBlank"&&(d+=2),d<y.length&&y[d][1].type==="content"))for(;++d<y.length&&y[d][1].type!=="content";)y[d][1].type==="chunkText"&&(y[d][1]._isInFirstContentOfListItem=!0,d++);if(s[0]==="enter")s[1].contentType&&(Object.assign(r,cWr(k,o)),o=r[o],E=!0);else if(s[1]._container){for(d=o,c=void 0;d--;)if(h=k.get(d),h[1].type==="lineEnding"||h[1].type==="lineEndingBlank")h[0]==="enter"&&(c&&(k.get(c)[1].type="lineEndingBlank"),h[1].type="lineEnding",c=d);else if(!(h[1].type==="linePrefix"||h[1].type==="listItemIndent"))break;c&&(s[1].end={...k.get(c)[1].start},b=k.slice(c,o),b.unshift(s),k.splice(c,o-c+1,b))}}return p2(n,0,Number.POSITIVE_INFINITY,k.slice(0)),!E}function cWr(n,r){const o=n.get(r)[1],s=n.get(r)[2];let c=r-1;const d=[];let h=o._tokenizer;h||(h=s.parser[o.contentType](o.start),o._contentTypeTextTrailing&&(h._contentTypeTextTrailing=!0));const b=h.events,y=[],E={};let k,x,O=-1,P=o,D=0,$=0;const U=[$];for(;P;){for(;n.get(++c)[1]!==P;);d.push(c),P._tokenizer||(k=s.sliceStream(P),P.next||k.push(null),x&&h.defineSkip(P.start),P._isInFirstContentOfListItem&&(h._gfmTasklistFirstContentOfListItem=!0),h.write(k),P._isInFirstContentOfListItem&&(h._gfmTasklistFirstContentOfListItem=void 0)),x=P,P=P.next}for(P=o;++O<b.length;)b[O][0]==="exit"&&b[O-1][0]==="enter"&&b[O][1].type===b[O-1][1].type&&b[O][1].start.line!==b[O][1].end.line&&($=O+1,U.push($),P._tokenizer=void 0,P.previous=void 0,P=P.next);for(h.events=[],P?(P._tokenizer=void 0,P.previous=void 0):U.pop(),O=U.length;O--;){const B=b.slice(U[O],U[O+1]),G=d.pop();y.push([G,G+B.length-1]),n.splice(G,2,B)}for(y.reverse(),O=-1;++O<y.length;)E[D+y[O][0]]=D+y[O][1],D+=y[O][1]-y[O][0]-1;return E}const uWr={resolve:dWr,tokenize:fWr},lWr={partial:!0,tokenize:hWr};function dWr(n){return sfn(n),n}function fWr(n,r){let o;return s;function s(b){return n.enter("content"),o=n.enter("chunkContent",{contentType:"content"}),c(b)}function c(b){return b===null?d(b):ea(b)?n.check(lWr,h,d)(b):(n.consume(b),c)}function d(b){return n.exit("chunkContent"),n.exit("content"),r(b)}function h(b){return n.consume(b),n.exit("chunkContent"),o.next=n.enter("chunkContent",{contentType:"content",previous:o}),o=o.next,c}}function hWr(n,r,o){const s=this;return c;function c(h){return n.exit("chunkContent"),n.enter("lineEnding"),n.consume(h),n.exit("lineEnding"),Kc(n,d,"linePrefix")}function d(h){if(h===null||ea(h))return o(h);const b=s.events[s.events.length-1];return!s.parser.constructs.disable.null.includes("codeIndented")&&b&&b[1].type==="linePrefix"&&b[2].sliceSerialize(b[1],!0).length>=4?r(h):n.interrupt(s.parser.constructs.flow,o,r)(h)}}function cfn(n,r,o,s,c,d,h,b,y){const E=y||Number.POSITIVE_INFINITY;let k=0;return x;function x(B){return B===60?(n.enter(s),n.enter(c),n.enter(d),n.consume(B),n.exit(d),O):B===null||B===32||B===41||Kwe(B)?o(B):(n.enter(s),n.enter(h),n.enter(b),n.enter("chunkString",{contentType:"string"}),$(B))}function O(B){return B===62?(n.enter(d),n.consume(B),n.exit(d),n.exit(c),n.exit(s),r):(n.enter(b),n.enter("chunkString",{contentType:"string"}),P(B))}function P(B){return B===62?(n.exit("chunkString"),n.exit(b),O(B)):B===null||B===60||ea(B)?o(B):(n.consume(B),B===92?D:P)}function D(B){return B===60||B===62||B===92?(n.consume(B),P):P(B)}function $(B){return!k&&(B===null||B===41||cd(B))?(n.exit("chunkString"),n.exit(b),n.exit(h),n.exit(s),r(B)):k<E&&B===40?(n.consume(B),k++,$):B===41?(n.consume(B),k--,$):B===null||B===32||B===40||Kwe(B)?o(B):(n.consume(B),B===92?U:$)}function U(B){return B===40||B===41||B===92?(n.consume(B),$):$(B)}}function ufn(n,r,o,s,c,d){const h=this;let b=0,y;return E;function E(P){return n.enter(s),n.enter(c),n.consume(P),n.exit(c),n.enter(d),k}function k(P){return b>999||P===null||P===91||P===93&&!y||P===94&&!b&&"_hiddenFootnoteSupport"in h.parser.constructs?o(P):P===93?(n.exit(d),n.enter(c),n.consume(P),n.exit(c),n.exit(s),r):ea(P)?(n.enter("lineEnding"),n.consume(P),n.exit("lineEnding"),k):(n.enter("chunkString",{contentType:"string"}),x(P))}function x(P){return P===null||P===91||P===93||ea(P)||b++>999?(n.exit("chunkString"),k(P)):(n.consume(P),y||(y=!mc(P)),P===92?O:x)}function O(P){return P===91||P===92||P===93?(n.consume(P),b++,x):x(P)}}function lfn(n,r,o,s,c,d){let h;return b;function b(O){return O===34||O===39||O===40?(n.enter(s),n.enter(c),n.consume(O),n.exit(c),h=O===40?41:O,y):o(O)}function y(O){return O===h?(n.enter(c),n.consume(O),n.exit(c),n.exit(s),r):(n.enter(d),E(O))}function E(O){return O===h?(n.exit(d),y(h)):O===null?o(O):ea(O)?(n.enter("lineEnding"),n.consume(O),n.exit("lineEnding"),Kc(n,E,"linePrefix")):(n.enter("chunkString",{contentType:"string"}),k(O))}function k(O){return O===h||O===null||ea(O)?(n.exit("chunkString"),E(O)):(n.consume(O),O===92?x:k)}function x(O){return O===h||O===92?(n.consume(O),k):k(O)}}function Eie(n,r){let o;return s;function s(c){return ea(c)?(n.enter("lineEnding"),n.consume(c),n.exit("lineEnding"),o=!0,s):mc(c)?Kc(n,s,o?"linePrefix":"lineSuffix")(c):r(c)}}const pWr={name:"definition",tokenize:mWr},gWr={partial:!0,tokenize:bWr};function mWr(n,r,o){const s=this;let c;return d;function d(P){return n.enter("definition"),h(P)}function h(P){return ufn.call(s,n,b,o,"definitionLabel","definitionLabelMarker","definitionLabelString")(P)}function b(P){return c=zx(s.sliceSerialize(s.events[s.events.length-1][1]).slice(1,-1)),P===58?(n.enter("definitionMarker"),n.consume(P),n.exit("definitionMarker"),y):o(P)}function y(P){return cd(P)?Eie(n,E)(P):E(P)}function E(P){return cfn(n,k,o,"definitionDestination","definitionDestinationLiteral","definitionDestinationLiteralMarker","definitionDestinationRaw","definitionDestinationString")(P)}function k(P){return n.attempt(gWr,x,x)(P)}function x(P){return mc(P)?Kc(n,O,"whitespace")(P):O(P)}function O(P){return P===null||ea(P)?(n.exit("definition"),s.parser.defined.push(c),r(P)):o(P)}}function bWr(n,r,o){return s;function s(b){return cd(b)?Eie(n,c)(b):o(b)}function c(b){return lfn(n,d,o,"definitionTitle","definitionTitleMarker","definitionTitleString")(b)}function d(b){return mc(b)?Kc(n,h,"whitespace")(b):h(b)}function h(b){return b===null||ea(b)?r(b):o(b)}}const vWr={name:"hardBreakEscape",tokenize:wWr};function wWr(n,r,o){return s;function s(d){return n.enter("hardBreakEscape"),n.consume(d),c}function c(d){return ea(d)?(n.exit("hardBreakEscape"),r(d)):o(d)}}const yWr={name:"headingAtx",resolve:EWr,tokenize:SWr};function EWr(n,r){let o=n.length-2,s=3,c,d;return n[s][1].type==="whitespace"&&(s+=2),o-2>s&&n[o][1].type==="whitespace"&&(o-=2),n[o][1].type==="atxHeadingSequence"&&(s===o-1||o-4>s&&n[o-2][1].type==="whitespace")&&(o-=s+1===o?2:4),o>s&&(c={type:"atxHeadingText",start:n[s][1].start,end:n[o][1].end},d={type:"chunkText",start:n[s][1].start,end:n[o][1].end,contentType:"text"},p2(n,s,o-s+1,[["enter",c,r],["enter",d,r],["exit",d,r],["exit",c,r]])),n}function SWr(n,r,o){let s=0;return c;function c(k){return n.enter("atxHeading"),d(k)}function d(k){return n.enter("atxHeadingSequence"),h(k)}function h(k){return k===35&&s++<6?(n.consume(k),h):k===null||cd(k)?(n.exit("atxHeadingSequence"),b(k)):o(k)}function b(k){return k===35?(n.enter("atxHeadingSequence"),y(k)):k===null||ea(k)?(n.exit("atxHeading"),r(k)):mc(k)?Kc(n,b,"whitespace")(k):(n.enter("atxHeadingText"),E(k))}function y(k){return k===35?(n.consume(k),y):(n.exit("atxHeadingSequence"),b(k))}function E(k){return k===null||k===35||cd(k)?(n.exit("atxHeadingText"),b(k)):(n.consume(k),E)}}const _Wr=["address","article","aside","base","basefont","blockquote","body","caption","center","col","colgroup","dd","details","dialog","dir","div","dl","dt","fieldset","figcaption","figure","footer","form","frame","frameset","h1","h2","h3","h4","h5","h6","head","header","hr","html","iframe","legend","li","link","main","menu","menuitem","nav","noframes","ol","optgroup","option","p","param","search","section","summary","table","tbody","td","tfoot","th","thead","title","tr","track","ul"],_Wt=["pre","script","style","textarea"],kWr={concrete:!0,name:"htmlFlow",resolveTo:CWr,tokenize:AWr},TWr={partial:!0,tokenize:OWr},xWr={partial:!0,tokenize:IWr};function CWr(n){let r=n.length;for(;r--&&!(n[r][0]==="enter"&&n[r][1].type==="htmlFlow"););return r>1&&n[r-2][1].type==="linePrefix"&&(n[r][1].start=n[r-2][1].start,n[r+1][1].start=n[r-2][1].start,n.splice(r-2,2)),n}function AWr(n,r,o){const s=this;let c,d,h,b,y;return E;function E(Ie){return k(Ie)}function k(Ie){return n.enter("htmlFlow"),n.enter("htmlFlowData"),n.consume(Ie),x}function x(Ie){return Ie===33?(n.consume(Ie),O):Ie===47?(n.consume(Ie),d=!0,$):Ie===63?(n.consume(Ie),c=3,s.interrupt?r:de):sv(Ie)?(n.consume(Ie),h=String.fromCharCode(Ie),U):o(Ie)}function O(Ie){return Ie===45?(n.consume(Ie),c=2,P):Ie===91?(n.consume(Ie),c=5,b=0,D):sv(Ie)?(n.consume(Ie),c=4,s.interrupt?r:de):o(Ie)}function P(Ie){return Ie===45?(n.consume(Ie),s.interrupt?r:de):o(Ie)}function D(Ie){const $t="CDATA[";return Ie===$t.charCodeAt(b++)?(n.consume(Ie),b===$t.length?s.interrupt?r:xe:D):o(Ie)}function $(Ie){return sv(Ie)?(n.consume(Ie),h=String.fromCharCode(Ie),U):o(Ie)}function U(Ie){if(Ie===null||Ie===47||Ie===62||cd(Ie)){const $t=Ie===47,An=h.toLowerCase();return!$t&&!d&&_Wt.includes(An)?(c=1,s.interrupt?r(Ie):xe(Ie)):_Wr.includes(h.toLowerCase())?(c=6,$t?(n.consume(Ie),B):s.interrupt?r(Ie):xe(Ie)):(c=7,s.interrupt&&!s.parser.lazy[s.now().line]?o(Ie):d?G(Ie):K(Ie))}return Ie===45||K1(Ie)?(n.consume(Ie),h+=String.fromCharCode(Ie),U):o(Ie)}function B(Ie){return Ie===62?(n.consume(Ie),s.interrupt?r:xe):o(Ie)}function G(Ie){return mc(Ie)?(n.consume(Ie),G):ge(Ie)}function K(Ie){return Ie===47?(n.consume(Ie),ge):Ie===58||Ie===95||sv(Ie)?(n.consume(Ie),X):mc(Ie)?(n.consume(Ie),K):ge(Ie)}function X(Ie){return Ie===45||Ie===46||Ie===58||Ie===95||K1(Ie)?(n.consume(Ie),X):ae(Ie)}function ae(Ie){return Ie===61?(n.consume(Ie),oe):mc(Ie)?(n.consume(Ie),ae):K(Ie)}function oe(Ie){return Ie===null||Ie===60||Ie===61||Ie===62||Ie===96?o(Ie):Ie===34||Ie===39?(n.consume(Ie),y=Ie,he):mc(Ie)?(n.consume(Ie),oe):ye(Ie)}function he(Ie){return Ie===y?(n.consume(Ie),y=null,we):Ie===null||ea(Ie)?o(Ie):(n.consume(Ie),he)}function ye(Ie){return Ie===null||Ie===34||Ie===39||Ie===47||Ie===60||Ie===61||Ie===62||Ie===96||cd(Ie)?ae(Ie):(n.consume(Ie),ye)}function we(Ie){return Ie===47||Ie===62||mc(Ie)?K(Ie):o(Ie)}function ge(Ie){return Ie===62?(n.consume(Ie),Te):o(Ie)}function Te(Ie){return Ie===null||ea(Ie)?xe(Ie):mc(Ie)?(n.consume(Ie),Te):o(Ie)}function xe(Ie){return Ie===45&&c===2?(n.consume(Ie),_e):Ie===60&&c===1?(n.consume(Ie),qe):Ie===62&&c===4?(n.consume(Ie),We):Ie===63&&c===3?(n.consume(Ie),de):Ie===93&&c===5?(n.consume(Ie),Xe):ea(Ie)&&(c===6||c===7)?(n.exit("htmlFlowData"),n.check(TWr,ht,Ue)(Ie)):Ie===null||ea(Ie)?(n.exit("htmlFlowData"),Ue(Ie)):(n.consume(Ie),xe)}function Ue(Ie){return n.check(xWr,Z,ht)(Ie)}function Z(Ie){return n.enter("lineEnding"),n.consume(Ie),n.exit("lineEnding"),Re}function Re(Ie){return Ie===null||ea(Ie)?Ue(Ie):(n.enter("htmlFlowData"),xe(Ie))}function _e(Ie){return Ie===45?(n.consume(Ie),de):xe(Ie)}function qe(Ie){return Ie===47?(n.consume(Ie),h="",it):xe(Ie)}function it(Ie){if(Ie===62){const $t=h.toLowerCase();return _Wt.includes($t)?(n.consume(Ie),We):xe(Ie)}return sv(Ie)&&h.length<8?(n.consume(Ie),h+=String.fromCharCode(Ie),it):xe(Ie)}function Xe(Ie){return Ie===93?(n.consume(Ie),de):xe(Ie)}function de(Ie){return Ie===62?(n.consume(Ie),We):Ie===45&&c===2?(n.consume(Ie),de):xe(Ie)}function We(Ie){return Ie===null||ea(Ie)?(n.exit("htmlFlowData"),ht(Ie)):(n.consume(Ie),We)}function ht(Ie){return n.exit("htmlFlow"),r(Ie)}}function IWr(n,r,o){const s=this;return c;function c(h){return ea(h)?(n.enter("lineEnding"),n.consume(h),n.exit("lineEnding"),d):o(h)}function d(h){return s.parser.lazy[s.now().line]?o(h):r(h)}}function OWr(n,r,o){return s;function s(c){return n.enter("lineEnding"),n.consume(c),n.exit("lineEnding"),n.attempt(Joe,r,o)}}const NWr={name:"htmlText",tokenize:RWr};function RWr(n,r,o){const s=this;let c,d,h;return b;function b(de){return n.enter("htmlText"),n.enter("htmlTextData"),n.consume(de),y}function y(de){return de===33?(n.consume(de),E):de===47?(n.consume(de),ae):de===63?(n.consume(de),K):sv(de)?(n.consume(de),ye):o(de)}function E(de){return de===45?(n.consume(de),k):de===91?(n.consume(de),d=0,D):sv(de)?(n.consume(de),G):o(de)}function k(de){return de===45?(n.consume(de),P):o(de)}function x(de){return de===null?o(de):de===45?(n.consume(de),O):ea(de)?(h=x,qe(de)):(n.consume(de),x)}function O(de){return de===45?(n.consume(de),P):x(de)}function P(de){return de===62?_e(de):de===45?O(de):x(de)}function D(de){const We="CDATA[";return de===We.charCodeAt(d++)?(n.consume(de),d===We.length?$:D):o(de)}function $(de){return de===null?o(de):de===93?(n.consume(de),U):ea(de)?(h=$,qe(de)):(n.consume(de),$)}function U(de){return de===93?(n.consume(de),B):$(de)}function B(de){return de===62?_e(de):de===93?(n.consume(de),B):$(de)}function G(de){return de===null||de===62?_e(de):ea(de)?(h=G,qe(de)):(n.consume(de),G)}function K(de){return de===null?o(de):de===63?(n.consume(de),X):ea(de)?(h=K,qe(de)):(n.consume(de),K)}function X(de){return de===62?_e(de):K(de)}function ae(de){return sv(de)?(n.consume(de),oe):o(de)}function oe(de){return de===45||K1(de)?(n.consume(de),oe):he(de)}function he(de){return ea(de)?(h=he,qe(de)):mc(de)?(n.consume(de),he):_e(de)}function ye(de){return de===45||K1(de)?(n.consume(de),ye):de===47||de===62||cd(de)?we(de):o(de)}function we(de){return de===47?(n.consume(de),_e):de===58||de===95||sv(de)?(n.consume(de),ge):ea(de)?(h=we,qe(de)):mc(de)?(n.consume(de),we):_e(de)}function ge(de){return de===45||de===46||de===58||de===95||K1(de)?(n.consume(de),ge):Te(de)}function Te(de){return de===61?(n.consume(de),xe):ea(de)?(h=Te,qe(de)):mc(de)?(n.consume(de),Te):we(de)}function xe(de){return de===null||de===60||de===61||de===62||de===96?o(de):de===34||de===39?(n.consume(de),c=de,Ue):ea(de)?(h=xe,qe(de)):mc(de)?(n.consume(de),xe):(n.consume(de),Z)}function Ue(de){return de===c?(n.consume(de),c=void 0,Re):de===null?o(de):ea(de)?(h=Ue,qe(de)):(n.consume(de),Ue)}function Z(de){return de===null||de===34||de===39||de===60||de===61||de===96?o(de):de===47||de===62||cd(de)?we(de):(n.consume(de),Z)}function Re(de){return de===47||de===62||cd(de)?we(de):o(de)}function _e(de){return de===62?(n.consume(de),n.exit("htmlTextData"),n.exit("htmlText"),r):o(de)}function qe(de){return n.exit("htmlTextData"),n.enter("lineEnding"),n.consume(de),n.exit("lineEnding"),it}function it(de){return mc(de)?Kc(n,Xe,"linePrefix",s.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(de):Xe(de)}function Xe(de){return n.enter("htmlTextData"),h(de)}}const $Qe={name:"labelEnd",resolveAll:MWr,resolveTo:jWr,tokenize:$Wr},PWr={tokenize:FWr},LWr={tokenize:BWr},DWr={tokenize:UWr};function MWr(n){let r=-1;const o=[];for(;++r<n.length;){const s=n[r][1];if(o.push(n[r]),s.type==="labelImage"||s.type==="labelLink"||s.type==="labelEnd"){const c=s.type==="labelImage"?4:2;s.type="data",r+=c}}return n.length!==o.length&&p2(n,0,n.length,o),n}function jWr(n,r){let o=n.length,s=0,c,d,h,b;for(;o--;)if(c=n[o][1],d){if(c.type==="link"||c.type==="labelLink"&&c._inactive)break;n[o][0]==="enter"&&c.type==="labelLink"&&(c._inactive=!0)}else if(h){if(n[o][0]==="enter"&&(c.type==="labelImage"||c.type==="labelLink")&&!c._balanced&&(d=o,c.type!=="labelLink")){s=2;break}}else c.type==="labelEnd"&&(h=o);const y={type:n[d][1].type==="labelLink"?"link":"image",start:{...n[d][1].start},end:{...n[n.length-1][1].end}},E={type:"label",start:{...n[d][1].start},end:{...n[h][1].end}},k={type:"labelText",start:{...n[d+s+2][1].end},end:{...n[h-2][1].start}};return b=[["enter",y,r],["enter",E,r]],b=L_(b,n.slice(d+1,d+s+3)),b=L_(b,[["enter",k,r]]),b=L_(b,HEe(r.parser.constructs.insideSpan.null,n.slice(d+s+4,h-3),r)),b=L_(b,[["exit",k,r],n[h-2],n[h-1],["exit",E,r]]),b=L_(b,n.slice(h+1)),b=L_(b,[["exit",y,r]]),p2(n,d,n.length,b),n}function $Wr(n,r,o){const s=this;let c=s.events.length,d,h;for(;c--;)if((s.events[c][1].type==="labelImage"||s.events[c][1].type==="labelLink")&&!s.events[c][1]._balanced){d=s.events[c][1];break}return b;function b(O){return d?d._inactive?x(O):(h=s.parser.defined.includes(zx(s.sliceSerialize({start:d.end,end:s.now()}))),n.enter("labelEnd"),n.enter("labelMarker"),n.consume(O),n.exit("labelMarker"),n.exit("labelEnd"),y):o(O)}function y(O){return O===40?n.attempt(PWr,k,h?k:x)(O):O===91?n.attempt(LWr,k,h?E:x)(O):h?k(O):x(O)}function E(O){return n.attempt(DWr,k,x)(O)}function k(O){return r(O)}function x(O){return d._balanced=!0,o(O)}}function FWr(n,r,o){return s;function s(x){return n.enter("resource"),n.enter("resourceMarker"),n.consume(x),n.exit("resourceMarker"),c}function c(x){return cd(x)?Eie(n,d)(x):d(x)}function d(x){return x===41?k(x):cfn(n,h,b,"resourceDestination","resourceDestinationLiteral","resourceDestinationLiteralMarker","resourceDestinationRaw","resourceDestinationString",32)(x)}function h(x){return cd(x)?Eie(n,y)(x):k(x)}function b(x){return o(x)}function y(x){return x===34||x===39||x===40?lfn(n,E,o,"resourceTitle","resourceTitleMarker","resourceTitleString")(x):k(x)}function E(x){return cd(x)?Eie(n,k)(x):k(x)}function k(x){return x===41?(n.enter("resourceMarker"),n.consume(x),n.exit("resourceMarker"),n.exit("resource"),r):o(x)}}function BWr(n,r,o){const s=this;return c;function c(b){return ufn.call(s,n,d,h,"reference","referenceMarker","referenceString")(b)}function d(b){return s.parser.defined.includes(zx(s.sliceSerialize(s.events[s.events.length-1][1]).slice(1,-1)))?r(b):o(b)}function h(b){return o(b)}}function UWr(n,r,o){return s;function s(d){return n.enter("reference"),n.enter("referenceMarker"),n.consume(d),n.exit("referenceMarker"),c}function c(d){return d===93?(n.enter("referenceMarker"),n.consume(d),n.exit("referenceMarker"),n.exit("reference"),r):o(d)}}const zWr={name:"labelStartImage",resolveAll:$Qe.resolveAll,tokenize:HWr};function HWr(n,r,o){const s=this;return c;function c(b){return n.enter("labelImage"),n.enter("labelImageMarker"),n.consume(b),n.exit("labelImageMarker"),d}function d(b){return b===91?(n.enter("labelMarker"),n.consume(b),n.exit("labelMarker"),n.exit("labelImage"),h):o(b)}function h(b){return b===94&&"_hiddenFootnoteSupport"in s.parser.constructs?o(b):r(b)}}const GWr={name:"labelStartLink",resolveAll:$Qe.resolveAll,tokenize:qWr};function qWr(n,r,o){const s=this;return c;function c(h){return n.enter("labelLink"),n.enter("labelMarker"),n.consume(h),n.exit("labelMarker"),n.exit("labelLink"),d}function d(h){return h===94&&"_hiddenFootnoteSupport"in s.parser.constructs?o(h):r(h)}}const dqe={name:"lineEnding",tokenize:VWr};function VWr(n,r){return o;function o(s){return n.enter("lineEnding"),n.consume(s),n.exit("lineEnding"),Kc(n,r,"linePrefix")}}const Pve={name:"thematicBreak",tokenize:WWr};function WWr(n,r,o){let s=0,c;return d;function d(E){return n.enter("thematicBreak"),h(E)}function h(E){return c=E,b(E)}function b(E){return E===c?(n.enter("thematicBreakSequence"),y(E)):s>=3&&(E===null||ea(E))?(n.exit("thematicBreak"),r(E)):o(E)}function y(E){return E===c?(n.consume(E),s++,y):(n.exit("thematicBreakSequence"),mc(E)?Kc(n,b,"whitespace")(E):b(E))}}const Hw={continuation:{tokenize:JWr},exit:QWr,name:"list",tokenize:XWr},KWr={partial:!0,tokenize:eKr},YWr={partial:!0,tokenize:ZWr};function XWr(n,r,o){const s=this,c=s.events[s.events.length-1];let d=c&&c[1].type==="linePrefix"?c[2].sliceSerialize(c[1],!0).length:0,h=0;return b;function b(P){const D=s.containerState.type||(P===42||P===43||P===45?"listUnordered":"listOrdered");if(D==="listUnordered"?!s.containerState.marker||P===s.containerState.marker:dKe(P)){if(s.containerState.type||(s.containerState.type=D,n.enter(D,{_container:!0})),D==="listUnordered")return n.enter("listItemPrefix"),P===42||P===45?n.check(Pve,o,E)(P):E(P);if(!s.interrupt||P===49)return n.enter("listItemPrefix"),n.enter("listItemValue"),y(P)}return o(P)}function y(P){return dKe(P)&&++h<10?(n.consume(P),y):(!s.interrupt||h<2)&&(s.containerState.marker?P===s.containerState.marker:P===41||P===46)?(n.exit("listItemValue"),E(P)):o(P)}function E(P){return n.enter("listItemMarker"),n.consume(P),n.exit("listItemMarker"),s.containerState.marker=s.containerState.marker||P,n.check(Joe,s.interrupt?o:k,n.attempt(KWr,O,x))}function k(P){return s.containerState.initialBlankLine=!0,d++,O(P)}function x(P){return mc(P)?(n.enter("listItemPrefixWhitespace"),n.consume(P),n.exit("listItemPrefixWhitespace"),O):o(P)}function O(P){return s.containerState.size=d+s.sliceSerialize(n.exit("listItemPrefix"),!0).length,r(P)}}function JWr(n,r,o){const s=this;return s.containerState._closeFlow=void 0,n.check(Joe,c,d);function c(b){return s.containerState.furtherBlankLines=s.containerState.furtherBlankLines||s.containerState.initialBlankLine,Kc(n,r,"listItemIndent",s.containerState.size+1)(b)}function d(b){return s.containerState.furtherBlankLines||!mc(b)?(s.containerState.furtherBlankLines=void 0,s.containerState.initialBlankLine=void 0,h(b)):(s.containerState.furtherBlankLines=void 0,s.containerState.initialBlankLine=void 0,n.attempt(YWr,r,h)(b))}function h(b){return s.containerState._closeFlow=!0,s.interrupt=void 0,Kc(n,n.attempt(Hw,r,o),"linePrefix",s.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(b)}}function ZWr(n,r,o){const s=this;return Kc(n,c,"listItemIndent",s.containerState.size+1);function c(d){const h=s.events[s.events.length-1];return h&&h[1].type==="listItemIndent"&&h[2].sliceSerialize(h[1],!0).length===s.containerState.size?r(d):o(d)}}function QWr(n){n.exit(this.containerState.type)}function eKr(n,r,o){const s=this;return Kc(n,c,"listItemPrefixWhitespace",s.parser.constructs.disable.null.includes("codeIndented")?void 0:5);function c(d){const h=s.events[s.events.length-1];return!mc(d)&&h&&h[1].type==="listItemPrefixWhitespace"?r(d):o(d)}}const kWt={name:"setextUnderline",resolveTo:tKr,tokenize:nKr};function tKr(n,r){let o=n.length,s,c,d;for(;o--;)if(n[o][0]==="enter"){if(n[o][1].type==="content"){s=o;break}n[o][1].type==="paragraph"&&(c=o)}else n[o][1].type==="content"&&n.splice(o,1),!d&&n[o][1].type==="definition"&&(d=o);const h={type:"setextHeading",start:{...n[s][1].start},end:{...n[n.length-1][1].end}};return n[c][1].type="setextHeadingText",d?(n.splice(c,0,["enter",h,r]),n.splice(d+1,0,["exit",n[s][1],r]),n[s][1].end={...n[d][1].end}):n[s][1]=h,n.push(["exit",h,r]),n}function nKr(n,r,o){const s=this;let c;return d;function d(E){let k=s.events.length,x;for(;k--;)if(s.events[k][1].type!=="lineEnding"&&s.events[k][1].type!=="linePrefix"&&s.events[k][1].type!=="content"){x=s.events[k][1].type==="paragraph";break}return!s.parser.lazy[s.now().line]&&(s.interrupt||x)?(n.enter("setextHeadingLine"),c=E,h(E)):o(E)}function h(E){return n.enter("setextHeadingLineSequence"),b(E)}function b(E){return E===c?(n.consume(E),b):(n.exit("setextHeadingLineSequence"),mc(E)?Kc(n,y,"lineSuffix")(E):y(E))}function y(E){return E===null||ea(E)?(n.exit("setextHeadingLine"),r(E)):o(E)}}const rKr={tokenize:iKr};function iKr(n){const r=this,o=n.attempt(Joe,s,n.attempt(this.parser.constructs.flowInitial,c,Kc(n,n.attempt(this.parser.constructs.flow,c,n.attempt(uWr,c)),"linePrefix")));return o;function s(d){if(d===null){n.consume(d);return}return n.enter("lineEndingBlank"),n.consume(d),n.exit("lineEndingBlank"),r.currentConstruct=void 0,o}function c(d){if(d===null){n.consume(d);return}return n.enter("lineEnding"),n.consume(d),n.exit("lineEnding"),r.currentConstruct=void 0,o}}const oKr={resolveAll:ffn()},aKr=dfn("string"),sKr=dfn("text");function dfn(n){return{resolveAll:ffn(n==="text"?cKr:void 0),tokenize:r};function r(o){const s=this,c=this.parser.constructs[n],d=o.attempt(c,h,b);return h;function h(k){return E(k)?d(k):b(k)}function b(k){if(k===null){o.consume(k);return}return o.enter("data"),o.consume(k),y}function y(k){return E(k)?(o.exit("data"),d(k)):(o.consume(k),y)}function E(k){if(k===null)return!0;const x=c[k];let O=-1;if(x)for(;++O<x.length;){const P=x[O];if(!P.previous||P.previous.call(s,s.previous))return!0}return!1}}}function ffn(n){return r;function r(o,s){let c=-1,d;for(;++c<=o.length;)d===void 0?o[c]&&o[c][1].type==="data"&&(d=c,c++):(!o[c]||o[c][1].type!=="data")&&(c!==d+2&&(o[d][1].end=o[c-1][1].end,o.splice(d+2,c-d-2),c=d+2),d=void 0);return n?n(o,s):o}}function cKr(n,r){let o=0;for(;++o<=n.length;)if((o===n.length||n[o][1].type==="lineEnding")&&n[o-1][1].type==="data"){const s=n[o-1][1],c=r.sliceStream(s);let d=c.length,h=-1,b=0,y;for(;d--;){const E=c[d];if(typeof E=="string"){for(h=E.length;E.charCodeAt(h-1)===32;)b++,h--;if(h)break;h=-1}else if(E===-2)y=!0,b++;else if(E!==-1){d++;break}}if(r._contentTypeTextTrailing&&o===n.length&&(b=0),b){const E={type:o===n.length||y||b<2?"lineSuffix":"hardBreakTrailing",start:{_bufferIndex:d?h:s.start._bufferIndex+h,_index:s.start._index+d,line:s.end.line,column:s.end.column-b,offset:s.end.offset-b},end:{...s.end}};s.end={...E.start},s.start.offset===s.end.offset?Object.assign(s,E):(n.splice(o,0,["enter",E,r],["exit",E,r]),o+=2)}o++}return n}const uKr={42:Hw,43:Hw,45:Hw,48:Hw,49:Hw,50:Hw,51:Hw,52:Hw,53:Hw,54:Hw,55:Hw,56:Hw,57:Hw,62:ifn},lKr={91:pWr},dKr={[-2]:lqe,[-1]:lqe,32:lqe},fKr={35:yWr,42:Pve,45:[kWt,Pve],60:kWr,61:kWt,95:Pve,96:SWt,126:SWt},hKr={38:afn,92:ofn},pKr={[-5]:dqe,[-4]:dqe,[-3]:dqe,33:zWr,38:afn,42:fKe,60:[GVr,NWr],91:GWr,92:[vWr,ofn],93:$Qe,95:fKe,96:rWr},gKr={null:[fKe,oKr]},mKr={null:[42,95]},bKr={null:[]},vKr=Object.freeze(Object.defineProperty({__proto__:null,attentionMarkers:mKr,contentInitial:lKr,disable:bKr,document:uKr,flow:fKr,flowInitial:dKr,insideSpan:gKr,string:hKr,text:pKr},Symbol.toStringTag,{value:"Module"}));function wKr(n,r,o){let s={_bufferIndex:-1,_index:0,line:o&&o.line||1,column:o&&o.column||1,offset:o&&o.offset||0};const c={},d=[];let h=[],b=[];const y={attempt:he(ae),check:he(oe),consume:G,enter:K,exit:X,interrupt:he(oe,{interrupt:!0})},E={code:null,containerState:{},defineSkip:$,events:[],now:D,parser:n,previous:null,sliceSerialize:O,sliceStream:P,write:x};let k=r.tokenize.call(E,y);return r.resolveAll&&d.push(r),E;function x(Te){return h=L_(h,Te),U(),h[h.length-1]!==null?[]:(ye(r,0),E.events=HEe(d,E.events,E),E.events)}function O(Te,xe){return EKr(P(Te),xe)}function P(Te){return yKr(h,Te)}function D(){const{_bufferIndex:Te,_index:xe,line:Ue,column:Z,offset:Re}=s;return{_bufferIndex:Te,_index:xe,line:Ue,column:Z,offset:Re}}function $(Te){c[Te.line]=Te.column,ge()}function U(){let Te;for(;s._index<h.length;){const xe=h[s._index];if(typeof xe=="string")for(Te=s._index,s._bufferIndex<0&&(s._bufferIndex=0);s._index===Te&&s._bufferIndex<xe.length;)B(xe.charCodeAt(s._bufferIndex));else B(xe)}}function B(Te){k=k(Te)}function G(Te){ea(Te)?(s.line++,s.column=1,s.offset+=Te===-3?2:1,ge()):Te!==-1&&(s.column++,s.offset++),s._bufferIndex<0?s._index++:(s._bufferIndex++,s._bufferIndex===h[s._index].length&&(s._bufferIndex=-1,s._index++)),E.previous=Te}function K(Te,xe){const Ue=xe||{};return Ue.type=Te,Ue.start=D(),E.events.push(["enter",Ue,E]),b.push(Ue),Ue}function X(Te){const xe=b.pop();return xe.end=D(),E.events.push(["exit",xe,E]),xe}function ae(Te,xe){ye(Te,xe.from)}function oe(Te,xe){xe.restore()}function he(Te,xe){return Ue;function Ue(Z,Re,_e){let qe,it,Xe,de;return Array.isArray(Z)?ht(Z):"tokenize"in Z?ht([Z]):We(Z);function We(Yt){return Or;function Or(Nt){const on=Nt!==null&&Yt[Nt],br=Nt!==null&&Yt.null,wi=[...Array.isArray(on)?on:on?[on]:[],...Array.isArray(br)?br:br?[br]:[]];return ht(wi)(Nt)}}function ht(Yt){return qe=Yt,it=0,Yt.length===0?_e:Ie(Yt[it])}function Ie(Yt){return Or;function Or(Nt){return de=we(),Xe=Yt,Yt.partial||(E.currentConstruct=Yt),Yt.name&&E.parser.constructs.disable.null.includes(Yt.name)?An():Yt.tokenize.call(xe?Object.assign(Object.create(E),xe):E,y,$t,An)(Nt)}}function $t(Yt){return Te(Xe,de),Re}function An(Yt){return de.restore(),++it<qe.length?Ie(qe[it]):_e}}}function ye(Te,xe){Te.resolveAll&&!d.includes(Te)&&d.push(Te),Te.resolve&&p2(E.events,xe,E.events.length-xe,Te.resolve(E.events.slice(xe),E)),Te.resolveTo&&(E.events=Te.resolveTo(E.events,E))}function we(){const Te=D(),xe=E.previous,Ue=E.currentConstruct,Z=E.events.length,Re=Array.from(b);return{from:Z,restore:_e};function _e(){s=Te,E.previous=xe,E.currentConstruct=Ue,E.events.length=Z,b=Re,ge()}}function ge(){s.line in c&&s.column<2&&(s.column=c[s.line],s.offset+=c[s.line]-1)}}function yKr(n,r){const o=r.start._index,s=r.start._bufferIndex,c=r.end._index,d=r.end._bufferIndex;let h;if(o===c)h=[n[o].slice(s,d)];else{if(h=n.slice(o,c),s>-1){const b=h[0];typeof b=="string"?h[0]=b.slice(s):h.shift()}d>0&&h.push(n[c].slice(0,d))}return h}function EKr(n,r){let o=-1;const s=[];let c;for(;++o<n.length;){const d=n[o];let h;if(typeof d=="string")h=d;else switch(d){case-5:{h="\r";break}case-4:{h=`
`;break}case-3:{h=`\r
`;break}case-2:{h=r?" ":"	";break}case-1:{if(!r&&c)continue;h=" ";break}default:h=String.fromCharCode(d)}c=d===-2,s.push(h)}return s.join("")}function SKr(n){const s={constructs:nfn([vKr,...(n||{}).extensions||[]]),content:c(jVr),defined:[],document:c(FVr),flow:c(rKr),lazy:{},string:c(aKr),text:c(sKr)};return s;function c(d){return h;function h(b){return wKr(s,d,b)}}}function _Kr(n){for(;!sfn(n););return n}const TWt=/[\0\t\n\r]/g;function kKr(){let n=1,r="",o=!0,s;return c;function c(d,h,b){const y=[];let E,k,x,O,P;for(d=r+(typeof d=="string"?d.toString():new TextDecoder(h||void 0).decode(d)),x=0,r="",o&&(d.charCodeAt(0)===65279&&x++,o=void 0);x<d.length;){if(TWt.lastIndex=x,E=TWt.exec(d),O=E&&E.index!==void 0?E.index:d.length,P=d.charCodeAt(O),!E){r=d.slice(x);break}if(P===10&&x===O&&s)y.push(-3),s=void 0;else switch(s&&(y.push(-5),s=void 0),x<O&&(y.push(d.slice(x,O)),n+=O-x),P){case 0:{y.push(65533),n++;break}case 9:{for(k=Math.ceil(n/4)*4,y.push(-2);n++<k;)y.push(-1);break}case 10:{y.push(-4),n=1;break}default:s=!0,n=1}x=O+1}return b&&(s&&y.push(-5),r&&y.push(r),y.push(null)),y}}const TKr=/\\([!-/:-@[-`{-~])|&(#(?:\d{1,7}|x[\da-f]{1,6})|[\da-z]{1,31});/gi;function xKr(n){return n.replace(TKr,CKr)}function CKr(n,r,o){if(r)return r;if(o.charCodeAt(0)===35){const c=o.charCodeAt(1),d=c===120||c===88;return rfn(o.slice(d?2:1),d?16:10)}return jQe(o)||n}const hfn={}.hasOwnProperty;function AKr(n,r,o){return typeof r!="string"&&(o=r,r=void 0),IKr(o)(_Kr(SKr(o).document().write(kKr()(n,r,!0))))}function IKr(n){const r={transforms:[],canContainEols:["emphasis","fragment","heading","paragraph","strong"],enter:{autolink:d(lo),autolinkProtocol:we,autolinkEmail:we,atxHeading:d(Ko),blockQuote:d(br),characterEscape:we,characterReference:we,codeFenced:d(wi),codeFencedFenceInfo:h,codeFencedFenceMeta:h,codeIndented:d(wi,h),codeText:d(no,h),codeTextData:we,data:we,codeFlowValue:we,definition:d(Ri),definitionDestinationString:h,definitionLabelString:h,definitionTitleString:h,emphasis:d(ro),hardBreakEscape:d(Ca),hardBreakTrailing:d(Ca),htmlFlow:d(qi,h),htmlFlowData:we,htmlText:d(qi,h),htmlTextData:we,image:d(yi),label:h,link:d(lo),listItem:d(Fc),listItemValue:O,listOrdered:d(Eo,x),listUnordered:d(Eo),paragraph:d(Yo),reference:Ie,referenceString:h,resourceDestinationString:h,resourceTitleString:h,setextHeading:d(Ko),strong:d(zl),thematicBreak:d(Yc)},exit:{atxHeading:y(),atxHeadingSequence:ae,autolink:y(),autolinkEmail:on,autolinkProtocol:Nt,blockQuote:y(),characterEscapeValue:ge,characterReferenceMarkerHexadecimal:An,characterReferenceMarkerNumeric:An,characterReferenceValue:Yt,characterReference:Or,codeFenced:y(U),codeFencedFence:$,codeFencedFenceInfo:P,codeFencedFenceMeta:D,codeFlowValue:ge,codeIndented:y(B),codeText:y(Re),codeTextData:ge,data:ge,definition:y(),definitionDestinationString:X,definitionLabelString:G,definitionTitleString:K,emphasis:y(),hardBreakEscape:y(xe),hardBreakTrailing:y(xe),htmlFlow:y(Ue),htmlFlowData:ge,htmlText:y(Z),htmlTextData:ge,image:y(qe),label:Xe,labelText:it,lineEnding:Te,link:y(_e),listItem:y(),listOrdered:y(),listUnordered:y(),paragraph:y(),referenceString:$t,resourceDestinationString:de,resourceTitleString:We,resource:ht,setextHeading:y(ye),setextHeadingLineSequence:he,setextHeadingText:oe,strong:y(),thematicBreak:y()}};pfn(r,(n||{}).mdastExtensions||[]);const o={};return s;function s(Sn){let Pn={type:"root",children:[]};const Kr={stack:[Pn],tokenStack:[],config:r,enter:b,exit:E,buffer:h,resume:k,data:o},Ei=[];let Sa=-1;for(;++Sa<Sn.length;)if(Sn[Sa][1].type==="listOrdered"||Sn[Sa][1].type==="listUnordered")if(Sn[Sa][0]==="enter")Ei.push(Sa);else{const bs=Ei.pop();Sa=c(Sn,bs,Sa)}for(Sa=-1;++Sa<Sn.length;){const bs=r[Sn[Sa][0]];hfn.call(bs,Sn[Sa][1].type)&&bs[Sn[Sa][1].type].call(Object.assign({sliceSerialize:Sn[Sa][2].sliceSerialize},Kr),Sn[Sa][1])}if(Kr.tokenStack.length>0){const bs=Kr.tokenStack[Kr.tokenStack.length-1];(bs[1]||xWt).call(Kr,void 0,bs[0])}for(Pn.position={start:KD(Sn.length>0?Sn[0][1].start:{line:1,column:1,offset:0}),end:KD(Sn.length>0?Sn[Sn.length-2][1].end:{line:1,column:1,offset:0})},Sa=-1;++Sa<r.transforms.length;)Pn=r.transforms[Sa](Pn)||Pn;return Pn}function c(Sn,Pn,Kr){let Ei=Pn-1,Sa=-1,bs=!1,Cc,Ac,Xc,rc;for(;++Ei<=Kr;){const js=Sn[Ei];switch(js[1].type){case"listUnordered":case"listOrdered":case"blockQuote":{js[0]==="enter"?Sa++:Sa--,rc=void 0;break}case"lineEndingBlank":{js[0]==="enter"&&(Cc&&!rc&&!Sa&&!Xc&&(Xc=Ei),rc=void 0);break}case"linePrefix":case"listItemValue":case"listItemMarker":case"listItemPrefix":case"listItemPrefixWhitespace":break;default:rc=void 0}if(!Sa&&js[0]==="enter"&&js[1].type==="listItemPrefix"||Sa===-1&&js[0]==="exit"&&(js[1].type==="listUnordered"||js[1].type==="listOrdered")){if(Cc){let Jc=Ei;for(Ac=void 0;Jc--;){const Id=Sn[Jc];if(Id[1].type==="lineEnding"||Id[1].type==="lineEndingBlank"){if(Id[0]==="exit")continue;Ac&&(Sn[Ac][1].type="lineEndingBlank",bs=!0),Id[1].type="lineEnding",Ac=Jc}else if(!(Id[1].type==="linePrefix"||Id[1].type==="blockQuotePrefix"||Id[1].type==="blockQuotePrefixWhitespace"||Id[1].type==="blockQuoteMarker"||Id[1].type==="listItemIndent"))break}Xc&&(!Ac||Xc<Ac)&&(Cc._spread=!0),Cc.end=Object.assign({},Ac?Sn[Ac][1].start:js[1].end),Sn.splice(Ac||Ei,0,["exit",Cc,js[2]]),Ei++,Kr++}if(js[1].type==="listItemPrefix"){const Jc={type:"listItem",_spread:!1,start:Object.assign({},js[1].start),end:void 0};Cc=Jc,Sn.splice(Ei,0,["enter",Jc,js[2]]),Ei++,Kr++,Xc=void 0,rc=!0}}}return Sn[Pn][1]._spread=bs,Kr}function d(Sn,Pn){return Kr;function Kr(Ei){b.call(this,Sn(Ei),Ei),Pn&&Pn.call(this,Ei)}}function h(){this.stack.push({type:"fragment",children:[]})}function b(Sn,Pn,Kr){this.stack[this.stack.length-1].children.push(Sn),this.stack.push(Sn),this.tokenStack.push([Pn,Kr||void 0]),Sn.position={start:KD(Pn.start),end:void 0}}function y(Sn){return Pn;function Pn(Kr){Sn&&Sn.call(this,Kr),E.call(this,Kr)}}function E(Sn,Pn){const Kr=this.stack.pop(),Ei=this.tokenStack.pop();if(Ei)Ei[0].type!==Sn.type&&(Pn?Pn.call(this,Sn,Ei[0]):(Ei[1]||xWt).call(this,Sn,Ei[0]));else throw new Error("Cannot close `"+Sn.type+"` ("+yie({start:Sn.start,end:Sn.end})+"): it’s not open");Kr.position.end=KD(Sn.end)}function k(){return MQe(this.stack.pop())}function x(){this.data.expectingFirstListItemValue=!0}function O(Sn){if(this.data.expectingFirstListItemValue){const Pn=this.stack[this.stack.length-2];Pn.start=Number.parseInt(this.sliceSerialize(Sn),10),this.data.expectingFirstListItemValue=void 0}}function P(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.lang=Sn}function D(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.meta=Sn}function $(){this.data.flowCodeInside||(this.buffer(),this.data.flowCodeInside=!0)}function U(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.value=Sn.replace(/^(\r?\n|\r)|(\r?\n|\r)$/g,""),this.data.flowCodeInside=void 0}function B(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.value=Sn.replace(/(\r?\n|\r)$/g,"")}function G(Sn){const Pn=this.resume(),Kr=this.stack[this.stack.length-1];Kr.label=Pn,Kr.identifier=zx(this.sliceSerialize(Sn)).toLowerCase()}function K(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.title=Sn}function X(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.url=Sn}function ae(Sn){const Pn=this.stack[this.stack.length-1];if(!Pn.depth){const Kr=this.sliceSerialize(Sn).length;Pn.depth=Kr}}function oe(){this.data.setextHeadingSlurpLineEnding=!0}function he(Sn){const Pn=this.stack[this.stack.length-1];Pn.depth=this.sliceSerialize(Sn).codePointAt(0)===61?1:2}function ye(){this.data.setextHeadingSlurpLineEnding=void 0}function we(Sn){const Kr=this.stack[this.stack.length-1].children;let Ei=Kr[Kr.length-1];(!Ei||Ei.type!=="text")&&(Ei=ld(),Ei.position={start:KD(Sn.start),end:void 0},Kr.push(Ei)),this.stack.push(Ei)}function ge(Sn){const Pn=this.stack.pop();Pn.value+=this.sliceSerialize(Sn),Pn.position.end=KD(Sn.end)}function Te(Sn){const Pn=this.stack[this.stack.length-1];if(this.data.atHardBreak){const Kr=Pn.children[Pn.children.length-1];Kr.position.end=KD(Sn.end),this.data.atHardBreak=void 0;return}!this.data.setextHeadingSlurpLineEnding&&r.canContainEols.includes(Pn.type)&&(we.call(this,Sn),ge.call(this,Sn))}function xe(){this.data.atHardBreak=!0}function Ue(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.value=Sn}function Z(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.value=Sn}function Re(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.value=Sn}function _e(){const Sn=this.stack[this.stack.length-1];if(this.data.inReference){const Pn=this.data.referenceType||"shortcut";Sn.type+="Reference",Sn.referenceType=Pn,delete Sn.url,delete Sn.title}else delete Sn.identifier,delete Sn.label;this.data.referenceType=void 0}function qe(){const Sn=this.stack[this.stack.length-1];if(this.data.inReference){const Pn=this.data.referenceType||"shortcut";Sn.type+="Reference",Sn.referenceType=Pn,delete Sn.url,delete Sn.title}else delete Sn.identifier,delete Sn.label;this.data.referenceType=void 0}function it(Sn){const Pn=this.sliceSerialize(Sn),Kr=this.stack[this.stack.length-2];Kr.label=xKr(Pn),Kr.identifier=zx(Pn).toLowerCase()}function Xe(){const Sn=this.stack[this.stack.length-1],Pn=this.resume(),Kr=this.stack[this.stack.length-1];if(this.data.inReference=!0,Kr.type==="link"){const Ei=Sn.children;Kr.children=Ei}else Kr.alt=Pn}function de(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.url=Sn}function We(){const Sn=this.resume(),Pn=this.stack[this.stack.length-1];Pn.title=Sn}function ht(){this.data.inReference=void 0}function Ie(){this.data.referenceType="collapsed"}function $t(Sn){const Pn=this.resume(),Kr=this.stack[this.stack.length-1];Kr.label=Pn,Kr.identifier=zx(this.sliceSerialize(Sn)).toLowerCase(),this.data.referenceType="full"}function An(Sn){this.data.characterReferenceType=Sn.type}function Yt(Sn){const Pn=this.sliceSerialize(Sn),Kr=this.data.characterReferenceType;let Ei;Kr?(Ei=rfn(Pn,Kr==="characterReferenceMarkerNumeric"?10:16),this.data.characterReferenceType=void 0):Ei=jQe(Pn);const Sa=this.stack[this.stack.length-1];Sa.value+=Ei}function Or(Sn){const Pn=this.stack.pop();Pn.position.end=KD(Sn.end)}function Nt(Sn){ge.call(this,Sn);const Pn=this.stack[this.stack.length-1];Pn.url=this.sliceSerialize(Sn)}function on(Sn){ge.call(this,Sn);const Pn=this.stack[this.stack.length-1];Pn.url="mailto:"+this.sliceSerialize(Sn)}function br(){return{type:"blockquote",children:[]}}function wi(){return{type:"code",lang:null,meta:null,value:""}}function no(){return{type:"inlineCode",value:""}}function Ri(){return{type:"definition",identifier:"",label:null,title:null,url:""}}function ro(){return{type:"emphasis",children:[]}}function Ko(){return{type:"heading",depth:0,children:[]}}function Ca(){return{type:"break"}}function qi(){return{type:"html",value:""}}function yi(){return{type:"image",title:null,url:"",alt:null}}function lo(){return{type:"link",title:null,url:"",children:[]}}function Eo(Sn){return{type:"list",ordered:Sn.type==="listOrdered",start:null,spread:Sn._spread,children:[]}}function Fc(Sn){return{type:"listItem",spread:Sn._spread,checked:null,children:[]}}function Yo(){return{type:"paragraph",children:[]}}function zl(){return{type:"strong",children:[]}}function ld(){return{type:"text",value:""}}function Yc(){return{type:"thematicBreak"}}}function KD(n){return{line:n.line,column:n.column,offset:n.offset}}function pfn(n,r){let o=-1;for(;++o<r.length;){const s=r[o];Array.isArray(s)?pfn(n,s):OKr(n,s)}}function OKr(n,r){let o;for(o in r)if(hfn.call(r,o))switch(o){case"canContainEols":{const s=r[o];s&&n[o].push(...s);break}case"transforms":{const s=r[o];s&&n[o].push(...s);break}case"enter":case"exit":{const s=r[o];s&&Object.assign(n[o],s);break}}}function xWt(n,r){throw n?new Error("Cannot close `"+n.type+"` ("+yie({start:n.start,end:n.end})+"): a different token (`"+r.type+"`, "+yie({start:r.start,end:r.end})+") is open"):new Error("Cannot close document, a token (`"+r.type+"`, "+yie({start:r.start,end:r.end})+") is still open")}function NKr(n){const r=this;r.parser=o;function o(s){return AKr(s,{...r.data("settings"),...n,extensions:r.data("micromarkExtensions")||[],mdastExtensions:r.data("fromMarkdownExtensions")||[]})}}function RKr(n,r){const o={type:"element",tagName:"blockquote",properties:{},children:n.wrap(n.all(r),!0)};return n.patch(r,o),n.applyData(r,o)}function PKr(n,r){const o={type:"element",tagName:"br",properties:{},children:[]};return n.patch(r,o),[n.applyData(r,o),{type:"text",value:`
`}]}function LKr(n,r){const o=r.value?r.value+`
`:"",s={};r.lang&&(s.className=["language-"+r.lang]);let c={type:"element",tagName:"code",properties:s,children:[{type:"text",value:o}]};return r.meta&&(c.data={meta:r.meta}),n.patch(r,c),c=n.applyData(r,c),c={type:"element",tagName:"pre",properties:{},children:[c]},n.patch(r,c),c}function DKr(n,r){const o={type:"element",tagName:"del",properties:{},children:n.all(r)};return n.patch(r,o),n.applyData(r,o)}function MKr(n,r){const o={type:"element",tagName:"em",properties:{},children:n.all(r)};return n.patch(r,o),n.applyData(r,o)}function jKr(n,r){const o=typeof n.options.clobberPrefix=="string"?n.options.clobberPrefix:"user-content-",s=String(r.identifier).toUpperCase(),c=FK(s.toLowerCase()),d=n.footnoteOrder.indexOf(s);let h,b=n.footnoteCounts.get(s);b===void 0?(b=0,n.footnoteOrder.push(s),h=n.footnoteOrder.length):h=d+1,b+=1,n.footnoteCounts.set(s,b);const y={type:"element",tagName:"a",properties:{href:"#"+o+"fn-"+c,id:o+"fnref-"+c+(b>1?"-"+b:""),dataFootnoteRef:!0,ariaDescribedBy:["footnote-label"]},children:[{type:"text",value:String(h)}]};n.patch(r,y);const E={type:"element",tagName:"sup",properties:{},children:[y]};return n.patch(r,E),n.applyData(r,E)}function $Kr(n,r){const o={type:"element",tagName:"h"+r.depth,properties:{},children:n.all(r)};return n.patch(r,o),n.applyData(r,o)}function FKr(n,r){if(n.options.allowDangerousHtml){const o={type:"raw",value:r.value};return n.patch(r,o),n.applyData(r,o)}}function gfn(n,r){const o=r.referenceType;let s="]";if(o==="collapsed"?s+="[]":o==="full"&&(s+="["+(r.label||r.identifier)+"]"),r.type==="imageReference")return[{type:"text",value:"!["+r.alt+s}];const c=n.all(r),d=c[0];d&&d.type==="text"?d.value="["+d.value:c.unshift({type:"text",value:"["});const h=c[c.length-1];return h&&h.type==="text"?h.value+=s:c.push({type:"text",value:s}),c}function BKr(n,r){const o=String(r.identifier).toUpperCase(),s=n.definitionById.get(o);if(!s)return gfn(n,r);const c={src:FK(s.url||""),alt:r.alt};s.title!==null&&s.title!==void 0&&(c.title=s.title);const d={type:"element",tagName:"img",properties:c,children:[]};return n.patch(r,d),n.applyData(r,d)}function UKr(n,r){const o={src:FK(r.url)};r.alt!==null&&r.alt!==void 0&&(o.alt=r.alt),r.title!==null&&r.title!==void 0&&(o.title=r.title);const s={type:"element",tagName:"img",properties:o,children:[]};return n.patch(r,s),n.applyData(r,s)}function zKr(n,r){const o={type:"text",value:r.value.replace(/\r?\n|\r/g," ")};n.patch(r,o);const s={type:"element",tagName:"code",properties:{},children:[o]};return n.patch(r,s),n.applyData(r,s)}function HKr(n,r){const o=String(r.identifier).toUpperCase(),s=n.definitionById.get(o);if(!s)return gfn(n,r);const c={href:FK(s.url||"")};s.title!==null&&s.title!==void 0&&(c.title=s.title);const d={type:"element",tagName:"a",properties:c,children:n.all(r)};return n.patch(r,d),n.applyData(r,d)}function GKr(n,r){const o={href:FK(r.url)};r.title!==null&&r.title!==void 0&&(o.title=r.title);const s={type:"element",tagName:"a",properties:o,children:n.all(r)};return n.patch(r,s),n.applyData(r,s)}function qKr(n,r,o){const s=n.all(r),c=o?VKr(o):mfn(r),d={},h=[];if(typeof r.checked=="boolean"){const k=s[0];let x;k&&k.type==="element"&&k.tagName==="p"?x=k:(x={type:"element",tagName:"p",properties:{},children:[]},s.unshift(x)),x.children.length>0&&x.children.unshift({type:"text",value:" "}),x.children.unshift({type:"element",tagName:"input",properties:{type:"checkbox",checked:r.checked,disabled:!0},children:[]}),d.className=["task-list-item"]}let b=-1;for(;++b<s.length;){const k=s[b];(c||b!==0||k.type!=="element"||k.tagName!=="p")&&h.push({type:"text",value:`
`}),k.type==="element"&&k.tagName==="p"&&!c?h.push(...k.children):h.push(k)}const y=s[s.length-1];y&&(c||y.type!=="element"||y.tagName!=="p")&&h.push({type:"text",value:`
`});const E={type:"element",tagName:"li",properties:d,children:h};return n.patch(r,E),n.applyData(r,E)}function VKr(n){let r=!1;if(n.type==="list"){r=n.spread||!1;const o=n.children;let s=-1;for(;!r&&++s<o.length;)r=mfn(o[s])}return r}function mfn(n){const r=n.spread;return r??n.children.length>1}function WKr(n,r){const o={},s=n.all(r);let c=-1;for(typeof r.start=="number"&&r.start!==1&&(o.start=r.start);++c<s.length;){const h=s[c];if(h.type==="element"&&h.tagName==="li"&&h.properties&&Array.isArray(h.properties.className)&&h.properties.className.includes("task-list-item")){o.className=["contains-task-list"];break}}const d={type:"element",tagName:r.ordered?"ol":"ul",properties:o,children:n.wrap(s,!0)};return n.patch(r,d),n.applyData(r,d)}function KKr(n,r){const o={type:"element",tagName:"p",properties:{},children:n.all(r)};return n.patch(r,o),n.applyData(r,o)}function YKr(n,r){const o={type:"root",children:n.wrap(n.all(r))};return n.patch(r,o),n.applyData(r,o)}function XKr(n,r){const o={type:"element",tagName:"strong",properties:{},children:n.all(r)};return n.patch(r,o),n.applyData(r,o)}function JKr(n,r){const o=n.all(r),s=o.shift(),c=[];if(s){const h={type:"element",tagName:"thead",properties:{},children:n.wrap([s],!0)};n.patch(r.children[0],h),c.push(h)}if(o.length>0){const h={type:"element",tagName:"tbody",properties:{},children:n.wrap(o,!0)},b=WI(r.children[1]),y=UEe(r.children[r.children.length-1]);b&&y&&(h.position={start:b,end:y}),c.push(h)}const d={type:"element",tagName:"table",properties:{},children:n.wrap(c,!0)};return n.patch(r,d),n.applyData(r,d)}function ZKr(n,r,o){const s=o?o.children:void 0,d=(s?s.indexOf(r):1)===0?"th":"td",h=o&&o.type==="table"?o.align:void 0,b=h?h.length:r.children.length;let y=-1;const E=[];for(;++y<b;){const x=r.children[y],O={},P=h?h[y]:void 0;P&&(O.align=P);let D={type:"element",tagName:d,properties:O,children:[]};x&&(D.children=n.all(x),n.patch(x,D),D=n.applyData(x,D)),E.push(D)}const k={type:"element",tagName:"tr",properties:{},children:n.wrap(E,!0)};return n.patch(r,k),n.applyData(r,k)}function QKr(n,r){const o={type:"element",tagName:"td",properties:{},children:n.all(r)};return n.patch(r,o),n.applyData(r,o)}const CWt=9,AWt=32;function eYr(n){const r=String(n),o=/\r?\n|\r/g;let s=o.exec(r),c=0;const d=[];for(;s;)d.push(IWt(r.slice(c,s.index),c>0,!0),s[0]),c=s.index+s[0].length,s=o.exec(r);return d.push(IWt(r.slice(c),c>0,!1)),d.join("")}function IWt(n,r,o){let s=0,c=n.length;if(r){let d=n.codePointAt(s);for(;d===CWt||d===AWt;)s++,d=n.codePointAt(s)}if(o){let d=n.codePointAt(c-1);for(;d===CWt||d===AWt;)c--,d=n.codePointAt(c-1)}return c>s?n.slice(s,c):""}function tYr(n,r){const o={type:"text",value:eYr(String(r.value))};return n.patch(r,o),n.applyData(r,o)}function nYr(n,r){const o={type:"element",tagName:"hr",properties:{},children:[]};return n.patch(r,o),n.applyData(r,o)}const rYr={blockquote:RKr,break:PKr,code:LKr,delete:DKr,emphasis:MKr,footnoteReference:jKr,heading:$Kr,html:FKr,imageReference:BKr,image:UKr,inlineCode:zKr,linkReference:HKr,link:GKr,listItem:qKr,list:WKr,paragraph:KKr,root:YKr,strong:XKr,table:JKr,tableCell:QKr,tableRow:ZKr,text:tYr,thematicBreak:nYr,toml:q0e,yaml:q0e,definition:q0e,footnoteDefinition:q0e};function q0e(){}const bfn=-1,GEe=0,Sie=1,Ywe=2,FQe=3,BQe=4,UQe=5,zQe=6,vfn=7,wfn=8,OWt=typeof self=="object"?self:globalThis,iYr=(n,r)=>{const o=(c,d)=>(n.set(d,c),c),s=c=>{if(n.has(c))return n.get(c);const[d,h]=r[c];switch(d){case GEe:case bfn:return o(h,c);case Sie:{const b=o([],c);for(const y of h)b.push(s(y));return b}case Ywe:{const b=o({},c);for(const[y,E]of h)b[s(y)]=s(E);return b}case FQe:return o(new Date(h),c);case BQe:{const{source:b,flags:y}=h;return o(new RegExp(b,y),c)}case UQe:{const b=o(new Map,c);for(const[y,E]of h)b.set(s(y),s(E));return b}case zQe:{const b=o(new Set,c);for(const y of h)b.add(s(y));return b}case vfn:{const{name:b,message:y}=h;return o(new OWt[b](y),c)}case wfn:return o(BigInt(h),c);case"BigInt":return o(Object(BigInt(h)),c);case"ArrayBuffer":return o(new Uint8Array(h).buffer,h);case"DataView":{const{buffer:b}=new Uint8Array(h);return o(new DataView(b),h)}}return o(new OWt[d](h),c)};return s},NWt=n=>iYr(new Map,n)(0),ZV="",{toString:oYr}={},{keys:aYr}=Object,Cre=n=>{const r=typeof n;if(r!=="object"||!n)return[GEe,r];const o=oYr.call(n).slice(8,-1);switch(o){case"Array":return[Sie,ZV];case"Object":return[Ywe,ZV];case"Date":return[FQe,ZV];case"RegExp":return[BQe,ZV];case"Map":return[UQe,ZV];case"Set":return[zQe,ZV];case"DataView":return[Sie,o]}return o.includes("Array")?[Sie,o]:o.includes("Error")?[vfn,o]:[Ywe,o]},V0e=([n,r])=>n===GEe&&(r==="function"||r==="symbol"),sYr=(n,r,o,s)=>{const c=(h,b)=>{const y=s.push(h)-1;return o.set(b,y),y},d=h=>{if(o.has(h))return o.get(h);let[b,y]=Cre(h);switch(b){case GEe:{let k=h;switch(y){case"bigint":b=wfn,k=h.toString();break;case"function":case"symbol":if(n)throw new TypeError("unable to serialize "+y);k=null;break;case"undefined":return c([bfn],h)}return c([b,k],h)}case Sie:{if(y){let O=h;return y==="DataView"?O=new Uint8Array(h.buffer):y==="ArrayBuffer"&&(O=new Uint8Array(h)),c([y,[...O]],h)}const k=[],x=c([b,k],h);for(const O of h)k.push(d(O));return x}case Ywe:{if(y)switch(y){case"BigInt":return c([y,h.toString()],h);case"Boolean":case"Number":case"String":return c([y,h.valueOf()],h)}if(r&&"toJSON"in h)return d(h.toJSON());const k=[],x=c([b,k],h);for(const O of aYr(h))(n||!V0e(Cre(h[O])))&&k.push([d(O),d(h[O])]);return x}case FQe:return c([b,h.toISOString()],h);case BQe:{const{source:k,flags:x}=h;return c([b,{source:k,flags:x}],h)}case UQe:{const k=[],x=c([b,k],h);for(const[O,P]of h)(n||!(V0e(Cre(O))||V0e(Cre(P))))&&k.push([d(O),d(P)]);return x}case zQe:{const k=[],x=c([b,k],h);for(const O of h)(n||!V0e(Cre(O)))&&k.push(d(O));return x}}const{message:E}=h;return c([b,{name:y,message:E}],h)};return d},RWt=(n,{json:r,lossy:o}={})=>{const s=[];return sYr(!(r||o),!!r,new Map,s)(n),s},hK=typeof structuredClone=="function"?(n,r)=>r&&("json"in r||"lossy"in r)?NWt(RWt(n,r)):structuredClone(n):(n,r)=>NWt(RWt(n,r));function cYr(n,r){const o=[{type:"text",value:"↩"}];return r>1&&o.push({type:"element",tagName:"sup",properties:{},children:[{type:"text",value:String(r)}]}),o}function uYr(n,r){return"Back to reference "+(n+1)+(r>1?"-"+r:"")}function lYr(n){const r=typeof n.options.clobberPrefix=="string"?n.options.clobberPrefix:"user-content-",o=n.options.footnoteBackContent||cYr,s=n.options.footnoteBackLabel||uYr,c=n.options.footnoteLabel||"Footnotes",d=n.options.footnoteLabelTagName||"h2",h=n.options.footnoteLabelProperties||{className:["sr-only"]},b=[];let y=-1;for(;++y<n.footnoteOrder.length;){const E=n.footnoteById.get(n.footnoteOrder[y]);if(!E)continue;const k=n.all(E),x=String(E.identifier).toUpperCase(),O=FK(x.toLowerCase());let P=0;const D=[],$=n.footnoteCounts.get(x);for(;$!==void 0&&++P<=$;){D.length>0&&D.push({type:"text",value:" "});let G=typeof o=="string"?o:o(y,P);typeof G=="string"&&(G={type:"text",value:G}),D.push({type:"element",tagName:"a",properties:{href:"#"+r+"fnref-"+O+(P>1?"-"+P:""),dataFootnoteBackref:"",ariaLabel:typeof s=="string"?s:s(y,P),className:["data-footnote-backref"]},children:Array.isArray(G)?G:[G]})}const U=k[k.length-1];if(U&&U.type==="element"&&U.tagName==="p"){const G=U.children[U.children.length-1];G&&G.type==="text"?G.value+=" ":U.children.push({type:"text",value:" "}),U.children.push(...D)}else k.push(...D);const B={type:"element",tagName:"li",properties:{id:r+"fn-"+O},children:n.wrap(k,!0)};n.patch(E,B),b.push(B)}if(b.length!==0)return{type:"element",tagName:"section",properties:{dataFootnotes:!0,className:["footnotes"]},children:[{type:"element",tagName:d,properties:{...hK(h),id:"footnote-label"},children:[{type:"text",value:c}]},{type:"text",value:`
`},{type:"element",tagName:"ol",properties:{},children:n.wrap(b,!0)},{type:"text",value:`
`}]}}const qEe=function(n){if(n==null)return pYr;if(typeof n=="function")return VEe(n);if(typeof n=="object")return Array.isArray(n)?dYr(n):fYr(n);if(typeof n=="string")return hYr(n);throw new Error("Expected function, string, or object as test")};function dYr(n){const r=[];let o=-1;for(;++o<n.length;)r[o]=qEe(n[o]);return VEe(s);function s(...c){let d=-1;for(;++d<r.length;)if(r[d].apply(this,c))return!0;return!1}}function fYr(n){const r=n;return VEe(o);function o(s){const c=s;let d;for(d in n)if(c[d]!==r[d])return!1;return!0}}function hYr(n){return VEe(r);function r(o){return o&&o.type===n}}function VEe(n){return r;function r(o,s,c){return!!(gYr(o)&&n.call(this,o,typeof s=="number"?s:void 0,c||void 0))}}function pYr(){return!0}function gYr(n){return n!==null&&typeof n=="object"&&"type"in n}const yfn=[],mYr=!0,hKe=!1,bYr="skip";function Efn(n,r,o,s){let c;typeof r=="function"&&typeof o!="function"?(s=o,o=r):c=r;const d=qEe(c),h=s?-1:1;b(n,void 0,[])();function b(y,E,k){const x=y&&typeof y=="object"?y:{};if(typeof x.type=="string"){const P=typeof x.tagName=="string"?x.tagName:typeof x.name=="string"?x.name:void 0;Object.defineProperty(O,"name",{value:"node ("+(y.type+(P?"<"+P+">":""))+")"})}return O;function O(){let P=yfn,D,$,U;if((!r||d(y,E,k[k.length-1]||void 0))&&(P=vYr(o(y,k)),P[0]===hKe))return P;if("children"in y&&y.children){const B=y;if(B.children&&P[0]!==bYr)for($=(s?B.children.length:-1)+h,U=k.concat(B);$>-1&&$<B.children.length;){const G=B.children[$];if(D=b(G,$,U)(),D[0]===hKe)return D;$=typeof D[1]=="number"?D[1]:$+h}}return P}}}function vYr(n){return Array.isArray(n)?n:typeof n=="number"?[mYr,n]:n==null?yfn:[n]}function WEe(n,r,o,s){let c,d,h;typeof r=="function"&&typeof o!="function"?(d=void 0,h=r,c=o):(d=r,h=o,c=s),Efn(n,d,b,c);function b(y,E){const k=E[E.length-1],x=k?k.children.indexOf(y):void 0;return h(y,x,k)}}const pKe={}.hasOwnProperty,wYr={};function yYr(n,r){const o=r||wYr,s=new Map,c=new Map,d=new Map,h={...rYr,...o.handlers},b={all:E,applyData:SYr,definitionById:s,footnoteById:c,footnoteCounts:d,footnoteOrder:[],handlers:h,one:y,options:o,patch:EYr,wrap:kYr};return WEe(n,function(k){if(k.type==="definition"||k.type==="footnoteDefinition"){const x=k.type==="definition"?s:c,O=String(k.identifier).toUpperCase();x.has(O)||x.set(O,k)}}),b;function y(k,x){const O=k.type,P=b.handlers[O];if(pKe.call(b.handlers,O)&&P)return P(b,k,x);if(b.options.passThrough&&b.options.passThrough.includes(O)){if("children"in k){const{children:$,...U}=k,B=hK(U);return B.children=b.all(k),B}return hK(k)}return(b.options.unknownHandler||_Yr)(b,k,x)}function E(k){const x=[];if("children"in k){const O=k.children;let P=-1;for(;++P<O.length;){const D=b.one(O[P],k);if(D){if(P&&O[P-1].type==="break"&&(!Array.isArray(D)&&D.type==="text"&&(D.value=PWt(D.value)),!Array.isArray(D)&&D.type==="element")){const $=D.children[0];$&&$.type==="text"&&($.value=PWt($.value))}Array.isArray(D)?x.push(...D):x.push(D)}}}return x}}function EYr(n,r){n.position&&(r.position=uVr(n))}function SYr(n,r){let o=r;if(n&&n.data){const s=n.data.hName,c=n.data.hChildren,d=n.data.hProperties;if(typeof s=="string")if(o.type==="element")o.tagName=s;else{const h="children"in o?o.children:[o];o={type:"element",tagName:s,properties:{},children:h}}o.type==="element"&&d&&Object.assign(o.properties,hK(d)),"children"in o&&o.children&&c!==null&&c!==void 0&&(o.children=c)}return o}function _Yr(n,r){const o=r.data||{},s="value"in r&&!(pKe.call(o,"hProperties")||pKe.call(o,"hChildren"))?{type:"text",value:r.value}:{type:"element",tagName:"div",properties:{},children:n.all(r)};return n.patch(r,s),n.applyData(r,s)}function kYr(n,r){const o=[];let s=-1;for(r&&o.push({type:"text",value:`
`});++s<n.length;)s&&o.push({type:"text",value:`
`}),o.push(n[s]);return r&&n.length>0&&o.push({type:"text",value:`
`}),o}function PWt(n){let r=0,o=n.charCodeAt(r);for(;o===9||o===32;)r++,o=n.charCodeAt(r);return n.slice(r)}function LWt(n,r){const o=yYr(n,r),s=o.one(n,void 0),c=lYr(o),d=Array.isArray(s)?{type:"root",children:s}:s||{type:"root",children:[]};return c&&d.children.push({type:"text",value:`
`},c),d}function TYr(n,r){return n&&"run"in n?async function(o,s){const c=LWt(o,{file:s,...r});await n.run(c,s)}:function(o,s){return LWt(o,{file:s,...n||r})}}function DWt(n){if(n)throw n}var fqe,MWt;function xYr(){if(MWt)return fqe;MWt=1;var n=Object.prototype.hasOwnProperty,r=Object.prototype.toString,o=Object.defineProperty,s=Object.getOwnPropertyDescriptor,c=function(E){return typeof Array.isArray=="function"?Array.isArray(E):r.call(E)==="[object Array]"},d=function(E){if(!E||r.call(E)!=="[object Object]")return!1;var k=n.call(E,"constructor"),x=E.constructor&&E.constructor.prototype&&n.call(E.constructor.prototype,"isPrototypeOf");if(E.constructor&&!k&&!x)return!1;var O;for(O in E);return typeof O>"u"||n.call(E,O)},h=function(E,k){o&&k.name==="__proto__"?o(E,k.name,{enumerable:!0,configurable:!0,value:k.newValue,writable:!0}):E[k.name]=k.newValue},b=function(E,k){if(k==="__proto__")if(n.call(E,k)){if(s)return s(E,k).value}else return;return E[k]};return fqe=function y(){var E,k,x,O,P,D,$=arguments[0],U=1,B=arguments.length,G=!1;for(typeof $=="boolean"&&(G=$,$=arguments[1]||{},U=2),($==null||typeof $!="object"&&typeof $!="function")&&($={});U<B;++U)if(E=arguments[U],E!=null)for(k in E)x=b($,k),O=b(E,k),$!==O&&(G&&O&&(d(O)||(P=c(O)))?(P?(P=!1,D=x&&c(x)?x:[]):D=x&&d(x)?x:{},h($,{name:k,newValue:y(G,D,O)})):typeof O<"u"&&h($,{name:k,newValue:O}));return $},fqe}var CYr=xYr();const hqe=Y4(CYr);function gKe(n){if(typeof n!="object"||n===null)return!1;const r=Object.getPrototypeOf(n);return(r===null||r===Object.prototype||Object.getPrototypeOf(r)===null)&&!(Symbol.toStringTag in n)&&!(Symbol.iterator in n)}function AYr(){const n=[],r={run:o,use:s};return r;function o(...c){let d=-1;const h=c.pop();if(typeof h!="function")throw new TypeError("Expected function as last argument, not "+h);b(null,...c);function b(y,...E){const k=n[++d];let x=-1;if(y){h(y);return}for(;++x<c.length;)(E[x]===null||E[x]===void 0)&&(E[x]=c[x]);c=E,k?IYr(k,b)(...E):h(null,...E)}}function s(c){if(typeof c!="function")throw new TypeError("Expected `middelware` to be a function, not "+c);return n.push(c),r}}function IYr(n,r){let o;return s;function s(...h){const b=n.length>h.length;let y;b&&h.push(c);try{y=n.apply(this,h)}catch(E){const k=E;if(b&&o)throw k;return c(k)}b||(y&&y.then&&typeof y.then=="function"?y.then(d,c):y instanceof Error?c(y):d(y))}function c(h,...b){o||(o=!0,r(h,...b))}function d(h){c(null,h)}}const oI={basename:OYr,dirname:NYr,extname:RYr,join:PYr,sep:"/"};function OYr(n,r){if(r!==void 0&&typeof r!="string")throw new TypeError('"ext" argument must be a string');Zoe(n);let o=0,s=-1,c=n.length,d;if(r===void 0||r.length===0||r.length>n.length){for(;c--;)if(n.codePointAt(c)===47){if(d){o=c+1;break}}else s<0&&(d=!0,s=c+1);return s<0?"":n.slice(o,s)}if(r===n)return"";let h=-1,b=r.length-1;for(;c--;)if(n.codePointAt(c)===47){if(d){o=c+1;break}}else h<0&&(d=!0,h=c+1),b>-1&&(n.codePointAt(c)===r.codePointAt(b--)?b<0&&(s=c):(b=-1,s=h));return o===s?s=h:s<0&&(s=n.length),n.slice(o,s)}function NYr(n){if(Zoe(n),n.length===0)return".";let r=-1,o=n.length,s;for(;--o;)if(n.codePointAt(o)===47){if(s){r=o;break}}else s||(s=!0);return r<0?n.codePointAt(0)===47?"/":".":r===1&&n.codePointAt(0)===47?"//":n.slice(0,r)}function RYr(n){Zoe(n);let r=n.length,o=-1,s=0,c=-1,d=0,h;for(;r--;){const b=n.codePointAt(r);if(b===47){if(h){s=r+1;break}continue}o<0&&(h=!0,o=r+1),b===46?c<0?c=r:d!==1&&(d=1):c>-1&&(d=-1)}return c<0||o<0||d===0||d===1&&c===o-1&&c===s+1?"":n.slice(c,o)}function PYr(...n){let r=-1,o;for(;++r<n.length;)Zoe(n[r]),n[r]&&(o=o===void 0?n[r]:o+"/"+n[r]);return o===void 0?".":LYr(o)}function LYr(n){Zoe(n);const r=n.codePointAt(0)===47;let o=DYr(n,!r);return o.length===0&&!r&&(o="."),o.length>0&&n.codePointAt(n.length-1)===47&&(o+="/"),r?"/"+o:o}function DYr(n,r){let o="",s=0,c=-1,d=0,h=-1,b,y;for(;++h<=n.length;){if(h<n.length)b=n.codePointAt(h);else{if(b===47)break;b=47}if(b===47){if(!(c===h-1||d===1))if(c!==h-1&&d===2){if(o.length<2||s!==2||o.codePointAt(o.length-1)!==46||o.codePointAt(o.length-2)!==46){if(o.length>2){if(y=o.lastIndexOf("/"),y!==o.length-1){y<0?(o="",s=0):(o=o.slice(0,y),s=o.length-1-o.lastIndexOf("/")),c=h,d=0;continue}}else if(o.length>0){o="",s=0,c=h,d=0;continue}}r&&(o=o.length>0?o+"/..":"..",s=2)}else o.length>0?o+="/"+n.slice(c+1,h):o=n.slice(c+1,h),s=h-c-1;c=h,d=0}else b===46&&d>-1?d++:d=-1}return o}function Zoe(n){if(typeof n!="string")throw new TypeError("Path must be a string. Received "+JSON.stringify(n))}const MYr={cwd:jYr};function jYr(){return"/"}function mKe(n){return!!(n!==null&&typeof n=="object"&&"href"in n&&n.href&&"protocol"in n&&n.protocol&&n.auth===void 0)}function $Yr(n){if(typeof n=="string")n=new URL(n);else if(!mKe(n)){const r=new TypeError('The "path" argument must be of type string or an instance of URL. Received `'+n+"`");throw r.code="ERR_INVALID_ARG_TYPE",r}if(n.protocol!=="file:"){const r=new TypeError("The URL must be of scheme file");throw r.code="ERR_INVALID_URL_SCHEME",r}return FYr(n)}function FYr(n){if(n.hostname!==""){const s=new TypeError('File URL host must be "localhost" or empty on darwin');throw s.code="ERR_INVALID_FILE_URL_HOST",s}const r=n.pathname;let o=-1;for(;++o<r.length;)if(r.codePointAt(o)===37&&r.codePointAt(o+1)===50){const s=r.codePointAt(o+2);if(s===70||s===102){const c=new TypeError("File URL path must not include encoded / characters");throw c.code="ERR_INVALID_FILE_URL_PATH",c}}return decodeURIComponent(r)}const pqe=["history","path","basename","stem","extname","dirname"];class Sfn{constructor(r){let o;r?mKe(r)?o={path:r}:typeof r=="string"||BYr(r)?o={value:r}:o=r:o={},this.cwd="cwd"in o?"":MYr.cwd(),this.data={},this.history=[],this.messages=[],this.value,this.map,this.result,this.stored;let s=-1;for(;++s<pqe.length;){const d=pqe[s];d in o&&o[d]!==void 0&&o[d]!==null&&(this[d]=d==="history"?[...o[d]]:o[d])}let c;for(c in o)pqe.includes(c)||(this[c]=o[c])}get basename(){return typeof this.path=="string"?oI.basename(this.path):void 0}set basename(r){mqe(r,"basename"),gqe(r,"basename"),this.path=oI.join(this.dirname||"",r)}get dirname(){return typeof this.path=="string"?oI.dirname(this.path):void 0}set dirname(r){jWt(this.basename,"dirname"),this.path=oI.join(r||"",this.basename)}get extname(){return typeof this.path=="string"?oI.extname(this.path):void 0}set extname(r){if(gqe(r,"extname"),jWt(this.dirname,"extname"),r){if(r.codePointAt(0)!==46)throw new Error("`extname` must start with `.`");if(r.includes(".",1))throw new Error("`extname` cannot contain multiple dots")}this.path=oI.join(this.dirname,this.stem+(r||""))}get path(){return this.history[this.history.length-1]}set path(r){mKe(r)&&(r=$Yr(r)),mqe(r,"path"),this.path!==r&&this.history.push(r)}get stem(){return typeof this.path=="string"?oI.basename(this.path,this.extname):void 0}set stem(r){mqe(r,"stem"),gqe(r,"stem"),this.path=oI.join(this.dirname||"",r+(this.extname||""))}fail(r,o,s){const c=this.message(r,o,s);throw c.fatal=!0,c}info(r,o,s){const c=this.message(r,o,s);return c.fatal=void 0,c}message(r,o,s){const c=new Z1(r,o,s);return this.path&&(c.name=this.path+":"+c.name,c.file=this.path),c.fatal=!1,this.messages.push(c),c}toString(r){return this.value===void 0?"":typeof this.value=="string"?this.value:new TextDecoder(r||void 0).decode(this.value)}}function gqe(n,r){if(n&&n.includes(oI.sep))throw new Error("`"+r+"` cannot be a path: did not expect `"+oI.sep+"`")}function mqe(n,r){if(!n)throw new Error("`"+r+"` cannot be empty")}function jWt(n,r){if(!n)throw new Error("Setting `"+r+"` requires `path` to be set too")}function BYr(n){return!!(n&&typeof n=="object"&&"byteLength"in n&&"byteOffset"in n)}const UYr=function(n){const s=this.constructor.prototype,c=s[n],d=function(){return c.apply(d,arguments)};return Object.setPrototypeOf(d,s),d},zYr={}.hasOwnProperty;class HQe extends UYr{constructor(){super("copy"),this.Compiler=void 0,this.Parser=void 0,this.attachers=[],this.compiler=void 0,this.freezeIndex=-1,this.frozen=void 0,this.namespace={},this.parser=void 0,this.transformers=AYr()}copy(){const r=new HQe;let o=-1;for(;++o<this.attachers.length;){const s=this.attachers[o];r.use(...s)}return r.data(hqe(!0,{},this.namespace)),r}data(r,o){return typeof r=="string"?arguments.length===2?(wqe("data",this.frozen),this.namespace[r]=o,this):zYr.call(this.namespace,r)&&this.namespace[r]||void 0:r?(wqe("data",this.frozen),this.namespace=r,this):this.namespace}freeze(){if(this.frozen)return this;const r=this;for(;++this.freezeIndex<this.attachers.length;){const[o,...s]=this.attachers[this.freezeIndex];if(s[0]===!1)continue;s[0]===!0&&(s[0]=void 0);const c=o.call(r,...s);typeof c=="function"&&this.transformers.use(c)}return this.frozen=!0,this.freezeIndex=Number.POSITIVE_INFINITY,this}parse(r){this.freeze();const o=W0e(r),s=this.parser||this.Parser;return bqe("parse",s),s(String(o),o)}process(r,o){const s=this;return this.freeze(),bqe("process",this.parser||this.Parser),vqe("process",this.compiler||this.Compiler),o?c(void 0,o):new Promise(c);function c(d,h){const b=W0e(r),y=s.parse(b);s.run(y,b,function(k,x,O){if(k||!x||!O)return E(k);const P=x,D=s.stringify(P,O);qYr(D)?O.value=D:O.result=D,E(k,O)});function E(k,x){k||!x?h(k):d?d(x):o(void 0,x)}}}processSync(r){let o=!1,s;return this.freeze(),bqe("processSync",this.parser||this.Parser),vqe("processSync",this.compiler||this.Compiler),this.process(r,c),FWt("processSync","process",o),s;function c(d,h){o=!0,DWt(d),s=h}}run(r,o,s){$Wt(r),this.freeze();const c=this.transformers;return!s&&typeof o=="function"&&(s=o,o=void 0),s?d(void 0,s):new Promise(d);function d(h,b){const y=W0e(o);c.run(r,y,E);function E(k,x,O){const P=x||r;k?b(k):h?h(P):s(void 0,P,O)}}}runSync(r,o){let s=!1,c;return this.run(r,o,d),FWt("runSync","run",s),c;function d(h,b){DWt(h),c=b,s=!0}}stringify(r,o){this.freeze();const s=W0e(o),c=this.compiler||this.Compiler;return vqe("stringify",c),$Wt(r),c(r,s)}use(r,...o){const s=this.attachers,c=this.namespace;if(wqe("use",this.frozen),r!=null)if(typeof r=="function")y(r,o);else if(typeof r=="object")Array.isArray(r)?b(r):h(r);else throw new TypeError("Expected usable value, not `"+r+"`");return this;function d(E){if(typeof E=="function")y(E,[]);else if(typeof E=="object")if(Array.isArray(E)){const[k,...x]=E;y(k,x)}else h(E);else throw new TypeError("Expected usable value, not `"+E+"`")}function h(E){if(!("plugins"in E)&&!("settings"in E))throw new Error("Expected usable value but received an empty preset, which is probably a mistake: presets typically come with `plugins` and sometimes with `settings`, but this has neither");b(E.plugins),E.settings&&(c.settings=hqe(!0,c.settings,E.settings))}function b(E){let k=-1;if(E!=null)if(Array.isArray(E))for(;++k<E.length;){const x=E[k];d(x)}else throw new TypeError("Expected a list of plugins, not `"+E+"`")}function y(E,k){let x=-1,O=-1;for(;++x<s.length;)if(s[x][0]===E){O=x;break}if(O===-1)s.push([E,...k]);else if(k.length>0){let[P,...D]=k;const $=s[O][1];gKe($)&&gKe(P)&&(P=hqe(!0,$,P)),s[O]=[E,P,...D]}}}}const HYr=new HQe().freeze();function bqe(n,r){if(typeof r!="function")throw new TypeError("Cannot `"+n+"` without `parser`")}function vqe(n,r){if(typeof r!="function")throw new TypeError("Cannot `"+n+"` without `compiler`")}function wqe(n,r){if(r)throw new Error("Cannot call `"+n+"` on a frozen processor.\nCreate a new processor first, by calling it: use `processor()` instead of `processor`.")}function $Wt(n){if(!gKe(n)||typeof n.type!="string")throw new TypeError("Expected node, got `"+n+"`")}function FWt(n,r,o){if(!o)throw new Error("`"+n+"` finished async. Use `"+r+"` instead")}function W0e(n){return GYr(n)?n:new Sfn(n)}function GYr(n){return!!(n&&typeof n=="object"&&"message"in n&&"messages"in n)}function qYr(n){return typeof n=="string"||VYr(n)}function VYr(n){return!!(n&&typeof n=="object"&&"byteLength"in n&&"byteOffset"in n)}const WYr="https://github.com/remarkjs/react-markdown/blob/main/changelog.md",BWt=[],UWt={allowDangerousHtml:!0},KYr=/^(https?|ircs?|mailto|xmpp)$/i,YYr=[{from:"astPlugins",id:"remove-buggy-html-in-markdown-parser"},{from:"allowDangerousHtml",id:"remove-buggy-html-in-markdown-parser"},{from:"allowNode",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"allowElement"},{from:"allowedTypes",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"allowedElements"},{from:"className",id:"remove-classname"},{from:"disallowedTypes",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"disallowedElements"},{from:"escapeHtml",id:"remove-buggy-html-in-markdown-parser"},{from:"includeElementIndex",id:"#remove-includeelementindex"},{from:"includeNodeIndex",id:"change-includenodeindex-to-includeelementindex"},{from:"linkTarget",id:"remove-linktarget"},{from:"plugins",id:"change-plugins-to-remarkplugins",to:"remarkPlugins"},{from:"rawSourcePos",id:"#remove-rawsourcepos"},{from:"renderers",id:"change-renderers-to-components",to:"components"},{from:"source",id:"change-source-to-children",to:"children"},{from:"sourcePos",id:"#remove-sourcepos"},{from:"transformImageUri",id:"#add-urltransform",to:"urlTransform"},{from:"transformLinkUri",id:"#add-urltransform",to:"urlTransform"}];function XYr(n){const r=JYr(n),o=ZYr(n);return QYr(r.runSync(r.parse(o),o),n)}function JYr(n){const r=n.rehypePlugins||BWt,o=n.remarkPlugins||BWt,s=n.remarkRehypeOptions?{...n.remarkRehypeOptions,...UWt}:UWt;return HYr().use(NKr).use(o).use(TYr,s).use(r)}function ZYr(n){const r=n.children||"",o=new Sfn;return typeof r=="string"&&(o.value=r),o}function QYr(n,r){const o=r.allowedElements,s=r.allowElement,c=r.components,d=r.disallowedElements,h=r.skipHtml,b=r.unwrapDisallowed,y=r.urlTransform||eXr;for(const k of YYr)Object.hasOwn(r,k.from)&&(""+k.from+(k.to?"use `"+k.to+"` instead":"remove it")+WYr+k.id,void 0);return WEe(n,E),pVr(n,{Fragment:H.Fragment,components:c,ignoreInvalidStyle:!0,jsx:H.jsx,jsxs:H.jsxs,passKeys:!0,passNode:!0});function E(k,x,O){if(k.type==="raw"&&O&&typeof x=="number")return h?O.children.splice(x,1):O.children[x]={type:"text",value:k.value},x;if(k.type==="element"){let P;for(P in uqe)if(Object.hasOwn(uqe,P)&&Object.hasOwn(k.properties,P)){const D=k.properties[P],$=uqe[P];($===null||$.includes(k.tagName))&&(k.properties[P]=y(String(D||""),P,k))}}if(k.type==="element"){let P=o?!o.includes(k.tagName):d?d.includes(k.tagName):!1;if(!P&&s&&typeof x=="number"&&(P=!s(k,x,O)),P&&O&&typeof x=="number")return b&&k.children?O.children.splice(x,1,...k.children):O.children.splice(x,1),x}}}function eXr(n){const r=n.indexOf(":"),o=n.indexOf("?"),s=n.indexOf("#"),c=n.indexOf("/");return r===-1||c!==-1&&r>c||o!==-1&&r>o||s!==-1&&r>s||KYr.test(n.slice(0,r))?n:""}const zWt=/[#.]/g;function tXr(n,r){const o=n||"",s={};let c=0,d,h;for(;c<o.length;){zWt.lastIndex=c;const b=zWt.exec(o),y=o.slice(c,b?b.index:o.length);y&&(d?d==="#"?s.id=y:Array.isArray(s.className)?s.className.push(y):s.className=[y]:h=y,c+=y.length),b&&(d=b[0],c++)}return{type:"element",tagName:h||r||"div",properties:s,children:[]}}function _fn(n,r,o){const s=o?oXr(o):void 0;function c(d,h,...b){let y;if(d==null){y={type:"root",children:[]};const E=h;b.unshift(E)}else{y=tXr(d,r);const E=y.tagName.toLowerCase(),k=s?s.get(E):void 0;if(y.tagName=k||E,nXr(h))b.unshift(h);else for(const[x,O]of Object.entries(h))rXr(n,y.properties,x,O)}for(const E of b)bKe(y.children,E);return y.type==="element"&&y.tagName==="template"&&(y.content={type:"root",children:y.children},y.children=[]),y}return c}function nXr(n){if(n===null||typeof n!="object"||Array.isArray(n))return!0;if(typeof n.type!="string")return!1;const r=n,o=Object.keys(n);for(const s of o){const c=r[s];if(c&&typeof c=="object"){if(!Array.isArray(c))return!0;const d=c;for(const h of d)if(typeof h!="number"&&typeof h!="string")return!0}}return!!("children"in n&&Array.isArray(n.children))}function rXr(n,r,o,s){const c=RQe(n,o);let d;if(s!=null){if(typeof s=="number"){if(Number.isNaN(s))return;d=s}else typeof s=="boolean"?d=s:typeof s=="string"?c.spaceSeparated?d=uWt(s):c.commaSeparated?d=iWt(s):c.commaOrSpaceSeparated?d=uWt(iWt(s).join(" ")):d=HWt(c,c.property,s):Array.isArray(s)?d=[...s]:d=c.property==="style"?iXr(s):String(s);if(Array.isArray(d)){const h=[];for(const b of d)h.push(HWt(c,c.property,b));d=h}c.property==="className"&&Array.isArray(r.className)&&(d=r.className.concat(d)),r[c.property]=d}}function bKe(n,r){if(r!=null)if(typeof r=="number"||typeof r=="string")n.push({type:"text",value:String(r)});else if(Array.isArray(r))for(const o of r)bKe(n,o);else if(typeof r=="object"&&"type"in r)r.type==="root"?bKe(n,r.children):n.push(r);else throw new Error("Expected node, nodes, or string, got `"+r+"`")}function HWt(n,r,o){if(typeof o=="string"){if(n.number&&o&&!Number.isNaN(Number(o)))return Number(o);if((n.boolean||n.overloadedBoolean)&&(o===""||poe(o)===poe(r)))return!0}return o}function iXr(n){const r=[];for(const[o,s]of Object.entries(n))r.push([o,s].join(": "));return r.join("; ")}function oXr(n){const r=new Map;for(const o of n)r.set(o.toLowerCase(),o);return r}const aXr=["altGlyph","altGlyphDef","altGlyphItem","animateColor","animateMotion","animateTransform","clipPath","feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feDistantLight","feDropShadow","feFlood","feFuncA","feFuncB","feFuncG","feFuncR","feGaussianBlur","feImage","feMerge","feMergeNode","feMorphology","feOffset","fePointLight","feSpecularLighting","feSpotLight","feTile","feTurbulence","foreignObject","glyphRef","linearGradient","radialGradient","solidColor","textArea","textPath"],sXr=_fn(BEe,"div"),cXr=_fn($K,"g",aXr);function uXr(n){const r=String(n),o=[];return{toOffset:c,toPoint:s};function s(d){if(typeof d=="number"&&d>-1&&d<=r.length){let h=0;for(;;){let b=o[h];if(b===void 0){const y=GWt(r,o[h-1]);b=y===-1?r.length+1:y+1,o[h]=b}if(b>d)return{line:h+1,column:d-(h>0?o[h-1]:0)+1,offset:d};h++}}}function c(d){if(d&&typeof d.line=="number"&&typeof d.column=="number"&&!Number.isNaN(d.line)&&!Number.isNaN(d.column)){for(;o.length<d.line;){const b=o[o.length-1],y=GWt(r,b),E=y===-1?r.length+1:y+1;if(b===E)break;o.push(E)}const h=(d.line>1?o[d.line-2]:0)+d.column-1;if(h<o[d.line-1])return h}}}function GWt(n,r){const o=n.indexOf("\r",r),s=n.indexOf(`
`,r);return s===-1?o:o===-1||o+1===s?s:o<s?o:s}const jB={html:"http://www.w3.org/1999/xhtml",mathml:"http://www.w3.org/1998/Math/MathML",svg:"http://www.w3.org/2000/svg",xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"},kfn={}.hasOwnProperty,lXr=Object.prototype;function dXr(n,r){const o=r||{};return GQe({file:o.file||void 0,location:!1,schema:o.space==="svg"?$K:BEe,verbose:o.verbose||!1},n)}function GQe(n,r){let o;switch(r.nodeName){case"#comment":{const s=r;return o={type:"comment",value:s.data},Lve(n,s,o),o}case"#document":case"#document-fragment":{const s=r,c="mode"in s?s.mode==="quirks"||s.mode==="limited-quirks":!1;if(o={type:"root",children:Tfn(n,r.childNodes),data:{quirksMode:c}},n.file&&n.location){const d=String(n.file),h=uXr(d),b=h.toPoint(0),y=h.toPoint(d.length);o.position={start:b,end:y}}return o}case"#documentType":{const s=r;return o={type:"doctype"},Lve(n,s,o),o}case"#text":{const s=r;return o={type:"text",value:s.value},Lve(n,s,o),o}default:return o=fXr(n,r),o}}function Tfn(n,r){let o=-1;const s=[];for(;++o<r.length;){const c=GQe(n,r[o]);s.push(c)}return s}function fXr(n,r){const o=n.schema;n.schema=r.namespaceURI===jB.svg?$K:BEe;let s=-1;const c={};for(;++s<r.attrs.length;){const b=r.attrs[s],y=(b.prefix?b.prefix+":":"")+b.name;kfn.call(lXr,y)||(c[y]=b.value)}const h=(n.schema.space==="svg"?cXr:sXr)(r.tagName,c,Tfn(n,r.childNodes));if(Lve(n,r,h),h.tagName==="template"){const b=r,y=b.sourceCodeLocation,E=y&&y.startTag&&vW(y.startTag),k=y&&y.endTag&&vW(y.endTag),x=GQe(n,b.content);E&&k&&n.file&&(x.position={start:E.end,end:k.start}),h.content=x}return n.schema=o,h}function Lve(n,r,o){if("sourceCodeLocation"in r&&r.sourceCodeLocation&&n.file){const s=hXr(n,o,r.sourceCodeLocation);s&&(n.location=!0,o.position=s)}}function hXr(n,r,o){const s=vW(o);if(r.type==="element"){const c=r.children[r.children.length-1];if(s&&!o.endTag&&c&&c.position&&c.position.end&&(s.end=Object.assign({},c.position.end)),n.verbose){const d={};let h;if(o.attrs)for(h in o.attrs)kfn.call(o.attrs,h)&&(d[RQe(n.schema,h).property]=vW(o.attrs[h]));o.startTag;const b=vW(o.startTag),y=o.endTag?vW(o.endTag):void 0,E={opening:b};y&&(E.closing=y),E.properties=d,r.data={position:E}}}return s}function vW(n){const r=qWt({line:n.startLine,column:n.startCol,offset:n.startOffset}),o=qWt({line:n.endLine,column:n.endCol,offset:n.endOffset});return r||o?{start:r,end:o}:void 0}function qWt(n){return n.line&&n.column?n:void 0}class Qoe{constructor(r,o,s){this.property=r,this.normal=o,s&&(this.space=s)}}Qoe.prototype.property={};Qoe.prototype.normal={};Qoe.prototype.space=null;function xfn(n,r){const o={},s={};let c=-1;for(;++c<n.length;)Object.assign(o,n[c].property),Object.assign(s,n[c].normal);return new Qoe(o,s,r)}function vKe(n){return n.toLowerCase()}class W_{constructor(r,o){this.property=r,this.attribute=o}}W_.prototype.space=null;W_.prototype.boolean=!1;W_.prototype.booleanish=!1;W_.prototype.overloadedBoolean=!1;W_.prototype.number=!1;W_.prototype.commaSeparated=!1;W_.prototype.spaceSeparated=!1;W_.prototype.commaOrSpaceSeparated=!1;W_.prototype.mustUseProperty=!1;W_.prototype.defined=!1;let pXr=0;const ms=yU(),Gp=yU(),Cfn=yU(),Wr=yU(),xd=yU(),UW=yU(),ZE=yU();function yU(){return 2**++pXr}const wKe=Object.freeze(Object.defineProperty({__proto__:null,boolean:ms,booleanish:Gp,commaOrSpaceSeparated:ZE,commaSeparated:UW,number:Wr,overloadedBoolean:Cfn,spaceSeparated:xd},Symbol.toStringTag,{value:"Module"})),yqe=Object.keys(wKe);class qQe extends W_{constructor(r,o,s,c){let d=-1;if(super(r,o),VWt(this,"space",c),typeof s=="number")for(;++d<yqe.length;){const h=yqe[d];VWt(this,yqe[d],(s&wKe[h])===wKe[h])}}}qQe.prototype.defined=!0;function VWt(n,r,o){o&&(n[r]=o)}const gXr={}.hasOwnProperty;function BK(n){const r={},o={};let s;for(s in n.properties)if(gXr.call(n.properties,s)){const c=n.properties[s],d=new qQe(s,n.transform(n.attributes||{},s),c,n.space);n.mustUseProperty&&n.mustUseProperty.includes(s)&&(d.mustUseProperty=!0),r[s]=d,o[vKe(s)]=s,o[vKe(d.attribute)]=s}return new Qoe(r,o,n.space)}const Afn=BK({space:"xlink",transform(n,r){return"xlink:"+r.slice(5).toLowerCase()},properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null}}),Ifn=BK({space:"xml",transform(n,r){return"xml:"+r.slice(3).toLowerCase()},properties:{xmlLang:null,xmlBase:null,xmlSpace:null}});function Ofn(n,r){return r in n?n[r]:r}function Nfn(n,r){return Ofn(n,r.toLowerCase())}const Rfn=BK({space:"xmlns",attributes:{xmlnsxlink:"xmlns:xlink"},transform:Nfn,properties:{xmlns:null,xmlnsXLink:null}}),Pfn=BK({transform(n,r){return r==="role"?r:"aria-"+r.slice(4).toLowerCase()},properties:{ariaActiveDescendant:null,ariaAtomic:Gp,ariaAutoComplete:null,ariaBusy:Gp,ariaChecked:Gp,ariaColCount:Wr,ariaColIndex:Wr,ariaColSpan:Wr,ariaControls:xd,ariaCurrent:null,ariaDescribedBy:xd,ariaDetails:null,ariaDisabled:Gp,ariaDropEffect:xd,ariaErrorMessage:null,ariaExpanded:Gp,ariaFlowTo:xd,ariaGrabbed:Gp,ariaHasPopup:null,ariaHidden:Gp,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:xd,ariaLevel:Wr,ariaLive:null,ariaModal:Gp,ariaMultiLine:Gp,ariaMultiSelectable:Gp,ariaOrientation:null,ariaOwns:xd,ariaPlaceholder:null,ariaPosInSet:Wr,ariaPressed:Gp,ariaReadOnly:Gp,ariaRelevant:null,ariaRequired:Gp,ariaRoleDescription:xd,ariaRowCount:Wr,ariaRowIndex:Wr,ariaRowSpan:Wr,ariaSelected:Gp,ariaSetSize:Wr,ariaSort:null,ariaValueMax:Wr,ariaValueMin:Wr,ariaValueNow:Wr,ariaValueText:null,role:null}}),mXr=BK({space:"html",attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},transform:Nfn,mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:UW,acceptCharset:xd,accessKey:xd,action:null,allow:null,allowFullScreen:ms,allowPaymentRequest:ms,allowUserMedia:ms,alt:null,as:null,async:ms,autoCapitalize:null,autoComplete:xd,autoFocus:ms,autoPlay:ms,blocking:xd,capture:null,charSet:null,checked:ms,cite:null,className:xd,cols:Wr,colSpan:null,content:null,contentEditable:Gp,controls:ms,controlsList:xd,coords:Wr|UW,crossOrigin:null,data:null,dateTime:null,decoding:null,default:ms,defer:ms,dir:null,dirName:null,disabled:ms,download:Cfn,draggable:Gp,encType:null,enterKeyHint:null,fetchPriority:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:ms,formTarget:null,headers:xd,height:Wr,hidden:ms,high:Wr,href:null,hrefLang:null,htmlFor:xd,httpEquiv:xd,id:null,imageSizes:null,imageSrcSet:null,inert:ms,inputMode:null,integrity:null,is:null,isMap:ms,itemId:null,itemProp:xd,itemRef:xd,itemScope:ms,itemType:xd,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:ms,low:Wr,manifest:null,max:null,maxLength:Wr,media:null,method:null,min:null,minLength:Wr,multiple:ms,muted:ms,name:null,nonce:null,noModule:ms,noValidate:ms,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforeMatch:null,onBeforePrint:null,onBeforeToggle:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextLost:null,onContextMenu:null,onContextRestored:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onScrollEnd:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:ms,optimum:Wr,pattern:null,ping:xd,placeholder:null,playsInline:ms,popover:null,popoverTarget:null,popoverTargetAction:null,poster:null,preload:null,readOnly:ms,referrerPolicy:null,rel:xd,required:ms,reversed:ms,rows:Wr,rowSpan:Wr,sandbox:xd,scope:null,scoped:ms,seamless:ms,selected:ms,shadowRootClonable:ms,shadowRootDelegatesFocus:ms,shadowRootMode:null,shape:null,size:Wr,sizes:null,slot:null,span:Wr,spellCheck:Gp,src:null,srcDoc:null,srcLang:null,srcSet:null,start:Wr,step:null,style:null,tabIndex:Wr,target:null,title:null,translate:null,type:null,typeMustMatch:ms,useMap:null,value:Gp,width:Wr,wrap:null,writingSuggestions:null,align:null,aLink:null,archive:xd,axis:null,background:null,bgColor:null,border:Wr,borderColor:null,bottomMargin:Wr,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:ms,declare:ms,event:null,face:null,frame:null,frameBorder:null,hSpace:Wr,leftMargin:Wr,link:null,longDesc:null,lowSrc:null,marginHeight:Wr,marginWidth:Wr,noResize:ms,noHref:ms,noShade:ms,noWrap:ms,object:null,profile:null,prompt:null,rev:null,rightMargin:Wr,rules:null,scheme:null,scrolling:Gp,standby:null,summary:null,text:null,topMargin:Wr,valueType:null,version:null,vAlign:null,vLink:null,vSpace:Wr,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:ms,disableRemotePlayback:ms,prefix:null,property:null,results:Wr,security:null,unselectable:null}}),bXr=BK({space:"svg",attributes:{accentHeight:"accent-height",alignmentBaseline:"alignment-baseline",arabicForm:"arabic-form",baselineShift:"baseline-shift",capHeight:"cap-height",className:"class",clipPath:"clip-path",clipRule:"clip-rule",colorInterpolation:"color-interpolation",colorInterpolationFilters:"color-interpolation-filters",colorProfile:"color-profile",colorRendering:"color-rendering",crossOrigin:"crossorigin",dataType:"datatype",dominantBaseline:"dominant-baseline",enableBackground:"enable-background",fillOpacity:"fill-opacity",fillRule:"fill-rule",floodColor:"flood-color",floodOpacity:"flood-opacity",fontFamily:"font-family",fontSize:"font-size",fontSizeAdjust:"font-size-adjust",fontStretch:"font-stretch",fontStyle:"font-style",fontVariant:"font-variant",fontWeight:"font-weight",glyphName:"glyph-name",glyphOrientationHorizontal:"glyph-orientation-horizontal",glyphOrientationVertical:"glyph-orientation-vertical",hrefLang:"hreflang",horizAdvX:"horiz-adv-x",horizOriginX:"horiz-origin-x",horizOriginY:"horiz-origin-y",imageRendering:"image-rendering",letterSpacing:"letter-spacing",lightingColor:"lighting-color",markerEnd:"marker-end",markerMid:"marker-mid",markerStart:"marker-start",navDown:"nav-down",navDownLeft:"nav-down-left",navDownRight:"nav-down-right",navLeft:"nav-left",navNext:"nav-next",navPrev:"nav-prev",navRight:"nav-right",navUp:"nav-up",navUpLeft:"nav-up-left",navUpRight:"nav-up-right",onAbort:"onabort",onActivate:"onactivate",onAfterPrint:"onafterprint",onBeforePrint:"onbeforeprint",onBegin:"onbegin",onCancel:"oncancel",onCanPlay:"oncanplay",onCanPlayThrough:"oncanplaythrough",onChange:"onchange",onClick:"onclick",onClose:"onclose",onCopy:"oncopy",onCueChange:"oncuechange",onCut:"oncut",onDblClick:"ondblclick",onDrag:"ondrag",onDragEnd:"ondragend",onDragEnter:"ondragenter",onDragExit:"ondragexit",onDragLeave:"ondragleave",onDragOver:"ondragover",onDragStart:"ondragstart",onDrop:"ondrop",onDurationChange:"ondurationchange",onEmptied:"onemptied",onEnd:"onend",onEnded:"onended",onError:"onerror",onFocus:"onfocus",onFocusIn:"onfocusin",onFocusOut:"onfocusout",onHashChange:"onhashchange",onInput:"oninput",onInvalid:"oninvalid",onKeyDown:"onkeydown",onKeyPress:"onkeypress",onKeyUp:"onkeyup",onLoad:"onload",onLoadedData:"onloadeddata",onLoadedMetadata:"onloadedmetadata",onLoadStart:"onloadstart",onMessage:"onmessage",onMouseDown:"onmousedown",onMouseEnter:"onmouseenter",onMouseLeave:"onmouseleave",onMouseMove:"onmousemove",onMouseOut:"onmouseout",onMouseOver:"onmouseover",onMouseUp:"onmouseup",onMouseWheel:"onmousewheel",onOffline:"onoffline",onOnline:"ononline",onPageHide:"onpagehide",onPageShow:"onpageshow",onPaste:"onpaste",onPause:"onpause",onPlay:"onplay",onPlaying:"onplaying",onPopState:"onpopstate",onProgress:"onprogress",onRateChange:"onratechange",onRepeat:"onrepeat",onReset:"onreset",onResize:"onresize",onScroll:"onscroll",onSeeked:"onseeked",onSeeking:"onseeking",onSelect:"onselect",onShow:"onshow",onStalled:"onstalled",onStorage:"onstorage",onSubmit:"onsubmit",onSuspend:"onsuspend",onTimeUpdate:"ontimeupdate",onToggle:"ontoggle",onUnload:"onunload",onVolumeChange:"onvolumechange",onWaiting:"onwaiting",onZoom:"onzoom",overlinePosition:"overline-position",overlineThickness:"overline-thickness",paintOrder:"paint-order",panose1:"panose-1",pointerEvents:"pointer-events",referrerPolicy:"referrerpolicy",renderingIntent:"rendering-intent",shapeRendering:"shape-rendering",stopColor:"stop-color",stopOpacity:"stop-opacity",strikethroughPosition:"strikethrough-position",strikethroughThickness:"strikethrough-thickness",strokeDashArray:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeLineCap:"stroke-linecap",strokeLineJoin:"stroke-linejoin",strokeMiterLimit:"stroke-miterlimit",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",tabIndex:"tabindex",textAnchor:"text-anchor",textDecoration:"text-decoration",textRendering:"text-rendering",transformOrigin:"transform-origin",typeOf:"typeof",underlinePosition:"underline-position",underlineThickness:"underline-thickness",unicodeBidi:"unicode-bidi",unicodeRange:"unicode-range",unitsPerEm:"units-per-em",vAlphabetic:"v-alphabetic",vHanging:"v-hanging",vIdeographic:"v-ideographic",vMathematical:"v-mathematical",vectorEffect:"vector-effect",vertAdvY:"vert-adv-y",vertOriginX:"vert-origin-x",vertOriginY:"vert-origin-y",wordSpacing:"word-spacing",writingMode:"writing-mode",xHeight:"x-height",playbackOrder:"playbackorder",timelineBegin:"timelinebegin"},transform:Ofn,properties:{about:ZE,accentHeight:Wr,accumulate:null,additive:null,alignmentBaseline:null,alphabetic:Wr,amplitude:Wr,arabicForm:null,ascent:Wr,attributeName:null,attributeType:null,azimuth:Wr,bandwidth:null,baselineShift:null,baseFrequency:null,baseProfile:null,bbox:null,begin:null,bias:Wr,by:null,calcMode:null,capHeight:Wr,className:xd,clip:null,clipPath:null,clipPathUnits:null,clipRule:null,color:null,colorInterpolation:null,colorInterpolationFilters:null,colorProfile:null,colorRendering:null,content:null,contentScriptType:null,contentStyleType:null,crossOrigin:null,cursor:null,cx:null,cy:null,d:null,dataType:null,defaultAction:null,descent:Wr,diffuseConstant:Wr,direction:null,display:null,dur:null,divisor:Wr,dominantBaseline:null,download:ms,dx:null,dy:null,edgeMode:null,editable:null,elevation:Wr,enableBackground:null,end:null,event:null,exponent:Wr,externalResourcesRequired:null,fill:null,fillOpacity:Wr,fillRule:null,filter:null,filterRes:null,filterUnits:null,floodColor:null,floodOpacity:null,focusable:null,focusHighlight:null,fontFamily:null,fontSize:null,fontSizeAdjust:null,fontStretch:null,fontStyle:null,fontVariant:null,fontWeight:null,format:null,fr:null,from:null,fx:null,fy:null,g1:UW,g2:UW,glyphName:UW,glyphOrientationHorizontal:null,glyphOrientationVertical:null,glyphRef:null,gradientTransform:null,gradientUnits:null,handler:null,hanging:Wr,hatchContentUnits:null,hatchUnits:null,height:null,href:null,hrefLang:null,horizAdvX:Wr,horizOriginX:Wr,horizOriginY:Wr,id:null,ideographic:Wr,imageRendering:null,initialVisibility:null,in:null,in2:null,intercept:Wr,k:Wr,k1:Wr,k2:Wr,k3:Wr,k4:Wr,kernelMatrix:ZE,kernelUnitLength:null,keyPoints:null,keySplines:null,keyTimes:null,kerning:null,lang:null,lengthAdjust:null,letterSpacing:null,lightingColor:null,limitingConeAngle:Wr,local:null,markerEnd:null,markerMid:null,markerStart:null,markerHeight:null,markerUnits:null,markerWidth:null,mask:null,maskContentUnits:null,maskUnits:null,mathematical:null,max:null,media:null,mediaCharacterEncoding:null,mediaContentEncodings:null,mediaSize:Wr,mediaTime:null,method:null,min:null,mode:null,name:null,navDown:null,navDownLeft:null,navDownRight:null,navLeft:null,navNext:null,navPrev:null,navRight:null,navUp:null,navUpLeft:null,navUpRight:null,numOctaves:null,observer:null,offset:null,onAbort:null,onActivate:null,onAfterPrint:null,onBeforePrint:null,onBegin:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnd:null,onEnded:null,onError:null,onFocus:null,onFocusIn:null,onFocusOut:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadStart:null,onMessage:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onMouseWheel:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRepeat:null,onReset:null,onResize:null,onScroll:null,onSeeked:null,onSeeking:null,onSelect:null,onShow:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnload:null,onVolumeChange:null,onWaiting:null,onZoom:null,opacity:null,operator:null,order:null,orient:null,orientation:null,origin:null,overflow:null,overlay:null,overlinePosition:Wr,overlineThickness:Wr,paintOrder:null,panose1:null,path:null,pathLength:Wr,patternContentUnits:null,patternTransform:null,patternUnits:null,phase:null,ping:xd,pitch:null,playbackOrder:null,pointerEvents:null,points:null,pointsAtX:Wr,pointsAtY:Wr,pointsAtZ:Wr,preserveAlpha:null,preserveAspectRatio:null,primitiveUnits:null,propagate:null,property:ZE,r:null,radius:null,referrerPolicy:null,refX:null,refY:null,rel:ZE,rev:ZE,renderingIntent:null,repeatCount:null,repeatDur:null,requiredExtensions:ZE,requiredFeatures:ZE,requiredFonts:ZE,requiredFormats:ZE,resource:null,restart:null,result:null,rotate:null,rx:null,ry:null,scale:null,seed:null,shapeRendering:null,side:null,slope:null,snapshotTime:null,specularConstant:Wr,specularExponent:Wr,spreadMethod:null,spacing:null,startOffset:null,stdDeviation:null,stemh:null,stemv:null,stitchTiles:null,stopColor:null,stopOpacity:null,strikethroughPosition:Wr,strikethroughThickness:Wr,string:null,stroke:null,strokeDashArray:ZE,strokeDashOffset:null,strokeLineCap:null,strokeLineJoin:null,strokeMiterLimit:Wr,strokeOpacity:Wr,strokeWidth:null,style:null,surfaceScale:Wr,syncBehavior:null,syncBehaviorDefault:null,syncMaster:null,syncTolerance:null,syncToleranceDefault:null,systemLanguage:ZE,tabIndex:Wr,tableValues:null,target:null,targetX:Wr,targetY:Wr,textAnchor:null,textDecoration:null,textRendering:null,textLength:null,timelineBegin:null,title:null,transformBehavior:null,type:null,typeOf:ZE,to:null,transform:null,transformOrigin:null,u1:null,u2:null,underlinePosition:Wr,underlineThickness:Wr,unicode:null,unicodeBidi:null,unicodeRange:null,unitsPerEm:Wr,values:null,vAlphabetic:Wr,vMathematical:Wr,vectorEffect:null,vHanging:Wr,vIdeographic:Wr,version:null,vertAdvY:Wr,vertOriginX:Wr,vertOriginY:Wr,viewBox:null,viewTarget:null,visibility:null,width:null,widths:null,wordSpacing:null,writingMode:null,x:null,x1:null,x2:null,xChannelSelector:null,xHeight:Wr,y:null,y1:null,y2:null,yChannelSelector:null,z:null,zoomAndPan:null}}),vXr=/^data[-\w.:]+$/i,WWt=/-[a-z]/g,wXr=/[A-Z]/g;function yXr(n,r){const o=vKe(r);let s=r,c=W_;if(o in n.normal)return n.property[n.normal[o]];if(o.length>4&&o.slice(0,4)==="data"&&vXr.test(r)){if(r.charAt(4)==="-"){const d=r.slice(5).replace(WWt,SXr);s="data"+d.charAt(0).toUpperCase()+d.slice(1)}else{const d=r.slice(4);if(!WWt.test(d)){let h=d.replace(wXr,EXr);h.charAt(0)!=="-"&&(h="-"+h),r="data"+h}}c=qQe}return new c(s,r)}function EXr(n){return"-"+n.toLowerCase()}function SXr(n){return n.charAt(1).toUpperCase()}const _Xr=xfn([Ifn,Afn,Rfn,Pfn,mXr],"html"),Lfn=xfn([Ifn,Afn,Rfn,Pfn,bXr],"svg"),KWt={}.hasOwnProperty;function Dfn(n,r){const o=r||{};function s(c,...d){let h=s.invalid;const b=s.handlers;if(c&&KWt.call(c,n)){const y=String(c[n]);h=KWt.call(b,y)?b[y]:s.unknown}if(h)return h.call(this,c,...d)}return s.handlers=o.handlers||{},s.invalid=o.invalid,s.unknown=o.unknown,s}const kXr={},TXr={}.hasOwnProperty,Mfn=Dfn("type",{handlers:{root:CXr,element:RXr,text:OXr,comment:NXr,doctype:IXr}});function xXr(n,r){const s=(r||kXr).space;return Mfn(n,s==="svg"?Lfn:_Xr)}function CXr(n,r){const o={nodeName:"#document",mode:(n.data||{}).quirksMode?"quirks":"no-quirks",childNodes:[]};return o.childNodes=VQe(n.children,o,r),UK(n,o),o}function AXr(n,r){const o={nodeName:"#document-fragment",childNodes:[]};return o.childNodes=VQe(n.children,o,r),UK(n,o),o}function IXr(n){const r={nodeName:"#documentType",name:"html",publicId:"",systemId:"",parentNode:null};return UK(n,r),r}function OXr(n){const r={nodeName:"#text",value:n.value,parentNode:null};return UK(n,r),r}function NXr(n){const r={nodeName:"#comment",data:n.value,parentNode:null};return UK(n,r),r}function RXr(n,r){const o=r;let s=o;n.type==="element"&&n.tagName.toLowerCase()==="svg"&&o.space==="html"&&(s=Lfn);const c=[];let d;if(n.properties){for(d in n.properties)if(d!=="children"&&TXr.call(n.properties,d)){const y=PXr(s,d,n.properties[d]);y&&c.push(y)}}const h=s.space,b={nodeName:n.tagName,tagName:n.tagName,attrs:c,namespaceURI:jB[h],childNodes:[],parentNode:null};return b.childNodes=VQe(n.children,b,s),UK(n,b),n.tagName==="template"&&n.content&&(b.content=AXr(n.content,s)),b}function PXr(n,r,o){const s=yXr(n,r);if(o===!1||o===null||o===void 0||typeof o=="number"&&Number.isNaN(o)||!o&&s.boolean)return;Array.isArray(o)&&(o=s.commaSeparated?Udn(o):Ydn(o));const c={name:s.attribute,value:o===!0?"":String(o)};if(s.space&&s.space!=="html"&&s.space!=="svg"){const d=c.name.indexOf(":");d<0?c.prefix="":(c.name=c.name.slice(d+1),c.prefix=s.attribute.slice(0,d)),c.namespace=jB[s.space]}return c}function VQe(n,r,o){let s=-1;const c=[];if(n)for(;++s<n.length;){const d=Mfn(n[s],o);d.parentNode=r,c.push(d)}return c}function UK(n,r){const o=n.position;o&&o.start&&o.end&&(o.start.offset,o.end.offset,r.sourceCodeLocation={startLine:o.start.line,startCol:o.start.column,startOffset:o.start.offset,endLine:o.end.line,endCol:o.end.column,endOffset:o.end.offset})}const LXr=["area","base","basefont","bgsound","br","col","command","embed","frame","hr","image","img","input","keygen","link","meta","param","source","track","wbr"],DXr=new Set([65534,65535,131070,131071,196606,196607,262142,262143,327678,327679,393214,393215,458750,458751,524286,524287,589822,589823,655358,655359,720894,720895,786430,786431,851966,851967,917502,917503,983038,983039,1048574,1048575,1114110,1114111]),Sf="�";var Ze;(function(n){n[n.EOF=-1]="EOF",n[n.NULL=0]="NULL",n[n.TABULATION=9]="TABULATION",n[n.CARRIAGE_RETURN=13]="CARRIAGE_RETURN",n[n.LINE_FEED=10]="LINE_FEED",n[n.FORM_FEED=12]="FORM_FEED",n[n.SPACE=32]="SPACE",n[n.EXCLAMATION_MARK=33]="EXCLAMATION_MARK",n[n.QUOTATION_MARK=34]="QUOTATION_MARK",n[n.AMPERSAND=38]="AMPERSAND",n[n.APOSTROPHE=39]="APOSTROPHE",n[n.HYPHEN_MINUS=45]="HYPHEN_MINUS",n[n.SOLIDUS=47]="SOLIDUS",n[n.DIGIT_0=48]="DIGIT_0",n[n.DIGIT_9=57]="DIGIT_9",n[n.SEMICOLON=59]="SEMICOLON",n[n.LESS_THAN_SIGN=60]="LESS_THAN_SIGN",n[n.EQUALS_SIGN=61]="EQUALS_SIGN",n[n.GREATER_THAN_SIGN=62]="GREATER_THAN_SIGN",n[n.QUESTION_MARK=63]="QUESTION_MARK",n[n.LATIN_CAPITAL_A=65]="LATIN_CAPITAL_A",n[n.LATIN_CAPITAL_Z=90]="LATIN_CAPITAL_Z",n[n.RIGHT_SQUARE_BRACKET=93]="RIGHT_SQUARE_BRACKET",n[n.GRAVE_ACCENT=96]="GRAVE_ACCENT",n[n.LATIN_SMALL_A=97]="LATIN_SMALL_A",n[n.LATIN_SMALL_Z=122]="LATIN_SMALL_Z"})(Ze||(Ze={}));const zw={DASH_DASH:"--",CDATA_START:"[CDATA[",DOCTYPE:"doctype",SCRIPT:"script",PUBLIC:"public",SYSTEM:"system"};function jfn(n){return n>=55296&&n<=57343}function MXr(n){return n>=56320&&n<=57343}function jXr(n,r){return(n-55296)*1024+9216+r}function $fn(n){return n!==32&&n!==10&&n!==13&&n!==9&&n!==12&&n>=1&&n<=31||n>=127&&n<=159}function Ffn(n){return n>=64976&&n<=65007||DXr.has(n)}var On;(function(n){n.controlCharacterInInputStream="control-character-in-input-stream",n.noncharacterInInputStream="noncharacter-in-input-stream",n.surrogateInInputStream="surrogate-in-input-stream",n.nonVoidHtmlElementStartTagWithTrailingSolidus="non-void-html-element-start-tag-with-trailing-solidus",n.endTagWithAttributes="end-tag-with-attributes",n.endTagWithTrailingSolidus="end-tag-with-trailing-solidus",n.unexpectedSolidusInTag="unexpected-solidus-in-tag",n.unexpectedNullCharacter="unexpected-null-character",n.unexpectedQuestionMarkInsteadOfTagName="unexpected-question-mark-instead-of-tag-name",n.invalidFirstCharacterOfTagName="invalid-first-character-of-tag-name",n.unexpectedEqualsSignBeforeAttributeName="unexpected-equals-sign-before-attribute-name",n.missingEndTagName="missing-end-tag-name",n.unexpectedCharacterInAttributeName="unexpected-character-in-attribute-name",n.unknownNamedCharacterReference="unknown-named-character-reference",n.missingSemicolonAfterCharacterReference="missing-semicolon-after-character-reference",n.unexpectedCharacterAfterDoctypeSystemIdentifier="unexpected-character-after-doctype-system-identifier",n.unexpectedCharacterInUnquotedAttributeValue="unexpected-character-in-unquoted-attribute-value",n.eofBeforeTagName="eof-before-tag-name",n.eofInTag="eof-in-tag",n.missingAttributeValue="missing-attribute-value",n.missingWhitespaceBetweenAttributes="missing-whitespace-between-attributes",n.missingWhitespaceAfterDoctypePublicKeyword="missing-whitespace-after-doctype-public-keyword",n.missingWhitespaceBetweenDoctypePublicAndSystemIdentifiers="missing-whitespace-between-doctype-public-and-system-identifiers",n.missingWhitespaceAfterDoctypeSystemKeyword="missing-whitespace-after-doctype-system-keyword",n.missingQuoteBeforeDoctypePublicIdentifier="missing-quote-before-doctype-public-identifier",n.missingQuoteBeforeDoctypeSystemIdentifier="missing-quote-before-doctype-system-identifier",n.missingDoctypePublicIdentifier="missing-doctype-public-identifier",n.missingDoctypeSystemIdentifier="missing-doctype-system-identifier",n.abruptDoctypePublicIdentifier="abrupt-doctype-public-identifier",n.abruptDoctypeSystemIdentifier="abrupt-doctype-system-identifier",n.cdataInHtmlContent="cdata-in-html-content",n.incorrectlyOpenedComment="incorrectly-opened-comment",n.eofInScriptHtmlCommentLikeText="eof-in-script-html-comment-like-text",n.eofInDoctype="eof-in-doctype",n.nestedComment="nested-comment",n.abruptClosingOfEmptyComment="abrupt-closing-of-empty-comment",n.eofInComment="eof-in-comment",n.incorrectlyClosedComment="incorrectly-closed-comment",n.eofInCdata="eof-in-cdata",n.absenceOfDigitsInNumericCharacterReference="absence-of-digits-in-numeric-character-reference",n.nullCharacterReference="null-character-reference",n.surrogateCharacterReference="surrogate-character-reference",n.characterReferenceOutsideUnicodeRange="character-reference-outside-unicode-range",n.controlCharacterReference="control-character-reference",n.noncharacterCharacterReference="noncharacter-character-reference",n.missingWhitespaceBeforeDoctypeName="missing-whitespace-before-doctype-name",n.missingDoctypeName="missing-doctype-name",n.invalidCharacterSequenceAfterDoctypeName="invalid-character-sequence-after-doctype-name",n.duplicateAttribute="duplicate-attribute",n.nonConformingDoctype="non-conforming-doctype",n.missingDoctype="missing-doctype",n.misplacedDoctype="misplaced-doctype",n.endTagWithoutMatchingOpenElement="end-tag-without-matching-open-element",n.closingOfElementWithOpenChildElements="closing-of-element-with-open-child-elements",n.disallowedContentInNoscriptInHead="disallowed-content-in-noscript-in-head",n.openElementsLeftAfterEof="open-elements-left-after-eof",n.abandonedHeadElementChild="abandoned-head-element-child",n.misplacedStartTagForHeadElement="misplaced-start-tag-for-head-element",n.nestedNoscriptInHead="nested-noscript-in-head",n.eofInElementThatCanContainOnlyText="eof-in-element-that-can-contain-only-text"})(On||(On={}));const $Xr=65536;class FXr{constructor(r){this.handler=r,this.html="",this.pos=-1,this.lastGapPos=-2,this.gapStack=[],this.skipNextNewLine=!1,this.lastChunkWritten=!1,this.endOfChunkHit=!1,this.bufferWaterline=$Xr,this.isEol=!1,this.lineStartPos=0,this.droppedBufferSize=0,this.line=1,this.lastErrOffset=-1}get col(){return this.pos-this.lineStartPos+ +(this.lastGapPos!==this.pos)}get offset(){return this.droppedBufferSize+this.pos}getError(r,o){const{line:s,col:c,offset:d}=this,h=c+o,b=d+o;return{code:r,startLine:s,endLine:s,startCol:h,endCol:h,startOffset:b,endOffset:b}}_err(r){this.handler.onParseError&&this.lastErrOffset!==this.offset&&(this.lastErrOffset=this.offset,this.handler.onParseError(this.getError(r,0)))}_addGap(){this.gapStack.push(this.lastGapPos),this.lastGapPos=this.pos}_processSurrogate(r){if(this.pos!==this.html.length-1){const o=this.html.charCodeAt(this.pos+1);if(MXr(o))return this.pos++,this._addGap(),jXr(r,o)}else if(!this.lastChunkWritten)return this.endOfChunkHit=!0,Ze.EOF;return this._err(On.surrogateInInputStream),r}willDropParsedChunk(){return this.pos>this.bufferWaterline}dropParsedChunk(){this.willDropParsedChunk()&&(this.html=this.html.substring(this.pos),this.lineStartPos-=this.pos,this.droppedBufferSize+=this.pos,this.pos=0,this.lastGapPos=-2,this.gapStack.length=0)}write(r,o){this.html.length>0?this.html+=r:this.html=r,this.endOfChunkHit=!1,this.lastChunkWritten=o}insertHtmlAtCurrentPos(r){this.html=this.html.substring(0,this.pos+1)+r+this.html.substring(this.pos+1),this.endOfChunkHit=!1}startsWith(r,o){if(this.pos+r.length>this.html.length)return this.endOfChunkHit=!this.lastChunkWritten,!1;if(o)return this.html.startsWith(r,this.pos);for(let s=0;s<r.length;s++)if((this.html.charCodeAt(this.pos+s)|32)!==r.charCodeAt(s))return!1;return!0}peek(r){const o=this.pos+r;if(o>=this.html.length)return this.endOfChunkHit=!this.lastChunkWritten,Ze.EOF;const s=this.html.charCodeAt(o);return s===Ze.CARRIAGE_RETURN?Ze.LINE_FEED:s}advance(){if(this.pos++,this.isEol&&(this.isEol=!1,this.line++,this.lineStartPos=this.pos),this.pos>=this.html.length)return this.endOfChunkHit=!this.lastChunkWritten,Ze.EOF;let r=this.html.charCodeAt(this.pos);return r===Ze.CARRIAGE_RETURN?(this.isEol=!0,this.skipNextNewLine=!0,Ze.LINE_FEED):r===Ze.LINE_FEED&&(this.isEol=!0,this.skipNextNewLine)?(this.line--,this.skipNextNewLine=!1,this._addGap(),this.advance()):(this.skipNextNewLine=!1,jfn(r)&&(r=this._processSurrogate(r)),this.handler.onParseError===null||r>31&&r<127||r===Ze.LINE_FEED||r===Ze.CARRIAGE_RETURN||r>159&&r<64976||this._checkForProblematicCharacters(r),r)}_checkForProblematicCharacters(r){$fn(r)?this._err(On.controlCharacterInInputStream):Ffn(r)&&this._err(On.noncharacterInInputStream)}retreat(r){for(this.pos-=r;this.pos<this.lastGapPos;)this.lastGapPos=this.gapStack.pop(),this.pos--;this.isEol=!1}}var nc;(function(n){n[n.CHARACTER=0]="CHARACTER",n[n.NULL_CHARACTER=1]="NULL_CHARACTER",n[n.WHITESPACE_CHARACTER=2]="WHITESPACE_CHARACTER",n[n.START_TAG=3]="START_TAG",n[n.END_TAG=4]="END_TAG",n[n.COMMENT=5]="COMMENT",n[n.DOCTYPE=6]="DOCTYPE",n[n.EOF=7]="EOF",n[n.HIBERNATION=8]="HIBERNATION"})(nc||(nc={}));function Bfn(n,r){for(let o=n.attrs.length-1;o>=0;o--)if(n.attrs[o].name===r)return n.attrs[o].value;return null}const BXr=new Uint16Array('ᵁ<Õıʊҝջאٵ۞ޢߖࠏ੊ઑඡ๭༉༦჊ረዡᐕᒝᓃᓟᔥ\0\0\0\0\0\0ᕫᛍᦍᰒᷝ὾⁠↰⊍⏀⏻⑂⠤⤒ⴈ⹈⿎〖㊺㘹㞬㣾㨨㩱㫠㬮ࠀEMabcfglmnoprstu\\bfms¦³¹ÈÏlig耻Æ䃆P耻&䀦cute耻Á䃁reve;䄂Āiyx}rc耻Â䃂;䐐r;쀀𝔄rave耻À䃀pha;䎑acr;䄀d;橓Āgp¡on;䄄f;쀀𝔸plyFunction;恡ing耻Å䃅Ācs¾Ãr;쀀𝒜ign;扔ilde耻Ã䃃ml耻Ä䃄ЀaceforsuåûþėĜĢħĪĀcrêòkslash;或Ŷöø;櫧ed;挆y;䐑ƀcrtąċĔause;戵noullis;愬a;䎒r;쀀𝔅pf;쀀𝔹eve;䋘còēmpeq;扎܀HOacdefhilorsuōőŖƀƞƢƵƷƺǜȕɳɸɾcy;䐧PY耻©䂩ƀcpyŝŢźute;䄆Ā;iŧŨ拒talDifferentialD;慅leys;愭ȀaeioƉƎƔƘron;䄌dil耻Ç䃇rc;䄈nint;戰ot;䄊ĀdnƧƭilla;䂸terDot;䂷òſi;䎧rcleȀDMPTǇǋǑǖot;抙inus;抖lus;投imes;抗oĀcsǢǸkwiseContourIntegral;戲eCurlyĀDQȃȏoubleQuote;思uote;怙ȀlnpuȞȨɇɕonĀ;eȥȦ户;橴ƀgitȯȶȺruent;扡nt;戯ourIntegral;戮ĀfrɌɎ;愂oduct;成nterClockwiseContourIntegral;戳oss;樯cr;쀀𝒞pĀ;Cʄʅ拓ap;才րDJSZacefiosʠʬʰʴʸˋ˗ˡ˦̳ҍĀ;oŹʥtrahd;椑cy;䐂cy;䐅cy;䐏ƀgrsʿ˄ˇger;怡r;憡hv;櫤Āayː˕ron;䄎;䐔lĀ;t˝˞戇a;䎔r;쀀𝔇Āaf˫̧Ācm˰̢riticalȀADGT̖̜̀̆cute;䂴oŴ̋̍;䋙bleAcute;䋝rave;䁠ilde;䋜ond;拄ferentialD;慆Ѱ̽\0\0\0͔͂\0Ѕf;쀀𝔻ƀ;DE͈͉͍䂨ot;惜qual;扐blèCDLRUVͣͲ΂ϏϢϸontourIntegraìȹoɴ͹\0\0ͻ»͉nArrow;懓Āeo·ΤftƀARTΐΖΡrrow;懐ightArrow;懔eåˊngĀLRΫτeftĀARγιrrow;柸ightArrow;柺ightArrow;柹ightĀATϘϞrrow;懒ee;抨pɁϩ\0\0ϯrrow;懑ownArrow;懕erticalBar;戥ǹABLRTaВЪаўѿͼrrowƀ;BUНОТ憓ar;椓pArrow;懵reve;䌑eft˒к\0ц\0ѐightVector;楐eeVector;楞ectorĀ;Bљњ憽ar;楖ightǔѧ\0ѱeeVector;楟ectorĀ;BѺѻ懁ar;楗eeĀ;A҆҇护rrow;憧ĀctҒҗr;쀀𝒟rok;䄐ࠀNTacdfglmopqstuxҽӀӄӋӞӢӧӮӵԡԯԶՒ՝ՠեG;䅊H耻Ð䃐cute耻É䃉ƀaiyӒӗӜron;䄚rc耻Ê䃊;䐭ot;䄖r;쀀𝔈rave耻È䃈ement;戈ĀapӺӾcr;䄒tyɓԆ\0\0ԒmallSquare;旻erySmallSquare;斫ĀgpԦԪon;䄘f;쀀𝔼silon;䎕uĀaiԼՉlĀ;TՂՃ橵ilde;扂librium;懌Āci՗՚r;愰m;橳a;䎗ml耻Ë䃋Āipժկsts;戃onentialE;慇ʀcfiosօֈ֍ֲ׌y;䐤r;쀀𝔉lledɓ֗\0\0֣mallSquare;旼erySmallSquare;斪Ͱֺ\0ֿ\0\0ׄf;쀀𝔽All;戀riertrf;愱cò׋؀JTabcdfgorstר׬ׯ׺؀ؒؖ؛؝أ٬ٲcy;䐃耻>䀾mmaĀ;d׷׸䎓;䏜reve;䄞ƀeiy؇،ؐdil;䄢rc;䄜;䐓ot;䄠r;쀀𝔊;拙pf;쀀𝔾eater̀EFGLSTصلَٖٛ٦qualĀ;Lؾؿ扥ess;招ullEqual;执reater;檢ess;扷lantEqual;橾ilde;扳cr;쀀𝒢;扫ЀAacfiosuڅڋږڛڞڪھۊRDcy;䐪Āctڐڔek;䋇;䁞irc;䄤r;愌lbertSpace;愋ǰگ\0ڲf;愍izontalLine;攀Āctۃۅòکrok;䄦mpńېۘownHumðįqual;扏܀EJOacdfgmnostuۺ۾܃܇܎ܚܞܡܨ݄ݸދޏޕcy;䐕lig;䄲cy;䐁cute耻Í䃍Āiyܓܘrc耻Î䃎;䐘ot;䄰r;愑rave耻Ì䃌ƀ;apܠܯܿĀcgܴܷr;䄪inaryI;慈lieóϝǴ݉\0ݢĀ;eݍݎ戬Āgrݓݘral;戫section;拂isibleĀCTݬݲomma;恣imes;恢ƀgptݿރވon;䄮f;쀀𝕀a;䎙cr;愐ilde;䄨ǫޚ\0ޞcy;䐆l耻Ï䃏ʀcfosuެ޷޼߂ߐĀiyޱ޵rc;䄴;䐙r;쀀𝔍pf;쀀𝕁ǣ߇\0ߌr;쀀𝒥rcy;䐈kcy;䐄΀HJacfosߤߨ߽߬߱ࠂࠈcy;䐥cy;䐌ppa;䎚Āey߶߻dil;䄶;䐚r;쀀𝔎pf;쀀𝕂cr;쀀𝒦րJTaceflmostࠥࠩࠬࡐࡣ঳সে্਷ੇcy;䐉耻<䀼ʀcmnpr࠷࠼ࡁࡄࡍute;䄹bda;䎛g;柪lacetrf;愒r;憞ƀaeyࡗ࡜ࡡron;䄽dil;䄻;䐛Āfsࡨ॰tԀACDFRTUVarࡾࢩࢱࣦ࣠ࣼयज़ΐ४Ānrࢃ࢏gleBracket;柨rowƀ;BR࢙࢚࢞憐ar;懤ightArrow;懆eiling;挈oǵࢷ\0ࣃbleBracket;柦nǔࣈ\0࣒eeVector;楡ectorĀ;Bࣛࣜ懃ar;楙loor;挊ightĀAV࣯ࣵrrow;憔ector;楎Āerँगeƀ;AVउऊऐ抣rrow;憤ector;楚iangleƀ;BEतथऩ抲ar;槏qual;抴pƀDTVषूौownVector;楑eeVector;楠ectorĀ;Bॖॗ憿ar;楘ectorĀ;B॥०憼ar;楒ightáΜs̀EFGLSTॾঋকঝঢভqualGreater;拚ullEqual;扦reater;扶ess;檡lantEqual;橽ilde;扲r;쀀𝔏Ā;eঽা拘ftarrow;懚idot;䄿ƀnpw৔ਖਛgȀLRlr৞৷ਂਐeftĀAR০৬rrow;柵ightArrow;柷ightArrow;柶eftĀarγਊightáοightáϊf;쀀𝕃erĀLRਢਬeftArrow;憙ightArrow;憘ƀchtਾੀੂòࡌ;憰rok;䅁;扪Ѐacefiosuਗ਼੝੠੷੼અઋ઎p;椅y;䐜Ādl੥੯iumSpace;恟lintrf;愳r;쀀𝔐nusPlus;戓pf;쀀𝕄cò੶;䎜ҀJacefostuણધભીଔଙඑ඗ඞcy;䐊cute;䅃ƀaey઴હાron;䅇dil;䅅;䐝ƀgswે૰଎ativeƀMTV૓૟૨ediumSpace;怋hiĀcn૦૘ë૙eryThiî૙tedĀGL૸ଆreaterGreateòٳessLesóੈLine;䀊r;쀀𝔑ȀBnptଢନଷ଺reak;恠BreakingSpace;䂠f;愕ڀ;CDEGHLNPRSTV୕ୖ୪୼஡௫ఄ౞಄ದ೘ൡඅ櫬Āou୛୤ngruent;扢pCap;扭oubleVerticalBar;戦ƀlqxஃஊ஛ement;戉ualĀ;Tஒஓ扠ilde;쀀≂̸ists;戄reater΀;EFGLSTஶஷ஽௉௓௘௥扯qual;扱ullEqual;쀀≧̸reater;쀀≫̸ess;批lantEqual;쀀⩾̸ilde;扵umpń௲௽ownHump;쀀≎̸qual;쀀≏̸eĀfsఊధtTriangleƀ;BEచఛడ拪ar;쀀⧏̸qual;括s̀;EGLSTవశ఼ౄోౘ扮qual;扰reater;扸ess;쀀≪̸lantEqual;쀀⩽̸ilde;扴estedĀGL౨౹reaterGreater;쀀⪢̸essLess;쀀⪡̸recedesƀ;ESಒಓಛ技qual;쀀⪯̸lantEqual;拠ĀeiಫಹverseElement;戌ghtTriangleƀ;BEೋೌ೒拫ar;쀀⧐̸qual;拭ĀquೝഌuareSuĀbp೨೹setĀ;E೰ೳ쀀⊏̸qual;拢ersetĀ;Eഃആ쀀⊐̸qual;拣ƀbcpഓതൎsetĀ;Eഛഞ쀀⊂⃒qual;抈ceedsȀ;ESTലള഻െ抁qual;쀀⪰̸lantEqual;拡ilde;쀀≿̸ersetĀ;E൘൛쀀⊃⃒qual;抉ildeȀ;EFT൮൯൵ൿ扁qual;扄ullEqual;扇ilde;扉erticalBar;戤cr;쀀𝒩ilde耻Ñ䃑;䎝܀Eacdfgmoprstuvලෂ෉෕ෛ෠෧෼ขภยา฿ไlig;䅒cute耻Ó䃓Āiy෎ීrc耻Ô䃔;䐞blac;䅐r;쀀𝔒rave耻Ò䃒ƀaei෮ෲ෶cr;䅌ga;䎩cron;䎟pf;쀀𝕆enCurlyĀDQฎบoubleQuote;怜uote;怘;橔Āclวฬr;쀀𝒪ash耻Ø䃘iŬื฼de耻Õ䃕es;樷ml耻Ö䃖erĀBP๋๠Āar๐๓r;怾acĀek๚๜;揞et;掴arenthesis;揜Ҁacfhilors๿ງຊຏຒດຝະ໼rtialD;戂y;䐟r;쀀𝔓i;䎦;䎠usMinus;䂱Āipຢອncareplanåڝf;愙Ȁ;eio຺ູ໠໤檻cedesȀ;EST່້໏໚扺qual;檯lantEqual;扼ilde;找me;怳Ādp໩໮uct;戏ortionĀ;aȥ໹l;戝Āci༁༆r;쀀𝒫;䎨ȀUfos༑༖༛༟OT耻"䀢r;쀀𝔔pf;愚cr;쀀𝒬؀BEacefhiorsu༾གྷཇའཱིྦྷྪྭ႖ႩႴႾarr;椐G耻®䂮ƀcnrཎནབute;䅔g;柫rĀ;tཛྷཝ憠l;椖ƀaeyཧཬཱron;䅘dil;䅖;䐠Ā;vླྀཹ愜erseĀEUྂྙĀlq྇ྎement;戋uilibrium;懋pEquilibrium;楯r»ཹo;䎡ghtЀACDFTUVa࿁࿫࿳ဢဨၛႇϘĀnr࿆࿒gleBracket;柩rowƀ;BL࿜࿝࿡憒ar;懥eftArrow;懄eiling;按oǵ࿹\0စbleBracket;柧nǔည\0နeeVector;楝ectorĀ;Bဝသ懂ar;楕loor;挋Āerိ၃eƀ;AVဵံြ抢rrow;憦ector;楛iangleƀ;BEၐၑၕ抳ar;槐qual;抵pƀDTVၣၮၸownVector;楏eeVector;楜ectorĀ;Bႂႃ憾ar;楔ectorĀ;B႑႒懀ar;楓Āpuႛ႞f;愝ndImplies;楰ightarrow;懛ĀchႹႼr;愛;憱leDelayed;槴ڀHOacfhimoqstuფჱჷჽᄙᄞᅑᅖᅡᅧᆵᆻᆿĀCcჩხHcy;䐩y;䐨FTcy;䐬cute;䅚ʀ;aeiyᄈᄉᄎᄓᄗ檼ron;䅠dil;䅞rc;䅜;䐡r;쀀𝔖ortȀDLRUᄪᄴᄾᅉownArrow»ОeftArrow»࢚ightArrow»࿝pArrow;憑gma;䎣allCircle;战pf;쀀𝕊ɲᅭ\0\0ᅰt;戚areȀ;ISUᅻᅼᆉᆯ斡ntersection;抓uĀbpᆏᆞsetĀ;Eᆗᆘ抏qual;抑ersetĀ;Eᆨᆩ抐qual;抒nion;抔cr;쀀𝒮ar;拆ȀbcmpᇈᇛሉላĀ;sᇍᇎ拐etĀ;Eᇍᇕqual;抆ĀchᇠህeedsȀ;ESTᇭᇮᇴᇿ扻qual;檰lantEqual;扽ilde;承Tháྌ;我ƀ;esሒሓሣ拑rsetĀ;Eሜም抃qual;抇et»ሓրHRSacfhiorsሾቄ቉ቕ቞ቱቶኟዂወዑORN耻Þ䃞ADE;愢ĀHc቎ቒcy;䐋y;䐦Ābuቚቜ;䀉;䎤ƀaeyብቪቯron;䅤dil;䅢;䐢r;쀀𝔗Āeiቻ኉ǲኀ\0ኇefore;戴a;䎘Ācn኎ኘkSpace;쀀  Space;怉ldeȀ;EFTካኬኲኼ戼qual;扃ullEqual;扅ilde;扈pf;쀀𝕋ipleDot;惛Āctዖዛr;쀀𝒯rok;䅦ૡዷጎጚጦ\0ጬጱ\0\0\0\0\0ጸጽ፷ᎅ\0᏿ᐄᐊᐐĀcrዻጁute耻Ú䃚rĀ;oጇገ憟cir;楉rǣጓ\0጖y;䐎ve;䅬Āiyጞጣrc耻Û䃛;䐣blac;䅰r;쀀𝔘rave耻Ù䃙acr;䅪Ādiፁ፩erĀBPፈ፝Āarፍፐr;䁟acĀekፗፙ;揟et;掵arenthesis;揝onĀ;P፰፱拃lus;抎Āgp፻፿on;䅲f;쀀𝕌ЀADETadps᎕ᎮᎸᏄϨᏒᏗᏳrrowƀ;BDᅐᎠᎤar;椒ownArrow;懅ownArrow;憕quilibrium;楮eeĀ;AᏋᏌ报rrow;憥ownáϳerĀLRᏞᏨeftArrow;憖ightArrow;憗iĀ;lᏹᏺ䏒on;䎥ing;䅮cr;쀀𝒰ilde;䅨ml耻Ü䃜ҀDbcdefosvᐧᐬᐰᐳᐾᒅᒊᒐᒖash;披ar;櫫y;䐒ashĀ;lᐻᐼ抩;櫦Āerᑃᑅ;拁ƀbtyᑌᑐᑺar;怖Ā;iᑏᑕcalȀBLSTᑡᑥᑪᑴar;戣ine;䁼eparator;杘ilde;所ThinSpace;怊r;쀀𝔙pf;쀀𝕍cr;쀀𝒱dash;抪ʀcefosᒧᒬᒱᒶᒼirc;䅴dge;拀r;쀀𝔚pf;쀀𝕎cr;쀀𝒲Ȁfiosᓋᓐᓒᓘr;쀀𝔛;䎞pf;쀀𝕏cr;쀀𝒳ҀAIUacfosuᓱᓵᓹᓽᔄᔏᔔᔚᔠcy;䐯cy;䐇cy;䐮cute耻Ý䃝Āiyᔉᔍrc;䅶;䐫r;쀀𝔜pf;쀀𝕐cr;쀀𝒴ml;䅸ЀHacdefosᔵᔹᔿᕋᕏᕝᕠᕤcy;䐖cute;䅹Āayᕄᕉron;䅽;䐗ot;䅻ǲᕔ\0ᕛoWidtè૙a;䎖r;愨pf;愤cr;쀀𝒵௡ᖃᖊᖐ\0ᖰᖶᖿ\0\0\0\0ᗆᗛᗫᙟ᙭\0ᚕ᚛ᚲᚹ\0ᚾcute耻á䃡reve;䄃̀;Ediuyᖜᖝᖡᖣᖨᖭ戾;쀀∾̳;房rc耻â䃢te肻´̆;䐰lig耻æ䃦Ā;r²ᖺ;쀀𝔞rave耻à䃠ĀepᗊᗖĀfpᗏᗔsym;愵èᗓha;䎱ĀapᗟcĀclᗤᗧr;䄁g;樿ɤᗰ\0\0ᘊʀ;adsvᗺᗻᗿᘁᘇ戧nd;橕;橜lope;橘;橚΀;elmrszᘘᘙᘛᘞᘿᙏᙙ戠;榤e»ᘙsdĀ;aᘥᘦ戡ѡᘰᘲᘴᘶᘸᘺᘼᘾ;榨;榩;榪;榫;榬;榭;榮;榯tĀ;vᙅᙆ戟bĀ;dᙌᙍ抾;榝Āptᙔᙗh;戢»¹arr;捼Āgpᙣᙧon;䄅f;쀀𝕒΀;Eaeiop዁ᙻᙽᚂᚄᚇᚊ;橰cir;橯;扊d;手s;䀧roxĀ;e዁ᚒñᚃing耻å䃥ƀctyᚡᚦᚨr;쀀𝒶;䀪mpĀ;e዁ᚯñʈilde耻ã䃣ml耻ä䃤Āciᛂᛈoninôɲnt;樑ࠀNabcdefiklnoprsu᛭ᛱᜰ᜼ᝃᝈ᝸᝽០៦ᠹᡐᜍ᤽᥈ᥰot;櫭Ācrᛶ᜞kȀcepsᜀᜅᜍᜓong;扌psilon;䏶rime;怵imĀ;e᜚᜛戽q;拍Ŷᜢᜦee;抽edĀ;gᜬᜭ挅e»ᜭrkĀ;t፜᜷brk;掶Āoyᜁᝁ;䐱quo;怞ʀcmprtᝓ᝛ᝡᝤᝨausĀ;eĊĉptyv;榰séᜌnoõēƀahwᝯ᝱ᝳ;䎲;愶een;扬r;쀀𝔟g΀costuvwឍឝឳេ៕៛៞ƀaiuបពរðݠrc;旯p»፱ƀdptឤឨឭot;樀lus;樁imes;樂ɱឹ\0\0ើcup;樆ar;昅riangleĀdu៍្own;施p;斳plus;樄eåᑄåᒭarow;植ƀako៭ᠦᠵĀcn៲ᠣkƀlst៺֫᠂ozenge;槫riangleȀ;dlr᠒᠓᠘᠝斴own;斾eft;旂ight;斸k;搣Ʊᠫ\0ᠳƲᠯ\0ᠱ;斒;斑4;斓ck;斈ĀeoᠾᡍĀ;qᡃᡆ쀀=⃥uiv;쀀≡⃥t;挐Ȁptwxᡙᡞᡧᡬf;쀀𝕓Ā;tᏋᡣom»Ꮜtie;拈؀DHUVbdhmptuvᢅᢖᢪᢻᣗᣛᣬ᣿ᤅᤊᤐᤡȀLRlrᢎᢐᢒᢔ;敗;敔;敖;敓ʀ;DUduᢡᢢᢤᢦᢨ敐;敦;敩;敤;敧ȀLRlrᢳᢵᢷᢹ;敝;敚;敜;教΀;HLRhlrᣊᣋᣍᣏᣑᣓᣕ救;敬;散;敠;敫;敢;敟ox;槉ȀLRlrᣤᣦᣨᣪ;敕;敒;攐;攌ʀ;DUduڽ᣷᣹᣻᣽;敥;敨;攬;攴inus;抟lus;択imes;抠ȀLRlrᤙᤛᤝ᤟;敛;敘;攘;攔΀;HLRhlrᤰᤱᤳᤵᤷ᤻᤹攂;敪;敡;敞;攼;攤;攜Āevģ᥂bar耻¦䂦Ȁceioᥑᥖᥚᥠr;쀀𝒷mi;恏mĀ;e᜚᜜lƀ;bhᥨᥩᥫ䁜;槅sub;柈Ŭᥴ᥾lĀ;e᥹᥺怢t»᥺pƀ;Eeįᦅᦇ;檮Ā;qۜۛೡᦧ\0᧨ᨑᨕᨲ\0ᨷᩐ\0\0᪴\0\0᫁\0\0ᬡᬮ᭍᭒\0᯽\0ᰌƀcpr᦭ᦲ᧝ute;䄇̀;abcdsᦿᧀᧄ᧊᧕᧙戩nd;橄rcup;橉Āau᧏᧒p;橋p;橇ot;橀;쀀∩︀Āeo᧢᧥t;恁îړȀaeiu᧰᧻ᨁᨅǰ᧵\0᧸s;橍on;䄍dil耻ç䃧rc;䄉psĀ;sᨌᨍ橌m;橐ot;䄋ƀdmnᨛᨠᨦil肻¸ƭptyv;榲t脀¢;eᨭᨮ䂢räƲr;쀀𝔠ƀceiᨽᩀᩍy;䑇ckĀ;mᩇᩈ朓ark»ᩈ;䏇r΀;Ecefms᩟᩠ᩢᩫ᪤᪪᪮旋;槃ƀ;elᩩᩪᩭ䋆q;扗eɡᩴ\0\0᪈rrowĀlr᩼᪁eft;憺ight;憻ʀRSacd᪒᪔᪖᪚᪟»ཇ;擈st;抛irc;抚ash;抝nint;樐id;櫯cir;槂ubsĀ;u᪻᪼晣it»᪼ˬ᫇᫔᫺\0ᬊonĀ;eᫍᫎ䀺Ā;qÇÆɭ᫙\0\0᫢aĀ;t᫞᫟䀬;䁀ƀ;fl᫨᫩᫫戁îᅠeĀmx᫱᫶ent»᫩eóɍǧ᫾\0ᬇĀ;dኻᬂot;橭nôɆƀfryᬐᬔᬗ;쀀𝕔oäɔ脀©;sŕᬝr;愗Āaoᬥᬩrr;憵ss;朗Ācuᬲᬷr;쀀𝒸Ābpᬼ᭄Ā;eᭁᭂ櫏;櫑Ā;eᭉᭊ櫐;櫒dot;拯΀delprvw᭠᭬᭷ᮂᮬᯔ᯹arrĀlr᭨᭪;椸;椵ɰ᭲\0\0᭵r;拞c;拟arrĀ;p᭿ᮀ憶;椽̀;bcdosᮏᮐᮖᮡᮥᮨ截rcap;橈Āauᮛᮞp;橆p;橊ot;抍r;橅;쀀∪︀Ȁalrv᮵ᮿᯞᯣrrĀ;mᮼᮽ憷;椼yƀevwᯇᯔᯘqɰᯎ\0\0ᯒreã᭳uã᭵ee;拎edge;拏en耻¤䂤earrowĀlrᯮ᯳eft»ᮀight»ᮽeäᯝĀciᰁᰇoninôǷnt;戱lcty;挭ঀAHabcdefhijlorstuwz᰸᰻᰿ᱝᱩᱵᲊᲞᲬᲷ᳻᳿ᴍᵻᶑᶫᶻ᷆᷍rò΁ar;楥Ȁglrs᱈ᱍ᱒᱔ger;怠eth;愸òᄳhĀ;vᱚᱛ怐»ऊūᱡᱧarow;椏aã̕Āayᱮᱳron;䄏;䐴ƀ;ao̲ᱼᲄĀgrʿᲁr;懊tseq;橷ƀglmᲑᲔᲘ耻°䂰ta;䎴ptyv;榱ĀirᲣᲨsht;楿;쀀𝔡arĀlrᲳᲵ»ࣜ»သʀaegsv᳂͸᳖᳜᳠mƀ;oș᳊᳔ndĀ;ș᳑uit;晦amma;䏝in;拲ƀ;io᳧᳨᳸䃷de脀÷;o᳧ᳰntimes;拇nø᳷cy;䑒cɯᴆ\0\0ᴊrn;挞op;挍ʀlptuwᴘᴝᴢᵉᵕlar;䀤f;쀀𝕕ʀ;emps̋ᴭᴷᴽᵂqĀ;d͒ᴳot;扑inus;戸lus;戔quare;抡blebarwedgåúnƀadhᄮᵝᵧownarrowóᲃarpoonĀlrᵲᵶefôᲴighôᲶŢᵿᶅkaro÷གɯᶊ\0\0ᶎrn;挟op;挌ƀcotᶘᶣᶦĀryᶝᶡ;쀀𝒹;䑕l;槶rok;䄑Ādrᶰᶴot;拱iĀ;fᶺ᠖斿Āah᷀᷃ròЩaòྦangle;榦Āci᷒ᷕy;䑟grarr;柿ऀDacdefglmnopqrstuxḁḉḙḸոḼṉṡṾấắẽỡἪἷὄ὎὚ĀDoḆᴴoôᲉĀcsḎḔute耻é䃩ter;橮ȀaioyḢḧḱḶron;䄛rĀ;cḭḮ扖耻ê䃪lon;払;䑍ot;䄗ĀDrṁṅot;扒;쀀𝔢ƀ;rsṐṑṗ檚ave耻è䃨Ā;dṜṝ檖ot;檘Ȁ;ilsṪṫṲṴ檙nters;揧;愓Ā;dṹṺ檕ot;檗ƀapsẅẉẗcr;䄓tyƀ;svẒẓẕ戅et»ẓpĀ1;ẝẤĳạả;怄;怅怃ĀgsẪẬ;䅋p;怂ĀgpẴẸon;䄙f;쀀𝕖ƀalsỄỎỒrĀ;sỊị拕l;槣us;橱iƀ;lvỚớở䎵on»ớ;䏵ȀcsuvỪỳἋἣĀioữḱrc»Ḯɩỹ\0\0ỻíՈantĀglἂἆtr»ṝess»Ṻƀaeiἒ἖Ἒls;䀽st;扟vĀ;DȵἠD;橸parsl;槥ĀDaἯἳot;打rr;楱ƀcdiἾὁỸr;愯oô͒ĀahὉὋ;䎷耻ð䃰Āmrὓὗl耻ë䃫o;悬ƀcipὡὤὧl;䀡sôծĀeoὬὴctatioîՙnentialåչৡᾒ\0ᾞ\0ᾡᾧ\0\0ῆῌ\0ΐ\0ῦῪ \0 ⁚llingdotseñṄy;䑄male;晀ƀilrᾭᾳ῁lig;耀ﬃɩᾹ\0\0᾽g;耀ﬀig;耀ﬄ;쀀𝔣lig;耀ﬁlig;쀀fjƀaltῙ῜ῡt;晭ig;耀ﬂns;斱of;䆒ǰ΅\0ῳf;쀀𝕗ĀakֿῷĀ;vῼ´拔;櫙artint;樍Āao‌⁕Ācs‑⁒α‚‰‸⁅⁈\0⁐β•‥‧‪‬\0‮耻½䂽;慓耻¼䂼;慕;慙;慛Ƴ‴\0‶;慔;慖ʴ‾⁁\0\0⁃耻¾䂾;慗;慜5;慘ƶ⁌\0⁎;慚;慝8;慞l;恄wn;挢cr;쀀𝒻ࢀEabcdefgijlnorstv₂₉₟₥₰₴⃰⃵⃺⃿℃ℒℸ̗ℾ⅒↞Ā;lٍ₇;檌ƀcmpₐₕ₝ute;䇵maĀ;dₜ᳚䎳;檆reve;䄟Āiy₪₮rc;䄝;䐳ot;䄡Ȁ;lqsؾق₽⃉ƀ;qsؾٌ⃄lanô٥Ȁ;cdl٥⃒⃥⃕c;檩otĀ;o⃜⃝檀Ā;l⃢⃣檂;檄Ā;e⃪⃭쀀⋛︀s;檔r;쀀𝔤Ā;gٳ؛mel;愷cy;䑓Ȁ;Eajٚℌℎℐ;檒;檥;檤ȀEaesℛℝ℩ℴ;扩pĀ;p℣ℤ檊rox»ℤĀ;q℮ℯ檈Ā;q℮ℛim;拧pf;쀀𝕘Āci⅃ⅆr;愊mƀ;el٫ⅎ⅐;檎;檐茀>;cdlqr׮ⅠⅪⅮⅳⅹĀciⅥⅧ;檧r;橺ot;拗Par;榕uest;橼ʀadelsↄⅪ←ٖ↛ǰ↉\0↎proø₞r;楸qĀlqؿ↖lesó₈ií٫Āen↣↭rtneqq;쀀≩︀Å↪ԀAabcefkosy⇄⇇⇱⇵⇺∘∝∯≨≽ròΠȀilmr⇐⇔⇗⇛rsðᒄf»․ilôکĀdr⇠⇤cy;䑊ƀ;cwࣴ⇫⇯ir;楈;憭ar;意irc;䄥ƀalr∁∎∓rtsĀ;u∉∊晥it»∊lip;怦con;抹r;쀀𝔥sĀew∣∩arow;椥arow;椦ʀamopr∺∾≃≞≣rr;懿tht;戻kĀlr≉≓eftarrow;憩ightarrow;憪f;쀀𝕙bar;怕ƀclt≯≴≸r;쀀𝒽asè⇴rok;䄧Ābp⊂⊇ull;恃hen»ᱛૡ⊣\0⊪\0⊸⋅⋎\0⋕⋳\0\0⋸⌢⍧⍢⍿\0⎆⎪⎴cute耻í䃭ƀ;iyݱ⊰⊵rc耻î䃮;䐸Ācx⊼⊿y;䐵cl耻¡䂡ĀfrΟ⋉;쀀𝔦rave耻ì䃬Ȁ;inoܾ⋝⋩⋮Āin⋢⋦nt;樌t;戭fin;槜ta;愩lig;䄳ƀaop⋾⌚⌝ƀcgt⌅⌈⌗r;䄫ƀelpܟ⌏⌓inåގarôܠh;䄱f;抷ed;䆵ʀ;cfotӴ⌬⌱⌽⍁are;愅inĀ;t⌸⌹戞ie;槝doô⌙ʀ;celpݗ⍌⍐⍛⍡al;抺Āgr⍕⍙eróᕣã⍍arhk;樗rod;樼Ȁcgpt⍯⍲⍶⍻y;䑑on;䄯f;쀀𝕚a;䎹uest耻¿䂿Āci⎊⎏r;쀀𝒾nʀ;EdsvӴ⎛⎝⎡ӳ;拹ot;拵Ā;v⎦⎧拴;拳Ā;iݷ⎮lde;䄩ǫ⎸\0⎼cy;䑖l耻ï䃯̀cfmosu⏌⏗⏜⏡⏧⏵Āiy⏑⏕rc;䄵;䐹r;쀀𝔧ath;䈷pf;쀀𝕛ǣ⏬\0⏱r;쀀𝒿rcy;䑘kcy;䑔Ѐacfghjos␋␖␢␧␭␱␵␻ppaĀ;v␓␔䎺;䏰Āey␛␠dil;䄷;䐺r;쀀𝔨reen;䄸cy;䑅cy;䑜pf;쀀𝕜cr;쀀𝓀஀ABEHabcdefghjlmnoprstuv⑰⒁⒆⒍⒑┎┽╚▀♎♞♥♹♽⚚⚲⛘❝❨➋⟀⠁⠒ƀart⑷⑺⑼rò৆òΕail;椛arr;椎Ā;gঔ⒋;檋ar;楢ॣ⒥\0⒪\0⒱\0\0\0\0\0⒵Ⓔ\0ⓆⓈⓍ\0⓹ute;䄺mptyv;榴raîࡌbda;䎻gƀ;dlࢎⓁⓃ;榑åࢎ;檅uo耻«䂫rЀ;bfhlpst࢙ⓞⓦⓩ⓫⓮⓱⓵Ā;f࢝ⓣs;椟s;椝ë≒p;憫l;椹im;楳l;憢ƀ;ae⓿─┄檫il;椙Ā;s┉┊檭;쀀⪭︀ƀabr┕┙┝rr;椌rk;杲Āak┢┬cĀek┨┪;䁻;䁛Āes┱┳;榋lĀdu┹┻;榏;榍Ȁaeuy╆╋╖╘ron;䄾Ādi═╔il;䄼ìࢰâ┩;䐻Ȁcqrs╣╦╭╽a;椶uoĀ;rนᝆĀdu╲╷har;楧shar;楋h;憲ʀ;fgqs▋▌উ◳◿扤tʀahlrt▘▤▷◂◨rrowĀ;t࢙□aé⓶arpoonĀdu▯▴own»њp»०eftarrows;懇ightƀahs◍◖◞rrowĀ;sࣴࢧarpoonó྘quigarro÷⇰hreetimes;拋ƀ;qs▋ও◺lanôবʀ;cdgsব☊☍☝☨c;檨otĀ;o☔☕橿Ā;r☚☛檁;檃Ā;e☢☥쀀⋚︀s;檓ʀadegs☳☹☽♉♋pproøⓆot;拖qĀgq♃♅ôউgtò⒌ôছiíলƀilr♕࣡♚sht;楼;쀀𝔩Ā;Eজ♣;檑š♩♶rĀdu▲♮Ā;l॥♳;楪lk;斄cy;䑙ʀ;achtੈ⚈⚋⚑⚖rò◁orneòᴈard;楫ri;旺Āio⚟⚤dot;䅀ustĀ;a⚬⚭掰che»⚭ȀEaes⚻⚽⛉⛔;扨pĀ;p⛃⛄檉rox»⛄Ā;q⛎⛏檇Ā;q⛎⚻im;拦Ѐabnoptwz⛩⛴⛷✚✯❁❇❐Ānr⛮⛱g;柬r;懽rëࣁgƀlmr⛿✍✔eftĀar০✇ightá৲apsto;柼ightá৽parrowĀlr✥✩efô⓭ight;憬ƀafl✶✹✽r;榅;쀀𝕝us;樭imes;樴š❋❏st;戗áፎƀ;ef❗❘᠀旊nge»❘arĀ;l❤❥䀨t;榓ʀachmt❳❶❼➅➇ròࢨorneòᶌarĀ;d྘➃;業;怎ri;抿̀achiqt➘➝ੀ➢➮➻quo;怹r;쀀𝓁mƀ;egল➪➬;檍;檏Ābu┪➳oĀ;rฟ➹;怚rok;䅂萀<;cdhilqrࠫ⟒☹⟜⟠⟥⟪⟰Āci⟗⟙;檦r;橹reå◲mes;拉arr;楶uest;橻ĀPi⟵⟹ar;榖ƀ;ef⠀भ᠛旃rĀdu⠇⠍shar;楊har;楦Āen⠗⠡rtneqq;쀀≨︀Å⠞܀Dacdefhilnopsu⡀⡅⢂⢎⢓⢠⢥⢨⣚⣢⣤ઃ⣳⤂Dot;戺Ȁclpr⡎⡒⡣⡽r耻¯䂯Āet⡗⡙;時Ā;e⡞⡟朠se»⡟Ā;sျ⡨toȀ;dluျ⡳⡷⡻owîҌefôएðᏑker;斮Āoy⢇⢌mma;権;䐼ash;怔asuredangle»ᘦr;쀀𝔪o;愧ƀcdn⢯⢴⣉ro耻µ䂵Ȁ;acdᑤ⢽⣀⣄sôᚧir;櫰ot肻·Ƶusƀ;bd⣒ᤃ⣓戒Ā;uᴼ⣘;横ţ⣞⣡p;櫛ò−ðઁĀdp⣩⣮els;抧f;쀀𝕞Āct⣸⣽r;쀀𝓂pos»ᖝƀ;lm⤉⤊⤍䎼timap;抸ఀGLRVabcdefghijlmoprstuvw⥂⥓⥾⦉⦘⧚⧩⨕⨚⩘⩝⪃⪕⪤⪨⬄⬇⭄⭿⮮ⰴⱧⱼ⳩Āgt⥇⥋;쀀⋙̸Ā;v⥐௏쀀≫⃒ƀelt⥚⥲⥶ftĀar⥡⥧rrow;懍ightarrow;懎;쀀⋘̸Ā;v⥻ే쀀≪⃒ightarrow;懏ĀDd⦎⦓ash;抯ash;抮ʀbcnpt⦣⦧⦬⦱⧌la»˞ute;䅄g;쀀∠⃒ʀ;Eiop඄⦼⧀⧅⧈;쀀⩰̸d;쀀≋̸s;䅉roø඄urĀ;a⧓⧔普lĀ;s⧓ସǳ⧟\0⧣p肻 ଷmpĀ;e௹ఀʀaeouy⧴⧾⨃⨐⨓ǰ⧹\0⧻;橃on;䅈dil;䅆ngĀ;dൾ⨊ot;쀀⩭̸p;橂;䐽ash;怓΀;Aadqsxஒ⨩⨭⨻⩁⩅⩐rr;懗rĀhr⨳⨶k;椤Ā;oᏲᏰot;쀀≐̸uiöୣĀei⩊⩎ar;椨í஘istĀ;s஠டr;쀀𝔫ȀEest௅⩦⩹⩼ƀ;qs஼⩭௡ƀ;qs஼௅⩴lanô௢ií௪Ā;rஶ⪁»ஷƀAap⪊⪍⪑rò⥱rr;憮ar;櫲ƀ;svྍ⪜ྌĀ;d⪡⪢拼;拺cy;䑚΀AEadest⪷⪺⪾⫂⫅⫶⫹rò⥦;쀀≦̸rr;憚r;急Ȁ;fqs఻⫎⫣⫯tĀar⫔⫙rro÷⫁ightarro÷⪐ƀ;qs఻⪺⫪lanôౕĀ;sౕ⫴»శiíౝĀ;rవ⫾iĀ;eచథiäඐĀpt⬌⬑f;쀀𝕟膀¬;in⬙⬚⬶䂬nȀ;Edvஉ⬤⬨⬮;쀀⋹̸ot;쀀⋵̸ǡஉ⬳⬵;拷;拶iĀ;vಸ⬼ǡಸ⭁⭃;拾;拽ƀaor⭋⭣⭩rȀ;ast୻⭕⭚⭟lleì୻l;쀀⫽⃥;쀀∂̸lint;樔ƀ;ceಒ⭰⭳uåಥĀ;cಘ⭸Ā;eಒ⭽ñಘȀAait⮈⮋⮝⮧rò⦈rrƀ;cw⮔⮕⮙憛;쀀⤳̸;쀀↝̸ghtarrow»⮕riĀ;eೋೖ΀chimpqu⮽⯍⯙⬄୸⯤⯯Ȁ;cerല⯆ഷ⯉uå൅;쀀𝓃ortɭ⬅\0\0⯖ará⭖mĀ;e൮⯟Ā;q൴൳suĀbp⯫⯭å೸åഋƀbcp⯶ⰑⰙȀ;Ees⯿ⰀഢⰄ抄;쀀⫅̸etĀ;eഛⰋqĀ;qണⰀcĀ;eലⰗñസȀ;EesⰢⰣൟⰧ抅;쀀⫆̸etĀ;e൘ⰮqĀ;qൠⰣȀgilrⰽⰿⱅⱇìௗlde耻ñ䃱çృiangleĀlrⱒⱜeftĀ;eచⱚñదightĀ;eೋⱥñ೗Ā;mⱬⱭ䎽ƀ;esⱴⱵⱹ䀣ro;愖p;怇ҀDHadgilrsⲏⲔⲙⲞⲣⲰⲶⳓⳣash;抭arr;椄p;쀀≍⃒ash;抬ĀetⲨⲬ;쀀≥⃒;쀀>⃒nfin;槞ƀAetⲽⳁⳅrr;椂;쀀≤⃒Ā;rⳊⳍ쀀<⃒ie;쀀⊴⃒ĀAtⳘⳜrr;椃rie;쀀⊵⃒im;쀀∼⃒ƀAan⳰⳴ⴂrr;懖rĀhr⳺⳽k;椣Ā;oᏧᏥear;椧ቓ᪕\0\0\0\0\0\0\0\0\0\0\0\0\0ⴭ\0ⴸⵈⵠⵥ⵲ⶄᬇ\0\0ⶍⶫ\0ⷈⷎ\0ⷜ⸙⸫⸾⹃Ācsⴱ᪗ute耻ó䃳ĀiyⴼⵅrĀ;c᪞ⵂ耻ô䃴;䐾ʀabios᪠ⵒⵗǈⵚlac;䅑v;樸old;榼lig;䅓Ācr⵩⵭ir;榿;쀀𝔬ͯ⵹\0\0⵼\0ⶂn;䋛ave耻ò䃲;槁Ābmⶈ෴ar;榵Ȁacitⶕ⶘ⶥⶨrò᪀Āir⶝ⶠr;榾oss;榻nå๒;槀ƀaeiⶱⶵⶹcr;䅍ga;䏉ƀcdnⷀⷅǍron;䎿;榶pf;쀀𝕠ƀaelⷔ⷗ǒr;榷rp;榹΀;adiosvⷪⷫⷮ⸈⸍⸐⸖戨rò᪆Ȁ;efmⷷⷸ⸂⸅橝rĀ;oⷾⷿ愴f»ⷿ耻ª䂪耻º䂺gof;抶r;橖lope;橗;橛ƀclo⸟⸡⸧ò⸁ash耻ø䃸l;折iŬⸯ⸴de耻õ䃵esĀ;aǛ⸺s;樶ml耻ö䃶bar;挽ૡ⹞\0⹽\0⺀⺝\0⺢⺹\0\0⻋ຜ\0⼓\0\0⼫⾼\0⿈rȀ;astЃ⹧⹲຅脀¶;l⹭⹮䂶leìЃɩ⹸\0\0⹻m;櫳;櫽y;䐿rʀcimpt⺋⺏⺓ᡥ⺗nt;䀥od;䀮il;怰enk;怱r;쀀𝔭ƀimo⺨⺰⺴Ā;v⺭⺮䏆;䏕maô੶ne;明ƀ;tv⺿⻀⻈䏀chfork»´;䏖Āau⻏⻟nĀck⻕⻝kĀ;h⇴⻛;愎ö⇴sҀ;abcdemst⻳⻴ᤈ⻹⻽⼄⼆⼊⼎䀫cir;樣ir;樢Āouᵀ⼂;樥;橲n肻±ຝim;樦wo;樧ƀipu⼙⼠⼥ntint;樕f;쀀𝕡nd耻£䂣Ԁ;Eaceinosu່⼿⽁⽄⽇⾁⾉⾒⽾⾶;檳p;檷uå໙Ā;c໎⽌̀;acens່⽙⽟⽦⽨⽾pproø⽃urlyeñ໙ñ໎ƀaes⽯⽶⽺pprox;檹qq;檵im;拨iíໟmeĀ;s⾈ຮ怲ƀEas⽸⾐⽺ð⽵ƀdfp໬⾙⾯ƀals⾠⾥⾪lar;挮ine;挒urf;挓Ā;t໻⾴ï໻rel;抰Āci⿀⿅r;쀀𝓅;䏈ncsp;怈̀fiopsu⿚⋢⿟⿥⿫⿱r;쀀𝔮pf;쀀𝕢rime;恗cr;쀀𝓆ƀaeo⿸〉〓tĀei⿾々rnionóڰnt;樖stĀ;e【】䀿ñἙô༔઀ABHabcdefhilmnoprstux぀けさすムㄎㄫㅇㅢㅲㆎ㈆㈕㈤㈩㉘㉮㉲㊐㊰㊷ƀartぇおがròႳòϝail;検aròᱥar;楤΀cdenqrtとふへみわゔヌĀeuねぱ;쀀∽̱te;䅕iãᅮmptyv;榳gȀ;del࿑らるろ;榒;榥å࿑uo耻»䂻rր;abcfhlpstw࿜ガクシスゼゾダッデナp;極Ā;f࿠ゴs;椠;椳s;椞ë≝ð✮l;楅im;楴l;憣;憝Āaiパフil;椚oĀ;nホボ戶aló༞ƀabrョリヮrò៥rk;杳ĀakンヽcĀekヹ・;䁽;䁝Āes㄂㄄;榌lĀduㄊㄌ;榎;榐Ȁaeuyㄗㄜㄧㄩron;䅙Ādiㄡㄥil;䅗ì࿲âヺ;䑀Ȁclqsㄴㄷㄽㅄa;椷dhar;楩uoĀ;rȎȍh;憳ƀacgㅎㅟངlȀ;ipsླྀㅘㅛႜnåႻarôྩt;断ƀilrㅩဣㅮsht;楽;쀀𝔯ĀaoㅷㆆrĀduㅽㅿ»ѻĀ;l႑ㆄ;楬Ā;vㆋㆌ䏁;䏱ƀgns㆕ㇹㇼht̀ahlrstㆤㆰ㇂㇘㇤㇮rrowĀ;t࿜ㆭaéトarpoonĀduㆻㆿowîㅾp»႒eftĀah㇊㇐rrowó࿪arpoonóՑightarrows;應quigarro÷ニhreetimes;拌g;䋚ingdotseñἲƀahm㈍㈐㈓rò࿪aòՑ;怏oustĀ;a㈞㈟掱che»㈟mid;櫮Ȁabpt㈲㈽㉀㉒Ānr㈷㈺g;柭r;懾rëဃƀafl㉇㉊㉎r;榆;쀀𝕣us;樮imes;樵Āap㉝㉧rĀ;g㉣㉤䀩t;榔olint;樒arò㇣Ȁachq㉻㊀Ⴜ㊅quo;怺r;쀀𝓇Ābu・㊊oĀ;rȔȓƀhir㊗㊛㊠reåㇸmes;拊iȀ;efl㊪ၙᠡ㊫方tri;槎luhar;楨;愞ൡ㋕㋛㋟㌬㌸㍱\0㍺㎤\0\0㏬㏰\0㐨㑈㑚㒭㒱㓊㓱\0㘖\0\0㘳cute;䅛quï➺Ԁ;Eaceinpsyᇭ㋳㋵㋿㌂㌋㌏㌟㌦㌩;檴ǰ㋺\0㋼;檸on;䅡uåᇾĀ;dᇳ㌇il;䅟rc;䅝ƀEas㌖㌘㌛;檶p;檺im;择olint;樓iíሄ;䑁otƀ;be㌴ᵇ㌵担;橦΀Aacmstx㍆㍊㍗㍛㍞㍣㍭rr;懘rĀhr㍐㍒ë∨Ā;oਸ਼਴t耻§䂧i;䀻war;椩mĀin㍩ðnuóñt;朶rĀ;o㍶⁕쀀𝔰Ȁacoy㎂㎆㎑㎠rp;景Āhy㎋㎏cy;䑉;䑈rtɭ㎙\0\0㎜iäᑤaraì⹯耻­䂭Āgm㎨㎴maƀ;fv㎱㎲㎲䏃;䏂Ѐ;deglnprካ㏅㏉㏎㏖㏞㏡㏦ot;橪Ā;q኱ኰĀ;E㏓㏔檞;檠Ā;E㏛㏜檝;檟e;扆lus;樤arr;楲aròᄽȀaeit㏸㐈㐏㐗Āls㏽㐄lsetmé㍪hp;樳parsl;槤Ādlᑣ㐔e;挣Ā;e㐜㐝檪Ā;s㐢㐣檬;쀀⪬︀ƀflp㐮㐳㑂tcy;䑌Ā;b㐸㐹䀯Ā;a㐾㐿槄r;挿f;쀀𝕤aĀdr㑍ЂesĀ;u㑔㑕晠it»㑕ƀcsu㑠㑹㒟Āau㑥㑯pĀ;sᆈ㑫;쀀⊓︀pĀ;sᆴ㑵;쀀⊔︀uĀbp㑿㒏ƀ;esᆗᆜ㒆etĀ;eᆗ㒍ñᆝƀ;esᆨᆭ㒖etĀ;eᆨ㒝ñᆮƀ;afᅻ㒦ְrť㒫ֱ»ᅼaròᅈȀcemt㒹㒾㓂㓅r;쀀𝓈tmîñiì㐕aræᆾĀar㓎㓕rĀ;f㓔ឿ昆Āan㓚㓭ightĀep㓣㓪psiloîỠhé⺯s»⡒ʀbcmnp㓻㕞ሉ㖋㖎Ҁ;Edemnprs㔎㔏㔑㔕㔞㔣㔬㔱㔶抂;櫅ot;檽Ā;dᇚ㔚ot;櫃ult;櫁ĀEe㔨㔪;櫋;把lus;檿arr;楹ƀeiu㔽㕒㕕tƀ;en㔎㕅㕋qĀ;qᇚ㔏eqĀ;q㔫㔨m;櫇Ābp㕚㕜;櫕;櫓c̀;acensᇭ㕬㕲㕹㕻㌦pproø㋺urlyeñᇾñᇳƀaes㖂㖈㌛pproø㌚qñ㌗g;晪ڀ123;Edehlmnps㖩㖬㖯ሜ㖲㖴㗀㗉㗕㗚㗟㗨㗭耻¹䂹耻²䂲耻³䂳;櫆Āos㖹㖼t;檾ub;櫘Ā;dሢ㗅ot;櫄sĀou㗏㗒l;柉b;櫗arr;楻ult;櫂ĀEe㗤㗦;櫌;抋lus;櫀ƀeiu㗴㘉㘌tƀ;enሜ㗼㘂qĀ;qሢ㖲eqĀ;q㗧㗤m;櫈Ābp㘑㘓;櫔;櫖ƀAan㘜㘠㘭rr;懙rĀhr㘦㘨ë∮Ā;oਫ਩war;椪lig耻ß䃟௡㙑㙝㙠ዎ㙳㙹\0㙾㛂\0\0\0\0\0㛛㜃\0㜉㝬\0\0\0㞇ɲ㙖\0\0㙛get;挖;䏄rë๟ƀaey㙦㙫㙰ron;䅥dil;䅣;䑂lrec;挕r;쀀𝔱Ȁeiko㚆㚝㚵㚼ǲ㚋\0㚑eĀ4fኄኁaƀ;sv㚘㚙㚛䎸ym;䏑Ācn㚢㚲kĀas㚨㚮pproø዁im»ኬsðኞĀas㚺㚮ð዁rn耻þ䃾Ǭ̟㛆⋧es膀×;bd㛏㛐㛘䃗Ā;aᤏ㛕r;樱;樰ƀeps㛡㛣㜀á⩍Ȁ;bcf҆㛬㛰㛴ot;挶ir;櫱Ā;o㛹㛼쀀𝕥rk;櫚á㍢rime;怴ƀaip㜏㜒㝤dåቈ΀adempst㜡㝍㝀㝑㝗㝜㝟ngleʀ;dlqr㜰㜱㜶㝀㝂斵own»ᶻeftĀ;e⠀㜾ñम;扜ightĀ;e㊪㝋ñၚot;旬inus;樺lus;樹b;槍ime;樻ezium;揢ƀcht㝲㝽㞁Āry㝷㝻;쀀𝓉;䑆cy;䑛rok;䅧Āio㞋㞎xô᝷headĀlr㞗㞠eftarro÷ࡏightarrow»ཝऀAHabcdfghlmoprstuw㟐㟓㟗㟤㟰㟼㠎㠜㠣㠴㡑㡝㡫㢩㣌㣒㣪㣶ròϭar;楣Ācr㟜㟢ute耻ú䃺òᅐrǣ㟪\0㟭y;䑞ve;䅭Āiy㟵㟺rc耻û䃻;䑃ƀabh㠃㠆㠋ròᎭlac;䅱aòᏃĀir㠓㠘sht;楾;쀀𝔲rave耻ù䃹š㠧㠱rĀlr㠬㠮»ॗ»ႃlk;斀Āct㠹㡍ɯ㠿\0\0㡊rnĀ;e㡅㡆挜r»㡆op;挏ri;旸Āal㡖㡚cr;䅫肻¨͉Āgp㡢㡦on;䅳f;쀀𝕦̀adhlsuᅋ㡸㡽፲㢑㢠ownáᎳarpoonĀlr㢈㢌efô㠭ighô㠯iƀ;hl㢙㢚㢜䏅»ᏺon»㢚parrows;懈ƀcit㢰㣄㣈ɯ㢶\0\0㣁rnĀ;e㢼㢽挝r»㢽op;挎ng;䅯ri;旹cr;쀀𝓊ƀdir㣙㣝㣢ot;拰lde;䅩iĀ;f㜰㣨»᠓Āam㣯㣲rò㢨l耻ü䃼angle;榧ހABDacdeflnoprsz㤜㤟㤩㤭㦵㦸㦽㧟㧤㧨㧳㧹㧽㨁㨠ròϷarĀ;v㤦㤧櫨;櫩asèϡĀnr㤲㤷grt;榜΀eknprst㓣㥆㥋㥒㥝㥤㦖appá␕othinçẖƀhir㓫⻈㥙opô⾵Ā;hᎷ㥢ïㆍĀiu㥩㥭gmá㎳Ābp㥲㦄setneqĀ;q㥽㦀쀀⊊︀;쀀⫋︀setneqĀ;q㦏㦒쀀⊋︀;쀀⫌︀Āhr㦛㦟etá㚜iangleĀlr㦪㦯eft»थight»ၑy;䐲ash»ံƀelr㧄㧒㧗ƀ;beⷪ㧋㧏ar;抻q;扚lip;拮Ābt㧜ᑨaòᑩr;쀀𝔳tré㦮suĀbp㧯㧱»ജ»൙pf;쀀𝕧roð໻tré㦴Ācu㨆㨋r;쀀𝓋Ābp㨐㨘nĀEe㦀㨖»㥾nĀEe㦒㨞»㦐igzag;榚΀cefoprs㨶㨻㩖㩛㩔㩡㩪irc;䅵Ādi㩀㩑Ābg㩅㩉ar;機eĀ;qᗺ㩏;扙erp;愘r;쀀𝔴pf;쀀𝕨Ā;eᑹ㩦atèᑹcr;쀀𝓌ૣណ㪇\0㪋\0㪐㪛\0\0㪝㪨㪫㪯\0\0㫃㫎\0㫘ៜ៟tré៑r;쀀𝔵ĀAa㪔㪗ròσrò৶;䎾ĀAa㪡㪤ròθrò৫að✓is;拻ƀdptឤ㪵㪾Āfl㪺ឩ;쀀𝕩imåឲĀAa㫇㫊ròώròਁĀcq㫒ីr;쀀𝓍Āpt៖㫜ré។Ѐacefiosu㫰㫽㬈㬌㬑㬕㬛㬡cĀuy㫶㫻te耻ý䃽;䑏Āiy㬂㬆rc;䅷;䑋n耻¥䂥r;쀀𝔶cy;䑗pf;쀀𝕪cr;쀀𝓎Ācm㬦㬩y;䑎l耻ÿ䃿Ԁacdefhiosw㭂㭈㭔㭘㭤㭩㭭㭴㭺㮀cute;䅺Āay㭍㭒ron;䅾;䐷ot;䅼Āet㭝㭡træᕟa;䎶r;쀀𝔷cy;䐶grarr;懝pf;쀀𝕫cr;쀀𝓏Ājn㮅㮇;怍j;怌'.split("").map(n=>n.charCodeAt(0))),UXr=new Map([[0,65533],[128,8364],[130,8218],[131,402],[132,8222],[133,8230],[134,8224],[135,8225],[136,710],[137,8240],[138,352],[139,8249],[140,338],[142,381],[145,8216],[146,8217],[147,8220],[148,8221],[149,8226],[150,8211],[151,8212],[152,732],[153,8482],[154,353],[155,8250],[156,339],[158,382],[159,376]]);function zXr(n){var r;return n>=55296&&n<=57343||n>1114111?65533:(r=UXr.get(n))!==null&&r!==void 0?r:n}var Hm;(function(n){n[n.NUM=35]="NUM",n[n.SEMI=59]="SEMI",n[n.EQUALS=61]="EQUALS",n[n.ZERO=48]="ZERO",n[n.NINE=57]="NINE",n[n.LOWER_A=97]="LOWER_A",n[n.LOWER_F=102]="LOWER_F",n[n.LOWER_X=120]="LOWER_X",n[n.LOWER_Z=122]="LOWER_Z",n[n.UPPER_A=65]="UPPER_A",n[n.UPPER_F=70]="UPPER_F",n[n.UPPER_Z=90]="UPPER_Z"})(Hm||(Hm={}));const HXr=32;var fM;(function(n){n[n.VALUE_LENGTH=49152]="VALUE_LENGTH",n[n.BRANCH_LENGTH=16256]="BRANCH_LENGTH",n[n.JUMP_TABLE=127]="JUMP_TABLE"})(fM||(fM={}));function yKe(n){return n>=Hm.ZERO&&n<=Hm.NINE}function GXr(n){return n>=Hm.UPPER_A&&n<=Hm.UPPER_F||n>=Hm.LOWER_A&&n<=Hm.LOWER_F}function qXr(n){return n>=Hm.UPPER_A&&n<=Hm.UPPER_Z||n>=Hm.LOWER_A&&n<=Hm.LOWER_Z||yKe(n)}function VXr(n){return n===Hm.EQUALS||qXr(n)}var Fm;(function(n){n[n.EntityStart=0]="EntityStart",n[n.NumericStart=1]="NumericStart",n[n.NumericDecimal=2]="NumericDecimal",n[n.NumericHex=3]="NumericHex",n[n.NamedEntity=4]="NamedEntity"})(Fm||(Fm={}));var R4;(function(n){n[n.Legacy=0]="Legacy",n[n.Strict=1]="Strict",n[n.Attribute=2]="Attribute"})(R4||(R4={}));class WXr{constructor(r,o,s){this.decodeTree=r,this.emitCodePoint=o,this.errors=s,this.state=Fm.EntityStart,this.consumed=1,this.result=0,this.treeIndex=0,this.excess=1,this.decodeMode=R4.Strict}startEntity(r){this.decodeMode=r,this.state=Fm.EntityStart,this.result=0,this.treeIndex=0,this.excess=1,this.consumed=1}write(r,o){switch(this.state){case Fm.EntityStart:return r.charCodeAt(o)===Hm.NUM?(this.state=Fm.NumericStart,this.consumed+=1,this.stateNumericStart(r,o+1)):(this.state=Fm.NamedEntity,this.stateNamedEntity(r,o));case Fm.NumericStart:return this.stateNumericStart(r,o);case Fm.NumericDecimal:return this.stateNumericDecimal(r,o);case Fm.NumericHex:return this.stateNumericHex(r,o);case Fm.NamedEntity:return this.stateNamedEntity(r,o)}}stateNumericStart(r,o){return o>=r.length?-1:(r.charCodeAt(o)|HXr)===Hm.LOWER_X?(this.state=Fm.NumericHex,this.consumed+=1,this.stateNumericHex(r,o+1)):(this.state=Fm.NumericDecimal,this.stateNumericDecimal(r,o))}addToNumericResult(r,o,s,c){if(o!==s){const d=s-o;this.result=this.result*Math.pow(c,d)+Number.parseInt(r.substr(o,d),c),this.consumed+=d}}stateNumericHex(r,o){const s=o;for(;o<r.length;){const c=r.charCodeAt(o);if(yKe(c)||GXr(c))o+=1;else return this.addToNumericResult(r,s,o,16),this.emitNumericEntity(c,3)}return this.addToNumericResult(r,s,o,16),-1}stateNumericDecimal(r,o){const s=o;for(;o<r.length;){const c=r.charCodeAt(o);if(yKe(c))o+=1;else return this.addToNumericResult(r,s,o,10),this.emitNumericEntity(c,2)}return this.addToNumericResult(r,s,o,10),-1}emitNumericEntity(r,o){var s;if(this.consumed<=o)return(s=this.errors)===null||s===void 0||s.absenceOfDigitsInNumericCharacterReference(this.consumed),0;if(r===Hm.SEMI)this.consumed+=1;else if(this.decodeMode===R4.Strict)return 0;return this.emitCodePoint(zXr(this.result),this.consumed),this.errors&&(r!==Hm.SEMI&&this.errors.missingSemicolonAfterCharacterReference(),this.errors.validateNumericCharacterReference(this.result)),this.consumed}stateNamedEntity(r,o){const{decodeTree:s}=this;let c=s[this.treeIndex],d=(c&fM.VALUE_LENGTH)>>14;for(;o<r.length;o++,this.excess++){const h=r.charCodeAt(o);if(this.treeIndex=KXr(s,c,this.treeIndex+Math.max(1,d),h),this.treeIndex<0)return this.result===0||this.decodeMode===R4.Attribute&&(d===0||VXr(h))?0:this.emitNotTerminatedNamedEntity();if(c=s[this.treeIndex],d=(c&fM.VALUE_LENGTH)>>14,d!==0){if(h===Hm.SEMI)return this.emitNamedEntityData(this.treeIndex,d,this.consumed+this.excess);this.decodeMode!==R4.Strict&&(this.result=this.treeIndex,this.consumed+=this.excess,this.excess=0)}}return-1}emitNotTerminatedNamedEntity(){var r;const{result:o,decodeTree:s}=this,c=(s[o]&fM.VALUE_LENGTH)>>14;return this.emitNamedEntityData(o,c,this.consumed),(r=this.errors)===null||r===void 0||r.missingSemicolonAfterCharacterReference(),this.consumed}emitNamedEntityData(r,o,s){const{decodeTree:c}=this;return this.emitCodePoint(o===1?c[r]&~fM.VALUE_LENGTH:c[r+1],s),o===3&&this.emitCodePoint(c[r+2],s),s}end(){var r;switch(this.state){case Fm.NamedEntity:return this.result!==0&&(this.decodeMode!==R4.Attribute||this.result===this.treeIndex)?this.emitNotTerminatedNamedEntity():0;case Fm.NumericDecimal:return this.emitNumericEntity(0,2);case Fm.NumericHex:return this.emitNumericEntity(0,3);case Fm.NumericStart:return(r=this.errors)===null||r===void 0||r.absenceOfDigitsInNumericCharacterReference(this.consumed),0;case Fm.EntityStart:return 0}}}function KXr(n,r,o,s){const c=(r&fM.BRANCH_LENGTH)>>7,d=r&fM.JUMP_TABLE;if(c===0)return d!==0&&s===d?o:-1;if(d){const y=s-d;return y<0||y>=c?-1:n[o+y]-1}let h=o,b=h+c-1;for(;h<=b;){const y=h+b>>>1,E=n[y];if(E<s)h=y+1;else if(E>s)b=y-1;else return n[y+c]}return-1}var Er;(function(n){n.HTML="http://www.w3.org/1999/xhtml",n.MATHML="http://www.w3.org/1998/Math/MathML",n.SVG="http://www.w3.org/2000/svg",n.XLINK="http://www.w3.org/1999/xlink",n.XML="http://www.w3.org/XML/1998/namespace",n.XMLNS="http://www.w3.org/2000/xmlns/"})(Er||(Er={}));var GB;(function(n){n.TYPE="type",n.ACTION="action",n.ENCODING="encoding",n.PROMPT="prompt",n.NAME="name",n.COLOR="color",n.FACE="face",n.SIZE="size"})(GB||(GB={}));var D_;(function(n){n.NO_QUIRKS="no-quirks",n.QUIRKS="quirks",n.LIMITED_QUIRKS="limited-quirks"})(D_||(D_={}));var nn;(function(n){n.A="a",n.ADDRESS="address",n.ANNOTATION_XML="annotation-xml",n.APPLET="applet",n.AREA="area",n.ARTICLE="article",n.ASIDE="aside",n.B="b",n.BASE="base",n.BASEFONT="basefont",n.BGSOUND="bgsound",n.BIG="big",n.BLOCKQUOTE="blockquote",n.BODY="body",n.BR="br",n.BUTTON="button",n.CAPTION="caption",n.CENTER="center",n.CODE="code",n.COL="col",n.COLGROUP="colgroup",n.DD="dd",n.DESC="desc",n.DETAILS="details",n.DIALOG="dialog",n.DIR="dir",n.DIV="div",n.DL="dl",n.DT="dt",n.EM="em",n.EMBED="embed",n.FIELDSET="fieldset",n.FIGCAPTION="figcaption",n.FIGURE="figure",n.FONT="font",n.FOOTER="footer",n.FOREIGN_OBJECT="foreignObject",n.FORM="form",n.FRAME="frame",n.FRAMESET="frameset",n.H1="h1",n.H2="h2",n.H3="h3",n.H4="h4",n.H5="h5",n.H6="h6",n.HEAD="head",n.HEADER="header",n.HGROUP="hgroup",n.HR="hr",n.HTML="html",n.I="i",n.IMG="img",n.IMAGE="image",n.INPUT="input",n.IFRAME="iframe",n.KEYGEN="keygen",n.LABEL="label",n.LI="li",n.LINK="link",n.LISTING="listing",n.MAIN="main",n.MALIGNMARK="malignmark",n.MARQUEE="marquee",n.MATH="math",n.MENU="menu",n.META="meta",n.MGLYPH="mglyph",n.MI="mi",n.MO="mo",n.MN="mn",n.MS="ms",n.MTEXT="mtext",n.NAV="nav",n.NOBR="nobr",n.NOFRAMES="noframes",n.NOEMBED="noembed",n.NOSCRIPT="noscript",n.OBJECT="object",n.OL="ol",n.OPTGROUP="optgroup",n.OPTION="option",n.P="p",n.PARAM="param",n.PLAINTEXT="plaintext",n.PRE="pre",n.RB="rb",n.RP="rp",n.RT="rt",n.RTC="rtc",n.RUBY="ruby",n.S="s",n.SCRIPT="script",n.SEARCH="search",n.SECTION="section",n.SELECT="select",n.SOURCE="source",n.SMALL="small",n.SPAN="span",n.STRIKE="strike",n.STRONG="strong",n.STYLE="style",n.SUB="sub",n.SUMMARY="summary",n.SUP="sup",n.TABLE="table",n.TBODY="tbody",n.TEMPLATE="template",n.TEXTAREA="textarea",n.TFOOT="tfoot",n.TD="td",n.TH="th",n.THEAD="thead",n.TITLE="title",n.TR="tr",n.TRACK="track",n.TT="tt",n.U="u",n.UL="ul",n.SVG="svg",n.VAR="var",n.WBR="wbr",n.XMP="xmp"})(nn||(nn={}));var re;(function(n){n[n.UNKNOWN=0]="UNKNOWN",n[n.A=1]="A",n[n.ADDRESS=2]="ADDRESS",n[n.ANNOTATION_XML=3]="ANNOTATION_XML",n[n.APPLET=4]="APPLET",n[n.AREA=5]="AREA",n[n.ARTICLE=6]="ARTICLE",n[n.ASIDE=7]="ASIDE",n[n.B=8]="B",n[n.BASE=9]="BASE",n[n.BASEFONT=10]="BASEFONT",n[n.BGSOUND=11]="BGSOUND",n[n.BIG=12]="BIG",n[n.BLOCKQUOTE=13]="BLOCKQUOTE",n[n.BODY=14]="BODY",n[n.BR=15]="BR",n[n.BUTTON=16]="BUTTON",n[n.CAPTION=17]="CAPTION",n[n.CENTER=18]="CENTER",n[n.CODE=19]="CODE",n[n.COL=20]="COL",n[n.COLGROUP=21]="COLGROUP",n[n.DD=22]="DD",n[n.DESC=23]="DESC",n[n.DETAILS=24]="DETAILS",n[n.DIALOG=25]="DIALOG",n[n.DIR=26]="DIR",n[n.DIV=27]="DIV",n[n.DL=28]="DL",n[n.DT=29]="DT",n[n.EM=30]="EM",n[n.EMBED=31]="EMBED",n[n.FIELDSET=32]="FIELDSET",n[n.FIGCAPTION=33]="FIGCAPTION",n[n.FIGURE=34]="FIGURE",n[n.FONT=35]="FONT",n[n.FOOTER=36]="FOOTER",n[n.FOREIGN_OBJECT=37]="FOREIGN_OBJECT",n[n.FORM=38]="FORM",n[n.FRAME=39]="FRAME",n[n.FRAMESET=40]="FRAMESET",n[n.H1=41]="H1",n[n.H2=42]="H2",n[n.H3=43]="H3",n[n.H4=44]="H4",n[n.H5=45]="H5",n[n.H6=46]="H6",n[n.HEAD=47]="HEAD",n[n.HEADER=48]="HEADER",n[n.HGROUP=49]="HGROUP",n[n.HR=50]="HR",n[n.HTML=51]="HTML",n[n.I=52]="I",n[n.IMG=53]="IMG",n[n.IMAGE=54]="IMAGE",n[n.INPUT=55]="INPUT",n[n.IFRAME=56]="IFRAME",n[n.KEYGEN=57]="KEYGEN",n[n.LABEL=58]="LABEL",n[n.LI=59]="LI",n[n.LINK=60]="LINK",n[n.LISTING=61]="LISTING",n[n.MAIN=62]="MAIN",n[n.MALIGNMARK=63]="MALIGNMARK",n[n.MARQUEE=64]="MARQUEE",n[n.MATH=65]="MATH",n[n.MENU=66]="MENU",n[n.META=67]="META",n[n.MGLYPH=68]="MGLYPH",n[n.MI=69]="MI",n[n.MO=70]="MO",n[n.MN=71]="MN",n[n.MS=72]="MS",n[n.MTEXT=73]="MTEXT",n[n.NAV=74]="NAV",n[n.NOBR=75]="NOBR",n[n.NOFRAMES=76]="NOFRAMES",n[n.NOEMBED=77]="NOEMBED",n[n.NOSCRIPT=78]="NOSCRIPT",n[n.OBJECT=79]="OBJECT",n[n.OL=80]="OL",n[n.OPTGROUP=81]="OPTGROUP",n[n.OPTION=82]="OPTION",n[n.P=83]="P",n[n.PARAM=84]="PARAM",n[n.PLAINTEXT=85]="PLAINTEXT",n[n.PRE=86]="PRE",n[n.RB=87]="RB",n[n.RP=88]="RP",n[n.RT=89]="RT",n[n.RTC=90]="RTC",n[n.RUBY=91]="RUBY",n[n.S=92]="S",n[n.SCRIPT=93]="SCRIPT",n[n.SEARCH=94]="SEARCH",n[n.SECTION=95]="SECTION",n[n.SELECT=96]="SELECT",n[n.SOURCE=97]="SOURCE",n[n.SMALL=98]="SMALL",n[n.SPAN=99]="SPAN",n[n.STRIKE=100]="STRIKE",n[n.STRONG=101]="STRONG",n[n.STYLE=102]="STYLE",n[n.SUB=103]="SUB",n[n.SUMMARY=104]="SUMMARY",n[n.SUP=105]="SUP",n[n.TABLE=106]="TABLE",n[n.TBODY=107]="TBODY",n[n.TEMPLATE=108]="TEMPLATE",n[n.TEXTAREA=109]="TEXTAREA",n[n.TFOOT=110]="TFOOT",n[n.TD=111]="TD",n[n.TH=112]="TH",n[n.THEAD=113]="THEAD",n[n.TITLE=114]="TITLE",n[n.TR=115]="TR",n[n.TRACK=116]="TRACK",n[n.TT=117]="TT",n[n.U=118]="U",n[n.UL=119]="UL",n[n.SVG=120]="SVG",n[n.VAR=121]="VAR",n[n.WBR=122]="WBR",n[n.XMP=123]="XMP"})(re||(re={}));const YXr=new Map([[nn.A,re.A],[nn.ADDRESS,re.ADDRESS],[nn.ANNOTATION_XML,re.ANNOTATION_XML],[nn.APPLET,re.APPLET],[nn.AREA,re.AREA],[nn.ARTICLE,re.ARTICLE],[nn.ASIDE,re.ASIDE],[nn.B,re.B],[nn.BASE,re.BASE],[nn.BASEFONT,re.BASEFONT],[nn.BGSOUND,re.BGSOUND],[nn.BIG,re.BIG],[nn.BLOCKQUOTE,re.BLOCKQUOTE],[nn.BODY,re.BODY],[nn.BR,re.BR],[nn.BUTTON,re.BUTTON],[nn.CAPTION,re.CAPTION],[nn.CENTER,re.CENTER],[nn.CODE,re.CODE],[nn.COL,re.COL],[nn.COLGROUP,re.COLGROUP],[nn.DD,re.DD],[nn.DESC,re.DESC],[nn.DETAILS,re.DETAILS],[nn.DIALOG,re.DIALOG],[nn.DIR,re.DIR],[nn.DIV,re.DIV],[nn.DL,re.DL],[nn.DT,re.DT],[nn.EM,re.EM],[nn.EMBED,re.EMBED],[nn.FIELDSET,re.FIELDSET],[nn.FIGCAPTION,re.FIGCAPTION],[nn.FIGURE,re.FIGURE],[nn.FONT,re.FONT],[nn.FOOTER,re.FOOTER],[nn.FOREIGN_OBJECT,re.FOREIGN_OBJECT],[nn.FORM,re.FORM],[nn.FRAME,re.FRAME],[nn.FRAMESET,re.FRAMESET],[nn.H1,re.H1],[nn.H2,re.H2],[nn.H3,re.H3],[nn.H4,re.H4],[nn.H5,re.H5],[nn.H6,re.H6],[nn.HEAD,re.HEAD],[nn.HEADER,re.HEADER],[nn.HGROUP,re.HGROUP],[nn.HR,re.HR],[nn.HTML,re.HTML],[nn.I,re.I],[nn.IMG,re.IMG],[nn.IMAGE,re.IMAGE],[nn.INPUT,re.INPUT],[nn.IFRAME,re.IFRAME],[nn.KEYGEN,re.KEYGEN],[nn.LABEL,re.LABEL],[nn.LI,re.LI],[nn.LINK,re.LINK],[nn.LISTING,re.LISTING],[nn.MAIN,re.MAIN],[nn.MALIGNMARK,re.MALIGNMARK],[nn.MARQUEE,re.MARQUEE],[nn.MATH,re.MATH],[nn.MENU,re.MENU],[nn.META,re.META],[nn.MGLYPH,re.MGLYPH],[nn.MI,re.MI],[nn.MO,re.MO],[nn.MN,re.MN],[nn.MS,re.MS],[nn.MTEXT,re.MTEXT],[nn.NAV,re.NAV],[nn.NOBR,re.NOBR],[nn.NOFRAMES,re.NOFRAMES],[nn.NOEMBED,re.NOEMBED],[nn.NOSCRIPT,re.NOSCRIPT],[nn.OBJECT,re.OBJECT],[nn.OL,re.OL],[nn.OPTGROUP,re.OPTGROUP],[nn.OPTION,re.OPTION],[nn.P,re.P],[nn.PARAM,re.PARAM],[nn.PLAINTEXT,re.PLAINTEXT],[nn.PRE,re.PRE],[nn.RB,re.RB],[nn.RP,re.RP],[nn.RT,re.RT],[nn.RTC,re.RTC],[nn.RUBY,re.RUBY],[nn.S,re.S],[nn.SCRIPT,re.SCRIPT],[nn.SEARCH,re.SEARCH],[nn.SECTION,re.SECTION],[nn.SELECT,re.SELECT],[nn.SOURCE,re.SOURCE],[nn.SMALL,re.SMALL],[nn.SPAN,re.SPAN],[nn.STRIKE,re.STRIKE],[nn.STRONG,re.STRONG],[nn.STYLE,re.STYLE],[nn.SUB,re.SUB],[nn.SUMMARY,re.SUMMARY],[nn.SUP,re.SUP],[nn.TABLE,re.TABLE],[nn.TBODY,re.TBODY],[nn.TEMPLATE,re.TEMPLATE],[nn.TEXTAREA,re.TEXTAREA],[nn.TFOOT,re.TFOOT],[nn.TD,re.TD],[nn.TH,re.TH],[nn.THEAD,re.THEAD],[nn.TITLE,re.TITLE],[nn.TR,re.TR],[nn.TRACK,re.TRACK],[nn.TT,re.TT],[nn.U,re.U],[nn.UL,re.UL],[nn.SVG,re.SVG],[nn.VAR,re.VAR],[nn.WBR,re.WBR],[nn.XMP,re.XMP]]);function zK(n){var r;return(r=YXr.get(n))!==null&&r!==void 0?r:re.UNKNOWN}const Nr=re,XXr={[Er.HTML]:new Set([Nr.ADDRESS,Nr.APPLET,Nr.AREA,Nr.ARTICLE,Nr.ASIDE,Nr.BASE,Nr.BASEFONT,Nr.BGSOUND,Nr.BLOCKQUOTE,Nr.BODY,Nr.BR,Nr.BUTTON,Nr.CAPTION,Nr.CENTER,Nr.COL,Nr.COLGROUP,Nr.DD,Nr.DETAILS,Nr.DIR,Nr.DIV,Nr.DL,Nr.DT,Nr.EMBED,Nr.FIELDSET,Nr.FIGCAPTION,Nr.FIGURE,Nr.FOOTER,Nr.FORM,Nr.FRAME,Nr.FRAMESET,Nr.H1,Nr.H2,Nr.H3,Nr.H4,Nr.H5,Nr.H6,Nr.HEAD,Nr.HEADER,Nr.HGROUP,Nr.HR,Nr.HTML,Nr.IFRAME,Nr.IMG,Nr.INPUT,Nr.LI,Nr.LINK,Nr.LISTING,Nr.MAIN,Nr.MARQUEE,Nr.MENU,Nr.META,Nr.NAV,Nr.NOEMBED,Nr.NOFRAMES,Nr.NOSCRIPT,Nr.OBJECT,Nr.OL,Nr.P,Nr.PARAM,Nr.PLAINTEXT,Nr.PRE,Nr.SCRIPT,Nr.SECTION,Nr.SELECT,Nr.SOURCE,Nr.STYLE,Nr.SUMMARY,Nr.TABLE,Nr.TBODY,Nr.TD,Nr.TEMPLATE,Nr.TEXTAREA,Nr.TFOOT,Nr.TH,Nr.THEAD,Nr.TITLE,Nr.TR,Nr.TRACK,Nr.UL,Nr.WBR,Nr.XMP]),[Er.MATHML]:new Set([Nr.MI,Nr.MO,Nr.MN,Nr.MS,Nr.MTEXT,Nr.ANNOTATION_XML]),[Er.SVG]:new Set([Nr.TITLE,Nr.FOREIGN_OBJECT,Nr.DESC]),[Er.XLINK]:new Set,[Er.XML]:new Set,[Er.XMLNS]:new Set},EKe=new Set([Nr.H1,Nr.H2,Nr.H3,Nr.H4,Nr.H5,Nr.H6]);nn.STYLE,nn.SCRIPT,nn.XMP,nn.IFRAME,nn.NOEMBED,nn.NOFRAMES,nn.PLAINTEXT;var at;(function(n){n[n.DATA=0]="DATA",n[n.RCDATA=1]="RCDATA",n[n.RAWTEXT=2]="RAWTEXT",n[n.SCRIPT_DATA=3]="SCRIPT_DATA",n[n.PLAINTEXT=4]="PLAINTEXT",n[n.TAG_OPEN=5]="TAG_OPEN",n[n.END_TAG_OPEN=6]="END_TAG_OPEN",n[n.TAG_NAME=7]="TAG_NAME",n[n.RCDATA_LESS_THAN_SIGN=8]="RCDATA_LESS_THAN_SIGN",n[n.RCDATA_END_TAG_OPEN=9]="RCDATA_END_TAG_OPEN",n[n.RCDATA_END_TAG_NAME=10]="RCDATA_END_TAG_NAME",n[n.RAWTEXT_LESS_THAN_SIGN=11]="RAWTEXT_LESS_THAN_SIGN",n[n.RAWTEXT_END_TAG_OPEN=12]="RAWTEXT_END_TAG_OPEN",n[n.RAWTEXT_END_TAG_NAME=13]="RAWTEXT_END_TAG_NAME",n[n.SCRIPT_DATA_LESS_THAN_SIGN=14]="SCRIPT_DATA_LESS_THAN_SIGN",n[n.SCRIPT_DATA_END_TAG_OPEN=15]="SCRIPT_DATA_END_TAG_OPEN",n[n.SCRIPT_DATA_END_TAG_NAME=16]="SCRIPT_DATA_END_TAG_NAME",n[n.SCRIPT_DATA_ESCAPE_START=17]="SCRIPT_DATA_ESCAPE_START",n[n.SCRIPT_DATA_ESCAPE_START_DASH=18]="SCRIPT_DATA_ESCAPE_START_DASH",n[n.SCRIPT_DATA_ESCAPED=19]="SCRIPT_DATA_ESCAPED",n[n.SCRIPT_DATA_ESCAPED_DASH=20]="SCRIPT_DATA_ESCAPED_DASH",n[n.SCRIPT_DATA_ESCAPED_DASH_DASH=21]="SCRIPT_DATA_ESCAPED_DASH_DASH",n[n.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN=22]="SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN",n[n.SCRIPT_DATA_ESCAPED_END_TAG_OPEN=23]="SCRIPT_DATA_ESCAPED_END_TAG_OPEN",n[n.SCRIPT_DATA_ESCAPED_END_TAG_NAME=24]="SCRIPT_DATA_ESCAPED_END_TAG_NAME",n[n.SCRIPT_DATA_DOUBLE_ESCAPE_START=25]="SCRIPT_DATA_DOUBLE_ESCAPE_START",n[n.SCRIPT_DATA_DOUBLE_ESCAPED=26]="SCRIPT_DATA_DOUBLE_ESCAPED",n[n.SCRIPT_DATA_DOUBLE_ESCAPED_DASH=27]="SCRIPT_DATA_DOUBLE_ESCAPED_DASH",n[n.SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH=28]="SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH",n[n.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN=29]="SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN",n[n.SCRIPT_DATA_DOUBLE_ESCAPE_END=30]="SCRIPT_DATA_DOUBLE_ESCAPE_END",n[n.BEFORE_ATTRIBUTE_NAME=31]="BEFORE_ATTRIBUTE_NAME",n[n.ATTRIBUTE_NAME=32]="ATTRIBUTE_NAME",n[n.AFTER_ATTRIBUTE_NAME=33]="AFTER_ATTRIBUTE_NAME",n[n.BEFORE_ATTRIBUTE_VALUE=34]="BEFORE_ATTRIBUTE_VALUE",n[n.ATTRIBUTE_VALUE_DOUBLE_QUOTED=35]="ATTRIBUTE_VALUE_DOUBLE_QUOTED",n[n.ATTRIBUTE_VALUE_SINGLE_QUOTED=36]="ATTRIBUTE_VALUE_SINGLE_QUOTED",n[n.ATTRIBUTE_VALUE_UNQUOTED=37]="ATTRIBUTE_VALUE_UNQUOTED",n[n.AFTER_ATTRIBUTE_VALUE_QUOTED=38]="AFTER_ATTRIBUTE_VALUE_QUOTED",n[n.SELF_CLOSING_START_TAG=39]="SELF_CLOSING_START_TAG",n[n.BOGUS_COMMENT=40]="BOGUS_COMMENT",n[n.MARKUP_DECLARATION_OPEN=41]="MARKUP_DECLARATION_OPEN",n[n.COMMENT_START=42]="COMMENT_START",n[n.COMMENT_START_DASH=43]="COMMENT_START_DASH",n[n.COMMENT=44]="COMMENT",n[n.COMMENT_LESS_THAN_SIGN=45]="COMMENT_LESS_THAN_SIGN",n[n.COMMENT_LESS_THAN_SIGN_BANG=46]="COMMENT_LESS_THAN_SIGN_BANG",n[n.COMMENT_LESS_THAN_SIGN_BANG_DASH=47]="COMMENT_LESS_THAN_SIGN_BANG_DASH",n[n.COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH=48]="COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH",n[n.COMMENT_END_DASH=49]="COMMENT_END_DASH",n[n.COMMENT_END=50]="COMMENT_END",n[n.COMMENT_END_BANG=51]="COMMENT_END_BANG",n[n.DOCTYPE=52]="DOCTYPE",n[n.BEFORE_DOCTYPE_NAME=53]="BEFORE_DOCTYPE_NAME",n[n.DOCTYPE_NAME=54]="DOCTYPE_NAME",n[n.AFTER_DOCTYPE_NAME=55]="AFTER_DOCTYPE_NAME",n[n.AFTER_DOCTYPE_PUBLIC_KEYWORD=56]="AFTER_DOCTYPE_PUBLIC_KEYWORD",n[n.BEFORE_DOCTYPE_PUBLIC_IDENTIFIER=57]="BEFORE_DOCTYPE_PUBLIC_IDENTIFIER",n[n.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED=58]="DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED",n[n.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED=59]="DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED",n[n.AFTER_DOCTYPE_PUBLIC_IDENTIFIER=60]="AFTER_DOCTYPE_PUBLIC_IDENTIFIER",n[n.BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS=61]="BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS",n[n.AFTER_DOCTYPE_SYSTEM_KEYWORD=62]="AFTER_DOCTYPE_SYSTEM_KEYWORD",n[n.BEFORE_DOCTYPE_SYSTEM_IDENTIFIER=63]="BEFORE_DOCTYPE_SYSTEM_IDENTIFIER",n[n.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED=64]="DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED",n[n.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED=65]="DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED",n[n.AFTER_DOCTYPE_SYSTEM_IDENTIFIER=66]="AFTER_DOCTYPE_SYSTEM_IDENTIFIER",n[n.BOGUS_DOCTYPE=67]="BOGUS_DOCTYPE",n[n.CDATA_SECTION=68]="CDATA_SECTION",n[n.CDATA_SECTION_BRACKET=69]="CDATA_SECTION_BRACKET",n[n.CDATA_SECTION_END=70]="CDATA_SECTION_END",n[n.CHARACTER_REFERENCE=71]="CHARACTER_REFERENCE",n[n.AMBIGUOUS_AMPERSAND=72]="AMBIGUOUS_AMPERSAND"})(at||(at={}));const gp={DATA:at.DATA,RCDATA:at.RCDATA,RAWTEXT:at.RAWTEXT,SCRIPT_DATA:at.SCRIPT_DATA,PLAINTEXT:at.PLAINTEXT,CDATA_SECTION:at.CDATA_SECTION};function JXr(n){return n>=Ze.DIGIT_0&&n<=Ze.DIGIT_9}function eie(n){return n>=Ze.LATIN_CAPITAL_A&&n<=Ze.LATIN_CAPITAL_Z}function ZXr(n){return n>=Ze.LATIN_SMALL_A&&n<=Ze.LATIN_SMALL_Z}function eM(n){return ZXr(n)||eie(n)}function YWt(n){return eM(n)||JXr(n)}function K0e(n){return n+32}function Ufn(n){return n===Ze.SPACE||n===Ze.LINE_FEED||n===Ze.TABULATION||n===Ze.FORM_FEED}function XWt(n){return Ufn(n)||n===Ze.SOLIDUS||n===Ze.GREATER_THAN_SIGN}function QXr(n){return n===Ze.NULL?On.nullCharacterReference:n>1114111?On.characterReferenceOutsideUnicodeRange:jfn(n)?On.surrogateCharacterReference:Ffn(n)?On.noncharacterCharacterReference:$fn(n)||n===Ze.CARRIAGE_RETURN?On.controlCharacterReference:null}class eJr{constructor(r,o){this.options=r,this.handler=o,this.paused=!1,this.inLoop=!1,this.inForeignNode=!1,this.lastStartTagName="",this.active=!1,this.state=at.DATA,this.returnState=at.DATA,this.entityStartPos=0,this.consumedAfterSnapshot=-1,this.currentCharacterToken=null,this.currentToken=null,this.currentAttr={name:"",value:""},this.preprocessor=new FXr(o),this.currentLocation=this.getCurrentLocation(-1),this.entityDecoder=new WXr(BXr,(s,c)=>{this.preprocessor.pos=this.entityStartPos+c-1,this._flushCodePointConsumedAsCharacterReference(s)},o.onParseError?{missingSemicolonAfterCharacterReference:()=>{this._err(On.missingSemicolonAfterCharacterReference,1)},absenceOfDigitsInNumericCharacterReference:s=>{this._err(On.absenceOfDigitsInNumericCharacterReference,this.entityStartPos-this.preprocessor.pos+s)},validateNumericCharacterReference:s=>{const c=QXr(s);c&&this._err(c,1)}}:void 0)}_err(r,o=0){var s,c;(c=(s=this.handler).onParseError)===null||c===void 0||c.call(s,this.preprocessor.getError(r,o))}getCurrentLocation(r){return this.options.sourceCodeLocationInfo?{startLine:this.preprocessor.line,startCol:this.preprocessor.col-r,startOffset:this.preprocessor.offset-r,endLine:-1,endCol:-1,endOffset:-1}:null}_runParsingLoop(){if(!this.inLoop){for(this.inLoop=!0;this.active&&!this.paused;){this.consumedAfterSnapshot=0;const r=this._consume();this._ensureHibernation()||this._callState(r)}this.inLoop=!1}}pause(){this.paused=!0}resume(r){if(!this.paused)throw new Error("Parser was already resumed");this.paused=!1,!this.inLoop&&(this._runParsingLoop(),this.paused||r?.())}write(r,o,s){this.active=!0,this.preprocessor.write(r,o),this._runParsingLoop(),this.paused||s?.()}insertHtmlAtCurrentPos(r){this.active=!0,this.preprocessor.insertHtmlAtCurrentPos(r),this._runParsingLoop()}_ensureHibernation(){return this.preprocessor.endOfChunkHit?(this.preprocessor.retreat(this.consumedAfterSnapshot),this.consumedAfterSnapshot=0,this.active=!1,!0):!1}_consume(){return this.consumedAfterSnapshot++,this.preprocessor.advance()}_advanceBy(r){this.consumedAfterSnapshot+=r;for(let o=0;o<r;o++)this.preprocessor.advance()}_consumeSequenceIfMatch(r,o){return this.preprocessor.startsWith(r,o)?(this._advanceBy(r.length-1),!0):!1}_createStartTagToken(){this.currentToken={type:nc.START_TAG,tagName:"",tagID:re.UNKNOWN,selfClosing:!1,ackSelfClosing:!1,attrs:[],location:this.getCurrentLocation(1)}}_createEndTagToken(){this.currentToken={type:nc.END_TAG,tagName:"",tagID:re.UNKNOWN,selfClosing:!1,ackSelfClosing:!1,attrs:[],location:this.getCurrentLocation(2)}}_createCommentToken(r){this.currentToken={type:nc.COMMENT,data:"",location:this.getCurrentLocation(r)}}_createDoctypeToken(r){this.currentToken={type:nc.DOCTYPE,name:r,forceQuirks:!1,publicId:null,systemId:null,location:this.currentLocation}}_createCharacterToken(r,o){this.currentCharacterToken={type:r,chars:o,location:this.currentLocation}}_createAttr(r){this.currentAttr={name:r,value:""},this.currentLocation=this.getCurrentLocation(0)}_leaveAttrName(){var r,o;const s=this.currentToken;if(Bfn(s,this.currentAttr.name)===null){if(s.attrs.push(this.currentAttr),s.location&&this.currentLocation){const c=(r=(o=s.location).attrs)!==null&&r!==void 0?r:o.attrs=Object.create(null);c[this.currentAttr.name]=this.currentLocation,this._leaveAttrValue()}}else this._err(On.duplicateAttribute)}_leaveAttrValue(){this.currentLocation&&(this.currentLocation.endLine=this.preprocessor.line,this.currentLocation.endCol=this.preprocessor.col,this.currentLocation.endOffset=this.preprocessor.offset)}prepareToken(r){this._emitCurrentCharacterToken(r.location),this.currentToken=null,r.location&&(r.location.endLine=this.preprocessor.line,r.location.endCol=this.preprocessor.col+1,r.location.endOffset=this.preprocessor.offset+1),this.currentLocation=this.getCurrentLocation(-1)}emitCurrentTagToken(){const r=this.currentToken;this.prepareToken(r),r.tagID=zK(r.tagName),r.type===nc.START_TAG?(this.lastStartTagName=r.tagName,this.handler.onStartTag(r)):(r.attrs.length>0&&this._err(On.endTagWithAttributes),r.selfClosing&&this._err(On.endTagWithTrailingSolidus),this.handler.onEndTag(r)),this.preprocessor.dropParsedChunk()}emitCurrentComment(r){this.prepareToken(r),this.handler.onComment(r),this.preprocessor.dropParsedChunk()}emitCurrentDoctype(r){this.prepareToken(r),this.handler.onDoctype(r),this.preprocessor.dropParsedChunk()}_emitCurrentCharacterToken(r){if(this.currentCharacterToken){switch(r&&this.currentCharacterToken.location&&(this.currentCharacterToken.location.endLine=r.startLine,this.currentCharacterToken.location.endCol=r.startCol,this.currentCharacterToken.location.endOffset=r.startOffset),this.currentCharacterToken.type){case nc.CHARACTER:{this.handler.onCharacter(this.currentCharacterToken);break}case nc.NULL_CHARACTER:{this.handler.onNullCharacter(this.currentCharacterToken);break}case nc.WHITESPACE_CHARACTER:{this.handler.onWhitespaceCharacter(this.currentCharacterToken);break}}this.currentCharacterToken=null}}_emitEOFToken(){const r=this.getCurrentLocation(0);r&&(r.endLine=r.startLine,r.endCol=r.startCol,r.endOffset=r.startOffset),this._emitCurrentCharacterToken(r),this.handler.onEof({type:nc.EOF,location:r}),this.active=!1}_appendCharToCurrentCharacterToken(r,o){if(this.currentCharacterToken)if(this.currentCharacterToken.type===r){this.currentCharacterToken.chars+=o;return}else this.currentLocation=this.getCurrentLocation(0),this._emitCurrentCharacterToken(this.currentLocation),this.preprocessor.dropParsedChunk();this._createCharacterToken(r,o)}_emitCodePoint(r){const o=Ufn(r)?nc.WHITESPACE_CHARACTER:r===Ze.NULL?nc.NULL_CHARACTER:nc.CHARACTER;this._appendCharToCurrentCharacterToken(o,String.fromCodePoint(r))}_emitChars(r){this._appendCharToCurrentCharacterToken(nc.CHARACTER,r)}_startCharacterReference(){this.returnState=this.state,this.state=at.CHARACTER_REFERENCE,this.entityStartPos=this.preprocessor.pos,this.entityDecoder.startEntity(this._isCharacterReferenceInAttribute()?R4.Attribute:R4.Legacy)}_isCharacterReferenceInAttribute(){return this.returnState===at.ATTRIBUTE_VALUE_DOUBLE_QUOTED||this.returnState===at.ATTRIBUTE_VALUE_SINGLE_QUOTED||this.returnState===at.ATTRIBUTE_VALUE_UNQUOTED}_flushCodePointConsumedAsCharacterReference(r){this._isCharacterReferenceInAttribute()?this.currentAttr.value+=String.fromCodePoint(r):this._emitCodePoint(r)}_callState(r){switch(this.state){case at.DATA:{this._stateData(r);break}case at.RCDATA:{this._stateRcdata(r);break}case at.RAWTEXT:{this._stateRawtext(r);break}case at.SCRIPT_DATA:{this._stateScriptData(r);break}case at.PLAINTEXT:{this._statePlaintext(r);break}case at.TAG_OPEN:{this._stateTagOpen(r);break}case at.END_TAG_OPEN:{this._stateEndTagOpen(r);break}case at.TAG_NAME:{this._stateTagName(r);break}case at.RCDATA_LESS_THAN_SIGN:{this._stateRcdataLessThanSign(r);break}case at.RCDATA_END_TAG_OPEN:{this._stateRcdataEndTagOpen(r);break}case at.RCDATA_END_TAG_NAME:{this._stateRcdataEndTagName(r);break}case at.RAWTEXT_LESS_THAN_SIGN:{this._stateRawtextLessThanSign(r);break}case at.RAWTEXT_END_TAG_OPEN:{this._stateRawtextEndTagOpen(r);break}case at.RAWTEXT_END_TAG_NAME:{this._stateRawtextEndTagName(r);break}case at.SCRIPT_DATA_LESS_THAN_SIGN:{this._stateScriptDataLessThanSign(r);break}case at.SCRIPT_DATA_END_TAG_OPEN:{this._stateScriptDataEndTagOpen(r);break}case at.SCRIPT_DATA_END_TAG_NAME:{this._stateScriptDataEndTagName(r);break}case at.SCRIPT_DATA_ESCAPE_START:{this._stateScriptDataEscapeStart(r);break}case at.SCRIPT_DATA_ESCAPE_START_DASH:{this._stateScriptDataEscapeStartDash(r);break}case at.SCRIPT_DATA_ESCAPED:{this._stateScriptDataEscaped(r);break}case at.SCRIPT_DATA_ESCAPED_DASH:{this._stateScriptDataEscapedDash(r);break}case at.SCRIPT_DATA_ESCAPED_DASH_DASH:{this._stateScriptDataEscapedDashDash(r);break}case at.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN:{this._stateScriptDataEscapedLessThanSign(r);break}case at.SCRIPT_DATA_ESCAPED_END_TAG_OPEN:{this._stateScriptDataEscapedEndTagOpen(r);break}case at.SCRIPT_DATA_ESCAPED_END_TAG_NAME:{this._stateScriptDataEscapedEndTagName(r);break}case at.SCRIPT_DATA_DOUBLE_ESCAPE_START:{this._stateScriptDataDoubleEscapeStart(r);break}case at.SCRIPT_DATA_DOUBLE_ESCAPED:{this._stateScriptDataDoubleEscaped(r);break}case at.SCRIPT_DATA_DOUBLE_ESCAPED_DASH:{this._stateScriptDataDoubleEscapedDash(r);break}case at.SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH:{this._stateScriptDataDoubleEscapedDashDash(r);break}case at.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN:{this._stateScriptDataDoubleEscapedLessThanSign(r);break}case at.SCRIPT_DATA_DOUBLE_ESCAPE_END:{this._stateScriptDataDoubleEscapeEnd(r);break}case at.BEFORE_ATTRIBUTE_NAME:{this._stateBeforeAttributeName(r);break}case at.ATTRIBUTE_NAME:{this._stateAttributeName(r);break}case at.AFTER_ATTRIBUTE_NAME:{this._stateAfterAttributeName(r);break}case at.BEFORE_ATTRIBUTE_VALUE:{this._stateBeforeAttributeValue(r);break}case at.ATTRIBUTE_VALUE_DOUBLE_QUOTED:{this._stateAttributeValueDoubleQuoted(r);break}case at.ATTRIBUTE_VALUE_SINGLE_QUOTED:{this._stateAttributeValueSingleQuoted(r);break}case at.ATTRIBUTE_VALUE_UNQUOTED:{this._stateAttributeValueUnquoted(r);break}case at.AFTER_ATTRIBUTE_VALUE_QUOTED:{this._stateAfterAttributeValueQuoted(r);break}case at.SELF_CLOSING_START_TAG:{this._stateSelfClosingStartTag(r);break}case at.BOGUS_COMMENT:{this._stateBogusComment(r);break}case at.MARKUP_DECLARATION_OPEN:{this._stateMarkupDeclarationOpen(r);break}case at.COMMENT_START:{this._stateCommentStart(r);break}case at.COMMENT_START_DASH:{this._stateCommentStartDash(r);break}case at.COMMENT:{this._stateComment(r);break}case at.COMMENT_LESS_THAN_SIGN:{this._stateCommentLessThanSign(r);break}case at.COMMENT_LESS_THAN_SIGN_BANG:{this._stateCommentLessThanSignBang(r);break}case at.COMMENT_LESS_THAN_SIGN_BANG_DASH:{this._stateCommentLessThanSignBangDash(r);break}case at.COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH:{this._stateCommentLessThanSignBangDashDash(r);break}case at.COMMENT_END_DASH:{this._stateCommentEndDash(r);break}case at.COMMENT_END:{this._stateCommentEnd(r);break}case at.COMMENT_END_BANG:{this._stateCommentEndBang(r);break}case at.DOCTYPE:{this._stateDoctype(r);break}case at.BEFORE_DOCTYPE_NAME:{this._stateBeforeDoctypeName(r);break}case at.DOCTYPE_NAME:{this._stateDoctypeName(r);break}case at.AFTER_DOCTYPE_NAME:{this._stateAfterDoctypeName(r);break}case at.AFTER_DOCTYPE_PUBLIC_KEYWORD:{this._stateAfterDoctypePublicKeyword(r);break}case at.BEFORE_DOCTYPE_PUBLIC_IDENTIFIER:{this._stateBeforeDoctypePublicIdentifier(r);break}case at.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED:{this._stateDoctypePublicIdentifierDoubleQuoted(r);break}case at.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED:{this._stateDoctypePublicIdentifierSingleQuoted(r);break}case at.AFTER_DOCTYPE_PUBLIC_IDENTIFIER:{this._stateAfterDoctypePublicIdentifier(r);break}case at.BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS:{this._stateBetweenDoctypePublicAndSystemIdentifiers(r);break}case at.AFTER_DOCTYPE_SYSTEM_KEYWORD:{this._stateAfterDoctypeSystemKeyword(r);break}case at.BEFORE_DOCTYPE_SYSTEM_IDENTIFIER:{this._stateBeforeDoctypeSystemIdentifier(r);break}case at.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED:{this._stateDoctypeSystemIdentifierDoubleQuoted(r);break}case at.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED:{this._stateDoctypeSystemIdentifierSingleQuoted(r);break}case at.AFTER_DOCTYPE_SYSTEM_IDENTIFIER:{this._stateAfterDoctypeSystemIdentifier(r);break}case at.BOGUS_DOCTYPE:{this._stateBogusDoctype(r);break}case at.CDATA_SECTION:{this._stateCdataSection(r);break}case at.CDATA_SECTION_BRACKET:{this._stateCdataSectionBracket(r);break}case at.CDATA_SECTION_END:{this._stateCdataSectionEnd(r);break}case at.CHARACTER_REFERENCE:{this._stateCharacterReference();break}case at.AMBIGUOUS_AMPERSAND:{this._stateAmbiguousAmpersand(r);break}default:throw new Error("Unknown state")}}_stateData(r){switch(r){case Ze.LESS_THAN_SIGN:{this.state=at.TAG_OPEN;break}case Ze.AMPERSAND:{this._startCharacterReference();break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._emitCodePoint(r);break}case Ze.EOF:{this._emitEOFToken();break}default:this._emitCodePoint(r)}}_stateRcdata(r){switch(r){case Ze.AMPERSAND:{this._startCharacterReference();break}case Ze.LESS_THAN_SIGN:{this.state=at.RCDATA_LESS_THAN_SIGN;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._emitChars(Sf);break}case Ze.EOF:{this._emitEOFToken();break}default:this._emitCodePoint(r)}}_stateRawtext(r){switch(r){case Ze.LESS_THAN_SIGN:{this.state=at.RAWTEXT_LESS_THAN_SIGN;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._emitChars(Sf);break}case Ze.EOF:{this._emitEOFToken();break}default:this._emitCodePoint(r)}}_stateScriptData(r){switch(r){case Ze.LESS_THAN_SIGN:{this.state=at.SCRIPT_DATA_LESS_THAN_SIGN;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._emitChars(Sf);break}case Ze.EOF:{this._emitEOFToken();break}default:this._emitCodePoint(r)}}_statePlaintext(r){switch(r){case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._emitChars(Sf);break}case Ze.EOF:{this._emitEOFToken();break}default:this._emitCodePoint(r)}}_stateTagOpen(r){if(eM(r))this._createStartTagToken(),this.state=at.TAG_NAME,this._stateTagName(r);else switch(r){case Ze.EXCLAMATION_MARK:{this.state=at.MARKUP_DECLARATION_OPEN;break}case Ze.SOLIDUS:{this.state=at.END_TAG_OPEN;break}case Ze.QUESTION_MARK:{this._err(On.unexpectedQuestionMarkInsteadOfTagName),this._createCommentToken(1),this.state=at.BOGUS_COMMENT,this._stateBogusComment(r);break}case Ze.EOF:{this._err(On.eofBeforeTagName),this._emitChars("<"),this._emitEOFToken();break}default:this._err(On.invalidFirstCharacterOfTagName),this._emitChars("<"),this.state=at.DATA,this._stateData(r)}}_stateEndTagOpen(r){if(eM(r))this._createEndTagToken(),this.state=at.TAG_NAME,this._stateTagName(r);else switch(r){case Ze.GREATER_THAN_SIGN:{this._err(On.missingEndTagName),this.state=at.DATA;break}case Ze.EOF:{this._err(On.eofBeforeTagName),this._emitChars("</"),this._emitEOFToken();break}default:this._err(On.invalidFirstCharacterOfTagName),this._createCommentToken(2),this.state=at.BOGUS_COMMENT,this._stateBogusComment(r)}}_stateTagName(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this.state=at.BEFORE_ATTRIBUTE_NAME;break}case Ze.SOLIDUS:{this.state=at.SELF_CLOSING_START_TAG;break}case Ze.GREATER_THAN_SIGN:{this.state=at.DATA,this.emitCurrentTagToken();break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.tagName+=Sf;break}case Ze.EOF:{this._err(On.eofInTag),this._emitEOFToken();break}default:o.tagName+=String.fromCodePoint(eie(r)?K0e(r):r)}}_stateRcdataLessThanSign(r){r===Ze.SOLIDUS?this.state=at.RCDATA_END_TAG_OPEN:(this._emitChars("<"),this.state=at.RCDATA,this._stateRcdata(r))}_stateRcdataEndTagOpen(r){eM(r)?(this.state=at.RCDATA_END_TAG_NAME,this._stateRcdataEndTagName(r)):(this._emitChars("</"),this.state=at.RCDATA,this._stateRcdata(r))}handleSpecialEndTag(r){if(!this.preprocessor.startsWith(this.lastStartTagName,!1))return!this._ensureHibernation();this._createEndTagToken();const o=this.currentToken;switch(o.tagName=this.lastStartTagName,this.preprocessor.peek(this.lastStartTagName.length)){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:return this._advanceBy(this.lastStartTagName.length),this.state=at.BEFORE_ATTRIBUTE_NAME,!1;case Ze.SOLIDUS:return this._advanceBy(this.lastStartTagName.length),this.state=at.SELF_CLOSING_START_TAG,!1;case Ze.GREATER_THAN_SIGN:return this._advanceBy(this.lastStartTagName.length),this.emitCurrentTagToken(),this.state=at.DATA,!1;default:return!this._ensureHibernation()}}_stateRcdataEndTagName(r){this.handleSpecialEndTag(r)&&(this._emitChars("</"),this.state=at.RCDATA,this._stateRcdata(r))}_stateRawtextLessThanSign(r){r===Ze.SOLIDUS?this.state=at.RAWTEXT_END_TAG_OPEN:(this._emitChars("<"),this.state=at.RAWTEXT,this._stateRawtext(r))}_stateRawtextEndTagOpen(r){eM(r)?(this.state=at.RAWTEXT_END_TAG_NAME,this._stateRawtextEndTagName(r)):(this._emitChars("</"),this.state=at.RAWTEXT,this._stateRawtext(r))}_stateRawtextEndTagName(r){this.handleSpecialEndTag(r)&&(this._emitChars("</"),this.state=at.RAWTEXT,this._stateRawtext(r))}_stateScriptDataLessThanSign(r){switch(r){case Ze.SOLIDUS:{this.state=at.SCRIPT_DATA_END_TAG_OPEN;break}case Ze.EXCLAMATION_MARK:{this.state=at.SCRIPT_DATA_ESCAPE_START,this._emitChars("<!");break}default:this._emitChars("<"),this.state=at.SCRIPT_DATA,this._stateScriptData(r)}}_stateScriptDataEndTagOpen(r){eM(r)?(this.state=at.SCRIPT_DATA_END_TAG_NAME,this._stateScriptDataEndTagName(r)):(this._emitChars("</"),this.state=at.SCRIPT_DATA,this._stateScriptData(r))}_stateScriptDataEndTagName(r){this.handleSpecialEndTag(r)&&(this._emitChars("</"),this.state=at.SCRIPT_DATA,this._stateScriptData(r))}_stateScriptDataEscapeStart(r){r===Ze.HYPHEN_MINUS?(this.state=at.SCRIPT_DATA_ESCAPE_START_DASH,this._emitChars("-")):(this.state=at.SCRIPT_DATA,this._stateScriptData(r))}_stateScriptDataEscapeStartDash(r){r===Ze.HYPHEN_MINUS?(this.state=at.SCRIPT_DATA_ESCAPED_DASH_DASH,this._emitChars("-")):(this.state=at.SCRIPT_DATA,this._stateScriptData(r))}_stateScriptDataEscaped(r){switch(r){case Ze.HYPHEN_MINUS:{this.state=at.SCRIPT_DATA_ESCAPED_DASH,this._emitChars("-");break}case Ze.LESS_THAN_SIGN:{this.state=at.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._emitChars(Sf);break}case Ze.EOF:{this._err(On.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break}default:this._emitCodePoint(r)}}_stateScriptDataEscapedDash(r){switch(r){case Ze.HYPHEN_MINUS:{this.state=at.SCRIPT_DATA_ESCAPED_DASH_DASH,this._emitChars("-");break}case Ze.LESS_THAN_SIGN:{this.state=at.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.state=at.SCRIPT_DATA_ESCAPED,this._emitChars(Sf);break}case Ze.EOF:{this._err(On.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break}default:this.state=at.SCRIPT_DATA_ESCAPED,this._emitCodePoint(r)}}_stateScriptDataEscapedDashDash(r){switch(r){case Ze.HYPHEN_MINUS:{this._emitChars("-");break}case Ze.LESS_THAN_SIGN:{this.state=at.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN;break}case Ze.GREATER_THAN_SIGN:{this.state=at.SCRIPT_DATA,this._emitChars(">");break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.state=at.SCRIPT_DATA_ESCAPED,this._emitChars(Sf);break}case Ze.EOF:{this._err(On.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break}default:this.state=at.SCRIPT_DATA_ESCAPED,this._emitCodePoint(r)}}_stateScriptDataEscapedLessThanSign(r){r===Ze.SOLIDUS?this.state=at.SCRIPT_DATA_ESCAPED_END_TAG_OPEN:eM(r)?(this._emitChars("<"),this.state=at.SCRIPT_DATA_DOUBLE_ESCAPE_START,this._stateScriptDataDoubleEscapeStart(r)):(this._emitChars("<"),this.state=at.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(r))}_stateScriptDataEscapedEndTagOpen(r){eM(r)?(this.state=at.SCRIPT_DATA_ESCAPED_END_TAG_NAME,this._stateScriptDataEscapedEndTagName(r)):(this._emitChars("</"),this.state=at.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(r))}_stateScriptDataEscapedEndTagName(r){this.handleSpecialEndTag(r)&&(this._emitChars("</"),this.state=at.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(r))}_stateScriptDataDoubleEscapeStart(r){if(this.preprocessor.startsWith(zw.SCRIPT,!1)&&XWt(this.preprocessor.peek(zw.SCRIPT.length))){this._emitCodePoint(r);for(let o=0;o<zw.SCRIPT.length;o++)this._emitCodePoint(this._consume());this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED}else this._ensureHibernation()||(this.state=at.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(r))}_stateScriptDataDoubleEscaped(r){switch(r){case Ze.HYPHEN_MINUS:{this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED_DASH,this._emitChars("-");break}case Ze.LESS_THAN_SIGN:{this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN,this._emitChars("<");break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._emitChars(Sf);break}case Ze.EOF:{this._err(On.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break}default:this._emitCodePoint(r)}}_stateScriptDataDoubleEscapedDash(r){switch(r){case Ze.HYPHEN_MINUS:{this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH,this._emitChars("-");break}case Ze.LESS_THAN_SIGN:{this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN,this._emitChars("<");break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitChars(Sf);break}case Ze.EOF:{this._err(On.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break}default:this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitCodePoint(r)}}_stateScriptDataDoubleEscapedDashDash(r){switch(r){case Ze.HYPHEN_MINUS:{this._emitChars("-");break}case Ze.LESS_THAN_SIGN:{this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN,this._emitChars("<");break}case Ze.GREATER_THAN_SIGN:{this.state=at.SCRIPT_DATA,this._emitChars(">");break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitChars(Sf);break}case Ze.EOF:{this._err(On.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break}default:this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitCodePoint(r)}}_stateScriptDataDoubleEscapedLessThanSign(r){r===Ze.SOLIDUS?(this.state=at.SCRIPT_DATA_DOUBLE_ESCAPE_END,this._emitChars("/")):(this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED,this._stateScriptDataDoubleEscaped(r))}_stateScriptDataDoubleEscapeEnd(r){if(this.preprocessor.startsWith(zw.SCRIPT,!1)&&XWt(this.preprocessor.peek(zw.SCRIPT.length))){this._emitCodePoint(r);for(let o=0;o<zw.SCRIPT.length;o++)this._emitCodePoint(this._consume());this.state=at.SCRIPT_DATA_ESCAPED}else this._ensureHibernation()||(this.state=at.SCRIPT_DATA_DOUBLE_ESCAPED,this._stateScriptDataDoubleEscaped(r))}_stateBeforeAttributeName(r){switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.SOLIDUS:case Ze.GREATER_THAN_SIGN:case Ze.EOF:{this.state=at.AFTER_ATTRIBUTE_NAME,this._stateAfterAttributeName(r);break}case Ze.EQUALS_SIGN:{this._err(On.unexpectedEqualsSignBeforeAttributeName),this._createAttr("="),this.state=at.ATTRIBUTE_NAME;break}default:this._createAttr(""),this.state=at.ATTRIBUTE_NAME,this._stateAttributeName(r)}}_stateAttributeName(r){switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:case Ze.SOLIDUS:case Ze.GREATER_THAN_SIGN:case Ze.EOF:{this._leaveAttrName(),this.state=at.AFTER_ATTRIBUTE_NAME,this._stateAfterAttributeName(r);break}case Ze.EQUALS_SIGN:{this._leaveAttrName(),this.state=at.BEFORE_ATTRIBUTE_VALUE;break}case Ze.QUOTATION_MARK:case Ze.APOSTROPHE:case Ze.LESS_THAN_SIGN:{this._err(On.unexpectedCharacterInAttributeName),this.currentAttr.name+=String.fromCodePoint(r);break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.currentAttr.name+=Sf;break}default:this.currentAttr.name+=String.fromCodePoint(eie(r)?K0e(r):r)}}_stateAfterAttributeName(r){switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.SOLIDUS:{this.state=at.SELF_CLOSING_START_TAG;break}case Ze.EQUALS_SIGN:{this.state=at.BEFORE_ATTRIBUTE_VALUE;break}case Ze.GREATER_THAN_SIGN:{this.state=at.DATA,this.emitCurrentTagToken();break}case Ze.EOF:{this._err(On.eofInTag),this._emitEOFToken();break}default:this._createAttr(""),this.state=at.ATTRIBUTE_NAME,this._stateAttributeName(r)}}_stateBeforeAttributeValue(r){switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.QUOTATION_MARK:{this.state=at.ATTRIBUTE_VALUE_DOUBLE_QUOTED;break}case Ze.APOSTROPHE:{this.state=at.ATTRIBUTE_VALUE_SINGLE_QUOTED;break}case Ze.GREATER_THAN_SIGN:{this._err(On.missingAttributeValue),this.state=at.DATA,this.emitCurrentTagToken();break}default:this.state=at.ATTRIBUTE_VALUE_UNQUOTED,this._stateAttributeValueUnquoted(r)}}_stateAttributeValueDoubleQuoted(r){switch(r){case Ze.QUOTATION_MARK:{this.state=at.AFTER_ATTRIBUTE_VALUE_QUOTED;break}case Ze.AMPERSAND:{this._startCharacterReference();break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.currentAttr.value+=Sf;break}case Ze.EOF:{this._err(On.eofInTag),this._emitEOFToken();break}default:this.currentAttr.value+=String.fromCodePoint(r)}}_stateAttributeValueSingleQuoted(r){switch(r){case Ze.APOSTROPHE:{this.state=at.AFTER_ATTRIBUTE_VALUE_QUOTED;break}case Ze.AMPERSAND:{this._startCharacterReference();break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.currentAttr.value+=Sf;break}case Ze.EOF:{this._err(On.eofInTag),this._emitEOFToken();break}default:this.currentAttr.value+=String.fromCodePoint(r)}}_stateAttributeValueUnquoted(r){switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this._leaveAttrValue(),this.state=at.BEFORE_ATTRIBUTE_NAME;break}case Ze.AMPERSAND:{this._startCharacterReference();break}case Ze.GREATER_THAN_SIGN:{this._leaveAttrValue(),this.state=at.DATA,this.emitCurrentTagToken();break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),this.currentAttr.value+=Sf;break}case Ze.QUOTATION_MARK:case Ze.APOSTROPHE:case Ze.LESS_THAN_SIGN:case Ze.EQUALS_SIGN:case Ze.GRAVE_ACCENT:{this._err(On.unexpectedCharacterInUnquotedAttributeValue),this.currentAttr.value+=String.fromCodePoint(r);break}case Ze.EOF:{this._err(On.eofInTag),this._emitEOFToken();break}default:this.currentAttr.value+=String.fromCodePoint(r)}}_stateAfterAttributeValueQuoted(r){switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this._leaveAttrValue(),this.state=at.BEFORE_ATTRIBUTE_NAME;break}case Ze.SOLIDUS:{this._leaveAttrValue(),this.state=at.SELF_CLOSING_START_TAG;break}case Ze.GREATER_THAN_SIGN:{this._leaveAttrValue(),this.state=at.DATA,this.emitCurrentTagToken();break}case Ze.EOF:{this._err(On.eofInTag),this._emitEOFToken();break}default:this._err(On.missingWhitespaceBetweenAttributes),this.state=at.BEFORE_ATTRIBUTE_NAME,this._stateBeforeAttributeName(r)}}_stateSelfClosingStartTag(r){switch(r){case Ze.GREATER_THAN_SIGN:{const o=this.currentToken;o.selfClosing=!0,this.state=at.DATA,this.emitCurrentTagToken();break}case Ze.EOF:{this._err(On.eofInTag),this._emitEOFToken();break}default:this._err(On.unexpectedSolidusInTag),this.state=at.BEFORE_ATTRIBUTE_NAME,this._stateBeforeAttributeName(r)}}_stateBogusComment(r){const o=this.currentToken;switch(r){case Ze.GREATER_THAN_SIGN:{this.state=at.DATA,this.emitCurrentComment(o);break}case Ze.EOF:{this.emitCurrentComment(o),this._emitEOFToken();break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.data+=Sf;break}default:o.data+=String.fromCodePoint(r)}}_stateMarkupDeclarationOpen(r){this._consumeSequenceIfMatch(zw.DASH_DASH,!0)?(this._createCommentToken(zw.DASH_DASH.length+1),this.state=at.COMMENT_START):this._consumeSequenceIfMatch(zw.DOCTYPE,!1)?(this.currentLocation=this.getCurrentLocation(zw.DOCTYPE.length+1),this.state=at.DOCTYPE):this._consumeSequenceIfMatch(zw.CDATA_START,!0)?this.inForeignNode?this.state=at.CDATA_SECTION:(this._err(On.cdataInHtmlContent),this._createCommentToken(zw.CDATA_START.length+1),this.currentToken.data="[CDATA[",this.state=at.BOGUS_COMMENT):this._ensureHibernation()||(this._err(On.incorrectlyOpenedComment),this._createCommentToken(2),this.state=at.BOGUS_COMMENT,this._stateBogusComment(r))}_stateCommentStart(r){switch(r){case Ze.HYPHEN_MINUS:{this.state=at.COMMENT_START_DASH;break}case Ze.GREATER_THAN_SIGN:{this._err(On.abruptClosingOfEmptyComment),this.state=at.DATA;const o=this.currentToken;this.emitCurrentComment(o);break}default:this.state=at.COMMENT,this._stateComment(r)}}_stateCommentStartDash(r){const o=this.currentToken;switch(r){case Ze.HYPHEN_MINUS:{this.state=at.COMMENT_END;break}case Ze.GREATER_THAN_SIGN:{this._err(On.abruptClosingOfEmptyComment),this.state=at.DATA,this.emitCurrentComment(o);break}case Ze.EOF:{this._err(On.eofInComment),this.emitCurrentComment(o),this._emitEOFToken();break}default:o.data+="-",this.state=at.COMMENT,this._stateComment(r)}}_stateComment(r){const o=this.currentToken;switch(r){case Ze.HYPHEN_MINUS:{this.state=at.COMMENT_END_DASH;break}case Ze.LESS_THAN_SIGN:{o.data+="<",this.state=at.COMMENT_LESS_THAN_SIGN;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.data+=Sf;break}case Ze.EOF:{this._err(On.eofInComment),this.emitCurrentComment(o),this._emitEOFToken();break}default:o.data+=String.fromCodePoint(r)}}_stateCommentLessThanSign(r){const o=this.currentToken;switch(r){case Ze.EXCLAMATION_MARK:{o.data+="!",this.state=at.COMMENT_LESS_THAN_SIGN_BANG;break}case Ze.LESS_THAN_SIGN:{o.data+="<";break}default:this.state=at.COMMENT,this._stateComment(r)}}_stateCommentLessThanSignBang(r){r===Ze.HYPHEN_MINUS?this.state=at.COMMENT_LESS_THAN_SIGN_BANG_DASH:(this.state=at.COMMENT,this._stateComment(r))}_stateCommentLessThanSignBangDash(r){r===Ze.HYPHEN_MINUS?this.state=at.COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH:(this.state=at.COMMENT_END_DASH,this._stateCommentEndDash(r))}_stateCommentLessThanSignBangDashDash(r){r!==Ze.GREATER_THAN_SIGN&&r!==Ze.EOF&&this._err(On.nestedComment),this.state=at.COMMENT_END,this._stateCommentEnd(r)}_stateCommentEndDash(r){const o=this.currentToken;switch(r){case Ze.HYPHEN_MINUS:{this.state=at.COMMENT_END;break}case Ze.EOF:{this._err(On.eofInComment),this.emitCurrentComment(o),this._emitEOFToken();break}default:o.data+="-",this.state=at.COMMENT,this._stateComment(r)}}_stateCommentEnd(r){const o=this.currentToken;switch(r){case Ze.GREATER_THAN_SIGN:{this.state=at.DATA,this.emitCurrentComment(o);break}case Ze.EXCLAMATION_MARK:{this.state=at.COMMENT_END_BANG;break}case Ze.HYPHEN_MINUS:{o.data+="-";break}case Ze.EOF:{this._err(On.eofInComment),this.emitCurrentComment(o),this._emitEOFToken();break}default:o.data+="--",this.state=at.COMMENT,this._stateComment(r)}}_stateCommentEndBang(r){const o=this.currentToken;switch(r){case Ze.HYPHEN_MINUS:{o.data+="--!",this.state=at.COMMENT_END_DASH;break}case Ze.GREATER_THAN_SIGN:{this._err(On.incorrectlyClosedComment),this.state=at.DATA,this.emitCurrentComment(o);break}case Ze.EOF:{this._err(On.eofInComment),this.emitCurrentComment(o),this._emitEOFToken();break}default:o.data+="--!",this.state=at.COMMENT,this._stateComment(r)}}_stateDoctype(r){switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this.state=at.BEFORE_DOCTYPE_NAME;break}case Ze.GREATER_THAN_SIGN:{this.state=at.BEFORE_DOCTYPE_NAME,this._stateBeforeDoctypeName(r);break}case Ze.EOF:{this._err(On.eofInDoctype),this._createDoctypeToken(null);const o=this.currentToken;o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.missingWhitespaceBeforeDoctypeName),this.state=at.BEFORE_DOCTYPE_NAME,this._stateBeforeDoctypeName(r)}}_stateBeforeDoctypeName(r){if(eie(r))this._createDoctypeToken(String.fromCharCode(K0e(r))),this.state=at.DOCTYPE_NAME;else switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.NULL:{this._err(On.unexpectedNullCharacter),this._createDoctypeToken(Sf),this.state=at.DOCTYPE_NAME;break}case Ze.GREATER_THAN_SIGN:{this._err(On.missingDoctypeName),this._createDoctypeToken(null);const o=this.currentToken;o.forceQuirks=!0,this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.EOF:{this._err(On.eofInDoctype),this._createDoctypeToken(null);const o=this.currentToken;o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._createDoctypeToken(String.fromCodePoint(r)),this.state=at.DOCTYPE_NAME}}_stateDoctypeName(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this.state=at.AFTER_DOCTYPE_NAME;break}case Ze.GREATER_THAN_SIGN:{this.state=at.DATA,this.emitCurrentDoctype(o);break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.name+=Sf;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:o.name+=String.fromCodePoint(eie(r)?K0e(r):r)}}_stateAfterDoctypeName(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.GREATER_THAN_SIGN:{this.state=at.DATA,this.emitCurrentDoctype(o);break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._consumeSequenceIfMatch(zw.PUBLIC,!1)?this.state=at.AFTER_DOCTYPE_PUBLIC_KEYWORD:this._consumeSequenceIfMatch(zw.SYSTEM,!1)?this.state=at.AFTER_DOCTYPE_SYSTEM_KEYWORD:this._ensureHibernation()||(this._err(On.invalidCharacterSequenceAfterDoctypeName),o.forceQuirks=!0,this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r))}}_stateAfterDoctypePublicKeyword(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this.state=at.BEFORE_DOCTYPE_PUBLIC_IDENTIFIER;break}case Ze.QUOTATION_MARK:{this._err(On.missingWhitespaceAfterDoctypePublicKeyword),o.publicId="",this.state=at.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED;break}case Ze.APOSTROPHE:{this._err(On.missingWhitespaceAfterDoctypePublicKeyword),o.publicId="",this.state=at.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED;break}case Ze.GREATER_THAN_SIGN:{this._err(On.missingDoctypePublicIdentifier),o.forceQuirks=!0,this.state=at.DATA,this.emitCurrentDoctype(o);break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.missingQuoteBeforeDoctypePublicIdentifier),o.forceQuirks=!0,this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r)}}_stateBeforeDoctypePublicIdentifier(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.QUOTATION_MARK:{o.publicId="",this.state=at.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED;break}case Ze.APOSTROPHE:{o.publicId="",this.state=at.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED;break}case Ze.GREATER_THAN_SIGN:{this._err(On.missingDoctypePublicIdentifier),o.forceQuirks=!0,this.state=at.DATA,this.emitCurrentDoctype(o);break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.missingQuoteBeforeDoctypePublicIdentifier),o.forceQuirks=!0,this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r)}}_stateDoctypePublicIdentifierDoubleQuoted(r){const o=this.currentToken;switch(r){case Ze.QUOTATION_MARK:{this.state=at.AFTER_DOCTYPE_PUBLIC_IDENTIFIER;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.publicId+=Sf;break}case Ze.GREATER_THAN_SIGN:{this._err(On.abruptDoctypePublicIdentifier),o.forceQuirks=!0,this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:o.publicId+=String.fromCodePoint(r)}}_stateDoctypePublicIdentifierSingleQuoted(r){const o=this.currentToken;switch(r){case Ze.APOSTROPHE:{this.state=at.AFTER_DOCTYPE_PUBLIC_IDENTIFIER;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.publicId+=Sf;break}case Ze.GREATER_THAN_SIGN:{this._err(On.abruptDoctypePublicIdentifier),o.forceQuirks=!0,this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:o.publicId+=String.fromCodePoint(r)}}_stateAfterDoctypePublicIdentifier(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this.state=at.BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS;break}case Ze.GREATER_THAN_SIGN:{this.state=at.DATA,this.emitCurrentDoctype(o);break}case Ze.QUOTATION_MARK:{this._err(On.missingWhitespaceBetweenDoctypePublicAndSystemIdentifiers),o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break}case Ze.APOSTROPHE:{this._err(On.missingWhitespaceBetweenDoctypePublicAndSystemIdentifiers),o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.missingQuoteBeforeDoctypeSystemIdentifier),o.forceQuirks=!0,this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r)}}_stateBetweenDoctypePublicAndSystemIdentifiers(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.GREATER_THAN_SIGN:{this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.QUOTATION_MARK:{o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break}case Ze.APOSTROPHE:{o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.missingQuoteBeforeDoctypeSystemIdentifier),o.forceQuirks=!0,this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r)}}_stateAfterDoctypeSystemKeyword(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:{this.state=at.BEFORE_DOCTYPE_SYSTEM_IDENTIFIER;break}case Ze.QUOTATION_MARK:{this._err(On.missingWhitespaceAfterDoctypeSystemKeyword),o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break}case Ze.APOSTROPHE:{this._err(On.missingWhitespaceAfterDoctypeSystemKeyword),o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break}case Ze.GREATER_THAN_SIGN:{this._err(On.missingDoctypeSystemIdentifier),o.forceQuirks=!0,this.state=at.DATA,this.emitCurrentDoctype(o);break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.missingQuoteBeforeDoctypeSystemIdentifier),o.forceQuirks=!0,this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r)}}_stateBeforeDoctypeSystemIdentifier(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.QUOTATION_MARK:{o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break}case Ze.APOSTROPHE:{o.systemId="",this.state=at.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break}case Ze.GREATER_THAN_SIGN:{this._err(On.missingDoctypeSystemIdentifier),o.forceQuirks=!0,this.state=at.DATA,this.emitCurrentDoctype(o);break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.missingQuoteBeforeDoctypeSystemIdentifier),o.forceQuirks=!0,this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r)}}_stateDoctypeSystemIdentifierDoubleQuoted(r){const o=this.currentToken;switch(r){case Ze.QUOTATION_MARK:{this.state=at.AFTER_DOCTYPE_SYSTEM_IDENTIFIER;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.systemId+=Sf;break}case Ze.GREATER_THAN_SIGN:{this._err(On.abruptDoctypeSystemIdentifier),o.forceQuirks=!0,this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:o.systemId+=String.fromCodePoint(r)}}_stateDoctypeSystemIdentifierSingleQuoted(r){const o=this.currentToken;switch(r){case Ze.APOSTROPHE:{this.state=at.AFTER_DOCTYPE_SYSTEM_IDENTIFIER;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter),o.systemId+=Sf;break}case Ze.GREATER_THAN_SIGN:{this._err(On.abruptDoctypeSystemIdentifier),o.forceQuirks=!0,this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:o.systemId+=String.fromCodePoint(r)}}_stateAfterDoctypeSystemIdentifier(r){const o=this.currentToken;switch(r){case Ze.SPACE:case Ze.LINE_FEED:case Ze.TABULATION:case Ze.FORM_FEED:break;case Ze.GREATER_THAN_SIGN:{this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.EOF:{this._err(On.eofInDoctype),o.forceQuirks=!0,this.emitCurrentDoctype(o),this._emitEOFToken();break}default:this._err(On.unexpectedCharacterAfterDoctypeSystemIdentifier),this.state=at.BOGUS_DOCTYPE,this._stateBogusDoctype(r)}}_stateBogusDoctype(r){const o=this.currentToken;switch(r){case Ze.GREATER_THAN_SIGN:{this.emitCurrentDoctype(o),this.state=at.DATA;break}case Ze.NULL:{this._err(On.unexpectedNullCharacter);break}case Ze.EOF:{this.emitCurrentDoctype(o),this._emitEOFToken();break}}}_stateCdataSection(r){switch(r){case Ze.RIGHT_SQUARE_BRACKET:{this.state=at.CDATA_SECTION_BRACKET;break}case Ze.EOF:{this._err(On.eofInCdata),this._emitEOFToken();break}default:this._emitCodePoint(r)}}_stateCdataSectionBracket(r){r===Ze.RIGHT_SQUARE_BRACKET?this.state=at.CDATA_SECTION_END:(this._emitChars("]"),this.state=at.CDATA_SECTION,this._stateCdataSection(r))}_stateCdataSectionEnd(r){switch(r){case Ze.GREATER_THAN_SIGN:{this.state=at.DATA;break}case Ze.RIGHT_SQUARE_BRACKET:{this._emitChars("]");break}default:this._emitChars("]]"),this.state=at.CDATA_SECTION,this._stateCdataSection(r)}}_stateCharacterReference(){let r=this.entityDecoder.write(this.preprocessor.html,this.preprocessor.pos);if(r<0)if(this.preprocessor.lastChunkWritten)r=this.entityDecoder.end();else{this.active=!1,this.preprocessor.pos=this.preprocessor.html.length-1,this.consumedAfterSnapshot=0,this.preprocessor.endOfChunkHit=!0;return}r===0?(this.preprocessor.pos=this.entityStartPos,this._flushCodePointConsumedAsCharacterReference(Ze.AMPERSAND),this.state=!this._isCharacterReferenceInAttribute()&&YWt(this.preprocessor.peek(1))?at.AMBIGUOUS_AMPERSAND:this.returnState):this.state=this.returnState}_stateAmbiguousAmpersand(r){YWt(r)?this._flushCodePointConsumedAsCharacterReference(r):(r===Ze.SEMICOLON&&this._err(On.unknownNamedCharacterReference),this.state=this.returnState,this._callState(r))}}const zfn=new Set([re.DD,re.DT,re.LI,re.OPTGROUP,re.OPTION,re.P,re.RB,re.RP,re.RT,re.RTC]),JWt=new Set([...zfn,re.CAPTION,re.COLGROUP,re.TBODY,re.TD,re.TFOOT,re.TH,re.THEAD,re.TR]),Xwe=new Set([re.APPLET,re.CAPTION,re.HTML,re.MARQUEE,re.OBJECT,re.TABLE,re.TD,re.TEMPLATE,re.TH]),tJr=new Set([...Xwe,re.OL,re.UL]),nJr=new Set([...Xwe,re.BUTTON]),ZWt=new Set([re.ANNOTATION_XML,re.MI,re.MN,re.MO,re.MS,re.MTEXT]),QWt=new Set([re.DESC,re.FOREIGN_OBJECT,re.TITLE]),rJr=new Set([re.TR,re.TEMPLATE,re.HTML]),iJr=new Set([re.TBODY,re.TFOOT,re.THEAD,re.TEMPLATE,re.HTML]),oJr=new Set([re.TABLE,re.TEMPLATE,re.HTML]),aJr=new Set([re.TD,re.TH]);class sJr{get currentTmplContentOrNode(){return this._isInTemplate()?this.treeAdapter.getTemplateContent(this.current):this.current}constructor(r,o,s){this.treeAdapter=o,this.handler=s,this.items=[],this.tagIDs=[],this.stackTop=-1,this.tmplCount=0,this.currentTagId=re.UNKNOWN,this.current=r}_indexOf(r){return this.items.lastIndexOf(r,this.stackTop)}_isInTemplate(){return this.currentTagId===re.TEMPLATE&&this.treeAdapter.getNamespaceURI(this.current)===Er.HTML}_updateCurrentElement(){this.current=this.items[this.stackTop],this.currentTagId=this.tagIDs[this.stackTop]}push(r,o){this.stackTop++,this.items[this.stackTop]=r,this.current=r,this.tagIDs[this.stackTop]=o,this.currentTagId=o,this._isInTemplate()&&this.tmplCount++,this.handler.onItemPush(r,o,!0)}pop(){const r=this.current;this.tmplCount>0&&this._isInTemplate()&&this.tmplCount--,this.stackTop--,this._updateCurrentElement(),this.handler.onItemPop(r,!0)}replace(r,o){const s=this._indexOf(r);this.items[s]=o,s===this.stackTop&&(this.current=o)}insertAfter(r,o,s){const c=this._indexOf(r)+1;this.items.splice(c,0,o),this.tagIDs.splice(c,0,s),this.stackTop++,c===this.stackTop&&this._updateCurrentElement(),this.current&&this.currentTagId!==void 0&&this.handler.onItemPush(this.current,this.currentTagId,c===this.stackTop)}popUntilTagNamePopped(r){let o=this.stackTop+1;do o=this.tagIDs.lastIndexOf(r,o-1);while(o>0&&this.treeAdapter.getNamespaceURI(this.items[o])!==Er.HTML);this.shortenToLength(Math.max(o,0))}shortenToLength(r){for(;this.stackTop>=r;){const o=this.current;this.tmplCount>0&&this._isInTemplate()&&(this.tmplCount-=1),this.stackTop--,this._updateCurrentElement(),this.handler.onItemPop(o,this.stackTop<r)}}popUntilElementPopped(r){const o=this._indexOf(r);this.shortenToLength(Math.max(o,0))}popUntilPopped(r,o){const s=this._indexOfTagNames(r,o);this.shortenToLength(Math.max(s,0))}popUntilNumberedHeaderPopped(){this.popUntilPopped(EKe,Er.HTML)}popUntilTableCellPopped(){this.popUntilPopped(aJr,Er.HTML)}popAllUpToHtmlElement(){this.tmplCount=0,this.shortenToLength(1)}_indexOfTagNames(r,o){for(let s=this.stackTop;s>=0;s--)if(r.has(this.tagIDs[s])&&this.treeAdapter.getNamespaceURI(this.items[s])===o)return s;return-1}clearBackTo(r,o){const s=this._indexOfTagNames(r,o);this.shortenToLength(s+1)}clearBackToTableContext(){this.clearBackTo(oJr,Er.HTML)}clearBackToTableBodyContext(){this.clearBackTo(iJr,Er.HTML)}clearBackToTableRowContext(){this.clearBackTo(rJr,Er.HTML)}remove(r){const o=this._indexOf(r);o>=0&&(o===this.stackTop?this.pop():(this.items.splice(o,1),this.tagIDs.splice(o,1),this.stackTop--,this._updateCurrentElement(),this.handler.onItemPop(r,!1)))}tryPeekProperlyNestedBodyElement(){return this.stackTop>=1&&this.tagIDs[1]===re.BODY?this.items[1]:null}contains(r){return this._indexOf(r)>-1}getCommonAncestor(r){const o=this._indexOf(r)-1;return o>=0?this.items[o]:null}isRootHtmlElementCurrent(){return this.stackTop===0&&this.tagIDs[0]===re.HTML}hasInDynamicScope(r,o){for(let s=this.stackTop;s>=0;s--){const c=this.tagIDs[s];switch(this.treeAdapter.getNamespaceURI(this.items[s])){case Er.HTML:{if(c===r)return!0;if(o.has(c))return!1;break}case Er.SVG:{if(QWt.has(c))return!1;break}case Er.MATHML:{if(ZWt.has(c))return!1;break}}}return!0}hasInScope(r){return this.hasInDynamicScope(r,Xwe)}hasInListItemScope(r){return this.hasInDynamicScope(r,tJr)}hasInButtonScope(r){return this.hasInDynamicScope(r,nJr)}hasNumberedHeaderInScope(){for(let r=this.stackTop;r>=0;r--){const o=this.tagIDs[r];switch(this.treeAdapter.getNamespaceURI(this.items[r])){case Er.HTML:{if(EKe.has(o))return!0;if(Xwe.has(o))return!1;break}case Er.SVG:{if(QWt.has(o))return!1;break}case Er.MATHML:{if(ZWt.has(o))return!1;break}}}return!0}hasInTableScope(r){for(let o=this.stackTop;o>=0;o--)if(this.treeAdapter.getNamespaceURI(this.items[o])===Er.HTML)switch(this.tagIDs[o]){case r:return!0;case re.TABLE:case re.HTML:return!1}return!0}hasTableBodyContextInTableScope(){for(let r=this.stackTop;r>=0;r--)if(this.treeAdapter.getNamespaceURI(this.items[r])===Er.HTML)switch(this.tagIDs[r]){case re.TBODY:case re.THEAD:case re.TFOOT:return!0;case re.TABLE:case re.HTML:return!1}return!0}hasInSelectScope(r){for(let o=this.stackTop;o>=0;o--)if(this.treeAdapter.getNamespaceURI(this.items[o])===Er.HTML)switch(this.tagIDs[o]){case r:return!0;case re.OPTION:case re.OPTGROUP:break;default:return!1}return!0}generateImpliedEndTags(){for(;this.currentTagId!==void 0&&zfn.has(this.currentTagId);)this.pop()}generateImpliedEndTagsThoroughly(){for(;this.currentTagId!==void 0&&JWt.has(this.currentTagId);)this.pop()}generateImpliedEndTagsWithExclusion(r){for(;this.currentTagId!==void 0&&this.currentTagId!==r&&JWt.has(this.currentTagId);)this.pop()}}const Eqe=3;var hI;(function(n){n[n.Marker=0]="Marker",n[n.Element=1]="Element"})(hI||(hI={}));const eKt={type:hI.Marker};class cJr{constructor(r){this.treeAdapter=r,this.entries=[],this.bookmark=null}_getNoahArkConditionCandidates(r,o){const s=[],c=o.length,d=this.treeAdapter.getTagName(r),h=this.treeAdapter.getNamespaceURI(r);for(let b=0;b<this.entries.length;b++){const y=this.entries[b];if(y.type===hI.Marker)break;const{element:E}=y;if(this.treeAdapter.getTagName(E)===d&&this.treeAdapter.getNamespaceURI(E)===h){const k=this.treeAdapter.getAttrList(E);k.length===c&&s.push({idx:b,attrs:k})}}return s}_ensureNoahArkCondition(r){if(this.entries.length<Eqe)return;const o=this.treeAdapter.getAttrList(r),s=this._getNoahArkConditionCandidates(r,o);if(s.length<Eqe)return;const c=new Map(o.map(h=>[h.name,h.value]));let d=0;for(let h=0;h<s.length;h++){const b=s[h];b.attrs.every(y=>c.get(y.name)===y.value)&&(d+=1,d>=Eqe&&this.entries.splice(b.idx,1))}}insertMarker(){this.entries.unshift(eKt)}pushElement(r,o){this._ensureNoahArkCondition(r),this.entries.unshift({type:hI.Element,element:r,token:o})}insertElementAfterBookmark(r,o){const s=this.entries.indexOf(this.bookmark);this.entries.splice(s,0,{type:hI.Element,element:r,token:o})}removeEntry(r){const o=this.entries.indexOf(r);o!==-1&&this.entries.splice(o,1)}clearToLastMarker(){const r=this.entries.indexOf(eKt);r===-1?this.entries.length=0:this.entries.splice(0,r+1)}getElementEntryInScopeWithTagName(r){const o=this.entries.find(s=>s.type===hI.Marker||this.treeAdapter.getTagName(s.element)===r);return o&&o.type===hI.Element?o:null}getElementEntry(r){return this.entries.find(o=>o.type===hI.Element&&o.element===r)}}const tM={createDocument(){return{nodeName:"#document",mode:D_.NO_QUIRKS,childNodes:[]}},createDocumentFragment(){return{nodeName:"#document-fragment",childNodes:[]}},createElement(n,r,o){return{nodeName:n,tagName:n,attrs:o,namespaceURI:r,childNodes:[],parentNode:null}},createCommentNode(n){return{nodeName:"#comment",data:n,parentNode:null}},createTextNode(n){return{nodeName:"#text",value:n,parentNode:null}},appendChild(n,r){n.childNodes.push(r),r.parentNode=n},insertBefore(n,r,o){const s=n.childNodes.indexOf(o);n.childNodes.splice(s,0,r),r.parentNode=n},setTemplateContent(n,r){n.content=r},getTemplateContent(n){return n.content},setDocumentType(n,r,o,s){const c=n.childNodes.find(d=>d.nodeName==="#documentType");if(c)c.name=r,c.publicId=o,c.systemId=s;else{const d={nodeName:"#documentType",name:r,publicId:o,systemId:s,parentNode:null};tM.appendChild(n,d)}},setDocumentMode(n,r){n.mode=r},getDocumentMode(n){return n.mode},detachNode(n){if(n.parentNode){const r=n.parentNode.childNodes.indexOf(n);n.parentNode.childNodes.splice(r,1),n.parentNode=null}},insertText(n,r){if(n.childNodes.length>0){const o=n.childNodes[n.childNodes.length-1];if(tM.isTextNode(o)){o.value+=r;return}}tM.appendChild(n,tM.createTextNode(r))},insertTextBefore(n,r,o){const s=n.childNodes[n.childNodes.indexOf(o)-1];s&&tM.isTextNode(s)?s.value+=r:tM.insertBefore(n,tM.createTextNode(r),o)},adoptAttributes(n,r){const o=new Set(n.attrs.map(s=>s.name));for(let s=0;s<r.length;s++)o.has(r[s].name)||n.attrs.push(r[s])},getFirstChild(n){return n.childNodes[0]},getChildNodes(n){return n.childNodes},getParentNode(n){return n.parentNode},getAttrList(n){return n.attrs},getTagName(n){return n.tagName},getNamespaceURI(n){return n.namespaceURI},getTextNodeContent(n){return n.value},getCommentNodeContent(n){return n.data},getDocumentTypeNodeName(n){return n.name},getDocumentTypeNodePublicId(n){return n.publicId},getDocumentTypeNodeSystemId(n){return n.systemId},isTextNode(n){return n.nodeName==="#text"},isCommentNode(n){return n.nodeName==="#comment"},isDocumentTypeNode(n){return n.nodeName==="#documentType"},isElementNode(n){return Object.prototype.hasOwnProperty.call(n,"tagName")},setNodeSourceCodeLocation(n,r){n.sourceCodeLocation=r},getNodeSourceCodeLocation(n){return n.sourceCodeLocation},updateNodeSourceCodeLocation(n,r){n.sourceCodeLocation={...n.sourceCodeLocation,...r}}},Hfn="html",uJr="about:legacy-compat",lJr="http://www.ibm.com/data/dtd/v11/ibmxhtml1-transitional.dtd",Gfn=["+//silmaril//dtd html pro v0r11 19970101//","-//as//dtd html 3.0 aswedit + extensions//","-//advasoft ltd//dtd html 3.0 aswedit + extensions//","-//ietf//dtd html 2.0 level 1//","-//ietf//dtd html 2.0 level 2//","-//ietf//dtd html 2.0 strict level 1//","-//ietf//dtd html 2.0 strict level 2//","-//ietf//dtd html 2.0 strict//","-//ietf//dtd html 2.0//","-//ietf//dtd html 2.1e//","-//ietf//dtd html 3.0//","-//ietf//dtd html 3.2 final//","-//ietf//dtd html 3.2//","-//ietf//dtd html 3//","-//ietf//dtd html level 0//","-//ietf//dtd html level 1//","-//ietf//dtd html level 2//","-//ietf//dtd html level 3//","-//ietf//dtd html strict level 0//","-//ietf//dtd html strict level 1//","-//ietf//dtd html strict level 2//","-//ietf//dtd html strict level 3//","-//ietf//dtd html strict//","-//ietf//dtd html//","-//metrius//dtd metrius presentational//","-//microsoft//dtd internet explorer 2.0 html strict//","-//microsoft//dtd internet explorer 2.0 html//","-//microsoft//dtd internet explorer 2.0 tables//","-//microsoft//dtd internet explorer 3.0 html strict//","-//microsoft//dtd internet explorer 3.0 html//","-//microsoft//dtd internet explorer 3.0 tables//","-//netscape comm. corp.//dtd html//","-//netscape comm. corp.//dtd strict html//","-//o'reilly and associates//dtd html 2.0//","-//o'reilly and associates//dtd html extended 1.0//","-//o'reilly and associates//dtd html extended relaxed 1.0//","-//sq//dtd html 2.0 hotmetal + extensions//","-//softquad software//dtd hotmetal pro 6.0::19990601::extensions to html 4.0//","-//softquad//dtd hotmetal pro 4.0::19971010::extensions to html 4.0//","-//spyglass//dtd html 2.0 extended//","-//sun microsystems corp.//dtd hotjava html//","-//sun microsystems corp.//dtd hotjava strict html//","-//w3c//dtd html 3 1995-03-24//","-//w3c//dtd html 3.2 draft//","-//w3c//dtd html 3.2 final//","-//w3c//dtd html 3.2//","-//w3c//dtd html 3.2s draft//","-//w3c//dtd html 4.0 frameset//","-//w3c//dtd html 4.0 transitional//","-//w3c//dtd html experimental 19960712//","-//w3c//dtd html experimental 970421//","-//w3c//dtd w3 html//","-//w3o//dtd w3 html 3.0//","-//webtechs//dtd mozilla html 2.0//","-//webtechs//dtd mozilla html//"],dJr=[...Gfn,"-//w3c//dtd html 4.01 frameset//","-//w3c//dtd html 4.01 transitional//"],fJr=new Set(["-//w3o//dtd w3 html strict 3.0//en//","-/w3c/dtd html 4.0 transitional/en","html"]),qfn=["-//w3c//dtd xhtml 1.0 frameset//","-//w3c//dtd xhtml 1.0 transitional//"],hJr=[...qfn,"-//w3c//dtd html 4.01 frameset//","-//w3c//dtd html 4.01 transitional//"];function tKt(n,r){return r.some(o=>n.startsWith(o))}function pJr(n){return n.name===Hfn&&n.publicId===null&&(n.systemId===null||n.systemId===uJr)}function gJr(n){if(n.name!==Hfn)return D_.QUIRKS;const{systemId:r}=n;if(r&&r.toLowerCase()===lJr)return D_.QUIRKS;let{publicId:o}=n;if(o!==null){if(o=o.toLowerCase(),fJr.has(o))return D_.QUIRKS;let s=r===null?dJr:Gfn;if(tKt(o,s))return D_.QUIRKS;if(s=r===null?qfn:hJr,tKt(o,s))return D_.LIMITED_QUIRKS}return D_.NO_QUIRKS}const nKt={TEXT_HTML:"text/html",APPLICATION_XML:"application/xhtml+xml"},mJr="definitionurl",bJr="definitionURL",vJr=new Map(["attributeName","attributeType","baseFrequency","baseProfile","calcMode","clipPathUnits","diffuseConstant","edgeMode","filterUnits","glyphRef","gradientTransform","gradientUnits","kernelMatrix","kernelUnitLength","keyPoints","keySplines","keyTimes","lengthAdjust","limitingConeAngle","markerHeight","markerUnits","markerWidth","maskContentUnits","maskUnits","numOctaves","pathLength","patternContentUnits","patternTransform","patternUnits","pointsAtX","pointsAtY","pointsAtZ","preserveAlpha","preserveAspectRatio","primitiveUnits","refX","refY","repeatCount","repeatDur","requiredExtensions","requiredFeatures","specularConstant","specularExponent","spreadMethod","startOffset","stdDeviation","stitchTiles","surfaceScale","systemLanguage","tableValues","targetX","targetY","textLength","viewBox","viewTarget","xChannelSelector","yChannelSelector","zoomAndPan"].map(n=>[n.toLowerCase(),n])),wJr=new Map([["xlink:actuate",{prefix:"xlink",name:"actuate",namespace:Er.XLINK}],["xlink:arcrole",{prefix:"xlink",name:"arcrole",namespace:Er.XLINK}],["xlink:href",{prefix:"xlink",name:"href",namespace:Er.XLINK}],["xlink:role",{prefix:"xlink",name:"role",namespace:Er.XLINK}],["xlink:show",{prefix:"xlink",name:"show",namespace:Er.XLINK}],["xlink:title",{prefix:"xlink",name:"title",namespace:Er.XLINK}],["xlink:type",{prefix:"xlink",name:"type",namespace:Er.XLINK}],["xml:lang",{prefix:"xml",name:"lang",namespace:Er.XML}],["xml:space",{prefix:"xml",name:"space",namespace:Er.XML}],["xmlns",{prefix:"",name:"xmlns",namespace:Er.XMLNS}],["xmlns:xlink",{prefix:"xmlns",name:"xlink",namespace:Er.XMLNS}]]),yJr=new Map(["altGlyph","altGlyphDef","altGlyphItem","animateColor","animateMotion","animateTransform","clipPath","feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feDistantLight","feFlood","feFuncA","feFuncB","feFuncG","feFuncR","feGaussianBlur","feImage","feMerge","feMergeNode","feMorphology","feOffset","fePointLight","feSpecularLighting","feSpotLight","feTile","feTurbulence","foreignObject","glyphRef","linearGradient","radialGradient","textPath"].map(n=>[n.toLowerCase(),n])),EJr=new Set([re.B,re.BIG,re.BLOCKQUOTE,re.BODY,re.BR,re.CENTER,re.CODE,re.DD,re.DIV,re.DL,re.DT,re.EM,re.EMBED,re.H1,re.H2,re.H3,re.H4,re.H5,re.H6,re.HEAD,re.HR,re.I,re.IMG,re.LI,re.LISTING,re.MENU,re.META,re.NOBR,re.OL,re.P,re.PRE,re.RUBY,re.S,re.SMALL,re.SPAN,re.STRONG,re.STRIKE,re.SUB,re.SUP,re.TABLE,re.TT,re.U,re.UL,re.VAR]);function SJr(n){const r=n.tagID;return r===re.FONT&&n.attrs.some(({name:s})=>s===GB.COLOR||s===GB.SIZE||s===GB.FACE)||EJr.has(r)}function Vfn(n){for(let r=0;r<n.attrs.length;r++)if(n.attrs[r].name===mJr){n.attrs[r].name=bJr;break}}function Wfn(n){for(let r=0;r<n.attrs.length;r++){const o=vJr.get(n.attrs[r].name);o!=null&&(n.attrs[r].name=o)}}function WQe(n){for(let r=0;r<n.attrs.length;r++){const o=wJr.get(n.attrs[r].name);o&&(n.attrs[r].prefix=o.prefix,n.attrs[r].name=o.name,n.attrs[r].namespace=o.namespace)}}function _Jr(n){const r=yJr.get(n.tagName);r!=null&&(n.tagName=r,n.tagID=zK(n.tagName))}function kJr(n,r){return r===Er.MATHML&&(n===re.MI||n===re.MO||n===re.MN||n===re.MS||n===re.MTEXT)}function TJr(n,r,o){if(r===Er.MATHML&&n===re.ANNOTATION_XML){for(let s=0;s<o.length;s++)if(o[s].name===GB.ENCODING){const c=o[s].value.toLowerCase();return c===nKt.TEXT_HTML||c===nKt.APPLICATION_XML}}return r===Er.SVG&&(n===re.FOREIGN_OBJECT||n===re.DESC||n===re.TITLE)}function xJr(n,r,o,s){return(!s||s===Er.HTML)&&TJr(n,r,o)||(!s||s===Er.MATHML)&&kJr(n,r)}const CJr="hidden",AJr=8,IJr=3;var gt;(function(n){n[n.INITIAL=0]="INITIAL",n[n.BEFORE_HTML=1]="BEFORE_HTML",n[n.BEFORE_HEAD=2]="BEFORE_HEAD",n[n.IN_HEAD=3]="IN_HEAD",n[n.IN_HEAD_NO_SCRIPT=4]="IN_HEAD_NO_SCRIPT",n[n.AFTER_HEAD=5]="AFTER_HEAD",n[n.IN_BODY=6]="IN_BODY",n[n.TEXT=7]="TEXT",n[n.IN_TABLE=8]="IN_TABLE",n[n.IN_TABLE_TEXT=9]="IN_TABLE_TEXT",n[n.IN_CAPTION=10]="IN_CAPTION",n[n.IN_COLUMN_GROUP=11]="IN_COLUMN_GROUP",n[n.IN_TABLE_BODY=12]="IN_TABLE_BODY",n[n.IN_ROW=13]="IN_ROW",n[n.IN_CELL=14]="IN_CELL",n[n.IN_SELECT=15]="IN_SELECT",n[n.IN_SELECT_IN_TABLE=16]="IN_SELECT_IN_TABLE",n[n.IN_TEMPLATE=17]="IN_TEMPLATE",n[n.AFTER_BODY=18]="AFTER_BODY",n[n.IN_FRAMESET=19]="IN_FRAMESET",n[n.AFTER_FRAMESET=20]="AFTER_FRAMESET",n[n.AFTER_AFTER_BODY=21]="AFTER_AFTER_BODY",n[n.AFTER_AFTER_FRAMESET=22]="AFTER_AFTER_FRAMESET"})(gt||(gt={}));const OJr={startLine:-1,startCol:-1,startOffset:-1,endLine:-1,endCol:-1,endOffset:-1},Kfn=new Set([re.TABLE,re.TBODY,re.TFOOT,re.THEAD,re.TR]),rKt={scriptingEnabled:!0,sourceCodeLocationInfo:!1,treeAdapter:tM,onParseError:null};class iKt{constructor(r,o,s=null,c=null){this.fragmentContext=s,this.scriptHandler=c,this.currentToken=null,this.stopped=!1,this.insertionMode=gt.INITIAL,this.originalInsertionMode=gt.INITIAL,this.headElement=null,this.formElement=null,this.currentNotInHTML=!1,this.tmplInsertionModeStack=[],this.pendingCharacterTokens=[],this.hasNonWhitespacePendingCharacterToken=!1,this.framesetOk=!0,this.skipNextNewLine=!1,this.fosterParentingEnabled=!1,this.options={...rKt,...r},this.treeAdapter=this.options.treeAdapter,this.onParseError=this.options.onParseError,this.onParseError&&(this.options.sourceCodeLocationInfo=!0),this.document=o??this.treeAdapter.createDocument(),this.tokenizer=new eJr(this.options,this),this.activeFormattingElements=new cJr(this.treeAdapter),this.fragmentContextID=s?zK(this.treeAdapter.getTagName(s)):re.UNKNOWN,this._setContextModes(s??this.document,this.fragmentContextID),this.openElements=new sJr(this.document,this.treeAdapter,this)}static parse(r,o){const s=new this(o);return s.tokenizer.write(r,!0),s.document}static getFragmentParser(r,o){const s={...rKt,...o};r??(r=s.treeAdapter.createElement(nn.TEMPLATE,Er.HTML,[]));const c=s.treeAdapter.createElement("documentmock",Er.HTML,[]),d=new this(s,c,r);return d.fragmentContextID===re.TEMPLATE&&d.tmplInsertionModeStack.unshift(gt.IN_TEMPLATE),d._initTokenizerForFragmentParsing(),d._insertFakeRootElement(),d._resetInsertionMode(),d._findFormInFragmentContext(),d}getFragment(){const r=this.treeAdapter.getFirstChild(this.document),o=this.treeAdapter.createDocumentFragment();return this._adoptNodes(r,o),o}_err(r,o,s){var c;if(!this.onParseError)return;const d=(c=r.location)!==null&&c!==void 0?c:OJr,h={code:o,startLine:d.startLine,startCol:d.startCol,startOffset:d.startOffset,endLine:s?d.startLine:d.endLine,endCol:s?d.startCol:d.endCol,endOffset:s?d.startOffset:d.endOffset};this.onParseError(h)}onItemPush(r,o,s){var c,d;(d=(c=this.treeAdapter).onItemPush)===null||d===void 0||d.call(c,r),s&&this.openElements.stackTop>0&&this._setContextModes(r,o)}onItemPop(r,o){var s,c;if(this.options.sourceCodeLocationInfo&&this._setEndLocation(r,this.currentToken),(c=(s=this.treeAdapter).onItemPop)===null||c===void 0||c.call(s,r,this.openElements.current),o){let d,h;this.openElements.stackTop===0&&this.fragmentContext?(d=this.fragmentContext,h=this.fragmentContextID):{current:d,currentTagId:h}=this.openElements,this._setContextModes(d,h)}}_setContextModes(r,o){const s=r===this.document||r&&this.treeAdapter.getNamespaceURI(r)===Er.HTML;this.currentNotInHTML=!s,this.tokenizer.inForeignNode=!s&&r!==void 0&&o!==void 0&&!this._isIntegrationPoint(o,r)}_switchToTextParsing(r,o){this._insertElement(r,Er.HTML),this.tokenizer.state=o,this.originalInsertionMode=this.insertionMode,this.insertionMode=gt.TEXT}switchToPlaintextParsing(){this.insertionMode=gt.TEXT,this.originalInsertionMode=gt.IN_BODY,this.tokenizer.state=gp.PLAINTEXT}_getAdjustedCurrentElement(){return this.openElements.stackTop===0&&this.fragmentContext?this.fragmentContext:this.openElements.current}_findFormInFragmentContext(){let r=this.fragmentContext;for(;r;){if(this.treeAdapter.getTagName(r)===nn.FORM){this.formElement=r;break}r=this.treeAdapter.getParentNode(r)}}_initTokenizerForFragmentParsing(){if(!(!this.fragmentContext||this.treeAdapter.getNamespaceURI(this.fragmentContext)!==Er.HTML))switch(this.fragmentContextID){case re.TITLE:case re.TEXTAREA:{this.tokenizer.state=gp.RCDATA;break}case re.STYLE:case re.XMP:case re.IFRAME:case re.NOEMBED:case re.NOFRAMES:case re.NOSCRIPT:{this.tokenizer.state=gp.RAWTEXT;break}case re.SCRIPT:{this.tokenizer.state=gp.SCRIPT_DATA;break}case re.PLAINTEXT:{this.tokenizer.state=gp.PLAINTEXT;break}}}_setDocumentType(r){const o=r.name||"",s=r.publicId||"",c=r.systemId||"";if(this.treeAdapter.setDocumentType(this.document,o,s,c),r.location){const h=this.treeAdapter.getChildNodes(this.document).find(b=>this.treeAdapter.isDocumentTypeNode(b));h&&this.treeAdapter.setNodeSourceCodeLocation(h,r.location)}}_attachElementToTree(r,o){if(this.options.sourceCodeLocationInfo){const s=o&&{...o,startTag:o};this.treeAdapter.setNodeSourceCodeLocation(r,s)}if(this._shouldFosterParentOnInsertion())this._fosterParentElement(r);else{const s=this.openElements.currentTmplContentOrNode;this.treeAdapter.appendChild(s??this.document,r)}}_appendElement(r,o){const s=this.treeAdapter.createElement(r.tagName,o,r.attrs);this._attachElementToTree(s,r.location)}_insertElement(r,o){const s=this.treeAdapter.createElement(r.tagName,o,r.attrs);this._attachElementToTree(s,r.location),this.openElements.push(s,r.tagID)}_insertFakeElement(r,o){const s=this.treeAdapter.createElement(r,Er.HTML,[]);this._attachElementToTree(s,null),this.openElements.push(s,o)}_insertTemplate(r){const o=this.treeAdapter.createElement(r.tagName,Er.HTML,r.attrs),s=this.treeAdapter.createDocumentFragment();this.treeAdapter.setTemplateContent(o,s),this._attachElementToTree(o,r.location),this.openElements.push(o,r.tagID),this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(s,null)}_insertFakeRootElement(){const r=this.treeAdapter.createElement(nn.HTML,Er.HTML,[]);this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(r,null),this.treeAdapter.appendChild(this.openElements.current,r),this.openElements.push(r,re.HTML)}_appendCommentNode(r,o){const s=this.treeAdapter.createCommentNode(r.data);this.treeAdapter.appendChild(o,s),this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(s,r.location)}_insertCharacters(r){let o,s;if(this._shouldFosterParentOnInsertion()?({parent:o,beforeElement:s}=this._findFosterParentingLocation(),s?this.treeAdapter.insertTextBefore(o,r.chars,s):this.treeAdapter.insertText(o,r.chars)):(o=this.openElements.currentTmplContentOrNode,this.treeAdapter.insertText(o,r.chars)),!r.location)return;const c=this.treeAdapter.getChildNodes(o),d=s?c.lastIndexOf(s):c.length,h=c[d-1];if(this.treeAdapter.getNodeSourceCodeLocation(h)){const{endLine:y,endCol:E,endOffset:k}=r.location;this.treeAdapter.updateNodeSourceCodeLocation(h,{endLine:y,endCol:E,endOffset:k})}else this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(h,r.location)}_adoptNodes(r,o){for(let s=this.treeAdapter.getFirstChild(r);s;s=this.treeAdapter.getFirstChild(r))this.treeAdapter.detachNode(s),this.treeAdapter.appendChild(o,s)}_setEndLocation(r,o){if(this.treeAdapter.getNodeSourceCodeLocation(r)&&o.location){const s=o.location,c=this.treeAdapter.getTagName(r),d=o.type===nc.END_TAG&&c===o.tagName?{endTag:{...s},endLine:s.endLine,endCol:s.endCol,endOffset:s.endOffset}:{endLine:s.startLine,endCol:s.startCol,endOffset:s.startOffset};this.treeAdapter.updateNodeSourceCodeLocation(r,d)}}shouldProcessStartTagTokenInForeignContent(r){if(!this.currentNotInHTML)return!1;let o,s;return this.openElements.stackTop===0&&this.fragmentContext?(o=this.fragmentContext,s=this.fragmentContextID):{current:o,currentTagId:s}=this.openElements,r.tagID===re.SVG&&this.treeAdapter.getTagName(o)===nn.ANNOTATION_XML&&this.treeAdapter.getNamespaceURI(o)===Er.MATHML?!1:this.tokenizer.inForeignNode||(r.tagID===re.MGLYPH||r.tagID===re.MALIGNMARK)&&s!==void 0&&!this._isIntegrationPoint(s,o,Er.HTML)}_processToken(r){switch(r.type){case nc.CHARACTER:{this.onCharacter(r);break}case nc.NULL_CHARACTER:{this.onNullCharacter(r);break}case nc.COMMENT:{this.onComment(r);break}case nc.DOCTYPE:{this.onDoctype(r);break}case nc.START_TAG:{this._processStartTag(r);break}case nc.END_TAG:{this.onEndTag(r);break}case nc.EOF:{this.onEof(r);break}case nc.WHITESPACE_CHARACTER:{this.onWhitespaceCharacter(r);break}}}_isIntegrationPoint(r,o,s){const c=this.treeAdapter.getNamespaceURI(o),d=this.treeAdapter.getAttrList(o);return xJr(r,c,d,s)}_reconstructActiveFormattingElements(){const r=this.activeFormattingElements.entries.length;if(r){const o=this.activeFormattingElements.entries.findIndex(c=>c.type===hI.Marker||this.openElements.contains(c.element)),s=o===-1?r-1:o-1;for(let c=s;c>=0;c--){const d=this.activeFormattingElements.entries[c];this._insertElement(d.token,this.treeAdapter.getNamespaceURI(d.element)),d.element=this.openElements.current}}}_closeTableCell(){this.openElements.generateImpliedEndTags(),this.openElements.popUntilTableCellPopped(),this.activeFormattingElements.clearToLastMarker(),this.insertionMode=gt.IN_ROW}_closePElement(){this.openElements.generateImpliedEndTagsWithExclusion(re.P),this.openElements.popUntilTagNamePopped(re.P)}_resetInsertionMode(){for(let r=this.openElements.stackTop;r>=0;r--)switch(r===0&&this.fragmentContext?this.fragmentContextID:this.openElements.tagIDs[r]){case re.TR:{this.insertionMode=gt.IN_ROW;return}case re.TBODY:case re.THEAD:case re.TFOOT:{this.insertionMode=gt.IN_TABLE_BODY;return}case re.CAPTION:{this.insertionMode=gt.IN_CAPTION;return}case re.COLGROUP:{this.insertionMode=gt.IN_COLUMN_GROUP;return}case re.TABLE:{this.insertionMode=gt.IN_TABLE;return}case re.BODY:{this.insertionMode=gt.IN_BODY;return}case re.FRAMESET:{this.insertionMode=gt.IN_FRAMESET;return}case re.SELECT:{this._resetInsertionModeForSelect(r);return}case re.TEMPLATE:{this.insertionMode=this.tmplInsertionModeStack[0];return}case re.HTML:{this.insertionMode=this.headElement?gt.AFTER_HEAD:gt.BEFORE_HEAD;return}case re.TD:case re.TH:{if(r>0){this.insertionMode=gt.IN_CELL;return}break}case re.HEAD:{if(r>0){this.insertionMode=gt.IN_HEAD;return}break}}this.insertionMode=gt.IN_BODY}_resetInsertionModeForSelect(r){if(r>0)for(let o=r-1;o>0;o--){const s=this.openElements.tagIDs[o];if(s===re.TEMPLATE)break;if(s===re.TABLE){this.insertionMode=gt.IN_SELECT_IN_TABLE;return}}this.insertionMode=gt.IN_SELECT}_isElementCausesFosterParenting(r){return Kfn.has(r)}_shouldFosterParentOnInsertion(){return this.fosterParentingEnabled&&this.openElements.currentTagId!==void 0&&this._isElementCausesFosterParenting(this.openElements.currentTagId)}_findFosterParentingLocation(){for(let r=this.openElements.stackTop;r>=0;r--){const o=this.openElements.items[r];switch(this.openElements.tagIDs[r]){case re.TEMPLATE:{if(this.treeAdapter.getNamespaceURI(o)===Er.HTML)return{parent:this.treeAdapter.getTemplateContent(o),beforeElement:null};break}case re.TABLE:{const s=this.treeAdapter.getParentNode(o);return s?{parent:s,beforeElement:o}:{parent:this.openElements.items[r-1],beforeElement:null}}}}return{parent:this.openElements.items[0],beforeElement:null}}_fosterParentElement(r){const o=this._findFosterParentingLocation();o.beforeElement?this.treeAdapter.insertBefore(o.parent,r,o.beforeElement):this.treeAdapter.appendChild(o.parent,r)}_isSpecialElement(r,o){const s=this.treeAdapter.getNamespaceURI(r);return XXr[s].has(o)}onCharacter(r){if(this.skipNextNewLine=!1,this.tokenizer.inForeignNode){aQr(this,r);return}switch(this.insertionMode){case gt.INITIAL:{Are(this,r);break}case gt.BEFORE_HTML:{_ie(this,r);break}case gt.BEFORE_HEAD:{kie(this,r);break}case gt.IN_HEAD:{Tie(this,r);break}case gt.IN_HEAD_NO_SCRIPT:{xie(this,r);break}case gt.AFTER_HEAD:{Cie(this,r);break}case gt.IN_BODY:case gt.IN_CAPTION:case gt.IN_CELL:case gt.IN_TEMPLATE:{Xfn(this,r);break}case gt.TEXT:case gt.IN_SELECT:case gt.IN_SELECT_IN_TABLE:{this._insertCharacters(r);break}case gt.IN_TABLE:case gt.IN_TABLE_BODY:case gt.IN_ROW:{Sqe(this,r);break}case gt.IN_TABLE_TEXT:{nhn(this,r);break}case gt.IN_COLUMN_GROUP:{Jwe(this,r);break}case gt.AFTER_BODY:{Zwe(this,r);break}case gt.AFTER_AFTER_BODY:{Dve(this,r);break}}}onNullCharacter(r){if(this.skipNextNewLine=!1,this.tokenizer.inForeignNode){oQr(this,r);return}switch(this.insertionMode){case gt.INITIAL:{Are(this,r);break}case gt.BEFORE_HTML:{_ie(this,r);break}case gt.BEFORE_HEAD:{kie(this,r);break}case gt.IN_HEAD:{Tie(this,r);break}case gt.IN_HEAD_NO_SCRIPT:{xie(this,r);break}case gt.AFTER_HEAD:{Cie(this,r);break}case gt.TEXT:{this._insertCharacters(r);break}case gt.IN_TABLE:case gt.IN_TABLE_BODY:case gt.IN_ROW:{Sqe(this,r);break}case gt.IN_COLUMN_GROUP:{Jwe(this,r);break}case gt.AFTER_BODY:{Zwe(this,r);break}case gt.AFTER_AFTER_BODY:{Dve(this,r);break}}}onComment(r){if(this.skipNextNewLine=!1,this.currentNotInHTML){SKe(this,r);return}switch(this.insertionMode){case gt.INITIAL:case gt.BEFORE_HTML:case gt.BEFORE_HEAD:case gt.IN_HEAD:case gt.IN_HEAD_NO_SCRIPT:case gt.AFTER_HEAD:case gt.IN_BODY:case gt.IN_TABLE:case gt.IN_CAPTION:case gt.IN_COLUMN_GROUP:case gt.IN_TABLE_BODY:case gt.IN_ROW:case gt.IN_CELL:case gt.IN_SELECT:case gt.IN_SELECT_IN_TABLE:case gt.IN_TEMPLATE:case gt.IN_FRAMESET:case gt.AFTER_FRAMESET:{SKe(this,r);break}case gt.IN_TABLE_TEXT:{Ire(this,r);break}case gt.AFTER_BODY:{jJr(this,r);break}case gt.AFTER_AFTER_BODY:case gt.AFTER_AFTER_FRAMESET:{$Jr(this,r);break}}}onDoctype(r){switch(this.skipNextNewLine=!1,this.insertionMode){case gt.INITIAL:{FJr(this,r);break}case gt.BEFORE_HEAD:case gt.IN_HEAD:case gt.IN_HEAD_NO_SCRIPT:case gt.AFTER_HEAD:{this._err(r,On.misplacedDoctype);break}case gt.IN_TABLE_TEXT:{Ire(this,r);break}}}onStartTag(r){this.skipNextNewLine=!1,this.currentToken=r,this._processStartTag(r),r.selfClosing&&!r.ackSelfClosing&&this._err(r,On.nonVoidHtmlElementStartTagWithTrailingSolidus)}_processStartTag(r){this.shouldProcessStartTagTokenInForeignContent(r)?sQr(this,r):this._startTagOutsideForeignContent(r)}_startTagOutsideForeignContent(r){switch(this.insertionMode){case gt.INITIAL:{Are(this,r);break}case gt.BEFORE_HTML:{BJr(this,r);break}case gt.BEFORE_HEAD:{zJr(this,r);break}case gt.IN_HEAD:{nC(this,r);break}case gt.IN_HEAD_NO_SCRIPT:{qJr(this,r);break}case gt.AFTER_HEAD:{WJr(this,r);break}case gt.IN_BODY:{Q1(this,r);break}case gt.IN_TABLE:{pK(this,r);break}case gt.IN_TABLE_TEXT:{Ire(this,r);break}case gt.IN_CAPTION:{HZr(this,r);break}case gt.IN_COLUMN_GROUP:{XQe(this,r);break}case gt.IN_TABLE_BODY:{XEe(this,r);break}case gt.IN_ROW:{JEe(this,r);break}case gt.IN_CELL:{VZr(this,r);break}case gt.IN_SELECT:{ohn(this,r);break}case gt.IN_SELECT_IN_TABLE:{KZr(this,r);break}case gt.IN_TEMPLATE:{XZr(this,r);break}case gt.AFTER_BODY:{ZZr(this,r);break}case gt.IN_FRAMESET:{QZr(this,r);break}case gt.AFTER_FRAMESET:{tQr(this,r);break}case gt.AFTER_AFTER_BODY:{rQr(this,r);break}case gt.AFTER_AFTER_FRAMESET:{iQr(this,r);break}}}onEndTag(r){this.skipNextNewLine=!1,this.currentToken=r,this.currentNotInHTML?cQr(this,r):this._endTagOutsideForeignContent(r)}_endTagOutsideForeignContent(r){switch(this.insertionMode){case gt.INITIAL:{Are(this,r);break}case gt.BEFORE_HTML:{UJr(this,r);break}case gt.BEFORE_HEAD:{HJr(this,r);break}case gt.IN_HEAD:{GJr(this,r);break}case gt.IN_HEAD_NO_SCRIPT:{VJr(this,r);break}case gt.AFTER_HEAD:{KJr(this,r);break}case gt.IN_BODY:{YEe(this,r);break}case gt.TEXT:{PZr(this,r);break}case gt.IN_TABLE:{moe(this,r);break}case gt.IN_TABLE_TEXT:{Ire(this,r);break}case gt.IN_CAPTION:{GZr(this,r);break}case gt.IN_COLUMN_GROUP:{qZr(this,r);break}case gt.IN_TABLE_BODY:{_Ke(this,r);break}case gt.IN_ROW:{ihn(this,r);break}case gt.IN_CELL:{WZr(this,r);break}case gt.IN_SELECT:{ahn(this,r);break}case gt.IN_SELECT_IN_TABLE:{YZr(this,r);break}case gt.IN_TEMPLATE:{JZr(this,r);break}case gt.AFTER_BODY:{chn(this,r);break}case gt.IN_FRAMESET:{eQr(this,r);break}case gt.AFTER_FRAMESET:{nQr(this,r);break}case gt.AFTER_AFTER_BODY:{Dve(this,r);break}}}onEof(r){switch(this.insertionMode){case gt.INITIAL:{Are(this,r);break}case gt.BEFORE_HTML:{_ie(this,r);break}case gt.BEFORE_HEAD:{kie(this,r);break}case gt.IN_HEAD:{Tie(this,r);break}case gt.IN_HEAD_NO_SCRIPT:{xie(this,r);break}case gt.AFTER_HEAD:{Cie(this,r);break}case gt.IN_BODY:case gt.IN_TABLE:case gt.IN_CAPTION:case gt.IN_COLUMN_GROUP:case gt.IN_TABLE_BODY:case gt.IN_ROW:case gt.IN_CELL:case gt.IN_SELECT:case gt.IN_SELECT_IN_TABLE:{ehn(this,r);break}case gt.TEXT:{LZr(this,r);break}case gt.IN_TABLE_TEXT:{Ire(this,r);break}case gt.IN_TEMPLATE:{shn(this,r);break}case gt.AFTER_BODY:case gt.IN_FRAMESET:case gt.AFTER_FRAMESET:case gt.AFTER_AFTER_BODY:case gt.AFTER_AFTER_FRAMESET:{YQe(this,r);break}}}onWhitespaceCharacter(r){if(this.skipNextNewLine&&(this.skipNextNewLine=!1,r.chars.charCodeAt(0)===Ze.LINE_FEED)){if(r.chars.length===1)return;r.chars=r.chars.substr(1)}if(this.tokenizer.inForeignNode){this._insertCharacters(r);return}switch(this.insertionMode){case gt.IN_HEAD:case gt.IN_HEAD_NO_SCRIPT:case gt.AFTER_HEAD:case gt.TEXT:case gt.IN_COLUMN_GROUP:case gt.IN_SELECT:case gt.IN_SELECT_IN_TABLE:case gt.IN_FRAMESET:case gt.AFTER_FRAMESET:{this._insertCharacters(r);break}case gt.IN_BODY:case gt.IN_CAPTION:case gt.IN_CELL:case gt.IN_TEMPLATE:case gt.AFTER_BODY:case gt.AFTER_AFTER_BODY:case gt.AFTER_AFTER_FRAMESET:{Yfn(this,r);break}case gt.IN_TABLE:case gt.IN_TABLE_BODY:case gt.IN_ROW:{Sqe(this,r);break}case gt.IN_TABLE_TEXT:{thn(this,r);break}}}}function NJr(n,r){let o=n.activeFormattingElements.getElementEntryInScopeWithTagName(r.tagName);return o?n.openElements.contains(o.element)?n.openElements.hasInScope(r.tagID)||(o=null):(n.activeFormattingElements.removeEntry(o),o=null):Qfn(n,r),o}function RJr(n,r){let o=null,s=n.openElements.stackTop;for(;s>=0;s--){const c=n.openElements.items[s];if(c===r.element)break;n._isSpecialElement(c,n.openElements.tagIDs[s])&&(o=c)}return o||(n.openElements.shortenToLength(Math.max(s,0)),n.activeFormattingElements.removeEntry(r)),o}function PJr(n,r,o){let s=r,c=n.openElements.getCommonAncestor(r);for(let d=0,h=c;h!==o;d++,h=c){c=n.openElements.getCommonAncestor(h);const b=n.activeFormattingElements.getElementEntry(h),y=b&&d>=IJr;!b||y?(y&&n.activeFormattingElements.removeEntry(b),n.openElements.remove(h)):(h=LJr(n,b),s===r&&(n.activeFormattingElements.bookmark=b),n.treeAdapter.detachNode(s),n.treeAdapter.appendChild(h,s),s=h)}return s}function LJr(n,r){const o=n.treeAdapter.getNamespaceURI(r.element),s=n.treeAdapter.createElement(r.token.tagName,o,r.token.attrs);return n.openElements.replace(r.element,s),r.element=s,s}function DJr(n,r,o){const s=n.treeAdapter.getTagName(r),c=zK(s);if(n._isElementCausesFosterParenting(c))n._fosterParentElement(o);else{const d=n.treeAdapter.getNamespaceURI(r);c===re.TEMPLATE&&d===Er.HTML&&(r=n.treeAdapter.getTemplateContent(r)),n.treeAdapter.appendChild(r,o)}}function MJr(n,r,o){const s=n.treeAdapter.getNamespaceURI(o.element),{token:c}=o,d=n.treeAdapter.createElement(c.tagName,s,c.attrs);n._adoptNodes(r,d),n.treeAdapter.appendChild(r,d),n.activeFormattingElements.insertElementAfterBookmark(d,c),n.activeFormattingElements.removeEntry(o),n.openElements.remove(o.element),n.openElements.insertAfter(r,d,c.tagID)}function KQe(n,r){for(let o=0;o<AJr;o++){const s=NJr(n,r);if(!s)break;const c=RJr(n,s);if(!c)break;n.activeFormattingElements.bookmark=s;const d=PJr(n,c,s.element),h=n.openElements.getCommonAncestor(s.element);n.treeAdapter.detachNode(d),h&&DJr(n,h,d),MJr(n,c,s)}}function SKe(n,r){n._appendCommentNode(r,n.openElements.currentTmplContentOrNode)}function jJr(n,r){n._appendCommentNode(r,n.openElements.items[0])}function $Jr(n,r){n._appendCommentNode(r,n.document)}function YQe(n,r){if(n.stopped=!0,r.location){const o=n.fragmentContext?0:2;for(let s=n.openElements.stackTop;s>=o;s--)n._setEndLocation(n.openElements.items[s],r);if(!n.fragmentContext&&n.openElements.stackTop>=0){const s=n.openElements.items[0],c=n.treeAdapter.getNodeSourceCodeLocation(s);if(c&&!c.endTag&&(n._setEndLocation(s,r),n.openElements.stackTop>=1)){const d=n.openElements.items[1],h=n.treeAdapter.getNodeSourceCodeLocation(d);h&&!h.endTag&&n._setEndLocation(d,r)}}}}function FJr(n,r){n._setDocumentType(r);const o=r.forceQuirks?D_.QUIRKS:gJr(r);pJr(r)||n._err(r,On.nonConformingDoctype),n.treeAdapter.setDocumentMode(n.document,o),n.insertionMode=gt.BEFORE_HTML}function Are(n,r){n._err(r,On.missingDoctype,!0),n.treeAdapter.setDocumentMode(n.document,D_.QUIRKS),n.insertionMode=gt.BEFORE_HTML,n._processToken(r)}function BJr(n,r){r.tagID===re.HTML?(n._insertElement(r,Er.HTML),n.insertionMode=gt.BEFORE_HEAD):_ie(n,r)}function UJr(n,r){const o=r.tagID;(o===re.HTML||o===re.HEAD||o===re.BODY||o===re.BR)&&_ie(n,r)}function _ie(n,r){n._insertFakeRootElement(),n.insertionMode=gt.BEFORE_HEAD,n._processToken(r)}function zJr(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.HEAD:{n._insertElement(r,Er.HTML),n.headElement=n.openElements.current,n.insertionMode=gt.IN_HEAD;break}default:kie(n,r)}}function HJr(n,r){const o=r.tagID;o===re.HEAD||o===re.BODY||o===re.HTML||o===re.BR?kie(n,r):n._err(r,On.endTagWithoutMatchingOpenElement)}function kie(n,r){n._insertFakeElement(nn.HEAD,re.HEAD),n.headElement=n.openElements.current,n.insertionMode=gt.IN_HEAD,n._processToken(r)}function nC(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.BASE:case re.BASEFONT:case re.BGSOUND:case re.LINK:case re.META:{n._appendElement(r,Er.HTML),r.ackSelfClosing=!0;break}case re.TITLE:{n._switchToTextParsing(r,gp.RCDATA);break}case re.NOSCRIPT:{n.options.scriptingEnabled?n._switchToTextParsing(r,gp.RAWTEXT):(n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_HEAD_NO_SCRIPT);break}case re.NOFRAMES:case re.STYLE:{n._switchToTextParsing(r,gp.RAWTEXT);break}case re.SCRIPT:{n._switchToTextParsing(r,gp.SCRIPT_DATA);break}case re.TEMPLATE:{n._insertTemplate(r),n.activeFormattingElements.insertMarker(),n.framesetOk=!1,n.insertionMode=gt.IN_TEMPLATE,n.tmplInsertionModeStack.unshift(gt.IN_TEMPLATE);break}case re.HEAD:{n._err(r,On.misplacedStartTagForHeadElement);break}default:Tie(n,r)}}function GJr(n,r){switch(r.tagID){case re.HEAD:{n.openElements.pop(),n.insertionMode=gt.AFTER_HEAD;break}case re.BODY:case re.BR:case re.HTML:{Tie(n,r);break}case re.TEMPLATE:{EU(n,r);break}default:n._err(r,On.endTagWithoutMatchingOpenElement)}}function EU(n,r){n.openElements.tmplCount>0?(n.openElements.generateImpliedEndTagsThoroughly(),n.openElements.currentTagId!==re.TEMPLATE&&n._err(r,On.closingOfElementWithOpenChildElements),n.openElements.popUntilTagNamePopped(re.TEMPLATE),n.activeFormattingElements.clearToLastMarker(),n.tmplInsertionModeStack.shift(),n._resetInsertionMode()):n._err(r,On.endTagWithoutMatchingOpenElement)}function Tie(n,r){n.openElements.pop(),n.insertionMode=gt.AFTER_HEAD,n._processToken(r)}function qJr(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.BASEFONT:case re.BGSOUND:case re.HEAD:case re.LINK:case re.META:case re.NOFRAMES:case re.STYLE:{nC(n,r);break}case re.NOSCRIPT:{n._err(r,On.nestedNoscriptInHead);break}default:xie(n,r)}}function VJr(n,r){switch(r.tagID){case re.NOSCRIPT:{n.openElements.pop(),n.insertionMode=gt.IN_HEAD;break}case re.BR:{xie(n,r);break}default:n._err(r,On.endTagWithoutMatchingOpenElement)}}function xie(n,r){const o=r.type===nc.EOF?On.openElementsLeftAfterEof:On.disallowedContentInNoscriptInHead;n._err(r,o),n.openElements.pop(),n.insertionMode=gt.IN_HEAD,n._processToken(r)}function WJr(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.BODY:{n._insertElement(r,Er.HTML),n.framesetOk=!1,n.insertionMode=gt.IN_BODY;break}case re.FRAMESET:{n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_FRAMESET;break}case re.BASE:case re.BASEFONT:case re.BGSOUND:case re.LINK:case re.META:case re.NOFRAMES:case re.SCRIPT:case re.STYLE:case re.TEMPLATE:case re.TITLE:{n._err(r,On.abandonedHeadElementChild),n.openElements.push(n.headElement,re.HEAD),nC(n,r),n.openElements.remove(n.headElement);break}case re.HEAD:{n._err(r,On.misplacedStartTagForHeadElement);break}default:Cie(n,r)}}function KJr(n,r){switch(r.tagID){case re.BODY:case re.HTML:case re.BR:{Cie(n,r);break}case re.TEMPLATE:{EU(n,r);break}default:n._err(r,On.endTagWithoutMatchingOpenElement)}}function Cie(n,r){n._insertFakeElement(nn.BODY,re.BODY),n.insertionMode=gt.IN_BODY,KEe(n,r)}function KEe(n,r){switch(r.type){case nc.CHARACTER:{Xfn(n,r);break}case nc.WHITESPACE_CHARACTER:{Yfn(n,r);break}case nc.COMMENT:{SKe(n,r);break}case nc.START_TAG:{Q1(n,r);break}case nc.END_TAG:{YEe(n,r);break}case nc.EOF:{ehn(n,r);break}}}function Yfn(n,r){n._reconstructActiveFormattingElements(),n._insertCharacters(r)}function Xfn(n,r){n._reconstructActiveFormattingElements(),n._insertCharacters(r),n.framesetOk=!1}function YJr(n,r){n.openElements.tmplCount===0&&n.treeAdapter.adoptAttributes(n.openElements.items[0],r.attrs)}function XJr(n,r){const o=n.openElements.tryPeekProperlyNestedBodyElement();o&&n.openElements.tmplCount===0&&(n.framesetOk=!1,n.treeAdapter.adoptAttributes(o,r.attrs))}function JJr(n,r){const o=n.openElements.tryPeekProperlyNestedBodyElement();n.framesetOk&&o&&(n.treeAdapter.detachNode(o),n.openElements.popAllUpToHtmlElement(),n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_FRAMESET)}function ZJr(n,r){n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._insertElement(r,Er.HTML)}function QJr(n,r){n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n.openElements.currentTagId!==void 0&&EKe.has(n.openElements.currentTagId)&&n.openElements.pop(),n._insertElement(r,Er.HTML)}function eZr(n,r){n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._insertElement(r,Er.HTML),n.skipNextNewLine=!0,n.framesetOk=!1}function tZr(n,r){const o=n.openElements.tmplCount>0;(!n.formElement||o)&&(n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._insertElement(r,Er.HTML),o||(n.formElement=n.openElements.current))}function nZr(n,r){n.framesetOk=!1;const o=r.tagID;for(let s=n.openElements.stackTop;s>=0;s--){const c=n.openElements.tagIDs[s];if(o===re.LI&&c===re.LI||(o===re.DD||o===re.DT)&&(c===re.DD||c===re.DT)){n.openElements.generateImpliedEndTagsWithExclusion(c),n.openElements.popUntilTagNamePopped(c);break}if(c!==re.ADDRESS&&c!==re.DIV&&c!==re.P&&n._isSpecialElement(n.openElements.items[s],c))break}n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._insertElement(r,Er.HTML)}function rZr(n,r){n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._insertElement(r,Er.HTML),n.tokenizer.state=gp.PLAINTEXT}function iZr(n,r){n.openElements.hasInScope(re.BUTTON)&&(n.openElements.generateImpliedEndTags(),n.openElements.popUntilTagNamePopped(re.BUTTON)),n._reconstructActiveFormattingElements(),n._insertElement(r,Er.HTML),n.framesetOk=!1}function oZr(n,r){const o=n.activeFormattingElements.getElementEntryInScopeWithTagName(nn.A);o&&(KQe(n,r),n.openElements.remove(o.element),n.activeFormattingElements.removeEntry(o)),n._reconstructActiveFormattingElements(),n._insertElement(r,Er.HTML),n.activeFormattingElements.pushElement(n.openElements.current,r)}function aZr(n,r){n._reconstructActiveFormattingElements(),n._insertElement(r,Er.HTML),n.activeFormattingElements.pushElement(n.openElements.current,r)}function sZr(n,r){n._reconstructActiveFormattingElements(),n.openElements.hasInScope(re.NOBR)&&(KQe(n,r),n._reconstructActiveFormattingElements()),n._insertElement(r,Er.HTML),n.activeFormattingElements.pushElement(n.openElements.current,r)}function cZr(n,r){n._reconstructActiveFormattingElements(),n._insertElement(r,Er.HTML),n.activeFormattingElements.insertMarker(),n.framesetOk=!1}function uZr(n,r){n.treeAdapter.getDocumentMode(n.document)!==D_.QUIRKS&&n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._insertElement(r,Er.HTML),n.framesetOk=!1,n.insertionMode=gt.IN_TABLE}function Jfn(n,r){n._reconstructActiveFormattingElements(),n._appendElement(r,Er.HTML),n.framesetOk=!1,r.ackSelfClosing=!0}function Zfn(n){const r=Bfn(n,GB.TYPE);return r!=null&&r.toLowerCase()===CJr}function lZr(n,r){n._reconstructActiveFormattingElements(),n._appendElement(r,Er.HTML),Zfn(r)||(n.framesetOk=!1),r.ackSelfClosing=!0}function dZr(n,r){n._appendElement(r,Er.HTML),r.ackSelfClosing=!0}function fZr(n,r){n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._appendElement(r,Er.HTML),n.framesetOk=!1,r.ackSelfClosing=!0}function hZr(n,r){r.tagName=nn.IMG,r.tagID=re.IMG,Jfn(n,r)}function pZr(n,r){n._insertElement(r,Er.HTML),n.skipNextNewLine=!0,n.tokenizer.state=gp.RCDATA,n.originalInsertionMode=n.insertionMode,n.framesetOk=!1,n.insertionMode=gt.TEXT}function gZr(n,r){n.openElements.hasInButtonScope(re.P)&&n._closePElement(),n._reconstructActiveFormattingElements(),n.framesetOk=!1,n._switchToTextParsing(r,gp.RAWTEXT)}function mZr(n,r){n.framesetOk=!1,n._switchToTextParsing(r,gp.RAWTEXT)}function oKt(n,r){n._switchToTextParsing(r,gp.RAWTEXT)}function bZr(n,r){n._reconstructActiveFormattingElements(),n._insertElement(r,Er.HTML),n.framesetOk=!1,n.insertionMode=n.insertionMode===gt.IN_TABLE||n.insertionMode===gt.IN_CAPTION||n.insertionMode===gt.IN_TABLE_BODY||n.insertionMode===gt.IN_ROW||n.insertionMode===gt.IN_CELL?gt.IN_SELECT_IN_TABLE:gt.IN_SELECT}function vZr(n,r){n.openElements.currentTagId===re.OPTION&&n.openElements.pop(),n._reconstructActiveFormattingElements(),n._insertElement(r,Er.HTML)}function wZr(n,r){n.openElements.hasInScope(re.RUBY)&&n.openElements.generateImpliedEndTags(),n._insertElement(r,Er.HTML)}function yZr(n,r){n.openElements.hasInScope(re.RUBY)&&n.openElements.generateImpliedEndTagsWithExclusion(re.RTC),n._insertElement(r,Er.HTML)}function EZr(n,r){n._reconstructActiveFormattingElements(),Vfn(r),WQe(r),r.selfClosing?n._appendElement(r,Er.MATHML):n._insertElement(r,Er.MATHML),r.ackSelfClosing=!0}function SZr(n,r){n._reconstructActiveFormattingElements(),Wfn(r),WQe(r),r.selfClosing?n._appendElement(r,Er.SVG):n._insertElement(r,Er.SVG),r.ackSelfClosing=!0}function aKt(n,r){n._reconstructActiveFormattingElements(),n._insertElement(r,Er.HTML)}function Q1(n,r){switch(r.tagID){case re.I:case re.S:case re.B:case re.U:case re.EM:case re.TT:case re.BIG:case re.CODE:case re.FONT:case re.SMALL:case re.STRIKE:case re.STRONG:{aZr(n,r);break}case re.A:{oZr(n,r);break}case re.H1:case re.H2:case re.H3:case re.H4:case re.H5:case re.H6:{QJr(n,r);break}case re.P:case re.DL:case re.OL:case re.UL:case re.DIV:case re.DIR:case re.NAV:case re.MAIN:case re.MENU:case re.ASIDE:case re.CENTER:case re.FIGURE:case re.FOOTER:case re.HEADER:case re.HGROUP:case re.DIALOG:case re.DETAILS:case re.ADDRESS:case re.ARTICLE:case re.SEARCH:case re.SECTION:case re.SUMMARY:case re.FIELDSET:case re.BLOCKQUOTE:case re.FIGCAPTION:{ZJr(n,r);break}case re.LI:case re.DD:case re.DT:{nZr(n,r);break}case re.BR:case re.IMG:case re.WBR:case re.AREA:case re.EMBED:case re.KEYGEN:{Jfn(n,r);break}case re.HR:{fZr(n,r);break}case re.RB:case re.RTC:{wZr(n,r);break}case re.RT:case re.RP:{yZr(n,r);break}case re.PRE:case re.LISTING:{eZr(n,r);break}case re.XMP:{gZr(n,r);break}case re.SVG:{SZr(n,r);break}case re.HTML:{YJr(n,r);break}case re.BASE:case re.LINK:case re.META:case re.STYLE:case re.TITLE:case re.SCRIPT:case re.BGSOUND:case re.BASEFONT:case re.TEMPLATE:{nC(n,r);break}case re.BODY:{XJr(n,r);break}case re.FORM:{tZr(n,r);break}case re.NOBR:{sZr(n,r);break}case re.MATH:{EZr(n,r);break}case re.TABLE:{uZr(n,r);break}case re.INPUT:{lZr(n,r);break}case re.PARAM:case re.TRACK:case re.SOURCE:{dZr(n,r);break}case re.IMAGE:{hZr(n,r);break}case re.BUTTON:{iZr(n,r);break}case re.APPLET:case re.OBJECT:case re.MARQUEE:{cZr(n,r);break}case re.IFRAME:{mZr(n,r);break}case re.SELECT:{bZr(n,r);break}case re.OPTION:case re.OPTGROUP:{vZr(n,r);break}case re.NOEMBED:case re.NOFRAMES:{oKt(n,r);break}case re.FRAMESET:{JJr(n,r);break}case re.TEXTAREA:{pZr(n,r);break}case re.NOSCRIPT:{n.options.scriptingEnabled?oKt(n,r):aKt(n,r);break}case re.PLAINTEXT:{rZr(n,r);break}case re.COL:case re.TH:case re.TD:case re.TR:case re.HEAD:case re.FRAME:case re.TBODY:case re.TFOOT:case re.THEAD:case re.CAPTION:case re.COLGROUP:break;default:aKt(n,r)}}function _Zr(n,r){if(n.openElements.hasInScope(re.BODY)&&(n.insertionMode=gt.AFTER_BODY,n.options.sourceCodeLocationInfo)){const o=n.openElements.tryPeekProperlyNestedBodyElement();o&&n._setEndLocation(o,r)}}function kZr(n,r){n.openElements.hasInScope(re.BODY)&&(n.insertionMode=gt.AFTER_BODY,chn(n,r))}function TZr(n,r){const o=r.tagID;n.openElements.hasInScope(o)&&(n.openElements.generateImpliedEndTags(),n.openElements.popUntilTagNamePopped(o))}function xZr(n){const r=n.openElements.tmplCount>0,{formElement:o}=n;r||(n.formElement=null),(o||r)&&n.openElements.hasInScope(re.FORM)&&(n.openElements.generateImpliedEndTags(),r?n.openElements.popUntilTagNamePopped(re.FORM):o&&n.openElements.remove(o))}function CZr(n){n.openElements.hasInButtonScope(re.P)||n._insertFakeElement(nn.P,re.P),n._closePElement()}function AZr(n){n.openElements.hasInListItemScope(re.LI)&&(n.openElements.generateImpliedEndTagsWithExclusion(re.LI),n.openElements.popUntilTagNamePopped(re.LI))}function IZr(n,r){const o=r.tagID;n.openElements.hasInScope(o)&&(n.openElements.generateImpliedEndTagsWithExclusion(o),n.openElements.popUntilTagNamePopped(o))}function OZr(n){n.openElements.hasNumberedHeaderInScope()&&(n.openElements.generateImpliedEndTags(),n.openElements.popUntilNumberedHeaderPopped())}function NZr(n,r){const o=r.tagID;n.openElements.hasInScope(o)&&(n.openElements.generateImpliedEndTags(),n.openElements.popUntilTagNamePopped(o),n.activeFormattingElements.clearToLastMarker())}function RZr(n){n._reconstructActiveFormattingElements(),n._insertFakeElement(nn.BR,re.BR),n.openElements.pop(),n.framesetOk=!1}function Qfn(n,r){const o=r.tagName,s=r.tagID;for(let c=n.openElements.stackTop;c>0;c--){const d=n.openElements.items[c],h=n.openElements.tagIDs[c];if(s===h&&(s!==re.UNKNOWN||n.treeAdapter.getTagName(d)===o)){n.openElements.generateImpliedEndTagsWithExclusion(s),n.openElements.stackTop>=c&&n.openElements.shortenToLength(c);break}if(n._isSpecialElement(d,h))break}}function YEe(n,r){switch(r.tagID){case re.A:case re.B:case re.I:case re.S:case re.U:case re.EM:case re.TT:case re.BIG:case re.CODE:case re.FONT:case re.NOBR:case re.SMALL:case re.STRIKE:case re.STRONG:{KQe(n,r);break}case re.P:{CZr(n);break}case re.DL:case re.UL:case re.OL:case re.DIR:case re.DIV:case re.NAV:case re.PRE:case re.MAIN:case re.MENU:case re.ASIDE:case re.BUTTON:case re.CENTER:case re.FIGURE:case re.FOOTER:case re.HEADER:case re.HGROUP:case re.DIALOG:case re.ADDRESS:case re.ARTICLE:case re.DETAILS:case re.SEARCH:case re.SECTION:case re.SUMMARY:case re.LISTING:case re.FIELDSET:case re.BLOCKQUOTE:case re.FIGCAPTION:{TZr(n,r);break}case re.LI:{AZr(n);break}case re.DD:case re.DT:{IZr(n,r);break}case re.H1:case re.H2:case re.H3:case re.H4:case re.H5:case re.H6:{OZr(n);break}case re.BR:{RZr(n);break}case re.BODY:{_Zr(n,r);break}case re.HTML:{kZr(n,r);break}case re.FORM:{xZr(n);break}case re.APPLET:case re.OBJECT:case re.MARQUEE:{NZr(n,r);break}case re.TEMPLATE:{EU(n,r);break}default:Qfn(n,r)}}function ehn(n,r){n.tmplInsertionModeStack.length>0?shn(n,r):YQe(n,r)}function PZr(n,r){var o;r.tagID===re.SCRIPT&&((o=n.scriptHandler)===null||o===void 0||o.call(n,n.openElements.current)),n.openElements.pop(),n.insertionMode=n.originalInsertionMode}function LZr(n,r){n._err(r,On.eofInElementThatCanContainOnlyText),n.openElements.pop(),n.insertionMode=n.originalInsertionMode,n.onEof(r)}function Sqe(n,r){if(n.openElements.currentTagId!==void 0&&Kfn.has(n.openElements.currentTagId))switch(n.pendingCharacterTokens.length=0,n.hasNonWhitespacePendingCharacterToken=!1,n.originalInsertionMode=n.insertionMode,n.insertionMode=gt.IN_TABLE_TEXT,r.type){case nc.CHARACTER:{nhn(n,r);break}case nc.WHITESPACE_CHARACTER:{thn(n,r);break}}else eae(n,r)}function DZr(n,r){n.openElements.clearBackToTableContext(),n.activeFormattingElements.insertMarker(),n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_CAPTION}function MZr(n,r){n.openElements.clearBackToTableContext(),n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_COLUMN_GROUP}function jZr(n,r){n.openElements.clearBackToTableContext(),n._insertFakeElement(nn.COLGROUP,re.COLGROUP),n.insertionMode=gt.IN_COLUMN_GROUP,XQe(n,r)}function $Zr(n,r){n.openElements.clearBackToTableContext(),n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_TABLE_BODY}function FZr(n,r){n.openElements.clearBackToTableContext(),n._insertFakeElement(nn.TBODY,re.TBODY),n.insertionMode=gt.IN_TABLE_BODY,XEe(n,r)}function BZr(n,r){n.openElements.hasInTableScope(re.TABLE)&&(n.openElements.popUntilTagNamePopped(re.TABLE),n._resetInsertionMode(),n._processStartTag(r))}function UZr(n,r){Zfn(r)?n._appendElement(r,Er.HTML):eae(n,r),r.ackSelfClosing=!0}function zZr(n,r){!n.formElement&&n.openElements.tmplCount===0&&(n._insertElement(r,Er.HTML),n.formElement=n.openElements.current,n.openElements.pop())}function pK(n,r){switch(r.tagID){case re.TD:case re.TH:case re.TR:{FZr(n,r);break}case re.STYLE:case re.SCRIPT:case re.TEMPLATE:{nC(n,r);break}case re.COL:{jZr(n,r);break}case re.FORM:{zZr(n,r);break}case re.TABLE:{BZr(n,r);break}case re.TBODY:case re.TFOOT:case re.THEAD:{$Zr(n,r);break}case re.INPUT:{UZr(n,r);break}case re.CAPTION:{DZr(n,r);break}case re.COLGROUP:{MZr(n,r);break}default:eae(n,r)}}function moe(n,r){switch(r.tagID){case re.TABLE:{n.openElements.hasInTableScope(re.TABLE)&&(n.openElements.popUntilTagNamePopped(re.TABLE),n._resetInsertionMode());break}case re.TEMPLATE:{EU(n,r);break}case re.BODY:case re.CAPTION:case re.COL:case re.COLGROUP:case re.HTML:case re.TBODY:case re.TD:case re.TFOOT:case re.TH:case re.THEAD:case re.TR:break;default:eae(n,r)}}function eae(n,r){const o=n.fosterParentingEnabled;n.fosterParentingEnabled=!0,KEe(n,r),n.fosterParentingEnabled=o}function thn(n,r){n.pendingCharacterTokens.push(r)}function nhn(n,r){n.pendingCharacterTokens.push(r),n.hasNonWhitespacePendingCharacterToken=!0}function Ire(n,r){let o=0;if(n.hasNonWhitespacePendingCharacterToken)for(;o<n.pendingCharacterTokens.length;o++)eae(n,n.pendingCharacterTokens[o]);else for(;o<n.pendingCharacterTokens.length;o++)n._insertCharacters(n.pendingCharacterTokens[o]);n.insertionMode=n.originalInsertionMode,n._processToken(r)}const rhn=new Set([re.CAPTION,re.COL,re.COLGROUP,re.TBODY,re.TD,re.TFOOT,re.TH,re.THEAD,re.TR]);function HZr(n,r){const o=r.tagID;rhn.has(o)?n.openElements.hasInTableScope(re.CAPTION)&&(n.openElements.generateImpliedEndTags(),n.openElements.popUntilTagNamePopped(re.CAPTION),n.activeFormattingElements.clearToLastMarker(),n.insertionMode=gt.IN_TABLE,pK(n,r)):Q1(n,r)}function GZr(n,r){const o=r.tagID;switch(o){case re.CAPTION:case re.TABLE:{n.openElements.hasInTableScope(re.CAPTION)&&(n.openElements.generateImpliedEndTags(),n.openElements.popUntilTagNamePopped(re.CAPTION),n.activeFormattingElements.clearToLastMarker(),n.insertionMode=gt.IN_TABLE,o===re.TABLE&&moe(n,r));break}case re.BODY:case re.COL:case re.COLGROUP:case re.HTML:case re.TBODY:case re.TD:case re.TFOOT:case re.TH:case re.THEAD:case re.TR:break;default:YEe(n,r)}}function XQe(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.COL:{n._appendElement(r,Er.HTML),r.ackSelfClosing=!0;break}case re.TEMPLATE:{nC(n,r);break}default:Jwe(n,r)}}function qZr(n,r){switch(r.tagID){case re.COLGROUP:{n.openElements.currentTagId===re.COLGROUP&&(n.openElements.pop(),n.insertionMode=gt.IN_TABLE);break}case re.TEMPLATE:{EU(n,r);break}case re.COL:break;default:Jwe(n,r)}}function Jwe(n,r){n.openElements.currentTagId===re.COLGROUP&&(n.openElements.pop(),n.insertionMode=gt.IN_TABLE,n._processToken(r))}function XEe(n,r){switch(r.tagID){case re.TR:{n.openElements.clearBackToTableBodyContext(),n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_ROW;break}case re.TH:case re.TD:{n.openElements.clearBackToTableBodyContext(),n._insertFakeElement(nn.TR,re.TR),n.insertionMode=gt.IN_ROW,JEe(n,r);break}case re.CAPTION:case re.COL:case re.COLGROUP:case re.TBODY:case re.TFOOT:case re.THEAD:{n.openElements.hasTableBodyContextInTableScope()&&(n.openElements.clearBackToTableBodyContext(),n.openElements.pop(),n.insertionMode=gt.IN_TABLE,pK(n,r));break}default:pK(n,r)}}function _Ke(n,r){const o=r.tagID;switch(r.tagID){case re.TBODY:case re.TFOOT:case re.THEAD:{n.openElements.hasInTableScope(o)&&(n.openElements.clearBackToTableBodyContext(),n.openElements.pop(),n.insertionMode=gt.IN_TABLE);break}case re.TABLE:{n.openElements.hasTableBodyContextInTableScope()&&(n.openElements.clearBackToTableBodyContext(),n.openElements.pop(),n.insertionMode=gt.IN_TABLE,moe(n,r));break}case re.BODY:case re.CAPTION:case re.COL:case re.COLGROUP:case re.HTML:case re.TD:case re.TH:case re.TR:break;default:moe(n,r)}}function JEe(n,r){switch(r.tagID){case re.TH:case re.TD:{n.openElements.clearBackToTableRowContext(),n._insertElement(r,Er.HTML),n.insertionMode=gt.IN_CELL,n.activeFormattingElements.insertMarker();break}case re.CAPTION:case re.COL:case re.COLGROUP:case re.TBODY:case re.TFOOT:case re.THEAD:case re.TR:{n.openElements.hasInTableScope(re.TR)&&(n.openElements.clearBackToTableRowContext(),n.openElements.pop(),n.insertionMode=gt.IN_TABLE_BODY,XEe(n,r));break}default:pK(n,r)}}function ihn(n,r){switch(r.tagID){case re.TR:{n.openElements.hasInTableScope(re.TR)&&(n.openElements.clearBackToTableRowContext(),n.openElements.pop(),n.insertionMode=gt.IN_TABLE_BODY);break}case re.TABLE:{n.openElements.hasInTableScope(re.TR)&&(n.openElements.clearBackToTableRowContext(),n.openElements.pop(),n.insertionMode=gt.IN_TABLE_BODY,_Ke(n,r));break}case re.TBODY:case re.TFOOT:case re.THEAD:{(n.openElements.hasInTableScope(r.tagID)||n.openElements.hasInTableScope(re.TR))&&(n.openElements.clearBackToTableRowContext(),n.openElements.pop(),n.insertionMode=gt.IN_TABLE_BODY,_Ke(n,r));break}case re.BODY:case re.CAPTION:case re.COL:case re.COLGROUP:case re.HTML:case re.TD:case re.TH:break;default:moe(n,r)}}function VZr(n,r){const o=r.tagID;rhn.has(o)?(n.openElements.hasInTableScope(re.TD)||n.openElements.hasInTableScope(re.TH))&&(n._closeTableCell(),JEe(n,r)):Q1(n,r)}function WZr(n,r){const o=r.tagID;switch(o){case re.TD:case re.TH:{n.openElements.hasInTableScope(o)&&(n.openElements.generateImpliedEndTags(),n.openElements.popUntilTagNamePopped(o),n.activeFormattingElements.clearToLastMarker(),n.insertionMode=gt.IN_ROW);break}case re.TABLE:case re.TBODY:case re.TFOOT:case re.THEAD:case re.TR:{n.openElements.hasInTableScope(o)&&(n._closeTableCell(),ihn(n,r));break}case re.BODY:case re.CAPTION:case re.COL:case re.COLGROUP:case re.HTML:break;default:YEe(n,r)}}function ohn(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.OPTION:{n.openElements.currentTagId===re.OPTION&&n.openElements.pop(),n._insertElement(r,Er.HTML);break}case re.OPTGROUP:{n.openElements.currentTagId===re.OPTION&&n.openElements.pop(),n.openElements.currentTagId===re.OPTGROUP&&n.openElements.pop(),n._insertElement(r,Er.HTML);break}case re.HR:{n.openElements.currentTagId===re.OPTION&&n.openElements.pop(),n.openElements.currentTagId===re.OPTGROUP&&n.openElements.pop(),n._appendElement(r,Er.HTML),r.ackSelfClosing=!0;break}case re.INPUT:case re.KEYGEN:case re.TEXTAREA:case re.SELECT:{n.openElements.hasInSelectScope(re.SELECT)&&(n.openElements.popUntilTagNamePopped(re.SELECT),n._resetInsertionMode(),r.tagID!==re.SELECT&&n._processStartTag(r));break}case re.SCRIPT:case re.TEMPLATE:{nC(n,r);break}}}function ahn(n,r){switch(r.tagID){case re.OPTGROUP:{n.openElements.stackTop>0&&n.openElements.currentTagId===re.OPTION&&n.openElements.tagIDs[n.openElements.stackTop-1]===re.OPTGROUP&&n.openElements.pop(),n.openElements.currentTagId===re.OPTGROUP&&n.openElements.pop();break}case re.OPTION:{n.openElements.currentTagId===re.OPTION&&n.openElements.pop();break}case re.SELECT:{n.openElements.hasInSelectScope(re.SELECT)&&(n.openElements.popUntilTagNamePopped(re.SELECT),n._resetInsertionMode());break}case re.TEMPLATE:{EU(n,r);break}}}function KZr(n,r){const o=r.tagID;o===re.CAPTION||o===re.TABLE||o===re.TBODY||o===re.TFOOT||o===re.THEAD||o===re.TR||o===re.TD||o===re.TH?(n.openElements.popUntilTagNamePopped(re.SELECT),n._resetInsertionMode(),n._processStartTag(r)):ohn(n,r)}function YZr(n,r){const o=r.tagID;o===re.CAPTION||o===re.TABLE||o===re.TBODY||o===re.TFOOT||o===re.THEAD||o===re.TR||o===re.TD||o===re.TH?n.openElements.hasInTableScope(o)&&(n.openElements.popUntilTagNamePopped(re.SELECT),n._resetInsertionMode(),n.onEndTag(r)):ahn(n,r)}function XZr(n,r){switch(r.tagID){case re.BASE:case re.BASEFONT:case re.BGSOUND:case re.LINK:case re.META:case re.NOFRAMES:case re.SCRIPT:case re.STYLE:case re.TEMPLATE:case re.TITLE:{nC(n,r);break}case re.CAPTION:case re.COLGROUP:case re.TBODY:case re.TFOOT:case re.THEAD:{n.tmplInsertionModeStack[0]=gt.IN_TABLE,n.insertionMode=gt.IN_TABLE,pK(n,r);break}case re.COL:{n.tmplInsertionModeStack[0]=gt.IN_COLUMN_GROUP,n.insertionMode=gt.IN_COLUMN_GROUP,XQe(n,r);break}case re.TR:{n.tmplInsertionModeStack[0]=gt.IN_TABLE_BODY,n.insertionMode=gt.IN_TABLE_BODY,XEe(n,r);break}case re.TD:case re.TH:{n.tmplInsertionModeStack[0]=gt.IN_ROW,n.insertionMode=gt.IN_ROW,JEe(n,r);break}default:n.tmplInsertionModeStack[0]=gt.IN_BODY,n.insertionMode=gt.IN_BODY,Q1(n,r)}}function JZr(n,r){r.tagID===re.TEMPLATE&&EU(n,r)}function shn(n,r){n.openElements.tmplCount>0?(n.openElements.popUntilTagNamePopped(re.TEMPLATE),n.activeFormattingElements.clearToLastMarker(),n.tmplInsertionModeStack.shift(),n._resetInsertionMode(),n.onEof(r)):YQe(n,r)}function ZZr(n,r){r.tagID===re.HTML?Q1(n,r):Zwe(n,r)}function chn(n,r){var o;if(r.tagID===re.HTML){if(n.fragmentContext||(n.insertionMode=gt.AFTER_AFTER_BODY),n.options.sourceCodeLocationInfo&&n.openElements.tagIDs[0]===re.HTML){n._setEndLocation(n.openElements.items[0],r);const s=n.openElements.items[1];s&&!(!((o=n.treeAdapter.getNodeSourceCodeLocation(s))===null||o===void 0)&&o.endTag)&&n._setEndLocation(s,r)}}else Zwe(n,r)}function Zwe(n,r){n.insertionMode=gt.IN_BODY,KEe(n,r)}function QZr(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.FRAMESET:{n._insertElement(r,Er.HTML);break}case re.FRAME:{n._appendElement(r,Er.HTML),r.ackSelfClosing=!0;break}case re.NOFRAMES:{nC(n,r);break}}}function eQr(n,r){r.tagID===re.FRAMESET&&!n.openElements.isRootHtmlElementCurrent()&&(n.openElements.pop(),!n.fragmentContext&&n.openElements.currentTagId!==re.FRAMESET&&(n.insertionMode=gt.AFTER_FRAMESET))}function tQr(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.NOFRAMES:{nC(n,r);break}}}function nQr(n,r){r.tagID===re.HTML&&(n.insertionMode=gt.AFTER_AFTER_FRAMESET)}function rQr(n,r){r.tagID===re.HTML?Q1(n,r):Dve(n,r)}function Dve(n,r){n.insertionMode=gt.IN_BODY,KEe(n,r)}function iQr(n,r){switch(r.tagID){case re.HTML:{Q1(n,r);break}case re.NOFRAMES:{nC(n,r);break}}}function oQr(n,r){r.chars=Sf,n._insertCharacters(r)}function aQr(n,r){n._insertCharacters(r),n.framesetOk=!1}function uhn(n){for(;n.treeAdapter.getNamespaceURI(n.openElements.current)!==Er.HTML&&n.openElements.currentTagId!==void 0&&!n._isIntegrationPoint(n.openElements.currentTagId,n.openElements.current);)n.openElements.pop()}function sQr(n,r){if(SJr(r))uhn(n),n._startTagOutsideForeignContent(r);else{const o=n._getAdjustedCurrentElement(),s=n.treeAdapter.getNamespaceURI(o);s===Er.MATHML?Vfn(r):s===Er.SVG&&(_Jr(r),Wfn(r)),WQe(r),r.selfClosing?n._appendElement(r,s):n._insertElement(r,s),r.ackSelfClosing=!0}}function cQr(n,r){if(r.tagID===re.P||r.tagID===re.BR){uhn(n),n._endTagOutsideForeignContent(r);return}for(let o=n.openElements.stackTop;o>0;o--){const s=n.openElements.items[o];if(n.treeAdapter.getNamespaceURI(s)===Er.HTML){n._endTagOutsideForeignContent(r);break}const c=n.treeAdapter.getTagName(s);if(c.toLowerCase()===r.tagName){r.tagName=c,n.openElements.shortenToLength(o);break}}}nn.AREA,nn.BASE,nn.BASEFONT,nn.BGSOUND,nn.BR,nn.COL,nn.EMBED,nn.FRAME,nn.HR,nn.IMG,nn.INPUT,nn.KEYGEN,nn.LINK,nn.META,nn.PARAM,nn.SOURCE,nn.TRACK,nn.WBR;const uQr=/<(\/?)(iframe|noembed|noframes|plaintext|script|style|textarea|title|xmp)(?=[\t\n\f\r />])/gi,lQr=new Set(["mdxFlowExpression","mdxJsxFlowElement","mdxJsxTextElement","mdxTextExpression","mdxjsEsm"]),sKt={sourceCodeLocationInfo:!0,scriptingEnabled:!1};function lhn(n,r){const o=yQr(n),s=Dfn("type",{handlers:{root:dQr,element:fQr,text:hQr,comment:fhn,doctype:pQr,raw:mQr},unknown:bQr}),c={parser:o?new iKt(sKt):iKt.getFragmentParser(void 0,sKt),handle(b){s(b,c)},stitches:!1,options:r||{}};s(n,c),HK(c,WI());const d=o?c.parser.document:c.parser.getFragment(),h=dXr(d,{file:c.options.file});return c.stitches&&WEe(h,"comment",function(b,y,E){const k=b;if(k.value.stitch&&E&&y!==void 0){const x=E.children;return x[y]=k.value.stitch,y}}),h.type==="root"&&h.children.length===1&&h.children[0].type===n.type?h.children[0]:h}function dhn(n,r){let o=-1;if(n)for(;++o<n.length;)r.handle(n[o])}function dQr(n,r){dhn(n.children,r)}function fQr(n,r){vQr(n,r),dhn(n.children,r),wQr(n,r)}function hQr(n,r){r.parser.tokenizer.state>4&&(r.parser.tokenizer.state=0);const o={type:nc.CHARACTER,chars:n.value,location:tae(n)};HK(r,WI(n)),r.parser.currentToken=o,r.parser._processToken(r.parser.currentToken)}function pQr(n,r){const o={type:nc.DOCTYPE,name:"html",forceQuirks:!1,publicId:"",systemId:"",location:tae(n)};HK(r,WI(n)),r.parser.currentToken=o,r.parser._processToken(r.parser.currentToken)}function gQr(n,r){r.stitches=!0;const o=EQr(n);if("children"in n&&"children"in o){const s=lhn({type:"root",children:n.children},r.options);o.children=s.children}fhn({type:"comment",value:{stitch:o}},r)}function fhn(n,r){const o=n.value,s={type:nc.COMMENT,data:o,location:tae(n)};HK(r,WI(n)),r.parser.currentToken=s,r.parser._processToken(r.parser.currentToken)}function mQr(n,r){if(r.parser.tokenizer.preprocessor.html="",r.parser.tokenizer.preprocessor.pos=-1,r.parser.tokenizer.preprocessor.lastGapPos=-2,r.parser.tokenizer.preprocessor.gapStack=[],r.parser.tokenizer.preprocessor.skipNextNewLine=!1,r.parser.tokenizer.preprocessor.lastChunkWritten=!1,r.parser.tokenizer.preprocessor.endOfChunkHit=!1,r.parser.tokenizer.preprocessor.isEol=!1,hhn(r,WI(n)),r.parser.tokenizer.write(r.options.tagfilter?n.value.replace(uQr,"&lt;$1$2"):n.value,!1),r.parser.tokenizer._runParsingLoop(),r.parser.tokenizer.state===72||r.parser.tokenizer.state===78){r.parser.tokenizer.preprocessor.lastChunkWritten=!0;const o=r.parser.tokenizer._consume();r.parser.tokenizer._callState(o)}}function bQr(n,r){const o=n;if(r.options.passThrough&&r.options.passThrough.includes(o.type))gQr(o,r);else{let s="";throw lQr.has(o.type)&&(s=". It looks like you are using MDX nodes with `hast-util-raw` (or `rehype-raw`). If you use this because you are using remark or rehype plugins that inject `'html'` nodes, then please raise an issue with that plugin, as its a bad and slow idea. If you use this because you are using markdown syntax, then you have to configure this utility (or plugin) to pass through these nodes (see `passThrough` in docs), but you can also migrate to use the MDX syntax"),new Error("Cannot compile `"+o.type+"` node"+s)}}function HK(n,r){hhn(n,r);const o=n.parser.tokenizer.currentCharacterToken;o&&o.location&&(o.location.endLine=n.parser.tokenizer.preprocessor.line,o.location.endCol=n.parser.tokenizer.preprocessor.col+1,o.location.endOffset=n.parser.tokenizer.preprocessor.offset+1,n.parser.currentToken=o,n.parser._processToken(n.parser.currentToken)),n.parser.tokenizer.paused=!1,n.parser.tokenizer.inLoop=!1,n.parser.tokenizer.active=!1,n.parser.tokenizer.returnState=gp.DATA,n.parser.tokenizer.charRefCode=-1,n.parser.tokenizer.consumedAfterSnapshot=-1,n.parser.tokenizer.currentLocation=null,n.parser.tokenizer.currentCharacterToken=null,n.parser.tokenizer.currentToken=null,n.parser.tokenizer.currentAttr={name:"",value:""}}function hhn(n,r){if(r&&r.offset!==void 0){const o={startLine:r.line,startCol:r.column,startOffset:r.offset,endLine:-1,endCol:-1,endOffset:-1};n.parser.tokenizer.preprocessor.lineStartPos=-r.column+1,n.parser.tokenizer.preprocessor.droppedBufferSize=r.offset,n.parser.tokenizer.preprocessor.line=r.line,n.parser.tokenizer.currentLocation=o}}function vQr(n,r){const o=n.tagName.toLowerCase();if(r.parser.tokenizer.state===gp.PLAINTEXT)return;HK(r,WI(n));const s=r.parser.openElements.current;let c="namespaceURI"in s?s.namespaceURI:jB.html;c===jB.html&&o==="svg"&&(c=jB.svg);const d=xXr({...n,children:[]},{space:c===jB.svg?"svg":"html"}),h={type:nc.START_TAG,tagName:o,tagID:zK(o),selfClosing:!1,ackSelfClosing:!1,attrs:"attrs"in d?d.attrs:[],location:tae(n)};r.parser.currentToken=h,r.parser._processToken(r.parser.currentToken),r.parser.tokenizer.lastStartTagName=o}function wQr(n,r){const o=n.tagName.toLowerCase();if(!r.parser.tokenizer.inForeignNode&&LXr.includes(o)||r.parser.tokenizer.state===gp.PLAINTEXT)return;HK(r,UEe(n));const s={type:nc.END_TAG,tagName:o,tagID:zK(o),selfClosing:!1,ackSelfClosing:!1,attrs:[],location:tae(n)};r.parser.currentToken=s,r.parser._processToken(r.parser.currentToken),o===r.parser.tokenizer.lastStartTagName&&(r.parser.tokenizer.state===gp.RCDATA||r.parser.tokenizer.state===gp.RAWTEXT||r.parser.tokenizer.state===gp.SCRIPT_DATA)&&(r.parser.tokenizer.state=gp.DATA)}function yQr(n){const r=n.type==="root"?n.children[0]:n;return!!(r&&(r.type==="doctype"||r.type==="element"&&r.tagName.toLowerCase()==="html"))}function tae(n){const r=WI(n)||{line:void 0,column:void 0,offset:void 0},o=UEe(n)||{line:void 0,column:void 0,offset:void 0};return{startLine:r.line,startCol:r.column,startOffset:r.offset,endLine:o.line,endCol:o.column,endOffset:o.offset}}function EQr(n){return"children"in n?hK({...n,children:[]}):hK(n)}function SQr(n){return function(r,o){return lhn(r,{...n,file:o})}}function cKt(n,r){const o=String(n);if(typeof r!="string")throw new TypeError("Expected character");let s=0,c=o.indexOf(r);for(;c!==-1;)s++,c=o.indexOf(r,c+r.length);return s}function _Qr(n){if(typeof n!="string")throw new TypeError("Expected a string");return n.replace(/[|\\{}()[\]^$+*?.]/g,"\\$&").replace(/-/g,"\\x2d")}function kQr(n,r,o){const c=qEe((o||{}).ignore||[]),d=TQr(r);let h=-1;for(;++h<d.length;)Efn(n,"text",b);function b(E,k){let x=-1,O;for(;++x<k.length;){const P=k[x],D=O?O.children:void 0;if(c(P,D?D.indexOf(P):void 0,O))return;O=P}if(O)return y(E,k)}function y(E,k){const x=k[k.length-1],O=d[h][0],P=d[h][1];let D=0;const U=x.children.indexOf(E);let B=!1,G=[];O.lastIndex=0;let K=O.exec(E.value);for(;K;){const X=K.index,ae={index:K.index,input:K.input,stack:[...k,E]};let oe=P(...K,ae);if(typeof oe=="string"&&(oe=oe.length>0?{type:"text",value:oe}:void 0),oe===!1?O.lastIndex=X+1:(D!==X&&G.push({type:"text",value:E.value.slice(D,X)}),Array.isArray(oe)?G.push(...oe):oe&&G.push(oe),D=X+K[0].length,B=!0),!O.global)break;K=O.exec(E.value)}return B?(D<E.value.length&&G.push({type:"text",value:E.value.slice(D)}),x.children.splice(U,1,...G)):G=[E],U+G.length}}function TQr(n){const r=[];if(!Array.isArray(n))throw new TypeError("Expected find and replace tuple or list of tuples");const o=!n[0]||Array.isArray(n[0])?n:[n];let s=-1;for(;++s<o.length;){const c=o[s];r.push([xQr(c[0]),CQr(c[1])])}return r}function xQr(n){return typeof n=="string"?new RegExp(_Qr(n),"g"):n}function CQr(n){return typeof n=="function"?n:function(){return n}}const _qe="phrasing",kqe=["autolink","link","image","label"];function AQr(){return{transforms:[DQr],enter:{literalAutolink:OQr,literalAutolinkEmail:Tqe,literalAutolinkHttp:Tqe,literalAutolinkWww:Tqe},exit:{literalAutolink:LQr,literalAutolinkEmail:PQr,literalAutolinkHttp:NQr,literalAutolinkWww:RQr}}}function IQr(){return{unsafe:[{character:"@",before:"[+\\-.\\w]",after:"[\\-.\\w]",inConstruct:_qe,notInConstruct:kqe},{character:".",before:"[Ww]",after:"[\\-.\\w]",inConstruct:_qe,notInConstruct:kqe},{character:":",before:"[ps]",after:"\\/",inConstruct:_qe,notInConstruct:kqe}]}}function OQr(n){this.enter({type:"link",title:null,url:"",children:[]},n)}function Tqe(n){this.config.enter.autolinkProtocol.call(this,n)}function NQr(n){this.config.exit.autolinkProtocol.call(this,n)}function RQr(n){this.config.exit.data.call(this,n);const r=this.stack[this.stack.length-1];r.type,r.url="http://"+this.sliceSerialize(n)}function PQr(n){this.config.exit.autolinkEmail.call(this,n)}function LQr(n){this.exit(n)}function DQr(n){kQr(n,[[/(https?:\/\/|www(?=\.))([-.\w]+)([^ \t\r\n]*)/gi,MQr],[new RegExp("(?<=^|\\s|\\p{P}|\\p{S})([-.\\w+]+)@([-\\w]+(?:\\.[-\\w]+)+)","gu"),jQr]],{ignore:["link","linkReference"]})}function MQr(n,r,o,s,c){let d="";if(!phn(c)||(/^w/i.test(r)&&(o=r+o,r="",d="http://"),!$Qr(o)))return!1;const h=FQr(o+s);if(!h[0])return!1;const b={type:"link",title:null,url:d+r+h[0],children:[{type:"text",value:r+h[0]}]};return h[1]?[b,{type:"text",value:h[1]}]:b}function jQr(n,r,o,s){return!phn(s,!0)||/[-\d_]$/.test(o)?!1:{type:"link",title:null,url:"mailto:"+r+"@"+o,children:[{type:"text",value:r+"@"+o}]}}function $Qr(n){const r=n.split(".");return!(r.length<2||r[r.length-1]&&(/_/.test(r[r.length-1])||!/[a-zA-Z\d]/.test(r[r.length-1]))||r[r.length-2]&&(/_/.test(r[r.length-2])||!/[a-zA-Z\d]/.test(r[r.length-2])))}function FQr(n){const r=/[!"&'),.:;<>?\]}]+$/.exec(n);if(!r)return[n,void 0];n=n.slice(0,r.index);let o=r[0],s=o.indexOf(")");const c=cKt(n,"(");let d=cKt(n,")");for(;s!==-1&&c>d;)n+=o.slice(0,s+1),o=o.slice(s+1),s=o.indexOf(")"),d++;return[n,o]}function phn(n,r){const o=n.input.charCodeAt(n.index-1);return(n.index===0||sU(o)||zEe(o))&&(!r||o!==47)}ghn.peek=KQr;function BQr(){this.buffer()}function UQr(n){this.enter({type:"footnoteReference",identifier:"",label:""},n)}function zQr(){this.buffer()}function HQr(n){this.enter({type:"footnoteDefinition",identifier:"",label:"",children:[]},n)}function GQr(n){const r=this.resume(),o=this.stack[this.stack.length-1];o.type,o.identifier=zx(this.sliceSerialize(n)).toLowerCase(),o.label=r}function qQr(n){this.exit(n)}function VQr(n){const r=this.resume(),o=this.stack[this.stack.length-1];o.type,o.identifier=zx(this.sliceSerialize(n)).toLowerCase(),o.label=r}function WQr(n){this.exit(n)}function KQr(){return"["}function ghn(n,r,o,s){const c=o.createTracker(s);let d=c.move("[^");const h=o.enter("footnoteReference"),b=o.enter("reference");return d+=c.move(o.safe(o.associationId(n),{after:"]",before:d})),b(),h(),d+=c.move("]"),d}function YQr(){return{enter:{gfmFootnoteCallString:BQr,gfmFootnoteCall:UQr,gfmFootnoteDefinitionLabelString:zQr,gfmFootnoteDefinition:HQr},exit:{gfmFootnoteCallString:GQr,gfmFootnoteCall:qQr,gfmFootnoteDefinitionLabelString:VQr,gfmFootnoteDefinition:WQr}}}function XQr(n){let r=!1;return n&&n.firstLineBlank&&(r=!0),{handlers:{footnoteDefinition:o,footnoteReference:ghn},unsafe:[{character:"[",inConstruct:["label","phrasing","reference"]}]};function o(s,c,d,h){const b=d.createTracker(h);let y=b.move("[^");const E=d.enter("footnoteDefinition"),k=d.enter("label");return y+=b.move(d.safe(d.associationId(s),{before:y,after:"]"})),k(),y+=b.move("]:"),s.children&&s.children.length>0&&(b.shift(4),y+=b.move((r?`
`:" ")+d.indentLines(d.containerFlow(s,b.current()),r?mhn:JQr))),E(),y}}function JQr(n,r,o){return r===0?n:mhn(n,r,o)}function mhn(n,r,o){return(o?"":"    ")+n}const ZQr=["autolink","destinationLiteral","destinationRaw","reference","titleQuote","titleApostrophe"];bhn.peek=rei;function QQr(){return{canContainEols:["delete"],enter:{strikethrough:tei},exit:{strikethrough:nei}}}function eei(){return{unsafe:[{character:"~",inConstruct:"phrasing",notInConstruct:ZQr}],handlers:{delete:bhn}}}function tei(n){this.enter({type:"delete",children:[]},n)}function nei(n){this.exit(n)}function bhn(n,r,o,s){const c=o.createTracker(s),d=o.enter("strikethrough");let h=c.move("~~");return h+=o.containerPhrasing(n,{...c.current(),before:h,after:"~"}),h+=c.move("~~"),d(),h}function rei(){return"~"}function iei(n){return n.length}function oei(n,r){const o=r||{},s=(o.align||[]).concat(),c=o.stringLength||iei,d=[],h=[],b=[],y=[];let E=0,k=-1;for(;++k<n.length;){const $=[],U=[];let B=-1;for(n[k].length>E&&(E=n[k].length);++B<n[k].length;){const G=aei(n[k][B]);if(o.alignDelimiters!==!1){const K=c(G);U[B]=K,(y[B]===void 0||K>y[B])&&(y[B]=K)}$.push(G)}h[k]=$,b[k]=U}let x=-1;if(typeof s=="object"&&"length"in s)for(;++x<E;)d[x]=uKt(s[x]);else{const $=uKt(s);for(;++x<E;)d[x]=$}x=-1;const O=[],P=[];for(;++x<E;){const $=d[x];let U="",B="";$===99?(U=":",B=":"):$===108?U=":":$===114&&(B=":");let G=o.alignDelimiters===!1?1:Math.max(1,y[x]-U.length-B.length);const K=U+"-".repeat(G)+B;o.alignDelimiters!==!1&&(G=U.length+G+B.length,G>y[x]&&(y[x]=G),P[x]=G),O[x]=K}h.splice(1,0,O),b.splice(1,0,P),k=-1;const D=[];for(;++k<h.length;){const $=h[k],U=b[k];x=-1;const B=[];for(;++x<E;){const G=$[x]||"";let K="",X="";if(o.alignDelimiters!==!1){const ae=y[x]-(U[x]||0),oe=d[x];oe===114?K=" ".repeat(ae):oe===99?ae%2?(K=" ".repeat(ae/2+.5),X=" ".repeat(ae/2-.5)):(K=" ".repeat(ae/2),X=K):X=" ".repeat(ae)}o.delimiterStart!==!1&&!x&&B.push("|"),o.padding!==!1&&!(o.alignDelimiters===!1&&G==="")&&(o.delimiterStart!==!1||x)&&B.push(" "),o.alignDelimiters!==!1&&B.push(K),B.push(G),o.alignDelimiters!==!1&&B.push(X),o.padding!==!1&&B.push(" "),(o.delimiterEnd!==!1||x!==E-1)&&B.push("|")}D.push(o.delimiterEnd===!1?B.join("").replace(/ +$/,""):B.join(""))}return D.join(`
`)}function aei(n){return n==null?"":String(n)}function uKt(n){const r=typeof n=="string"?n.codePointAt(0):0;return r===67||r===99?99:r===76||r===108?108:r===82||r===114?114:0}function sei(n,r,o,s){const c=o.enter("blockquote"),d=o.createTracker(s);d.move("> "),d.shift(2);const h=o.indentLines(o.containerFlow(n,d.current()),cei);return c(),h}function cei(n,r,o){return">"+(o?"":" ")+n}function uei(n,r){return lKt(n,r.inConstruct,!0)&&!lKt(n,r.notInConstruct,!1)}function lKt(n,r,o){if(typeof r=="string"&&(r=[r]),!r||r.length===0)return o;let s=-1;for(;++s<r.length;)if(n.includes(r[s]))return!0;return!1}function dKt(n,r,o,s){let c=-1;for(;++c<o.unsafe.length;)if(o.unsafe[c].character===`
`&&uei(o.stack,o.unsafe[c]))return/[ \t]/.test(s.before)?"":" ";return`\\
`}function lei(n,r){const o=String(n);let s=o.indexOf(r),c=s,d=0,h=0;if(typeof r!="string")throw new TypeError("Expected substring");for(;s!==-1;)s===c?++d>h&&(h=d):d=1,c=s+r.length,s=o.indexOf(r,c);return h}function dei(n,r){return!!(r.options.fences===!1&&n.value&&!n.lang&&/[^ \r\n]/.test(n.value)&&!/^[\t ]*(?:[\r\n]|$)|(?:^|[\r\n])[\t ]*$/.test(n.value))}function fei(n){const r=n.options.fence||"`";if(r!=="`"&&r!=="~")throw new Error("Cannot serialize code with `"+r+"` for `options.fence`, expected `` ` `` or `~`");return r}function hei(n,r,o,s){const c=fei(o),d=n.value||"",h=c==="`"?"GraveAccent":"Tilde";if(dei(n,o)){const x=o.enter("codeIndented"),O=o.indentLines(d,pei);return x(),O}const b=o.createTracker(s),y=c.repeat(Math.max(lei(d,c)+1,3)),E=o.enter("codeFenced");let k=b.move(y);if(n.lang){const x=o.enter(`codeFencedLang${h}`);k+=b.move(o.safe(n.lang,{before:k,after:" ",encode:["`"],...b.current()})),x()}if(n.lang&&n.meta){const x=o.enter(`codeFencedMeta${h}`);k+=b.move(" "),k+=b.move(o.safe(n.meta,{before:k,after:`
`,encode:["`"],...b.current()})),x()}return k+=b.move(`
`),d&&(k+=b.move(d+`
`)),k+=b.move(y),E(),k}function pei(n,r,o){return(o?"":"    ")+n}function JQe(n){const r=n.options.quote||'"';if(r!=='"'&&r!=="'")throw new Error("Cannot serialize title with `"+r+"` for `options.quote`, expected `\"`, or `'`");return r}function gei(n,r,o,s){const c=JQe(o),d=c==='"'?"Quote":"Apostrophe",h=o.enter("definition");let b=o.enter("label");const y=o.createTracker(s);let E=y.move("[");return E+=y.move(o.safe(o.associationId(n),{before:E,after:"]",...y.current()})),E+=y.move("]: "),b(),!n.url||/[\0- \u007F]/.test(n.url)?(b=o.enter("destinationLiteral"),E+=y.move("<"),E+=y.move(o.safe(n.url,{before:E,after:">",...y.current()})),E+=y.move(">")):(b=o.enter("destinationRaw"),E+=y.move(o.safe(n.url,{before:E,after:n.title?" ":`
`,...y.current()}))),b(),n.title&&(b=o.enter(`title${d}`),E+=y.move(" "+c),E+=y.move(o.safe(n.title,{before:E,after:c,...y.current()})),E+=y.move(c),b()),h(),E}function mei(n){const r=n.options.emphasis||"*";if(r!=="*"&&r!=="_")throw new Error("Cannot serialize emphasis with `"+r+"` for `options.emphasis`, expected `*`, or `_`");return r}function boe(n){return"&#x"+n.toString(16).toUpperCase()+";"}function Qwe(n,r,o){const s=fK(n),c=fK(r);return s===void 0?c===void 0?o==="_"?{inside:!0,outside:!0}:{inside:!1,outside:!1}:c===1?{inside:!0,outside:!0}:{inside:!1,outside:!0}:s===1?c===void 0?{inside:!1,outside:!1}:c===1?{inside:!0,outside:!0}:{inside:!1,outside:!1}:c===void 0?{inside:!1,outside:!1}:c===1?{inside:!0,outside:!1}:{inside:!1,outside:!1}}vhn.peek=bei;function vhn(n,r,o,s){const c=mei(o),d=o.enter("emphasis"),h=o.createTracker(s),b=h.move(c);let y=h.move(o.containerPhrasing(n,{after:c,before:b,...h.current()}));const E=y.charCodeAt(0),k=Qwe(s.before.charCodeAt(s.before.length-1),E,c);k.inside&&(y=boe(E)+y.slice(1));const x=y.charCodeAt(y.length-1),O=Qwe(s.after.charCodeAt(0),x,c);O.inside&&(y=y.slice(0,-1)+boe(x));const P=h.move(c);return d(),o.attentionEncodeSurroundingInfo={after:O.outside,before:k.outside},b+y+P}function bei(n,r,o){return o.options.emphasis||"*"}function vei(n,r){let o=!1;return WEe(n,function(s){if("value"in s&&/\r?\n|\r/.test(s.value)||s.type==="break")return o=!0,hKe}),!!((!n.depth||n.depth<3)&&MQe(n)&&(r.options.setext||o))}function wei(n,r,o,s){const c=Math.max(Math.min(6,n.depth||1),1),d=o.createTracker(s);if(vei(n,o)){const k=o.enter("headingSetext"),x=o.enter("phrasing"),O=o.containerPhrasing(n,{...d.current(),before:`
`,after:`
`});return x(),k(),O+`
`+(c===1?"=":"-").repeat(O.length-(Math.max(O.lastIndexOf("\r"),O.lastIndexOf(`
`))+1))}const h="#".repeat(c),b=o.enter("headingAtx"),y=o.enter("phrasing");d.move(h+" ");let E=o.containerPhrasing(n,{before:"# ",after:`
`,...d.current()});return/^[\t ]/.test(E)&&(E=boe(E.charCodeAt(0))+E.slice(1)),E=E?h+" "+E:h,o.options.closeAtx&&(E+=" "+h),y(),b(),E}whn.peek=yei;function whn(n){return n.value||""}function yei(){return"<"}yhn.peek=Eei;function yhn(n,r,o,s){const c=JQe(o),d=c==='"'?"Quote":"Apostrophe",h=o.enter("image");let b=o.enter("label");const y=o.createTracker(s);let E=y.move("![");return E+=y.move(o.safe(n.alt,{before:E,after:"]",...y.current()})),E+=y.move("]("),b(),!n.url&&n.title||/[\0- \u007F]/.test(n.url)?(b=o.enter("destinationLiteral"),E+=y.move("<"),E+=y.move(o.safe(n.url,{before:E,after:">",...y.current()})),E+=y.move(">")):(b=o.enter("destinationRaw"),E+=y.move(o.safe(n.url,{before:E,after:n.title?" ":")",...y.current()}))),b(),n.title&&(b=o.enter(`title${d}`),E+=y.move(" "+c),E+=y.move(o.safe(n.title,{before:E,after:c,...y.current()})),E+=y.move(c),b()),E+=y.move(")"),h(),E}function Eei(){return"!"}Ehn.peek=Sei;function Ehn(n,r,o,s){const c=n.referenceType,d=o.enter("imageReference");let h=o.enter("label");const b=o.createTracker(s);let y=b.move("![");const E=o.safe(n.alt,{before:y,after:"]",...b.current()});y+=b.move(E+"]["),h();const k=o.stack;o.stack=[],h=o.enter("reference");const x=o.safe(o.associationId(n),{before:y,after:"]",...b.current()});return h(),o.stack=k,d(),c==="full"||!E||E!==x?y+=b.move(x+"]"):c==="shortcut"?y=y.slice(0,-1):y+=b.move("]"),y}function Sei(){return"!"}Shn.peek=_ei;function Shn(n,r,o){let s=n.value||"",c="`",d=-1;for(;new RegExp("(^|[^`])"+c+"([^`]|$)").test(s);)c+="`";for(/[^ \r\n]/.test(s)&&(/^[ \r\n]/.test(s)&&/[ \r\n]$/.test(s)||/^`|`$/.test(s))&&(s=" "+s+" ");++d<o.unsafe.length;){const h=o.unsafe[d],b=o.compilePattern(h);let y;if(h.atBreak)for(;y=b.exec(s);){let E=y.index;s.charCodeAt(E)===10&&s.charCodeAt(E-1)===13&&E--,s=s.slice(0,E)+" "+s.slice(y.index+1)}}return c+s+c}function _ei(){return"`"}function _hn(n,r){const o=MQe(n);return!!(!r.options.resourceLink&&n.url&&!n.title&&n.children&&n.children.length===1&&n.children[0].type==="text"&&(o===n.url||"mailto:"+o===n.url)&&/^[a-z][a-z+.-]+:/i.test(n.url)&&!/[\0- <>\u007F]/.test(n.url))}khn.peek=kei;function khn(n,r,o,s){const c=JQe(o),d=c==='"'?"Quote":"Apostrophe",h=o.createTracker(s);let b,y;if(_hn(n,o)){const k=o.stack;o.stack=[],b=o.enter("autolink");let x=h.move("<");return x+=h.move(o.containerPhrasing(n,{before:x,after:">",...h.current()})),x+=h.move(">"),b(),o.stack=k,x}b=o.enter("link"),y=o.enter("label");let E=h.move("[");return E+=h.move(o.containerPhrasing(n,{before:E,after:"](",...h.current()})),E+=h.move("]("),y(),!n.url&&n.title||/[\0- \u007F]/.test(n.url)?(y=o.enter("destinationLiteral"),E+=h.move("<"),E+=h.move(o.safe(n.url,{before:E,after:">",...h.current()})),E+=h.move(">")):(y=o.enter("destinationRaw"),E+=h.move(o.safe(n.url,{before:E,after:n.title?" ":")",...h.current()}))),y(),n.title&&(y=o.enter(`title${d}`),E+=h.move(" "+c),E+=h.move(o.safe(n.title,{before:E,after:c,...h.current()})),E+=h.move(c),y()),E+=h.move(")"),b(),E}function kei(n,r,o){return _hn(n,o)?"<":"["}Thn.peek=Tei;function Thn(n,r,o,s){const c=n.referenceType,d=o.enter("linkReference");let h=o.enter("label");const b=o.createTracker(s);let y=b.move("[");const E=o.containerPhrasing(n,{before:y,after:"]",...b.current()});y+=b.move(E+"]["),h();const k=o.stack;o.stack=[],h=o.enter("reference");const x=o.safe(o.associationId(n),{before:y,after:"]",...b.current()});return h(),o.stack=k,d(),c==="full"||!E||E!==x?y+=b.move(x+"]"):c==="shortcut"?y=y.slice(0,-1):y+=b.move("]"),y}function Tei(){return"["}function ZQe(n){const r=n.options.bullet||"*";if(r!=="*"&&r!=="+"&&r!=="-")throw new Error("Cannot serialize items with `"+r+"` for `options.bullet`, expected `*`, `+`, or `-`");return r}function xei(n){const r=ZQe(n),o=n.options.bulletOther;if(!o)return r==="*"?"-":"*";if(o!=="*"&&o!=="+"&&o!=="-")throw new Error("Cannot serialize items with `"+o+"` for `options.bulletOther`, expected `*`, `+`, or `-`");if(o===r)throw new Error("Expected `bullet` (`"+r+"`) and `bulletOther` (`"+o+"`) to be different");return o}function Cei(n){const r=n.options.bulletOrdered||".";if(r!=="."&&r!==")")throw new Error("Cannot serialize items with `"+r+"` for `options.bulletOrdered`, expected `.` or `)`");return r}function xhn(n){const r=n.options.rule||"*";if(r!=="*"&&r!=="-"&&r!=="_")throw new Error("Cannot serialize rules with `"+r+"` for `options.rule`, expected `*`, `-`, or `_`");return r}function Aei(n,r,o,s){const c=o.enter("list"),d=o.bulletCurrent;let h=n.ordered?Cei(o):ZQe(o);const b=n.ordered?h==="."?")":".":xei(o);let y=r&&o.bulletLastUsed?h===o.bulletLastUsed:!1;if(!n.ordered){const k=n.children?n.children[0]:void 0;if((h==="*"||h==="-")&&k&&(!k.children||!k.children[0])&&o.stack[o.stack.length-1]==="list"&&o.stack[o.stack.length-2]==="listItem"&&o.stack[o.stack.length-3]==="list"&&o.stack[o.stack.length-4]==="listItem"&&o.indexStack[o.indexStack.length-1]===0&&o.indexStack[o.indexStack.length-2]===0&&o.indexStack[o.indexStack.length-3]===0&&(y=!0),xhn(o)===h&&k){let x=-1;for(;++x<n.children.length;){const O=n.children[x];if(O&&O.type==="listItem"&&O.children&&O.children[0]&&O.children[0].type==="thematicBreak"){y=!0;break}}}}y&&(h=b),o.bulletCurrent=h;const E=o.containerFlow(n,s);return o.bulletLastUsed=h,o.bulletCurrent=d,c(),E}function Iei(n){const r=n.options.listItemIndent||"one";if(r!=="tab"&&r!=="one"&&r!=="mixed")throw new Error("Cannot serialize items with `"+r+"` for `options.listItemIndent`, expected `tab`, `one`, or `mixed`");return r}function Oei(n,r,o,s){const c=Iei(o);let d=o.bulletCurrent||ZQe(o);r&&r.type==="list"&&r.ordered&&(d=(typeof r.start=="number"&&r.start>-1?r.start:1)+(o.options.incrementListMarker===!1?0:r.children.indexOf(n))+d);let h=d.length+1;(c==="tab"||c==="mixed"&&(r&&r.type==="list"&&r.spread||n.spread))&&(h=Math.ceil(h/4)*4);const b=o.createTracker(s);b.move(d+" ".repeat(h-d.length)),b.shift(h);const y=o.enter("listItem"),E=o.indentLines(o.containerFlow(n,b.current()),k);return y(),E;function k(x,O,P){return O?(P?"":" ".repeat(h))+x:(P?d:d+" ".repeat(h-d.length))+x}}function Nei(n,r,o,s){const c=o.enter("paragraph"),d=o.enter("phrasing"),h=o.containerPhrasing(n,s);return d(),c(),h}const Rei=qEe(["break","delete","emphasis","footnote","footnoteReference","image","imageReference","inlineCode","inlineMath","link","linkReference","mdxJsxTextElement","mdxTextExpression","strong","text","textDirective"]);function Pei(n,r,o,s){return(n.children.some(function(h){return Rei(h)})?o.containerPhrasing:o.containerFlow).call(o,n,s)}function Lei(n){const r=n.options.strong||"*";if(r!=="*"&&r!=="_")throw new Error("Cannot serialize strong with `"+r+"` for `options.strong`, expected `*`, or `_`");return r}Chn.peek=Dei;function Chn(n,r,o,s){const c=Lei(o),d=o.enter("strong"),h=o.createTracker(s),b=h.move(c+c);let y=h.move(o.containerPhrasing(n,{after:c,before:b,...h.current()}));const E=y.charCodeAt(0),k=Qwe(s.before.charCodeAt(s.before.length-1),E,c);k.inside&&(y=boe(E)+y.slice(1));const x=y.charCodeAt(y.length-1),O=Qwe(s.after.charCodeAt(0),x,c);O.inside&&(y=y.slice(0,-1)+boe(x));const P=h.move(c+c);return d(),o.attentionEncodeSurroundingInfo={after:O.outside,before:k.outside},b+y+P}function Dei(n,r,o){return o.options.strong||"*"}function Mei(n,r,o,s){return o.safe(n.value,s)}function jei(n){const r=n.options.ruleRepetition||3;if(r<3)throw new Error("Cannot serialize rules with repetition `"+r+"` for `options.ruleRepetition`, expected `3` or more");return r}function $ei(n,r,o){const s=(xhn(o)+(o.options.ruleSpaces?" ":"")).repeat(jei(o));return o.options.ruleSpaces?s.slice(0,-1):s}const Ahn={blockquote:sei,break:dKt,code:hei,definition:gei,emphasis:vhn,hardBreak:dKt,heading:wei,html:whn,image:yhn,imageReference:Ehn,inlineCode:Shn,link:khn,linkReference:Thn,list:Aei,listItem:Oei,paragraph:Nei,root:Pei,strong:Chn,text:Mei,thematicBreak:$ei};function Fei(){return{enter:{table:Bei,tableData:fKt,tableHeader:fKt,tableRow:zei},exit:{codeText:Hei,table:Uei,tableData:xqe,tableHeader:xqe,tableRow:xqe}}}function Bei(n){const r=n._align;this.enter({type:"table",align:r.map(function(o){return o==="none"?null:o}),children:[]},n),this.data.inTable=!0}function Uei(n){this.exit(n),this.data.inTable=void 0}function zei(n){this.enter({type:"tableRow",children:[]},n)}function xqe(n){this.exit(n)}function fKt(n){this.enter({type:"tableCell",children:[]},n)}function Hei(n){let r=this.resume();this.data.inTable&&(r=r.replace(/\\([\\|])/g,Gei));const o=this.stack[this.stack.length-1];o.type,o.value=r,this.exit(n)}function Gei(n,r){return r==="|"?r:n}function qei(n){const r=n||{},o=r.tableCellPadding,s=r.tablePipeAlign,c=r.stringLength,d=o?" ":"|";return{unsafe:[{character:"\r",inConstruct:"tableCell"},{character:`
`,inConstruct:"tableCell"},{atBreak:!0,character:"|",after:"[	 :-]"},{character:"|",inConstruct:"tableCell"},{atBreak:!0,character:":",after:"-"},{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{inlineCode:O,table:h,tableCell:y,tableRow:b}};function h(P,D,$,U){return E(k(P,$,U),P.align)}function b(P,D,$,U){const B=x(P,$,U),G=E([B]);return G.slice(0,G.indexOf(`
`))}function y(P,D,$,U){const B=$.enter("tableCell"),G=$.enter("phrasing"),K=$.containerPhrasing(P,{...U,before:d,after:d});return G(),B(),K}function E(P,D){return oei(P,{align:D,alignDelimiters:s,padding:o,stringLength:c})}function k(P,D,$){const U=P.children;let B=-1;const G=[],K=D.enter("table");for(;++B<U.length;)G[B]=x(U[B],D,$);return K(),G}function x(P,D,$){const U=P.children;let B=-1;const G=[],K=D.enter("tableRow");for(;++B<U.length;)G[B]=y(U[B],P,D,$);return K(),G}function O(P,D,$){let U=Ahn.inlineCode(P,D,$);return $.stack.includes("tableCell")&&(U=U.replace(/\|/g,"\\$&")),U}}function Vei(){return{exit:{taskListCheckValueChecked:hKt,taskListCheckValueUnchecked:hKt,paragraph:Kei}}}function Wei(){return{unsafe:[{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{listItem:Yei}}}function hKt(n){const r=this.stack[this.stack.length-2];r.type,r.checked=n.type==="taskListCheckValueChecked"}function Kei(n){const r=this.stack[this.stack.length-2];if(r&&r.type==="listItem"&&typeof r.checked=="boolean"){const o=this.stack[this.stack.length-1];o.type;const s=o.children[0];if(s&&s.type==="text"){const c=r.children;let d=-1,h;for(;++d<c.length;){const b=c[d];if(b.type==="paragraph"){h=b;break}}h===o&&(s.value=s.value.slice(1),s.value.length===0?o.children.shift():o.position&&s.position&&typeof s.position.start.offset=="number"&&(s.position.start.column++,s.position.start.offset++,o.position.start=Object.assign({},s.position.start)))}}this.exit(n)}function Yei(n,r,o,s){const c=n.children[0],d=typeof n.checked=="boolean"&&c&&c.type==="paragraph",h="["+(n.checked?"x":" ")+"] ",b=o.createTracker(s);d&&b.move(h);let y=Ahn.listItem(n,r,o,{...s,...b.current()});return d&&(y=y.replace(/^(?:[*+-]|\d+\.)([\r\n]| {1,3})/,E)),y;function E(k){return k+h}}function Xei(){return[AQr(),YQr(),QQr(),Fei(),Vei()]}function Jei(n){return{extensions:[IQr(),XQr(n),eei(),qei(n),Wei()]}}const Zei={tokenize:iti,partial:!0},Ihn={tokenize:oti,partial:!0},Ohn={tokenize:ati,partial:!0},Nhn={tokenize:sti,partial:!0},Qei={tokenize:cti,partial:!0},Rhn={name:"wwwAutolink",tokenize:nti,previous:Lhn},Phn={name:"protocolAutolink",tokenize:rti,previous:Dhn},t6={name:"emailAutolink",tokenize:tti,previous:Mhn},KI={};function eti(){return{text:KI}}let yB=48;for(;yB<123;)KI[yB]=t6,yB++,yB===58?yB=65:yB===91&&(yB=97);KI[43]=t6;KI[45]=t6;KI[46]=t6;KI[95]=t6;KI[72]=[t6,Phn];KI[104]=[t6,Phn];KI[87]=[t6,Rhn];KI[119]=[t6,Rhn];function tti(n,r,o){const s=this;let c,d;return h;function h(x){return!kKe(x)||!Mhn.call(s,s.previous)||QQe(s.events)?o(x):(n.enter("literalAutolink"),n.enter("literalAutolinkEmail"),b(x))}function b(x){return kKe(x)?(n.consume(x),b):x===64?(n.consume(x),y):o(x)}function y(x){return x===46?n.check(Qei,k,E)(x):x===45||x===95||K1(x)?(d=!0,n.consume(x),y):k(x)}function E(x){return n.consume(x),c=!0,y}function k(x){return d&&c&&sv(s.previous)?(n.exit("literalAutolinkEmail"),n.exit("literalAutolink"),r(x)):o(x)}}function nti(n,r,o){const s=this;return c;function c(h){return h!==87&&h!==119||!Lhn.call(s,s.previous)||QQe(s.events)?o(h):(n.enter("literalAutolink"),n.enter("literalAutolinkWww"),n.check(Zei,n.attempt(Ihn,n.attempt(Ohn,d),o),o)(h))}function d(h){return n.exit("literalAutolinkWww"),n.exit("literalAutolink"),r(h)}}function rti(n,r,o){const s=this;let c="",d=!1;return h;function h(x){return(x===72||x===104)&&Dhn.call(s,s.previous)&&!QQe(s.events)?(n.enter("literalAutolink"),n.enter("literalAutolinkHttp"),c+=String.fromCodePoint(x),n.consume(x),b):o(x)}function b(x){if(sv(x)&&c.length<5)return c+=String.fromCodePoint(x),n.consume(x),b;if(x===58){const O=c.toLowerCase();if(O==="http"||O==="https")return n.consume(x),y}return o(x)}function y(x){return x===47?(n.consume(x),d?E:(d=!0,y)):o(x)}function E(x){return x===null||Kwe(x)||cd(x)||sU(x)||zEe(x)?o(x):n.attempt(Ihn,n.attempt(Ohn,k),o)(x)}function k(x){return n.exit("literalAutolinkHttp"),n.exit("literalAutolink"),r(x)}}function iti(n,r,o){let s=0;return c;function c(h){return(h===87||h===119)&&s<3?(s++,n.consume(h),c):h===46&&s===3?(n.consume(h),d):o(h)}function d(h){return h===null?o(h):r(h)}}function oti(n,r,o){let s,c,d;return h;function h(E){return E===46||E===95?n.check(Nhn,y,b)(E):E===null||cd(E)||sU(E)||E!==45&&zEe(E)?y(E):(d=!0,n.consume(E),h)}function b(E){return E===95?s=!0:(c=s,s=void 0),n.consume(E),h}function y(E){return c||s||!d?o(E):r(E)}}function ati(n,r){let o=0,s=0;return c;function c(h){return h===40?(o++,n.consume(h),c):h===41&&s<o?d(h):h===33||h===34||h===38||h===39||h===41||h===42||h===44||h===46||h===58||h===59||h===60||h===63||h===93||h===95||h===126?n.check(Nhn,r,d)(h):h===null||cd(h)||sU(h)?r(h):(n.consume(h),c)}function d(h){return h===41&&s++,n.consume(h),c}}function sti(n,r,o){return s;function s(b){return b===33||b===34||b===39||b===41||b===42||b===44||b===46||b===58||b===59||b===63||b===95||b===126?(n.consume(b),s):b===38?(n.consume(b),d):b===93?(n.consume(b),c):b===60||b===null||cd(b)||sU(b)?r(b):o(b)}function c(b){return b===null||b===40||b===91||cd(b)||sU(b)?r(b):s(b)}function d(b){return sv(b)?h(b):o(b)}function h(b){return b===59?(n.consume(b),s):sv(b)?(n.consume(b),h):o(b)}}function cti(n,r,o){return s;function s(d){return n.consume(d),c}function c(d){return K1(d)?o(d):r(d)}}function Lhn(n){return n===null||n===40||n===42||n===95||n===91||n===93||n===126||cd(n)}function Dhn(n){return!sv(n)}function Mhn(n){return!(n===47||kKe(n))}function kKe(n){return n===43||n===45||n===46||n===95||K1(n)}function QQe(n){let r=n.length,o=!1;for(;r--;){const s=n[r][1];if((s.type==="labelLink"||s.type==="labelImage")&&!s._balanced){o=!0;break}if(s._gfmAutolinkLiteralWalkedInto){o=!1;break}}return n.length>0&&!o&&(n[n.length-1][1]._gfmAutolinkLiteralWalkedInto=!0),o}const uti={tokenize:bti,partial:!0};function lti(){return{document:{91:{name:"gfmFootnoteDefinition",tokenize:pti,continuation:{tokenize:gti},exit:mti}},text:{91:{name:"gfmFootnoteCall",tokenize:hti},93:{name:"gfmPotentialFootnoteCall",add:"after",tokenize:dti,resolveTo:fti}}}}function dti(n,r,o){const s=this;let c=s.events.length;const d=s.parser.gfmFootnotes||(s.parser.gfmFootnotes=[]);let h;for(;c--;){const y=s.events[c][1];if(y.type==="labelImage"){h=y;break}if(y.type==="gfmFootnoteCall"||y.type==="labelLink"||y.type==="label"||y.type==="image"||y.type==="link")break}return b;function b(y){if(!h||!h._balanced)return o(y);const E=zx(s.sliceSerialize({start:h.end,end:s.now()}));return E.codePointAt(0)!==94||!d.includes(E.slice(1))?o(y):(n.enter("gfmFootnoteCallLabelMarker"),n.consume(y),n.exit("gfmFootnoteCallLabelMarker"),r(y))}}function fti(n,r){let o=n.length;for(;o--;)if(n[o][1].type==="labelImage"&&n[o][0]==="enter"){n[o][1];break}n[o+1][1].type="data",n[o+3][1].type="gfmFootnoteCallLabelMarker";const s={type:"gfmFootnoteCall",start:Object.assign({},n[o+3][1].start),end:Object.assign({},n[n.length-1][1].end)},c={type:"gfmFootnoteCallMarker",start:Object.assign({},n[o+3][1].end),end:Object.assign({},n[o+3][1].end)};c.end.column++,c.end.offset++,c.end._bufferIndex++;const d={type:"gfmFootnoteCallString",start:Object.assign({},c.end),end:Object.assign({},n[n.length-1][1].start)},h={type:"chunkString",contentType:"string",start:Object.assign({},d.start),end:Object.assign({},d.end)},b=[n[o+1],n[o+2],["enter",s,r],n[o+3],n[o+4],["enter",c,r],["exit",c,r],["enter",d,r],["enter",h,r],["exit",h,r],["exit",d,r],n[n.length-2],n[n.length-1],["exit",s,r]];return n.splice(o,n.length-o+1,...b),n}function hti(n,r,o){const s=this,c=s.parser.gfmFootnotes||(s.parser.gfmFootnotes=[]);let d=0,h;return b;function b(x){return n.enter("gfmFootnoteCall"),n.enter("gfmFootnoteCallLabelMarker"),n.consume(x),n.exit("gfmFootnoteCallLabelMarker"),y}function y(x){return x!==94?o(x):(n.enter("gfmFootnoteCallMarker"),n.consume(x),n.exit("gfmFootnoteCallMarker"),n.enter("gfmFootnoteCallString"),n.enter("chunkString").contentType="string",E)}function E(x){if(d>999||x===93&&!h||x===null||x===91||cd(x))return o(x);if(x===93){n.exit("chunkString");const O=n.exit("gfmFootnoteCallString");return c.includes(zx(s.sliceSerialize(O)))?(n.enter("gfmFootnoteCallLabelMarker"),n.consume(x),n.exit("gfmFootnoteCallLabelMarker"),n.exit("gfmFootnoteCall"),r):o(x)}return cd(x)||(h=!0),d++,n.consume(x),x===92?k:E}function k(x){return x===91||x===92||x===93?(n.consume(x),d++,E):E(x)}}function pti(n,r,o){const s=this,c=s.parser.gfmFootnotes||(s.parser.gfmFootnotes=[]);let d,h=0,b;return y;function y(D){return n.enter("gfmFootnoteDefinition")._container=!0,n.enter("gfmFootnoteDefinitionLabel"),n.enter("gfmFootnoteDefinitionLabelMarker"),n.consume(D),n.exit("gfmFootnoteDefinitionLabelMarker"),E}function E(D){return D===94?(n.enter("gfmFootnoteDefinitionMarker"),n.consume(D),n.exit("gfmFootnoteDefinitionMarker"),n.enter("gfmFootnoteDefinitionLabelString"),n.enter("chunkString").contentType="string",k):o(D)}function k(D){if(h>999||D===93&&!b||D===null||D===91||cd(D))return o(D);if(D===93){n.exit("chunkString");const $=n.exit("gfmFootnoteDefinitionLabelString");return d=zx(s.sliceSerialize($)),n.enter("gfmFootnoteDefinitionLabelMarker"),n.consume(D),n.exit("gfmFootnoteDefinitionLabelMarker"),n.exit("gfmFootnoteDefinitionLabel"),O}return cd(D)||(b=!0),h++,n.consume(D),D===92?x:k}function x(D){return D===91||D===92||D===93?(n.consume(D),h++,k):k(D)}function O(D){return D===58?(n.enter("definitionMarker"),n.consume(D),n.exit("definitionMarker"),c.includes(d)||c.push(d),Kc(n,P,"gfmFootnoteDefinitionWhitespace")):o(D)}function P(D){return r(D)}}function gti(n,r,o){return n.check(Joe,r,n.attempt(uti,r,o))}function mti(n){n.exit("gfmFootnoteDefinition")}function bti(n,r,o){const s=this;return Kc(n,c,"gfmFootnoteDefinitionIndent",5);function c(d){const h=s.events[s.events.length-1];return h&&h[1].type==="gfmFootnoteDefinitionIndent"&&h[2].sliceSerialize(h[1],!0).length===4?r(d):o(d)}}function vti(n){let o=(n||{}).singleTilde;const s={name:"strikethrough",tokenize:d,resolveAll:c};return o==null&&(o=!0),{text:{126:s},insideSpan:{null:[s]},attentionMarkers:{null:[126]}};function c(h,b){let y=-1;for(;++y<h.length;)if(h[y][0]==="enter"&&h[y][1].type==="strikethroughSequenceTemporary"&&h[y][1]._close){let E=y;for(;E--;)if(h[E][0]==="exit"&&h[E][1].type==="strikethroughSequenceTemporary"&&h[E][1]._open&&h[y][1].end.offset-h[y][1].start.offset===h[E][1].end.offset-h[E][1].start.offset){h[y][1].type="strikethroughSequence",h[E][1].type="strikethroughSequence";const k={type:"strikethrough",start:Object.assign({},h[E][1].start),end:Object.assign({},h[y][1].end)},x={type:"strikethroughText",start:Object.assign({},h[E][1].end),end:Object.assign({},h[y][1].start)},O=[["enter",k,b],["enter",h[E][1],b],["exit",h[E][1],b],["enter",x,b]],P=b.parser.constructs.insideSpan.null;P&&p2(O,O.length,0,HEe(P,h.slice(E+1,y),b)),p2(O,O.length,0,[["exit",x,b],["enter",h[y][1],b],["exit",h[y][1],b],["exit",k,b]]),p2(h,E-1,y-E+3,O),y=E+O.length-2;break}}for(y=-1;++y<h.length;)h[y][1].type==="strikethroughSequenceTemporary"&&(h[y][1].type="data");return h}function d(h,b,y){const E=this.previous,k=this.events;let x=0;return O;function O(D){return E===126&&k[k.length-1][1].type!=="characterEscape"?y(D):(h.enter("strikethroughSequenceTemporary"),P(D))}function P(D){const $=fK(E);if(D===126)return x>1?y(D):(h.consume(D),x++,P);if(x<2&&!o)return y(D);const U=h.exit("strikethroughSequenceTemporary"),B=fK(D);return U._open=!B||B===2&&!!$,U._close=!$||$===2&&!!B,b(D)}}}class wti{constructor(){this.map=[]}add(r,o,s){yti(this,r,o,s)}consume(r){if(this.map.sort(function(d,h){return d[0]-h[0]}),this.map.length===0)return;let o=this.map.length;const s=[];for(;o>0;)o-=1,s.push(r.slice(this.map[o][0]+this.map[o][1]),this.map[o][2]),r.length=this.map[o][0];s.push(r.slice()),r.length=0;let c=s.pop();for(;c;){for(const d of c)r.push(d);c=s.pop()}this.map.length=0}}function yti(n,r,o,s){let c=0;if(!(o===0&&s.length===0)){for(;c<n.map.length;){if(n.map[c][0]===r){n.map[c][1]+=o,n.map[c][2].push(...s);return}c+=1}n.map.push([r,o,s])}}function Eti(n,r){let o=!1;const s=[];for(;r<n.length;){const c=n[r];if(o){if(c[0]==="enter")c[1].type==="tableContent"&&s.push(n[r+1][1].type==="tableDelimiterMarker"?"left":"none");else if(c[1].type==="tableContent"){if(n[r-1][1].type==="tableDelimiterMarker"){const d=s.length-1;s[d]=s[d]==="left"?"center":"right"}}else if(c[1].type==="tableDelimiterRow")break}else c[0]==="enter"&&c[1].type==="tableDelimiterRow"&&(o=!0);r+=1}return s}function Sti(){return{flow:{null:{name:"table",tokenize:_ti,resolveAll:kti}}}}function _ti(n,r,o){const s=this;let c=0,d=0,h;return b;function b(ge){let Te=s.events.length-1;for(;Te>-1;){const Z=s.events[Te][1].type;if(Z==="lineEnding"||Z==="linePrefix")Te--;else break}const xe=Te>-1?s.events[Te][1].type:null,Ue=xe==="tableHead"||xe==="tableRow"?oe:y;return Ue===oe&&s.parser.lazy[s.now().line]?o(ge):Ue(ge)}function y(ge){return n.enter("tableHead"),n.enter("tableRow"),E(ge)}function E(ge){return ge===124||(h=!0,d+=1),k(ge)}function k(ge){return ge===null?o(ge):ea(ge)?d>1?(d=0,s.interrupt=!0,n.exit("tableRow"),n.enter("lineEnding"),n.consume(ge),n.exit("lineEnding"),P):o(ge):mc(ge)?Kc(n,k,"whitespace")(ge):(d+=1,h&&(h=!1,c+=1),ge===124?(n.enter("tableCellDivider"),n.consume(ge),n.exit("tableCellDivider"),h=!0,k):(n.enter("data"),x(ge)))}function x(ge){return ge===null||ge===124||cd(ge)?(n.exit("data"),k(ge)):(n.consume(ge),ge===92?O:x)}function O(ge){return ge===92||ge===124?(n.consume(ge),x):x(ge)}function P(ge){return s.interrupt=!1,s.parser.lazy[s.now().line]?o(ge):(n.enter("tableDelimiterRow"),h=!1,mc(ge)?Kc(n,D,"linePrefix",s.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(ge):D(ge))}function D(ge){return ge===45||ge===58?U(ge):ge===124?(h=!0,n.enter("tableCellDivider"),n.consume(ge),n.exit("tableCellDivider"),$):ae(ge)}function $(ge){return mc(ge)?Kc(n,U,"whitespace")(ge):U(ge)}function U(ge){return ge===58?(d+=1,h=!0,n.enter("tableDelimiterMarker"),n.consume(ge),n.exit("tableDelimiterMarker"),B):ge===45?(d+=1,B(ge)):ge===null||ea(ge)?X(ge):ae(ge)}function B(ge){return ge===45?(n.enter("tableDelimiterFiller"),G(ge)):ae(ge)}function G(ge){return ge===45?(n.consume(ge),G):ge===58?(h=!0,n.exit("tableDelimiterFiller"),n.enter("tableDelimiterMarker"),n.consume(ge),n.exit("tableDelimiterMarker"),K):(n.exit("tableDelimiterFiller"),K(ge))}function K(ge){return mc(ge)?Kc(n,X,"whitespace")(ge):X(ge)}function X(ge){return ge===124?D(ge):ge===null||ea(ge)?!h||c!==d?ae(ge):(n.exit("tableDelimiterRow"),n.exit("tableHead"),r(ge)):ae(ge)}function ae(ge){return o(ge)}function oe(ge){return n.enter("tableRow"),he(ge)}function he(ge){return ge===124?(n.enter("tableCellDivider"),n.consume(ge),n.exit("tableCellDivider"),he):ge===null||ea(ge)?(n.exit("tableRow"),r(ge)):mc(ge)?Kc(n,he,"whitespace")(ge):(n.enter("data"),ye(ge))}function ye(ge){return ge===null||ge===124||cd(ge)?(n.exit("data"),he(ge)):(n.consume(ge),ge===92?we:ye)}function we(ge){return ge===92||ge===124?(n.consume(ge),ye):ye(ge)}}function kti(n,r){let o=-1,s=!0,c=0,d=[0,0,0,0],h=[0,0,0,0],b=!1,y=0,E,k,x;const O=new wti;for(;++o<n.length;){const P=n[o],D=P[1];P[0]==="enter"?D.type==="tableHead"?(b=!1,y!==0&&(pKt(O,r,y,E,k),k=void 0,y=0),E={type:"table",start:Object.assign({},D.start),end:Object.assign({},D.end)},O.add(o,0,[["enter",E,r]])):D.type==="tableRow"||D.type==="tableDelimiterRow"?(s=!0,x=void 0,d=[0,0,0,0],h=[0,o+1,0,0],b&&(b=!1,k={type:"tableBody",start:Object.assign({},D.start),end:Object.assign({},D.end)},O.add(o,0,[["enter",k,r]])),c=D.type==="tableDelimiterRow"?2:k?3:1):c&&(D.type==="data"||D.type==="tableDelimiterMarker"||D.type==="tableDelimiterFiller")?(s=!1,h[2]===0&&(d[1]!==0&&(h[0]=h[1],x=Y0e(O,r,d,c,void 0,x),d=[0,0,0,0]),h[2]=o)):D.type==="tableCellDivider"&&(s?s=!1:(d[1]!==0&&(h[0]=h[1],x=Y0e(O,r,d,c,void 0,x)),d=h,h=[d[1],o,0,0])):D.type==="tableHead"?(b=!0,y=o):D.type==="tableRow"||D.type==="tableDelimiterRow"?(y=o,d[1]!==0?(h[0]=h[1],x=Y0e(O,r,d,c,o,x)):h[1]!==0&&(x=Y0e(O,r,h,c,o,x)),c=0):c&&(D.type==="data"||D.type==="tableDelimiterMarker"||D.type==="tableDelimiterFiller")&&(h[3]=o)}for(y!==0&&pKt(O,r,y,E,k),O.consume(r.events),o=-1;++o<r.events.length;){const P=r.events[o];P[0]==="enter"&&P[1].type==="table"&&(P[1]._align=Eti(r.events,o))}return n}function Y0e(n,r,o,s,c,d){const h=s===1?"tableHeader":s===2?"tableDelimiter":"tableData",b="tableContent";o[0]!==0&&(d.end=Object.assign({},sW(r.events,o[0])),n.add(o[0],0,[["exit",d,r]]));const y=sW(r.events,o[1]);if(d={type:h,start:Object.assign({},y),end:Object.assign({},y)},n.add(o[1],0,[["enter",d,r]]),o[2]!==0){const E=sW(r.events,o[2]),k=sW(r.events,o[3]),x={type:b,start:Object.assign({},E),end:Object.assign({},k)};if(n.add(o[2],0,[["enter",x,r]]),s!==2){const O=r.events[o[2]],P=r.events[o[3]];if(O[1].end=Object.assign({},P[1].end),O[1].type="chunkText",O[1].contentType="text",o[3]>o[2]+1){const D=o[2]+1,$=o[3]-o[2]-1;n.add(D,$,[])}}n.add(o[3]+1,0,[["exit",x,r]])}return c!==void 0&&(d.end=Object.assign({},sW(r.events,c)),n.add(c,0,[["exit",d,r]]),d=void 0),d}function pKt(n,r,o,s,c){const d=[],h=sW(r.events,o);c&&(c.end=Object.assign({},h),d.push(["exit",c,r])),s.end=Object.assign({},h),d.push(["exit",s,r]),n.add(o+1,0,d)}function sW(n,r){const o=n[r],s=o[0]==="enter"?"start":"end";return o[1][s]}const Tti={name:"tasklistCheck",tokenize:Cti};function xti(){return{text:{91:Tti}}}function Cti(n,r,o){const s=this;return c;function c(y){return s.previous!==null||!s._gfmTasklistFirstContentOfListItem?o(y):(n.enter("taskListCheck"),n.enter("taskListCheckMarker"),n.consume(y),n.exit("taskListCheckMarker"),d)}function d(y){return cd(y)?(n.enter("taskListCheckValueUnchecked"),n.consume(y),n.exit("taskListCheckValueUnchecked"),h):y===88||y===120?(n.enter("taskListCheckValueChecked"),n.consume(y),n.exit("taskListCheckValueChecked"),h):o(y)}function h(y){return y===93?(n.enter("taskListCheckMarker"),n.consume(y),n.exit("taskListCheckMarker"),n.exit("taskListCheck"),b):o(y)}function b(y){return ea(y)?r(y):mc(y)?n.check({tokenize:Ati},r,o)(y):o(y)}}function Ati(n,r,o){return Kc(n,s,"whitespace");function s(c){return c===null?o(c):r(c)}}function Iti(n){return nfn([eti(),lti(),vti(n),Sti(),xti()])}const Oti={};function Nti(n){const r=this,o=n||Oti,s=r.data(),c=s.micromarkExtensions||(s.micromarkExtensions=[]),d=s.fromMarkdownExtensions||(s.fromMarkdownExtensions=[]),h=s.toMarkdownExtensions||(s.toMarkdownExtensions=[]);c.push(Iti(o)),d.push(Xei()),h.push(Jei(o))}const Rti=n=>n.replace(/^\/+/,"").trim(),Pti=(n,r)=>{if(!r?.position?.start?.offset)return!1;const o=r.position.start.offset,s=n.slice(Math.max(0,o-20),o);return/\*\*File:\s*`?$|File:\s*`?$/.test(s)},Lti=({filePath:n,children:r})=>{const{setOpenFile:o}=bU(),s=Io(gc,x=>x.files),c=y2(),{projectId:d}=G_(),{getPath:h}=MK(),b=Q.useCallback(x=>{o(x),c(h(q4.Code.toLowerCase()))},[o,c,h]),y=Q.useCallback(x=>{x.preventDefault();const O=Rti(n),P=s?.some(D=>D.path===O);if(d)if(z_(H_.fileLinkClicked,{filePath:O,fileExists:P}),P)b(O);else try{_qr(O,s??null),b(O)}catch(D){Va(D,{tags:{component:"FilePathLink",action:"createFile"}})}},[n,s,d,b]),E=x=>{if(typeof x=="string")return x;if(Un.isValidElement(x)){const O=x.props;return E(O.children)}return""},k=E(r);return H.jsx(_I,{href:"#",onClick:y,cursor:"pointer",textDecoration:"none",display:"inline-flex",_hover:{textDecoration:"none"},children:H.jsxs($c,{as:"span",gap:"0.5rem",px:"0.5rem",bg:"white",border:"1px solid",borderColor:"rasaNeutral.400",borderRadius:"0.5rem",fontSize:"0.813rem",fontWeight:"normal",color:"rasawebDeepPurple.800",_hover:{bg:"#f6f7fb"},lineHeight:"150%",children:[H.jsx(xa,{icon:Oor}),k]})})},Dti=n=>typeof n=="string"&&/^\d+$/.test(n),gKt=n=>typeof n=="string"?n:n&&Array.isArray(n)?n.join(""):"",Mti=({content:n})=>H.jsx(ai,{children:H.jsx(XYr,{remarkPlugins:[Nti],rehypePlugins:[SQr],components:{p:({children:r})=>H.jsx(Fi,{as:"p",fontSize:QE,pb:"0.5rem",color:"rasawebDeepPurple.800",lineHeight:aU,children:r}),h1:({children:r})=>H.jsx(Fi,{as:"h1",fontSize:"lg",color:"rasawebDeepPurple.800",children:r}),h2:({children:r})=>H.jsx(Fi,{as:"h2",fontSize:"md",color:"rasawebDeepPurple.800",children:r}),h3:({children:r})=>H.jsx(Fi,{as:"h3",fontSize:QE,color:"rasawebDeepPurple.800",children:r}),h4:({children:r})=>H.jsx(Fi,{as:"h4",fontSize:QE,color:"rasawebDeepPurple.800",children:r}),h5:({children:r})=>H.jsx(Fi,{as:"h5",fontSize:QE,color:"rasawebDeepPurple.800",children:r}),h6:({children:r})=>H.jsx(Fi,{as:"h6",fontSize:QE,color:"rasawebDeepPurple.800",children:r}),ul:({children:r})=>H.jsx(Xve,{as:"ul",pl:"1.5rem",fontSize:QE,color:"rasawebDeepPurple.800",children:r}),ol:({children:r})=>H.jsx(Xve,{as:"ol",pl:"1.5rem",fontSize:QE,color:"rasawebDeepPurple.800",children:r}),li:({children:r})=>H.jsx(CYe,{as:"li",pb:"0.25rem",fontSize:QE,color:"rasawebDeepPurple.800",lineHeight:aU,children:r}),code:({className:r,children:o,node:s})=>H.jsx(jti,{className:r||"",content:n,node:s,children:o}),hr:()=>H.jsx(ai,{h:"0.5rem"}),a:({children:r,href:o})=>H.jsx($ti,{href:o||"",children:r})},children:n})}),jti=({children:n,className:r,content:o,node:s})=>{const c=/language-(\w+)/.exec(r||"");if(c){const d=gKt(n);return H.jsx(FW,{language:c[1],code:d})}else{const d=gKt(n);return Pti(o,s)?H.jsx(Lti,{filePath:d,children:H.jsx(lVe,{fontSize:Wwe,px:"0.5rem",borderRadius:"0.25rem",children:n})}):H.jsx(lVe,{fontSize:Wwe,px:"0.5rem",borderRadius:"0.5rem",bgColor:"rasaNeutral.200",children:n})}},$ti=({children:n,href:r})=>{const o=Un.Children.toArray(n)[0],s=Dti(o);return H.jsx(_I,{href:r,target:"_blank",variant:"underline",fontSize:QE,color:"rasawebDeepPurple.800",children:s?`[${o}]`:H.jsxs(H.Fragment,{children:[n,H.jsx(xa,{icon:Uye})]})})},Fti=({references:n})=>{if(!n||n.length===0)return null;const r=n.map((o,s)=>H.jsx(CYe,{children:H.jsxs(_I,{href:o.url,target:"_blank",variant:"underline",fontSize:QE,color:"rasawebDeepPurple.800",lineHeight:aU,children:["[",o.index,"] ",o.title,H.jsx(xa,{color:"rasawebDeepPurple.800",size:"xs",icon:Uye})]})},`${o.index}-${s}`));return H.jsxs(ai,{pt:"1rem",children:[H.jsx(Fi,{fontSize:"md",fontWeight:"bold",mb:"0.5rem",children:"References"}),H.jsx(Xve,{children:r})]})},Bti=({selectedIssue:n,otherText:r,setSelectedIssue:o,setOtherText:s})=>H.jsxs(ai,{onClick:c=>c.stopPropagation(),children:[H.jsx(Vd,{gap:"0.5rem",align:"start",pb:"1rem",children:H.jsx(Fi,{children:"Tell us why this response wasn't helpful to you."})}),H.jsx(ai,{w:"100%","data-testid":"copilot-feedback-form",children:H.jsx(Sen,{children:H.jsxs(ai,{children:[H.jsx(sie,{required:!0,children:H.jsx(NYe,{value:n,onValueChange:c=>o(c.value||""),w:"100%",children:H.jsx(ai,{display:"flex",flexWrap:"wrap",gap:"0.75rem",children:Uqr.map(c=>H.jsx(RYe,{value:c.value,asChild:!0,children:H.jsx(Vln,{item:{value:c.value,label:c.label},isActive:n===c.value,onActivate:d=>o(d),size:"sm"})},c.value))})})}),n===Ox.Other&&H.jsx(sie,{mt:"1.75rem",children:H.jsx(DYe,{name:"otherFeedback",value:r,onChange:c=>s(c.target.value),placeholder:"Please describe the issue...",size:"sm",resize:"none",rows:2,color:"rasawebDeepPurple.800",borderColor:"rasaNeutral.300",borderRadius:"0.5rem",_focus:{outline:"none"},_placeholder:{color:"rasaNeutral.700"},css:{animation:"fadeIn 0.2s ease-out","@keyframes fadeIn":{"0%":{opacity:0,transform:"translateY(-4px)"},"100%":{opacity:1,transform:"translateY(0)"}}}})})]})})})]}),Uti=[{label:"Positive",value:Kie.Positive,icon:inn},{label:"Negative",value:Kie.Negative,icon:Ior}],zti=({messageId:n})=>{const[r,o]=Q.useState(null),{open:s,onOpen:c,onClose:d}=_M(),[h,b]=Q.useState(""),[y,E]=Q.useState(""),[k,x]=Q.useState(!1),O=()=>{b(""),E(""),x(!1)},P=async G=>{if(o(G),G===Kie.Positive){d(),Kp.create({title:_0e.positive.title,description:_0e.positive.description,type:"success",duration:4e3});const K={messageId:n||"unknown",rating:"positive",issue:Ox.None,timestamp:new Date().toISOString()};await z_(H_.copilotFeedback,K)}else G===Kie.Negative&&c()},D=async()=>{const G={messageId:n||"unknown",rating:"negative",issue:h||Ox.None,otherFeedback:h===Ox.Other?y:void 0,timestamp:new Date().toISOString()};try{await z_(H_.copilotFeedback,G),Kp.create({title:_0e.negative.title,description:_0e.negative.description,type:"success",duration:4e3}),d(),O()}catch(K){Kp.create({title:"Failed to submit feedback",description:"Please try again or contact support if the issue persists.",type:"error",duration:5e3}),Va(K,{tags:{component:"CopilotFeedback",action:"handleSubmit"}})}finally{x(!1)}},$=()=>{d(),o(null),O()},U=G=>G?{transform:"scale(1.05)",color:"colorPalette.solid",opacity:1}:{transform:"scale(1)"},B=[{id:"Cancel",title:"Cancel",onClick:G=>{G.stopPropagation(),$()},"aria-label":"Cancel",variant:"outline",colorPalette:"dark"},{id:"Submit",title:"Submit",onClick:G=>{G.stopPropagation(),D().catch(K=>Va(K))},"aria-label":"Submit",variant:"solid",colorPalette:"dark",disabled:!h||k}];return H.jsxs(ai,{"data-testid":"copilot-feedback",position:"relative",w:"100%",display:"flex",justifyContent:"flex-end",children:[H.jsx(NYe,{value:r||"",display:"flex",w:"100%",justifyContent:"flex-end",children:H.jsx($c,{gap:0,position:"relative",children:Uti.map(G=>{const K=r===G.value,X=!!(r&&!K);return H.jsx(RYe,{value:G.value,asChild:!0,children:H.jsx(Gm,{"aria-label":G.label,size:"xs",variant:"subtle",colorPalette:"dark",onClick:()=>void P(G.value),_hover:r?{}:{transform:"scale(1.1)"},_active:r?{}:{transform:"scale(0.95)"},transition:"all 0.25s cubic-bezier(0.4, 0, 0.2, 1)",display:X?"none":"block",disabled:K,...U(K),children:H.jsx(xa,{icon:G.icon})})},G.value)})})}),H.jsx(K4,{open:s,onClose:$,onOpen:c,title:"Help us improve",displayOverlay:!0,size:"sm",actions:B,children:H.jsx(Bti,{selectedIssue:h,otherText:y,setSelectedIssue:b,setOtherText:E})})]})},Hti=({links:n})=>{if(!n||n.length===0)return null;const r=n.map((o,s)=>H.jsx(CYe,{children:H.jsx(_I,{href:o.url,variant:"underline",fontSize:QE,color:"rasawebDeepPurple.800",lineHeight:aU,children:o.label})},`${o.url}-${s}`));return H.jsx(ai,{pt:"1rem",children:H.jsx(Xve,{children:r})})};function Gti({onClick:n,hide:r=!1,tooltipText:o}){return r?null:H.jsx(Wg,{positioning:{placement:"top"},content:o??"See what's changed",bgColor:"rasawebDeepPurple.800",showArrow:!0,children:H.jsx(Gm,{"aria-label":"View diff",variant:"ghost",size:"sm",onClick:s=>{s.stopPropagation(),n?.()},borderRadius:"8px",_hover:{bg:"rasaNeutral.300"},children:H.jsx(xa,{icon:NK,color:"rasaNeutral.700",size:"sm"})})})}function qti({onClick:n,hide:r=!1}){return r?null:H.jsx(Wg,{positioning:{placement:"top"},content:"Show error log",bgColor:"rasawebDeepPurple.800",showArrow:!0,children:H.jsx(Gm,{"aria-label":"Show error log",variant:"ghost",size:"sm",onClick:o=>{o.stopPropagation(),n?.()},borderRadius:"8px",_hover:{bg:"rasaNeutral.300"},children:H.jsx(xa,{icon:$or,color:"rasaNeutral.700",size:"sm"})})})}function Vti({onClick:n,hide:r=!1}){return r?null:H.jsx(Wg,{positioning:{placement:"top"},content:"Restore this version",bgColor:"rasawebDeepPurple.800",showArrow:!0,children:H.jsx(Gm,{"aria-label":"Restore this version",variant:"ghost",size:"sm",onClick:o=>{o.stopPropagation(),n?.()},borderRadius:"8px",_hover:{bg:"rasaNeutral.300"},children:H.jsx(xa,{icon:Mor,color:"rasaNeutral.700",size:"sm"})})})}var Vp=(n=>(n.Latest="Latest",n.Successful="Successful",n.Error="Error",n.OldError="OldError",n))(Vp||{});function Wti({version:n}){const r=Kti[n];return H.jsxs($c,{gap:2,children:[H.jsx(ai,{width:"8px",height:"8px",borderRadius:"full",bg:r.dotColor,flexShrink:0}),H.jsx(Kx,{size:"sm",color:r.textColor,children:r.text})]})}const Kti={[Vp.Latest]:{dotColor:"rasaGreen.800",text:"Live",textColor:"rasaGreen.800"},[Vp.Successful]:{dotColor:"rasaNeutral.700",text:"Old version",textColor:"rasaNeutral.700"},[Vp.Error]:{dotColor:"rasaRed.800",text:"Chat unavailable",textColor:"rasaRed.800"},[Vp.OldError]:{dotColor:"rasaNeutral.700",text:"Old version",textColor:"rasaNeutral.700"}};function Yti({title:n,version:r,isFirstMessageWithCommit:o=!1,onUndo:s,onDiffView:c,onWarning:d,onNavigateToAgent:h}){const b=o?"See what's new":"See what's changed",y=(r===Vp.Successful||r===Vp.OldError||r===Vp.Error)&&s,E=r!==Vp.OldError,k=(r===Vp.Error||r===Vp.OldError)&&d;n=Xti(r,n);const x=r!==Vp.Latest,O={position:"relative",width:"100%",height:"62px",bg:"rasaNeutral.50",borderRadius:"0.75rem",border:"1px solid",borderColor:"rasaNeutral.400",px:"1rem",pt:"12px",pb:"11px",transition:"all 0.2s"};return H.jsx(Wg,{positioning:{placement:"bottom"},content:"Try this version",bgColor:"rasawebDeepPurple.800",showArrow:!0,disabled:x,children:H.jsx(ai,{role:x?void 0:"button","aria-label":x?void 0:"Try this version",onClick:x?void 0:h??void 0,_hover:Jti(x),...O,children:H.jsxs(Iu,{justify:"space-between",align:"center",height:"100%",children:[H.jsxs(Iu,{direction:"column",gap:.5,flex:1,minWidth:0,children:[H.jsx(Kx,{size:"md",overflow:"hidden",textOverflow:"ellipsis",whiteSpace:"nowrap",truncate:!0,children:n}),H.jsx(Wti,{version:r})]}),H.jsxs($c,{gap:0,flexShrink:0,children:[H.jsx(Vti,{hide:!y,onClick:s}),H.jsx(qti,{hide:!k,onClick:d}),H.jsx(Gti,{hide:!E,onClick:c,tooltipText:b})]})]})})})}function Xti(n,r){if(r)return r;switch(n){case Vp.Latest:case Vp.Successful:return"Update files";case Vp.OldError:case Vp.Error:return"Failed to update files";default:return"Unknown version"}}const Jti=n=>({bg:"rasaNeutral.100",cursor:n?"default":"pointer"}),Zti=({isLastMessageWithCommit:n,isFirstMessageWithCommit:r,isSuccessfulTraining:o,commit:s,error:c})=>{const{open:d,onOpen:h,onClose:b}=_M(),y=n?void 0:h,{open:E,onOpen:k,onClose:x}=_M(),{getFiles:O}=FEe(),{setOpenFile:P,setVersion:D}=bU(),{getPath:$}=MK(),U=Io(da,we=>we.projectUrl),B=Io(bEe,we=>we.loading),G=y2(),K=Q.useCallback(()=>{G($("chat"),{replace:!0})},[G,$]),X=Q.useCallback(()=>{if(!s?.sha){Kp.create({title:"Error",description:"Failed to navigate to diff",type:"error",duration:4e3}),Va(new Error("Commit SHA is missing"),{tags:{component:"CopilotButtonActions",action:"navigateToDiff"}});return}G($(`code/diff/${s.sha}`),{replace:!0})},[G,$,s?.sha]),ae=Q.useMemo(()=>Qti(n,o),[n,o]),oe=async()=>{await O(),P(void 0),D(we=>we+1)},he=[{id:"Cancel",title:"Cancel",onClick:()=>b(),"aria-label":"Cancel",variant:"outline",colorPalette:"dark",disabled:B},{id:"Confirm",title:"Yes, restore version",onClick:async()=>{if(!s?.sha||!U){b();return}try{const we=await khr({projectUrl:U,commit_sha:s?.sha});T0e(!0),fPr({jobId:we.job_id,projectUrl:U,onSuccess:()=>{T0e(!1),b()},onError:ge=>{Va(ge,{tags:{component:"CopilotButtonActions",action:"createCommitEventSource"}}),T0e(!1),b()},onFirstEvent:()=>void oe()})}catch(we){Va(we,{tags:{component:"CopilotButtonActions",action:"postRollback"}}),T0e(!1),b()}},"aria-label":"Confirm",variant:"solid",colorPalette:"dark",loading:B}],ye=[{id:"Cancel",title:"Cancel",onClick:()=>x(),"aria-label":"Cancel",variant:"outline",colorPalette:"dark"}];return H.jsxs(H.Fragment,{children:[H.jsx(ai,{width:"100%",p:"0",mx:"0",mt:"0.5rem",mb:"1.5rem","data-testid":"copilot-version-button-container",children:H.jsx(Yti,{title:s?.title,isFirstMessageWithCommit:r,version:ae,onNavigateToAgent:K,onDiffView:X,onWarning:c?k:void 0,onUndo:s?y:void 0})}),s&&H.jsx(K4,{canBeClosed:!1,open:d,onOpen:h,onClose:b,title:"Restore previous version",actions:he||[],children:H.jsxs(Fi,{size:"md",children:["Restoring ",H.jsx("b",{children:s.title})," will discard all unsaved or staged changes. Are you sure you want to continue?"]})}),c&&H.jsxs(K4,{open:E,onOpen:k,onClose:x,title:c.title,size:"lg",actions:ye||[],children:[c.description,c.codeSample?H.jsx(FW,{language:"error",code:c.codeSample}):null]})]})};function Qti(n,r){return n?r?Vp.Latest:Vp.Error:r?Vp.Successful:Vp.OldError}const eni=({commit:n,logs:r,isLastMessageWithCommit:o,isFirstMessageWithCommit:s})=>{const c=n.training_success===!0,d=r?{title:"Error log",description:"This is the error output from the framework. Get more guidance from Copilot by pasting a specific line into the chat.",codeSample:r}:void 0,h={title:n.message,sha:n.sha};return H.jsx(Zti,{isLastMessageWithCommit:o,isFirstMessageWithCommit:s,isSuccessfulTraining:c,commit:h,error:d})},tni=({status:n})=>{switch(n){case"completed":return H.jsx(ai,{flexShrink:0,w:"1rem",h:"1rem",display:"flex",alignItems:"center",children:H.jsx(xa,{icon:sXe,color:"rasawebDeepPurple.800",style:{width:"1rem",height:"1rem"}})});case"in_progress":return H.jsx(ai,{flexShrink:0,w:"1rem",h:"1rem",display:"flex",alignItems:"center",children:H.jsx(kK,{size:"sm",color:"rasawebDeepPurple.800",css:{width:"1rem",height:"1rem",borderWidth:"0.125rem"}})});case"cancelled":case"pending":default:return H.jsx(ai,{flexShrink:0,w:"1rem",h:"1rem",borderRadius:"full",border:"0.125rem solid",borderColor:"rasaNeutral.400",bg:"transparent"})}},nni=({task:n})=>{const r=n.status==="cancelled";return H.jsxs($c,{gap:"0.5rem",alignItems:"flex-start",opacity:r?.5:1,children:[H.jsx(ai,{pt:"0.125rem",children:H.jsx(tni,{status:n.status})}),H.jsx(Fi,{fontSize:QE,color:"rasawebDeepPurple.800",lineHeight:aU,letterSpacing:"0.025rem",children:n.content})]})},rni=({tasks:n})=>{if(!n||n.length===0)return null;const r=n.some(o=>o.status==="in_progress");return H.jsxs(Vd,{align:"stretch",gap:"0.5rem",w:"100%",mb:"1rem",children:[n.map(o=>H.jsx(nni,{task:o},o.id)),r&&H.jsx(ai,{mt:"0.5rem",children:H.jsx(Fdn,{size:"sm",colorPalette:"copilot",children:"Implementing the changes..."})})]})},mKt=({message:n,isStreaming:r=!1,isLastMessageWithCommit:o=!1,isFirstMessageWithCommit:s=!1})=>{const c=n.role===kI.User,d=n.role===kI.Internal;if(c)return H.jsx(oni,{message:n});if(d)return null;const h=n.content.find(x=>x.type==="references"),b=n.content.filter(x=>x.type==="link"),y=n.content.find(x=>x.type==="plan"),E=ani(n),k=sni(n);return H.jsxs(Vd,{alignSelf:"start",maxW:"100%",w:"100%",gap:"1rem",alignItems:"start","data-testid":"copilot-response",children:[H.jsx(ini,{message:n}),H.jsxs(ai,{width:"100%",maxW:"100%",overflow:"hidden","data-testid":"copilot-response-content",children:[y&&H.jsx(rni,{tasks:y.tasks}),E&&H.jsx(eni,{commit:E,logs:k,isLastMessageWithCommit:o,isFirstMessageWithCommit:s}),n.content.filter(x=>x.type==="text").map((x,O)=>H.jsx(Mti,{content:x.text},`text-${O}-${x.text.slice(0,20)}`)),b?H.jsx(Hti,{links:b}):null,h?H.jsx(Fti,{references:h.references}):null]}),!r&&H.jsx(zti,{messageId:n.timestamp.getTime().toString()})]})},ini=({message:n})=>{const r=H.jsx(NEe,{name:"Copilot",bg:"transparent",shape:"square",size:"2xs",src:"/rasa_logo_square_white.svg",objectFit:"fill"});return H.jsxs($c,{alignSelf:"start",gap:"0.5rem",w:"100%",children:[r,H.jsx(Kx,{fontSize:"sm",children:"Copilot"}),H.jsx(Pen,{}),H.jsx(Fi,{fontSize:Wwe,color:"rasaNeutral.700",children:rsn(n.timestamp)})]})},oni=({message:n})=>{const o=n.content.filter(c=>c.type==="event").map(RIr),s=nsn(o);return H.jsxs(Vd,{alignSelf:"end",maxW:"100%",justifyItems:"flex-end","data-testid":"copilot-user-message",children:[H.jsx(ai,{maxW:"100%",overflow:"hidden",alignSelf:"flex-end",children:s.length>0&&H.jsx($c,{gap:"0.25rem",flexWrap:"wrap",justifyContent:"flex-end",alignItems:"flex-end",children:s.map(c=>H.jsx(Bdn,{attachment:c},c.id))})}),H.jsx(ai,{bgColor:"rasaNeutral.200",py:"0.625rem",px:"1.25rem",borderTopEndRadius:"0.25rem",borderBottomEndRadius:"1.25rem",borderTopStartRadius:"1.25rem",borderBottomStartRadius:"1.25rem",w:"fit-content",alignSelf:"flex-end",children:n.content.filter(c=>c.type==="text").map((c,d)=>H.jsx(Fi,{fontSize:QE,color:"rasawebDeepPurple.800",lineHeight:aU,whiteSpace:"pre-wrap",children:c.text},`text-${d}-${c.text.slice(0,20)}`))})]})};function ani(n){const r=n.content.filter(o=>o.type==="commit");if(r.length!==0){if(r.length>1){Va("Multiple commit contents found in message",{tags:{component:"CopilotChatMessage",action:"getCommitFromMessage"},extra:{message:JSON.stringify(n)}});return}return r[0].commit}}function sni(n){const r=n.content.filter(o=>o.type==="logs");if(!(r.length===0||r[0].logs.length===0))return r.length>1&&Va("Multiple logs contents found in message",{tags:{component:"CopilotChatMessage",action:"getLogsFromMessage"},extra:{message:JSON.stringify(n)}}),r[0].logs.length>1&&Va("Multiple logs found in message",{tags:{component:"CopilotChatMessage",action:"getLogsFromMessage"},extra:{message:JSON.stringify(n)}}),r[0].logs[0].content}const cni=600,uni=n=>{const r=n.current;if(!r)return!0;const{scrollTop:o,scrollHeight:s,clientHeight:c}=r;return s-o-c<50},bKt=n=>{const r=n.current;if(!r)return cni;const{clientHeight:o}=r;return o<300?Math.round(o/2):Math.round(o-300)};function lni(n){return n.role!==kI.Copilot?!1:n.content.some(r=>r.type==="commit")}const dni=()=>{const n=Io(il,ge=>ge.isLoading),r=Io(il,ge=>ge.streamingMessage),o=Io(il,ge=>ge.messages),s=Io(il,ge=>!!ge.shouldScrollOnStreamStart),c=Q.useRef(null),d=Io(da,ge=>ge.loading),h=Io(da,ge=>ge.loadingProject),b=Io(da,ge=>ge.isRestoring),y=Io(gc,ge=>ge.trainingLoading),E=Q.useRef(null),k=Q.useRef(null),x=Q.useRef(null),O=Q.useRef(null),[P,D]=Q.useState(!0),[$,U]=Q.useState(bKt(k)),B=Q.useCallback(()=>{const ge=uni(k);D(ge)},[]),G=Q.useCallback(()=>{const ge=bKt(k);U(ge)},[]),K=Q.useMemo(()=>{const ge={height:"0px",transition:"height 0.3s ease-in-out"};return r?ge.height=`${$}px`:(ge.height="0px",ge.transition="none"),ge},[r,$]),X=Q.useCallback(()=>{x.current&&clearTimeout(x.current),x.current=setTimeout(()=>{B(),G(),x.current=null},100)},[B,G]),ae=()=>{E.current?.scrollIntoView({behavior:"smooth"})};Q.useEffect(()=>{o.length>0&&o.at(-1)?.role===kI.User&&(c.current&&clearTimeout(c.current),c.current=setTimeout(()=>{ae(),c.current=null},100))},[o]),Q.useEffect(()=>{B(),G()},[o,r,B,G]),Q.useEffect(()=>{const ge=r?.timestamp??null,Te=ge!==O.current;r&&Te&&s&&(vWe(!1),c.current&&clearTimeout(c.current),c.current=setTimeout(()=>{ae(),c.current=null},450)),O.current=ge},[r,s]),Q.useEffect(()=>()=>{x.current&&clearTimeout(x.current),c.current&&clearTimeout(c.current)},[]);let oe="Thinking...";h?oe="Building your project...":b?oe="Restoring your project...":y?oe="Saving the changes...":d&&(oe="Loading...");const he=Q.useMemo(()=>o.findLastIndex(ge=>ge.role===kI.Copilot&&ge.content.some(Te=>Te.type==="commit")),[o]),ye=Q.useMemo(()=>o.findIndex(ge=>ge.role===kI.Copilot&&ge.content.some(Te=>Te.type==="commit")),[o]),we=Q.useMemo(()=>(r&&lni(r))??!1,[r]);return H.jsxs(ai,{position:"relative",w:"100%",h:"100%",children:[H.jsxs(Ien,{w:"100%",h:"100%",overflow:"hidden",variant:"hover",children:[H.jsx(Oen,{w:"100%",h:"100%",ref:k,onScroll:X,children:H.jsxs(Vd,{flex:1,p:"1.5rem",gap:"2rem",align:"stretch",w:"100%",minW:"0","data-testid":"copilot-chat",children:[o.map((ge,Te)=>H.jsx(mKt,{message:ge,isStreaming:!1,isFirstMessageWithCommit:ye===Te,isLastMessageWithCommit:he===Te&&!we},`${Te}-${ge.content.length}`)),r&&H.jsx(mKt,{message:r,isStreaming:!0,isFirstMessageWithCommit:ye===-1&&we,isLastMessageWithCommit:we},`${r.content.length}`),(n||h||b||y)&&!r&&H.jsxs(Vd,{alignSelf:"start",maxW:"100%",gap:"1rem",alignItems:"start",children:[H.jsxs($c,{alignSelf:"start",gap:"0.5rem",children:[H.jsx(NEe,{name:"Copilot",bg:"transparent",shape:"square",size:"2xs",src:"/rasa_logo_square_white.svg",objectFit:"fill"}),H.jsx(Fi,{fontSize:"sm",fontWeight:"bold",color:"rasawebDeepPurple.800",children:"Copilot"})]}),H.jsx(Fdn,{size:"sm",children:oe})]}),H.jsx(ai,{height:K.height,m:"0",p:"0",transition:K.transition,overflow:"hidden"}),H.jsx("div",{ref:E})]})}),H.jsx(Nen,{orientation:"vertical",children:H.jsx(PYe,{})})]}),!P&&H.jsx(ai,{position:"absolute",bottom:"1rem",left:"50%",transform:"translateX(-50%)",zIndex:"10",children:H.jsx(Gm,{size:"xs",onClick:()=>{ae()},variant:"solid",colorPalette:"dark",rounded:"full",children:H.jsx(xa,{icon:Nor,size:"xs"})})})]})},fni=()=>{const[n,r]=Q.useState(""),o=Q.useRef(null),{getIdTokenClaims:s}=ty(),{getFiles:c}=FEe(),d=Io(il,U=>U.attachments),h=Io(il,U=>U.isLoading),b=Io(il,U=>U.streamingMessage),y=Io(gc,PK),E=Io(da,U=>U.loading||U.loadingProject||U.inactive||U.isRestoring);Q.useEffect(()=>{!E&&!h&&!y&&o.current&&o.current.focus()},[E,h,y]);const k=()=>{c().catch(U=>{const B=U instanceof Error?U.message:"Failed to refetch files after commit";hie(B)})},x=U=>{U.key==="Enter"&&!U.shiftKey&&!b&&(U.preventDefault(),O())},O=async()=>{const U=n.trim();if(U){r(""),o.current&&(o.current.style.height="auto");try{const G=(await s())?.__raw;await NFt({message:U,token:G,onCommit:k})}catch{await NFt({message:U,onCommit:k})}z_(H_.userMessageSent,{target:"copilot"})}},P=()=>{o.current&&(o.current.style.height="auto")},D=U=>{U.style.height=`${Math.min(U.scrollHeight,240)}px`},$={color:"rasawebDeepPurple.800",border:"none",resize:"none",p:"0",overflowY:"auto",minH:"1lh",maxH:"15rem",height:"auto",lineHeight:`${aU}`,_focus:{boxShadow:"none",outline:"none"},"&::placeholder":{color:"rasaNeutral.700"}};return H.jsxs(ai,{borderTop:"1px solid",borderColor:"rasaNeutral.400",p:"1.5rem","data-testid":"copilot-input",children:[d.length>0?H.jsx($c,{gap:"0.25rem",mb:"1rem",flexWrap:"wrap",children:d.map(U=>H.jsx(Bdn,{attachment:U,removable:!0},U.id))}):null,H.jsx(CQe,{step:Qf.Step2,children:H.jsxs($c,{gap:"1.5rem",alignItems:"center",children:[H.jsx(DYe,{ref:o,value:n,css:$,rows:1,onChange:U=>{r(U.target.value),U.target.value===""&&P()},onKeyDown:x,placeholder:"Ask for help, plan a feature, or learn about Rasa...",disabled:h||y||E,onInput:U=>{D(U.target)}}),H.jsx(Gm,{"aria-label":"Send",borderRadius:"full",size:"2xs",colorPalette:"dark",onClick:()=>void O(),loading:h,disabled:n.length===0||h||y||E||!!b,children:H.jsx(xa,{icon:Hor,color:"white"})})]})})]})},jhn=({text:n,...r})=>{const o={pointerEvents:"none"};return H.jsx($c,{...r,children:H.jsxs(_I,{href:"/",textDecoration:"none",children:[H.jsx(Iye,{src:"/rasa_horizontal_logo_white.svg",alt:"logo",h:"2rem"}),H.jsx(Fi,{size:"sm",css:o,children:n??"Hello Rasa"})]})})},hni=()=>{const{showRasaLogo:n}=GM(),r={flexBasis:`${Scn}rem`,flexShrink:0,width:"100%",borderBottom:"1px solid",borderColor:"rasaNeutral.400"};return H.jsx($c,{alignItems:"center",justifyContent:"space-between",gap:2,px:4,css:r,children:n&&H.jsxs(H.Fragment,{children:[H.jsx(jhn,{}),H.jsx(Men,{colorPalette:"purple",children:H.jsx(jen,{fontSize:"13px",children:"Beta"})})]})})},pni=n=>{const r=Roe(),{projectUrl:o,projectId:s,projectInactive:c}=Io(da,b=>({projectUrl:b.projectUrl,projectId:b.projectId,projectInactive:b.inactive})),{data:d,error:h}=xK({queryKey:["CopilotMode",o,s],queryFn:()=>phr({projectUrl:o||""}),enabled:!!o&&!!s&&!c});h&&console.error(`Error fetching copilot mode for project ${s}:`,h),Q.useEffect(()=>{d&&d.mode!==n&&o&&ghr({projectUrl:o,mode:n}).then(()=>r.invalidateQueries({queryKey:["CopilotMode",o,s]})).then(()=>console.log(`🤖 Copilot mode: ${n}`)).catch(b=>{console.error(`Failed to set copilot mode for project ${s}:`,b)})},[d,n,o,s,r])},gni=()=>{const{isStandalone:n,isStudioEmbed:r}=GM(),o=AEe("useAgentSdk",!1)[0];return pni(n||r||o?"agent_sdk":"legacy"),FEe(),H.jsxs(Vd,{gap:0,height:"100%",overflow:"auto",width:"100%",children:[H.jsx(hni,{}),H.jsxs(Iu,{direction:"column",bg:"white",boxShadow:"lg",flexGrow:1,overflow:"auto",width:"100%","data-testid":"copilot-chat-panel",children:[H.jsx(Vd,{flex:1,overflowY:"auto",align:"stretch",children:H.jsx(dni,{})}),H.jsx(fni,{})]})]})},mni=()=>{const{getIdTokenClaims:n}=ty(),{sessionId:r,messages:o}=Io(il,E=>({sessionId:E.sessionId,messages:E.messages})),{projectUrl:s,projectLoading:c,jobId:d}=Io(da,E=>({projectUrl:E.projectUrl,projectLoading:E.loading,jobId:E.jobId})),{data:h,isLoading:b,error:y}=xK({queryKey:["copilotHistory",s,r],queryFn:async()=>{const k=(await n())?.__raw;return s?(await Thr({projectUrl:s,token:k})).map(O=>({...O,timestamp:new Date(Number.parseFloat(O.timestamp)*1e3),content:PIr(O.content)})):[]},enabled:!!r&&!!s&&!c&&!d&&o.length===0,refetchOnWindowFocus:!1,retry:3});return Q.useEffect(()=>{h&&il.setState(E=>({...E,messages:h}))},[h]),{isLoading:b,error:y,historyData:h}},bni=({panels:n,initialLeft:r=h5r,minLeft:o=p5r,maxLeft:s=g5r})=>{const[c,d]=Q.useState(r),[h,b]=Q.useState(!1),y=()=>{b(!0)},E=()=>{b(!1)},k=Q.useCallback(O=>{if(!h)return;const P=Math.min(Math.max(O.clientX,o),s);d(P)},[o,s,h]);Q.useEffect(()=>{if(h)return globalThis.addEventListener("mousemove",k),globalThis.addEventListener("mouseup",E),()=>{globalThis.removeEventListener("mousemove",k),globalThis.removeEventListener("mouseup",E)}},[h,k]);const x={width:"9px",cursor:"col-resize",display:"flex",justifyContent:"center",marginLeft:"-4px",marginRight:"-4px",zIndex:"1","& > div":{width:"1px",bg:"rasaNeutral.400"},"&:hover > div":{bgImage:"linear-gradient({colors.rasaNeutral.400}, {colors.rasawebPurple.800}, {colors.rasaNeutral.400})"}};return H.jsxs($c,{width:"100%",height:"100%",alignItems:"stretch",gap:0,userSelect:h?"none":"auto","data-testid":"split-layout",children:[H.jsx(ai,{width:`${c}px`,minWidth:`${o}px`,maxWidth:`${s}px`,children:n.left}),H.jsx(ai,{css:x,onMouseDown:y,"data-testid":"resize-gutter",children:H.jsx(ai,{})}),H.jsx(ai,{flex:"1",minWidth:"0",children:n.right})]})};function Cqe(){const n=y2(),{projectId:r}=G_(),[o]=Dye(),{projectUrl:s,loading:c}=Io(da,$=>({projectUrl:$.projectUrl,loading:$.loading})),{user:d}=ty(),{data:h}=IEe(),{updateOnboardingStatus:b,initializeIfNeeded:y,onboardingState:E}=qoe(),{showCopilot:k,showProjectActions:x,isCloud:O}=GM(),P=o.get("forceProjectUrl");Q.useEffect(()=>{P&&Icn(P)},[P]),mni();const D=Q.useCallback($=>{y(),Ocn($)},[y]);return Q.useEffect(()=>{if(!O||P){y();return}r&&!s&&D(r)},[r,D,n,s,y,O,P]),Q.useEffect(()=>{s&&!c&&(E.step===Qf.Loading||E.step===null)&&(b(Qf.Step1),GZe(s))},[s,c]),Q.useEffect(()=>{Rcn("Home")},[]),Q.useEffect(()=>{Pcn(d)},[d]),H.jsxs(Cln,{children:[k?H.jsx(bni,{panels:{left:H.jsx(gni,{}),right:H.jsx(rWt,{})}}):H.jsx(rWt,{}),x&&H.jsx(mQe,{isRequired:h?.auth_required===!0})]})}const vni=()=>{const n=Io(da,o=>o.assistantName),r={position:"sticky",borderBottom:"1px solid",borderColor:"rasaNeutral.400",height:"3.25rem",maxW:"100vw",px:"4",justifyContent:"space-between"};return H.jsxs($c,{css:r,children:[H.jsx(jhn,{text:n||""}),H.jsx(ud,{size:"sm",asChild:!0,children:H.jsxs(_I,{target:"_blank",colorPalette:"dark",textDecoration:"none",href:globalThis.location.origin,children:["Build your own agent",H.jsx(xa,{icon:Uye})]})})]})},wni=({minimal:n,children:r})=>H.jsxs(Cye,{height:"100vh",overflowX:"hidden",gap:0,children:[n?H.jsx(vni,{}):H.jsx(pdn,{}),H.jsx(sQn,{flex:1,minH:0,maxH:"100vh",maxW:"100vw",px:0,display:"flex",flexDirection:"column",children:r})]}),yni=()=>H.jsx(Bx,{shortLabel:"Unable to connect to the agent",longLabel:"Something went wrong and we couldn’t connect. Try refreshing the page — if the issue continues, the link may be outdated or the agent unavailable.",image:zm.CubesError,children:H.jsx(Vd,{children:H.jsx(ud,{onClick:()=>globalThis.location.reload(),variant:"solid",colorPalette:"dark",mb:"0.5rem",children:"Refresh page"})})}),Eni=()=>{zln();const{inspectEnabled:n}=bU(),{error:r,inactive:o,isRestoring:s}=Io(da,X=>({error:X.error,inactive:X.inactive,isRestoring:X.isRestoring})),{sendMessage:c,setUrl:d,startNewConversation:h,conversationList:b,inputDisabled:y,sessionId:E,stack:k,slots:x,flowHistory:O,replayingConversation:P,waitingForUserInput:D,replayConversation:$,cleanup:U}=Uln({useMemoryOnly:!0});Q.useEffect(()=>{o&&U()},[o,U]);const B={background:"url(/bg.svg)",backgroundSize:"cover",height:"100%",alignItems:"center"},G={width:"66%",height:"calc(100% - 4rem)",minWidth:"618px",margin:"0 auto",borderColor:"gray.200",bg:"rasaNeutral.50",borderRadius:"1rem",justifyContent:"center"},K=Q.useCallback(()=>s?H.jsx(gdn,{}):o?H.jsx(mdn,{}):r?H.jsx(yni,{}):H.jsx(Tdn,{flowView:n,sendMessage:c,setUrl:d,startNewConversation:h,conversationList:b,inputDisabled:y,sessionId:E,stack:k,slots:x,flowHistory:O,replayingConversation:P,waitingForUserInput:D,replayConversation:$}),[o,r,s,n,c,d,h,b,y,E,k,x,O,P,D,$]);return H.jsx(Iu,{css:B,children:H.jsx(Iu,{css:G,children:K()})})},Sni=()=>{const{projectId:n}=G_();return Q.useEffect(()=>(PFt(!1),Rcn("Shared Chat"),()=>{PFt(!0)}),[]),Q.useEffect(()=>{n&&Ocn(n)},[n]),H.jsx(Cln,{children:H.jsx(wni,{minimal:!0,children:H.jsx(Eni,{})})})},_ni=new Yer;function kni(){return H.jsxs(inr,{children:[H.jsx(SB,{path:"/",element:H.jsx(K7t,{to:"/chat",replace:!0})}),H.jsx(SB,{path:"/:view",element:H.jsx(Cqe,{})}),H.jsx(SB,{path:"/:view/diff/:commitSha",element:H.jsx(Cqe,{})}),H.jsx(SB,{path:"/:view/:pathToFile",element:H.jsx(Cqe,{})}),H.jsx(SB,{path:"/agent/:projectId",element:H.jsx(Sni,{})}),H.jsx(SB,{path:"*",element:H.jsx(K7t,{to:"/chat",replace:!0})})]})}function Tni(){return wMr(),H.jsxs(etr,{client:_ni,children:[H.jsxs(hFn,{value:Edn,children:[H.jsx(Anr,{children:H.jsx(pPr,{children:H.jsx(kni,{})})}),H.jsx(Gor,{})]}),H.jsx(ptr,{initialIsOpen:!1})]})}const xni="login.hello.rasa.ai",Cni="Gq5RdRwp174OFIfTz6Re9TZUseYDXUYE",Ani="development";console.log(`🚀 Hello Rasa App Version: ${Ani}`);console.log("🛠️ Build Mode: production");Ojn.createRoot(document.getElementById("root")).render(H.jsx(Q.StrictMode,{children:H.jsx(f$n,{domain:xni,clientId:Cni,authorizationParams:{redirect_uri:gQe()},children:H.jsx(YDr,{children:H.jsx(Tni,{})})})}));export{eK as C,eZe as P,wo as _,yo as a,Vg as b,zr as c,Van as d,gB as e,Uni as f,zni as g,Fx as h,Xie as i,wan as j,iZe as k,R$e as l,Xdr as m,lie as n,$4 as o,Fdr as p,$lr as q,f0e as r,Ix as s,Elr as t,Hni as u,Jdr as v,Y4 as w};
